Package: np
Version: 0.60-20
Date: 2026-02-11
Imports: boot, cubature, methods, quadprog, quantreg, stats
Suggests: crs, MASS, logspline, ks, testthat
Title: Nonparametric Kernel Smoothing Methods for Mixed Data Types
Authors@R: c(person(given = "Jeffrey S.", family = "Racine", role = c("aut","cre"), email = "racinej@mcmaster.ca"),
             person(given = "Tristen", family = "Hayfield", role = c("aut"), email = "tristen.hayfield@gmail.com"))
Maintainer: Jeffrey S. Racine <racinej@mcmaster.ca>
Description: Nonparametric (and semiparametric) kernel methods that seamlessly handle a mix of continuous, unordered, and ordered factor data types. We would like to gratefully acknowledge support from the Natural Sciences and Engineering Research Council of Canada (NSERC, <https://www.nserc-crsng.gc.ca/>), the Social Sciences and Humanities Research Council of Canada (SSHRC, <https://www.sshrc-crsh.gc.ca/>), and the Shared Hierarchical Academic Research Computing Network (SHARCNET, <https://sharcnet.ca/>). We would also like to acknowledge the contributions of the GNU GSL authors. In particular, we adapt the GNU GSL B-spline routine gsl_bspline.c adding automated support for quantile knots (in addition to uniform knots), providing missing functionality for derivatives, and for extending the splines beyond their endpoints.
License: GPL
URL: https://github.com/JeffreyRacine/R-Package-np
BugReports: https://github.com/JeffreyRacine/R-Package-np/issues
Repository: CRAN
NeedsCompilation: yes
Packaged: 2026-02-11 21:12:26 UTC; jracine
Author: Jeffrey S. Racine [aut, cre],
  Tristen Hayfield [aut]
Date/Publication: 2026-02-12 06:10:12 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 05:43:45 UTC; unix
Archs: np.so.dSYM

useDynLib(np, .registration = TRUE)
import(stats, boot, cubature)
importFrom("grDevices", "boxplot.stats", "extendrange", "n2mfrow", "grey", "rgb")
importFrom("graphics", "axis", "lines", "mtext", "par", "persp", "plot", "points", "plot.new", "plot.window", "polygon")
importFrom("utils", "combn", "flush.console")
importFrom("methods", "is")
importFrom("quantreg", "rq")
importFrom("quadprog", "solve.QP")
export(b.star
       ,npuniden.boundary
       ,npuniden.reflect
       ,npuniden.sc
       ,npudensbw
       ,npudistbw
       ,npudens
       ,npudist
       ,npregbw
       ,npreg
       ,npregiv
       ,npregivderiv              
       ,npplreg
       ,npplregbw
       ,npcdensbw
       ,npcdens
       ,npcdist
       ,npcdistbw
       ,npcopula
       ,npdeptest       
       ,npsdeptest
       ,npplot
       ,npksum
       ,npqreg
       ,npquantile
       ,npcmstest
       ,npconmode
       ,npdeneqtest
       ,npindex
       ,npindexbw
       ,npqcmstest       
       ,npseed
       ,npscoef
       ,npscoefbw
       ,npsigtest
       ,npsymtest
       ,npunitest       
       ,uocquantile
       ,se
       ,gradients
       ,nptgauss
       )

S3method(print,bandwidth)
S3method(print,dbandwidth)
S3method(print,npdensity)
S3method(print,npdistribution)
S3method(print,rbandwidth)
S3method(print,npregression)
S3method(print,plbandwidth)
S3method(print,plregression)
S3method(print,conbandwidth)
S3method(print,condbandwidth)
S3method(print,condensity)
S3method(print,condistribution)
S3method(print,npkernelsum)
S3method(print,qregression)
S3method(print, cmstest)
S3method(print, npregiv)
S3method(print, npregivderiv)
S3method(print, conmode)
S3method(print,deptest)
S3method(print,sdeptest)
S3method(print,sibandwidth)
S3method(print,singleindex)
S3method(print,sigtest)
S3method(print,scbandwidth)
S3method(print,smoothcoefficient)
S3method(print,symtest)
S3method(print,deneqtest)
S3method(print,unitest)
##


S3method(plot, bandwidth)
S3method(plot, dbandwidth)
S3method(plot, rbandwidth)
S3method(plot, conbandwidth)
S3method(plot, condbandwidth)
S3method(plot, plbandwidth)
S3method(plot, sibandwidth)
S3method(plot, scbandwidth)

S3method(plot, npdensity)
S3method(plot, npdistribution)
S3method(plot, npregression)
S3method(plot, npregiv)
S3method(plot, npregivderiv)
S3method(plot, plregression)
S3method(plot, condensity)
S3method(plot, condistribution)
S3method(plot, qregression)
S3method(plot, singleindex)
S3method(plot, smoothcoefficient)

S3method(fitted, npdensity)
S3method(fitted, npdistribution)
S3method(fitted, npregression)
S3method(fitted, plregression)
S3method(fitted, condensity)
S3method(fitted, condistribution)
S3method(fitted, conmode)
S3method(fitted, qregression)
S3method(fitted, singleindex)
S3method(fitted, smoothcoefficient)

S3method(residuals, npregression)
S3method(residuals, plregression)
S3method(residuals, singleindex)
S3method(residuals, smoothcoefficient)

S3method(coef, plregression)
S3method(coef, singleindex)
S3method(coef, sibandwidth)
S3method(coef, smoothcoefficient)

S3method(vcov, plregression)
S3method(vcov, singleindex)

S3method(predict, npregression)
S3method(predict, bandwidth)
S3method(predict, dbandwidth)
S3method(predict, conbandwidth)
S3method(predict, condbandwidth)
S3method(predict, condensity)
S3method(predict, condistribution)
S3method(predict, npdensity)
S3method(predict, npdistribution)
S3method(predict, plbandwidth)
S3method(predict, plregression)
S3method(predict, qregression)
S3method(predict, rbandwidth)
S3method(predict, scbandwidth)
S3method(predict, sibandwidth)
S3method(predict, singleindex)
S3method(predict, smoothcoefficient)

S3method(summary, bandwidth)
S3method(summary, dbandwidth)
S3method(summary, conbandwidth)
S3method(summary, condbandwidth)
S3method(summary, rbandwidth)
S3method(summary, plbandwidth)
S3method(summary, sibandwidth)
S3method(summary, scbandwidth)

S3method(summary, npdensity)
S3method(summary, npdistribution)
S3method(summary, condensity)
S3method(summary, condistribution)
S3method(summary, npregression)
S3method(summary, npregiv)
S3method(summary, npregivderiv)
S3method(summary, cmstest)
S3method(summary, conmode)
S3method(summary, deptest)
S3method(summary, sdeptest)
S3method(summary, sigtest)
S3method(summary, plregression)
S3method(summary, qregression)
S3method(summary, singleindex)
S3method(summary, smoothcoefficient)
S3method(summary, symtest)
S3method(summary, deneqtest)
S3method(summary, unitest)

S3method(se, npdensity)
S3method(se, npdistribution)
S3method(se, condensity)
S3method(se, condistribution)
S3method(se, npregression)
S3method(se, qregression)
S3method(se, singleindex)
S3method(se, smoothcoefficient)

S3method(gradients, condensity)
S3method(gradients, condistribution)
S3method(gradients, npregression)
S3method(gradients, qregression)
S3method(gradients, singleindex)

S3method(mode, conmode)

S3method(quantile, qregression)

S3method(npudensbw, NULL)
S3method(npudensbw, formula)
S3method(npudensbw, default)
S3method(npudensbw, bandwidth)

S3method(as.double,bandwidth)

S3method(npudens, formula)
S3method(npudens, call)
S3method(npudens, default)
S3method(npudens, bandwidth)

S3method(npudistbw, NULL)
S3method(npudistbw, formula)
S3method(npudistbw, default)
S3method(npudistbw, dbandwidth)

S3method(as.double, dbandwidth)

S3method(npudist, formula)
S3method(npudist, call)
S3method(npudist, default)
S3method(npudist, dbandwidth)

S3method(npregbw, NULL)
S3method(npregbw, formula)
S3method(npregbw, default)
S3method(npregbw, rbandwidth)

S3method(npreg, formula)
S3method(npreg, call)
S3method(npreg, default)
S3method(npreg, rbandwidth)

S3method(npsigtest, formula)
S3method(npsigtest, call)
S3method(npsigtest, default)
S3method(npsigtest, npregression)
S3method(npsigtest, rbandwidth)

S3method(npplreg, formula)
S3method(npplreg, call)
S3method(npplreg, default)
S3method(npplreg, plbandwidth)

S3method(npplregbw, NULL)
S3method(npplregbw, formula)
S3method(npplregbw, default)
S3method(npplregbw, plbandwidth)

S3method(npcdensbw, NULL)
S3method(npcdensbw, formula)
S3method(npcdensbw, default)
S3method(npcdensbw, conbandwidth)

S3method(npcdens, formula)
S3method(npcdens, call)
S3method(npcdens, default)
S3method(npcdens, conbandwidth)

S3method(npcdist, formula)
S3method(npcdist, default)
S3method(npcdist, call)
S3method(npcdist, condbandwidth)

S3method(npcdistbw, NULL)
S3method(npcdistbw, formula)
S3method(npcdistbw, default)
S3method(npcdistbw, condbandwidth)

S3method(npplot, rbandwidth)
S3method(npplot, bandwidth)
S3method(npplot, dbandwidth)
S3method(npplot, plbandwidth)
S3method(npplot, conbandwidth)
S3method(npplot, condbandwidth)
S3method(npplot, sibandwidth)
S3method(npplot, scbandwidth)

S3method(compute.bootstrap.errors, rbandwidth)
S3method(compute.bootstrap.errors, bandwidth)
S3method(compute.bootstrap.errors, dbandwidth)
S3method(compute.bootstrap.errors, conbandwidth)
S3method(compute.bootstrap.errors, condbandwidth)

S3method(npksum, integer)
S3method(npksum, numeric)
S3method(npksum, formula)
S3method(npksum, default)

S3method(npqreg, formula)
S3method(npqreg, call)
S3method(npqreg, default)
S3method(npqreg, condbandwidth)
S3method(npqreg, conbandwidth)

S3method(npconmode, formula)
S3method(npconmode, call)
S3method(npconmode, default)
S3method(npconmode, conbandwidth)

S3method(npindexbw, NULL)
S3method(npindexbw, formula)
S3method(npindexbw, default)
S3method(npindexbw, sibandwidth)

S3method(npscoefbw, NULL)
S3method(npscoefbw, formula)
S3method(npscoefbw, default)
S3method(npscoefbw, scbandwidth)

S3method(npindex, default)
S3method(npindex, call)
S3method(npindex, formula)
S3method(npindex, sibandwidth)

S3method(npscoef, default)
S3method(npscoef, call)
S3method(npscoef, formula)
S3method(npscoef, scbandwidth)

S3method(gsl.bs, default)
S3method(predict, gsl.bs)

# np

![CRAN downloads](https://cranlogs.r-pkg.org/badges/grand-total/np)

This is the R package `np` (Nonparametric Kernel Methods for Mixed Datatypes) written and maintained by Jeffrey S. Racine (racinej@mcmaster.ca) and co-authored by Tristen Hayfield (tristen.hayfield@gmail.com).

## Installation

You can install the stable version on [CRAN](https://cran.r-project.org/package=np):

```r
install.packages('np', dependencies = TRUE)
```

Or download the [zip ball](https://github.com/JeffreyRacine/R-Package-np/zipball/master)
or [tar ball](https://github.com/JeffreyRacine/R-Package-np/tarball/master),
decompress and run `R CMD INSTALL` on it.

Alternatively, you can install the development version but before
doing so Windows users have to first install
[Rtools](https://cran.r-project.org/bin/windows/Rtools/), while macOS
users have to first install
[Xcode](https://apps.apple.com/us/app/xcode/id497799835?mt=12) and the
command line tools (once you have Xcode
installed, open a terminal and run `xcode-select --install`). Note also
that versions of e.g. Rtools are paired with versions of R so ensure
you have the latest version of R installed prior to commencing this
process.

After installing
[Rtools](https://cran.r-project.org/bin/windows/Rtools/)/[Xcode](https://apps.apple.com/us/app/xcode/id497799835?mt=12)
and **devtools** (via install.packages("devtools")), install the
development package using the following command:

```r
library(devtools); install_github('JeffreyRacine/R-Package-np')
```

Note also that if you wish a fast install without the building of
vignettes (or if you do not have TeX installed on your system), add
the option build_vignettes=FALSE to the install_github() call.

Note that if you wish to install the MPI-enabled development version
of the package (i.e. the package `npRmpi`), you can add the option
`ref='npRmpi'` to the `install_github()` call above, presuming that your
system has a working MPI subsystem installed. See `WORKTREES.md` and
`BUILD.md` in this repo for local build details.

For more information on this project please visit the maintainer's website (https://experts.mcmaster.ca/people/racinej).

## $Id: lc_k1_mean.R,v 1.11 2013/12/01 16:07:16 jracine Exp jracine $

## We illustrate constrained kernel estimation using the local
## constant estimator where the constraints are l(x) <= \hat g(x) <=
## u(x) (see Du, P. and C. Parmeter and J.S. Racine (2013),
## "Nonparametric Kernel Regression with Multiple Predictors and
## Multiple Shape Constraints," Statistica Sinica, Volume 23, Number
## 3, 1343-1372).

rm(list=ls())

## Load required packages, set options.

library(np)
library(quadprog)
library(crs)
options(np.tree=TRUE,np.messages=FALSE)

## Set the kernel function.

ckertype <- "epanechnikov"

## Set the order of the local polynomial `p' (p=0 is the
## Nadaraya-Watson estimator, p=1 local linear, p=2 local quadratic
## etc.). If p < 0 both p and the bandwidth (bw) are determined by
## cross-validation, otherwise only bw is determined.

p <- 1

## Simulate a sample of data. We can control signal/noise ratio by
## multiplying epsilon by sd(dgp), so rnorm(n,sd=...) can be set to
## sd=(.25,.5,1,2) which would yield an R-squared for the Oracle model
## of (.95,.8,.5,and .2). Changing n changes the sample size,
## modifying cos() the data generating process (dgp).

n <- 1000
x <- sort(runif(n))
dgp <- cos(2*pi*x)
y <- dgp + sd(dgp)*rnorm(n,sd=0.25)

## Set up bounds for the quadratic program. We are going to require
## the lower and upper constraints l(x) and u(x) for g(x). Here the
## upper bound is a constant, the lower bound depends on x.

lower <- -1.5+2*x
upper <- rep(0.5,n)
if(any(lower>upper)) stop(" constraints are inconsistent")

## X (data frame of regressors) and y are passed below, so if you add
## extra regressors simply add them to X here and to the formula (used
## to compute cross-validated smoothing parameters) and be done.

X <- data.frame(x)

formula.glp <- formula(y~x)

model.glp <- crs:::npglpreg(formula=formula.glp,
                            cv=ifelse(p>=0,"bandwidth","degree-bandwidth"),
                            degree=ifelse(p>=0,rep(p,NCOL(X)),rep(0,NCOL(X))),
                            ckertype=ckertype,
                            nmulti=min(NCOL(X),5))

bw <- model.glp$bws
p <- model.glp$degree

## W is the local polynomial model matrix (includes an intercept).

W <- crs:::W.glp(xdat=X,
                 degree=rep(p,NCOL(X)))

## Generate the matrix of kernel weights using data-driven bandwidths
## that are optimal for the unconstrained model.

K <- npksum(txdat=X,
            bws=bw,
            ckertype=ckertype,
            return.kernel.weights=TRUE)$kw

## Compute the matrix for which \hat g(x|p) = t(A)%*%p (i.e. the local
## polynomial fit).

A <- sapply(1:n,function(i){W[i,,drop=FALSE]%*%chol2inv(chol(t(W)%*%(K[,i]*W)))%*%t(W)*y*K[,i]})

## Compute the unconstrained estimator.

p.u <- rep(1,n)
fit.unres <- t(A)%*%p.u

## Solve the quadratic program. The function solve.QP in the quadprog
## package solves the problem min (p-p.u)'(p-p.u) subject to the
## constraints Amat^T p >= bvec. Note that we construct Amat to
## contain the constraints a) ) A^T p >= lower, and b) -A^Tp >= -upper
## (here A^T = \hat g(x|p)).

output.QP <- solve.QP(Dmat=diag(n),
                      dvec=rep(1,n),
                      Amat=cbind(A,-A),
                      bvec=c(lower,-upper))

if(is.nan(output.QP$value)) stop(" solve.QP was unable to find a solution: adjust constraints and restart")

p.hat <- output.QP$solution

## Compute the constrained estimator.

fit.res <- t(A)%*%p.hat

## Plot the DGP, unrestricted, and restricted fits along with the
## constraints and data.

ylim <- c(min(dgp,lower,fit.unres,fit.res,y),
          max(dgp,upper,fit.unres,fit.res,y))
subtext <- paste("Local polynomial order = ", p,", bandwidth = ",formatC(bw,digits=3,format="f"),", n = ",n,sep="")
plot(x,y,cex=.25,ylab="g(x)",ylim=ylim,col="grey",sub=subtext)
lines(x,dgp,col=1,lty=1)
lines(x,fit.unres,col=2,lty=1,lwd=2)
lines(x,fit.res,col=3,lty=1,lwd=2)
lines(x,lower,lty=1,col="lightgrey")
lines(x,upper,lty=1,col="lightgrey")

legend("topleft",
       c("DGP","Unconstrained [g(x)]","Constrained [g(x|p)]","Constraints [l(x) <= g(x|p) <= u(x)]"),
       lty=c(1,1,1,1),
       col=c(1:3,"lightgrey"),
       bty="n",
       cex=0.75)

## this uses the opposite of the standard row/col convention... to be fixed
build.kdtree <- function(p, nbucket = 4*nrow(p)){
  nd <- nrow(p)
  n <- ncol(p)

  rbb <- c(paste(c('bbl','bbu'),rep(1:nd, each = 2),sep = '.'))
  rows <- rbb
  ntrd <- length(rows)

  rchi <- c('child.l', 'child.u')
  rowsi <- c(rchi,'nlev','istart')

  ntri <- length(rowsi)

  ##numnode <- 2^(1+ceiling(log(n/nbucket)/log(2))) - 1
  nf <- 2^(1+floor(log(n/nbucket)/log(2)))
  nmax <- 2^(1+ceiling(log(n/nbucket)/log(2))) - 1
  numnode <- min(2*n - (nbucket-1)*nf - 1, nmax)
  
  trd <- matrix(nrow = ntrd, ncol = numnode, dimnames = list(rows, NULL))
  tri <- matrix(data = -1, nrow = ntri, ncol = numnode, dimnames = list(rowsi, NULL))

## a basic kd tree node keeps track of:
## 1. box centre
## 2. edge lengths
## 3. children
## 4. points
##   - if we use a selection function which finds the median and
##     puts points in the array so that they are high-low sorted
##     then every node knows how to find all leaves contained by it
##     quite simply
##     -> sort(x, partial = )

## ROOT STARTS AT LEVEL 1, NODE INDEX 1
## node: current node 
## lev:  level in tree (somewhat redundant)
## nlev: number of poins at this level
## istart: starting index in point array
## nbucket: maximum number of points in a leaf
## nn: total number of nodes in your subtree
  
  build.tree <- function(node = 1, lev = 1, nlev = n, istart = 1, nbucket = nbucket){
    ip <- istart + 0:(nlev-1)
    
    trd[rows,node] <<- sapply(1:nd, function(i) range(p[i,ip]))

    ## set some accounting info

    tri['nlev', node] <<- nlev
    tri['istart', node] <<- istart

    ni <- 1

    if(nlev > nbucket){
      ## next we split the node, cycling through dimensions equally
      d.split <- (lev-1)%%nd + 1

      ## This is more efficiently done in the c backend
      med <- median(p[d.split, ip])
      i.lo <- which(p[d.split,ip] <= med)
      
      p[,ip] <<- cbind(p[,ip[i.lo],drop=FALSE],p[,ip[-i.lo],drop=FALSE])

      cnl2 <- ceiling(nlev/2)

      tri['child.l', node] <<- node + ni

      ni <- ni + build.tree(node = tri['child.l',node], lev = lev+1,
                            nlev = cnl2, istart = istart, nbucket = nbucket)

      tri['child.u', node] <<- node + ni
      
      ni <- ni + build.tree(node = tri['child.u',node], lev = lev+1,
                            nlev = nlev - cnl2, istart = istart + cnl2, nbucket = nbucket)
    }
    return(ni)
  }

  build.tree(nbucket=nbucket)

  tree <- list(p = p, trd = trd, tri = tri, nbucket = nbucket, numnode = numnode, rbb = rbb)
  invisible(tree)
}

## Two problems need be addressed:
## a) orientation of the two rectangles wrt one another
## b) nature of intersection
## 1. No overlap.
## 2. Search box partially contained in node 
## 3. Search box totally contained in node
## 4. Search box partially contains node.
## 5. Search box totally contains node.
## 6. Intersection of search boxes with no contained corners.

## This can be reduced to three states:
## miss, hit-open (partial intersection), and hit-done (the search box envelopes the node)

KD.NOCHILD <- -1

KD.MISS <- 0
KD.HITOPEN <- 1
KD.HITDONE <- 2

## This might look a little mysterious, but it's the result of manually mapping out
## the 6 different scenarios above, so it should look mysterious
boxIntersect <- function(bba, bbb){
  ## bounding boxes - lower.1, upper.1, lower.2, upper.2, ...
  pa <- matrix(bba, byrow = TRUE, ncol = 2)
  pb <- matrix(bbb, byrow = TRUE, ncol = 2)

  ## not bothering to be tricky here - just do two tables of relevant comparisons
  tta <- pa[,1] < pb
  ttb <- pa[,2] < pb

  tt <- cbind(tta,ttb)

  miss <- any(sapply(1:nrow(tt), function(i){ (sum(tt[i,])%%4) == 0 }))

  if(miss)
    return(KD.MISS)

  done <- all(sapply(1:nrow(tt), function(i){ all(xor(tt[i,c(1,2)],tt[i,c(3,4)])) }))
  return(ifelse(done, KD.HITDONE, KD.HITOPEN))
}

boxSearch <- function(tree, box, node = 1){
  ihit <- node

  if((res <- boxIntersect(box, tree$trd[tree$rbb,node])) == KD.MISS)
    return(integer(0))

  is.leaf <- tree$tri['child.l',node] == KD.NOCHILD
  
  if((res == KD.HITOPEN) && (!is.leaf))
    ihit <- c(boxSearch(tree, box, node = tree$tri['child.l',node]),
              boxSearch(tree, box, node = tree$tri['child.u',node]))

  return(ihit)
}

## playing around with the tree 
if(FALSE){
  set.seed(42)

  n <- 67
  nd <- 2
  nbucket <- 8
  
  q <- matrix(runif(nd*n), nrow = nd, ncol = n)
  
  mytree <- build.kdtree(q, nbucket = nbucket)

  attach(mytree)

  op <- par(mfrow = c(2, 2), pty = "s")

  for(i in 1:4){
    plot(p[1,],p[2,],pch=19, type = 'n', xlab = 'x', ylab = 'y', main = paste("level",i))
    ilev <- round(log(n / tri["nlev",])/log(2))+1
    lev.plot <- c(i)
    i.plot <- ilev %in% lev.plot
    colors <- c("blue","green","orange", "red", "purple")
    rect(trd["bbl.1",i.plot], trd["bbl.2",i.plot], trd["bbu.1",i.plot], trd["bbu.2",i.plot],lwd = 4, border = (colors[ilev])[i.plot])
    points(p[1,],p[2,],pch=19)
  }
  
  detach(mytree)

  ## do a box search
  x11()
  plot(q[1,],q[2,], pch = 19, xlab ='x', ylab='y')

  box <- c(0.25,0.5,0.25,0.5)
  nb <- boxSearch(mytree, box)

  rect(mytree$trd[1,nb],mytree$trd[3,nb],mytree$trd[2,nb],mytree$trd[4,nb], border = 'blue', lwd = 3)
  rect(box[1],box[3],box[2],box[4], border = 'red', lwd = 3)
  
  if(FALSE){
    set.seed(42)
    x11()
    while(readline(prompt='continue?') == 'y'){
      bba <- c(range(runif(2)),range(runif(2)))
      bbb <- c(range(runif(2)),range(runif(2)))
      
      res <- boxIntersect(bba,bbb)
      
      plot(NA, main = res, xlab = 'x', ylab = 'y', xlim = c(0,1), ylim = c(0,1), type = 'n')
      rect(bba[1],bba[3],bba[2],bba[4], border = 'red')
      rect(bbb[1],bbb[3],bbb[2],bbb[4], border = 'green')
    }
  }

}

### R code from vignette source 'np.Rnw'

###################################################
### code chunk number 1: np.Rnw:90-91
###################################################
options(prompt = "R> ", np.messages = FALSE, digits = 3, warn = -1)


###################################################
### code chunk number 2: np.Rnw:554-558
###################################################
library("np")
data("cps71")
model.par <- lm(logwage ~ age + I(age^2), data = cps71)
summary(model.par)


###################################################
### code chunk number 3: np.Rnw:570-576
###################################################
model.np <- npreg(logwage ~ age, 
                  regtype = "ll", 
                  bwmethod = "cv.aic",
                  gradients = TRUE,
                  data = cps71)
summary(model.np)


###################################################
### code chunk number 4: np.Rnw:592-593
###################################################
npsigtest(model.np)


###################################################
### code chunk number 5: np.Rnw:651-661 (eval = FALSE)
###################################################
## plot(cps71$age, cps71$logwage, xlab = "age", ylab = "log(wage)", cex=.1)
## lines(cps71$age, fitted(model.np), lty = 1, col = "blue")
## lines(cps71$age, fitted(model.par), lty = 2, col = " red")
## plot(model.np, plot.errors.method = "asymptotic")
## 
## plot(model.np, gradients = TRUE)
## lines(cps71$age, coef(model.par)[2]+2*cps71$age*coef(model.par)[3],
##       lty = 2,
##       col = "red")
## plot(model.np, gradients = TRUE, plot.errors.method = "asymptotic")


###################################################
### code chunk number 6: np.Rnw:666-691
###################################################
getOption("SweaveHooks")[["multifig"]]()
options(SweaveHooks = list(multifig = function() par(mfrow=c(2,2),mar=c(4,4,3,2))))

# Plot 1

plot(cps71$age,cps71$logwage,xlab="age",ylab="log(wage)",cex=.1)
lines(cps71$age,fitted(model.np),lty=1,col="blue")
lines(cps71$age,fitted(model.par),lty=2,col="red")

# Plot 2

plot(cps71$age,gradients(model.np),xlab="age",ylab="gradient",type="l",lty=1,col="blue")
lines(cps71$age,coef(model.par)[2]+2*cps71$age*coef(model.par)[3],lty=2,col="red")

# Plot 3

plot(cps71$age,fitted(model.np),xlab="age",ylab="log(wage)",ylim=c(min(fitted(model.np)-2*model.np$merr),max(fitted(model.np)+2*model.np$merr)),type="l")
lines(cps71$age,fitted(model.np)+2*model.np$merr,lty=2,col="red")
lines(cps71$age,fitted(model.np)-2*model.np$merr,lty=2,col="red")

# Plot 4

plot(cps71$age,gradients(model.np),xlab="age",ylab="gradient",ylim=c(min(gradients(model.np)-2*model.np$gerr),max(gradients(model.np)+2*model.np$gerr)),type="l",lty=1,col="blue")
lines(cps71$age,gradients(model.np)+2*model.np$gerr,lty=2,col="red")
lines(cps71$age,gradients(model.np)-2*model.np$gerr,lty=2,col="red")



###################################################
### code chunk number 7: np.Rnw:752-755
###################################################
cps.eval <- data.frame(age = seq(10,70, by=10))
predict(model.par, newdata = cps.eval)
predict(model.np, newdata = cps.eval)


###################################################
### code chunk number 8: np.Rnw:797-805
###################################################
data("wage1")
model.ols <- lm(lwage ~ female +
                married +
                educ +
                exper +
                tenure,
                data = wage1)
summary(model.ols)


###################################################
### code chunk number 9: np.Rnw:828-848
###################################################
model.ols <- lm(lwage ~ female +
                married +
                educ +
                exper +
                tenure,
                x = TRUE,
                y = TRUE,
                data = wage1)
X <- data.frame(wage1$female,
                wage1$married,
                wage1$educ,
                wage1$exper,
                wage1$tenure)
output <- npcmstest(model = model.ols, 
                    xdat = X, 
                    ydat = wage1$lwage, 
                    nmulti = 1,
                    tol = 0.1, 
                    ftol = 0.1)
summary(output)


###################################################
### code chunk number 10: np.Rnw:903-913
###################################################
#bw.all <- npregbw(formula = lwage ~ female +
#                  married +
#                  educ +
#                  exper +
#                  tenure,
#                  regtype = "ll",
#                  bwmethod = "cv.aic",
#                  data = wage1)
model.np <- npreg(bws = bw.all)
summary(model.np)


###################################################
### code chunk number 11: np.Rnw:946-981
###################################################
set.seed(123)
ii <- sample(seq(1, nrow(wage1)), replace=FALSE)
wage1.train <- wage1[ii[1:400],]
wage1.eval <- wage1[ii[401:nrow(wage1)],]
model.ols <- lm(lwage ~ female +
                married +
                educ +
                exper +
                tenure,
                data = wage1.train)
fit.ols <- predict(model.ols,
                   data = wage1.train,
                   newdata = wage1.eval)
pse.ols <- mean((wage1.eval$lwage - fit.ols)^2)
#bw.subset <- npregbw(formula = lwage ~ female +
#                     married +
#                     educ +
#                     exper +
#                     tenure,
#                     regtype = "ll",
#                     bwmethod = "cv.aic",
#                     data = wage1.train)
model.np <- npreg(bws = bw.subset)
fit.np <- predict(model.np,
                  data = wage1.train,
                  newdata = wage1.eval)
pse.np <- mean((wage1.eval$lwage - fit.np)^2)
bw.freq <- bw.subset
bw.freq$bw[1] <- 0
bw.freq$bw[2] <- 0
model.np.freq <- npreg(bws = bw.freq)
fit.np.freq <- predict(model.np.freq,
                       data = wage1.train,
                       newdata = wage1.eval)
pse.np.freq <- mean((wage1.eval$lwage - fit.np.freq)^2)


###################################################
### code chunk number 12: np.Rnw:1017-1020 (eval = FALSE)
###################################################
## plot(model.np,
##      plot.errors.method = "bootstrap",
##      plot.errors.boot.num = 25)


###################################################
### code chunk number 13: np.Rnw:1025-1028
###################################################
plot(model.np,
     plot.errors.method = "bootstrap",
     plot.errors.boot.num = 25)


###################################################
### code chunk number 14: np.Rnw:1073-1104
###################################################
data("birthwt", package = "MASS")
birthwt$low <- factor(birthwt$low)
birthwt$smoke <- factor(birthwt$smoke)
birthwt$race <- factor(birthwt$race)
birthwt$ht <- factor(birthwt$ht)
birthwt$ui <- factor(birthwt$ui)
birthwt$ftv <- factor(birthwt$ftv)
model.logit <- glm(low ~ smoke +
                   race +
                   ht +
                   ui +
                   ftv +
                   age +
                   lwt, 
                   family = binomial(link = logit),
                   data = birthwt)
model.np <- npconmode(low ~
                      smoke +
                      race +
                      ht +
                      ui +
                      ftv +
                      age +
                      lwt,
                      tol = 0.1,
                      ftol = 0.1,
                      data = birthwt)
cm <- table(birthwt$low, 
            ifelse(fitted(model.logit) > 0.5, 1, 0))
cm
summary(model.np)


###################################################
### code chunk number 15: np.Rnw:1132-1137
###################################################
data("faithful", package = "datasets")
f.faithful <- npudens(~ eruptions + waiting, data = faithful)
F.faithful <- npudist(~ eruptions + waiting, data = faithful)
summary(f.faithful)
summary(F.faithful)


###################################################
### code chunk number 16: np.Rnw:1142-1144 (eval = FALSE)
###################################################
## plot(f.faithful, xtrim = -0.2, view = "fixed", main = "")
## plot(F.faithful, xtrim = -0.2, view = "fixed", main = "")


###################################################
### code chunk number 17: np.Rnw:1149-1163
###################################################
getOption("SweaveHooks")[["multifig"]]()
options(SweaveHooks = list(multifig = function() par(mfrow=c(1,2),mar=rep(1.5,4))))
plot.output <- plot(f.faithful, xtrim=-0.2, view="fixed", main = "",plot.behavior="data")
# Retrieve data from plot() to create multiple figures
f <- matrix(plot.output$d1$dens, 50, 50)
plot.x1 <- unique(plot.output$d1$eval[,1])
plot.x2 <- unique(plot.output$d1$eval[,2])
persp(plot.x1,plot.x2,f,xlab="eruptions",ylab="waiting",zlab="Joint Density",col="lightblue", ticktype="detailed")

plot.output <- plot(F.faithful, xtrim = -0.2, view = "fixed", main="",plot.behavior="data")
# Retrieve data from plot() to create multiple figures
F <- matrix(plot.output$d1$dist, 50, 50)
plot.x1 <- unique(plot.output$d1$eval[,1])
plot.x2 <- unique(plot.output$d1$eval[,2])
persp(plot.x1,plot.x2,F,xlab="eruptions",ylab="waiting",zlab="Joint Distribution",col="lightblue", ticktype="detailed")


###################################################
### code chunk number 18: np.Rnw:1189-1200
###################################################
data("Italy")
fhat <- npcdens(gdp ~ year,
                tol = 0.1,
                ftol = 0.1,
                data = Italy)
summary(fhat)
Fhat <- npcdist(gdp ~ year,
                tol = 0.1,
                ftol = 0.1,
                data = Italy)
summary(Fhat)


###################################################
### code chunk number 19: np.Rnw:1207-1209 (eval = FALSE)
###################################################
## plot(fhat, view = "fixed", main = "", theta = 300, phi = 50)
## plot(Fhat, view = "fixed", main = "", theta = 300, phi = 50)


###################################################
### code chunk number 20: np.Rnw:1214-1229
###################################################
getOption("SweaveHooks")[["multifig"]]()
options(SweaveHooks = list(multifig = function() par(mfrow=c(1,2),mar=rep(1.25,4))))
plot.output <- plot(fhat, view="fixed", main="",plot.behavior="data")
# Retrieve data from plot() to create multiple figures
f <- matrix(plot.output$cd1$condens, 48, 50)
plot.y1 <- unique(plot.output$cd1$yeval)
plot.x1 <- unique(plot.output$cd1$xeval)
persp(as.integer(levels(plot.x1)),plot.y1,f,xlab="year",ylab="gdp",zlab="Conditional Density",col="lightblue", ticktype="detailed",theta=300,phi=50)

plot.output <- plot(Fhat, view="fixed", main="",plot.behavior="data")
# Retrieve data from plot() to create multiple figures
F <- matrix(plot.output$cd1$condist, 48, 50)
plot.y1 <- unique(plot.output$cd1$yeval)
plot.x1 <- unique(plot.output$cd1$xeval)
persp(as.numeric(plot.x1)+1951,plot.y1,F,xlab="year",ylab="gdp",zlab="Conditional Distribution",col="lightblue", ticktype="detailed",theta=300,phi=50)



###################################################
### code chunk number 21: np.Rnw:1259-1266
###################################################
bw <- npcdistbw(formula = gdp ~ year,
                tol = 0.1,
                ftol = 0.1,
                data = Italy)
model.q0.25 <- npqreg(bws = bw, tau = 0.25)
model.q0.50 <- npqreg(bws = bw, tau = 0.50)
model.q0.75 <- npqreg(bws = bw, tau = 0.75)


###################################################
### code chunk number 22: np.Rnw:1273-1280 (eval = FALSE)
###################################################
## plot(Italy$year, Italy$gdp, main = "", 
##      xlab = "Year", ylab = "GDP Quantiles")
## lines(Italy$year, model.q0.25$quantile, col = "red", lty = 1, lwd = 2)
## lines(Italy$year, model.q0.50$quantile, col = "blue", lty = 2, lwd = 2)
## lines(Italy$year, model.q0.75$quantile, col = "red", lty = 3, lwd = 2)
## legend(ordered(1951), 32, c("tau = 0.25", "tau = 0.50", "tau = 0.75"), 
##        lty = c(1, 2, 3), col = c("red", "blue", "red"))


###################################################
### code chunk number 23: np.Rnw:1289-1298
###################################################
plot(Italy$year, Italy$gdp, 
     main = "",
     xlab = "Year", 
     ylab = "GDP Quantiles")
lines(Italy$year, model.q0.25$quantile, col = "red", lty = 1, lwd = 2)
lines(Italy$year, model.q0.50$quantile, col = "blue", lty = 2, lwd = 2)
lines(Italy$year, model.q0.75$quantile, col = "red", lty = 3, lwd = 2)
legend(ordered(1951), 32, c("tau = 0.25", "tau = 0.50", "tau = 0.75"), 
       lty = c(1, 2, 3), col = c("red", "blue", "red"))


###################################################
### code chunk number 24: np.Rnw:1341-1347
###################################################
model.pl <- npplreg(lwage ~ female +
                    married +
                    educ +
                    tenure | exper,
                    data = wage1)
summary(model.pl)


###################################################
### code chunk number 25: np.Rnw:1367-1379
###################################################
model.index <- npindex(low ~
                       smoke +
                       race +
                       ht +
                       ui +
                       ftv +
                       age +
                       lwt,
                       method = "kleinspady",
                       gradients = TRUE,
                       data = birthwt)
summary(model.index)


###################################################
### code chunk number 26: np.Rnw:1399-1407
###################################################
model <- npindex(lwage ~ female +
                 married +
                 educ +
                 exper +
                 tenure,
                 data = wage1,
                 nmulti = 1)
summary(model)


###################################################
### code chunk number 27: np.Rnw:1433-1452
###################################################
model.ols <- lm(lwage ~ female +
                married +
                educ +
                exper +
                tenure,
                data = wage1)
wage1.augmented <- wage1
wage1.augmented$dfemale <- as.integer(wage1$female == "Male")
wage1.augmented$dmarried <- as.integer(wage1$married == "Notmarried")
model.scoef <- npscoef(lwage ~ dfemale +
                       dmarried +
                       educ +
                       exper +
                       tenure | female,
                       betas = TRUE,
                       data = wage1.augmented)
summary(model.scoef)
colMeans(coef(model.scoef))
coef(model.ols)


###################################################
### code chunk number 28: np.Rnw:1478-1480
###################################################
fit.lc <- npksum(txdat = cps71$age, tydat = cps71$logwage, bws = 2)$ksum/
  npksum(txdat = cps71$age, bws = 2)$ksum


###################################################
### code chunk number 29: np.Rnw:1485-1487 (eval = FALSE)
###################################################
## plot(cps71$age, cps71$logwage, xlab = "Age", ylab = "log(wage)")
## lines(cps71$age, fit.lc, col = "blue")


###################################################
### code chunk number 30: np.Rnw:1492-1494
###################################################
plot(cps71$age, cps71$logwage, xlab = "Age", ylab = "log(wage)", cex=.1)
lines(cps71$age, fit.lc, col = "blue")

### R code from vignette source 'entropy_np.Rnw'

###################################################
### code chunk number 1: entropy_np.Rnw:70-72
###################################################
library(np)
options(prompt = "R> ", np.messages = FALSE, digits = 3, warn = -1)


###################################################
### code chunk number 2: entropy_np.Rnw:179-194
###################################################
set.seed(1234)
n <- 250
## Distributions are equal

sample.A <- data.frame(a=rnorm(n),b=factor(rbinom(n,2,.5)))
sample.B <- data.frame(a=rnorm(n),b=factor(rbinom(n,2,.5)))

npdeneqtest(sample.A,sample.B,boot.num=99)

## Distributions are unequal

sample.A <- data.frame(a=rnorm(n),b=factor(rbinom(n,2,.5)))
sample.B <- data.frame(a=rnorm(n,sd=10),b=factor(rbinom(n,2,.25)))

npdeneqtest(sample.A,sample.B,boot.num=99)


###################################################
### code chunk number 3: entropy_np.Rnw:226-254
###################################################
set.seed(1234)
n <- 1000
     
## Compute the statistic only, different distributions
     
x <- rchisq(n,df=10)
y <- rnorm(n,sd=10000,mean=-10000)
     
npunitest(x,y,bootstrap=FALSE)

## Data drawn from same continuous distribution

x <- rnorm(n)
y <- rnorm(n)
npunitest(x,y,boot.num=99)

## Data drawn from different continuous distributions having the
## same mean and variance

x <- rchisq(n,df=5)
y <- rnorm(n,mean=5,sd=sqrt(10))
npunitest(x,y,boot.num=99)

## Data drawn from different discrete distributions
     
x <- factor(rbinom(n,2,.5))
y <- factor(rbinom(n,2,.1))
npunitest(x,y,boot.num=99)


###################################################
### code chunk number 4: entropy_np.Rnw:312-325
###################################################
set.seed(1234)
     
n <- 100

## Asymmetric discrete probability distribution

x <- factor(rbinom(n,2,.8))
npsymtest(x,boot.num=99)

## Symmetric continuous distribution
     
y <- rnorm(n)
npsymtest(y,boot.num=99)


###################################################
### code chunk number 5: entropy_np.Rnw:365-374
###################################################
set.seed(123)
## Test/measure lack of fit between y and its fitted value from a
## regression model when x is relevant.
n <- 100
x <- rnorm(n)
y <- 1 + x + rnorm(n)
model <- lm(y~x)
y.fit <- fitted(model)
npdeptest(y,y.fit,boot.num=99,method="summation")


###################################################
### code chunk number 6: entropy_np.Rnw:408-423
###################################################
set.seed(123)
## A function to create a time series
ar.series <- function(phi,epsilon) {
  n <- length(epsilon)
  series <- numeric(n)
  series[1] <- epsilon[1]/(1-phi)
  for(i in 2:n) {
    series[i] <- phi*series[i-1] + epsilon[i]
  }
  return(series)
}
n <- 100
## Stationary persistent time-series
yt <- ar.series(0.95,rnorm(n))
npsdeptest(yt,lag.num=2,boot.num=99,method="summation")


###################################################
### code chunk number 7: entropy_np.Rnw:447-477
###################################################
Srho <- function(x,y,...) {
  ## First write a function to compute the integrand (this is fed to
  ## the `integrate' function). This function's first argument is the
  ## point at which the two densities are computed (the remaining
  ## arguments are the data vectors).
  integrand <- function(t,x,y) {
    ## First, nonparametrically estimate the density of the x data
    ## using a plug-in bandwidth and evaluate the density at the point
    ## `t'.
    f.x <- fitted(npudens(tdat=x,edat=t,bws=bw.SJ(x),...))
    ## Next, estimate the parametric density of the data y using the
    ## Gaussian distribution and evaluate the density at the point
    ## `t'.
    f.y <- dnorm(t,mean=mean(y),sd=sd(y))
    ## Compute and return the integrand evaluated at the point `t'.
    return(0.5*(sqrt(f.x)-sqrt(f.y))**2)
  }
  ## Feed the integrand function to integrate() and return the value
  ## of the integral.
  return(integrate(integrand,-Inf,Inf,x=x,y=y)$value)
}
set.seed(123)
n <- 1000
## Data drawn from the same distribution
x <- rnorm(n)
y <- rnorm(n)
Srho(x,y)
## Data drawn from different distributions
y <- rnorm(n,sd=100)
Srho(x,y)

require(np)

## Example - compute nonparametric instrumental regression using
## Landweber-Fridman iteration of Fredholm integral equations of the
## first kind.

## We consider an equation with an endogenous regressor (`z') and an
## instrument (`w'). Let y = phi(z) + u where phi(z) is the function of
## interest. Here E(u|z) is not zero hence the conditional mean E(y|z)
## does not coincide with the function of interest, but if there exists
## an instrument w such that E(u|w) = 0, then we can recover the
## function of interest by solving an ill-posed inverse problem.

data(Engel95)

## Sort on logexp (the endogenous regressor) for plotting purposes

Engel95 <- Engel95[order(Engel95$logexp),] 

attach(Engel95)

model.iv <- npregiv(y=food,z=logexp,w=logwages,method="Landweber-Fridman")
phi <- model.iv$phi

## Compute the non-IV regression (i.e. regress y on z)

ghat <- npreg(food~logexp,regtype="ll")

## For the plots, restrict focal attention to the bulk of the data
## (i.e. for the plotting area trim out 1/4 of one percent from each
## tail of y and z)

trim <- 0.0025

plot(logexp,food,
     ylab="Food Budget Share",
     xlab="log(Total Expenditure)",
     xlim=quantile(logexp,c(trim,1-trim)),
     ylim=quantile(food,c(trim,1-trim)),
     main="Nonparametric Instrumental Kernel Regression",
     type="p",
     cex=.5,
     col="lightgrey")

lines(logexp,phi,col="blue",lwd=2,lty=2)

lines(logexp,fitted(ghat),col="red",lwd=2,lty=4)

legend(quantile(logexp,trim),quantile(food,1-trim),
       c(expression(paste("Nonparametric IV: ",hat(varphi)(logexp))),
         "Nonparametric Regression: E(food | logexp)"),
       lty=c(2,4),
       col=c("blue","red"),
       lwd=c(2,2))

npcdistbw <-
  function(...){
    mc <- match.call(expand.dots = FALSE)
    npRejectRenamedScaleFactorSearchArgs(names(mc$...), where = "npcdistbw")
    target <- .np_bw_dispatch_target(dots = mc$...,
                                     data_arg_names = c("xdat", "ydat", "gydat"),
                                     eval_env = parent.frame())
    UseMethod("npcdistbw", target)
  }

npcdistbw.formula <-
  function(formula, data, subset, na.action, call, gdata = NULL, ...){
    orig.ts <- tryCatch({
        if (missing(data))
            .np_terms_ts_mask(terms_obj = terms(formula),
                              data = environment(formula),
                              eval_env = environment(formula))
        else .np_terms_ts_mask(terms_obj = terms(formula, data = data),
                               data = data,
                               eval_env = environment(formula))
    }, error = function(e) FALSE)

    has.gval <- !is.null(gdata)
    
    gmf <- mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(mf), nomatch = 0)
    gm <- match(c("formula", "gdata"),
               names(gmf), nomatch = 0)

    mf <- mf[c(1,m)]
    gmf <- gmf[c(1,gm)]

    formula.call <- .np_bw_formula_from_call(call_obj = call, eval_env = parent.frame())
    if (!is.null(formula.call)) {
      mf[[2]] <- formula.call
      gmf[[2]] <- formula.call
    }
                     

    mf[[1]] <- as.name("model.frame")
    gmf[[1]] <- as.name("model.frame")
    formula.obj <- .np_bw_resolve_formula(formula_obj = formula,
                                        formula_call = formula.call,
                                        eval_env = parent.frame())

    variableNames <- if(m[2] > 0) explodeFormula(formula.obj, data = data) else explodeFormula(formula.obj)
    
    ## make formula evaluable, then eval
    varsPlus <- lapply(variableNames, paste, collapse=" + ")
    mf[["formula"]] <- as.formula(paste(" ~ ", varsPlus[[1]]," + ",
                                        varsPlus[[2]]),
                                  env = environment(formula))
    gmf[["formula"]] <- mf[["formula"]]

    mf[["formula"]] <- terms(mf[["formula"]])
    if(all(orig.ts)){
      args <- (as.list(attr(mf[["formula"]], "variables"))[-1])
      attr(mf[["formula"]], "predvars") <- as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), args))))
    }else if(any(orig.ts)){
      arguments <- (as.list(attr(mf[["formula"]], "variables"))[-1])
      arguments.normal <- arguments[which(!orig.ts)]
      arguments.timeseries <- arguments[which(orig.ts)]

      ix <- sort(c(which(orig.ts),which(!orig.ts)),index.return = TRUE)$ix
      attr(mf[["formula"]], "predvars") <- bquote(.(as.call(c(quote(cbind),as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments.timeseries)))),arguments.normal,check.rows = TRUE)))[,.(ix)])
    }
    
    mf.args <- as.list(mf[-1L])
    mf <- do.call(stats::model.frame, mf.args, envir = parent.frame())
    
    ydat <- mf[, variableNames[[1]], drop = FALSE]
    xdat <- mf[, variableNames[[2]], drop = FALSE]

    if (has.gval) {
      gmf.args <- as.list(gmf[-1L])
      names(gmf.args)[names(gmf.args) == "gdata"] <- "data"
      gmf <- do.call(stats::model.frame, gmf.args, envir = parent.frame())
      gydat <- gmf[, variableNames[[1]], drop = FALSE]
    }
    
    bw.args <- list(xdat = xdat, ydat = ydat)
    if (has.gval)
      bw.args$gydat <- gydat
    tbw <- do.call(npcdistbw, c(bw.args, list(...)))

    ## clean up (possible) inconsistencies due to recursion ...
    tbw$call <- match.call(expand.dots = FALSE)
    environment(tbw$call) <- parent.frame()
    tbw$formula <- formula
    tbw$rows.omit <- as.vector(attr(mf,"na.action"))
    tbw$nobs.omit <- length(tbw$rows.omit)
    tbw$terms <- attr(mf,"terms")
    tbw$variableNames <- variableNames

    tbw
  }

npcdistbw.condbandwidth <-
  function(xdat = stop("data 'xdat' missing"),
           ydat = stop("data 'ydat' missing"),
           gydat = NULL,
           bws,
           bandwidth.compute = TRUE,
           cfac.dir = 2.5*(3.0-sqrt(5)),
           scale.factor.init = 0.5,
           dfac.dir = 0.25*(3.0-sqrt(5)),
           dfac.init = 0.375,
           dfc.dir = 3,
           do.full.integral = FALSE,
           ftol = 1.490116e-07,
           scale.factor.init.upper = 2.0,
           hbd.dir = 1,
           hbd.init = 0.9,
           initc.dir = 1.0,
           initd.dir = 1.0,
           invalid.penalty = c("baseline","dbmax"),
           itmax = 10000,
           lbc.dir = 0.5,
           scale.factor.init.lower = 0.1,
           lbd.dir = 0.1,
           lbd.init = 0.1,
           memfac = 500.0,
           ngrid = 100,
           nmulti,
           penalty.multiplier = 10,
           powell.remin = TRUE,
           scale.init.categorical.sample = FALSE,
           scale.factor.search.lower = NULL,
           small = 1.490116e-05,
           tol = 1.490116e-04,
           transform.bounds = FALSE,
           ...){

    elapsed.start <- proc.time()[3]

    ydat = toFrame(ydat)
    xdat = toFrame(xdat)

    if (missing(nmulti)){
      nmulti <- npDefaultNmulti(dim(ydat)[2]+dim(xdat)[2])
    }
    bandwidth.compute <- npValidateScalarLogical(bandwidth.compute, "bandwidth.compute")
    remin <- npValidateScalarLogical(powell.remin, "powell.remin")
    do.full.integral <- npValidateScalarLogical(do.full.integral, "do.full.integral")
    scale.init.categorical.sample <-
      npValidateScalarLogical(scale.init.categorical.sample, "scale.init.categorical.sample")
    transform.bounds <- npValidateScalarLogical(transform.bounds, "transform.bounds")
    itmax <- npValidatePositiveInteger(itmax, "itmax")
    ngrid <- npValidatePositiveInteger(ngrid, "ngrid")
    ftol <- npValidatePositiveFiniteNumeric(ftol, "ftol")
    tol <- npValidatePositiveFiniteNumeric(tol, "tol")
    small <- npValidatePositiveFiniteNumeric(small, "small")
    memfac <- npValidatePositiveFiniteNumeric(memfac, "memfac")
    penalty.multiplier <- npValidatePositiveFiniteNumeric(penalty.multiplier, "penalty.multiplier")
    scale.factor.search.lower <- npResolveScaleFactorLowerBound(
      if (is.null(scale.factor.search.lower)) npGetScaleFactorSearchLower(bws) else scale.factor.search.lower
    )
    nmulti <- npValidateNmulti(nmulti)
    .np_progress_bandwidth_set_total(nmulti)

    if (length(bws$ybw) != dim(ydat)[2])
      stop(paste("length of bandwidth vector does not match number of columns of", "'ydat'"))

    if (length(bws$xbw) != dim(xdat)[2])
      stop(paste("length of bandwidth vector does not match number of columns of", "'xdat'"))

    if (dim(ydat)[1] != dim(xdat)[1])
      stop(paste("number of rows of", "'ydat'", "does not match", "'xdat'"))

    if ((any(bws$iycon) &&
         !all(vapply(as.data.frame(ydat[, bws$iycon]), inherits, logical(1), c("integer", "numeric")))) ||
        (any(bws$iyord) &&
         !all(vapply(as.data.frame(ydat[, bws$iyord]), inherits, logical(1), "ordered"))) ||
        (any(bws$iyuno) &&
         !all(vapply(as.data.frame(ydat[, bws$iyuno]), inherits, logical(1), "factor"))))
      stop(paste("supplied bandwidths do not match", "'ydat'", "in type"))

    if ((any(bws$ixcon) &&
         !all(vapply(as.data.frame(xdat[, bws$ixcon]), inherits, logical(1), c("integer", "numeric")))) ||
        (any(bws$ixord) &&
         !all(vapply(as.data.frame(xdat[, bws$ixord]), inherits, logical(1), "ordered"))) ||
        (any(bws$ixuno) &&
         !all(vapply(as.data.frame(xdat[, bws$ixuno]), inherits, logical(1), "factor"))))
      stop(paste("supplied bandwidths do not match", "'xdat'", "in type"))

    ##if (bws$type != 'fixed')
    ##stop("only fixed bandwidths currently supported with ccdf bandwidth selection")

    ## catch and destroy NA's
    goodrows <- seq_len(nrow(xdat))
    rows.omit <- unclass(na.action(na.omit(data.frame(xdat,ydat))))
    goodrows[rows.omit] <- 0

    if (all(goodrows==0))
      stop("Data has no rows without NAs")

    xdat = xdat[goodrows,,drop = FALSE]
    ydat = ydat[goodrows,,drop = FALSE]

    
    nrow = nrow(ydat)
    yncol = ncol(ydat)
    xncol = ncol(xdat)

    ## at this stage, data to be sent to the c routines must be converted to
    ## numeric type.

    oydat <- ydat
    
    ydat = toMatrix(ydat)

    yuno = ydat[, bws$iyuno, drop = FALSE]
    ycon = ydat[, bws$iycon, drop = FALSE]
    yord = ydat[, bws$iyord, drop = FALSE]

    xdat = toMatrix(xdat)

    xuno = xdat[, bws$ixuno, drop = FALSE]
    xcon = xdat[, bws$ixcon, drop = FALSE]
    xord = xdat[, bws$ixord, drop = FALSE]

    tbw <- bws
    spec <- npCanonicalConditionalRegSpec(
      regtype = if (is.null(tbw$regtype)) "lc" else tbw$regtype,
      basis = if (is.null(tbw$basis)) "glp" else tbw$basis,
      degree = if (is.null(tbw$degree)) NULL else tbw$degree,
      bernstein.basis = isTRUE(tbw$bernstein.basis),
      ncon = tbw$xncon,
      where = "npcdistbw"
    )
    tbw$regtype <- spec$regtype
    tbw$pregtype <- switch(spec$regtype,
                           lc = "Local-Constant",
                           ll = "Local-Linear",
                           lp = "Local-Polynomial")
    tbw$basis <- spec$basis
    tbw$degree <- spec$degree
    tbw$bernstein.basis <- spec$bernstein.basis
    tbw$regtype.engine <- spec$regtype.engine
    tbw$basis.engine <- spec$basis.engine
    tbw$degree.engine <- spec$degree.engine
    tbw$bernstein.basis.engine <- spec$bernstein.basis.engine
    reg.code <- if (identical(spec$regtype.engine, "lp")) REGTYPE_LP else REGTYPE_LC
    degree.code <- if (tbw$xncon > 0L) as.integer(spec$degree.engine) else integer(0)
    basis.code <- as.integer(npLpBasisCode(spec$basis.engine))
    bernstein.engine <- isTRUE(spec$bernstein.basis.engine)

    if(!is.null(gydat)){
      gydat <- toFrame(gydat)
      if(any(is.na(gydat)))
        stop("na's not allowed to be present in cdf gdata")

      gydat <- toMatrix(gydat)

      gyuno = gydat[, bws$iyuno, drop = FALSE]
      gyord = gydat[, bws$iyord, drop = FALSE]
      gycon = gydat[, bws$iycon, drop = FALSE]
      cdf_on_train = FALSE
      nog = nrow(gydat)

    } else {
      if(do.full.integral) {
        cdf_on_train = TRUE
        nog = 0

        gyuno = data.frame()
        gyord = data.frame()
        gycon = data.frame()

      } else {
        cdf_on_train = FALSE
        nog = ngrid
        probs <- seq(0,1,length.out = nog)

        evy <- oydat[seq_len(nog),,drop = FALSE]
        for (i in seq_len(ncol(evy))) {
          evy[,i] <- uocquantile(oydat[,i], probs)
        }

        evy <- toMatrix(evy)

        gyuno = evy[, bws$iyuno, drop = FALSE]
        gyord = evy[, bws$iyord, drop = FALSE]
        gycon = evy[, bws$iycon, drop = FALSE]

      }

    }

    mysd <- EssDee(data.frame(xcon,ycon))
    nconfac <- nrow^(-1.0/(2.0*bws$cxkerorder+bws$ncon))
    ncatfac <- nrow^(-2.0/(2.0*bws$cxkerorder+bws$ncon))

    invalid.penalty <- match.arg(invalid.penalty)
    penalty_mode <- (if (invalid.penalty == "baseline") 1L else 0L)

    if (bandwidth.compute){
      cont.start <- npContinuousSearchStartControls(
        scale.factor.init.lower,
        scale.factor.init.upper,
        scale.factor.init,
        scale.factor.search.lower,
        where = "npcdistbw"
      )
      myopti = list(num_obs_train = nrow,
        num_obs_grid = nog,
        iMultistart = IMULTI_TRUE,
        iNum_Multistart = nmulti,
        int_use_starting_values = (if (all(bws$ybw==0) && all(bws$xbw==0)) USE_START_NO else USE_START_YES),
        int_LARGE_SF = (if (bws$scaling) SF_NORMAL else SF_ARB),
        BANDWIDTH_den_extern = switch(bws$type,
          fixed = BW_FIXED,
          generalized_nn = BW_GEN_NN,
          adaptive_nn = BW_ADAP_NN),
        itmax=itmax, int_RESTART_FROM_MIN=(if (remin) RE_MIN_TRUE else RE_MIN_FALSE), 
        int_MINIMIZE_IO=if (isTRUE(getOption("np.messages"))) IO_MIN_FALSE else IO_MIN_TRUE, 
        bwmethod = switch(bws$method,
          cv.ls = CDBWM_CVLS),
        xkerneval = switch(bws$cxkertype,
          gaussian = CKER_GAUSS + bws$cxkerorder/2 - 1,
          epanechnikov = CKER_EPAN + bws$cxkerorder/2 - 1,
          uniform = CKER_UNI,
          "truncated gaussian" = CKER_TGAUSS),
        ykerneval = switch(bws$cykertype,
          gaussian = CKER_GAUSS + bws$cykerorder/2 - 1,
          epanechnikov = CKER_EPAN + bws$cykerorder/2 - 1,
          uniform = CKER_UNI,
          "truncated gaussian" = CKER_TGAUSS),
        uxkerneval = switch(bws$uxkertype,
          aitchisonaitken = UKER_AIT,
          liracine = UKER_LR),
        uykerneval = switch(bws$uykertype,
          aitchisonaitken = UKER_AIT,
          liracine = UKER_LR),
        oxkerneval = switch(bws$oxkertype,
          wangvanryzin = OKER_WANG,
          liracine = OKER_LR,
        "racineliyan" = OKER_RLY),
        oykerneval = switch(bws$oykertype,
          wangvanryzin = OKER_WANG,
          liracine = OKER_NLR,
        "racineliyan" = OKER_RLY),
        ynuno = dim(yuno)[2],
        ynord = dim(yord)[2],
        yncon = dim(ycon)[2],
        xnuno = dim(xuno)[2],
        xnord = dim(xord)[2],
        xncon = dim(xcon)[2],
        cdf_on_train = cdf_on_train,
        int_do_tree = if (isTRUE(getOption("np.tree"))) DO_TREE_YES else DO_TREE_NO,
        scale.init.categorical.sample=scale.init.categorical.sample,
        dfc.dir = dfc.dir,
        transform.bounds = transform.bounds)
      
      myoptd = list(ftol=ftol, tol=tol, small=small, memfac = memfac,
        lbc.dir = lbc.dir, cfac.dir = cfac.dir, initc.dir = initc.dir, 
        lbd.dir = lbd.dir, hbd.dir = hbd.dir, dfac.dir = dfac.dir, initd.dir = initd.dir, 
        lbc.init = cont.start$scale.factor.init.lower, hbc.init = cont.start$scale.factor.init.upper, cfac.init = cont.start$scale.factor.init, 
        lbd.init = lbd.init, hbd.init = hbd.init, dfac.init = dfac.init, 
        nconfac = nconfac, ncatfac = ncatfac,
        scale.factor.lower.bound = scale.factor.search.lower)

      cxker.bounds.c <- npKernelBoundsMarshal(bws$cxkerlb[bws$ixcon], bws$cxkerub[bws$ixcon])
      cyker.bounds.c <- npKernelBoundsMarshal(bws$cykerlb[bws$iycon], bws$cykerub[bws$iycon])

      if (bws$method != "normal-reference"){
        myout <-
          npWithLocalLinearRawBasisSearchError(
            .Call("C_np_distribution_conditional_bw",
                  as.double(yuno), as.double(yord), as.double(ycon),
                  as.double(xuno), as.double(xord), as.double(xcon),
                  as.double(gyuno), as.double(gyord), as.double(gycon),
                  as.double(mysd),
                  as.integer(myopti), as.double(myoptd),
                  as.double(c(bws$xbw[bws$ixcon], bws$ybw[bws$iycon],
                              bws$ybw[bws$iyuno], bws$ybw[bws$iyord],
                              bws$xbw[bws$ixuno], bws$xbw[bws$ixord])),
                  as.integer(nmulti),
                  as.integer(penalty_mode),
                  as.double(penalty.multiplier),
                  as.integer(degree.code),
                  as.integer(bernstein.engine),
                  as.integer(basis.code),
                  as.integer(reg.code),
                  as.double(cxker.bounds.c$lb),
                  as.double(cxker.bounds.c$ub),
                  as.double(cyker.bounds.c$lb),
                  as.double(cyker.bounds.c$ub),
                  PACKAGE="np"),
            where = "npcdistbw",
            spec = spec,
            bwmethod = bws$method,
            ncon = tbw$xncon
          )
        total.time <- proc.time()[3] - elapsed.start
      } else {
        nbw = double(yncol+xncol)
        gbw = bws$yncon+bws$xncon
        if (gbw > 0){
          xcon_idx <- seq_len(bws$xncon)
          ycon_idx <- seq.int(from = bws$xncon + 1L, length.out = bws$yncon)
          if (length(xcon_idx) > 0L)
            nbw[xcon_idx] <- 1.06
          if (length(ycon_idx) > 0L)
            nbw[ycon_idx] <- 1.587
          if(!bws$scaling){
            gbw_idx <- seq_len(gbw)
            nbw[gbw_idx]=nbw[gbw_idx]*mysd*nconfac
          }
        }
        myout= list( bw = nbw, fval = c(NA,NA) )
        total.time <- NA
      }

      yr = seq_len(yncol)
      xr = seq_len(xncol)
      rorder = numeric(yncol + xncol)

      ## bandwidths are passed back from the C routine in an unusual order
      ## xc, y[cuo], x[uo]
      
      rxcon = xr[bws$ixcon]
      rxuno = xr[bws$ixuno] 
      rxord = xr[bws$ixord] 

      rycon = yr[bws$iycon] 
      ryuno = yr[bws$iyuno] 
      ryord = yr[bws$iyord] 


      ## rorder[c(rxcon,rycon,ryuno,ryord,rxuno,rxord)]=1:(yncol+xncol)

      tbw <- bws
      tbw$ybw[c(rycon,ryuno,ryord)] <- myout$bw[yr+bws$xncon]
      tbw$xbw[c(rxcon,rxuno,rxord)] <- myout$bw[setdiff(seq_len(yncol + xncol), yr + bws$xncon)]

      tbw$fval = myout$fval[1]
      tbw$ifval = myout$fval[2]
      tbw$num.feval <- sum(myout$eval.history[is.finite(myout$eval.history)])
      tbw$num.feval.fast <- myout$fast.history[1]
      tbw$fval.history <- myout$fval.history
      tbw$eval.history <- myout$eval.history
      tbw$invalid.history <- myout$invalid.history
      tbw$timing <- myout$timing
      tbw$total.time <- total.time
    }
    
    ## bandwidth metadata
    tbw$sfactor <- tbw$bandwidth <- list(x = tbw$xbw, y = tbw$ybw)

    apply_bw_meta <- function(tl, dfactor){
      for (nm in names(tl)) {
        idx <- tl[[nm]]
        if (length(idx) == 0L)
          next
        if (tbw$scaling) {
          tbw$bandwidth[[nm]][idx] <- tbw$bandwidth[[nm]][idx] * dfactor[[nm]]
        } else {
          tbw$sfactor[[nm]][idx] <- tbw$sfactor[[nm]][idx] / dfactor[[nm]]
        }
      }
    }
    
    if ((tbw$xnuno+tbw$ynuno) > 0){
      dfactor <- ncatfac
      dfactor <- list(x = dfactor, y = dfactor)

      tl <- list(x = tbw$xdati$iuno, y = tbw$ydati$iuno)

      apply_bw_meta(tl = tl, dfactor = dfactor)
    }

    if ((tbw$xnord+tbw$ynord) > 0){
      dfactor <- ncatfac
      dfactor <- list(x = dfactor, y = dfactor)

      tl <- list(x = tbw$xdati$iord, y = tbw$ydati$iord)

      apply_bw_meta(tl = tl, dfactor = dfactor)
    }

      
    if (tbw$ncon > 0){
      dfactor <- nconfac
      dfactor <- list(x = EssDee(xcon)*dfactor, y = EssDee(ycon)*dfactor)

      tl <- list(x = tbw$xdati$icon, y = tbw$ydati$icon)

      apply_bw_meta(tl = tl, dfactor = dfactor)
    }
  
    tbw <- condbandwidth(xbw = tbw$xbw,
                         ybw = tbw$ybw,
                         bwmethod = tbw$method,
                         bwscaling = tbw$scaling,
                         bwtype = tbw$type,
                         cxkertype = tbw$cxkertype,
                         cxkerorder = tbw$cxkerorder,
                         cxkerbound = tbw$cxkerbound,
                         cxkerlb = tbw$cxkerlb,
                         cxkerub = tbw$cxkerub,
                         uxkertype = tbw$uxkertype,
                         oxkertype = tbw$oxkertype,
                         cykertype = tbw$cykertype,
                         cykerorder = tbw$cykerorder,
                         cykerbound = tbw$cykerbound,
                         cykerlb = tbw$cykerlb,
                         cykerub = tbw$cykerub,
                         uykertype = tbw$uykertype,
                         oykertype = tbw$oykertype,
                        fval = tbw$fval,
                        ifval = tbw$ifval,
                        num.feval = tbw$num.feval,
                        num.feval.fast = tbw$num.feval.fast,
                        fval.history = tbw$fval.history,
                         eval.history = tbw$eval.history,
                         invalid.history = tbw$invalid.history,
                         nobs = tbw$nobs,
                         xdati = tbw$xdati,
                         ydati = tbw$ydati,      
                         xnames = tbw$xnames,
                         ynames = tbw$ynames,
                         sfactor = tbw$sfactor,
                         bandwidth = tbw$bandwidth,
                         rows.omit = rows.omit,
                         nconfac = nconfac,
                         ncatfac = ncatfac,
                         sdev = mysd,
                         bandwidth.compute = bandwidth.compute,
                         timing = tbw$timing,
                         total.time = tbw$total.time,
                         regtype = tbw$regtype,
                         pregtype = tbw$pregtype,
                         basis = tbw$basis,
                         degree = tbw$degree,
                         bernstein.basis = tbw$bernstein.basis,
                         regtype.engine = tbw$regtype.engine,
                         basis.engine = tbw$basis.engine,
                         degree.engine = tbw$degree.engine,
                         bernstein.basis.engine = tbw$bernstein.basis.engine)
    tbw <- npSetScaleFactorSearchLower(tbw, scale.factor.search.lower)

    tbw <- .np_refresh_xy_bandwidth_metadata(tbw)

    tbw
  }

.npcdistbw_build_condbandwidth <- function(xdat,
                                           ydat,
                                           bws,
                                           bandwidth.compute,
                                           reg.args) {
  x.info <- untangle(xdat)
  y.info <- untangle(ydat)
  y.idx <- seq_len(ncol(ydat))
  x.idx <- seq_len(ncol(xdat))

  bw.args <- c(
    list(
      xbw = bws[length(y.idx) + x.idx],
      ybw = bws[y.idx],
      nobs = nrow(xdat),
      xdati = x.info,
      ydati = y.info,
      xnames = names(xdat),
      ynames = names(ydat),
      bandwidth.compute = bandwidth.compute
    ),
    reg.args
  )

  out <- do.call(condbandwidth, bw.args)
  if (!is.null(reg.args$scale.factor.search.lower))
    out$scale.factor.search.lower <- npResolveScaleFactorLowerBound(
      reg.args$scale.factor.search.lower
    )
  out
}

.npcdistbw_eval_only <- function(xdat,
                                 ydat,
                                 gydat = NULL,
                                 bws,
                                 do.full.integral = FALSE,
                                 ngrid = 100L,
                                 invalid.penalty = c("baseline", "dbmax"),
                                 penalty.multiplier = 10) {
  invalid.penalty <- match.arg(invalid.penalty)

  ydat <- toFrame(ydat)
  xdat <- toFrame(xdat)

  if (length(bws$ybw) != dim(ydat)[2])
    stop("length of bandwidth vector does not match number of columns of 'ydat'")
  if (length(bws$xbw) != dim(xdat)[2])
    stop("length of bandwidth vector does not match number of columns of 'xdat'")
  if (dim(ydat)[1] != dim(xdat)[1])
    stop("number of rows of 'ydat' does not match 'xdat'")

  goodrows <- seq_len(nrow(xdat))
  rows.omit <- unclass(na.action(na.omit(data.frame(xdat, ydat))))
  goodrows[rows.omit] <- 0

  xdat <- xdat[goodrows,, drop = FALSE]
  ydat <- ydat[goodrows,, drop = FALSE]

  oydat <- ydat
  ymat <- toMatrix(ydat)
  xmat <- toMatrix(xdat)

  yuno <- ymat[, bws$iyuno, drop = FALSE]
  ycon <- ymat[, bws$iycon, drop = FALSE]
  yord <- ymat[, bws$iyord, drop = FALSE]
  xuno <- xmat[, bws$ixuno, drop = FALSE]
  xcon <- xmat[, bws$ixcon, drop = FALSE]
  xord <- xmat[, bws$ixord, drop = FALSE]

  if (!is.null(gydat)) {
    gydat <- toFrame(gydat)
    if (any(is.na(gydat)))
      stop("na's not allowed to be present in cdf gdata")
    gmat <- toMatrix(gydat)
    gyuno <- gmat[, bws$iyuno, drop = FALSE]
    gyord <- gmat[, bws$iyord, drop = FALSE]
    gycon <- gmat[, bws$iycon, drop = FALSE]
    cdf_on_train <- FALSE
    nog <- nrow(gmat)
  } else if (isTRUE(do.full.integral)) {
    cdf_on_train <- TRUE
    nog <- 0L
    gyuno <- data.frame()
    gyord <- data.frame()
    gycon <- data.frame()
  } else {
    cdf_on_train <- FALSE
    nog <- npValidatePositiveInteger(ngrid, "ngrid")
    probs <- seq(0, 1, length.out = nog)
    evy <- oydat[seq_len(nog),, drop = FALSE]
    for (i in seq_len(ncol(evy)))
      evy[, i] <- uocquantile(oydat[, i], probs)
    evy <- toMatrix(evy)
    gyuno <- evy[, bws$iyuno, drop = FALSE]
    gyord <- evy[, bws$iyord, drop = FALSE]
    gycon <- evy[, bws$iycon, drop = FALSE]
  }

  mysd <- EssDee(data.frame(xcon, ycon))
  nrow <- nrow(ymat)
  nconfac <- nrow^(-1.0 / (2.0 * bws$cxkerorder + bws$ncon))
  ncatfac <- nrow^(-2.0 / (2.0 * bws$cxkerorder + bws$ncon))

  penalty_mode <- if (invalid.penalty == "baseline") 1L else 0L
  reg.code <- if (identical(bws$regtype.engine, "lp")) REGTYPE_LP else REGTYPE_LC
  degree.code <- if (bws$xncon > 0L) as.integer(bws$degree.engine) else integer(0L)
  basis.code <- as.integer(npLpBasisCode(bws$basis.engine))
  bernstein.engine <- isTRUE(bws$bernstein.basis.engine)

  myopti <- list(
    num_obs_train = nrow,
    num_obs_grid = nog,
    iMultistart = IMULTI_FALSE,
    iNum_Multistart = 0L,
    int_use_starting_values = USE_START_YES,
    int_LARGE_SF = if (bws$scaling) SF_NORMAL else SF_ARB,
    BANDWIDTH_den_extern = switch(bws$type,
      fixed = BW_FIXED,
      generalized_nn = BW_GEN_NN,
      adaptive_nn = BW_ADAP_NN),
    itmax = 0L,
    int_RESTART_FROM_MIN = RE_MIN_FALSE,
    int_MINIMIZE_IO = IO_MIN_TRUE,
    bwmethod = CDBWM_CVLS,
    xkerneval = switch(bws$cxkertype,
      gaussian = CKER_GAUSS + bws$cxkerorder/2 - 1,
      epanechnikov = CKER_EPAN + bws$cxkerorder/2 - 1,
      uniform = CKER_UNI,
      "truncated gaussian" = CKER_TGAUSS),
    ykerneval = switch(bws$cykertype,
      gaussian = CKER_GAUSS + bws$cykerorder/2 - 1,
      epanechnikov = CKER_EPAN + bws$cykerorder/2 - 1,
      uniform = CKER_UNI,
      "truncated gaussian" = CKER_TGAUSS),
    uxkerneval = switch(bws$uxkertype,
      aitchisonaitken = UKER_AIT,
      liracine = UKER_LR),
    uykerneval = switch(bws$uykertype,
      aitchisonaitken = UKER_AIT,
      liracine = UKER_LR),
    oxkerneval = switch(bws$oxkertype,
      wangvanryzin = OKER_WANG,
      liracine = OKER_LR,
      racineliyan = OKER_RLY),
    oykerneval = switch(bws$oykertype,
      wangvanryzin = OKER_WANG,
      liracine = OKER_NLR,
      racineliyan = OKER_RLY),
    ynuno = dim(yuno)[2],
    ynord = dim(yord)[2],
    yncon = dim(ycon)[2],
    xnuno = dim(xuno)[2],
    xnord = dim(xord)[2],
    xncon = dim(xcon)[2],
    cdf_on_train = cdf_on_train,
    int_do_tree = if (isTRUE(getOption("np.tree"))) DO_TREE_YES else DO_TREE_NO,
    scale.init.categorical.sample = FALSE,
    dfc.dir = 0L,
    transform.bounds = FALSE
  )

  myoptd <- list(
    ftol = 0,
    tol = 0,
    small = 0,
    memfac = 0,
    lbc.dir = 0,
    cfac.dir = 0,
    initc.dir = 0,
    lbd.dir = 0,
    hbd.dir = 0,
    dfac.dir = 0,
    initd.dir = 0,
    lbc.init = 0,
    hbc.init = 0,
    cfac.init = 0,
    lbd.init = 0,
    hbd.init = 0,
    dfac.init = 0,
    nconfac = nconfac,
    ncatfac = ncatfac,
    scale.factor.lower.bound = npResolveScaleFactorLowerBound(
      npGetScaleFactorSearchLower(bws),
      argname = "bws$scale.factor.search.lower"
    )
  )

  cxker.bounds.c <- npKernelBoundsMarshal(bws$cxkerlb[bws$ixcon], bws$cxkerub[bws$ixcon])
  cyker.bounds.c <- npKernelBoundsMarshal(bws$cykerlb[bws$iycon], bws$cykerub[bws$iycon])

  out <- .Call(
    "C_np_distribution_conditional_bw_eval",
    as.double(yuno),
    as.double(yord),
    as.double(ycon),
    as.double(xuno),
    as.double(xord),
    as.double(xcon),
    as.double(gyuno),
    as.double(gyord),
    as.double(gycon),
    as.double(mysd),
    as.integer(myopti),
    as.double(myoptd),
    as.double(c(bws$xbw[bws$ixcon], bws$ybw[bws$iycon],
                bws$ybw[bws$iyuno], bws$ybw[bws$iyord],
                bws$xbw[bws$ixuno], bws$xbw[bws$ixord])),
    as.integer(1L),
    as.integer(penalty_mode),
    as.double(penalty.multiplier),
    as.integer(degree.code),
    as.integer(bernstein.engine),
    as.integer(basis.code),
    as.integer(reg.code),
    as.double(cxker.bounds.c$lb),
    as.double(cxker.bounds.c$ub),
    as.double(cyker.bounds.c$lb),
    as.double(cyker.bounds.c$ub),
    PACKAGE = "np"
  )

  list(
    objective = as.numeric(out$fval[1L]),
    num.feval = 1L,
    num.feval.fast = as.numeric(as.numeric(out$fast.history[1L]) > 0)
  )
}

.npcdistbw_run_fixed_degree <- function(xdat, ydat, bws, reg.args, opt.args) {
  tbw <- .npcdistbw_build_condbandwidth(
    xdat = xdat,
    ydat = ydat,
    bws = bws,
    bandwidth.compute = opt.args$bandwidth.compute,
    reg.args = reg.args
  )

  do.call(npcdistbw.condbandwidth, c(list(xdat = xdat, ydat = ydat, bws = tbw), opt.args))
}

.npcdistbw_nomad_bw_setup <- function(xdat,
                                      ydat,
                                      template,
                                      bandwidth.scale.categorical = 1e4) {
  xdat <- toFrame(xdat)
  ydat <- toFrame(ydat)
  xmat <- toMatrix(xdat)
  ymat <- toMatrix(ydat)
  xcon <- xmat[, template$ixcon, drop = FALSE]
  ycon <- ymat[, template$iycon, drop = FALSE]
  nrow <- nrow(xmat)
  nconfac <- nrow^(-1.0 / (2.0 * template$cxkerorder + template$ncon))
  ncatfac <- nrow^(-2.0 / (2.0 * template$cxkerorder + template$ncon))

  x_offset <- length(template$ybw)
  y_cont_flat <- which(template$iycon)
  x_cont_flat <- x_offset + which(template$ixcon)
  y_uno_flat <- which(template$iyuno)
  y_ord_flat <- which(template$iyord)
  x_uno_flat <- x_offset + which(template$ixuno)
  x_ord_flat <- x_offset + which(template$ixord)

  cat_upper_one <- function(values, kernel) {
    if (identical(kernel, "aitchisonaitken")) {
      nlev <- length(unique(values))
      return((nlev - 1) / nlev)
    }
    1
  }

  cat_upper <- c(
    if (length(y_uno_flat)) vapply(which(template$iyuno), function(i) cat_upper_one(ydat[[i]], template$uykertype), numeric(1L)) else numeric(0L),
    rep.int(1, length(y_ord_flat)),
    if (length(x_uno_flat)) vapply(which(template$ixuno), function(i) cat_upper_one(xdat[[i]], template$uxkertype), numeric(1L)) else numeric(0L),
    rep.int(1, length(x_ord_flat))
  )

  setup <- list(
    cont_flat = c(y_cont_flat, x_cont_flat),
    cont_scale = .npConditionalNomadContScale(
      ycon = ycon,
      xcon = xcon,
      iycon = template$iycon,
      ixcon = template$ixcon,
      nconfac = nconfac,
      where = "npcdistbw"
    ),
    cat_flat = c(y_uno_flat, y_ord_flat, x_uno_flat, x_ord_flat),
    ncatfac = ncatfac,
    bandwidth.scale.categorical = bandwidth.scale.categorical,
    cat_upper = cat_upper
  )
  .npAssertConditionalNomadSetup(setup, where = "npcdistbw")
  setup
}

.npcdistbw_nomad_point_to_bw <- function(point, template, setup) {
  .npAssertConditionalNomadSetup(setup, where = "npcdistbw")
  point <- as.numeric(point)
  ncont <- length(setup$cont_flat)
  ncat <- length(setup$cat_flat)
  bws <- numeric(length(template$ybw) + length(template$xbw))

  if (ncont > 0L) {
    gamma <- point[seq_len(ncont)]
    ext_bw <- gamma * setup$cont_scale
    bws[setup$cont_flat] <- if (isTRUE(template$scaling)) gamma else ext_bw
  }

  if (ncat > 0L) {
    lambda_scaled <- point[ncont + seq_len(ncat)]
    ext_bw <- lambda_scaled / setup$bandwidth.scale.categorical
    bws[setup$cat_flat] <- if (isTRUE(template$scaling)) ext_bw / setup$ncatfac else ext_bw
  }

  bws
}

.npcdistbw_nomad_bw_to_point <- function(bws, template, setup) {
  .npAssertConditionalNomadSetup(setup, where = "npcdistbw")
  point <- numeric(length(setup$cont_flat) + length(setup$cat_flat))

  if (length(setup$cont_flat) > 0L) {
    raw <- bws[setup$cont_flat]
    point[seq_along(setup$cont_flat)] <- if (isTRUE(template$scaling)) {
      raw
    } else {
      raw / setup$cont_scale
    }
  }

  if (length(setup$cat_flat) > 0L) {
    raw <- bws[setup$cat_flat]
    ext_bw <- if (isTRUE(template$scaling)) raw * setup$ncatfac else raw
    point[length(setup$cont_flat) + seq_along(setup$cat_flat)] <- ext_bw * setup$bandwidth.scale.categorical
  }

  point
}

.npcdistbw_nomad_search <- function(xdat,
                                    ydat,
                                    bws,
                                    reg.args,
                                    opt.args,
                                    degree.search,
                                    nomad.inner.nmulti = 0L,
                                    random.seed = 42L) {
  if (isTRUE(degree.search$verify))
    stop("automatic degree search with search.engine='nomad' does not support degree.verify")

  template.reg.args <- reg.args
  template.reg.args$regtype <- "lp"
  template.reg.args$pregtype <- "Local-Polynomial"
  template.reg.args$degree <- as.integer(degree.search$start.degree)
  template.reg.args$bernstein.basis <- degree.search$bernstein.basis
  template.reg.args$regtype.engine <- "lp"
  template.reg.args$degree.engine <- as.integer(degree.search$start.degree)
  template.reg.args$bernstein.basis.engine <- degree.search$bernstein.basis

  template <- .npcdistbw_build_condbandwidth(
    xdat = xdat,
    ydat = ydat,
    bws = bws,
    bandwidth.compute = FALSE,
    reg.args = template.reg.args
  )

  if (!identical(template$type, "fixed"))
    stop("automatic degree search with search.engine='nomad' currently requires bwtype='fixed'")

  setup <- .npcdistbw_nomad_bw_setup(xdat = xdat, ydat = ydat, template = template)
  bwdim <- length(setup$cont_flat) + length(setup$cat_flat)
  ndeg <- length(degree.search$start.degree)
  nomad.nmulti <- if (is.null(opt.args$nmulti)) npDefaultNmulti(dim(ydat)[2]+dim(xdat)[2]) else npValidateNmulti(opt.args$nmulti[1L])
  cont_lower <- npGetScaleFactorSearchLower(
    template,
    argname = "template$scale.factor.search.lower"
  )
  bw_lower <- c(rep.int(cont_lower, length(setup$cont_flat)), rep.int(0, length(setup$cat_flat)))
  bw_upper <- c(rep.int(1e6, length(setup$cont_flat)), setup$cat_upper * setup$bandwidth.scale.categorical)

  x0 <- c(
    .np_nomad_complete_start_point(
      point = {
        raw <- c(template$ybw, template$xbw)
        if (all(raw == 0)) NULL else .npcdistbw_nomad_bw_to_point(raw, template = template, setup = setup)
      },
      lower = bw_lower,
      upper = bw_upper,
      ncont = length(setup$cont_flat)
    ),
    as.integer(degree.search$start.degree)
  )
  lb <- c(bw_lower, degree.search$lower)
  ub <- c(bw_upper, degree.search$upper)
  bbin <- c(rep.int(0L, bwdim), rep.int(1L, ndeg))
  baseline.record <- NULL
  nomad.num.feval.total <- 0
  nomad.num.feval.fast.total <- 0

  .np_nomad_baseline_note(degree.search$start.degree)

  eval_fun <- function(point) {
    point <- as.numeric(point)
    degree <- as.integer(round(point[bwdim + seq_len(ndeg)]))
    degree <- .np_degree_clip_to_grid(degree, degree.search$candidates)
    bw_vec <- .npcdistbw_nomad_point_to_bw(point[seq_len(bwdim)], template = template, setup = setup)

    eval.reg.args <- reg.args
    eval.reg.args$regtype <- "lp"
    eval.reg.args$pregtype <- "Local-Polynomial"
    eval.reg.args$degree <- degree
    eval.reg.args$bernstein.basis <- degree.search$bernstein.basis
    eval.reg.args$regtype.engine <- "lp"
    eval.reg.args$degree.engine <- degree
    eval.reg.args$bernstein.basis.engine <- degree.search$bernstein.basis

    tbw <- .npcdistbw_build_condbandwidth(
      xdat = xdat,
      ydat = ydat,
      bws = bw_vec,
      bandwidth.compute = FALSE,
      reg.args = eval.reg.args
    )

    out <- .npcdistbw_eval_only(
      xdat = xdat,
      ydat = ydat,
      gydat = opt.args$gydat,
      bws = tbw,
      do.full.integral = if (is.null(opt.args$do.full.integral)) FALSE else opt.args$do.full.integral,
      ngrid = if (is.null(opt.args$ngrid)) 100L else opt.args$ngrid,
      invalid.penalty = "baseline",
      penalty.multiplier = if (is.null(opt.args$penalty.multiplier)) 10 else opt.args$penalty.multiplier
    )
    nomad.num.feval.total <<- nomad.num.feval.total + as.numeric(out$num.feval[1L])
    nomad.num.feval.fast.total <<- nomad.num.feval.fast.total + as.numeric(out$num.feval.fast[1L])

    list(
      objective = out$objective,
      degree = degree,
      num.feval = out$num.feval
    )
  }

  build_payload <- function(point, best_record, solution, interrupted) {
    point <- as.numeric(point)
    degree <- as.integer(best_record$degree)
    bw_vec <- .npcdistbw_nomad_point_to_bw(point[seq_len(bwdim)], template = template, setup = setup)
    powell.elapsed <- NA_real_

    build_direct_payload <- function() {
      final.reg.args <- reg.args
      final.reg.args$regtype <- "lp"
      final.reg.args$pregtype <- "Local-Polynomial"
      final.reg.args$degree <- degree
      final.reg.args$bernstein.basis <- degree.search$bernstein.basis
      final.reg.args$regtype.engine <- "lp"
      final.reg.args$degree.engine <- degree
      final.reg.args$bernstein.basis.engine <- degree.search$bernstein.basis

      tbw <- .npcdistbw_build_condbandwidth(
        xdat = xdat,
        ydat = ydat,
        bws = bw_vec,
        bandwidth.compute = FALSE,
        reg.args = final.reg.args
      )
      tbw$fval <- as.numeric(best_record$objective)
      tbw$ifval <- as.numeric(best_record$objective)
      tbw$num.feval <- as.numeric(nomad.num.feval.total)
      tbw$num.feval.fast <- as.numeric(nomad.num.feval.fast.total)
      tbw$fval.history <- as.numeric(best_record$objective)
      tbw$eval.history <- if (!is.null(solution$bbe)) rep(1, max(1L, as.integer(solution$bbe))) else 1
      tbw$invalid.history <- 0
      tbw$timing <- NA_real_
      tbw$total.time <- NA_real_

      payload <- npcdistbw.condbandwidth(
        xdat = xdat,
        ydat = ydat,
        bws = tbw,
        bandwidth.compute = FALSE
      )
      if (!is.null(payload$method) && length(payload$method))
        payload$pmethod <- bwmToPrint(as.character(payload$method[1L]))
      payload
    }

    direct.payload <- build_direct_payload()
    direct.objective <- as.numeric(best_record$objective)

    if (identical(degree.search$engine, "nomad+powell")) {
      hot.reg.args <- reg.args
      hot.reg.args$regtype <- "lp"
      hot.reg.args$pregtype <- "Local-Polynomial"
      hot.reg.args$degree <- degree
      hot.reg.args$bernstein.basis <- degree.search$bernstein.basis
      hot.reg.args$regtype.engine <- "lp"
      hot.reg.args$degree.engine <- degree
      hot.reg.args$bernstein.basis.engine <- degree.search$bernstein.basis
      hot.opt.args <- .np_nomad_powell_hotstart_opt_args(
        opt.args,
        strategy = "disable_multistart",
        remin = isTRUE(opt.args$powell.remin)
      )
      powell.start <- proc.time()[3L]
      hot.payload <- .np_nomad_with_powell_progress(
        degree = degree,
        best_record = best_record,
        expr = .npcdistbw_run_fixed_degree(
          xdat = xdat,
          ydat = ydat,
          bws = bw_vec,
          reg.args = hot.reg.args,
          opt.args = hot.opt.args
        )
      )
      powell.elapsed <- proc.time()[3L] - powell.start
      direct.payload$num.feval <- as.numeric(direct.payload$num.feval[1L]) + as.numeric(hot.payload$num.feval[1L])
      direct.payload$num.feval.fast <- as.numeric(direct.payload$num.feval.fast[1L]) + as.numeric(hot.payload$num.feval.fast[1L])
      hot.payload$num.feval <- direct.payload$num.feval
      hot.payload$num.feval.fast <- direct.payload$num.feval.fast
      if (!is.null(hot.payload$method) && length(hot.payload$method))
        hot.payload$pmethod <- bwmToPrint(as.character(hot.payload$method[1L]))
      hot.objective <- as.numeric(hot.payload$fval[1L])
      if (is.finite(hot.objective) &&
          .np_degree_better(hot.objective, direct.objective, direction = "min")) {
        return(list(payload = hot.payload, objective = hot.objective, powell.time = powell.elapsed))
      }
    }

    list(payload = direct.payload, objective = direct.objective, powell.time = powell.elapsed)
  }

  .np_nomad_search(
    engine = degree.search$engine,
    baseline_record = baseline.record,
    start_degree = degree.search$start.degree,
    x0 = x0,
    bbin = bbin,
    lb = lb,
    ub = ub,
    eval_fun = eval_fun,
    build_payload = build_payload,
    direction = "min",
    objective_name = "fval",
    nmulti = nomad.nmulti,
    nomad.inner.nmulti = nomad.inner.nmulti,
    random.seed = random.seed,
    remin = isTRUE(opt.args$nomad.remin),
    nomad.opts = list(
      DIRECTION_TYPE = "ORTHO 2N",
      QUAD_MODEL_SEARCH = "no",
      NM_SEARCH = "no",
      SPECULATIVE_SEARCH = "no",
      EVAL_OPPORTUNISTIC = "no"
    ),
    degree_spec = list(
      initial = degree.search$start.degree,
      lower = degree.search$lower,
      upper = degree.search$upper,
      basis = degree.search$basis,
      nobs = degree.search$nobs,
      user_supplied = degree.search$start.user
    )
  )
}

.npcdistbw_degree_search_controls <- function(regtype,
                                              regtype.named,
                                              ncon,
                                              nobs,
                                              basis,
                                              degree.select,
                                              search.engine,
                                              degree.min,
                                              degree.max,
                                              degree.start,
                                              degree.restarts,
                                              degree.max.cycles,
                                              degree.verify,
                                              bernstein.basis,
                                              bernstein.named) {
  degree.select <- match.arg(degree.select, c("manual", "coordinate", "exhaustive"))
  if (identical(degree.select, "manual"))
    return(NULL)
  search.engine <- .np_degree_search_engine_controls(search.engine)

  regtype.requested <- if (isTRUE(regtype.named)) match.arg(regtype, c("lc", "ll", "lp")) else "lc"
  if (!identical(regtype.requested, "lp"))
    stop("automatic degree search currently requires regtype='lp'")
  if (ncon < 1L)
    stop("automatic degree search requires at least one continuous conditioning predictor")

  bern.auto <- if (isTRUE(bernstein.named)) bernstein.basis else TRUE
  bern.auto <- npValidateGlpBernstein(regtype = "lp", bernstein.basis = bern.auto)

  bounds <- .np_degree_normalize_bounds(
    ncon = ncon,
    degree.min = degree.min,
    degree.max = degree.max,
    default.max = 3L
  )

  baseline.degree <- rep.int(0L, ncon)
  # Density/distribution degree search should anchor on the local-constant
  # baseline unless the user explicitly requests another start.
  default.start.degree <- baseline.degree
  start.degree <- if (is.null(degree.start)) {
    pmax(bounds$lower, pmin(bounds$upper, default.start.degree))
  } else {
    start.raw <- npValidateGlpDegree(regtype = "lp", degree = degree.start, ncon = ncon, argname = "degree.start")
    out.of.range <- vapply(seq_len(ncon), function(j) !(start.raw[j] %in% bounds$candidates[[j]]), logical(1))
    if (any(out.of.range))
      stop("degree.start must lie within the searched degree candidates for every continuous conditioning predictor")
    start.raw
  }

  list(
    method = if (identical(search.engine, "cell")) degree.select else search.engine,
    engine = search.engine,
    candidates = bounds$candidates,
    lower = bounds$lower,
    upper = bounds$upper,
    baseline.degree = baseline.degree,
    start.degree = start.degree,
    start.user = !is.null(degree.start),
    basis = if (missing(basis) || is.null(basis)) "glp" else as.character(basis[1L]),
    nobs = as.integer(nobs[1L]),
    restarts = npValidateNonNegativeInteger(degree.restarts, "degree.restarts"),
    max.cycles = npValidatePositiveInteger(degree.max.cycles, "degree.max.cycles"),
    verify = npValidateScalarLogical(degree.verify, "degree.verify"),
    bernstein.basis = bern.auto
  )
}

.npcdistbw_attach_degree_search <- function(bws, search_result) {
  metadata <- list(
    mode = search_result$method,
    verify = isTRUE(search_result$verify),
    completed = isTRUE(search_result$completed),
    certified = isTRUE(search_result$certified),
    interrupted = isTRUE(search_result$interrupted),
    baseline.degree = search_result$baseline$degree,
    baseline.fval = search_result$baseline$objective,
    best.degree = search_result$best$degree,
    best.fval = search_result$best$objective,
    nomad.time = search_result$nomad.time,
    powell.time = search_result$powell.time,
    optim.time = search_result$optim.time,
    n.unique = search_result$n.unique,
    n.visits = search_result$n.visits,
    n.cached = search_result$n.cached,
    grid.size = search_result$grid.size,
    best.restart = search_result$best.restart,
    restart.starts = search_result$restart.starts,
    restart.degree.starts = search_result$restart.degree.starts,
    restart.bandwidth.starts = search_result$restart.bandwidth.starts,
    restart.start.info = search_result$restart.start.info,
    restart.results = search_result$restart.results,
    trace = search_result$trace
  )

  if (!is.null(search_result$nomad.time))
    bws$nomad.time <- as.numeric(search_result$nomad.time[1L])
  if (!is.null(search_result$powell.time))
    bws$powell.time <- as.numeric(search_result$powell.time[1L])
  if (!is.null(search_result$optim.time) && is.finite(search_result$optim.time))
    bws$total.time <- as.numeric(search_result$optim.time[1L])
  bws <- .np_attach_nomad_restart_summary(bws, search_result)
  bws$degree.search <- metadata
  bws
}

npcdistbw.NULL <-
  function(xdat = stop("data 'xdat' missing"),
           ydat = stop("data 'ydat' missing"),
           bws, ...){

    ## maintain x names and 'toFrame'
    xdat <- toFrame(xdat)

    ## maintain y names and 'toFrame'
    ydat <- toFrame(ydat)

    ## do bandwidths
    
    bws = double(ncol(ydat)+ncol(xdat))

    tbw <- npcdistbw.default(xdat = xdat, ydat = ydat, bws = bws, ...)

    ## clean up (possible) inconsistencies due to recursion ...
    mc <- match.call(expand.dots = FALSE)
    environment(mc) <- parent.frame()
    tbw$call <- mc

    tbw
  }

npcdistbw.default <-
  function(xdat = stop("data 'xdat' missing"),
           ydat = stop("data 'ydat' missing"),
           gydat,
           bws, 
           bandwidth.compute = TRUE,
           bwmethod,
           bwscaling,
           bwtype,
           cfac.dir,
           scale.factor.init,
           cxkerbound,
           cxkerlb,
           cxkerorder,
           cxkertype,
           cxkerub,
           cykerbound,
           cykerlb,
           cykerorder,
           cykertype,
           cykerub,
           dfac.dir,
           dfac.init,
           dfc.dir,
           do.full.integral,
           ftol,
           scale.factor.init.upper,
           hbd.dir,
           hbd.init,
           initc.dir,
           initd.dir,
           invalid.penalty,
           itmax,
           lbc.dir,
           scale.factor.init.lower,
           lbd.dir,
           lbd.init,
           memfac,
           ngrid,
           nmulti,
           oxkertype,
           oykertype,
           penalty.multiplier,
           nomad.remin = FALSE,
           powell.remin,
           scale.init.categorical.sample,
           scale.factor.search.lower = NULL,
           small,
           tol,
           transform.bounds,
           uxkertype,
           regtype = c("lc", "ll", "lp"),
           basis = c("glp", "additive", "tensor"),
           degree = NULL,
           degree.select = c("manual", "coordinate", "exhaustive"),
           search.engine = c("nomad+powell", "cell", "nomad"),
           nomad = FALSE,
           nomad.nmulti = 0L,
           degree.min = NULL,
           degree.max = NULL,
           degree.start = NULL,
           degree.restarts = 0L,
           degree.max.cycles = 20L,
           degree.verify = FALSE,
           bernstein.basis = FALSE,
           ## dummy arguments for condbandwidth() function call
           ...){

    ## maintain x names and 'toFrame'
    xdat <- toFrame(xdat)

    ## maintain y names and 'toFrame'
    ydat <- toFrame(ydat)

    x.info <- untangle(xdat)
    y.info <- untangle(ydat)

    mc <- match.call(expand.dots = FALSE)
    mc.names <- names(mc)
    nomad.shortcut <- .np_prepare_nomad_shortcut(
      nomad = nomad,
      call_names = mc.names,
      preset = list(
        regtype = "lp",
        search.engine = "nomad+powell",
        degree.select = "coordinate",
        bernstein.basis = TRUE,
        degree.min = 0L,
        degree.max = 10L,
        degree.verify = FALSE,
        bwtype = "fixed"
      ),
      values = list(
        regtype = if ("regtype" %in% mc.names) regtype else NULL,
        search.engine = if ("search.engine" %in% mc.names) search.engine else NULL,
        degree.select = if ("degree.select" %in% mc.names) degree.select else NULL,
        bernstein.basis = if ("bernstein.basis" %in% mc.names) bernstein.basis else NULL,
        degree.min = if ("degree.min" %in% mc.names) degree.min else NULL,
        degree.max = if ("degree.max" %in% mc.names) degree.max else NULL,
        degree.verify = if ("degree.verify" %in% mc.names) degree.verify else NULL,
        bwtype = if ("bwtype" %in% mc.names) bwtype else NULL,
        degree = if ("degree" %in% mc.names) degree else NULL
      ),
      where = "npcdistbw"
    )

    if (isTRUE(nomad.shortcut$enabled)) {
      if (sum(x.info$icon) == 0L)
        stop("nomad=TRUE requires at least one continuous predictor for degree search",
             call. = FALSE)
      if ("degree" %in% mc.names)
        stop("nomad=TRUE does not support an explicit degree; remove degree or set nomad=FALSE")
      if ("regtype" %in% mc.names &&
          !identical(as.character(match.arg(nomad.shortcut$values$regtype, c("lc", "ll", "lp")))[1L], "lp"))
        stop("nomad=TRUE requires regtype='lp'")
      if ("bwtype" %in% mc.names &&
          !identical(as.character(match.arg(nomad.shortcut$values$bwtype, c("fixed", "generalized_nn", "adaptive_nn")))[1L], "fixed"))
        stop("nomad=TRUE currently requires bwtype='fixed'")
      if ("degree.select" %in% mc.names &&
          identical(as.character(match.arg(nomad.shortcut$values$degree.select, c("manual", "coordinate", "exhaustive")))[1L], "manual"))
        stop("nomad=TRUE requires automatic degree search; use degree.select='coordinate' or 'exhaustive'")
      if ("search.engine" %in% mc.names &&
          !(as.character(match.arg(nomad.shortcut$values$search.engine, c("nomad+powell", "cell", "nomad")))[1L] %in%
              c("nomad", "nomad+powell")))
        stop("nomad=TRUE requires search.engine='nomad' or 'nomad+powell'")
      if ("bernstein.basis" %in% mc.names &&
          !isTRUE(npValidateGlpBernstein(regtype = "lp",
                                        bernstein.basis = nomad.shortcut$values$bernstein.basis)))
        stop("nomad=TRUE currently requires bernstein.basis=TRUE")
      if ("degree.verify" %in% mc.names &&
          isTRUE(npValidateScalarLogical(nomad.shortcut$values$degree.verify, "degree.verify")))
        stop("nomad=TRUE currently requires degree.verify=FALSE")
    }

    regtype.named <- isTRUE(nomad.shortcut$enabled) || any(mc.names == "regtype")
    basis.named <- any(mc.names == "basis")
    degree.named <- any(mc.names == "degree")
    bernstein.named <- isTRUE(nomad.shortcut$enabled) || any(mc.names == "bernstein.basis")

    regtype <- if (!is.null(nomad.shortcut$values$regtype)) {
      match.arg(nomad.shortcut$values$regtype, c("lc", "ll", "lp"))
    } else {
      "lc"
    }
    if (identical(regtype, "lc") && (basis.named || degree.named || bernstein.named))
      stop("regtype='lc' does not accept basis/degree/bernstein.basis; use regtype='lp' for local-polynomial controls")
    if (identical(regtype, "ll")) {
      if (degree.named)
        stop("regtype='ll' uses canonical LP(degree=1, basis='glp'); remove 'degree' or use regtype='lp'")
      if (basis.named && !identical(match.arg(basis), "glp"))
        stop("regtype='ll' uses canonical basis='glp'; use regtype='lp' for alternate LP bases")
      if (bernstein.named && isTRUE(bernstein.basis))
        stop("regtype='ll' uses canonical bernstein.basis=FALSE; use regtype='lp' for Bernstein LP")
    }

    bernstein.value <- if (!is.null(nomad.shortcut$values$bernstein.basis)) {
      nomad.shortcut$values$bernstein.basis
    } else {
      bernstein.basis
    }
    degree.select.value <- if (!is.null(nomad.shortcut$values$degree.select)) nomad.shortcut$values$degree.select else "manual"
    degree.setup <- npSetupGlpDegree(
      regtype = regtype,
      degree = degree,
      ncon = sum(x.info$icon),
      degree.select = degree.select.value
    )
    spec <- npCanonicalConditionalRegSpec(
      regtype = regtype,
      basis = basis,
      degree = degree.setup,
      bernstein.basis = bernstein.value,
      ncon = sum(x.info$icon),
      where = "npcdistbw"
    )
    pregtype <- switch(spec$regtype,
                       lc = "Local-Constant",
                       ll = "Local-Linear",
                       lp = "Local-Polynomial")

    search.mc.names <- names(mc)
    lp.dot.args <- list(...)
    .np_degree_reject_unknown_dots(lp.dot.args, "npcdistbw")
    random.seed.value <- .np_degree_extract_random_seed(lp.dot.args)
    search.engine.value <- if (!is.null(nomad.shortcut$values$search.engine)) nomad.shortcut$values$search.engine else "nomad+powell"
    scale.factor.search.lower <- npResolveScaleFactorLowerBound(scale.factor.search.lower)
    degree.min.value <- nomad.shortcut$values$degree.min
    degree.max.value <- nomad.shortcut$values$degree.max
    degree.start.value <- if ("degree.start" %in% search.mc.names) degree.start else NULL
    degree.restarts.value <- if ("degree.restarts" %in% search.mc.names) degree.restarts else 0L
    degree.max.cycles.value <- if ("degree.max.cycles" %in% search.mc.names) degree.max.cycles else 20L
    degree.verify.value <- if (!is.null(nomad.shortcut$values$degree.verify)) nomad.shortcut$values$degree.verify else FALSE
    degree.search <- .npcdistbw_degree_search_controls(
      regtype = regtype,
      regtype.named = regtype.named,
      ncon = sum(x.info$icon),
      nobs = NROW(xdat),
      basis = if (basis.named) basis else "glp",
      degree.select = degree.select.value,
      search.engine = search.engine.value,
      degree.min = degree.min.value,
      degree.max = degree.max.value,
      degree.start = degree.start.value,
      degree.restarts = degree.restarts.value,
      degree.max.cycles = degree.max.cycles.value,
      degree.verify = degree.verify.value,
      bernstein.basis = bernstein.value,
      bernstein.named = bernstein.named
    )
    nomad.inner.named <- "nomad.nmulti" %in% search.mc.names
    nomad.inner.nmulti <- if (nomad.inner.named) {
      npValidateNonNegativeInteger(nomad.nmulti, "nomad.nmulti")
    } else {
      0L
    }
    if (nomad.inner.named &&
        (is.null(degree.search) || !(degree.search$engine %in% c("nomad", "nomad+powell")))) {
      stop("nomad.nmulti is only supported when regtype='lp', automatic degree search is active, and search.engine is 'nomad' or 'nomad+powell'")
    }

    if (!is.null(degree.search)) {
      spec$bernstein.basis <- degree.search$bernstein.basis
      spec$bernstein.basis.engine <- degree.search$bernstein.basis
    }

    ## first grab dummy args for bandwidth() and perform 'bootstrap'
    ## bandwidth() call

    mc.names <- names(mc)
    margs <- c("bwmethod", "bwscaling", "bwtype", "cxkertype", "cxkerorder",
               "cxkerbound", "cxkerlb", "cxkerub",
               "cykertype", "cykerorder", "cykerbound", "cykerlb", "cykerub",
               "uxkertype", "oxkertype", "oykertype")

    m <- match(margs, mc.names, nomatch = 0)
    any.m <- any(m != 0)

    y.idx <- seq_len(length(ydat))
    x.idx <- seq_len(length(xdat))
    bw.args <- list(
      xbw = bws[length(ydat) + x.idx],
      ybw = bws[y.idx],
      uykertype = "aitchisonaitken",
      nobs = nrow(xdat),
      xdati = x.info,
      ydati = y.info,
      xnames = names(xdat),
      ynames = names(ydat),
      bandwidth.compute = bandwidth.compute,
      regtype = spec$regtype,
      pregtype = pregtype,
      basis = spec$basis,
      degree = spec$degree,
      bernstein.basis = spec$bernstein.basis,
      regtype.engine = spec$regtype.engine,
      basis.engine = spec$basis.engine,
      degree.engine = spec$degree.engine,
      bernstein.basis.engine = spec$bernstein.basis.engine
    )
    if (any.m) {
      nms <- mc.names[m]
      bw.args[nms] <- mget(nms, envir = environment(), inherits = FALSE)
    }
    reg.args <- bw.args[setdiff(names(bw.args), c("xbw", "ybw", "nobs", "xdati", "ydati", "xnames", "ynames", "bandwidth.compute"))]
                        
    ## next grab dummies for actual bandwidth selection and perform call

    mc.names <- names(mc)
    margs <- c("gydat", "nmulti", "nomad.remin", "powell.remin", "itmax", "do.full.integral", "ngrid", "ftol",
               "tol", "small", "memfac",
               "lbc.dir", "dfc.dir", "cfac.dir","initc.dir", 
               "lbd.dir", "hbd.dir", "dfac.dir", "initd.dir", 
               "scale.factor.init.lower", "scale.factor.init.upper", "scale.factor.init", 
               "lbd.init", "hbd.init", "dfac.init", 
               "scale.factor.search.lower",
               "scale.init.categorical.sample",
               "transform.bounds",
               "invalid.penalty",
               "penalty.multiplier")
    m <- match(margs, mc.names, nomatch = 0)
    any.m <- any(m != 0)

    if (any.m) {
      nms <- mc.names[m]
      opt.args <- mget(nms, envir = environment(), inherits = FALSE)
    } else {
      opt.args <- list()
    }
    opt.args <- c(list(bandwidth.compute = bandwidth.compute), opt.args)
    reg.args$scale.factor.search.lower <- scale.factor.search.lower
    opt.args$scale.factor.search.lower <- scale.factor.search.lower

    if (!is.null(degree.search)) {
      if (identical(degree.search$engine, "cell")) {
        eval_fun <- function(degree.vec) {
          cell.reg.args <- reg.args
          cell.reg.args$regtype <- "lp"
          cell.reg.args$pregtype <- "Local-Polynomial"
          cell.reg.args$degree <- as.integer(degree.vec)
          cell.reg.args$bernstein.basis <- degree.search$bernstein.basis
          cell.reg.args$regtype.engine <- "lp"
          cell.reg.args$degree.engine <- as.integer(degree.vec)
          cell.reg.args$bernstein.basis.engine <- degree.search$bernstein.basis
          cell.bws <- .npcdistbw_run_fixed_degree(
            xdat = xdat,
            ydat = ydat,
            bws = bws,
            reg.args = cell.reg.args,
            opt.args = opt.args
          )
          list(
            objective = as.numeric(cell.bws$fval[1L]),
            payload = cell.bws,
            num.feval = if (!is.null(cell.bws$num.feval)) as.numeric(cell.bws$num.feval[1L]) else NA_real_
          )
        }

        search.result <- .np_degree_search(
          method = degree.search$method,
          candidates = degree.search$candidates,
          baseline_degree = degree.search$baseline.degree,
          start_degree = degree.search$start.degree,
          restarts = degree.search$restarts,
          max_cycles = degree.search$max.cycles,
          verify = degree.search$verify,
          eval_fun = eval_fun,
          direction = "min",
          trace_level = "full",
          objective_name = "fval"
        )
      } else {
        search.result <- .npcdistbw_nomad_search(
          xdat = xdat,
          ydat = ydat,
          bws = bws,
          reg.args = reg.args,
          opt.args = opt.args,
          degree.search = degree.search,
          nomad.inner.nmulti = nomad.inner.nmulti,
          random.seed = random.seed.value
        )
      }
      tbw <- .npcdistbw_attach_degree_search(
        bws = search.result$best_payload,
        search_result = search.result
      )
    } else {
      tbw <- .npcdistbw_build_condbandwidth(
        xdat = xdat,
        ydat = ydat,
        bws = bws,
        bandwidth.compute = bandwidth.compute,
        reg.args = reg.args
      )
      bwsel.args <- c(list(xdat = xdat, ydat = ydat, bws = tbw), opt.args)
      tbw <- .np_progress_select_bandwidth_enhanced(
        "Selecting conditional distribution bandwidth",
        do.call(npcdistbw.condbandwidth, bwsel.args)
      )
    }

    mc <- match.call(expand.dots = FALSE)
    environment(mc) <- parent.frame()
    tbw$call <- mc
    tbw <- .np_attach_nomad_shortcut(tbw, nomad.shortcut$metadata)

    return(tbw)
  }

library(np)

quiet_nomad_eval <- function(expr) {
  out <- NULL
  capture.output(out <- force(expr))
  out
}

expect_nomad_bw_timing <- function(bw) {
  expect_true(is.finite(bw$nomad.time))
  expect_true(is.finite(bw$powell.time))
  expect_true(is.finite(bw$total.time))
  expect_equal(as.double(bw$total.time),
               as.double(bw$nomad.time + bw$powell.time),
               tolerance = 1e-8)
  expect_equal(as.double(bw$degree.search$optim.time),
               as.double(bw$nomad.time + bw$powell.time),
               tolerance = 1e-8)
}

expect_nomad_fit_timing <- function(fit) {
  expect_true(is.finite(fit$bws$nomad.time))
  expect_true(is.finite(fit$bws$powell.time))
  expect_true(is.finite(fit$bws$total.time))
  expect_equal(as.double(fit$bws$total.time),
               as.double(fit$bws$nomad.time + fit$bws$powell.time),
               tolerance = 1e-8)

  if (!is.null(fit$optim.time) && length(fit$optim.time))
    expect_equal(as.double(fit$optim.time),
                 as.double(fit$bws$total.time),
                 tolerance = 1e-8)

  if (!is.null(fit$total.time) && length(fit$total.time) &&
      !is.null(fit$fit.time) && length(fit$fit.time) &&
      !is.null(fit$optim.time) && length(fit$optim.time))
    expect_equal(as.double(fit$total.time),
                 as.double(fit$optim.time + fit$fit.time),
                 tolerance = 1e-8)
}

test_that("NOMAD timing is consistent across supported serial families", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260401)
  d1 <- data.frame(x = sort(runif(12)))
  d1$y <- d1$x + 0.35 * d1$x^2 + rnorm(nrow(d1), sd = 0.12)

  d2 <- data.frame(x = runif(12), z = runif(12, -1, 1))
  d2$y <- 1 + d2$x + sin(pi * d2$z) + rnorm(nrow(d2), sd = 0.18)

  d3 <- data.frame(x1 = runif(12, -1, 1), x2 = runif(12, -1, 1))
  idx <- d3$x1 + 0.75 * d3$x2
  d3$y <- sin(idx) + 0.25 * idx^2 + rnorm(nrow(d3), sd = 0.08)

  bw_reg <- quiet_nomad_eval(npregbw(
    y ~ x,
    data = d1,
    regtype = "lp",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  ))
  fit_reg <- quiet_nomad_eval(npreg(y ~ x, data = d1, nomad = TRUE, degree.max = 1L, nmulti = 1L))
  expect_nomad_bw_timing(bw_reg)
  expect_nomad_fit_timing(fit_reg)

  bw_cdens <- quiet_nomad_eval(npcdensbw(
    y ~ x,
    data = d1,
    regtype = "lp",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  ))
  fit_cdens <- quiet_nomad_eval(npcdens(y ~ x, data = d1, nomad = TRUE, degree.max = 1L, nmulti = 1L))
  expect_nomad_bw_timing(bw_cdens)
  expect_nomad_fit_timing(fit_cdens)

  bw_cdist <- quiet_nomad_eval(npcdistbw(
    y ~ x,
    data = d1,
    regtype = "lp",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  ))
  fit_cdist <- quiet_nomad_eval(npcdist(y ~ x, data = d1, nomad = TRUE, degree.max = 1L, nmulti = 1L, ngrid = 7L))
  expect_nomad_bw_timing(bw_cdist)
  expect_nomad_fit_timing(fit_cdist)

  bw_pl <- quiet_nomad_eval(npplregbw(
    y ~ x | z,
    data = d2,
    regtype = "lp",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  ))
  fit_pl <- quiet_nomad_eval(npplreg(y ~ x | z, data = d2, nomad = TRUE, degree.max = 1L, nmulti = 1L))
  expect_nomad_bw_timing(bw_pl)
  expect_nomad_fit_timing(fit_pl)

  bw_sc <- quiet_nomad_eval(npscoefbw(
    y ~ x | z,
    data = d2,
    regtype = "lp",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  ))
  fit_sc <- quiet_nomad_eval(npscoef(y ~ x | z, data = d2, nomad = TRUE, degree.max = 1L, nmulti = 1L))
  expect_nomad_bw_timing(bw_sc)
  expect_nomad_fit_timing(fit_sc)

  bw_si <- quiet_nomad_eval(npindexbw(
    y ~ x1 + x2,
    data = d3,
    method = "ichimura",
    regtype = "lp",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    nmulti = 1L
  ))
  fit_si <- quiet_nomad_eval(npindex(y ~ x1 + x2, data = d3, method = "ichimura", nomad = TRUE, degree.max = 1L, nmulti = 1L))
  expect_nomad_bw_timing(bw_si)
  expect_nomad_fit_timing(fit_si)
})

library(np)

test_that("bounded cv.ls remains finite for gaussian order 2 and 4", {
  set.seed(20260224)
  x <- runif(80)
  dat <- data.frame(x = x)

  for (ord in c(2L, 4L)) {
    bw <- npudensbw(
      dat = dat,
      bwmethod = "cv.ls",
      bwtype = "fixed",
      ckertype = "gaussian",
      ckerorder = ord,
      ckerbound = "range",
      nmulti = 1
    )

    expect_true(is.finite(as.numeric(bw$bw[1])))
    expect_true(is.finite(as.numeric(bw$fval)))
  }
})

test_that("bounded unconditional cv.ls scalar quadrature supports generalized and adaptive NN bwtypes", {
  set.seed(20260421)
  dat <- data.frame(x = runif(48L))

  bw.gnn <- npudensbw(
    dat = dat,
    bwmethod = "cv.ls",
    bwtype = "generalized_nn",
    ckertype = "epanechnikov",
    ckerorder = 6L,
    ckerbound = "range",
    nmulti = 1
  )

  bw.ad <- npudensbw(
    dat = dat,
    bwmethod = "cv.ls",
    bwtype = "adaptive_nn",
    ckertype = "gaussian",
    ckerorder = 8L,
    ckerbound = "range",
    nmulti = 1
  )

  expect_true(is.finite(as.numeric(bw.gnn$bw[1])))
  expect_true(is.finite(as.numeric(bw.gnn$fval)))
  expect_true(is.finite(as.numeric(bw.ad$bw[1])))
  expect_true(is.finite(as.numeric(bw.ad$fval)))
})

test_that("bounded unconditional cv.ls admits mixed and bivariate continuous bounded data", {
  set.seed(20260421)
  mixed <- data.frame(x = runif(24L), g = factor(sample(c("a", "b"), 24L, replace = TRUE)))
  multi <- data.frame(x1 = runif(24L), x2 = runif(24L))
  combo <- data.frame(
    x1 = runif(24L),
    x2 = runif(24L),
    u = factor(sample(c("a", "b"), 24L, replace = TRUE)),
    o = ordered(sample(1:3, 24L, replace = TRUE))
  )

  bw.mixed <- npudensbw(
    dat = mixed,
    bwmethod = "cv.ls",
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1
  )

  bw.multi <- npudensbw(
    dat = multi,
    bwmethod = "cv.ls",
    bwtype = "adaptive_nn",
    ckerbound = "range",
    nmulti = 1
  )

  bw.combo <- npudensbw(
    dat = combo,
    bwmethod = "cv.ls",
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1
  )

  expect_true(all(is.finite(as.numeric(bw.mixed$bw))))
  expect_true(is.finite(as.numeric(bw.mixed$fval)))
  expect_true(all(is.finite(as.numeric(bw.multi$bw))))
  expect_true(is.finite(as.numeric(bw.multi$fval)))
  expect_true(all(is.finite(as.numeric(bw.combo$bw))))
  expect_true(is.finite(as.numeric(bw.combo$fval)))
})

test_that("bounded conditional cv.ls remains finite for gaussian order 2 and 4", {
  set.seed(20260224)
  n <- 70
  x <- runif(n)
  y <- runif(n)

  xdat <- data.frame(x = x)
  ydat <- data.frame(y = y)

  for (ord in c(2L, 4L)) {
    bw <- npcdensbw(
      xdat = xdat,
      ydat = ydat,
      bwmethod = "cv.ls",
      bwtype = "fixed",
      cxkertype = "gaussian",
      cykertype = "gaussian",
      cxkerorder = ord,
      cykerorder = ord,
      cxkerbound = "range",
      cykerbound = "range",
      nmulti = 1
    )

    expect_true(all(is.finite(as.numeric(bw$xbw))))
    expect_true(all(is.finite(as.numeric(bw$ybw))))
    expect_true(is.finite(as.numeric(bw$fval)))
  }
})

test_that("bounded conditional distribution cv.ls remains finite for gaussian order 2 and 4", {
  set.seed(20260224)
  n <- 70
  x <- runif(n)
  y <- runif(n)

  xdat <- data.frame(x = x)
  ydat <- data.frame(y = y)

  for (ord in c(2L, 4L)) {
    bw <- npcdistbw(
      xdat = xdat,
      ydat = ydat,
      bwmethod = "cv.ls",
      bwtype = "fixed",
      cxkertype = "gaussian",
      cykertype = "gaussian",
      cxkerorder = ord,
      cykerorder = ord,
      cxkerbound = "range",
      cykerbound = "range",
      nmulti = 1
    )

    expect_true(all(is.finite(as.numeric(bw$xbw))))
    expect_true(all(is.finite(as.numeric(bw$ybw))))
    expect_true(is.finite(as.numeric(bw$fval)))
  }
})

test_that("bounded unconditional cv.ls remains finite after conditional bounded selectors", {
  set.seed(20260316)
  n <- 48
  x <- runif(n)
  y <- runif(n)

  xdat <- data.frame(x = x)
  ydat <- data.frame(y = y)

  bw.cd <- npcdensbw(
    xdat = xdat,
    ydat = ydat,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    cxkertype = "gaussian",
    cykertype = "gaussian",
    cxkerorder = 2L,
    cykerorder = 2L,
    cxkerbound = "range",
    cykerbound = "range",
    nmulti = 1
  )
  expect_true(all(is.finite(as.numeric(bw.cd$xbw))))
  expect_true(all(is.finite(as.numeric(bw.cd$ybw))))

  bw.ud <- npudensbw(
    dat = xdat,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    ckertype = "gaussian",
    ckerorder = 4L,
    ckerbound = "range",
    nmulti = 1
  )

  expect_true(is.finite(as.numeric(bw.ud$bw[1])))
  expect_true(is.finite(as.numeric(bw.ud$fval)))
})

test_that("bounded conditional cv.ls scalar quadrature supports generalized and adaptive NN bwtypes", {
  set.seed(20260421)
  n <- 48L
  xdat <- data.frame(x = runif(n))
  ydat <- data.frame(y = runif(n))

  bw.gnn <- npcdensbw(
    xdat = xdat,
    ydat = ydat,
    bwmethod = "cv.ls",
    bwtype = "generalized_nn",
    cxkertype = "epanechnikov",
    cykertype = "gaussian",
    cxkerorder = 4L,
    cykerorder = 8L,
    cxkerbound = "range",
    cykerbound = "range",
    nmulti = 1
  )

  bw.ad <- npcdensbw(
    xdat = xdat,
    ydat = ydat,
    bwmethod = "cv.ls",
    bwtype = "adaptive_nn",
    cxkertype = "gaussian",
    cykertype = "epanechnikov",
    cxkerorder = 2L,
    cykerorder = 6L,
    cxkerbound = "range",
    cykerbound = "range",
    nmulti = 1
  )

  expect_true(all(is.finite(as.numeric(bw.gnn$xbw))))
  expect_true(all(is.finite(as.numeric(bw.gnn$ybw))))
  expect_true(is.finite(as.numeric(bw.gnn$fval)))
  expect_true(all(is.finite(as.numeric(bw.ad$xbw))))
  expect_true(all(is.finite(as.numeric(bw.ad$ybw))))
  expect_true(is.finite(as.numeric(bw.ad$fval)))
})

test_that("bounded conditional cv.ls admits mixed and bivariate continuous bounded responses", {
  set.seed(20260421)
  n <- 32L
  xdat <- data.frame(x = runif(n))
  ymixed <- data.frame(y = runif(n), g = factor(sample(c("a", "b"), n, replace = TRUE)))
  ymulti <- data.frame(y1 = runif(n), y2 = runif(n))
  ycombo <- data.frame(
    y1 = runif(n),
    y2 = runif(n),
    u = factor(sample(c("a", "b"), n, replace = TRUE)),
    o = ordered(sample(1:3, n, replace = TRUE))
  )

  bw.mixed <- npcdensbw(
    xdat = xdat,
    ydat = ymixed,
    bwmethod = "cv.ls",
    bwtype = "generalized_nn",
    cykerbound = "range",
    nmulti = 1
  )

  bw.multi <- npcdensbw(
    xdat = xdat,
    ydat = ymulti,
    bwmethod = "cv.ls",
    bwtype = "adaptive_nn",
    cykerbound = "range",
    nmulti = 1
  )

  bw.combo <- npcdensbw(
    xdat = xdat,
    ydat = ycombo,
    bwmethod = "cv.ls",
    bwtype = "generalized_nn",
    cykerbound = "range",
    nmulti = 1
  )

  expect_true(all(is.finite(as.numeric(bw.mixed$xbw))))
  expect_true(all(is.finite(as.numeric(bw.mixed$ybw))))
  expect_true(is.finite(as.numeric(bw.mixed$fval)))
  expect_true(all(is.finite(as.numeric(bw.multi$xbw))))
  expect_true(all(is.finite(as.numeric(bw.multi$ybw))))
  expect_true(is.finite(as.numeric(bw.multi$fval)))
  expect_true(all(is.finite(as.numeric(bw.combo$xbw))))
  expect_true(all(is.finite(as.numeric(bw.combo$ybw))))
  expect_true(is.finite(as.numeric(bw.combo$fval)))
})

test_that("bounded cv.ls still fails fast beyond two continuous bounded variables", {
  set.seed(20260421)
  n <- 24L
  xdat <- data.frame(x = runif(n))
  y3 <- data.frame(y1 = runif(n), y2 = runif(n), y3 = runif(n))
  d3 <- data.frame(x1 = runif(n), x2 = runif(n), x3 = runif(n))

  expect_error(
    npcdensbw(
      xdat = xdat,
      ydat = y3,
      bwmethod = "cv.ls",
      bwtype = "fixed",
      cykerbound = "range",
      nmulti = 1
    ),
    "supports up to two continuous response variables"
  )

  expect_error(
    npudensbw(
      dat = d3,
      bwmethod = "cv.ls",
      bwtype = "fixed",
      ckerbound = "range",
      nmulti = 1
    ),
    "supports up to two continuous variables"
  )
})

library(np)

test_that("npcdens proper apply='both' combines fitted and predict slice behavior", {
  set.seed(20260324)
  x <- runif(40, -1, 1)
  y <- sin(2 * pi * x) + rnorm(40, sd = 0.18)
  nd <- rbind(
    data.frame(y = c(-0.75, -0.2, 0.4), x = rep(-0.35, 3L)),
    data.frame(y = c(-0.1, 0.3, 0.75, 1.0), x = rep(0.45, 4L))
  )

  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.28, 0.22),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )

  ctrl.both <- list(mode = "slice", apply = "both", slice.grid.size = 21L, slice.extend.factor = 0)
  ctrl.fit <- modifyList(ctrl.both, list(apply = "fitted"))

  fit.raw <- npcdens(bws = bw, txdat = data.frame(x = x), tydat = data.frame(y = y))
  fit.fitted <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE,
    proper.control = ctrl.fit
  )
  fit.both <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE,
    proper.control = ctrl.both
  )

  pred.base <- predict(fit.raw, newdata = nd, proper = FALSE)
  pred.ref <- predict(fit.raw, newdata = nd, proper = TRUE, proper.control = ctrl.both)
  pred.both <- predict(fit.both, newdata = nd, proper.control = ctrl.both)

  expect_true(isTRUE(fit.both$proper.requested))
  expect_true(isTRUE(fit.both$proper.applied))
  expect_identical(fit.both$proper.info$route, "slice")
  expect_equal(fit.both$condens.raw, fit.raw$condens, tolerance = 1e-12)
  expect_equal(fit.both$condens, fit.fitted$condens, tolerance = 1e-10)
  expect_equal(as.numeric(pred.both), as.numeric(pred.ref), tolerance = 1e-10)
  expect_true(any(abs(as.numeric(pred.both) - as.numeric(pred.base)) > 1e-8))
})

test_that("npcdist proper apply='both' combines fitted and predict slice behavior", {
  set.seed(20260324)
  x <- runif(40, -1, 1)
  y <- sin(2 * pi * x) + rnorm(40, sd = 0.18)
  nd <- rbind(
    data.frame(y = c(-0.75, -0.2, 0.4), x = rep(-0.35, 3L)),
    data.frame(y = c(-0.1, 0.3, 0.75, 1.0), x = rep(0.45, 4L))
  )

  bw <- npcdistbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.28, 0.22),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )

  ctrl.both <- list(mode = "slice", apply = "both", slice.grid.size = 21L, slice.extend.factor = 0)
  ctrl.fit <- modifyList(ctrl.both, list(apply = "fitted"))

  fit.raw <- npcdist(bws = bw, txdat = data.frame(x = x), tydat = data.frame(y = y))
  fit.fitted <- npcdist(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE,
    proper.control = ctrl.fit
  )
  fit.both <- npcdist(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE,
    proper.control = ctrl.both
  )

  pred.base <- predict(fit.raw, newdata = nd, proper = FALSE)
  pred.ref <- predict(fit.raw, newdata = nd, proper = TRUE, proper.control = ctrl.both)
  pred.both <- predict(fit.both, newdata = nd, proper.control = ctrl.both)

  expect_true(isTRUE(fit.both$proper.requested))
  expect_true(isTRUE(fit.both$proper.applied))
  expect_identical(fit.both$proper.info$route, "slice")
  expect_equal(fit.both$condist.raw, fit.raw$condist, tolerance = 1e-12)
  expect_equal(fit.both$condist, fit.fitted$condist, tolerance = 1e-10)
  expect_equal(as.numeric(pred.both), as.numeric(pred.ref), tolerance = 1e-10)
})

test_that("quantile categorical gradient bootstrap helper matches explicit local refits", {
  skip_if_not_installed("np")

  library(np)

  helper <- getFromNamespace(".np_inid_boot_from_quantile_gradient_local", "np")
  eval_fun <- getFromNamespace(".np_plot_quantile_eval", "np")

  set.seed(20260312)
  n <- 24L
  x1 <- factor(sample(c("a", "b"), n, replace = TRUE))
  x2 <- rnorm(n)
  y <- rnorm(n)
  xdat <- data.frame(x1 = x1, x2 = x2)
  ydat <- data.frame(y = y)
  bw <- suppressWarnings(npcdistbw(xdat = xdat, ydat = ydat, nmulti = 1L))
  tau <- 0.5

  slice <- suppressWarnings(plot(
    bw,
    xdat = xdat,
    ydat = ydat,
    tau = tau,
    gradients = TRUE,
    perspective = FALSE,
    output = "data"
  ))[[1L]]

  counts <- cbind(
    rep(1L, n),
    c(rep(2L, 4), rep(0L, 4), rep(1L, n - 8)),
    c(rep(0L, 3), rep(3L, 3), rep(1L, n - 6))
  )
  if (!is.matrix(counts))
    counts <- matrix(counts, nrow = n)

  boot <- helper(
    xdat = xdat,
    ydat = y,
    exdat = slice$xeval,
    bws = bw,
    B = ncol(counts),
    tau = tau,
    gradient.index = 1L,
    counts = counts
  )

  explicit_fit <- function(idx) {
    as.vector(eval_fun(
      bws = bw,
      txdat = xdat[idx, , drop = FALSE],
      tydat = y[idx],
      exdat = slice$xeval,
      tau = tau,
      gradients = TRUE
    )$quantgrad[, 1L])
  }

  oracle.t0 <- explicit_fit(seq_len(n))
  oracle.t <- vapply(
    seq_len(ncol(counts)),
    function(j) explicit_fit(rep.int(seq_len(n), counts[, j])),
    numeric(length(oracle.t0))
  )

  expect_equal(as.vector(boot$t0), oracle.t0, tolerance = 1e-8)
  expect_equal(boot$t, t(oracle.t), tolerance = 1e-8)
})

test_that("quantile categorical bootstrap gradients work again", {
  skip_if_not_installed("np")

  library(np)

  set.seed(20260312)
  n <- 24L
  x1 <- factor(sample(c("a", "b"), n, replace = TRUE))
  x2 <- rnorm(n)
  y <- rnorm(n)
  xdat <- data.frame(x1 = x1, x2 = x2)
  ydat <- data.frame(y = y)
  bw <- suppressWarnings(npcdistbw(xdat = xdat, ydat = ydat, nmulti = 1L))
  fit <- npqreg(bws = bw, txdat = xdat, tydat = y, tau = 0.5)

  for (case in list(
    list(label = "bw", obj = bw, args = list(xdat = xdat, ydat = ydat)),
    list(label = "fit", obj = fit, args = list())
  )) {
    for (boot.method in c("inid", "fixed", "geom")) {
      plot.args <- c(
        list(
          case$obj,
          tau = 0.5,
          gradients = TRUE,
          perspective = FALSE,
          output = "data",
          errors = "bootstrap",
          bootstrap = boot.method,
          B = 5L
        ),
        case$args
      )
      out <- suppressWarnings(do.call(plot, plot.args))
      expect_true(is.list(out), info = paste(case$label, boot.method))
      expect_true(length(out) >= 1L, info = paste(case$label, boot.method))
      expect_true(length(out[[1L]]$bxp) > 0L, info = paste(case$label, boot.method))
      expect_equal(length(out[[1L]]$bxp$names), 2L, info = paste(case$label, boot.method))
    }
  }
})

test_that("npqreg basic functionality works", {
  data("cps71")
  cps71_sub <- cps71[1:50, ]
  
  # Quantile regression needs a condbandwidth object
  bw <- npcdistbw(xdat=cps71_sub$age, ydat=cps71_sub$logwage, 
                  bws=c(0.5, 5), bandwidth.compute=FALSE)
  
  # Median regression
  model <- npqreg(bws=bw, tau=0.5)
  
  expect_s3_class(model, "qregression")
  expect_type(predict(model), "double")
  expect_equal(length(predict(model)), 50)
  expect_equal(model$tau, 0.5)
  expect_error(gradients(model), "fit the model with gradients=TRUE", fixed = TRUE)
  expect_error(gradients(model, errors = TRUE), "fit the model with gradients=TRUE", fixed = TRUE)
  
  expect_output(summary(model))
})

test_that("npqreg works with multiple taus", {
  data("cps71")
  cps71_sub <- cps71[1:30, ]
  bw <- npcdistbw(xdat=cps71_sub$age, ydat=cps71_sub$logwage, 
                  bws=c(0.5, 5), bandwidth.compute=FALSE)
  
  # npqreg only takes a single tau at a time according to some versions, 
  # but let's see if it works as extra arg or in bws.
  # Actually, npqreg usage says tau is an argument.
  
  model_q25 <- npqreg(bws=bw, tau=0.25)
  expect_equal(model_q25$tau, 0.25)
})

test_that("npqreg formula route strips fit controls before bandwidth selection", {
  set.seed(20260511)
  d <- data.frame(x = runif(36), y = rnorm(36))

  fit <- npqreg(y ~ x, data = d, tau = c(0.25, 0.5),
                gradients = TRUE, tol = 1e-4, small = 1e-5,
                itmax = 1000L, nmulti = 1L)

  expect_s3_class(fit, "qregression")
  expect_equal(fit$tau, c(0.25, 0.5))
  expect_true(fit$gradients)
  expect_equal(ncol(fitted(fit)), 2L)
})

test_that("npqreg formula route honors manual bws with gradients", {
  set.seed(20260513)
  d <- data.frame(x = runif(35), y = rnorm(35))

  fit <- npqreg(y ~ x, data = d, tau = c(0.25, 0.5),
                gradients = TRUE, bws = c(0.3, 0.3),
                bandwidth.compute = FALSE)

  expect_s3_class(fit, "qregression")
  expect_equal(dim(fitted(fit)), c(nrow(d), 2L))
  expect_equal(dim(gradients(fit)), c(nrow(d), 1L, 2L))
})

test_that("npqreg validates tau and continuous responses before fitting", {
  set.seed(20260513)
  d <- data.frame(x = runif(24), y = rnorm(24))
  bw <- npcdistbw(
    xdat = d["x"],
    ydat = d["y"],
    bws = c(0.4, 0.4),
    bandwidth.compute = FALSE
  )

  expect_error(
    npqreg(bws = bw, txdat = d["x"], tydat = d$y, tau = 0),
    "'tau' must contain numeric values in (0,1)",
    fixed = TRUE
  )
  expect_error(
    npqreg(bws = bw, txdat = d["x"], tydat = d$y, tau = c(0.25, 1)),
    "'tau' must contain numeric values in (0,1)",
    fixed = TRUE
  )
  expect_error(
    npqreg(bws = bw, txdat = d["x"], tydat = d$y, tau = NA_real_),
    "'tau' must contain numeric values in (0,1)",
    fixed = TRUE
  )
  expect_error(
    npqreg(bws = bw, txdat = d["x"], tydat = d$y, tau = "0.5"),
    "'tau' must contain numeric values in (0,1)",
    fixed = TRUE
  )

  d$y_factor <- factor(rep(0:1, length.out = nrow(d)))
  expect_error(
    npqreg(y_factor ~ x, data = d, bws = c(0.4, 0.4),
           bandwidth.compute = FALSE),
    "'tydat' is not continuous",
    fixed = TRUE
  )
})

test_that("npqreg vector-tau plots support legend controls", {
  set.seed(20260511)
  n <- 24L
  xdat <- data.frame(x = seq(0.05, 0.95, length.out = n))
  ydat <- sin(2 * pi * xdat$x) + rnorm(n, sd = 0.05)
  bw <- npcdistbw(
    xdat = xdat,
    ydat = ydat,
    bws = c(0.5, 0.5),
    bandwidth.compute = FALSE
  )
  fit <- npqreg(bws = bw, tau = c(0.25, 0.5, 0.75))

  old.dev <- grDevices::dev.cur()
  grDevices::pdf(file = tempfile(fileext = ".pdf"))
  on.exit({
    grDevices::dev.off()
    if (old.dev > 1L)
      grDevices::dev.set(old.dev)
  }, add = TRUE)

  expect_silent(plot(fit, perspective = FALSE, errors = "none", neval = 6L,
                     legend = NULL))
  expect_silent(plot(fit, perspective = FALSE, errors = "none", neval = 6L,
                     legend = NA))
  expect_silent(plot(fit, perspective = FALSE, errors = "none", neval = 6L,
                     legend = "bottomright"))
  expect_silent(plot(fit, perspective = FALSE, errors = "none", neval = 6L,
                     legend = list(x = "bottomleft",
                                   legend = c("low", "mid", "high"),
                                   bty = "o")))
  expect_error(
    plot(fit, perspective = FALSE, errors = "none", neval = 6L,
         legend = 1),
    "legend must be TRUE/FALSE"
  )
})

test_that("npqreg plot preserves NOMAD-selected LP degree metadata", {
  set.seed(70511)
  n <- 24L
  d <- data.frame(
    x = runif(n),
    y = sin(runif(n)) + rnorm(n, sd = 0.05)
  )
  bw <- npcdistbw(
    y ~ x,
    data = d,
    nomad = TRUE,
    search.engine = "nomad",
    degree.min = 0L,
    degree.max = 1L,
    degree.start = 1L,
    degree.verify = FALSE,
    nmulti = 1L,
    nomad.nmulti = 0L
  )
  expect_identical(bw$regtype, "lp")
  expect_identical(bw$regtype.engine, "lp")
  expect_false(is.null(bw$degree))
  expect_false(is.null(bw$degree.engine))

  fit <- npqreg(bws = bw, tau = 0.5)
  out <- plot(
    fit,
    output = "data",
    perspective = FALSE,
    tau = c(0.25, 0.5),
    errors = "none",
    neval = 5L
  )

  expect_true(all(vapply(out, inherits, logical(1), "qregression")))
  expect_true(all(vapply(out, function(xi) identical(xi$bws$regtype.engine, "lp"), logical(1))))
  expect_true(all(vapply(out, function(xi) identical(xi$bws$degree.engine, bw$degree.engine), logical(1))))
})

test_that("npqreg gradients populate qregression objects", {
  set.seed(70507)
  n <- 150L
  xdat <- data.frame(
    x1 = runif(n, -1, 1),
    x2 = runif(n, -1, 1)
  )
  ydat <- xdat$x1 - 0.5 * xdat$x2 + rnorm(n, sd = 0.15)
  exdat <- data.frame(
    x1 = seq(-0.40, 0.40, length.out = 5L),
    x2 = seq(0.35, -0.35, length.out = 5L)
  )
  bw <- npcdistbw(
    xdat = xdat,
    ydat = ydat,
    bws = c(0.80, 0.80, 0.80),
    bandwidth.compute = FALSE
  )

  fit <- npqreg(bws = bw, txdat = xdat, tydat = ydat, exdat = exdat,
                tau = 0.45, gradients = TRUE, tol = 1e-6, small = 1e-7)
  ref <- getFromNamespace(".np_plot_quantile_eval", "np")(
    bws = bw,
    txdat = xdat,
    tydat = ydat,
    exdat = exdat,
    tau = 0.45,
    gradients = TRUE,
    tol = 1e-6,
    small = 1e-7
  )

  expect_s3_class(fit, "qregression")
  expect_true(fit$gradients)
  expect_equal(dim(fit$quantgrad), c(nrow(exdat), ncol(xdat)))
  expect_equal(dim(fit$quantgerr), c(nrow(exdat), ncol(xdat)))
  expect_true(all(is.finite(fit$quantgrad)))
  expect_true(all(is.finite(fit$quantgerr)))
  expect_true(all(is.finite(se(fit))))
  expect_true(all(se(fit) > 0))
  expect_equal(fit$quantgrad, ref$quantgrad, tolerance = 0)
  expect_equal(gradients(fit), fit$quantgrad, tolerance = 0)
  expect_equal(gradients(fit, errors = TRUE), fit$quantgerr, tolerance = 0)

  qdelta <- getFromNamespace(".npqreg_quantile_delta_from_conditional", "np")(
    bws = bw,
    xdat = xdat,
    ydat = ydat,
    exdat = exdat,
    quantile = fitted(fit),
    gradients = TRUE
  )
  expect_equal(se(fit), qdelta$quanterr, tolerance = 0)
  expect_equal(fit$quantgrad, qdelta$quantgrad, tolerance = 0)
  expect_equal(fit$quantgerr, qdelta$quantgerr, tolerance = 0)

  eps <- 1e-2
  for (j in seq_along(exdat)) {
    xp <- xm <- exdat
    xp[[j]] <- xp[[j]] + eps
    xm[[j]] <- xm[[j]] - eps
    qp <- fitted(npqreg(bws = bw, txdat = xdat, tydat = ydat, exdat = xp,
                        tau = 0.45, gradients = FALSE, tol = 1e-6, small = 1e-7))
    qm <- fitted(npqreg(bws = bw, txdat = xdat, tydat = ydat, exdat = xm,
                        tau = 0.45, gradients = FALSE, tol = 1e-6, small = 1e-7))
    fd <- (qp - qm) / (2 * eps)
    expect_lt(max(abs(fit$quantgrad[, j] - fd), na.rm = TRUE), 1e-3)
  }
})

test_that("npqreg gradient plots support asymptotic errors", {
  set.seed(70512)
  n <- 48L
  xdat <- data.frame(x = sort(runif(n)))
  ydat <- sin(2 * pi * xdat$x) + rnorm(n, sd = 0.15)
  bw <- npcdistbw(
    xdat = xdat,
    ydat = ydat,
    bws = c(0.45, 0.45),
    bandwidth.compute = FALSE,
    cxkerbound = "range",
    cykerbound = "range"
  )
  fit <- npqreg(bws = bw, tau = c(0.25, 0.5, 0.75), gradients = TRUE)

  out <- plot(
    fit,
    gradients = TRUE,
    output = "data",
    perspective = FALSE,
    errors = "asymptotic",
    neval = 8L
  )

  expect_true(all(vapply(out, inherits, logical(1), "qregression")))
  expect_true(all(is.finite(gradients(fit))))
  expect_true(all(is.finite(gradients(fit, errors = TRUE))))
})

test_that("named bws formula dispatch matches explicit npqreg bandwidth route", {
  set.seed(20260323)
  d <- data.frame(
    x = seq(0.1, 0.9, length.out = 9),
    y = seq(0.2, 1.0, length.out = 9)
  )
  nd <- data.frame(x = c(0.2, 0.5, 0.8))

  bw <- npcdistbw(y ~ x, data = d)
  fit.pos <- npqreg(bws = bw, data = d, newdata = nd, tau = 0.4)
  fit.named <- npqreg(bws = y ~ x, data = d, newdata = nd, tau = 0.4)

  expect_s3_class(fit.named, "qregression")
  expect_equal(length(fit.named$quantile), nrow(nd), tolerance = 0)
  expect_lt(max(abs(as.numeric(fit.named$quantile) - as.numeric(fit.pos$quantile))), 1e-2)
  expect_lt(max(abs(as.numeric(fit.named$quanterr) - as.numeric(fit.pos$quanterr))), 1e-2)
})

test_that("formula npqreg keeps native exdat out of bandwidth selection", {
  set.seed(20260514)
  d <- data.frame(
    x = seq(0.1, 0.9, length.out = 30)
  )
  d$y <- sin(2 * pi * d$x) + rnorm(nrow(d), sd = 0.08)
  nd <- data.frame(x = c(0.2, 0.5, 0.8))

  fit <- npqreg(
    y ~ x,
    data = d,
    exdat = nd,
    tau = 0.5,
    nmulti = 1
  )

  expect_s3_class(fit, "qregression")
  expect_equal(length(fitted(fit)), nrow(nd))
})

test_that("named bws formula dispatch forwards NOMAD shortcut controls", {
  skip_if_not_installed("crs")

  set.seed(20260511)
  d <- data.frame(
    x = seq(0.05, 0.95, length.out = 30),
    y = sin(seq(0.05, 0.95, length.out = 30)) + rnorm(30, sd = 0.05)
  )

  capture.output(
    fit.named <- npqreg(
      bws = y ~ x,
      data = d,
      tau = 0.5,
      nomad = TRUE,
      nmulti = 1L,
      nomad.nmulti = 1L,
      degree.max = 1L
    )
  )

  expect_s3_class(fit.named, "qregression")
  expect_true(isTRUE(fit.named$bws$nomad.shortcut$enabled))
  expect_identical(as.character(fit.named$bws$regtype.engine), "lp")
})

test_that("npqreg NOMAD fit metadata propagates through prediction, gradients, and plot", {
  skip_if_not_installed("crs")

  count_namespace_calls <- function(fun, expr) {
    ns <- asNamespace("np")
    ctr <- new.env(parent = emptyenv())
    ctr$n <- 0L
    trace(
      what = fun,
      where = ns,
      tracer = bquote(.(ctr)$n <- .(ctr)$n + 1L),
      print = FALSE
    )
    on.exit(try(untrace(fun, where = ns), silent = TRUE), add = TRUE)
    force(expr)
    ctr$n
  }

  set.seed(20260511)
  n <- 36L
  d <- data.frame(x = seq(0.05, 0.95, length.out = n))
  d$y <- sin(2 * pi * d$x) + rnorm(n, sd = 0.08)
  nd <- data.frame(x = c(0.2, 0.4, 0.7))

  capture.output(
    fit <- npqreg(
      y ~ x,
      data = d,
      newdata = nd,
      tau = 0.5,
      gradients = TRUE,
      nomad = TRUE,
      nmulti = 1L,
      nomad.nmulti = 1L,
      degree.max = 1L
    )
  )

  expect_s3_class(fit, "qregression")
  expect_true(isTRUE(fit$bws$nomad.shortcut$enabled))
  expect_identical(as.character(fit$bws$regtype.engine), "lp")
  expect_length(fitted(fit), nrow(nd))
  expect_equal(nrow(gradients(fit)), nrow(nd))
  expect_equal(nrow(gradients(fit, errors = TRUE)), nrow(nd))

  bw.calls <- count_namespace_calls("npcdistbw", {
    pred <- predict(fit, newdata = nd, tau = c(0.25, 0.5, 0.75))
    expect_equal(nrow(pred), nrow(nd))
    expect_equal(ncol(pred), 3L)
  })
  expect_identical(bw.calls, 0L)

  qreg.calls <- count_namespace_calls("npqreg", {
    out <- plot(
      fit,
      output = "data",
      perspective = FALSE,
      errors = "none",
      tau = c(0.25, 0.5, 0.75),
      neval = 6L
    )
    expect_type(out, "list")
  })
  expect_identical(qreg.calls, 0L)
})

library(np)

test_that("plot condistribution 2D data payload applies proper repair on supported grids", {
  set.seed(2)
  n <- 80L
  x <- runif(n, -1, 1)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.20)

  bw <- npcdistbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.28, 0.22),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )
  fit <- npcdist(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE
  )

  out <- suppressWarnings(plot(
    fit,
    output = "data",
    perspective = TRUE,
    view = "fixed"
  ))

  expect_type(out, "list")
  expect_true("cd1" %in% names(out))
  expect_s3_class(out$cd1, "condistribution")
  expect_true(isTRUE(out$cd1$proper.requested))
  expect_true(isTRUE(out$cd1$proper.applied))
  expect_true(all(out$cd1$condist >= -1e-8))
  expect_true(all(out$cd1$condist <= 1 + 1e-8))

  split.idx <- split(seq_along(out$cd1$condist), out$cd1$xeval)
  for (idx in split.idx)
    expect_true(all(diff(out$cd1$condist[idx]) >= -1e-8))
})

test_that("plot condistribution 1D data payload repairs only y-varying panels", {
  set.seed(2)
  n <- 80L
  x <- runif(n, -1, 1)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.20)

  bw <- npcdistbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.28, 0.22),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )
  fit <- npcdist(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE
  )

  out <- suppressWarnings(plot(
    fit,
    output = "data",
    perspective = FALSE
  ))

  expect_length(out, 2L)
  expect_s3_class(out[[1]], "condistribution")
  expect_s3_class(out[[2]], "condistribution")
  expect_true(isTRUE(out[[1]]$proper.requested))
  expect_false(isTRUE(out[[1]]$proper.applied))
  expect_true(isTRUE(out[[2]]$proper.requested))
  expect_true(isTRUE(out[[2]]$proper.applied))
  expect_true(all(out[[2]]$condist >= -1e-8))
  expect_true(all(out[[2]]$condist <= 1 + 1e-8))
  expect_true(all(diff(out[[2]]$condist) >= -1e-8))
})

test_that("plot condistribution supports asymptotic and bootstrap errors on proper grids", {
  set.seed(2)
  x <- runif(60, -1, 1)
  y <- sin(2 * pi * x) + rnorm(60, sd = 0.2)

  bw <- npcdistbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.28, 0.22),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )
  fit <- npcdist(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE
  )

  asym <- suppressWarnings(plot(
    fit,
    output = "data",
    perspective = TRUE,
    view = "fixed",
    errors = "asymptotic"
  ))
  boot <- suppressWarnings(plot(
    fit,
    output = "data",
    perspective = TRUE,
    view = "fixed",
    errors = "bootstrap",
    bootstrap = "inid",
    B = 19,
    band = "pointwise"
  ))

  expect_true(isTRUE(asym$cd1$proper.applied))
  expect_true(all(asym$cd1$condist >= -1e-8))
  expect_true(all(asym$cd1$condist <= 1 + 1e-8))
  expect_true(all(is.finite(asym$cd1$conderr)))

  expect_true(isTRUE(boot$cd1$proper.applied))
  expect_true(all(boot$cd1$condist >= -1e-8))
  expect_true(all(boot$cd1$condist <= 1 + 1e-8))
  expect_true(all(is.finite(boot$cd1$conderr)))

  xeval <- if (is.data.frame(boot$cd1$xeval)) boot$cd1$xeval[[1L]] else as.vector(boot$cd1$xeval)
  split.idx <- split(seq_along(boot$cd1$condist), xeval)
  for (idx in split.idx)
    expect_true(all(diff(boot$cd1$condist[idx]) >= -1e-8))
})

test_that("plot condistribution bootstrap supports mixed-bound fixed helper path", {
  set.seed(42)
  n <- 60L
  x <- runif(n)
  y <- runif(n)

  fit <- npcdist(y ~ x, cykerbound = "range")
  out <- suppressWarnings(plot(
    fit,
    output = "data",
    view = "fixed",
    data_overlay = FALSE,
    errors = "bootstrap",
    bootstrap = "inid",
    B = 5L,
    band = "pointwise"
  ))

  expect_s3_class(out$cd1, "condistribution")
  expect_true(all(is.finite(out$cd1$condist)))
  expect_true(all(is.finite(out$cd1$conderr)))
})

test_that("bandwidth predict methods forward object through bws", {
  old.options <- options(np.messages = FALSE)
  on.exit(options(old.options), add = TRUE)

  check_numeric_equal <- function(a, b, tolerance = 1e-10) {
    expect_equal(as.numeric(a), as.numeric(b), tolerance = tolerance)
  }

  set.seed(20260512)
  n <- 70
  x <- data.frame(x = runif(n))
  y <- sin(2 * pi * x$x) + rnorm(n, sd = 0.2)
  ey <- data.frame(y = y)
  z <- data.frame(z = runif(n))
  w <- data.frame(w = runif(n))
  nd <- data.frame(x = c(0.2, 0.5, 0.8))
  nd.yx <- data.frame(y = c(-0.4, 0, 0.4), x = c(0.2, 0.5, 0.8))
  nd.xw <- data.frame(x = c(0.2, 0.5, 0.8), w = c(0.1, 0.4, 0.7))

  bw.udens <- npudensbw(dat = x, bws = 0.25, bandwidth.compute = FALSE)
  check_numeric_equal(predict(bw.udens)$dens, npudens(bws = bw.udens)$dens)
  check_numeric_equal(
    predict(bw.udens, newdata = nd)$dens,
    npudens(bws = bw.udens, newdata = nd)$dens
  )

  bw.udist <- npudistbw(dat = x, bws = 0.25, bandwidth.compute = FALSE)
  check_numeric_equal(predict(bw.udist)$dist, npudist(bws = bw.udist)$dist)
  check_numeric_equal(
    predict(bw.udist, newdata = nd)$dist,
    npudist(bws = bw.udist, newdata = nd)$dist
  )

  bw.cdens <- npcdensbw(
    xdat = x,
    ydat = ey,
    bws = c(0.25, 0.2),
    bandwidth.compute = FALSE
  )
  check_numeric_equal(predict(bw.cdens)$condens, npcdens(bws = bw.cdens)$condens)
  check_numeric_equal(
    predict(bw.cdens, newdata = nd.yx)$condens,
    npcdens(bws = bw.cdens, newdata = nd.yx)$condens
  )

  bw.cdist <- npcdistbw(
    xdat = x,
    ydat = ey,
    bws = c(0.25, 0.2),
    bandwidth.compute = FALSE
  )
  check_numeric_equal(predict(bw.cdist)$condist, npcdist(bws = bw.cdist)$condist)
  check_numeric_equal(
    predict(bw.cdist, newdata = nd.yx)$condist,
    npcdist(bws = bw.cdist, newdata = nd.yx)$condist
  )

  bw.reg <- npregbw(xdat = x, ydat = y, bws = 0.25, bandwidth.compute = FALSE)
  check_numeric_equal(predict(bw.reg)$mean, npreg(bws = bw.reg)$mean)
  check_numeric_equal(
    predict(bw.reg, newdata = nd)$mean,
    npreg(bws = bw.reg, newdata = nd)$mean
  )

  bw.si <- npindexbw(
    xdat = data.frame(x = x$x, w = w$w),
    ydat = y,
    bws = c(0.25, 1, 0),
    bandwidth.compute = FALSE
  )
  check_numeric_equal(predict(bw.si)$mean, npindex(bws = bw.si)$mean)
  check_numeric_equal(
    predict(bw.si, newdata = nd.xw)$mean,
    npindex(bws = bw.si, newdata = nd.xw)$mean
  )

  semidat <- data.frame(
    y = y[seq_len(30)],
    x = x$x[seq_len(30)],
    z = z$z[seq_len(30)]
  )
  nd.xz <- data.frame(x = semidat$x[1:3], z = c(0.2, 0.5, 0.8))

  bw.pl <- npplregbw(y ~ x | z, data = semidat, nmulti = 1, tol = 0.1, ftol = 0.1)
  check_numeric_equal(predict(bw.pl)$mean, npplreg(bws = bw.pl)$mean)
  check_numeric_equal(
    predict(bw.pl, newdata = nd.xz)$mean,
    npplreg(bws = bw.pl, newdata = nd.xz)$mean
  )

  bw.sc <- npscoefbw(y ~ x | z, data = semidat, nmulti = 1, tol = 0.1, ftol = 0.1)
  check_numeric_equal(predict(bw.sc)$mean, npscoef(bws = bw.sc)$mean)
  check_numeric_equal(
    predict(bw.sc, newdata = nd.xz)$mean,
    npscoef(bws = bw.sc, newdata = nd.xz)$mean
  )
})

test_that("npqreg fixed no-error slice prototype matches current fitted-object route", {
  proto <- getFromNamespace(".np_plot_proto_npqreg_fixed_none_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(2801)

  n <- 70L
  x <- data.frame(x1 = runif(n, -1, 1), x2 = runif(n, -0.5, 1.5))
  y <- data.frame(y = x$x1 - x$x2 + rnorm(n, sd = 0.15))
  bw <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.45, 0.50, 0.55),
    bandwidth.compute = FALSE,
    bwtype = "fixed"
  )
  fit <- npqreg(
    bws = bw,
    txdat = x,
    tydat = y,
    tau = 0.4
  )

  old <- suppressWarnings(plot(
    fit,
    xdat = x,
    ydat = y,
    output = "data",
    errors = "none",
    neval = 7L,
    perspective = FALSE
  ))
  candidate <- proto(
    fit,
    xdat = x,
    ydat = y,
    neval = 7L
  )
  stages <- proto(
    fit,
    xdat = x,
    ydat = y,
    neval = 7L,
    return.stages = TRUE
  )

  expect_named(candidate, names(old))
  expect_true(all(vapply(candidate, inherits, logical(1), "qregression")))
  for (nm in names(old)) {
    expect_named(candidate[[nm]], names(old[[nm]]), info = nm)
    expect_equal(candidate[[nm]]$xeval, old[[nm]]$xeval, info = nm)
    expect_equal(candidate[[nm]]$tau, old[[nm]]$tau, info = nm)
    expect_equal(candidate[[nm]]$quantile, old[[nm]]$quantile, info = nm)
    expect_equal(candidate[[nm]]$quanterr, old[[nm]]$quanterr, info = nm)
  }
  expect_named(stages, c("state", "target_grid", "evaluator", "intervals", "bootstrap", "plot_data"))
  for (nm in names(candidate)) {
    expect_equal(stages$plot_data[[nm]]$xeval, candidate[[nm]]$xeval, info = nm)
    expect_equal(stages$plot_data[[nm]]$quantile, candidate[[nm]]$quantile, info = nm)
    expect_equal(stages$plot_data[[nm]]$quanterr, candidate[[nm]]$quanterr, info = nm)
  }
  expect_true(all(vapply(stages$intervals, is.null, logical(1))))
  expect_true(all(vapply(stages$bootstrap, is.null, logical(1))))
})

test_that("npqreg plot prototype fails early outside its fitted-object fixed slice", {
  proto <- getFromNamespace(".np_plot_proto_npqreg_fixed_none_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(2802)

  n <- 55L
  x <- data.frame(x1 = runif(n), x2 = factor(rbinom(n, 1L, 0.5)))
  y <- data.frame(y = x$x1 + rnorm(n, sd = 0.1))
  bw <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.45, 0.49, 0.49),
    bandwidth.compute = FALSE,
    bwtype = "fixed"
  )
  fit <- npqreg(
    bws = bw,
    txdat = x,
    tydat = y,
    tau = 0.4
  )

  expect_error(
    proto(bw, xdat = x, ydat = y, neval = 7L),
    "quantile regression object",
    fixed = TRUE
  )

  expect_error(
    proto(fit, xdat = x, ydat = y, neval = 7L),
    "continuous/ordered slice variables",
    fixed = TRUE
  )

  expect_error(
    proto(fit, neval = 7L),
    "explicit xdat and ydat",
    fixed = TRUE
  )
})

test_that("npqreg fixed asymptotic slice prototype matches current fitted-object route", {
  proto <- getFromNamespace(".np_plot_proto_npqreg_fixed_asymptotic_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(2803)

  n <- 65L
  x <- data.frame(x1 = runif(n, -1, 1), x2 = runif(n, -1, 1))
  y <- data.frame(y = x$x1^2 - x$x2 + rnorm(n, sd = 0.2))
  bw <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.55, 0.50, 0.60),
    bandwidth.compute = FALSE,
    bwtype = "fixed"
  )
  fit <- npqreg(
    bws = bw,
    txdat = x,
    tydat = y,
    tau = 0.35
  )

  for (band in c("pmzsd", "pointwise", "bonferroni", "simultaneous", "all")) {
    old <- suppressWarnings(plot(
      fit,
      xdat = x,
      ydat = y,
      output = "data",
      errors = "asymptotic",
      band = band,
      neval = 6L,
      perspective = FALSE
    ))
    candidate <- proto(
      fit,
      xdat = x,
      ydat = y,
      neval = 6L,
      plot.errors.type = band
    )
    stages <- proto(
      fit,
      xdat = x,
      ydat = y,
      neval = 6L,
      plot.errors.type = band,
      return.stages = TRUE
    )

    for (nm in names(old)) {
      expect_equal(candidate[[nm]]$xeval, old[[nm]]$xeval, info = paste(nm, band))
      expect_equal(candidate[[nm]]$quantile, old[[nm]]$quantile, info = paste(nm, band))
      expect_equal(candidate[[nm]]$quanterr, old[[nm]]$quanterr, info = paste(nm, band))
      expect_equal(candidate[[nm]]$bias, old[[nm]]$bias, info = paste(nm, band))
      expect_equal(candidate[[nm]]$bxp, old[[nm]]$bxp, info = paste(nm, band))
    }
    expect_true(all(vapply(stages$bootstrap, is.null, logical(1))))
    for (nm in names(candidate)) {
      expect_equal(stages$plot_data[[nm]]$xeval, candidate[[nm]]$xeval, info = paste(nm, band))
      expect_equal(stages$plot_data[[nm]]$quantile, candidate[[nm]]$quantile, info = paste(nm, band))
      expect_equal(stages$plot_data[[nm]]$quanterr, candidate[[nm]]$quanterr, info = paste(nm, band))
      expect_equal(stages$plot_data[[nm]]$bias, candidate[[nm]]$bias, info = paste(nm, band))
      expect_equal(stages$plot_data[[nm]]$bxp, candidate[[nm]]$bxp, info = paste(nm, band))
    }
  }
})

test_that("npqreg fixed quantile-level bootstrap slice prototype matches current fitted-object route", {
  proto <- getFromNamespace(".np_plot_proto_npqreg_fixed_bootstrap_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(2804)

  n <- 50L
  x <- data.frame(x1 = runif(n, -1, 1), x2 = runif(n, -1, 1))
  y <- data.frame(y = cos(x$x1) + x$x2 + rnorm(n, sd = 0.25))
  bw <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.55, 0.60, 0.65),
    bandwidth.compute = FALSE,
    bwtype = "fixed"
  )
  fit <- npqreg(
    bws = bw,
    txdat = x,
    tydat = y,
    tau = 0.45
  )

  cases <- expand.grid(
    method = c("inid", "fixed", "geom"),
    center = c("estimate", "bias-corrected"),
    stringsAsFactors = FALSE
  )
  for (ii in seq_len(nrow(cases))) {
    method <- cases$method[ii]
    center <- cases$center[ii]
    boot.seed <- 8200L + ii
    plot.extra <- if (identical(method, "inid")) list() else list(boot_control = np_boot_control(blocklen = 3L))
    proto.extra <- if (identical(method, "inid")) list() else list(plot.errors.boot.blocklen = 3L)
    set.seed(boot.seed)
    old <- suppressWarnings(do.call(plot, c(
      list(
        x = fit,
        xdat = x,
        ydat = y,
        output = "data",
        errors = "bootstrap",
        bootstrap = method,
        B = 11L,
        center = center,
        band = "pmzsd",
        neval = 5L,
        perspective = FALSE,
        random.seed = boot.seed
      ),
      plot.extra
    )))
    set.seed(boot.seed)
    candidate <- suppressWarnings(do.call(proto, c(
      list(
        object = fit,
        xdat = x,
        ydat = y,
        neval = 5L,
        plot.errors.boot.method = method,
        plot.errors.boot.num = 11L,
        plot.errors.center = center,
        plot.errors.type = "pmzsd"
      ),
      proto.extra
    )))
    set.seed(boot.seed)
    stages <- suppressWarnings(do.call(proto, c(
      list(
        object = fit,
        xdat = x,
        ydat = y,
        neval = 5L,
        plot.errors.boot.method = method,
        plot.errors.boot.num = 11L,
        plot.errors.center = center,
        plot.errors.type = "pmzsd",
        return.stages = TRUE
      ),
      proto.extra
    )))

    for (nm in names(old)) {
      expect_equal(candidate[[nm]]$xeval, old[[nm]]$xeval, info = paste(nm, method, center))
      expect_equal(candidate[[nm]]$quantile, old[[nm]]$quantile, info = paste(nm, method, center))
      expect_equal(candidate[[nm]]$quanterr, old[[nm]]$quanterr, info = paste(nm, method, center))
      expect_equal(candidate[[nm]]$bias, old[[nm]]$bias, info = paste(nm, method, center))
      expect_equal(candidate[[nm]]$bxp, old[[nm]]$bxp, info = paste(nm, method, center))
    }
    expect_true(all(vapply(stages$bootstrap, function(x) is.list(x) && !is.null(x$boot.err), logical(1))),
                info = paste(method, center))
  }
})

test_that("npqreg fixed gradient slice prototype matches current fitted-object route", {
  proto.none <- getFromNamespace(".np_plot_proto_npqreg_fixed_none_data", "np")
  proto.boot <- getFromNamespace(".np_plot_proto_npqreg_fixed_bootstrap_data", "np")
  proto.asym <- getFromNamespace(".np_plot_proto_npqreg_fixed_asymptotic_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(2805)

  n <- 45L
  x <- data.frame(x1 = runif(n, -1, 1), x2 = runif(n, -1, 1))
  y <- data.frame(y = x$x1^2 - x$x2 + rnorm(n, sd = 0.15))
  bw <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.50, 0.55, 0.60),
    bandwidth.compute = FALSE,
    bwtype = "fixed"
  )
  fit <- npqreg(
    bws = bw,
    txdat = x,
    tydat = y,
    tau = 0.45
  )

  old.none <- suppressWarnings(plot(
    fit,
    xdat = x,
    ydat = y,
    output = "data",
    errors = "none",
    gradients = TRUE,
    neval = 5L,
    perspective = FALSE
  ))
  candidate.none <- proto.none(
    fit,
    xdat = x,
    ydat = y,
    neval = 5L,
    gradients = TRUE
  )
  stages.none <- proto.none(
    fit,
    xdat = x,
    ydat = y,
    neval = 5L,
    gradients = TRUE,
    return.stages = TRUE
  )

  expect_named(candidate.none, names(old.none))
  expect_true(isTRUE(stages.none$state$gradients))
  for (nm in names(old.none)) {
    expect_equal(candidate.none[[nm]]$xeval, old.none[[nm]]$xeval, info = nm)
    expect_equal(candidate.none[[nm]]$quantile, old.none[[nm]]$quantile, info = nm)
    expect_equal(candidate.none[[nm]]$quantgrad, old.none[[nm]]$quantgrad, info = nm)
    expect_equal(candidate.none[[nm]]$quanterr, old.none[[nm]]$quanterr, info = nm)
  }

  old.asym <- suppressWarnings(plot(
    fit,
    xdat = x,
    ydat = y,
    output = "data",
    errors = "asymptotic",
    gradients = TRUE,
    neval = 5L,
    perspective = FALSE
  ))
  candidate.asym <- proto.asym(
    fit,
    xdat = x,
    ydat = y,
    neval = 5L,
    gradients = TRUE
  )
  expect_named(candidate.asym, names(old.asym))
  for (nm in names(old.asym)) {
    expect_equal(candidate.asym[[nm]]$xeval, old.asym[[nm]]$xeval, info = nm)
    expect_equal(candidate.asym[[nm]]$quantgrad, old.asym[[nm]]$quantgrad, info = nm)
    expect_equal(candidate.asym[[nm]]$quantgerr, old.asym[[nm]]$quantgerr, info = nm)
  }

  for (method in c("inid", "fixed")) {
    boot.seed <- if (identical(method, "inid")) 8301L else 902L
    plot.extra <- if (identical(method, "fixed")) list(boot_control = np_boot_control(blocklen = 2L)) else list()
    proto.extra <- if (identical(method, "fixed")) list(plot.errors.boot.blocklen = 2L) else list()
    boot.num <- if (identical(method, "fixed")) 5L else 7L
    set.seed(boot.seed)
    old.boot <- suppressWarnings(do.call(plot, c(
      list(
        x = fit,
        xdat = x,
        ydat = y,
        output = "data",
        errors = "bootstrap",
        bootstrap = method,
        B = boot.num,
        center = "estimate",
        band = "pmzsd",
        gradients = TRUE,
        neval = 5L,
        perspective = FALSE,
        random.seed = boot.seed
      ),
      plot.extra
    )))
    set.seed(boot.seed)
    candidate.boot <- suppressWarnings(do.call(proto.boot, c(
      list(
        object = fit,
        xdat = x,
        ydat = y,
        neval = 5L,
        plot.errors.boot.method = method,
        plot.errors.boot.num = boot.num,
        plot.errors.center = "estimate",
        plot.errors.type = "pmzsd",
        gradients = TRUE
      ),
      proto.extra
    )))
    set.seed(boot.seed)
    stages.boot <- suppressWarnings(do.call(proto.boot, c(
      list(
        object = fit,
        xdat = x,
        ydat = y,
        neval = 5L,
        plot.errors.boot.method = method,
        plot.errors.boot.num = boot.num,
        plot.errors.center = "estimate",
        plot.errors.type = "pmzsd",
        gradients = TRUE,
        return.stages = TRUE
      ),
      proto.extra
    )))

    for (nm in names(old.boot)) {
      expect_equal(candidate.boot[[nm]]$xeval, old.boot[[nm]]$xeval, info = paste(nm, method))
      expect_equal(candidate.boot[[nm]]$quantile, old.boot[[nm]]$quantile, info = paste(nm, method))
      expect_equal(candidate.boot[[nm]]$quantgrad, old.boot[[nm]]$quantgrad, info = paste(nm, method))
      expect_equal(candidate.boot[[nm]]$gc1err, old.boot[[nm]]$gc1err, info = paste(nm, method))
      expect_equal(candidate.boot[[nm]]$gc1bias, old.boot[[nm]]$gc1bias, info = paste(nm, method))
      expect_equal(candidate.boot[[nm]]$gc2err, old.boot[[nm]]$gc2err, info = paste(nm, method))
      expect_equal(candidate.boot[[nm]]$gc2bias, old.boot[[nm]]$gc2bias, info = paste(nm, method))
      expect_equal(candidate.boot[[nm]]$bxp, old.boot[[nm]]$bxp, info = paste(nm, method))
    }
    expect_true(all(vapply(stages.boot$bootstrap, is.list, logical(1))))
  }
})

library(np)

scale_floor_fixture <- function(n = 20L) {
  set.seed(8675309)
  data.frame(
    x = rnorm(n),
    y = rnorm(n)
  )
}

expect_bad_hbc_error <- function(expr) {
  expect_error(
    expr,
    regexp = "scale\\.factor\\.init\\.upper.*max\\('scale\\.factor\\.init\\.lower', 'scale\\.factor\\.search\\.lower'\\)"
  )
}

test_that("continuous search starts reject hbc below the effective lower endpoint", {
  dat <- scale_floor_fixture()

  expect_bad_hbc_error(
    npregbw(
      y ~ x,
      data = dat,
      scale.factor.search.lower = 1,
      scale.factor.init.lower = 0.1,
      scale.factor.init.upper = 0.5
    )
  )

  expect_bad_hbc_error(
    npudensbw(
      ~ x,
      data = dat,
      scale.factor.search.lower = 1,
      scale.factor.init.lower = 0.1,
      scale.factor.init.upper = 0.5
    )
  )

  expect_bad_hbc_error(
    npudistbw(
      ~ x,
      data = dat,
      scale.factor.search.lower = 1,
      scale.factor.init.lower = 0.1,
      scale.factor.init.upper = 0.5
    )
  )

  expect_bad_hbc_error(
    npcdensbw(
      y ~ x,
      data = dat,
      scale.factor.search.lower = 1,
      scale.factor.init.lower = 0.1,
      scale.factor.init.upper = 0.5
    )
  )

  expect_bad_hbc_error(
    npcdistbw(
      y ~ x,
      data = dat,
      scale.factor.search.lower = 1,
      scale.factor.init.lower = 0.1,
      scale.factor.init.upper = 0.5
    )
  )

  dat$z <- rnorm(nrow(dat))

  expect_bad_hbc_error(
    npscoefbw(
      y ~ x | z,
      data = dat,
      scale.factor.search.lower = 1,
      scale.factor.init.lower = 0.1,
      scale.factor.init.upper = 0.5
    )
  )

  expect_bad_hbc_error(
    npindexbw(
      y ~ x + z,
      data = dat,
      scale.factor.search.lower = 1,
      scale.factor.init.lower = 0.1,
      scale.factor.init.upper = 0.5
    )
  )
})

test_that("explicit bandwidth objects are not clamped by the search floor", {
  dat <- scale_floor_fixture()
  tiny <- 1e-8

  reg <- npregbw(
    xdat = data.frame(x = dat$x),
    ydat = dat$y,
    bws = tiny,
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    scale.factor.search.lower = 1
  )
  expect_equal(reg$bw[1L], tiny, tolerance = 0)

  dens <- npudensbw(
    dat = data.frame(x = dat$x),
    bws = tiny,
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    scale.factor.search.lower = 1
  )
  expect_equal(dens$bw[1L], tiny, tolerance = 0)

  dist <- npudistbw(
    dat = data.frame(x = dat$x),
    bws = tiny,
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    scale.factor.search.lower = 1
  )
  expect_equal(dist$bw[1L], tiny, tolerance = 0)

  cdens <- npcdensbw(
    xdat = data.frame(x = dat$x),
    ydat = data.frame(y = dat$y),
    bws = c(tiny, tiny),
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    scale.factor.search.lower = 1
  )
  expect_equal(cdens$ybw[1L], tiny, tolerance = 0)
  expect_equal(cdens$xbw[1L], tiny, tolerance = 0)

  cdist <- npcdistbw(
    xdat = data.frame(x = dat$x),
    ydat = data.frame(y = dat$y),
    bws = c(tiny, tiny),
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    scale.factor.search.lower = 1
  )
  expect_equal(cdist$ybw[1L], tiny, tolerance = 0)
  expect_equal(cdist$xbw[1L], tiny, tolerance = 0)
})

test_that("fixed unconditional density/distribution helper matches explicit refits", {
  skip_if_not_installed("np")

  set.seed(603101)
  n <- 70
  tx <- data.frame(x = rnorm(n))
  ex <- data.frame(x = seq(min(tx$x), max(tx$x), length.out = 19))
  B <- 7L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  dens.bw <- npudensbw(dat = tx, bws = 0.35, bandwidth.compute = FALSE)
  dist.bw <- npudistbw(dat = tx, bws = 0.35, bandwidth.compute = FALSE)
  fast.fun <- getFromNamespace(".np_inid_boot_from_ksum_unconditional", "np")

  dens.fast <- fast.fun(
    xdat = tx,
    exdat = ex,
    bws = dens.bw,
    B = B,
    operator = "normal",
    counts = counts
  )
  dist.fast <- fast.fun(
    xdat = tx,
    exdat = ex,
    bws = dist.bw,
    B = B,
    operator = "integral",
    counts = counts
  )

  dens.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  dist.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  for (b in seq_len(B)) {
    idx <- rep.int(seq_len(n), counts[, b])
    dens.explicit[b, ] <- npudens(tdat = tx[idx, , drop = FALSE], edat = ex, bws = dens.bw)$dens
    dist.explicit[b, ] <- npudist(tdat = tx[idx, , drop = FALSE], edat = ex, bws = dist.bw)$dist
  }

  expect_equal(dens.fast$t, dens.explicit, tolerance = 1e-10)
  expect_equal(dist.fast$t, dist.explicit, tolerance = 1e-10)
  expect_equal(dens.fast$t0, npudens(tdat = tx, edat = ex, bws = dens.bw)$dens, tolerance = 1e-12)
  expect_equal(dist.fast$t0, npudist(tdat = tx, edat = ex, bws = dist.bw)$dist, tolerance = 1e-12)
})

test_that("fixed conditional density/distribution helper matches explicit refits", {
  skip_if_not_installed("np")

  set.seed(603102)
  n <- 80
  tx <- data.frame(x = rnorm(n))
  ty <- data.frame(y = rnorm(n))
  ex <- data.frame(x = seq(min(tx$x), max(tx$x), length.out = 17))
  ey <- data.frame(y = seq(min(ty$y), max(ty$y), length.out = 17))
  B <- 7L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  fast.fun <- getFromNamespace(".np_inid_boot_from_ksum_conditional", "np")
  oracle.fun <- getFromNamespace(".np_inid_boot_from_conditional_localpoly_fixed_rowwise", "np")

  cfgs <- list(
    list(name = "lc", regtype = "lc", degree = NULL),
    list(name = "ll", regtype = "ll", degree = NULL),
    list(name = "lp", regtype = "lp", degree = 2L)
  )

  for (cfg in cfgs) {
    dens.args <- list(
      xdat = tx,
      ydat = ty,
      bws = c(0.45, 0.45),
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = cfg$regtype
    )
    dist.args <- dens.args
    if (!is.null(cfg$degree)) {
      dens.args$degree <- cfg$degree
      dens.args$basis <- "glp"
      dens.args$bernstein.basis <- FALSE
      dist.args$degree <- cfg$degree
      dist.args$basis <- "glp"
      dist.args$bernstein.basis <- FALSE
    }

    dens.bw <- do.call(npcdensbw, dens.args)
    dist.bw <- do.call(npcdistbw, dist.args)
    oracle.t0.tolerance <- if (identical(cfg$regtype, "lp")) 1e-10 else 1e-12

    dens.fast <- fast.fun(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = dens.bw,
      B = B,
      cdf = FALSE,
      counts = counts
    )
    dist.fast <- fast.fun(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = dist.bw,
      B = B,
      cdf = TRUE,
      counts = counts
    )

    dens.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
    dist.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
    for (b in seq_len(B)) {
      idx <- rep.int(seq_len(n), counts[, b])
      dens.explicit[b, ] <- npcdens(
        txdat = tx[idx, , drop = FALSE],
        tydat = ty[idx, , drop = FALSE],
        exdat = ex,
        eydat = ey,
        bws = dens.bw
      )$condens
      dist.explicit[b, ] <- npcdist(
        txdat = tx[idx, , drop = FALSE],
        tydat = ty[idx, , drop = FALSE],
        exdat = ex,
        eydat = ey,
        bws = dist.bw
      )$condist
    }

    expect_equal(
      dens.fast$t,
      dens.explicit,
      tolerance = 1e-10,
      info = paste("density explicit", cfg$name)
    )
    expect_equal(
      dist.fast$t,
      dist.explicit,
      tolerance = 1e-10,
      info = paste("distribution explicit", cfg$name)
    )
    expect_equal(
      dens.fast$t0,
      npcdens(txdat = tx, tydat = ty, exdat = ex, eydat = ey, bws = dens.bw)$condens,
      tolerance = 1e-12,
      info = paste("density t0", cfg$name)
    )
    expect_equal(
      dist.fast$t0,
      npcdist(txdat = tx, tydat = ty, exdat = ex, eydat = ey, bws = dist.bw)$condist,
      tolerance = 1e-12,
      info = paste("distribution t0", cfg$name)
    )

    if (!identical(cfg$regtype, "lc")) {
      dens.oracle <- oracle.fun(
        xdat = tx,
        ydat = ty,
        exdat = ex,
        eydat = ey,
        bws = dens.bw,
        B = B,
        cdf = FALSE,
        counts = counts
      )
      dist.oracle <- oracle.fun(
        xdat = tx,
        ydat = ty,
        exdat = ex,
        eydat = ey,
        bws = dist.bw,
        B = B,
        cdf = TRUE,
        counts = counts
      )

      expect_equal(
        dens.fast$t,
        dens.oracle$t,
        tolerance = 1e-12,
        info = paste("density oracle", cfg$name)
      )
      expect_equal(
        dist.fast$t,
        dist.oracle$t,
        tolerance = 1e-12,
        info = paste("distribution oracle", cfg$name)
      )
    expect_equal(
      dens.fast$t0,
      dens.oracle$t0,
      tolerance = oracle.t0.tolerance,
      info = paste("density oracle t0", cfg$name)
    )
    expect_equal(
      dist.fast$t0,
      dist.oracle$t0,
      tolerance = oracle.t0.tolerance,
      info = paste("distribution oracle t0", cfg$name)
    )

      drawer <- function(start, stopi) counts[, start:stopi, drop = FALSE]
      dens.fast.drawer <- fast.fun(
        xdat = tx,
        ydat = ty,
        exdat = ex,
        eydat = ey,
        bws = dens.bw,
        B = B,
        cdf = FALSE,
        counts.drawer = drawer
      )
      dist.fast.drawer <- fast.fun(
        xdat = tx,
        ydat = ty,
        exdat = ex,
        eydat = ey,
        bws = dist.bw,
        B = B,
        cdf = TRUE,
        counts.drawer = drawer
      )
      dens.oracle.drawer <- oracle.fun(
        xdat = tx,
        ydat = ty,
        exdat = ex,
        eydat = ey,
        bws = dens.bw,
        B = B,
        cdf = FALSE,
        counts.drawer = drawer
      )
      dist.oracle.drawer <- oracle.fun(
        xdat = tx,
        ydat = ty,
        exdat = ex,
        eydat = ey,
        bws = dist.bw,
        B = B,
        cdf = TRUE,
        counts.drawer = drawer
      )

      expect_equal(
        dens.fast.drawer$t,
        dens.oracle.drawer$t,
        tolerance = 1e-12,
        info = paste("density drawer oracle", cfg$name)
      )
      expect_equal(
        dist.fast.drawer$t,
        dist.oracle.drawer$t,
        tolerance = 1e-12,
        info = paste("distribution drawer oracle", cfg$name)
      )
    expect_equal(
      dens.fast.drawer$t0,
      dens.oracle.drawer$t0,
      tolerance = oracle.t0.tolerance,
      info = paste("density drawer oracle t0", cfg$name)
    )
    expect_equal(
      dist.fast.drawer$t0,
      dist.oracle.drawer$t0,
      tolerance = oracle.t0.tolerance,
      info = paste("distribution drawer oracle t0", cfg$name)
    )
    }
  }
})

test_that("fixed conditional ll/lp grouped inid helper matches the rowwise oracle on repeated evaluation x rows", {
  skip_if_not_installed("np")

  set.seed(6031021)
  n <- 72
  tx <- data.frame(x = rnorm(n))
  ty <- data.frame(y = rnorm(n))
  x.grid <- seq(min(tx$x), max(tx$x), length.out = 6L)
  y.grid <- seq(min(ty$y), max(ty$y), length.out = 5L)
  grid <- expand.grid(y = y.grid, x = x.grid)
  ex <- data.frame(x = grid$x)
  ey <- data.frame(y = grid$y)
  B <- 6L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  fast.fun <- getFromNamespace(".np_inid_boot_from_ksum_conditional", "np")
  oracle.fun <- getFromNamespace(".np_inid_boot_from_conditional_localpoly_fixed_rowwise", "np")

  cfgs <- list(
    list(name = "ll", regtype = "ll", degree = NULL),
    list(name = "lp", regtype = "lp", degree = 2L)
  )

  for (cfg in cfgs) {
    dens.args <- list(
      xdat = tx,
      ydat = ty,
      bws = c(0.40, 0.40),
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = cfg$regtype
    )
    dist.args <- dens.args
    if (!is.null(cfg$degree)) {
      dens.args$degree <- cfg$degree
      dens.args$basis <- "glp"
      dens.args$bernstein.basis <- FALSE
      dist.args$degree <- cfg$degree
      dist.args$basis <- "glp"
      dist.args$bernstein.basis <- FALSE
    }

    dens.bw <- do.call(npcdensbw, dens.args)
    dist.bw <- do.call(npcdistbw, dist.args)

    dens.fast <- fast.fun(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = dens.bw,
      B = B,
      cdf = FALSE,
      counts = counts
    )
    dens.oracle <- oracle.fun(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = dens.bw,
      B = B,
      cdf = FALSE,
      counts = counts
    )
    dist.fast <- fast.fun(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = dist.bw,
      B = B,
      cdf = TRUE,
      counts = counts
    )
    dist.oracle <- oracle.fun(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = dist.bw,
      B = B,
      cdf = TRUE,
      counts = counts
    )

    expect_equal(dens.fast$t, dens.oracle$t, tolerance = 1e-12, info = paste("density repeated-x oracle", cfg$name))
    expect_equal(dens.fast$t0, dens.oracle$t0, tolerance = 1e-10, info = paste("density repeated-x oracle t0", cfg$name))
    expect_equal(dist.fast$t, dist.oracle$t, tolerance = 1e-12, info = paste("distribution repeated-x oracle", cfg$name))
    expect_equal(dist.fast$t0, dist.oracle$t0, tolerance = 1e-10, info = paste("distribution repeated-x oracle t0", cfg$name))
  }
})

test_that("fixed conditional ll/lp grouped helper is drawer-equal and chunk-invariant on repeated evaluation x rows", {
  skip_if_not_installed("np")

  old.chunk <- getOption("np.plot.inid.chunk.size")
  options(np.plot.inid.chunk.size = 2L)
  on.exit(options(np.plot.inid.chunk.size = old.chunk), add = TRUE)

  set.seed(6031022)
  n <- 68
  tx <- data.frame(x = rnorm(n))
  ty <- data.frame(y = rnorm(n))
  x.grid <- seq(min(tx$x), max(tx$x), length.out = 7L)
  y.grid <- seq(min(ty$y), max(ty$y), length.out = 6L)
  grid <- expand.grid(y = y.grid, x = x.grid)
  ex <- data.frame(x = grid$x)
  ey <- data.frame(y = grid$y)
  B <- 7L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  fast.fun <- getFromNamespace(".np_inid_boot_from_ksum_conditional", "np")
  oracle.fun <- getFromNamespace(".np_inid_boot_from_conditional_localpoly_fixed_rowwise", "np")

  cfgs <- list(
    list(name = "ll", regtype = "ll", degree = NULL),
    list(name = "lp", regtype = "lp", degree = 2L)
  )

  for (cfg in cfgs) {
    dens.args <- list(
      xdat = tx,
      ydat = ty,
      bws = c(0.38, 0.38),
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = cfg$regtype
    )
    dist.args <- dens.args
    if (!is.null(cfg$degree)) {
      dens.args$degree <- cfg$degree
      dens.args$basis <- "glp"
      dens.args$bernstein.basis <- FALSE
      dist.args$degree <- cfg$degree
      dist.args$basis <- "glp"
      dist.args$bernstein.basis <- FALSE
    }

    dens.bw <- do.call(npcdensbw, dens.args)
    dist.bw <- do.call(npcdistbw, dist.args)
    drawer <- function(start, stopi) counts[, start:stopi, drop = FALSE]

    dens.fast <- fast.fun(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = dens.bw,
      B = B,
      cdf = FALSE,
      counts = counts
    )
    dens.drawer <- fast.fun(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = dens.bw,
      B = B,
      cdf = FALSE,
      counts.drawer = drawer
    )
    dens.oracle.drawer <- oracle.fun(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = dens.bw,
      B = B,
      cdf = FALSE,
      counts.drawer = drawer
    )

    dist.fast <- fast.fun(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = dist.bw,
      B = B,
      cdf = TRUE,
      counts = counts
    )
    dist.drawer <- fast.fun(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = dist.bw,
      B = B,
      cdf = TRUE,
      counts.drawer = drawer
    )
    dist.oracle.drawer <- oracle.fun(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = dist.bw,
      B = B,
      cdf = TRUE,
      counts.drawer = drawer
    )

    expect_equal(dens.drawer$t, dens.fast$t, tolerance = 1e-12, info = paste("density drawer chunk invariance", cfg$name))
    expect_equal(dens.drawer$t0, dens.fast$t0, tolerance = 1e-10, info = paste("density drawer chunk invariance t0", cfg$name))
    expect_equal(dens.drawer$t, dens.oracle.drawer$t, tolerance = 1e-8, info = paste("density drawer oracle repeated-x", cfg$name))
    expect_equal(dens.drawer$t0, dens.oracle.drawer$t0, tolerance = 1e-10, info = paste("density drawer oracle repeated-x t0", cfg$name))

    expect_equal(dist.drawer$t, dist.fast$t, tolerance = 1e-12, info = paste("distribution drawer chunk invariance", cfg$name))
    expect_equal(dist.drawer$t0, dist.fast$t0, tolerance = 1e-10, info = paste("distribution drawer chunk invariance t0", cfg$name))
    expect_equal(dist.drawer$t, dist.oracle.drawer$t, tolerance = 1e-8, info = paste("distribution drawer oracle repeated-x", cfg$name))
    expect_equal(dist.drawer$t0, dist.oracle.drawer$t0, tolerance = 1e-10, info = paste("distribution drawer oracle repeated-x t0", cfg$name))
  }
})

test_that("fixed density/distribution helpers preserve bounded kernel options", {
  skip_if_not_installed("np")

  set.seed(603103)
  n <- 60
  tx <- data.frame(x = runif(n))
  ty <- data.frame(y = runif(n))
  ex <- data.frame(x = seq(0.02, 0.98, length.out = 13))
  ey <- data.frame(y = seq(0.03, 0.97, length.out = 13))
  B <- 5L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  u.dens.bw <- npudensbw(
    dat = tx,
    bws = 0.20,
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerbound = "fixed",
    ckerlb = 0.0,
    ckerub = 1.0
  )
  u.dist.bw <- npudistbw(
    dat = tx,
    bws = 0.20,
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerbound = "fixed",
    ckerlb = 0.0,
    ckerub = 1.0
  )
  c.dens.bw <- npcdensbw(
    xdat = tx,
    ydat = ty,
    bws = c(0.25, 0.25),
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    cxkertype = "epanechnikov",
    cykertype = "epanechnikov",
    cxkerbound = "fixed",
    cykerbound = "fixed",
    cxkerlb = 0.0,
    cykerlb = 0.0,
    cxkerub = 1.0,
    cykerub = 1.0
  )
  c.dist.bw <- npcdistbw(
    xdat = tx,
    ydat = ty,
    bws = c(0.25, 0.25),
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    cxkertype = "epanechnikov",
    cykertype = "epanechnikov",
    cxkerbound = "fixed",
    cykerbound = "fixed",
    cxkerlb = 0.0,
    cykerlb = 0.0,
    cxkerub = 1.0,
    cykerub = 1.0
  )

  fast.u <- getFromNamespace(".np_inid_boot_from_ksum_unconditional", "np")
  fast.c <- getFromNamespace(".np_inid_boot_from_ksum_conditional", "np")

  u.dens.fast <- fast.u(
    xdat = tx,
    exdat = ex,
    bws = u.dens.bw,
    B = B,
    operator = "normal",
    counts = counts
  )
  u.dist.fast <- fast.u(
    xdat = tx,
    exdat = ex,
    bws = u.dist.bw,
    B = B,
    operator = "integral",
    counts = counts
  )
  c.dens.fast <- fast.c(
    xdat = tx,
    ydat = ty,
    exdat = ex,
    eydat = ey,
    bws = c.dens.bw,
    B = B,
    cdf = FALSE,
    counts = counts
  )
  c.dist.fast <- fast.c(
    xdat = tx,
    ydat = ty,
    exdat = ex,
    eydat = ey,
    bws = c.dist.bw,
    B = B,
    cdf = TRUE,
    counts = counts
  )

  u.dens.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  u.dist.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  c.dens.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  c.dist.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  for (b in seq_len(B)) {
    idx <- rep.int(seq_len(n), counts[, b])
    u.dens.explicit[b, ] <- npudens(tdat = tx[idx, , drop = FALSE], edat = ex, bws = u.dens.bw)$dens
    u.dist.explicit[b, ] <- npudist(tdat = tx[idx, , drop = FALSE], edat = ex, bws = u.dist.bw)$dist
    c.dens.explicit[b, ] <- npcdens(
      txdat = tx[idx, , drop = FALSE],
      tydat = ty[idx, , drop = FALSE],
      exdat = ex,
      eydat = ey,
      bws = c.dens.bw
    )$condens
    c.dist.explicit[b, ] <- npcdist(
      txdat = tx[idx, , drop = FALSE],
      tydat = ty[idx, , drop = FALSE],
      exdat = ex,
      eydat = ey,
      bws = c.dist.bw
    )$condist
  }

  expect_equal(u.dens.fast$t, u.dens.explicit, tolerance = 1e-10)
  expect_equal(u.dist.fast$t, u.dist.explicit, tolerance = 1e-10)
  expect_equal(c.dens.fast$t, c.dens.explicit, tolerance = 1e-10)
  expect_equal(c.dist.fast$t, c.dist.explicit, tolerance = 1e-10)
})

test_that("fixed-bwtype unsupervised plot bootstrap covers inid fixed and geom", {
  skip_if_not_installed("np")

  set.seed(603104)
  n <- 48
  xdat <- data.frame(x = rnorm(n))
  ydat <- data.frame(y = rnorm(n))

  u.dens.bw <- npudensbw(dat = xdat, bws = 0.30, bandwidth.compute = FALSE, bwtype = "fixed")
  u.dist.bw <- npudistbw(dat = xdat, bws = 0.30, bandwidth.compute = FALSE, bwtype = "fixed")
  c.dens.bw <- npcdensbw(xdat = xdat, ydat = ydat, bws = c(0.35, 0.35), bandwidth.compute = FALSE, bwtype = "fixed")
  c.dist.bw <- npcdistbw(xdat = xdat, ydat = ydat, bws = c(0.35, 0.35), bandwidth.compute = FALSE, bwtype = "fixed")

  run_unsup_plot <- function(bw, ..., boot.method) {
    suppressWarnings(plot(
      bw,
      output = "data",
      perspective = FALSE,
      errors = "bootstrap",
      bootstrap = boot.method,
      boot_control = np_boot_control(blocklen = 3L),
      B = 5L,
      band = "pointwise",
      neval = 11L,
      ...
    ))
  }

  for (boot.method in c("inid", "fixed", "geom")) {
    expect_type(run_unsup_plot(u.dens.bw, xdat = xdat, boot.method = boot.method), "list")
    expect_type(run_unsup_plot(u.dist.bw, xdat = xdat, boot.method = boot.method), "list")
    expect_type(run_unsup_plot(c.dens.bw, xdat = xdat, ydat = ydat, view = "fixed", boot.method = boot.method), "list")
    expect_type(run_unsup_plot(c.dist.bw, xdat = xdat, ydat = ydat, view = "fixed", boot.method = boot.method), "list")
  }
})

test_that("formula npudensbw matches default interface with subset/na.action", {
  set.seed(20260222)
  dat <- data.frame(
    x1 = runif(30),
    x2 = rnorm(30)
  )
  dat$x2[c(4, 17)] <- NA_real_
  bw_formula <- np::npudensbw(
    ~ x1 + x2,
    data = dat,
    subset = x1 > 0.2,
    na.action = na.omit,
    bws = c(0.4, 0.6),
    bandwidth.compute = FALSE
  )

  mf <- model.frame(~ x1 + x2, data = dat, subset = x1 > 0.2, na.action = na.omit)
  bw_default <- np::npudensbw(
    dat = mf,
    bws = c(0.4, 0.6),
    bandwidth.compute = FALSE
  )

  expect_equal(as.numeric(bw_formula$bw), as.numeric(bw_default$bw))
})

test_that("formula npudistbw matches default interface with subset/na.action", {
  set.seed(20260222)
  dat <- data.frame(
    x1 = runif(30),
    x2 = rnorm(30)
  )
  dat$x1[c(2, 11)] <- NA_real_
  bw_formula <- np::npudistbw(
    ~ x1 + x2,
    data = dat,
    subset = x2 > -0.5,
    na.action = na.omit,
    bws = c(0.3, 0.5),
    bandwidth.compute = FALSE
  )

  mf <- model.frame(~ x1 + x2, data = dat, subset = x2 > -0.5, na.action = na.omit)
  bw_default <- np::npudistbw(
    dat = mf,
    bws = c(0.3, 0.5),
    bandwidth.compute = FALSE
  )

  expect_equal(as.numeric(bw_formula$bw), as.numeric(bw_default$bw))
})

test_that("formula npudistbw gdata path matches default interface", {
  set.seed(20260222)
  dat <- data.frame(
    x1 = runif(30),
    x2 = rnorm(30)
  )
  gdat <- data.frame(
    x1 = runif(22),
    x2 = rnorm(22)
  )

  bw_formula <- np::npudistbw(
    ~ x1 + x2,
    data = dat,
    gdata = gdat,
    bws = c(0.3, 0.5),
    bandwidth.compute = FALSE
  )

  mf <- model.frame(~ x1 + x2, data = dat)
  gmf <- model.frame(~ x1 + x2, data = gdat)
  bw_default <- np::npudistbw(
    dat = mf,
    gdat = gmf,
    bws = c(0.3, 0.5),
    bandwidth.compute = FALSE
  )

  expect_equal(as.numeric(bw_formula$bw), as.numeric(bw_default$bw))
})

test_that("formula npregbw matches default interface with subset/na.action", {
  set.seed(20260222)
  dat <- data.frame(
    y = rnorm(40),
    x1 = runif(40),
    x2 = rnorm(40)
  )
  dat$y[c(3, 9)] <- NA_real_
  bw_formula <- np::npregbw(
    y ~ x1 + x2,
    data = dat,
    subset = x1 < 0.85,
    na.action = na.omit,
    bws = c(0.45, 0.55),
    bandwidth.compute = FALSE,
    regtype = "ll"
  )

  mf <- model.frame(y ~ x1 + x2, data = dat, subset = x1 < 0.85, na.action = na.omit)
  bw_default <- np::npregbw(
    xdat = mf[, c("x1", "x2"), drop = FALSE],
    ydat = model.response(mf),
    bws = c(0.45, 0.55),
    bandwidth.compute = FALSE,
    regtype = "ll"
  )

  expect_equal(as.numeric(bw_formula$bw), as.numeric(bw_default$bw))
})

test_that("formula npcdistbw gdata path matches default interface", {
  set.seed(20260222)
  dat <- data.frame(
    y = rnorm(36),
    x = runif(36)
  )
  gdat <- data.frame(
    y = rnorm(18),
    x = runif(18)
  )

  bw_formula <- np::npcdistbw(
    y ~ x,
    data = dat,
    gdata = gdat,
    bws = c(0.35, 0.45),
    bandwidth.compute = FALSE
  )

  mf <- model.frame(y ~ x, data = dat)
  gmf <- model.frame(y ~ x, data = gdat)
  bw_default <- np::npcdistbw(
    xdat = mf[, "x", drop = FALSE],
    ydat = mf[, "y", drop = FALSE],
    gydat = gmf[, "y", drop = FALSE],
    bws = c(0.35, 0.45),
    bandwidth.compute = FALSE
  )

  expect_equal(as.numeric(bw_formula$bw), as.numeric(bw_default$bw))
})

test_that("formula npcdensbw matches default interface with subset/na.action", {
  set.seed(20260222)
  dat <- data.frame(
    y = rnorm(36),
    x = runif(36)
  )
  dat$y[c(5, 19)] <- NA_real_

  bw_formula <- np::npcdensbw(
    y ~ x,
    data = dat,
    subset = x > 0.25,
    na.action = na.omit,
    bws = c(0.35, 0.45),
    bandwidth.compute = FALSE
  )

  mf <- model.frame(y ~ x, data = dat, subset = x > 0.25, na.action = na.omit)
  bw_default <- np::npcdensbw(
    xdat = mf[, "x", drop = FALSE],
    ydat = mf[, "y", drop = FALSE],
    bws = c(0.35, 0.45),
    bandwidth.compute = FALSE
  )

  expect_equal(as.numeric(bw_formula$bw), as.numeric(bw_default$bw))
})

test_that("formula npindexbw matches default interface with subset/na.action", {
  set.seed(20260222)
  dat <- data.frame(
    y = rnorm(42),
    x1 = runif(42),
    x2 = rnorm(42)
  )
  dat$y[c(4, 21)] <- NA_real_

  bw_formula <- np::npindexbw(
    y ~ x1 + x2,
    data = dat,
    subset = x1 < 0.9,
    na.action = na.omit,
    bws = c(0.2, 0.3, 0.4),
    bandwidth.compute = FALSE,
    method = "ichimura",
    nmulti = 1
  )

  mf <- model.frame(y ~ x1 + x2, data = dat, subset = x1 < 0.9, na.action = na.omit)
  bw_default <- np::npindexbw(
    xdat = mf[, c("x1", "x2"), drop = FALSE],
    ydat = model.response(mf),
    bws = c(0.2, 0.3, 0.4),
    bandwidth.compute = FALSE,
    method = "ichimura",
    nmulti = 1
  )

  expect_equal(as.numeric(bw_formula$bw), as.numeric(bw_default$bw))
})

test_that("formula npplregbw matches default interface with subset/na.action", {
  set.seed(20260222)
  dat <- data.frame(
    y = rnorm(44),
    x = runif(44),
    z = rnorm(44)
  )
  bws_mat <- matrix(c(0.25, 0.35), nrow = 2, ncol = 1)

  bw_formula <- np::npplregbw(
    y ~ z | x,
    data = dat,
    subset = x < 0.92,
    na.action = na.omit,
    bws = bws_mat,
    bandwidth.compute = FALSE
  )

  mf <- model.frame(y ~ x + z, data = dat, subset = x < 0.92, na.action = na.omit)
  bw_default <- np::npplregbw(
    xdat = mf[, "z", drop = FALSE],
    ydat = model.response(mf),
    zdat = mf[, "x", drop = FALSE],
    bws = bws_mat,
    bandwidth.compute = FALSE
  )

  bw_formula_vec <- vapply(bw_formula$bw, function(bwi) as.numeric(bwi$bw)[1L], numeric(1))
  bw_default_vec <- vapply(bw_default$bw, function(bwi) as.numeric(bwi$bw)[1L], numeric(1))
  expect_equal(bw_formula_vec, bw_default_vec)
})

test_that("formula npscoefbw matches default interface with subset/na.action", {
  set.seed(20260222)
  dat <- data.frame(
    y = rnorm(44),
    x = runif(44),
    z = rnorm(44)
  )
  dat$y[c(8, 29)] <- NA_real_

  bw_formula <- np::npscoefbw(
    y ~ x | z,
    data = dat,
    subset = x > 0.08,
    na.action = na.omit,
    bws = 0.4,
    bandwidth.compute = FALSE
  )

  mf <- model.frame(y ~ x + z, data = dat, subset = x > 0.08, na.action = na.omit)
  bw_default <- np::npscoefbw(
    xdat = mf[, "x", drop = FALSE],
    ydat = model.response(mf),
    zdat = mf[, "z", drop = FALSE],
    bws = 0.4,
    bandwidth.compute = FALSE
  )

  expect_equal(as.numeric(bw_formula$bw), as.numeric(bw_default$bw))
})

npudensbw_bandwidth <- getFromNamespace("npudensbw.bandwidth", "np")
npudistbw_dbandwidth <- getFromNamespace("npudistbw.dbandwidth", "np")
npregbw_rbandwidth <- getFromNamespace("npregbw.rbandwidth", "np")
npcdensbw_conbandwidth <- getFromNamespace("npcdensbw.conbandwidth", "np")
npcdistbw_condbandwidth <- getFromNamespace("npcdistbw.condbandwidth", "np")
npindexbw_sibandwidth <- getFromNamespace("npindexbw.sibandwidth", "np")
npscoefbw_scbandwidth <- getFromNamespace("npscoefbw.scbandwidth", "np")
npplregbw_default <- getFromNamespace("npplregbw.default", "np")
npplregbw_plbandwidth <- getFromNamespace("npplregbw.plbandwidth", "np")

test_that("bandwidth wrappers reject invalid scalar control flags", {
  dat <- data.frame(x = c(0.1, 0.2))
  bws <- list()
  cdens.bws <- npcdensbw(xdat = dat, ydat = dat, bws = c(1, 1),
                         bandwidth.compute = FALSE)
  cdist.bws <- npcdistbw(xdat = dat, ydat = dat, bws = c(1, 1),
                         bandwidth.compute = FALSE)

  expect_error(
    npudensbw_bandwidth(dat = dat, bws = bws, bandwidth.compute = c(TRUE, FALSE)),
    "'bandwidth.compute' must be TRUE or FALSE"
  )
  expect_error(
    npudensbw_bandwidth(dat = dat, bws = bws, nmulti = -1),
    "'nmulti' must be a positive integer"
  )
  expect_error(
    npudensbw_bandwidth(dat = dat, bws = bws, itmax = 0),
    "'itmax' must be a positive integer"
  )
  expect_error(
    npudensbw_bandwidth(dat = dat, bws = bws, ftol = 0),
    "'ftol' must be a positive finite numeric scalar"
  )
  expect_error(
    npudensbw_bandwidth(dat = dat, bws = bws, penalty.multiplier = 0),
    "'penalty.multiplier' must be a positive finite numeric scalar"
  )

  expect_error(
    npudistbw_dbandwidth(dat = dat, bws = bws, do.full.integral = c(TRUE, FALSE)),
    "'do.full.integral' must be TRUE or FALSE"
  )
  expect_error(
    npudistbw_dbandwidth(dat = dat, bws = bws, ngrid = 0),
    "'ngrid' must be a positive integer"
  )
  expect_error(
    npudistbw_dbandwidth(dat = dat, bws = bws, memfac = 0),
    "'memfac' must be a positive finite numeric scalar"
  )

  expect_error(
    npregbw_rbandwidth(xdat = dat, ydat = dat$x, bws = bws, bandwidth.compute = c(TRUE, FALSE)),
    "'bandwidth.compute' must be TRUE or FALSE"
  )
  expect_error(
    npregbw_rbandwidth(xdat = dat, ydat = dat$x, bws = bws, itmax = 0),
    "'itmax' must be a positive integer"
  )

  expect_error(
    npcdensbw_conbandwidth(xdat = dat, ydat = dat, bws = cdens.bws, memfac = 0),
    "'memfac' must be a positive finite numeric scalar"
  )
  expect_error(
    npcdensbw_conbandwidth(xdat = dat, ydat = dat, bws = cdens.bws, nmulti = -1),
    "'nmulti' must be a positive integer"
  )

  expect_error(
    npcdistbw_condbandwidth(xdat = dat, ydat = dat, bws = cdist.bws, do.full.integral = c(TRUE, FALSE)),
    "'do.full.integral' must be TRUE or FALSE"
  )
  expect_error(
    npcdistbw_condbandwidth(xdat = dat, ydat = dat, bws = cdist.bws, ngrid = 0),
    "'ngrid' must be a positive integer"
  )

  expect_error(
    npindexbw_sibandwidth(xdat = dat, ydat = dat$x, bws = bws, bandwidth.compute = c(TRUE, FALSE)),
    "'bandwidth.compute' must be TRUE or FALSE"
  )
  expect_error(
    npindexbw_sibandwidth(xdat = dat, ydat = dat$x, bws = bws, optim.maxit = 0),
    "'optim.maxit' must be a positive integer"
  )
  expect_error(
    npindexbw_sibandwidth(xdat = dat, ydat = dat$x, bws = bws, scale.factor.init.lower = 0.6, scale.factor.init.upper = 0.5),
    "'scale.factor.init.upper' must be greater than or equal to max"
  )

  expect_error(
    npscoefbw_scbandwidth(xdat = dat, ydat = dat$x, bws = bws, cv.iterate = c(TRUE, FALSE)),
    "'cv.iterate' must be TRUE or FALSE"
  )
  expect_error(
    npscoefbw_scbandwidth(xdat = dat, ydat = dat$x, bws = bws, backfit.tol = 0),
    "'backfit.tol' must be a positive finite numeric scalar"
  )
  expect_error(
    npscoefbw_scbandwidth(xdat = dat, ydat = dat$x, bws = bws, scale.factor.init.lower = 0.8, scale.factor.init.upper = 0.6),
    "'scale.factor.init.upper' must be greater than or equal to max"
  )
  expect_error(
    npscoefbw_scbandwidth(xdat = dat, ydat = dat$x, bws = bws, hbd.init = 2.1),
    "categorical start factors must be less than or equal to 2"
  )

  expect_error(
    npplregbw_default(xdat = dat, ydat = dat$x, zdat = dat,
                      bws = matrix(0, nrow = 2, ncol = 1),
                      bandwidth.compute = c(TRUE, FALSE)),
    "'bandwidth.compute' must be TRUE or FALSE"
  )
  expect_error(
    npplregbw_plbandwidth(xdat = dat, ydat = dat$x, zdat = dat, bws = bws, nmulti = -1),
    "'nmulti' must be a positive integer"
  )
})

test_that("conditional density/distribution gradient bootstrap inid works for bandwidth objects", {
  skip_if_not_installed("np")

  library(np)

  set.seed(20260225)
  n <- 24L
  x1 <- factor(sample(c("a", "b"), n, replace = TRUE))
  x2 <- rnorm(n)
  y <- rnorm(n)

  xdat <- data.frame(x1 = x1, x2 = x2)
  ydat <- data.frame(y = y)

  bw.cd <- suppressWarnings(npcdensbw(xdat = xdat, ydat = ydat, nmulti = 1L))
  bw.cdist <- suppressWarnings(npcdistbw(xdat = xdat, ydat = ydat, nmulti = 1L))

  for (obj in list(bw.cd, bw.cdist)) {
    out <- suppressWarnings(
      plot(
        obj,
        xdat = xdat,
        ydat = ydat,
        output = "data",
        perspective = FALSE,
        gradients = TRUE,
        errors = "bootstrap",
        bootstrap = "inid",
        B = 5L
      )
    )
    expect_type(out, "list")
    expect_true(length(out[[1L]]$bxp) > 0L)
  }
})

test_that("conditional density/distribution gradient bootstrap inid works for fitted objects", {
  skip_if_not_installed("np")

  library(np)

  set.seed(20260226)
  n <- 24L
  x1 <- factor(sample(c("a", "b"), n, replace = TRUE))
  x2 <- rnorm(n)
  y <- rnorm(n)

  xdat <- data.frame(x1 = x1, x2 = x2)
  ydat <- data.frame(y = y)

  bw.cd <- suppressWarnings(npcdensbw(xdat = xdat, ydat = ydat, nmulti = 1L))
  bw.cdist <- suppressWarnings(npcdistbw(xdat = xdat, ydat = ydat, nmulti = 1L))
  fit.cd <- npcdens(bws = bw.cd)
  fit.cdist <- npcdist(bws = bw.cdist)

  for (obj in list(fit.cd, fit.cdist)) {
    out <- suppressWarnings(
      plot(
        obj,
        output = "data",
        perspective = FALSE,
        gradients = TRUE,
        errors = "bootstrap",
        bootstrap = "inid",
        B = 5L
      )
    )
    expect_type(out, "list")
    expect_true(length(out[[1L]]$bxp) > 0L)
  }
})

library(np)

empty_shadow_matrix <- function(n) {
  matrix(numeric(0), nrow = n, ncol = 0)
}

shadow_cker_code <- function(kernel) {
  switch(kernel,
    gaussian = 0L,
    epanechnikov = 4L,
    uniform = 8L,
    truncated = 9L,
    stop("unsupported continuous kernel for shadow test")
  )
}

shadow_uker_code <- function(kernel) {
  switch(kernel,
    aitchisonaitken = 0L,
    liracine = 1L,
    stop("unsupported unordered kernel for shadow test")
  )
}

shadow_oker_code <- function(kernel) {
  switch(kernel,
    wangvanryzin = 0L,
    liracine = 2L,
    racineliyan = 3L,
    stop("unsupported ordered kernel for shadow test")
  )
}

shadow_basis_code <- function(basis) {
  switch(basis,
    additive = 0L,
    glp = 1L,
    tensor = 2L,
    stop("unsupported LP basis for shadow test")
  )
}

shadow_rbw <- function(bw) {
  c(
    bw$xbw[bw$ixcon],
    bw$ybw[bw$iycon],
    bw$ybw[bw$iyuno],
    bw$ybw[bw$iyord],
    bw$xbw[bw$ixuno],
    bw$xbw[bw$ixord]
  )
}

shadow_bwtype_code <- function(bw) {
  switch(bw$type,
    fixed = 0L,
    generalized_nn = 1L,
    adaptive_nn = 2L,
    stop("unsupported bandwidth type for shadow test")
  )
}

shadow_reg_code <- function(bw) {
  if (identical(bw$regtype.engine, "lp")) 2L else 0L
}

shadow_degree <- function(bw) {
  if (identical(bw$regtype.engine, "lp")) as.integer(bw$degree.engine) else integer(0)
}

shadow_safe_call <- function(name, ...) {
  on.exit(
    tryCatch(.Call("C_np_shadow_reset_state", PACKAGE = "np"),
             error = function(e) NULL),
    add = TRUE
  )
  .Call(name, ..., PACKAGE = "np")
}

call_shadow_density <- function(bw, x, y, criterion = c("cv.ml", "cv.ls"),
                                use_tree = FALSE, compare_old = TRUE) {
  criterion <- match.arg(criterion)
  n <- nrow(x)
  shadow_safe_call(
    "C_np_shadow_cv_density_conditional",
    empty_shadow_matrix(n), empty_shadow_matrix(n), as.matrix(y),
    empty_shadow_matrix(n), empty_shadow_matrix(n), as.matrix(x),
    as.double(shadow_rbw(bw)),
    shadow_bwtype_code(bw),
    shadow_cker_code(bw$cykertype),
    shadow_uker_code(bw$uykertype),
    shadow_oker_code(bw$oykertype),
    shadow_cker_code(bw$cxkertype),
    shadow_uker_code(bw$uxkertype),
    shadow_oker_code(bw$oxkertype),
    use_tree,
    if (identical(criterion, "cv.ml")) 0L else 1L,
    shadow_reg_code(bw),
    shadow_degree(bw),
    isTRUE(bw$bernstein.basis.engine),
    shadow_basis_code(bw$basis.engine),
    compare_old
  )
}

call_shadow_distribution <- function(bw, x, ytrain, yeval = ytrain,
                                     use_tree = FALSE, cdfontrain = FALSE,
                                     compare_old = TRUE) {
  n <- nrow(x)
  ne <- nrow(yeval)
  shadow_safe_call(
    "C_np_shadow_cv_distribution_conditional",
    empty_shadow_matrix(n), empty_shadow_matrix(n), as.matrix(ytrain),
    empty_shadow_matrix(ne), empty_shadow_matrix(ne), as.matrix(yeval),
    empty_shadow_matrix(n), empty_shadow_matrix(n), as.matrix(x),
    as.double(shadow_rbw(bw)),
    shadow_bwtype_code(bw),
    shadow_cker_code(bw$cykertype),
    shadow_uker_code(bw$uykertype),
    shadow_oker_code(bw$oykertype),
    shadow_cker_code(bw$cxkertype),
    shadow_uker_code(bw$uxkertype),
    shadow_oker_code(bw$oxkertype),
    use_tree,
    shadow_reg_code(bw),
    shadow_degree(bw),
    isTRUE(bw$bernstein.basis.engine),
    shadow_basis_code(bw$basis.engine),
    cdfontrain,
    compare_old
  )
}

call_shadow_xweights_row <- function(bw, x, y, row_index, use_tree = FALSE) {
  n <- nrow(x)
  shadow_safe_call(
    "C_np_shadow_cv_xweights_conditional",
    empty_shadow_matrix(n), empty_shadow_matrix(n), as.matrix(y),
    empty_shadow_matrix(n), empty_shadow_matrix(n), as.matrix(x),
    as.double(shadow_rbw(bw)),
    shadow_bwtype_code(bw),
    shadow_cker_code(bw$cxkertype),
    shadow_uker_code(bw$uxkertype),
    shadow_oker_code(bw$oxkertype),
    use_tree,
    shadow_reg_code(bw),
    shadow_degree(bw),
    isTRUE(bw$bernstein.basis.engine),
    shadow_basis_code(bw$basis.engine),
    as.integer(row_index)
  )
}

test_that("shadow reset state is a harmless no-op when inactive", {
  expect_null(.Call("C_np_shadow_reset_state", PACKAGE = "np"))
  expect_null(.Call("C_np_shadow_reset_state", PACKAGE = "np"))
})

test_that("shadow density lc matches legacy cv objectives", {
  set.seed(42)
  n <- 40L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = rnorm(n))
  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.35, 0.45, 0.55),
    bandwidth.compute = FALSE,
    regtype = "lc"
  )

  res_ml <- call_shadow_density(bw, x, y, criterion = "cv.ml", compare_old = TRUE)
  res_ls <- call_shadow_density(bw, x, y, criterion = "cv.ls", compare_old = TRUE)

  expect_equal(res_ml$new, res_ml$old, tolerance = 1e-12)
  expect_equal(res_ls$new, res_ls$old, tolerance = 2e-3)
})

test_that("shadow density lp preserves ll canonicalization and tree parity", {
  set.seed(33)
  n <- 42L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = x$x1 + rnorm(n, sd = 0.15))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.31, 0.43, 0.57),
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  bw.lp <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.31, 0.43, 0.57),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = degree
  )

  ll_ml <- call_shadow_density(bw.ll, x, y, criterion = "cv.ml", compare_old = FALSE)
  lp_ml <- call_shadow_density(bw.lp, x, y, criterion = "cv.ml", compare_old = FALSE)
  ll_ls <- call_shadow_density(bw.ll, x, y, criterion = "cv.ls", compare_old = FALSE)
  lp_ls <- call_shadow_density(bw.lp, x, y, criterion = "cv.ls", compare_old = FALSE)
  lp_ml_tree <- call_shadow_density(bw.lp, x, y, criterion = "cv.ml", use_tree = TRUE, compare_old = FALSE)
  lp_ls_tree <- call_shadow_density(bw.lp, x, y, criterion = "cv.ls", use_tree = TRUE, compare_old = FALSE)

  expect_equal(ll_ml$new, lp_ml$new, tolerance = 1e-12)
  expect_equal(ll_ls$new, lp_ls$new, tolerance = 1e-12)
  expect_equal(lp_ml_tree$new, lp_ml$new, tolerance = 1e-10)
  expect_equal(lp_ls_tree$new, lp_ls$new, tolerance = 1e-12)
  expect_equal(ll_ml$prod, ll_ml$new, tolerance = 1e-12)
  expect_equal(lp_ml$prod, lp_ml$new, tolerance = 1e-12)
  expect_equal(lp_ml_tree$prod, lp_ml_tree$new, tolerance = 1e-10)
  expect_equal(ll_ls$prod, ll_ls$new, tolerance = 1e-12)
  expect_equal(lp_ls$prod, lp_ls$new, tolerance = 1e-12)
  expect_equal(lp_ls_tree$prod, lp_ls_tree$new, tolerance = 1e-12)
})

test_that("shadow density cvml preserves the large-kernel X collapse", {
  set.seed(11)
  n <- 30L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = rnorm(n))
  big <- 1e6
  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(big, big, big),
    bandwidth.compute = FALSE,
    regtype = "lc"
  )

  res <- call_shadow_density(bw, x, y, criterion = "cv.ml", compare_old = FALSE)
  collapsed <- -n * log(mean(dnorm(outer(y$y1, y$y1, "-"), sd = big)))

  expect_equal(res$new, collapsed, tolerance = 1e-9)
})

test_that("shadow density generalized-nn cvml preserves ll canonicalization", {
  set.seed(202)
  n <- 38L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = sin(2 * pi * x$x1) + x$x2 + rnorm(n, sd = 0.12))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(4, 7, 5),
    bwtype = "generalized_nn",
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  bw.lp <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(4, 7, 5),
    bwtype = "generalized_nn",
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = degree
  )

  ll_ml <- call_shadow_density(bw.ll, x, y, criterion = "cv.ml", compare_old = FALSE)
  lp_ml <- call_shadow_density(bw.lp, x, y, criterion = "cv.ml", compare_old = FALSE)

  expect_true(is.finite(ll_ml$new))
  expect_true(is.finite(lp_ml$new))
  expect_equal(ll_ml$new, lp_ml$new, tolerance = 1e-10)
  expect_true(is.finite(ll_ml$prod))
  expect_true(is.finite(lp_ml$prod))
  expect_equal(ll_ml$prod, lp_ml$prod, tolerance = 1e-10)
})

test_that("shadow density generalized-nn LP cures the legacy penalty collapse", {
  set.seed(203)
  n <- 38L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = sin(2 * pi * x$x1) + x$x2 + rnorm(n, sd = 0.12))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(4, 7, 5),
    bwtype = "generalized_nn",
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  bw.lp <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(4, 7, 5),
    bwtype = "generalized_nn",
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = degree
  )

  ll_ls <- call_shadow_density(bw.ll, x, y, criterion = "cv.ls", compare_old = FALSE)
  lp_ls <- call_shadow_density(bw.lp, x, y, criterion = "cv.ls", compare_old = FALSE)

  expect_true(is.finite(ll_ls$new))
  expect_true(is.finite(lp_ls$new))
  expect_equal(ll_ls$new, lp_ls$new, tolerance = 1e-10)
  expect_equal(ll_ls$prod, ll_ls$new, tolerance = 1e-10)
  expect_equal(lp_ls$prod, lp_ls$new, tolerance = 1e-10)
  expect_gt(ll_ls$new, -1e6)
})

test_that("shadow fixed-bandwidth X-side row helper matches dense oracle and leaves self out", {
  set.seed(512)
  n <- 18L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = sin(2 * pi * x$x1) + x$x2 + rnorm(n, sd = 0.08))
  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.28, 0.37, 0.41),
    bandwidth.compute = FALSE,
    regtype = "ll"
  )

  for (row in seq_len(n)) {
    res <- call_shadow_xweights_row(bw, x, y, row_index = row)
    expect_equal(res$streamed, res$dense, tolerance = 1e-12)
    expect_equal(res$streamed[row], 0, tolerance = 1e-12)
    expect_equal(sum(res$streamed), 1, tolerance = 1e-10)
  }
})

test_that("shadow fixed-bandwidth X-side row helper preserves ll == lp degree-1 and tree parity", {
  set.seed(513)
  n <- 20L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = x$x1 - x$x2 + rnorm(n, sd = 0.06))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.31, 0.36, 0.42),
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  bw.lp <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.31, 0.36, 0.42),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = degree
  )

  res.ll <- call_shadow_xweights_row(bw.ll, x, y, row_index = 7L)
  res.lp <- call_shadow_xweights_row(bw.lp, x, y, row_index = 7L)
  res.tree <- call_shadow_xweights_row(bw.lp, x, y, row_index = 7L, use_tree = TRUE)

  expect_equal(res.ll$streamed, res.lp$streamed, tolerance = 1e-12)
  expect_equal(res.lp$streamed, res.tree$streamed, tolerance = 1e-12)
})

test_that("shadow generalized-nn X-side row helper matches dense oracle and preserves ll == lp", {
  set.seed(514)
  n <- 22L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = x$x1 + x$x2 + rnorm(n, sd = 0.07))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(5, 7, 6),
    bwtype = "generalized_nn",
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  bw.lp <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(5, 7, 6),
    bwtype = "generalized_nn",
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = degree
  )

  res.ll <- call_shadow_xweights_row(bw.ll, x, y, row_index = 9L)
  res.lp <- call_shadow_xweights_row(bw.lp, x, y, row_index = 9L)

  expect_equal(res.ll$streamed, res.ll$dense, tolerance = 1e-12)
  expect_equal(res.lp$streamed, res.lp$dense, tolerance = 1e-12)
  expect_equal(res.ll$streamed, res.lp$streamed, tolerance = 1e-10)
  expect_equal(sum(res.ll$streamed), 1, tolerance = 1e-10)
  expect_equal(res.ll$streamed[9L], 0, tolerance = 1e-12)
})

locate_shadow_proof_src <- function(filename) {
  candidates <- c(
    testthat::test_path("..", "..", "src", filename),
    testthat::test_path("..", "..", "..", "src", filename),
    file.path(Sys.getenv("R_PACKAGE_DIR", ""), "src", filename),
    file.path(Sys.getenv("R_PACKAGE_SOURCE", ""), "src", filename),
    file.path(getwd(), "src", filename),
    file.path(getwd(), "..", "src", filename)
  )
  candidates <- unique(candidates[nzchar(candidates)])
  hits <- candidates[file.exists(candidates)]
  if (length(hits) == 0L) {
    return(NULL)
  }
  hits[[1L]]
}

read_shadow_proof_src <- function(filename) {
  src_file <- locate_shadow_proof_src(filename)
  skip_if(
    is.null(src_file),
    sprintf("source file src/%s unavailable in this test context", filename)
  )
  readLines(src_file, warn = FALSE)
}

test_that("shadow fixed-bandwidth X-side row helper stays row-streamed", {
  lines <- read_shadow_proof_src("jksum.c")
  start <- grep("^int np_shadow_proof_conditional_x_weight_row_fixed\\(", lines)
  stop <- grep("^static int np_shadow_conditional_build_y_matrix\\(", lines)

  expect_length(start, 1L)
  expect_length(stop, 1L)
  expect_lt(start, stop)

  body <- paste(lines[(start + 1L):(stop - 1L)], collapse = "\n")

  expect_false(grepl("(malloc|calloc|alloc_matd|alloc_tmatd)\\([^\\)]*(num_train|num_obs_train_extern)[^\\)]*(num_train|num_obs_train_extern)", body))
})

test_that("shadow generalized-nn X-side row helper stays row-streamed", {
  lines <- read_shadow_proof_src("jksum.c")
  start <- grep("^int np_shadow_proof_conditional_x_weight_row_stream\\(", lines)
  stop <- grep("^static int np_shadow_conditional_build_y_matrix\\(", lines)

  expect_length(start, 1L)
  expect_length(stop, 1L)
  expect_lt(start, stop)

  body <- paste(lines[(start + 1L):(stop - 1L)], collapse = "\n")

  expect_false(grepl("(malloc|calloc|alloc_matd|alloc_tmatd)\\([^\\)]*(num_train|num_obs_train_extern)[^\\)]*(num_train|num_obs_train_extern)", body))
})

test_that("fixed-bandwidth cvml LP stream avoids dense row storage", {
  lines <- read_shadow_proof_src("jksum.c")
  start <- grep("^int np_conditional_density_cvml_lp_stream\\(", lines)
  stop <- grep("^static int np_shadow_conditional_build_y_matrix\\(", lines)

  expect_length(start, 1L)
  expect_length(stop, 1L)
  expect_lt(start, stop)

  body <- paste(lines[start:(stop - 1L)], collapse = "\n")

  expect_false(grepl("(malloc|calloc|alloc_matd|alloc_tmatd)\\([^\\)]*(num_obs|num_obs_train_extern)[^\\)]*(num_obs|num_obs_train_extern)", body))
})

test_that("generalized-nn cvml LP stream avoids dense row storage", {
  lines <- read_shadow_proof_src("jksum.c")
  start <- grep("^int np_conditional_density_cvml_lp_stream\\(", lines)
  stop <- grep("^static int np_shadow_conditional_build_y_matrix\\(", lines)

  expect_length(start, 1L)
  expect_length(stop, 1L)
  expect_lt(start, stop)

  body <- paste(lines[start:(stop - 1L)], collapse = "\n")

  expect_false(grepl("(malloc|calloc|alloc_matd|alloc_tmatd)\\([^\\)]*(num_obs|num_obs_train_extern)[^\\)]*(num_obs|num_obs_train_extern)", body))
})

test_that("generalized-nn cvls LP stream avoids dense row storage", {
  lines <- read_shadow_proof_src("jksum.c")
  start <- grep("^int np_conditional_density_cvls_lp_stream\\(", lines)
  stop <- grep("^static int np_shadow_conditional_build_y_matrix\\(", lines)

  expect_length(start, 1L)
  expect_length(stop, 1L)
  expect_lt(start, stop)

  body <- paste(lines[start:(stop - 1L)], collapse = "\n")

  expect_false(grepl("(malloc|calloc|alloc_matd|alloc_tmatd)\\([^\\)]*(num_obs|num_obs_train_extern)[^\\)]*(num_obs|num_obs_train_extern)", body))
})

test_that("conditional density cvls production bypasses shadow block helpers", {
  lines <- read_shadow_proof_src("jksum.c")
  start <- grep("^int np_conditional_density_cvls_lp_stream\\(", lines)
  stop <- grep("^int np_conditional_distribution_cvls_lp_stream\\(", lines)

  expect_length(start, 1L)
  expect_gte(length(stop), 1L)
  expect_lt(start, stop[1L])

  body <- paste(lines[start:(stop[1L] - 1L)], collapse = "\n")

  expect_false(grepl("np_shadow_proof_conditional_x_weight_block_stream\\s*\\(", body))
  expect_false(grepl("np_shadow_conditional_y_block_stream_op\\s*\\(", body))
  expect_true(grepl("np_conditional_x_weight_block_stream_core\\s*\\(", body))
  expect_true(grepl("np_conditional_y_block_stream_op_core\\s*\\(", body))
})

test_that("fixed-bandwidth cvls LP stream avoids dense row storage", {
  lines <- read_shadow_proof_src("jksum.c")
  start_matches <- grep("^int np_conditional_density_cvls_lp_stream\\(", lines)
  stop <- grep("^static int np_shadow_conditional_build_y_matrix\\(", lines)

  expect_gte(length(start_matches), 1L)
  expect_length(stop, 1L)
  start <- tail(start_matches, 1L)
  expect_lt(start, stop)

  body <- paste(lines[start:(stop - 1L)], collapse = "\n")

  expect_false(grepl("(malloc|calloc|alloc_matd|alloc_tmatd)\\([^\\)]*(num_obs|num_obs_train_extern)[^\\)]*(num_obs|num_obs_train_extern)", body))
})

test_that("shadow LP objectives are sensitive to degree on oracle cells", {
  set.seed(123)
  n <- 35L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = x$x1^2 - x$x2 + rnorm(n, sd = 0.05))

  bw.d1 <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.24, 0.31, 0.42),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = rep.int(1L, ncol(x))
  )
  bw.d2 <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.24, 0.31, 0.42),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = rep.int(2L, ncol(x))
  )
  dens.d1 <- call_shadow_density(bw.d1, x, y, criterion = "cv.ml", compare_old = FALSE)
  dens.d2 <- call_shadow_density(bw.d2, x, y, criterion = "cv.ml", compare_old = FALSE)

  bw.c1 <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.24, 0.31, 0.42),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = rep.int(1L, ncol(x))
  )
  bw.c2 <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.24, 0.31, 0.42),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = rep.int(2L, ncol(x))
  )
  dist.d1 <- call_shadow_distribution(bw.c1, x, y, compare_old = FALSE)
  dist.d2 <- call_shadow_distribution(bw.c2, x, y, compare_old = FALSE)

  expect_gt(abs(dens.d2$new - dens.d1$new), 1e-6)
  expect_gt(abs(dist.d2$new - dist.d1$new), 1e-8)
})

test_that("shadow distribution lc matches legacy cvls for both cdfontrain modes", {
  set.seed(55)
  n <- 28L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = rnorm(n))
  bw <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.33, 0.44, 0.51),
    bandwidth.compute = FALSE,
    regtype = "lc"
  )

  res_false <- call_shadow_distribution(bw, x, y, cdfontrain = FALSE, compare_old = TRUE)
  res_true <- call_shadow_distribution(bw, x, y, cdfontrain = TRUE, compare_old = TRUE)

  expect_equal(res_false$new, res_false$old, tolerance = 1e-12)
  expect_equal(res_true$new, res_true$old, tolerance = 1e-12)
})

test_that("shadow distribution generalized-nn LP preserves ll canonicalization", {
  set.seed(263)
  n <- 36L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = sin(2 * pi * x$x1) + x$x2 + rnorm(n, sd = 0.1))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(4, 6, 5),
    bwtype = "generalized_nn",
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  bw.lp <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(4, 6, 5),
    bwtype = "generalized_nn",
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = degree
  )

  dist.ll <- call_shadow_distribution(bw.ll, x, y, compare_old = FALSE)
  dist.lp <- call_shadow_distribution(bw.lp, x, y, compare_old = FALSE)

  expect_true(is.finite(dist.ll$new))
  expect_true(is.finite(dist.lp$new))
  expect_equal(dist.ll$new, dist.lp$new, tolerance = 1e-10)
  expect_equal(dist.ll$prod, dist.ll$new, tolerance = 1e-10)
  expect_equal(dist.lp$prod, dist.lp$new, tolerance = 1e-10)
})

test_that("shadow distribution lp preserves ll canonicalization and tree parity", {
  set.seed(88)
  n <- 34L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = x$x1 - x$x2 + rnorm(n, sd = 0.2))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.28, 0.39, 0.52),
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  bw.lp <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.28, 0.39, 0.52),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = degree
  )

  ll_res <- call_shadow_distribution(bw.ll, x, y, compare_old = FALSE)
  lp_res <- call_shadow_distribution(bw.lp, x, y, compare_old = FALSE)
  lp_tree <- call_shadow_distribution(bw.lp, x, y, use_tree = TRUE, compare_old = FALSE)

  expect_equal(ll_res$new, lp_res$new, tolerance = 1e-12)
  expect_equal(lp_tree$new, lp_res$new, tolerance = 1e-12)
  expect_equal(ll_res$prod, ll_res$new, tolerance = 1e-12)
  expect_equal(lp_res$prod, lp_res$new, tolerance = 1e-12)
})

test_that("fixed-bandwidth cdist cvls LP stream avoids dense row storage", {
  lines <- read_shadow_proof_src("jksum.c")
  start_matches <- grep("^int np_conditional_distribution_cvls_lp_stream\\(", lines)
  stop <- grep("^static int np_shadow_conditional_build_y_matrix\\(", lines)

  expect_gte(length(start_matches), 1L)
  expect_length(stop, 1L)
  start <- tail(start_matches, 1L)
  expect_lt(start, stop)

  body <- paste(lines[start:(stop - 1L)], collapse = "\n")

  expect_false(grepl("(malloc|calloc|alloc_matd|alloc_tmatd)\\([^\\)]*(num_obs|num_obs_train_extern)[^\\)]*(num_obs|num_obs_train_extern)", body))
})

test_that("generalized-nn cdist cvls LP stream avoids dense row storage", {
  lines <- read_shadow_proof_src("jksum.c")
  start_matches <- grep("^int np_conditional_distribution_cvls_lp_stream\\(", lines)
  stop <- grep("^static int np_shadow_conditional_build_y_matrix\\(", lines)

  expect_gte(length(start_matches), 1L)
  expect_length(stop, 1L)
  start <- tail(start_matches, 1L)
  expect_lt(start, stop)

  body <- paste(lines[start:(stop - 1L)], collapse = "\n")

  expect_false(grepl("(malloc|calloc|alloc_matd|alloc_tmatd)\\([^\\)]*(num_obs|num_obs_train_extern)[^\\)]*(num_obs|num_obs_train_extern)", body))
})

test_that("LP all-large conditional fast helpers stay memory-lean", {
  lines <- read_shadow_proof_src("jksum.c")
  start <- grep("^static int np_conditional_lp_all_large_ctx_prepare\\(", lines)
  stop <- grep("^int np_conditional_density_cvml_lp_stream\\(", lines)

  expect_length(start, 1L)
  expect_length(stop, 1L)
  expect_lt(start, stop)

  body <- paste(lines[start:(stop - 1L)], collapse = "\n")

  expect_false(grepl("(malloc|calloc|alloc_matd|alloc_tmatd)\\([^\\)]*(num_train|num_obs_train_extern)[^\\)]*(num_train|num_obs_train_extern)", body))
  expect_false(grepl("\\bdiag\\s*\\(", body))
})

test_that("shadow density LP all-large fixed objectives stay exact and count fast hits", {
  set.seed(615)
  n <- 48L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rbeta(n, 1 + x$x, 2 - x$x))
  degree <- 1L

  bw.ll.ml <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.17, 1e6),
    bandwidth.compute = FALSE,
    regtype = "ll",
    bwmethod = "cv.ml"
  )
  bw.lp.ml <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.17, 1e6),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = degree,
    bwmethod = "cv.ml"
  )
  bw.ll.ls <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.17, 1e6),
    bandwidth.compute = FALSE,
    regtype = "ll",
    bwmethod = "cv.ls"
  )
  bw.lp.ls <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.17, 1e6),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = degree,
    bwmethod = "cv.ls"
  )

  sh.ll.ml <- call_shadow_density(bw.ll.ml, x, y, criterion = "cv.ml", compare_old = FALSE)
  sh.lp.ml <- call_shadow_density(bw.lp.ml, x, y, criterion = "cv.ml", compare_old = FALSE)
  sh.ll.ls <- call_shadow_density(bw.ll.ls, x, y, criterion = "cv.ls", compare_old = FALSE)
  sh.lp.ls <- call_shadow_density(bw.lp.ls, x, y, criterion = "cv.ls", compare_old = FALSE)

  ev.ll.ml <- np:::.npcdensbw_eval_only(x, y, bw.ll.ml)
  ev.lp.ml <- np:::.npcdensbw_eval_only(x, y, bw.lp.ml)
  ev.ll.ls <- np:::.npcdensbw_eval_only(x, y, bw.ll.ls)
  ev.lp.ls <- np:::.npcdensbw_eval_only(x, y, bw.lp.ls)

  expect_equal(-ev.ll.ml$objective, sh.ll.ml$new, tolerance = 1e-10)
  expect_equal(-ev.lp.ml$objective, sh.lp.ml$new, tolerance = 1e-10)
  expect_equal(-ev.ll.ls$objective, sh.ll.ls$new, tolerance = 1e-10)
  expect_equal(-ev.lp.ls$objective, sh.lp.ls$new, tolerance = 1e-10)
  expect_equal(ev.ll.ml$objective, ev.lp.ml$objective, tolerance = 1e-10)
  expect_equal(ev.ll.ls$objective, ev.lp.ls$objective, tolerance = 1e-10)
  expect_gt(ev.ll.ml$num.feval.fast, 0)
  expect_gt(ev.lp.ml$num.feval.fast, 0)
  expect_gt(ev.ll.ls$num.feval.fast, 0)
  expect_gt(ev.lp.ls$num.feval.fast, 0)
})

test_that("shadow distribution LP all-large fixed objective stays exact and counts fast hits", {
  set.seed(616)
  n <- 44L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rbeta(n, 1 + x$x, 2 - x$x))
  degree <- 1L

  bw.ll <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.17, 1e6),
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  bw.lp <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.17, 1e6),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = degree
  )

  sh.ll <- call_shadow_distribution(
    bw.ll,
    x,
    y,
    yeval = y,
    cdfontrain = TRUE,
    compare_old = FALSE
  )
  sh.lp <- call_shadow_distribution(
    bw.lp,
    x,
    y,
    yeval = y,
    cdfontrain = TRUE,
    compare_old = FALSE
  )

  ev.ll <- np:::.npcdistbw_eval_only(x, y, bws = bw.ll, do.full.integral = TRUE)
  ev.lp <- np:::.npcdistbw_eval_only(x, y, bws = bw.lp, do.full.integral = TRUE)

  expect_equal(ev.ll$objective, sh.ll$new, tolerance = 1e-10)
  expect_equal(ev.lp$objective, sh.lp$new, tolerance = 1e-10)
  expect_equal(ev.ll$objective, ev.lp$objective, tolerance = 1e-10)
  expect_gt(ev.ll$num.feval.fast, 0)
  expect_gt(ev.lp$num.feval.fast, 0)
})

test_that("kernelcv no longer references dense shadow proof helpers", {
  lines <- read_shadow_proof_src("kernelcv.c")

  expect_false(any(grepl("np_shadow_cv_con_density_ml\\s*\\(", lines)))
  expect_false(any(grepl("np_shadow_cv_con_density_ls\\s*\\(", lines)))
  expect_false(any(grepl("np_shadow_cv_con_distribution_ls\\s*\\(", lines)))
})

test_that("fixed-bwtype plot bootstrap covers supervised wild and unsupervised inid fixed geom families", {
  skip_if_not_installed("np")

  set.seed(603106)
  n <- 48
  xdat <- data.frame(x = rnorm(n))
  zdat <- data.frame(z = runif(n))
  xidat <- data.frame(x1 = xdat$x, x2 = runif(n))
  yreg <- sin(xdat$x) + rnorm(n, sd = 0.15)
  ydat <- data.frame(y = rnorm(n))
  yscoef <- (0.4 + xdat$x) * cos(2 * pi * zdat$z) + rnorm(n, sd = 0.05)
  yplreg <- sin(2 * pi * zdat$z) + 1.2 * xdat$x + rnorm(n, sd = 0.05)
  yindex <- sin(xidat$x1 + 0.5 * xidat$x2) + rnorm(n, sd = 0.05)

  reg.bws <- list(
    lc = npregbw(
      xdat = xdat,
      ydat = yreg,
      bws = 0.30,
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = "lc"
    ),
    ll = npregbw(
      xdat = xdat,
      ydat = yreg,
      bws = 0.30,
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = "ll"
    ),
    lp = npregbw(
      xdat = xdat,
      ydat = yreg,
      bws = 0.30,
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = "lp",
      degree = 2L,
      basis = "glp",
      bernstein.basis = FALSE
    )
  )
  u.dens.bw <- npudensbw(dat = xdat, bws = 0.30, bandwidth.compute = FALSE, bwtype = "fixed")
  u.dist.bw <- npudistbw(dat = xdat, bws = 0.30, bandwidth.compute = FALSE, bwtype = "fixed")
  c.dens.bw <- npcdensbw(xdat = xdat, ydat = ydat, bws = c(0.35, 0.35), bandwidth.compute = FALSE, bwtype = "fixed")
  c.dist.bw <- npcdistbw(xdat = xdat, ydat = ydat, bws = c(0.35, 0.35), bandwidth.compute = FALSE, bwtype = "fixed")
  sc.bws <- list(
    lc = npscoefbw(
      xdat = xdat,
      zdat = zdat,
      ydat = yscoef,
      bws = 0.30,
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = "lc"
    ),
    ll = npscoefbw(
      xdat = xdat,
      zdat = zdat,
      ydat = yscoef,
      bws = 0.30,
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = "ll"
    ),
    lp = npscoefbw(
      xdat = xdat,
      zdat = zdat,
      ydat = yscoef,
      bws = 0.30,
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = "lp",
      degree = 2L,
      basis = "glp",
      bernstein.basis = FALSE
    )
  )
  pl.bws <- list(
    lc = npplregbw(
      xdat = xdat,
      zdat = zdat,
      ydat = yplreg,
      bws = matrix(c(0.30, 0.30), nrow = 2L, ncol = 1L),
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = "lc"
    ),
    ll = npplregbw(
      xdat = xdat,
      zdat = zdat,
      ydat = yplreg,
      bws = matrix(c(0.30, 0.30), nrow = 2L, ncol = 1L),
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = "ll"
    ),
    lp = npplregbw(
      xdat = xdat,
      zdat = zdat,
      ydat = yplreg,
      bws = matrix(c(0.30, 0.30), nrow = 2L, ncol = 1L),
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = "lp",
      degree = 2L,
      basis = "glp",
      bernstein.basis = FALSE
    )
  )
  si.bws <- list(
    lc = npindexbw(
      xdat = xidat,
      ydat = yindex,
      bws = c(1, 1, 0.30),
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = "lc"
    ),
    ll = npindexbw(
      xdat = xidat,
      ydat = yindex,
      bws = c(1, 1, 0.30),
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = "ll"
    ),
    lp = npindexbw(
      xdat = xidat,
      ydat = yindex,
      bws = c(1, 1, 0.30),
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = "lp",
      degree = 2L,
      basis = "glp",
      bernstein.basis = FALSE
    )
  )

  run_plot <- function(bw, ..., boot.method) {
    suppressWarnings(plot(
      bw,
      output = "data",
      perspective = FALSE,
      errors = "bootstrap",
      bootstrap = boot.method,
      boot_control = np_boot_control(blocklen = 3L),
      B = 5L,
      band = "pointwise",
      neval = 11L,
      ...
    ))
  }

  for (boot.method in c("inid", "fixed", "geom")) {
    for (reg.bw in reg.bws) {
      expect_type(run_plot(reg.bw, xdat = xdat, ydat = yreg, boot.method = boot.method), "list")
    }
    expect_type(run_plot(u.dens.bw, xdat = xdat, boot.method = boot.method), "list")
    expect_type(run_plot(u.dist.bw, xdat = xdat, boot.method = boot.method), "list")
    expect_type(run_plot(c.dens.bw, xdat = xdat, ydat = ydat, view = "fixed", boot.method = boot.method), "list")
    expect_type(run_plot(c.dist.bw, xdat = xdat, ydat = ydat, view = "fixed", boot.method = boot.method), "list")
    for (sc.bw in sc.bws) {
      expect_type(run_plot(sc.bw, xdat = xdat, ydat = yscoef, zdat = zdat, boot.method = boot.method), "list")
    }
    for (pl.bw in pl.bws) {
      expect_type(run_plot(pl.bw, xdat = xdat, ydat = yplreg, zdat = zdat, boot.method = boot.method), "list")
    }
    for (si.bw in si.bws) {
      expect_type(run_plot(si.bw, xdat = xidat, ydat = yindex, boot.method = boot.method), "list")
    }
  }

  for (reg.bw in reg.bws) {
    expect_type(run_plot(reg.bw, xdat = xdat, ydat = yreg, boot.method = "wild"), "list")
  }
  for (sc.bw in sc.bws) {
    expect_type(run_plot(sc.bw, xdat = xdat, ydat = yscoef, zdat = zdat, boot.method = "wild"), "list")
  }
  for (pl.bw in pl.bws) {
    expect_type(run_plot(pl.bw, xdat = xdat, ydat = yplreg, zdat = zdat, boot.method = "wild"), "list")
  }
  for (si.bw in si.bws) {
    expect_type(run_plot(si.bw, xdat = xidat, ydat = yindex, boot.method = "wild"), "list")
  }

  expect_error(
    run_plot(u.dens.bw, xdat = xdat, boot.method = "wild"),
    "not supported"
  )
  expect_error(
    run_plot(u.dist.bw, xdat = xdat, boot.method = "wild"),
    "not supported"
  )
  expect_error(
    run_plot(c.dens.bw, xdat = xdat, ydat = ydat, view = "fixed", boot.method = "wild"),
    "not supported"
  )
  expect_error(
    run_plot(c.dist.bw, xdat = xdat, ydat = ydat, view = "fixed", boot.method = "wild"),
    "not supported"
  )
})

with_public_trace_counter <- function(pkg, fun, expr) {
  ns <- asNamespace(pkg)
  ctr <- new.env(parent = emptyenv())
  ctr$n <- 0L

  trace(
    what = fun,
    where = ns,
    tracer = bquote(.(ctr)$n <- .(ctr)$n + 1L),
    print = FALSE
  )
  on.exit(try(untrace(fun, where = ns), silent = TRUE), add = TRUE)

  force(expr)
  ctr$n
}

test_that("object-fed regression plots avoid public npreg re-entry", {
  set.seed(9210)
  n <- 48
  xdat <- data.frame(x = runif(n))
  ydat <- sin(2 * pi * xdat$x) + rnorm(n, sd = 0.05)

  bw <- npregbw(xdat = xdat, ydat = ydat, regtype = "lc", nmulti = 1)
  fit <- npreg(txdat = xdat, tydat = ydat, bws = bw)

  calls <- with_public_trace_counter("np", "npreg", {
    out.bw <- plot(bw, xdat = xdat, ydat = ydat, output = "data")
    out.fit <- plot(fit, output = "data")
    expect_type(out.bw, "list")
    expect_type(out.fit, "list")
  })

  expect_identical(calls, 0L)
})

test_that("object-fed unconditional density/distribution plots avoid public re-entry", {
  set.seed(9211)
  n <- 42
  dat <- data.frame(x = runif(n))

  dens.bw <- npudensbw(dat = dat, bws = 0.18, bandwidth.compute = FALSE, bwtype = "fixed")
  dist.bw <- npudistbw(dat = dat, bws = 0.18, bandwidth.compute = FALSE, bwtype = "fixed")
  dens.fit <- npudens(tdat = dat, bws = dens.bw)
  dist.fit <- npudist(tdat = dat, bws = dist.bw)

  dens.calls <- with_public_trace_counter("np", "npudens", {
    out.bw <- plot(dens.bw, xdat = dat, output = "data")
    out.fit <- plot(dens.fit, output = "data")
    expect_type(out.bw, "list")
    expect_type(out.fit, "list")
  })

  dist.calls <- with_public_trace_counter("np", "npudist", {
    out.bw <- plot(dist.bw, xdat = dat, output = "data")
    out.fit <- plot(dist.fit, output = "data")
    expect_type(out.bw, "list")
    expect_type(out.fit, "list")
  })

  expect_identical(dens.calls, 0L)
  expect_identical(dist.calls, 0L)
})

test_that("object-fed partially linear plots avoid public npplreg re-entry", {
  set.seed(9212)
  n <- 54
  xdat <- data.frame(x = runif(n))
  zdat <- data.frame(z = runif(n))
  ydat <- 1.5 * xdat$x + cos(2 * pi * zdat$z) + rnorm(n, sd = 0.05)

  bw <- npplregbw(xdat = xdat, ydat = ydat, zdat = zdat, regtype = "lc", nmulti = 1)
  fit <- npplreg(txdat = xdat, tydat = ydat, tzdat = zdat, bws = bw)

  calls <- with_public_trace_counter("np", "npplreg", {
    out.bw <- plot(bw, xdat = xdat, ydat = ydat, zdat = zdat, output = "data")
    out.fit <- plot(fit, output = "data")
    expect_type(out.bw, "list")
    expect_type(out.fit, "list")
  })

  expect_identical(calls, 0L)
})

test_that("object-fed conditional density/distribution plots avoid public re-entry", {
  set.seed(9213)
  n <- 46
  xdat <- data.frame(x = runif(n))
  ydat <- data.frame(y = rnorm(n))

  dens.bw <- npcdensbw(xdat = xdat, ydat = ydat, bws = c(0.22, 0.22), bandwidth.compute = FALSE)
  dist.bw <- npcdistbw(xdat = xdat, ydat = ydat, bws = c(0.22, 0.22), bandwidth.compute = FALSE)
  dens.fit <- npcdens(txdat = xdat, tydat = ydat, bws = dens.bw)
  dist.fit <- npcdist(txdat = xdat, tydat = ydat, bws = dist.bw)

  dens.calls <- with_public_trace_counter("np", "npcdens", {
    out.grid.bw <- plot(dens.bw, xdat = xdat, ydat = ydat, output = "data", view = "fixed")
    out.grid.fit <- plot(dens.fit, output = "data", view = "fixed")
    out.slice.bw <- plot(dens.bw, xdat = xdat, ydat = ydat, output = "data", perspective = FALSE, errors = "asymptotic")
    out.slice.fit <- plot(dens.fit, output = "data", perspective = FALSE, errors = "asymptotic")
    expect_type(out.grid.bw, "list")
    expect_type(out.grid.fit, "list")
    expect_type(out.slice.bw, "list")
    expect_type(out.slice.fit, "list")
  })

  dist.calls <- with_public_trace_counter("np", "npcdist", {
    out.grid.bw <- plot(dist.bw, xdat = xdat, ydat = ydat, output = "data", view = "fixed")
    out.grid.fit <- plot(dist.fit, output = "data", view = "fixed")
    out.slice.bw <- plot(dist.bw, xdat = xdat, ydat = ydat, output = "data", perspective = FALSE, errors = "asymptotic")
    out.slice.fit <- plot(dist.fit, output = "data", perspective = FALSE, errors = "asymptotic")
    expect_type(out.grid.bw, "list")
    expect_type(out.grid.fit, "list")
    expect_type(out.slice.bw, "list")
    expect_type(out.slice.fit, "list")
  })

  expect_identical(dens.calls, 0L)
  expect_identical(dist.calls, 0L)
})

test_that("object-fed quantile plots avoid public npqreg re-entry", {
  set.seed(9214)
  n <- 46
  xdat <- data.frame(x = runif(n))
  ydat <- data.frame(y = rnorm(n))

  bw <- npcdistbw(xdat = xdat, ydat = ydat, bws = c(0.22, 0.22), bandwidth.compute = FALSE)
  fit <- npqreg(txdat = xdat, tydat = ydat, bws = bw, tau = 0.4)

  calls <- with_public_trace_counter("np", "npqreg", {
    out.grid.bw <- plot(bw, xdat = xdat, ydat = ydat, output = "data", view = "fixed", quantreg = TRUE, tau = 0.4)
    out.grid.fit <- plot(fit, output = "data", view = "fixed")
    out.slice.bw <- plot(bw, xdat = xdat, ydat = ydat, output = "data", perspective = FALSE, quantreg = TRUE, tau = 0.4)
    out.slice.fit <- plot(fit, output = "data", perspective = FALSE)
    expect_type(out.grid.bw, "list")
    expect_type(out.grid.fit, "list")
    expect_type(out.slice.bw, "list")
    expect_type(out.slice.fit, "list")
  })

  expect_identical(calls, 0L)
})

test_that("smooth coefficient coef/asymptotic plots avoid public npscoef re-entry", {
  set.seed(9215)
  n <- 52
  xdat <- data.frame(x = runif(n))
  zdat <- data.frame(z = runif(n))
  ydat <- sin(2 * pi * zdat$z) + xdat$x * (1 + zdat$z) + rnorm(n, sd = 0.05)

  bw <- npscoefbw(xdat = xdat, ydat = ydat, zdat = zdat, regtype = "ll", nmulti = 1)
  fit <- npscoef(bws = bw, txdat = xdat, tydat = ydat, tzdat = zdat, errors = TRUE, betas = TRUE)

  calls <- with_public_trace_counter("np", "npscoef", {
    out.bw.coef <- suppressWarnings(
      plot(
        bw,
        xdat = xdat,
        ydat = ydat,
        zdat = zdat,
        coef = TRUE,
        perspective = FALSE,
        output = "data",
        errors = "none"
      )
    )
    out.fit.coef <- suppressWarnings(
      plot(
        fit,
        xdat = xdat,
        ydat = ydat,
        zdat = zdat,
        coef = TRUE,
        perspective = FALSE,
        output = "data",
        errors = "none"
      )
    )
    out.bw.asym <- suppressWarnings(
      plot(
        bw,
        xdat = xdat,
        ydat = ydat,
        zdat = zdat,
        coef = FALSE,
        perspective = FALSE,
        output = "data",
        errors = "asymptotic"
      )
    )
    out.fit.asym <- suppressWarnings(
      plot(
        fit,
        xdat = xdat,
        ydat = ydat,
        zdat = zdat,
        coef = FALSE,
        perspective = FALSE,
        output = "data",
        errors = "asymptotic"
      )
    )
    expect_type(out.bw.coef, "list")
    expect_type(out.fit.coef, "list")
    expect_type(out.bw.asym, "list")
    expect_type(out.fit.asym, "list")
  })

  expect_identical(calls, 0L)
})

with_npindex_nomad_payload_bindings <- function(bindings, code) {
  code <- substitute(code)
  ns <- asNamespace("np")
  old <- lapply(names(bindings), function(name) get(name, envir = ns, inherits = FALSE))
  names(old) <- names(bindings)

  for (name in names(bindings)) {
    was_locked <- bindingIsLocked(name, ns)
    if (was_locked)
      unlockBinding(name, ns)
    assign(name, bindings[[name]], envir = ns)
    if (was_locked)
      lockBinding(name, ns)
  }

  on.exit({
    for (name in names(old)) {
      was_locked <- bindingIsLocked(name, ns)
      if (was_locked)
        unlockBinding(name, ns)
      assign(name, old[[name]], envir = ns)
      if (was_locked)
        lockBinding(name, ns)
    }
  }, add = TRUE)

  eval(code, envir = parent.frame())
}

test_that("npindexbw retains Powell evaluation counts when Powell does not improve", {
  ns <- asNamespace("np")

  result <- with_npindex_nomad_payload_bindings(
    list(
      .npindexbw_build_sibandwidth = function(xdat, ydat, bws, template, bandwidth.compute, reg.args) {
        list(
          beta = c(1, 0.5),
          bw = 0.3,
          type = "fixed",
          method = "ichimura",
          ckertype = "gaussian",
          ckerorder = 2L,
          ckerbound = "none",
          ckerlb = NA_real_,
          ckerub = NA_real_,
          ynames = "y"
        )
      },
      .npindexbw_eval_objective = function(param, xmat, ydat, bws, spec) {
        list(objective = 10, num.feval.fast = 3L)
      },
      .np_nomad_baseline_note = function(degree) invisible(NULL),
      .np_nomad_search = function(engine,
                                  baseline_record,
                                  start_degree,
                                  x0,
                                  bbin,
                                  lb,
                                  ub,
                                  eval_fun,
                                  build_payload,
                                  direction,
                                  objective_name,
                                  nmulti,
                                  nomad.inner.nmulti,
                                  random.seed,
                                  remin = FALSE,
                                  degree_spec,
                                  progress_label) {
        eval_fun(x0)
        build_payload(
          x0,
          best_record = list(degree = start_degree, objective = 10, num.feval = 1L),
          solution = list(bbe = 1L),
          interrupted = FALSE
        )
      },
      .np_nomad_with_powell_progress = function(degree, expr, best_record = NULL) expr,
      .npindexbw_run_fixed_degree = function(xdat, ydat, bws, template, reg.args, opt.args) {
        list(
          method = "ichimura",
          fval = 11,
          num.feval = 7L,
          num.feval.fast = 5L
        )
      },
      npindexbw.sibandwidth = function(xdat, ydat, bws, bandwidth.compute, ...) {
        bws
      }
    ),
    get(".npindexbw_nomad_search", envir = ns, inherits = FALSE)(
      xdat = data.frame(x1 = c(0.1, 0.2), x2 = c(0.3, 0.4)),
      ydat = c(1, 2),
      bws = c(1, 0.5, 0.3),
      template = list(method = "ichimura"),
      reg.args = list(
        method = "ichimura",
        regtype = "lp",
        basis.engine = "tensor"
      ),
      opt.args = list(nmulti = 1L),
      degree.search = list(
        verify = FALSE,
        start.degree = 1L,
        bernstein.basis = TRUE,
        lower = 0L,
        upper = 2L,
        candidates = list(0:2),
        engine = "nomad+powell",
        basis = "glp",
        nobs = 2L,
        start.user = FALSE
      ),
      nomad.inner.nmulti = 0L
    )
  )

  expect_equal(result$payload$fval, 10)
  expect_equal(result$payload$num.feval, 8)
  expect_equal(result$payload$num.feval.fast, 8)
  expect_equal(result$objective, 10)
})

test_that("npindexbw fixed NOMAD route normalizes internal h starts but keeps public restart starts raw", {
  ns <- asNamespace("np")
  xdat <- data.frame(x1 = c(-0.3, 0.1, 0.4, 0.8), x2 = c(0.5, -0.1, 0.2, -0.4))
  ydat <- c(0.2, -0.1, 0.8, 1.1)
  baseline.bws <- list(
    beta = c(1, 0),
    bw = 0,
    type = "fixed",
    method = "ichimura",
    ckertype = "gaussian",
    ckerorder = 2L,
    ckerbound = "none",
    ckerlb = NA_real_,
    ckerub = NA_real_,
    ynames = "y"
  )
  degree.search <- list(
    verify = FALSE,
    start.degree = 1L,
    bernstein.basis = TRUE,
    lower = 0L,
    upper = 2L,
    candidates = list(0:2),
    engine = "nomad+powell",
    basis = "glp",
    nobs = nrow(xdat),
    start.user = FALSE
  )
  setup <- get(".npindexbw_nomad_fixed_start_setup", envir = ns, inherits = FALSE)(
    xmat = as.matrix(xdat),
    ydat = ydat,
    baseline.bws = baseline.bws,
    degree.search = degree.search,
    nmulti = 1L,
    random.seed = 42L
  )
  captured <- NULL

  result <- with_npindex_nomad_payload_bindings(
    list(
      .npindexbw_build_sibandwidth = function(xdat, ydat, bws, template, bandwidth.compute, reg.args) {
        baseline.bws
      },
      .np_nomad_baseline_note = function(degree) invisible(NULL),
      .np_nomad_search = function(engine,
                                  baseline_record,
                                  start_degree,
                                  x0,
                                  bbin,
                                  lb,
                                  ub,
                                  eval_fun,
                                  build_payload,
                                  direction,
                                  objective_name,
                                  nmulti,
                                  nomad.inner.nmulti,
                                  random.seed,
                                  degree_spec,
                                  ...) {
        captured <<- list(
          x0 = as.numeric(x0),
          lb = as.numeric(lb),
          ub = as.numeric(ub)
        )
        list(
          method = engine,
          direction = direction,
          verify = FALSE,
          completed = TRUE,
          certified = FALSE,
          interrupted = FALSE,
          baseline = list(degree = start_degree, objective = 10),
          best = list(degree = start_degree, objective = 10, num.feval = 1L),
          best_payload = list(fval = 10, num.feval = 1L, num.feval.fast = 0L, method = "ichimura"),
          best_point = as.numeric(x0),
          n.unique = 1L,
          n.visits = 1L,
          n.cached = 0L,
          nomad.time = 1,
          powell.time = 0,
          optim.time = 1,
          grid.size = NA_integer_,
          best.restart = 1L,
          restart.starts = list(as.numeric(x0)),
          restart.degree.starts = list(as.integer(x0[length(x0)])),
          restart.bandwidth.starts = list(as.numeric(x0[seq_len(length(x0) - 1L)])),
          restart.start.info = list(user_supplied_start = FALSE),
          restart.results = list(list(
            restart = 1L,
            start = as.numeric(x0),
            degree.start = start_degree,
            objective = 10,
            bbe = 1L,
            iterations = 0L,
            solution = as.numeric(x0)
          )),
          trace = data.frame()
        )
      }
    ),
    get(".npindexbw_nomad_search", envir = ns, inherits = FALSE)(
      xdat = xdat,
      ydat = ydat,
      bws = c(0, 0, 0),
      template = list(method = "ichimura"),
      reg.args = list(
        method = "ichimura",
        regtype = "lp",
        basis.engine = "tensor"
      ),
      opt.args = list(nmulti = 1L, random.seed = 42L),
      degree.search = degree.search,
      nomad.inner.nmulti = 0L
    )
  )

  expect_equal(captured$x0, as.numeric(setup$start_matrix.point[1L, ]), tolerance = 1e-12)
  expect_equal(result$restart.starts[[1]], as.numeric(setup$start_matrix.raw[1L, ]), tolerance = 1e-12)
  expect_equal(result$restart.bandwidth.starts[[1]], as.numeric(setup$start_matrix.raw[1L, seq_len(ncol(setup$start_matrix.raw) - 1L)]), tolerance = 1e-12)
  expect_equal(result$restart.results[[1]]$start, as.numeric(setup$start_matrix.raw[1L, ]), tolerance = 1e-12)
  expect_equal(result$restart.results[[1]]$solution, as.numeric(setup$start_matrix.raw[1L, ]), tolerance = 1e-12)
})

test_that("npindexbw fixed NOMAD route preserves explicit user fixed starts", {
  ns <- asNamespace("np")
  xdat <- data.frame(x1 = c(-0.3, 0.1, 0.4, 0.8), x2 = c(0.5, -0.1, 0.2, -0.4))
  ydat <- c(0.2, -0.1, 0.8, 1.1)
  baseline.bws <- list(
    beta = c(1, 1.75),
    bw = 0.25,
    type = "fixed",
    method = "ichimura",
    ckertype = "gaussian",
    ckerorder = 2L,
    ckerbound = "none",
    ckerlb = NA_real_,
    ckerub = NA_real_,
    ynames = "y"
  )
  degree.search <- list(
    verify = FALSE,
    start.degree = 1L,
    bernstein.basis = TRUE,
    lower = 0L,
    upper = 2L,
    candidates = list(0:2),
    engine = "nomad+powell",
    basis = "glp",
    nobs = nrow(xdat),
    start.user = FALSE
  )
  setup <- get(".npindexbw_nomad_fixed_start_setup", envir = ns, inherits = FALSE)(
    xmat = as.matrix(xdat),
    ydat = ydat,
    baseline.bws = baseline.bws,
    degree.search = degree.search,
    nmulti = 1L,
    random.seed = 42L
  )
  captured <- NULL

  result <- with_npindex_nomad_payload_bindings(
    list(
      .npindexbw_build_sibandwidth = function(xdat, ydat, bws, template, bandwidth.compute, reg.args) {
        baseline.bws
      },
      .np_nomad_baseline_note = function(degree) invisible(NULL),
      .np_nomad_search = function(engine,
                                  baseline_record,
                                  start_degree,
                                  x0,
                                  bbin,
                                  lb,
                                  ub,
                                  eval_fun,
                                  build_payload,
                                  direction,
                                  objective_name,
                                  nmulti,
                                  nomad.inner.nmulti,
                                  random.seed,
                                  degree_spec,
                                  ...) {
        captured <<- list(
          x0 = as.numeric(x0),
          lb = as.numeric(lb),
          ub = as.numeric(ub)
        )
        list(
          method = engine,
          direction = direction,
          verify = FALSE,
          completed = TRUE,
          certified = FALSE,
          interrupted = FALSE,
          baseline = list(degree = start_degree, objective = 10),
          best = list(degree = start_degree, objective = 10, num.feval = 1L),
          best_payload = list(fval = 10, num.feval = 1L, num.feval.fast = 0L, method = "ichimura"),
          best_point = as.numeric(x0),
          n.unique = 1L,
          n.visits = 1L,
          n.cached = 0L,
          nomad.time = 1,
          powell.time = 0,
          optim.time = 1,
          grid.size = NA_integer_,
          best.restart = 1L,
          restart.starts = list(as.numeric(x0)),
          restart.degree.starts = list(as.integer(x0[length(x0)])),
          restart.bandwidth.starts = list(as.numeric(x0[seq_len(length(x0) - 1L)])),
          restart.start.info = list(user_supplied_start = FALSE),
          restart.results = list(list(
            restart = 1L,
            start = as.numeric(x0),
            degree.start = start_degree,
            objective = 10,
            bbe = 1L,
            iterations = 0L,
            solution = as.numeric(x0)
          )),
          trace = data.frame()
        )
      }
    ),
    get(".npindexbw_nomad_search", envir = ns, inherits = FALSE)(
      xdat = xdat,
      ydat = ydat,
      bws = c(0, 0, 0),
      template = list(method = "ichimura"),
      reg.args = list(
        method = "ichimura",
        regtype = "lp",
        basis.engine = "tensor"
      ),
      opt.args = list(nmulti = 1L, random.seed = 42L),
      degree.search = degree.search,
      nomad.inner.nmulti = 0L
    )
  )

  expect_equal(setup$start_matrix.raw[1L, 1L], 1.75, tolerance = 1e-12)
  expect_equal(setup$start_matrix.raw[1L, 2L], 0.25, tolerance = 1e-12)
  expect_equal(captured$x0, as.numeric(setup$start_matrix.point[1L, ]), tolerance = 1e-12)
  expect_equal(result$restart.starts[[1]], as.numeric(setup$start_matrix.raw[1L, ]), tolerance = 1e-12)
  expect_equal(result$restart.results[[1]]$start, as.numeric(setup$start_matrix.raw[1L, ]), tolerance = 1e-12)
})

condensdist_fit_progress_time_counter <- function(start = 0, by = 1.7) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

condensdist_fit_progress_lines <- function(shadow) {
  vapply(shadow$trace, `[[`, character(1L), "line")
}

expect_condensdist_clean_powell_surface <- function(lines, pkg_pattern = "np", info = NULL) {
  powell.lines <- grep(sprintf("^\\[%s\\] Refining bandwidth \\(", pkg_pattern), lines, value = TRUE)
  detail.info <- if (!is.null(info)) info else paste(lines, collapse = "\n")

  expect_true(length(powell.lines) > 0L, info = detail.info)
  expect_true(
    any(grepl(
      sprintf("^\\[%s\\] Refining bandwidth \\(elapsed [0-9]+\\.[0-9]s, degree \\([0-9]+\\)(, iter [0-9]+)?\\)$", pkg_pattern),
      powell.lines
    )),
    info = paste(c(detail.info, powell.lines), collapse = "\n")
  )
  expect_false(any(grepl("best \\(", powell.lines)), info = paste(c(detail.info, powell.lines), collapse = "\n"))
  expect_false(
    any(grepl(sprintf("^\\[%s\\] Bandwidth selection \\(Refining NOMAD solution", pkg_pattern), lines)),
    info = detail.info
  )
}

make_condensdist_fit_progress_fixture <- function() {
  set.seed(20260404)
  n <- 18L
  dat <- data.frame(
    x = seq(-0.9, 0.9, length.out = n),
    y = 0.4 * seq(-0.9, 0.9, length.out = n) + sin(seq(-0.9, 0.9, length.out = n))
  )

  list(
    dat = dat,
    tx = dat["x"],
    ty = dat["y"],
    n = n,
    cdens.bw = npcdensbw(
      xdat = dat["x"],
      ydat = dat["y"],
      bws = c(0.24, 0.24),
      bandwidth.compute = FALSE
    ),
    cdist.bw = npcdistbw(
      xdat = dat["x"],
      ydat = dat["y"],
      bws = c(0.24, 0.24),
      bandwidth.compute = FALSE
    )
  )
}

test_that("npcdens direct bws fit emits single-line fit progress", {
  fixture <- make_condensdist_fit_progress_fixture()
  total <- 2L * fixture$n

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npcdens(
      bws = fixture$cdens.bw,
      txdat = fixture$tx,
      tydat = fixture$ty
    ),
    force_renderer = "single_line",
    now = condensdist_fit_progress_time_counter()
  )

  lines <- condensdist_fit_progress_lines(actual)

  expect_s3_class(actual$value, "condensity")
  expect_true(any(grepl(
    sprintf("^\\[np\\] Fitting conditional density 1/%d \\([0-9]+\\.[0-9]%%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", total),
    lines
  )))
  expect_true(any(grepl(
    sprintf("^\\[np\\] Fitting conditional density %d/%d \\(100\\.0%%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$", total, total),
    lines
  )))
})

test_that("npcdens bw to fit route hands off immediately into single-line fit progress", {
  fixture <- make_condensdist_fit_progress_fixture()
  total <- 2L * fixture$n

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npcdens(
      txdat = fixture$tx,
      tydat = fixture$ty,
      nmulti = 1L
    ),
    force_renderer = "single_line",
    now = condensdist_fit_progress_time_counter()
  )

  lines <- condensdist_fit_progress_lines(actual)
  bandwidth.pos <- grep("^\\[np\\] Bandwidth selection \\(", lines)
  fit.start.pos <- grep(
    sprintf("^\\[np\\] Fitting conditional density 0/%d \\(0\\.0%%, elapsed 0\\.0s, eta 0\\.0s\\): starting$", total),
    lines
  )
  fit.finish.pos <- grep(
    sprintf("^\\[np\\] Fitting conditional density %d/%d \\(100\\.0%%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$", total, total),
    lines
  )

  expect_s3_class(actual$value, "condensity")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(fit.start.pos) == 1L)
  expect_true(length(fit.finish.pos) >= 1L)
  expect_lt(max(bandwidth.pos), fit.start.pos[[1L]])
  expect_lt(fit.start.pos[[1L]], fit.finish.pos[[1L]])
})

test_that("npcdens nomad to powell to fit route preserves single-line fit handoff", {
  skip_if_not_installed("crs")

  fixture <- make_condensdist_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npcdens(
      y ~ x,
      data = fixture$dat,
      nomad = TRUE,
      degree.max = 1L,
      nmulti = 1L
    ),
    force_renderer = "single_line",
    now = condensdist_fit_progress_time_counter()
  )

  lines <- condensdist_fit_progress_lines(actual)
  fit.start.pos <- grep(
    sprintf("^\\[np\\] Fitting conditional density 0/%d \\(0\\.0%%, elapsed 0\\.0s, eta 0\\.0s\\): starting$", fixture$n),
    lines
  )
  fit.finish.pos <- grep(
    sprintf("^\\[np\\] Fitting conditional density %d/%d \\(100\\.0%%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$", fixture$n, fixture$n),
    lines
  )
  powell.pos <- grep("^\\[np\\] Refining bandwidth \\(", lines)
  bandwidth.pos <- grep("^\\[np\\] Selecting degree and bandwidth", lines)

  expect_s3_class(actual$value, "condensity")
  expect_condensdist_clean_powell_surface(lines, pkg_pattern = "np")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(powell.pos) > 0L)
  expect_true(length(fit.start.pos) == 1L)
  expect_true(length(fit.finish.pos) >= 1L)
  expect_lt(max(powell.pos), fit.start.pos[[1L]])
  expect_lt(max(bandwidth.pos), fit.start.pos[[1L]])
  expect_lt(fit.start.pos[[1L]], fit.finish.pos[[1L]])
})

test_that("npcdist direct bws fit emits single-line fit progress", {
  fixture <- make_condensdist_fit_progress_fixture()
  total <- 2L * fixture$n

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npcdist(
      bws = fixture$cdist.bw,
      txdat = fixture$tx,
      tydat = fixture$ty
    ),
    force_renderer = "single_line",
    now = condensdist_fit_progress_time_counter()
  )

  lines <- condensdist_fit_progress_lines(actual)

  expect_s3_class(actual$value, "condistribution")
  expect_true(any(grepl(
    sprintf("^\\[np\\] Fitting conditional distribution 1/%d \\([0-9]+\\.[0-9]%%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", total),
    lines
  )))
  expect_true(any(grepl(
    sprintf("^\\[np\\] Fitting conditional distribution %d/%d \\(100\\.0%%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$", total, total),
    lines
  )))
})

test_that("npcdist bw to fit route hands off immediately into single-line fit progress", {
  fixture <- make_condensdist_fit_progress_fixture()
  total <- 2L * fixture$n

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npcdist(
      txdat = fixture$tx,
      tydat = fixture$ty,
      nmulti = 1L
    ),
    force_renderer = "single_line",
    now = condensdist_fit_progress_time_counter()
  )

  lines <- condensdist_fit_progress_lines(actual)
  bandwidth.pos <- grep("^\\[np\\] Bandwidth selection \\(", lines)
  fit.start.pos <- grep(
    sprintf("^\\[np\\] Fitting conditional distribution 0/%d \\(0\\.0%%, elapsed 0\\.0s, eta 0\\.0s\\): starting$", total),
    lines
  )
  fit.finish.pos <- grep(
    sprintf("^\\[np\\] Fitting conditional distribution %d/%d \\(100\\.0%%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$", total, total),
    lines
  )

  expect_s3_class(actual$value, "condistribution")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(fit.start.pos) == 1L)
  expect_true(length(fit.finish.pos) >= 1L)
  expect_lt(max(bandwidth.pos), fit.start.pos[[1L]])
  expect_lt(fit.start.pos[[1L]], fit.finish.pos[[1L]])
})

test_that("npcdist nomad to powell to fit route preserves single-line fit handoff", {
  skip_if_not_installed("crs")

  fixture <- make_condensdist_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npcdist(
      y ~ x,
      data = fixture$dat,
      nomad = TRUE,
      degree.max = 1L,
      nmulti = 1L
    ),
    force_renderer = "single_line",
    now = condensdist_fit_progress_time_counter()
  )

  lines <- condensdist_fit_progress_lines(actual)
  fit.start.pos <- grep(
    sprintf("^\\[np\\] Fitting conditional distribution 0/%d \\(0\\.0%%, elapsed 0\\.0s, eta 0\\.0s\\): starting$", fixture$n),
    lines
  )
  fit.finish.pos <- grep(
    sprintf("^\\[np\\] Fitting conditional distribution %d/%d \\(100\\.0%%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$", fixture$n, fixture$n),
    lines
  )
  powell.pos <- grep("^\\[np\\] Refining bandwidth \\(", lines)
  bandwidth.pos <- grep("^\\[np\\] Selecting degree and bandwidth", lines)

  expect_s3_class(actual$value, "condistribution")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(powell.pos) > 0L)
  expect_true(length(fit.start.pos) == 1L)
  expect_true(length(fit.finish.pos) >= 1L)
  expect_lt(max(powell.pos), fit.start.pos[[1L]])
  expect_lt(max(bandwidth.pos), fit.start.pos[[1L]])
  expect_lt(fit.start.pos[[1L]], fit.finish.pos[[1L]])
})

test_that("nearest-neighbor plot helpers run for regression and density/distribution families", {
  skip_if_not_installed("np")

  set.seed(329)
  n <- 45
  x <- data.frame(x = runif(n))
  y <- sin(2 * pi * x$x) + rnorm(n, sd = 0.08)
  yframe <- data.frame(y = y)

  run_plot <- function(bw, ...) {
    suppressWarnings(plot(
      bw,
      output = "data",
      perspective = FALSE,
      errors = "bootstrap",
      bootstrap = "inid",
      B = 5,
      ...
    ))
  }

  for (bt in c("generalized_nn", "adaptive_nn")) {
    rbw <- npregbw(xdat = x, ydat = y, regtype = "ll", nmulti = 1, bwtype = bt)
    expect_type(run_plot(rbw, xdat = x, ydat = y), "list")
    expect_type(run_plot(rbw, xdat = x, ydat = y, gradients = TRUE), "list")

    ubw <- npudensbw(dat = x, nmulti = 1, bwtype = bt)
    expect_type(run_plot(ubw), "list")

    dbw <- npudistbw(dat = x, nmulti = 1, bwtype = bt)
    expect_type(run_plot(dbw), "list")

    cbw <- npcdensbw(xdat = x, ydat = yframe, nmulti = 1, bwtype = bt)
    expect_type(run_plot(cbw, xdat = x, ydat = yframe, view = "fixed"), "list")

    cdbw <- npcdistbw(xdat = x, ydat = yframe, nmulti = 1, bwtype = bt)
    expect_type(run_plot(cdbw, xdat = x, ydat = yframe, view = "fixed"), "list")
  }
})

test_that("nearest-neighbor frozen bootstrap plots run across regression and unsupervised families", {
  skip_if_not_installed("np")

  set.seed(32908)
  n <- 45
  x <- data.frame(x = runif(n))
  y <- sin(2 * pi * x$x) + rnorm(n, sd = 0.08)
  yframe <- data.frame(y = y)

  run_plot <- function(bw, ...) {
    suppressWarnings(plot(
      bw,
      output = "data",
      perspective = FALSE,
      errors = "bootstrap",
      boot_control = np_boot_control(nonfixed = "frozen"),
      B = 5,
      ...
    ))
  }

  for (bt in c("generalized_nn", "adaptive_nn")) {
    rbw <- npregbw(xdat = x, ydat = y, regtype = "ll", nmulti = 1, bwtype = bt)

    for (boot.method in c("inid", "fixed", "geom")) {
      expect_type(
        run_plot(rbw, xdat = x, ydat = y, bootstrap = boot.method),
        "list"
      )

      ubw <- npudensbw(dat = x, nmulti = 1, bwtype = bt)
      expect_type(run_plot(ubw, bootstrap = boot.method), "list")

      dbw <- npudistbw(dat = x, nmulti = 1, bwtype = bt)
      expect_type(run_plot(dbw, bootstrap = boot.method), "list")

      cbw <- npcdensbw(xdat = x, ydat = yframe, nmulti = 1, bwtype = bt)
      expect_type(
        run_plot(cbw, xdat = x, ydat = yframe, view = "fixed", bootstrap = boot.method),
        "list"
      )

      cdbw <- npcdistbw(xdat = x, ydat = yframe, nmulti = 1, bwtype = bt)
      expect_type(
        run_plot(cdbw, xdat = x, ydat = yframe, view = "fixed", bootstrap = boot.method),
        "list"
      )
    }
  }
})

test_that("lp entry points require explicit degree", {
  set.seed(1)
  n <- 24L
  x <- runif(n)
  z <- rnorm(n)
  y <- x + z + rnorm(n, sd = 0.1)

  calls <- list(
    npregbw = quote(npregbw(xdat = data.frame(x = x), ydat = y, regtype = "lp")),
    npcdensbw = quote(npcdensbw(xdat = data.frame(x = x), ydat = y, regtype = "lp")),
    npcdistbw = quote(npcdistbw(xdat = data.frame(x = x), ydat = y, regtype = "lp")),
    npindexbw = quote(npindexbw(xdat = data.frame(x = x), ydat = y, regtype = "lp")),
    npplregbw = quote(npplregbw(xdat = data.frame(x = x), ydat = y, zdat = data.frame(z = z), regtype = "lp")),
    npscoefbw = quote(npscoefbw(xdat = data.frame(x = x), ydat = y, zdat = data.frame(z = z), regtype = "lp")),
    npreg = quote(npreg(y ~ x, regtype = "lp")),
    npcdens = quote(npcdens(y ~ x, regtype = "lp")),
    npcdist = quote(npcdist(y ~ x, regtype = "lp")),
    npindex = quote(npindex(y ~ x, regtype = "lp")),
    npplreg = quote(npplreg(y ~ x | z, regtype = "lp")),
    npscoef = quote(npscoef(y ~ x | z, regtype = "lp"))
  )

  for (nm in names(calls)) {
    expect_error(
      suppressWarnings(eval(calls[[nm]])),
      regexp = "degree must be supplied explicitly when regtype='lp'",
      info = nm
    )
  }
})

test_that("lp degree length validation remains in force", {
  set.seed(1)
  n <- 20L
  x1 <- runif(n)
  x2 <- runif(n)
  y <- x1 + x2 + rnorm(n, sd = 0.1)

  expect_error(
    npregbw(xdat = data.frame(x1 = x1, x2 = x2), ydat = y, regtype = "lp", degree = 1L),
    regexp = "degree must have one entry per continuous predictor",
    fixed = FALSE
  )

  expect_error(
    npreg(y ~ x1 + x2, regtype = "lp", degree = 1L),
    regexp = "degree must have one entry per continuous predictor",
    fixed = FALSE
  )
})

test_that("automatic lp degree search still works without explicit degree", {
  set.seed(7)
  n <- 18L
  x1 <- runif(n)
  x2 <- runif(n)
  z <- rnorm(n)
  y <- x1 + x2 + z + rnorm(n, sd = 0.1)

  common.args <- list(
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "cell",
    degree.min = 1L,
    degree.max = 1L,
    bwtype = "fixed"
  )

  fits <- list(
    npregbw = do.call(npregbw, c(list(
      xdat = data.frame(x = x1),
      ydat = y,
      nmulti = 1L
    ), common.args)),
    npcdensbw = do.call(npcdensbw, c(list(
      xdat = data.frame(x = x1),
      ydat = y
    ), common.args)),
    npcdistbw = do.call(npcdistbw, c(list(
      xdat = data.frame(x = x1),
      ydat = y
    ), common.args)),
    npindexbw = do.call(npindexbw, c(list(
      xdat = data.frame(x1 = x1, x2 = x2),
      ydat = y,
      nmulti = 1L
    ), common.args)),
    npplregbw = do.call(npplregbw, c(list(
      xdat = data.frame(x = x1),
      ydat = y,
      zdat = data.frame(z = z),
      nmulti = 1L
    ), common.args)),
    npscoefbw = do.call(npscoefbw, c(list(
      xdat = data.frame(x = x1),
      ydat = y,
      zdat = data.frame(z = z),
      nmulti = 1L
    ), common.args))
  )

  expect_equal(length(as.integer(fits$npregbw$degree)), 1L)
  expect_equal(length(as.integer(fits$npcdensbw$degree)), 1L)
  expect_equal(length(as.integer(fits$npcdistbw$degree)), 1L)
  expect_equal(length(as.integer(fits$npindexbw$degree)), 1L)
  expect_equal(length(as.integer(fits$npplregbw$degree)), 1L)
  expect_equal(length(as.integer(fits$npscoefbw$degree)), 1L)

  for (fit in fits)
    expect_true(all(as.integer(fit$degree) >= 0L))
})

library(np)

phase1_npcdist_cvls_gnn_fixture <- function() {
  set.seed(20260309)
  n <- 24L
  x <- data.frame(
    x1 = runif(n),
    x2 = runif(n)
  )
  y <- data.frame(
    y1 = x$x1^2 - 0.35 * x$x2 + 0.2 * sin(2 * pi * x$x1) + rnorm(n, sd = 0.08)
  )
  list(x = x, y = y)
}

phase1_npcdist_cvls_gnn_cases <- local({
  cache <- NULL

  function() {
    if (!is.null(cache))
      return(cache)

    dat <- phase1_npcdist_cvls_gnn_fixture()
    degree1 <- rep.int(1L, ncol(dat$x))
    degree2 <- rep.int(2L, ncol(dat$x))

    cache <<- list(
      dat = dat,
      degree1 = degree1,
      degree2 = degree2,
      bw.lc = npcdistbw(
        xdat = dat$x,
        ydat = dat$y,
        regtype = "lc",
        bwtype = "generalized_nn",
        bwmethod = "cv.ls",
        nmulti = 1L,
        itmax = 1L
      ),
      bw.ll = npcdistbw(
        xdat = dat$x,
        ydat = dat$y,
        regtype = "ll",
        bwtype = "generalized_nn",
        bwmethod = "cv.ls",
        nmulti = 1L,
        itmax = 1L
      ),
      bw.lp = npcdistbw(
        xdat = dat$x,
        ydat = dat$y,
        regtype = "lp",
        basis = "glp",
        degree = degree1,
        bwtype = "generalized_nn",
        bwmethod = "cv.ls",
        nmulti = 1L,
        itmax = 1L
      ),
      bw.d2 = npcdistbw(
        xdat = dat$x,
        ydat = dat$y,
        regtype = "lp",
        basis = "glp",
        degree = degree2,
        bwtype = "generalized_nn",
        bwmethod = "cv.ls",
        nmulti = 1L,
        itmax = 1L
      )
    )

    cache
  }
})

test_that("phase1 npcdistbw cv.ls generalized-nn lc matches the frozen public baseline", {
  bw.lc <- phase1_npcdist_cvls_gnn_cases()$bw.lc

  expect_true(is.finite(bw.lc$fval))
  expect_equal(bw.lc$fval, 0.1096745754682811, tolerance = 1e-10)
})

test_that("phase1 npcdistbw cv.ls generalized-nn keeps ll on canonical lp degree-1 glp", {
  cases <- phase1_npcdist_cvls_gnn_cases()
  degree <- cases$degree1
  bw.ll <- cases$bw.ll
  bw.lp <- cases$bw.lp

  expect_identical(bw.ll$regtype.engine, "lp")
  expect_identical(bw.ll$basis.engine, "glp")
  expect_identical(as.integer(bw.ll$degree.engine), degree)
  expect_true(is.finite(bw.ll$fval))
  expect_true(is.finite(bw.lp$fval))
  expect_equal(bw.ll$fval, 0.10210193919818145, tolerance = 1e-10)
  expect_equal(bw.lp$fval, 0.10210193919818145, tolerance = 1e-10)
  expect_equal(bw.ll$fval, bw.lp$fval, tolerance = 1e-10)
})

test_that("phase1 npcdistbw cv.ls generalized-nn lp degree-2 succeeds on a higher-order fixture", {
  cases <- phase1_npcdist_cvls_gnn_cases()
  bw.d1 <- cases$bw.lp
  bw.d2 <- cases$bw.d2

  expect_identical(as.integer(bw.d2$degree.engine), cases$degree2)
  expect_true(is.finite(bw.d2$fval))
  expect_equal(bw.d2$fval, 0.093788759039826752, tolerance = 1e-10)
  expect_gt(abs(bw.d2$fval - bw.d1$fval), 1e-6)
})

test_that("phase1 npcdistbw cv.ls generalized-nn avoids search-boundary collapse", {
  set.seed(20260309)
  n <- 30L
  x <- data.frame(x1 = runif(n))
  y <- data.frame(y1 = rbeta(n, shape1 = 1 + 2 * x$x1, shape2 = 2))

  bw <- npcdistbw(
    xdat = x,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = 1L,
    bwtype = "generalized_nn",
    bwmethod = "cv.ls",
    nmulti = 1L,
    itmax = 1L
  )
  fit <- npcdist(
    bws = bw,
    exdat = data.frame(x1 = rep(0.5, 4L)),
    eydat = data.frame(y1 = c(0, 0.02, 0.98, 1))
  )
  pred <- fitted(fit)

  expect_true(is.finite(bw$fval))
  expect_true(all(is.finite(pred)))
  expect_true(all(pred >= -1e-6))
  expect_true(all(pred <= 1 + 1e-6))
  expect_gt(pred[3] - pred[1], 0.5)
  expect_true(all(bw$xbw > 1))
  expect_true(all(bw$ybw > 1))
  expect_true(all(bw$xbw < n))
  expect_true(all(bw$ybw < n))
})

library(np)

test_that("predict aliases newdata to native eval args for default npreg/npudens/npudist/npindex", {
  set.seed(20260224)
  x <- runif(70)
  y <- rnorm(70)
  nd <- data.frame(x = c(0.1, 0.3, 0.7))

  bw.reg <- npregbw(
    xdat = data.frame(x = x),
    ydat = y,
    bws = 0.25,
    bandwidth.compute = FALSE
  )
  fit.reg <- npreg(bws = bw.reg, txdat = data.frame(x = x), tydat = y)
  expect_equal(
    as.numeric(predict(fit.reg, newdata = nd)),
    as.numeric(predict(fit.reg, exdat = nd)),
    tolerance = 1e-12
  )

  bw.den <- npudensbw(
    dat = data.frame(x = x),
    bws = 0.25,
    bandwidth.compute = FALSE
  )
  fit.den <- npudens(bws = bw.den, tdat = data.frame(x = x))
  expect_equal(
    as.numeric(predict(fit.den, newdata = nd)),
    as.numeric(predict(fit.den, edat = nd)),
    tolerance = 1e-12
  )

  bw.dist <- npudistbw(
    dat = data.frame(x = x),
    bws = 0.25,
    bandwidth.compute = FALSE
  )
  fit.dist <- npudist(bws = bw.dist, tdat = data.frame(x = x))
  expect_equal(
    as.numeric(predict(fit.dist, newdata = nd)),
    as.numeric(predict(fit.dist, edat = nd)),
    tolerance = 1e-12
  )

  x2 <- runif(70)
  nd.si <- data.frame(x = c(0.15, 0.35), x2 = c(0.4, 0.8))
  bw.si <- npindexbw(
    xdat = data.frame(x = x, x2 = x2),
    ydat = y,
    bws = c(0.25, 0.25, 1),
    bandwidth.compute = FALSE
  )
  fit.si <- npindex(
    bws = bw.si,
    txdat = data.frame(x = x, x2 = x2),
    tydat = y
  )
  expect_equal(
    as.numeric(predict(fit.si, newdata = nd.si)),
    as.numeric(predict(fit.si, exdat = nd.si)),
    tolerance = 1e-12
  )
})

test_that("predict aliases newdata to exdat/eydat for default npcdens/npcdist", {
  set.seed(20260224)
  x <- runif(60)
  y <- runif(60)
  nd <- data.frame(y = c(0.2, 0.5), x = c(0.1, 0.8))

  bw.cd <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.25, 0.2),
    bandwidth.compute = FALSE
  )
  fit.cd <- npcdens(
    bws = bw.cd,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y)
  )
  expect_equal(
    as.numeric(predict(fit.cd, newdata = nd)),
    as.numeric(predict(fit.cd, exdat = nd["x"], eydat = nd["y"])),
    tolerance = 1e-12
  )
  expect_error(
    predict(fit.cd, newdata = data.frame(x = c(0.1, 0.2))),
    "must include columns"
  )

  bw.cdist <- npcdistbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.25, 0.2),
    bandwidth.compute = FALSE
  )
  fit.cdist <- npcdist(
    bws = bw.cdist,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y)
  )
  expect_equal(
    as.numeric(predict(fit.cdist, newdata = nd)),
    as.numeric(predict(fit.cdist, exdat = nd["x"], eydat = nd["y"])),
    tolerance = 1e-12
  )
})

test_that("predict aliases newdata to exdat for default npqreg", {
  set.seed(20260512)
  x <- data.frame(x = runif(50))
  y <- rnorm(50)
  nd <- data.frame(x = c(0.2, 0.5, 0.8))

  bw <- npcdistbw(
    xdat = x,
    ydat = data.frame(y = y),
    bws = c(0.25, 0.25),
    bandwidth.compute = FALSE
  )
  fit <- npqreg(bws = bw, txdat = x, tydat = y, tau = 0.5)

  expect_equal(
    as.numeric(predict(fit, newdata = nd)),
    as.numeric(predict(fit, exdat = nd)),
    tolerance = 1e-12
  )

  nd.native <- data.frame(x = c(0.15, 0.35, 0.65))
  nd.standard <- data.frame(x = c(0.25, 0.55, 0.85))
  expect_equal(
    as.numeric(predict(fit, newdata = nd.standard, exdat = nd.native)),
    as.numeric(predict(fit, exdat = nd.native)),
    tolerance = 1e-12
  )
  expect_error(
    predict(fit, newdata = data.frame(z = c(0.2, 0.5, 0.8))),
    "newdata must contain columns"
  )
})

test_that("predict supports formula newdata and vector tau for npqreg", {
  set.seed(20260512)
  dat <- data.frame(
    y = sin(runif(70)) + rnorm(70, sd = 0.1),
    x = runif(70)
  )
  nd <- data.frame(x = c(0.2, 0.5, 0.8))

  bw <- npcdistbw(
    y ~ x,
    data = dat,
    bws = c(0.25, 0.25),
    bandwidth.compute = FALSE
  )
  fit <- npqreg(bws = bw, tau = c(0.25, 0.5), gradients = TRUE)

  pred <- predict(fit, newdata = nd)
  expect_equal(dim(pred), c(nrow(nd), 2L))

  se.pred <- predict(fit, newdata = nd, se.fit = TRUE)
  expect_named(se.pred, c("fit", "se.fit", "df", "residual.scale"))
  expect_equal(dim(se.pred$fit), c(nrow(nd), 2L))
  expect_equal(dim(se.pred$se.fit), c(nrow(nd), 2L))

  nd.native <- data.frame(x = c(0.15, 0.35))
  expect_equal(
    as.numeric(predict(fit, exdat = nd.native, newdata = nd)),
    as.numeric(predict(fit, newdata = nd.native)),
    tolerance = 1e-12
  )
  expect_error(
    predict(fit, se.fit = "yes"),
    "'se.fit' must be TRUE or FALSE",
    fixed = TRUE
  )
  x <- rep(0.1, nrow(nd))
  expect_error(
    npqreg(bws = bw, newdata = data.frame(z = x), tau = 0.5),
    "newdata must contain columns"
  )
  expect_error(
    predict(fit, newdata = data.frame(z = x)),
    "newdata must contain columns"
  )
})

test_that("predict supports conmode formula newdata without changing extraction behavior", {
  set.seed(20260512)
  dat <- data.frame(
    y = factor(rbinom(80, 1, 0.45)),
    x = runif(80),
    z = factor(rbinom(80, 1, 0.5))
  )
  bw <- npcdensbw(
    y ~ x + z,
    data = dat,
    bws = c(0.25, 0.3, 0.4),
    bandwidth.compute = FALSE
  )
  fit <- npconmode(bws = bw, probabilities = TRUE)
  fit.no.prob <- npconmode(bws = bw, probabilities = FALSE)
  nd <- dat[1:5, c("x", "z")]
  nd.with.y <- dat[1:5, c("y", "x", "z")]

  expect_equal(predict(fit), fit$conmode)
  expect_equal(predict(fit, type = "prob"), fit$probabilities)
  expect_error(
    predict(fit.no.prob, type = "prob"),
    "class probabilities are not stored"
  )

  ref <- npconmode(bws = fit$bws, newdata = nd)
  expect_equal(as.character(predict(fit, newdata = nd)), as.character(ref$conmode))

  ref.prob <- npconmode(bws = fit$bws, newdata = nd, probabilities = TRUE)
  expect_equal(
    unname(predict(fit, newdata = nd, type = "prob")),
    unname(ref.prob$probabilities),
    tolerance = 1e-12
  )

  ref.with.y <- npconmode(bws = fit$bws, newdata = nd.with.y)
  expect_equal(
    as.character(predict(fit, newdata = nd.with.y)),
    as.character(ref.with.y$conmode)
  )
  expect_error(
    predict(fit, newdata = nd, type = "prob", probabilities = FALSE),
    "requires probabilities=TRUE"
  )
  x <- rep(0.1, nrow(nd))
  expect_error(
    npconmode(bws = bw, newdata = data.frame(wrong = x)),
    "newdata must contain columns"
  )
  expect_error(
    predict(fit, newdata = data.frame(wrong = x)),
    "newdata must contain columns"
  )
})

test_that("predict supports conmode native evaluation and native precedence", {
  set.seed(20260512)
  xdat <- data.frame(
    x = runif(70),
    z = factor(rbinom(70, 1, 0.5))
  )
  ydat <- factor(rbinom(70, 1, 0.4))
  bw <- npcdensbw(
    xdat = xdat,
    ydat = data.frame(y = ydat),
    bws = c(0.25, 0.3, 0.4),
    bandwidth.compute = FALSE
  )
  fit <- npconmode(bws = bw, txdat = xdat, tydat = ydat, probabilities = TRUE)
  nd <- xdat[1:5, , drop = FALSE]
  nd.native <- xdat[6:10, , drop = FALSE]

  expect_equal(
    as.character(predict(fit, newdata = nd)),
    as.character(predict(fit, exdat = nd))
  )
  expect_equal(
    unname(predict(fit, newdata = nd, type = "prob")),
    unname(predict(fit, exdat = nd, type = "prob")),
    tolerance = 1e-12
  )
  expect_equal(
    as.character(predict(fit, newdata = nd, exdat = NULL)),
    as.character(predict(fit, newdata = nd))
  )
  expect_error(
    predict(fit, eydat = data.frame(y = ydat[1:5])),
    "eydat.*exdat"
  )
  expect_equal(
    as.character(predict(fit, newdata = nd, exdat = nd.native)),
    as.character(predict(fit, exdat = nd.native))
  )
})

test_that("predict aliases newdata to the explicit-evaluation slice route for npcdens", {
  set.seed(20260322)
  x <- runif(70, -1, 1)
  y <- sin(2 * pi * x) + rnorm(70, sd = 0.2)
  nd <- rbind(
    data.frame(y = c(-0.7, -0.15, 0.45), x = rep(-0.35, 3L)),
    data.frame(y = c(-0.25, 0.2, 0.75, 1.0), x = rep(0.4, 4L))
  )
  ctrl <- list(mode = "slice", slice.grid.size = 21L, slice.extend.factor = 0)

  bw.cd <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.27, 0.21),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )
  fit.cd <- npcdens(
    bws = bw.cd,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE
  )

  expect_equal(
    as.numeric(predict(fit.cd, newdata = nd, proper.control = ctrl)),
    as.numeric(predict(
      fit.cd,
      exdat = nd["x"],
      eydat = nd["y"],
      proper.control = ctrl
    )),
    tolerance = 1e-10
  )
})

test_that("predict newdata is unchanged when proper apply='fitted' targets only fitted values", {
  set.seed(20260322)
  x <- runif(70, -1, 1)
  y <- sin(2 * pi * x) + rnorm(70, sd = 0.2)
  nd <- rbind(
    data.frame(y = c(-0.7, -0.15, 0.45), x = rep(-0.35, 3L)),
    data.frame(y = c(-0.25, 0.2, 0.75, 1.0), x = rep(0.4, 4L))
  )
  ctrl <- list(mode = "slice", apply = "fitted", slice.grid.size = 21L, slice.extend.factor = 0)

  bw.cd <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.27, 0.21),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )
  fit.cd <- npcdens(
    bws = bw.cd,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE
  )

  expect_equal(
    as.numeric(predict(fit.cd, newdata = nd, proper = FALSE)),
    as.numeric(predict(fit.cd, newdata = nd, proper.control = ctrl)),
    tolerance = 1e-12
  )
})

test_that("predict aliases newdata to the explicit-evaluation slice route for npcdist", {
  set.seed(20260322)
  x <- runif(70, -1, 1)
  y <- sin(2 * pi * x) + rnorm(70, sd = 0.2)
  nd <- rbind(
    data.frame(y = c(-0.7, -0.15, 0.45), x = rep(-0.35, 3L)),
    data.frame(y = c(-0.25, 0.2, 0.75, 1.0), x = rep(0.4, 4L))
  )
  ctrl <- list(mode = "slice", slice.grid.size = 21L, slice.extend.factor = 0)

  bw.cd <- npcdistbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.27, 0.21),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )
  fit.cd <- npcdist(
    bws = bw.cd,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE
  )

  expect_equal(
    as.numeric(predict(fit.cd, newdata = nd, proper.control = ctrl)),
    as.numeric(predict(
      fit.cd,
      exdat = nd["x"],
      eydat = nd["y"],
      proper.control = ctrl
    )),
    tolerance = 1e-10
  )
})

test_that("predict newdata for npcdist is unchanged when proper apply='fitted' targets only fitted values", {
  set.seed(20260322)
  x <- runif(70, -1, 1)
  y <- sin(2 * pi * x) + rnorm(70, sd = 0.2)
  nd <- rbind(
    data.frame(y = c(-0.7, -0.15, 0.45), x = rep(-0.35, 3L)),
    data.frame(y = c(-0.25, 0.2, 0.75, 1.0), x = rep(0.4, 4L))
  )
  ctrl <- list(mode = "slice", apply = "fitted", slice.grid.size = 21L, slice.extend.factor = 0)

  bw.cd <- npcdistbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.27, 0.21),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )
  fit.cd <- npcdist(
    bws = bw.cd,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE
  )

  expect_equal(
    as.numeric(predict(fit.cd, newdata = nd, proper = FALSE)),
    as.numeric(predict(fit.cd, newdata = nd, proper.control = ctrl)),
    tolerance = 1e-12
  )
})

test_that("predict aliases newdata to exdat/ezdat for default npscoef/npplreg", {
  set.seed(20260227)
  n <- 80
  x <- runif(n)
  z <- runif(n)
  y <- sin(2 * pi * x) + 1.5 * z + rnorm(n, sd = 0.05)
  nd <- data.frame(x = c(0.2, 0.5, 0.8), z = c(0.1, 0.4, 0.9))

  bw.sc <- npscoefbw(
    xdat = data.frame(x = x),
    ydat = y,
    zdat = data.frame(z = z),
    bws = 0.25,
    bandwidth.compute = FALSE
  )
  fit.sc <- npscoef(
    bws = bw.sc,
    txdat = data.frame(x = x),
    tydat = y,
    tzdat = data.frame(z = z)
  )
  expect_equal(
    as.numeric(predict(fit.sc, newdata = nd)),
    as.numeric(predict(fit.sc, exdat = nd["x"], ezdat = nd["z"])),
    tolerance = 1e-12
  )
  expect_error(
    predict(fit.sc, newdata = data.frame(x = nd$x)),
    "must include columns"
  )

  bw.pl <- npplregbw(
    xdat = data.frame(z = z),
    ydat = y,
    zdat = data.frame(x = x),
    bws = matrix(c(0.25, 0.25), nrow = 2),
    bandwidth.compute = FALSE
  )
  fit.pl <- npplreg(
    bws = bw.pl,
    txdat = data.frame(z = z),
    tydat = y,
    tzdat = data.frame(x = x)
  )
  expect_equal(
    as.numeric(predict(fit.pl, newdata = nd)),
    as.numeric(predict(fit.pl, exdat = nd["z"], ezdat = nd["x"])),
    tolerance = 1e-12
  )
  expect_error(
    predict(fit.pl, newdata = data.frame(x = nd$x)),
    "must include columns"
  )
})

library(np)

set.seed(101)

make_xy <- function(n = 24L) {
  x <- data.frame(x = stats::runif(n))
  y <- data.frame(y = x$x + stats::rnorm(n, sd = 0.1))
  list(x = x, y = y)
}

make_cat_xy <- function(n = 80L) {
  x <- data.frame(
    f = factor(stats::rbinom(n, 1L, 0.45)),
    g = ordered(sample(letters[1:3], n, replace = TRUE))
  )
  y <- data.frame(
    y = 0.4 * as.integer(x$f) + 0.2 * as.integer(x$g) +
      stats::rnorm(n, sd = 0.3)
  )
  list(x = x, y = y)
}

bw_num <- function(bw) as.numeric(unlist(bw$bw, use.names = FALSE))

test_that("npcdensbw stores canonical ll/lp metadata", {
  d <- make_xy()

  bw.lc <- npcdensbw(
    xdat = d$x,
    ydat = d$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE
  )
  expect_identical(bw.lc$regtype, "lc")
  expect_identical(bw.lc$regtype.engine, "lc")

  bw.ll <- npcdensbw(
    xdat = d$x,
    ydat = d$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  expect_identical(bw.ll$regtype, "ll")
  expect_identical(bw.ll$regtype.engine, "lp")
  expect_identical(bw.ll$basis.engine, "glp")
  expect_identical(as.integer(bw.ll$degree.engine), 1L)
  expect_false(isTRUE(bw.ll$bernstein.basis.engine))

  bw.lp <- npcdensbw(
    xdat = d$x,
    ydat = d$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "tensor",
    degree = 2L,
    bernstein.basis = TRUE
  )
  expect_identical(bw.lp$regtype, "lp")
  expect_identical(bw.lp$regtype.engine, "lp")
  expect_identical(bw.lp$basis.engine, "tensor")
  expect_identical(as.integer(bw.lp$degree.engine), 2L)
  expect_true(isTRUE(bw.lp$bernstein.basis.engine))
})

test_that("npcdistbw stores canonical ll/lp metadata", {
  d <- make_xy()

  bw.ll <- npcdistbw(
    xdat = d$x,
    ydat = d$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  expect_identical(bw.ll$regtype, "ll")
  expect_identical(bw.ll$regtype.engine, "lp")
  expect_identical(bw.ll$basis.engine, "glp")
  expect_identical(as.integer(bw.ll$degree.engine), 1L)

  bw.lp <- npcdistbw(
    xdat = d$x,
    ydat = d$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "additive",
    degree = 3L
  )
  expect_identical(bw.lp$regtype.engine, "lp")
  expect_identical(bw.lp$basis.engine, "additive")
  expect_identical(as.integer(bw.lp$degree.engine), 3L)
})

test_that("npc* conditional regtype argument contracts fail fast", {
  d <- make_xy()

  expect_error(
    npcdensbw(
      xdat = d$x,
      ydat = d$y,
      bws = c(0.35, 0.35),
      bandwidth.compute = FALSE,
      regtype = "ll",
      degree = 2L
    ),
    "canonical LP\\(degree=1"
  )

  expect_error(
    npcdistbw(
      xdat = d$x,
      ydat = d$y,
      bws = c(0.35, 0.35),
      bandwidth.compute = FALSE,
      regtype = "lc",
      basis = "glp"
    ),
    "regtype='lc' does not accept basis/degree/bernstein.basis"
  )
})

test_that("npcdens ll matches lp(degree=1, basis='glp')", {
  d <- make_xy()

  bw.ll <- npcdensbw(
    xdat = d$x,
    ydat = d$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  bw.lp <- npcdensbw(
    xdat = d$x,
    ydat = d$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = 1L
  )

  fit.ll <- npcdens(bws = bw.ll, txdat = d$x, tydat = d$y, gradients = TRUE)
  fit.lp <- npcdens(bws = bw.lp, txdat = d$x, tydat = d$y, gradients = TRUE)

  expect_equal(fitted(fit.ll), fitted(fit.lp), tolerance = 1e-10)
  expect_equal(fit.ll$congrad, fit.lp$congrad, tolerance = 1e-10)
})

test_that("npcdist ll matches lp(degree=1, basis='glp')", {
  d <- make_xy()

  bw.ll <- npcdistbw(
    xdat = d$x,
    ydat = d$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  bw.lp <- npcdistbw(
    xdat = d$x,
    ydat = d$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = 1L
  )

  fit.ll <- npcdist(bws = bw.ll, txdat = d$x, tydat = d$y, gradients = TRUE)
  fit.lp <- npcdist(bws = bw.lp, txdat = d$x, tydat = d$y, gradients = TRUE)

  expect_equal(fitted(fit.ll), fitted(fit.lp), tolerance = 1e-10)
  expect_equal(fit.ll$congrad, fit.lp$congrad, tolerance = 1e-10)
})

test_that("npcdens categorical-only predictors reject impossible degree structure", {
  d <- make_cat_xy()

  set.seed(90210)
  bw.lc <- npcdensbw(
    xdat = d$x,
    ydat = d$y,
    regtype = "lc",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  expect_error(
    npcdensbw(xdat = d$x, ydat = d$y, regtype = "ll",
              bwmethod = "cv.ls", nmulti = 1L),
    "requires at least one continuous predictor"
  )
  expect_error(
    npcdensbw(xdat = d$x, ydat = d$y, regtype = "lp", degree = 1L,
              bwmethod = "cv.ls", nmulti = 1L),
    "degree must be 0"
  )
  expect_error(
    npcdensbw(xdat = d$x, ydat = d$y, regtype = "lp",
              bwmethod = "cv.ls", nmulti = 1L),
    "degree must be 0"
  )
  expect_error(
    npcdensbw(xdat = d$x, ydat = d$y, regtype = "lp",
              degree.select = "coordinate", search.engine = "cell",
              bwmethod = "cv.ls", nmulti = 1L),
    "automatic degree search requires at least one continuous"
  )
  set.seed(90210)
  bw.lp <- npcdensbw(
    xdat = d$x,
    ydat = d$y,
    regtype = "lp",
    degree = 0L,
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_identical(bw.lp$regtype, "lp")
  expect_identical(bw.lp$regtype.engine, "lc")
  expect_equal(as.numeric(bw.lc$fval), as.numeric(bw.lp$fval), tolerance = 1e-12)
  expect_equal(bw_num(bw.lc), bw_num(bw.lp), tolerance = 1e-12)
  expect_error(
    npcdensbw(xdat = d$x, ydat = d$y, nomad = TRUE),
    "nomad=TRUE requires at least one continuous predictor"
  )
})

test_that("npcdist categorical-only predictors reject impossible degree structure", {
  d <- make_cat_xy()

  set.seed(90210)
  bw.lc <- npcdistbw(
    xdat = d$x,
    ydat = d$y,
    regtype = "lc",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  expect_error(
    npcdistbw(xdat = d$x, ydat = d$y, regtype = "ll",
              bwmethod = "cv.ls", nmulti = 1L),
    "requires at least one continuous predictor"
  )
  expect_error(
    npcdistbw(xdat = d$x, ydat = d$y, regtype = "lp", degree = 1L,
              bwmethod = "cv.ls", nmulti = 1L),
    "degree must be 0"
  )
  expect_error(
    npcdistbw(xdat = d$x, ydat = d$y, regtype = "lp",
              bwmethod = "cv.ls", nmulti = 1L),
    "degree must be 0"
  )
  expect_error(
    npcdistbw(xdat = d$x, ydat = d$y, regtype = "lp",
              degree.select = "coordinate", search.engine = "cell",
              bwmethod = "cv.ls", nmulti = 1L),
    "automatic degree search requires at least one continuous"
  )
  set.seed(90210)
  bw.lp <- npcdistbw(
    xdat = d$x,
    ydat = d$y,
    regtype = "lp",
    degree = 0L,
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_identical(bw.lp$regtype, "lp")
  expect_identical(bw.lp$regtype.engine, "lc")
  expect_equal(as.numeric(bw.lc$fval), as.numeric(bw.lp$fval), tolerance = 1e-12)
  expect_equal(bw_num(bw.lc), bw_num(bw.lp), tolerance = 1e-12)
  expect_error(
    npcdistbw(xdat = d$x, ydat = d$y, nomad = TRUE),
    "nomad=TRUE requires at least one continuous predictor"
  )
})

test_that("npcdens LC fixed no-error plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npcdens_lc_fixed_none_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(123)

  n <- 80L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    nmulti = 1L,
    regtype = "lc",
    bwtype = "fixed"
  )

  old <- plot(
    bw,
    xdat = x,
    ydat = y,
    output = "data",
    errors = "none",
    view = "fixed",
    neval = 9L,
    perspective = TRUE
  )
  candidate <- proto(bw, xdat = x, ydat = y, neval = 9L)
  stages <- proto(bw, xdat = x, ydat = y, neval = 9L, return.stages = TRUE)

  expect_named(candidate, names(old))
  expect_s3_class(candidate$cd1, "condensity")
  expect_named(candidate$cd1, names(old$cd1))
  expect_equal(candidate$cd1$xeval, old$cd1$xeval)
  expect_equal(candidate$cd1$yeval, old$cd1$yeval)
  expect_equal(candidate$cd1$condens, old$cd1$condens)
  expect_equal(candidate$cd1$conderr, old$cd1$conderr)
  expect_equal(candidate$cd1$bias, old$cd1$bias)
  expect_identical(candidate$cd1$proper.requested, old$cd1$proper.requested)
  expect_identical(candidate$cd1$proper.applied, old$cd1$proper.applied)
  expect_named(stages, c("state", "target_grid", "evaluator", "intervals", "bootstrap", "proper_projection", "plot_data"))
  expect_null(stages$intervals)
  expect_null(stages$bootstrap)
  expect_null(stages$proper_projection)
  expect_equal(stages$plot_data, candidate)
  expect_equal(stages$evaluator$condens, old$cd1$condens)
})

test_that("npcdens plot prototype fails early outside its vertical slice", {
  proto <- getFromNamespace(".np_plot_proto_npcdens_lc_fixed_none_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(124)

  n <- 60L
  x <- data.frame(x = factor(rbinom(n, 1L, 0.5)))
  y <- data.frame(y = rnorm(n))
  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    nmulti = 1L,
    regtype = "lc",
    bwtype = "fixed"
  )

  expect_error(
    proto(bw, xdat = x, ydat = y, neval = 9L),
    "continuous/ordered surface variables",
    fixed = TRUE
  )

  expect_error(
    proto(bw, neval = 9L),
    "explicit xdat and ydat",
    fixed = TRUE
  )
})

test_that("npcdens LC fixed asymptotic plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npcdens_lc_fixed_asymptotic_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(125)

  n <- 80L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    nmulti = 1L,
    regtype = "lc",
    bwtype = "fixed"
  )

  for (band in c("pmzsd", "pointwise", "bonferroni", "simultaneous", "all")) {
    old <- suppressWarnings(plot(
      bw,
      xdat = x,
      ydat = y,
      output = "data",
      errors = "asymptotic",
      band = band,
      view = "fixed",
      neval = 9L,
      perspective = TRUE
    ))
    candidate <- proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 9L,
      band = band
    )
    stages <- proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 9L,
      band = band,
      return.stages = TRUE
    )

    expect_equal(candidate$cd1$xeval, old$cd1$xeval, info = band)
    expect_equal(candidate$cd1$yeval, old$cd1$yeval, info = band)
    expect_equal(candidate$cd1$condens, old$cd1$condens, info = band)
    expect_equal(candidate$cd1$conderr, old$cd1$conderr, info = band)
    expect_named(stages, c("state", "target_grid", "evaluator", "intervals", "bootstrap", "proper_projection", "plot_data"))
    expect_equal(stages$plot_data, candidate, info = band)
    expect_identical(stages$intervals$method, "asymptotic")
    expect_identical(stages$intervals$type, band)
    expect_null(stages$bootstrap)
    expect_null(stages$proper_projection)
  }
})

test_that("npcdens LC fixed inid bootstrap plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npcdens_lc_fixed_bootstrap_inid_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(126)

  n <- 70L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    nmulti = 1L,
    regtype = "lc",
    bwtype = "fixed"
  )

  cases <- expand.grid(
    band = c("pmzsd", "pointwise", "bonferroni"),
    center = c("estimate", "bias-corrected"),
    stringsAsFactors = FALSE
  )
  for (ii in seq_len(nrow(cases))) {
    band <- cases$band[ii]
    center <- cases$center[ii]
    boot.seed <- 1000 + ii
    set.seed(boot.seed)
    old <- suppressWarnings(plot(
      bw,
      xdat = x,
      ydat = y,
      output = "data",
      errors = "bootstrap",
      bootstrap = "inid",
      B = 19L,
      center = center,
      band = band,
      view = "fixed",
      neval = 7L,
      perspective = TRUE,
      random.seed = boot.seed
    ))
    set.seed(boot.seed)
    candidate <- suppressWarnings(proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 7L,
      B = 19L,
      center = center,
      band = band
    ))
    set.seed(boot.seed)
    stages <- suppressWarnings(proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 7L,
      B = 19L,
      center = center,
      band = band,
      return.stages = TRUE
    ))

    expect_equal(candidate$cd1$xeval, old$cd1$xeval, info = paste(band, center))
    expect_equal(candidate$cd1$yeval, old$cd1$yeval, info = paste(band, center))
    expect_equal(candidate$cd1$condens, old$cd1$condens, info = paste(band, center))
    expect_equal(candidate$cd1$conderr, old$cd1$conderr, info = paste(band, center))
    expect_named(stages, c("state", "target_grid", "evaluator", "intervals", "bootstrap", "proper_projection", "plot_data"))
    expect_equal(stages$plot_data, candidate, info = paste(band, center))
    expect_identical(stages$intervals$method, "bootstrap")
    expect_identical(stages$intervals$type, band)
    expect_identical(stages$bootstrap$method, "inid")
    expect_identical(stages$bootstrap$B, 19L)
    expect_equal(stages$bootstrap$boot.err[, 1:2, drop = FALSE],
                 candidate$cd1$conderr,
                 info = paste(band, center))
    expect_null(stages$proper_projection)
  }
})

test_that("npcdens LC fixed proper plot-data prototype matches current route", {
  none_proto <- getFromNamespace(".np_plot_proto_npcdens_lc_fixed_none_data", "np")
  boot_proto <- getFromNamespace(".np_plot_proto_npcdens_lc_fixed_bootstrap_inid_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(127)

  n <- 75L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    nmulti = 1L,
    regtype = "lc",
    bwtype = "fixed"
  )

  old_none <- plot(
    bw,
    xdat = x,
    ydat = y,
    output = "data",
    errors = "none",
    proper = TRUE,
    view = "fixed",
    neval = 7L,
    perspective = TRUE
  )
  cand_none <- none_proto(
    bw,
    xdat = x,
    ydat = y,
    neval = 7L,
    proper = TRUE
  )
  stages_none <- none_proto(
    bw,
    xdat = x,
    ydat = y,
    neval = 7L,
    proper = TRUE,
    return.stages = TRUE
  )

  expect_equal(cand_none$cd1$condens, old_none$cd1$condens)
  expect_equal(cand_none$cd1$conderr, old_none$cd1$conderr)
  expect_identical(cand_none$cd1$proper.requested, old_none$cd1$proper.requested)
  expect_identical(cand_none$cd1$proper.applied, old_none$cd1$proper.applied)
  expect_identical(stages_none$proper_projection$requested, TRUE)
  expect_identical(stages_none$proper_projection$applied, cand_none$cd1$proper.applied)

  boot.seed <- 9001L
  set.seed(boot.seed)
  old_boot <- suppressWarnings(plot(
    bw,
    xdat = x,
    ydat = y,
    output = "data",
    errors = "bootstrap",
    bootstrap = "inid",
    B = 19L,
    band = "pmzsd",
    proper = TRUE,
    view = "fixed",
    neval = 7L,
    perspective = TRUE,
    random.seed = boot.seed
  ))
  set.seed(boot.seed)
  cand_boot <- suppressWarnings(boot_proto(
    bw,
    xdat = x,
    ydat = y,
    neval = 7L,
    B = 19L,
    band = "pmzsd",
    proper = TRUE
  ))
  set.seed(boot.seed)
  stages_boot <- suppressWarnings(boot_proto(
    bw,
    xdat = x,
    ydat = y,
    neval = 7L,
    B = 19L,
    band = "pmzsd",
    proper = TRUE,
    return.stages = TRUE
  ))

  expect_equal(cand_boot$cd1$condens, old_boot$cd1$condens)
  expect_equal(cand_boot$cd1$conderr, old_boot$cd1$conderr)
  expect_identical(cand_boot$cd1$proper.requested, old_boot$cd1$proper.requested)
  expect_identical(cand_boot$cd1$proper.applied, old_boot$cd1$proper.applied)
  expect_identical(stages_boot$proper_projection$requested, TRUE)
  expect_identical(stages_boot$proper_projection$applied, cand_boot$cd1$proper.applied)
  expect_equal(stages_boot$bootstrap$boot.err[, 1:2, drop = FALSE],
               cand_boot$cd1$conderr)
})

test_that("npcdens LC fixed block bootstrap plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npcdens_lc_fixed_bootstrap_block_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(128)

  n <- 70L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    nmulti = 1L,
    regtype = "lc",
    bwtype = "fixed"
  )

  for (method in c("fixed", "geom")) {
    boot.seed <- if (identical(method, "fixed")) 9101L else 9102L
    set.seed(boot.seed)
    old <- suppressWarnings(plot(
      bw,
      xdat = x,
      ydat = y,
      output = "data",
      errors = "bootstrap",
      bootstrap = method,
      boot_control = np_boot_control(blocklen = 5L),
      B = 19L,
      band = "pmzsd",
      view = "fixed",
      neval = 7L,
      perspective = TRUE,
      random.seed = boot.seed
    ))
    set.seed(boot.seed)
    candidate <- suppressWarnings(proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 7L,
      bootstrap = method,
      boot_control = np_boot_control(blocklen = 5L),
      B = 19L,
      band = "pmzsd"
    ))
    set.seed(boot.seed)
    stages <- suppressWarnings(proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 7L,
      bootstrap = method,
      boot_control = np_boot_control(blocklen = 5L),
      B = 19L,
      band = "pmzsd",
      return.stages = TRUE
    ))

    expect_equal(candidate$cd1$condens, old$cd1$condens, info = method)
    expect_equal(candidate$cd1$conderr, old$cd1$conderr, info = method)
    expect_identical(stages$bootstrap$method, method)
    expect_identical(stages$bootstrap$blocklen, 5L)
    expect_equal(stages$bootstrap$boot.err[, 1:2, drop = FALSE],
                 candidate$cd1$conderr,
                 info = method)
  }
})

test_that("npcdens fixed LL/LP no-error plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npcdens_fixed_none_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(129)

  n <- 75L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  cases <- list(
    ll = list(regtype = "ll"),
    lp1 = list(regtype = "lp", degree = 1L),
    lp2 = list(regtype = "lp", degree = 2L)
  )

  for (nm in names(cases)) {
    bw <- do.call(
      npcdensbw,
      c(list(xdat = x, ydat = y, nmulti = 1L, bwtype = "fixed"), cases[[nm]])
    )
    old <- plot(
      bw,
      xdat = x,
      ydat = y,
      output = "data",
      errors = "none",
      view = "fixed",
      neval = 7L,
      perspective = TRUE
    )
    candidate <- proto(bw, xdat = x, ydat = y, neval = 7L)
    stages <- proto(bw, xdat = x, ydat = y, neval = 7L, return.stages = TRUE)

    expect_equal(candidate$cd1$xeval, old$cd1$xeval, info = nm)
    expect_equal(candidate$cd1$yeval, old$cd1$yeval, info = nm)
    expect_equal(candidate$cd1$condens, old$cd1$condens, info = nm)
    expect_equal(candidate$cd1$conderr, old$cd1$conderr, info = nm)
    expect_equal(stages$plot_data, candidate, info = nm)
  }
})

test_that("npcdens fixed LL/LP inid bootstrap plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npcdens_fixed_bootstrap_inid_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(130)

  n <- 70L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  cases <- list(
    ll = list(regtype = "ll"),
    lp1 = list(regtype = "lp", degree = 1L),
    lp2 = list(regtype = "lp", degree = 2L)
  )

  for (ii in seq_along(cases)) {
    nm <- names(cases)[ii]
    bw <- do.call(
      npcdensbw,
      c(list(xdat = x, ydat = y, nmulti = 1L, bwtype = "fixed"), cases[[ii]])
    )
    boot.seed <- 9200L + ii
    set.seed(boot.seed)
    old <- suppressWarnings(plot(
      bw,
      xdat = x,
      ydat = y,
      output = "data",
      errors = "bootstrap",
      bootstrap = "inid",
      B = 17L,
      band = "pmzsd",
      view = "fixed",
      neval = 7L,
      perspective = TRUE,
      random.seed = boot.seed
    ))
    set.seed(boot.seed)
    candidate <- suppressWarnings(proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 7L,
      B = 17L,
      band = "pmzsd"
    ))
    set.seed(boot.seed)
    stages <- suppressWarnings(proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 7L,
      B = 17L,
      band = "pmzsd",
      return.stages = TRUE
    ))

    expect_equal(candidate$cd1$condens, old$cd1$condens, info = nm)
    expect_equal(candidate$cd1$conderr, old$cd1$conderr, info = nm)
    expect_identical(stages$bootstrap$method, "inid")
    expect_equal(stages$bootstrap$boot.err[, 1:2, drop = FALSE],
                 candidate$cd1$conderr,
                 info = nm)
  }
})

test_that("npcdens generalized/adaptive NN no-error plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npcdens_fixed_none_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(131)

  n <- 70L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  cases <- c("generalized_nn", "adaptive_nn")

  for (bwtype in cases) {
    bw <- npcdensbw(
      xdat = x,
      ydat = y,
      nmulti = 1L,
      regtype = "lc",
      bwtype = bwtype
    )
    old <- plot(
      bw,
      xdat = x,
      ydat = y,
      output = "data",
      errors = "none",
      view = "fixed",
      neval = 7L,
      perspective = TRUE
    )
    candidate <- proto(bw, xdat = x, ydat = y, neval = 7L)
    stages <- proto(bw, xdat = x, ydat = y, neval = 7L, return.stages = TRUE)

    expect_equal(candidate$cd1$xeval, old$cd1$xeval, info = bwtype)
    expect_equal(candidate$cd1$yeval, old$cd1$yeval, info = bwtype)
    expect_equal(candidate$cd1$condens, old$cd1$condens, info = bwtype)
    expect_equal(candidate$cd1$conderr, old$cd1$conderr, info = bwtype)
    expect_equal(stages$plot_data, candidate, info = bwtype)
  }
})

test_that("npcdens generalized/adaptive NN inid bootstrap plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npcdens_fixed_bootstrap_inid_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(132)

  n <- 55L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  cases <- c("generalized_nn", "adaptive_nn")

  for (ii in seq_along(cases)) {
    bwtype <- cases[ii]
    bw <- npcdensbw(
      xdat = x,
      ydat = y,
      nmulti = 1L,
      regtype = "lc",
      bwtype = bwtype
    )
    boot.seed <- 9300L + ii
    set.seed(boot.seed)
    old <- suppressWarnings(plot(
      bw,
      xdat = x,
      ydat = y,
      output = "data",
      errors = "bootstrap",
      bootstrap = "inid",
      B = 11L,
      band = "pmzsd",
      view = "fixed",
      neval = 5L,
      perspective = TRUE,
      random.seed = boot.seed
    ))
    set.seed(boot.seed)
    candidate <- suppressWarnings(proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 5L,
      B = 11L,
      band = "pmzsd"
    ))
    set.seed(boot.seed)
    stages <- suppressWarnings(proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 5L,
      B = 11L,
      band = "pmzsd",
      return.stages = TRUE
    ))

    expect_equal(candidate$cd1$condens, old$cd1$condens, info = bwtype)
    expect_equal(candidate$cd1$conderr, old$cd1$conderr, info = bwtype)
    expect_identical(stages$bootstrap$method, "inid")
    expect_equal(stages$bootstrap$boot.err[, 1:2, drop = FALSE],
                 candidate$cd1$conderr,
                 info = bwtype)
  }
})

test_that("npcdist fixed no-error plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npcdist_fixed_none_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(134)

  n <- 75L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  cases <- list(
    lc = list(regtype = "lc"),
    ll = list(regtype = "ll"),
    lp1 = list(regtype = "lp", degree = 1L),
    lp2 = list(regtype = "lp", degree = 2L)
  )

  for (nm in names(cases)) {
    bw <- do.call(
      npcdistbw,
      c(list(xdat = x, ydat = y, nmulti = 1L, bwtype = "fixed"), cases[[nm]])
    )
    old <- plot(
      bw,
      xdat = x,
      ydat = y,
      output = "data",
      errors = "none",
      view = "fixed",
      neval = 7L,
      perspective = TRUE
    )
    candidate <- proto(bw, xdat = x, ydat = y, neval = 7L)
    stages <- proto(bw, xdat = x, ydat = y, neval = 7L, return.stages = TRUE)

    expect_s3_class(candidate$cd1, "condistribution")
    expect_equal(candidate$cd1$xeval, old$cd1$xeval, info = nm)
    expect_equal(candidate$cd1$yeval, old$cd1$yeval, info = nm)
    expect_equal(candidate$cd1$condist, old$cd1$condist, info = nm)
    expect_equal(candidate$cd1$conderr, old$cd1$conderr, info = nm)
    expect_identical(stages$state$family, "npcdist")
    expect_true(stages$state$cdf)
    expect_equal(stages$plot_data, candidate, info = nm)
  }
})

test_that("npcdist fixed inid bootstrap plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npcdist_fixed_bootstrap_inid_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(135)

  n <- 65L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  cases <- list(
    lc = list(regtype = "lc"),
    ll = list(regtype = "ll"),
    lp1 = list(regtype = "lp", degree = 1L),
    lp2 = list(regtype = "lp", degree = 2L)
  )

  for (ii in seq_along(cases)) {
    nm <- names(cases)[ii]
    bw <- do.call(
      npcdistbw,
      c(list(xdat = x, ydat = y, nmulti = 1L, bwtype = "fixed"), cases[[ii]])
    )
    boot.seed <- 9400L + ii
    set.seed(boot.seed)
    old <- suppressWarnings(plot(
      bw,
      xdat = x,
      ydat = y,
      output = "data",
      errors = "bootstrap",
      bootstrap = "inid",
      B = 13L,
      band = "pmzsd",
      view = "fixed",
      neval = 6L,
      perspective = TRUE,
      random.seed = boot.seed
    ))
    set.seed(boot.seed)
    candidate <- suppressWarnings(proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 6L,
      B = 13L,
      band = "pmzsd"
    ))
    set.seed(boot.seed)
    stages <- suppressWarnings(proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 6L,
      B = 13L,
      band = "pmzsd",
      return.stages = TRUE
    ))

    expect_equal(candidate$cd1$condist, old$cd1$condist, info = nm)
    expect_equal(candidate$cd1$conderr, old$cd1$conderr, info = nm)
    expect_identical(stages$bootstrap$method, "inid")
    expect_equal(stages$bootstrap$boot.err[, 1:2, drop = FALSE],
                 candidate$cd1$conderr,
                 info = nm)
  }
})

test_that("npcdist generalized/adaptive NN plot-data prototype matches current route", {
  none_proto <- getFromNamespace(".np_plot_proto_npcdist_fixed_none_data", "np")
  boot_proto <- getFromNamespace(".np_plot_proto_npcdist_fixed_bootstrap_inid_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(136)

  n <- 60L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))

  for (ii in seq_along(c("generalized_nn", "adaptive_nn"))) {
    bwtype <- c("generalized_nn", "adaptive_nn")[ii]
    bw <- npcdistbw(
      xdat = x,
      ydat = y,
      nmulti = 1L,
      regtype = "lc",
      bwtype = bwtype
    )
    old <- plot(
      bw,
      xdat = x,
      ydat = y,
      output = "data",
      errors = "none",
      view = "fixed",
      neval = 6L,
      perspective = TRUE
    )
    candidate <- none_proto(bw, xdat = x, ydat = y, neval = 6L)
    expect_equal(candidate$cd1$xeval, old$cd1$xeval, info = bwtype)
    expect_equal(candidate$cd1$yeval, old$cd1$yeval, info = bwtype)
    expect_equal(candidate$cd1$condist, old$cd1$condist, info = bwtype)
    expect_equal(candidate$cd1$conderr, old$cd1$conderr, info = bwtype)

    boot.seed <- 9500L + ii
    set.seed(boot.seed)
    old.boot <- suppressWarnings(plot(
      bw,
      xdat = x,
      ydat = y,
      output = "data",
      errors = "bootstrap",
      bootstrap = "inid",
      B = 11L,
      band = "pmzsd",
      view = "fixed",
      neval = 5L,
      perspective = TRUE,
      random.seed = boot.seed
    ))
    set.seed(boot.seed)
    candidate.boot <- suppressWarnings(boot_proto(
      bw,
      xdat = x,
      ydat = y,
      neval = 5L,
      B = 11L,
      band = "pmzsd"
    ))
    expect_equal(candidate.boot$cd1$condist, old.boot$cd1$condist, info = bwtype)
    expect_equal(candidate.boot$cd1$conderr, old.boot$cd1$conderr, info = bwtype)
  }
})

test_that("npcdist proper projection plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npcdist_fixed_none_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(137)

  n <- 70L
  x <- data.frame(x = runif(n, -1, 1))
  y <- data.frame(y = sin(2 * pi * x$x) + rnorm(n, sd = 0.2))
  bw <- npcdistbw(
    xdat = x,
    ydat = y,
    bws = c(0.28, 0.22),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 2L
  )
  old <- suppressWarnings(plot(
    bw,
    xdat = x,
    ydat = y,
    output = "data",
    errors = "none",
    view = "fixed",
    neval = 8L,
    perspective = TRUE,
    proper = TRUE
  ))
  candidate <- suppressWarnings(proto(
    bw,
    xdat = x,
    ydat = y,
    neval = 8L,
    proper = TRUE
  ))
  stages <- suppressWarnings(proto(
    bw,
    xdat = x,
    ydat = y,
    neval = 8L,
    proper = TRUE,
    return.stages = TRUE
  ))

  expect_true(isTRUE(candidate$cd1$proper.requested))
  expect_equal(candidate$cd1$proper.applied, old$cd1$proper.applied)
  expect_equal(candidate$cd1$condist, old$cd1$condist)
  expect_equal(candidate$cd1$conderr, old$cd1$conderr)
  expect_named(stages$proper_projection, c("requested", "applied", "method", "info"))
})

test_that("npcdens staged plot-data object renders through base renderer smoke", {
  proto <- getFromNamespace(".np_plot_proto_npcdens_fixed_none_data", "np")
  render <- getFromNamespace(".np_plot_proto_npcdens_surface_base_render", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(133)

  n <- 65L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    nmulti = 1L,
    regtype = "lp",
    degree = 1L,
    bwtype = "fixed"
  )
  candidate <- proto(bw, xdat = x, ydat = y, neval = 7L)

  out.file <- tempfile(fileext = ".pdf")
  grDevices::pdf(out.file)
  on.exit({
    if (grDevices::dev.cur() > 1L)
      grDevices::dev.off()
  }, add = TRUE)
  expect_identical(render(candidate, perspective = FALSE), candidate)
  expect_identical(render(candidate, perspective = TRUE), candidate)
  grDevices::dev.off()

  expect_true(file.exists(out.file))
  expect_gt(file.info(out.file)$size, 0)
})

test_that("npcdist staged plot-data object renders through base renderer smoke", {
  proto <- getFromNamespace(".np_plot_proto_npcdist_fixed_none_data", "np")
  render <- getFromNamespace(".np_plot_proto_npcdens_surface_base_render", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(138)

  n <- 65L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rnorm(n))
  bw <- npcdistbw(
    xdat = x,
    ydat = y,
    nmulti = 1L,
    regtype = "lp",
    degree = 1L,
    bwtype = "fixed"
  )
  candidate <- proto(bw, xdat = x, ydat = y, neval = 7L)

  out.file <- tempfile(fileext = ".pdf")
  grDevices::pdf(out.file)
  on.exit({
    if (grDevices::dev.cur() > 1L)
      grDevices::dev.off()
  }, add = TRUE)
  expect_identical(render(candidate, perspective = FALSE), candidate)
  expect_identical(render(candidate, perspective = TRUE), candidate)
  grDevices::dev.off()

  expect_true(file.exists(out.file))
  expect_gt(file.info(out.file)$size, 0)
})

test_that("npcdens nomad shortcut matches the explicit density preset", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260322)
  dat <- data.frame(x = sort(runif(14)), y = sort(runif(14)))

  bw_short <- np::npcdensbw(
    y ~ x,
    data = dat,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L
  )

  bw_long <- np::npcdensbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    search.engine = "nomad+powell",
    degree.select = "coordinate",
    bernstein.basis = TRUE,
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = FALSE,
    bwtype = "fixed",
    nmulti = 1L
  )

  fit_short <- local({
    npcdensbw <- np::npcdensbw
    np::npcdens(
      y ~ x,
      data = dat,
      nomad = TRUE,
      degree.max = 1L,
      nmulti = 1L
    )
  })

  fit_long <- local({
    npcdensbw <- np::npcdensbw
    np::npcdens(
      y ~ x,
      data = dat,
      regtype = "lp",
      search.engine = "nomad+powell",
      degree.select = "coordinate",
      bernstein.basis = TRUE,
      degree.min = 0L,
      degree.max = 1L,
      degree.verify = FALSE,
      bwtype = "fixed",
      nmulti = 1L
    )
  })

  expect_lt(max(abs(unname(bw_short$xbw) - unname(bw_long$xbw))), 5e-4)
  expect_lt(max(abs(unname(bw_short$ybw) - unname(bw_long$ybw))), 5e-4)
  expect_identical(as.integer(bw_short$degree), as.integer(bw_long$degree))
  expect_true(isTRUE(bw_short$bernstein.basis))
  expect_true(is.list(bw_short$nomad.shortcut))
  expect_equal(fitted(fit_short), fitted(fit_long), tolerance = 5e-4)
  expect_true(is.list(fit_short$bws$nomad.shortcut))
})

test_that("npcdist nomad shortcut matches the explicit distribution preset", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260322)
  dat <- data.frame(x = sort(runif(14)), y = sort(runif(14)))

  bw_short <- np::npcdistbw(
    y ~ x,
    data = dat,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L
  )

  bw_long <- np::npcdistbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    search.engine = "nomad+powell",
    degree.select = "coordinate",
    bernstein.basis = TRUE,
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = FALSE,
    bwtype = "fixed",
    nmulti = 1L
  )

  fit_short <- local({
    npcdistbw <- np::npcdistbw
    np::npcdist(
      y ~ x,
      data = dat,
      nomad = TRUE,
      degree.max = 1L,
      nmulti = 1L
    )
  })

  fit_long <- local({
    npcdistbw <- np::npcdistbw
    np::npcdist(
      y ~ x,
      data = dat,
      regtype = "lp",
      search.engine = "nomad+powell",
      degree.select = "coordinate",
      bernstein.basis = TRUE,
      degree.min = 0L,
      degree.max = 1L,
      degree.verify = FALSE,
      bwtype = "fixed",
      nmulti = 1L
    )
  })

  expect_equal(unname(bw_short$xbw), unname(bw_long$xbw), tolerance = 1e-8)
  expect_equal(unname(bw_short$ybw), unname(bw_long$ybw), tolerance = 1e-8)
  expect_identical(as.integer(bw_short$degree), as.integer(bw_long$degree))
  expect_true(isTRUE(bw_short$bernstein.basis))
  expect_true(is.list(bw_short$nomad.shortcut))
  expect_equal(fitted(fit_short), fitted(fit_long), tolerance = 1e-8)
  expect_true(is.list(fit_short$bws$nomad.shortcut))
})

test_that("npcdens and npcdist accept nomad.nmulti through the NOMAD shortcut route", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260323)
  dat <- data.frame(x = sort(runif(14)), y = sort(runif(14)))

  bw_cd <- np::npcdensbw(
    y ~ x,
    data = dat,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L,
    nomad.nmulti = 2L
  )
  bw_cdf <- np::npcdistbw(
    y ~ x,
    data = dat,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L,
    nomad.nmulti = 2L
  )

  expect_true(is.list(bw_cd$nomad.shortcut))
  expect_true(is.list(bw_cdf$nomad.shortcut))
  expect_true(isTRUE(bw_cd$bernstein.basis))
  expect_true(isTRUE(bw_cdf$bernstein.basis))
})

test_that("npcdens and npcdist nomad shortcuts fail fast on incompatible settings", {
  x <- data.frame(x = seq(0, 1, length.out = 8))
  y <- data.frame(y = seq(0, 1, length.out = 8))

  expect_error(
    np::npcdensbw(xdat = x, ydat = y, nomad = TRUE, regtype = "ll"),
    "nomad=TRUE requires regtype='lp'"
  )
  expect_error(
    np::npcdensbw(xdat = x, ydat = y, nomad = TRUE, bernstein.basis = FALSE),
    "requires bernstein.basis=TRUE"
  )
  expect_error(
    np::npcdistbw(xdat = x, ydat = y, nomad = TRUE, bwtype = "adaptive_nn"),
    "requires bwtype='fixed'"
  )
  expect_error(
    np::npcdistbw(xdat = x, ydat = y, nomad = TRUE, search.engine = "cell"),
    "requires search.engine='nomad' or 'nomad\\+powell'"
  )
  expect_error(
    np::npcdensbw(xdat = x, ydat = y, nomad.nmulti = 1L),
    "nomad.nmulti is only supported when regtype='lp', automatic degree search is active, and search.engine is 'nomad' or 'nomad\\+powell'"
  )
})

library(np)

test_that("npcdens exposes lp higher-order continuous gradients via hat operators", {
  npcdenshat <- getFromNamespace("npcdenshat", "np")

  set.seed(20260514)
  n <- 70
  x <- data.frame(x = seq(-1, 1, length.out = n))
  y <- data.frame(y = 1 + x$x + x$x^2 + 0.05 * sin(3 * x$x))
  ex <- data.frame(x = seq(-0.8, 0.8, length.out = 18))
  ey <- data.frame(y = seq(min(y$y) + 0.05, max(y$y) - 0.05, length.out = 18))
  rhs <- rep.int(1.0, n)

  for (basis in c("glp", "additive", "tensor")) {
    bw <- npcdensbw(
      xdat = x,
      ydat = y,
      regtype = "lp",
      basis = basis,
      degree = 2L,
      bws = c(0.35, 0.45),
      bandwidth.compute = FALSE
    )

    fit <- npcdens(
      bws = bw,
      txdat = x,
      tydat = y,
      exdat = ex,
      eydat = ey,
      gradients = TRUE,
      gradient.order = 2L
    )
    H2 <- npcdenshat(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey, s = 2L)

    expect_equal(as.vector(fit$congrad[, 1L]), as.vector(H2 %*% rhs),
                 tolerance = 1e-9, info = basis)
    expect_true(all(is.na(fit$congerr[, 1L])), info = basis)
    expect_equal(gradients(fit, gradient.order = 2L), fit$congrad,
                 tolerance = 0, info = basis)
    expect_error(gradients(fit, gradient.order = 1L),
                 "differs from the derivative order stored",
                 info = basis)
  }
})

test_that("npcdist exposes lp higher-order continuous gradients via hat operators", {
  npcdisthat <- getFromNamespace("npcdisthat", "np")

  set.seed(20260514)
  n <- 70
  x <- data.frame(x = seq(-1, 1, length.out = n))
  y <- data.frame(y = 1 + x$x + x$x^2 + 0.05 * sin(3 * x$x))
  ex <- data.frame(x = seq(-0.8, 0.8, length.out = 18))
  ey <- data.frame(y = seq(min(y$y) + 0.05, max(y$y) - 0.05, length.out = 18))
  rhs <- rep.int(1.0, n)

  for (basis in c("glp", "additive", "tensor")) {
    bw <- npcdistbw(
      xdat = x,
      ydat = y,
      regtype = "lp",
      basis = basis,
      degree = 2L,
      bws = c(0.35, 0.45),
      bandwidth.compute = FALSE
    )

    fit <- npcdist(
      bws = bw,
      txdat = x,
      tydat = y,
      exdat = ex,
      eydat = ey,
      gradients = TRUE,
      gradient.order = 2L
    )
    H2 <- npcdisthat(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey, s = 2L)

    expect_equal(as.vector(fit$congrad[, 1L]), as.vector(H2 %*% rhs),
                 tolerance = 1e-9, info = basis)
    expect_true(all(is.na(fit$congerr[, 1L])), info = basis)
    expect_equal(gradients(fit, gradient.order = 2L), fit$congrad,
                 tolerance = 0, info = basis)
    expect_error(gradients(fit, gradient.order = 1L),
                 "differs from the derivative order stored",
                 info = basis)
  }
})

test_that("conditional higher-order gradient requests validate scope", {
  set.seed(20260514)
  x <- data.frame(x = seq(0, 1, length.out = 24))
  y <- data.frame(y = x$x + rnorm(24, sd = 0.01))

  bw.ll <- npcdensbw(
    xdat = x,
    ydat = y,
    regtype = "ll",
    bws = c(0.3, 0.3),
    bandwidth.compute = FALSE
  )
  expect_warning(
    fit.ll <- npcdens(bws = bw.ll, txdat = x, tydat = y, gradients = TRUE,
                      gradient.order = 2L),
    "exceed polynomial degree"
  )
  expect_warning(
    expect_true(all(is.na(gradients(fit.ll, gradient.order = 2L)[, 1L]))),
    "exceed polynomial degree"
  )

  bw.lp1 <- npcdistbw(
    xdat = x,
    ydat = y,
    regtype = "lp",
    degree = 1L,
    bws = c(0.3, 0.3),
    bandwidth.compute = FALSE
  )
  expect_warning(
    fit <- npcdist(bws = bw.lp1, txdat = x, tydat = y, gradients = TRUE,
                   gradient.order = 2L),
    "exceed polynomial degree"
  )
  expect_warning(
    expect_true(all(is.na(gradients(fit, gradient.order = 2L)[, 1L]))),
    "exceed polynomial degree"
  )
})

test_that("formula routes keep gradient.order out of bandwidth selection", {
  set.seed(20260514)
  dat <- data.frame(
    x = seq(0, 1, length.out = 28),
    y = seq(0, 1, length.out = 28)^2 + rnorm(28, sd = 0.02)
  )

  fit.dens <- npcdens(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree = 2L,
    gradients = TRUE,
    gradient.order = 2L,
    nmulti = 1L
  )
  expect_equal(fit.dens$gradient.order, 2L)

  fit.dist <- npcdist(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree = 2L,
    gradients = TRUE,
    gradient.order = 2L,
    nmulti = 1L
  )
  expect_equal(fit.dist$gradient.order, 2L)
})

test_that("conditional plot data honors higher-order gradient requests", {
  set.seed(20260514)
  n <- 30
  x <- data.frame(x = seq(-1, 1, length.out = n))
  y <- data.frame(y = 1 + x$x + x$x^2)

  for (constructor in list(npcdensbw, npcdistbw)) {
    bw <- constructor(
      xdat = x,
      ydat = y,
      regtype = "lp",
      degree = 2L,
      basis = "tensor",
      bernstein.basis = FALSE,
      bws = c(10, 10),
      bandwidth.compute = FALSE
    )

    plot.out <- plot(
      bw,
      xdat = x,
      ydat = y,
      gradients = TRUE,
      gradient.order = 2L,
      xq = 0.5,
      yq = 0.5,
      neval = 8,
      perspective = FALSE,
      view = "fixed",
      plot.behavior = "data",
      plot.errors.method = "none"
    )

    slice <- plot.out[[1L]]
    eval.fun <- if (inherits(bw, "condbandwidth")) npcdist else npcdens
    fit <- eval.fun(
      bws = bw,
      txdat = x,
      tydat = y,
      exdat = slice$xeval,
      eydat = slice$yeval,
      gradients = TRUE,
      gradient.order = 2L
    )

    expect_equal(slice$gradient.order, 2L)
    expect_equal(as.vector(slice$congrad[, 1L]), as.vector(fit$congrad[, 1L]),
                 tolerance = 1e-10)
    expect_true(all(is.na(slice$congerr[, 1L])))
  }
})

library(np)

adaptive_shadow_empty <- function(n) {
  matrix(numeric(0), nrow = n, ncol = 0)
}

adaptive_shadow_cker <- function(kernel) {
  switch(kernel,
    gaussian = 0L,
    epanechnikov = 4L,
    uniform = 8L,
    truncated = 9L,
    stop("unsupported continuous kernel")
  )
}

adaptive_shadow_uker <- function(kernel) {
  switch(kernel,
    aitchisonaitken = 0L,
    liracine = 1L,
    stop("unsupported unordered kernel")
  )
}

adaptive_shadow_oker <- function(kernel) {
  switch(kernel,
    wangvanryzin = 0L,
    liracine = 2L,
    racineliyan = 3L,
    stop("unsupported ordered kernel")
  )
}

adaptive_shadow_rbw <- function(bw) {
  c(
    bw$xbw[bw$ixcon],
    bw$ybw[bw$iycon],
    bw$ybw[bw$iyuno],
    bw$ybw[bw$iyord],
    bw$xbw[bw$ixuno],
    bw$xbw[bw$ixord]
  )
}

adaptive_shadow_bwtype <- function(bw) {
  switch(bw$type,
    fixed = 0L,
    generalized_nn = 1L,
    adaptive_nn = 2L,
    stop("unsupported bandwidth type")
  )
}

adaptive_shadow_regtype <- function(bw) {
  if (identical(bw$regtype.engine, "lp")) 2L else 0L
}

adaptive_shadow_degree <- function(bw) {
  if (identical(bw$regtype.engine, "lp")) as.integer(bw$degree.engine) else integer(0)
}

adaptive_shadow_basis <- function(basis_engine, regtype_engine) {
  if (!identical(regtype_engine, "lp")) {
    return(0L)
  }
  switch(basis_engine,
    additive = 0L,
    glp = 1L,
    tensor = 2L,
    stop("unsupported LP basis")
  )
}

adaptive_shadow_safe_call <- function(name, ...) {
  on.exit(
    tryCatch(.Call("C_np_shadow_reset_state", PACKAGE = "np"),
             error = function(e) NULL),
    add = TRUE
  )
  .Call(name, ..., PACKAGE = "np")
}

call_adaptive_density_shadow <- function(bw, x, y, criterion = c("cv.ml", "cv.ls")) {
  criterion <- match.arg(criterion)
  n <- nrow(x)
  adaptive_shadow_safe_call(
    "C_np_shadow_cv_density_conditional",
    adaptive_shadow_empty(n), adaptive_shadow_empty(n), as.matrix(y),
    adaptive_shadow_empty(n), adaptive_shadow_empty(n), as.matrix(x),
    as.double(adaptive_shadow_rbw(bw)),
    adaptive_shadow_bwtype(bw),
    adaptive_shadow_cker(bw$cykertype),
    adaptive_shadow_uker(bw$uykertype),
    adaptive_shadow_oker(bw$oykertype),
    adaptive_shadow_cker(bw$cxkertype),
    adaptive_shadow_uker(bw$uxkertype),
    adaptive_shadow_oker(bw$oxkertype),
    FALSE,
    if (identical(criterion, "cv.ml")) 0L else 1L,
    adaptive_shadow_regtype(bw),
    adaptive_shadow_degree(bw),
    isTRUE(bw$bernstein.basis.engine),
    adaptive_shadow_basis(bw$basis.engine, bw$regtype.engine),
    identical(bw$regtype.engine, "lc")
  )
}

call_adaptive_distribution_shadow <- function(bw, x, ytrain, yeval = ytrain, cdfontrain = FALSE) {
  n <- nrow(x)
  ne <- nrow(yeval)
  adaptive_shadow_safe_call(
    "C_np_shadow_cv_distribution_conditional",
    adaptive_shadow_empty(n), adaptive_shadow_empty(n), as.matrix(ytrain),
    adaptive_shadow_empty(ne), adaptive_shadow_empty(ne), as.matrix(yeval),
    adaptive_shadow_empty(n), adaptive_shadow_empty(n), as.matrix(x),
    as.double(adaptive_shadow_rbw(bw)),
    adaptive_shadow_bwtype(bw),
    adaptive_shadow_cker(bw$cykertype),
    adaptive_shadow_uker(bw$uykertype),
    adaptive_shadow_oker(bw$oykertype),
    adaptive_shadow_cker(bw$cxkertype),
    adaptive_shadow_uker(bw$uxkertype),
    adaptive_shadow_oker(bw$oxkertype),
    FALSE,
    adaptive_shadow_regtype(bw),
    adaptive_shadow_degree(bw),
    isTRUE(bw$bernstein.basis.engine),
    adaptive_shadow_basis(bw$basis.engine, bw$regtype.engine),
    cdfontrain,
    identical(bw$regtype.engine, "lc")
  )
}

adaptive_cdist_eval_grid <- function(ydat, ngrid = 48L) {
  probs <- seq(0, 1, length.out = ngrid)
  evy <- ydat[rep(1L, ngrid), , drop = FALSE]
  for (i in seq_len(ncol(evy))) {
    evy[, i] <- uocquantile(ydat[, i], probs)
  }
  evy
}

adaptive_cv_fixture <- function(seed) {
  set.seed(seed)
  n <- 36L
  x <- data.frame(
    x1 = runif(n),
    x2 = runif(n)
  )
  y <- data.frame(
    y1 = sin(2 * pi * x$x1) + 0.5 * x$x2^2 + rnorm(n, sd = 0.08)
  )
  list(
    x = x,
    y = y,
    degree1 = rep.int(1L, ncol(x)),
    degree2 = rep.int(2L, ncol(x)),
    bws = c(5, 7, 6)
  )
}

make_adaptive_npcdens_bw <- function(x, y, regtype, bws, degree = NULL, bandwidth.compute = FALSE) {
  args <- list(
    xdat = x,
    ydat = y,
    bwtype = "adaptive_nn",
    bws = bws,
    bandwidth.compute = bandwidth.compute,
    regtype = regtype
  )
  if (identical(regtype, "lp")) {
    args$basis <- "glp"
    args$degree <- degree
  }
  do.call(npcdensbw, args)
}

make_adaptive_npcdist_bw <- function(x, y, regtype, bws, degree = NULL, bandwidth.compute = FALSE) {
  args <- list(
    xdat = x,
    ydat = y,
    bwtype = "adaptive_nn",
    bws = bws,
    bandwidth.compute = bandwidth.compute,
    regtype = regtype,
    itmax = 1L
  )
  if (identical(regtype, "lp")) {
    args$basis <- "glp"
    args$degree <- degree
  }
  do.call(npcdistbw, args)
}

test_that("adaptive conditional density shadow CV LS activates LP production route", {
  fixture <- adaptive_cv_fixture(4101)
  bw.lc <- make_adaptive_npcdens_bw(fixture$x, fixture$y, "lc", fixture$bws)
  bw.ll <- make_adaptive_npcdens_bw(fixture$x, fixture$y, "ll", fixture$bws)
  bw.lp1 <- make_adaptive_npcdens_bw(fixture$x, fixture$y, "lp", fixture$bws, degree = fixture$degree1)
  bw.lp2 <- make_adaptive_npcdens_bw(fixture$x, fixture$y, "lp", fixture$bws, degree = fixture$degree2)

  shadow.lc <- call_adaptive_density_shadow(bw.lc, fixture$x, fixture$y, criterion = "cv.ls")
  shadow.ll <- call_adaptive_density_shadow(bw.ll, fixture$x, fixture$y, criterion = "cv.ls")
  shadow.lp1 <- call_adaptive_density_shadow(bw.lp1, fixture$x, fixture$y, criterion = "cv.ls")
  shadow.lp2 <- call_adaptive_density_shadow(bw.lp2, fixture$x, fixture$y, criterion = "cv.ls")

  expect_true(is.finite(shadow.ll$new))
  expect_true(is.finite(shadow.lp1$new))
  expect_true(is.finite(shadow.lp2$new))
  expect_true(is.finite(shadow.lp1$prod))
  expect_equal(shadow.ll$new, shadow.lp1$new, tolerance = 1e-10)
  expect_equal(shadow.ll$prod, shadow.lp1$prod, tolerance = 1e-10)
  expect_gt(abs(shadow.lc$new - shadow.lp1$new), 1e-6)
  expect_gt(abs(shadow.lp2$new - shadow.lp1$new), 1e-6)
})

test_that("adaptive conditional density shadow CV ML activates LP production route", {
  fixture <- adaptive_cv_fixture(4102)
  bw.lc <- make_adaptive_npcdens_bw(fixture$x, fixture$y, "lc", fixture$bws)
  bw.ll <- make_adaptive_npcdens_bw(fixture$x, fixture$y, "ll", fixture$bws)
  bw.lp1 <- make_adaptive_npcdens_bw(fixture$x, fixture$y, "lp", fixture$bws, degree = fixture$degree1)
  bw.lp2 <- make_adaptive_npcdens_bw(fixture$x, fixture$y, "lp", fixture$bws, degree = fixture$degree2)

  shadow.lc <- call_adaptive_density_shadow(bw.lc, fixture$x, fixture$y, criterion = "cv.ml")
  shadow.ll <- call_adaptive_density_shadow(bw.ll, fixture$x, fixture$y, criterion = "cv.ml")
  shadow.lp1 <- call_adaptive_density_shadow(bw.lp1, fixture$x, fixture$y, criterion = "cv.ml")
  shadow.lp2 <- call_adaptive_density_shadow(bw.lp2, fixture$x, fixture$y, criterion = "cv.ml")

  expect_true(is.finite(shadow.ll$new))
  expect_true(is.finite(shadow.lp1$new))
  expect_true(is.finite(shadow.lp2$new))
  expect_true(is.finite(shadow.lp1$prod))
  expect_equal(shadow.ll$new, shadow.lp1$new, tolerance = 1e-10)
  expect_equal(shadow.ll$prod, shadow.lp1$prod, tolerance = 1e-10)
  expect_gt(abs(shadow.lc$new - shadow.lp1$new), 1e-6)
  expect_gt(abs(shadow.lp2$new - shadow.lp1$new), 1e-6)
})

test_that("adaptive conditional distribution shadow CV LS activates LP production route", {
  fixture <- adaptive_cv_fixture(4103)
  yeval <- adaptive_cdist_eval_grid(fixture$y)
  bw.lc <- make_adaptive_npcdist_bw(fixture$x, fixture$y, "lc", fixture$bws)
  bw.ll <- make_adaptive_npcdist_bw(fixture$x, fixture$y, "ll", fixture$bws)
  bw.lp1 <- make_adaptive_npcdist_bw(fixture$x, fixture$y, "lp", fixture$bws, degree = fixture$degree1)
  bw.lp2 <- make_adaptive_npcdist_bw(fixture$x, fixture$y, "lp", fixture$bws, degree = fixture$degree2)

  shadow.lc <- call_adaptive_distribution_shadow(bw.lc, fixture$x, fixture$y, yeval = yeval)
  shadow.ll <- call_adaptive_distribution_shadow(bw.ll, fixture$x, fixture$y, yeval = yeval)
  shadow.lp1 <- call_adaptive_distribution_shadow(bw.lp1, fixture$x, fixture$y, yeval = yeval)
  shadow.lp2 <- call_adaptive_distribution_shadow(bw.lp2, fixture$x, fixture$y, yeval = yeval)

  expect_true(is.finite(shadow.ll$new))
  expect_true(is.finite(shadow.lp1$new))
  expect_true(is.finite(shadow.lp2$new))
  expect_true(is.finite(shadow.lp1$prod))
  expect_equal(shadow.ll$new, shadow.lp1$new, tolerance = 1e-10)
  expect_equal(shadow.lp1$new, shadow.lp1$prod, tolerance = 1e-10)
  expect_gt(abs(shadow.lc$new - shadow.lp1$new), 1e-6)
  expect_gt(abs(shadow.lp2$new - shadow.lp1$new), 1e-6)
})

test_that("adaptive public conditional density CV LS separates lc from LP while preserving ll canonicalization", {
  fixture <- adaptive_cv_fixture(4201)
  bw.lc <- npcdensbw(
    xdat = fixture$x,
    ydat = fixture$y,
    regtype = "lc",
    bwtype = "adaptive_nn",
    bwmethod = "cv.ls",
    nmulti = 1
  )
  bw.ll <- npcdensbw(
    xdat = fixture$x,
    ydat = fixture$y,
    regtype = "ll",
    bwtype = "adaptive_nn",
    bwmethod = "cv.ls",
    nmulti = 1
  )
  bw.lp1 <- npcdensbw(
    xdat = fixture$x,
    ydat = fixture$y,
    regtype = "lp",
    basis = "glp",
    degree = fixture$degree1,
    bwtype = "adaptive_nn",
    bwmethod = "cv.ls",
    nmulti = 1
  )
  bw.lp2 <- npcdensbw(
    xdat = fixture$x,
    ydat = fixture$y,
    regtype = "lp",
    basis = "glp",
    degree = fixture$degree2,
    bwtype = "adaptive_nn",
    bwmethod = "cv.ls",
    nmulti = 1
  )

  expect_equal(bw.ll$fval, bw.lp1$fval, tolerance = 1e-8)
  expect_gt(abs(bw.lc$fval - bw.lp1$fval), 1e-6)
  expect_gt(abs(bw.lp2$fval - bw.lp1$fval), 1e-6)
})

test_that("adaptive public conditional density CV ML separates lc from LP while preserving ll canonicalization", {
  fixture <- adaptive_cv_fixture(4202)
  bw.lc <- npcdensbw(
    xdat = fixture$x,
    ydat = fixture$y,
    regtype = "lc",
    bwtype = "adaptive_nn",
    bwmethod = "cv.ml",
    nmulti = 1
  )
  bw.ll <- npcdensbw(
    xdat = fixture$x,
    ydat = fixture$y,
    regtype = "ll",
    bwtype = "adaptive_nn",
    bwmethod = "cv.ml",
    nmulti = 1
  )
  bw.lp1 <- npcdensbw(
    xdat = fixture$x,
    ydat = fixture$y,
    regtype = "lp",
    basis = "glp",
    degree = fixture$degree1,
    bwtype = "adaptive_nn",
    bwmethod = "cv.ml",
    nmulti = 1
  )
  bw.lp2 <- npcdensbw(
    xdat = fixture$x,
    ydat = fixture$y,
    regtype = "lp",
    basis = "glp",
    degree = fixture$degree2,
    bwtype = "adaptive_nn",
    bwmethod = "cv.ml",
    nmulti = 1
  )

  expect_equal(bw.ll$fval, bw.lp1$fval, tolerance = 1e-8)
  expect_gt(abs(bw.lc$fval - bw.lp1$fval), 1e-6)
  expect_gt(abs(bw.lp2$fval - bw.lp1$fval), 1e-6)
})

test_that("adaptive public conditional distribution CV LS separates lc from LP while preserving ll canonicalization", {
  fixture <- adaptive_cv_fixture(4203)
  bw.lc <- npcdistbw(
    xdat = fixture$x,
    ydat = fixture$y,
    regtype = "lc",
    bwtype = "adaptive_nn",
    bwmethod = "cv.ls",
    nmulti = 1,
    itmax = 1L
  )
  bw.ll <- npcdistbw(
    xdat = fixture$x,
    ydat = fixture$y,
    regtype = "ll",
    bwtype = "adaptive_nn",
    bwmethod = "cv.ls",
    nmulti = 1,
    itmax = 1L
  )
  bw.lp1 <- npcdistbw(
    xdat = fixture$x,
    ydat = fixture$y,
    regtype = "lp",
    basis = "glp",
    degree = fixture$degree1,
    bwtype = "adaptive_nn",
    bwmethod = "cv.ls",
    nmulti = 1,
    itmax = 1L
  )
  bw.lp2 <- npcdistbw(
    xdat = fixture$x,
    ydat = fixture$y,
    regtype = "lp",
    basis = "glp",
    degree = fixture$degree2,
    bwtype = "adaptive_nn",
    bwmethod = "cv.ls",
    nmulti = 1,
    itmax = 1L
  )

  expect_equal(bw.ll$fval, bw.lp1$fval, tolerance = 1e-8)
  expect_gt(abs(bw.lc$fval - bw.lp1$fval), 1e-6)
  expect_gt(abs(bw.lp2$fval - bw.lp1$fval), 1e-6)
})

library(np)

test_that("core formula estimators reject malformed newdata before parent-frame lookup", {
  set.seed(20260513)
  dat <- data.frame(x = runif(45), y = runif(45))
  bad <- data.frame(z = dat$x[1:3])

  bw.reg <- npregbw(y ~ x, data = dat, bws = 0.25, bandwidth.compute = FALSE)
  fit.reg <- npreg(bws = bw.reg, gradients = TRUE)
  expect_error(npreg(bws = bw.reg, newdata = bad), "newdata must contain columns")
  expect_error(predict(fit.reg, newdata = bad), "newdata must contain columns")

  bw.den <- npudensbw(~ x, data = dat, bws = 0.25, bandwidth.compute = FALSE)
  fit.den <- npudens(bws = bw.den)
  expect_error(npudens(bws = bw.den, newdata = bad), "newdata must contain columns")
  expect_error(predict(fit.den, newdata = bad), "newdata must contain columns")

  bw.dist <- npudistbw(~ x, data = dat, bws = 0.25, bandwidth.compute = FALSE)
  fit.dist <- npudist(bws = bw.dist)
  expect_error(npudist(bws = bw.dist, newdata = bad), "newdata must contain columns")
  expect_error(predict(fit.dist, newdata = bad), "newdata must contain columns")

  bw.cd <- npcdensbw(y ~ x, data = dat, bws = c(0.25, 0.25), bandwidth.compute = FALSE)
  fit.cd <- npcdens(bws = bw.cd, gradients = TRUE)
  expect_error(npcdens(bws = bw.cd, newdata = bad), "newdata must contain columns")
  expect_error(predict(fit.cd, newdata = bad), "newdata must contain columns")

  bw.cdist <- npcdistbw(y ~ x, data = dat, bws = c(0.25, 0.25), bandwidth.compute = FALSE)
  fit.cdist <- npcdist(bws = bw.cdist, gradients = TRUE)
  expect_error(npcdist(bws = bw.cdist, newdata = bad), "newdata must contain columns")
  expect_error(predict(fit.cdist, newdata = bad), "newdata must contain columns")
})

test_that("core predict methods retain native evaluation argument precedence", {
  set.seed(20260513)
  dat <- data.frame(x = runif(45), y = runif(45))
  nd.standard <- data.frame(x = c(0.2, 0.5, 0.8))
  nd.native <- data.frame(x = c(0.15, 0.35, 0.65))
  cd.standard <- data.frame(y = c(0.2, 0.4, 0.7), x = c(0.2, 0.5, 0.8))
  cd.native <- data.frame(y = c(0.1, 0.3, 0.6), x = c(0.15, 0.35, 0.65))

  bw.reg <- npregbw(y ~ x, data = dat, bws = 0.25, bandwidth.compute = FALSE)
  fit.reg <- npreg(bws = bw.reg)
  expect_equal(
    as.numeric(predict(fit.reg, newdata = nd.standard, exdat = nd.native)),
    as.numeric(predict(fit.reg, exdat = nd.native)),
    tolerance = 1e-12
  )

  bw.den <- npudensbw(~ x, data = dat, bws = 0.25, bandwidth.compute = FALSE)
  fit.den <- npudens(bws = bw.den)
  expect_equal(
    as.numeric(predict(fit.den, newdata = nd.standard, edat = nd.native)),
    as.numeric(predict(fit.den, edat = nd.native)),
    tolerance = 1e-12
  )

  bw.dist <- npudistbw(~ x, data = dat, bws = 0.25, bandwidth.compute = FALSE)
  fit.dist <- npudist(bws = bw.dist)
  expect_equal(
    as.numeric(predict(fit.dist, newdata = nd.standard, edat = nd.native)),
    as.numeric(predict(fit.dist, edat = nd.native)),
    tolerance = 1e-12
  )

  bw.cd <- npcdensbw(y ~ x, data = dat, bws = c(0.25, 0.25), bandwidth.compute = FALSE)
  fit.cd <- npcdens(bws = bw.cd)
  expect_equal(
    as.numeric(predict(fit.cd, newdata = cd.standard, eydat = cd.native["y"], exdat = cd.native["x"])),
    as.numeric(predict(fit.cd, eydat = cd.native["y"], exdat = cd.native["x"])),
    tolerance = 1e-12
  )

  bw.cdist <- npcdistbw(y ~ x, data = dat, bws = c(0.25, 0.25), bandwidth.compute = FALSE)
  fit.cdist <- npcdist(bws = bw.cdist)
  expect_equal(
    as.numeric(predict(fit.cdist, newdata = cd.standard, eydat = cd.native["y"], exdat = cd.native["x"])),
    as.numeric(predict(fit.cdist, eydat = cd.native["y"], exdat = cd.native["x"])),
    tolerance = 1e-12
  )
})

test_that("core S3 logical arguments fail clearly", {
  set.seed(20260513)
  dat <- data.frame(x = runif(45), y = runif(45))

  bw.reg <- npregbw(y ~ x, data = dat, bws = 0.25, bandwidth.compute = FALSE)
  fit.reg <- npreg(bws = bw.reg, gradients = TRUE)
  expect_error(predict(fit.reg, se.fit = "yes"), "'se.fit' must be TRUE or FALSE", fixed = TRUE)
  expect_error(gradients(fit.reg, errors = "yes"), "'errors' must be TRUE or FALSE", fixed = TRUE)

  bw.den <- npudensbw(~ x, data = dat, bws = 0.25, bandwidth.compute = FALSE)
  fit.den <- npudens(bws = bw.den)
  expect_error(predict(fit.den, se.fit = "yes"), "'se.fit' must be TRUE or FALSE", fixed = TRUE)

  bw.dist <- npudistbw(~ x, data = dat, bws = 0.25, bandwidth.compute = FALSE)
  fit.dist <- npudist(bws = bw.dist)
  expect_error(predict(fit.dist, se.fit = "yes"), "'se.fit' must be TRUE or FALSE", fixed = TRUE)

  bw.cd <- npcdensbw(y ~ x, data = dat, bws = c(0.25, 0.25), bandwidth.compute = FALSE)
  fit.cd <- npcdens(bws = bw.cd, gradients = TRUE)
  expect_error(predict(fit.cd, se.fit = "yes"), "'se.fit' must be TRUE or FALSE", fixed = TRUE)
  expect_error(gradients(fit.cd, errors = "yes"), "'errors' must be TRUE or FALSE", fixed = TRUE)

  bw.cdist <- npcdistbw(y ~ x, data = dat, bws = c(0.25, 0.25), bandwidth.compute = FALSE)
  fit.cdist <- npcdist(bws = bw.cdist, gradients = TRUE)
  expect_error(predict(fit.cdist, se.fit = "yes"), "'se.fit' must be TRUE or FALSE", fixed = TRUE)
  expect_error(gradients(fit.cdist, errors = "yes"), "'errors' must be TRUE or FALSE", fixed = TRUE)
})

library(np)

quiet_eval <- function(expr) {
  value <- NULL
  capture.output(value <- force(expr))
  value
}

expect_bw_timing_contract <- function(obj, summary_label = "Estimation Time:") {
  expect_true(is.finite(obj$total.time))
  txt <- paste(capture.output(summary(obj)), collapse = "\n")
  expect_match(txt, summary_label, fixed = TRUE)
  if (!is.null(obj$nomad.time) && is.finite(obj$nomad.time))
    expect_match(txt, "NOMAD ", fixed = TRUE)
  if (!is.null(obj$powell.time) && is.finite(obj$powell.time))
    expect_match(txt, "Powell ", fixed = TRUE)
}

expect_fit_timing_contract <- function(obj, summary_label = "Estimation Time:") {
  expect_true(is.finite(obj$total.time))
  expect_true(is.finite(obj$fit.time))
  expect_true(is.finite(obj$optim.time))
  expect_equal(obj$total.time, obj$optim.time + obj$fit.time, tolerance = 1e-8)

  txt <- paste(capture.output(summary(obj)), collapse = "\n")
  expect_match(txt, summary_label, fixed = TRUE)
  if (!is.null(obj$nomad.time) && is.finite(obj$nomad.time)) {
    expect_match(txt, "NOMAD ", fixed = TRUE)
    if (!is.null(obj$powell.time) && is.finite(obj$powell.time))
      expect_match(txt, "Powell ", fixed = TRUE)
    expect_match(txt, "fit ", fixed = TRUE)
  } else {
    expect_match(txt, "optim ", fixed = TRUE)
    expect_match(txt, "fit ", fixed = TRUE)
  }
}

test_that("direct density and distribution families report elapsed optimization time", {
  set.seed(42)
  n <- 120L
  x <- runif(n)
  y <- x^2 + rnorm(n, sd = 0.1)

  bw_udens <- quiet_eval(npudensbw(~x, nmulti = 1))
  fit_udens <- quiet_eval(npudens(bws = bw_udens))

  bw_udist <- quiet_eval(npudistbw(~x, nmulti = 1))
  fit_udist <- quiet_eval(npudist(bws = bw_udist))

  bw_cdens <- quiet_eval(npcdensbw(y ~ x, regtype = "ll", bwtype = "adaptive_nn", nmulti = 1))
  fit_cdens <- quiet_eval(npcdens(bws = bw_cdens, txdat = data.frame(x = x), tydat = data.frame(y = y)))

  bw_cdist <- quiet_eval(npcdistbw(y ~ x, regtype = "ll", bwtype = "adaptive_nn", nmulti = 1))
  fit_cdist <- quiet_eval(npcdist(bws = bw_cdist, txdat = data.frame(x = x), tydat = data.frame(y = y)))

  for (obj in list(bw_udens, bw_udist, bw_cdens, bw_cdist))
    expect_bw_timing_contract(obj)

  for (obj in list(fit_udens, fit_udist, fit_cdens, fit_cdist))
    expect_fit_timing_contract(obj)

  expect_gt(fit_cdens$optim.time, 0)
  expect_gt(fit_cdist$optim.time, 0)
})

test_that("regression and quantile regression remain timing-contract compliant", {
  set.seed(99)
  n <- 100L
  x <- runif(n)
  y <- x^2 + rnorm(n, sd = 0.15)

  bw_reg <- quiet_eval(npregbw(xdat = data.frame(x = x), ydat = y, regtype = "ll", bwtype = "adaptive_nn", nmulti = 1))
  fit_reg <- quiet_eval(npreg(bws = bw_reg, txdat = data.frame(x = x), tydat = y))

  bw_qreg <- quiet_eval(npcdistbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    regtype = "ll",
    bwtype = "adaptive_nn",
    nmulti = 1
  ))
  fit_qreg <- quiet_eval(npqreg(
    bws = bw_qreg,
    tau = 0.5
  ))

  expect_bw_timing_contract(bw_reg)
  expect_bw_timing_contract(bw_qreg)
  expect_fit_timing_contract(fit_reg)
  expect_fit_timing_contract(fit_qreg)
})

test_that("composite core families remain timing-contract compliant", {
  set.seed(123)

  n <- 50L
  x <- runif(n)
  z <- runif(n)
  y <- x^2 + 2 * z + rnorm(n, sd = 0.1)
  bw_pl <- quiet_eval(npplregbw(xdat = z, zdat = x, ydat = y, regtype = "ll", nmulti = 1))
  fit_pl <- quiet_eval(npplreg(bws = bw_pl))

  x1 <- runif(40L)
  x2 <- runif(40L)
  y_si <- (x1 + x2)^2 + rnorm(40L, sd = 0.1)
  bw_si <- quiet_eval(npindexbw(xdat = data.frame(x1 = x1, x2 = x2), ydat = y_si, method = "ichimura", nmulti = 1))
  fit_si <- quiet_eval(npindex(bws = bw_si))

  xs <- runif(n)
  zs <- runif(n)
  y_sc <- (0.5 + xs^2) * zs + rnorm(n, sd = 0.1)
  bw_sc <- quiet_eval(npscoefbw(xdat = xs, zdat = zs, ydat = y_sc, regtype = "ll", nmulti = 1))
  fit_sc <- quiet_eval(npscoef(bws = bw_sc))

  for (obj in list(bw_pl, bw_si, bw_sc))
    expect_bw_timing_contract(obj)

  for (obj in list(fit_pl, fit_si, fit_sc))
    expect_fit_timing_contract(obj)
})

test_that("NOMAD plus Powell timing is carried through bandwidth and fit summaries", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(x = sort(runif(24)))
  dat$y <- sin(2 * pi * dat$x) + rnorm(nrow(dat), sd = 0.05)

  bw <- quiet_eval(npregbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 2L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  ))
  fit <- quiet_eval(npreg(bws = bw, txdat = data.frame(x = dat$x), tydat = dat$y))

  expect_identical(bw$degree.search$mode, "nomad+powell")
  expect_true(is.finite(bw$nomad.time))
  expect_true(is.finite(bw$powell.time))
  expect_true(is.finite(bw$total.time))
  expect_gte(bw$total.time + 1e-8, bw$nomad.time + bw$powell.time)
  expect_equal(fit$optim.time, bw$total.time, tolerance = 1e-8)
  expect_equal(fit$nomad.time, bw$nomad.time, tolerance = 1e-8)
  expect_equal(fit$powell.time, bw$powell.time, tolerance = 1e-8)

  bw_txt <- paste(capture.output(summary(bw)), collapse = "\n")
  fit_txt <- paste(capture.output(summary(fit)), collapse = "\n")
  expect_match(bw_txt, "NOMAD ", fixed = TRUE)
  expect_match(bw_txt, "Powell ", fixed = TRUE)
  expect_match(fit_txt, "NOMAD ", fixed = TRUE)
  expect_match(fit_txt, "Powell ", fixed = TRUE)
  expect_match(fit_txt, "fit ", fixed = TRUE)
})

quiet_capture <- function(expr) {
  out <- NULL
  sinkfile <- tempfile()
  on.exit(unlink(sinkfile), add = TRUE)
  invisible(capture.output(out <- eval.parent(substitute(expr)), file = sinkfile))
  out
}

flatten_plot_mean <- function(plot_obj) {
  vals <- unlist(lapply(plot_obj, function(comp) {
    keys <- c("mean", "dens", "dist", "condens", "condist")
    hit <- keys[keys %in% names(comp)]
    if (!length(hit))
      return(numeric(0))
    as.double(comp[[hit[1L]]])
  }), use.names = FALSE)
  vals[is.finite(vals)]
}

run_bootstrap_plot <- function(bw,
                               xdat = NULL,
                               ydat = NULL,
                               zdat = NULL,
                               boot_num = 9L,
                               boot_method = "wild") {
  args <- list(
    bw,
    output = "data",
    perspective = FALSE,
    errors = "bootstrap",
    bootstrap = boot_method,
    B = boot_num
  )
  if (!is.null(xdat))
    args$xdat <- xdat
  if (!is.null(ydat))
    args$ydat <- ydat
  if (!is.null(zdat))
    args$zdat <- zdat

  quiet_capture(suppressWarnings(do.call(plot, args)))
}

test_that("rbandwidth plot helper uses method-defining kernel options", {
  set.seed(9301)
  n <- 50L
  x <- runif(n)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.1)
  xdat <- data.frame(x = x)

  bw.gaussian <- npregbw(
    y ~ x,
    data = data.frame(y = y, x = x),
    bws = 0.2,
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "gaussian"
  )

  bw.epan <- npregbw(
    y ~ x,
    data = data.frame(y = y, x = x),
    bws = 0.2,
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerorder = 2L
  )

  set.seed(9302)
  out.gaussian <- run_bootstrap_plot(bw.gaussian, xdat = xdat, ydat = y, boot_num = 9L)
  set.seed(9302)
  out.epan <- run_bootstrap_plot(bw.epan, xdat = xdat, ydat = y, boot_num = 9L)

  mean.gaussian <- flatten_plot_mean(out.gaussian)
  mean.epan <- flatten_plot_mean(out.epan)

  expect_gt(length(mean.gaussian), 0L)
  expect_equal(length(mean.gaussian), length(mean.epan))
  expect_gt(max(abs(mean.gaussian - mean.epan)), 1e-6)
})

test_that("sibandwidth plot helper uses method-defining kernel options", {
  set.seed(9303)
  n <- 60L
  xdat <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- xdat$x1^2 + 0.5 * xdat$x2 + rnorm(n, sd = 0.1)

  bw.gaussian <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = c(0.2, 0.2, 0.25),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "gaussian"
  )

  bw.epan <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = c(0.2, 0.2, 0.25),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerorder = 2L
  )

  set.seed(9304)
  out.gaussian <- run_bootstrap_plot(bw.gaussian, xdat = xdat, ydat = y, boot_num = 9L)
  set.seed(9304)
  out.epan <- run_bootstrap_plot(bw.epan, xdat = xdat, ydat = y, boot_num = 9L)

  mean.gaussian <- flatten_plot_mean(out.gaussian)
  mean.epan <- flatten_plot_mean(out.epan)

  expect_gt(length(mean.gaussian), 0L)
  expect_equal(length(mean.gaussian), length(mean.epan))
  expect_gt(max(abs(mean.gaussian - mean.epan)), 1e-6)
})

test_that("scbandwidth plot helper uses method-defining kernel options", {
  set.seed(9305)
  n <- 55L
  xdat <- data.frame(x = runif(n))
  zdat <- data.frame(z = runif(n))
  y <- sin(2 * pi * xdat$x) * zdat$z + rnorm(n, sd = 0.1)

  bw.gaussian <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    bws = 0.2,
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "gaussian"
  )

  bw.epan <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    bws = 0.2,
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerorder = 2L
  )

  set.seed(9306)
  out.gaussian <- run_bootstrap_plot(bw.gaussian, xdat = xdat, ydat = y, zdat = zdat, boot_num = 9L)
  set.seed(9306)
  out.epan <- run_bootstrap_plot(bw.epan, xdat = xdat, ydat = y, zdat = zdat, boot_num = 9L)

  mean.gaussian <- flatten_plot_mean(out.gaussian)
  mean.epan <- flatten_plot_mean(out.epan)

  expect_gt(length(mean.gaussian), 0L)
  expect_equal(length(mean.gaussian), length(mean.epan))
  expect_gt(max(abs(mean.gaussian - mean.epan)), 1e-6)
})

test_that("plbandwidth plot helper uses method-defining kernel options", {
  set.seed(9307)
  n <- 40L
  xdat <- data.frame(x = runif(n))
  zdat <- data.frame(z = runif(n))
  y <- 2 * xdat$x + sin(2 * pi * zdat$z) + rnorm(n, sd = 0.1)

  bw.gaussian <- quiet_capture(npplregbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    nmulti = 1L,
    regtype = "lp",
    basis = "glp",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "gaussian"
  ))

  bw.epan <- quiet_capture(npplregbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    nmulti = 1L,
    regtype = "lp",
    basis = "glp",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerorder = 2L
  ))

  set.seed(9308)
  out.gaussian <- run_bootstrap_plot(bw.gaussian, xdat = xdat, ydat = y, zdat = zdat, boot_num = 7L)
  set.seed(9308)
  out.epan <- run_bootstrap_plot(bw.epan, xdat = xdat, ydat = y, zdat = zdat, boot_num = 7L)

  mean.gaussian <- flatten_plot_mean(out.gaussian)
  mean.epan <- flatten_plot_mean(out.epan)

  expect_gt(length(mean.gaussian), 0L)
  expect_equal(length(mean.gaussian), length(mean.epan))
  expect_gt(max(abs(mean.gaussian - mean.epan)), 1e-6)
})

test_that("bandwidth plot helper uses method-defining kernel options", {
  set.seed(9309)
  n <- 70L
  xdat <- data.frame(x = runif(n))

  bw.gaussian <- npudensbw(
    dat = xdat,
    bws = 0.25,
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    ckertype = "gaussian"
  )

  bw.epan <- npudensbw(
    dat = xdat,
    bws = 0.25,
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerorder = 2L
  )

  set.seed(9310)
  out.gaussian <- run_bootstrap_plot(
    bw.gaussian,
    xdat = xdat,
    boot_num = 9L,
    boot_method = "inid"
  )
  set.seed(9310)
  out.epan <- run_bootstrap_plot(
    bw.epan,
    xdat = xdat,
    boot_num = 9L,
    boot_method = "inid"
  )

  mean.gaussian <- flatten_plot_mean(out.gaussian)
  mean.epan <- flatten_plot_mean(out.epan)

  expect_gt(length(mean.gaussian), 0L)
  expect_equal(length(mean.gaussian), length(mean.epan))
  expect_gt(max(abs(mean.gaussian - mean.epan)), 1e-6)
})

test_that("dbandwidth plot helper uses method-defining kernel options", {
  set.seed(9311)
  n <- 70L
  xdat <- data.frame(x = runif(n))

  bw.gaussian <- npudistbw(
    dat = xdat,
    bws = 0.25,
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    ckertype = "gaussian"
  )

  bw.epan <- npudistbw(
    dat = xdat,
    bws = 0.25,
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerorder = 2L
  )

  set.seed(9312)
  out.gaussian <- run_bootstrap_plot(
    bw.gaussian,
    xdat = xdat,
    boot_num = 9L,
    boot_method = "inid"
  )
  set.seed(9312)
  out.epan <- run_bootstrap_plot(
    bw.epan,
    xdat = xdat,
    boot_num = 9L,
    boot_method = "inid"
  )

  mean.gaussian <- flatten_plot_mean(out.gaussian)
  mean.epan <- flatten_plot_mean(out.epan)

  expect_gt(length(mean.gaussian), 0L)
  expect_equal(length(mean.gaussian), length(mean.epan))
  expect_gt(max(abs(mean.gaussian - mean.epan)), 1e-6)
})

test_that("conbandwidth plot helper uses method-defining x/y kernel options", {
  set.seed(9313)
  n <- 64L
  xdat <- data.frame(x = runif(n))
  ydat <- data.frame(y = runif(n))

  bw.gaussian <- npcdensbw(
    xdat = xdat,
    ydat = ydat,
    bws = c(0.22, 0.22),
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    cxkertype = "gaussian",
    cykertype = "gaussian"
  )

  bw.epan <- npcdensbw(
    xdat = xdat,
    ydat = ydat,
    bws = c(0.22, 0.22),
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    cxkertype = "epanechnikov",
    cykertype = "epanechnikov",
    cxkerorder = 2L,
    cykerorder = 2L
  )

  set.seed(9314)
  out.gaussian <- run_bootstrap_plot(
    bw.gaussian,
    xdat = xdat,
    ydat = ydat,
    boot_num = 9L,
    boot_method = "inid"
  )
  set.seed(9314)
  out.epan <- run_bootstrap_plot(
    bw.epan,
    xdat = xdat,
    ydat = ydat,
    boot_num = 9L,
    boot_method = "inid"
  )

  mean.gaussian <- flatten_plot_mean(out.gaussian)
  mean.epan <- flatten_plot_mean(out.epan)

  expect_gt(length(mean.gaussian), 0L)
  expect_equal(length(mean.gaussian), length(mean.epan))
  expect_gt(max(abs(mean.gaussian - mean.epan)), 1e-6)
})

test_that("condbandwidth plot helper uses method-defining x/y kernel options", {
  set.seed(9315)
  n <- 64L
  xdat <- data.frame(x = runif(n))
  ydat <- data.frame(y = runif(n))

  bw.gaussian <- npcdistbw(
    xdat = xdat,
    ydat = ydat,
    bws = c(0.22, 0.22),
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    cxkertype = "gaussian",
    cykertype = "gaussian"
  )

  bw.epan <- npcdistbw(
    xdat = xdat,
    ydat = ydat,
    bws = c(0.22, 0.22),
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    cxkertype = "epanechnikov",
    cykertype = "epanechnikov",
    cxkerorder = 2L,
    cykerorder = 2L
  )

  set.seed(9316)
  out.gaussian <- run_bootstrap_plot(
    bw.gaussian,
    xdat = xdat,
    ydat = ydat,
    boot_num = 9L,
    boot_method = "inid"
  )
  set.seed(9316)
  out.epan <- run_bootstrap_plot(
    bw.epan,
    xdat = xdat,
    ydat = ydat,
    boot_num = 9L,
    boot_method = "inid"
  )

  mean.gaussian <- flatten_plot_mean(out.gaussian)
  mean.epan <- flatten_plot_mean(out.epan)

  expect_gt(length(mean.gaussian), 0L)
  expect_equal(length(mean.gaussian), length(mean.epan))
  expect_gt(max(abs(mean.gaussian - mean.epan)), 1e-6)
})

test_that("conditional categorical gradient bootstrap helper matches explicit local refits", {
  skip_if_not_installed("np")

  library(np)

  helper <- getFromNamespace(".np_inid_boot_from_conditional_gradient_local", "np")
  eval_fun <- getFromNamespace(".np_plot_conditional_eval", "np")

  set.seed(20260312)
  n <- 24L
  x1 <- factor(sample(c("a", "b"), n, replace = TRUE))
  x2 <- rnorm(n)
  y <- rnorm(n)
  xdat <- data.frame(x1 = x1, x2 = x2)
  ydat <- data.frame(y = y)

  fixtures <- list(
    list(
      label = "dens",
      cdf = FALSE,
      bw = suppressWarnings(npcdensbw(xdat = xdat, ydat = ydat, nmulti = 1L))
    ),
    list(
      label = "dist",
      cdf = TRUE,
      bw = suppressWarnings(npcdistbw(xdat = xdat, ydat = ydat, nmulti = 1L))
    )
  )

  counts <- cbind(
    rep(1L, n),
    c(rep(2L, 4), rep(0L, 4), rep(1L, n - 8)),
    c(rep(0L, 3), rep(3L, 3), rep(1L, n - 6))
  )
  if (!is.matrix(counts))
    counts <- matrix(counts, nrow = n)

  for (fixture in fixtures) {
    slice <- suppressWarnings(plot(
      fixture$bw,
      xdat = xdat,
      ydat = ydat,
      gradients = TRUE,
      perspective = FALSE,
      output = "data"
    ))[[1L]]

    boot <- helper(
      xdat = xdat,
      ydat = ydat,
      exdat = slice$xeval,
      eydat = slice$yeval,
      bws = fixture$bw,
      B = ncol(counts),
      cdf = fixture$cdf,
      gradient.index = 1L,
      counts = counts
    )

    explicit_fit <- function(idx) {
      as.vector(eval_fun(
        bws = fixture$bw,
        xdat = xdat[idx, , drop = FALSE],
        ydat = ydat[idx, , drop = FALSE],
        exdat = slice$xeval,
        eydat = slice$yeval,
        cdf = fixture$cdf,
        gradients = TRUE
      )$congrad[, 1L])
    }

    oracle.t0 <- explicit_fit(seq_len(n))
    oracle.t <- vapply(
      seq_len(ncol(counts)),
      function(j) explicit_fit(rep.int(seq_len(n), counts[, j])),
      numeric(length(oracle.t0))
    )

    expect_equal(as.vector(boot$t0), oracle.t0, tolerance = 1e-8, info = fixture$label)
    expect_equal(boot$t, t(oracle.t), tolerance = 1e-8, info = fixture$label)
  }
})

test_that("conditional density/distribution categorical bootstrap gradients work again", {
  skip_if_not_installed("np")

  library(np)

  set.seed(20260312)
  n <- 24L
  x1 <- factor(sample(c("a", "b"), n, replace = TRUE))
  x2 <- rnorm(n)
  y <- rnorm(n)
  xdat <- data.frame(x1 = x1, x2 = x2)
  ydat <- data.frame(y = y)

  cd.bw <- suppressWarnings(npcdensbw(xdat = xdat, ydat = ydat, nmulti = 1L))
  cdist.bw <- suppressWarnings(npcdistbw(xdat = xdat, ydat = ydat, nmulti = 1L))
  cases <- list(
    list(label = "npcdens-bw", obj = cd.bw, args = list(xdat = xdat, ydat = ydat)),
    list(label = "npcdens-fit", obj = npcdens(bws = cd.bw), args = list()),
    list(label = "npcdist-bw", obj = cdist.bw, args = list(xdat = xdat, ydat = ydat)),
    list(label = "npcdist-fit", obj = npcdist(bws = cdist.bw), args = list())
  )

  for (case in cases) {
    for (boot.method in c("inid", "fixed", "geom")) {
      plot.args <- c(
        list(
          case$obj,
          perspective = FALSE,
          output = "data",
          gradients = TRUE,
          errors = "bootstrap",
          bootstrap = boot.method,
          B = 7L
        ),
        case$args
      )
      out <- suppressWarnings(do.call(plot, plot.args))
      expect_true(is.list(out), info = paste(case$label, boot.method))
      expect_true(length(out) >= 1L, info = paste(case$label, boot.method))
      expect_true(length(out[[1L]]$bxp) > 0L, info = paste(case$label, boot.method))
      expect_equal(length(out[[1L]]$bxp$names), 2L, info = paste(case$label, boot.method))
    }
  }
})

test_that("named bws formula dispatch matches positional density/distribution routes", {
  set.seed(20260323)
  d <- data.frame(
    x = seq(0.1, 0.9, length.out = 9),
    y = seq(0.2, 1.0, length.out = 9)
  )
  nd.u <- data.frame(x = c(0.2, 0.5, 0.8))
  nd.c <- data.frame(x = c(0.2, 0.5, 0.8), y = c(0.25, 0.55, 0.85))

  bw.ud <- npudensbw(~ x, data = d, newdata = nd.u)
  ud.pos <- npudens(bws = bw.ud, newdata = nd.u)
  ud.named <- npudens(bws = ~ x, data = d, newdata = nd.u)
  expect_equal(as.numeric(ud.named$dens), as.numeric(ud.pos$dens), tolerance = 0)
  expect_equal(as.numeric(ud.named$derr), as.numeric(ud.pos$derr), tolerance = 0)

  bw.udist <- npudistbw(~ x, data = d, newdata = nd.u)
  udist.pos <- npudist(bws = bw.udist, newdata = nd.u)
  udist.named <- npudist(bws = ~ x, data = d, newdata = nd.u)
  expect_equal(as.numeric(udist.named$dist), as.numeric(udist.pos$dist), tolerance = 0)
  expect_equal(as.numeric(udist.named$derr), as.numeric(udist.pos$derr), tolerance = 0)

  bw.cd <- npcdensbw(y ~ x, data = d, nmulti = 1, tol = 1, ftol = 1)
  cd.pos <- npcdens(bws = bw.cd, data = d, newdata = nd.c)
  cd.named <- npcdens(bws = y ~ x, data = d, newdata = nd.c, nmulti = 1, tol = 1, ftol = 1)
  expect_lt(max(abs(as.numeric(cd.named$condens) - as.numeric(cd.pos$condens))), 1e-2)
  expect_lt(max(abs(as.numeric(cd.named$conderr) - as.numeric(cd.pos$conderr))), 1e-2)

  bw.cdist <- npcdistbw(y ~ x, data = d, nmulti = 1, tol = 1, ftol = 1)
  cdist.pos <- npcdist(bws = bw.cdist, data = d, newdata = nd.c)
  cdist.named <- npcdist(bws = y ~ x, data = d, newdata = nd.c, nmulti = 1, tol = 1, ftol = 1)
  expect_lt(max(abs(as.numeric(cdist.named$condist) - as.numeric(cdist.pos$condist))), 1e-2)
  expect_lt(max(abs(as.numeric(cdist.named$conderr) - as.numeric(cdist.pos$conderr))), 1e-2)
})

test_that("inid ksum fast path matches explicit resample refits for npudens/npudist", {
  skip_if_not_installed("np")

  set.seed(325)
  n <- 70
  x <- rnorm(n)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 19))
  B <- 9L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  dens.bw <- npudensbw(dat = tx, bws = 0.35, bandwidth.compute = FALSE)
  dist.bw <- npudistbw(dat = tx, bws = 0.35, bandwidth.compute = FALSE)
  fast.fun <- getFromNamespace(".np_inid_boot_from_ksum_unconditional", "np")

  dens.fast <- fast.fun(
    xdat = tx,
    exdat = ex,
    bws = dens.bw,
    B = B,
    operator = "normal",
    counts = counts
  )
  dist.fast <- fast.fun(
    xdat = tx,
    exdat = ex,
    bws = dist.bw,
    B = B,
    operator = "integral",
    counts = counts
  )

  dens.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  dist.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  for (b in seq_len(B)) {
    idx <- rep.int(seq_len(n), counts[, b])
    dens.explicit[b, ] <- npudens(tdat = tx[idx, , drop = FALSE], edat = ex, bws = dens.bw)$dens
    dist.explicit[b, ] <- npudist(tdat = tx[idx, , drop = FALSE], edat = ex, bws = dist.bw)$dist
  }

  expect_equal(dens.fast$t, dens.explicit, tolerance = 1e-10)
  expect_equal(dist.fast$t, dist.explicit, tolerance = 1e-10)
  expect_equal(dens.fast$t0, npudens(tdat = tx, edat = ex, bws = dens.bw)$dens, tolerance = 1e-12)
  expect_equal(dist.fast$t0, npudist(tdat = tx, edat = ex, bws = dist.bw)$dist, tolerance = 1e-12)
})

test_that("inid ksum fast path matches explicit resample refits for npcdens/npcdist", {
  skip_if_not_installed("np")

  set.seed(326)
  n <- 80
  x <- rnorm(n)
  y <- rnorm(n)
  tx <- data.frame(x = x)
  ty <- data.frame(y = y)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 17))
  ey <- data.frame(y = seq(min(y), max(y), length.out = 17))
  B <- 7L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  dens.bw <- npcdensbw(xdat = tx, ydat = ty, bws = c(0.45, 0.45), bandwidth.compute = FALSE)
  dist.bw <- npcdistbw(xdat = tx, ydat = ty, bws = c(0.45, 0.45), bandwidth.compute = FALSE)
  fast.fun <- getFromNamespace(".np_inid_boot_from_ksum_conditional", "np")

  dens.fast <- fast.fun(
    xdat = tx,
    ydat = ty,
    exdat = ex,
    eydat = ey,
    bws = dens.bw,
    B = B,
    cdf = FALSE,
    counts = counts
  )
  dist.fast <- fast.fun(
    xdat = tx,
    ydat = ty,
    exdat = ex,
    eydat = ey,
    bws = dist.bw,
    B = B,
    cdf = TRUE,
    counts = counts
  )

  expect_true(is.list(dens.fast))
  expect_true(is.list(dist.fast))

  dens.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  dist.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  for (b in seq_len(B)) {
    idx <- rep.int(seq_len(n), counts[, b])
    dens.explicit[b, ] <- npcdens(
      txdat = tx[idx, , drop = FALSE],
      tydat = ty[idx, , drop = FALSE],
      exdat = ex,
      eydat = ey,
      bws = dens.bw
    )$condens
    dist.explicit[b, ] <- npcdist(
      txdat = tx[idx, , drop = FALSE],
      tydat = ty[idx, , drop = FALSE],
      exdat = ex,
      eydat = ey,
      bws = dist.bw
    )$condist
  }

  expect_equal(dens.fast$t, dens.explicit, tolerance = 1e-10)
  expect_equal(dist.fast$t, dist.explicit, tolerance = 1e-10)
  expect_equal(dens.fast$t0, npcdens(txdat = tx, tydat = ty, exdat = ex, eydat = ey, bws = dens.bw)$condens, tolerance = 1e-12)
  expect_equal(dist.fast$t0, npcdist(txdat = tx, tydat = ty, exdat = ex, eydat = ey, bws = dist.bw)$condist, tolerance = 1e-12)
})

test_that("ksum fast paths honor non-default kernel/bound options for density/distribution families", {
  skip_if_not_installed("np")

  set.seed(3261)
  n <- 60
  x <- runif(n)
  y <- runif(n)
  tx <- data.frame(x = x)
  ty <- data.frame(y = y)
  ex <- data.frame(x = seq(0.02, 0.98, length.out = 13))
  ey <- data.frame(y = seq(0.03, 0.97, length.out = 13))
  B <- 6L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  u.dens.bw <- npudensbw(
    dat = tx,
    bws = 0.20,
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerbound = "fixed",
    ckerlb = 0.0,
    ckerub = 1.0
  )
  u.dist.bw <- npudistbw(
    dat = tx,
    bws = 0.20,
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerbound = "fixed",
    ckerlb = 0.0,
    ckerub = 1.0
  )
  c.dens.bw <- npcdensbw(
    xdat = tx,
    ydat = ty,
    bws = c(0.25, 0.25),
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    cxkertype = "epanechnikov",
    cykertype = "epanechnikov",
    cxkerbound = "fixed",
    cykerbound = "fixed",
    cxkerlb = 0.0,
    cykerlb = 0.0,
    cxkerub = 1.0,
    cykerub = 1.0
  )
  c.dist.bw <- npcdistbw(
    xdat = tx,
    ydat = ty,
    bws = c(0.25, 0.25),
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    cxkertype = "epanechnikov",
    cykertype = "epanechnikov",
    cxkerbound = "fixed",
    cykerbound = "fixed",
    cxkerlb = 0.0,
    cykerlb = 0.0,
    cxkerub = 1.0,
    cykerub = 1.0
  )

  fast.u <- getFromNamespace(".np_inid_boot_from_ksum_unconditional", "np")
  fast.c <- getFromNamespace(".np_inid_boot_from_ksum_conditional", "np")

  u.dens.fast <- fast.u(
    xdat = tx,
    exdat = ex,
    bws = u.dens.bw,
    B = B,
    operator = "normal",
    counts = counts
  )
  u.dist.fast <- fast.u(
    xdat = tx,
    exdat = ex,
    bws = u.dist.bw,
    B = B,
    operator = "integral",
    counts = counts
  )
  c.dens.fast <- fast.c(
    xdat = tx,
    ydat = ty,
    exdat = ex,
    eydat = ey,
    bws = c.dens.bw,
    B = B,
    cdf = FALSE,
    counts = counts
  )
  c.dist.fast <- fast.c(
    xdat = tx,
    ydat = ty,
    exdat = ex,
    eydat = ey,
    bws = c.dist.bw,
    B = B,
    cdf = TRUE,
    counts = counts
  )

  u.dens.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  u.dist.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  c.dens.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  c.dist.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  for (b in seq_len(B)) {
    idx <- rep.int(seq_len(n), counts[, b])
    u.dens.explicit[b, ] <- npudens(tdat = tx[idx, , drop = FALSE], edat = ex, bws = u.dens.bw)$dens
    u.dist.explicit[b, ] <- npudist(tdat = tx[idx, , drop = FALSE], edat = ex, bws = u.dist.bw)$dist
    c.dens.explicit[b, ] <- npcdens(
      txdat = tx[idx, , drop = FALSE],
      tydat = ty[idx, , drop = FALSE],
      exdat = ex,
      eydat = ey,
      bws = c.dens.bw
    )$condens
    c.dist.explicit[b, ] <- npcdist(
      txdat = tx[idx, , drop = FALSE],
      tydat = ty[idx, , drop = FALSE],
      exdat = ex,
      eydat = ey,
      bws = c.dist.bw
    )$condist
  }

  expect_equal(u.dens.fast$t, u.dens.explicit, tolerance = 1e-10)
  expect_equal(u.dist.fast$t, u.dist.explicit, tolerance = 1e-10)
  expect_equal(c.dens.fast$t, c.dens.explicit, tolerance = 1e-10)
  expect_equal(c.dist.fast$t, c.dist.explicit, tolerance = 1e-10)
})

test_that("conditional helper constructors forward kernel and bwscaling options", {
  skip_if_not_installed("np")

  set.seed(3262)
  n <- 55
  tx <- data.frame(x = runif(n))
  ty <- data.frame(y = runif(n))
  bw.base <- npcdensbw(
    xdat = tx,
    ydat = ty,
    bws = c(0.22, 0.22),
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    bwscaling = FALSE,
    cxkertype = "epanechnikov",
    cykertype = "epanechnikov",
    cxkerbound = "fixed",
    cykerbound = "fixed",
    cxkerlb = 0.0,
    cykerlb = 0.0,
    cxkerub = 1.0,
    cykerub = 1.0
  )
  bw <- bw.base
  bw$bwscaling <- TRUE

  np.ns <- asNamespace("np")
  cap <- new.env(parent = emptyenv())
  cap$calls <- list()

  trace(
    what = "kbandwidth.numeric",
    where = np.ns,
    tracer = bquote({
      assign(
        "calls",
        c(
          get("calls", envir = .(cap)),
          list(list(
            bwscaling = bwscaling,
            ckertype = ckertype,
            ckerorder = ckerorder,
            ckerbound = ckerbound,
            ckerlb = ckerlb,
            ckerub = ckerub
          ))
        ),
        envir = .(cap)
      )
    }),
    print = FALSE
  )
  on.exit(untrace("kbandwidth.numeric", where = np.ns), add = TRUE)

  make.kx <- getFromNamespace(".np_con_make_kbandwidth_x", "np")
  make.kxy <- getFromNamespace(".np_con_make_kbandwidth_xy", "np")
  kx <- make.kx(bws = bw, xdat = tx)
  kxy <- make.kxy(bws = bw, xdat = tx, ydat = ty)
  expect_false(is.null(kx))
  expect_false(is.null(kxy))
  expect_true(length(cap$calls) >= 2L)

  for (call in cap$calls) {
    expect_identical(isTRUE(call$bwscaling), FALSE)
    expect_identical(as.character(call$ckertype), as.character(bw$cxkertype))
    expect_identical(as.character(call$ckerorder), as.character(bw$cxkerorder))
    expect_identical(as.character(call$ckerbound), as.character(bw$cxkerbound))
  }
})

test_that("manual bws bwscaling toggle is invariant for unconditional ksum helper output", {
  skip_if_not_installed("np")

  set.seed(3263)
  n <- 70
  tx <- data.frame(x = runif(n))
  ex <- data.frame(x = seq(0.04, 0.96, length.out = 11))
  B <- 7L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  bw.base <- npudensbw(
    dat = tx,
    bws = 0.21,
    bandwidth.compute = FALSE,
    bwtype = "fixed",
    bwscaling = FALSE
  )
  bw.alt <- bw.base
  bw.alt$bwscaling <- TRUE

  fast.u <- getFromNamespace(".np_inid_boot_from_ksum_unconditional", "np")
  out.base <- fast.u(
    xdat = tx,
    exdat = ex,
    bws = bw.base,
    B = B,
    operator = "normal",
    counts = counts
  )
  out.alt <- fast.u(
    xdat = tx,
    exdat = ex,
    bws = bw.alt,
    B = B,
    operator = "normal",
    counts = counts
  )

  expect_equal(out.base$t0, out.alt$t0, tolerance = 1e-12)
  expect_equal(out.base$t, out.alt$t, tolerance = 1e-12)
})

test_that("density/distribution plot bootstrap rejects wild selector", {
  skip_if_not_installed("np")

  set.seed(328)
  n <- 50
  x <- rnorm(n)
  y <- rnorm(n)

  ubw <- npudensbw(dat = data.frame(x = x), bws = 0.5, bandwidth.compute = FALSE)
  expect_error(
    suppressWarnings(
      plot(
        ubw,
        output = "data",
        perspective = FALSE,
        errors = "bootstrap",
        bootstrap = "wild",
        B = 9
      )
    ),
    "not supported for unconditional density/distribution estimators"
  )

  cbw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.6, 0.6),
    bandwidth.compute = FALSE
  )
  expect_error(
    suppressWarnings(
      plot(
        cbw,
        output = "data",
        perspective = FALSE,
        errors = "bootstrap",
        bootstrap = "wild",
        B = 9
      )
    ),
    "not supported for conditional density/distribution estimators"
  )
})

test_that("conditional bandwidth selectors reject unused named dots", {
  set.seed(20260511)
  d <- data.frame(x = runif(30), y = rnorm(30))

  expect_error(
    npcdensbw(y ~ x, data = d, ckertype = "banana", nmulti = 1L),
    "unused argument"
  )
  expect_error(
    npcdistbw(y ~ x, data = d, ckertype = "banana", nmulti = 1L),
    "unused argument"
  )
})

test_that("nearest-neighbor regression helper matches explicit resample refits", {
  skip_if_not_installed("np")

  set.seed(32905)
  n <- 45
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(2 * pi * x1) + 0.6 * x2 + rnorm(n, sd = 0.08)
  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- tx[seq_len(15), , drop = FALSE]
  B <- 9L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))
  fast.fun <- getFromNamespace(".np_inid_boot_from_reghat_exact", "np")

  cfgs <- list(
    list(regtype = "lc", basis = NULL, degree = NULL),
    list(regtype = "ll", basis = NULL, degree = NULL),
    list(regtype = "lp", basis = "glp", degree = c(2L, 2L))
  )

  for (bt in c("generalized_nn", "adaptive_nn")) {
    bw.val <- if (identical(bt, "adaptive_nn")) c(5, 5) else c(2, 2)

    for (cfg in cfgs) {
      bw.args <- list(
        xdat = tx,
        ydat = y,
        regtype = cfg$regtype,
        bws = bw.val,
        bwtype = bt,
        bandwidth.compute = FALSE
      )
      if (!is.null(cfg$basis)) {
        bw.args$basis <- cfg$basis
        bw.args$degree <- cfg$degree
      }
      bw <- do.call(npregbw, bw.args)

      fast.out <- fast.fun(
        xdat = tx,
        exdat = ex,
        bws = bw,
        ydat = y,
        B = B,
        counts = counts
      )

      explicit.t <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
      for (b in seq_len(B)) {
        idx <- rep.int(seq_len(n), counts[, b])
        explicit.t[b, ] <- npreg(
          txdat = tx[idx, , drop = FALSE],
          tydat = y[idx],
          exdat = ex,
          bws = bw,
          gradients = FALSE,
          warn.glp.gradient = FALSE
        )$mean
      }

      expect_equal(
        fast.out$t,
        explicit.t,
        tolerance = 1e-10,
        info = paste(bt, cfg$regtype)
      )
      expect_equal(
        fast.out$t0,
        npreg(txdat = tx, tydat = y, exdat = ex, bws = bw, gradients = FALSE, warn.glp.gradient = FALSE)$mean,
        tolerance = 1e-10,
        info = paste(bt, cfg$regtype)
      )
    }

    bw.grad <- npregbw(
      xdat = tx,
      ydat = y,
      regtype = "ll",
      bws = bw.val,
      bwtype = bt,
      bandwidth.compute = FALSE
    )
    fast.grad <- fast.fun(
      xdat = tx,
      exdat = ex,
      bws = bw.grad,
      ydat = y,
      B = B,
      counts = counts,
      gradients = TRUE,
      gradient.order = 1L,
      slice.index = 1L
    )

    explicit.grad <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
    for (b in seq_len(B)) {
      idx <- rep.int(seq_len(n), counts[, b])
      explicit.grad[b, ] <- npreg(
        txdat = tx[idx, , drop = FALSE],
        tydat = y[idx],
        exdat = ex,
        bws = bw.grad,
        gradients = TRUE,
        gradient.order = 1L,
        warn.glp.gradient = FALSE
      )$grad[, 1L]
    }

    expect_equal(
      fast.grad$t,
      explicit.grad,
      tolerance = 1e-10,
      info = paste(bt, "ll-grad")
    )
    expect_equal(
      fast.grad$t0,
      npreg(txdat = tx, tydat = y, exdat = ex, bws = bw.grad, gradients = TRUE, gradient.order = 1L, warn.glp.gradient = FALSE)$grad[, 1L],
      tolerance = 1e-10,
      info = paste(bt, "ll-grad")
    )
  }
})

test_that("nearest-neighbor ksum helpers match explicit resample refits", {
  skip_if_not_installed("np")

  set.seed(32906)
  n <- 40
  tx <- data.frame(x = runif(n))
  ty <- data.frame(y = runif(n))
  ex <- data.frame(x = seq(0.1, 0.9, length.out = 11))
  ey <- data.frame(y = seq(0.1, 0.9, length.out = 11))
  B <- 9L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  fast.u <- getFromNamespace(".np_inid_boot_from_ksum_unconditional", "np")
  fast.c <- getFromNamespace(".np_inid_boot_from_ksum_conditional", "np")

  for (bt in c("generalized_nn", "adaptive_nn")) {
    bw.val <- if (identical(bt, "adaptive_nn")) 5 else 2
    cbw.val <- rep.int(bw.val, 2L)

    u.dens.bw <- npudensbw(dat = tx, bws = bw.val, bwtype = bt, bandwidth.compute = FALSE)
    u.dist.bw <- npudistbw(dat = tx, bws = bw.val, bwtype = bt, bandwidth.compute = FALSE)
    c.dens.bw <- npcdensbw(xdat = tx, ydat = ty, bws = cbw.val, bwtype = bt, bandwidth.compute = FALSE)
    c.dist.bw <- npcdistbw(xdat = tx, ydat = ty, bws = cbw.val, bwtype = bt, bandwidth.compute = FALSE)

    u.dens.fast <- fast.u(
      xdat = tx,
      exdat = ex,
      bws = u.dens.bw,
      B = B,
      operator = "normal",
      counts = counts
    )
    u.dist.fast <- fast.u(
      xdat = tx,
      exdat = ex,
      bws = u.dist.bw,
      B = B,
      operator = "integral",
      counts = counts
    )
    c.dens.fast <- fast.c(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = c.dens.bw,
      B = B,
      cdf = FALSE,
      counts = counts
    )
    c.dist.fast <- fast.c(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = c.dist.bw,
      B = B,
      cdf = TRUE,
      counts = counts
    )

    u.dens.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
    u.dist.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
    c.dens.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
    c.dist.explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
    for (b in seq_len(B)) {
      idx <- rep.int(seq_len(n), counts[, b])
      u.dens.explicit[b, ] <- npudens(tdat = tx[idx, , drop = FALSE], edat = ex, bws = u.dens.bw)$dens
      u.dist.explicit[b, ] <- npudist(tdat = tx[idx, , drop = FALSE], edat = ex, bws = u.dist.bw)$dist
      c.dens.explicit[b, ] <- npcdens(
        txdat = tx[idx, , drop = FALSE],
        tydat = ty[idx, , drop = FALSE],
        exdat = ex,
        eydat = ey,
        bws = c.dens.bw
      )$condens
      c.dist.explicit[b, ] <- npcdist(
        txdat = tx[idx, , drop = FALSE],
        tydat = ty[idx, , drop = FALSE],
        exdat = ex,
        eydat = ey,
        bws = c.dist.bw
      )$condist
    }

    expect_equal(u.dens.fast$t, u.dens.explicit, tolerance = 1e-10, info = paste(bt, "udens"))
    expect_equal(u.dist.fast$t, u.dist.explicit, tolerance = 1e-10, info = paste(bt, "udist"))
    expect_equal(c.dens.fast$t, c.dens.explicit, tolerance = 1e-10, info = paste(bt, "npcdens"))
    expect_equal(c.dist.fast$t, c.dist.explicit, tolerance = 1e-10, info = paste(bt, "npcdist"))
    expect_equal(u.dens.fast$t0, npudens(tdat = tx, edat = ex, bws = u.dens.bw)$dens, tolerance = 1e-12, info = paste(bt, "udens-t0"))
    expect_equal(u.dist.fast$t0, npudist(tdat = tx, edat = ex, bws = u.dist.bw)$dist, tolerance = 1e-12, info = paste(bt, "udist-t0"))
    expect_equal(c.dens.fast$t0, npcdens(txdat = tx, tydat = ty, exdat = ex, eydat = ey, bws = c.dens.bw)$condens, tolerance = 1e-12, info = paste(bt, "npcdens-t0"))
    expect_equal(c.dist.fast$t0, npcdist(txdat = tx, tydat = ty, exdat = ex, eydat = ey, bws = c.dist.bw)$condist, tolerance = 1e-12, info = paste(bt, "npcdist-t0"))
  }
})

test_that("nearest-neighbor frozen hat helpers match frozen operator application", {
  skip_if_not_installed("np")

  set.seed(32907)
  n <- 36
  tx <- data.frame(x = runif(n))
  ty <- data.frame(y = runif(n))
  ex <- data.frame(x = seq(0.1, 0.9, length.out = 9))
  ey <- data.frame(y = seq(0.15, 0.85, length.out = 9))
  B <- 7L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  frozen.u <- getFromNamespace(".np_inid_boot_from_hat_unconditional_frozen", "np")
  frozen.c <- getFromNamespace(".np_inid_boot_from_hat_conditional_frozen", "np")
  udens.hat <- getFromNamespace("npudenshat", "np")
  udist.hat <- getFromNamespace("npudisthat", "np")
  cdens.hat <- getFromNamespace("npcdenshat", "np")
  cdist.hat <- getFromNamespace("npcdisthat", "np")
  make_xkbw <- getFromNamespace(".npcdhat_make_xkbw", "np")
  make_ybw <- getFromNamespace(".npcdhat_make_ybw", "np")
  make_kmat <- getFromNamespace(".npcdhat_make_kernel_matrix", "np")

  for (bt in c("generalized_nn", "adaptive_nn")) {
    bw.val <- if (identical(bt, "adaptive_nn")) 5 else 2
    cbw.val <- rep.int(bw.val, 2L)

    u.dens.bw <- npudensbw(dat = tx, bws = bw.val, bwtype = bt, bandwidth.compute = FALSE)
    u.dist.bw <- npudistbw(dat = tx, bws = bw.val, bwtype = bt, bandwidth.compute = FALSE)
    c.dens.bw <- npcdensbw(xdat = tx, ydat = ty, bws = cbw.val, bwtype = bt, bandwidth.compute = FALSE)
    c.dist.bw <- npcdistbw(xdat = tx, ydat = ty, bws = cbw.val, bwtype = bt, bandwidth.compute = FALSE)

    H.udens <- unclass(udens.hat(bws = u.dens.bw, tdat = tx, edat = ex, output = "matrix"))
    H.udist <- unclass(udist.hat(bws = u.dist.bw, tdat = tx, edat = ex, output = "matrix"))
    H.cdens <- unclass(cdens.hat(bws = c.dens.bw, txdat = tx, tydat = ty, exdat = ex, eydat = ey, output = "matrix"))
    H.cdist <- unclass(cdist.hat(bws = c.dist.bw, txdat = tx, tydat = ty, exdat = ex, eydat = ey, output = "matrix"))
    xkbw.dens <- make_xkbw(bws = c.dens.bw, txdat = tx)
    xkbw.dist <- make_xkbw(bws = c.dist.bw, txdat = tx)
    ybw.dens <- make_ybw(bws = c.dens.bw, tydat = ty)
    ybw.dist <- make_ybw(bws = c.dist.bw, tydat = ty)
    Kx.dens <- make_kmat(
      kbw = xkbw.dens,
      txdat = tx,
      exdat = ex,
      operator = rep.int("normal", ncol(tx))
    )
    Ky.dens <- make_kmat(
      kbw = ybw.dens,
      txdat = ty,
      exdat = ey,
      operator = rep.int("normal", ncol(ty))
    )
    Kx.dist <- make_kmat(
      kbw = xkbw.dist,
      txdat = tx,
      exdat = ex,
      operator = rep.int("normal", ncol(tx))
    )
    Ky.dist <- make_kmat(
      kbw = ybw.dist,
      txdat = ty,
      exdat = ey,
      operator = rep.int("integral", ncol(ty))
    )
    den.op.dens <- Kx.dens / n
    num.op.dens <- (Kx.dens * Ky.dens) / n
    den.op.dist <- Kx.dist / n
    num.op.dist <- (Kx.dist * Ky.dist) / n

    u.dens.frozen <- frozen.u(
      xdat = tx,
      exdat = ex,
      bws = u.dens.bw,
      B = B,
      operator = "normal",
      counts = counts
    )
    u.dist.frozen <- frozen.u(
      xdat = tx,
      exdat = ex,
      bws = u.dist.bw,
      B = B,
      operator = "integral",
      counts = counts
    )
    c.dens.frozen <- frozen.c(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = c.dens.bw,
      B = B,
      cdf = FALSE,
      counts = counts
    )
    c.dist.frozen <- frozen.c(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = c.dist.bw,
      B = B,
      cdf = TRUE,
      counts = counts
    )

    expect_equal(u.dens.frozen$t, t(H.udens %*% counts), tolerance = 1e-10, info = paste(bt, "udens"))
    expect_equal(u.dist.frozen$t, t(H.udist %*% counts), tolerance = 1e-10, info = paste(bt, "udist"))
    expect_equal(
      c.dens.frozen$t,
      t(num.op.dens %*% counts) / pmax(t(den.op.dens %*% counts), .Machine$double.eps),
      tolerance = 1e-10,
      info = paste(bt, "npcdens")
    )
    expect_equal(
      c.dist.frozen$t,
      t(num.op.dist %*% counts) / pmax(t(den.op.dist %*% counts), .Machine$double.eps),
      tolerance = 1e-10,
      info = paste(bt, "npcdist")
    )

    expect_equal(u.dens.frozen$t0, rowSums(H.udens), tolerance = 1e-12, info = paste(bt, "udens-t0"))
    expect_equal(u.dist.frozen$t0, rowSums(H.udist), tolerance = 1e-12, info = paste(bt, "udist-t0"))
    expect_equal(
      c.dens.frozen$t0,
      rowSums(num.op.dens) / pmax(rowSums(den.op.dens), .Machine$double.eps),
      tolerance = 1e-12,
      info = paste(bt, "npcdens-t0")
    )
    expect_equal(
      c.dist.frozen$t0,
      rowSums(num.op.dist) / pmax(rowSums(den.op.dist), .Machine$double.eps),
      tolerance = 1e-12,
      info = paste(bt, "npcdist-t0")
    )
  }
})

test_that("npindex nearest-neighbor inid helper matches explicit resample refits", {
  skip_if_not_installed("np")

  set.seed(3292)
  n <- 45
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.08)
  tx <- data.frame(x1 = x1, x2 = x2)
  B <- 9L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))
  fast.fun <- getFromNamespace(".np_inid_boot_from_index", "np")

  cfgs <- list(
    list(regtype = "lc", basis = NULL, degree = NULL, h = 5L),
    list(regtype = "ll", basis = NULL, degree = NULL, h = 5L),
    list(regtype = "lp", basis = "tensor", degree = 2L, h = 5L)
  )

  for (bt in c("generalized_nn", "adaptive_nn")) {
    for (cfg in cfgs) {
      bw.args <- list(
        xdat = tx,
        ydat = y,
        bws = c(1, 1, cfg$h),
        bandwidth.compute = FALSE,
        regtype = cfg$regtype,
        bwtype = bt
      )
      if (!is.null(cfg$basis)) {
        bw.args$basis <- cfg$basis
        bw.args$degree <- cfg$degree
      }
      bw <- do.call(npindexbw, bw.args)

      fast.out <- fast.fun(
        xdat = tx,
        ydat = y,
        bws = bw,
        B = B,
        counts = counts,
        gradients = FALSE
      )

      explicit.t <- matrix(NA_real_, nrow = B, ncol = n)
      for (b in seq_len(B)) {
        idx <- rep.int(seq_len(n), counts[, b])
        explicit.t[b, ] <- npindex(
          txdat = tx[idx, , drop = FALSE],
          tydat = y[idx],
          exdat = tx,
          bws = bw,
          gradients = FALSE
        )$mean
      }

      expect_equal(
        fast.out$t,
        explicit.t,
        tolerance = 1e-8,
        info = paste(bt, cfg$regtype)
      )
      expect_equal(
        fast.out$t0,
        npindex(txdat = tx, tydat = y, exdat = tx, bws = bw, gradients = FALSE)$mean,
        tolerance = 1e-8,
        info = paste(bt, cfg$regtype)
      )
    }
  }
})

progress_time_counter <- function(start = 0, by = 1.3) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

shadow_lines <- function(shadow) {
  vapply(shadow$trace, `[[`, character(1L), "line")
}

test_that("distribution and conditional public routes use the generic bandwidth selection label", {
  set.seed(321)
  n <- 24
  x <- runif(n)
  y <- x^2 + rnorm(n, sd = 0.1)

  cases <- list(
    quote(npudistbw(dat = data.frame(x = x), nmulti = 2)),
    quote(npudist(tdat = data.frame(x = x), nmulti = 2)),
    quote(npcdensbw(xdat = data.frame(x = x), ydat = data.frame(y = y), regtype = "lc", nmulti = 2)),
    quote(npcdens(txdat = data.frame(x = x), tydat = data.frame(y = y), regtype = "lc", nmulti = 2)),
    quote(npcdistbw(xdat = data.frame(x = x), ydat = data.frame(y = y), regtype = "lc", nmulti = 2)),
    quote(npcdist(txdat = data.frame(x = x), tydat = data.frame(y = y), regtype = "lc", nmulti = 2)),
    quote(npplregbw(xdat = data.frame(x = x), zdat = data.frame(z = x), ydat = y, nmulti = 2)),
    quote(npplreg(y ~ x | z, data = data.frame(x = x, z = x, y = y), nmulti = 2))
  )

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  for (expr in cases) {
    actual <- capture_progress_shadow_trace(
      eval(expr),
      force_renderer = "single_line",
      now = progress_time_counter()
    )

    lines <- shadow_lines(actual)
    info <- paste(deparse(expr), collapse = "")

    expect_true(any(grepl("^\\[np\\] Bandwidth selection", lines)), info = info)
    expect_true(any(grepl("100\\.0%, eta 0\\.0s\\)$", lines)), info = info)
  }
})

test_that("npcdist basic functionality works", {
  data("faithful")
  bw <- npcdistbw(xdat=faithful$eruptions, ydat=faithful$waiting, 
                  bws=c(0.5, 5), bandwidth.compute=FALSE)
  
  fit <- npcdist(bws=bw)
  
  expect_s3_class(fit, "condistribution")
  expect_type(predict(fit), "double")
  expect_equal(length(predict(fit)), nrow(faithful))
  
  expect_output(summary(fit))
})

test_that("extended bw generics route named data args without bws to NULL methods", {
  set.seed(20260222)
  n <- 35
  x <- rnorm(n)
  z <- rnorm(n)
  y <- x + z + rnorm(n)

  x1 <- data.frame(x = x)
  x2 <- data.frame(x = x, z = z)
  y1 <- data.frame(y = y)
  z1 <- data.frame(z = z)

  bw_cdens <- npcdensbw(xdat = x1, ydat = y1, bwmethod = "normal-reference")
  bw_cdist <- npcdistbw(xdat = x1, ydat = y1, bwmethod = "normal-reference")
  bw_pl <- npplregbw(xdat = x1, ydat = y, zdat = z1, nmulti = 1)
  bw_si <- npindexbw(xdat = x2, ydat = y, nmulti = 1)
  bw_sc <- npscoefbw(xdat = x1, ydat = y, zdat = z1, nmulti = 1)

  expect_s3_class(bw_cdens, "conbandwidth")
  expect_s3_class(bw_cdist, "condbandwidth")
  expect_s3_class(bw_pl, "plbandwidth")
  expect_s3_class(bw_si, "sibandwidth")
  expect_s3_class(bw_sc, "scbandwidth")
})

test_that("plot contract: bootstrap args require explicit bootstrap mode across engine families", {
  skip_if_not_installed("np")

  msg <- "bootstrap controls require errors = \"bootstrap\""

  set.seed(811)
  n <- 60
  x <- runif(n)
  x2 <- runif(n)
  z <- runif(n)
  y <- sin(2 * pi * x) + 0.3 * z + rnorm(n, sd = 0.08)

  xdat <- data.frame(x = x)
  ydat <- data.frame(y = y)
  idxdat <- data.frame(y = y, x = x, x2 = x2)

  rbw <- npregbw(xdat = xdat, ydat = y, bws = 0.25, bandwidth.compute = FALSE)
  ubw <- npudensbw(dat = xdat, bws = 0.25, bandwidth.compute = FALSE)
  dbw <- npudistbw(dat = xdat, bws = 0.25, bandwidth.compute = FALSE)
  cbw <- npcdensbw(xdat = xdat, ydat = ydat, bws = c(0.30, 0.30), bandwidth.compute = FALSE)
  cdbw <- npcdistbw(xdat = xdat, ydat = ydat, bws = c(0.30, 0.30), bandwidth.compute = FALSE)
  pbw <- npplregbw(
    xdat = data.frame(x = z),
    zdat = data.frame(z = x),
    ydat = y,
    bws = matrix(c(0.30, 0.30), nrow = 2),
    bandwidth.compute = FALSE
  )
  sbw <- npindexbw(
    y ~ x + x2,
    data = idxdat,
    bws = c(1, 0.30, 0.30),
    bandwidth.compute = FALSE,
    method = "ichimura"
  )
  scbw <- npscoefbw(
    xdat = data.frame(x = x),
    zdat = data.frame(z = z),
    ydat = y,
    bws = 0.20,
    bandwidth.compute = FALSE
  )

  expect_error(
    suppressWarnings(plot(rbw, output = "data", perspective = FALSE, B = 5)),
    msg,
    fixed = TRUE
  )
  expect_error(
    suppressWarnings(plot(ubw, output = "data", perspective = FALSE, B = 5)),
    msg,
    fixed = TRUE
  )
  expect_error(
    suppressWarnings(plot(dbw, output = "data", perspective = FALSE, B = 5)),
    msg,
    fixed = TRUE
  )
  expect_error(
    suppressWarnings(plot(cbw, output = "data", perspective = FALSE, B = 5)),
    msg,
    fixed = TRUE
  )
  expect_error(
    suppressWarnings(plot(cdbw, output = "data", perspective = FALSE, B = 5)),
    msg,
    fixed = TRUE
  )
  expect_error(
    suppressWarnings(
      plot(
        pbw,
        xdat = data.frame(x = z),
        ydat = y,
        zdat = data.frame(z = x),
        output = "data",
        perspective = FALSE,
        B = 5
      )
    ),
    msg,
    fixed = TRUE
  )
  expect_error(
    suppressWarnings(
      plot(
        sbw,
        xdat = data.frame(x = x, x2 = x2),
        ydat = y,
        output = "data",
        perspective = FALSE,
        B = 5
      )
    ),
    msg,
    fixed = TRUE
  )
  expect_error(
    suppressWarnings(
      plot(
        scbw,
        xdat = data.frame(x = x),
        ydat = y,
        zdat = data.frame(z = z),
        output = "data",
        perspective = FALSE,
        B = 5
      )
    ),
    msg,
    fixed = TRUE
  )
})

library(np)

test_that("fixed +/-Inf bounds are parity-equivalent to none at fixed bandwidth", {
  set.seed(20260224)
  x <- runif(80)
  dat <- data.frame(x = x)

  bw.none <- npudensbw(
    dat = dat,
    bws = 0.2,
    bandwidth.compute = FALSE,
    ckerbound = "none"
  )

  bw.inf <- npudensbw(
    dat = dat,
    bws = 0.2,
    bandwidth.compute = FALSE,
    ckerbound = "fixed",
    ckerlb = -Inf,
    ckerub = Inf
  )

  fit.none <- npudens(bws = bw.none, tdat = dat)
  fit.inf <- npudens(bws = bw.inf, tdat = dat)

  expect_equal(as.numeric(fit.none$dens), as.numeric(fit.inf$dens), tolerance = 1e-12)
})

test_that("scalar fixed bounds recycle over multiple continuous variables", {
  set.seed(20260224)
  dat <- data.frame(x1 = runif(64), x2 = runif(64))

  bw.scalar <- npudensbw(
    dat = dat,
    bws = c(0.2, 0.25),
    bandwidth.compute = FALSE,
    ckerbound = "fixed",
    ckerlb = 0,
    ckerub = 1
  )

  bw.vector <- npudensbw(
    dat = dat,
    bws = c(0.2, 0.25),
    bandwidth.compute = FALSE,
    ckerbound = "fixed",
    ckerlb = c(0, 0),
    ckerub = c(1, 1)
  )

  fit.scalar <- npudens(bws = bw.scalar, tdat = dat)
  fit.vector <- npudens(bws = bw.vector, tdat = dat)

  expect_equal(as.numeric(fit.scalar$dens), as.numeric(fit.vector$dens), tolerance = 1e-12)
})

test_that("invalid fixed bounds are rejected with clear diagnostics", {
  set.seed(20260224)
  x <- runif(50)
  dat <- data.frame(x = x)

  expect_error(
    npudensbw(
      dat = dat,
      bws = 0.2,
      bandwidth.compute = FALSE,
      ckerbound = "fixed",
      ckerlb = 1,
      ckerub = 1
    ),
    "lower < upper"
  )

  expect_error(
    npudensbw(
      dat = dat,
      bws = 0.2,
      bandwidth.compute = FALSE,
      ckerbound = "fixed",
      ckerlb = 0.2,
      ckerub = 0.8
    ),
    "Violations: x"
  )
})

test_that("bounded generalized_nn is available for certified core public routes", {
  set.seed(20260224)
  x <- runif(36)
  y <- cos(2 * pi * x) + rnorm(36, sd = 0.05)
  xy <- data.frame(x = x)
  yy <- data.frame(y = runif(36))

  bw.ud.cvml <- npudensbw(
    dat = xy,
    bwmethod = "cv.ml",
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1
  )
  bw.ud.cvls <- npudensbw(
    dat = xy,
    bwmethod = "cv.ls",
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1
  )
  fit.ud <- npudens(bws = bw.ud.cvls, tdat = xy)

  bw.dist <- npudistbw(
    dat = xy,
    bwmethod = "cv.cdf",
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1
  )
  fit.dist <- npudist(bws = bw.dist, tdat = xy)

  bw.reg <- npregbw(
    xdat = xy,
    ydat = y,
    regtype = "ll",
    bwmethod = "cv.ls",
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1
  )
  fit.reg <- npreg(bws = bw.reg, txdat = xy, tydat = y)

  bw.cd.cvml <- npcdensbw(
    xdat = xy,
    ydat = yy,
    bwmethod = "cv.ml",
    bwtype = "generalized_nn",
    cxkerbound = "range",
    cykerbound = "range",
    nmulti = 1
  )
  bw.cd.cvls <- npcdensbw(
    xdat = xy,
    ydat = yy,
    bwmethod = "cv.ls",
    bwtype = "generalized_nn",
    cxkerbound = "range",
    cykerbound = "range",
    nmulti = 1
  )
  fit.cd <- npcdens(bws = bw.cd.cvls, txdat = xy, tydat = yy)

  bw.cdist <- npcdistbw(
    xdat = xy,
    ydat = yy,
    bwmethod = "cv.ls",
    bwtype = "generalized_nn",
    cxkerbound = "range",
    cykerbound = "range",
    nmulti = 1
  )
  fit.cdist <- npcdist(bws = bw.cdist, txdat = xy, tydat = yy)

  expect_true(all(is.finite(as.numeric(bw.ud.cvml$bw))))
  expect_true(is.finite(bw.ud.cvml$fval))
  expect_true(all(is.finite(as.numeric(bw.ud.cvls$bw))))
  expect_true(is.finite(bw.ud.cvls$fval))
  expect_true(all(is.finite(as.numeric(fit.ud$dens))))

  expect_true(all(is.finite(as.numeric(bw.dist$bw))))
  expect_true(is.finite(bw.dist$fval))
  expect_true(all(is.finite(as.numeric(fit.dist$dist))))

  expect_true(all(is.finite(as.numeric(bw.reg$bw))))
  expect_true(is.finite(bw.reg$fval))
  expect_true(all(is.finite(as.numeric(fit.reg$mean))))

  expect_true(all(is.finite(as.numeric(bw.cd.cvml$xbw))))
  expect_true(all(is.finite(as.numeric(bw.cd.cvml$ybw))))
  expect_true(is.finite(bw.cd.cvml$fval))
  expect_true(all(is.finite(as.numeric(bw.cd.cvls$xbw))))
  expect_true(all(is.finite(as.numeric(bw.cd.cvls$ybw))))
  expect_true(is.finite(bw.cd.cvls$fval))
  expect_true(all(is.finite(as.numeric(fit.cd$condens))))

  expect_true(all(is.finite(as.numeric(bw.cdist$xbw))))
  expect_true(all(is.finite(as.numeric(bw.cdist$ybw))))
  expect_true(is.finite(bw.cdist$fval))
  expect_true(all(is.finite(as.numeric(fit.cdist$condist))))
})

test_that("npplreg, npindex, and npscoef bounded generalized_nn are available", {
  set.seed(20260224)
  x <- runif(32)
  y <- cos(2 * pi * x)
  y_sc <- (1 + sin(2 * pi * x)) * x + rnorm(32, sd = 0.05)
  xy <- data.frame(x = x)

  bw.pl <- npplregbw(
    xdat = xy,
    ydat = y,
    zdat = xy,
    bwmethod = "cv.ls",
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1
  )
  fit.pl <- npplreg(bws = bw.pl, txdat = xy, tydat = y, tzdat = xy)

  bw.index <- npindexbw(
    xdat = data.frame(x1 = x, x2 = x^2),
    ydat = y,
    method = "ichimura",
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1
  )
  fit.index <- npindex(bws = bw.index, txdat = data.frame(x1 = x, x2 = x^2), tydat = y)
  bw.sc <- npscoefbw(
    xdat = xy,
    ydat = y_sc,
    zdat = xy,
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1
  )
  fit.sc <- npscoef(bws = bw.sc, txdat = xy, tydat = y_sc, tzdat = xy, iterate = FALSE)

  expect_true(all(is.finite(as.numeric(unlist(lapply(bw.pl$bw, function(obj) obj$bw))))))
  expect_true(all(is.finite(as.numeric(fit.pl$mean))))
  expect_true(all(is.finite(as.numeric(bw.index$beta))))
  expect_true(is.finite(as.numeric(bw.index$bw)))
  expect_true(all(is.finite(as.numeric(fit.index$mean))))
  expect_true(all(is.finite(as.numeric(bw.sc$bw))))
  expect_true(all(is.finite(as.numeric(fit.sc$mean))))
})

test_that("evaluation support violations are caught before native execution", {
  set.seed(20260224)
  x <- runif(70)
  dat <- data.frame(x = x)

  bw <- npudensbw(
    dat = dat,
    bws = 0.2,
    bandwidth.compute = FALSE,
    ckerbound = "range"
  )

  expect_error(
    npudens(
      bws = bw,
      tdat = dat,
      edat = data.frame(x = max(x) + 0.05)
    ),
    "x >"
  )
})

test_that("predict paths enforce bounded eval checks with variable diagnostics", {
  set.seed(20260224)
  x <- runif(80)
  y <- runif(80)
  dat.x <- data.frame(x = x)
  dat.y <- data.frame(y = y)

  bw.den <- npudensbw(
    dat = dat.x,
    bws = 0.2,
    bandwidth.compute = FALSE,
    ckerbound = "range"
  )
  fit.den <- npudens(bws = bw.den, tdat = dat.x)
  expect_error(
    predict(fit.den, edat = data.frame(x = max(x) + 0.01)),
    "Evaluation data violate 'ckerbound' bounds: x >"
  )

  bw.dist <- npudistbw(
    dat = dat.x,
    bws = 0.2,
    bandwidth.compute = FALSE,
    ckerbound = "range"
  )
  fit.dist <- npudist(bws = bw.dist, tdat = dat.x)
  expect_error(
    predict(fit.dist, edat = data.frame(x = max(x) + 0.01)),
    "Evaluation data violate 'ckerbound' bounds: x >"
  )

  bw.reg <- npregbw(
    xdat = dat.x,
    ydat = y,
    bws = 0.2,
    bandwidth.compute = FALSE,
    ckerbound = "range"
  )
  fit.reg <- npreg(bws = bw.reg, txdat = dat.x, tydat = y)
  expect_error(
    predict(fit.reg, exdat = data.frame(x = max(x) + 0.01)),
    "Evaluation data violate 'ckerbound' bounds: x >"
  )

  bw.cd <- npcdensbw(
    xdat = dat.x,
    ydat = dat.y,
    bws = c(0.2, 0.2),
    bandwidth.compute = FALSE,
    cxkerbound = "range",
    cykerbound = "range"
  )
  fit.cd <- npcdens(bws = bw.cd, txdat = dat.x, tydat = dat.y)
  expect_error(
    predict(fit.cd, exdat = data.frame(x = max(x) + 0.01), eydat = data.frame(y = mean(y))),
    "Evaluation data violate 'cxkerbound' bounds: x >"
  )
  expect_error(
    predict(fit.cd, exdat = data.frame(x = mean(x)), eydat = data.frame(y = max(y) + 0.01)),
    "Evaluation data violate 'cykerbound' bounds: y >"
  )

  bw.cdist <- npcdistbw(
    xdat = dat.x,
    ydat = dat.y,
    bws = c(0.2, 0.2),
    bandwidth.compute = FALSE,
    cxkerbound = "range",
    cykerbound = "range"
  )
  fit.cdist <- npcdist(bws = bw.cdist, txdat = dat.x, tydat = dat.y)
  expect_error(
    predict(fit.cdist, exdat = data.frame(x = max(x) + 0.01), eydat = data.frame(y = mean(y))),
    "Evaluation data violate 'cxkerbound' bounds: x >"
  )
  expect_error(
    predict(fit.cdist, exdat = data.frame(x = mean(x)), eydat = data.frame(y = max(y) + 0.01)),
    "Evaluation data violate 'cykerbound' bounds: y >"
  )
})

library(np)

quiet_eval <- function(expr) {
  value <- NULL
  capture.output(value <- force(expr))
  value
}

manual_conditional_frozen_boot <- function(bw, xdat, ydat, exdat, eydat, counts, cdf) {
  ns <- asNamespace("np")
  make_xkbw <- get(".npcdhat_make_xkbw", envir = ns)
  make_ybw <- get(".npcdhat_make_ybw", envir = ns)
  make_kmat <- get(".npcdhat_make_kernel_matrix", envir = ns)

  xkbw <- make_xkbw(bws = bw, txdat = xdat)
  ykbw <- make_ybw(bws = bw, tydat = ydat)
  Kx <- make_kmat(
    kbw = xkbw,
    txdat = xdat,
    exdat = exdat,
    operator = rep.int("normal", ncol(xdat))
  )
  Ky <- make_kmat(
    kbw = ykbw,
    txdat = ydat,
    exdat = eydat,
    operator = rep.int(if (isTRUE(cdf)) "integral" else "normal", ncol(ydat))
  )

  n <- nrow(xdat)
  den.op <- Kx / n
  num.op <- (Kx * Ky) / n
  den <- t(den.op %*% counts)
  num <- t(num.op %*% counts)

  list(
    t = num / pmax(den, .Machine$double.eps),
    t0 = rowSums(num.op) / pmax(rowSums(den.op), .Machine$double.eps)
  )
}

manual_conditional_frozen_weighted_state <- function(bw, xdat, ydat, exdat, eydat, counts, cdf) {
  ns <- asNamespace("np")
  make_kbx <- get(".np_con_make_kbandwidth_x", envir = ns)
  make_kbxy <- get(".np_con_make_kbandwidth_xy", envir = ns)
  exact_state_build <- get(".np_ksum_exact_state_build", envir = ns)
  exact_state_eval <- get(".np_ksum_eval_exact_state", envir = ns)
  bind_fast <- get(".np_bind_data_frames_fast", envir = ns)

  kbx <- make_kbx(bws = bw, xdat = xdat)
  kbxy <- make_kbxy(bws = bw, xdat = xdat, ydat = ydat)
  den.state <- exact_state_build(
    bws = kbx,
    exdat = exdat,
    operator = rep.int("normal", ncol(xdat))
  )
  num.state <- exact_state_build(
    bws = kbxy,
    exdat = bind_fast(exdat, eydat),
    operator = c(
      rep.int("normal", ncol(xdat)),
      rep.int(if (isTRUE(cdf)) "integral" else "normal", ncol(ydat))
    )
  )

  n.total <- colSums(counts)
  tmat <- t(vapply(seq_len(ncol(counts)), function(j) {
    den <- as.numeric(exact_state_eval(
      state = den.state,
      txdat = xdat,
      weights = matrix(as.double(counts[, j]), ncol = 1L)
    )) / n.total[j]
    num <- as.numeric(exact_state_eval(
      state = num.state,
      txdat = bind_fast(xdat, ydat),
      weights = matrix(as.double(counts[, j]), ncol = 1L)
    )) / n.total[j]
    num / pmax(den, .Machine$double.eps)
  }, numeric(nrow(exdat))))

  den0 <- as.numeric(exact_state_eval(
    state = den.state,
    txdat = xdat,
    weights = matrix(1.0, nrow = nrow(xdat), ncol = 1L)
  )) / nrow(xdat)
  num0 <- as.numeric(exact_state_eval(
    state = num.state,
    txdat = bind_fast(xdat, ydat),
    weights = matrix(1.0, nrow = nrow(xdat), ncol = 1L)
  )) / nrow(xdat)

  list(
    t = tmat,
    t0 = num0 / pmax(den0, .Machine$double.eps)
  )
}

test_that("conditional frozen helper preserves numerator/denominator recombination for nonfixed dens/dist", {
  ns <- asNamespace("np")
  frozen.fun <- get(".np_inid_boot_from_hat_conditional_frozen", envir = ns)

  xdat <- data.frame(x = c(0.05, 0.15, 0.30, 0.45, 0.70, 0.90))
  ydat <- data.frame(y = c(0.10, 0.20, 0.35, 0.50, 0.72, 0.88))
  counts <- cbind(
    c(2, 0, 1, 1, 1, 1),
    c(0, 2, 1, 1, 1, 1),
    c(1, 1, 2, 0, 1, 1)
  )
  exdat <- data.frame(x = c(0.10, 0.35, 0.80))
  eydat <- data.frame(y = c(0.12, 0.40, 0.82))

  cases <- expand.grid(
    bwtype = c("generalized_nn", "adaptive_nn"),
    cdf = c(FALSE, TRUE),
    stringsAsFactors = FALSE
  )

  for (ii in seq_len(nrow(cases))) {
    bw <- quiet_eval(
      if (isTRUE(cases$cdf[ii])) {
        npcdistbw(
          xdat = xdat,
          ydat = ydat,
          bws = c(3L, 3L),
          bwtype = cases$bwtype[ii],
          bandwidth.compute = FALSE
        )
      } else {
        npcdensbw(
          xdat = xdat,
          ydat = ydat,
          bws = c(3L, 3L),
          bwtype = cases$bwtype[ii],
          bandwidth.compute = FALSE
        )
      }
    )

    helper.out <- frozen.fun(
      xdat = xdat,
      ydat = ydat,
      exdat = exdat,
      eydat = eydat,
      bws = bw,
      B = ncol(counts),
      cdf = cases$cdf[ii],
      counts = counts
    )
    manual.out <- manual_conditional_frozen_boot(
      bw = bw,
      xdat = xdat,
      ydat = ydat,
      exdat = exdat,
      eydat = eydat,
      counts = counts,
      cdf = cases$cdf[ii]
    )
    weighted.out <- manual_conditional_frozen_weighted_state(
      bw = bw,
      xdat = xdat,
      ydat = ydat,
      exdat = exdat,
      eydat = eydat,
      counts = counts,
      cdf = cases$cdf[ii]
    )

    expect_equal(
      helper.out$t0,
      manual.out$t0,
      tolerance = 1e-12,
      info = paste(cases$bwtype[ii], if (isTRUE(cases$cdf[ii])) "dist" else "dens", "t0")
    )
    expect_equal(
      helper.out$t,
      manual.out$t,
      tolerance = 1e-12,
      info = paste(cases$bwtype[ii], if (isTRUE(cases$cdf[ii])) "dist" else "dens", "t")
    )
    if (!identical(cases$bwtype[ii], "adaptive_nn")) {
      expect_equal(
        helper.out$t0,
        weighted.out$t0,
        tolerance = 1e-12,
        info = paste(cases$bwtype[ii], if (isTRUE(cases$cdf[ii])) "dist" else "dens", "weighted-state t0")
      )
      expect_equal(
        helper.out$t,
        weighted.out$t,
        tolerance = 1e-12,
        info = paste(cases$bwtype[ii], if (isTRUE(cases$cdf[ii])) "dist" else "dens", "weighted-state t")
      )
    }
  }
})

test_that("plot contract: regression and conditional estimators return data payloads", {
  skip_if_not_installed("np")

  set.seed(101)
  n <- 80
  x <- runif(n)
  z <- runif(n)
  y <- rnorm(n)

  rbw <- npregbw(y ~ x + z, nmulti = 1)
  rout <- suppressWarnings(
    plot(
      rbw,
      output = "data",
      perspective = FALSE,
      errors = "bootstrap",
      B = 9,
      band = "all",
      view = "fixed"
    )
  )

  expect_type(rout, "list")
  expect_true(length(rout) > 0)
  expect_true(all(vapply(rout, inherits, logical(1), "npregression")))
  expect_true(all(vapply(rout, function(xi) !is.null(xi$mean), logical(1))))
  expect_true(all(vapply(rout, function(xi) !is.null(xi$merr), logical(1))))
})

test_that("plot contract: package legends are user-controllable on band-all plots", {
  set.seed(20260511)
  n <- 36L
  x <- runif(n)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.1)
  rbw <- npregbw(
    xdat = data.frame(x = x),
    ydat = y,
    bws = 0.35,
    bandwidth.compute = FALSE
  )

  old.dev <- grDevices::dev.cur()
  grDevices::pdf(file = tempfile(fileext = ".pdf"))
  on.exit({
    grDevices::dev.off()
    if (old.dev > 1L)
      grDevices::dev.set(old.dev)
  }, add = TRUE)

  expect_silent(suppressWarnings(plot(
    rbw, xdat = data.frame(x = x), ydat = y, perspective = FALSE,
    errors = "bootstrap", bootstrap = "wild", B = 9L, band = "all",
    neval = 6L, legend = FALSE
  )))
  expect_silent(suppressWarnings(plot(
    rbw, xdat = data.frame(x = x), ydat = y, perspective = FALSE,
    errors = "bootstrap", bootstrap = "wild", B = 9L, band = "all",
    neval = 6L, legend = "bottomright"
  )))
  expect_silent(suppressWarnings(plot(
    rbw, xdat = data.frame(x = x), ydat = y, perspective = FALSE,
    errors = "bootstrap", bootstrap = "wild", B = 9L, band = "all",
    neval = 6L, legend = list(x = "bottomleft", bty = "o")
  )))
  expect_error(
    suppressWarnings(plot(
      rbw, xdat = data.frame(x = x), ydat = y, perspective = FALSE,
      errors = "bootstrap", bootstrap = "wild", B = 9L, band = "all",
      neval = 6L, legend = 1
    )),
    "legend must be TRUE/FALSE"
  )
})

test_that("plot contract: npcdens/npcdist support bootstrap all in data mode", {
  skip_if_not_installed("np")

  set.seed(102)
  n <- 80
  x <- runif(n)
  y <- rnorm(n)

  cbw <- npcdensbw(y ~ x, nmulti = 1, bwmethod = "cv.ls")
  cout <- suppressWarnings(
    plot(
      cbw,
      output = "data",
      perspective = FALSE,
      errors = "bootstrap",
      B = 9,
      band = "all",
      view = "fixed"
    )
  )

  expect_type(cout, "list")
  expect_true(length(cout) > 0)
  expect_true(all(vapply(cout, inherits, logical(1), "condensity")))
  expect_true(all(vapply(cout, function(xi) !is.null(xi$condens), logical(1))))
  expect_true(all(vapply(cout, function(xi) !is.null(xi$conderr), logical(1))))

  dbw <- npcdistbw(y ~ x, nmulti = 1, bwmethod = "cv.ls")
  dout <- suppressWarnings(
    plot(
      dbw,
      output = "data",
      perspective = FALSE,
      errors = "bootstrap",
      B = 9,
      band = "all",
      view = "fixed"
    )
  )

  expect_type(dout, "list")
  expect_true(length(dout) > 0)
  expect_true(all(vapply(dout, inherits, logical(1), "condistribution")))
  expect_true(all(vapply(dout, function(xi) !is.null(xi$condist), logical(1))))
  expect_true(all(vapply(dout, function(xi) !is.null(xi$conderr), logical(1))))
})

test_that("plot contract: 3D plot-data matches data mode for regression and conditional estimators", {
  skip_if_not_installed("np")

  with_plot_device <- function(expr) {
    pdf(file = tempfile(fileext = ".pdf"))
    on.exit(dev.off(), add = TRUE)
    suppressWarnings(force(expr))
  }

  set.seed(107)
  n <- 80
  x <- rnorm(n)
  z <- rnorm(n)
  y <- x - 0.5 * z + rnorm(n, sd = 0.2)

  rfit <- npreg(
    bws = npregbw(
      xdat = data.frame(x = x, z = z),
      ydat = y,
      bws = c(0.55, 0.55),
      bandwidth.compute = FALSE
    ),
    txdat = data.frame(x = x, z = z),
    tydat = y
  )
  rdata <- plot(rfit, output = "data", view = "fixed")
  expect_warning(plot(rfit, output = "data", view = "fixed"), NA)
  rplotdata <- with_plot_device(plot(rfit, output = "plot-data", view = "fixed"))

  expect_type(rplotdata, "list")
  expect_named(rplotdata, names(rdata))
  expect_s3_class(rplotdata$r1, "npregression")
  expect_equal(rplotdata$r1$mean, rdata$r1$mean)
  expect_equal(rplotdata$r1$eval, rdata$r1$eval)

  cfit <- npcdens(
    bws = npcdensbw(
      xdat = data.frame(x = x),
      ydat = data.frame(y = y),
      bws = c(0.55, 0.55),
      bandwidth.compute = FALSE,
      regtype = "lp",
      degree = 4L,
      bernstein = TRUE
    ),
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE
  )
  cdata <- plot(cfit, output = "data", view = "fixed")
  expect_warning(plot(cfit, output = "data", view = "fixed"), NA)
  cplotdata <- with_plot_device(plot(cfit, output = "plot-data", view = "fixed"))

  expect_type(cplotdata, "list")
  expect_named(cplotdata, names(cdata))
  expect_s3_class(cplotdata$cd1, "condensity")
  expect_equal(cplotdata$cd1$condens, cdata$cd1$condens)
  expect_equal(cplotdata$cd1$condens.raw, cdata$cd1$condens.raw)
  expect_equal(cplotdata$cd1$xeval, cdata$cd1$xeval)
  expect_equal(cplotdata$cd1$yeval, cdata$cd1$yeval)

  dfit <- npcdist(
    bws = npcdistbw(
      xdat = data.frame(x = x),
      ydat = data.frame(y = y),
      bws = c(0.55, 0.55),
      bandwidth.compute = FALSE,
      regtype = "lp",
      degree = 4L,
      bernstein = TRUE
    ),
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE
  )
  ddata <- suppressWarnings(plot(dfit, output = "data", view = "fixed"))
  dplotdata <- with_plot_device(plot(dfit, output = "plot-data", view = "fixed"))

  expect_type(dplotdata, "list")
  expect_named(dplotdata, names(ddata))
  expect_s3_class(dplotdata$cd1, "condistribution")
  expect_equal(dplotdata$cd1$condist, ddata$cd1$condist)
  expect_equal(dplotdata$cd1$condist.raw, ddata$cd1$condist.raw)
  expect_equal(dplotdata$cd1$xeval, ddata$cd1$xeval)
  expect_equal(dplotdata$cd1$yeval, ddata$cd1$yeval)
})

test_that("plot contract: remaining public plot families return plot-data payloads", {
  skip_if_not_installed("np")

  with_plot_device <- function(expr) {
    pdf(file = tempfile(fileext = ".pdf"))
    on.exit(dev.off(), add = TRUE)
    suppressWarnings(force(expr))
  }

  expect_plot_modes_match <- function(object, fields, ...) {
    data.out <- suppressWarnings(plot(object, output = "data", ...))
    plot.out <- with_plot_device(plot(object, output = "plot-data", ...))

    expect_type(plot.out, "list")
    expect_named(plot.out, names(data.out))

    for (nm in names(data.out)) {
      expect_s3_class(plot.out[[nm]], class(data.out[[nm]])[1L])
      for (field in fields) {
        if (field %in% names(data.out[[nm]]) && field %in% names(plot.out[[nm]])) {
          expect_equal(plot.out[[nm]][[field]], data.out[[nm]][[field]])
        }
      }
    }
  }

  set.seed(108)
  n <- 60
  x <- runif(n)
  x2 <- runif(n)
  z <- runif(n)
  y <- sin(2 * pi * x) + 0.5 * x2 + 1.2 * z + rnorm(n, sd = 0.05)

  bw.ud <- npudensbw(
    dat = data.frame(x = x),
    bws = 0.25,
    bandwidth.compute = FALSE
  )
  expect_plot_modes_match(bw.ud, fields = c("eval", "dens"), perspective = FALSE)

  fit.ud <- npudens(bws = bw.ud, tdat = data.frame(x = x))
  expect_plot_modes_match(fit.ud, fields = c("eval", "dens"), perspective = FALSE)

  bw.ui <- npudistbw(
    dat = data.frame(x = x),
    bws = 0.25,
    bandwidth.compute = FALSE
  )
  expect_plot_modes_match(bw.ui, fields = c("eval", "dist"), perspective = FALSE)

  fit.ui <- npudist(bws = bw.ui, tdat = data.frame(x = x))
  expect_plot_modes_match(fit.ui, fields = c("eval", "dist"), perspective = FALSE)

  bw.reg <- npregbw(
    xdat = data.frame(x = x),
    ydat = y,
    bws = 0.25,
    bandwidth.compute = FALSE
  )
  expect_plot_modes_match(bw.reg, fields = c("eval", "mean", "merr"), perspective = FALSE)

  bw.cd <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.25, 0.25),
    bandwidth.compute = FALSE
  )
  expect_plot_modes_match(
    bw.cd,
    fields = c("condens", "xeval", "yeval", "conderr"),
    perspective = TRUE,
    view = "fixed"
  )

  bw.cf <- npcdistbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.25, 0.25),
    bandwidth.compute = FALSE
  )
  expect_plot_modes_match(
    bw.cf,
    fields = c("condist", "xeval", "yeval", "conderr"),
    perspective = TRUE,
    view = "fixed"
  )

  fit.qr <- npqreg(
    bws = bw.cf,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    tau = 0.5
  )
  expect_plot_modes_match(
    fit.qr,
    fields = c("quantile", "xeval", "quanterr"),
    perspective = TRUE,
    view = "fixed"
  )

  fit.qr.alt <- npqreg(
    bws = bw.cf,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    tau = 0.4
  )
  bw.qr.alt.fixed <- plot(
    bw.cf,
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    output = "data",
    perspective = TRUE,
    view = "fixed",
    quantreg = TRUE,
    tau = 0.4
  )
  fit.qr.alt.fixed <- plot(
    fit.qr.alt,
    output = "data",
    perspective = TRUE,
    view = "fixed"
  )
  bw.qr.alt.slice <- plot(
    bw.cf,
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    output = "data",
    perspective = FALSE,
    quantreg = TRUE,
    tau = 0.4
  )
  fit.qr.alt.slice <- plot(
    fit.qr.alt,
    output = "data",
    perspective = FALSE
  )

  expect_true(all(vapply(fit.qr.alt.fixed, inherits, logical(1), "qregression")))
  expect_true(all(vapply(fit.qr.alt.slice, inherits, logical(1), "qregression")))
  expect_true(all(vapply(fit.qr.alt.fixed, function(xi) identical(xi$tau, 0.4), logical(1))))
  expect_true(all(vapply(fit.qr.alt.slice, function(xi) identical(xi$tau, 0.4), logical(1))))
  expect_equal(fit.qr.alt.fixed[[1L]]$quantile, bw.qr.alt.fixed[[1L]]$quantile)
  expect_equal(fit.qr.alt.fixed[[1L]]$xeval, bw.qr.alt.fixed[[1L]]$xeval)
  expect_equal(fit.qr.alt.fixed[[1L]]$quanterr, bw.qr.alt.fixed[[1L]]$quanterr)
  expect_equal(fit.qr.alt.slice[[1L]]$quantile, bw.qr.alt.slice[[1L]]$quantile)
  expect_equal(fit.qr.alt.slice[[1L]]$xeval, bw.qr.alt.slice[[1L]]$xeval)
  expect_equal(fit.qr.alt.slice[[1L]]$quanterr, bw.qr.alt.slice[[1L]]$quanterr)

  bw.si <- npindexbw(
    xdat = data.frame(x = x, x2 = x2),
    ydat = y,
    bws = c(0.25, 0.25, 1),
    bandwidth.compute = FALSE
  )
  expect_plot_modes_match(bw.si, fields = c("index", "mean", "grad"))

  fit.si <- npindex(bws = bw.si, txdat = data.frame(x = x, x2 = x2), tydat = y)
  expect_plot_modes_match(fit.si, fields = c("index", "mean", "grad"))

  bw.sc <- npscoefbw(
    xdat = data.frame(x = x),
    zdat = data.frame(z = z),
    ydat = y,
    bws = 0.25,
    bandwidth.compute = FALSE
  )
  expect_plot_modes_match(bw.sc, fields = c("eval", "mean", "grad"), perspective = FALSE)

  fit.sc <- npscoef(
    bws = bw.sc,
    txdat = data.frame(x = x),
    tzdat = data.frame(z = z),
    tydat = y
  )
  expect_plot_modes_match(fit.sc, fields = c("eval", "mean", "grad"), perspective = FALSE)

  bw.pl <- npplregbw(
    xdat = data.frame(z = z),
    zdat = data.frame(x = x),
    ydat = y,
    bws = matrix(c(0.25, 0.25), nrow = 2),
    bandwidth.compute = FALSE
  )
  expect_plot_modes_match(bw.pl, fields = c("mean", "merr"), perspective = FALSE)

  fit.pl <- npplreg(bws = bw.pl, txdat = data.frame(z = z), tzdat = data.frame(x = x), tydat = y)
  expect_plot_modes_match(fit.pl, fields = c("mean", "merr"), perspective = FALSE)
})

test_that("plot contract: rgl plot-data returns the usual data payload", {
  skip_if_not_installed("np")
  skip_if_not_installed("rgl")

  old.opts <- options(rgl.useNULL = TRUE, rgl.printRglwidget = TRUE)
  on.exit(options(old.opts), add = TRUE)

  set.seed(110)
  n <- 50
  x <- rnorm(n)
  z <- rnorm(n)
  y <- x - 0.5 * z + rnorm(n, sd = 0.1)

  rfit <- npreg(
    bws = npregbw(
      xdat = data.frame(x = x, z = z),
      ydat = y,
      bws = c(0.55, 0.55),
      bandwidth.compute = FALSE
    ),
    txdat = data.frame(x = x, z = z),
    tydat = y
  )

  rdata <- suppressWarnings(plot(
    rfit,
    output = "data",
    renderer = "rgl",
    view = "fixed",
    data_overlay = FALSE
  ))

  rplotdata <- suppressWarnings(plot(
    rfit,
    output = "plot-data",
    renderer = "rgl",
    view = "fixed",
    data_overlay = FALSE
  ))

  expect_type(rplotdata, "list")
  expect_named(rplotdata, names(rdata))
  expect_s3_class(rplotdata$r1, "npregression")
  expect_equal(rplotdata$r1$mean, rdata$r1$mean)
  expect_equal(rplotdata$r1$eval, rdata$r1$eval)

  cfit <- npcdens(
    bws = npcdensbw(
      xdat = data.frame(x = x),
      ydat = data.frame(y = y),
      bws = c(0.55, 0.55),
      bandwidth.compute = FALSE,
      regtype = "lp",
      degree = 4L,
      bernstein = TRUE
    ),
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE
  )

  cdata <- suppressWarnings(plot(
    cfit,
    output = "data",
    renderer = "rgl",
    view = "fixed"
  ))

  cplotdata <- suppressWarnings(plot(
    cfit,
    output = "plot-data",
    renderer = "rgl",
    view = "fixed"
  ))

  expect_type(cplotdata, "list")
  expect_named(cplotdata, names(cdata))
  expect_s3_class(cplotdata$cd1, "condensity")
  expect_equal(cplotdata$cd1$condens, cdata$cd1$condens)
  expect_equal(cplotdata$cd1$xeval, cdata$cd1$xeval)
  expect_equal(cplotdata$cd1$yeval, cdata$cd1$yeval)
})

test_that("plot contract: alpha is enforced in (0,0.5)", {
  skip_if_not_installed("np")

  set.seed(103)
  n <- 60
  x <- runif(n)
  y <- rnorm(n)
  bw <- npregbw(y ~ x, nmulti = 1)

  expect_error(
    suppressWarnings(
      plot(
        bw,
        output = "data",
        perspective = FALSE,
        errors = "bootstrap",
        alpha = 0.5
      )
    ),
    "must lie in \\(0, 0\\.5\\)"
  )

  expect_error(
    suppressWarnings(
      plot(
        bw,
        output = "data",
        perspective = FALSE,
        errors = "bootstrap",
        alpha = 0
      )
    ),
    "must lie in \\(0, 0\\.5\\)"
  )
})

test_that("plot contract: default condensity plot call stays scalar-safe", {
  skip_if_not_installed("np")

  set.seed(104)
  n <- 80
  x <- rnorm(n)
  y <- rnorm(n)

  bw <- npcdensbw(y ~ x, bws = c(1.0, 1.0), bandwidth.compute = FALSE)
  fit <- npcdens(bws = bw)

  pdf(file = tempfile(fileext = ".pdf"))
  on.exit(dev.off(), add = TRUE)
  expect_error(plot(fit, perspective = FALSE), NA)
})

test_that("plot contract: npscoef supports coef=TRUE plot path", {
  skip_if_not_installed("np")

  set.seed(105)
  n <- 60
  x <- runif(n)
  z <- runif(n, -2, 2)
  y <- x * exp(z) * (1 + rnorm(n, sd = 0.15))

  fit <- npscoef(y ~ x | z, regtype = "ll", betas = TRUE)
  out <- suppressWarnings(
    plot(
      fit,
      coef = TRUE,
      coef_index = 1,
      perspective = FALSE,
      neval = 20,
      output = "plot-data",
      errors = "none"
    )
  )

  expect_type(out, "list")
  expect_true(length(out) > 0)
  expect_true(all(vapply(out, inherits, logical(1), "smoothcoefficient")))
  expect_true(all(vapply(out, function(xi) !is.null(xi$mean), logical(1))))
})

test_that("plot contract: npplreg supports coef=TRUE plot-data payload", {
  skip_if_not_installed("np")

  set.seed(106)
  n <- 80
  x <- runif(n)
  z <- runif(n, -2, 2)
  y <- 1 + 0.7 * x + sin(z) + rnorm(n, sd = 0.15)
  xdat <- data.frame(x = x)
  zdat <- data.frame(z = z)

  fit <- npplreg(y ~ x | z, regtype = "ll")
  out <- suppressWarnings(
    plot(
      fit,
      xdat = xdat,
      ydat = y,
      zdat = zdat,
      coef = TRUE,
      output = "plot-data",
      errors = "none"
    )
  )

  expect_type(out, "list")
  expect_true(all(c("coefficients", "coefficient.stderr", "fit") %in% names(out)))
  expect_true(is.numeric(out$coefficients))
  expect_true(length(out$coefficients) >= 1L)
})

test_that("plot contract: scbandwidth perspective bootstrap uses z evaluation column", {
  skip_if_not_installed("np")

  set.seed(107)
  n <- 50
  x <- runif(n)
  z <- runif(n)
  y <- sin(2 * pi * z) + x * (1 + z) + rnorm(n, sd = 0.05)
  xdat <- data.frame(x = x)
  zdat <- data.frame(z = z)

  bw <- npscoefbw(xdat = xdat, ydat = y, zdat = zdat, regtype = "lc", nmulti = 1)
  neval <- 12L

  np.ns <- asNamespace("np")
  cap <- new.env(parent = emptyenv())
  cap$exdat <- NULL
  cap$ezdat <- NULL

  tq <- getFromNamespace("trim.quantiles", "np")
  x1.eval <- seq(tq(xdat[, 1L], 0.0)[1L], tq(xdat[, 1L], 0.0)[2L], length.out = neval)
  x2.eval <- seq(tq(zdat[, 1L], 0.0)[1L], tq(zdat[, 1L], 0.0)[2L], length.out = neval)
  x.eval <- expand.grid(x1.eval, x2.eval)
  colnames(x.eval) <- c("x", "z")

  trace(
    what = "compute.bootstrap.errors.scbandwidth",
    where = np.ns,
    tracer = bquote({
      assign("exdat", exdat, envir = .(cap))
      assign("ezdat", ezdat, envir = .(cap))
    }),
    print = FALSE
  )
  on.exit(try(untrace("compute.bootstrap.errors.scbandwidth", where = np.ns), silent = TRUE), add = TRUE)

  suppressWarnings(
    plot(
      bw,
      xdat = xdat,
      ydat = y,
      zdat = zdat,
      perspective = TRUE,
      view = "fixed",
      output = "data",
      errors = "bootstrap",
      bootstrap = "inid",
      B = 9,
      band = "pointwise",
      neval = neval
    )
  )

  expect_equal(cap$exdat, x.eval[, 1, drop = FALSE])
  expect_equal(cap$ezdat, x.eval[, 2, drop = FALSE])
})

test_that("plot contract: npscoef fitted perspective path preserves semantic z eval name", {
  skip_if_not_installed("np")

  set.seed(108)
  n <- 120
  x <- runif(n)
  z <- runif(n, min = -2, max = 2)
  y <- x * exp(z) * (1 + rnorm(n, sd = 0.15))

  fit <- npscoef(
    y ~ x | z,
    nomad = TRUE,
    nmulti = 1L
  )

  np.ns <- asNamespace("np")
  cap <- new.env(parent = emptyenv())
  cap$tzdat.names <- NULL
  cap$ezdat.names <- NULL

  trace(
    what = ".np_scoef_fit_internal",
    where = np.ns,
    tracer = bquote({
      if (!missing(tzdat) && !missing(ezdat) && !is.null(tzdat) && !is.null(ezdat)) {
        assign("tzdat.names", colnames(as.data.frame(tzdat)), envir = .(cap))
        assign("ezdat.names", colnames(as.data.frame(ezdat)), envir = .(cap))
      }
    }),
    print = FALSE
  )
  on.exit(try(untrace(".np_scoef_fit_internal", where = np.ns), silent = TRUE), add = TRUE)

  out <- suppressWarnings(plot(
    fit,
    view = "fixed",
    renderer = "base",
    output = "data",
    data_overlay = FALSE,
    errors = "none"
  ))

  expect_type(out, "list")
  expect_identical(cap$tzdat.names, "z")
  expect_identical(cap$ezdat.names, "z")
})

test_that("plot contract: bootstrap defaults are wild for regression-class and inid for unsupervised engines", {
  skip_if_not_installed("np")

  reg.engines <- c(
    ".np_plot_rbandwidth_engine",
    ".np_plot_scbandwidth_engine",
    ".np_plot_plbandwidth_engine",
    ".np_plot_sibandwidth_engine"
  )
  unsup.engines <- c(
    ".np_plot_bandwidth_engine",
    ".np_plot_dbandwidth_engine",
    ".np_plot_conbandwidth_engine",
    ".np_plot_condbandwidth_engine"
  )

  for (nm in reg.engines) {
    fn <- getFromNamespace(nm, "np")
    defaults <- eval(formals(fn)$plot.errors.boot.method)
    expect_identical(defaults[1L], "wild")
  }

  for (nm in unsup.engines) {
    fn <- getFromNamespace(nm, "np")
    defaults <- eval(formals(fn)$plot.errors.boot.method)
    expect_identical(defaults[1L], "inid")
  }

  for (nm in c(reg.engines, unsup.engines)) {
    fn <- getFromNamespace(nm, "np")
    defaults <- eval(formals(fn)$plot.errors.boot.nonfixed)
    expect_identical(defaults[1L], "exact")
  }
})

test_that("npudens core smoke stays alive", {
  set.seed(1)
  x <- data.frame(x = seq(0.1, 0.9, length.out = 24L))

  bw <- npudensbw(dat = x, bws = 0.2, bandwidth.compute = FALSE)
  fit <- npudens(bws = bw)

  expect_s3_class(fit, "npdensity")
  expect_equal(length(predict(fit)), nrow(x))
  expect_true(all(is.finite(predict(fit))))
})

test_that("npudist core smoke stays alive", {
  set.seed(2)
  x <- data.frame(x = seq(0.1, 0.9, length.out = 24L))

  bw <- npudistbw(dat = x, bws = 0.2, bandwidth.compute = FALSE)
  fit <- npudist(bws = bw)

  expect_s3_class(fit, "npdistribution")
  expect_equal(length(predict(fit)), nrow(x))
  expect_true(all(predict(fit) >= 0 & predict(fit) <= 1))
})

test_that("npreg core smoke stays alive", {
  set.seed(3)
  x <- data.frame(x = seq(0.1, 1.0, length.out = 28L))
  y <- sin(2 * pi * x$x)

  bw <- npregbw(xdat = x, ydat = y, bws = 0.25, bandwidth.compute = FALSE)
  fit <- npreg(bws = bw)

  expect_s3_class(fit, "npregression")
  expect_equal(length(predict(fit)), nrow(x))
  expect_true(all(is.finite(predict(fit))))
})

test_that("npcdens core smoke stays alive", {
  set.seed(4)
  x <- data.frame(x = seq(0.1, 1.0, length.out = 24L))
  y <- data.frame(y = x$x^2)

  bw <- npcdensbw(xdat = x, ydat = y, bws = c(0.25, 0.25), bandwidth.compute = FALSE)
  fit <- npcdens(bws = bw)

  expect_s3_class(fit, "condensity")
  expect_equal(length(predict(fit)), nrow(x))
  expect_true(all(is.finite(predict(fit))))
})

test_that("npcdist core smoke stays alive", {
  set.seed(5)
  x <- data.frame(x = seq(0.1, 1.0, length.out = 24L))
  y <- data.frame(y = x$x^2)

  bw <- npcdistbw(xdat = x, ydat = y, bws = c(0.25, 0.25), bandwidth.compute = FALSE)
  fit <- npcdist(bws = bw)

  expect_s3_class(fit, "condistribution")
  expect_equal(length(predict(fit)), nrow(x))
  expect_true(all(predict(fit) >= 0 & predict(fit) <= 1))
})

test_that("npplreg core smoke stays alive", {
  set.seed(6)
  n <- 24L
  z <- data.frame(z = seq(0.1, 1.0, length.out = n))
  x <- data.frame(x = seq(0.2, 1.1, length.out = n))
  y <- z$z^2 + 2 * x$x

  bw <- npplregbw(
    xdat = x,
    zdat = z,
    ydat = y,
    bws = matrix(c(0.25, 0.25), nrow = 2L),
    bandwidth.compute = FALSE
  )
  fit <- npplreg(bws = bw)

  expect_s3_class(fit, "plregression")
  expect_equal(length(predict(fit)), n)
  expect_true(all(is.finite(predict(fit))))
})

test_that("npindex core smoke stays alive", {
  set.seed(7)
  n <- 24L
  x <- data.frame(
    x1 = seq(0.1, 1.0, length.out = n),
    x2 = seq(1.0, 0.1, length.out = n)
  )
  y <- x$x1 - x$x2

  bw <- npindexbw(
    xdat = x,
    ydat = y,
    method = "ichimura",
    bws = c(1, 0.25, 0.25),
    bandwidth.compute = FALSE
  )
  fit <- npindex(bws = bw)

  expect_s3_class(fit, "singleindex")
  expect_equal(length(predict(fit)), n)
  expect_true(all(is.finite(predict(fit))))
})

test_that("npscoef core smoke stays alive", {
  set.seed(8)
  n <- 24L
  x <- data.frame(x = seq(0.1, 1.0, length.out = n))
  z <- data.frame(z = seq(1.0, 0.1, length.out = n))
  y <- x$x * z$z

  bw <- npscoefbw(
    xdat = x,
    zdat = z,
    ydat = y,
    bws = 0.25,
    bandwidth.compute = FALSE
  )
  fit <- npscoef(bws = bw, iterate = FALSE, errors = FALSE)

  expect_s3_class(fit, "smoothcoefficient")
  expect_equal(length(predict(fit)), n)
  expect_true(all(is.finite(predict(fit))))
})

test_that("npcdist NOMAD accounting is owner-level on fast-path fits", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)
  suppressPackageStartupMessages(library(np))

  set.seed(42)
  n <- 300L
  x <- runif(n)
  y <- runif(n)

  fit <- npcdist(
    y ~ x,
    nomad = TRUE,
    search.engine = "nomad",
    cykerbound = "range"
  )

  expect_gt(as.numeric(fit$bws$num.feval.fast[1L]), 0)
  expect_lte(
    as.numeric(fit$bws$num.feval.fast[1L]),
    as.numeric(fit$bws$num.feval[1L])
  )

  ev <- np:::.npcdistbw_eval_only(
    data.frame(x = x),
    data.frame(y = y),
    NULL,
    fit$bws,
    invalid.penalty = "baseline",
    penalty.multiplier = 10
  )

  expect_equal(as.numeric(ev$num.feval), 1)
  expect_equal(as.numeric(ev$num.feval.fast), 1)
})

library(np)

cdist_shadow_empty <- function(n) {
  matrix(numeric(0), nrow = n, ncol = 0)
}

cdist_shadow_cker <- function(kernel) {
  switch(kernel,
    gaussian = 0L,
    epanechnikov = 4L,
    uniform = 8L,
    truncated = 9L,
    stop("unsupported continuous kernel")
  )
}

cdist_shadow_uker <- function(kernel) {
  switch(kernel,
    aitchisonaitken = 0L,
    liracine = 1L,
    stop("unsupported unordered kernel")
  )
}

cdist_shadow_oker <- function(kernel) {
  switch(kernel,
    wangvanryzin = 0L,
    liracine = 2L,
    racineliyan = 3L,
    stop("unsupported ordered kernel")
  )
}

cdist_shadow_rbw <- function(bw) {
  c(
    bw$xbw[bw$ixcon],
    bw$ybw[bw$iycon],
    bw$ybw[bw$iyuno],
    bw$ybw[bw$iyord],
    bw$xbw[bw$ixuno],
    bw$xbw[bw$ixord]
  )
}

cdist_shadow_bwtype <- function(bw) {
  switch(bw$type,
    fixed = 0L,
    generalized_nn = 1L,
    adaptive_nn = 2L,
    stop("unsupported bandwidth type")
  )
}

cdist_shadow_safe_call <- function(name, ...) {
  on.exit(
    tryCatch(.Call("C_np_shadow_reset_state", PACKAGE = "np"),
             error = function(e) NULL),
    add = TRUE
  )
  .Call(name, ..., PACKAGE = "np")
}

call_public_cdist_cvls_shadow <- function(bw, x, ytrain, yeval = ytrain, cdfontrain = FALSE) {
  n <- nrow(x)
  ne <- nrow(yeval)
  cdist_shadow_safe_call(
    "C_np_shadow_cv_distribution_conditional",
    cdist_shadow_empty(n), cdist_shadow_empty(n), as.matrix(ytrain),
    cdist_shadow_empty(ne), cdist_shadow_empty(ne), as.matrix(yeval),
    cdist_shadow_empty(n), cdist_shadow_empty(n), as.matrix(x),
    as.double(cdist_shadow_rbw(bw)),
    cdist_shadow_bwtype(bw),
    cdist_shadow_cker(bw$cykertype),
    cdist_shadow_uker(bw$uykertype),
    cdist_shadow_oker(bw$oykertype),
    cdist_shadow_cker(bw$cxkertype),
    cdist_shadow_uker(bw$uxkertype),
    cdist_shadow_oker(bw$oxkertype),
    FALSE,
    0L,
    integer(0),
    FALSE,
    0L,
    cdfontrain,
    TRUE
  )
}

public_cdist_eval_grid <- function(ydat, ngrid = 100L) {
  probs <- seq(0, 1, length.out = ngrid)
  evy <- ydat[seq_len(ngrid), , drop = FALSE]
  for (i in seq_len(ncol(evy))) {
    evy[, i] <- uocquantile(ydat[, i], probs)
  }
  evy
}

test_that("public npcdistbw cv.ls keeps lc on the legacy objective", {
  set.seed(303)
  n <- 32L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = rnorm(n))

  bw.lc <- npcdistbw(
    xdat = x,
    ydat = y,
    regtype = "lc",
    bwmethod = "cv.ls",
    nmulti = 1,
    itmax = 1L
  )
  shadow <- call_public_cdist_cvls_shadow(bw.lc, x, y, yeval = public_cdist_eval_grid(y))

  expect_equal(bw.lc$fval, shadow$old, tolerance = 1e-10)
})

test_that("public npcdistbw cv.ls fixed LP/LL route activates with ll == lp parity", {
  set.seed(20260311)
  n <- 24L
  x <- data.frame(x1 = runif(n))
  y <- data.frame(y1 = x$x1 + rnorm(n, sd = 0.1))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdistbw(
    xdat = x,
    ydat = y,
    regtype = "ll",
    bwmethod = "cv.ls",
    nmulti = 1,
    itmax = 1L
  )
  bw.lp <- npcdistbw(
    xdat = x,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = degree,
    bwmethod = "cv.ls",
    nmulti = 1,
    itmax = 1L
  )

  expect_equal(bw.ll$fval, bw.lp$fval, tolerance = 1e-8)
})

test_that("public npcdistbw cv.ls generalized-nn LP route activates with ll == lp parity", {
  set.seed(20260312)
  n <- 24L
  x <- data.frame(x1 = runif(n))
  y <- data.frame(y1 = x$x1 + rnorm(n, sd = 0.1))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdistbw(
    xdat = x,
    ydat = y,
    regtype = "ll",
    bwtype = "generalized_nn",
    bwmethod = "cv.ls",
    nmulti = 1,
    itmax = 1L
  )
  bw.lp <- npcdistbw(
    xdat = x,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = degree,
    bwtype = "generalized_nn",
    bwmethod = "cv.ls",
    nmulti = 1,
    itmax = 1L
  )

  expect_true(is.finite(bw.ll$fval))
  expect_true(is.finite(bw.lp$fval))
  expect_equal(bw.ll$fval, bw.lp$fval, tolerance = 1e-8)
})

test_that("public npcdistbw cv.ls adaptive-nn LP route activates with ll == lp parity", {
  set.seed(20260313)
  n <- 24L
  x <- data.frame(x1 = runif(n))
  y <- data.frame(y1 = x$x1 + rnorm(n, sd = 0.1))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdistbw(
    xdat = x,
    ydat = y,
    regtype = "ll",
    bwtype = "adaptive_nn",
    bwmethod = "cv.ls",
    nmulti = 1,
    itmax = 1L
  )
  bw.lp <- npcdistbw(
    xdat = x,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = degree,
    bwtype = "adaptive_nn",
    bwmethod = "cv.ls",
    nmulti = 1,
    itmax = 1L
  )

  expect_true(is.finite(bw.ll$fval))
  expect_true(is.finite(bw.lp$fval))
  expect_equal(bw.ll$fval, bw.lp$fval, tolerance = 1e-8)
})

## Function that implements the entropy metric test for serial
## dependence described in "A Dependence Metric For Possibly Nonlinear
## Processes By C. W. Granger, E. Maasoumi and J. Racine Journal of
## Time Series Analysis (2004), Vol. 25, No. 5, 649-669.

npsdeptest <- function(data = NULL,
                       lag.num = 1,
                       method=c("integration","summation"),
                       bootstrap = TRUE,
                       boot.num = 399,
                       random.seed = 42) {
  
  ## Trap fatal errors

  if(is.data.frame(data)) stop(" you must enter a data vector (not data frame)")
  if(is.null(data)) stop(" you must enter a data vector")
  if((lag.num < 1) || (lag.num > length(data))) stop(" lag.num must be a positive integer less than the number of observations")
  if(ncol(data.frame(data)) != 1) stop(" data must have one dimension only")
  if(boot.num < 9) stop(" number of bootstrap replications must be >= 9")

  method <- match.arg(method)

  ## Save seed prior to setting

  seed.state <- .np_seed_enter(random.seed)
  .np_progress_note("Computing bandwidths")


  ## If the variable is a time series convert to type numeric

  if(is.ts(data)) data <- as.numeric(data)

  ## Omit NAs

  data <- na.omit(data)

  ## Define the metric entropy function

  Srho.bivar <- function(x.dat = NULL,
                         y.dat = NULL,
                         bw.x = NULL,
                         bw.y = NULL,
                         bw.joint = NULL,
                         method=c("integration","summation")) {
    
    if(is.null(bw.x)||is.null(bw.y)||is.null(bw.joint)) stop(" you must provide numeric bandwidths for f(x), f(y) and f(x,y)")
    
    method <- match.arg(method)
    
    if(method=="summation") {
      
      ## Summation version:
      ## \sum_i(1-sqrt(f(x_i)f(y_i)/f(x_i,y_i)))^2
      ## We could code this by hand per below, however, there is no
      ## performance penalty imposed by sum() so no need.

      ## Issue of common support points for evaluation
      
      f.x <- fitted(npudens(tdat=x.dat,bws=bw.x))
      f.y <- fitted(npudens(tdat=y.dat,bws=bw.y))
      f.xy <- fitted(npudens(tdat=cbind(x.dat,y.dat),bws=bw.joint))
      summand <- f.x*f.y/f.xy

      ## In summation version we divide densities which can lead to
      ## numerical instability issues not present in the integrand
      ## version (which uses differences instead). We check for this
      ## case, remove offending points, and warn. This traps -Inf,
      ## Inf, and NaN.

      if(!all(is.finite(summand))) {
        .np_warning(" non-finite value in summation-based statistic: integration recommended")
        summand <- summand[is.finite(summand)]
      }

      return(0.5*mean((1-sqrt(summand))**2))
      
    } else {
      
      ## Integration version:
      ## \int\int (sqrt(f(x,y))-sqrt(f(x))sqrt(f(y)))^2 dx dy
      
      lo.default <- c(min(x.dat)-10.0*IQR(x.dat),min(y.dat)-10.0*IQR(y.dat))
      up.default <- c(max(x.dat)+10.0*IQR(x.dat),max(y.dat)+10.0*IQR(y.dat))
      
      Srho.integrand <- function(xy) {
        
        ## We code this up by hand using a second order gaussian
        ## kernel. The multidimensional integration routines for some
        ## reason crawl to almost a halt if we call the functions in the
        ## np package. Good to have noticed this at this juncture, bad
        ## to not have general code though could branch at this point?
        
        ##      f.x <- fitted(npudens(tdat=x.dat,edat=xy[1],bws=bw.x))
        ##      f.y <- fitted(npudens(tdat=y.dat,edat=xy[2],bws=bw.y))
        ##      f.xy <- fitted(npudens(tdat=cbind(x.dat,y.dat),edat=cbind(xy[1],xy[2]),bws=bw.joint))


        f.x <- mean(dnorm((xy[1]-x.dat)/bw.x))/bw.x
        f.y <- mean(dnorm((xy[2]-y.dat)/bw.y))/bw.y
        f.xy <- mean(dnorm((xy[1]-x.dat)/bw.joint[1])*dnorm((xy[2]-y.dat)/bw.joint[2])
                    / (bw.joint[1]*bw.joint[2]))
        
        return((sqrt(f.xy)-sqrt(f.x)*sqrt(f.y))**2)
        
      }
      
      return(0.5*adaptIntegrate(Srho.integrand,
                                lowerLimit=lo.default,
                                upperLimit=up.default)$integral)
      
    }

  } ## end of Srho.bivar function

  ## Compute the metric entropy for lags 1 through lag.num

  Srho.vec <- numeric()
  ## `Portmanteau' cumulant of all lags
  Srho.cumulant.vec <- numeric()
  
  bw.y <- numeric()
  bw.y.lag <- numeric()
  bw.joint.y <- numeric()
  bw.joint.y.lag <- numeric()      
  
  ## Save the bandwidths for resampling exercise...
  lag.progress <- .np_progress_begin("Constructing metric entropy by lag", total = lag.num, surface = "lag")
  
  for (k in seq_len(lag.num)) {
    ## Create y and y.lag
    
    tmp <- ts.intersect(as.ts(data),lag(as.ts(data),k))
    y <- as.numeric(tmp[,1])
    y.lag <- as.numeric(tmp[,2])
    rm(tmp)

    ## Compute and save bandwidths (save for bootstrapping if
    ## requested)

    bw.y[k] <- .np_progress_with_legacy_suppressed(npudensbw(~y))$bw
    bw.y.lag[k] <- .np_progress_with_legacy_suppressed(npudensbw(~y.lag))$bw
    bw.joint <- .np_progress_with_legacy_suppressed(npudensbw(~y+y.lag))$bw
    bw.joint.y[k] <- bw.joint[1]
    bw.joint.y.lag[k] <- bw.joint[2]
    
    Srho.vec[k] <- Srho.bivar(y,y.lag,bw.y[k],bw.y.lag[k],bw.joint,method=method)
    lag.progress <- .np_progress_step(lag.progress, done = k, detail = paste("lag", k))

  }

  lag.progress <- .np_progress_end(lag.progress, detail = paste("lag", lag.num))

  for (k in seq_len(lag.num)) Srho.cumulant.vec[k] <- sum(Srho.vec[seq_len(k)])

  ## Bootstrap if requested - null is independence so simple iid
  ## index resampling under replacement is sufficient

  if(bootstrap) {

    Srho.vec.boot <- numeric()
    ## `Portmanteau' cumulant of all lags
    Srho.cumulant.vec.boot <- numeric()

    ## Matrix for resamples

    Srho.bootstrap.mat <- matrix(NA,boot.num,(lag.num))
		Srho.cumulant.bootstrap.mat <- matrix(NA,boot.num,(lag.num))
    progress <- .np_progress_begin("Bootstrap replications", total = boot.num, surface = "bootstrap")

    for (b in seq_len(boot.num)) {

      ## Resample under the null

      resampled.ts <- as.ts(data[sample.int(length(data), replace = TRUE)])

      for (k in seq_len(lag.num)) {
        tmp <- ts.intersect(resampled.ts, lag(resampled.ts, k))
        y <- as.numeric(tmp[,1])
        y.lag <- as.numeric(tmp[,2])
        Srho.vec.boot[k] <- Srho.bivar(y, y.lag, bw.y[k], bw.y.lag[k], c(bw.joint.y[k], bw.joint.y.lag[k]), method = method)
        Srho.bootstrap.mat[b,k] <- Srho.vec.boot[k]
        ## `Portmanteau' cumulant of all lags
        Srho.cumulant.vec.boot[k] <- sum(Srho.vec.boot[seq_len(k)])
        Srho.cumulant.bootstrap.mat[b,k] <- Srho.cumulant.vec.boot[k]
      }

      progress <- .np_progress_step(progress, done = b)

    }

    progress <- .np_progress_end(progress)

    ## Compute P-values

    P.vec <- numeric()
    P.cumulant.vec <- numeric()
    
    for (k in seq_len(lag.num)) {
      P.vec[k] <- mean(Srho.bootstrap.mat[,k] > Srho.vec[k])
      P.cumulant.vec[k] <- mean(Srho.cumulant.bootstrap.mat[,k] > Srho.cumulant.vec[k])
    }

  }

  ## Restore seed
  
  .np_seed_exit(seed.state)
  
  if(bootstrap) {
    
    sdeptest(Srho = Srho.vec,
             Srho.cumulant = Srho.cumulant.vec,
             Srho.bootstrap.mat = Srho.bootstrap.mat,
             Srho.cumulant.bootstrap.mat = Srho.cumulant.bootstrap.mat,
             P = P.vec,
             P.cumulant = P.cumulant.vec,
             bootstrap = bootstrap,
             boot.num = boot.num,
             lag.num = lag.num,
             bw.y = bw.y,
             bw.y.lag = bw.y.lag,
             bw.joint = cbind(bw.joint.y,bw.joint.y.lag))
    
  } else {
    
    sdeptest(Srho = Srho.vec,
             Srho.cumulant = Srho.cumulant.vec,
             Srho.bootstrap.mat = NULL,
             Srho.cumulant.bootstrap.mat = NULL,
             P = NULL,
             P.cumulant = NULL,
             bootstrap = bootstrap,
             boot.num = NULL,
             lag.num = lag.num,
             bw.y = bw.y,
             bw.y.lag = bw.y.lag,
             bw.joint = cbind(bw.joint.y,bw.joint.y.lag))
    
  }

}

test_that("npdeptest basic functionality works", {
  set.seed(42)
  n <- 50
  x <- rnorm(n)
  y <- x + rnorm(n, sd=0.1)
  
  # Use summation and small boot.num for speed
  test <- npdeptest(x, y, method="summation", boot.num=9)
  
  expect_s3_class(test, "deptest")
  expect_output(summary(test))
})

test_that("npsdeptest basic functionality works", {
  set.seed(42)
  n <- 50
  y <- arima.sim(n=n, list(ar=0.5))
  
  test <- npsdeptest(y, lag.num=1, method="summation", boot.num=9)
  
  expect_s3_class(test, "sdeptest")
  expect_output(summary(test))
})

npudens <-
  function(bws, ...){
    args <- list(...)

    if (!missing(bws)){
      if (is.recursive(bws)){
        if (!is.null(bws$formula) && is.null(args$tdat))
          UseMethod("npudens",bws$formula)
        else if (!is.null(bws$call) && is.null(args$tdat))
          UseMethod("npudens",bws$call)
        else if (!is.call(bws))
          UseMethod("npudens",bws)
        else
          UseMethod("npudens",NULL)
      } else {
        UseMethod("npudens", NULL)
      }
    } else {
      UseMethod("npudens", NULL)
    }
  }

npudens.formula <-
  function(bws, data = NULL, newdata = NULL, ...){

    tt <- terms(bws)

    m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)

    tmf <- bws$call[c(1,m)]
    tmf[[1]] <- as.name("model.frame")
    tmf[["formula"]] <- tt
    mf.args <- as.list(tmf)[-1L]
    umf <- tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

    tdat <- tmf[, attr(attr(tmf, "terms"),"term.labels"), drop = FALSE]

    has.eval <- !is.null(newdata)
    if (has.eval) {
      npValidateNewdataFormula(newdata, tt, include.response = TRUE)
      umf.args <- list(formula = tt, data = newdata)
      umf <- do.call(stats::model.frame, umf.args, envir = parent.frame())
      emf <- umf

      edat <- emf[, attr(attr(emf, "terms"),"term.labels"), drop = FALSE]
    }

    ud.args <- list(tdat = tdat)
    if (has.eval)
      ud.args$edat <- edat
    ud.args$bws <- bws
    ev <- do.call(npudens, c(ud.args, list(...)))

    ev$omit <- attr(umf,"na.action")
    ev$rows.omit <- as.vector(ev$omit)
    ev$nobs.omit <- length(ev$rows.omit)

    ev$dens <- napredict(ev$omit, ev$dens)
    ev$derr <- napredict(ev$omit, ev$derr)

    return(ev)
  }

npudens.call <-
  function(bws, ...) {
    npudens(bws, tdat = .np_eval_bws_call_arg(bws, "dat"),
            ...)
  }

npudens.bandwidth <-
  function(bws,
           tdat = stop("invoked without training data 'tdat'"),
           edat, ...){

  dots <- list(...)
  fit.start <- proc.time()[3]
  fit.progress.handoff <- isTRUE(dots$.np_fit_progress_handoff)

  no.e = missing(edat)

  tdat = toFrame(tdat)

  if (!no.e)
    edat = toFrame(edat)

  if (!(no.e || tdat %~% edat ))
    stop("tdat and edat are not similar data frames!")

  if (length(bws$bw) != length(tdat))
    stop("length of bandwidth vector does not match number of columns of 'tdat'")

  if ((any(bws$icon) &&
       !all(vapply(as.data.frame(tdat[, bws$icon]), inherits, logical(1), c("integer", "numeric")))) ||
      (any(bws$iord) &&
       !all(vapply(as.data.frame(tdat[, bws$iord]), inherits, logical(1), "ordered"))) ||
      (any(bws$iuno) &&
       !all(vapply(as.data.frame(tdat[, bws$iuno]), inherits, logical(1), "factor"))))
    stop("supplied bandwidths do not match 'tdat' in type")

  tdat <- na.omit(tdat)
  rows.omit <- unclass(na.action(tdat))

  if (!no.e){
    edat <- na.omit(edat)
    rows.omit <- unclass(na.action(edat))
  }

  tnrow = nrow(tdat)
  enrow = (if (no.e) tnrow else nrow(edat))

  ## re-assign levels in training and evaluation data to ensure correct
  ## conversion to numeric type.
    
  tdat <- adjustLevels(tdat, bws$xdati)
  
  if (!no.e)
    edat <- adjustLevels(edat, bws$xdati, allowNewCells = TRUE)

  if (!no.e)
    npKernelBoundsCheckEval(edat, bws$icon, bws$ckerlb, bws$ckerub, argprefix = "cker")

  ## grab the evaluation data before it is converted to numeric
  if(no.e)
    teval <- tdat
  else
    teval <- edat

  ## put the unordered, ordered, and continuous data in their own objects
  ## data that is not a factor is continuous.
  
  tdat = toMatrix(tdat)

  tuno = tdat[, bws$iuno, drop = FALSE]
  tcon = tdat[, bws$icon, drop = FALSE]
  tord = tdat[, bws$iord, drop = FALSE]

  if (!no.e){
    edat = toMatrix(edat)

    euno = edat[, bws$iuno, drop = FALSE]
    econ = edat[, bws$icon, drop = FALSE]
    eord = edat[, bws$iord, drop = FALSE]

  } else {
    euno = data.frame()
    eord = data.frame()
    econ = data.frame()
  }

  myopti = list(
    num_obs_train = tnrow,
    num_obs_eval = enrow,
    num_uno = bws$nuno,
    num_ord = bws$nord,
    num_con = bws$ncon,
    int_LARGE_SF = (if (bws$scaling) SF_NORMAL else SF_ARB),
    BANDWIDTH_den_extern = switch(bws$type,
      fixed = BW_FIXED,
      generalized_nn = BW_GEN_NN,
      adaptive_nn = BW_ADAP_NN),
    int_MINIMIZE_IO=if (isTRUE(getOption("np.messages"))) IO_MIN_FALSE else IO_MIN_TRUE, 
    ckerneval = switch(bws$ckertype,
      gaussian = CKER_GAUSS + bws$ckerorder/2 - 1,
      epanechnikov = CKER_EPAN + bws$ckerorder/2 - 1,
      uniform = CKER_UNI,
      "truncated gaussian" = CKER_TGAUSS),
    ukerneval = switch(bws$ukertype,
      aitchisonaitken = UKER_AIT,
      liracine = UKER_LR),
    okerneval = switch(bws$okertype,
      wangvanryzin = OKER_WANG,
      liracine = OKER_NLR,
        "racineliyan" = OKER_RLY),
    no.e = no.e,
    mcv.numRow = attr(bws$xmcv, "num.row"),
      densOrDist = NP_DO_DENS,
      old.dens = FALSE,
      int_do_tree = if (isTRUE(getOption("np.tree"))) DO_TREE_YES else DO_TREE_NO)
  cker.bounds.c <- npKernelBoundsMarshal(bws$ckerlb[bws$icon], bws$ckerub[bws$icon])

  myout <- .np_with_compiled_fit_progress(
    label = "Fitting density",
    total = .np_densdist_fit_total(bws = bws, tnrow = tnrow, enrow = enrow),
    handoff = fit.progress.handoff,
    handoff.detail = if (fit.progress.handoff) "starting" else NULL,
    .Call("C_np_density",
          as.double(tuno), as.double(tord), as.double(tcon),
          as.double(euno), as.double(eord), as.double(econ),
          as.double(c(bws$bw[bws$icon], bws$bw[bws$iuno], bws$bw[bws$iord])),
          as.double(bws$xmcv), as.double(attr(bws$xmcv, "pad.num")),
          as.double(bws$nconfac), as.double(bws$ncatfac), as.double(bws$sdev),
          as.integer(myopti),
          as.integer(enrow),
          as.double(cker.bounds.c$lb),
          as.double(cker.bounds.c$ub),
          PACKAGE = "np")
  )

  ## For purely categorical density with zero bandwidths, the variance of
  ## the sample proportion is p(1-p)/n. The C routine returns p/n; fix here.
  if (bws$ncon == 0 && bws$nord == 0 && bws$nuno > 0 && all(bws$bw[bws$iuno] == 0)) {
    p <- pmin(pmax(myout$dens, 0), 1)
    myout$derr <- sqrt(p * (1 - p) / tnrow)
  }

  fit.elapsed <- proc.time()[3] - fit.start
  optim.time <- if (!is.null(bws$total.time) && is.finite(bws$total.time)) as.double(bws$total.time) else NA_real_
  total.time <- fit.elapsed + (if (is.na(optim.time)) 0.0 else optim.time)

  ev <- npdensity(bws=bws, eval=teval, dens = myout$dens,
                  derr = myout$derr, ll = myout$log_likelihood,
                  ntrain = tnrow, trainiseval = no.e,
                  rows.omit = rows.omit,
                  timing = bws$timing, total.time = total.time,
                  optim.time = optim.time, fit.time = fit.elapsed)
  return(ev)
}

npudens.default <- function(bws, tdat, ...){
  sc <- sys.call()
  sc.names <- names(sc)

  ## here we check to see if the function was called with tdat =
  ## if it was, we need to catch that and map it to dat =
  ## otherwise the call is passed unadulterated to npudensbw

  bws.named <- any(sc.names == "bws")
  tdat.named <- any(sc.names == "tdat")

  no.bws <- missing(bws)
  no.tdat <- missing(tdat)
  has.explicit.bws <- (!no.bws) && isa(bws, "bandwidth")
  bws.formula <- (!no.bws) && inherits(bws, "formula")

  if (bws.named && no.tdat && bws.formula) {
    sc$`bws` <- NULL
    sc$formula <- bws
    sc.bw <- sc
    sc.bw[[1]] <- quote(npudensbw)
    bws.named <- FALSE
  } else {
    sc.bw <- sc
    sc.bw[[1]] <- quote(npudensbw)
  }

  ## if bws was passed in explicitly, do not compute bandwidths
    
  if(tdat.named)
    tdat <- toFrame(tdat)

  if(bws.named){
    sc.bw$bandwidth.compute <- FALSE
  }

  ostxy <- c('tdat')
  nstxy <- c('dat')
  
  m.txy <- match(ostxy, names(sc.bw), nomatch = 0)

  if(any(m.txy > 0)) {
    names(sc.bw)[m.txy] <- nstxy[m.txy > 0]
  }
    
  tbw <- if (!has.explicit.bws) {
    .np_progress_select_bandwidth_enhanced(
      "Selecting density bandwidth",
      .np_eval_bw_call(sc.bw, caller_env = parent.frame())
    )
  } else {
    .np_eval_bw_call(sc.bw, caller_env = parent.frame())
  }

  ## convention: first argument is always dropped, second, if present, propagated
  call.args <- list(bws = tbw)
  if (!no.tdat) {
    if (tdat.named) {
      call.args$tdat <- tdat
    } else {
      call.args <- c(call.args, list(tdat))
    }
  }
  if (!has.explicit.bws)
    call.args$.np_fit_progress_handoff <- TRUE
  do.call(npudens, c(call.args, list(...)))
}

densdist_fit_progress_time_counter <- function(start = 0, by = 1.7) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

densdist_fit_progress_lines <- function(shadow) {
  vapply(shadow$trace, `[[`, character(1L), "line")
}

make_densdist_fit_progress_fixture <- function() {
  set.seed(20260404)
  n <- 24L
  dat <- data.frame(x = seq(-0.9, 0.9, length.out = n))

  list(
    dat = dat,
    tdat = dat["x"],
    n = n,
    dens.bw = npudensbw(
      dat = dat["x"],
      bws = 0.24,
      bandwidth.compute = FALSE
    ),
    dist.bw = npudistbw(
      dat = dat["x"],
      bws = 0.24,
      bandwidth.compute = FALSE
    )
  )
}

test_that("npudens direct bws fit emits single-line fit progress", {
  fixture <- make_densdist_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0,
    np.progress.interval.unknown.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npudens(
      bws = fixture$dens.bw,
      tdat = fixture$tdat
    ),
    force_renderer = "single_line",
    now = densdist_fit_progress_time_counter()
  )

  lines <- densdist_fit_progress_lines(actual)

  expect_s3_class(actual$value, "npdensity")
  expect_true(any(grepl(
    sprintf("^\\[np\\] Fitting density 1/%d \\([0-9]+\\.[0-9]%%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", fixture$n),
    lines
  )))
  expect_true(any(grepl(
    sprintf("^\\[np\\] Fitting density %d/%d \\(100\\.0%%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$", fixture$n, fixture$n),
    lines
  )))
})

test_that("npudens bw to fit route hands off immediately into single-line fit progress", {
  fixture <- make_densdist_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npudens(
      tdat = fixture$tdat,
      nmulti = 2L
    ),
    force_renderer = "single_line",
    now = densdist_fit_progress_time_counter()
  )

  lines <- densdist_fit_progress_lines(actual)
  bandwidth.pos <- grep("^\\[np\\] Bandwidth selection \\(", lines)
  fit.start.pos <- grep(
    sprintf("^\\[np\\] Fitting density 0/%d \\(0\\.0%%, elapsed 0\\.0s, eta 0\\.0s\\): starting$", fixture$n),
    lines
  )
  fit.finish.pos <- grep(
    sprintf("^\\[np\\] Fitting density %d/%d \\(100\\.0%%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$", fixture$n, fixture$n),
    lines
  )

  expect_s3_class(actual$value, "npdensity")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(fit.start.pos) == 1L)
  expect_true(length(fit.finish.pos) >= 1L)
  expect_lt(max(bandwidth.pos), fit.start.pos[[1L]])
  expect_lt(fit.start.pos[[1L]], fit.finish.pos[[1L]])
})

test_that("npudist direct bws fit emits single-line fit progress", {
  fixture <- make_densdist_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0,
    np.progress.interval.unknown.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npudist(
      bws = fixture$dist.bw,
      tdat = fixture$tdat
    ),
    force_renderer = "single_line",
    now = densdist_fit_progress_time_counter()
  )

  lines <- densdist_fit_progress_lines(actual)

  expect_s3_class(actual$value, "npdistribution")
  expect_true(any(grepl(
    sprintf("^\\[np\\] Fitting distribution 1/%d \\([0-9]+\\.[0-9]%%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", fixture$n),
    lines
  )))
  expect_true(any(grepl(
    sprintf("^\\[np\\] Fitting distribution %d/%d \\(100\\.0%%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$", fixture$n, fixture$n),
    lines
  )))
})

test_that("npudist bw to fit route hands off immediately into single-line fit progress", {
  fixture <- make_densdist_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npudist(
      tdat = fixture$tdat,
      nmulti = 2L
    ),
    force_renderer = "single_line",
    now = densdist_fit_progress_time_counter()
  )

  lines <- densdist_fit_progress_lines(actual)
  bandwidth.pos <- grep("^\\[np\\] Bandwidth selection \\(", lines)
  fit.start.pos <- grep(
    sprintf("^\\[np\\] Fitting distribution 0/%d \\(0\\.0%%, elapsed 0\\.0s, eta 0\\.0s\\): starting$", fixture$n),
    lines
  )
  fit.finish.pos <- grep(
    sprintf("^\\[np\\] Fitting distribution %d/%d \\(100\\.0%%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$", fixture$n, fixture$n),
    lines
  )

  expect_s3_class(actual$value, "npdistribution")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(fit.start.pos) == 1L)
  expect_true(length(fit.finish.pos) >= 1L)
  expect_lt(max(bandwidth.pos), fit.start.pos[[1L]])
  expect_lt(fit.start.pos[[1L]], fit.finish.pos[[1L]])
})

test_that("bw generics route named data args without bws to NULL methods", {
  set.seed(20260222)
  x <- rnorm(30)
  y <- x + rnorm(30)
  dat <- data.frame(x = x, y = y)

  bw_reg <- npregbw(xdat = x, ydat = y, regtype = "lc", bwmethod = "cv.aic", nmulti = 1)
  bw_dens <- npudensbw(dat = dat, bwmethod = "normal-reference")
  bw_dist <- npudistbw(dat = dat, bwmethod = "normal-reference")

  expect_s3_class(bw_reg, "rbandwidth")
  expect_s3_class(bw_dens, "bandwidth")
  expect_s3_class(bw_dist, "dbandwidth")
})

test_that("bw generics route unnamed data args without bws to NULL methods", {
  set.seed(20260501)
  x <- rnorm(30)
  y <- x + rnorm(30)
  dat <- data.frame(x = x, y = y)

  bw_reg <- npregbw(x, y, regtype = "lc", bwmethod = "cv.aic", nmulti = 1)
  bw_dens <- npudensbw(dat, bwmethod = "normal-reference")
  bw_dist <- npudistbw(dat, bwmethod = "normal-reference")

  expect_s3_class(bw_reg, "rbandwidth")
  expect_s3_class(bw_dens, "bandwidth")
  expect_s3_class(bw_dist, "dbandwidth")
})

test_that("bw object dispatch remains intact", {
  set.seed(20260222)
  x <- rnorm(25)
  y <- x + rnorm(25)
  dat <- data.frame(x = x, y = y)

  bw_reg0 <- npregbw(xdat = x, ydat = y, bws = 0.5, bandwidth.compute = FALSE, regtype = "lc")
  bw_reg1 <- npregbw(bws = bw_reg0, xdat = x, ydat = y, bandwidth.compute = FALSE, regtype = "lc")
  expect_equal(as.numeric(bw_reg1$bw), as.numeric(bw_reg0$bw))

  bw_dens0 <- npudensbw(dat = dat, bws = c(0.4, 0.6), bandwidth.compute = FALSE)
  bw_dens1 <- npudensbw(bws = bw_dens0, dat = dat, bandwidth.compute = FALSE)
  expect_equal(as.numeric(bw_dens1$bw), as.numeric(bw_dens0$bw))

  bw_dist0 <- npudistbw(dat = dat, bws = c(0.4, 0.6), bandwidth.compute = FALSE)
  bw_dist1 <- npudistbw(bws = bw_dist0, dat = dat, bandwidth.compute = FALSE)
  expect_equal(as.numeric(bw_dist1$bw), as.numeric(bw_dist0$bw))
})

test_that("formula bandwidth reuse preserves explicit bws objects", {
  set.seed(20260501)
  n <- 30L
  x <- rnorm(n)
  y <- x + rnorm(n)
  dat <- data.frame(y = y, x = x, g = factor(sample(0:1, n, TRUE)))

  bw_dens0 <- npudensbw(~ x, data = dat, bwmethod = "normal-reference")
  bw_dens1 <- npudensbw(~ x, data = dat, bws = bw_dens0,
                        bandwidth.compute = FALSE)
  expect_s3_class(bw_dens1, "bandwidth")
  expect_equal(as.numeric(bw_dens1$bw), as.numeric(bw_dens0$bw))

  bw_cdens0 <- npcdensbw(y ~ g, data = dat, bwtype = "fixed", nmulti = 1)
  bw_cdens1 <- npcdensbw(y ~ g, data = dat, bws = bw_cdens0,
                         bandwidth.compute = FALSE)
  expect_s3_class(bw_cdens1, "conbandwidth")
  expect_equal(as.numeric(bw_cdens1$xbw), as.numeric(bw_cdens0$xbw))
  expect_equal(as.numeric(bw_cdens1$ybw), as.numeric(bw_cdens0$ybw))
})

test_that("formula estimator fronts tolerate omitted legacy regtype defaults", {
  set.seed(20260322)
  n <- 40L
  x1 <- runif(n)
  x2 <- runif(n)
  y <- x1^2 + rnorm(n, sd = 0.1)

  expect_s3_class(npreg(y ~ x1 + x2, nmulti = 1), "npregression")
  expect_s3_class(npudens(~ y + x1, nmulti = 1), "npdensity")
  expect_s3_class(npscoef(y ~ x1 | x2, nmulti = 1), "smoothcoefficient")
  expect_s3_class(npindex(y ~ x1 + x2, nmulti = 1), "singleindex")
  expect_s3_class(npplreg(y ~ x1 | x2, nmulti = 1), "plregression")
})

test_that("npregbw preserves deprecated remin compatibility", {
  set.seed(20260515)
  n <- 40L
  x <- rnorm(n)
  y <- x + rnorm(n)

  set.seed(20260516)
  expect_warning(
    bw.legacy <- npregbw(y ~ x, nmulti = 1, remin = FALSE),
    "deprecated"
  )
  set.seed(20260516)
  bw.modern <- npregbw(y ~ x, nmulti = 1, powell.remin = FALSE)

  expect_equal(as.numeric(bw.legacy$bw), as.numeric(bw.modern$bw),
               tolerance = 1e-12)
  expect_warning(
    fit.legacy <- npreg(y ~ x, nmulti = 1, remin = FALSE),
    "deprecated"
  )
  expect_s3_class(fit.legacy, "npregression")
})

library(np)

collect_plot_payload_fields <- function(x, path = character()) {
  out <- list()

  if (is.null(x)) {
    return(out)
  }

  if (is.data.frame(x)) {
    x <- as.list(x)
  }

  if (is.list(x)) {
    nm <- names(x)
    if (is.null(nm)) {
      nm <- as.character(seq_along(x))
    }

    for (i in seq_along(x)) {
      out <- c(out, collect_plot_payload_fields(x[[i]], c(path, nm[i])))
    }
    return(out)
  }

  if (!is.numeric(x)) {
    return(out)
  }

  leaf <- if (length(path)) path[[length(path)]] else ""
  if (!(leaf %in% c("mean", "merr", "evalx", "evaly", "evalz"))) {
    return(out)
  }

  out[[paste(path, collapse = ".")]] <- as.numeric(x)
  out
}

run_exact_frozen_plot_pair <- function(fit, ..., seed = 9001L) {
  extra <- list(...)
  exact <- NULL
  frozen <- NULL

  set.seed(seed)
  suppressWarnings(capture.output(
    exact <- do.call(plot, c(list(fit), extra, list(boot_control = np_boot_control(nonfixed = "exact"))))
  ))

  set.seed(seed)
  suppressWarnings(capture.output(
    frozen <- do.call(plot, c(list(fit), extra, list(boot_control = np_boot_control(nonfixed = "frozen"))))
  ))

  list(exact = exact, frozen = frozen)
}

expect_plot_payload_comparable <- function(pair,
                                           label,
                                           min.merr.corr,
                                           max.merr.rel) {
  exact.fields <- collect_plot_payload_fields(pair$exact)
  frozen.fields <- collect_plot_payload_fields(pair$frozen)

  expect_equal(sort(names(exact.fields)), sort(names(frozen.fields)), info = label)

  for (nm in names(exact.fields)) {
    exact.val <- exact.fields[[nm]]
    frozen.val <- frozen.fields[[nm]]
    leaf <- sub("^.*\\.", "", nm)

    expect_equal(length(frozen.val), length(exact.val), info = sprintf("%s %s length", label, nm))
    if (leaf %in% c("mean", "evalx", "evaly", "evalz")) {
      expect_true(all(is.finite(exact.val)), info = sprintf("%s %s exact finite", label, nm))
      expect_true(all(is.finite(frozen.val)), info = sprintf("%s %s frozen finite", label, nm))
      expect_equal(frozen.val, exact.val, tolerance = 1e-10, info = sprintf("%s %s", label, nm))
      next
    }

    expect_equal(is.na(frozen.val), is.na(exact.val), info = sprintf("%s %s NA mask", label, nm))
    keep <- is.finite(exact.val) & is.finite(frozen.val)
    if (!any(keep)) {
      next
    }

    exact.keep <- exact.val[keep]
    frozen.keep <- frozen.val[keep]
    scale <- max(1e-8, max(abs(exact.keep)))
    rel.max <- max(abs(frozen.keep - exact.keep)) / scale
    corr <- if (length(exact.keep) > 1L &&
                stats::sd(exact.keep) > 0 &&
                stats::sd(frozen.keep) > 0) {
      stats::cor(exact.keep, frozen.keep)
    } else {
      1
    }

    expect_true(corr >= min.merr.corr, info = sprintf("%s %s corr", label, nm))
    expect_true(rel.max <= max.merr.rel, info = sprintf("%s %s rel", label, nm))
  }
}

test_that("exact and frozen plot payloads stay comparable for regression and semiparametric families", {
  set.seed(20260322)

  n.reg <- 70L
  x.reg <- runif(n.reg, -1, 1)
  y.reg <- x.reg + rnorm(n.reg, sd = 0.15)
  reg.fit <- npreg(y.reg ~ x.reg,
                   nmulti = 1,
                   regtype = "lp",
                   degree = 1,
                   bwtype = "adaptive_nn")
  reg.pair <- run_exact_frozen_plot_pair(
    reg.fit,
    view = "fixed",
    gradients = TRUE,
    output = "data",
    errors = "bootstrap",
    bootstrap = "inid",
    B = 41L,
    band = "pointwise",
    neval = 8L
  )
  expect_plot_payload_comparable(reg.pair, "npreg", min.merr.corr = 0.90, max.merr.rel = 0.75)

  n.idx <- 70L
  x.idx <- runif(n.idx, -1, 1)
  z.idx <- rnorm(n.idx)
  y.idx <- x.idx + 0.5 * z.idx + rnorm(n.idx, sd = 0.15)
  idx.fit <- npindex(y.idx ~ x.idx + z.idx,
                     nmulti = 1,
                     gradients = TRUE,
                     bwtype = "adaptive_nn")
  idx.pair <- run_exact_frozen_plot_pair(
    idx.fit,
    view = "fixed",
    output = "data",
    errors = "bootstrap",
    bootstrap = "inid",
    B = 41L,
    band = "pointwise",
    neval = 8L
  )
  expect_plot_payload_comparable(idx.pair, "npindex", min.merr.corr = 0.90, max.merr.rel = 0.80)

  n.pl <- 75L
  x.pl <- runif(n.pl, -1, 1)
  z.pl <- rnorm(n.pl)
  y.pl <- x.pl^2 + z.pl + rnorm(n.pl, sd = 0.2)
  pl.fit <- npplreg(y.pl ~ x.pl | z.pl, nmulti = 1, bwtype = "adaptive_nn")
  pl.pair <- run_exact_frozen_plot_pair(
    pl.fit,
    view = "fixed",
    output = "data",
    errors = "bootstrap",
    bootstrap = "geom",
    B = 41L,
    band = "pointwise",
    neval = 8L
  )
  expect_plot_payload_comparable(pl.pair, "npplreg", min.merr.corr = 0.70, max.merr.rel = 0.55)

  n.sc <- 75L
  x.sc <- runif(n.sc, -1, 1)
  z.sc <- rnorm(n.sc)
  y.sc <- x.sc^2 + z.sc + rnorm(n.sc, sd = 0.2 * stats::sd(x.sc))
  sc.fit <- npscoef(y.sc ~ x.sc | z.sc,
                    nmulti = 1,
                    regtype = "ll",
                    bwtype = "adaptive_nn")
  sc.pair <- run_exact_frozen_plot_pair(
    sc.fit,
    view = "fixed",
    coef = FALSE,
    output = "data",
    errors = "bootstrap",
    bootstrap = "inid",
    B = 41L,
    band = "pointwise",
    neval = 8L
  )
  expect_plot_payload_comparable(sc.pair, "npscoef", min.merr.corr = 0.95, max.merr.rel = 0.35)
})

test_that("exact and frozen plot payloads stay comparable for density and distribution families", {
  set.seed(20260323)

  n.u <- 75L
  y.u <- runif(n.u, -1, 1) + rnorm(n.u, sd = 0.1)
  ud.fit <- npudens(~ y.u, nmulti = 1, bwtype = "adaptive_nn")
  ud.pair <- run_exact_frozen_plot_pair(
    ud.fit,
    output = "data",
    errors = "bootstrap",
    bootstrap = "inid",
    B = 41L,
    band = "pointwise",
    neval = 8L
  )
  expect_plot_payload_comparable(ud.pair, "npudens", min.merr.corr = 0.95, max.merr.rel = 0.75)

  uf.fit <- npudist(~ y.u, nmulti = 1, bwtype = "adaptive_nn")
  uf.pair <- run_exact_frozen_plot_pair(
    uf.fit,
    output = "data",
    errors = "bootstrap",
    bootstrap = "inid",
    B = 41L,
    band = "pointwise",
    neval = 8L
  )
  expect_plot_payload_comparable(uf.pair, "npudist", min.merr.corr = 0.95, max.merr.rel = 0.45)

  n.c <- 75L
  x.c <- runif(n.c, -1, 1)
  y.c <- x.c + rnorm(n.c, sd = 0.2)
  c.dat <- data.frame(x = x.c, y = y.c)
  cd.fit <- npcdens(y ~ x, data = c.dat, nmulti = 1, bwtype = "generalized_nn")
  cd.pair <- run_exact_frozen_plot_pair(
    cd.fit,
    view = "fixed",
    output = "data",
    errors = "bootstrap",
    bootstrap = "inid",
    B = 41L,
    band = "pointwise",
    neval = 8L
  )
  expect_plot_payload_comparable(cd.pair, "npcdens", min.merr.corr = 0.55, max.merr.rel = 1.20)

  cf.fit <- npcdist(y ~ x, data = c.dat, nmulti = 1, bwtype = "generalized_nn")
  cf.pair <- run_exact_frozen_plot_pair(
    cf.fit,
    view = "fixed",
    output = "data",
    errors = "bootstrap",
    bootstrap = "inid",
    B = 41L,
    band = "pointwise",
    neval = 8L
  )
  expect_plot_payload_comparable(cf.pair, "npcdist", min.merr.corr = 0.90, max.merr.rel = 0.75)
})

library(np)

test_that("bounded adaptive_nn is available for unconditional density", {
  set.seed(20260325)
  x <- runif(36)
  xy <- data.frame(x = x)

  bw.fixed <- npudensbw(
    dat = xy,
    bwmethod = "cv.ml",
    bwtype = "fixed",
    ckerbound = "range",
    nmulti = 1
  )
  bw.generalized <- npudensbw(
    dat = xy,
    bwmethod = "cv.ml",
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1
  )
  bw.adaptive <- npudensbw(
    dat = xy,
    bwmethod = "cv.ml",
    bwtype = "adaptive_nn",
    ckerbound = "range",
    nmulti = 1
  )

  fit.adaptive <- npudens(bws = bw.adaptive, tdat = xy)

  expect_true(all(is.finite(as.numeric(bw.adaptive$bw))))
  expect_true(is.finite(bw.adaptive$fval))
  expect_true(all(is.finite(as.numeric(fit.adaptive$dens))))

  expect_true(all(is.finite(as.numeric(bw.fixed$bw))))
  expect_true(is.finite(bw.fixed$fval))
  expect_true(all(is.finite(as.numeric(bw.generalized$bw))))
  expect_true(is.finite(bw.generalized$fval))

  expect_false(isTRUE(all.equal(as.numeric(bw.fixed$bw), as.numeric(bw.adaptive$bw))))
  expect_false(isTRUE(all.equal(as.numeric(bw.generalized$bw), as.numeric(bw.adaptive$bw))))
})

library(np)

test_that("bounded gaussian kernel matches direct normalization at fixed bandwidth", {
  x <- sort(c(5, 11, 21, 31, 46, 75, 98, 122, 145, 165, 195, 224, 245, 293, 321, 330, 350, 420))
  h <- 100
  a <- min(x)
  b <- max(x)

  bw <- npudensbw(dat = data.frame(x = x),
                  bws = h,
                  bandwidth.compute = FALSE,
                  ckertype = "gaussian",
                  ckerbound = "range")
  fit <- npudens(tdat = data.frame(x = x), bws = bw)

  fdirect <- sapply(x, function(xx) {
    mean(dnorm((x - xx)/h)/(h * (pnorm((b - xx)/h) - pnorm((a - xx)/h))))
  })

  expect_equal(as.numeric(fit$dens), as.numeric(fdirect), tolerance = 1e-7)
})

test_that("cv bandwidth with range bounds is larger than unbounded for stress example", {
  x <- sort(c(5, 11, 21, 31, 46, 75, 98, 122, 145, 165, 195, 224, 245, 293, 321, 330, 350, 420))

  bw_bound <- npudensbw(dat = data.frame(x = x),
                        bwmethod = "cv.ml",
                        ckertype = "gaussian",
                        ckerbound = "range",
                        nmulti = 1)

  bw_inf <- npudensbw(dat = data.frame(x = x),
                      bwmethod = "cv.ml",
                      ckertype = "gaussian",
                      ckerbound = "fixed",
                      ckerlb = -Inf,
                      ckerub = Inf,
                      nmulti = 1)

  expect_gt(as.numeric(bw_bound$bw[1]), as.numeric(bw_inf$bw[1]))
})

progress_time_counter <- function(start = 0, by = 1.1) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

shadow_lines <- function(shadow) {
  vapply(shadow$trace, `[[`, character(1L), "line")
}

shadow_signature <- function(shadow) {
  trace <- shadow$trace[vapply(shadow$trace, `[[`, character(1L), "event") == "render"]
  data.frame(
    event = vapply(trace, `[[`, character(1L), "event"),
    line = vapply(trace, `[[`, character(1L), "line"),
    stringsAsFactors = FALSE
  )
}

shadow_stage_signature <- function(shadow) {
  sig <- shadow_signature(shadow)
  sig$line <- sub(", elapsed .*", "", sig$line)
  sig
}

test_that("npudensbw bandwidth progress uses the enhanced multistart handoff", {
  set.seed(42)
  x <- rnorm(35)

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  legacy <- capture_progress_shadow_trace(
    npudensbw(
      dat = data.frame(x = x),
      bwmethod = "cv.ml",
      nmulti = 3
    ),
    force_renderer = "legacy",
    now = progress_time_counter()
  )

  set.seed(42)
  single_line <- capture_progress_shadow_trace(
    npudensbw(
      dat = data.frame(x = x),
      bwmethod = "cv.ml",
      nmulti = 3
    ),
    force_renderer = "single_line",
    now = progress_time_counter()
  )

  lines <- shadow_lines(single_line)

  expect_s3_class(single_line$value, "bandwidth")
  expect_equal(shadow_stage_signature(single_line), shadow_stage_signature(legacy))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 1/3\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/3, elapsed [0-9]+\\.[0-9]s, [0-9]+\\.[0-9]%, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 3/3, elapsed [0-9]+\\.[0-9]s, [0-9]+\\.[0-9]%, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 3/3, elapsed [0-9]+\\.[0-9]s, 100\\.0%, eta 0\\.0s\\)$", lines)))
})

test_that("npudens indirect entry inherits the enhanced density bandwidth progress", {
  set.seed(101)
  x <- rnorm(35)

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npudens(
      tdat = data.frame(x = x),
      bwmethod = "cv.ml",
      nmulti = 3
    ),
    force_renderer = "single_line",
    now = progress_time_counter()
  )

  lines <- shadow_lines(actual)

  expect_s3_class(actual$value, "npdensity")
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 1/3\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/3, elapsed [0-9]+\\.[0-9]s, [0-9]+\\.[0-9]%, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 3/3, elapsed [0-9]+\\.[0-9]s, 100\\.0%, eta 0\\.0s\\)$", lines)))
})

test_that("npregbw adopts the generic bandwidth selection line", {
  set.seed(7)
  x <- runif(30)
  y <- sin(2 * pi * x) + rnorm(30, sd = 0.1)

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  legacy <- capture_progress_shadow_trace(
    npregbw(
      xdat = data.frame(x = x),
      ydat = y,
      regtype = "lc",
      bwmethod = "cv.aic",
      nmulti = 3
    ),
    force_renderer = "legacy",
    now = progress_time_counter()
  )

  set.seed(7)
  single_line <- capture_progress_shadow_trace(
    npregbw(
      xdat = data.frame(x = x),
      ydat = y,
      regtype = "lc",
      bwmethod = "cv.aic",
      nmulti = 3
    ),
    force_renderer = "single_line",
    now = progress_time_counter()
  )

  lines <- shadow_lines(single_line)

  expect_s3_class(single_line$value, "rbandwidth")
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 1/3\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/3, elapsed [0-9]+\\.[0-9]s, [0-9]+\\.[0-9]%, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 3/3, elapsed [0-9]+\\.[0-9]s, 100\\.0%, eta 0\\.0s\\)$", lines)))
})

test_that(".Call wrappers tolerate empty continuous bounds", {
  set.seed(20260222)
  x <- rnorm(60)
  y <- x + rnorm(60)

  bw_reg <- npregbw(xdat = x, ydat = y, bws = 0.5, bandwidth.compute = FALSE, regtype = "lc")
  fit_reg <- npreg(bws = bw_reg)
  expect_s3_class(fit_reg, "npregression")

  bw_dens <- npudensbw(dat = data.frame(x = x), bws = 0.6, bandwidth.compute = FALSE)
  fit_dens <- npudens(tdat = data.frame(x = x), bws = bw_dens,
                      ckerlb = numeric(0), ckerub = numeric(0))
  expect_s3_class(fit_dens, "npdensity")

  bw_dist <- npudistbw(dat = data.frame(x = x), bws = 0.6, bandwidth.compute = FALSE)
  fit_dist <- npudist(tdat = data.frame(x = x), bws = bw_dist,
                      ckerlb = numeric(0), ckerub = numeric(0))
  expect_s3_class(fit_dist, "npdistribution")

  bw_cd <- npcdensbw(xdat = data.frame(x = x), ydat = data.frame(y = y),
                     bws = c(0.7, 0.5), bandwidth.compute = FALSE)
  fit_cd <- npcdens(
    txdat = data.frame(x = x), tydat = data.frame(y = y), bws = bw_cd,
    cxkerlb = numeric(0), cxkerub = numeric(0),
    cykerlb = numeric(0), cykerub = numeric(0)
  )
  expect_s3_class(fit_cd, "condensity")
})

test_that("npudens exposes unconditional fast counts in bandwidth summary", {
  set.seed(42)
  n <- 100L
  y <- runif(n)

  fit <- npudens(
    ~ y,
    ckerbound = "fixed",
    ckerlb = 0,
    ckerub = 1,
    nomad = TRUE
  )

  expect_true(is.finite(as.numeric(fit$bws$num.feval[1L])))
  expect_true(is.finite(as.numeric(fit$bws$num.feval.fast[1L])))
  expect_gt(as.numeric(fit$bws$num.feval.fast[1L]), 0)
  expect_lte(as.numeric(fit$bws$num.feval.fast[1L]), as.numeric(fit$bws$num.feval[1L]))

  txt <- paste(capture.output(summary(fit$bws)), collapse = "\n")
  expect_true(grepl("Number of Function Evaluations:", txt, fixed = TRUE))
  expect_true(grepl("fast =", txt, fixed = TRUE))
  expect_true(grepl(sprintf("fast = %s", format(fit$bws$num.feval.fast[1L])), txt, fixed = TRUE))
})

test_that("predict returns log-likelihood metadata for npudens and npcdens", {
  set.seed(20260323)

  x <- data.frame(x = c(-1.2, -0.5, 0.0, 0.4, 0.9, 1.4))
  y <- c(-0.9, -0.2, 0.1, 0.5, 1.0, 1.6)

  bw.u <- npudensbw(dat = x, bws = 0.6, bandwidth.compute = FALSE)
  fit.u <- npudens(bws = bw.u, tdat = x)
  out.u <- predict(fit.u, se.fit = TRUE)

  expect_false(is.null(out.u$log.likelihood))
  expect_equal(out.u$log.likelihood, fit.u$log_likelihood)

  bw.c <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.6, 0.7),
    bandwidth.compute = FALSE
  )
  fit.c <- npcdens(bws = bw.c, txdat = x, tydat = y)
  out.c <- predict(fit.c, se.fit = TRUE)

  expect_false(is.null(out.c$log.likelihood))
  expect_equal(out.c$log.likelihood, fit.c$log_likelihood)
})

test_that("npudens default constructs bandwidth call in caller frame", {
  run_case <- function() {
    set.seed(101)
    x <- runif(40)
    npudens(tdat = x, bwmethod = "normal-reference")
  }

  fit <- run_case()
  expect_s3_class(fit, "npdensity")
  expect_true(any(grepl("bandwidth$", class(fit$bws))))
  expect_equal(NROW(fit$dens), NROW(fit$eval))
})

test_that("npudist default constructs bandwidth call in caller frame", {
  run_case <- function() {
    set.seed(202)
    x <- runif(45)
    npudist(tdat = x, bwmethod = "normal-reference")
  }

  fit <- run_case()
  expect_s3_class(fit, "npdistribution")
  expect_true(any(grepl("bandwidth$", class(fit$bws))))
  expect_equal(length(fit$dist), NROW(fit$eval))
})

test_that("npreg default constructs bandwidth call in caller frame", {
  run_case <- function() {
    set.seed(303)
    x <- runif(55)
    y <- sin(2 * pi * x) + rnorm(55, sd = 0.2)
    npreg(txdat = x, tydat = y, regtype = "lc", bwmethod = "cv.ls", nmulti = 1)
  }

  fit <- run_case()
  expect_s3_class(fit, "npregression")
  expect_true(any(grepl("bandwidth$", class(fit$bws))))
  expect_equal(length(fit$mean), NROW(fit$eval))
})

test_that("npreg default bandwidth selection remains finite after bounded conditional selectors", {
  set.seed(20260316)
  n <- 48
  x <- runif(n)
  y <- runif(n)

  bw.cd <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bwmethod = "cv.ls",
    bwtype = "fixed",
    cxkertype = "gaussian",
    cykertype = "gaussian",
    cxkerorder = 2L,
    cykerorder = 2L,
    cxkerbound = "range",
    cykerbound = "range",
    nmulti = 1
  )

  expect_true(all(is.finite(as.numeric(bw.cd$xbw))))
  expect_true(all(is.finite(as.numeric(bw.cd$ybw))))

  fit.reg <- npreg(
    txdat = x,
    tydat = sin(2 * pi * x) + rnorm(n, sd = 0.15),
    regtype = "lc",
    bwmethod = "cv.ls",
    nmulti = 1
  )

  expect_s3_class(fit.reg, "npregression")
  expect_true(any(grepl("bandwidth$", class(fit.reg$bws))))
  expect_equal(length(fit.reg$mean), NROW(fit.reg$eval))
})

library(np)

bounded_gaussian_kernel <- function(eval_x, train_x, h, lb, ub, order) {
  z <- (train_x - eval_x) / h
  pdf <- switch(
    as.character(as.integer(order)),
    `2` = dnorm(z),
    `4` = (1.5 - 0.5 * z^2) * dnorm(z),
    stop("unsupported order")
  )
  cdf <- switch(
    as.character(as.integer(order)),
    `2` = pnorm,
    `4` = function(v) pnorm(v) + 0.5 * v * dnorm(v),
    stop("unsupported order")
  )
  denom <- h * (cdf((ub - eval_x) / h) - cdf((lb - eval_x) / h))
  pdf / denom
}

reconstruct_npudens_cvls_fixed <- function(x, bw, q = 1201L) {
  stopifnot(
    identical(bw$method, "cv.ls"),
    identical(bw$type, "fixed"),
    identical(bw$ckertype, "gaussian"),
    bw$ncon == 1L,
    bw$nuno == 0L,
    bw$nord == 0L
  )

  h <- as.numeric(bw$bw[1])
  lb <- as.numeric(bw$ckerlb[bw$icon][1])
  ub <- as.numeric(bw$ckerub[bw$icon][1])
  grid <- seq(lb, ub, length.out = q)
  trap <- diff(grid)[1] * c(0.5, rep(1, q - 2L), 0.5)

  fit <- npudens(
    bws = bw,
    tdat = data.frame(x = x),
    edat = data.frame(x = grid)
  )

  i1 <- sum(trap * as.numeric(fit$dens)^2)
  i2 <- mean(vapply(seq_along(x), function(i) {
    mean(bounded_gaussian_kernel(x[i], x[-i], h, lb, ub, bw$ckerorder))
  }, numeric(1)))

  list(i1 = i1, i2 = i2, objective = i1 - 2 * i2)
}

test_that("phase1 npudensbw cv.ls fixed bounded objective matches direct reconstruction", {
  set.seed(20260421)
  x <- rbeta(64L, 2, 5)
  dat <- data.frame(x = x)

  for (ord in c(2L, 4L)) {
    bw <- npudensbw(
      dat = dat,
      bwmethod = "cv.ls",
      bwtype = "fixed",
      ckertype = "gaussian",
      ckerorder = ord,
      ckerbound = "range",
      nmulti = 1
    )

    rec <- reconstruct_npudens_cvls_fixed(x, bw)

    expect_true(is.finite(as.numeric(bw$bw[1])))
    expect_true(as.numeric(bw$bw[1]) > 0)
    expect_equal(as.numeric(bw$fval), -rec$objective, tolerance = 1e-2)
  }
})

library(np)

test_that("npudenshat matches npudens and preserves matrix/apply parity across bwtypes", {
  npudenshat <- getFromNamespace("npudenshat", "np")

  set.seed(20260310)
  n <- 60
  x <- sort(runif(n))
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(0.05, 0.95, length.out = 21))
  iota <- rep.int(1, n)

  for (bwtype in c("fixed", "generalized_nn", "adaptive_nn")) {
    bw <- npudensbw(
      dat = tx,
      bwtype = bwtype,
      bws = if (identical(bwtype, "fixed")) 0.18 else 9,
      bandwidth.compute = FALSE
    )

    fit.in <- npudens(bws = bw, tdat = tx)
    fit.ex <- npudens(bws = bw, tdat = tx, edat = ex)

    H.in <- npudenshat(bws = bw, tdat = tx, output = "matrix")
    H.ex <- npudenshat(bws = bw, tdat = tx, edat = ex, output = "matrix")

    apply.in <- npudenshat(bws = bw, tdat = tx, y = iota, output = "apply")
    apply.ex <- npudenshat(bws = bw, tdat = tx, edat = ex, y = iota, output = "apply")

    expect_s3_class(H.in, "npudenshat")
    expect_equal(as.vector(H.in %*% iota), as.vector(fit.in$dens), tolerance = 1e-10, info = bwtype)
    expect_equal(as.vector(H.ex %*% iota), as.vector(fit.ex$dens), tolerance = 1e-10, info = bwtype)
    expect_equal(as.vector(apply.in), as.vector(H.in %*% iota), tolerance = 1e-12, info = bwtype)
    expect_equal(as.vector(apply.ex), as.vector(H.ex %*% iota), tolerance = 1e-12, info = bwtype)
  }
})

test_that("npudenshat fixed-bandwidth count vectors reproduce resampled npudens fits", {
  npudenshat <- getFromNamespace("npudenshat", "np")

  set.seed(20260311)
  n <- 45
  x <- sort(runif(n))
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(0.1, 0.9, length.out = 17))
  bw <- npudensbw(
    dat = tx,
    bws = 0.16,
    bwtype = "fixed",
    bandwidth.compute = FALSE
  )
  H <- npudenshat(bws = bw, tdat = tx, edat = ex, output = "matrix")
  counts <- rmultinom(n = 3L, size = n, prob = rep.int(1 / n, n))

  for (b in seq_len(ncol(counts))) {
    idx <- rep.int(seq_len(n), counts[, b])
    fit <- npudens(bws = bw, tdat = tx[idx, , drop = FALSE], edat = ex)
    apply.out <- npudenshat(
      bws = bw,
      tdat = tx,
      edat = ex,
      y = counts[, b],
      output = "apply"
    )

    expect_equal(as.vector(H %*% counts[, b]), as.vector(apply.out), tolerance = 1e-12, info = b)
    expect_equal(as.vector(apply.out), as.vector(fit$dens), tolerance = 1e-10, info = b)
  }
})

test_that("npudenshat fixed apply mode matches matrix RHS multiplication", {
  npudenshat <- getFromNamespace("npudenshat", "np")

  set.seed(20260311)
  n <- 48
  x <- sort(runif(n))
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(0.1, 0.9, length.out = 19))
  bw <- npudensbw(
    dat = tx,
    bws = 0.16,
    bwtype = "fixed",
    bandwidth.compute = FALSE
  )
  rhs <- cbind(seq_len(n) / n, sin(seq_len(n) / 7))
  H <- npudenshat(bws = bw, tdat = tx, edat = ex, output = "matrix")
  apply.out <- npudenshat(bws = bw, tdat = tx, edat = ex, y = rhs, output = "apply")

  expect_equal(apply.out, H %*% rhs, tolerance = 1e-12)
})

test_that("npudenshat preserves bounded gaussian manual-bandwidth semantics", {
  npudenshat <- getFromNamespace("npudenshat", "np")

  x <- sort(c(5, 11, 21, 31, 46, 75, 98, 122, 145, 165, 195, 224, 245, 293, 321, 330, 350, 420))
  tx <- data.frame(x = x)
  iota <- rep.int(1, nrow(tx))

  bw <- npudensbw(
    dat = tx,
    bws = 100,
    bandwidth.compute = FALSE,
    ckertype = "gaussian",
    ckerbound = "range"
  )

  fit <- npudens(bws = bw, tdat = tx)
  H <- npudenshat(bws = bw, tdat = tx, output = "matrix")
  apply.out <- npudenshat(bws = bw, tdat = tx, y = iota, output = "apply")

  expect_equal(as.vector(H %*% iota), as.vector(fit$dens), tolerance = 1e-7)
  expect_equal(as.vector(apply.out), as.vector(fit$dens), tolerance = 1e-7)
})

.np_plot_sibandwidth_engine <-
  function(bws,
           xdat,
           ydat,
           data = NULL,
           common.scale = TRUE,
           gradients = FALSE,
           main = NULL,
           type = NULL,
           cex.axis = NULL,
           cex.lab = NULL,
           cex.main = NULL,
           cex.sub = NULL,
           col = NULL,
           ylab = NULL,
           xlab = NULL,
           sub = NULL,
           ylim = NULL,
           xlim = NULL,
           lty = NULL,
           lwd = NULL,
           plot.behavior = c("plot","plot-data","data"),
           neval = 50,
           plot.errors.method = c("none","bootstrap","asymptotic"),
           plot.errors.boot.num = 1999,
           plot.errors.boot.method = c("wild", "inid", "fixed", "geom"),
           plot.errors.boot.nonfixed = c("exact", "frozen"),
           plot.errors.boot.wild = c("rademacher", "mammen"),
           plot.errors.boot.blocklen = NULL,
           plot.errors.center = c("estimate","bias-corrected"),
           plot.errors.type = c("pmzsd","pointwise","bonferroni","simultaneous","all"),
           plot.errors.alpha = 0.05,
           plot.errors.style = c("band","bar"),
           plot.errors.bar = c("|","I"),
           plot.errors.bar.num = NULL,
           plot.par.mfrow = TRUE,
           plot.rug = FALSE,
           ...,
           random.seed){

    engine.ctx <- .np_plot_engine_begin(plot.par.mfrow = plot.par.mfrow)
    on.exit(.np_plot_restore_par(engine.ctx$oldpar), add = TRUE)
    plot.par.mfrow <- engine.ctx$plot.par.mfrow

    miss.xy = c(missing(xdat),missing(ydat))
    xy <- .np_plot_resolve_xydat(bws = bws, xdat = xdat, ydat = ydat, miss.xy = miss.xy)
    xdat <- xy$xdat
    ydat <- xy$ydat


    if (missing(plot.errors.method) &
        any(!missing(plot.errors.boot.num), !missing(plot.errors.boot.method),
            !missing(plot.errors.boot.nonfixed),
            !missing(plot.errors.boot.blocklen))){
      stop(
        "errors must be set to \"bootstrap\" when bootstrap controls are supplied",
        call. = FALSE
      )
    }

    normalized.opts <- .np_plot_normalize_common_options(
      plot.behavior = plot.behavior,
      plot.errors.method = plot.errors.method,
      plot.errors.boot.method = plot.errors.boot.method,
      plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
      plot.errors.boot.wild = plot.errors.boot.wild,
      plot.errors.boot.blocklen = plot.errors.boot.blocklen,
      plot.errors.center = plot.errors.center,
      plot.errors.type = plot.errors.type,
      plot.errors.alpha = plot.errors.alpha,
      plot.errors.style = plot.errors.style,
      plot.errors.bar = plot.errors.bar,
      xdat = xdat,
      common.scale = common.scale,
      ylim = ylim,
      allow_asymptotic_quantile = FALSE
    )

    plot.behavior <- normalized.opts$plot.behavior
    plot.errors.method <- normalized.opts$plot.errors.method
    plot.errors.boot.method <- normalized.opts$plot.errors.boot.method
    plot.errors.boot.nonfixed <- normalized.opts$plot.errors.boot.nonfixed
    plot.errors.boot.wild <- normalized.opts$plot.errors.boot.wild
    plot.errors.boot.blocklen <- normalized.opts$plot.errors.boot.blocklen
    plot.errors.center <- normalized.opts$plot.errors.center
    plot.errors.type <- normalized.opts$plot.errors.type
    plot.errors.alpha <- normalized.opts$plot.errors.alpha
    plot.errors.style <- normalized.opts$plot.errors.style
    plot.errors.bar <- normalized.opts$plot.errors.bar
    common.scale <- normalized.opts$common.scale

    plot.errors = (plot.errors.method != "none")
    first.render <- .np_plot_first_render_state()
    on.exit(.np_plot_activity_end(first.render$activity), add = TRUE)
    plot.rug <- .np_plot_validate_rug_request(
      plot.rug = plot.rug,
      route = "plot.sibandwidth()",
      supported.route = TRUE,
      renderer = "base",
      reason = "supported base plot routes"
    )

    plot.layout <- .np_plot_layout_begin(
      plot.behavior = plot.behavior,
      plot.par.mfrow = plot.par.mfrow,
      mfrow = if (gradients) n2mfrow(bws$ndim) else c(1, 1)
    )

    plot.out = list()

    scalar_default <- function(value, default) {
      if (is.null(value)) default else value
    }

    dots <- list(...)
    plot.legend <- if (!is.null(dots$legend)) dots$legend else TRUE
    plot.user.args <- .np_plot_user_args(dots, "plot")
    bxp.user.args <- .np_plot_user_args(dots, "bxp")
    bxp.args <- bxp.user.args
    if (!is.null(col)) bxp.args$col <- col
    if (!is.null(lty)) bxp.args$lty <- lty
    if (!is.null(lwd)) bxp.args$lwd <- lwd

    make_singleindex_payload <- function(index,
                                         mean,
                                         grad = NA,
                                         gradients.flag = FALSE,
                                         trainiseval = FALSE) {
      do.call(singleindex, list(
        bws = bws,
        index = index,
        mean = mean,
        grad = grad,
        ntrain = nrow(xdat),
        trainiseval = trainiseval,
        gradients = gradients.flag
      ))
    }

    xdat <- toFrame(xdat)
    neval <- .np_plot_validate_neval(neval, where = "plot.sibandwidth")
    eval.info <- .np_plot_singleindex_eval_grid(
      bws = bws,
      xdat = xdat,
      neval = neval,
      trim = 0.0,
      where = "plot.sibandwidth"
    )
    maxneval <- nrow(eval.info$idx.eval)

    if (is.null(plot.errors.bar.num))
      plot.errors.bar.num = min(maxneval, 25L)

    if (identical(plot.errors.method, "asymptotic")) {
      tobj <- .np_plot_singleindex_asymptotic_eval(
        bws = bws,
        txdat = xdat,
        tydat = ydat,
        gradients = gradients,
        index.eval = eval.info$index.eval
      )
    } else {
      tobj <- .np_plot_singleindex_local_eval(
        bws = bws,
        idx.train = eval.info$idx.train,
        idx.eval = eval.info$idx.eval,
        ydat = ydat,
        gradients = gradients
      )
    }
    
    temp.err = matrix(data = NA, nrow = maxneval, ncol = 3)
    temp.mean = rep(NA_real_, maxneval)
    temp.all.err <- NULL
    

    temp.mean[] = if(gradients) tobj$grad.index else tobj$mean

    if (plot.errors){
      if (plot.errors.method == "asymptotic") {
        asym.obj <- .np_plot_asymptotic_error_from_se(
          se = if (gradients) tobj$gerr.index else tobj$merr,
          alpha = plot.errors.alpha,
          band.type = plot.errors.type,
          m = maxneval
        )
        temp.err[,1:2] <- asym.obj$err
        temp.all.err <- asym.obj$all.err
      } else if (plot.errors.method == "bootstrap") {
        temp.boot.raw <- compute.bootstrap.errors(
                  xdat = xdat, ydat = ydat,
                  gradients = gradients,
                  plot.errors.boot.method = plot.errors.boot.method,
                  plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
                  plot.errors.boot.wild = plot.errors.boot.wild,
                  plot.errors.boot.blocklen = plot.errors.boot.blocklen,
                  plot.errors.boot.num = plot.errors.boot.num,
                  plot.errors.center = plot.errors.center,
                  plot.errors.type = plot.errors.type,
                  plot.errors.alpha = plot.errors.alpha,
                  idx.eval = eval.info$idx.eval,
                  bws = bws)
        temp.err[,] <- temp.boot.raw[["boot.err"]]
        temp.all.err <- temp.boot.raw[["boot.all.err"]]
      }
    }

    i.sort = sort(tobj$index, index.return=TRUE)$ix

    if (!gradients){
      if(!is.null(ylim)){
        ymin = ylim[1]
        ymax = ylim[2]
      } else {
        if (plot.errors && plot.errors.type == "all") {
          yr <- compute.all.error.range(if (plot.errors.center == "estimate") temp.mean else temp.err[,3],
                                        temp.all.err)
          ymin <- yr[1]
          ymax <- yr[2]
        } else {
          if (plot.errors){
            ymin <- min(na.omit(c(temp.mean - temp.err[,1],
                                  temp.err[,3] - temp.err[,1])))
            ymax <- max(na.omit(c(temp.mean + temp.err[,2],
                                  temp.err[,3] + temp.err[,2])))
          } else {
            ymin <- min(c(temp.mean, temp.err[,3]))
            ymax <- max(c(temp.mean, temp.err[,3]))
          }
        }
      }


      if (plot.behavior != "data"){      
        plot.layout <- .np_plot_layout_activate(plot.layout)
        if (plot.errors){
          plot.args <- list(x = tobj$index[i.sort],
                            y = temp.mean[i.sort],
                            ylim = if (!is.null(ylim)) ylim else c(ymin, ymax),
                            xlim = xlim,
                            cex.axis = scalar_default(cex.axis, par()$cex.axis),
                            cex.lab =  scalar_default(cex.lab, par()$cex.lab),
                            cex.main = scalar_default(cex.main, par()$cex.main),
                            cex.sub = scalar_default(cex.sub, par()$cex.sub),
                            xlab = scalar_default(xlab, "index"),
                            ylab = scalar_default(ylab, gen.label(bws$ynames, "Conditional Mean")),
                            type = scalar_default(type, "l"),
                            lty = scalar_default(lty, par()$lty),
                            col = scalar_default(col, par()$col),
                            main = main,
                            sub = sub,
                            lwd = scalar_default(lwd, par()$lwd))
          plot.args <- .np_plot_merge_user_args(plot.args, plot.user.args)
          .np_plot_first_render_begin(first.render)
          do.call(plot, plot.args)
          .np_plot_first_render_end(first.render)
          if (plot.rug)
            .np_plot_draw_rug_1d(tobj$index)
          if (plot.errors.type == "all") {
            sorted.all.err <- lapply(temp.all.err, function(err) {
              if (is.null(err)) return(NULL)
              err[i.sort,,drop = FALSE]
            })
            draw.all.error.types(
              ex = na.omit(tobj$index[i.sort]),
              center = na.omit(if (plot.errors.center == "estimate") temp.mean[i.sort] else temp.err[i.sort,3]),
              all.err = sorted.all.err,
              plot.errors.style = plot.errors.style,
              plot.errors.bar = plot.errors.bar,
              plot.errors.bar.num = plot.errors.bar.num,
              lty = 2,
              legend = plot.legend)
          } else if (plot.errors.center == "estimate") {
            draw.errors(ex = na.omit(tobj$index[i.sort]),
                        ely = na.omit(temp.mean[i.sort] - temp.err[i.sort,1]),
                        ehy = na.omit(temp.mean[i.sort] + temp.err[i.sort,2]),
                        plot.errors.style = plot.errors.style,
                        plot.errors.bar = plot.errors.bar,
                        plot.errors.bar.num = plot.errors.bar.num,
                        lty = 2)
          } else if (plot.errors.center == "bias-corrected") {
            lines(na.omit(tobj$index[i.sort]), na.omit(temp.err[i.sort,3]), lty = 3)
            draw.errors(ex = na.omit(tobj$index[i.sort]),
                        ely = na.omit(temp.err[i.sort,3] - temp.err[i.sort,1]),
                        ehy = na.omit(temp.err[i.sort,3] + temp.err[i.sort,2]),
                        plot.errors.style  = plot.errors.style,
                        plot.errors.bar = plot.errors.bar,
                        plot.errors.bar.num = plot.errors.bar.num,
                        lty = 2)
          }
        } else {
          plot.args <- list(x = tobj$index[i.sort],
                            y = temp.mean[i.sort],
                            cex.axis = scalar_default(cex.axis, par()$cex.axis),
                            cex.lab =  scalar_default(cex.lab, par()$cex.lab),
                            cex.main = scalar_default(cex.main, par()$cex.main),
                            cex.sub = scalar_default(cex.sub, par()$cex.sub),
                            xlab = scalar_default(xlab, "Index"),
                            ylab = scalar_default(ylab, gen.label(bws$ynames, "Conditional Mean")),
                            type = scalar_default(type, "l"),
                            lty = scalar_default(lty, par()$lty),
                            col = scalar_default(col, par()$col),
                            main = main,
                            sub = sub,
                            xlim = xlim,
                            ylim = ylim,
                            lwd = scalar_default(lwd, par()$lwd))
          plot.args <- .np_plot_merge_user_args(plot.args, plot.user.args)
          .np_plot_first_render_begin(first.render)
          do.call(plot, plot.args)
          .np_plot_first_render_end(first.render)
          if (plot.rug)
            .np_plot_draw_rug_1d(tobj$index)
        }
      }

      if (plot.behavior != "plot") {
        plot.out[1] = NA
        plot.out[[1]] <- make_singleindex_payload(
          index = tobj$index,
          mean = tobj$mean,
          gradients.flag = FALSE,
          trainiseval = eval.info$trainiseval
        )
        if (plot.errors) {
          plot.out[[1]]$merr <- cbind(-temp.err[,1], temp.err[,2])
          if (plot.errors.center == "bias-corrected") {
            plot.out[[1]]$bias <- temp.err[,3] - temp.mean
          }
        }
      }

    } else {

        bmax = max(bws$beta)
        bmin = min(bws$beta)

        if (is.null(ylim)) {
          if (plot.errors && plot.errors.type == "all"){
            yr <- compute.all.error.range(if (plot.errors.center == "estimate") temp.mean else temp.err[,3],
                                          temp.all.err)
            ymax = yr[2]
            ymin = yr[1]
          } else if (plot.errors){
            ymax = max(temp.mean + temp.err[,2], na.rm = TRUE)
            ymin = min(temp.mean - temp.err[,1], na.rm = TRUE)
          } else {
            ymax = max(temp.mean, na.rm = TRUE)
            ymin = min(temp.mean, na.rm = TRUE)
          }
          common.ylim <- c(min(bmin*ymax,bmax*ymin),max(bmax*ymax,bmin*ymin))
        } else {
          common.ylim <- ylim
        }


        if (plot.behavior != "plot"){
          plot.out[1] = NA
          plot.out[[1]] <- make_singleindex_payload(
            index = tobj$index[i.sort],
            mean = tobj$mean[i.sort],
            grad = matrix(data = 0, nrow = maxneval, ncol = ncol(xdat)),
            gradients.flag = TRUE,
            trainiseval = eval.info$trainiseval
          )
          plot.out[[1]]$glerr = matrix(data=0,nrow = maxneval, ncol = ncol(xdat))
          plot.out[[1]]$gherr = matrix(data=0,nrow = maxneval, ncol = ncol(xdat))
          plot.out[[1]]$gbias = matrix(data=0,nrow = maxneval, ncol = ncol(xdat))
          
        }


        for (i in seq_len(ncol(xdat))) {
          if (plot.behavior != "data"){
            plot.layout <- .np_plot_layout_activate(plot.layout)

            if (is.null(ylim)) {
              if (!common.scale) {
                center.base.i <- if (plot.errors.center == "estimate")
                  temp.mean[i.sort] else temp.err[i.sort,3]
                center.i <- bws$beta[i] * center.base.i
                if (plot.errors && plot.errors.type == "all" && !is.null(temp.all.err)) {
                  sorted.all.err <- lapply(temp.all.err, function(err) {
                    if (is.null(err)) return(NULL)
                    abs(bws$beta[i]) * err[i.sort,,drop = FALSE]
                  })
                  panel.range <- compute.all.error.range(center.i, sorted.all.err)
                } else if (plot.errors) {
                  lo.i <- bws$beta[i] * (center.base.i - temp.err[i.sort,1])
                  hi.i <- bws$beta[i] * (center.base.i + temp.err[i.sort,2])
                  panel.range <- c(min(pmin(lo.i, hi.i), na.rm = TRUE),
                                   max(pmax(lo.i, hi.i), na.rm = TRUE))
                } else {
                  panel.range <- c(min(center.i, na.rm = TRUE), max(center.i, na.rm = TRUE))
                }
                panel.ylim <- panel.range
              } else {
                panel.ylim <- common.ylim
              }
            } else {
              panel.ylim <- common.ylim
            }
            
            plot.args <- list(x = tobj$index[i.sort],
                              y = temp.mean[i.sort] * bws$beta[i],
                              ylim = panel.ylim,
                              cex.axis = scalar_default(cex.axis, par()$cex.axis),
                              cex.lab =  scalar_default(cex.lab, par()$cex.lab),
                              cex.main = scalar_default(cex.main, par()$cex.main),
                              cex.sub = scalar_default(cex.sub, par()$cex.sub),
                              xlab = scalar_default(xlab, "index"),
                              ylab = paste("Gradient Component", i, "of", gen.label(bws$ynames, "Conditional Mean")),
                              lty = scalar_default(lty, par()$lty),
                              col = scalar_default(col, par()$col),
                              type = scalar_default(type, "l"),
                              main = main,
                              sub = sub,
                              lwd = scalar_default(lwd, par()$lwd))
            plot.args <- .np_plot_merge_user_args(plot.args, plot.user.args)
            .np_plot_first_render_begin(first.render)
            do.call(plot, plot.args)
            .np_plot_first_render_end(first.render)
            if (plot.rug)
              .np_plot_draw_rug_1d(tobj$index)
            
            if (plot.errors){
              if (plot.errors.type == "all") {
                scaled.all.err <- lapply(temp.all.err, function(err) {
                  if (is.null(err)) return(NULL)
                  abs(bws$beta[i]) * err[i.sort,,drop = FALSE]
                })
                draw.all.error.types(
                  ex = na.omit(tobj$index[i.sort]),
                  center = na.omit(bws$beta[i] * if (plot.errors.center == "estimate") temp.mean[i.sort] else temp.err[i.sort,3]),
                  all.err = scaled.all.err,
                  plot.errors.style = plot.errors.style,
                  plot.errors.bar = plot.errors.bar,
                  plot.errors.bar.num = plot.errors.bar.num,
                  lty = 2,
                  legend = plot.legend)
              } else if (plot.errors.center == "estimate") {
                lo.i <- bws$beta[i] * (temp.mean[i.sort] - temp.err[i.sort,1])
                hi.i <- bws$beta[i] * (temp.mean[i.sort] + temp.err[i.sort,2])
                draw.errors(ex = na.omit(tobj$index[i.sort]),
                            ely = na.omit(pmin(lo.i, hi.i)),
                            ehy = na.omit(pmax(lo.i, hi.i)),
                            plot.errors.style = plot.errors.style,
                            plot.errors.bar = plot.errors.bar,
                            plot.errors.bar.num = plot.errors.bar.num,
                            lty = 2)
              } else if (plot.errors.center == "bias-corrected") {
                lo.i <- bws$beta[i] * (temp.err[i.sort,3] - temp.err[i.sort,1])
                hi.i <- bws$beta[i] * (temp.err[i.sort,3] + temp.err[i.sort,2])
                lines(na.omit(tobj$index[i.sort]), na.omit(bws$beta[i] * temp.err[i.sort,3]), lty = 3)
                draw.errors(ex = na.omit(tobj$index[i.sort]),
                            ely = na.omit(pmin(lo.i, hi.i)),
                            ehy = na.omit(pmax(lo.i, hi.i)),
                            plot.errors.style  = plot.errors.style,
                            plot.errors.bar = plot.errors.bar,
                            plot.errors.bar.num = plot.errors.bar.num,
                            lty = 2)
              }
            }
          }

          if (plot.behavior != "plot"){
            plot.out[[1]]$grad[,i] = bws$beta[i]*temp.mean[i.sort]
            center.out.i <- if (plot.errors.center == "estimate")
              temp.mean[i.sort] else temp.err[i.sort,3]
            lo.i <- bws$beta[i] * (center.out.i - temp.err[i.sort,1])
            hi.i <- bws$beta[i] * (center.out.i + temp.err[i.sort,2])
            plot.out[[1]]$glerr[,i] = pmin(lo.i, hi.i)
            plot.out[[1]]$gherr[,i] = pmax(lo.i, hi.i)
            plot.out[[1]]$gbias[,i] = bws$beta[i]*temp.err[i.sort,3]
          }

        }
      }
    
    if (plot.behavior != "data" && plot.par.mfrow)
      par(mfrow=c(1,1),cex=par()$cex)
    
    if (plot.behavior != "plot"){
      names(plot.out) = paste(if (gradients) "si.grad" else "si", seq_along(plot.out), sep = "")
      
      return (plot.out)
    }

    
  }

## This function will compute the cumulative integral at each sample
## realization using the trapezoidal rule and the cumsum function as
## we need to compute this in a computationally efficient manner.

# integrate.trapezoidal <- function(x,y) {
#   n <- length(x)
#   rank.x <- rank(x)
#   order.x <- order(x)
#   y <- y[order.x]
#   x <- x[order.x]
#   int.vec <- numeric(length(x))
#   ## Use a correction term at the boundary: -cx^2/12*(f'(b)-f'(a)),
#   ## check for NaN case
#   cx  <- x[2]-x[1]
#   ca <- (y[2]-y[1])/cx
#   cb <- (y[n]-y[n-1])/cx
#   cf <- cx^2/12*(cb-ca)
#   if(!is.finite(cf)) cf <- 0
#   int.vec[1] <- 0
#   int.vec[2:n] <- cumsum((x[2:n]-x[2:n-1])*(y[2:n]+y[2:n-1])/2)
#   return(int.vec[rank.x]-cf)
# }

# Benches 1.3-1.7 times faster (mean/median) produces identical results

integrate.trapezoidal <- function(x, y) {
  n <- length(x)
  order.x <- order(x)
  x <- x[order.x]
  y <- y[order.x]
  dx <- diff(x)
  dy <- diff(y)
  cx <- dx[1]
  ca <- dy[1] / cx
  cb <- dy[n - 1] / cx
  cf <- cx^2 / 12 * (cb - ca)
  if (!is.finite(cf)) cf <- 0
  int.vec <- c(0, cumsum(dx * (y[-n] + y[-1]) / 2))
  int.vec <- int.vec - cf
  int.vec[order(order.x)]  # inverse permutation == rank(x) when no ties
}

## No Zero Denominator, used in C code for kernel estimation...

## Original, better, best

# NZD <- function(a) {
#   ifelse(a<0,pmin(-.Machine$double.eps,a),pmax(.Machine$double.eps,a))
# }
  
# NZD <- function(a) {
#   if(length(a) == 1) {
#     if(is.na(a)) return(a)
#     if(a < 0) return(min(-.Machine$double.eps, a))
#     return(max(.Machine$double.eps, a))
#   }
#   ifelse(a<0,pmin(-.Machine$double.eps,a),pmax(.Machine$double.eps,a))
# }

# Benches 5.5-7.9 times faster (mean/median) produces identical results

NZD <- function(a) {
  eps <- .Machine$double.eps
  if (length(a) == 1) {
    if (a >= 0) {
      if (a < eps) return(eps)
    } else {
      if (a > -eps) return(-eps)
    }
    return(a)
  }
  idx <- which(abs(a) < eps)
  if (length(idx) > 0) {
    small <- a[idx]
    small[small >= 0] <- eps
    small[small < 0] <- -eps
    a[idx] <- small
  }
  a
}

## New function when only positive values are guaranteed, better, best

# NZD_pos <- function(a) {
#   if(length(a) == 1) {
#     if(is.na(a)) return(a)
#     return(max(.Machine$double.eps, a))
#   }
#   pmax(.Machine$double.eps,a)
# }

# Benches 1.3-1.8 times faster (mean/median) produces identical results

NZD_pos <- function(a) {
  eps <- .Machine$double.eps
  if (length(a) == 1)
    return(if (a < eps) eps else a)
  idx <- which(a < eps)
  if (length(idx) > 0)
    a[idx] <- eps
  a
}

npRidgeSequenceAdditive <- function(n.train, cap = 1.0) {
  if (!is.numeric(n.train) || length(n.train) != 1L || is.na(n.train) ||
      !is.finite(n.train) || n.train < 1 || n.train != floor(n.train))
    stop("'n.train' must be a positive integer")
  if (!is.numeric(cap) || length(cap) != 1L || is.na(cap) ||
      !is.finite(cap) || cap <= 0)
    stop("'cap' must be a positive finite numeric scalar")

  step <- 1.0 / as.double(n.train)
  seq.out <- seq.int(from = 0.0, to = as.double(cap), by = step)
  if (utils::tail(seq.out, 1L) < cap)
    seq.out <- c(seq.out, as.double(cap))
  unique(as.double(seq.out))
}

npRidgeSequenceFromBase <- function(n.train, ridge.base = 0.0, cap = 1.0) {
  if (!is.numeric(ridge.base) || length(ridge.base) != 1L || is.na(ridge.base) ||
      !is.finite(ridge.base) || ridge.base < 0)
    stop("'ridge.base' must be a non-negative finite numeric scalar")

  base <- as.double(ridge.base)
  step <- 1.0 / as.double(n.train)
  cap <- as.double(cap)

  if (base >= cap)
    return(unique(c(base, cap)))

  seq.out <- seq.int(from = base, to = cap, by = step)
  if (utils::tail(seq.out, 1L) < cap)
    seq.out <- c(seq.out, cap)
  unique(as.double(seq.out))
}

.np_glp_degree_hard_max <- 100L
.np_glp_degree_warn_threshold <- 25L

.np_warn_high_glp_degree <- function(value, argname = "degree") {
  value <- as.integer(value)
  if (!length(value))
    return(invisible(NULL))

  threshold <- .np_glp_degree_warn_threshold
  if (any(value > threshold)) {
    warning(
      sprintf(
        "%s contains unusually large polynomial degree values above %d; numerical instability is possible",
        argname,
        threshold
      ),
      call. = FALSE
    )
  }

  invisible(NULL)
}

npValidateGlpDegree <- function(regtype, degree, ncon, argname = "degree") {
  degree.max <- .np_glp_degree_hard_max

  if (!identical(regtype, "lp"))
    return(NULL)

  if (ncon == 0L) {
    if (is.null(degree))
      stop(sprintf("%s must be 0 when regtype='lp' has no continuous predictors",
                   argname),
           call. = FALSE)
    if (!length(degree))
      return(integer(0))
    if (!is.numeric(degree) ||
        length(degree) != 1L ||
        anyNA(degree) ||
        any(!is.finite(degree)) ||
        degree != 0L)
      stop(sprintf("%s must be 0 when regtype='lp' has no continuous predictors",
                   argname),
           call. = FALSE)
    return(integer(0))
  }

  if (is.null(degree)) {
    stop(sprintf("%s must be supplied explicitly when regtype='lp'", argname))
  }

  if (length(degree) != ncon)
    stop(sprintf("%s must have one entry per continuous predictor (%d expected, got %d)",
                 argname, ncon, length(degree)))

  if (!is.numeric(degree))
    stop(sprintf("%s must contain finite non-negative integers", argname))

  if (anyNA(degree) || any(!is.finite(degree)))
    stop(sprintf("%s must contain finite non-negative integers", argname))

  if (any(degree < 0))
    stop(sprintf("%s must contain finite non-negative integers", argname))

  if (any(degree != floor(degree)))
    stop(sprintf("%s must contain finite non-negative integers", argname))

  if (any(degree > degree.max))
    stop(sprintf("%s must contain finite non-negative integers in [0,%d]",
                 argname, degree.max))

  .np_warn_high_glp_degree(degree, argname = argname)

  as.integer(degree)
}

npSetupGlpDegree <- function(regtype,
                             degree,
                             ncon,
                             degree.select = c("manual", "coordinate", "exhaustive")) {
  if (!identical(regtype, "lp"))
    return(degree)

  degree.select <- match.arg(degree.select, c("manual", "coordinate", "exhaustive"))
  if (!is.null(degree) || identical(degree.select, "manual"))
    return(degree)

  if (ncon == 0L)
    return(integer(0))

  rep.int(1L, ncon)
}

npValidateGlpBernstein <- function(regtype, bernstein.basis, argname = "bernstein.basis") {
  if (!identical(regtype, "lp"))
    return(FALSE)

  if (is.null(bernstein.basis))
    bernstein.basis <- FALSE

  if (!is.logical(bernstein.basis) || length(bernstein.basis) != 1L || is.na(bernstein.basis))
    stop(sprintf("%s must be TRUE or FALSE", argname))

  isTRUE(bernstein.basis)
}

npValidateScalarLogical <- function(value, argname) {
  value <- as.logical(value)
  if (length(value) != 1L || is.na(value))
    stop(sprintf("'%s' must be TRUE or FALSE", argname))
  value
}

npValidateNewdataColumns <- function(newdata, required, argname = "newdata") {
  nd <- toFrame(newdata)
  required <- unique(as.character(required))
  required <- required[nzchar(required)]
  missing.names <- setdiff(required, names(nd))
  if (length(missing.names) > 0L) {
    stop(sprintf("%s must contain columns: %s",
                 argname, paste(shQuote(required), collapse = ", ")),
         call. = FALSE)
  }
  invisible(TRUE)
}

npValidateNewdataFormula <- function(newdata, formula, include.response = TRUE,
                                     argname = "newdata") {
  tt <- terms(formula)
  if (!isTRUE(include.response))
    tt <- delete.response(tt)
  npValidateNewdataColumns(newdata, all.vars(tt), argname = argname)
}

.np_prepare_nomad_shortcut <- function(nomad,
                                       call_names,
                                       preset,
                                       values,
                                       where = "npregbw") {
  call_names <- if (is.null(call_names)) character(0) else call_names[nzchar(call_names)]
  nomad.enabled <- if ("nomad" %in% call_names) {
    npValidateScalarLogical(nomad, "nomad")
  } else {
    FALSE
  }

  metadata <- list(
    enabled = isTRUE(nomad.enabled),
    where = where,
    preset = "lp_nomad",
    auto.filled = character(0),
    user.supplied = character(0),
    normalized.values = preset
  )

  if (!isTRUE(nomad.enabled))
    return(list(enabled = FALSE, values = values, metadata = metadata))

  auto.filled <- character(0)
  user.supplied <- intersect(names(preset), call_names)

  for (arg in names(preset)) {
    if (!(arg %in% call_names)) {
      values[[arg]] <- preset[[arg]]
      auto.filled <- c(auto.filled, arg)
    }
  }

  metadata$auto.filled <- auto.filled
  metadata$user.supplied <- user.supplied
  metadata$normalized.values <- values[names(preset)]

  list(enabled = TRUE, values = values, metadata = metadata)
}

.np_attach_nomad_shortcut <- function(obj, metadata) {
  if (isTRUE(metadata$enabled))
    obj$nomad.shortcut <- metadata
  obj
}

npValidateNonNegativeInteger <- function(value, argname) {
  if (!is.numeric(value) || length(value) != 1L || is.na(value) ||
      !is.finite(value) || value < 0 || value != floor(value))
    stop(sprintf("'%s' must be a non-negative integer", argname))
  as.integer(value)
}

npDefaultNmulti <- function(p) {
  if (!is.numeric(p) || length(p) != 1L || is.na(p) ||
      !is.finite(p) || p < 1 || p != floor(p))
    stop("'p' must be a positive integer")
  min(2L, as.integer(p))
}

npValidatePositiveInteger <- function(value, argname) {
  if (!is.numeric(value) || length(value) != 1L || is.na(value) ||
      !is.finite(value) || value < 1 || value != floor(value))
    stop(sprintf("'%s' must be a positive integer", argname))
  as.integer(value)
}

npValidateNmulti <- function(value, argname = "nmulti") {
  npValidatePositiveInteger(value, argname)
}

npValidatePositiveFiniteNumeric <- function(value, argname) {
  if (!is.numeric(value) || length(value) != 1L || is.na(value) ||
      !is.finite(value) || value <= 0)
    stop(sprintf("'%s' must be a positive finite numeric scalar", argname))
  as.double(value)
}

npScaleFactorSearchOldNames <- c(
  "cfac.init",
  "lbc.init",
  "hbc.init",
  "scale.factor.lower.bound"
)

npScaleFactorSearchNewNames <- c(
  cfac.init = "scale.factor.init",
  lbc.init = "scale.factor.init.lower",
  hbc.init = "scale.factor.init.upper",
  scale.factor.lower.bound = "scale.factor.search.lower"
)

npRejectRenamedScaleFactorSearchArgs <- function(arg.names,
                                                 where = "bandwidth search") {
  old <- intersect(npScaleFactorSearchOldNames, arg.names)
  if (!length(old))
    return(invisible(NULL))

  messages <- sprintf("'%s' has been renamed to '%s'",
                      old,
                      unname(npScaleFactorSearchNewNames[old]))
  stop(sprintf("%s: %s", where, paste(messages, collapse = "; ")),
       call. = FALSE)
}

npValidateScaleFactorLowerBound <- function(value,
                                            argname = "scale.factor.search.lower") {
  if (!is.numeric(value) || length(value) != 1L || is.na(value) ||
      !is.finite(value) || value < 0) {
    stop(sprintf("'%s' must be a nonnegative finite numeric scalar", argname),
         call. = FALSE)
  }

  as.double(value)
}

npResolveScaleFactorLowerBound <- function(value,
                                           fallback = 0.1,
                                           argname = "scale.factor.search.lower") {
  if (is.null(value))
    return(as.double(fallback))

  npValidateScaleFactorLowerBound(value, argname = argname)
}

npResolveScaleFactorSearchLower <- function(value,
                                            fallback = 0.1,
                                            argname = "scale.factor.search.lower") {
  npResolveScaleFactorLowerBound(value, fallback = fallback, argname = argname)
}

npGetScaleFactorSearchLower <- function(object,
                                        fallback = 0.1,
                                        argname = "scale.factor.search.lower") {
  new <- object[["scale.factor.search.lower"]]
  old <- object[["scale.factor.lower.bound"]]

  if (!is.null(new) && !is.null(old)) {
    new.value <- npResolveScaleFactorSearchLower(new, fallback = fallback,
                                                 argname = "scale.factor.search.lower")
    old.value <- npResolveScaleFactorSearchLower(old, fallback = fallback,
                                                 argname = "scale.factor.lower.bound")
    if (!identical(new.value, old.value)) {
      stop(sprintf(
        "%s conflicts with legacy 'scale.factor.lower.bound' (%.15g versus %.15g)",
        argname,
        new.value,
        old.value
      ), call. = FALSE)
    }
    return(new.value)
  }

  if (!is.null(new))
    return(npResolveScaleFactorSearchLower(new, fallback = fallback,
                                           argname = argname))

  npResolveScaleFactorSearchLower(old, fallback = fallback,
                                  argname = "scale.factor.lower.bound")
}

npSetScaleFactorSearchLower <- function(object, value, fallback = 0.1) {
  object[["scale.factor.search.lower"]] <-
    npResolveScaleFactorSearchLower(value, fallback = fallback)
  object[["scale.factor.lower.bound"]] <- NULL
  object
}

npEffectiveContinuousStartLower <- function(scale.factor.init.lower,
                                            scale.factor.search.lower) {
  max(as.double(scale.factor.init.lower), as.double(scale.factor.search.lower))
}

npEffectiveContinuousStartPoint <- function(scale.factor.init,
                                            scale.factor.search.lower) {
  max(as.double(scale.factor.init), as.double(scale.factor.search.lower))
}

npContinuousSearchStartControls <- function(scale.factor.init.lower,
                                            scale.factor.init.upper,
                                            scale.factor.init,
                                            scale.factor.search.lower,
                                            where = "bandwidth search") {
  lbc.raw <- npValidatePositiveFiniteNumeric(scale.factor.init.lower,
                                             "scale.factor.init.lower")
  scale.factor.init.upper <- npValidatePositiveFiniteNumeric(scale.factor.init.upper,
                                             "scale.factor.init.upper")
  cfac.raw <- npValidatePositiveFiniteNumeric(scale.factor.init,
                                             "scale.factor.init")
  scale.factor.search.lower <- npValidateScaleFactorLowerBound(
    scale.factor.search.lower,
    "scale.factor.search.lower"
  )

  lbc.effective <- npEffectiveContinuousStartLower(
    lbc.raw,
    scale.factor.search.lower
  )
  cfac.effective <- npEffectiveContinuousStartPoint(
    cfac.raw,
    scale.factor.search.lower
  )

  if (scale.factor.init.upper < lbc.effective) {
    stop(sprintf(
      "%s: 'scale.factor.init.upper' must be greater than or equal to max('scale.factor.init.lower', 'scale.factor.search.lower') (effective lower %.15g; scale.factor.init.upper %.15g)",
      where,
      lbc.effective,
      scale.factor.init.upper
    ), call. = FALSE)
  }

  list(
    scale.factor.init.lower = lbc.effective,
    scale.factor.init.upper = scale.factor.init.upper,
    scale.factor.init = cfac.effective
  )
}

npValidateLpBasis <- function(regtype, basis, argname = "basis") {
  if (!identical(regtype, "lp"))
    return("glp")

  if (is.null(basis))
    basis <- "glp"

  basis <- match.arg(as.character(basis), c("glp", "additive", "tensor"))
  basis
}

npLpBasisCode <- function(basis) {
  basis.value <- if (is.null(basis) || !length(basis)) "glp" else basis
  switch(tolower(basis.value),
         additive = 0L,
         glp = 1L,
         tensor = 2L,
         1L)
}

npValidateGlpGradientOrder <- function(regtype,
                                       gradient.order,
                                       ncon,
                                       argname = "gradient.order") {
  degree.max <- 12L

  if (!identical(regtype, "lp"))
    return(NULL)

  if (is.null(gradient.order))
    gradient.order <- rep.int(1L, ncon)

  if (!length(gradient.order) && ncon == 0L)
    return(integer(0))

  if (length(gradient.order) == 1L && ncon > 1L)
    gradient.order <- rep.int(gradient.order, ncon)

  if (length(gradient.order) != ncon)
    stop(sprintf("%s must have one entry per continuous predictor (%d expected, got %d)",
                 argname, ncon, length(gradient.order)))

  if (!is.numeric(gradient.order))
    stop(sprintf("%s must contain finite positive integers", argname))

  if (anyNA(gradient.order) || any(!is.finite(gradient.order)))
    stop(sprintf("%s must contain finite positive integers", argname))

  if (any(gradient.order <= 0))
    stop(sprintf("%s must contain finite positive integers", argname))

  if (any(gradient.order != floor(gradient.order)))
    stop(sprintf("%s must contain finite positive integers", argname))

  if (any(gradient.order > degree.max))
    stop(sprintf("%s must contain integers in [1,%d]", argname, degree.max))

  as.integer(gradient.order)
}

npConditionalRegEngineSpec <- function(bws, where = "conditional estimator") {
  reg.engine <- if (is.null(bws$regtype.engine)) {
    if (is.null(bws$regtype)) "lc" else as.character(bws$regtype)
  } else {
    as.character(bws$regtype.engine)
  }
  basis.engine <- if (is.null(bws$basis.engine)) {
    if (is.null(bws$basis)) "glp" else bws$basis
  } else {
    bws$basis.engine
  }
  degree.engine <- if (is.null(bws$degree.engine)) {
    if (bws$xncon > 0L) {
      if (identical(reg.engine, "lc")) {
        rep.int(0L, bws$xncon)
      } else {
        npValidateGlpDegree(
          regtype = "lp",
          degree = bws$degree,
          ncon = bws$xncon
        )
      }
    } else {
      integer(0)
    }
  } else {
    as.integer(bws$degree.engine)
  }
  bernstein.engine <- if (is.null(bws$bernstein.basis.engine)) {
    isTRUE(bws$bernstein.basis)
  } else {
    isTRUE(bws$bernstein.basis.engine)
  }

  if (!identical(reg.engine, "lp") && bws$xncon > 0L)
    degree.engine <- rep.int(if (identical(reg.engine, "ll")) 1L else 0L, bws$xncon)

  list(
    reg.engine = reg.engine,
    basis.engine = basis.engine,
    degree.engine = degree.engine,
    bernstein.engine = bernstein.engine
  )
}

npConditionalGradientOrder <- function(bws,
                                       reg.engine,
                                       gradient.order,
                                       where = "conditional estimator") {
  if (bws$xncon == 0L)
    return(integer(0))

  gorder <- if (is.null(gradient.order)) {
    rep.int(1L, bws$xncon)
  } else {
    npValidateGlpGradientOrder(
      regtype = "lp",
      gradient.order = gradient.order,
      ncon = bws$xncon
    )
  }

  if (!identical(reg.engine, "lp") && any(gorder > 1L)) {
    stop(sprintf(
      "%s supports gradient.order > 1 only for regtype='lp' continuous explanatory predictors",
      where
    ), call. = FALSE)
  }

  gorder
}

npCheckRegressionDesignCondition <- function(reg.code,
                                             xcon,
                                             basis = "glp",
                                             degree = NULL,
                                             bernstein.basis = FALSE,
                                             where = "npregbw") {
  kappa.warn <- 1e8
  kappa.stop <- 1e12

  if (!(reg.code %in% c(REGTYPE_LL, REGTYPE_LP)))
    return(invisible(NULL))

  xcon <- as.data.frame(xcon)
  n <- nrow(xcon)
  if (is.null(n) || n <= 0L)
    return(invisible(NULL))

  B <- if (identical(reg.code, REGTYPE_LP)) {
    if (is.null(degree))
      stop(sprintf("%s: LP degree vector missing for design-conditioning check", where))
    W.lp(xdat = xcon,
         degree = degree,
         basis = basis,
         bernstein.basis = isTRUE(bernstein.basis))
  } else {
    cbind(1, as.matrix(xcon))
  }

  p <- ncol(B)
  if (is.null(p) || p <= 0L)
    return(invisible(NULL))

  sv <- suppressWarnings(tryCatch(svd(B, nu = 0L, nv = 0L)$d, error = function(e) NULL))
  if (is.null(sv) || !length(sv))
    stop(sprintf("%s: unable to compute singular values for design-conditioning check", where))
  tol.rank <- max(dim(B)) * max(sv) * .Machine$double.eps
  r <- sum(sv > tol.rank)
  if (r < p) {
    stop(sprintf("%s: regression design matrix is rank deficient (rank=%d < p=%d). Reduce polynomial degree or remove collinear continuous predictors.",
                 where, r, p))
  }

  kB <- suppressWarnings(tryCatch(kappa(B), error = function(e) Inf))
  if (!is.finite(kB))
    kB <- Inf

  if (kB > kappa.stop) {
    stop(sprintf("%s: regression design matrix is severely ill-conditioned (kappa(B)=%.3e > %.1e). Reduce polynomial degree or remove collinear continuous predictors.",
                 where, kB, kappa.stop))
  }

  if (kB > kappa.warn) {
    .np_warning(sprintf("%s: regression design matrix is ill-conditioned (kappa(B)=%.3e > %.1e). Estimation may rely heavily on ridging; consider lower degree or less collinear predictors.",
                    where, kB, kappa.warn),
            call. = FALSE, immediate. = TRUE)
  }

  invisible(NULL)
}

npRegtypeToC <- function(regtype, degree, ncon, context = "npreg") {
  # Internal regression-type codes:
  # lc -> REGTYPE_LC, ll -> REGTYPE_LL, lp -> REGTYPE_LP.
  if (identical(regtype, "lc"))
    return(list(code = REGTYPE_LC, degree = NULL))

  if (identical(regtype, "ll"))
    return(list(code = REGTYPE_LL, degree = NULL))

  degree <- npValidateGlpDegree(regtype, degree, ncon)

  list(code = REGTYPE_LP, degree = degree)
}

npCanonicalConditionalRegSpec <- function(regtype = c("lc", "ll", "lp"),
                                          basis = c("glp", "additive", "tensor"),
                                          degree = NULL,
                                          bernstein.basis = FALSE,
                                          ncon,
                                          where = "npc*") {
  regtype <- match.arg(regtype)
  ncon <- npValidateNonNegativeInteger(ncon, "ncon")
  basis <- npValidateLpBasis(regtype = "lp", basis = basis)
  bernstein.basis <- npValidateGlpBernstein(regtype = "lp",
                                            bernstein.basis = bernstein.basis)

  if (identical(regtype, "lc")) {
    degree <- rep.int(0L, ncon)
    return(list(
      regtype = "lc",
      basis = basis,
      degree = degree,
      bernstein.basis = FALSE,
      regtype.engine = "lc",
      basis.engine = basis,
      degree.engine = degree,
      bernstein.basis.engine = FALSE
    ))
  }

  if (identical(regtype, "ll")) {
    if (ncon == 0L)
      stop(sprintf("%s: regtype='ll' requires at least one continuous predictor; use regtype='lc' for categorical-only predictors",
                   where),
           call. = FALSE)

    if (ncon > 0L) {
      degree <- rep.int(1L, ncon)
      basis <- "glp"
    } else {
      degree <- integer(0)
      basis <- "glp"
    }

    return(list(
      regtype = "ll",
      basis = basis,
      degree = degree,
      bernstein.basis = FALSE,
      regtype.engine = if (ncon > 0L) "lp" else "lc",
      basis.engine = basis,
      degree.engine = degree,
      bernstein.basis.engine = FALSE
    ))
  }

  degree <- npValidateGlpDegree(regtype = "lp",
                                degree = degree,
                                ncon = ncon)
  if (ncon == 0L && length(degree) == 0L)
    bernstein.basis <- FALSE
  list(
    regtype = "lp",
    basis = basis,
    degree = degree,
    bernstein.basis = bernstein.basis,
    regtype.engine = if (ncon > 0L) "lp" else "lc",
    basis.engine = basis,
    degree.engine = degree,
    bernstein.basis.engine = bernstein.basis
  )
}

npWithLocalLinearRawBasisSearchError <- function(expr,
                                                 where,
                                                 spec,
                                                 bwmethod,
                                                 ncon) {
  expected <- switch(where,
    npcdensbw = "C_np_density_conditional_bw: optimizer returned a fixed-bandwidth candidate with invalid raw objective",
    npcdistbw = "C_np_distribution_conditional_bw: optimizer returned a fixed-bandwidth candidate with invalid raw objective",
    NULL
  )
  tryCatch(
    force(expr),
    error = function(e) {
      msg <- conditionMessage(e)
      degree <- if (is.null(spec$degree.engine)) integer(0) else as.integer(spec$degree.engine)
      ncon <- as.integer(ncon)
      targeted <- !is.null(expected) &&
        identical(msg, expected) &&
        identical(spec$regtype, "ll") &&
        identical(bwmethod, "cv.ls") &&
        identical(spec$regtype.engine, "lp") &&
        !isTRUE(spec$bernstein.basis.engine) &&
        length(degree) == ncon &&
        ncon > 0L &&
        all(degree == 1L)

      if (targeted) {
        stop(sprintf(
          "%s() local-linear cv.ls failed while using the canonical raw degree-1 basis. Try regtype = \"lp\", degree = 1, bernstein.basis = TRUE, or center/scale continuous regressors.",
          where
        ), call. = FALSE)
      }
      stop(e)
    }
  )
}

npResolveCanonicalConditionalRegSpec <- function(mc.names,
                                                 regtype = c("lc", "ll", "lp"),
                                                 basis = c("glp", "additive", "tensor"),
                                                 degree = NULL,
                                                 bernstein.basis = FALSE,
                                                 ncon,
                                                 where = "npreg") {
  mc.names <- if (is.null(mc.names)) character(0) else as.character(mc.names)
  regtype.named <- any(mc.names == "regtype")
  basis.named <- any(mc.names == "basis")
  degree.named <- any(mc.names == "degree")
  bernstein.named <- any(mc.names == "bernstein.basis")

  regtype <- if (regtype.named) match.arg(regtype) else "lc"

  if (identical(regtype, "lc") && (basis.named || degree.named || bernstein.named)) {
    stop("regtype='lc' does not accept basis/degree/bernstein.basis; use regtype='lp' for local-polynomial controls")
  }

  if (identical(regtype, "ll")) {
    if (degree.named) {
      degree <- npValidateGlpDegree(regtype = "lp", degree = degree, ncon = ncon)
      if (!identical(as.integer(degree), rep.int(1L, ncon)))
        stop("regtype='ll' uses canonical LP(degree=1, basis='glp'); remove 'degree' or use regtype='lp'")
    }
    if (basis.named) {
      basis <- match.arg(basis)
      if (!identical(basis, "glp"))
        stop("regtype='ll' uses canonical basis='glp'; use regtype='lp' for alternate LP bases")
    }
    if (bernstein.named) {
      bernstein.basis <- npValidateGlpBernstein(regtype = "lp",
                                                bernstein.basis = bernstein.basis)
      if (isTRUE(bernstein.basis))
        stop("regtype='ll' uses canonical bernstein.basis=FALSE; use regtype='lp' for Bernstein LP")
    }
  }

  npCanonicalConditionalRegSpec(
    regtype = regtype,
    basis = basis,
    degree = degree,
    bernstein.basis = bernstein.basis,
    ncon = ncon,
    where = where
  )
}

npRejectLegacyLpArgs <- function(dotnames, where = "npreg") {
  if (is.null(dotnames) || !length(dotnames))
    return(invisible(NULL))
  bad <- intersect(dotnames, c("glp.degree", "glp.bernstein", "glp.basis"))
  if (length(bad))
    stop(sprintf("%s: legacy arguments %s are no longer supported; use basis, degree and bernstein.basis",
                 where, paste(sprintf("'%s'", bad), collapse = ", ")))
  if ("remin" %in% dotnames)
    stop(sprintf("%s: argument 'remin' has been replaced by 'powell.remin' and 'nomad.remin'",
                 where))
  invisible(NULL)
}

dim_basis <- function(basis = c("glp", "additive", "tensor"),
                      kernel = TRUE,
                      degree = NULL,
                      segments = NULL,
                      include = NULL,
                      categories = NULL) {
  basis <- match.arg(basis)

  if (is.null(degree))
    degree <- integer(0)
  if (is.null(segments))
    segments <- rep.int(1L, length(degree))

  if (length(degree) != length(segments))
    stop("degree and segments must have the same length")
  if (!is.numeric(degree) ||
      anyNA(degree) ||
      any(!is.finite(degree)) ||
      any(degree < 0L) ||
      any(degree != floor(degree)))
    stop("degree must contain finite non-negative integers")
  if (!is.numeric(segments) ||
      anyNA(segments) ||
      any(!is.finite(segments)) ||
      any(segments <= 0L) ||
      any(segments != floor(segments)))
    stop("segments must contain finite positive integers")

  degree <- as.integer(degree)
  segments <- as.integer(segments)

  if (is.null(include))
    include <- integer(0)
  if (is.null(categories))
    categories <- integer(0)

  if (length(include) != length(categories))
    stop("include and categories must have the same length")
  if (!is.numeric(include) ||
      anyNA(include) ||
      any(!is.finite(include)) ||
      any(include < 0L) ||
      any(include != floor(include)))
    stop("include must contain finite non-negative integers")
  if (!is.numeric(categories) ||
      anyNA(categories) ||
      any(!is.finite(categories)) ||
      any(categories < 0L) ||
      any(categories != floor(categories)))
    stop("categories must contain finite non-negative integers")

  include <- as.integer(include)
  categories <- as.integer(categories)

  basis.code <- switch(basis, additive = 0L, glp = 1L, tensor = 2L)

  .Call("C_np_dim_basis",
        as.integer(basis.code),
        as.integer(isTRUE(kernel)),
        degree,
        segments,
        include,
        categories,
        PACKAGE = "np")
}

dimBS <- function(basis = "additive",
                  kernel = TRUE,
                  degree = NULL,
                  segments = NULL,
                  include = NULL,
                  categories = NULL) {
  dim_basis(basis = basis,
            kernel = kernel,
            degree = degree,
            segments = segments,
            include = include,
            categories = categories)
}

## Function to test for monotone increasing vector

is.monotone.increasing <- function(x) {
  ## Sorted and last value > first value
  !is.unsorted(x) && x[length(x)] > x[1]
}
  
## what is a badord? ... an ordered factor of numeric values to treat
## them properly one must preserve the numeric value, ie. scale not
## just their sorted order Actually, the ord/badord paradigm must go,
## in place of levels caching

matrix.sd <- function(x, na.rm=FALSE) {
  if(is.matrix(x)) apply(x, 2, sd, na.rm=na.rm)
  else if(is.vector(x)) sd(x, na.rm=na.rm)
  else if(is.data.frame(x)) sapply(x, sd, na.rm=na.rm)
  else sd(as.vector(x), na.rm=na.rm)
}

.np_validate_seed_scalar <- function(seed) {
  if (is.null(seed) || length(seed) != 1L || !is.numeric(seed) || is.na(seed) || !is.finite(seed))
    stop("'seed' must be a single finite numeric value", call. = FALSE)

  normalized <- abs(as.double(seed)[1L])

  if (normalized > .Machine$integer.max || normalized != floor(normalized))
    stop("'seed' must be representable as a non-negative integer after abs()", call. = FALSE)

  as.integer(normalized)
}

npseed <- function(seed){
  .Call("C_np_set_seed", .np_validate_seed_scalar(seed), PACKAGE = "np")
  invisible()
}

erf <- function(z) { 2 * pnorm(z*sqrt(2)) - 1 }

nptgauss <- function(b){

  rel.tol <- sqrt(.Machine$double.eps)

  b.max <- sqrt(-2*log(.Machine$double.eps))

  if((b < 0) || (b > b.max))
    stop(paste("b must be between 0 and",b.max))

  alpha <- 1.0/(pnorm(b)-pnorm(-b)-2*b*dnorm(b))

  tgauss <- function(z) {
    out <- alpha * (dnorm(z) - dnorm(b))
    out[abs(z) >= b] <- 0.0
    out
  }

  c0 <- alpha*dnorm(b)

  k <- integrate(f = function(z) { tgauss(z)^2 }, -b, b)$value
  k2 <- integrate(f = function(z) { z^2*tgauss(z) }, -b, b)$value
  k22 <- integrate(f = function(z) { (z*tgauss(z))^2 }, -b, b)$value
  km <- integrate(f = function(z) { tgauss(z-0.5)*tgauss(z+0.5) }, -b+0.5, b-0.5)$value

  a0 <- (0.5 + 2*b*c0)/integrate(f = function(z){ erf(z/2 + b)*exp(-0.25*z^2) }, -2*b, 0)$value
  a2 <- (c0 + k - a0*erf(b))/erf(b/sqrt(2))
  a1 <- -(a2*erf(b/sqrt(2)) + c0)/(2*b)

  int.kernels[CKER_TGAUSS + 1] <- k
  
  invisible(.Call("C_np_set_tgauss2",
                  as.double(c(b, alpha, c0, a0, a1, a2, k, k2, k22, km)),
                  PACKAGE = "np"))

}

numNotIn <- function(x){
  while(is.element(num <- rnorm(1),x)){}
  num
}

dlev <- function(x){
  if(is.ordered(x))
    x.dlev <- suppressWarnings(as.numeric(levels(x)))
  if (!is.ordered(x) || any(is.na(x.dlev)))
    x.dlev <- as.numeric(seq_len(nlevels(x)))
  x.dlev
}

isNum <- function(x){
  return(!any(is.na(suppressWarnings(as.numeric(x)))))
}

untangle <- function(frame){
  if (is.null(frame))
    return(NULL)
  
  iord <- unlist(lapply(frame, is.ordered))
  iuno <- unlist(lapply(frame, is.factor)) & !iord
  icon <- unlist(lapply(frame, is.numeric))

  if(!all(iord | iuno | icon)) 
    stop("non-allowed data type in data frame")

  inumord <-
    suppressWarnings(unlist(lapply(frame,
    function (z) {
      is.ordered(z) && is.numeric(tryCatch(as.numeric(levels(z)), warning = function (y) {
        FALSE }))
    })))

  all.lev <- lapply(frame, function(y){
    t.ret <- NULL
    if (is.factor(y))
      t.ret <- levels(y)
    t.ret
  })

  all.ulev <- lapply(frame, function (y) {
    t.ret <- NULL
    if (is.factor(y))
      t.ret <- sort(unique(y))
    t.ret
  })

  all.dlev <- lapply(frame, function (y) {
    t.ret <- NULL
    if (is.factor(y))
      t.ret <- dlev(y)
    t.ret
  })

  all.nlev <- lapply(frame, function(y) {
    t.ret <- NULL
    if (is.factor(y))
      t.ret <- nlevels(y)
    t.ret
  })

  all.min <- lapply(frame, function(y) {
    t.ret <- NULL
    if (is.numeric(y))
      t.ret <- min(y, na.rm = TRUE)
    t.ret
  })

  all.max <- lapply(frame, function(y) {
    t.ret <- NULL
    if (is.numeric(y))
      t.ret <- max(y, na.rm = TRUE)
    t.ret
  })

  list(iord = iord,
       iuno = iuno,
       icon = icon,
       inumord = inumord,
       all.lev = all.lev,
       all.ulev = all.ulev,
       all.dlev = all.dlev,
       all.nlev = all.nlev,
       all.min = all.min,
       all.max = all.max)
}

npKernelBoundsResolve <- function(dati,
                                  varnames,
                                  kerbound = c("none", "range", "fixed"),
                                  kerlb = NULL,
                                  kerub = NULL,
                                  argprefix = "cker") {
  kerbound <- match.arg(kerbound)
  icon.idx <- which(dati$icon)
  ncon <- length(icon.idx)
  if (is.null(varnames) || length(varnames) != length(dati$icon))
    varnames <- paste0("V", seq_along(dati$icon))

  full.lb <- rep(-Inf, length(dati$icon))
  full.ub <- rep(Inf, length(dati$icon))

  if (ncon == 0L) {
    return(list(bound = "none", lb = full.lb, ub = full.ub))
  }

  mins <- unlist(dati$all.min[icon.idx], use.names = FALSE)
  maxs <- unlist(dati$all.max[icon.idx], use.names = FALSE)
  cnames <- varnames[icon.idx]

  recycleBounds <- function(x, nm) {
    if (is.null(x))
      stop(sprintf("'%s' requires '%s' and '%s'.",
                   paste0(argprefix, "bound"),
                   paste0(argprefix, "lb"),
                   paste0(argprefix, "ub")))
    x <- as.numeric(x)
    if (!(length(x) %in% c(1L, ncon))) {
      stop(sprintf("length(%s) must be 1 or equal to the number of continuous variables (%d).",
                   nm, ncon))
    }
    if (length(x) == 1L)
      rep(x, ncon)
    else
      x
  }

  if (kerbound == "none") {
    lb <- rep(-Inf, ncon)
    ub <- rep(Inf, ncon)
  } else if (kerbound == "range") {
    lb <- mins
    ub <- maxs
  } else {
    lb <- recycleBounds(kerlb, paste0(argprefix, "lb"))
    ub <- recycleBounds(kerub, paste0(argprefix, "ub"))
  }

  if (any(!is.finite(lb) & !is.infinite(lb)) || any(!is.finite(ub) & !is.infinite(ub))) {
    stop(sprintf("'%s' and '%s' must be finite values or +/-Inf.",
                 paste0(argprefix, "lb"),
                 paste0(argprefix, "ub")))
  }

  if (any(lb >= ub)) {
    stop(sprintf("Invalid bounds for '%s': require lower < upper for every continuous variable.",
                 paste0(argprefix, "bound")))
  }

  bad.cover <- (lb > mins) | (ub < maxs)
  if (any(bad.cover)) {
    bad.vars <- paste(cnames[bad.cover], collapse = ", ")
    stop(sprintf("Invalid bounds for '%s': require lower <= min(training) and upper >= max(training) for each continuous variable. Violations: %s",
                 paste0(argprefix, "bound"), bad.vars))
  }

  if (kerbound == "fixed" && all(is.infinite(lb) & (lb < 0)) && all(is.infinite(ub) & (ub > 0))) {
    kerbound <- "none"
  }

  full.lb[icon.idx] <- lb
  full.ub[icon.idx] <- ub

  list(bound = kerbound, lb = full.lb, ub = full.ub)
}

npKernelBoundsCheckEval <- function(evaldat,
                                    icon,
                                    kerlb,
                                    kerub,
                                    argprefix = "cker") {
  if (is.null(evaldat) || is.null(icon) || !any(icon))
    return(invisible(TRUE))

  if (is.null(kerlb) || is.null(kerub))
    return(invisible(TRUE))

  icon.idx <- which(icon)
  ncon <- length(icon.idx)
  if (ncon == 0L)
    return(invisible(TRUE))

  lb <- as.numeric(kerlb[icon.idx])
  ub <- as.numeric(kerub[icon.idx])

  if (all(is.infinite(lb) & lb < 0) && all(is.infinite(ub) & ub > 0))
    return(invisible(TRUE))

  enames <- names(evaldat)
  if (is.null(enames) || length(enames) != length(icon))
    enames <- paste0("V", seq_along(icon))
  cnames <- enames[icon.idx]

  econ <- as.matrix(evaldat[, icon.idx, drop = FALSE])
  if (!is.double(econ))
    storage.mode(econ) <- "double"

  bad.low <- vapply(seq_len(ncon), function(i) {
    is.finite(lb[i]) && any(econ[, i] < lb[i], na.rm = TRUE)
  }, logical(1L))
  bad.high <- vapply(seq_len(ncon), function(i) {
    is.finite(ub[i]) && any(econ[, i] > ub[i], na.rm = TRUE)
  }, logical(1L))

  if (any(bad.low) || any(bad.high)) {
    low.msg <- if (any(bad.low)) {
      paste(sprintf("%s < %s", cnames[bad.low], format(lb[bad.low], digits = 12)), collapse = ", ")
    } else {
      NULL
    }
    high.msg <- if (any(bad.high)) {
      paste(sprintf("%s > %s", cnames[bad.high], format(ub[bad.high], digits = 12)), collapse = ", ")
    } else {
      NULL
    }
    msg <- paste(c(low.msg, high.msg), collapse = "; ")
    stop(sprintf("Evaluation data violate '%s' bounds: %s",
                 paste0(argprefix, "bound"), msg))
  }

  invisible(TRUE)
}

npKernelBoundsMarshal <- function(kerlb, kerub) {
  lb <- as.double(kerlb)
  ub <- as.double(kerub)
  big <- .Machine$double.xmax
  lb[is.infinite(lb) & lb < 0] <- -big
  ub[is.infinite(ub) & ub > 0] <- big
  list(lb = lb, ub = ub)
}

validateBandwidth <- function(bws){
  vari <- names(bws$bandwidth)
  bchecker <- function(j){
    v <- vari[j]
    dati <- bws$dati[[v]]
    bwv <- bws$bandwidth[[j]]
    stopifnot(length(bwv) == length(dati$iord))

    cd <- function(a,b){
      (a-b)/(a+b+.Machine$double.eps) > 5.0*.Machine$double.eps
    }
    
    nn.lower.bound <- if (!identical(v, "x")) 1L else npRegressionNnLowerBound(bws)

    vb <- sapply(seq_along(bwv), function(i){
      falg <- (bwv[i] < 0)

      if (dati$icon[i] && (falg || (!is.finite(bwv[i])))){
        stop(paste("Invalid bandwidth supplied for continuous",
                   "variable", bws$varnames[[v]][i], ":",bwv[i]))
      }

      if (dati$icon[i] &&
          !identical(bws$type, "fixed") &&
          (bwv[i] > 0) &&
          (bwv[i] < nn.lower.bound)) {
        stop(sprintf(
          "%s: nearest-neighbor bandwidth must be at least %d for continuous nonfixed regression",
          bws$varnames[[v]][i],
          nn.lower.bound
        ))
      }
      
      if (dati$iord[i] &&
          (falg || cd(bwv[i],oMaxL(dati$all.nlev[[i]],
                         kertype = bws$klist[[v]]$okertype)))){
        stop(paste("Invalid bandwidth supplied for ordered",
                   "variable", bws$varnames[[v]][i], ":",bwv[i]))
      }
      
      if (dati$iuno[i] &&
          (falg || cd(bwv[i],uMaxL(dati$all.nlev[[i]],
                         kertype = bws$klist[[v]]$ukertype)))){
        stop(paste("Invalid bandwidth supplied for unordered",
                   "variable", bws$varnames[[v]][i], ":",bwv[i]))
      }
      return(TRUE)
    })
    
    return(vb)
  }
  vbl <- lapply(seq_along(vari), bchecker)
  invisible(vbl)
}

validateBandwidthTF <- function(bws){
  vari <- names(bws$bandwidth)
  bchecker <- function(j){
    v <- vari[j]
    dati <- bws$dati[[v]]
    bwv <- bws$bandwidth[[j]]

    if(length(bwv) != length(dati$iord))
      return(FALSE)

    cd <- function(a,b){
      (a-b)/(a+b+.Machine$double.eps) > 5.0*.Machine$double.eps
    }
    
    nn.lower.bound <- if (!identical(v, "x")) 1L else npRegressionNnLowerBound(bws)

    vb <- sapply(seq_along(bwv), function(i){
      falg <- (bwv[i] < 0)

      if (dati$icon[i]) {
        if(bws$type == "fixed") {
          if(falg || (!is.finite(bwv[i]))){
            return(FALSE)
          }
        } else if((((bwv[i] > 0) && (bwv[i] < nn.lower.bound))) || (!is.finite(bwv[i]))) {
          return(FALSE)
        }
      }
      
      if (dati$iord[i] &&
          (falg || cd(bwv[i],oMaxL(dati$all.nlev[[i]],
                         kertype = bws$klist[[v]]$okertype)))){
        return(FALSE)
      }
      
      if (dati$iuno[i] &&
          (falg || cd(bwv[i],uMaxL(dati$all.nlev[[i]],
                         kertype = bws$klist[[v]]$ukertype)))){
        return(FALSE)
      }
      return(TRUE)
    })
    
    return(all(vb))
  }
  vbl <- all(unlist(lapply(seq_along(vari), bchecker)))
  return(vbl)
}

.np_refresh_xy_bandwidth_metadata <- function(bws) {
  if (is.null(bws$xbw) || is.null(bws$ybw) ||
      is.null(bws$xdati) || is.null(bws$ydati) ||
      is.null(bws$nconfac) || is.null(bws$ncatfac) ||
      is.null(bws$sdev) || is.null(bws$dati) || is.null(bws$klist)) {
    return(bws)
  }

  bandwidth <- list(x = bws$xbw, y = bws$ybw)
  sfactor <- bandwidth

  if ((bws$xnuno + bws$ynuno) > 0L) {
    if (isTRUE(bws$scaling)) {
      if (bws$xnuno > 0L)
        bandwidth$x[bws$xdati$iuno] <- bandwidth$x[bws$xdati$iuno] * bws$ncatfac
      if (bws$ynuno > 0L)
        bandwidth$y[bws$ydati$iuno] <- bandwidth$y[bws$ydati$iuno] * bws$ncatfac
    } else {
      if (bws$xnuno > 0L)
        sfactor$x[bws$xdati$iuno] <- sfactor$x[bws$xdati$iuno] / bws$ncatfac
      if (bws$ynuno > 0L)
        sfactor$y[bws$ydati$iuno] <- sfactor$y[bws$ydati$iuno] / bws$ncatfac
    }
  }

  if ((bws$xnord + bws$ynord) > 0L) {
    if (isTRUE(bws$scaling)) {
      if (bws$xnord > 0L)
        bandwidth$x[bws$xdati$iord] <- bandwidth$x[bws$xdati$iord] * bws$ncatfac
      if (bws$ynord > 0L)
        bandwidth$y[bws$ydati$iord] <- bandwidth$y[bws$ydati$iord] * bws$ncatfac
    } else {
      if (bws$xnord > 0L)
        sfactor$x[bws$xdati$iord] <- sfactor$x[bws$xdati$iord] / bws$ncatfac
      if (bws$ynord > 0L)
        sfactor$y[bws$ydati$iord] <- sfactor$y[bws$ydati$iord] / bws$ncatfac
    }
  }

  if ((bws$xncon + bws$yncon) > 0L) {
    sdev <- as.numeric(bws$sdev)
    sx <- if (bws$xncon > 0L) sdev[seq_len(bws$xncon)] else numeric(0L)
    sy <- if (bws$yncon > 0L) sdev[bws$xncon + seq_len(bws$yncon)] else numeric(0L)

    if (isTRUE(bws$scaling)) {
      if (bws$xncon > 0L)
        bandwidth$x[bws$xdati$icon] <- bandwidth$x[bws$xdati$icon] * (sx * bws$nconfac)
      if (bws$yncon > 0L)
        bandwidth$y[bws$ydati$icon] <- bandwidth$y[bws$ydati$icon] * (sy * bws$nconfac)
    } else {
      if (bws$xncon > 0L)
        sfactor$x[bws$xdati$icon] <- sfactor$x[bws$xdati$icon] / (sx * bws$nconfac)
      if (bws$yncon > 0L)
        sfactor$y[bws$ydati$icon] <- sfactor$y[bws$ydati$icon] / (sy * bws$nconfac)
    }
  }

  sumNum <- lapply(c("x", "y"), function(v) {
    dati <- bws$dati[[v]]
    sfv <- sfactor[[v]]
    sapply(seq_along(sfv), function(i) {
      if (dati$icon[i])
        return(sfv[i])
      if (dati$iord[i])
        return(oMaxL(dati$all.nlev[[i]], kertype = bws$klist[[v]]$okertype))
      if (dati$iuno[i])
        return(uMaxL(dati$all.nlev[[i]], kertype = bws$klist[[v]]$ukertype))
      NA_real_
    })
  })
  names(sumNum) <- c("x", "y")

  bws$bandwidth <- bandwidth
  bws$sfactor <- sfactor
  bws$sumNum <- sumNum
  bws
}

npRegressionNnLowerBound <- function(bws) {
  if (!inherits(bws, "rbandwidth"))
    return(1L)

  if (is.null(bws$type) || identical(bws$type, "fixed"))
    return(1L)

  if (is.null(bws$icon) || !any(bws$icon))
    return(1L)

  2L
}

npValidateRegressionNnLowerBound <- function(bws,
                                             where,
                                             allow.zero.placeholder = FALSE) {
  lower <- npRegressionNnLowerBound(bws)
  if (lower <= 1L)
    return(invisible(bws))

  bw <- as.double(bws$bw)
  icon <- which(as.logical(bws$icon))
  if (!length(icon))
    return(invisible(bws))

  offenders <- bw[icon] < lower
  if (allow.zero.placeholder)
    offenders <- offenders & (bw[icon] > 0)
  if (!any(offenders))
    return(invisible(bws))

  upper <- max(lower, as.integer(bws$nobs) - 1L)
  stop(
    sprintf(
      "%s: nearest-neighbor bandwidth must be in [%d, %d] for continuous nonfixed regression",
      where,
      lower,
      upper
    ),
    call. = FALSE
  )
}


explodeFormula <- function(formula, data=NULL){
  if(any(grepl("\\.",deparse(formula)))) {
      if(is.null(data)) stop("'.' in formula and no 'data' argument")
      formula <- terms(formula, data=data)
  }
  res <- strsplit(strsplit(paste(deparse(formula), collapse=""),
                           " *[~] *")[[1]], " *[+] *")
  stopifnot(all(sapply(res,length) > 0))
  names(res) <- c("response","terms")
  res
}


explodePipe <- function(formula, env = parent.frame()){
  if (!inherits(formula, "formula")) {
    if (is.symbol(formula) && is.environment(env)) {
      not_found <- .np_missing_binding_sentinel
      sym_val <- get0(as.character(formula), envir = env, inherits = TRUE, ifnotfound = not_found)
      if (!identical(sym_val, not_found)) {
        formula <- sym_val
      } else {
        out <- .np_try_eval_in_frames(formula, eval_env = env, search_frames = FALSE)
        if (isTRUE(out$ok))
          formula <- out$value
        else
          formula <- out$error
      }
    } else {
      out <- .np_try_eval_in_frames(formula, eval_env = env, search_frames = FALSE)
      if (isTRUE(out$ok))
        formula <- out$value
      else
        formula <- out$error
    }
    if (inherits(formula, "error"))
      stop(conditionMessage(formula), call. = FALSE)
  }
  tf <- as.character(formula)  
  tf <- tf[length(tf)]
  lhs <- if (length(as.character(formula)) == 3) {
    strsplit(as.character(formula)[2], " *[+] *")
  } else {
    list()
  }
  rhs <- strsplit(strsplit(tf, " *[|] *")[[1]], " *[+] *")
  c(lhs, rhs)
}

"%~%" <- function(a,b) {
  identical(class(a), class(b)) && (length(a) == length(b)) &&
  all(vapply(a, coarseclass, character(1)) == vapply(b, coarseclass, character(1)))
}

coarseclass <- function(a) {
  if (inherits(a, "integer")) return("numeric")
  return(class(a)[1])
}

toFrame <- function(frame) {
  if(!is.data.frame(frame)){
    t.names <- NULL

    if(!(is.vector(frame) || is.factor(frame) || is.matrix(frame)))
      stop(deparse(substitute(frame))," must be a data frame, matrix, vector, or factor")

    if(!is.matrix(frame))
      t.names <- paste(deparse(substitute(frame)),
                       collapse = "")
    
    frame <- data.frame(frame, check.names=FALSE)
    
    if(!is.null(t.names))
      names(frame) <- t.names
  }
  return(frame)
}


cast <- function(a, b, same.levels = TRUE){
  if(is.ordered(b)){
    if(same.levels)
      ordered(a, levels = levels(b))
    else
      ordered(a)
  }   
  else if(is.factor(b)){
    if(same.levels)
      factor(a, levels = levels(b))
    else
      factor(a)
  }
  else if (coarseclass(b) == "numeric")
    as.double(a)
  else if (is.data.frame(b)) {
    if (dim(a)[2] == dim(b)[2]){
      r = data.frame(a)
      for (i in seq_along(b))
        r[,i] = cast(a[,i],b[,i], same.levels = same.levels)
      r
    } else { stop("a could not be cast as b") }
  }
}

subcol <- function(x, v, i){
  x[,i] = cast(v,x[,i])
  x
}

mcvConstruct <- function(dati){
  nuno <- sum(dati$iuno)
  nord <- sum(dati$iord)

  num.row <- max(sapply(dati$all.lev,length))

  pad.num <- numNotIn(unlist(dati$all.dlev))

  mcv <- matrix(data = pad.num, nrow = num.row, ncol = (nuno+nord))
  attr(mcv, "num.row") <- num.row
  attr(mcv, "pad.num") <- pad.num

  cnt <- 0
  if (nuno > 0)
    for (i in which(dati$iuno)) 
      mcv[seq_along(dati$all.lev[[i]]), (cnt <- cnt+1)] <- dati$all.dlev[[i]]

  cnt <- 0
  if (nord > 0)
    for (i in which(dati$iord))
      mcv[seq_along(dati$all.lev[[i]]), (cnt <- cnt+1)+nuno] <- dati$all.dlev[[i]]

  mcv
}

## when admitting new categories, adjustLevels will attempt to catch possible mistakes:
## if an unordered variable contains more than one new category, warn
## if an ordered, but scaleless variable contains a new category, error
## if an ordered, scale-possessing variable contains a new category lacking scale, error

adjustLevels <- function(data, dati, allowNewCells = FALSE){
  for (i in which(dati$iord | dati$iuno)){
    if (allowNewCells){
      newCats <- setdiff(levels(data[,i]),dati$all.lev[[i]])
      if (length(newCats) >= 1){
        if (dati$iuno[i]){
          if (length(newCats) > 1)
            .np_warning(paste("more than one 'new' category is redundant when estimating on unordered data.\n",
                          "training data categories: ", paste(dati$all.lev[[i]], collapse=" "),"\n",
                          "redundant estimation data categories: ", paste(newCats, collapse=" "), "\n", sep=""))
          data[,i] <- factor(data[,i], levels = c(dati$all.lev[[i]], newCats))
        } else {
          if (dati$inumord[i]){
            if (!isNum(newCats))
              stop(paste("estimation data contains a new qualitative category, but training data is\n",
                         "ordered, and numeric.\n",
                         "training data categories: ", paste(dati$all.lev[[i]], collapse=" "),"\n",
                         "conflicting estimation data categories: ", paste(newCats, collapse=" "), "\n", sep=""))
          } else {
            stop(paste("estimation beyond the support of training data of an ordered,\n",
                       "categorical, qualitative variable is not supported.\n"))
          }

          data[,i] <- ordered(data[,i], levels = sort(as.numeric(c(dati$all.lev[[i]], newCats))))
        }
      } else {
        data[,i] <- factor(data[,i], levels = dati$all.lev[[i]])
      }
    } else {
      if (!all(is.element(levels(data[,i]), dati$all.lev[[i]])))
        stop("data contains unknown factors (wrong dataset provided?)")
      data[,i] <- factor(data[,i], levels = dati$all.lev[[i]])
    }
  }

  data
}

toMatrix <- function(data) {
  tq <- sapply(data, function(y) {
    if(is.factor(y))
      y <- dlev(y)[as.integer(y)]
    y})
  dim(tq) <- dim(data) ## cover the corner case of single element d.f.
  tq
}

## this doesn't just strictly check for the response, but does tell you
## that evaluating with response fails ... in principle the evaluation
## could fail without the response too, but then the calling routine is about
## to die a noisy death anyhow ...
succeedWithResponse <- function(tt, frame){
  vars <- attr(tt, "variables")
  if (is.null(vars))
    return(FALSE)
  out <- .np_try_eval_in_frames(vars, eval_env = frame, enclos = NULL, search_frames = FALSE)
  isTRUE(out$ok)
}

## determine whether a bandwidth
## matches a data set
bwMatch <- function(data, dati){
  if (length(dati$icon) != ncol(data))
    stop("bandwidth vector is of improper length")

  test.dati <- untangle(data)

  if (any(xor(dati$icon,test.dati$icon)) ||
      any(xor(dati$iord,test.dati$iord)) ||
      any(xor(dati$iuno,test.dati$iuno)))
    stop(paste("supplied bandwidths do not match","data", "in type"))
}

uMaxL <- function(c, kertype = c("aitchisonaitken","liracine")){
  switch(kertype,
         aitchisonaitken = (c-1.0)/c,
         liracine = 1.0)
}

oMaxL <- function(c, kertype = c("wangvanryzin", "liracine", "nliracine", "racineliyan")){
  switch(kertype,
         wangvanryzin = 1.0,
         liracine = 1.0,
         nliracine = 1.0,
         "racineliyan" = 1.0)
}

## tested with: rbandwidth
## right now all bandwidth objects have some crusty
## vestiges of their evolution, ie. non-list metadata
## such as xnames or ynames. The new metadata system is
## for the most part list based and facilitates generic
## operations

updateBwNameMetadata <- function(nameList, bws){
  ## names of 'old' metadata in bw object
  onames <- names(nameList)
  if (length(onames) == 0L)
    return(bws)

  for (nm in onames) {
    bws[[nm]] <- nameList[[nm]]
    bws$varnames[[substr(nm, 1L, 1L)]] <- nameList[[nm]]
  }

  bws
}

## some string utility functions

pad <- function(s){
  idx <- nchar(s) > 0
  out <- rep.int(" ", length(s))
  out[idx] <- paste("", s[idx], "")
  names(out) <- names(s)
  out
}

rpad <- function(s){
  idx <- nchar(s) > 0
  out <- rep.int("", length(s))
  out[idx] <- paste(s[idx], "")
  names(out) <- names(s)
  out
}

blank <- function(len){
  sapply(len, function(nb){
    paste(rep(' ', times = nb), collapse='')
  })
}

formatv <- function(v){
  sapply(seq_along(v), function(j){ format(v[j]) })
}

## strings used in various report generating functions

genOmitStr <- function(x){
  t.str <- ''
  if(!is.null(x$rows.omit) && !identical(x$rows.omit, NA))
    t.str <- paste("\nNo. Complete Observations: ", x$nobs,
                   "\nNo. Incomplete (NA) Observations: ", x$nobs.omit,
                   "\nObservations omitted or excluded: ", paste(x$rows.omit, collapse=" "),
                   "\n")
  return(t.str)
}

## Estimation-related rgf's
genGofStr <- function(x){
  mse.str <- ""
  if (!is.na(x$MSE))
    mse.str <- paste("\nResidual standard error:", format(sqrt(x$MSE)))

  r2.str <- ""
  if (!is.na(x$R2))
    r2.str <- paste("\nR-squared:", format(x$R2))

  paste(mse.str, r2.str, sep = "")
}

genTimingStr <- function(x){
  if (is.null(x$total.time) || is.na(x$total.time))
    return("")

  nomad.time <- if (!is.null(x$nomad.time) && !is.na(x$nomad.time))
    as.double(x$nomad.time) else NA_real_
  powell.time <- if (!is.null(x$powell.time) && !is.na(x$powell.time))
    as.double(x$powell.time) else NA_real_
  fit.time <- if (!is.null(x$fit.time) && !is.na(x$fit.time))
    as.double(x$fit.time) else NA_real_

  if (is.finite(nomad.time) || is.finite(powell.time)) {
    detail <- character(0)
    if (is.finite(nomad.time))
      detail <- c(detail, paste("NOMAD ", format(nomad.time), "s", sep = ""))
    if (is.finite(powell.time))
      detail <- c(detail, paste("Powell ", format(powell.time), "s", sep = ""))
    if (is.finite(fit.time))
      detail <- c(detail, paste("fit ", format(fit.time), "s", sep = ""))

    return(paste("\nEstimation Time: ", format(x$total.time), " seconds (",
                 paste(detail, collapse = ", "), ")", sep = ""))
  }

  if (!is.null(x$optim.time) && !is.na(x$optim.time) &&
      !is.null(x$fit.time) && !is.na(x$fit.time))
    return(paste("\nEstimation Time: ", format(x$total.time), " seconds (optim ",
                 format(x$optim.time), "s, fit ", format(x$fit.time), "s)", sep = ""))

  paste("\nEstimation Time: ",format(x$total.time)," seconds",sep = "")
}

.np_attach_nomad_restart_summary <- function(bws, search_result, tol = 1e-10) {
  restart.results <- search_result$restart.results
  direction <- if (!is.null(search_result$direction) && length(search_result$direction)) {
    as.character(search_result$direction[1L])
  } else {
    "min"
  }
  restart.fval <- if (!is.null(restart.results) && length(restart.results)) {
    vapply(
      restart.results,
      function(x) {
        if (is.null(x$objective) || !length(x$objective)) NA_real_ else as.numeric(x$objective[1L])
      },
      numeric(1L)
    )
  } else {
    numeric(0L)
  }

  best.objective <- if (!is.null(search_result$best$objective) && length(search_result$best$objective)) {
    as.numeric(search_result$best$objective[1L])
  } else {
    NA_real_
  }

  best.restart <- NA_integer_
  if (!is.null(search_result$best.restart) &&
      length(search_result$best.restart) &&
      is.finite(search_result$best.restart[1L])) {
    best.restart <- as.integer(search_result$best.restart[1L])
  } else if (length(restart.fval) && is.finite(best.objective)) {
    diffs <- abs(restart.fval - best.objective)
    diffs[!is.finite(diffs)] <- Inf
    if (any(is.finite(diffs))) {
      idx <- which.min(diffs)
      scale <- max(1, abs(best.objective))
      if (is.finite(diffs[idx]) && diffs[idx] <= tol * scale)
        best.restart <- as.integer(idx)
    }
  } else if (length(restart.fval) && any(is.finite(restart.fval))) {
    best.restart <- as.integer(if (identical(direction, "max")) which.max(restart.fval) else which.min(restart.fval))
  }

  bws$nomad.restart.results <- restart.results
  bws$nomad.restart.fval <- restart.fval
  bws$nomad.best.restart <- best.restart
  bws$nomad.restart.starts <- search_result$restart.starts
  bws$nomad.restart.degree.starts <- search_result$restart.degree.starts
  bws$nomad.restart.bandwidth.starts <- search_result$restart.bandwidth.starts
  bws$nomad.remin <- isTRUE(search_result$nomad.remin)
  bws$nomad.remin.index <- search_result$nomad.remin.index
  bws$nomad.remin.roundtrip <- search_result$nomad.remin.roundtrip
  bws
}

.np_multistart_label <- function(x, tol = sqrt(.Machine$double.eps)) {
  if (!is.null(x$nomad.best.restart) &&
      length(x$nomad.best.restart) == 1L &&
      is.finite(x$nomad.best.restart)) {
    idx <- as.numeric(x$nomad.best.restart[1L])
    if (abs(idx - round(idx)) <= tol && idx >= 1)
      return(as.integer(round(idx)))
  }

  if (!is.null(x$ifval) &&
      length(x$ifval) == 1L &&
      is.finite(x$ifval)) {
    idx <- as.numeric(x$ifval[1L])
    if (abs(idx - round(idx)) <= tol) {
      idx <- as.integer(round(idx))
      return(if (idx <= 0L) 1L else idx)
    }
  }

  NA_integer_
}
  
pCatGofStr <- function(x){
  if(!identical(x$confusion.matrix, NA)){
    cat("\nConfusion Matrix\n")
    print(x$confusion.matrix)
  }

  if (!identical(x$CCR.overall,NA))
    cat("\nOverall Correct Classification Ratio: ", format(x$CCR.overall))

  if (!identical(x$CCR.byoutcome,NA)){
    cat("\nCorrect Classification Ratio By Outcome:\n")
    print(x$CCR.byoutcome)
  }

  if (!identical(x$fit.mcfadden,NA))
    cat("\nMcFadden-Puig-Kerschner performance measure: ", format(x$fit.mcfadden))

}

genDenEstStr <- function(x){
  loglik.str <- ""
  if (!(is.null(x$log_likelihood) || identical(x$log_likelihood, NA))) {
    loglik.str <- paste("\nLog Likelihood:", format(x$log_likelihood))
  }
  paste("\nBandwidth Type: ", x$ptype, loglik.str, sep = "")
}

genRegEstStr <- function(x){
  regtype <- if (!is.null(x$regtype)) x$regtype else if (!is.null(x$bws)) x$bws$regtype else NULL
  basis <- if (!is.null(x$basis)) x$basis else if (!is.null(x$bws)) x$bws$basis else NULL
  bern <- if (!is.null(x$bernstein.basis)) x$bernstein.basis else if (!is.null(x$bws)) x$bws$bernstein.basis else NULL
  est.label <- if (identical(regtype, "lp")) npFormatRegressionType(x) else x$pregtype
  basis.family <- if (identical(regtype, "lp")) npLpBasisFamilyLabel(basis) else NULL
  basis.rep <- if (identical(regtype, "lp")) npLpBasisRepresentationLabel(bern) else NULL
  est.label.str <- if (is.null(est.label)) "" else paste("\nKernel Regression Estimator:", est.label)
  basis.family.str <- if (is.null(basis.family)) "" else paste("\nLP Basis Family:", basis.family)
  basis.rep.str <- if (is.null(basis.rep)) "" else paste("\nLP Basis Representation:", basis.rep)
  ptype.str <- if (is.null(x$ptype)) "" else paste("\nBandwidth Type:", x$ptype)
  tau.str <- if (is.null(x$tau)) "" else paste("\nTau:", paste(x$tau, collapse = ", "))
  paste(est.label.str, basis.family.str, basis.rep.str, ptype.str, tau.str, sep = "")
}

npLpBasisFamilyLabel <- function(basis){
  basis.value <- if (is.null(basis) || !length(basis)) "glp" else basis
  b <- tolower(basis.value)
  switch(b,
         glp = "Generalized",
         additive = "Additive",
         tensor = "Tensor",
         tools::toTitleCase(b))
}

npLpBasisRepresentationLabel <- function(bernstein){
  if (isTRUE(bernstein)) "Bernstein" else "Raw"
}

npLpBasisNcol <- function(basis = "glp", degree){
  if (is.null(degree) || !length(degree))
    return(NA_real_)
  d <- dim_basis(basis = basis,
                 kernel = TRUE,
                 degree = as.integer(degree),
                 segments = rep.int(1L, length(degree)))
  if (identical(tolower(basis), "tensor")) d else d + 1.0
}

npFormatRegressionType <- function(x){
  regtype <- if (!is.null(x$regtype)) {
    x$regtype
  } else if (!is.null(x$bws) && !is.null(x$bws$regtype)) {
    x$bws$regtype
  } else {
    NULL
  }

  pregtype <- if (!is.null(x$pregtype)) {
    x$pregtype
  } else if (!is.null(x$bws) && !is.null(x$bws$pregtype)) {
    x$bws$pregtype
  } else {
    NULL
  }

  if (!identical(regtype, "lp"))
    return(pregtype)

  degree <- if (!is.null(x$degree)) {
    x$degree
  } else if (!is.null(x$bws) && !is.null(x$bws$degree)) {
    x$bws$degree
  } else {
    NULL
  }

  if (is.null(degree) || length(degree) == 0)
    return("Local-Polynomial")

  basis <- if (!is.null(x$basis)) {
    x$basis
  } else if (!is.null(x$bws) && !is.null(x$bws$basis)) {
    x$bws$basis
  } else {
    "glp"
  }
  basis.family <- npLpBasisFamilyLabel(basis)

  if (!is.null(x$child.degree.common) && !isTRUE(x$child.degree.common))
    return(sprintf("Local-Polynomial (%s basis; child-specific degree)",
                   basis.family))

  sprintf("Local-Polynomial (%s basis; degree = %s)",
          basis.family, paste(degree, collapse = ","))
}

npBandwidthSummaryLabel <- function(bwtype, bwscaling = FALSE){
  if (isTRUE(bwscaling))
    return("Scale Factor(s)")

  if (identical(bwtype, "fixed"))
    return("Bandwidth(s)")

  "Bandwidth Nearest Neighbor(s)"
}


## bandwidth-related report generating functions
genBwSelStr <- function(x){
  fval.str <- ""
  if (!identical(x$fval, NA)) {
    ms.label <- .np_multistart_label(x)
    fval.str <- if (is.na(ms.label)) {
      paste("\nObjective Function Value: ", format(x$fval), sep = "")
    } else {
      paste("\nObjective Function Value: ", format(x$fval),
            " (achieved on multistart ", ms.label, ")", sep = "")
    }
  }

  nfe.str <- ""
  if(!(is.null(x$num.feval) || (length(x$num.feval) == 1L && is.na(x$num.feval)))){
    nfe.str <- paste("\nNumber of Function Evaluations: ", format(x$num.feval), sep="")
    if(!(is.null(x$num.feval.fast) || (length(x$num.feval.fast) == 1L && is.na(x$num.feval.fast)))){
      nfe.str <- paste(nfe.str, " (fast = ", format(x$num.feval.fast), ")", sep="")
    }
  }

  pregtype <- npFormatRegressionType(x)

  pregtype.str <- if (is.null(pregtype)) "" else paste("\nRegression Type:", pregtype)
  pmethod.str <- if (is.null(x$pmethod)) "" else paste("\nBandwidth Selection Method:", x$pmethod)
  formula.str <- if (!identical(x$formula, NULL)) paste("\nFormula:", paste(deparse(x$formula), collapse = "\n")) else ""
  ptype.str <- if (is.null(x$ptype)) "" else paste("\nBandwidth Type: ", x$ptype, sep = "")

  paste(pregtype.str,
        pmethod.str,
        formula.str,
        ptype.str,
        fval.str,
        nfe.str,
        sep = "")
}

genBwScaleStrs <- function(x){
  ## approach is to take metadata and flatten it so it then can be
  ## processed into a single string

  vari <- names(x$sumNum)

  flat_varnames <- unlist(x$varnames[vari], use.names = FALSE)
  flat_vartitles <- unlist(
    lapply(vari, function(v) rep.int(x$vartitleabb[[v]], length(x$varnames[[v]]))),
    use.names = FALSE
  )
  flat_bandwidth <- unlist(x$bandwidth[vari], use.names = FALSE)
  flat_sum <- unlist(x$sumNum[vari], use.names = FALSE)
  flat_icon <- unlist(lapply(vari, function(v) x$dati[[v]]$icon), use.names = FALSE)

  sumText <- lapply(seq_along(flat_varnames), function(i) {
    if (isTRUE(flat_icon[[i]])) {
      if (x$type == "fixed") "Scale Factor:" else ""
    } else {
      "Lambda Max:"
    }
  })

  maxNameLen <- max(nchar(unlist(sumText)))
  print.sumText <- lapply(sumText, '!=', "")

  sumText <- lapply(seq_along(sumText), function(i){
    paste(blank(maxNameLen - nchar(sumText[[i]])), sumText[[i]], sep="")
  })

  t.nchar <- lapply(flat_varnames, nchar)
                    
  maxNameLen <- max(unlist(t.nchar))

  vatText <- lapply(seq_along(t.nchar), function(j){
    paste("\n", rpad(flat_vartitles[[j]]), "Var. Name: ",
          flat_varnames[[j]],
          sapply(t.nchar[[j]], function(nc){
            paste(rep(' ', maxNameLen - nc), collapse='')
          }), sep='')
  })

  return(sapply(seq_along(t.nchar), function(j){
    sum.str <- ""
    if (isTRUE(print.sumText[[j]]))
      sum.str <- paste(sumText[[j]], " ", npFormat(flat_sum[[j]]), sep = "")
    paste(vatText[[j]], " Bandwidth: ", npFormat(flat_bandwidth[[j]]), " ",
          sum.str, sep = "", collapse = "")
  }))
}

npFormat <- function(x){
  format(sapply(x,format))
}

genBwKerStrs <- function(x){
  vari <- names(x$klist)

  ncon <- sapply(vari, function(v){
    sum(x$dati[[v]]$icon)
  })

  nuno <- sapply(vari, function(v){
    sum(x$dati[[v]]$iuno)
  })

  nord <- sapply(vari, function(v){
    sum(x$dati[[v]]$iord)
  })

  cktype <- sapply(vari, function(v){
    x$klist[[v]]$ckertype
  })

  uktype <- sapply(vari, function(v){
    x$klist[[v]]$ukertype
  })

  oktype <- sapply(vari, function(v){
    x$klist[[v]]$okertype
  })

  tt <- ''

  if(any(ncon > 0)){
    ctype.str <- ""
    if (length(unique(cktype)) == 1) {
      ctype.str <- paste("\nContinuous Kernel Type:",
                         x$klist[[vari[1]]]$pckertype)
    } else {
      ctype.str <- paste(sapply(seq_along(vari), function(v){
        if (ncon[v] > 0)
          paste("\nContinuous Kernel Type (",
                x$vartitleabb[[vari[v]]],
                " Var.): ", x$klist[[vari[v]]]$pckertype, sep = "")
        else ""
      }), collapse = "")
    }
    tt <- paste("\n", ctype.str, sep = "")
    cont.str <- paste(sapply(seq_along(vari), function(i){
      if (ncon[i] > 0)
        paste("\nNo. Continuous", pad(x$vartitle[[vari[i]]]), "Vars.: ",
              ncon[i], sep = "")
      else ""
    }), collapse = "")
    tt <- paste(tt, cont.str, sep = "")
  }
                
    
  if(any(nuno > 0)) {
    utype.str <- ""
    if (length(unique(uktype)) == 1) {
      utype.str <- paste("\nUnordered Categorical Kernel Type:",
                         x$klist[[vari[1]]]$pukertype)
    } else {
      utype.str <- paste(sapply(seq_along(vari), function(i){
        if (nuno[i] > 0)
          paste("\nUnordered Categorical Kernel Type (",
                x$vartitleabb[[vari[i]]],
                " Var.): ", x$klist[[vari[i]]]$pukertype, sep = "")
        else ""
      }), collapse = "")
    }
    tt <- paste(tt, "\n", utype.str, sep = "")
    uno.str <- paste(sapply(seq_along(vari), function(i){
      if (nuno[i] > 0)
        paste("\nNo. Unordered Categorical", pad(x$vartitle[[vari[i]]]), "Vars.: ",
              nuno[i], sep = "")
      else ""
    }), collapse = "")
    tt <- paste(tt, uno.str, sep = "")

  }

  if(any(nord > 0)) {
    otype.str <- ""
    if (length(unique(oktype)) == 1) {
      otype.str <- paste("\nOrdered Categorical Kernel Type:",
                         x$klist[[vari[1]]]$pokertype)
    } else {
      otype.str <- paste(sapply(seq_along(vari), function(i){
        if (nord[i] > 0)
          paste("\nOrdered Categorical Kernel Type (",
                x$vartitleabb[[vari[i]]],
                " Var.): ", x$klist[[vari[i]]]$pokertype, sep = "")
        else ""
      }), collapse = "")
    }
    tt <- paste(tt, "\n", otype.str, sep = "")
    ord.str <- paste(sapply(seq_along(vari), function(i){
      if (nord[i] > 0)
        paste("\nNo. Ordered Categorical", pad(x$vartitle[[vari[i]]]), "Vars.: ",
              nord[i], sep = "")
      else ""
    }), collapse = "")
    tt <- paste(tt, ord.str, sep = "")

  }

  return(tt)
}

genBwKerStrsXY <- function(x){
  t.str <- ''
  cnt <- 0
  
  if (x$xncon + x$yncon > 0){
    cker.str <- ""
    if (x$pcxkertype == x$pcykertype) {
      cker.str <- paste("\n\nContinuous Kernel Type:", x$pcxkertype)
    } else {
      exp.str <- if (x$xncon > 0)
        paste("\nContinuous Kernel Type (Exp. Var.):", x$pcxkertype)
      else ""
      dep.str <- if (x$yncon > 0)
        paste("\nContinuous Kernel Type (Dep. Var.):", x$pcykertype)
      else ""
      cker.str <- paste("\n", exp.str, dep.str)
    }
    ycon.str <- if (x$yncon > 0) paste("\nNo. Continuous Dependent Vars.:", x$yncon) else ""
    xcon.str <- if (x$xncon > 0) paste("\nNo. Continuous Explanatory Vars.:", x$xncon) else ""
    t.str[cnt <- cnt + 1] <-
      paste(cker.str, ycon.str, xcon.str)
  }

  if (x$xnuno + x$ynuno > 0){
    uker.str <- ""
    if (x$puxkertype == x$puykertype) {
      uker.str <- paste("\n\nUnordered Categorical Kernel Type:", x$puxkertype)
    } else {
      exp.str <- if (x$xnuno > 0)
        paste("\nUnordered Categorical Kernel Type (Exp. Var.):", x$puxkertype)
      else ""
      dep.str <- if (x$ynuno > 0)
        paste("\nUnordered Categorical Kernel Type (Dep. Var.):", x$puykertype)
      else ""
      uker.str <- paste("\n", exp.str, dep.str)
    }
    yuno.str <- if (x$ynuno > 0) paste("\nNo. Unordered Categorical Dependent Vars.:", x$ynuno) else ""
    xuno.str <- if (x$xnuno > 0) paste("\nNo. Unordered Categorical Explanatory Vars.:", x$xnuno) else ""
    t.str[cnt <- cnt + 1] <-
      paste(uker.str, yuno.str, xuno.str)
  }

  if (x$xnord + x$ynord > 0){
    oker.str <- ""
    if (x$poxkertype == x$poykertype) {
      oker.str <- paste("\n\nOrdered Categorical Kernel Type:", x$poxkertype)
    } else {
      exp.str <- if (x$xnord > 0)
        paste("\nOrdered Categorical Kernel Type (Exp. Var.):", x$poxkertype)
      else ""
      dep.str <- if (x$ynord > 0)
        paste("\nOrdered Categorical Kernel Type (Dep. Var.):", x$poykertype)
      else ""
      oker.str <- paste("\n", exp.str, dep.str)
    }
    yord.str <- if (x$ynord > 0) paste("\nNo. Ordered Categorical Dependent Vars.:", x$ynord) else ""
    xord.str <- if (x$xnord > 0) paste("\nNo. Ordered Categorical Explanatory Vars.:", x$xnord) else ""
    t.str[cnt <- cnt + 1] <-
      paste(oker.str, yord.str, xord.str)
  }
  return(t.str)
}

genBwGOFStrs <- function(x) {
  ###paste("Residual standard error:", sqrt
}
  
## statistical functions
RSQfunc <- function(y,y.pred) {
  y.mean <- mean(y)
  (sum((y-y.mean)*(y.pred-y.mean))^2)/(sum((y-y.mean)^2)*sum((y.pred-y.mean)^2))
}

MSEfunc <- function(y,y.fit) {
  mean((y-y.fit)^2)
}

MAEfunc <- function(y,y.fit) {
  mean(abs(y-y.fit))
}

MAPEfunc <- function(y,y.fit) {
  jj = which(y != 0)
  
  mean(c(abs((y[jj]-y.fit[jj])/y[jj]), as.numeric(replicate(length(y)-length(jj),2))))
}

CORRfunc <- function(y,y.fit) {
  abs(corr(cbind(y,y.fit)))
}

SIGNfunc <- function(y,y.fit) {
  sum(sign(y) == sign(y.fit))/length(y)
}

EssDee <- function(y){
  if(any(dim(as.matrix(y)) == 0))
      return(0)
  sd.vec <- apply(as.matrix(y),2,sd)
  IQR.vec <- apply(as.matrix(y),2,IQR)/QFAC
  mad.vec <- apply(as.matrix(y),2,mad)
  a <- apply(cbind(sd.vec,IQR.vec,mad.vec),1, function(x) max(x))
  if(any(a<=0)) .np_warning(paste("variable ",which(a<=0)," appears to be constant",sep=""))
  a <- apply(cbind(sd.vec,IQR.vec,mad.vec),1, function(x) min(x[x>0]))  
  return(a)
}

.npConditionalNomadContScale <- function(ycon, xcon, iycon, ixcon, nconfac, where) {
  y_scale <- if (length(which(iycon)) > 0L) EssDee(ycon) else numeric(0L)
  x_scale <- if (length(which(ixcon)) > 0L) EssDee(xcon) else numeric(0L)
  cont_scale <- c(y_scale, x_scale) * nconfac
  expected <- length(which(iycon)) + length(which(ixcon))

  if (length(cont_scale) != expected) {
    stop(sprintf(
      "%s: internal NOMAD bandwidth scale mismatch (%d scale values for %d continuous bandwidth slots)",
      where, length(cont_scale), expected
    ))
  }

  cont_scale
}

.npAssertConditionalNomadSetup <- function(setup, where) {
  if (length(setup$cont_scale) != length(setup$cont_flat)) {
    stop(sprintf(
      "%s: internal NOMAD bandwidth transform mismatch (%d continuous scales for %d continuous slots)",
      where, length(setup$cont_scale), length(setup$cont_flat)
    ))
  }
  if (length(setup$cat_upper) != length(setup$cat_flat)) {
    stop(sprintf(
      "%s: internal NOMAD bandwidth transform mismatch (%d categorical bounds for %d categorical slots)",
      where, length(setup$cat_upper), length(setup$cat_flat)
    ))
  }
  invisible(setup)
}


##EssDee <- function(y){
##
##  sd.vec <- apply(as.matrix(y),2,sd)
##  IQR.vec <- apply(as.matrix(y),2,IQR)/QFAC
##  return(ifelse(sd.vec<IQR.vec|IQR.vec==0,sd.vec,IQR.vec))
##  
##}

## consolidating various bits of code related to converting internal settings
## to printable strings

bwmToPrint <- function(s){
  switch(s,
         manual = "Manual",
         cv.aic = "Expected Kullback-Leibler Cross-Validation",
         cv.ml = "Maximum Likelihood Cross-Validation",
         cv.cdf = "Least Squares Cross-Validation",
         cv.ls = "Least Squares Cross-Validation",
         "normal-reference" = "Normal Reference")
}

bwtToPrint <- function(s){
  switch(s,
         fixed = "Fixed",
         generalized_nn = "Generalized Nearest Neighbour",
         adaptive_nn = "Adaptive Nearest Neighbour" )
}

cktToPrint <- function(s, order = "", kerbound = "none"){
  pck <- switch(s,
                gaussian = paste(order,"Gaussian"),
                epanechnikov =  paste(order,"Epanechnikov"),
                uniform = "Uniform",
                "truncated gaussian" = "Truncated Gaussian")
  if (!is.null(kerbound) && !identical(kerbound, "none"))
    pck <- paste0(pck, " (bounded/", kerbound, ")")
  pck
}

uktToPrint <- function(s){
  switch(s,
         aitchisonaitken = "Aitchison and Aitken",
         liracine = "Li and Racine (normalized)")
}

oktToPrint <- function(s, normalized = FALSE) {
  if(normalized){
    pok <- 
      switch(s,
             wangvanryzin = "Wang and Van Ryzin", 
             liracine = "Li and Racine (normalized)",
             nliracine = "Li and Racine (normalized)",
             "racineliyan" = "Racine, Li, and Yan")
  } else {
    pok <- 
      switch(s,
             wangvanryzin = "Wang and Van Ryzin", 
             liracine = "Li and Racine",
             nliracine = "Li and Racine (normalized)",
             "racineliyan" = "Racine, Li, and Yan")
  }
  return(pok)
}

### holding place for some generic methods

se <- function(x){
  UseMethod("se",x)
}

gradients <- function(x, ...){
  UseMethod("gradients",x)
}

.np_reject_unused_dots <- function(dots, where) {
  if (length(dots) == 0L)
    return(invisible(TRUE))

  dot.names <- names(dots)
  if (is.null(dot.names))
    dot.names <- rep("", length(dots))

  labels <- ifelse(nzchar(dot.names), dot.names, "<unnamed>")
  labels <- paste(sprintf("'%s'", labels), collapse = ", ")
  stop(sprintf("unused argument%s in %s: %s",
               if (length(dots) == 1L) "" else "s",
               where,
               labels),
       call. = FALSE)
}

## From crs to avoid crs:::W.glp

mypoly <- function(x,
                   ex=NULL,
                   degree,
                   gradient.compute = FALSE,
                   r=0,
                   Bernstein = TRUE) {

  if(missing(x)) stop(" Error: x required")
  if(missing(degree)) stop(" Error: degree required")
  if(degree < 1) stop(" Error: degree must be a positive integer")
  if(!is.logical(Bernstein)) stop(" Error: Bernstein must be logical")

  if(!Bernstein) {

    ## Raw polynomials and their derivatives

    if(!is.null(ex)) x <- ex

    if(gradient.compute) {
      Z <- NULL
      for(i in seq_len(degree)) {
        if((i-r) >= 0) {
          tmp <- (factorial(i)/factorial(i-r))*x^max(0,i-r)
        } else {
          tmp <- rep(0,length(x))
        }
        Z <- cbind(Z,tmp)
      }
    } else {
      Z <- outer(x,1L:degree,"^")
    }

  } else {
    ## Bernstein polynomial basis (no interior knots) aligned with C-side LP code.
    x.train <- as.double(x)
    x.use <- if (is.null(ex)) x.train else as.double(ex)
    r <- if (gradient.compute) as.integer(r) else 0L
    if (is.na(r) || r < 0L) stop(" Error: derivative order must be a non-negative integer")

    xmin <- min(x.train)
    xmax <- max(x.train)
    xrange <- xmax - xmin
    if (!is.finite(xrange) || xrange <= 0) {
      Z <- matrix(0.0, nrow = length(x.use), ncol = degree)
    } else if (r > degree) {
      Z <- matrix(0.0, nrow = length(x.use), ncol = degree)
    } else {
      u <- (x.use - xmin) / xrange
      m <- degree - r
      coef.deriv <- if (r == 0L) {
        1.0
      } else {
        exp(lfactorial(degree) - lfactorial(m)) / (xrange^r)
      }

      z.list <- lapply(seq_len(degree), function(idx) {
        out <- rep(0.0, length(u))
        for (k in 0:r) {
          j <- idx - k
          if (j >= 0L && j <= m) {
            out <- out + ((-1.0)^(r - k)) * choose(r, k) * choose(m, j) * (u^j) * ((1.0 - u)^(m - j))
          }
        }
        coef.deriv * out
      })
      Z <- do.call(cbind, z.list)
    }
  }

  return(as.matrix(Z))

}

## W.lp accepts a vector of degrees and provides local-polynomial bases
## with selectable term structure (glp/additive/tensor).

npBuildLpTerms <- function(degree, basis = c("glp", "additive", "tensor")) {
  basis <- match.arg(basis)
  k <- length(degree)
  if (k == 0L)
    return(matrix(integer(0), nrow = 1L, ncol = 0L))

  degree <- as.integer(degree)
  degree.list <- lapply(degree, function(d) 0:d)
  z <- as.matrix(do.call(base::expand.grid, degree.list))
  s <- rowSums(z)

  if (identical(basis, "glp")) {
    ind <- (s > 0) & (s <= max(degree))
    z <- z[ind, , drop = FALSE]
    if (!all(degree == max(degree))) {
      for (j in seq_along(degree)) {
        d <- degree[j]
        if ((d < max(degree)) && (d > 0)) {
          s <- rowSums(z)
          dropj <- (s > d) & (z[, j, drop = FALSE] == matrix(d, nrow(z), 1, byrow = TRUE))
          z <- z[!dropj, , drop = FALSE]
        }
      }
    }
  } else if (identical(basis, "additive")) {
    ind <- (s > 0) & (rowSums(z > 0) == 1L)
    z <- z[ind, , drop = FALSE]
  } else if (identical(basis, "tensor")) {
    z <- z[s > 0, , drop = FALSE]
  }

  rbind(matrix(0L, nrow = 1L, ncol = k), z)
}

W.lp <- function(xdat = NULL,
                 exdat = NULL,
                 degree = NULL,
                 gradient.vec = NULL,
                 basis = c("glp", "additive", "tensor"),
                 bernstein.basis = TRUE,
                 Bernstein = bernstein.basis) {

  if(is.null(xdat)) stop(" Error: You must provide data")
  if(is.null(degree) || any(degree < 0)) stop(paste(" Error: degree vector must contain non-negative integers\ndegree is (", degree, ")\n",sep=""))
  basis <- match.arg(basis)

  xdat <- as.data.frame(xdat)

  xdat.col.numeric <- vapply(seq_len(ncol(xdat)), function(i) is.numeric(xdat[,i]), logical(1))
  k <- ncol(as.data.frame(xdat[,xdat.col.numeric]))

  xdat.numeric <- NULL
  if(k > 0) {
    xdat.numeric <- as.data.frame(xdat[,xdat.col.numeric])
    if(!is.null(exdat)) {
      exdat.numeric <- as.data.frame(exdat[,xdat.col.numeric])
    } else {
      exdat.numeric <- NULL
    }
  }

  if(!is.null(gradient.vec) && (length(gradient.vec) != k)) stop(paste(" Error: gradient vector and number of numeric predictors must be conformable\n",sep=""))
  if(!is.null(gradient.vec) && any(gradient.vec < 0)) stop(paste(" Error: gradient vector must contain non-negative integers\n",sep=""))

  if(!is.null(gradient.vec)) {
    gradient.compute <- TRUE
  } else {
    gradient.compute <- FALSE
    gradient.vec <- rep(NA,k)
  }

  if(length(degree) != k) stop(" Error: degree vector and number of numeric predictors incompatible")

  if(all(degree == 0) || (k == 0)) {

    ## Local constant OR no continuous variables

    if(is.null(exdat)) {
      return(matrix(1,nrow=nrow(as.data.frame(xdat)),ncol=1))
    } else {
      return(matrix(1,nrow=nrow(as.data.frame(exdat)),ncol=1))
    }

  } else {

    z <- npBuildLpTerms(degree = degree, basis = basis)
    z.noi <- z[-1L, , drop = FALSE]
    if(is.null(exdat)) {
      res <- rep.int(1,nrow(xdat.numeric))
    } else {
      res <- rep.int(1,nrow(exdat.numeric))
    }
    res.deriv <- 1
    if(degree[1] > 0) {
      res <- cbind(1, mypoly(x=xdat.numeric[,1],
                             ex=exdat.numeric[,1],
                             degree=degree[1],
                             gradient.compute=gradient.compute,
                             r=gradient.vec[1],
                             Bernstein=Bernstein))[, 1 + z.noi[, 1]]

      if(gradient.compute && gradient.vec[1] != 0) res.deriv <- cbind(1,matrix(NA,1,degree[1]))[, 1 + z.noi[, 1],drop=FALSE]
      if(gradient.compute && gradient.vec[1] == 0) res.deriv <- cbind(1,matrix(0,1,degree[1]))[, 1 + z.noi[, 1],drop=FALSE]
    }
    if(k > 1) for (i in 2:k) if(degree[i] > 0) {
      res <- res * cbind(1, mypoly(x=xdat.numeric[,i],
                                   ex=exdat.numeric[,i],
                                   degree=degree[i],
                                   gradient.compute=gradient.compute,
                                   r=gradient.vec[i],
                                   Bernstein=Bernstein))[, 1 + z.noi[, i]]
      if(gradient.compute && gradient.vec[i] != 0) res.deriv <- res.deriv * cbind(1,matrix(NA,1,degree[i]))[, 1 + z.noi[, i],drop=FALSE]
      if(gradient.compute && gradient.vec[i] == 0) res.deriv <- res.deriv *cbind(1,matrix(0,1,degree[i]))[, 1 + z.noi[, i],drop=FALSE]
    }

    if(is.null(exdat)) {
      res <- matrix(res,nrow=NROW(xdat))
    } else {
      res <- matrix(res,nrow=NROW(exdat))
    }
    if(gradient.compute) res.deriv <- matrix(res.deriv,nrow=1)
    colnames(res) <- apply(z.noi, 1L, function(x) paste(x, collapse = "."))
    if(gradient.compute) colnames(res.deriv) <- apply(z.noi, 1L, function(x) paste(x, collapse = "."))

    if(gradient.compute) {
      res[,!is.na(as.numeric(res.deriv))] <- 0
      return(cbind(0,res))
    } else {
      return(cbind(1,res))
    }

  }

}

### internal constants used in the c backend

SF_NORMAL = 0
SF_ARB = 1

BW_FIXED = 0
BW_GEN_NN = 1
BW_ADAP_NN = 2

IMULTI_TRUE = 1
IMULTI_FALSE = 0

RE_MIN_TRUE = 0
RE_MIN_FALSE = 1

IO_MIN_TRUE = 1
IO_MIN_FALSE = 0

USE_START_NO = 0
USE_START_YES = 1

NP_DO_DENS = 1
NP_DO_DIST = 0

## initially making an np-wide option via the 'options' mechanism
DO_TREE_NO = 0
DO_TREE_YES = 1

##kernel defs
CKER_GAUSS = 0
CKER_EPAN  = 4
CKER_UNI   = 8
CKER_TGAUSS = 9

UKER_AIT = 0
UKER_LR = 1

OKER_WANG = 0
OKER_LR = 1
OKER_NLR = 2
OKER_RLY = 3

##density 
BWM_CVML = 0
BWM_CVLS = 1
BWM_CVML_NP= 2

##distribution
DBWM_CVLS = 0

##regression
BWM_CVAIC = 0

REGTYPE_LC = 0
REGTYPE_LL = 1
REGTYPE_LP = 2
# legacy alias retained for internal/backward compatibility
REGTYPE_GLP = REGTYPE_LP

##conditional density/distribution
CBWM_CVML = 0
CBWM_CVLS = 1
CBWM_NPLS = 2
CBWM_CCDF = 3 # Added 7/2/2010 jracine

##conditional distribution
CDBWM_CVLS = 0

##integral operators on kernels
OP_NOOP        = -1
OP_NORMAL      = 0
OP_CONVOLUTION = 1
OP_DERIVATIVE  = 2
OP_INTEGRAL    = 3

ALL_OPERATORS = c(OP_NORMAL, OP_CONVOLUTION, OP_DERIVATIVE, OP_INTEGRAL)
names(ALL_OPERATORS) <- c("normal","convolution", "derivative", "integral")

PERMUTATION_OPERATORS <- c(OP_NOOP, OP_NORMAL, OP_DERIVATIVE, OP_INTEGRAL)
names(PERMUTATION_OPERATORS) <- c("none", "normal", "derivative", "integral")

## useful numerical constants of kernel integrals
int.kernels <- c(0.28209479177387814348, 0.47603496111841936711, 0.62396943688265038571, 0.74785078617543927990,
                 0.26832815729997476357, 0.55901699437494742410, 0.84658823667359826246, 1.1329342579014329689,
                 0.5, 2.90113075268188e-01)

QFAC <- qnorm(.25,lower.tail=FALSE)*2

.np_eval_bws_call_arg <- function(bws, arg) {
  if (is.null(bws$call))
    stop("bandwidth object does not contain a call component")

  expr <- bws$call[[arg]]
  if (is.null(expr))
    stop(sprintf("bandwidth call does not contain '%s'", arg))

  eval.env <- environment(bws$call)
  if (is.null(eval.env))
    eval.env <- parent.frame()

  if (!is.language(expr))
    return(expr)

  val <- .np_try_eval_in_frames(expr, eval_env = eval.env, search_frames = FALSE)
  if (isTRUE(val$ok))
    return(val$value)

  fallback <- bws[[arg]]
  if (!is.null(fallback))
    return(fallback)

  if (inherits(val$error, "error"))
    stop(conditionMessage(val$error), call. = FALSE)
  stop(sprintf("unable to evaluate '%s' in bandwidth call", arg), call. = FALSE)
}

.np_eval_call_arg <- function(call_obj, arg, caller_env = parent.frame()) {
  if (is.null(call_obj))
    stop("object does not contain a call component")

  expr <- call_obj[[arg]]
  if (is.null(expr))
    stop(sprintf("call does not contain '%s'", arg))

  eval.env <- environment(call_obj)
  if (is.null(eval.env))
    eval.env <- caller_env

  if (!is.language(expr))
    return(expr)

  if (is.symbol(expr)) {
    val <- .np_try_eval_in_frames(expr, eval_env = eval.env, search_frames = FALSE)
    if (isTRUE(val$ok))
      return(val$value)

    if (is.environment(caller_env)) {
      not_found <- .np_missing_binding_sentinel
      sym_val <- get0(as.character(expr), envir = caller_env, inherits = TRUE, ifnotfound = not_found)
      if (!identical(sym_val, not_found))
        return(sym_val)
    }
  }

  val <- .np_try_eval_in_frames(expr, eval_env = eval.env)
  if (isTRUE(val$ok))
    return(val$value)

  if (inherits(val$error, "error"))
    stop(conditionMessage(val$error), call. = FALSE)
  stop(sprintf("unable to evaluate call argument '%s'", arg), call. = FALSE)
}

library(np)

test_that("old continuous scale-factor search names fail with rename diagnostics", {
  dat <- data.frame(x = rnorm(12), y = rnorm(12))

  expect_error(
    npregbw(y ~ x, data = dat, cfac.init = 0.5),
    "'cfac.init' has been renamed to 'scale.factor.init'",
    fixed = TRUE
  )
  expect_error(
    npudensbw(~ x, data = dat, lbc.init = 0.1),
    "'lbc.init' has been renamed to 'scale.factor.init.lower'",
    fixed = TRUE
  )
  expect_error(
    npcdensbw(y ~ x, data = dat, hbc.init = 2),
    "'hbc.init' has been renamed to 'scale.factor.init.upper'",
    fixed = TRUE
  )
  expect_error(
    npindexbw(y ~ x + I(x^2), data = dat, scale.factor.lower.bound = 0.2),
    "'scale.factor.lower.bound' has been renamed to 'scale.factor.search.lower'",
    fixed = TRUE
  )
})

test_that("scale-factor search lower metadata uses canonical field with legacy read fallback", {
  canonical <- list(scale.factor.search.lower = 0.25)
  legacy <- list(scale.factor.lower.bound = 0.25)
  conflict <- list(scale.factor.search.lower = 0.25, scale.factor.lower.bound = 0.5)

  expect_equal(np:::npGetScaleFactorSearchLower(canonical), 0.25, tolerance = 0)
  expect_equal(np:::npGetScaleFactorSearchLower(legacy), 0.25, tolerance = 0)
  expect_error(
    np:::npGetScaleFactorSearchLower(conflict),
    "conflicts with legacy 'scale.factor.lower.bound'",
    fixed = TRUE
  )

  object <- np:::npSetScaleFactorSearchLower(list(scale.factor.lower.bound = 0.1), 0.3)
  expect_equal(object$scale.factor.search.lower, 0.3, tolerance = 0)
  expect_null(object$scale.factor.lower.bound)
})

test_that("protected non-continuous initialization names remain separate", {
  npscoefbw_default <- getFromNamespace("npscoefbw.default", "np")
  npudensbw_default <- getFromNamespace("npudensbw.default", "np")

  expect_true(all(c("dfac.init", "lbd.init", "hbd.init") %in%
                    names(formals(npscoefbw_default))))
  expect_true(all(c("dfac.init", "lbd.init", "hbd.init",
                    "initc.dir", "initd.dir",
                    "scale.init.categorical.sample") %in%
                    names(formals(npudensbw_default))))
})

test_that("gradient accessors fail fast when gradients are unavailable", {
  npreg.obj <- structure(
    list(
      grad = NA,
      gerr = NA,
      bws = list(regtype = "ll", ncon = 1L, icon = TRUE, degree = 1L)
    ),
    class = "npregression"
  )
  expect_error(gradients(npreg.obj), "gradients are not available")
  expect_error(gradients(npreg.obj, errors = TRUE), "gradient standard errors are not available")

  si.obj <- structure(list(grad = NA, gerr = NA), class = "singleindex")
  expect_error(gradients(si.obj), "gradients are not available")
  expect_error(gradients(si.obj, errors = TRUE), "gradient standard errors are not available")

  cd.obj <- structure(list(congrad = NA, congerr = NA), class = "condensity")
  expect_error(gradients(cd.obj), "gradients are not available")
  expect_error(gradients(cd.obj, errors = TRUE), "gradient standard errors are not available")

  cdf.obj <- structure(list(congrad = NA, congerr = NA), class = "condistribution")
  expect_error(gradients(cdf.obj), "gradients are not available")
  expect_error(gradients(cdf.obj, errors = TRUE), "gradient standard errors are not available")
})

dim_basis <- getFromNamespace("dim_basis", "np")
dimBS <- getFromNamespace("dimBS", "np")

test_that("GLP validators handle non-numeric and missing inputs deterministically", {
  expect_null(np:::npValidateGlpDegree("lc", "a", 1L))
  expect_identical(np:::npValidateGlpDegree("lp", NULL, 0L), integer(0))

  expect_error(
    np:::npValidateGlpDegree("lp", "a", 1L),
    "finite non-negative integers"
  )
  expect_error(
    np:::npValidateGlpDegree("lp", c(1, NA_real_), 2L),
    "finite non-negative integers"
  )
  expect_warning(
    expect_identical(np:::npValidateGlpDegree("lp", 30L, 1L), 30L),
    "unusually large polynomial degree values above 25"
  )
  expect_error(
    np:::npValidateGlpDegree("lp", 101L, 1L),
    "\\[0,100\\]"
  )

  expect_null(np:::npValidateGlpGradientOrder("lc", "a", 1L))
  expect_identical(np:::npValidateGlpGradientOrder("lp", NULL, 0L), integer(0))

  expect_error(
    np:::npValidateGlpGradientOrder("lp", "a", 1L),
    "finite positive integers"
  )
  expect_error(
    np:::npValidateGlpGradientOrder("lp", c(1, NA_real_), 2L),
    "finite positive integers"
  )
})

test_that("dim_basis validates integer-like vectors before coercion", {
  expect_error(dim_basis(degree = "a"), "degree must contain finite non-negative integers")
  expect_error(dim_basis(degree = 1L, segments = "a"), "segments must contain finite positive integers")
  expect_error(
    dim_basis(degree = integer(0), include = c("a"), categories = c(0L)),
    "include must contain finite non-negative integers"
  )
  expect_error(
    dim_basis(degree = integer(0), include = c(0L), categories = c("a")),
    "categories must contain finite non-negative integers"
  )
})

test_that("dimBS compatibility wrapper matches dim_basis", {
  expect_identical(
    dimBS(basis = "tensor", degree = c(2L, 3L), segments = c(1L, 2L)),
    dim_basis(basis = "tensor", degree = c(2L, 3L), segments = c(1L, 2L))
  )
  expect_identical(
    dimBS(basis = "glp",
          kernel = FALSE,
          degree = c(2L, 1L),
          segments = c(1L, 1L),
          include = c(1L, 0L, 1L),
          categories = c(3L, 2L, 4L)),
    dim_basis(basis = "glp",
              kernel = FALSE,
              degree = c(2L, 1L),
              segments = c(1L, 1L),
              include = c(1L, 0L, 1L),
              categories = c(3L, 2L, 4L))
  )
})

test_that("npksum basic functionality works", {
  data("cps71")
  cps71_sub <- cps71[1:50,]
  
  k <- npksum(txdat=cps71_sub$age, tydat=cps71_sub$logwage, bws=1.0)
  
  expect_s3_class(k, "npkernelsum")
  expect_type(k$ksum, "double")
})

test_that("npseed works", {
  # npseed sets seed for C backend, doesn't affect R's runif
  expect_silent(npseed(42))
})

test_that("b.star works", {
  data("Italy")
  # b.star is for block size selection in bootstrap
  b <- b.star(Italy$gdp)
  expect_type(b, "double")
  expect_equal(nrow(b), 1)
})

test_that("b.star round=TRUE applies elementwise bounds for multivariate input", {
  set.seed(1)
  x <- cbind(arima.sim(n = 200, list(ar = 0.5)),
             arima.sim(n = 200, list(ar = 0.8)))

  b.raw <- b.star(x, round = FALSE)
  b.round <- b.star(x, round = TRUE)
  b.max <- ceiling(min(3 * sqrt(nrow(x)), nrow(x) / 3))

  expected <- cbind(
    pmax(1, pmin(b.max, round(b.raw[, "BstarSB"]))),
    pmax(1, pmin(b.max, round(b.raw[, "BstarCB"])))
  )
  colnames(expected) <- colnames(b.raw)

  expect_equal(b.round, expected)
})

library(np)

test_that("omitted generic scale-factor floor defaults to 0.1", {
  expect_equal(np:::npResolveScaleFactorLowerBound(NULL), 0.1, tolerance = 0)
  expect_equal(np:::npResolveScaleFactorLowerBound(0.01), 0.01, tolerance = 0)
})

test_that("scale.factor.search.lower is accepted across bandwidth selectors", {
  set.seed(20260423)
  n <- 32L
  x <- runif(n)
  z <- runif(n)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.25)

  expect_floor <- function(object, expected = 0.2) {
    expect_equal(object$scale.factor.search.lower, expected, tolerance = 0)
    invisible(object)
  }

  common <- list(
    bwscaling = TRUE,
    bwtype = "fixed",
    nmulti = 1L,
    itmax = 10L,
    scale.factor.search.lower = 0.2
  )

  expect_floor(do.call(npregbw, c(
    list(xdat = data.frame(x = x), ydat = y, bwmethod = "cv.ls"),
    common
  )))

  expect_floor(do.call(npudensbw, c(
    list(dat = data.frame(y = y), bwmethod = "cv.ls"),
    common
  )))

  expect_floor(do.call(npudistbw, c(
    list(dat = data.frame(y = y), bwmethod = "cv.cdf"),
    common
  )))

  expect_floor(do.call(npcdensbw, c(
    list(xdat = data.frame(x = x), ydat = y, bwmethod = "cv.ls"),
    common
  )))

  expect_floor(do.call(npcdistbw, c(
    list(xdat = data.frame(x = x), ydat = y, bwmethod = "cv.ls"),
    common
  )))
})

test_that("scale.factor.search.lower propagates through semiparametric selectors", {
  set.seed(20260424)
  n <- 32L
  x <- runif(n)
  z <- runif(n)
  y <- 1 + x + sin(2 * pi * z) + rnorm(n, sd = 0.25)

  pl <- npplregbw(
    xdat = data.frame(x = x),
    ydat = y,
    zdat = data.frame(z = z),
    bwmethod = "cv.ls",
    bwtype = "fixed",
    nmulti = 1L,
    itmax = 10L,
    scale.factor.search.lower = 0.2
  )
  expect_equal(
    unname(vapply(pl$bw, function(bwi) bwi$scale.factor.search.lower, numeric(1L))),
    rep(0.2, length(pl$bw)),
    tolerance = 0
  )

  si <- npindexbw(
    xdat = data.frame(x = x, z = z),
    ydat = y,
    bws = c(0, 0, 0),
    bwmethod = "ichimura",
    bwtype = "fixed",
    nmulti = 1L,
    optim.maxit = 10L,
    scale.factor.search.lower = 0.2
  )
  expect_equal(si$scale.factor.search.lower, 0.2, tolerance = 0)

  sc <- npscoefbw(
    xdat = data.frame(x = x),
    ydat = y,
    zdat = data.frame(z = z),
    bwmethod = "cv.ls",
    bwtype = "fixed",
    nmulti = 1L,
    optim.maxit = 10L,
    scale.factor.search.lower = 0.2
  )
  expect_equal(sc$scale.factor.search.lower, 0.2, tolerance = 0)
})

shadow_conditional_xweights_lp <- function(bw, xdat, ydat, row_index) {
  make_mat <- function(z) {
    z <- as.matrix(z)
    storage.mode(z) <- "double"
    z
  }

  xmat <- np:::toMatrix(xdat)
  ymat <- np:::toMatrix(np:::toFrame(ydat))
  ycon <- make_mat(ymat[, bw$iycon, drop = FALSE])
  yuno <- make_mat(ymat[, bw$iyuno, drop = FALSE])
  yord <- make_mat(ymat[, bw$iyord, drop = FALSE])
  xcon <- make_mat(xmat[, bw$ixcon, drop = FALSE])
  xuno <- make_mat(xmat[, bw$ixuno, drop = FALSE])
  xord <- make_mat(xmat[, bw$ixord, drop = FALSE])

  bw_vec <- c(
    bw$xbw[bw$ixcon],
    bw$ybw[bw$iycon],
    bw$ybw[bw$iyuno],
    bw$ybw[bw$iyord],
    bw$xbw[bw$ixuno],
    bw$xbw[bw$ixord]
  )

  .Call(
    "C_np_shadow_cv_xweights_conditional",
    yuno,
    yord,
    ycon,
    xuno,
    xord,
    xcon,
    as.double(bw_vec),
    as.integer(np:::BW_FIXED),
    as.integer(switch(
      bw$cxkertype,
      gaussian = np:::CKER_GAUSS + bw$cxkerorder / 2 - 1,
      epanechnikov = np:::CKER_EPAN + bw$cxkerorder / 2 - 1,
      uniform = np:::CKER_UNI,
      "truncated gaussian" = np:::CKER_TGAUSS
    )),
    as.integer(switch(
      bw$uxkertype,
      aitchisonaitken = np:::UKER_AIT,
      liracine = np:::UKER_LR
    )),
    as.integer(switch(
      bw$oxkertype,
      wangvanryzin = np:::OKER_WANG,
      liracine = np:::OKER_LR,
      racineliyan = np:::OKER_RLY
    )),
    as.integer(FALSE),
    as.integer(np:::REGTYPE_LP),
    as.integer(bw$degree.engine),
    as.integer(isTRUE(bw$bernstein.basis.engine)),
    as.integer(np:::npLpBasisCode(bw$basis.engine)),
    as.integer(row_index),
    PACKAGE = "np"
  )$streamed
}

bounded_gaussian_kernel <- function(x0, X, h, lower, upper) {
  denom <- h * (pnorm((upper - x0) / h) - pnorm((lower - x0) / h))
  dnorm((x0 - X) / h) / denom
}

smooth_cv_loglik <- function(fit_values, cutoff = .Machine$double.xmin) {
  out <- numeric(length(fit_values))
  log_cutoff <- log(cutoff)

  pos <- fit_values > cutoff
  neg <- fit_values < -cutoff
  mid <- !(pos | neg)

  out[pos] <- log(fit_values[pos])
  out[neg] <- -log(abs(fit_values[neg])) + 2 * log_cutoff
  out[mid] <- log_cutoff
  out
}

test_that("npcdensbw cv.ml LP degree-0 bounded objective matches delete-one reconstruction", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(1)
  n <- 80L
  xdat <- data.frame(x = runif(n))
  ydat <- rbeta(n, 1, 1)

  bw <- np::npcdensbw(
    xdat = xdat,
    ydat = ydat,
    bwmethod = "cv.ml",
    regtype = "lp",
    degree = 0L,
    bws = c(0.15, 0.12),
    bandwidth.compute = FALSE,
    cxkerbound = "range",
    cykerbound = "range"
  )

  manual_rows <- vapply(seq_len(n), function(i) {
    weights <- shadow_conditional_xweights_lp(bw, xdat, ydat, i)
    ky <- bounded_gaussian_kernel(
      ydat[i],
      ydat,
      bw$ybw[1L],
      bw$cykerlb[bw$iycon][1L],
      bw$cykerub[bw$iycon][1L]
    )
    sum(weights * ky)
  }, numeric(1))

  expect_true(all(manual_rows > .Machine$double.xmin))

  manual_objective <- sum(log(manual_rows))
  np_objective <- np:::.npcdensbw_eval_only(xdat, ydat, bw)$objective

  expect_equal(np_objective, manual_objective, tolerance = 1e-5)
})

test_that("npcdensbw cv.ml LP objective uses smooth penalty for negative delete-one fits", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(1)
  n <- 80L
  xdat <- data.frame(x = runif(n))
  ydat <- rbeta(n, 1, 1)

  bw <- np::npcdensbw(
    xdat = xdat,
    ydat = ydat,
    bwmethod = "cv.ml",
    regtype = "lp",
    degree = 3L,
    bws = c(0.15, 0.12),
    bandwidth.compute = FALSE,
    cxkerbound = "range",
    cykerbound = "range"
  )

  manual_rows <- vapply(seq_len(n), function(i) {
    weights <- shadow_conditional_xweights_lp(bw, xdat, ydat, i)
    ky <- bounded_gaussian_kernel(
      ydat[i],
      ydat,
      bw$ybw[1L],
      bw$cykerlb[bw$iycon][1L],
      bw$cykerub[bw$iycon][1L]
    )
    sum(weights * ky)
  }, numeric(1))

  expect_gt(sum(manual_rows < 0), 0L)

  smooth_objective <- sum(smooth_cv_loglik(manual_rows))
  constant_terms <- rep.int(log(.Machine$double.xmin), length(manual_rows))
  pos <- manual_rows > .Machine$double.xmin
  constant_terms[pos] <- log(manual_rows[pos])
  constant_objective <- sum(constant_terms)
  np_objective <- np:::.npcdensbw_eval_only(xdat, ydat, bw)$objective

  expect_lt(abs(np_objective - smooth_objective), 20)
  expect_gt(abs(np_objective - constant_objective), 1000)
})

suppressPackageStartupMessages(library(np))

test_that("Powell hot-start respects fixed continuous lower bound", {
  skip_if_not_installed("crs")

  floor_value <- function(y, n, p = 2L, q = 1L) {
    0.1 * np:::EssDee(y) * n^(-1 / (2 * p + q))
  }

  set.seed(600007)
  n <- 400
  x <- runif(n)
  y <- rchisq(n, df = 2 + 4 * (x - 0.5)^2)
  floor_y <- floor_value(y, n)

  nom <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bwmethod = "cv.ls",
    regtype = "lp",
    bwtype = "fixed",
    search.engine = "nomad",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 10L,
    bernstein.basis = TRUE,
    nmulti = 2,
    nomad = TRUE,
    cxkerbound = "fixed", cxkerlb = 0, cxkerub = 1,
    cykerbound = "fixed", cykerlb = 0, cykerub = Inf
  )

  expect_gte(nom$ybw[1], floor_y)

  nom$degree <- 0L
  nom$degree.engine <- 0L
  nom$regtype <- "lp"
  nom$pregtype <- "Local-Polynomial"

  hot <- np:::npcdensbw.conbandwidth(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = nom,
    bandwidth.compute = TRUE,
    nmulti = 1
  )

  expect_gte(hot$ybw[1], floor_y)
})

test_that("non-search explicit bandwidths are not clamped", {
  set.seed(77)
  n <- 80
  x <- runif(n)
  y <- rexp(n)

  base <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bwmethod = "cv.ls",
    regtype = "lp",
    degree = 0L,
    bwtype = "fixed",
    bandwidth.compute = FALSE,
    cxkerbound = "fixed", cxkerlb = 0, cxkerub = 1,
    cykerbound = "fixed", cykerlb = 0, cykerub = Inf
  )

  base$xbw[1] <- 0.12345
  base$ybw[1] <- 1e-06

  out <- np:::npcdensbw.conbandwidth(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = base,
    bandwidth.compute = FALSE
  )

  expect_equal(out$xbw[1], 0.12345, tolerance = 0)
  expect_equal(out$ybw[1], 1e-06, tolerance = 0)
})

test_that("public eval-only bounded cv.ls objective still accepts tiny fixed bandwidths", {
  set.seed(42)
  n <- 350L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rbeta(n, 1, 1))

  bw.bad <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.0610566549, 0.0009802178),
    bwtype = "fixed",
    bwmethod = "cv.ls",
    regtype = "lp",
    degree = 3,
    basis = "glp",
    bandwidth.compute = FALSE,
    cxkerbound = "range",
    cykerbound = "range"
  )

  obj.bad <- np:::.npcdensbw_eval_only(x, y, bw.bad)$objective
  expect_true(is.finite(obj.bad))
})

library(np)

make_regression_summary_label_fixture <- function(type) {
  untangle <- getFromNamespace("untangle", "np")
  rbandwidth <- getFromNamespace("rbandwidth", "np")
  npregression <- getFromNamespace("npregression", "np")

  xdat <- data.frame(x = c(0.1, 0.4, 0.9))
  ydat <- data.frame(y = c(1, 2, 3))
  xdati <- untangle(xdat)
  ydati <- untangle(ydat)

  bws <- rbandwidth(
    bw = 2,
    regtype = "lc",
    bwmethod = "cv.ls",
    bwscaling = FALSE,
    bwtype = type,
    ckertype = "gaussian",
    ckerorder = 2,
    ukertype = "aitchisonaitken",
    okertype = "liracine",
    nobs = nrow(xdat),
    xdati = xdati,
    ydati = ydati,
    xnames = names(xdat),
    ynames = names(ydat),
    sfactor = 1,
    bandwidth = if (identical(type, "fixed")) 0.25 else 2,
    nconfac = 1L,
    ncatfac = 0L,
    sdev = stats::sd(xdat$x),
    bandwidth.compute = FALSE
  )

  npregression(
    bws = bws,
    eval = as.matrix(xdat),
    mean = ydat$y,
    ntrain = nrow(xdat),
    trainiseval = TRUE
  )
}

test_that("fixed mixed-data bandwidth summaries label continuous scale factors correctly", {
  set.seed(123)
  n <- 40L
  xdat <- data.frame(
    x = runif(n),
    z1 = runif(n),
    z2 = factor(sample(c("a", "b"), n, replace = TRUE))
  )
  ydat <- xdat$x + xdat$z1 + as.numeric(xdat$z2 == "b") + rnorm(n)

  bw <- npregbw(
    xdat = xdat,
    ydat = ydat,
    regtype = "lc",
    bwmethod = "cv.aic",
    nmulti = 1
  )

  s <- paste(capture.output(summary(bw)), collapse = "\n")

  expect_match(s, "Exp\\. Var\\. Name: x\\s+Bandwidth: .*Scale Factor:")
  expect_match(s, "Exp\\. Var\\. Name: z1\\s+Bandwidth: .*Scale Factor:")
  expect_match(s, "Exp\\. Var\\. Name: z2\\s+Bandwidth: .*Lambda Max:")
})

test_that("regression summaries distinguish nearest-neighbor selectors from fixed bandwidths", {
  fixed <- paste(capture.output(summary(make_regression_summary_label_fixture("fixed"))), collapse = "\n")
  adaptive <- paste(capture.output(summary(make_regression_summary_label_fixture("adaptive_nn"))), collapse = "\n")
  generalized <- paste(capture.output(summary(make_regression_summary_label_fixture("generalized_nn"))), collapse = "\n")

  expect_match(fixed, "Bandwidth\\(s\\):\\s+2")
  expect_match(adaptive, "Bandwidth Nearest Neighbor\\(s\\):\\s+2")
  expect_match(generalized, "Bandwidth Nearest Neighbor\\(s\\):\\s+2")
})

test_that("npseed accepts valid integer-representable scalars", {
  expect_silent(npseed(42))
  expect_silent(npseed(-42))
  expect_silent(npseed(42.0))
})

test_that("npseed rejects invalid scalar inputs before reaching C", {
  expect_error(npseed(NA_integer_), "'seed' must be a single finite numeric value")
  expect_error(npseed(Inf), "'seed' must be a single finite numeric value")
  expect_error(npseed(NaN), "'seed' must be a single finite numeric value")
  expect_error(npseed(numeric()), "'seed' must be a single finite numeric value")
  expect_error(npseed(c(1, 2)), "'seed' must be a single finite numeric value")
  expect_error(npseed(1.5), "'seed' must be representable as a non-negative integer after abs\\(\\)")
  expect_error(npseed(.Machine$integer.max + 1), "'seed' must be representable as a non-negative integer after abs\\(\\)")
  expect_error(npseed("1"), "'seed' must be a single finite numeric value")
})

test_that("C_np_set_seed rejects direct misuse", {
  expect_error(.Call("C_np_set_seed", NA_integer_, PACKAGE = "np"), "seed must be finite")
  expect_error(.Call("C_np_set_seed", Inf, PACKAGE = "np"), "seed must be finite")
  expect_error(.Call("C_np_set_seed", 1.5, PACKAGE = "np"), "seed must be representable as a non-negative integer after abs\\(\\)")
  expect_error(.Call("C_np_set_seed", c(1L, 2L), PACKAGE = "np"), "seed must have length 1")
  expect_error(.Call("C_np_set_seed", TRUE, PACKAGE = "np"), "seed must be numeric")
})

test_that("uocquantile basic functionality works", {
  x_num <- c(1, 2, 3, 4, 5)
  expect_equal(unname(uocquantile(x_num, 0.5)), 3)
  
  x_fact <- factor(c("A", "A", "B", "C"))
  expect_equal(as.character(uocquantile(x_fact, 0.5)), "A")
  
  x_ord <- ordered(c("Low", "Medium", "High"), levels=c("Low", "Medium", "High"))
  expect_equal(as.character(uocquantile(x_ord, 0.5)), "Medium")
})

test_that("nptgauss basic functionality works", {
  # nptgauss is called for its side effect (setting C variables) 
  # or returns NULL/constants.
  expect_silent(nptgauss(3.0))
})

test_that("plot basic functionality works", {
  data("faithful")
  bw <- npudensbw(dat=faithful, bws=c(0.5, 5), bandwidth.compute=FALSE)
  
  # Use pdf(NULL) to avoid opening a window
  pdf(NULL)
  on.exit(dev.off())
  
  expect_silent(plot(bw))
})

test_that("se and gradients methods work", {
  data("cps71")
  cps71_sub <- cps71[1:50, ]
  bw <- npregbw(logwage~age, data=cps71_sub, bws=1.0, bandwidth.compute=FALSE)
  model <- npreg(bws=bw, gradients=TRUE)
  
  expect_type(se(model), "double")
  expect_length(se(model), 50)
  
  expect_type(gradients(model), "double")
  expect_length(gradients(model), 50)
})

test_that("internal kbandwidth S3 methods are registered but not exported", {
  ns <- asNamespace(getNamespaceName(environment(npksum)))
  s3 <- getNamespaceInfo(ns, "S3methods")

  expect_false("kbandwidth" %in% getNamespaceExports(ns))
  expect_true(any(s3[, 1L] == "kbandwidth" & s3[, 2L] == "numeric"))
  expect_true(any(s3[, 1L] == "kbandwidth" & s3[, 2L] == "integer"))
  expect_true(any(s3[, 1L] == "kbandwidth" & s3[, 2L] == "default"))
  expect_true(any(s3[, 1L] == "print" & s3[, 2L] == "kbandwidth"))
  expect_true(any(s3[, 1L] == "as.double" & s3[, 2L] == "kbandwidth"))

  expect_true(!is.null(getS3method("print", "kbandwidth", optional = TRUE)))
  expect_true(!is.null(getS3method("as.double", "kbandwidth", optional = TRUE)))
})

test_that("internal kbandwidth objects print and coerce consistently", {
  kbandwidth <- getFromNamespace("kbandwidth", getNamespaceName(environment(npksum)))
  untangle <- getFromNamespace("untangle", getNamespaceName(environment(npksum)))

  x <- data.frame(x = seq(0, 1, length.out = 5))
  kbw <- kbandwidth(
    bw = 0.5,
    xdati = untangle(x),
    xnames = names(x),
    nobs = nrow(x)
  )

  expect_s3_class(kbw, "kbandwidth")
  expect_equal(as.double(kbw), 0.5)
  out <- capture.output(print(kbw))
  expect_true(any(grepl("Kernel Sum Bandwidth", out, fixed = TRUE)))
  expect_true(any(grepl("Bandwidth Type: Fixed", out, fixed = TRUE)))
})

test_that("npreg lp higher-order gradients match npreghat across lp bases", {
  set.seed(99)
  n <- 140
  x <- sort(runif(n))
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.02)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 35))

  for (basis in c("glp", "additive", "tensor")) {
    bw <- npregbw(
      xdat = tx,
      ydat = y,
      regtype = "lp",
      degree = 2L,
      basis = basis,
      bws = 0.25,
      bandwidth.compute = FALSE
    )

    fit <- suppressWarnings(npreg(
      txdat = tx,
      tydat = y,
      exdat = ex,
      bws = bw,
      gradients = TRUE,
      gradient.order = 2L
    ))

    H2 <- npreghat(bws = bw, txdat = tx, exdat = ex, s = 2L)
    g2 <- as.vector(H2 %*% y)

    expect_false(all(is.na(fit$grad[, 1])), info = basis)
    expect_equal(as.vector(fit$grad[, 1]), g2, tolerance = 1e-6, info = basis)
  }
})

test_that("npreg lp supports per-variable derivative orders", {
  set.seed(123)
  n <- 180
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(2 * pi * x1) + x2^3 + rnorm(n, sd = 0.03)
  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- data.frame(x1 = seq(0.05, 0.95, length.out = 20),
                   x2 = seq(0.1, 0.9, length.out = 20))

  bw <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "lp",
    degree = c(2L, 3L),
    basis = "glp",
    bws = c(0.2, 0.2),
    bandwidth.compute = FALSE
  )

  gorder <- c(2L, 1L)
  fit <- suppressWarnings(npreg(
    txdat = tx,
    tydat = y,
    exdat = ex,
    bws = bw,
    gradients = TRUE,
    gradient.order = gorder
  ))

  Hx1 <- npreghat(bws = bw, txdat = tx, exdat = ex, s = c(2L, 0L))
  Hx2 <- npreghat(bws = bw, txdat = tx, exdat = ex, s = c(0L, 1L))
  expect_equal(as.vector(fit$grad[, 1]), as.vector(Hx1 %*% y), tolerance = 1e-6)
  expect_equal(as.vector(fit$grad[, 2]), as.vector(Hx2 %*% y), tolerance = 1e-6)
})

test_that("npreg lp gradients accessor honors stored derivative order", {
  set.seed(20260514)
  n <- 55
  x <- seq(-1, 1, length.out = n)
  y <- 1 + 2 * x + 3 * x^2
  tx <- data.frame(x = x)

  bw2 <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "lp",
    degree = 2L,
    basis = "glp",
    bws = 10,
    bandwidth.compute = FALSE
  )

  fit1 <- npreg(
    txdat = tx,
    tydat = y,
    bws = bw2,
    gradients = TRUE,
    gradient.order = 1L,
    warn.glp.gradient = FALSE
  )

  expect_error(
    gradients(fit1, gradient.order = 2L),
    "differs from the derivative order stored"
  )
  expect_error(
    gradients(fit1, errors = TRUE, gradient.order = 2L),
    "differs from the derivative order stored"
  )

  fit2 <- npreg(
    txdat = tx,
    tydat = y,
    bws = bw2,
    gradients = TRUE,
    gradient.order = 2L,
    warn.glp.gradient = FALSE
  )

  expect_equal(gradients(fit2, gradient.order = 2L), fit2$grad, tolerance = 0)
  expect_equal(gradients(fit2, errors = TRUE, gradient.order = 2L),
               fit2$gerr, tolerance = 0)

  bw1 <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "lp",
    degree = 1L,
    basis = "glp",
    bws = 10,
    bandwidth.compute = FALSE
  )
  fit.degree1 <- npreg(
    txdat = tx,
    tydat = y,
    bws = bw1,
    gradients = TRUE,
    gradient.order = 1L,
    warn.glp.gradient = FALSE
  )
  expect_warning(
    undefined <- gradients(fit.degree1, gradient.order = 2L),
    "exceed polynomial degree"
  )
  expect_true(all(is.na(undefined[, 1L])))
})

test_that("npreg lp Bernstein derivatives are returned on original scale", {
  n <- 50
  x <- seq(-1, 1, length.out = n)
  y <- 1 + 2 * x + 3 * x^2
  tx <- data.frame(x = x)

  bw <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "lp",
    degree = 2L,
    basis = "tensor",
    bernstein.basis = TRUE,
    bws = 100,
    bandwidth.compute = FALSE
  )

  fit <- npreg(
    txdat = tx,
    tydat = y,
    bws = bw,
    gradients = TRUE,
    gradient.order = 2L,
    warn.glp.gradient = FALSE
  )

  expect_equal(as.vector(fitted(fit)), y, tolerance = 1e-8)
  expect_equal(as.vector(gradients(fit, gradient.order = 2L)[, 1L]),
               rep(6, n), tolerance = 1e-8)
  expect_true(all(is.finite(gradients(fit, errors = TRUE,
                                      gradient.order = 2L)[, 1L])))
})

test_that("core fitted and se extract stored quantities without recomputation", {
  dens <- structure(list(dens = c(1, 2), derr = c(0.1, 0.2)), class = "npdensity")
  expect_equal(fitted(dens), dens$dens)
  expect_equal(se(dens), dens$derr)

  dist <- structure(list(dist = c(0.2, 0.8), derr = c(0.03, 0.04)), class = "npdistribution")
  expect_equal(fitted(dist), dist$dist)
  expect_equal(se(dist), dist$derr)

  reg <- structure(
    list(mean = c(2, 3), merr = c(0.2, 0.3), residuals = TRUE, resid = c(-0.1, 0.1)),
    class = "npregression"
  )
  expect_equal(fitted(reg), reg$mean)
  expect_equal(se(reg), reg$merr)
  expect_equal(residuals(reg), reg$resid)

  qreg <- structure(
    list(quantile = matrix(1:4, nrow = 2), quanterr = matrix(0.1, nrow = 2, ncol = 2)),
    class = "qregression"
  )
  expect_equal(fitted(qreg), qreg$quantile)
  expect_equal(quantile(qreg), qreg$quantile)
  expect_equal(se(qreg), qreg$quanterr)
})

test_that("conditional and semiparametric extractors preserve stored shapes", {
  cdens <- structure(
    list(condens = c(1, 2), conderr = c(0.1, 0.2), proper.applied = FALSE),
    class = "condensity"
  )
  expect_equal(fitted(cdens), cdens$condens)
  expect_equal(se(cdens), cdens$conderr)

  cdist <- structure(
    list(condist = c(0.1, 0.9), conderr = c(0.02, 0.03), proper.applied = FALSE),
    class = "condistribution"
  )
  expect_equal(fitted(cdist), cdist$condist)
  expect_equal(se(cdist), cdist$conderr)

  pl <- structure(
    list(mean = c(1, 3), merr = c(0.3, 0.4), residuals = TRUE, resid = c(0.1, -0.1)),
    class = "plregression"
  )
  expect_equal(fitted(pl), pl$mean)
  expect_equal(residuals(pl), pl$resid)

  si <- structure(
    list(
      mean = c(1, 2),
      merr = c(0.2, 0.3),
      residuals = TRUE,
      resid = c(0.1, 0.2),
      beta = c(1, -0.5),
      xnames = c("x1", "x2"),
      betavcov = diag(2)
    ),
    class = "singleindex"
  )
  expect_equal(fitted(si), si$mean)
  expect_equal(se(si), si$merr)
  expect_equal(residuals(si), si$resid)
  expect_equal(unname(coef(si)), si$beta)
  expect_equal(names(coef(si)), si$xnames)
  expect_equal(vcov(si), si$betavcov)

  sc <- structure(
    list(mean = c(2, 4), merr = c(0.1, 0.2), residuals = TRUE, resid = c(0.2, 0.3)),
    class = "smoothcoefficient"
  )
  expect_equal(fitted(sc), sc$mean)
  expect_equal(se(sc), sc$merr)
  expect_equal(residuals(sc), sc$resid)
})

test_that("extractors fail clearly when optional quantities were not stored", {
  qreg <- structure(list(quantgrad = NA, quantgerr = NA), class = "qregression")
  expect_error(gradients(qreg), "gradients are not available")
  expect_error(gradients(qreg, errors = TRUE), "gradient standard errors are not available")
  expect_error(gradients(qreg, errors = "yes"), "'errors' must be TRUE or FALSE", fixed = TRUE)

  conmode <- structure(list(), class = "conmode")
  expect_error(gradients(conmode), "class-probability gradients/effects are not available")
  expect_error(gradients(conmode, errors = TRUE), "gradient standard errors are not available")
  expect_error(gradients(conmode, errors = "yes"), "'errors' must be TRUE or FALSE", fixed = TRUE)

  cdens <- structure(
    list(conderr = c(0.1, 0.2), proper.applied = TRUE),
    class = "condensity"
  )
  expect_error(se(cdens), "standard errors are unavailable for repaired conditional densities")

  cdist <- structure(
    list(conderr = c(0.1, 0.2), proper.applied = TRUE),
    class = "condistribution"
  )
  expect_error(se(cdist), "standard errors are unavailable for repaired conditional distributions")
})

test_that("conmode modal-class extraction uses predict rather than a fake mode method", {
  ns <- asNamespace(getNamespaceName(environment(npconmode)))
  s3 <- getNamespaceInfo(ns, "S3methods")

  expect_false(exists("mode.conmode", envir = ns, inherits = FALSE))
  expect_false(any(s3[, 1L] == "mode" & s3[, 2L] == "conmode"))

  fit <- structure(list(conmode = factor(c("a", "b", "a"))), class = "conmode")
  expect_equal(predict(fit), fit$conmode)
  expect_identical(mode(fit), "list")
})

test_that("npreg rejects non-logical gradients and residuals", {
  set.seed(20260223)
  dat <- data.frame(y = rnorm(30), x = runif(30))
  bw <- np::npregbw(
    y ~ x,
    data = dat,
    bws = 0.45,
    bandwidth.compute = FALSE,
    regtype = "lc"
  )

  expect_error(np::npreg(bws = bw, data = dat, gradients = "foo"),
               "'gradients' must be TRUE or FALSE")
  expect_error(np::npreg(bws = bw, data = dat, residuals = "bar"),
               "'residuals' must be TRUE or FALSE")
})

test_that("npreg.rbandwidth validates gradient flags as logical scalars", {
  npreg_rbandwidth <- getFromNamespace("npreg.rbandwidth", "np")
  fn.body <- paste(deparse(body(npreg_rbandwidth), width.cutoff = 500L), collapse = " ")
  expect_match(fn.body, "gradients <- npValidateScalarLogical\\(gradients, \"gradients\"\\)")
  expect_match(fn.body, "residuals <- npValidateScalarLogical\\(residuals, \"residuals\"\\)")
})

test_that("lp regtype remains lp internally for degree 0/1", {
  npRegtypeToC <- getFromNamespace("npRegtypeToC", "np")
  REGTYPE_LP <- getFromNamespace("REGTYPE_LP", "np")

  expect_identical(npRegtypeToC(regtype = "lp", degree = 0L, ncon = 1L)$code, REGTYPE_LP)
  expect_identical(npRegtypeToC(regtype = "lp", degree = 1L, ncon = 1L)$code, REGTYPE_LP)
})

test_that("lp degree-0 gradients fail fast while value path remains available", {
  set.seed(20260305)
  dat <- data.frame(y = rnorm(30), x = runif(30))
  bw <- np::npregbw(
    y ~ x,
    data = dat,
    bws = 0.45,
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 0
  )

  expect_identical(bw$regtype, "lp")
  expect_error(
    np::npreg(bws = bw, data = dat, gradients = TRUE),
    "regtype='lp' with degree=0 does not support derivatives"
  )
  expect_no_error(
    fit <- np::npreg(bws = bw, data = dat, gradients = FALSE)
  )
  expect_identical(fit$bws$regtype, "lp")
})

test_that("lp bernstein OOS warns but keeps canonical npreg path", {
  set.seed(20260305)
  dat <- data.frame(y = rnorm(40), x = runif(40))
  bw <- np::npregbw(
    y ~ x,
    data = dat,
    bws = 0.35,
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 2,
    bernstein.basis = TRUE
  )
  ex <- data.frame(x = c(-0.2, 1.2))
  expect_warning(
    fit <- np::npreg(bws = bw, txdat = dat["x"], tydat = dat$y, exdat = ex),
    "outside training support"
  )
  expect_identical(fit$bws$regtype, "lp")
})

test_that("npreg.rbandwidth no longer contains bernstein OOS direct fallback", {
  npreg_rbandwidth <- getFromNamespace("npreg.rbandwidth", "np")
  fn.body <- paste(deparse(body(npreg_rbandwidth), width.cutoff = 500L), collapse = " ")
  expect_no_match(fn.body, "bernstein\\.oos")
  expect_no_match(fn.body, "\\.np_kernel_weights_direct")
  expect_no_match(fn.body, "\\.npreghat_solve_eval")
})

test_that("npconmode proper helper enforces binary complement probabilities", {
  helper <- getFromNamespace(".npConmodeProperProbabilities", "np")
  raw <- matrix(c(-0.2, 1.2,
                  0.4, 0.8,
                  0.3, 0.3), ncol = 2L, byrow = TRUE)

  out <- helper(raw, levels = c("0", "1"), proper = TRUE)

  expect_true(isTRUE(out$proper.requested))
  expect_true(isTRUE(out$proper.applied))
  expect_equal(rowSums(out$probabilities), rep(1, nrow(raw)), tolerance = 1e-12)
  expect_true(all(out$probabilities >= -1e-12))
  expect_equal(out$probabilities[, 2L], 1 - out$probabilities[, 1L], tolerance = 1e-12)
  expect_identical(out$proper.info$reason, "projected")
})

test_that("npconmode proper controls fail fast on invalid control objects", {
  helper <- getFromNamespace(".npConmodeProperProbabilities", "np")
  raw <- matrix(c(0.2, 0.8, 0.4, 0.6), ncol = 2L, byrow = TRUE)

  expect_error(helper(raw, levels = c("0", "1"), proper.control = 1),
               "'proper.control' must be a list", fixed = TRUE)
  expect_error(helper(raw, levels = c("0", "1"), proper.control = list(foo = 1)),
               "unused argument")
  expect_error(helper(raw, levels = c("0", "1"), proper.control = list(tol = -1)),
               "'proper.control$tol' must be a non-negative scalar", fixed = TRUE)
})

test_that("npconmode bandwidth route reports all-NA training data clearly", {
  set.seed(20260511)
  d <- data.frame(x = runif(30), y = factor(rbinom(30, 1L, .5), levels = 0:1))
  bw <- npcdensbw(y ~ x, data = d, nmulti = 1L)
  badx <- data.frame(x = rep(NA_real_, 10))
  bady <- data.frame(y = factor(rep(NA, 10), levels = 0:1))

  expect_error(npconmode(bws = bw, txdat = badx, tydat = bady),
               "Training data has no rows without NAs", fixed = TRUE)
})

test_that("npconmode rejects non-categorical responses before modal fitting", {
  set.seed(20260513)
  d <- data.frame(x = runif(30), y = rnorm(30))

  expect_error(
    npconmode(y ~ x, data = d, bws = c(0.4, 0.4),
              bandwidth.compute = FALSE),
    "'tydat' must consist of one (1) discrete variable",
    fixed = TRUE
  )
})

test_that("npconmode proper defaults follow the canonical regression type", {
  effective <- getFromNamespace(".npConmodeEffectiveProper", "np")

  expect_false(effective(list(regtype = "lc"), NULL))
  expect_true(effective(list(regtype = "ll"), NULL))
  expect_true(effective(list(regtype = "lp"), NULL))
  expect_false(effective(list(regtype = "lp", regtype.engine = "lc"), NULL))
  expect_true(effective(list(regtype = "lc", regtype.engine = "lp"), NULL))
  expect_false(effective(list(regtype = "lp"), FALSE))
  expect_true(effective(list(regtype = "lc"), TRUE))
})

test_that("npconmode formula route forwards NOMAD shortcut controls", {
  skip_if_not_installed("crs")

  set.seed(20260511)
  n <- 30L
  d <- data.frame(
    x = seq(-1, 1, length.out = n)
  )
  d$y <- factor(rbinom(n, 1L, plogis(1.5 * d$x)))

  capture.output(
    fit <- npconmode(
      y ~ x,
      data = d,
      nomad = TRUE,
      nmulti = 1L,
      nomad.nmulti = 1L,
      degree.max = 1L
    )
  )

  expect_s3_class(fit, "conmode")
  expect_true(isTRUE(fit$bws$nomad.shortcut$enabled))
  expect_identical(as.character(fit$bws$regtype.engine), "lp")
})

test_that("npconmode formula route treats response-free newdata strictly", {
  nested_call <- function() {
    set.seed(20260511)
    d <- data.frame(x = sort(runif(50)))
    d$y <- factor(rbinom(50, 1L, plogis(1.2 * d$x)), levels = 0:1)
    nd <- data.frame(x = c(.1, .4, .8))

    npconmode(
      y ~ x,
      data = d,
      newdata = nd,
      regtype = "ll",
      bwmethod = "cv.ls",
      nmulti = 1L,
      probabilities = TRUE,
      gradients = TRUE,
      level = "1"
    )
  }

  fit <- nested_call()
  expect_s3_class(fit, "conmode")
  expect_equal(length(fitted(fit)), 3L)
  expect_equal(nrow(predict(fit, type = "prob")), 3L)
  expect_equal(NROW(gradients(fit)), 3L)
})

test_that("npconmode formula route keeps native exdat out of bandwidth selection", {
  set.seed(20260514)
  d <- data.frame(x = sort(runif(40)))
  d$y <- factor(rbinom(40, 1L, plogis(0.5 + d$x)), levels = 0:1)
  nd <- data.frame(x = c(0.2, 0.5, 0.8))

  fit <- npconmode(
    y ~ x,
    data = d,
    exdat = nd,
    probabilities = TRUE,
    nmulti = 1
  )

  expect_s3_class(fit, "conmode")
  expect_equal(length(fitted(fit)), nrow(nd))
  expect_equal(nrow(predict(fit, type = "prob")), nrow(nd))
})

test_that("npconmode NOMAD fit metadata propagates through evaluation and bandwidth plots", {
  skip_if_not_installed("crs")

  count_namespace_calls <- function(fun, expr) {
    ns <- asNamespace("np")
    ctr <- new.env(parent = emptyenv())
    ctr$n <- 0L
    trace(
      what = fun,
      where = ns,
      tracer = bquote(.(ctr)$n <- .(ctr)$n + 1L),
      print = FALSE
    )
    on.exit(try(untrace(fun, where = ns), silent = TRUE), add = TRUE)
    force(expr)
    ctr$n
  }

  set.seed(20260511)
  n <- 36L
  d <- data.frame(x = seq(-1, 1, length.out = n))
  d$y <- factor(rbinom(n, 1L, plogis(1.5 * d$x)))
  nd <- data.frame(x = c(-0.75, -0.1, 0.35, 0.8))

  capture.output(
    fit <- npconmode(
      y ~ x,
      data = d,
      newdata = nd,
      nomad = TRUE,
      nmulti = 1L,
      nomad.nmulti = 1L,
      degree.max = 1L,
      probabilities = TRUE
    )
  )

  expect_s3_class(fit, "conmode")
  expect_true(isTRUE(fit$bws$nomad.shortcut$enabled))
  expect_identical(as.character(fit$bws$regtype.engine), "lp")
  expect_length(fit$conmode, nrow(nd))
  expect_length(fitted(fit), nrow(nd))
  expect_equal(nrow(fit$probabilities), nrow(nd))
  expect_true(all(fit$probabilities >= -1e-12))
  expect_equal(rowSums(fit$probabilities), rep(1, nrow(nd)), tolerance = 1e-12)

  bw.calls <- count_namespace_calls("npcdensbw", {
    out <- plot(
      fit$bws,
      xdat = d["x"],
      ydat = d["y"],
      output = "data",
      perspective = FALSE,
      neval = 6L
    )
    expect_type(out, "list")
  })
  expect_identical(bw.calls, 0L)
})

test_that("npconmode binary class-probability gradients are stored and plotted", {
  set.seed(20260511)
  n <- 70L
  d <- data.frame(x = seq(-1, 1, length.out = n))
  d$y <- factor(rbinom(n, 1L, plogis(1.25 * d$x)), levels = 0:1)

  capture.output(
    fit <- npconmode(
      y ~ x,
      data = d,
      regtype = "ll",
      bwmethod = "cv.ls",
      nmulti = 1L,
      probabilities = TRUE,
      gradients = TRUE
    )
  )

  expect_s3_class(fit, "conmode")
  expect_true(isTRUE(fit$gradients))
  expect_equal(rowSums(fit$probabilities), rep(1, n), tolerance = 1e-10)
  expect_equal(dim(fit$probability.errors), dim(fit$probabilities))
  expect_equal(colnames(fit$probability.errors), colnames(fit$probabilities))
  expect_equal(as.character(fit$probability.gradient.level), "0")
  expect_equal(predict(fit, type = "class"), fit$conmode)
  expect_equal(predict(fit, type = "prob"), fit$probabilities)

  gout <- gradients(fit)
  expect_equal(dim(gout), c(n, 1L))

  direct <- npcdens(
    bws = fit$bws,
    txdat = d["x"],
    tydat = d["y"],
    exdat = d["x"],
    eydat = factor(rep("0", n), levels = levels(d$y)),
    gradients = TRUE
  )
  expect_equal(as.vector(gout), as.vector(direct$congrad), tolerance = 1e-10)
  expect_error(gradients(fit, errors = TRUE), "standard errors")
  expect_error(gradients(fit, level = "1"), "stored class-probability gradients")

  pdata <- plot(fit, output = "data")
  expect_named(pdata, "x")
  expect_equal(nrow(pdata$x), n)
  expect_equal(plot(fit, output = "both"), pdata)

  perr <- plot(fit, output = "data", errors = "asymptotic",
               band = "all", level = "0")
  expect_true(all(c("stderr", "lower", "upper",
                    "pointwise.lower", "bonferroni.upper") %in%
                    names(perr$x)))
  expect_equal(perr$x$stderr, fit$probability.errors[, "0"],
               tolerance = 1e-12, ignore_attr = TRUE)

  grdata <- plot(fit, gradients = TRUE, output = "data")
  expect_named(grdata, "x")
  expect_true("effect" %in% names(grdata$x))
  expect_error(plot(fit, gradients = TRUE, level = "1", output = "data"),
               "stored class-probability gradients")

  grid.data <- plot(fit, view = "fixed", neval = 9L, output = "data")
  expect_named(grid.data, "x")
  expect_equal(nrow(grid.data$x), 9L)
  expect_equal(unique(grid.data$x$view), "fixed")
  grid.prob <- predict(fit,
                       newdata = data.frame(x = grid.data$x$x),
                       type = "prob")[, "0"]
  expect_equal(grid.data$x$probability, grid.prob, tolerance = 1e-5)

  grid.err <- plot(fit, view = "fixed", neval = 9L, output = "data",
                   errors = "asymptotic", band = "pointwise", level = "0")
  expect_true(all(c("stderr", "lower", "upper") %in% names(grid.err$x)))
  expect_true(any(is.finite(grid.err$x$stderr)))
  pred.se <- predict(fit,
                     newdata = data.frame(x = grid.data$x$x),
                     type = "prob",
                     se.fit = TRUE)
  expect_named(pred.se, c("fit", "se.fit"))
  expect_equal(pred.se$fit[, "0"], grid.prob, tolerance = 1e-5)
  expect_true(any(is.finite(pred.se$se.fit[, "0"])))
  expect_error(predict(fit, type = "class", se.fit = TRUE),
               "type=\"prob\"", fixed = TRUE)

  set.seed(20260514)
  grid.boot <- plot(fit, view = "fixed", neval = 7L, output = "data",
                    errors = "bootstrap", bootstrap = "inid", B = 39L,
                    band = "pointwise", level = "0")
  expect_true(all(c("stderr", "center", "lower", "upper") %in%
                    names(grid.boot$x)))
  expect_true(any(is.finite(grid.boot$x$stderr)))
  expect_true(all(is.finite(grid.boot$x$lower)))
  expect_true(all(is.finite(grid.boot$x$upper)))
  expect_error(plot(fit, view = "fixed", neval = 7L, output = "data",
                    errors = "bootstrap", bootstrap = "wild"),
               "wild bootstrap is not available")

  grid.grad <- plot(fit, view = "fixed", neval = 9L,
                    gradients = TRUE, output = "data")
  expect_equal(nrow(grid.grad$x), 9L)
  expect_true("effect" %in% names(grid.grad$x))

  pdf(tempfile(fileext = ".pdf"))
  expect_silent(plot(fit))
  expect_silent(plot(fit, view = "fixed", neval = 9L))
  expect_silent(plot(fit, col = "red", lwd = 2, lty = 2, type = "l"))
  expect_silent(plot(fit, gradients = TRUE))
  expect_silent(plot(fit, gradients = TRUE, view = "fixed", neval = 9L))
  expect_silent(plot(fit, gradients = TRUE, col = "blue", lwd = 2, lty = 3,
                     type = "l"))
  expect_silent(plot(fit, errors = "bootstrap", B = 39L))
  expect_error(plot(fit, gradients = TRUE, errors = "asymptotic"),
               "not probability gradients")
  expect_error(plot(fit, renderer = "rgl"),
               "renderer='rgl' is supported only")
  grDevices::dev.off()
})

test_that("npconmode fixed surface payload is object-fed and predict-aligned", {
  set.seed(20260514)
  n <- 60L
  d <- data.frame(
    x1 = runif(n, -1, 1),
    x2 = runif(n, -1, 1),
    z = factor(sample(c("a", "b"), n, replace = TRUE))
  )
  eta <- 0.5 + d$x1 - 0.75 * d$x2 + 0.25 * (d$z == "b")
  d$y <- factor(rbinom(n, 1L, plogis(eta)), levels = 0:1)

  capture.output(
    fit <- npconmode(
      y ~ x1 + x2 + z,
      data = d,
      regtype = "ll",
      bwmethod = "cv.ls",
      nmulti = 1L,
      probabilities = TRUE
    )
  )

  surf <- plot(fit, view = "fixed", perspective = TRUE,
               neval = 7L, output = "data")
  expect_s3_class(surf, "np_conmode_surface_data")
  expect_named(surf, c("surface", "grid", "held", "level", "levels", "proper"))
  expect_equal(surf$grid$variables, c("x1", "x2"))
  expect_equal(surf$grid$dims, c(7L, 7L))
  expect_equal(nrow(surf$surface), 49L)
  expect_true("z" %in% names(surf$held))
  expect_equal(unique(surf$surface$view), "fixed")
  expect_true(all(is.finite(surf$surface$probability)))
  expect_true(all(surf$surface$probability >= -1e-10))
  expect_true(all(surf$surface$probability <= 1 + 1e-10))

  surf.err <- plot(fit, view = "fixed", perspective = TRUE,
                   neval = 5L, output = "data",
                   errors = "asymptotic", level = surf$level)
  expect_true(all(c("stderr", "lower", "upper") %in% names(surf.err$surface)))
  expect_true(any(is.finite(surf.err$surface$stderr)))
  expect_error(plot(fit, view = "fixed", perspective = TRUE,
                    neval = 5L, errors = "asymptotic"),
               "output=\"data\" only", fixed = TRUE)
  expect_error(plot(fit, view = "fixed", perspective = TRUE,
                    neval = 5L, errors = "bootstrap", output = "data",
                    B = 39L),
               "surface bootstrap intervals")

  nd <- data.frame(
    x1 = surf$surface$x1,
    x2 = surf$surface$x2,
    z = surf$held$z[rep(1L, nrow(surf$surface))]
  )
  pred <- predict(fit, newdata = nd, type = "prob")[, surf$level]
  expect_equal(surf$surface$probability, pred, tolerance = 1e-5)

  surf.alias <- plot(fit, view = "fixed", persp = TRUE,
                     neval = 5L, output = "data",
                     plot.vars = c("x2", "x1"))
  expect_equal(surf.alias$grid$variables, c("x2", "x1"))
  expect_equal(surf.alias$grid$dims, c(5L, 5L))

  expect_silent({
    grDevices::pdf(tempfile(fileext = ".pdf"))
    on.exit(grDevices::dev.off(), add = TRUE)
    plot(fit, perspective = TRUE, neval = 5L)
    rendered <- plot(fit, perspective = TRUE, neval = 5L,
                     output = "plot-data")
  })
  data.only <- plot(fit, perspective = TRUE, neval = 5L,
                    output = "data")
  expect_equal(rendered, data.only)
  expect_equal(
    plot(fit, perspective = TRUE, renderer = "rgl", neval = 5L,
         output = "data"),
    data.only
  )
  expect_error(plot(fit, renderer = "rgl", output = "data"),
               "renderer='rgl' is supported only")
  if (suppressWarnings(requireNamespace("rgl", quietly = TRUE))) {
    rgl.rendered <- suppressWarnings(
      plot(fit, perspective = TRUE, renderer = "rgl", neval = 5L,
           output = "plot-data")
    )
    expect_equal(rgl.rendered, data.only)
  }
  expect_error(plot(fit, plot.vars = c("x1", "x2"), output = "data"),
               "'plot.vars' is only supported")
  expect_error(plot(fit, perspective = TRUE, persp = FALSE, output = "data"),
               "conflicting plot.conmode arguments")
  expect_error(plot(fit, perspective = TRUE, gradients = TRUE, output = "data"),
               "not gradients")
  expect_error(plot(fit, perspective = TRUE, plot.vars = c("x1", "z"),
                    output = "data"),
               "currently require continuous variables")

  fit.one <- npconmode(
    y ~ x1 + z,
    data = d,
    bws = c(0.5, 0.5, 0.5),
    bandwidth.compute = FALSE,
    probabilities = TRUE
  )
  expect_error(plot(fit.one, perspective = TRUE, output = "data"),
               "require exactly two continuous conditioning variables")
})

test_that("npconmode class-probability gradients are level-specific for multinomial responses", {
  set.seed(20260511)
  n <- 45L
  d <- data.frame(x = seq(-1, 1, length.out = n))
  d$y <- factor(sample(letters[1:3], n, replace = TRUE))

  capture.output(
    fit <- npconmode(
      y ~ x,
      data = d,
      regtype = "ll",
      bwmethod = "cv.ls",
      nmulti = 1L,
      probabilities = TRUE,
      level = "b",
      gradients = TRUE
    )
  )

  expect_s3_class(fit, "conmode")
  expect_equal(as.character(fit$probability.gradient.level), "b")
  expect_equal(dim(gradients(fit)), c(n, 1L))
  expect_true(all(fit$probabilities >= -1e-12))
  expect_equal(rowSums(fit$probabilities), rep(1, n), tolerance = 1e-10)
  expect_error(
    npconmode(
      y ~ x,
      data = d,
      regtype = "ll",
      bwmethod = "cv.ls",
      nmulti = 1L,
      probabilities = TRUE,
      level = "z",
      gradients = TRUE
    ),
    "'level' must identify one response level",
    fixed = TRUE
  )
  expect_error(gradients(fit, level = "c"), "stored class-probability gradients")
  grdata <- plot(fit, gradients = TRUE, output = "data")
  expect_equal(unique(grdata$x$level), "b")
})

library(np)

test_that("semiparametric formula estimators reject malformed newdata before parent-frame lookup", {
  set.seed(20260513)
  dat <- data.frame(
    y = rnorm(60),
    x1 = runif(60),
    x2 = runif(60),
    z = runif(60)
  )

  bw.si <- npindexbw(y ~ x1 + x2, data = dat,
                     bws = c(0.25, 0.25, 1),
                     bandwidth.compute = FALSE)
  fit.si <- npindex(bws = bw.si, gradients = TRUE, errors = TRUE)
  expect_error(npindex(bws = bw.si, newdata = data.frame(q = 1:2)),
               "newdata must contain columns")
  expect_error(predict(fit.si, newdata = data.frame(q = 1:2)),
               "newdata must contain columns")

  bw.sc <- npscoefbw(y ~ x1 | z, data = dat, bws = 0.25,
                     bandwidth.compute = FALSE)
  fit.sc <- npscoef(bws = bw.sc, errors = TRUE, iterate = FALSE)
  expect_error(npscoef(bws = bw.sc, newdata = data.frame(q = 1:2),
                       errors = FALSE, iterate = FALSE),
               "newdata must contain columns")
  expect_error(predict(fit.sc, newdata = data.frame(q = 1:2)),
               "newdata must contain columns")

  bw.pl <- npplregbw(y ~ x2 | z, data = dat,
                     bws = matrix(c(0.25, 0.25), nrow = 2),
                     bandwidth.compute = FALSE)
  fit.pl <- npplreg(bws = bw.pl)
  expect_error(npplreg(bws = bw.pl, newdata = data.frame(z = c(0.3, 0.5))),
               "newdata must contain columns")
  expect_error(predict(fit.pl, newdata = data.frame(x2 = c(0.3, 0.5))),
               "newdata must contain columns")
})

test_that("semiparametric predict methods retain native evaluation argument precedence", {
  set.seed(20260513)
  dat <- data.frame(
    y = rnorm(60),
    x1 = runif(60),
    x2 = runif(60),
    z = runif(60)
  )
  nd.standard <- data.frame(x1 = c(0.2, 0.4, 0.6),
                            x2 = c(0.1, 0.3, 0.5),
                            z = c(0.7, 0.5, 0.3))
  nd.native <- data.frame(x1 = c(0.15, 0.35, 0.55),
                          x2 = c(0.12, 0.32, 0.52),
                          z = c(0.65, 0.45, 0.25))

  bw.si <- npindexbw(y ~ x1 + x2, data = dat,
                     bws = c(0.25, 0.25, 1),
                     bandwidth.compute = FALSE)
  fit.si <- npindex(bws = bw.si)
  expect_equal(
    as.numeric(predict(fit.si,
                       newdata = nd.standard[c("x1", "x2")],
                       exdat = nd.native[c("x1", "x2")])),
    as.numeric(predict(fit.si, exdat = nd.native[c("x1", "x2")])),
    tolerance = 1e-12
  )

  bw.sc <- npscoefbw(y ~ x1 | z, data = dat, bws = 0.25,
                     bandwidth.compute = FALSE)
  fit.sc <- npscoef(bws = bw.sc, errors = TRUE, iterate = FALSE)
  expect_equal(
    as.numeric(predict(fit.sc,
                       newdata = nd.standard[c("x1", "z")],
                       exdat = nd.native["x1"],
                       ezdat = nd.native["z"])),
    as.numeric(predict(fit.sc, exdat = nd.native["x1"], ezdat = nd.native["z"])),
    tolerance = 1e-12
  )

  bw.pl <- npplregbw(y ~ x2 | z, data = dat,
                     bws = matrix(c(0.25, 0.25), nrow = 2),
                     bandwidth.compute = FALSE)
  fit.pl <- npplreg(bws = bw.pl)
  expect_equal(
    as.numeric(predict(fit.pl,
                       newdata = nd.standard[c("x2", "z")],
                       exdat = nd.native["x2"],
                       ezdat = nd.native["z"])),
    as.numeric(predict(fit.pl, exdat = nd.native["x2"], ezdat = nd.native["z"])),
    tolerance = 1e-12
  )
})

test_that("semiparametric S3 logical arguments fail clearly", {
  set.seed(20260513)
  dat <- data.frame(
    y = rnorm(60),
    x1 = runif(60),
    x2 = runif(60),
    z = runif(60)
  )

  bw.si <- npindexbw(y ~ x1 + x2, data = dat,
                     bws = c(0.25, 0.25, 1),
                     bandwidth.compute = FALSE)
  fit.si <- npindex(bws = bw.si, gradients = TRUE, errors = TRUE)
  expect_error(predict(fit.si, se.fit = "yes"),
               "'se.fit' must be TRUE or FALSE", fixed = TRUE)
  expect_error(gradients(fit.si, errors = "yes"),
               "'errors' must be TRUE or FALSE", fixed = TRUE)

  bw.sc <- npscoefbw(y ~ x1 | z, data = dat, bws = 0.25,
                     bandwidth.compute = FALSE)
  fit.sc <- npscoef(bws = bw.sc, errors = TRUE, iterate = FALSE)
  expect_error(predict(fit.sc, se.fit = "yes"),
               "'se.fit' must be TRUE or FALSE", fixed = TRUE)

  bw.pl <- npplregbw(y ~ x2 | z, data = dat,
                     bws = matrix(c(0.25, 0.25), nrow = 2),
                     bandwidth.compute = FALSE)
  fit.pl <- npplreg(bws = bw.pl)
  expect_error(predict(fit.pl, se.fit = "yes"),
               "'se.fit' must be TRUE or FALSE", fixed = TRUE)
})

library(np)

local_linear_guidance_dens_msg <- "C_np_density_conditional_bw: optimizer returned a fixed-bandwidth candidate with invalid raw objective"
local_linear_guidance_dist_msg <- "C_np_distribution_conditional_bw: optimizer returned a fixed-bandwidth candidate with invalid raw objective"

local_linear_guidance_ll_spec <- function(ncon = 1L) {
  np:::npCanonicalConditionalRegSpec(regtype = "ll", ncon = ncon)
}

local_linear_guidance_lp_spec <- function(ncon = 1L) {
  np:::npCanonicalConditionalRegSpec(
    regtype = "lp",
    degree = rep.int(1L, ncon),
    bernstein.basis = FALSE,
    ncon = ncon
  )
}

test_that("local-linear raw-basis guidance is targeted to conditional density ll cv.ls", {
  expect_error(
    np:::npWithLocalLinearRawBasisSearchError(
      stop(local_linear_guidance_dens_msg, call. = FALSE),
      where = "npcdensbw",
      spec = local_linear_guidance_ll_spec(),
      bwmethod = "cv.ls",
      ncon = 1L
    ),
    regexp = "npcdensbw\\(\\) local-linear cv\\.ls failed while using the canonical raw degree-1 basis"
  )

  expect_error(
    np:::npWithLocalLinearRawBasisSearchError(
      stop(local_linear_guidance_dens_msg, call. = FALSE),
      where = "npcdensbw",
      spec = local_linear_guidance_lp_spec(),
      bwmethod = "cv.ls",
      ncon = 1L
    ),
    regexp = local_linear_guidance_dens_msg
  )
})

test_that("local-linear raw-basis guidance is targeted to conditional distribution ll cv.ls", {
  expect_error(
    np:::npWithLocalLinearRawBasisSearchError(
      stop(local_linear_guidance_dist_msg, call. = FALSE),
      where = "npcdistbw",
      spec = local_linear_guidance_ll_spec(),
      bwmethod = "cv.ls",
      ncon = 1L
    ),
    regexp = "npcdistbw\\(\\) local-linear cv\\.ls failed while using the canonical raw degree-1 basis"
  )
})

test_that("local-linear raw-basis guidance leaves non-targeted errors unchanged", {
  expect_error(
    np:::npWithLocalLinearRawBasisSearchError(
      stop(local_linear_guidance_dist_msg, call. = FALSE),
      where = "npcdensbw",
      spec = local_linear_guidance_ll_spec(),
      bwmethod = "cv.ls",
      ncon = 1L
    ),
    regexp = local_linear_guidance_dist_msg
  )

  expect_error(
    np:::npWithLocalLinearRawBasisSearchError(
      stop(local_linear_guidance_dens_msg, call. = FALSE),
      where = "npcdensbw",
      spec = local_linear_guidance_ll_spec(),
      bwmethod = "cv.ml",
      ncon = 1L
    ),
    regexp = local_linear_guidance_dens_msg
  )

  expect_error(
    np:::npWithLocalLinearRawBasisSearchError(
      stop(local_linear_guidance_dens_msg, call. = FALSE),
      where = "npcdensbw",
      spec = local_linear_guidance_ll_spec(ncon = 0L),
      bwmethod = "cv.ls",
      ncon = 0L
    ),
    regexp = local_linear_guidance_dens_msg
  )
})

.np_semihat_require_class <- function(bws, class.name, fn.name) {
  if (!inherits(bws, class.name))
    stop(sprintf("argument 'bws' must inherit from class '%s' in %s()", class.name, fn.name))
}

.np_semihat_make_regbw_args <- function(source, xdat, ydat, bw) {
  xdat <- toFrame(xdat)
  ncon.target <- sum(untangle(xdat)$icon)
  spec <- npCanonicalConditionalRegSpec(
    regtype = if (is.null(source$regtype)) "lc" else as.character(source$regtype),
    basis = if (is.null(source$basis)) "glp" else as.character(source$basis),
    degree = source$degree,
    bernstein.basis = isTRUE(source$bernstein.basis),
    ncon = ncon.target,
    where = "npscoef"
  )

  collapse_bound <- function(v, nm) {
    if (is.null(v))
      return(NULL)
    vv <- as.double(v)
    if (ncon.target <= 0L)
      return(NULL)
    if (length(vv) <= 1L || length(vv) == ncon.target)
      return(vv)
    if (ncon.target == 1L) {
      uu <- unique(vv)
      if (length(uu) == 1L)
        return(uu)
      stop(sprintf("cannot collapse %s with %d distinct values to scalar helper bound", nm, length(uu)),
           call. = FALSE)
    }
    stop(sprintf("%s length (%d) is incompatible with helper continuous dimension (%d)",
                 nm, length(vv), ncon.target),
         call. = FALSE)
  }

  args <- list(
    xdat = xdat,
    ydat = ydat,
    bws = as.double(bw),
    regtype = spec$regtype.engine,
    bandwidth.compute = FALSE
  )

  # Forward the full compatible estimator/bandwidth option contract.
  args$basis <- spec$basis.engine
  args$degree <- spec$degree.engine
  args$bernstein.basis <- spec$bernstein.basis.engine
  if (!is.null(source$method) && source$method %in% c("cv.ls", "cv.aic"))
    args$bwmethod <- source$method
  if (!is.null(source$scaling))
    args$bwscaling <- isTRUE(source$scaling)
  if (!is.null(source$type))
    args$bwtype <- source$type
  if (!is.null(source$ckertype))
    args$ckertype <- source$ckertype
  if (!is.null(source$ckerorder))
    args$ckerorder <- source$ckerorder
  if (!is.null(source$ckerbound))
    args$ckerbound <- source$ckerbound
  if (!is.null(source$ckerlb))
    args$ckerlb <- collapse_bound(source$ckerlb, "ckerlb")
  if (!is.null(source$ckerub))
    args$ckerub <- collapse_bound(source$ckerub, "ckerub")
  if (!is.null(source$ukertype))
    args$ukertype <- source$ukertype
  if (!is.null(source$okertype))
    args$okertype <- source$okertype

  args
}

.np_semihat_make_regbw_state <- function(source, xdat, ydat, bw) {
  resolve_choice <- function(value, choices, default = choices[1L]) {
    if (is.null(value) || !length(value))
      return(default)
    cand <- as.character(value)[1L]
    if (!cand %in% choices)
      return(default)
    cand
  }

  xdat <- toFrame(xdat)
  if (!(is.vector(ydat) || is.factor(ydat)))
    stop("'ydat' must be a vector")

  args <- .np_semihat_make_regbw_args(source = source, xdat = xdat, ydat = ydat, bw = bw)
  xdati <- untangle(xdat)
  ydati <- untangle(data.frame(ydat))
  xmat <- toMatrix(xdat)
  rcon <- xmat[, xdati$icon, drop = FALSE]
  nobs <- nrow(xdat)
  ncon <- sum(xdati$icon)
  method <- resolve_choice(if (!is.null(args$bwmethod)) args$bwmethod else source$method,
                           c("cv.ls", "cv.aic"))
  bwtype <- resolve_choice(if (!is.null(args$bwtype)) args$bwtype else source$type,
                           c("fixed", "generalized_nn", "adaptive_nn"))
  ckertype <- resolve_choice(if (!is.null(args$ckertype)) args$ckertype else source$ckertype,
                             c("gaussian", "truncated gaussian", "epanechnikov", "uniform"))
  ckerorder <- if (!is.null(args$ckerorder)) args$ckerorder else source$ckerorder
  ckerbound <- resolve_choice(if (!is.null(args$ckerbound)) args$ckerbound else source$ckerbound,
                              c("none", "range", "fixed"))
  ukertype <- resolve_choice(if (!is.null(args$ukertype)) args$ukertype else source$ukertype,
                             c("aitchisonaitken", "liracine"))
  okertype <- resolve_choice(if (!is.null(args$okertype)) args$okertype else source$okertype,
                             c("liracine", "wangvanryzin", "racineliyan", "nliracine"))
  scaling <- if (!is.null(args$bwscaling)) isTRUE(args$bwscaling) else FALSE
  bw.vec <- stats::setNames(as.double(args$bws), names(xdat))
  sfactor <- bandwidth <- bw.vec
  nconfac <- nobs^(-1.0 / (2.0 * ckerorder + ncon))
  ncatfac <- nobs^(-2.0 / (2.0 * ckerorder + ncon))
  sdev <- if (ncon > 0L) EssDee(rcon) else numeric(0)

  if (sum(xdati$iuno) > 0L) {
    if (scaling) {
      bandwidth[xdati$iuno] <- bandwidth[xdati$iuno] * ncatfac
    } else {
      sfactor[xdati$iuno] <- sfactor[xdati$iuno] / ncatfac
    }
  }

  if (sum(xdati$iord) > 0L) {
    if (scaling) {
      bandwidth[xdati$iord] <- bandwidth[xdati$iord] * ncatfac
    } else {
      sfactor[xdati$iord] <- sfactor[xdati$iord] / ncatfac
    }
  }

  if (ncon > 0L) {
    dfactor <- sdev * nconfac
    if (scaling) {
      bandwidth[xdati$icon] <- bandwidth[xdati$icon] * dfactor
    } else {
      sfactor[xdati$icon] <- sfactor[xdati$icon] / dfactor
    }
  }

  out <- rbandwidth(
    bw = bw.vec,
    regtype = args$regtype,
    basis = args$basis,
    degree = args$degree,
    bernstein.basis = args$bernstein.basis,
    bwmethod = method,
    bwscaling = scaling,
    bwtype = bwtype,
    ckertype = ckertype,
    ckerorder = ckerorder,
    ckerbound = ckerbound,
    ckerlb = args$ckerlb,
    ckerub = args$ckerub,
    ukertype = ukertype,
    okertype = okertype,
    fval = NA,
    ifval = NA,
    num.feval = NA,
    num.feval.fast = NA,
    fval.history = NA,
    eval.history = NA,
    invalid.history = NA,
    nobs = nobs,
    xdati = xdati,
    ydati = ydati,
    xnames = names(xdat),
    ynames = if (!is.null(source$ynames) && length(source$ynames)) source$ynames else "y",
    sfactor = sfactor,
    bandwidth = bandwidth,
    rows.omit = NA,
    nconfac = nconfac,
    ncatfac = ncatfac,
    sdev = sdev,
    bandwidth.compute = FALSE,
    timing = NA,
    total.time = c(elapsed = 0.0)
  )
  out$call <- call(".np_semihat_make_regbw_state")
  out
}

.npscoef_canonical_spec <- function(source, zdat, where = "npscoef") {
  zdat <- toFrame(zdat)
  npCanonicalConditionalRegSpec(
    regtype = if (is.null(source$regtype)) "lc" else as.character(source$regtype),
    basis = if (is.null(source$basis)) "glp" else as.character(source$basis),
    degree = source$degree,
    bernstein.basis = isTRUE(source$bernstein.basis),
    ncon = sum(untangle(zdat)$icon),
    where = where
  )
}

.np_indexhat_rbw <- function(bws, idx.train) {
  .np_semihat_make_regbw_state(
    source = bws,
    xdat = idx.train,
    ydat = rep.int(0.0, nrow(idx.train)),
    bw = bws$bw
  )
}

.np_indexhat_ll_owner_rbw <- function(bws, idx.train) {
  base <- .np_indexhat_rbw(bws = bws, idx.train = idx.train)
  npregbw(
    xdat = idx.train,
    ydat = rep.int(0.0, nrow(idx.train)),
    bws = base$bw,
    regtype = "ll",
    bwtype = base$type,
    bandwidth.compute = FALSE,
    ckertype = base$ckertype,
    ckerorder = base$ckerorder,
    ckerbound = base$ckerbound,
    ckerlb = base$ckerlb,
    ckerub = base$ckerub,
    ukertype = base$ukertype,
    okertype = base$okertype
  )
}

.np_indexhat_numeric_y <- function(y) {
  if (is.factor(y) || is.vector(y))
    return(matrix(as.double(y), ncol = 1L))

  as.matrix(y)
}

.np_indexhat_kbw <- function(bws, idx.train) {
  resolve_choice <- function(value, choices, default = choices[1L]) {
    if (is.null(value) || !length(value))
      return(default)
    cand <- as.character(value)[1L]
    if (!cand %in% choices)
      return(default)
    cand
  }

  collapse_bound <- function(v, nm) {
    if (is.null(v))
      return(NULL)
    vv <- as.double(v)
    if (length(vv) <= 1L)
      return(vv)
    uu <- unique(vv)
    if (length(uu) == 1L)
      return(uu)
    stop(sprintf("cannot collapse %s with %d distinct values to scalar helper bound",
                 nm, length(uu)),
         call. = FALSE)
  }

  bwtype <- resolve_choice(bws$type, c("fixed", "generalized_nn", "adaptive_nn"))
  ckertype <- resolve_choice(bws$ckertype, c("gaussian", "truncated gaussian", "epanechnikov", "uniform"))
  ckerbound <- resolve_choice(bws$ckerbound, c("none", "range", "fixed"))
  ukertype <- resolve_choice(bws$ukertype, c("aitchisonaitken", "liracine"))
  okertype <- resolve_choice(bws$okertype, c("liracine", "wangvanryzin", "racineliyan", "nliracine"))
  kbandwidth.numeric(
    bw = c(bws$bw),
    bwtype = bwtype,
    bwscaling = FALSE,
    ckertype = ckertype,
    ckerorder = bws$ckerorder,
    ckerbound = ckerbound,
    ckerlb = collapse_bound(bws$ckerlb, "ckerlb"),
    ckerub = collapse_bound(bws$ckerub, "ckerub"),
    ukertype = ukertype,
    okertype = okertype,
    nobs = nrow(idx.train),
    xdati = untangle(idx.train),
    ydati = bws$ydati,
    xnames = "index",
    ynames = bws$ynames,
    xdat = idx.train
  )
}

.np_indexhat_core <- function(bws,
                              idx.train,
                              idx.eval,
                              y = NULL,
                              output = c("matrix", "apply"),
                              ridge = 0.0) {
  output <- match.arg(output)
  kbw <- .np_indexhat_kbw(bws = bws, idx.train = idx.train)
  spec <- .npindex_resolve_spec(bws, where = "npindexhat")
  regtype <- spec$regtype.engine

  kw <- .np_kernel_weights_direct(
    bws = kbw,
    txdat = idx.train,
    exdat = idx.eval,
    bandwidth.divide = TRUE,
    kernel.pow = 1.0
  )

  if (!is.matrix(kw))
    kw <- matrix(kw, nrow = nrow(idx.train))

  ntrain <- nrow(idx.train)
  neval <- nrow(idx.eval)
  if (nrow(kw) != ntrain || ncol(kw) != neval)
    stop("single-index hat kernel-weight matrix shape mismatch", call. = FALSE)

  if (!is.null(y)) {
    y <- .np_indexhat_numeric_y(y)
    if (nrow(y) != ntrain)
      stop("number of rows in 'y' must equal number of training rows")
  }

  if (identical(regtype, "lc")) {
    den <- pmax(colSums(kw), .Machine$double.eps)
    if (identical(output, "matrix"))
      return(sweep(t(kw), 1L, den, "/", check.margin = FALSE))

    out <- sweep(t(kw), 1L, den, "/", check.margin = FALSE) %*% y
    return(if (ncol(out) == 1L) as.vector(out) else out)
  }

  degree <- if (identical(regtype, "ll")) {
    1L
  } else {
    spec$degree.engine
  }

  W <- W.lp(
    xdat = idx.train,
    degree = degree,
    basis = spec$basis.engine,
    bernstein.basis = spec$bernstein.basis.engine
  )
  W.eval <- W.lp(
    xdat = idx.train,
    exdat = idx.eval,
    degree = degree,
    basis = spec$basis.engine,
    bernstein.basis = spec$bernstein.basis.engine
  )

  if (!is.matrix(W))
    W <- matrix(W, nrow = ntrain)
  if (!is.matrix(W.eval))
    W.eval <- matrix(W.eval, nrow = neval)
  if (nrow(W.eval) != neval)
    W.eval <- matrix(W.eval, nrow = neval, byrow = FALSE)

  if (identical(output, "matrix")) {
    H <- matrix(NA_real_, nrow = neval, ncol = ntrain)
  } else {
    out <- matrix(0.0, nrow = neval, ncol = ncol(y))
  }

  for (i in seq_len(neval)) {
    solve.out <- .npreghat_solve_eval(
      W = W,
      w.eval = W.eval[i, ],
      k = kw[, i],
      ridge.base = ridge
    )

    if (is.null(solve.out))
      stop(sprintf("failed to solve single-index hat system at evaluation row %d", i),
           call. = FALSE)

    h.row <- kw[, i] * drop(W %*% solve.out$v)

    if (identical(output, "matrix")) {
      H[i, ] <- h.row
    } else {
      out[i, ] <- drop(crossprod(h.row, y))
    }
  }

  if (identical(output, "matrix"))
    H
  else if (ncol(out) == 1L)
    as.vector(out)
  else
    out
}

.np_indexhat_lc_kernel_weights <- function(bws, idx.train, idx.eval) {
  kw <- npksum(
    txdat = idx.train,
    exdat = idx.eval,
    bws = bws$bw,
    bwtype = bws$type,
    ckertype = bws$ckertype,
    ckerorder = bws$ckerorder,
    return.kernel.weights = TRUE
  )$kw

  if (!is.matrix(kw))
    kw <- matrix(kw, nrow = nrow(idx.train))

  if (nrow(kw) != nrow(idx.train) || ncol(kw) != nrow(idx.eval))
    stop("single-index lc kernel-weight matrix shape mismatch", call. = FALSE)

  kw
}

.np_indexhat_lc_mean <- function(bws,
                                 idx.train,
                                 idx.eval,
                                 y = NULL,
                                 output = c("matrix", "apply")) {
  output <- match.arg(output)
  kw <- .np_indexhat_lc_kernel_weights(
    bws = bws,
    idx.train = idx.train,
    idx.eval = idx.eval
  )
  H <- sweep(
    t(kw),
    1L,
    pmax(colSums(kw), .Machine$double.eps),
    "/",
    check.margin = FALSE
  )

  if (identical(output, "matrix"))
    return(H)

  if (is.null(y))
    stop("argument 'y' is required when output='apply'")

  y <- .np_indexhat_numeric_y(y)
  if (nrow(y) != nrow(idx.train))
    stop("number of rows in 'y' must equal number of training rows")

  out <- H %*% y
  if (ncol(out) == 1L) as.vector(out) else out
}

.np_indexhat_lc_derivative <- function(bws,
                                       idx.train,
                                       idx.eval,
                                       y = NULL,
                                       output = c("matrix", "apply")) {
  output <- match.arg(output)
  out <- npreghat(
    bws = .np_indexhat_rbw(bws = bws, idx.train = idx.train),
    txdat = idx.train,
    exdat = idx.eval,
    y = y,
    output = output,
    s = 1L
  )

  if (!identical(output, "matrix"))
    return(out)

  matrix(
    as.double(out),
    nrow = nrow(out),
    ncol = ncol(out),
    dimnames = dimnames(out)
  )
}

.np_indexhat_lp_mean_matrix <- function(bws, idx.train, idx.eval) {
  out <- npreghat(
    bws = .np_indexhat_rbw(bws = bws, idx.train = idx.train),
    txdat = idx.train,
    exdat = idx.eval,
    output = "matrix",
    s = 0L
  )

  matrix(
    as.double(out),
    nrow = nrow(out),
    ncol = ncol(out),
    dimnames = dimnames(out)
  )
}

.np_indexhat_gradient_matrix <- function(bws, idx.train, idx.eval) {
  regtype <- if (!is.null(bws$regtype)) as.character(bws$regtype)[1L] else "lc"
  owner.bw <- if (identical(regtype, "ll")) {
    .np_indexhat_ll_owner_rbw(bws = bws, idx.train = idx.train)
  } else {
    .np_indexhat_rbw(bws = bws, idx.train = idx.train)
  }

  out <- npreghat(
    bws = owner.bw,
    txdat = idx.train,
    exdat = idx.eval,
    output = "matrix",
    s = 1L
  )

  matrix(
    as.double(out),
    nrow = nrow(out),
    ncol = ncol(out),
    dimnames = dimnames(out)
  )
}

.np_indexhat_exact <- function(bws,
                               idx.train,
                               idx.eval,
                               y = NULL,
                               output = c("matrix", "apply"),
                               s = 0L) {
  output <- match.arg(output)
  spec <- .npindex_resolve_spec(bws, where = "npindexhat")
  regtype <- if (!is.null(bws$regtype)) as.character(bws$regtype)[1L] else spec$regtype.engine
  regtype.engine <- spec$regtype.engine
  if (identical(regtype, "ll")) {
    out <- npreghat(
      bws = .np_indexhat_ll_owner_rbw(bws = bws, idx.train = idx.train),
      txdat = idx.train,
      exdat = idx.eval,
      y = y,
      output = output,
      s = s
    )

    if (!identical(output, "matrix"))
      return(out)

    return(matrix(
      as.double(out),
      nrow = nrow(out),
      ncol = ncol(out),
      dimnames = dimnames(out)
    ))
  }

  if (identical(regtype.engine, "lc")) {
    if (s == 1L) {
      return(.np_indexhat_lc_derivative(
        bws = bws,
        idx.train = idx.train,
        idx.eval = idx.eval,
        y = y,
        output = output
      ))
    }

    return(.np_indexhat_lc_mean(
      bws = bws,
      idx.train = idx.train,
      idx.eval = idx.eval,
      y = y,
      output = output
    ))
  }

  rbw <- .np_indexhat_rbw(bws = bws, idx.train = idx.train)

  lp.mean.owner.safe <- identical(regtype.engine, "lp") &&
    identical(output, "matrix") &&
    s == 0L &&
    !identical(bws$type, "fixed") &&
    !(
      identical(bws$type, "generalized_nn") &&
        all(as.integer(spec$degree.engine) == 1L) &&
        (
          !identical(spec$basis.engine, "glp") ||
            isTRUE(spec$bernstein.basis.engine)
        )
    )

  if (lp.mean.owner.safe) {
    return(.np_indexhat_lp_mean_matrix(
      bws = bws,
      idx.train = idx.train,
      idx.eval = idx.eval
    ))
  }

  lp.grad.owner.safe <- identical(output, "matrix") &&
    s == 1L &&
    (
      identical(regtype, "ll") ||
        (
          identical(regtype.engine, "lp") &&
            !(
              identical(bws$type, "generalized_nn") &&
                all(as.integer(spec$degree.engine) == 1L) &&
                (
                  !identical(spec$basis.engine, "glp") ||
                    isTRUE(spec$bernstein.basis.engine)
                )
            )
        )
    )

  if (lp.grad.owner.safe) {
    return(.np_indexhat_gradient_matrix(
      bws = bws,
      idx.train = idx.train,
      idx.eval = idx.eval
    ))
  }

  if (s == 1L) {
    fit_one <- function(ycol) {
      fit <- .np_regression_direct(
        bws = rbw,
        txdat = idx.train,
        tydat = ycol,
        exdat = idx.eval,
        gradients = TRUE,
        gradient.order = 1L
      )
      fit$grad[, 1L]
    }

    if (identical(output, "apply")) {
      if (is.null(y))
        stop("argument 'y' is required when output='apply'")

      y <- .np_indexhat_numeric_y(y)
      if (nrow(y) != nrow(idx.train))
        stop("number of rows in 'y' must equal number of training rows")

      out <- vapply(
        seq_len(ncol(y)),
        function(j) fit_one(y[, j]),
        numeric(nrow(idx.eval))
      )

      if (is.null(dim(out)))
        return(as.vector(out))
      if (ncol(out) == 1L)
        return(as.vector(out[, 1L]))
      return(out)
    }

    neval <- nrow(idx.eval)
    ntrain <- nrow(idx.train)
    H <- matrix(NA_real_, nrow = neval, ncol = ntrain)
    for (j in seq_len(ntrain)) {
      yj <- numeric(ntrain)
      yj[j] <- 1.0
      H[, j] <- fit_one(yj)
    }
    return(H)
  }

  if (identical(output, "apply")) {
    if (is.null(y))
      stop("argument 'y' is required when output='apply'")

    fit <- .np_regression_direct(
      bws = rbw,
      txdat = idx.train,
      tydat = y,
      exdat = idx.eval,
      gradients = FALSE
    )
    return(fit$mean)
  }

  fit_one <- function(ycol) {
    fit <- .np_regression_direct(
      bws = rbw,
      txdat = idx.train,
      tydat = ycol,
      exdat = idx.eval,
      gradients = FALSE
    )
    fit$mean
  }

  neval <- nrow(idx.eval)
  ntrain <- nrow(idx.train)
  H <- matrix(NA_real_, nrow = neval, ncol = ntrain)
  for (j in seq_len(ntrain)) {
    yj <- numeric(ntrain)
    yj[j] <- 1.0
    H[, j] <- fit_one(yj)
  }
  H
}

.np_indexhat_apply_exact <- function(bws,
                                     idx.train,
                                     idx.eval,
                                     y,
                                     s = 0L) {
  .np_indexhat_exact(
    bws = bws,
    idx.train = idx.train,
    idx.eval = idx.eval,
    y = y,
    output = "apply",
    s = s
  )
}

.npscoef_make_regbw <- function(bws, zdat, bw = bws$bw) {
  .np_semihat_make_regbw_state(
    source = bws,
    xdat = zdat,
    ydat = rep.int(0.0, nrow(zdat)),
    bw = bw
  )
}

.npscoef_lp_state <- function(bws, tzdat, ezdat, leave.one.out = FALSE, where = "npscoef") {
  tzdat <- toFrame(tzdat)
  ezdat <- toFrame(ezdat)
  leave.one.out <- npValidateScalarLogical(leave.one.out, "leave.one.out")
  spec <- .npscoef_canonical_spec(source = bws, zdat = tzdat, where = where)
  same.eval <- isTRUE(all.equal(tzdat, ezdat, check.attributes = FALSE))

  if (leave.one.out && !same.eval) {
    stop("leave.one.out=TRUE requires evaluation 'z' data to match training 'z' data")
  }

  state <- list(
    spec = spec,
    leave.one.out = leave.one.out
  )

  if (identical(spec$regtype.engine, "lc")) {
    state$z.train <- adjustLevels(tzdat, bws$zdati)
    state$z.eval <- if (leave.one.out) state$z.train else adjustLevels(ezdat, bws$zdati, allowNewCells = TRUE)
    return(state)
  }

  rbw <- .npscoef_make_regbw(bws = bws, zdat = tzdat)
  z.train <- adjustLevels(tzdat, rbw$xdati)
  z.eval <- if (leave.one.out) z.train else adjustLevels(ezdat, rbw$xdati, allowNewCells = TRUE)
  W.train <- W.lp(
    xdat = z.train,
    degree = spec$degree.engine,
    basis = spec$basis.engine,
    bernstein.basis = spec$bernstein.basis.engine
  )
  W.eval <- W.lp(
    xdat = z.train,
    exdat = if (leave.one.out) NULL else z.eval,
    degree = spec$degree.engine,
    basis = spec$basis.engine,
    bernstein.basis = spec$bernstein.basis.engine
  )
  neval <- nrow(z.eval)
  if (!is.matrix(W.eval))
    W.eval <- matrix(W.eval, nrow = neval)
  if (nrow(W.eval) != neval)
    W.eval <- matrix(W.eval, nrow = neval, byrow = FALSE)

  state$rbw <- rbw
  state$z.train <- z.train
  state$z.eval <- z.eval
  state$W.train <- W.train
  state$W.eval <- W.eval
  state
}

.npscoef_row_tensor_design <- function(X, Z) {
  X <- as.matrix(X)
  Z <- as.matrix(Z)
  if (nrow(X) != nrow(Z))
    stop("tensor design inputs must have the same number of rows", call. = FALSE)
  do.call(cbind, lapply(seq_len(ncol(X)), function(j) Z * X[, j]))
}

.npscoef_lp1_largeh_global_fit <- function(bws,
                                           tzdat,
                                           ezdat,
                                           W.train,
                                           tydat,
                                           u2 = NULL,
                                           leave.one.out = FALSE,
                                           where = "npscoef",
                                           solver) {
  if (missing(solver) || !is.function(solver))
    stop("LP1 large-h global fit requires a solver function", call. = FALSE)

  lp_state <- .npscoef_lp_state(
    bws = bws,
    tzdat = tzdat,
    ezdat = ezdat,
    leave.one.out = leave.one.out,
    where = where
  )
  spec <- lp_state$spec
  if (!identical(spec$regtype.engine, "lp") ||
      !identical(spec$basis.engine, "glp") ||
      isTRUE(spec$bernstein.basis.engine) ||
      any(as.integer(spec$degree.engine) != 1L)) {
    stop("LP1 large-h global fit called outside canonical degree-1 raw glp regime",
         call. = FALSE)
  }

  tensor.train <- .npscoef_row_tensor_design(W.train, lp_state$W.train)
  z.eval.one <- lp_state$z.eval[1L, , drop = FALSE]

  ytensor <- cbind(tydat, tensor.train)
  main.ks <- npksum(
    txdat = lp_state$z.train,
    tydat = ytensor,
    weights = ytensor,
    exdat = z.eval.one,
    bws = lp_state$rbw,
    bandwidth.divide = TRUE
  )$ksum
  tyw <- as.double(main.ks[-1L, 1L, 1L])
  tww <- main.ks[-1L, -1L, 1L, drop = TRUE]

  solve.out <- solver(tyw, tww)
  theta <- as.double(solve.out$coef)
  theta.mat <- matrix(theta,
                      nrow = ncol(lp_state$W.eval),
                      ncol = ncol(W.train))
  coef <- t(lp_state$W.eval %*% theta.mat)

  out <- list(
    lp_state = lp_state,
    coef = coef,
    theta = theta,
    ridge = solve.out$ridge,
    tww = tww
  )

  if (!is.null(u2)) {
    out$s <- npksum(
      txdat = lp_state$z.train,
      tydat = tensor.train,
      weights = tensor.train * as.double(u2),
      exdat = z.eval.one,
      bws = lp_state$rbw,
      bandwidth.divide = TRUE,
      kernel.pow = 2
    )$ksum[, , 1L, drop = TRUE]
  } else {
    out$s <- NULL
  }

  out
}

.npscoef_effective_weight_state <- function(bws, tzdat, ezdat, leave.one.out = FALSE) {
  state <- .npscoef_lp_state(
    bws = bws,
    tzdat = tzdat,
    ezdat = ezdat,
    leave.one.out = leave.one.out,
    where = "npscoefhat"
  )
  state$bws <- bws
  state$regtype <- state$spec$regtype.engine
  state
}

.npscoef_effective_weight_one <- function(state, eval.index) {
  .npscoef_effective_weight_chunk(state = state, eval.indices = eval.index)[, 1L]
}

.npscoef_effective_weight_chunk_size <- function(ntrain, neval) {
  chunk.opt <- getOption("np.scoef.weight.chunk.size")
  if (!is.null(chunk.opt)) {
    chunk.opt <- as.integer(chunk.opt)[1L]
    if (is.na(chunk.opt) || chunk.opt < 1L)
      stop("option 'np.scoef.weight.chunk.size' must be a positive integer")
    return(min(as.integer(neval), chunk.opt))
  }

  target.bytes <- 32L * 1024L * 1024L
  chunk <- as.integer(floor(target.bytes / (8 * max(1L, as.integer(ntrain)))))
  if (!is.finite(chunk) || is.na(chunk) || chunk < 1L)
    chunk <- 1L
  min(as.integer(neval), max(1L, chunk))
}

.npscoef_effective_weight_chunk <- function(state, eval.indices) {
  eval.indices <- as.integer(eval.indices)
  if (!length(eval.indices))
    return(matrix(0.0, nrow = nrow(state$z.train), ncol = 0L))
  if (anyNA(eval.indices) || any(!is.finite(eval.indices)) ||
      any(eval.indices < 1L) || any(eval.indices > nrow(state$z.eval))) {
    stop("invalid evaluation index for smooth-coefficient effective weights", call. = FALSE)
  }

  eval.rows <- state$z.eval[eval.indices, , drop = FALSE]
  bw.use <- if (identical(state$regtype, "lc")) state$bws else state$rbw
  kw <- .np_kernel_weights_direct(
    txdat = state$z.train,
    exdat = eval.rows,
    bws = bw.use,
    bandwidth.divide = TRUE,
    kernel.pow = 1.0
  )

  if (isTRUE(state$leave.one.out)) {
    for (jj in seq_along(eval.indices))
      kw[eval.indices[[jj]], jj] <- 0.0
  }

  if (identical(state$regtype, "lc"))
    return(kw)

  out <- matrix(0.0, nrow = nrow(kw), ncol = ncol(kw))
  for (jj in seq_along(eval.indices)) {
    ii <- eval.indices[[jj]]
    solve.out <- .npreghat_solve_eval(
      W = state$W.train,
      w.eval = state$W.eval[ii, ],
      k = kw[, jj],
      ridge.base = 0.0
    )
    if (is.null(solve.out)) {
      stop(sprintf("failed to solve smooth-coefficient effective-weight system at evaluation row %d", ii))
    }
    out[, jj] <- kw[, jj] * drop(state$W.train %*% solve.out$v)
  }

  out
}

.npscoef_weight_matrix <- function(bws, tzdat, ezdat, leave.one.out = FALSE) {
  state <- .npscoef_effective_weight_state(
    bws = bws,
    tzdat = tzdat,
    ezdat = ezdat,
    leave.one.out = leave.one.out
  )
  neval <- nrow(state$z.eval)
  ntrain <- nrow(state$z.train)
  H <- matrix(0.0, nrow = ntrain, ncol = neval)
  for (ii in seq_len(neval))
    H[, ii] <- .npscoef_effective_weight_one(state = state, eval.index = ii)
  H
}

npindexhat <-
  function(bws,
           txdat = stop("training data 'txdat' missing"),
           exdat = txdat,
           y = NULL,
           output = c("matrix", "apply", "constraint"),
           s = 0L,
           fd.step = NULL,
           ...){

    .np_reject_unused_dots(list(...), "npindexhat")
    output <- match.arg(output)
    constraint.output <- identical(output, "constraint")
    operator.output <- if (constraint.output) "matrix" else output
    .np_semihat_require_class(bws, "sibandwidth", "npindexhat")
    if (is.null(y) && constraint.output)
      stop("argument 'y' is required when output='constraint'")
    s <- as.integer(s)
    if (length(s) != 1L || is.na(s) || !(s %in% c(0L, 1L)))
      stop("argument 's' must be 0 (fit) or 1 (index derivative)")
    if (!is.null(fd.step)) {
      fd.step <- as.double(fd.step)
      if (length(fd.step) != 1L || is.na(fd.step) || !is.finite(fd.step) || fd.step <= 0)
        stop("argument 'fd.step' must be a positive finite scalar")
    }

    txdat <- toFrame(txdat)
    exdat <- toFrame(exdat)

    npKernelBoundsCheckEval(exdat, bws$xdati$icon, bws$ckerlb, bws$ckerub, argprefix = "cker")

    txdat <- adjustLevels(txdat, bws$xdati)
    exdat <- adjustLevels(exdat, bws$xdati, allowNewCells = TRUE)

    txm <- toMatrix(txdat)
    exm <- toMatrix(exdat)

    index.train <- as.vector(txm %*% bws$beta)
    index.eval <- as.vector(exm %*% bws$beta)

    idx.train <- data.frame(index = index.train)
    idx.eval <- data.frame(index = index.eval)
    spec <- .npindex_resolve_spec(bws, where = "npindexhat")

    if (identical(spec$regtype.engine, "lc")) {
      H <- .np_indexhat_exact(
        bws = bws,
        idx.train = idx.train,
        idx.eval = idx.eval,
        y = y,
        output = operator.output,
        s = s
      )
      if (constraint.output)
        return(.np_hat_constraint_from_matrix(H, y, "npindexhat"))
      return(H)
    }

    if (s == 1L) {
      H <- .np_indexhat_exact(
        bws = bws,
        idx.train = idx.train,
        idx.eval = idx.eval,
        y = y,
        output = operator.output,
        s = s
      )
      if (constraint.output)
        return(.np_hat_constraint_from_matrix(H, y, "npindexhat"))
      return(H)
    }

    if (identical(output, "apply")) {
      return(.np_indexhat_apply_exact(
        bws = bws,
        idx.train = idx.train,
        idx.eval = idx.eval,
        y = y,
        s = s
      ))
    }

    if (!identical(bws$type, "fixed")) {
      H <- .np_indexhat_exact(
        bws = bws,
        idx.train = idx.train,
        idx.eval = idx.eval,
        y = y,
        output = operator.output,
        s = s
      )
      if (constraint.output)
        return(.np_hat_constraint_from_matrix(H, y, "npindexhat"))
      return(H)
    }

    if (s == 0L) {
      H <- .np_indexhat_core(
        bws = bws,
        idx.train = idx.train,
        idx.eval = idx.eval,
        y = y,
        output = operator.output,
        ridge = 0.0
      )
      if (constraint.output)
        return(.np_hat_constraint_from_matrix(H, y, "npindexhat"))
      return(H)
    }

    step <- if (is.null(fd.step)) {
      span <- diff(range(index.train))
      span.scale <- if (is.finite(span) && span > 0) span else 1.0
      max(1.0e-6, 1.0e-5 * span.scale)
    } else {
      fd.step <- as.double(fd.step)
      if (length(fd.step) != 1L || is.na(fd.step) || !is.finite(fd.step) || fd.step <= 0)
        stop("argument 'fd.step' must be a positive finite scalar")
      fd.step
    }
    if (is.na(step) || !is.finite(step) || step <= 0)
      stop("argument 'fd.step' must be a positive finite scalar")

    if (identical(operator.output, "matrix")) {
      H.plus <- .np_indexhat_core(
        bws = bws,
        idx.train = idx.train,
        idx.eval = data.frame(index = index.eval + step),
        output = "matrix",
        ridge = 0.0
      )
      H.minus <- .np_indexhat_core(
        bws = bws,
        idx.train = idx.train,
        idx.eval = data.frame(index = index.eval - step),
        output = "matrix",
        ridge = 0.0
      )
      H <- (H.plus - H.minus) / (2.0 * step)
      if (constraint.output)
        return(.np_hat_constraint_from_matrix(H, y, "npindexhat"))
      return(H)
    }

    if (is.null(y))
      stop("argument 'y' is required when output='apply'")

    out.plus <- .np_indexhat_core(
      bws = bws,
      idx.train = idx.train,
      idx.eval = data.frame(index = index.eval + step),
      y = y,
      output = "apply",
      ridge = 0.0
    )
    out.minus <- .np_indexhat_core(
      bws = bws,
      idx.train = idx.train,
      idx.eval = data.frame(index = index.eval - step),
      y = y,
      output = "apply",
      ridge = 0.0
    )
    (out.plus - out.minus) / (2.0 * step)
  }

npplreghat <-
  function(bws,
           txdat = stop("training data 'txdat' missing"),
           tzdat = stop("training data 'tzdat' missing"),
           exdat = txdat,
           ezdat = tzdat,
           y = NULL,
           output = c("apply", "matrix", "constraint"),
           ...){

    .np_reject_unused_dots(list(...), "npplreghat")
    output <- match.arg(output)
    constraint.output <- identical(output, "constraint")
    matrix.output <- identical(output, "matrix") || constraint.output
    .np_semihat_require_class(bws, "plbandwidth", "npplreghat")
    txdat <- toFrame(txdat)
    tzdat <- toFrame(tzdat)
    exdat <- toFrame(exdat)
    ezdat <- toFrame(ezdat)

    if (nrow(txdat) != nrow(tzdat))
      stop("training data 'txdat' and 'tzdat' must have same number of rows")
    if (nrow(exdat) != nrow(ezdat))
      stop("evaluation data 'exdat' and 'ezdat' must have same number of rows")
    if (ncol(txdat) != ncol(exdat))
      stop("'txdat' and 'exdat' must have same number of columns")

    npKernelBoundsCheckEval(ezdat, bws$zdati$icon, bws$ckerlb, bws$ckerub, argprefix = "cker")

    txdat <- adjustLevels(txdat, bws$xdati)
    exdat <- adjustLevels(exdat, bws$xdati, allowNewCells = TRUE)
    tzdat <- adjustLevels(tzdat, bws$zdati)
    ezdat <- adjustLevels(ezdat, bws$zdati, allowNewCells = TRUE)

    if (is.null(y) && identical(output, "apply"))
      stop("argument 'y' is required when output='apply'")
    if (is.null(y) && constraint.output)
      stop("argument 'y' is required when output='constraint'")

    x.train.num <- matrix(0.0, nrow = nrow(txdat), ncol = ncol(txdat))
    x.eval.num <- matrix(0.0, nrow = nrow(exdat), ncol = ncol(txdat))

    for (j in seq_len(ncol(txdat))) {
      bw.xj <- bws$bw[[j + 1L]]
      if (is.factor(txdat[[j]])) {
        trj <- adjustLevels(txdat[, j, drop = FALSE], bw.xj$ydati)
        evj <- adjustLevels(exdat[, j, drop = FALSE], bw.xj$ydati, allowNewCells = TRUE)
        lev <- bws$bw[[j + 1L]]$ydati$all.dlev[[1L]]
        x.train.num[, j] <- lev[as.integer(trj[, 1L])]
        x.eval.num[, j] <- lev[as.integer(evj[, 1L])]
      } else {
        x.train.num[, j] <- as.double(txdat[[j]])
        x.eval.num[, j] <- as.double(exdat[[j]])
      }
    }

    n <- nrow(txdat)
    p <- ncol(txdat)
    m <- nrow(exdat)

    resx.train <- matrix(0.0, nrow = n, ncol = p)
    resx.eval <- matrix(0.0, nrow = m, ncol = p)

    if (matrix.output) {
      H.y.eval <- npreghat(
        bws = bws$bw$yzbw,
        txdat = tzdat,
        exdat = ezdat,
        output = "matrix"
      )

      for (j in seq_len(p)) {
        xhat.train <- as.vector(npreghat(
          bws = bws$bw[[j + 1L]],
          txdat = tzdat,
          y = x.train.num[, j],
          output = "apply"
        ))
        xhat.eval <- as.vector(npreghat(
          bws = bws$bw[[j + 1L]],
          txdat = tzdat,
          exdat = ezdat,
          y = x.train.num[, j],
          output = "apply"
        ))

        resx.train[, j] <- x.train.num[, j] - xhat.train
        resx.eval[, j] <- x.eval.num[, j] - xhat.eval
      }

      qrR <- qr(resx.train, tol = .Machine$double.eps)
      H.y.train <- npreghat(
        bws = bws$bw$yzbw,
        txdat = tzdat,
        output = "matrix"
      )
      H.y.train[] <- -H.y.train
      diag(H.y.train) <- diag(H.y.train) + 1.0
      A <- qr.coef(qrR, H.y.train)
      A[is.na(A)] <- 0.0
      H <- H.y.eval + resx.eval %*% A
      if (constraint.output)
        return(.np_hat_constraint_from_matrix(H, y, "npplreghat"))
      return(H)
    }

    yy <- as.matrix(y)
    if (nrow(yy) != n)
      stop("number of rows in 'y' must equal number of training rows")

    for (j in seq_len(p)) {
      xhat.train <- npreghat(
        bws = bws$bw[[j + 1L]],
        txdat = tzdat,
        y = x.train.num[, j],
        output = "apply"
      )
      xhat.eval <- npreghat(
        bws = bws$bw[[j + 1L]],
        txdat = tzdat,
        exdat = ezdat,
        y = x.train.num[, j],
        output = "apply"
      )
      resx.train[, j] <- x.train.num[, j] - as.vector(xhat.train)
      resx.eval[, j] <- x.eval.num[, j] - as.vector(xhat.eval)
    }

    qrR <- qr(resx.train, tol = .Machine$double.eps)

    Hy.eval <- npreghat(
      bws = bws$bw$yzbw,
      txdat = tzdat,
      exdat = ezdat,
      y = yy,
      output = "apply"
    )
    if (!is.matrix(Hy.eval))
      Hy.eval <- matrix(Hy.eval, ncol = ncol(yy))

    Hy.train <- npreghat(
      bws = bws$bw$yzbw,
      txdat = tzdat,
      y = yy,
      output = "apply"
    )
    if (!is.matrix(Hy.train))
      Hy.train <- matrix(Hy.train, ncol = ncol(yy))

    B <- qr.coef(qrR, yy - Hy.train)
    B[is.na(B)] <- 0.0

    out <- Hy.eval + resx.eval %*% B
    if (ncol(out) == 1L) as.vector(out) else out
  }

npscoefhat <-
  function(bws,
           txdat = stop("training data 'txdat' missing"),
           tzdat = NULL,
           exdat = txdat,
           ezdat = tzdat,
           y = NULL,
           output = c("matrix", "apply", "constraint"),
           ridge = 0.0,
           iterate = FALSE,
           leave.one.out = FALSE,
           ...){

    .np_reject_unused_dots(list(...), "npscoefhat")
    output <- match.arg(output)
    constraint.output <- identical(output, "constraint")
    .np_semihat_require_class(bws, "scbandwidth", "npscoefhat")
    iterate <- npValidateScalarLogical(iterate, "iterate")
    leave.one.out <- npValidateScalarLogical(leave.one.out, "leave.one.out")
    ridge <- as.double(ridge)
    if (length(ridge) != 1L || is.na(ridge) || !is.finite(ridge) || ridge < 0)
      stop("argument 'ridge' must be a non-negative finite scalar")
    if (iterate)
      stop("iterate=TRUE is not supported in npscoefhat")

    miss.z <- missing(tzdat) || is.null(tzdat)

    txdat <- toFrame(txdat)
    if (miss.z)
      tzdat <- txdat
    else
      tzdat <- toFrame(tzdat)

    exdat <- toFrame(exdat)
    if (missing(ezdat) || is.null(ezdat))
      ezdat <- if (miss.z) exdat else tzdat
    else
      ezdat <- toFrame(ezdat)

    if (nrow(txdat) != nrow(tzdat))
      stop("'txdat' and 'tzdat' must have same number of training rows")
    if (nrow(exdat) != nrow(ezdat))
      stop("'exdat' and 'ezdat' must have same number of evaluation rows")
    if (ncol(txdat) != ncol(exdat))
      stop("'txdat' and 'exdat' must have same number of columns")

    npKernelBoundsCheckEval(ezdat, bws$zdati$icon, bws$ckerlb, bws$ckerub, argprefix = "cker")

    txdat <- adjustLevels(txdat, bws$xdati)
    exdat <- adjustLevels(exdat, bws$xdati, allowNewCells = TRUE)
    if (!miss.z) {
      tzdat <- adjustLevels(tzdat, bws$zdati)
      ezdat <- adjustLevels(ezdat, bws$zdati, allowNewCells = TRUE)
    }

    X.train <- toMatrix(txdat)
    X.eval <- toMatrix(exdat)

    W.train <- cbind(1.0, X.train)
    W.eval <- cbind(1.0, X.eval)

    n <- nrow(W.train)
    m <- nrow(W.eval)
    spec <- .npscoef_canonical_spec(source = bws, zdat = tzdat, where = "npscoefhat")

    if (is.null(y) && constraint.output)
      stop("argument 'y' is required when output='constraint'")

    if (identical(output, "apply")) {
      if (is.null(y))
        stop("argument 'y' is required when output='apply'")
      yy <- as.matrix(y)
      if (nrow(yy) != n)
        stop("number of rows in 'y' must equal number of training rows")
      out <- matrix(0.0, nrow = m, ncol = ncol(yy))

      if (identical(spec$regtype.engine, "lc")) {
        state <- .npscoef_effective_weight_state(
          bws = bws,
          tzdat = tzdat,
          ezdat = ezdat,
          leave.one.out = leave.one.out
        )
        chunk.size <- .npscoef_effective_weight_chunk_size(ntrain = n, neval = m)

        for (start in seq.int(1L, m, by = chunk.size)) {
          stopi <- min(m, start + chunk.size - 1L)
          idx <- seq.int(start, stopi)
          kw.chunk <- .npscoef_effective_weight_chunk(state = state, eval.indices = idx)

          for (jj in seq_along(idx)) {
            ii <- idx[[jj]]
            solve.out <- .npreghat_solve_eval(
              W = W.train,
              w.eval = W.eval[ii, ],
              k = kw.chunk[, jj],
              ridge.base = ridge
            )
            if (is.null(solve.out))
              stop(sprintf("failed to solve local hat system at evaluation row %d", ii))
            h.row <- kw.chunk[, jj] * drop(W.train %*% solve.out$v)
            out[ii, ] <- drop(crossprod(h.row, yy))
          }
        }
      } else {
        lp_state <- .npscoef_lp_state(
          bws = bws,
          tzdat = tzdat,
          ezdat = ezdat,
          leave.one.out = leave.one.out,
          where = "npscoefhat"
        )
        tensor.train <- .npscoef_row_tensor_design(W.train, lp_state$W.train)
        tensor.eval <- .npscoef_row_tensor_design(W.eval, lp_state$W.eval)
        chunk.size <- .npscoef_effective_weight_chunk_size(ntrain = n, neval = m)

        for (start in seq.int(1L, m, by = chunk.size)) {
          stopi <- min(m, start + chunk.size - 1L)
          idx <- seq.int(start, stopi)
          kw.chunk <- .np_kernel_weights_direct(
            txdat = lp_state$z.train,
            exdat = lp_state$z.eval[idx, , drop = FALSE],
            bws = lp_state$rbw,
            bandwidth.divide = TRUE,
            kernel.pow = 1.0
          )

          if (leave.one.out) {
            for (jj in seq_along(idx))
              kw.chunk[idx[[jj]], jj] <- 0.0
          }

          for (jj in seq_along(idx)) {
            ii <- idx[[jj]]
            solve.out <- .npreghat_solve_eval(
              W = tensor.train,
              w.eval = tensor.eval[ii, ],
              k = kw.chunk[, jj],
              ridge.base = ridge
            )
            if (is.null(solve.out))
              stop(sprintf("failed to solve smooth-coefficient local hat system at evaluation row %d", ii))
            h.row <- kw.chunk[, jj] * drop(tensor.train %*% solve.out$v)
            out[ii, ] <- drop(crossprod(h.row, yy))
          }
        }
      }

      if (ncol(out) == 1L) return(as.vector(out))
      return(out)
    }

    H <- matrix(0.0, nrow = m, ncol = n)

    if (identical(spec$regtype.engine, "lc")) {
      state <- .npscoef_effective_weight_state(
        bws = bws,
        tzdat = tzdat,
        ezdat = ezdat,
        leave.one.out = leave.one.out
      )
      chunk.size <- .npscoef_effective_weight_chunk_size(ntrain = n, neval = m)

      for (start in seq.int(1L, m, by = chunk.size)) {
        stopi <- min(m, start + chunk.size - 1L)
        idx <- seq.int(start, stopi)
        kw.chunk <- .npscoef_effective_weight_chunk(state = state, eval.indices = idx)

        for (jj in seq_along(idx)) {
          ii <- idx[[jj]]
          solve.out <- .npreghat_solve_eval(
            W = W.train,
            w.eval = W.eval[ii, ],
            k = kw.chunk[, jj],
            ridge.base = ridge
          )
          if (is.null(solve.out))
            stop(sprintf("failed to solve local hat system at evaluation row %d", ii))
          H[ii, ] <- kw.chunk[, jj] * drop(W.train %*% solve.out$v)
        }
      }
    } else {
      lp_state <- .npscoef_lp_state(
        bws = bws,
        tzdat = tzdat,
        ezdat = ezdat,
        leave.one.out = leave.one.out,
        where = "npscoefhat"
      )
      tensor.train <- .npscoef_row_tensor_design(W.train, lp_state$W.train)
      tensor.eval <- .npscoef_row_tensor_design(W.eval, lp_state$W.eval)
      chunk.size <- .npscoef_effective_weight_chunk_size(ntrain = n, neval = m)

      for (start in seq.int(1L, m, by = chunk.size)) {
        stopi <- min(m, start + chunk.size - 1L)
        idx <- seq.int(start, stopi)
        kw.chunk <- .np_kernel_weights_direct(
          txdat = lp_state$z.train,
          exdat = lp_state$z.eval[idx, , drop = FALSE],
          bws = lp_state$rbw,
          bandwidth.divide = TRUE,
          kernel.pow = 1.0
        )

        if (leave.one.out) {
          for (jj in seq_along(idx))
            kw.chunk[idx[[jj]], jj] <- 0.0
        }

        for (jj in seq_along(idx)) {
          ii <- idx[[jj]]
          solve.out <- .npreghat_solve_eval(
            W = tensor.train,
            w.eval = tensor.eval[ii, ],
            k = kw.chunk[, jj],
            ridge.base = ridge
          )
          if (is.null(solve.out))
            stop(sprintf("failed to solve smooth-coefficient local hat system at evaluation row %d", ii))
          H[ii, ] <- kw.chunk[, jj] * drop(tensor.train %*% solve.out$v)
        }
      }
    }

    if (constraint.output)
      return(.np_hat_constraint_from_matrix(H, y, "npscoefhat"))

    H
  }

library(np)

test_that("npindexhat lp mean owner preserves fit and matrix/apply parity for nonfixed bwtypes", {
  set.seed(20260315)
  n <- 90L
  x1 <- runif(n, -1, 1)
  x2 <- rnorm(n)
  y <- sin(2 * (0.7 * x1 - 0.3 * x2)) + 0.35 * x1 * x2 + rnorm(n, sd = 0.04)

  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- data.frame(
    x1 = seq(min(x1) * 0.9, max(x1) * 0.9, length.out = 30L),
    x2 = seq(quantile(x2, 0.2), quantile(x2, 0.8), length.out = 30L)
  )

  cases <- list(
    list(label = "lp1 generalized canonical", bwtype = "generalized_nn", degree = 1L, basis = "glp", bern = FALSE, bws = c(1, 1, 9L)),
    list(label = "lp1 generalized tensor legacy", bwtype = "generalized_nn", degree = 1L, basis = "tensor", bern = FALSE, bws = c(1, 1, 9L)),
    list(label = "lp1 generalized bernstein legacy", bwtype = "generalized_nn", degree = 1L, basis = "glp", bern = TRUE, bws = c(1, 1, 9L)),
    list(label = "lp1 adaptive tensor", bwtype = "adaptive_nn", degree = 1L, basis = "tensor", bern = FALSE, bws = c(1, 1, 9L)),
    list(label = "lp2 generalized tensor", bwtype = "generalized_nn", degree = 2L, basis = "tensor", bern = FALSE, bws = c(1, 1, 11L)),
    list(label = "lp2 generalized bernstein", bwtype = "generalized_nn", degree = 2L, basis = "glp", bern = TRUE, bws = c(1, 1, 11L)),
    list(label = "lp2 adaptive tensor", bwtype = "adaptive_nn", degree = 2L, basis = "tensor", bern = FALSE, bws = c(1, 1, 11L))
  )

  for (cfg in cases) {
    bw <- npindexbw(
      xdat = tx,
      ydat = y,
      regtype = "lp",
      degree = cfg$degree,
      basis = cfg$basis,
      bernstein.basis = cfg$bern,
      bwtype = cfg$bwtype,
      bandwidth.compute = FALSE,
      bws = cfg$bws
    )

    fit.mean <- npindex(
      bws = bw,
      txdat = tx,
      tydat = y,
      exdat = ex,
      gradients = FALSE,
      errors = FALSE
    )

    apply.mean <- npindexhat(
      bws = bw,
      txdat = tx,
      exdat = ex,
      y = y,
      output = "apply",
      s = 0L
    )
    matrix.mean <- npindexhat(
      bws = bw,
      txdat = tx,
      exdat = ex,
      output = "matrix",
      s = 0L
    )

    expect_equal(as.vector(apply.mean), as.vector(fit.mean$mean), tolerance = 1e-8, info = paste("mean", cfg$label))
    expect_equal(as.vector(matrix.mean %*% y), as.vector(fit.mean$mean), tolerance = 1e-8, info = paste("mean matrix", cfg$label))
    expect_equal(as.vector(apply.mean), as.vector(matrix.mean %*% y), tolerance = 1e-10, info = paste("mean parity", cfg$label))
  }
})

test_that("npscoefhat reproduces npscoef fitted values and supports matrix RHS", {
  set.seed(2468)
  n <- 90
  x <- runif(n)
  z <- runif(n)
  y <- (x^2) * z + 0.3 * x + rnorm(n, sd = 0.05)

  bw <- npscoefbw(
    xdat = x,
    zdat = z,
    ydat = y,
    bws = 0.15,
    bandwidth.compute = FALSE
  )

  tx <- data.frame(x = x)
  tz <- data.frame(z = z)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 40))
  ez <- data.frame(z = seq(min(z), max(z), length.out = 40))

  fit.train <- npscoef(
    bws = bw,
    txdat = tx,
    tydat = y,
    tzdat = tz,
    iterate = FALSE
  )
  fit.eval <- npscoef(
    bws = bw,
    txdat = tx,
    tydat = y,
    tzdat = tz,
    exdat = ex,
    ezdat = ez,
    iterate = FALSE
  )

  H.train <- npscoefhat(
    bws = bw,
    txdat = tx,
    tzdat = tz,
    output = "matrix",
    iterate = FALSE
  )
  H.eval <- npscoefhat(
    bws = bw,
    txdat = tx,
    tzdat = tz,
    exdat = ex,
    ezdat = ez,
    output = "matrix",
    iterate = FALSE
  )

  expect_equal(as.vector(H.train %*% y), as.vector(fit.train$mean), tolerance = 1e-8)
  expect_equal(as.vector(H.eval %*% y), as.vector(fit.eval$mean), tolerance = 1e-8)

  ystar <- cbind(y, y + 0.1)
  hy <- npscoefhat(
    bws = bw,
    txdat = tx,
    tzdat = tz,
    exdat = ex,
    ezdat = ez,
    y = ystar,
    output = "apply",
    iterate = FALSE
  )
  expect_true(isTRUE(all.equal(
    hy,
    H.eval %*% ystar,
    tolerance = 1e-10,
    check.attributes = FALSE
  )))
})

test_that("npscoef and npscoefhat support ll/lp basis variants", {
  set.seed(2469)
  n <- 95
  x <- runif(n)
  z <- runif(n)
  y <- (0.4 + x) * sin(2 * pi * z) + rnorm(n, sd = 0.04)
  tx <- data.frame(x = x)
  tz <- data.frame(z = z)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 32))
  ez <- data.frame(z = seq(min(z), max(z), length.out = 32))

  cfgs <- list(
    list(regtype = "ll", basis = NULL, label = "ll"),
    list(regtype = "lp", basis = "glp", label = "lp-glp"),
    list(regtype = "lp", basis = "additive", label = "lp-additive"),
    list(regtype = "lp", basis = "tensor", label = "lp-tensor")
  )

  for (cfg in cfgs) {
    bw.args <- list(
      xdat = x,
      zdat = z,
      ydat = y,
      bws = 0.18,
      bandwidth.compute = FALSE,
      regtype = cfg$regtype
    )
    if (!is.null(cfg$basis)) {
      bw.args$basis <- cfg$basis
      bw.args$degree <- 2L
    }
    bw <- do.call(npscoefbw, bw.args)
    expect_identical(bw$regtype, cfg$regtype, info = cfg$label)
    if (!is.null(cfg$basis))
      expect_identical(bw$basis, cfg$basis, info = cfg$label)

    fit.eval <- npscoef(
      bws = bw,
      txdat = tx,
      tydat = y,
      tzdat = tz,
      exdat = ex,
      ezdat = ez,
      errors = FALSE,
      iterate = FALSE
    )
    H.eval <- npscoefhat(
      bws = bw,
      txdat = tx,
      tzdat = tz,
      exdat = ex,
      ezdat = ez,
      output = "matrix",
      iterate = FALSE
    )
    expect_equal(as.vector(H.eval %*% y), as.vector(fit.eval$mean), tolerance = 1e-8, info = cfg$label)
  }
})

test_that("npscoefhat leave.one.out honors lc and ll paths", {
  set.seed(97533)
  n <- 85
  x <- runif(n)
  z <- runif(n)
  y <- (0.5 + x) * cos(2 * pi * z) + rnorm(n, sd = 0.04)
  tx <- data.frame(x = x)
  tz <- data.frame(z = z)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 20))
  ez <- data.frame(z = seq(min(z), max(z), length.out = 20))

  bw.lc <- npscoefbw(
    xdat = x, zdat = z, ydat = y,
    bws = 0.16, regtype = "lc", bandwidth.compute = FALSE
  )
  H0.lc <- npscoefhat(
    bws = bw.lc, txdat = tx, tzdat = tz,
    output = "matrix", iterate = FALSE, leave.one.out = FALSE
  )
  H1.lc <- npscoefhat(
    bws = bw.lc, txdat = tx, tzdat = tz,
    output = "matrix", iterate = FALSE, leave.one.out = TRUE
  )
  expect_gt(max(abs(H0.lc - H1.lc)), 1e-8)

  bw.ll <- npscoefbw(
    xdat = x, zdat = z, ydat = y,
    bws = 0.16, regtype = "ll", bandwidth.compute = FALSE
  )
  H0.ll <- npscoefhat(
    bws = bw.ll, txdat = tx, tzdat = tz,
    output = "matrix", iterate = FALSE, leave.one.out = FALSE
  )
  H1.ll <- npscoefhat(
    bws = bw.ll, txdat = tx, tzdat = tz,
    output = "matrix", iterate = FALSE, leave.one.out = TRUE
  )
  expect_gt(max(abs(H0.ll - H1.ll)), 1e-8)
  fit0.ll <- npscoef(
    bws = bw.ll, txdat = tx, tydat = y, tzdat = tz,
    iterate = FALSE, errors = FALSE, leave.one.out = FALSE
  )
  fit1.ll <- npscoef(
    bws = bw.ll, txdat = tx, tydat = y, tzdat = tz,
    iterate = FALSE, errors = FALSE, leave.one.out = TRUE
  )
  expect_gt(max(abs(fit0.ll$mean - fit1.ll$mean)), 1e-8)

  expect_error(
    npscoefhat(
      bws = bw.ll,
      txdat = tx, tzdat = tz,
      exdat = ex, ezdat = ez,
      output = "matrix",
      iterate = FALSE,
      leave.one.out = TRUE
    ),
    "requires evaluation 'z' data to match training 'z' data"
  )
})

test_that("npplreghat reproduces npplreg fitted values and supports matrix RHS", {
  set.seed(97531)
  n <- 120
  x <- runif(n)
  z <- runif(n)
  y <- sin(z) + 2.0 * x + rnorm(n, sd = 0.05)

  bw <- npplregbw(
    xdat = x,
    zdat = z,
    ydat = y,
    bws = matrix(c(0.2, 0.2), nrow = 2, ncol = 1),
    bandwidth.compute = FALSE
  )

  tx <- data.frame(x = x)
  tz <- data.frame(z = z)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 35))
  ez <- data.frame(z = seq(min(z), max(z), length.out = 35))

  fit.train <- npplreg(
    bws = bw,
    txdat = tx,
    tydat = y,
    tzdat = tz
  )
  fit.eval <- npplreg(
    bws = bw,
    txdat = tx,
    tydat = y,
    tzdat = tz,
    exdat = ex,
    ezdat = ez
  )

  H.train <- npplreghat(
    bws = bw,
    txdat = tx,
    tzdat = tz,
    output = "matrix"
  )
  H.eval <- npplreghat(
    bws = bw,
    txdat = tx,
    tzdat = tz,
    exdat = ex,
    ezdat = ez,
    output = "matrix"
  )

  expect_equal(as.vector(H.train %*% y), as.vector(fit.train$mean), tolerance = 1e-8)
  expect_equal(as.vector(H.eval %*% y), as.vector(fit.eval$mean), tolerance = 1e-8)

  ystar <- cbind(y, y + 0.05)
  hy <- npplreghat(
    bws = bw,
    txdat = tx,
    tzdat = tz,
    exdat = ex,
    ezdat = ez,
    y = ystar,
    output = "apply"
  )
  expect_equal(hy, H.eval %*% ystar, tolerance = 1e-10)
})

test_that("npplreghat supports ll/lp with lp basis variants", {
  set.seed(97532)
  n <- 90
  x <- runif(n)
  z <- runif(n)
  y <- sin(2 * z) + 1.25 * x + rnorm(n, sd = 0.04)
  tx <- data.frame(x = x)
  tz <- data.frame(z = z)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 27))
  ez <- data.frame(z = seq(min(z), max(z), length.out = 27))

  cfgs <- list(
    list(regtype = "ll", basis = NULL, label = "ll"),
    list(regtype = "lp", basis = "glp", label = "lp-glp"),
    list(regtype = "lp", basis = "additive", label = "lp-additive"),
    list(regtype = "lp", basis = "tensor", label = "lp-tensor")
  )

  for (cfg in cfgs) {
    bw.args <- list(
      xdat = x,
      zdat = z,
      ydat = y,
      bws = matrix(c(0.22, 0.22), nrow = 2, ncol = 1),
      bandwidth.compute = FALSE,
      regtype = cfg$regtype
    )
    if (!is.null(cfg$basis)) {
      bw.args$basis <- cfg$basis
      bw.args$degree <- 2L
    }
    bw <- do.call(npplregbw, bw.args)
    if (!is.null(cfg$basis))
      expect_identical(bw$basis, cfg$basis, info = cfg$label)

    fit.eval <- npplreg(
      bws = bw,
      txdat = tx,
      tydat = y,
      tzdat = tz,
      exdat = ex,
      ezdat = ez
    )
    H.eval <- npplreghat(
      bws = bw,
      txdat = tx,
      tzdat = tz,
      exdat = ex,
      ezdat = ez,
      output = "matrix"
    )
    expect_equal(
      as.vector(H.eval %*% y),
      as.vector(fit.eval$mean),
      tolerance = 1e-8,
      info = cfg$label
    )
  }
})

test_that("npindexhat reproduces npindex fit and approximates gradient", {
  set.seed(314159)
  n <- 110
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(2 * (x1 + x2)) + rnorm(n, sd = 0.05)

  tx <- data.frame(x1 = x1, x2 = x2)
  bw <- npindexbw(xdat = tx, ydat = y, method = "ichimura", nmulti = 1)

  fit.mean <- npindex(
    bws = bw,
    txdat = tx,
    tydat = y,
    exdat = tx,
    gradients = FALSE
  )
  fit.grad <- npindex(
    bws = bw,
    txdat = tx,
    tydat = y,
    exdat = tx,
    gradients = TRUE
  )

  H0 <- npindexhat(
    bws = bw,
    txdat = tx,
    exdat = tx,
    s = 0L,
    output = "matrix"
  )
  H1 <- npindexhat(
    bws = bw,
    txdat = tx,
    exdat = tx,
    s = 1L,
    output = "matrix"
  )

  expect_equal(as.vector(H0 %*% y), as.vector(fit.mean$mean), tolerance = 1e-8)
  expect_equal(as.vector(H1 %*% y), as.vector(fit.grad$grad[, 1]), tolerance = 5e-3)

  ystar <- cbind(y, y - 0.05)
  hy <- npindexhat(
    bws = bw,
    txdat = tx,
    exdat = tx,
    y = ystar,
    s = 0L,
    output = "apply"
  )
  expect_true(isTRUE(all.equal(
    hy,
    H0 %*% ystar,
    tolerance = 1e-10,
    check.attributes = FALSE
  )))
})

test_that("npindex and npindexhat support ll/lp basis variants", {
  set.seed(314160)
  n <- 90
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(2 * (x1 + x2)) + rnorm(n, sd = 0.05)
  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- tx[seq_len(30), , drop = FALSE]

  cfgs <- list(
    list(regtype = "ll", basis = NULL, label = "ll"),
    list(regtype = "lp", basis = "glp", label = "lp-glp"),
    list(regtype = "lp", basis = "additive", label = "lp-additive"),
    list(regtype = "lp", basis = "tensor", label = "lp-tensor")
  )

  for (cfg in cfgs) {
    bw.args <- list(
      xdat = tx,
      ydat = y,
      bws = c(1, 1, 0.25),
      bandwidth.compute = FALSE,
      regtype = cfg$regtype
    )
    if (!is.null(cfg$basis)) {
      bw.args$basis <- cfg$basis
      bw.args$degree <- 2L
    }
    bw <- do.call(npindexbw, bw.args)
    expect_identical(bw$regtype, cfg$regtype, info = cfg$label)
    if (!is.null(cfg$basis))
      expect_identical(bw$basis, cfg$basis, info = cfg$label)

    fit <- npindex(
      bws = bw,
      txdat = tx,
      tydat = y,
      exdat = ex,
      gradients = FALSE
    )
    H <- npindexhat(
      bws = bw,
      txdat = tx,
      exdat = ex,
      s = 0L,
      output = "matrix"
    )
    expect_equal(as.vector(H %*% y), as.vector(fit$mean), tolerance = 1e-8, info = cfg$label)
  }
})

test_that("npindex and npindexhat preserve nearest-neighbor bwtype semantics", {
  set.seed(314162)
  n <- 80
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(2 * (x1 + x2)) + rnorm(n, sd = 0.05)
  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- tx[seq_len(25), , drop = FALSE]

  cfgs <- list(
    list(regtype = "lc", basis = NULL, degree = NULL, h = 5L),
    list(regtype = "ll", basis = NULL, degree = NULL, h = 5L),
    list(regtype = "lp", basis = "tensor", degree = 2L, h = 5L)
  )

  for (bt in c("generalized_nn", "adaptive_nn")) {
    for (cfg in cfgs) {
      bw.args <- list(
        xdat = tx,
        ydat = y,
        bws = c(1, 1, cfg$h),
        bandwidth.compute = FALSE,
        regtype = cfg$regtype,
        bwtype = bt
      )
      if (!is.null(cfg$basis)) {
        bw.args$basis <- cfg$basis
        bw.args$degree <- cfg$degree
      }
      bw <- do.call(npindexbw, bw.args)

      fit <- npindex(
        bws = bw,
        txdat = tx,
        tydat = y,
        exdat = ex,
        gradients = FALSE
      )
      H <- npindexhat(
        bws = bw,
        txdat = tx,
        exdat = ex,
        s = 0L,
        output = "matrix"
      )
      hy <- npindexhat(
        bws = bw,
        txdat = tx,
        exdat = ex,
        y = y,
        s = 0L,
        output = "apply"
      )

      expect_equal(as.vector(H %*% y), as.vector(fit$mean), tolerance = 1e-8, info = paste(bt, cfg$regtype))
      expect_equal(as.vector(hy), as.vector(fit$mean), tolerance = 1e-8, info = paste(bt, cfg$regtype))
    }
  }
})

test_that("npindexhat exact apply matches npindex on resampled nearest-neighbor lp fits", {
  set.seed(3292)
  n <- 45
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.08)
  tx <- data.frame(x1 = x1, x2 = x2)
  counts <- rmultinom(n = 3L, size = n, prob = rep.int(1 / n, n))

  for (bt in c("generalized_nn", "adaptive_nn")) {
    bw <- npindexbw(
      xdat = tx,
      ydat = y,
      bws = c(1, 1, 5L),
      bandwidth.compute = FALSE,
      regtype = "lp",
      basis = "tensor",
      degree = 2L,
      bwtype = bt
    )

    for (b in seq_len(ncol(counts))) {
      idx <- rep.int(seq_len(n), counts[, b])
      fit <- npindex(
        bws = bw,
        txdat = tx[idx, , drop = FALSE],
        tydat = y[idx],
        exdat = tx,
        gradients = FALSE
      )
      hy <- npindexhat(
        bws = bw,
        txdat = tx[idx, , drop = FALSE],
        exdat = tx,
        y = y[idx],
        output = "apply",
        s = 0L
      )

      expect_equal(as.vector(hy), as.vector(fit$mean), tolerance = 1e-8, info = paste(bt, "apply/public", b))
    }
  }
})

test_that("semihat apply mode matches core fits across bwtypes", {
  set.seed(20260308)
  n <- 70
  x <- runif(n)
  x2 <- runif(n)
  z <- runif(n)
  y <- (0.4 + x) * sin(2 * pi * z) + 0.3 * x2 + rnorm(n, sd = 0.04)

  tx.sc <- data.frame(x = x)
  tz.sc <- data.frame(z = z)
  ex.sc <- data.frame(x = seq(0.1, 0.9, length.out = 18))
  ez.sc <- data.frame(z = seq(0.1, 0.9, length.out = 18))

  tx.pl <- data.frame(x = x)
  tz.pl <- data.frame(z = z)
  ex.pl <- data.frame(x = seq(0.1, 0.9, length.out = 18))
  ez.pl <- data.frame(z = seq(0.1, 0.9, length.out = 18))

  tx.si <- data.frame(x1 = x, x2 = x2)
  ex.si <- tx.si[seq_len(18), , drop = FALSE]

  for (regtype in c("lc", "ll", "lp")) {
    for (bwtype in c("fixed", "generalized_nn", "adaptive_nn")) {
      sc.args <- list(
        xdat = tx.sc,
        zdat = tz.sc,
        ydat = y,
        regtype = regtype,
        bwtype = bwtype,
        bandwidth.compute = FALSE,
        bws = if (identical(bwtype, "fixed")) 0.18 else 9
      )
      pl.args <- list(
        xdat = tx.pl,
        zdat = tz.pl,
        ydat = y,
        regtype = regtype,
        bwtype = bwtype,
        bandwidth.compute = FALSE,
        bws = matrix(rep(if (identical(bwtype, "fixed")) 0.18 else 9, 2L), nrow = 2L, ncol = 1L)
      )
      si.args <- list(
        xdat = tx.si,
        ydat = y,
        regtype = regtype,
        bwtype = bwtype,
        bandwidth.compute = FALSE,
        bws = c(1, 1, if (identical(bwtype, "fixed")) 0.18 else 9)
      )
      if (identical(regtype, "lp")) {
        sc.args$degree <- 1L
        sc.args$basis <- "glp"
        sc.args$bernstein.basis <- FALSE
        pl.args$degree <- 1L
        pl.args$basis <- "glp"
        pl.args$bernstein.basis <- FALSE
        si.args$degree <- 1L
        si.args$basis <- "glp"
        si.args$bernstein.basis <- FALSE
      }

      sc.bw <- do.call(npscoefbw, sc.args)
      sc.fit <- npscoef(
        bws = sc.bw,
        txdat = tx.sc,
        tzdat = tz.sc,
        tydat = y,
        exdat = ex.sc,
        ezdat = ez.sc,
        iterate = FALSE,
        errors = FALSE
      )
      sc.apply <- npscoefhat(
        bws = sc.bw,
        txdat = tx.sc,
        tzdat = tz.sc,
        exdat = ex.sc,
        ezdat = ez.sc,
        y = y,
        output = "apply",
        iterate = FALSE
      )
      sc.H <- npscoefhat(
        bws = sc.bw,
        txdat = tx.sc,
        tzdat = tz.sc,
        exdat = ex.sc,
        ezdat = ez.sc,
        output = "matrix",
        iterate = FALSE
      )
      sc.constraint <- npscoefhat(
        bws = sc.bw,
        txdat = tx.sc,
        tzdat = tz.sc,
        exdat = ex.sc,
        ezdat = ez.sc,
        y = y,
        output = "constraint",
        iterate = FALSE
      )
      expect_equal(
        as.vector(sc.apply),
        as.vector(sc.H %*% y),
        tolerance = 1e-10,
        info = paste("npscoefhat helper parity", regtype, bwtype)
      )
      expect_equal(
        sc.constraint,
        t(sc.H) * y,
        tolerance = 0,
        ignore_attr = TRUE,
        info = paste("npscoefhat constraint", regtype, bwtype)
      )
      if (identical(bwtype, "fixed")) {
        expect_equal(
          as.vector(sc.apply),
          as.vector(sc.fit$mean),
          tolerance = 1e-8,
          info = paste("npscoefhat core parity", regtype, bwtype)
        )
      }

      pl.bw <- do.call(npplregbw, pl.args)
      pl.fit <- npplreg(
        bws = pl.bw,
        txdat = tx.pl,
        tzdat = tz.pl,
        tydat = y,
        exdat = ex.pl,
        ezdat = ez.pl
      )
      pl.apply <- npplreghat(
        bws = pl.bw,
        txdat = tx.pl,
        tzdat = tz.pl,
        exdat = ex.pl,
        ezdat = ez.pl,
        y = y,
        output = "apply"
      )
      pl.H <- npplreghat(
        bws = pl.bw,
        txdat = tx.pl,
        tzdat = tz.pl,
        exdat = ex.pl,
        ezdat = ez.pl,
        output = "matrix"
      )
      pl.constraint <- npplreghat(
        bws = pl.bw,
        txdat = tx.pl,
        tzdat = tz.pl,
        exdat = ex.pl,
        ezdat = ez.pl,
        y = y,
        output = "constraint"
      )
      expect_equal(
        as.vector(pl.apply),
        as.vector(pl.fit$mean),
        tolerance = 1e-8,
        info = paste("npplreghat", regtype, bwtype)
      )
      expect_equal(
        pl.constraint,
        t(pl.H) * y,
        tolerance = 0,
        ignore_attr = TRUE,
        info = paste("npplreghat constraint", regtype, bwtype)
      )

      si.bw <- do.call(npindexbw, si.args)
      si.fit.mean <- npindex(
        bws = si.bw,
        txdat = tx.si,
        tydat = y,
        exdat = ex.si,
        gradients = FALSE
      )
      si.fit.grad <- npindex(
        bws = si.bw,
        txdat = tx.si,
        tydat = y,
        exdat = ex.si,
        gradients = TRUE
      )
      si.apply.mean <- npindexhat(
        bws = si.bw,
        txdat = tx.si,
        exdat = ex.si,
        y = y,
        output = "apply",
        s = 0L
      )
      si.H.mean <- npindexhat(
        bws = si.bw,
        txdat = tx.si,
        exdat = ex.si,
        output = "matrix",
        s = 0L
      )
      si.constraint.mean <- npindexhat(
        bws = si.bw,
        txdat = tx.si,
        exdat = ex.si,
        y = y,
        output = "constraint",
        s = 0L
      )
      si.apply.grad <- npindexhat(
        bws = si.bw,
        txdat = tx.si,
        exdat = ex.si,
        y = y,
        output = "apply",
        s = 1L
      )
      si.H.grad <- npindexhat(
        bws = si.bw,
        txdat = tx.si,
        exdat = ex.si,
        output = "matrix",
        s = 1L
      )
      si.constraint.grad <- npindexhat(
        bws = si.bw,
        txdat = tx.si,
        exdat = ex.si,
        y = y,
        output = "constraint",
        s = 1L
      )
      expect_equal(
        as.vector(si.apply.mean),
        as.vector(si.fit.mean$mean),
        tolerance = 1e-8,
        info = paste("npindexhat mean", regtype, bwtype)
      )
      expect_equal(
        as.vector(si.H.mean %*% y),
        as.vector(si.fit.mean$mean),
        tolerance = 1e-8,
        info = paste("npindexhat mean matrix", regtype, bwtype)
      )
      expect_equal(
        as.vector(si.H.mean %*% y),
        as.vector(si.apply.mean),
        tolerance = 1e-10,
        info = paste("npindexhat mean matrix/apply", regtype, bwtype)
      )
      expect_equal(
        si.constraint.mean,
        t(si.H.mean) * y,
        tolerance = 0,
        ignore_attr = TRUE,
        info = paste("npindexhat mean constraint", regtype, bwtype)
      )
      expect_equal(
        as.vector(si.apply.grad),
        as.vector(si.fit.grad$grad[, 1]),
        tolerance = 1e-8,
        info = paste("npindexhat grad", regtype, bwtype)
      )
      expect_equal(
        as.vector(si.H.grad %*% y),
        as.vector(si.fit.grad$grad[, 1]),
        tolerance = 1e-8,
        info = paste("npindexhat grad matrix", regtype, bwtype)
      )
      expect_equal(
        as.vector(si.H.grad %*% y),
        as.vector(si.apply.grad),
        tolerance = 1e-10,
        info = paste("npindexhat grad matrix/apply", regtype, bwtype)
      )
      expect_equal(
        si.constraint.grad,
        t(si.H.grad) * y,
        tolerance = 0,
        ignore_attr = TRUE,
        info = paste("npindexhat grad constraint", regtype, bwtype)
      )
    }
  }
})

test_that("npscoefbw adaptive-nn manual bandwidth must be integer support", {
  set.seed(20260309)
  n <- 40L
  x <- runif(n)
  z <- runif(n)
  y <- (0.4 + x) * sin(2 * pi * z) + rnorm(n, sd = 0.04)
  tx <- data.frame(x = x)
  tz <- data.frame(z = z)

  expect_error(
    npscoefbw(
      xdat = tx,
      zdat = tz,
      ydat = y,
      regtype = "lc",
      bwtype = "adaptive_nn",
      bandwidth.compute = FALSE,
      bws = 0.13
    ),
    "nearest-neighbor bandwidth must be an integer"
  )
})

test_that("npscoefhat selected adaptive-nn owner preserves integer support and helper parity", {
  set.seed(20260308)
  n <- 50L
  x <- runif(n)
  z <- runif(n)
  y <- (0.4 + x) * sin(2 * pi * z) + rnorm(n, sd = 0.04)

  tx <- data.frame(x = x)
  tz <- data.frame(z = z)
  ex <- data.frame(x = seq(0.1, 0.9, length.out = 12L))
  ez <- data.frame(z = seq(0.1, 0.9, length.out = 12L))

  tol <- sqrt(.Machine$double.eps)
  upper <- n - 1L
  round_half_to_even <- getFromNamespace(".np_round_half_to_even", "np")

  for (regtype in c("lc", "ll", "lp")) {
    bw_args <- list(
      xdat = tx,
      zdat = tz,
      ydat = y,
      regtype = regtype,
      bwtype = "adaptive_nn",
      nmulti = 1L
    )
    if (identical(regtype, "lp")) {
      bw_args$degree <- 1L
      bw_args$basis <- "glp"
      bw_args$bernstein.basis <- FALSE
    }

    bw <- do.call(npscoefbw, bw_args)
    expect_true(all(abs(bw$bw - round_half_to_even(bw$bw)) <= tol), info = regtype)
    expect_true(all(bw$bw >= 1 & bw$bw <= upper), info = regtype)

    hat.apply <- npscoefhat(
      bws = bw,
      txdat = tx,
      tzdat = tz,
      exdat = ex,
      ezdat = ez,
      y = y,
      output = "apply",
      iterate = FALSE
    )
    hat.matrix <- npscoefhat(
      bws = bw,
      txdat = tx,
      tzdat = tz,
      exdat = ex,
      ezdat = ez,
      output = "matrix",
      iterate = FALSE
    )

    expect_equal(as.vector(hat.apply), as.vector(hat.matrix %*% y), tolerance = 1e-10, info = regtype)
  }
})

test_that("npindexbw lc selection no longer collapses nearest-neighbor bwtypes to fixed", {
  set.seed(314163)
  n <- 70
  x1 <- rnorm(n)
  x2 <- rnorm(n)
  y <- x1 - x2 + rnorm(n, sd = 0.2)
  tx <- data.frame(x1 = x1, x2 = x2)

  bw.fixed <- npindexbw(xdat = tx, ydat = y, regtype = "lc", bwtype = "fixed", nmulti = 1)
  bw.gen <- npindexbw(xdat = tx, ydat = y, regtype = "lc", bwtype = "generalized_nn", nmulti = 1)
  bw.adap <- npindexbw(xdat = tx, ydat = y, regtype = "lc", bwtype = "adaptive_nn", nmulti = 1)

  fit.fixed <- npindex(bws = bw.fixed, txdat = tx, tydat = y)$mean
  fit.gen <- npindex(bws = bw.gen, txdat = tx, tydat = y)$mean
  fit.adap <- npindex(bws = bw.adap, txdat = tx, tydat = y)$mean

  expect_gt(max(abs(fit.fixed - fit.gen)), 1e-6)
  expect_gt(max(abs(fit.fixed - fit.adap)), 1e-6)
})

test_that("semihat helper routes preserve LP and kernel-bound option contracts", {
  set.seed(314161)
  n <- 75
  x1 <- runif(n)
  x2 <- runif(n)
  z <- runif(n)
  y <- sin(2 * pi * z) + 0.8 * x1 - 0.3 * x2 + rnorm(n, sd = 0.04)

  tx <- data.frame(x1 = x1, x2 = x2)
  tz <- data.frame(z = z)
  ex <- tx[seq_len(25), , drop = FALSE]
  ez <- tz[seq_len(25), , drop = FALSE]

  sibw <- npindexbw(
    xdat = tx,
    ydat = y,
    bws = c(1, 1, 0.22),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "tensor",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerorder = 2L,
    ckerbound = "none"
  )
  expect_identical(sibw$regtype, "lp")
  expect_identical(sibw$basis, "tensor")
  expect_true(isTRUE(sibw$bernstein.basis))
  expect_identical(sibw$ckerbound, "none")

  si.fit <- npindex(
    bws = sibw,
    txdat = tx,
    tydat = y,
    exdat = ex,
    gradients = FALSE
  )
  expect_true(all(is.finite(si.fit$mean)))
  si.H <- npindexhat(
    bws = sibw,
    txdat = tx,
    exdat = ex,
    s = 0L,
    output = "matrix"
  )
  si.apply <- npindexhat(
    bws = sibw,
    txdat = tx,
    exdat = ex,
    y = y,
    s = 0L,
    output = "apply"
  )
  expect_equal(as.vector(si.H %*% y), as.vector(si.apply), tolerance = 1e-10)

  sibw.bad.bounds <- npindexbw(
    xdat = tx,
    ydat = y,
    bws = c(1, 1, 0.22),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "tensor",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerorder = 2L,
    ckerbound = "fixed",
    ckerlb = 0.0,
    ckerub = 1.0
  )
  expect_error(
    npindexhat(
      bws = sibw.bad.bounds,
      txdat = tx,
      exdat = ex,
      s = 0L,
      output = "matrix"
    ),
    "Invalid bounds for 'ckerbound'|Evaluation data violate 'ckerbound' bounds"
  )

  scbw <- npscoefbw(
    xdat = tx[, 1, drop = FALSE],
    zdat = tz,
    ydat = y,
    bws = 0.20,
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "tensor",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerorder = 2L,
    ckerbound = "fixed",
    ckerlb = 0.0,
    ckerub = 1.0
  )
  expect_identical(scbw$regtype, "lp")
  expect_identical(scbw$basis, "tensor")
  expect_true(isTRUE(scbw$bernstein.basis))
  expect_identical(scbw$ckerbound, "fixed")

  sc.fit <- npscoef(
    bws = scbw,
    txdat = tx[, 1, drop = FALSE],
    tydat = y,
    tzdat = tz,
    exdat = ex[, 1, drop = FALSE],
    ezdat = ez,
    iterate = FALSE
  )
  sc.H <- npscoefhat(
    bws = scbw,
    txdat = tx[, 1, drop = FALSE],
    tzdat = tz,
    exdat = ex[, 1, drop = FALSE],
    ezdat = ez,
    output = "matrix",
    iterate = FALSE
  )
  expect_equal(as.vector(sc.H %*% y), as.vector(sc.fit$mean), tolerance = 1e-8)

  plbw <- npplregbw(
    xdat = tx[, 1, drop = FALSE],
    zdat = tz,
    ydat = y,
    bws = matrix(c(0.20, 0.20), nrow = 2, ncol = 1),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "tensor",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerorder = 2L,
    ckerbound = "fixed",
    ckerlb = 0.0,
    ckerub = 1.0
  )
  expect_identical(plbw$regtype, "lp")
  expect_identical(plbw$basis, "tensor")
  expect_true(isTRUE(plbw$bernstein.basis))
  expect_identical(plbw$ckerbound, "fixed")

  pl.fit <- npplreg(
    bws = plbw,
    txdat = tx[, 1, drop = FALSE],
    tydat = y,
    tzdat = tz,
    exdat = ex[, 1, drop = FALSE],
    ezdat = ez
  )
  pl.H <- npplreghat(
    bws = plbw,
    txdat = tx[, 1, drop = FALSE],
    tzdat = tz,
    exdat = ex[, 1, drop = FALSE],
    ezdat = ez,
    output = "matrix"
  )
  expect_true(all(is.finite(pl.fit$mean)))
  pl.apply <- npplreghat(
    bws = plbw,
    txdat = tx[, 1, drop = FALSE],
    tzdat = tz,
    exdat = ex[, 1, drop = FALSE],
    ezdat = ez,
    y = y,
    output = "apply"
  )
  expect_equal(as.vector(pl.H %*% y), as.vector(pl.apply), tolerance = 1e-10)
})

test_that("semihat validates class and scalar controls", {
  set.seed(27182)
  n <- 40
  x <- runif(n)
  y <- rnorm(n)
  z <- runif(n)

  rbw <- npregbw(y ~ x, bws = 0.2, bandwidth.compute = FALSE)
  expect_error(npindexhat(bws = rbw, txdat = data.frame(x = x)), "sibandwidth")
  expect_error(npplreghat(bws = rbw, txdat = data.frame(x = x), tzdat = data.frame(z = z)), "plbandwidth")
  expect_error(npscoefhat(bws = rbw, txdat = data.frame(x = x), tzdat = data.frame(z = z)), "scbandwidth")
  expect_error(
    npindexhat(bws = structure(list(), class = "sibandwidth"),
               txdat = data.frame(x = x), foo = TRUE),
    "unused argument in npindexhat: 'foo'"
  )
  expect_error(
    npplreghat(bws = structure(list(), class = "plbandwidth"),
               txdat = data.frame(x = x), tzdat = data.frame(z = z), foo = TRUE),
    "unused argument in npplreghat: 'foo'"
  )
  expect_error(
    npscoefhat(bws = structure(list(), class = "scbandwidth"),
               txdat = data.frame(x = x), tzdat = data.frame(z = z), foo = TRUE),
    "unused argument in npscoefhat: 'foo'"
  )

  sibw <- npindexbw(
    xdat = data.frame(x = x, x2 = x^2),
    ydat = y,
    bws = c(1, 1, 0.2),
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  expect_error(
    npindexhat(bws = sibw, txdat = data.frame(x = x, x2 = x^2), s = 1L, fd.step = 0),
    "argument 'fd.step' must be a positive finite scalar"
  )

  scbw <- npscoefbw(
    xdat = x,
    zdat = z,
    ydat = y,
    bws = 0.2,
    bandwidth.compute = FALSE
  )
  h0 <- npscoefhat(
    bws = scbw,
    txdat = data.frame(x = x),
    tzdat = data.frame(z = z),
    ridge = 0
  )
  expect_true(is.matrix(h0))
})

test_that("plot bootstrap supports wild for sc/pl/si bandwidth objects", {
  skip_on_cran()

  old.chunk <- getOption("np.plot.wild.chunk.size")
  on.exit(options(np.plot.wild.chunk.size = old.chunk), add = TRUE)
  options(np.plot.wild.chunk.size = 5L)

  set.seed(20260223)
  n <- 80
  x <- runif(n)
  z <- runif(n)
  y <- sin(2 * pi * z) + 1.5 * x + rnorm(n, sd = 0.05)

  scbw <- npscoefbw(
    xdat = x,
    zdat = z,
    ydat = y,
    bws = 0.2,
    bandwidth.compute = FALSE
  )
  sc.out <- suppressWarnings(
    plot(
      scbw,
      xdat = data.frame(x = x),
      ydat = y,
      zdat = data.frame(z = z),
      perspective = FALSE,
      output = "data",
      errors = "bootstrap",
      bootstrap = "wild",
      B = 19
    )
  )
  expect_type(sc.out, "list")
  expect_true(length(sc.out) > 0)

  sc.out.rad <- suppressWarnings(
    plot(
      scbw,
      xdat = data.frame(x = x),
      ydat = y,
      zdat = data.frame(z = z),
      perspective = FALSE,
      output = "data",
      errors = "bootstrap",
      bootstrap = "wild",
      boot_control = np_boot_control(wild = "rademacher"),
      band = "pointwise",
      B = 19
    )
  )
  expect_type(sc.out.rad, "list")
  expect_true(length(sc.out.rad) > 0)
  expect_false(isTRUE(all.equal(
    sc.out[[1]]$merr,
    sc.out.rad[[1]]$merr,
    tolerance = 0,
    check.attributes = FALSE
  )))

  sc.cfgs <- list(
    list(regtype = "ll", basis = NULL, label = "ll"),
    list(regtype = "lp", basis = "glp", label = "lp-glp"),
    list(regtype = "lp", basis = "additive", label = "lp-additive"),
    list(regtype = "lp", basis = "tensor", label = "lp-tensor")
  )
  for (cfg in sc.cfgs) {
    sc.args <- list(
      xdat = x,
      zdat = z,
      ydat = y,
      bws = 0.2,
      bandwidth.compute = FALSE,
      regtype = cfg$regtype
    )
    if (!is.null(cfg$basis)) {
      sc.args$basis <- cfg$basis
      sc.args$degree <- 2L
    }
    scbw.cfg <- do.call(npscoefbw, sc.args)
    sc.out.cfg <- suppressWarnings(
      plot(
        scbw.cfg,
        xdat = data.frame(x = x),
        ydat = y,
        zdat = data.frame(z = z),
        perspective = FALSE,
        output = "data",
        errors = "bootstrap",
        bootstrap = "wild",
        B = 11
      )
    )
    expect_type(sc.out.cfg, "list")
    expect_true(length(sc.out.cfg) > 0, info = cfg$label)
  }

  pl.cfgs <- list(
    list(regtype = "lc", basis = NULL, label = "lc"),
    list(regtype = "ll", basis = NULL, label = "ll"),
    list(regtype = "lp", basis = "glp", label = "lp-glp"),
    list(regtype = "lp", basis = "additive", label = "lp-additive"),
    list(regtype = "lp", basis = "tensor", label = "lp-tensor")
  )
  for (cfg in pl.cfgs) {
    pl.args <- list(
      xdat = x,
      zdat = z,
      ydat = y,
      bws = matrix(c(0.2, 0.2), nrow = 2, ncol = 1),
      bandwidth.compute = FALSE,
      regtype = cfg$regtype
    )
    if (!is.null(cfg$basis)) {
      pl.args$basis <- cfg$basis
      pl.args$degree <- 2L
    }
    plbw <- do.call(npplregbw, pl.args)
    pl.out <- suppressWarnings(
      plot(
        plbw,
        xdat = data.frame(x = x),
        ydat = y,
        zdat = data.frame(z = z),
        perspective = FALSE,
        output = "data",
        errors = "bootstrap",
        bootstrap = "wild",
        B = 9
      )
    )
    expect_true(is.list(pl.out), info = cfg$label)
    expect_true(length(pl.out) > 0, info = cfg$label)
  }

  sibw <- npindexbw(
    xdat = data.frame(x1 = x, x2 = z),
    ydat = y,
    method = "ichimura",
    nmulti = 1
  )
  si.out <- suppressWarnings(
    plot(
      sibw,
      xdat = data.frame(x1 = x, x2 = z),
      ydat = y,
      output = "data",
      errors = "bootstrap",
      bootstrap = "wild",
      B = 19
    )
  )
  expect_type(si.out, "list")
  expect_true(length(si.out) > 0)
  expect_true(is.matrix(si.out[[1]]$merr))
  expect_equal(ncol(si.out[[1]]$merr), 2L)
  expect_false(all(is.na(si.out[[1]]$merr)))
})

library(np)

test_that("generalized-nn ll and canonical lp degree-1 owners stay exact at NN floor", {
  set.seed(20260315)

  n <- 60L
  x <- rnorm(n)
  y <- x^2 + rnorm(n, sd = 0.2)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 31L))

  make_bw <- function(regtype) {
    args <- list(
      xdat = tx,
      ydat = y,
      regtype = regtype,
      bwtype = "generalized_nn",
      bws = 2L,
      bandwidth.compute = FALSE
    )
    if (identical(regtype, "lp")) {
      args$basis <- "glp"
      args$degree <- 1L
      args$bernstein.basis <- FALSE
    }
    do.call(npregbw, args)
  }

  bw.ll <- make_bw("ll")
  bw.lp <- make_bw("lp")

  fit.ll <- npreg(bws = bw.ll, txdat = tx, tydat = y, exdat = ex)
  fit.lp <- npreg(bws = bw.lp, txdat = tx, tydat = y, exdat = ex)

  H.ll <- npreghat(bws = bw.ll, txdat = tx, exdat = ex, output = "matrix")
  H.lp <- npreghat(bws = bw.lp, txdat = tx, exdat = ex, output = "matrix")
  hat.apply.ll <- npreghat(bws = bw.ll, txdat = tx, exdat = ex, y = y, output = "apply")
  hat.apply.lp <- npreghat(bws = bw.lp, txdat = tx, exdat = ex, y = y, output = "apply")

  expect_equal(as.vector(fit.ll$mean), as.vector(fit.lp$mean), tolerance = 1e-10)
  expect_equal(as.vector(hat.apply.ll), as.vector(hat.apply.lp), tolerance = 1e-10)
  expect_equal(as.vector(H.ll %*% y), as.vector(hat.apply.ll), tolerance = 1e-10)
  expect_equal(as.vector(H.lp %*% y), as.vector(hat.apply.lp), tolerance = 1e-10)
  expect_equal(as.vector(H.ll %*% y), as.vector(fit.ll$mean), tolerance = 1e-10)
  expect_equal(as.vector(H.lp %*% y), as.vector(fit.lp$mean), tolerance = 1e-10)
})

test_that("npplreg generalized-nn evaluation hats stay aligned at NN floor", {
  set.seed(20260315)

  n <- 80L
  z <- runif(n)
  x <- 0.7 * z + rnorm(n, sd = 0.2)
  y <- 1 + 1.1 * x + sin(2 * pi * z) + rnorm(n, sd = 0.05)

  tx <- data.frame(x = x)
  tz <- data.frame(z = z)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 31L))
  ez <- data.frame(z = seq(min(z), max(z), length.out = 31L))
  bws.fixed <- matrix(c(2L, 9L), nrow = 2L)

  make_bw <- function(regtype) {
    args <- list(
      xdat = tx,
      zdat = tz,
      ydat = y,
      regtype = regtype,
      bwtype = "generalized_nn",
      bws = bws.fixed,
      bandwidth.compute = FALSE
    )
    if (identical(regtype, "lp")) {
      args$basis <- "glp"
      args$degree <- 1L
      args$bernstein.basis <- FALSE
    }
    do.call(npplregbw, args)
  }

  bw.ll <- make_bw("ll")
  bw.lp <- make_bw("lp")

  fit.ll <- npplreg(bws = bw.ll, txdat = tx, tzdat = tz, tydat = y, exdat = ex, ezdat = ez)
  fit.lp <- npplreg(bws = bw.lp, txdat = tx, tzdat = tz, tydat = y, exdat = ex, ezdat = ez)

  H.ll <- npplreghat(bws = bw.ll, txdat = tx, tzdat = tz, exdat = ex, ezdat = ez, output = "matrix")
  H.lp <- npplreghat(bws = bw.lp, txdat = tx, tzdat = tz, exdat = ex, ezdat = ez, output = "matrix")
  hat.apply.ll <- npplreghat(bws = bw.ll, txdat = tx, tzdat = tz, exdat = ex, ezdat = ez, y = y, output = "apply")
  hat.apply.lp <- npplreghat(bws = bw.lp, txdat = tx, tzdat = tz, exdat = ex, ezdat = ez, y = y, output = "apply")

  expect_equal(as.vector(fit.ll$mean), as.vector(fit.lp$mean), tolerance = 1e-10)
  expect_equal(as.vector(hat.apply.ll), as.vector(hat.apply.lp), tolerance = 1e-10)
  expect_equal(as.vector(H.ll %*% y), as.vector(hat.apply.ll), tolerance = 1e-7)
  expect_equal(as.vector(H.lp %*% y), as.vector(hat.apply.lp), tolerance = 1e-7)
  expect_equal(as.vector(H.ll %*% y), as.vector(fit.ll$mean), tolerance = 1e-7)
  expect_equal(as.vector(H.lp %*% y), as.vector(fit.lp$mean), tolerance = 1e-7)
})

test_that("nonfixed regression NN bandwidths enforce k >= 2 across lc ll and lp", {
  set.seed(42)
  n <- 80
  x <- sort(rnorm(n))
  y <- x^2 + rnorm(n, sd = 0.15)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 19))

  expect_error(
    npregbw(xdat = tx, ydat = y, regtype = "lc", bwtype = "generalized_nn",
            bandwidth.compute = FALSE, bws = 1),
    "nearest-neighbor bandwidth must be at least 2|nearest-neighbor bandwidth must be in \\[2,"
  )

  expect_error(
    npregbw(xdat = tx, ydat = y, regtype = "ll", bwtype = "generalized_nn",
            bandwidth.compute = FALSE, bws = 1),
    "nearest-neighbor bandwidth must be at least 2|nearest-neighbor bandwidth must be in \\[2,"
  )

  expect_error(
    npregbw(xdat = tx, ydat = y, regtype = "lp", degree = 2L, bwtype = "generalized_nn",
            bandwidth.compute = FALSE, bws = 1),
    "nearest-neighbor bandwidth must be at least 2|nearest-neighbor bandwidth must be in \\[2,"
  )

  bw.ll <- npregbw(xdat = tx, ydat = y, regtype = "ll", bwtype = "generalized_nn",
                   bandwidth.compute = FALSE, bws = 2)
  fit.ll <- npreg(bws = bw.ll, exdat = ex)
  hat.ll <- npreghat(bws = bw.ll, txdat = tx, exdat = ex)
  apply.ll <- npreghat(bws = bw.ll, txdat = tx, exdat = ex, y = y, output = "apply")
  expect_equal(drop(hat.ll %*% y), drop(fit.ll$mean), tolerance = 1e-10)
  expect_equal(drop(apply.ll), drop(fit.ll$mean), tolerance = 1e-10)

  bw.auto <- npregbw(xdat = tx, ydat = y, regtype = "ll", bwtype = "generalized_nn", nmulti = 1)
  expect_gte(as.integer(bw.auto$bw[1]), 2L)

  bw.bad <- bw.ll
  bw.bad$bw[1] <- 1
  bw.bad$bandwidth$x[1] <- 1
  expect_error(npreg(bws = bw.bad), "nearest-neighbor bandwidth must be in \\[2,")
  expect_error(npreghat(bws = bw.bad, txdat = tx), "nearest-neighbor bandwidth must be in \\[2,")
})

test_that("semiparametric NN selectors honor the same floor on common-use routes", {
  set.seed(9202)
  n <- 60L
  z <- sort(runif(n))
  x <- 0.7 * z + rnorm(n, sd = 0.18)
  y <- 1 + 1.1 * x + sin(2 * pi * z) + rnorm(n, sd = 0.05)
  tx <- data.frame(x = x)
  tz <- data.frame(z = z)
  ex <- data.frame(x = seq(quantile(x, 0.2), quantile(x, 0.8), length.out = 25L))
  ez <- data.frame(z = seq(0.1, 0.9, length.out = 25L))

  bw.pl <- npplregbw(xdat = tx, zdat = tz, ydat = y,
                     regtype = "ll", bwtype = "generalized_nn", nmulti = 1)
  expect_gte(min(vapply(bw.pl$bw, function(one) min(as.integer(one$bw[one$icon])), integer(1))), 2L)
  fit.pl <- npplreg(bws = bw.pl, txdat = tx, tzdat = tz, tydat = y, exdat = ex, ezdat = ez)$mean
  apply.pl <- npplreghat(bws = bw.pl, txdat = tx, tzdat = tz, exdat = ex, ezdat = ez, y = y, output = "apply")
  matrix.pl <- drop(npplreghat(bws = bw.pl, txdat = tx, tzdat = tz, exdat = ex, ezdat = ez, output = "matrix") %*% y)
  expect_equal(drop(apply.pl), drop(fit.pl), tolerance = 1e-10)
  expect_equal(drop(matrix.pl), drop(fit.pl), tolerance = 1e-10)

  set.seed(9203)
  n <- 80L
  x1 <- runif(n, -1, 1)
  x2 <- rnorm(n)
  y <- sin(2 * (0.7 * x1 - 0.3 * x2)) + 0.2 * x1 * x2 + rnorm(n, sd = 0.05)
  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- data.frame(x1 = seq(-0.75, 0.75, length.out = 20L),
                   x2 = seq(-0.8, 0.8, length.out = 20L))

  bw.si <- npindexbw(xdat = tx, ydat = y, regtype = "ll",
                     bwtype = "generalized_nn", nmulti = 1)
  expect_gte(as.integer(bw.si$bw[1]), 2L)
  fit.si <- npindex(bws = bw.si, txdat = tx, tydat = y, exdat = ex,
                    gradients = FALSE, errors = FALSE)$mean
  apply.si <- npindexhat(bws = bw.si, txdat = tx, exdat = ex, y = y, output = "apply", s = 0L)
  matrix.si <- drop(npindexhat(bws = bw.si, txdat = tx, exdat = ex, output = "matrix", s = 0L) %*% y)
  expect_equal(drop(apply.si), drop(fit.si), tolerance = 1e-10)
  expect_equal(drop(matrix.si), drop(fit.si), tolerance = 1e-10)

  set.seed(9204)
  n <- 80L
  x <- runif(n, -1, 1)
  z <- rnorm(n)
  y <- x^2 + rnorm(n, sd = 0.25 * sd(x))
  tx <- data.frame(x = x)
  tz <- data.frame(z = z)

  bw.sc <- npscoefbw(xdat = tx, zdat = tz, ydat = y, nmulti = 1, bwtype = "adaptive_nn")
  expect_gte(min(as.integer(bw.sc$bw)), 2L)
  fit.sc <- npscoef(bws = bw.sc, txdat = tx, tzdat = tz, tydat = y)$mean
  matrix.sc <- drop(npscoefhat(bws = bw.sc, txdat = tx, tzdat = tz) %*% y)
  expect_equal(drop(matrix.sc), drop(fit.sc), tolerance = 1e-10)
})

library(np)

test_that("npindexhat lc owner preserves fit and matrix/apply parity across bwtypes", {
  set.seed(20260315)
  n <- 90L
  x1 <- runif(n, -1, 1)
  x2 <- rnorm(n)
  y <- sin(2 * (0.7 * x1 - 0.3 * x2)) + 0.25 * x1 + rnorm(n, sd = 0.04)

  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- data.frame(
    x1 = seq(min(x1) * 0.9, max(x1) * 0.9, length.out = 30L),
    x2 = seq(quantile(x2, 0.2), quantile(x2, 0.8), length.out = 30L)
  )

  for (bwtype in c("fixed", "generalized_nn", "adaptive_nn")) {
    bw <- npindexbw(
      xdat = tx,
      ydat = y,
      regtype = "lc",
      bwtype = bwtype,
      bandwidth.compute = FALSE,
      bws = c(1, 1, if (identical(bwtype, "fixed")) 0.22 else 9L)
    )

    fit.mean <- npindex(
      bws = bw,
      txdat = tx,
      tydat = y,
      exdat = ex,
      gradients = FALSE,
      errors = FALSE
    )
    fit.grad <- npindex(
      bws = bw,
      txdat = tx,
      tydat = y,
      exdat = ex,
      gradients = TRUE,
      errors = FALSE
    )

    apply.mean <- npindexhat(
      bws = bw,
      txdat = tx,
      exdat = ex,
      y = y,
      output = "apply",
      s = 0L
    )
    matrix.mean <- npindexhat(
      bws = bw,
      txdat = tx,
      exdat = ex,
      output = "matrix",
      s = 0L
    )
    apply.grad <- npindexhat(
      bws = bw,
      txdat = tx,
      exdat = ex,
      y = y,
      output = "apply",
      s = 1L
    )
    matrix.grad <- npindexhat(
      bws = bw,
      txdat = tx,
      exdat = ex,
      output = "matrix",
      s = 1L
    )

    expect_equal(as.vector(apply.mean), as.vector(fit.mean$mean), tolerance = 1e-8, info = paste("mean", bwtype))
    expect_equal(as.vector(matrix.mean %*% y), as.vector(fit.mean$mean), tolerance = 1e-8, info = paste("mean matrix", bwtype))
    expect_equal(as.vector(apply.mean), as.vector(matrix.mean %*% y), tolerance = 1e-10, info = paste("mean parity", bwtype))

    expect_equal(as.vector(apply.grad), as.vector(fit.grad$grad[, 1L]), tolerance = 1e-8, info = paste("grad", bwtype))
    expect_equal(as.vector(matrix.grad %*% y), as.vector(fit.grad$grad[, 1L]), tolerance = 1e-8, info = paste("grad matrix", bwtype))
    expect_equal(as.vector(apply.grad), as.vector(matrix.grad %*% y), tolerance = 1e-10, info = paste("grad parity", bwtype))
  }
})

expected_singleindex_plot <- function(bw, tx, y, neval = 50L, gradients = FALSE) {
  eval_grid <- getFromNamespace(".np_plot_singleindex_eval_grid", "np")
  local_eval <- getFromNamespace(".np_plot_singleindex_local_eval", "np")

  info <- eval_grid(bws = bw, xdat = tx, neval = neval, where = "test")
  out <- local_eval(
    bws = bw,
    idx.train = info$idx.train,
    idx.eval = info$idx.eval,
    ydat = y,
    gradients = gradients
  )
  out$trainiseval <- info$trainiseval
  out
}

with_plot_device <- function(expr) {
  pdf(file = tempfile(fileext = ".pdf"))
  on.exit(dev.off(), add = TRUE)
  suppressWarnings(force(expr))
}

run_singleindex_bootstrap_plot <- function(bw,
                                           xdat,
                                           ydat,
                                           boot_method,
                                           gradients = FALSE,
                                           output = "data",
                                           boot_num = 9L,
                                           neval = 50L) {
  suppressWarnings(plot(
    bw,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    output = behavior,
    perspective = FALSE,
    gradients = gradients,
    errors = "bootstrap",
    bootstrap = boot_method,
    B = boot_num
  ))[[1]]
}

test_that("sibandwidth plot/helper route is free of public core re-entry", {
  skip_if_not_installed("np")

  touched <- c(
    ".np_indexhat_rbw",
    ".np_indexhat_kbw",
    ".np_indexhat_core",
    ".np_indexhat_exact",
    "npindexhat",
    ".np_plot_sibandwidth_engine",
    ".np_inid_boot_from_index",
    ".np_inid_boot_from_index_exact",
    "compute.bootstrap.errors.sibandwidth"
  )
  forbidden <- c(
    "\\bnpindex\\(",
    "\\bnpindexbw\\(",
    "\\bnpreg\\(",
    "\\bnpregbw\\(",
    "\\bnpplreg\\(",
    "\\bnpplregbw\\(",
    "\\bnpscoef\\(",
    "\\bnpscoefbw\\(",
    "\\bnpcdens\\(",
    "\\bnpcdensbw\\(",
    "\\bnpksum\\(",
    "\\bnpregbw\\(",
    "\\brbandwidth\\(",
    "\\bkbandwidth\\("
  )

  for (nm in touched) {
    fn <- getFromNamespace(nm, "np")
    txt <- paste(deparse(body(fn), width.cutoff = 500L), collapse = "\n")
    for (pat in forbidden) {
      expect_false(
        grepl(pat, txt, perl = TRUE),
        info = paste(nm, "contains forbidden public call pattern", pat)
      )
    }
  }
})

test_that("npindex helpers match fitted and gradient outputs across regtype and bwtype", {
  skip_if_not_installed("np")

  indexhat_rbw <- getFromNamespace(".np_indexhat_rbw", "np")
  regression_direct <- getFromNamespace(".np_regression_direct", "np")

  set.seed(20260307)
  n <- 45
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.06)
  tx <- data.frame(x1 = x1, x2 = x2)

  cfgs <- list(
    list(regtype = "lc", basis = NULL, degree = NULL, label = "lc"),
    list(regtype = "ll", basis = NULL, degree = NULL, label = "ll"),
    list(regtype = "lp", basis = "tensor", degree = 2L, label = "lp")
  )

  for (bt in c("fixed", "generalized_nn", "adaptive_nn")) {
    h <- if (identical(bt, "fixed")) 0.25 else 5L
    for (cfg in cfgs) {
      bw.args <- list(
        xdat = tx,
        ydat = y,
        bws = c(1, 1, h),
        bandwidth.compute = FALSE,
        regtype = cfg$regtype,
        bwtype = bt
      )
      if (!is.null(cfg$basis)) {
        bw.args$basis <- cfg$basis
        bw.args$degree <- cfg$degree
      }

      bw <- do.call(npindexbw, bw.args)
      fit.mean <- npindex(
        bws = bw,
        txdat = tx,
        tydat = y,
        gradients = FALSE
      )
      fit.grad <- npindex(
        bws = bw,
        txdat = tx,
        tydat = y,
        gradients = TRUE
      )
      helper.ref <- list(index = as.vector(as.matrix(tx) %*% bw$beta))
      fit.mean.aligned <- fit.mean$mean
      fit.grad.mean.aligned <- fit.grad$mean
      fit.grad.aligned <- fit.grad$grad
      idx.train <- data.frame(index = helper.ref$index)
      rbw <- indexhat_rbw(bws = bw, idx.train = idx.train)
      helper.grad <- regression_direct(
        bws = rbw,
        txdat = idx.train,
        tydat = y,
        exdat = idx.train,
        gradients = TRUE,
        gradient.order = 1L
      )

      hat.mean <- npindexhat(
        bws = bw,
        txdat = tx,
        exdat = tx,
        y = y,
        output = "apply",
        s = 0L
      )
      hat.mean.matrix <- npindexhat(
        bws = bw,
        txdat = tx,
        exdat = tx,
        output = "matrix",
        s = 0L
      )

      expect_equal(
        hat.mean,
        fit.mean.aligned,
        tolerance = 1e-10,
        info = paste(bt, cfg$label, "hat mean")
      )
      expect_equal(
        as.vector(hat.mean.matrix %*% y),
        hat.mean,
        tolerance = 1e-10,
        info = paste(bt, cfg$label, "hat mean matrix")
      )

      expect_equal(
        helper.grad$mean,
        fit.grad.mean.aligned,
        tolerance = 1e-10,
        info = paste(bt, cfg$label, "helper gradient mean")
      )
      expect_equal(
        helper.grad$grad[, 1L],
        fit.grad.aligned[, 1L] / bw$beta[1L],
        tolerance = 1e-10,
        info = paste(bt, cfg$label, "helper gradient apply")
      )
      expect_equal(
        fit.grad.aligned,
        helper.grad$grad[, 1L] %o% as.vector(bw$beta),
        tolerance = 1e-10,
        info = paste(bt, cfg$label, "full gradient reconstruction")
      )
    }
  }
})

test_that("sibandwidth plot payload matches single-index fits across regtype and bwtype", {
  skip_if_not_installed("np")

  set.seed(20260307)
  n <- 45
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.06)
  tx <- data.frame(x1 = x1, x2 = x2)

  cfgs <- list(
    list(regtype = "lc", basis = NULL, degree = NULL, label = "lc"),
    list(regtype = "ll", basis = NULL, degree = NULL, label = "ll"),
    list(regtype = "lp", basis = "tensor", degree = 2L, label = "lp")
  )

  for (bt in c("fixed", "generalized_nn", "adaptive_nn")) {
    h <- if (identical(bt, "fixed")) 0.25 else 5L
    for (cfg in cfgs) {
      bw.args <- list(
        xdat = tx,
        ydat = y,
        bws = c(1, 1, h),
        bandwidth.compute = FALSE,
        regtype = cfg$regtype,
        bwtype = bt
      )
      if (!is.null(cfg$basis)) {
        bw.args$basis <- cfg$basis
        bw.args$degree <- cfg$degree
      }

      bw <- do.call(npindexbw, bw.args)
      neval <- 17L
      expected.mean <- expected_singleindex_plot(
        bw = bw,
        tx = tx,
        y = y,
        neval = neval,
        gradients = FALSE
      )

      out.data.mean <- suppressWarnings(plot(
        bw,
        xdat = tx,
        ydat = y,
        neval = neval,
        output = "data",
        perspective = FALSE,
        gradients = FALSE
      ))[[1]]
      out.plot.mean <- with_plot_device(plot(
        bw,
        xdat = tx,
        ydat = y,
        neval = neval,
        output = "plot-data",
        perspective = FALSE,
        gradients = FALSE
      ))[[1]]

      expect_equal(
        out.data.mean$index,
        expected.mean$index,
        tolerance = 1e-12,
        info = paste(bt, cfg$label, "index")
      )
      expect_equal(
        out.data.mean$mean,
        expected.mean$mean,
        tolerance = 1e-10,
        info = paste(bt, cfg$label, "mean")
      )
      expect_equal(
        out.data.mean$mean,
        out.plot.mean$mean,
        tolerance = 1e-12,
        info = paste(bt, cfg$label, "mean plot-data")
      )
      expect_false(out.data.mean$trainiseval)
      expect_equal(length(out.data.mean$mean), neval)

      expected.grad <- expected_singleindex_plot(
        bw = bw,
        tx = tx,
        y = y,
        neval = neval,
        gradients = TRUE
      )

      out.data.grad <- suppressWarnings(plot(
        bw,
        xdat = tx,
        ydat = y,
        neval = neval,
        output = "data",
        perspective = FALSE,
        gradients = TRUE
      ))[[1]]
      out.plot.grad <- with_plot_device(plot(
        bw,
        xdat = tx,
        ydat = y,
        neval = neval,
        output = "plot-data",
        perspective = FALSE,
        gradients = TRUE
      ))[[1]]

      expect_equal(
        out.data.grad$index,
        expected.grad$index,
        tolerance = 1e-12,
        info = paste(bt, cfg$label, "grad index")
      )
      expect_equal(
        out.data.grad$mean,
        expected.grad$mean,
        tolerance = 1e-10,
        info = paste(bt, cfg$label, "grad mean")
      )
      expect_equal(
        out.data.grad$grad,
        expected.grad$grad,
        tolerance = 1e-10,
        info = paste(bt, cfg$label, "grad")
      )
      expect_equal(
        out.data.grad$grad,
        out.plot.grad$grad,
        tolerance = 1e-12,
        info = paste(bt, cfg$label, "grad plot-data")
      )
      expect_false(out.data.grad$trainiseval)
      expect_equal(nrow(out.data.grad$grad), neval)
    }
  }
})

test_that("sibandwidth bootstrap helper routes work across bwtype, regtype, and method", {
  skip_if_not_installed("np")

  set.seed(20260307)
  n <- 35
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.08)
  tx <- data.frame(x1 = x1, x2 = x2)

  cfgs <- list(
    list(regtype = "lc", basis = NULL, degree = NULL, label = "lc"),
    list(regtype = "ll", basis = NULL, degree = NULL, label = "ll"),
    list(regtype = "lp", basis = "tensor", degree = 2L, label = "lp")
  )

  for (bt in c("fixed", "generalized_nn", "adaptive_nn")) {
    h <- if (identical(bt, "fixed")) 0.25 else 5L
    for (cfg in cfgs) {
      bw.args <- list(
        xdat = tx,
        ydat = y,
        bws = c(1, 1, h),
        bandwidth.compute = FALSE,
        regtype = cfg$regtype,
        bwtype = bt
      )
      if (!is.null(cfg$basis)) {
        bw.args$basis <- cfg$basis
        bw.args$degree <- cfg$degree
      }

      bw <- do.call(npindexbw, bw.args)
      neval <- 19L
      for (boot_method in c("wild", "inid", "fixed", "geom")) {
        out.data <- run_singleindex_bootstrap_plot(
          bw = bw,
          xdat = tx,
          ydat = y,
          boot_method = boot_method,
          gradients = FALSE,
          output = "data",
          neval = neval
        )
        out.plot <- with_plot_device(run_singleindex_bootstrap_plot(
          bw = bw,
          xdat = tx,
          ydat = y,
          boot_method = boot_method,
          gradients = FALSE,
          output = "plot-data",
          neval = neval
        ))

        expect_equal(
          length(out.data$mean),
          neval,
          info = paste(bt, cfg$label, boot_method, "mean length")
        )
        expect_equal(
          dim(out.data$merr),
          c(neval, 2L),
          info = paste(bt, cfg$label, boot_method, "merr shape")
        )
        expect_true(
          all(is.finite(out.data$mean)),
          info = paste(bt, cfg$label, boot_method, "mean finite")
        )
        expect_true(
          all(is.finite(out.data$merr)),
          info = paste(bt, cfg$label, boot_method, "merr finite")
        )
        expect_equal(
          out.data$mean,
          out.plot$mean,
          tolerance = 1e-12,
          info = paste(bt, cfg$label, boot_method, "plot-data center")
        )
      }
    }
  }
})

test_that("sibandwidth nonfixed bootstrap helpers fail fast for gradients", {
  skip_if_not_installed("np")

  set.seed(20260307)
  n <- 30
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.08)
  tx <- data.frame(x1 = x1, x2 = x2)

  for (bt in c("generalized_nn", "adaptive_nn")) {
    h <- 5L
    bw <- npindexbw(
      xdat = tx,
      ydat = y,
      bws = c(1, 1, h),
      bandwidth.compute = FALSE,
      regtype = "ll",
      bwtype = bt
    )

    for (boot_method in c("inid", "fixed", "geom")) {
      expect_error(
        run_singleindex_bootstrap_plot(
          bw = bw,
          xdat = tx,
          ydat = y,
          boot_method = boot_method,
          gradients = TRUE
        ),
        "requires helper mode with gradients=FALSE",
        info = paste(bt, boot_method, "gradient bootstrap guard")
      )
    }
  }
})

library(np)

test_that("semihat wrappers preserve infinite-bound parity and finite-bound evaluation checks", {
  set.seed(20260320)
  n <- 90L
  x <- seq(0, 1, length.out = n)
  z <- seq(0, 1, length.out = n)
  y <- sin(2 * pi * z) + 0.8 * x + rnorm(n, sd = 0.03)

  tx <- data.frame(x = x)
  tz <- data.frame(z = z)
  ex <- data.frame(x = c(0, 0.5, 1))
  ez <- data.frame(z = c(0, 0.5, 1))
  ex.bad <- data.frame(x = c(-0.01, 0.5, 1.01))
  ez.bad <- data.frame(z = c(-0.01, 0.5, 1.01))

  sibw.none <- suppressWarnings(npindexbw(
    xdat = tx,
    ydat = y,
    bws = c(1, 0.4),
    bandwidth.compute = FALSE,
    regtype = "lc",
    bwtype = "fixed",
    ckertype = "gaussian",
    ckerorder = 2L,
    ckerbound = "none"
  ))
  sibw.inf <- suppressWarnings(npindexbw(
    xdat = tx,
    ydat = y,
    bws = c(1, 0.4),
    bandwidth.compute = FALSE,
    regtype = "lc",
    bwtype = "fixed",
    ckertype = "gaussian",
    ckerorder = 2L,
    ckerbound = "fixed",
    ckerlb = -Inf,
    ckerub = Inf
  ))
  sibw.fixed <- suppressWarnings(npindexbw(
    xdat = tx,
    ydat = y,
    bws = c(1, 0.4),
    bandwidth.compute = FALSE,
    regtype = "lc",
    bwtype = "fixed",
    ckertype = "gaussian",
    ckerorder = 2L,
    ckerbound = "fixed",
    ckerlb = 0,
    ckerub = 1
  ))

  si.none <- npindexhat(bws = sibw.none, txdat = tx, exdat = ex, y = y, s = 0L, output = "apply")
  si.inf <- npindexhat(bws = sibw.inf, txdat = tx, exdat = ex, y = y, s = 0L, output = "apply")
  si.fixed <- npindexhat(bws = sibw.fixed, txdat = tx, exdat = ex, y = y, s = 0L, output = "apply")

  expect_equal(as.vector(si.none), as.vector(si.inf), tolerance = 1e-10)
  expect_equal(as.vector(si.none), as.vector(si.fixed), tolerance = 1e-10)
  expect_error(
    npindexhat(bws = sibw.fixed, txdat = tx, exdat = ex.bad, y = y, s = 0L, output = "apply"),
    "Invalid bounds for 'ckerbound'|Evaluation data violate 'ckerbound' bounds"
  )

  scbw.none <- npscoefbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    bws = 0.4,
    bandwidth.compute = FALSE,
    regtype = "lc",
    bwtype = "fixed",
    ckertype = "gaussian",
    ckerorder = 2L,
    ckerbound = "none"
  )
  scbw.inf <- npscoefbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    bws = 0.4,
    bandwidth.compute = FALSE,
    regtype = "lc",
    bwtype = "fixed",
    ckertype = "gaussian",
    ckerorder = 2L,
    ckerbound = "fixed",
    ckerlb = -Inf,
    ckerub = Inf
  )
  scbw.fixed <- npscoefbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    bws = 0.4,
    bandwidth.compute = FALSE,
    regtype = "lc",
    bwtype = "fixed",
    ckertype = "gaussian",
    ckerorder = 2L,
    ckerbound = "fixed",
    ckerlb = 0,
    ckerub = 1
  )

  sc.none <- npscoefhat(
    bws = scbw.none,
    txdat = tx,
    tzdat = tz,
    exdat = ex,
    ezdat = ez,
    y = y,
    output = "apply",
    iterate = FALSE
  )
  sc.inf <- npscoefhat(
    bws = scbw.inf,
    txdat = tx,
    tzdat = tz,
    exdat = ex,
    ezdat = ez,
    y = y,
    output = "apply",
    iterate = FALSE
  )
  sc.fixed <- npscoefhat(
    bws = scbw.fixed,
    txdat = tx,
    tzdat = tz,
    exdat = ex,
    ezdat = ez,
    y = y,
    output = "apply",
    iterate = FALSE
  )

  expect_equal(as.vector(sc.none), as.vector(sc.inf), tolerance = 1e-10)
  expect_equal(as.vector(sc.none), as.vector(sc.fixed), tolerance = 1e-10)
  expect_error(
    npscoefhat(
      bws = scbw.fixed,
      txdat = tx,
      tzdat = tz,
      exdat = ex,
      ezdat = ez.bad,
      y = y,
      output = "apply",
      iterate = FALSE
    ),
    "Invalid bounds for 'ckerbound'|Evaluation data violate 'ckerbound' bounds"
  )

  plbw.none <- npplregbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    bws = matrix(c(0.4, 0.4), nrow = 2L, ncol = 1L),
    bandwidth.compute = FALSE,
    regtype = "lc",
    bwtype = "fixed",
    ckertype = "gaussian",
    ckerorder = 2L,
    ckerbound = "none"
  )
  plbw.inf <- npplregbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    bws = matrix(c(0.4, 0.4), nrow = 2L, ncol = 1L),
    bandwidth.compute = FALSE,
    regtype = "lc",
    bwtype = "fixed",
    ckertype = "gaussian",
    ckerorder = 2L,
    ckerbound = "fixed",
    ckerlb = -Inf,
    ckerub = Inf
  )
  plbw.fixed <- npplregbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    bws = matrix(c(0.4, 0.4), nrow = 2L, ncol = 1L),
    bandwidth.compute = FALSE,
    regtype = "lc",
    bwtype = "fixed",
    ckertype = "gaussian",
    ckerorder = 2L,
    ckerbound = "fixed",
    ckerlb = 0,
    ckerub = 1
  )

  pl.none <- npplreghat(
    bws = plbw.none,
    txdat = tx,
    tzdat = tz,
    exdat = ex,
    ezdat = ez,
    y = y,
    output = "apply"
  )
  pl.inf <- npplreghat(
    bws = plbw.inf,
    txdat = tx,
    tzdat = tz,
    exdat = ex,
    ezdat = ez,
    y = y,
    output = "apply"
  )
  pl.fixed <- npplreghat(
    bws = plbw.fixed,
    txdat = tx,
    tzdat = tz,
    exdat = ex,
    ezdat = ez,
    y = y,
    output = "apply"
  )

  expect_equal(as.vector(pl.none), as.vector(pl.inf), tolerance = 1e-10)
  expect_equal(as.vector(pl.none), as.vector(pl.fixed), tolerance = 1e-10)
  expect_error(
    npplreghat(
      bws = plbw.fixed,
      txdat = tx,
      tzdat = tz,
      exdat = ex,
      ezdat = ez.bad,
      y = y,
      output = "apply"
    ),
    "Invalid bounds for 'ckerbound'|Evaluation data violate 'ckerbound' bounds"
  )
})

test_that("bounded nonfixed semihat support is widened for pl, index, and scoef", {
  set.seed(20260325)
  n <- 48L
  x1 <- runif(n)
  x2 <- runif(n)
  z <- runif(n)
  y_index <- sin(2 * pi * (0.7 * x1 - 0.4 * x2)) + rnorm(n, sd = 0.05)
  y_pl <- 0.8 * x1 + cos(2 * pi * z) + rnorm(n, sd = 0.05)
  y_sc <- (1 + sin(2 * pi * z)) * x1 + rnorm(n, sd = 0.05)

  tx_index <- data.frame(x1 = x1, x2 = x2)
  tx1 <- data.frame(x = x1)
  tz <- data.frame(z = z)

  bw.idx.gnn <- npindexbw(
    xdat = tx_index,
    ydat = y_index,
    method = "ichimura",
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1,
    itmax = 40,
    tol = 0.1
  )
  bw.idx.adapt <- npindexbw(
    xdat = tx_index,
    ydat = y_index,
    method = "ichimura",
    bwtype = "adaptive_nn",
    ckerbound = "range",
    nmulti = 1,
    itmax = 40,
    tol = 0.1
  )
  fit.idx.gnn <- npindex(bws = bw.idx.gnn, txdat = tx_index, tydat = y_index)
  fit.idx.adapt <- npindex(bws = bw.idx.adapt, txdat = tx_index, tydat = y_index)

  expect_true(all(is.finite(as.numeric(bw.idx.gnn$beta))))
  expect_true(is.finite(as.numeric(bw.idx.gnn$bw)))
  expect_true(all(is.finite(as.numeric(fit.idx.gnn$mean))))
  expect_true(all(is.finite(as.numeric(bw.idx.adapt$beta))))
  expect_true(is.finite(as.numeric(bw.idx.adapt$bw)))
  expect_true(all(is.finite(as.numeric(fit.idx.adapt$mean))))

  bw.pl.gnn <- npplregbw(
    xdat = tx1,
    ydat = y_pl,
    zdat = tz,
    regtype = "lc",
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1,
    itmax = 40,
    tol = 0.1
  )
  bw.pl.adapt <- npplregbw(
    xdat = tx1,
    ydat = y_pl,
    zdat = tz,
    regtype = "lc",
    bwtype = "adaptive_nn",
    ckerbound = "range",
    nmulti = 1,
    itmax = 40,
    tol = 0.1
  )
  fit.pl.gnn <- npplreg(bws = bw.pl.gnn, txdat = tx1, tydat = y_pl, tzdat = tz)
  fit.pl.adapt <- npplreg(bws = bw.pl.adapt, txdat = tx1, tydat = y_pl, tzdat = tz)

  expect_true(all(is.finite(as.numeric(unlist(lapply(bw.pl.gnn$bw, function(obj) obj$bw))))))
  expect_true(all(is.finite(as.numeric(unlist(lapply(bw.pl.adapt$bw, function(obj) obj$bw))))))
  expect_true(all(is.finite(as.numeric(fit.pl.gnn$mean))))
  expect_true(all(is.finite(as.numeric(fit.pl.adapt$mean))))

  bw.sc.gnn <- npscoefbw(
    xdat = tx1,
    ydat = y_sc,
    zdat = tz,
    regtype = "lc",
    bwtype = "generalized_nn",
    ckerbound = "range",
    nmulti = 1,
    itmax = 40,
    tol = 0.1
  )
  bw.sc.adapt <- npscoefbw(
    xdat = tx1,
    ydat = y_sc,
    zdat = tz,
    regtype = "lc",
    bwtype = "adaptive_nn",
    ckerbound = "range",
    nmulti = 1,
    itmax = 40,
    tol = 0.1
  )
  fit.sc.gnn <- npscoef(bws = bw.sc.gnn, txdat = tx1, tydat = y_sc, tzdat = tz, iterate = FALSE)
  fit.sc.adapt <- npscoef(bws = bw.sc.adapt, txdat = tx1, tydat = y_sc, tzdat = tz, iterate = FALSE)

  expect_true(all(is.finite(as.numeric(bw.sc.gnn$bw))))
  expect_true(all(is.finite(as.numeric(fit.sc.gnn$mean))))
  expect_true(all(is.finite(as.numeric(bw.sc.adapt$bw))))
  expect_true(all(is.finite(as.numeric(fit.sc.adapt$mean))))
})

library(np)

test_that("npindexhat ll owner preserves fit and matrix/apply parity across bwtypes", {
  set.seed(20260314)
  n <- 70
  x1 <- runif(n)
  x2 <- runif(n)
  y <- x1 + 0.5 * x2 + rnorm(n, sd = 0.04)

  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- tx[seq_len(18), , drop = FALSE]

  for (bwtype in c("fixed", "generalized_nn", "adaptive_nn")) {
    bw <- npindexbw(
      xdat = tx,
      ydat = y,
      regtype = "ll",
      bwtype = bwtype,
      bandwidth.compute = FALSE,
      bws = c(1, 1, if (identical(bwtype, "fixed")) 0.18 else 9)
    )

    fit.mean <- npindex(
      bws = bw,
      txdat = tx,
      tydat = y,
      exdat = ex,
      gradients = FALSE
    )
    fit.grad <- npindex(
      bws = bw,
      txdat = tx,
      tydat = y,
      exdat = ex,
      gradients = TRUE
    )

    apply.mean <- npindexhat(
      bws = bw,
      txdat = tx,
      exdat = ex,
      y = y,
      output = "apply",
      s = 0L
    )
    matrix.mean <- npindexhat(
      bws = bw,
      txdat = tx,
      exdat = ex,
      output = "matrix",
      s = 0L
    )

    apply.grad <- npindexhat(
      bws = bw,
      txdat = tx,
      exdat = ex,
      y = y,
      output = "apply",
      s = 1L
    )
    matrix.grad <- npindexhat(
      bws = bw,
      txdat = tx,
      exdat = ex,
      output = "matrix",
      s = 1L
    )

    expect_equal(as.vector(apply.mean), as.vector(fit.mean$mean), tolerance = 1e-8, info = paste("mean", bwtype))
    expect_equal(as.vector(matrix.mean %*% y), as.vector(fit.mean$mean), tolerance = 1e-8, info = paste("mean matrix", bwtype))
    expect_equal(as.vector(apply.mean), as.vector(matrix.mean %*% y), tolerance = 1e-10, info = paste("mean parity", bwtype))

    expect_equal(as.vector(apply.grad), as.vector(fit.grad$grad[, 1]), tolerance = 1e-8, info = paste("grad", bwtype))
    expect_equal(as.vector(matrix.grad %*% y), as.vector(fit.grad$grad[, 1]), tolerance = 1e-8, info = paste("grad matrix", bwtype))
    expect_equal(as.vector(apply.grad), as.vector(matrix.grad %*% y), tolerance = 1e-10, info = paste("grad parity", bwtype))
  }
})

test_that("npreg cv objective and bandwidths match for ll and lp(degree=1)", {
  set.seed(20260227)
  x <- runif(70)
  y <- sin(2 * pi * x) + rnorm(70, sd = 0.05)
  tx <- data.frame(x = x)

  for (m in c("cv.ls", "cv.aic")) {
    for (bt in c("fixed", "generalized_nn")) {
      set.seed(90210)
      bw.ll <- npregbw(
        xdat = tx,
        ydat = y,
        regtype = "ll",
        bwmethod = m,
        bwtype = bt,
        nmulti = 1L
      )
      set.seed(90210)
      bw.lp <- npregbw(
        xdat = tx,
        ydat = y,
        regtype = "lp",
        basis = "glp",
        degree = 1L,
        bwmethod = m,
        bwtype = bt,
        nmulti = 1L
      )

      expect_identical(bw.ll$regtype, "ll")
      expect_identical(bw.ll$pregtype, "Local-Linear")
      expect_identical(bw.ll$regtype.engine, "lp")
      expect_identical(as.integer(bw.ll$degree.engine), rep.int(1L, bw.ll$ncon))
      expect_equal(as.numeric(bw.ll$fval), as.numeric(bw.lp$fval), tolerance = 1e-10)
      expect_equal(as.numeric(bw.ll$bw), as.numeric(bw.lp$bw), tolerance = 1e-9)
    }
  }
})

test_that("npreg categorical-only predictors reject impossible degree structure", {
  set.seed(20260509)
  n <- 120
  tx <- data.frame(
    f = factor(rbinom(n, 1L, 0.45)),
    g = ordered(sample(letters[1:3], n, replace = TRUE))
  )
  y <- 0.5 * as.integer(tx$f) + 0.2 * as.integer(tx$g) + rnorm(n, sd = 0.35)

  set.seed(90210)
  bw.lc <- npregbw(xdat = tx, ydat = y, regtype = "lc",
                   bwmethod = "cv.aic", nmulti = 1L)
  expect_error(
    npregbw(xdat = tx, ydat = y, regtype = "ll",
            bwmethod = "cv.aic", nmulti = 1L),
    "requires at least one continuous predictor"
  )
  expect_error(
    npregbw(xdat = tx, ydat = y, regtype = "lp", degree = 1L,
            bwmethod = "cv.aic", nmulti = 1L),
    "degree must be 0"
  )
  expect_error(
    npregbw(xdat = tx, ydat = y, regtype = "lp",
            bwmethod = "cv.aic", nmulti = 1L),
    "degree must be 0"
  )
  expect_error(
    npregbw(xdat = tx, ydat = y, regtype = "lp",
            degree.select = "coordinate", search.engine = "cell",
            bwmethod = "cv.aic", nmulti = 1L),
    "automatic degree search requires at least one continuous"
  )
  set.seed(90210)
  bw.lp <- npregbw(xdat = tx, ydat = y, regtype = "lp", degree = 0L,
                   bwmethod = "cv.aic", nmulti = 1L)

  expect_identical(bw.lp$regtype, "lp")
  expect_identical(bw.lp$regtype.engine, "lc")
  expect_equal(as.numeric(bw.lc$fval), as.numeric(bw.lp$fval), tolerance = 1e-12)
  expect_equal(as.numeric(bw.lc$bw), as.numeric(bw.lp$bw), tolerance = 1e-12)
  expect_error(
    npregbw(xdat = tx, ydat = y, nomad = TRUE),
    "nomad=TRUE requires at least one continuous predictor"
  )
})

test_that("npreg and npreghat match for ll and lp(degree=1) in 1D", {
  set.seed(20260227)
  n <- 140
  x <- sort(runif(n))
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.02)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 50))

  bw.ll <- npregbw(
    xdat = tx,
    ydat = y,
    bws = 0.2,
    regtype = "ll",
    bandwidth.compute = FALSE
  )
  bw.lp <- npregbw(
    xdat = tx,
    ydat = y,
    bws = 0.2,
    regtype = "lp",
    basis = "glp",
    degree = 1L,
    bandwidth.compute = FALSE
  )

  fit.ll <- npreg(
    txdat = tx,
    tydat = y,
    exdat = ex,
    bws = bw.ll,
    gradients = TRUE,
    gradient.order = 1L,
    warn.glp.gradient = FALSE
  )
  fit.lp <- npreg(
    txdat = tx,
    tydat = y,
    exdat = ex,
    bws = bw.lp,
    gradients = TRUE,
    gradient.order = 1L,
    warn.glp.gradient = FALSE
  )

  expect_equal(as.numeric(fit.ll$mean), as.numeric(fit.lp$mean), tolerance = 1e-10)
  expect_equal(as.numeric(fit.ll$grad), as.numeric(fit.lp$grad), tolerance = 1e-10)

  H0.ll <- npreghat(bws = bw.ll, txdat = tx, exdat = ex)
  H0.lp <- npreghat(bws = bw.lp, txdat = tx, exdat = ex)
  H1.ll <- npreghat(bws = bw.ll, txdat = tx, exdat = ex, s = 1L)
  H1.lp <- npreghat(bws = bw.lp, txdat = tx, exdat = ex, s = 1L)

  expect_equal(as.numeric(H0.ll %*% y), as.numeric(H0.lp %*% y), tolerance = 1e-10)
  expect_equal(as.numeric(H1.ll %*% y), as.numeric(H1.lp %*% y), tolerance = 1e-10)
})

test_that("npscoef cv and estimation match for ll and lp(degree=1) in 1D", {
  set.seed(20260227)
  n <- 80
  x <- runif(n)
  z <- runif(n)
  y <- (0.5 + sin(2 * pi * x)) * z + rnorm(n, sd = 0.04)

  set.seed(7701)
  bw.ll.cv <- npscoefbw(
    xdat = x,
    zdat = z,
    ydat = y,
    regtype = "ll",
    nmulti = 1L,
    cv.iterate = FALSE
  )
  set.seed(7701)
  bw.lp.cv <- npscoefbw(
    xdat = x,
    zdat = z,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = 1L,
    nmulti = 1L,
    cv.iterate = FALSE
  )

  expect_equal(as.numeric(bw.ll.cv$fval), as.numeric(bw.lp.cv$fval), tolerance = 1e-10)
  expect_equal(as.numeric(bw.ll.cv$bw), as.numeric(bw.lp.cv$bw), tolerance = 1e-9)

  bw.ll <- npscoefbw(
    xdat = x,
    zdat = z,
    ydat = y,
    regtype = "ll",
    bws = 0.2,
    bandwidth.compute = FALSE
  )
  bw.lp <- npscoefbw(
    xdat = x,
    zdat = z,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = 1L,
    bws = 0.2,
    bandwidth.compute = FALSE
  )

  fit.ll <- npscoef(
    bws = bw.ll,
    txdat = data.frame(x = x),
    tzdat = data.frame(z = z),
    tydat = y,
    errors = FALSE
  )
  fit.lp <- npscoef(
    bws = bw.lp,
    txdat = data.frame(x = x),
    tzdat = data.frame(z = z),
    tydat = y,
    errors = FALSE
  )

  expect_equal(as.numeric(fit.ll$mean), as.numeric(fit.lp$mean), tolerance = 1e-10)
})

test_that("npindex cv, fit, and npindexhat(s=1) match for ll and lp(degree=1) in 1D", {
  set.seed(20260227)
  n <- 100
  x <- runif(n)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.04)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(0, 1, length.out = 40))

  set.seed(8801)
  bw.ll.cv <- suppressWarnings(npindexbw(
    xdat = tx,
    ydat = y,
    regtype = "ll",
    method = "ichimura",
    nmulti = 1L
  ))
  set.seed(8801)
  bw.lp.cv <- suppressWarnings(npindexbw(
    xdat = tx,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = 1L,
    method = "ichimura",
    nmulti = 1L
  ))

  expect_equal(as.numeric(bw.ll.cv$fval), as.numeric(bw.lp.cv$fval), tolerance = 1e-10)
  expect_equal(as.numeric(bw.ll.cv$bw), as.numeric(bw.lp.cv$bw), tolerance = 1e-9)

  bw.ll <- suppressWarnings(npindexbw(
    xdat = tx,
    ydat = y,
    bws = c(1, 0.25),
    regtype = "ll",
    method = "ichimura",
    bandwidth.compute = FALSE
  ))
  bw.lp <- suppressWarnings(npindexbw(
    xdat = tx,
    ydat = y,
    bws = c(1, 0.25),
    regtype = "lp",
    basis = "glp",
    degree = 1L,
    method = "ichimura",
    bandwidth.compute = FALSE
  ))

  fit.ll <- npindex(bws = bw.ll, txdat = tx, tydat = y, exdat = ex)
  fit.lp <- npindex(bws = bw.lp, txdat = tx, tydat = y, exdat = ex)
  expect_equal(as.numeric(fit.ll$mean), as.numeric(fit.lp$mean), tolerance = 1e-10)

  H0.ll <- npindexhat(bws = bw.ll, txdat = tx)
  H0.lp <- npindexhat(bws = bw.lp, txdat = tx)
  H1.ll <- npindexhat(bws = bw.ll, txdat = tx, s = 1L)
  H1.lp <- npindexhat(bws = bw.lp, txdat = tx, s = 1L)
  expect_equal(as.numeric(H0.ll %*% y), as.numeric(H0.lp %*% y), tolerance = 1e-10)
  expect_equal(as.numeric(H1.ll %*% y), as.numeric(H1.lp %*% y), tolerance = 1e-10)
})

test_that("npplreg cv and estimation match for ll and lp(degree=1) in 1D", {
  set.seed(20260227)
  n <- 120
  z <- runif(n)
  x <- runif(n)
  y <- sin(2 * pi * z) + 1.2 * x + rnorm(n, sd = 0.04)

  set.seed(9901)
  bw.ll.cv <- npplregbw(
    xdat = data.frame(x = x),
    zdat = data.frame(z = z),
    ydat = y,
    regtype = "ll",
    nmulti = 1L
  )
  set.seed(9901)
  bw.lp.cv <- npplregbw(
    xdat = data.frame(x = x),
    zdat = data.frame(z = z),
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = 1L,
    nmulti = 1L
  )

  expect_equal(as.numeric(bw.ll.cv$fval), as.numeric(bw.lp.cv$fval), tolerance = 1e-10)
  expect_equal(as.numeric(bw.ll.cv$bw$yzbw$bw), as.numeric(bw.lp.cv$bw$yzbw$bw), tolerance = 1e-9)
  expect_equal(as.numeric(bw.ll.cv$bw[[2]]$bw), as.numeric(bw.lp.cv$bw[[2]]$bw), tolerance = 1e-9)

  bw.ll <- npplregbw(
    xdat = data.frame(x = x),
    zdat = data.frame(z = z),
    ydat = y,
    regtype = "ll",
    bws = matrix(c(0.2, 0.2), nrow = 2),
    bandwidth.compute = FALSE
  )
  bw.lp <- npplregbw(
    xdat = data.frame(x = x),
    zdat = data.frame(z = z),
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = 1L,
    bws = matrix(c(0.2, 0.2), nrow = 2),
    bandwidth.compute = FALSE
  )

  ex <- data.frame(x = runif(30))
  ez <- data.frame(z = runif(30))
  fit.ll <- npplreg(
    bws = bw.ll,
    txdat = data.frame(x = x),
    tzdat = data.frame(z = z),
    tydat = y,
    exdat = ex,
    ezdat = ez
  )
  fit.lp <- npplreg(
    bws = bw.lp,
    txdat = data.frame(x = x),
    tzdat = data.frame(z = z),
    tydat = y,
    exdat = ex,
    ezdat = ez
  )

  expect_equal(as.numeric(fit.ll$mean), as.numeric(fit.lp$mean), tolerance = 1e-10)
  expect_equal(as.numeric(fit.ll$xcoef), as.numeric(fit.lp$xcoef), tolerance = 1e-10)
})

test_that("npscoef multivariate cv and estimation match for ll and lp(degree=1)", {
  set.seed(20260306)
  n <- 65
  x1 <- rnorm(n)
  x2 <- rnorm(n)
  z1 <- runif(n, -1, 1)
  z2 <- runif(n, -1, 1)
  b1 <- 1 + 0.6 * z1 - 0.2 * z2
  b2 <- -0.5 + 0.4 * z1 + 0.3 * z2
  y <- b1 * x1 + b2 * x2 + rnorm(n, sd = 0.08)

  tx <- data.frame(x1 = x1, x2 = x2)
  tz <- data.frame(z1 = z1, z2 = z2)

  set.seed(7702)
  bw.ll.cv <- npscoefbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    regtype = "ll",
    nmulti = 1L,
    cv.iterate = FALSE,
    backfit.iterate = FALSE
  )
  set.seed(7702)
  bw.lp.cv <- npscoefbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = c(1L, 1L),
    nmulti = 1L,
    cv.iterate = FALSE,
    backfit.iterate = FALSE
  )

  expect_equal(as.numeric(bw.ll.cv$fval), as.numeric(bw.lp.cv$fval), tolerance = 1e-10)
  expect_equal(as.numeric(bw.ll.cv$bw), as.numeric(bw.lp.cv$bw), tolerance = 1e-9)
  expect_equal(as.numeric(bw.ll.cv$num.feval), as.numeric(bw.lp.cv$num.feval), tolerance = 0)

  bw.ll <- npscoefbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    regtype = "ll",
    bws = c(0.7, 0.7),
    bandwidth.compute = FALSE
  )
  bw.lp <- npscoefbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = c(1L, 1L),
    bws = c(0.7, 0.7),
    bandwidth.compute = FALSE
  )

  fit.ll <- npscoef(bws = bw.ll, txdat = tx, tzdat = tz, tydat = y, errors = FALSE)
  fit.lp <- npscoef(bws = bw.lp, txdat = tx, tzdat = tz, tydat = y, errors = FALSE)

  expect_equal(as.numeric(fit.ll$mean), as.numeric(fit.lp$mean), tolerance = 1e-10)
  expect_equal(as.numeric(fit.ll$beta), as.numeric(fit.lp$beta), tolerance = 1e-10)
})

test_that("npindex multivariate cv and estimation match for ll and lp(degree=1)", {
  n <- 75

  run_case <- function(method) {
    set.seed(if (identical(method, "ichimura")) 20260307 else 20260308)
    x1 <- runif(n, -1, 1)
    x2 <- runif(n, -1, 1)
    x3 <- runif(n, -1, 1)
    eta <- x1 + 0.5 * x2 - 0.4 * x3
    y <- if (identical(method, "ichimura")) {
      eta + 0.3 * eta^2 + rnorm(n, sd = 0.1)
    } else {
      rbinom(n, size = 1L, prob = plogis(eta))
    }
    tx <- data.frame(x1 = x1, x2 = x2, x3 = x3)

    set.seed(8802)
    bw.ll.cv <- suppressWarnings(npindexbw(
      xdat = tx,
      ydat = y,
      regtype = "ll",
      method = method,
      nmulti = 1L
    ))
    set.seed(8802)
    bw.lp.cv <- suppressWarnings(npindexbw(
      xdat = tx,
      ydat = y,
      regtype = "lp",
      basis = "glp",
      degree = 1L,
      method = method,
      nmulti = 1L
    ))

    expect_equal(as.numeric(bw.ll.cv$fval), as.numeric(bw.lp.cv$fval), tolerance = 1e-10)
    expect_equal(
      c(as.numeric(bw.ll.cv$beta), as.numeric(bw.ll.cv$bw)),
      c(as.numeric(bw.lp.cv$beta), as.numeric(bw.lp.cv$bw)),
      tolerance = 1e-9
    )
    expect_equal(as.numeric(bw.ll.cv$num.feval), as.numeric(bw.lp.cv$num.feval), tolerance = 0)

    bws.fixed <- c(1, 0.3, -0.2, 0.6)
    bw.ll <- suppressWarnings(npindexbw(
      xdat = tx,
      ydat = y,
      bws = bws.fixed,
      regtype = "ll",
      method = method,
      bandwidth.compute = FALSE
    ))
    bw.lp <- suppressWarnings(npindexbw(
      xdat = tx,
      ydat = y,
      bws = bws.fixed,
      regtype = "lp",
      basis = "glp",
      degree = 1L,
      method = method,
      bandwidth.compute = FALSE
    ))

    fit.ll <- npindex(bws = bw.ll, txdat = tx, tydat = y)
    fit.lp <- npindex(bws = bw.lp, txdat = tx, tydat = y)
    expect_equal(as.numeric(fit.ll$mean), as.numeric(fit.lp$mean), tolerance = 1e-10)
    expect_equal(as.numeric(fit.ll$beta), as.numeric(fit.lp$beta), tolerance = 1e-10)
  }

  run_case("ichimura")
  run_case("kleinspady")
})

test_that("npplreg multivariate cv and estimation match for ll and lp(degree=1)", {
  set.seed(20260309)
  n <- 70
  z1 <- runif(n, -1, 1)
  z2 <- runif(n, -1, 1)
  x1 <- 0.8 * z1 + rnorm(n, sd = 0.2)
  x2 <- -0.6 * z2 + rnorm(n, sd = 0.2)
  y <- 1 + 1.2 * x1 - 0.7 * x2 + 0.4 * z1^2 - 0.3 * z1 * z2 + rnorm(n, sd = 0.1)

  tx <- data.frame(x1 = x1, x2 = x2)
  tz <- data.frame(z1 = z1, z2 = z2)

  set.seed(9902)
  bw.ll.cv <- npplregbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    regtype = "ll",
    nmulti = 1L
  )
  set.seed(9902)
  bw.lp.cv <- npplregbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = c(1L, 1L),
    nmulti = 1L
  )

  expect_equal(as.numeric(bw.ll.cv$fval), as.numeric(bw.lp.cv$fval), tolerance = 1e-10)
  expect_equal(as.numeric(bw.ll.cv$bw$yzbw$bw), as.numeric(bw.lp.cv$bw$yzbw$bw), tolerance = 1e-9)
  expect_equal(as.numeric(bw.ll.cv$bw[[2L]]$bw), as.numeric(bw.lp.cv$bw[[2L]]$bw), tolerance = 1e-9)
  expect_equal(as.numeric(bw.ll.cv$bw[[3L]]$bw), as.numeric(bw.lp.cv$bw[[3L]]$bw), tolerance = 1e-9)
  expect_equal(as.numeric(bw.ll.cv$num.feval), as.numeric(bw.lp.cv$num.feval), tolerance = 0)

  bws.fixed <- matrix(c(0.3, 0.3, 0.3, 0.3, 0.3, 0.3), nrow = 3L)
  bw.ll <- npplregbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    regtype = "ll",
    bws = bws.fixed,
    bandwidth.compute = FALSE
  )
  bw.lp <- npplregbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = c(1L, 1L),
    bws = bws.fixed,
    bandwidth.compute = FALSE
  )

  fit.ll <- npplreg(bws = bw.ll, txdat = tx, tzdat = tz, tydat = y)
  fit.lp <- npplreg(bws = bw.lp, txdat = tx, tzdat = tz, tydat = y)

  expect_equal(as.numeric(fit.ll$mean), as.numeric(fit.lp$mean), tolerance = 1e-10)
  expect_equal(as.numeric(fit.ll$xcoef), as.numeric(fit.lp$xcoef), tolerance = 1e-10)
})

test_that("npplreg generalized-nn semiparametric hat invariants match for ll and canonical lp degree-1", {
  set.seed(20260308)
  n <- 80
  z <- runif(n)
  x <- 0.7 * z + rnorm(n, sd = 0.2)
  y <- 1 + 1.1 * x + sin(2 * pi * z) + rnorm(n, sd = 0.05)

  tx <- data.frame(x = x)
  tz <- data.frame(z = z)
  bws.fixed <- matrix(c(2, 9), nrow = 2L)

  bw.ll <- npplregbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    regtype = "ll",
    bwtype = "generalized_nn",
    bws = bws.fixed,
    bandwidth.compute = FALSE
  )
  bw.lp <- npplregbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = 1L,
    bernstein.basis = FALSE,
    bwtype = "generalized_nn",
    bws = bws.fixed,
    bandwidth.compute = FALSE
  )

  fit.ll <- npplreg(bws = bw.ll, txdat = tx, tzdat = tz, tydat = y)
  fit.lp <- npplreg(bws = bw.lp, txdat = tx, tzdat = tz, tydat = y)
  hat.apply.ll <- npplreghat(bws = bw.ll, txdat = tx, tzdat = tz, y = y, output = "apply")
  hat.apply.lp <- npplreghat(bws = bw.lp, txdat = tx, tzdat = tz, y = y, output = "apply")
  hat.matrix.ll <- npplreghat(bws = bw.ll, txdat = tx, tzdat = tz, output = "matrix")
  hat.matrix.lp <- npplreghat(bws = bw.lp, txdat = tx, tzdat = tz, output = "matrix")

  expect_equal(as.numeric(fit.ll$mean), as.numeric(fit.lp$mean), tolerance = 1e-10)
  expect_equal(as.numeric(fit.ll$xcoef), as.numeric(fit.lp$xcoef), tolerance = 1e-10)
  expect_equal(as.numeric(hat.apply.ll), as.numeric(hat.apply.lp), tolerance = 1e-10)
  expect_equal(as.numeric(hat.apply.ll), as.numeric(hat.matrix.ll %*% y), tolerance = 1e-10)
  expect_equal(as.numeric(hat.apply.lp), as.numeric(hat.matrix.lp %*% y), tolerance = 1e-10)
})

test_that("npindex inid fast path matches explicit resample refits", {
  skip_if_not_installed("np")

  set.seed(3231)
  n <- 40
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.1)
  tx <- data.frame(x1 = x1, x2 = x2)
  bw <- npindexbw(xdat = tx, ydat = y, method = "ichimura", nmulti = 1)
  B <- 11L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  fast.fun <- getFromNamespace(".np_inid_lc_boot_from_hat", "np")
  H <- npindexhat(bws = bw, txdat = tx, exdat = tx, output = "matrix", s = 0L)
  fast.out <- fast.fun(H = H, ydat = y, B = B, counts = counts)

  explicit.t <- matrix(NA_real_, nrow = B, ncol = n)
  for (b in seq_len(B)) {
    idx <- rep.int(seq_len(n), counts[, b])
    explicit.t[b, ] <- npindex(
      txdat = tx[idx, , drop = FALSE],
      tydat = y[idx],
      exdat = tx,
      bws = bw,
      gradients = FALSE
    )$mean
  }

  expect_equal(fast.out$t, explicit.t, tolerance = 1e-10)
  expect_equal(fast.out$t0, as.vector(H %*% y), tolerance = 1e-12)
})

test_that("npindex ll/lp inid fast path matches explicit resample refits", {
  skip_if_not_installed("np")

  set.seed(32315)
  n <- 45
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.1)
  tx <- data.frame(x1 = x1, x2 = x2)
  B <- 9L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  fast.fun <- getFromNamespace(".np_inid_boot_from_regression", "np")
  rbw.fun <- getFromNamespace(".np_indexhat_rbw", "np")

  cfgs <- list(
    list(regtype = "ll", basis = NULL, label = "ll"),
    list(regtype = "lp", basis = "additive", label = "lp-additive"),
    list(regtype = "lp", basis = "tensor", label = "lp-tensor"),
    list(regtype = "lp", basis = "glp", label = "lp-glp")
  )

  for (cfg in cfgs) {
    bw.args <- list(
      xdat = tx,
      ydat = y,
      bws = c(1, 1, 0.25),
      bandwidth.compute = FALSE,
      regtype = cfg$regtype
    )
    if (!is.null(cfg$basis)) {
      bw.args$basis <- cfg$basis
      bw.args$degree <- 2L
    }
    bw <- do.call(npindexbw, bw.args)

    tx.index <- data.frame(index = as.vector(as.matrix(tx) %*% bw$beta))
    rbw <- rbw.fun(bws = bw, idx.train = tx.index)

    fast.out <- fast.fun(
      xdat = tx.index,
      exdat = tx.index,
      bws = rbw,
      ydat = y,
      B = B,
      counts = counts
    )

    explicit.t <- matrix(NA_real_, nrow = B, ncol = n)
    for (b in seq_len(B)) {
      idx <- rep.int(seq_len(n), counts[, b])
      explicit.t[b, ] <- npindex(
        txdat = tx[idx, , drop = FALSE],
        tydat = y[idx],
        exdat = tx,
        bws = bw,
        gradients = FALSE
      )$mean
    }

    expect_equal(fast.out$t, explicit.t, tolerance = 1e-6, info = cfg$label)
    expect_equal(
      fast.out$t0,
      npindex(txdat = tx, tydat = y, exdat = tx, bws = bw, gradients = FALSE)$mean,
      tolerance = 1e-6,
      info = cfg$label
    )
  }
})

test_that("npindex plot bootstrap inid supports ll/lp basis variants", {
  skip_if_not_installed("np")

  set.seed(3232)
  n <- 70
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.1)
  tx <- data.frame(x1 = x1, x2 = x2)

  cfgs <- list(
    list(regtype = "ll", basis = NULL, label = "ll"),
    list(regtype = "lp", basis = "glp", label = "lp-glp"),
    list(regtype = "lp", basis = "additive", label = "lp-additive"),
    list(regtype = "lp", basis = "tensor", label = "lp-tensor")
  )

  for (cfg in cfgs) {
    bw.args <- list(
      xdat = tx,
      ydat = y,
      bws = c(1, 1, 0.25),
      bandwidth.compute = FALSE,
      regtype = cfg$regtype
    )
    if (!is.null(cfg$basis)) {
      bw.args$basis <- cfg$basis
      bw.args$degree <- 2L
    }
    bw <- do.call(npindexbw, bw.args)
    out <- suppressWarnings(
      plot(
        bw,
        xdat = tx,
        ydat = y,
        output = "data",
        perspective = FALSE,
        errors = "bootstrap",
        bootstrap = "inid",
        B = 7
      )
    )
    expect_type(out, "list")
    expect_true(length(out) > 0, info = cfg$label)
  }
})

test_that("npindex plot bootstrap inid fails fast for unsupported nonfixed gradients", {
  skip_if_not_installed("np")

  set.seed(32321)
  n <- 60
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.1)
  tx <- data.frame(x1 = x1, x2 = x2)

  for (bt in c("generalized_nn", "adaptive_nn")) {
    bw <- npindexbw(
      xdat = tx,
      ydat = y,
      bws = c(1, 1, 5L),
      bandwidth.compute = FALSE,
      bwtype = bt
    )

    expect_error(
      suppressWarnings(
        plot(
          bw,
          xdat = tx,
          ydat = y,
          output = "data",
          perspective = FALSE,
          gradients = TRUE,
          errors = "bootstrap",
          bootstrap = "inid",
          B = 7
        )
      ),
      "gradients=FALSE",
      info = bt
    )
  }
})

test_that("npindex bounded bootstrap plot-data supports bw and fit objects", {
  skip_if_not_installed("np")

  set.seed(32322)
  n <- 36
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.08)
  tx <- data.frame(x1 = x1, x2 = x2)

  bw <- npindexbw(
    xdat = tx,
    ydat = y,
    method = "ichimura",
    ckerbound = "range",
    bws = c(1, 0.25, 0.25),
    bandwidth.compute = FALSE
  )
  fit <- npindex(bws = bw, txdat = tx, tydat = y)

  run_plot <- function(obj, boot.method) {
    suppressWarnings(
      plot(
        obj,
        xdat = tx,
        ydat = y,
        output = "data",
        perspective = FALSE,
        errors = "bootstrap",
        bootstrap = boot.method,
        B = 5L,
        band = "pointwise"
      )
    )
  }

  for (boot.method in c("inid", "geom", "wild")) {
    out.bw <- run_plot(bw, boot.method)
    out.fit <- run_plot(fit, boot.method)

    expect_type(out.bw, "list")
    expect_type(out.fit, "list")
    expect_true(length(out.bw) > 0L, info = paste("bw", boot.method))
    expect_true(length(out.fit) > 0L, info = paste("fit", boot.method))
  }
})

## Function that implements the test for asymmetry applicable to both
## real-values and categorical univariate data described in Maasoumi,
## E. and J.S. Racine (2009), "A Robust Entropy-Based Test of
## Asymmetry for Discrete and Continuous Processes," Econometric
## Reviews, Volume 28, pp 246-261.

npsymtest <- function(data = NULL,
                      method = c("integration","summation"),
                      boot.num = 399,
                      bw = NULL,
                      boot.method = c("iid", "geom"),
                      random.seed = 42,
                      ...) {

  if(is.data.frame(data)) stop(" you must enter a data vector (not data frame)")
  if(is.null(data)) stop(" you must enter a data vector")
  if(ncol(data.frame(data)) != 1) stop(" data must have one dimension only")
  if(boot.num < 9) stop(" number of bootstrap replications must be >= 9")

  boot.method <- match.arg(boot.method)
  method <- match.arg(method)

  ## Save seed prior to setting

  seed.state <- .np_seed_enter(random.seed)
  .np_progress_note("Computing bandwidths")

  ## If of type ts convert to numeric to handle time series data

  if(is.ts(data)) data <- as.numeric(data)

  ## Remove NAs

  data <- na.omit(data)

  ## Note - this function accepts numeric and factors/ordered,
  ## however, factors must be integers.

  if(is.numeric(data)) {
    ## Numerical stability - integrate does best centered around zero,
    ## so standardize numeric data - this affects nothing else.
    data <- (data-mean(data))/sd(data)
    data.rotate <- -(data-mean(data))+mean(data)
    if(is.null(bw)) bw <- bw.SJ(data)
  } else {
    if(is.ordered(data)) {
      ## Rotate around the median, ordered case.
      data.levels <- levels(data)
      tmp <- as.numeric(data.matrix(data))
      location <- median(sort(unique(tmp)))
      data.rotate <- ordered(-(tmp-location)+location,levels=data.levels)
    } else {
      ## unordered case.
      data.levels <- levels(data)
      tmp <- as.numeric(data.matrix(data))
      location <- median(sort(unique(tmp)))
      data.rotate <- factor(-(tmp-location)+location,levels=data.levels)
    }
    ## Optimal bandwidth for factor (unordered) used for both
    ## (plug-in).
    if(is.null(bw)) {
      n <- length(data)
      c <- length(unique(data))
      if(c <= 1) stop("data must contain at least two distinct factor levels")
      xeval <- unique(data)
      p <- fitted(npudens(tdat=data,edat=xeval,bws=0,...))
      sum.Lambda3 <- c/(c-1)*sum(p*(1-p))
      sum.Lambda2.minus.Lambda1.sq <- c^2/((c-1)^2)*sum(p*(1-p))
      n.sum.Lambda1.sq <- n*sum(((1-c*p)/(c-1))^2)
      bw <- sum.Lambda3/(sum.Lambda2.minus.Lambda1.sq+n.sum.Lambda1.sq)
    }
  }
	
  ##  Impose the null and generate a data of resampled statistics

  if(is.numeric(data)) {
    data.null <- c(data,data.rotate)
  } else {
    if(is.ordered(data)) {
      data.null <- ordered(c(as.character(data),as.character(data.rotate)))
    } else {
      data.null <- factor(c(as.character(data),as.character(data.rotate)))
    }
  }

  ## Hellinger distance (Srho) between the densities for the data and
  ## the rotated data. This function accepts numeric and
  ## factor/ordered.
  
  Srho.sym <- function(data,
                       data.rotate,
                       bw,
                       method=c("integration","summation")) {
    
    if(ncol(data.frame(data)) != 1)  stop(" data must have one dimension only") 

    if(is.numeric(data)) {
      if(method=="summation") {
        ## Summation version uses
        ##  0.5\sum_i(1-sqrt(f.data.rotate/f.data)^2). Note that here
        ##  we must evaluate on common sample, so we use `data' for
        ##  the evaluation points. We could also combine both data and
        ##  data.rotate for the evaluation points if we chose (would
        ##  this improve power?).
        f.data <- fitted(npudens(tdat=data,edat=data,bws=bw,...))
        f.data.rotate <- fitted(npudens(tdat=data.rotate,edat=data,bws=bw,...))
        ## In summation version we divide densities which can lead to
        ## numerical instability issues not present in the integrand
        ## version (which uses differences instead). We check for this
        ## case, remove offending points, and warn. This traps -Inf,
        ## Inf, and NaN.
        summand <- f.data.rotate/f.data
        if(!all(is.finite(summand))) {
          .np_warning(" non-finite value in summation-based statistic: integration recommended")
          summand <- summand[is.finite(summand)]
        }
        return(0.5*mean((1-sqrt(summand))**2))
      } else {
        ## Integration version
        h <- function(x,data,data.rotate) {
          f.data <- fitted(npudens(tdat=data,edat=x,bws=bw,...))
          f.data.rotate <- fitted(npudens(tdat=data.rotate,edat=x,bws=bw,...))
          return(0.5*(sqrt(f.data)-sqrt(f.data.rotate))**2)
        }
        return.integrate <- integrate(h,-Inf,Inf,subdivisions=1e+05,stop.on.error=FALSE,data=data,data.rotate=data.rotate)
        if(return.integrate$message != "OK") .np_warning(return.integrate$message)
        return(return.integrate$value)
      }
    } else {
      xeval <- unique(data)
      p.data <- fitted(npudens(tdat=data,edat=xeval,bws=bw,...))
      p.data.rotate <- fitted(npudens(tdat=data.rotate,edat=xeval,bws=bw,...))   
      ## Sum the information function over the unique probabilities and
      ## return.
      return(sum(0.5*(sqrt(p.data)-sqrt(p.data.rotate))**2))
    }
  }
  
  ## Compute the test statistic

  test.stat <- Srho.sym(data,data.rotate,bw,method=method)

  ## Function to be fed to boot - accepts data that gets
  ## permuted/rearranged to define resampled data. The sole difference
  ## between boot.fun.boot and boot.fun.tsboot is the order of
  ## arguments.

  boot.state <- new.env(parent = emptyenv())
  boot.state$counter <- 0L
  boot.state$progress <- .np_progress_begin("Bootstrap replications", total = boot.num, surface = "bootstrap")

  ## Function to be fed to tsboot - accepts a vector of integers
  ## corresponding to all observations in the sample (1,2,...) that
  ## get permuted/rearranged to define resampled data.

	boot.fun <- function(ii,data.null,bw) {
    boot.state$counter <- boot.state$counter + 1L
    boot.state$progress <- .np_progress_step(boot.state$progress, done = boot.state$counter)
    null.sample1 <- data.null[ii]
    if(is.numeric(data.null)) {
      null.sample2 <- -(null.sample1-mean(null.sample1))+mean(null.sample1)
    } else {
      if(is.ordered(data)) {
        ## Rotate around the median, ordered case
        null.sample1.levels <- levels(null.sample1)
        tmp <- as.numeric(data.matrix(null.sample1))
        location <- median(sort(unique(tmp)))
        null.sample2 <- ordered(-(tmp-location)+location,levels=data.levels)
      } else {
        ## unordered case
        null.sample1.levels <- levels(null.sample1)
        tmp <- as.numeric(data.matrix(null.sample1))
        location <- median(sort(unique(tmp)))
        null.sample2 <- factor(-(tmp-location)+location,levels=data.levels)
      }
    }
    return(Srho.sym(null.sample1,null.sample2,bw,method=method))
	}

  ## Need to bootstrap integers for data.null to accommodate both
  ## numeric and factor/ordered.

  tseries.idx <- seq_len(2L * length(data))

  if(boot.method == "iid")  {

    ## data.null is of length 2*n - if we use boot() we automatically
    ## get resamples of length 2*n - however, we want resamples from
    ## data.null of length n. Solution here is to use the block
    ## bootstrap with a block length of 1 which is presumed to
    ## generate an iid bootstrap.

    resampled.stat <- tsboot(tseries = tseries.idx,
                             statistic = boot.fun,
                             R = boot.num,
                             n.sim = length(data),
                             l = 1,
                             sim = "fixed",
                             data.null = data.null,
                             bw = bw)$t

  } else {

    if(is.numeric(data)) {
      ## Optimal block length is based upon `data' (not data.null).
      boot.blocklen <- b.star(data,round=TRUE)[1,1]
    } else {
      boot.blocklen <- b.star(as.numeric(data.matrix(data)),round=TRUE)[1,1]
    }

    resampled.stat <- tsboot(tseries = tseries.idx,
                             statistic = boot.fun,
                             R = boot.num,
                             n.sim = length(data),
                             l = boot.blocklen,
                             sim = "geom",
                             data.null = data.null,
                             bw = bw)$t

  }

  boot.state$progress <- .np_progress_end(boot.state$progress)

  p.value <- mean(resampled.stat > test.stat)

  ## Restore seed

  .np_seed_exit(seed.state)
  
  symtest(Srho = test.stat,
          Srho.bootstrap = resampled.stat,
          P = p.value,
          boot.num = boot.num,
          data.rotate = data.rotate,
          bw = bw)

}

library(np)

quiet_eval <- function(expr) {
  value <- NULL
  capture.output(value <- force(expr))
  value
}

test_that("npplreg exact helper matches duplicate-row oracle and adaptive frozen stays finite", {
  boot.fun <- getFromNamespace(".np_inid_boot_from_plreg", "np")

  txdat <- data.frame(x = c(0.05, 0.05, 0.25, 0.25, 0.60, 0.60, 0.90, 0.90))
  tzdat <- data.frame(z = c(0.10, 0.10, 0.35, 0.35, 0.65, 0.65, 0.95, 0.95))
  y <- with(txdat, 0.5 + 0.8 * x + sin(2 * pi * tzdat$z))
  exdat <- data.frame(x = c(0.10, 0.30, 0.70))
  ezdat <- data.frame(z = c(0.12, 0.40, 0.85))
  counts <- cbind(
    c(2, 0, 2, 0, 1, 1, 1, 1),
    c(0, 2, 0, 2, 1, 1, 1, 1),
    c(1, 1, 3, 0, 0, 2, 1, 0)
  )

  bw <- npplregbw(
    xdat = txdat,
    zdat = tzdat,
    ydat = y,
    bws = matrix(c(3L, 3L), nrow = 2L, ncol = 1L),
    bwtype = "adaptive_nn",
    bandwidth.compute = FALSE,
    regtype = "ll"
  )

  exact.out <- boot.fun(
    txdat = txdat,
    ydat = y,
    tzdat = tzdat,
    exdat = exdat,
    ezdat = ezdat,
    bws = bw,
    B = ncol(counts),
    counts = counts,
    mode = "exact"
  )
  frozen.out <- boot.fun(
    txdat = txdat,
    ydat = y,
    tzdat = tzdat,
    exdat = exdat,
    ezdat = ezdat,
    bws = bw,
    B = ncol(counts),
    counts = counts,
    mode = "frozen"
  )

  exact.oracle <- vapply(seq_len(ncol(counts)), function(j) {
    idx <- np:::.np_counts_to_indices(counts[, j])
    as.vector(npplreg(
      bws = bw,
      txdat = txdat[idx, , drop = FALSE],
      tydat = y[idx],
      tzdat = tzdat[idx, , drop = FALSE],
      exdat = exdat,
      ezdat = ezdat
    )$mean)
  }, numeric(nrow(exdat)))
  exact.oracle <- t(exact.oracle)

  t0.oracle <- as.vector(npplreg(
    bws = bw,
    txdat = txdat,
    tydat = y,
    tzdat = tzdat,
    exdat = exdat,
    ezdat = ezdat
  )$mean)

  expect_equal(exact.out$t, exact.oracle, tolerance = 1e-10)
  expect_equal(as.vector(exact.out$t0), t0.oracle, tolerance = 1e-10)
  expect_equal(as.vector(frozen.out$t0), t0.oracle, tolerance = 1e-10)
  expect_true(all(is.finite(frozen.out$t)))
  expect_lt(max(abs(frozen.out$t)), 10 * max(abs(exact.out$t)) + 1e-8)
  expect_false(isTRUE(all.equal(
    exact.out$t,
    frozen.out$t,
    tolerance = 1e-10,
    check.attributes = FALSE
  )))
})

test_that("npplreg bootstrap plot payload stays finite across frozen nonfixed bw types", {
  set.seed(42)
  n <- 100
  x <- runif(n, -1, 1)
  z <- rnorm(n)
  y <- x + rnorm(n, sd = 0.25 * sd(x))

  for (bwtype in c("fixed", "generalized_nn", "adaptive_nn")) {
    fit <- npplreg(y ~ x | z, nmulti = 1, bwtype = bwtype)

    for (mode in c("exact", "frozen")) {
      out <- quiet_eval(plot(
        fit,
        errors = "bootstrap",
        view = "fixed",
        bootstrap = "geom",
        neval = 12,
        B = 61,
        band = "pointwise",
        boot_control = np_boot_control(nonfixed = mode),
        output = "data"
      ))

      expect_true(all(is.finite(out$r1$merr)), info = paste(bwtype, mode))
      expect_true(max(abs(out$r1$merr)) < 1e4, info = paste(bwtype, mode))
    }
  }
})

test_that("npplreg frozen bootstrap mode is forwarded into the helper", {
  txdat <- data.frame(x = c(0.05, 0.05, 0.25, 0.25, 0.60, 0.60, 0.90, 0.90))
  tzdat <- data.frame(z = c(0.10, 0.10, 0.35, 0.35, 0.65, 0.65, 0.95, 0.95))
  y <- with(txdat, 0.5 + 0.8 * x + sin(2 * pi * tzdat$z))

  bw <- npplregbw(
    xdat = txdat,
    zdat = tzdat,
    ydat = y,
    bws = matrix(c(3L, 3L), nrow = 2L, ncol = 1L),
    bwtype = "adaptive_nn",
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  ns <- asNamespace("np")
  orig <- getFromNamespace(".np_inid_boot_from_plreg", "np")
  modes <- character()
  assignInNamespace(".np_inid_boot_from_plreg", function(..., mode = c("exact", "frozen")) {
    mode <- match.arg(mode)
    modes <<- c(modes, mode)
    orig(..., mode = mode)
  }, ns = "np")
  on.exit(assignInNamespace(".np_inid_boot_from_plreg", orig, ns = "np"), add = TRUE)

  compute.boot <- getFromNamespace("compute.bootstrap.errors.plbandwidth", "np")

  expect_no_error(compute.boot(
    xdat = txdat,
    ydat = y,
    zdat = tzdat,
    exdat = data.frame(x = seq(0.10, 0.80, length.out = 6L)),
    ezdat = data.frame(z = seq(0.12, 0.92, length.out = 6L)),
    gradients = FALSE,
    slice.index = 0L,
    progress.target = NULL,
    plot.errors.boot.method = "inid",
    plot.errors.boot.nonfixed = "frozen",
    plot.errors.boot.num = 41L,
    plot.errors.center = "estimate",
    plot.errors.type = "pointwise",
    plot.errors.alpha = 0.05,
    bws = bw
  ))

  expect_gte(length(modes), 1L)
  expect_true(all(modes == "frozen"))
})

test_that("npcmstest basic functionality works", {
  set.seed(42)
  n <- 50
  x <- rnorm(n)
  y <- 1 + x + rnorm(n, sd=0.1)
  # Parametric model - need x=TRUE, y=TRUE for npcmstest
  model <- lm(y~x, x=TRUE, y=TRUE)
  
  # Asymptotic for speed
  test <- npcmstest(model=model, xdat=x, ydat=y, distribution="asymptotic")
  
  expect_s3_class(test, "cmstest")
  expect_output(summary(test))
})

test_that("npqcmstest basic functionality works", {
  library(quantreg)
  set.seed(42)
  n <- 50
  x <- rnorm(n)
  y <- 1 + x + rnorm(n, sd=0.1)
  # npqcmstest needs an rq model with model=TRUE
  model <- rq(y~x, tau=0.5, model=TRUE)
  
  # Asymptotic for speed
  test <- npqcmstest(model=model, xdat=x, ydat=y, distribution="asymptotic")
  
  expect_s3_class(test, "cmstest") 
})

test_that("npdeneqtest basic functionality works", {
  set.seed(42)
  n <- 50
  x <- data.frame(v1=rnorm(n))
  y <- data.frame(v1=rnorm(n, mean=0.5))
  
  # Use small boot.num
  test <- npdeneqtest(x, y, boot.num=9)
  
  expect_s3_class(test, "deneqtest")
  expect_output(summary(test))
})

test_that("npsymtest basic functionality works", {
  set.seed(42)
  n <- 50
  x <- rgamma(n, shape=2)
  
  test <- npsymtest(x, method="summation", boot.num=9)
  
  expect_s3_class(test, "symtest")
  expect_output(summary(test))
})

test_that("npunitest basic functionality works", {
  set.seed(42)
  n <- 50
  x <- rnorm(n)
  y <- rnorm(n, mean=0.5)
  
  test <- npunitest(x, y, method="summation", boot.num=9)
  
  expect_s3_class(test, "unitest")
  expect_output(summary(test))
})

library(np)

quiet_eval <- function(expr) {
  value <- NULL
  capture.output(value <- force(expr))
  value
}

test_that("npscoef exact helper matches duplicate-row oracle and frozen stays finite", {
  boot.fun <- getFromNamespace(".np_inid_boot_from_scoef", "np")

  xdat <- data.frame(x = c(0.05, 0.05, 0.25, 0.25, 0.60, 0.60, 0.90, 0.90))
  zdat <- data.frame(z = c(0.10, 0.10, 0.35, 0.35, 0.65, 0.65, 0.95, 0.95))
  y <- with(xdat, 1 + x * (1 + zdat$z))
  exdat <- data.frame(x = c(0.10, 0.30, 0.70))
  ezdat <- data.frame(z = c(0.12, 0.40, 0.85))
  counts <- cbind(
    c(2, 0, 2, 0, 1, 1, 1, 1),
    c(0, 2, 0, 2, 1, 1, 1, 1),
    c(1, 1, 3, 0, 0, 2, 1, 0)
  )

  bw <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    bws = 3L,
    bwtype = "adaptive_nn",
    bandwidth.compute = FALSE,
    regtype = "ll"
  )

  exact.out <- boot.fun(
    txdat = xdat,
    ydat = y,
    tzdat = zdat,
    exdat = exdat,
    ezdat = ezdat,
    bws = bw,
    B = ncol(counts),
    counts = counts,
    mode = "exact"
  )
  frozen.out <- boot.fun(
    txdat = xdat,
    ydat = y,
    tzdat = zdat,
    exdat = exdat,
    ezdat = ezdat,
    bws = bw,
    B = ncol(counts),
    counts = counts,
    mode = "frozen"
  )

  exact.oracle <- vapply(seq_len(ncol(counts)), function(j) {
    idx <- np:::.np_counts_to_indices(counts[, j])
    as.vector(npscoef(
      bws = bw,
      txdat = xdat[idx, , drop = FALSE],
      tzdat = zdat[idx, , drop = FALSE],
      tydat = y[idx],
      exdat = exdat,
      ezdat = ezdat,
      iterate = FALSE,
      errors = FALSE
    )$mean)
  }, numeric(nrow(exdat)))
  exact.oracle <- t(exact.oracle)

  t0.oracle <- as.vector(npscoef(
    bws = bw,
    txdat = xdat,
    tzdat = zdat,
    tydat = y,
    exdat = exdat,
    ezdat = ezdat,
    iterate = FALSE,
    errors = FALSE
  )$mean)

  expect_equal(exact.out$t, exact.oracle, tolerance = 1e-10)
  expect_equal(as.vector(exact.out$t0), t0.oracle, tolerance = 1e-10)
  expect_equal(as.vector(frozen.out$t0), t0.oracle, tolerance = 1e-10)
  expect_true(all(is.finite(frozen.out$t)))
  expect_true(max(abs(frozen.out$t)) < 1e4)
})

test_that("npscoef frozen surface plot mode is forwarded", {
  xdat <- data.frame(x = c(0.05, 0.05, 0.25, 0.25, 0.60, 0.60, 0.90, 0.90))
  zdat <- data.frame(z = c(0.10, 0.10, 0.35, 0.35, 0.65, 0.65, 0.95, 0.95))
  y <- with(xdat, 1 + x * (1 + zdat$z))

  bw <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    bws = 3L,
    bwtype = "adaptive_nn",
    bandwidth.compute = FALSE,
    regtype = "ll"
  )
  ns <- asNamespace("np")
  orig <- getFromNamespace(".np_inid_boot_from_scoef", "np")
  modes <- character()
  assignInNamespace(".np_inid_boot_from_scoef", function(..., mode = c("exact", "frozen")) {
    mode <- match.arg(mode)
    modes <<- c(modes, mode)
    orig(..., mode = mode)
  }, ns = "np")
  on.exit(assignInNamespace(".np_inid_boot_from_scoef", orig, ns = "np"), add = TRUE)

  expect_no_error(capture.output(plot(
    bw,
    xdat = xdat,
    ydat = y,
    zdat = zdat,
    neval = 6L,
    coef = FALSE,
    output = "data",
    errors = "bootstrap",
    bootstrap = "inid",
    boot_control = np_boot_control(nonfixed = "frozen"),
    B = 41L,
    band = "pointwise"
  )))

  expect_gte(length(modes), 1L)
  expect_true(all(modes == "frozen"))
})

test_that("npscoef fixed helper treats exact and frozen identically", {
  boot.fun <- getFromNamespace(".np_inid_boot_from_scoef", "np")

  set.seed(42)
  n <- 30L
  xdat <- data.frame(x = runif(n, -1, 1))
  zdat <- data.frame(z = rnorm(n))
  y <- with(xdat, x^2 + rnorm(n, sd = 0.1))
  exdat <- data.frame(x = seq(-0.9, 0.9, length.out = 9L))
  ezdat <- data.frame(z = seq(-1.0, 1.0, length.out = 9L))
  counts <- rmultinom(n = 5L, size = n, prob = rep.int(1 / n, n))

  bw <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    bws = c(0.6),
    bwtype = "fixed",
    bandwidth.compute = FALSE,
    regtype = "ll"
  )

  exact.out <- boot.fun(
    txdat = xdat,
    ydat = y,
    tzdat = zdat,
    exdat = exdat,
    ezdat = ezdat,
    bws = bw,
    B = ncol(counts),
    counts = counts,
    mode = "exact"
  )
  frozen.out <- boot.fun(
    txdat = xdat,
    ydat = y,
    tzdat = zdat,
    exdat = exdat,
    ezdat = ezdat,
    bws = bw,
    B = ncol(counts),
    counts = counts,
    mode = "frozen"
  )

  expect_equal(exact.out$t0, frozen.out$t0, tolerance = 1e-12)
  expect_equal(exact.out$t, frozen.out$t, tolerance = 1e-12)
})

test_that("npscoef fixed lp helper matches duplicate-row oracle", {
  boot.fun <- getFromNamespace(".np_inid_boot_from_scoef", "np")

  set.seed(20260318)
  n <- 28L
  xdat <- data.frame(x = runif(n, -1, 1))
  zdat <- data.frame(z = rnorm(n))
  y <- with(xdat, x^2 + zdat$z + rnorm(n, sd = 0.1))

  ex.seq <- seq(
    np:::trim.quantiles(xdat[, 1L], 0.05)[1L],
    np:::trim.quantiles(xdat[, 1L], 0.05)[2L],
    length.out = 4L
  )
  ez.seq <- seq(
    np:::trim.quantiles(zdat[, 1L], 0.05)[1L],
    np:::trim.quantiles(zdat[, 1L], 0.05)[2L],
    length.out = 4L
  )
  eval.grid <- expand.grid(x = ex.seq, z = ez.seq)

  bw <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    bws = c(0.7),
    bwtype = "fixed",
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 1,
    bernstein.basis = FALSE
  )

  counts <- cbind(
    c(rep(2L, 6L), rep(1L, n - 12L), rep(0L, 6L)),
    c(rep(0L, 4L), rep(1L, n - 8L), rep(2L, 4L)),
    c(3L, rep(1L, n - 3L), 2L, 0L)
  )

  helper.out <- boot.fun(
    txdat = xdat,
    ydat = y,
    tzdat = zdat,
    exdat = eval.grid["x"],
    ezdat = eval.grid["z"],
    bws = bw,
    B = ncol(counts),
    counts = counts,
    mode = "exact"
  )

  oracle <- vapply(seq_len(ncol(counts)), function(j) {
    idx <- np:::.np_counts_to_indices(counts[, j])
    as.vector(npscoef(
      bws = bw,
      txdat = xdat[idx, , drop = FALSE],
      tzdat = zdat[idx, , drop = FALSE],
      tydat = y[idx],
      exdat = eval.grid["x"],
      ezdat = eval.grid["z"],
      iterate = FALSE,
      errors = FALSE
    )$mean)
  }, numeric(nrow(eval.grid)))
  oracle <- t(oracle)

  expect_equal(helper.out$t, oracle, tolerance = 1e-10)
})

test_that("npscoef frozen nonfixed local-polynomial helpers stay finite", {
  boot.fun <- getFromNamespace(".np_inid_boot_from_scoef", "np")

  set.seed(20260318)
  n <- 40L
  xdat <- data.frame(x = runif(n, -1, 1))
  zdat <- data.frame(z = rnorm(n))
  y <- with(xdat, x^2 + zdat$z + rnorm(n, sd = 0.1))
  exdat <- data.frame(x = seq(
    np:::trim.quantiles(xdat[, 1L], 0.05)[1L],
    np:::trim.quantiles(xdat[, 1L], 0.05)[2L],
    length.out = 5L
  ))
  ezdat <- data.frame(z = seq(
    np:::trim.quantiles(zdat[, 1L], 0.05)[1L],
    np:::trim.quantiles(zdat[, 1L], 0.05)[2L],
    length.out = 5L
  ))
  counts <- rmultinom(n = 5L, size = n, prob = rep.int(1 / n, n))

  configs <- list(
    list(
      label = "adaptive_lp_bernstein",
      bw = npscoefbw(
        xdat = xdat,
        zdat = zdat,
        ydat = y,
        nmulti = 1L,
        regtype = "lp",
        degree = 1L,
        bernstein.basis = TRUE,
        bwtype = "adaptive_nn"
      )
    ),
    list(
      label = "generalized_ll",
      bw = npscoefbw(
        xdat = xdat,
        zdat = zdat,
        ydat = y,
        nmulti = 1L,
        regtype = "ll",
        bwtype = "generalized_nn"
      )
    )
  )

  for (cfg in configs) {
    exact.out <- boot.fun(
      txdat = xdat,
      ydat = y,
      tzdat = zdat,
      exdat = exdat,
      ezdat = ezdat,
      bws = cfg$bw,
      B = ncol(counts),
      counts = counts,
      mode = "exact"
    )
    frozen.out <- boot.fun(
      txdat = xdat,
      ydat = y,
      tzdat = zdat,
      exdat = exdat,
      ezdat = ezdat,
      bws = cfg$bw,
      B = ncol(counts),
      counts = counts,
      mode = "frozen"
    )

    expect_equal(frozen.out$t0, exact.out$t0, tolerance = 1e-10, info = cfg$label)
    expect_true(all(is.finite(frozen.out$t)), info = cfg$label)
    expect_true(max(abs(frozen.out$t)) < 1e4, info = cfg$label)
  }
})

.onAttach <- function (lib, pkg) {
	packageStartupMessage(
    sprintf(
      paste(
        "np %s",
        "Examples and guides at https://jeffreyracine.github.io/gallery/",
        'See also vignette("np_getting_started", package = "np")',
        sep = "\n"
      ),
      utils::packageDescription(pkg, fields = "Version")
    ),
    domain = NULL,
    appendLF = TRUE
  )
}

.onLoad <- function (lib, pkg) {
  if (is.null(getOption("np.messages")))
    options(np.messages = TRUE)
  if (is.null(getOption("np.tree")))
    options(np.tree = FALSE)
  if (is.null(getOption("np.largeh.rel.tol")))
    options(np.largeh.rel.tol = 1e-3)
  if (is.null(getOption("np.disc.upper.rel.tol")))
    options(np.disc.upper.rel.tol = 1e-2)
}

.onUnload <- function (lpath){
  tryCatch(.Call("C_np_release_static_buffers", PACKAGE = "np"), error = function(e) NULL)
  library.dynam.unload("np", libpath=lpath) 
}

condistribution <- 
    function(bws, xeval, yeval, condist, conderr = NA,
             congrad = NA, congerr = NA,           
             ntrain, trainiseval = FALSE, gradients = FALSE,
             gradient.order = NULL,
             proper.requested = FALSE,
             proper.applied = FALSE,
             proper.method = NULL,
             condist.raw = NULL,
             proper.info = NULL,
             rows.omit = NA,
             timing = NA, total.time = NA,
             optim.time = NA, fit.time = NA){

        if (missing(bws) || missing(xeval) || missing(yeval) || missing(condist) || missing(ntrain))
            stop("improper invocation of condistribution constructor")

        if (length(rows.omit) == 0)
            rows.omit <- NA

        d <- list(
            xbw = bws$xbw,
            ybw = bws$ybw,
            bws = bws,
            xnames = bws$xnames,
            ynames = bws$ynames,
            nobs = dim(xeval)[1],
            xndim = bws$xndim,
            yndim = bws$yndim,
            xnord = bws$xnord,
            xnuno = bws$xnuno,
            xncon = bws$xncon,
            ynord = bws$ynord,
            ynuno = bws$ynuno,
            yncon = bws$yncon,
            pscaling = bws$pscaling,
            ptype = bws$ptype,
            pcxkertype = bws$pcxkertype,
            puxkertype = bws$puxkertype,
            poxkertype = bws$poxkertype,
            pcykertype = bws$pcykertype,
            puykertype = bws$puykertype,
            poykertype = bws$poykertype,
            xeval = xeval,
            yeval = yeval,
            condist = condist,
            conderr = conderr,
            congrad = congrad,
            congerr = congerr,
            ntrain = ntrain,
            trainiseval = trainiseval,
            gradients = gradients,
            gradient.order = gradient.order,
            proper.requested = proper.requested,
            proper.applied = proper.applied,
            proper.method = proper.method,
            condist.raw = condist.raw,
            proper.info = proper.info,
            rows.omit = rows.omit,
            nobs.omit = if (identical(rows.omit, NA)) 0 else length(rows.omit),
            timing = timing, total.time = total.time,
            optim.time = optim.time, fit.time = fit.time)


        class(d) <- "condistribution"

        return(d)
    }

print.condistribution <- function(x, digits=NULL, ...){
  cat("\nConditional distribution data: ", x$ntrain, " training points,",
      if (x$trainiseval) "" else paste(" and ", x$nobs, " evaluation points,\n", sep=""),
      " in ", x$xndim + x$yndim, " variable(s)",
      "\n(", x$yndim, " dependent variable(s), and ", x$xndim, " explanatory variable(s))\n\n",
      sep="")
  print(matrix(x$ybw,ncol=x$yndim,dimnames=list(paste("Dep. Var. ",x$pscaling,":",sep=""),x$ynames)))

  print(matrix(x$xbw,ncol=x$xndim,dimnames=list(paste("Exp. Var. ",x$pscaling,":",sep=""),x$xnames)))

  cat(genDenEstStr(x))
  cat(genBwKerStrs(x$bws))
  if (!is.null(x$proper.requested) && !is.null(x$proper.applied)) {
    proper.state <- if (isTRUE(x$proper.applied)) {
      sprintf("requested and applied (%s)", x$proper.method)
    } else if (isTRUE(x$proper.requested)) {
      "requested but not applied"
    } else {
      "not requested"
    }
    cat("\nProper distribution repair:", proper.state)
  }


  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

fitted.condistribution <- function(object, ...){
 object$condist 
}
se.condistribution <- function(x){
  if (isTRUE(x$proper.applied)) {
    stop("standard errors are unavailable for repaired conditional distributions in tranche 1")
  }
  x$conderr
}
gradients.condistribution <- function(x, errors = FALSE, gradient.order = NULL, ...) {
  errors <- npValidateScalarLogical(errors, "errors")
  gout <- if (!errors) x$congrad else x$congerr
  if (is.null(gout) || (length(gout) == 1L && is.logical(gout) && is.na(gout)))
    stop(if (!errors)
      "gradients are not available: fit the model with gradients=TRUE"
    else
      "gradient standard errors are not available: fit the model with gradients=TRUE")

  reg.spec <- npConditionalRegEngineSpec(x$bws, where = "gradients.condistribution")
  if (!identical(reg.spec$reg.engine, "lp") || is.null(gradient.order))
    return(gout)

  if (!is.matrix(gout))
    return(gout)

  gorder <- npValidateGlpGradientOrder(regtype = "lp",
                                       gradient.order = gradient.order,
                                       ncon = x$bws$xncon)
  stored.order <- x$gradient.order
  if (is.null(stored.order) && x$bws$xncon > 0L)
    stored.order <- rep.int(1L, x$bws$xncon)
  stored.order <- npValidateGlpGradientOrder(regtype = "lp",
                                             gradient.order = stored.order,
                                             ncon = x$bws$xncon)
  if (length(gorder)) {
    defined.request <- (gorder <= reg.spec$degree.engine)
    if (any(defined.request & (gorder != stored.order))) {
      stop("requested gradient.order differs from the derivative order stored in this condistribution object; refit or predict/evaluate with gradients=TRUE and the desired gradient.order",
           call. = FALSE)
    }
  }

  gout.masked <- gout
  gout.masked[,] <- NA_real_
  cont.idx <- which(x$bws$ixcon)
  if (length(cont.idx)) {
    keep.cont <- (gorder <= reg.spec$degree.engine)
    if (any(keep.cont)) {
      keep.idx <- cont.idx[keep.cont]
      gout.masked[, keep.idx] <- gout[, keep.idx, drop = FALSE]
    }
    if (any(gorder > reg.spec$degree.engine))
      .np_warning("some requested glp derivatives exceed polynomial degree; returning NA for those components")
  }
  gout.masked
}

predict.condistribution <- function(object, se.fit = FALSE, ...) {
  se.fit <- npValidateScalarLogical(se.fit, "se.fit")
  dots <- list(...)
  has.formula.route <- !is.null(object$bws$formula)
  proper_arg <- dots[["proper", exact = TRUE]]

  if ((!is.null(dots$exdat) || !is.null(dots$eydat)) && !is.null(dots$newdata))
    dots$newdata <- NULL

  if (!has.formula.route &&
      is.null(dots$exdat) &&
      is.null(dots$eydat) &&
      !is.null(dots$newdata)) {
    nd <- toFrame(dots$newdata)
    req <- c(object$ynames, object$xnames)
    miss <- setdiff(req, names(nd))
    if (length(miss) > 0L) {
      stop(sprintf("'newdata' must include columns %s, or supply both 'exdat' and 'eydat'.",
                   paste(shQuote(req), collapse = ", ")))
    }
    dots$eydat <- nd[, object$ynames, drop = FALSE]
    dots$exdat <- nd[, object$xnames, drop = FALSE]
    dots$newdata <- NULL
  }

  if (is.null(proper_arg) && isTRUE(object$proper.requested)) {
    dots$proper <- TRUE
    proper_arg <- TRUE
  }
  if (isTRUE(proper_arg)) {
    proper.control <- dots[["proper.control", exact = TRUE]]
    if (is.null(proper.control))
      proper.control <- list()
    if (is.null(proper.control$fail.on.unsupported))
      proper.control$fail.on.unsupported <- TRUE
    dots$proper.control <- proper.control
  }

  tr <- do.call(npcdist, c(list(bws = object$bws), dots))
  if(se.fit)
    return(list(fit = fitted(tr), se.fit = se(tr), 
                df = tr$nobs))
  else
    return(fitted(tr))
}


summary.condistribution <- function(object, ...){
  cat("\nConditional Distribution Data: ", object$ntrain, " training points,",
      if (object$trainiseval) "" else paste(" and ", object$nobs, " evaluation points,\n", sep=""),
      " in ", object$xndim + object$yndim, " variable(s)",
      "\n(", object$yndim, " dependent variable(s), and ", object$xndim, " explanatory variable(s))\n\n",
      sep="")

  cat(genOmitStr(object))
  print(matrix(object$ybw,ncol=object$yndim,dimnames=list(paste("Dep. Var. ",object$pscaling,":",sep=""),object$ynames)))

  print(matrix(object$xbw,ncol=object$xndim,dimnames=list(paste("Exp. Var. ",object$pscaling,":",sep=""),object$xnames)))

  cat(genDenEstStr(object))

  cat(genBwKerStrs(object$bws))
  if (!is.null(object$proper.requested) && !is.null(object$proper.applied)) {
    proper.state <- if (isTRUE(object$proper.applied)) {
      sprintf("requested and applied (%s)", object$proper.method)
    } else if (isTRUE(object$proper.requested)) {
      "requested but not applied"
    } else {
      "not requested"
    }
    cat("\nProper distribution repair:", proper.state)
  }
  cat(genTimingStr(object))
  cat('\n\n')  
}

# This function implements an individual test of significance for both
# discrete (Racine, hart, Li, 2006, ER) and continuous variables
# (Racine, 1997, JBES). It accepts a data frame for explanatory data
# (mixed datatypes allowed), a vector for y for a regression model, an
# npregbw object, and a set of indices for the columns of X for which
# the test is to be run (default = all).

npsigtest <-
  function(bws, ...){
    args <- list(...)

    if (!missing(bws)){
      if (is.recursive(bws)){
        if (!is.null(bws$formula) && is.null(args$xdat))
          UseMethod("npsigtest",bws$formula)
#        else if (!is.null(bws$call) && is.null(args$xdat) && (class(bws) != "npregression"))
        else if (!is.null(bws$call) && is.null(args$xdat) && (!isa(bws,"npregression")))
          UseMethod("npsigtest",bws$call)
        else if (!is.call(bws))
          UseMethod("npsigtest",bws)
        else
          UseMethod("npsigtest",NULL)
      } else {
        UseMethod("npsigtest", NULL)
      }
    } else {
      UseMethod("npsigtest", NULL)
    }
  }

npsigtest.formula <-
  function(bws, data = NULL, ...){

    tt <- terms(bws)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)
    tmf <- bws$call[c(1,m)]
    tmf[[1]] <- as.name("model.frame")
    tmf[["formula"]] <- tt
    mf.args <- as.list(tmf)[-1L]
    umf <- tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

    ydat <- model.response(tmf)
    xdat <- tmf[, attr(attr(tmf, "terms"),"term.labels"), drop = FALSE]

    ev <- npsigtest(xdat = xdat, ydat = ydat, bws = bws, ...)
    ev$call <- match.call(expand.dots = FALSE)
    environment(ev$call) <- parent.frame()
    ev$rows.omit <- as.vector(attr(umf,"na.action"))
    ev$nobs.omit <- length(ev$rows.omit)
    return(ev)
  }

npsigtest.call <-
  function(bws, ...) {
    ev <- npsigtest(xdat = .np_eval_bws_call_arg(bws, "xdat"),
                    ydat = .np_eval_bws_call_arg(bws, "ydat"),
                    bws = bws, ...)
    ev$call <- match.call(expand.dots = FALSE)
    environment(ev$call) <- parent.frame()
    return(ev)
  }

npsigtest.npregression <-
  function(bws, ...){
    ev <- npsigtest(bws$bws, ...)
    ev$call <- match.call(expand.dots = FALSE)
    environment(ev$call) <- parent.frame()
    return(ev)
  }

.np_npsig_bootstrap_bw_reselect <- function(xdat,
                                            ydat,
                                            bws.seed,
                                            extra.args = list(),
                                            bootstrap.iter,
                                            bw.fun = npregbw) {
  bw.args <- if (length(extra.args)) extra.args else list()
  bw.args[c("xdat", "ydat", "bws")] <- NULL

  user.nmulti <- !is.null(names(bw.args)) &&
    "nmulti" %in% names(bw.args) &&
    !is.null(bw.args$nmulti)

  if (!user.nmulti && bootstrap.iter > 1L)
    bw.args$nmulti <- 1L

  .np_progress_with_legacy_suppressed(
    do.call(bw.fun, c(list(xdat = xdat, ydat = ydat, bws = bws.seed), bw.args))
  )
}

npsigtest.rbandwidth <- function(bws,
                                 xdat = stop("data xdat missing"),
                                 ydat = stop("data ydat missing"),
                                 boot.num = 399,
                                 boot.method = c("iid","wild","wild-rademacher","pairwise"),
                                 boot.type = c("I","II"),
                                 pivot = TRUE,
                                 joint = FALSE,
                                 index = seq_len(ncol(xdat)),
                                 random.seed = 42,
                                 ...) {

  xdat <- toFrame(xdat)

  if(boot.num < 9) stop("number of bootstrap replications must be >= 9")

  ## catch and destroy NA's
  goodrows <- seq_len(nrow(xdat))
  rows.omit <- attr(na.omit(data.frame(xdat,ydat)), "na.action")
  goodrows[rows.omit] <- 0

  if (all(goodrows==0))
    stop("Data has no rows without NAs")

  xdat <- xdat[goodrows,,drop = FALSE]
  ydat <- ydat[goodrows]

  if (is.factor(ydat))
    stop("dependent variable must be continuous.")

  ## Save seed prior to setting

  seed.state <- .np_seed_enter(random.seed)
  extra.args <- list(...)


  boot.type <- match.arg(boot.type)
  boot.method <- match.arg(boot.method)

  if(boot.type=="II") {
    ## Store a copy of the bandwidths passed in
    bws.original <- bws
  }

  num.obs <- nrow(xdat)

  ## Test for valid entries in index

  if(anyNA(index)) stop("index must not contain missing values")
  if(any(index < 1 | index > NCOL(xdat), na.rm = TRUE)) stop(paste("invalid index provided: index entries must lie between 1 and ",NCOL(xdat),sep=""))
  if(length(unique(index)) < length(index)) stop("index contains repeated values (must be unique)")

  if(!joint) {

    In <- numeric(length(index))
    P <- numeric(length(index))

  }

  ## Some constants for the wild bootstrap

  a <- -0.6180339887499  # (1-sqrt(5))/2
  b <- 1.6180339887499   # (1+sqrt(5))/2
  P.a <-0.72360679774998 # (1+sqrt(5))/(2*sqrt(5))

  draw.wild.mult <- function(n.obs, a, b, p.a) {
    u <- stats::runif(n.obs)
    mult <- rep.int(b, n.obs)
    mult[u <= p.a] <- a
    mult
  }

  ## A vector for storing the resampled statistics

  In.vec <- numeric(boot.num)

  if(joint==TRUE) {

    ## Joint test

    In.mat = matrix(data = 0, ncol = 1, nrow = boot.num)

    if(boot.type=="II") {

      ## Reset bw vals to original as the ith component of bws gets
      ## overwritten when index changes so needs to be set to its
      ## original value

      bws <- bws.original

    }

    ## Note - xdat must be a data frame

    ## Construct In, the average value of the squared derivatives of
    ## the jth element, discrete or continuous

    npreg.out <- npreg(txdat = xdat,
                       tydat = ydat,
                       bws = bws,
                       gradients = TRUE,
                       ...)

    In <- if(!pivot) {
      mean(npreg.out$grad[,index]^2)
    } else {
      ## Temporarily trap NaN XXX
      npreg.out$gerr[is.nan(npreg.out$gerr)] <- .Machine$double.xmax
      mean((npreg.out$grad[,index]/NZD(npreg.out$gerr[,index]))^2)
    }

    if(boot.method != "pairwise") {

      ## Compute scale and mean of unrestricted residuals

      ei.unres <- scale(residuals(npreg(bws=bws)))
      ei.unres.scale <- attr(ei.unres,"scaled:scale")
      ei.unres.center <- attr(ei.unres,"scaled:center")      

      ## We now construct mhat.xi holding constant the variable whose
      ## significance is being tested at its median. First, make a copy
      ## of the data frame xdat
      
      xdat.eval <- xdat
      
      ## Impose the null by evaluating the conditional mean holding
      ## xdat[,i] constant at its median (numeric) or mode
      ## (factor/ordered) using uocquantile()

      for(i in index) {
        xq <- uocquantile(xdat[,i], 0.5)
        if (is.factor(xdat[,i]) || is.ordered(xdat[,i])) {
          xdat.eval[,i] <- cast(xq, xdat[,i], same.levels = TRUE)
        } else {
          xdat.eval[,i] <- xq
        }
      }
      
      mhat.xi <-  npreg(txdat = xdat,
                        tydat = ydat,
                        exdat = xdat.eval,
                        bws = bws,
                        ...)$mean

      ## Rescale and recenter the residuals under the null to those
      ## under the alternative
      
      ei <- as.numeric(scale(ydat-mhat.xi)*ei.unres.scale+ei.unres.center)
      
      ## Recenter the residuals to have mean zero

      ei <- ei - mean(ei)
      
    }
    
    if(boot.type=="II")
      bws.boot.prev <- bws.original

    .np_progress_note("Testing joint significance")
    progress <- .np_progress_begin("Bootstrap replications", total = boot.num, surface = "bootstrap")

    for (i.star in seq_len(boot.num)) {
      if(boot.method == "iid") {

        ydat.star <- mhat.xi + ei[sample.int(num.obs, replace = TRUE)]

      } else if(boot.method == "wild") {

        ## Conduct a wild bootstrap. We generate a sample for ydat
        ## (ydat.star) drawn from the conditional mean evaluated
        ## holding the variable tested at its median, and add to that
        ## a wild bootstrap draw from the original disturbance vector

        ydat.star <- mhat.xi + ei * draw.wild.mult(num.obs, a, b, P.a)

      } else if(boot.method == "wild-rademacher") {

        ## Conduct a wild bootstrap. We generate a sample for ydat
        ## (ydat.star) drawn from the conditional mean evaluated
        ## holding the variable tested at its median, and add to that
        ## a wild bootstrap draw from the original disturbance vector

        ydat.star <- mhat.xi + ei * draw.wild.mult(num.obs, -1, 1, P.a)

      } else if(boot.method =="pairwise") {

        ## Leave variable being tested untouched, resample remaining
        ## pairs of y,X thereby breaking any systematic relationship
        ## between variable being tested in y
        boot.index <- sample.int(num.obs, replace = TRUE)
        ydat.star <- ydat[boot.index]
        xdat.star <- xdat[boot.index,]
        for(i in index) xdat.star[,i] <- xdat[,i]

      }

      if(boot.type=="II") {

        ## Bootstrap II reuses the previous bootstrap optimum as a
        ## hot start and drops to nmulti=1 after the first
        ## re-selection unless the user explicitly supplied nmulti.

        if(boot.method == "pairwise") {

          bws.boot <- .np_npsig_bootstrap_bw_reselect(
            xdat = xdat.star,
            ydat = ydat.star,
            bws.seed = bws.boot.prev,
            extra.args = extra.args,
            bootstrap.iter = i.star
          )

        } else {

          bws.boot <- .np_npsig_bootstrap_bw_reselect(
            xdat = xdat,
            ydat = ydat.star,
            bws.seed = bws.boot.prev,
            extra.args = extra.args,
            bootstrap.iter = i.star
          )

        }

        bws.boot.prev <- bws.boot

        ## Copy the new cross-validated bandwidth for variable i into
        ## bw.original and use this below.

        bws <- bws.original

        bws$bw[index] <- bws.boot$bw[index]

      }

      if(boot.method == "pairwise") {

        npreg.boot <- npreg(txdat = xdat.star,
                            tydat = ydat.star,
                            bws = bws,
                            gradients = TRUE,
                            ...)

      } else {

        npreg.boot <- npreg(txdat = xdat,
                            tydat = ydat.star,
                            bws = bws,
                            gradients = TRUE,
                            ...)

      }

      In.vec[i.star] <- if(!pivot) {
        mean(npreg.boot$grad[,index]^2)
      } else {
        ## Temporarily trap NaN XXX
        npreg.boot$gerr[is.nan(npreg.boot$gerr)] <- .Machine$double.xmax
        mean((npreg.boot$grad[,index]/NZD(npreg.boot$gerr[,index]))^2)
      }
      progress <- .np_progress_step(progress, done = i.star)
    }

    progress <- .np_progress_end(progress)

    ## Compute the P-value

    P <- mean(In.vec > In)

    In.mat[,1] = In.vec

  } else {

    ## Individual test

    ## ii is the counter for successive elements of In and P...

    In.mat = matrix(data = 0, ncol = length(index), nrow = boot.num)

    ii <- 0

    for(i in index) {
      
      ## Increment counter...
      
      ii <- ii + 1
      
      if(boot.type=="II") {
        
        ## Reset bw vals to original as the ith component of bws gets
        ## overwritten when index changes so needs to be set to its
        ## original value
        
        bws <- bws.original
        
      }
      
      ## Note - xdat must be a data frame
      
      ## Construct In, the average value of the squared derivatives of
      ## the jth element, discrete or continuous
      
      npreg.out <- npreg(txdat = xdat,
                         tydat = ydat,
                         bws = bws,
                         gradients = TRUE,
                         ...)
      
      In[ii] <- if(!pivot) {
        mean(npreg.out$grad[,i]^2)
      } else {
        ## Temporarily trap NaN XXX
        npreg.out$gerr[is.nan(npreg.out$gerr)] <- .Machine$double.xmax
        mean((npreg.out$grad[,i]/NZD(npreg.out$gerr[,i]))^2)
      }
      
      if(boot.method != "pairwise") {

        ## Compute scale and mean of unrestricted residuals

        ei.unres <- scale(residuals(npreg(bws=bws)))
        ei.unres.scale <- attr(ei.unres,"scaled:scale")
        ei.unres.center <- attr(ei.unres,"scaled:center")      

        ## We now construct mhat.xi holding constant the variable whose
        ## significance is being tested at its median. First, make a copy
        ## of the data frame xdat
        
        xdat.eval <- xdat
        
        ## Impose the null by evaluating the conditional mean holding
        ## xdat[,i] constant at its median (numeric) or mode
        ## (factor/ordered) using uocquantile()
        
        xq <- uocquantile(xdat[,i], 0.5)
        if (is.factor(xdat[,i]) || is.ordered(xdat[,i])) {
          xdat.eval[,i] <- cast(xq, xdat[,i], same.levels = TRUE)
        } else {
          xdat.eval[,i] <- xq
        }
        
        mhat.xi <-  npreg(txdat = xdat,
                          tydat = ydat,
                          exdat = xdat.eval,
                          bws = bws,
                          ...)$mean
        
        ## Rescale and recenter the residuals under the null to those
        ## under the alternative
        
        ei <- as.numeric(scale(ydat-mhat.xi)*ei.unres.scale+ei.unres.center)
        
        ## Recenter the residuals to have mean zero
        
        ei <- ei - mean(ei)
        
      }
      
      if(boot.type=="II")
        bws.boot.prev <- bws.original

      .np_progress_note(sprintf("Testing variable %s of (%s)", i, paste(index, collapse = ",")))
      progress <- .np_progress_begin("Bootstrap replications", total = boot.num, surface = "bootstrap")

      for (i.star in seq_len(boot.num)) {
        if(boot.method == "iid") {
          
          ydat.star <- mhat.xi + ei[sample.int(num.obs, replace = TRUE)]
          
        } else if(boot.method == "wild") {
          
          ## Conduct a wild bootstrap. We generate a sample for ydat
          ## (ydat.star) drawn from the conditional mean evaluated
          ## holding the variable tested at its median, and add to that
          ## a wild bootstrap draw from the original disturbance vector
          
          ydat.star <- mhat.xi + ei * draw.wild.mult(num.obs, a, b, P.a)
          
        } else if(boot.method == "wild-rademacher") {
          
          ## Conduct a wild bootstrap. We generate a sample for ydat
          ## (ydat.star) drawn from the conditional mean evaluated
          ## holding the variable tested at its median, and add to that
          ## a wild bootstrap draw from the original disturbance vector
          
          ydat.star <- mhat.xi + ei * draw.wild.mult(num.obs, -1, 1, P.a)
          
        } else if(boot.method =="pairwise") {
          
          ## Leave variable being tested untouched, resample remaining
          ## pairs of y,X thereby breaking any systematic relationship
          ## between variable being tested in y
          boot.index <- sample.int(num.obs, replace = TRUE)
          ydat.star <- ydat[boot.index]
          xdat.star <- xdat
          xdat.star[,-i] <- xdat[boot.index,-i]
          
        }
        
        if(boot.type=="II") {
          
          ## Bootstrap II reuses the previous bootstrap optimum as a
          ## hot start and drops to nmulti=1 after the first
          ## re-selection unless the user explicitly supplied nmulti.
          
          if(boot.method == "pairwise") {
            
            bws.boot <- .np_npsig_bootstrap_bw_reselect(
              xdat = xdat.star,
              ydat = ydat.star,
              bws.seed = bws.boot.prev,
              extra.args = extra.args,
              bootstrap.iter = i.star
            )
            
          } else {
            
            bws.boot <- .np_npsig_bootstrap_bw_reselect(
              xdat = xdat,
              ydat = ydat.star,
              bws.seed = bws.boot.prev,
              extra.args = extra.args,
              bootstrap.iter = i.star
            )
            
          }

          bws.boot.prev <- bws.boot
          
          ## Copy the new cross-validated bandwidth for variable i into
          ## bw.original and use this below.
          
          bws <- bws.original
          
          bws$bw[i] <- bws.boot$bw[i]
          
        }
        
        if(boot.method == "pairwise") {
          
          npreg.boot <- npreg(txdat = xdat.star,
                              tydat = ydat.star,
                              bws = bws,
                              gradients = TRUE,
                              ...)
          
        } else {
          
          npreg.boot <- npreg(txdat = xdat,
                              tydat = ydat.star,
                              bws = bws,
                              gradients = TRUE,
                              ...)
          
        }
        
        In.vec[i.star] <- if(!pivot) {
          mean(npreg.boot$grad[,i]^2)
        } else {
          ## Temporarily trap NaN XXX
          npreg.boot$gerr[is.nan(npreg.boot$gerr)] <- .Machine$double.xmax
          mean((npreg.boot$grad[,i]/NZD(npreg.boot$gerr[,i]))^2)
        }
        progress <- .np_progress_step(progress, done = i.star)

      }

      progress <- .np_progress_end(progress)
      
      ## Compute the P-value
      
      P[ii] <- mean(In.vec > In[ii])
      
      In.mat[,ii] = In.vec
      
    }
    
  } ## End invididual test

  ## Return a list containing the statistic and its P-value
  ## bootstrapped In.vec for each variable...

  ## Restore seed

  .np_seed_exit(seed.state)

  sigtest(In=In,
          In.mat,
          P=P,
          bws = bws,
          ixvar = index,
          boot.method,
          pivot,
          joint,
          boot.type,
          boot.num)

}

npsigtest.default <- function(bws, xdat, ydat, ...){
  sc <- sys.call()
  sc.names <- names(sc)

  ## here we check to see if the function was called with tdat = if it
  ## was, we need to catch that and map it to dat = otherwise the call
  ## is passed unadulterated to npudensbw

  bws.named <- any(sc.names == "bws")
  xdat.named <- any(sc.names == "xdat")
  ydat.named <- any(sc.names == "ydat")

  no.bws <- missing(bws)
  no.xdat <- missing(xdat)
  no.ydat <- missing(ydat)

  ## if bws was passed in explicitly, do not compute bandwidths

  if(xdat.named)
    xdat <- toFrame(xdat)

  sc.bw <- sc
  
  sc.bw[[1]] <- quote(npregbw)

  if(bws.named){
    sc.bw$bandwidth.compute <- FALSE
  }

  tbw <- .np_eval_bw_call(sc.bw, caller_env = parent.frame())
  
  call.args <- list(bws = tbw)
  if(!no.xdat)
    call.args$xdat <- xdat
  if(!no.ydat)
    call.args$ydat <- ydat

  dots <- list(...)
  dots[c("bws", "bandwidth.compute", "formula", "data", "xdat", "ydat")] <- NULL

  ev <- do.call(npsigtest, c(call.args, dots))

  ev$call <- match.call(expand.dots = FALSE)
  environment(ev$call) <- parent.frame()
  return(ev)
}

test_that("npsigtest basic functionality works", {
  set.seed(42)
  n <- 50 # Keep it small for speed
  x1 <- runif(n)
  x2 <- runif(n)
  y <- x1^2 + rnorm(n, sd=0.1) # x2 is irrelevant
  
  bw <- npregbw(y~x1+x2, bws=c(0.1, 0.5), bandwidth.compute=FALSE)
  
  # Significance test can be slow, use few boot replications
  sig <- npsigtest(bws=bw, boot.num=19)
  
  expect_s3_class(sig, "sigtest")
  expect_output(summary(sig))
})

test_that("npsigtest Type II hot-start helper honors nmulti contract", {
  helper <- getFromNamespace(".np_npsig_bootstrap_bw_reselect", "np")
  xdat <- data.frame(x = c(0.1, 0.4, 0.8))
  ydat <- c(1, 2, 3)
  calls <- list()

  stub_bw <- function(...) {
    args <- list(...)
    calls[[length(calls) + 1L]] <<- args
    structure(list(bw = 0.25, seed_id = length(calls)), class = "rbandwidth")
  }

  seed0 <- structure(list(bw = 0.5, seed_id = 0L), class = "rbandwidth")
  seed1 <- helper(xdat = xdat,
                  ydat = ydat,
                  bws.seed = seed0,
                  bootstrap.iter = 1L,
                  extra.args = list(regtype = "ll"),
                  bw.fun = stub_bw)
  helper(xdat = xdat,
         ydat = ydat,
         bws.seed = seed1,
         bootstrap.iter = 2L,
         extra.args = list(regtype = "ll"),
         bw.fun = stub_bw)

  expect_false("nmulti" %in% names(calls[[1L]]))
  expect_identical(calls[[1L]]$bws$seed_id, 0L)
  expect_identical(calls[[2L]]$nmulti, 1L)
  expect_identical(calls[[2L]]$bws$seed_id, 1L)
})

test_that("npsigtest Type II hot-start helper preserves explicit nmulti", {
  helper <- getFromNamespace(".np_npsig_bootstrap_bw_reselect", "np")
  xdat <- data.frame(x = c(0.1, 0.4, 0.8))
  ydat <- c(1, 2, 3)
  calls <- list()

  stub_bw <- function(...) {
    args <- list(...)
    calls[[length(calls) + 1L]] <<- args
    structure(list(bw = 0.25), class = "rbandwidth")
  }

  seed0 <- structure(list(bw = 0.5), class = "rbandwidth")
  helper(xdat = xdat,
         ydat = ydat,
         bws.seed = seed0,
         bootstrap.iter = 1L,
         extra.args = list(nmulti = 7L),
         bw.fun = stub_bw)
  helper(xdat = xdat,
         ydat = ydat,
         bws.seed = seed0,
         bootstrap.iter = 2L,
         extra.args = list(nmulti = 7L),
         bw.fun = stub_bw)

  expect_identical(calls[[1L]]$nmulti, 7L)
  expect_identical(calls[[2L]]$nmulti, 7L)
})

test_that("npsigtest formula interface path works", {
  set.seed(7)
  n <- 40
  x1 <- runif(n)
  x2 <- runif(n)
  y <- x1 + rnorm(n, sd = 0.1)
  mydat <- data.frame(y, x1, x2)

  sig <- npsigtest(y ~ x1 + x2,
                   data = mydat,
                   boot.num = 9)

  expect_s3_class(sig, "sigtest")
  expect_true(is.numeric(sig$P))
})

test_that("npsigtest Type II remains functional with hot-start reselection", {
  set.seed(17)
  n <- 30
  x1 <- runif(n)
  x2 <- runif(n)
  y <- x1 + rnorm(n, sd = 0.15)

  bw <- npregbw(y ~ x1 + x2, bws = c(0.2, 0.4), bandwidth.compute = FALSE)
  sig <- npsigtest(bws = bw, boot.num = 9, boot.type = "II")

  expect_s3_class(sig, "sigtest")
  expect_true(is.numeric(sig$P))
})

test_that("npsigtest rejects duplicate index entries", {
  set.seed(11)
  n <- 30
  x1 <- runif(n)
  x2 <- runif(n)
  y <- x1 + rnorm(n, sd = 0.1)

  bw <- npregbw(y ~ x1 + x2, bws = c(0.2, 0.4), bandwidth.compute = FALSE)

  expect_error(
    npsigtest(bws = bw, boot.num = 9, index = c(1, 1)),
    "repeated values"
  )
})

## This illustration was made possible by Samuele Centorrino
## <samuele.centorrino@univ-tlse1.fr>

suppressPackageStartupMessages(library(np))

set.seed(42)
read_numeric_or_default <- function(prompt, default) {
  val <- suppressWarnings(as.numeric(readline(prompt = prompt)))
  if (is.na(val)) {
    if (!interactive()) {
      val <- default
    } else {
      stop("Invalid input for prompt: ", prompt)
    }
  }
  val
}

n <- read_numeric_or_default("Input the number of observations desired: ", 30)
method <- read_numeric_or_default("Input the method (0=Landweber-Fridman, 1=Tikhonov): ", 1)
method <- ifelse(method == 0, "Landweber-Fridman", "Tikhonov")
nmulti <- read_numeric_or_default("Input the number of multistarts desired (e.g. 2): ", 1)

v  <- rnorm(n,mean=0,sd=.27)
eps <- rnorm(n,mean=0,sd=0.05)
u <- -0.5*v + eps
w <- rnorm(n,mean=0,sd=1)
z <- 0.2*w + v

## In Darolles et al (2011) there exist two DGPs. The first is
## phi(z)=z^2.

phi <- function(z) { z^2 }
eyz <- function(z) { z^2 -0.325*z }

y <- phi(z) + u

## Sort on z (for plotting)

ivdata <- data.frame(y,z,w)
ivdata <- ivdata[order(ivdata$z),]
rm(y,z,w)
attach(ivdata)

model.iv <- npregiv(y=y,z=z,w=w,nmulti=nmulti,method=method)
phi.iv <- model.iv$phi

## Now the non-iv local linear estimator of E(y|z)

ll.mean <- fitted(npreg(y~z,nmulti=nmulti,regtype="ll"))

## For the plots, restrict focal attention to the bulk of the data
## (i.e. for the plotting area trim out 1/4 of one percent from each
## tail of y and z)

trim <- 0.0025

if(method=="Tikhonov")  {

  subtext <- paste("Tikhonov alpha = ",
                   formatC(model.iv$alpha,digits=3,format="fg"),
                   ", n = ", n, sep="")

} else {

  subtext <- paste("Landweber-Fridman iterations = ",
                   model.iv$num.iterations,
                   ", n = ", n,sep="")

}

curve(phi,min(z),max(z),
      xlim=quantile(z,c(trim,1-trim)),
      ylim=quantile(y,c(trim,1-trim)),
      ylab="Y",
      xlab="Z",
      main="Nonparametric Instrumental Kernel Regression",
      sub=subtext,
      lwd=1,lty=1)
  
points(z,y,type="p",cex=.25,col="grey")

lines(z,eyz(z))

lines(z,phi.iv,col="blue",lwd=2,lty=2)

lines(z,ll.mean,col="red",lwd=2,lty=4)

legend(x="top",inset=c(.01,.01),
       c(expression(paste(varphi(z),", E(y|z)",sep="")),
         expression(paste("Nonparametric ",hat(varphi)(z))),
         "Nonparametric E(y|z)"),
       lty=c(1,2,4),
       col=c("black","blue","red"),
       lwd=c(1,2,2))

# We implement Ichimura's single index model and Klein and Spady's
# single index model using npksum() and the nlm() minimization
# routine in R. These semiparametric models are used to reduce
# dimensionality to a one-dimensional nonparametric estimator, though
# at the potential cost of misspecification.

# Note also that we will use the so-called scale normalization, i.e.,
# that beta_1=1 (no need to estimate) which reduces search by 1
# parameter (this is obviously restricted search subject to beta_1=1).

# Define the index function model... it is a simple local constant
# estimator of y on a linear index X\beta where beta_1 is presumed to
# be 1 by restriction though, at this stage, the user may feed in any
# value they so desire.

npindex <-
  function(bws, ...){
    args <- list(...)

    if (!missing(bws)){
      if (is.recursive(bws)){
        if (!is.null(bws$formula) && is.null(args$txdat))
          UseMethod("npindex",bws$formula)
        else if (!is.null(bws$call) && is.null(args$txdat))
          UseMethod("npindex",bws$call)
        else if (!is.call(bws))
          UseMethod("npindex",bws)
        else
          UseMethod("npindex",NULL)
      } else {
        UseMethod("npindex", NULL)
      }
    } else {
      UseMethod("npindex", NULL)
    }
  }

npindex.formula <-
    function(bws, data = NULL, newdata = NULL, y.eval = FALSE, ...){

        tt <- terms(bws)
        m <- match(c("formula", "data", "subset", "na.action"),
                   names(bws$call), nomatch = 0)
        tmf <- bws$call[c(1,m)]
        tmf[[1]] <- as.name("model.frame")
        tmf[["formula"]] <- tt
        mf.args <- as.list(tmf)[-1L]
        umf <- tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

        response.name <- attr(tmf, "names")[attr(attr(tmf, "terms"), "response")]
        tydat <- model.response(tmf)
        txdat <- tmf[, attr(attr(tmf, "terms"),"term.labels"), drop = FALSE]
        has.eval <- !is.null(newdata)
        if (has.eval) {
          if (!y.eval){
            npValidateNewdataFormula(newdata, tt, include.response = FALSE)
            tt <- delete.response(tt)

            orig.ts <- .np_terms_ts_mask(terms_obj = tt, data = newdata)
            
            ## delete.response clobbers predvars, which is used for timeseries objects
            ## so we need to reconstruct it

            if(all(orig.ts)){
              args <- (as.list(attr(tt, "variables"))[-1])
              attr(tt, "predvars") <- as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), args))))
            }else if(any(orig.ts)){
              arguments <- (as.list(attr(tt, "variables"))[-1])
              arguments.normal <- arguments[which(!orig.ts)]
              arguments.timeseries <- arguments[which(orig.ts)]

              ix <- sort(c(which(orig.ts),which(!orig.ts)),index.return = TRUE)$ix
              attr(tt, "predvars") <- bquote(.(as.call(c(quote(cbind),as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments.timeseries)))),arguments.normal,check.rows = TRUE)))[,.(ix)])
            }else{
              attr(tt, "predvars") <- attr(tt, "variables")
            }
          }
          
          if (y.eval)
            npValidateNewdataFormula(newdata, tt, include.response = TRUE)
          umf.args <- list(formula = tt, data = newdata)
          umf <- do.call(stats::model.frame, umf.args, envir = parent.frame())
          emf <- umf

          if (y.eval)
            eydat <- model.response(emf)
          
          exdat <- emf[, attr(attr(emf, "terms"),"term.labels"), drop = FALSE]
        }

        si.args <- list(txdat = txdat, tydat = tydat)
        if (has.eval) {
          si.args$exdat <- exdat
          if (y.eval)
            si.args$eydat <- eydat
        }
        si.args$bws <- bws
        ev <- do.call(npindex, c(si.args, list(...)))

        if (length(response.name) == 1L && !is.na(response.name) && nzchar(response.name)) {
            if (!is.null(ev$bws))
                ev$bws$ynames <- response.name
        }

        ev$omit <- attr(umf,"na.action")
        ev$rows.omit <- as.vector(ev$omit)
        ev$nobs.omit <- length(ev$rows.omit)

        ev$mean <- napredict(ev$omit, ev$mean)
        ev$merr <- napredict(ev$omit, ev$merr)

        if(ev$gradients){
            ev$grad <- napredict(ev$omit, ev$grad)
            ev$gerr <- napredict(ev$omit, ev$gerr)
        }

        if(ev$residuals){
            ev$resid <- naresid(ev$omit, ev$resid)
        }    
        return(ev)
    }

npindex.call <-
  function(bws, ...) {
    npindex(txdat = .np_eval_bws_call_arg(bws, "xdat"),
            tydat = .np_eval_bws_call_arg(bws, "ydat"),
            bws = bws, ...)
  }

.np_index_regression_bandwidth <- function(index.df, ydat, bws, spec) {
  bw.args <- list(
    xdat = index.df,
    ydat = ydat,
    bws = bws$bw,
    bandwidth.compute = FALSE,
    bwtype = bws$type,
    ckertype = bws$ckertype,
    ckerorder = bws$ckerorder,
    regtype = spec$regtype.engine
  )
  if (identical(spec$regtype.engine, "lp")) {
    bw.args$basis <- spec$basis.engine
    bw.args$degree <- spec$degree.engine
    bw.args$bernstein.basis <- spec$bernstein.basis.engine
  }
  do.call(npregbw, bw.args)
}

npindex.default <- function(bws, txdat, tydat, nomad = FALSE, ...){
  sc <- sys.call()
  sc.names <- names(sc)
  nomad <- npValidateScalarLogical(nomad, "nomad")

  ## here we check to see if the function was called with tdat =
  ## if it was, we need to catch that and map it to dat =
  ## otherwise the call is passed unadulterated to npudensbw

  bws.named <- any(sc.names == "bws")
  txdat.named <- any(sc.names == "txdat")
  tydat.named <- any(sc.names == "tydat")

  no.bws <- missing(bws)
  no.txdat <- missing(txdat)
  no.tydat <- missing(tydat)
  has.explicit.bws <- (!no.bws) && isa(bws, "sibandwidth")

  ## if bws was passed in explicitly, do not compute bandwidths

  if(txdat.named)
    txdat <- toFrame(txdat)

  sc.bw <- sc
  
  sc.bw[[1]] <- quote(npindexbw)

  bws.formula <- (!no.bws) && inherits(bws, "formula")
  if (bws.formula) {
    ib <- match("bws", names(sc.bw), nomatch = 0L)
    if (ib > 0L) names(sc.bw)[ib] <- "formula"
  }

  if(bws.named && !bws.formula){
    sc.bw$bandwidth.compute <- FALSE
  }

  ostxy <- c('txdat','tydat')
  nstxy <- c('xdat','ydat')
  
  m.txy <- match(ostxy, names(sc.bw), nomatch = 0)

  if(any(m.txy > 0)) {
    names(sc.bw)[m.txy] <- nstxy[m.txy > 0]
  }
    
  use.outer.bandwidth.progress <- !.np_bw_call_uses_nomad_degree_search(
    sc.bw,
    caller_env = parent.frame()
  )

  tbw <- if (!has.explicit.bws) {
    if (use.outer.bandwidth.progress) {
      .np_progress_select_bandwidth_enhanced(
        "Selecting single-index bandwidth",
        .np_eval_bw_call(sc.bw, caller_env = parent.frame())
      )
    } else {
      .np_eval_bw_call(sc.bw, caller_env = parent.frame())
    }
  } else {
    .np_eval_bw_call(sc.bw, caller_env = parent.frame())
  }

  call.args <- list(bws = tbw)
  if (no.bws) {
    call.args$txdat <- txdat
    call.args$tydat <- tydat
  } else {
    if (txdat.named) call.args$txdat <- txdat
    if (tydat.named) call.args$tydat <- tydat
    if ((!bws.named) && (!txdat.named) && (!no.tydat) && (!tydat.named)) {
      call.args <- c(call.args, list(tydat))
    }
  }
  if (no.bws || bws.formula || is.call(bws))
    call.args$.np_fit_progress_handoff <- TRUE
  do.call(npindex, c(call.args, list(...)))
}

npindex.sibandwidth <-
  function(bws,
           txdat = stop("training data 'txdat' missing"),
           tydat = stop("training data 'tydat' missing"),
           exdat,
           eydat,
           boot.num = 399,
           errors = FALSE,
           gradients = FALSE,
           residuals = FALSE, ...) {

    fit.start <- proc.time()[3]
    dots <- list(...)
    fit.progress.handoff <- isTRUE(dots$.np_fit_progress_handoff)
    fit.progress.allow <- isTRUE(.np_progress_enabled(domain = "bandwidth"))
    gradients <- npValidateScalarLogical(gradients, "gradients")
    residuals <- npValidateScalarLogical(residuals, "residuals")
    errors <- npValidateScalarLogical(errors, "errors")
    if (!is.numeric(boot.num) || length(boot.num) != 1L || is.na(boot.num) ||
        !is.finite(boot.num) || boot.num < 1 || boot.num != floor(boot.num))
      stop("'boot.num' must be a positive integer")
    boot.num <- as.integer(boot.num)

    no.ex = missing(exdat)
    no.ey = missing(eydat)

    ## if no.ex then if !no.ey then ey and tx must match, to get
    ## oos errors alternatively if no.ey you get is errors if
    ## !no.ex then if !no.ey then ey and ex must match, to get
    ## oos errors alternatively if no.ey you get NO errors since we
    ## don't evaluate on the training data

    txdat = toFrame(txdat)

    if (!(is.vector(tydat) || is.factor(tydat)))
      stop("'tydat' must be a vector or a factor")

    tydat =
      if (is.factor(tydat))
        as.numeric(levels(tydat))[as.integer(tydat)]
      else
        as.double(tydat)

    if (!no.ex){
      exdat = toFrame(exdat)

      if (! txdat %~% exdat )
        stop("'txdat' and 'exdat' are not similar data frames!")

      if (!no.ey){
        if (dim(exdat)[1] != length(eydat))
          stop("number of evaluation data 'exdat' and dependent data 'eydat' do not match")
      }

    } else if(!no.ey) {
      if (dim(txdat)[1] != length(eydat))
        stop("number of training data 'txdat' and dependent data 'eydat' do not match")
    }

    ## catch and destroy NA's
    keep.rows <- rep_len(TRUE, nrow(txdat))
    rows.omit <- attr(na.omit(data.frame(txdat,tydat)), "na.action")
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Training data has no rows without NAs")

    txdat <- txdat[keep.rows,,drop = FALSE]
    tydat <- tydat[keep.rows]

    if (!no.ex){
      keep.eval <- rep_len(TRUE, nrow(exdat))
      eval.df <- data.frame(exdat)
      if (!no.ey)
        eval.df <- data.frame(eval.df, eydat)
      rows.omit <- attr(na.omit(eval.df), "na.action")
      if (length(rows.omit) > 0L)
        keep.eval[as.integer(rows.omit)] <- FALSE

      if (!any(keep.eval))
        stop("Evaluation data has no rows without NAs")

      exdat <- exdat[keep.eval,,drop = FALSE]
      if (!no.ey)
        eydat <- eydat[keep.eval]
    }

    ## convert tydat, eydat to numeric, from a factor with levels from the y-data
    ## used during bandwidth selection.

    if (is.factor(tydat)){
      tydat <- adjustLevels(data.frame(tydat), bws$ydati)[,1]
      tydat <- (bws$ydati$all.dlev[[1]])[as.integer(tydat)]
    }
    else
      tydat <- as.double(tydat)

    if (no.ey)
      eydat <- double()
    else {
      if (is.factor(eydat)){
        eydat <- adjustLevels(data.frame(eydat), bws$ydati)[,1]
        eydat <- (bws$ydati$all.dlev[[1]])[as.integer(eydat)]
      }
      else
        eydat <- as.double(eydat)
    }

    ## re-assign levels in training and evaluation data to ensure correct
    ## conversion to numeric type.

    txdat <- adjustLevels(txdat, bws$xdati)

    if (!no.ex)
      exdat <- adjustLevels(exdat, bws$xdati)

    ## grab the evaluation data before it is converted to numeric
    if(no.ex)
      teval <- txdat
    else
      teval <- exdat

    ## put the unordered, ordered, and continuous data in their own objects
    ## data that is not a factor is continuous.

    txdat = toMatrix(txdat)

    if (!no.ex){
      exdat = toMatrix(exdat)
    }

    ## from this point on txdat and exdat have been recast as matrices

    ## First, create the scalar index (n \times 1 vector)

    index <- as.vector(txdat %*% bws$beta)

    if(no.ex) {
      index.eval <- index
      exdat <- txdat
      eydat <- tydat
    } else {
      index.eval <- as.vector(exdat %*% bws$beta)
    }
    index.df <- data.frame(index = index)
    index.eval.df <- data.frame(index = index.eval)

    spec <- .npindex_resolve_spec(bws, where = "npindex")
    regtype <- spec$regtype.engine
    lc.fixed.progress.route <- identical(bws$method, "ichimura") &&
      identical(regtype, "lc") &&
      identical(bws$type, "fixed") &&
      !gradients &&
      !errors &&
      !residuals &&
      (no.ex || (!no.ex && no.ey)) &&
      (fit.progress.allow || fit.progress.handoff)
    npreg.idx.args <- list(
      txdat = index.df,
      tydat = tydat,
      bws = bws$bw,
      bwtype = bws$type,
      ckertype = bws$ckertype,
      ckerorder = bws$ckerorder,
      regtype = regtype,
      warn.glp.gradient = FALSE
    )
    if (identical(regtype, "lp")) {
      npreg.idx.args$basis <- spec$basis.engine
      npreg.idx.args$degree <- spec$degree.engine
      npreg.idx.args$bernstein.basis <- spec$bernstein.basis.engine
    }
    npreg.idx.bw <- if (identical(regtype, "lp") || lc.fixed.progress.route) {
      .np_index_regression_bandwidth(
        index.df = index.df,
        ydat = tydat,
        bws = bws,
        spec = spec
      )
    } else {
      NULL
    }
    next_npreg_fit_args <- function(exdat = NULL, gradients = FALSE) {
      args <- if (identical(regtype, "lp") || lc.fixed.progress.route) {
        c(
          list(
            bws = npreg.idx.bw,
            txdat = index.df,
            tydat = tydat,
            gradients = gradients,
            warn.glp.gradient = FALSE
          ),
          if (!is.null(exdat)) list(exdat = exdat) else list()
        )
      } else {
        c(
          npreg.idx.args,
          list(gradients = gradients),
          if (!is.null(exdat)) list(exdat = exdat) else list()
        )
      }
      if (fit.progress.handoff) {
        args$.np_fit_progress_handoff <- TRUE
        fit.progress.handoff <<- FALSE
      }
      args
    }

    fast.largeh <- FALSE
    fast.largeh.eval.mean <- NULL
    fast.largeh.train.mean <- NULL
    if (identical(regtype, "lc") && !gradients && !errors && !lc.fixed.progress.route) {
      gate.index <- if (no.ex) index else c(index, index.eval)
      fast.largeh <- .npindexbw_fast_eligible(
        h = as.double(bws$bw),
        bws = bws,
        eval.index = gate.index
      )
      if (fast.largeh) {
        fast.largeh.eval.mean <- {
          tww.fast <- npksum(
            txdat = index.df,
            tydat = as.matrix(data.frame(tydat, 1)),
            weights = as.matrix(data.frame(tydat, 1)),
            exdat = index.eval.df[1L, , drop = FALSE],
            bws = bws$bw,
            bwtype = bws$type,
            ckertype = bws$ckertype,
            ckerorder = bws$ckerorder
          )$ksum
          as.double(tww.fast[1, 2, 1L] / NZD(tww.fast[2, 2, 1L]))
        }

        if (!no.ex && (no.ey || residuals)) {
          fast.largeh.train.mean <- {
            tww.fast <- npksum(
              txdat = index.df,
              tydat = as.matrix(data.frame(tydat, 1)),
              weights = as.matrix(data.frame(tydat, 1)),
              exdat = index.df[1L, , drop = FALSE],
              bws = bws$bw,
              bwtype = bws$type,
              ckertype = bws$ckertype,
              ckerorder = bws$ckerorder
            )$ksum
            as.double(tww.fast[1, 2, 1L] / NZD(tww.fast[2, 2, 1L]))
          }
        }
      }
    }

    ## Next, if no gradients are requested, use (faster) npksum

    if(gradients==FALSE) {
      if (identical(regtype, "lc") && !lc.fixed.progress.route) {
        if (fast.largeh) {
          index.mean <- rep.int(fast.largeh.eval.mean, length(index.eval))
        } else {
          tww <- npksum(txdat=index.df,
                        tydat=as.matrix(data.frame(tydat,1)),
                        weights=as.matrix(data.frame(tydat,1)),
                        exdat=index.eval.df,
                        bws=bws$bw,
                        bwtype = bws$type,
                        ckertype = bws$ckertype,
                        ckerorder = bws$ckerorder)$ksum

          index.mean <- tww[1,2,]/NZD(tww[2,2,])
        }

        if (!no.ex && (no.ey || residuals)) {

          ## want to evaluate on training data for in sample errors even
          ## if evaluation x's are different from training but no y's
          ## are specified

          if (fast.largeh) {
            index.tmean <- rep.int(fast.largeh.train.mean, length(tydat))
          } else {
            tww <- npksum(txdat=index.df,
                          tydat=as.matrix(data.frame(tydat,1)),
                          weights=as.matrix(data.frame(tydat,1)),
                          exdat=index.df,
                          bws=bws$bw,
                          bwtype = bws$type,
                          ckertype = bws$ckertype,
                          ckerorder = bws$ckerorder)$ksum

            index.tmean <- tww[1,2,]/NZD(tww[2,2,])
          }

        }
      } else {
        model <- do.call(npreg, next_npreg_fit_args(
          exdat = index.eval.df,
          gradients = FALSE
        ))
        index.mean <- model$mean

        if (!no.ex && (no.ey || residuals)) {
          model <- do.call(npreg, next_npreg_fit_args(
            gradients = FALSE
          ))
          index.tmean <- model$mean
        }
      }

    } else if(gradients==TRUE) {
      model <- do.call(npreg, next_npreg_fit_args(
        exdat = index.eval.df,
        gradients = TRUE
      ))

      index.mean <- model$mean

      ## index.grad is a matrix, one column for each variable, each
      ## equal to its coefficient beta_i times the first derivative of
      ## the local-constant model

      index.grad <- as.matrix(model$grad)%*%t(as.vector(bws$beta))

      if (!no.ex) {

        ## Want to evaluate on training data for in sample errors even
        ## if evaluation x's are different from training but no y's
        ## are specified. Also, needed for variance-covariance matrix
        ## (uses on ly the training data)

        model <- do.call(npreg, next_npreg_fit_args(
          gradients = TRUE
        ))

        index.tmean <- model$mean

        index.tgrad <- model$grad

      }

    }

    if (no.ex) {
      index.tmean <- index.mean
    }

    if (no.ex && gradients) {
      index.tgrad <- index.grad
    }

    ## 5/3/2010, jracine, added vcov methods... thanks to Juan Carlos
    ## Escanciano <jescanci@indiana.edu> for pushing me on this for
    ## the Klein and Spady estimator... use index.tmean, index.tgrad
    ## (training X) - need gradients == TRUE in order for this to
    ## work.

    if (bws$method == "ichimura" && gradients) {

      ## First row & column of covariance matrix `Bvcov' are zero due
      ## to identification condition that beta_1=0. Note the n n^{-1}
      ## n in V^{-1}\Sigma V^{-1} and the \sqrt{n} in the
      ## normalization of \hat\beta will cancel.

      q <- ncol(txdat)
      Bvcov <- matrix(0,q,q)
      dimnames(Bvcov) <- list(bws$xnames,bws$xnames)

      ## Use the weight matrix so we can compute all expectations with
      ## only one call to npksum (the kernel arguments x\beta do not
      ## change, only the j for X_{ij} in E(X_{ij}|X_i'\beta)

      W <- txdat[,-1,drop=FALSE]

      tyindex <- npksum(txdat = index.df,
                        tydat = rep(1,length(tydat)),
                        weights = W,
                        bws = bws$bw,
                        bwtype = bws$type,
                        ckertype = bws$ckertype,
                        ckerorder = bws$ckerorder)$ksum

      tindex <- npksum(txdat = index.df,
                       bws = bws$bw,
                       bwtype = bws$type,
                       ckertype = bws$ckertype,
                       ckerorder = bws$ckerorder)$ksum

      ## Need to trap case where k-1=1... ksum will return a 1 D
      ## array, need a 1 x n matrix

      if(length(dim(tyindex))==1) tyindex <- matrix(tyindex,nrow=1,ncol=dim(tyindex))

      ## xmex = X_i-\hat E(X_i|X_i'\beta), dimension k\times n.

      xmex <- sapply(seq_along(tydat),function(i){W[i,]-tyindex[,i]/tindex[i]})

      ## Need to trap case where k-1=1..., sapply will return a
      ## vector, need a 1 x n matrix

      ## Need to trap case where k-1=1..., sapply will return a
      ## vector, need a 1 x n matrix
      if(is.vector(xmex)) {
        dg.db.xmex <- matrix(index.tgrad[,1]*xmex,nrow=1,ncol=length(xmex))
      } else {
        dg.db.xmex <- index.tgrad[,1]*xmex
      }

      uhat <- tydat - index.tmean ## Training y and training mean

      Vinv <- chol2inv(chol(dg.db.xmex%*%t(dg.db.xmex)))

      Sigma <- (uhat*dg.db.xmex)%*%t(uhat*dg.db.xmex)

      Bvcov[-1,-1] <- Vinv %*% Sigma %*% Vinv

      dimnames(Bvcov) <- list(bws$xnames,bws$xnames)

      ## Now export this in an S3 method...

    } else if (bws$method == "kleinspady" && gradients) {

      ## We divide by P(1-P) so test for P=0 or 1...

      keep <- which(index.tmean < 1 & index.tmean > 0)
      dg.db <- txdat[,-1,drop=FALSE]*index.tgrad[,1]

      ## First row & column of covariance matrix are zero due to
      ## identification condition that beta_1=0. Note the n^{-1} in
      ## the E and the \sqrt{n} in the normalization of \hat\beta will
      ## cancel.

      q <- ncol(txdat)
      Bvcov <- matrix(0,q,q)
      Bvcov[-1,-1] <- chol2inv(chol(t(dg.db[keep,])%*%(dg.db[keep,]/(index.tmean[keep]*
        (1-index.tmean[keep])))))

      dimnames(Bvcov) <- list(bws$xnames,bws$xnames)

      ## Now export this in an S3 method...

    }

    if (gradients){
      boofun = function(data, indices){
        rindex <- txdat[indices,] %*% bws$beta
        rindex.df <- data.frame(index = as.vector(rindex))
        boot.args <- list(
          txdat = rindex.df,
          tydat = tydat[indices],
          exdat = index.eval.df,
          bws = bws$bw,
          bwtype = bws$type,
          ckertype = bws$ckertype,
          ckerorder = bws$ckerorder,
          regtype = regtype,
          gradients = TRUE,
          warn.glp.gradient = FALSE
        )
        if (identical(regtype, "lp")) {
          boot.args$basis <- spec$basis.engine
          boot.args$degree <- spec$degree.engine
          boot.args$bernstein.basis <- spec$bernstein.basis.engine
        }
        model <- do.call(npreg, boot.args)[c('mean','grad')]
        
        c(model$mean, model$grad, mean(model$grad))
      }

    } else {
      boofun = function(data, indices){
        rindex = txdat[indices,] %*% bws$beta
        if (identical(regtype, "lc")) {
          rindex.df <- data.frame(index = as.vector(rindex))
          tww <- npksum(txdat = rindex.df,
                        tydat = cbind(tydat[indices],1),
                        weights = cbind(tydat[indices],1),
                        exdat = index.eval.df,
                        bws = bws$bw,
                        bwtype = bws$type,
                        ckertype = bws$ckertype,
                        ckerorder = bws$ckerorder)$ksum

          tww[1,2,]/NZD(tww[2,2,])
        } else {
          rindex.df <- data.frame(index = as.vector(rindex))
          boot.args <- list(
            txdat = rindex.df,
            tydat = tydat[indices],
            exdat = index.eval.df,
            bws = bws$bw,
            bwtype = bws$type,
            ckertype = bws$ckertype,
            ckerorder = bws$ckerorder,
            regtype = regtype,
            gradients = FALSE,
            warn.glp.gradient = FALSE
          )
          if (identical(regtype, "lp")) {
            boot.args$basis <- spec$basis.engine
            boot.args$degree <- spec$degree.engine
            boot.args$bernstein.basis <- spec$bernstein.basis.engine
          }
          do.call(npreg, boot.args)$mean
        }
        
      }
    }

    if (errors){

      boot.out = suppressWarnings(boot(data.frame(txdat,tydat), boofun, R = boot.num))

      index.merr = matrix(data = 0, ncol = 1, nrow = length(index.eval))
      index.merr[,] = .np_plot_bootstrap_col_sds(boot.out$t[, seq_len(length(index.eval)), drop = FALSE])

      if (gradients) {
        index.gerr = matrix(data = 0, ncol = ncol(txdat), nrow = length(index.eval))
        index.gerr[,] = .np_plot_bootstrap_col_sds(
          boot.out$t[, (length(index.eval) + 1):(2 * length(index.eval)), drop = FALSE]
        )

        for (i in seq_len(ncol(txdat)))
          index.gerr[,i] = abs(bws$beta[i])*index.gerr[,i]

        index.mgerr = sd(boot.out$t[,2*length(index.eval)+1])
        index.mgerr = abs(bws$beta)*index.mgerr
      }
    }
    ## goodness of fit

    if(bws$method == "ichimura") {
      if (!no.ey) {
        RSQ = RSQfunc(eydat,index.mean)
        MSE = MSEfunc(eydat,index.mean)
        MAE = MAEfunc(eydat,index.mean)
        MAPE = MAPEfunc(eydat,index.mean)
        CORR = if (fast.largeh) suppressWarnings(CORRfunc(eydat,index.mean)) else CORRfunc(eydat,index.mean)
        SIGN = SIGNfunc(eydat,index.mean)
      } else {
        RSQ = RSQfunc(tydat,index.tmean)
        MSE = MSEfunc(tydat,index.tmean)
        MAE = MAEfunc(tydat,index.tmean)
        MAPE = MAPEfunc(tydat,index.tmean)
        CORR = if (fast.largeh) suppressWarnings(CORRfunc(tydat,index.tmean)) else CORRfunc(tydat,index.tmean)
        SIGN = SIGNfunc(tydat,index.tmean)
      }
      strgof = "xtra=c(RSQ,MSE,MAE,MAPE,CORR,SIGN),"
      strres = (if (residuals) "resid = tydat - index.tmean," else "")
    } else if(bws$method == "kleinspady") {
      index.pred =
        if (!no.ey) round(index.mean)
        else round(index.tmean)

      confusion.matrix =
        table(if (!no.ey) eydat else tydat,
              index.pred, dnn=c("Actual", "Predicted"))

      CCR.overall <- sum(diag(confusion.matrix))/sum(confusion.matrix)
      CCR.byoutcome <- diag(confusion.matrix)/rowSums(confusion.matrix)

      fit.mcfadden <- confusion.matrix/sum(confusion.matrix)

      fit.mcfadden <- sum(diag(fit.mcfadden)) -
        (sum(fit.mcfadden^2)-sum(diag(fit.mcfadden)^2))

      strgof = "confusion.matrix = confusion.matrix, CCR.overall = CCR.overall,
           CCR.byoutcome =  CCR.byoutcome, fit.mcfadden = fit.mcfadden,"
      strres = ""
    }

    ev.args <- list(
      bws = bws,
      index = index.eval,
      mean = index.mean,
      ntrain = nrow(txdat),
      trainiseval = no.ex,
      residuals = residuals,
      gradients = gradients
    )
    if (errors)
      ev.args$merr <- index.merr
    if (gradients) {
      ev.args$grad <- index.grad
      ev.args$mean.grad <- colMeans(index.grad)
      ev.args$betavcov <- Bvcov
    }
    if (errors && gradients) {
      ev.args$gerr <- index.gerr
      ev.args$mean.gerr <- index.mgerr
    }
    if (bws$method == "ichimura") {
      if (residuals)
        ev.args$resid <- tydat - index.tmean
      ev.args$xtra <- c(RSQ, MSE, MAE, MAPE, CORR, SIGN)
    } else if (bws$method == "kleinspady") {
      ev.args$confusion.matrix <- confusion.matrix
      ev.args$CCR.overall <- CCR.overall
      ev.args$CCR.byoutcome <- CCR.byoutcome
      ev.args$fit.mcfadden <- fit.mcfadden
    }
    ev <- do.call(singleindex, ev.args)
    fit.elapsed <- proc.time()[3] - fit.start
    optim.time <- if (!is.null(bws$total.time) && is.finite(bws$total.time)) as.double(bws$total.time) else NA_real_
    total.time <- fit.elapsed + (if (is.na(optim.time)) 0.0 else optim.time)
    ev$timing <- bws$timing
    ev$total.time <- total.time
    ev$optim.time <- optim.time
    ev$fit.time <- fit.elapsed
    ev$nomad.time <- if (!is.null(bws$nomad.time) && is.finite(bws$nomad.time)) as.double(bws$nomad.time) else NA_real_
    ev$powell.time <- if (!is.null(bws$powell.time) && is.finite(bws$powell.time)) as.double(bws$powell.time) else NA_real_
    ev
  }

test_that("npplreg nomad shortcut matches the explicit partially linear preset", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260322)
  n <- 22
  dat <- data.frame(x = rnorm(n), z = sort(runif(n)))
  dat$y <- 1 + 0.75 * dat$x + sin(2 * pi * dat$z) + rnorm(n, sd = 0.08)

  bw_short <- np::npplregbw(
    y ~ x | z,
    data = dat,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L
  )
  bw_long <- np::npplregbw(
    y ~ x | z,
    data = dat,
    regtype = "lp",
    search.engine = "nomad+powell",
    degree.select = "coordinate",
    bernstein.basis = TRUE,
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = FALSE,
    bwtype = "fixed",
    nmulti = 1L
  )

  fit_short <- np::npplreg(
    y ~ x | z,
    data = dat,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L
  )
  fit_long <- np::npplreg(
    y ~ x | z,
    data = dat,
    regtype = "lp",
    search.engine = "nomad+powell",
    degree.select = "coordinate",
    bernstein.basis = TRUE,
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = FALSE,
    bwtype = "fixed",
    nmulti = 1L
  )

  expect_identical(as.integer(bw_short$degree), as.integer(bw_long$degree))
  expect_true(is.list(bw_short$nomad.shortcut))
  expect_equal(fitted(fit_short), fitted(fit_long), tolerance = 5e-4)
  expect_true(is.list(fit_short$bws$nomad.shortcut))
})

test_that("npscoef nomad shortcut matches the explicit smooth-coefficient preset", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260322)
  n <- 22
  xdat <- data.frame(x = runif(n))
  zdat <- data.frame(z = sort(runif(n)))
  y <- (1 + zdat$z^2) * xdat$x + rnorm(n, sd = 0.08)

  bw_short <- np::npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L
  )
  bw_long <- np::npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    search.engine = "nomad+powell",
    degree.select = "coordinate",
    bernstein.basis = TRUE,
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = FALSE,
    bwtype = "fixed",
    nmulti = 1L
  )

  fit_short <- np::npscoef(
    txdat = xdat,
    tzdat = zdat,
    tydat = y,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L,
    errors = FALSE,
    betas = FALSE
  )
  fit_long <- np::npscoef(
    txdat = xdat,
    tzdat = zdat,
    tydat = y,
    regtype = "lp",
    search.engine = "nomad+powell",
    degree.select = "coordinate",
    bernstein.basis = TRUE,
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = FALSE,
    bwtype = "fixed",
    nmulti = 1L,
    errors = FALSE,
    betas = FALSE
  )

  expect_identical(as.integer(bw_short$degree), as.integer(bw_long$degree))
  expect_true(is.list(bw_short$nomad.shortcut))
  expect_equal(fitted(fit_short), fitted(fit_long), tolerance = 5e-4)
  expect_true(is.list(fit_short$bws$nomad.shortcut))
})

test_that("npindex nomad shortcut matches the explicit single-index preset", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260322)
  n <- 24
  xdat <- data.frame(x1 = runif(n, -1, 1), x2 = runif(n, -1, 1))
  index <- xdat$x1 + 0.75 * xdat$x2
  y <- sin(index) + 0.25 * index^2 + rnorm(n, sd = 0.05)

  bw_short <- np::npindexbw(
    xdat = xdat,
    ydat = y,
    method = "ichimura",
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L
  )
  bw_long <- np::npindexbw(
    xdat = xdat,
    ydat = y,
    method = "ichimura",
    regtype = "lp",
    search.engine = "nomad+powell",
    degree.select = "coordinate",
    bernstein.basis = TRUE,
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = FALSE,
    bwtype = "fixed",
    nmulti = 1L
  )

  fit_short <- np::npindex(
    txdat = xdat,
    tydat = y,
    method = "ichimura",
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L
  )
  fit_long <- np::npindex(
    txdat = xdat,
    tydat = y,
    method = "ichimura",
    regtype = "lp",
    search.engine = "nomad+powell",
    degree.select = "coordinate",
    bernstein.basis = TRUE,
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = FALSE,
    bwtype = "fixed",
    nmulti = 1L
  )

  expect_identical(as.integer(bw_short$degree), as.integer(bw_long$degree))
  expect_true(is.list(bw_short$nomad.shortcut))
  expect_equal(fitted(fit_short), fitted(fit_long), tolerance = 5e-4)
  expect_true(is.list(fit_short$bws$nomad.shortcut))

  dat <- data.frame(y = y, x1 = xdat$x1, x2 = xdat$x2)
  fit_formula <- np::npindex(
    y ~ x1 + x2,
    data = dat,
    method = "ichimura",
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L
  )
  expect_true(is.list(fit_formula$bws$nomad.shortcut))
  expect_true(isTRUE(fit_formula$bws$nomad.shortcut$enabled))
})

test_that("semiparametric NOMAD shortcut routes accept nomad.nmulti", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260323)
  n <- 18
  dat <- data.frame(x = rnorm(n), z = sort(runif(n)))
  dat$y <- 1 + 0.5 * dat$x + sin(2 * pi * dat$z) + rnorm(n, sd = 0.05)

  fit_pl <- np::npplreg(
    y ~ x | z,
    data = dat,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L,
    nomad.nmulti = 2L
  )
  fit_sc <- np::npscoef(
    txdat = data.frame(x = dat$x),
    tzdat = data.frame(z = dat$z),
    tydat = dat$y,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L,
    nomad.nmulti = 2L,
    errors = FALSE,
    betas = FALSE
  )
  fit_si <- np::npindex(
    txdat = data.frame(x1 = dat$x, x2 = dat$z),
    tydat = dat$y,
    method = "ichimura",
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L,
    nomad.nmulti = 2L
  )

  expect_true(is.list(fit_pl$bws$nomad.shortcut))
  expect_true(is.list(fit_sc$bws$nomad.shortcut))
  expect_true(is.list(fit_si$bws$nomad.shortcut))
})

test_that("semiparametric nomad shortcuts fail fast on incompatible settings", {
  xdat <- data.frame(x = runif(10))
  zdat <- data.frame(z = runif(10))
  y <- runif(10)

  expect_error(
    np::npplregbw(xdat = xdat, zdat = zdat, ydat = y, nomad = TRUE, regtype = "ll"),
    "nomad=TRUE requires regtype='lp'"
  )
  expect_error(
    np::npscoefbw(xdat = xdat, zdat = zdat, ydat = y, nomad = TRUE, bernstein.basis = FALSE),
    "requires bernstein.basis=TRUE"
  )
  expect_error(
    np::npindexbw(xdat = data.frame(x1 = runif(10), x2 = runif(10)), ydat = y, nomad = TRUE, bwtype = "adaptive_nn"),
    "requires bwtype='fixed'"
  )
  expect_error(
    np::npplregbw(xdat = xdat, zdat = zdat, ydat = y, nomad.nmulti = 1L),
    "nomad.nmulti is only supported when regtype='lp', automatic degree search is active, and search.engine is 'nomad' or 'nomad\\+powell'"
  )
})

align_singleindex_grad_index <- function(fit, target_index, beta) {
  fit_index <- as.vector(fit$index)
  ord <- match(target_index, fit_index)
  if (anyNA(ord))
    stop("failed to align single-index gradient fit to target index", call. = FALSE)
  as.matrix(fit$grad)[ord, 1L] / as.numeric(beta[1L])
}

explicit_singleindex_grad_index <- function(bw, tx, y, counts_vec) {
  idx <- rep.int(seq_len(nrow(tx)), counts_vec)
  target_index <- as.vector(as.matrix(tx) %*% bw$beta)
  fit <- npindex(
    bws = bw,
    txdat = tx[idx, , drop = FALSE],
    tydat = y[idx],
    exdat = tx,
    gradients = TRUE
  )
  align_singleindex_grad_index(fit = fit, target_index = target_index, beta = bw$beta)
}

test_that("fixed single-index gradient helper matches duplicate-sample refits", {
  helper <- getFromNamespace(".np_inid_boot_from_index", "np")

  set.seed(20260310)
  n <- 36
  tx <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- sin(2 * tx$x1 - tx$x2) + rnorm(n, sd = 0.05)

  counts <- cbind(
    rep(1L, n),
    c(rep(2L, 6), rep(0L, 6), rep(1L, n - 12)),
    c(rep(0L, 3), rep(3L, 3), rep(1L, n - 6)),
    sample.int(3L, n, replace = TRUE)
  )

  normalize_counts <- function(v, n) {
    out <- as.integer(v)
    while (sum(out) < n)
      out[which.min(out)] <- out[which.min(out)] + 1L
    while (sum(out) > n) {
      hit <- which(out > 0L)[1L]
      out[hit] <- out[hit] - 1L
    }
    out
  }

  counts <- apply(counts, 2L, normalize_counts, n = n)
  if (!is.matrix(counts))
    counts <- matrix(counts, nrow = n)

  cfgs <- list(
    list(regtype = "lc", label = "lc"),
    list(regtype = "ll", label = "ll"),
    list(regtype = "lp", basis = "tensor", degree = 2L, label = "lp")
  )

  for (cfg in cfgs) {
    bw.args <- list(
      xdat = tx,
      ydat = y,
      bws = c(1, 1, 0.25),
      bandwidth.compute = FALSE,
      regtype = cfg$regtype,
      bwtype = "fixed"
    )
    if (!is.null(cfg$basis)) {
      bw.args$basis <- cfg$basis
      bw.args$degree <- cfg$degree
    }
    bw <- do.call(npindexbw, bw.args)

    boot <- helper(
      xdat = tx,
      ydat = y,
      bws = bw,
      B = ncol(counts),
      counts = counts,
      gradients = TRUE
    )

    oracle_t0 <- explicit_singleindex_grad_index(
      bw = bw,
      tx = tx,
      y = y,
      counts_vec = rep(1L, n)
    )
    oracle_t <- vapply(
      seq_len(ncol(counts)),
      function(j) explicit_singleindex_grad_index(
        bw = bw,
        tx = tx,
        y = y,
        counts_vec = counts[, j]
      ),
      numeric(n)
    )

    expect_equal(
      as.vector(boot$t0),
      oracle_t0,
      tolerance = 1e-10,
      info = paste(cfg$label, "t0 parity")
    )
    expect_equal(
      boot$t,
      t(oracle_t),
      tolerance = 1e-8,
      info = paste(cfg$label, "counts parity")
    )
  }
})

test_that("fixed single-index gradient helper counts.drawer matches counts matrix", {
  helper <- getFromNamespace(".np_inid_boot_from_index", "np")

  set.seed(20260310)
  n <- 30
  tx <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- sin(tx$x1 + tx$x2) + rnorm(n, sd = 0.06)
  bw <- npindexbw(
    xdat = tx,
    ydat = y,
    bws = c(1, 1, 0.25),
    bandwidth.compute = FALSE,
    regtype = "ll",
    bwtype = "fixed"
  )

  counts <- cbind(
    rep(1L, n),
    c(rep(2L, 5), rep(0L, 5), rep(1L, n - 10)),
    c(rep(0L, 4), rep(2L, 4), rep(1L, n - 8))
  )
  if (!is.matrix(counts))
    counts <- matrix(counts, nrow = n)

  drawer <- function(start, stop) counts[, start:stop, drop = FALSE]

  boot.counts <- helper(
    xdat = tx,
    ydat = y,
    bws = bw,
    B = ncol(counts),
    counts = counts,
    gradients = TRUE
  )
  boot.drawer <- helper(
    xdat = tx,
    ydat = y,
    bws = bw,
    B = ncol(counts),
    counts.drawer = drawer,
    gradients = TRUE
  )

  expect_equal(boot.drawer$t0, boot.counts$t0, tolerance = 1e-12)
  expect_equal(boot.drawer$t, boot.counts$t, tolerance = 1e-12)
})

test_that("sibandwidth fixed gradient bootstrap now works for helper methods", {
  set.seed(20260310)
  n <- 30
  tx <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- sin(tx$x1 + tx$x2) + rnorm(n, sd = 0.07)
  bw <- npindexbw(
    xdat = tx,
    ydat = y,
    bws = c(1, 1, 0.25),
    bandwidth.compute = FALSE,
    regtype = "ll",
    bwtype = "fixed"
  )

  for (boot.method in c("inid", "fixed", "geom")) {
    out <- suppressWarnings(plot(
      bw,
      xdat = tx,
      ydat = y,
      output = "data",
      perspective = FALSE,
      gradients = TRUE,
      errors = "bootstrap",
      bootstrap = boot.method,
      B = 9L
    ))[[1]]

    expect_true(all(is.finite(out$grad)))
    expect_true(all(is.finite(out$glerr)))
    expect_true(all(is.finite(out$gherr)))
  }
})

test_that("single-index nonfixed gradient helper methods still fail fast", {
  set.seed(20260310)
  n <- 30
  tx <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- sin(tx$x1 + tx$x2) + rnorm(n, sd = 0.08)

  for (bt in c("generalized_nn", "adaptive_nn")) {
    bw <- npindexbw(
      xdat = tx,
      ydat = y,
      bws = c(1, 1, 5L),
      bandwidth.compute = FALSE,
      regtype = "ll",
      bwtype = bt
    )

    for (boot.method in c("inid", "fixed", "geom")) {
      expect_error(
        suppressWarnings(plot(
          bw,
          xdat = tx,
          ydat = y,
          output = "data",
          perspective = FALSE,
          gradients = TRUE,
          errors = "bootstrap",
          bootstrap = boot.method,
          B = 9L
        )),
        "requires helper mode with gradients=FALSE",
        info = paste(bt, boot.method, "gradient helper guard")
      )
    }
  }
})

test_that("named bws formula dispatch matches positional npindex route", {
  set.seed(20260323)
  dat <- data.frame(
    y = rnorm(24),
    x1 = runif(24),
    x2 = runif(24)
  )
  ex <- data.frame(
    x1 = seq(0.1, 0.9, length.out = 8),
    x2 = seq(0.9, 0.1, length.out = 8)
  )

  fit.pos <- npindex(
    y ~ x1 + x2,
    data = dat,
    newdata = ex,
    method = "ichimura"
  )
  fit.named <- npindex(
    bws = y ~ x1 + x2,
    data = dat,
    newdata = ex,
    method = "ichimura"
  )

  expect_equal(as.numeric(fit.named$mean), as.numeric(fit.pos$mean), tolerance = 0)
  expect_equal(as.numeric(fit.named$merr), as.numeric(fit.pos$merr), tolerance = 0)
})

test_that("npindex basic functionality works", {
  set.seed(42)
  n <- 100
  x1 <- runif(n)
  x2 <- runif(n)
  # Single index model: y = g(x1 + x2) + e
  y <- (x1 + x2)^2 + rnorm(n, sd=0.1)
  
  mydat <- data.frame(y, x1, x2)
  bw <- npindexbw(
    y ~ x1 + x2,
    data = mydat,
    bws = c(1, 0.35, 0.45),
    bandwidth.compute = FALSE,
    method = "ichimura"
  )
  
  model <- npindex(bws=bw)
  
  expect_s3_class(model, "singleindex")
  expect_type(predict(model), "double")
  expect_output(summary(model))
})

test_that("npindexbw nearest-neighbor selection stores integer support and exact fits stay public-green", {
  set.seed(314163)
  n <- 70L
  x1 <- rnorm(n)
  x2 <- rnorm(n)
  y <- x1 - x2 + rnorm(n, sd = 0.2)
  tx <- data.frame(x1 = x1, x2 = x2)

  for (bt in c("generalized_nn", "adaptive_nn")) {
    bw <- npindexbw(xdat = tx, ydat = y, regtype = "lc", bwtype = bt, nmulti = 1)
    fit <- npindex(bws = bw, txdat = tx, tydat = y, gradients = FALSE)

    expect_true(bw$bw >= 1, info = bt)
    expect_equal(bw$bw, as.double(as.integer(bw$bw)), tolerance = 0, info = bt)
    expect_true(all(is.finite(fit$mean)), info = bt)
  }
})

test_that("manual single-index nearest-neighbor bandwidths fail fast when not integer support", {
  tx <- data.frame(x1 = seq(0.1, 0.9, length.out = 8L), x2 = seq(0.9, 0.1, length.out = 8L))
  y <- tx$x1 - tx$x2

  expect_error(
    npindexbw(
      xdat = tx,
      ydat = y,
      bws = c(1, 1, 0.5),
      bandwidth.compute = FALSE,
      bwtype = "adaptive_nn"
    ),
    "nearest-neighbor bandwidth must be an integer"
  )
})

test_that("npindex supports residual and error branches with evaluation y data", {
  set.seed(20260309)
  n <- 56L
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.05)
  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- tx[seq_len(18L), , drop = FALSE]
  ey <- sin(ex$x1 + ex$x2) + rnorm(nrow(ex), sd = 0.05)

  for (cfg in list(
    list(regtype = "lc"),
    list(regtype = "ll"),
    list(regtype = "lp", basis = "glp", degree = 1L)
  )) {
    bw.args <- list(
      xdat = tx,
      ydat = y,
      regtype = cfg$regtype,
      bwtype = "adaptive_nn",
      nmulti = 1L
    )
    if (!is.null(cfg$basis)) {
      bw.args$basis <- cfg$basis
      bw.args$degree <- cfg$degree
      bw.args$bernstein.basis <- FALSE
    }

    bw <- do.call(npindexbw, bw.args)

    fit.is <- npindex(bws = bw, txdat = tx, tydat = y, gradients = FALSE, residuals = TRUE)
    expect_equal(length(fit.is$resid), nrow(tx), info = cfg$regtype)
    expect_equal(as.vector(fit.is$resid), y - as.vector(fit.is$mean), tolerance = 1e-8, info = cfg$regtype)

    fit.oos <- npindex(bws = bw, txdat = tx, tydat = y, exdat = ex, eydat = ey, gradients = FALSE, errors = TRUE)
    expect_equal(length(fit.oos$mean), nrow(ex), info = cfg$regtype)
    expect_equal(length(fit.oos$merr), nrow(ex), info = cfg$regtype)
    expect_true(all(is.finite(fit.oos$mean)), info = cfg$regtype)
    expect_true(all(is.finite(fit.oos$merr)), info = cfg$regtype)

    fit.grad <- npindex(bws = bw, txdat = tx, tydat = y, exdat = ex, eydat = ey, gradients = TRUE, errors = TRUE)
    expect_equal(dim(fit.grad$grad), c(nrow(ex), ncol(tx)), info = cfg$regtype)
    expect_equal(dim(fit.grad$gerr), c(nrow(ex), ncol(tx)), info = cfg$regtype)
    expect_true(all(is.finite(fit.grad$grad)), info = cfg$regtype)
    expect_true(all(is.finite(fit.grad$gerr)), info = cfg$regtype)
  }
})

test_that("npindex bootstrap error SD recovery matches covariance diagonal reduction", {
  boot.sds <- getFromNamespace(".np_plot_bootstrap_col_sds", "np")

  set.seed(20260311)
  boot.t <- matrix(rnorm(250 * 24), nrow = 250, ncol = 24)

  ref.mean <- sqrt(diag(cov(boot.t[, 1:12, drop = FALSE])))
  ref.grad <- sqrt(diag(cov(boot.t[, 13:24, drop = FALSE])))

  expect_equal(
    boot.sds(boot.t[, 1:12, drop = FALSE]),
    ref.mean,
    tolerance = 1e-15
  )
  expect_equal(
    boot.sds(boot.t[, 13:24, drop = FALSE]),
    ref.grad,
    tolerance = 1e-15
  )
})

test_that("predict.singleindex forwards explicit boot.num without changing fitted predictions", {
  set.seed(20260323)
  n <- 60L
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.05)
  tx <- data.frame(x1 = x1, x2 = x2)
  nd <- tx[seq_len(8L), , drop = FALSE]

  bw <- npindexbw(
    xdat = tx,
    ydat = y,
    bws = c(1, 0.35, 0.45),
    bandwidth.compute = FALSE,
    method = "ichimura"
  )
  fit <- npindex(bws = bw, txdat = tx, tydat = y)

  pred <- predict(fit, newdata = nd)
  pred.se <- predict(fit, newdata = nd, se.fit = TRUE, boot.num = 10)

  expect_equal(as.numeric(pred.se$fit), as.numeric(pred), tolerance = 0)
  expect_equal(length(pred.se$se.fit), nrow(nd))
})

test_that("npindex NOMAD shortcut stays near the Powell basin on the canonical Ichimura repro", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(42)
  n <- 500
  dat <- data.frame(
    x1 = runif(n, min = -1, max = 1),
    x2 = runif(n, min = -1, max = 1)
  )
  dat$y <- dat$x1 - dat$x2 + rnorm(n)

  fit_powell <- np::npindex(
    txdat = dat[c("x1", "x2")],
    tydat = dat$y,
    method = "ichimura",
    regtype = "lp",
    bernstein = TRUE,
    degree = 1L
  )
  fit_nomad <- np::npindex(
    txdat = dat[c("x1", "x2")],
    tydat = dat$y,
    method = "ichimura",
    nomad = TRUE
  )

  expect_identical(as.integer(fit_nomad$bws$degree), 1L)
  expect_equal(as.numeric(fit_nomad$bws$fval), as.numeric(fit_powell$bws$fval), tolerance = 0.02)
  expect_equal(as.numeric(fit_nomad$bws$beta[2L]), as.numeric(fit_powell$bws$beta[2L]), tolerance = 0.1)
  expect_true(is.finite(fit_nomad$bws$bw))
  expect_lt(as.numeric(fit_nomad$bws$bw), 5)
  expect_lt(as.numeric(fit_nomad$bws$beta[2L]), 0)
})

library(np)

quiet_eval <- function(expr) {
  value <- NULL
  capture.output(value <- force(expr))
  value
}

test_that("npindex frozen mean helper stays on the exact single-index scale", {
  set.seed(42)
  n <- 200L
  B <- 25L
  x <- runif(n, -1, 1)
  z <- rnorm(n)
  y <- x^2 + rnorm(n, sd = 0.25 * stats::sd(x))
  xdat <- data.frame(x = x, z = z)

  fit <- quiet_eval(
    npindex(
      y ~ x + z,
      bwtype = "adaptive_nn",
      nmulti = 1L
    )
  )

  eval_grid_fun <- getFromNamespace(".np_plot_singleindex_eval_grid", "np")
  to_frame <- getFromNamespace("toFrame", "np")
  exact_helper <- getFromNamespace(".np_inid_boot_from_index", "np")

  eval.info <- eval_grid_fun(
    bws = fit$bws,
    xdat = to_frame(xdat),
    neval = 40L,
    trim = 0.0
  )
  counts <- stats::rmultinom(B, size = n, prob = rep.int(1 / n, n))

  frozen <- exact_helper(
    xdat = xdat,
    ydat = y,
    bws = fit$bws,
    B = B,
    counts = counts,
    frozen = TRUE,
    idx.eval = eval.info$idx.eval,
    progress.label = "test frozen"
  )
  exact <- exact_helper(
    xdat = xdat,
    ydat = y,
    bws = fit$bws,
    B = B,
    counts = counts,
    frozen = FALSE,
    idx.eval = eval.info$idx.eval,
    progress.label = "test exact"
  )

  sd.ratio <- stats::na.omit(apply(frozen$t, 2L, stats::sd) / apply(exact$t, 2L, stats::sd))

  expect_equal(frozen$t0, exact$t0, tolerance = 1e-10)
  expect_true(length(sd.ratio) > 0L)
  expect_gt(stats::median(sd.ratio), 0.75)
  expect_lt(stats::median(sd.ratio), 1.25)
})

test_that("npindex public frozen plot-data mean stays on the exact scale", {
  set.seed(42)
  n <- 120L
  x <- runif(n, -1, 1)
  z <- rnorm(n)
  y <- x^2 + rnorm(n, sd = 0.25 * stats::sd(x))

  fit <- quiet_eval(
    npindex(
      y ~ x + z,
      bwtype = "adaptive_nn",
      nmulti = 1L
    )
  )

  get_obj <- function(mode) {
    suppressWarnings(plot(
      fit,
      output = "data",
      neval = 40L,
      errors = "bootstrap",
      bootstrap = "inid",
      boot_control = np_boot_control(nonfixed = mode),
      B = 39L,
      band = "pointwise"
    ))[[1L]]
  }

  frozen <- get_obj("frozen")
  exact <- get_obj("exact")
  ratio <- stats::median(abs(exact$merr[, 1L]) / pmax(abs(frozen$merr[, 1L]), 1e-12), na.rm = TRUE)

  expect_equal(frozen$mean, exact$mean, tolerance = 1e-12)
  expect_gt(ratio, 0.5)
  expect_lt(ratio, 2.0)
})

test_that("npindex nonfixed frozen bootstrap still supports gradient slices", {
  set.seed(42)
  n <- 80L
  x <- runif(n)
  z <- rnorm(n)
  y <- x + rnorm(n)

  fit <- quiet_eval(
    npindex(
      y ~ x + z,
      regtype = "ll",
      bwtype = "adaptive_nn",
      nmulti = 1
    )
  )

  tf <- tempfile(fileext = ".pdf")
  grDevices::pdf(tf)
  on.exit({
    try(grDevices::dev.off(), silent = TRUE)
    unlink(tf)
  }, add = TRUE)

  expect_no_error(
    capture.output(plot(
      fit,
      neval = 20L,
      errors = "bootstrap",
      bootstrap = "inid",
      boot_control = np_boot_control(nonfixed = "frozen"),
      B = 41L,
      band = "pointwise",
      gradients = TRUE
    ))
  )
})

npindex_fit_progress_time_counter <- function(start = 0, by = 1.7) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

npindex_fit_progress_time_values <- function(values) {
  force(values)
  i <- 0L
  function() {
    i <<- min(i + 1L, length(values))
    values[[i]]
  }
}

npindex_fit_progress_lines <- function(shadow) {
  vapply(shadow$trace, `[[`, character(1L), "line")
}

make_npindex_fit_progress_fixture <- function() {
  set.seed(20260404)
  n <- 24L
  dat <- data.frame(
    x1 = runif(n, -1, 1),
    x2 = runif(n, -1, 1)
  )
  index <- dat$x1 + 0.6 * dat$x2
  dat$y <- sin(index) + 0.2 * index^2 + rnorm(n, sd = 0.05)

  list(
    dat = dat,
    tx = dat[c("x1", "x2")],
    y = dat$y,
    bw = npindexbw(
      xdat = dat[c("x1", "x2")],
      ydat = dat$y,
      bws = c(1, 0.6, 0.35),
      method = "ichimura",
      regtype = "lp",
      degree = 1L,
      bernstein.basis = TRUE,
      bwtype = "fixed",
      bandwidth.compute = FALSE
    ),
    bw_lc_fixed = npindexbw(
      xdat = dat[c("x1", "x2")],
      ydat = dat$y,
      bws = c(1, 0.6, 0.35),
      method = "ichimura",
      regtype = "lc",
      bwtype = "fixed",
      bandwidth.compute = FALSE
    )
  )
}

test_that("npindex direct lp fit emits known-total fit progress without handoff", {
  fixture <- make_npindex_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npindex(
      bws = fixture$bw,
      txdat = fixture$tx,
      tydat = fixture$y
    ),
    force_renderer = "single_line",
    now = npindex_fit_progress_time_counter()
  )

  lines <- npindex_fit_progress_lines(actual)

  expect_s3_class(actual$value, "singleindex")
  expect_false(any(grepl(": starting$", lines)))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 1/24 \\([0-9]+\\.[0-9]%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 24/24 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )))
})

test_that("npindex direct lp fit stays silent below start grace without handoff", {
  fixture <- make_npindex_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0.75,
    np.progress.interval.known.sec = 0.5
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npindex(
      bws = fixture$bw,
      txdat = fixture$tx,
      tydat = fixture$y
    ),
    force_renderer = "single_line",
    now = npindex_fit_progress_time_values(c(0, 0.2, 0.4, 0.6))
  )

  expect_length(actual$trace, 0L)
})

test_that("npindex lp bw to fit route hands off into the regression fit surface", {
  fixture <- make_npindex_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npindex(
      y ~ x1 + x2,
      data = fixture$dat,
      method = "ichimura",
      regtype = "lp",
      degree.select = "coordinate",
      search.engine = "cell",
      degree.min = 0L,
      degree.max = 1L,
      bwtype = "fixed",
      nmulti = 1L
    ),
    force_renderer = "single_line",
    now = npindex_fit_progress_time_counter()
  )

  lines <- npindex_fit_progress_lines(actual)
  bandwidth.pos <- grep("^\\[np\\] Bandwidth selection \\(", lines)
  fit.start.pos <- grep(
    "^\\[np\\] Fitting regression 0/24 \\(0\\.0%, elapsed 0\\.0s, eta 0\\.0s\\): starting$",
    lines
  )
  fit.finish.pos <- grep(
    "^\\[np\\] Fitting regression 24/24 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )

  expect_s3_class(actual$value, "singleindex")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(fit.start.pos) == 1L)
  expect_true(length(fit.finish.pos) >= 1L)
  expect_lt(max(bandwidth.pos), fit.start.pos[[1L]])
  expect_lt(fit.start.pos[[1L]], fit.finish.pos[[1L]])
})

test_that("npindex lp nomad to powell to fit route preserves single-line handoff", {
  skip_if_not_installed("crs")

  fixture <- make_npindex_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npindex(
      y ~ x1 + x2,
      data = fixture$dat,
      method = "ichimura",
      nomad = TRUE,
      degree.min = 0L,
      degree.max = 1L,
      nmulti = 1L
    ),
    force_renderer = "single_line",
    now = npindex_fit_progress_time_counter()
  )

  lines <- npindex_fit_progress_lines(actual)
  bandwidth.pos <- grep("^\\[np\\] Selecting degree and bandwidth", lines)
  powell.pos <- grep("^\\[np\\] Refining bandwidth \\(", lines)
  fit.start.pos <- grep(
    "^\\[np\\] Fitting regression 0/24 \\(0\\.0%, elapsed 0\\.0s, eta 0\\.0s\\): starting$",
    lines
  )
  fit.finish.pos <- grep(
    "^\\[np\\] Fitting regression 24/24 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )

  expect_s3_class(actual$value, "singleindex")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(powell.pos) > 0L)
  expect_true(length(fit.start.pos) == 1L)
  expect_true(length(fit.finish.pos) >= 1L)
  expect_lt(max(bandwidth.pos), fit.start.pos[[1L]])
  expect_lt(max(powell.pos), fit.start.pos[[1L]])
  expect_lt(fit.start.pos[[1L]], fit.finish.pos[[1L]])
})

test_that("predict.singleindex lp re-entry emits evaluation and training fit progress without handoff", {
  fixture <- make_npindex_fit_progress_fixture()

  fit <- npindex(
    bws = fixture$bw,
    txdat = fixture$tx,
    tydat = fixture$y
  )

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    predict(fit, newdata = fixture$dat[c(2L, 7L), c("x1", "x2")]),
    force_renderer = "single_line",
    now = npindex_fit_progress_time_counter()
  )

  lines <- npindex_fit_progress_lines(actual)

  expect_false(any(grepl(": starting$", lines)))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 1/2 \\(50\\.0%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 2/2 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 1/24 \\([0-9]+\\.[0-9]%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 24/24 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )))
})

test_that("npindex direct fixed lc fit emits known-total fit progress without handoff", {
  fixture <- make_npindex_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npindex(
      bws = fixture$bw_lc_fixed,
      txdat = fixture$tx,
      tydat = fixture$y
    ),
    force_renderer = "single_line",
    now = npindex_fit_progress_time_counter()
  )

  lines <- npindex_fit_progress_lines(actual)

  expect_s3_class(actual$value, "singleindex")
  expect_false(any(grepl(": starting$", lines)))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 1/24 \\([0-9]+\\.[0-9]%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 24/24 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )))
})

test_that("npindex direct fixed lc fit stays silent below start grace without handoff", {
  fixture <- make_npindex_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0.75,
    np.progress.interval.known.sec = 0.5
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npindex(
      bws = fixture$bw_lc_fixed,
      txdat = fixture$tx,
      tydat = fixture$y
    ),
    force_renderer = "single_line",
    now = npindex_fit_progress_time_values(c(0, 0.2, 0.4, 0.6))
  )

  expect_length(actual$trace, 0L)
})

test_that("npindex fixed lc bw to fit route hands off into the regression fit surface", {
  fixture <- make_npindex_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npindex(
      y ~ x1 + x2,
      data = fixture$dat,
      method = "ichimura",
      regtype = "lc",
      bwtype = "fixed",
      nmulti = 1L
    ),
    force_renderer = "single_line",
    now = npindex_fit_progress_time_counter()
  )

  lines <- npindex_fit_progress_lines(actual)
  bandwidth.pos <- grep("^\\[np\\] Bandwidth selection \\(", lines)
  fit.start.pos <- grep(
    "^\\[np\\] Fitting regression 0/24 \\(0\\.0%, elapsed 0\\.0s, eta 0\\.0s\\): starting$",
    lines
  )
  fit.finish.pos <- grep(
    "^\\[np\\] Fitting regression 24/24 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )

  expect_s3_class(actual$value, "singleindex")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(fit.start.pos) == 1L)
  expect_true(length(fit.finish.pos) >= 1L)
  expect_lt(max(bandwidth.pos), fit.start.pos[[1L]])
  expect_lt(fit.start.pos[[1L]], fit.finish.pos[[1L]])
})

test_that("predict.singleindex fixed lc re-entry emits evaluation and training fit progress without handoff", {
  fixture <- make_npindex_fit_progress_fixture()

  fit <- npindex(
    bws = fixture$bw_lc_fixed,
    txdat = fixture$tx,
    tydat = fixture$y
  )

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    predict(fit, newdata = fixture$dat[c(2L, 7L), c("x1", "x2")]),
    force_renderer = "single_line",
    now = npindex_fit_progress_time_counter()
  )

  lines <- npindex_fit_progress_lines(actual)

  expect_false(any(grepl(": starting$", lines)))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 1/2 \\(50\\.0%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 2/2 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 1/24 \\([0-9]+\\.[0-9]%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 24/24 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )))
})

make_npindex_lc_fixed_progress_fixture <- function() {
  set.seed(20260405)
  n <- 24L
  dat <- data.frame(
    x1 = runif(n, -1, 1),
    x2 = runif(n, -1, 1)
  )
  index <- dat$x1 + 0.6 * dat$x2
  dat$y <- sin(index) + 0.2 * index^2 + rnorm(n, sd = 0.05)

  list(
    dat = dat,
    tx = dat[c("x1", "x2")],
    y = dat$y,
    newdata = dat[c(2L, 7L), c("x1", "x2")],
    bw_fixed = npindexbw(
      xdat = dat[c("x1", "x2")],
      ydat = dat$y,
      bws = c(1, 0.6, 0.35),
      method = "ichimura",
      regtype = "lc",
      bwtype = "fixed",
      bandwidth.compute = FALSE
    ),
    bw_largeh = npindexbw(
      xdat = dat[c("x1", "x2")],
      ydat = dat$y,
      bws = c(1, 0.6, 5),
      method = "ichimura",
      regtype = "lc",
      bwtype = "fixed",
      bandwidth.compute = FALSE
    )
  )
}

npindex_lc_fixed_scalar_index <- function(txdat, beta) {
  data.frame(index = as.vector(as.matrix(txdat) %*% beta))
}

npindex_lc_fixed_oracle <- function(index.tx, y, bws, index.ex = index.tx) {
  spec <- getFromNamespace(".npindex_resolve_spec", "np")(bws, where = "npindex")
  regbw <- getFromNamespace(".np_index_regression_bandwidth", "np")(
    index.df = index.tx,
    ydat = y,
    bws = bws,
    spec = spec
  )
  npreg(
    exdat = index.ex,
    bws = regbw,
    txdat = index.tx,
    tydat = y
  )
}

test_that("npindex fixed lc direct and predict routes preserve the scalar-index oracle", {
  fixture <- make_npindex_lc_fixed_progress_fixture()

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  fit <- npindex(
    bws = fixture$bw_fixed,
    txdat = fixture$tx,
    tydat = fixture$y
  )

  index.tx <- npindex_lc_fixed_scalar_index(fixture$tx, fit$beta)
  oracle <- npindex_lc_fixed_oracle(index.tx, fixture$y, fit$bws)

  pred <- predict(fit, newdata = fixture$newdata)
  index.ex <- npindex_lc_fixed_scalar_index(fixture$newdata, fit$beta)
  pred.oracle <- npindex_lc_fixed_oracle(index.tx, fixture$y, fit$bws, index.ex = index.ex)

  expect_equal(as.numeric(fit$mean), as.numeric(oracle$mean), tolerance = 1e-12)
  expect_equal(as.numeric(pred), as.numeric(pred.oracle$mean), tolerance = 1e-12)
})

test_that("npindex fixed lc large-h route preserves the scalar-index oracle", {
  fixture <- make_npindex_lc_fixed_progress_fixture()

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  fit <- npindex(
    bws = fixture$bw_largeh,
    txdat = fixture$tx,
    tydat = fixture$y
  )

  index.tx <- npindex_lc_fixed_scalar_index(fixture$tx, fit$beta)
  oracle <- npindex_lc_fixed_oracle(index.tx, fixture$y, fit$bws)

  expect_equal(as.numeric(fit$mean), as.numeric(oracle$mean), tolerance = 1e-12)
})

test_that("npindex fixed lc bw-to-fit route preserves the scalar-index oracle", {
  fixture <- make_npindex_lc_fixed_progress_fixture()

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260405)
  fit <- npindex(
    y ~ x1 + x2,
    data = fixture$dat,
    method = "ichimura",
    regtype = "lc",
    bwtype = "fixed",
    nmulti = 1L
  )

  index.tx <- npindex_lc_fixed_scalar_index(fixture$tx, fit$beta)
  oracle <- npindex_lc_fixed_oracle(index.tx, fixture$y, fit$bws)

  expect_equal(as.numeric(fit$mean), as.numeric(oracle$mean), tolerance = 1e-12)
})

test_that("npindex formula newdata path matches explicit data path", {
  set.seed(20260223)
  dat <- data.frame(
    y = rnorm(36),
    x1 = runif(36),
    x2 = runif(36)
  )
  ex <- data.frame(
    x1 = seq(0.1, 0.9, length.out = 10),
    x2 = seq(0.9, 0.1, length.out = 10)
  )

  bw <- np::npindexbw(
    y ~ x1 + x2,
    data = dat,
    bws = c(1, 0.35, 0.45),
    bandwidth.compute = FALSE,
    method = "ichimura"
  )

  fit_formula <- np::npindex(bws = bw, data = dat, newdata = ex)
  fit_default <- np::npindex(
    bws = bw,
    txdat = dat[c("x1", "x2")],
    tydat = dat$y,
    exdat = ex[c("x1", "x2")]
  )

  expect_equal(fit_formula$mean, fit_default$mean)
  expect_equal(fit_formula$merr, fit_default$merr)
})

test_that("npscoef formula newdata path matches explicit data path", {
  set.seed(20260223)
  dat <- data.frame(
    y = rnorm(34),
    x1 = runif(34),
    z1 = runif(34)
  )
  ex <- data.frame(
    x1 = seq(0.05, 0.95, length.out = 11),
    z1 = seq(0.9, 0.2, length.out = 11)
  )

  bw <- np::npscoefbw(
    y ~ x1 | z1,
    data = dat,
    bws = 0.4,
    bandwidth.compute = FALSE
  )

  fit_formula <- np::npscoef(bws = bw, data = dat, newdata = ex)
  fit_default <- np::npscoef(
    bws = bw,
    txdat = dat["x1"],
    tydat = dat$y,
    tzdat = dat["z1"],
    exdat = ex["x1"],
    ezdat = ex["z1"]
  )

  expect_equal(fit_formula$mean, fit_default$mean)
  expect_equal(fit_formula$merr, fit_default$merr)
})

test_that("npplreg formula newdata path matches explicit data path", {
  set.seed(20260223)
  dat <- data.frame(
    y = rnorm(30),
    x1 = runif(30),
    z1 = runif(30)
  )
  ex <- data.frame(
    x1 = seq(0.1, 0.8, length.out = 9),
    z1 = seq(0.8, 0.1, length.out = 9)
  )

  bw <- np::npplregbw(
    y ~ x1 | z1,
    data = dat,
    bws = matrix(c(0.45, 0.35), nrow = 2, ncol = 1),
    bandwidth.compute = FALSE
  )

  fit_formula <- np::npplreg(bws = bw, data = dat, newdata = ex)
  fit_default <- np::npplreg(
    bws = bw,
    txdat = dat["x1"],
    tydat = dat$y,
    tzdat = dat["z1"],
    exdat = ex["x1"],
    ezdat = ex["z1"]
  )

  expect_equal(fit_formula$mean, fit_default$mean)
  expect_equal(fit_formula$merr, fit_default$merr)
})

with_npindex_degree_bindings <- function(bindings, code) {
  code <- substitute(code)
  ns <- asNamespace("np")
  old <- lapply(names(bindings), function(name) get(name, envir = ns, inherits = FALSE))
  names(old) <- names(bindings)

  for (name in names(bindings)) {
    was_locked <- bindingIsLocked(name, ns)
    if (was_locked)
      unlockBinding(name, ns)
    assign(name, bindings[[name]], envir = ns)
    if (was_locked)
      lockBinding(name, ns)
  }

  on.exit({
    for (name in names(old)) {
      was_locked <- bindingIsLocked(name, ns)
      if (was_locked)
        unlockBinding(name, ns)
      assign(name, old[[name]], envir = ns)
      if (was_locked)
        lockBinding(name, ns)
    }
  }, add = TRUE)

  eval(code, envir = parent.frame())
}

capture_npindex_degree_messages_only <- function(expr) {
  messages <- character()
  withCallingHandlers(
    expr,
    message = function(m) {
      messages <<- c(messages, conditionMessage(m))
      invokeRestart("muffleMessage")
    }
  )
  messages
}

npindex_degree_progress_time_values <- function(values) {
  force(values)
  i <- 0L
  function() {
    i <<- min(i + 1L, length(values))
    values[[i]]
  }
}

test_that("npindexbw exhaustive degree search matches manual Ichimura profile minimum", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 30
  xdat <- data.frame(
    x1 = runif(n, -1, 1),
    x2 = runif(n, -1, 1)
  )
  index <- xdat$x1 + 0.75 * xdat$x2
  y <- sin(index) + 0.25 * index^2 + rnorm(n, sd = 0.05)

  start.bws <- c(1, 0.75, 0.35)
  bw0 <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = start.bws,
    method = "ichimura",
    regtype = "lp",
    degree = 0L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    nmulti = 1L
  )
  bw1 <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = start.bws,
    method = "ichimura",
    regtype = "lp",
    degree = 1L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    nmulti = 1L
  )
  auto <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = start.bws,
    method = "ichimura",
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    nmulti = 1L
  )

  expect_s3_class(auto, "sibandwidth")
  expect_true(isTRUE(auto$bernstein.basis))
  expect_identical(auto$degree.search$mode, "exhaustive")
  expect_true(isTRUE(auto$degree.search$completed))
  expect_true(isTRUE(auto$degree.search$certified))
  expect_lte(auto$fval, min(bw0$fval, bw1$fval) + 1e-10)
  expect_lte(auto$degree.search$best.fval, auto$degree.search$baseline.fval + 1e-10)
  expect_true(all(c("degree", "fval", "status", "cached") %in% names(auto$degree.search$trace)))

  manual <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = start.bws,
    method = "ichimura",
    regtype = "lp",
    degree = 1L,
    bwtype = "fixed",
    nmulti = 1L
  )
  expect_null(manual$degree.search)
})

test_that("npindexbw coordinate search can be exhaustively certified on a small grid", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 28
  xdat <- data.frame(
    x1 = runif(n, -1, 1),
    x2 = runif(n, -1, 1)
  )
  index <- xdat$x1 + 0.5 * xdat$x2
  y <- cos(index) + 0.2 * index^2 + rnorm(n, sd = 0.05)
  start.bws <- c(1, 0.5, 0.3)

  exhaustive <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = start.bws,
    method = "ichimura",
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    nmulti = 1L
  )
  coordinate <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = start.bws,
    method = "ichimura",
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = TRUE,
    degree.restarts = 1L,
    degree.max.cycles = 4L,
    bwtype = "fixed",
    nmulti = 1L
  )

  expect_identical(coordinate$degree.search$mode, "coordinate")
  expect_true(isTRUE(coordinate$degree.search$completed))
  expect_true(isTRUE(coordinate$degree.search$certified))
  expect_equal(as.integer(coordinate$degree), as.integer(exhaustive$degree))
  expect_equal(coordinate$fval, exhaustive$fval, tolerance = 1e-10)
})

test_that("npindexbw automatic degree search honors Klein-Spady objective direction", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 32
  xdat <- data.frame(
    x1 = runif(n, -1, 1),
    x2 = runif(n, -1, 1)
  )
  index <- xdat$x1 - 0.6 * xdat$x2
  p <- plogis(1.25 * index - 0.4 * index^2)
  y <- rbinom(n, size = 1L, prob = p)
  start.bws <- c(1, -0.6, 0.35)

  bw0 <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = start.bws,
    method = "kleinspady",
    regtype = "lp",
    degree = 0L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    nmulti = 1L
  )
  bw1 <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = start.bws,
    method = "kleinspady",
    regtype = "lp",
    degree = 1L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    nmulti = 1L
  )
  auto <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = start.bws,
    method = "kleinspady",
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    nmulti = 1L
  )

  expect_identical(auto$method, "kleinspady")
  expect_lte(auto$fval, min(bw0$fval, bw1$fval) + 1e-10)
})

test_that("npindexbw automatic degree search enforces pilot guardrails", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 24
  xdat <- data.frame(
    x1 = runif(n, -1, 1),
    x2 = runif(n, -1, 1)
  )
  y <- xdat$x1 + xdat$x2 + rnorm(n, sd = 0.05)

  expect_error(
    npindexbw(
      xdat = xdat,
      ydat = y,
      bws = c(1, 1, 0.3),
      method = "ichimura",
      regtype = "lc",
      degree.select = "exhaustive",
      search.engine = "cell",
      degree.min = 0L,
      degree.max = 1L,
      bwtype = "fixed",
      nmulti = 1L
    ),
    "automatic degree search currently requires regtype='lp'"
  )

  expect_error(
    npindexbw(
      xdat = xdat,
      ydat = y,
      bws = c(1, 1, 0.3),
      method = "ichimura",
      regtype = "lp",
      bandwidth.compute = FALSE,
      degree.select = "exhaustive",
      search.engine = "cell",
      degree.min = 0L,
      degree.max = 1L,
      bwtype = "fixed",
      nmulti = 1L
    ),
    "bandwidth.compute=TRUE"
  )

  bw <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = c(1, 1, 0.3),
    method = "ichimura",
    regtype = "lp",
    bernstein.basis = FALSE,
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 4L,
    bwtype = "fixed",
    nmulti = 1L
  )

  expect_s3_class(bw, "sibandwidth")
  expect_false(isTRUE(bw$bernstein.basis))
  expect_lte(max(as.integer(bw$degree)), 4L)
})

test_that("npindex forwards automatic LP degree search through npindexbw", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 24
  dat <- data.frame(
    x1 = runif(n, -1, 1),
    x2 = runif(n, -1, 1)
  )
  index <- dat$x1 + 0.5 * dat$x2
  dat$y <- sin(index) + rnorm(n, sd = 0.05)

  fit <- npindex(
    y ~ x1 + x2,
    data = dat,
    method = "ichimura",
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    nmulti = 1L
  )

  expect_s3_class(fit, "singleindex")
  expect_s3_class(fit$bws, "sibandwidth")
  expect_false(is.null(fit$bws$degree.search))
  expect_identical(fit$bws$degree.search$mode, "exhaustive")
})

test_that("npindexbw automatic degree search emits staged progress output", {
  old_opts <- options(np.messages = TRUE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 20
  xdat <- data.frame(
    x1 = runif(n, -1, 1),
    x2 = runif(n, -1, 1)
  )
  y <- xdat$x1 + 0.5 * xdat$x2 + rnorm(n, sd = 0.05)

  msgs <- with_npindex_degree_bindings(
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_renderer_for_surface = function(surface, capability) "legacy",
      .np_progress_now = npindex_degree_progress_time_values(seq(0, 20, by = 0.5))
    ),
    capture_npindex_degree_messages_only(
      get("npindexbw", envir = asNamespace("np"), inherits = FALSE)(
        xdat = xdat,
        ydat = y,
        bws = c(1, 0.5, 0.3),
        method = "ichimura",
        regtype = "lp",
        degree.select = "exhaustive",
        search.engine = "cell",
        degree.min = 0L,
        degree.max = 1L,
        bwtype = "fixed",
        nmulti = 1L
      )
    )
  )

  expect_true(any(grepl("Automatic polynomial degree search baseline \\(0\\)", msgs)))
  expect_true(any(grepl("Selecting degree and bandwidth", msgs, fixed = TRUE)))
  expect_true(any(grepl("exhaustive", msgs)))
  expect_true(any(grepl("best (", msgs, fixed = TRUE)))
})

test_that("npindexbw automatic degree search defaults to NOMAD plus Powell", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 26
  xdat <- data.frame(
    x1 = runif(n, -1, 1),
    x2 = runif(n, -1, 1)
  )
  index <- xdat$x1 + 0.5 * xdat$x2
  y <- sin(index) + 0.15 * index^2 + rnorm(n, sd = 0.05)

  auto <- npindexbw(
    xdat = xdat,
    ydat = y,
    bws = c(1, 0.5, 0.3),
    method = "ichimura",
    regtype = "lp",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 2L,
    bwtype = "fixed",
    nmulti = 1L
  )

  expect_s3_class(auto, "sibandwidth")
  expect_false(is.null(auto$degree.search))
  expect_identical(auto$degree.search$mode, "nomad+powell")
  expect_true(is.finite(auto$fval))
  expect_lte(auto$fval, auto$degree.search$baseline.fval + 1e-8)
  expect_equal(auto$fval, auto$degree.search$best.fval, tolerance = 1e-8)
})

test_that("npindex ichimura lc gradients preserve the scalar-index fit contract", {
  old_opts <- options(np.messages = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(42)
  n <- 1000L
  x1 <- runif(n, min = -1, max = 1)
  x2 <- runif(n, min = -1, max = 1)
  y <- x1 - x2 + rnorm(n)
  dat <- data.frame(x1 = x1, x2 = x2, y = y)

  bw <- npindexbw(formula = y ~ x1 + x2, data = dat)
  fit <- npindex(bws = bw, gradients = TRUE)

  index.df <- data.frame(index = as.vector(as.matrix(dat[c("x1", "x2")]) %*% bw$beta))
  oracle <- do.call(npreg, list(
    txdat = index.df,
    tydat = dat$y,
    exdat = index.df,
    bws = bw$bw,
    bwtype = bw$type,
    ckertype = bw$ckertype,
    ckerorder = bw$ckerorder,
    regtype = bw$regtype,
    gradients = TRUE,
    warn.glp.gradient = FALSE
  ))

  expect_true(all(is.finite(fit$grad)))
  expect_true(all(is.finite(fit$betavcov)))
  expect_equal(as.numeric(fit$mean), as.numeric(oracle$mean), tolerance = 0)
  expect_equal(
    unname(as.matrix(fit$grad)),
    unname(as.matrix(oracle$grad) %*% t(as.vector(bw$beta))),
    tolerance = 0
  )
})

make_semiparam_formula_fits <- function(seed = 123, n = 24L) {
  set.seed(seed)

  dat_pl <- data.frame(x = rnorm(n), z = sort(runif(n)))
  dat_pl$y <- 1 + dat_pl$x + sin(2 * pi * dat_pl$z) + rnorm(n, sd = 0.08)
  bw_pl <- np::npplregbw(
    y ~ x | z,
    data = dat_pl,
    bws = matrix(c(0.7, 0.5), nrow = 2),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 1,
    bwtype = "fixed"
  )
  fit_pl <- np::npplreg(bws = bw_pl)

  dat_sc <- data.frame(x = runif(n), z = sort(runif(n)))
  dat_sc$y <- (1 + dat_sc$z^2) * dat_sc$x + rnorm(n, sd = 0.08)
  bw_sc <- np::npscoefbw(
    y ~ x | z,
    data = dat_sc,
    bws = 0.5,
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 1,
    bwtype = "fixed"
  )
  fit_sc <- np::npscoef(bws = bw_sc, errors = FALSE, betas = FALSE)

  dat_si <- data.frame(x1 = runif(n, -1, 1), x2 = runif(n, -1, 1))
  idx <- dat_si$x1 + 0.5 * dat_si$x2
  dat_si$y <- sin(idx) + 0.25 * idx^2 + rnorm(n, sd = 0.05)
  bw_si <- np::npindexbw(
    y ~ x1 + x2,
    data = dat_si,
    bws = c(1, 0.7, 0.7),
    bandwidth.compute = FALSE,
    method = "ichimura",
    regtype = "lp",
    degree = 1,
    bwtype = "fixed"
  )
  fit_si <- np::npindex(bws = bw_si)

  list(pl = fit_pl, sc = fit_sc, si = fit_si)
}

test_that("semiparametric formula fits preserve response-name metadata for plotting", {
  fits <- make_semiparam_formula_fits(seed = 123)

  expect_identical(fits$pl$bws$ynames, "y")
  expect_identical(fits$sc$bws$ynames, "y")
  expect_identical(fits$si$bws$ynames, "y")
})

test_that("semiparametric formula-fit plots do not regress on response labels", {
  fits <- make_semiparam_formula_fits(seed = 456)
  out.pl <- expect_no_error(plot(fits$pl, output = "data", perspective = FALSE))
  out.sc <- expect_no_error(plot(fits$sc, output = "data", perspective = FALSE))
  out.si <- expect_no_error(plot(fits$si, output = "data", perspective = FALSE))

  expect_type(out.pl, "list")
  expect_type(out.sc, "list")
  expect_type(out.si, "list")
  expect_true(length(out.pl) > 0L)
  expect_true(length(out.sc) > 0L)
  expect_true(length(out.si) > 0L)
})

.np_plot_conbandwidth_engine <-
  function(bws,
           xdat,
           ydat,
           data = NULL,
           xq = 0.5,
           yq = 0.5,
           xtrim = 0.0,
           ytrim = 0.0,
           neval = 50,
           gradients = FALSE,
           gradient.order = 1L,
           common.scale = TRUE,
           perspective = TRUE,
           renderer = c("base", "rgl"),
           main = NULL,
           type = NULL,
           border = NULL,
           cex.axis = NULL,
           cex.lab = NULL,
           cex.main = NULL,
           cex.sub = NULL,
           col = NULL,
           ylab = NULL,
           xlab = NULL,
           zlab = NULL,
           sub = NULL,
           ylim = NULL,
           xlim = NULL,
           zlim = NULL,
           lty = NULL,
           lwd = NULL,
           theta = 0.0,
           phi = 20.0,
           tau = 0.5,
           view = c("rotate","fixed"),
           plot.behavior = c("plot","plot-data","data"),
           plot.errors.method = c("none","bootstrap","asymptotic"),
           plot.errors.boot.method = c("inid", "fixed", "geom"),
           plot.errors.boot.nonfixed = c("exact", "frozen"),
           plot.errors.boot.blocklen = NULL,
           plot.errors.boot.num = 1999,
           plot.errors.center = c("estimate","bias-corrected"),
           plot.errors.type = c("pmzsd","pointwise","bonferroni","simultaneous","all"),
           plot.errors.alpha = 0.05,
           plot.errors.style = c("band","bar"),
           plot.errors.bar = c("|","I"),
           plot.errors.bar.num = min(neval,25),
           plot.bxp = FALSE,
           plot.bxp.out = TRUE,
           plot.par.mfrow = TRUE,
           plot.rug = FALSE,
           proper = FALSE,
           proper.method = c("project"),
           proper.control = list(),
           ...,
           random.seed){

    engine.ctx <- .np_plot_engine_begin(plot.par.mfrow = plot.par.mfrow)
    on.exit(.np_plot_restore_par(engine.ctx$oldpar), add = TRUE)
    plot.par.mfrow <- engine.ctx$plot.par.mfrow

    scalar_default <- function(value, default) {
      if (is.null(value)) default else value
    }

    dots <- list(...)
    plot.legend <- if (!is.null(dots$legend)) dots$legend else TRUE
    plot.user.args <- .np_plot_user_args(dots, "plot")
    bxp.user.args <- .np_plot_user_args(dots, "bxp")
    rgl.persp3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.persp3d", "rgl.persp3d.")
    rgl.view3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.view3d", "rgl.view3d.")
    rgl.par3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.par3d", "rgl.par3d.")
    rgl.grid3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.grid3d", "rgl.grid3d.")
    rgl.widget.user.args <- .np_plot_collect_rgl_args(dots, "rgl.widget", "rgl.widget.")
    rgl.legend3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.legend3d", "rgl.legend3d.")
    rgl.legend3d.user.args <- .np_plot_merge_rgl_legend_control(rgl.legend3d.user.args, plot.legend)
    rgl.surface3d.user.args <- .np_plot_extract_prefixed_args(dots, "rgl.surface3d.")
    bxp.args <- bxp.user.args
    if (!is.null(col)) bxp.args$col <- col
    if (!is.null(lty)) bxp.args$lty <- lty
    if (!is.null(lwd)) bxp.args$lwd <- lwd
    if (!is.null(border)) bxp.args$border <- border

    cdf <- FALSE
    quantreg <- FALSE
    miss.xy = c(missing(xdat),missing(ydat))
    
    if (any(miss.xy) && !all(miss.xy))
      stop("one of, but not both, xdat and ydat was specified")
    else if(all(miss.xy) && !is.null(bws$formula)){
      tt <- terms(bws)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)
    tmf <- bws$call[c(1,m)]
    tmf[[1]] <- as.name("model.frame")
    tmf[["formula"]] <- tt
    mf.args <- as.list(tmf)[-1L]
    umf <- tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

      ydat <- tmf[, bws$variableNames[["response"]], drop = FALSE]
      xdat <- tmf[, bws$variableNames[["terms"]], drop = FALSE]
    } else {
      if(all(miss.xy) && !is.null(bws$call)){
        xdat <- data.frame(.np_eval_bws_call_arg(bws, "xdat"))
        ydat <- data.frame(.np_eval_bws_call_arg(bws, "ydat"))
      }

      ## catch and destroy NA's
      xdat = toFrame(xdat)
      ydat = toFrame(ydat)
      
      keep.rows <- rep_len(TRUE, nrow(xdat))
      rows.omit <- attr(na.omit(data.frame(xdat, ydat)), "na.action")
      if (length(rows.omit) > 0L)
        keep.rows[as.integer(rows.omit)] <- FALSE

      if (!any(keep.rows))
        stop("Data has no rows without NAs")

      xdat <- xdat[keep.rows,,drop = FALSE]
      ydat <- ydat[keep.rows,,drop = FALSE]

    }

    if (quantreg && dim(ydat)[2] != 1)
      stop("'ydat' must have one column for quantile regression")
    
    xq = double(bws$xndim)+xq
    yq = double(bws$yndim)+yq
    
    xtrim = double(bws$xndim)+xtrim
    ytrim = double(bws$yndim)+ytrim

    if (missing(plot.errors.method) &
        any(!missing(plot.errors.boot.num), !missing(plot.errors.boot.method),
            !missing(plot.errors.boot.nonfixed),
            !missing(plot.errors.boot.blocklen))){
      stop(
        "errors must be set to \"bootstrap\" when bootstrap controls are supplied",
        call. = FALSE
      )
    }
    
    normalized.opts <- .np_plot_normalize_common_options(
      plot.behavior = plot.behavior,
      plot.errors.method = plot.errors.method,
      plot.errors.boot.method = plot.errors.boot.method,
      plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
      plot.errors.boot.blocklen = plot.errors.boot.blocklen,
      plot.errors.center = plot.errors.center,
      plot.errors.type = plot.errors.type,
      plot.errors.alpha = plot.errors.alpha,
      plot.errors.style = plot.errors.style,
      plot.errors.bar = plot.errors.bar,
      xdat = xdat,
      common.scale = common.scale,
      ylim = ylim
    )

    plot.behavior <- normalized.opts$plot.behavior
    plot.errors.method <- normalized.opts$plot.errors.method
    plot.errors.boot.method <- normalized.opts$plot.errors.boot.method
    plot.errors.boot.nonfixed <- normalized.opts$plot.errors.boot.nonfixed
    plot.errors.boot.blocklen <- normalized.opts$plot.errors.boot.blocklen
    plot.errors.center <- normalized.opts$plot.errors.center
    plot.errors.type <- normalized.opts$plot.errors.type
    plot.errors.alpha <- normalized.opts$plot.errors.alpha
    plot.errors.style <- normalized.opts$plot.errors.style
    plot.errors.bar <- normalized.opts$plot.errors.bar
    common.scale <- normalized.opts$common.scale

    plot.errors = (plot.errors.method != "none")
    proper.args <- .np_condens_validate_proper_args(
      proper = proper,
      proper.method = proper.method,
      proper.control = proper.control
    )
    surface.supported <- isTRUE((bws$xncon + bws$xnord + bws$yncon + bws$ynord - quantreg == 2) &&
                                (bws$xnuno + bws$ynuno == 0) &&
                                !any(xor(bws$xdati$iord, bws$xdati$inumord)))
    renderer <- .np_plot_validate_renderer_request(
      renderer = renderer,
      route = "plot.conbandwidth()",
      perspective = perspective,
      supported.route = surface.supported,
      view = as.character(view)[1L],
      gradients = gradients,
      plot.errors.method = plot.errors.method,
      plot.behavior = plot.behavior,
      allow.plot.errors = TRUE
    )
    plot.rug <- .np_plot_validate_rug_request(
      plot.rug = plot.rug,
      route = "plot.conbandwidth()",
      supported.route = if (identical(renderer, "rgl")) {
        isTRUE(surface.supported && perspective)
      } else {
        TRUE
      },
      renderer = renderer,
      reason = if (identical(renderer, "rgl")) {
        "supported rgl surface routes"
      } else {
        "supported base plot routes"
      }
    )

    if (surface.supported && perspective && !gradients){
      view = match.arg(view)
      rotate = (view == "rotate")
      
      if (is.ordered(xdat[,1])){
        x1.eval = bws$xdati$all.ulev[[1]]
        x1.neval = length(x1.eval)
      } else {
        x1.neval = neval
        qi = trim.quantiles(xdat[,1], xtrim[1])
        x1.eval = seq(qi[1], qi[2], length.out = x1.neval)
      }

      ## if we are doing quantile regression then we are dealing
      ## with 2 x variables ...

      if (quantreg){
        tx2 <- xdat[,2]
        txi <- 2
        txdati <- bws$xdati
        txtrim <- xtrim
      }
      else{
        tx2 <- ydat[,1]
        txi <- 1
        txdati <- bws$ydati
        txtrim <- ytrim
      }
      
      if (txdati$iord[txi]){
        x2.eval = txdati$all.ulev[[txi]]
        x2.neval = length(x2.eval)
      } else {
        x2.neval = neval
        qi = trim.quantiles(tx2, txtrim[txi])
        x2.eval = seq(qi[1], qi[2], length.out = x2.neval)
      }

      x.eval <- expand.grid(x1.eval, x2.eval)

      if (bws$xdati$iord[1])
        x1.eval <- (bws$xdati$all.dlev[[1]])[as.integer(x1.eval)]
      
      if (txdati$iord[txi])
        x2.eval <- (txdati$all.dlev[[txi]])[as.integer(x2.eval)]


      tboo =
        if(quantreg) "quant"
        else if (cdf) "dist"
        else "dens"

      if (quantreg) {
        tobj <- .np_plot_quantile_eval(
          txdat = xdat,
          tydat = ydat,
          exdat = x.eval,
          tau = tau,
          bws = bws
        )
      } else {
        tobj <- .np_plot_conditional_eval(
          bws = bws,
          xdat = xdat,
          ydat = ydat,
          exdat = x.eval[,1, drop = FALSE],
          eydat = x.eval[,2, drop = FALSE],
          cdf = FALSE,
          gradients = FALSE,
          proper = isTRUE(proper.args$proper.requested),
          proper.method = proper.args$proper.method,
          proper.control = proper.args$proper.control
        )
      }
      tcomp <- switch(tboo,
                      "quant" = tobj$quantile,
                      "dist" = tobj$condist,
                      "dens" = tobj$condens)
      tcerr <- if (quantreg) tobj$quanterr else tobj$conderr
      tex <- if (quantreg) x.eval else x.eval[,1]
      tey <- if (quantreg) NA else x.eval[,2]

      tdens = matrix(data = tcomp,
        nrow = x1.neval, ncol = x2.neval, byrow = FALSE)

      terr = matrix(data = tcerr, nrow = length(tcomp), ncol = 3)
      terr[,3] = NA
      lerr.all <- NULL
      herr.all <- NULL
      
      if (plot.errors.method == "bootstrap"){
        terr.obj <- compute.bootstrap.errors(xdat = xdat, ydat = ydat,
          exdat = tex, eydat = tey,
          cdf = cdf,
          quantreg = quantreg,
          tau = tau,
          gradients = FALSE,
          gradient.index = 0,
          slice.index = 0,
          plot.errors.boot.method = plot.errors.boot.method,
          plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
          plot.errors.boot.blocklen = plot.errors.boot.blocklen,
          plot.errors.boot.num = plot.errors.boot.num,
          plot.errors.center = plot.errors.center,
          plot.errors.type = plot.errors.type,
          plot.errors.alpha = plot.errors.alpha,
          progress.target = NULL,
          proper = isTRUE(proper.args$proper.requested),
          proper.method = proper.args$proper.method,
          proper.control = proper.args$proper.control,
          bws = bws)
        terr <- terr.obj[["boot.err"]]
        terr.all <- terr.obj[["boot.all.err"]]

        pc = (plot.errors.center == "bias-corrected")
        center.val <- if(pc) terr[,3] else tcomp

        lerr = matrix(data = center.val - terr[,1],
          nrow = x1.neval, ncol = x2.neval, byrow = FALSE)

        herr = matrix(data = center.val + terr[,2],
          nrow = x1.neval, ncol = x2.neval, byrow = FALSE)

        if (plot.errors.type == "all" && !is.null(terr.all)) {
          lerr.all <- lapply(terr.all, function(te)
            matrix(data = center.val - te[,1], nrow = x1.neval, ncol = x2.neval, byrow = FALSE))
          herr.all <- lapply(terr.all, function(te)
            matrix(data = center.val + te[,2], nrow = x1.neval, ncol = x2.neval, byrow = FALSE))
        }

      } else if (plot.errors.method == "asymptotic") {
        terr.obj <- .np_plot_asymptotic_error_from_se(
          se = tcerr,
          alpha = plot.errors.alpha,
          band.type = plot.errors.type,
          m = nrow(tex)
        )
        terr[,1:2] <- terr.obj$err
        terr.all <- terr.obj$all.err
        center.val <- tcomp
        lerr = matrix(data = center.val - terr[,1],
          nrow = x1.neval, ncol = x2.neval, byrow = FALSE)

        herr = matrix(data = center.val + terr[,2],
          nrow = x1.neval, ncol = x2.neval, byrow = FALSE)

        if (plot.errors.type == "all" && !is.null(terr.all)) {
          lerr.all <- lapply(terr.all, function(te)
            matrix(data = center.val - te[,1], nrow = x1.neval, ncol = x2.neval, byrow = FALSE))
          herr.all <- lapply(terr.all, function(te)
            matrix(data = center.val + te[,2], nrow = x1.neval, ncol = x2.neval, byrow = FALSE))
        }
      }

      if(is.null(zlim)) {
          zlim =
              if (plot.errors){
                  if (plot.errors.type == "all" && !is.null(lerr.all))
                    c(min(c(unlist(lerr.all), lerr)), max(c(unlist(herr.all), herr)))
                  else
                    c(min(lerr),max(herr))
              } else
                c(min(tcomp),max(tcomp))
      }

      if (plot.behavior != "plot"){
        ret.fun <- switch(tboo,
                          "quant" = qregression,
                          "dist" = condistribution,
                          "dens" = condensity)
        ret.args <- list(bws = bws, xeval = tex, ntrain = dim(xdat)[1])
        if (quantreg) {
          ret.args$tau <- tau
          ret.args$quantile <- tcomp
          ret.args$quanterr <- terr[,1:2]
        } else {
          ret.args$yeval <- tey
          if (cdf) ret.args$condist <- tcomp else ret.args$condens <- tcomp
          ret.args$conderr <- terr[,1:2]
          ret.args$proper.requested <- tobj$proper.requested
          ret.args$proper.applied <- tobj$proper.applied
          ret.args$proper.method <- tobj$proper.method
          ret.args$condens.raw <- tobj$condens.raw
          ret.args$proper.info <- tobj$proper.info
        }
        cd1 <- do.call(ret.fun, ret.args)
        cd1$bias = NA

        if (plot.errors.center == "bias-corrected")
          cd1$bias = terr[,3] - tcomp
        
        if (plot.behavior == "data")
          return ( list(cd1 = cd1) )
      }


      # rows = constant x2
      # cols = constant x1

      xlab.val <- scalar_default(xlab, gen.label(names(xdat)[1], "X"))
      ylab.val <- scalar_default(ylab, gen.label(names(ydat)[1], "Y"))
      zlab.val <- scalar_default(zlab, paste("Conditional", if (cdf) "Distribution" else "Density"))

      if (identical(renderer, "rgl")) {
        rgl.first.render <- .np_plot_first_render_state()
        on.exit(.np_plot_activity_end(rgl.first.render$activity), add = TRUE)
        rgl.view <- .np_plot_rgl_view_angles(theta = theta, phi = phi)
        main.val <- if (!is.null(main)) main else NULL
        .np_plot_first_render_begin(rgl.first.render)
        rgl.out <- .np_plot_render_surface_rgl(
          x = x1.eval,
          y = x2.eval,
          z = tdens,
          zlim = zlim,
          col = col,
          border = scalar_default(border, "black"),
          xlab = xlab.val,
          ylab = ylab.val,
          zlab = zlab.val,
          theta = rgl.view$theta,
          phi = rgl.view$phi,
          main = main.val,
          par3d.args = rgl.par3d.user.args,
          view3d.args = rgl.view3d.user.args,
          persp3d.args = rgl.persp3d.user.args,
          grid3d.args = rgl.grid3d.user.args,
          widget.args = rgl.widget.user.args,
          draw.extras = function() {
            if (plot.rug) {
              .np_plot_draw_floor_rug_rgl(
                x1 = xdat[,1],
                x2 = if (quantreg) xdat[,2] else ydat[,1],
                zlim = zlim
              )
            }
            if (plot.errors) {
              .np_plot_error_surfaces_rgl(
                x = x1.eval,
                y = x2.eval,
                plot.errors.type = plot.errors.type,
                lerr = lerr,
                herr = herr,
                lerr.all = lerr.all,
                herr.all = herr.all,
                surface3d.args = rgl.surface3d.user.args,
                legend3d.args = rgl.legend3d.user.args
              )
            }
          }
        )
        .np_plot_first_render_end(rgl.first.render)
        return(.np_plot_rgl_finalize(
          rgl.out = rgl.out,
          plot.behavior = plot.behavior,
          plot.data = list(cd1 = cd1)
        ))
      }

      rotate.defaults <- .np_plot_rotate_defaults()
      dtheta = rotate.defaults$dtheta
      persp.col = grDevices::adjustcolor(
        .np_plot_persp_surface_colors(z = tdens, col = col),
        alpha.f = 0.5
      )
      first.render.activity <- NULL
      first.render.pending <- TRUE
      on.exit(.np_plot_activity_end(first.render.activity), add = TRUE)
      frame.theta <- (0:((360 %/% dtheta - 1L) * rotate)) * dtheta + theta
      rotation.progress <- .np_plot_rotation_progress_begin(length(frame.theta))
      on.exit(.np_plot_rotation_progress_end(rotation.progress), add = TRUE)
      
      for (frame.idx in seq_along(frame.theta)){
          i <- frame.theta[[frame.idx]]
          if (isTRUE(first.render.pending))
            first.render.activity <- .np_plot_activity_begin("Rendering plot surface")
          persp.mat <- persp(x1.eval,
                             x2.eval,
                             tdens,
                             zlim = zlim,
                             col = persp.col,
                             border = scalar_default(border, "black"),
                             ticktype = "detailed",
                             cex.axis = scalar_default(cex.axis, par()$cex.axis),
                             cex.lab = scalar_default(cex.lab, par()$cex.lab),
                             cex.main = scalar_default(cex.main, par()$cex.main),
                             cex.sub = scalar_default(cex.sub, par()$cex.sub),
                             lwd = 0.8 * scalar_default(lwd, par()$lwd),
                             xlab = xlab.val,
                             ylab = ylab.val,
                             zlab = zlab.val,
                             theta = i,
                             phi = phi,
                             main = scalar_default(main, ""))

          if (isTRUE(first.render.pending)) {
            .np_plot_activity_end(first.render.activity)
            first.render.activity <- NULL
            first.render.pending <- FALSE
          }
          .np_plot_draw_box_grid_persp(
            xlim = range(x1.eval, finite = TRUE),
            ylim = range(x2.eval, finite = TRUE),
            zlim = zlim,
            persp.mat = persp.mat
          )
          if (plot.rug) {
            .np_plot_draw_floor_rug_persp(
              x1 = xdat[,1],
              x2 = if (quantreg) xdat[,2] else ydat[,1],
              zlim = zlim,
              persp.mat = persp.mat
            )
          }

          if (plot.errors){
            .np_plot_draw_error_wireframes_persp(
              x = x1.eval,
              y = x2.eval,
              persp.mat = persp.mat,
              plot.errors.type = plot.errors.type,
              lerr = lerr,
              herr = herr,
              lerr.all = lerr.all,
              herr.all = herr.all,
              border = scalar_default(border, "grey"),
              lwd = scalar_default(lwd, par()$lwd)
            )
            if (plot.errors.type == "all" && !is.null(lerr.all) && !is.null(herr.all)) {
              .np_plot_draw_all_band_legend(
                legend = plot.legend,
                x = "topright",
                lty = 1,
                lwd = 2.15 * scalar_default(lwd, par()$lwd)
              )
            }
          }

          rotation.progress <- .np_plot_rotation_progress_tick(rotation.progress, done = frame.idx)
          Sys.sleep(if (isTRUE(rotate)) rotate.defaults$sleep else 0.5)
      }

      if (plot.behavior == "plot-data")
        return ( list(cd1 = cd1) )
    } else {

      dsf = if (gradients) bws$xndim else 1
      tot.dim = bws$xndim + bws$yndim - quantreg

      plot.layout <- .np_plot_layout_begin(
        plot.behavior = plot.behavior,
        plot.par.mfrow = plot.par.mfrow,
        mfrow = n2mfrow(dsf * tot.dim)
      )

      x.ev = xdat[1,,drop = FALSE]
      y.ev = ydat[1,,drop = FALSE]

      for (i in seq_len(bws$xndim))
        x.ev[1,i] = uocquantile(xdat[,i], prob=xq[i])

      for (i in seq_len(bws$yndim))
        y.ev[1,i] = uocquantile(ydat[,i], prob=yq[i])


      maxneval = max(c(sapply(xdat,nlevels), sapply(ydat,nlevels), neval))

      exdat = xdat[rep(1, maxneval), , drop = FALSE]
      eydat = as.data.frame(matrix(data = 0, nrow = maxneval, ncol = bws$yndim))

      for (i in seq_len(bws$xndim))
        exdat[,i] = x.ev[1,i]

      for (i in seq_len(bws$yndim))
        eydat[,i] = y.ev[1,i]

      if (common.scale){
        data.eval = matrix(data = NA, nrow = maxneval,
          ncol = tot.dim*dsf)
        
        data.err = matrix(data = NA, nrow = maxneval,
          ncol = 3*tot.dim*dsf)
        data.err.all = vector("list", tot.dim*dsf)

        allei = as.data.frame(matrix(data = NA, nrow = maxneval,
          ncol = tot.dim))

        all.bxp = list()
      }

      all.isFactor = c(vapply(xdat, is.factor, logical(1)), vapply(ydat, is.factor, logical(1)))

      plot.out = list()

      temp.err = matrix(data = NA, nrow = maxneval, ncol = 3)
      temp.dens = replicate(maxneval, NA)
      ## plotting controls
      plot.bootstrap <- plot.errors.method == "bootstrap"
      tylabE <- if (quantreg) paste(tau, "quantile") else paste("Conditional", if (cdf) "Distribution" else "Density")
      plotOnEstimate <- (plot.errors.center == "estimate")

      plot.index = 0
      xOrY = "x"

      for (i in seq_len(bws$xndim)){
        plot.index = plot.index + 1
        temp.err[,] = NA
        temp.dens[] =  NA
        temp.boot = list()

        xi.factor = all.isFactor[plot.index]
        
        if (xi.factor){
          ei = levels(xdat[,i])
          ei = factor(ei, levels = ei)
          xi.neval = length(ei)
        } else {
          xi.neval = neval
          qi = trim.quantiles(xdat[,i], xtrim[i])
          ei = seq(qi[1], qi[2], length.out = neval)
        }

        if (xi.neval < maxneval){
          ei[(xi.neval+1):maxneval] = NA
        }

        if (quantreg) {
          tobj <- .np_plot_quantile_eval(
            txdat = xdat,
            tydat = ydat,
            exdat = subcol(exdat,ei,i)[seq_len(xi.neval),, drop = FALSE],
            gradients = gradients,
            tau = tau,
            bws = bws
          )
        } else {
          tobj <- .np_plot_conditional_eval(
            bws = bws,
            xdat = xdat,
            ydat = ydat,
            exdat = subcol(exdat,ei,i)[seq_len(xi.neval),, drop = FALSE],
            eydat = eydat[seq_len(xi.neval),, drop = FALSE],
            cdf = cdf,
            gradients = gradients,
            gradient.order = gradient.order
          )
        }
        if (!quantreg && isTRUE(proper.args$proper.requested)) {
          tobj$proper.requested <- TRUE
          tobj$proper.applied <- FALSE
          tobj$proper.method <- proper.args$proper.method
          tobj$proper.info <- .np_condens_make_reason_info(
            reason = "x_slices_not_repeated",
            supported = FALSE
          )
        }

        
        ## if there are gradients then we need to repeat the process for each component

        eval.extract <- function(obj, jj){
          if (gradients) {
            if (quantreg) obj$quantgrad[,jj] else obj$congrad[,jj]
          } else if (quantreg) {
            obj$quantile
          } else if (cdf) {
            obj$condist
          } else {
            obj$condens
          }
        }
        err.extract <- function(obj, jj){
          if (gradients) {
            if (quantreg) rep(NA_real_, length(eval.extract(obj, jj))) else obj$congerr[,jj]
          } else if (quantreg) {
            obj$quanterr
          } else {
            obj$conderr
          }
        }

        if (plot.behavior != "plot"){
          plot.out[plot.index] = NA
          plot.out[[plot.index]] = tobj
        }

        for (j in seq_len(dsf)){
          temp.boot = list()
          temp.all.err <- NULL
          temp.dens[seq_len(xi.neval)] <- eval.extract(tobj, j)
          
          if (plot.errors){
            if (plot.errors.method == "asymptotic") {
              terr.j <- err.extract(tobj, j)
              asym.obj <- .np_plot_asymptotic_error_from_se(
                se = terr.j,
                alpha = plot.errors.alpha,
                band.type = plot.errors.type,
                m = xi.neval
              )
              temp.err[seq_len(xi.neval),1:2] <- asym.obj$err
              temp.all.err <- asym.obj$all.err
            }
            else if (plot.errors.method == "bootstrap"){
              temp.boot <- compute.bootstrap.errors(
                        xdat = xdat,
                        ydat = ydat,
                        exdat = subcol(exdat,ei,i)[seq_len(xi.neval),, drop = FALSE],
                        eydat = eydat[seq_len(xi.neval),, drop = FALSE],
                        cdf = cdf,
                        quantreg = quantreg,
                        tau = tau,
                        gradients = gradients,
                        gradient.order = gradient.order,
                        gradient.index = j,
                        slice.index = plot.index,
                        plot.errors.boot.method = plot.errors.boot.method,
                        plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
                        plot.errors.boot.blocklen = plot.errors.boot.blocklen,
                        plot.errors.boot.num = plot.errors.boot.num,
                        plot.errors.center = plot.errors.center,
                        plot.errors.type = plot.errors.type,
                        plot.errors.alpha = plot.errors.alpha,
                        progress.target = .np_plot_conditional_bootstrap_target_label(
                          bws = bws,
                          slice.index = plot.index,
                          gradients = gradients,
                          gradient.index = j
                        ),
                        bws = bws)
              temp.err[seq_len(xi.neval),] <- temp.boot[["boot.err"]]
              temp.all.err <- temp.boot[["boot.all.err"]]
              temp.boot <- temp.boot[["bxp"]]
              if (!plot.bxp.out){
                temp.boot$out <- numeric()
                temp.boot$group <- integer()
              }
            }
          }
          
          if (common.scale){
            allei[,plot.index] = ei
            data.eval[,(plot.index-1)*dsf+j] = temp.dens
            if (plot.errors){
              all.bxp[plot.index] = NA
              all.bxp[[plot.index]] = temp.boot

              data.err[,seq(3*((plot.index-1)*dsf+j)-2,length=3)] = temp.err
              data.err.all[[(plot.index-1)*dsf+j]] = temp.all.err
            }
          } else if (plot.behavior != "data") {
            plot.layout <- .np_plot_layout_activate(plot.layout)
            ## plot evaluation
            plot.fun <- if (xi.factor) {
              .np_plot_panel_fun(plot.bootstrap = plot.bootstrap, plot.bxp = plot.bxp)
            } else {
              plot
            }
            plot.args <- list()
            if (xi.factor) {
              if (plot.bootstrap && plot.bxp) plot.args$z <- temp.boot else plot.args$f <- ei
            } else {
              plot.args$x <- ei
            }
            if (!(xi.factor && plot.bootstrap && plot.bxp))
              plot.args$y <- temp.dens
            if (plot.errors)
              plot.args$ylim <- if (plot.errors.type == "all")
                compute.all.error.range(if (plotOnEstimate) temp.dens else temp.err[,3], temp.all.err)
              else
                c(min(na.omit(c(temp.dens - temp.err[,1], temp.err[,3] - temp.err[,1]))),
                  max(na.omit(c(temp.dens + temp.err[,2], temp.err[,3] + temp.err[,2]))))
            plot.args$xlab <- scalar_default(xlab, gen.label(if (xOrY == "x") bws$xnames[i] else bws$ynames[i], paste(toupper(xOrY), i, sep = "")))
            plot.args$ylab <- scalar_default(ylab, if (gradients) paste("GC", j, "of", tylabE) else tylabE)
            if (!xi.factor) {
              plot.args$type <- scalar_default(type, "l")
              plot.args$lty <- scalar_default(lty, par()$lty)
              plot.args$col <- scalar_default(col, par()$col)
              plot.args$lwd <- scalar_default(lwd, par()$lwd)
              plot.args$cex.axis <- scalar_default(cex.axis, par()$cex.axis)
              plot.args$cex.lab <- scalar_default(cex.lab, par()$cex.lab)
              plot.args$cex.main <- scalar_default(cex.main, par()$cex.main)
              plot.args$cex.sub <- scalar_default(cex.sub, par()$cex.sub)
            } else {
              if (!is.null(col)) plot.args$col <- col
              if (!is.null(lty)) plot.args$lty <- lty
              if (!is.null(lwd)) plot.args$lwd <- lwd
            }
            plot.args$main <- scalar_default(main, "")
            plot.args$sub <- scalar_default(sub, "")
            plot.args <- .np_plot_merge_user_args(
              plot.args,
              if (xi.factor && plot.bootstrap && plot.bxp) bxp.args else plot.user.args
            )
            do.call(plot.fun, plot.args)
            if (plot.rug && !xi.factor)
              .np_plot_draw_rug_1d(if (xOrY == "x") xdat[,i] else ydat[,i])

            ## error plotting evaluation
            if (plot.errors && !(xi.factor && plot.bootstrap && plot.bxp)){
              if (plot.errors.type == "all") {
                draw.all.error.types(
                  ex = as.numeric(na.omit(ei)),
                  center = na.omit(if (plotOnEstimate) temp.dens else temp.err[,3]),
                  all.err = temp.all.err,
                  xi.factor = xi.factor,
                  legend = plot.legend)
              } else {
                if (!xi.factor && !plotOnEstimate)
                  lines(na.omit(ei), na.omit(temp.err[,3]), lty = 3)
                draw.args <- list(
                  ex = as.numeric(na.omit(ei)),
                  ely = if (plotOnEstimate) na.omit(temp.dens - temp.err[,1]) else na.omit(temp.err[,3] - temp.err[,1]),
                  ehy = if (plotOnEstimate) na.omit(temp.dens + temp.err[,2]) else na.omit(temp.err[,3] + temp.err[,2]),
                  plot.errors.style = if (xi.factor) "bar" else plot.errors.style,
                  plot.errors.bar = if (xi.factor) "I" else plot.errors.bar,
                  plot.errors.bar.num = plot.errors.bar.num,
                  lty = if (xi.factor) 1 else 2
                )
                do.call(draw.errors, draw.args)
              }
            }
          }
        
          if (plot.behavior != "plot" && plot.errors) {
            err.name <- if (gradients) paste("gc", j, "err", sep = "") else if (quantreg) "quanterr" else "conderr"
            bias.name <- if (gradients) paste("gc", j, "bias", sep = "") else "bias"
            plot.out[[plot.index]][[err.name]] <- na.omit(cbind(-temp.err[,1], temp.err[,2]))
            plot.out[[plot.index]][[bias.name]] <- na.omit(temp.dens - temp.err[,3])
            plot.out[[plot.index]]$bxp <- temp.boot
          }
        }
      }

      if (!quantreg){
        xOrY = "y"
        for (i in seq_len(bws$yndim)){
          plot.index = plot.index + 1
          temp.err[,] = NA
          temp.dens[] =  NA
          temp.boot = list()

          xi.factor = all.isFactor[plot.index]
          
          if (xi.factor){
            ei = bws$ydati$all.ulev[[i]]
            xi.neval = length(ei)
          } else {
            xi.neval = neval
            qi = trim.quantiles(ydat[,i], ytrim[i])
            ei = seq(qi[1], qi[2], length.out = neval)
          }

          if (xi.neval < maxneval){
            ei[(xi.neval+1):maxneval] = NA
          }

          if (quantreg) {
            tobj <- .np_plot_quantile_eval(
              txdat = xdat,
              tydat = ydat,
              eydat = subcol(eydat,ei,i)[seq_len(xi.neval),, drop = FALSE],
              gradients = gradients,
              tau = tau,
              bws = bws
            )
          } else {
            tobj <- .np_plot_conditional_eval(
              bws = bws,
              xdat = xdat,
              ydat = ydat,
              exdat = exdat[seq_len(xi.neval),, drop = FALSE],
              eydat = subcol(eydat,ei,i)[seq_len(xi.neval),, drop = FALSE],
              cdf = cdf,
              gradients = gradients,
              gradient.order = gradient.order,
              proper = isTRUE(proper.args$proper.requested),
              proper.method = proper.args$proper.method,
              proper.control = proper.args$proper.control
            )
          }

          
          ## if there are gradients then we need to repeat the process for each component

          eval.extract <- function(obj, jj){
            if (gradients) {
              if (quantreg) obj$quantgrad[,jj] else obj$congrad[,jj]
            } else if (quantreg) {
              obj$quantile
            } else if (cdf) {
              obj$condist
            } else {
              obj$condens
            }
          }
          err.extract <- function(obj, jj){
            if (gradients) {
              if (quantreg) rep(NA_real_, length(eval.extract(obj, jj))) else obj$congerr[,jj]
            } else if (quantreg) {
              obj$quanterr
            } else {
              obj$conderr
            }
          }

          if (plot.behavior != "plot"){
            plot.out[plot.index] = NA
            plot.out[[plot.index]] = tobj
          }

          for (j in seq_len(dsf)){
            temp.boot = list()
            temp.all.err <- NULL
            temp.dens[seq_len(xi.neval)] <- eval.extract(tobj, j)
            
            if (plot.errors){
              if (plot.errors.method == "asymptotic") {
                terr.j <- err.extract(tobj, j)
                asym.obj <- .np_plot_asymptotic_error_from_se(
                  se = terr.j,
                  alpha = plot.errors.alpha,
                  band.type = plot.errors.type,
                  m = xi.neval
                )
                temp.err[seq_len(xi.neval),1:2] <- asym.obj$err
                temp.all.err <- asym.obj$all.err
              }
              else if (plot.errors.method == "bootstrap"){
                temp.boot <- compute.bootstrap.errors(
                          xdat = xdat,
                          ydat = ydat,
                          exdat = exdat[seq_len(xi.neval),, drop = FALSE],
                          eydat = subcol(eydat,ei,i)[seq_len(xi.neval),, drop = FALSE],
                          cdf = cdf,
                          quantreg = quantreg,
                          tau = tau,
                          gradients = gradients,
                          gradient.order = gradient.order,
                          gradient.index = j,
                          slice.index = plot.index,
                          plot.errors.boot.method = plot.errors.boot.method,
                          plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
                          plot.errors.boot.blocklen = plot.errors.boot.blocklen,
                          plot.errors.boot.num = plot.errors.boot.num,
                          plot.errors.center = plot.errors.center,
                          plot.errors.type = plot.errors.type,
                          plot.errors.alpha = plot.errors.alpha,
                          progress.target = .np_plot_conditional_bootstrap_target_label(
                            bws = bws,
                            slice.index = plot.index,
                            gradients = gradients,
                            gradient.index = j
                          ),
                          proper = isTRUE(proper.args$proper.requested),
                          proper.method = proper.args$proper.method,
                          proper.control = proper.args$proper.control,
                          bws = bws)
                temp.err[seq_len(xi.neval),] <- temp.boot[["boot.err"]]
                temp.all.err <- temp.boot[["boot.all.err"]]
                temp.boot <- temp.boot[["bxp"]]
                if (!plot.bxp.out){
                  temp.boot$out <- numeric()
                  temp.boot$group <- integer()
                }
              }
            }
            
            if (common.scale){
              allei[,plot.index] = ei
              data.eval[,(plot.index-1)*dsf+j] = temp.dens
              if (plot.errors){
                all.bxp[plot.index] = NA
                all.bxp[[plot.index]] = temp.boot

                data.err[,seq(3*((plot.index-1)*dsf+j)-2,length=3)] = temp.err
                data.err.all[[(plot.index-1)*dsf+j]] = temp.all.err
              }
            } else if (plot.behavior != "data") {
              plot.layout <- .np_plot_layout_activate(plot.layout)
              ## plot evaluation
              plot.fun <- if (xi.factor) {
                .np_plot_panel_fun(plot.bootstrap = plot.bootstrap, plot.bxp = plot.bxp)
              } else {
                plot
              }
              plot.args <- list()
              if (xi.factor) {
                if (plot.bootstrap && plot.bxp) plot.args$z <- temp.boot else plot.args$f <- ei
              } else {
                plot.args$x <- ei
              }
              if (!(xi.factor && plot.bootstrap && plot.bxp))
                plot.args$y <- temp.dens
              if (plot.errors)
                plot.args$ylim <- if (plot.errors.type == "all")
                  compute.all.error.range(if (plotOnEstimate) temp.dens else temp.err[,3], temp.all.err)
                else
                  c(min(na.omit(c(temp.dens - temp.err[,1], temp.err[,3] - temp.err[,1]))),
                    max(na.omit(c(temp.dens + temp.err[,2], temp.err[,3] + temp.err[,2]))))
              plot.args$xlab <- scalar_default(xlab, gen.label(if (xOrY == "x") bws$xnames[i] else bws$ynames[i], paste(toupper(xOrY), i, sep = "")))
              plot.args$ylab <- scalar_default(ylab, if (gradients) paste("GC", j, "of", tylabE) else tylabE)
              if (!xi.factor) {
                plot.args$type <- scalar_default(type, "l")
                plot.args$lty <- scalar_default(lty, par()$lty)
                plot.args$col <- scalar_default(col, par()$col)
                plot.args$lwd <- scalar_default(lwd, par()$lwd)
                plot.args$cex.axis <- scalar_default(cex.axis, par()$cex.axis)
                plot.args$cex.lab <- scalar_default(cex.lab, par()$cex.lab)
                plot.args$cex.main <- scalar_default(cex.main, par()$cex.main)
                plot.args$cex.sub <- scalar_default(cex.sub, par()$cex.sub)
              } else {
                if (!is.null(col)) plot.args$col <- col
                if (!is.null(lty)) plot.args$lty <- lty
                if (!is.null(lwd)) plot.args$lwd <- lwd
              }
              plot.args$main <- scalar_default(main, "")
              plot.args$sub <- scalar_default(sub, "")
              plot.args <- .np_plot_merge_user_args(
                plot.args,
                if (xi.factor && plot.bootstrap && plot.bxp) bxp.args else plot.user.args
              )
              do.call(plot.fun, plot.args)
              if (plot.rug && !xi.factor)
                .np_plot_draw_rug_1d(if (xOrY == "x") xdat[,i] else ydat[,i])

              ## error plotting evaluation
              if (plot.errors && !(xi.factor && plot.bootstrap && plot.bxp)){
                if (plot.errors.type == "all") {
                  draw.all.error.types(
                    ex = as.numeric(na.omit(ei)),
                    center = na.omit(if (plotOnEstimate) temp.dens else temp.err[,3]),
                    all.err = temp.all.err,
                    xi.factor = xi.factor,
                  legend = plot.legend)
                } else {
                  if (!xi.factor && !plotOnEstimate)
                    lines(na.omit(ei), na.omit(temp.err[,3]), lty = 3)
                  draw.args <- list(
                    ex = as.numeric(na.omit(ei)),
                    ely = if (plotOnEstimate) na.omit(temp.dens - temp.err[,1]) else na.omit(temp.err[,3] - temp.err[,1]),
                    ehy = if (plotOnEstimate) na.omit(temp.dens + temp.err[,2]) else na.omit(temp.err[,3] + temp.err[,2]),
                    plot.errors.style = if (xi.factor) "bar" else plot.errors.style,
                    plot.errors.bar = if (xi.factor) "I" else plot.errors.bar,
                    plot.errors.bar.num = plot.errors.bar.num,
                    lty = if (xi.factor) 1 else 2
                  )
                  do.call(draw.errors, draw.args)
                }
              }
            }
              
            if (plot.behavior != "plot" && plot.errors) {
              err.name <- if (gradients) paste("gc", j, "err", sep = "") else if (quantreg) "quanterr" else "conderr"
              bias.name <- if (gradients) paste("gc", j, "bias", sep = "") else "bias"
              plot.out[[plot.index]][[err.name]] <- na.omit(cbind(-temp.err[,1], temp.err[,2]))
              plot.out[[plot.index]][[bias.name]] <- na.omit(temp.dens - temp.err[,3])
              plot.out[[plot.index]]$bxp <- temp.boot
            }
          }
        }
      }
      
      if (common.scale && (plot.behavior != "data")){
        if (plot.errors && plot.errors.type == "all") {
          y.min <- Inf
          y.max <- -Inf
          for (k in seq_len(tot.dim*dsf)) {
            if (is.null(data.err.all[[k]])) next
            nkeep.k <- nrow(data.err.all[[k]]$pointwise)
            if (nkeep.k == 0) next
            center.k <- if (plot.errors.center == "estimate")
              na.omit(data.eval[seq_len(nkeep.k), k])
            else
              na.omit(data.err[seq_len(nkeep.k), 3*k])
            range.k <- compute.all.error.range(center.k, data.err.all[[k]])
            y.min <- min(y.min, range.k[1], na.rm = TRUE)
            y.max <- max(y.max, range.k[2], na.rm = TRUE)
          }
          if (!is.finite(y.min) || !is.finite(y.max)) {
            y.min <- min(na.omit(as.double(data.eval)))
            y.max <- max(na.omit(as.double(data.eval)))
          }
        } else {
          jj = seq_len(dsf*tot.dim)*3
          if (plot.errors.center == "estimate" || !plot.errors) {
            y.max = max(na.omit(as.double(data.eval)) +
              if (plot.errors) na.omit(as.double(data.err[,jj-1]))
              else 0)
            y.min = min(na.omit(as.double(data.eval)) -
              if (plot.errors) na.omit(as.double(data.err[,jj-2]))
              else 0)
          } else if (plot.errors.center == "bias-corrected") {
            y.max = max(na.omit(as.double(data.err[,jj] + data.err[,jj-1])))
            y.min = min(na.omit(as.double(data.err[,jj] - data.err[,jj-2])))
          }
        }

        if(!is.null(ylim)){
          y.min = ylim[1]
          y.max = ylim[2]
        }
        
        xOrY = "x"
        
        for (plot.index in seq_len(tot.dim)){
          i = if (plot.index <= bws$xndim) plot.index else plot.index - bws$xndim

          if (plot.index > bws$xndim)
            xOrY <- "y"
            
          xi.factor = all.isFactor[plot.index]

          for (j in seq_len(dsf)){
            plot.layout <- .np_plot_layout_activate(plot.layout)
            ## plot evaluation
            idx <- (plot.index-1)*dsf+j
            plot.fun <- if (xi.factor) {
              .np_plot_panel_fun(plot.bootstrap = plot.bootstrap, plot.bxp = plot.bxp)
            } else {
              plot
            }
            plot.args <- list()
            if (xi.factor) {
              if (plot.bootstrap && plot.bxp) plot.args$z <- all.bxp[[plot.index]] else plot.args$f <- allei[,plot.index]
            } else {
              plot.args$x <- allei[,plot.index]
            }
            if (!(xi.factor && plot.bootstrap && plot.bxp))
              plot.args$y <- data.eval[,idx]
            plot.args$ylim <- c(y.min, y.max)
            plot.args$xlab <- scalar_default(xlab, gen.label(if (xOrY == "x") bws$xnames[i] else bws$ynames[i], paste(toupper(xOrY), i, sep = "")))
            plot.args$ylab <- scalar_default(ylab, if (gradients) paste("GC", j, "of", tylabE) else tylabE)
            if (!xi.factor) {
              plot.args$type <- scalar_default(type, "l")
              plot.args$lty <- scalar_default(lty, par()$lty)
              plot.args$col <- scalar_default(col, par()$col)
              plot.args$lwd <- scalar_default(lwd, par()$lwd)
              plot.args$cex.axis <- scalar_default(cex.axis, par()$cex.axis)
              plot.args$cex.lab <- scalar_default(cex.lab, par()$cex.lab)
              plot.args$cex.main <- scalar_default(cex.main, par()$cex.main)
              plot.args$cex.sub <- scalar_default(cex.sub, par()$cex.sub)
            } else {
              if (!is.null(col)) plot.args$col <- col
              if (!is.null(lty)) plot.args$lty <- lty
              if (!is.null(lwd)) plot.args$lwd <- lwd
            }
            plot.args$main <- scalar_default(main, "")
            plot.args$sub <- scalar_default(sub, "")
            plot.args <- .np_plot_merge_user_args(
              plot.args,
              if (xi.factor && plot.bootstrap && plot.bxp) bxp.args else plot.user.args
            )
            do.call(plot.fun, plot.args)
            if (plot.rug && !xi.factor)
              .np_plot_draw_rug_1d(if (xOrY == "x") xdat[,i] else ydat[,i])

            ## error plotting evaluation
            if (plot.errors && !(xi.factor && plot.bootstrap && plot.bxp)){
              if (plot.errors.type == "all") {
                draw.all.error.types(
                  ex = as.numeric(na.omit(allei[,plot.index])),
                  center = if (plotOnEstimate)
                    na.omit(data.eval[,idx])
                  else
                    na.omit(data.err[,3*idx]),
                  all.err = data.err.all[[idx]],
                  xi.factor = xi.factor,
                  legend = plot.legend)
              } else {
                if (!xi.factor && !plotOnEstimate)
                  lines(na.omit(ei), na.omit(temp.err[,3]), lty = 3)
                draw.args <- list(
                  ex = as.numeric(na.omit(allei[,plot.index])),
                  ely = if (plotOnEstimate) na.omit(data.eval[,idx] - data.err[,3*idx-2]) else na.omit(data.err[,3*idx] - data.err[,3*idx-2]),
                  ehy = if (plotOnEstimate) na.omit(data.eval[,idx] + data.err[,3*idx-1]) else na.omit(data.err[,3*idx] + data.err[,3*idx-1]),
                  plot.errors.style = if (xi.factor) "bar" else plot.errors.style,
                  plot.errors.bar = if (xi.factor) "I" else plot.errors.bar,
                  plot.errors.bar.num = plot.errors.bar.num,
                  lty = if (xi.factor) 1 else 2
                )
                do.call(draw.errors, draw.args)
              }
            }
          }
        }
      }

      if (plot.behavior != "data" && plot.par.mfrow)
        par(mfrow=c(1,1),cex=par()$cex)

      if (plot.behavior != "plot"){
        names(plot.out) = paste("cd", seq_len(tot.dim), sep="")
        return (plot.out)
      }
    }
  }

npuniden.reflect <- function(X = NULL,
                             Y = NULL,
                             h = NULL,
                             a = 0,
                             b = 1,
                             ...) {
    ## Data reflection - note that the bandwidth should be based on
    ## the original sample of n observations (SCOTT (1992), Page 149)
    if(is.null(X)) stop("you must pass a vector X")
    if(anyNA(X)) stop("X must not contain missing values")
    if(!is.finite(a) && !is.finite(b)) stop("one of either a or b must be finite")
    if(a>=b) stop("a must be less than b")
    if(any(X<a, na.rm = TRUE)) stop("X must be >= a")
    if(any(X>b, na.rm = TRUE)) stop("X must be <= b")
    if(!is.null(Y) && anyNA(Y)) stop("Y must not contain missing values")
    if(!is.null(Y) && any(Y<a, na.rm = TRUE)) stop("Y must be >= a")
    if(!is.null(Y) && any(Y>b, na.rm = TRUE)) stop("Y must be <= b")
    if(is.null(Y)) Y <- X
    if(!is.null(h) && h <= 0) stop("bandwidth h must be positive")
    if(is.null(h)) {
        bw <- npudensbw(~X,...)
        hh <- bw$bw
    } else {
        hh <- h
    }
    if(is.finite(a) && is.finite(b)) {
        X.reflect <- c(X,-X+2*a,-X+2*b)
        f.reflect <- npudens(tdat=X.reflect,edat=Y,bws=hh,...)
        f <- 3*fitted(f.reflect)
        cdf <- 3*fitted(npudist(tdat=X.reflect,edat=Y,bws=hh,...))-1
        std <- 3*se(f.reflect)
    } else if(is.finite(a) && !is.finite(b)) {
        X.reflect <- c(X,-X+2*a)
        f.reflect <- npudens(tdat=X.reflect,edat=Y,bws=hh,...)
        f <- 2*fitted(f.reflect)
        cdf <- 2*fitted(npudist(tdat=X.reflect,edat=Y,bws=hh,...))-1      
        std <- 2*se(f.reflect)
    } else if(!is.finite(a) && is.finite(b)) {
        X.reflect <- c(X,-X+2*b)
        f.reflect <- npudens(tdat=X.reflect,edat=Y,bws=hh,...)
        f <- 2*fitted(f.reflect)
        cdf <- 2*fitted(npudist(tdat=X.reflect,edat=Y,bws=hh,...))      
        std <- 2*se(f.reflect)
    }
    if(is.null(h)) {
        return(list(f=f,
                    F=cdf,
                    sd.f=std,
                    sd.F=sqrt(abs(cdf*(1-cdf)/length(cdf))),
                    h=hh,
                    nmulti=length(bw$fval.history),
                    cv.opt=bw$fval))
    } else {
        return(list(f=f,
                    F=cdf,
                    sd.f=std,
                    sd.F=sqrt(abs(cdf*(1-cdf)/length(cdf))),
                    h=hh))
    }
}

test_that("npuniden.boundary basic functionality works", {
  set.seed(42)
  n <- 50
  X <- rbeta(n, 5, 1)
  
  # Use fixed bandwidth for speed
  model <- npuniden.boundary(X, h=0.05, a=0, b=1)
  
  expect_type(model, "list")
  expect_true("f" %in% names(model))
  expect_length(model$f, n)
  expect_true(all(model$f >= 0))
})

test_that("npuniden.reflect basic functionality works", {
  set.seed(42)
  n <- 50
  X <- rbeta(n, 5, 1)
  
  model <- npuniden.reflect(X, h=0.05, a=0, b=1)
  
  expect_type(model, "list")
  expect_true("f" %in% names(model))
})

test_that("npuniden.sc basic functionality works", {
  set.seed(42)
  n <- 50
  X <- rbeta(n, 5, 1)
  
  # Shape constrained density estimation
  model <- npuniden.sc(X, h=0.05, a=0, b=1, lb=0, ub=Inf, constraint="density")
  
  expect_type(model, "list")
  expect_true("f" %in% names(model))
})

## This functions accepts the following arguments:

## y: univariate outcome
## z: endogenous predictors
## w: instruments
## x: exogenous predictors

## zeval: optional evaluation data for the endogenous predictors
## weval: optional evaluation data for the instruments
## xeval: optional evaluation data for the exogenous predictors

## ... optional arguments for npreg()

## This function returns a list with the following elements:

## phi: the IV estimator of phi(z) corresponding to the estimated
## derivative phi(z)
## phi.prime: the IV derivative estimator
## phi.mat: the matrix with colums phi_1, phi_2 etc. over all iterations
## phi.prime.mat: the matrix with colums phi'_1, phi'_2 etc. over all iterations
## num.iterations: number of iterations taken by Landweber-Fridman
## norm.stop: the stopping rule for each Landweber-Fridman iteration
## convergence: a character string indicating whether/why iteration terminated

## First, a series of functions for local polynomial kernel regression
## (regtype = "lp", with lc/ll as special degree-restricted cases)

## This function returns the weight matrix for a local polynomial

## supports mixed data types. It presumes that Y is in column 1. Basic
## error checking is undertaken. j.reg= strips off weights for mean
## (1), partials up to order p, and cross-partials. All partials and
## cross partials are wrt continuous regressors, and cross-partials
## require k > 1 and p > 1. Shrinking towards the local constant mean,
## first, and second partial derivatives is implemented for regions
## where the local polynomial estimator is ill-conditioned (sparse
## data, small h etc.).

npregivderiv <- function(y,
                         z,
                         w,
                         x=NULL,
                         zeval=NULL,
                         weval=NULL,
                         xeval=NULL,
                         constant=0.5,
                         iterate.break=TRUE,
                         iterate.max=1000,
                         nmulti=NULL,
                         random.seed=42,
                         smooth.residuals=TRUE,
                         start.from=c("Eyz","EEywz"),
                         starting.values=NULL,
                         stop.on.increase=TRUE,
                         ...) {

  ptm.start <- proc.time()
  cl <- match.call()

  ## Basic error checking

  if(!is.logical(stop.on.increase)) stop("stop.on.increase must be logical (TRUE/FALSE)")
  if(!is.logical(smooth.residuals)) stop("smooth.residuals must be logical (TRUE/FALSE)")

  start.from <- match.arg(start.from)

  if(iterate.max < 2) stop("iterate.max must be at least 2")
  if(constant <= 0 || constant >= 1) stop("constant must lie in the range (0,1)")

  if(missing(y)) stop("You must provide y")
  if(missing(z)) stop("You must provide z")
  if(missing(w)) stop("You must provide w")
  if(NCOL(y) > 1) stop("y must be univariate")
  if(NROW(y) != NROW(z) || NROW(y) != NROW(w)) stop("y, z, and w have differing numbers of rows")
  if(!is.null(x) && NROW(y) != NROW(x)) stop("y and x have differing numbers of rows")

  ## Check for evaluation data

  if(is.null(zeval)) zeval <- z
  if(is.null(weval)) weval <- w
  if(is.null(weval)) xeval <- x

  if(!is.null(starting.values) && (NROW(starting.values) != NROW(zeval))) stop(paste("starting.values must be of length",NROW(zeval)))

  ## Need to determine how many x, w, z are numeric

  z <- data.frame(z)
  w <- data.frame(w)
  if(!is.null(x)) z <- data.frame(z,x)
  if(!is.null(xeval)) zeval <- data.frame(zeval,xeval)

  if(!is.null(nmulti)) nmulti <- npValidateNmulti(nmulti)
  nmulti.loop <- if(!is.null(nmulti)) nmulti else 1
  nmulti <- if(!is.null(nmulti)) nmulti else npDefaultNmulti(max(NCOL(z), NCOL(w)))

  z.numeric <- sapply(seq_len(NCOL(z)),function(i){is.numeric(z[,i])})
  num.z.numeric <- NCOL(as.data.frame(z[,z.numeric]))

  w.numeric <- sapply(seq_len(NCOL(w)),function(i){is.numeric(w[,i])})
  num.w.numeric <- NCOL(as.data.frame(w[,w.numeric]))

  ## Landweber-Fridman

  ## We begin the iteration computing phi.prime.0

  ## Note - here I am only treating the univariate case, so let's
  ## throw a stop with warning for now...

  if(NCOL(z) > 1) stop(" This version supports univariate z only (beta after all)")

  ## For all results we need the density function for Z and the
  ## survivor function for Z (1-CDF of Z)

  .np_progress_note("Preparing IV derivative regression")

  ## Let's compute the bandwidth object for the unconditional
  ## density for the moment. Use the normal-reference rule for speed
  ## considerations, same smoothing for PDF and CDF.

  bw <- .np_progress_with_legacy_suppressed(npudensbw(dat=z, bwmethod="normal-reference"))
  model.fz <- .np_progress_with_legacy_suppressed(npudens(tdat=z, bws=bw$bw))
  f.z <- predict(model.fz, newdata=zeval)
  model.Sz <- .np_progress_with_legacy_suppressed(npudist(tdat=z, bws=bw$bw))
  S.z <- 1-predict(model.Sz, newdata=zeval)

  ## For stopping rule...

  model.E.y.w <- .np_progress_with_legacy_suppressed(npreg(tydat=y,
                                                           txdat=w,
                                                           exdat=weval,
                                                           nmulti=nmulti,
                                                           ...))
  E.y.w <- model.E.y.w$mean
  bw.E.y.w <- model.E.y.w$bws

  ## Potential alternative starting rule (consistent with
  ## npregiv). Here we start with E(Y|Z) rather than zero

  if(is.null(starting.values)) {

    model.phi.prime <- .np_progress_with_legacy_suppressed(npreg(tydat=if(start.from=="Eyz") y else E.y.w,
                                                                 txdat=z,
                                                                 exdat=zeval,
                                                                 gradients=TRUE,
                                                                 nmulti=nmulti,
                                                                 ...))
    phi.prime <- model.phi.prime$grad[,1]
    bw.E.y.z <- model.phi.prime$bws

  } else {

    phi.prime <- starting.values
    bw.E.y.z <- NULL

  }

  ## Now we can compute phi.0 by integrating phi.prime.0 up to each
  ## sample realization (here we use the trapezoidal rule)

  ## NOTE - this presumes univariate z case... in general this would
  ## be a continuous variable's index

  # Pre-calculate components for integrate.trapezoidal
  z.val <- z[,1]
  n.z <- length(z.val)
  order.z <- order(z.val)
  z.sorted <- z.val[order.z]
  dz.sorted <- diff(z.sorted)
  cz.z <- dz.sorted[1]
  inv.order.z <- order(order.z)

  integrate.trapezoidal.internal <- function(y.val) {
      y.sorted <- y.val[order.z]
      dy.sorted <- diff(y.sorted)
      ca.z <- dy.sorted[1] / cz.z
      cb.z <- dy.sorted[n.z - 1] / cz.z
      cf.z <- cz.z^2 / 12 * (cb.z - ca.z)
      if (!is.finite(cf.z)) cf.z <- 0
      int.vec <- c(0, cumsum(dz.sorted * (y.sorted[-n.z] + y.sorted[-1]) / 2))
      int.vec <- int.vec - cf.z
      int.vec[inv.order.z]
  }

  phi <- integrate.trapezoidal.internal(phi.prime)

  starting.values.phi <- phi
  starting.values.phi.prime <- phi.prime

  ## In the definition of phi we have the integral minus the mean of
  ## the integral with respect to z, so subtract the mean here

  phi <- phi - mean(phi) + mean(y)

  phi.mat <- matrix(NA, length(phi), iterate.max)
  phi.mat[,1] <- phi
  phi.prime.mat <- matrix(NA, length(phi.prime), iterate.max)
  phi.prime.mat[,1] <- phi.prime

  norm.stop <- numeric(iterate.max)

  ## Now we compute mu.0 (a residual of sorts)

  mu <- y - phi

  ## Now we repeat this entire process using mu = y = phi.0 rather than y

  progress <- .np_progress_begin("Iterating Landweber-Fridman derivative solve", surface = "iv_solve")

  if(smooth.residuals) {
    progress <- .np_progress_step(progress, done = 1, detail = "updating E(mu|w)")

    ## Additional smoothing on top of the stopping rule required, but
    ## we have computed the stopping rule so reuse the bandwidth
    ## vector to be passed below. Here we compute the bandwidth
    ## optimal for the regression of mu on w.

    ## Next, we regress require \mu_{0,i} W using bws optimal for phi on w

    model.mu.w <- .np_progress_with_legacy_suppressed(npreg(tydat=mu,
                                                            txdat=w,
                                                            exdat=weval,
                                                            ...))
    predicted.E.mu.w <- model.mu.w$mean
    bw.mu.w <- model.mu.w$bws

  } else {

    model.phi.w <- .np_progress_with_legacy_suppressed(npreg(tydat=phi,
                                                             txdat=w,
                                                             exdat=weval,
                                                             ...))
    E.phi.w <- model.phi.w$mean
    bw.mu.w <- model.phi.w$bws

    predicted.E.mu.w <- E.y.w - E.phi.w

  }


  norm.stop[1] <- sum(predicted.E.mu.w^2)/NZD_pos(sum(E.y.w^2))
  
  ## We again require the mean of the fitted values

  mean.predicted.E.mu.w <- mean(predicted.E.mu.w)

  ## Now we compute T^* applied to mu

  cdf.weighted.average <- npksum(txdat=z,
                                 exdat=zeval,
                                 tydat=as.matrix(predicted.E.mu.w),
                                 operator="integral",
                                 ukertype="liracine",
                                 okertype="liracine",
                                 bws=bw$bw,
                                 ...)$ksum/length(y)

  survivor.weighted.average <- mean.predicted.E.mu.w - cdf.weighted.average

  T.star.mu <- (survivor.weighted.average-S.z*mean.predicted.E.mu.w)/NZD(f.z)

  ## Now we update phi.prime.0, this provides phi.prime.1, and now
  ## we can iterate until convergence... note we replace phi.prime.0
  ## with phi.prime.1 (i.e. overwrite phi.prime)

  phi.prime <- phi.prime + constant*T.star.mu

  phi.mat <- matrix(NA, length(phi), iterate.max)
  phi.mat[,1] <- phi
  phi.prime.mat <- matrix(NA, length(phi.prime), iterate.max)
  phi.prime.mat[,1] <- phi.prime

  norm.stop <- numeric(iterate.max)
  norm.stop[1] <- sum(predicted.E.mu.w^2)/NZD_pos(sum(E.y.w^2))
  
  ## We again require the mean of the fitted values

  mean.predicted.E.mu.w <- mean(predicted.E.mu.w)

  ## This we iterate...

  for(j in 2:iterate.max) {

    progress <- .np_progress_step(progress, done = j, detail = "updating phi")

    ## NOTE - this presumes univariate z case... in general this would
    ## be a continuous variable's index

    phi <- integrate.trapezoidal.internal(phi.prime)

    ## In the definition of phi we have the integral minus the mean of
    ## the integral with respect to z, so subtract the mean here

    phi <- phi - mean(phi) + mean(y)

    ## Now we compute mu.0 (a residual of sorts)

    mu <- y - phi

    ## Now we repeat this entire process using mu = y = phi.0 rather than y

    mean.mu <- mean(mu)

    if(smooth.residuals) {
      progress <- .np_progress_step(progress, done = j, detail = "updating E(mu|w)")


      ## Additional smoothing on top of the stopping rule required, but
      ## we have computed the stopping rule so reuse the bandwidth
      ## vector to be passed below. Here we compute the bandwidth
      ## optimal for the regression of mu on w.

      ## Next, we regress require \mu_{0,i} W using bws optimal for phi on w

      model.mu.w <- .np_progress_with_legacy_suppressed(npreg(tydat=mu,
                                                              txdat=w,
                                                              eydat=mu,
                                                              exdat=weval,
                                                              bws=bw.mu.w,
                                                              nmulti=nmulti.loop,
                                                              ...))
      predicted.E.mu.w <- model.mu.w$mean
      bw.mu.w <- model.mu.w$bws

    } else {

      model.phi.w <- .np_progress_with_legacy_suppressed(npreg(tydat=phi,
                                                               txdat=w,
                                                               eydat=phi,
                                                               exdat=weval,
                                                               bws=bw.mu.w,
                                                               nmulti=nmulti.loop,
                                                               ...))
      E.phi.w <- model.phi.w$mean
      bw.mu.w <- model.phi.w$bws

      predicted.E.mu.w <- E.y.w - E.phi.w

    }

    norm.stop[j] <- j*sum(predicted.E.mu.w^2)/NZD_pos(sum(E.y.w^2))
    
    mean.predicted.E.mu.w <- mean(predicted.E.mu.w)

    ## Now we compute T^* applied to mu

    cdf.weighted.average <- npksum(txdat=z,
                                   exdat=zeval,
                                   tydat=as.matrix(predicted.E.mu.w),
                                   operator="integral",
                                   ukertype="liracine",
                                   okertype="liracine",
                                   bws=bw$bw,
                                   ...)$ksum/length(y)

    survivor.weighted.average <- mean.predicted.E.mu.w - cdf.weighted.average

    T.star.mu <- (survivor.weighted.average-S.z*mean.mu)/NZD(f.z)

    ## Now we update, this provides phi.prime.1, and now we can
    ## iterate until convergence...

    phi.prime <- phi.prime + constant*T.star.mu
    phi.mat[,j] <- phi
    phi.prime.mat[,j] <- phi.prime

    ## The number of iterations in LF is asymptotically equivalent to
    ## 1/alpha (where alpha is the regularization parameter in
    ## Tikhonov).  Plus the criterion function we use is increasing
    ## for very small number of iterations. So we need a threshold
    ## after which we can pretty much confidently say that the
    ## stopping criterion is decreasing.  In Darolles et al. (2011)
    ## \alpha ~ O(N^(-1/(min(beta,2)+2)), where beta is the so called
    ## qualification of your regularization method. Take the worst
    ## case in which beta = 0 and then the number of iterations is ~
    ## N^0.5. Note that derivative estimation seems to require more
    ## iterations hence the heuristic sqrt(N)

    if(j > round(sqrt(nrow(z))) && !is.monotone.increasing(norm.stop)) {

      ## If stopping rule criterion increases or we are below stopping
      ## tolerance then break

      if(stop.on.increase && norm.stop[j] > norm.stop[j-1]) {
        convergence <- "STOP_ON_INCREASE"
        if(iterate.break) break()
      }

    }

    convergence <- "ITERATE_MAX"

  }

  ## Trim matrices and norm.stop to the actual number of iterations performed
  if(j < iterate.max) {
      phi.mat <- phi.mat[, 1:j, drop=FALSE]
      phi.prime.mat <- phi.prime.mat[, 1:j, drop=FALSE]
      norm.stop <- norm.stop[1:j]
  }

  ## Extract minimum, and check for monotone increasing function and
  ## issue warning in that case. Otherwise allow for an increasing
  ## then decreasing (and potentially increasing thereafter) portion
  ## of the stopping function, ignore the initial increasing portion,
  ## and take the min from where the initial inflection point occurs
  ## to the length of norm.stop.

  if(is.monotone.increasing(norm.stop)) {
    .np_warning("Stopping rule increases monotonically (consult model$norm.stop):\nThis could be the result of an inspired initial value (unlikely)\nNote: we suggest manually choosing phi.0 and restarting (e.g., instead set `start.from' to EEywz or provide a vector of starting values")
    convergence <- "FAILURE_MONOTONE_INCREASING"
    j <- length(norm.stop)
    phi <- phi.mat[,1]
    phi.prime <- phi.prime.mat[,1]
  } else {
    ## Ignore the initial increasing portion, take the min to the
    ## right of where the initial inflection point occurs.
    j <- 1
    ## Climb the initial hill...
    while(norm.stop[j+1] >= norm.stop[j] && j < length(norm.stop)) j <- j + 1
    ## Descend into the first valley
    while(norm.stop[j+1] < norm.stop[j] && j < length(norm.stop)) j <- j + 1
    ## When you start to climb again, stop, previous location was min
    phi <- phi.mat[,j-1]
    phi.prime <- phi.prime.mat[,j-1]
  }
  
  progress <- .np_progress_end(progress, detail = "updating E(mu|w)")

  if(j == iterate.max) .np_warning(" iterate.max reached: increase iterate.max or inspect norm.stop vector")

  ret <- list(phi=phi,
              phi.prime=phi.prime,
              phi.mat=phi.mat,
              phi.prime.mat=phi.prime.mat,
              num.iterations=j,
              norm.stop=norm.stop,
              convergence=convergence,
              starting.values.phi=starting.values.phi,
              starting.values.phi.prime=starting.values.phi.prime,
              bw.E.y.w=bw.E.y.w,
              bw.E.y.z=bw.E.y.z,
              call=cl,
              y=y,
              z=z,
              w=w,
              x=x,
              zeval=zeval,
              weval=weval,
              xeval=xeval,
              nmulti=nmulti,
              ptm=proc.time() - ptm.start)
  class(ret) <- "npregivderiv"
  return(ret)

}

print.npregivderiv <- function(x, ...) {
  summary.npregivderiv(x, ...)
  invisible(x)
}

summary.npregivderiv <- function(object, ...) {
  format_bw <- function(bw, label, names = NULL) {
    if(is.null(bw)) return()
    if(is.matrix(bw)) bw <- bw[nrow(bw), , drop = TRUE]
    bw <- as.numeric(bw)
    if(!is.null(names) && length(names) == length(bw)) {
      vals <- paste(paste(names, formatC(bw, digits=8, format="g"), sep=": "), collapse=", ")
    } else {
      vals <- paste(formatC(bw, digits=8, format="g"), collapse=", ")
    }
    cat(paste("\n", label, " ", vals, sep=""))
  }

  cat("Call:\n")
  print(object$call)

  cat("\nNonparametric Instrumental Kernel Derivative Estimation\n",sep="")

  cat(paste("\nNumber of continuous endogenous predictors: ",format(NCOL(object$z)),sep=""),sep="")
  cat(paste("\nNumber of continuous instruments: ",format(NCOL(object$w)),sep=""),sep="")
  if(!is.null(object$x)) cat(paste("\nNumber of continuous exogenous predictors: ",format(NCOL(object$x)),sep=""),sep="")

  cat(paste("\nTraining observations: ", format(NROW(object$y)), sep=""))

  cat(paste("\n\nRegularization method: Landweber-Fridman",sep=""))
  cat(paste("\nNumber of iterations: ", format(object$num.iterations), sep=""))
  cat(paste("\nStopping rule value: ", format(object$norm.stop[length(object$norm.stop)],digits=8), sep=""))

  w.names <- if(!is.null(object$w)) colnames(object$w) else NULL
  z.names <- if(!is.null(object$z)) colnames(object$z) else NULL
  format_bw(object$bw.E.y.w, "Bandwidth for E(y|w):", w.names)
  format_bw(object$bw.E.y.z, "Bandwidth for E(y|z):", z.names)

  cat(paste("\nNumber of multistarts: ", format(object$nmulti), sep=""))
  cat(paste("\nEstimation time: ", formatC(object$ptm[1],digits=1,format="f"), " seconds",sep=""))
  cat("\n\n")
}

plot.npregivderiv <- function(x,
                              plot.data = FALSE,
                              phi = FALSE,
                              ...) {

  object <- x
  .np_plot_validate_npregiv_call(
    sys.call(),
    method_args = c("plot.data", "phi"),
    context = "plot.npregivderiv"
  )
  dots <- list(...)
  take_arg <- function(name, default = NULL) {
    if (!is.null(dots[[name]])) {
      val <- dots[[name]]
      dots[[name]] <<- NULL
      return(val)
    }
    default
  }

  ## We only support univariate endogenous predictor z
  if(NCOL(object$z) > 1) stop(" only univariate z is supported")

  z <- object$z[,1]
  y <- object$y
  zname <- names(object$z)[1]
  yname <- "y"

  if(phi) {
    ## Plot the structural function phi
    fit <- object$phi
    ylab <- yname
  } else {
    ## Plot the derivative phi.prime (default for npregivderiv)
    fit <- object$phi.prime
    ylab <- paste("d", yname, "/d", zname, sep="")
  }

  if(plot.data && !phi) {
      ## Scatter data doesn't make sense for derivative plots directly
      plot.data <- FALSE
  }

  if(plot.data) {
    plot.type <- take_arg("type", "p")
    plot.xlab <- take_arg("xlab", zname)
    plot.ylab <- take_arg("ylab", yname)
    user.col <- take_arg("col", NULL)
    line.lwd <- take_arg("lwd", 2)
    plot.args <- c(list(x = z,
                        y = y,
                        xlab = plot.xlab,
                        ylab = plot.ylab,
                        type = plot.type,
                        col = if (is.null(user.col)) "lightgrey" else user.col),
                   dots)
    do.call(plot, plot.args)
    line.args <- list(x = z[order(z)],
                      y = fit[order(z)],
                      lwd = line.lwd)
    if (!is.null(user.col)) {
      line.args$col <- user.col
    }
    do.call(lines, c(line.args, dots))
  } else {
    plot.type <- take_arg("type", "l")
    plot.xlab <- take_arg("xlab", zname)
    plot.ylab <- take_arg("ylab", ylab)
    user.col <- take_arg("col", NULL)
    line.lwd <- take_arg("lwd", 2)
    plot.args <- list(x = z[order(z)],
                      y = fit[order(z)],
                      type = plot.type,
                      xlab = plot.xlab,
                      ylab = plot.ylab,
                      lwd = line.lwd)
    if (!is.null(user.col)) {
      plot.args$col <- user.col
    }
    do.call(plot, c(plot.args, dots))
  }
}

capture_progress_shadow_with_conditions <- function(expr, force_renderer = NULL, now = function() 0) {
  messages <- character()
  warnings <- character()

  shadow <- withCallingHandlers(
    capture_progress_shadow_trace(expr, force_renderer = force_renderer, now = now),
    message = function(m) {
      messages <<- c(messages, conditionMessage(m))
      invokeRestart("muffleMessage")
    },
    warning = function(w) {
      warnings <<- c(warnings, conditionMessage(w))
      invokeRestart("muffleWarning")
    }
  )

  list(
    value = shadow$value,
    trace = shadow$trace,
    final_line = shadow$final_line,
    messages = sub("\n$", "", messages),
    warnings = warnings
  )
}

progress_time_counter <- function(start = 0, by = 2.1) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

make_iv_data <- function(n = 18) {
  set.seed(42)
  z <- runif(n)
  w <- z + rnorm(n, sd = 0.1)
  y <- z^2 + rnorm(n, sd = 0.1)
  list(y = y, z = z, w = w)
}

cached_landweber_single_line <- local({
  cache <- NULL
  function() {
    if (is.null(cache)) {
      dat <- make_iv_data()
      old_opts <- options(np.messages = TRUE)
      on.exit(options(old_opts), add = TRUE)
      cache <<- capture_progress_shadow_with_conditions(
        npregiv(y = dat$y, z = dat$z, w = dat$w, method = "Landweber-Fridman", iterate.max = 2),
        force_renderer = "single_line",
        now = progress_time_counter()
      )
    }
    cache
  }
})

cached_tikhonov_single_line <- local({
  cache <- NULL
  function() {
    if (is.null(cache)) {
      dat <- make_iv_data()
      old_opts <- options(np.messages = TRUE)
      on.exit(options(old_opts), add = TRUE)
      cache <<- capture_progress_shadow_with_conditions(
        npregiv(
          y = dat$y,
          z = dat$z,
          w = dat$w,
          method = "Tikhonov",
          iterate.Tikhonov = TRUE,
          iterate.Tikhonov.num = 2
        ),
        force_renderer = "single_line",
        now = progress_time_counter()
      )
    }
    cache
  }
})

cached_landweber_no_residual_smoothing <- local({
  cache <- NULL
  function() {
    if (is.null(cache)) {
      dat <- make_iv_data()
      old_opts <- options(np.messages = TRUE)
      on.exit(options(old_opts), add = TRUE)
      cache <<- capture_progress_shadow_with_conditions(
        npregiv(
          y = dat$y,
          z = dat$z,
          w = dat$w,
          method = "Landweber-Fridman",
          smooth.residuals = FALSE,
          iterate.max = 2
        ),
        force_renderer = "single_line",
        now = progress_time_counter()
      )
    }
    cache
  }
})

cached_npregivderiv_parity <- local({
  cache <- NULL
  function() {
    if (is.null(cache)) {
      old_opts <- options(np.messages = TRUE)
      on.exit(options(old_opts), add = TRUE)

      dat <- make_iv_data()
      legacy <- capture_progress_shadow_with_conditions(
        npregivderiv(y = dat$y, z = dat$z, w = dat$w, iterate.max = 2),
        force_renderer = "legacy",
        now = progress_time_counter()
      )

      single_line <- capture_progress_shadow_with_conditions(
        npregivderiv(y = dat$y, z = dat$z, w = dat$w, iterate.max = 2),
        force_renderer = "single_line",
        now = progress_time_counter()
      )

      cache <<- list(legacy = legacy, single_line = single_line)
    }
    cache
  }
})

shadow_lines_matching <- function(shadow, pattern) {
  lines <- vapply(shadow$trace, `[[`, character(1L), "line")
  lines[grepl(pattern, lines)]
}

shadow_signature <- function(shadow, pattern) {
  lines <- vapply(shadow$trace, `[[`, character(1L), "line")
  events <- vapply(shadow$trace, `[[`, character(1L), "event")
  keep <- grepl(pattern, lines) & events == "render"

  data.frame(
    event = events[keep],
    line = lines[keep],
    stringsAsFactors = FALSE
  )
}

test_that("Landweber npregiv single-line progress reports object labels with outer iterations", {
  single_line <- cached_landweber_single_line()

  lines <- vapply(single_line$trace, `[[`, character(1L), "line")

  expect_s3_class(single_line$value, "npregiv")
  expect_true(any(grepl("^\\[np\\] IV regression \\(E\\[y\\|w\\], elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] IV regression \\(E\\[[^)]*\\|z\\], elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] IV regression \\(E\\[y-phi\\(z\\)\\|w\\], iteration 1, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] IV regression \\(E\\[E\\[y-phi\\(z\\)\\|w\\]\\|z\\], iteration 1, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_false(any(grepl("Iterating Landweber-Fridman solve", lines, fixed = TRUE)))
  expect_false(any(grepl("%|eta ", lines)))
})

test_that("Tikhonov npregiv single-line progress restores historical object labels", {
  single_line <- cached_tikhonov_single_line()

  lines <- vapply(single_line$trace, `[[`, character(1L), "line")

  expect_s3_class(single_line$value, "npregiv")
  expect_true(any(grepl("^\\[np\\] IV regression \\(E\\[y\\|w\\], elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] IV regression \\(E\\[E\\[y\\|w\\]\\|z\\], elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] IV regression \\(alpha, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] IV regression \\(E\\[phi\\(z\\)\\|w\\], iteration [0-9]+, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] IV regression \\(E\\[E\\[phi\\(z\\)\\|w\\]\\|z\\], iteration [0-9]+, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] IV regression \\(phi\\(z\\), iteration [0-9]+, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_false(any(grepl("Iterating Tikhonov solve", lines, fixed = TRUE)))
  expect_false(any(grepl("%|eta ", lines)))
})

test_that("Landweber npregiv without residual smoothing reports the alternate historical objects", {
  single_line <- cached_landweber_no_residual_smoothing()

  lines <- vapply(single_line$trace, `[[`, character(1L), "line")

  expect_true(any(grepl("^\\[np\\] IV regression \\(E\\[phi\\(z\\)\\|w\\], iteration 1, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] IV regression \\(E\\[E\\[y\\|w\\]-E\\[phi\\(z\\)\\|w\\]\\|z\\], iteration 1, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
})

test_that("npregivderiv single-line progress matches legacy semantics", {
  parity <- cached_npregivderiv_parity()
  legacy <- parity$legacy
  single_line <- parity$single_line

  pattern <- "^\\[np\\] Iterating Landweber-Fridman derivative solve"
  lines <- shadow_lines_matching(single_line, pattern)

  expect_s3_class(single_line$value, "npregivderiv")
  expect_equal(shadow_signature(single_line, pattern), shadow_signature(legacy, pattern))
  expect_true(any(grepl("^\\[np\\] Preparing IV derivative regression$", single_line$messages)))
  expect_true(any(grepl("^\\[np\\] Iterating Landweber-Fridman derivative solve\\.\\.\\. iteration [0-9]+, elapsed [0-9]+\\.[0-9]s", lines)))
})

test_that("npregiv proof-slice progress respects np.messages FALSE", {
  dat <- make_iv_data()
  old_opts <- options(np.messages = FALSE)
  on.exit(options(old_opts), add = TRUE)

  res <- capture_progress_shadow_with_conditions(
    npregiv(y = dat$y, z = dat$z, w = dat$w, method = "Landweber-Fridman", iterate.max = 2),
    now = progress_time_counter()
  )

  expect_length(res$messages, 0)
  expect_length(res$trace, 0)
})

test_that("npregivderiv proof-slice progress respects suppressMessages", {
  dat <- make_iv_data()
  old_opts <- options(np.messages = TRUE)
  on.exit(options(old_opts), add = TRUE)

  res <- capture_progress_shadow_with_conditions(
    suppressMessages(npregivderiv(y = dat$y, z = dat$z, w = dat$w, iterate.max = 2)),
    now = progress_time_counter()
  )

  expect_length(res$messages, 0)
  expect_length(res$trace, 0)
})

test_that("npregiv basic functionality works", {
  set.seed(42)
  n <- 100
  # Simple IV setup
  z <- runif(n) # instrument
  w <- z + rnorm(n, sd=0.1) # endogenous regressor
  y <- w^2 + rnorm(n, sd=0.1) # outcome
  
  # This can be slow and sensitive to parameters
  # Use simple settings
  # Note: npregiv doesn't have a bandwidth.compute=FALSE in the same way, 
  # but we can try to keep it simple.
  
  # For the sake of a fast unit test, maybe we just check if it runs
  # without error with very small number of iterations if applicable,
  # but npregiv uses Tikhonov or Landweber-Fridman.
  
  # Let's try a very small example
  model <- npregiv(y=y, w=w, z=z, method="Landweber-Fridman", iterate.max=10)
  
  expect_s3_class(model, "npregiv")
  expect_true("phi" %in% names(model))
  expect_false(is.null(model$call))
})

test_that("npregivderiv basic functionality works", {
  set.seed(42)
  n <- 50
  z <- runif(n)
  w <- z + rnorm(n, sd=0.1)
  y <- z^2 + rnorm(n, sd=0.1)
  
  # Just check if it runs
  model <- npregivderiv(y=y, z=z, w=w, iterate.max=2)
  expect_s3_class(model, "npregivderiv")
  expect_true("phi.prime" %in% names(model))
  expect_false(is.null(model$call))
})

library(testthat)
library(np)

np_check_filter <- Sys.getenv("NP_CHECK_FILTER", unset = "check-core-smoke")
np_check_full <- identical(Sys.getenv("NP_CHECK_FULL", unset = ""), "1")

if (isTRUE(np_check_full) || identical(np_check_filter, "")) {
  test_check("np")
} else {
  test_check("np", filter = np_check_filter)
}

condbandwidth <-
  function(xbw,
           ybw,
           bwmethod = c("cv.ls","normal-reference", "manual"),
           bwscaling = FALSE,
           bwtype = c("fixed","generalized_nn","adaptive_nn"),
           cxkertype = c("gaussian","truncated gaussian","epanechnikov","uniform"), 
           cxkerorder = c(2,4,6,8),
           cxkerbound = c("none","range","fixed"),
           cxkerlb = NULL,
           cxkerub = NULL,
           uxkertype = c("aitchisonaitken","liracine"),
           oxkertype = c("liracine","wangvanryzin","racineliyan"),
           cykertype = c("gaussian","truncated gaussian","epanechnikov","uniform"), 
           cykerorder = c(2,4,6,8),
           cykerbound = c("none","range","fixed"),
           cykerlb = NULL,
           cykerub = NULL,
           uykertype = c("aitchisonaitken","liracine"),
           oykertype = c("liracine","wangvanryzin","racineliyan"),
           fval = NA,
           ifval = NA,
           num.feval = NA,
           num.feval.fast = NA,
           fval.history = NA,
           eval.history = NA,
           invalid.history = NA,
           nobs = NA,
           xdati, ydati,
           xnames = character(length(xbw)),
           ynames = character(length(ybw)),
           sfactor = NA, bandwidth = NA,
           rows.omit = NA,
           nconfac = NA,
           ncatfac = NA,
           sdev = NA,
           bandwidth.compute = TRUE,
           timing = NA,
           total.time = NA,
           regtype = "lc",
           pregtype = "Local-Constant",
           basis = "glp",
           degree = integer(0),
           bernstein.basis = FALSE,
           regtype.engine = "lc",
           basis.engine = "glp",
           degree.engine = integer(0),
           bernstein.basis.engine = FALSE,
           ...){

  if (missing(xbw) || missing(ybw))
    stop("improper invocation of condbandwidth constructor: 'bw' or i[cuo]* missing")
  
  xndim = length(xbw)
  yndim = length(ybw)
  
  bwmethod = match.arg(bwmethod)
  bwtype = match.arg(bwtype)

  cxkertype = match.arg(cxkertype)
  cykertype = match.arg(cykertype)
  cxkerbound = match.arg(cxkerbound)
  cykerbound = match.arg(cykerbound)

  if(missing(cxkerorder))
    cxkerorder = 2
  else if (cxkertype == "uniform")
    .np_warning("ignoring kernel order specified with uniform kernel type")
  else {
    kord = c(2,4,6,8) 
    if (!any(kord == cxkerorder))
      stop("cxkerorder must be one of ", paste(kord,collapse=" "))
  }

  if (cxkertype == "truncated gaussian" && cxkerorder != 2)
    .np_warning("using truncated gaussian of order 2, higher orders not yet implemented")

  if (bwmethod == "normal-reference" && (cxkertype != "gaussian" || bwtype != "fixed")){    
    .np_warning("normal-reference bandwidth selection assumes gaussian kernel with fixed bandwidth")
    bwtype = "fixed"
    cxkertype = "gaussian"
  }

  if(missing(cykerorder))
    cykerorder = 2
  else if (cykertype == "uniform")
    .np_warning("ignoring kernel order specified with uniform kernel type")
  else {
    kord = c(2,4,6,8) 
    if (!any(kord == cykerorder))
      stop("cykerorder must be one of ", paste(kord,collapse=" "))
  }

  if (cykertype == "truncated gaussian" && cykerorder != 2)
    .np_warning("using truncated gaussian of order 2, higher orders not yet implemented")

  if (bwmethod == "normal-reference" && (cykertype != "gaussian" || bwtype != "fixed")){    
    .np_warning("normal-reference bandwidth selection assumes gaussian kernel with fixed bandwidth")
    bwtype = "fixed"
    cykertype = "gaussian"
  }

  if (cxkerorder != cykerorder && bwscaling)
    stop("scale factors with different order kernels for dependent and explanatory variables is unsupported")
  
  uxkertype = match.arg(uxkertype)
  uykertype = match.arg(uykertype)
  
  oxkertype = match.arg(oxkertype)
  oykertype = match.arg(oykertype)
  cxbounds <- npKernelBoundsResolve(
    dati = xdati,
    varnames = xnames,
    kerbound = cxkerbound,
    kerlb = cxkerlb,
    kerub = cxkerub,
    argprefix = "cxker")
  cybounds <- npKernelBoundsResolve(
    dati = ydati,
    varnames = ynames,
    kerbound = cykerbound,
    kerlb = cykerlb,
    kerub = cykerub,
    argprefix = "cyker")
  bounded_nonfixed_supported <- bwtype %in% c("generalized_nn", "adaptive_nn")
  if (bwtype != "fixed" &&
      (cxbounds$bound != "none" || cybounds$bound != "none") &&
      !bounded_nonfixed_supported)
    stop("finite continuous kernel bounds require bwtype = \"fixed\"")

  pxorder = switch( cxkerorder/2, "Second-Order", "Fourth-Order", "Sixth-Order", "Eighth-Order" )
  pyorder = switch( cykerorder/2, "Second-Order", "Fourth-Order", "Sixth-Order", "Eighth-Order" )

  dati <- list(x = xdati, y = ydati)
  
  if (!identical(sfactor,NA)){
    ## using the new model for generically accessing bandwidth objects

    okertype <- list(x = oxkertype, y = oykertype)
    ukertype <- list(x = uxkertype, y = uykertype)

    scaleOrMax <- function(i, j) {
      if (dati[[j]]$icon[i])
        return((sfactor[[j]])[i])

      if (dati[[j]]$iord[i])
        return(oMaxL(dati[[j]]$all.nlev[[i]], kertype = okertype[[j]]))
      
      if (dati[[j]]$iuno[i])
        return(uMaxL(dati[[j]]$all.nlev[[i]], kertype = ukertype[[j]]))
    }

    sumNum <- list(x = NA, y = NA)
    sumNum[] <- lapply(seq_along(dati), function(i) {
      sapply(seq_along(dati[[i]]$icon), scaleOrMax, j = i)
    })
  } else {
    sumNum <- NA
  }

  if (length(rows.omit) == 0)
    rows.omit <- NA

  mybw = list(
    xbw=xbw,
    ybw=ybw,
    method = bwmethod,
    pmethod = bwmToPrint(bwmethod),
    fval = fval,
    ifval = ifval,
    num.feval = num.feval,
    num.feval.fast = num.feval.fast,
    fval.history = fval.history,
    eval.history = eval.history,
    invalid.history = invalid.history,
    scaling = bwscaling,
    pscaling = npBandwidthSummaryLabel(bwtype = bwtype, bwscaling = bwscaling),
    type = bwtype,
    ptype = bwtToPrint(bwtype),
    cxkertype = cxkertype,
    cykertype = cykertype,
    cxkerorder = cxkerorder,
    cykerorder = cykerorder,
    cxkerbound = cxbounds$bound,
    cxkerlb = cxbounds$lb,
    cxkerub = cxbounds$ub,
    cykerbound = cybounds$bound,
    cykerlb = cybounds$lb,
    cykerub = cybounds$ub,
    pcxkertype = cktToPrint(cxkertype, order = pxorder, kerbound = cxbounds$bound),
    pcykertype = cktToPrint(cykertype, order = pyorder, kerbound = cybounds$bound),
    uxkertype = uxkertype,
    uykertype = uykertype,
    puxkertype = uktToPrint(uxkertype),
    puykertype = uktToPrint(uykertype),
    oxkertype = oxkertype,
    oykertype = oykertype,
    poxkertype = oktToPrint(oxkertype),
    poykertype = oktToPrint(oykertype, normalized = TRUE),
    nobs = nobs,
    xndim = xndim,
    yndim = yndim,
    ndim = xndim + yndim,
    xncon = sum(xdati$icon),
    xnuno = sum(xdati$iuno),
    xnord = sum(xdati$iord),
    yncon = sum(ydati$icon),
    ynuno = sum(ydati$iuno),
    ynord = sum(ydati$iord),
    ncon = sum(c(xdati$icon, ydati$icon)),
    ixcon = xdati$icon,
    ixuno = xdati$iuno,
    ixord = xdati$iord,
    iycon = ydati$icon,
    iyuno = ydati$iuno,
    iyord = ydati$iord,
    xnames = xnames,
    ynames = ynames,
    xdati = xdati,
    ydati = ydati,
    xmcv = mcvConstruct(xdati),
    ymcv = mcvConstruct(ydati),
    sfactor = sfactor,
    bandwidth = bandwidth,
    nconfac = nconfac,
    ncatfac = ncatfac,
    sdev = sdev,
    sumNum = sumNum,
    dati = dati, 
    varnames = list(x = xnames, y = ynames),
    vartitle = list(x = "Explanatory", y = "Dependent"),
    vartitleabb = list(x = "Exp.", y = "Dep."),
    rows.omit = rows.omit,
    nobs.omit = if (identical(rows.omit, NA)) 0 else length(rows.omit),
    timing = timing,
    total.time = total.time,
    regtype = regtype,
    pregtype = pregtype,
    basis = basis,
    degree = degree,
    bernstein.basis = bernstein.basis,
    regtype.engine = regtype.engine,
    basis.engine = basis.engine,
    degree.engine = degree.engine,
    bernstein.basis.engine = bernstein.basis.engine)

  mybw$klist = list(
    x =
    list(ckertype = cxkertype,
         ckerbound = cxbounds$bound,
         ckerlb = cxbounds$lb,
         ckerub = cxbounds$ub,
         pckertype = mybw$pcxkertype,
         ukertype = uxkertype,
         pukertype = mybw$puxkertype,
         okertype = oxkertype,
         pokertype = mybw$poxkertype),
    y =
    list(ckertype = cykertype,
         ckerbound = cybounds$bound,
         ckerlb = cybounds$lb,
         ckerub = cybounds$ub,
         pckertype = mybw$pcykertype,
         ukertype = uykertype,
         pukertype = mybw$puykertype,
         okertype = oykertype,
         pokertype = mybw$poykertype))

  if(!bandwidth.compute)
    mybw$pmethod <- "Manual"


  class(mybw) = "condbandwidth"
  if(!any(is.na(mybw$bandwidth)))
    validateBandwidth(mybw)
  mybw
}

print.condbandwidth <- function(x, digits=NULL, ...){
  cat("\nConditional distribution data (",x$nobs," observations, ",
      (x$xndim+x$yndim)," variable(s))",
      "\n(", x$yndim, " dependent variable(s), and ", x$xndim, " explanatory variable(s))\n\n",
      sep="")
  print(matrix(x$ybw,ncol=x$yndim,dimnames=list(paste("Dep. Var. ",x$pscaling,":",sep=""),x$ynames)))

  print(matrix(x$xbw,ncol=x$xndim,dimnames=list(paste("Exp. Var. ",x$pscaling,":",sep=""),x$xnames)))

  cat(genBwSelStr(x))
  cat(genBwKerStrsXY(x))
  
  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}


summary.condbandwidth <- function(object, ...) {
  cat("\nConditional distribution data (",object$nobs," observations, ",
      (object$xndim+object$yndim)," variable(s))",
      "\n(", object$yndim, " dependent variable(s), and ", object$xndim, " explanatory variable(s))\n",
      sep="")

  cat(genOmitStr(object))
  cat(genBwSelStr(object))

  cat(paste("\n", genBwScaleStrs(object), sep=""))
  cat(genBwKerStrs(object))

  cat(genTimingStr(object))
  
  cat("\n\n")
}

predict.condbandwidth <- function(object, ...) { npcdist(bws = object, ...) }

 .np_scbandwidth_manual_nn_validate <- function(bw, nobs, where = "scbandwidth") {
  vapply(bw, function(h) {
    .np_sibandwidth_manual_nn_validate(h = h, nobs = nobs, where = where)
  }, numeric(1))
}

scbandwidth <-
  function(bw = stop("scbandwidth:argument 'bw' missing"),
           regtype = c("lc","ll","lp"),
           basis = c("glp","additive","tensor"),
           degree = NULL,
           bernstein.basis = FALSE,
           bwmethod = c("cv.ls", "manual"),
           bwscaling = FALSE,
           bwtype = c("fixed","generalized_nn","adaptive_nn"),
           ckertype = c("gaussian","truncated gaussian","epanechnikov","uniform"), 
           ckerorder = c(2,4,6,8),
           ckerbound = c("none","range","fixed"),
           ckerlb = NULL,
           ckerub = NULL,
           ukertype = c("aitchisonaitken", "liracine"),
           okertype = c("liracine","wangvanryzin","racineliyan"),
           fval = NA,
           ifval = NA,
           num.feval = NA,
           num.feval.fast = NA,
           nobs = NA,
           numimp = NA,
           fval.vector = NA,
           xdati = stop("scbandwidth:argument 'xdati' missing"),
           ydati = stop("scbandwidth:argument 'ydati' missing"),
           zdati = NULL,
           xnames = character(length(bw)),
           ynames = character(1),
           znames = NULL,
           sfactor = NA, bandwidth = NA,
           rows.omit = NA,
           nconfac = NA,
           ncatfac = NA,
           sdev = NA,
           bandwidth.compute = TRUE,
           optim.method = "NA",
           total.time = NA,
           ...){

  ndim = length(bw)
  npRejectLegacyLpArgs(names(list(...)), where = "scbandwidth")
  regtype = match.arg(regtype)
  basis <- npValidateLpBasis(regtype = regtype, basis = basis)
  bwmethod = match.arg(bwmethod)
  bwtype = match.arg(bwtype)
  ckertype = match.arg(ckertype)
  ckerbound = match.arg(ckerbound)

  if(missing(ckerorder))
    ckerorder = 2
  else if (ckertype == "uniform")
    .np_warning("ignoring kernel order specified with uniform kernel type")
  else {
    kord = c(2,4,6,8) 
    if (!any(kord == ckerorder))
      stop("ckerorder must be one of ", paste(kord,collapse=" "))
  }

  if (ckertype == "truncated gaussian" && ckerorder != 2)
    .np_warning("using truncated gaussian of order 2, higher orders not yet implemented")

  ukertype = match.arg(ukertype)
  okertype = match.arg(okertype)

  tdati <- zdati

  if(is.null(zdati))
    tdati <- xdati

  cbounds <- npKernelBoundsResolve(
    dati = tdati,
    varnames = if(is.null(znames)) xnames else znames,
    kerbound = ckerbound,
    kerlb = ckerlb,
    kerub = ckerub,
    argprefix = "cker")
  bounded_nonfixed_supported <- bwtype %in% c("generalized_nn", "adaptive_nn")
  if (bwtype != "fixed" && cbounds$bound != "none" && !bounded_nonfixed_supported)
    stop("finite continuous kernel bounds require bwtype = \"fixed\"")
  if (bwtype != "fixed" && (!bandwidth.compute || any(bw != 0)))
    bw <- .np_scbandwidth_manual_nn_validate(bw = bw, nobs = nobs, where = "scbandwidth")
  ncon <- sum(tdati$icon)
  degree <- npValidateGlpDegree(regtype = regtype,
                                degree = degree,
                                ncon = ncon)
  bernstein.basis <- npValidateGlpBernstein(regtype = regtype,
                                            bernstein.basis = bernstein.basis)
  if (identical(regtype, "lp") && ncon > 0L && is.finite(nobs)) {
    lp.dim <- dim_basis(basis = basis,
                        kernel = TRUE,
                        degree = degree,
                        segments = rep.int(1L, ncon))
    if (is.finite(lp.dim) && lp.dim > (nobs - 1.0))
      stop(sprintf("LP basis dimension (%s) exceeds nobs - 1 (%s); reduce degree",
                   format(lp.dim, trim = TRUE, scientific = FALSE),
                   format(nobs - 1.0, trim = TRUE, scientific = FALSE)))
  }

  porder = switch( ckerorder/2, "Second-Order", "Fourth-Order", "Sixth-Order", "Eighth-Order" )

  if (!identical(sfactor,NA)){
    sumNum <- sapply(seq_len(ndim), function(i) {
      if (tdati$icon[i])
        return(sfactor[i])

      if (tdati$iord[i])
        return(oMaxL(tdati$all.nlev[[i]], kertype = okertype))
      
      if (tdati$iuno[i])
        return(uMaxL(tdati$all.nlev[[i]], kertype = ukertype))
    })
  } else {
    sumNum <- NA
  }

  if (length(rows.omit) == 0)
    rows.omit <- NA

  
  mybw = list(
    bw=bw,
    regtype = regtype,
    pregtype = switch(regtype,
      lc = "Local-Constant",
      ll = "Local-Linear",
      lp = "Local-Polynomial"),
    basis = basis,
    degree = degree,
    bernstein.basis = bernstein.basis,
    method = bwmethod,
    pmethod = bwmToPrint(bwmethod),
    pomethod = switch(optim.method,
      "Nelder-Mead" = "Nelder-Mead",
      "BFGS" = "BFGS",
      "CG" = "CG", "NA"),
    fval = fval,
    ifval = ifval,
    num.feval = num.feval,
    num.feval.fast = num.feval.fast,
    scaling = bwscaling,
    pscaling = npBandwidthSummaryLabel(bwtype = bwtype, bwscaling = bwscaling),
    type = bwtype,
    ptype = bwtToPrint(bwtype),
    ckertype = ckertype,    
    ckerorder = ckerorder,
    ckerbound = cbounds$bound,
    ckerlb = cbounds$lb,
    ckerub = cbounds$ub,
    pckertype = cktToPrint(ckertype, order = porder, kerbound = cbounds$bound),
    ukertype = ukertype,
    pukertype = uktToPrint(ukertype),
    okertype = okertype,
    pokertype = oktToPrint(okertype),
    nobs = nobs,
    ndim = ndim,
    ncon = ncon,
    nuno = sum(tdati$iuno),
    nord = sum(tdati$iord),
    icon = tdati$icon,
    iuno = tdati$iuno,
    iord = tdati$iord,
    xnames = xnames,
    ynames = ynames,
    znames = znames,
    xdati = xdati,
    ydati = ydati,
    zdati = zdati,
    xmcv = mcvConstruct(xdati),
    sfactor = list(sfactor),
    bandwidth = list(bandwidth),
    sumNum = list(sumNum),
    dati = list(x = xdati, y = ydati, z = zdati),
    varnames = list(x = xnames, y = ynames, z = znames),
    vartitle = list(x = "Explanatory", y = "Dependent", z = "Explanatory"),
    vartitleabb = list(x = "Exp.", y = "Dep.", z = "Exp."),
    rows.omit = rows.omit,
    nobs.omit = if (identical(rows.omit, NA)) 0 else length(rows.omit),
    total.time = total.time)

  mybw$klist <-
      list(list(ckertype = ckertype,
                pckertype = mybw$pckertype,
                ukertype = ukertype,
                pukertype = mybw$pukertype,
                okertype = okertype,
                pokertype = mybw$pokertype))

  zorx <- if (is.null(zdati)) "x" else "z"
  names(mybw$sfactor) <- zorx
  names(mybw$bandwidth) <- zorx 
  names(mybw$sumNum) <- zorx
  names(mybw$klist) <- zorx
  
  if(!bandwidth.compute)
    mybw$pmethod <- "Manual"

  class(mybw) = "scbandwidth"
  if(!any(is.na(mybw$bandwidth)))
    validateBandwidth(mybw)
  mybw
  
}

as.double.scbandwidth <- function(x, ...){
  x$bw
}

print.scbandwidth <- function(x, digits=NULL, ...){
  cat("\nSmooth Coefficient Model",
      "\nRegression Data (",x$nobs," observations, ",x$ndim," variable(s)):\n\n",sep="")
  print(matrix(x$bw,ncol=x$ndim,dimnames=list(paste(x$pscaling,":",sep=""),
                                  if(is.null(x$znames)) x$xnames else x$znames)))

  cat(genBwSelStr(x))
  cat(genBwKerStrs(x))
  
  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

predict.scbandwidth <- function(object, ...) { npscoef(bws = object, ...) }

summary.scbandwidth <- function(object, ...){
  cat("\nSmooth Coefficient Regression",
      "\nRegression Data (",object$nobs," observations, ",object$ndim," variable(s)):\n",sep="")

  cat("\nOptimisation Method: ", object$pomethod)

  cat(genOmitStr(object))
  cat(genBwSelStr(object))

  cat('\n')
  cat(genBwScaleStrs(object))

  cat(genBwKerStrs(object))
  cat(genTimingStr(object))
  cat("\n\n")

}

npuniden.sc <- function(X=NULL,
                        Y=NULL,
                        h=NULL,
                        a=0,
                        b=1,
                        lb=NULL,
                        ub=NULL,
                        extend.range=0,
                        num.grid=0,
                        function.distance=TRUE,
                        integral.equal=FALSE,
                        constraint=c("density",
                            "mono.incr",
                            "mono.decr",
                            "concave",
                            "convex",
                            "log-concave",
                            "log-convex")) {

    constraint <- match.arg(constraint)

    ## Gaussian1 kernel function, derivatives up to order two

    kernel <- function(x,X,h,a=0,b=1,deriv=0) {

        if(deriv < 0 || deriv > 2) stop("deriv must be one of 0, 1, or 2")
        
        ## Note: the different sign/position of x in z, z.a, and z.b
        ## changes the sign of the derivative from 1/h to -1/h
        
        z <- (x-X)/h
        z.a <- (a-x)/h
        z.b <- (b-x)/h
        
        K.z.a <- dnorm(z.a)
        K.z.b <- dnorm(z.b)
        
        G.z.a <- pnorm(z.a)
        G.z.b <- pnorm(z.b)
        
        G.z.b.m.G.z.a <- G.z.b-G.z.a
        
        G.z.a.d1 <- -K.z.a/h
        G.z.b.d1 <- -K.z.b/h
        
        G.z.b.d1.m.G.z.a.d1 <- G.z.b.d1-G.z.a.d1
        
        G.z.a.d2 <- -z.a*K.z.a/h^2
        G.z.b.d2 <- -z.b*K.z.b/h^2
        
        G.z.b.d2.m.G.z.a.d2 <- G.z.b.d2-G.z.a.d2
        
        K.z <- dnorm(z)
        K.z.d1 <- -z*K.z/h
        K.z.d2 <- (z^2-1)*K.z/h^2

        ## If the above change due to the use of different base kernel
        ## functions, the formulae below remain unaffected
        
        if(deriv==0) {
            K <- K.z/G.z.b.m.G.z.a
            return(K/h)
        } else if(deriv==1) {
            K <- K.z.d1/G.z.b.m.G.z.a - K.z*G.z.b.d1.m.G.z.a.d1/G.z.b.m.G.z.a^2
            return(K/h)        
        } else if(deriv==2) {
            K <- K.z.d2/G.z.b.m.G.z.a -
                (2*K.z.d1*G.z.b.d1.m.G.z.a.d1+K.z*G.z.b.d2.m.G.z.a.d2)/G.z.b.m.G.z.a^2 +
                    2*K.z*G.z.b.d1.m.G.z.a.d1^2/G.z.b.m.G.z.a^3
            return(K/h)
        }
    }
    
    int.kernel.squared <- function(X,h,a=a,b=b) {
        ## Use numeric integration to compute Kappa, the integral of
        ## the square of the kernel function needed for the asymptotic
        ## standard error of the density estimate seq(a,b) will barf
        ## on -Inf or Inf, trap these cases and use extendrange
        if(is.finite(a) && is.finite(b)) X.seq <- seq(a,b,length=1000)
        if(is.finite(a) && !is.finite(b)) X.seq <- seq(a,extendrange(X,f=10)[2],length=1000)
        if(!is.finite(a) && is.finite(b)) X.seq <- seq(extendrange(X,f=10)[1],b,length=1000)
        if(!is.finite(a) && !is.finite(b)) X.seq <- seq(extendrange(X,f=10)[1],extendrange(X,f=10)[2],length=1000)
        sapply(seq_along(X), function(i){integrate.trapezoidal(X.seq,h*kernel(X[i],X.seq,h,a,b,deriv=0)^2)[length(X.seq)]})
    }

    W.kernel <- function(x,X,h,a=0,b=1,deriv=0) {
        sapply(seq_along(x), function(i){kernel(x[i],X,h,a,b,deriv)})
    }
    
    if(is.null(X)) stop("you must pass a vector X")
    if(anyNA(X)) stop("X must not contain missing values")
    if(a>=b) stop("a must be less than b")
    if(any(X<a, na.rm = TRUE)) stop("X must be >= a")
    if(any(X>b, na.rm = TRUE)) stop("X must be <= b")
    if(!is.null(Y) && anyNA(Y)) stop("Y must not contain missing values")
    if(!is.null(Y) && any(Y<a, na.rm = TRUE)) stop("Y must be >= a")
    if(!is.null(Y) && any(Y>b, na.rm = TRUE)) stop("Y must be <= b")
    if(is.null(h)) stop("you must provide a bandwidth")
    if(h <= 0) stop("bandwidth h must be positive")
    if(num.grid < 0) stop("num.grid must be a non-negative integer")
    if(constraint=="density" && is.null(lb) && is.null(ub)) stop("you must provide lower and/or upper bounds when constraining the density")
    if(!is.null(lb) && anyNA(lb)) stop("lower bound must not contain missing values")
    if(!is.null(lb) && any(lb<0, na.rm = TRUE)) stop("lower bound must be non-negative")
    if(!is.null(ub) && anyNA(ub)) stop("upper bound must not contain missing values")
    if(!is.null(ub) && any(ub<0, na.rm = TRUE)) stop("upper bound must be non-negative")
    if(!is.null(lb) && !is.null(ub) && any(ub<lb, na.rm = TRUE)) stop("upper bound must be greater than or equal to lower bound")

    ## First elements are X if Y=NULL or Y, rest are to make sure
    ## constraints are imposed on the bulk of the support and a bit
    ## outside governed by f=extend.range

    if(num.grid==0) {
        grid <- NULL
    } else {
        grid <- seq(max(c(a,extendrange(X,f=extend.range)[1])),min(c(b,extendrange(X,f=extend.range)[2])),,num.grid)
    }

    X.grid <- c(Y,X,grid)
    
    ## We always need the unconstrained weight matrix and estimate

    A <- W.kernel(X.grid,X,h,a=a,b=b,deriv=0)
    f <- colMeans(A)

    ## First or second order derivative needed, place them both in *.deriv

    deriv <- 0
    if(constraint=="mono.decr" || constraint=="mono.incr") deriv <- 1
    if(constraint=="concave" || constraint=="convex" || constraint=="log-concave" || constraint=="log-convex") deriv <- 2

    ## Second derivative uses z.a and z.b times other arguments, which
    ## must be finite, so set to very small and large values depending
    ## on the range of the training data
    
    if(deriv==2 && a==-Inf) a <- extendrange(X,f=100)[1]
    if(deriv==2 && b==Inf) b <- extendrange(X,f=100)[2]

    ## The derivative for the log-density estimate differs from that
    ## for the raw density, and also requires the first derivative

    A.deriv <- W.kernel(X.grid,X,h,a=a,b=b,deriv=deriv)
    if(constraint=="log-concave" || constraint=="log-convex") {
        f.deriv <- (colMeans(A.deriv)*f - colMeans(W.kernel(X.grid,X,h,a=a,b=b,deriv=1))^2)/(f^2)
    } else {
        f.deriv <- colMeans(A.deriv)
    }

    ## The sign of the constraint matrix Amat and bvec switch
    ## depending on whether the constraint is >= or <=, so automate
    ## this
    
    sign.deriv <- 1
    if(constraint=="mono.decr" || constraint=="concave" || constraint=="log-concave") sign.deriv <- -1

    ## Length of the grid vector and training data vector
    
    n.grid <- length(X.grid)
    n.train <- length(X)

    ## Solve the quadratic program

    solve.QP.flag <- TRUE
    output.QP <- NULL
    constant <- c(1,1/10,10,1/100,100,1/1000,1000,1/10000,10000,1/100000,100000)
    attempts.max <- length(constant)
    attempts <- 1
    while((is.null(output.QP) || any(is.na(output.QP$solution))) && attempts <= attempts.max) {
        if(attempts==attempts.max && !function.distance) {
            .np_warning("solve.QP was unable to find a solution with function.distance=FALSE, restarting with function.distance=TRUE", immediate. = TRUE)
            attempts <- 1
            function.distance <- TRUE
        }
        if(function.distance) {
            ## Non-identity forcing matrix minimizes the squared
            ## function difference distance
            Dmat <- (A%*%t(A)+sqrt(.Machine$double.eps)*diag(n.train))/constant[attempts]
        } else {
            ## Identity forcing matrix minimizes the squared weight
            ## distance
            Dmat <- diag(n.train)/constant[attempts]
        }
        ## The unconstrained weight vector contains zeros
        dvec <- rep(0,n.train)
        if(constraint=="log-concave" || constraint=="log-convex") {
            ## The log-concave/convex estimate will be non-negative,
            ## no need to impose this constraint (will be slack)
            Amat <- cbind(rep(1,n.train),sign.deriv*A.deriv)
            bvec <- c(0,-sign.deriv*f.deriv)
        } else if(constraint!="density") {
            ## For other constraints we need to enforce non-negativity
            ## of the estimate - in either case, the unconstrained
            ## weights are zero hence the "sum to zero" constraint
            Amat <- cbind(rep(1,n.train),A,sign.deriv*A.deriv)
            bvec <- c(0,-f,-sign.deriv*f.deriv)
        } else {
            ## Constrain the density
            Amat <- cbind(rep(1,n.train),A,-A)
            bvec <- c(0,lb-f,f-ub)
        }
        output.QP <- tryCatch(solve.QP(Dmat=Dmat,
                                       dvec=dvec,
                                       Amat=Amat,
                                       bvec=bvec,
                                       meq=1),
                              error = function(e) NULL)
        ## Sometimes rescaling Dmat can overcome deficiencies in
        ## solve.QP
        attempts <- attempts+1
    }

    ## If solve.QP cannot find a solution issue an immediate warning
    ## but return the unconstrained vector

    if(is.null(output.QP) || any(is.na(output.QP$solution))) {
        .np_warning("solve.QP was unable to find a solution, unconstrained estimate returned", immediate. = TRUE)
        output.QP$solution <- rep(0,n.train)
        solve.QP.flag <- FALSE
    }
    
    if(constraint=="log-concave" || constraint=="log-convex") {
        f.sc <- as.numeric(exp(log(f)+t(A)%*%output.QP$solution))
        ## Constrained derivatives, i.e., derivatives of log(f), not f
        ## - if you wish derivatives of f uncomment the following
        ## line
        ## f.deriv <- colMeans(A.deriv)
        f.sc.deriv <- f.deriv+t(A.deriv)%*%output.QP$solution
    } else {
        f.sc <- as.numeric(f+t(A)%*%output.QP$solution)
        f.sc.deriv <- as.numeric(f.deriv+t(A.deriv)%*%output.QP$solution)
    }

    ## If the option integral.equal=TRUE is set, adjust the
    ## constrained estimate to have the same integral as the
    ## unconstrained estimate

    corr.factor <- 1
    int.f.sc <- integrate.trapezoidal(X.grid,f.sc)[n.grid]
    int.f <- integrate.trapezoidal(X.grid,f)[n.grid]
    if(integral.equal) corr.factor <- int.f.sc/int.f
    f.sc <- f.sc/corr.factor

    ## The constraints are imposed on X.grid, but we only want the
    ## constrained estimate at the sample points X (or evaluation
    ## points Y if Y is specified)

    if(is.null(Y)) {
        index <- seq_len(n.train)
    } else {
        index <- seq_along(Y)
    }

    se.f <- sqrt(abs(f*int.kernel.squared(X.grid,h,a,b)/(h*length(f))))[index]
    se.f.sc <- sqrt(abs(f.sc*int.kernel.squared(X.grid,h,a,b)/(h*length(f.sc))))[index]    

    F <- integrate.trapezoidal(X.grid[index],f[index])
    F.sc <- integrate.trapezoidal(X.grid[index],f.sc[index])
    f <- f[index]
    f.sc <- f.sc[index]
    f.deriv <- f.deriv[index]
    f.sc.deriv <- f.sc.deriv[index]

    ## Return a list
    
    return(list(f=f,
                f.sc=f.sc,
                se.f=se.f,
                se.f.sc=se.f.sc,
                f.deriv=f.deriv,
                f.sc.deriv=f.sc.deriv,
                F=F,
                F.sc=F.sc,
                se.F=sqrt(abs(F*(1-F)/length(F))),
                se.F.sc=sqrt(abs(F.sc*(1-F.sc)/length(F.sc))),                
                f.integral=int.f,
                f.sc.integral=int.f.sc,
                solve.QP=solve.QP.flag,
                attempts=attempts))

}

npudistbw <- function(...){
  mc <- match.call(expand.dots = FALSE)
  npRejectRenamedScaleFactorSearchArgs(names(mc$...), where = "npudistbw")
  target <- .np_bw_dispatch_target(dots = mc$...,
                                   data_arg_names = c("dat", "gdat"),
                                   eval_env = parent.frame())
  UseMethod("npudistbw", target)
}

npudistbw.formula <-
  function(formula, data, subset, na.action, call, gdata = NULL, ...){
    formula.terms <- terms(formula)
    orig.ts <- if (missing(data))
      .np_terms_ts_mask(terms_obj = formula.terms,
                        data = environment(formula),
                        eval_env = environment(formula))
    else .np_terms_ts_mask(terms_obj = formula.terms,
                           data = data,
                           eval_env = environment(formula))

    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(mf), nomatch = 0)
    mf <- mf[c(1,m)]

    if(all(orig.ts)){
      args <- (as.list(attr(formula.terms, "variables"))[-1])
      formula <- formula.terms
      attr(formula, "predvars") <- as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), args))))
      mf[["formula"]] <- formula
    }else if(any(orig.ts)){
      arguments <- (as.list(attr(formula.terms, "variables"))[-1])
      arguments.normal <- arguments[which(!orig.ts)]
      arguments.timeseries <- arguments[which(orig.ts)]

      ix <- sort(c(which(orig.ts),which(!orig.ts)),index.return = TRUE)$ix
      formula <- formula.terms
      attr(formula, "predvars") <- bquote(.(as.call(c(quote(cbind),as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments.timeseries)))),arguments.normal,check.rows = TRUE)))[,.(ix)])
      mf[["formula"]] <- formula
    }
    
    mf[[1]] <- as.name("model.frame")
    mf.args <- as.list(mf[-1L])
    mf <- do.call(stats::model.frame, mf.args, envir = parent.frame())

    if (attr(attr(mf, "terms"), "response") != 0)
      stop("invalid distribution formula")
    
    dat <- mf[, attr(attr(mf, "terms"),"term.labels"), drop = FALSE]

    has.gval <- !is.null(gdata)
    if (has.gval) {
      gmf <- match.call(expand.dots = FALSE)
      gm <- match(c("formula", "gdata"), names(gmf), nomatch = 0)
      gmf <- gmf[c(1,gm)]
                     
      gmf[[1]] <- as.name("model.frame")
      names(gmf)[3] <- "data"
      gmf.args <- as.list(gmf[-1L])
      gmf <- do.call(stats::model.frame, gmf.args, envir = parent.frame())

      gdat <- gmf[, attr(attr(gmf, "terms"),"term.labels"), drop = FALSE]

    }

    bw.args <- list(dat = dat)
    if (has.gval)
      bw.args$gdat <- gdat
    tbw <- do.call(npudistbw, c(bw.args, list(...)))
    tbw$call <- match.call(expand.dots = FALSE)
    environment(tbw$call) <- parent.frame()
    tbw$formula <- formula
    tbw$terms <- attr(mf,"terms")
    tbw$rows.omit <- as.vector(attr(mf,"na.action"))
    tbw$nobs.omit <- length(tbw$rows.omit)
    tbw
  }


npudistbw.NULL <-
  function(dat = stop("invoked without input data 'dat'"),
           bws, ...){

    t.names <- NULL
    if(!is.data.frame(dat) && !is.matrix(dat))
      t.names <- deparse(substitute(dat))

    dat = toFrame(dat)
    
    if(!is.null(t.names))
      names(dat) <- t.names

    bws = double(dim(dat)[2])

    tbw <- npudistbw.default(dat = dat, bws = bws, ...)

    ## clean up (possible) inconsistencies due to recursion ...
    mc <- match.call(expand.dots = FALSE)
    environment(mc) <- parent.frame()
    tbw$call <- mc

    tbw
  }

npudistbw.dbandwidth <- 
  function(dat = stop("invoked without input data 'dat'"),
           bws,
           gdat = NULL,
           bandwidth.compute = TRUE,
           cfac.dir = 2.5*(3.0-sqrt(5)),
           scale.factor.init = 0.5,
           dfac.dir = 0.25*(3.0-sqrt(5)),
           dfac.init = 0.375,
           dfc.dir = 3,
           do.full.integral = FALSE,
           ftol = 1.490116e-07,
           scale.factor.init.upper = 2.0,
           hbd.dir = 1,
           hbd.init = 0.9,
           initc.dir = 1.0,
           initd.dir = 1.0,
           invalid.penalty = c("baseline","dbmax"),
           itmax = 10000,
           lbc.dir = 0.5,
           scale.factor.init.lower = 0.1,
           lbd.dir = 0.1,
           lbd.init = 0.1,
           memfac = 500.0,
           ngrid = 100,
           nmulti,
           penalty.multiplier = 10,
           powell.remin = TRUE,
           scale.init.categorical.sample = FALSE,
           scale.factor.search.lower = NULL,
           small = 1.490116e-05,
           tol = 1.490116e-04,
           transform.bounds = FALSE,
           ...){

    elapsed.start <- proc.time()[3]

    dat = toFrame(dat)
    bandwidth.compute <- npValidateScalarLogical(bandwidth.compute, "bandwidth.compute")
    remin <- npValidateScalarLogical(powell.remin, "powell.remin")
    do.full.integral <- npValidateScalarLogical(do.full.integral, "do.full.integral")
    scale.init.categorical.sample <-
      npValidateScalarLogical(scale.init.categorical.sample, "scale.init.categorical.sample")
    transform.bounds <- npValidateScalarLogical(transform.bounds, "transform.bounds")
    itmax <- npValidatePositiveInteger(itmax, "itmax")
    ngrid <- npValidatePositiveInteger(ngrid, "ngrid")
    ftol <- npValidatePositiveFiniteNumeric(ftol, "ftol")
    tol <- npValidatePositiveFiniteNumeric(tol, "tol")
    small <- npValidatePositiveFiniteNumeric(small, "small")
    memfac <- npValidatePositiveFiniteNumeric(memfac, "memfac")
    penalty.multiplier <- npValidatePositiveFiniteNumeric(penalty.multiplier, "penalty.multiplier")
    scale.factor.search.lower <- npResolveScaleFactorLowerBound(
      if (is.null(scale.factor.search.lower)) npGetScaleFactorSearchLower(bws) else scale.factor.search.lower
    )

    nofi <- missing(do.full.integral)
    nogi <- missing(ngrid)
    
    if (missing(nmulti)){
      nmulti <- npDefaultNmulti(dim(dat)[2])
    }
    nmulti <- npValidateNmulti(nmulti)
    .np_progress_bandwidth_set_total(nmulti)

    if (length(bws$bw) != dim(dat)[2])
      stop(paste("length of bandwidth vector does not match number of columns of",
           "'dat'"))

    if ((any(bws$icon) &&
         !all(vapply(as.data.frame(dat[, bws$icon]), inherits, logical(1), c("integer", "numeric")))) ||
        (any(bws$iord) &&
         !all(vapply(as.data.frame(dat[, bws$iord]), inherits, logical(1), "ordered"))) ||
        (any(bws$iuno) &&
         !all(vapply(as.data.frame(dat[, bws$iuno]), inherits, logical(1), "factor"))))
      stop(paste("supplied bandwidths do not match", "'dat'", "in type"))

    if(any(bws$iuno))
      stop("distribution bandwidth selection does not support unordered data types")

    dat <- na.omit(dat)
    rows.omit <- unclass(na.action(dat))

    nrow = dim(dat)[1]
    ncol = dim(dat)[2]

    ## at this stage, data to be sent to the c routines must be converted to
    ## numeric type.

    odat <- dat
    dat = toMatrix(dat)

    duno = dat[, bws$iuno, drop = FALSE]
    dcon = dat[, bws$icon, drop = FALSE]
    dord = dat[, bws$iord, drop = FALSE]

    tbw <- bws

    mysd <- EssDee(dcon)
    nconfac <- nrow^(-1.0/(1.0+bws$ckerorder))
    ncatfac <- nrow^(-2.0/(1.0+bws$ckerorder))

    ## these are the points where we evaluate the CDF
    ## for now we default to the training points (hence cdf_on_train = TRUE below)
    if(!is.null(gdat)){
      gdat <- toFrame(gdat)
      if(any(is.na(gdat)))
        stop("na's not allowed to be present in cdf gdata")
      gdat <- toMatrix(gdat)

      guno = gdat[, bws$iuno, drop = FALSE]
      gord = gdat[, bws$iord, drop = FALSE]
      gcon = gdat[, bws$icon, drop = FALSE]
      cdf_on_train = FALSE
      nog = nrow(gdat)
      
    } else {
      if(do.full.integral) {
        cdf_on_train = TRUE
        nog = 0
        guno = data.frame()
        gord = data.frame()
        gcon = data.frame()
      } else {
        cdf_on_train = FALSE
        nog = ngrid
        probs <- seq(0,1,length.out = nog)
        ev <- odat[seq_len(nog),,drop = FALSE]
        for (i in seq_len(ncol(ev))) {
          ev[,i] <- uocquantile(odat[,i], probs)
        }

        ev <- toMatrix(ev)
        
        guno = ev[, bws$iuno, drop = FALSE]
        gord = ev[, bws$iord, drop = FALSE]
        gcon = ev[, bws$icon, drop = FALSE]
      }
    }

    invalid.penalty <- match.arg(invalid.penalty)
    penalty_mode <- (if (invalid.penalty == "baseline") 1L else 0L)

    if (bandwidth.compute){
      cont.start <- npContinuousSearchStartControls(
        scale.factor.init.lower,
        scale.factor.init.upper,
        scale.factor.init,
        scale.factor.search.lower,
        where = "npudistbw"
      )
      myopti = list(num_obs_train = dim(dat)[1],
        num_obs_eval = nog,
        iMultistart = IMULTI_TRUE,
        iNum_Multistart = nmulti,
        int_use_starting_values = (if (all(bws$bw==0)) USE_START_NO else USE_START_YES),
        int_LARGE_SF = (if (bws$scaling) SF_NORMAL else SF_ARB),
        BANDWIDTH_den_extern = switch(bws$type,
          fixed = BW_FIXED,
          generalized_nn = BW_GEN_NN,
          adaptive_nn = BW_ADAP_NN),
        itmax=itmax, int_RESTART_FROM_MIN=(if (remin) RE_MIN_TRUE else RE_MIN_FALSE), 
        int_MINIMIZE_IO=if (isTRUE(getOption("np.messages"))) IO_MIN_FALSE else IO_MIN_TRUE, 
        bwmethod = switch(bws$method,
          cv.cdf = DBWM_CVLS),
        ckerneval = switch(bws$ckertype,
          gaussian = CKER_GAUSS + bws$ckerorder/2 - 1,
          epanechnikov = CKER_EPAN + bws$ckerorder/2 - 1,
          uniform = CKER_UNI,
          "truncated gaussian" = CKER_TGAUSS),
        ukerneval = switch(bws$ukertype,
          aitchisonaitken = UKER_AIT,
          liracine = UKER_LR),
        okerneval = switch(bws$okertype,
          wangvanryzin = OKER_WANG,
          liracine = OKER_NLR,
        "racineliyan" = OKER_RLY),
        cdf_on_train = cdf_on_train,
        nuno = dim(duno)[2],
        nord = dim(dord)[2],
        ncon = dim(dcon)[2],
        int_do_tree = if (isTRUE(getOption("np.tree"))) DO_TREE_YES else DO_TREE_NO,
        scale.init.categorical.sample = scale.init.categorical.sample,
        dfc.dir = dfc.dir,
        transform.bounds = transform.bounds)
      
      myoptd = list(ftol=ftol, tol=tol, small=small,
        lbc.dir = lbc.dir, cfac.dir = cfac.dir, initc.dir = initc.dir, 
        lbd.dir = lbd.dir, hbd.dir = hbd.dir, dfac.dir = dfac.dir, initd.dir = initd.dir, 
        lbc.init = cont.start$scale.factor.init.lower, hbc.init = cont.start$scale.factor.init.upper, cfac.init = cont.start$scale.factor.init, 
        lbd.init = lbd.init, hbd.init = hbd.init, dfac.init = dfac.init, 
        nconfac = nconfac, ncatfac = ncatfac, memfac = memfac,
        scale.factor.lower.bound = scale.factor.search.lower)
      cker.bounds.c <- npKernelBoundsMarshal(bws$ckerlb[bws$icon], bws$ckerub[bws$icon])

      if (bws$method != "normal-reference"){
        myout <-
          .Call("C_np_distribution_bw",
                as.double(duno), as.double(dord), as.double(dcon),
                as.double(guno), as.double(gord), as.double(gcon), as.double(mysd),
                as.integer(myopti), as.double(myoptd),
                as.double(c(bws$bw[bws$icon], bws$bw[bws$iuno], bws$bw[bws$iord])),
                as.integer(nmulti),
                as.integer(penalty_mode),
                as.double(penalty.multiplier),
                as.double(cker.bounds.c$lb),
                as.double(cker.bounds.c$ub),
                PACKAGE="np")
        total.time <- proc.time()[3] - elapsed.start
      } else {
        nbw = double(ncol)
        gbw = bws$ncon
        if (gbw > 0){
          gbw_idx <- seq_len(gbw)
          nbw[gbw_idx] = 1.587
          if(!bws$scaling)
            nbw[gbw_idx]=nbw[gbw_idx]*mysd*nconfac
        }
        myout= list( bw = nbw, fval = c(NA,NA) )
        total.time <- NA
      }

      rorder = numeric(ncol)
      ord_idx <- seq_len(ncol)
      rorder[c(ord_idx[bws$icon], ord_idx[bws$iuno], ord_idx[bws$iord])] <- ord_idx

      tbw$bw <- myout$bw[rorder]

      tbw$fval = myout$fval[1]
      tbw$ifval = myout$fval[2]
      tbw$num.feval <- sum(myout$eval.history[is.finite(myout$eval.history)])
      tbw$num.feval.fast <- myout$fast.history[1]
      tbw$fval.history <- myout$fval.history
      tbw$eval.history <- myout$eval.history
      tbw$invalid.history <- myout$invalid.history
      tbw$timing <- myout$timing
      tbw$total.time <- total.time
    }
    
    tbw$sfactor <- tbw$bandwidth <- tbw$bw
    
    if (tbw$nuno > 0){
      if(tbw$scaling){ 
        tbw$bandwidth[tbw$xdati$iuno] <- tbw$bandwidth[tbw$xdati$iuno]*ncatfac
      } else {
        tbw$sfactor[tbw$xdati$iuno] <- tbw$sfactor[tbw$xdati$iuno]/ncatfac
      }
    }
    
    if (tbw$nord > 0){
      if(tbw$scaling){
        tbw$bandwidth[tbw$xdati$iord] <- tbw$bandwidth[tbw$xdati$iord]*ncatfac
      } else {
        tbw$sfactor[tbw$xdati$iord] <- tbw$sfactor[tbw$xdati$iord]/ncatfac
      }
    }


    if (tbw$ncon > 0){
      dfactor <- mysd*nconfac

      if (tbw$scaling) {
        tbw$bandwidth[tbw$xdati$icon] <- tbw$bandwidth[tbw$xdati$icon]*dfactor
      } else {
        tbw$sfactor[tbw$xdati$icon] <- tbw$sfactor[tbw$xdati$icon]/dfactor
      }
    }

    tbw <- dbandwidth(bw = tbw$bw,
                      bwmethod = tbw$method,
                      bwscaling = tbw$scaling,
                      bwtype = tbw$type,
                      ckertype = tbw$ckertype,
                      ckerorder = tbw$ckerorder,
                      ckerbound = tbw$ckerbound,
                      ckerlb = tbw$ckerlb,
                      ckerub = tbw$ckerub,
                      ukertype = c("aitchisonaitken"),
                      okertype = tbw$okertype,
                      fval = tbw$fval,
                      ifval = tbw$ifval,
                      num.feval = tbw$num.feval,
                      num.feval.fast = tbw$num.feval.fast,
                      fval.history = tbw$fval.history,
                      eval.history = tbw$eval.history,
                      invalid.history = tbw$invalid.history,
                      nobs = tbw$nobs,
                      xdati = tbw$xdati,
                      xnames = tbw$xnames,
                      sfactor = tbw$sfactor,
                      bandwidth = tbw$bandwidth,
                      rows.omit = rows.omit,
                      nconfac = nconfac,
                      ncatfac = ncatfac,
                      sdev = mysd,
                      bandwidth.compute = bandwidth.compute,
                      timing = tbw$timing,
                      total.time = tbw$total.time)
    tbw <- npSetScaleFactorSearchLower(tbw, scale.factor.search.lower)
    
    tbw
  }

npudistbw.default <-
  function(dat = stop("invoked without input data 'dat'"),
           bws,
           gdat,
           bandwidth.compute = TRUE,
           bwmethod,
           bwscaling,
           bwtype,
           cfac.dir,
           scale.factor.init,
           ckerbound,
           ckerlb,
           ckerorder,
           ckertype,
           ckerub,
           dfac.dir,
           dfac.init,
           dfc.dir,
           do.full.integral,
           ftol,
           scale.factor.init.upper,
           hbd.dir,
           hbd.init,
           initc.dir,
           initd.dir,
           invalid.penalty,
           itmax,
           lbc.dir,
           scale.factor.init.lower,
           lbd.dir,
           lbd.init,
           memfac,
           ngrid,
           nmulti,
           okertype,
           penalty.multiplier,
           powell.remin,
           scale.init.categorical.sample,
           scale.factor.search.lower = NULL,
           small,
           tol,
           transform.bounds,
           ...){

    t.names <- NULL
    if(!is.data.frame(dat) && !is.matrix(dat))
      t.names <- deparse(substitute(dat))

    dat <- toFrame(dat)
    
    if(!is.null(t.names))
      names(dat) <- t.names

    ## first grab dummy args for bandwidth() and perform 'bootstrap'
    ## bandwidth() call

    mc.names <- names(match.call(expand.dots = FALSE))
    margs <- c("bwmethod", "bwscaling", "bwtype", "ckertype", "ckerorder",
               "ckerbound", "ckerlb", "ckerub", "okertype")


    m <- match(margs, mc.names, nomatch = 0)
    any.m <- any(m != 0)

    bw.args <- list(
      bw = bws,
      nobs = dim(dat)[1],
      xdati = untangle(dat),
      xnames = names(dat),
      bandwidth.compute = bandwidth.compute
    )
    if (any.m) {
      nms <- mc.names[m]
      bw.args[nms] <- mget(nms, envir = environment(), inherits = FALSE)
    }
    tbw <- do.call(dbandwidth, bw.args)


    ## next grab dummies for actual bandwidth selection and perform call

    mc.names <- names(match.call(expand.dots = FALSE))
    margs <- c("gdat","bandwidth.compute", "nmulti", "powell.remin", "itmax",
               "do.full.integral", "ngrid", "ftol", "tol",
               "small", "lbc.dir", "dfc.dir", "cfac.dir", "initc.dir", 
               "lbd.dir", "hbd.dir", "dfac.dir", "initd.dir", 
               "scale.factor.init.lower", "scale.factor.init.upper", "scale.factor.init", 
               "lbd.init", "hbd.init", "dfac.init", 
               "scale.init.categorical.sample", "scale.factor.search.lower", "memfac",
               "transform.bounds",
               "invalid.penalty",
               "penalty.multiplier")
    m <- match(margs, mc.names, nomatch = 0)
    any.m <- any(m != 0)

    bwsel.args <- list(dat = dat, bws = tbw)
    if (any.m) {
      nms <- mc.names[m]
      bwsel.args[nms] <- mget(nms, envir = environment(), inherits = FALSE)
    }
    tbw <- .np_progress_select_bandwidth_enhanced(
      "Selecting distribution bandwidth",
      do.call(npudistbw.dbandwidth, bwsel.args)
    )

    mc <- match.call(expand.dots = FALSE)
    environment(mc) <- parent.frame()
    tbw$call <- mc

    return(tbw)
  }

test_that("npudist exposes unconditional fast counts in bandwidth summary", {
  library(np)
  set.seed(42)
  n <- 100L
  y <- runif(n)

  fit <- npudist(
    ~ y,
    ckerbound = "fixed",
    ckerlb = 0,
    ckerub = 1,
    nomad = TRUE
  )

  expect_true(is.finite(as.numeric(fit$bws$num.feval[1L])))
  expect_true(is.finite(as.numeric(fit$bws$num.feval.fast[1L])))
  expect_lte(as.numeric(fit$bws$num.feval.fast[1L]), as.numeric(fit$bws$num.feval[1L]))

  txt <- paste(capture.output(summary(fit$bws)), collapse = "\n")
  expect_true(grepl("Number of Function Evaluations:", txt, fixed = TRUE))
  expect_true(grepl("fast =", txt, fixed = TRUE))
  expect_true(grepl(sprintf("fast = %s", format(fit$bws$num.feval.fast[1L])), txt, fixed = TRUE))
})

test_that("npudist counts forced large-h unconditional evaluations as fast", {
  set.seed(42)
  y <- runif(100L)
  dat <- data.frame(y = y)

  bw0 <- npudistbw(
    dat = dat,
    bwmethod = "cv.cdf",
    ckerbound = "fixed",
    ckerlb = 0,
    ckerub = 1,
    bandwidth.compute = FALSE
  )

  bw_big <- bw0
  bw_big$bw[] <- 1e10
  bw_big$bandwidth[] <- 1e10
  bw_big$sfactor[] <- 1e10

  forced <- npudistbw(
    dat = dat,
    bws = bw_big,
    bandwidth.compute = TRUE,
    nmulti = 1L,
    itmax = 1L,
    ckerbound = "fixed",
    ckerlb = 0,
    ckerub = 1
  )

  expect_true(is.finite(as.numeric(forced$num.feval[1L])))
  expect_true(is.finite(as.numeric(forced$num.feval.fast[1L])))
  expect_gt(as.numeric(forced$num.feval.fast[1L]), 0)
  expect_lte(as.numeric(forced$num.feval.fast[1L]), as.numeric(forced$num.feval[1L]))
})

test_that("npcopula basic functionality works", {
  data("faithful")
  # npcopula can take a dbandwidth (from npudistbw) for copula 
  # or bandwidth (from npudensbw) for copula density
  
  bw <- npudistbw(dat=faithful, bws=c(0.5, 5), bandwidth.compute=FALSE)
  
  # Copula CDF
  cop <- npcopula(data=faithful, bws=bw)
  expect_s3_class(cop, "npcopula")
  expect_true(is.data.frame(as.data.frame(cop)))
  expect_true("copula" %in% names(cop))
  expect_identical(cop$target, "distribution")
  expect_identical(cop$evaluation, "sample")
  expect_true(all(cop$copula >= 0 & cop$copula <= 1))
  
  # Copula density
  bw_dens <- npudensbw(dat=faithful, bws=c(0.5, 5), bandwidth.compute=FALSE)
  cop_dens <- npcopula(data=faithful, bws=bw_dens)
  expect_s3_class(cop_dens, "npcopula")
  expect_true(is.data.frame(as.data.frame(cop_dens)))
  expect_true("copula" %in% names(cop_dens))
  expect_identical(cop_dens$target, "density")
  expect_identical(cop_dens$evaluation, "sample")
  expect_true(all(cop_dens$copula >= 0))
})

test_that("npcopula formula route builds a plot-ready grid by default", {
  set.seed(42)
  dat <- data.frame(x = rnorm(50), y = 0.5 * rnorm(50) + rnorm(50, sd = 0.1))

  fit <- npcopula(~ x + y, data = dat, neval = 4, nmulti = 1)

  expect_s3_class(fit, "npcopula")
  expect_equal(nrow(as.data.frame(fit)), 16L)
  expect_identical(fit$target, "distribution")
  expect_identical(fit$evaluation, "grid")
  expect_true(isTRUE(fit$u.auto))
  expect_equal(fit$grid.dim, c(4L, 4L))
  expect_true(inherits(fit$bws, "dbandwidth"))
  expect_true(is.data.frame(fit$eval))
  expect_equal(as.data.frame(fit), fit$eval, ignore_attr = TRUE)

  fixed.fit <- npcopula(
    ~ x + y,
    data = dat,
    bws = c(0.7, 0.7),
    bandwidth.compute = FALSE,
    neval = 3
  )
  expect_s3_class(fixed.fit, "npcopula")
  expect_equal(nrow(as.data.frame(fixed.fit)), 9L)
  expect_true(inherits(fixed.fit$bws, "dbandwidth"))

  u <- data.frame(x = c(0.25, 0.75), y = c(0.25, 0.75))
  explicit.u.fit <- npcopula(
    ~ x + y,
    data = dat,
    bws = c(0.7, 0.7),
    bandwidth.compute = FALSE,
    u = u,
    n.quasi.inv = 40
  )
  expect_s3_class(explicit.u.fit, "npcopula")
  expect_equal(nrow(as.data.frame(explicit.u.fit)), 4L)
  expect_true(inherits(explicit.u.fit$bws, "dbandwidth"))
})

test_that("npcopula formula route can request sample evaluation and density target", {
  set.seed(43)
  dat <- data.frame(x = rnorm(40), y = 0.5 * rnorm(40) + rnorm(40, sd = 0.1))

  sample.fit <- npcopula(~ x + y, data = dat, evaluation = "sample", nmulti = 1)
  expect_equal(nrow(as.data.frame(sample.fit)), nrow(dat))
  expect_identical(sample.fit$evaluation, "sample")

  dens.fit <- npcopula(~ x + y, data = dat, target = "density", neval = 3, nmulti = 1)
  expect_equal(nrow(as.data.frame(dens.fit)), 9L)
  expect_identical(dens.fit$target, "density")
  expect_true(inherits(dens.fit$bws, "bandwidth"))

  dens.fixed <- npcopula(
    ~ x + y,
    data = dat,
    target = "density",
    bws = c(0.7, 0.7),
    bandwidth.compute = FALSE,
    neval = 3
  )
  expect_equal(nrow(as.data.frame(dens.fixed)), 9L)
  expect_identical(dens.fixed$target, "density")
  expect_true(inherits(dens.fixed$bws, "bandwidth"))
})

test_that("npcopula formula route rejects unsafe automatic high-dimensional grids", {
  dat <- data.frame(x = rnorm(20), y = rnorm(20), z = rnorm(20))

  expect_error(
    npcopula(~ x + y + z, data = dat, nmulti = 1),
    "automatic copula probability grids"
  )
})

test_that("npcopula target conflicts with explicit bandwidth object fail clearly", {
  data("faithful")
  bw <- npudistbw(dat=faithful, bws=c(0.5, 5), bandwidth.compute=FALSE)
  expect_error(
    npcopula(data = faithful, bws = bw, target = "density"),
    "target.*conflicts"
  )
})

test_that("npcopula validates probability-grid and quasi-inverse controls", {
  data("faithful")
  bw <- npudistbw(dat = faithful, bws = c(0.5, 5), bandwidth.compute = FALSE)

  expect_error(
    npcopula(data = faithful, bws = bw,
             u = data.frame(eruptions = "a", waiting = "b")),
    "numeric probability"
  )
  expect_error(
    npcopula(data = faithful, bws = bw,
             u = data.frame(eruptions = 1.2, waiting = 0.5)),
    "u must lie in \\[0,1\\]"
  )
  expect_error(
    npcopula(data = faithful, bws = bw,
             u = data.frame(eruptions = 0.5)),
    "u and bws are incompatible"
  )
  expect_error(
    npcopula(data = faithful, bws = bw, u = c(0.5, 0.5),
             n.quasi.inv = 2.5),
    "integer greater than one"
  )
  expect_error(
    npcopula(data = faithful, bws = bw, u = c(0.5, 0.5),
             n.quasi.inv = NA),
    "integer greater than one"
  )
  expect_error(
    npcopula(data = faithful, bws = bw, u = c(0.5, 0.5),
             er.quasi.inv = -1),
    "non-negative finite"
  )
  expect_error(
    npcopula(data = faithful, bws = bw, u = c(0.5, 0.5),
             er.quasi.inv = NA),
    "non-negative finite"
  )
})

test_that("predict.npcopula evaluates stored bandwidths on probability grids", {
  data("faithful")
  bw <- npudistbw(dat = faithful, bws = c(0.5, 5), bandwidth.compute = FALSE)
  u <- data.frame(eruptions = seq(0.2, 0.8, length.out = 3),
                  waiting = seq(0.2, 0.8, length.out = 3))
  fit <- npcopula(data = faithful, bws = bw, u = u, n.quasi.inv = 40)

  expect_equal(predict(fit), fitted(fit))
  expect_null(attr(fit, "bws"))

  u.new <- data.frame(eruptions = c(0.25, 0.75),
                      waiting = c(0.25, 0.75))
  expected <- npcopula(bws = fit$bws, data = faithful, u = u.new,
                       n.quasi.inv = 40)
  expect_equal(predict(fit, u = u.new, n.quasi.inv = 40), fitted(expected))

  u.alias <- data.frame(u1 = c(0.25, 0.75), u2 = c(0.25, 0.75))
  expect_equal(predict(fit, newdata = u.alias, n.quasi.inv = 40),
               fitted(expected))

  pred.object <- predict(fit, u = u.new, n.quasi.inv = 40, output = "object")
  expect_s3_class(pred.object, "npcopula")
  expect_equal(as.data.frame(predict(fit, u = u.new, n.quasi.inv = 40,
                                     output = "data")),
               as.data.frame(expected),
               ignore_attr = TRUE)

  pred.se <- predict(fit, se.fit = TRUE)
  expect_equal(pred.se$fit, fitted(fit))
  expect_equal(pred.se$se.fit, se(fit))
  expect_length(pred.se$se.fit, length(fitted(fit)))
  expect_true(all(is.finite(pred.se$se.fit)))
  expect_error(predict(fit, se.fit = "yes"),
               "'se.fit' must be TRUE or FALSE", fixed = TRUE)
})

test_that("npcopula fitted and basic plot methods work", {
  data("faithful")
  bw <- npudistbw(dat=faithful, bws=c(0.5, 5), bandwidth.compute=FALSE)
  u <- data.frame(eruptions = seq(0, 1, length.out = 4),
                  waiting = seq(0, 1, length.out = 4))
  fit <- npcopula(data = faithful, bws = bw, u = u, n.quasi.inv = 40)

  expect_equal(fitted(fit), fit$copula)
  expect_silent(plot(fit, view = "fixed"))
  expect_silent(plot(fit, view = "fixed", col = NULL))
  expect_silent(plot(fit, col = "lightblue", border = "grey40",
                     theta = 20, phi = 25, view = "fixed", shade = 0.3))
  expect_silent(plot(fit, view = "image", perspective = FALSE))
  expect_silent(plot(fit, view = "contour", perspective = FALSE))
  expect_true("copula" %in%
                getFromNamespace(".np_progress_single_line_surfaces", "np")())
  expect_error(plot(fit, view = "image", renderer = "rgl"),
               "surface views")

  one.point <- npcopula(data = faithful, bws = bw, u = c(0.5, 0.5), n.quasi.inv = 40)
  expect_equal(nrow(as.data.frame(one.point)), 1L)
  expect_equal(one.point$grid.dim, c(1L, 1L))
})

test_that("npcopula contour and image plots use descriptive default titles", {
  data("faithful")
  bw <- npudistbw(dat = faithful, bws = c(0.5, 5), bandwidth.compute = FALSE)
  u <- data.frame(eruptions = seq(0, 1, length.out = 3),
                  waiting = seq(0, 1, length.out = 3))
  fit <- npcopula(data = faithful, bws = bw, u = u, n.quasi.inv = 40)

  captured <- new.env(parent = emptyenv())
  captured$contour <- NULL
  captured$image <- NULL
  pdf(file = tempfile(fileext = ".pdf"))
  on.exit(dev.off(), add = TRUE)
  trace(
    what = "contour",
    where = asNamespace("graphics"),
    tracer = bquote({
      assign("contour", as.list(match.call()), envir = .(captured))
    }),
    print = FALSE
  )
  on.exit(
    try(untrace("contour", where = asNamespace("graphics")), silent = TRUE),
    add = TRUE
  )
  trace(
    what = "image",
    where = asNamespace("graphics"),
    tracer = bquote({
      assign("image", as.list(match.call()), envir = .(captured))
    }),
    print = FALSE
  )
  on.exit(
    try(untrace("image", where = asNamespace("graphics")), silent = TRUE),
    add = TRUE
  )

  expect_silent(plot(fit, view = "contour", perspective = FALSE))
  expect_silent(plot(fit, view = "image", perspective = FALSE))
  expect_identical(eval(captured$contour$main), "Copula Contour")
  expect_identical(eval(captured$image$main), "Contour Heatmap")
})

test_that("npcopula perspective defaults follow shared surface convention", {
  plot.method <- getS3method("plot", "npcopula")
  form <- formals(plot.method)
  expect_equal(form$theta, 0.0)
  expect_equal(form$phi, 20.0)
  expect_identical(eval(form$view)[1L], "rotate")

  data("faithful")
  bw <- npudistbw(dat = faithful, bws = c(0.5, 5), bandwidth.compute = FALSE)
  u <- data.frame(eruptions = seq(0, 1, length.out = 3),
                  waiting = seq(0, 1, length.out = 3))
  fit <- npcopula(data = faithful, bws = bw, u = u, n.quasi.inv = 40)

  captured <- new.env(parent = emptyenv())
  captured$args <- list()
  pdf(file = tempfile(fileext = ".pdf"))
  on.exit(dev.off(), add = TRUE)
  trace(
    what = "persp",
    where = asNamespace("graphics"),
    tracer = bquote({
      assign("args",
             as.list(match.call()),
             envir = .(captured))
    }),
    print = FALSE
  )
  on.exit(
    try(untrace("persp", where = asNamespace("graphics")), silent = TRUE),
    add = TRUE
  )

  expect_silent(plot(fit, view = "fixed"))
  expect_equal(eval(captured$args$theta), 0.0)
  expect_equal(eval(captured$args$phi), 20.0)
  expect_identical(captured$args$ticktype, "detailed")
})

test_that("npcopula zlim caps plotted surface without changing data", {
  data("faithful")
  bw <- npudistbw(dat = faithful, bws = c(0.5, 5), bandwidth.compute = FALSE)
  u <- data.frame(eruptions = seq(0, 1, length.out = 3),
                  waiting = seq(0, 1, length.out = 3))
  fit <- npcopula(data = faithful, bws = bw, u = u, n.quasi.inv = 40)
  fit$eval$copula[1L] <- 5
  fit$copula <- fit$eval$copula

  expect_gt(max(plot(fit, output = "data")$copula), 1)
  pdf(file = tempfile(fileext = ".pdf"))
  on.exit(dev.off(), add = TRUE)
  expect_warning(plot(fit, view = "fixed", zlim = c(0, 1)), NA)
})

test_that("npcopula plot intervals work for grid surfaces", {
  set.seed(44)
  dat <- data.frame(x = rnorm(45), y = rnorm(45))
  fit <- npcopula(~ x + y, data = dat, neval = 3, nmulti = 1)

  asym <- plot(fit, output = "data", errors = "asymptotic", band = "all")
  expect_true(all(c("lower", "upper", "pointwise.lower",
                    "bonferroni.upper") %in% names(asym)))
  expect_equal(nrow(asym), nrow(as.data.frame(fit)))

  boot <- plot(fit, output = "data", errors = "bootstrap",
               bootstrap = "inid", B = 3, band = "pmzsd")
  expect_true(all(c("center", "lower", "upper") %in% names(boot)))
  expect_equal(nrow(boot), nrow(as.data.frame(fit)))

  dens.bw <- npudensbw(dat = dat, bws = c(0.7, 0.7),
                       bandwidth.compute = FALSE)
  dens.fit <- npcopula(
    bws = dens.bw,
    data = dat,
    u = data.frame(x = seq(0.2, 0.8, length.out = 3),
                   y = seq(0.2, 0.8, length.out = 3)),
    n.quasi.inv = 40
  )
  dens <- plot(dens.fit, output = "data", errors = "bootstrap",
               bootstrap = "inid", B = 3, band = "pmzsd")
  expect_true(all(c("center", "lower", "upper") %in% names(dens)))
  expect_equal(nrow(dens), nrow(as.data.frame(dens.fit)))

  sample.fit <- npcopula(~ x + y, data = dat, evaluation = "sample", nmulti = 1)
  expect_error(plot(sample.fit, errors = "asymptotic"), "grid evaluation")
})

test_that("npcopula plots mixed ordered grids on requested probability axes", {
  set.seed(42)
  n <- 80
  n.eval <- 12
  rho <- 0.99
  x <- rnorm(n)
  ylatent <- rho * x + sqrt(1 - rho^2) * rnorm(n)
  dat <- data.frame(
    x = x,
    y = ordered(as.integer(cut(
      ylatent,
      quantile(ylatent, seq(0, 1, by = 0.1)),
      include.lowest = TRUE
    )) - 1)
  )
  fit <- npcopula(~ x + y, data = dat, neval = n.eval, nmulti = 1)

  expect_equal(nrow(as.data.frame(fit)), n.eval * n.eval)
  expect_equal(length(unique(fit$u.grid$u1)), n.eval)
  expect_equal(length(unique(fit$u.grid$u2)), n.eval)
  expect_lt(length(unique(as.data.frame(fit)$u2)), n.eval)
  expect_silent(plot(fit, view = "contour", output = "data"))

  empirical <- plot(fit, view = "empirical", output = "data")
  expect_equal(nrow(empirical), n)
  expect_true(all(c("u1", "u2") %in% names(empirical)))

  all.payload <- plot(fit, view = "all", output = "data")
  expect_equal(nrow(all.payload$copula), n.eval * n.eval)
  expect_equal(nrow(all.payload$empirical), n)
  expect_equal(nrow(all.payload$density), n.eval * n.eval)
  expect_error(plot(fit, view = "all", renderer = "rgl"), "renderer='base'")

  dens <- npcopula(~ x + y, data = dat, target = "density",
                   neval = n.eval, nmulti = 1)
  expect_error(plot(dens, view = "all"),
               "requires a copula distribution object")
})

with_plot_contract_device <- function(expr) {
  pdf(file = tempfile(fileext = ".pdf"))
  on.exit(dev.off(), add = TRUE)
  suppressWarnings(force(expr))
}

expect_plot_modes_match_unconditional <- function(bw,
                                                  xdat,
                                                  method_label,
                                                  yfield,
                                                  errors,
                                                  band) {
  data.out <- suppressWarnings(
    plot(
      bw,
      xdat = xdat,
      output = "data",
      perspective = FALSE,
      errors = errors,
      band = band,
      B = 5,
      view = "fixed"
    )
  )

  plot.out <- with_plot_contract_device(
    plot(
      bw,
      xdat = xdat,
      output = "plot-data",
      perspective = FALSE,
      errors = errors,
      band = band,
      B = 5,
      view = "fixed"
    )
  )

  expect_true(is.list(plot.out), info = method_label)
  expect_named(plot.out, names(data.out), info = method_label)

  for (nm in names(data.out)) {
    expect_true(
      inherits(plot.out[[nm]], class(data.out[[nm]])[1L]),
      info = paste(method_label, nm, "class")
    )
    expect_equal(plot.out[[nm]]$eval, data.out[[nm]]$eval, info = paste(method_label, nm, "eval"))
    expect_equal(plot.out[[nm]][[yfield]], data.out[[nm]][[yfield]], info = paste(method_label, nm, yfield))
    expect_equal(plot.out[[nm]]$derr, data.out[[nm]]$derr, info = paste(method_label, nm, "derr"))
    expect_equal(plot.out[[nm]]$bias, data.out[[nm]]$bias, info = paste(method_label, nm, "bias"))
  }
}

test_that("plot engine startup helper respects plot.par.mfrow option override", {
  skip_if_not_installed("np")

  begin <- getFromNamespace(".np_plot_engine_begin", "np")
  restore <- getFromNamespace(".np_plot_restore_par", "np")

  pdf(file = tempfile(fileext = ".pdf"))
  on.exit(dev.off(), add = TRUE)

  old_opts <- options(plot.par.mfrow = FALSE)
  on.exit(options(old_opts), add = TRUE)

  state <- begin(plot.par.mfrow = TRUE)
  on.exit(restore(state$oldpar), add = TRUE)

  expect_identical(state$plot.par.mfrow, FALSE)
  expect_true(is.numeric(state$oldpar))
  expect_length(state$oldpar, 1L)
})

test_that("unconditional interval payload helper preserves bootstrap and asymptotic semantics", {
  skip_if_not_installed("np")

  payload <- getFromNamespace(".np_plot_interval_payload", "np")

  boot.raw <- list(
    boot.err = cbind(c(0.2, 0.3), c(0.4, 0.5), c(1.2, 1.8)),
    boot.all.err = list(pointwise = cbind(c(0.2, 0.3), c(0.4, 0.5))),
    bxp = list(stats = matrix(1, nrow = 5, ncol = 1))
  )

  boot.out <- payload(
    estimate = c(1, 2),
    se = c(0.1, 0.2),
    errors = "bootstrap",
    alpha = 0.05,
    band = "all",
    center = "bias-corrected",
    bootstrap_raw = boot.raw
  )

  expect_equal(boot.out$err, boot.raw$boot.err)
  expect_identical(boot.out$all.err, boot.raw$boot.all.err)
  expect_identical(boot.out$center, boot.raw$boot.err[,3])
  expect_identical(boot.out$bxp, boot.raw$bxp)

  asym.out <- payload(
    estimate = c(1, 2),
    se = c(0.1, 0.2),
    errors = "asymptotic",
    alpha = 0.05,
    band = "pointwise",
    center = "estimate",
    bootstrap_raw = NULL
  )

  expect_identical(asym.out$center, c(1, 2))
  expect_equal(dim(asym.out$err), c(2L, 3L))
  expect_true(all(is.na(asym.out$err[,3])))
  expect_identical(asym.out$bxp, list())
})

test_that("bandwidth and dbandwidth engines preserve data vs plot-data payloads", {
  skip_if_not_installed("np")

  set.seed(20260313)
  xdat <- data.frame(x = rnorm(18), z = runif(18))

  dens.bw <- npudensbw(
    dat = xdat,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE
  )
  dist.bw <- npudistbw(
    dat = xdat,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE
  )

  expect_plot_modes_match_unconditional(
    bw = dens.bw,
    xdat = xdat,
    method_label = "npudensbw bootstrap",
    yfield = "dens",
    errors = "bootstrap",
    band = "all"
  )
  expect_plot_modes_match_unconditional(
    bw = dist.bw,
    xdat = xdat,
    method_label = "npudistbw bootstrap",
    yfield = "dist",
    errors = "bootstrap",
    band = "all"
  )
  expect_plot_modes_match_unconditional(
    bw = dens.bw,
    xdat = xdat,
    method_label = "npudensbw asymptotic",
    yfield = "dens",
    errors = "asymptotic",
    band = "pointwise"
  )
  expect_plot_modes_match_unconditional(
    bw = dist.bw,
    xdat = xdat,
    method_label = "npudistbw asymptotic",
    yfield = "dist",
    errors = "asymptotic",
    band = "pointwise"
  )
})

with_np_bindings <- function(bindings, code) {
  code <- substitute(code)
  ns <- asNamespace("np")
  old <- lapply(names(bindings), function(name) get(name, envir = ns, inherits = FALSE))
  names(old) <- names(bindings)

  for (name in names(bindings)) {
    was_locked <- bindingIsLocked(name, ns)
    if (was_locked) {
      unlockBinding(name, ns)
    }
    assign(name, bindings[[name]], envir = ns)
    if (was_locked) {
      lockBinding(name, ns)
    }
  }

  on.exit({
    for (name in names(old)) {
      was_locked <- bindingIsLocked(name, ns)
      if (was_locked) {
        unlockBinding(name, ns)
      }
      assign(name, old[[name]], envir = ns)
      if (was_locked) {
        lockBinding(name, ns)
      }
    }
  }, add = TRUE)

  eval(code, envir = parent.frame())
}

capture_progress_conditions <- function(expr) {
  messages <- character()
  warnings <- character()
  value <- NULL

  message.output <- utils::capture.output(
    value <- withCallingHandlers(
      expr,
      message = function(m) {
        messages <<- c(messages, conditionMessage(m))
        invokeRestart("muffleMessage")
      },
      warning = function(w) {
        warnings <<- c(warnings, conditionMessage(w))
        invokeRestart("muffleWarning")
      }
    ),
    type = "message"
  )

  list(value = value, messages = c(messages, message.output), warnings = warnings)
}

normalize_messages <- function(x) {
  sub("\n$", "", x)
}

make_npcopula_fixture <- function(seed = 42, n = 30) {
  set.seed(seed)
  data.frame(
    x = runif(n),
    y = 0.7 * runif(n) + 0.3 * rnorm(n, sd = 0.1)
  )
}

test_that("npcopula sample-realization path emits compact staged progress", {
  skip_if_not(exists("npcopula", mode = "function"), "np package not attached")
  mydat <- make_npcopula_fixture()
  bw <- npudistbw(dat = mydat, bws = c(0.2, 0.2), bandwidth.compute = FALSE)

  old_opts <- options(np.messages = TRUE)
  on.exit(options(old_opts), add = TRUE)

  res <- with_np_bindings(
    list(.np_progress_is_interactive = function() TRUE),
    capture_progress_conditions(
      npcopula(bws = bw, data = mydat)
    )
  )

  messages <- normalize_messages(res$messages)

  expect_s3_class(res$value, "npcopula")
  expect_true(is.data.frame(as.data.frame(res$value)))
  expect_true(any(grepl("\\[np\\] Copula (distribution|dist) sample", messages)))
  expect_false(any(grepl("^\\[np\\] Computing the marginal of", messages)))
})

test_that("npcopula u-grid density path emits compact staged progress", {
  skip_if_not(exists("npcopula", mode = "function"), "np package not attached")
  mydat <- make_npcopula_fixture(seed = 99)
  bw <- npudensbw(dat = mydat, bws = c(0.2, 0.2), bandwidth.compute = FALSE)
  u <- as.matrix(data.frame(x = c(0.25, 0.75), y = c(0.25, 0.75)))

  old_opts <- options(np.messages = TRUE)
  on.exit(options(old_opts), add = TRUE)

  res <- with_np_bindings(
    list(.np_progress_is_interactive = function() TRUE),
    capture_progress_conditions(
      npcopula(
        bws = bw,
        data = mydat,
        u = u,
        n.quasi.inv = 40
      )
    )
  )

  messages <- normalize_messages(res$messages)

  expect_s3_class(res$value, "npcopula")
  expect_true(is.data.frame(as.data.frame(res$value)))
  expect_true(any(grepl("\\[np\\] Copula (density|dens) grid", messages)))
  expect_false(any(grepl("^\\[np\\] Computing the quasi-inverse", messages)))
  expect_false(any(grepl("^\\[np\\] Computing the marginal of", messages)))
})

test_that("npcopula progress respects np.messages FALSE", {
  skip_if_not(exists("npcopula", mode = "function"), "np package not attached")
  mydat <- make_npcopula_fixture(seed = 17)
  bw <- npudistbw(dat = mydat, bws = c(0.2, 0.2), bandwidth.compute = FALSE)

  old_opts <- options(np.messages = FALSE)
  on.exit(options(old_opts), add = TRUE)

  res <- with_np_bindings(
    list(.np_progress_is_interactive = function() TRUE),
    capture_progress_conditions(
      npcopula(bws = bw, data = mydat)
    )
  )

  expect_length(res$messages, 0)
})

test_that("npcopula progress respects suppressMessages", {
  skip_if_not(exists("npcopula", mode = "function"), "np package not attached")
  mydat <- make_npcopula_fixture(seed = 23)
  bw <- npudistbw(dat = mydat, bws = c(0.2, 0.2), bandwidth.compute = FALSE)

  old_opts <- options(np.messages = TRUE)
  on.exit(options(old_opts), add = TRUE)

  res <- with_np_bindings(
    list(.np_progress_is_interactive = function() TRUE),
    capture_progress_conditions(
      suppressMessages(npcopula(bws = bw, data = mydat))
    )
  )

  expect_length(res$messages, 0)
})

test_that("bounded adaptive_nn is available for unconditional distribution", {
  set.seed(20260325)
  n <- 40
  x <- runif(n)
  xdf <- data.frame(x = x)

  bw.fixed <- npudistbw(
    dat = xdf,
    bwmethod = "cv.cdf",
    bwtype = "fixed",
    ckerorder = 2,
    ckerbound = "range",
    nmulti = 1,
    powell.remin = FALSE,
    itmax = 50,
    tol = 0.1
  )

  bw.gnn <- npudistbw(
    dat = xdf,
    bwmethod = "cv.cdf",
    bwtype = "generalized_nn",
    ckerorder = 2,
    ckerbound = "range",
    nmulti = 1,
    powell.remin = FALSE,
    itmax = 50,
    tol = 0.1
  )

  bw.ad <- npudistbw(
    dat = xdf,
    bwmethod = "cv.cdf",
    bwtype = "adaptive_nn",
    ckerorder = 2,
    ckerbound = "range",
    nmulti = 1,
    powell.remin = FALSE,
    itmax = 50,
    tol = 0.1
  )

  fit.ad <- npudist(bws = bw.ad)
  eval.ad <- predict(fit.ad, newdata = data.frame(x = sort(x)))

  expect_true(is.finite(bw.fixed$fval))
  expect_true(is.finite(bw.gnn$fval))
  expect_true(is.finite(bw.ad$fval))
  expect_true(all(is.finite(as.numeric(bw.ad$bw))))
  expect_true(all(is.finite(as.numeric(fitted(fit.ad)))))
  expect_true(all(is.finite(as.numeric(eval.ad))))
  expect_true(all(as.numeric(eval.ad) >= -0.05))
  expect_true(all(as.numeric(eval.ad) <= 1.05))
  expect_true(all(diff(as.numeric(eval.ad)) >= -1e-8))
  expect_false(isTRUE(all.equal(as.numeric(bw.ad$bw), as.numeric(bw.fixed$bw))))
})

test_that("npudist basic functionality works", {
  data("faithful")
  # Use fixed bandwidths for speed
  bw <- npudistbw(dat=faithful, bws=c(0.5, 5), bandwidth.compute=FALSE)
  
  fit <- npudist(bws=bw)
  
  expect_s3_class(fit, "npdistribution")
  expect_type(predict(fit), "double")
  expect_true(all(predict(fit) >= 0 & predict(fit) <= 1))
  expect_equal(length(predict(fit)), nrow(faithful))
  
  expect_output(summary(fit))
})

test_that("npudist works with formula and factors", {
  data("Italy")
  Italy_sub <- Italy[1:100, ]
  bw <- npudistbw(formula=~ordered(year)+gdp, data=Italy_sub, 
                  bws=c(0.5, 1.0), bandwidth.compute=FALSE)
  
  fit <- npudist(bws=bw)
  expect_s3_class(fit, "npdistribution")
})

test_that("npconmode basic functionality works", {
  data("faithful")
  # Use small subset
  # npconmode needs a conditional density bandwidth object
  # where ydat is discrete.
  y <- factor(ifelse(faithful$eruptions[1:50] > 3, "long", "short"))
  x <- faithful$waiting[1:50]
  # bws[1] is for ydat (discrete), bws[2] is for xdat (continuous)
  bw <- npcdensbw(xdat=x, ydat=y, 
                  bws=c(0.1, 5), bandwidth.compute=FALSE)
  
  mode_fit <- npconmode(bws=bw)
  
  expect_s3_class(mode_fit, "conmode")
  expect_type(fitted(mode_fit), "double")
})

test_that("npquantile basic functionality works", {
  data("faithful")
  x <- faithful$waiting[1:50]
  bw <- npudistbw(dat=x, bws=5, bandwidth.compute=FALSE)
  
  # npquantile is for unconditional quantile
  q <- npquantile(x, bws=bw, tau=c(0.25, 0.5, 0.75))
  
  expect_type(q, "double")
  expect_length(q, 3)
})

test_that("npudens and npudist fixed no-error plot-data prototypes match current routes", {
  dens.proto <- getFromNamespace(".np_plot_proto_npudens_fixed_none_data", "np")
  dist.proto <- getFromNamespace(".np_plot_proto_npudist_fixed_none_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(2401)

  n <- 75L
  x <- data.frame(x1 = rnorm(n), x2 = runif(n, -1, 1))
  dens.bw <- npudensbw(
    dat = x,
    bws = c(0.55, 0.60),
    bandwidth.compute = FALSE,
    bwtype = "fixed"
  )
  dist.bw <- npudistbw(
    dat = x,
    bws = c(0.55, 0.60),
    bandwidth.compute = FALSE,
    bwtype = "fixed"
  )

  old.dens <- suppressWarnings(plot(
    dens.bw,
    xdat = x,
    output = "data",
    errors = "none",
    view = "fixed",
    neval = 7L,
    perspective = TRUE
  ))
  cand.dens <- dens.proto(
    dens.bw,
    xdat = x,
    neval = 7L
  )
  stages.dens <- dens.proto(
    dens.bw,
    xdat = x,
    neval = 7L,
    return.stages = TRUE
  )

  expect_named(cand.dens, names(old.dens))
  expect_s3_class(cand.dens$d1, "npdensity")
  expect_named(cand.dens$d1, names(old.dens$d1))
  expect_equal(cand.dens$d1$eval, old.dens$d1$eval)
  expect_equal(cand.dens$d1$dens, old.dens$d1$dens)
  expect_equal(cand.dens$d1$derr, old.dens$d1$derr)
  expect_equal(cand.dens$d1$bias, old.dens$d1$bias)
  expect_named(stages.dens, c("state", "target_grid", "evaluator", "intervals", "bootstrap", "plot_data"))
  expect_equal(stages.dens$plot_data, cand.dens)
  expect_null(stages.dens$intervals)
  expect_null(stages.dens$bootstrap)
  expect_equal(stages.dens$evaluator$dens, old.dens$d1$dens)

  old.dist <- suppressWarnings(plot(
    dist.bw,
    xdat = x,
    output = "data",
    errors = "none",
    view = "fixed",
    neval = 7L,
    perspective = TRUE
  ))
  cand.dist <- dist.proto(
    dist.bw,
    xdat = x,
    neval = 7L
  )
  stages.dist <- dist.proto(
    dist.bw,
    xdat = x,
    neval = 7L,
    return.stages = TRUE
  )

  expect_named(cand.dist, names(old.dist))
  expect_s3_class(cand.dist$d1, "npdistribution")
  expect_named(cand.dist$d1, names(old.dist$d1))
  expect_equal(cand.dist$d1$eval, old.dist$d1$eval)
  expect_equal(cand.dist$d1$dist, old.dist$d1$dist)
  expect_equal(cand.dist$d1$derr, old.dist$d1$derr)
  expect_equal(cand.dist$d1$bias, old.dist$d1$bias)
  expect_named(stages.dist, c("state", "target_grid", "evaluator", "intervals", "bootstrap", "plot_data"))
  expect_equal(stages.dist$plot_data, cand.dist)
  expect_null(stages.dist$intervals)
  expect_null(stages.dist$bootstrap)
  expect_equal(stages.dist$evaluator$dist, old.dist$d1$dist)
})

test_that("npudens and npudist fixed asymptotic and bootstrap prototypes match current routes", {
  proto <- getFromNamespace(".np_plot_proto_unconditional_fixed_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(2402)

  n <- 65L
  x <- data.frame(x1 = rnorm(n), x2 = runif(n, -1, 1))
  dens.bw <- npudensbw(
    dat = x,
    bws = c(0.55, 0.60),
    bandwidth.compute = FALSE,
    bwtype = "fixed"
  )
  dist.bw <- npudistbw(
    dat = x,
    bws = c(0.55, 0.60),
    bandwidth.compute = FALSE,
    bwtype = "fixed"
  )

  for (case in c("dens", "dist")) {
    bws <- if (identical(case, "dens")) dens.bw else dist.bw
    cdf <- identical(case, "dist")
    value.name <- if (cdf) "dist" else "dens"
    error.name <- if (cdf) "derr" else "derr"

    for (band in c("pmzsd", "pointwise", "bonferroni")) {
      old <- suppressWarnings(plot(
        bws,
        xdat = x,
        output = "data",
        errors = "asymptotic",
        band = band,
        view = "fixed",
        neval = 6L,
        perspective = TRUE
      ))
      candidate <- proto(
        bws,
        xdat = x,
        neval = 6L,
        errors = "asymptotic",
        band = band,
        cdf = cdf
      )
      expect_equal(candidate$d1[[value.name]], old$d1[[value.name]], info = paste(case, band))
      expect_equal(candidate$d1[[error.name]], old$d1[[error.name]], info = paste(case, band))
    }

    boot.seed <- if (cdf) 4202L else 4201L
    set.seed(boot.seed)
    old <- suppressWarnings(plot(
      bws,
      xdat = x,
      output = "data",
      errors = "bootstrap",
      bootstrap = "inid",
      B = 13L,
      center = "bias-corrected",
      band = "pointwise",
      view = "fixed",
      neval = 5L,
      perspective = TRUE,
      random.seed = boot.seed
    ))
    set.seed(boot.seed)
    candidate <- suppressWarnings(proto(
      bws,
      xdat = x,
      neval = 5L,
      errors = "bootstrap",
      bootstrap = "inid",
      B = 13L,
      center = "bias-corrected",
      band = "pointwise",
      cdf = cdf
    ))
    set.seed(boot.seed)
    stages <- suppressWarnings(proto(
      bws,
      xdat = x,
      neval = 5L,
      errors = "bootstrap",
      bootstrap = "inid",
      B = 13L,
      center = "bias-corrected",
      band = "pointwise",
      cdf = cdf,
      return.stages = TRUE
    ))

    expect_equal(candidate$d1[[value.name]], old$d1[[value.name]], info = case)
    expect_equal(candidate$d1[[error.name]], old$d1[[error.name]], info = case)
    expect_equal(candidate$d1$bias, old$d1$bias, info = case)
    expect_identical(stages$intervals$method, "bootstrap")
    expect_identical(stages$bootstrap$method, "inid")
    expect_identical(stages$bootstrap$B, 13L)
    expect_equal(stages$plot_data, candidate, info = case)
  }
})

test_that("npudens and npudist staged plot data can be rendered without estimator re-entry", {
  dens.proto <- getFromNamespace(".np_plot_proto_npudens_fixed_none_data", "np")
  dist.proto <- getFromNamespace(".np_plot_proto_npudist_fixed_none_data", "np")
  render <- getFromNamespace(".np_plot_proto_rectangular_surface_base_render", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(2403)

  n <- 55L
  x <- data.frame(x1 = rnorm(n), x2 = runif(n, -1, 1))
  dens.bw <- npudensbw(
    dat = x,
    bws = c(0.55, 0.60),
    bandwidth.compute = FALSE,
    bwtype = "fixed"
  )
  dist.bw <- npudistbw(
    dat = x,
    bws = c(0.55, 0.60),
    bandwidth.compute = FALSE,
    bwtype = "fixed"
  )
  dens.data <- dens.proto(dens.bw, xdat = x, neval = 5L)
  dist.data <- dist.proto(dist.bw, xdat = x, neval = 5L)
  pdf.file <- tempfile(fileext = ".pdf")
  grDevices::pdf(pdf.file)
  on.exit(if (grDevices::dev.cur() > 1L) grDevices::dev.off(), add = TRUE)

  expect_identical(render(dens.data, perspective = FALSE), dens.data)
  expect_identical(render(dens.data, perspective = TRUE), dens.data)
  expect_identical(render(dist.data, perspective = FALSE), dist.data)
  expect_identical(render(dist.data, perspective = TRUE), dist.data)
  grDevices::dev.off()
  expect_true(file.exists(pdf.file))
  expect_gt(file.info(pdf.file)$size, 0)
})

sdeptest <- function(Srho,
                    Srho.cumulant,
                    Srho.bootstrap.mat,
                    Srho.cumulant.bootstrap.mat,
                    P,
                    P.cumulant,
                    bootstrap,
                    boot.num,
                    lag.num,
                    bw.y,
                    bw.y.lag,
                    bw.joint) {
  
  tsdep = list(Srho = Srho,
    Srho.cumulant = Srho.cumulant,
    Srho.bootstrap.mat = Srho.bootstrap.mat,
    Srho.cumulant.bootstrap.mat = Srho.cumulant.bootstrap.mat,
    P = P,
    P.cumulant = P.cumulant,
    bootstrap = bootstrap,
    boot.num = boot.num,
    lag.num = lag.num,
    bw.y = bw.y,
    bw.y.lag = bw.y.lag,
    bw.joint = bw.joint)

  if(tsdep$bootstrap) {

    tsdep$reject <- numeric()
    tsdep$rejectNum <- character()

    for(k in seq_len(lag.num)) {

      reject <- ' '
      
      if (P[k] < 0.1)
        reject <- '.'
      
      if (P[k] < 0.05)
        reject <- '*'
      
      if (P[k] < 0.01)
        reject <- '**'
      
      if (P[k] < 0.001)
        reject <- '***'

      tsdep$reject[k] <- reject
      tsdep$rejectNum[k] <- switch(reject,
                                   ' ' = 100,
                                   '.' = 10,
                                   '*' = 5,
                                   '**' = 1,
                                   '***' = 0.1)

    }
    
  }
    
  class(tsdep) <- "sdeptest"

  tsdep
  
}

print.sdeptest <- function(x, ...){
  if(x$bootstrap) {
    cat("\nConsistent Metric Entropy Test for Nonlinear Dependence",
        paste("\n", format(x$boot.num), " Bootstrap Replications, ",
              format(x$lag.num), " Lags\n",sep=""))
    for(k in seq_len(x$lag.num)) {
      cat(paste("\nTest Statistic ", sQuote(paste("Srho[", k , "]",sep="")), ": ",
          format(x$Srho[k]), "\tP Value: ", format.pval(x$P[k])," ", x$reject[k],sep=""))
    }

    cat("\n---\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")
    
    for(k in seq_len(x$lag.num)) {
      reject.msg <- if (x$reject[k] == ' ') {
        paste("\nFail to reject the null of independence at lag ", k, " at the 10% level", sep = "")
      } else {
        paste("\nNull of independence is rejected at lag ", k, " at the ", x$rejectNum[k], "% level", sep = "")
      }
      cat(reject.msg)
    }
  } else {
    cat("\nConsistent Nonlinear Dependence Metric Entropy",
        paste("\n", format(x$lag.num), " Lags\n",sep=""))
    for(k in seq_len(x$lag.num)) {
      cat(paste("\nStatistic ", sQuote(paste("Srho[", k , "]",sep="")), ": ",
          format(x$Srho[k]),sep=""))
    }
  }
  cat("\n\n")
}

summary.sdeptest <- function(object, ...){
  print(object)
}

with_np_degree_bindings <- function(bindings, code) {
  code <- substitute(code)
  ns <- asNamespace("np")
  old <- lapply(names(bindings), function(name) get(name, envir = ns, inherits = FALSE))
  names(old) <- names(bindings)

  for (name in names(bindings)) {
    was_locked <- bindingIsLocked(name, ns)
    if (was_locked) {
      unlockBinding(name, ns)
    }
    assign(name, bindings[[name]], envir = ns)
    if (was_locked) {
      lockBinding(name, ns)
    }
  }

  on.exit({
    for (name in names(old)) {
      was_locked <- bindingIsLocked(name, ns)
      if (was_locked) {
        unlockBinding(name, ns)
      }
      assign(name, old[[name]], envir = ns)
      if (was_locked) {
        lockBinding(name, ns)
      }
    }
  }, add = TRUE)

  eval(code, envir = parent.frame())
}

capture_degree_messages_only <- function(expr) {
  messages <- character()
  withCallingHandlers(
    expr,
    message = function(m) {
      messages <<- c(messages, conditionMessage(m))
      invokeRestart("muffleMessage")
    }
  )
  messages
}

degree_progress_time_values <- function(values) {
  force(values)
  i <- 0L
  function() {
    i <<- min(i + 1L, length(values))
    values[[i]]
  }
}

expect_nomad_unknown_bound_progress <- function(msgs, pkg_pattern, info = NULL) {
  expect_false(any(grepl("nomad\\+powell", msgs, ignore.case = TRUE)), info = info)
  expect_false(any(grepl("eval [0-9]+", msgs)), info = info)
  expect_false(any(grepl("fval=", msgs, fixed = TRUE)), info = info)
  expect_false(any(grepl("%|eta ", msgs)), info = info)
  expect_true(any(grepl(sprintf("^\\[%s\\] Selecting degree and bandwidth \\(", pkg_pattern), msgs)), info = info)
  expect_true(any(grepl(sprintf("^\\[%s\\] Refining bandwidth \\(", pkg_pattern), msgs)), info = info)
  expect_true(any(grepl("multistart [12]/2", msgs)), info = info)
  expect_true(any(grepl("iteration [0-9]+", msgs)), info = info)
  expect_true(any(grepl("deg \\(", msgs)), info = info)
  expect_true(any(grepl("best \\(", msgs)), info = info)
}

test_that("remaining serial NOMAD families use unknown-bound restart progress", {
  skip_if_not_installed("crs")

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0,
    np.progress.interval.known.sec = 0,
    np.progress.interval.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260401)
  n <- 18L
  x <- data.frame(x = sort(runif(n)))
  z <- data.frame(z = sort(runif(n)))
  y_reg <- sin(2 * pi * x$x) + rnorm(n, sd = 0.05)
  y_beta <- rbeta(n, 1 + x$x, 2 - x$x)
  y_pl <- 1 + 0.5 * x$x + cos(2 * pi * z$z) + rnorm(n, sd = 0.05)
  y_sc <- (1 + z$z^2) * x$x + rnorm(n, sd = 0.05)
  y_idx <- sin(x$x + 0.5 * z$z) + rnorm(n, sd = 0.05)

  cases <- list(
    npcdistbw = function() np::npcdistbw(
      y_reg ~ x,
      data = data.frame(y_reg = y_reg, x = x$x),
      regtype = "lp",
      degree.select = "coordinate",
      search.engine = "nomad+powell",
      degree.min = 0L,
      degree.max = 1L,
      degree.verify = FALSE,
      bwtype = "fixed",
      bwmethod = "cv.ls",
      nmulti = 2L,
      ngrid = 30L,
      max.bb.eval = 8L
    ),
    npplregbw = function() np::npplregbw(
      xdat = x,
      zdat = z,
      ydat = y_pl,
      regtype = "lp",
      bernstein.basis = TRUE,
      degree.select = "coordinate",
      search.engine = "nomad+powell",
      degree.min = 0L,
      degree.max = 1L,
      degree.verify = FALSE,
      bwtype = "fixed",
      bwmethod = "cv.ls",
      nmulti = 2L,
      max.bb.eval = 8L
    ),
    npscoefbw = function() np::npscoefbw(
      xdat = x,
      zdat = z,
      ydat = y_sc,
      regtype = "lp",
      bernstein.basis = TRUE,
      degree.select = "coordinate",
      search.engine = "nomad+powell",
      degree.min = 0L,
      degree.max = 1L,
      degree.verify = FALSE,
      bwtype = "fixed",
      bwmethod = "cv.ls",
      nmulti = 2L,
      max.bb.eval = 8L
    ),
    npindexbw = function() np::npindexbw(
      xdat = data.frame(x1 = x$x, x2 = z$z),
      ydat = y_idx,
      bws = c(1, 0.5, 0.35),
      method = "ichimura",
      regtype = "lp",
      bernstein.basis = TRUE,
      degree.select = "coordinate",
      search.engine = "nomad+powell",
      degree.min = 0L,
      degree.max = 1L,
      degree.verify = FALSE,
      bwtype = "fixed",
      nmulti = 2L,
      max.bb.eval = 8L
    )
  )

  for (case in names(cases)) {
    msgs <- with_np_degree_bindings(
      list(
        .np_progress_is_interactive = function() TRUE,
        .np_progress_renderer_for_surface = function(surface, capability) "legacy",
        .np_progress_now = degree_progress_time_values(seq(0, 60, by = 0.25))
      ),
      capture_degree_messages_only(cases[[case]]())
    )
    expect_nomad_unknown_bound_progress(msgs, pkg_pattern = "np", info = case)
  }
})

test_that("npcdensbw nomad+powell payload does not inject phantom multistart totals", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260320)
  dat <- data.frame(x = runif(20))
  dat$y <- rbeta(nrow(dat), 1, 1)

  trace_env <- new.env(parent = emptyenv())
  trace_env$totals <- integer()
  trace_env$powell.count <- 0L

  trace(
    np:::.np_progress_bandwidth_set_total,
    tracer = eval(substitute(
      quote({
        assign(
          "totals",
          c(get("totals", envir = .trace_env, inherits = FALSE), as.integer(total)),
          envir = .trace_env
        )
      }),
      list(.trace_env = trace_env)
    )),
    print = FALSE
  )
  trace(
    np:::.npcdensbw_with_powell_refinement_progress,
    tracer = eval(substitute(
      quote({
        assign(
          "powell.count",
          get("powell.count", envir = .trace_env, inherits = FALSE) + 1L,
          envir = .trace_env
        )
      }),
      list(.trace_env = trace_env)
    )),
    print = FALSE
  )
  on.exit(untrace(np:::.np_progress_bandwidth_set_total), add = TRUE)
  on.exit(untrace(np:::.npcdensbw_with_powell_refinement_progress), add = TRUE)

  bw <- np::npcdensbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "nomad+powell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ml",
    nmulti = 2L,
    max.bb.eval = 20L,
    cxkerbound = "range",
    cykerbound = "range"
  )

  expect_s3_class(bw, "conbandwidth")
  expect_identical(get("powell.count", envir = trace_env, inherits = FALSE), 1L)
  expect_false(any(get("totals", envir = trace_env, inherits = FALSE) == 2L))
})

test_that("npcdensbw NOMAD plus Powell progress mirrors shared restart detail", {
  skip_if_not_installed("crs")

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0,
    np.progress.interval.known.sec = 0,
    np.progress.interval.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260323)
  dat <- data.frame(x = runif(20))
  dat$y <- rbeta(nrow(dat), 1, 1)

  msgs <- with_np_degree_bindings(
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_renderer_for_surface = function(surface, capability) "legacy",
      .np_progress_now = degree_progress_time_values(seq(0, 60, by = 0.25))
    ),
    capture_degree_messages_only(
      np::npcdensbw(
        y ~ x,
        data = dat,
        regtype = "lp",
        degree.select = "coordinate",
        search.engine = "nomad+powell",
        degree.min = 0L,
        degree.max = 1L,
        bwtype = "fixed",
        bwmethod = "cv.ml",
        nmulti = 2L,
        max.bb.eval = 12L,
        cxkerbound = "range",
        cykerbound = "range"
      )
    )
  )

  expect_false(any(grepl("nomad\\+powell", msgs, ignore.case = TRUE)))
  expect_false(any(grepl("eval [0-9]+", msgs)))
  expect_false(any(grepl("fval=", msgs, fixed = TRUE)))
  expect_false(any(grepl("%|eta ", msgs)))
  expect_true(any(grepl("^\\[np\\] Selecting degree and bandwidth \\(", msgs)))
  expect_true(any(grepl("^\\[np\\] Refining bandwidth \\(", msgs)))
  expect_true(any(grepl("multistart [12]/2", msgs)))
  expect_true(any(grepl("iteration [0-9]+", msgs)))
  if (any(grepl("multistart 2/2", msgs)))
    expect_true(any(grepl("iteration [0-9]+( \\([0-9]+\\))?", msgs)))
  expect_true(any(grepl("deg \\(", msgs)))
  expect_true(any(grepl("best \\(", msgs)))
})

expect_np_npregbw_powell_progress_surface <- function(lines) {
  powell.lines <- lines[grepl("^\\[np\\] Refining bandwidth \\(", lines)]
  info <- paste(lines, collapse = "\n")

  expect_true(length(powell.lines) > 0L, info = info)
  expect_false(any(grepl(
    "^\\[np\\] Refining NOMAD solution with one Powell hot start at degree ",
    lines
  )), info = info)
  expect_false(any(grepl("best (", powell.lines, fixed = TRUE)), info = info)
  powell.iter.lines <- powell.lines[grepl("iter [0-9]+", powell.lines)]
  expect_true(length(powell.iter.lines) > 0L, info = info)
  expect_true(any(grepl("^\\[np\\] Refining bandwidth \\(elapsed ", powell.iter.lines)), info = info)
  expect_true(any(grepl(", degree \\(", powell.iter.lines)), info = info)
}

test_that("npregbw NOMAD plus Powell progress keeps lines compact and restart-oriented", {
  skip_if_not_installed("crs")

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0,
    np.progress.interval.known.sec = 0,
    np.progress.interval.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260321)
  dat <- data.frame(x = sort(runif(16)))
  dat$y <- sin(2 * pi * dat$x) + rnorm(nrow(dat), sd = 0.05)

  msgs <- with_np_degree_bindings(
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_renderer_for_surface = function(surface, capability) "legacy",
      .np_progress_now = degree_progress_time_values(seq(0, 40, by = 0.25))
    ),
    capture_degree_messages_only(
      get("npregbw", envir = asNamespace("np"), inherits = FALSE)(
        y ~ x,
        data = dat,
        nomad = TRUE,
        degree.max = 1L,
        nmulti = 2L,
        max.bb.eval = 8
      )
    )
  )

  expect_false(any(grepl("^\\[np\\] Starting NOMAD automatic polynomial degree search at degree ", msgs)))
  expect_false(any(grepl("^\\[np\\] Refining NOMAD solution with one Powell hot start at degree ", msgs)))
  expect_false(any(grepl("starting at degree \\(", msgs)))
  expect_false(any(grepl("nomad\\+powell", msgs, ignore.case = TRUE)))
  expect_false(any(grepl("eval [0-9]+", msgs)))
  expect_false(any(grepl("fval=", msgs, fixed = TRUE)))
  expect_false(any(grepl("%|eta ", msgs)))
  expect_true(any(grepl("^\\[np\\] Selecting degree and bandwidth \\(", msgs)))
  expect_true(any(grepl("^\\[np\\] Refining bandwidth \\(", msgs)))
  expect_np_npregbw_powell_progress_surface(msgs)
  expect_true(any(grepl("multistart [12]/2", msgs)))
  expect_true(any(grepl("iteration [0-9]+", msgs)))
  expect_true(any(grepl("deg \\(", msgs)))
  expect_true(any(grepl("best \\(", msgs)))
})

with_np_degree_bindings <- function(bindings, code) {
  code <- substitute(code)
  ns <- asNamespace("np")
  old <- lapply(names(bindings), function(name) get(name, envir = ns, inherits = FALSE))
  names(old) <- names(bindings)

  for (name in names(bindings)) {
    was_locked <- bindingIsLocked(name, ns)
    if (was_locked) {
      unlockBinding(name, ns)
    }
    assign(name, bindings[[name]], envir = ns)
    if (was_locked) {
      lockBinding(name, ns)
    }
  }

  on.exit({
    for (name in names(old)) {
      was_locked <- bindingIsLocked(name, ns)
      if (was_locked) {
        unlockBinding(name, ns)
      }
      assign(name, old[[name]], envir = ns)
      if (was_locked) {
        lockBinding(name, ns)
      }
    }
  }, add = TRUE)

  eval(code, envir = parent.frame())
}

test_that("npscoefbw exhaustive degree search matches manual profile minimum", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 28
  xdat <- data.frame(x = runif(n))
  zdat <- data.frame(z = sort(runif(n)))
  y <- (1 + zdat$z^2) * xdat$x + rnorm(n, sd = 0.08)

  bw0 <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree = 0L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  bw1 <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree = 1L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  auto <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_s3_class(auto, "scbandwidth")
  expect_true(isTRUE(auto$bernstein.basis))
  expect_identical(auto$degree.search$mode, "exhaustive")
  expect_true(isTRUE(auto$degree.search$completed))
  expect_true(isTRUE(auto$degree.search$certified))
  expect_lte(auto$fval, min(bw0$fval, bw1$fval) + 1e-10)
  expect_lte(auto$degree.search$best.fval, auto$degree.search$baseline.fval + 1e-10)
  expect_true(all(c("degree", "fval", "status", "cached") %in% names(auto$degree.search$trace)))

  manual <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  expect_null(manual$degree.search)
})

test_that("npscoefbw coordinate search can be exhaustively certified on a small grid", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 26
  xdat <- data.frame(x = runif(n))
  zdat <- data.frame(
    z1 = runif(n),
    z2 = runif(n)
  )
  y <- (1 + zdat$z1 + zdat$z2^2) * xdat$x + rnorm(n, sd = 0.08)

  exhaustive <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  coordinate <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = TRUE,
    degree.restarts = 1L,
    degree.max.cycles = 4L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_identical(coordinate$degree.search$mode, "coordinate")
  expect_true(isTRUE(coordinate$degree.search$completed))
  expect_true(isTRUE(coordinate$degree.search$certified))
  expect_equal(as.integer(coordinate$degree), as.integer(exhaustive$degree))
  expect_equal(coordinate$fval, exhaustive$fval, tolerance = 1e-10)
})

test_that("npscoefbw automatic degree search enforces pilot guardrails", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 24
  xdat <- data.frame(x = runif(n))
  zdat <- data.frame(z = runif(n))
  y <- (1 + zdat$z) * xdat$x + rnorm(n, sd = 0.08)

  expect_error(
    npscoefbw(
      xdat = xdat,
      zdat = zdat,
      ydat = y,
      regtype = "lc",
      degree.select = "exhaustive",
      degree.min = 0L,
      degree.max = 1L,
      bwtype = "fixed",
      bwmethod = "cv.ls",
      nmulti = 1L
    ),
    "automatic degree search currently requires regtype='lp'"
  )

  expect_error(
    npscoefbw(
      xdat = xdat,
      zdat = zdat,
      ydat = y,
      regtype = "lp",
      bandwidth.compute = FALSE,
      degree.select = "exhaustive",
      degree.min = 0L,
      degree.max = 1L,
      bws = 0.2
    ),
    "bandwidth.compute=TRUE"
  )

  bw <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    bernstein.basis = FALSE,
    degree.select = "exhaustive",
    degree.min = 0L,
    degree.max = 4L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_s3_class(bw, "scbandwidth")
  expect_false(isTRUE(bw$bernstein.basis))
  expect_lte(max(as.integer(bw$degree)), 4L)

  expect_error(
    npscoefbw(
      xdat = xdat,
      zdat = zdat,
      ydat = y,
      regtype = "lp",
      degree.select = "coordinate",
      degree.min = 0L,
      degree.max = 1L,
      cv.iterate = TRUE,
      bwtype = "fixed",
      bwmethod = "cv.ls",
      nmulti = 1L
    ),
    "cv.iterate=FALSE"
  )
})

test_that("npscoef forwards automatic LP degree search through npscoefbw", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 24
  dat <- data.frame(
    x = runif(n),
    z = runif(n)
  )
  dat$y <- (1 + dat$z^2) * dat$x + rnorm(n, sd = 0.08)

  fit <- npscoef(
    y ~ x | z,
    data = dat,
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_s3_class(fit, "smoothcoefficient")
  expect_s3_class(fit$bws, "scbandwidth")
  expect_false(is.null(fit$bws$degree.search))
  expect_identical(fit$bws$degree.search$mode, "exhaustive")
})

test_that("npscoefbw automatic degree search defaults to NOMAD plus Powell", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 24
  xdat <- data.frame(x = runif(n))
  zdat <- data.frame(z = sort(runif(n)))
  y <- (1 + zdat$z^2) * xdat$x + rnorm(n, sd = 0.08)

  bw <- npscoefbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_s3_class(bw, "scbandwidth")
  expect_identical(bw$degree.search$mode, "nomad+powell")
  expect_true(is.finite(bw$nomad.time))
})

test_that("npscoefbw NOMAD degree search fails fast when crs is unavailable", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 24
  xdat <- data.frame(x = runif(n))
  zdat <- data.frame(z = sort(runif(n)))
  y <- (1 + zdat$z) * xdat$x + rnorm(n, sd = 0.08)

  expect_error(
    with_np_degree_bindings(
      list(.np_nomad_require_crs = function() stop("crs missing", call. = FALSE)),
      npscoefbw(
        xdat = xdat,
        zdat = zdat,
        ydat = y,
        regtype = "lp",
        degree.select = "coordinate",
        search.engine = "nomad",
        degree.min = 0L,
        degree.max = 1L,
        bwtype = "fixed",
        bwmethod = "cv.ls",
        nmulti = 1L
      )
    ),
    "crs missing"
  )
})

test_that("npregbw automatic degree search defaults to NOMAD plus Powell", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(x = sort(runif(16)))
  dat$y <- sin(2 * pi * dat$x) + rnorm(nrow(dat), sd = 0.05)

  bw <- np::npregbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_identical(bw$degree.search$mode, "nomad+powell")
  expect_true(isTRUE(bw$degree.search$completed))
})

test_that("npregbw NOMAD degree search fails fast when crs is unavailable", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(x = runif(16), y = rnorm(16))

  expect_error(
    with_np_degree_bindings(
      list(.np_nomad_require_crs = function() stop("crs missing", call. = FALSE)),
      np::npregbw(
        y ~ x,
        data = dat,
        regtype = "lp",
        degree.select = "coordinate",
        search.engine = "nomad",
        degree.min = 0L,
        degree.max = 1L,
        bwtype = "fixed",
        bwmethod = "cv.ls",
        nmulti = 1L
      )
    ),
    "crs missing"
  )
})

test_that("npregbw forwards nomad.nmulti to inner crs::snomadr multistarts", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260323)
  dat <- data.frame(x = sort(runif(16)))
  dat$y <- sin(2 * pi * dat$x) + rnorm(nrow(dat), sd = 0.05)

  assign("np_nomad_inner_nmulti_seen", integer(0), envir = .GlobalEnv)
  on.exit(rm(list = "np_nomad_inner_nmulti_seen", envir = .GlobalEnv), add = TRUE)
  trace(
    what = crs::snomadr,
    tracer = quote({
      .GlobalEnv[["np_nomad_inner_nmulti_seen"]] <- c(
        .GlobalEnv[["np_nomad_inner_nmulti_seen"]],
        as.integer(nmulti)
      )
    }),
    print = FALSE
  )
  on.exit(untrace(crs::snomadr), add = TRUE)

  np::npregbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "nomad",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L,
    nomad.nmulti = 2L
  )

  expect_identical(get("np_nomad_inner_nmulti_seen", envir = .GlobalEnv), 2L)
})

test_that("internal degree search returns best-so-far on interrupt", {
  degree_search <- getFromNamespace(".np_degree_search", "np")

  eval_count <- 0L
  result <- degree_search(
    method = "exhaustive",
    candidates = list(0:1),
    baseline_degree = 0L,
    start_degree = 0L,
    eval_fun = function(degree) {
      eval_count <<- eval_count + 1L
      if (eval_count >= 2L) {
        stop(structure(list(message = "interrupt"), class = c("interrupt", "condition")))
      }
      list(
        objective = as.numeric(sum(degree)),
        payload = list(degree = as.integer(degree)),
        num.feval = 1L
      )
    },
    direction = "min",
    trace_level = "full"
  )

  expect_true(isTRUE(result$interrupted))
  expect_false(isTRUE(result$completed))
  expect_identical(as.integer(result$best$degree), 0L)
  expect_identical(as.integer(result$best_payload$degree), 0L)
})

test_that("restart starts are deterministic and RNG-independent", {
  restart_starts <- getFromNamespace(".np_degree_restart_starts", "np")

  set.seed(1)
  junk1 <- runif(10)
  starts1 <- restart_starts(
    candidates = list(0:2, 0:2),
    restarts = 4L,
    exclude = list(c(0L, 0L))
  )

  set.seed(999)
  junk2 <- runif(10)
  starts2 <- restart_starts(
    candidates = list(0:2, 0:2),
    restarts = 4L,
    exclude = list(c(0L, 0L))
  )

  expect_false(identical(junk1, junk2))
  expect_identical(starts1, starts2)
})

test_that("NOMAD LP degree starts are deterministic, safe, and prefix-stable", {
  build_degree_starts <- getFromNamespace(".np_lp_nomad_build_degree_starts", "np")

  starts3 <- build_degree_starts(
    initial = c(1L, 1L, 1L, 1L),
    lower = rep(0L, 4L),
    upper = rep(10L, 4L),
    basis = "tensor",
    nobs = 100L,
    nmulti = 3L,
    random.seed = 42L
  )
  starts5 <- build_degree_starts(
    initial = c(1L, 1L, 1L, 1L),
    lower = rep(0L, 4L),
    upper = rep(10L, 4L),
    basis = "tensor",
    nobs = 100L,
    nmulti = 5L,
    random.seed = 42L
  )

  expect_identical(starts5[seq_len(3L), , drop = FALSE], starts3)
  expect_identical(as.integer(starts3[1L, ]), c(1L, 1L, 1L, 1L))
  expect_true(all(apply(starts5, 1L, function(d) np:::dim_basis(basis = "tensor", degree = d) <= floor(0.25 * (100L - 1L)))))
})

test_that("NOMAD mixed starts preserve user start 1 and expose prefix-stable restart points", {
  build_starts <- getFromNamespace(".np_nomad_build_starts", "np")

  x0 <- c(1.5, 0.25, 1, 1, 1)
  spec <- list(
    initial = c(1L, 1L, 1L),
    lower = c(0L, 0L, 0L),
    upper = c(10L, 10L, 10L),
    basis = "glp",
    nobs = 80L,
    user_supplied = TRUE
  )

  starts2 <- build_starts(
    x0 = x0,
    bbin = c(0L, 0L, 1L, 1L, 1L),
    lb = c(1e-2, 0, 0, 0, 0),
    ub = c(1e6, 1, 10, 10, 10),
    nmulti = 2L,
    random.seed = 99L,
    degree_spec = spec
  )
  starts4 <- build_starts(
    x0 = x0,
    bbin = c(0L, 0L, 1L, 1L, 1L),
    lb = c(1e-2, 0, 0, 0, 0),
    ub = c(1e6, 1, 10, 10, 10),
    nmulti = 4L,
    random.seed = 99L,
    degree_spec = spec
  )

  expect_equal(starts2[1L, ], x0)
  expect_identical(starts4[seq_len(2L), , drop = FALSE], starts2)
})

test_that("NOMAD Powell hot-start helpers never emit zero public multistarts", {
  hot_nmulti <- getFromNamespace(".np_nomad_powell_hotstart_nmulti", "np")

  expect_identical(hot_nmulti("disable_multistart"), 1L)
  expect_identical(hot_nmulti("single_iteration"), 1L)
})

test_that("NOMAD Powell hot-start option helper controls internal remin explicitly", {
  hot_opt_args <- getFromNamespace(".np_nomad_powell_hotstart_opt_args", "np")

  opt.args <- list(
    nmulti = 7L,
    powell.remin = TRUE,
    ftol = 1e-6,
    tol = 1e-5,
    custom = "preserve-me"
  )

  out <- hot_opt_args(opt.args, strategy = "disable_multistart")

  expect_identical(out$nmulti, 1L)
  expect_false(out$powell.remin)
  expect_identical(out$ftol, opt.args$ftol)
  expect_identical(out$tol, opt.args$tol)
  expect_identical(out$custom, opt.args$custom)

  out2 <- hot_opt_args(opt.args, strategy = "single_iteration")
  expect_identical(out2$nmulti, 1L)
  expect_false(out2$powell.remin)

  out3 <- hot_opt_args(opt.args, strategy = "single_iteration", remin = TRUE)
  expect_identical(out3$nmulti, 1L)
  expect_true(out3$powell.remin)
})

test_that("NOMAD restart summary records remin metadata", {
  attach_summary <- getFromNamespace(".np_attach_nomad_restart_summary", "np")

  bws <- list(fval = 1)
  search_result <- list(
    direction = "min",
    best = list(objective = 0.5),
    best.restart = 2L,
    restart.results = list(
      list(objective = 1.0),
      list(objective = 0.5, remin = TRUE)
    ),
    restart.starts = list(10, 20),
    restart.degree.starts = list(1L, 2L),
    restart.bandwidth.starts = list(0.1, 0.2),
    nomad.remin = TRUE,
    nomad.remin.index = 2L,
    nomad.remin.roundtrip = list(objective = 1.0, degree = 1L)
  )

  out <- attach_summary(bws, search_result)

  expect_true(out$nomad.remin)
  expect_identical(out$nomad.remin.index, 2L)
  expect_identical(out$nomad.remin.roundtrip$degree, 1L)
  expect_identical(out$nomad.best.restart, 2L)
  expect_equal(out$nomad.restart.fval, c(1.0, 0.5))
})

test_that("coordinate search skips incumbent cell revisits within a sweep", {
  degree_search <- getFromNamespace(".np_degree_search", "np")

  result <- degree_search(
    method = "coordinate",
    candidates = list(0:2, 0:2),
    baseline_degree = c(0L, 0L),
    start_degree = c(0L, 0L),
    restarts = 0L,
    max_cycles = 1L,
    eval_fun = function(degree) {
      list(
        objective = as.numeric(sum(degree)),
        payload = list(degree = as.integer(degree)),
        num.feval = 1L
      )
    },
    direction = "min",
    trace_level = "full"
  )

  expect_identical(result$n.unique, 5L)
  expect_identical(result$n.visits, 5L)
  expect_identical(result$n.cached, 0L)
  expect_identical(nrow(result$trace), result$n.unique)
  expect_false(any(result$trace$cached))
})

test_that("automatic exhaustive search emits a safety warning on large grids", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE, np.degree.search.warn.grid = 3L)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(
    x1 = runif(20),
    x2 = runif(20)
  )
  dat$y <- dat$x1 + dat$x2 + rnorm(nrow(dat), sd = 0.05)

  expect_warning(
    np::npregbw(
      y ~ x1 + x2,
      data = dat,
      regtype = "lp",
      degree.select = "exhaustive",
      search.engine = "cell",
      degree.min = 0L,
      degree.max = 1L,
      bwtype = "fixed",
      bwmethod = "cv.ls",
      nmulti = 1L
    ),
    "exhaustive degree search will evaluate 4 degree combinations"
  )
})

test_that("automatic exhaustive search honors internal hard grid limits", {
  old_opts <- options(
    np.messages = FALSE,
    np.tree = FALSE,
    np.degree.search.warn.grid = Inf,
    np.degree.search.max.grid = 3L
  )
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(
    x1 = runif(20),
    x2 = runif(20)
  )
  dat$y <- dat$x1 + dat$x2 + rnorm(nrow(dat), sd = 0.05)

  expect_error(
    np::npregbw(
      y ~ x1 + x2,
      data = dat,
      regtype = "lp",
      degree.select = "exhaustive",
      search.engine = "cell",
      degree.min = 0L,
      degree.max = 1L,
      bwtype = "fixed",
      bwmethod = "cv.ls",
      nmulti = 1L
    ),
    "exceeding the configured limit of 3"
  )
})

test_that("automatic degree search emits staged progress output", {
  old_opts <- options(np.messages = TRUE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(x = sort(runif(20)))
  dat$y <- dat$x + rnorm(nrow(dat), sd = 0.05)

  msgs <- with_np_degree_bindings(
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_renderer_for_surface = function(surface, capability) "legacy",
      .np_progress_now = degree_progress_time_values(seq(0, 20, by = 0.5))
    ),
    capture_degree_messages_only(
      get("npregbw", envir = asNamespace("np"), inherits = FALSE)(
        y ~ x,
        data = dat,
        regtype = "lp",
        degree.select = "exhaustive",
        search.engine = "cell",
        degree.min = 0L,
        degree.max = 1L,
        bwtype = "fixed",
        bwmethod = "cv.ls",
        nmulti = 1L
      )
    )
  )

  expect_true(any(grepl("Automatic polynomial degree search baseline \\(0\\)", msgs)))
  expect_true(any(grepl("Selecting degree and bandwidth", msgs, fixed = TRUE)))
  expect_true(any(grepl("exhaustive", msgs)))
  expect_true(any(grepl("best (", msgs, fixed = TRUE)))

  coord_msgs <- with_np_degree_bindings(
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_renderer_for_surface = function(surface, capability) "legacy",
      .np_progress_now = degree_progress_time_values(seq(0, 40, by = 0.5))
    ),
    capture_degree_messages_only(
      get("npregbw", envir = asNamespace("np"), inherits = FALSE)(
        y ~ x,
        data = dat,
        regtype = "lp",
        degree.select = "coordinate",
        search.engine = "cell",
        degree.min = 0L,
        degree.max = 1L,
        degree.verify = TRUE,
        degree.max.cycles = 2L,
        bwtype = "fixed",
        bwmethod = "cv.ls",
        nmulti = 1L
      )
    )
  )

  expect_true(any(grepl("Coordinate automatic polynomial degree search over 0:1", coord_msgs)))
  expect_true(any(grepl("max 3 search evaluations", coord_msgs)))
  expect_true(any(grepl("Exhaustively certifying automatic polynomial degree search over 2 degree combinations \\(re-optimizing bandwidths\\)", coord_msgs)))
})

bandwidth <-
  function(bw = stop("bandwidth: argument 'bw' missing"),
           bwmethod = c("cv.ml","cv.ls","normal-reference"),
           bwscaling = FALSE,
           bwtype = c("fixed","generalized_nn","adaptive_nn"),
           ckertype = c("gaussian","truncated gaussian","epanechnikov","uniform"), 
           ckerorder = c(2,4,6,8),
           ckerbound = c("none","range","fixed"),
           ckerlb = NULL,
           ckerub = NULL,
           ukertype = c("aitchisonaitken","liracine"),
           okertype = c("liracine","wangvanryzin","racineliyan"),
           fval = NA,
           ifval = NA,
           num.feval = NA,
           num.feval.fast = NA,
           fval.history = NA,
           eval.history = NA,
           invalid.history = NA,
           nobs = NA,
           xdati = stop("bandwidth:argument 'xdati' missing"),
           xnames = character(length(bw)),
           sfactor = NA, bandwidth = NA,
           rows.omit = NA, 
           nconfac = NA, ncatfac = NA, sdev = NA,
           bandwidth.compute = TRUE,
           timing = NA,
           total.time = NA,
           ...){
    
    ndim = length(bw)
    bwmethod = match.arg(bwmethod)
    bwtype = match.arg(bwtype)
    ckertype = match.arg(ckertype)
    ckerbound = match.arg(ckerbound)

    if(missing(ckerorder))
      ckerorder = 2
    else if (ckertype == "uniform")
      .np_warning("ignoring kernel order specified with uniform kernel type")
    else {
      kord = c(2,4,6,8) 
      if (!any(kord == ckerorder))
        stop("ckerorder must be one of ", paste(kord,collapse=" "))
    }

    if (ckertype == "truncated gaussian" && ckerorder != 2)
      .np_warning("using truncated gaussian of order 2, higher orders not yet implemented")

    if (bwmethod == "normal-reference" && (ckertype != "gaussian" || bwtype != "fixed")){    
      .np_warning("normal-reference bandwidth selection assumes gaussian kernel with fixed bandwidth")
      bwtype = "fixed"
      ckertype = "gaussian"
    }

    ukertype = match.arg(ukertype)
    okertype = match.arg(okertype)
    cbounds <- npKernelBoundsResolve(
      dati = xdati,
      varnames = xnames,
      kerbound = ckerbound,
      kerlb = ckerlb,
      kerub = ckerub,
      argprefix = "cker")
    bounded_nonfixed_supported <- bwtype %in% c("generalized_nn", "adaptive_nn")
    if (bwtype != "fixed" &&
        cbounds$bound != "none" &&
        !bounded_nonfixed_supported)
      stop("finite continuous kernel bounds require bwtype = \"fixed\"")

    porder = switch( ckerorder/2, "Second-Order", "Fourth-Order", "Sixth-Order", "Eighth-Order" )

    if (!identical(sfactor,NA)){
      sumNum <- sapply(seq_len(ndim), function(i) {
        if (xdati$icon[i])
          return(sfactor[i])

        if (xdati$iord[i])
          return(oMaxL(xdati$all.nlev[[i]], kertype = okertype))
        
        if (xdati$iuno[i])
          return(uMaxL(xdati$all.nlev[[i]], kertype = ukertype))
      })
    } else {
      sumNum <- NA
    }

    if (length(rows.omit) == 0)
        rows.omit <- NA

    mybw = list(
      bw=bw,
      method = bwmethod,
      pmethod = bwmToPrint(bwmethod),
      fval = fval,
      ifval = ifval,
      num.feval = num.feval,
      num.feval.fast = num.feval.fast,
      fval.history = fval.history,
      eval.history = eval.history,
      invalid.history = invalid.history,
      scaling = bwscaling,
      pscaling = npBandwidthSummaryLabel(bwtype = bwtype, bwscaling = bwscaling),
      type = bwtype,
      ptype = bwtToPrint(bwtype),
      ckertype = ckertype,    
      ckerorder = ckerorder,
      ckerbound = cbounds$bound,
      ckerlb = cbounds$lb,
      ckerub = cbounds$ub,
      pckertype = cktToPrint(ckertype, order = porder, kerbound = cbounds$bound),
      ukertype = ukertype,
      pukertype = uktToPrint(ukertype),
      okertype = okertype,
      pokertype = oktToPrint(okertype, normalized = TRUE),
      nobs = nobs,
      ndim = ndim,
      ncon = sum(xdati$icon),
      nuno = sum(xdati$iuno),
      nord = sum(xdati$iord),
      icon = xdati$icon,
      iuno = xdati$iuno,
      iord = xdati$iord,
      xnames = xnames,
      xdati = xdati,
      sfactor = list(x = sfactor),
      bandwidth = list(x = bandwidth),
      nconfac = nconfac,
      ncatfac = ncatfac,
      sdev = sdev,
      sumNum = list(x = sumNum),
      xmcv = mcvConstruct(xdati),
      dati = list(x = xdati),
      varnames = list(x = xnames),
      vartitle = list(x = ""),
      vartitleabb = list(x = ""),
      rows.omit = rows.omit,
      nobs.omit = if (identical(rows.omit, NA)) 0 else length(rows.omit),
      timing = timing,
      total.time = total.time)

    mybw$klist = list(
      x =
      list(ckertype = ckertype,
           ckerbound = cbounds$bound,
           ckerlb = cbounds$lb,
           ckerub = cbounds$ub,
           pckertype = mybw$pckertype,
           ukertype = ukertype,
           pukertype = mybw$pukertype,
           okertype = okertype,
           pokertype = mybw$pokertype))

    if(!bandwidth.compute)
      mybw$pmethod <- "Manual"
    

    class(mybw) = "bandwidth"
    if(!any(is.na(mybw$bandwidth)))
    validateBandwidth(mybw)
    mybw
    
  }

as.double.bandwidth <- function(x, ...){
  x$bw
}

print.bandwidth <- function(x, digits=NULL, ...){
  cat("\nData (",x$nobs," observations, ",x$ndim," variable(s)):\n\n",sep="")
  print(matrix(x$bw,ncol=x$ndim,dimnames=list(paste(x$pscaling,":",sep=""),x$xnames)))

  cat(genBwSelStr(x))

  cat(genBwKerStrs(x))

  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

summary.bandwidth <- function(object, ...) {
  cat("\nData (",object$nobs," observations, ",object$ndim," variable(s)):\n",sep="")

  cat(genOmitStr(object))
  
  cat(genBwSelStr(object))

  cat('\n')
  cat(genBwScaleStrs(object))

  cat(genBwKerStrs(object))

  cat(genTimingStr(object))
  
  cat("\n\n")
}


predict.bandwidth <- function(object, ...) { npudens(bws = object, ...) }

progress_time_values <- function(values) {
  force(values)
  i <- 0L
  function() {
    i <<- min(i + 1L, length(values))
    values[[i]]
  }
}

shadow_lines <- function(shadow) {
  vapply(shadow$trace, `[[`, character(1L), "line")
}

test_that("bandwidth compaction never abbreviates the Bandwidth selection prefix", {
  compact_bandwidth <- getFromNamespace(".np_progress_compact_bandwidth_line", "np")

  compacted <- compact_bandwidth(
    "[np] Bandwidth selection (multistart 2/2, iteration 84, elapsed 10.0s, 99.9%, eta 0.0s)",
    max_width = 74L
  )

  expect_true(grepl("Bandwidth selection \\(", compacted))
  expect_false(grepl("Bandwidth sel \\(", compacted))
})

test_that("unknown-bound NOMAD restart detail drops synthetic percent and eta", {
  nomad_detail <- getFromNamespace(".np_nomad_progress_detail", "np")

  line <- nomad_detail(
    current_degree = 5L,
    best_record = list(degree = 1L),
    iteration = 77L,
    cumulative_iteration = 1234L,
    restart_index = 2L,
    nmulti = 2L,
    restart_durations = c(18.8),
    elapsed = 18.8
  )

  expect_identical(
    line,
    "multistart 2/2, iteration 77 (1234), elapsed 18.8s, deg (5), best (1)"
  )
  expect_false(grepl("%|eta ", line))
})

test_that("NOMAD progress fuses component context with dynamic fields", {
  nomad_detail <- getFromNamespace(".np_nomad_progress_detail", "np")
  powell_detail <- getFromNamespace(".np_nomad_powell_progress_detail", "np")
  npreg_powell_fields <- getFromNamespace(".npregbw_powell_progress_fields", "np")
  fit_line <- getFromNamespace(".np_progress_fit_single_line", "np")
  set_context <- getFromNamespace(".np_progress_bandwidth_set_context", "np")

  set_context("E[y|z] (1/3)")
  on.exit(set_context(NULL), add = TRUE)

  line <- nomad_detail(
    current_degree = 5L,
    best_record = list(degree = 1L),
    iteration = 77L,
    cumulative_iteration = 1234L,
    restart_index = 2L,
    nmulti = 2L,
    restart_durations = c(18.8),
    elapsed = 18.8
  )

  expect_identical(
    line,
    "E[y|z] (1/3), multistart 2/2, iteration 77 (1234), elapsed 18.8s, deg (5), best (1)"
  )

  compact <- fit_line(
    sprintf("[np] Selecting degree/bandwidth (%s)", line),
    max_width = 80L
  )

  expect_true(startsWith(compact, "[np] Degree/bw search (E[y|z] (1/3),"))
  expect_true(grepl("iter 77", compact, fixed = TRUE))
  expect_true(grepl("elap 18.8s", compact, fixed = TRUE))
  expect_true(grepl("deg (5)", compact, fixed = TRUE))
  expect_false(identical(compact, "[np] E[y|z] (1/3)"))

  powell <- powell_detail(
    current_degree = 5L,
    best_record = list(degree = 1L),
    iteration = 9L,
    elapsed = 2.5
  )

  expect_identical(
    powell,
    "E[y|z] (1/3), elapsed 2.5s, iter 9, deg (5), best (1)"
  )

  expect_identical(
    npreg_powell_fields(
      state = list(started = 0, nomad_current_degree = 5L),
      done = 9L,
      now = 2.5
    ),
    c("E[y|z] (1/3)", "elapsed 2.5s", "degree (5)", "iter 9")
  )
})

test_that("dark-launched bandwidth engine preserves nmulti=1 iteration heartbeats", {
  select_bw <- getFromNamespace(".np_progress_select_bandwidth", "np")
  activity_bw <- getFromNamespace(".np_progress_bandwidth_activity_step", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.progress.bandwidth.enhanced = TRUE,
    np.progress.start.grace.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      value <- select_bw("Selecting density bandwidth", {
        activity_bw(28L)
        activity_bw(64L)
        7
      })
      expect_identical(value, 7)
    },
    now = progress_time_values(c(0, 1.2, 3.4, 5.6))
  )

  lines <- shadow_lines(actual)

  expect_true(any(grepl("^\\[np\\] Bandwidth selection\\.\\.\\.$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(iteration 28, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(iteration 64, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_false(any(grepl("multistart", lines, fixed = TRUE)))
})

test_that("dark-launched bandwidth engine switches from iteration to estimate mode after first completion", {
  select_bw <- getFromNamespace(".np_progress_select_bandwidth", "np")
  set_total <- getFromNamespace(".np_progress_bandwidth_set_total", "np")
  activity_bw <- getFromNamespace(".np_progress_bandwidth_activity_step", "np")
  step_bw <- getFromNamespace(".np_progress_bandwidth_multistart_step", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.progress.bandwidth.enhanced = TRUE,
    np.progress.start.grace.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      value <- select_bw("Selecting density bandwidth", {
        set_total(2L)
        activity_bw(28L)
        step_bw(1L, 2L)
        activity_bw(56L)
        activity_bw(84L)
        11
      })
      expect_identical(value, 11)
    },
    now = progress_time_values(c(0, 1.0, 4.0, 7.0, 10.0, 12.0))
  )

  lines <- shadow_lines(actual)

  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 1/2, iteration 28, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/2, elapsed 4\\.0s, 50\\.0%, eta 4\\.0s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/2, iteration 56, elapsed 7\\.0s, 87\\.5%, eta 1\\.0s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/2, iteration 84, elapsed 10\\.0s, 99\\.9%, eta 0\\.0s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/2, elapsed 12\\.0s, 100\\.0%, eta 0\\.0s\\)$", lines)))
})

test_that("dark-launched completion-estimate heartbeats keep the unknown-total throttle cadence", {
  select_bw <- getFromNamespace(".np_progress_select_bandwidth", "np")
  set_total <- getFromNamespace(".np_progress_bandwidth_set_total", "np")
  activity_bw <- getFromNamespace(".np_progress_bandwidth_activity_step", "np")
  step_bw <- getFromNamespace(".np_progress_bandwidth_multistart_step", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.progress.bandwidth.enhanced = TRUE,
    np.progress.start.grace.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      value <- select_bw("Selecting regression bandwidth", {
        set_total(2L)
        activity_bw(20L)
        step_bw(1L, 2L)
        activity_bw(21L)
        activity_bw(22L)
        13
      })
      expect_identical(value, 13)
    },
    now = progress_time_values(c(0, 1.0, 4.0, 5.0, 6.2, 8.4))
  )

  complete_trace <- actual$trace[grepl("eta ", shadow_lines(actual), fixed = TRUE)]
  complete_lines <- vapply(complete_trace, `[[`, character(1L), "line")
  complete_times <- vapply(complete_trace, `[[`, numeric(1L), "now")

  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/2, elapsed 4\\.0s, 50\\.0%, eta 4\\.0s\\)$", complete_lines)))
  expect_false(any(grepl("elapsed 5\\.0s", complete_lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/2, iteration 22, elapsed 6\\.2s, 77\\.5%, eta 1\\.8s\\)$", complete_lines)))
  expect_true(all(diff(complete_times[seq_len(min(2L, length(complete_times)))]) >= 2.0))
})

test_that("bandwidth progress can carry a coordinator context label such as degree", {
  select_bw <- getFromNamespace(".np_progress_select_bandwidth", "np")
  set_total <- getFromNamespace(".np_progress_bandwidth_set_total", "np")
  activity_bw <- getFromNamespace(".np_progress_bandwidth_activity_step", "np")
  step_bw <- getFromNamespace(".np_progress_bandwidth_multistart_step", "np")
  set_context <- getFromNamespace(".np_progress_bandwidth_set_context", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.progress.bandwidth.enhanced = TRUE,
    np.progress.start.grace.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      set_context("deg (1,0)")
      on.exit(set_context(NULL), add = TRUE)
      value <- select_bw("Selecting regression bandwidth", {
        set_total(2L)
        activity_bw(20L)
        step_bw(1L, 2L)
        activity_bw(21L)
        13
      })
      expect_identical(value, 13)
    },
    now = progress_time_values(c(0, 1.0, 4.0, 6.2, 8.4))
  )

  lines <- shadow_lines(actual)

  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(deg \\(1,0\\)\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(deg \\(1,0\\), multistart 1/2, iteration 20, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(deg \\(1,0\\), multistart 2/2, elapsed [0-9]+\\.[0-9]s, [0-9]+\\.[0-9]%, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(deg \\(1,0\\), multistart 2/2, iteration 21, elapsed [0-9]+\\.[0-9]s, [0-9]+\\.[0-9]%, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(deg \\(1,0\\), multistart 2/2, elapsed [0-9]+\\.[0-9]s, 100\\.0%, eta 0\\.0s\\)$", lines)))
})

.np_round_half_to_even <- function(x) {
  intpart <- trunc(x)
  fracpart <- x - intpart

  if (fracpart < 0.5) {
    intpart
  } else if (fracpart > 0.5) {
    intpart + 1
  } else if ((intpart %% 2) != 0) {
    intpart + 1
  } else {
    intpart
  }
}

.np_sibandwidth_manual_nn_validate <- function(h, nobs, where = "sibandwidth") {
  if (!is.finite(h))
    stop(sprintf("%s: nearest-neighbor bandwidth must be finite", where), call. = FALSE)

  upper <- max(1L, as.integer(nobs) - 1L)
  tol <- sqrt(.Machine$double.eps)
  rounded <- .np_round_half_to_even(h)
  if ((h < 1) || (h > upper) || (abs(h - rounded) > tol)) {
    stop(
      sprintf(
        "%s: nearest-neighbor bandwidth must be an integer in [1, %d]",
        where,
        upper
      ),
      call. = FALSE
    )
  }

  invisible(as.double(rounded))
}

sibandwidth <-
  function(beta, h,
           method=c("ichimura","kleinspady"),
           regtype = c("lc", "ll", "lp"),
           basis = c("glp", "additive", "tensor"),
           degree = NULL,
           bernstein.basis = FALSE,
           bwtype = c("fixed","generalized_nn","adaptive_nn"),
           ckertype = c("gaussian","truncated gaussian","epanechnikov","uniform"), 
           ckerorder = c(2,4,6,8),
           ckerbound = c("none","range","fixed"),
           ckerlb = NULL,
           ckerub = NULL,
           fval = NA,
           ifval = NA,
           num.feval = NA,
           num.feval.fast = NA,
           numimp = NA,
           fval.vector = NA,
           nobs = NA,
           xdati, ydati, idati = untangle(data.frame(double(1))),
           xnames = character(length(beta)),
           ynames = character(1),
           sfactor = NA, bandwidth = NA,
           rows.omit = NA, bandwidth.compute = TRUE,
           optim.method = "NA",
           only.optimize.beta = FALSE,
           total.time = NA,
           ...){

  ndim = length(beta)
  spec <- npCanonicalConditionalRegSpec(
    regtype = regtype,
    basis = basis,
    degree = degree,
    bernstein.basis = bernstein.basis,
    ncon = 1L,
    where = "sibandwidth"
  )
  regtype <- spec$regtype
  basis <- spec$basis
  degree <- spec$degree
  bernstein.basis <- spec$bernstein.basis
  method = match.arg(method)
  ckertype = match.arg(ckertype)
  ckerbound = match.arg(ckerbound)
  bwtype <- match.arg(bwtype)
  
  if(missing(ckerorder))
    ckerorder = 2
  else if (ckertype == "uniform")
    .np_warning("ignoring kernel order specified with uniform kernel type")
  else {
    kord = c(2,4,6,8) 
    if (!any(kord == ckerorder))
      stop("ckerorder must be one of ", paste(kord,collapse=" "))
  }

  if (ckertype == "truncated gaussian" && ckerorder != 2)
    .np_warning("using truncated gaussian of order 2, higher orders not yet implemented")

  porder = switch( ckerorder/2, "Second-Order", "Fourth-Order", "Sixth-Order", "Eighth-Order" )
  cbounds <- npKernelBoundsResolve(
    dati = xdati,
    varnames = xnames,
    kerbound = ckerbound,
    kerlb = ckerlb,
    kerub = ckerub,
    argprefix = "cker")
  bounded_nonfixed_supported <- bwtype %in% c("generalized_nn", "adaptive_nn")
  if (bwtype != "fixed" && cbounds$bound != "none" && !bounded_nonfixed_supported)
    stop("finite continuous kernel bounds require bwtype = \"fixed\"")
  if (bwtype != "fixed" && (!bandwidth.compute || h != 0))
    .np_sibandwidth_manual_nn_validate(h = h, nobs = nobs, where = "sibandwidth")

  sumNum <- sfactor
  ##idati <- NA
  
  if (length(rows.omit) == 0)
    rows.omit <- NA
  
  mybw = list(
    beta=beta,
    bw = h,
    regtype = regtype,
    pregtype = switch(regtype,
      lc = "Local-Constant",
      ll = "Local-Linear",
      lp = "Local-Polynomial"),
    basis = basis,
    degree = degree,
    bernstein.basis = bernstein.basis,
    regtype.engine = spec$regtype.engine,
    basis.engine = spec$basis.engine,
    degree.engine = spec$degree.engine,
    bernstein.basis.engine = spec$bernstein.basis.engine,
    method = method,
    pmethod = switch( method,
      ichimura = "Ichimura",
      "kleinspady" = "Klein and Spady"
      ),
    pomethod = switch(optim.method,
      "Nelder-Mead" = "Nelder-Mead",
      "BFGS" = "BFGS",
      "CG" = "CG", "NA"),
    fval = fval,
    ifval = ifval,
    num.feval = num.feval,
    num.feval.fast = num.feval.fast,
    numimp = numimp,
    fval.vector = fval.vector,
    pscaling = npBandwidthSummaryLabel(bwtype = bwtype),
    type = bwtype,
    ptype = switch( bwtype,
      fixed = "Fixed",
      generalized_nn = "Generalized Nearest Neighbour",
      adaptive_nn = "Adaptive Nearest Neighbour" ),
    ckertype = ckertype,    
    ckerorder = ckerorder,
    ckerbound = cbounds$bound,
    ckerlb = cbounds$lb,
    ckerub = cbounds$ub,
    pckertype = cktToPrint(ckertype, order = porder, kerbound = cbounds$bound),
    nobs = nobs,
    ndim = ndim,
    ncon = sum(xdati$icon),
    nuno = sum(xdati$iuno),
    nord = sum(xdati$iord),
    xdati = xdati,
    ydati = ydati,
    xnames = xnames,
    ynames = ynames,
    sfactor = list(index = sfactor),
    bandwidth = list(index = bandwidth),
    sumNum = list(index = sumNum),
    dati = list(x = xdati, y = ydati, index = idati),
    varnames = list(x = xnames, y = ynames, index = "index"),
    vartitle = list(x = "Explanatory", y = "Dependent", index = "Explanatory"),
    vartitleabb = list(x = "Exp.", y = "Dep.", index = "Exp."),
    rows.omit = rows.omit,
    nobs.omit = if (identical(rows.omit, NA)) 0 else length(rows.omit),
    total.time = total.time)

  mybw$klist <- list(
    index =
    list(ckertype = ckertype,
         ckerbound = cbounds$bound,
         ckerlb = cbounds$lb,
         ckerub = cbounds$ub,
         pckertype = mybw$pckertype))

  if(only.optimize.beta)
    mybw$pmethod <- paste("Pilot (bandwidth) +", mybw$pmethod, "(beta)")
  
  if(!bandwidth.compute)
    mybw$pmethod <- "Manual"

  class(mybw) = "sibandwidth"
  if(!any(is.na(mybw$bandwidth)))
    validateBandwidth(mybw)
  mybw
  
}

print.sibandwidth <- function(x, digits=NULL, ...){
  cat("\nSingle Index Model",
      "\nRegression data (",x$nobs,
      " observations, ",x$ndim," variable(s)):\n\n",sep="")

  print(matrix(x$beta,ncol=x$ndim,dimnames=list(paste("Beta",":",sep=""),x$xnames)))
  cat("Bandwidth: ",x$bw)
  cat(genBwSelStr(x))
  cat(genBwKerStrs(x))

  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

coef.sibandwidth <- function(object, ...) {
 tc <- object$beta
 names(tc) <- object$xnames
 return(tc)
}
predict.sibandwidth <- function(object, ...) { npindex(bws = object, ...) }

summary.sibandwidth <- function(object, ...){
  cat("\nSingle Index Model",
      "\nRegression data (",object$nobs,
      " observations, ",object$ndim," variable(s)):\n\n",sep="")

  print(matrix(object$beta,ncol=object$ndim,dimnames=list(paste("Beta",":",sep=""),object$xnames)))
  cat("Bandwidth: ",object$bw)
  cat("\nOptimisation Method: ", object$pomethod)

  cat(genOmitStr(object))
  cat(genBwSelStr(object))
  cat(genBwKerStrs(object))
  cat(genTimingStr(object))
  cat("\n\n")
}

## Function that implements the bivariate dependence metric described
## in Maasoumi, E.~and J.S.~Racine (2002), "Entropy and Predictability
## of Stock Market Returns," Journal of Econometrics, March, Volume
## 107, Issue 2, pp 291-312.

npdeptest <- function(data.x = NULL,
                      data.y = NULL,
                      method=c("integration","summation"),
                      bootstrap = TRUE,
                      boot.num = 399,
                      random.seed = 42) {
  
  ## Trap fatal errors

  if(is.data.frame(data.x)||is.data.frame(data.y)) stop(" you must enter two data vectors (and not data frames)")
  if(is.factor(data.x)||is.factor(data.y)) stop(" does not support factors")
  if(is.null(data.x)||is.null(data.y)) stop(" you must enter x and y data vectors")
  if(ncol(data.frame(data.x)) != 1) stop(" data must have one dimension only")
  if(length(data.x)!=length(data.y)) stop(" data vectors must be of equal length")
  if(boot.num < 9) stop(" number of bootstrap replications must be >= 9")

  method <- match.arg(method)

  ## Save seed prior to setting

  seed.state <- .np_seed_enter(random.seed)

  .np_progress_note("Computing bandwidths")

  ## If the variable is a time series convert to type numeric

  if(is.ts(data.x)) data.x <- as.numeric(data.x)
  if(is.ts(data.y)) data.y <- as.numeric(data.y)

  ## Remove any NAs from paired data

  tmp <- na.omit(data.frame(data.x,data.y))
  data.x <- tmp$data.x
  data.y <- tmp$data.y
  rm(tmp)

  ## Define the metric entropy function

  Srho.bivar <- function(x.dat = NULL,
                         y.dat = NULL,
                         bw.x = NULL,
                         bw.y = NULL,
                         bw.joint = NULL,
                         method=c("integration","summation")) {
    
    if(is.null(bw.x)||is.null(bw.y)||is.null(bw.joint)) stop(" you must provide numeric bandwidths for f(x), f(y) and f(x,y)")
    
    method <- match.arg(method)
    
    if(method=="summation") {
      
      ## Summation version: \sum_i(1-sqrt(f(x_i)f(y_i)/f(x_i,y_i)))^2
      ## We could code this by hand per below, however, there is no
      ## performance penalty imposed by sum() so no need.
      
      f.x <- fitted(npudens(tdat=x.dat,bws=bw.x))
      f.y <- fitted(npudens(tdat=y.dat,bws=bw.y))
      f.xy <- fitted(npudens(tdat=cbind(x.dat,y.dat),bws=bw.joint))
      summand <- f.x*f.y/f.xy
      
      ## In summation version we divide densities which can lead to
      ## numerical instability issues not present in the integrand
      ## version (which uses differences instead). We check for this
      ## case, remove offending points, and warn. This traps -Inf,
      ## Inf, and NaN.  

      if(!all(is.finite(summand))) {
        .np_warning(" non-finite value in summation-based statistic: integration recommended")
        summand <- summand[is.finite(summand)]
      }

      return(0.5*mean((1-sqrt(summand))**2))
      
    } else {
      
      ## Integration version:
      ## \int\int (sqrt(f(x,y))-sqrt(f(x))sqrt(f(y)))^2 dx dy
      
      lo.default <- c(min(x.dat)-10.0*IQR(x.dat),min(y.dat)-10.0*IQR(y.dat))
      up.default <- c(max(x.dat)+10.0*IQR(x.dat),max(y.dat)+10.0*IQR(y.dat))
      
      Srho.integrand <- function(xy) {
        
        ## We code this up by hand using a second order gaussian
        ## kernel. The multidimensional integration routines for some
        ## reason crawl to almost a halt if we call the functions in the
        ## np package. Good to have noticed this at this juncture, bad
        ## to not have general code though could branch at this point?
        
        ##      f.x <- fitted(npudens(tdat=x.dat,edat=xy[1],bws=bw.x))
        ##      f.y <- fitted(npudens(tdat=y.dat,edat=xy[2],bws=bw.y))
        ##      f.xy <- fitted(npudens(tdat=cbind(x.dat,y.dat),edat=cbind(xy[1],xy[2]),bws=bw.joint))
        
        f.x <- mean(dnorm((xy[1]-x.dat)/bw.x))/bw.x
        f.y <- mean(dnorm((xy[2]-y.dat)/bw.y))/bw.y
        f.xy <- mean(dnorm((xy[1]-x.dat)/bw.joint[1])*dnorm((xy[2]-y.dat)/bw.joint[2])
                    / (bw.joint[1]*bw.joint[2]))
        
        return((sqrt(f.xy)-sqrt(f.x)*sqrt(f.y))**2)
        
      }
      
      return(0.5*adaptIntegrate(Srho.integrand,
                                lowerLimit=lo.default,
                                upperLimit=up.default)$integral)
      
    }

  } ## end of Srho.bivar function
  
  ## Compute and save bandwidths (save for bootstrapping if requested)

  bw.data.x <- .np_progress_with_legacy_suppressed(npudensbw(~data.x))$bw
  bw.data.y <- .np_progress_with_legacy_suppressed(npudensbw(~data.y))$bw
  bw.joint <- .np_progress_with_legacy_suppressed(npudensbw(~data.x+data.y))$bw

  .np_progress_note("Constructing metric entropy")
  
  Srho.vec <- Srho.bivar(data.x,data.y,bw.data.x,bw.data.y,bw.joint,method=method)

  ## Bootstrap if requested - null is independence so simple iid
  ## index resampling under replacement is sufficient

  if(bootstrap) {

    Srho.vec.boot <- numeric()
    progress <- .np_progress_begin("Bootstrap replications", total = boot.num, surface = "bootstrap")

    for (b in seq_len(boot.num)) {
      ## Break systematic relationship between x and y (null)
      
      data.x.boot <- data.x[sample.int(length(data.x), replace = TRUE)]
      
      Srho.vec.boot[b] <- Srho.bivar(data.x.boot,data.y,bw.data.x,bw.data.y,bw.joint,method=method)
      progress <- .np_progress_step(progress, done = b)

    }

    progress <- .np_progress_end(progress)

    ## Compute P-values

    P <- mean(Srho.vec.boot > Srho.vec)

  }

  ## Restore seed

  .np_seed_exit(seed.state)

  if(bootstrap) {

    deptest(Srho = Srho.vec,
            Srho.bootstrap.vec = Srho.vec.boot,
            P = P,
            bootstrap = bootstrap,
            boot.num = boot.num,
            bw.data.x = bw.data.x,
            bw.data.y = bw.data.y,
            bw.joint = bw.joint)

  } else {

    deptest(Srho = Srho.vec,
            Srho.bootstrap.vec = NULL,
            P = NULL,
            bootstrap = bootstrap,
            boot.num = NULL,
            bw.data.x = bw.data.x,
            bw.data.y = bw.data.y,
            bw.joint = bw.joint)

  }

}

npquantile <- function(x=NULL,
                       tau=c(0.01,0.05,0.25,0.50,0.75,0.95,0.99),
                       num.eval=10000,
                       bws=NULL,
                       f=1,
                       ...) {

  ## Some basic error checking.

  if(is.null(x)) stop("must provide data")
#  if(class(x) != "numeric") stop("x must be numeric and univariate")
  if(!isa(x,"numeric")) stop("x must be numeric and univariate")

  if(anyNA(tau)) stop("tau must not contain missing values")
  if(any(tau<0 | tau>1, na.rm = TRUE)) stop("tau must lie in the closed interval [0,1]")
  if(num.eval < 100) stop("num.eval must be >= 100")

  if(is.null(bws)) bws <- npudistbw(~x,...)
#  if(class(bws)!="dbandwidth") stop("bw object must be a npudistbw() object")
  if(!isa(bws,"dbandwidth")) stop("bw object must be a npudistbw() object")
  if(length(bws$xnames)>1) stop("bw object must be univariate")

  ## Create grid from which quasi-inverse is extracted - extend the
  ## range of x for evaluation grid, also add empirical quantiles to
  ## grid.

  x.er <- extendrange(x,f=f)
  x.eval <- na.omit(sort(c(seq(x.er[1],x.er[2],length=num.eval),
                           quantile(x,tau,na.rm=TRUE))))

  F <- fitted(npudist(tdat=x,
                      edat=x.eval,
                      bws=bws))

  ## Now compute the quasi-inverse from the estimated F for the
  ## evaluation points. If tau is input and any value lies beyond the
  ## CDF values for the evaluation points, reset them to the min/max
  ## CDF values for the evaluation data (otherwise the quantiles are
  ## undefined).

  x.tau <- numeric(length(tau))
  
  for (i in seq_along(tau)) {
    tau[tau<min(F)] <- min(F)
    tau[tau>max(F)] <- max(F)        
    x.tau[i] <-  min(x.eval[F>=tau[i]])
  }

  ## Asymptotic standard errors could be passed as an attribute
  
  ## f <- fitted(npudens(tdat=x,
  ##                     edat=x.tau,
  ##                     bws=bws$bw))
  ## asy.var <- tau*(1-tau)/(length(x)*f**2)
  ## attr(x.tau,"asy.var") <- asy.var

  names(x.tau) <- paste(tau*100,"%",sep="")

  return(x.tau)

}

npindexbw <-
  function(...){
    mc <- match.call(expand.dots = FALSE)
    npRejectRenamedScaleFactorSearchArgs(names(mc$...), where = "npindexbw")
    target <- .np_bw_dispatch_target(dots = mc$...,
                                     data_arg_names = c("xdat", "ydat"),
                                     eval_env = parent.frame())
    UseMethod("npindexbw", target)
  }

npindexbw.formula <-
  function(formula, data, subset, na.action, call, ...){
    formula.terms <- terms(formula)
    orig.ts <- if (missing(data))
      .np_terms_ts_mask(terms_obj = formula.terms,
                        data = environment(formula),
                        eval_env = environment(formula))
    else .np_terms_ts_mask(terms_obj = formula.terms,
                           data = data,
                           eval_env = environment(formula))

    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(mf), nomatch = 0)
    mf <- mf[c(1,m)]

    mf[[1]] <- as.name("model.frame")

    if(all(orig.ts)){
      args <- (as.list(attr(formula.terms, "variables"))[-1])
      formula <- formula.terms
      attr(formula, "predvars") <- as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), args))))
      mf[["formula"]] <- formula
    }else if(any(orig.ts)){
      arguments <- (as.list(attr(formula.terms, "variables"))[-1])
      arguments.normal <- arguments[which(!orig.ts)]
      arguments.timeseries <- arguments[which(orig.ts)]

      ix <- sort(c(which(orig.ts),which(!orig.ts)),index.return = TRUE)$ix
      formula <- formula.terms
      attr(formula, "predvars") <- bquote(.(as.call(c(quote(cbind),as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments.timeseries)))),arguments.normal,check.rows = TRUE)))[,.(ix)])
      mf[["formula"]] <- formula
    }
    
    mf.args <- as.list(mf[-1L])
    mf <- do.call(stats::model.frame, mf.args, envir = parent.frame())

    ydat <- model.response(mf)
    xdat <- mf[, attr(attr(mf, "terms"),"term.labels"), drop = FALSE]

    tbw <- do.call(npindexbw, c(list(xdat = xdat, ydat = ydat), list(...)))

    ## clean up (possible) inconsistencies due to recursion ...
    tbw$call <- match.call(expand.dots = FALSE)
    environment(tbw$call) <- parent.frame()
    tbw$formula <- formula
    tbw$rows.omit <- as.vector(attr(mf,"na.action"))
    tbw$nobs.omit <- length(tbw$rows.omit)
    tbw$terms <- attr(mf,"terms")

    tbw <-
      updateBwNameMetadata(nameList =
                           list(ynames =
                                attr(mf, "names")[attr(tbw$terms, "response")]),
                           bws = tbw)

    tbw
  }

npindexbw.NULL <-
  function(xdat = stop("training data xdat missing"),
           ydat = stop("training data ydat missing"),
           bws, ...){

    xdat <- toFrame(xdat)

    bws <- double(ncol(xdat)+1)

    tbw <- npindexbw.default(xdat = xdat,
                             ydat = ydat,
                             bws = bws, ...)

    ## clean up (possible) inconsistencies due to recursion ...
    mc <- match.call(expand.dots = FALSE)
    environment(mc) <- parent.frame()
    tbw$call <- mc

    tbw <- updateBwNameMetadata(nameList =
                                list(ynames = deparse(substitute(ydat))),
                                bws = tbw)

    tbw
  }

.npindex_resolve_spec <- function(source, where = "npindex") {
  if (!is.null(source$regtype.engine)) {
    return(list(
      regtype = if (is.null(source$regtype)) "lc" else as.character(source$regtype),
      basis = if (is.null(source$basis)) "glp" else as.character(source$basis),
      degree = if (is.null(source$degree)) integer(0) else as.integer(source$degree),
      bernstein.basis = isTRUE(source$bernstein.basis),
      regtype.engine = as.character(source$regtype.engine),
      basis.engine = if (is.null(source$basis.engine)) "glp" else as.character(source$basis.engine),
      degree.engine = if (is.null(source$degree.engine)) integer(0) else as.integer(source$degree.engine),
      bernstein.basis.engine = isTRUE(source$bernstein.basis.engine)
    ))
  }

  npCanonicalConditionalRegSpec(
    regtype = if (is.null(source$regtype)) "lc" else as.character(source$regtype),
    basis = if (is.null(source$basis)) "glp" else as.character(source$basis),
    degree = source$degree,
    bernstein.basis = isTRUE(source$bernstein.basis),
    ncon = 1L,
    where = where
  )
}

.npindex_nn_candidate_bandwidth <- function(h, bwtype, nobs) {
  if (identical(bwtype, "fixed")) {
    return(list(ok = is.finite(h) && (h > 0), value = as.double(h)))
  }

  if (!is.finite(h)) {
    return(list(ok = FALSE, value = NA_real_))
  }

  lower <- 2L
  upper <- max(1L, as.integer(nobs) - 1L)
  k <- .np_round_half_to_even(h)
  list(ok = (k >= lower) && (k <= upper), value = as.double(k))
}

.npindexbw_h_start_controls <- function(scale.factor.init.lower = 0.1,
                                        scale.factor.init.upper = 2.0,
                                        scale.factor.init = 0.5,
                                        scale.factor.search.lower = 0,
                                        where = "npindexbw") {
  cont.start <- npContinuousSearchStartControls(
    scale.factor.init.lower,
    scale.factor.init.upper,
    scale.factor.init,
    scale.factor.search.lower,
    where = where
  )
  cont.start$scale.factor.search.lower <- as.double(scale.factor.search.lower)
  cont.start
}

.npindex_start_bandwidth_scale <- function(fit, nobs) {
  EssDee(fit) * nobs^(-1 / 5)
}

.npindex_default_start_bandwidth <- function(fit,
                                             bwtype,
                                             nobs,
                                             start.controls = .npindexbw_h_start_controls()) {
  if (identical(bwtype, "fixed")) {
    return(start.controls$scale.factor.init * .npindex_start_bandwidth_scale(fit = fit, nobs = nobs))
  }

  lower <- 2L
  max(lower, min(max(1L, as.integer(nobs) - 1L), .np_round_half_to_even(sqrt(nobs))))
}

.npindex_random_start_bandwidth <- function(fit,
                                            bwtype,
                                            nobs,
                                            start.controls = .npindexbw_h_start_controls()) {
  if (identical(bwtype, "fixed")) {
    return(runif(1, min = start.controls$scale.factor.init.lower, max = start.controls$scale.factor.init.upper) *
             .npindex_start_bandwidth_scale(fit = fit, nobs = nobs))
  }

  upper <- max(1L, as.integer(nobs) - 1L)
  runif(1, min = 2, max = max(2L, upper))
}

.npindex_finalize_bandwidth <- function(h,
                                        bwtype,
                                        nobs,
                                        lower = NULL,
                                        where = "npindexbw") {
  candidate <- .npindex_nn_candidate_bandwidth(h = h, bwtype = bwtype, nobs = nobs)
  if (!candidate$ok) {
    if (identical(bwtype, "fixed")) {
      stop(sprintf("%s: bandwidth must be positive and finite", where), call. = FALSE)
    }
    stop(
      sprintf(
        "%s: nearest-neighbor bandwidth candidate must map to an integer in [2, %d]",
        where,
        max(2L, as.integer(nobs) - 1L)
      ),
      call. = FALSE
    )
  }

  if (identical(bwtype, "fixed") && !is.null(lower) && candidate$value < lower) {
    stop(sprintf("%s: bandwidth is below the continuous scale-factor lower bound", where),
         call. = FALSE)
  }

  candidate$value
}

.npindexbw_nomad_fixed_h_scale <- function(fit,
                                           h.start.raw,
                                           nobs,
                                           start.controls = .npindexbw_h_start_controls()) {
  scale <- .npindex_default_start_bandwidth(
    fit = fit,
    bwtype = "fixed",
    nobs = nobs,
    start.controls = start.controls
  )
  if (!is.finite(scale) || scale <= 0) {
    scale <- max(
      if (is.finite(h.start.raw)) abs(as.double(h.start.raw)) else 0,
      1e-3
    )
  }
  as.double(scale)
}

.npindexbw_nomad_fixed_start_setup <- function(xmat,
                                               ydat,
                                               baseline.bws,
                                               degree.search,
                                               nmulti,
                                               random.seed,
                                               h.start.controls = .npindexbw_h_start_controls()) {
  p <- ncol(xmat)
  nobs <- nrow(xmat)
  beta.free <- if (p > 1L) seq_len(p - 1L) else integer(0)
  h.col <- length(beta.free) + 1L
  degree.col <- h.col + 1L

  ols.fit <- lm(ydat ~ xmat, x = TRUE)
  fit.proxy <- as.double(fitted(ols.fit))
  fit.proxy[!is.finite(fit.proxy)] <- mean(ydat[is.finite(ydat)])
  fit.proxy[!is.finite(fit.proxy)] <- 0

  ols.beta <- if (length(beta.free)) {
    as.double(coef(ols.fit)[3:ncol(ols.fit$x)])
  } else {
    numeric(0)
  }
  if (length(ols.beta))
    ols.beta[!is.finite(ols.beta)] <- 0

  beta.start.raw <- if (length(beta.free)) {
    beta.user <- as.double(baseline.bws$beta[beta.free + 1L])
    if (setequal(beta.user, c(0))) ols.beta else beta.user
  } else {
    numeric(0)
  }
  if (length(beta.start.raw))
    beta.start.raw[!is.finite(beta.start.raw)] <- 0

  if (isTRUE(all.equal(as.double(baseline.bws$bw[1L]), 0))) {
    h.start.raw <- .npindex_default_start_bandwidth(
      fit = fit.proxy,
      bwtype = "fixed",
      nobs = nobs,
      start.controls = h.start.controls
    )
  } else {
    h.lower.raw <- h.start.controls$scale.factor.search.lower *
      .npindex_start_bandwidth_scale(fit = fit.proxy, nobs = nobs)
    h.start.raw <- tryCatch(
      .npindex_finalize_bandwidth(
        h = baseline.bws$bw[1L],
        bwtype = "fixed",
        nobs = nobs,
        lower = h.lower.raw,
        where = "npindexbw"
      ),
      error = function(e) .npindex_default_start_bandwidth(
        fit = fit.proxy,
        bwtype = "fixed",
        nobs = nobs,
        start.controls = h.start.controls
      )
    )
  }
  if (!is.finite(h.start.raw) || h.start.raw <= 0)
    h.start.raw <- 1e-3

  h.scale <- .npindexbw_nomad_fixed_h_scale(
    fit = fit.proxy,
    h.start.raw = h.start.raw,
    nobs = nobs,
    start.controls = h.start.controls
  )
  degree.starts <- .np_lp_nomad_build_degree_starts(
    initial = degree.search$start.degree,
    lower = degree.search$lower,
    upper = degree.search$upper,
    basis = degree.search$basis,
    nobs = degree.search$nobs,
    nmulti = nmulti,
    random.seed = random.seed,
    user_supplied = isTRUE(degree.search$start.user)
  )

  start_matrix.raw <- matrix(0, nrow = nmulti, ncol = degree.col)
  if (length(beta.free))
    start_matrix.raw[1L, beta.free] <- beta.start.raw
  start_matrix.raw[1L, h.col] <- h.start.raw
  start_matrix.raw[, degree.col] <- as.integer(degree.starts[, 1L])

  if (nmulti > 1L) {
    seed.state <- .np_seed_enter(random.seed)
    on.exit(.np_seed_exit(seed.state, remove_if_absent = TRUE), add = TRUE)

    for (j in 2:nmulti) {
      if (length(beta.free)) {
        beta.rand <- runif(length(ols.beta), min = 0.5, max = 1.5) * ols.beta
        beta.rand[!is.finite(beta.rand)] <- 0
        start_matrix.raw[j, beta.free] <- beta.rand
      }

      h.rand <- .npindex_random_start_bandwidth(
        fit = fit.proxy,
        bwtype = "fixed",
        nobs = nobs,
        start.controls = h.start.controls
      )
      if (!is.finite(h.rand) || h.rand <= 0)
        h.rand <- h.start.raw
      start_matrix.raw[j, h.col] <- h.rand
    }
  }

  start_matrix.point <- start_matrix.raw
  start_matrix.point[, h.col] <- start_matrix.point[, h.col] / h.scale

  list(
    beta.free = beta.free,
    ols.beta = ols.beta,
    h.col = h.col,
    degree.col = degree.col,
    h.scale = h.scale,
    start_matrix.raw = start_matrix.raw,
    start_matrix.point = start_matrix.point
  )
}

.npindexbw_fast_eligible <- function(h, bws, eval.index) {
  if (!identical(bws$type, "fixed"))
    return(FALSE)

  fast_largeh_tol <- getOption("np.largeh.rel.tol", 1e-3)
  if (!is.numeric(fast_largeh_tol) || length(fast_largeh_tol) != 1L ||
      is.na(fast_largeh_tol) || !is.finite(fast_largeh_tol) ||
      fast_largeh_tol <= 0 || fast_largeh_tol >= 0.1)
    fast_largeh_tol <- 1e-3

  cont_utol <- switch(
    bws$ckertype,
    gaussian = sqrt(-2.0 * log(1.0 - fast_largeh_tol)),
    "truncated gaussian" = sqrt(-2.0 * log(1.0 - fast_largeh_tol)),
    epanechnikov = sqrt(fast_largeh_tol),
    uniform = 1.0 - 32.0 * .Machine$double.eps,
    0.0
  )

  if (!is.finite(cont_utol) || cont_utol <= 0 || !is.finite(h) || h <= 0)
    return(FALSE)

  vals <- as.double(eval.index)
  vals <- vals[is.finite(vals)]
  if (!length(vals))
    return(FALSE)

  h >= (diff(range(vals)) / cont_utol)
}

.npindex_lp_loo_fit <- function(index,
                                ydat,
                                h,
                                bws,
                                spec,
                                chunk.size = 128L) {
  index <- as.double(index)
  ydat <- as.double(ydat)
  n <- length(index)
  if (length(ydat) != n)
    stop("index and ydat must have the same length")

  index.df <- data.frame(index = index)
  kbw <- kbandwidth(
    bw = c(h),
    bwtype = bws$type,
    ckertype = bws$ckertype,
    ckerorder = bws$ckerorder,
    ckerbound = bws$ckerbound,
    ckerlb = bws$ckerlb,
    ckerub = bws$ckerub,
    nobs = n,
    xdati = untangle(index.df),
    ydati = untangle(data.frame(ydat)),
    xnames = "index",
    ynames = bws$ynames
  )
  degree <- as.integer(spec$degree.engine)
  W <- W.lp(
    xdat = index.df,
    degree = degree,
    basis = spec$basis.engine,
    bernstein.basis = spec$bernstein.basis.engine
  )
  W.eval <- W.lp(
    xdat = index.df,
    exdat = index.df,
    degree = degree,
    basis = spec$basis.engine,
    bernstein.basis = spec$bernstein.basis.engine
  )
  if (!is.matrix(W))
    W <- matrix(W, nrow = n)
  if (!is.matrix(W.eval))
    W.eval <- matrix(W.eval, nrow = n)

  ridge.grid <- npRidgeSequenceFromBase(n.train = n, ridge.base = 0.0, cap = 1.0)
  fit.loo <- double(n)
  chunk.size <- max(1L, npValidatePositiveInteger(chunk.size, "chunk.size"))

  for (start in seq.int(1L, n, by = chunk.size)) {
    idx <- seq.int(start, min(n, start + chunk.size - 1L))
    kw <- .np_kernel_weights_direct(
      bws = kbw,
      txdat = index.df,
      exdat = index.df[idx, , drop = FALSE],
      bandwidth.divide = TRUE,
      kernel.pow = 1.0
    )

    kw[cbind(idx, seq_along(idx))] <- 0.0
    W.eval.block <- W.eval[idx, , drop = FALSE]

    for (jj in seq_along(idx)) {
      w <- kw[, jj]
      XtWy0 <- sum(w * ydat)
      lc.mean <- XtWy0 / NZD(sum(w))
      A <- crossprod(W, W * w)
      z <- crossprod(W, ydat * w)
      solved <- FALSE

      for (ridge in ridge.grid) {
        Ar <- A
        zr <- z
        if (ridge > 0) {
          diag(Ar) <- diag(Ar) + ridge
          zr[1L] <- XtWy0 + ridge * XtWy0 / NZD(A[1L, 1L])
        }

        coef <- tryCatch(chol2inv(chol(Ar)) %*% zr, error = function(e) NULL)
        if (!is.null(coef) && all(is.finite(coef))) {
          alpha <- min(1.0, ridge)
          fit.loo[idx[jj]] <- (1.0 - alpha) * drop(W.eval.block[jj, , drop = FALSE] %*% coef) +
            alpha * lc.mean
          solved <- TRUE
          break
        }
      }

      if (!solved)
        fit.loo[idx[jj]] <- lc.mean
    }
  }

  fit.loo
}

.npindexbw_build_sibandwidth <- function(xdat,
                                         ydat,
                                         bws,
                                         template,
                                         bandwidth.compute,
                                         reg.args) {
  p <- ncol(xdat)
  out <- do.call(
    sibandwidth,
    c(
      list(
        beta = bws[seq_len(p)],
        h = bws[p + 1L],
        method = template$method,
        ckertype = template$ckertype,
        ckerorder = template$ckerorder,
        ckerbound = template$ckerbound,
        ckerlb = template$ckerlb,
        ckerub = template$ckerub,
        bwtype = template$type,
        nobs = nrow(xdat),
        xdati = untangle(xdat),
        ydati = untangle(data.frame(ydat)),
        xnames = names(xdat),
        ynames = template$ynames,
        bandwidth = bws[p + 1L],
        bandwidth.compute = bandwidth.compute
      ),
      reg.args
    )
  )
  if (!is.null(reg.args$scale.factor.search.lower))
    out$scale.factor.search.lower <- npResolveScaleFactorLowerBound(
      reg.args$scale.factor.search.lower
    )
  out
}

.npindexbw_run_fixed_degree <- function(xdat, ydat, bws, template, reg.args, opt.args) {
  tbw <- .npindexbw_build_sibandwidth(
    xdat = xdat,
    ydat = ydat,
    bws = bws,
    template = template,
    bandwidth.compute = opt.args$bandwidth.compute,
    reg.args = reg.args
  )

  do.call(npindexbw.sibandwidth, c(list(xdat = xdat, ydat = ydat, bws = tbw), opt.args))
}

.npindexbw_eval_objective <- function(param,
                                      xmat,
                                      ydat,
                                      bws,
                                      spec) {
  p <- ncol(xmat)
  beta.idx <- if (p > 1L) seq_len(p - 1L) else integer(0)
  beta <- if (length(beta.idx)) as.double(param[beta.idx]) else numeric(0)
  h <- as.double(param[p])
  nobs <- nrow(xmat)

  if (identical(bws$method, "ichimura")) {
    invalid.penalty <- 10 * mean(ydat^2)
  } else {
    invalid.penalty <- sqrt(.Machine$double.xmax)
  }

  h.candidate <- .npindex_nn_candidate_bandwidth(h = h, bwtype = bws$type, nobs = nobs)
  if (!h.candidate$ok)
    return(list(objective = invalid.penalty, num.feval.fast = 0L))
  h <- h.candidate$value

  index <- xmat %*% c(1, beta)
  wmat <- cbind(ydat, 1.0)

  fit.loo <- if (identical(spec$regtype.engine, "lc")) {
    tww <- tryCatch(
      npksum(
        txdat = index,
        tydat = wmat,
        weights = wmat,
        leave.one.out = TRUE,
        bandwidth.divide = TRUE,
        bws = c(h),
        bwtype = bws$type,
        ckertype = bws$ckertype,
        ckerorder = bws$ckerorder,
        ckerbound = bws$ckerbound,
        ckerlb = bws$ckerlb,
        ckerub = bws$ckerub
      )$ksum,
      error = function(e) NULL
    )
    if (is.null(tww))
      return(list(objective = invalid.penalty, num.feval.fast = 0L))
    tww[1, 2, ] / NZD(tww[2, 2, ])
  } else {
    ok.design <- tryCatch({
      npCheckRegressionDesignCondition(
        reg.code = REGTYPE_LP,
        xcon = data.frame(index = index),
        basis = spec$basis.engine,
        degree = spec$degree.engine,
        bernstein.basis = spec$bernstein.basis.engine,
        where = "npindexbw"
      )
      TRUE
    }, error = function(e) FALSE)

    if (!ok.design)
      return(list(objective = invalid.penalty, num.feval.fast = 0L))

    .npindex_lp_loo_fit(
      index = index,
      ydat = ydat,
      h = h,
      bws = bws,
      spec = spec
    )
  }

  if (any(!is.finite(fit.loo)))
    return(list(objective = invalid.penalty, num.feval.fast = 0L))

  if (identical(bws$method, "ichimura")) {
    return(list(
      objective = mean((ydat - fit.loo)^2),
      num.feval.fast = if (.npindexbw_fast_eligible(h = h, bws = bws, eval.index = index)) 1L else 0L
    ))
  }

  floor <- sqrt(.Machine$double.eps)
  fit.loo[fit.loo < floor] <- floor
  fit.loo[fit.loo > 1 - floor] <- 1 - floor
  list(
    objective = -mean(ydat * log(fit.loo) + (1.0 - ydat) * log1p(-fit.loo)),
    num.feval.fast = if (.npindexbw_fast_eligible(h = h, bws = bws, eval.index = index)) 1L else 0L
  )
}

.npindexbw_nomad_search <- function(xdat,
                                    ydat,
                                    bws,
                                    template,
                                    reg.args,
                                    opt.args,
                                    degree.search,
                                    nomad.inner.nmulti = 0L) {
  if (isTRUE(degree.search$verify))
    stop("automatic degree search with search.engine='nomad' does not support degree.verify")

  template.reg.args <- reg.args
  template.reg.args$regtype <- "lp"
  template.reg.args$degree <- as.integer(degree.search$start.degree)
  template.reg.args$bernstein.basis <- degree.search$bernstein.basis
  template.reg.args$regtype.engine <- "lp"
  template.reg.args$degree.engine <- as.integer(degree.search$start.degree)
  template.reg.args$bernstein.basis.engine <- degree.search$bernstein.basis

  baseline.bws <- .npindexbw_build_sibandwidth(
    xdat = xdat,
    ydat = ydat,
    bws = bws,
    template = template,
    bandwidth.compute = FALSE,
    reg.args = template.reg.args
  )

  keep.rows <- rep_len(TRUE, nrow(xdat))
  rows.omit <- attr(na.omit(data.frame(xdat, ydat)), "na.action")
  if (length(rows.omit) > 0L)
    keep.rows[as.integer(rows.omit)] <- FALSE
  x.clean <- toMatrix(xdat[keep.rows, , drop = FALSE])
  y.clean <- ydat[keep.rows]
  if (is.factor(y.clean))
    y.clean <- dlev(y.clean)[as.integer(y.clean)]
  else
    y.clean <- as.double(y.clean)

  p <- ncol(x.clean)
  nomad.nmulti <- if (is.null(opt.args$nmulti)) npDefaultNmulti(ncol(xdat)) else npValidateNmulti(opt.args$nmulti[1L])
  scale.factor.search.lower <- npResolveScaleFactorLowerBound(opt.args$scale.factor.search.lower)
  h.start.controls <- .npindexbw_h_start_controls(
    scale.factor.init.lower = if (is.null(opt.args$scale.factor.init.lower)) 0.1 else opt.args$scale.factor.init.lower,
    scale.factor.init.upper = if (is.null(opt.args$scale.factor.init.upper)) 2.0 else opt.args$scale.factor.init.upper,
    scale.factor.init = if (is.null(opt.args$scale.factor.init)) 0.5 else opt.args$scale.factor.init,
    scale.factor.search.lower = scale.factor.search.lower,
    where = "npindexbw"
  )
  fixed.nomad <- identical(baseline.bws$type, "fixed")
  fixed.setup <- if (fixed.nomad) {
    .npindexbw_nomad_fixed_start_setup(
      xmat = x.clean,
      ydat = y.clean,
      baseline.bws = baseline.bws,
      degree.search = degree.search,
      nmulti = nomad.nmulti,
      random.seed = if (!is.null(opt.args$random.seed)) opt.args$random.seed else 42L,
      h.start.controls = h.start.controls
    )
  } else {
    NULL
  }
  beta.free <- if (fixed.nomad) fixed.setup$beta.free else if (p > 1L) seq_len(p - 1L) else integer(0)
  beta.start <- if (fixed.nomad) {
    if (length(beta.free)) as.double(fixed.setup$start_matrix.raw[1L, beta.free]) else numeric(0)
  } else {
    if (length(beta.free)) as.double(baseline.bws$beta[beta.free + 1L]) else numeric(0)
  }
  h.start.raw <- if (fixed.nomad) {
    as.double(fixed.setup$start_matrix.raw[1L, fixed.setup$h.col])
  } else {
    as.double(baseline.bws$bw[1L])
  }
  beta.lower <- if (length(beta.start)) {
    if (fixed.nomad) {
      pmin(0.5 * fixed.setup$ols.beta, 1.5 * fixed.setup$ols.beta)
    } else {
      -pmax(10, 10 * abs(beta.start))
    }
  } else {
    numeric(0)
  }
  beta.upper <- if (length(beta.start)) {
    if (fixed.nomad) {
      pmax(0.5 * fixed.setup$ols.beta, 1.5 * fixed.setup$ols.beta)
    } else {
      pmax(10, 10 * abs(beta.start))
    }
  } else {
    numeric(0)
  }
  h.lower.raw <- if (fixed.nomad) 1e-3 else 2
  h.upper.raw <- if (fixed.nomad) {
    max(1e6, abs(h.start.raw) * 1e3, 1)
  } else {
    max(2L, as.integer(nrow(x.clean)) - 1L)
  }
  gamma.start <- if (fixed.nomad) fixed.setup$start_matrix.point[1L, fixed.setup$h.col] else NA_real_
  h.lower <- if (fixed.nomad) {
    h.start.controls$scale.factor.search.lower
  } else {
    h.lower.raw
  }
  h.upper <- if (fixed.nomad) {
    h.start.controls$scale.factor.init.upper
  } else {
    h.upper.raw
  }
  h.integer <- !identical(baseline.bws$type, "fixed")

  x0 <- if (fixed.nomad) {
    as.numeric(fixed.setup$start_matrix.point[1L, ])
  } else {
    c(beta.start, h.start.raw, as.integer(degree.search$start.degree))
  }
  lb <- c(beta.lower, h.lower, degree.search$lower)
  ub <- c(beta.upper, h.upper, degree.search$upper)
  bbin <- c(rep.int(0L, length(beta.start)), if (isTRUE(h.integer)) 1L else 0L, 1L)
  baseline.record <- NULL
  nomad.num.feval.total <- 0
  nomad.num.feval.fast.total <- 0

  .np_nomad_baseline_note(degree.search$start.degree)

  point_h_to_raw <- function(h.point) {
    h.raw <- if (fixed.nomad) as.double(h.point) * fixed.setup$h.scale else as.double(h.point)
    as.double(h.raw)
  }

  point_to_public <- function(point) {
    point <- as.numeric(point)
    if (length(point) >= (length(beta.free) + 1L))
      point[length(beta.free) + 1L] <- point_h_to_raw(point[length(beta.free) + 1L])
    point
  }

  bandwidth_point_to_public <- function(point) {
    point <- as.numeric(point)
    if (length(point))
      point[length(point)] <- point_h_to_raw(point[length(point)])
    point
  }

  point_to_param <- function(point) {
    beta.tail <- if (length(beta.free)) as.double(point[seq_along(beta.free)]) else numeric(0)
    h <- point_h_to_raw(point[length(beta.free) + 1L])
    h <- .npindex_finalize_bandwidth(
      h = h,
      bwtype = baseline.bws$type,
      nobs = nrow(x.clean),
      lower = if (fixed.nomad) h.start.controls$scale.factor.search.lower * fixed.setup$h.scale else NULL,
      where = "npindexbw"
    )
    c(beta.tail, h)
  }

  point_to_bws <- function(point) {
    param <- point_to_param(point)
    beta.tail <- if (length(beta.free)) param[seq_along(beta.free)] else numeric(0)
    h <- param[length(beta.free) + 1L]
    c(1.0, beta.tail, h)
  }

  eval_fun <- function(point) {
    point <- as.numeric(point)
    degree <- .np_degree_clip_to_grid(
      as.integer(round(point[length(point)])),
      degree.search$candidates
    )
    eval.spec <- reg.args
    eval.spec$regtype.engine <- "lp"
    eval.spec$degree.engine <- degree
    eval.spec$bernstein.basis.engine <- degree.search$bernstein.basis
    eval.spec$basis.engine <- reg.args$basis.engine
    objective <- .npindexbw_eval_objective(
      param = point_to_param(point),
      xmat = x.clean,
      ydat = y.clean,
      bws = baseline.bws,
      spec = eval.spec
    )
    nomad.num.feval.total <<- nomad.num.feval.total + 1L
    nomad.num.feval.fast.total <<- nomad.num.feval.fast.total + as.numeric(objective$num.feval.fast[1L])
    list(
      objective = as.numeric(objective$objective[1L]),
      degree = as.integer(degree),
      num.feval = 1,
      num.feval.fast = as.numeric(objective$num.feval.fast[1L])
    )
  }

  build_payload <- function(point, best_record, solution, interrupted) {
    point <- as.numeric(point)
    degree <- as.integer(best_record$degree)
    bw.vec <- point_to_bws(point)
    powell.elapsed <- NA_real_

    build_direct_payload <- function() {
      final.reg.args <- reg.args
      final.reg.args$regtype <- "lp"
      final.reg.args$degree <- degree
      final.reg.args$bernstein.basis <- degree.search$bernstein.basis
      final.reg.args$regtype.engine <- "lp"
      final.reg.args$degree.engine <- degree
      final.reg.args$bernstein.basis.engine <- degree.search$bernstein.basis

      tbw <- .npindexbw_build_sibandwidth(
        xdat = xdat,
        ydat = ydat,
        bws = bw.vec,
        template = baseline.bws,
        bandwidth.compute = FALSE,
        reg.args = final.reg.args
      )
      tbw$fval <- as.numeric(best_record$objective)
      tbw$num.feval <- if (!is.null(solution$bbe)) as.numeric(solution$bbe) else as.numeric(best_record$num.feval)
      tbw$num.feval.fast <- as.numeric(nomad.num.feval.fast.total)
      tbw$total.time <- NA_real_
      payload <- npindexbw.sibandwidth(
        xdat = xdat,
        ydat = ydat,
        bws = tbw,
        bandwidth.compute = FALSE
      )
      if (!is.null(payload$method) && length(payload$method))
        payload$pmethod <- bwmToPrint(as.character(payload$method[1L]))
      payload
    }

    direct.payload <- build_direct_payload()
    direct.objective <- as.numeric(best_record$objective)

    if (identical(degree.search$engine, "nomad+powell")) {
      hot.reg.args <- reg.args
      hot.reg.args$regtype <- "lp"
      hot.reg.args$degree <- degree
      hot.reg.args$bernstein.basis <- degree.search$bernstein.basis
      hot.reg.args$regtype.engine <- "lp"
      hot.reg.args$degree.engine <- degree
      hot.reg.args$bernstein.basis.engine <- degree.search$bernstein.basis
      hot.opt.args <- .np_nomad_powell_hotstart_opt_args(
        opt.args,
        strategy = "single_iteration",
        remin = isTRUE(opt.args$powell.remin)
      )
      powell.start <- proc.time()[3L]
      hot.payload <- .np_nomad_with_powell_progress(
        degree = degree,
        best_record = best_record,
        expr = .npindexbw_run_fixed_degree(
          xdat = xdat,
          ydat = ydat,
          bws = bw.vec,
          template = template,
          reg.args = hot.reg.args,
          opt.args = hot.opt.args
        )
      )
      powell.elapsed <- proc.time()[3L] - powell.start
      direct.payload$num.feval <- as.numeric(nomad.num.feval.total) + as.numeric(hot.payload$num.feval[1L])
      direct.payload$num.feval.fast <- as.numeric(nomad.num.feval.fast.total) + as.numeric(hot.payload$num.feval.fast[1L])
      hot.payload$num.feval <- direct.payload$num.feval
      hot.payload$num.feval.fast <- direct.payload$num.feval.fast
      if (!is.null(hot.payload$method) && length(hot.payload$method))
        hot.payload$pmethod <- bwmToPrint(as.character(hot.payload$method[1L]))
      hot.objective <- as.numeric(hot.payload$fval[1L])
      if (is.finite(hot.objective) &&
          .np_degree_better(hot.objective, direct.objective, direction = "min")) {
        return(list(payload = hot.payload, objective = hot.objective, powell.time = powell.elapsed))
      }
    }

    list(payload = direct.payload, objective = direct.objective, powell.time = powell.elapsed)
  }

  search.result <- .np_nomad_search(
    engine = degree.search$engine,
    baseline_record = baseline.record,
    start_degree = degree.search$start.degree,
    x0 = x0,
    bbin = bbin,
    lb = lb,
    ub = ub,
    eval_fun = eval_fun,
    build_payload = build_payload,
    direction = "min",
    objective_name = "fval",
    nmulti = nomad.nmulti,
    nomad.inner.nmulti = nomad.inner.nmulti,
    random.seed = if (!is.null(opt.args$random.seed)) opt.args$random.seed else 42L,
    remin = isTRUE(opt.args$nomad.remin),
    degree_spec = list(
      initial = degree.search$start.degree,
      lower = degree.search$lower,
      upper = degree.search$upper,
      basis = degree.search$basis,
      nobs = degree.search$nobs,
      user_supplied = degree.search$start.user
    )
  )

  if (fixed.nomad) {
    if (!is.null(search.result$restart.starts)) {
      search.result$restart.starts <- lapply(search.result$restart.starts, point_to_public)
    }
    if (!is.null(search.result$restart.bandwidth.starts)) {
      search.result$restart.bandwidth.starts <- lapply(search.result$restart.bandwidth.starts, bandwidth_point_to_public)
    }
    if (!is.null(search.result$best_point) && length(search.result$best_point)) {
      search.result$best_point <- point_to_public(search.result$best_point)
    }
    if (!is.null(search.result$restart.results) && length(search.result$restart.results)) {
      search.result$restart.results <- lapply(search.result$restart.results, function(rec) {
        if (!is.null(rec$start) && length(rec$start))
          rec$start <- point_to_public(rec$start)
        if (!is.null(rec$solution) && length(rec$solution))
          rec$solution <- point_to_public(rec$solution)
        rec
      })
    }
  }

  search.result
}

.npindexbw_degree_search_controls <- function(regtype,
                                              regtype.named,
                                              bandwidth.compute,
                                              nobs,
                                              basis,
                                              degree.select,
                                              search.engine,
                                              degree.min,
                                              degree.max,
                                              degree.start,
                                              degree.restarts,
                                              degree.max.cycles,
                                              degree.verify,
                                              bernstein.basis,
                                              bernstein.named) {
  degree.select <- match.arg(degree.select, c("manual", "coordinate", "exhaustive"))
  if (identical(degree.select, "manual"))
    return(NULL)
  search.engine <- .np_degree_search_engine_controls(search.engine)

  regtype.requested <- if (isTRUE(regtype.named)) match.arg(regtype, c("lc", "ll", "lp")) else "lc"
  if (!identical(regtype.requested, "lp"))
    stop("automatic degree search currently requires regtype='lp'")
  if (!isTRUE(bandwidth.compute))
    stop("automatic degree search requires bandwidth.compute=TRUE")

  bern.auto <- if (isTRUE(bernstein.named)) bernstein.basis else TRUE
  bern.auto <- npValidateGlpBernstein(regtype = "lp", bernstein.basis = bern.auto)

  bounds <- .np_degree_normalize_bounds(
    ncon = 1L,
    degree.min = degree.min,
    degree.max = degree.max,
    default.max = 3L
  )

  baseline.degree <- 0L
  default.start.degree <- if (identical(search.engine, "cell")) baseline.degree else 1L
  start.degree <- if (is.null(degree.start)) {
    pmax(bounds$lower, pmin(bounds$upper, default.start.degree))
  } else {
    start.raw <- npValidateGlpDegree(
      regtype = "lp",
      degree = degree.start,
      ncon = 1L,
      argname = "degree.start"
    )
    if (!(start.raw[1L] %in% bounds$candidates[[1L]]))
      stop("degree.start must lie within the searched degree candidates")
    start.raw
  }

  list(
    method = if (identical(search.engine, "cell")) degree.select else search.engine,
    engine = search.engine,
    candidates = bounds$candidates,
    lower = bounds$lower,
    upper = bounds$upper,
    baseline.degree = as.integer(baseline.degree),
    start.degree = as.integer(start.degree),
    start.user = !is.null(degree.start),
    basis = if (missing(basis) || is.null(basis)) "glp" else as.character(basis[1L]),
    nobs = as.integer(nobs[1L]),
    restarts = npValidateNonNegativeInteger(degree.restarts, "degree.restarts"),
    max.cycles = npValidatePositiveInteger(degree.max.cycles, "degree.max.cycles"),
    verify = npValidateScalarLogical(degree.verify, "degree.verify"),
    bernstein.basis = bern.auto
  )
}

.npindexbw_attach_degree_search <- function(bws, search_result) {
  metadata <- list(
    mode = search_result$method,
    verify = isTRUE(search_result$verify),
    completed = isTRUE(search_result$completed),
    certified = isTRUE(search_result$certified),
    interrupted = isTRUE(search_result$interrupted),
    baseline.degree = search_result$baseline$degree,
    baseline.fval = search_result$baseline$objective,
    best.degree = search_result$best$degree,
    best.fval = search_result$best$objective,
    nomad.time = search_result$nomad.time,
    powell.time = search_result$powell.time,
    optim.time = search_result$optim.time,
    n.unique = search_result$n.unique,
    n.visits = search_result$n.visits,
    n.cached = search_result$n.cached,
    grid.size = search_result$grid.size,
    best.restart = search_result$best.restart,
    restart.starts = search_result$restart.starts,
    restart.degree.starts = search_result$restart.degree.starts,
    restart.bandwidth.starts = search_result$restart.bandwidth.starts,
    restart.start.info = search_result$restart.start.info,
    restart.results = search_result$restart.results,
    trace = search_result$trace
  )

  if (!is.null(search_result$nomad.time))
    bws$nomad.time <- as.numeric(search_result$nomad.time[1L])
  if (!is.null(search_result$powell.time))
    bws$powell.time <- as.numeric(search_result$powell.time[1L])
  if (!is.null(search_result$optim.time) && is.finite(search_result$optim.time))
    bws$total.time <- as.numeric(search_result$optim.time[1L])
  bws <- .np_attach_nomad_restart_summary(bws, search_result)
  bws$degree.search <- metadata
  bws
}

npindexbw.default <-
  function(xdat = stop("training data xdat missing"),
           ydat = stop("training data ydat missing"),
           bws,
           bandwidth.compute = TRUE,
           basis = c("glp", "additive", "tensor"),
           bernstein.basis = FALSE,
           degree = NULL,
           degree.select = c("manual", "coordinate", "exhaustive"),
           search.engine = c("nomad+powell", "cell", "nomad"),
           nomad = FALSE,
           nomad.nmulti = 0L,
           degree.min = NULL,
           degree.max = NULL,
           degree.start = NULL,
           degree.restarts = 0L,
           degree.max.cycles = 20L,
           degree.verify = FALSE,
           nmulti,
           nomad.remin = FALSE,
           powell.remin = TRUE,
           only.optimize.beta,
           optim.abstol,
           optim.maxattempts,
           optim.maxit,
           optim.method,
           optim.reltol,
           random.seed,
           regtype = c("lc", "ll", "lp"),
           scale.factor.init.lower = 0.1,
           scale.factor.init.upper = 2.0,
           scale.factor.init = 0.5,
           scale.factor.search.lower = NULL,
           ...){

    xdat <- toFrame(xdat)

    if (!(is.vector(ydat) || is.factor(ydat)))
      stop("'ydat' must be a vector")

    if (ncol(xdat) < 2) {
      if (coarseclass(xdat[,1]) != "numeric")
        stop("xdat must contain at least one continuous variable")

      .np_warning(paste("xdat has one dimension. Using a single index model to reduce",
                    "dimensionality is unnecessary."))
    }

    if (coarseclass(bws) != "numeric" || length(bws) != ncol(xdat)+1)
      stop(paste("manually specified 'bws' must be a numeric vector of length ncol(xdat)+1.",
                 "See documentation for details."))

    p <- ncol(xdat)
    mc <- match.call(expand.dots = FALSE)
    mc.names <- names(mc)
    dots <- list(...)
    dot.names <- names(dots)
    nomad.shortcut <- .np_prepare_nomad_shortcut(
      nomad = nomad,
      call_names = unique(c(mc.names, dot.names)),
      preset = list(
        regtype = "lp",
        search.engine = "nomad+powell",
        degree.select = "coordinate",
        bernstein.basis = TRUE,
        degree.min = 0L,
        degree.max = 10L,
        degree.verify = FALSE,
        bwtype = "fixed"
      ),
      values = list(
        regtype = if ("regtype" %in% mc.names) regtype else NULL,
        search.engine = if ("search.engine" %in% mc.names) search.engine else NULL,
        degree.select = if ("degree.select" %in% mc.names) degree.select else NULL,
        bernstein.basis = if ("bernstein.basis" %in% mc.names) bernstein.basis else NULL,
        degree.min = if ("degree.min" %in% mc.names) degree.min else NULL,
        degree.max = if ("degree.max" %in% mc.names) degree.max else NULL,
        degree.verify = if ("degree.verify" %in% mc.names) degree.verify else NULL,
        bwtype = if ("bwtype" %in% dot.names) dots$bwtype else NULL,
        degree = if ("degree" %in% mc.names) degree else NULL
      ),
      where = "npindexbw"
    )

    if (isTRUE(nomad.shortcut$enabled)) {
      if ("degree" %in% mc.names)
        stop("nomad=TRUE does not support an explicit degree; remove degree or set nomad=FALSE")
      if ("regtype" %in% mc.names &&
          !identical(as.character(match.arg(nomad.shortcut$values$regtype, c("lc", "ll", "lp")))[1L], "lp"))
        stop("nomad=TRUE requires regtype='lp'")
      if ("bwtype" %in% dot.names &&
          !identical(as.character(match.arg(nomad.shortcut$values$bwtype, c("fixed", "generalized_nn", "adaptive_nn")))[1L], "fixed"))
        stop("nomad=TRUE currently requires bwtype='fixed'")
      if ("degree.select" %in% mc.names &&
          identical(as.character(match.arg(nomad.shortcut$values$degree.select, c("manual", "coordinate", "exhaustive")))[1L], "manual"))
        stop("nomad=TRUE requires automatic degree search; use degree.select='coordinate' or 'exhaustive'")
      if ("search.engine" %in% mc.names &&
          !(as.character(match.arg(nomad.shortcut$values$search.engine, c("nomad+powell", "cell", "nomad")))[1L] %in%
              c("nomad", "nomad+powell")))
        stop("nomad=TRUE requires search.engine='nomad' or 'nomad+powell'")
      if ("bernstein.basis" %in% mc.names &&
          !isTRUE(npValidateGlpBernstein(regtype = "lp",
                                        bernstein.basis = nomad.shortcut$values$bernstein.basis)))
        stop("nomad=TRUE currently requires bernstein.basis=TRUE")
      if ("degree.verify" %in% mc.names &&
          isTRUE(npValidateScalarLogical(nomad.shortcut$values$degree.verify, "degree.verify")))
        stop("nomad=TRUE currently requires degree.verify=FALSE")
    }

    random.seed.value <- if ("random.seed" %in% mc.names) {
      npValidateNonNegativeInteger(random.seed, "random.seed")
    } else {
      42L
    }
    degree.select.value <- if (!is.null(nomad.shortcut$values$degree.select)) nomad.shortcut$values$degree.select else "manual"
    degree.search <- .npindexbw_degree_search_controls(
      regtype = if (!is.null(nomad.shortcut$values$regtype)) nomad.shortcut$values$regtype else regtype,
      regtype.named = isTRUE(nomad.shortcut$enabled) || ("regtype" %in% mc.names),
      bandwidth.compute = bandwidth.compute,
      nobs = NROW(xdat),
      basis = basis,
      degree.select = degree.select.value,
      search.engine = if (!is.null(nomad.shortcut$values$search.engine)) nomad.shortcut$values$search.engine else "nomad+powell",
      degree.min = nomad.shortcut$values$degree.min,
      degree.max = nomad.shortcut$values$degree.max,
      degree.start = if ("degree.start" %in% mc.names) degree.start else NULL,
      degree.restarts = if ("degree.restarts" %in% mc.names) degree.restarts else 0L,
      degree.max.cycles = if ("degree.max.cycles" %in% mc.names) degree.max.cycles else 20L,
      degree.verify = if (!is.null(nomad.shortcut$values$degree.verify)) nomad.shortcut$values$degree.verify else FALSE,
      bernstein.basis = if (!is.null(nomad.shortcut$values$bernstein.basis)) nomad.shortcut$values$bernstein.basis else bernstein.basis,
      bernstein.named = isTRUE(nomad.shortcut$enabled) || ("bernstein.basis" %in% mc.names)
    )
    nomad.inner.named <- "nomad.nmulti" %in% mc.names
    nomad.inner.nmulti <- if (nomad.inner.named) {
      npValidateNonNegativeInteger(nomad.nmulti, "nomad.nmulti")
    } else {
      0L
    }
    if (nomad.inner.named &&
        (is.null(degree.search) || !(degree.search$engine %in% c("nomad", "nomad+powell")))) {
      stop("nomad.nmulti is only supported when regtype='lp', automatic degree search is active, and search.engine is 'nomad' or 'nomad+powell'")
    }
    degree.setup <- npSetupGlpDegree(
      regtype = if (!is.null(nomad.shortcut$values$regtype)) nomad.shortcut$values$regtype else regtype,
      degree = degree,
      ncon = 1L,
      degree.select = degree.select.value
    )
    scale.factor.search.lower <- npResolveScaleFactorLowerBound(scale.factor.search.lower)
    spec <- npResolveCanonicalConditionalRegSpec(
      mc.names = mc.names,
      regtype = if (!is.null(nomad.shortcut$values$regtype)) nomad.shortcut$values$regtype else regtype,
      basis = basis,
      degree = degree.setup,
      bernstein.basis = if (!is.null(nomad.shortcut$values$bernstein.basis)) nomad.shortcut$values$bernstein.basis else bernstein.basis,
      ncon = 1L,
      where = "npindexbw"
    )
    if (!is.null(degree.search)) {
      spec$bernstein.basis <- degree.search$bernstein.basis
      spec$bernstein.basis.engine <- degree.search$bernstein.basis
    }
    tbw <- sibandwidth(beta = bws[seq_len(p)],
                       h = bws[p+1L], ...,
                       regtype = spec$regtype,
                       basis = spec$basis,
                       degree = spec$degree,
                       bernstein.basis = spec$bernstein.basis,
                       nobs = dim(xdat)[1],
                       xdati = untangle(xdat),
                       ydati = untangle(data.frame(ydat)),
                       xnames = names(xdat),
                       ynames = deparse(substitute(ydat)),
                       bandwidth = bws[p+1L],
                       bandwidth.compute = bandwidth.compute)
    tbw <- npSetScaleFactorSearchLower(tbw, scale.factor.search.lower)

    if (tbw$method == "kleinspady" && !setequal(ydat,c(0,1)))
      stop("Klein and Spady's estimator requires binary ydat with 0/1 values only")

    mc.names <- names(match.call(expand.dots = FALSE))
    margs <- c("nmulti", "nomad.remin", "powell.remin", "random.seed", "optim.method", "optim.maxattempts",
               "optim.reltol", "optim.abstol", "optim.maxit", "only.optimize.beta",
               "scale.factor.init.lower", "scale.factor.init.upper", "scale.factor.init",
               "scale.factor.search.lower")

    m <- match(margs, mc.names, nomatch = 0)
    any.m <- any(m != 0)
    bwsel.args <- list(xdat = xdat, ydat = ydat, bws = tbw)
    if (any.m) {
      nms <- mc.names[m]
      bwsel.args[nms] <- mget(nms, envir = environment(), inherits = FALSE)
    }

    if (bandwidth.compute && !is.null(degree.search)) {
      reg.args <- list(
        regtype = spec$regtype,
        basis = spec$basis,
        degree = spec$degree,
        bernstein.basis = spec$bernstein.basis,
        regtype.engine = spec$regtype.engine,
        basis.engine = spec$basis.engine,
        degree.engine = spec$degree.engine,
        bernstein.basis.engine = spec$bernstein.basis.engine,
        scale.factor.search.lower = scale.factor.search.lower
      )
      opt.args <- c(
        list(bandwidth.compute = bandwidth.compute),
        bwsel.args[setdiff(names(bwsel.args), c("xdat", "ydat", "bws"))]
      )
      opt.args$scale.factor.search.lower <- scale.factor.search.lower

      if (identical(degree.search$engine, "cell")) {
        eval_fun <- function(degree.vec) {
          cell.reg.args <- reg.args
          cell.reg.args$regtype <- "lp"
          cell.reg.args$degree <- as.integer(degree.vec)
          cell.reg.args$bernstein.basis <- degree.search$bernstein.basis
          cell.reg.args$regtype.engine <- "lp"
          cell.reg.args$degree.engine <- as.integer(degree.vec)
          cell.reg.args$bernstein.basis.engine <- degree.search$bernstein.basis
          cell.bws <- .npindexbw_run_fixed_degree(
            xdat = xdat,
            ydat = ydat,
            bws = bws,
            template = tbw,
            reg.args = cell.reg.args,
            opt.args = opt.args
          )
          list(
            objective = as.numeric(cell.bws$fval[1L]),
            payload = cell.bws,
            num.feval = if (!is.null(cell.bws$num.feval)) as.numeric(cell.bws$num.feval[1L]) else NA_real_
          )
        }

        search.result <- .np_degree_search(
          method = degree.search$method,
          candidates = degree.search$candidates,
          baseline_degree = degree.search$baseline.degree,
          start_degree = degree.search$start.degree,
          restarts = degree.search$restarts,
          max_cycles = degree.search$max.cycles,
          verify = degree.search$verify,
          eval_fun = eval_fun,
          direction = "min",
          trace_level = "full",
          objective_name = "fval"
        )
      } else {
        search.result <- .npindexbw_nomad_search(
          xdat = xdat,
          ydat = ydat,
          bws = bws,
          template = tbw,
          reg.args = reg.args,
          opt.args = utils::modifyList(opt.args, list(random.seed = random.seed.value)),
          degree.search = degree.search,
          nomad.inner.nmulti = nomad.inner.nmulti
        )
      }
      tbw <- .npindexbw_attach_degree_search(
        bws = search.result$best_payload,
        search_result = search.result
      )
    } else if (bandwidth.compute) {
      tbw <- .np_progress_select_bandwidth_enhanced(
        "Selecting single-index bandwidth",
        do.call(npindexbw.sibandwidth, bwsel.args)
      )
    }

    mc <- match.call(expand.dots = FALSE)
    environment(mc) <- parent.frame()
    tbw$call <- mc
    tbw <- .np_attach_nomad_shortcut(tbw, nomad.shortcut$metadata)

    return(tbw)
  }

npindexbw.sibandwidth <-
  function(xdat = stop("training data xdat missing"),
           ydat = stop("training data ydat missing"),
           bws,
           bandwidth.compute = TRUE,
           nmulti,
           only.optimize.beta = FALSE,
           optim.abstol = .Machine$double.eps,
           optim.maxattempts = 10,
           optim.maxit = 500,
           optim.method = c("Nelder-Mead", "BFGS", "CG"),
           optim.reltol = sqrt(.Machine$double.eps),
           random.seed = 42,
           scale.factor.init.lower = 0.1,
           scale.factor.init.upper = 2.0,
           scale.factor.init = 0.5,
           scale.factor.search.lower = NULL,
           ...){

    ## Save seed prior to setting

    seed.state <- .np_seed_enter(random.seed)


    xdat = toFrame(xdat)

    if (missing(nmulti)){
      nmulti <- npDefaultNmulti(ncol(xdat))
    }
    bandwidth.compute <- npValidateScalarLogical(bandwidth.compute, "bandwidth.compute")
    only.optimize.beta <- npValidateScalarLogical(only.optimize.beta, "only.optimize.beta")
    nmulti <- npValidateNmulti(nmulti)
    .np_progress_bandwidth_set_total(nmulti)
    optim.maxattempts <- npValidatePositiveInteger(optim.maxattempts, "optim.maxattempts")
    optim.maxit <- npValidatePositiveInteger(optim.maxit, "optim.maxit")
    optim.reltol <- npValidatePositiveFiniteNumeric(optim.reltol, "optim.reltol")
    optim.abstol <- npValidatePositiveFiniteNumeric(optim.abstol, "optim.abstol")
    scale.factor.search.lower <- npResolveScaleFactorLowerBound(
      if (is.null(scale.factor.search.lower)) npGetScaleFactorSearchLower(bws) else scale.factor.search.lower
    )
    h.start.controls <- .npindexbw_h_start_controls(
      scale.factor.init.lower = scale.factor.init.lower,
      scale.factor.init.upper = scale.factor.init.upper,
      scale.factor.init = scale.factor.init,
      scale.factor.search.lower = scale.factor.search.lower,
      where = "npindexbw"
    )

    if (bws$method == "kleinspady" && !setequal(ydat,c(0,1)))
      stop("Klein and Spady's estimator requires binary ydat with 0/1 values only")

    if (ncol(xdat) < 2) {
      if (coarseclass(xdat[,1]) != "numeric")
        stop("xdat must contain at least one continuous variable")

      .np_warning(paste("xdat has one dimension. Using a single index model to reduce",
                    "dimensionality is unnecessary."))
    }

    optim.method <- match.arg(optim.method)

    ## catch and destroy NA's
    keep.rows <- rep_len(TRUE, nrow(xdat))
    rows.omit <- attr(na.omit(data.frame(xdat,ydat)), "na.action")
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Data has no rows without NAs")

    xdat <- xdat[keep.rows,,drop = FALSE]
    ydat <- ydat[keep.rows]

    ## convert to numeric
    if (is.factor(ydat))
      ydat <- dlev(ydat)[as.integer(ydat)]
    else
      ydat <- as.double(ydat)

    xdat = toMatrix(xdat)
    p <- ncol(xdat)
    beta.idx <- if (p > 1L) seq_len(p - 1L) else integer(0)
    nobs <- nrow(xdat)
    spec <- .npindex_resolve_spec(bws, where = "npindexbw")

    total.time <-
      system.time({

        if(bandwidth.compute){

          ## Invariant objects used by objective evaluations.
          xmat <- xdat
          wmat <- cbind(ydat, 1.0)
          bandwidth_eval_count <- 0L

          bandwidth_progress_step <- function() {
            bandwidth_eval_count <<- bandwidth_eval_count + 1L
            .np_progress_bandwidth_activity_step(done = bandwidth_eval_count)
            invisible(NULL)
          }
          fixed.h.lower <- NULL

          ## Note - there are two methods currently implemented, Ichimura's
          ## least squares approach and Klein and Spady's likelihood approach.

          ##We define ichimura's objective function. Since we normalize beta_1
          ##to equal 1, we only worry about beta_2...beta_k in the index
          ##function for these internals, and when computing the leave-one-out
          ##objective function use c(1,beta). However, we do indeed return
          ##c(1,beta) which can be used in the index.model function above.

          ichimuraMaxPenalty <- 10*mean(ydat^2)

          ichimura <- function(param) {
            bandwidth_progress_step()

            ##Define the leave-one-out objective function, sum (y - \hat
            ## G(X\hat\beta))^2. We let beta denote beta_2...beta_k (first k-1
            ## parameters in `param') and then let h denote the kth column.
            beta <- param[beta.idx]
            h <- param[p]
            h.candidate <- .npindex_nn_candidate_bandwidth(h = h, bwtype = bws$type, nobs = nobs)
            if (!h.candidate$ok)
              return(ichimuraMaxPenalty)
            h <- h.candidate$value
            if (!is.null(fixed.h.lower) && h < fixed.h.lower)
              return(ichimuraMaxPenalty)

            ## Next we define the sum of squared leave-one-out residuals

            sum.squares.leave.one.out <- function(beta,h) {

              ## Normalize beta_1 = 1 hence multiply X by c(1,beta)

              index <- xmat %*% c(1, beta)

              fit.loo <- if (identical(spec$regtype.engine, "lc")) {
                ## One call to npksum to avoid repeated computation of the
                ## product kernel (the expensive part)
                tww <- tryCatch(
                  npksum(txdat=index,
                         tydat=wmat,
                         weights=wmat,
                         leave.one.out=TRUE,
                         bandwidth.divide=TRUE,
                         bws=c(h),
                         bwtype = bws$type,
                         ckertype = bws$ckertype,
                         ckerorder = bws$ckerorder,
                         ckerbound = bws$ckerbound,
                         ckerlb = bws$ckerlb,
                         ckerub = bws$ckerub)$ksum,
                  error = function(e) NULL
                )
              if (is.null(tww))
                  return(rep.int(NA_real_, length(ydat)))
                tww[1,2,]/NZD(tww[2,2,])
              } else {
                ok.design <- tryCatch({
                  npCheckRegressionDesignCondition(
                    reg.code = REGTYPE_LP,
                    xcon = data.frame(index = index),
                    basis = spec$basis.engine,
                    degree = spec$degree.engine,
                    bernstein.basis = spec$bernstein.basis.engine,
                    where = "npindexbw"
                  )
                  TRUE
                }, error = function(e) FALSE)
                if (!ok.design)
                  return(ichimuraMaxPenalty)

                .npindex_lp_loo_fit(
                  index = index,
                  ydat = ydat,
                  h = h,
                  bws = bws,
                  spec = spec
                )
              }

              if (any(!is.finite(fit.loo)))
                return(ichimuraMaxPenalty)

              t.ret <- mean((ydat-fit.loo)^2)
              if (is.finite(t.ret) && .npindexbw_fast_eligible(h = h, bws = bws, eval.index = index))
                num.feval.fast.overall <<- num.feval.fast.overall + 1L
              return(t.ret)

            }

            ## For the objective function, we require a positive bandwidth, so
            ## return an infinite penalty for negative h

            if(h > 0) {
              return(sum.squares.leave.one.out(beta,h))
            } else {
              return(ichimuraMaxPenalty)
            }

          }

          ichimura.nobw <- function(param,h){ return(ichimura(c(param,h))) }

          ## We define ichimura's objective function. Since we normalize beta_1
          ## to equal 1, we only worry about beta_2...beta_k in the index
          ## function for these internals, and when computing the leave-one-out
          ## objective function use c(1,beta). However, we do indeed return
          ## c(1,beta) which can be used in the index.model function above.

          kleinspadyFloor <- sqrt(.Machine$double.eps)

          kleinspady <- function(param) {
            bandwidth_progress_step()

            ## Define the leave-one-out objective function, sum (y - \hat
            ## G(X\hat\beta))^2. We let beta denote beta_2...beta_k (first k-1
            ## parameters in `param') and then let h denote the kth column.
            beta <- param[beta.idx]
            h <- param[p]
            h.candidate <- .npindex_nn_candidate_bandwidth(h = h, bwtype = bws$type, nobs = nobs)
            if (!h.candidate$ok)
              return(sqrt(.Machine$double.xmax))
            h <- h.candidate$value
            if (!is.null(fixed.h.lower) && h < fixed.h.lower)
              return(sqrt(.Machine$double.xmax))

            ## Next we define the sum of logs

            sum.log.leave.one.out <- function(beta,h) {

              ## Normalize beta_1 = 1 hence multiply X by c(1,beta)

              index <- xmat %*% c(1, beta)

              ks.loo <- if (identical(spec$regtype.engine, "lc")) {
                ## One call to npksum to avoid repeated computation of the
                ## product kernel (the expensive part)
                tww <- tryCatch(
                  npksum(txdat=index,
                         tydat=wmat,
                         weights=wmat,
                         leave.one.out=TRUE,
                         bandwidth.divide=TRUE,
                         bws=c(h),
                         bwtype = bws$type,
                         ckertype = bws$ckertype,
                         ckerorder = bws$ckerorder,
                         ckerbound = bws$ckerbound,
                         ckerlb = bws$ckerlb,
                         ckerub = bws$ckerub)$ksum,
                  error = function(e) NULL
                )
              if (is.null(tww))
                  return(rep.int(NA_real_, length(ydat)))
                tww[1,2,]/NZD(tww[2,2,])
              } else {
                ok.design <- tryCatch({
                  npCheckRegressionDesignCondition(
                    reg.code = REGTYPE_LP,
                    xcon = data.frame(index = index),
                    basis = spec$basis.engine,
                    degree = spec$degree.engine,
                    bernstein.basis = spec$bernstein.basis.engine,
                    where = "npindexbw"
                  )
                  TRUE
                }, error = function(e) FALSE)
                if (!ok.design)
                  return(sqrt(.Machine$double.xmax))

                .npindex_lp_loo_fit(
                  index = index,
                  ydat = ydat,
                  h = h,
                  bws = bws,
                  spec = spec
                )
              }

              if (any(!is.finite(ks.loo)))
                return(sqrt(.Machine$double.xmax))

              ## Avoid taking log of zero (ks.loo = 0 or 1 since we take
              ## the log of ks.loo and the log of 1-ks.loo)

              ks.loo[ks.loo < kleinspadyFloor] <- kleinspadyFloor
              ks.loo[ks.loo > 1 - kleinspadyFloor] <- 1 - kleinspadyFloor

              ## Maximize the log likelihood, therefore minimize minus.
              ## Here ydat is binary (0/1), so this is equivalent to
              ## ifelse(ydat==1, log(ks.loo), log(1-ks.loo)) but avoids
              ## branchy ifelse and uses stable log1p for the second term.
              t.ret <- -mean(ydat * log(ks.loo) + (1.0 - ydat) * log1p(-ks.loo))
              if (is.finite(t.ret) && .npindexbw_fast_eligible(h = h, bws = bws, eval.index = index))
                num.feval.fast.overall <<- num.feval.fast.overall + 1L
              return(t.ret)

            }

            ## For the objective function, we require a positive bandwidth, so
            ## return an infinite penalty for negative h

            if(h > 0) {
              return(sum.log.leave.one.out(beta,h))
            } else {
              ## No natural counterpart to var of y here, unlike Ichimura above...
              return(sqrt(.Machine$double.xmax))
            }

          }

          kleinspady.nobw <- function(param,h){ return(kleinspady(c(param,h))) }
          ## Now we implement multistarting

          fval.min <- .Machine$double.xmax
          numimp <- 0
          fval.value <- numeric(nmulti)
          num.feval.overall <- 0
          num.feval.fast.overall <- 0L

          if(bws$method == "ichimura"){
            optim.fn <- if(only.optimize.beta) ichimura.nobw else ichimura
            optim.control <- list(abstol=optim.abstol,
                                  reltol=optim.reltol,
                                  maxit=optim.maxit)
          } else if(bws$method == "kleinspady"){
            optim.fn <- if(only.optimize.beta) kleinspady.nobw else  kleinspady
            optim.control <- list(reltol=optim.reltol,maxit=optim.maxit)
          }

          for (i in seq_len(nmulti)) {
            bandwidth_eval_count <- 0L
            ## We use the nlm command to minimize the objective function using
            ## starting values. Note that since we normalize beta_1=1 here beta
            ## is the k-1 vector containing beta_2...beta_k

            if(i == 1) {

              ## Initial values taken from OLS fit with a constant used for
              ## multistart 1
              ols.fit <- lm(ydat~xdat,x=TRUE)
              fit <- fitted(ols.fit)
              fixed.h.lower <- if (identical(bws$type, "fixed")) {
                h.start.controls$scale.factor.search.lower * .npindex_start_bandwidth_scale(fit = fit, nobs = nobs)
              } else {
                NULL
              }

              if (p != 1L){
                if (setequal(bws$beta[2:p], c(0)))
                  beta <- coef(ols.fit)[3:ncol(ols.fit$x)]
                else
                  beta = bws$beta[2:p]
              } else { beta = numeric(0) }

              if (bws$bw == 0)
                h <- .npindex_default_start_bandwidth(
                  fit = fit,
                  bwtype = bws$type,
                  nobs = nobs,
                  start.controls = h.start.controls
                )
              else
                h <- tryCatch(
                  .npindex_finalize_bandwidth(
                    h = bws$bw,
                    bwtype = bws$type,
                    nobs = nobs,
                    lower = fixed.h.lower,
                    where = "npindexbw"
                  ),
                  error = function(e) .npindex_default_start_bandwidth(
                    fit = fit,
                    bwtype = bws$type,
                    nobs = nobs,
                    start.controls = h.start.controls
                  )
                )
            } else {
              ## Random initialization used for remaining multistarts

              beta.length <- length(coef(ols.fit)[3:ncol(ols.fit$x)])
              beta <- runif(beta.length,min=0.5,max=1.5)*coef(ols.fit)[3:ncol(ols.fit$x)]
              if (!only.optimize.beta)
                h <- .npindex_random_start_bandwidth(
                  fit = fit,
                  bwtype = bws$type,
                  nobs = nobs,
                  start.controls = h.start.controls
                )
            }

            optim.parm <- if(only.optimize.beta) beta else c(beta,h)

            optim.base.args <- list(
              par = optim.parm,
              fn = optim.fn,
              gr = NULL,
              method = optim.method,
              control = optim.control
            )
            if (only.optimize.beta) {
              optim.base.args$h <- h
            }
            suppressWarnings(optim.return <- do.call(optim, optim.base.args))
            if(!is.null(optim.return$counts) && length(optim.return$counts) > 0)
              num.feval.overall <- num.feval.overall + optim.return$counts[1]
            attempts <- 0
            while((optim.return$convergence != 0) && (attempts <= optim.maxattempts)) {
              attempts <- attempts + 1
              beta.length <- length(coef(ols.fit)[3:ncol(ols.fit$x)])
              beta <- runif(beta.length,min=0.5,max=1.5)*coef(ols.fit)[3:ncol(ols.fit$x)]
              if(!only.optimize.beta)
                h <- .npindex_random_start_bandwidth(
                  fit = fit,
                  bwtype = bws$type,
                  nobs = nobs,
                  start.controls = h.start.controls
                )

              if(optim.return$convergence == 1){
                if(optim.control$maxit < (2^32/10))
                  optim.control$maxit <- 10*optim.control$maxit
                else
                  stop(paste("optim failed to converge after optim.maxattempts = ", optim.maxattempts, " iterations."))
              }

              if(optim.return$convergence == 10){
                optim.control$reltol <-  10.0*optim.control$reltol
                if(!is.null(optim.control$abstol))
                  optim.control$abstol <-  10.0*optim.control$abstol
              }
              
              optim.parm <- if(only.optimize.beta) beta else c(beta,h)
              optim.base.args$par <- optim.parm
              optim.base.args$control <- optim.control
              if (!only.optimize.beta && ("h" %in% names(optim.base.args))) {
                optim.base.args$h <- NULL
              }
              if (only.optimize.beta) {
                optim.base.args$h <- h
              }
              suppressWarnings(optim.return <- do.call(optim, optim.base.args))
              if(!is.null(optim.return$counts) && length(optim.return$counts) > 0)
                num.feval.overall <- num.feval.overall + optim.return$counts[1]
            }

            if(optim.return$convergence != 0)
              stop(paste("optim failed to converge after optim.maxattempts = ", optim.maxattempts, " iterations."))

            fval.value[i] <- optim.return$value
            if(optim.return$value < fval.min) {
              param <- if(only.optimize.beta) c(optim.return$par,h) else optim.return$par
              fval.min <- optim.return$value
              numimp <- numimp + 1
              best <- i
            }

            .np_progress_bandwidth_multistart_step(done = i, total = nmulti)

          }

          bws$beta <- c(1.0, param[beta.idx])
          bws$bw <- .npindex_finalize_bandwidth(
            h = param[p],
            bwtype = bws$type,
            nobs = nobs,
            lower = fixed.h.lower,
            where = "npindexbw"
          )
          bws$fval <- fval.min
          bws$ifval <- best
          bws$num.feval <- num.feval.overall
          bws$num.feval.fast <- num.feval.fast.overall
          bws$numimp <- numimp
          bws$fval.vector <- fval.value
        }
      })[["elapsed"]]
    ## Return a list with beta (we append the restricted value of
    ## beta_1=1), the bandwidth h, the value of the objective function at
    ## its minimum, the number of restarts that resulted in an improved
    ## value of the objective function, the restart that resulted in the
    ## smallest value, and the vector of objective function values.

    ## Restore seed

    .np_seed_exit(seed.state)

    bws <- sibandwidth(beta = bws$beta,
                       h = bws$bw,
                       method = bws$method,
                       regtype = bws$regtype,
                       basis = bws$basis,
                       degree = bws$degree,
                       bernstein.basis = bws$bernstein.basis,
                       ckertype = bws$ckertype,
                       ckerorder = bws$ckerorder,
                       ckerbound = bws$ckerbound,
                       ckerlb = bws$ckerlb,
                       ckerub = bws$ckerub,
                       bwtype = bws$type,
                       fval = bws$fval,
                       ifval = bws$ifval,
                       num.feval = bws$num.feval,
                       num.feval.fast = bws$num.feval.fast,
                       numimp = bws$numimp,
                       fval.vector = bws$fval.vector,
                       nobs = bws$nobs,
                       xdati = bws$xdati,
                       ydati = bws$ydati,
                       xnames = bws$xnames,
                       ynames = bws$ynames,
                       bandwidth = bws$bw,
                       rows.omit = rows.omit,
                       bandwidth.compute = bandwidth.compute,
                       optim.method = optim.method,
                       only.optimize.beta = only.optimize.beta,
                       total.time = total.time)
    bws <- npSetScaleFactorSearchLower(bws, scale.factor.search.lower)

    bws

  }

test_that("npindex plot bootstrap inid supports nearest-neighbor bwtypes", {
  skip_if_not_installed("np")

  set.seed(3291)
  n <- 40
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(x1 + x2) + rnorm(n, sd = 0.08)
  tx <- data.frame(x1 = x1, x2 = x2)

  run_plot <- function(bw) {
    suppressWarnings(plot(
      bw,
      xdat = tx,
      ydat = y,
      output = "data",
      perspective = FALSE,
      errors = "bootstrap",
      bootstrap = "inid",
      B = 5
    ))
  }

  for (bt in c("generalized_nn", "adaptive_nn")) {
    bw.lc <- npindexbw(
      xdat = tx,
      ydat = y,
      bws = c(1, 1, 5L),
      bandwidth.compute = FALSE,
      bwtype = bt,
      regtype = "lc"
    )
    out.lc <- run_plot(bw.lc)
    expect_type(out.lc, "list")
    expect_true(length(out.lc) > 0)

    bw.ll <- npindexbw(
      xdat = tx,
      ydat = y,
      bws = c(1, 1, 5L),
      bandwidth.compute = FALSE,
      bwtype = bt,
      regtype = "ll"
    )
    out.ll <- run_plot(bw.ll)
    expect_type(out.ll, "list")
    expect_true(length(out.ll) > 0)
  }
})

to_matrix <- getFromNamespace("toMatrix", "np")

test_that("helper constructors forward kernel options and normalize bwscaling", {
  set.seed(8310)
  n <- 50
  xdat <- data.frame(x = runif(n))
  ydat <- data.frame(y = runif(n))

  for (builder in c("npcdensbw", "npcdistbw")) {
    bw <- do.call(builder, list(
      xdat = xdat,
      ydat = ydat,
      bws = c(0.22, 0.22),
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      bwscaling = TRUE,
      cxkertype = "epanechnikov",
      cykertype = "epanechnikov",
      cxkerorder = 2L,
      cykerorder = 2L,
      cxkerbound = "fixed",
      cykerbound = "fixed",
      cxkerlb = 0.0,
      cykerlb = 0.0,
      cxkerub = 1.0,
      cykerub = 1.0
    ))

    cap <- new.env(parent = emptyenv())
    cap$calls <- list()
    np.ns <- asNamespace("np")

    trace(
      what = "kbandwidth.numeric",
      where = np.ns,
      tracer = bquote({
        assign(
          "calls",
          c(
            get("calls", envir = .(cap)),
            list(list(
              bwscaling = bwscaling,
              ckertype = ckertype,
              ckerorder = ckerorder,
              ckerbound = ckerbound
            ))
          ),
          envir = .(cap)
        )
      }),
      print = FALSE
    )
    on.exit(try(untrace("kbandwidth.numeric", where = np.ns), silent = TRUE), add = TRUE)

    make.kx <- getFromNamespace(".np_con_make_kbandwidth_x", "np")
    make.kxy <- getFromNamespace(".np_con_make_kbandwidth_xy", "np")

    obj1 <- make.kx(bws = bw, xdat = xdat)
    obj2 <- make.kxy(bws = bw, xdat = xdat, ydat = ydat)

    expect_false(is.null(obj1))
    expect_false(is.null(obj2))
    expect_true(length(cap$calls) >= 2L)

    for (call in cap$calls) {
      expect_identical(isTRUE(call$bwscaling), FALSE)
      expect_identical(as.character(call$ckertype), as.character(bw$cxkertype))
      expect_identical(as.character(call$ckerorder), as.character(bw$cxkerorder))
      expect_identical(as.character(call$ckerbound), as.character(bw$cxkerbound))
    }

  }
})

test_that("semihat regbw args forward index LP/kernel options with bound collapse", {
  set.seed(8311)
  n <- 64
  xdat <- data.frame(x1 = runif(n), x2 = runif(n))
  ydat <- rnorm(n)

  bw <- npindexbw(
    xdat = xdat,
    ydat = ydat,
    bws = c(0.2, 0.2, 0.3),
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerorder = 2L,
    ckerbound = "fixed",
    ckerlb = c(0, 0),
    ckerub = c(1, 1)
  )
  idx.train <- data.frame(index = as.vector(to_matrix(xdat) %*% bw$beta))

  make.args <- getFromNamespace(".np_semihat_make_regbw_args", "np")
  args <- make.args(
    source = bw,
    xdat = idx.train,
    ydat = rep.int(0.0, nrow(idx.train)),
    bw = bw$bw
  )

  expect_identical(args$regtype, as.character(bw$regtype))
  expect_identical(args$basis, bw$basis)
  expect_equal(args$degree, bw$degree)
  expect_identical(isTRUE(args$bernstein.basis), isTRUE(bw$bernstein.basis))
  expect_identical(args$bwtype, bw$type)
  expect_identical(args$ckertype, bw$ckertype)
  expect_identical(args$ckerorder, bw$ckerorder)
  expect_identical(args$ckerbound, bw$ckerbound)
  expect_equal(as.double(args$ckerlb), 0)
  expect_equal(as.double(args$ckerub), 1)
  expect_identical(args$bandwidth.compute, FALSE)
})

test_that("semihat regbw args forward smooth-coef LP/kernel/scaling options", {
  set.seed(8312)
  n <- 58
  xdat <- data.frame(x = runif(n))
  zdat <- data.frame(z = runif(n))
  ydat <- rnorm(n)

  bw <- npscoefbw(
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    bws = 0.25,
    bandwidth.compute = FALSE,
    regtype = "lp",
    basis = "glp",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    bwscaling = TRUE,
    ckertype = "epanechnikov",
    ckerorder = 2L,
    ckerbound = "fixed",
    ckerlb = 0,
    ckerub = 1
  )

  make.args <- getFromNamespace(".np_semihat_make_regbw_args", "np")
  args <- make.args(
    source = bw,
    xdat = zdat,
    ydat = rep.int(0.0, nrow(zdat)),
    bw = bw$bw
  )

  expect_identical(args$regtype, as.character(bw$regtype))
  expect_identical(args$basis, bw$basis)
  expect_equal(args$degree, bw$degree)
  expect_identical(isTRUE(args$bernstein.basis), isTRUE(bw$bernstein.basis))
  expect_identical(args$bwscaling, isTRUE(bw$scaling))
  expect_identical(args$bwtype, bw$type)
  expect_identical(args$ckertype, bw$ckertype)
  expect_identical(args$ckerorder, bw$ckerorder)
  expect_identical(args$ckerbound, bw$ckerbound)
  expect_equal(as.double(args$ckerlb), 0)
  expect_equal(as.double(args$ckerub), 1)
  expect_identical(args$bandwidth.compute, FALSE)

  if (!is.null(bw$method) && bw$method %in% c("cv.ls", "cv.aic")) {
    expect_identical(args$bwmethod, bw$method)
  }
})

test_that("npplregbw ll stores canonical engine metadata and child regression bandwidths", {
  set.seed(20260306)
  n <- 40
  x <- data.frame(x = runif(n))
  z <- data.frame(z = runif(n))
  y <- sin(2 * pi * z$z) + x$x + rnorm(n, sd = 0.05)

  bw <- npplregbw(
    xdat = x,
    ydat = y,
    zdat = z,
    bws = matrix(0.25, nrow = 2L, ncol = 1L),
    regtype = "ll",
    bandwidth.compute = FALSE
  )

  expect_identical(bw$regtype, "ll")
  expect_identical(bw$basis, "glp")
  expect_identical(as.integer(bw$degree), 1L)
  expect_false(isTRUE(bw$bernstein.basis))
  expect_identical(bw$regtype.engine, "lp")
  expect_identical(bw$basis.engine, "glp")
  expect_identical(as.integer(bw$degree.engine), 1L)
  expect_false(isTRUE(bw$bernstein.basis.engine))

  expect_identical(bw$bw$yzbw$regtype, "lp")
  expect_identical(bw$bw$yzbw$basis, "glp")
  expect_identical(as.integer(bw$bw$yzbw$degree), 1L)
  expect_false(isTRUE(bw$bw$yzbw$bernstein.basis))

  expect_identical(bw$bw[[2]]$regtype, "lp")
  expect_identical(bw$bw[[2]]$basis, "glp")
  expect_identical(as.integer(bw$bw[[2]]$degree), 1L)
  expect_false(isTRUE(bw$bw[[2]]$bernstein.basis))
})

test_that("npplregbw ll rejects non-canonical LP controls", {
  set.seed(20260306)
  n <- 30
  x <- data.frame(x = runif(n))
  z <- data.frame(z = runif(n))
  y <- x$x + z$z + rnorm(n, sd = 0.05)

  expect_error(
    npplregbw(
      xdat = x,
      ydat = y,
      zdat = z,
      regtype = "ll",
      degree = 2L,
      nmulti = 1L
    ),
    "canonical LP\\(degree=1, basis='glp'\\)"
  )

  expect_error(
    npplregbw(
      xdat = x,
      ydat = y,
      zdat = z,
      regtype = "ll",
      basis = "tensor",
      nmulti = 1L
    ),
    "canonical basis='glp'"
  )

  expect_error(
    npplregbw(
      xdat = x,
      ydat = y,
      zdat = z,
      regtype = "ll",
      bernstein.basis = TRUE,
      nmulti = 1L
    ),
    "canonical bernstein.basis=FALSE"
  )
})

test_that("npindexbw ll stores canonical engine metadata", {
  set.seed(20260306)
  n <- 50
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- sin(2 * pi * x$x1) + 0.5 * x$x2 + rnorm(n, sd = 0.05)

  bw <- npindexbw(
    xdat = x,
    ydat = y,
    bws = c(1, 0, 0.25),
    regtype = "ll",
    bandwidth.compute = FALSE
  )

  expect_identical(bw$regtype, "ll")
  expect_identical(bw$basis, "glp")
  expect_identical(as.integer(bw$degree), 1L)
  expect_false(isTRUE(bw$bernstein.basis))
  expect_identical(bw$regtype.engine, "lp")
  expect_identical(bw$basis.engine, "glp")
  expect_identical(as.integer(bw$degree.engine), 1L)
  expect_false(isTRUE(bw$bernstein.basis.engine))
})

test_that("npindexbw ll rejects non-canonical LP controls", {
  set.seed(20260306)
  n <- 40
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- x$x1 + x$x2 + rnorm(n, sd = 0.05)

  expect_error(
    npindexbw(
      xdat = x,
      ydat = y,
      regtype = "ll",
      degree = 2L,
      method = "ichimura",
      nmulti = 1L
    ),
    "canonical LP\\(degree=1, basis='glp'\\)"
  )

  expect_error(
    npindexbw(
      xdat = x,
      ydat = y,
      regtype = "ll",
      basis = "tensor",
      method = "ichimura",
      nmulti = 1L
    ),
    "canonical basis='glp'"
  )

  expect_error(
    npindexbw(
      xdat = x,
      ydat = y,
      regtype = "ll",
      bernstein.basis = TRUE,
      method = "ichimura",
      nmulti = 1L
    ),
    "canonical bernstein.basis=FALSE"
  )
})

progress_time_counter <- function(start = 0, by = 0.6) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

shadow_lines <- function(shadow) {
  vapply(shadow$trace, `[[`, character(1L), "line")
}

shadow_signature <- function(shadow) {
  trace <- shadow$trace[vapply(shadow$trace, `[[`, character(1L), "event") == "render"]
  data.frame(
    event = vapply(trace, `[[`, character(1L), "event"),
    line = vapply(trace, `[[`, character(1L), "line"),
    stringsAsFactors = FALSE
  )
}

test_that("npindexbw adopts the generic bandwidth selection line", {
  set.seed(42)
  x1 <- runif(30)
  x2 <- runif(30)
  y <- x1 + 0.5 * x2 + rnorm(30, sd = 0.1)

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  legacy <- capture_progress_shadow_trace(
    npindexbw(
      xdat = data.frame(x1 = x1, x2 = x2),
      ydat = y,
      method = "ichimura",
      nmulti = 3,
      optim.maxit = 100
    ),
    force_renderer = "legacy",
    now = progress_time_counter()
  )

  set.seed(42)
  single_line <- capture_progress_shadow_trace(
    npindexbw(
      xdat = data.frame(x1 = x1, x2 = x2),
      ydat = y,
      method = "ichimura",
      nmulti = 3,
      optim.maxit = 100
    ),
    force_renderer = "single_line",
    now = progress_time_counter()
  )

  lines <- shadow_lines(single_line)

  expect_s3_class(single_line$value, "sibandwidth")
  expect_equal(shadow_signature(single_line), shadow_signature(legacy))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 1/3\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 1/3, iteration [0-9]+, elapsed [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/3, elapsed [0-9]+\\.[0-9]s, [0-9]+\\.[0-9]%, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/3, iteration [0-9]+, elapsed [0-9]+\\.[0-9]s, [0-9]+\\.[0-9]%, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 3/3, elapsed [0-9]+\\.[0-9]s, 100\\.0%, eta 0\\.0s\\)$", lines)))
})

test_that("npindexbw single-line progress respects np.messages FALSE and suppressMessages", {
  set.seed(7)
  x1 <- runif(25)
  x2 <- runif(25)
  y <- x1 + 0.5 * x2 + rnorm(25, sd = 0.1)

  old_opts <- options(np.messages = FALSE)
  on.exit(options(old_opts), add = TRUE)

  silent <- capture_progress_shadow_trace(
    npindexbw(
      xdat = data.frame(x1 = x1, x2 = x2),
      ydat = y,
      method = "ichimura",
      nmulti = 3,
      optim.maxit = 100
    ),
    force_renderer = "single_line",
    now = progress_time_counter()
  )

  expect_length(silent$trace, 0L)

  old_opts <- options(np.messages = TRUE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(9)
  x1 <- runif(25)
  x2 <- runif(25)
  y <- x1 + 0.5 * x2 + rnorm(25, sd = 0.1)

  suppressed <- capture_progress_shadow_trace(
    suppressMessages(
      npindexbw(
        xdat = data.frame(x1 = x1, x2 = x2),
        ydat = y,
        method = "ichimura",
        nmulti = 3,
        optim.maxit = 100
      )
    ),
    force_renderer = "single_line",
    now = progress_time_counter()
  )

  expect_length(suppressed$trace, 0L)
})

kbandwidth <-
  function(bw = stop("kbandwidth:argument 'bw' missing"), ...) {
    UseMethod("kbandwidth")
  }

kbandwidth.integer <-
  function(bw, ...) { kbandwidth.numeric(bw = bw, ...) }

kbandwidth.default <- function(bw, ...){
  kbandwidth.numeric(bw = unlist(bw$bandwidth),
                     bwscaling = FALSE,
                     bwtype = bw$type,
                     ckertype = bw$ckertype,
                     ckerorder = bw$ckerorder,
                     ckerbound = if (!is.null(bw$ckerbound)) bw$ckerbound else "none",
                     ckerlb = if (!is.null(bw$ckerlb)) bw$ckerlb else NULL,
                     ckerub = if (!is.null(bw$ckerub)) bw$ckerub else NULL,
                     ukertype = bw$ukertype,
                     okertype = bw$okertype,
                     nobs = bw$nobs,
                     xdati = if(is.null(bw$zdati)) bw$xdati else bw$zdati,
                     ydati = bw$ydati,
                     xnames = if(is.null(bw$zdati)) bw$xnames else bw$znames,
                     ynames = bw$ynames,
                     ...)
}

kbandwidth.numeric <-
  function(bw,
           bwscaling = FALSE,
           bwtype = c("fixed","generalized_nn","adaptive_nn"),
           ckertype = c("gaussian","truncated gaussian","epanechnikov","uniform"),
           ckerorder = c(2,4,6,8),
           ckerbound = c("none","range","fixed"),
           ckerlb = NULL,
           ckerub = NULL,
           ukertype = c("aitchisonaitken", "liracine"),
           okertype = c("liracine","wangvanryzin","racineliyan","nliracine"),
           nobs = NA,
           xdati = NULL,
           ydati = NULL,
           xnames = NULL,
           ynames = NULL,
           ...){

    ndim = length(bw)
    bwtype = match.arg(bwtype)
    ckertype = match.arg(ckertype)
    ckerbound = match.arg(ckerbound)

    if(missing(ckerorder))
      ckerorder = 2
    else if (ckertype == "uniform")
      .np_warning("ignoring kernel order specified with uniform kernel type")
    else {
      kord = c(2,4,6,8) 
      if (!any(kord == ckerorder))
        stop("ckerorder must be one of ", paste(kord,collapse=" "))
    }

    if(bwscaling != FALSE)
        stop("npksum only uses raw bandwidths, therefore bwscaling = TRUE is not allowed")
    
    ukertype = match.arg(ukertype)
    okertype = match.arg(okertype)
    cbounds <- npKernelBoundsResolve(
      dati = xdati,
      varnames = xnames,
      kerbound = ckerbound,
      kerlb = ckerlb,
      kerub = ckerub,
      argprefix = "cker")
    bounded_nonfixed_supported <- bwtype %in% c("generalized_nn", "adaptive_nn")
    if (bwtype != "fixed" && cbounds$bound != "none" && !bounded_nonfixed_supported)
      stop("finite continuous kernel bounds require bwtype = \"fixed\"")

    porder = switch( ckerorder/2, "Second-Order", "Fourth-Order", "Sixth-Order", "Eighth-Order" )
    
    mybw = list(
      bw=bw,
      scaling = bwscaling,
      pscaling = npBandwidthSummaryLabel(bwtype = bwtype, bwscaling = bwscaling),
      type = bwtype,
      ptype = bwtToPrint(bwtype),
      ckertype = ckertype,    
      ckerorder = ckerorder,
      ckerbound = cbounds$bound,
      ckerlb = cbounds$lb,
      ckerub = cbounds$ub,
      pckertype = cktToPrint(ckertype, order = porder, kerbound = cbounds$bound),
      ukertype = ukertype,
      pukertype = uktToPrint(ukertype),
      okertype = okertype,
      pokertype = oktToPrint(okertype),
      nobs = nobs,
      ndim = ndim,
      ncon = sum(xdati$icon),
      nuno = sum(xdati$iuno),
      nord = sum(xdati$iord),
      icon = xdati$icon,
      iuno = xdati$iuno,
      iord = xdati$iord,
      xnames = xnames,
      ynames = ynames,
      xdati = xdati,
      ydati = ydati,
      xmcv = mcvConstruct(xdati))

    class(mybw) = "kbandwidth"
    mybw
    
  }

as.double.kbandwidth <- function(x, ...){ x$bw }

print.kbandwidth <- function(x, digits=NULL, ...){
  cat("\nData (",x$nobs," observations, ",x$ndim," variable(s)):\n\n",sep="")
  print(matrix(x$bw,ncol=x$ndim,dimnames=list(paste(x$pscaling,":",sep=""),x$xnames)))
  
  cat("\nKernel Sum Bandwidth",
      if (!identical(x$formula,NULL)) paste("\nFormula:",
                                          deparse(x$formula)),
      "\nBandwidth Type:",x$ptype)


  if (x$ncon > 0)
    cat("\n\nContinuous Kernel Type:",x$pckertype,
        "\nNo. Continuous Vars.:",x$ncon)

  if (x$nuno > 0)
    cat("\n\nUnordered Categorical Kernel Type:",x$pukertype,
        "\nNo. Unordered Categorical Vars.:",x$nuno)

  if (x$nord > 0)
    cat("\n\nOrdered Categorical Kernel Type:",x$pokertype,
        "\nNo. Ordered Categorical Vars.:",x$nord)

  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

npscoefbw <-
  function(...){
    mc <- match.call(expand.dots = FALSE)
    npRejectRenamedScaleFactorSearchArgs(names(mc$...), where = "npscoefbw")
    target <- .np_bw_dispatch_target(dots = mc$...,
                                     data_arg_names = c("xdat", "ydat", "zdat"),
                                     eval_env = parent.frame())
    UseMethod("npscoefbw", target)
  }

npscoefbw.formula <-
  function(formula, data, subset, na.action, call, ...){
    orig.ts <- if (missing(data))
      .np_terms_ts_mask(terms_obj = terms(formula),
                        data = environment(formula),
                        eval_env = environment(formula))
    else .np_terms_ts_mask(terms_obj = terms(formula, data = data),
                           data = data,
                           eval_env = environment(formula))

    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(mf), nomatch = 0)
    mf <- mf[c(1,m)]

    formula.call <- .np_bw_formula_from_call(call_obj = call, eval_env = parent.frame())
    if (!is.null(formula.call))
      mf[[2]] <- formula.call

    mf[[1]] <- as.name("model.frame")

    formula.obj <- .np_bw_resolve_formula(formula_obj = formula,
                                        formula_call = formula.call,
                                        eval_env = parent.frame())
    chromoly <- explodePipe(formula.obj, env = environment(formula))

    bronze <- sapply(chromoly, paste, collapse = " + ")
    mf[["formula"]] <-
      as.formula(paste(bronze[1]," ~ ",
                       paste(bronze[2:length(bronze)],
                             collapse =" + ")),
                 env = environment(formula))

    mf[["formula"]] <- terms(mf[["formula"]])
    if(all(orig.ts)){
      args <- (as.list(attr(mf[["formula"]], "variables"))[-1])
      attr(mf[["formula"]], "predvars") <- as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), args))))
    }else if(any(orig.ts)){
      arguments <- (as.list(attr(mf[["formula"]], "variables"))[-1])
      arguments.normal <- arguments[which(!orig.ts)]
      arguments.timeseries <- arguments[which(orig.ts)]

      ix <- sort(c(which(orig.ts),which(!orig.ts)),index.return = TRUE)$ix
      attr(mf[["formula"]], "predvars") <- bquote(.(as.call(c(quote(cbind),as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments.timeseries)))),arguments.normal,check.rows = TRUE)))[,.(ix)])
    }
    
    mf.args <- as.list(mf[-1L])
    mf <- do.call(stats::model.frame, mf.args, envir = parent.frame())
    
    ydat <- model.response(mf)
    xdat <- mf[, chromoly[[2]], drop = FALSE]
    miss.z <- !(length(chromoly) == 3)
    if (!miss.z)
      zdat <- mf[, chromoly[[3]], drop = FALSE]
    
    bw.args <- list(xdat = xdat, ydat = ydat)
    if (!miss.z)
      bw.args$zdat <- zdat
    tbw <- do.call(npscoefbw, c(bw.args, list(...)))

    ## clean up (possible) inconsistencies due to recursion ...
    tbw$call <- match.call(expand.dots = FALSE)
    environment(tbw$call) <- parent.frame()
    tbw$formula <- formula
    tbw$rows.omit <- as.vector(attr(mf,"na.action"))
    tbw$nobs.omit <- length(tbw$rows.omit)
    tbw$terms <- attr(mf,"terms")
    tbw$chromoly <- chromoly

    tbw <-
      updateBwNameMetadata(nameList =
                           list(ynames =
                                attr(mf, "names")[attr(tbw$terms, "response")]),
                           bws = tbw)
    
    tbw
  }

npscoefbw.NULL <-
  function(xdat = stop("invoked without data 'xdat'"),
           ydat = stop("invoked without data 'ydat'"),
           zdat = NULL,
           bws, ...){

    miss.z <- missing(zdat)

    xdat <- toFrame(xdat)

    if(!miss.z)
      zdat <- toFrame(zdat)

    n.bw <- if (miss.z) ncol(xdat) else ncol(zdat)
    bws <- double(n.bw)

    bw.args <- list(xdat = xdat, ydat = ydat, bws = bws)
    if (!miss.z)
      bw.args$zdat <- zdat
    tbw <- do.call(npscoefbw.default, c(bw.args, list(...)))

    ## clean up (possible) inconsistencies due to recursion ...
    mc <- match.call(expand.dots = FALSE)
    environment(mc) <- parent.frame()
    tbw$call <- mc

    tbw <-
      updateBwNameMetadata(nameList = list(ynames = deparse(substitute(ydat))),
                           bws = tbw)

    tbw
  }

.npscoef_nn_candidate_bandwidth <- function(param, bwtype, nobs) {
  if (identical(bwtype, "fixed"))
    return(as.double(param))

  lower <- 2L
  upper <- max(1L, as.integer(nobs) - 1L)
  vapply(param, function(h) {
    if (!is.finite(h))
      return(NA_real_)
    as.double(max(lower, min(upper, .np_round_half_to_even(h))))
  }, numeric(1))
}

.npscoefbw_start_controls <- function(scale.factor.init.lower = 0.1,
                                     scale.factor.init.upper = 2.0,
                                     scale.factor.init = 0.5,
                                     lbd.init = 0.5,
                                     hbd.init = 1.5,
                                     dfac.init = 1.0,
                                     scale.factor.search.lower = 0,
                                     where = "npscoefbw") {
  cont.start <- npContinuousSearchStartControls(
    scale.factor.init.lower,
    scale.factor.init.upper,
    scale.factor.init,
    scale.factor.search.lower,
    where = where
  )
  lbd.init <- npValidatePositiveFiniteNumeric(lbd.init, "lbd.init")
  hbd.init <- npValidatePositiveFiniteNumeric(hbd.init, "hbd.init")
  dfac.init <- npValidatePositiveFiniteNumeric(dfac.init, "dfac.init")

  if (hbd.init < lbd.init) {
    stop(sprintf("%s: 'hbd.init' must be greater than or equal to 'lbd.init'", where),
         call. = FALSE)
  }
  if (lbd.init > 2 || hbd.init > 2 || dfac.init > 2) {
    stop(sprintf("%s: categorical start factors must be less than or equal to 2", where),
         call. = FALSE)
  }

  list(
    scale.factor.init.lower = cont.start$scale.factor.init.lower,
    scale.factor.init.upper = cont.start$scale.factor.init.upper,
    scale.factor.init = cont.start$scale.factor.init,
    scale.factor.search.lower = as.double(scale.factor.search.lower),
    lbd.init = as.double(lbd.init),
    hbd.init = as.double(hbd.init),
    dfac.init = as.double(dfac.init)
  )
}

.npscoef_start_factor_vector <- function(param,
                                         icon = NULL,
                                         iord = NULL,
                                         iuno = NULL,
                                         continuous.factor,
                                         categorical.factor,
                                         where = "npscoefbw") {
  ndim <- length(param)
  if (is.null(icon) || is.null(iord) || is.null(iuno))
    return(rep.int(as.double(continuous.factor), ndim))

  icon <- as.logical(icon)
  iord <- as.logical(iord)
  iuno <- as.logical(iuno)

  if (length(icon) != ndim || length(iord) != ndim || length(iuno) != ndim) {
    stop(sprintf("%s: invalid fixed-start coordinate map", where), call. = FALSE)
  }

  cat.mask <- iord | iuno
  factor <- rep_len(NA_real_, ndim)
  factor[icon] <- as.double(continuous.factor)
  factor[cat.mask] <- as.double(categorical.factor)

  if (anyNA(factor)) {
    stop(sprintf("%s: unsupported fixed-start coordinate type", where), call. = FALSE)
  }

  factor
}

.npscoef_default_start_bandwidth <- function(param,
                                             bwtype,
                                             nobs,
                                             start.controls = .npscoefbw_start_controls(),
                                             icon = NULL,
                                             iord = NULL,
                                             iuno = NULL) {
  if (identical(bwtype, "fixed"))
    return(as.double(
      param * .npscoef_start_factor_vector(
        param = param,
        icon = icon,
        iord = iord,
        iuno = iuno,
        continuous.factor = start.controls$scale.factor.init,
        categorical.factor = start.controls$dfac.init
      )
    ))

  lower <- 2L
  upper <- max(1L, as.integer(nobs) - 1L)
  start <- max(lower, min(upper, .np_round_half_to_even(sqrt(nobs))))
  rep.int(as.double(start), length(param))
}

.npscoef_random_start_bandwidth <- function(param,
                                            bwtype,
                                            nobs,
                                            start.controls = .npscoefbw_start_controls(),
                                            icon = NULL,
                                            iord = NULL,
                                            iuno = NULL) {
  if (identical(bwtype, "fixed")) {
    draws <- .npscoef_start_factor_vector(
      param = param,
      icon = icon,
      iord = iord,
      iuno = iuno,
      continuous.factor = start.controls$scale.factor.init.lower,
      categorical.factor = start.controls$lbd.init
    )

    for (ii in seq_along(draws)) {
      if (!is.null(icon) && !is.null(iord) && !is.null(iuno) &&
          !isTRUE(as.logical(icon)[ii])) {
        draws[ii] <- runif(1L, min = start.controls$lbd.init, max = start.controls$hbd.init)
      } else {
        draws[ii] <- runif(1L, min = start.controls$scale.factor.init.lower, max = start.controls$scale.factor.init.upper)
      }
    }

    return(as.double(draws * param))
  }

  upper <- max(1L, as.integer(nobs) - 1L)
  .npscoef_nn_candidate_bandwidth(
    param = runif(length(param), min = 2, max = max(2L, upper)),
    bwtype = bwtype,
    nobs = nobs
  )
}

.npscoef_candidate_is_admissible <- function(param,
                                            bwtype,
                                            nobs,
                                            lower = NULL) {
  candidate <- .npscoef_nn_candidate_bandwidth(
    param = param,
    bwtype = bwtype,
    nobs = nobs
  )
  if (any(!is.finite(candidate)))
    return(FALSE)
  if (identical(bwtype, "fixed")) {
    if (!is.null(lower))
      return(all(candidate >= lower))
    return(all(candidate > 0))
  }
  TRUE
}

.npscoef_finalize_bandwidth <- function(param,
                                        bwtype,
                                        nobs,
                                        lower = NULL,
                                        where = "npscoefbw") {
  candidate <- .npscoef_nn_candidate_bandwidth(param = param, bwtype = bwtype, nobs = nobs)
  if (any(!is.finite(candidate))) {
    if (identical(bwtype, "fixed")) {
      stop(sprintf("%s: bandwidth must be finite", where), call. = FALSE)
    }
    stop(
      sprintf(
        "%s: nearest-neighbor bandwidth must be an integer vector in [2, %d]",
        where,
        max(2L, as.integer(nobs) - 1L)
      ),
      call. = FALSE
    )
  }
  if (identical(bwtype, "fixed") && any(candidate <= 0)) {
    stop(sprintf("%s: bandwidth must be strictly positive", where), call. = FALSE)
  }
  if (identical(bwtype, "fixed") && !is.null(lower) && any(candidate < lower)) {
    stop(sprintf("%s: bandwidth is below the continuous scale-factor lower bound", where),
         call. = FALSE)
  }
  as.double(candidate)
}

npscoefbw.scbandwidth <- 
  function(xdat = stop("invoked without data 'xdat'"),
           ydat = stop("invoked without data 'ydat'"),
           zdat = NULL,
           bws,
           backfit.iterate = FALSE,
           backfit.maxiter = 100,
           backfit.tol = .Machine$double.eps,
           bandwidth.compute = TRUE,
           cv.iterate = FALSE,
           cv.num.iterations = 1,
           nmulti,
           optim.abstol = .Machine$double.eps,
           optim.maxattempts = 10,
           optim.maxit = 500,
           optim.method = c("Nelder-Mead", "BFGS", "CG"),
           optim.reltol = sqrt(.Machine$double.eps),
           random.seed = 42,
           scale.factor.init.lower = 0.1,
           scale.factor.init.upper = 2.0,
           scale.factor.init = 0.5,
           lbd.init = 0.5,
           hbd.init = 1.5,
           dfac.init = 1.0,
           scale.factor.search.lower = NULL,
           ...){
    
    ## Save seed prior to setting

    seed.state <- .np_seed_enter(random.seed)


    miss.z <- missing(zdat)
    
    xdat <- toFrame(xdat)

    if (!miss.z)
      zdat <- toFrame(zdat)
    
    if (missing(nmulti)){
      nmulti <- npDefaultNmulti(if (miss.z) NCOL(xdat) else NCOL(zdat))
    }
    regtype <- if (is.null(bws$regtype)) "lc" else bws$regtype
    cv.iterate <- npValidateScalarLogical(cv.iterate, "cv.iterate")
    backfit.iterate <- npValidateScalarLogical(backfit.iterate, "backfit.iterate")
    bandwidth.compute <- npValidateScalarLogical(bandwidth.compute, "bandwidth.compute")
    nmulti <- npValidateNmulti(nmulti)
    .np_progress_bandwidth_set_total(nmulti)
    backfit.maxiter <- npValidatePositiveInteger(backfit.maxiter, "backfit.maxiter")
    backfit.tol <- npValidatePositiveFiniteNumeric(backfit.tol, "backfit.tol")
    optim.maxattempts <- npValidatePositiveInteger(optim.maxattempts, "optim.maxattempts")
    optim.maxit <- npValidatePositiveInteger(optim.maxit, "optim.maxit")
    optim.reltol <- npValidatePositiveFiniteNumeric(optim.reltol, "optim.reltol")
    optim.abstol <- npValidatePositiveFiniteNumeric(optim.abstol, "optim.abstol")
    scale.factor.search.lower <- npResolveScaleFactorLowerBound(
      if (is.null(scale.factor.search.lower)) npGetScaleFactorSearchLower(bws) else scale.factor.search.lower
    )
    start.controls <- .npscoefbw_start_controls(
      scale.factor.init.lower = scale.factor.init.lower,
      scale.factor.init.upper = scale.factor.init.upper,
      scale.factor.init = scale.factor.init,
      lbd.init = lbd.init,
      hbd.init = hbd.init,
      dfac.init = dfac.init,
      scale.factor.search.lower = scale.factor.search.lower,
      where = "npscoefbw"
    )
    if (cv.iterate)
      cv.num.iterations <- npValidatePositiveInteger(cv.num.iterations, "cv.num.iterations")
    spec <- .npscoef_canonical_spec(
      source = bws,
      zdat = if (miss.z) xdat else zdat,
      where = "npscoefbw"
    )
    reg.engine <- spec$regtype.engine

    if (!identical(reg.engine, "lc") && cv.iterate)
      stop("cv.iterate currently supports regtype='lc' for npscoefbw")

    if (!(is.vector(ydat) || is.factor(ydat)))
      stop("'ydat' must be a vector or a factor")

    if (miss.z) {
      bwMatch(xdat, bws$xdati)
    } else {
      bwMatch(zdat, bws$zdati)
    }
    
    if (dim(xdat)[1] != length(ydat))
      stop("number of regression data and response data do not match")

    if (ncol(xdat) == 1 && missing(cv.iterate))
      cv.iterate = FALSE

    if (!all(bws$xdati$icon))
      stop("Only continuous 'x' regressors are supported in this version.")

    optim.method <- match.arg(optim.method)
      
    ## catch and destroy NA's
    keep.rows <- rep_len(TRUE, nrow(xdat))
    train.df <- data.frame(xdat, ydat)
    if (!miss.z)
      train.df <- data.frame(train.df, zdat)
    rows.omit <- attr(na.omit(train.df), "na.action")
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Data has no rows without NAs")

    xdat <- xdat[keep.rows,,drop = FALSE]
    ydat <- ydat[keep.rows]

    if(!miss.z)
      zdat <- zdat[keep.rows,, drop = FALSE]
    
    nrow = dim(xdat)[1]
    ncol = dim(xdat)[2]

    ## at this stage, data to be sent to the c routines must be converted to
    ## numeric type.

    if (is.factor(ydat))
      ydat <- dlev(ydat)[as.integer(ydat)]
    else
      ydat <- as.double(ydat)

    xdat <- toMatrix(xdat)

    ## if (!miss.z)
    ##  zdat <- toMatrix(zdat)
    
    ## bad data
    if (qr(xdat)$rank < ncol(xdat)){
      stop("columns of the independent variable (xdat) are linearly dependent") 
    }

    n <- nrow(xdat)
    
    ## ... do bandwidth selection
    
    ## construct 'W' matrix
    ## in the future one will be able to use a switch to npksum
    ## to emulate W

    W <- cbind(1.0, xdat)
    yW <- cbind(ydat, W)
    
    if (miss.z){
      zdat <- xdat
      dati <- bws$xdati
    }
    else
      dati <- bws$zdati
    zdat.df <- if (is.data.frame(zdat)) zdat else as.data.frame(zdat)

    mysd <- EssDee(zdat[, dati$icon, drop = FALSE])
    nconfac <- n^(-1.0/(2.0*bws$ckerorder+bws$ncon))
    ncatfac <- n^(-2.0/(2.0*bws$ckerorder+bws$ncon))

    bws$sdev <- mysd
    bws$nconfac <- nconfac
    bws$ncatfac <- ncatfac
    bw.scale.multiplier <- NULL
    if (bws$scaling) {
      bw.scale.multiplier <- rep(ncatfac, bws$ndim)
      if (any(bws$icon)) {
        icon.cumsum <- cumsum(dati$icon)
        bw.scale.multiplier[bws$icon] <- nconfac * bws$sdev[icon.cumsum[bws$icon]]
      }
    }
    apply_bw_to_scbw <- function(scbw, param) {
      param <- .npscoef_nn_candidate_bandwidth(param = param, bwtype = scbw$type, nobs = n)
      scbw$bw <- param
      if (scbw$scaling)
        scbw$bandwidth[[1]] <- scbw$bw * bw.scale.multiplier
      else
        scbw$bandwidth[[1]] <- scbw$bw
      scbw
    }

    fast_largeh_tol <- getOption("np.largeh.rel.tol", 1e-3)
    if (!is.numeric(fast_largeh_tol) || length(fast_largeh_tol) != 1L ||
        is.na(fast_largeh_tol) || !is.finite(fast_largeh_tol) ||
        fast_largeh_tol <= 0 || fast_largeh_tol >= 0.1)
      fast_largeh_tol <- 1e-3

    fast_disc_tol <- getOption("np.disc.upper.rel.tol", 1e-2)
    if (!is.numeric(fast_disc_tol) || length(fast_disc_tol) != 1L ||
        is.na(fast_disc_tol) || !is.finite(fast_disc_tol) ||
        fast_disc_tol <= 0 || fast_disc_tol >= 0.5)
      fast_disc_tol <- 1e-2

    cont_utol <- switch(
      bws$ckertype,
      gaussian = sqrt(-2.0 * log(1.0 - fast_largeh_tol)),
      "truncated gaussian" = sqrt(-2.0 * log(1.0 - fast_largeh_tol)),
      epanechnikov = sqrt(fast_largeh_tol),
      uniform = 1.0 - 32.0 * .Machine$double.eps,
      0.0
    )

    cont_hmin <- numeric(0)
    if (any(dati$icon) && is.finite(cont_utol) && cont_utol > 0) {
      zcon <- zdat.df[, dati$icon, drop = FALSE]
      cont_hmin <- vapply(zcon, function(col) {
        vals <- as.double(col)
        vals <- vals[is.finite(vals)]
        if (!length(vals))
          return(Inf)
        diff(range(vals)) / cont_utol
      }, numeric(1))
    }

    disc_upper_tol <- function(upper) {
      max(fast_disc_tol * abs(upper),
          16.0 * .Machine$double.eps * max(1.0, abs(upper)))
    }

    uno_upper <- numeric(0)
    if (any(dati$iuno)) {
      uno_idx <- which(dati$iuno)
      uno_upper <- vapply(uno_idx, function(i) {
        uMaxL(dati$all.nlev[[i]], kertype = bws$ukertype)
      }, numeric(1))
    }

    ord_upper <- numeric(0)
    if (any(dati$iord)) {
      ord_idx <- which(dati$iord)
      ord_upper <- vapply(ord_idx, function(i) {
        oMaxL(dati$all.nlev[[i]], kertype = bws$okertype)
      }, numeric(1))
    }

    npscoef_fast_eligible <- function(sbw) {
      .npscoefbw_fast_eligible(sbw = sbw, eval.zdat = zdat.df)
    }

    solve_cv_moment_system <- function(tyw, tww, W.eval.design, maxPenalty, Wz.eval = NULL) {
      neval <- ncol(tyw)
      ncoef <- nrow(tyw)
      pcoef <- ncol(W.eval.design)
      coef.out <- matrix(maxPenalty, nrow = pcoef, ncol = neval)
      ridge.grid <- npRidgeSequenceAdditive(n.train = n, cap = 1.0)
      ridge <- rep.int(ridge.grid[1L], neval)
      ridge.idx <- rep.int(1L, neval)
      doridge <- rep.int(TRUE, neval)

      while(any(doridge)){
        iloo <- seq_len(neval)[doridge]
        for (ii in iloo) {
          doridge[ii] <- FALSE
          ridge.val <- ridge[ii]*tyw[,ii][1]/NZD(tww[,,ii][1,1])
          theta.ii <- tryCatch(
            solve(tww[,,ii] + diag(rep(ridge[ii], ncoef)),
                  tyw[,ii] + c(ridge.val, rep(0, ncoef - 1))),
            error = function(e) e
          )
          if (inherits(theta.ii, "error")) {
            ridge.idx[ii] <- ridge.idx[ii] + 1L
            if (ridge.idx[ii] <= length(ridge.grid)) {
              ridge[ii] <- ridge.grid[ridge.idx[ii]]
              doridge[ii] <- TRUE
            }
            theta.ii <- rep(maxPenalty, ncoef)
          }

          if (is.null(Wz.eval)) {
            coef.out[,ii] <- theta.ii
          } else {
            coef.out[,ii] <- as.vector(crossprod(
              Wz.eval[ii,],
              matrix(theta.ii, nrow = ncol(Wz.eval), ncol = pcoef)
            ))
          }
        }
      }

      coef.out
    }

    lp_full_coef <- function(sbw, leave.one.out.eval) {
      lp_state <- .npscoef_lp_state(
        bws = sbw,
        tzdat = zdat.df,
        ezdat = zdat.df,
        leave.one.out = leave.one.out.eval,
        where = "npscoefbw"
      )
      tensor.train <- .npscoef_row_tensor_design(W, lp_state$W.train)
      ytensor <- cbind(ydat, tensor.train)
      ksum.args <- list(
        txdat = lp_state$z.train,
        tydat = ytensor,
        weights = ytensor,
        bws = lp_state$rbw,
        leave.one.out = leave.one.out.eval,
        bandwidth.divide = TRUE
      )
      main.ks <- do.call(npksum, ksum.args)$ksum
      tyw <- main.ks[-1L, 1L, , drop = FALSE]
      if (length(dim(tyw)) == 3L)
        dim(tyw) <- c(dim(tyw)[1L], dim(tyw)[3L])
      tww <- main.ks[-1L, -1L, , drop = FALSE]
      solve_cv_moment_system(
        tyw = tyw,
        tww = tww,
        W.eval.design = W,
        Wz.eval = lp_state$W.eval,
        maxPenalty = maxPenalty
      )
    }

    lp_partial_coef <- function(sbw, wj, partial.y, leave.one.out.eval) {
      lp_state <- .npscoef_lp_state(
        bws = sbw,
        tzdat = zdat.df,
        ezdat = zdat.df,
        leave.one.out = leave.one.out.eval,
        where = "npscoefbw"
      )
      U <- lp_state$W.train * wj
      yU <- cbind(partial.y, U)
      ksum.args <- list(
        txdat = lp_state$z.train,
        tydat = yU,
        weights = yU,
        bws = lp_state$rbw,
        leave.one.out = leave.one.out.eval,
        bandwidth.divide = TRUE
      )
      main.ks <- do.call(npksum, ksum.args)$ksum
      tyw <- main.ks[-1L, 1L, , drop = FALSE]
      if (length(dim(tyw)) == 3L)
        dim(tyw) <- c(dim(tyw)[1L], dim(tyw)[3L])
      tww <- main.ks[-1L, -1L, , drop = FALSE]
      as.vector(solve_cv_moment_system(
        tyw = tyw,
        tww = tww,
        W.eval.design = matrix(1.0, nrow = n, ncol = 1L),
        Wz.eval = lp_state$W.eval,
        maxPenalty = maxPenalty
      ))
    }

    total.time <-
      system.time({
        if (bandwidth.compute){
          maxPenalty <- sqrt(.Machine$double.xmax)
          cv_state <- new.env(parent = emptyenv())
          cv_state$fast_total <- 0L
          cv_state$objective_fast <- FALSE
          cv_state$optim_progress <- NULL
          cv_state$optim_eval <- 0L
          cv_state$multistart_index <- NA_integer_
          cv_state$partial_progress <- NULL
          cv_state$partial_eval <- 0L
          cv_state$backfit_iteration <- NA_integer_
          cv_state$partial_index <- NA_integer_

          cv_progress_detail <- function(ridging = FALSE) {
            detail <- sprintf("multistart %d", cv_state$multistart_index)
            if (isTRUE(ridging)) {
              paste(detail, "near-singular system encountered, ridging", sep = ", ")
            } else {
              detail
            }
          }

          cv_progress_begin <- function() {
            cv_state$optim_eval <- 0L
            cv_state$optim_progress <- .np_progress_begin("Optimizing smooth coefficient bandwidth")
            invisible(NULL)
          }

          cv_progress_step <- function(ridging = FALSE) {
            cv_state$optim_eval <- cv_state$optim_eval + 1L
            .np_progress_bandwidth_activity_step(done = cv_state$optim_eval)
            cv_state$optim_progress <- .np_progress_step(
              state = cv_state$optim_progress,
              done = cv_state$optim_eval,
              detail = cv_progress_detail(ridging = ridging)
            )
            invisible(NULL)
          }

          cv_progress_end <- function(state) {
            if (is.null(state))
              return(invisible(NULL))

            if (isTRUE(state$known_total) && identical(state$last_done, state$total))
              return(invisible(NULL))

            state$last_emit <- -Inf
            .np_progress_end(state)
            invisible(NULL)
          }

          cv_progress_finish <- function(ridging = FALSE) {
            if (is.null(cv_state$optim_progress))
              return(invisible(NULL))

            cv_state$optim_progress$last_emit <- -Inf
            cv_state$optim_progress <- .np_progress_end(
              cv_state$optim_progress,
              detail = cv_progress_detail(ridging = ridging)
            )
            cv_state$optim_progress <- NULL
            invisible(NULL)
          }

          partial_progress_detail <- function(fv = NULL) {
            detail <- sprintf(
              "backfitting iteration %d of %d, partial residual %d of %d",
              cv_state$backfit_iteration,
              cv.num.iterations,
              cv_state$partial_index,
              ncol(W)
            )
            if (!is.null(fv)) {
              detail <- paste(
                detail,
                sprintf(
                  "fval %s",
                  format(signif(fv, digits = getOption("digits", 7L)), trim = TRUE)
                ),
                sep = ", "
              )
            }
            detail
          }

          partial_progress_begin <- function(iteration, partial.index) {
            cv_state$backfit_iteration <- iteration
            cv_state$partial_index <- partial.index
            cv_state$partial_eval <- 0L
            cv_state$partial_progress <- .np_progress_begin("Optimizing partial residual bandwidth")
            invisible(NULL)
          }

          partial_progress_step <- function(fv) {
            cv_state$partial_eval <- cv_state$partial_eval + 1L
            cv_state$partial_progress <- .np_progress_step(
              state = cv_state$partial_progress,
              done = cv_state$partial_eval,
              detail = partial_progress_detail(fv = fv)
            )
            invisible(NULL)
          }

          partial_progress_finish <- function(fv = NULL) {
            if (is.null(cv_state$partial_progress))
              return(invisible(NULL))

            cv_state$partial_progress$last_emit <- -Inf
            cv_state$partial_progress <- .np_progress_end(
              cv_state$partial_progress,
              detail = partial_progress_detail(fv = fv)
            )
            cv_state$partial_progress <- NULL
            invisible(NULL)
          }

          overall.cv.ls <- function(param) {
            cv_state$objective_fast <- FALSE
            sbw <- apply_bw_to_scbw(bws, param)
            if (!validateBandwidthTF(sbw) ||
                (!is.null(fixed.lower) && any(param < fixed.lower)) ||
                ((bws$nord+bws$nuno > 0) && any(param[!bws$icon] > 2.0*x.scale[!bws$icon])))
              return(maxPenalty)
            cv_state$objective_fast <- npscoef_fast_eligible(sbw)

            if (identical(reg.engine, "lc")) {
              tww <- npksum(txdat = zdat, tydat = yW, weights = yW, bws = sbw,
                            leave.one.out = TRUE)$ksum

              mean.loo <- rep(maxPenalty,n)
              ridge.grid <- npRidgeSequenceAdditive(n.train = n, cap = 1.0)
              ridge <- rep.int(ridge.grid[1L], n)
              ridge.idx <- rep.int(1L, n)
              doridge <- rep.int(TRUE, n)

              nc <- ncol(tww[-1,-1,1])

              while(any(doridge)){
                iloo <- which(doridge)
                for (ii in iloo) {
                  doridge[ii] <- FALSE
                  ridge.val <- ridge[ii]*tww[-1,1,ii][1]/NZD(tww[-1,-1,ii][1,1])
                  beta.ii <- tryCatch(
                    solve(tww[-1,-1,ii] + diag(rep(ridge[ii], nc)),
                          tww[-1,1,ii] + c(ridge.val, rep(0, nc - 1))),
                    error = function(e) e
                  )
                  if (inherits(beta.ii, "error")) {
                    ridge.idx[ii] <- ridge.idx[ii] + 1L
                    if (ridge.idx[ii] <= length(ridge.grid)) {
                      ridge[ii] <- ridge.grid[ridge.idx[ii]]
                      doridge[ii] <- TRUE
                    }
                    beta.ii <- rep(maxPenalty, nc)
                  }
                  mean.loo[ii] <- W[ii,, drop = FALSE] %*% beta.ii
                }
              }
            } else {
              coef.loo <- lp_full_coef(sbw = sbw, leave.one.out.eval = TRUE)
              mean.loo <- rowSums(W * t(coef.loo))
            }

            stopifnot(all(is.finite(mean.loo)))

            if(!any(mean.loo == maxPenalty)){
              fv <- sum((ydat-mean.loo)^2)/n
              cv_progress_step()
            } else {
              cv_progress_step(ridging = TRUE)
              fv <- maxPenalty
            }

            if (isTRUE(cv_state$objective_fast))
              cv_state$fast_total <- cv_state$fast_total + 1L

            return((if (is.finite(fv)) fv else maxPenalty))

          }

          scoef.loo.args <- list(
            bws = bws, txdat = xdat, tydat = ydat,
            leave.one.out = TRUE, iterate = TRUE,
            maxiter = backfit.maxiter, tol = backfit.tol,
            betas = TRUE,
            .np_fit_progress_allow = FALSE
          )
          if (!miss.z)
            scoef.loo.args$tzdat <- zdat
          
          partial.cv.ls <- function(param, partial.index) {
            cv_state$objective_fast <- FALSE
            sbw <- apply_bw_to_scbw(bws, param)

            if (!validateBandwidthTF(sbw) ||
                (!is.null(fixed.lower) && any(param < fixed.lower)) ||
                ((bws$nord+bws$nuno > 0) && any(param[!bws$icon] > 2.0*x.scale[!bws$icon])))
              return(maxPenalty)
            cv_state$objective_fast <- npscoef_fast_eligible(sbw)
            
            if (backfit.iterate){
              bws$bw.fitted[,partial.index] <- param
              scoef.loo <- do.call(npscoef, scoef.loo.args)
              partial.loo <- W[,partial.index]*scoef.loo$beta[,partial.index]
            } else {
              wj <- W[,partial.index]
              if (identical(reg.engine, "lc")) {
                tww <- npksum(txdat=zdat,
                              tydat=cbind(partial.orig * wj, wj * wj),
                              weights=cbind(partial.orig * wj, 1),
                              bws=sbw,
                              leave.one.out=TRUE)$ksum

                partial.loo <- wj * tww[1,2,]/NZD(tww[2,2,])
              } else {
                partial.loo <- wj * lp_partial_coef(
                  sbw = sbw,
                  wj = wj,
                  partial.y = partial.orig,
                  leave.one.out.eval = TRUE
                )
              }
            }
            

            fv <- sum((partial.orig - partial.loo)^2)/n

            if (isTRUE(cv_state$objective_fast))
              cv_state$fast_total <- cv_state$fast_total + 1L

            partial_progress_step(fv = fv)
            return((if (is.finite(fv)) fv else maxPenalty))
          }

          ## Now we implement multistarting

          fval.min <- .Machine$double.xmax
          have_best <- FALSE
          numimp <- 0
          value.overall <- numeric(nmulti)
          num.feval.overall <- 0

          x.scale <- sapply(seq_len(bws$ndim), function(i){
            if (dati$icon[i]){
              return(1.059224*((if (bws$scaling) 1.0 else mysd[sum(dati$icon[seq_len(i)])]*nconfac)))
            }
            
            if (dati$iord[i])
              return(0.5*oMaxL(dati$all.nlev[[i]], kertype = bws$okertype)*
                     (if (bws$scaling) ncatfac else 1.0))
            
            if (dati$iuno[i])
              return(0.5*uMaxL(dati$all.nlev[[i]], kertype = bws$ukertype)*
                     (if (bws$scaling) ncatfac else 1.0))       
          })
          fixed.lower <- if (identical(bws$type, "fixed")) {
            out <- rep.int(0, length(x.scale))
            out[dati$icon] <- x.scale[dati$icon] * start.controls$scale.factor.search.lower
            out
          } else {
            NULL
          }

          optim.control <- list(abstol = optim.abstol,
                                reltol = optim.reltol,
                                maxit = optim.maxit)

          for (i in seq_len(nmulti)) {
            cv_state$multistart_index <- i
            cv_progress_begin()
            
            if (i == 1) {
              tbw <- .npscoef_default_start_bandwidth(
                param = x.scale,
                bwtype = bws$type,
                nobs = n,
                start.controls = start.controls,
                icon = dati$icon,
                iord = dati$iord,
                iuno = dati$iuno
              )
              if (all(bws$bw != 0) &&
                  .npscoef_candidate_is_admissible(param = bws$bw, bwtype = bws$type, nobs = n,
                                                   lower = fixed.lower)) {
                tbw <- .npscoef_finalize_bandwidth(
                  param = bws$bw,
                  bwtype = bws$type,
                  nobs = n,
                  lower = fixed.lower,
                  where = "npscoefbw"
                )
              }
            } else {
              tbw <- .npscoef_random_start_bandwidth(
                param = x.scale,
                bwtype = bws$type,
                nobs = n,
                start.controls = start.controls,
                icon = dati$icon,
                iord = dati$iord,
                iuno = dati$iuno
              )
            }

            suppressWarnings(optim.return <- optim(tbw,
                                                   fn = overall.cv.ls,
                                                   control = optim.control))
            if(!is.null(optim.return$counts) && length(optim.return$counts) > 0)
              num.feval.overall <- num.feval.overall + optim.return$counts[1]
            attempts <- 0
            while((optim.return$convergence != 0) && (attempts <= optim.maxattempts)) {
              attempts <- attempts + 1
              tbw <- .npscoef_random_start_bandwidth(
                param = x.scale,
                bwtype = bws$type,
                nobs = n,
                start.controls = start.controls,
                icon = dati$icon,
                iord = dati$iord,
                iuno = dati$iuno
              )
              optim.control <- lapply(optim.control, '*', 10.0)
              suppressWarnings(optim.return <- optim(tbw,
                                                     fn = overall.cv.ls,
                                                     control = optim.control))
              if(!is.null(optim.return$counts) && length(optim.return$counts) > 0)
                num.feval.overall <- num.feval.overall + optim.return$counts[1]

            }

            cv_progress_finish()

            value.overall[i] <- optim.return$value

            if (.npscoef_candidate_is_admissible(
              param = optim.return$par,
              bwtype = bws$type,
              nobs = n,
              lower = fixed.lower
            ) && (!have_best || optim.return$value < fval.min)) {
              param <- .npscoef_finalize_bandwidth(
                param = optim.return$par,
                bwtype = bws$type,
                nobs = n,
                lower = fixed.lower,
                where = "npscoefbw"
              )
              min.overall <- optim.return$value
              fval.min <- min.overall ## Added by jracine Jul 22 2010
              numimp.overall <- numimp + 1
              best.overall <- i
              have_best <- TRUE
            }

            .np_progress_bandwidth_multistart_step(done = i, total = nmulti)
          }

          if (!have_best) {
            if (identical(bws$type, "fixed")) {
              stop("npscoefbw: no feasible fixed bandwidths found", call. = FALSE)
            }
            stop("npscoefbw: no feasible bandwidths found", call. = FALSE)
          }

          param.overall <- bws$bw <- .npscoef_finalize_bandwidth(
            param = param,
            bwtype = bws$type,
            nobs = n,
            where = "npscoefbw"
          )
          bws <- apply_bw_to_scbw(bws, bws$bw)

          if(cv.iterate){
            n.part <- (ncol(xdat)+1)
            backfit.progress <- .np_progress_begin(
              "Backfitting smooth coefficient bandwidth",
              total = cv.num.iterations
            )
            on.exit(cv_progress_end(backfit.progress), add = TRUE)
            bws$bw.fitted <- matrix(data = bws$bw, nrow = length(bws$bw), ncol = n.part)
            ## obtain matrix of alpha.hat | h0 and beta.hat | h0

            scoef.args <- list(
              bws = bws,
              txdat = xdat,
              tydat = ydat,
              iterate = FALSE,
              betas = TRUE,
              .np_fit_progress_allow = FALSE
            )
            if (!miss.z)
              scoef.args$tzdat <- zdat
            scoef <- do.call(npscoef, scoef.args)
            
            resid.full <- ydat - scoef$mean

            
            for (i in seq_len(cv.num.iterations)) {
              backfit.progress <- .np_progress_step(
                state = backfit.progress,
                done = i,
                detail = sprintf("iteration %d of %d", i, cv.num.iterations)
              )

              for (j in seq_len(n.part)) {
                ## estimate partial residuals
                partial.orig <- W[,j] * scoef$beta[,j] + resid.full
                partial_progress_begin(iteration = i, partial.index = j)
                
                ## minimise
                suppressWarnings(optim.return <-
                                 optim(tbw, fn = partial.cv.ls,
                                       control = optim.control,
                                       partial.index = j))
                if(!is.null(optim.return$counts) && length(optim.return$counts) > 0)
                  num.feval.overall <- num.feval.overall + optim.return$counts[1]
                partial_progress_finish(fv = optim.return$value)
                
                ## grab parameter
                bws$bw.fitted[,j] <- optim.return$par

                if (backfit.iterate){
                  ## re-estimate all betas
                  scoef.args <- list(
                    bws = bws, txdat = xdat, tydat = ydat,
                    iterate = TRUE, maxiter = backfit.maxiter,
                    tol = backfit.tol, betas = TRUE,
                    .np_fit_progress_allow = FALSE
                  )
                  if (!miss.z)
                    scoef.args$tzdat <- zdat
                  scoef <- do.call(npscoef, scoef.args)
                  resid.full <- ydat - scoef$mean
                } else {
                  bws$bw <- bws$bw.fitted[,j]
                  ## estimate new beta.hats

                  bws <- apply_bw_to_scbw(bws, bws$bw)

                  if (identical(reg.engine, "lc")) {
                    wj <- W[,j]
                    tww <- npksum(txdat=zdat,
                                  tydat=cbind(partial.orig * wj, wj * wj),
                                  weights=cbind(partial.orig * wj, 1),
                                  bws=bws)$ksum
                    scoef$beta[,j] <- tww[1,2,]/NZD(tww[2,2,])
                  } else {
                    wj <- W[,j]
                    scoef$beta[,j] <- lp_partial_coef(
                      sbw = bws,
                      wj = wj,
                      partial.y = partial.orig,
                      leave.one.out.eval = FALSE
                    )
                  }
                  
                  bws$bw <- param.overall
                  bws <- apply_bw_to_scbw(bws, bws$bw)
                  ## estimate new full residuals 
                  resid.full <- partial.orig - W[,j] * scoef$beta[,j]
                }
              }
            }
            scoef.loo.args <- list(
              bws = bws, txdat = xdat, tydat = ydat,
              iterate = TRUE, maxiter = backfit.maxiter,
              tol = backfit.tol, leave.one.out = TRUE,
              .np_fit_progress_allow = FALSE
            )
            if (!miss.z)
              scoef.loo.args$tzdat <- zdat
            scoef.loo <- do.call(npscoef, scoef.loo.args)$mean
            bws$fval.fitted <- sum((ydat - scoef.loo)^2)/n
          }

          bws$fval = min.overall
          bws$ifval = best.overall
          bws$num.feval = num.feval.overall
          bws$num.feval.fast = cv_state$fast_total
          bws$numimp = numimp.overall
          bws$fval.vector = value.overall
        }
      })[["elapsed"]]
    
    bws$sfactor <- bws$bandwidth <- bws$bw
    nfactor <- nrow^(-2.0/(2.0*bws$ckerorder+bws$ncon))
    dfactor <- EssDee(zdat[, dati$icon, drop = FALSE])*nrow^(-1.0/(2.0*bws$ckerorder+sum(dati$icon)))

    if (bws$scaling) {
      bws$bandwidth[dati$icon] <- bws$bandwidth[dati$icon]*dfactor

      if(bws$nuno > 0)
        bws$bandwidth[dati$iuno] <- bws$bandwidth[dati$iuno]*nfactor

      if(bws$nord > 0)
        bws$bandwidth[dati$iord] <- bws$bandwidth[dati$iord]*nfactor
      
    } else {
      bws$sfactor[dati$icon] <- bws$sfactor[dati$icon]/dfactor

      if(bws$nuno > 0)
        bws$sfactor[dati$iuno] <- bws$sfactor[dati$iuno]/nfactor

      if(bws$nord > 0)
        bws$sfactor[dati$iord] <- bws$sfactor[dati$iord]/nfactor
    }

    ## Restore seed

    .np_seed_exit(seed.state)
    
    bws <- scbandwidth(bw = bws$bw,
                       regtype = regtype,
                       basis = if (is.null(bws$basis)) "glp" else bws$basis,
                       degree = bws$degree,
                       bernstein.basis = bws$bernstein.basis,
                       bwmethod = bws$method,
                       bwscaling = bws$scaling,
                       bwtype = bws$type,
                       ckertype = bws$ckertype,
                       ckerorder = bws$ckerorder,
                       ckerbound = bws$ckerbound,
                       ckerlb = bws$ckerlb,
                       ckerub = bws$ckerub,
                       ukertype = bws$ukertype,
                       okertype = bws$okertype,
                       fval = bws$fval,
                       ifval = bws$ifval,
                       num.feval = bws$num.feval,
                       num.feval.fast = bws$num.feval.fast,
                       numimp = bws$numimp,
                       fval.vector = bws$fval.vector,
                       nobs = bws$nobs,
                       xdati = bws$xdati,
                       ydati = bws$ydati,
                       zdati = bws$zdati,
                       xnames = bws$xnames,
                       ynames = bws$ynames,
                       znames = bws$znames,
                       sfactor = bws$sfactor,
                       bandwidth = bws$bandwidth,
                       rows.omit = rows.omit,
                       bandwidth.compute = bandwidth.compute,
                       optim.method = optim.method,
                       total.time = total.time)
    bws <- npSetScaleFactorSearchLower(bws, scale.factor.search.lower)

    bws
  }

.npscoefbw_build_scbandwidth <- function(xdat,
                                         ydat,
                                         zdat,
                                         bws,
                                         bandwidth.compute,
                                         reg.args) {
  miss.z <- is.null(zdat)
  zdati <- if (miss.z) NULL else untangle(zdat)
  znames <- if (miss.z) NULL else names(zdat)

  sbw.args <- c(
    list(
      bw = bws,
      nobs = dim(xdat)[1],
      xdati = untangle(xdat),
      ydati = untangle(data.frame(ydat)),
      zdati = zdati,
      xnames = names(xdat),
      ynames = deparse(substitute(ydat)),
      znames = znames,
      bandwidth.compute = bandwidth.compute
    ),
    reg.args
  )

  out <- do.call(scbandwidth, sbw.args)
  if (!is.null(reg.args$scale.factor.search.lower))
    out$scale.factor.search.lower <- npResolveScaleFactorLowerBound(
      reg.args$scale.factor.search.lower
    )
  out
}

.npscoefbw_run_fixed_degree <- function(xdat, ydat, zdat, bws, reg.args, opt.args) {
  tbw <- .npscoefbw_build_scbandwidth(
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    bws = bws,
    bandwidth.compute = opt.args$bandwidth.compute,
    reg.args = reg.args
  )

  scbw.args <- c(list(xdat = xdat, ydat = ydat, bws = tbw), opt.args)
  if (!is.null(zdat))
    scbw.args$zdat <- zdat
  do.call(npscoefbw.scbandwidth, scbw.args)
}

.npscoefbw_nomad_controls <- function(search.engine) {
  .np_degree_search_engine_controls(search.engine)
}

.npscoefbw_fast_eligible <- function(sbw, eval.zdat) {
  if (!identical(sbw$type, "fixed"))
    return(FALSE)

  tdati <- if (is.null(sbw$zdati)) sbw$xdati else sbw$zdati
  eval.zdat <- toFrame(eval.zdat)

  fast_largeh_tol <- getOption("np.largeh.rel.tol", 1e-3)
  if (!is.numeric(fast_largeh_tol) || length(fast_largeh_tol) != 1L ||
      is.na(fast_largeh_tol) || !is.finite(fast_largeh_tol) ||
      fast_largeh_tol <= 0 || fast_largeh_tol >= 0.1)
    fast_largeh_tol <- 1e-3

  fast_disc_tol <- getOption("np.disc.upper.rel.tol", 1e-2)
  if (!is.numeric(fast_disc_tol) || length(fast_disc_tol) != 1L ||
      is.na(fast_disc_tol) || !is.finite(fast_disc_tol) ||
      fast_disc_tol <= 0 || fast_disc_tol >= 0.5)
    fast_disc_tol <- 1e-2

  cont_utol <- switch(
    sbw$ckertype,
    gaussian = sqrt(-2.0 * log(1.0 - fast_largeh_tol)),
    "truncated gaussian" = sqrt(-2.0 * log(1.0 - fast_largeh_tol)),
    epanechnikov = sqrt(fast_largeh_tol),
    uniform = 1.0 - 32.0 * .Machine$double.eps,
    0.0
  )

  cont_hmin <- numeric(0)
  if (any(sbw$icon) && is.finite(cont_utol) && cont_utol > 0) {
    zcon <- eval.zdat[, sbw$icon, drop = FALSE]
    cont_hmin <- vapply(zcon, function(col) {
      vals <- as.double(col)
      vals <- vals[is.finite(vals)]
      if (!length(vals))
        return(Inf)
      diff(range(vals)) / cont_utol
    }, numeric(1))
  }

  disc_upper_tol <- function(upper) {
    max(fast_disc_tol * abs(upper),
        16.0 * .Machine$double.eps * max(1.0, abs(upper)))
  }

  uno_upper <- numeric(0)
  if (any(sbw$iuno)) {
    uno_idx <- which(sbw$iuno)
    uno_upper <- vapply(uno_idx, function(i) {
      uMaxL(tdati$all.nlev[[i]], kertype = sbw$ukertype)
    }, numeric(1))
  }

  ord_upper <- numeric(0)
  if (any(sbw$iord)) {
    ord_idx <- which(sbw$iord)
    ord_upper <- vapply(ord_idx, function(i) {
      oMaxL(tdati$all.nlev[[i]], kertype = sbw$okertype)
    }, numeric(1))
  }

  bwv <- sbw$bandwidth[[1L]]
  if (!length(bwv) || length(bwv) != length(sbw$icon))
    return(FALSE)

  if (any(sbw$icon)) {
    bw_cont <- bwv[sbw$icon]
    if (any(!is.finite(bw_cont)) || any(bw_cont <= 0) ||
        any(bw_cont < cont_hmin))
      return(FALSE)
  }

  if (any(sbw$iuno)) {
    bw_uno <- bwv[sbw$iuno]
    ok_uno <- mapply(function(bw, upper) {
      is.finite(bw) && abs(bw - upper) <= disc_upper_tol(upper)
    }, bw = bw_uno, upper = uno_upper, SIMPLIFY = TRUE, USE.NAMES = FALSE)
    if (!all(ok_uno))
      return(FALSE)
  }

  if (any(sbw$iord)) {
    bw_ord <- bwv[sbw$iord]
    ok_ord <- mapply(function(bw, upper) {
      is.finite(bw) && abs(bw - upper) <= disc_upper_tol(upper)
    }, bw = bw_ord, upper = ord_upper, SIMPLIFY = TRUE, USE.NAMES = FALSE)
    if (!all(ok_ord))
      return(FALSE)
  }

  TRUE
}

.npscoefbw_eval_only <- function(xdat,
                                 ydat,
                                 zdat,
                                 bws,
                                 invalid.penalty = c("baseline", "large"),
                                 penalty.multiplier = 10) {
  invalid.penalty <- match.arg(invalid.penalty)
  base.penalty <- switch(
    invalid.penalty,
    baseline = if (is.finite(bws$fval[1L])) as.numeric(bws$fval[1L]) else 1,
    large = 1
  )
  base.penalty <- max(abs(base.penalty), 1)
  penalty <- penalty.multiplier * base.penalty

  fit <- tryCatch(
    npscoef(
      bws = bws,
      txdat = xdat,
      tydat = ydat,
      tzdat = zdat,
      leave.one.out = TRUE,
      iterate = FALSE,
      betas = FALSE,
      errors = FALSE,
      .np_fit_progress_allow = FALSE
    ),
    error = function(e) e
  )

  if (inherits(fit, "error") || is.null(fit$mean) || any(!is.finite(fit$mean))) {
    return(list(objective = penalty, num.feval = 1L, num.feval.fast = 0L))
  }

  list(
    objective = as.numeric(mean((as.double(ydat) - as.double(fit$mean))^2)),
    num.feval = 1L,
    num.feval.fast = if (.npscoefbw_fast_eligible(bws, eval.zdat = if (is.null(zdat)) xdat else zdat)) 1L else 0L
  )
}

.npscoefbw_nomad_search <- function(xdat,
                                    ydat,
                                    zdat,
                                    bws,
                                    reg.args,
                                    opt.args,
                                    degree.search,
                                    nomad.inner.nmulti = 0L) {
  if (isTRUE(degree.search$verify))
    stop("automatic degree search with search.engine='nomad' does not support degree.verify")

  if (!identical(opt.args$bandwidth.compute, TRUE))
    stop("automatic degree search with search.engine='nomad' requires bandwidth.compute=TRUE")

  eval.zdat <- if (is.null(zdat)) xdat else zdat

  template.reg.args <- reg.args
  template.reg.args$regtype <- "lp"
  template.reg.args$degree <- as.integer(degree.search$start.degree)
  template.reg.args$bernstein.basis <- degree.search$bernstein.basis

  template <- .npscoefbw_build_scbandwidth(
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    bws = bws,
    bandwidth.compute = FALSE,
    reg.args = template.reg.args
  )

  if (!identical(template$type, "fixed"))
    stop("automatic degree search with search.engine='nomad' currently requires bwtype='fixed'")
  setup <- .npregbw_nomad_bw_setup(xdat = eval.zdat, template = template)
  ncon <- length(setup$cont_idx)
  ncat <- length(setup$cat_idx)
  ndeg <- length(degree.search$start.degree)
  nomad.nmulti <- if (is.null(opt.args$nmulti)) npDefaultNmulti(NCOL(eval.zdat)) else npValidateNmulti(opt.args$nmulti[1L])

  cont_lower <- npGetScaleFactorSearchLower(
    template,
    argname = "template$scale.factor.search.lower"
  )
  bw_lower <- c(rep.int(cont_lower, ncon), rep.int(0, ncat))
  bw_upper <- c(rep.int(1e6, ncon), setup$cat_upper * setup$bandwidth.scale.categorical)

  x0 <- c(
    .np_nomad_complete_start_point(
      point = if (all(template$bw == 0)) NULL else .npregbw_nomad_bw_to_point(template$bw, template = template, setup = setup),
      lower = bw_lower,
      upper = bw_upper,
      ncont = ncon
    ),
    as.integer(degree.search$start.degree)
  )
  lb <- c(bw_lower, degree.search$lower)
  ub <- c(bw_upper, degree.search$upper)
  bbin <- c(rep.int(0L, ncon + ncat), rep.int(1L, ndeg))
  baseline.record <- NULL
  nomad.num.feval.total <- 0
  nomad.num.feval.fast.total <- 0

  .np_nomad_baseline_note(degree.search$start.degree)

  eval_fun <- function(point) {
    point <- as.numeric(point)
    degree <- as.integer(round(point[ncon + ncat + seq_len(ndeg)]))
    degree <- .np_degree_clip_to_grid(degree, degree.search$candidates)
    bw_vec <- .npregbw_nomad_point_to_bw(point[seq_len(ncon + ncat)], template = template, setup = setup)

    eval.reg.args <- reg.args
    eval.reg.args$regtype <- "lp"
    eval.reg.args$degree <- degree
    eval.reg.args$bernstein.basis <- degree.search$bernstein.basis

    tbw <- .npscoefbw_build_scbandwidth(
      xdat = xdat,
      ydat = ydat,
      zdat = zdat,
      bws = bw_vec,
      bandwidth.compute = FALSE,
      reg.args = eval.reg.args
    )

    out <- .npscoefbw_eval_only(
      xdat = xdat,
      ydat = ydat,
      zdat = zdat,
      bws = tbw,
      invalid.penalty = "baseline",
      penalty.multiplier = if (is.null(opt.args$penalty.multiplier)) 10 else opt.args$penalty.multiplier
    )
    nomad.num.feval.total <<- nomad.num.feval.total + as.numeric(out$num.feval[1L])
    nomad.num.feval.fast.total <<- nomad.num.feval.fast.total + as.numeric(out$num.feval.fast[1L])

    list(
      objective = out$objective,
      degree = degree,
      num.feval = out$num.feval
    )
  }

  build_payload <- function(point, best_record, solution, interrupted) {
    point <- as.numeric(point)
    degree <- as.integer(best_record$degree)
    bw_vec <- .npregbw_nomad_point_to_bw(point[seq_len(ncon + ncat)], template = template, setup = setup)
    powell.elapsed <- NA_real_

    build_direct_payload <- function() {
      final.reg.args <- reg.args
      final.reg.args$regtype <- "lp"
      final.reg.args$degree <- degree
      final.reg.args$bernstein.basis <- degree.search$bernstein.basis

      tbw <- .npscoefbw_build_scbandwidth(
        xdat = xdat,
        ydat = ydat,
        zdat = zdat,
        bws = bw_vec,
        bandwidth.compute = FALSE,
        reg.args = final.reg.args
      )
      tbw$fval <- as.numeric(best_record$objective)
      tbw$ifval <- as.numeric(best_record$objective)
      tbw$num.feval <- as.numeric(nomad.num.feval.total)
      tbw$num.feval.fast <- as.numeric(nomad.num.feval.fast.total)
      tbw$numimp <- 0
      tbw$fval.vector <- as.numeric(best_record$objective)
      tbw$total.time <- NA_real_
      if (!is.null(tbw$method) && length(tbw$method))
        tbw$pmethod <- bwmToPrint(as.character(tbw$method[1L]))
      tbw
    }

    direct.payload <- build_direct_payload()
    direct.objective <- as.numeric(best_record$objective)

    if (identical(degree.search$engine, "nomad+powell")) {
      hot.reg.args <- reg.args
      hot.reg.args$regtype <- "lp"
      hot.reg.args$degree <- degree
      hot.reg.args$bernstein.basis <- degree.search$bernstein.basis
      hot.opt.args <- .np_nomad_powell_hotstart_opt_args(
        opt.args,
        strategy = "single_iteration",
        remin = isTRUE(opt.args$powell.remin)
      )
      powell.start <- proc.time()[3L]
      hot.payload <- .np_nomad_with_powell_progress(
        degree = degree,
        best_record = best_record,
        expr = local({
          .npscoefbw_run_fixed_degree(
            xdat = xdat,
            ydat = ydat,
            zdat = zdat,
            bws = bw_vec,
            reg.args = hot.reg.args,
            opt.args = hot.opt.args
          )
        })
      )
      powell.elapsed <- proc.time()[3L] - powell.start
      direct.payload$num.feval <- as.numeric(direct.payload$num.feval[1L]) + as.numeric(hot.payload$num.feval[1L])
      direct.payload$num.feval.fast <- as.numeric(direct.payload$num.feval.fast[1L]) + as.numeric(hot.payload$num.feval.fast[1L])
      hot.payload$num.feval <- direct.payload$num.feval
      hot.payload$num.feval.fast <- direct.payload$num.feval.fast
      if (!is.null(hot.payload$method) && length(hot.payload$method))
        hot.payload$pmethod <- bwmToPrint(as.character(hot.payload$method[1L]))
      hot.objective <- as.numeric(hot.payload$fval[1L])
      if (is.finite(hot.objective) &&
          .np_degree_better(hot.objective, direct.objective, direction = "min")) {
        return(list(payload = hot.payload, objective = hot.objective, powell.time = powell.elapsed))
      }
    }

    list(payload = direct.payload, objective = direct.objective, powell.time = powell.elapsed)
  }

  .np_nomad_search(
    engine = degree.search$engine,
    baseline_record = baseline.record,
    start_degree = degree.search$start.degree,
    x0 = x0,
    bbin = bbin,
    lb = lb,
    ub = ub,
    eval_fun = eval_fun,
    build_payload = build_payload,
    direction = "min",
    objective_name = "fval",
    nmulti = nomad.nmulti,
    nomad.inner.nmulti = nomad.inner.nmulti,
    random.seed = if (!is.null(opt.args$random.seed)) opt.args$random.seed else 42L,
    remin = isTRUE(opt.args$nomad.remin),
    degree_spec = list(
      initial = degree.search$start.degree,
      lower = degree.search$lower,
      upper = degree.search$upper,
      basis = degree.search$basis,
      nobs = degree.search$nobs,
      user_supplied = degree.search$start.user
    )
  )
}

.npscoefbw_degree_search_controls <- function(regtype,
                                              regtype.named,
                                              cv.iterate,
                                              cv.iterate.named,
                                              bandwidth.compute,
                                              ncon,
                                              nobs,
                                              basis,
                                              degree.select,
                                              search.engine,
                                              degree.min,
                                              degree.max,
                                              degree.start,
                                              degree.restarts,
                                              degree.max.cycles,
                                              degree.verify,
                                              bernstein.basis,
                                              bernstein.named) {
  degree.select <- match.arg(degree.select, c("manual", "coordinate", "exhaustive"))
  if (identical(degree.select, "manual"))
    return(NULL)
  search.engine <- .npscoefbw_nomad_controls(search.engine)

  regtype.requested <- if (isTRUE(regtype.named)) match.arg(regtype, c("lc", "ll", "lp")) else "lc"
  if (!identical(regtype.requested, "lp"))
    stop("automatic degree search currently requires regtype='lp'")
  if (!isTRUE(bandwidth.compute))
    stop("automatic degree search requires bandwidth.compute=TRUE")
  if (isTRUE(cv.iterate.named) && isTRUE(cv.iterate))
    stop("automatic degree search currently requires cv.iterate=FALSE")
  if (ncon < 1L)
    stop("automatic degree search requires at least one continuous smoothing predictor")

  bern.auto <- if (isTRUE(bernstein.named)) bernstein.basis else TRUE
  bern.auto <- npValidateGlpBernstein(regtype = "lp", bernstein.basis = bern.auto)

  bounds <- .np_degree_normalize_bounds(
    ncon = ncon,
    degree.min = degree.min,
    degree.max = degree.max,
    default.max = 3L
  )

  baseline.degree <- rep.int(0L, ncon)
  default.start.degree <- if (identical(search.engine, "cell")) {
    baseline.degree
  } else {
    rep.int(1L, ncon)
  }
  start.degree <- if (is.null(degree.start)) {
    pmax(bounds$lower, pmin(bounds$upper, default.start.degree))
  } else {
    start.raw <- npValidateGlpDegree(regtype = "lp", degree = degree.start, ncon = ncon, argname = "degree.start")
    out.of.range <- vapply(seq_len(ncon), function(j) !(start.raw[j] %in% bounds$candidates[[j]]), logical(1))
    if (any(out.of.range))
      stop("degree.start must lie within the searched degree candidates for every continuous smoothing predictor")
    start.raw
  }

  list(
    method = if (identical(search.engine, "cell")) degree.select else search.engine,
    engine = search.engine,
    candidates = bounds$candidates,
    lower = bounds$lower,
    upper = bounds$upper,
    baseline.degree = baseline.degree,
    start.degree = start.degree,
    start.user = !is.null(degree.start),
    basis = if (missing(basis) || is.null(basis)) "glp" else as.character(basis[1L]),
    nobs = as.integer(nobs[1L]),
    restarts = npValidateNonNegativeInteger(degree.restarts, "degree.restarts"),
    max.cycles = npValidatePositiveInteger(degree.max.cycles, "degree.max.cycles"),
    verify = npValidateScalarLogical(degree.verify, "degree.verify"),
    bernstein.basis = bern.auto
  )
}

.npscoefbw_attach_degree_search <- function(bws, search_result) {
  metadata <- list(
    mode = search_result$method,
    verify = isTRUE(search_result$verify),
    completed = isTRUE(search_result$completed),
    certified = isTRUE(search_result$certified),
    interrupted = isTRUE(search_result$interrupted),
    baseline.degree = search_result$baseline$degree,
    baseline.fval = search_result$baseline$objective,
    best.degree = search_result$best$degree,
    best.fval = search_result$best$objective,
    nomad.time = search_result$nomad.time,
    powell.time = search_result$powell.time,
    optim.time = search_result$optim.time,
    n.unique = search_result$n.unique,
    n.visits = search_result$n.visits,
    n.cached = search_result$n.cached,
    grid.size = search_result$grid.size,
    best.restart = search_result$best.restart,
    restart.starts = search_result$restart.starts,
    restart.degree.starts = search_result$restart.degree.starts,
    restart.bandwidth.starts = search_result$restart.bandwidth.starts,
    restart.start.info = search_result$restart.start.info,
    restart.results = search_result$restart.results,
    trace = search_result$trace
  )

  if (!is.null(search_result$nomad.time))
    bws$nomad.time <- as.numeric(search_result$nomad.time[1L])
  if (!is.null(search_result$powell.time))
    bws$powell.time <- as.numeric(search_result$powell.time[1L])
  if (!is.null(search_result$optim.time) && is.finite(search_result$optim.time))
    bws$total.time <- as.numeric(search_result$optim.time[1L])
  bws <- .np_attach_nomad_restart_summary(bws, search_result)
  bws$degree.search <- metadata
  bws
}

npscoefbw.default <-
  function(xdat = stop("invoked without data 'xdat'"),
           ydat = stop("invoked without data 'ydat'"),
           zdat = NULL,
           bws,
           backfit.iterate,
           backfit.maxiter,
           backfit.tol,
           bandwidth.compute = TRUE,
           basis,
           bernstein.basis,
           bwmethod,
           bwscaling,
           bwtype,
           ckerbound,
           ckerlb,
           ckerorder,
           ckertype,
           ckerub,
           cv.iterate,
           cv.num.iterations,
           degree,
           degree.select = c("manual", "coordinate", "exhaustive"),
           search.engine = c("nomad+powell", "cell", "nomad"),
           nomad = FALSE,
           nomad.nmulti = 0L,
           degree.min = NULL,
           degree.max = NULL,
           degree.start = NULL,
           degree.restarts = 0L,
           degree.max.cycles = 20L,
           degree.verify = FALSE,
           nmulti,
           nomad.remin = FALSE,
           powell.remin = TRUE,
           okertype,
           optim.abstol,
           optim.maxattempts,
           optim.maxit,
           optim.method,
           optim.reltol,
           random.seed,
           regtype,
           ukertype,
           scale.factor.init.lower = 0.1,
           scale.factor.init.upper = 2.0,
           scale.factor.init = 0.5,
           lbd.init = 0.5,
           hbd.init = 1.5,
           dfac.init = 1.0,
           scale.factor.search.lower = NULL,
           ...){

    if (!missing(bwmethod) && identical(match.arg(bwmethod, c("cv.ls", "manual")), "manual") &&
        missing(bws))
      stop("bwmethod='manual' requires argument 'bws'")

    miss.z <- missing(zdat)
    xdat <- toFrame(xdat)
    
    if (!(is.vector(ydat) || is.factor(ydat)))
      stop("'ydat' must be a vector or a factor")

    if(!miss.z)
      zdat <- toFrame(zdat)

    mc <- match.call(expand.dots = FALSE)
    mc.names <- names(mc)
    nomad.shortcut <- .np_prepare_nomad_shortcut(
      nomad = nomad,
      call_names = mc.names,
      preset = list(
        regtype = "lp",
        search.engine = "nomad+powell",
        degree.select = "coordinate",
        bernstein.basis = TRUE,
        degree.min = 0L,
        degree.max = 10L,
        degree.verify = FALSE,
        bwtype = "fixed"
      ),
      values = list(
        regtype = if ("regtype" %in% mc.names) regtype else NULL,
        search.engine = if ("search.engine" %in% mc.names) search.engine else NULL,
        degree.select = if ("degree.select" %in% mc.names) degree.select else NULL,
        bernstein.basis = if ("bernstein.basis" %in% mc.names) bernstein.basis else NULL,
        degree.min = if ("degree.min" %in% mc.names) degree.min else NULL,
        degree.max = if ("degree.max" %in% mc.names) degree.max else NULL,
        degree.verify = if ("degree.verify" %in% mc.names) degree.verify else NULL,
        bwtype = if ("bwtype" %in% mc.names) bwtype else NULL,
        degree = if ("degree" %in% mc.names) degree else NULL
      ),
      where = "npscoefbw"
    )

    if (isTRUE(nomad.shortcut$enabled)) {
      if ("degree" %in% mc.names)
        stop("nomad=TRUE does not support an explicit degree; remove degree or set nomad=FALSE")
      if ("regtype" %in% mc.names &&
          !identical(as.character(match.arg(nomad.shortcut$values$regtype, c("lc", "ll", "lp")))[1L], "lp"))
        stop("nomad=TRUE requires regtype='lp'")
      if ("bwtype" %in% mc.names &&
          !identical(as.character(match.arg(nomad.shortcut$values$bwtype, c("fixed", "generalized_nn", "adaptive_nn")))[1L], "fixed"))
        stop("nomad=TRUE currently requires bwtype='fixed'")
      if ("degree.select" %in% mc.names &&
          identical(as.character(match.arg(nomad.shortcut$values$degree.select, c("manual", "coordinate", "exhaustive")))[1L], "manual"))
        stop("nomad=TRUE requires automatic degree search; use degree.select='coordinate' or 'exhaustive'")
      if ("search.engine" %in% mc.names &&
          !(as.character(match.arg(nomad.shortcut$values$search.engine, c("nomad+powell", "cell", "nomad")))[1L] %in%
              c("nomad", "nomad+powell")))
        stop("nomad=TRUE requires search.engine='nomad' or 'nomad+powell'")
      if ("bernstein.basis" %in% mc.names &&
          !isTRUE(npValidateGlpBernstein(regtype = "lp",
                                        bernstein.basis = nomad.shortcut$values$bernstein.basis)))
        stop("nomad=TRUE currently requires bernstein.basis=TRUE")
      if ("degree.verify" %in% mc.names &&
          isTRUE(npValidateScalarLogical(nomad.shortcut$values$degree.verify, "degree.verify")))
        stop("nomad=TRUE currently requires degree.verify=FALSE")
    }

    regtype.named <- isTRUE(nomad.shortcut$enabled) || any(mc.names == "regtype")
    bernstein.named <- isTRUE(nomad.shortcut$enabled) || any(mc.names == "bernstein.basis")
    cv.iterate.named <- any(mc.names == "cv.iterate")
    regtype.value <- if (!is.null(nomad.shortcut$values$regtype)) nomad.shortcut$values$regtype else "lc"
    degree.select.value <- if (!is.null(nomad.shortcut$values$degree.select)) nomad.shortcut$values$degree.select else "manual"
    degree.search <- .npscoefbw_degree_search_controls(
      regtype = regtype.value,
      regtype.named = regtype.named,
      cv.iterate = cv.iterate,
      cv.iterate.named = cv.iterate.named,
      bandwidth.compute = bandwidth.compute,
      ncon = sum(if (miss.z) untangle(xdat)$icon else untangle(zdat)$icon),
      nobs = NROW(xdat),
      basis = if ("basis" %in% mc.names) basis else "glp",
      degree.select = degree.select.value,
      search.engine = if (!is.null(nomad.shortcut$values$search.engine)) nomad.shortcut$values$search.engine else "nomad+powell",
      degree.min = nomad.shortcut$values$degree.min,
      degree.max = nomad.shortcut$values$degree.max,
      degree.start = if ("degree.start" %in% mc.names) degree.start else NULL,
      degree.restarts = if ("degree.restarts" %in% mc.names) degree.restarts else 0L,
      degree.max.cycles = if ("degree.max.cycles" %in% mc.names) degree.max.cycles else 20L,
      degree.verify = if (!is.null(nomad.shortcut$values$degree.verify)) nomad.shortcut$values$degree.verify else FALSE,
      bernstein.basis = if (!is.null(nomad.shortcut$values$bernstein.basis)) nomad.shortcut$values$bernstein.basis else bernstein.basis,
      bernstein.named = bernstein.named
    )
    nomad.inner.named <- "nomad.nmulti" %in% mc.names
    nomad.inner.nmulti <- if (nomad.inner.named) {
      npValidateNonNegativeInteger(nomad.nmulti, "nomad.nmulti")
    } else {
      0L
    }
    if (nomad.inner.named &&
        (is.null(degree.search) || !(degree.search$engine %in% c("nomad", "nomad+powell")))) {
      stop("nomad.nmulti is only supported when regtype='lp', automatic degree search is active, and search.engine is 'nomad' or 'nomad+powell'")
    }
    degree.setup <- npSetupGlpDegree(
      regtype = regtype.value,
      degree = if ("degree" %in% mc.names) degree else NULL,
      ncon = sum(if (miss.z) untangle(xdat)$icon else untangle(zdat)$icon),
      degree.select = degree.select.value
    )
    scale.factor.search.lower <- npResolveScaleFactorLowerBound(scale.factor.search.lower)

    ## first grab dummy args for scbandwidth() and perform 'bootstrap'
    ## bandwidth call

    margs <- c("regtype", "basis", "degree", "bernstein.basis",
               "bwmethod", "bwscaling", "bwtype", "ckertype", "ckerorder",
               "ckerbound", "ckerlb", "ckerub", "ukertype", "okertype",
               "scale.factor.search.lower")

    m <- match(margs, mc.names, nomatch = 0)
    any.m <- any(m != 0)

    sbw.args <- list(
      bw = bws,
      nobs = dim(xdat)[1],
      xdati = untangle(xdat),
      ydati = untangle(data.frame(ydat)),
      zdati = untangle(zdat),
      xnames = names(xdat),
      ynames = deparse(substitute(ydat)),
      znames = names(zdat),
      bandwidth.compute = bandwidth.compute
    )
    if (any.m) {
      nms <- mc.names[m]
      sbw.args[nms] <- mget(nms, envir = environment(), inherits = FALSE)
    }
    if (!("degree" %in% names(sbw.args)) && !is.null(degree.setup))
      sbw.args$degree <- degree.setup
    reg.args <- sbw.args[setdiff(names(sbw.args), c("bw", "nobs", "xdati", "ydati", "zdati", "xnames", "ynames", "znames", "bandwidth.compute"))]
    if (!is.null(degree.search))
      reg.args$bernstein.basis <- degree.search$bernstein.basis
    tbw <- do.call(scbandwidth, sbw.args)
    tbw <- npSetScaleFactorSearchLower(tbw, scale.factor.search.lower)

    ## next grab dummies for actual bandwidth selection and perform call
    margs <- c("zdat",
               "nmulti",
               "powell.remin",
               "random.seed",
               "scale.factor.init.lower", "scale.factor.init.upper", "scale.factor.init",
               "lbd.init", "hbd.init", "dfac.init",
               "scale.factor.search.lower",
               "cv.iterate",
               "cv.num.iterations",
               "backfit.iterate",
               "backfit.maxiter",
               "backfit.tol",
               "optim.method", "optim.maxattempts",
               "optim.reltol", "optim.abstol", "optim.maxit")
    m <- match(margs, mc.names, nomatch = 0)
    any.m <- any(m != 0)

    if (any.m) {
      nms <- mc.names[m]
      opt.args <- mget(nms, envir = environment(), inherits = FALSE)
    } else {
      opt.args <- list()
    }
    opt.args <- c(list(bandwidth.compute = bandwidth.compute), opt.args)
    reg.args$scale.factor.search.lower <- scale.factor.search.lower
    opt.args$scale.factor.search.lower <- scale.factor.search.lower

    if (!is.null(degree.search)) {
      if (identical(degree.search$engine, "cell")) {
        eval_fun <- function(degree.vec) {
          cell.reg.args <- reg.args
          cell.reg.args$regtype <- "lp"
          cell.reg.args$degree <- as.integer(degree.vec)
          cell.reg.args$bernstein.basis <- degree.search$bernstein.basis
          cell.bws <- .npscoefbw_run_fixed_degree(
            xdat = xdat,
            ydat = ydat,
            zdat = if (miss.z) NULL else zdat,
            bws = bws,
            reg.args = cell.reg.args,
            opt.args = opt.args
          )
          list(
            objective = as.numeric(cell.bws$fval[1L]),
            payload = cell.bws,
            num.feval = if (!is.null(cell.bws$num.feval)) as.numeric(cell.bws$num.feval[1L]) else NA_real_
          )
        }

        search.result <- .np_degree_search(
          method = degree.search$method,
          candidates = degree.search$candidates,
          baseline_degree = degree.search$baseline.degree,
          start_degree = degree.search$start.degree,
          restarts = degree.search$restarts,
          max_cycles = degree.search$max.cycles,
          verify = degree.search$verify,
          eval_fun = eval_fun,
          direction = "min",
          trace_level = "full",
          objective_name = "fval"
        )
      } else {
        search.result <- .npscoefbw_nomad_search(
          xdat = xdat,
          ydat = ydat,
          zdat = if (miss.z) NULL else zdat,
          bws = bws,
          reg.args = reg.args,
          opt.args = opt.args,
          degree.search = degree.search,
          nomad.inner.nmulti = nomad.inner.nmulti
        )
      }
      tbw <- .npscoefbw_attach_degree_search(
        bws = search.result$best_payload,
        search_result = search.result
      )
    } else {
      scbw.args <- c(list(xdat = xdat, ydat = ydat, bws = tbw), opt.args)
      if (!miss.z)
        scbw.args$zdat <- zdat
      tbw <- .np_progress_select_bandwidth_enhanced(
        "Selecting smooth coefficient bandwidth",
        do.call(npscoefbw.scbandwidth, scbw.args)
      )
    }

    environment(mc) <- parent.frame()
    tbw$call <- mc
    tbw <- .np_attach_nomad_shortcut(tbw, nomad.shortcut$metadata)

    return(tbw)
    
  }

test_that("rbandwidth plot bootstrap supports gradients across methods", {
  skip_if_not_installed("np")

  set.seed(322)
  n <- 60
  x <- runif(n)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.15)
  bw <- npregbw(y ~ x, regtype = "ll", nmulti = 1)

  run_plot <- function(method) {
    suppressWarnings(
      plot(
        bw,
        output = "data",
        perspective = FALSE,
        gradients = TRUE,
        errors = "bootstrap",
        bootstrap = method,
        B = 9
      )
    )
  }

  methods <- c("inid", "fixed", "geom", "wild")
  for (m in methods) {
    set.seed(9320 + match(m, methods))
    out <- run_plot(method = m)
    expect_type(out, "list")
    expect_true(length(out) > 0)
  }
})

test_that("plot bootstrap accepts wild selector", {
  skip_if_not_installed("np")

  set.seed(323)
  n <- 80
  x <- runif(n)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.1)
  bw <- npregbw(y ~ x, bws = 0.2, bandwidth.compute = FALSE)

  out <- suppressWarnings(
    plot(
      bw,
      xdat = data.frame(x = x),
      ydat = y,
      output = "data",
      perspective = FALSE,
      errors = "bootstrap",
      bootstrap = "wild",
      B = 9
    )
  )

  expect_type(out, "list")
  expect_true(length(out) > 0)
})

test_that("scbandwidth bootstrap non-coef path avoids npscoef refits", {
  skip_if_not_installed("np")

  set.seed(3240)
  n <- 65
  x <- runif(n)
  z <- runif(n)
  y <- sin(2 * pi * z) + x * (1 + z) + rnorm(n, sd = 0.1)
  xdat <- data.frame(x = x)
  zdat <- data.frame(z = z)
  bw <- npscoefbw(xdat = xdat, ydat = y, zdat = zdat, regtype = "lc", nmulti = 1)

  np.ns <- asNamespace("np")
  ctr <- new.env(parent = emptyenv())
  ctr$n <- 0L
  trace(
    what = "npscoef",
    where = np.ns,
    tracer = bquote(.(ctr)$n <- .(ctr)$n + 1L),
    print = FALSE
  )
  on.exit(untrace("npscoef", where = np.ns), add = TRUE)

  out <- suppressWarnings(
    plot(
      bw,
      xdat = xdat,
      ydat = y,
      zdat = zdat,
      coef = FALSE,
      perspective = FALSE,
      output = "data",
      errors = "bootstrap",
      bootstrap = "inid",
      B = 7
    )
  )

  expect_type(out, "list")
  expect_true(length(out) > 0)
  expect_identical(ctr$n, 0L)
})

test_that("inid lc chunked generation matches explicit chunked draws", {
  skip_if_not_installed("np")

  set.seed(324)
  n <- 30
  x <- runif(n)
  y <- cos(2 * pi * x) + rnorm(n, sd = 0.1)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 17))
  bw <- npregbw(y ~ x, regtype = "lc", bws = c(0.2), bandwidth.compute = FALSE)
  H <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix")
  B <- 11L

  fast.fun <- getFromNamespace(".np_inid_lc_boot_from_hat", "np")
  old.chunk <- getOption("np.plot.inid.chunk.size")
  on.exit(options(np.plot.inid.chunk.size = old.chunk), add = TRUE)
  options(np.plot.inid.chunk.size = 3L)

  set.seed(7001)
  auto <- fast.fun(H = H, ydat = y, B = B, counts = NULL)

  set.seed(7001)
  counts <- matrix(NA_integer_, nrow = n, ncol = B)
  start <- 1L
  while (start <= B) {
    stopi <- min(B, start + 3L - 1L)
    bsz <- stopi - start + 1L
    counts[, start:stopi] <- rmultinom(n = bsz, size = n, prob = rep.int(1 / n, n))
    start <- stopi + 1L
  }
  explicit <- fast.fun(H = H, ydat = y, B = B, counts = counts)

  expect_equal(auto$t, explicit$t, tolerance = 1e-12)
  expect_equal(auto$t0, explicit$t0, tolerance = 1e-12)
})

test_that("lp degree is used by regression inid helper construction", {
  skip_if_not_installed("np")

  set.seed(3264)
  n <- 75
  x1 <- runif(n)
  x2 <- runif(n)
  tx <- data.frame(x1 = x1, x2 = x2)
  y <- sin(3 * x1) + 0.25 * x2^2 + rnorm(n, sd = 0.04)
  ex <- tx[seq_len(20), , drop = FALSE]
  B <- 6L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  bw2 <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = c(2L, 2L),
    bws = c(0.28, 0.28),
    bandwidth.compute = FALSE
  )
  bw3 <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = c(3L, 3L),
    bws = c(0.28, 0.28),
    bandwidth.compute = FALSE
  )

  fast.fun <- getFromNamespace(".np_inid_boot_from_regression", "np")
  out2 <- fast.fun(xdat = tx, exdat = ex, bws = bw2, ydat = y, B = B, counts = counts)
  out3 <- fast.fun(xdat = tx, exdat = ex, bws = bw3, ydat = y, B = B, counts = counts)

  fit2 <- npreg(txdat = tx, tydat = y, exdat = ex, bws = bw2, gradients = FALSE, warn.glp.gradient = FALSE)$mean
  fit3 <- npreg(txdat = tx, tydat = y, exdat = ex, bws = bw3, gradients = FALSE, warn.glp.gradient = FALSE)$mean

  expect_equal(out2$t0, fit2, tolerance = 1e-6)
  expect_equal(out3$t0, fit3, tolerance = 1e-6)
  expect_gt(max(abs(out2$t0 - out3$t0)), 1e-6)
})

test_that("npreg plot bootstrap inid supports lp basis variants", {
  skip_if_not_installed("np")

  set.seed(327)
  n <- 70
  x <- runif(n)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.1)
  tx <- data.frame(x = x)

  for (basis in c("additive", "tensor", "glp")) {
    bw <- npregbw(
      xdat = tx,
      ydat = y,
      regtype = "lp",
      degree = 2,
      basis = basis,
      bws = 0.2,
      bandwidth.compute = FALSE
    )

    out <- suppressWarnings(
      plot(
        bw,
        xdat = tx,
        ydat = y,
        output = "data",
        perspective = FALSE,
        errors = "bootstrap",
        bootstrap = "inid",
        B = 7
      )
    )
    expect_type(out, "list")
    expect_true(length(out) > 0)
  }
})

test_that("npscoef fixed no-error plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npscoef_fixed_none_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(2601)

  n <- 70L
  x <- data.frame(x = runif(n, -1, 1))
  z <- data.frame(z = runif(n, -0.5, 1.5))
  y <- (1 + 0.5 * x$x) * sin(z$z) + rnorm(n, sd = 0.15)
  cases <- list(
    lc = list(regtype = "lc"),
    ll = list(regtype = "ll"),
    lp1 = list(regtype = "lp", degree = 1L),
    lp2 = list(regtype = "lp", degree = 2L)
  )

  for (case.name in names(cases)) {
    bw.args <- c(
      list(
        xdat = x,
        zdat = z,
        ydat = y,
        bws = 0.45,
        bandwidth.compute = FALSE,
        bwtype = "fixed"
      ),
      cases[[case.name]]
    )
    bw <- do.call(npscoefbw, bw.args)

    old <- suppressWarnings(plot(
      bw,
      xdat = x,
      zdat = z,
      ydat = y,
      output = "data",
      errors = "none",
      neval = 7L,
      perspective = TRUE
    ))
    candidate <- proto(
      bw,
      xdat = x,
      zdat = z,
      ydat = y,
      neval = 7L
    )
    stages <- proto(
      bw,
      xdat = x,
      zdat = z,
      ydat = y,
      neval = 7L,
      return.stages = TRUE
    )

    expect_named(candidate, names(old), info = case.name)
    expect_s3_class(candidate$r1, "smoothcoefficient")
    expect_named(candidate$r1, names(old$r1), info = case.name)
    expect_equal(candidate$r1$eval, old$r1$eval, info = case.name)
    expect_equal(candidate$r1$mean, old$r1$mean, info = case.name)
    expect_equal(candidate$r1$merr, old$r1$merr, info = case.name)
    expect_equal(candidate$r1$bias, old$r1$bias, info = case.name)
    expect_named(stages, c("state", "target_grid", "evaluator", "intervals", "bootstrap", "plot_data"),
                 info = case.name)
    expect_equal(stages$plot_data, candidate, info = case.name)
    expect_null(stages$intervals, info = case.name)
    expect_null(stages$bootstrap, info = case.name)
    expect_equal(stages$evaluator$mean, old$r1$mean, info = case.name)
  }
})

test_that("npscoef plot prototype fails early outside its explicit fixed surface slice", {
  proto <- getFromNamespace(".np_plot_proto_npscoef_fixed_none_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(2602)

  n <- 55L
  x <- data.frame(x = runif(n))
  z <- data.frame(z = factor(rbinom(n, 1L, 0.5)))
  y <- x$x + rnorm(n, sd = 0.1)
  bw <- npscoefbw(
    xdat = x,
    zdat = z,
    ydat = y,
    bws = 0.45,
    bandwidth.compute = FALSE,
    regtype = "lc",
    bwtype = "fixed"
  )

  expect_error(
    proto(bw, xdat = x, zdat = z, ydat = y, neval = 7L),
    "continuous/ordered surface variables",
    fixed = TRUE
  )

  expect_error(
    proto(bw, neval = 7L),
    "explicit xdat, ydat, and zdat",
    fixed = TRUE
  )
})

test_that("npscoef fixed asymptotic plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npscoef_fixed_asymptotic_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(2603)

  n <- 65L
  x <- data.frame(x = runif(n, -1, 1))
  z <- data.frame(z = runif(n, -1, 1))
  y <- (1 + x$x) * cos(z$z) + rnorm(n, sd = 0.2)
  bw <- npscoefbw(
    xdat = x,
    zdat = z,
    ydat = y,
    bws = 0.55,
    bandwidth.compute = FALSE,
    regtype = "ll",
    bwtype = "fixed"
  )

  for (band in c("pmzsd", "pointwise", "bonferroni", "simultaneous", "all")) {
    old <- suppressWarnings(plot(
      bw,
      xdat = x,
      zdat = z,
      ydat = y,
      output = "data",
      errors = "asymptotic",
      band = band,
      neval = 6L,
      perspective = TRUE
    ))
    candidate <- proto(
      bw,
      xdat = x,
      zdat = z,
      ydat = y,
      neval = 6L,
      band = band
    )
    stages <- proto(
      bw,
      xdat = x,
      zdat = z,
      ydat = y,
      neval = 6L,
      band = band,
      return.stages = TRUE
    )

    expect_equal(candidate$r1$eval, old$r1$eval, info = band)
    expect_equal(candidate$r1$mean, old$r1$mean, info = band)
    expect_equal(candidate$r1$merr, old$r1$merr, info = band)
    expect_equal(candidate$r1$bias, old$r1$bias, info = band)
    expect_identical(stages$intervals$method, "asymptotic")
    expect_identical(stages$intervals$type, band)
    expect_null(stages$bootstrap, info = band)
    expect_equal(stages$plot_data, candidate, info = band)
  }
})

test_that("npscoef fixed bootstrap plot-data prototype matches current route", {
  proto <- getFromNamespace(".np_plot_proto_npscoef_fixed_bootstrap_data", "np")
  withr::local_options(np.messages = FALSE)
  set.seed(2604)

  n <- 50L
  x <- data.frame(x = runif(n, -1, 1))
  z <- data.frame(z = runif(n, -1, 1))
  y <- (1 - 0.5 * x$x) * sin(z$z) + rnorm(n, sd = 0.25)
  bw <- npscoefbw(
    xdat = x,
    zdat = z,
    ydat = y,
    bws = 0.55,
    bandwidth.compute = FALSE,
    regtype = "ll",
    bwtype = "fixed"
  )
  cases <- expand.grid(
    method = c("wild", "inid"),
    center = c("estimate", "bias-corrected"),
    stringsAsFactors = FALSE
  )

  for (ii in seq_len(nrow(cases))) {
    method <- cases$method[ii]
    center <- cases$center[ii]
    boot.seed <- 6200L + ii
    set.seed(boot.seed)
    old <- suppressWarnings(plot(
      bw,
      xdat = x,
      zdat = z,
      ydat = y,
      output = "data",
      errors = "bootstrap",
      bootstrap = method,
      B = 11L,
      center = center,
      band = "pointwise",
      neval = 5L,
      perspective = TRUE,
      random.seed = boot.seed
    ))
    set.seed(boot.seed)
    candidate <- suppressWarnings(proto(
      bw,
      xdat = x,
      zdat = z,
      ydat = y,
      neval = 5L,
      bootstrap = method,
      B = 11L,
      center = center,
      band = "pointwise"
    ))
    set.seed(boot.seed)
    stages <- suppressWarnings(proto(
      bw,
      xdat = x,
      zdat = z,
      ydat = y,
      neval = 5L,
      bootstrap = method,
      B = 11L,
      center = center,
      band = "pointwise",
      return.stages = TRUE
    ))

    expect_equal(candidate$r1$eval, old$r1$eval, info = paste(method, center))
    expect_equal(candidate$r1$mean, old$r1$mean, info = paste(method, center))
    expect_equal(candidate$r1$merr, old$r1$merr, info = paste(method, center))
    expect_equal(candidate$r1$bias, old$r1$bias, info = paste(method, center))
    expect_identical(stages$bootstrap$method, method)
    expect_identical(stages$bootstrap$center, center)
    expect_equal(stages$plot_data, candidate, info = paste(method, center))
  }
})

test_that("npscoef basic functionality works", {
  set.seed(42)
  n <- 100
  x1 <- runif(n) # smoothing variable
  z1 <- runif(n) # parametric variable
  # Smooth coefficient model: y = beta(x1) * z1 + e
  y <- (x1^2) * z1 + rnorm(n, sd=0.1)
  
  bw <- npscoefbw(xdat=x1, zdat=z1, ydat=y, bws=0.1, bandwidth.compute=FALSE)
  
  model <- npscoef(bws=bw)
  
  expect_s3_class(model, "smoothcoefficient")
  expect_type(predict(model), "double")
  expect_output(summary(model))
})

test_that("npscoefbw records ll/lp controls", {
  set.seed(43)
  n <- 80
  x1 <- runif(n)
  z1 <- runif(n)
  y <- (0.2 + x1) * cos(2 * pi * z1) + rnorm(n, sd = 0.08)

  bw.ll <- npscoefbw(
    xdat = x1,
    zdat = z1,
    ydat = y,
    regtype = "ll",
    bws = 0.15,
    bandwidth.compute = FALSE
  )
  expect_identical(bw.ll$regtype, "ll")

  bw.lp <- npscoefbw(
    xdat = x1,
    zdat = z1,
    ydat = y,
    regtype = "lp",
    basis = "tensor",
    degree = 2L,
    bws = 0.15,
    bandwidth.compute = FALSE
  )
  expect_identical(bw.lp$regtype, "lp")
  expect_identical(bw.lp$basis, "tensor")
})

test_that("npscoef direct route accepts omitted tzdat with explicit bandwidths", {
  x <- data.frame(x = c(0.1, 0.3, 0.5, 0.8, 0.9, 1.1))
  y <- c(0.12, 0.25, 0.44, 0.61, 0.73, 0.95)

  bw <- npscoefbw(xdat = x, ydat = y, bws = 0.35, bandwidth.compute = FALSE)
  fit <- npscoef(bws = bw, txdat = x, tydat = y, errors = FALSE, iterate = FALSE)

  expect_s3_class(fit, "smoothcoefficient")
  expect_equal(length(fit$mean), nrow(x))
})

test_that("npscoef explicit tzdat direct route matches stored-data route", {
  x <- data.frame(x = c(0.1, 0.3, 0.5, 0.8, 0.9, 1.1))
  z <- data.frame(z = c(1.0, 0.5, 1.5, 0.8, 1.2, 0.7))
  y <- c(0.12, 0.25, 0.44, 0.61, 0.73, 0.95)

  bw <- npscoefbw(
    xdat = x,
    zdat = z,
    ydat = y,
    bws = 0.35,
    bandwidth.compute = FALSE
  )

  fit.stored <- npscoef(bws = bw, errors = FALSE, iterate = FALSE)
  fit.direct <- npscoef(
    bws = bw,
    txdat = x,
    tydat = y,
    tzdat = z,
    errors = FALSE,
    iterate = FALSE
  )

  expect_equal(fit.direct$mean, fit.stored$mean, tolerance = 0)
  expect_equal(fit.direct$bws$bw, fit.stored$bws$bw, tolerance = 0)
})

progress_time_counter <- function(start = 0, by = 2.1) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

shadow_lines <- function(shadow) {
  vapply(shadow$trace, `[[`, character(1L), "line")
}

shadow_render_lines <- function(shadow) {
  trace <- shadow$trace[vapply(shadow$trace, `[[`, character(1L), "event") == "render"]
  vapply(trace, `[[`, character(1L), "line")
}

installed_function_text <- function(name, package = "np") {
  paste(deparse(getFromNamespace(name, package)), collapse = "\n")
}

test_that("npscoefbw adopts the generic bandwidth selection line", {
  set.seed(3240)
  n <- 28
  x <- runif(n)
  z <- runif(n)
  y <- sin(2 * pi * z) + x * (1 + z) + rnorm(n, sd = 0.1)

  old_opts <- options(np.messages = TRUE)
  on.exit(options(old_opts), add = TRUE)

  legacy <- capture_progress_shadow_trace(
    npscoefbw(
      xdat = data.frame(x = x),
      zdat = data.frame(z = z),
      ydat = y,
      regtype = "lc",
      nmulti = 2,
      optim.maxit = 3,
      cv.iterate = FALSE
    ),
    force_renderer = "legacy",
    now = progress_time_counter()
  )

  set.seed(3240)
  actual <- capture_progress_shadow_trace(
    npscoefbw(
      xdat = data.frame(x = x),
      zdat = data.frame(z = z),
      ydat = y,
      regtype = "lc",
      nmulti = 2,
      optim.maxit = 3,
      cv.iterate = FALSE
    ),
    now = progress_time_counter()
  )

  bandwidth_lines <- shadow_lines(actual)[grepl("^\\[np\\] Bandwidth selection", shadow_lines(actual))]
  render_bandwidth_lines <- shadow_render_lines(actual)[grepl("^\\[np\\] Bandwidth selection", shadow_render_lines(actual))]
  legacy_bandwidth_lines <- shadow_render_lines(legacy)[grepl("^\\[np\\] Bandwidth selection", shadow_render_lines(legacy))]

  expect_s3_class(actual$value, "scbandwidth")
  expect_equal(render_bandwidth_lines, legacy_bandwidth_lines)
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 1/2\\)$", bandwidth_lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 1/2, iteration [0-9]+, elapsed [0-9]+\\.[0-9]s\\)$", bandwidth_lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/2, elapsed [0-9]+\\.[0-9]s, 50\\.0%, eta [0-9]+\\.[0-9]s\\)$", bandwidth_lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/2, iteration [0-9]+, elapsed [0-9]+\\.[0-9]s, [0-9]+\\.[0-9]%, eta [0-9]+\\.[0-9]s\\)$", bandwidth_lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(multistart 2/2, elapsed [0-9]+\\.[0-9]s, 100\\.0%, eta 0\\.0s\\)$", bandwidth_lines)))
})

test_that("npscoefbw progress respects np.messages FALSE", {
  set.seed(3240)
  n <- 24
  x <- runif(n)
  z <- runif(n)
  y <- sin(2 * pi * z) + x * (1 + z) + rnorm(n, sd = 0.1)

  old_opts <- options(np.messages = FALSE)
  on.exit(options(old_opts), add = TRUE)

  silent <- capture_progress_shadow_trace(
    npscoefbw(
      xdat = data.frame(x = x),
      zdat = data.frame(z = z),
      ydat = y,
      regtype = "lc",
      nmulti = 1,
      optim.maxit = 2,
      cv.iterate = FALSE
    ),
    now = progress_time_counter()
  )

  expect_length(silent$trace, 0L)
})

test_that("npscoefbw cv.iterate path retains backfitting progress hooks", {
  src <- installed_function_text("npscoefbw.scbandwidth")

  expect_true(grepl("Backfitting smooth coefficient bandwidth", src, fixed = TRUE))
  expect_true(grepl("Optimizing partial residual bandwidth", src, fixed = TRUE))
  expect_true(grepl("\\.np_progress_begin\\(\"Backfitting smooth coefficient bandwidth\"", src))
  expect_true(grepl("\\.np_progress_begin\\(\"Optimizing partial residual bandwidth\"", src))
})

library(np)

test_that("npscoefbw fast counter fires only on the z-irrelevant sentinel", {
  old_opts <- options(
    np.messages = FALSE,
    np.largeh.rel.tol = 0.05,
    np.disc.upper.rel.tol = 0.05
  )
  on.exit(options(old_opts), add = TRUE)

  set.seed(77)
  n <- 200L
  x <- runif(n)
  z <- runif(n)

  dat_lc <- data.frame(
    y = rnorm(n, sd = 0.5 * sd(x)),
    x = x,
    z = z
  )
  dat_ll <- data.frame(
    y = x + rnorm(n, sd = 0.5 * sd(x)),
    x = x,
    z = z
  )

  bw_lc <- npscoefbw(y ~ x | z, data = dat_lc, regtype = "lc", bwmethod = "cv.ls", nmulti = 1)
  bw_ll <- npscoefbw(y ~ x | z, data = dat_ll, regtype = "ll", bwmethod = "cv.ls", nmulti = 1)
  bw_lp <- npscoefbw(
    y ~ x | z,
    data = dat_ll,
    regtype = "lp",
    degree = 1,
    bwmethod = "cv.ls",
    nmulti = 1
  )

  expect_true(is.finite(as.numeric(bw_lc$num.feval.fast)))
  expect_true(is.finite(as.numeric(bw_ll$num.feval.fast)))
  expect_true(is.finite(as.numeric(bw_lp$num.feval.fast)))

  expect_gt(as.integer(bw_lc$num.feval.fast), 0L)
  expect_gt(as.integer(bw_ll$num.feval.fast), 0L)
  expect_gt(as.integer(bw_lp$num.feval.fast), 0L)

  expect_equal(as.numeric(bw_ll$fval), as.numeric(bw_lp$fval), tolerance = 1e-10)
  expect_equal(as.integer(bw_ll$num.feval.fast), as.integer(bw_lp$num.feval.fast))
})

smoothcoefficient <- 
  function(bws, eval, mean, merr = NA, beta = NA,
           grad = NA, gerr = NA, resid = NA,
           ntrain, trainiseval = FALSE, residuals = FALSE,
           betas = FALSE,
           xtra = rep(NA, 6),
           timing = NA, total.time = NA,
           optim.time = NA, fit.time = NA){

    if (missing(bws) || missing(eval) || missing(ntrain))
      stop("improper invocation of smoothcoefficient constructor")

    d = list(
      bw = bws$bw,
      bws = bws,
      xnames = bws$xnames,
      ynames = bws$ynames,
      znames = bws$znames,
      nobs = if(is.data.frame(eval)) nrow(eval) else nrow(eval[[1]]),
      ndim = bws$ndim,
      nord = bws$nord,
      nuno = bws$nuno,
      ncon = bws$ncon,
      pscaling = bws$pscaling,
      ptype = bws$ptype,
      pckertype = bws$pckertype,
      pukertype = bws$pukertype,
      pokertype = bws$pokertype,
      eval = eval,
      mean = mean,
      merr = merr,
      ntrain = ntrain,
      trainiseval = trainiseval,
      residuals = residuals,
      betas = betas,
      beta = beta,
      grad = grad,
      gerr = gerr,
      resid = resid,
      R2 = xtra[1],
      MSE = xtra[2],
      MAE = xtra[3],
      MAPE = xtra[4],
      CORR = xtra[5],
      SIGN = xtra[6],
      timing = timing,
      total.time = total.time,
      optim.time = optim.time,
      fit.time = fit.time
      )

    class(d) = "smoothcoefficient"

    d
  }

print.smoothcoefficient <- function(x, digits=NULL, ...){
  cat("\nSmooth Coefficient Model",
      "\nRegression data: ", x$ntrain, " training points,",
      if (x$trainiseval) "" else paste(" and ", x$nobs," evaluation points,", sep=""),
      " in ",x$ndim," variable(s)\n",sep="")
  print(matrix(x$bw,ncol=x$ndim,dimnames=list(paste(x$pscaling,":",sep=""),
                                  if(is.null(x$znames)) x$xnames else x$znames)))

  ## print(matrix(x$bw,ncol=x$ndim,dimnames=list(paste(x$pscaling,":",sep=""),x$xnames)))

  cat(genRegEstStr(x))

  cat(genBwKerStrs(x$bws))
  cat('\n\n')  

  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

coef.smoothcoefficient <- function(object, ...) {
  tc <- object$beta
  if(object$betas)
    dimnames(tc) <- list(NULL,c("Intercept",object$xnames))
  return(tc)
}

fitted.smoothcoefficient <- function(object, ...){
 object$mean 
}
residuals.smoothcoefficient <- function(object, ...) {
 if(object$residuals) { return(object$resid) } else { return(npscoef(bws = object$bws, residuals =TRUE)$resid) } 
}
se.smoothcoefficient <- function(x){ x$merr }
predict.smoothcoefficient <- function(object, se.fit = FALSE, ...) {
  se.fit <- npValidateScalarLogical(se.fit, "se.fit")
  dots <- list(...)
  has.formula.route <- !is.null(object$bws$formula)

  if ((!is.null(dots$exdat) || !is.null(dots$ezdat)) && !is.null(dots$newdata)) {
    dots$newdata <- NULL
  } else if (!has.formula.route && is.null(dots$exdat) && !is.null(dots$newdata)) {
    nd <- toFrame(dots$newdata)
    if (!is.null(object$bws$znames)) {
      need <- c(object$bws$xnames, object$bws$znames)
      if (!all(need %in% names(nd)))
        stop("'newdata' must include columns: ", paste(need, collapse = ", "))
      dots$exdat <- nd[, object$bws$xnames, drop = FALSE]
      dots$ezdat <- nd[, object$bws$znames, drop = FALSE]
    } else {
      if (!all(object$bws$xnames %in% names(nd)))
        stop("'newdata' must include columns: ", paste(object$bws$xnames, collapse = ", "))
      dots$exdat <- nd[, object$bws$xnames, drop = FALSE]
    }
    dots$newdata <- NULL
  }

  tr <- do.call(npscoef, c(list(bws = object$bws), dots))
  if(se.fit)
    return(list(fit = fitted(tr), se.fit = se(tr), 
                df = tr$nobs, residual.scale = tr$MSE))
  else
    return(fitted(tr))
}

summary.smoothcoefficient <- function(object, ...){
  cat("\nSmooth Coefficient Model",
      "\nRegression data: ", object$ntrain, " training points,",
      if (object$trainiseval) "" else paste(" and ", object$nobs," evaluation points,", sep=""),
      " in ",object$ndim," variable(s)\n",sep="")

  cat(genOmitStr(object))
  cat("\n")

  print(matrix(object$bw,ncol=object$ndim,dimnames=list(paste(object$pscaling,":",sep=""),
                                  if(is.null(object$znames)) object$xnames else object$znames)))

  cat(genRegEstStr(object))
  cat("\n")
  cat(genGofStr(object))

  cat(genBwKerStrs(object$bws))
  cat(genTimingStr(object))
  cat('\n\n')  
}

.np_progress_is_interactive <- function() {
  interactive()
}

.np_progress_now <- function() {
  as.numeric(proc.time()[["elapsed"]])
}

.np_progress_enabled <- function(domain = "general") {
  (isTRUE(.np_progress_runtime$force_enabled) || isTRUE(getOption("np.messages", TRUE))) &&
    isTRUE(.np_progress_is_interactive())
}

.np_io_pkg_prefix <- function() {
  "[np]"
}

.np_io_prefix_text <- function(text) {
  prefix <- .np_io_pkg_prefix()
  if (!is.character(text)) {
    return(text)
  }

  vapply(
    text,
    function(line) {
      if (is.na(line) || startsWith(line, prefix)) {
        return(line)
      }

      paste(prefix, line)
    },
    character(1L)
  )
}

.np_message <- function(...) {
  message(.np_io_prefix_text(paste(..., sep = "", collapse = "")))
  invisible(NULL)
}

.np_warning <- function(..., call. = TRUE, immediate. = FALSE, noBreaks. = FALSE) {
  warning(
    .np_io_prefix_text(paste(..., sep = "", collapse = "")),
    call. = call.,
    immediate. = immediate.,
    noBreaks. = noBreaks.
  )
}

.np_progress_pkg_prefix <- function() {
  .np_io_pkg_prefix()
}

.np_progress_emit <- function(line) {
  .np_message(line)
  invisible(NULL)
}

.np_progress_call_name <- function(call) {
  head <- call[[1L]]

  if (is.symbol(head)) {
    return(as.character(head))
  }

  if (is.call(head) &&
      identical(head[[1L]], as.name("::")) &&
      length(head) >= 3L &&
      is.symbol(head[[2L]]) &&
      is.symbol(head[[3L]])) {
    return(paste(as.character(head[[2L]]), as.character(head[[3L]]), sep = "::"))
  }

  ""
}

.np_progress_is_message_muffled <- function() {
  calls <- sys.calls()
  if (!length(calls)) {
    return(FALSE)
  }

  any(vapply(
    calls,
    function(call) {
      .np_progress_call_name(call) %in% c(
        "suppressMessages",
        "base::suppressMessages",
        "suppressPackageStartupMessages",
        "base::suppressPackageStartupMessages"
      )
    },
    logical(1L)
  ))
}

.np_progress_resolve_message_muffling <- function(state) {
  if (!identical(state$renderer, "single_line")) {
    return(state)
  }

  if (isTRUE(state$message_muffled_checked)) {
    return(state)
  }

  state$message_muffled <- isTRUE(.np_progress_is_message_muffled())
  state$message_muffled_checked <- TRUE
  state
}

.np_progress_fmt_num <- function(x) {
  formatC(x, digits = 1L, format = "f")
}

.np_progress_runtime <- local({
  env <- new.env(parent = emptyenv())
  env$bandwidth_depth <- 0L
  env$bandwidth_label <- NULL
  env$bandwidth_state <- NULL
  env$bandwidth_old_messages <- NULL
  env$bandwidth_context_label <- NULL
  env$fit_state <- NULL
  env$iv_depth <- 0L
  env$iv_label <- NULL
  env$iv_state <- NULL
  env$iv_old_messages <- NULL
  env$force_enabled <- FALSE
  env
})

.np_progress_bandwidth_set_context <- function(label = NULL) {
  if (is.null(label)) {
    .np_progress_runtime$bandwidth_context_label <- NULL
    return(invisible(NULL))
  }

  label <- as.character(label)[1L]
  if (is.na(label) || !nzchar(label)) {
    .np_progress_runtime$bandwidth_context_label <- NULL
    return(invisible(NULL))
  }

  .np_progress_runtime$bandwidth_context_label <- label
  invisible(NULL)
}

.np_progress_make_registry <- function() {
  env <- new.env(parent = emptyenv())
  env$next_id <- 0L
  env$active_id <- NULL
  env$last_single_line_width <- 0L
  env
}

.np_progress_registry <- .np_progress_make_registry()

.np_progress_is_rstudio_console <- function() {
  identical(.Platform$GUI, "RStudio") || identical(Sys.getenv("RSTUDIO"), "1")
}

.np_progress_reset_registry <- function() {
  .np_progress_registry$next_id <- 0L
  .np_progress_registry$active_id <- NULL
  .np_progress_registry$last_single_line_width <- 0L
  invisible(NULL)
}

.np_progress_next_id <- function() {
  .np_progress_registry$next_id <- as.integer(.np_progress_registry$next_id) + 1L
  sprintf("progress-%d", .np_progress_registry$next_id)
}

.np_progress_capability <- function(domain = "general") {
  interactive_console <- isTRUE(.np_progress_is_interactive())
  rstudio_console <- isTRUE(.np_progress_is_rstudio_console())

  list(
    domain = domain,
    interactive = interactive_console,
    rstudio = rstudio_console,
    single_line_viable = interactive_console
  )
}

.np_progress_single_line_surfaces <- function() {
  c("bandwidth", "plot_activity", "plot_bounded", "bootstrap", "lag", "iv_solve",
    "copula")
}

.np_progress_renderer_for_surface <- function(surface, capability) {
  if (surface %in% .np_progress_single_line_surfaces() &&
      isTRUE(capability$single_line_viable)) {
    return("single_line")
  }

  "legacy"
}

.np_progress_claim_owner <- function(session_id) {
  active_id <- .np_progress_registry$active_id
  if (is.null(active_id) || identical(active_id, session_id)) {
    .np_progress_registry$active_id <- session_id
    return(TRUE)
  }

  FALSE
}

.np_progress_release_owner <- function(session_id) {
  if (identical(.np_progress_registry$active_id, session_id)) {
    .np_progress_registry$active_id <- NULL
  }

  invisible(NULL)
}

.np_progress_make_snapshot <- function(state, line, event, now, done = NULL, detail = NULL, render_line = line) {
  list(
    id = state$id,
    surface = state$surface,
    renderer = state$renderer,
    label = state$label,
    kind = if (isTRUE(state$known_total)) "known" else "unknown",
    current = done,
    total = state$total,
    detail = detail,
    line = line,
    render_line = render_line,
    event = event,
    started_at = state$started,
    now = now,
    last_width = state$last_render_width
  )
}

.np_progress_render_legacy <- function(snapshot, event = c("render", "finish", "abort")) {
  event <- match.arg(event)
  .np_progress_emit(snapshot$line)
  invisible(snapshot)
}

.np_progress_single_line_connection <- function() {
  stderr()
}

.np_progress_single_line_supports_ansi <- function(con = .np_progress_single_line_connection()) {
  if (identical(.Platform$OS.type, "windows")) {
    return(FALSE)
  }

  tryCatch(isTRUE(isatty(con)), error = function(...) FALSE)
}

.np_progress_has_tty <- function() {
  any(vapply(
    list(stdin(), stdout(), stderr()),
    function(con) {
      tryCatch(isTRUE(isatty(con)), error = function(...) FALSE)
    },
    logical(1L)
  ))
}

.np_progress_parse_width <- function(value) {
  value <- suppressWarnings(as.integer(value)[1L])
  if (!is.finite(value) || is.na(value) || value < 20L) {
    return(NA_integer_)
  }

  value
}

.np_progress_terminal_width_probe <- function() {
  if (!isTRUE(.np_progress_is_interactive()) || isTRUE(.np_progress_is_rstudio_console())) {
    return(NA_integer_)
  }
  if (!isTRUE(.np_progress_has_tty())) {
    return(NA_integer_)
  }

  if (identical(.Platform$OS.type, "windows")) {
    return(.np_progress_parse_width(Sys.getenv("COLUMNS", "")))
  }

  probe <- tryCatch(
    suppressWarnings(system2(
      "sh",
      c("-c", "stty size < /dev/tty 2>/dev/null"),
      stdout = TRUE,
      stderr = TRUE
    )),
    error = function(...) character()
  )
  if (length(probe)) {
    fields <- strsplit(trimws(probe[[1L]]), "[[:space:]]+")[[1L]]
    width <- .np_progress_parse_width(fields[[length(fields)]])
    if (!is.na(width)) {
      return(width)
    }
  }

  probe <- tryCatch(
    suppressWarnings(system2("tput", "cols", stdout = TRUE, stderr = TRUE)),
    error = function(...) character()
  )
  if (length(probe)) {
    width <- .np_progress_parse_width(probe[[1L]])
    if (!is.na(width)) {
      return(width)
    }
  }

  .np_progress_parse_width(Sys.getenv("COLUMNS", ""))
}

.np_progress_output_width <- function() {
  width <- if (isTRUE(.np_progress_is_rstudio_console())) {
    .np_progress_parse_width(getOption("width", 80))
  } else {
    .np_progress_terminal_width_probe()
  }
  if (is.na(width)) {
    width <- .np_progress_parse_width(getOption("width", 80))
  }
  if (is.na(width)) {
    width <- 80L
  }

  reserve <- if (isTRUE(.np_progress_is_rstudio_console())) 4L else 0L
  max(20L, width - reserve)
}

.np_progress_ellipsize_middle <- function(text, max_width) {
  max_width <- suppressWarnings(as.integer(max_width)[1L])
  if (is.na(max_width) || max_width < 1L) {
    return("")
  }

  if (nchar(text, type = "width") <= max_width) {
    return(text)
  }

  if (max_width <= 3L) {
    return(substr("...", 1L, max_width))
  }

  keep_right <- max(1L, floor((max_width - 3L) / 2L))
  keep_left <- max_width - 3L - keep_right
  paste0(
    substr(text, 1L, keep_left),
    "...",
    substr(text, nchar(text, type = "chars") - keep_right + 1L, nchar(text, type = "chars"))
  )
}

.np_progress_compact_single_line <- function(line, max_width) {
  if (!is.character(line) || length(line) != 1L || is.na(line)) {
    return(line)
  }

  max_width <- suppressWarnings(as.integer(max_width)[1L])
  if (is.na(max_width) || max_width < 1L) {
    return(line)
  }

  if (nchar(line, type = "width") <= max_width) {
    return(line)
  }

  compacted <- line
  replacements <- list(
    c(" Preparing plot bootstrap ", " Prep plot "),
    c(" Computing regression plot fit", " Regression plot fit"),
    c(" Computing bootstrap pmzsd errors", " Bootstrap pmzsd"),
    c(" Constructing bootstrap all bands", " Bootstrap all bands"),
    c(" smooth coefficient ", " smooth coef "),
    c(" coefficient ", " coef "),
    c(" single-index ", " SI "),
    c(" regression ", " reg "),
    c(" density ", " dens "),
    c(" distribution ", " dist "),
    c(" bandwidth", " bw")
  )

  for (rule in replacements) {
    if (nchar(compacted, type = "width") <= max_width) {
      break
    }
    compacted <- gsub(rule[[1L]], rule[[2L]], compacted, fixed = TRUE)
  }

  compacted
}

.np_progress_compact_bandwidth_line <- function(line, max_width) {
  if (!is.character(line) || length(line) != 1L || is.na(line)) {
    return(line)
  }

  max_width <- suppressWarnings(as.integer(max_width)[1L])
  if (is.na(max_width) || max_width < 1L) {
    return(line)
  }

  if (nchar(line, type = "width") <= max_width) {
    return(line)
  }

  matches <- regexec("^(.+Bandwidth selection) \\((.+)\\)$", line)
  capture <- regmatches(line, matches)[[1L]]
  if (length(capture) != 3L) {
    return(line)
  }

  prefix <- capture[[2L]]
  fields <- strsplit(capture[[3L]], ", ", fixed = TRUE)[[1L]]
  if (!length(fields)) {
    return(line)
  }

  render <- function(parts) {
    sprintf("%s (%s)", prefix, paste(parts, collapse = ", "))
  }

  abbreviate_fields <- function(parts) {
    compacted <- parts
    compacted <- sub("^multistart ([0-9]+/[0-9]+)$", "\\1", compacted)
    compacted <- sub("^iteration ", "iter ", compacted)
    compacted
  }

  candidates <- list(
    render(fields),
    render(abbreviate_fields(fields))
  )

  for (candidate in candidates) {
    if (nchar(candidate, type = "width") <= max_width) {
      return(candidate)
    }
  }

  line
}

.np_progress_compact_plot_bootstrap_line <- function(line, max_width) {
  if (!is.character(line) || length(line) != 1L || is.na(line)) {
    return(line)
  }

  max_width <- suppressWarnings(as.integer(max_width)[1L])
  if (is.na(max_width) || max_width < 1L) {
    return(line)
  }

  if (nchar(line, type = "width") <= max_width) {
    return(line)
  }

  matches <- regexec("^(.+Plot bootstrap(?: \\([^)]+\\))?) ([0-9]+/[0-9]+) \\((.+)\\)$", line)
  capture <- regmatches(line, matches)[[1L]]
  if (length(capture) != 4L) {
    return(line)
  }

  prefix <- capture[[2L]]
  counter <- capture[[3L]]
  fields <- strsplit(capture[[4L]], ", ", fixed = TRUE)[[1L]]
  if (!length(fields)) {
    return(line)
  }

  render <- function(prefix_text, parts) {
    sprintf("%s %s (%s)", prefix_text, counter, paste(parts, collapse = ", "))
  }

  compact_prefix <- prefix
  compact_prefix <- sub("\\(grad index ", "(grad idx ", compact_prefix)
  compact_prefix <- sub("\\(index ", "(idx ", compact_prefix)

  compact_fields <- function(parts) {
    compacted <- parts
    compacted <- sub("^elapsed ", "elap ", compacted)
    compacted
  }

  candidates <- list(
    render(prefix, fields),
    render(compact_prefix, fields),
    render(compact_prefix, compact_fields(fields))
  )

  for (candidate in candidates) {
    if (nchar(candidate, type = "width") <= max_width) {
      return(candidate)
    }
  }

  line
}

.np_progress_compact_degree_line <- function(line, max_width) {
  if (!is.character(line) || length(line) != 1L || is.na(line)) {
    return(line)
  }

  max_width <- suppressWarnings(as.integer(max_width)[1L])
  if (is.na(max_width) || max_width < 1L) {
    return(line)
  }

  if (!grepl("Selecting polynomial degree and bandwidth", line, fixed = TRUE) &&
      !grepl("Selecting polynomial degree and bw", line, fixed = TRUE) &&
      !grepl("Selecting degree and bandwidth", line, fixed = TRUE) &&
      !grepl("Selecting degree/bandwidth", line, fixed = TRUE) &&
      !grepl("Refining bandwidth", line, fixed = TRUE)) {
    return(line)
  }

  compact <- function(text) {
    out <- text
    out <- sub("Selecting polynomial degree and bandwidth", "Degree/bw search", out, fixed = TRUE)
    out <- sub("Selecting polynomial degree and bw", "Degree/bw search", out, fixed = TRUE)
    out <- sub("Selecting degree and bandwidth", "Degree/bw search", out, fixed = TRUE)
    out <- sub("Selecting degree/bandwidth", "Degree/bw search", out, fixed = TRUE)
    out <- sub("Refining bandwidth", "Refining bw", out, fixed = TRUE)
    out <- gsub("elapsed ", "elap ", out, fixed = TRUE)
    out <- gsub("restart ", "r ", out, fixed = TRUE)
    out <- gsub("cycle ", "cy ", out, fixed = TRUE)
    out <- gsub("coord ", "c ", out, fixed = TRUE)
    out <- gsub("degree ", "deg ", out, fixed = TRUE)
    out
  }

  compact_fields <- function(text) {
    out <- compact(text)
    out <- gsub("multistart ", "ms ", out, fixed = TRUE)
    out <- gsub("iteration ", "iter ", out, fixed = TRUE)
    out
  }

  field_drop_candidates <- function(text) {
    if (!endsWith(text, ")")) {
      return(character(0L))
    }

    open <- regexpr(" (", text, fixed = TRUE)[1L]
    if (open <= 0L) {
      return(character(0L))
    }

    prefix <- substr(text, 1L, open - 1L)
    fields <- substr(text, open + 2L, nchar(text, type = "chars") - 1L)
    parts <- strsplit(fields, ", ", fixed = TRUE)[[1L]]
    if (!length(parts)) {
      return(character(0L))
    }

    render <- function(values) {
      sprintf("%s (%s)", prefix, paste(values, collapse = ", "))
    }

    out <- character(0L)
    drop_sets <- list(
      grep("^ms ", parts),
      grep("^best ", parts),
      grep("^deg ", parts)
    )
    keep <- seq_along(parts)
    for (drop in drop_sets) {
      if (length(drop)) {
        keep <- setdiff(keep, drop)
        if (length(keep)) {
          out <- c(out, render(parts[keep]))
        }
      }
    }

    out
  }

  compacted <- compact(line)
  compacted_fields <- compact_fields(line)
  candidates <- unlist(list(
    line,
    compacted,
    compacted_fields,
    field_drop_candidates(compacted_fields)
  ), use.names = FALSE)

  for (candidate in candidates) {
    if (nchar(candidate, type = "width") <= max_width) {
      return(candidate)
    }
  }

  line
}

.np_progress_fit_single_line <- function(line, max_width = .np_progress_output_width()) {
  if (!is.character(line) || length(line) != 1L || is.na(line)) {
    return(line)
  }

  max_width <- suppressWarnings(as.integer(max_width)[1L])
  if (is.na(max_width) || max_width < 1L) {
    return(line)
  }

  if (nchar(line, type = "width") <= max_width) {
    return(line)
  }

  line <- .np_progress_compact_single_line(line, max_width = max_width)
  if (nchar(line, type = "width") <= max_width) {
    return(line)
  }

  line <- .np_progress_compact_bandwidth_line(line, max_width = max_width)
  if (nchar(line, type = "width") <= max_width) {
    return(line)
  }

  line <- .np_progress_compact_plot_bootstrap_line(line, max_width = max_width)
  if (nchar(line, type = "width") <= max_width) {
    return(line)
  }

  line <- .np_progress_compact_degree_line(line, max_width = max_width)
  if (nchar(line, type = "width") <= max_width) {
    return(line)
  }

  detail_pos <- regexpr(": ", line, fixed = TRUE)[1L]
  if (detail_pos > 0L) {
    without_detail <- substr(line, 1L, detail_pos - 1L)
    if (nchar(without_detail, type = "width") <= max_width) {
      return(without_detail)
    }
    line <- without_detail
  }

  .np_progress_ellipsize_middle(line, max_width = max_width)
}

.np_progress_render_single_line <- function(snapshot, event = c("render", "finish", "abort")) {
  event <- match.arg(event)
  render_line <- snapshot$render_line
  con <- .np_progress_single_line_connection()
  use_ansi <- isTRUE(.np_progress_single_line_supports_ansi(con))
  width <- nchar(render_line, type = "width")
  global_width <- suppressWarnings(as.integer(.np_progress_registry$last_single_line_width)[1L])
  if (is.na(global_width) || global_width < 0L) {
    global_width <- 0L
  }
  if (identical(event, "finish")) {
    if (use_ansi) {
      .np_progress_registry$last_single_line_width <- 0L
      base::cat("\r\033[2K\r", file = con, sep = "")
      flush(con)
      flush.console()
      return(invisible(snapshot))
    }
    clear_width <- max(snapshot$last_width, width, global_width)
    clear_line <- if (clear_width > 0L) strrep(" ", clear_width) else ""
    .np_progress_registry$last_single_line_width <- 0L
    base::cat("\r", clear_line, "\r", file = con, sep = "")
    flush(con)
    flush.console()
    return(invisible(snapshot))
  }

  pad <- max(0L, max(snapshot$last_width, global_width) - width)
  suffix <- if (pad > 0L) paste(rep(" ", pad), collapse = "") else ""

  if (use_ansi) {
    base::cat("\r\033[2K", render_line, file = con, sep = "")
  } else {
    base::cat("\r", render_line, suffix, file = con, sep = "")
  }
  if (identical(event, "abort")) {
    .np_progress_registry$last_single_line_width <- 0L
    base::cat("\n", file = con, sep = "")
  } else {
    .np_progress_registry$last_single_line_width <- width
  }
  flush(con)
  flush.console()
  invisible(snapshot)
}

.np_progress_render <- function(state, line, event, now, done = NULL, detail = NULL) {
  if (!isTRUE(state$enabled) || !isTRUE(state$visible)) {
    return(state)
  }

  state <- .np_progress_resolve_message_muffling(state)
  if (isTRUE(state$message_muffled)) {
    return(state)
  }

  render_line <- if (identical(state$renderer, "single_line")) {
    .np_progress_fit_single_line(line)
  } else {
    line
  }

  snapshot <- .np_progress_make_snapshot(
    state = state,
    line = line,
    render_line = render_line,
    event = event,
    now = now,
    done = done,
    detail = detail
  )

  if (identical(state$renderer, "single_line")) {
    .np_progress_render_single_line(
      snapshot = snapshot,
      event = if (identical(event, "finish")) "finish" else if (identical(event, "abort")) "abort" else "render"
    )
  } else {
    .np_progress_render_legacy(
      snapshot = snapshot,
      event = if (identical(event, "finish")) "finish" else if (identical(event, "abort")) "abort" else "render"
    )
  }

  state$rendered <- TRUE
  state$last_render_width <- nchar(render_line, type = "width")
  state$last_line <- line
  state
}

.np_progress_show_now <- function(state, done = NULL, detail = NULL) {
  if (is.null(state) || !isTRUE(state$enabled) || !isTRUE(state$visible)) {
    return(state)
  }

  if (!is.null(done)) {
    state$last_done <- done
  }

  now <- state$started
  line <- .np_progress_format_line(
    state = state,
    done = state$last_done,
    detail = detail,
    now = now
  )
  state <- .np_progress_render(
    state = state,
    line = line,
    event = "start",
    now = now,
    done = state$last_done,
    detail = detail
  )
  state$start_note_pending <- FALSE
  state$last_emit <- now
  state$last_emitted_done <- state$last_done
  state$last_emitted_detail <- detail
  state
}

.np_progress_start_grace_sec <- function(known_total = FALSE, domain = "general") {
  default <- if (identical(domain, "plot")) {
    0.75
  } else if (isTRUE(known_total)) {
    0.75
  } else {
    1.0
  }

  option_name <- if (identical(domain, "plot")) {
    "np.plot.progress.start.grace.sec"
  } else if (isTRUE(known_total)) {
    "np.progress.start.grace.known.sec"
  } else {
    "np.progress.start.grace.unknown.sec"
  }

  val <- suppressWarnings(as.numeric(getOption(option_name, default))[1L])
  if (!is.finite(val) || is.na(val) || val < 0) {
    val <- default
  }

  val
}

.np_progress_interval_sec <- function(known_total = FALSE, domain = "general") {
  default <- if (identical(domain, "plot")) {
    # Plot bootstrap progress should feel like the accepted bandwidth heartbeat.
    2.0
  } else if (isTRUE(known_total)) {
    0.5
  } else {
    2.0
  }

  option_name <- if (identical(domain, "plot")) {
    "np.plot.progress.interval.sec"
  } else if (isTRUE(known_total)) {
    "np.progress.interval.known.sec"
  } else {
    "np.progress.interval.unknown.sec"
  }

  val <- suppressWarnings(as.numeric(getOption(option_name, default))[1L])
  if (!is.finite(val) || is.na(val) || val < 0) {
    val <- default
  }

  val
}

.np_progress_bandwidth_enhanced_enabled <- function() {
  isTRUE(getOption("np.progress.bandwidth.enhanced", FALSE))
}

.np_progress_iv_enhanced_state <- function(state) {
  isTRUE(state$iv_progress_common)
}

.np_progress_bandwidth_enhanced_state <- function(state) {
  isTRUE(state$bandwidth_progress_common)
}

.np_progress_bandwidth_coordinator_active <- function(state) {
  isTRUE(state$bandwidth_coordinator_active)
}

.np_progress_bandwidth_title <- function() {
  "Bandwidth selection"
}

.np_progress_iv_title <- function() {
  "IV regression"
}

.np_progress_iv_initialize_state <- function(state) {
  state$iv_progress_common <- TRUE
  state$iv_object_label <- NULL
  state$iv_iteration <- NULL
  state$start_note <- sprintf(
    "%s %s...",
    state$pkg_prefix,
    .np_progress_iv_title()
  )
  state
}

.np_progress_iv_format_line <- function(state, now = .np_progress_now()) {
  elapsed <- max(0, now - state$started)
  fields <- character()

  if (!is.null(state$iv_object_label) && nzchar(state$iv_object_label)) {
    fields <- c(fields, state$iv_object_label)
  }

  if (!is.null(state$iv_iteration)) {
    fields <- c(fields, sprintf("iteration %s", format(state$iv_iteration)))
  }

  fields <- c(fields, sprintf("elapsed %ss", .np_progress_fmt_num(elapsed)))

  sprintf(
    "%s %s (%s)",
    state$pkg_prefix,
    .np_progress_iv_title(),
    paste(fields, collapse = ", ")
  )
}

.np_progress_bandwidth_initialize_state <- function(state) {
  state$bandwidth_progress_common <- TRUE
  state$bandwidth_nmulti_total <- 1L
  state$bandwidth_multistart_current <- 1L
  state$bandwidth_multistart_completed <- 0L
  state$bandwidth_current_started <- state$started
  state$bandwidth_multistart_durations <- numeric()
  state$bandwidth_mode <- "iteration"
  state$bandwidth_last_iteration <- NULL
  state$bandwidth_estimate_last_emit <- NULL
  state$bandwidth_coordinator_active <- FALSE
  state$bandwidth_coordinator_groups <- NULL
  state$bandwidth_coordinator_local_total <- NULL
  state$bandwidth_coordinator_group_index <- NULL
  state$bandwidth_coordinator_group_label <- NULL
  state$bandwidth_coordinator_offset <- 0L
  state$bandwidth_coordinator_local_current <- 1L
  state$start_note <- sprintf(
    "%s %s...",
    state$pkg_prefix,
    .np_progress_bandwidth_title()
  )
  state
}

.np_progress_bandwidth_set_coordinator <- function(total_groups, local_total) {
  state <- .np_progress_runtime$bandwidth_state
  total_groups <- suppressWarnings(as.integer(total_groups)[1L])
  local_total <- suppressWarnings(as.integer(local_total)[1L])

  if (is.null(state) ||
      !.np_progress_bandwidth_enhanced_state(state) ||
      is.na(total_groups) || total_groups < 1L ||
      is.na(local_total) || local_total < 1L) {
    return(invisible(NULL))
  }

  state$bandwidth_coordinator_active <- TRUE
  state$bandwidth_coordinator_groups <- total_groups
  state$bandwidth_coordinator_local_total <- local_total
  state$bandwidth_coordinator_group_index <- 1L
  state$bandwidth_coordinator_offset <- 0L
  state$bandwidth_nmulti_total <- total_groups * local_total
  state$bandwidth_multistart_current <- 1L
  state$bandwidth_multistart_completed <- 0L
  .np_progress_runtime$bandwidth_state <- state
  invisible(NULL)
}

.np_progress_bandwidth_set_coordinator_group <- function(group_index, group_label = NULL) {
  state <- .np_progress_runtime$bandwidth_state
  group_index <- suppressWarnings(as.integer(group_index)[1L])

  if (is.null(state) ||
      !.np_progress_bandwidth_coordinator_active(state) ||
      is.na(group_index) || group_index < 1L) {
    return(invisible(NULL))
  }

  local_total <- suppressWarnings(as.integer(state$bandwidth_coordinator_local_total)[1L])
  total <- suppressWarnings(as.integer(state$bandwidth_nmulti_total)[1L])
  if (is.na(local_total) || local_total < 1L || is.na(total) || total < 1L) {
    return(invisible(NULL))
  }

  offset <- (group_index - 1L) * local_total
  state$bandwidth_coordinator_group_index <- group_index
  state$bandwidth_coordinator_group_label <- if (is.null(group_label)) NULL else as.character(group_label)[1L]
  state$bandwidth_coordinator_offset <- offset
  state$bandwidth_coordinator_local_current <- 1L
  state$bandwidth_multistart_completed <- min(offset, total)
  state$bandwidth_multistart_current <- min(total, offset + 1L)
  state$bandwidth_current_started <- .np_progress_now()
  state$last_done <- NULL
  if (offset > 0L) {
    state$bandwidth_mode <- "complete_estimate"
  }
  if (!is.na(local_total) && local_total > 1L) {
    state$start_note <- if (!is.null(state$bandwidth_coordinator_group_label)) {
      sprintf(
        "%s %s (%s, multistart 1/%s)",
        state$pkg_prefix,
        .np_progress_bandwidth_title(),
        state$bandwidth_coordinator_group_label,
        format(local_total)
      )
    } else {
      sprintf(
        "%s %s (multistart 1/%s)",
        state$pkg_prefix,
        .np_progress_bandwidth_title(),
        format(local_total)
      )
    }
  } else {
    state$start_note <- if (!is.null(state$bandwidth_coordinator_group_label)) {
      sprintf(
        "%s %s (%s)",
        state$pkg_prefix,
        .np_progress_bandwidth_title(),
        state$bandwidth_coordinator_group_label
      )
    } else {
      sprintf(
        "%s %s...",
        state$pkg_prefix,
        .np_progress_bandwidth_title()
      )
    }
  }
  .np_progress_runtime$bandwidth_state <- state
  invisible(NULL)
}

.np_progress_bandwidth_register_nmulti <- function(state, total) {
  total <- suppressWarnings(as.integer(total)[1L])
  if (is.null(state) || is.na(total) || total < 1L) {
    return(state)
  }

  if (.np_progress_bandwidth_coordinator_active(state)) {
    state$bandwidth_coordinator_local_total <- total
    groups <- suppressWarnings(as.integer(state$bandwidth_coordinator_groups)[1L])
    if (!is.na(groups) && groups >= 1L) {
      state$bandwidth_nmulti_total <- groups * total
    }
    return(state)
  }

  state$bandwidth_nmulti_total <- total
  state$bandwidth_multistart_current <- min(
    max(1L, suppressWarnings(as.integer(state$bandwidth_multistart_current)[1L])),
    total
  )
  if (total > 1L) {
    state$start_note <- sprintf(
      "%s %s (multistart 1/%s)",
      state$pkg_prefix,
      .np_progress_bandwidth_title(),
      format(total)
    )
  } else {
    state$start_note <- sprintf(
      "%s %s...",
      state$pkg_prefix,
      .np_progress_bandwidth_title()
    )
  }
  state
}

.np_progress_bandwidth_estimated_total_time <- function(state) {
  durations <- state$bandwidth_multistart_durations
  total <- suppressWarnings(as.integer(state$bandwidth_nmulti_total)[1L])
  if (!length(durations) || is.na(total) || total < 1L) {
    return(NA_real_)
  }

  mean(durations) * total
}

.np_progress_bandwidth_format_iteration <- function(state, done = NULL, now = .np_progress_now()) {
  elapsed <- max(0, now - state$started)
  total <- suppressWarnings(as.integer(state$bandwidth_nmulti_total)[1L])
  current <- suppressWarnings(as.integer(state$bandwidth_multistart_current)[1L])
  fields <- character()

  if (.np_progress_bandwidth_coordinator_active(state)) {
    label <- state$bandwidth_coordinator_group_label
    local_total <- suppressWarnings(as.integer(state$bandwidth_coordinator_local_total)[1L])
    local_current <- suppressWarnings(as.integer(state$bandwidth_coordinator_local_current)[1L])
    if (!is.null(label) && nzchar(label)) {
      fields <- c(fields, label)
    }
    if (!is.na(local_total) && local_total > 1L && !is.na(local_current) && local_current >= 1L) {
      fields <- c(fields, sprintf("multistart %s/%s", format(local_current), format(local_total)))
    }
  } else if (!is.na(total) && total > 1L && !is.na(current) && current >= 1L) {
    fields <- c(fields, sprintf("multistart %s/%s", format(current), format(total)))
  }

  if (!is.null(done)) {
    fields <- c(fields, sprintf("iteration %s", format(done)))
  }

  fields <- c(fields, sprintf("elapsed %ss", .np_progress_fmt_num(elapsed)))

  sprintf(
    "%s %s (%s)",
    state$pkg_prefix,
    .np_progress_bandwidth_title(),
    paste(fields, collapse = ", ")
  )
}

.np_progress_bandwidth_format_estimate <- function(state, now = .np_progress_now()) {
  elapsed <- max(0, now - state$started)
  total <- suppressWarnings(as.integer(state$bandwidth_nmulti_total)[1L])
  completed <- suppressWarnings(as.integer(state$bandwidth_multistart_completed)[1L])
  current <- suppressWarnings(as.integer(state$bandwidth_multistart_current)[1L])
  if (is.na(total) || total < 1L || completed < 1L) {
    return(.np_progress_bandwidth_format_iteration(state = state, done = state$last_done, now = now))
  }

  est.total <- .np_progress_bandwidth_estimated_total_time(state)
  base.share <- 100 * completed / total
  pct <- base.share
  eta <- 0

  if (is.finite(est.total) && !is.na(est.total) && est.total > 0) {
    pct.calc <- 100 * elapsed / est.total
    pct.upper <- if (completed < total) 99.9 else 100.0
    pct <- max(base.share, min(pct.upper, pct.calc))
    if (completed < total) {
      eta <- max(0, est.total - elapsed)
    }
  }

  fields <- character()

  if (.np_progress_bandwidth_coordinator_active(state)) {
    label <- state$bandwidth_coordinator_group_label
    local_total <- suppressWarnings(as.integer(state$bandwidth_coordinator_local_total)[1L])
    local_current <- suppressWarnings(as.integer(state$bandwidth_coordinator_local_current)[1L])
    if (!is.null(label) && nzchar(label)) {
      fields <- c(fields, label)
    }
    if (!is.na(local_current) && local_current >= 1L && !is.na(local_total) && local_total >= 1L) {
      fields <- c(fields, sprintf("multistart %s/%s", format(local_current), format(local_total)))
    }
  } else if (!is.na(current) && current >= 1L) {
    fields <- c(fields, sprintf("multistart %s/%s", format(current), format(total)))
  }

  if (completed < total && !is.null(state$last_done)) {
    fields <- c(fields, sprintf("iteration %s", format(state$last_done)))
  }

  fields <- c(
    fields,
    sprintf("elapsed %ss", .np_progress_fmt_num(elapsed)),
    sprintf("%s%%", .np_progress_fmt_num(pct)),
    sprintf("eta %ss", .np_progress_fmt_num(eta))
  )

  sprintf(
    "%s %s (%s)",
    state$pkg_prefix,
    .np_progress_bandwidth_title(),
    paste(fields, collapse = ", ")
  )
}

.np_progress_bandwidth_format_line <- function(state, done = NULL, now = .np_progress_now()) {
  mode <- state$bandwidth_mode
  if (identical(mode, "complete_estimate")) {
    .np_progress_bandwidth_format_estimate(state = state, now = now)
  } else {
    .np_progress_bandwidth_format_iteration(
      state = state,
      done = if (is.null(done)) state$last_done else done,
      now = now
    )
  }
}

.np_progress_format_known_total <- function(state, done, detail = NULL, now = .np_progress_now()) {
  total <- state$total
  pct <- if (isTRUE(total > 0)) 100 * done / total else 0
  elapsed <- max(0, now - state$started)
  eta <- if (isTRUE(done > 0) && isTRUE(total >= done)) elapsed * (total - done) / done else 0

  progress_token <- if (isTRUE(state$bandwidth_multistart)) {
    sprintf("multistart %s/%s", format(done), format(total))
  } else {
    sprintf("%s/%s", format(done), format(total))
  }

  line <- sprintf(
    "%s %s %s (%s%%, elapsed %ss, eta %ss)",
    state$pkg_prefix,
    state$label,
    progress_token,
    .np_progress_fmt_num(pct),
    .np_progress_fmt_num(elapsed),
    .np_progress_fmt_num(eta)
  )

  if (!is.null(detail)) {
    line <- paste0(line, ": ", detail)
  }

  line
}

.np_progress_format_unknown_total <- function(state, done = NULL, detail = NULL, now = .np_progress_now()) {
  if (is.function(state$unknown_total_fields)) {
    fields <- state$unknown_total_fields(
      state = state,
      done = done,
      detail = detail,
      now = now
    )
    if (length(fields)) {
      return(sprintf(
        "%s %s (%s)",
        state$pkg_prefix,
        state$label,
        paste(as.character(fields), collapse = ", ")
      ))
    }
  }

  elapsed <- max(0, now - state$started)
  line <- sprintf("%s %s... ", state$pkg_prefix, state$label)

  if (!is.null(done)) {
    line <- paste0(line, "iteration ", format(done), ", ")
  }

  line <- paste0(line, "elapsed ", .np_progress_fmt_num(elapsed), "s")

  if (!is.null(detail)) {
    line <- paste0(line, ": ", detail)
  }

  line
}

.np_progress_format_line <- function(state, done = NULL, detail = NULL, now = .np_progress_now()) {
  if (.np_progress_iv_enhanced_state(state)) {
    return(.np_progress_iv_format_line(
      state = state,
      now = now
    ))
  }

  if (.np_progress_bandwidth_enhanced_state(state)) {
    return(.np_progress_bandwidth_format_line(
      state = state,
      done = done,
      now = now
    ))
  }

  if (isTRUE(state$known_total)) {
    .np_progress_format_known_total(
      state = state,
      done = if (is.null(done)) state$last_done else done,
      detail = detail,
      now = now
    )
  } else {
    .np_progress_format_unknown_total(
      state = state,
      done = if (is.null(done)) state$last_done else done,
      detail = detail,
      now = now
    )
  }
}

.np_progress_begin <- function(label, total = NULL, domain = "general", surface = domain) {
  known_total <- !is.null(total)
  throttle_sec <- .np_progress_interval_sec(known_total = known_total, domain = domain)
  started <- .np_progress_now()
  capability <- .np_progress_capability(domain = domain)
  session_id <- .np_progress_next_id()
  enabled <- .np_progress_enabled(domain = domain)
  renderer <- .np_progress_renderer_for_surface(surface = surface, capability = capability)
  visible <- if (!isTRUE(enabled)) {
    FALSE
  } else {
    isTRUE(.np_progress_claim_owner(session_id))
  }

  list(
    id = session_id,
    enabled = enabled,
    visible = visible,
    pkg_prefix = .np_progress_pkg_prefix(),
    label = label,
    surface = surface,
    total = total,
    known_total = known_total,
    started = started,
    last_emit = started - throttle_sec,
    throttle_sec = throttle_sec,
    last_done = if (known_total) 0 else NULL,
    domain = domain,
    renderer = renderer,
    capability = capability,
    rendered = FALSE,
    start_note = sprintf("%s %s...", .np_progress_pkg_prefix(), label),
    start_note_pending = TRUE,
    start_note_grace_sec = .np_progress_start_grace_sec(known_total = known_total, domain = domain),
    last_line = NULL,
    last_render_width = 0L,
    last_emitted_done = NULL,
    last_emitted_detail = NULL,
    unknown_total_fields = NULL,
    message_muffled = FALSE,
    message_muffled_checked = FALSE
  )
}

.np_progress_maybe_emit_start_note <- function(state, now = .np_progress_now()) {
  if (!isTRUE(state$enabled) || !isTRUE(state$visible) || !isTRUE(state$start_note_pending)) {
    return(state)
  }

  if ((now - state$started) < state$start_note_grace_sec) {
    return(state)
  }

  state <- .np_progress_render(
    state = state,
    line = state$start_note,
    event = "start",
    now = now
  )
  state$start_note_pending <- FALSE

  if (isTRUE(state$start_note_consumes_throttle)) {
    state$last_emit <- now
  }

  state
}

.np_progress_step_at <- function(state, now, done = NULL, detail = NULL, force = FALSE) {
  if (!is.null(done)) {
    state$last_done <- done
  }

  if (!isTRUE(state$enabled)) {
    return(state)
  }

  state <- .np_progress_maybe_emit_start_note(state = state, now = now)
  if (isTRUE(state$start_note_pending)) {
    return(state)
  }

  if (!isTRUE(force) && (now - state$last_emit) < state$throttle_sec) {
    return(state)
  }

  line <- .np_progress_format_line(
    state = state,
    done = if (is.null(done)) state$last_done else done,
    detail = detail,
    now = now
  )

  if (!identical(line, state$last_line)) {
    state <- .np_progress_render(
      state = state,
      line = line,
      event = "update",
      now = now,
      done = if (is.null(done)) state$last_done else done,
      detail = detail
    )
    state$last_emit <- now
    state$last_emitted_done <- if (is.null(done)) state$last_done else done
    state$last_emitted_detail <- detail
  }

  state
}

.np_progress_step <- function(state, done = NULL, detail = NULL) {
  .np_progress_step_at(
    state = state,
    now = .np_progress_now(),
    done = done,
    detail = detail
  )
}

.np_progress_end <- function(state, detail = NULL) {
  if (!isTRUE(state$enabled)) {
    .np_progress_release_owner(state$id)
    return(invisible(state))
  }

  now <- .np_progress_now()
  if (is.null(state$last_line) &&
      (now - state$started) < state$start_note_grace_sec) {
    .np_progress_release_owner(state$id)
    return(invisible(state))
  }

  must_clear <- identical(state$renderer, "single_line") && isTRUE(state$rendered)

  if (.np_progress_bandwidth_enhanced_state(state)) {
    total <- suppressWarnings(as.integer(state$bandwidth_nmulti_total)[1L])
    if (!is.na(total) && total > 1L) {
      completed <- suppressWarnings(as.integer(state$bandwidth_multistart_completed)[1L])
      if (completed < total) {
        gap <- total - completed
        dur <- max(0, now - state$bandwidth_current_started)
        if (gap > 0L) {
          state$bandwidth_multistart_durations <- c(
            state$bandwidth_multistart_durations,
            rep.int(dur, gap)
          )
        }
        state$bandwidth_multistart_completed <- total
        state$bandwidth_multistart_current <- total
      }
      state$bandwidth_mode <- "complete_estimate"
      line <- .np_progress_bandwidth_format_estimate(state = state, now = now)
      if (isTRUE(must_clear) || !identical(line, state$last_line)) {
        state <- .np_progress_render(
          state = state,
          line = line,
          event = "finish",
          now = now,
          done = state$last_done,
          detail = detail
        )
      }
      .np_progress_release_owner(state$id)
      return(invisible(state))
    }
  }

  if (isTRUE(state$known_total)) {
    done <- if (is.null(state$total)) state$last_done else state$total
    line <- .np_progress_format_line(state = state, done = done, detail = detail, now = now)
    if (isTRUE(must_clear) ||
        !(identical(done, state$last_emitted_done) && identical(detail, state$last_emitted_detail))) {
      state <- .np_progress_render(
        state = state,
        line = line,
        event = "finish",
        now = now,
        done = done,
        detail = detail
      )
      state$last_emit <- now
      state$last_done <- done
      state$last_emitted_done <- done
      state$last_emitted_detail <- detail
    }
    .np_progress_release_owner(state$id)
    return(invisible(state))
  }

  if (!is.null(state$last_done)) {
    line <- .np_progress_format_line(state = state, done = state$last_done, detail = detail, now = now)
    if (isTRUE(must_clear) ||
        !(identical(state$last_done, state$last_emitted_done) && identical(detail, state$last_emitted_detail))) {
      state <- .np_progress_render(
        state = state,
        line = line,
        event = "finish",
        now = now,
        done = state$last_done,
        detail = detail
      )
      state$last_emit <- now
      state$last_emitted_done <- state$last_done
      state$last_emitted_detail <- detail
    }
  }

  if (isTRUE(must_clear) && is.null(state$last_done) && !is.null(state$last_line)) {
    state <- .np_progress_render(
      state = state,
      line = state$last_line,
      event = "finish",
      now = now,
      detail = detail
    )
  }

  .np_progress_release_owner(state$id)
  invisible(state)
}

.np_progress_abort <- function(state, detail = NULL) {
  if (!isTRUE(state$enabled)) {
    .np_progress_release_owner(state$id)
    return(invisible(state))
  }

  now <- .np_progress_now()
  if (!isTRUE(state$rendered)) {
    .np_progress_release_owner(state$id)
    return(invisible(state))
  }

  line <- if (!is.null(detail)) paste0(state$pkg_prefix, " ", detail) else state$last_line
  state <- .np_progress_render(
    state = state,
    line = line,
    event = "abort",
    now = now,
    done = state$last_done,
    detail = detail
  )
  .np_progress_release_owner(state$id)
  invisible(state)
}

.np_progress_note <- function(label) {
  if (.np_progress_enabled()) {
    .np_progress_emit(sprintf("%s %s", .np_progress_pkg_prefix(), label))
  }

  invisible(NULL)
}

.np_progress_bandwidth_active <- function() {
  isTRUE(.np_progress_runtime$bandwidth_depth > 0L)
}

.np_progress_iv_active <- function() {
  isTRUE(.np_progress_runtime$iv_depth > 0L)
}

.np_progress_iv_set_object <- function(label = NULL, iteration = NULL) {
  state <- .np_progress_runtime$iv_state

  if (is.null(state) || !.np_progress_iv_enhanced_state(state)) {
    return(invisible(NULL))
  }

  if (!is.null(label)) {
    label <- as.character(label)[1L]
    if (is.na(label) || !nzchar(label)) {
      label <- NULL
    }
  }

  if (!is.null(iteration)) {
    iteration <- suppressWarnings(as.integer(iteration)[1L])
    if (is.na(iteration) || iteration < 1L) {
      iteration <- NULL
    }
  }

  state$iv_object_label <- label
  state$iv_iteration <- iteration
  .np_progress_runtime$iv_state <- .np_progress_step_at(
    state = state,
    now = .np_progress_now(),
    done = iteration,
    force = TRUE
  )

  invisible(NULL)
}

.np_progress_iv_activity_step <- function(iteration = NULL) {
  state <- .np_progress_runtime$iv_state

  if (is.null(state) || !.np_progress_iv_enhanced_state(state)) {
    return(invisible(NULL))
  }

  if (!is.null(iteration)) {
    iteration <- suppressWarnings(as.integer(iteration)[1L])
    if (is.na(iteration) || iteration < 1L) {
      iteration <- NULL
    }
    state$iv_iteration <- iteration
  }

  .np_progress_runtime$iv_state <- .np_progress_step(
    state = state,
    done = state$iv_iteration
  )

  invisible(NULL)
}

.np_progress_iv_finish <- function() {
  state <- .np_progress_runtime$iv_state

  if (is.null(state)) {
    return(invisible(NULL))
  }

  .np_progress_runtime$iv_state <- .np_progress_end(state)
  invisible(NULL)
}

.np_progress_iv_clear <- function() {
  .np_progress_runtime$iv_depth <- 0L
  .np_progress_runtime$iv_label <- NULL
  .np_progress_runtime$iv_state <- NULL
  .np_progress_runtime$iv_old_messages <- NULL
  .np_progress_runtime$force_enabled <- FALSE
  invisible(NULL)
}

.np_fit_progress_begin <- function(label, total, handoff = FALSE, detail = NULL) {
  total <- suppressWarnings(as.integer(total)[1L])
  if (is.na(total) || total < 1L) {
    .np_progress_runtime$fit_state <- NULL
    return(invisible(NULL))
  }

  state <- .np_progress_begin(
    label = label,
    total = total,
    domain = "bandwidth",
    surface = "bandwidth"
  )

  if (isTRUE(handoff)) {
    state <- .np_progress_show_now(
      state = state,
      done = 0L,
      detail = detail
    )
  }

  .np_progress_runtime$fit_state <- state
  invisible(state)
}

.np_fit_progress_step <- function(done = NULL, detail = NULL) {
  state <- .np_progress_runtime$fit_state

  if (is.null(state)) {
    return(invisible(NULL))
  }

  if (!is.null(done)) {
    done <- suppressWarnings(as.integer(done)[1L])
    if (is.na(done) || done < 1L) {
      done <- NULL
    }
  }

  .np_progress_runtime$fit_state <- .np_progress_step(
    state = state,
    done = done,
    detail = detail
  )

  invisible(NULL)
}

.np_fit_progress_finish <- function(detail = NULL) {
  state <- .np_progress_runtime$fit_state

  if (is.null(state)) {
    return(invisible(NULL))
  }

  .np_progress_end(state, detail = detail)
  .np_progress_runtime$fit_state <- NULL
  invisible(NULL)
}

.np_fit_progress_abort <- function(detail = NULL) {
  state <- .np_progress_runtime$fit_state

  if (is.null(state)) {
    return(invisible(NULL))
  }

  .np_progress_abort(state, detail = detail)
  .np_progress_runtime$fit_state <- NULL
  invisible(NULL)
}

.np_with_compiled_fit_progress <- function(label, total, handoff = FALSE, handoff.detail = NULL, expr) {
  total <- suppressWarnings(as.integer(total)[1L])
  fit.progress.active <- FALSE
  c.progress.active <- FALSE

  on.exit({
    if (isTRUE(c.progress.active)) {
      .Call("C_np_progress_fit_end", PACKAGE = "np")
      c.progress.active <- FALSE
    }

    if (isTRUE(fit.progress.active)) {
      .np_fit_progress_abort()
      fit.progress.active <- FALSE
    }
  }, add = TRUE)

  if (isTRUE(.np_progress_enabled(domain = "bandwidth")) &&
      !is.na(total) && total >= 1L) {
    .np_fit_progress_begin(
      label = label,
      total = total,
      handoff = handoff,
      detail = handoff.detail
    )
    fit.progress.active <- TRUE
    .Call("C_np_progress_fit_begin", total, PACKAGE = "np")
    c.progress.active <- TRUE
  }

  value <- force(expr)

  if (isTRUE(c.progress.active)) {
    .Call("C_np_progress_fit_end", PACKAGE = "np")
    c.progress.active <- FALSE
  }

  if (isTRUE(fit.progress.active)) {
    .np_fit_progress_finish()
    fit.progress.active <- FALSE
  }

  value
}

.np_densdist_fit_total <- function(bws, tnrow, enrow) {
  if (identical(as.character(bws$type)[1L], "adaptive_nn")) {
    as.integer(tnrow)
  } else {
    as.integer(enrow)
  }
}

.np_condensdist_fit_total <- function(bws, tnrow, enrow) {
  reg.engine <- if (is.null(bws$regtype.engine)) {
    if (is.null(bws$regtype)) "lc" else as.character(bws$regtype)
  } else {
    as.character(bws$regtype.engine)
  }
  base.total <- .np_densdist_fit_total(bws = bws, tnrow = tnrow, enrow = enrow)
  lp.route <- identical(reg.engine, "lp") && isTRUE(bws$xncon > 0L)

  if (isTRUE(lp.route)) {
    as.integer(enrow)
  } else {
    as.integer(2L * base.total)
  }
}

.np_progress_bandwidth_detail <- function(done, total) {
  NULL
}

.np_progress_bandwidth_set_total <- function(total) {
  state <- .np_progress_runtime$bandwidth_state
  total <- suppressWarnings(as.integer(total)[1L])

  if (is.null(state) || is.na(total) || total <= 1L) {
    return(invisible(NULL))
  }

  # NOMAD restart search is unknown-bound even when inner helpers expose a
  # local multistart count; do not synthesize percent/ETA from that count.
  if (!is.null(state$nomad_nmulti) && is.function(state$unknown_total_fields)) {
    return(invisible(NULL))
  }

  if (.np_progress_bandwidth_enhanced_state(state)) {
    .np_progress_runtime$bandwidth_state <- .np_progress_bandwidth_register_nmulti(
      state = state,
      total = total
    )
    return(invisible(NULL))
  }

  if (isTRUE(state$known_total) && identical(as.integer(state$total), total)) {
    return(invisible(NULL))
  }

  label <- .np_progress_runtime$bandwidth_label
  if (is.null(label))
    label <- state$label

  upgraded <- .np_progress_begin(
    label = label,
    total = total,
    domain = state$domain,
    surface = state$surface
  )
  upgraded$id <- state$id
  upgraded$visible <- state$visible
  upgraded$started <- state$started
  upgraded$last_emit <- state$started - upgraded$throttle_sec
  upgraded$renderer <- state$renderer
  upgraded$capability <- state$capability
  upgraded$rendered <- state$rendered
  upgraded$start_note_pending <- state$start_note_pending
  upgraded$last_line <- state$last_line
  upgraded$last_render_width <- state$last_render_width
  upgraded$last_emitted_done <- state$last_emitted_done
  upgraded$last_emitted_detail <- state$last_emitted_detail
  upgraded$message_muffled <- state$message_muffled
  upgraded$message_muffled_checked <- state$message_muffled_checked
  upgraded$bandwidth_multistart <- TRUE
  if (!isTRUE(state$start_note_pending) && identical(state$last_line, state$start_note)) {
    upgraded$start_note_pending <- FALSE
  }

  .np_progress_runtime$bandwidth_state <- upgraded
  invisible(NULL)
}

.np_progress_bandwidth_multistart_step <- function(done, total = NULL) {
  state <- .np_progress_runtime$bandwidth_state

  if (is.null(state)) {
    return(invisible(NULL))
  }

  if (!is.null(total)) {
    .np_progress_bandwidth_set_total(total = total)
    state <- .np_progress_runtime$bandwidth_state
  }

  if (.np_progress_bandwidth_enhanced_state(state)) {
    done <- suppressWarnings(as.integer(done)[1L])
    if (is.null(state) || is.na(done) || done < 1L) {
      return(invisible(NULL))
    }

    total <- suppressWarnings(as.integer(state$bandwidth_nmulti_total)[1L])
    local_total <- if (.np_progress_bandwidth_coordinator_active(state)) {
      suppressWarnings(as.integer(state$bandwidth_coordinator_local_total)[1L])
    } else {
      total
    }
    if (is.na(local_total) || local_total < 2L || is.na(total) || total < 2L) {
      return(invisible(NULL))
    }

    done <- min(done, local_total)
    now <- .np_progress_now()
    completed <- suppressWarnings(as.integer(state$bandwidth_multistart_completed)[1L])
    completed <- max(0L, completed)
    global_done <- if (.np_progress_bandwidth_coordinator_active(state)) {
      offset <- suppressWarnings(as.integer(state$bandwidth_coordinator_offset)[1L])
      offset + done
    } else {
      done
    }
    gap <- global_done - completed
    if (gap > 0L) {
      dur <- max(0, now - state$bandwidth_current_started)
      state$bandwidth_multistart_durations <- c(
        state$bandwidth_multistart_durations,
        rep.int(dur, gap)
      )
    }
    state$bandwidth_multistart_completed <- global_done
    state$bandwidth_multistart_current <- if (done < local_total) global_done + 1L else min(total, global_done)
    if (.np_progress_bandwidth_coordinator_active(state)) {
      state$bandwidth_coordinator_local_current <- if (done < local_total) done + 1L else local_total
    }
    if (done < local_total) {
      state$bandwidth_current_started <- now
      state$last_done <- NULL
    }
    state$bandwidth_mode <- "complete_estimate"
    state$last_emit <- -Inf
    .np_progress_runtime$bandwidth_state <- .np_progress_step_at(
      state = state,
      now = now,
      done = state$last_done,
      force = TRUE
    )
    return(invisible(NULL))
  }

  if (is.null(state) || !isTRUE(state$known_total)) {
    return(invisible(NULL))
  }

  done <- suppressWarnings(as.integer(done)[1L])
  if (is.na(done) || done < 1L) {
    return(invisible(NULL))
  }

  total <- as.integer(state$total)
  done <- min(done, total)
  .np_progress_runtime$bandwidth_state <- .np_progress_step(
    state = state,
    done = done,
    detail = .np_progress_bandwidth_detail(done = done, total = total)
  )

  invisible(NULL)
}

.np_progress_bandwidth_activity_step <- function(done = NULL) {
  state <- .np_progress_runtime$bandwidth_state

  if (.np_progress_bandwidth_enhanced_state(state)) {
    if (!is.null(done)) {
      done <- suppressWarnings(as.integer(done)[1L])
      if (is.na(done) || done < 1L) {
        done <- NULL
      }
    }
    .np_progress_runtime$bandwidth_state <- .np_progress_step(
      state = state,
      done = done
    )
    return(invisible(NULL))
  }

  if (is.null(state) || isTRUE(state$known_total)) {
    return(invisible(NULL))
  }

  if (!is.null(done)) {
    done <- suppressWarnings(as.integer(done)[1L])
    if (is.na(done) || done < 1L) {
      done <- NULL
    }
  }

  .np_progress_runtime$bandwidth_state <- .np_progress_step(
    state = state,
    done = done
  )

  invisible(NULL)
}

.np_progress_signal_from_c <- function(event, surface, current = NULL, total = NULL) {
  event <- as.character(event)[1L]
  surface <- as.character(surface)[1L]

  if (is.na(event) || is.na(surface)) {
    return(invisible(FALSE))
  }

  if (identical(event, "bandwidth_multistart_step") && identical(surface, "bandwidth")) {
    .np_progress_bandwidth_multistart_step(done = current, total = total)
    return(invisible(TRUE))
  }

  if (identical(event, "bandwidth_activity_step") && identical(surface, "bandwidth")) {
    .np_progress_bandwidth_activity_step(done = current)
    return(invisible(TRUE))
  }

  if (identical(event, "fit_step") && identical(surface, "bandwidth")) {
    .np_fit_progress_step(done = current)
    return(invisible(TRUE))
  }

  invisible(FALSE)
}

.np_progress_bandwidth_finish <- function() {
  state <- .np_progress_runtime$bandwidth_state

  if (is.null(state)) {
    return(invisible(NULL))
  }

  if (isTRUE(state$known_total) && isTRUE(state$total > 1L)) {
    total <- as.integer(state$total)
    .np_progress_runtime$bandwidth_state <- .np_progress_end(
      state,
      detail = .np_progress_bandwidth_detail(done = total, total = total)
    )
  } else {
    .np_progress_runtime$bandwidth_state <- .np_progress_end(state)
  }

  invisible(NULL)
}

.np_progress_bandwidth_clear <- function() {
  .np_progress_runtime$bandwidth_depth <- 0L
  .np_progress_runtime$bandwidth_label <- NULL
  .np_progress_runtime$bandwidth_state <- NULL
  .np_progress_runtime$bandwidth_old_messages <- NULL
  .np_progress_runtime$bandwidth_context_label <- NULL
  .np_progress_runtime$force_enabled <- FALSE
  invisible(NULL)
}

.np_progress_with_bandwidth_enhanced <- function(expr) {
  old <- getOption("np.progress.bandwidth.enhanced", FALSE)
  on.exit(options(np.progress.bandwidth.enhanced = old), add = TRUE)
  options(np.progress.bandwidth.enhanced = TRUE)
  force(expr)
}

.np_progress_select_bandwidth <- function(label, expr) {
  starting <- !.np_progress_bandwidth_active()
  if (starting) {
    .np_progress_runtime$bandwidth_old_messages <- getOption("np.messages", TRUE)
    .np_progress_runtime$force_enabled <- isTRUE(.np_progress_runtime$bandwidth_old_messages)
    options(np.messages = FALSE)
    .np_progress_runtime$bandwidth_label <- as.character(label)[1L]
    .np_progress_runtime$bandwidth_state <- .np_progress_begin(
      label = .np_progress_runtime$bandwidth_label,
      domain = "general",
      surface = "bandwidth"
    )
    if (.np_progress_bandwidth_enhanced_enabled()) {
      .np_progress_runtime$bandwidth_state <- .np_progress_bandwidth_initialize_state(
        .np_progress_runtime$bandwidth_state
      )
      context_label <- .np_progress_runtime$bandwidth_context_label
      if (!is.null(context_label) && nzchar(context_label)) {
        .np_progress_bandwidth_set_coordinator(total_groups = 1L, local_total = 1L)
        .np_progress_bandwidth_set_coordinator_group(1L, context_label)
        .np_progress_runtime$bandwidth_state$start_note_grace_sec <- 0
        .np_progress_runtime$bandwidth_state <- .np_progress_maybe_emit_start_note(
          .np_progress_runtime$bandwidth_state,
          now = .np_progress_runtime$bandwidth_state$started
        )
      }
    }
  }

  .np_progress_runtime$bandwidth_depth <- as.integer(.np_progress_runtime$bandwidth_depth) + 1L
  on.exit({
    .np_progress_runtime$bandwidth_depth <- max(
      0L,
      as.integer(.np_progress_runtime$bandwidth_depth) - 1L
    )
    if (starting) {
      .np_progress_bandwidth_finish()
      options(np.messages = .np_progress_runtime$bandwidth_old_messages)
      .np_progress_bandwidth_clear()
    }
  }, add = TRUE)

  force(expr)
}

.np_progress_select_bandwidth_enhanced <- function(label, expr) {
  .np_progress_with_bandwidth_enhanced(
    .np_progress_select_bandwidth(label = label, expr = expr)
  )
}

.np_progress_select_iv <- function(label = .np_progress_iv_title(), expr) {
  starting <- !.np_progress_iv_active()
  if (starting) {
    .np_progress_runtime$iv_old_messages <- getOption("np.messages", TRUE)
    .np_progress_runtime$force_enabled <- isTRUE(.np_progress_runtime$iv_old_messages)
    options(np.messages = FALSE)
    .np_progress_runtime$iv_label <- as.character(label)[1L]
    .np_progress_runtime$iv_state <- .np_progress_iv_initialize_state(
      .np_progress_begin(
        label = .np_progress_runtime$iv_label,
        domain = "general",
        surface = "iv_solve"
      )
    )
  }

  .np_progress_runtime$iv_depth <- as.integer(.np_progress_runtime$iv_depth) + 1L
  on.exit({
    .np_progress_runtime$iv_depth <- max(
      0L,
      as.integer(.np_progress_runtime$iv_depth) - 1L
    )
    if (starting) {
      .np_progress_iv_finish()
      options(np.messages = .np_progress_runtime$iv_old_messages)
      .np_progress_iv_clear()
    }
  }, add = TRUE)

  force(expr)
}

.np_progress_with_legacy_suppressed <- function(expr) {
  old_np_messages <- getOption("np.messages")
  on.exit(options(np.messages = old_np_messages), add = TRUE)
  options(np.messages = FALSE)
  force(expr)
}

with_np_bindings <- function(bindings, code) {
  code <- substitute(code)
  ns <- asNamespace("np")
  old <- lapply(names(bindings), function(name) get(name, envir = ns, inherits = FALSE))
  names(old) <- names(bindings)

  for (name in names(bindings)) {
    was_locked <- bindingIsLocked(name, ns)
    if (was_locked) {
      unlockBinding(name, ns)
    }
    assign(name, bindings[[name]], envir = ns)
    if (was_locked) {
      lockBinding(name, ns)
    }
  }

  on.exit({
    for (name in names(old)) {
      was_locked <- bindingIsLocked(name, ns)
      if (was_locked) {
        unlockBinding(name, ns)
      }
      assign(name, old[[name]], envir = ns)
      if (was_locked) {
        lockBinding(name, ns)
      }
    }
  }, add = TRUE)

  eval(code, envir = parent.frame())
}

capture_messages_only <- function(expr) {
  messages <- character()
  withCallingHandlers(
    expr,
    message = function(m) {
      messages <<- c(messages, conditionMessage(m))
      invokeRestart("muffleMessage")
    }
  )
  messages
}

capture_warnings_only <- function(expr) {
  warnings <- character()
  withCallingHandlers(
    expr,
    warning = function(w) {
      warnings <<- c(warnings, conditionMessage(w))
      invokeRestart("muffleWarning")
    }
  )
  warnings
}

normalize_messages <- function(x) {
  sub("\n$", "", x)
}

progress_time_values <- function(values) {
  force(values)
  i <- 0L
  function() {
    i <<- min(i + 1L, length(values))
    values[[i]]
  }
}

capture_single_line_output <- function(pkg, bindings, code) {
  path <- tempfile(fileext = ".log")
  con <- file(path, open = "wb")
  closed <- FALSE

  on.exit({
    if (!closed) {
      close(con)
    }
    unlink(path)
  }, add = TRUE)

  bindings$.np_progress_single_line_connection <- function() con

  if (identical(pkg, "np")) {
    with_np_bindings(bindings, code)
  } else {
    stop("unknown package: ", pkg)
  }

  close(con)
  closed <- TRUE
  readChar(path, nchars = file.info(path)$size, useBytes = TRUE)
}

test_that("progress begin returns disabled state when messages are off", {
  begin <- getFromNamespace(".np_progress_begin", "np")

  old_opts <- options(np.messages = FALSE)
  on.exit(options(old_opts), add = TRUE)

  state <- with_np_bindings(
    list(.np_progress_is_interactive = function() TRUE,
         .np_progress_now = function() 10),
    begin("Iterating Landweber-Fridman solve")
  )

  expect_false(state$enabled)
  expect_identical(state$label, "Iterating Landweber-Fridman solve")
  expect_null(state$total)
  expect_false(state$known_total)
})

test_that("known-total formatting is deterministic", {
  fmt <- getFromNamespace(".np_progress_format_known_total", "np")
  state <- list(
    enabled = TRUE,
    pkg_prefix = "[np]",
    label = "Iterating Tikhonov solve",
    total = 5,
    known_total = TRUE,
    started = 10,
    last_emit = 9.5,
    throttle_sec = 0.5,
    last_done = 2,
    domain = "general"
  )

  line <- fmt(state, done = 2, detail = "recomputing alpha", now = 14)

  expect_identical(
    line,
    "[np] Iterating Tikhonov solve 2/5 (40.0%, elapsed 4.0s, eta 6.0s): recomputing alpha"
  )
})

test_that("unknown-total formatting is deterministic", {
  fmt <- getFromNamespace(".np_progress_format_unknown_total", "np")
  state <- list(
    enabled = TRUE,
    pkg_prefix = "[np]",
    label = "Iterating Landweber-Fridman solve",
    total = NULL,
    known_total = FALSE,
    started = 10,
    last_emit = 8,
    throttle_sec = 2,
    last_done = 3,
    domain = "general"
  )

  line <- fmt(state, done = 3, detail = "evaluating stopping rule", now = 14)

  expect_identical(
    line,
    "[np] Iterating Landweber-Fridman solve... iteration 3, elapsed 4.0s: evaluating stopping rule"
  )
})

test_that("known-total progress delays start note and avoids duplicate completion", {
  begin <- getFromNamespace(".np_progress_begin", "np")
  step <- getFromNamespace(".np_progress_step", "np")
  finish <- getFromNamespace(".np_progress_end", "np")

  old_opts <- options(np.messages = TRUE, np.progress.start.grace.known.sec = 0.75)
  on.exit(options(old_opts), add = TRUE)

  messages <- with_np_bindings(
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_now = progress_time_values(c(0, 0.2, 0.8, 0.9))
    ),
    capture_messages_only({
      state <- begin("Bootstrap replications", total = 2)
      state <- step(state, done = 1)
      state <- step(state, done = 2)
      finish(state)
    })
  )

  messages <- normalize_messages(messages)

  expect_identical(messages[[1L]], "[np] Bootstrap replications...")
  expect_false(any(grepl("^\\[np\\] Bootstrap replications 1/2 ", messages)))
  expect_equal(sum(grepl("^\\[np\\] Bootstrap replications 2/2 ", messages)), 1L)
})

test_that("unknown-total progress delays start note until grace elapses", {
  begin <- getFromNamespace(".np_progress_begin", "np")
  step <- getFromNamespace(".np_progress_step", "np")

  old_opts <- options(np.messages = TRUE, np.progress.start.grace.unknown.sec = 1.0)
  on.exit(options(old_opts), add = TRUE)

  messages <- with_np_bindings(
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_now = progress_time_values(c(0, 0.4, 1.2))
    ),
    capture_messages_only({
      state <- begin("Iterating Landweber-Fridman solve")
      state <- step(state, done = 3, detail = "updating residual smoothing")
      step(state, done = 4, detail = "evaluating stopping rule")
    })
  )

  messages <- normalize_messages(messages)

  expect_identical(messages[[1L]], "[np] Iterating Landweber-Fridman solve...")
  expect_identical(
    messages[[2L]],
    "[np] Iterating Landweber-Fridman solve... iteration 4, elapsed 1.2s: evaluating stopping rule"
  )
  expect_length(messages, 2L)
})

test_that("fast progress stays silent below start grace", {
  begin <- getFromNamespace(".np_progress_begin", "np")
  finish <- getFromNamespace(".np_progress_end", "np")

  old_opts <- options(np.messages = TRUE, np.progress.start.grace.known.sec = 0.75)
  on.exit(options(old_opts), add = TRUE)

  messages <- with_np_bindings(
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_now = progress_time_values(c(0, 0.2))
    ),
    capture_messages_only({
      state <- begin("Bootstrap replications", total = 5)
      finish(state)
    })
  )

  expect_length(messages, 0)
})

test_that("suppressMessages suppresses progress notes", {
  note <- getFromNamespace(".np_progress_note", "np")

  old_opts <- options(np.messages = TRUE)
  on.exit(options(old_opts), add = TRUE)

  messages <- with_np_bindings(
    list(.np_progress_is_interactive = function() TRUE),
    capture_messages_only(suppressMessages(note("Preparing IV derivative regression")))
  )

  expect_length(messages, 0)
})

test_that("legacy suppression restores np.messages on success and error", {
  suppress_legacy <- getFromNamespace(".np_progress_with_legacy_suppressed", "np")

  old_opts <- options(np.messages = TRUE)
  on.exit(options(old_opts), add = TRUE)

  value <- suppress_legacy({
    expect_false(isTRUE(getOption("np.messages")))
    42
  })

  expect_identical(value, 42)
  expect_true(isTRUE(getOption("np.messages")))

  expect_error(
    suppress_legacy({
      stop("boom")
    }),
    "boom"
  )
  expect_true(isTRUE(getOption("np.messages")))
})

test_that("warning helper prefixes package warnings", {
  warn <- getFromNamespace(".np_warning", "np")

  warnings <- capture_warnings_only(warn("kernel order ignored"))

  expect_identical(warnings, "[np] kernel order ignored")
})

test_that("bandwidth selection helper suppresses legacy output and drives bounded updates", {
  select_bw <- getFromNamespace(".np_progress_select_bandwidth", "np")
  set_total <- getFromNamespace(".np_progress_bandwidth_set_total", "np")
  step_bw <- getFromNamespace(".np_progress_bandwidth_multistart_step", "np")

  old_opts <- options(np.messages = TRUE, np.progress.start.grace.known.sec = 0)
  on.exit(options(old_opts), add = TRUE)

  seen <- NULL
  legacy <- capture_progress_shadow_trace(
    {
      value <- select_bw("Selecting regression bandwidth", {
        seen <<- getOption("np.messages")
        set_total(3)
        step_bw(1, 3)
        step_bw(3, 3)
        7
      })
      expect_identical(value, 7)
    },
    force_renderer = "legacy",
    now = progress_time_values(c(0, 1.1, 2.2, 3.3, 4.4))
  )

  seen <- NULL
  actual <- capture_progress_shadow_trace(
    {
      value <- select_bw("Selecting regression bandwidth", {
        seen <<- getOption("np.messages")
        set_total(3)
        step_bw(1, 3)
        step_bw(3, 3)
        7
      })
      expect_identical(value, 7)
    },
    now = progress_time_values(c(0, 1.1, 2.2, 3.3, 4.4))
  )
  lines <- vapply(actual$trace, `[[`, character(1L), "line")
  legacy_lines <- vapply(
    legacy$trace[vapply(legacy$trace, `[[`, character(1L), "event") == "render"],
    `[[`,
    character(1L),
    "line"
  )
  render_lines <- vapply(
    actual$trace[vapply(actual$trace, `[[`, character(1L), "event") == "render"],
    `[[`,
    character(1L),
    "line"
  )

  expect_false(isTRUE(seen))
  expect_equal(render_lines, legacy_lines)
  expect_true(any(grepl("^\\[np\\] Selecting regression bandwidth\\.\\.\\.$", lines)))
  expect_true(any(grepl("^\\[np\\] Selecting regression bandwidth multistart 1/3 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Selecting regression bandwidth multistart 3/3 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(isTRUE(getOption("np.messages")))
})

test_that("compiled progress bridge feeds the bounded bandwidth sink", {
  select_bw <- getFromNamespace(".np_progress_select_bandwidth", "np")

  old_opts <- options(np.messages = TRUE, np.progress.start.grace.known.sec = 0)
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      value <- select_bw("Selecting regression bandwidth", {
        .Call(
          "C_np_progress_signal",
          "bandwidth_multistart_step",
          "bandwidth",
          1L,
          3L,
          PACKAGE = "np"
        )
        .Call(
          "C_np_progress_signal",
          "bandwidth_multistart_step",
          "bandwidth",
          3L,
          3L,
          PACKAGE = "np"
        )
        11
      })
      expect_identical(value, 11)
    },
    now = progress_time_values(c(0, 1.1, 2.2, 3.3))
  )

  lines <- vapply(actual$trace, `[[`, character(1L), "line")

  expect_true(any(grepl("^\\[np\\] Selecting regression bandwidth\\.\\.\\.$", lines)))
  expect_true(any(grepl("^\\[np\\] Selecting regression bandwidth multistart 1/3 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Selecting regression bandwidth multistart 3/3 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
})

test_that("compiled progress bridge feeds unknown-total bandwidth activity updates", {
  select_bw <- getFromNamespace(".np_progress_select_bandwidth", "np")

  old_opts <- options(np.messages = TRUE, np.progress.start.grace.unknown.sec = 0)
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      value <- select_bw("Selecting conditional density bandwidth", {
        .Call(
          "C_np_progress_signal",
          "bandwidth_activity_step",
          "bandwidth",
          64L,
          0L,
          PACKAGE = "np"
        )
        .Call(
          "C_np_progress_signal",
          "bandwidth_activity_step",
          "bandwidth",
          128L,
          0L,
          PACKAGE = "np"
        )
        13
      })
      expect_identical(value, 13)
    },
    now = progress_time_values(c(0, 1.2, 3.4, 5.6))
  )

  lines <- vapply(actual$trace, `[[`, character(1L), "line")

  expect_true(any(grepl("^\\[np\\] Selecting conditional density bandwidth\\.\\.\\.$", lines)))
  expect_true(any(grepl("^\\[np\\] Selecting conditional density bandwidth\\.\\.\\. iteration 64, elapsed [0-9]+\\.[0-9]s$", lines)))
  expect_true(any(grepl("^\\[np\\] Selecting conditional density bandwidth\\.\\.\\. iteration 128, elapsed [0-9]+\\.[0-9]s$", lines)))
})

test_that("compiled progress bridge feeds fit updates on the bandwidth surface", {
  fit_begin <- getFromNamespace(".np_fit_progress_begin", "np")
  fit_finish <- getFromNamespace(".np_fit_progress_finish", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      fit_begin("Fitting density", total = 3L, handoff = TRUE, detail = "starting")
      .Call(
        "C_np_progress_signal",
        "fit_step",
        "bandwidth",
        1L,
        3L,
        PACKAGE = "np"
      )
      .Call(
        "C_np_progress_signal",
        "fit_step",
        "bandwidth",
        3L,
        3L,
        PACKAGE = "np"
      )
      fit_finish()
    },
    now = progress_time_values(c(0, 1.1, 2.2, 3.3, 4.4))
  )

  lines <- vapply(actual$trace, `[[`, character(1L), "line")

  expect_true(any(grepl(
    "^\\[np\\] Fitting density 0/3 \\(0\\.0%, elapsed 0\\.0s, eta 0\\.0s\\): starting$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting density 1/3 \\([0-9]+\\.[0-9]%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting density 3/3 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )))
})

test_that("single-line fit drops detail before truncating", {
  fit <- getFromNamespace(".np_progress_fit_single_line", "np")

  line <- "[np] Constructing metric entropy by lag 1/2 (50.0%, elapsed 7.2s, eta 7.2s): lag 1"
  fitted <- fit(line, max_width = 78)

  expect_identical(
    fitted,
    "[np] Constructing metric entropy by lag 1/2 (50.0%, elapsed 7.2s, eta 7.2s)"
  )
})

test_that("single-line fit compacts bandwidth labels before ellipsizing", {
  fit <- getFromNamespace(".np_progress_fit_single_line", "np")

  line <- "[np] Selecting regression bandwidth multistart 2/3 (66.7%, elapsed 6.3s, eta 3.2s)"
  fitted <- fit(line, max_width = 80)

  expect_lte(nchar(fitted, type = "width"), 80L)
  expect_identical(
    fitted,
    "[np] Selecting reg bandwidth multistart 2/3 (66.7%, elapsed 6.3s, eta 3.2s)"
  )
})

test_that("single-line fit keeps smooth coefficient bandwidth lines readable at 80 columns", {
  fit <- getFromNamespace(".np_progress_fit_single_line", "np")

  line <- "[np] Selecting smooth coefficient bandwidth multistart 2/2 (100.0%, elapsed 6.3s, eta 0.0s)"
  fitted <- fit(line, max_width = 80)

  expect_lte(nchar(fitted, type = "width"), 80L)
  expect_identical(
    fitted,
    "[np] Selecting smooth coef bw multistart 2/2 (100.0%, elapsed 6.3s, eta 0.0s)"
  )
})

test_that("single-line fit preserves readable bandwidth fields at 80 columns", {
  fit <- getFromNamespace(".np_progress_fit_single_line", "np")

  line <- "[np] Bandwidth selection (multistart 2/2, iteration 31, elapsed 10.6s, 62.4%, eta 6.4s)"
  fitted <- fit(line, max_width = 80)

  expect_lte(nchar(fitted, type = "width"), 80L)
  expect_identical(
    fitted,
    "[np] Bandwidth selection (2/2, iter 31, elapsed 10.6s, 62.4%, eta 6.4s)"
  )
})

test_that("single-line fit preserves readable coordinator bandwidth fields at 80 columns", {
  fit <- getFromNamespace(".np_progress_fit_single_line", "np")

  line <- "[np] Bandwidth selection (y~z, multistart 2/2, iteration 31, elapsed 10.6s, 62.4%, eta 6.4s)"
  fitted <- fit(line, max_width = 80)

  expect_lte(nchar(fitted, type = "width"), 80L)
  expect_identical(
    fitted,
    "[np] Bandwidth selection (y~z, 2/2, iter 31, elapsed 10.6s, 62.4%, eta 6.4s)"
  )
})

test_that("single-line fit preserves restart and cumulative iteration in NOMAD lines", {
  fit <- getFromNamespace(".np_progress_fit_single_line", "np")

  line <- "[np] Bandwidth selection (multistart 2/2, iteration 31 (128), elapsed 10.6s, deg (1), best (0))"
  fitted <- fit(line, max_width = 80)

  expect_lte(nchar(fitted, type = "width"), 80L)
  expect_identical(
    fitted,
    "[np] Bandwidth selection (2/2, iter 31 (128), elapsed 10.6s, deg (1), best (0))"
  )
})

test_that("single-line fit compacts plot bootstrap labels so prep detail survives", {
  fit <- getFromNamespace(".np_progress_fit_single_line", "np")

  line <- "[np] Preparing plot bootstrap inid 0/50 (0.0%, elapsed 0.0s, eta 0.0s): eval 1/50"
  fitted <- fit(line, max_width = 80)

  expect_lte(nchar(fitted, type = "width"), 80L)
  expect_identical(
    fitted,
    "[np] Prep plot inid 0/50 (0.0%, elapsed 0.0s, eta 0.0s): eval 1/50"
  )
})

test_that("single-line fit preserves readable plot bootstrap counters at 80 columns", {
  fit <- getFromNamespace(".np_progress_fit_single_line", "np")

  line <- "[np] Plot bootstrap (grad index 1/1) 12345/80000 (15.4%, elapsed 8.7s, eta 44.9s)"
  fitted <- fit(line, max_width = 80)

  expect_lte(nchar(fitted, type = "width"), 80L)
  expect_identical(
    fitted,
    "[np] Plot bootstrap (grad idx 1/1) 12345/80000 (15.4%, elapsed 8.7s, eta 44.9s)"
  )
})

test_that("single-line fit preserves both ends when truncation is still required", {
  fit <- getFromNamespace(".np_progress_fit_single_line", "np")

  line <- "[np] Bootstrap replications 123/999 (12.3%, elapsed 12.3s, eta 87.7s)"
  fitted <- fit(line, max_width = 40)

  expect_lte(nchar(fitted, type = "width"), 40L)
  expect_true(startsWith(fitted, "[np]"))
  expect_true(endsWith(fitted, "eta 87.7s)"))
  expect_true(grepl("\\.\\.\\.", fitted))
})

test_that("RStudio capability keeps single-line viable at the boundary", {
  capability <- getFromNamespace(".np_progress_capability", "np")

  actual <- with_np_bindings(
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_is_rstudio_console = function() TRUE
    ),
    capability()
  )

  expect_true(isTRUE(actual$interactive))
  expect_true(isTRUE(actual$rstudio))
  expect_true(isTRUE(actual$single_line_viable))
})

test_that("RStudio width budget reserves redraw margin centrally", {
  output_width <- getFromNamespace(".np_progress_output_width", "np")

  actual <- with_np_bindings(list(.np_progress_is_rstudio_console = function() TRUE), {
    old_options <- options(width = 82)
    on.exit(options(old_options), add = TRUE)
    output_width()
  })

  expect_identical(actual, 78L)
})

test_that("terminal width probe takes precedence over stale option width", {
  output_width <- getFromNamespace(".np_progress_output_width", "np")

  actual <- with_np_bindings(
    list(
      .np_progress_is_rstudio_console = function() FALSE,
      .np_progress_terminal_width_probe = function() 132L
    ),
    {
      old_options <- options(width = 80)
      on.exit(options(old_options), add = TRUE)
      output_width()
    }
  )

  expect_identical(actual, 132L)
})

test_that("single-line finish clears the rendered line without leaving a newline", {
  render <- getFromNamespace(".np_progress_render_single_line", "np")
  line <- "[np] Bootstrap replications 2/2 (100.0%, elapsed 1.0s, eta 0.0s)"

  output <- capture_single_line_output(
    "np",
    list(.np_progress_output_width = function() 80L),
    {
      render(list(render_line = line, last_width = 0L), event = "render")
      render(list(render_line = line, last_width = nchar(line, type = "width")), event = "finish")
    }
  )

  expect_identical(
    output,
    paste0("\r", line, "\r", strrep(" ", nchar(line, type = "width")), "\r")
  )
})

test_that("single-line render uses ANSI erase when the console supports it", {
  render <- getFromNamespace(".np_progress_render_single_line", "np")
  line <- "[np] Rotating plot 18/72 (25.0%, elapsed 10.9s, eta 32.6s)"

  output <- capture_single_line_output(
    "np",
    list(.np_progress_single_line_supports_ansi = function(con) TRUE),
    {
      render(list(render_line = line, last_width = 0L), event = "render")
      render(list(render_line = line, last_width = nchar(line, type = "width")), event = "finish")
    }
  )

  expect_identical(output, paste0("\r\033[2K", line, "\r\033[2K\r"))
})

test_that("single-line render clears stale suffix across state handoff in non-ANSI mode", {
  render <- getFromNamespace(".np_progress_render_single_line", "np")
  output_width <- 120L
  long <- "[np] Calling NOMAD (Nonsmooth Optimization by Mesh Adaptive Direct Search)... elapsed 0.0s"
  short <- "[np] NOMAD search..."

  output <- capture_single_line_output(
    "np",
    list(
      .np_progress_output_width = function() output_width,
      .np_progress_single_line_supports_ansi = function(con) FALSE
    ),
    {
      render(list(render_line = long, last_width = 0L), event = "render")
      render(list(render_line = short, last_width = 0L), event = "render")
    }
  )

  expect_identical(
    output,
    paste0(
      "\r", long,
      "\r", short, strrep(" ", nchar(long, type = "width") - nchar(short, type = "width"))
    )
  )
})

test_that("progress end clears single-line output even when the final state already rendered", {
  begin <- getFromNamespace(".np_progress_begin", "np")
  step <- getFromNamespace(".np_progress_step", "np")
  finish <- getFromNamespace(".np_progress_end", "np")
  reset <- getFromNamespace(".np_progress_reset_registry", "np")

  output <- capture_single_line_output(
    "np",
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_is_rstudio_console = function() FALSE,
      .np_progress_now = progress_time_values(c(0, 0, 1, 1)),
      .np_progress_output_width = function() 120L
    ),
    {
      reset()
      on.exit(reset(), add = TRUE)
      state <- begin("Bootstrap replications", total = 2, surface = "bootstrap")
      state <- step(state, done = 2)
      finish(state)
    }
  )

  line <- "[np] Bootstrap replications 2/2 (100.0%, elapsed 1.0s, eta 0.0s)"
  expect_identical(
    output,
    paste0(
      "\r",
      strrep(" ", nchar(line, type = "width")),
      "\r"
    )
  )
})

test_that("plot progress can repaint on elapsed time before the first checkpoint", {
  begin <- getFromNamespace(".np_plot_progress_begin", "np")
  tick <- getFromNamespace(".np_plot_progress_tick", "np")
  finish <- getFromNamespace(".np_plot_progress_end", "np")
  reset <- getFromNamespace(".np_progress_reset_registry", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress.start.grace.sec = 0,
    np.plot.progress.interval.sec = 0.5
  )
  on.exit(options(old_opts), add = TRUE)

  output <- capture_single_line_output(
    "np",
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_is_rstudio_console = function() FALSE,
      .np_progress_now = progress_time_values(c(0, 0, 0.6, 1.2)),
      .np_progress_output_width = function() 120L
    ),
    {
      reset()
      on.exit(reset(), add = TRUE)
      state <- begin(total = 50L, label = "Preparing plot bootstrap inid")
      state <- tick(state, done = 1L)
      state <- tick(state, done = 2L)
      finish(state)
    }
  )

  expect_true(grepl("\\[np\\] Preparing plot bootstrap inid\\.\\.\\.", output))
  expect_true(grepl("\\[np\\] Preparing plot bootstrap inid 2/50", output))
})

test_that("plot progress default cadence matches bandwidth heartbeat", {
  plot_interval <- getFromNamespace(".np_plot_progress_interval_sec", "np")
  begin <- getFromNamespace(".np_progress_begin", "np")

  old_opts <- options(
    np.plot.progress.interval.sec = NULL,
    np.progress.interval.unknown.sec = NULL,
    np.progress.interval.known.sec = NULL
  )
  on.exit(options(old_opts), add = TRUE)

  expect_identical(plot_interval(), 2.0)
  expect_identical(begin("Bandwidth selection")$throttle_sec, 2.0)
  expect_identical(begin("Bootstrap replications", total = 10L)$throttle_sec, 0.5)
})

test_that("interactive bootstrap chunk sizes leave room for intermediate progress", {
  progress_chunk_cap <- getFromNamespace(".np_plot_progress_chunk_cap", "np")
  wild_chunk <- getFromNamespace(".np_wild_chunk_size", "np")
  enabled_chunk <- with_np_bindings(
    list(.np_progress_is_interactive = function() TRUE),
    wild_chunk(n = 500L, B = 9999L)
  )
  disabled_chunk <- with_np_bindings(
    list(.np_progress_is_interactive = function() FALSE),
    wild_chunk(n = 500L, B = 9999L)
  )

  expect_identical(progress_chunk_cap(9999L), 2500L)
  expect_identical(enabled_chunk, 16L)
  expect_identical(disabled_chunk, 9999L)
})

test_that("single-line abort preserves the final line and terminates it", {
  render <- getFromNamespace(".np_progress_render_single_line", "np")
  line <- "[np] Bootstrap replications aborted"

  output <- capture_single_line_output(
    "np",
    list(),
    render(list(render_line = line, last_width = 0L), event = "abort")
  )

  expect_identical(output, paste0("\r", line, "\n"))
})

test_that("progress ownership suppresses nested visibility for legacy renderer too", {
  begin <- getFromNamespace(".np_progress_begin", "np")
  finish <- getFromNamespace(".np_progress_end", "np")
  reset <- getFromNamespace(".np_progress_reset_registry", "np")

  reset()
  on.exit(reset(), add = TRUE)

  states <- with_np_bindings(
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_is_rstudio_console = function() TRUE,
      .np_progress_now = function() 1
    ),
    {
      outer <- begin("Outer task", total = 2)
      inner <- begin("Inner task", total = 2)
      finish(outer)
      list(outer = outer, inner = inner)
    }
  )

  expect_true(isTRUE(states$outer$visible))
  expect_false(isTRUE(states$inner$visible))
})

rbandwidth <-
  function(bw = stop("rbandwidth:argument 'bw' missing"),
           regtype = c("lc","ll","lp"),
           basis = c("glp","additive","tensor"),
           degree = NULL,
           bernstein.basis = FALSE,
           bwmethod = c("cv.ls","cv.aic"),
           bwscaling = FALSE,
           bwtype = c("fixed","generalized_nn","adaptive_nn"),
           ckertype = c("gaussian","truncated gaussian","epanechnikov","uniform"),
           ckerorder = c(2,4,6,8),
           ckerbound = c("none","range","fixed"),
           ckerlb = NULL,
           ckerub = NULL,
           ukertype = c("aitchisonaitken", "liracine"),
           okertype = c("liracine","wangvanryzin","racineliyan"),
           fval = NA,
           ifval = NA,
           num.feval = NA,
           num.feval.fast = NA,
           fval.history = NA,
           eval.history = NA,
           invalid.history = NA,
           nobs = NA,
           xdati = stop("rbandwidth:argument 'xdati' missing"),
           ydati = stop("rbandwidth:argument 'ydati' missing"),
           xnames = character(length(bw)),
           ynames = character(1),
           sfactor = NA, bandwidth = NA,
           rows.omit = NA,
           nconfac = NA,
           ncatfac = NA,
           sdev = NA,
           bandwidth.compute = TRUE,
           timing = NA,
           total.time = NA,
           ...){

  ndim = length(bw)
  npRejectLegacyLpArgs(names(list(...)), where = "rbandwidth")
  regtype = match.arg(regtype)
  basis <- npValidateLpBasis(regtype = regtype, basis = basis)
  bwmethod = match.arg(bwmethod)
  bwtype = match.arg(bwtype)
  ckertype = match.arg(ckertype)
  ckerbound = match.arg(ckerbound)

  if(missing(ckerorder))
    ckerorder = 2
  else if (ckertype == "uniform")
    .np_warning("ignoring kernel order specified with uniform kernel type")
  else {
    kord = c(2,4,6,8) 
    if (!any(kord == ckerorder))
      stop("ckerorder must be one of ", paste(kord,collapse=" "))
  }

  if (ckertype == "truncated gaussian" && ckerorder != 2)
    .np_warning("using truncated gaussian of order 2, higher orders not yet implemented")

  ukertype = match.arg(ukertype)
  okertype = match.arg(okertype)
  cbounds <- npKernelBoundsResolve(
    dati = xdati,
    varnames = xnames,
    kerbound = ckerbound,
    kerlb = ckerlb,
    kerub = ckerub,
    argprefix = "cker")
  bounded_nonfixed_supported <- bwtype %in% c("generalized_nn", "adaptive_nn")
  if (bwtype != "fixed" &&
      cbounds$bound != "none" &&
      !bounded_nonfixed_supported)
    stop("finite continuous kernel bounds require bwtype = \"fixed\"")

  ncon <- sum(xdati$icon)
  degree <- npValidateGlpDegree(regtype = regtype,
                                    degree = degree,
                                    ncon = ncon)
  bernstein.basis <- npValidateGlpBernstein(regtype = regtype,
                                          bernstein.basis = bernstein.basis)
  reg.spec <- npCanonicalConditionalRegSpec(
    regtype = regtype,
    basis = basis,
    degree = degree,
    bernstein.basis = bernstein.basis,
    ncon = ncon,
    where = "rbandwidth"
  )
  if (identical(regtype, "lp") && ncon > 0L && is.finite(nobs)) {
    lp.dim <- dim_basis(basis = basis,
                        kernel = TRUE,
                        degree = degree,
                        segments = rep.int(1L, ncon))
    if (is.finite(lp.dim) && lp.dim > (nobs - 1.0))
      stop(sprintf("LP basis dimension (%s) exceeds nobs - 1 (%s); reduce degree",
                   format(lp.dim, trim = TRUE, scientific = FALSE),
                   format(nobs - 1.0, trim = TRUE, scientific = FALSE)))
  }

  porder = switch( ckerorder/2, "Second-Order", "Fourth-Order", "Sixth-Order", "Eighth-Order" )
  ## calculate some info to be pretty-printed

  if (!identical(sfactor,NA)){
    sumNum <- sapply(seq_len(ndim), function(i) {
      if (xdati$icon[i])
        return(sfactor[i])

      if (xdati$iord[i])
        return(oMaxL(xdati$all.nlev[[i]], kertype = okertype))
      
      if (xdati$iuno[i])
        return(uMaxL(xdati$all.nlev[[i]], kertype = ukertype))
    })
  } else {
    sumNum <- NA
  }

  if (length(rows.omit) == 0)
    rows.omit <- NA
  
  mybw = list(
    bw=bw,
    regtype = regtype,
    pregtype = switch(regtype,
      lc = "Local-Constant",
      ll = "Local-Linear",
      lp = "Local-Polynomial"),
    basis = basis,
    degree = degree,
    bernstein.basis = bernstein.basis,
    regtype.engine = reg.spec$regtype.engine,
    basis.engine = reg.spec$basis.engine,
    degree.engine = reg.spec$degree.engine,
    bernstein.basis.engine = reg.spec$bernstein.basis.engine,
    method = bwmethod,
    pmethod = bwmToPrint(bwmethod),
    fval = fval,
    ifval = ifval,
    num.feval = num.feval,
    num.feval.fast = num.feval.fast,
    fval.history = fval.history,
    eval.history = eval.history,
    invalid.history = invalid.history,
    scaling = bwscaling,
    pscaling = npBandwidthSummaryLabel(bwtype = bwtype, bwscaling = bwscaling),
    type = bwtype,
    ptype = bwtToPrint(bwtype),
    ckertype = ckertype,    
    ckerorder = ckerorder,
    ckerbound = cbounds$bound,
    ckerlb = cbounds$lb,
    ckerub = cbounds$ub,
    pckertype = cktToPrint(ckertype, order = porder, kerbound = cbounds$bound),
    ukertype = ukertype,
    pukertype = uktToPrint(ukertype),
    okertype = okertype,
    pokertype = oktToPrint(okertype),
    nobs = nobs,
    ndim = ndim,
    ncon = ncon,
    nuno = sum(xdati$iuno),
    nord = sum(xdati$iord),
    icon = xdati$icon,
    iuno = xdati$iuno,
    iord = xdati$iord,
    xnames = xnames,
    ynames = ynames,
    xdati = xdati,
    ydati = ydati,
    xmcv = mcvConstruct(xdati),
    sfactor = list(x = sfactor),
    bandwidth = list(x = bandwidth),
    nconfac = nconfac,
    ncatfac = ncatfac,
    sdev = sdev,
    sumNum = list(x = sumNum),
    dati = list(x = xdati, y = ydati),
    varnames = list(x = xnames, y = ynames),
    vartitle = list(x = "Explanatory", y = "Dependent"),
    vartitleabb = list(x = "Exp.", y = "Dep."),
    rows.omit = rows.omit,
    nobs.omit = if (identical(rows.omit, NA)) 0 else length(rows.omit),
    timing = timing,
    total.time = total.time)

  
  mybw$klist = list(
    x =
    list(ckertype = ckertype,
         pckertype = mybw$pckertype,
         ukertype = ukertype,
         pukertype = mybw$pukertype,
         okertype = okertype,
         pokertype = mybw$pokertype))

  if(!bandwidth.compute)
    mybw$pmethod <- "Manual"

  class(mybw) = "rbandwidth"
  if(!any(is.na(mybw$bandwidth)))
    validateBandwidth(mybw)
  mybw
  
}

as.double.rbandwidth <- function(x, ...){
  x$bw
}

npInsertGlpSummary <- function(txt, basis, degree, bernstein){
  lines <- strsplit(txt, "\n", fixed = TRUE)[[1]]
  reg.idx <- grep("^Regression Type:", lines)

  if (length(reg.idx) != 1L)
    return(txt)
  if (is.null(basis) || !length(basis))
    basis <- "glp"

  basis.family <- npLpBasisFamilyLabel(basis)
  basis.rep <- npLpBasisRepresentationLabel(bernstein)
  basis.dim <- npLpBasisNcol(basis = basis, degree = degree)

  glp.lines <- c(
    paste("LP Basis Family:", basis.family),
    paste("LP Basis Representation:", basis.rep),
    paste("LP Basis Dimension (NCOL(B)):",
          format(basis.dim, scientific = FALSE, trim = TRUE))
  )

  lines <- append(lines, glp.lines, after = reg.idx)
  paste(lines, collapse = "\n")
}

## feature: when using dataframe interface, summary and print methods don't 
## provide info on the dependent variable
print.rbandwidth <- function(x, digits=NULL, ...){
  cat("\nRegression Data (",x$nobs," observations, ",x$ndim," variable(s)):\n\n",sep="")
  print(matrix(x$bw,ncol=x$ndim,dimnames=list(paste(x$pscaling,":",sep=""),x$xnames)))

  bw.sel.str <- genBwSelStr(x)
  if (identical(x$regtype, "lp") && x$ncon > 0)
    bw.sel.str <- npInsertGlpSummary(txt = bw.sel.str,
                                     basis = x$basis,
                                     degree = x$degree,
                                     bernstein = x$bernstein.basis)
  cat(bw.sel.str)
  cat(genBwKerStrs(x))

  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}


predict.rbandwidth <- function(object, ...) { npreg(bws = object, ...) }

summary.rbandwidth <- function(object, ...){
  cat("\nRegression Data (",object$nobs," observations, ",object$ndim," variable(s)):\n",sep="")

  cat(genOmitStr(object))
  bw.sel.str <- genBwSelStr(object)
  if (identical(object$regtype, "lp") && object$ncon > 0)
    bw.sel.str <- npInsertGlpSummary(txt = bw.sel.str,
                                     basis = object$basis,
                                     degree = object$degree,
                                     bernstein = object$bernstein.basis)
  cat(bw.sel.str)

  cat('\n')
  cat(genBwScaleStrs(object))
  cat(genBwKerStrs(object))

  cat(genTimingStr(object))
  cat("\n\n")

}

.np_condist_proper_shadow_enabled <- function() {
  isTRUE(getOption("np.condist.proper.shadow", FALSE))
}

.np_condist_proper_reason_message <- function(reason, where = "npcdist()") {
  switch(
    as.character(reason),
    not_requested = "proper-distribution repair was not requested",
    already_proper = "proper=TRUE was requested, but the estimator is already proper by construction",
    no_eval_grid = "proper=TRUE requires explicit evaluation data that define a full y-grid for each fixed x; paired row-wise evaluation is insufficient",
    scope_not_selected = "proper=TRUE was requested, but 'proper.control$apply' does not target the values returned by this object",
    slice_disabled = "proper=TRUE slice repair is disabled by internal dispatcher controls",
    slice_context_missing = "proper=TRUE slice repair requires explicit evaluation and training data context",
    slice_invalid_master_grid = "proper=TRUE slice repair could not construct a valid internal y-grid",
    slice_eval_failed = "proper=TRUE slice repair failed while evaluating the internal explicit-grid oracle",
    y_not_univariate_continuous = "proper=TRUE currently supports only univariate continuous y",
    gradients_unsupported = "proper=TRUE is currently unsupported when gradients=TRUE",
    x_slices_not_repeated = "proper=TRUE requires repeated fixed-x slices in the evaluation data",
    y_grid_not_common = "proper=TRUE requires every fixed-x slice to share a common ordered y-grid",
    y_grid_not_ordered = "proper=TRUE requires at least two strictly increasing y-grid points in each fixed-x slice",
    invalid_quadrature = "proper=TRUE could not construct valid grid weights from the evaluation y-grid",
    repair_disabled_shadow_mode = sprintf("proper=TRUE repair is disabled in internal shadow-validation mode in %s", where),
    sprintf("proper=TRUE is unsupported in %s for reason '%s'", where, as.character(reason))
  )
}

.np_condist_normalize_proper_control <- function(proper.control) {
  if (is.null(proper.control))
    proper.control <- list()

  if (!is.list(proper.control))
    stop("'proper.control' must be a list")

  ctrl <- list(
    tol = 1e-10,
    grid.check = TRUE,
    store.raw = TRUE,
    fail.on.unsupported = FALSE,
    mode = "grid",
    apply = "evaluation",
    slice.grid.size = 101L,
    slice.extend.factor = 0.1
  )

  known <- names(ctrl)
  if (length(proper.control)) {
    extra <- setdiff(names(proper.control), known)
    if (length(extra))
      stop(sprintf("unsupported entries in 'proper.control': %s",
                   paste(shQuote(extra), collapse = ", ")))
    ctrl[names(proper.control)] <- proper.control
  }

  ctrl$tol <- npValidatePositiveFiniteNumeric(ctrl$tol, "proper.control$tol")
  ctrl$grid.check <- npValidateScalarLogical(ctrl$grid.check, "proper.control$grid.check")
  ctrl$store.raw <- npValidateScalarLogical(ctrl$store.raw, "proper.control$store.raw")
  ctrl$fail.on.unsupported <- npValidateScalarLogical(
    ctrl$fail.on.unsupported,
    "proper.control$fail.on.unsupported"
  )
  ctrl$mode <- match.arg(as.character(ctrl$mode)[1L], c("grid", "slice"))
  ctrl$apply <- match.arg(as.character(ctrl$apply)[1L], c("evaluation", "fitted", "both"))
  ctrl$`slice.grid.size` <- npValidatePositiveInteger(
    ctrl$`slice.grid.size`,
    "proper.control$slice.grid.size"
  )
  ctrl$`slice.extend.factor` <- .np_condist_validate_nonnegative_finite_numeric(
    ctrl$`slice.extend.factor`,
    "proper.control$slice.extend.factor"
  )

  ctrl
}

.np_condist_validate_proper_args <- function(proper = FALSE,
                                             proper.method = c("isotonic"),
                                             proper.control = list()) {
  proper <- npValidateScalarLogical(proper, "proper")
  proper.method <- match.arg(as.character(proper.method)[1L], c("isotonic"))
  proper.control <- .np_condist_normalize_proper_control(proper.control)

  list(
    proper.requested = proper,
    proper.method = proper.method,
    proper.control = proper.control
  )
}

.np_condist_make_reason_info <- function(reason,
                                         supported = FALSE,
                                         slice.count = 0L,
                                         grid.common = FALSE,
                                         lower.bound = 0,
                                         upper.bound = 1,
                                         monotone.violations.raw = NULL,
                                         range.raw = NULL,
                                         projection.distance.l2 = NULL) {
  list(
    supported = supported,
    reason = as.character(reason)[1L],
    slice.count = as.integer(slice.count),
    grid.common = isTRUE(grid.common),
    lower.bound = lower.bound,
    upper.bound = upper.bound,
    monotone.violations.raw = monotone.violations.raw,
    range.raw = range.raw,
    projection.distance.l2 = projection.distance.l2
  )
}

.np_condist_weighted_pava <- function(f, w, tol = 1e-10) {
  f <- as.double(f)
  w <- as.double(w)

  if (length(f) != length(w))
    stop("length mismatch between distribution values and grid weights")
  if (any(!is.finite(f)))
    stop("distribution values must be finite")
  if (any(!is.finite(w)) || any(w <= 0))
    stop("grid weights must be finite and strictly positive")
  if (!is.finite(tol) || tol <= 0)
    stop("projection tolerance must be positive and finite")

  block.start <- seq_along(f)
  block.end <- seq_along(f)
  block.weight <- w
  block.value <- f
  nblock <- length(f)

  i <- 1L
  while (i < nblock) {
    if (block.value[i] <= block.value[i + 1L] + tol) {
      i <- i + 1L
      next
    }

    new.weight <- block.weight[i] + block.weight[i + 1L]
    new.value <- (block.weight[i] * block.value[i] +
                    block.weight[i + 1L] * block.value[i + 1L]) / new.weight

    block.end[i] <- block.end[i + 1L]
    block.weight[i] <- new.weight
    block.value[i] <- new.value

    if (i + 1L < nblock) {
      keep <- seq.int(i + 2L, nblock)
      block.start[(i + 1L):(nblock - 1L)] <- block.start[keep]
      block.end[(i + 1L):(nblock - 1L)] <- block.end[keep]
      block.weight[(i + 1L):(nblock - 1L)] <- block.weight[keep]
      block.value[(i + 1L):(nblock - 1L)] <- block.value[keep]
    }

    nblock <- nblock - 1L
    if (i > 1L)
      i <- i - 1L
  }

  fitted <- numeric(length(f))
  for (j in seq_len(nblock))
    fitted[block.start[j]:block.end[j]] <- block.value[j]

  fitted
}

.np_condist_project_bounded_isotonic <- function(f,
                                                 w,
                                                 lower = 0,
                                                 upper = 1,
                                                 tol = 1e-10) {
  f <- as.double(f)
  w <- as.double(w)

  if (!is.finite(lower) || !is.finite(upper) || lower > upper)
    stop("invalid lower/upper bounds for isotonic projection")

  big.weight <- max(1, sum(w)) * 1e12
  aug.fit <- .np_condist_weighted_pava(
    f = c(lower, f, upper),
    w = c(big.weight, w, big.weight),
    tol = tol
  )

  out <- aug.fit[-c(1L, length(aug.fit))]
  out <- pmin(pmax(out, lower), upper)

  if (any(diff(out) < -10 * tol))
    stop("bounded isotonic projection failed monotonicity check")

  out
}

.np_condist_prepare_proper_plan <- function(object, proper.control = list()) {
  proper.control <- .np_condist_normalize_proper_control(proper.control)

  grid <- .np_condens_detect_proper_grid(object = object, tol = proper.control$tol)
  if (!isTRUE(grid$supported)) {
    info <- .np_condist_make_reason_info(
      reason = grid$reason,
      supported = FALSE,
      slice.count = 0L,
      grid.common = FALSE
    )
    return(list(
      supported = FALSE,
      reason = grid$reason,
      proper.info = info
    ))
  }

  weights <- tryCatch(
    .np_condens_trapezoid_weights(grid$y.grid, tol = proper.control$tol),
    error = function(e) e
  )
  if (inherits(weights, "error")) {
    info <- .np_condist_make_reason_info(
      reason = "invalid_quadrature",
      supported = FALSE,
      slice.count = length(grid$slices),
      grid.common = TRUE
    )
    return(list(
      supported = FALSE,
      reason = "invalid_quadrature",
      proper.info = info
    ))
  }

  list(
    supported = TRUE,
    slices = grid$slices,
    weights = weights,
    y.grid = grid$y.grid,
    proper.control = proper.control
  )
}

.np_condist_project_values_with_plan <- function(values,
                                                 plan,
                                                 progress.label = NULL) {
  if (!isTRUE(plan$supported))
    stop("proper projection plan is not supported")

  is.vector.input <- is.null(dim(values))
  values.mat <- if (is.vector.input) {
    matrix(as.double(values), nrow = 1L)
  } else {
    data.matrix(values)
  }

  if (ncol(values.mat) != sum(lengths(plan$slices)))
    stop("value length mismatch for proper distribution projection")

  out <- values.mat
  progress <- NULL
  if (!is.vector.input && !is.null(progress.label) && nrow(values.mat) > 1L) {
    progress <- .np_plot_stage_progress_begin(
      total = nrow(values.mat),
      label = as.character(progress.label)[1L]
    )
    on.exit(.np_plot_progress_end(progress), add = TRUE)
  }
  for (row in seq_len(nrow(values.mat))) {
    for (idx in plan$slices) {
      out[row, idx] <- .np_condist_project_bounded_isotonic(
        f = values.mat[row, idx],
        w = plan$weights,
        lower = 0,
        upper = 1,
        tol = plan$proper.control$tol
      )
    }
    if (!is.null(progress))
      progress <- .np_plot_progress_tick(state = progress, done = row)
  }

  if (is.vector.input) as.vector(out[1L, ]) else out
}

.np_condist_apply_proper_grid <- function(object,
                                          proper.method = "isotonic",
                                          proper.control = list()) {
  proper.method <- match.arg(as.character(proper.method)[1L], c("isotonic"))
  proper.control <- .np_condist_normalize_proper_control(proper.control)

  plan <- .np_condist_prepare_proper_plan(
    object = object,
    proper.control = proper.control
  )
  if (!isTRUE(plan$supported)) {
    return(list(
      applied = FALSE,
      reason = plan$reason,
      proper.info = plan$proper.info
    ))
  }

  raw <- as.double(object$condist)
  repaired <- .np_condist_project_values_with_plan(raw, plan)
  monotone.violations.raw <- integer(length(plan$slices))
  range.raw <- matrix(NA_real_, nrow = length(plan$slices), ncol = 2L,
                      dimnames = list(NULL, c("min", "max")))
  projection.distance <- numeric(length(plan$slices))

  for (i in seq_along(plan$slices)) {
    idx <- plan$slices[[i]]
    f.slice <- raw[idx]
    g.slice <- repaired[idx]
    monotone.violations.raw[i] <- sum(diff(f.slice) < -proper.control$tol)
    range.raw[i, ] <- c(min(f.slice), max(f.slice))
    projection.distance[i] <- sqrt(sum(plan$weights * (g.slice - f.slice)^2))
  }

  info <- .np_condist_make_reason_info(
    reason = "applied",
    supported = TRUE,
    slice.count = length(plan$slices),
    grid.common = TRUE,
    monotone.violations.raw = monotone.violations.raw,
    range.raw = range.raw,
    projection.distance.l2 = projection.distance
  )

  list(
    applied = TRUE,
    condist = repaired,
    condist.raw = if (isTRUE(proper.control$store.raw)) raw else NULL,
    proper.info = info
  )
}

.np_condist_apply_proper <- function(object,
                                     proper.method = "isotonic",
                                     proper.control = list(),
                                     slice.context = NULL) {
  proper.method <- match.arg(as.character(proper.method)[1L], c("isotonic"))
  proper.control <- .np_condist_normalize_proper_control(proper.control)

  if (!isTRUE(object$trainiseval) && identical(proper.control$apply, "fitted")) {
    info <- .np_condist_make_reason_info(
      reason = "scope_not_selected",
      supported = TRUE,
      slice.count = 0L,
      grid.common = FALSE
    )
    return(list(applied = FALSE, reason = "scope_not_selected", proper.info = info))
  }

  grid.out <- .np_condist_apply_proper_grid(
    object = object,
    proper.method = proper.method,
    proper.control = proper.control
  )
  if (isTRUE(grid.out$applied))
    return(grid.out)

  if (!identical(proper.control$mode, "slice"))
    return(grid.out)

  .np_condist_apply_proper_slice(
    object = object,
    proper.method = proper.method,
    proper.control = proper.control,
    slice.context = slice.context,
    grid.out = grid.out
  )
}

.np_condist_finalize_proper_object <- function(object,
                                               proper = FALSE,
                                               proper.method = c("isotonic"),
                                               proper.control = list(),
                                               slice.context = NULL,
                                               where = "npcdist()") {
  args <- .np_condist_validate_proper_args(
    proper = proper,
    proper.method = proper.method,
    proper.control = proper.control
  )

  object$proper.requested <- args$proper.requested
  object$proper.applied <- FALSE
  object$proper.method <- if (isTRUE(args$proper.requested)) args$proper.method else NULL
  object$condist.raw <- NULL
  object$proper.info <- .np_condist_make_reason_info(
    reason = if (isTRUE(args$proper.requested)) "pending" else "not_requested",
    supported = FALSE
  )

  if (!isTRUE(args$proper.requested))
    return(object)

  if (.np_condist_is_already_proper_by_design(object$bws)) {
    object$proper.info <- .np_condist_make_reason_info(
      reason = "already_proper",
      supported = TRUE
    )
    return(object)
  }

  if (.np_condist_proper_shadow_enabled()) {
    object$proper.info <- .np_condist_make_reason_info(
      reason = "repair_disabled_shadow_mode",
      supported = FALSE
    )
    return(object)
  }

  proper.out <- .np_condist_apply_proper(
    object = object,
    proper.method = args$proper.method,
    proper.control = args$proper.control,
    slice.context = slice.context
  )

  if (!isTRUE(proper.out$applied)) {
    object$proper.info <- proper.out$proper.info
    if (isTRUE(args$proper.control$fail.on.unsupported) &&
        !isTRUE(proper.out$proper.info$supported)) {
      stop(.np_condist_proper_reason_message(
        reason = proper.out$reason,
        where = where
      ), call. = FALSE)
    }
    return(object)
  }

  object$condist <- proper.out$condist
  object$proper.applied <- TRUE
  object$proper.info <- proper.out$proper.info
  if (!is.null(proper.out$condist.raw))
    object$condist.raw <- proper.out$condist.raw

  object
}

deneqtest <- function(Tn,
                      In,
                      Tn.bootstrap,
                      In.bootstrap,
                      Tn.P,
                      In.P,                      
                      boot.num) {

  tdeneq = list(Tn=Tn,
    In=In,
    Tn.bootstrap=Tn.bootstrap,
    In.bootstrap=In.bootstrap,                
    Tn.P=Tn.P,
    In.P=In.P,    
    boot.num=boot.num)

  reject <- ' '
  
  if (Tn.P < 0.1)
    reject <- '.'

  if (Tn.P < 0.05)
    reject <- '*'

  if (Tn.P < 0.01)
    reject <- '**'

  if (Tn.P < 0.001)
    reject <- '***'

  tdeneq$reject <- reject
  tdeneq$rejectNum <- switch(reject,
                             ' ' = 100,
                             '.' = 10,
                             '*' = 5,
                             '**' = 1,
                             '***' = 0.1)
  
  class(tdeneq) <- "deneqtest"

  tdeneq
}

print.deneqtest <- function(x, ...){
  reject.msg <- if (x$reject == ' ') {
    "\nFail to reject the null of equality at the 10% level"
  } else {
    paste("\nNull of equality is rejected at the ", x$rejectNum, "% level", sep = "")
  }
  cat("\nConsistent Density Equality Test",
      paste("\n", format(x$boot.num), " Bootstrap Replications",sep=""),
      "\n\nTest Statistic ", sQuote("Tn"), ": ",
      format(x$Tn), "\tP Value: ", format.pval(x$Tn.P)," ", x$reject,
      "\n---\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1",
      reject.msg,
      "\n\n", sep="")
}

summary.deneqtest <- function(object, ...){
  print(object)
}

npregression <- 
    function(bws, eval, mean, merr = NA, grad = NA, gerr = NA,
             resid = NA,
             ntrain, trainiseval = FALSE, gradients = FALSE, residuals = FALSE,
             gradient.order = NULL,
             xtra = rep(NA, 6),
             rows.omit = NA,
             timing = NA,
             total.time = NA,
             optim.time = NA,
             fit.time = NA){

        if (missing(bws) || missing(eval) || missing(ntrain))
            stop("improper invocation of npregression constructor")

        if (length(rows.omit) == 0)
            rows.omit <- NA

        d <- list(
            bw = bws$bw,
            bws = bws,
            pregtype = bws$pregtype,
            xnames = bws$xnames,
            ynames = bws$ynames,
            nobs = dim(eval)[1],
            ndim = bws$ndim,
            nord = bws$nord,
            nuno = bws$nuno,
            ncon = bws$ncon,
            pscaling = bws$pscaling,
            ptype = bws$ptype,
            pckertype = bws$pckertype,
            pukertype = bws$pukertype,
            pokertype = bws$pokertype,
            eval = eval,
            mean = mean,
            merr = merr,
            grad = grad,
            gerr = gerr,
            resid = resid,
            ntrain = ntrain,
            trainiseval = trainiseval,
            gradients = gradients,
            gradient.order = gradient.order,
            residuals = residuals,
            R2 = xtra[1],
            MSE = xtra[2],
            MAE = xtra[3],
            MAPE = xtra[4],
            CORR = xtra[5],
            SIGN = xtra[6],
            rows.omit = rows.omit,
            nobs.omit = if (identical(rows.omit, NA)) 0 else length(rows.omit),
            timing = timing,
            total.time = total.time,
            optim.time = optim.time,
            fit.time = fit.time)

        class(d) <- "npregression"

        return(d)
    }

print.npregression <- function(x, digits=NULL, ...){
  cat("\nRegression Data: ", x$ntrain, " training points,",
      if (x$trainiseval) "" else paste(" and ", x$nobs," evaluation points,",
                                      sep=""),
      " in ",x$ndim," variable(s)\n",sep="")

  print(matrix(x$bw,ncol=x$ndim,dimnames=list(paste(x$pscaling,":",sep=""),x$xnames)))
  
  cat(genRegEstStr(x))
  cat(genBwKerStrs(x$bws))
  
  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

fitted.npregression <- function(object, ...){
 object$mean 
}
residuals.npregression <- function(object, ...) {
 if(object$residuals) { return(object$resid) } else { return(npreg(bws = object$bws, residuals =TRUE)$resid) } 
}
se.npregression <- function(x) { x$merr }
gradients.npregression <- function(x, errors = FALSE, gradient.order = NULL, ...) {
  errors <- npValidateScalarLogical(errors, "errors")
  gout <- if (!errors) x$grad else x$gerr
  if (is.null(gout) || (length(gout) == 1L && is.logical(gout) && is.na(gout)))
    stop(if (!errors)
      "gradients are not available: fit the model with gradients=TRUE"
    else
      "gradient standard errors are not available: fit the model with gradients=TRUE")

  if (!identical(x$bws$regtype, "lp") || is.null(gradient.order))
    return(gout)

  if (!is.matrix(gout))
    return(gout)

  gorder <- npValidateGlpGradientOrder(regtype = x$bws$regtype,
                                       gradient.order = gradient.order,
                                       ncon = x$bws$ncon)
  stored.order <- x$gradient.order
  if (is.null(stored.order) && x$bws$ncon > 0L)
    stored.order <- rep.int(1L, x$bws$ncon)
  stored.order <- npValidateGlpGradientOrder(regtype = x$bws$regtype,
                                             gradient.order = stored.order,
                                             ncon = x$bws$ncon)
  if (length(gorder)) {
    defined.request <- (gorder <= x$bws$degree)
    if (any(defined.request & (gorder != stored.order))) {
      stop("requested gradient.order differs from the derivative order stored in this npregression object; refit or predict/evaluate with gradients=TRUE and the desired gradient.order",
           call. = FALSE)
    }
  }

  gout.masked <- gout
  gout.masked[,] <- NA_real_
  cont.idx <- which(x$bws$icon)
  if (length(cont.idx)) {
    keep.cont <- (gorder <= x$bws$degree)
    if (any(keep.cont)) {
      keep.idx <- cont.idx[keep.cont]
      gout.masked[, keep.idx] <- gout[, keep.idx, drop = FALSE]
    }
    if (any(gorder > x$bws$degree))
      .np_warning("some requested glp derivatives exceed polynomial degree; returning NA for those components")
  }
  gout.masked
}
predict.npregression <- function(object, se.fit = FALSE, ...) {
  se.fit <- npValidateScalarLogical(se.fit, "se.fit")
  dots <- list(...)
  has.formula.route <- !is.null(object$bws$formula)

  if (!is.null(dots$exdat) && !is.null(dots$newdata)) {
    dots$newdata <- NULL
  } else if (!has.formula.route && is.null(dots$exdat) && !is.null(dots$newdata)) {
    dots$exdat <- dots$newdata
    dots$newdata <- NULL
  }

  tr <- do.call(npreg, c(list(bws = object$bws), dots))
  if(se.fit)
    return(list(fit = fitted(tr), se.fit = se(tr), 
                df = tr$nobs, residual.scale = tr$MSE))
  else
    return(fitted(tr))
}

summary.npregression <- function(object, ...) {
  cat("\nRegression Data: ", object$ntrain, " training points,",
      if (object$trainiseval) "" else paste(" and ", object$nobs," evaluation points,",
                                      sep=""),
      " in ",object$ndim," variable(s)\n",sep="")

  cat(genOmitStr(object))

  print(matrix(object$bw,ncol=object$ndim,dimnames=list(paste(object$pscaling,":",sep=""),object$xnames)))

  cat(genRegEstStr(object))
  cat(genGofStr(object))

  cat(genBwKerStrs(object$bws))
  cat(genTimingStr(object))
  cat('\n\n')  
}

npuniden.boundary <- function(X=NULL,
                              Y=NULL,
                              h=NULL,
                              a=min(X),
                              b=max(X),
                              bwmethod=c("cv.ls","cv.ml"),
                              cv=c("grid-hybrid","numeric"),
                              grid=NULL,
                              kertype=c("gaussian1","gaussian2",
                                  "beta1","beta2",
                                  "fb","fbl","fbu",
                                  "rigaussian","gamma"),
                              nmulti=1,
                              proper=FALSE) {
    kertype <- match.arg(kertype)
    cv <- match.arg(cv)
    bwmethod <- match.arg(bwmethod)
    if(!is.null(grid) && anyNA(grid)) stop("grid must not contain missing values")
    if(!is.null(grid) && any(grid<=0, na.rm = TRUE)) stop(" the grid vector must contain positive values")
    if(is.null(X)) stop("you must pass a vector X")
    if(anyNA(X)) stop("X must not contain missing values")
    if(kertype=="gamma" || kertype=="rigaussian") b <- Inf
    if(kertype=="fbl") b <- Inf
    if(kertype=="fbu") a <- -Inf
    if(a>=b) stop("a must be less than b")
    if(any(X<a, na.rm = TRUE)) stop("X must be >= a")
    if(any(X>b, na.rm = TRUE)) stop("X must be <= b")
    if(!is.null(Y) && anyNA(Y)) stop("Y must not contain missing values")
    if(!is.null(Y) && any(Y<a, na.rm = TRUE)) stop("Y must be >= a")
    if(!is.null(Y) && any(Y>b, na.rm = TRUE)) stop("Y must be <= b")
    if(is.null(Y)) Y <- X
    if(!is.null(h) && h <= 0) stop("bandwidth h must be positive")
    if(nmulti < 1) stop("number of multistarts nmulti must be positive")
    if(!is.logical(proper)) stop("proper must be either TRUE or FALSE")
    if(kertype=="gaussian2" && (!is.finite(a) || !is.finite(b))) stop("finite bounds are required for kertype gaussian2")
    h.opt <- NULL
    if(kertype=="gaussian1") {
        ## Gaussian reweighted boundary kernel function (bias of O(h))
        kernel <- function(x,X,h,a=0,b=1) {
            dnorm((x-X)/h)/(h*(pnorm((b-x)/h)-pnorm((a-x)/h)))
        }
    } else if(kertype=="gaussian2") {
        ## Gaussian reweighted second-order boundary kernel function
        ## (bias of O(h^2)). Instability surfaces for extremely large
        ## bandwidths relative to range of the data, so we shrink to
        ## the uniform when h exceeds 10,000 times the range (b-a)
        kernel <- function(x,X,h,a=0,b=1) {
            z <- (x-X)/h
            z.a <- (a-x)/h
            z.b <- (b-x)/h
            pnorm.zb.m.pnorm.za <- (pnorm(z.b)-pnorm(z.a))
            mu.1 <- (dnorm(z.a)-dnorm(z.b))/(pnorm.zb.m.pnorm.za)
            mu.2 <- 1+(z.a*dnorm(z.a)-z.b*dnorm(z.b))/(pnorm.zb.m.pnorm.za)
            mu.3 <- ((z.a**2+2)*dnorm(z.a)-(z.b**2+2)*dnorm(z.b))/(pnorm.zb.m.pnorm.za)
            aa <- mu.3/(mu.3-mu.1*mu.2)
            bb <- -mu.1/(mu.3-mu.1*mu.2)
            if((b-a)/h > 1e-04) {
                (aa+bb*z**2)*dnorm(z)/(h*pnorm.zb.m.pnorm.za)
            } else {
                rep(1/(b-a),length(X))
            }
        }
    } else if(kertype=="beta1") {
        ## Chen (1999), Beta 1 kernel function (bias of O(h), function
        ## of f' and f'', no division by h), need to rescale to
        ## integrate to 1 on [a,b]
        kernel <- function(x,X,h,a=0,b=1) {
            X <- (X-a)/(b-a)
            x <- (x-a)/(b-a)
            dbeta(X,x/h+1,(1-x)/h+1)/(b-a)
        }
    } else if(kertype=="beta2") {
        ## Chen (1999), Beta 2 kernel function (bias of O(h), function
        ## of f'' only, no division by h), need to rescale to
        ## integrate to 1 on [a,b]
        rho <- function(x,h) {2*h**2+2.5-sqrt(4*h**4+6*h**2+2.25-x**2-x/h)}
        kernel <- function(x,X,h,a=0,b=1) {
            X <- (X-a)/(b-a)
            x <- (x-a)/(b-a)
            if(x < 2*h && h < (b-a)) {
                dbeta(X,rho(x,h),(1-x)/h)/(b-a)
            } else if((2*h <= x && x <= 1-2*h) || h >= (b-a)) {
                dbeta(X,x/h,(1-x)/h)/(b-a)
            } else if(x > 1-2*h && h < (b-a)) {
                dbeta(X,x/h,rho(1-x,h))/(b-a)
            }
        }
    } else if(kertype=="gamma") {
        ## Gamma kernel function for x in [a,Inf]
        kernel <- function(x,X,h,a=0,b=1) {
            ## No division by h, rescale to lie in [0,Inf], b is a
            ## dummy, not used but needed to avoid warning about
            ## function kernel having different named arguments
            X <- X-a
            x <- x-a
            dgamma(X,x/h+1,1/h)
        }
    } else if(kertype=="rigaussian") {
        ## Reverse inverse Gaussian for x in [a,Inf]
        kernel <- function(x,X,h,a=0,b=1) {
            ## No division by h, rescale to lie in [0,Inf], b is a
            ## dummy, not used but needed to avoid warning about
            ## function kernel having different named arguments
            X <- X - a
            x <- x - a
            x.res <- sqrt(x**2+h*x)
            k <- exp(-x.res/(2*h)*(X/x.res+x.res/X-2))/sqrt(2*pi*h*X)
            k[is.nan(k)] <- 0
            k
        }
    } else if(kertype=="fb") {
        ## Floating boundary kernel (Scott (1992), Page 46), left and
        ## right bound, truncated biweight in interior
        kernel <- function(x,X,h,a=0,b=1) {
            t <- (X-x)/h
            if(x < a+h && h < (b-a)) {
                c <- (a-x)/h
                out <- numeric(length(t))
                mask <- (c <= t) & (t <= 2 + c)
                if(any(mask)) {
                    tm <- t[mask]
                    out[mask] <- .75 * (c + 1 - 1.25 * (1 + 2 * c) * (tm - c)^2) * (tm - (c + 2))^2
                }
                out / h
            } else if((a+h <= x && x <= b-h) || h >= (b-a)) {
                z.a <- (a-x)/h
                z.b <- (b-x)/h  
                rw <- (3*(z.b^5-z.a^5)-10*(z.b^3-z.a^3)+15*(z.b-z.a))/16
                rw[rw>1] <- 1
                out <- numeric(length(t))
                mask <- abs(t) < 1
                if(any(mask)) {
                    tm <- t[mask]
                    out[mask] <- (15 / 16) * (1 - tm^2)^2 / (h * rw)
                }
                out
            } else if(x > b-h && h < (b-a)) {
                c <- (b-x)/h
                out <- numeric(length(t))
                mask <- (c - 2 <= t) & (t <= c)
                if(any(mask)) {
                    tm <- t[mask]
                    out[mask] <- .75 * (1 - c + 1.25 * (-1 + 2 * c) * (tm - c)^2) * (tm - (c - 2))^2
                }
                out / h
            }
        }
    } else if(kertype=="fbl") {
        ## Floating boundary kernel (Scott (1992), Page 46), left bound
        kernel <- function(x,X,h,a=0,b=1) {
            t <- (X-x)/h
            if(x < a+h) {
                c <- (a-x)/h
                out <- numeric(length(t))
                mask <- (c <= t) & (t <= 2 + c)
                if(any(mask)) {
                    tm <- t[mask]
                    out[mask] <- .75 * (c + 1 - 1.25 * (1 + 2 * c) * (tm - c)^2) * (tm - (c + 2))^2
                }
                out / h
            } else {
                out <- numeric(length(t))
                mask <- abs(t) < 1
                if(any(mask)) {
                    tm <- t[mask]
                    out[mask] <- (15 / 16) * (1 - tm^2)^2 / h
                }
                out
            }
        }
    } else if(kertype=="fbu") {
        kernel <- function(x,X,h,a=0,b=1) {
            ## Floating boundary kernel (Scott (1992), Page 46), right bound
            t <- (X-x)/h
            if(x <= b-h) {
                out <- numeric(length(t))
                mask <- abs(t) < 1
                if(any(mask)) {
                    tm <- t[mask]
                    out[mask] <- (15 / 16) * (1 - tm^2)^2 / h
                }
                out
            } else {
                c <- (b-x)/h
                out <- numeric(length(t))
                mask <- (c - 2 <= t) & (t <= c)
                if(any(mask)) {
                    tm <- t[mask]
                    out[mask] <- .75 * (1 - c + 1.25 * (-1 + 2 * c) * (tm - c)^2) * (tm - (c - 2))^2
                }
                out / h
            }
        }
    }
    int.kernel.squared <- function(X,h,a=a,b=b) {
        ## Use numeric integration to compute Kappa, the integral of
        ## the square of the kernel function needed for the asymptotic
        ## standard error of the density estimate seq(a,b) will barf
        ## on -Inf or Inf, trap these cases and use extendrange
        if(is.finite(a) && is.finite(b)) X.seq <- seq(a,b,length=1000)
        if(is.finite(a) && !is.finite(b)) X.seq <- seq(a,extendrange(X,f=10)[2],length=1000)
        if(!is.finite(a) && is.finite(b)) X.seq <- seq(extendrange(X,f=10)[1],b,length=1000)
        if(!is.finite(a) && !is.finite(b)) X.seq <- seq(extendrange(X,f=10)[1],extendrange(X,f=10)[2],length=1000)
        sapply(seq_along(X), function(i){integrate.trapezoidal(X.seq,h*kernel(X[i],X.seq,h,a,b)**2)[length(X.seq)]})
    }
    fhat <- function(X,Y,h,a=0,b=1,proper=FALSE) {
        f <- sapply(seq_along(Y), function(i){mean(kernel(Y[i],X,h,a,b))})
        if(proper) {
            if(is.finite(a) && is.finite(b)) X.seq <- seq(a,b,length=1000)
            if(is.finite(a) && !is.finite(b)) X.seq <- seq(a,extendrange(X,f=10)[2],length=1000)
            if(!is.finite(a) && is.finite(b)) X.seq <- seq(extendrange(X,f=10)[1],b,length=1000)
            if(!is.finite(a) && !is.finite(b)) X.seq <- seq(extendrange(X,f=10)[1],extendrange(X,f=10)[2],length=1000)
            f.seq <- sapply(seq_along(X.seq), function(i){mean(kernel(X.seq[i],X,h,a,b))})
            if(any(f.seq<0)) {
                f <- f - min(f.seq)
                f.seq <- f.seq - min(f.seq)
            }
            int.f.seq <- integrate.trapezoidal(X.seq,f.seq)[length(X.seq)]
            f <- f/int.f.seq
        }
        return(f)
    }
    Fhat <- function(Y,f,a,b,proper=FALSE) {
        ## Numerical integration of f, check for aberrant values, if
        ## on range of data ensure F\in[0,1], if not make sure value
        ## is proper (negative boundary kernel functions can cause
        ## unwanted artifacts)
        f[is.na(f)] <- 0
        F <- integrate.trapezoidal(Y,f)
        if(proper) {
            if(min(Y)==a && max(Y)==b) {
                F <- (F-min(F))/(max(F)-min(F))
            } else {
                if(min(F)<0) F <- F+min(F)
                if(max(F)>1) F <- F/max(F)
            }
        }
        F
    }
    fhat.loo <- function(X,h,a=0,b=1) {
        n <- length(X)
        if (n <= 1L) return(rep(NA_real_, n))
        sapply(seq_along(X), function(i){
            kv <- kernel(X[i], X, h, a, b)
            (sum(kv) - kv[i])/(n - 1L)
        })
    }
    if(bwmethod=="cv.ml") {
        ## Likelihood cross-validation function (maximizing)
        fnscale <- list(fnscale = -1)
        cv.function <- function(h,X,a=0,b=1) {
            f.loo <- fhat.loo(X,h,a,b)
            good <- (f.loo > 0) & is.finite(f.loo)
            f.safe <- f.loo
            f.safe[!good] <- .Machine$double.xmin
            return(sum(log(f.safe)))
        }
    } else {
        ## Least-squares cross-validation function (minimizing)
        fnscale <- list(fnscale = 1) 
        cv.function <- function(h,X,a=0,b=1) {
            cv.ls <- (integrate.trapezoidal(X,fhat(X,X,h,a,b)**2)[order(X)])[length(X)]-2*mean(fhat.loo(X,h,a,b))
            if (is.finite(cv.ls)) cv.ls else sqrt(sqrt(.Machine$double.xmax))
        }
    }
    ## Grid search and then numeric optimization search (no
    ## multistarting, but sound starting point always used for
    ## subsequent refinement by optim)
    if(is.null(h) && cv == "grid-hybrid") {
        ## First establish a sound starting value using grid search,
        ## then use that starting value for numeric search
        if(is.null(grid)) {
            rob.spread <- c(sd(X),IQR(X)/1.349)
            rob.spread <- min(rob.spread[rob.spread>0])
            constant <- rob.spread*length(X)**(-0.2)
            h.vec <- c(seq(0.25,1.75,length=10),2^(1:25))*constant
            cv.vec <- sapply(seq_along(h.vec), function(i){cv.function(h.vec[i],X,a,b)})
            start.idx <- if (bwmethod=="cv.ml") which.max(cv.vec) else which.min(cv.vec)
            upper.bound <- if (kertype=="beta2") (b-a)/4 else Inf
            foo <- optim(h.vec[start.idx],
                         cv.function,
                         method="L-BFGS-B",
                         lower=sqrt(.Machine$double.eps),
                         upper=upper.bound,
                         control = fnscale,
                         X=X,
                         a=a,
                         b=b)
            h.opt <- foo$par
            cv.opt <- foo$value
        } else {
            cv.vec <- sapply(seq_along(grid), function(i){cv.function(grid[i],X,a,b)})
            start.idx <- if (bwmethod=="cv.ml") which.max(cv.vec) else which.min(cv.vec)
            upper.bound <- if (kertype=="beta2") (b-a)/4 else Inf
            foo <- optim(grid[start.idx],
                         cv.function,
                         method="L-BFGS-B",
                         lower=sqrt(.Machine$double.eps),
                         upper=upper.bound,
                         control = fnscale,
                         X=X,
                         a=a,
                         b=b)
            h.opt <- foo$par
            cv.opt <- foo$value
        }
    }
    if(is.null(h.opt)) {
        ## Manual inputted bandwidth
        f <- fhat(X,Y,h,a,b,proper=proper)
        ## Numerical integration via the trapezoidal rule
        F <- Fhat(Y,f,a,b,proper=proper)
        return(list(f=f,
                    F=F,
                    sd.f=sqrt(abs(f*int.kernel.squared(Y,h,a,b)/(h*length(f)))),
                    sd.F=sqrt(abs(F*(1-F)/length(F))),
                    h=h))
    } else {
        ## Search bandwidth
        f <- fhat(X,Y,h.opt,a,b,proper=proper)
        ## Numerical integration via the trapezoidal rule
        F <- Fhat(Y,f,a,b,proper=proper)
        return(list(f=f,
                    F=F,
                    sd.f=sqrt(abs(f*int.kernel.squared(Y,h.opt,a,b)/(h.opt*length(f)))),
                    sd.F=sqrt(abs(F*(1-F)/length(F))),
                    h=h.opt,
                    nmulti=nmulti,
                    cv.opt=cv.opt))
    }
}

npcdist <-
  function(bws, ...){
    args <- list(...)

    if (!missing(bws)){
      if (is.recursive(bws)){
        if (!is.null(bws$formula) && is.null(args$txdat))
          UseMethod("npcdist",bws$formula)
        else if (!is.null(bws$call) && is.null(args$txdat))
          UseMethod("npcdist",bws$call)
        else if (!is.call(bws))
          UseMethod("npcdist",bws)
        else
          UseMethod("npcdist",NULL)
      } else {
        UseMethod("npcdist", NULL)
      }
    } else {
      UseMethod("npcdist", NULL)
    }
  }

npcdist.formula <-
  function(bws, data = NULL, newdata = NULL, ...){
    tt <- terms(bws)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)
    tmf <- bws$call[c(1,m)]
    tmf[[1]] <- as.name("model.frame")
    tmf[["formula"]] <- tt
    mf.args <- as.list(tmf)[-1L]
    umf <- tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

    tydat <- tmf[, bws$variableNames[["response"]], drop = FALSE]
    txdat <- tmf[, bws$variableNames[["terms"]], drop = FALSE]

    has.eval <- !is.null(newdata)
    if (has.eval) {
      npValidateNewdataFormula(newdata, tt, include.response = TRUE)
      umf.args <- list(formula = tt, data = newdata)
      umf <- do.call(stats::model.frame, umf.args, envir = parent.frame())
      emf <- umf
      
      eydat <- emf[, bws$variableNames[["response"]], drop = FALSE]
      exdat <- emf[, bws$variableNames[["terms"]], drop = FALSE]
    }

    cd.args <- list(txdat = txdat, tydat = tydat)
    if (has.eval) {
      cd.args$exdat <- exdat
      cd.args$eydat <- eydat
    }
    cd.args$bws <- bws
    ev <- do.call(npcdist, c(cd.args, list(...)))

    ev$omit <- attr(umf,"na.action")
    ev$rows.omit <- as.vector(ev$omit)
    ev$nobs.omit <- length(ev$rows.omit)

    ev$condist <- napredict(ev$omit, ev$condist)
    ev$conderr <- napredict(ev$omit, ev$conderr)
    if (!is.null(ev$condist.raw))
      ev$condist.raw <- napredict(ev$omit, ev$condist.raw)

    if(ev$gradients){
        ev$congrad <- napredict(ev$omit, ev$congrad)
        ev$congerr <- napredict(ev$omit, ev$congerr)
    }

    return(ev)
  }

npcdist.call <-
  function(bws, ...) {
    npcdist(txdat = .np_eval_bws_call_arg(bws, "xdat"),
            tydat = .np_eval_bws_call_arg(bws, "ydat"),
            bws = bws, ...)
  }


npcdist.condbandwidth <-
  function(bws,
           txdat = stop("invoked without training data 'txdat'"),
           tydat = stop("invoked without training data 'tydat'"),
           exdat, eydat, gradients = FALSE,
           gradient.order = 1L,
           proper = FALSE,
           proper.method = c("isotonic"),
           proper.control = list(),
           ...){

    dots <- list(...)
    fit.start <- proc.time()[3]
    fit.progress.handoff <- isTRUE(dots$.np_fit_progress_handoff)
    gradients <- npValidateScalarLogical(gradients, "gradients")
    proper.args <- .np_condist_validate_proper_args(
      proper = proper,
      proper.method = proper.method,
      proper.control = proper.control
    )

    if (xor(missing(exdat),missing(eydat)))
      stop("evaluation data must be supplied for both 'exdat' and 'eydat'")

    no.exy = missing(exdat)

    txdat = toFrame(txdat)
    tydat = toFrame(tydat)

    if (!no.exy){
      exdat = toFrame(exdat)
      eydat = toFrame(eydat)

      if (! txdat %~% exdat )
        stop("'txdat' and 'exdat' are not similar data frames!")

      if (! tydat %~% eydat )
        stop("'tydat' and 'eydat' are not similar data frames!")

    }

    if (length(bws$xbw) != length(txdat))
      stop("length of bandwidth vector does not match number of columns of 'txdat'")

    if (length(bws$ybw) != length(tydat))
      stop("length of bandwidth vector does not match number of columns of 'tydat'")

    if ((any(bws$ixcon) &&
         !all(vapply(txdat[, bws$ixcon, drop = FALSE], inherits, logical(1), c("integer", "numeric")))) ||
        (any(bws$ixord) &&
         !all(vapply(txdat[, bws$ixord, drop = FALSE], inherits, logical(1), "ordered"))) ||
        (any(bws$ixuno) &&
         !all(vapply(txdat[, bws$ixuno, drop = FALSE], inherits, logical(1), "factor"))))
      stop("supplied bandwidths do not match 'txdat' in type")

    if ((any(bws$iycon) &&
         !all(vapply(tydat[, bws$iycon, drop = FALSE], inherits, logical(1), c("integer", "numeric")))) ||
        (any(bws$iyord) &&
         !all(vapply(tydat[, bws$iyord, drop = FALSE], inherits, logical(1), "ordered"))) ||
        (any(bws$iyuno) &&
         !all(vapply(tydat[, bws$iyuno, drop = FALSE], inherits, logical(1), "factor"))))
      stop("supplied bandwidths do not match 'tydat' in type")
    
    ## catch and destroy NA's
    keep.rows <- rep_len(TRUE, nrow(txdat))
    rows.omit <- attr(na.omit(data.frame(txdat, tydat)), "na.action")
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Data has no rows without NAs")

    txdat <- txdat[keep.rows,,drop = FALSE]
    tydat <- tydat[keep.rows,,drop = FALSE]

    if (!no.exy){
      keep.eval <- rep_len(TRUE, nrow(exdat))
      rows.omit <- attr(na.omit(data.frame(exdat, eydat)), "na.action")
      if (length(rows.omit) > 0L)
        keep.eval[as.integer(rows.omit)] <- FALSE

      if (!any(keep.eval))
        stop("Data has no rows without NAs")

      exdat <- exdat[keep.eval,,drop = FALSE]
      eydat <- eydat[keep.eval,,drop = FALSE]
    }


    tnrow = nrow(txdat)
    enrow = (if (no.exy) tnrow else nrow(exdat))

    ## re-assign levels in training and evaluation data to ensure correct
    ## conversion to numeric type.
    
    txdat <- adjustLevels(txdat, bws$xdati)
    tydat <- adjustLevels(tydat, bws$ydati)
    
    if (!no.exy){
      exdat <- adjustLevels(exdat, bws$xdati, allowNewCells = TRUE)
      eydat <- adjustLevels(eydat, bws$ydati, allowNewCells = TRUE)
      npKernelBoundsCheckEval(exdat, bws$ixcon, bws$cxkerlb, bws$cxkerub, argprefix = "cxker")
      npKernelBoundsCheckEval(eydat, bws$iycon, bws$cykerlb, bws$cykerub, argprefix = "cyker")
    }

    proper.slice.context <- list(
      txdat = txdat,
      tydat = tydat,
      exdat = if (no.exy) NULL else exdat,
      eydat = if (no.exy) NULL else eydat
    )

    ## grab the evaluation data before it is converted to numeric
    if(no.exy){
      txeval <- txdat
      tyeval <- tydat
    } else {
      txeval <- exdat
      tyeval <- eydat
    }

    ## at this stage, data to be sent to the c routines must be converted to
    ## numeric type.
    
    tydat = toMatrix(tydat)

    tyuno = tydat[, bws$iyuno, drop = FALSE]
    tycon = tydat[, bws$iycon, drop = FALSE]
    tyord = tydat[, bws$iyord, drop = FALSE]


    txdat = toMatrix(txdat)

    txuno = txdat[, bws$ixuno, drop = FALSE]
    txcon = txdat[, bws$ixcon, drop = FALSE]
    txord = txdat[, bws$ixord, drop = FALSE]

    if (!no.exy){
      eydat = toMatrix(eydat)

      eyuno = eydat[, bws$iyuno, drop = FALSE]
      eycon = eydat[, bws$iycon, drop = FALSE]
      eyord = eydat[, bws$iyord, drop = FALSE]


      exdat = toMatrix(exdat)

      exuno = exdat[, bws$ixuno, drop = FALSE]
      excon = exdat[, bws$ixcon, drop = FALSE]
      exord = exdat[, bws$ixord, drop = FALSE]
    } else {
      eyuno = data.frame()
      eycon = data.frame()
      eyord = data.frame()

      exuno = data.frame()
      excon = data.frame()
      exord = data.frame()
    }

    reg.spec <- npConditionalRegEngineSpec(bws, where = "npcdist")
    reg.engine <- reg.spec$reg.engine
    basis.engine <- reg.spec$basis.engine
    degree.engine <- reg.spec$degree.engine
    bernstein.engine <- reg.spec$bernstein.engine
    glp.gradient.order <- npConditionalGradientOrder(
      bws = bws,
      reg.engine = reg.engine,
      gradient.order = gradient.order,
      where = "npcdist"
    )
    if (isTRUE(gradients) &&
        identical(reg.engine, "lp") &&
        (bws$xncon > 0L) &&
        all(degree.engine == 0L)) {
      stop("regtype='lp' with degree=0 does not support derivatives; use gradients=FALSE for fitted/predicted values")
    }

    reg.c <- npRegtypeToC(
      regtype = if (identical(reg.engine, "lp")) "lp" else "lc",
      degree = degree.engine,
      ncon = bws$xncon,
      context = "npcdist"
    )
    degree.c <- if (bws$xncon > 0L) {
      as.integer(if (is.null(reg.c$degree)) rep.int(0L, bws$xncon) else reg.c$degree)
    } else {
      integer(0)
    }
    basis.code <- as.integer(npLpBasisCode(basis.engine))

    myopti <- list(
        num_obs_train = tnrow,
        num_obs_eval = enrow,
        int_LARGE_SF = (if (bws$scaling) SF_NORMAL else SF_ARB),
        BANDWIDTH_den_extern = switch(bws$type,
            fixed = BW_FIXED,
            generalized_nn = BW_GEN_NN,
            adaptive_nn = BW_ADAP_NN),
        int_MINIMIZE_IO=if (isTRUE(getOption("np.messages"))) IO_MIN_FALSE else IO_MIN_TRUE,
        xkerneval = switch(bws$cxkertype,
            gaussian = CKER_GAUSS + bws$cxkerorder/2 - 1,
            epanechnikov = CKER_EPAN + bws$cxkerorder/2 - 1,
            uniform = CKER_UNI,
            "truncated gaussian" = CKER_TGAUSS),
        ykerneval = switch(bws$cykertype,
            gaussian = CKER_GAUSS + bws$cykerorder/2 - 1,
            epanechnikov = CKER_EPAN + bws$cykerorder/2 - 1,
            uniform = CKER_UNI,
            "truncated gaussian" = CKER_TGAUSS),
        uxkerneval = switch(bws$uxkertype,
            aitchisonaitken = UKER_AIT,
            liracine = UKER_LR),
        uykerneval = switch(bws$uykertype,
            aitchisonaitken = UKER_AIT,
            liracine = UKER_LR),
        oxkerneval = switch(bws$oxkertype,
            wangvanryzin = OKER_WANG,
            liracine = OKER_NLR,
        "racineliyan" = OKER_RLY),
        oykerneval = switch(bws$oykertype,
            wangvanryzin = OKER_WANG,
            liracine = OKER_NLR,
        "racineliyan" = OKER_RLY),
        num_yuno = bws$ynuno,
        num_yord = bws$ynord,
        num_ycon = bws$yncon,
        num_xuno = bws$xnuno,
        num_xord = bws$xnord,
        num_xcon = bws$xncon,
        no.exy = no.exy,
        gradients = gradients,
        ymcv.numRow = attr(bws$ymcv, "num.row"),
        xmcv.numRow = attr(bws$xmcv, "num.row"),
        densOrDist = NP_DO_DIST,
        int_do_tree = if (isTRUE(getOption("np.tree"))) DO_TREE_YES else DO_TREE_NO)

    cxker.bounds.c <- npKernelBoundsMarshal(bws$cxkerlb[bws$ixcon], bws$cxkerub[bws$ixcon])
    cyker.bounds.c <- npKernelBoundsMarshal(bws$cykerlb[bws$iycon], bws$cykerub[bws$iycon])

    myout <- .np_with_compiled_fit_progress(
      label = "Fitting conditional distribution",
      total = .np_condensdist_fit_total(bws = bws, tnrow = tnrow, enrow = enrow),
      handoff = fit.progress.handoff,
      handoff.detail = if (fit.progress.handoff) "starting" else NULL,
      .Call("C_np_density_conditional",
            as.double(tyuno), as.double(tyord), as.double(tycon),
            as.double(txuno), as.double(txord), as.double(txcon),
            as.double(eyuno), as.double(eyord), as.double(eycon),
            as.double(exuno), as.double(exord), as.double(excon),
            as.double(c(bws$xbw[bws$ixcon], bws$ybw[bws$iycon],
                        bws$ybw[bws$iyuno], bws$ybw[bws$iyord],
                        bws$xbw[bws$ixuno], bws$xbw[bws$ixord])),
            as.double(bws$ymcv), as.double(attr(bws$ymcv, "pad.num")),
            as.double(bws$xmcv), as.double(attr(bws$xmcv, "pad.num")),
            as.double(bws$nconfac), as.double(bws$ncatfac), as.double(bws$sdev),
            as.integer(myopti),
            as.integer(enrow),
            as.integer(bws$xndim),
            as.double(cxker.bounds.c$lb),
            as.double(cxker.bounds.c$ub),
            as.double(cyker.bounds.c$lb),
            as.double(cyker.bounds.c$ub),
            as.integer(reg.c$code),
            as.integer(degree.c),
            as.integer(bernstein.engine),
            basis.code,
            PACKAGE = "np")
    )
    names(myout)[1] <- "condist"

    if(gradients){
      myout$congrad = matrix(data=myout$congrad, nrow = enrow, ncol = bws$xndim, byrow = FALSE) 
      rorder = numeric(bws$xndim)
      xidx <- seq_len(bws$xndim)
      rorder[c(xidx[bws$ixcon], xidx[bws$ixuno], xidx[bws$ixord])] <- xidx
      myout$congrad = myout$congrad[, rorder, drop = FALSE]

      myout$congerr = matrix(data=myout$congerr, nrow = enrow, ncol = bws$xndim, byrow = FALSE)
      myout$congerr = myout$congerr[, rorder, drop = FALSE]

      if (identical(reg.engine, "lp") && bws$xncon > 0L) {
        cont.idx <- which(bws$ixcon)
        invalid.order <- glp.gradient.order > degree.engine
        if (any(invalid.order)) {
          myout$congrad[, cont.idx[invalid.order]] <- NA_real_
          myout$congerr[, cont.idx[invalid.order]] <- NA_real_
          .np_warning("some requested glp derivatives exceed polynomial degree; returning NA for those components")
        }

        higher.order <- (glp.gradient.order > 1L) & !invalid.order
        if (any(higher.order)) {
          rhs <- rep.int(1.0, nrow(proper.slice.context$txdat))
          for (jj in which(higher.order)) {
            svec <- integer(bws$xncon)
            svec[jj] <- glp.gradient.order[jj]
            hat.args <- list(
              bws = bws,
              txdat = proper.slice.context$txdat,
              tydat = proper.slice.context$tydat,
              y = rhs,
              output = "apply",
              s = svec
            )
            if (!no.exy) {
              hat.args$exdat <- proper.slice.context$exdat
              hat.args$eydat <- proper.slice.context$eydat
            }
            myout$congrad[, cont.idx[jj]] <- as.vector(do.call(npcdisthat, hat.args))
            myout$congerr[, cont.idx[jj]] <- NA_real_
          }
        }
      }
    } else {
      myout$congrad = NA
      myout$congerr = NA
    }


    fit.elapsed <- proc.time()[3] - fit.start
    optim.time <- if (!is.null(bws$total.time) && is.finite(bws$total.time)) as.double(bws$total.time) else NA_real_
    total.time <- fit.elapsed + (if (is.na(optim.time)) 0.0 else optim.time)

    out <- condistribution(bws = bws,
                           xeval = txeval,
                           yeval = tyeval,
                           condist = myout$condist, conderr = myout$conderr,
                           congrad = myout$congrad, congerr = myout$congerr,
                           ntrain = tnrow, trainiseval = no.exy, gradients = gradients,
                           gradient.order = if (identical(reg.engine, "lp")) glp.gradient.order else NULL,
                           rows.omit = rows.omit,
                           timing = bws$timing, total.time = total.time,
                           optim.time = optim.time, fit.time = fit.elapsed)
    out$nomad.time <- if (!is.null(bws$nomad.time) && is.finite(bws$nomad.time)) as.double(bws$nomad.time) else NA_real_
    out$powell.time <- if (!is.null(bws$powell.time) && is.finite(bws$powell.time)) as.double(bws$powell.time) else NA_real_

    .np_condist_finalize_proper_object(
      object = out,
      proper = proper.args$proper.requested,
      proper.method = proper.args$proper.method,
      proper.control = proper.args$proper.control,
      slice.context = proper.slice.context,
      where = "npcdist()"
    )

  }

npcdist.default <- function(bws, txdat, tydat, nomad = FALSE, ...){
  sc <- sys.call()
  sc.names <- names(sc)
  nomad <- npValidateScalarLogical(nomad, "nomad")

  ## here we check to see if the function was called with tdat =
  ## if it was, we need to catch that and map it to dat =
  ## otherwise the call is passed unadulterated to npudensbw

  bws.named <- any(sc.names == "bws")
  txdat.named <- any(sc.names == "txdat")
  tydat.named <- any(sc.names == "tydat")

  no.bws <- missing(bws)
  no.txdat <- missing(txdat)
  no.tydat <- missing(tydat)
  has.explicit.bws <- (!no.bws) && isa(bws, "condbandwidth")
  bws.formula <- (!no.bws) && inherits(bws, "formula")

  if (bws.named && no.txdat && no.tydat && bws.formula) {
    sc$`bws` <- NULL
    sc$formula <- bws
    sc.bw <- sc
    sc.bw[[1]] <- quote(npcdistbw)
    bws.named <- FALSE
  } else {
    sc.bw <- sc
    sc.bw[[1]] <- quote(npcdistbw)
  }

  ## if bws was passed in explicitly, do not compute bandwidths
    
  if(txdat.named)
    txdat <- toFrame(txdat)

  if(tydat.named)
    tydat <- toFrame(tydat)

  if(bws.named){
    sc.bw$bandwidth.compute <- FALSE
  }

  ostxy <- c('txdat','tydat')
  nstxy <- c('xdat','ydat')
  
  m.txy <- match(ostxy, names(sc.bw), nomatch = 0)

  if(any(m.txy > 0)) {
    names(sc.bw)[m.txy] <- nstxy[m.txy > 0]
  }
  sc.bw$newdata <- NULL
  sc.bw$exdat <- NULL
  sc.bw$eydat <- NULL
  sc.bw$gradients <- NULL
  sc.bw$gradient.order <- NULL
  sc.bw$proper <- NULL
  sc.bw$proper.method <- NULL
  sc.bw$proper.control <- NULL
    
  use.outer.bandwidth.progress <- !.np_bw_call_uses_nomad_degree_search(
    sc.bw,
    caller_env = parent.frame()
  )

  tbw <- if (!has.explicit.bws) {
    if (use.outer.bandwidth.progress) {
      .np_progress_select_bandwidth_enhanced(
        "Selecting conditional distribution bandwidth",
        .np_eval_bw_call(sc.bw, caller_env = parent.frame())
      )
    } else {
      .np_eval_bw_call(sc.bw, caller_env = parent.frame())
    }
  } else {
    .np_eval_bw_call(sc.bw, caller_env = parent.frame())
  }

  call.args <- list(bws = tbw)
  if (no.bws) {
    call.args$txdat <- txdat
    call.args$tydat <- tydat
  } else {
    if (txdat.named) call.args$txdat <- txdat
    if (tydat.named) call.args$tydat <- tydat
    if ((!bws.named) && (!txdat.named) && (!no.tydat) && (!tydat.named)) {
      call.args <- c(call.args, list(tydat))
    }
  }
  if (!has.explicit.bws)
    call.args$.np_fit_progress_handoff <- TRUE
  do.call(npcdist, c(call.args, list(...)))
}

test_that("fixed conditional ll/lp density-distribution helper matches duplicate-sample refits", {
  skip_if_not_installed("np")

  helper <- getFromNamespace(".np_inid_boot_from_ksum_conditional", "np")

  run_case <- function(family = c("dens", "dist"), regtype = c("ll", "lp")) {
    family <- match.arg(family)
    regtype <- match.arg(regtype)

    set.seed(switch(
      paste(family, regtype, sep = "_"),
      dens_ll = 603301L,
      dist_ll = 603302L,
      dens_lp = 603303L,
      dist_lp = 603304L
    ))

    n <- 55L
    B <- 7L
    tx <- data.frame(x = sort(runif(n)))
    ty <- data.frame(y = 0.5 * tx$x + rnorm(n, sd = 0.2))
    ex <- data.frame(x = seq(min(tx$x), max(tx$x), length.out = 13L))
    ey <- data.frame(y = seq(as.numeric(quantile(ty$y, 0.15)), as.numeric(quantile(ty$y, 0.85)), length.out = 13L))
    counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

    bw.args <- list(
      xdat = tx,
      ydat = ty,
      bws = c(0.28, 0.28),
      bandwidth.compute = FALSE,
      bwtype = "fixed",
      regtype = regtype
    )
    if (identical(regtype, "lp")) {
      bw.args$basis <- "glp"
      bw.args$degree <- 2L
    }

    bws <- if (identical(family, "dens")) {
      do.call(npcdensbw, bw.args)
    } else {
      do.call(npcdistbw, bw.args)
    }

    fast <- helper(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = bws,
      B = B,
      cdf = identical(family, "dist"),
      counts = counts
    )
    fast.drawer <- helper(
      xdat = tx,
      ydat = ty,
      exdat = ex,
      eydat = ey,
      bws = bws,
      B = B,
      cdf = identical(family, "dist"),
      counts.drawer = function(start, stop) counts[, start:stop, drop = FALSE]
    )

    explicit <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
    for (b in seq_len(B)) {
      idx <- rep.int(seq_len(n), counts[, b])
      explicit[b, ] <- if (identical(family, "dens")) {
        npcdens(
          txdat = tx[idx, , drop = FALSE],
          tydat = ty[idx, , drop = FALSE],
          exdat = ex,
          eydat = ey,
          bws = bws
        )$condens
      } else {
        npcdist(
          txdat = tx[idx, , drop = FALSE],
          tydat = ty[idx, , drop = FALSE],
          exdat = ex,
          eydat = ey,
          bws = bws
        )$condist
      }
    }

    t0 <- if (identical(family, "dens")) {
      npcdens(txdat = tx, tydat = ty, exdat = ex, eydat = ey, bws = bws)$condens
    } else {
      npcdist(txdat = tx, tydat = ty, exdat = ex, eydat = ey, bws = bws)$condist
    }

    expect_equal(fast$t, explicit, tolerance = 1e-10, info = paste(family, regtype))
    expect_equal(fast$t0, t0, tolerance = 1e-12, info = paste(family, regtype))
    expect_equal(fast.drawer$t, explicit, tolerance = 1e-10, info = paste(family, regtype, "drawer"))
    expect_equal(fast.drawer$t0, t0, tolerance = 1e-12, info = paste(family, regtype, "drawer"))
  }

  run_case("dens", "ll")
  run_case("dist", "ll")
  run_case("dens", "lp")
  run_case("dist", "lp")
})

sigtest <- function(In,
                    In.bootstrap,
                    P,
                    bws = NA,
                    ixvar,
                    boot.method,
                    pivot,
                    joint,
                    boot.type,
                    boot.num){

  tsig <- list(In = In,
               In.bootstrap = In.bootstrap,
               P = P,
               bws = bws,
               ixvar = ixvar,
               pmethod = switch(boot.method,
                 "pairwise" = "Pairwise",
                 "iid" = "IID",
                 "wild" ="Wild",
                 "wild-rademacher" = "Rademacher Wild"),
               pivot = pivot,
               joint = joint,
               ptype = boot.type,
               boot.num = boot.num)
  
  tsig$reject <- rep('', length(In))
  tsig$rejectNum <- rep(NA, length(In))

  tsig$reject[a <- (P < 0.1)] <- '.'
  tsig$rejectNum[a] <- 10

  tsig$reject[a <- (P < 0.05)] <- '*'
  tsig$rejectNum[a] <- 5

  tsig$reject[a <- (P < 0.01)] <- '**'
  tsig$rejectNum[a] <- 1

  tsig$reject[a <- (P < 0.001)] <- '***'
  tsig$rejectNum[a] <- 0.1

  class(tsig) = "sigtest"

  tsig
}

print.sigtest <- function(x, ...){
  cat("\nKernel Regression Significance Test",
      "\nType ", x$ptype,
      " Test with ", x$pmethod,
      " Bootstrap (",x$boot.num," replications,",
      " Pivot = ", x$pivot,", joint = ",x$joint,")",
      "\nExplanatory variables tested for significance:\n",
      paste(paste(x$bws$xnames[x$ixvar]," (",x$ixvar,")", sep=""), collapse=", "),"\n\n",
      sep="")
      

  if (!identical(x$bws,NA))
    print(matrix(x$bws$bw,ncol=x$bws$ndim,dimnames=list(paste(x$bws$pscaling,":",sep=""),
                                            x$bws$xnames)))

  maxNameLen <- max(nc <- nchar(nm <- x$bws$xnames[x$ixvar]))

  if(!x$joint) {
    cat("\nIndividual Significance Tests\n")
    cat("P Value:", paste("\n", nm," ", blank(maxNameLen-nc), format.pval(x$P),
                          " ", x$reject, sep=''))
  } else {
    cat("\nJoint Significance Test\n")
    cat("P Value:", paste(" ", format.pval(x$P)," ", x$reject, sep=''))
  }
  cat("\n---\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1\n\n")
}

summary.sigtest <- function(object, ...) {
  print(object)
}

#!/usr/bin/env Rscript

parse_args <- function(args) {
  cfg <- list(
    repo = getwd(),
    baseline = NULL,
    tests = character(),
    out_dir = tempfile('testthat_delta_gate_')
  )

  if (length(args) == 0L) return(cfg)

  for (a in args) {
    if (!startsWith(a, '--')) stop('Bad arg: ', a)
    key <- sub('^--([^=]+)=?.*$', '\\1', a)
    has_eq <- grepl('=', a, fixed = TRUE)
    val <- if (has_eq) sub('^--[^=]+=','', a) else 'TRUE'

    if (key == 'repo') cfg$repo <- val
    else if (key == 'baseline') cfg$baseline <- val
    else if (key == 'tests') cfg$tests <- strsplit(val, ',', fixed = TRUE)[[1L]]
    else if (key == 'out_dir') cfg$out_dir <- val
    else stop('Unknown arg: ', key)
  }

  cfg$repo <- normalizePath(cfg$repo, winslash = '/', mustWork = TRUE)
  cfg$out_dir <- normalizePath(cfg$out_dir, winslash = '/', mustWork = FALSE)
  cfg$tests <- trimws(cfg$tests)
  cfg$tests <- cfg$tests[nzchar(cfg$tests)]
  cfg
}

resolve_tests <- function(repo, tests_arg) {
  test_dir <- file.path(repo, 'tests', 'testthat')
  if (!dir.exists(test_dir)) stop('Missing testthat dir: ', test_dir)

  all_files <- list.files(test_dir, pattern = '^test-.*\\.R$', full.names = TRUE)
  if (length(tests_arg) == 0L) return(all_files)

  out <- character()
  for (t in tests_arg) {
    if (file.exists(t)) {
      out <- c(out, normalizePath(t, winslash = '/', mustWork = TRUE))
      next
    }
    cand <- file.path(test_dir, t)
    if (file.exists(cand)) {
      out <- c(out, normalizePath(cand, winslash = '/', mustWork = TRUE))
      next
    }
    stop('Unable to resolve test file: ', t)
  }
  unique(out)
}

is_mpi_sensitive_test <- function(path) {
  txt <- tryCatch(readLines(path, warn = FALSE), error = function(e) character())
  if (!length(txt)) return(FALSE)
  any(grepl("spawn_mpi_slaves\\(|npRmpi\\.init\\(|mpi\\.", txt))
}

read_baseline <- function(path) {
  if (is.null(path) || !nzchar(path) || !file.exists(path)) {
    return(data.frame(
      file = character(),
      allow_failures = integer(),
      allow_errors = integer(),
      stringsAsFactors = FALSE
    ))
  }

  b <- read.csv(path, stringsAsFactors = FALSE)
  req <- c('file', 'allow_failures', 'allow_errors')
  miss <- setdiff(req, names(b))
  if (length(miss)) stop('Baseline file missing columns: ', paste(miss, collapse = ','))
  b$file <- basename(trimws(b$file))
  b$allow_failures <- as.integer(b$allow_failures)
  b$allow_errors <- as.integer(b$allow_errors)
  b[req]
}

count_expectations <- function(results) {
  failures <- 0L
  errors <- 0L
  warnings <- 0L
  skips <- 0L

  for (tr in results) {
    rr <- tr$results
    if (is.null(rr) || length(rr) == 0L) next
    failures <- failures + sum(vapply(rr, function(x) inherits(x, 'expectation_failure'), logical(1)))
    errors <- errors + sum(vapply(rr, function(x) inherits(x, 'expectation_error'), logical(1)))
    warnings <- warnings + sum(vapply(rr, function(x) inherits(x, 'expectation_warning'), logical(1)))
    skips <- skips + sum(vapply(rr, function(x) inherits(x, 'expectation_skip'), logical(1)))
  }

  list(failures = failures, errors = errors, warnings = warnings, skips = skips)
}

run_one <- function(path) {
  r <- testthat::ListReporter$new()
  run_error <- ''

  tryCatch(
    testthat::test_file(path, reporter = r),
    error = function(e) {
      run_error <<- conditionMessage(e)
    }
  )

  counts <- count_expectations(r$get_results())
  data.frame(
    file = basename(path),
    path = path,
    failures = counts$failures,
    errors = counts$errors,
    warnings = counts$warnings,
    skips = counts$skips,
    run_error = run_error,
    stringsAsFactors = FALSE
  )
}

main <- function(args = commandArgs(trailingOnly = TRUE)) {
  cfg <- parse_args(args)
  dir.create(cfg$out_dir, recursive = TRUE, showWarnings = FALSE)

  baseline <- read_baseline(cfg$baseline)
  tests <- resolve_tests(cfg$repo, cfg$tests)

  if (basename(cfg$repo) == "np-npRmpi" &&
      !identical(toupper(Sys.getenv("NP_RMPI_ALLOW_INPROC_TESTS", "FALSE")), "TRUE")) {
    sens <- tests[vapply(tests, is_mpi_sensitive_test, logical(1))]
    if (length(sens) > 0L) {
      cat("Refusing in-process testthat run for MPI-sensitive files in npRmpi:\n")
      cat(paste0(" - ", basename(sens), collapse = "\n"), "\n")
      stop("Use session/attach/manual route validators (or set NP_RMPI_ALLOW_INPROC_TESTS=TRUE explicitly).")
    }
  }

  suppressPackageStartupMessages(library(pkgload))
  suppressPackageStartupMessages(library(testthat))
  old_skip <- Sys.getenv("NP_RMPI_SKIP_INIT", unset = NA_character_)
  Sys.setenv(NP_RMPI_SKIP_INIT = "1")
  pkgload::load_all(cfg$repo, quiet = TRUE, export_all = FALSE)
  if (is.na(old_skip)) {
    Sys.unsetenv("NP_RMPI_SKIP_INIT")
  } else {
    Sys.setenv(NP_RMPI_SKIP_INIT = old_skip)
  }

  rows <- lapply(tests, run_one)
  res <- do.call(rbind, rows)

  if (nrow(baseline) > 0L) {
    res <- merge(res, baseline, by = 'file', all.x = TRUE)
  } else {
    res$allow_failures <- NA_integer_
    res$allow_errors <- NA_integer_
  }
  res$allow_failures[is.na(res$allow_failures)] <- 0L
  res$allow_errors[is.na(res$allow_errors)] <- 0L

  res$violate_failures <- res$failures > res$allow_failures
  res$violate_errors <- res$errors > res$allow_errors
  res$violate_run_error <- nzchar(res$run_error)
  res$status <- ifelse(
    res$violate_failures | res$violate_errors | res$violate_run_error,
    'FAIL',
    'PASS'
  )

  summary_path <- file.path(cfg$out_dir, 'testthat_delta_summary.csv')
  write.csv(res[order(res$file), ], summary_path, row.names = FALSE)

  cat('repo:', cfg$repo, '\n')
  cat('baseline:', if (is.null(cfg$baseline)) '<none>' else cfg$baseline, '\n')
  cat('summary:', summary_path, '\n')
  cat('status table:\n')
  print(table(res$status, useNA = 'ifany'))

  bad <- res[res$status == 'FAIL', c('file', 'failures', 'allow_failures', 'errors', 'allow_errors', 'run_error')]
  if (nrow(bad) > 0L) {
    cat('\nviolations:\n')
    print(bad, row.names = FALSE)
    stop('testthat delta gate failed')
  }

  cat('TESTTHAT_DELTA_GATE_OK\n')
}

main()

npregbw <-
  function(...){
    mc <- match.call(expand.dots = FALSE)
    npRejectRenamedScaleFactorSearchArgs(names(mc$...), where = "npregbw")
    target <- .np_bw_dispatch_target(dots = mc$...,
                                     data_arg_names = c("xdat", "ydat"),
                                     eval_env = parent.frame())
    UseMethod("npregbw", target)
  }

npregbw.formula <-
  function(formula, data, subset, na.action, call, ...){

    formula.terms <- terms(formula)

    orig.ts <- if (missing(data))
      .np_terms_ts_mask(terms_obj = formula.terms,
                        data = environment(formula),
                        eval_env = environment(formula))
    else .np_terms_ts_mask(terms_obj = formula.terms,
                           data = data,
                           eval_env = environment(formula))

    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(mf), nomatch = 0)
    mf <- mf[c(1,m)]

    if(all(orig.ts)){
      args <- (as.list(attr(formula.terms, "variables"))[-1])
      formula <- formula.terms
      attr(formula, "predvars") <- as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), args))))
      mf[["formula"]] <- formula
    }else if(any(orig.ts)){
      arguments <- (as.list(attr(formula.terms, "variables"))[-1])
      arguments.normal <- arguments[which(!orig.ts)]
      arguments.timeseries <- arguments[which(orig.ts)]

      ix <- sort(c(which(orig.ts),which(!orig.ts)),index.return = TRUE)$ix
      formula <- formula.terms
      attr(formula, "predvars") <- bquote(.(as.call(c(quote(cbind),as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments.timeseries)))),arguments.normal,check.rows = TRUE)))[,.(ix)])
      mf[["formula"]] <- formula
    }
      
    mf[[1]] <- as.name("model.frame")
    mf.args <- as.list(mf[-1L])
    mf <- do.call(stats::model.frame, mf.args, envir = parent.frame())

    ydat <- model.response(mf)
    xdat <- mf[, attr(attr(mf, "terms"),"term.labels"), drop = FALSE]
    
    tbw <- do.call(npregbw, c(list(xdat = xdat, ydat = ydat), list(...)))

    ## clean up (possible) inconsistencies due to recursion ...
    tbw$call <- match.call(expand.dots = FALSE)
    environment(tbw$call) <- parent.frame()
    tbw$formula <- formula
    tbw$rows.omit <- as.vector(attr(mf,"na.action"))
    tbw$nobs.omit <- length(tbw$rows.omit)
    tbw$terms <- attr(mf,"terms")

    tbw <-
      updateBwNameMetadata(nameList =
                           list(ynames =
                                attr(mf, "names")[attr(tbw$terms, "response")]),
                           bws = tbw)
    
    tbw
  }

npregbw.NULL <-
  function(xdat = stop("invoked without data 'xdat'"),
           ydat = stop("invoked without data 'ydat'"),
           bws, ...){

    xdat <- toFrame(xdat)

    bws = double(dim(xdat)[2])
    
    tbw <- npregbw.default(xdat = xdat, ydat = ydat, bws = bws, ...)

    ## clean up (possible) inconsistencies due to recursion ...
    mc <- match.call(expand.dots = FALSE)
    environment(mc) <- parent.frame()
    tbw$call <- mc

    tbw <- updateBwNameMetadata(nameList =
                                list(ynames = deparse(substitute(ydat))),
                                bws = tbw)
    
    tbw
  }

npregbw.rbandwidth <- 
  function(xdat = stop("invoked without data 'xdat'"),
           ydat = stop("invoked without data 'ydat'"),
           bws,
           bandwidth.compute = TRUE,
           cfac.dir = 2.5*(3.0-sqrt(5)),
           scale.factor.init = 0.5,
           dfac.dir = 0.25*(3.0-sqrt(5)),
           dfac.init = 0.375,
           dfc.dir = 3,
           ftol = 1.490116e-07,
           scale.factor.init.upper = 2.0,
           hbd.dir = 1,
           hbd.init = 0.9,
           initc.dir = 1.0,
           initd.dir = 1.0,
           invalid.penalty = c("baseline","dbmax"),
           itmax = 10000,
           lbc.dir = 0.5,
           scale.factor.init.lower = 0.1,
           lbd.dir = 0.1,
           lbd.init = 0.1,
           nmulti,
           penalty.multiplier = 10,
           powell.remin = TRUE,
           scale.init.categorical.sample = FALSE,
           scale.factor.search.lower = NULL,
           small = 1.490116e-05,
           tol = 1.490116e-04,
           transform.bounds = FALSE,
           ...){
    elapsed.start <- proc.time()[3]

    xdat <- toFrame(xdat)

    if (missing(nmulti)){
      nmulti <- npDefaultNmulti(dim(xdat)[2])
    }
    bandwidth.compute <- npValidateScalarLogical(bandwidth.compute, "bandwidth.compute")
    remin <- npValidateScalarLogical(powell.remin, "powell.remin")
    scale.init.categorical.sample <-
      npValidateScalarLogical(scale.init.categorical.sample, "scale.init.categorical.sample")
    transform.bounds <- npValidateScalarLogical(transform.bounds, "transform.bounds")
    itmax <- npValidatePositiveInteger(itmax, "itmax")
    ftol <- npValidatePositiveFiniteNumeric(ftol, "ftol")
    tol <- npValidatePositiveFiniteNumeric(tol, "tol")
    small <- npValidatePositiveFiniteNumeric(small, "small")
    penalty.multiplier <- npValidatePositiveFiniteNumeric(penalty.multiplier, "penalty.multiplier")
    scale.factor.search.lower <- npResolveScaleFactorLowerBound(
      if (is.null(scale.factor.search.lower)) npGetScaleFactorSearchLower(bws) else scale.factor.search.lower
    )
    nmulti <- npValidateNmulti(nmulti)
    .np_progress_bandwidth_set_total(nmulti)

    if (!(is.vector(ydat) || is.factor(ydat)))
      stop("'ydat' must be a vector")

    if (length(bws$bw) != dim(xdat)[2])
      stop("length of bandwidth vector does not match number of columns of 'xdat'")

    npValidateRegressionNnLowerBound(
      bws,
      where = "npregbw",
      allow.zero.placeholder = TRUE
    )

    if ((any(bws$icon) &&
         !all(vapply(xdat[, bws$icon, drop = FALSE], inherits, logical(1), c("integer", "numeric")))) ||
        (any(bws$iord) &&
         !all(vapply(xdat[, bws$iord, drop = FALSE], inherits, logical(1), "ordered"))) ||
        (any(bws$iuno) &&
         !all(vapply(xdat[, bws$iuno, drop = FALSE], inherits, logical(1), "factor"))))
      stop("supplied bandwidths do not match 'xdat' in type")

    if (dim(xdat)[1] != length(ydat))
      stop("number of regression data and response data do not match")

    ## catch and destroy NA's
    keep.rows <- rep_len(TRUE, nrow(xdat))
    rows.omit <- attr(na.omit(data.frame(xdat,ydat)), "na.action")
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Data has no rows without NAs")

    xdat <- xdat[keep.rows,,drop = FALSE]
    ydat <- ydat[keep.rows]
    
    nrow = dim(xdat)[1]
    ncol = dim(xdat)[2]

    ## at this stage, data to be sent to the c routines must be converted to
    ## numeric type.

    if (is.factor(ydat))
      ydat <- dlev(ydat)[as.integer(ydat)]
    else
      ydat <- as.double(ydat)

    xdat = toMatrix(xdat)

    runo = xdat[, bws$iuno, drop = FALSE]
    rcon = xdat[, bws$icon, drop = FALSE]
    rord = xdat[, bws$iord, drop = FALSE]

    tbw <- bws
    tbw$basis <- npValidateLpBasis(regtype = tbw$regtype,
                                   basis = tbw$basis)
    tbw$degree <- npValidateGlpDegree(regtype = tbw$regtype,
                                          degree = tbw$degree,
                                          ncon = tbw$ncon)
    tbw$bernstein.basis <- npValidateGlpBernstein(regtype = tbw$regtype,
                                                bernstein.basis = tbw$bernstein.basis)
    reg.spec <- npCanonicalConditionalRegSpec(
      regtype = tbw$regtype,
      basis = tbw$basis,
      degree = tbw$degree,
      bernstein.basis = tbw$bernstein.basis,
      ncon = tbw$ncon,
      where = "npregbw"
    )

    mysd <- EssDee(rcon)
    nconfac <- nrow^(-1.0/(2.0*bws$ckerorder+bws$ncon))
    ncatfac <- nrow^(-2.0/(2.0*bws$ckerorder+bws$ncon))

    invalid.penalty <- match.arg(invalid.penalty)
    penalty_mode <- (if (invalid.penalty == "baseline") 1L else 0L)

    reg.c <- npRegtypeToC(regtype = reg.spec$regtype.engine,
                          degree = reg.spec$degree.engine,
                          ncon = tbw$ncon,
                          context = "npregbw")
    npCheckRegressionDesignCondition(reg.code = reg.c$code,
                                     xcon = rcon,
                                     basis = reg.spec$basis.engine,
                                     degree = reg.spec$degree.engine,
                                     bernstein.basis = reg.spec$bernstein.basis.engine,
                                     where = "npregbw")
    degree.c <- if (tbw$ncon > 0) {
      as.integer(if (is.null(reg.c$degree)) rep.int(0L, tbw$ncon) else reg.c$degree)
    } else {
      integer(1)
    }

    if (bandwidth.compute){
      cont.start <- npContinuousSearchStartControls(
        scale.factor.init.lower,
        scale.factor.init.upper,
        scale.factor.init,
        scale.factor.search.lower,
        where = "npregbw"
      )
      myopti = list(num_obs_train = dim(xdat)[1], 
        iMultistart = IMULTI_TRUE,
        iNum_Multistart = nmulti,
        int_use_starting_values = (if (all(bws$bw==0)) USE_START_NO else USE_START_YES),
        int_LARGE_SF = (if (bws$scaling) SF_NORMAL else SF_ARB),
        BANDWIDTH_reg_extern = switch(bws$type,
          fixed = BW_FIXED,
          generalized_nn = BW_GEN_NN,
          adaptive_nn = BW_ADAP_NN),
        itmax=itmax, int_RESTART_FROM_MIN=(if (remin) RE_MIN_TRUE else RE_MIN_FALSE), 
        int_MINIMIZE_IO=if (isTRUE(getOption("np.messages"))) IO_MIN_FALSE else IO_MIN_TRUE, 
        bwmethod = switch(bws$method,
          cv.aic = BWM_CVAIC,
          cv.ls = BWM_CVLS),
        kerneval = switch(bws$ckertype,
          gaussian = CKER_GAUSS + bws$ckerorder/2 - 1,
          epanechnikov = CKER_EPAN + bws$ckerorder/2 - 1,
          uniform = CKER_UNI,
          "truncated gaussian" = CKER_TGAUSS),
        ukerneval = switch(bws$ukertype,
          aitchisonaitken = UKER_AIT,
          liracine = UKER_LR),
        okerneval = switch(bws$okertype,
          wangvanryzin = OKER_WANG,
          liracine = OKER_LR,
        "racineliyan" = OKER_RLY),
        nuno = bws$nuno,
        nord = bws$nord,
        ncon = bws$ncon,
        regtype = reg.c$code,
        int_do_tree = if (isTRUE(getOption("np.tree"))) DO_TREE_YES else DO_TREE_NO,
        scale.init.categorical.sample = scale.init.categorical.sample,
        dfc.dir = dfc.dir,
        transform.bounds = transform.bounds)
      
      myoptd = list(ftol=ftol, tol=tol, small=small,
        lbc.dir = lbc.dir, cfac.dir = cfac.dir, initc.dir = initc.dir, 
        lbd.dir = lbd.dir, hbd.dir = hbd.dir, dfac.dir = dfac.dir, initd.dir = initd.dir, 
        lbc.init = cont.start$scale.factor.init.lower, hbc.init = cont.start$scale.factor.init.upper, cfac.init = cont.start$scale.factor.init, 
        lbd.init = lbd.init, hbd.init = hbd.init, dfac.init = dfac.init, 
        nconfac = nconfac, ncatfac = ncatfac,
        scale.factor.lower.bound = scale.factor.search.lower)

        cker.bounds.c <- npKernelBoundsMarshal(bws$ckerlb[bws$icon], bws$ckerub[bws$icon])

        system.time(myout <-
          .Call("C_np_regression_bw",
                as.double(runo), as.double(rord), as.double(rcon), as.double(ydat),
                as.double(mysd),
                as.integer(myopti), as.double(myoptd),
                as.double(c(bws$bw[bws$icon], bws$bw[bws$iuno], bws$bw[bws$iord])),
                as.integer(nmulti),
                as.integer(penalty_mode),
                as.double(penalty.multiplier),
                as.integer(degree.c),
                as.integer(isTRUE(reg.spec$bernstein.basis.engine)),
                as.integer(npLpBasisCode(reg.spec$basis.engine)),
                as.double(cker.bounds.c$lb),
                as.double(cker.bounds.c$ub),
                PACKAGE = "np"))[1]
      

      rorder = numeric(ncol)
      ord_idx <- seq_len(ncol)
      rorder[c(ord_idx[bws$icon], ord_idx[bws$iuno], ord_idx[bws$iord])] <- ord_idx

      tbw$bw <- myout$bw[rorder]
      tbw$fval <- myout$fval[1]
      tbw$ifval <- myout$fval[2]
      tbw$num.feval <- sum(myout$eval.history[is.finite(myout$eval.history)])
      tbw$num.feval.fast <- myout$fast.history[1]
      tbw$fval.history <- myout$fval.history
      tbw$eval.history <- myout$eval.history
      tbw$invalid.history <- myout$invalid.history
      tbw$timing <- myout$timing
    }

    tbw$total.time <- proc.time()[3] - elapsed.start

    tbw$sfactor <- tbw$bandwidth <- tbw$bw

    if (tbw$nuno > 0){
      if(tbw$scaling){ 
        tbw$bandwidth[tbw$xdati$iuno] <- tbw$bandwidth[tbw$xdati$iuno]*ncatfac
      } else {
        tbw$sfactor[tbw$xdati$iuno] <- tbw$sfactor[tbw$xdati$iuno]/ncatfac
      }
    }
    
    if (tbw$nord > 0){
      if(tbw$scaling){
        tbw$bandwidth[tbw$xdati$iord] <- tbw$bandwidth[tbw$xdati$iord]*ncatfac
      } else {
        tbw$sfactor[tbw$xdati$iord] <- tbw$sfactor[tbw$xdati$iord]/ncatfac
      }
    }

    if (tbw$ncon > 0){
      dfactor <- mysd*nconfac

      if (tbw$scaling) {
        tbw$bandwidth[tbw$xdati$icon] <- tbw$bandwidth[tbw$xdati$icon]*dfactor
      } else {
        tbw$sfactor[tbw$xdati$icon] <- tbw$sfactor[tbw$xdati$icon]/dfactor
      }
    }
    


    tbw <- rbandwidth(bw = tbw$bw,
                      regtype = tbw$regtype,
                      basis = tbw$basis,
                      degree = tbw$degree,
                      bernstein.basis = tbw$bernstein.basis,
                      bwmethod = tbw$method,
                      bwscaling = tbw$scaling,
                      bwtype = tbw$type,
                      ckertype = tbw$ckertype,
                      ckerorder = tbw$ckerorder,
                      ckerbound = tbw$ckerbound,
                      ckerlb = tbw$ckerlb,
                      ckerub = tbw$ckerub,
                      ukertype = tbw$ukertype,
                      okertype = tbw$okertype,
                      fval = tbw$fval,
                      ifval = tbw$ifval,
                      num.feval = tbw$num.feval,
                      num.feval.fast = tbw$num.feval.fast,
                      fval.history = tbw$fval.history,
                      eval.history = tbw$eval.history,
                      invalid.history = tbw$invalid.history,
                      nobs = tbw$nobs,
                      xdati = tbw$xdati,
                      ydati = tbw$ydati,
                      xnames = tbw$xnames,
                      ynames = tbw$ynames,
                      sfactor = tbw$sfactor,
                      bandwidth = tbw$bandwidth,
                      rows.omit = rows.omit,
                      nconfac = nconfac,
                      ncatfac = ncatfac,
                      sdev = mysd,
                      bandwidth.compute = bandwidth.compute,
                      timing = tbw$timing,
                      total.time = tbw$total.time)
    tbw <- npSetScaleFactorSearchLower(tbw, scale.factor.search.lower)
    tbw
  }

.npregbw_build_rbandwidth <- function(xdat,
                                      ydat,
                                      bws,
                                      bandwidth.compute,
                                      reg.args,
                                      yname) {
  bw.args <- c(
    list(
      bw = bws,
      nobs = dim(xdat)[1],
      xdati = untangle(xdat),
      ydati = untangle(data.frame(ydat)),
      xnames = names(xdat),
      ynames = yname,
      bandwidth.compute = bandwidth.compute
    ),
    reg.args
  )
  out <- do.call(rbandwidth, bw.args)
  if (!is.null(reg.args$scale.factor.search.lower)) {
    out$scale.factor.search.lower <- npResolveScaleFactorLowerBound(
      reg.args$scale.factor.search.lower
    )
  }
  out
}

.npregbw_eval_only <- function(xdat,
                               ydat,
                               bws,
                               invalid.penalty = c("baseline", "dbmax"),
                               penalty.multiplier = 10) {
  out <- .npregbw_call_fixed_degree_core(
    xdat = xdat,
    ydat = ydat,
    bws = bws,
    invalid.penalty = invalid.penalty,
    penalty.multiplier = penalty.multiplier,
    eval.only = TRUE
  )

  list(
    objective = out$objective,
    num.feval = out$num.feval,
    num.feval.fast = out$num.feval.fast
  )
}

.npregbw_call_fixed_degree_core <- function(xdat,
                                            ydat,
                                            bws,
                                            nmulti = 1L,
                                            itmax = 0L,
                                            remin = FALSE,
                                            scale.init.categorical.sample = FALSE,
                                            ftol = 0,
                                            tol = 0,
                                            small = 0,
                                            lbc.dir = 0,
                                            dfc.dir = 0,
                                            cfac.dir = 0,
                                            initc.dir = 0,
                                            lbd.dir = 0,
                                            hbd.dir = 0,
                                            dfac.dir = 0,
                                            initd.dir = 0,
                                            scale.factor.init.lower = 0,
                                            scale.factor.init.upper = 0,
                                            scale.factor.init = 0,
                                            lbd.init = 0,
                                            hbd.init = 0,
                                            dfac.init = 0,
                                            invalid.penalty = c("baseline", "dbmax"),
                                            penalty.multiplier = 10,
                                            transform.bounds = FALSE,
                                            scale.factor.search.lower = NULL,
                                            eval.only = FALSE) {
  invalid.penalty <- match.arg(invalid.penalty)
  scale.factor.search.lower <- npResolveScaleFactorLowerBound(
    if (is.null(scale.factor.search.lower)) npGetScaleFactorSearchLower(bws) else scale.factor.search.lower
  )
  if (!isTRUE(eval.only)) {
    cont.start <- npContinuousSearchStartControls(
      scale.factor.init.lower,
      scale.factor.init.upper,
      scale.factor.init,
      scale.factor.search.lower,
      where = "npregbw"
    )
    scale.factor.init.lower <- cont.start$scale.factor.init.lower
    scale.factor.init.upper <- cont.start$scale.factor.init.upper
    scale.factor.init <- cont.start$scale.factor.init
  }

  xdat <- toFrame(xdat)
  if (!(is.vector(ydat) || is.factor(ydat)))
    stop("'ydat' must be a vector")

  if (length(bws$bw) != dim(xdat)[2])
    stop("length of bandwidth vector does not match number of columns of 'xdat'")

  keep.rows <- rep_len(TRUE, nrow(xdat))
  rows.omit <- attr(na.omit(data.frame(xdat, ydat)), "na.action")
  if (length(rows.omit) > 0L)
    keep.rows[as.integer(rows.omit)] <- FALSE

  xdat <- xdat[keep.rows,, drop = FALSE]
  ydat <- ydat[keep.rows]

  if (is.factor(ydat))
    ydat <- dlev(ydat)[as.integer(ydat)]
  else
    ydat <- as.double(ydat)

  xmat <- toMatrix(xdat)
  runo <- xmat[, bws$iuno, drop = FALSE]
  rcon <- xmat[, bws$icon, drop = FALSE]
  rord <- xmat[, bws$iord, drop = FALSE]

  mysd <- EssDee(rcon)
  nrow <- dim(xmat)[1L]
  nconfac <- nrow^(-1.0 / (2.0 * bws$ckerorder + bws$ncon))
  ncatfac <- nrow^(-2.0 / (2.0 * bws$ckerorder + bws$ncon))

  penalty_mode <- if (invalid.penalty == "baseline") 1L else 0L
  reg.spec <- npCanonicalConditionalRegSpec(
    regtype = bws$regtype,
    basis = bws$basis,
    degree = bws$degree,
    bernstein.basis = bws$bernstein.basis,
    ncon = bws$ncon,
    where = "npregbw"
  )
  reg.c <- npRegtypeToC(regtype = reg.spec$regtype.engine,
                        degree = reg.spec$degree.engine,
                        ncon = bws$ncon,
                        context = "npregbw")
  degree.c <- if (bws$ncon > 0) as.integer(reg.spec$degree.engine) else integer(1L)
  nmulti <- as.integer(nmulti[1L])

  myopti <- list(
    num_obs_train = nrow,
    iMultistart = IMULTI_TRUE,
    iNum_Multistart = nmulti,
    int_use_starting_values = if (all(bws$bw == 0)) USE_START_NO else USE_START_YES,
    int_LARGE_SF = if (bws$scaling) SF_NORMAL else SF_ARB,
    BANDWIDTH_reg_extern = switch(bws$type,
      fixed = BW_FIXED,
      generalized_nn = BW_GEN_NN,
      adaptive_nn = BW_ADAP_NN),
    itmax = itmax,
    int_RESTART_FROM_MIN = if (isTRUE(remin)) RE_MIN_TRUE else RE_MIN_FALSE,
    int_MINIMIZE_IO = if (isTRUE(eval.only) || !isTRUE(getOption("np.messages"))) IO_MIN_TRUE else IO_MIN_FALSE,
    bwmethod = switch(bws$method,
      cv.aic = BWM_CVAIC,
      cv.ls = BWM_CVLS),
    kerneval = switch(bws$ckertype,
      gaussian = CKER_GAUSS + bws$ckerorder/2 - 1,
      epanechnikov = CKER_EPAN + bws$ckerorder/2 - 1,
      uniform = CKER_UNI,
      "truncated gaussian" = CKER_TGAUSS),
    ukerneval = switch(bws$ukertype,
      aitchisonaitken = UKER_AIT,
      liracine = UKER_LR),
    okerneval = switch(bws$okertype,
      wangvanryzin = OKER_WANG,
      liracine = OKER_LR,
      racineliyan = OKER_RLY),
    nuno = bws$nuno,
    nord = bws$nord,
    ncon = bws$ncon,
    regtype = reg.c$code,
    int_do_tree = if (isTRUE(getOption("np.tree"))) DO_TREE_YES else DO_TREE_NO,
    scale.init.categorical.sample = scale.init.categorical.sample,
    dfc.dir = dfc.dir,
    transform.bounds = transform.bounds
  )

  myoptd <- list(
    ftol = ftol,
    tol = tol,
    small = small,
    lbc.dir = lbc.dir,
    cfac.dir = cfac.dir,
    initc.dir = initc.dir,
    lbd.dir = lbd.dir,
    hbd.dir = hbd.dir,
    dfac.dir = dfac.dir,
    initd.dir = initd.dir,
    lbc.init = scale.factor.init.lower,
    hbc.init = scale.factor.init.upper,
    cfac.init = scale.factor.init,
    lbd.init = lbd.init,
    hbd.init = hbd.init,
    dfac.init = dfac.init,
    nconfac = nconfac,
    ncatfac = ncatfac,
    scale.factor.lower.bound = scale.factor.search.lower
  )

  cker.bounds.c <- npKernelBoundsMarshal(bws$ckerlb[bws$icon], bws$ckerub[bws$icon])

  if (isTRUE(eval.only)) {
    out <- .Call(
      C_np_regression_bw_eval,
      as.double(runo),
      as.double(rord),
      as.double(rcon),
      as.double(ydat),
      as.double(mysd),
      as.integer(myopti),
      as.double(myoptd),
      as.double(c(bws$bw[bws$icon], bws$bw[bws$iuno], bws$bw[bws$iord])),
      as.integer(1L),
      as.integer(penalty_mode),
      as.double(penalty.multiplier),
      as.integer(degree.c),
      as.integer(isTRUE(reg.spec$bernstein.basis.engine)),
      as.integer(npLpBasisCode(reg.spec$basis.engine)),
      as.double(cker.bounds.c$lb),
      as.double(cker.bounds.c$ub),
      PACKAGE = "np"
    )
  } else {
    out <- .Call(
      C_np_regression_bw,
      as.double(runo),
      as.double(rord),
      as.double(rcon),
      as.double(ydat),
      as.double(mysd),
      as.integer(myopti),
      as.double(myoptd),
      as.double(c(bws$bw[bws$icon], bws$bw[bws$iuno], bws$bw[bws$iord])),
      as.integer(max(1L, nmulti)),
      as.integer(penalty_mode),
      as.double(penalty.multiplier),
      as.integer(degree.c),
      as.integer(isTRUE(reg.spec$bernstein.basis.engine)),
      as.integer(npLpBasisCode(reg.spec$basis.engine)),
      as.double(cker.bounds.c$lb),
      as.double(cker.bounds.c$ub),
      PACKAGE = "np"
    )
  }

  rorder <- numeric(length(bws$bw))
  ord.idx <- seq_along(bws$bw)
  rorder[c(ord.idx[bws$icon], ord.idx[bws$iuno], ord.idx[bws$iord])] <- ord.idx

  list(
    objective = as.numeric(out$fval[1L]),
    ifval = as.numeric(out$fval[2L]),
    bw = as.numeric(out$bw[rorder]),
    num.feval = sum(out$eval.history[is.finite(out$eval.history)]),
    num.feval.fast = as.numeric(out$fast.history[1L]),
    fval.history = out$fval.history,
    eval.history = out$eval.history,
    invalid.history = out$invalid.history,
    timing = out$timing
  )
}

.npregbw_finalize_fixed_degree_payload <- function(xdat,
                                                   ydat,
                                                   bws,
                                                   core,
                                                   total.time) {
  tbw <- bws
  tbw$bw <- core$bw
  tbw$fval <- core$objective
  tbw$ifval <- core$ifval
  tbw$num.feval <- core$num.feval
  tbw$num.feval.fast <- core$num.feval.fast
  tbw$fval.history <- core$fval.history
  tbw$eval.history <- core$eval.history
  tbw$invalid.history <- core$invalid.history
  tbw$timing <- core$timing

  payload <- npregbw.rbandwidth(
    xdat = xdat,
    ydat = ydat,
    bws = tbw,
    bandwidth.compute = FALSE
  )
  if (!is.null(payload$method) && length(payload$method))
    payload$pmethod <- bwmToPrint(as.character(payload$method[1L]))
  payload$timing <- core$timing
  payload$total.time <- total.time
  payload
}

.npregbw_run_fixed_degree_source_of_truth <- function(xdat,
                                                      ydat,
                                                      bws,
                                                      opt.args) {
  opt.value <- function(name, default) {
    if (is.null(opt.args[[name]])) default else opt.args[[name]]
  }

  elapsed.start <- proc.time()[3L]
  core <- .npregbw_call_fixed_degree_core(
    xdat = xdat,
    ydat = ydat,
    bws = bws,
    nmulti = opt.value("nmulti", npDefaultNmulti(dim(toFrame(xdat))[2L])),
    itmax = opt.value("itmax", 10000L),
    remin = opt.value("powell.remin", TRUE),
    scale.init.categorical.sample = opt.value("scale.init.categorical.sample", FALSE),
    ftol = opt.value("ftol", 1.490116e-07),
    tol = opt.value("tol", 1.490116e-04),
    small = opt.value("small", 1.490116e-05),
    lbc.dir = opt.value("lbc.dir", 0.5),
    dfc.dir = opt.value("dfc.dir", 3),
    cfac.dir = opt.value("cfac.dir", 2.5 * (3.0 - sqrt(5))),
    initc.dir = opt.value("initc.dir", 1.0),
    lbd.dir = opt.value("lbd.dir", 0.1),
    hbd.dir = opt.value("hbd.dir", 1),
    dfac.dir = opt.value("dfac.dir", 0.25 * (3.0 - sqrt(5))),
    initd.dir = opt.value("initd.dir", 1.0),
    scale.factor.init.lower = opt.value("scale.factor.init.lower", 0.1),
    scale.factor.init.upper = opt.value("scale.factor.init.upper", 2.0),
    scale.factor.init = opt.value("scale.factor.init", 0.5),
    lbd.init = opt.value("lbd.init", 0.1),
    hbd.init = opt.value("hbd.init", 0.9),
    dfac.init = opt.value("dfac.init", 0.375),
    invalid.penalty = opt.value("invalid.penalty", "baseline"),
    penalty.multiplier = opt.value("penalty.multiplier", 10),
    transform.bounds = opt.value("transform.bounds", FALSE),
    scale.factor.search.lower = opt.value("scale.factor.search.lower", NULL),
    eval.only = FALSE
  )

  .npregbw_finalize_fixed_degree_payload(
    xdat = xdat,
    ydat = ydat,
    bws = bws,
    core = core,
    total.time = proc.time()[3L] - elapsed.start
  )
}

.npregbw_nomad_controls <- function(search.engine) {
  .np_degree_search_engine_controls(search.engine)
}

.npregbw_nomad_bw_setup <- function(xdat, template, bandwidth.scale.categorical = 1e4) {
  xdat <- toFrame(xdat)
  xmat <- toMatrix(xdat)
  rcon <- xmat[, template$icon, drop = FALSE]
  mysd <- EssDee(rcon)
  nrow <- dim(xmat)[1L]
  nconfac <- nrow^(-1.0 / (2.0 * template$ckerorder + template$ncon))
  ncatfac <- nrow^(-2.0 / (2.0 * template$ckerorder + template$ncon))

  cont_scale <- mysd * nconfac
  cont_idx <- which(template$icon)
  cat_idx <- c(which(template$iuno), which(template$iord))

  cat_upper <- numeric(length(cat_idx))
  for (k in seq_along(cat_idx)) {
    i <- cat_idx[k]
    if (template$iuno[i] && identical(template$ukertype, "aitchisonaitken")) {
      nlev <- length(unique(xdat[[i]]))
      cat_upper[k] <- (nlev - 1) / nlev
    } else {
      cat_upper[k] <- 1
    }
  }

  list(
    cont_scale = cont_scale,
    cont_idx = cont_idx,
    cat_idx = cat_idx,
    ncatfac = ncatfac,
    bandwidth.scale.categorical = bandwidth.scale.categorical,
    cat_upper = cat_upper
  )
}

.npregbw_nomad_point_to_bw <- function(point, template, setup) {
  point <- as.numeric(point)
  bws <- numeric(length(template$bw))
  ncon <- length(setup$cont_idx)
  ncat <- length(setup$cat_idx)

  if (ncon > 0L) {
    gamma <- point[seq_len(ncon)]
    ext_bw <- gamma * setup$cont_scale
    bws[setup$cont_idx] <- if (isTRUE(template$scaling)) gamma else ext_bw
  }

  if (ncat > 0L) {
    lambda_scaled <- point[ncon + seq_len(ncat)]
    ext_bw <- lambda_scaled / setup$bandwidth.scale.categorical
    bws[setup$cat_idx] <- if (isTRUE(template$scaling)) ext_bw / setup$ncatfac else ext_bw
  }

  bws
}

.npregbw_nomad_bw_to_point <- function(bws, template, setup) {
  point <- numeric(length(setup$cont_idx) + length(setup$cat_idx))

  if (length(setup$cont_idx) > 0L) {
    raw <- bws[setup$cont_idx]
    point[seq_along(setup$cont_idx)] <- if (isTRUE(template$scaling)) {
      raw
    } else {
      raw / setup$cont_scale
    }
  }

  if (length(setup$cat_idx) > 0L) {
    raw <- bws[setup$cat_idx]
    ext_bw <- if (isTRUE(template$scaling)) raw * setup$ncatfac else raw
    point[length(setup$cont_idx) + seq_along(setup$cat_idx)] <- ext_bw * setup$bandwidth.scale.categorical
  }

  point
}

.npregbw_powell_progress_fields <- function(state,
                                            done = NULL,
                                            detail = NULL,
                                            now = .np_progress_now()) {
  fields <- .np_degree_progress_context_fields()
  elapsed <- max(0, now - state$started)

  fields <- c(fields, sprintf("elapsed %ss", .np_progress_fmt_num(elapsed)))

  if (!is.null(state$nomad_current_degree)) {
    fields <- c(
      fields,
      sprintf("degree %s", .np_degree_format_degree(state$nomad_current_degree))
    )
  }

  if (!is.null(done)) {
    done <- suppressWarnings(as.integer(done)[1L])
    if (!is.na(done) && done >= 1L) {
      fields <- c(fields, sprintf("iter %s", format(done)))
    }
  }

  fields
}

.npregbw_with_powell_refinement_progress <- function(degree, expr) {
  old.state <- .np_progress_runtime$bandwidth_state
  active.state <- .np_progress_begin(
    label = .np_nomad_powell_progress_label(),
    domain = "general",
    surface = "bandwidth"
  )

  on.exit({
    current.state <- .np_progress_runtime$bandwidth_state
    if (!is.null(current.state)) {
      .np_progress_end(current.state)
    }
    .np_progress_runtime$bandwidth_state <- old.state
  }, add = TRUE)

  active.state$unknown_total_fields <- .npregbw_powell_progress_fields
  active.state$nomad_nmulti <- 1L
  active.state$nomad_current_degree <- as.integer(degree)
  active.state <- .np_progress_show_now(active.state)
  .np_progress_runtime$bandwidth_state <- active.state

  value <- force(expr)

  if (!is.null(active.state) && is.list(value) && !is.null(value$num.feval)) {
    done <- suppressWarnings(as.integer(value$num.feval[1L]))
    if (!is.na(done) && done >= 1L && !is.null(.np_progress_runtime$bandwidth_state)) {
      .np_progress_runtime$bandwidth_state <- .np_progress_step_at(
        state = .np_progress_runtime$bandwidth_state,
        now = .np_progress_now(),
        done = done,
        force = TRUE
      )
    }
  }

  value
}

.npregbw_nomad_search <- function(xdat,
                                  ydat,
                                  bws,
                                  reg.args,
                                  opt.args,
                                  yname,
                                  degree.search,
                                  nomad.inner.nmulti = 0L,
                                  random.seed = 42L,
                                  nomad.opts = list()) {
  if (isTRUE(degree.search$verify))
    stop("automatic degree search with search.engine='nomad' does not support degree.verify")

  template.reg.args <- reg.args
  template.reg.args$regtype <- "lp"
  template.reg.args$degree <- as.integer(degree.search$start.degree)
  template.reg.args$bernstein.basis <- degree.search$bernstein.basis
  template <- .npregbw_build_rbandwidth(
    xdat = xdat,
    ydat = ydat,
    bws = bws,
    bandwidth.compute = FALSE,
    reg.args = template.reg.args,
    yname = yname
  )
  if (!identical(template$type, "fixed"))
    stop("automatic degree search with search.engine='nomad' currently requires bwtype='fixed'")

  setup <- .npregbw_nomad_bw_setup(xdat = xdat, template = template)
  ncon <- length(setup$cont_idx)
  ncat <- length(setup$cat_idx)
  nomad.nmulti <- if (is.null(opt.args$nmulti)) npDefaultNmulti(dim(xdat)[2]) else npValidateNmulti(opt.args$nmulti[1L])

  cont_lower <- npGetScaleFactorSearchLower(template,
                                            argname = "template$scale.factor.search.lower")
  bw_lower <- c(rep.int(cont_lower, ncon), rep.int(0, ncat))
  bw_upper <- c(rep.int(1e6, ncon), setup$cat_upper * setup$bandwidth.scale.categorical)

  point.start <- if (all(template$bw == 0)) {
    NULL
  } else {
    .npregbw_nomad_bw_to_point(template$bw, template = template, setup = setup)
  }
  x0 <- c(
    .np_nomad_complete_start_point(
      point = point.start,
      lower = bw_lower,
      upper = bw_upper,
      ncont = ncon
    ),
    as.integer(degree.search$start.degree)
  )
  lb <- c(bw_lower, degree.search$lower)
  ub <- c(bw_upper, degree.search$upper)
  bbin <- c(rep.int(0L, ncon + ncat), rep.int(1L, ncon))
  baseline.record <- NULL
  nomad.num.feval.total <- 0
  nomad.num.feval.fast.total <- 0

  .np_nomad_baseline_note(degree.search$start.degree)

  eval_fun <- function(point) {
    point <- as.numeric(point)
    degree <- as.integer(round(point[ncon + ncat + seq_len(ncon)]))
    degree <- .np_degree_clip_to_grid(degree, degree.search$candidates)
    bw_vec <- .npregbw_nomad_point_to_bw(point[seq_len(ncon + ncat)], template = template, setup = setup)

    eval.reg.args <- reg.args
    eval.reg.args$regtype <- "lp"
    eval.reg.args$degree <- degree
    eval.reg.args$bernstein.basis <- degree.search$bernstein.basis

    tbw <- .npregbw_build_rbandwidth(
      xdat = xdat,
      ydat = ydat,
      bws = bw_vec,
      bandwidth.compute = FALSE,
      reg.args = eval.reg.args,
      yname = yname
    )

    out <- .npregbw_eval_only(
      xdat = xdat,
      ydat = ydat,
      bws = tbw,
      invalid.penalty = "baseline",
      penalty.multiplier = if (is.null(opt.args$penalty.multiplier)) 10 else opt.args$penalty.multiplier
    )
    nomad.num.feval.total <<- nomad.num.feval.total + as.numeric(out$num.feval[1L])
    nomad.num.feval.fast.total <<- nomad.num.feval.fast.total + as.numeric(out$num.feval.fast[1L])

    list(
      objective = out$objective,
      degree = degree,
      num.feval = out$num.feval
    )
  }

  build_payload <- function(point, best_record, solution, interrupted) {
    point <- as.numeric(point)
    degree <- as.integer(best_record$degree)
    bw_vec <- .npregbw_nomad_point_to_bw(point[seq_len(ncon + ncat)], template = template, setup = setup)
    powell.elapsed <- NA_real_
    final.reg.args <- reg.args
    final.reg.args$regtype <- "lp"
    final.reg.args$degree <- degree
    final.reg.args$bernstein.basis <- degree.search$bernstein.basis
    final.tbw <- .npregbw_build_rbandwidth(
      xdat = xdat,
      ydat = ydat,
      bws = bw_vec,
      bandwidth.compute = FALSE,
      reg.args = final.reg.args,
      yname = yname
    )

    direct.payload <- .npregbw_finalize_fixed_degree_payload(
      xdat = xdat,
      ydat = ydat,
      bws = final.tbw,
      core = list(
        bw = as.numeric(final.tbw$bw),
        objective = as.numeric(best_record$objective),
        ifval = as.numeric(best_record$objective),
        num.feval = as.numeric(nomad.num.feval.total),
        num.feval.fast = as.numeric(nomad.num.feval.fast.total),
        fval.history = as.numeric(best_record$objective),
        eval.history = if (!is.null(solution$bbe)) rep(1, max(1L, as.integer(solution$bbe))) else 1,
        invalid.history = 0,
        timing = NA_real_
      ),
      total.time = NA_real_
    )
    direct.objective <- as.numeric(best_record$objective)

    if (identical(degree.search$engine, "nomad+powell")) {
      hot.opt.args <- .np_nomad_powell_hotstart_opt_args(
        opt.args,
        strategy = "disable_multistart",
        remin = isTRUE(opt.args$powell.remin)
      )

      powell.start <- proc.time()[3L]
      hot.payload <- .npregbw_with_powell_refinement_progress(degree, local({
        .npregbw_run_fixed_degree_source_of_truth(
          xdat = xdat,
          ydat = ydat,
          bws = final.tbw,
          opt.args = hot.opt.args
        )
      }))
      powell.elapsed <- proc.time()[3L] - powell.start

      direct.payload$num.feval <- as.numeric(direct.payload$num.feval[1L]) + as.numeric(hot.payload$num.feval[1L])
      direct.payload$num.feval.fast <- as.numeric(direct.payload$num.feval.fast[1L]) + as.numeric(hot.payload$num.feval.fast[1L])
      hot.payload$num.feval <- direct.payload$num.feval
      hot.payload$num.feval.fast <- direct.payload$num.feval.fast
      if (!is.null(hot.payload$method) && length(hot.payload$method))
        hot.payload$pmethod <- bwmToPrint(as.character(hot.payload$method[1L]))

      hot.objective <- as.numeric(hot.payload$fval[1L])
      if (is.finite(hot.objective) &&
          .np_degree_better(hot.objective, direct.objective, direction = "min")) {
        return(list(payload = hot.payload, objective = hot.objective, powell.time = powell.elapsed))
      }
    }

    list(payload = direct.payload, objective = direct.objective, powell.time = powell.elapsed)
  }

  search.engine.used <- if (identical(degree.search$engine, "nomad+powell")) {
    "nomad"
  } else {
    degree.search$engine
  }

  search.result <- .np_nomad_search(
    engine = search.engine.used,
    baseline_record = baseline.record,
    start_degree = degree.search$start.degree,
    x0 = x0,
    bbin = bbin,
    lb = lb,
    ub = ub,
    eval_fun = eval_fun,
    build_payload = build_payload,
    direction = "min",
    objective_name = "fval",
    nmulti = nomad.nmulti,
    nomad.inner.nmulti = nomad.inner.nmulti,
    random.seed = random.seed,
    handoff_before_build = identical(degree.search$engine, "nomad+powell"),
    remin = isTRUE(opt.args$nomad.remin),
    nomad.opts = nomad.opts,
    degree_spec = list(
      initial = degree.search$start.degree,
      lower = degree.search$lower,
      upper = degree.search$upper,
      basis = degree.search$basis,
      nobs = degree.search$nobs,
      user_supplied = degree.search$start.user
    )
  )

  if (!identical(search.engine.used, degree.search$engine)) {
    search.result$method <- degree.search$engine
  }

  search.result
}

.npregbw_run_fixed_degree <- function(xdat, ydat, bws, reg.args, opt.args, yname) {
  tbw <- .npregbw_build_rbandwidth(
    xdat = xdat,
    ydat = ydat,
    bws = bws,
    bandwidth.compute = opt.args$bandwidth.compute,
    reg.args = reg.args,
    yname = yname
  )

  do.call(npregbw.rbandwidth, c(list(xdat = xdat, ydat = ydat, bws = tbw), opt.args))
}

.npregbw_degree_search_controls <- function(regtype,
                                            regtype.named,
                                            ncon,
                                            nobs,
                                            basis,
                                            degree.select,
                                            search.engine,
                                            degree.min,
                                            degree.max,
                                            degree.start,
                                            degree.restarts,
                                            degree.max.cycles,
                                            degree.verify,
                                            bernstein.basis,
                                            bernstein.named) {
  degree.select <- match.arg(degree.select, c("manual", "coordinate", "exhaustive"))
  if (identical(degree.select, "manual"))
    return(NULL)
  search.engine <- .npregbw_nomad_controls(search.engine)

  regtype.requested <- if (isTRUE(regtype.named)) match.arg(regtype, c("lc", "ll", "lp")) else "lc"
  if (!identical(regtype.requested, "lp"))
    stop("automatic degree search currently requires regtype='lp'")
  if (ncon < 1L)
    stop("automatic degree search requires at least one continuous regressor")

  bern.auto <- if (isTRUE(bernstein.named)) bernstein.basis else TRUE
  bern.auto <- npValidateGlpBernstein(regtype = "lp", bernstein.basis = bern.auto)

  bounds <- .np_degree_normalize_bounds(
    ncon = ncon,
    degree.min = degree.min,
    degree.max = degree.max,
    default.max = 3L
  )

  baseline.degree <- rep.int(0L, ncon)
  default.start.degree <- if (identical(search.engine, "cell")) {
    baseline.degree
  } else {
    rep.int(1L, ncon)
  }
  start.degree <- if (is.null(degree.start)) {
    pmax(bounds$lower, pmin(bounds$upper, default.start.degree))
  } else {
    start.raw <- npValidateGlpDegree(regtype = "lp", degree = degree.start, ncon = ncon, argname = "degree.start")
    out.of.range <- vapply(seq_len(ncon), function(j) !(start.raw[j] %in% bounds$candidates[[j]]), logical(1))
    if (any(out.of.range))
      stop("degree.start must lie within the searched degree candidates for every continuous predictor")
    start.raw
  }

  list(
    method = if (identical(search.engine, "cell")) degree.select else search.engine,
    engine = search.engine,
    candidates = bounds$candidates,
    lower = bounds$lower,
    upper = bounds$upper,
    baseline.degree = baseline.degree,
    start.degree = start.degree,
    start.user = !is.null(degree.start),
    basis = if (missing(basis) || is.null(basis)) "glp" else as.character(basis[1L]),
    nobs = as.integer(nobs[1L]),
    restarts = npValidateNonNegativeInteger(degree.restarts, "degree.restarts"),
    max.cycles = npValidatePositiveInteger(degree.max.cycles, "degree.max.cycles"),
    verify = npValidateScalarLogical(degree.verify, "degree.verify"),
    bernstein.basis = bern.auto
  )
}

.npregbw_attach_degree_search <- function(bws, search_result) {
  metadata <- list(
    mode = search_result$method,
    verify = isTRUE(search_result$verify),
    completed = isTRUE(search_result$completed),
    certified = isTRUE(search_result$certified),
    interrupted = isTRUE(search_result$interrupted),
    baseline.degree = search_result$baseline$degree,
    baseline.fval = search_result$baseline$objective,
    best.degree = search_result$best$degree,
    best.fval = search_result$best$objective,
    nomad.time = search_result$nomad.time,
    powell.time = search_result$powell.time,
    optim.time = search_result$optim.time,
    n.unique = search_result$n.unique,
    n.visits = search_result$n.visits,
    n.cached = search_result$n.cached,
    grid.size = search_result$grid.size,
    best.restart = search_result$best.restart,
    restart.starts = search_result$restart.starts,
    restart.degree.starts = search_result$restart.degree.starts,
    restart.bandwidth.starts = search_result$restart.bandwidth.starts,
    restart.start.info = search_result$restart.start.info,
    restart.results = search_result$restart.results,
    trace = search_result$trace
  )

  if (!is.null(search_result$nomad.time))
    bws$nomad.time <- as.numeric(search_result$nomad.time[1L])
  if (!is.null(search_result$powell.time))
    bws$powell.time <- as.numeric(search_result$powell.time[1L])
  if (!is.null(search_result$optim.time) && is.finite(search_result$optim.time))
    bws$total.time <- as.numeric(search_result$optim.time[1L])
  bws <- .np_attach_nomad_restart_summary(bws, search_result)
  bws$degree.search <- metadata
  bws
}

npregbw.default <-
  function(xdat = stop("invoked without data 'xdat'"),
           ydat = stop("invoked without data 'ydat'"),
           bws,
           bandwidth.compute = TRUE,
           basis,
           bernstein.basis,
           bwmethod,
           bwscaling,
           bwtype,
           cfac.dir,
           scale.factor.init,
           ckerbound,
           ckerlb,
           ckerorder,
           ckertype,
           ckerub,
           degree,
           degree.select = c("manual", "coordinate", "exhaustive"),
           search.engine = c("nomad+powell", "cell", "nomad"),
           nomad = FALSE,
           nomad.nmulti = 0L,
           degree.min = NULL,
           degree.max = NULL,
           degree.start = NULL,
           degree.restarts = 0L,
           degree.max.cycles = 20L,
           degree.verify = FALSE,
           dfac.dir,
           dfac.init,
           dfc.dir,
           ftol,
           scale.factor.init.upper,
           hbd.dir,
           hbd.init,
           initc.dir,
           initd.dir,
           invalid.penalty = c("baseline","dbmax"),
           itmax,
           lbc.dir,
           scale.factor.init.lower,
           lbd.dir,
           lbd.init,
           nmulti,
           okertype,
           penalty.multiplier = 10,
           regtype,
           nomad.remin = FALSE,
           powell.remin,
           scale.init.categorical.sample,
           scale.factor.search.lower = NULL,
           small,
           tol,
           transform.bounds = FALSE,
           ukertype,
           ...){

    xdat <- toFrame(xdat)
    yname <- deparse(substitute(ydat))
    lp.dot.args <- list(...)
    if ("remin" %in% names(lp.dot.args)) {
      legacy.remin <- npValidateScalarLogical(lp.dot.args$remin, "remin")
      warning("npregbw: argument 'remin' is deprecated; use 'powell.remin' and 'nomad.remin'",
              call. = FALSE)
      if (missing(powell.remin))
        powell.remin <- legacy.remin
      if (missing(nomad.remin))
        nomad.remin <- legacy.remin
      lp.dot.args$remin <- NULL
    }
    npRejectLegacyLpArgs(names(lp.dot.args), where = "npregbw")
    random.seed.value <- .np_degree_extract_random_seed(lp.dot.args)
    scale.factor.search.lower <- npResolveScaleFactorLowerBound(scale.factor.search.lower)

    if (!(is.vector(ydat) || is.factor(ydat)))
      stop("'ydat' must be a vector")

    ## first grab dummy args for bandwidth() and perform 'bootstrap'
    ## bandwidth() call

    mc.names <- names(match.call(expand.dots = FALSE))
    nomad.shortcut <- .np_prepare_nomad_shortcut(
      nomad = nomad,
      call_names = mc.names,
      preset = list(
        regtype = "lp",
        search.engine = "nomad+powell",
        degree.select = "coordinate",
        bernstein.basis = TRUE,
        degree.min = 0L,
        degree.max = 10L,
        degree.verify = FALSE,
        bwtype = "fixed"
      ),
      values = list(
        regtype = if ("regtype" %in% mc.names) regtype else NULL,
        search.engine = if ("search.engine" %in% mc.names) search.engine else NULL,
        degree.select = if ("degree.select" %in% mc.names) degree.select else NULL,
        bernstein.basis = if ("bernstein.basis" %in% mc.names) bernstein.basis else NULL,
        degree.min = if ("degree.min" %in% mc.names) degree.min else NULL,
        degree.max = if ("degree.max" %in% mc.names) degree.max else NULL,
        degree.verify = if ("degree.verify" %in% mc.names) degree.verify else NULL,
        bwtype = if ("bwtype" %in% mc.names) bwtype else NULL,
        degree = if ("degree" %in% mc.names) degree else NULL
      ),
      where = "npregbw"
    )

    if (isTRUE(nomad.shortcut$enabled)) {
      if (sum(untangle(xdat)$icon) == 0L)
        stop("nomad=TRUE requires at least one continuous predictor for degree search",
             call. = FALSE)
      if ("degree" %in% mc.names)
        stop("nomad=TRUE does not support an explicit degree; remove degree or set nomad=FALSE")
      if ("regtype" %in% mc.names &&
          !identical(as.character(match.arg(nomad.shortcut$values$regtype, c("lc", "ll", "lp")))[1L], "lp"))
        stop("nomad=TRUE requires regtype='lp'")
      if ("bwtype" %in% mc.names &&
          !identical(as.character(match.arg(nomad.shortcut$values$bwtype, c("fixed", "generalized_nn", "adaptive_nn")))[1L], "fixed"))
        stop("nomad=TRUE currently requires bwtype='fixed'")
      if ("degree.select" %in% mc.names &&
          identical(as.character(match.arg(nomad.shortcut$values$degree.select, c("manual", "coordinate", "exhaustive")))[1L], "manual"))
        stop("nomad=TRUE requires automatic degree search; use degree.select='coordinate' or 'exhaustive'")
      if ("search.engine" %in% mc.names &&
          !(as.character(match.arg(nomad.shortcut$values$search.engine, c("nomad+powell", "cell", "nomad")))[1L] %in%
              c("nomad", "nomad+powell")))
        stop("nomad=TRUE requires search.engine='nomad' or 'nomad+powell'")
      if ("bernstein.basis" %in% mc.names &&
          !isTRUE(npValidateGlpBernstein(regtype = "lp",
                                        bernstein.basis = nomad.shortcut$values$bernstein.basis)))
        stop("nomad=TRUE currently requires bernstein.basis=TRUE")
      if ("degree.verify" %in% mc.names &&
          isTRUE(npValidateScalarLogical(nomad.shortcut$values$degree.verify, "degree.verify")))
        stop("nomad=TRUE currently requires degree.verify=FALSE")
    }

    margs <- c("regtype", "basis", "degree", "bernstein.basis", "bwmethod", "bwscaling", "bwtype",
               "ckertype", "ckerorder", "ckerbound", "ckerlb", "ckerub", "ukertype", "okertype")

    m <- match(margs, mc.names, nomatch = 0)
    any.m <- any(m != 0)

    bw.args <- list(
      bw = bws,
      nobs = dim(xdat)[1],
      xdati = untangle(xdat),
      ydati = untangle(data.frame(ydat)),
      xnames = names(xdat),
      ynames = yname,
      bandwidth.compute = bandwidth.compute
    )
    if (any.m) {
      nms <- mc.names[m]
      bw.args[nms] <- mget(nms, envir = environment(), inherits = FALSE)
    }

    mc.names <- names(match.call(expand.dots = FALSE))
    margs <- c("nmulti", "nomad.remin", "powell.remin", "itmax", "ftol", "tol",
               "small",
               "lbc.dir", "dfc.dir", "cfac.dir","initc.dir", 
               "lbd.dir", "hbd.dir", "dfac.dir", "initd.dir", 
               "scale.factor.init.lower", "scale.factor.init.upper", "scale.factor.init", 
               "lbd.init", "hbd.init", "dfac.init", 
               "scale.init.categorical.sample",
               "transform.bounds",
               "scale.factor.search.lower",
               "invalid.penalty",
               "penalty.multiplier")
    m <- match(margs, mc.names, nomatch = 0)
    any.m <- any(m != 0)

    if (any.m) {
      nms <- mc.names[m]
      opt.args <- mget(nms, envir = environment(), inherits = FALSE)
    } else {
      opt.args <- list()
    }

    reg.args <- bw.args[setdiff(names(bw.args), c("bw", "nobs", "xdati", "ydati", "xnames", "ynames", "bandwidth.compute"))]
    opt.args <- c(list(bandwidth.compute = bandwidth.compute), opt.args)
    reg.args$scale.factor.search.lower <- scale.factor.search.lower
    opt.args$scale.factor.search.lower <- scale.factor.search.lower

    ncon <- sum(untangle(xdat)$icon)
    search.mc.names <- mc.names
    regtype.value <- if (!is.null(nomad.shortcut$values$regtype)) nomad.shortcut$values$regtype else "lc"
    bernstein.value <- if (!is.null(nomad.shortcut$values$bernstein.basis)) nomad.shortcut$values$bernstein.basis else TRUE
    degree.select.value <- if (!is.null(nomad.shortcut$values$degree.select)) nomad.shortcut$values$degree.select else "manual"
    search.engine.value <- if (!is.null(nomad.shortcut$values$search.engine)) nomad.shortcut$values$search.engine else "nomad+powell"
    degree.min.value <- nomad.shortcut$values$degree.min
    degree.max.value <- nomad.shortcut$values$degree.max
    degree.start.value <- if ("degree.start" %in% search.mc.names) degree.start else NULL
    degree.restarts.value <- if ("degree.restarts" %in% search.mc.names) degree.restarts else 0L
    degree.max.cycles.value <- if ("degree.max.cycles" %in% search.mc.names) degree.max.cycles else 20L
    degree.verify.value <- if (!is.null(nomad.shortcut$values$degree.verify)) nomad.shortcut$values$degree.verify else FALSE
    degree.search <- .npregbw_degree_search_controls(
      regtype = regtype.value,
      regtype.named = isTRUE(nomad.shortcut$enabled) || ("regtype" %in% search.mc.names),
      ncon = ncon,
      nobs = NROW(xdat),
      basis = if ("basis" %in% search.mc.names) basis else "glp",
      degree.select = degree.select.value,
      search.engine = search.engine.value,
      degree.min = degree.min.value,
      degree.max = degree.max.value,
      degree.start = degree.start.value,
      degree.restarts = degree.restarts.value,
      degree.max.cycles = degree.max.cycles.value,
      degree.verify = degree.verify.value,
      bernstein.basis = bernstein.value,
      bernstein.named = isTRUE(nomad.shortcut$enabled) || ("bernstein.basis" %in% search.mc.names)
    )
    nomad.inner.named <- "nomad.nmulti" %in% search.mc.names
    nomad.inner.nmulti <- if (nomad.inner.named) {
      npValidateNonNegativeInteger(nomad.nmulti, "nomad.nmulti")
    } else {
      0L
    }
    if (nomad.inner.named &&
        (is.null(degree.search) || !(degree.search$engine %in% c("nomad", "nomad+powell")))) {
      stop("nomad.nmulti is only supported when regtype='lp', automatic degree search is active, and search.engine is 'nomad' or 'nomad+powell'")
    }
    if (!is.null(degree.search) && is.null(reg.args$degree)) {
      reg.args$degree <- npSetupGlpDegree(
        regtype = regtype.value,
        degree = NULL,
        ncon = ncon,
        degree.select = degree.select.value
      )
    }

    if (!is.null(degree.search)) {
      if (identical(degree.search$engine, "cell")) {
        eval_fun <- function(degree.vec) {
          cell.reg.args <- reg.args
          cell.reg.args$regtype <- "lp"
          cell.reg.args$degree <- as.integer(degree.vec)
          cell.reg.args$bernstein.basis <- degree.search$bernstein.basis
          cell.bws <- .npregbw_run_fixed_degree(
            xdat = xdat,
            ydat = ydat,
            bws = bws,
            reg.args = cell.reg.args,
            opt.args = opt.args,
            yname = yname
          )
          list(
            objective = as.numeric(cell.bws$fval[1L]),
            payload = cell.bws,
            num.feval = if (!is.null(cell.bws$num.feval)) as.numeric(cell.bws$num.feval[1L]) else NA_real_
          )
        }

        search.result <- .np_degree_search(
          method = degree.search$method,
          candidates = degree.search$candidates,
          baseline_degree = degree.search$baseline.degree,
          start_degree = degree.search$start.degree,
          restarts = degree.search$restarts,
          max_cycles = degree.search$max.cycles,
          verify = degree.search$verify,
          eval_fun = eval_fun,
          direction = "min",
          trace_level = "full",
          objective_name = "fval"
        )
      } else {
        search.result <- .npregbw_nomad_search(
          xdat = xdat,
          ydat = ydat,
          bws = bws,
          reg.args = reg.args,
          opt.args = opt.args,
          yname = yname,
          degree.search = degree.search,
          nomad.inner.nmulti = nomad.inner.nmulti,
          random.seed = random.seed.value,
          nomad.opts = list(
            DIRECTION_TYPE = "ORTHO 2N",
            QUAD_MODEL_SEARCH = "no",
            NM_SEARCH = "no",
            SPECULATIVE_SEARCH = "no",
            EVAL_OPPORTUNISTIC = "no"
          )
        )
      }
      tbw <- .npregbw_attach_degree_search(
        bws = search.result$best_payload,
        search_result = search.result
      )
      tbw <- .np_attach_nomad_shortcut(tbw, nomad.shortcut$metadata)
      mc <- match.call(expand.dots = FALSE)
      environment(mc) <- parent.frame()
      tbw$call <- mc
      return(tbw)
    }

    tbw <- .npregbw_build_rbandwidth(
      xdat = xdat,
      ydat = ydat,
      bws = bws,
      bandwidth.compute = bandwidth.compute,
      reg.args = reg.args,
      yname = yname
    )
    bwsel.args <- c(list(xdat = xdat, ydat = ydat, bws = tbw), opt.args)
    tbw <- .np_progress_select_bandwidth_enhanced(
      "Selecting regression bandwidth",
      do.call(npregbw.rbandwidth, bwsel.args)
    )
    tbw <- .np_attach_nomad_shortcut(tbw, nomad.shortcut$metadata)

    mc <- match.call(expand.dots = FALSE)
    environment(mc) <- parent.frame()
    tbw$call <- mc

    return(tbw)
    
  }

test_that("inid lc fast path matches explicit resample refits", {
  skip_if_not_installed("np")

  set.seed(321)
  n <- 40
  x <- runif(n)
  y <- cos(2 * pi * x) + rnorm(n, sd = 0.1)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 21))

  bw <- npregbw(
    y ~ x,
    regtype = "lc",
    bws = c(0.2),
    bandwidth.compute = FALSE
  )

  H <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix")
  B <- 13L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  fast.fun <- getFromNamespace(".np_inid_lc_boot_from_hat", "np")
  fast.out <- fast.fun(H = H, ydat = y, B = B, counts = counts)

  explicit.t <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  for (b in seq_len(B)) {
    idx <- rep.int(seq_len(n), counts[, b])
    fit.b <- npreg(
      txdat = tx[idx, , drop = FALSE],
      tydat = y[idx],
      exdat = ex,
      bws = bw,
      gradients = FALSE,
      warn.glp.gradient = FALSE
    )
    explicit.t[b, ] <- fit.b$mean
  }

  expect_equal(fast.out$t, explicit.t, tolerance = 1e-10)
  expect_equal(fast.out$t0, as.vector(H %*% y), tolerance = 1e-12)
})

test_that("inid ll/lp fast path matches explicit resample refits", {
  skip_if_not_installed("np")

  set.seed(3211)
  n <- 45
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(2 * pi * x1) + 0.6 * x2 + rnorm(n, sd = 0.08)
  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- tx[seq_len(15), , drop = FALSE]
  B <- 9L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  fast.fun <- getFromNamespace(".np_inid_boot_from_regression", "np")
  cfgs <- list(
    list(regtype = "ll", basis = NULL, degree = NULL, label = "ll"),
    list(regtype = "lp", basis = "additive", degree = c(2L, 2L), label = "lp-additive"),
    list(regtype = "lp", basis = "tensor", degree = c(2L, 2L), label = "lp-tensor"),
    list(regtype = "lp", basis = "glp", degree = c(2L, 2L), label = "lp-glp")
  )

  for (cfg in cfgs) {
    bw.args <- list(
      xdat = tx,
      ydat = y,
      regtype = cfg$regtype,
      bws = c(0.3, 0.3),
      bandwidth.compute = FALSE
    )
    if (!is.null(cfg$basis)) {
      bw.args$basis <- cfg$basis
      bw.args$degree <- cfg$degree
    }
    bw <- do.call(npregbw, bw.args)

    fast.out <- fast.fun(
      xdat = tx,
      exdat = ex,
      bws = bw,
      ydat = y,
      B = B,
      counts = counts
    )

    explicit.t <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
    for (b in seq_len(B)) {
      idx <- rep.int(seq_len(n), counts[, b])
      explicit.t[b, ] <- npreg(
        txdat = tx[idx, , drop = FALSE],
        tydat = y[idx],
        exdat = ex,
        bws = bw,
        gradients = FALSE,
        warn.glp.gradient = FALSE
      )$mean
    }

    expect_equal(fast.out$t, explicit.t, tolerance = 1e-6, info = cfg$label)
    expect_equal(
      fast.out$t0,
      npreg(txdat = tx, tydat = y, exdat = ex, bws = bw, gradients = FALSE, warn.glp.gradient = FALSE)$mean,
      tolerance = 1e-6,
      info = cfg$label
    )
  }
})

test_that("npplreg inid fast path matches explicit resample refits", {
  skip_if_not_installed("np")

  set.seed(32316)
  n <- 40
  x1 <- runif(n)
  x2 <- runif(n)
  z1 <- runif(n)
  z2 <- runif(n)
  tx <- data.frame(x1 = x1, x2 = x2)
  tz <- data.frame(z1 = z1, z2 = z2)
  y <- sin(2 * pi * z1) + 0.5 * x1 - 0.2 * x2 + rnorm(n, sd = 0.08)
  B <- 9L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  bw <- npplregbw(xdat = tx, ydat = y, zdat = tz, regtype = "lc", nmulti = 1)
  fast.fun <- getFromNamespace(".np_inid_boot_from_plreg", "np")
  fast.out <- fast.fun(
    txdat = tx,
    ydat = y,
    tzdat = tz,
    exdat = tx,
    ezdat = tz,
    bws = bw,
    B = B,
    counts = counts
  )

  explicit.t <- matrix(NA_real_, nrow = B, ncol = n)
  for (b in seq_len(B)) {
    idx <- rep.int(seq_len(n), counts[, b])
    explicit.t[b, ] <- npplreg(
      bws = bw,
      txdat = tx[idx, , drop = FALSE],
      tydat = y[idx],
      tzdat = tz[idx, , drop = FALSE],
      exdat = tx,
      ezdat = tz
    )$mean
  }

  fit0 <- npplreg(
    bws = bw,
    txdat = tx,
    tydat = y,
    tzdat = tz,
    exdat = tx,
    ezdat = tz
  )$mean

  expect_equal(fast.out$t, explicit.t, tolerance = 1e-6)
  expect_equal(as.vector(fast.out$t0), as.vector(fit0), tolerance = 1e-7)
})

test_that("npreg inid fast path supports continuous-slice gradients", {
  skip_if_not_installed("np")

  set.seed(3212)
  n <- 45
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(2 * pi * x1) + 0.4 * x2 + rnorm(n, sd = 0.08)
  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- tx[seq_len(15), , drop = FALSE]
  B <- 9L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  bw <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "ll",
    bws = c(0.3, 0.3),
    bandwidth.compute = FALSE
  )

  fast.fun <- getFromNamespace(".np_inid_boot_from_regression", "np")
  fast.out <- fast.fun(
    xdat = tx,
    exdat = ex,
    bws = bw,
    ydat = y,
    B = B,
    counts = counts,
    gradients = TRUE,
    gradient.order = 1L,
    slice.index = 1L
  )

  explicit.t <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  for (b in seq_len(B)) {
    idx <- rep.int(seq_len(n), counts[, b])
    explicit.t[b, ] <- npreg(
      txdat = tx[idx, , drop = FALSE],
      tydat = y[idx],
      exdat = ex,
      bws = bw,
      gradients = TRUE,
      gradient.order = 1L,
      warn.glp.gradient = FALSE
    )$grad[, 1L]
  }

  fit0 <- npreg(
    txdat = tx,
    tydat = y,
    exdat = ex,
    bws = bw,
    gradients = TRUE,
    gradient.order = 1L,
    warn.glp.gradient = FALSE
  )$grad[, 1L]

  expect_equal(fast.out$t, explicit.t, tolerance = 1e-6)
  expect_equal(as.vector(fast.out$t0), as.vector(fit0), tolerance = 1e-6)
})

test_that("direct kernel-weight helper matches npksum.default with evaluation data", {
  kw.fun <- getFromNamespace(".np_kernel_weights_direct", "np")
  ksum.fun <- getFromNamespace("npksum.default", "np")
  set.seed(4021)
  n <- 60
  x <- runif(n)
  z <- rnorm(n)
  tx <- data.frame(x = x, z = z)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 21), z = seq(min(z), max(z), length.out = 21))
  y <- sin(x) + 0.1 * z + rnorm(n, sd = 0.05)

  bw <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "ll",
    bws = c(0.3, 0.4),
    bandwidth.compute = FALSE
  )

  kw.direct <- kw.fun(
    bws = bw,
    txdat = tx,
    exdat = ex,
    leave.one.out = FALSE,
    bandwidth.divide = TRUE
  )

  kw.ref <- ksum.fun(
    bws = bw,
    txdat = tx,
    exdat = ex,
    return.kernel.weights = TRUE,
    bandwidth.divide = TRUE
  )$kw

  expect_equal(kw.direct, kw.ref, tolerance = 1e-12)
})

test_that("direct kernel-weight helper matches npksum.default leave-one-out", {
  kw.fun <- getFromNamespace(".np_kernel_weights_direct", "np")
  ksum.fun <- getFromNamespace("npksum.default", "np")
  set.seed(4022)
  n <- 50
  x <- runif(n)
  y <- cos(2 * pi * x) + rnorm(n, sd = 0.04)
  tx <- data.frame(x = x)

  bw <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "lc",
    bws = 0.25,
    bandwidth.compute = FALSE
  )

  kw.direct <- kw.fun(
    bws = bw,
    txdat = tx,
    exdat = NULL,
    leave.one.out = TRUE,
    bandwidth.divide = TRUE
  )

  kw.ref <- ksum.fun(
    bws = bw,
    txdat = tx,
    leave.one.out = TRUE,
    return.kernel.weights = TRUE,
    bandwidth.divide = TRUE
  )$kw
  if (nrow(kw.ref) == ncol(kw.ref))
    diag(kw.ref) <- 0

  expect_equal(kw.direct, kw.ref, tolerance = 1e-12)
})

test_that("fixed ll gradient helper matches explicit refits without direct regression calls", {
  skip_if_not_installed("np")

  set.seed(20260310)
  n <- 45
  x1 <- runif(n)
  x2 <- runif(n)
  y <- sin(2 * pi * x1) + 0.4 * x2 + rnorm(n, sd = 0.08)
  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- tx[seq_len(15), , drop = FALSE]
  B <- 9L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  bw <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "ll",
    bws = c(0.3, 0.3),
    bandwidth.compute = FALSE
  )

  fast.fun <- getFromNamespace(".np_inid_boot_from_regression", "np")
  np.ns <- asNamespace("np")
  old.calls <- getOption("np.regression.direct.calls")
  options(np.regression.direct.calls = 0L)
  on.exit(options(np.regression.direct.calls = old.calls), add = TRUE)
  trace(
    what = ".np_regression_direct",
    where = np.ns,
    tracer = quote(options(np.regression.direct.calls = getOption("np.regression.direct.calls", 0L) + 1L)),
    print = FALSE
  )
  on.exit(untrace(".np_regression_direct", where = np.ns), add = TRUE)

  options(np.regression.direct.calls = 0L)
  fast.out <- fast.fun(
    xdat = tx,
    exdat = ex,
    bws = bw,
    ydat = y,
    B = B,
    counts = counts,
    gradients = TRUE,
    gradient.order = 1L,
    slice.index = 1L
  )
  fast.calls <- getOption("np.regression.direct.calls")

  explicit.t <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  for (b in seq_len(B)) {
    idx <- rep.int(seq_len(n), counts[, b])
    explicit.t[b, ] <- npreg(
      txdat = tx[idx, , drop = FALSE],
      tydat = y[idx],
      exdat = ex,
      bws = bw,
      gradients = TRUE,
      gradient.order = 1L,
      warn.glp.gradient = FALSE
    )$grad[, 1L]
  }

  fit0 <- npreg(
    txdat = tx,
    tydat = y,
    exdat = ex,
    bws = bw,
    gradients = TRUE,
    gradient.order = 1L,
    warn.glp.gradient = FALSE
  )$grad[, 1L]

  expect_equal(fast.out$t, explicit.t, tolerance = 1e-10)
  expect_equal(as.vector(fast.out$t0), as.vector(fit0), tolerance = 1e-10)
  expect_identical(fast.calls, 0L)
})

test_that("fixed lp gradient helper preserves derivative order and counts drawer semantics", {
  skip_if_not_installed("np")

  set.seed(20260310 + 1L)
  n <- 40
  x1 <- runif(n)
  x2 <- runif(n)
  y <- cos(2 * pi * x1) + x2^2 + rnorm(n, sd = 0.06)
  tx <- data.frame(x1 = x1, x2 = x2)
  ex <- tx[seq_len(12), , drop = FALSE]
  B <- 7L
  counts <- rmultinom(n = B, size = n, prob = rep.int(1 / n, n))

  bw <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "lp",
    degree = c(2L, 2L),
    basis = "glp",
    bernstein.basis = FALSE,
    bws = c(0.3, 0.3),
    bandwidth.compute = FALSE
  )

  fast.fun <- getFromNamespace(".np_inid_boot_from_regression", "np")
  fast.counts <- fast.fun(
    xdat = tx,
    exdat = ex,
    bws = bw,
    ydat = y,
    B = B,
    counts = counts,
    gradients = TRUE,
    gradient.order = c(2L, 1L),
    slice.index = 1L
  )

  drawer <- function(start, stopi) counts[, start:stopi, drop = FALSE]
  fast.drawer <- fast.fun(
    xdat = tx,
    exdat = ex,
    bws = bw,
    ydat = y,
    B = B,
    counts.drawer = drawer,
    gradients = TRUE,
    gradient.order = c(2L, 1L),
    slice.index = 1L
  )

  explicit.t <- matrix(NA_real_, nrow = B, ncol = nrow(ex))
  for (b in seq_len(B)) {
    idx <- rep.int(seq_len(n), counts[, b])
    explicit.t[b, ] <- npreg(
      txdat = tx[idx, , drop = FALSE],
      tydat = y[idx],
      exdat = ex,
      bws = bw,
      gradients = TRUE,
      gradient.order = c(2L, 1L),
      warn.glp.gradient = FALSE
    )$grad[, 1L]
  }

  fit0 <- npreg(
    txdat = tx,
    tydat = y,
    exdat = ex,
    bws = bw,
    gradients = TRUE,
    gradient.order = c(2L, 1L),
    warn.glp.gradient = FALSE
  )$grad[, 1L]

  expect_equal(fast.counts$t, explicit.t, tolerance = 1e-8)
  expect_equal(as.vector(fast.counts$t0), as.vector(fit0), tolerance = 1e-8)
  expect_equal(fast.drawer$t, fast.counts$t, tolerance = 1e-10)
  expect_equal(as.vector(fast.drawer$t0), as.vector(fast.counts$t0), tolerance = 0)
})

test_that("fixed lp plot output records requested gradient order", {
  skip_if_not_installed("np")

  n <- 45
  x <- seq(-1, 1, length.out = n)
  y <- 1 + 2 * x + 3 * x^2
  tx <- data.frame(x = x)

  bw <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "lp",
    degree = 2L,
    basis = "glp",
    bws = 100,
    bwscaling = FALSE,
    bandwidth.compute = FALSE
  )

  fit <- npreg(
    bws = bw,
    gradients = TRUE,
    gradient.order = 1L,
    warn.glp.gradient = FALSE
  )

  pdf(NULL)
  on.exit(dev.off(), add = TRUE)
  out <- plot(
    fit,
    gradients = TRUE,
    gradient_order = 2L,
    errors = "none",
    output = "data",
    neval = 20
  )

  expect_equal(as.vector(out[[1L]]$grad), rep(6, 20), tolerance = 1e-8)
  expect_identical(out[[1L]]$gradient.order, 2L)
})

test_that("npregbw exhaustive degree search matches manual profile minimum", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(x = sort(runif(28)))
  dat$y <- dat$x^2 + rnorm(nrow(dat), sd = 0.05)

  bw0 <- np::npregbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree = 0L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  bw1 <- np::npregbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  auto <- np::npregbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expected_fval <- min(bw0$fval, bw1$fval)

  expect_s3_class(auto, "rbandwidth")
  expect_true(isTRUE(auto$bernstein.basis))
  expect_identical(auto$degree.search$mode, "exhaustive")
  expect_true(isTRUE(auto$degree.search$completed))
  expect_true(isTRUE(auto$degree.search$certified))
  expect_lte(auto$fval, expected_fval + 1e-10)
  expect_lte(auto$degree.search$best.fval, auto$degree.search$baseline.fval + 1e-10)
  expect_true(all(c("degree", "fval", "status", "cached") %in% names(auto$degree.search$trace)))
  expect_identical(nrow(auto$degree.search$trace), auto$degree.search$n.unique)
  expect_identical(auto$degree.search$n.cached, auto$degree.search$n.visits - auto$degree.search$n.unique)

  manual <- np::npregbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  expect_null(manual$degree.search)
})

test_that("npregbw coordinate search can be exhaustively certified on a small grid", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(
    x1 = runif(26),
    x2 = runif(26)
  )
  dat$y <- dat$x1 + dat$x2^2 + rnorm(nrow(dat), sd = 0.05)

  exhaustive <- np::npregbw(
    y ~ x1 + x2,
    data = dat,
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  coordinate <- np::npregbw(
    y ~ x1 + x2,
    data = dat,
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = TRUE,
    degree.restarts = 1L,
    degree.max.cycles = 4L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_identical(coordinate$degree.search$mode, "coordinate")
  expect_true(isTRUE(coordinate$degree.search$completed))
  expect_true(isTRUE(coordinate$degree.search$certified))
  expect_equal(as.integer(coordinate$degree), as.integer(exhaustive$degree))
  expect_equal(coordinate$fval, exhaustive$fval, tolerance = 1e-10)
  expect_lte(coordinate$degree.search$best.fval, coordinate$degree.search$baseline.fval + 1e-10)
  expect_identical(nrow(coordinate$degree.search$trace), coordinate$degree.search$n.unique)
  expect_identical(coordinate$degree.search$n.cached, coordinate$degree.search$n.visits - coordinate$degree.search$n.unique)
})

test_that("npregbw automatic degree search enforces pilot guardrails", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(y = rnorm(24), x = runif(24))

  expect_error(
    np::npregbw(
      y ~ x,
      data = dat,
      regtype = "lc",
      degree.select = "exhaustive",
      degree.min = 0L,
      degree.max = 1L,
      bwtype = "fixed",
      bwmethod = "cv.ls",
      nmulti = 1L
    ),
    "automatic degree search currently requires regtype='lp'"
  )

  bw <- np::npregbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    bernstein.basis = FALSE,
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 4L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_s3_class(bw, "rbandwidth")
  expect_false(isTRUE(bw$bernstein.basis))
  expect_lte(max(as.integer(bw$degree)), 4L)

  bw <- np::npregbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    bernstein.basis = FALSE,
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_false(isTRUE(bw$bernstein.basis))
})

test_that("npreg forwards automatic LP degree search through npregbw", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(x = runif(24))
  dat$y <- dat$x + rnorm(nrow(dat), sd = 0.05)

  fit <- local({
    suppressPackageStartupMessages(library(np))
    npreg(
      y ~ x,
      data = dat,
      regtype = "lp",
      degree.select = "exhaustive",
      search.engine = "cell",
      degree.min = 0L,
      degree.max = 1L,
      bwtype = "fixed",
      bwmethod = "cv.ls",
      nmulti = 1L
    )
  })

  expect_s3_class(fit, "npregression")
  expect_s3_class(fit$bws, "rbandwidth")
  expect_false(is.null(fit$bws$degree.search))
  expect_identical(fit$bws$degree.search$mode, "exhaustive")
})

library(np)

with_npreg_plot_device <- function(expr) {
  tf <- tempfile(fileext = ".pdf")
  grDevices::pdf(tf)
  on.exit({
    grDevices::dev.off()
    unlink(tf)
  }, add = TRUE)
  force(expr)
}

test_that("npreg adaptive-nn wild gradient plot-data stays available in serial", {
  set.seed(42)
  n <- 120L
  x <- sort(stats::runif(n))
  y <- sin(2 * pi * x) + stats::rnorm(n, sd = 0.1)
  xdat <- data.frame(x = x)

  bw <- npregbw(
    xdat = xdat,
    ydat = y,
    regtype = "ll",
    bwtype = "adaptive_nn",
    bws = 35L,
    bandwidth.compute = FALSE
  )
  fit <- npreg(
    bws = bw,
    txdat = xdat,
    tydat = y,
    warn.glp.gradient = FALSE
  )

  out <- with_npreg_plot_device(suppressWarnings(plot(
    fit,
    xdat = xdat,
    ydat = y,
    output = "data",
    perspective = FALSE,
    neval = 40L,
    errors = "bootstrap",
    bootstrap = "wild",
    B = 39L,
    band = "all",
    gradients = TRUE
  )))

  expect_type(out, "list")
  expect_s3_class(out[[1L]], "npregression")
  expect_length(out[[1L]]$mean, 40L)
})

test_that("bounded adaptive_nn is available for unconditional regression", {
  set.seed(20260325)
  n <- 36
  x <- runif(n)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.15)
  xdf <- data.frame(x = x)

  bw.fixed <- npregbw(
    xdat = xdf,
    ydat = y,
    regtype = "ll",
    bwmethod = "cv.ls",
    bwtype = "fixed",
    ckerorder = 2,
    ckerbound = "range",
    nmulti = 1,
    powell.remin = FALSE,
    itmax = 50,
    tol = 0.1
  )

  bw.gnn <- npregbw(
    xdat = xdf,
    ydat = y,
    regtype = "ll",
    bwmethod = "cv.ls",
    bwtype = "generalized_nn",
    ckerorder = 2,
    ckerbound = "range",
    nmulti = 1,
    powell.remin = FALSE,
    itmax = 50,
    tol = 0.1
  )

  bw.ad <- npregbw(
    xdat = xdf,
    ydat = y,
    regtype = "ll",
    bwmethod = "cv.ls",
    bwtype = "adaptive_nn",
    ckerorder = 2,
    ckerbound = "range",
    nmulti = 1,
    powell.remin = FALSE,
    itmax = 50,
    tol = 0.1
  )

  fit.ad <- npreg(bws = bw.ad, gradients = TRUE)

  expect_true(is.finite(bw.fixed$fval))
  expect_true(is.finite(bw.gnn$fval))
  expect_true(is.finite(bw.ad$fval))
  expect_true(all(is.finite(as.numeric(bw.ad$bw))))
  expect_true(all(is.finite(as.numeric(fitted(fit.ad)))))
  expect_true(all(is.finite(as.numeric(fit.ad$gradients))))
  expect_false(isTRUE(all.equal(as.numeric(bw.ad$bw), as.numeric(bw.fixed$bw))))
  expect_false(isTRUE(all.equal(as.numeric(bw.ad$bw), as.numeric(bw.gnn$bw))))
})

test_that("npregbw NOMAD degree search backend improves over the baseline", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(x = sort(runif(16)))
  dat$y <- sin(2 * pi * dat$x) + rnorm(nrow(dat), sd = 0.05)

  bw <- np::npregbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "nomad",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 2L
  )

  expect_s3_class(bw, "rbandwidth")
  expect_identical(bw$degree.search$mode, "nomad")
  expect_true(isTRUE(bw$degree.search$completed))
  expect_gte(bw$degree.search$n.unique, 1L)
  expect_lte(bw$degree.search$best.fval, bw$degree.search$baseline.fval + 1e-10)
  expect_identical(length(bw$degree.search$restart.degree.starts), 2L)
  expect_identical(length(bw$degree.search$restart.results), 2L)
  expect_identical(bw$degree.search$restart.results[[1L]]$restart, 1L)
  expect_identical(length(bw$nomad.restart.fval), 2L)
  expect_identical(length(bw$nomad.restart.results), 2L)
  expect_identical(length(bw$nomad.restart.degree.starts), 2L)
  restart.objectives <- unlist(lapply(
    bw$degree.search$restart.results,
    function(x) if (is.null(x$objective)) NA_real_ else as.numeric(x$objective[1L])
  ))
  restart.objectives <- restart.objectives[is.finite(restart.objectives)]
  expect_true(length(restart.objectives) >= 1L)
  expect_lte(bw$degree.search$best.fval, min(restart.objectives) + 1e-6)
  expect_identical(as.numeric(bw$nomad.restart.fval), as.numeric(unlist(lapply(
    bw$nomad.restart.results,
    function(x) if (is.null(x$objective)) NA_real_ else as.numeric(x$objective[1L])
  ))))
  expect_identical(bw$nomad.best.restart, which.min(bw$nomad.restart.fval))
})

suppressPackageStartupMessages(library(np))

test_that("nonfixed search uses observation-count bounds for adaptive and generalized", {
  set.seed(123)

  x <- sample(rep(1:8, each = 10), 60, replace = FALSE)
  y <- x + rnorm(60)
  n_kmax <- length(x) - 1L

  bw.adap <- npregbw(ydat = y, xdat = data.frame(x = x), bwtype = "adaptive_nn")
  bw.gen <- npregbw(ydat = y, xdat = data.frame(x = x), bwtype = "generalized_nn")

  expect_gte(unname(bw.adap$bw), 1)
  expect_lte(unname(bw.adap$bw), n_kmax)
  expect_gte(unname(bw.gen$bw), 1)
  expect_lte(unname(bw.gen$bw), n_kmax)
})

test_that("explicit k may exceed empirical support size when observation count allows it", {
  x <- c(0, 0, 0, 1)
  y <- c(0, 1, 2, 3)

  bw.adap <- npregbw(
    ydat = y,
    xdat = data.frame(x = x),
    bwtype = "adaptive_nn",
    bws = length(x) - 1L,
    bandwidth.compute = FALSE
  )
  bw.gen <- npregbw(
    ydat = y,
    xdat = data.frame(x = x),
    bwtype = "generalized_nn",
    bws = length(x) - 1L,
    bandwidth.compute = FALSE
  )

  expect_no_error(
    npreg(bws = bw.adap, exdat = data.frame(x = x))
  )

  expect_no_error(
    npreg(bws = bw.gen, exdat = data.frame(x = x)),
  )
})

test_that("nonfixed regression bandwidth search fails fast for constant regressors", {
  expect_error(
    suppressWarnings(
      npregbw(ydat = rnorm(50), xdat = data.frame(x = rep(1, 50)), bwtype = "adaptive_nn")
    ),
    "at least two distinct continuous regressor values per dimension"
  )
})

test_that("nonfixed conditional bandwidth search fails fast for constant response-side support", {
  x <- data.frame(x = rep(1:5, 8))
  y <- rep(1, 40)

  expect_error(
    suppressWarnings(
      npcdensbw(ydat = y, xdat = x, bwtype = "adaptive_nn")
    ),
    "at least two distinct continuous variable values per dimension"
  )
})

fixed_cvls_objective <- function(xdat,
                                 ydat,
                                 regtype,
                                 degree,
                                 bws,
                                 basis = "glp",
                                 bernstein.basis = FALSE) {
  ns <- asNamespace("np")
  xdat <- as.data.frame(xdat)
  bw0 <- np::npregbw(
    xdat = xdat,
    ydat = ydat,
    bws = bws,
    regtype = regtype,
    degree = degree,
    basis = basis,
    bernstein.basis = bernstein.basis,
    bwmethod = "cv.ls",
    bandwidth.compute = FALSE
  )

  xmat <- get("toMatrix", ns)(xdat)
  runo <- xmat[, bw0$iuno, drop = FALSE]
  rcon <- xmat[, bw0$icon, drop = FALSE]
  rord <- xmat[, bw0$iord, drop = FALSE]
  mysd <- get("EssDee", ns)(rcon)
  n <- nrow(xdat)
  nconfac <- n^(-1.0 / (2.0 * bw0$ckerorder + bw0$ncon))
  ncatfac <- n^(-2.0 / (2.0 * bw0$ckerorder + bw0$ncon))
  reg.c <- get("npRegtypeToC", ns)(
    regtype = bw0$regtype,
    degree = bw0$degree,
    ncon = bw0$ncon,
    context = "npregbw"
  )

  myopti <- as.integer(list(
    n,
    get("IMULTI_FALSE", ns),
    1L,
    get("USE_START_YES", ns),
    if (bw0$scaling) get("SF_NORMAL", ns) else get("SF_ARB", ns),
    get("BW_FIXED", ns),
    0L,
    get("RE_MIN_FALSE", ns),
    get("IO_MIN_TRUE", ns),
    get("BWM_CVLS", ns),
    get("CKER_GAUSS", ns) + bw0$ckerorder / 2 - 1,
    get("UKER_AIT", ns),
    get("OKER_WANG", ns),
    bw0$nuno,
    bw0$nord,
    bw0$ncon,
    reg.c$code,
    get("DO_TREE_NO", ns),
    FALSE,
    3L,
    FALSE
  ))
  myoptd <- as.double(list(
    1.490116e-07,
    1.490116e-04,
    1.490116e-05,
    0.5,
    2.5 * (3.0 - sqrt(5)),
    1.0,
    0.1,
    1,
    0.25 * (3.0 - sqrt(5)),
    1.0,
    0.1,
    2.0,
    0.5,
    0.1,
    0.9,
    0.375,
    nconfac,
    ncatfac
  ))
  ck <- get("npKernelBoundsMarshal", ns)(bw0$ckerlb[bw0$icon], bw0$ckerub[bw0$icon])

  out <- .Call(
    "C_np_regression_bw_eval",
    as.double(runo),
    as.double(rord),
    as.double(rcon),
    as.double(ydat),
    as.double(mysd),
    myopti,
    myoptd,
    as.double(c(bw0$bw[bw0$icon], bw0$bw[bw0$iuno], bw0$bw[bw0$iord])),
    as.integer(1L),
    as.integer(1L),
    as.double(10),
    as.integer(if (bw0$ncon > 0) bw0$degree else integer(1)),
    as.integer(isTRUE(bw0$bernstein.basis)),
    as.integer(get("npLpBasisCode", ns)(bw0$basis)),
    as.double(ck$lb),
    as.double(ck$ub),
    PACKAGE = "np"
  )

  unname(out$fval[1])
}

test_that("fixed cv.ls is exact on in-class local polynomial fixtures", {
  tol <- 1e-20

  set.seed(20260308)
  x1 <- sort(runif(30))
  y1 <- 1 + 2 * x1
  obj.ll <- fixed_cvls_objective(
    xdat = data.frame(x = x1),
    ydat = y1,
    regtype = "ll",
    degree = 1L,
    bws = 0.25
  )
  obj.lp1 <- fixed_cvls_objective(
    xdat = data.frame(x = x1),
    ydat = y1,
    regtype = "lp",
    degree = 1L,
    bws = 0.25
  )
  expect_lt(abs(obj.ll), tol)
  expect_lt(abs(obj.lp1), tol)
  expect_equal(obj.ll, obj.lp1, tolerance = 1e-24)

  set.seed(20260309)
  x2 <- runif(36)
  z2 <- runif(36)
  y2 <- 0.5 + 1.2 * x2 - 0.7 * z2 + 0.8 * x2^2 - 0.3 * x2 * z2 + 0.6 * z2^2
  obj.lp2 <- fixed_cvls_objective(
    xdat = data.frame(x = x2, z = z2),
    ydat = y2,
    regtype = "lp",
    degree = c(2L, 2L),
    bws = c(0.3, 0.3)
  )
  expect_lt(abs(obj.lp2), tol)

  set.seed(20260310)
  x3 <- runif(40)
  z3 <- runif(40)
  w3 <- runif(40)
  y3 <- 0.2 + 0.9 * x3 + 0.6 * z3 - 0.4 * w3 + 0.7 * x3^2
  obj.lp211 <- fixed_cvls_objective(
    xdat = data.frame(x = x3, z = z3, w = w3),
    ydat = y3,
    regtype = "lp",
    degree = c(2L, 1L, 1L),
    bws = c(0.35, 0.35, 0.35)
  )
  expect_lt(abs(obj.lp211), tol)
})

test_that("fixed cv.ls keeps ll and canonical lp degree-1 aligned off-model", {
  set.seed(20260311)
  x <- sort(runif(48))
  y <- sin(2 * pi * x) + 0.15 * x

  obj.ll <- fixed_cvls_objective(
    xdat = data.frame(x = x),
    ydat = y,
    regtype = "ll",
    degree = 1L,
    bws = 0.22
  )
  obj.lp <- fixed_cvls_objective(
    xdat = data.frame(x = x),
    ydat = y,
    regtype = "lp",
    degree = 1L,
    bws = 0.22
  )

  expect_equal(obj.ll, obj.lp, tolerance = 1e-12)

  set.seed(20260312)
  x1 <- runif(52)
  x2 <- runif(52)
  u <- factor(sample(c("a", "b", "c"), 52, replace = TRUE))
  o <- ordered(sample(1:3, 52, replace = TRUE))
  y2 <- sin(2 * pi * x1) + 0.2 * x2 + as.numeric(u) / 9 + as.numeric(o) / 11

  obj2.ll <- fixed_cvls_objective(
    xdat = data.frame(x1 = x1, x2 = x2, u = u, o = o),
    ydat = y2,
    regtype = "ll",
    degree = c(1L, 1L),
    bws = c(0.28, 0.31, 0.45, 0.55)
  )
  obj2.lp <- fixed_cvls_objective(
    xdat = data.frame(x1 = x1, x2 = x2, u = u, o = o),
    ydat = y2,
    regtype = "lp",
    degree = c(1L, 1L),
    bws = c(0.28, 0.31, 0.45, 0.55)
  )

  expect_equal(obj2.ll, obj2.lp, tolerance = 1e-12)
})

library(np)

test_that("generalized higher-degree lp owner matches npreg on evaluation data", {
  run_case <- function(train_df, y, eval_df, degree, basis = "glp", bern = FALSE, tol = 1e-10) {
    frame <- train_df
    frame$y <- y
    fml <- as.formula(paste("y~", paste(names(train_df), collapse = "+")))
    bw <- npregbw(fml,
                  data = frame,
                  regtype = "lp",
                  degree = degree,
                  basis = basis,
                  bernstein.basis = bern,
                  bwtype = "generalized_nn")

    H <- npreghat(bws = bw, txdat = train_df, exdat = eval_df)
    a <- npreghat(bws = bw, txdat = train_df, exdat = eval_df, y = y, output = "apply")
    g <- npreg(bws = bw, exdat = eval_df)

    expect_equal(drop(H %*% y), a, tolerance = tol)
    expect_equal(drop(H %*% y), g$mean, tolerance = tol)
  }

  set.seed(101)
  n <- 80
  x <- runif(n, -1, 1)
  y <- x^2 + rnorm(n, sd = 0.15)
  xe <- data.frame(x = seq(-1, 1, length.out = 41))
  run_case(data.frame(x = x), y, xe, degree = 2L)
  run_case(data.frame(x = x), y, xe, degree = 3L)

  set.seed(102)
  n <- 70
  x1 <- runif(n, -1, 1)
  x2 <- runif(n, -1, 1)
  y <- x1^2 - x2 + x1 * x2 + rnorm(n, sd = 0.1)
  xe <- expand.grid(x1 = seq(-1, 1, length.out = 6),
                    x2 = seq(-1, 1, length.out = 6))
  for (b in c("glp", "additive", "tensor")) {
    run_case(data.frame(x1 = x1, x2 = x2), y, xe, degree = c(2L, 2L), basis = b)
  }

  set.seed(103)
  n <- 70
  x1 <- c(0, 1, runif(n - 2L))
  x2 <- c(1, 0, runif(n - 2L))
  y <- sin(pi * x1) + x2^2 + rnorm(n, sd = 0.1)
  xe <- expand.grid(x1 = seq(0.1, 0.9, length.out = 6),
                    x2 = seq(0.1, 0.9, length.out = 6))
  run_case(data.frame(x1 = x1, x2 = x2),
           y,
           xe,
           degree = c(2L, 2L),
           basis = "tensor",
           bern = TRUE,
           tol = 1e-9)

  set.seed(104)
  n <- 70
  x1 <- runif(n, -1, 1)
  x2 <- runif(n, -1, 1)
  u <- factor(sample(c("a", "b", "c"), n, replace = TRUE))
  o <- ordered(sample(1:3, n, replace = TRUE))
  y <- x1^2 + x2 - 0.5 * (u == "b") + 0.3 * (o == 3) + rnorm(n, sd = 0.1)
  xe <- data.frame(
    x1 = seq(-1, 1, length.out = 12),
    x2 = seq(-1, 1, length.out = 12),
    u = factor(rep(c("a", "b", "c"), length.out = 12), levels = levels(u)),
    o = ordered(rep(1:3, length.out = 12), levels = levels(o))
  )
  run_case(data.frame(x1 = x1, x2 = x2, u = u, o = o),
           y,
           xe,
           degree = c(2L, 2L),
           basis = "tensor")
})

test_that("tree-enabled generalized higher-degree lp owner stays exact via core fallback", {
  old_tree <- getOption("np.tree")
  options(np.tree = TRUE)
  on.exit(options(np.tree = old_tree), add = TRUE)

  set.seed(203)
  n <- 80
  x <- runif(n, -1, 1)
  y <- x^2 + rnorm(n, sd = 0.2)
  d <- data.frame(x = x)
  bw <- npregbw(y ~ x,
                data = data.frame(d, y = y),
                regtype = "lp",
                degree = 2L,
                basis = "glp",
                bwtype = "generalized_nn")

  H <- npreghat(bws = bw, txdat = d)
  a <- npreghat(bws = bw, txdat = d, y = y, output = "apply")
  g <- npreg(bws = bw)

  expect_equal(drop(H %*% y), a, tolerance = 1e-10)
  expect_equal(drop(H %*% y), g$mean, tolerance = 1e-10)
})

test_that("generalized higher-degree lp derivative owner matches npreg on evaluation data", {
  run_case <- function(train_df, y, eval_df, degree, s, basis = "glp", bern = FALSE, tol = 1e-9) {
    frame <- train_df
    frame$y <- y
    fml <- as.formula(paste("y~", paste(names(train_df), collapse = "+")))
    bw <- npregbw(fml,
                  data = frame,
                  regtype = "lp",
                  degree = degree,
                  basis = basis,
                  bernstein.basis = bern,
                  bwtype = "generalized_nn")

    H <- npreghat(bws = bw, txdat = train_df, exdat = eval_df, s = s)
    a <- npreghat(bws = bw, txdat = train_df, exdat = eval_df, y = y, output = "apply", s = s)
    g <- npreg(bws = bw, exdat = eval_df, gradients = TRUE)
    target.col <- which(as.integer(s) == 1L)[1L]

    expect_equal(drop(H %*% y), a, tolerance = tol)
    expect_equal(drop(H %*% y), g$grad[, target.col], tolerance = tol)
  }

  set.seed(301)
  n <- 80
  x <- runif(n, -1, 1)
  y <- x^2 + rnorm(n, sd = 0.15)
  xe <- data.frame(x = seq(-1, 1, length.out = 41))
  run_case(data.frame(x = x), y, xe, degree = 2L, s = 1L)
  run_case(data.frame(x = x), y, xe, degree = 3L, s = 1L)

  set.seed(302)
  n <- 70
  x1 <- runif(n, -1, 1)
  x2 <- runif(n, -1, 1)
  y <- x1^2 - x2 + x1 * x2 + rnorm(n, sd = 0.1)
  xe <- expand.grid(x1 = seq(-1, 1, length.out = 6),
                    x2 = seq(-1, 1, length.out = 6))
  run_case(data.frame(x1 = x1, x2 = x2),
           y,
           xe,
           degree = c(2L, 2L),
           s = c(1L, 0L),
           basis = "tensor")
})

test_that("generalized NN regression bandwidth is stable on duplicated continuous data", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  x <- data.frame(x = rep(1:5, each = 4))
  y <- c(
    1, 1.1, 0.9, 1.05,
    2, 2.1, 1.9, 2.05,
    3, 3.1, 2.9, 3.05,
    4, 4.2, 3.8, 4.1,
    5, 5.2, 4.8, 5.1
  )

  bw <- npregbw(
    xdat = x,
    ydat = y,
    bwtype = "generalized_nn",
    nmulti = 1
  )

  expect_equal(as.numeric(bw$bw), 2)
  expect_equal(as.numeric(bw$fval), 0.1717472951192434, tolerance = 1e-14)
})

test_that("adaptive NN regression bandwidth is stable on duplicated continuous data", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  x <- data.frame(x = rep(1:5, each = 4))
  y <- c(
    1, 1.1, 0.9, 1.05,
    2, 2.1, 1.9, 2.05,
    3, 3.1, 2.9, 3.05,
    4, 4.2, 3.8, 4.1,
    5, 5.2, 4.8, 5.1
  )

  bw <- npregbw(
    xdat = x,
    ydat = y,
    bwtype = "adaptive_nn",
    nmulti = 1
  )

  expect_equal(as.numeric(bw$bw), 3)
  expect_equal(as.numeric(bw$fval), 0.1717472951192434, tolerance = 1e-14)
})

test_that("wild categorical regression gradient helper matches explicit refits", {
  skip_if_not_installed("np")

  library(np)

  wild_helper <- getFromNamespace(".np_plot_boot_from_hat_wild_factor_effects", "np")
  rad_draws <- getFromNamespace(".np_rademacher_draws", "np")

  set.seed(20260312)
  n <- 36L
  g <- factor(sample(c("a", "b"), n, replace = TRUE))
  x <- runif(n)
  y <- 1 + 0.5 * (g == "b") + sin(2 * pi * x) + rnorm(n, sd = 0.05)
  xdat <- data.frame(g = g, x = x)
  bw <- npregbw(
    xdat = xdat,
    ydat = y,
    regtype = "ll",
    bwtype = "fixed",
    bws = c(0.25, 0.3),
    bandwidth.compute = FALSE
  )

  exdat <- plot(
    bw,
    xdat = xdat,
    ydat = y,
    gradients = TRUE,
    output = "data",
    perspective = FALSE
  )[[1L]]$eval
  H <- npreghat(
    bws = bw,
    txdat = xdat,
    exdat = exdat,
    output = "matrix"
  )
  fit.mean <- as.vector(npreghat(
    bws = bw,
    txdat = xdat,
    exdat = xdat,
    y = y,
    output = "apply"
  ))

  B <- 7L
  set.seed(11)
  helper.out <- wild_helper(
    H = H,
    ydat = y,
    fit.mean = fit.mean,
    B = B,
    wild = "rademacher"
  )

  set.seed(11)
  draws <- rad_draws(n = n, B = B)
  explicit.t <- matrix(NA_real_, nrow = B, ncol = nrow(exdat))
  for (b in seq_len(B)) {
    ystar <- fit.mean + (y - fit.mean) * draws[, b]
    mean.b <- npreg(
      txdat = xdat,
      tydat = ystar,
      exdat = exdat,
      bws = bw,
      gradients = FALSE,
      warn.glp.gradient = FALSE
    )$mean
    explicit.t[b, ] <- mean.b - mean.b[1L]
  }
  fit0 <- npreg(
    txdat = xdat,
    tydat = y,
    exdat = exdat,
    bws = bw,
    gradients = FALSE,
    warn.glp.gradient = FALSE
  )$mean

  expect_equal(helper.out$t, explicit.t, tolerance = 1e-6)
  expect_equal(as.vector(helper.out$t0), as.vector(fit0 - fit0[1L]), tolerance = 1e-6)
})

test_that("categorical regression gradient bootstrap works for default, inid, and wild routes", {
  skip_if_not_installed("np")

  library(np)

  set.seed(20260312)
  n <- 36L
  g <- factor(sample(c("a", "b"), n, replace = TRUE))
  x <- runif(n)
  y <- 1 + 0.4 * (g == "b") + cos(2 * pi * x) + rnorm(n, sd = 0.05)
  xdat <- data.frame(g = g, x = x)
  bw <- npregbw(
    xdat = xdat,
    ydat = y,
    regtype = "ll",
    bwtype = "fixed",
    bws = c(0.25, 0.3),
    bandwidth.compute = FALSE
  )
  fit <- npreg(
    bws = bw,
    txdat = xdat,
    tydat = y,
    gradients = TRUE,
    errors = TRUE
  )

  for (boot.method in c("default", "inid", "wild")) {
    args <- list(
      fit,
      output = "data",
      perspective = FALSE,
      gradients = TRUE,
      errors = "bootstrap",
      B = 11L
    )
    if (!identical(boot.method, "default"))
      args$bootstrap <- boot.method

    out <- suppressWarnings(do.call(plot, args))
    expect_type(out, "list")
    expect_true(length(out) >= 1L, info = boot.method)
    expect_true(length(out[[1L]]$bxp) > 0L, info = boot.method)
    expect_equal(length(out[[1L]]$bxp$names), 2L, info = boot.method)
    expect_true(all(is.finite(out[[1L]]$grad)), info = boot.method)
    expect_true(all(is.finite(out[[1L]]$glerr)), info = boot.method)
    expect_true(all(is.finite(out[[1L]]$gherr)), info = boot.method)
  }
})

npreg_fit_progress_time_counter <- function(start = 0, by = 1.7) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

npreg_fit_progress_time_values <- function(values) {
  force(values)
  i <- 0L
  function() {
    i <<- min(i + 1L, length(values))
    values[[i]]
  }
}

npreg_fit_progress_lines <- function(shadow) {
  vapply(shadow$trace, `[[`, character(1L), "line")
}

expect_np_npreg_powell_progress_surface <- function(lines) {
  powell.lines <- lines[grepl("^\\[np\\] Refining bandwidth \\(", lines)]
  info <- paste(lines, collapse = "\n")

  expect_true(length(powell.lines) > 0L, info = info)
  expect_false(any(grepl(
    "^\\[np\\] Refining NOMAD solution with one Powell hot start at degree ",
    lines
  )), info = info)
  expect_false(any(grepl("best (", powell.lines, fixed = TRUE)), info = info)
  expect_true(any(grepl(
    "^\\[np\\] Refining bandwidth \\(elapsed [0-9]+\\.[0-9]s, degree \\([^)]*\\), iter [0-9]+\\)$",
    powell.lines
  )), info = info)
}

make_npreg_fit_progress_fixture <- function() {
  set.seed(20260404)
  n <- 24L
  dat <- data.frame(x = seq(-0.9, 0.9, length.out = n))
  dat$y <- sin(2 * pi * dat$x) + 0.35 * dat$x

  list(
    dat = dat,
    tx = dat["x"],
    y = dat$y,
    bw = npregbw(
      xdat = dat["x"],
      ydat = dat$y,
      bws = 0.22,
      bandwidth.compute = FALSE
    )
  )
}

test_that("npreg direct bws fit emits known-total fit progress", {
  fixture <- make_npreg_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npreg(
      bws = fixture$bw,
      txdat = fixture$tx,
      tydat = fixture$y
    ),
    force_renderer = "single_line",
    now = npreg_fit_progress_time_counter()
  )

  lines <- npreg_fit_progress_lines(actual)

  expect_s3_class(actual$value, "npregression")
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 1/24 \\([0-9]+\\.[0-9]%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 24/24 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )))
})

test_that("npreg direct bws fit stays silent below start grace without handoff", {
  fixture <- make_npreg_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0.75,
    np.progress.interval.known.sec = 0.5
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npreg(
      bws = fixture$bw,
      txdat = fixture$tx,
      tydat = fixture$y
    ),
    force_renderer = "single_line",
    now = npreg_fit_progress_time_values(c(0, 0.2, 0.4, 0.6))
  )

  expect_length(actual$trace, 0L)
})

test_that("npreg bw to fit route hands off immediately into single-line fit progress", {
  fixture <- make_npreg_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npreg(
      y ~ x,
      data = fixture$dat,
      nmulti = 1L
    ),
    force_renderer = "single_line",
    now = npreg_fit_progress_time_counter()
  )

  lines <- npreg_fit_progress_lines(actual)
  bandwidth.pos <- grep("^\\[np\\] Bandwidth selection \\(", lines)
  fit.start.pos <- grep(
    "^\\[np\\] Fitting regression 0/24 \\(0\\.0%, elapsed 0\\.0s, eta 0\\.0s\\): starting$",
    lines
  )
  fit.finish.pos <- grep(
    "^\\[np\\] Fitting regression 24/24 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )

  expect_s3_class(actual$value, "npregression")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(fit.start.pos) == 1L)
  expect_true(length(fit.finish.pos) >= 1L)
  expect_lt(max(bandwidth.pos), fit.start.pos[[1L]])
  expect_lt(fit.start.pos[[1L]], fit.finish.pos[[1L]])
})

test_that("npreg nomad to powell to fit route preserves single-line fit handoff", {
  skip_if_not_installed("crs")

  fixture <- make_npreg_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0,
    np.progress.interval.known.sec = 0,
    np.progress.interval.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npreg(
      y ~ x,
      data = fixture$dat,
      nomad = TRUE,
      degree.max = 1L,
      nmulti = 1L
    ),
    force_renderer = "single_line",
    now = npreg_fit_progress_time_counter()
  )

  lines <- npreg_fit_progress_lines(actual)
  bandwidth.pos <- grep("^\\[np\\] Selecting degree and bandwidth", lines)
  powell.pos <- grep("^\\[np\\] Refining bandwidth \\(", lines)
  fit.start.pos <- grep(
    "^\\[np\\] Fitting regression 0/24 \\(0\\.0%, elapsed 0\\.0s, eta 0\\.0s\\): starting$",
    lines
  )
  fit.finish.pos <- grep(
    "^\\[np\\] Fitting regression 24/24 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )

  expect_s3_class(actual$value, "npregression")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(powell.pos) > 0L)
  expect_true(length(fit.start.pos) == 1L)
  expect_true(length(fit.finish.pos) >= 1L)
  expect_np_npreg_powell_progress_surface(lines)
  expect_lt(max(bandwidth.pos), fit.start.pos[[1L]])
  expect_lt(max(powell.pos), fit.start.pos[[1L]])
  expect_lt(fit.start.pos[[1L]], fit.finish.pos[[1L]])
})

test_that("predict.npregression re-entry emits fit progress", {
  fixture <- make_npreg_fit_progress_fixture()

  fit <- npreg(
    bws = fixture$bw,
    txdat = fixture$tx,
    tydat = fixture$y
  )

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    predict(fit, newdata = fixture$dat[c(2L, 7L), "x", drop = FALSE]),
    force_renderer = "single_line",
    now = npreg_fit_progress_time_counter()
  )

  lines <- npreg_fit_progress_lines(actual)

  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 1/2 \\(50\\.0%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting regression 2/2 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\)$",
    lines
  )))
})

library(np)

make_jksum_mixed_data <- function(n = 220L, seed = 42L) {
  set.seed(seed)
  x1 <- runif(n)
  x2 <- runif(n)
  z1 <- rbinom(n, 1, 0.5)
  z2 <- rbinom(n, 1, 0.5)
  y <- cos(2 * pi * x1) + 0.5 * sin(2 * pi * x2) + z1 + rnorm(n, sd = 0.20)
  data.frame(
    y = y,
    x1 = x1,
    x2 = x2,
    z1 = factor(z1),
    z2 = ordered(z2)
  )
}

run_reg_cv_once <- function(dat, regtype, bwmethod) {
  t_bw <- system.time(
    bw <- npregbw(
      y ~ x1 + x2 + z1 + z2,
      regtype = regtype,
      bwmethod = bwmethod,
      nmulti = 1,
      data = dat
    )
  )

  list(
    fval = as.numeric(bw$fval),
    nfe = as.integer(bw$num.feval),
    elapsed = as.numeric(t_bw[["elapsed"]])
  )
}

test_that("jksum regression CV parity is deterministic for mixed data", {
  old_opts <- options(
    np.messages = FALSE,
    np.largeh.rel.tol = 0.05,
    np.disc.upper.rel.tol = 0.05
  )
  on.exit(options(old_opts), add = TRUE)

  dat <- make_jksum_mixed_data(n = 220L, seed = 100L)
  combos <- expand.grid(
    regtype = c("ll", "lc"),
    bwmethod = c("cv.ls", "cv.aic"),
    stringsAsFactors = FALSE
  )

  for (i in seq_len(nrow(combos))) {
    regtype <- combos$regtype[[i]]
    bwmethod <- combos$bwmethod[[i]]

    set.seed(123)
    r1 <- run_reg_cv_once(dat, regtype, bwmethod)
    set.seed(123)
    r2 <- run_reg_cv_once(dat, regtype, bwmethod)

    expect_true(is.finite(r1$fval))
    expect_true(is.finite(r2$fval))
    expect_true(r1$nfe > 0L)
    expect_true(r2$nfe > 0L)
    expect_equal(r1$nfe, r2$nfe)
    expect_equal(r1$fval, r2$fval, tolerance = 1e-12)
  }
})

test_that("jksum regression CV smoke performance remains bounded", {
  skip_on_cran()

  old_opts <- options(
    np.messages = FALSE,
    np.largeh.rel.tol = 0.05,
    np.disc.upper.rel.tol = 0.05
  )
  on.exit(options(old_opts), add = TRUE)

  dat <- make_jksum_mixed_data(n = 220L, seed = 101L)

  set.seed(321)
  r_ll <- run_reg_cv_once(dat, "ll", "cv.ls")
  set.seed(321)
  r_lc <- run_reg_cv_once(dat, "lc", "cv.ls")

  expect_true(r_ll$nfe > 0L)
  expect_true(r_lc$nfe > 0L)
  expect_true(is.finite(r_ll$fval))
  expect_true(is.finite(r_lc$fval))

  # Guardrail: this should remain a small smoke test in CI-scale environments.
  expect_lt(r_ll$elapsed + r_lc$elapsed, 20)
})

test_that("large-h fast gateway stays active for lc/ll/lp under canonical DGPs", {
  old_opts <- options(
    np.messages = FALSE,
    np.largeh.rel.tol = 0.05,
    np.disc.upper.rel.tol = 0.05
  )
  on.exit(options(old_opts), add = TRUE)

  set.seed(42)
  n <- 200L
  x <- runif(n)
  y_lc <- rnorm(n, sd = 0.5 * sd(x))
  y_ll <- x + rnorm(n, sd = 0.5 * sd(x))
  dat_lc <- data.frame(y = y_lc, x = x)
  dat_ll <- data.frame(y = y_ll, x = x)

  set.seed(42)
  bw_ll <- npregbw(y ~ x, data = dat_ll, regtype = "ll", bwmethod = "cv.ls", nmulti = 1)
  set.seed(42)
  bw_lc <- npregbw(y ~ x, data = dat_lc, regtype = "lc", bwmethod = "cv.ls", nmulti = 1)
  set.seed(42)
  bw_lp <- npregbw(
    y ~ x,
    data = dat_ll,
    regtype = "lp",
    degree = 1,
    bwmethod = "cv.ls",
    nmulti = 1
  )

  expect_true(is.finite(as.numeric(bw_ll$fval)))
  expect_true(is.finite(as.numeric(bw_lc$fval)))
  expect_true(is.finite(as.numeric(bw_lp$fval)))

  expect_gt(as.integer(bw_ll$num.feval.fast), 0L)
  expect_gt(as.integer(bw_lp$num.feval.fast), 0L)
  expect_gt(as.integer(bw_lc$num.feval.fast), 0L)
  expect_equal(as.numeric(bw_ll$fval), as.numeric(bw_lp$fval), tolerance = 1e-10)
})

test_that("adaptive ll matrix hat matches apply and npreg for mean and gradient", {
  set.seed(42)

  n <- 80
  x <- runif(n)
  y <- x + rnorm(n)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 25))

  bw <- npregbw(y ~ x, regtype = "ll", bwtype = "adaptive_nn")

  mean.apply <- npreghat(
    bws = bw,
    txdat = tx,
    exdat = ex,
    y = y,
    output = "apply"
  )
  mean.matrix <- npreghat(
    bws = bw,
    txdat = tx,
    exdat = ex,
    output = "matrix"
  )

  grad.apply <- npreghat(
    bws = bw,
    txdat = tx,
    exdat = ex,
    y = y,
    output = "apply",
    s = 1
  )
  grad.matrix <- npreghat(
    bws = bw,
    txdat = tx,
    exdat = ex,
    output = "matrix",
    s = 1
  )

  np.mean <- npreg(txdat = tx, tydat = y, bws = bw, exdat = ex)$mean
  np.grad <- npreg(txdat = tx, tydat = y, bws = bw, exdat = ex, gradients = TRUE)$grad[, 1L]

  expect_equal(drop(mean.matrix %*% y), mean.apply, tolerance = 1e-12)
  expect_equal(drop(mean.matrix %*% y), np.mean, tolerance = 1e-12)
  expect_equal(drop(grad.matrix %*% y), grad.apply, tolerance = 1e-12)
  expect_equal(drop(grad.matrix %*% y), np.grad, tolerance = 1e-12)
})

test_that("adaptive lc matrix hat matches fitted values and predict on train and eval data", {
  set.seed(20260315)

  n <- 64L
  x <- sort(runif(n))
  y <- sin(2 * pi * x) + 0.2 * x + rnorm(n, sd = 0.03)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(0.05, 0.95, length.out = 21L))

  bw <- npregbw(y ~ x, regtype = "lc", bwtype = "adaptive_nn", nmulti = 1L)

  fit.train <- npreg(bws = bw, txdat = tx, tydat = y)
  fit.eval <- npreg(bws = bw, txdat = tx, tydat = y, exdat = ex)

  hat.train.matrix <- npreghat(bws = bw, txdat = tx, output = "matrix")
  hat.train.apply <- npreghat(bws = bw, txdat = tx, y = y, output = "apply")
  hat.eval.matrix <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix")
  hat.eval.apply <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply")

  expect_equal(drop(hat.train.matrix %*% y), as.vector(fit.train$mean), tolerance = 1e-12)
  expect_equal(as.vector(hat.train.apply), as.vector(fit.train$mean), tolerance = 1e-12)
  expect_equal(drop(hat.train.matrix %*% y), as.vector(hat.train.apply), tolerance = 1e-12)

  expect_equal(drop(hat.eval.matrix %*% y), as.vector(fit.eval$mean), tolerance = 1e-12)
  expect_equal(as.vector(hat.eval.apply), as.vector(fit.eval$mean), tolerance = 1e-12)
  expect_equal(drop(hat.eval.matrix %*% y), as.vector(hat.eval.apply), tolerance = 1e-12)
  expect_equal(drop(hat.eval.matrix %*% y), as.vector(predict(fit.train, newdata = ex)), tolerance = 1e-12)
})

test_that("adaptive lp matrix hat matches apply and npreg on train and eval data across degrees", {
  set.seed(20260315)

  n <- 72L
  x <- sort(rnorm(n))
  y <- x^3 - 0.4 * x + rnorm(n, sd = 0.04)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 27L))

  for (degree in 1:3) {
    bw <- npregbw(
      xdat = tx,
      ydat = y,
      regtype = "lp",
      degree = degree,
      bwtype = "adaptive_nn",
      bandwidth.compute = FALSE,
      bws = 18L
    )

    fit.train <- npreg(bws = bw, txdat = tx, tydat = y, gradients = TRUE)
    fit.eval <- npreg(bws = bw, txdat = tx, tydat = y, exdat = ex, gradients = TRUE)

    mean.train.apply <- npreghat(bws = bw, txdat = tx, y = y, output = "apply")
    mean.train.matrix <- npreghat(bws = bw, txdat = tx, output = "matrix")
    mean.eval.apply <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply")
    mean.eval.matrix <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix")

    grad.train.apply <- npreghat(bws = bw, txdat = tx, y = y, output = "apply", s = 1L)
    grad.train.matrix <- npreghat(bws = bw, txdat = tx, output = "matrix", s = 1L)
    grad.eval.apply <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply", s = 1L)
    grad.eval.matrix <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix", s = 1L)

    expect_equal(drop(mean.train.matrix %*% y), as.vector(fit.train$mean), tolerance = 1e-10)
    expect_equal(as.vector(mean.train.apply), as.vector(fit.train$mean), tolerance = 1e-10)
    expect_equal(drop(mean.train.matrix %*% y), as.vector(mean.train.apply), tolerance = 1e-10)

    expect_equal(drop(mean.eval.matrix %*% y), as.vector(fit.eval$mean), tolerance = 1e-10)
    expect_equal(as.vector(mean.eval.apply), as.vector(fit.eval$mean), tolerance = 1e-10)
    expect_equal(drop(mean.eval.matrix %*% y), as.vector(mean.eval.apply), tolerance = 1e-10)

    expect_equal(drop(grad.train.matrix %*% y), as.vector(fit.train$grad[, 1L]), tolerance = 1e-8)
    expect_equal(as.vector(grad.train.apply), as.vector(fit.train$grad[, 1L]), tolerance = 1e-8)
    expect_equal(drop(grad.train.matrix %*% y), as.vector(grad.train.apply), tolerance = 1e-8)

    expect_equal(drop(grad.eval.matrix %*% y), as.vector(fit.eval$grad[, 1L]), tolerance = 1e-8)
    expect_equal(as.vector(grad.eval.apply), as.vector(fit.eval$grad[, 1L]), tolerance = 1e-8)
    expect_equal(drop(grad.eval.matrix %*% y), as.vector(grad.eval.apply), tolerance = 1e-8)
  }
})

test_that("fixed and generalized lc/ll matrix hats match apply and npreg on train and eval data", {
  set.seed(20260316)

  n <- 96L
  x <- sort(rnorm(n))
  y <- x^2 + 0.15 * x + rnorm(n, sd = 0.05)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 31L))

  for (regtype in c("lc", "ll")) {
    for (bwtype in c("fixed", "generalized_nn")) {
      bw <- npregbw(
        xdat = tx,
        ydat = y,
        regtype = regtype,
        bwtype = bwtype,
        bandwidth.compute = FALSE,
        bws = if (identical(bwtype, "fixed")) 0.55 else 9L
      )

      fit.train <- npreg(bws = bw, txdat = tx, tydat = y, gradients = identical(regtype, "ll"))
      fit.eval <- npreg(bws = bw, txdat = tx, tydat = y, exdat = ex, gradients = identical(regtype, "ll"))

      mean.train.matrix <- npreghat(bws = bw, txdat = tx, output = "matrix")
      mean.train.apply <- npreghat(bws = bw, txdat = tx, y = y, output = "apply")
      mean.eval.matrix <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix")
      mean.eval.apply <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply")

      expect_equal(drop(mean.train.matrix %*% y), as.vector(fit.train$mean), tolerance = 1e-10)
      expect_equal(as.vector(mean.train.apply), as.vector(fit.train$mean), tolerance = 1e-10)
      expect_equal(drop(mean.eval.matrix %*% y), as.vector(fit.eval$mean), tolerance = 1e-10)
      expect_equal(as.vector(mean.eval.apply), as.vector(fit.eval$mean), tolerance = 1e-10)

      if (identical(regtype, "ll")) {
        grad.train.matrix <- npreghat(bws = bw, txdat = tx, output = "matrix", s = 1L)
        grad.train.apply <- npreghat(bws = bw, txdat = tx, y = y, output = "apply", s = 1L)
        grad.eval.matrix <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix", s = 1L)
        grad.eval.apply <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply", s = 1L)

        expect_equal(drop(grad.train.matrix %*% y), as.vector(fit.train$grad[, 1L]), tolerance = 1e-8)
        expect_equal(as.vector(grad.train.apply), as.vector(fit.train$grad[, 1L]), tolerance = 1e-8)
        expect_equal(drop(grad.eval.matrix %*% y), as.vector(fit.eval$grad[, 1L]), tolerance = 1e-8)
        expect_equal(as.vector(grad.eval.apply), as.vector(fit.eval$grad[, 1L]), tolerance = 1e-8)
      }
    }
  }
})

test_that("fixed lp matrix hat matches apply and npreg across higher degrees", {
  set.seed(20260317)

  n <- 96L
  x <- sort(rnorm(n))
  y <- x^3 - 0.4 * x + rnorm(n, sd = 0.05)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 31L))

  for (degree in 2:3) {
    bw <- npregbw(
      xdat = tx,
      ydat = y,
      regtype = "lp",
      degree = degree,
      basis = "glp",
      bernstein.basis = FALSE,
      bwtype = "fixed",
      bandwidth.compute = FALSE,
      bws = 0.55
    )

    fit.train <- npreg(bws = bw, txdat = tx, tydat = y, gradients = TRUE)
    fit.eval <- npreg(bws = bw, txdat = tx, tydat = y, exdat = ex, gradients = TRUE)

    mean.train.matrix <- npreghat(bws = bw, txdat = tx, output = "matrix")
    mean.train.apply <- npreghat(bws = bw, txdat = tx, y = y, output = "apply")
    mean.eval.matrix <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix")
    mean.eval.apply <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply")

    grad.train.matrix <- npreghat(bws = bw, txdat = tx, output = "matrix", s = 1L)
    grad.train.apply <- npreghat(bws = bw, txdat = tx, y = y, output = "apply", s = 1L)
    grad.eval.matrix <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix", s = 1L)
    grad.eval.apply <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply", s = 1L)

    expect_equal(drop(mean.train.matrix %*% y), as.vector(fit.train$mean), tolerance = 1e-10)
    expect_equal(as.vector(mean.train.apply), as.vector(fit.train$mean), tolerance = 1e-10)
    expect_equal(drop(mean.eval.matrix %*% y), as.vector(fit.eval$mean), tolerance = 1e-10)
    expect_equal(as.vector(mean.eval.apply), as.vector(fit.eval$mean), tolerance = 1e-10)

    expect_equal(drop(grad.train.matrix %*% y), as.vector(fit.train$grad[, 1L]), tolerance = 1e-8)
    expect_equal(as.vector(grad.train.apply), as.vector(fit.train$grad[, 1L]), tolerance = 1e-8)
    expect_equal(drop(grad.eval.matrix %*% y), as.vector(fit.eval$grad[, 1L]), tolerance = 1e-8)
    expect_equal(as.vector(grad.eval.apply), as.vector(fit.eval$grad[, 1L]), tolerance = 1e-8)
  }
})

suppressPackageStartupMessages(library(np))

test_that("adaptive duplicate mass promotes zero-radius neighborhoods to the nearest positive radius", {
  x <- data.frame(x = c(0, 0, 0, 1))
  y <- c(0, 1, 2, 3)

  bw2 <- npregbw(
    ydat = y,
    xdat = x,
    bwtype = "adaptive_nn",
    bws = 2,
    bandwidth.compute = FALSE
  )
  bw3 <- npregbw(
    ydat = y,
    xdat = x,
    bwtype = "adaptive_nn",
    bws = 3,
    bandwidth.compute = FALSE
  )

  fit2 <- npreg(bws = bw2, exdat = x)
  fit3 <- npreg(bws = bw3, exdat = x)

  expect_equal(fit2$mean, fit3$mean, tolerance = 0)
})

test_that("adaptive observation-count bandwidths handle tied supports that formerly failed", {
  x <- data.frame(x = c(0, 1, 1, 1, 2))
  y <- c(0, 1, 2, 3, 4)

  bw <- npregbw(
    ydat = y,
    xdat = x,
    bwtype = "adaptive_nn",
    bws = 3,
    bandwidth.compute = FALSE
  )

  fit <- npreg(bws = bw, exdat = x)

  expect_length(fit$mean, nrow(x))
  expect_true(all(is.finite(fit$mean)))
})

test_that("generalized duplicate mass promotes zero-radius neighborhoods to the nearest positive radius", {
  x <- data.frame(x = c(0, 0, 0, 1))
  y <- c(0, 1, 2, 3)

  bw2 <- npregbw(
    ydat = y,
    xdat = x,
    bwtype = "generalized_nn",
    bws = 2,
    bandwidth.compute = FALSE
  )
  bw3 <- npregbw(
    ydat = y,
    xdat = x,
    bwtype = "generalized_nn",
    bws = 3,
    bandwidth.compute = FALSE
  )

  fit2 <- npreg(bws = bw2, exdat = x)
  fit3 <- npreg(bws = bw3, exdat = x)

  expect_equal(fit2$mean, fit3$mean, tolerance = 0)
})

test_that("generalized observation-count bandwidths handle low-support ties and out-of-support evaluation", {
  x <- data.frame(x = c(0, 0, 0, 1, 1, 2))
  y <- c(0, 1, 2, 3, 4, 5)
  ex <- data.frame(x = c(0, 0.5, 1, 1.5, 2))

  bw <- npregbw(
    ydat = y,
    xdat = x,
    bwtype = "generalized_nn",
    bws = 3,
    bandwidth.compute = FALSE
  )

  fit <- npreg(bws = bw, exdat = ex)

  expect_length(fit$mean, nrow(ex))
  expect_true(all(is.finite(fit$mean)))
})

test_that("npregbw formula and default paths preserve current metadata shape", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(42)
  dat <- data.frame(
    y = rnorm(12),
    x = runif(12)
  )

  bw_formula <- np::npregbw(
    y ~ x,
    data = dat,
    regtype = "lc",
    bwmethod = "cv.ls",
    bwtype = "fixed",
    nmulti = 1
  )
  bw_default <- np::npregbw(
    xdat = dat["x"],
    ydat = dat$y,
    regtype = "lc",
    bwmethod = "cv.ls",
    bwtype = "fixed",
    nmulti = 1
  )

  expect_s3_class(bw_formula, "rbandwidth")
  expect_s3_class(bw_default, "rbandwidth")

  expect_true(all(c("call", "formula", "terms", "nobs.omit") %in% names(bw_formula)))
  expect_false("rows.omit" %in% names(bw_formula))
  expect_false("formula" %in% names(bw_default))
  expect_false("terms" %in% names(bw_default))
  expect_true(all(c("call", "rows.omit", "nobs.omit") %in% names(bw_default)))

  expect_identical(as.character(bw_formula$call[[1L]]), "npregbw.formula")
  expect_identical(as.character(bw_default$call[[1L]]), "npregbw.NULL")
  expect_identical(attr(bw_formula$terms, "term.labels"), "x")

  expect_identical(bw_formula$nobs.omit, 0L)
  expect_identical(bw_default$rows.omit, NA)
  expect_identical(bw_default$nobs.omit, 0)

  expect_identical(bw_formula$varnames$x, "x")
  expect_identical(bw_formula$varnames$y, "y")
  expect_identical(bw_default$varnames$x, "x")
  expect_identical(bw_default$varnames$y, "dat$y")
})

test_that("npregbw formula path preserves omitted-row metadata", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  dat <- data.frame(
    y = c(1, 2, NA, 4, 5, 6),
    x = c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
  )

  bw <- np::npregbw(
    y ~ x,
    data = dat,
    regtype = "lc",
    bwmethod = "cv.ls",
    bwtype = "fixed",
    nmulti = 1
  )

  expect_identical(bw$rows.omit, 3L)
  expect_identical(bw$nobs.omit, 1L)
  expect_identical(bw$nobs, 5L)
})

test_that("npregbw rejects mismatched bandwidth object types with current message", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(1)
  y <- rnorm(8)
  numeric_bw <- np::npregbw(
    xdat = data.frame(x = runif(8)),
    ydat = y,
    regtype = "lc",
    bwtype = "fixed",
    bandwidth.compute = FALSE,
    bws = 0.4
  )

  expect_error(
    np::npregbw(
      xdat = data.frame(x = factor(letters[1:8])),
      ydat = y,
      bws = numeric_bw,
      bandwidth.compute = FALSE
    ),
    "supplied bandwidths do not match 'xdat' in type"
  )
})

test_that("npreghat reproduces npreg fitted values for mixed-data local constant", {
  set.seed(20260223)
  n <- 120
  x <- runif(n)
  u <- factor(sample(c("a", "b", "c"), n, replace = TRUE))
  o <- ordered(sample(1:3, n, replace = TRUE))
  y <- sin(2 * pi * x) + as.numeric(u) / 3 + as.numeric(o) / 5 + rnorm(n, sd = 0.05)

  tx <- data.frame(x = x, u = u, o = o)
  bw <- npregbw(
    xdat = tx,
    ydat = y,
    bws = c(0.2, 0.4, 0.4),
    regtype = "lc",
    bandwidth.compute = FALSE
  )

  fit <- npreg(txdat = tx, tydat = y, bws = bw, warn.glp.gradient = FALSE)
  H <- npreghat(bws = bw, txdat = tx)

  expect_s3_class(H, "npreghat")
  expect_equal(as.vector(H %*% y), as.vector(fit$mean), tolerance = 1e-8)
})

test_that("npreghat supports lp/ll derivatives and matrix apply mode", {
  set.seed(777)
  n <- 150
  x <- sort(runif(n))
  y <- cos(2 * pi * x) + rnorm(n, sd = 0.03)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 40))

  bw <- npregbw(
    xdat = tx,
    ydat = y,
    bws = 0.2,
    regtype = "ll",
    bandwidth.compute = FALSE
  )

  fit <- npreg(
    txdat = tx,
    tydat = y,
    exdat = ex,
    bws = bw,
    gradients = TRUE,
    gradient.order = 1L,
    warn.glp.gradient = FALSE
  )

  H0 <- npreghat(bws = bw, txdat = tx, exdat = ex)
  H1 <- npreghat(bws = bw, txdat = tx, exdat = ex, s = 1L)

  expect_equal(as.vector(H0 %*% y), as.vector(fit$mean), tolerance = 1e-8)
  expect_equal(as.vector(H1 %*% y), as.vector(fit$grad[, 1]), tolerance = 1e-6)

  yboot <- cbind(y, y + 0.1)
  hy.apply <- npreghat(
    bws = bw,
    txdat = tx,
    exdat = ex,
    y = yboot,
    output = "apply"
  )

  expect_true(isTRUE(all.equal(
    hy.apply,
    H0 %*% yboot,
    tolerance = 1e-10,
    check.attributes = FALSE
  )))
})

test_that("npreghat lc derivative matrix matches analytic npreg for fixed and generalized-nn", {
  set.seed(20260311)
  n <- 70L
  x <- sort(runif(n))
  y <- sin(2 * pi * x) + 0.3 * x + rnorm(n, sd = 0.04)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(0.1, 0.9, length.out = 20L))

  for (bwtype in c("fixed", "generalized_nn")) {
    bw <- npregbw(
      xdat = tx,
      ydat = y,
      regtype = "lc",
      bwtype = bwtype,
      bandwidth.compute = FALSE,
      bws = if (identical(bwtype, "fixed")) 0.2 else 9L
    )

    fit <- npreg(
      bws = bw,
      txdat = tx,
      tydat = y,
      exdat = ex,
      gradients = TRUE,
      warn.glp.gradient = FALSE
    )
    H <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix", s = 1L)
    grad.apply <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply", s = 1L)

    expect_equal(as.vector(H %*% y), as.vector(fit$grad[, 1]), tolerance = 1e-6)
    expect_equal(as.vector(H %*% y), as.vector(grad.apply), tolerance = 1e-8)
  }
})

test_that("npreghat leave.one.out is honored and predict reuses it safely", {
  set.seed(90210)
  n <- 80
  x <- runif(n)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.02)
  tx <- data.frame(x = x)
  bw <- npregbw(
    xdat = tx,
    ydat = y,
    bws = 0.18,
    regtype = "lc",
    bandwidth.compute = FALSE
  )

  H.in <- npreghat(bws = bw, txdat = tx, leave.one.out = FALSE)
  H.loo <- npreghat(bws = bw, txdat = tx, leave.one.out = TRUE)

  expect_gt(max(abs(H.in - H.loo)), 1e-8)
  expect_lt(max(abs(diag(H.loo))), 1e-12)
  expect_error(
    npreghat(bws = bw, txdat = tx, exdat = tx[1:10, , drop = FALSE], leave.one.out = TRUE),
    "you may not specify 'leave.one.out = TRUE' and provide evaluation data"
  )

  hy <- predict(H.loo, y = y, output = "apply")
  expect_equal(as.vector(hy), as.vector(H.loo %*% y), tolerance = 1e-10)
})

test_that("npreghat lp bernstein path matches predict semantics", {
  set.seed(20260305)
  n <- 220
  x <- sort(runif(n))
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.08)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(min(x), max(x), length.out = 60))

  bw.raw <- npregbw(
    xdat = tx,
    ydat = y,
    bws = 0.18,
    regtype = "lp",
    degree = 3L,
    bernstein = FALSE,
    bandwidth.compute = FALSE
  )
  bw.bern <- npregbw(
    xdat = tx,
    ydat = y,
    bws = 0.18,
    regtype = "lp",
    degree = 3L,
    bernstein = TRUE,
    bandwidth.compute = FALSE
  )

  fit.raw <- npreg(txdat = tx, tydat = y, exdat = ex, bws = bw.raw, gradients = FALSE, warn.glp.gradient = FALSE)
  fit.bern <- npreg(txdat = tx, tydat = y, exdat = ex, bws = bw.bern, gradients = FALSE, warn.glp.gradient = FALSE)
  grad.raw <- npreg(txdat = tx, tydat = y, exdat = ex, bws = bw.raw, gradients = TRUE, warn.glp.gradient = FALSE)
  grad.bern <- npreg(txdat = tx, tydat = y, exdat = ex, bws = bw.bern, gradients = TRUE, warn.glp.gradient = FALSE)

  hat.raw <- npreghat(bws = bw.raw, txdat = tx, exdat = ex, y = y, output = "apply", s = 0L)
  hat.bern <- npreghat(bws = bw.bern, txdat = tx, exdat = ex, y = y, output = "apply", s = 0L)
  dhat.raw <- npreghat(bws = bw.raw, txdat = tx, exdat = ex, y = y, output = "apply", s = 1L)
  dhat.bern <- npreghat(bws = bw.bern, txdat = tx, exdat = ex, y = y, output = "apply", s = 1L)

  expect_equal(as.vector(hat.raw), as.vector(fit.raw$mean), tolerance = 1e-8)
  expect_equal(as.vector(hat.bern), as.vector(fit.bern$mean), tolerance = 1e-8)
  expect_equal(as.vector(fit.bern$mean), as.vector(fit.raw$mean), tolerance = 1e-8)
  expect_equal(as.vector(dhat.raw), as.vector(grad.raw$grad[, 1L]), tolerance = 1e-8)
  expect_equal(as.vector(dhat.bern), as.vector(grad.bern$grad[, 1L]), tolerance = 1e-8)
  expect_equal(as.vector(grad.bern$grad[, 1L]), as.vector(grad.raw$grad[, 1L]), tolerance = 1e-8)
})

test_that("npreghat mean and supported gradients match npreg across bwtypes", {
  set.seed(20260308)
  n <- 90
  x <- sort(runif(n))
  y <- sin(2 * pi * x) + 0.25 * x + rnorm(n, sd = 0.04)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(0.05, 0.95, length.out = 25))

  make_bw <- function(regtype, bwtype) {
    bw_args <- list(
      xdat = tx,
      ydat = y,
      regtype = regtype,
      bwtype = bwtype,
      bandwidth.compute = FALSE,
      bws = if (identical(bwtype, "fixed")) 0.18 else 9
    )
    if (identical(regtype, "lp")) {
      bw_args$degree <- 1L
      bw_args$basis <- "glp"
      bw_args$bernstein.basis <- FALSE
    }
    do.call(npregbw, bw_args)
  }

  for (regtype in c("lc", "ll", "lp")) {
    for (bwtype in c("fixed", "generalized_nn", "adaptive_nn")) {
      bw <- make_bw(regtype, bwtype)

      fit.in <- npreg(
        bws = bw,
        txdat = tx,
        tydat = y,
        gradients = TRUE,
        warn.glp.gradient = FALSE
      )
      fit.ex <- npreg(
        bws = bw,
        txdat = tx,
        tydat = y,
        exdat = ex,
        gradients = TRUE,
        warn.glp.gradient = FALSE
      )
      pred.ex <- predict(fit.in, newdata = ex)

      hat.in.mean <- npreghat(bws = bw, txdat = tx, y = y, output = "apply")
      hat.ex.mean <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply")

      expect_equal(
        as.vector(hat.in.mean),
        as.vector(fit.in$mean),
        tolerance = 1e-8,
        info = paste("mean in-sample", regtype, bwtype)
      )
      expect_equal(
        as.vector(hat.ex.mean),
        as.vector(fit.ex$mean),
        tolerance = 1e-8,
        info = paste("mean exdat", regtype, bwtype)
      )
      expect_equal(
        as.vector(hat.ex.mean),
        as.vector(pred.ex),
        tolerance = 1e-8,
        info = paste("mean predict", regtype, bwtype)
      )

      hat.in.grad <- npreghat(bws = bw, txdat = tx, y = y, output = "apply", s = 1L)
      hat.ex.grad <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply", s = 1L)

      expect_equal(
        as.vector(hat.in.grad),
        as.vector(fit.in$grad[, 1]),
        tolerance = 1e-6,
        info = paste("grad in-sample", regtype, bwtype)
      )
      expect_equal(
        as.vector(hat.ex.grad),
        as.vector(fit.ex$grad[, 1]),
        tolerance = 1e-6,
        info = paste("grad exdat", regtype, bwtype)
      )
    }
  }
})

test_that("generalized-nn ll and canonical lp degree-1 share the same public/apply regression route", {
  set.seed(20260308)
  n <- 90
  x <- sort(runif(n))
  y <- sin(2 * pi * x) + 0.25 * x + rnorm(n, sd = 0.04)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(0.05, 0.95, length.out = 25))

  bw.ll <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "ll",
    bwtype = "generalized_nn",
    bws = 9,
    bandwidth.compute = FALSE
  )
  bw.lp <- npregbw(
    xdat = tx,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = 1L,
    bernstein.basis = FALSE,
    bwtype = "generalized_nn",
    bws = 9,
    bandwidth.compute = FALSE
  )

  fit.ll <- npreg(bws = bw.ll, txdat = tx, tydat = y, exdat = ex, warn.glp.gradient = FALSE)
  fit.lp <- npreg(bws = bw.lp, txdat = tx, tydat = y, exdat = ex, warn.glp.gradient = FALSE)

  hat.apply.ll <- npreghat(bws = bw.ll, txdat = tx, exdat = ex, y = y, output = "apply")
  hat.apply.lp <- npreghat(bws = bw.lp, txdat = tx, exdat = ex, y = y, output = "apply")
  hat.matrix.ll <- npreghat(bws = bw.ll, txdat = tx, exdat = ex)
  hat.matrix.lp <- npreghat(bws = bw.lp, txdat = tx, exdat = ex)

  expect_equal(as.vector(fit.ll$mean), as.vector(fit.lp$mean), tolerance = 1e-10)
  expect_equal(as.vector(hat.apply.ll), as.vector(hat.apply.lp), tolerance = 1e-10)
  expect_equal(as.vector(hat.apply.ll), as.vector(hat.matrix.ll %*% y), tolerance = 1e-10)
  expect_equal(as.vector(hat.apply.lp), as.vector(hat.matrix.lp %*% y), tolerance = 1e-10)
})

test_that("npreghat nonfixed higher-order lp operator matches npreg and matrix apply semantics", {
  make_case <- function(seed, bwtype) {
    set.seed(seed)
    n <- 28
    x <- data.frame(x = runif(n))
    y <- cos(2 * pi * x$x) + rnorm(n, sd = if (identical(bwtype, "generalized_nn")) 0.05 else 0.03)
    Y <- cbind(y, y^2)
    bw <- npregbw(
      xdat = x,
      ydat = y,
      regtype = "lp",
      degree = 2,
      basis = "tensor",
      bernstein.basis = TRUE,
      bwmethod = "cv.ls",
      bwtype = bwtype
    )
    fit <- npreg(bws = bw, txdat = x, tydat = y)
    H <- npreghat(bws = bw, txdat = x, output = "matrix")
    a.vec <- npreghat(bws = bw, txdat = x, y = y, output = "apply")
    a.mat <- npreghat(bws = bw, txdat = x, y = Y, output = "apply")
    list(fit = fit, H = H, a.vec = a.vec, a.mat = a.mat, y = y, Y = Y)
  }

  for (case in list(
    make_case(3, "generalized_nn"),
    make_case(5, "adaptive_nn")
  )) {
    expect_equal(as.vector(case$H %*% case$y), as.vector(case$fit$mean), tolerance = 1e-8)
    expect_equal(as.vector(case$a.vec), as.vector(case$fit$mean), tolerance = 1e-8)
    expect_equal(as.vector(case$H %*% case$y), as.vector(case$a.vec), tolerance = 1e-10)
    expect_equal(case$H %*% case$Y, case$a.mat, tolerance = 1e-10, ignore_attr = TRUE)
  }
})

test_that("adaptive-nn npreghat matrix owner matches apply and npreg on manual bandwidth routes", {
  set.seed(20260309)
  n <- 40L
  x <- sort(runif(n))
  y <- sin(2 * pi * x) + 0.25 * x + rnorm(n, sd = 0.04)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(0.05, 0.95, length.out = 12L))

  make_bw <- function(regtype) {
    bw_args <- list(
      xdat = tx,
      ydat = y,
      regtype = regtype,
      bwtype = "adaptive_nn",
      bandwidth.compute = FALSE,
      bws = 9L
    )
    if (identical(regtype, "lp")) {
      bw_args$degree <- 1L
      bw_args$basis <- "glp"
      bw_args$bernstein.basis <- FALSE
    }
    do.call(npregbw, bw_args)
  }

  for (regtype in c("lc", "ll", "lp")) {
    bw <- make_bw(regtype)
    fit <- npreg(
      bws = bw,
      txdat = tx,
      tydat = y,
      exdat = ex,
      gradients = !identical(regtype, "lc"),
      warn.glp.gradient = FALSE
    )
    hat.apply <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply")
    hat.matrix <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix")

    expect_equal(as.vector(hat.matrix %*% y), as.vector(hat.apply), tolerance = 1e-8)
    expect_equal(as.vector(hat.matrix %*% y), as.vector(fit$mean), tolerance = 1e-8)

    if (!identical(regtype, "lc")) {
      grad.apply <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply", s = 1L)
      grad.matrix <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix", s = 1L)

      expect_equal(as.vector(grad.matrix %*% y), as.vector(grad.apply), tolerance = 1e-8)
      expect_equal(as.vector(grad.matrix %*% y), as.vector(fit$grad[, 1]), tolerance = 1e-6)
    }
  }
})

test_that("adaptive-nn npreghat matrix owner matches apply and npreg on selected bandwidth routes", {
  set.seed(20260309)
  n <- 32L
  x <- sort(runif(n))
  y <- sin(2 * pi * x) + 0.25 * x + rnorm(n, sd = 0.04)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(0.05, 0.95, length.out = 10L))

  make_bw <- function(regtype) {
    bw_args <- list(
      xdat = tx,
      ydat = y,
      regtype = regtype,
      bwtype = "adaptive_nn",
      nmulti = 1L
    )
    if (identical(regtype, "lp")) {
      bw_args$degree <- 1L
      bw_args$basis <- "glp"
      bw_args$bernstein.basis <- FALSE
    }
    do.call(npregbw, bw_args)
  }

  for (regtype in c("lc", "ll", "lp")) {
    bw <- make_bw(regtype)
    fit <- npreg(
      bws = bw,
      txdat = tx,
      tydat = y,
      exdat = ex,
      gradients = !identical(regtype, "lc"),
      warn.glp.gradient = FALSE
    )
    hat.apply <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply")
    hat.matrix <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix")

    expect_equal(as.vector(hat.matrix %*% y), as.vector(hat.apply), tolerance = 1e-8)
    expect_equal(as.vector(hat.matrix %*% y), as.vector(fit$mean), tolerance = 1e-8)

    if (!identical(regtype, "lc")) {
      grad.apply <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "apply", s = 1L)
      grad.matrix <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix", s = 1L)

      expect_equal(as.vector(grad.matrix %*% y), as.vector(grad.apply), tolerance = 1e-8)
      expect_equal(as.vector(grad.matrix %*% y), as.vector(fit$grad[, 1]), tolerance = 1e-6)
    }
  }
})

test_that("npreghat constraint output is exact row-weighted transpose", {
  set.seed(20260513)
  n <- 36L
  x <- sort(runif(n))
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.05)
  tx <- data.frame(x = x)
  ex <- data.frame(x = seq(0.05, 0.95, length.out = 9L))

  make_bw <- function(regtype, bwtype) {
    args <- list(
      xdat = tx,
      ydat = y,
      regtype = regtype,
      bwtype = bwtype,
      bandwidth.compute = FALSE,
      bws = if (identical(bwtype, "fixed")) 0.25 else 7L
    )
    if (identical(regtype, "lp")) {
      args$degree <- 2L
      args$basis <- "glp"
      args$bernstein.basis <- FALSE
    }
    do.call(npregbw, args)
  }

  for (regtype in c("lc", "ll", "lp")) {
    for (bwtype in c("fixed", "generalized_nn", "adaptive_nn")) {
      bw <- make_bw(regtype, bwtype)
      H <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix")
      A <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y, output = "constraint")
      expect_equal(
        A,
        t(H) * y,
        tolerance = 0,
        ignore_attr = TRUE,
        info = paste("mean", regtype, bwtype)
      )

      if (!identical(regtype, "lc")) {
        H.grad <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix", s = 1L)
        A.grad <- npreghat(bws = bw, txdat = tx, exdat = ex, y = y,
                           output = "constraint", s = 1L)
        expect_equal(
          A.grad,
          t(H.grad) * y,
          tolerance = 0,
          ignore_attr = TRUE,
          info = paste("gradient", regtype, bwtype)
        )
      }
    }
  }

  bw <- make_bw("ll", "fixed")
  H <- npreghat(bws = bw, txdat = tx, exdat = ex, output = "matrix")
  expect_error(
    npreghat(bws = bw, txdat = tx, exdat = ex, output = "constraint"),
    "argument 'y' is required"
  )
  expect_error(
    npreghat(bws = bw, txdat = tx, exdat = ex, y = cbind(y, y),
             output = "constraint"),
    "one-column"
  )
  expect_error(
    npreghat(bws = bw, txdat = tx, exdat = ex, y = y[-1L],
             output = "constraint"),
    "must match"
  )

  H.obj <- npreghat(bws = bw, txdat = tx, output = "matrix")
  A.pred <- predict(H.obj, newdata = ex, y = y, output = "constraint")
  expect_equal(A.pred, t(H) * y, tolerance = 0, ignore_attr = TRUE)
})

test_that("default nmulti cap follows min(2, p) on public bandwidth routes", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260322)
  dat <- data.frame(
    y = rnorm(14),
    x1 = runif(14),
    x2 = runif(14)
  )

  bw_reg_uni <- np::npregbw(
    y ~ x1,
    data = dat,
    regtype = "lc",
    bwtype = "fixed",
    bwmethod = "cv.aic"
  )
  bw_reg_multi <- np::npregbw(
    y ~ x1 + x2,
    data = dat,
    regtype = "lc",
    bwtype = "fixed",
    bwmethod = "cv.aic"
  )
  bw_cd_multi <- np::npcdensbw(
    xdat = dat[c("x1", "x2")],
    ydat = dat["y"],
    regtype = "lc",
    bwtype = "fixed",
    bwmethod = "cv.ls"
  )

  expect_identical(length(bw_reg_uni$fval.history), 1L)
  expect_identical(length(bw_reg_multi$fval.history), 2L)
  expect_identical(length(bw_cd_multi$fval.history), 2L)
})

np_try_eval_in_frames <- getFromNamespace(".np_try_eval_in_frames", "np")
np_eval_bw_call <- getFromNamespace(".np_eval_bw_call", "np")
np_bw_call_uses_nomad_degree_search <- getFromNamespace(".np_bw_call_uses_nomad_degree_search", "np")
np_bw_dispatch_target <- getFromNamespace(".np_bw_dispatch_target", "np")
np_bw_formula_from_call <- getFromNamespace(".np_bw_formula_from_call", "np")

test_that(".np_try_eval_in_frames resolves symbols from eval env first", {
  env <- new.env(parent = emptyenv())
  env$x <- 11L
  out <- np_try_eval_in_frames(quote(x), eval_env = env)
  expect_true(out$ok)
  expect_identical(out$value, 11L)
})

test_that(".np_try_eval_in_frames preserves NULL symbol values", {
  env <- new.env(parent = emptyenv())
  env$x <- NULL
  out <- np_try_eval_in_frames(quote(x), eval_env = env)
  expect_true(out$ok)
  expect_null(out$value)
})

test_that(".np_try_eval_in_frames resolves symbols from caller frames", {
  val <- local({
    x <- 17L
    np_try_eval_in_frames(quote(x), eval_env = new.env(parent = emptyenv()))
  })
  expect_true(val$ok)
  expect_identical(val$value, 17L)
})

test_that(".np_try_eval_in_frames can disable caller frame fallback", {
  out <- local({
    x <- 19L
    np_try_eval_in_frames(quote(x), eval_env = new.env(parent = emptyenv()), search_frames = FALSE)
  })
  expect_false(out$ok)
  expect_true(inherits(out$error, "error"))
})

test_that(".np_try_eval_in_frames returns an error object when resolution fails", {
  out <- np_try_eval_in_frames(quote(np_missing_symbol_contract), eval_env = new.env(parent = emptyenv()))
  expect_false(out$ok)
  expect_true(inherits(out$error, "error"))
  expect_match(conditionMessage(out$error), "np_missing_symbol_contract")
})

test_that(".np_try_eval_in_frames evaluates non-symbol expressions in caller frames", {
  out <- local({
    x <- 4L
    np_try_eval_in_frames(quote(x + 1L), eval_env = new.env(parent = emptyenv()))
  })
  expect_true(out$ok)
  expect_identical(out$value, 5L)
})

test_that(".np_try_eval_in_frames can disable caller fallback for non-symbol expressions", {
  out <- local({
    x <- 4L
    np_try_eval_in_frames(
      quote(x + 1L),
      eval_env = new.env(parent = emptyenv()),
      search_frames = FALSE
    )
  })
  expect_false(out$ok)
  expect_true(inherits(out$error, "error"))
})

test_that(".np_try_eval_in_frames honors enclos for non-symbol expressions", {
  eval_env <- list()
  enclos <- new.env(parent = baseenv())
  enclos$z <- 8L
  out <- np_try_eval_in_frames(
    quote(z + 2L),
    eval_env = eval_env,
    enclos = enclos,
    search_frames = FALSE
  )
  expect_true(out$ok)
  expect_identical(out$value, 10L)
})

test_that(".np_eval_bw_call reports underlying evaluation errors", {
  call_obj <- as.call(list(as.name("identity"), as.name("np_missing_bw_arg_contract")))
  environment(call_obj) <- new.env(parent = baseenv())

  expect_error(
    np_eval_bw_call(call_obj, caller_env = new.env(parent = baseenv())),
    "np_missing_bw_arg_contract"
  )
})

test_that(".np_eval_bw_call evaluates successful calls in caller context", {
  out <- local({
    z <- 23L
    np_eval_bw_call(quote(identity(z)), caller_env = environment())
  })

  expect_identical(out, 23L)
})

test_that(".np_bw_call_uses_nomad_degree_search tolerates missing dummy formals", {
  out <- local({
    x1 <- rnorm(10)
    x2 <- rnorm(10)
    y <- x1 + rnorm(10)
    np_bw_call_uses_nomad_degree_search(
      quote(npregbw(y ~ x1 + x2, nmulti = 2)),
      caller_env = environment()
    )
  })

  expect_false(out)
})

test_that(".np_bw_call_uses_nomad_degree_search classifies supported cases narrowly", {
  out <- local({
    x1 <- rnorm(10)
    x2 <- rnorm(10)
    y <- x1 + rnorm(10)

    c(
      lc = np_bw_call_uses_nomad_degree_search(
        quote(npregbw(y ~ x1 + x2, regtype = "lc", nmulti = 2)),
        caller_env = environment()
      ),
      ll = np_bw_call_uses_nomad_degree_search(
        quote(npregbw(y ~ x1 + x2, regtype = "ll", nmulti = 2)),
        caller_env = environment()
      ),
      lp_manual = np_bw_call_uses_nomad_degree_search(
        quote(npregbw(y ~ x1 + x2, regtype = "lp", degree = c(1L, 1L), nmulti = 2)),
        caller_env = environment()
      ),
      lp_cell = np_bw_call_uses_nomad_degree_search(
        quote(npregbw(
          y ~ x1 + x2,
          regtype = "lp",
          degree.select = "coordinate",
          search.engine = "cell",
          nmulti = 2
        )),
        caller_env = environment()
      ),
      lp_nomad = np_bw_call_uses_nomad_degree_search(
        quote(npregbw(
          y ~ x1 + x2,
          regtype = "lp",
          degree.select = "coordinate",
          search.engine = "nomad",
          nmulti = 2
        )),
        caller_env = environment()
      ),
      lp_nomad_powell = np_bw_call_uses_nomad_degree_search(
        quote(npregbw(
          y ~ x1 + x2,
          regtype = "lp",
          degree.select = "exhaustive",
          search.engine = "nomad+powell",
          nmulti = 2
        )),
        caller_env = environment()
      )
    )
  })

  expect_false(out[["lc"]])
  expect_false(out[["ll"]])
  expect_false(out[["lp_manual"]])
  expect_false(out[["lp_cell"]])
  expect_true(out[["lp_nomad"]])
  expect_true(out[["lp_nomad_powell"]])
})

test_that(".np_bw_dispatch_target distinguishes formula, named-data, and bws entry shapes", {
  formula.out <- local({
    y <- rnorm(5)
    x <- rnorm(5)
    np_bw_dispatch_target(
      dots = list(quote(y ~ x)),
      data_arg_names = c("xdat", "ydat"),
      eval_env = environment()
    )
  })
  expect_true(inherits(formula.out, "formula"))

  data.out <- np_bw_dispatch_target(
    dots = list(xdat = quote(1:3), ydat = quote(4:6)),
    data_arg_names = c("xdat", "ydat"),
    eval_env = new.env(parent = baseenv())
  )
  expect_null(data.out)

  bw.out <- np_bw_dispatch_target(
    dots = list(bws = quote(structure(list(), class = "rbandwidth")), xdat = quote(1:3)),
    data_arg_names = c("xdat", "ydat"),
    eval_env = new.env(parent = baseenv())
  )
  expect_s3_class(bw.out, "rbandwidth")
})

test_that(".np_bw_dispatch_target resolves bws forwarded through dots", {
  bw <- structure(list(), class = "conbandwidth")

  forwarder <- function(...) {
    np_bw_dispatch_target(
      dots = list(xdat = quote(xdat), ydat = quote(ydat), bws = quote(bws)),
      data_arg_names = c("xdat", "ydat"),
      eval_env = environment()
    )
  }

  xdat <- data.frame(x = 1:3)
  ydat <- data.frame(y = 4:6)
  out <- forwarder(xdat = xdat, ydat = ydat, bws = bw)

  expect_s3_class(out, "conbandwidth")
})

test_that(".np_bw_formula_from_call extracts the formula argument call when present", {
  out <- np_bw_formula_from_call(quote(npregbw(y ~ x1 + x2, nmulti = 2)))
  expect_true(is.call(out))
  expect_identical(deparse(out), "y ~ x1 + x2")

  expect_null(np_bw_formula_from_call(quote(npregbw(xdat = x1, ydat = y, nmulti = 2))))
})

test_that("npreg formula newdata path matches explicit data path", {
  set.seed(20260222)
  dat <- data.frame(
    y = rnorm(32),
    x = runif(32)
  )
  ex <- data.frame(x = seq(0.1, 0.9, length.out = 9))

  bw <- np::npregbw(
    y ~ x,
    data = dat,
    bws = 0.45,
    bandwidth.compute = FALSE,
    regtype = "lc"
  )

  fit_formula <- np::npreg(bws = bw, data = dat, newdata = ex)
  fit_default <- np::npreg(
    bws = bw,
    txdat = dat["x"],
    tydat = dat$y,
    exdat = ex["x"]
  )

  expect_equal(fit_formula$mean, fit_default$mean)
  expect_equal(fit_formula$merr, fit_default$merr)
})

test_that("npreg formula y.eval path matches explicit data path", {
  set.seed(20260222)
  dat <- data.frame(
    y = rnorm(36),
    x = runif(36)
  )
  ex <- data.frame(
    y = rnorm(12),
    x = seq(0.1, 0.9, length.out = 12)
  )

  bw <- np::npregbw(
    y ~ x,
    data = dat,
    bws = 0.4,
    bandwidth.compute = FALSE,
    regtype = "lc"
  )

  fit_formula <- np::npreg(bws = bw, data = dat, newdata = ex, y.eval = TRUE)
  fit_default <- np::npreg(
    bws = bw,
    txdat = dat["x"],
    tydat = dat$y,
    exdat = ex["x"],
    eydat = ex$y
  )

  expect_equal(fit_formula$mean, fit_default$mean)
  expect_equal(fit_formula$merr, fit_default$merr)
})

np_prepare_nomad_shortcut <- getFromNamespace(".np_prepare_nomad_shortcut", "np")
np_bw_call_uses_nomad_degree_search <- getFromNamespace(".np_bw_call_uses_nomad_degree_search", "np")

test_that(".np_prepare_nomad_shortcut fills only missing preset arguments", {
  out <- np_prepare_nomad_shortcut(
    nomad = TRUE,
    call_names = c("xdat", "ydat", "nomad", "degree.max"),
    preset = list(
      regtype = "lp",
      search.engine = "nomad+powell",
      degree.select = "coordinate",
      bernstein.basis = TRUE,
      degree.min = 0L,
      degree.max = 10L,
      degree.verify = FALSE,
      bwtype = "fixed"
    ),
    values = list(
      regtype = NULL,
      search.engine = NULL,
      degree.select = NULL,
      bernstein.basis = NULL,
      degree.min = NULL,
      degree.max = 1L,
      degree.verify = NULL,
      bwtype = NULL
    ),
    where = "npregbw"
  )

  expect_true(out$enabled)
  expect_identical(out$values$degree.max, 1L)
  expect_identical(out$values$regtype, "lp")
  expect_identical(out$metadata$user.supplied, "degree.max")
  expect_true(all(c("regtype", "search.engine", "degree.select", "bernstein.basis",
                    "degree.min", "degree.verify", "bwtype") %in%
                    out$metadata$auto.filled))
})

test_that("nomad shortcut marks automatic bandwidth calls for progress routing", {
  out <- local({
    x <- rnorm(10)
    y <- x + rnorm(10)
    np_bw_call_uses_nomad_degree_search(
      quote(npreg(y ~ x, nomad = TRUE, degree.max = 1L, nmulti = 1L)),
      caller_env = environment()
    )
  })

  expect_true(out)
})

test_that("npregbw nomad shortcut matches the explicit regression preset", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260322)
  dat <- data.frame(x = sort(runif(14)))
  dat$y <- sin(2 * pi * dat$x) + rnorm(nrow(dat), sd = 0.05)

  bw_short <- np::npregbw(
    y ~ x,
    data = dat,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L
  )

  bw_long <- np::npregbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    search.engine = "nomad+powell",
    degree.select = "coordinate",
    bernstein.basis = TRUE,
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = FALSE,
    bwtype = "fixed",
    nmulti = 1L
  )

  expect_s3_class(bw_short, "rbandwidth")
  expect_equal(unname(bw_short$bw), unname(bw_long$bw), tolerance = 1e-8)
  expect_identical(as.integer(bw_short$degree), as.integer(bw_long$degree))
  expect_identical(bw_short$degree.search$mode, bw_long$degree.search$mode)
  expect_true(is.list(bw_short$nomad.shortcut))
  expect_identical(bw_short$nomad.shortcut$user.supplied, "degree.max")
  expect_true(all(c("regtype", "search.engine", "degree.select", "bernstein.basis",
                    "degree.min", "degree.verify", "bwtype") %in%
                    bw_short$nomad.shortcut$auto.filled))
})

test_that("npreg and npregbw accept nomad.nmulti through the NOMAD shortcut route", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260323)
  dat <- data.frame(x = sort(runif(14)))
  dat$y <- sin(2 * pi * dat$x) + rnorm(nrow(dat), sd = 0.05)

  bw <- np::npregbw(
    y ~ x,
    data = dat,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L,
    nomad.nmulti = 2L
  )
  fit <- np::npreg(
    y ~ x,
    data = dat,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L,
    nomad.nmulti = 2L
  )

  expect_s3_class(bw, "rbandwidth")
  expect_true(is.list(fit$bws$nomad.shortcut))
  expect_identical(as.integer(fit$bws$degree), as.integer(bw$degree))
})

test_that("npreg nomad shortcut matches the explicit regression preset", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260322)
  dat <- data.frame(x = sort(runif(14)))
  dat$y <- sin(2 * pi * dat$x) + rnorm(nrow(dat), sd = 0.05)

  fit_short <- np::npreg(
    y ~ x,
    data = dat,
    nomad = TRUE,
    degree.max = 1L,
    nmulti = 1L
  )

  fit_long <- np::npreg(
    y ~ x,
    data = dat,
    regtype = "lp",
    search.engine = "nomad+powell",
    degree.select = "coordinate",
    bernstein.basis = TRUE,
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = FALSE,
    bwtype = "fixed",
    nmulti = 1L
  )

  expect_equal(fitted(fit_short), fitted(fit_long), tolerance = 1e-8)
  expect_identical(as.integer(fit_short$bws$degree), as.integer(fit_long$bws$degree))
  expect_true(is.list(fit_short$bws$nomad.shortcut))
})

test_that("npregbw nomad shortcut fails fast on incompatible explicit settings", {
  x <- data.frame(x = seq(0, 1, length.out = 8))
  y <- sin(2 * pi * x$x)

  expect_error(
    np::npregbw(xdat = x, ydat = y, nomad = TRUE, regtype = "ll"),
    "nomad=TRUE requires regtype='lp'"
  )

  expect_error(
    np::npregbw(xdat = x, ydat = y, nomad = TRUE, bwtype = "adaptive_nn"),
    "nomad=TRUE currently requires bwtype='fixed'"
  )

  expect_error(
    np::npregbw(xdat = x, ydat = y, nomad = TRUE, degree = 1L),
    "does not support an explicit degree"
  )

  expect_error(
    np::npregbw(xdat = x, ydat = y, nomad = TRUE, bernstein.basis = FALSE),
    "requires bernstein.basis=TRUE"
  )

  expect_error(
    np::npregbw(xdat = x, ydat = y, nomad = TRUE, search.engine = "cell"),
    "requires search.engine='nomad' or 'nomad\\+powell'"
  )

  expect_error(
    np::npregbw(xdat = x, ydat = y, nomad.nmulti = 1L),
    "nomad.nmulti is only supported when regtype='lp', automatic degree search is active, and search.engine is 'nomad' or 'nomad\\+powell'"
  )

  expect_error(
    np::npregbw(
      xdat = x,
      ydat = y,
      regtype = "lp",
      degree.select = "coordinate",
      search.engine = "cell",
      degree.min = 0L,
      degree.max = 1L,
      nomad.nmulti = 1L
    ),
    "nomad.nmulti is only supported when regtype='lp', automatic degree search is active, and search.engine is 'nomad' or 'nomad\\+powell'"
  )
})

test_that("npreg basic functionality works", {
  data("cps71")
  # Use a small subset for speed
  cps71_sub <- cps71[1:100, ]
  
  # Local constant
  bw_lc <- npregbw(xdat=cps71_sub$age, ydat=cps71_sub$logwage, 
                    bws=1.0, bandwidth.compute=FALSE, regtype="lc")
  model_lc <- npreg(bws=bw_lc)
  
  expect_s3_class(model_lc, "npregression")
  expect_type(predict(model_lc), "double")
  expect_equal(length(predict(model_lc)), 100)
  
  # Local linear
  bw_ll <- npregbw(xdat=cps71_sub$age, ydat=cps71_sub$logwage, 
                    bws=1.0, bandwidth.compute=FALSE, regtype="ll")
  model_ll <- npreg(bws=bw_ll)
  
  expect_s3_class(model_ll, "npregression")
  expect_type(predict(model_ll), "double")
})

test_that("npreg works with formula and mixed data", {
  data("Italy")
  Italy_sub <- Italy[1:50, ]
  
  bw <- npregbw(gdp~ordered(year), data=Italy_sub, 
                bws=0.5, bandwidth.compute=FALSE)
  model <- npreg(bws=bw)
  
  expect_s3_class(model, "npregression")
  expect_output(summary(model))
})

test_that("npregbw cv.ls handles all-categorical fixed bandwidths", {
  old_rng <- RNGkind()
  on.exit(do.call(RNGkind, as.list(old_rng)), add = TRUE)
  RNGkind("Mersenne-Twister", "Inversion", "Rejection")
  set.seed(5000001L)

  n <- 200L
  xbar <- factor(rbinom(n, 1L, 0.5), levels = 0:1)
  xtilde <- factor(rbinom(n, 1L, 0.5), levels = 0:1)
  y <- ifelse(as.integer(as.character(xbar)) == 1L, 1, 0) + rnorm(n)

  bw <- npregbw(
    xdat = data.frame(xbar = xbar, xtilde = xtilde),
    ydat = y,
    bwmethod = "cv.ls",
    bwscaling = FALSE,
    bwtype = "fixed",
    ukertype = "aitchisonaitken"
  )

  expect_equal(unname(bw$bw),
               c(0.016073908490923, 0.499999933944028),
               tolerance = 1e-10)
  expect_equal(unname(bw$fval), 1.0135341078806, tolerance = 1e-10)
  expect_true(is.finite(unname(bw$num.feval)))
  expect_gt(unname(bw$num.feval), 0)
})

library(np)

test_that("lc first-derivative owner matches npreg and apply on common cells", {
  run_case <- function(bwtype, bwval, tol = 1e-10) {
    train <- data.frame(x = x)
    frame <- data.frame(train, y = y)
    bw <- npregbw(y ~ x,
                  data = frame,
                  regtype = "lc",
                  bwtype = bwtype,
                  bandwidth.compute = FALSE,
                  bws = bwval)

    H.train <- npreghat(bws = bw, txdat = train, s = 1L)
    H.eval <- npreghat(bws = bw, txdat = train, exdat = eval, s = 1L)
    a.train <- npreghat(bws = bw, txdat = train, y = y, s = 1L, output = "apply")
    a.eval <- npreghat(bws = bw, txdat = train, exdat = eval, y = y, s = 1L, output = "apply")
    g.train <- npreg(bws = bw, gradients = TRUE)
    g.eval <- npreg(bws = bw, exdat = eval, gradients = TRUE)

    expect_equal(drop(H.train %*% y), a.train, tolerance = tol)
    expect_equal(drop(H.eval %*% y), a.eval, tolerance = tol)
    expect_equal(drop(H.train %*% y), g.train$grad[, 1L], tolerance = tol)
    expect_equal(drop(H.eval %*% y), g.eval$grad[, 1L], tolerance = tol)
  }

  set.seed(410)
  n <- 60L
  x <- rnorm(n)
  y <- x^2 + rnorm(n, sd = 0.2)
  eval <- data.frame(x = seq(min(x), max(x), length.out = 31L))

  run_case("fixed", 0.45)
  run_case("generalized_nn", 9)
  run_case("adaptive_nn", 9)
})

.np_plreg_bws <- function(object, where = "plregression") {
  if (!is.null(object$bws))
    return(object$bws)

  if (!is.null(object$bw) && inherits(object$bw, "plbandwidth"))
    return(object$bw)

  stop(sprintf("%s object does not contain a usable partially linear bandwidth object", where),
       call. = FALSE)
}

plregression = 
  function(bws, xcoef, xcoeferr = 0, xcoefvcov, evalx, evalz, mean, resid = NA,
           ntrain, trainiseval = FALSE, residuals = FALSE,
           xtra = double(6),
           timing = NA, total.time = NA,
           optim.time = NA, fit.time = NA){

    if (missing(bws) || missing(evalx) || missing(evalz) || missing(mean) || missing(ntrain) || missing(xcoef))
      stop("improper invocation of plregression constructor")

    d = list(
      bws = bws,
      bw = bws,
      xcoef = xcoef,
      xcoeferr = xcoeferr,
      xcoefvcov = xcoefvcov,      
      pregtype = bws$pregtype,
      data.znames = names(evalz),
      data.xnames = names(evalx),
      nobs = dim(evalz)[1],
      zndim = dim(evalz)[2],
      xndim = dim(evalx)[2],
      pscaling = bws$pscaling,
      ptype = bws$ptype,
      pckertype = bws$pckertype,
      pukertype = bws$pukertype,
      pokertype = bws$pokertype,
      evalx = evalx,
      evalz = evalz,
      mean = mean,
      resid = resid,
      ntrain = ntrain,
      trainiseval = trainiseval,
      residuals = residuals,
      R2 = xtra[1],
      MSE = xtra[2],
      MAE = xtra[3],
      MAPE = xtra[4],
      CORR = xtra[5],
      SIGN = xtra[6],
      timing = timing,
      total.time = total.time,
      optim.time = optim.time,
      fit.time = fit.time
      )

    names(d$xcoeferr) <- names(d$xcoef) <- d$data.xnames
    dimnames(d$xcoefvcov) <- list(d$data.xnames, d$data.xnames)
    
    class(d) = "plregression"
    
    d
  }

print.plregression <- function(x, digits=NULL, ...){
  obj.bws <- .np_plreg_bws(x, where = "print.plregression")

  cat("\nPartially Linear Model",
      "\nRegression data: ", x$ntrain, " training points,",
      if (x$trainiseval) "" else paste(" and ", x$nobs, " evaluation points,",
                                      sep = ""),
      " in ",(x$zndim+x$xndim)," variable(s)",
      "\nWith ", x$xndim, " linear parametric regressor(s), ",
      x$zndim, " nonparametric regressor(s)\n\n", sep="")

  bwmat = matrix(data = 0, nrow = x$xndim+1, ncol = obj.bws$bw$yzbw$ndim)
  
  for (i in seq_along(obj.bws$bw))
    bwmat[i,] = obj.bws$bw[[i]]$bw
  
  print(matrix(bwmat[1,], ncol=x$zndim,
               dimnames=list(paste(x$pscaling,":",sep=""),
                 c("y(z)", replicate(x$zndim-1,"")))))
  cat("\n")
  print(matrix(bwmat[2:(1+x$xndim),], ncol=x$zndim,
               dimnames=list(c(paste(x$pscaling,":",sep=""), replicate(x$xndim-1,"")),
                 c("x(z)", replicate(x$zndim-1,"")))))

  print(matrix(x$xcoef,ncol=x$xndim,
               dimnames=list("Coefficient(s):",x$data.xnames)))
  
  cat(genRegEstStr(x))

  cat(genBwKerStrs(obj.bws))
  cat('\n\n')  

  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

coef.plregression <- function(object, errors = FALSE, ...) {
  errors <- npValidateScalarLogical(errors, "errors")
  if(!errors)
    return(object$xcoef)
  else
    return(object$xcoeferr)
}

vcov.plregression <- function(object,...) {
  return(object$xcoefvcov)
}

fitted.plregression <- function(object, ...){
 object$mean 
}
residuals.plregression <- function(object, ...) {
 if(object$residuals) {
   return(object$resid)
 } else {
   return(npplreg(bws = .np_plreg_bws(object, where = "residuals.plregression"),
                  residuals = TRUE)$resid)
 }
}

.np_plreg_predict_se_block_size <- function(ntrain) {
  ntrain <- max(1L, as.integer(ntrain)[1L])
  max(1L, min(512L, floor(5.0e7 / ntrain)))
}

.np_plreg_predict_train_data <- function(bws) {
  if (!is.null(bws$formula)) {
    tt <- terms(bws)
    tt.xf <- bws$xterms

    m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)
    tmf <- bws$call[c(1, m)]
    tmf.xf <- bws$call[c(1, m)]

    tmf[[1L]] <- as.name("model.frame")
    tmf.xf[[1L]] <- as.name("model.frame")
    tmf[["formula"]] <- tt
    tmf.xf[["formula"]] <- tt.xf

    mf.args <- as.list(tmf)[-1L]
    mf.xf.args <- as.list(tmf.xf)[-1L]
    train.mf <- do.call(stats::model.frame, mf.args, envir = environment(tt))
    train.xf <- do.call(stats::model.frame, mf.xf.args, envir = environment(tt.xf))

    list(
      txdat = train.xf,
      tydat = model.response(train.mf),
      tzdat = train.mf[, bws$chromoly[[3L]], drop = FALSE]
    )
  } else {
    list(
      txdat = toFrame(.np_eval_bws_call_arg(bws, "xdat")),
      tydat = .np_eval_bws_call_arg(bws, "ydat"),
      tzdat = toFrame(.np_eval_bws_call_arg(bws, "zdat"))
    )
  }
}

.np_plreg_predict_se_data <- function(bws, fit) {
  train <- .np_plreg_predict_train_data(bws)
  txdat <- toFrame(train$txdat)
  tydat <- train$tydat
  tzdat <- toFrame(train$tzdat)

  keep.rows <- rep_len(TRUE, nrow(txdat))
  rows.omit <- attr(na.omit(data.frame(txdat, tydat, tzdat)), "na.action")
  if (length(rows.omit) > 0L)
    keep.rows[as.integer(rows.omit)] <- FALSE

  if (!any(keep.rows))
    stop("Training data has no rows without NAs")

  txdat <- txdat[keep.rows, , drop = FALSE]
  tydat <- tydat[keep.rows]
  tzdat <- tzdat[keep.rows, , drop = FALSE]

  exdat <- toFrame(fit$evalx)
  ezdat <- toFrame(fit$evalz)

  list(txdat = txdat, tydat = tydat, tzdat = tzdat, exdat = exdat, ezdat = ezdat)
}

.np_plreg_predict_se <- function(bws, fit) {
  dat <- .np_plreg_predict_se_data(bws = bws, fit = fit)
  train.fit <- npplreg(bws = bws, residuals = TRUE)
  u <- as.numeric(residuals(train.fit))
  if (length(u) != nrow(dat$txdat))
    stop("internal error: residual length does not match training rows")

  u2 <- u^2
  neval <- nrow(dat$exdat)
  block.size <- .np_plreg_predict_se_block_size(nrow(dat$txdat))
  se.fit <- numeric(neval)
  starts <- seq.int(1L, neval, by = block.size)

  for (st in starts) {
    en <- min(neval, st + block.size - 1L)
    ii <- st:en
    H <- npplreghat(
      bws = bws,
      txdat = dat$txdat,
      tzdat = dat$tzdat,
      exdat = dat$exdat[ii, , drop = FALSE],
      ezdat = dat$ezdat[ii, , drop = FALSE],
      output = "matrix"
    )
    se.fit[ii] <- sqrt(pmax(drop((H^2) %*% u2), 0.0))
  }

  se.fit
}

predict.plregression <- function(object, se.fit = FALSE, ...) {
  se.fit <- npValidateScalarLogical(se.fit, "se.fit")
  obj.bws <- .np_plreg_bws(object, where = "predict.plregression")
  dots <- list(...)
  has.formula.route <- !is.null(obj.bws$formula)

  if ((!is.null(dots$exdat) || !is.null(dots$ezdat)) && !is.null(dots$newdata)) {
    dots$newdata <- NULL
  } else if (!has.formula.route && is.null(dots$exdat) && !is.null(dots$newdata)) {
    nd <- toFrame(dots$newdata)
    need <- c(obj.bws$xnames, obj.bws$znames)
    if (!all(need %in% names(nd)))
      stop("'newdata' must include columns: ", paste(need, collapse = ", "))
    dots$exdat <- nd[, obj.bws$xnames, drop = FALSE]
    dots$ezdat <- nd[, obj.bws$znames, drop = FALSE]
    dots$newdata <- NULL
  }

  tr <- do.call(npplreg, c(list(bws = obj.bws), dots))
  fit <- fitted(tr)

  if (se.fit) {
    se.out <- .np_plreg_predict_se(bws = obj.bws, fit = tr)
    return(list(fit = fit, se.fit = se.out))
  }

  fit
}
summary.plregression <- function(object, ...){
  obj.bws <- .np_plreg_bws(object, where = "summary.plregression")

  cat("\nPartially Linear Model",
      "\nRegression data: ", object$ntrain, " training points,",
      if (object$trainiseval) "" else paste(" and ", object$nobs, " evaluation points,",
                                      sep = ""),
      " in ",(object$zndim+object$xndim)," variable(s)",
      "\nWith ", object$xndim, " linear parametric regressor(s), ",
      object$zndim, " nonparametric regressor(s)\n", sep="")

  cat(genOmitStr(object))
  cat("\n")

  bwmat = matrix(data = 0, nrow = object$xndim+1, ncol = obj.bws$bw$yzbw$ndim)
  
  for (i in seq_along(obj.bws$bw))
    bwmat[i,] = obj.bws$bw[[i]]$bw
  
  print(matrix(bwmat[1,], ncol=object$zndim,
               dimnames=list(paste(object$pscaling,":",sep=""),
                 c("y(z)", replicate(object$zndim-1,"")))))
  cat("\n")
  print(matrix(bwmat[2:(1+object$xndim),], ncol=object$zndim,
               dimnames=list(c(paste(object$pscaling,":",sep=""), replicate(object$xndim-1,"")),
                 c("x(z)", replicate(object$zndim-1,"")))))
  cat("\n")

  print(matrix(object$xcoef,ncol=object$xndim,
               dimnames=list("Coefficient(s):",object$data.xnames)))
  
  cat(genRegEstStr(object))
  cat("\n")
  cat(genGofStr(object))

  cat(genBwKerStrs(obj.bws))
  cat(genTimingStr(object))
  cat('\n\n')  
}  

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
options(np.messages = FALSE)

## -----------------------------------------------------------------------------
library(np)
data(cps71, package = "np")

bw <- npregbw(logwage ~ age, data = cps71)
summary(bw)

fit <- npreg(bws = bw)
summary(fit)

## ----fig.width = 6, fig.height = 4--------------------------------------------
plot(cps71$age, cps71$logwage, cex = 0.25, col = "grey")
lines(cps71$age, fitted(fit), col = 2, lwd = 2)

## -----------------------------------------------------------------------------
set.seed(42)
mydat <- data.frame(
  y = rnorm(200),
  x_cont = runif(200),
  x_unordered = factor(sample(c("a", "b", "c"), 200, replace = TRUE)),
  x_ordered = ordered(sample(1:4, 200, replace = TRUE))
)

bw_mixed <- npregbw(y ~ x_cont + x_unordered + x_ordered, data = mydat)
fit_mixed <- npreg(bws = bw_mixed)
summary(fit_mixed)

## -----------------------------------------------------------------------------
if (requireNamespace("crs", quietly = TRUE) &&
    utils::packageVersion("crs") >= package_version("0.15-41")) {
  set.seed(7)
  n <- 120
  x <- runif(n, -1, 1)
  y <- x + 0.4 * x^2 + rnorm(n, sd = 0.18)

  fit_nomad <- npreg(y ~ x, nomad = TRUE, degree.max = 1L, nmulti = 1L)
  fit_nomad$bws$nomad.shortcut

  # Tune one component explicitly while leaving the rest of the preset in place.
  fit_nomad_direct <- npreg(
    y ~ x,
    nomad = TRUE,
    search.engine = "nomad",
    degree.max = 1L,
    nmulti = 1L
  )
}

.np_missing_binding_sentinel <- new.env(parent = emptyenv())

.np_try_eval_in_frames <- function(expr, eval_env = parent.frame(), enclos = NULL, search_frames = TRUE) {
  if (!is.language(expr))
    return(list(ok = TRUE, value = expr, error = NULL))

  sym <- NULL
  not_found <- .np_missing_binding_sentinel
  if (is.symbol(expr)) {
    sym <- as.character(expr)
  }

  if (!is.null(sym) && is.environment(eval_env)) {
    get_dot_value <- function(env) {
      dots <- tryCatch(
        eval(substitute(list(...)), envir = env),
        error = function(e) not_found
      )
      if (identical(dots, not_found) || !is.list(dots))
        return(not_found)
      dot_names <- names(dots)
      if (is.null(dot_names) || !(sym %in% dot_names))
        return(not_found)
      dots[[which(dot_names == sym)[1L]]]
    }
    eval_symbol <- function(env) {
      tryCatch(eval(expr, envir = env), error = function(e) not_found)
    }

    sym_val <- get0(sym, envir = eval_env, inherits = TRUE, ifnotfound = not_found)
    if (!identical(sym_val, not_found))
      return(list(ok = TRUE, value = sym_val, error = NULL))
    sym_val <- get_dot_value(eval_env)
    if (!identical(sym_val, not_found))
      return(list(ok = TRUE, value = sym_val, error = NULL))
    sym_val <- eval_symbol(eval_env)
    if (!identical(sym_val, not_found))
      return(list(ok = TRUE, value = sym_val, error = NULL))

    first_error <- simpleError(sprintf("object '%s' not found", sym))
    if (!isTRUE(search_frames))
      return(list(ok = FALSE, value = NULL, error = first_error))

    frames <- sys.frames()
    for (i in rev(seq_along(frames))) {
      env_i <- frames[[i]]
      if (identical(env_i, eval_env))
        next
      if (is.environment(env_i)) {
        sym_val <- get0(sym, envir = env_i, inherits = TRUE, ifnotfound = not_found)
        if (!identical(sym_val, not_found))
          return(list(ok = TRUE, value = sym_val, error = NULL))
        sym_val <- get_dot_value(env_i)
        if (!identical(sym_val, not_found))
          return(list(ok = TRUE, value = sym_val, error = NULL))
        sym_val <- eval_symbol(env_i)
        if (!identical(sym_val, not_found))
          return(list(ok = TRUE, value = sym_val, error = NULL))
      }
    }

    return(list(ok = FALSE, value = NULL, error = first_error))
  }

  eval_once <- if (is.null(enclos)) {
    # Intentional dynamic evaluation: selector/formula NSE semantics require
    # frame-aware call evaluation across caller frames.
    # Guarded by helper contracts in tests/testthat:
    # - test-call-eval-helpers-contract.R
    # - test-bw-eval-helper-contract.R
    function(env) tryCatch(eval(expr, envir = env), error = function(e) e)
  } else {
    # Same contract path when an explicit enclos is supplied.
    function(env) tryCatch(eval(expr, envir = env, enclos = enclos), error = function(e) e)
  }

  val <- eval_once(eval_env)
  if (!inherits(val, "error"))
    return(list(ok = TRUE, value = val, error = NULL))

  first_error <- val
  if (!isTRUE(search_frames))
    return(list(ok = FALSE, value = NULL, error = first_error))

  frames <- sys.frames()
  for (i in rev(seq_along(frames))) {
    env_i <- frames[[i]]
    if (identical(env_i, eval_env))
      next
    val_i <- eval_once(env_i)
    if (!inherits(val_i, "error"))
      return(list(ok = TRUE, value = val_i, error = NULL))
  }

  list(ok = FALSE, value = NULL, error = first_error)
}

.np_eval_bw_call <- function(call_obj, caller_env = parent.frame()) {
  if (!is.call(call_obj))
    stop("bandwidth selector call is malformed", call. = FALSE)

  fallback <- .np_try_eval_in_frames(call_obj, eval_env = caller_env)
  if (isTRUE(fallback$ok))
    return(fallback$value)

  if (inherits(fallback$error, "error"))
    stop(conditionMessage(fallback$error), call. = FALSE)
  stop("bandwidth selector call evaluation failed", call. = FALSE)
}

.np_bw_call_uses_nomad_degree_search <- function(call_obj, caller_env = parent.frame()) {
  if (!is.call(call_obj) || length(call_obj) < 1L)
    return(FALSE)

  fn_expr <- call_obj[[1L]]
  fn <- .np_try_eval_in_frames(fn_expr, eval_env = caller_env)
  if (!isTRUE(fn$ok) || !is.function(fn$value))
    return(FALSE)

  fn_def <- fn$value
  fn_name <- if (is.symbol(fn_expr)) as.character(fn_expr) else NULL
  if (!is.null(fn_name) && nzchar(fn_name)) {
    default_name <- paste0(fn_name, ".default")
    default_fn <- get0(default_name, envir = environment(fn_def), inherits = TRUE)
    if (is.function(default_fn))
      fn_def <- default_fn
  }

  matched <- tryCatch(
    match.call(definition = fn_def, call = call_obj, expand.dots = FALSE),
    error = function(e) NULL
  )
  if (is.null(matched))
    return(FALSE)

  defaults <- formals(fn_def)
  is_missing_arg <- function(z) missing(z)
  arg_expr <- function(source, name) {
    if (is.null(source) || is.null(names(source)) || !(name %in% names(source)))
      return(NULL)
    expr <- source[[name]]
    if (is_missing_arg(expr))
      return(NULL)
    expr
  }
  arg_value <- function(name) {
    expr <- arg_expr(matched, name)
    if (is.null(expr))
      expr <- arg_expr(defaults, name)
    if (is.null(expr))
      return(NULL)
    value <- .np_try_eval_in_frames(expr, eval_env = caller_env)
    if (!isTRUE(value$ok))
      return(NULL)
    value$value
  }

  nomad <- arg_value("nomad")
  if (!is.null(nomad) && isTRUE(as.logical(nomad)[1L]))
    return(TRUE)

  regtype <- arg_value("regtype")
  if (is.null(regtype) || !identical(as.character(regtype)[1L], "lp"))
    return(FALSE)

  degree.select <- arg_value("degree.select")
  if (is.null(degree.select))
    return(FALSE)
  degree.select <- as.character(degree.select)[1L]
  if (!nzchar(degree.select) || identical(degree.select, "manual"))
    return(FALSE)

  search.engine <- arg_value("search.engine")
  if (is.null(search.engine))
    return(FALSE)
  search.engine <- as.character(search.engine)[1L]

  search.engine %in% c("nomad", "nomad+powell")
}

.np_bw_dispatch_target <- function(dots, data_arg_names = character(), eval_env = parent.frame()) {
  if (length(dots) == 0L)
    stop("invoked without arguments")

  dot.names <- names(dots)
  has.named.bws <- !is.null(dot.names) && any(dot.names == "bws")

  if (!is.null(dot.names) && any(dot.names == "formula")) {
    fval <- .np_try_eval_in_frames(dots[[which(dot.names == "formula")[1L]]], eval_env = eval_env)
    if (isTRUE(fval$ok))
      return(fval$value)
  }

  first.eval <- .np_try_eval_in_frames(dots[[1L]], eval_env = eval_env)
  if (!isTRUE(first.eval$ok))
    return(NULL)
  first.val <- first.eval$value
  if (inherits(first.val, "formula"))
    return(first.val)

  if (!has.named.bws)
    return(NULL)

  if (has.named.bws) {
    bval <- .np_try_eval_in_frames(dots[[which(dot.names == "bws")[1L]]], eval_env = eval_env)
    if (isTRUE(bval$ok))
      return(bval$value)
  }

  first.val
}

.np_bw_formula_from_call <- function(call_obj, eval_env = parent.frame()) {
  if (missing(call_obj) || !is.call(call_obj))
    return(NULL)

  if (length(call_obj) < 2L)
    return(NULL)

  for (i in 2:length(call_obj)) {
    val <- .np_try_eval_in_frames(call_obj[[i]], eval_env = eval_env)
    if (isTRUE(val$ok) && inherits(val$value, "formula"))
      return(call_obj[[i]])
  }

  NULL
}

.np_bw_resolve_formula <- function(formula_obj, formula_call = NULL, eval_env = parent.frame()) {
  if (is.null(formula_call))
    return(formula_obj)

  resolved <- .np_try_eval_in_frames(formula_call, eval_env = eval_env)
  if (isTRUE(resolved$ok) && inherits(resolved$value, "formula"))
    return(resolved$value)

  formula_obj
}

.np_terms_variable_values <- function(terms_obj, data, eval_env = environment(terms_obj)) {
  if (missing(terms_obj) || is.null(terms_obj))
    return(list())

  vars <- attr(terms_obj, "variables")
  if (is.null(vars))
    return(list())

  out <- .np_try_eval_in_frames(vars, eval_env = data, enclos = eval_env)
  if (!isTRUE(out$ok) || is.null(out$value))
    return(list())
  out <- out$value
  if (!is.list(out))
    out <- as.list(out)

  out
}

.np_terms_ts_mask <- function(terms_obj, data, eval_env = environment(terms_obj)) {
  vals <- .np_terms_variable_values(terms_obj = terms_obj, data = data, eval_env = eval_env)
  if (length(vals) == 0L)
    return(logical(0))

  vapply(vals, inherits, logical(1), "ts")
}

npreghat <-
  function(bws, ...){
    args <- list(...)

    if (!missing(bws)){
      if (is.recursive(bws)){
        if (!is.null(bws$formula) && is.null(args$txdat))
          UseMethod("npreghat", bws$formula)
        else if (!is.null(bws$call) && is.null(args$txdat))
          UseMethod("npreghat", bws$call)
        else if (!is.call(bws))
          UseMethod("npreghat", bws)
        else
          UseMethod("npreghat", NULL)
      } else {
        UseMethod("npreghat", NULL)
      }
    } else {
      UseMethod("npreghat", NULL)
    }
  }

.np_hat_constraint_from_matrix <- function(H, y, where) {
  if (is.null(y))
    stop(sprintf("argument 'y' is required when output='constraint' in %s", where),
         call. = FALSE)

  if (is.factor(y) || is.vector(y)) {
    y <- matrix(as.double(y), ncol = 1L)
  } else {
    y <- as.matrix(y)
  }

  if (ncol(y) != 1L)
    stop(sprintf("output='constraint' currently requires vector or one-column 'y' in %s", where),
         call. = FALSE)

  if (nrow(y) != ncol(H))
    stop(sprintf("length of 'y' must match the number of training columns in %s", where),
         call. = FALSE)

  t(H) * as.vector(y)
}

.npreghat_resolve_s <- function(s, deriv, ncon, con.names) {
  if (ncon == 0L)
    return(integer(0))

  if (is.null(s)) {
    if (is.null(deriv)) {
      s <- integer(ncon)
    } else {
      s <- deriv
    }
  }

  if (length(s) == 1L)
    s <- c(as.integer(s), rep.int(0L, ncon - 1L))

  if (!is.null(names(s))) {
    sout <- integer(ncon)
    names(sout) <- con.names
    keep <- intersect(names(s), con.names)
    if (length(keep))
      sout[keep] <- as.integer(s[keep])
    s <- sout
  }

  if (length(s) != ncon)
    stop("argument 's' must have length equal to the number of continuous predictors")

  if (anyNA(s) || any(s < 0) || any(s != as.integer(s)))
    stop("argument 's' must be a non-negative integer vector")

  as.integer(s)
}

.npreghat_solve_eval <- function(W, w.eval, k, ridge.base) {
  XtWX <- crossprod(W, W * k)
  p <- nrow(XtWX)
  diag.loc <- cbind(seq_len(p), seq_len(p))
  XtWX.diag <- XtWX[diag.loc]
  ridge.grid <- npRidgeSequenceFromBase(
    n.train = nrow(W),
    ridge.base = max(0.0, as.double(ridge.base)),
    cap = 1.0
  )
  solved <- FALSE
  ridge <- ridge.grid[1L]

  for (ridge.try in ridge.grid) {
    ridge <- ridge.try
    A <- XtWX
    if (ridge > 0)
      A[diag.loc] <- XtWX.diag + ridge

    v <- tryCatch(
      drop(solve(t(A), matrix(w.eval, ncol = 1L))),
      error = function(e) NULL
    )

    if (!is.null(v) && all(is.finite(v))) {
      solved <- TRUE
      break
    }
  }

  if (!solved)
    return(NULL)

  list(v = v, ridge = ridge)
}

.npreghat_exact_matrix_from_core <- function(bws, txdat, exdat = NULL, s = NULL) {
  miss.ex <- is.null(exdat)
  neval <- if (miss.ex) nrow(txdat) else nrow(exdat)
  ntrain <- nrow(txdat)
  H <- matrix(NA_real_, nrow = neval, ncol = ntrain)
  if (is.null(s))
    s <- integer(bws$ncon)

  want.grad <- length(s) > 0L && any(s > 0L)
  target.col <- NULL
  if (want.grad) {
    if (!(sum(s) == 1L && all(s %in% c(0L, 1L))))
      stop("exact direct hat matrix supports only mean and first derivatives")
    target.cont <- which(s == 1L)
    target.col <- which(bws$icon)[target.cont]
  }

  fit_one <- function(ycol) {
    direct.out <- .np_regression_direct(
      bws = bws,
      txdat = txdat,
      tydat = ycol,
      exdat = exdat,
      gradients = want.grad,
      gradient.order = 1L
    )
    if (is.null(target.col)) {
      direct.out$mean
    } else {
      direct.out$grad[, target.col]
    }
  }

  for (j in seq_len(ntrain)) {
    yj <- numeric(ntrain)
    yj[j] <- 1.0
    H[, j] <- fit_one(yj)
  }

  H
}

.npreghat_exact_lc_matrix_from_kernel_weights <- function(bws, txdat, exdat = NULL) {
  kw <- .np_kernel_weights_direct(
    bws = bws,
    txdat = txdat,
    exdat = exdat,
    leave.one.out = FALSE,
    bandwidth.divide = TRUE,
    kernel.pow = 1.0
  )

  denom <- colSums(kw)
  denom[denom == 0.0] <- .Machine$double.xmin
  sweep(t(kw), 1L, denom, "/")
}

.npreghat_exact_lc_derivative_matrix_from_npksum_chunked <- function(bws,
                                                                     txdat,
                                                                     exdat = NULL,
                                                                     s) {
  no.ex <- is.null(exdat)
  txdat <- toFrame(txdat)
  if (!no.ex) {
    exdat <- toFrame(exdat)
    if (!(txdat %~% exdat))
      stop("'txdat' and 'exdat' are not similar data frames!")
  }

  target.cont <- which(s == 1L)
  if (length(target.cont) != 1L)
    stop("lc derivative hat matrix requires exactly one continuous first derivative")

  eval.data <- if (no.ex) txdat else exdat
  ntrain <- nrow(txdat)
  neval <- nrow(eval.data)

  if (identical(bws$type, "adaptive_nn")) {
    out <- npksum.default(
      bws = bws,
      txdat = txdat,
      exdat = if (no.ex) txdat else eval.data,
      bandwidth.divide = TRUE,
      return.kernel.weights = TRUE,
      return.derivative.kernel.weights = TRUE,
      permutation.operator = "derivative"
    )

    kw <- out$kw
    pkw <- out$p.kw
    if (is.null(kw) || is.null(pkw))
      stop("adaptive lc derivative hat matrix requires kernel weights and derivative kernel weights")

    if (!is.matrix(kw))
      kw <- matrix(kw, nrow = ntrain, ncol = neval)

    if (length(dim(pkw)) == 3L) {
      pkw <- pkw[, , target.cont, drop = TRUE]
    } else {
      pkw <- as.matrix(pkw)
    }

    sk <- as.vector(out$ksum)
    dsk <- out$p.ksum
    if (length(dim(dsk)) == 3L) {
      dsk <- dsk[, 1L, target.cont, drop = TRUE]
    } else if (is.matrix(dsk)) {
      dsk <- dsk[, target.cont, drop = TRUE]
    }
    dsk <- as.vector(dsk)

    return(t(
      sweep(pkw, 2L, sk, "/") -
      sweep(kw, 2L, dsk / (sk^2), "*")
    ))
  }

  block.size <- min(512L, ntrain)
  ones <- rep.int(1.0, ntrain)
  H <- matrix(0.0, nrow = neval, ncol = ntrain)

  for (start in seq.int(1L, ntrain, by = block.size)) {
    stop.col <- min(ntrain, start + block.size - 1L)
    cols <- seq.int(start, stop.col)
    ib <- length(cols)
    W <- matrix(0.0, nrow = ntrain, ncol = ib + 1L)
    W[cbind(cols, seq_len(ib))] <- 1.0
    W[, ib + 1L] <- 1.0

    # The derivative call already returns both ksum and p.ksum.
    out <- npksum.default(
      bws = bws,
      txdat = txdat,
      exdat = if (no.ex) txdat else eval.data,
      tydat = ones,
      weights = W,
      bandwidth.divide = TRUE,
      permutation.operator = "derivative"
    )

    ks <- out$ksum
    ps <- out$p.ksum

    ks <- if (is.null(dim(ks))) {
      matrix(ks, nrow = ib + 1L, ncol = neval)
    } else {
      as.matrix(ks)
    }
    ps <- if (is.null(dim(ps))) {
      matrix(ps, nrow = ib + 1L, ncol = neval)
    } else if (length(dim(ps)) == 3L && dim(ps)[3L] == 1L) {
      ps[, , 1L, drop = TRUE]
    } else {
      as.matrix(ps)
    }

    sk <- ks[nrow(ks), ]
    dsk <- ps[nrow(ps), ]

    H[, cols] <- t(
      (ps[seq_len(ib), , drop = FALSE] / rep(sk, each = ib)) -
      (ks[seq_len(ib), , drop = FALSE] * rep(dsk / (sk^2), each = ib))
    )
  }

  H
}

.npreghat_exact_ll_matrix_from_kernel_weights <- function(bws, txdat, exdat = NULL, s = NULL) {
  miss.ex <- is.null(exdat)
  eval.data <- if (miss.ex) txdat else exdat
  ntrain <- nrow(txdat)
  neval <- nrow(eval.data)
  kw <- .np_kernel_weights_direct(
    bws = bws,
    txdat = txdat,
    exdat = if (miss.ex) NULL else eval.data,
    leave.one.out = FALSE,
    bandwidth.divide = identical(bws$type, "adaptive_nn"),
    kernel.pow = 1.0
  )

  xcon.train <- as.matrix(txdat[, bws$icon, drop = FALSE])
  xcon.eval <- as.matrix(eval.data[, bws$icon, drop = FALSE])
  design <- cbind(1.0, xcon.train)
  H <- matrix(NA_real_, nrow = neval, ncol = ntrain)

  want.grad <- length(s) > 0L && any(s > 0L)
  target.cont <- if (want.grad) which(s == 1L) else integer(0)

  for (j in seq_len(neval)) {
    w <- kw[, j]
    A.base <- crossprod(design, design * w)
    rhs.base <- t(design * w)
    solved <- tryCatch(solve(A.base, rhs.base), error = function(e) NULL)

    if (is.null(solved) || !all(is.finite(solved))) {
      eps <- 1.0 / ntrain
      nepsilon <- 0.0
      A.try <- A.base
      rhs.try <- rhs.base
      diag(A.try) <- diag(A.try) + eps
      solved <- tryCatch(solve(A.try, rhs.try), error = function(e) NULL)

      while (is.null(solved) || !all(is.finite(solved))) {
        diag(A.try) <- diag(A.try) + eps
        nepsilon <- nepsilon + eps
        solved <- tryCatch(solve(A.try, rhs.try), error = function(e) NULL)
      }

      if (nepsilon > 0.0) {
        sumw <- sum(w)
        if (sumw == 0.0)
          sumw <- .Machine$double.xmin
        rhs.try[1L, ] <- rhs.try[1L, ] + nepsilon * (w / sumw)
        solved <- solve(A.try, rhs.try)
      }
    }

    if (!want.grad) {
      fit.weights <- c(1.0, xcon.eval[j, , drop = TRUE])
      H[j, ] <- drop(fit.weights %*% solved)
    } else {
      H[j, ] <- solved[1L + target.cont, ]
    }
  }

  H
}

.npreghat_exact_lp_matrix_from_kernel_weights <- function(bws,
                                                          txdat,
                                                          exdat = NULL,
                                                          s = NULL,
                                                          basis = "glp",
                                                          degree = integer(0),
                                                          bernstein.basis = FALSE) {
  miss.ex <- is.null(exdat)
  eval.data <- if (miss.ex) txdat else exdat
  ntrain <- nrow(txdat)
  neval <- nrow(eval.data)
  want.grad <- length(s) > 0L && any(s > 0L)

  if (identical(bws$type, "generalized_nn") &&
      any(degree > 1L) &&
      isTRUE(getOption("np.tree"))) {
    return(.npreghat_exact_matrix_from_core(
      bws = bws,
      txdat = txdat,
      exdat = if (miss.ex) NULL else exdat,
      s = s
    ))
  }

  if (identical(bws$type, "generalized_nn") &&
      any(degree > 1L) &&
      !isTRUE(getOption("np.tree"))) {
    return(.npreghat_exact_lp_matrix_from_regression_core_chunked(
      bws = bws,
      txdat = txdat,
      exdat = if (miss.ex) NULL else exdat,
      basis = basis,
      degree = degree,
      bernstein.basis = bernstein.basis,
      s = s
    ))
  }

  kw <- .np_kernel_weights_direct(
    bws = bws,
    txdat = txdat,
    exdat = if (miss.ex) NULL else eval.data,
    leave.one.out = FALSE,
    bandwidth.divide = TRUE,
    kernel.pow = 1.0
  )

  W.train <- W.lp(
    xdat = txdat[, bws$icon, drop = FALSE],
    degree = degree,
    basis = basis,
    bernstein.basis = bernstein.basis
  )
  W.eval <- W.lp(
    xdat = txdat[, bws$icon, drop = FALSE],
    exdat = if (miss.ex) NULL else eval.data[, bws$icon, drop = FALSE],
    degree = degree,
    gradient.vec = if (want.grad) s else NULL,
    basis = basis,
    bernstein.basis = bernstein.basis
  )

  H <- matrix(NA_real_, nrow = neval, ncol = ntrain)
  eps <- 1.0 / max(1L, ntrain)

  for (j in seq_len(neval)) {
    w <- kw[, j]
    A.base <- crossprod(W.train, W.train * w)
    rhs <- W.eval[j, ]
    solved <- tryCatch(solve(A.base, rhs), error = function(e) NULL)

    if (is.null(solved) || !all(is.finite(solved))) {
      A.try <- A.base
      nepsilon <- 0.0

      repeat {
        diag(A.try) <- diag(A.try) + eps
        nepsilon <- nepsilon + eps
        solved <- tryCatch(solve(A.try, rhs), error = function(e) NULL)
        if (!is.null(solved) && all(is.finite(solved)))
          break
      }

      denom <- A.try[1L, 1L]
      if (!is.finite(denom) || abs(denom) < .Machine$double.xmin)
        denom <- .Machine$double.xmin
      solved[1L] <- solved[1L] * (1.0 + nepsilon / denom)
    }

    H[j, ] <- w * drop(W.train %*% solved)
  }

  H
}

.npreghat_exact_lp_matrix_from_regression_core_chunked <- function(bws,
                                                                   txdat,
                                                                   exdat = NULL,
                                                                   basis = "glp",
                                                                   degree = integer(0),
                                                                   bernstein.basis = FALSE,
                                                                   s = NULL,
                                                                   chunk.size = 128L) {
  no.ex <- is.null(exdat)

  txdat <- toFrame(txdat)
  if (!no.ex) {
    exdat <- toFrame(exdat)
    if (!(txdat %~% exdat))
      stop("'txdat' and 'exdat' are not similar data frames!")
  }

  if (length(bws$bw) != length(txdat))
    stop("length of bandwidth vector does not match number of columns of 'txdat'")

  if ((any(bws$icon) &&
       !all(vapply(txdat[, bws$icon, drop = FALSE], inherits, logical(1), c("integer", "numeric")))) ||
      (any(bws$iord) &&
       !all(vapply(txdat[, bws$iord, drop = FALSE], inherits, logical(1), "ordered"))) ||
      (any(bws$iuno) &&
       !all(vapply(txdat[, bws$iuno, drop = FALSE], inherits, logical(1), "factor")))) {
    stop("supplied bandwidths do not match 'txdat' in type")
  }

  txdat <- adjustLevels(txdat, bws$xdati)
  if (!no.ex) {
    exdat <- adjustLevels(exdat, bws$xdati, allowNewCells = TRUE)
    npKernelBoundsCheckEval(exdat, bws$icon, bws$ckerlb, bws$ckerub, argprefix = "cker")
  }

  txmat <- toMatrix(txdat)
  tuno <- txmat[, bws$iuno, drop = FALSE]
  tcon <- txmat[, bws$icon, drop = FALSE]
  tord <- txmat[, bws$iord, drop = FALSE]
  reg.c <- npRegtypeToC(regtype = "lp",
                        degree = degree,
                        ncon = bws$ncon,
                        context = ".npreghat_exact_lp_matrix_from_regression_core_chunked")

  npCheckRegressionDesignCondition(reg.code = reg.c$code,
                                   xcon = tcon,
                                   basis = basis,
                                   degree = degree,
                                   bernstein.basis = bernstein.basis,
                                   where = ".npreghat_exact_lp_matrix_from_regression_core_chunked")

  if (!no.ex) {
    exmat <- toMatrix(exdat)
    euno <- exmat[, bws$iuno, drop = FALSE]
    econ <- exmat[, bws$icon, drop = FALSE]
    eord <- exmat[, bws$iord, drop = FALSE]
  } else {
    euno <- tuno
    econ <- tcon
    eord <- tord
  }

  tuno <- if (bws$nuno > 0L) as.matrix(tuno) else matrix(double(), nrow = nrow(txdat), ncol = 0L)
  tcon <- if (bws$ncon > 0L) as.matrix(tcon) else matrix(double(), nrow = nrow(txdat), ncol = 0L)
  tord <- if (bws$nord > 0L) as.matrix(tord) else matrix(double(), nrow = nrow(txdat), ncol = 0L)
  euno <- if (bws$nuno > 0L) as.matrix(euno) else matrix(double(), nrow = if (no.ex) nrow(txdat) else nrow(exdat), ncol = 0L)
  econ <- if (bws$ncon > 0L) as.matrix(econ) else matrix(double(), nrow = if (no.ex) nrow(txdat) else nrow(exdat), ncol = 0L)
  eord <- if (bws$nord > 0L) as.matrix(eord) else matrix(double(), nrow = if (no.ex) nrow(txdat) else nrow(exdat), ncol = 0L)

  bwtype.c <- switch(bws$type,
    fixed = BW_FIXED,
    generalized_nn = BW_GEN_NN,
    adaptive_nn = BW_ADAP_NN
  )
  kernel.x.c <- switch(bws$ckertype,
    gaussian = CKER_GAUSS + bws$ckerorder / 2 - 1,
    epanechnikov = CKER_EPAN + bws$ckerorder / 2 - 1,
    uniform = CKER_UNI,
    "truncated gaussian" = CKER_TGAUSS
  )
  kernel.xu.c <- switch(bws$ukertype,
    aitchisonaitken = UKER_AIT,
    liracine = UKER_LR
  )
  kernel.xo.c <- switch(bws$okertype,
    wangvanryzin = OKER_WANG,
    liracine = OKER_LR,
    racineliyan = OKER_RLY
  )

  ntrain <- nrow(txdat)
  neval <- if (no.ex) ntrain else nrow(exdat)
  H <- matrix(0.0, nrow = neval, ncol = ntrain)
  chunk.size <- max(1L, min(as.integer(chunk.size), ntrain))
  bw.vec <- as.double(c(bws$bw[bws$icon], bws$bw[bws$iuno], bws$bw[bws$iord]))
  tree.flag <- isTRUE(getOption("np.tree"))
  grad.vec <- if (length(s) && any(s > 0L)) as.integer(s) else integer(0L)
  on.exit(
    tryCatch(.Call("C_np_shadow_reset_state", PACKAGE = "np"),
             error = function(e) NULL),
    add = TRUE
  )

  for (start in seq.int(1L, ntrain, by = chunk.size)) {
    stop.col <- min(ntrain, start + chunk.size - 1L)
    cols <- seq.int(start, stop.col)
    rhs <- matrix(0.0, nrow = ntrain, ncol = length(cols))
    rhs[cbind(cols, seq_along(cols))] <- 1.0
    H[, cols] <- .Call(
      "C_np_regression_lp_apply_conditional",
      tuno,
      tord,
      tcon,
      euno,
      eord,
      econ,
      rhs,
      bw.vec,
      as.integer(bwtype.c),
      as.integer(kernel.x.c),
        as.integer(kernel.xu.c),
        as.integer(kernel.xo.c),
        as.logical(tree.flag),
        as.integer(degree),
        grad.vec,
        as.integer(isTRUE(bernstein.basis)),
        as.integer(npLpBasisCode(basis)),
        PACKAGE = "np"
      )
  }

  H
}

.np_kernel_weights_direct <- function(bws,
                                      txdat,
                                      exdat = NULL,
                                      leave.one.out = FALSE,
                                      bandwidth.divide = TRUE,
                                      kernel.pow = 1.0,
                                      operator = NULL) {
  miss.ex <- is.null(exdat)
  txdat <- toFrame(txdat)
  if (!miss.ex) {
    exdat <- toFrame(exdat)
    if (!(txdat %~% exdat))
      stop("'txdat' and 'exdat' are not similar data frames!")
  }

  if (!isa(bws, "kbandwidth"))
    bws <- kbandwidth(bws)

  if (length(bws$bw) != length(txdat))
    stop("length of bandwidth vector does not match number of columns of 'txdat'")

  leave.one.out <- npValidateScalarLogical(leave.one.out, "leave.one.out")
  bandwidth.divide <- npValidateScalarLogical(bandwidth.divide, "bandwidth.divide")
  if (!miss.ex && leave.one.out)
    stop("you may not specify 'leave.one.out = TRUE' and provide evaluation data")

  if (is.null(operator)) {
    operator <- rep.int("normal", length(txdat))
  } else {
    if (length(operator) == 1L)
      operator <- rep.int(operator, length(txdat))
    if (length(operator) != length(txdat))
      stop("operator not specified for all variables")
  }

  uo.operators <- c("normal", "convolution", "integral")
  if (!all(operator %in% names(ALL_OPERATORS)))
    stop("invalid operator specification")
  if (!all(operator[bws$iuno | bws$iord] %in% uo.operators))
    stop("unordered and ordered variables may only use normal, convolution, or integral operators")
  operator.num <- ALL_OPERATORS[operator]

  txdat <- adjustLevels(txdat, bws$xdati, allowNewCells = TRUE)
  if (!miss.ex) {
    exdat <- adjustLevels(exdat, bws$xdati, allowNewCells = TRUE)
    npKernelBoundsCheckEval(exdat, bws$icon, bws$ckerlb, bws$ckerub, argprefix = "cker")
  }

  txm <- toMatrix(txdat)
  tuno <- txm[, bws$iuno, drop = FALSE]
  tcon <- txm[, bws$icon, drop = FALSE]
  tord <- txm[, bws$iord, drop = FALSE]

  if (!miss.ex) {
    exm <- toMatrix(exdat)
    euno <- exm[, bws$iuno, drop = FALSE]
    econ <- exm[, bws$icon, drop = FALSE]
    eord <- exm[, bws$iord, drop = FALSE]
  } else {
    euno <- data.frame()
    eord <- data.frame()
    econ <- data.frame()
  }

  tnrow <- nrow(txdat)
  enrow <- if (miss.ex) tnrow else nrow(exdat)
  nkw <- tnrow * enrow

  myopti <- list(
    num_obs_train = tnrow,
    num_obs_eval = enrow,
    num_uno = bws$nuno,
    num_ord = bws$nord,
    num_con = bws$ncon,
    int_LARGE_SF = SF_ARB,
    BANDWIDTH_reg_extern = switch(bws$type,
      fixed = BW_FIXED,
      generalized_nn = BW_GEN_NN,
      adaptive_nn = BW_ADAP_NN
    ),
    int_MINIMIZE_IO = if (isTRUE(getOption("np.messages"))) IO_MIN_FALSE else IO_MIN_TRUE,
    kerneval = switch(bws$ckertype,
      gaussian = CKER_GAUSS + bws$ckerorder / 2 - 1,
      epanechnikov = CKER_EPAN + bws$ckerorder / 2 - 1,
      uniform = CKER_UNI,
      "truncated gaussian" = CKER_TGAUSS
    ),
    ukerneval = switch(bws$ukertype,
      aitchisonaitken = UKER_AIT,
      liracine = UKER_LR
    ),
    okerneval = switch(bws$okertype,
      wangvanryzin = OKER_WANG,
      liracine = OKER_LR,
      nliracine = OKER_NLR,
      racineliyan = OKER_RLY
    ),
    miss.ex = miss.ex,
    leave.one.out = leave.one.out,
    bandwidth.divide = bandwidth.divide,
    mcv.numRow = attr(bws$xmcv, "num.row"),
    wncol = 0L,
    yncol = 0L,
    int_do_tree = if (isTRUE(getOption("np.tree"))) DO_TREE_YES else DO_TREE_NO,
    return.kernel.weights = TRUE,
    permutation.operator = PERMUTATION_OPERATORS[["none"]],
    compute.score = FALSE,
    compute.ocg = FALSE
  )

  cker.bounds.c <- npKernelBoundsMarshal(bws$ckerlb[bws$icon], bws$ckerub[bws$icon])

  asDouble <- function(data) {
    if (is.null(data)) as.double(0.0) else as.double(data)
  }

  myout <- .Call(
    "C_np_kernelsum",
    asDouble(tuno), asDouble(tord), asDouble(tcon),
    as.double(0.0), as.double(0.0),
    asDouble(euno), asDouble(eord), asDouble(econ),
    as.double(c(bws$bw[bws$icon], bws$bw[bws$iuno], bws$bw[bws$iord])),
    as.double(bws$xmcv), as.double(attr(bws$xmcv, "pad.num")),
    as.integer(c(operator.num[bws$icon], operator.num[bws$iuno], operator.num[bws$iord])),
    as.integer(myopti), as.double(kernel.pow),
    as.integer(enrow),
    as.integer(0L),
    as.integer(nkw),
    as.double(cker.bounds.c$lb),
    as.double(cker.bounds.c$ub),
    PACKAGE = "np"
  )

  kw <- matrix(as.double(myout[["kernel.weights"]]), nrow = tnrow, ncol = enrow)
  if (leave.one.out && miss.ex && nrow(kw) == ncol(kw))
    diag(kw) <- 0.0
  kw
}

.np_regression_direct <- function(bws,
                                  txdat,
                                  tydat,
                                  exdat = NULL,
                                  gradients = FALSE,
                                  gradient.order = 1L,
                                  local.mode = FALSE) {
  no.ex <- is.null(exdat)
  gradients <- npValidateScalarLogical(gradients, "gradients")
  local.mode <- npValidateScalarLogical(local.mode, "local.mode")

  txdat <- toFrame(txdat)
  if (!no.ex) {
    exdat <- toFrame(exdat)
    if (!(txdat %~% exdat))
      stop("'txdat' and 'exdat' are not similar data frames!")
  }

  if (length(bws$bw) != length(txdat))
    stop("length of bandwidth vector does not match number of columns of 'txdat'")

  regtype <- if (is.null(bws$regtype)) "lc" else as.character(bws$regtype)
  basis <- npValidateLpBasis(regtype = regtype, basis = bws$basis)
  degree <- npValidateGlpDegree(regtype = regtype,
                                degree = bws$degree,
                                ncon = bws$ncon)
  bernstein.basis <- npValidateGlpBernstein(regtype = regtype,
                                            bernstein.basis = bws$bernstein.basis)
  reg.spec <- npCanonicalConditionalRegSpec(
    regtype = regtype,
    basis = basis,
    degree = degree,
    bernstein.basis = bernstein.basis,
    ncon = bws$ncon,
    where = ".np_regression_direct"
  )
  glp.gradient.order <- if (identical(reg.spec$regtype.engine, "lp")) {
    if (identical(regtype, "lp")) {
      npValidateGlpGradientOrder(regtype = regtype,
                                 gradient.order = gradient.order,
                                 ncon = bws$ncon)
    } else if (bws$ncon > 0L) {
      rep.int(1L, bws$ncon)
    } else {
      integer(0)
    }
  } else {
    NULL
  }
  if (isTRUE(gradients) &&
      identical(reg.spec$regtype.engine, "lp") &&
      (bws$ncon > 0L) &&
      all(reg.spec$degree.engine == 0L)) {
    stop("regtype='lp' with degree=0 does not support derivatives; use gradients=FALSE for fitted/predicted values")
  }

  reg.c <- npRegtypeToC(regtype = reg.spec$regtype.engine,
                        degree = reg.spec$degree.engine,
                        ncon = bws$ncon,
                        context = ".np_regression_direct")
  degree.c <- if (bws$ncon > 0L) {
    as.integer(if (is.null(reg.c$degree)) rep.int(0L, bws$ncon) else reg.c$degree)
  } else {
    integer(1L)
  }

  if ((any(bws$icon) &&
       !all(vapply(txdat[, bws$icon, drop = FALSE], inherits, logical(1), c("integer", "numeric")))) ||
      (any(bws$iord) &&
       !all(vapply(txdat[, bws$iord, drop = FALSE], inherits, logical(1), "ordered"))) ||
      (any(bws$iuno) &&
       !all(vapply(txdat[, bws$iuno, drop = FALSE], inherits, logical(1), "factor")))) {
    stop("supplied bandwidths do not match 'txdat' in type")
  }

  if (!(is.vector(tydat) || is.factor(tydat)))
    stop("'tydat' must be a vector or a factor")
  if (nrow(txdat) != length(tydat))
    stop("number of explanatory data 'txdat' and dependent data 'tydat' do not match")

  if (is.factor(tydat)) {
    tydat <- adjustLevels(data.frame(tydat), bws$ydati)[, 1L]
    tydat <- (bws$ydati$all.dlev[[1L]])[as.integer(tydat)]
  } else {
    tydat <- as.double(tydat)
  }

  txdat <- adjustLevels(txdat, bws$xdati)
  if (!no.ex) {
    exdat <- adjustLevels(exdat, bws$xdati, allowNewCells = TRUE)
    npKernelBoundsCheckEval(exdat, bws$icon, bws$ckerlb, bws$ckerub, argprefix = "cker")
  }

  txmat <- toMatrix(txdat)
  tuno <- txmat[, bws$iuno, drop = FALSE]
  tcon <- txmat[, bws$icon, drop = FALSE]
  tord <- txmat[, bws$iord, drop = FALSE]

  npCheckRegressionDesignCondition(reg.code = reg.c$code,
                                   xcon = tcon,
                                   basis = reg.spec$basis.engine,
                                   degree = reg.spec$degree.engine,
                                   bernstein.basis = reg.spec$bernstein.basis.engine,
                                   where = ".np_regression_direct")

  if (!no.ex) {
    exmat <- toMatrix(exdat)
    euno <- exmat[, bws$iuno, drop = FALSE]
    econ <- exmat[, bws$icon, drop = FALSE]
    eord <- exmat[, bws$iord, drop = FALSE]
  } else {
    euno <- data.frame()
    econ <- data.frame()
    eord <- data.frame()
  }

  tnrow <- nrow(txdat)
  enrow <- if (no.ex) tnrow else nrow(exdat)
  ncol.x <- ncol(txdat)

  myopti <- list(
    num_obs_train = tnrow,
    num_obs_eval = enrow,
    num_uno = bws$nuno,
    num_ord = bws$nord,
    num_con = bws$ncon,
    int_LARGE_SF = if (bws$scaling) SF_NORMAL else SF_ARB,
    BANDWIDTH_reg_extern = switch(bws$type,
      fixed = BW_FIXED,
      generalized_nn = BW_GEN_NN,
      adaptive_nn = BW_ADAP_NN
    ),
    int_MINIMIZE_IO = if (isTRUE(getOption("np.messages"))) IO_MIN_FALSE else IO_MIN_TRUE,
    kerneval = switch(bws$ckertype,
      gaussian = CKER_GAUSS + bws$ckerorder / 2 - 1,
      epanechnikov = CKER_EPAN + bws$ckerorder / 2 - 1,
      uniform = CKER_UNI,
      "truncated gaussian" = CKER_TGAUSS
    ),
    ukerneval = switch(bws$ukertype,
      aitchisonaitken = UKER_AIT,
      liracine = UKER_LR
    ),
    okerneval = switch(bws$okertype,
      wangvanryzin = OKER_WANG,
      liracine = OKER_LR,
      "racineliyan" = OKER_RLY
    ),
    ey_is_ty = TRUE,
    do_grad = gradients,
    regtype = reg.c$code,
    no.ex = no.ex,
    mcv.numRow = attr(bws$xmcv, "num.row"),
    int_do_tree = if (isTRUE(getOption("np.tree"))) DO_TREE_YES else DO_TREE_NO,
    old.reg = FALSE
  )

  cker.bounds.c <- npKernelBoundsMarshal(bws$ckerlb[bws$icon], bws$ckerub[bws$icon])

  asDouble <- function(data) {
    if (is.null(data)) as.double(0.0) else as.double(data)
  }

  glp.gradient.order.c <- if (bws$ncon > 0L) {
    as.integer(if (is.null(glp.gradient.order)) rep.int(1L, bws$ncon) else glp.gradient.order)
  } else {
    integer(1L)
  }

  myout <- .Call(
    "C_np_regression",
    asDouble(tuno), asDouble(tord), asDouble(tcon), as.double(tydat),
    asDouble(euno), asDouble(eord), asDouble(econ), as.double(double()),
    asDouble(c(bws$bw[bws$icon], bws$bw[bws$iuno], bws$bw[bws$iord])),
    asDouble(bws$xmcv), asDouble(attr(bws$xmcv, "pad.num")),
    asDouble(bws$nconfac), asDouble(bws$ncatfac), asDouble(bws$sdev),
    as.integer(myopti),
    as.integer(degree.c),
    as.integer(glp.gradient.order.c),
    as.integer(isTRUE(reg.spec$bernstein.basis.engine)),
    as.integer(npLpBasisCode(reg.spec$basis.engine)),
    as.integer(enrow),
    as.integer(ncol.x),
    as.logical(gradients),
    as.double(cker.bounds.c$lb),
    as.double(cker.bounds.c$ub),
    PACKAGE = "np"
  )

  out <- list(mean = as.double(myout$mean))

  if (gradients) {
    grad <- matrix(data = myout$g, nrow = enrow, ncol = ncol.x, byrow = FALSE)
    rorder <- numeric(ncol.x)
    ord.idx <- seq_len(ncol.x)
    rorder[c(ord.idx[bws$icon], ord.idx[bws$iuno], ord.idx[bws$iord])] <- ord.idx
    grad <- as.matrix(grad[, rorder, drop = FALSE])

    if (identical(regtype, "lp")) {
      cont.idx <- which(bws$icon)
      if (length(cont.idx)) {
        invalid.order <- glp.gradient.order > degree
        if (any(invalid.order))
          grad[, cont.idx[invalid.order]] <- NA_real_
      }
    }

    out$grad <- grad
  }

  out
}

npreghat.formula <-
  function(bws, data = NULL, newdata = NULL, ...){

    tt <- terms(bws)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)
    tmf <- bws$call[c(1, m)]
    tmf[[1]] <- as.name("model.frame")
    tmf[["formula"]] <- tt
    mf.args <- as.list(tmf)[-1L]
    mf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

    y <- model.response(mf)
    txdat <- mf[, attr(attr(mf, "terms"), "term.labels"), drop = FALSE]

    has.eval <- !is.null(newdata)
    if (has.eval) {
      et <- delete.response(tt)
      emf <- model.frame(et, data = newdata)
      exdat <- emf[, attr(attr(emf, "terms"), "term.labels"), drop = FALSE]
    }

    hat.args <- list(bws = bws, txdat = txdat)
    if (!is.null(y))
      hat.args$y <- y
    if (has.eval)
      hat.args$exdat <- exdat

    ev <- do.call(npreghat, c(hat.args, list(...)))
    attr(ev, "call") <- match.call(expand.dots = FALSE)
    ev
  }

npreghat.call <-
  function(bws, ...) {
    ev <- npreghat(
      txdat = .np_eval_bws_call_arg(bws, "xdat"),
      y = .np_eval_bws_call_arg(bws, "ydat"),
      bws = bws,
      ...
    )
    attr(ev, "call") <- match.call(expand.dots = FALSE)
    ev
  }

npreghat.npregression <-
  function(bws, txdat, y, ...){
    model <- bws
    rbw <- model$bws

    if (missing(txdat))
      txdat <- .np_eval_bws_call_arg(rbw, "xdat")

    if (missing(y))
      y <- .np_eval_bws_call_arg(rbw, "ydat")

    npreghat(bws = rbw, txdat = txdat, y = y, ...)
  }

npreghat.rbandwidth <-
  function(bws,
           txdat = stop("training data 'txdat' missing"),
           exdat,
           y = NULL,
           output = c("matrix", "apply", "constraint"),
           basis = NULL,
           bernstein.basis = NULL,
           degree = NULL,
           deriv = NULL,
           leave.one.out = FALSE,
           ridge = 0.0,
           s = NULL,
           ...){

    output <- match.arg(output)
    constraint.output <- identical(output, "constraint")
    matrix.output <- identical(output, "matrix") || constraint.output
    dots <- list(...)
    npRejectLegacyLpArgs(names(dots), where = "npreghat")

    txdat <- toFrame(txdat)
    no.ex <- missing(exdat)
    leave.one.out <- npValidateScalarLogical(leave.one.out, "leave.one.out")

    if (!no.ex) {
      exdat <- toFrame(exdat)
      if (!(txdat %~% exdat))
        stop("'txdat' and 'exdat' are not similar data frames!")
    }

    if (!no.ex && leave.one.out)
      stop("you may not specify 'leave.one.out = TRUE' and provide evaluation data")

    if (length(bws$bw) != ncol(txdat))
      stop("length of bandwidth vector does not match number of columns of 'txdat'")

    npValidateRegressionNnLowerBound(bws, where = "npreghat")

    if ((any(bws$icon) &&
         !all(vapply(txdat[, bws$icon, drop = FALSE], inherits, logical(1), c("integer", "numeric")))) ||
        (any(bws$iord) &&
         !all(vapply(txdat[, bws$iord, drop = FALSE], inherits, logical(1), "ordered"))) ||
        (any(bws$iuno) &&
         !all(vapply(txdat[, bws$iuno, drop = FALSE], inherits, logical(1), "factor"))))
      stop("supplied bandwidths do not match 'txdat' in type")

    if (!is.null(y)) {
      if (is.factor(y) || is.vector(y)) {
        if (length(y) != nrow(txdat))
          stop("length of 'y' must match the number of training rows in 'txdat'")
        y <- matrix(as.double(y), ncol = 1L)
      } else {
        y <- as.matrix(y)
        if (nrow(y) != nrow(txdat))
          stop("number of rows in 'y' must match the number of training rows in 'txdat'")
      }
    }

    keep.rows <- rep_len(TRUE, nrow(txdat))
    rows.omit <- attr(na.omit(txdat), "na.action")
    if (!is.null(y))
      rows.omit <- union(rows.omit, attr(na.omit(as.data.frame(y)), "na.action"))

    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Training data has no rows without NAs")

    txdat <- txdat[keep.rows, , drop = FALSE]
    if (!is.null(y))
      y <- y[keep.rows, , drop = FALSE]

    if (!no.ex) {
      keep.eval <- rep_len(TRUE, nrow(exdat))
      rows.omit.eval <- attr(na.omit(exdat), "na.action")
      if (length(rows.omit.eval) > 0L)
        keep.eval[as.integer(rows.omit.eval)] <- FALSE
      if (!any(keep.eval))
        stop("Evaluation data has no rows without NAs")
      exdat <- exdat[keep.eval, , drop = FALSE]
    }

    regtype <- if (is.null(bws$regtype)) "lc" else as.character(bws$regtype)
    ncon <- bws$ncon
    con.names <- names(txdat)[which(bws$icon)]

    degree <- if (identical(regtype, "lc")) {
      rep.int(0L, ncon)
    } else if (identical(regtype, "ll")) {
      rep.int(1L, ncon)
    } else {
      npValidateGlpDegree(regtype = "lp",
                          degree = if (is.null(degree)) bws$degree else degree,
                          ncon = ncon)
    }

    basis <- npValidateLpBasis(
      regtype = "lp",
      basis = if (is.null(basis)) {
        if (is.null(bws$basis)) "glp" else bws$basis
      } else {
        basis
      }
    )

    bernstein.basis <- npValidateGlpBernstein(
      regtype = "lp",
      bernstein.basis = if (is.null(bernstein.basis)) {
        isTRUE(bws$bernstein.basis)
      } else {
        bernstein.basis
      }
    )

    s <- .npreghat_resolve_s(s = s, deriv = deriv, ncon = ncon, con.names = con.names)
    reg.spec <- npCanonicalConditionalRegSpec(
      regtype = regtype,
      basis = basis,
      degree = degree,
      bernstein.basis = bernstein.basis,
      ncon = ncon,
      where = "npreghat"
    )

    first.derivative.request <- (sum(s) == 1L) && all(s %in% c(0L, 1L))
    simple.operator.request <- (sum(s) == 0L) || first.derivative.request

    direct.apply <- identical(output, "apply") &&
      !is.null(y) &&
      !isTRUE(leave.one.out) &&
      (ncol(y) == 1L) &&
      simple.operator.request

    lc.derivative.exact.route <- identical(regtype, "lc") &&
      first.derivative.request

    exact.lc.kernel.route <- !isTRUE(leave.one.out) &&
      !any(s > 0L) &&
      identical(regtype, "lc") &&
      (bws$ncon > 0L)

    exact.ll.kernel.route <- !isTRUE(leave.one.out) &&
      simple.operator.request &&
      identical(regtype, "ll") &&
      (bws$ncon > 0L)

    exact.lp.kernel.route <- !isTRUE(leave.one.out) &&
      simple.operator.request &&
      identical(regtype, "lp") &&
      identical(reg.spec$regtype.engine, "lp") &&
      (bws$ncon > 0L)

    exact.core.route <- !isTRUE(leave.one.out) &&
      simple.operator.request &&
      (
        exact.lc.kernel.route ||
        exact.ll.kernel.route ||
        exact.lp.kernel.route ||
        lc.derivative.exact.route ||
        FALSE
      )

    if (direct.apply) {
      direct.out <- .np_regression_direct(
        bws = bws,
        txdat = txdat,
        tydat = as.vector(y[, 1L]),
        exdat = if (no.ex) NULL else exdat,
        gradients = any(s > 0L),
        gradient.order = 1L
      )

      if (!any(s > 0L))
        return(as.vector(direct.out$mean))

      target.cont <- which(s == 1L)
      target.col <- which(bws$icon)[target.cont]
      return(as.vector(direct.out$grad[, target.col]))
    }

    if (exact.core.route) {
      H <- if (lc.derivative.exact.route) {
        .npreghat_exact_lc_derivative_matrix_from_npksum_chunked(
          bws = bws,
          txdat = txdat,
          exdat = if (no.ex) NULL else exdat,
          s = s
        )
      } else if (exact.lc.kernel.route) {
        .npreghat_exact_lc_matrix_from_kernel_weights(
          bws = bws,
          txdat = txdat,
          exdat = if (no.ex) NULL else exdat
        )
      } else if (exact.ll.kernel.route) {
        .npreghat_exact_ll_matrix_from_kernel_weights(
          bws = bws,
          txdat = txdat,
          exdat = if (no.ex) NULL else exdat,
          s = s
        )
      } else if (exact.lp.kernel.route) {
        .npreghat_exact_lp_matrix_from_kernel_weights(
          bws = bws,
          txdat = txdat,
          exdat = if (no.ex) NULL else exdat,
          s = s,
          basis = reg.spec$basis.engine,
          degree = reg.spec$degree.engine,
          bernstein.basis = reg.spec$bernstein.basis.engine
        )
      } else {
        .npreghat_exact_matrix_from_core(
          bws = bws,
          txdat = txdat,
          exdat = if (no.ex) NULL else exdat,
          s = s
        )
      }

      if (identical(output, "apply")) {
        if (is.null(y))
          stop("argument 'y' is required when output='apply'")
        out <- H %*% y
        if (ncol(out) == 1L)
          return(as.vector(out))
        return(out)
      }

      if (constraint.output)
        return(.np_hat_constraint_from_matrix(H, y, "npreghat"))

      class(H) <- c("npreghat", "matrix")
      attr(H, "bws") <- bws
      attr(H, "txdat") <- txdat
      attr(H, "exdat") <- if (no.ex) txdat else exdat
      attr(H, "trainiseval") <- no.ex
      attr(H, "regtype") <- regtype
      attr(H, "degree") <- degree
      attr(H, "basis") <- basis
      attr(H, "bernstein.basis") <- bernstein.basis
      attr(H, "s") <- s
      attr(H, "leave.one.out") <- leave.one.out
      attr(H, "ridge.used") <- rep.int(0.0, nrow(H))
      attr(H, "rows.omit") <- rows.omit
      attr(H, "call") <- match.call(expand.dots = FALSE)

      if (!is.null(y)) {
        Hy <- H %*% y
        if (ncol(Hy) == 1L)
          Hy <- as.vector(Hy)
        attr(H, "Hy") <- Hy
      }

      return(H)
    }

    if (any(s > degree))
      stop("requested derivative order in 's' exceeds local polynomial degree")

    kw <- .np_kernel_weights_direct(
      bws = bws,
      txdat = txdat,
      exdat = if (no.ex) NULL else exdat,
      leave.one.out = leave.one.out,
      bandwidth.divide = TRUE,
      kernel.pow = 1.0
    )

    ntrain <- nrow(txdat)
    neval <- ncol(kw)

    W <- W.lp(
      xdat = txdat,
      degree = degree,
      basis = basis,
      bernstein.basis = bernstein.basis
    )

    eval.data <- if (no.ex) txdat else exdat
    W.eval <- W.lp(
      xdat = txdat,
      exdat = if (no.ex) NULL else eval.data,
      degree = degree,
      gradient.vec = if (any(s > 0L)) s else NULL,
      basis = basis,
      bernstein.basis = bernstein.basis
    )

    if (!is.matrix(W.eval))
      W.eval <- matrix(W.eval, nrow = neval)

    if (nrow(W.eval) != neval)
      W.eval <- matrix(W.eval, nrow = neval, byrow = FALSE)

    ridge.used <- rep.int(0.0, neval)

    if (matrix.output) {
      H <- matrix(NA_real_, nrow = neval, ncol = ntrain)
    } else {
      if (is.null(y))
        stop("argument 'y' is required when output='apply'")
      out <- matrix(0.0, nrow = neval, ncol = ncol(y))
    }

    for (i in seq_len(neval)) {
      solve.out <- .npreghat_solve_eval(
        W = W,
        w.eval = W.eval[i, ],
        k = kw[, i],
        ridge.base = ridge
      )

      if (is.null(solve.out))
        stop(sprintf("failed to solve local hat system at evaluation row %d", i))

      ridge.used[i] <- solve.out$ridge
      h.row <- kw[, i] * drop(W %*% solve.out$v)

      if (matrix.output) {
        H[i, ] <- h.row
      } else {
        out[i, ] <- drop(crossprod(h.row, y))
      }
    }

    if (identical(output, "apply")) {
      if (ncol(out) == 1L)
        return(as.vector(out))
      return(out)
    }

    if (constraint.output)
      return(.np_hat_constraint_from_matrix(H, y, "npreghat"))

    class(H) <- c("npreghat", "matrix")
    attr(H, "bws") <- bws
    attr(H, "txdat") <- txdat
    attr(H, "exdat") <- if (no.ex) txdat else exdat
    attr(H, "trainiseval") <- no.ex
    attr(H, "regtype") <- regtype
    attr(H, "degree") <- degree
    attr(H, "basis") <- basis
    attr(H, "bernstein.basis") <- bernstein.basis
    attr(H, "s") <- s
    attr(H, "leave.one.out") <- leave.one.out
    attr(H, "ridge.used") <- ridge.used
    attr(H, "rows.omit") <- rows.omit
    attr(H, "call") <- match.call(expand.dots = FALSE)

    if (!is.null(y)) {
      Hy <- H %*% y
      if (ncol(Hy) == 1L)
        Hy <- as.vector(Hy)
      attr(H, "Hy") <- Hy
    }

    H
  }

npreghat.default <-
  function(bws, ...){
    stop("unsupported 'bws' type; provide an 'rbandwidth' or 'npregression' object")
  }

npreghat.NULL <- npreghat.default

predict.npreghat <-
  function(object,
           newdata = NULL,
           y = NULL,
           output = c("matrix", "apply", "constraint"),
           s = attr(object, "s"),
           leave.one.out = attr(object, "leave.one.out"),
           deriv = NULL,
           ...){
    output <- match.arg(output)
    bws <- attr(object, "bws")
    txdat <- attr(object, "txdat")
    dots <- list(...)

    if (is.null(bws) || is.null(txdat))
      stop("object does not carry 'bws' and training data attributes")

    leave.one.out <- if (is.null(leave.one.out)) FALSE else leave.one.out
    call.args <- list(
      bws = bws,
      txdat = txdat,
      y = y,
      output = output,
      s = s,
      deriv = deriv,
      degree = attr(object, "degree"),
      basis = attr(object, "basis"),
      bernstein.basis = attr(object, "bernstein.basis"),
      leave.one.out = leave.one.out
    )
    if (!is.null(newdata) || !isTRUE(leave.one.out))
      call.args$exdat <- if (is.null(newdata)) attr(object, "exdat") else newdata
    do.call(npreghat, c(call.args, dots))
  }

print.npreghat <- function(x, ...) {
  s <- attr(x, "s")
  cat("\nNonparametric hat operator:", nrow(x), "evaluation x", ncol(x), "training\n")
  cat("regtype:", attr(x, "regtype"),
      "| basis:", attr(x, "basis"),
      "| s:", if (length(s)) paste(s, collapse = ",") else "none", "\n")
  invisible(x)
}

npksum <-
  function(...){
    mc <- match.call(expand.dots = FALSE)
    target <- .np_bw_dispatch_target(dots = mc$..., eval_env = parent.frame())
    UseMethod("npksum", target)
  }

npksum.formula <-
  function(formula, data, newdata, subset, na.action, ...){

    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(mf), nomatch = 0)
    mf <- mf[c(1,m)]
    mf[[1]] <- as.name("model.frame")
    mf.args <- as.list(mf[-1L])
    mf <- do.call(stats::model.frame, mf.args, envir = parent.frame())
    
    tydat <- model.response(mf)
    txdat <- mf[, attr(attr(mf, "terms"),"term.labels"), drop = FALSE]

    miss.new <- missing(newdata)
    if (!miss.new){
      tt <- delete.response(attr(mf,"terms"))
      umf.args <- list(formula = tt, data = newdata)
      umf <- do.call(stats::model.frame, umf.args, envir = parent.frame())
      exdat <- umf[, attr(attr(umf, "terms"),"term.labels"), drop = FALSE]
    }

    call_args <- list(txdat = txdat)
    if (!is.null(tydat))
      call_args$tydat <- tydat
    if (!miss.new)
      call_args$exdat <- exdat

    tobj <- do.call(npksum, c(call_args, list(...)))

    tobj$formula <- formula
    tobj$na.action <- attr(mf, "na.action")
    tobj$terms <- attr(mf,"terms")
    tobj
  }

npksum.integer <-
  function(...) { npksum.numeric(...) }

npksum.numeric <-
  function(bws,
           txdat = stop("training data 'txdat' missing"),
           tydat,
           exdat,
           weights,
           bandwidth.divide, compute.ocg, compute.score, kernel.pow,
           leave.one.out, operator, permutation.operator, return.kernel.weights,
           ...){
    dots <- list(...)
    return.derivative.kernel.weights <- isTRUE(dots$return.derivative.kernel.weights)
    dots$return.derivative.kernel.weights <- NULL

    txdat <- toFrame(txdat)
    if (!missing(exdat)) {
      exdat_was_df <- is.data.frame(exdat)
      exdat_was_matrix <- is.matrix(exdat)
      exdat <- toFrame(exdat)
      # Preserve historical naming: `npksum.numeric()` used to forward `exdat=exdat`
      # into `npksum.default()`, so one-dimensional evaluation data typically carried
      # the column name "exdat" regardless of the caller's symbol.
      if (!exdat_was_df && !exdat_was_matrix && ncol(exdat) == 1L)
        names(exdat) <- "exdat"
    }

    kbw_args <- list(
      bw = bws,
      xdati = untangle(txdat),
      xnames = names(txdat)
    )
    if (!missing(tydat))
      kbw_args$ydati <- untangle(as.data.frame(tydat))

    kbw_args <- c(kbw_args, dots)
    tbw <- do.call(kbandwidth, kbw_args)

    call_args <- list(txdat = txdat, bws = tbw)
    if (!missing(tydat))
      call_args$tydat <- tydat
    if (!missing(exdat))
      call_args$exdat <- exdat
    if (!missing(weights))
      call_args$weights <- weights
    if (!missing(leave.one.out))
      call_args$leave.one.out <- leave.one.out
    if (!missing(kernel.pow))
      call_args$kernel.pow <- kernel.pow
    if (!missing(bandwidth.divide))
      call_args$bandwidth.divide <- bandwidth.divide
    if (!missing(operator))
      call_args$operator <- operator
    if (!missing(permutation.operator))
      call_args$permutation.operator <- permutation.operator
    if (!missing(compute.score))
      call_args$compute.score <- compute.score
    if (!missing(compute.ocg))
      call_args$compute.ocg <- compute.ocg
    if (!missing(return.kernel.weights))
      call_args$return.kernel.weights <- return.kernel.weights
    if (return.derivative.kernel.weights)
      call_args$return.derivative.kernel.weights <- return.derivative.kernel.weights

    do.call(npksum.default, call_args)
  }
    
npksum.default <- 
  function(bws,
           txdat = stop("training data 'txdat' missing"),
           tydat = NULL,
           exdat = NULL,
           weights = NULL,
           bandwidth.divide = FALSE,
           compute.ocg = FALSE,
           compute.score = FALSE,
           kernel.pow = 1.0,
           leave.one.out = FALSE,
           operator = names(ALL_OPERATORS),
           permutation.operator = names(PERMUTATION_OPERATORS),
           return.kernel.weights = FALSE,
           ...){
    dots <- list(...)
    return.derivative.kernel.weights <- isTRUE(dots$return.derivative.kernel.weights)

    miss.ty <- missing(tydat)
    miss.ex <- missing(exdat)
    miss.weights <- missing(weights)

    uo.operators <- c("normal","convolution","integral")

    if(missing(operator))
      operator <- match.arg(operator, choices = names(ALL_OPERATORS))
    else
      operator <- match.arg(operator, choices = names(ALL_OPERATORS), several.ok = TRUE)

    permutation.operator <- match.arg(permutation.operator, choices = names(PERMUTATION_OPERATORS) )
    
    txdat = toFrame(txdat)

    if (!miss.ty && is.vector(tydat))
      tydat = as.matrix(tydat)

#    if(class(bws) != "kbandwidth")
    if(!isa(bws,"kbandwidth"))
        bws = kbandwidth(bws)

    if (!miss.ex){
      exdat = toFrame(exdat)

      if (! txdat %~% exdat )
        stop("'txdat' and 'exdat' are not similar data frames!")
    }

    if (length(bws$bw) != length(txdat))
      stop("length of bandwidth vector does not match number of columns of 'txdat'")

    if(length(operator) == 1)
      operator = rep(operator, length(txdat))
    
    if(length(operator) != length(txdat))
      stop("operator not specified for all variables")

    if(compute.score && compute.ocg)
      stop("compute.score and compute.ocg are mutually exclusive, and cannot be enabled simultaneously")
    if(!all(operator[bws$iuno | bws$iord] %in% uo.operators) && !compute.score)
      stop("unordered and ordered variables may only make use of 'normal', 'convolution' and 'integral' operator types")
    
    operator.num <- ALL_OPERATORS[operator]
    poperator.num <- PERMUTATION_OPERATORS[permutation.operator]
    
    if ((any(bws$icon) &&
         !all(vapply(txdat[, bws$icon, drop = FALSE], inherits, logical(1), c("integer", "numeric")))) ||
        (any(bws$iord) &&
         !all(vapply(txdat[, bws$iord, drop = FALSE], inherits, logical(1), "ordered"))) ||
        (any(bws$iuno) &&
         !all(vapply(txdat[, bws$iuno, drop = FALSE], inherits, logical(1), "factor"))))
      stop("supplied bandwidths do not match 'txdat' in type")

    if (!miss.ty && (nrow(txdat) != nrow(tydat)))
      stop("number of explanatory data 'txdat' and dependent data 'tydat' do not match")

    integer.pow = identical(as.double(as.integer(kernel.pow)),as.double(kernel.pow))
    
    if(!integer.pow)
      stop("'kernel.pow' is not an integer")

    if (!miss.ex && leave.one.out)
      stop("you may not specify 'leave.one.out = TRUE' and provide evaluation data")

    if (!miss.weights && !(is.matrix(weights) && nrow(weights) == nrow(txdat)))
      stop("improperly specified weight matrix")
    ## catch and destroy NA's

    keep.rows <- rep_len(TRUE, nrow(txdat))
    rows.omit <- attr(na.omit(txdat), "na.action")

    if (!miss.ty)
      rows.omit <- union(rows.omit, attr(na.omit(tydat), "na.action"))

    if (!miss.weights)
      rows.omit <- union(rows.omit, attr(na.omit(weights), "na.action"))

    
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Data has no rows without NAs")

    txdat <- txdat[keep.rows,,drop = FALSE]

    if (!miss.ty)
      tydat <- tydat[keep.rows,, drop = FALSE]

    if (!miss.weights)
      weights <- weights[keep.rows,, drop = FALSE]

    if (!miss.ex){
      exdat = na.omit(exdat)
      
      if (nrow(exdat)==0)
        stop("Data has no rows without NAs")
    }

    ## end catch and destroy NA's

    tnrow = nrow(txdat)
    enrow = (if (miss.ex) tnrow else nrow(exdat))
    twncol = (if (miss.weights) 0 else ncol(weights))
    tyncol = (if (miss.ty) 0 else ncol(tydat))

    dim.in = c(twncol, tyncol, enrow)

    dim.out = c(twncol, tyncol, enrow)
    
    length.out = prod(dim.out[which(dim.out > 0)])

    has.permutation <- permutation.operator != "none"
    has.pksum <- has.permutation || compute.ocg

    if (has.pksum){
      npvar <- (if (has.permutation) bws$ncon else 0L) +
        (if (compute.score || compute.ocg) bws$nuno + bws$nord else 0L)
      p.length.out <- npvar*length.out
      p.dim.out <- c(dim.out, max(npvar, 0))      
    }
    else
      p.length.out <- 0

    
    ##dim.out = dim.out[dim.out > 1]

    ## re-assign levels in training and evaluation data to ensure correct
    ## conversion to numeric type.
    
    txdat <- adjustLevels(txdat, bws$xdati, allowNewCells = TRUE)
    
    if (!miss.ex)
      exdat <- adjustLevels(exdat, bws$xdati, allowNewCells = TRUE)

    if (!miss.ex)
      npKernelBoundsCheckEval(exdat, bws$icon, bws$ckerlb, bws$ckerub, argprefix = "cker")

    ## grab the evaluation data before it is converted to numeric
    if(miss.ex)
      teval <- txdat
    else
      teval <- exdat

    ## put the unordered, ordered, and continuous data in their own objects
    ## data that is not a factor is continuous.
    
    txdat = toMatrix(txdat)

    tuno = txdat[, bws$iuno, drop = FALSE]
    tcon = txdat[, bws$icon, drop = FALSE]
    tord = txdat[, bws$iord, drop = FALSE]

    if (!miss.ex){
      exdat = toMatrix(exdat)

      euno = exdat[, bws$iuno, drop = FALSE]
      econ = exdat[, bws$icon, drop = FALSE]
      eord = exdat[, bws$iord, drop = FALSE]

    } else {
      euno = data.frame()
      eord = data.frame()
      econ = data.frame()
    }

    nkw <- (if (return.kernel.weights) tnrow*enrow else 0)

    need.pkw <- isTRUE(return.derivative.kernel.weights) &&
      return.kernel.weights &&
      has.pksum &&
      (p.length.out > 0L)

    return.names <- c("ksum", "kernel.weights", "p.ksum")
    if (need.pkw)
      return.names <- c(return.names, "p.kernel.weights")
      
	    myopti = list(
      num_obs_train = tnrow,
      num_obs_eval = enrow,
      num_uno = bws$nuno,
      num_ord = bws$nord,
      num_con = bws$ncon,
      int_LARGE_SF = SF_ARB,
      BANDWIDTH_reg_extern = switch(bws$type,
        fixed = BW_FIXED,
        generalized_nn = BW_GEN_NN,
        adaptive_nn = BW_ADAP_NN),
      int_MINIMIZE_IO=if (isTRUE(getOption("np.messages"))) IO_MIN_FALSE else IO_MIN_TRUE, 
      kerneval = switch(bws$ckertype,
        gaussian = CKER_GAUSS + bws$ckerorder/2 - 1,
        epanechnikov = CKER_EPAN + bws$ckerorder/2 - 1,
        uniform = CKER_UNI,
        "truncated gaussian" = CKER_TGAUSS),
      ukerneval = switch(bws$ukertype,
        aitchisonaitken = UKER_AIT,
        liracine = UKER_LR),
      okerneval = switch(bws$okertype,
        wangvanryzin = OKER_WANG,
        liracine = OKER_LR,
        nliracine = OKER_NLR,
        "racineliyan" = OKER_RLY),
      miss.ex = miss.ex,
      leave.one.out = leave.one.out, 
      bandwidth.divide = bandwidth.divide,
      mcv.numRow = attr(bws$xmcv, "num.row"),
      wncol = dim.in[1],
      yncol = dim.in[2],
      int_do_tree = if (isTRUE(getOption("np.tree"))) DO_TREE_YES else DO_TREE_NO,
      return.kernel.weights = return.kernel.weights,
      permutation.operator = poperator.num,
	      compute.score = compute.score,
	      compute.ocg = compute.ocg)

	    cker.bounds.c <- npKernelBoundsMarshal(bws$ckerlb[bws$icon], bws$ckerub[bws$icon])
    
   asDouble <- function(data){
	   if (is.null(data)){
	 	   result <- as.double(0.0)
	   }
	   else {
		   result <- as.double(data)
	   }
	   return(result)
   }

    myout <-
      .Call("C_np_kernelsum",
            asDouble(tuno), asDouble(tord), asDouble(tcon),
            asDouble(tydat), asDouble(weights),
            asDouble(euno), asDouble(eord), asDouble(econ),
            as.double(c(bws$bw[bws$icon], bws$bw[bws$iuno], bws$bw[bws$iord])),
            as.double(bws$xmcv), as.double(attr(bws$xmcv, "pad.num")),
            as.integer(c(operator.num[bws$icon], operator.num[bws$iuno], operator.num[bws$iord])),
            as.integer(myopti), as.double(kernel.pow),
            as.integer(length.out),
            as.integer(p.length.out),
            as.integer(nkw),
            as.double(cker.bounds.c$lb),
            as.double(cker.bounds.c$ub),
            PACKAGE="np")[return.names]

    if (dim.out[1] > 1){
      dim(myout[["ksum"]]) <- dim.out
      if (dim.out[2] < 2)
        dim(myout[["ksum"]]) <- dim(myout[["ksum"]])[-2]
    } else if (max(dim.out) > 1) {
      dim(myout[["ksum"]]) <- dim.out[dim.out > 1]
      if (miss.weights)
        myout[["ksum"]] <- aperm(myout[["ksum"]])
    }

    if(return.kernel.weights){
      kw <- matrix(data = myout[["kernel.weights"]], nrow = tnrow, ncol = enrow)
    } else {
      kw <- NULL
    }

    if(need.pkw){
      raw.pkw <- myout[["p.kernel.weights"]]
      if(length(raw.pkw) > 0L){
        if(npvar == 1L){
          p.kw <- matrix(data = raw.pkw, nrow = tnrow, ncol = enrow)
        } else {
          p.kw <- array(data = raw.pkw, dim = c(tnrow, enrow, npvar))
        }
      } else {
        p.kw <- NULL
      }
    } else {
      p.kw <- NULL
    }

    if(has.pksum && (p.length.out > 0)) {
      dim.p <- p.dim.out[which(p.dim.out > 1)]
      if(length(dim.p) == 0) dim.p <- 1

      p.myout <- matrix(data = myout[["p.ksum"]], ncol = npvar)

      ip <- integer(0)

      if(has.permutation && (bws$ncon > 0))
        ip <- which(bws$icon)

      if(compute.ocg || compute.score){
        if(bws$nuno > 0)
          ip <- c(ip,which(bws$iuno))

        if(bws$nord > 0)
          ip <- c(ip,which(bws$iord))
      }

      p.myout[,ip] <- p.myout

      p.myout <- array(data = as.vector(p.myout), dim = dim.p)
    } else {
      p.myout <- NULL
    }
    return( npkernelsum(bws = bws,
                        eval = teval,
                        ksum = myout[["ksum"]],
                        kw = kw,
                        p.ksum = p.myout,
                        p.kw = p.kw,
                        ntrain = tnrow, trainiseval = miss.ex) )

  }

test_that("npksum numeric and formula interfaces agree", {
  set.seed(1)
  n <- 30
  x <- runif(n)
  y <- rnorm(n)
  dat <- data.frame(y = y, x = x)

  k1 <- npksum(txdat = x, tydat = y, bws = 0.5)
  k2 <- npksum(y ~ x, data = dat, bws = 0.5)

  expect_s3_class(k1, "npkernelsum")
  expect_s3_class(k2, "npkernelsum")
  expect_identical(k1$ksum, k2$ksum)
})

test_that("npksum preserves 1D exdat column naming", {
  set.seed(1)
  n <- 10
  x <- runif(n)
  ex <- seq(0, 1, length.out = 3)

  k <- npksum(txdat = x, exdat = ex, bws = 0.5)
  expect_identical(colnames(k$eval), "exdat")
})

test_that("npksum positional bws dispatch matches named bws", {
  set.seed(42)
  n <- 20
  x <- runif(n)
  y <- rnorm(n)

  k_named <- npksum(txdat = x, tydat = y, bws = 0.4)
  k_pos <- npksum(0.4, txdat = x, tydat = y)

  expect_s3_class(k_pos, "npkernelsum")
  expect_identical(k_named$ksum, k_pos$ksum)
})

test_that("npksum formula subset and na.action match explicit data path", {
  set.seed(7)
  n <- 24
  dat <- data.frame(x = runif(n), y = rnorm(n))
  dat$y[c(3, 12)] <- NA_real_

  k_formula <- npksum(y ~ x, data = dat, subset = x > 0.2, na.action = na.omit, bws = 0.45)

  dat2 <- na.omit(subset(dat, x > 0.2))
  k_explicit <- npksum(txdat = dat2$x, tydat = dat2$y, bws = 0.45)

  expect_identical(k_formula$ksum, k_explicit$ksum)
})

test_that("npksum bounded nonfixed helper path works for nearest-neighbor bandwidths", {
  set.seed(20260325)
  n <- 24
  x <- runif(n)
  y <- rnorm(n)
  w <- cbind(y, 1.0)

  k_gnn <- npksum(
    txdat = x,
    tydat = w,
    weights = w,
    leave.one.out = TRUE,
    bandwidth.divide = TRUE,
    bws = 6,
    bwtype = "generalized_nn",
    ckerbound = "range"
  )
  k_adapt <- npksum(
    txdat = x,
    tydat = w,
    weights = w,
    leave.one.out = TRUE,
    bandwidth.divide = TRUE,
    bws = 7,
    bwtype = "adaptive_nn",
    ckerbound = "range"
  )

  expect_s3_class(k_gnn, "npkernelsum")
  expect_s3_class(k_adapt, "npkernelsum")
  expect_true(all(is.finite(as.numeric(k_gnn$ksum))))
  expect_true(all(is.finite(as.numeric(k_adapt$ksum))))
})

.npcdhat_make_xbw <- function(bws, txdat) {
  regtype <- if (is.null(bws$regtype.engine)) bws$regtype else bws$regtype.engine
  basis <- if (is.null(bws$basis.engine)) bws$basis else bws$basis.engine
  degree <- if (is.null(bws$degree.engine)) bws$degree else bws$degree.engine
  bernstein <- if (is.null(bws$bernstein.basis.engine)) bws$bernstein.basis else bws$bernstein.basis.engine

  xbw.args <- list(
    xdat = txdat,
    ydat = rep.int(0.0, nrow(txdat)),
    bws = bws$xbw,
    regtype = regtype,
    bwtype = bws$type,
    bandwidth.compute = FALSE,
    ckertype = bws$cxkertype,
    ckerorder = bws$cxkerorder,
    ckerbound = bws$cxkerbound,
    ckerlb = bws$cxkerlb,
    ckerub = bws$cxkerub,
    ukertype = bws$uxkertype,
    okertype = bws$oxkertype
  )

  if (!is.null(basis))
    xbw.args$basis <- basis
  if (!is.null(degree))
    xbw.args$degree <- degree
  if (!is.null(bernstein))
    xbw.args$bernstein.basis <- bernstein

  do.call(npregbw, xbw.args)
}

.npcdhat_make_xkbw <- function(bws, txdat) {
  kbandwidth.numeric(
    bw = bws$xbw,
    bwscaling = FALSE,
    bwtype = bws$type,
    ckertype = bws$cxkertype,
    ckerorder = bws$cxkerorder,
    ckerbound = bws$cxkerbound,
    ckerlb = bws$cxkerlb,
    ckerub = bws$cxkerub,
    ukertype = bws$uxkertype,
    okertype = bws$oxkertype,
    nobs = nrow(txdat),
    xdati = untangle(txdat),
    ydati = NULL,
    xnames = names(txdat),
    ynames = NULL
  )
}

.npcdhat_make_ybw <- function(bws, tydat) {
  kbandwidth.numeric(
    bw = bws$ybw,
    bwscaling = FALSE,
    bwtype = bws$type,
    ckertype = bws$cykertype,
    ckerorder = bws$cykerorder,
    ckerbound = bws$cykerbound,
    ckerlb = bws$cykerlb,
    ckerub = bws$cykerub,
    ukertype = bws$uykertype,
    okertype = bws$oykertype,
    nobs = nrow(tydat),
    xdati = untangle(tydat),
    ydati = NULL,
    xnames = names(tydat),
    ynames = NULL
  )
}

.npcdhat_resolve_x_s <- function(bws, txdat, s = NULL, deriv = NULL, where = "npcdhat") {
  if (is.null(s) && is.null(deriv))
    return(NULL)

  txdat <- toFrame(txdat)
  ncon <- sum(bws$ixcon)
  con.names <- names(txdat)[bws$ixcon]
  if (!is.null(s) && !is.null(names(s))) {
    sout <- integer(ncon)
    names(sout) <- con.names
    keep <- intersect(names(s), con.names)
    if (length(keep))
      sout[keep] <- as.integer(s[keep])
    s <- sout
  }
  .npreghat_resolve_s(s = s, deriv = deriv, ncon = ncon, con.names = con.names)
}

.npcdhat_has_x_derivative <- function(s) {
  length(s) > 0L && any(s > 0L)
}

.npcdhat_use_adaptive_ratio <- function(bws, x.s = NULL) {
  regtype <- if (is.null(bws$regtype.engine)) bws$regtype else bws$regtype.engine
  identical(bws$type, "adaptive_nn") &&
    !.npcdhat_has_x_derivative(x.s) &&
    identical(as.character(regtype), "lc")
}

.npcdhat_make_xhat_matrix <- function(bws, txdat, exdat, s = NULL) {
  xbw <- .npcdhat_make_xbw(bws = bws, txdat = txdat)
  spec <- npCanonicalConditionalRegSpec(
    regtype = if (is.null(xbw$regtype)) "lc" else as.character(xbw$regtype),
    basis = if (is.null(xbw$basis)) "glp" else xbw$basis,
    degree = xbw$degree,
    bernstein.basis = isTRUE(xbw$bernstein.basis),
    ncon = xbw$ncon,
    where = "npcdhat"
  )
  regtype <- if (is.null(xbw$regtype.engine)) {
    as.character(spec$regtype.engine)
  } else {
    as.character(xbw$regtype.engine)
  }
  basis <- if (is.null(xbw$basis.engine)) spec$basis.engine else xbw$basis.engine
  degree <- if (is.null(xbw$degree.engine)) as.integer(spec$degree.engine) else as.integer(xbw$degree.engine)
  bernstein <- if (is.null(xbw$bernstein.basis.engine)) {
    isTRUE(spec$bernstein.basis.engine)
  } else {
    isTRUE(xbw$bernstein.basis.engine)
  }

  if (identical(regtype, "lc")) {
    if (.npcdhat_has_x_derivative(s)) {
      return(.npreghat_exact_lc_derivative_matrix_from_npksum_chunked(
        bws = xbw,
        txdat = txdat,
        exdat = exdat,
        s = s
      ))
    }

    return(.npreghat_exact_lc_matrix_from_kernel_weights(
      bws = xbw,
      txdat = txdat,
      exdat = exdat
    ))
  }

  if (identical(regtype, "ll")) {
    return(.npreghat_exact_ll_matrix_from_kernel_weights(
      bws = xbw,
      txdat = txdat,
      exdat = exdat,
      s = s
    ))
  }

  if (identical(regtype, "lp")) {
    if (identical(xbw$type, "generalized_nn") && any(degree > 1L)) {
      H <- matrix(NA_real_, nrow = nrow(exdat), ncol = nrow(txdat))
      for (i in seq_len(nrow(exdat))) {
        H[i, ] <- .npreghat_exact_matrix_from_core(
          bws = xbw,
          txdat = txdat,
          exdat = exdat[i, , drop = FALSE],
          s = s
        )[1L, ]
      }
      return(H)
    }

    return(.npreghat_exact_lp_matrix_from_kernel_weights(
      bws = xbw,
      txdat = txdat,
      exdat = exdat,
      s = s,
      basis = basis,
      degree = degree,
      bernstein.basis = bernstein
    ))
  }

  .npreghat_exact_matrix_from_core(
    bws = xbw,
    txdat = txdat,
    exdat = exdat,
    s = s
  )
}

.np_direct_operator_matrix <- function(kbw, txdat, exdat, operator, where) {
  txdat <- toFrame(txdat)
  exdat <- toFrame(exdat)
  op.info <- .np_operator_kernel_weight_scale(
    bws = kbw,
    operator = operator,
    nvars = ncol(txdat),
    where = where
  )
  kw <- .np_kernel_weights_direct(
    bws = op.info$bws,
    txdat = txdat,
    exdat = exdat,
    bandwidth.divide = TRUE,
    operator = op.info$operator
  )

  if (!is.matrix(kw))
    kw <- matrix(kw, nrow = nrow(txdat))
  if (nrow(kw) != nrow(txdat) || ncol(kw) != nrow(exdat))
    stop(sprintf("%s returned unexpected operator shape", where))

  t(kw) / op.info$scale
}

.np_direct_operator_apply <- function(kbw, txdat, exdat, operator, rhs, where) {
  txdat <- toFrame(txdat)
  exdat <- toFrame(exdat)
  rhs <- as.matrix(rhs)
  storage.mode(rhs) <- "double"

  if (nrow(rhs) != nrow(txdat))
    stop(sprintf("%s received RHS with unexpected number of rows", where))

  op.info <- .np_operator_kernel_weight_scale(
    bws = kbw,
    operator = operator,
    nvars = ncol(txdat),
    where = where
  )
  kw <- .np_kernel_weights_direct(
    bws = op.info$bws,
    txdat = txdat,
    exdat = exdat,
    bandwidth.divide = TRUE,
    operator = op.info$operator
  )

  if (!is.matrix(kw))
    kw <- matrix(kw, nrow = nrow(txdat))
  if (nrow(kw) != nrow(txdat) || ncol(kw) != nrow(exdat))
    stop(sprintf("%s returned unexpected operator shape", where))

  crossprod(kw, rhs) / op.info$scale
}

.np_local_operator_ksum <- function(kbw,
                                    txdat,
                                    exdat,
                                    operator,
                                    rhs,
                                    return.kernel.weights = FALSE,
                                    where) {
  txdat <- toFrame(txdat)
  exdat <- toFrame(exdat)
  rhs <- as.matrix(rhs)
  storage.mode(rhs) <- "double"

  if (nrow(rhs) != nrow(txdat))
    stop(sprintf("%s received RHS with unexpected number of rows", where))

  op.info <- .np_operator_kernel_weight_scale(
    bws = kbw,
    operator = operator,
    nvars = ncol(txdat),
    where = where
  )
  bws <- op.info$bws

  txdat <- adjustLevels(txdat, bws$xdati, allowNewCells = TRUE)
  exdat <- adjustLevels(exdat, bws$xdati, allowNewCells = TRUE)
  npKernelBoundsCheckEval(exdat, bws$icon, bws$ckerlb, bws$ckerub, argprefix = "cker")

  txm <- toMatrix(txdat)
  exm <- toMatrix(exdat)
  tuno <- txm[, bws$iuno, drop = FALSE]
  tcon <- txm[, bws$icon, drop = FALSE]
  tord <- txm[, bws$iord, drop = FALSE]
  euno <- exm[, bws$iuno, drop = FALSE]
  econ <- exm[, bws$icon, drop = FALSE]
  eord <- exm[, bws$iord, drop = FALSE]

  tnrow <- nrow(txdat)
  enrow <- nrow(exdat)
  nkw <- if (isTRUE(return.kernel.weights)) tnrow * enrow else 0L

  operator.num <- ALL_OPERATORS[op.info$operator]
  myopti <- list(
    num_obs_train = tnrow,
    num_obs_eval = enrow,
    num_uno = bws$nuno,
    num_ord = bws$nord,
    num_con = bws$ncon,
    int_LARGE_SF = SF_ARB,
    BANDWIDTH_reg_extern = switch(bws$type,
      fixed = BW_FIXED,
      generalized_nn = BW_GEN_NN,
      adaptive_nn = BW_ADAP_NN
    ),
    int_MINIMIZE_IO = if (isTRUE(getOption("np.messages"))) IO_MIN_FALSE else IO_MIN_TRUE,
    kerneval = switch(bws$ckertype,
      gaussian = CKER_GAUSS + bws$ckerorder / 2 - 1,
      epanechnikov = CKER_EPAN + bws$ckerorder / 2 - 1,
      uniform = CKER_UNI,
      "truncated gaussian" = CKER_TGAUSS
    ),
    ukerneval = switch(bws$ukertype,
      aitchisonaitken = UKER_AIT,
      liracine = UKER_LR
    ),
    okerneval = switch(bws$okertype,
      wangvanryzin = OKER_WANG,
      liracine = OKER_LR,
      nliracine = OKER_NLR,
      racineliyan = OKER_RLY
    ),
    miss.ex = FALSE,
    leave.one.out = FALSE,
    bandwidth.divide = TRUE,
    mcv.numRow = attr(bws$xmcv, "num.row"),
    wncol = 0L,
    yncol = ncol(rhs),
    int_do_tree = if (isTRUE(getOption("np.tree"))) DO_TREE_YES else DO_TREE_NO,
    return.kernel.weights = isTRUE(return.kernel.weights),
    permutation.operator = PERMUTATION_OPERATORS[["none"]],
    compute.score = FALSE,
    compute.ocg = FALSE
  )

  cker.bounds.c <- npKernelBoundsMarshal(bws$ckerlb[bws$icon], bws$ckerub[bws$icon])

  asDouble <- function(data) {
    if (is.null(data)) as.double(0.0) else as.double(data)
  }

  myout <- .Call(
    "C_np_kernelsum",
    asDouble(tuno), asDouble(tord), asDouble(tcon),
    asDouble(rhs), as.double(0.0),
    asDouble(euno), asDouble(eord), asDouble(econ),
    as.double(c(bws$bw[bws$icon], bws$bw[bws$iuno], bws$bw[bws$iord])),
    as.double(bws$xmcv), as.double(attr(bws$xmcv, "pad.num")),
    as.integer(c(operator.num[bws$icon], operator.num[bws$iuno], operator.num[bws$iord])),
    as.integer(myopti), as.double(1.0),
    as.integer(enrow),
    as.integer(0L),
    as.integer(nkw),
    as.double(cker.bounds.c$lb),
    as.double(cker.bounds.c$ub),
    PACKAGE = "np"
  )

  out <- myout[["ksum"]]
  if (!is.matrix(out))
    out <- matrix(out, nrow = enrow, ncol = ncol(rhs))

  kw <- NULL
  if (isTRUE(return.kernel.weights))
    kw <- matrix(as.double(myout[["kernel.weights"]]), nrow = tnrow, ncol = enrow)

  list(ksum = out, kernel.weights = kw)
}

.np_exact_operator_apply <- function(kbw, txdat, exdat, operator, rhs, where) {
  rhs <- as.matrix(rhs)
  storage.mode(rhs) <- "double"

  out <- .np_local_operator_ksum(
    kbw = kbw,
    txdat = txdat,
    exdat = exdat,
    operator = operator,
    rhs = rhs,
    return.kernel.weights = FALSE,
    where = where
  )$ksum

  if (!is.matrix(out))
    out <- matrix(out, nrow = nrow(exdat), ncol = ncol(rhs))

  if (ncol(out) == 1L)
    out[, 1L, drop = FALSE]
  else
    out
}

.np_exact_operator_matrix <- function(kbw, txdat, exdat, operator, where) {
  txdat <- toFrame(txdat)
  exdat <- toFrame(exdat)

  probe <- .np_local_operator_ksum(
    kbw = kbw,
    txdat = txdat,
    exdat = exdat,
    operator = operator,
    rhs = matrix(1.0, nrow = nrow(txdat), ncol = 1L),
    return.kernel.weights = TRUE,
    where = where
  )

  kw <- probe$kernel.weights
  if (!is.matrix(kw))
    kw <- matrix(kw, nrow = nrow(txdat))
  if (nrow(kw) != nrow(txdat) || ncol(kw) != nrow(exdat))
    stop(sprintf("%s returned unexpected kernel-weight shape", where))

  H.raw <- t(kw)
  row.sum <- drop(H.raw %*% rep.int(1.0, nrow(txdat)))
  row.sum[abs(row.sum) < .Machine$double.xmin] <- .Machine$double.xmin
  row.scale <- as.vector(probe$ksum[, 1L]) / row.sum

  sweep(H.raw, 1L, row.scale, "*")
}

.npcdhat_make_kernel_matrix <- function(kbw, txdat, exdat, operator) {
  if (!identical(kbw$type, "fixed")) {
    return(.np_exact_operator_matrix(
      kbw = kbw,
      txdat = txdat,
      exdat = exdat,
      operator = operator,
      where = "conditional hat exact kernel matrix"
    ))
  }

  .np_direct_operator_matrix(
    kbw = kbw,
    txdat = txdat,
    exdat = exdat,
    operator = operator,
    where = "conditional hat direct kernel matrix"
  )
}

.npcdhat_ratio_matrix <- function(bws, txdat, tydat, exdat, eydat, operator) {
  xkbw <- .npcdhat_make_xkbw(bws = bws, txdat = txdat)
  ybw <- .npcdhat_make_ybw(bws = bws, tydat = tydat)
  Kx <- .npcdhat_make_kernel_matrix(
    kbw = xkbw,
    txdat = txdat,
    exdat = exdat,
    operator = rep.int("normal", ncol(txdat))
  )
  Ky <- .npcdhat_make_kernel_matrix(
    kbw = ybw,
    txdat = tydat,
    exdat = eydat,
    operator = rep.int(operator, ncol(tydat))
  )

  denom <- rowSums(Kx) / nrow(txdat)
  sweep((Kx * Ky) / nrow(txdat), 1L, pmax(denom, .Machine$double.eps), "/")
}

.npcdhat_exact_matrix <- function(bws, txdat, tydat, exdat, eydat, operator, x.s = NULL) {
  if (.npcdhat_use_adaptive_ratio(bws = bws, x.s = x.s)) {
    return(.npcdhat_ratio_matrix(
      bws = bws,
      txdat = txdat,
      tydat = tydat,
      exdat = exdat,
      eydat = eydat,
      operator = operator
    ))
  }

  ybw <- .npcdhat_make_ybw(bws = bws, tydat = tydat)
  Hx <- .npcdhat_make_xhat_matrix(
    bws = bws,
    txdat = txdat,
    exdat = exdat,
    s = x.s
  )
  Gy <- .npcdhat_make_kernel_matrix(
    kbw = ybw,
    txdat = tydat,
    exdat = eydat,
    operator = rep.int(operator, ncol(tydat))
  )

  Hx * Gy
}

.npcdhat_exact_apply <- function(bws, txdat, tydat, exdat, eydat, rhs, operator, x.s = NULL) {
  if (.npcdhat_use_adaptive_ratio(bws = bws, x.s = x.s)) {
    H <- .npcdhat_ratio_matrix(
      bws = bws,
      txdat = txdat,
      tydat = tydat,
      exdat = exdat,
      eydat = eydat,
      operator = operator
    )
    return(H %*% rhs)
  }

  ybw <- .npcdhat_make_ybw(bws = bws, tydat = tydat)
  Hx <- .npcdhat_make_xhat_matrix(
    bws = bws,
    txdat = txdat,
    exdat = exdat,
    s = x.s
  )
  Gy <- .npcdhat_make_kernel_matrix(
    kbw = ybw,
    txdat = tydat,
    exdat = eydat,
    operator = rep.int(operator, ncol(tydat))
  )

  (Hx * Gy) %*% rhs
}

.npcdhat_core <- function(bws,
                          txdat,
                          tydat,
                          exdat,
                          eydat,
                          y,
                          output,
                          operator,
                          x.deriv = NULL,
                          x.s = NULL,
                          class_name,
                          where) {
  output <- match.arg(output, c("matrix", "apply"))

  if (xor(is.null(exdat), is.null(eydat)))
    stop("evaluation data must be supplied for both 'exdat' and 'eydat'")

  no.exy <- is.null(exdat)

  txdat <- toFrame(txdat)
  tydat <- toFrame(tydat)

  if (!no.exy) {
    exdat <- toFrame(exdat)
    eydat <- toFrame(eydat)

    if (!(txdat %~% exdat))
      stop("'txdat' and 'exdat' are not similar data frames!")
    if (!(tydat %~% eydat))
      stop("'tydat' and 'eydat' are not similar data frames!")
  }

  if (!is.null(y)) {
    if (is.factor(y) || is.vector(y)) {
      y <- matrix(as.double(y), ncol = 1L)
    } else {
      y <- as.matrix(y)
      storage.mode(y) <- "double"
    }

    if (nrow(y) != nrow(txdat))
      stop("number of rows in 'y' must match the number of training rows in 'txdat'")
  }

  keep.rows <- rep_len(TRUE, nrow(txdat))
  rows.omit.train <- attr(stats::na.omit(data.frame(txdat, tydat)), "na.action")
  if (!is.null(y))
    rows.omit.train <- union(rows.omit.train, attr(stats::na.omit(as.data.frame(y)), "na.action"))

  if (length(rows.omit.train) > 0L)
    keep.rows[as.integer(rows.omit.train)] <- FALSE

  if (!any(keep.rows))
    stop("Training data has no rows without NAs")

  txdat <- txdat[keep.rows, , drop = FALSE]
  tydat <- tydat[keep.rows, , drop = FALSE]
  if (!is.null(y))
    y <- y[keep.rows, , drop = FALSE]

  if (!no.exy) {
    keep.eval <- rep_len(TRUE, nrow(exdat))
    rows.omit.eval <- attr(stats::na.omit(data.frame(exdat, eydat)), "na.action")
    if (length(rows.omit.eval) > 0L)
      keep.eval[as.integer(rows.omit.eval)] <- FALSE

    if (!any(keep.eval))
      stop("Evaluation data has no rows without NAs")

    exdat <- exdat[keep.eval, , drop = FALSE]
    eydat <- eydat[keep.eval, , drop = FALSE]
  } else {
    exdat <- txdat
    eydat <- tydat
  }

  if (!is.null(x.s) || !is.null(x.deriv)) {
    x.s <- .npcdhat_resolve_x_s(
      bws = bws,
      txdat = txdat,
      s = x.s,
      deriv = x.deriv,
      where = where
    )
  } else {
    x.s <- NULL
  }

  if (identical(output, "apply")) {
    if (is.null(y))
      stop("argument 'y' is required when output='apply'")

    out <- .npcdhat_exact_apply(
      bws = bws,
      txdat = txdat,
      tydat = tydat,
      exdat = exdat,
      eydat = eydat,
      rhs = y,
      operator = operator,
      x.s = x.s
    )
    if (ncol(out) == 1L)
      return(as.vector(out))
    return(out)
  }

  H <- .npcdhat_exact_matrix(
    bws = bws,
    txdat = txdat,
    tydat = tydat,
    exdat = exdat,
    eydat = eydat,
    operator = operator,
    x.s = x.s
  )

  class(H) <- c(class_name, "matrix")
  attr(H, "bws") <- bws
  attr(H, "txdat") <- txdat
  attr(H, "tydat") <- tydat
  attr(H, "exdat") <- exdat
  attr(H, "eydat") <- eydat
  attr(H, "trainiseval") <- no.exy
  attr(H, "rows.omit.train") <- rows.omit.train
  attr(H, "s") <- x.s
  attr(H, "call") <- match.call(expand.dots = FALSE)

  if (!is.null(y)) {
    Hy <- H %*% y
    if (ncol(Hy) == 1L)
      Hy <- as.vector(Hy)
    attr(H, "Hy") <- Hy
  }

  H
}

.np_degree_eval_key <- function(degree) {
  paste(as.integer(degree), collapse = ",")
}

.np_degree_better <- function(candidate, incumbent, direction = c("min", "max")) {
  direction <- match.arg(direction)
  if (!is.finite(candidate))
    return(FALSE)
  if (!is.finite(incumbent))
    return(TRUE)
  if (identical(direction, "min")) candidate < incumbent else candidate > incumbent
}

.np_degree_normalize_bounds <- function(ncon,
                                        degree.min = NULL,
                                        degree.max = NULL,
                                        default.max = 3L) {
  if (!is.numeric(ncon) || length(ncon) != 1L || is.na(ncon) || ncon < 1L)
    stop("automatic degree search requires at least one continuous regressor")

  normalize_one <- function(x, nm, default) {
    if (is.null(x))
      x <- default
    if (!is.numeric(x) || anyNA(x) || any(!is.finite(x)) || any(x != floor(x)))
      stop(sprintf("%s must contain finite integers", nm))
    if (length(x) == 1L)
      x <- rep.int(as.integer(x), ncon)
    if (length(x) != ncon)
      stop(sprintf("%s must have length 1 or %d", nm, ncon))
    as.integer(x)
  }

  lower <- normalize_one(degree.min, "degree.min", 0L)
  upper <- normalize_one(degree.max, "degree.max", default.max)

  if (any(lower < 0L))
    stop("degree.min must contain non-negative integers")
  if (any(upper < lower))
    stop("degree.max must be greater than or equal to degree.min coordinate-wise")
  if (any(upper > .np_glp_degree_hard_max))
    stop(sprintf("degree.max must contain integers in [0,%d]", .np_glp_degree_hard_max))

  .np_warn_high_glp_degree(upper, argname = "degree.max")

  candidates <- lapply(seq_len(ncon), function(j) seq.int(lower[j], upper[j]))

  list(lower = lower, upper = upper, candidates = candidates)
}

.np_degree_search_engine_controls <- function(search.engine) {
  match.arg(search.engine, c("nomad+powell", "cell", "nomad"))
}

.np_degree_trace_to_frame <- function(records, objective_name = "objective") {
  if (!length(records)) {
    out <- data.frame(
      trace_id = integer(0),
      eval_id = integer(0),
      degree = character(0),
      objective = numeric(0),
      status = character(0),
      cached = logical(0),
      message = character(0),
      elapsed = numeric(0),
      num.feval = numeric(0),
      stringsAsFactors = FALSE
    )
  } else {
    out <- data.frame(
      trace_id = vapply(records, `[[`, integer(1), "trace_id"),
      eval_id = vapply(records, `[[`, integer(1), "eval_id"),
      degree = vapply(records, function(x) paste(x$degree, collapse = ","), character(1)),
      objective = vapply(records, `[[`, numeric(1), "objective"),
      status = vapply(records, `[[`, character(1), "status"),
      cached = vapply(records, `[[`, logical(1), "cached"),
      message = vapply(records, function(x) {
        if (is.null(x$message)) "" else as.character(x$message)
      }, character(1)),
      elapsed = vapply(records, `[[`, numeric(1), "elapsed"),
      num.feval = vapply(records, function(x) {
        if (is.null(x$num.feval) || length(x$num.feval) != 1L || is.na(x$num.feval))
          NA_real_
        else as.numeric(x$num.feval)
      }, numeric(1)),
      stringsAsFactors = FALSE
    )
  }

  if (!identical(objective_name, "objective"))
    names(out)[names(out) == "objective"] <- objective_name

  out
}

.np_degree_clip_to_grid <- function(degree, candidates) {
  out <- as.integer(degree)
  for (j in seq_along(candidates)) {
    cand <- as.integer(candidates[[j]])
    out[j] <- cand[which.min(abs(cand - out[j]))]
  }
  out
}

.np_degree_grid_size <- function(candidates) {
  prod(vapply(candidates, length, integer(1)))
}

.np_degree_format_degree <- function(degree) {
  sprintf("(%s)", paste(as.integer(degree), collapse = ","))
}

.np_degree_format_objective <- function(value) {
  value <- as.numeric(value)[1L]
  if (!is.finite(value))
    return("NA")

  formatC(value, digits = 6L, format = "fg", flag = "#")
}

.np_degree_format_candidate_set <- function(cand) {
  cand <- as.integer(cand)
  if (!length(cand))
    return("c()")

  if (length(cand) >= 2L && all(diff(cand) == 1L))
    return(sprintf("%s:%s", cand[1L], cand[length(cand)]))

  sprintf("c(%s)", paste(cand, collapse = ","))
}

.np_degree_format_search_space <- function(candidates) {
  paste(
    vapply(candidates, .np_degree_format_candidate_set, character(1L)),
    collapse = " x "
  )
}

.np_degree_coordinate_visit_cap <- function(candidates,
                                            max_cycles,
                                            restart_total) {
  restart_total <- npValidatePositiveInteger(restart_total, "restart_total")
  max_cycles <- npValidatePositiveInteger(max_cycles, "degree.max.cycles")
  per_cycle <- sum(vapply(candidates, function(cand) {
    max(0L, length(cand) - 1L)
  }, integer(1L)))
  restart_total * (1L + max_cycles * per_cycle)
}

.np_degree_progress_context <- function() {
  label <- .np_progress_runtime$bandwidth_context_label
  if (is.null(label))
    return(NULL)

  label <- as.character(label)[1L]
  if (is.na(label) || !nzchar(label))
    return(NULL)

  label
}

.np_degree_progress_label <- function() {
  "Selecting degree and bandwidth"
}

.np_degree_progress_context_fields <- function() {
  context <- .np_degree_progress_context()
  if (is.null(context)) character(0L) else context
}

.np_degree_progress_best_detail <- function(best_record,
                                            objective_name = "objective") {
  if (is.null(best_record))
    return("best pending")

  sprintf(
    "best %s, %s=%s",
    .np_degree_format_degree(best_record$degree),
    objective_name,
    .np_degree_format_objective(best_record$objective)
  )
}

.np_degree_progress_best_degree_detail <- function(best_record) {
  if (is.null(best_record))
    return("best pending")

  sprintf("best %s", .np_degree_format_degree(best_record$degree))
}

.np_degree_progress_detail <- function(phase,
                                       best_record,
                                       objective_name = "objective",
                                       step = NULL,
                                       step_max = NULL,
                                       cycle = NULL,
                                       max_cycles = NULL,
                                       coordinate = NULL,
                                       ncon = NULL,
                                       restart_index = NULL,
                                       restart_total = NULL,
                                       evaluating = NULL) {
  phase <- switch(
    as.character(phase)[1L],
    coordinate = "coord",
    exhaustive = "exhaustive",
    verify = "verify",
    as.character(phase)[1L]
  )
  fields <- c(.np_degree_progress_context_fields(), phase)

  if (!is.null(step) && !is.na(step) && step >= 1L) {
    if (!is.null(step_max) && !is.na(step_max) && step_max >= step) {
      fields <- c(fields, sprintf("step %s/%s", format(step), format(step_max)))
    } else {
      fields <- c(fields, sprintf("step %s", format(step)))
    }
  }

  if (!is.null(restart_total) && !is.na(restart_total) && restart_total > 1L &&
      !is.null(restart_index) && !is.na(restart_index) && restart_index >= 1L) {
    fields <- c(fields, sprintf("restart %s/%s", format(restart_index), format(restart_total)))
  }

  if (!is.null(cycle) && !is.na(cycle) && cycle >= 1L &&
      !is.null(max_cycles) && !is.na(max_cycles) && max_cycles >= 1L) {
    fields <- c(fields, sprintf("cycle %s/%s", format(cycle), format(max_cycles)))
  }

  if (!is.null(coordinate) && !is.na(coordinate) && coordinate >= 1L &&
      !is.null(ncon) && !is.na(ncon) && ncon >= 1L) {
    fields <- c(fields, sprintf("coord %s/%s", format(coordinate), format(ncon)))
  }

  if (!is.null(evaluating)) {
    fields <- c(fields, sprintf("deg %s", .np_degree_format_degree(evaluating)))
  }

  fields <- c(fields, .np_degree_progress_best_detail(
    best_record = best_record,
    objective_name = objective_name
  ))

  paste(fields, collapse = ", ")
}

.np_degree_progress_begin <- function(total = NULL, detail = NULL) {
  state <- .np_progress_begin(
    label = .np_degree_progress_label(),
    total = total,
    domain = "general",
    surface = "bandwidth"
  )

  .np_progress_show_now(
    state = state,
    done = if (is.null(total)) NULL else 0L,
    detail = detail
  )
}

.np_degree_progress_step <- function(state,
                                     done = NULL,
                                     detail = NULL,
                                     force = FALSE) {
  if (is.null(state))
    return(state)

  if (isTRUE(force)) {
    return(.np_progress_step_at(
      state = state,
      now = .np_progress_now(),
      done = done,
      detail = detail,
      force = TRUE
    ))
  }

  .np_progress_step(
    state = state,
    done = done,
    detail = detail
  )
}

.np_degree_progress_end <- function(state,
                                    detail = NULL,
                                    interrupted = FALSE) {
  if (is.null(state))
    return(invisible(state))

  if (isTRUE(interrupted)) {
    return(invisible(.np_progress_abort(
      state = state,
      detail = paste("Polynomial degree search interrupted;", detail)
    )))
  }

  invisible(.np_progress_end(state = state, detail = detail))
}

.np_degree_check_grid_budget <- function(candidates,
                                         method,
                                         verify = FALSE,
                                         warn_threshold = getOption("np.degree.search.warn.grid", 10000L),
                                         max_threshold = getOption("np.degree.search.max.grid", Inf)) {
  grid.size <- .np_degree_grid_size(candidates)

  if (!is.finite(grid.size) || grid.size < 1)
    stop("searched degree grid is invalid")

  if ((identical(method, "exhaustive") || isTRUE(verify)) &&
      is.finite(max_threshold) &&
      (grid.size > max_threshold)) {
    stop(
      sprintf(
        "exhaustive degree search requires %s degree combinations, exceeding the configured limit of %s",
        format(grid.size, scientific = FALSE, trim = TRUE),
        format(max_threshold, scientific = FALSE, trim = TRUE)
      )
    )
  }

  if ((identical(method, "exhaustive") || isTRUE(verify)) &&
      is.finite(warn_threshold) &&
      (grid.size > warn_threshold)) {
    warning(
      sprintf(
        "exhaustive degree search will evaluate %s degree combinations; this may take considerable time",
        format(grid.size, scientific = FALSE, trim = TRUE)
      ),
      call. = FALSE
    )
  }

  grid.size
}

.np_degree_iterate_grid <- function(candidates, visit, prefix = integer(0), index = 1L) {
  if (index > length(candidates))
    return(isTRUE(visit(as.integer(prefix))))

  for (d in candidates[[index]]) {
    keep_going <- .np_degree_iterate_grid(
      candidates = candidates,
      visit = visit,
      prefix = c(prefix, as.integer(d)),
      index = index + 1L
    )
    if (!isTRUE(keep_going))
      return(FALSE)
  }

  TRUE
}

.np_degree_restart_starts <- function(candidates, restarts, exclude = NULL) {
  restarts <- npValidateNonNegativeInteger(restarts, "degree.restarts")
  if (restarts == 0L)
    return(list())

  ncon <- length(candidates)
  lower <- vapply(candidates, function(cand) as.integer(cand[1L]), integer(1))
  upper <- vapply(candidates, function(cand) as.integer(cand[length(cand)]), integer(1))
  midpoint <- vapply(candidates, function(cand) {
    cand <- as.integer(cand)
    cand[ceiling(length(cand) / 2)]
  }, integer(1))

  seen <- new.env(hash = TRUE, parent = emptyenv())
  starts <- list()

  remember <- function(degree) {
    assign(.np_degree_eval_key(degree), TRUE, envir = seen)
    invisible(NULL)
  }

  already_seen <- function(degree) {
    exists(.np_degree_eval_key(degree), envir = seen, inherits = FALSE)
  }

  if (!is.null(exclude)) {
    if (is.list(exclude)) {
      for (deg in exclude)
        remember(as.integer(deg))
    } else {
      remember(as.integer(exclude))
    }
  }

  add_start <- function(degree) {
    degree <- as.integer(degree)
    if (already_seen(degree))
      return(FALSE)
    remember(degree)
    starts[[length(starts) + 1L]] <<- degree
    TRUE
  }

  seeds <- list(lower, upper, midpoint)
  for (seed in seeds) {
    add_start(seed)
    if (length(starts) >= restarts)
      return(starts)
  }

  k <- 1L
  while (length(starts) < restarts) {
    degree <- vapply(seq_len(ncon), function(j) {
      cand <- as.integer(candidates[[j]])
      idx <- ((k * j) %% length(cand)) + 1L
      cand[idx]
    }, integer(1))
    add_start(degree)
    k <- k + 1L
  }

  starts
}

.np_degree_search_exhaustive <- function(state,
                                         candidates,
                                         phase = "exhaustive",
                                         objective_name = "objective") {
  visited <- 0L

  .np_degree_iterate_grid(candidates, function(degree) {
    if (!is.null(state$progress_state) &&
        identical(state$progress_state$renderer, "single_line")) {
      state$progress_state <- .np_degree_progress_step(
        state = state$progress_state,
        done = visited,
        detail = .np_degree_progress_detail(
          phase = phase,
          best_record = state$best_record,
          objective_name = objective_name,
          evaluating = degree
        ),
        force = TRUE
      )
    }

    state$evaluate(degree)
    visited <<- visited + 1L
    state$progress_state <- .np_degree_progress_step(
      state = state$progress_state,
      done = visited,
      detail = .np_degree_progress_detail(
        phase = phase,
        best_record = state$best_record,
        objective_name = objective_name,
        evaluating = degree
      )
    )
    !isTRUE(state$interrupted)
  })

  invisible(visited)
}

.np_degree_search_coordinate <- function(state,
                                         candidates,
                                         start_degree,
                                         restart_starts,
                                         max_cycles,
                                         direction,
                                         objective_name = "objective") {
  restart_total <- length(restart_starts) + 1L
  step_max <- .np_degree_coordinate_visit_cap(
    candidates = candidates,
    max_cycles = max_cycles,
    restart_total = restart_total
  )

  run_coordinate <- function(initial_degree, restart_index) {
    current <- .np_degree_clip_to_grid(initial_degree, candidates)
    current_record <- state$get_cached(current)
    if (!is.null(state$progress_state) &&
        identical(state$progress_state$renderer, "single_line")) {
      state$progress_state <- .np_degree_progress_step(
        state = state$progress_state,
        done = NULL,
        detail = .np_degree_progress_detail(
          phase = "coordinate",
          best_record = state$best_record,
          objective_name = objective_name,
          step = state$visit_id,
          step_max = step_max,
          restart_index = restart_index,
          restart_total = restart_total,
          evaluating = current
        ),
        force = TRUE
      )
    }
    if (is.null(current_record))
      current_record <- state$evaluate(current)
    state$progress_state <- .np_degree_progress_step(
      state = state$progress_state,
      done = NULL,
      detail = .np_degree_progress_detail(
        phase = "coordinate",
        best_record = state$best_record,
        objective_name = objective_name,
        step = max(1L, state$visit_id - 1L),
        step_max = step_max,
        restart_index = restart_index,
        restart_total = restart_total,
        evaluating = current
      )
    )
    for (cycle in seq_len(max_cycles)) {
      prior <- current
      for (j in seq_along(candidates)) {
        best_local <- current[j]
        best_local_value <- if (!is.null(current_record) &&
                                identical(current_record$status, "ok") &&
                                is.finite(current_record$objective)) {
          current_record$objective
        } else if (identical(direction, "min")) {
          Inf
        } else {
          -Inf
        }
        best_local_record <- current_record
        for (d in candidates[[j]]) {
          if (identical(as.integer(d), current[j]))
            next
          trial <- current
          trial[j] <- as.integer(d)
          if (!is.null(state$progress_state) &&
              identical(state$progress_state$renderer, "single_line")) {
            state$progress_state <- .np_degree_progress_step(
              state = state$progress_state,
              done = NULL,
              detail = .np_degree_progress_detail(
                phase = "coordinate",
                best_record = state$best_record,
                objective_name = objective_name,
                step = state$visit_id,
                step_max = step_max,
                cycle = cycle,
                max_cycles = max_cycles,
                coordinate = j,
                ncon = length(candidates),
                restart_index = restart_index,
                restart_total = restart_total,
                evaluating = trial
              ),
              force = TRUE
            )
          }
          rec <- state$evaluate(trial)
          state$progress_state <- .np_degree_progress_step(
            state = state$progress_state,
            done = NULL,
            detail = .np_degree_progress_detail(
              phase = "coordinate",
              best_record = state$best_record,
              objective_name = objective_name,
              step = max(1L, state$visit_id - 1L),
              step_max = step_max,
              cycle = cycle,
              max_cycles = max_cycles,
              coordinate = j,
              ncon = length(candidates),
              restart_index = restart_index,
              restart_total = restart_total,
              evaluating = trial
            )
          )
          if (isTRUE(state$interrupted))
            return(invisible(NULL))
          if (!identical(rec$status, "ok") || !is.finite(rec$objective))
            next
          if (.np_degree_better(rec$objective, best_local_value, direction = direction)) {
            best_local <- as.integer(d)
            best_local_value <- rec$objective
            best_local_record <- rec
          }
        }
        current[j] <- best_local
        current_record <- best_local_record
      }
      if (identical(current, prior) || isTRUE(state$interrupted))
        break
    }
    invisible(NULL)
  }

  run_coordinate(start_degree, restart_index = 1L)
  if (isTRUE(state$interrupted) || !length(restart_starts))
    return(invisible(NULL))

  for (i in seq_along(restart_starts)) {
    run_coordinate(restart_starts[[i]], restart_index = i + 1L)
    if (isTRUE(state$interrupted))
      break
  }

  invisible(NULL)
}

.np_degree_search <- function(method = c("coordinate", "exhaustive"),
                              candidates,
                              baseline_degree,
                              start_degree,
                              restarts = 0L,
                              max_cycles = 20L,
                              verify = FALSE,
                              eval_fun,
                              direction = c("min", "max"),
                              trace_level = c("full", "none"),
                              objective_name = "objective") {
  method <- match.arg(method)
  direction <- match.arg(direction)
  trace_level <- match.arg(trace_level)
  restarts <- npValidateNonNegativeInteger(restarts, "degree.restarts")
  max_cycles <- npValidatePositiveInteger(max_cycles, "degree.max.cycles")
  grid.size <- .np_degree_check_grid_budget(
    candidates = candidates,
    method = method,
    verify = verify
  )

  state <- new.env(parent = emptyenv())
  state$cache <- new.env(hash = TRUE, parent = emptyenv())
  state$trace_records <- list()
  state$trace_enabled <- identical(trace_level, "full")
  state$trace_id <- 0L
  state$eval_id <- 0L
  state$visit_id <- 0L
  state$cached_visits <- 0L
  state$best_record <- NULL
  state$best_payload <- NULL
  state$baseline_record <- NULL
  state$interrupted <- FALSE
  state$progress_state <- NULL

  state$record_trace <- function(rec) {
    if (!isTRUE(state$trace_enabled))
      return(invisible(rec))
    state$trace_id <- state$trace_id + 1L
    rec$trace_id <- state$trace_id
    state$trace_records[[length(state$trace_records) + 1L]] <- rec
    invisible(rec)
  }

  state$update_best <- function(rec, payload = NULL) {
    if (!identical(rec$status, "ok") || !is.finite(rec$objective))
      return(invisible(NULL))
    incumbent <- if (is.null(state$best_record)) {
      if (identical(direction, "min")) Inf else -Inf
    } else {
      state$best_record$objective
    }
    if (.np_degree_better(rec$objective, incumbent, direction = direction)) {
      state$best_record <- rec
      if (!is.null(payload))
        state$best_payload <- payload
    }
    invisible(NULL)
  }

  state$get_cached <- function(degree) {
    key <- .np_degree_eval_key(as.integer(degree))
    if (!exists(key, envir = state$cache, inherits = FALSE))
      return(NULL)
    get(key, envir = state$cache, inherits = FALSE)
  }

  state$evaluate <- function(degree) {
    degree <- as.integer(degree)
    state$visit_id <- state$visit_id + 1L
    key <- .np_degree_eval_key(degree)

    if (exists(key, envir = state$cache, inherits = FALSE)) {
      cached <- get(key, envir = state$cache, inherits = FALSE)
      state$cached_visits <- state$cached_visits + 1L
      cached$cached <- TRUE
      return(cached)
    }

    started <- proc.time()[3]
    payload <- NULL
    status <- "ok"
    msg <- NULL
    objective <- NA_real_
    num.feval <- NA_real_

    result <- tryCatch(
      {
        .np_progress_bandwidth_set_context(
          sprintf("deg %s", .np_degree_format_degree(degree))
        )
        on.exit(.np_progress_bandwidth_set_context(NULL), add = TRUE)
        eval_fun(degree)
      },
      interrupt = function(e) {
        state$interrupted <- TRUE
        NULL
      },
      error = function(e) {
        status <<- "error"
        msg <<- conditionMessage(e)
        NULL
      }
    )

    if (isTRUE(state$interrupted)) {
      rec <- list(
        eval_id = NA_integer_,
        degree = degree,
        objective = NA_real_,
        status = "interrupt",
        cached = FALSE,
        message = "search interrupted",
        elapsed = proc.time()[3] - started,
        num.feval = NA_real_
      )
      return(state$record_trace(rec))
    }

    if (!is.null(result)) {
      if (!is.list(result) || is.null(result$objective))
        stop("fixed-degree evaluator must return a list containing 'objective'")
      objective <- as.numeric(result$objective[1L])
      payload <- result$payload
      if (!is.null(result$num.feval))
        num.feval <- as.numeric(result$num.feval[1L])
    }

    state$eval_id <- state$eval_id + 1L
    rec <- list(
      eval_id = state$eval_id,
      degree = degree,
      objective = objective,
      status = status,
      cached = FALSE,
      message = msg,
      elapsed = proc.time()[3] - started,
      num.feval = num.feval
    )
    assign(key, rec, envir = state$cache)
    state$record_trace(rec)
    if (identical(status, "ok"))
      state$update_best(rec, payload = payload)
    rec
  }

  baseline_degree <- as.integer(baseline_degree)
  start_degree <- as.integer(start_degree)
  restart_starts <- .np_degree_restart_starts(
    candidates = candidates,
    restarts = restarts,
    exclude = list(start_degree, baseline_degree)
  )

  tryCatch({
    .np_progress_note(sprintf(
      "Automatic polynomial degree search baseline %s",
      .np_degree_format_degree(baseline_degree)
    ))
    state$baseline_record <- state$evaluate(baseline_degree)

    if (identical(method, "exhaustive")) {
      .np_progress_note(sprintf(
        "Exhaustive automatic polynomial degree search over %s degree combinations on %s",
        format(grid.size, scientific = FALSE, trim = TRUE),
        .np_degree_format_search_space(candidates)
      ))
      state$progress_state <- .np_degree_progress_begin(
        total = grid.size,
        detail = .np_degree_progress_detail(
          phase = "exhaustive",
          best_record = state$best_record,
          objective_name = objective_name
        )
      )
      .np_degree_search_exhaustive(
        state = state,
        candidates = candidates,
        phase = "exhaustive",
        objective_name = objective_name
      )
    } else {
      restart_total <- length(restart_starts) + 1L
      .np_progress_note(sprintf(
        "Coordinate automatic polynomial degree search over %s (max %s search evaluations)",
        .np_degree_format_search_space(candidates),
        format(
          .np_degree_coordinate_visit_cap(
            candidates = candidates,
            max_cycles = max_cycles,
            restart_total = restart_total
          ),
          scientific = FALSE,
          trim = TRUE
        )
      ))
      state$progress_state <- .np_degree_progress_begin(
        detail = .np_degree_progress_detail(
          phase = "coordinate",
          best_record = state$best_record,
          objective_name = objective_name
        )
      )
      .np_degree_search_coordinate(
        state = state,
        candidates = candidates,
        start_degree = start_degree,
        restart_starts = restart_starts,
        max_cycles = max_cycles,
        direction = direction,
        objective_name = objective_name
      )
      if (!isTRUE(state$interrupted) && isTRUE(verify)) {
        state$progress_state <- .np_degree_progress_end(
          state = state$progress_state,
          detail = .np_degree_progress_detail(
            phase = "coordinate",
            best_record = state$best_record,
            objective_name = objective_name
          )
        )
        .np_progress_note(sprintf(
          "Exhaustively certifying automatic polynomial degree search over %s degree combinations (re-optimizing bandwidths)",
          format(grid.size, scientific = FALSE, trim = TRUE)
        ))
        state$progress_state <- .np_degree_progress_begin(
          total = grid.size,
          detail = .np_degree_progress_detail(
            phase = "verify",
            best_record = state$best_record,
            objective_name = objective_name
          )
        )
        .np_degree_search_exhaustive(
          state = state,
          candidates = candidates,
          phase = "verify",
          objective_name = objective_name
        )
      }
    }
  }, interrupt = function(e) {
    state$interrupted <- TRUE
    NULL
  })

  state$progress_state <- .np_degree_progress_end(
    state = state$progress_state,
    detail = .np_degree_progress_best_detail(
      best_record = state$best_record,
      objective_name = objective_name
    ),
    interrupted = state$interrupted
  )

  if (is.null(state$best_payload))
    stop("automatic degree search failed to obtain any admissible fitted model")

  list(
    method = method,
    verify = isTRUE(verify),
    completed = !isTRUE(state$interrupted),
    certified = !isTRUE(state$interrupted) && (identical(method, "exhaustive") || isTRUE(verify)),
    interrupted = isTRUE(state$interrupted),
    baseline = state$baseline_record,
    best = state$best_record,
    best_payload = state$best_payload,
    n.unique = state$eval_id,
    n.visits = state$visit_id,
    n.cached = state$cached_visits,
    grid.size = grid.size,
    restart.starts = restart_starts,
    trace = .np_degree_trace_to_frame(state$trace_records, objective_name = objective_name)
  )
}

.np_nomad_require_crs <- function(version_fn = utils::packageVersion,
                                  load_namespace = loadNamespace,
                                  minimum_version = "0.15-41") {
  minimum_version_label <- as.character(minimum_version)[1L]
  minimum_version <- package_version(minimum_version_label)

  current_version <- tryCatch(
    version_fn("crs"),
    error = function(e) NULL
  )

  if (is.null(current_version)) {
    stop(
      sprintf(
        paste(
          "automatic degree search with search.engine='nomad' requires the",
          "suggested package 'crs' (>= %s) to provide the NOMAD backend;",
          "install.packages('crs')"
        ),
        minimum_version_label
      ),
      call. = FALSE
    )
  }

  current_version <- package_version(as.character(current_version))

  if (current_version < minimum_version) {
    stop(
      sprintf(
        "automatic degree search with search.engine='nomad' requires 'crs' (>= %s); installed version is %s",
        minimum_version_label,
        as.character(current_version)
      ),
      call. = FALSE
    )
  }

  load_result <- tryCatch(
    {
      suppressPackageStartupMessages(load_namespace("crs"))
      TRUE
    },
    error = function(e) e
  )

  if (inherits(load_result, "error")) {
    stop(
      sprintf(
        "automatic degree search with search.engine='nomad' requires 'crs' (>= %s); failed to load installed version %s: %s",
        minimum_version_label,
        as.character(current_version),
        conditionMessage(load_result)
      ),
      call. = FALSE
    )
  }

  invisible(TRUE)
}

.np_degree_extract_random_seed <- function(dots, default = 42L) {
  if (is.null(dots) || length(dots) == 0L)
    return(as.integer(default)[1L])

  if (is.null(names(dots)) || !"random.seed" %in% names(dots))
    return(as.integer(default)[1L])

  npValidateNonNegativeInteger(dots[["random.seed"]], "random.seed")
}

.np_degree_reject_unknown_dots <- function(dots,
                                           where,
                                           allowed = c("random.seed")) {
  if (is.null(dots) || length(dots) == 0L)
    return(invisible(TRUE))

  dot.names <- names(dots)
  if (is.null(dot.names))
    dot.names <- rep("", length(dots))

  bad <- dot.names == "" | !(dot.names %in% allowed)
  if (any(bad))
    .np_reject_unused_dots(dots[bad], where)

  invisible(TRUE)
}

.np_nomad_coerce_start_value <- function(x, type, lb, ub) {
  if (identical(as.integer(type), 3L)) {
    x <- if (is.finite(x) && x >= 0.5) 1 else 0
  } else if (identical(as.integer(type), 1L) || identical(as.integer(type), 2L)) {
    x <- round(x)
  }

  if (is.finite(lb) && x < lb)
    x <- lb
  if (is.finite(ub) && x > ub)
    x <- ub

  as.numeric(x)
}

.np_nomad_complete_start_point <- function(point,
                                           lower,
                                           upper,
                                           ncont = 0L) {
  n <- length(lower)
  out <- rep(NA_real_, n)

  if (!is.null(point)) {
    point <- as.numeric(point)
    if (length(point) == n)
      out <- point
  }

  ncont <- max(0L, min(as.integer(ncont), n))
  if (ncont > 0L) {
    cont.default <- pmax(lower[seq_len(ncont)], pmin(upper[seq_len(ncont)], 1.5))
    bad <- !is.finite(out[seq_len(ncont)]) | out[seq_len(ncont)] <= 0
    out[seq_len(ncont)][bad] <- cont.default[bad]
  }

  if (n > ncont) {
    cat.idx <- (ncont + 1L):n
    cat.default <- pmax(lower[cat.idx], pmin(upper[cat.idx], 0.5 * upper[cat.idx]))
    bad <- !is.finite(out[cat.idx]) | out[cat.idx] < lower[cat.idx] | out[cat.idx] > upper[cat.idx]
    out[cat.idx][bad] <- cat.default[bad]
  }

  pmax(lower, pmin(upper, out))
}

.np_lp_nomad_dim_budget <- function(nobs) {
  nobs <- as.integer(nobs[1L])
  if (!is.finite(nobs) || is.na(nobs) || nobs <= 1L)
    return(1L)
  max(1L, as.integer(floor(0.25 * (nobs - 1L))))
}

.np_lp_nomad_proposal_upper <- function(lower, upper) {
  q <- length(lower)
  if (!q)
    return(integer(0))

  lower <- as.integer(lower)
  upper <- as.integer(upper)
  as.integer(pmin(
    upper,
    pmax(lower + 1L, ceiling(upper / max(1L, q)))
  ))
}

.np_lp_nomad_reduce_degree_start <- function(degree,
                                             lower,
                                             basis = c("glp", "additive", "tensor"),
                                             dim_budget = Inf) {
  basis <- match.arg(basis)
  out <- as.integer(degree)
  lower <- as.integer(lower)

  if (!is.finite(dim_budget))
    return(out)

  repeat {
    current.dim <- tryCatch(
      dim_basis(basis = basis, degree = out),
      error = function(e) Inf
    )
    if (is.finite(current.dim) && current.dim <= dim_budget)
      break

    idx <- which(out > lower)
    if (!length(idx))
      break

    drop.idx <- idx[which.max(out[idx])][1L]
    out[drop.idx] <- out[drop.idx] - 1L
  }

  out
}

.np_lp_nomad_build_degree_starts <- function(initial,
                                             lower,
                                             upper,
                                             basis = c("glp", "additive", "tensor"),
                                             nobs,
                                             nmulti = 1L,
                                             random.seed = 42L,
                                             max.tries = 256L,
                                             user_supplied = FALSE) {
  basis <- match.arg(basis)
  lower <- as.integer(lower)
  upper <- as.integer(upper)
  q <- length(lower)
  nstart <- npValidateNmulti(nmulti)

  if (!q)
    return(matrix(integer(0), nrow = nstart, ncol = 0L))

  starts <- matrix(0L, nrow = nstart, ncol = q)
  dim_budget <- .np_lp_nomad_dim_budget(nobs)
  proposal.upper <- .np_lp_nomad_proposal_upper(lower, upper)

  initial <- as.integer(pmax(lower, pmin(upper, initial)))
  if (!isTRUE(user_supplied))
    initial <- .np_lp_nomad_reduce_degree_start(
      degree = initial,
      lower = lower,
      basis = basis,
      dim_budget = dim_budget
    )
  starts[1L, ] <- initial

  if (nstart <= 1L)
    return(starts)

  seed.state <- .np_seed_enter(random.seed)
  on.exit(.np_seed_exit(seed.state, remove_if_absent = TRUE), add = TRUE)

  max.tries <- max(1L, as.integer(max.tries[1L]))
  for (j in 2:nstart) {
    accepted <- FALSE
    fallback <- starts[1L, ]
    for (k in seq_len(max.tries)) {
      candidate <- vapply(
        seq_len(q),
        function(i) sample.int(proposal.upper[i] - lower[i] + 1L, 1L) + lower[i] - 1L,
        integer(1L)
      )
      fallback <- candidate
      candidate.dim <- tryCatch(
        dim_basis(basis = basis, degree = candidate),
        error = function(e) Inf
      )
      if (is.finite(candidate.dim) && candidate.dim <= dim_budget) {
        starts[j, ] <- as.integer(candidate)
        accepted <- TRUE
        break
      }
    }

    if (!accepted) {
      starts[j, ] <- .np_lp_nomad_reduce_degree_start(
        degree = fallback,
        lower = lower,
        basis = basis,
        dim_budget = dim_budget
      )
    }
  }

  starts
}

.np_nomad_build_starts <- function(x0,
                                   bbin,
                                   lb,
                                   ub,
                                   nmulti = 1L,
                                   random.seed = 42L,
                                   degree_spec = NULL) {
  n <- length(lb)
  if (length(ub) != n || length(bbin) != n)
    stop("NOMAD start construction requires matching lengths for x0/bbin/lb/ub")

  nstart <- npValidateNmulti(nmulti)
  starts <- matrix(0, nrow = nstart, ncol = n)

  seed.state <- .np_seed_enter(random.seed)
  on.exit(.np_seed_exit(seed.state, remove_if_absent = TRUE), add = TRUE)

  for (j in seq_len(nstart)) {
    for (i in seq_len(n)) {
      lo <- if (is.finite(lb[i])) lb[i] else -1
      hi <- if (is.finite(ub[i])) ub[i] else 1
      if (hi < lo) {
        tmp <- lo
        lo <- hi
        hi <- tmp
      }
      starts[j, i] <- .np_nomad_coerce_start_value(
        runif(1L, min = lo, max = hi),
        type = bbin[i],
        lb = lb[i],
        ub = ub[i]
      )
    }
  }

  if (!is.null(degree_spec)) {
    q <- length(degree_spec$lower)
    if (q > 0L) {
      degree.idx <- (n - q + 1L):n
      degree.starts <- .np_lp_nomad_build_degree_starts(
        initial = if (is.null(degree_spec$initial)) rep.int(1L, q) else degree_spec$initial,
        lower = degree_spec$lower,
        upper = degree_spec$upper,
        basis = if (is.null(degree_spec$basis)) "glp" else degree_spec$basis,
        nobs = degree_spec$nobs,
        nmulti = nstart,
        random.seed = random.seed,
        user_supplied = isTRUE(degree_spec$user_supplied)
      )
      starts[, degree.idx] <- degree.starts
    }
  }

  if (is.null(x0))
    return(starts)

  x0 <- as.numeric(x0)
  if (length(x0) == n) {
    starts[1L, ] <- vapply(
      seq_len(n),
      function(i) .np_nomad_coerce_start_value(x0[i], type = bbin[i], lb = lb[i], ub = ub[i]),
      numeric(1L)
    )
    if (!is.null(degree_spec)) {
      q <- length(degree_spec$lower)
      if (q > 0L) {
        degree.idx <- (n - q + 1L):n
        starts[1L, degree.idx] <- .np_lp_nomad_build_degree_starts(
          initial = if (is.null(degree_spec$initial)) rep.int(1L, q) else degree_spec$initial,
          lower = degree_spec$lower,
          upper = degree_spec$upper,
          basis = if (is.null(degree_spec$basis)) "glp" else degree_spec$basis,
          nobs = degree_spec$nobs,
          nmulti = 1L,
          random.seed = random.seed,
          user_supplied = isTRUE(degree_spec$user_supplied)
        )[1L, ]
      }
    }
    return(starts)
  }

  if (length(x0) >= n * nstart) {
    for (j in seq_len(nstart)) {
      starts[j, ] <- vapply(
        seq_len(n),
        function(i) .np_nomad_coerce_start_value(
          x0[j + (i - 1L) * nstart],
          type = bbin[i],
          lb = lb[i],
          ub = ub[i]
        ),
        numeric(1L)
      )
    }
  }

  starts
}

.np_nomad_progress_detail <- function(current_degree,
                                      best_record,
                                      iteration = NULL,
                                      cumulative_iteration = NULL,
                                      restart_index = NULL,
                                      nmulti = 1L,
                                      restart_durations = numeric(),
                                      elapsed = NULL) {
  fields <- .np_degree_progress_context_fields()
  nmulti <- suppressWarnings(as.integer(nmulti)[1L])
  restart_index <- suppressWarnings(as.integer(restart_index)[1L])
  if (!is.na(nmulti) && nmulti > 1L) {
    if (!is.na(restart_index) && restart_index >= 1L) {
      fields <- c(fields, sprintf("multistart %s/%s", format(restart_index), format(nmulti)))
    } else {
      fields <- c(fields, sprintf("multistarts %s", format(nmulti)))
    }
  }

  cumulative_iteration <- suppressWarnings(as.integer(cumulative_iteration)[1L])
  if (!is.null(iteration) && is.finite(iteration) && iteration >= 1L) {
    if (!is.na(cumulative_iteration) &&
        cumulative_iteration >= 1L &&
        cumulative_iteration != iteration) {
      fields <- c(
        fields,
        sprintf("iteration %s (%s)", format(iteration), format(cumulative_iteration))
      )
    } else {
      fields <- c(fields, sprintf("iteration %s", format(iteration)))
    }
  }

  if (is.finite(elapsed) && !is.na(elapsed) && elapsed >= 0)
    fields <- c(fields, sprintf("elapsed %ss", .np_progress_fmt_num(elapsed)))

  if (!is.null(current_degree))
    fields <- c(fields, sprintf("deg %s", .np_degree_format_degree(current_degree)))

  fields <- c(fields, .np_degree_progress_best_degree_detail(best_record = best_record))

  paste(fields, collapse = ", ")
}

.np_nomad_progress_start_detail <- function(baseline_degree,
                                            restart_index = 1L,
                                            nmulti,
                                            best_record,
                                            restart_durations = numeric(),
                                            elapsed = NULL) {
  .np_nomad_progress_detail(
    current_degree = baseline_degree,
    best_record = best_record,
    restart_index = restart_index,
    nmulti = nmulti,
    restart_durations = restart_durations,
    elapsed = elapsed
  )
}

.np_nomad_progress_fields <- function(state,
                                      done = NULL,
                                      detail = NULL,
                                      now = .np_progress_now()) {
  .np_nomad_progress_detail(
    current_degree = state$nomad_current_degree,
    best_record = state$nomad_best_record,
    iteration = done,
    cumulative_iteration = state$nomad_eval_id,
    restart_index = state$nomad_restart_index,
    nmulti = state$nomad_nmulti,
    restart_durations = state$nomad_restart_durations,
    elapsed = max(0, now - state$started)
  )
}

.np_nomad_progress_begin <- function(nmulti,
                                     baseline_degree,
                                     best_record) {
  state <- .np_progress_begin(
    label = .np_degree_progress_label(),
    domain = "general",
    surface = "bandwidth"
  )

  state$unknown_total_fields <- .np_nomad_progress_fields
  state$nomad_nmulti <- npValidateNmulti(nmulti)
  state$nomad_restart_index <- 1L
  state$nomad_restart_durations <- numeric()
  state$nomad_current_degree <- as.integer(baseline_degree)
  state$nomad_best_record <- best_record
  state$nomad_eval_id <- 0L
  state$start_note <- if (state$nomad_nmulti > 1L) {
    sprintf(
      "%s %s (multistart 1/%s)",
      state$pkg_prefix,
      state$label,
      format(state$nomad_nmulti)
    )
  } else {
    sprintf("%s %s...", state$pkg_prefix, state$label)
  }

  state
}

.np_nomad_baseline_note <- function(degree) {
  invisible(NULL)
}

.np_nomad_powell_note <- function(degree) {
  invisible(NULL)
}

.np_nomad_powell_progress_label <- function() {
  "Refining bandwidth"
}

.np_nomad_powell_hotstart_nmulti <- function(strategy = c("disable_multistart",
                                                          "single_iteration")) {
  strategy <- match.arg(strategy)
  switch(strategy,
         disable_multistart = 1L,
         single_iteration = 1L)
}

.np_nomad_powell_hotstart_opt_args <- function(opt.args,
                                               strategy = c("disable_multistart",
                                                            "single_iteration"),
                                               remin = FALSE) {
  out <- opt.args
  out$nmulti <- .np_nomad_powell_hotstart_nmulti(strategy)
  out$powell.remin <- isTRUE(remin)
  out
}

.np_nomad_powell_context_label <- function(degree) {
  sprintf(
    "Refining NOMAD solution with one Powell hot start at degree %s",
    .np_degree_format_degree(degree)
  )
}

.np_nomad_powell_progress_detail <- function(current_degree,
                                             best_record,
                                             iteration = NULL,
                                             elapsed = NULL) {
  fields <- .np_degree_progress_context_fields()

  if (is.finite(elapsed) && !is.na(elapsed) && elapsed >= 0)
    fields <- c(fields, sprintf("elapsed %ss", .np_progress_fmt_num(elapsed)))

  if (!is.null(iteration)) {
    iteration <- suppressWarnings(as.integer(iteration)[1L])
    if (!is.na(iteration) && iteration >= 1L)
      fields <- c(fields, sprintf("iter %s", format(iteration)))
  }

  if (!is.null(current_degree))
    fields <- c(fields, sprintf("deg %s", .np_degree_format_degree(current_degree)))

  fields <- c(fields, .np_degree_progress_best_degree_detail(best_record = best_record))

  paste(fields, collapse = ", ")
}

.np_nomad_powell_progress_fields <- function(state,
                                             done = NULL,
                                             detail = NULL,
                                             now = .np_progress_now()) {
  .np_nomad_powell_progress_detail(
    current_degree = state$nomad_current_degree,
    best_record = state$nomad_best_record,
    iteration = done,
    elapsed = max(0, now - state$started)
  )
}

.np_nomad_progress_enter_powell <- function(state,
                                            degree,
                                            best_record) {
  if (is.null(state))
    return(state)

  state$label <- .np_nomad_powell_progress_label()
  state$unknown_total_fields <- .np_nomad_powell_progress_fields
  state$nomad_current_degree <- as.integer(degree)
  state$nomad_best_record <- best_record
  .np_progress_step_at(
    state = state,
    now = .np_progress_now(),
    done = state$last_done,
    force = TRUE
  )
}

.np_nomad_with_powell_progress <- function(degree, expr, best_record = NULL) {
  old.context <- .np_progress_runtime$bandwidth_context_label
  old.state <- .np_progress_runtime$bandwidth_state
  local.state <- NULL
  progress.enabled <- isTRUE(.np_progress_enabled(domain = "general"))

  powell.context <- if (!is.null(old.context) && nzchar(old.context)) old.context else NULL
  .np_progress_bandwidth_set_context(powell.context)
  on.exit({
    if (!is.null(local.state) && !is.null(.np_progress_runtime$bandwidth_state)) {
      .np_progress_end(.np_progress_runtime$bandwidth_state)
    }
    if (!is.null(local.state)) {
      .np_progress_runtime$bandwidth_state <- old.state
    }
    .np_progress_bandwidth_set_context(old.context)
  }, add = TRUE)

  if (is.null(old.state) && isTRUE(progress.enabled)) {
    local.state <- .np_progress_begin(
      label = .np_nomad_powell_progress_label(),
      domain = "general",
      surface = "bandwidth"
    )
    local.state$unknown_total_fields <- .np_nomad_powell_progress_fields
    local.state$nomad_current_degree <- as.integer(degree)
    local.state$nomad_best_record <- best_record
    local.state <- .np_progress_show_now(local.state)
    .np_progress_runtime$bandwidth_state <- local.state
  }

  value <- force(expr)

  if (!is.null(local.state) && is.list(value) && !is.null(value$num.feval)) {
    done <- suppressWarnings(as.integer(value$num.feval[1L]))
    if (!is.na(done) && done >= 1L && !is.null(.np_progress_runtime$bandwidth_state)) {
      .np_progress_runtime$bandwidth_state <- .np_progress_step_at(
        state = .np_progress_runtime$bandwidth_state,
        now = .np_progress_now(),
        done = done,
        force = TRUE
      )
    }
  }

  value
}

.np_nomad_search <- function(engine = c("nomad", "nomad+powell"),
                             baseline_record,
                             start_degree = NULL,
                             x0,
                             bbin,
                             lb,
                             ub,
                             eval_fun,
                             build_payload,
                             direction = c("min", "max"),
                             objective_name = "objective",
                             nmulti = 1L,
                             nomad.inner.nmulti = 0L,
                             random.seed = 42L,
                             display.nomad.progress = FALSE,
                             handoff_before_build = FALSE,
                             remin = FALSE,
                             degree_spec = NULL,
                             nomad.opts = list()) {
  engine <- match.arg(engine)
  direction <- match.arg(direction)
  .np_nomad_require_crs()

  state <- new.env(parent = emptyenv())
  state$trace_records <- list()
  state$trace_id <- 0L
  state$eval_id <- 0L
  state$visit_id <- 0L
  state$best_record <- baseline_record
  state$baseline_record <- baseline_record
  state$best_point <- NULL
  state$best_payload <- NULL
  state$interrupted <- FALSE
  state$error <- NULL
  state$progress_state <- NULL
  state$nomad.time <- NA_real_
  state$powell.time <- NA_real_
  state$restart_starts <- NULL
  state$restart_results <- NULL
  state$nomad.remin <- isTRUE(remin)
  state$nomad.remin.index <- NA_integer_
  state$nomad.remin.roundtrip <- NULL
  state$current_restart <- NA_integer_
  state$best_restart_index <- NA_integer_
  state$restart_durations <- numeric()
  state$restart_eval_id <- 0L

  nomad.nmulti <- npValidateNmulti(nmulti)
  nomad.inner.nmulti <- npValidateNonNegativeInteger(nomad.inner.nmulti, "nomad.inner.nmulti")
  nomad.opts <- if (is.null(nomad.opts)) list() else nomad.opts
  start_matrix <- .np_nomad_build_starts(
    x0 = x0,
    bbin = bbin,
    lb = lb,
    ub = ub,
    nmulti = nomad.nmulti,
    random.seed = random.seed,
    degree_spec = degree_spec
  )
  state$restart_starts <- lapply(
    seq_len(nrow(start_matrix)),
    function(i) as.numeric(start_matrix[i, ])
  )
  if (!is.null(degree_spec) && length(degree_spec$lower) > 0L) {
    q <- length(degree_spec$lower)
    degree.idx <- (ncol(start_matrix) - q + 1L):ncol(start_matrix)
    state$restart_degree_starts <- lapply(
      seq_len(nrow(start_matrix)),
      function(i) as.integer(start_matrix[i, degree.idx])
    )
    if (degree.idx[1L] > 1L) {
      state$restart_bandwidth_starts <- lapply(
        seq_len(nrow(start_matrix)),
        function(i) as.numeric(start_matrix[i, seq_len(degree.idx[1L] - 1L)])
      )
    } else {
      state$restart_bandwidth_starts <- replicate(nrow(start_matrix), numeric(0), simplify = FALSE)
    }
    state$restart_start_info <- list(
      basis = if (is.null(degree_spec$basis)) "glp" else degree_spec$basis,
      dim_budget = .np_lp_nomad_dim_budget(degree_spec$nobs),
      proposal.upper = .np_lp_nomad_proposal_upper(
        lower = degree_spec$lower,
        upper = degree_spec$upper
      ),
      user_supplied_start = isTRUE(degree_spec$user_supplied)
    )
  } else {
    state$restart_degree_starts <- NULL
    state$restart_bandwidth_starts <- NULL
    state$restart_start_info <- NULL
  }

  state$record_trace <- function(rec) {
    state$trace_id <- state$trace_id + 1L
    rec$trace_id <- state$trace_id
    state$trace_records[[length(state$trace_records) + 1L]] <- rec
    invisible(rec)
  }

  state$update_best <- function(rec, point) {
    if (!identical(rec$status, "ok") || !is.finite(rec$objective))
      return(invisible(NULL))

    incumbent <- if (is.null(state$best_record)) {
      if (identical(direction, "min")) Inf else -Inf
    } else {
      state$best_record$objective
    }

    if (.np_degree_better(rec$objective, incumbent, direction = direction)) {
      state$best_record <- rec
      state$best_point <- point
      state$best_restart_index <- state$current_restart
    }

    invisible(NULL)
  }

  wrapped_eval <- function(point) {
    point <- as.numeric(point)
    state$visit_id <- state$visit_id + 1L
    started <- proc.time()[3L]
    status <- "ok"
    msg <- NULL
    objective <- NA_real_
    degree <- integer(0)
    num.feval <- NA_real_

    result <- tryCatch(
      eval_fun(point),
      interrupt = function(e) {
        state$interrupted <- TRUE
        NULL
      },
      error = function(e) {
        status <<- "error"
        msg <<- conditionMessage(e)
        NULL
      }
    )

    if (isTRUE(state$interrupted))
      stop(structure(list(message = "interrupt"), class = c("interrupt", "condition")))

    if (!is.null(result)) {
      if (!is.list(result) || is.null(result$objective) || is.null(result$degree))
        stop("NOMAD fixed-point evaluator must return a list containing 'objective' and 'degree'")

      objective <- as.numeric(result$objective[1L])
      degree <- as.integer(result$degree)
      if (!is.null(result$num.feval))
        num.feval <- as.numeric(result$num.feval[1L])
    }

    state$eval_id <- state$eval_id + 1L
    rec <- list(
      eval_id = state$eval_id,
      degree = degree,
      objective = objective,
      status = status,
      cached = FALSE,
      message = msg,
      elapsed = proc.time()[3L] - started,
      num.feval = num.feval
    )
    if (is.null(state$baseline_record))
      state$baseline_record <- rec
    state$record_trace(rec)
    state$update_best(rec, point = point)
    state$restart_eval_id <- state$restart_eval_id + 1L
    state$progress_state$nomad_current_degree <- degree
    state$progress_state$nomad_best_record <- state$best_record
    state$progress_state$nomad_eval_id <- state$eval_id
    state$progress_state$nomad_restart_index <- state$current_restart
    state$progress_state$nomad_restart_durations <- state$restart_durations
    state$progress_state <- .np_degree_progress_step(
      state = state$progress_state,
      done = state$restart_eval_id,
      detail = NULL
    )

    if (identical(status, "ok")) {
      if (identical(direction, "min")) {
        return(objective)
      } else {
        return(-objective)
      }
    }

    if (identical(direction, "min")) Inf else .Machine$double.xmax
  }

  state$progress_state <- .np_nomad_progress_begin(
    nmulti = nomad.nmulti,
    baseline_degree = if (is.null(start_degree)) {
      if (is.null(baseline_record)) integer(0) else baseline_record$degree
    } else {
      as.integer(start_degree)
    },
    best_record = state$best_record
  )

  restart_results <- vector("list", nomad.nmulti)
  best_solution <- NULL
  nomad.elapsed <- 0
  for (i in seq_len(nomad.nmulti)) {
    state$current_restart <- as.integer(i)
    state$restart_eval_id <- 0L
    restart_degree <- if (!is.null(state$restart_degree_starts) &&
                          length(state$restart_degree_starts) >= i) {
      state$restart_degree_starts[[i]]
    } else if (is.null(start_degree)) {
      if (is.null(baseline_record)) integer(0) else baseline_record$degree
    } else {
      as.integer(start_degree)
    }

    if (i > 1L) {
      state$progress_state$nomad_current_degree <- restart_degree
      state$progress_state$nomad_best_record <- state$best_record
      state$progress_state$nomad_restart_index <- i
      state$progress_state$nomad_restart_durations <- state$restart_durations
      state$progress_state <- .np_degree_progress_step(
        state = state$progress_state,
        done = NULL,
        detail = NULL,
        force = TRUE
      )
    }

    pre_restart_best <- if (is.null(state$best_record) || is.null(state$best_record$objective)) {
      if (identical(direction, "min")) Inf else -Inf
    } else {
      state$best_record$objective
    }

    nomad.start <- proc.time()[3L]
    solution_i <- tryCatch(
      {
        solver.opts <- utils::modifyList(
          list(
            SEED = as.integer(random.seed),
            RNG_ALT_SEEDING = TRUE
          ),
          nomad.opts
        )
        crs::snomadr(
          eval.f = wrapped_eval,
          n = length(x0),
          bbin = as.integer(bbin),
          bbout = 0L,
          x0 = as.numeric(start_matrix[i, ]),
          lb = as.double(lb),
          ub = as.double(ub),
          nmulti = nomad.inner.nmulti,
          random.seed = as.integer(random.seed),
          opts = solver.opts,
          display.nomad.progress = display.nomad.progress,
          snomadr.environment = environment(wrapped_eval)
        )
      },
      interrupt = function(e) {
        state$interrupted <- TRUE
        NULL
      },
      error = function(e) {
        state$error <- conditionMessage(e)
        NULL
      }
    )
    restart.elapsed <- proc.time()[3L] - nomad.start
    nomad.elapsed <- nomad.elapsed + restart.elapsed
    state$restart_durations <- c(state$restart_durations, restart.elapsed)
    state$progress_state$nomad_restart_durations <- state$restart_durations

    restart_results[[i]] <- list(
      restart = i,
      start = as.numeric(start_matrix[i, ]),
      degree.start = restart_degree,
      elapsed = restart.elapsed,
      status = if (is.null(solution_i)) {
        if (isTRUE(state$interrupted)) "interrupt" else "error"
      } else if (!is.null(solution_i$status)) {
        as.character(solution_i$status)
      } else {
        "ok"
      },
      message = if (is.null(solution_i)) state$error else solution_i$message,
      objective = if (!is.null(solution_i) && !is.null(solution_i$objective)) {
        raw.objective <- as.numeric(solution_i$objective[1L])
        if (identical(direction, "min")) raw.objective else -raw.objective
      } else {
        NA_real_
      },
      bbe = if (!is.null(solution_i) && !is.null(solution_i$bbe)) {
        as.numeric(solution_i$bbe[1L])
      } else {
        NA_real_
      },
      iterations = if (!is.null(solution_i) && !is.null(solution_i$iterations)) {
        as.numeric(solution_i$iterations[1L])
      } else {
        NA_real_
      },
      solution = if (!is.null(solution_i) && !is.null(solution_i$solution)) {
        as.numeric(solution_i$solution)
      } else {
        NULL
      }
    )

    if (isTRUE(state$interrupted))
      break

    post_restart_best <- if (is.null(state$best_record) || is.null(state$best_record$objective)) {
      if (identical(direction, "min")) Inf else -Inf
    } else {
      state$best_record$objective
    }
    if (!is.null(solution_i) &&
        .np_degree_better(post_restart_best, pre_restart_best, direction = direction)) {
      best_solution <- solution_i
    }
  }

  if (isTRUE(state$nomad.remin) && !isTRUE(state$interrupted) &&
      !is.null(state$best_point) && length(state$best_point) == length(x0)) {
    remin.index <- length(restart_results) + 1L
    remin.start <- as.numeric(state$best_point)
    remin.degree <- if (!is.null(state$best_record$degree)) {
      as.integer(state$best_record$degree)
    } else {
      integer(0)
    }
    state$nomad.remin.roundtrip <- list(
      objective = as.numeric(state$best_record$objective[1L]),
      degree = remin.degree,
      note = "accepted best point reused; generic eval_fun is not side-effect-free"
    )
    state$current_restart <- as.integer(remin.index)
    state$restart_eval_id <- 0L
    state$nomad.remin.index <- as.integer(remin.index)
    state$restart_starts[[remin.index]] <- remin.start
    if (!is.null(state$restart_degree_starts))
      state$restart_degree_starts[[remin.index]] <- remin.degree
    if (!is.null(state$restart_bandwidth_starts)) {
      q <- length(remin.degree)
      state$restart_bandwidth_starts[[remin.index]] <- if (q > 0L) {
        as.numeric(remin.start[seq_len(length(remin.start) - q)])
      } else {
        as.numeric(remin.start)
      }
    }
    if (!is.null(state$progress_state)) {
      state$progress_state$nomad_current_degree <- remin.degree
      state$progress_state$nomad_best_record <- state$best_record
      state$progress_state$nomad_restart_index <- remin.index
      state$progress_state$nomad_restart_durations <- state$restart_durations
      state$progress_state <- .np_degree_progress_step(
        state = state$progress_state,
        done = NULL,
        detail = NULL,
        force = TRUE
      )
    }
    pre_restart_best <- state$best_record$objective
    nomad.start <- proc.time()[3L]
    solution_i <- tryCatch(
      {
        solver.opts <- utils::modifyList(
          list(
            SEED = as.integer(random.seed),
            RNG_ALT_SEEDING = TRUE
          ),
          nomad.opts
        )
        crs::snomadr(
          eval.f = wrapped_eval,
          n = length(x0),
          bbin = as.integer(bbin),
          bbout = 0L,
          x0 = remin.start,
          lb = as.double(lb),
          ub = as.double(ub),
          nmulti = nomad.inner.nmulti,
          random.seed = as.integer(random.seed),
          opts = solver.opts,
          display.nomad.progress = display.nomad.progress,
          snomadr.environment = environment(wrapped_eval)
        )
      },
      interrupt = function(e) {
        state$interrupted <- TRUE
        NULL
      },
      error = function(e) {
        state$error <- conditionMessage(e)
        NULL
      }
    )
    restart.elapsed <- proc.time()[3L] - nomad.start
    nomad.elapsed <- nomad.elapsed + restart.elapsed
    state$restart_durations <- c(state$restart_durations, restart.elapsed)
    state$progress_state$nomad_restart_durations <- state$restart_durations
    restart_results[[remin.index]] <- list(
      restart = remin.index,
      remin = TRUE,
      start = remin.start,
      degree.start = remin.degree,
      elapsed = restart.elapsed,
      status = if (is.null(solution_i)) {
        if (isTRUE(state$interrupted)) "interrupt" else "error"
      } else if (!is.null(solution_i$status)) {
        as.character(solution_i$status)
      } else {
        "ok"
      },
      message = if (is.null(solution_i)) state$error else solution_i$message,
      objective = if (!is.null(solution_i) && !is.null(solution_i$objective)) {
        raw.objective <- as.numeric(solution_i$objective[1L])
        if (identical(direction, "min")) raw.objective else -raw.objective
      } else {
        NA_real_
      },
      bbe = if (!is.null(solution_i) && !is.null(solution_i$bbe)) {
        as.numeric(solution_i$bbe[1L])
      } else {
        NA_real_
      },
      iterations = if (!is.null(solution_i) && !is.null(solution_i$iterations)) {
        as.numeric(solution_i$iterations[1L])
      } else {
        NA_real_
      },
      solution = if (!is.null(solution_i) && !is.null(solution_i$solution)) {
        as.numeric(solution_i$solution)
      } else {
        NULL
      }
    )
    post_restart_best <- if (is.null(state$best_record) || is.null(state$best_record$objective)) {
      if (identical(direction, "min")) Inf else -Inf
    } else {
      state$best_record$objective
    }
    if (!is.null(solution_i) &&
        .np_degree_better(post_restart_best, pre_restart_best, direction = direction)) {
      best_solution <- solution_i
    }
  }
  state$nomad.time <- nomad.elapsed
  state$restart_results <- restart_results

  if (!is.null(best_solution) &&
      is.null(state$best_point) &&
      length(best_solution$solution) == length(x0)) {
    state$best_point <- as.numeric(best_solution$solution)
  }

  if (is.null(state$best_point))
    stop(if (is.null(state$error)) {
      "automatic degree search failed to obtain any admissible fitted model"
    } else {
      state$error
    })

  if (identical(engine, "nomad+powell") && !is.null(state$progress_state)) {
    state$progress_state <- .np_nomad_progress_enter_powell(
      state = state$progress_state,
      degree = state$best_record$degree,
      best_record = state$best_record
    )
  }

  if (isTRUE(handoff_before_build) && !is.null(state$progress_state)) {
    state$progress_state$nomad_current_degree <- state$best_record$degree
    state$progress_state$nomad_best_record <- state$best_record
    state$progress_state <- .np_degree_progress_end(
      state = state$progress_state,
      detail = NULL,
      interrupted = state$interrupted
    )
    state$progress_state <- NULL
  }

  payload_result <- tryCatch(
    build_payload(
      point = state$best_point,
      best_record = state$best_record,
      solution = best_solution,
      interrupted = state$interrupted
    ),
    error = function(e) {
      if (!is.null(state$progress_state)) {
        state$progress_state <- .np_progress_abort(
          state = state$progress_state,
          detail = conditionMessage(e)
        )
      }
      stop(e)
    }
  )
  if (is.list(payload_result) && !is.null(payload_result$payload)) {
    state$best_payload <- payload_result$payload
    if (!is.null(payload_result$powell.time))
      state$powell.time <- as.numeric(payload_result$powell.time[1L])
    if (!is.null(payload_result$objective) &&
        .np_degree_better(payload_result$objective, state$best_record$objective, direction = direction)) {
      state$best_record$objective <- as.numeric(payload_result$objective[1L])
    }
  } else {
    state$best_payload <- payload_result
  }

  if (!is.null(state$progress_state)) {
    state$progress_state$nomad_current_degree <- state$best_record$degree
    state$progress_state$nomad_best_record <- state$best_record
    state$progress_state <- .np_degree_progress_end(
      state = state$progress_state,
      detail = NULL,
      interrupted = state$interrupted
    )
  }

  list(
    method = engine,
    direction = direction,
    verify = FALSE,
    completed = !isTRUE(state$interrupted),
    certified = FALSE,
    interrupted = isTRUE(state$interrupted),
    baseline = state$baseline_record,
    best = state$best_record,
    best_payload = state$best_payload,
    best_point = state$best_point,
    n.unique = state$eval_id,
    n.visits = state$visit_id,
    n.cached = 0L,
    nomad.time = state$nomad.time,
    powell.time = state$powell.time,
    optim.time = sum(c(state$nomad.time, state$powell.time), na.rm = TRUE),
    grid.size = NA_integer_,
    best.restart = state$best_restart_index,
    nomad.remin = state$nomad.remin,
    nomad.remin.index = state$nomad.remin.index,
    nomad.remin.roundtrip = state$nomad.remin.roundtrip,
    restart.starts = state$restart_starts,
    restart.degree.starts = state$restart_degree_starts,
    restart.bandwidth.starts = state$restart_bandwidth_starts,
    restart.start.info = state$restart_start_info,
    restart.results = state$restart_results,
    trace = .np_degree_trace_to_frame(state$trace_records, objective_name = objective_name)
  )
}

npplregbw <-
  function(...){
    mc <- match.call(expand.dots = FALSE)
    npRejectRenamedScaleFactorSearchArgs(names(mc$...), where = "npplregbw")
    target <- .np_bw_dispatch_target(dots = mc$...,
                                     data_arg_names = c("xdat", "ydat", "zdat"),
                                     eval_env = parent.frame())
    UseMethod("npplregbw", target)
  }

npplregbw.formula <-
  function(formula, data, subset, na.action, call, ...){
    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(mf), nomatch = 0)
    mf <- mf[c(1,m)]

    formula.call <- .np_bw_formula_from_call(call_obj = call, eval_env = parent.frame())
    if (!is.null(formula.call))
      mf[[2]] <- formula.call

    mf.xf <- mf
    
    mf[[1]] <- as.name("model.frame")
    mf.xf[[1]] <- as.name("model.frame")
    
    ## mangle formula ...
    formula.obj <- .np_bw_resolve_formula(formula_obj = formula,
                                        formula_call = formula.call,
                                        eval_env = parent.frame())
    chromoly <- explodePipe(formula.obj, env = environment(formula))

    if (length(chromoly) != 3) ## stop if malformed formula
      stop("invoked with improper formula, please see npplregbw documentation for proper use")

    ## make formula evaluable, then eval
    bronze <- lapply(chromoly, paste, collapse = " + ")

    mf.xf[["formula"]] <- as.formula(paste(" ~ ", bronze[[2]]),
                                    env = environment(formula))

    mf[["formula"]] <- as.formula(paste(bronze[[1]]," ~ ", bronze[[3]]),
                                  env = environment(formula))

    formula.all <- if(missing(data)) {
        terms(as.formula(paste(" ~ ",bronze[[1]]," + ",bronze[[2]], " + ",bronze[[3]]),
                                  env = environment(formula)))
    } else {
        terms(as.formula(paste(" ~ ",bronze[[1]]," + ",bronze[[2]], " + ",bronze[[3]]),
                                  env = environment(formula)), data = data)
    }

    orig.ts <- if (missing(data))
      .np_terms_ts_mask(terms_obj = formula.all,
                        data = environment(formula.all),
                        eval_env = environment(formula.all))
    else .np_terms_ts_mask(terms_obj = formula.all,
                           data = data,
                           eval_env = environment(formula.all))

    arguments.mfx <- chromoly[[2]]
    arguments.mf <- c(chromoly[[1]],chromoly[[3]])

    mf[["formula"]] <- terms(mf[["formula"]])
    mf.xf[["formula"]] <- terms(mf.xf[["formula"]])
    
    if(all(orig.ts)){
      arguments <- (as.list(attr(formula.all, "variables"))[-1])
      attr(mf[["formula"]], "predvars") <- bquote(.(as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments)))))[,.(match(arguments.mf,arguments)),drop = FALSE])
      attr(mf.xf[["formula"]], "predvars") <- bquote(.(as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments)))))[,.(match(arguments.mfx,arguments)),drop = FALSE])
    }else if(any(orig.ts)){
      arguments <- (as.list(attr(formula.all, "variables"))[-1])
      arguments.normal <- arguments[which(!orig.ts)]
      arguments.timeseries <- arguments[which(orig.ts)]

      ix <- sort(c(which(orig.ts),which(!orig.ts)),index.return = TRUE)$ix
      attr(mf[["formula"]], "predvars") <- bquote((.(as.call(c(quote(cbind),as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments.timeseries)))),arguments.normal,check.rows = TRUE)))[,.(ix)])[,.(match(arguments.mf,arguments)),drop = FALSE])
      attr(mf.xf[["formula"]], "predvars") <- bquote((.(as.call(c(quote(cbind),as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments.timeseries)))),arguments.normal,check.rows = TRUE)))[,.(ix)])[,.(match(arguments.mfx,arguments)),drop = FALSE])
    }
    
    mf.args <- as.list(mf[-1L])
    mf.xf.args <- as.list(mf.xf[-1L])
    mf <- do.call(stats::model.frame, mf.args, envir = parent.frame())
    mf.xf <- do.call(stats::model.frame, mf.xf.args, envir = parent.frame())

    ydat <- model.response(mf)
    xdat <- mf.xf
    zdat <- mf[, chromoly[[3]], drop = FALSE]

    tbw <- do.call(npplregbw,
                   c(list(xdat = xdat, ydat = ydat, zdat = zdat), list(...)))

    ## clean up (possible) inconsistencies due to recursion ...
    tbw$call <- match.call(expand.dots = FALSE)
    environment(tbw$call) <- parent.frame()
    tbw$formula <- formula
    tbw$rows.omit <- as.vector(attr(mf,"na.action"))
    tbw$nobs.omit <- length(tbw$rows.omit)
    tbw$terms <- attr(mf,"terms")
    tbw$xterms <- attr(mf.xf,"terms")
    tbw$chromoly <- chromoly

    tbw <-
      updateBwNameMetadata(nameList =
                           list(ynames =
                                attr(mf, "names")[attr(tbw$terms, "response")]),
                           bws = tbw)

    tbw
  }


npplregbw.NULL =
  function(xdat = stop("invoked without data `xdat'"),
           ydat = stop("invoked without data `ydat'"),
           zdat = stop("invoked without data `zdat'"),
           bws, ...){

    ## maintain x names and 'toFrame'
    xdat <- toFrame(xdat)

    ## maintain z names and 'toFrame'
    zdat <- toFrame(zdat)

    ## bandwidths

    bws = matrix(data = 0, nrow = 1+ncol(xdat), ncol = ncol(zdat))

    tbw <- npplregbw.default(xdat = xdat, ydat = ydat, zdat = zdat,
                             bws = bws, ...)

    ## clean up (possible) inconsistencies due to recursion ...
    mc <- match.call(expand.dots = FALSE)
    environment(mc) <- parent.frame()
    tbw$call <- mc

    tbw <-
      updateBwNameMetadata(nameList = list(ynames = .npplregbw_sanitize_yname(deparse(substitute(ydat)))),
                           bws = tbw)

    tbw
  }

npplregbw.plbandwidth =
  function(xdat = stop("invoked without data `xdat'"),
           ydat = stop("invoked without data `ydat'"),
           zdat = stop("invoked without data `zdat'"),
           bws, nmulti, ...){

    ## n observations
    ## rows are observations, columns are variables
    ## x is n x k
    ## y is n x 1
    ## z is n x p

    if (missing(nmulti)){
      nmulti <- npDefaultNmulti(dim(zdat)[2])
    }
    nmulti <- npValidateNmulti(nmulti)
    
    xdat = toFrame(xdat)
    zdat = toFrame(zdat)

    ## catch and destroy NA's
    keep.rows <- rep_len(TRUE, nrow(xdat))
    rows.omit <- na.action(na.omit(data.frame(xdat,ydat,zdat)))
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Training data has no rows without NAs")

    xdat <- xdat[keep.rows,,drop = FALSE]
    ydat <- ydat[keep.rows]
    zdat <- zdat[keep.rows,,drop = FALSE]

    ## y on z
    total.groups <- 1L + ncol(xdat)
    .np_progress_bandwidth_set_coordinator(
      total_groups = total.groups,
      local_total = nmulti
    )
    total.time <-
      system.time({
        .np_progress_bandwidth_set_coordinator_group(1L, "E[y|z]")
        bws$bw$yzbw  <- npregbw(xdat = zdat, ydat = ydat,
                                bws = bws$bw$yzbw, nmulti = nmulti, ...)
        
        ## x on z

        for (i in seq_len(ncol(xdat))) {
          .np_progress_bandwidth_set_coordinator_group(i + 1L, sprintf("E[%s|z]", names(xdat)[[i]]))
          bws$bw[[i+1]] <- npregbw(xdat=zdat, ydat=xdat[,i],
                  bws = bws$bw[[i+1]], nmulti = nmulti, ...)
        }
      })[["elapsed"]]
    num.feval <- sum(sapply(bws$bw, function(bwi) {
      if (is.null(bwi$num.feval) || identical(bwi$num.feval, NA)) 0 else bwi$num.feval
    }))
    num.feval.fast <- sum(sapply(bws$bw, function(bwi) {
      if (is.null(bwi$num.feval.fast) || identical(bwi$num.feval.fast, NA)) 0 else bwi$num.feval.fast
    }))
    fval <- {
      fv <- unlist(lapply(bws$bw, function(bwi) bwi$fval))
      if (length(fv) == 0 || all(!is.finite(fv))) NA else sum(fv[is.finite(fv)])
    }
    bws <- plbandwidth(bws = bws$bw,
                       regtype = bws$regtype,
                       basis = bws$basis,
                       degree = bws$degree,
                       bernstein.basis = bws$bernstein.basis,
                       bwmethod = bws$method,
                       bwscaling = bws$scaling,
                       bwtype = bws$type,
                       ckertype = bws$ckertype,
                       ckerorder = bws$ckerorder,
                       ckerbound = bws$ckerbound,
                       ckerlb = bws$ckerlb,
                       ckerub = bws$ckerub,
                       ukertype = bws$ukertype,
                       okertype = bws$okertype,
                       xdati = bws$xdati,
                       ydati = bws$ydati,
                       zdati = bws$zdati,
                       xnames = bws$xnames,
                       ynames = bws$ynames,
                       znames = bws$znames,
                       nobs = bws$nobs,
                       fval = fval,
                       num.feval = num.feval,
                       num.feval.fast = num.feval.fast,
                       rows.omit = rows.omit,
                       total.time = total.time)
    bws$degree.policy <- if (isTRUE(bws$child.degree.common)) "common-child-degree" else "child-specific"
    bws

  }

.npplregbw_sanitize_yname <- function(yname) {
  yname <- paste(yname, collapse = " ")
  if (!nzchar(yname) || nchar(yname) > 80L || grepl("^c\\(", yname))
    return("y")
  yname
}

.npplregbw_build_plbandwidth <- function(xdat,
                                         ydat,
                                         zdat,
                                         bws,
                                         bandwidth.compute,
                                         reg.args,
                                         outer.args,
                                         child.degree = NULL) {
  yname <- .npplregbw_sanitize_yname(deparse(substitute(ydat)))

  plband <- list()
  child.reg.args <- reg.args
  if (!is.null(child.degree))
    child.reg.args$degree <- child.degree[[1L]]
  plband$yzbw <- do.call(
    npregbw,
    c(list(xdat = zdat, ydat = ydat, bws = bws[1, ]), child.reg.args)
  )

  for (i in seq_len(dim(xdat)[2])) {
    child.reg.args <- reg.args
    if (!is.null(child.degree))
      child.reg.args$degree <- child.degree[[i + 1L]]
    plband[[i + 1L]] <- do.call(
      npregbw,
      c(list(xdat = zdat, ydat = xdat[, i], bws = bws[i + 1L, ]), child.reg.args)
    )
  }

  plbw.args <- c(
    list(
      bws = plband,
      nobs = dim(xdat)[1],
      fval = {
        fv <- unlist(lapply(plband, function(bwi) bwi$fval))
        if (length(fv) == 0 || all(!is.finite(fv))) NA else sum(fv[is.finite(fv)])
      },
      num.feval = sum(sapply(plband, function(bwi) {
        if (is.null(bwi$num.feval) || identical(bwi$num.feval, NA)) 0 else bwi$num.feval
      })),
      num.feval.fast = sum(sapply(plband, function(bwi) {
        if (is.null(bwi$num.feval.fast) || identical(bwi$num.feval.fast, NA)) 0 else bwi$num.feval.fast
      })),
      xdati = untangle(xdat),
      ydati = untangle(data.frame(ydat)),
      zdati = untangle(zdat),
      xnames = names(xdat),
      ynames = yname,
      znames = names(zdat),
      bandwidth.compute = bandwidth.compute
    ),
    outer.args
  )

  tbw <- do.call(plbandwidth, plbw.args)
  if (!is.null(child.degree)) {
    degree.keys <- vapply(child.degree, paste, collapse = ",", character(1L))
    tbw$degree.policy <- if (length(unique(degree.keys)) <= 1L) "common-child-degree" else "child-specific"
  }
  tbw
}

.npplregbw_run_fixed_degree <- function(xdat,
                                        ydat,
                                        zdat,
                                        bws,
                                        reg.args,
                                        outer.args,
                                        opt.args,
                                        child.degree = NULL) {
  tbw <- .npplregbw_build_plbandwidth(
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    bws = bws,
    bandwidth.compute = TRUE,
    reg.args = reg.args,
    outer.args = outer.args,
    child.degree = child.degree
  )

  do.call(
    npplregbw.plbandwidth,
    c(list(xdat = xdat, ydat = ydat, zdat = zdat, bws = tbw), opt.args)
  )
}

.npplregbw_child_responses <- function(xdat, ydat, yname = deparse(substitute(ydat))) {
  yname <- .npplregbw_sanitize_yname(yname)
  c(
    list(list(values = ydat, name = yname)),
    lapply(seq_len(ncol(xdat)), function(i) {
      list(values = xdat[, i], name = names(xdat)[i])
    })
  )
}

.npplregbw_child_degree_names <- function(xdat) {
  c("yzbw", names(toFrame(xdat)))
}

.npplregbw_validate_child_degree_entry <- function(value, ncon, child_name) {
  npValidateGlpDegree(
    regtype = "lp",
    degree = value,
    ncon = ncon,
    argname = sprintf("degree for npplreg child '%s'", child_name)
  )
}

.npplregbw_normalize_child_degree <- function(degree,
                                              regtype,
                                              ncon,
                                              child.names,
                                              degree.select) {
  out <- list(
    degree.for.spec = degree,
    child.degree = NULL,
    child.degree.common = TRUE
  )

  if (!identical(regtype, "lp") || is.null(degree))
    return(out)

  degree.select <- match.arg(degree.select, c("manual", "coordinate", "exhaustive"))
  nchild <- length(child.names)

  if (is.list(degree) && !is.data.frame(degree)) {
    if (!identical(degree.select, "manual"))
      stop("child-specific npplreg degree lists are fixed-degree controls; use degree.select='manual'")
    if (length(degree) != nchild)
      stop(sprintf("npplreg degree list must have one entry per child regression (%d expected, got %d)",
                   nchild, length(degree)))
    if (!is.null(names(degree)) && all(nzchar(names(degree)))) {
      missing.names <- setdiff(child.names, names(degree))
      if (length(missing.names))
        stop(sprintf("npplreg degree list is missing child %s",
                     paste(sQuote(missing.names), collapse = ", ")))
      degree <- degree[child.names]
    }
    child.degree <- Map(
      function(value, child_name) .npplregbw_validate_child_degree_entry(value, ncon, child_name),
      degree,
      child.names
    )
  } else if (is.matrix(degree) || is.data.frame(degree)) {
    if (!identical(degree.select, "manual"))
      stop("child-specific npplreg degree matrices are fixed-degree controls; use degree.select='manual'")
    degree.matrix <- as.matrix(degree)
    if (nrow(degree.matrix) != nchild || ncol(degree.matrix) != ncon) {
      stop(sprintf(
        "npplreg degree matrix must have %d rows (yzbw plus x children) and %d columns (continuous z predictors)",
        nchild,
        ncon
      ))
    }
    if (!is.null(rownames(degree.matrix)) && all(nzchar(rownames(degree.matrix)))) {
      missing.names <- setdiff(child.names, rownames(degree.matrix))
      if (length(missing.names))
        stop(sprintf("npplreg degree matrix is missing child row %s",
                     paste(sQuote(missing.names), collapse = ", ")))
      degree.matrix <- degree.matrix[child.names, , drop = FALSE]
    }
    child.degree <- lapply(seq_len(nchild), function(i) {
      .npplregbw_validate_child_degree_entry(degree.matrix[i, ], ncon, child.names[[i]])
    })
    names(child.degree) <- child.names
  } else {
    common.degree <- .npplregbw_validate_child_degree_entry(degree, ncon, "all")
    child.degree <- replicate(nchild, common.degree, simplify = FALSE)
    names(child.degree) <- child.names
  }

  degree.keys <- vapply(child.degree, paste, collapse = ",", character(1L))
  out$degree.for.spec <- child.degree[[1L]]
  out$child.degree <- child.degree
  out$child.degree.common <- length(unique(degree.keys)) <= 1L
  out
}

.npplregbw_degree_search_controls <- function(regtype,
                                              regtype.named,
                                              bandwidth.compute,
                                              ncon,
                                              nobs,
                                              basis,
                                              degree.select,
                                              search.engine,
                                              degree.min,
                                              degree.max,
                                              degree.start,
                                              degree.restarts,
                                              degree.max.cycles,
                                              degree.verify,
                                              bernstein.basis,
                                              bernstein.named) {
  degree.select <- match.arg(degree.select, c("manual", "coordinate", "exhaustive"))
  if (identical(degree.select, "manual"))
    return(NULL)
  search.engine <- .np_degree_search_engine_controls(search.engine)

  regtype.requested <- if (isTRUE(regtype.named)) match.arg(regtype, c("lc", "ll", "lp")) else "lc"
  if (!identical(regtype.requested, "lp"))
    stop("automatic degree search currently requires regtype='lp'")
  if (!isTRUE(bandwidth.compute))
    stop("automatic degree search requires bandwidth.compute=TRUE")
  if (ncon < 1L)
    stop("automatic degree search requires at least one continuous smoothing predictor")

  bern.auto <- if (isTRUE(bernstein.named)) bernstein.basis else TRUE
  bern.auto <- npValidateGlpBernstein(regtype = "lp", bernstein.basis = bern.auto)

  bounds <- .np_degree_normalize_bounds(
    ncon = ncon,
    degree.min = degree.min,
    degree.max = degree.max,
    default.max = 3L
  )

  baseline.degree <- rep.int(0L, ncon)
  default.start.degree <- if (identical(search.engine, "cell")) {
    baseline.degree
  } else {
    rep.int(1L, ncon)
  }
  start.degree <- if (is.null(degree.start)) {
    pmax(bounds$lower, pmin(bounds$upper, default.start.degree))
  } else {
    start.raw <- npValidateGlpDegree(regtype = "lp", degree = degree.start, ncon = ncon, argname = "degree.start")
    out.of.range <- vapply(seq_len(ncon), function(j) !(start.raw[j] %in% bounds$candidates[[j]]), logical(1))
    if (any(out.of.range))
      stop("degree.start must lie within the searched degree candidates for every continuous smoothing predictor")
    start.raw
  }

  list(
    method = if (identical(search.engine, "cell")) degree.select else search.engine,
    engine = search.engine,
    degree.select = degree.select,
    candidates = bounds$candidates,
    lower = bounds$lower,
    upper = bounds$upper,
    baseline.degree = baseline.degree,
    start.degree = start.degree,
    start.user = !is.null(degree.start),
    basis = if (missing(basis) || is.null(basis)) "glp" else as.character(basis[1L]),
    nobs = as.integer(nobs[1L]),
    restarts = npValidateNonNegativeInteger(degree.restarts, "degree.restarts"),
    max.cycles = npValidatePositiveInteger(degree.max.cycles, "degree.max.cycles"),
    verify = npValidateScalarLogical(degree.verify, "degree.verify"),
    bernstein.basis = bern.auto
  )
}

.npplregbw_attach_degree_search <- function(bws, search_result) {
  metadata <- list(
    mode = search_result$method,
    verify = isTRUE(search_result$verify),
    completed = isTRUE(search_result$completed),
    certified = isTRUE(search_result$certified),
    interrupted = isTRUE(search_result$interrupted),
    baseline.degree = search_result$baseline$degree,
    baseline.fval = search_result$baseline$objective,
    best.degree = search_result$best$degree,
    best.fval = search_result$best$objective,
    nomad.time = search_result$nomad.time,
    powell.time = search_result$powell.time,
    optim.time = search_result$optim.time,
    n.unique = search_result$n.unique,
    n.visits = search_result$n.visits,
    n.cached = search_result$n.cached,
    grid.size = search_result$grid.size,
    best.restart = search_result$best.restart,
    restart.starts = search_result$restart.starts,
    restart.degree.starts = search_result$restart.degree.starts,
    restart.bandwidth.starts = search_result$restart.bandwidth.starts,
    restart.start.info = search_result$restart.start.info,
    restart.results = search_result$restart.results,
    trace = search_result$trace
  )

  if (!is.null(search_result$nomad.time))
    bws$nomad.time <- as.numeric(search_result$nomad.time[1L])
  if (!is.null(search_result$powell.time))
    bws$powell.time <- as.numeric(search_result$powell.time[1L])
  if (!is.null(search_result$optim.time) && is.finite(search_result$optim.time))
    bws$total.time <- as.numeric(search_result$optim.time[1L])
  bws <- .np_attach_nomad_restart_summary(bws, search_result)
  bws$degree.search <- metadata
  bws
}

.npplregbw_child_nomad_call_args <- function(zdat,
                                             response,
                                             bw.start,
                                             reg.args,
                                             opt.args,
                                             degree.search,
                                             nomad.inner.nmulti,
                                             random.seed) {
  kernel.arg.names <- c(
    "basis", "bwmethod", "bwscaling", "bwtype", "ckertype", "ckerorder",
    "ckerbound", "ckerlb", "ckerub", "ukertype", "okertype",
    "scale.factor.search.lower"
  )
  optimizer.arg.names <- c(
    "ftol", "itmax", "nmulti", "nomad.remin", "powell.remin", "small", "tol",
    "scale.factor.search.lower"
  )
  child.args <- c(
    list(
      xdat = zdat,
      ydat = response,
      bws = bw.start,
      bandwidth.compute = TRUE,
      regtype = "lp",
      nomad = TRUE,
      search.engine = degree.search$engine,
      degree.select = degree.search$degree.select,
      bernstein.basis = degree.search$bernstein.basis,
      degree.min = degree.search$lower,
      degree.max = degree.search$upper,
      degree.start = degree.search$start.degree,
      degree.restarts = degree.search$restarts,
      degree.max.cycles = degree.search$max.cycles,
      degree.verify = degree.search$verify,
      nomad.nmulti = nomad.inner.nmulti,
      random.seed = random.seed
    ),
    reg.args[intersect(names(reg.args), kernel.arg.names)],
    opt.args[intersect(names(opt.args), optimizer.arg.names)]
  )
  child.args[c("degree", "degree.engine")] <- NULL
  child.args <- child.args[!duplicated(names(child.args), fromLast = TRUE)]
  child.args
}

.npplregbw_child_specific_nomad_search <- function(xdat,
                                                   ydat,
                                                   zdat,
                                                   bws,
                                                   reg.args,
                                                   outer.args,
                                                   opt.args,
                                                   degree.search,
                                                   nomad.inner.nmulti = 0L,
                                                   random.seed = 42L,
                                                   yname = deparse(substitute(ydat))) {
  if (isTRUE(degree.search$verify))
    stop("npplreg child-specific NOMAD does not support degree.verify")

  if (is.null(degree.search$degree.select) ||
      identical(degree.search$degree.select, "manual"))
    stop("npplreg child-specific NOMAD requires automatic child degree search")

  xdat <- toFrame(xdat)
  zdat <- toFrame(zdat)
  child.responses <- .npplregbw_child_responses(xdat = xdat, ydat = ydat, yname = yname)
  child.names <- c("yzbw", names(xdat))
  child.labels <- c("E[y|z]", sprintf("E[%s|z]", names(xdat)))
  child.list <- vector("list", length(child.responses))
  names(child.list) <- child.names

  total.start <- proc.time()[3L]
  .np_progress_bandwidth_set_coordinator(
    total_groups = length(child.responses),
    local_total = 1L
  )

  for (i in seq_along(child.responses)) {
    child.context <- sprintf("%s (%d/%d)", child.labels[[i]], i, length(child.responses))
    .np_progress_bandwidth_set_coordinator_group(i, child.context)
    old.context <- .np_progress_runtime$bandwidth_context_label
    .np_progress_bandwidth_set_context(child.context)
    child.list[[i]] <- tryCatch({
      child.args <- .npplregbw_child_nomad_call_args(
        zdat = zdat,
        response = child.responses[[i]]$values,
        bw.start = bws[i, ],
        reg.args = reg.args,
        opt.args = opt.args,
        degree.search = degree.search,
        nomad.inner.nmulti = nomad.inner.nmulti,
        random.seed = random.seed
      )
      do.call(npregbw, child.args)
    }, finally = {
      .np_progress_bandwidth_set_context(old.context)
    })
    child.list[[i]]$ynames <- child.responses[[i]]$name
  }

  total.time <- proc.time()[3L] - total.start
  child.fval <- vapply(child.list, function(bwi) {
    if (is.null(bwi$fval) || !length(bwi$fval)) NA_real_ else as.numeric(bwi$fval[1L])
  }, numeric(1L))
  child.baseline.fval <- vapply(child.list, function(bwi) {
    if (is.null(bwi$degree.search$baseline.fval) ||
        !length(bwi$degree.search$baseline.fval)) NA_real_ else as.numeric(bwi$degree.search$baseline.fval[1L])
  }, numeric(1L))
  child.feval <- vapply(child.list, function(bwi) {
    if (is.null(bwi$num.feval) || identical(bwi$num.feval, NA)) 0 else as.numeric(bwi$num.feval[1L])
  }, numeric(1L))
  child.feval.fast <- vapply(child.list, function(bwi) {
    if (is.null(bwi$num.feval.fast) || identical(bwi$num.feval.fast, NA)) 0 else as.numeric(bwi$num.feval.fast[1L])
  }, numeric(1L))
  child.nomad.time <- vapply(child.list, function(bwi) {
    if (is.null(bwi$nomad.time) || !length(bwi$nomad.time)) NA_real_ else as.numeric(bwi$nomad.time[1L])
  }, numeric(1L))
  child.powell.time <- vapply(child.list, function(bwi) {
    if (is.null(bwi$powell.time) || !length(bwi$powell.time)) NA_real_ else as.numeric(bwi$powell.time[1L])
  }, numeric(1L))

  common.degree <- {
    degree.keys <- vapply(child.list, function(bwi) paste(as.integer(bwi$degree), collapse = ","), character(1L))
    length(unique(degree.keys)) <= 1L
  }
  object.degree <- as.integer(child.list[[1L]]$degree)

  plbw.args <- c(
    list(
      bws = child.list,
      nobs = dim(xdat)[1],
      fval = if (all(!is.finite(child.fval))) NA_real_ else sum(child.fval[is.finite(child.fval)]),
      num.feval = sum(child.feval),
      num.feval.fast = sum(child.feval.fast),
      xdati = untangle(xdat),
      ydati = untangle(data.frame(ydat)),
      zdati = untangle(zdat),
      xnames = names(xdat),
      ynames = child.responses[[1L]]$name,
      znames = names(zdat),
      bandwidth.compute = TRUE,
      total.time = total.time
    ),
    utils::modifyList(
      outer.args,
      list(
        regtype = "lp",
        degree = object.degree,
        bernstein.basis = degree.search$bernstein.basis
      )
    )
  )
  tbw <- do.call(plbandwidth, plbw.args)
  tbw$degree.policy <- if (isTRUE(common.degree)) "common-child-degree" else "child-specific"
  tbw$nomad.time <- if (all(!is.finite(child.nomad.time))) NA_real_ else sum(child.nomad.time[is.finite(child.nomad.time)])
  tbw$powell.time <- if (all(!is.finite(child.powell.time))) NA_real_ else sum(child.powell.time[is.finite(child.powell.time)])
  tbw$total.time <- sum(c(tbw$nomad.time, tbw$powell.time), na.rm = TRUE)
  tbw$child.fval <- child.fval
  tbw$child.baseline.fval <- child.baseline.fval
  tbw$child.num.feval <- child.feval
  tbw$child.num.feval.fast <- child.feval.fast
  tbw$child.nomad.time <- child.nomad.time
  tbw$child.powell.time <- child.powell.time

  selected.degrees <- lapply(child.list, function(bwi) as.integer(bwi$degree))
  names(selected.degrees) <- child.names
  baseline.degrees <- lapply(child.list, function(bwi) {
    if (!is.null(bwi$degree.search$baseline.degree)) as.integer(bwi$degree.search$baseline.degree) else integer(0L)
  })
  names(baseline.degrees) <- child.names

  list(
    method = degree.search$engine,
    direction = "min",
    verify = FALSE,
    completed = TRUE,
    certified = TRUE,
    interrupted = FALSE,
    baseline = list(
      degree = baseline.degrees,
      objective = if (all(!is.finite(child.baseline.fval))) NA_real_ else sum(child.baseline.fval[is.finite(child.baseline.fval)])
    ),
    best = list(
      degree = selected.degrees,
      objective = as.numeric(tbw$fval[1L]),
      child.fval = child.fval
    ),
    best_payload = tbw,
    nomad.time = tbw$nomad.time,
    powell.time = tbw$powell.time,
    optim.time = tbw$total.time,
    n.unique = NA_integer_,
    n.visits = NA_integer_,
    n.cached = NA_integer_,
    grid.size = NA_integer_,
    best.restart = NA_integer_,
    restart.starts = NULL,
    restart.degree.starts = NULL,
    restart.bandwidth.starts = NULL,
    restart.start.info = NULL,
    restart.results = NULL,
    trace = data.frame(
      child = child.names,
      label = child.labels,
      degree = vapply(selected.degrees, paste, collapse = ",", character(1L)),
      objective = child.fval,
      num.feval = child.feval,
      num.feval.fast = child.feval.fast,
      stringsAsFactors = FALSE
    )
  )
}

.npplregbw_nomad_search <- function(xdat,
                                    ydat,
                                    zdat,
                                    bws,
                                    reg.args,
                                    outer.args,
                                    opt.args,
                                    degree.search,
                                    nomad.inner.nmulti = 0L,
                                    random.seed = 42L) {
  if (isTRUE(degree.search$verify))
    stop("automatic degree search with search.engine='nomad' does not support degree.verify")

  template.reg.args <- reg.args
  template.reg.args$regtype <- "lp"
  template.reg.args$degree <- as.integer(degree.search$start.degree)
  template.reg.args$bernstein.basis <- degree.search$bernstein.basis
  template.outer.args <- outer.args
  template.outer.args$regtype <- "lp"
  template.outer.args$degree <- as.integer(degree.search$start.degree)
  template.outer.args$bernstein.basis <- degree.search$bernstein.basis

  template <- .npplregbw_build_plbandwidth(
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    bws = bws,
    bandwidth.compute = FALSE,
    reg.args = template.reg.args,
    outer.args = template.outer.args
  )

  child.templates <- template$bw
  if (!length(child.templates))
    stop("automatic degree search with search.engine='nomad' could not initialize child bandwidth templates")
  if (any(vapply(child.templates, function(tb) !identical(tb$type, "fixed"), logical(1))))
    stop("automatic degree search with search.engine='nomad' currently requires bwtype='fixed'")

  child.responses <- .npplregbw_child_responses(
    xdat = xdat,
    ydat = ydat,
    yname = template$ynames
  )
  child.setup <- lapply(child.templates, function(tb) .npregbw_nomad_bw_setup(xdat = zdat, template = tb))
  child.point.length <- vapply(seq_along(child.templates), function(i) {
    length(.npregbw_nomad_bw_to_point(child.templates[[i]]$bw, template = child.templates[[i]], setup = child.setup[[i]]))
  }, integer(1L))

  child_cont_lower <- function(i) {
    npGetScaleFactorSearchLower(
      child.templates[[i]],
      argname = "child scale.factor.search.lower"
    )
  }

  child.lower <- unlist(lapply(seq_along(child.setup), function(i) {
    c(rep.int(child_cont_lower(i), length(child.setup[[i]]$cont_idx)),
      rep.int(0, length(child.setup[[i]]$cat_idx)))
  }), use.names = FALSE)
  child.upper <- unlist(lapply(child.setup, function(setup) {
    c(rep.int(1e6, length(setup$cont_idx)), setup$cat_upper * setup$bandwidth.scale.categorical)
  }), use.names = FALSE)
  child.start <- unlist(lapply(seq_along(child.templates), function(i) {
    raw <- child.templates[[i]]$bw
    point.start <- if (all(raw == 0)) NULL else .npregbw_nomad_bw_to_point(raw, template = child.templates[[i]], setup = child.setup[[i]])
    .np_nomad_complete_start_point(
      point = point.start,
      lower = c(rep.int(child_cont_lower(i), length(child.setup[[i]]$cont_idx)),
                rep.int(0, length(child.setup[[i]]$cat_idx))),
      upper = c(rep.int(1e6, length(child.setup[[i]]$cont_idx)), child.setup[[i]]$cat_upper * child.setup[[i]]$bandwidth.scale.categorical),
      ncont = length(child.setup[[i]]$cont_idx)
    )
  }), use.names = FALSE)
  bwdim <- length(child.start)
  ndeg <- length(degree.search$start.degree)
  nomad.nmulti <- if (is.null(opt.args$nmulti)) npDefaultNmulti(dim(zdat)[2]) else npValidateNmulti(opt.args$nmulti[1L])

  x0 <- c(child.start, as.integer(degree.search$start.degree))
  lb <- c(child.lower, degree.search$lower)
  ub <- c(child.upper, degree.search$upper)
  bbin <- c(rep.int(0L, bwdim), rep.int(1L, ndeg))
  baseline.record <- NULL
  nomad.num.feval.total <- 0
  nomad.num.feval.fast.total <- 0

  .np_nomad_baseline_note(degree.search$start.degree)

  point_to_matrix <- function(point) {
    cursor <- 1L
    out <- matrix(0, nrow = length(child.templates), ncol = ncol(zdat))
    for (i in seq_along(child.templates)) {
      seg <- point[cursor:(cursor + child.point.length[i] - 1L)]
      out[i, ] <- .npregbw_nomad_point_to_bw(seg, template = child.templates[[i]], setup = child.setup[[i]])
      cursor <- cursor + child.point.length[i]
    }
    out
  }

  child_eval_payload <- function(bw.matrix, degree, penalty.multiplier) {
    total.objective <- 0
    total.feval <- 0
    total.feval.fast <- 0
    child.payloads <- vector("list", length(child.templates))

    for (i in seq_along(child.templates)) {
      child.reg.args <- reg.args
      child.reg.args$regtype <- "lp"
      child.reg.args$degree <- degree
      child.reg.args$bernstein.basis <- degree.search$bernstein.basis
      child.reg.args$bandwidth.compute <- NULL
      resp <- child.responses[[i]]
      tbw <- .npregbw_build_rbandwidth(
        xdat = zdat,
        ydat = resp$values,
        bws = bw.matrix[i, ],
        bandwidth.compute = FALSE,
        reg.args = child.reg.args,
        yname = resp$name
      )
      out <- .npregbw_eval_only(
        xdat = zdat,
        ydat = resp$values,
        bws = tbw,
        invalid.penalty = "baseline",
        penalty.multiplier = penalty.multiplier
      )
      total.objective <- total.objective + out$objective
      total.feval <- total.feval + out$num.feval
      total.feval.fast <- total.feval.fast + out$num.feval.fast
      child.payloads[[i]] <- list(
        bws = tbw,
        objective = out$objective,
        response = resp,
        num.feval = out$num.feval,
        num.feval.fast = out$num.feval.fast
      )
    }

    list(
      objective = total.objective,
      num.feval = total.feval,
      num.feval.fast = total.feval.fast,
      child.payloads = child.payloads
    )
  }

  eval_fun <- function(point) {
    point <- as.numeric(point)
    degree <- as.integer(round(point[bwdim + seq_len(ndeg)]))
    degree <- .np_degree_clip_to_grid(degree, degree.search$candidates)
    bw.matrix <- point_to_matrix(point[seq_len(bwdim)])
    out <- child_eval_payload(
      bw.matrix = bw.matrix,
      degree = degree,
      penalty.multiplier = 10
    )
    nomad.num.feval.total <<- nomad.num.feval.total + as.numeric(out$num.feval[1L])
    nomad.num.feval.fast.total <<- nomad.num.feval.fast.total + as.numeric(out$num.feval.fast[1L])
    list(
      objective = out$objective,
      degree = degree,
      num.feval = out$num.feval,
      num.feval.fast = out$num.feval.fast
    )
  }

  build_payload <- function(point, best_record, solution, interrupted) {
    point <- as.numeric(point)
    degree <- as.integer(best_record$degree)
    bw.matrix <- point_to_matrix(point[seq_len(bwdim)])
    powell.elapsed <- NA_real_

    build_direct_payload <- function() {
      child.out <- child_eval_payload(
        bw.matrix = bw.matrix,
        degree = degree,
        penalty.multiplier = 10
      )
      child.list <- vector("list", length(child.templates))
      names(child.list) <- c("yzbw", names(xdat))

      for (i in seq_along(child.templates)) {
        resp <- child.responses[[i]]
        tbw <- child.out$child.payloads[[i]]$bws
        tbw$fval <- as.numeric(child.out$child.payloads[[i]]$objective)
        tbw$ifval <- as.numeric(child.out$child.payloads[[i]]$objective)
        tbw$num.feval <- as.numeric(child.out$child.payloads[[i]]$num.feval)
        tbw$num.feval.fast <- as.numeric(child.out$child.payloads[[i]]$num.feval.fast)
        tbw$fval.history <- as.numeric(child.out$child.payloads[[i]]$objective)
        tbw$eval.history <- 1
        tbw$invalid.history <- 0
        tbw$timing <- NA_real_
        tbw$total.time <- NA_real_
        child.list[[i]] <- npregbw.rbandwidth(
          xdat = zdat,
          ydat = resp$values,
          bws = tbw,
          bandwidth.compute = FALSE
        )
        if (!is.null(child.list[[i]]$method) && length(child.list[[i]]$method))
          child.list[[i]]$pmethod <- bwmToPrint(as.character(child.list[[i]]$method[1L]))
      }

      plbw.args <- c(
        list(
          bws = child.list,
          nobs = dim(xdat)[1],
          fval = child.out$objective,
          num.feval = as.numeric(nomad.num.feval.total),
          num.feval.fast = as.numeric(nomad.num.feval.fast.total),
          xdati = untangle(xdat),
          ydati = untangle(data.frame(ydat)),
          zdati = untangle(zdat),
          xnames = names(xdat),
          ynames = template$ynames,
          znames = names(zdat),
          bandwidth.compute = TRUE
        ),
        utils::modifyList(outer.args, list(regtype = "lp", degree = degree, bernstein.basis = degree.search$bernstein.basis))
      )
      do.call(plbandwidth, plbw.args)
    }

    direct.payload <- build_direct_payload()
    direct.objective <- as.numeric(best_record$objective)

    if (identical(degree.search$engine, "nomad+powell")) {
      hot.reg.args <- reg.args
      hot.reg.args$regtype <- "lp"
      hot.reg.args$degree <- degree
      hot.reg.args$bernstein.basis <- degree.search$bernstein.basis
      hot.outer.args <- outer.args
      hot.outer.args$regtype <- "lp"
      hot.outer.args$degree <- degree
      hot.outer.args$bernstein.basis <- degree.search$bernstein.basis
      hot.opt.args <- .np_nomad_powell_hotstart_opt_args(
        opt.args,
        strategy = "disable_multistart",
        remin = isTRUE(opt.args$powell.remin)
      )
      powell.start <- proc.time()[3L]
      hot.payload <- .np_nomad_with_powell_progress(
        degree = degree,
        best_record = best_record,
        expr = .npplregbw_run_fixed_degree(
          xdat = xdat,
          ydat = ydat,
          zdat = zdat,
          bws = bw.matrix,
          reg.args = hot.reg.args,
          outer.args = hot.outer.args,
          opt.args = hot.opt.args
        )
      )
      powell.elapsed <- proc.time()[3L] - powell.start
      direct.payload$num.feval <- as.numeric(direct.payload$num.feval[1L]) + as.numeric(hot.payload$num.feval[1L])
      direct.payload$num.feval.fast <- as.numeric(direct.payload$num.feval.fast[1L]) + as.numeric(hot.payload$num.feval.fast[1L])
      hot.payload$num.feval <- direct.payload$num.feval
      hot.payload$num.feval.fast <- direct.payload$num.feval.fast
      hot.objective <- as.numeric(hot.payload$fval[1L])
      if (is.finite(hot.objective) &&
          .np_degree_better(hot.objective, direct.objective, direction = "min")) {
        return(list(payload = hot.payload, objective = hot.objective, powell.time = powell.elapsed))
      }
    }

    list(payload = direct.payload, objective = direct.objective, powell.time = powell.elapsed)
  }

  .np_nomad_search(
    engine = degree.search$engine,
    baseline_record = baseline.record,
    start_degree = degree.search$start.degree,
    x0 = x0,
    bbin = bbin,
    lb = lb,
    ub = ub,
    eval_fun = eval_fun,
    build_payload = build_payload,
    direction = "min",
    objective_name = "fval",
    nmulti = nomad.nmulti,
    nomad.inner.nmulti = nomad.inner.nmulti,
    random.seed = random.seed,
    remin = isTRUE(opt.args$nomad.remin),
    nomad.opts = list(
      DIRECTION_TYPE = "ORTHO 2N",
      QUAD_MODEL_SEARCH = "no",
      NM_SEARCH = "no",
      SPECULATIVE_SEARCH = "no",
      EVAL_OPPORTUNISTIC = "no"
    ),
    degree_spec = list(
      initial = degree.search$start.degree,
      lower = degree.search$lower,
      upper = degree.search$upper,
      basis = degree.search$basis,
      nobs = degree.search$nobs,
      user_supplied = degree.search$start.user
    )
  )
}

npplregbw.default = 
  function(xdat = stop("invoked without data `xdat'"),
           ydat = stop("invoked without data `ydat'"),
           zdat = stop("invoked without data `zdat'"),
           bandwidth.compute = TRUE,
           bws,
           degree = NULL,
           degree.select = c("manual", "coordinate", "exhaustive"),
           search.engine = c("nomad+powell", "cell", "nomad"),
           nomad = FALSE,
           nomad.nmulti = 0L,
           degree.min = NULL,
           degree.max = NULL,
           degree.start = NULL,
           degree.restarts = 0L,
           degree.max.cycles = 20L,
           degree.verify = FALSE,
           scale.factor.search.lower = NULL,
           ftol, itmax, nmulti, nomad.remin, powell.remin, small, tol,
           ...){
    bandwidth.compute <- npValidateScalarLogical(bandwidth.compute, "bandwidth.compute")

    ## maintain x names and 'toFrame'
    xdat <- toFrame(xdat)

    ## maintain z names and 'toFrame'
    zdat <- toFrame(zdat)

    dots <- list(...)
    dot.names <- names(dots)
    mc <- match.call(expand.dots = FALSE)
    mc.names <- names(mc)
    nomad.shortcut <- .np_prepare_nomad_shortcut(
      nomad = nomad,
      call_names = unique(c(mc.names, dot.names)),
      preset = list(
        regtype = "lp",
        search.engine = "nomad+powell",
        degree.select = "coordinate",
        bernstein.basis = TRUE,
        degree.min = 0L,
        degree.max = 10L,
        degree.verify = FALSE,
        bwtype = "fixed"
      ),
      values = list(
        regtype = if ("regtype" %in% dot.names) dots$regtype else NULL,
        search.engine = if ("search.engine" %in% mc.names) search.engine else NULL,
        degree.select = if ("degree.select" %in% mc.names) degree.select else NULL,
        bernstein.basis = if ("bernstein.basis" %in% dot.names) dots$bernstein.basis else NULL,
        degree.min = if ("degree.min" %in% mc.names) degree.min else NULL,
        degree.max = if ("degree.max" %in% mc.names) degree.max else NULL,
        degree.verify = if ("degree.verify" %in% mc.names) degree.verify else NULL,
        bwtype = if ("bwtype" %in% dot.names) dots$bwtype else NULL,
        degree = if (!missing(degree)) degree else NULL
      ),
      where = "npplregbw"
    )

    if (isTRUE(nomad.shortcut$enabled)) {
      if (!missing(degree))
        stop("nomad=TRUE does not support an explicit degree; remove degree or set nomad=FALSE")
      if ("regtype" %in% dot.names &&
          !identical(as.character(match.arg(nomad.shortcut$values$regtype, c("lc", "ll", "lp")))[1L], "lp"))
        stop("nomad=TRUE requires regtype='lp'")
      if ("bwtype" %in% dot.names &&
          !identical(as.character(match.arg(nomad.shortcut$values$bwtype, c("fixed", "generalized_nn", "adaptive_nn")))[1L], "fixed"))
        stop("nomad=TRUE currently requires bwtype='fixed'")
      if ("degree.select" %in% mc.names &&
          identical(as.character(match.arg(nomad.shortcut$values$degree.select, c("manual", "coordinate", "exhaustive")))[1L], "manual"))
        stop("nomad=TRUE requires automatic degree search; use degree.select='coordinate' or 'exhaustive'")
      if ("search.engine" %in% mc.names &&
          !(as.character(match.arg(nomad.shortcut$values$search.engine, c("nomad+powell", "cell", "nomad")))[1L] %in%
              c("nomad", "nomad+powell")))
        stop("nomad=TRUE requires search.engine='nomad' or 'nomad+powell'")
      if ("bernstein.basis" %in% dot.names &&
          !isTRUE(npValidateGlpBernstein(regtype = "lp",
                                        bernstein.basis = nomad.shortcut$values$bernstein.basis)))
        stop("nomad=TRUE currently requires bernstein.basis=TRUE")
      if ("degree.verify" %in% mc.names &&
          isTRUE(npValidateScalarLogical(nomad.shortcut$values$degree.verify, "degree.verify")))
        stop("nomad=TRUE currently requires degree.verify=FALSE")
    }

    regtype.arg <- if (!is.null(nomad.shortcut$values$regtype)) nomad.shortcut$values$regtype else "lc"
    basis.arg <- if ("basis" %in% dot.names) dots$basis else "glp"
    degree.arg <- if (!missing(degree)) degree else NULL
    bernstein.arg <- if (!is.null(nomad.shortcut$values$bernstein.basis)) nomad.shortcut$values$bernstein.basis else FALSE
    ncon.z <- sum(untangle(zdat)$icon)
    child.degree.setup <- .npplregbw_normalize_child_degree(
      degree = degree.arg,
      regtype = regtype.arg,
      ncon = ncon.z,
      child.names = .npplregbw_child_degree_names(xdat),
      degree.select = if (!is.null(nomad.shortcut$values$degree.select)) nomad.shortcut$values$degree.select else "manual"
    )
    degree.for.spec <- child.degree.setup$degree.for.spec

    spec.mc.names <- dot.names
    if (!missing(degree))
      spec.mc.names <- c(spec.mc.names, "degree")
    if (isTRUE(nomad.shortcut$enabled))
      spec.mc.names <- unique(c(spec.mc.names, "regtype", "bernstein.basis"))
    degree.select.value <- if (!is.null(nomad.shortcut$values$degree.select)) nomad.shortcut$values$degree.select else "manual"
    degree.setup <- npSetupGlpDegree(
      regtype = regtype.arg,
      degree = degree.for.spec,
      ncon = ncon.z,
      degree.select = degree.select.value
    )
    scale.factor.search.lower <- npResolveScaleFactorLowerBound(scale.factor.search.lower)

    spec <- npResolveCanonicalConditionalRegSpec(
      mc.names = spec.mc.names,
      regtype = regtype.arg,
      basis = basis.arg,
      degree = degree.setup,
      bernstein.basis = bernstein.arg,
      ncon = ncon.z,
      where = "npplregbw"
    )

    random.seed.value <- .np_degree_extract_random_seed(dots)
    degree.search <- .npplregbw_degree_search_controls(
      regtype = regtype.arg,
      regtype.named = isTRUE(nomad.shortcut$enabled) || ("regtype" %in% dot.names),
      bandwidth.compute = bandwidth.compute,
      ncon = ncon.z,
      nobs = NROW(zdat),
      basis = basis.arg,
      degree.select = degree.select.value,
      search.engine = if (!is.null(nomad.shortcut$values$search.engine)) nomad.shortcut$values$search.engine else "nomad+powell",
      degree.min = nomad.shortcut$values$degree.min,
      degree.max = nomad.shortcut$values$degree.max,
      degree.start = if ("degree.start" %in% mc.names) degree.start else NULL,
      degree.restarts = if ("degree.restarts" %in% mc.names) degree.restarts else 0L,
      degree.max.cycles = if ("degree.max.cycles" %in% mc.names) degree.max.cycles else 20L,
      degree.verify = if (!is.null(nomad.shortcut$values$degree.verify)) nomad.shortcut$values$degree.verify else FALSE,
      bernstein.basis = bernstein.arg,
      bernstein.named = isTRUE(nomad.shortcut$enabled) || ("bernstein.basis" %in% dot.names)
    )
    nomad.inner.named <- "nomad.nmulti" %in% mc.names
    nomad.inner.nmulti <- if (nomad.inner.named) {
      npValidateNonNegativeInteger(nomad.nmulti, "nomad.nmulti")
    } else {
      0L
    }
    if (nomad.inner.named &&
        (is.null(degree.search) || !(degree.search$engine %in% c("nomad", "nomad+powell")))) {
      stop("nomad.nmulti is only supported when regtype='lp', automatic degree search is active, and search.engine is 'nomad' or 'nomad+powell'")
    }

    reg.args <- list(
      regtype = spec$regtype.engine,
      basis = spec$basis.engine,
      degree = spec$degree.engine,
      bernstein.basis = spec$bernstein.basis.engine,
      bandwidth.compute = FALSE,
      scale.factor.search.lower = scale.factor.search.lower
    )
    if (!is.null(degree.search))
      reg.args$bernstein.basis <- degree.search$bernstein.basis
    kernel.arg.names <- intersect(
      c("bwmethod", "bwscaling", "bwtype", "ckertype", "ckerorder",
        "ckerbound", "ckerlb", "ckerub", "ukertype", "okertype"),
      dot.names
    )
    if (length(kernel.arg.names))
      reg.args[kernel.arg.names] <- dots[kernel.arg.names]
    outer.args <- reg.args[setdiff(names(reg.args), "bandwidth.compute")]
    outer.args$regtype <- spec$regtype
    outer.args$basis <- spec$basis
    outer.args$degree <- spec$degree
    outer.args$bernstein.basis <- spec$bernstein.basis
    outer.args$scale.factor.search.lower <- scale.factor.search.lower

    opt.args <- list()
    margs <- c("regtype", "basis", "degree", "bernstein.basis",
               "bwmethod", "bwscaling", "bwtype", "ckertype", "ckerorder",
               "ckerbound", "ckerlb", "ckerub", "ukertype", "okertype",
               "scale.factor.search.lower",
               "ftol", "itmax", "nmulti", "nomad.remin", "powell.remin", "small", "tol")
    m <- match(margs, mc.names, nomatch = 0)
    any.m <- any(m != 0)

    if (bandwidth.compute) {
      if (!missing(nmulti)) nmulti <- npValidateNmulti(nmulti)
      if (!missing(powell.remin)) powell.remin <- npValidateScalarLogical(powell.remin, "powell.remin")
      if (!missing(nomad.remin)) nomad.remin <- npValidateScalarLogical(nomad.remin, "nomad.remin")
      if (!missing(itmax)) itmax <- npValidatePositiveInteger(itmax, "itmax")
      if (!missing(ftol)) ftol <- npValidatePositiveFiniteNumeric(ftol, "ftol")
      if (!missing(tol)) tol <- npValidatePositiveFiniteNumeric(tol, "tol")
      if (!missing(small)) small <- npValidatePositiveFiniteNumeric(small, "small")
      if (any.m) {
        nms <- mc.names[m]
        opt.args[nms] <- mget(nms, envir = environment(), inherits = FALSE)
      }
      opt.args$scale.factor.search.lower <- scale.factor.search.lower

      if (!is.null(degree.search)) {
        if (identical(degree.search$engine, "cell")) {
          eval_fun <- function(degree.vec) {
            cell.reg.args <- reg.args
            cell.reg.args$regtype <- "lp"
            cell.reg.args$degree <- as.integer(degree.vec)
            cell.reg.args$bernstein.basis <- degree.search$bernstein.basis
            cell.outer.args <- outer.args
            cell.outer.args$regtype <- "lp"
            cell.outer.args$degree <- as.integer(degree.vec)
            cell.outer.args$bernstein.basis <- degree.search$bernstein.basis
            cell.bws <- .npplregbw_run_fixed_degree(
              xdat = xdat,
              ydat = ydat,
              zdat = zdat,
              bws = bws,
              reg.args = cell.reg.args,
              outer.args = cell.outer.args,
              opt.args = opt.args
            )
            list(
              objective = as.numeric(cell.bws$fval[1L]),
              payload = cell.bws,
              num.feval = if (!is.null(cell.bws$num.feval)) as.numeric(cell.bws$num.feval[1L]) else NA_real_
            )
          }

          search.result <- .np_degree_search(
            method = degree.search$method,
            candidates = degree.search$candidates,
            baseline_degree = degree.search$baseline.degree,
            start_degree = degree.search$start.degree,
            restarts = degree.search$restarts,
            max_cycles = degree.search$max.cycles,
            verify = degree.search$verify,
            eval_fun = eval_fun,
            direction = "min",
            trace_level = "full",
            objective_name = "fval"
          )
        } else {
          search.result <- .npplregbw_child_specific_nomad_search(
            xdat = xdat,
            ydat = ydat,
            zdat = zdat,
            bws = bws,
            reg.args = reg.args,
            outer.args = outer.args,
            opt.args = opt.args,
            degree.search = degree.search,
            nomad.inner.nmulti = nomad.inner.nmulti,
            random.seed = random.seed.value,
            yname = deparse(substitute(ydat))
          )
        }
        tbw <- .npplregbw_attach_degree_search(
          bws = search.result$best_payload,
          search_result = search.result
        )
        mc <- match.call(expand.dots = FALSE)
        environment(mc) <- parent.frame()
        tbw$call <- mc
        tbw <- .np_attach_nomad_shortcut(tbw, nomad.shortcut$metadata)
        return(tbw)
      }
    }

    tbw <- .npplregbw_build_plbandwidth(
      xdat = xdat,
      ydat = ydat,
      zdat = zdat,
      bws = bws,
      bandwidth.compute = bandwidth.compute,
      reg.args = reg.args,
      outer.args = outer.args,
      child.degree = child.degree.setup$child.degree
    )

    if (bandwidth.compute) {
      tbw <- .np_progress_select_bandwidth_enhanced(
        "Selecting partially linear regression bandwidth",
        do.call(npplregbw.plbandwidth, c(list(xdat = xdat, ydat = ydat, zdat = zdat, bws = tbw), opt.args))
      )
    }

    mc <- match.call(expand.dots = FALSE)
    environment(mc) <- parent.frame()
    tbw$call <- mc
    tbw <- .np_attach_nomad_shortcut(tbw, nomad.shortcut$metadata)

    return(tbw)
  }

progress_time_counter <- function(start = 0, by = 1.7) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

shadow_lines <- function(shadow) {
  vapply(shadow$trace, `[[`, character(1L), "line")
}

test_that("npplregbw uses coordinated generic bandwidth selection progress", {
  set.seed(42)
  n <- 35
  x <- runif(n, -1, 1)
  z <- rnorm(n)
  y <- x^2 + z + rnorm(n, sd = 0.25 * sd(x))

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npplregbw(
      xdat = data.frame(x = x),
      zdat = data.frame(z = z),
      ydat = y,
      nmulti = 2
    ),
    force_renderer = "single_line",
    now = progress_time_counter()
  )

  lines <- shadow_lines(actual)

  expect_s3_class(actual$value, "plbandwidth")
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(y~z, multistart 1/2\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(y~z, multistart 2/2, elapsed [0-9]+\\.[0-9]s, [0-9]+\\.[0-9]%, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(x1~z, multistart 1/2, iteration [0-9]+, elapsed [0-9]+\\.[0-9]s, [0-9]+\\.[0-9]%, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(x1~z, multistart 2/2, elapsed [0-9]+\\.[0-9]s, 100\\.0%, eta 0\\.0s\\)$", lines)))
})

test_that("npplreg formula entry inherits coordinated generic bandwidth progress", {
  set.seed(42)
  n <- 35
  dat <- data.frame(x = runif(n, -1, 1), z = rnorm(n))
  dat$y <- dat$x^2 + dat$z + rnorm(n, sd = 0.25 * sd(dat$x))

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npplreg(
      y ~ x | z,
      data = dat,
      nmulti = 2
    ),
    force_renderer = "single_line",
    now = progress_time_counter()
  )

  lines <- shadow_lines(actual)

  expect_s3_class(actual$value, "plregression")
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(y~z, multistart 1/2\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(x1~z, multistart 1/2, iteration [0-9]+, elapsed [0-9]+\\.[0-9]s, [0-9]+\\.[0-9]%, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bandwidth selection \\(x1~z, multistart 2/2, elapsed [0-9]+\\.[0-9]s, 100\\.0%, eta 0\\.0s\\)$", lines)))
})

test_that("npplreg basic functionality works", {
  set.seed(42)
  n <- 100
  x1 <- runif(n) # nonparametric part
  z1 <- runif(n) # parametric part
  y <- x1^2 + 2*z1 + rnorm(n, sd=0.1)
  
  # Partially linear model: y = g(x1) + z1*beta + e
  # bws needs to be a matrix. Row 1: y on x1, Row 2: z1 on x1
  bw_mat <- matrix(c(0.1, 0.1), nrow=2, ncol=1)
  bw <- npplregbw(xdat=z1, zdat=x1, ydat=y, bws=bw_mat, bandwidth.compute=FALSE)
  
  model <- npplreg(bws=bw)
  
  expect_s3_class(model, "plregression")
  expect_type(predict(model), "double")
  expect_output(summary(model))
  expect_length(coef(model), 1)
})

test_that("npplreg direct-fit objects expose canonical and compatibility bandwidth slots", {
  set.seed(7)
  n <- 80
  x1 <- runif(n)
  z1 <- runif(n)
  y <- x1^2 + 2 * z1 + rnorm(n, sd = 0.1)

  bw_mat <- matrix(c(0.1, 0.1), nrow = 2, ncol = 1)
  bw <- npplregbw(xdat = z1, zdat = x1, ydat = y, bws = bw_mat, bandwidth.compute = FALSE)
  model <- npplreg(bws = bw)

  expect_true("bws" %in% names(model))
  expect_true("bw" %in% names(model))
  expect_s3_class(model$bws, "plbandwidth")
  expect_s3_class(model$bw, "plbandwidth")
  expect_equal(model$bw$bw, model$bws$bw)
  expect_equal(model$bw$fval, model$bws$fval)
})

test_that("plregression methods remain compatible with legacy objects lacking bws", {
  set.seed(11)
  n <- 80
  x1 <- runif(n)
  z1 <- runif(n)
  y <- x1^2 + 2 * z1 + rnorm(n, sd = 0.1)

  bw_mat <- matrix(c(0.1, 0.1), nrow = 2, ncol = 1)
  bw <- npplregbw(xdat = z1, zdat = x1, ydat = y, bws = bw_mat, bandwidth.compute = FALSE)
  model <- npplreg(bws = bw)
  legacy <- model
  legacy$bws <- NULL

  expect_silent(capture.output(print(legacy)))
  expect_silent(capture.output(summary(legacy)))
  expect_type(predict(legacy), "double")
  expect_type(residuals(legacy), "double")
})

test_that("predict.plregression se.fit returns prediction standard errors", {
  set.seed(13)
  n <- 80
  x1 <- runif(n)
  z1 <- runif(n)
  y <- x1^2 + 2 * z1 + rnorm(n, sd = 0.1)

  bw_mat <- matrix(c(0.1, 0.1), nrow = 2, ncol = 1)
  bw <- npplregbw(xdat = z1, zdat = x1, ydat = y, bws = bw_mat, bandwidth.compute = FALSE)
  model <- npplreg(bws = bw)

  expect_equal(as.numeric(predict(model)), as.numeric(fitted(model)), tolerance = 0)
  pred.se <- predict(model, se.fit = TRUE)
  expect_named(pred.se, c("fit", "se.fit"))
  expect_equal(as.numeric(pred.se$fit), as.numeric(fitted(model)), tolerance = 0)
  expect_length(pred.se$se.fit, length(fitted(model)))
  expect_true(all(is.finite(pred.se$se.fit)))
  expect_true(all(pred.se$se.fit >= 0))
})

test_that("predict.plregression se.fit supports evaluation data and mixed x/z", {
  set.seed(20260506)
  n <- 72L
  z1 <- runif(n)
  z2 <- factor(sample(c("a", "b", "c"), n, replace = TRUE))
  x1 <- rnorm(n)
  x2 <- factor(sample(c("low", "high"), n, replace = TRUE))
  x2num <- as.numeric(x2 == "high")
  y <- 0.8 * x1 - 0.4 * x2num + cos(2 * pi * z1) +
    0.25 * as.numeric(z2 == "b") + rnorm(n, sd = 0.08)

  tx <- data.frame(x1 = x1, x2 = x2)
  tz <- data.frame(z1 = z1, z2 = z2)
  ex <- data.frame(
    x1 = seq(-1, 1, length.out = 11L),
    x2 = factor(rep(c("low", "high", "low"), length.out = 11L), levels = levels(x2))
  )
  ez <- data.frame(
    z1 = seq(0.05, 0.95, length.out = 11L),
    z2 = factor(rep(c("a", "b", "c"), length.out = 11L), levels = levels(z2))
  )
  bw <- npplregbw(
    xdat = tx,
    zdat = tz,
    ydat = y,
    bws = matrix(c(0.25, 0.25, 0.35, 0.35, 0.3, 0.3), nrow = 3L, ncol = 2L),
    bandwidth.compute = FALSE
  )
  model <- npplreg(bws = bw)
  pred <- predict(model, exdat = ex, ezdat = ez)
  pred.se <- predict(model, exdat = ex, ezdat = ez, se.fit = TRUE)

  expect_equal(as.numeric(pred.se$fit), as.numeric(pred), tolerance = 0)
  expect_length(pred.se$se.fit, length(pred))
  expect_true(all(is.finite(pred.se$se.fit)))
  expect_true(all(pred.se$se.fit >= 0))
})

test_that("predict.plregression se.fit follows formula newdata evaluation rows", {
  set.seed(202605061)
  n <- 65L
  dat <- data.frame(
    y = rnorm(n),
    x = rnorm(n),
    z = runif(n)
  )
  dat$y <- 0.7 * dat$x + sin(2 * pi * dat$z) + rnorm(n, sd = 0.12)
  bw <- npplregbw(
    y ~ x | z,
    data = dat,
    bws = matrix(c(0.3, 0.3), nrow = 2L),
    bandwidth.compute = FALSE
  )
  fit <- npplreg(bws = bw)
  nd <- data.frame(
    x = seq(-1, 1, length.out = 9L),
    z = seq(0.1, 0.9, length.out = 9L)
  )

  pred <- predict(fit, newdata = nd)
  pred.se <- predict(fit, newdata = nd, se.fit = TRUE)

  expect_equal(as.numeric(pred.se$fit), as.numeric(pred), tolerance = 0)
  expect_length(pred.se$se.fit, nrow(nd))
  expect_true(all(is.finite(pred.se$se.fit)))
  expect_true(all(pred.se$se.fit >= 0))
})

test_that("npplreg plot-data asymptotic and bootstrap intervals are comparable", {
  set.seed(202605062)
  n <- 80L
  x <- rnorm(n)
  z <- runif(n)
  y <- 0.5 + x + sin(2 * pi * z) + rnorm(n, sd = 0.2)
  bw <- npplregbw(
    xdat = data.frame(x = x),
    zdat = data.frame(z = z),
    ydat = y,
    bws = matrix(c(0.3, 0.3), nrow = 2L),
    bandwidth.compute = FALSE
  )
  fit <- npplreg(bws = bw)
  extract_err <- function(obj)
    unlist(lapply(obj, function(x) rowMeans(abs(x$merr[, 1:2, drop = FALSE]))),
           use.names = FALSE)

  asym <- plot(
    fit,
    output = "data",
    perspective = FALSE,
    errors = "asymptotic",
    band = "pointwise",
    neval = 8L
  )
  boot <- plot(
    fit,
    output = "data",
    perspective = FALSE,
    errors = "bootstrap",
    band = "pointwise",
    B = 39L,
    random.seed = 202605062,
    neval = 8L
  )

  asym.err <- extract_err(asym)
  boot.err <- extract_err(boot)
  ratio <- median(asym.err, na.rm = TRUE) / median(boot.err, na.rm = TRUE)

  expect_true(is.finite(ratio))
  expect_true(ratio > 0.2)
  expect_true(ratio < 10)
})

test_that("npplreg residuals stay training-length when evaluation x/z are supplied", {
  set.seed(20260323)
  n <- 24L
  x1 <- seq(0.1, 1.2, length.out = n)
  z1 <- seq(0.2, 1.4, length.out = n)
  y <- x1^2 + 2 * z1

  bw <- npplregbw(
    xdat = z1,
    zdat = x1,
    ydat = y,
    bws = matrix(c(0.35, 0.35), nrow = 2),
    bandwidth.compute = FALSE
  )
  ex <- data.frame(x1 = z1[c(2, 5, 8)])
  ez <- data.frame(z1 = x1[c(3, 9, 15)])
  ey <- y[c(3, 9, 15)]

  train.fit <- npplreg(bws = bw, residuals = TRUE)
  eval.fit.noey <- npplreg(bws = bw, residuals = TRUE, exdat = ex, ezdat = ez)
  eval.fit.ey <- npplreg(bws = bw, residuals = TRUE, exdat = ex, ezdat = ez, eydat = ey)

  expect_equal(length(eval.fit.noey$mean), nrow(ex), tolerance = 0)
  expect_equal(length(eval.fit.ey$mean), nrow(ex), tolerance = 0)
  expect_equal(length(eval.fit.noey$resid), length(y), tolerance = 0)
  expect_equal(length(eval.fit.ey$resid), length(y), tolerance = 0)
  expect_equal(as.numeric(eval.fit.noey$resid), as.numeric(train.fit$resid), tolerance = 0)
  expect_equal(as.numeric(eval.fit.ey$resid), as.numeric(train.fit$resid), tolerance = 0)
})

test_that("npplregbw exhaustive degree search matches manual profile minimum", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 28
  xdat <- data.frame(x = rnorm(n))
  zdat <- data.frame(z = sort(runif(n)))
  y <- 1 + 0.75 * xdat$x + sin(2 * pi * zdat$z) + rnorm(n, sd = 0.08)

  bw0 <- npplregbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree = 0L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  bw1 <- npplregbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree = 1L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  auto <- npplregbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_s3_class(auto, "plbandwidth")
  expect_true(isTRUE(auto$bernstein.basis))
  expect_identical(auto$degree.search$mode, "exhaustive")
  expect_true(isTRUE(auto$degree.search$completed))
  expect_true(isTRUE(auto$degree.search$certified))
  expect_lte(auto$fval, min(bw0$fval, bw1$fval) + 1e-10)
  expect_lte(auto$degree.search$best.fval, auto$degree.search$baseline.fval + 1e-10)
  expect_true(all(c("degree", "fval", "status", "cached") %in% names(auto$degree.search$trace)))

  manual <- npplregbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  expect_null(manual$degree.search)
})

test_that("npplregbw coordinate search can be exhaustively certified on a small grid", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 26
  xdat <- data.frame(x = rnorm(n))
  zdat <- data.frame(
    z1 = runif(n),
    z2 = runif(n)
  )
  y <- 1 + 0.5 * xdat$x + sin(2 * pi * zdat$z1) + zdat$z2^2 + rnorm(n, sd = 0.08)

  exhaustive <- npplregbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  coordinate <- npplregbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = TRUE,
    degree.restarts = 1L,
    degree.max.cycles = 4L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_identical(coordinate$degree.search$mode, "coordinate")
  expect_true(isTRUE(coordinate$degree.search$completed))
  expect_true(isTRUE(coordinate$degree.search$certified))
  expect_equal(as.integer(coordinate$degree), as.integer(exhaustive$degree))
  expect_equal(coordinate$fval, exhaustive$fval, tolerance = 1e-10)
})

test_that("npplregbw automatic degree search enforces pilot guardrails", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 24
  xdat <- data.frame(x = rnorm(n))
  zdat <- data.frame(z = runif(n))
  y <- 1 + xdat$x + sin(2 * pi * zdat$z) + rnorm(n, sd = 0.08)

  expect_error(
    npplregbw(
      xdat = xdat,
      zdat = zdat,
      ydat = y,
      regtype = "lc",
      degree.select = "exhaustive",
      search.engine = "cell",
      degree.min = 0L,
      degree.max = 1L,
      bwtype = "fixed",
      bwmethod = "cv.ls",
      nmulti = 1L
    ),
    "automatic degree search currently requires regtype='lp'"
  )

  expect_error(
    npplregbw(
      xdat = xdat,
      zdat = zdat,
      ydat = y,
      regtype = "lp",
      bandwidth.compute = FALSE,
      degree.select = "exhaustive",
      search.engine = "cell",
      degree.min = 0L,
      degree.max = 1L,
      bws = matrix(0.2, nrow = 2L, ncol = 1L)
    ),
    "bandwidth.compute=TRUE"
  )

  bw <- npplregbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    bernstein.basis = FALSE,
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 4L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_s3_class(bw, "plbandwidth")
  expect_false(isTRUE(bw$bernstein.basis))
  expect_lte(max(as.integer(bw$degree)), 4L)
})

test_that("npplreg forwards automatic LP degree search through npplregbw", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 24
  dat <- data.frame(
    x = rnorm(n),
    z = runif(n)
  )
  dat$y <- 1 + 0.75 * dat$x + sin(2 * pi * dat$z) + rnorm(n, sd = 0.08)

  fit <- npplreg(
    y ~ x | z,
    data = dat,
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_s3_class(fit, "plregression")
  expect_s3_class(fit$bws, "plbandwidth")
  expect_false(is.null(fit$bws$degree.search))
  expect_identical(fit$bws$degree.search$mode, "exhaustive")
})

test_that("npplregbw automatic degree search defaults to NOMAD plus Powell", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  n <- 24
  xdat <- data.frame(x = rnorm(n))
  zdat <- data.frame(z = runif(n))
  y <- 1 + 0.75 * xdat$x + sin(2 * pi * zdat$z) + rnorm(n, sd = 0.08)

  auto <- npplregbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    regtype = "lp",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 2L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_s3_class(auto, "plbandwidth")
  expect_false(is.null(auto$degree.search))
  expect_identical(auto$degree.search$mode, "nomad+powell")
  expect_equal(length(auto$degree), 1L)
  expect_true(is.finite(auto$fval))
  expect_lte(auto$degree.search$best.fval, auto$degree.search$baseline.fval + 1e-8)
  expect_lte(auto$fval, auto$degree.search$baseline.fval + 1e-8)
  expect_equal(auto$fval, auto$degree.search$best.fval, tolerance = 1e-8)
})

test_that("npplreg explicit plbandwidth route preserves NOMAD child payload names", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260322)
  n <- 40L
  x1 <- runif(n)
  x2 <- runif(n)
  y <- x1^2 + rnorm(n, sd = 0.1)

  bw <- npplregbw(
    xdat = data.frame(x1 = x1),
    zdat = data.frame(x2 = x2),
    ydat = y,
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "nomad+powell",
    degree.max = 2L,
    nmulti = 1L
  )

  expect_identical(names(bw$bw), c("yzbw", "x1"))

  fit <- npplreg(
    bws = bw,
    txdat = data.frame(x1 = x1),
    tzdat = data.frame(x2 = x2),
    tydat = y
  )

  expect_s3_class(fit, "plregression")
  expect_equal(nrow(fit$evalx), n)
})

test_that("npplreg lp degree metadata is preserved in plbandwidth summary", {
  set.seed(42)
  n <- 120
  x1 <- rnorm(n)
  x2 <- rbinom(n, 1, 0.5)
  z1 <- rbinom(n, 1, 0.5)
  z2 <- rnorm(n)
  y <- 1 + x1 + x2 + z1 + sin(z2) + rnorm(n)

  for (deg in c(0L, 2L)) {
    fit <- npplreg(
      y ~ x1 + factor(x2) | factor(z1) + z2,
      regtype = "lp",
      degree = deg
    )

    expect_identical(as.integer(fit$bw$degree), as.integer(deg))
    expect_identical(as.integer(fit$bw$bw$yzbw$degree), as.integer(deg))

    s <- capture.output(summary(fit$bw))
    reg_line <- s[grep("Regression Type:", s)][1]
    expect_true(grepl(paste0("degree = ", deg), reg_line))
  }
})

test_that("npplregbw forwards kernel-defining options to child regression bandwidths", {
  set.seed(20260306)
  n <- 40
  xdat <- data.frame(x = runif(n))
  zdat <- data.frame(z = runif(n))
  y <- 2 * xdat$x + sin(2 * pi * zdat$z) + rnorm(n, sd = 0.1)

  bw <- npplregbw(
    xdat = xdat,
    zdat = zdat,
    ydat = y,
    nmulti = 1L,
    regtype = "lp",
    basis = "glp",
    degree = 2L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    ckertype = "epanechnikov",
    ckerorder = 2L
  )

  expect_identical(bw$ckertype, "epanechnikov")
  expect_identical(bw$bw$yzbw$ckertype, "epanechnikov")
  expect_identical(bw$bw[[2L]]$ckertype, "epanechnikov")
  expect_match(bw$bw$yzbw$pckertype, "Epanechnikov")
  expect_match(bw$bw[[2L]]$pckertype, "Epanechnikov")
})

npkernelsum = 
  function(bws, eval, ksum, kw, p.ksum, p.kw = NULL,
           ntrain, trainiseval = FALSE){

    if (missing(bws) || missing(eval) || missing(ksum) || missing(ntrain))
      stop("improper invocation of npkernelsum constructor")

    d = list(
      bw = bws$bw,
      data.names = bws$data.names,
      nobs = dim(eval)[1],
      ndim = bws$ndim,
      nord = bws$nord,
      nuno = bws$nuno,
      ncon = bws$ncon,
      pscaling = bws$pscaling,
      ptype = bws$ptype,
      pckertype = bws$pckertype,
      pukertype = bws$pukertype,
      pokertype = bws$pokertype,
      eval = eval,
      ksum = ksum,
      kw = kw,
      p.ksum = p.ksum,
      ntrain = ntrain,
      trainiseval = trainiseval
      )

    if (!is.null(p.kw))
      d$p.kw <- p.kw

    class(d) = "npkernelsum"

    d
  }

print.npkernelsum <- function(x, digits=NULL, ...){
  eval.points <- if (x$trainiseval) "" else paste0(" and ", x$nobs, " evaluation points,")
  cat("\nKernel Sum data: ", x$ntrain, " training points,",
      eval.points,
      " in ",x$ndim," variable(s)\n",sep="")

  print(matrix(x$bw,ncol=x$ndim,dimnames=list(paste(x$pscaling,":",sep=""),x$data.names)))
  
  cat("\nKernel Estimator:",x$pregtype,
      "\nBandwidth Type:",x$ptype,
      "\n\nKernel sum in component 'ksum'")

  if (!is.null(weights))
    cat("\nKernel weights available in component 'kw'")
  
  if (x$ncon > 0)
    cat("\n\nContinuous Kernel Type:",x$pckertype,
        "\nNo. Continuous Vars.:",x$ncon)

  if (x$nuno > 0)
    cat("\n\nUnordered Categorical Kernel Type:",x$pukertype,
        "\nNo. Unordered Categorical Vars.:",x$nuno)

  if (x$nord > 0)
    cat("\n\nOrdered Categorical Kernel Type:",x$pokertype,
        "\nNo. Ordered Categorical Vars.:",x$nord)

  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

npscoef <-
  function(bws, ...){
    args <- list(...)

    if (!missing(bws)){
      if (is.recursive(bws)){
        if (!is.null(bws$formula) && is.null(args$txdat))
          UseMethod("npscoef",bws$formula)
        else if (!is.null(bws$call) && is.null(args$txdat))
          UseMethod("npscoef",bws$call)
        else if (!is.call(bws))
          UseMethod("npscoef",bws)
        else
          UseMethod("npscoef",NULL)
      } else {
        UseMethod("npscoef", NULL)
      }
    } else {
      UseMethod("npscoef", NULL)
    }
  }

npscoef.formula <-
  function(bws, data = NULL, newdata = NULL, y.eval = FALSE, ...){

    tt <- terms(bws)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)
    tmf <- bws$call[c(1,m)]
    tmf[[1]] <- as.name("model.frame")
    tmf[["formula"]] <- tt
    mf.args <- as.list(tmf)[-1L]
    umf <- tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

    response.name <- attr(tmf, "names")[attr(attr(tmf, "terms"), "response")]
    tydat <- model.response(tmf)
    txdat <- tmf[, bws$chromoly[[2]], drop = FALSE]
    miss.z <- !(length(bws$chromoly) == 3)
    if (!miss.z)
      tzdat <- tmf[, bws$chromoly[[3]], drop = FALSE]

    has.eval <- !is.null(newdata)
    if (has.eval) {
      if (!y.eval){
        npValidateNewdataFormula(newdata, tt, include.response = FALSE)
        tt <- delete.response(tt)

        orig.ts <- .np_terms_ts_mask(terms_obj = tt, data = newdata)
        
        ## delete.response clobbers predvars, which is used for timeseries objects
        ## so we need to reconstruct it

        if(all(orig.ts)){
          args <- (as.list(attr(tt, "variables"))[-1])
          attr(tt, "predvars") <- as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), args))))
        }else if(any(orig.ts)){
          arguments <- (as.list(attr(tt, "variables"))[-1])
          arguments.normal <- arguments[which(!orig.ts)]
          arguments.timeseries <- arguments[which(orig.ts)]

          ix <- sort(c(which(orig.ts),which(!orig.ts)),index.return = TRUE)$ix
          attr(tt, "predvars") <- bquote(.(as.call(c(quote(cbind),as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments.timeseries)))),arguments.normal,check.rows = TRUE)))[,.(ix)])
        }else{
          attr(tt, "predvars") <- attr(tt, "variables")
        }
      }
      
      if (y.eval)
        npValidateNewdataFormula(newdata, tt, include.response = TRUE)
      umf.args <- list(formula = tt, data = newdata)
      umf <- do.call(stats::model.frame, umf.args, envir = parent.frame())
      emf <- umf

      if (y.eval)
        eydat <- model.response(emf)
      
      exdat <- emf[, bws$chromoly[[2]], drop = FALSE]
      if (!miss.z)
        ezdat <- emf[, bws$chromoly[[3]], drop = FALSE]
    }


    sc.args <- list(txdat = txdat, tydat = tydat)
    if (!miss.z)
      sc.args$tzdat <- tzdat
    if (has.eval) {
      sc.args$exdat <- exdat
      if (y.eval)
        sc.args$eydat <- eydat
      if (!miss.z)
        sc.args$ezdat <- ezdat
    }
    sc.args$bws <- bws
    ev <- do.call(npscoef, c(sc.args, list(...)))

    if (length(response.name) == 1L && !is.na(response.name) && nzchar(response.name)) {
      if (!is.null(ev$bws))
        ev$bws$ynames <- response.name
    }

    ev$omit <- attr(umf,"na.action")
    ev$rows.omit <- as.vector(ev$omit)
    ev$nobs.omit <- length(ev$rows.omit)

    ev$mean <- napredict(ev$omit, ev$mean)
    ev$merr <- napredict(ev$omit, ev$merr)

    if(ev$residuals){
        ev$resid <- naresid(ev$omit, ev$resid)
    }    
    return(ev)
  }

npscoef.call <-
  function(bws, ...) {
    call.args <- list(
      txdat = .np_eval_bws_call_arg(bws, "xdat"),
      tydat = .np_eval_bws_call_arg(bws, "ydat")
    )
    if (!is.null(bws$zdati))
      call.args$tzdat <- .np_eval_bws_call_arg(bws, "zdat")
    call.args$bws <- bws
    do.call(npscoef, c(call.args, list(...)))
  }

.np_scoef_fit_progress_begin <- function(handoff = FALSE, detail = NULL) {
  state <- .np_progress_begin(
    "Fitting smooth coefficient model",
    surface = "bandwidth"
  )

  if (isTRUE(handoff)) {
    state <- .np_progress_show_now(
      state = state,
      detail = detail
    )
  }

  state
}

npscoef.default <- function(bws, txdat, tydat, tzdat, nomad = FALSE, ...) {
  sc <- sys.call()
  sc.names <- names(sc)
  nomad <- npValidateScalarLogical(nomad, "nomad")

  ## here we check to see if the function was called with tdat =
  ## if it was, we need to catch that and map it to dat =
  ## otherwise the call is passed unadulterated to npudensbw

  bws.named <- any(sc.names == "bws")
  txdat.named <- any(sc.names == "txdat")
  tydat.named <- any(sc.names == "tydat")
  tzdat.named <- any(sc.names == "tzdat")

  no.bws <- missing(bws)
  no.txdat <- missing(txdat)
  no.tydat <- missing(tydat)
  no.tzdat <- missing(tzdat)
  has.explicit.bws <- (!no.bws) && isa(bws, "scbandwidth")

  ## if bws was passed in explicitly, do not compute bandwidths

  if(txdat.named)
    txdat <- toFrame(txdat)

  ## if(tydat.named)
  ## tydat <- toFrame(tydat)

  if(tzdat.named)
    tzdat <- toFrame(tzdat)

  sc.bw <- sc
  
  sc.bw[[1]] <- quote(npscoefbw)

  bws.formula <- (!no.bws) && inherits(bws, "formula")
  if (bws.formula) {
    ib <- match("bws", names(sc.bw), nomatch = 0L)
    if (ib > 0L) names(sc.bw)[ib] <- "formula"
  }

  if(bws.named && !bws.formula){
    sc.bw$bandwidth.compute <- FALSE
  }

  ostxy <- c('txdat','tydat','tzdat')
  nstxy <- c('xdat','ydat','zdat')
  
  m.txy <- match(ostxy, names(sc.bw), nomatch = 0)

  if(any(m.txy > 0)) {
    names(sc.bw)[m.txy] <- nstxy[m.txy > 0]
  }
    
  use.outer.bandwidth.progress <- !.np_bw_call_uses_nomad_degree_search(
    sc.bw,
    caller_env = parent.frame()
  )

  tbw <- if (!has.explicit.bws) {
    if (use.outer.bandwidth.progress) {
      .np_progress_select_bandwidth_enhanced(
        "Selecting smooth coefficient bandwidth",
        .np_eval_bw_call(sc.bw, caller_env = parent.frame())
      )
    } else {
      .np_eval_bw_call(sc.bw, caller_env = parent.frame())
    }
  } else {
    .np_eval_bw_call(sc.bw, caller_env = parent.frame())
  }

  ## because of some ambiguities in how the function might be called
  ## we only drop up to two unnamed arguments, when sometimes dropping
  ## three would be appropriate.  also, for simplicity, we don't allow
  ## for inconsistent mixes of named/unnamed arguments so bws is named
  ## or unnamed, and t[xyz]dat collectively either named or unnamed

  call.args <- list(bws = tbw)
  if (no.bws) {
    call.args$txdat <- txdat
    call.args$tydat <- tydat
    if (!no.tzdat) call.args$tzdat <- tzdat
  } else {
    if (txdat.named) call.args$txdat <- txdat
    if (tydat.named) call.args$tydat <- tydat
    if (tzdat.named) call.args$tzdat <- tzdat
    if ((!bws.named) && (!txdat.named) && (!no.tydat) && (!tydat.named)) {
      call.args <- c(call.args, list(tydat))
    }
    if ((!bws.named) && (!txdat.named) && (!no.tzdat) && (!tzdat.named)) {
      call.args <- c(call.args, list(tzdat))
    }
  }
  if (!has.explicit.bws)
    call.args$.np_fit_progress_handoff <- TRUE
  do.call(npscoef, c(call.args, list(...)))

}

.np_scoef_fit_internal <-
  function(bws,
           txdat = stop("training data 'txdat' missing"),
           tydat = stop("training data 'tydat' missing"),
           tzdat = NULL,
           exdat,
           eydat,
           ezdat,
           betas = FALSE,
           errors = TRUE,
           iterate = TRUE,
           leave.one.out = FALSE,
           maxiter = 100,
           residuals = FALSE,
           tol = .Machine$double.eps,
           ...){

    fit.start <- proc.time()[3]
    residuals <- npValidateScalarLogical(residuals, "residuals")
    errors <- npValidateScalarLogical(errors, "errors")
    iterate <- npValidateScalarLogical(iterate, "iterate")
    leave.one.out <- npValidateScalarLogical(leave.one.out, "leave.one.out")
    betas <- npValidateScalarLogical(betas, "betas")
    if (!is.numeric(maxiter) || length(maxiter) != 1L || is.na(maxiter) ||
        !is.finite(maxiter) || maxiter < 1 || maxiter != floor(maxiter))
      stop("'maxiter' must be a positive integer")
    if (!is.numeric(tol) || length(tol) != 1L || is.na(tol) ||
        !is.finite(tol) || tol < 0)
      stop("'tol' must be a finite numeric scalar >= 0")
    maxiter <- as.integer(maxiter)
    tol <- as.double(tol)
    regtype <- if (is.null(bws$regtype)) "lc" else bws$regtype
    dots <- list(...)
    fit.progress.allow <- !isFALSE(dots$.np_fit_progress_allow) &&
      isTRUE(.np_progress_enabled(domain = "bandwidth"))
    fit.progress.handoff <- fit.progress.allow && isTRUE(dots$.np_fit_progress_handoff)
    fit.progress <- NULL
    fit.progress.active <- FALSE
    fit.progress.step <- NULL
    if (isTRUE(fit.progress.allow)) {
      fit.progress <- .np_scoef_fit_progress_begin(
        handoff = fit.progress.handoff,
        detail = if (fit.progress.handoff) "building moments" else NULL
      )
      fit.progress.active <- TRUE
      fit.progress.counter <- 0L
      fit.progress.step <- function(detail) {
        fit.progress.counter <<- fit.progress.counter + 1L
        fit.progress <<- .np_progress_step(
          fit.progress,
          done = fit.progress.counter,
          detail = detail
        )
        invisible(NULL)
      }
      on.exit({
        if (isTRUE(fit.progress.active))
          .np_progress_abort(fit.progress)
      }, add = TRUE)
    }

    miss.z <- missing(tzdat) || is.null(tzdat)

    miss.ex = missing(exdat)
    miss.ey = missing(eydat)

    ## if miss.ex then if !miss.ey then ey and tx must match, to get
    ## oos errors alternatively if miss.ey you get is errors if
    ## !miss.ex then if !miss.ey then ey and ex must match, to get oos
    ## errors alternatively if miss.ey you get NO errors since we
    ## don't evaluate on the training data

    txdat <- toFrame(txdat)

    if (!(is.vector(tydat) || is.factor(tydat)))
      stop("'tydat' must be a vector or a factor")

    if (!miss.z)
      tzdat <- toFrame(tzdat)

    if (!miss.ex){
      exdat <- toFrame(exdat)

      if (!miss.z)
        ezdat <- toFrame(ezdat)

      if (! txdat %~% exdat )
        stop("'txdat' and 'exdat' are not similar data frames!")

      if (!miss.ey){
        if (dim(exdat)[1] != length(eydat))
          stop("number of evaluation data 'exdat' and dependent data 'eydat' do not match")
      }

    } else if(!miss.ey) {
      if (dim(txdat)[1] != length(eydat))
        stop("number of training data 'txdat' and dependent data 'eydat' do not match")
    }

    if(iterate && !is.null(bws$bw.fitted) && !miss.ex){
      .np_warning("iteration is not supported for out of sample evaluations; using overall bandwidths")
      iterate = FALSE
    }

    ## catch and destroy NA's
    keep.rows <- rep_len(TRUE, nrow(txdat))
    train.df <- data.frame(txdat, tydat)
    if (!miss.z)
      train.df <- data.frame(train.df, tzdat)
    rows.omit <- attr(na.omit(train.df), "na.action")
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Training data has no rows without NAs")

    txdat <- txdat[keep.rows,,drop = FALSE]
    tydat <- tydat[keep.rows]
    if (!miss.z)
      tzdat <- tzdat[keep.rows,, drop = FALSE]

    if (!miss.ex){
      keep.eval <- rep_len(TRUE, nrow(exdat))
      eval.df <- data.frame(exdat)
      if (!miss.ey)
        eval.df <- data.frame(eval.df, eydat)
      if (!miss.z)
        eval.df <- data.frame(eval.df, ezdat)
      rows.omit <- attr(na.omit(eval.df), "na.action")
      if (length(rows.omit) > 0L)
        keep.eval[as.integer(rows.omit)] <- FALSE

      exdat <- exdat[keep.eval,,drop = FALSE]
      if (!miss.ey)
        eydat <- eydat[keep.eval]
      if (!miss.z)
        ezdat <- ezdat[keep.eval,, drop = FALSE]

      if (!any(keep.eval))
        stop("Evaluation data has no rows without NAs")
    }

    ## convert tydat, eydat to numeric, from a factor with levels from the y-data
    ## used during bandwidth selection.

    if (is.factor(tydat)){
      tydat <- adjustLevels(as.data.frame(tydat), bws$ydati)[,1]
      tydat <- (bws$ydati$all.dlev[[1]])[as.integer(tydat)]
    }
    else
      tydat <- as.double(tydat)

    if (miss.ey)
      eydat <- double()
    else {
      if (is.factor(eydat)){
        eydat <- adjustLevels(as.data.frame(eydat), bws$ydati)[,1]
        eydat <- (bws$ydati$all.dlev[[1]])[as.integer(eydat)]
      }
      else
        eydat <- as.double(eydat)
    }

    ## re-assign levels in training and evaluation data to ensure correct
    ## conversion to numeric type.

    txdat <- adjustLevels(txdat, bws$xdati)
    if (!miss.z)
      tzdat <- adjustLevels(tzdat, bws$zdati)

    if (!miss.ex){
      exdat <- adjustLevels(exdat, bws$xdati)
      if (!miss.z)
        ezdat <- adjustLevels(ezdat, bws$zdati)
    }

    ## grab the evaluation data before it is converted to numeric
    if(miss.ex){
      teval <- txdat
      if (!miss.z)
        teval <- list(exdat = txdat, ezdat = tzdat)
    } else {
      teval <- exdat
      if (!miss.z)
        teval <- list(exdat = exdat, ezdat = ezdat)
    }

    ## put the unordered, ordered, and continuous data in their own objects
    ## data that is not a factor is continuous.

    txdat <- toMatrix(txdat)

    if (!miss.ex){
      exdat <- toMatrix(exdat)
    }

    if (miss.z){
      tzdat <- txdat
      if (!miss.ex)
        ezdat <- exdat
    }
    ## from this point on txdat and exdat have been recast as matrices
    ## construct 'W' matrix

    spec <- .npscoef_canonical_spec(source = bws, zdat = tzdat, where = "npscoef")
    reg.engine <- spec$regtype.engine
    W.train <- W <- as.matrix(data.frame(1,txdat))
    maxPenalty <- sqrt(.Machine$double.xmax)
    tnrow <- nrow(txdat)
    enrow <- (if (miss.ex) nrow(txdat) else nrow(exdat))

    if (!miss.ex)
      W <- as.matrix(data.frame(1,exdat))

    safe_chol2inv <- function(a, ridge0, eps, maxiter = 1000L){
      nc.local <- ncol(a)
      I.local <- diag(rep(1.0, nc.local))
      ridge.local <- max(as.double(ridge0), 0.0)
      for (iter in seq_len(maxiter)) {
        cm <- tryCatch(
          chol2inv(chol(a + ridge.local * I.local)),
          error = function(e) NULL
        )
        if (!is.null(cm))
          return(cm)
        ridge.local <- ridge.local + eps
      }
      NULL
    }

    fast_moment_solve <- function(tww.slice, tyw.slice, ridge.add, ridge.val) {
      ncoef.local <- nrow(tww.slice)

      if (ncoef.local == 1L) {
        denom <- tww.slice[1L, 1L] + ridge.add
        if (!is.finite(denom) || abs(denom) <= .Machine$double.eps)
          return(NULL)
        return((tyw.slice[1L] + ridge.val) / denom)
      }

      if (ncoef.local == 2L) {
        a11 <- tww.slice[1L, 1L] + ridge.add
        a12 <- tww.slice[1L, 2L]
        a21 <- tww.slice[2L, 1L]
        a22 <- tww.slice[2L, 2L] + ridge.add
        b1 <- tyw.slice[1L] + ridge.val
        b2 <- tyw.slice[2L]
        detA <- a11 * a22 - a12 * a21
        scale <- max(abs(c(a11, a12, a21, a22, b1, b2)), 1.0)
        if (!is.finite(detA) || abs(detA) <= .Machine$double.eps * scale)
          return(NULL)
        return(c(
          (b1 * a22 - a12 * b2) / detA,
          (a11 * b2 - b1 * a21) / detA
        ))
      }

      NULL
    }

    solve_moment_system <- function(tyw, tww, W.eval.design, Wz.eval = NULL, progress_detail = NULL) {
      neval.local <- ncol(tyw)
      ncoef <- nrow(tyw)
      pcoef <- ncol(W.eval.design)
      coef.out <- matrix(maxPenalty, nrow = pcoef, ncol = neval.local)
      theta.out <- if (is.null(Wz.eval)) NULL else matrix(NA_real_, nrow = ncoef, ncol = neval.local)
      ridge.grid <- npRidgeSequenceAdditive(n.train = tnrow, cap = 1.0)
      ridge <- rep.int(ridge.grid[1L], neval.local)
      ridge.idx <- rep.int(1L, neval.local)
      doridge <- rep.int(TRUE, neval.local)

      while(any(doridge)){
        iloo <- seq_len(neval.local)[doridge]
        for (ii in iloo) {
          doridge[ii] <- FALSE
          ridge.val <- ridge[ii]*tyw[,ii][1]/NZD(tww[,,ii][1,1])
          theta.ii <- fast_moment_solve(
            tww.slice = tww[, , ii],
            tyw.slice = tyw[, ii],
            ridge.add = ridge[ii],
            ridge.val = ridge.val
          )
          if (is.null(theta.ii)) {
            theta.ii <- tryCatch(
              solve(tww[,,ii] + diag(rep(ridge[ii], ncoef)),
                    tyw[,ii] + c(ridge.val, rep(0, ncoef - 1))),
              error = function(e) e
            )
          }
          if (inherits(theta.ii, "error")) {
            ridge.idx[ii] <- ridge.idx[ii] + 1L
            if (ridge.idx[ii] <= length(ridge.grid)) {
              ridge[ii] <- ridge.grid[ridge.idx[ii]]
              doridge[ii] <- TRUE
            }
            theta.ii <- rep(maxPenalty, ncoef)
          }

          if (is.null(Wz.eval)) {
            coef.out[,ii] <- theta.ii
          } else {
            theta.out[,ii] <- theta.ii
            coef.out[,ii] <- as.vector(crossprod(
              Wz.eval[ii,],
              matrix(theta.ii, nrow = ncol(Wz.eval), ncol = pcoef)
            ))
          }
          if (!is.null(fit.progress.step))
            fit.progress.step(progress_detail)
        }
      }

      list(coef = coef.out, theta = theta.out, ridge = ridge)
    }

    solve_single_moment_system <- function(tyw.vec, tww.mat) {
      ncoef <- length(tyw.vec)
      ridge.grid <- npRidgeSequenceAdditive(n.train = tnrow, cap = 1.0)
      ridge <- ridge.grid[1L]
      ridge.idx <- 1L

      repeat {
        ridge.val <- ridge * tyw.vec[1L] / NZD(tww.mat[1L, 1L])
        theta <- fast_moment_solve(
          tww.slice = tww.mat,
          tyw.slice = tyw.vec,
          ridge.add = ridge,
          ridge.val = ridge.val
        )
        if (is.null(theta)) {
          theta <- tryCatch(
            solve(
              tww.mat + diag(rep(ridge, ncoef)),
              tyw.vec + c(ridge.val, rep(0, ncoef - 1L))
            ),
            error = function(e) e
          )
        }
        if (!inherits(theta, "error"))
          return(list(coef = as.double(theta), ridge = ridge))

        ridge.idx <- ridge.idx + 1L
        if (ridge.idx > length(ridge.grid))
          return(list(coef = rep(maxPenalty, ncoef), ridge = ridge))
        ridge <- ridge.grid[ridge.idx]
      }
    }

    lc_moments <- function(z.eval, leave.one.out.eval, u2 = NULL) {
      yW.local <- cbind(tydat, W.train)
      ksum.args <- list(
        txdat = tzdat,
        tydat = yW.local,
        weights = yW.local,
        bws = bws,
        leave.one.out = leave.one.out.eval,
        bandwidth.divide = TRUE
      )
      if (!leave.one.out.eval && !is.null(z.eval))
        ksum.args$exdat <- z.eval
      main.ks <- do.call(npksum, ksum.args)$ksum
      tyw.out <- main.ks[-1L, 1L, , drop = FALSE]
      if (length(dim(tyw.out)) == 3L)
        dim(tyw.out) <- c(dim(tyw.out)[1L], dim(tyw.out)[3L])
      tww.out <- main.ks[-1L, -1L, , drop = FALSE]

      s.out <- NULL
      if (!is.null(u2)) {
        cov.args <- list(
          txdat = tzdat,
          tydat = W.train,
          weights = W.train * as.double(u2),
          bws = bws,
          leave.one.out = leave.one.out.eval,
          bandwidth.divide = TRUE,
          kernel.pow = 2
        )
        if (!leave.one.out.eval && !is.null(z.eval))
          cov.args$exdat <- z.eval
        s.out <- do.call(npksum, cov.args)$ksum
      }

      list(tyw = tyw.out, tww = tww.out, s = s.out)
    }

    do.iterate <- (iterate && !is.null(bws$bw.fitted) && miss.ex && identical(reg.engine, "lc"))
    gate.zdat <- if (miss.ex) tzdat else rbind(tzdat, ezdat)
    fast.largeh.lc <- identical(reg.engine, "lc") &&
      !leave.one.out &&
      identical(bws$type, "fixed") &&
      !do.iterate &&
      .npscoefbw_fast_eligible(bws, eval.zdat = gate.zdat)
    fast.largeh.lp1 <- identical(reg.engine, "lp") &&
      !leave.one.out &&
      identical(bws$type, "fixed") &&
      identical(spec$basis.engine, "glp") &&
      !isTRUE(spec$bernstein.basis.engine) &&
      all(as.integer(spec$degree.engine) == 1L) &&
      .npscoefbw_fast_eligible(bws, eval.zdat = gate.zdat)
    fast.largeh <- fast.largeh.lc || fast.largeh.lp1

    lc_fast_global_moments <- function(z.eval.one, u2 = NULL) {
      yW.local <- cbind(tydat, W.train)
      ksum.args <- list(
        txdat = tzdat,
        tydat = yW.local,
        weights = yW.local,
        exdat = z.eval.one,
        bws = bws,
        bandwidth.divide = TRUE
      )
      main.ks <- do.call(npksum, ksum.args)$ksum
      out <- list(
        tyw = as.double(main.ks[-1L, 1L, 1L]),
        tww = main.ks[-1L, -1L, 1L, drop = TRUE]
      )

      if (!is.null(u2)) {
        cov.args <- list(
          txdat = tzdat,
          tydat = W.train,
          weights = W.train * as.double(u2),
          exdat = z.eval.one,
          bws = bws,
          bandwidth.divide = TRUE,
          kernel.pow = 2
        )
        out$s <- do.call(npksum, cov.args)$ksum[, , 1L, drop = TRUE]
      } else {
        out$s <- NULL
      }

      out
    }

    lp_state <- if (identical(reg.engine, "lp")) {
      .npscoef_lp_state(
        bws = bws,
        tzdat = tzdat,
        ezdat = if (miss.ex) tzdat else ezdat,
        leave.one.out = leave.one.out,
        where = "npscoef"
      )
    } else {
      NULL
    }

    lp_tensor_moments <- function(state, u2 = NULL) {
      tensor.train <- .npscoef_row_tensor_design(W.train, state$W.train)
      ytensor <- cbind(tydat, tensor.train)
      ksum.args <- list(
        txdat = state$z.train,
        tydat = ytensor,
        weights = ytensor,
        bws = state$rbw,
        leave.one.out = state$leave.one.out,
        bandwidth.divide = TRUE
      )
      if (!state$leave.one.out)
        ksum.args$exdat <- state$z.eval
      main.ks <- do.call(npksum, ksum.args)$ksum
      tyw.out <- main.ks[-1L, 1L, , drop = FALSE]
      if (length(dim(tyw.out)) == 3L)
        dim(tyw.out) <- c(dim(tyw.out)[1L], dim(tyw.out)[3L])
      tww.out <- main.ks[-1L, -1L, , drop = FALSE]

      s.out <- NULL
      if (!is.null(u2)) {
        cov.args <- list(
          txdat = state$z.train,
          tydat = tensor.train,
          weights = tensor.train * as.double(u2),
          bws = state$rbw,
          leave.one.out = state$leave.one.out,
          bandwidth.divide = TRUE,
          kernel.pow = 2
        )
        if (!state$leave.one.out)
          cov.args$exdat <- state$z.eval
        s.out <- do.call(npksum, cov.args)$ksum
      }

      list(tyw = tyw.out, tww = tww.out, s = s.out)
    }

    moments <- NULL
    if (fast.largeh.lc) {
      eval.z.one <- if (miss.ex) tzdat[1L, , drop = FALSE] else ezdat[1L, , drop = FALSE]
      fast.eval <- lc_fast_global_moments(z.eval.one = eval.z.one)
      if (!is.null(fit.progress.step))
        fit.progress.step("solving global coefficients")
      fast.solve <- solve_single_moment_system(
        tyw.vec = fast.eval$tyw,
        tww.mat = fast.eval$tww
      )
      if (!is.null(fit.progress.step))
        fit.progress.step("assembling fitted values")
      coef.vec <- fast.solve$coef
      coef.mat <- matrix(coef.vec, nrow = length(coef.vec), ncol = enrow)
      ridge <- rep.int(fast.solve$ridge, enrow)
    } else if (fast.largeh.lp1) {
      fast.eval <- .npscoef_lp1_largeh_global_fit(
        bws = bws,
        tzdat = tzdat,
        ezdat = if (miss.ex) tzdat else ezdat,
        W.train = W.train,
        tydat = tydat,
        leave.one.out = leave.one.out,
        where = "npscoef",
        solver = solve_single_moment_system
      )
      if (!is.null(fit.progress.step))
        fit.progress.step("solving global coefficients")
      coef.mat <- fast.eval$coef
      ridge <- rep.int(fast.eval$ridge, enrow)
    } else if (identical(reg.engine, "lc")) {
      moments <- lc_moments(
        z.eval = if (miss.ex) NULL else ezdat,
        leave.one.out.eval = leave.one.out
      )
      solver <- solve_moment_system(
        tyw = moments$tyw,
        tww = moments$tww,
        W.eval.design = W,
        progress_detail = "solving coefficient rows"
      )
    } else {
      moments <- lp_tensor_moments(lp_state)
      solver <- solve_moment_system(
        tyw = moments$tyw,
        tww = moments$tww,
        W.eval.design = W,
        Wz.eval = lp_state$W.eval,
        progress_detail = "solving coefficient rows"
      )
    }

    if (!fast.largeh) {
      coef.mat <- solver$coef
      ridge <- solver$ridge
    }

    if (iterate && !is.null(bws$bw.fitted) && miss.ex && !identical(reg.engine, "lc"))
      .np_warning("iterate=TRUE currently supports regtype='lc' for npscoef; using iterate=FALSE")
    if (do.iterate){
      resid <- tydat - sapply(seq_len(enrow), function(i) { W[i,, drop = FALSE] %*% coef.mat[,i] })

      i = 0
      max.err <- .Machine$double.xmax
      aydat <- abs(tydat) + .Machine$double.eps

      n.part <- (ncol(txdat)+1)

      while((max.err > tol) && ((i <- i + 1) <= maxiter)){
        resid.old <- resid
        for (j in seq_len(n.part)) {
          ## estimate partial residuals
          partial <- W[,j] * coef.mat[j,] + resid

          ## use to calculate new beta implicitly

          twww <- npksum(txdat=tzdat,
                         tydat=cbind(partial * W[,j],W[,j]^2),
                         weights=cbind(partial * W[,j],1),
                         bws=bws,
                         leave.one.out=leave.one.out)$ksum

          coef.mat[j,] <- twww[1,2,]/NZD(twww[2,2,])

          ## estimate new full residuals
          resid <- partial - W[,j] * coef.mat[j,]
          if (!is.null(fit.progress.step))
            fit.progress.step(sprintf("backfit cycle %d partial %d/%d", i, j, n.part))
          ## repeat for consistency ?
        }
        max.err <- max(abs(resid.old - resid)/aydat)
      }
      if (max.err > tol)
        .np_warning(paste("backfit iterations did not converge. max err= ", max.err,", tol= ", tol,", maxiter= ", maxiter, sep=''))
      mean <- tydat - resid
    } else {
      mean <- sapply(seq_len(enrow), function(i) { W[i,, drop = FALSE] %*% coef.mat[,i] })
    }

    if (!miss.ey) {
      RSQ = RSQfunc(eydat, mean)
      MSE = MSEfunc(eydat, mean)
      MAE = MAEfunc(eydat, mean)
      MAPE = MAPEfunc(eydat, mean)
      CORR = CORRfunc(eydat, mean)
      SIGN = SIGNfunc(eydat, mean)
    } else if(miss.ex) {
      RSQ = RSQfunc(tydat, mean)
      MSE = MSEfunc(tydat, mean)
      MAE = MAEfunc(tydat, mean)
      MAPE = MAPEfunc(tydat, mean)
      CORR = CORRfunc(tydat, mean)
      SIGN = SIGNfunc(tydat, mean)
    }

    if (errors || (residuals && miss.ex)) {
      if (errors) {
        if (fast.largeh.lc) {
          if (miss.ex) {
            train.solve <- fast.solve
            mean.fit <- mean
          } else {
            train.fast <- lc_fast_global_moments(z.eval.one = tzdat[1L, , drop = FALSE])
            train.solve <- solve_single_moment_system(
              tyw.vec = train.fast$tyw,
              tww.mat = train.fast$tww
            )
            mean.fit <- as.vector(W.train %*% train.solve$coef)
          }
          resid <- tydat - mean.fit
          u2.W <- resid^2
          s.fast <- lc_fast_global_moments(
            z.eval.one = if (miss.ex) tzdat[1L, , drop = FALSE] else ezdat[1L, , drop = FALSE],
            u2 = u2.W
          )$s
          if (!is.null(fit.progress.step))
            fit.progress.step("estimating standard errors")
          cm.fast <- safe_chol2inv(fast.eval$tww, fast.solve$ridge, 1.0 / nrow(txdat))
          merr <- rep(NA_real_, enrow)
          beta.se <- matrix(NA_real_, nrow = enrow, ncol = nrow(coef.mat))
          if (!is.null(cm.fast)) {
            vcv.beta.fast <- cm.fast %*% s.fast %*% cm.fast
            merr <- sqrt(pmax(rowSums((W %*% vcv.beta.fast) * W), 0.0))
            beta.se[] <- rep(sqrt(pmax(diag(vcv.beta.fast), 0.0)), each = enrow)
          }
        } else if (fast.largeh.lp1) {
          if (miss.ex) {
            train.fast <- fast.eval
            mean.fit <- mean
          } else {
            train.fast <- .npscoef_lp1_largeh_global_fit(
              bws = bws,
              tzdat = tzdat,
              ezdat = tzdat,
              W.train = W.train,
              tydat = tydat,
              leave.one.out = leave.one.out,
              where = "npscoef",
              solver = solve_single_moment_system
            )
            mean.fit <- sapply(seq_len(tnrow), function(i) { W.train[i,, drop = FALSE] %*% train.fast$coef[,i] })
          }
          resid <- tydat - mean.fit
          u2.W <- resid^2
          s.fast <- .npscoef_lp1_largeh_global_fit(
            bws = bws,
            tzdat = tzdat,
            ezdat = if (miss.ex) tzdat else ezdat,
            W.train = W.train,
            tydat = tydat,
            u2 = u2.W,
            leave.one.out = leave.one.out,
            where = "npscoef",
            solver = solve_single_moment_system
          )$s
          if (!is.null(fit.progress.step))
            fit.progress.step("estimating standard errors")
          cm.fast <- safe_chol2inv(fast.eval$tww, fast.eval$ridge, 1.0 / nrow(txdat))
          merr <- rep(NA_real_, enrow)
          beta.se <- matrix(NA_real_, nrow = enrow, ncol = nrow(coef.mat))
          if (!is.null(cm.fast)) {
            vcv.theta.fast <- cm.fast %*% s.fast %*% cm.fast
            for (i in seq_len(enrow)) {
              trans.i <- kronecker(diag(ncol(W)), matrix(fast.eval$lp_state$W.eval[i,], nrow = 1L))
              vcv.beta.fast <- trans.i %*% vcv.theta.fast %*% t(trans.i)
              w.i <- W[i,, drop = FALSE]
              merr[i] <- sqrt(max(drop(w.i %*% vcv.beta.fast %*% t(w.i)), 0.0))
              beta.se[i,] <- sqrt(pmax(diag(vcv.beta.fast), 0.0))
            }
          }
        } else if (miss.ex && !do.iterate) {
          mean.fit <- mean
          resid <- tydat - mean.fit
          u2.W <- resid^2
          if (identical(reg.engine, "lc")) {
            moments$s <- lc_moments(
              z.eval = NULL,
              leave.one.out.eval = leave.one.out,
              u2 = u2.W
            )$s
          } else {
            moments$s <- lp_tensor_moments(lp_state, u2 = u2.W)$s
          }
        } else if (identical(reg.engine, "lc")) {
          train.moments <- lc_moments(z.eval = NULL, leave.one.out.eval = leave.one.out)
          train.solve <- solve_moment_system(
            tyw = train.moments$tyw,
            tww = train.moments$tww,
            W.eval.design = W.train
          )
          mean.fit <- sapply(seq_len(tnrow), function(i) { W.train[i,, drop = FALSE] %*% train.solve$coef[,i] })
          u2.W <- (resid <- tydat - mean.fit)^2
          moments$s <- lc_moments(
            z.eval = if (miss.ex) NULL else ezdat,
            leave.one.out.eval = leave.one.out,
            u2 = u2.W
          )$s
        } else {
          lp_state.err <- .npscoef_lp_state(
            bws = bws,
            tzdat = tzdat,
            ezdat = tzdat,
            leave.one.out = leave.one.out,
            where = "npscoef"
          )
          train.moments <- lp_tensor_moments(lp_state.err)
          train.solve <- solve_moment_system(
            tyw = train.moments$tyw,
            tww = train.moments$tww,
            W.eval.design = W.train,
            Wz.eval = lp_state.err$W.eval
          )
          mean.fit <- sapply(seq_len(tnrow), function(i) { W.train[i,, drop = FALSE] %*% train.solve$coef[,i] })
          u2.W <- (resid <- tydat - mean.fit)^2
          moments$s <- lp_tensor_moments(lp_state, u2 = u2.W)$s
        }
      } else if (residuals && miss.ex) {
        resid <- tydat - mean
      }
    }

    if (!errors)
      beta.se <- NULL
    if(errors && !fast.largeh){
      u2 <- as.double(u2.W)
      merr <- rep(NA_real_, enrow)
      beta.se <- matrix(NA_real_, nrow = enrow, ncol = nrow(coef.mat))
      for (i in seq_len(enrow)) {
        cm <- safe_chol2inv(moments$tww[,,i], ridge[i], 1.0 / nrow(txdat))
        if (is.null(cm))
          next
        s.mat <- moments$s[,,i]

        if (identical(reg.engine, "lc")) {
          vcv.beta <- cm %*% s.mat %*% cm
        } else {
          vcv.theta <- cm %*% s.mat %*% cm
          trans.i <- kronecker(diag(ncol(W)), matrix(lp_state$W.eval[i,], nrow = 1L))
          vcv.beta <- trans.i %*% vcv.theta %*% t(trans.i)
        }

        w.i <- W[i,,drop=FALSE]
        merr[i] <- sqrt(max(drop(w.i %*% vcv.beta %*% t(w.i)), 0.0))
        beta.se[i,] <- sqrt(pmax(diag(vcv.beta), 0.0))
        if (!is.null(fit.progress.step))
          fit.progress.step("estimating standard errors")
      }

    }

    sc.obj.args <- list(
      bws = bws,
      eval = teval,
      mean = mean,
      residuals = residuals,
      betas = betas,
      ntrain = nrow(txdat),
      trainiseval = miss.ex
    )
    if (errors && !do.iterate)
      sc.obj.args$merr <- merr
    if (ncol(txdat) > 0L) {
      sc.obj.args$grad <- t(coef.mat[-1,,drop = FALSE])
      if (errors && !do.iterate && !is.null(beta.se))
        sc.obj.args$gerr <- beta.se[, -1, drop = FALSE]
    }
    if (betas)
      sc.obj.args$beta <- t(coef.mat)
    if (residuals)
      sc.obj.args$resid <- resid
    if (!(miss.ey && !miss.ex))
      sc.obj.args$xtra <- c(RSQ, MSE, MAE, MAPE, CORR, SIGN)
    ev <- do.call(smoothcoefficient, sc.obj.args)
    fit.elapsed <- proc.time()[3] - fit.start
    optim.time <- if (!is.null(bws$total.time) && is.finite(bws$total.time)) as.double(bws$total.time) else NA_real_
    total.time <- fit.elapsed + (if (is.na(optim.time)) 0.0 else optim.time)
    ev$timing <- bws$timing
    ev$total.time <- total.time
    ev$optim.time <- optim.time
    ev$fit.time <- fit.elapsed
    ev$nomad.time <- if (!is.null(bws$nomad.time) && is.finite(bws$nomad.time)) as.double(bws$nomad.time) else NA_real_
    ev$powell.time <- if (!is.null(bws$powell.time) && is.finite(bws$powell.time)) as.double(bws$powell.time) else NA_real_
    if (isTRUE(fit.progress.active)) {
      fit.progress <- .np_progress_end(fit.progress)
      fit.progress.active <- FALSE
    }
    ev

  }

npscoef.scbandwidth <-
  function(bws,
           txdat = stop("training data 'txdat' missing"),
           tydat = stop("training data 'tydat' missing"),
           tzdat = NULL,
           exdat,
           eydat,
           ezdat,
           betas = FALSE,
           errors = TRUE,
           iterate = TRUE,
           leave.one.out = FALSE,
           maxiter = 100,
           residuals = FALSE,
           tol = .Machine$double.eps,
           ...){
    .np_scoef_fit_internal(
      bws = bws,
      txdat = txdat,
      tydat = tydat,
      tzdat = tzdat,
      exdat = exdat,
      eydat = eydat,
      ezdat = ezdat,
      betas = betas,
      errors = errors,
      iterate = iterate,
      leave.one.out = leave.one.out,
      maxiter = maxiter,
      residuals = residuals,
      tol = tol,
      ...
    )
  }

test_that("npscoef core path does not reference full-matrix helpers", {
  fn.body <- paste(deparse(body(np:::npscoef.scbandwidth)), collapse = "\n")
  expect_no_match(fn.body, "\\.npscoef_weight_matrix\\(")
  expect_no_match(fn.body, "\\.np_kernel_weights_direct")
  expect_no_match(fn.body, "npscoefhat\\(")
})

test_that("ordinary npscoef fits do not hit smoothcoef full-matrix helper", {
  old.opt <- getOption("np.messages")
  options(np.messages = FALSE)
  on.exit(options(np.messages = old.opt), add = TRUE)

  set.seed(20260306)
  n <- 70L
  x <- runif(n)
  z <- runif(n, min = -1.5, max = 1.5)
  y <- (0.3 + sin(1.4 * z)) * x + 0.15 * cos(z) + rnorm(n, sd = 0.04)

  assign(".np_trace_hits_npscoef", 0L, envir = .GlobalEnv)
  on.exit(rm(".np_trace_hits_npscoef", envir = .GlobalEnv), add = TRUE)
  trace(".npscoef_weight_matrix",
        tracer = quote(assign(".np_trace_hits_npscoef",
                              get(".np_trace_hits_npscoef", envir = .GlobalEnv) + 1L,
                              envir = .GlobalEnv)),
        print = FALSE,
        where = asNamespace("np"))
  on.exit(untrace(".npscoef_weight_matrix", where = asNamespace("np")), add = TRUE)

  run_case <- function(regtype) {
    bw.args <- list(
      xdat = x,
      zdat = z,
      ydat = y,
      bws = 0.28,
      regtype = regtype,
      bandwidth.compute = FALSE
    )
    if (identical(regtype, "lp")) {
      bw.args$basis <- "glp"
      bw.args$degree <- 2L
    }
    bw <- do.call(npscoefbw, bw.args)
    assign(".np_trace_hits_npscoef", 0L, envir = .GlobalEnv)
    fit <- npscoef(
      bws = bw,
      txdat = data.frame(x = x),
      tydat = y,
      tzdat = data.frame(z = z),
      gradients = TRUE,
      errors = TRUE,
      iterate = FALSE,
      betas = TRUE
    )
    expect_s3_class(fit, "smoothcoefficient")
    expect_identical(get(".np_trace_hits_npscoef", envir = .GlobalEnv), 0L, info = regtype)
  }

  run_case("lc")
  run_case("ll")
  run_case("lp")
})

test_that("named bws formula dispatch matches positional semiparametric routes", {
  set.seed(20260323)
  dat <- data.frame(
    y = rnorm(24),
    x1 = runif(24),
    z1 = runif(24)
  )
  ex <- data.frame(
    x1 = seq(0.1, 0.8, length.out = 7),
    z1 = seq(0.8, 0.1, length.out = 7)
  )

  sc.pos <- npscoef(y ~ x1 | z1, data = dat, newdata = ex, errors = FALSE, iterate = FALSE)
  sc.named <- npscoef(bws = y ~ x1 | z1, data = dat, newdata = ex, errors = FALSE, iterate = FALSE)
  expect_equal(as.numeric(sc.named$mean), as.numeric(sc.pos$mean), tolerance = 0)
  expect_equal(as.numeric(sc.named$merr), as.numeric(sc.pos$merr), tolerance = 0)

  pl.pos <- npplreg(y ~ x1 | z1, data = dat, newdata = ex)
  pl.named <- npplreg(bws = y ~ x1 | z1, data = dat, newdata = ex)
  expect_equal(as.numeric(pl.named$mean), as.numeric(pl.pos$mean), tolerance = 0)
  expect_equal(as.numeric(pl.named$merr), as.numeric(pl.pos$merr), tolerance = 0)
})

test_that("plot.plregression restores richer bws and infers overlay data from fitted object", {
  plot.plreg <- getFromNamespace(".np_plot_plregression", "np")

  bws.current <- structure(list(formula = NULL), class = "plbandwidth")
  bws.orig <- structure(list(formula = y ~ x | z), class = "plbandwidth")
  captured <- NULL

  local_mocked_bindings(
    .np_eval_call_arg = function(call, arg, caller_env = parent.frame()) bws.orig,
    .np_plot_call_method = function(method, bws, ...) {
      captured <<- list(method = method, bws = bws, dots = list(...))
      invisible("ok")
    },
    .package = "np"
  )

  object <- structure(
    list(
      bws = bws.current,
      bw = bws.current,
      call = quote(npplreg(bws = saved_bws)),
      trainiseval = TRUE,
      evalx = data.frame(x = c(1, 2, 3)),
      evalz = data.frame(z = c(4, 5, 6)),
      mean = c(0.5, 1.5, 2.5),
      resid = c(0.1, 0.2, 0.3)
    ),
    class = "plregression"
  )

  plot.plreg(object, main = "kept")

  expect_identical(captured$bws, bws.orig)
  expect_identical(captured$dots$xdat, object$evalx)
  expect_identical(captured$dots$zdat, object$evalz)
  expect_equal(captured$dots$ydat, object$mean + object$resid)
  expect_identical(captured$dots$main, "kept")
})

test_that("plot.smoothcoefficient restores richer bws and infers overlay data from fitted object", {
  plot.scoef <- getFromNamespace(".np_plot_smoothcoefficient", "np")

  bws.current <- structure(list(formula = NULL), class = "scbandwidth")
  bws.orig <- structure(list(formula = y ~ x | z), class = "scbandwidth")
  captured <- NULL

  local_mocked_bindings(
    .np_eval_call_arg = function(call, arg, caller_env = parent.frame()) bws.orig,
    .np_plot_call_method = function(method, bws, ...) {
      captured <<- list(method = method, bws = bws, dots = list(...))
      invisible("ok")
    },
    .package = "np"
  )

  object <- structure(
    list(
      bws = bws.current,
      call = quote(npscoef(bws = saved_bws)),
      trainiseval = TRUE,
      eval = list(
        exdat = data.frame(x = c(1, 2, 3)),
        ezdat = data.frame(z = c(4, 5, 6))
      ),
      mean = c(0.5, 1.5, 2.5),
      resid = c(0.1, 0.2, 0.3)
    ),
    class = "smoothcoefficient"
  )

  plot.scoef(object, col = "red")

  expect_identical(captured$bws, bws.orig)
  expect_identical(captured$dots$xdat, object$eval$exdat)
  expect_identical(captured$dots$zdat, object$eval$ezdat)
  expect_equal(captured$dots$ydat, object$mean + object$resid)
  expect_identical(captured$dots$col, "red")
})

test_that("plot.plregression recovers training data for direct formula fits", {
  skip_if_not_installed("np")

  set.seed(3001)
  n <- 50
  x <- runif(n)
  z <- runif(n, -2, 2)
  y <- 1 + 0.6 * x + sin(z) + rnorm(n, sd = 0.1)

  fit <- npplreg(y ~ x | z, regtype = "ll")

  assign("overlay_calls", 0L, envir = .GlobalEnv)
  trace(".np_plot_overlay_points_1d",
        tracer = quote({
          assign("overlay_calls", get("overlay_calls", envir = .GlobalEnv) + 1L, envir = .GlobalEnv)
        }),
        where = asNamespace("np"),
        print = FALSE)
  on.exit(untrace(".np_plot_overlay_points_1d", where = asNamespace("np")), add = TRUE)
  on.exit(rm("overlay_calls", envir = .GlobalEnv), add = TRUE)

  pdf(file = tempfile(fileext = ".pdf"))
  on.exit(dev.off(), add = TRUE)
  out <- suppressWarnings(plot(
    fit,
    perspective = FALSE,
    output = "plot-data",
    errors = "none",
    pch = 4,
    cex = 0.7
  ))

  expect_type(out, "list")
  overlay_calls <- get("overlay_calls", envir = .GlobalEnv)
  expect_gte(overlay_calls, 1L)
})

## Function that implements the univariate test for equality of
## densities described in Maasoumi, E.~and J.S.~Racine (2002),
## "Entropy and Predictability of Stock Market Returns," Journal of
## Econometrics, March, Volume 107, Issue 2, pp 291-312.

npunitest <- function(data.x = NULL,
                      data.y = NULL,
                      method = c("integration","summation"),
                      bootstrap = TRUE,
                      boot.num = 399,
                      bw.x = NULL,
                      bw.y = NULL,                         
                      random.seed = 42,
                      ...) {

  if(is.null(data.x) || is.null(data.y)) stop(" you must enter data vectors for x and y")
  if(is.data.frame(data.x) || is.data.frame(data.y)) stop(" you must enter data vectors (not data frames)")
  if(!identical(class(data.x), class(data.y))) stop(" data vectors must be of same data type")
  if((ncol(data.frame(data.x)) != 1) ||( ncol(data.frame(data.y)) != 1)) stop(" data vectors must have one dimension only")
  if(boot.num < 9) stop(" number of bootstrap replications must be >= 9")
  if(is.numeric(data.x) && (max(data.x) < min(data.y) || max(data.y) < min(data.x))) .np_warning("non-overlapping empirical distributions (see `Details' in ?npunidist)")

  method <- match.arg(method)

  ## Save seed prior to setting

  seed.state <- .np_seed_enter(random.seed)
  .np_progress_note("Computing bandwidths")

  ## If of type ts convert to numeric to handle time series data

  if(is.ts(data.x)) data.x <- as.numeric(data.x)
  if(is.ts(data.y)) data.y <- as.numeric(data.y)

  ## Remove NAs

  data.x <- na.omit(data.x)
  data.y <- na.omit(data.y)

  ## Note - this function accepts numeric and factors/ordered,
  ## however, factors must be integers.

  if(is.numeric(data.x)) {
    if(is.null(bw.x)) bw.x <- bw.SJ(data.x)
    if(is.null(bw.y)) bw.y <- bw.SJ(data.y)
  } else {
    ## Optimal bandwidth for factor (unordered) used for both
    ## (plug-in).
    if(is.null(bw.x)) {
      n <- length(data.x)
      c <- length(unique(data.x))
      if(c <= 1) stop("data.x must contain at least two distinct factor levels")
      xeval <- unique(data.x)
      p <- fitted(npudens(tdat=data.x,edat=xeval,bws=0,...))
      sum.Lambda3 <- c/(c-1)*sum(p*(1-p))
      sum.Lambda2.minus.Lambda1.sq <- c^2/((c-1)^2)*sum(p*(1-p))
      n.sum.Lambda1.sq <- n*sum(((1-c*p)/(c-1))^2)
      bw.x <- sum.Lambda3/(sum.Lambda2.minus.Lambda1.sq+n.sum.Lambda1.sq)
    }
    if(is.null(bw.y)) {
      n <- length(data.y)
      c <- length(unique(data.y))
      if(c <= 1) stop("data.y must contain at least two distinct factor levels")
      yeval <- unique(data.y)
      p <- fitted(npudens(tdat=data.y,edat=xeval,bws=0,...))
      sum.Lambda3 <- c/(c-1)*sum(p*(1-p))
      sum.Lambda2.minus.Lambda1.sq <- c^2/((c-1)^2)*sum(p*(1-p))
      n.sum.Lambda1.sq <- n*sum(((1-c*p)/(c-1))^2)
      bw.y <- sum.Lambda3/(sum.Lambda2.minus.Lambda1.sq+n.sum.Lambda1.sq)
    }
  }
	
  ## Hellinger distance (Srho) between the densities. This function
  ## accepts numeric and factor/ordered.
  
  Srho.univar <- function(data.x,
                          data.y,
                          bw.x,
                          bw.y,
                          method=c("integration","summation")) {

    if((ncol(data.frame(data.x)) != 1) || (ncol(data.frame(data.y)) != 1))  stop(" data vectors must have one dimension only") 
    if(is.numeric(data.x)) {
      if(method=="summation") {
        ## Summation - we use positive support data (data.x) when
        ## evaluating. This is proper and avoids division by
        ## zero. Note the difference between integration and summation
        ## in this case when the variables do not share common support.
        f.data.x <- fitted(npudens(tdat=data.x,edat=data.x,bws=bw.x,...))
        f.data.y <- fitted(npudens(tdat=data.y,edat=data.x,bws=bw.y,...))
        summand <- f.data.y/f.data.x
        ## In summation version we divide densities which can lead to
        ## numerical instability issues not present in the integrand
        ## version (which uses differences instead). We check for this
        ## case, remove offending points, and warn. This traps -Inf,
        ## Inf, and NaN.
        if(!all(is.finite(summand))) {
          .np_warning(" non-finite value in summation-based statistic: integration recommended")
          summand <- summand[is.finite(summand)]
        }
        Srho <- 0.5*mean((1-sqrt(summand))**2)
        if(Srho < 0 || Srho > 1) .np_warning(" numerical instability in summation-based statistic: integration recommended")
        return(Srho)
      } else {
        ## Integration
        h <- function(x,data.x,data.y) {
          f.data.x <- fitted(npudens(tdat=data.x,edat=x,bws=bw.x,...))
          f.data.y <- fitted(npudens(tdat=data.y,edat=x,bws=bw.y,...))
          return(0.5*(sqrt(f.data.x)-sqrt(f.data.y))**2)
        }
        return.integrate <- integrate(h,-Inf,Inf,subdivisions=1e+05,stop.on.error=FALSE,data.x=data.x,data.y=data.y)      
        if(return.integrate$message != "OK") .np_warning(return.integrate$message)
        return(return.integrate$value)
      }
    } else {
      xeval <- unique(data.x)
      p.data.x <- fitted(npudens(tdat=data.x,edat=xeval,bws=bw.x,...))
      p.data.y <- fitted(npudens(tdat=data.y,edat=xeval,bws=bw.y,...))   
      ## Sum the information function over the unique probabilities and
      ## return.
      return(sum(0.5*(sqrt(p.data.x)-sqrt(p.data.y))**2))
    }
  }

  
  ## Compute the test statistic

  .np_progress_note("Computing test statistic")

  test.stat <- Srho.univar(data.x,data.y,bw.x,bw.y,method=method)

  if(bootstrap) {
    
    if(method == "summation") {
      
      ## For summation since we evaluate Srho at the support data.x
      ## (not data.y) we resample from data.x only
      
      data.null <- data.x
      
    } else {
      
      ##  For integration impose the null by resampling from pooled
      ##  data.x and data.y
      
      if(is.numeric(data.x)) {
        data.null <- c(data.x,data.y)
      } else {
        if(is.ordered(data.x)) {
          data.null <- ordered(c(as.character(data.x),as.character(data.y)))
        } else {
          data.null <- factor(c(as.character(data.x),as.character(data.y)))
        }
      }
      
    }
    
    resampled.stat <- numeric(boot.num)
    progress <- .np_progress_begin("Bootstrap replications", total = boot.num, surface = "bootstrap")

    for (b in seq_len(boot.num)) {
      progress <- .np_progress_step(progress, done = b)

      ## Need to think this through... is the null one density? If so
      ## resample from that density for both x and y?
      
      ## Conduct simple iid bootstrap resamples
      
      data.null.x <- data.null[sample.int(length(data.null), size = length(data.x), replace = TRUE)]
      data.null.y <- data.null[sample.int(length(data.null), size = length(data.y), replace = TRUE)]

      resampled.stat[b] <- Srho.univar(data.null.x,data.null.y,bw.x,bw.y,method=method)

    }

    progress <- .np_progress_end(progress)
    
    p.value <- mean(resampled.stat > test.stat)
    
  }

  ## Restore seed

  .np_seed_exit(seed.state)

  if(bootstrap) {
    unitest(Srho = test.stat,
               Srho.bootstrap = resampled.stat,
               P = p.value,
               bootstrap = bootstrap,
               boot.num = boot.num,
               bw.x = bw.x,
               bw.y = bw.y)
  } else {
    unitest(Srho = test.stat,
               Srho.bootstrap = NULL,
               P = NULL,
               bootstrap = bootstrap,
               boot.num = NULL,
               bw.x = bw.x,
               bw.y = bw.y)
  }

}

.np_condist_slice_groups <- function(xeval) {
  xeval <- toFrame(xeval)

  if (ncol(xeval) == 0L) {
    split(seq_len(nrow(xeval)), factor(rep.int("all", nrow(xeval))), drop = TRUE)
  } else {
    groups <- do.call(interaction, c(unname(xeval), list(drop = TRUE, lex.order = TRUE)))
    split(seq_len(nrow(xeval)), groups, drop = TRUE)
  }
}

.np_condist_slice_bounds <- function(object, slice.context, proper.control) {
  yidx <- object$bws$iycon
  if (length(yidx) == 1L &&
      length(object$bws$cykerlb) >= yidx &&
      length(object$bws$cykerub) >= yidx &&
      is.finite(object$bws$cykerlb[yidx]) &&
      is.finite(object$bws$cykerub[yidx])) {
    bounds <- c(object$bws$cykerlb[yidx], object$bws$cykerub[yidx])
  } else {
    bounds <- grDevices::extendrange(
      c(as.double(slice.context$tydat[[1L]]), as.double(slice.context$eydat[[1L]])),
      f = proper.control$`slice.extend.factor`
    )
  }

  bounds <- as.double(bounds)
  if (length(bounds) != 2L ||
      any(!is.finite(bounds)) ||
      !(bounds[2L] > bounds[1L])) {
    stop("invalid slice master-grid bounds")
  }

  bounds
}

.np_condist_build_slice_master_grid <- function(object, slice.context, proper.control) {
  proper.control <- .np_condist_normalize_proper_control(proper.control)

  y.req <- as.double(slice.context$eydat[[1L]])
  if (!length(y.req) || any(!is.finite(y.req)))
    stop("requested evaluation y values must be finite for slice repair")

  bounds <- .np_condist_slice_bounds(
    object = object,
    slice.context = slice.context,
    proper.control = proper.control
  )

  base.grid <- seq(
    from = bounds[1L],
    to = bounds[2L],
    length.out = proper.control$`slice.grid.size`
  )
  y.grid <- sort(unique(c(base.grid, y.req)))

  if (length(y.grid) < 2L || any(diff(y.grid) <= proper.control$tol))
    stop("slice repair master-grid must contain at least two strictly increasing points")

  y.grid
}

.np_condist_build_slice_eval_grid <- function(object, slice.context, proper.control) {
  xeval <- toFrame(slice.context$exdat)
  eydat <- toFrame(slice.context$eydat)
  groups <- .np_condist_slice_groups(xeval)
  y.grid <- .np_condist_build_slice_master_grid(
    object = object,
    slice.context = slice.context,
    proper.control = proper.control
  )

  x.unique <- xeval[vapply(groups, `[`, integer(1), 1L), , drop = FALSE]
  ny <- length(y.grid)
  ng <- length(groups)

  if (ncol(x.unique) == 0L) {
    ex.grid <- data.frame(row.names = seq_len(ng * ny))
  } else {
    ex.grid <- x.unique[rep(seq_len(ng), each = ny), , drop = FALSE]
    rownames(ex.grid) <- NULL
  }

  ey.grid <- eydat[rep.int(1L, ng * ny), , drop = FALSE]
  ey.grid[[1L]] <- rep(y.grid, times = ng)
  rownames(ey.grid) <- NULL

  list(
    groups = groups,
    y.grid = y.grid,
    exdat = ex.grid,
    eydat = ey.grid,
    grid.slices = split(seq_len(ng * ny), rep(seq_len(ng), each = ny))
  )
}

.np_condist_apply_proper_slice <- function(object,
                                           proper.method = "isotonic",
                                           proper.control = list(),
                                           slice.context = NULL,
                                           grid.out = NULL) {
  proper.method <- match.arg(as.character(proper.method)[1L], c("isotonic"))
  proper.control <- .np_condist_normalize_proper_control(proper.control)

  if (!.np_condist_slice_dispatch_enabled()) {
    info <- .np_condist_make_reason_info(
      reason = "slice_disabled",
      supported = FALSE,
      slice.count = 0L,
      grid.common = FALSE
    )
    return(list(applied = FALSE, reason = "slice_disabled", proper.info = info))
  }

  target.context <- NULL
  target.scope <- NULL

  if (isTRUE(object$trainiseval)) {
    if (!identical(proper.control$apply, "fitted") &&
        !identical(proper.control$apply, "both")) {
      return(grid.out)
    }

    if (is.null(slice.context) ||
        is.null(slice.context$txdat) ||
        is.null(slice.context$tydat)) {
      info <- .np_condist_make_reason_info(
        reason = "slice_context_missing",
        supported = FALSE,
        slice.count = 0L,
        grid.common = FALSE
      )
      return(list(applied = FALSE, reason = "slice_context_missing", proper.info = info))
    }

    target.context <- list(
      txdat = slice.context$txdat,
      tydat = slice.context$tydat,
      exdat = slice.context$txdat,
      eydat = slice.context$tydat
    )
    target.scope <- "fitted"
  } else {
    if (identical(proper.control$apply, "fitted")) {
      info <- .np_condist_make_reason_info(
        reason = "scope_not_selected",
        supported = TRUE,
        slice.count = 0L,
        grid.common = FALSE
      )
      return(list(applied = FALSE, reason = "scope_not_selected", proper.info = info))
    }

    if (is.null(slice.context) ||
        is.null(slice.context$txdat) ||
        is.null(slice.context$tydat) ||
        is.null(slice.context$exdat) ||
        is.null(slice.context$eydat)) {
      info <- .np_condist_make_reason_info(
        reason = "slice_context_missing",
        supported = FALSE,
        slice.count = 0L,
        grid.common = FALSE
      )
      return(list(applied = FALSE, reason = "slice_context_missing", proper.info = info))
    }

    target.context <- slice.context
    target.scope <- "evaluation"
  }

  if (is.null(target.context) ||
      is.null(target.context$txdat) ||
      is.null(target.context$tydat) ||
      is.null(target.context$exdat) ||
      is.null(target.context$eydat)) {
    info <- .np_condist_make_reason_info(
      reason = "slice_context_missing",
      supported = FALSE,
      slice.count = 0L,
      grid.common = FALSE
    )
    return(list(applied = FALSE, reason = "slice_context_missing", proper.info = info))
  }

  grid.eval <- tryCatch(
    .np_condist_build_slice_eval_grid(
      object = object,
      slice.context = target.context,
      proper.control = proper.control
    ),
    error = function(e) e
  )
  if (inherits(grid.eval, "error")) {
    info <- .np_condist_make_reason_info(
      reason = "slice_invalid_master_grid",
      supported = FALSE,
      slice.count = 0L,
      grid.common = FALSE
    )
    return(list(
      applied = FALSE,
      reason = "slice_invalid_master_grid",
      proper.info = info
    ))
  }

  grid.fit <- tryCatch(
    npcdist(
      bws = object$bws,
      txdat = target.context$txdat,
      tydat = target.context$tydat,
      exdat = grid.eval$exdat,
      eydat = grid.eval$eydat,
      proper = FALSE
    ),
    error = function(e) e
  )
  if (inherits(grid.fit, "error")) {
    info <- .np_condist_make_reason_info(
      reason = "slice_eval_failed",
      supported = FALSE,
      slice.count = length(grid.eval$groups),
      grid.common = FALSE
    )
    return(list(applied = FALSE, reason = "slice_eval_failed", proper.info = info))
  }

  grid.proper <- .np_condist_apply_proper_grid(
    object = grid.fit,
    proper.method = proper.method,
    proper.control = proper.control
  )
  if (!isTRUE(grid.proper$applied))
    return(grid.proper)

  y.req <- as.double(target.context$eydat[[1L]])
  ypos <- match(y.req, grid.eval$y.grid)
  if (anyNA(ypos)) {
    info <- .np_condist_make_reason_info(
      reason = "slice_invalid_master_grid",
      supported = FALSE,
      slice.count = length(grid.eval$groups),
      grid.common = FALSE
    )
    return(list(
      applied = FALSE,
      reason = "slice_invalid_master_grid",
      proper.info = info
    ))
  }

  repaired <- numeric(length(y.req))
  for (i in seq_along(grid.eval$groups)) {
    idx.req <- grid.eval$groups[[i]]
    idx.grid <- grid.eval$grid.slices[[i]]
    repaired[idx.req] <- grid.proper$condist[idx.grid[ypos[idx.req]]]
  }

  info <- grid.proper$proper.info
  info$route <- "slice"
  info$apply.scope <- target.scope
  info$internal.grid.size <- length(grid.eval$y.grid)
  info$request.nobs <- nrow(target.context$eydat)

  list(
    applied = TRUE,
    condist = repaired,
    condist.raw = if (isTRUE(proper.control$store.raw)) as.double(object$condist) else NULL,
    proper.info = info
  )
}

## Kernel copula estimation.  The worker preserves the historical data-frame
## return columns while the public wrapper adds formula routing and S3 metadata.

npcopula <- function(bws, ...) {
  if (!missing(bws)) {
    UseMethod("npcopula", bws)
  } else {
    UseMethod("npcopula", NULL)
  }
}

.npcopula_validate_target <- function(target) {
  match.arg(target[1L], c("distribution", "density"))
}

.npcopula_validate_evaluation <- function(evaluation) {
  match.arg(evaluation[1L], c("grid", "sample"))
}

.npcopula_validate_neval <- function(neval) {
  if (!is.numeric(neval) || length(neval) != 1L || is.na(neval) ||
      !is.finite(neval) || neval != floor(neval) || neval < 2L)
    stop("'neval' must be a positive integer greater than one")
  as.integer(neval)
}

.npcopula_validate_n_quasi_inv <- function(n.quasi.inv) {
  if (!is.numeric(n.quasi.inv) || length(n.quasi.inv) != 1L ||
      is.na(n.quasi.inv) || !is.finite(n.quasi.inv) ||
      n.quasi.inv != floor(n.quasi.inv) || n.quasi.inv < 2L)
    stop("'n.quasi.inv' must be an integer greater than one")
  as.integer(n.quasi.inv)
}

.npcopula_validate_er_quasi_inv <- function(er.quasi.inv) {
  if (!is.numeric(er.quasi.inv) || length(er.quasi.inv) != 1L ||
      is.na(er.quasi.inv) || !is.finite(er.quasi.inv) ||
      er.quasi.inv < 0)
    stop("'er.quasi.inv' must be a non-negative finite numeric scalar")
  er.quasi.inv
}

.npcopula_validate_u <- function(u, num.var) {
  if (is.null(u))
    return(NULL)
  if (is.vector(u) && !is.list(u))
    u <- matrix(u, 1L, length(u))
  if (is.data.frame(u)) {
    if (!all(vapply(u, is.numeric, logical(1L))))
      stop("'u' must contain numeric probability values")
  } else if (!is.matrix(u)) {
    stop("'u' must be a numeric vector, matrix, or data frame")
  } else if (!is.numeric(u)) {
    stop("'u' must contain numeric probability values")
  }
  if (ncol(u) != num.var)
    stop("u and bws are incompatible")
  u.check <- as.matrix(u)
  if (anyNA(u.check))
    stop("u must not contain missing values")
  if (any(!is.finite(u.check)))
    stop("u must contain finite probability values")
  if (any(u.check > 1 | u.check < 0))
    stop("u must lie in [0,1]")
  u
}

.npcopula_make_auto_u <- function(xnames, neval) {
  out <- as.data.frame(
    replicate(length(xnames), seq(0, 1, length.out = neval), simplify = FALSE)
  )
  names(out) <- xnames
  out
}

.npcopula_grid_dim <- function(u, num.var) {
  if (is.null(u))
    return(NULL)
  rep.int(nrow(as.data.frame(u)), num.var)
}

.npcopula_progress_begin <- function(target, evaluation, total) {
  state <- .np_progress_begin(
    label = sprintf("Copula %s %s", target, evaluation),
    total = total,
    domain = "general",
    surface = "copula"
  )
  state$start_note_grace_sec <- 0
  .np_progress_show_now(state, done = if (!is.null(total)) 0L else NULL)
}

.npcopula_progress_step <- function(state, done, detail) {
  if (is.null(state))
    return(state)
  .np_progress_step_at(
    state = state,
    now = .np_progress_now(),
    done = done,
    detail = detail,
    force = TRUE
  )
}

.npcopula_progress_total <- function(density, u.provided, num.var) {
  if (!u.provided)
    return(1L + num.var)
  num.var + 2L
}

.npcopula_eval_expanded_grid <- function(bws, data, x.u, density) {
  if (!isTRUE(density))
    return(fitted(npudist(bws = bws, tdat = data, edat = x.u)))

  copula <- fitted(npudens(bws = bws, tdat = data, edat = x.u))
  for (j in seq_len(length(bws$xnames))) {
    bws.f.marginal <- npudensbw(formula(paste("~", bws$xnames[j])),
                                bws = bws$bw[j],
                                bandwidth.compute = FALSE,
                                bwtype = bws$type,
                                ckerorder = bws$ckerorder,
                                ckertype = bws$ckertype,
                                okertype = bws$okertype,
                                ukertype = bws$ukertype,
                                data = data)
    xeval <- data.frame(x.u[, j])
    names(xeval) <- bws$xnames[j]
    copula <- copula / NZD(fitted(npudens(
      bws = bws.f.marginal,
      tdat = data[, bws$xnames[j], drop = FALSE],
      edat = xeval
    )))
  }
  copula
}

.npcopula_object <- function(result,
                             bws,
                             data,
                             target,
                             evaluation,
                             u.provided,
                             u.auto,
                             grid.dim,
                             u.input = NULL,
                             u.grid = NULL,
                             neval,
                             n.quasi.inv,
                             er.quasi.inv,
                             timing = NULL) {
  d <- as.list(result)
  d$bw <- bws$bw
  d$bws <- bws
  d$xnames <- bws$xnames
  d$nobs <- nrow(result)
  d$ndim <- bws$ndim
  d$nord <- bws$nord
  d$nuno <- bws$nuno
  d$ncon <- bws$ncon
  d$pscaling <- bws$pscaling
  d$ptype <- bws$ptype
  d$pckertype <- bws$pckertype
  d$pukertype <- bws$pukertype
  d$pokertype <- bws$pokertype
  d$eval <- result
  d$copulaerr <- NULL
  d$data <- data
  d$target <- target
  d$density <- identical(target, "density")
  d$ntrain <- bws$nobs
  d$trainiseval <- identical(evaluation, "sample")
  d$u.provided <- isTRUE(u.provided)
  d$u.auto <- isTRUE(u.auto)
  d$evaluation <- evaluation
  d$grid.dim <- grid.dim
  d$u.input <- u.input
  d$u.grid <- u.grid
  d$neval <- neval
  d$n.quasi.inv <- n.quasi.inv
  d$er.quasi.inv <- er.quasi.inv
  d$timing <- timing
  class(d) <- "npcopula"
  d
}

.npcopula_marginal_bw <- function(bws, data, j) {
  dat <- data[, bws$xnames[j], drop = FALSE]
  continuous.slot <- if (!is.null(bws$icon)) match(j, bws$icon) else NA_integer_
  marginal <- list(
    bw = bws$bw[j],
    type = bws$type,
    ckerorder = bws$ckerorder,
    ckertype = bws$ckertype,
    ckerbound = bws$ckerbound,
    ckerlb = if (!is.null(bws$ckerlb) && !is.na(continuous.slot)) bws$ckerlb[continuous.slot] else NULL,
    ckerub = if (!is.null(bws$ckerub) && !is.na(continuous.slot)) bws$ckerub[continuous.slot] else NULL,
    ukertype = bws$ukertype,
    okertype = bws$okertype
  )
  .np_make_kbandwidth_unconditional(marginal, dat)
}

.npcopula_grid_eval <- function(x) {
  dat <- as.data.frame(x)
  xnames <- x$xnames
  if (length(xnames) != 2L)
    stop("plot.npcopula currently supports two-dimensional copula displays")
  grid.dim <- x$grid.dim
  if (is.null(grid.dim) || length(grid.dim) != 2L ||
      prod(grid.dim) != nrow(dat))
    stop("npcopula grid output cannot be reshaped into a two-dimensional surface")
  if (!is.null(x$u.grid)) {
    ugrid <- as.data.frame(x$u.grid)
    if (ncol(ugrid) != 2L || nrow(ugrid) != nrow(dat))
      stop("npcopula requested grid metadata is incompatible with the fitted surface")
    u1 <- sort(unique(ugrid[[1L]]))
    u2 <- sort(unique(ugrid[[2L]]))
  } else {
    u1 <- sort(unique(dat$u1))
    u2 <- sort(unique(dat$u2))
  }
  if (length(u1) != grid.dim[1L] || length(u2) != grid.dim[2L])
    stop("npcopula grid output is not rectangular")
  xgrid <- as.data.frame(dat[, xnames, drop = FALSE])
  list(
    u1 = u1,
    u2 = u2,
    xgrid = xgrid,
    z = matrix(dat$copula, nrow = length(u1), ncol = length(u2))
  )
}

.npcopula_empirical_eval <- function(x) {
  dat <- as.data.frame(x)
  if (identical(x$evaluation, "sample") &&
      all(c("u1", "u2") %in% names(dat))) {
    return(dat[, c("u1", "u2"), drop = FALSE])
  }
  data <- .npcopula_training_data(x)
  bws <- x$bws
  out <- vector("list", length(bws$xnames))
  for (j in seq_along(bws$xnames)) {
    mbw <- .npcopula_marginal_bw(bws, data, j)
    xdat <- data[, bws$xnames[j], drop = FALSE]
    out[[j]] <- .np_ksum_unconditional_eval_exact(
      xdat = xdat,
      exdat = xdat,
      bws = mbw,
      operator = "integral"
    )
  }
  out <- as.data.frame(out)
  names(out) <- paste0("u", seq_along(out))
  out
}

.npcopula_default_plot_main <- function(target.label, view) {
  switch(view,
         contour = paste(target.label, "Contour"),
         image = if (identical(target.label, "Copula")) {
           "Contour Heatmap"
         } else {
           paste(target.label, "Heatmap")
         },
         empirical = "Empirical Copula",
         target.label)
}

.npcopula_validate_zlim <- function(zlim) {
  if (is.null(zlim))
    return(NULL)
  if (!is.numeric(zlim) || length(zlim) != 2L ||
      anyNA(zlim) || any(!is.finite(zlim)) || zlim[1L] >= zlim[2L])
    stop("zlim must be a finite numeric vector of length two with zlim[1] < zlim[2]",
         call. = FALSE)
  zlim
}

.npcopula_clip_z_for_display <- function(z, zlim) {
  if (is.null(z))
    return(NULL)
  if (is.null(zlim))
    return(z)
  pmin(pmax(z, zlim[1L]), zlim[2L])
}

.npcopula_clip_z_list_for_display <- function(zlist, zlim) {
  if (is.null(zlist))
    return(NULL)
  lapply(zlist, .npcopula_clip_z_for_display, zlim = zlim)
}

.npcopula_density_companion <- function(x) {
  if (isTRUE(x$density))
    return(x)
  data <- .npcopula_training_data(x)
  f <- stats::as.formula(paste("~", paste(x$xnames, collapse = "+")))
  bws <- x$bws
  bw.args <- list(
    formula = f,
    data = data,
    bws = bws$bw,
    bandwidth.compute = FALSE,
    bwtype = bws$type,
    ckerorder = bws$ckerorder,
    ckertype = bws$ckertype,
    okertype = bws$okertype,
    ukertype = bws$ukertype
  )
  if (!is.null(bws$ckerbound))
    bw.args$ckerbound <- bws$ckerbound
  if (!is.null(bws$ckerlb))
    bw.args$ckerlb <- bws$ckerlb
  if (!is.null(bws$ckerub))
    bw.args$ckerub <- bws$ckerub
  dbw <- do.call(npudensbw, bw.args)
  npcopula.default(
    bws = dbw,
    data = data,
    u = x$u.input,
    target = "density",
    evaluation = x$evaluation,
    neval = x$neval,
    n.quasi.inv = x$n.quasi.inv,
    er.quasi.inv = x$er.quasi.inv,
    u.auto = isTRUE(x$u.auto)
  )
}

.npcopula_draw_surface_base <- function(grid,
                                        zlab,
                                        main,
                                        theta,
                                        phi,
                                        col,
                                        border,
                                        zlim,
                                        xlab,
                                        ylab,
                                        dots) {
  persp.args <- .np_plot_user_args(dots, "persp")
  z <- .npcopula_clip_z_for_display(grid$z, zlim)
  persp.col <- grDevices::adjustcolor(
    .np_plot_persp_surface_colors(z = z, col = col),
    alpha.f = 0.5
  )
  persp.call <- list(x = grid$u1, y = grid$u2, z = z,
                     theta = theta, phi = phi,
                     ticktype = "detailed",
                     xlab = xlab, ylab = ylab, zlab = zlab,
                     main = main, col = persp.col, border = border)
  if (!is.null(zlim))
    persp.call$zlim <- zlim
  do.call(graphics::persp, .np_plot_merge_user_args(persp.call, persp.args))
}

.npcopula_plot_all <- function(x,
                               xlab,
                               ylab,
                               theta,
                               phi,
                               col,
                               border,
                               zlim,
                               dots,
                               draw = TRUE) {
  if (!identical(x$evaluation, "grid"))
    stop("view='all' requires grid evaluation output", call. = FALSE)
  grid <- .npcopula_grid_eval(x)
  empirical <- .npcopula_empirical_eval(x)
  density.obj <- .npcopula_density_companion(x)
  density.grid <- .npcopula_grid_eval(density.obj)
  out <- list(
    copula = as.data.frame(x),
    empirical = empirical,
    density = as.data.frame(density.obj)
  )
  if (!isTRUE(draw))
    return(out)
  old.par <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old.par), add = TRUE)
  graphics::par(mfrow = c(2L, 2L))
  do.call(graphics::contour, c(list(x = grid$u1, y = grid$u2, z = grid$z,
                                    xlab = xlab, ylab = ylab,
                                    main = "Copula Contour"),
                               dots))
  .npcopula_draw_surface_base(
    grid = grid,
    zlab = "Copula",
    main = "Copula",
    theta = theta,
    phi = phi,
    col = col,
    border = border,
    zlim = if (is.null(zlim)) c(0, 1) else zlim,
    xlab = xlab,
    ylab = ylab,
    dots = dots
  )
  do.call(graphics::plot, c(list(x = empirical$u1, y = empirical$u2,
                                 xlab = xlab, ylab = ylab, cex = 0.25,
                                 main = "Empirical Copula"),
                            dots))
  .npcopula_draw_surface_base(
    grid = density.grid,
    zlab = "Copula Density",
    main = "Copula Density",
    theta = theta,
    phi = phi,
    col = col,
    border = border,
    zlim = NULL,
    xlab = xlab,
    ylab = ylab,
    dots = dots
  )
  invisible(out)
}

.npcopula_training_data <- function(x) {
  data <- x$data
  if (is.null(data))
    stop("npcopula object does not retain the training data needed for intervals; refit with npcopula()")
  as.data.frame(data)
}

.npcopula_marginal_density_product <- function(bws, data, xgrid) {
  out <- rep.int(1.0, nrow(xgrid))
  for (j in seq_along(bws$xnames)) {
    mbw <- .npcopula_marginal_bw(bws, data, j)
    xeval <- xgrid[, bws$xnames[j], drop = FALSE]
    out <- out * .np_ksum_unconditional_eval_exact(
      xdat = data[, bws$xnames[j], drop = FALSE],
      exdat = xeval,
      bws = mbw,
      operator = "normal"
    )
  }
  NZD(out)
}

.npcopula_asymptotic_se <- function(x, data, xgrid) {
  bws <- x$bws
  target <- x$target
  density <- identical(target, "density")
  if (density) {
    joint <- npudens(tdat = data, edat = xgrid, bws = bws)
    return(se(joint) / .npcopula_marginal_density_product(bws, data, xgrid))
  }
  joint <- npudist(tdat = data, edat = xgrid, bws = bws)
  se(joint)
}

.npcopula_eval_xgrid <- function(x) {
  dat <- as.data.frame(x)
  xnames <- x$xnames
  if (length(xnames) && all(xnames %in% names(dat)))
    return(as.data.frame(dat[, xnames, drop = FALSE]))
  data <- .npcopula_training_data(x)
  if (nrow(data) != length(x$copula))
    stop("npcopula object does not retain evaluation coordinates needed for standard errors")
  as.data.frame(data[, xnames, drop = FALSE])
}

.npcopula_boot_counts <- function(n, B, method, blocklen) {
  if (identical(method, "inid"))
    return(.np_inid_counts_matrix(n = n, B = B))
  drawer <- .np_block_counts_drawer(
    n = n,
    B = B,
    blocklen = blocklen,
    sim = method
  )
  .np_inid_counts_matrix(n = n, B = B, counts = drawer(1L, B))
}

.npcopula_boot_eval_weighted <- function(xdat, exdat, bws, operator, counts.col) {
  active <- .np_active_boot_sample(xdat = xdat, counts.col = counts.col)
  .np_ksum_unconditional_eval_exact(
    xdat = active$xdat,
    exdat = exdat,
    bws = bws,
    operator = operator,
    weights = active$weights,
    n.total = active$n.total
  )
}

.npcopula_bootstrap_values <- function(x,
                                       data,
                                       xgrid,
                                       B,
                                       method,
                                       blocklen) {
  bws <- x$bws
  target <- x$target
  density <- identical(target, "density")
  method <- match.arg(method, c("inid", "fixed", "geom"))
  B <- as.integer(B)
  if (B < 1L)
    stop("B must be a positive integer")
  if (!identical(method, "inid") &&
      (is.null(blocklen) || length(blocklen) != 1L || is.na(blocklen)))
    blocklen <- max(1L, floor(nrow(data)^(1/3)))
  marginal.bws <- if (density) {
    lapply(seq_along(bws$xnames), function(j) {
      .npcopula_marginal_bw(bws, data, j)
    })
  } else {
    NULL
  }

  counts <- .npcopula_boot_counts(
    n = nrow(data),
    B = B,
    method = method,
    blocklen = blocklen
  )
  tmat <- matrix(NA_real_, nrow = B, ncol = nrow(xgrid))
  progress <- .np_plot_bootstrap_progress_begin(
    total = B,
    label = if (identical(method, "inid")) "Plot bootstrap inid" else "Plot bootstrap block"
  )
  on.exit(.np_plot_progress_end(progress), add = TRUE)

  start <- 1L
  chunk.size <- .np_inid_chunk_size(n = nrow(data), B = B, progress_cap = TRUE)
  chunk.controller <- .np_plot_progress_chunk_controller(chunk.size = chunk.size, progress = progress)
  while (start <= B) {
    stopi <- min(B, start + chunk.controller$chunk.size - 1L)
    chunk.started <- .np_progress_now()
    for (bb in start:stopi) {
      joint <- .npcopula_boot_eval_weighted(
        xdat = data,
        exdat = xgrid,
        bws = bws,
        operator = if (density) "normal" else "integral",
        counts.col = counts[, bb]
      )
      if (density) {
        denom <- rep.int(1.0, nrow(xgrid))
        for (j in seq_along(bws$xnames)) {
          denom <- denom * .npcopula_boot_eval_weighted(
            xdat = data[, bws$xnames[j], drop = FALSE],
            exdat = xgrid[, bws$xnames[j], drop = FALSE],
            bws = marginal.bws[[j]],
            operator = "normal",
            counts.col = counts[, bb]
          )
        }
        joint <- joint / NZD(denom)
      }
      tmat[bb, ] <- joint
    }
    progress <- .np_plot_progress_tick(state = progress, done = stopi)
    chunk.controller <- .np_plot_progress_chunk_observe(
      controller = chunk.controller,
      bsz = stopi - start + 1L,
      elapsed.sec = .np_progress_now() - chunk.started
    )
    start <- stopi + 1L
  }

  list(t = tmat, t0 = x$copula)
}

.npcopula_interval_payload <- function(x,
                                       plot.errors.method,
                                       plot.errors.alpha,
                                       plot.errors.type,
                                       plot.errors.center,
                                       plot.errors.boot.method,
                                       plot.errors.boot.num,
                                       plot.errors.boot.blocklen) {
  if (identical(plot.errors.method, "none"))
    return(NULL)
  if (!identical(x$evaluation, "grid"))
    stop("npcopula intervals are available only for grid evaluation output")

  data <- .npcopula_training_data(x)
  grid <- .npcopula_grid_eval(x)
  se <- if (identical(plot.errors.method, "asymptotic")) {
    .npcopula_asymptotic_se(x, data = data, xgrid = grid$xgrid)
  } else {
    rep(NA_real_, nrow(grid$xgrid))
  }
  boot.raw <- if (identical(plot.errors.method, "bootstrap")) {
    boot.out <- .npcopula_bootstrap_values(
      x = x,
      data = data,
      xgrid = grid$xgrid,
      B = plot.errors.boot.num,
      method = plot.errors.boot.method,
      blocklen = plot.errors.boot.blocklen
    )
    if (identical(plot.errors.type, "pmzsd")) {
      boot.err <- matrix(NA_real_, nrow = nrow(grid$xgrid), ncol = 3L)
      boot.sd <- .np_plot_bootstrap_col_sds(boot.out$t)
      boot.err[, 1:2] <- qnorm(plot.errors.alpha / 2, lower.tail = FALSE) * boot.sd
      boot.all.err <- NULL
    } else {
      interval.summary <- .np_plot_bootstrap_interval_summary(
        boot.t = boot.out$t,
        t0 = boot.out$t0,
        alpha = plot.errors.alpha,
        band.type = plot.errors.type
      )
      boot.err <- matrix(NA_real_, nrow = nrow(grid$xgrid), ncol = 3L)
      boot.err[, 1:2] <- interval.summary$err
      boot.all.err <- interval.summary$all.err
    }
    if (identical(plot.errors.center, "bias-corrected"))
      boot.err[, 3L] <- 2 * boot.out$t0 - colMeans(boot.out$t)
    list(boot.err = boot.err, boot.all.err = boot.all.err, bxp = list())
  } else {
    NULL
  }
  .np_plot_interval_payload(
    estimate = x$copula,
    se = se,
    plot.errors.method = plot.errors.method,
    plot.errors.alpha = plot.errors.alpha,
    plot.errors.type = plot.errors.type,
    plot.errors.center = plot.errors.center,
    bootstrap_raw = boot.raw
  )
}

.npcopula_add_interval_columns <- function(x, payload) {
  x <- as.data.frame(x)
  if (is.null(payload))
    return(x)
  x$center <- payload$center
  x$lower <- payload$center - payload$err[, 1L]
  x$upper <- payload$center + payload$err[, 2L]
  if (!is.null(payload$all.err)) {
    for (nm in names(payload$all.err)) {
      x[[paste0(nm, ".lower")]] <- payload$center - payload$all.err[[nm]][, 1L]
      x[[paste0(nm, ".upper")]] <- payload$center + payload$all.err[[nm]][, 2L]
    }
  }
  x
}

print.npcopula <- function(x, ...) {
  target <- x$target
  evaluation <- x$evaluation
  ntrain <- x$ntrain
  xnames <- x$xnames
  grid.dim <- x$grid.dim

  cat("\nKernel Copula ", if (identical(target, "density")) "Density" else "Distribution",
      ": ", ntrain, " training points, in ", length(xnames),
      " variable(s)\n", sep = "")
  cat("Evaluation: ", evaluation, sep = "")
  if (!is.null(grid.dim))
    cat(" (grid ", paste(grid.dim, collapse = " x "), ")", sep = "")
  cat("\n\n")
  print(utils::head(as.data.frame(x), ...))
  invisible(x)
}

summary.npcopula <- function(object, ...) {
  print(object)
  cat("\nBandwidth summary:\n")
  summary(object$bws)
  cat("\nCopula value summary:\n")
  print(summary(object$copula))
  invisible(object)
}

fitted.npcopula <- function(object, ...) {
  object$copula
}

.npcopula_predict_newdata_to_u <- function(object, newdata) {
  if (is.null(newdata))
    return(NULL)
  if (is.vector(newdata) && !is.list(newdata))
    return(matrix(newdata, nrow = 1L))

  nd <- as.data.frame(newdata)
  xnames <- object$xnames
  unames <- paste0("u", seq_along(xnames))
  if (length(xnames) && all(xnames %in% names(nd)))
    return(nd[, xnames, drop = FALSE])
  if (length(unames) && all(unames %in% names(nd))) {
    out <- nd[, unames, drop = FALSE]
    names(out) <- xnames
    return(out)
  }
  nd
}

predict.npcopula <- function(object,
                             newdata = NULL,
                             u = NULL,
                             se.fit = FALSE,
                             output = c("vector", "object", "data"),
                             ...) {
  se.fit <- npValidateScalarLogical(se.fit, "se.fit")
  output <- match.arg(output)
  dots <- list(...)

  if (is.null(u) && !is.null(newdata)) {
    u <- .npcopula_predict_newdata_to_u(object, newdata)
  } else if (!is.null(u)) {
    u <- .npcopula_predict_newdata_to_u(object, u)
  }

  if (is.null(u) && !length(dots)) {
    ev <- object
  } else {
    data <- dots$data
    dots$data <- NULL
    if (is.null(data))
      data <- object$data
    if (is.null(data))
      stop("npcopula object does not retain the training data needed for prediction; refit with npcopula()")
    args <- c(list(bws = object$bws, data = data),
              if (is.null(u)) list() else list(u = u),
              dots)
    ev <- do.call(npcopula, args)
  }

  if (identical(output, "object"))
    return(ev)
  if (identical(output, "data"))
    return(as.data.frame(ev))
  if (se.fit)
    return(list(fit = fitted(ev), se.fit = se(ev), df = ev$ntrain))
  fitted(ev)
}

se.npcopula <- function(x) {
  if (!is.null(x$copulaerr) && length(x$copulaerr) == length(x$copula))
    return(x$copulaerr)
  .npcopula_asymptotic_se(
    x = x,
    data = .npcopula_training_data(x),
    xgrid = .npcopula_eval_xgrid(x)
  )
}

as.data.frame.npcopula <- function(x, row.names = NULL, optional = FALSE, ...) {
  out <- x$eval
  if (is.null(row.names)) {
    row.names <- .set_row_names(nrow(out))
  }
  attr(out, "row.names") <- row.names
  out
}

plot.npcopula <- function(x,
                          perspective = TRUE,
                          view = c("rotate", "fixed", "surface", "contour", "image",
                                   "empirical", "all"),
                          renderer = c("base", "rgl"),
                          errors = c("none", "bootstrap", "asymptotic"),
                          band = c("pointwise", "pmzsd", "bonferroni",
                                   "simultaneous", "all"),
                          alpha = 0.05,
                          bootstrap = c("inid", "fixed", "geom"),
                          B = 1999,
                          center = c("estimate", "bias-corrected"),
                          boot_control = np_boot_control(),
                          output = c("plot", "data", "plot-data", "both"),
                          legend = TRUE,
                          theta = 0.0,
                          phi = 20.0,
                          xlab = "u1",
                          ylab = "u2",
                          zlab = NULL,
                          main = NULL,
                          col = NULL,
                          border = "black",
                          zlim = NULL,
                          ...) {
  bootstrap.supplied <- !missing(bootstrap) || !missing(B) || !missing(center)
  interval.supplied <- !missing(band) || !missing(alpha)
  view <- match.arg(view)
  surface.view <- view %in% c("rotate", "fixed", "surface")
  rotate.surface <- identical(view, "rotate")
  renderer <- .np_plot_match_renderer(renderer)
  errors <- match.arg(errors)
  band <- match.arg(band)
  bootstrap <- match.arg(bootstrap)
  center <- match.arg(center)
  output <- match.arg(output)
  if (identical(output, "both"))
    output <- "plot-data"
  if (!is.numeric(alpha) || length(alpha) != 1L ||
      is.na(alpha) || alpha <= 0 || alpha >= 0.5)
    stop("alpha must lie in (0, 0.5)", call. = FALSE)
  if (!is.numeric(B) || length(B) != 1L || is.na(B) || B < 1L)
    stop("B must be a positive numeric scalar", call. = FALSE)
  zlim <- .npcopula_validate_zlim(zlim)
  if (!inherits(boot_control, "np_boot_control"))
    stop("boot_control must be created by np_boot_control()", call. = FALSE)
  if (!identical(errors, "bootstrap") && bootstrap.supplied)
    stop("bootstrap controls require errors = \"bootstrap\"", call. = FALSE)
  if (identical(errors, "none") && interval.supplied)
    stop("band and alpha require errors != \"none\"", call. = FALSE)
  if (identical(errors, "bootstrap"))
    .np_plot_reject_wild_unsupervised(bootstrap, "copula estimators")
  dots <- list(...)
  target <- x$target
  target.label <- if (identical(target, "density")) "Copula Density" else "Copula"
  main.supplied <- !missing(main) && !is.null(main)
  if (is.null(zlab))
    zlab <- target.label
  if (is.null(main))
    main <- .npcopula_default_plot_main(target.label, view)

  xnames <- x$xnames
  if (length(xnames) != 2L)
    stop("plot.npcopula currently supports two-dimensional copula displays")

  if (identical(view, "all")) {
    if (isTRUE(x$density))
      stop("view='all' requires a copula distribution object",
           call. = FALSE)
    if (!identical(errors, "none"))
      stop("view='all' does not currently support interval overlays",
           call. = FALSE)
    if (identical(renderer, "rgl"))
      stop("view='all' is currently supported only with renderer='base'",
           call. = FALSE)
    all.data <- .npcopula_plot_all(
      x = x,
      xlab = xlab,
      ylab = ylab,
      theta = theta,
      phi = phi,
      col = col,
      border = border,
      zlim = zlim,
      dots = dots,
      draw = !identical(output, "data")
    )
    if (identical(output, "data") || identical(output, "plot-data"))
      return(all.data)
    return(invisible(x))
  }

  if (identical(view, "empirical")) {
    if (!identical(errors, "none"))
      stop("view='empirical' does not support interval overlays",
           call. = FALSE)
    if (identical(renderer, "rgl"))
      stop("renderer='rgl' is supported only for grid surface displays in plot.npcopula",
           call. = FALSE)
    empirical <- .npcopula_empirical_eval(x)
    if (identical(output, "data"))
      return(empirical)
    do.call(graphics::plot, c(list(x = empirical$u1, y = empirical$u2,
                                   xlab = xlab, ylab = ylab, cex = 0.25,
                                   main = main),
                              dots))
    if (identical(output, "plot-data"))
      return(empirical)
    return(invisible(x))
  }

  evaluation <- x$evaluation
  if (identical(evaluation, "sample")) {
    if (!identical(errors, "none"))
      stop("npcopula intervals are available only for grid evaluation output",
           call. = FALSE)
    if (identical(renderer, "rgl"))
      stop("renderer='rgl' is supported only for grid surface displays in plot.npcopula",
           call. = FALSE)
    if (identical(output, "data"))
      return(as.data.frame(x))
    do.call(graphics::plot, c(list(x = x$u1, y = x$u2,
                                   xlab = xlab, ylab = ylab, main = main),
                              dots))
    if (identical(output, "plot-data"))
      return(as.data.frame(x))
    return(invisible(x))
  }

  if (!identical(errors, "none") &&
      !(isTRUE(perspective) && surface.view) &&
      !identical(output, "data"))
    stop("npcopula interval plotting requires a surface view and perspective=TRUE",
         call. = FALSE)

  grid <- .npcopula_grid_eval(x)
  u1 <- grid$u1
  u2 <- grid$u2
  z <- grid$z
  z.display <- .npcopula_clip_z_for_display(z, zlim)
  payload <- .npcopula_interval_payload(
    x = x,
    plot.errors.method = errors,
    plot.errors.alpha = alpha,
    plot.errors.type = band,
    plot.errors.center = center,
    plot.errors.boot.method = bootstrap,
    plot.errors.boot.num = as.integer(B),
    plot.errors.boot.blocklen = boot_control$blocklen
  )
  plot.data <- .npcopula_add_interval_columns(x, payload)
  if (identical(output, "data"))
    return(plot.data)

  lerr <- herr <- lerr.all <- herr.all <- NULL
  if (!is.null(payload)) {
    lerr <- matrix(payload$center - payload$err[, 1L],
                   nrow = length(u1), ncol = length(u2))
    herr <- matrix(payload$center + payload$err[, 2L],
                   nrow = length(u1), ncol = length(u2))
    if (identical(band, "all") && !is.null(payload$all.err)) {
      lerr.all <- lapply(payload$all.err, function(te)
        matrix(payload$center - te[, 1L], nrow = length(u1), ncol = length(u2)))
      herr.all <- lapply(payload$all.err, function(te)
        matrix(payload$center + te[, 2L], nrow = length(u1), ncol = length(u2)))
    }
  }
  if (is.null(zlim)) {
    zlim.values <- c(z, lerr, herr, unlist(lerr.all, use.names = FALSE),
                     unlist(herr.all, use.names = FALSE))
    zlim <- range(zlim.values, finite = TRUE)
    z.display <- z
  } else {
    lerr <- .npcopula_clip_z_for_display(lerr, zlim)
    herr <- .npcopula_clip_z_for_display(herr, zlim)
    lerr.all <- .npcopula_clip_z_list_for_display(lerr.all, zlim)
    herr.all <- .npcopula_clip_z_list_for_display(herr.all, zlim)
  }

  if (isTRUE(perspective) && surface.view) {
    renderer <- .np_plot_validate_renderer_request(
      renderer = renderer,
      route = "plot.npcopula",
      perspective = perspective,
      supported.route = TRUE,
      view = if (identical(renderer, "rgl")) {
        "fixed"
      } else if (rotate.surface) {
        "rotate"
      } else {
        "fixed"
      }
    )
    if (identical(renderer, "rgl")) {
      rgl.view <- .np_plot_rgl_view_angles(theta = theta, phi = phi)
      rgl.out <- .np_plot_render_surface_rgl(
        x = u1,
        y = u2,
        z = z.display,
        xlab = xlab,
        ylab = ylab,
        zlab = zlab,
        main = main,
        theta = rgl.view$theta,
        phi = rgl.view$phi,
        col = col,
        border = border,
        zlim = zlim,
        par3d.args = .np_plot_collect_rgl_args(dots, "rgl.par3d", "rgl.par3d."),
        view3d.args = .np_plot_collect_rgl_args(dots, "rgl.view3d", "rgl.view3d."),
        persp3d.args = .np_plot_collect_rgl_args(dots, "rgl.persp3d", "rgl.persp3d."),
        grid3d.args = .np_plot_collect_rgl_args(dots, "rgl.grid3d", "rgl.grid3d."),
        widget.args = .np_plot_collect_rgl_args(dots, "rgl.widget", "rgl.widget."),
        draw.extras = function() {
          if (!is.null(payload)) {
            .np_plot_error_surfaces_rgl(
              x = u1,
              y = u2,
              plot.errors.type = band,
              lerr = lerr,
              herr = herr,
              lerr.all = lerr.all,
              herr.all = herr.all,
              surface3d.args = .np_plot_collect_rgl_args(dots, "rgl.persp3d", "rgl.persp3d."),
              legend3d.args = .np_plot_merge_rgl_legend_control(
                .np_plot_collect_rgl_args(dots, "rgl.legend3d", "rgl.legend3d."),
                legend = legend
              )
            )
          }
        }
      )
      return(.np_plot_rgl_finalize(
        rgl.out = rgl.out,
        plot.behavior = output,
        plot.data = plot.data
      ))
    }
    persp.args <- .np_plot_user_args(dots, "persp")
    persp.col <- grDevices::adjustcolor(
      .np_plot_persp_surface_colors(z = z.display, col = col),
      alpha.f = 0.5
    )
    rotate.defaults <- .np_plot_rotate_defaults()
    dtheta <- rotate.defaults$dtheta
    frame.theta <- if (rotate.surface) {
      (0:((360 %/% dtheta - 1L))) * dtheta + theta
    } else {
      theta
    }
    rotation.progress <- .np_plot_rotation_progress_begin(length(frame.theta))
    on.exit(.np_plot_rotation_progress_end(rotation.progress), add = TRUE)

    for (frame.idx in seq_along(frame.theta)) {
      i <- frame.theta[[frame.idx]]
      persp.call <- list(x = u1, y = u2, z = z.display,
                         theta = i, phi = phi,
                         ticktype = "detailed",
                         xlab = xlab, ylab = ylab, zlab = zlab,
                         main = main, col = persp.col, border = border)
      if (!is.null(zlim))
        persp.call$zlim <- zlim
      persp.mat <- do.call(graphics::persp, .np_plot_merge_user_args(persp.call, persp.args))
      if (!is.null(payload)) {
        .np_plot_draw_error_wireframes_persp(
          x = u1,
          y = u2,
          persp.mat = persp.mat,
          plot.errors.type = band,
          lerr = lerr,
          herr = herr,
          lerr.all = lerr.all,
          herr.all = herr.all,
          border = "grey40",
          lwd = .np_plot_scalar_default(dots$lwd, par()$lwd)
        )
        if (identical(band, "all") && !is.null(lerr.all) && !is.null(herr.all) &&
            isTRUE(legend)) {
          .np_plot_draw_all_band_legend(legend = TRUE, x = "topright")
        }
      }
      rotation.progress <- .np_plot_rotation_progress_tick(rotation.progress, done = frame.idx)
      Sys.sleep(if (rotate.surface) rotate.defaults$sleep else 0.5)
    }
  } else if (identical(view, "contour")) {
    if (identical(renderer, "rgl"))
      stop("renderer='rgl' is supported only for surface views in plot.npcopula",
           call. = FALSE)
    do.call(graphics::contour, c(list(x = u1, y = u2, z = z,
                                      xlab = xlab, ylab = ylab, main = main),
                                 dots))
  } else {
    if (identical(renderer, "rgl"))
      stop("renderer='rgl' is supported only for surface views in plot.npcopula",
           call. = FALSE)
    image.col <- if (is.null(col)) {
      grDevices::hcl.colors(100L, palette = "viridis")
    } else {
      col
    }
    do.call(graphics::image, c(list(x = u1, y = u2, z = z,
                                    xlab = xlab, ylab = ylab, main = main,
                                    col = image.col),
                               dots))
  }
  if (identical(output, "plot-data"))
    return(plot.data)
  invisible(x)
}

npcopula.formula <- function(bws,
                             data = NULL,
                             u = NULL,
                             target = c("distribution", "density"),
                             evaluation = c("grid", "sample"),
                             neval = 30,
                             n.quasi.inv = 1000,
                             er.quasi.inv = 1,
                             ...) {
  target <- .npcopula_validate_target(target)
  evaluation <- .npcopula_validate_evaluation(evaluation)
  neval <- .npcopula_validate_neval(neval)

  mf <- stats::model.frame(bws, data = data)
  dat <- mf[, attr(attr(mf, "terms"), "term.labels"), drop = FALSE]
  num.var <- ncol(dat)

  u.auto <- FALSE
  if (is.null(u) && identical(evaluation, "grid")) {
    if (num.var != 2L)
      stop("automatic copula probability grids are supported only for two variables; supply 'u' or use evaluation='sample'")
    u <- .npcopula_make_auto_u(names(dat), neval)
    u.auto <- TRUE
  }

  bw.fun <- if (identical(target, "density")) npudensbw else npudistbw
  bw.args <- c(
    list(formula = bws),
    if (is.null(data)) list() else list(data = data),
    list(...)
  )
  bw <- .np_progress_select_bandwidth_enhanced(
    if (identical(target, "density")) "Selecting copula density bandwidth" else "Selecting copula distribution bandwidth",
    do.call(bw.fun, bw.args)
  )

  npcopula.default(
    bws = bw,
    data = dat,
    u = u,
    target = target,
    evaluation = if (is.null(u)) "sample" else "grid",
    neval = neval,
    n.quasi.inv = n.quasi.inv,
    er.quasi.inv = er.quasi.inv,
    u.auto = u.auto
  )
}

npcopula.default <- function(bws,
                             data,
                             u = NULL,
                             target = NULL,
                             evaluation = c("sample", "grid"),
                             neval = 30,
                             n.quasi.inv = 1000,
                             er.quasi.inv = 1,
                             ...) {
  dots <- list(...)
  u.auto <- isTRUE(dots$u.auto)
  dots$u.auto <- NULL
  formula.from.target <- !missing(target) && inherits(target, "formula")
  formula.from.u <- !missing(u) && inherits(u, "formula")
  formula.from.evaluation <- !missing(evaluation) && inherits(evaluation, "formula")
  formula.arg <- if (formula.from.target) {
    target
  } else if (formula.from.u) {
    u
  } else if (formula.from.evaluation) {
    evaluation
  } else if (length(dots) && inherits(dots[[1L]], "formula")) {
    dots[[1L]]
  } else {
    NULL
  }
  if (!is.null(formula.arg) &&
      !isa(bws, "dbandwidth") && !isa(bws, "bandwidth")) {
    if (length(dots) && inherits(dots[[1L]], "formula"))
      dots[[1L]] <- NULL
    target.formula <- .npcopula_validate_target(
      if (!missing(target) && !is.null(target) && !formula.from.target)
        target
      else
        c("distribution", "density")
    )
    evaluation.formula <- .npcopula_validate_evaluation(
      if (!missing(evaluation) && !formula.from.evaluation)
        evaluation
      else
        c("grid", "sample")
    )
    neval.formula <- .npcopula_validate_neval(if (!missing(neval)) neval else 30)
    mf <- if (missing(data)) {
      stats::model.frame(formula.arg)
    } else {
      stats::model.frame(formula.arg, data = data)
    }
    dat <- mf[, attr(attr(mf, "terms"), "term.labels"), drop = FALSE]
    u.formula <- if (formula.from.u || missing(u)) NULL else u
    auto.grid <- FALSE
    if (is.null(u.formula) && identical(evaluation.formula, "grid")) {
      if (ncol(dat) != 2L)
        stop("automatic copula probability grids are supported only for two variables; supply 'u' or use evaluation='sample'")
      u.formula <- .npcopula_make_auto_u(names(dat), neval.formula)
      auto.grid <- TRUE
    }
    bw.fun <- if (identical(target.formula, "density")) npudensbw else npudistbw
    bw.args <- c(
      list(formula = formula.arg),
      if (missing(data)) list() else list(data = data),
      list(bws = bws),
      dots
    )
    bw <- .np_progress_select_bandwidth_enhanced(
      if (identical(target.formula, "density")) "Selecting copula density bandwidth" else "Selecting copula distribution bandwidth",
      do.call(bw.fun, bw.args)
    )
    args <- c(
      list(bws = bw,
           data = dat,
           u = u.formula,
           target = target.formula,
           evaluation = if (is.null(u.formula)) "sample" else "grid",
           neval = neval.formula,
           n.quasi.inv = n.quasi.inv,
           er.quasi.inv = er.quasi.inv,
           u.auto = auto.grid)
    )
    return(do.call(npcopula.default, args))
  }

  evaluation <- .npcopula_validate_evaluation(evaluation)
  neval <- .npcopula_validate_neval(neval)
  start.time <- proc.time()[3]

  if(missing(data)) stop("You must provide a data frame")
  if(!is.data.frame(data)) stop("Object `data' must be a data frame")
  if(missing(bws)) stop("You must provide a bandwidth object")
  if(!isa(bws,"dbandwidth") && !isa(bws,"bandwidth"))
    stop("you must provide a density (npudensbw) or distribution (npudistbw) object")

  density <- isa(bws, "bandwidth")
  inferred.target <- if (density) "density" else "distribution"
  if (!is.null(target)) {
    target <- .npcopula_validate_target(target)
    if (!identical(target, inferred.target))
      stop("'target' conflicts with the supplied bandwidth object")
  } else {
    target <- inferred.target
  }

  result <- .npcopula_eval(
    bws = bws,
    data = data,
    u = u,
    target = target,
    evaluation = evaluation,
    neval = neval,
    n.quasi.inv = n.quasi.inv,
    er.quasi.inv = er.quasi.inv,
    u.auto = u.auto
  )
  result$timing <- proc.time()[3] - start.time
  result
}

.npcopula_eval <- function(bws,
                           data,
                           u = NULL,
                           target,
                           evaluation,
                           neval,
                           n.quasi.inv,
                           er.quasi.inv,
                           u.auto = FALSE) {
  density <- identical(target, "density")
  num.var <- length(bws$xnames)
  u <- .npcopula_validate_u(u, num.var)
  n.quasi.inv <- .npcopula_validate_n_quasi_inv(n.quasi.inv)
  er.quasi.inv <- .npcopula_validate_er_quasi_inv(er.quasi.inv)
  if(any(is.na(data))) stop("NA values present in data - recompute bw object with na.omit on data")
  if(bws$nuno>0) stop("unordered factors not suitable for copula estimation")

  u.provided <- !is.null(u)
  grid.dim <- .npcopula_grid_dim(u, num.var)
  u.input <- if (u.provided) as.data.frame(u) else NULL
  u.requested <- u

  progress.total <- .npcopula_progress_total(density = density,
                                             u.provided = u.provided,
                                             num.var = num.var)
  progress <- .npcopula_progress_begin(target = target,
                                       evaluation = if (u.provided) "grid" else "sample",
                                       total = progress.total)
  on.exit(.np_progress_end(progress), add = TRUE)
  stage <- 0L

  if(is.null(u)) {
    stage <- stage + 1L
    progress <- .npcopula_progress_step(
      progress, stage,
      if(!density) "joint distribution at sample realizations" else "joint density at sample realizations"
    )
    if(!density) {
      copula <- fitted(npudist(bws=bws,data=data))
    } else {
      copula <- fitted(npudens(bws=bws,data=data))
    }

    u <- matrix(NA,bws$nobs,num.var)
    for (j in seq_len(num.var)) {
      stage <- stage + 1L
      progress <- .npcopula_progress_step(
        progress, stage,
        sprintf("marginal %s at sample realizations", bws$xnames[j])
      )
      bws.F.marginal <- npudistbw(formula(paste("~",bws$xnames[j])),
                         bws=bws$bw[j],
                         bandwidth.compute=FALSE,
                         bwtype=bws$type,
                         ckerorder=bws$ckerorder,
                         ckertype=bws$ckertype,
                         okertype=bws$okertype,
                         data=data)

      u[,j] <- fitted(npudist(bws=bws.F.marginal,data=data))
      if(density) {
        bws.f.marginal <- npudensbw(formula(paste("~",bws$xnames[j])),
                           bws=bws$bw[j],
                           bandwidth.compute=FALSE,
                           bwtype=bws$type,
                           ckerorder=bws$ckerorder,
                           ckertype=bws$ckertype,
                           okertype=bws$okertype,
                           ukertype=bws$ukertype,
                           data=data)
        copula <- copula/NZD(fitted(npudens(bws=bws.f.marginal,data=data)))

      }
    }
  } else {
    n.u <- nrow(u)
    x.u <- data.frame(matrix(NA,n.u,num.var))
    names(x.u) <- bws$xnames
    for (j in seq_len(num.var)) {
      stage <- stage + 1L
      progress <- .npcopula_progress_step(
        progress, stage,
        sprintf("quasi-inverse marginal %s", bws$xnames[j])
      )
      x.marginal <- data[[bws$xnames[j]]]
      quantile.seq <- seq(0,1,length=round(n.quasi.inv/2))
      if(is.numeric(x.marginal)) {
        x.er <- extendrange(x.marginal,f=er.quasi.inv)
        x.q <- quantile(x.marginal,quantile.seq)
        x.eval <- sort(c(seq(x.er[1],x.er[2],length=round(n.quasi.inv/2)),x.q))
      } else {
        x.u[,j] <- ordered(x.u[,j],levels=levels(x.marginal))
        x.q <- sapply(seq_len(round(n.quasi.inv/2)), function(i) { uocquantile(x.marginal, quantile.seq[i]) })
        x.eval <- sort(ordered(c(as.character(x.q),as.character(x.q)),levels=levels(x.marginal)))
      }
      F <- fitted(npudist(tdat=x.marginal,
                          edat=x.eval,
                          bws=bws$bw[j],
                          bwtype=bws$type,
                          ckerorder=bws$ckerorder,
                          ckertype=bws$ckertype,
                          okertype=bws$okertype,
                          ukertype=bws$ukertype,data=data))
      for (i in seq_len(n.u)) {
        u[u[,j]<min(F),j] <- min(F)
        u[u[,j]>max(F),j] <- max(F)
        x.u[i,j] <-  min(x.eval[F>=u[i,j]])
      }
    }

    stage <- stage + 1L
    progress <- .npcopula_progress_step(progress, stage, "expand u grid")
    x.u <- expand.grid(data.frame(x.u))
    for (k in seq_len(ncol(x.u))) {
      if(is.ordered(data[,k])) x.u[,k] <- ordered(x.u[,k],levels=levels(data[,k]))
    }
    if(!density) {
      stage <- stage + 1L
      detail <- "joint distribution on expanded grid"
      progress <- .npcopula_progress_step(progress, stage, detail)
      copula <- .npcopula_eval_expanded_grid(
        bws = bws,
        data = data,
        x.u = x.u,
        density = density
      )
    } else {
      stage <- stage + 1L
      detail <- "copula density on expanded grid"
      progress <- .npcopula_progress_step(progress, stage, detail)
      copula <- .npcopula_eval_expanded_grid(
        bws = bws,
        data = data,
        x.u = x.u,
        density = density
      )
    }
  }

  if(!u.provided) {
    u <- data.frame(u)
    names(u) <- paste("u", seq_len(num.var), sep = "")
    out <- data.frame(copula,u)
  } else {
    u.grid <- expand.grid(data.frame(u.requested))
    names(u.grid) <- paste("u", seq_len(num.var), sep = "")
    u <- expand.grid(data.frame(u))
    names(u) <- paste("u", seq_len(num.var), sep = "")
    out <- data.frame(copula,u,x.u)
  }

  .npcopula_object(
    result = out,
    bws = bws,
    data = data,
    target = target,
    evaluation = if (u.provided) "grid" else "sample",
    u.provided = u.provided,
    u.auto = u.auto,
    grid.dim = grid.dim,
    u.input = u.input,
    u.grid = if (u.provided) u.grid else NULL,
    neval = neval,
    n.quasi.inv = n.quasi.inv,
    er.quasi.inv = er.quasi.inv
  )
}

.np_plot_plbandwidth_engine <-
  function(bws,
           xdat,
           ydat,
           zdat,
           data = NULL,
           xq = 0.5,
           zq = 0.5,
           xtrim = 0.0,
           ztrim = 0.0,
           neval = 50,
           common.scale = TRUE,
           perspective = TRUE,
           renderer = c("base", "rgl"),
           gradients = FALSE,
           coef = FALSE,
           main = NULL,
           type = NULL,
           border = NULL,
           cex = NULL,
           cex.axis = NULL,
           cex.lab = NULL,
           cex.main = NULL,
           cex.sub = NULL,
           col = NULL,
           ylab = NULL,
           xlab = NULL,
           zlab = NULL,
           sub = NULL,
           ylim = NULL,
           xlim = NULL,
           zlim = NULL,
           lty = NULL,
           lwd = NULL,
           theta = 0.0,
           phi = 20.0,
           view = c("rotate","fixed"),
           plot.behavior = c("plot","plot-data","data"),
           plot.errors.method = c("none","bootstrap","asymptotic"),
           plot.errors.boot.method = c("wild", "inid", "fixed", "geom"),
           plot.errors.boot.nonfixed = c("exact", "frozen"),
           plot.errors.boot.wild = c("rademacher", "mammen"),
           plot.errors.boot.blocklen = NULL,
           plot.errors.boot.num = 1999,
           plot.errors.center = c("estimate","bias-corrected"),
           plot.errors.type = c("pmzsd","pointwise","bonferroni","simultaneous","all"),
           plot.errors.alpha = 0.05,
           plot.errors.style = c("band","bar"),
           plot.errors.bar = c("|","I"),
           plot.errors.bar.num = min(neval,25),
           plot.bxp = FALSE,
           plot.bxp.out = TRUE,
           plot.par.mfrow = TRUE,
           plot.data.overlay = TRUE,
           plot.rug = FALSE,
           ...,
           random.seed){

    engine.ctx <- .np_plot_engine_begin(plot.par.mfrow = plot.par.mfrow)
    on.exit(.np_plot_restore_par(engine.ctx$oldpar), add = TRUE)
    plot.par.mfrow <- engine.ctx$plot.par.mfrow

    scalar_default <- function(value, default) {
      if (is.null(value)) default else value
    }

    dots <- list(...)
    plot.legend <- if (!is.null(dots$legend)) dots$legend else TRUE
    plot.user.args <- .np_plot_user_args(dots, "plot")
    points.user.args <- .np_plot_user_args(dots, "points")
    persp.user.args <- .np_plot_user_args(dots, "persp")
    bxp.user.args <- .np_plot_user_args(dots, "bxp")
    rgl.persp3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.persp3d", "rgl.persp3d.")
    rgl.view3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.view3d", "rgl.view3d.")
    rgl.par3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.par3d", "rgl.par3d.")
    rgl.grid3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.grid3d", "rgl.grid3d.")
    rgl.widget.user.args <- .np_plot_collect_rgl_args(dots, "rgl.widget", "rgl.widget.")
    rgl.legend3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.legend3d", "rgl.legend3d.")
    rgl.legend3d.user.args <- .np_plot_merge_rgl_legend_control(rgl.legend3d.user.args, plot.legend)
    rgl.points3d.user.args <- .np_plot_extract_prefixed_args(dots, "rgl.points3d.")
    rgl.surface3d.user.args <- .np_plot_extract_prefixed_args(dots, "rgl.surface3d.")
    if (!is.null(cex)) {
      if (is.null(plot.user.args$cex)) plot.user.args$cex <- cex
      if (is.null(points.user.args$cex)) points.user.args$cex <- cex
      if (is.null(bxp.user.args$cex)) bxp.user.args$cex <- cex
    }
    overlay.col <- points.user.args$col
    overlay.points.args <- points.user.args

    bxp.args <- bxp.user.args
    if (!is.null(col)) bxp.args$col <- col
    if (!is.null(lty)) bxp.args$lty <- lty
    if (!is.null(lwd)) bxp.args$lwd <- lwd
    if (!is.null(border)) bxp.args$border <- border

    if(!missing(gradients))
      stop("gradients not supported with partially linear models. Coefficients may be extracted with coef()")
    coef <- isTRUE(coef)

    miss.xyz = c(missing(xdat), missing(ydat), missing(zdat))
    
    if (any(miss.xyz) && !all(miss.xyz))
      stop("one of, but not both, xdat and ydat was specified")
    else if(all(miss.xyz) && !is.null(bws$formula)){
      tt <- terms(bws)
      m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)

      tmf.xf <- tmf.x <- tmf <- bws$call[c(1,m)]
      tmf.xf[[1]] <- tmf[[1]] <- as.name("model.frame")
      tmf[["formula"]] <- tt
      mf.args <- as.list(tmf)[-1L]
      umf <- tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

      bronze <- lapply(bws$chromoly, paste, collapse = " + ")

      tmf.xf[["formula"]] <- as.formula(paste(" ~ ", bronze[[2]]),
                                      env = environment(tt))
      mf.xf.args <- as.list(tmf.xf)[-1L]
      tmf.xf <- do.call(stats::model.frame, mf.xf.args, envir = environment(tt))
      
      ydat <- model.response(tmf)
      xdat <- tmf.xf

      zdat <- tmf[, bws$chromoly[[3]], drop = FALSE]
    } else {
      if(all(miss.xyz) && !is.null(bws$call)){
        xdat <- data.frame(.np_eval_bws_call_arg(bws, "xdat"))
        ydat = .np_eval_bws_call_arg(bws, "ydat")
        zdat <- data.frame(.np_eval_bws_call_arg(bws, "zdat"))
      }
      xdat = toFrame(xdat)
      zdat = toFrame(zdat)
      
      ## catch and destroy NA's
      keep.rows <- rep_len(TRUE, nrow(xdat))
      rows.omit <- attr(na.omit(data.frame(xdat, ydat, zdat)), "na.action")
      if (length(rows.omit) > 0L)
        keep.rows[as.integer(rows.omit)] <- FALSE

      if (!any(keep.rows))
        stop("Training data has no rows without NAs")

      xdat <- xdat[keep.rows,,drop = FALSE]
      ydat <- ydat[keep.rows]
      zdat <- zdat[keep.rows,,drop = FALSE]
    }


    nxcon = sum(bws$xdati$icon)
    nxuno = sum(bws$xdati$iuno)
    nxord = sum(bws$xdati$iord)

    nzcon = sum(bws$zdati$icon)
    nzuno = sum(bws$zdati$iuno)
    nzord = sum(bws$zdati$iord)

    xq = double(bws$xndim)+xq
    zq = double(bws$zndim)+zq
    
    xtrim = double(bws$xndim)+xtrim
    ztrim = double(bws$zndim)+ztrim


    if (missing(plot.errors.method) &
        any(!missing(plot.errors.boot.num), !missing(plot.errors.boot.method),
            !missing(plot.errors.boot.nonfixed),
            !missing(plot.errors.boot.blocklen))){
      stop(
        "errors must be set to \"bootstrap\" when bootstrap controls are supplied",
        call. = FALSE
      )
    }
    
    normalized.opts <- .np_plot_normalize_common_options(
      plot.behavior = plot.behavior,
      plot.errors.method = plot.errors.method,
      plot.errors.boot.method = plot.errors.boot.method,
      plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
      plot.errors.boot.wild = plot.errors.boot.wild,
      plot.errors.boot.blocklen = plot.errors.boot.blocklen,
      plot.errors.center = plot.errors.center,
      plot.errors.type = plot.errors.type,
      plot.errors.alpha = plot.errors.alpha,
      plot.errors.style = plot.errors.style,
      plot.errors.bar = plot.errors.bar,
      xdat = xdat,
      common.scale = common.scale,
      ylim = ylim,
      allow_asymptotic_quantile = FALSE
    )

    plot.behavior <- normalized.opts$plot.behavior
    plot.errors.method <- normalized.opts$plot.errors.method
    plot.errors.boot.method <- normalized.opts$plot.errors.boot.method
    plot.errors.boot.nonfixed <- normalized.opts$plot.errors.boot.nonfixed
    plot.errors.boot.wild <- normalized.opts$plot.errors.boot.wild
    plot.errors.boot.blocklen <- normalized.opts$plot.errors.boot.blocklen
    plot.errors.center <- normalized.opts$plot.errors.center
    plot.errors.type <- normalized.opts$plot.errors.type
    plot.errors.alpha <- normalized.opts$plot.errors.alpha
    plot.errors.style <- normalized.opts$plot.errors.style
    plot.errors.bar <- normalized.opts$plot.errors.bar
    common.scale <- normalized.opts$common.scale

    plot.errors = (plot.errors.method != "none")
    first.render <- .np_plot_first_render_state()
    on.exit(.np_plot_activity_end(first.render$activity), add = TRUE)
    if (identical(.np_plot_match_renderer(renderer), "rgl") && coef) {
      .np_plot_validate_renderer_request(
        renderer = renderer,
        route = "npplreg/npplregbw",
        perspective = perspective,
        supported.route = TRUE,
        view = if (missing(view)) "fixed" else view,
        gradients = FALSE,
        coef = coef,
        plot.errors.method = plot.errors.method,
        plot.data.overlay = FALSE,
        plot.behavior = plot.behavior
      )
    }

    if (coef) {
      fit.coef <- if (identical(plot.errors.method, "asymptotic")) {
        npplreg(
          bws = bws,
          txdat = xdat,
          tydat = ydat,
          tzdat = zdat
        )
      } else {
        .np_plot_plreg_local_fit(
          bws = bws,
          xdat = xdat,
          ydat = ydat,
          zdat = zdat
        )
      }
      cf <- as.double(fit.coef$xcoef)
      se <- as.double(fit.coef$xcoeferr)
      cf.names <- if (!is.null(names(fit.coef$xcoef))) names(fit.coef$xcoef) else bws$xnames
      if (is.null(cf.names) || length(cf.names) != length(cf))
        cf.names <- paste0("x", seq_along(cf))

      if (plot.behavior != "data") {
        bp <- barplot(cf,
                      names.arg = cf.names,
                      ylab = scalar_default(ylab, "Linear Coefficient"),
                      xlab = scalar_default(xlab, "Regressor"),
                      main = scalar_default(main, ""),
                      col = scalar_default(col, "gray70"),
                      border = scalar_default(border, par("fg")))
        if (plot.errors && length(se) == length(cf) && all(is.finite(se))) {
          arrows(bp, cf - se, bp, cf + se, angle = 90, code = 3, length = 0.05, lwd = 1)
        }
      }

      if (plot.behavior == "plot")
        return(invisible(NULL))

      return(list(coefficients = cf, coefficient.stderr = se, fit = fit.coef))
    }

    overlay.ok <- .np_plot_overlay_enabled(
      plot.data.overlay = plot.data.overlay,
      plot.behavior = plot.behavior,
      gradients = FALSE,
      coef = coef,
      plot.data.overlay.missing = missing(plot.data.overlay)
    )

    surface.supported <-
      (nxcon + nxord == 1) &&
      (nzcon + nzord == 1) &&
      (nxuno + nzuno == 0) &&
      !any(xor(bws$xdati$iord, bws$xdati$inumord)) &&
      !any(xor(bws$zdati$iord, bws$zdati$inumord))

    renderer <- .np_plot_validate_renderer_request(
      renderer = renderer,
      route = "npplreg/npplregbw",
      perspective = perspective,
      supported.route = surface.supported,
      view = if (missing(view)) "fixed" else view,
      gradients = FALSE,
      coef = coef,
      plot.errors.method = plot.errors.method,
      plot.data.overlay = overlay.ok,
      plot.behavior = plot.behavior,
      allow.plot.errors = TRUE,
      allow.plot.data.overlay = TRUE
    )
    plot.rug <- .np_plot_validate_rug_request(
      plot.rug = plot.rug,
      route = "npplreg/npplregbw",
      supported.route = if (identical(renderer, "rgl")) {
        isTRUE(surface.supported && perspective)
      } else {
        TRUE
      },
      renderer = renderer,
      reason = if (identical(renderer, "rgl")) {
        "supported rgl surface routes"
      } else {
        "supported base plot routes"
      }
    )

    if (surface.supported && perspective && !gradients){

      view = match.arg(view)
      rotate = (view == "rotate")
      
      if (is.ordered(xdat[,1])){
        x1.eval = bws$xdati$all.ulev[[1]]
        x1.neval = length(x1.eval)
      } else {
        x1.neval = neval
        qi = trim.quantiles(xdat[,1], xtrim[1])
        x1.eval = seq(qi[1], qi[2], length.out = x1.neval)
      }

      if (is.ordered(zdat[,1])){
        z1.eval = bws$zdati$all.ulev[[1]]
        z1.neval = length(z1.eval)
      } else {
        z1.neval = neval
        qi = trim.quantiles(zdat[,1], ztrim[1])
        z1.eval = seq(qi[1], qi[2], length.out = z1.neval)
      }

      x.eval <- expand.grid(x1.eval, z1.eval)

      if (bws$xdati$iord[1])
        x1.eval <- (bws$xdati$all.dlev[[1]])[as.integer(x1.eval)]
      
      if (bws$zdati$iord[1])
        z1.eval <- (bws$zdati$all.dlev[[1]])[as.integer(z1.eval)]

      tobj <- if (identical(plot.errors.method, "asymptotic")) {
        .np_plot_plreg_asymptotic_fit(
          bws = bws,
          xdat = xdat,
          ydat = ydat,
          zdat = zdat,
          exdat = x.eval[,1, drop = FALSE],
          ezdat = x.eval[,2, drop = FALSE]
        )
      } else {
        .np_plot_plreg_local_fit(
          bws = bws,
          xdat = xdat,
          ydat = ydat,
          zdat = zdat,
          exdat = x.eval[,1, drop = FALSE],
          ezdat = x.eval[,2, drop = FALSE]
        )
      }

      terr = matrix(data = NA, nrow = nrow(x.eval), ncol = 3)
      lerr.all <- NULL
      herr.all <- NULL
      
      treg = matrix(data = tobj$mean,
        nrow = x1.neval, ncol = z1.neval, byrow = FALSE)

      if (plot.errors.method == "asymptotic") {
        terr.obj <- .np_plot_asymptotic_error_from_se(
          se = tobj$merr,
          alpha = plot.errors.alpha,
          band.type = plot.errors.type,
          m = nrow(x.eval)
        )
        terr[,1:2] <- terr.obj$err
        terr.all <- terr.obj$all.err
        lerr = matrix(data = tobj$mean - terr[,1],
          nrow = x1.neval, ncol = z1.neval, byrow = FALSE)
        herr = matrix(data = tobj$mean + terr[,2],
          nrow = x1.neval, ncol = z1.neval, byrow = FALSE)
        if (plot.errors.type == "all" && !is.null(terr.all)) {
          lerr.all <- lapply(terr.all, function(te)
            matrix(data = tobj$mean - te[,1], nrow = x1.neval, ncol = z1.neval, byrow = FALSE))
          herr.all <- lapply(terr.all, function(te)
            matrix(data = tobj$mean + te[,2], nrow = x1.neval, ncol = z1.neval, byrow = FALSE))
        }

      } else if (plot.errors.method == "bootstrap"){
        terr.obj <- compute.bootstrap.errors(
          xdat = xdat, ydat = ydat, zdat = zdat,
          exdat = x.eval[,1, drop = FALSE], ezdat = x.eval[,2, drop = FALSE],
          gradients = FALSE,
          slice.index = 0,
          progress.target = NULL,
          plot.errors.boot.method = plot.errors.boot.method,
          plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
          plot.errors.boot.wild = plot.errors.boot.wild,
          plot.errors.boot.blocklen = plot.errors.boot.blocklen,
          plot.errors.boot.num = plot.errors.boot.num,
          plot.errors.center = plot.errors.center,
          plot.errors.type = plot.errors.type,
          plot.errors.alpha = plot.errors.alpha,
          bws = bws)
        terr <- terr.obj[["boot.err"]]
        terr.all <- terr.obj[["boot.all.err"]]

        pc = (plot.errors.center == "bias-corrected")
        center.val <- if(pc) terr[,3] else {tobj$mean}

        lerr = matrix(data = center.val - terr[,1],
          nrow = x1.neval, ncol = z1.neval, byrow = FALSE)

        herr = matrix(data = center.val + terr[,2],
          nrow = x1.neval, ncol = z1.neval, byrow = FALSE)
        if (plot.errors.type == "all" && !is.null(terr.all)) {
          lerr.all <- lapply(terr.all, function(te)
            matrix(data = center.val - te[,1], nrow = x1.neval, ncol = z1.neval, byrow = FALSE))
          herr.all <- lapply(terr.all, function(te)
            matrix(data = center.val + te[,2], nrow = x1.neval, ncol = z1.neval, byrow = FALSE))
        }

      }
      
      if(is.null(zlim)) {
          zlim =
              if (plot.errors && plot.errors.type == "all" && !is.null(lerr.all))
                  c(min(c(unlist(lerr.all), lerr)), max(c(unlist(herr.all), herr)))
              else if (plot.errors)
                  c(min(lerr),max(herr))
              else
                  c(min(tobj$mean),max(tobj$mean))
      }
      if (overlay.ok)
        zlim <- .np_plot_overlay_range(zlim, ydat)
        
      if (plot.behavior != "plot"){
        r1 = plregression(bws = bws, xcoef = tobj$xcoef,
          xcoefvcov = vcov(tobj),
          xcoeferr = tobj$xcoeferr,
          evalx =  x.eval[,1],
          evalz =  x.eval[,2],
          mean = tobj$mean,
          ntrain = dim(xdat)[1],
          trainiseval = FALSE,
          xtra=c(tobj$RSQ,tobj$MSE,0,0,0,0))

        r1$merr = NA
        r1$bias = NA

        if (plot.errors)
          r1$merr = terr[,1:2]
        
        if (plot.errors.center == "bias-corrected")
          r1$bias = terr[,3] - treg

        if (plot.behavior == "data")
          return ( list(r1 = r1) )

      }

      xlab.val <- scalar_default(xlab, gen.label(names(xdat)[1], "X1"))
      ylab.val <- scalar_default(ylab, gen.label(names(zdat)[1], "Z1"))
      zlab.val <- scalar_default(zlab, gen.label(names(ydat),"Conditional Mean"))

      if (identical(renderer, "rgl")) {
        rgl.view <- .np_plot_rgl_view_angles(theta = theta, phi = phi)
        main.val <- if (!is.null(main)) main else NULL
        overlay.x1 <- xdat[,1]
        if (is.factor(overlay.x1) || is.ordered(overlay.x1))
          overlay.x1 <- (bws$xdati$all.dlev[[1]])[as.integer(overlay.x1)]
        overlay.x2 <- zdat[,1]
        if (is.factor(overlay.x2) || is.ordered(overlay.x2))
          overlay.x2 <- (bws$zdati$all.dlev[[1]])[as.integer(overlay.x2)]
        .np_plot_first_render_begin(first.render)
        rgl.out <- .np_plot_render_surface_rgl(
          x = x1.eval,
          y = z1.eval,
          z = treg,
          zlim = zlim,
          col = col,
          border = scalar_default(border, "black"),
          xlab = xlab.val,
          ylab = ylab.val,
          zlab = zlab.val,
          theta = rgl.view$theta,
          phi = rgl.view$phi,
          main = main.val,
          par3d.args = rgl.par3d.user.args,
          view3d.args = rgl.view3d.user.args,
          persp3d.args = rgl.persp3d.user.args,
          grid3d.args = rgl.grid3d.user.args,
          widget.args = rgl.widget.user.args,
          draw.extras = function() {
            if (plot.rug) {
              .np_plot_draw_floor_rug_rgl(
                x1 = overlay.x1,
                x2 = overlay.x2,
                zlim = zlim
              )
            }
            if (plot.errors) {
              .np_plot_error_surfaces_rgl(
                x = x1.eval,
                y = z1.eval,
                plot.errors.type = plot.errors.type,
                lerr = lerr,
                herr = herr,
                lerr.all = lerr.all,
                herr.all = herr.all,
                surface3d.args = rgl.surface3d.user.args,
                legend3d.args = rgl.legend3d.user.args
              )
            }
            if (overlay.ok) {
              .np_plot_overlay_points_rgl(
                x1 = overlay.x1,
                x2 = overlay.x2,
                y = ydat,
                points3d.args = rgl.points3d.user.args
              )
            }
          }
        )
        .np_plot_first_render_end(first.render)
        return(.np_plot_rgl_finalize(
          rgl.out = rgl.out,
          plot.behavior = plot.behavior,
          plot.data = list(r1 = r1)
        ))
      }

      main.val <- if (!is.null(main)) main else NULL

      rotate.defaults <- .np_plot_rotate_defaults()
      dtheta = rotate.defaults$dtheta

      persp.col = grDevices::adjustcolor(
        .np_plot_persp_surface_colors(z = treg, col = col),
        alpha.f = 0.5
      )
      frame.theta <- (0:((360 %/% dtheta - 1L) * rotate)) * dtheta + theta
      rotation.progress <- .np_plot_rotation_progress_begin(length(frame.theta))
      on.exit(.np_plot_rotation_progress_end(rotation.progress), add = TRUE)
      overlay.x1 <- xdat[,1]
      if (is.factor(overlay.x1) || is.ordered(overlay.x1))
        overlay.x1 <- (bws$xdati$all.dlev[[1]])[as.integer(overlay.x1)]
      overlay.x2 <- zdat[,1]
      if (is.factor(overlay.x2) || is.ordered(overlay.x2))
        overlay.x2 <- (bws$zdati$all.dlev[[1]])[as.integer(overlay.x2)]
      
      ##for (j in 0:((50 %/% dphi - 1)*rotate)*dphi+phi){
        for (frame.idx in seq_along(frame.theta)){
          i <- frame.theta[[frame.idx]]
          .np_plot_first_render_begin(first.render)
          persp.args <- list(x = x1.eval,
                             y = z1.eval,
                             z = treg,
                             zlim = zlim,
                             col = persp.col,
                             border = scalar_default(border, "black"),
                             ticktype = "detailed",
                             cex.axis = scalar_default(cex.axis, par()$cex.axis),
                             cex.lab = scalar_default(cex.lab, par()$cex.lab),
                             cex.main = scalar_default(cex.main, par()$cex.main),
                             cex.sub = scalar_default(cex.sub, par()$cex.sub),
                             lwd = 0.8 * scalar_default(lwd, par()$lwd),
                             xlab = xlab.val,
                             ylab = ylab.val,
                             zlab = zlab.val,
                             theta = i,
                             phi = phi,
                             main = main.val)
          persp.args <- .np_plot_merge_user_args(persp.args, persp.user.args)
          persp.mat <- do.call(persp, persp.args)
          .np_plot_first_render_end(first.render)
          .np_plot_draw_box_grid_persp(
            xlim = range(x1.eval, finite = TRUE),
            ylim = range(z1.eval, finite = TRUE),
            zlim = zlim,
            persp.mat = persp.mat
          )
          if (plot.rug) {
            .np_plot_draw_floor_rug_persp(
              x1 = overlay.x1,
              x2 = overlay.x2,
              zlim = zlim,
              persp.mat = persp.mat
            )
          }

          if (plot.errors){
            .np_plot_draw_error_wireframes_persp(
              x = x1.eval,
              y = z1.eval,
              persp.mat = persp.mat,
              plot.errors.type = plot.errors.type,
              lerr = lerr,
              herr = herr,
              lerr.all = lerr.all,
              herr.all = herr.all,
              border = scalar_default(border, "grey"),
              lwd = scalar_default(lwd, par()$lwd)
            )
            if (plot.errors.type == "all" && !is.null(lerr.all) && !is.null(herr.all)) {
              .np_plot_draw_all_band_legend(
                legend = plot.legend,
                x = "topright",
                lty = 1,
                lwd = 2.15 * scalar_default(lwd, par()$lwd)
              )
            }
          }
          if (overlay.ok)
            do.call(.np_plot_overlay_points_persp,
                    c(list(x1 = overlay.x1,
                           x2 = overlay.x2,
                           y = ydat,
                           persp.mat = persp.mat),
                      overlay.points.args))

          rotation.progress <- .np_plot_rotation_progress_tick(rotation.progress, done = frame.idx)
          Sys.sleep(if (isTRUE(rotate)) rotate.defaults$sleep else 0.5)
        }
      ##}

      if (plot.behavior == "plot-data")
        return ( list(r1 = r1) )

    } else {
      plot.layout <- .np_plot_layout_begin(
        plot.behavior = plot.behavior,
        plot.par.mfrow = plot.par.mfrow,
        mfrow = n2mfrow(bws$xndim + bws$zndim)
      )

      x.ev = xdat[1,,drop = FALSE]
      z.ev = zdat[1,,drop = FALSE]

      for (i in seq_len(bws$xndim))
        x.ev[1,i] = uocquantile(xdat[,i], prob=xq[i])

      for (i in seq_len(bws$zndim))
        z.ev[1,i] = uocquantile(zdat[,i], prob=zq[i])


      maxneval = max(c(sapply(xdat,nlevels), sapply(zdat,nlevels), neval))

      ## Preserve original data types (e.g., factors) for evaluation data
      exdat = xdat[rep(1, maxneval), , drop = FALSE]
      ezdat = zdat[rep(1, maxneval), , drop = FALSE]

      for (i in seq_len(bws$xndim))
        exdat[,i] = x.ev[1,i]

      for (i in seq_len(bws$zndim))
        ezdat[,i] = z.ev[1,i]

      if (common.scale){
        data.eval = matrix(data = NA, nrow = maxneval,
          ncol = (bws$xndim + bws$zndim))
        
        data.err = matrix(data = NA, nrow = maxneval,
          ncol = 3*(bws$xndim + bws$zndim))
        data.err.all = vector("list", bws$xndim + bws$zndim)

        allei = as.data.frame(matrix(data = NA, nrow = maxneval,
          ncol = bws$xndim + bws$zndim))

        all.bxp = list()
      }

      all.isFactor = c(vapply(xdat, is.factor, logical(1)), vapply(zdat, is.factor, logical(1)))

      plot.out = list()

      temp.err = matrix(data = NA, nrow = maxneval, ncol = 3)
      temp.mean = replicate(maxneval, NA)

      ## plotting expressions
      
      plot.bootstrap = plot.errors.method == "bootstrap"

      plotOnEstimate = (plot.errors.center == "estimate")

      plot.index = 0
      xOrZ = "x"

      for (i in seq_len(bws$xndim)){
        plot.index = plot.index + 1
        temp.err[,] = NA
        temp.mean[] =  NA
        temp.boot = list()
        temp.all.err <- NULL

        xi.factor = all.isFactor[plot.index]
        
        if (xi.factor){
          ei = levels(xdat[,i])
          ei = factor(ei, levels = ei)
          xi.neval = length(ei)
        } else {
          xi.neval = neval
          qi = trim.quantiles(xdat[,i], xtrim[i])
          ei = seq(qi[1], qi[2], length.out = neval)
        }

        if (xi.neval < maxneval){
          ei[(xi.neval+1):maxneval] = NA
        }


        tobj <- if (identical(plot.errors.method, "asymptotic")) {
          .np_plot_plreg_asymptotic_fit(
            bws = bws,
            xdat = xdat,
            ydat = ydat,
            zdat = zdat,
            exdat = subcol(exdat,ei,i)[seq_len(xi.neval),, drop = FALSE],
            ezdat = ezdat[seq_len(xi.neval),, drop = FALSE]
          )
        } else {
          .np_plot_plreg_local_fit(
            bws = bws,
            xdat = xdat,
            ydat = ydat,
            zdat = zdat,
            exdat = subcol(exdat,ei,i)[seq_len(xi.neval),, drop = FALSE],
            ezdat = ezdat[seq_len(xi.neval),, drop = FALSE]
          )
        }

        temp.mean[seq_len(xi.neval)] = tobj$mean

        if (plot.errors){
          if (plot.errors.method == "asymptotic") {
            asym.obj <- .np_plot_asymptotic_error_from_se(
              se = tobj$merr,
              alpha = plot.errors.alpha,
              band.type = plot.errors.type,
              m = xi.neval
            )
            temp.err[seq_len(xi.neval),1:2] <- asym.obj$err
            temp.all.err <- asym.obj$all.err
          } else if (plot.errors.method == "bootstrap"){
            temp.boot <- compute.bootstrap.errors(
                      xdat = xdat,
                      ydat = ydat,
                      zdat = zdat,
                      exdat = subcol(exdat,ei,i)[seq_len(xi.neval),, drop = FALSE],
                      ezdat = ezdat[seq_len(xi.neval),, drop = FALSE],
                      gradients = gradients,
                      slice.index = plot.index,
                      progress.target = .np_plot_scoef_bootstrap_target_label(
                        bws = bws,
                        slice.index = plot.index
                      ),
                      plot.errors.boot.method = plot.errors.boot.method,
                      plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
                      plot.errors.boot.wild = plot.errors.boot.wild,
                      plot.errors.boot.blocklen = plot.errors.boot.blocklen,
                      plot.errors.boot.num = plot.errors.boot.num,
                      plot.errors.center = plot.errors.center,
                      plot.errors.type = plot.errors.type,
                      plot.errors.alpha = plot.errors.alpha,
                      bws = bws)
            temp.err[seq_len(xi.neval),] <- temp.boot[["boot.err"]]
            temp.all.err <- temp.boot[["boot.all.err"]]
            temp.boot <- temp.boot[["bxp"]]
            if (!plot.bxp.out){
              temp.boot$out <- numeric()
              temp.boot$group <- integer()
            }
          }
        }

        if (common.scale){
          allei[,plot.index] = ei
          data.eval[, plot.index] = temp.mean
          if (plot.errors){
            all.bxp[plot.index] = NA
            all.bxp[[plot.index]] = temp.boot

            data.err[, c(3*plot.index-2,3*plot.index-1,3*plot.index)] = temp.err
            data.err.all[[plot.index]] = temp.all.err
          }
        } else if (plot.behavior != "data") {
          plot.layout <- .np_plot_layout_activate(plot.layout)
          ## plot evaluation
          plot.fun <- if (xi.factor) {
            .np_plot_panel_fun(plot.bootstrap = plot.bootstrap, plot.bxp = plot.bxp)
          } else {
            plot
          }
          plot.args <- list()
          if (xi.factor) {
            if (plot.bootstrap && plot.bxp) plot.args$z <- temp.boot else plot.args$f <- ei
          } else {
            plot.args$x <- ei
          }
          if (!(xi.factor && plot.bootstrap && plot.bxp))
            plot.args$y <- temp.mean
          panel.ylim <- NULL
          if (plot.errors)
            panel.ylim <- if (plot.errors.type == "all")
              compute.all.error.range(if (plotOnEstimate) temp.mean else temp.err[,3], temp.all.err)
            else
              c(min(na.omit(c(temp.mean - temp.err[,1], temp.err[,3] - temp.err[,1]))),
                max(na.omit(c(temp.mean + temp.err[,2], temp.err[,3] + temp.err[,2]))))
          if (overlay.ok)
            panel.ylim <- .np_plot_overlay_range(panel.ylim, ydat)
          if (!is.null(panel.ylim))
            plot.args$ylim <- panel.ylim
          plot.args$xlab <- gen.label(if (xOrZ == "x") bws$xnames[i] else bws$znames[i],
                                      paste(toupper(xOrZ), i, sep = ""))
          plot.args$ylab <- paste(if (gradients) paste("Gradient Component ", i, " of", sep = "") else "",
                                  gen.label(bws$ynames, "Conditional Mean"))
          if (!xi.factor) {
            plot.args$type <- scalar_default(type, "l")
            plot.args$lty <- scalar_default(lty, par()$lty)
            plot.args$col <- scalar_default(col, par()$col)
            plot.args$lwd <- scalar_default(lwd, par()$lwd)
            plot.args$cex.axis <- scalar_default(cex.axis, par()$cex.axis)
            plot.args$cex.lab <- scalar_default(cex.lab, par()$cex.lab)
            plot.args$cex.main <- scalar_default(cex.main, par()$cex.main)
            plot.args$cex.sub <- scalar_default(cex.sub, par()$cex.sub)
          } else {
            if (!is.null(col)) plot.args$col <- col
            if (!is.null(lty)) plot.args$lty <- lty
            if (!is.null(lwd)) plot.args$lwd <- lwd
          }
          plot.args$main <- scalar_default(main, "")
          plot.args$sub <- scalar_default(sub, "")
          plot.args <- .np_plot_merge_user_args(
            plot.args,
            if (xi.factor && plot.bootstrap && plot.bxp) bxp.args else plot.user.args
          )
          if (overlay.ok && !xi.factor) {
            type.val <- plot.args$type
            plot.args$type <- "n"
            .np_plot_first_render_begin(first.render)
            do.call(plot.fun, plot.args)
            .np_plot_first_render_end(first.render)
            overlay.x <- if (xOrZ == "x") xdat[,i] else zdat[,i]
            do.call(.np_plot_overlay_points_1d,
                    c(list(x = overlay.x, y = ydat),
                      overlay.points.args))
            if (!identical(type.val, "n")) {
              ok.line <- is.finite(ei) & is.finite(temp.mean)
              line.args <- list(x = ei[ok.line],
                                y = temp.mean[ok.line],
                                type = type.val,
                                lty = plot.args$lty,
                                lwd = plot.args$lwd,
                                col = plot.args$col)
              do.call(lines, line.args)
            }
            if (plot.rug)
              .np_plot_draw_rug_1d(overlay.x)
          } else if (overlay.ok && xi.factor) {
            axis.labels <- levels(ei)
            axis.at <- seq_along(axis.labels)
            add.axis <- is.null(plot.user.args$xaxt)
            base.args <- list(x = as.numeric(ei),
                              y = temp.mean,
                              type = "n",
                              xlab = plot.args$xlab,
                              ylab = plot.args$ylab,
                              main = plot.args$main,
                              sub = plot.args$sub)
            if (!is.null(plot.args$ylim))
              base.args$ylim <- plot.args$ylim
            axis.lim <- if (xOrZ == "x") xlim else zlim
            if (!is.null(axis.lim)) {
              base.args$xlim <- axis.lim
            } else {
              base.args$xlim <- c(0.5, length(axis.labels) + 0.5)
            }
            if (add.axis)
              base.args$xaxt <- "n"
            base.args <- .np_plot_merge_user_args(base.args, plot.user.args)
            .np_plot_first_render_begin(first.render)
            do.call(graphics::plot.default, base.args)
            .np_plot_first_render_end(first.render)
            if (add.axis)
              axis(1, at = axis.at, labels = axis.labels)
            overlay.x <- if (xOrZ == "x") xdat[,i] else zdat[,i]
            do.call(.np_plot_overlay_points_factor,
                    c(list(x = overlay.x, y = ydat),
                      overlay.points.args))
            if (plot.bootstrap && plot.bxp) {
              do.call(bxp, c(list(z = temp.boot, add = TRUE), bxp.args))
            } else {
              l.f <- rep(ei, each = 3)
              l.f[3 * seq_along(ei)] <- NA
              l.y <- unlist(lapply(temp.mean, function(p) c(0, p, NA)))
              lines(x = l.f, y = l.y, lty = 2)
              point.args <- list(x = ei, y = temp.mean)
              if (!is.null(col)) point.args$col <- col
              if (!is.null(points.user.args$pch)) point.args$pch <- points.user.args$pch
              if (!is.null(points.user.args$cex)) point.args$cex <- points.user.args$cex
              if (!is.null(points.user.args$bg)) point.args$bg <- points.user.args$bg
              do.call(points, point.args)
            }
          } else {
            .np_plot_first_render_begin(first.render)
            do.call(plot.fun, plot.args)
            .np_plot_first_render_end(first.render)
            if (plot.rug && !xi.factor)
              .np_plot_draw_rug_1d(if (xOrZ == "x") xdat[,i] else zdat[,i])
          }

          ## error plotting evaluation
          if (plot.errors && !(xi.factor && plot.bootstrap && plot.bxp)){
            if (!xi.factor && !plotOnEstimate)
              lines(na.omit(ei), na.omit(temp.err[,3]), lty = 3)

            if (plot.errors.type == "all") {
              draw.all.error.types(
                ex = as.numeric(na.omit(ei)),
                center = na.omit(if (plotOnEstimate) temp.mean else temp.err[,3]),
                all.err = temp.all.err,
                xi.factor = xi.factor,
                  legend = plot.legend)
            } else {
              draw.args <- list(
                ex = as.numeric(na.omit(ei)),
                ely = if (plotOnEstimate) na.omit(temp.mean - temp.err[,1]) else na.omit(temp.err[,3] - temp.err[,1]),
                ehy = if (plotOnEstimate) na.omit(temp.mean + temp.err[,2]) else na.omit(temp.err[,3] + temp.err[,2]),
                plot.errors.style = if (xi.factor) "bar" else plot.errors.style,
                plot.errors.bar = if (xi.factor) "I" else plot.errors.bar,
                plot.errors.bar.num = plot.errors.bar.num,
                lty = if (xi.factor) 1 else 2
              )
              do.call(draw.errors, draw.args)
            }

          }
        }

        if (plot.behavior != "plot") {
          plot.out[plot.index] = NA
          if (gradients){
          } else {
            plot.out[[plot.index]] =
              plregression(bws = bws, xcoef = tobj$xcoef, xcoefvcov = vcov(tobj),
                           xcoeferr = tobj$xcoeferr,
                           evalx = subcol(exdat,ei,i)[seq_len(xi.neval),, drop = FALSE],
                           evalz = ezdat[seq_len(xi.neval),, drop = FALSE],
                           mean = na.omit(temp.mean),
                           ntrain = dim(xdat)[1],
                           trainiseval = FALSE,
                           xtra = c(tobj$RSQ, tobj$MSE, 0, 0, 0, 0))
            plot.out[[plot.index]]$merr = NA
            plot.out[[plot.index]]$bias = NA

            if (plot.errors)
              plot.out[[plot.index]]$merr = temp.err[,1:2]

            if (plot.errors.center == "bias-corrected")
              plot.out[[plot.index]]$bias = temp.err[,3] - temp.mean
            plot.out[[plot.index]]$bxp = temp.boot
          }
        }
      }

      xOrZ = "z"
      for (i in seq_len(bws$zndim)){
        plot.index = plot.index + 1
        temp.err[,] = NA
        temp.mean[] =  NA
        temp.boot = list()
        temp.all.err <- NULL

        xi.factor = all.isFactor[plot.index]
        
        if (xi.factor){
          ei = levels(zdat[,i])
          ei = factor(ei, levels = ei)
          xi.neval = length(ei)
        } else {
          xi.neval = neval
          qi = trim.quantiles(zdat[,i], ztrim[i])
          ei = seq(qi[1], qi[2], length.out = neval)
        }

        if (xi.neval < maxneval){
          ei[(xi.neval+1):maxneval] = NA
        }

        tobj <- if (identical(plot.errors.method, "asymptotic")) {
          .np_plot_plreg_asymptotic_fit(
            bws = bws,
            xdat = xdat,
            ydat = ydat,
            zdat = zdat,
            exdat = exdat[seq_len(xi.neval),, drop = FALSE],
            ezdat = subcol(ezdat,ei,i)[seq_len(xi.neval),, drop = FALSE]
          )
        } else {
          .np_plot_plreg_local_fit(
            bws = bws,
            xdat = xdat,
            ydat = ydat,
            zdat = zdat,
            exdat = exdat[seq_len(xi.neval),, drop = FALSE],
            ezdat = subcol(ezdat,ei,i)[seq_len(xi.neval),, drop = FALSE]
          )
        }

        temp.mean[seq_len(xi.neval)] = tobj$mean

        if (plot.errors){
          if (plot.errors.method == "asymptotic") {
            asym.obj <- .np_plot_asymptotic_error_from_se(
              se = tobj$merr,
              alpha = plot.errors.alpha,
              band.type = plot.errors.type,
              m = xi.neval
            )
            temp.err[seq_len(xi.neval),1:2] <- asym.obj$err
            temp.all.err <- asym.obj$all.err
          } else if (plot.errors.method == "bootstrap"){
            temp.boot <- compute.bootstrap.errors(
                      xdat = xdat,
                      ydat = ydat,
                      zdat = zdat,
                      exdat = exdat[seq_len(xi.neval),, drop = FALSE],
                      ezdat = subcol(ezdat,ei,i)[seq_len(xi.neval),, drop = FALSE],
                      gradients = gradients,
                      slice.index = plot.index,
                      progress.target = .np_plot_scoef_bootstrap_target_label(
                        bws = bws,
                        slice.index = plot.index
                      ),
                      plot.errors.boot.method = plot.errors.boot.method,
                      plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
                      plot.errors.boot.wild = plot.errors.boot.wild,
                      plot.errors.boot.blocklen = plot.errors.boot.blocklen,
                      plot.errors.boot.num = plot.errors.boot.num,
                      plot.errors.center = plot.errors.center,
                      plot.errors.type = plot.errors.type,
                      plot.errors.alpha = plot.errors.alpha,
                      bws = bws)
            temp.err[seq_len(xi.neval),] <- temp.boot[["boot.err"]]
            temp.all.err <- temp.boot[["boot.all.err"]]
            temp.boot <- temp.boot[["bxp"]]
            if (!plot.bxp.out){
              temp.boot$out <- numeric()
              temp.boot$group <- integer()
            }
          }
        }

        if (common.scale){
          allei[,plot.index] = ei
          data.eval[, plot.index] = temp.mean
          if (plot.errors){
            all.bxp[plot.index] = NA
            all.bxp[[plot.index]] = temp.boot

            data.err[, c(3*plot.index-2,3*plot.index-1,3*plot.index)] = temp.err
            data.err.all[[plot.index]] = temp.all.err
          }
        } else if (plot.behavior != "data") {
          plot.layout <- .np_plot_layout_activate(plot.layout)
          ## plot evaluation
          plot.fun <- if (xi.factor) {
            .np_plot_panel_fun(plot.bootstrap = plot.bootstrap, plot.bxp = plot.bxp)
          } else {
            plot
          }
          plot.args <- list()
          if (xi.factor) {
            if (plot.bootstrap && plot.bxp) plot.args$z <- temp.boot else plot.args$f <- ei
          } else {
            plot.args$x <- ei
          }
          if (!(xi.factor && plot.bootstrap && plot.bxp))
            plot.args$y <- temp.mean
          if (plot.errors)
            plot.args$ylim <- if (plot.errors.type == "all")
              compute.all.error.range(if (plotOnEstimate) temp.mean else temp.err[,3], temp.all.err)
            else
              c(min(na.omit(c(temp.mean - temp.err[,1], temp.err[,3] - temp.err[,1]))),
                max(na.omit(c(temp.mean + temp.err[,2], temp.err[,3] + temp.err[,2]))))
          plot.args$xlab <- gen.label(if (xOrZ == "x") bws$xnames[i] else bws$znames[i],
                                      paste(toupper(xOrZ), i, sep = ""))
          plot.args$ylab <- paste(if (gradients) paste("Gradient Component ", i, " of", sep = "") else "",
                                  gen.label(bws$ynames, "Conditional Mean"))
          if (!xi.factor) {
            plot.args$type <- scalar_default(type, "l")
            plot.args$lty <- scalar_default(lty, par()$lty)
            plot.args$col <- scalar_default(col, par()$col)
            plot.args$lwd <- scalar_default(lwd, par()$lwd)
            plot.args$cex.axis <- scalar_default(cex.axis, par()$cex.axis)
            plot.args$cex.lab <- scalar_default(cex.lab, par()$cex.lab)
            plot.args$cex.main <- scalar_default(cex.main, par()$cex.main)
            plot.args$cex.sub <- scalar_default(cex.sub, par()$cex.sub)
          } else {
            if (!is.null(col)) plot.args$col <- col
            if (!is.null(lty)) plot.args$lty <- lty
            if (!is.null(lwd)) plot.args$lwd <- lwd
          }
          plot.args$main <- scalar_default(main, "")
          plot.args$sub <- scalar_default(sub, "")
          plot.args <- .np_plot_merge_user_args(
            plot.args,
            if (xi.factor && plot.bootstrap && plot.bxp) bxp.args else plot.user.args
          )
          if (overlay.ok && !xi.factor) {
            type.val <- plot.args$type
            plot.args$type <- "n"
            .np_plot_first_render_begin(first.render)
            do.call(plot.fun, plot.args)
            .np_plot_first_render_end(first.render)
            overlay.x <- if (xOrZ == "x") xdat[,i] else zdat[,i]
            do.call(.np_plot_overlay_points_1d,
                    c(list(x = overlay.x, y = ydat),
                      overlay.points.args))
            if (!identical(type.val, "n")) {
              ok.line <- is.finite(allei[,plot.index]) & is.finite(data.eval[,plot.index])
              line.args <- list(x = allei[ok.line, plot.index],
                                y = data.eval[ok.line, plot.index],
                                type = type.val,
                                lty = plot.args$lty,
                                lwd = plot.args$lwd,
                                col = plot.args$col)
              do.call(lines, line.args)
            }
            if (plot.rug)
              .np_plot_draw_rug_1d(overlay.x)
          } else if (overlay.ok && xi.factor) {
            axis.labels <- levels(allei[,plot.index])
            axis.at <- seq_along(axis.labels)
            add.axis <- is.null(plot.user.args$xaxt)
            base.args <- list(x = as.numeric(allei[,plot.index]),
                              y = data.eval[,plot.index],
                              type = "n",
                              xlab = plot.args$xlab,
                              ylab = plot.args$ylab,
                              main = plot.args$main,
                              sub = plot.args$sub)
            if (!is.null(plot.args$ylim))
              base.args$ylim <- plot.args$ylim
            axis.lim <- if (xOrZ == "x") xlim else zlim
            if (!is.null(axis.lim)) {
              base.args$xlim <- axis.lim
            } else {
              base.args$xlim <- c(0.5, length(axis.labels) + 0.5)
            }
            if (add.axis)
              base.args$xaxt <- "n"
            base.args <- .np_plot_merge_user_args(base.args, plot.user.args)
            .np_plot_first_render_begin(first.render)
            do.call(graphics::plot.default, base.args)
            .np_plot_first_render_end(first.render)
            if (add.axis)
              axis(1, at = axis.at, labels = axis.labels)
            overlay.x <- if (xOrZ == "x") xdat[,i] else zdat[,i]
            do.call(.np_plot_overlay_points_factor,
                    c(list(x = overlay.x, y = ydat),
                      overlay.points.args))
            if (plot.bootstrap && plot.bxp) {
              do.call(bxp, c(list(z = all.bxp[[plot.index]], add = TRUE), bxp.args))
            } else {
              l.f <- rep(allei[,plot.index], each = 3)
              l.f[3 * seq_along(allei[,plot.index])] <- NA
              l.y <- unlist(lapply(data.eval[,plot.index], function(p) c(0, p, NA)))
              lines(x = l.f, y = l.y, lty = 2)
              point.args <- list(x = allei[,plot.index], y = data.eval[,plot.index])
              if (!is.null(col)) point.args$col <- col
              if (!is.null(points.user.args$pch)) point.args$pch <- points.user.args$pch
              if (!is.null(points.user.args$cex)) point.args$cex <- points.user.args$cex
              if (!is.null(points.user.args$bg)) point.args$bg <- points.user.args$bg
              do.call(points, point.args)
            }
          } else {
            .np_plot_first_render_begin(first.render)
            do.call(plot.fun, plot.args)
            .np_plot_first_render_end(first.render)
            if (plot.rug && !xi.factor)
              .np_plot_draw_rug_1d(if (xOrZ == "x") xdat[,i] else zdat[,i])
          }

          ## error plotting evaluation
          if (plot.errors && !(xi.factor && plot.bootstrap && plot.bxp)){
            if (!xi.factor && !plotOnEstimate)
              lines(na.omit(ei), na.omit(temp.err[,3]), lty = 3)

            if (plot.errors.type == "all") {
              draw.all.error.types(
                ex = as.numeric(na.omit(ei)),
                center = na.omit(if (plotOnEstimate) temp.mean else temp.err[,3]),
                all.err = temp.all.err,
                xi.factor = xi.factor,
                  legend = plot.legend)
            } else {
              draw.args <- list(
                ex = as.numeric(na.omit(ei)),
                ely = if (plotOnEstimate) na.omit(temp.mean - temp.err[,1]) else na.omit(temp.err[,3] - temp.err[,1]),
                ehy = if (plotOnEstimate) na.omit(temp.mean + temp.err[,2]) else na.omit(temp.err[,3] + temp.err[,2]),
                plot.errors.style = if (xi.factor) "bar" else plot.errors.style,
                plot.errors.bar = if (xi.factor) "I" else plot.errors.bar,
                plot.errors.bar.num = plot.errors.bar.num,
                lty = if (xi.factor) 1 else 2
              )
              do.call(draw.errors, draw.args)
            }

          }
        }

        
        if (plot.behavior != "plot") {
          plot.out[plot.index] = NA
          if (gradients){
          } else {
            plot.out[[plot.index]] =
              plregression(bws = bws, xcoef = tobj$xcoef,
                           xcoeferr = tobj$xcoeferr,
                           xcoefvcov = vcov(tobj),
                           evalx = exdat[seq_len(xi.neval),, drop = FALSE],
                           evalz = subcol(ezdat,ei,i)[seq_len(xi.neval),, drop = FALSE],
                           mean = na.omit(temp.mean),
                           ntrain = dim(zdat)[1],
                           trainiseval = FALSE,
                           xtra = c(tobj$RSQ, tobj$MSE, 0, 0, 0, 0))
            plot.out[[plot.index]]$merr = NA
            plot.out[[plot.index]]$bias = NA

            if (plot.errors)
              plot.out[[plot.index]]$merr = temp.err[,1:2]

            if (plot.errors.center == "bias-corrected")
              plot.out[[plot.index]]$bias = temp.err[,3] - temp.mean
            plot.out[[plot.index]]$bxp = temp.boot
          }
        }
      }

      if (common.scale && (plot.behavior != "data")){
        if (plot.errors && plot.errors.type == "all") {
          y.min <- Inf
          y.max <- -Inf
          for (k in seq_len(bws$xndim + bws$zndim)) {
            if (is.null(data.err.all[[k]])) next
            nkeep.k <- nrow(data.err.all[[k]]$pointwise)
            if (nkeep.k == 0) next
            center.k <- if (plot.errors.center == "estimate")
              na.omit(data.eval[seq_len(nkeep.k), k])
            else
              na.omit(data.err[seq_len(nkeep.k), 3*k])
            range.k <- compute.all.error.range(center.k, data.err.all[[k]])
            y.min <- min(y.min, range.k[1], na.rm = TRUE)
            y.max <- max(y.max, range.k[2], na.rm = TRUE)
          }
          if (!is.finite(y.min) || !is.finite(y.max)) {
            y.min <- min(na.omit(as.double(data.eval)))
            y.max <- max(na.omit(as.double(data.eval)))
          }
        } else {
          jj = seq_len(bws$xndim + bws$zndim)*3
          if (plot.errors.center == "estimate" || !plot.errors) {
            y.max = max(na.omit(as.double(data.eval)) +
              if (plot.errors) na.omit(as.double(data.err[,jj-1]))
              else 0)
            y.min = min(na.omit(as.double(data.eval)) -
              if (plot.errors) na.omit(as.double(data.err[,jj-2]))
              else 0)
          } else if (plot.errors.center == "bias-corrected") {
            y.max = max(na.omit(as.double(data.err[,jj] + data.err[,jj-1])))
            y.min = min(na.omit(as.double(data.err[,jj] - data.err[,jj-2])))
          }
        }

        if (overlay.ok) {
          y.range <- .np_plot_overlay_range(c(y.min, y.max), ydat)
          y.min <- y.range[1L]
          y.max <- y.range[2L]
        }

        if(!is.null(ylim)){
          y.min = ylim[1]
          y.max = ylim[2]
        }
        
        xOrZ = "x"
        
        for (plot.index in seq_len(bws$xndim + bws$zndim)){
          plot.layout <- .np_plot_layout_activate(plot.layout)
          i = if (plot.index <= bws$xndim) plot.index else plot.index - bws$xndim

          if (plot.index > bws$xndim)
            xOrZ <- "z"
            
          xi.factor = all.isFactor[plot.index]

          ## plot evaluation
          plot.fun <- if (xi.factor) {
            .np_plot_panel_fun(plot.bootstrap = plot.bootstrap, plot.bxp = plot.bxp)
          } else {
            plot
          }
          plot.args <- list()
          if (xi.factor) {
            if (plot.bootstrap && plot.bxp) plot.args$z <- all.bxp[[plot.index]] else plot.args$f <- allei[,plot.index]
          } else {
            plot.args$x <- allei[,plot.index]
          }
          if (!(xi.factor && plot.bootstrap && plot.bxp))
            plot.args$y <- data.eval[,plot.index]
          plot.args$ylim <- c(y.min, y.max)
          plot.args$xlab <- gen.label(if (xOrZ == "x") bws$xnames[i] else bws$znames[i],
                                      paste(toupper(xOrZ), i, sep = ""))
          plot.args$ylab <- paste(if (gradients) paste("Gradient Component ", i, " of", sep = "") else "",
                                  gen.label(bws$ynames, "Conditional Mean"))
          if (!xi.factor) {
            plot.args$type <- scalar_default(type, "l")
            plot.args$lty <- scalar_default(lty, par()$lty)
            plot.args$col <- scalar_default(col, par()$col)
            plot.args$lwd <- scalar_default(lwd, par()$lwd)
            plot.args$cex.axis <- scalar_default(cex.axis, par()$cex.axis)
            plot.args$cex.lab <- scalar_default(cex.lab, par()$cex.lab)
            plot.args$cex.main <- scalar_default(cex.main, par()$cex.main)
            plot.args$cex.sub <- scalar_default(cex.sub, par()$cex.sub)
          } else {
            if (!is.null(col)) plot.args$col <- col
            if (!is.null(lty)) plot.args$lty <- lty
            if (!is.null(lwd)) plot.args$lwd <- lwd
          }
          plot.args$main <- scalar_default(main, "")
          plot.args$sub <- scalar_default(sub, "")
          plot.args <- .np_plot_merge_user_args(
            plot.args,
            if (xi.factor && plot.bootstrap && plot.bxp) bxp.args else plot.user.args
          )
          if (overlay.ok && !xi.factor) {
            type.val <- plot.args$type
            plot.args$type <- "n"
            .np_plot_first_render_begin(first.render)
            do.call(plot.fun, plot.args)
            .np_plot_first_render_end(first.render)
            overlay.x <- if (xOrZ == "x") xdat[,i] else zdat[,i]
            do.call(.np_plot_overlay_points_1d,
                    c(list(x = overlay.x, y = ydat),
                      overlay.points.args))
            if (!identical(type.val, "n")) {
              ok.line <- is.finite(allei[,plot.index]) & is.finite(data.eval[,plot.index])
              line.args <- list(x = allei[ok.line, plot.index],
                                y = data.eval[ok.line, plot.index],
                                type = type.val,
                                lty = plot.args$lty,
                                lwd = plot.args$lwd,
                                col = plot.args$col)
              do.call(lines, line.args)
            }
            if (plot.rug)
              .np_plot_draw_rug_1d(overlay.x)
          } else if (overlay.ok && xi.factor) {
            axis.labels <- levels(allei[,plot.index])
            axis.at <- seq_along(axis.labels)
            add.axis <- is.null(plot.user.args$xaxt)
            base.args <- list(x = as.numeric(allei[,plot.index]),
                              y = data.eval[,plot.index],
                              type = "n",
                              xlab = plot.args$xlab,
                              ylab = plot.args$ylab,
                              main = plot.args$main,
                              sub = plot.args$sub,
                              ylim = plot.args$ylim)
            axis.lim <- if (xOrZ == "x") xlim else zlim
            if (!is.null(axis.lim)) {
              base.args$xlim <- axis.lim
            } else {
              base.args$xlim <- c(0.5, length(axis.labels) + 0.5)
            }
            if (add.axis)
              base.args$xaxt <- "n"
            base.args <- .np_plot_merge_user_args(base.args, plot.user.args)
            .np_plot_first_render_begin(first.render)
            do.call(graphics::plot.default, base.args)
            .np_plot_first_render_end(first.render)
            if (add.axis)
              axis(1, at = axis.at, labels = axis.labels)
            overlay.x <- if (xOrZ == "x") xdat[,i] else zdat[,i]
            do.call(.np_plot_overlay_points_factor,
                    c(list(x = overlay.x, y = ydat),
                      overlay.points.args))
            if (plot.bootstrap && plot.bxp) {
              do.call(bxp, c(list(z = all.bxp[[plot.index]], add = TRUE), bxp.args))
            } else {
              l.f <- rep(allei[,plot.index], each = 3)
              l.f[3 * seq_along(allei[,plot.index])] <- NA
              l.y <- unlist(lapply(data.eval[,plot.index], function(p) c(0, p, NA)))
              lines(x = l.f, y = l.y, lty = 2)
              point.args <- list(x = allei[,plot.index], y = data.eval[,plot.index])
              if (!is.null(col)) point.args$col <- col
              if (!is.null(points.user.args$pch)) point.args$pch <- points.user.args$pch
              if (!is.null(points.user.args$cex)) point.args$cex <- points.user.args$cex
              if (!is.null(points.user.args$bg)) point.args$bg <- points.user.args$bg
              do.call(points, point.args)
            }
          } else {
            .np_plot_first_render_begin(first.render)
            do.call(plot.fun, plot.args)
            .np_plot_first_render_end(first.render)
            if (plot.rug && !xi.factor)
              .np_plot_draw_rug_1d(if (xOrZ == "x") xdat[,i] else zdat[,i])
          }

          ## error plotting evaluation
          if (plot.errors && !(xi.factor && plot.bootstrap && plot.bxp)){
            if (plot.errors.type == "all") {
              draw.all.error.types(
                ex = as.numeric(na.omit(allei[,plot.index])),
                center = if (plotOnEstimate)
                  na.omit(data.eval[,plot.index])
                else
                  na.omit(data.err[,3*plot.index]),
                all.err = data.err.all[[plot.index]],
                xi.factor = xi.factor,
                  legend = plot.legend)
            } else {
              if (!xi.factor && !plotOnEstimate)
                lines(na.omit(ei), na.omit(temp.err[,3]), lty = 3)

              draw.args <- list(
                ex = as.numeric(na.omit(allei[,plot.index])),
                ely = if (plotOnEstimate) na.omit(data.eval[,plot.index] - data.err[,3*plot.index-2]) else na.omit(data.err[,3*plot.index] - data.err[,3*plot.index-2]),
                ehy = if (plotOnEstimate) na.omit(data.eval[,plot.index] + data.err[,3*plot.index-1]) else na.omit(data.err[,3*plot.index] + data.err[,3*plot.index-1]),
                plot.errors.style = if (xi.factor) "bar" else plot.errors.style,
                plot.errors.bar = if (xi.factor) "I" else plot.errors.bar,
                plot.errors.bar.num = plot.errors.bar.num,
                lty = if (xi.factor) 1 else 2
              )
              do.call(draw.errors, draw.args)
            }
            

          }
        }
      }

      if (plot.behavior != "data" && plot.par.mfrow)
        par(mfrow=c(1,1),cex=par()$cex)
      
      if (plot.behavior != "plot"){
        names(plot.out) =
          if (gradients){ }
          else
            paste("plr", seq_len(bws$xndim + bws$zndim), sep = "")
        
        return (plot.out)
      }
    }
  }

## Function that implements the multivariate density equality test
## described in Li, Q., E. Maasoumi, and J.S. Racine (2009), "A
## Nonparametric Test for Equality of Distributions with Mixed
## Categorical and Continuous Data," Journal of Econometrics, Volume
## 148, pp 186-200.

npdeneqtest <- function(x = NULL,
                        y = NULL,
                        bw.x = NULL,
                        bw.y = NULL,
                        boot.num = 399,
                        random.seed = 42,
                        ...) {

  ## Some testing of input values

  if(is.null(x) || is.null(y)) stop(" you must provide x and y data")
  if(!is.data.frame(x) || !is.data.frame(y)) stop(" x and y must be data frames")
  if(!identical(names(data.frame(x)),names(data.frame(y)))) stop(" data frames x and y must have identical variable names")
  if(boot.num < 9) stop(" number of bootstrap replications must be >= 9")

  .np_progress_note("Computing bandwidths")

  if(is.null(bw.x) || is.null(bw.y)) {
    bw.x <- .np_progress_with_legacy_suppressed(npudensbw(dat=x,...))
    bw.y <- .np_progress_with_legacy_suppressed(npudensbw(dat=y,...))
  }

  ## Save seed prior to setting

  seed.state <- .np_seed_enter(random.seed)


  ## First, define test statistic function. This will return the
  ## standardized and unstandardized test statistic along with its
  ## estimated variance.

  teststat <- function(x,y,bw.x,bw.y) {

    ## Get n1 and n2, number of rows in x and y

    n1 <- nrow(x)
    n2 <- nrow(y)

    ## First, compute the In statistic

    sum.1 <- sum(npksum(txdat=x,
                        bws=bw.x,
                        leave.one.out=TRUE,
                        bandwidth.divide=TRUE)$ksum)
    
    sum.2 <- sum(npksum(txdat=y,
                        bws=bw.y,
                        leave.one.out=TRUE,
                        bandwidth.divide=TRUE)$ksum)
    
    sum.3 <- sum(npksum(txdat=x,
                        exdat=y,
                        bws=bw.x,
                        leave.one.out=FALSE,
                        bandwidth.divide=TRUE)$ksum)

    ## sum.4 and sum.3 are identical...
    
    In <- sum.1/(n1*(n1-1))+sum.2/(n2*(n2-1))-2*sum.3/(n1*n2)

    ## Next, compute sigma^2_n

    sum.1 <- sum(npksum(txdat=x,
                        bws=bw.x,
                        kernel.pow=2,
                        leave.one.out=TRUE,
                        bandwidth.divide=TRUE)$ksum)
    
    sum.2 <- sum(npksum(txdat=y,
                        bws=bw.y,
                        kernel.pow=2,
                        leave.one.out=TRUE,
                        bandwidth.divide=TRUE)$ksum)
    
    sum.3 <- sum(npksum(txdat=x,
                        exdat=y,
                        bws=bw.x,
                        kernel.pow=2,
                        leave.one.out=FALSE,
                        bandwidth.divide=TRUE)$ksum)
    
    ## sum.4 and sum.3 are identical

    sigma2.n<- 2*(sum.1/(n1^2*(n1-1)^2)+sum.2/(n2^2*(n2-1)^2)+2*sum.3/(n1^2*n2^2))

    ## Finally, compute Tn, the standardized statistic

    Tn <- In/sqrt(sigma2.n)
    
    return(list(Tn=Tn,In=In))
    
  } ## End of test statistic

  ## Now write a bootstrap function for the test statistic
  
  teststat.boot <- function(x,y,bw.x,bw.y) {
    n1 <- nrow(x)
    n2 <- nrow(y)
    ## Resample from pooled data
    z <- data.frame(rbind(x,y))
    x.bootstrap <- data.frame(z[sample.int(nrow(z), size = n1, replace = TRUE), ])      
    y.bootstrap <- data.frame(z[sample.int(nrow(z), size = n2, replace = TRUE), ])      
    output.boot <- teststat(x.bootstrap,y.bootstrap,bw.x,bw.y)
    return(list(Tn=output.boot$Tn,
                In=output.boot$In))
  }
  
  Tn.vector <- numeric(boot.num)
  In.vector <- numeric(boot.num)

  progress <- .np_progress_begin("Bootstrap replications", total = boot.num, surface = "bootstrap")

  for (i in seq_len(boot.num)) {
    output.boot <- teststat.boot(x,y,bw.x,bw.y)
    Tn.vector[i] <- output.boot$Tn
    In.vector[i] <- output.boot$In
    progress <- .np_progress_step(progress, done = i)
  }

  progress <- .np_progress_end(progress)

  ## Compute the test statistic
  
  output <- teststat(x,y,bw.x,bw.y)
  
  ## Compute empirical P-values - the number of resampled statistics
  ## more extreme than the original statistic
  
  Tn.P <- mean(Tn.vector > output$Tn)
  In.P <- mean(In.vector > output$In)
  
  ## Restore seed

  .np_seed_exit(seed.state)
  
  deneqtest(Tn=output$Tn,
            In=output$In,
            Tn.bootstrap=Tn.vector,
            In.bootstrap=In.vector,                
            Tn.P=Tn.P,
            In.P=In.P,
            boot.num=boot.num)
  
}

gsl.bs <- function(...) UseMethod("gsl.bs")

gsl.bs.default <- function(x,
                           degree=3,
                           nbreak=2,
                           deriv=0,
                           x.min=NULL,
                           x.max=NULL,
                           intercept=FALSE,
                           knots=NULL,
                           ...) {

  x <- as.vector(x)
  ## Some error checking

  if (length(degree) != 1L || is.na(degree) || !is.finite(degree) ||
      degree != floor(degree))
    stop(" degree must be a positive integer")
  if (length(deriv) != 1L || is.na(deriv) || !is.finite(deriv) ||
      deriv != floor(deriv))
    stop(" deriv must be a non-negative integer")
  if (length(nbreak) != 1L || is.na(nbreak) || !is.finite(nbreak) ||
      nbreak != floor(nbreak))
    stop(" nbreak must be an integer at least 2")
  if(degree <= 0) stop(" degree must be a positive integer")
  if(deriv < 0) stop(" deriv must be a non-negative integer")
  if(deriv > degree +1 ) stop(" deriv must be smaller than degree plus 2") 
  if(nbreak <= 1) stop(" nbreak must be at least 2")

  if(!is.null(knots)) {
    if (length(knots) < 2L)
      stop(" knots must contain at least two values")
    if (anyNA(knots) || any(!is.finite(knots)))
      stop(" knots must contain finite non-missing values")
  }

#  if(!is.null(knots)) nbreak <- length(knots)
  if(!is.null(knots)&&length(knots)!=nbreak) {
    nbreak <- length(knots)
    .np_warning(paste(" nbreak and knots vector do not agree: resetting nbreak to", nbreak))
  }

  ## For evaluation (newx) must use min/max for x unless otherwise
  ## specified - check that mix < max


  if(!is.null(x.min) && !is.null(x.max)) if(x.min >= x.max) stop(" x.min must be less than x.max")
  if(is.null(x.min)) {
			x.min <- min(x)
			ol <- FALSE
	}
	else {
			ol <- x < x.min 
	}
  if(is.null(x.max)) {
			x.max <- max(x)
			or <- FALSE
	}
	else {
			or <- x > x.max 
	}

  ## We check whether the Xs are within knots. It seems that if 'knots'
  ## is given, we don't require x.min and x.max, otherwise, we check
  ## the boundary.

	if(!is.null(knots)) {
			ol <- ol | x < min(knots)
			or <- or | x > max(knots)
	}

	outside <- ol | or 

	if(any(outside)){
			.np_warning("some 'x' values beyond boundary knots may cause ill-conditioned bases")
			ord <- degree + 1
			derivs<- deriv:degree
			if(ord == deriv) 
					scalef = 1
			else
					scalef <- gamma(1L:(ord-deriv))
			B <- array(0, c(length(x), nbreak+degree-1))
			
			if(any(ol) && (ord>deriv)){ #ord <=deriv,  derivatives should be zero..#
					k.pivot <- x.min
					if(degree == deriv){
							xl <- cbind(1)
					}
					else {
							xl <- cbind(1, outer(x[ol]-k.pivot, 1L:(degree-deriv),  "^"))
					}
					tt <- bs.des(rep(k.pivot, ord-deriv), degree, nbreak, deriv=derivs, x.min, x.max, knots)  
					B[ol, ] <- xl %*% (tt/scalef)
			}
			if(any(or) && (ord>deriv)){
					k.pivot <- x.max
					if(degree == deriv){
							xr <- cbind(1)
					}
					else {
							xr <- cbind(1, outer(x[or]-k.pivot, 1L:(degree-deriv),  "^"))
					}
					tt <- bs.des(rep(k.pivot, ord-deriv), degree, nbreak,deriv= derivs, x.min, x.max, knots)
					B[or, ] <- xr %*% (tt/scalef)
			}
			inside <- !outside
			if(any(inside))
					B[inside, ] <- bs.des(x[inside], degree, nbreak, rep(deriv, length(x[inside])), x.min, x.max, knots)
	}
	else {
			B <- bs.des(x, degree, nbreak, rep(deriv, length(x)), x.min, x.max,knots)
	}


	if(!intercept) B <- B[,-1,drop=FALSE]

	attr(B, "degree") <- degree
	attr(B, "nbreak") <- nbreak
	attr(B, "deriv") <- deriv
	attr(B, "x.min") <- x.min
	attr(B, "x.max") <- x.max
	attr(B, "intercept") <- intercept
	attr(B, "knots") <- knots
	# Keep "gsl.bs" as a compatibility superclass while crs remains the
	# public owner of predict.gsl.bs.
	attr(B, "class") <- c("np_gsl.bs","gsl.bs","matrix")

	return(B)

}

bs.des     <- function(x,
											 degree=3,
											 nbreak=2,
											 deriv=integer(length(x)),   ## now deriv is a vector to represent the derivivative for each element of x 
											 x.min=NULL,
											 x.max=NULL,
											 knots=NULL,
											 ...) {

		x <- as.vector(x)
		n <- length(x)
		deriv <- as.vector(deriv)
    if (length(deriv) != n)
      stop(" deriv must have length equal to x")
    if (anyNA(deriv) || any(!is.finite(deriv)) || any(deriv != floor(deriv)))
      stop(" deriv must contain finite non-missing integers")
    if (any(deriv < 0))
      stop(" deriv must be non-negative")
    if (any(deriv > degree + 1))
      stop(" deriv must be smaller than degree plus 2")
    if(!is.null(knots)) {
      if (length(knots) != nbreak)
        stop(" knots length must match nbreak")
      if (anyNA(knots) || any(!is.finite(knots)))
        stop(" knots must contain finite non-missing values")
    }

		if(!is.null(x.min) && !is.null(x.max)) if(x.min >= x.max) stop(" x.min must be less than x.max")
		if(is.null(x.min)) x.min <- min(x)
		if(is.null(x.max)) x.max <- max(x)

		## 0 == don't use user supplied knots, 1 = use

		knots.int <- if (is.null(knots)) 0 else 1

		ncol <- nbreak+degree-1;

		if(all(deriv==0)) {

				B <- .Call("C_gsl_bspline",
									 as.double(x),
									 as.integer(degree),
									 as.integer(nbreak),
									 as.double(x.min),
									 as.double(x.max),
									 as.double(knots),
									 as.integer(knots.int),
									 PACKAGE="np" )

		} else {

				B <- .Call("C_gsl_bspline_deriv",
									 as.double(x),
									 as.integer(degree),
									 as.integer(nbreak),
									 as.integer(deriv),
									 as.double(x.min),
									 as.double(x.max),
									 as.double(knots),
									 as.integer(knots.int),
									 PACKAGE="np" )

		}

		return(B)

}

.np_predict_gsl_bs_impl <- function(object,
											 newx=NULL,
											 ...) {

		newx.ind <- NULL

		if(is.null(newx))

				## If No new data provided, return sample fit.
				B <- object

		else{

				x.min <- attr(object, "x.min")
				x.max <- attr(object, "x.max")

				newx <- as.numeric(newx)

#				if(min(newx)<x.min || max(newx)>x.max) {
#						.np_warning(" evaluation data lies beyond spline support: resetting those values to min/max")
#						newx[newx < x.min] <- x.min
#						newx[newx > x.max] <- x.max
#						newx.ind <- sort(c(which(newx < x.min),which(newx > x.max)))
#				}

				B <- gsl.bs(newx,
										degree=attr(object, "degree"),
										nbreak=attr(object, "nbreak"),
										deriv=attr(object, "deriv"),
										intercept=attr(object, "intercept"),
										knots=attr(object, "knots"),
										x.min=x.min,
										x.max=x.max)

		}

		attr(B, "newx") <- newx
		attr(B, "newx.trimmed") <- newx.ind

		return(B)

}

predict.np_gsl.bs <- function(object,
														 newx=NULL,
														 ...) {
		.np_predict_gsl_bs_impl(object = object, newx = newx, ...)
}

predict.gsl.bs <- function(object,
													 newx=NULL,
													 ...) {
		.np_predict_gsl_bs_impl(object = object, newx = newx, ...)
}

np.pairs <- function(y_vars, y_dat, ...) {
  if (missing(y_vars) || missing(y_dat))
    stop("'y_vars' and 'y_dat' are required")
  if (!is.data.frame(y_dat))
    y_dat <- as.data.frame(y_dat)
  if (is.null(names(y_vars)))
    names(y_vars) <- y_vars
  if (any(!y_vars %in% names(y_dat)))
    stop("all elements of 'y_vars' must be column names in 'y_dat'")

  pair_names <- expand.grid(y_vars, y_vars, stringsAsFactors = FALSE)
  pair_kerns <- lapply(seq_len(nrow(pair_names)), function(i) {
    y1 <- pair_names[i, 1]
    y2 <- pair_names[i, 2]
    if (y1 == y2) {
      npudens(tdat = y_dat[, y1], ...)
    } else {
      npreg(tydat = y_dat[, y2], txdat = y_dat[, y1], residuals = TRUE, ...)
    }
  })
  list(y_vars = y_vars, pair_names = pair_names, pair_kerns = pair_kerns)
}

np.pairs.plot <- function(pair_list) {
  if (length(pair_list) < 3)
    stop("pair_list must be created by np.pairs")

  pair_names <- pair_list[["pair_names"]]
  pair_kerns <- pair_list[["pair_kerns"]]
  y_vars <- pair_list[["y_vars"]]
  y_labels <- names(y_vars)

  oldpar <- .np_plot_capture_par(c("mfrow", "mar"))
  on.exit(.np_plot_restore_par(oldpar), add = TRUE)
  par(mfrow = c(length(y_vars), length(y_vars)), mar = c(4, 4, 2, 0))
  for (i in seq_len(nrow(pair_names))) {
    y1 <- pair_names[i, 1]
    y2 <- pair_names[i, 2]
    y1_label <- y_labels[match(y1, y_vars)]
    y2_label <- y_labels[match(y2, y_vars)]
    pair_i <- pair_kerns[[i]]
    if (y1 == y2) {
      ydat <- pair_i$eval[, 1]
      y_pred <- pair_i$dens
      y_ord <- order(ydat)
      y_pred <- y_pred[y_ord]
      ydat <- ydat[y_ord]
      plot.new()
      plot.window(xlim = range(ydat, na.rm = TRUE),
                  ylim = range(y_pred, na.rm = TRUE))
      polygon(c(ydat, rev(ydat)),
              c(y_pred, y_pred * 0),
              col = grey(0, 0.25), border = NA)
      axis(1)
      axis(2)
      mtext("Density Estimate", 2, line = 2.5, cex = 1)
      mtext(y1_label, 1, line = 2.5, cex = 1)
    } else {
      xdat <- pair_i$eval[, 1]
      y_pred <- pair_i$mean
      ydat <- y_pred + pair_i$resid
      x_ord <- order(xdat)
      y_pred <- y_pred[x_ord]
      ydat <- ydat[x_ord]
      xdat <- xdat[x_ord]
      plot.new()
      plot.window(xlim = range(xdat, na.rm = TRUE),
                  ylim = range(ydat, na.rm = TRUE))
      points(xdat, ydat, col = grey(0, 0.5), pch = 16, cex = 1.5)
      lines(xdat, y_pred, lwd = 2, col = rgb(0, 0, 1, 1))
      axis(1)
      axis(2)
      mtext(y2_label, 2, line = 2.5, cex = 1)
      mtext(y1_label, 1, line = 2.5, cex = 1)
    }
  }
  invisible(NULL)
}

with_plot_state <- function(expr) {
  pdf(file = tempfile(fileext = ".pdf"))
  on.exit(dev.off(), add = TRUE)

  before <- list(
    usr = par("usr"),
    mfrow = par("mfrow"),
    mar = par("mar"),
    new = par("new")
  )

  value <- suppressWarnings(force(expr))

  after <- list(
    usr = par("usr"),
    mfrow = par("mfrow"),
    mar = par("mar"),
    new = par("new")
  )

  list(value = value, before = before, after = after)
}

expect_plot_device_clean <- function(state, expect_mar_restore = FALSE, info = NULL) {
  expect_false(isTRUE(state$after$new), info = info)
  expect_equal(state$after$mfrow, state$before$mfrow, info = info)
  if (isTRUE(expect_mar_restore))
    expect_equal(state$after$mar, state$before$mar, info = info)
  expect_false(isTRUE(all.equal(state$after$usr, state$before$usr)), info = info)
}

test_that("plot methods keep device coordinates active while restoring layout state", {
  set.seed(20260310)
  n <- 60
  x <- runif(n)
  x2 <- runif(n)
  z <- runif(n)
  y <- sin(2 * pi * x) + 0.3 * z + rnorm(n, sd = 0.08)

  xdat <- data.frame(x = x)
  ydat <- data.frame(y = y)
  idxdat <- data.frame(y = y, x = x, x2 = x2)

  rfit <- npreg(
    txdat = xdat,
    tydat = y,
    bws = npregbw(xdat = xdat, ydat = y, bws = 0.25, bandwidth.compute = FALSE)
  )
  ubw <- npudensbw(dat = xdat, bws = 0.25, bandwidth.compute = FALSE)
  dbw <- npudistbw(dat = xdat, bws = 0.25, bandwidth.compute = FALSE)
  cbw <- npcdensbw(xdat = xdat, ydat = ydat, bws = c(0.30, 0.30), bandwidth.compute = FALSE)
  cdbw <- npcdistbw(xdat = xdat, ydat = ydat, bws = c(0.30, 0.30), bandwidth.compute = FALSE)
  pbw <- npplregbw(
    xdat = data.frame(x = z),
    zdat = data.frame(z = x),
    ydat = y,
    bws = matrix(c(0.30, 0.30), nrow = 2),
    bandwidth.compute = FALSE
  )
  sbw <- npindexbw(
    y ~ x + x2,
    data = idxdat,
    bws = c(1, 0.30, 0.30),
    bandwidth.compute = FALSE,
    method = "ichimura"
  )
  scbw <- npscoefbw(
    xdat = data.frame(x = x),
    zdat = data.frame(z = z),
    ydat = y,
    bws = 0.20,
    bandwidth.compute = FALSE
  )

  cases <- list(
    list(
      label = "npregression fit",
      expr = quote(plot(rfit))
    ),
    list(
      label = "npudensbw",
      expr = quote(plot(ubw, xdat = xdat))
    ),
    list(
      label = "npudistbw",
      expr = quote(plot(dbw, xdat = xdat))
    ),
    list(
      label = "npcdensbw",
      expr = quote(plot(cbw, xdat = xdat, ydat = ydat, perspective = FALSE))
    ),
    list(
      label = "npcdistbw",
      expr = quote(plot(cdbw, xdat = xdat, ydat = ydat, perspective = FALSE))
    ),
    list(
      label = "npplregbw",
      expr = quote(plot(pbw, xdat = data.frame(x = z), ydat = y, zdat = data.frame(z = x), perspective = FALSE))
    ),
    list(
      label = "npindexbw",
      expr = quote(plot(sbw, xdat = data.frame(x = x, x2 = x2), ydat = y, perspective = FALSE))
    ),
    list(
      label = "npscoefbw",
      expr = quote(plot(scbw, xdat = data.frame(x = x), ydat = y, zdat = data.frame(z = z), perspective = FALSE))
    )
  )

  for (case in cases) {
    state <- with_plot_state(eval(case$expr, envir = environment()))
    expect_plot_device_clean(state, info = case$label)
  }
})

test_that("np.pairs restores layout parameters without resetting plot coordinates", {
  set.seed(20260310 + 1L)
  dat <- data.frame(a = rnorm(40), b = runif(40), c = rnorm(40))

  pair_list <- np.pairs(y_vars = c("a", "b", "c"), y_dat = dat)
  state <- with_plot_state(np.pairs.plot(pair_list))

  expect_plot_device_clean(state, expect_mar_restore = TRUE, info = "np.pairs")
})

qregression <- 
    function(bws, xeval, tau, quantile, quanterr = NA, quantgrad = NA, quantgerr = NA, ntrain, trainiseval = FALSE, gradients = FALSE,
             timing = NA, total.time = NA, optim.time = NA, fit.time = NA){

        if (missing(bws) || missing(xeval) || missing(tau) || missing(quantile) || missing(ntrain))
            stop("improper invocation of qregression constructor")

        d <- list(
            xbw = bws$xbw,
            ybw = bws$ybw,
            bws = bws,
            xnames = bws$xnames,
            ynames = bws$ynames,
            nobs = nrow(xeval),
            xndim = bws$xndim,
            yndim = bws$yndim,
            xnord = bws$xnord,
            xnuno = bws$xnuno,
            xncon = bws$xncon,
            ynord = bws$ynord,
            ynuno = bws$ynuno,
            yncon = bws$yncon,
            pscaling = bws$pscaling,
            ptype = bws$ptype,
            pcxkertype = bws$pcxkertype,
            puxkertype = bws$puxkertype,
            poxkertype = bws$poxkertype,
            pcykertype = bws$pcykertype,
            puykertype = bws$puykertype,
            poykertype = bws$poykertype,
            xeval = xeval,
            tau = tau,
            quantile = quantile,
            quanterr = quanterr,
            quantgrad = quantgrad,
            quantgerr = quantgerr,
            ntrain = ntrain,
            trainiseval = trainiseval,
            gradients = gradients,
            timing = timing, total.time = total.time,
            optim.time = optim.time, fit.time = fit.time)

        class(d) <- "qregression"

        return(d)
    }

print.qregression <- function(x, digits=NULL, ...){
  cat("\nQuantile regression data: ", x$ntrain, " training points,",
      if (x$trainiseval) "" else paste(" and ", x$nobs, " evaluation points,\n", sep=""),
      " in ", x$xndim + x$yndim, " variable(s)",
      "\n(", x$yndim, " dependent variable(s), and ", x$xndim, " explanatory variable(s))\n\n",
      sep="")
  if (length(x$tau) > 1L)
    cat("Tau values:", paste(format(x$tau, trim = TRUE), collapse = ", "), "\n\n")
  print(matrix(x$ybw,ncol=x$yndim,dimnames=list(paste("Dep. Var. ",x$pscaling,":",sep=""),x$ynames)))

  print(matrix(x$xbw,ncol=x$xndim,dimnames=list(paste("Exp. Var. ",x$pscaling,":",sep=""),x$xnames)))

  cat(genRegEstStr(x))
  cat(genBwKerStrs(x$bws))
  
  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

fitted.qregression <- function(object, ...){
 object$quantile 
}
quantile.qregression <- function(x, ...){ x$quantile }
.npqreg_predict_newdata_to_exdat <- function(object, newdata) {
  if (is.data.frame(newdata) && !is.null(names(newdata))) {
    xnames <- object$bws$xnames
    if (length(xnames)) {
      missing.names <- setdiff(xnames, names(newdata))
      if (length(missing.names))
        stop(sprintf(
          "newdata must contain columns: %s",
          paste(shQuote(xnames), collapse = ", ")
        ), call. = FALSE)
      return(newdata[, xnames, drop = FALSE])
    }
  }
  newdata
}

predict.qregression <- function(object, se.fit = FALSE, ...) {
  se.fit <- npValidateScalarLogical(se.fit, "se.fit")
  dots <- list(...)
  has.formula.route <- !is.null(object$bws$formula)

  if (!is.null(dots$exdat) && !is.null(dots$newdata))
    dots$newdata <- NULL
  if (has.formula.route) {
    if (!is.null(dots$exdat)) {
      dots$newdata <- .npqreg_predict_newdata_to_exdat(object, dots$exdat)
      dots$exdat <- NULL
    } else if (!is.null(dots$newdata)) {
      dots$newdata <- .npqreg_predict_newdata_to_exdat(object, dots$newdata)
    }
  } else {
    if (!is.null(dots$newdata) && is.null(dots$exdat))
      dots$newdata <- .npqreg_predict_newdata_to_exdat(object, dots$newdata)
    if (is.null(dots$exdat) && !is.null(dots$newdata)) {
      dots$exdat <- dots$newdata
      dots$newdata <- NULL
    }
  }
  if (is.null(dots$tau))
    dots$tau <- object$tau
  tr <- do.call(npqreg, c(list(bws = object$bws), dots))
  if(se.fit)
    return(list(fit = fitted(tr), se.fit = se(tr), 
                df = tr$nobs, residual.scale = NA))
  else
    return(fitted(tr))
}

se.qregression <- function(x) { x$quanterr }
gradients.qregression <- function(x, errors = FALSE, ...) {
  errors <- npValidateScalarLogical(errors, "errors")
  gout <- if (!errors) x$quantgrad else x$quantgerr
  if (is.null(gout) || (length(gout) == 1L && is.logical(gout) && is.na(gout)))
    stop(if (!errors)
      "gradients are not available: fit the model with gradients=TRUE"
    else
      "gradient standard errors are not available: fit the model with gradients=TRUE")
  gout
}

summary.qregression <- function(object, ...) {
  cat("\nQuantile Regression Data: ", object$ntrain, " training points,",
      if (object$trainiseval) "" else paste(" and ", object$nobs, " evaluation points,\n", sep=""),
      " in ", object$xndim + object$yndim, " variable(s)",
      "\n(", object$yndim, " dependent variable(s), and ", object$xndim, " explanatory variable(s))\n\n",
      sep="")

  cat(genOmitStr(object))
  if (length(object$tau) > 1L)
    cat("Tau values:", paste(format(object$tau, trim = TRUE), collapse = ", "), "\n\n")

  print(matrix(object$ybw,ncol=object$yndim,dimnames=list(paste("Dep. Var. ",object$pscaling,":",sep=""),object$ynames)))

  print(matrix(object$xbw,ncol=object$xndim,dimnames=list(paste("Exp. Var. ",object$pscaling,":",sep=""),object$xnames)))

  cat(genRegEstStr(object))
  cat(genBwKerStrs(object$bws))
  cat(genTimingStr(object))
  
  cat("\n\n")
}

npplreg <-
  function(bws, ...){
    args <- list(...)

    if (!missing(bws)){
      if (is.recursive(bws)){
        if (!is.null(bws$formula) && is.null(args$txdat))
          UseMethod("npplreg",bws$formula)
        else if (!is.null(bws$call) && is.null(args$txdat))
          UseMethod("npplreg",bws$call)
        else if (!is.call(bws))
          UseMethod("npplreg",bws)
        else
          UseMethod("npplreg",NULL)
      } else {
        UseMethod("npplreg", NULL)
      }
    } else {
      UseMethod("npplreg", NULL)
    }
  }

npplreg.formula <-
  function(bws, data = NULL, newdata = NULL, y.eval = FALSE, ...){
    
    tt <- terms(bws)
    tt.xf <- bws$xterms
    
    m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)
    tmf.xf <- tmf <- bws$call[c(1,m)]
    
    tmf[[1]] <- as.name("model.frame")
    tmf.xf[[1]] <- as.name("model.frame")

    tmf.xf[["formula"]] <- tt.xf
    tmf[["formula"]] <- tt

    mf.args <- as.list(tmf)[-1L]
    mf.xf.args <- as.list(tmf.xf)[-1L]
    umf <- tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))
    tmf.xf <- do.call(stats::model.frame, mf.xf.args, envir = environment(tt.xf))
    
    response.name <- attr(tmf, "names")[attr(attr(tmf, "terms"), "response")]
    tydat <- model.response(tmf)
    txdat <- tmf.xf
    tzdat <- tmf[, bws$chromoly[[3]], drop = FALSE]

    has.eval <- !is.null(newdata)
    if (has.eval) {
      npValidateNewdataFormula(newdata, tt.xf, include.response = FALSE)
      if (!y.eval){
        npValidateNewdataFormula(newdata, tt, include.response = FALSE)
        tt <- delete.response(tt)
        
        bronze <- lapply(bws$chromoly, paste, collapse = " + ")
        formula.xz <- terms(as.formula(paste(" ~ ",bronze[[2]], " + ",bronze[[3]]),
                                       env = environment(bws$formula)))

        orig.ts <- .np_terms_ts_mask(terms_obj = formula.xz, data = newdata)

        arguments.mfx <- bws$chromoly[[2]]
        arguments.mf <- bws$chromoly[[3]]

        if(all(orig.ts)){
          arguments <- (as.list(attr(formula.xz, "variables"))[-1])
          attr(tt, "predvars") <- bquote(.(as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments)))))[,.(match(arguments.mf,arguments)),drop = FALSE])
          attr(tt.xf, "predvars") <- bquote(.(as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments)))))[,.(match(arguments.mfx,arguments)),drop = FALSE])
        }else if(any(orig.ts)){
          arguments <- (as.list(attr(formula.xz, "variables"))[-1])
          arguments.normal <- arguments[which(!orig.ts)]
          arguments.timeseries <- arguments[which(orig.ts)]

          ix <- sort(c(which(orig.ts),which(!orig.ts)),index.return = TRUE)$ix
          attr(tt, "predvars") <- bquote((.(as.call(c(quote(cbind),as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments.timeseries)))),arguments.normal,check.rows = TRUE)))[,.(ix)])[,.(match(arguments.mf,arguments)),drop = FALSE])
          attr(tt.xf, "predvars") <- bquote((.(as.call(c(quote(cbind),as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments.timeseries)))),arguments.normal,check.rows = TRUE)))[,.(ix)])[,.(match(arguments.mfx,arguments)),drop = FALSE])
        }else{
          attr(tt, "predvars") <- attr(tt, "variables")
          attr(tt.xf, "predvars") <- attr(tt.xf, "variables")
        }
          
      }
      
      if (y.eval)
        npValidateNewdataFormula(newdata, tt, include.response = TRUE)
      umf.args <- list(formula = tt, data = newdata)
      umf <- do.call(stats::model.frame, umf.args, envir = parent.frame())
      emf <- umf
      emf.xf.args <- list(formula = tt.xf, data = newdata)
      emf.xf <- do.call(stats::model.frame, emf.xf.args, envir = parent.frame())
      
      if (y.eval)
        eydat <- model.response(emf)

      exdat <- emf.xf
      ezdat <- emf[, bws$chromoly[[3]], drop = FALSE]
    }

    pl.args <- list(txdat = txdat, tydat = tydat, tzdat = tzdat)
    if (has.eval) {
      pl.args$exdat <- exdat
      pl.args$ezdat <- ezdat
      if (y.eval)
        pl.args$eydat <- eydat
    }
    pl.args$bws <- bws
    ev <- do.call(npplreg, c(pl.args, list(...)))

    if (length(response.name) == 1L && !is.na(response.name) && nzchar(response.name)) {
      if (!is.null(ev$bws))
        ev$bws$ynames <- response.name
    }

    ev$omit <- attr(umf,"na.action")
    ev$rows.omit <- as.vector(ev$omit)
    ev$nobs.omit <- length(ev$rows.omit)

    ev$mean <- napredict(ev$omit, ev$mean)
    ev$merr <- napredict(ev$omit, ev$merr)

    if(ev$residuals){
        ev$resid <- naresid(ev$omit, ev$resid)
    }    
    return(ev)
  }

npplreg.call <-
  function(bws, ...) {
    npplreg(txdat = .np_eval_bws_call_arg(bws, "xdat"),
            tydat = .np_eval_bws_call_arg(bws, "ydat"),
            tzdat = .np_eval_bws_call_arg(bws, "zdat"),
            bws = bws, ...)
  }

.np_plreg_fit_progress_targets <- function(xnames) {
  c("E[y|z]", sprintf("E[%s|z]", xnames), "final partially linear solve")
}

.np_plreg_fit_progress_begin <- function(xnames, handoff = FALSE) {
  state <- .np_progress_begin(
    "Fitting partially linear regression",
    total = length(.np_plreg_fit_progress_targets(xnames)),
    surface = "bandwidth"
  )

  if (isTRUE(handoff)) {
    state <- .np_progress_show_now(
      state = state,
      done = 0L,
      detail = paste("starting", .np_plreg_fit_progress_targets(xnames)[1L])
    )
  }

  state
}

.np_plot_plreg_local_fit <-
  function(bws,
           xdat,
           ydat,
           zdat,
           exdat,
           ezdat) {
    activity <- .np_plot_activity_begin("Computing partially linear plot fit")
    on.exit(.np_plot_activity_end(activity), add = TRUE)

    xdat <- toFrame(xdat)
    zdat <- toFrame(zdat)

    keep.rows <- rep_len(TRUE, nrow(xdat))
    rows.omit <- attr(na.omit(data.frame(xdat, ydat, zdat)), "na.action")
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Training data has no rows without NAs")

    xdat <- xdat[keep.rows, , drop = FALSE]
    ydat <- ydat[keep.rows]
    zdat <- zdat[keep.rows, , drop = FALSE]

    no.exz <- missing(exdat)
    if (!no.exz) {
      exdat <- toFrame(exdat)
      ezdat <- toFrame(ezdat)

      keep.eval <- rep_len(TRUE, nrow(exdat))
      rows.omit <- attr(na.omit(data.frame(exdat, ezdat)), "na.action")
      if (length(rows.omit) > 0L)
        keep.eval[as.integer(rows.omit)] <- FALSE

      if (!any(keep.eval))
        stop("Evaluation data has no rows without NAs")

      exdat <- exdat[keep.eval, , drop = FALSE]
      ezdat <- ezdat[keep.eval, , drop = FALSE]
    }

    if (is.factor(ydat)) {
      tmp.ty <- adjustLevels(data.frame(ydat), bws$bw$yzbw$ydati)
      tmp.ty <- (bws$bw$yzbw$ydati$all.dlev[[1L]])[as.integer(tmp.ty)]
    } else {
      tmp.ty <- as.double(ydat)
    }

    local.direct <- isTRUE(identical(bws$type, "generalized_nn"))

    reg_mean <- function(regbw, ytrain, zeval = NULL) {
      args <- list(
        bws = regbw,
        txdat = zdat,
        tydat = ytrain,
        local.mode = local.direct
      )
      if (!is.null(zeval))
        args$exdat <- zeval
      as.vector(do.call(.np_regression_direct, args)$mean)
    }

    yhat.train <- reg_mean(regbw = bws$bw$yzbw, ytrain = ydat)
    resy <- tmp.ty - yhat.train

    if (!no.exz)
      yhat.eval <- reg_mean(regbw = bws$bw$yzbw, ytrain = ydat, zeval = ezdat)

    ntrain <- nrow(xdat)
    neval <- if (no.exz) ntrain else nrow(exdat)
    p <- ncol(xdat)
    resx <- matrix(0.0, nrow = ntrain, ncol = p)
    resx.eval <- matrix(0.0, nrow = neval, ncol = p)

    for (j in seq_len(p)) {
      xhat.train <- reg_mean(regbw = bws$bw[[j + 1L]], ytrain = xdat[, j])

      if (is.factor(xdat[1L, j])) {
        tmp.dat <- adjustLevels(xdat[, j, drop = FALSE], bws$bw[[j + 1L]]$ydati)
        x.num.train <- (bws$bw[[j + 1L]]$ydati$all.dlev[[1L]])[as.integer(tmp.dat[, 1L])]
      } else {
        x.num.train <- as.double(xdat[, j])
      }
      resx[, j] <- x.num.train - xhat.train

      if (!no.exz) {
        xhat.eval <- reg_mean(regbw = bws$bw[[j + 1L]], ytrain = xdat[, j], zeval = ezdat)
        if (is.factor(xdat[1L, j])) {
          tmp.dat <- adjustLevels(exdat[, j, drop = FALSE], bws$bw[[j + 1L]]$ydati, allowNewCells = TRUE)
          x.num.eval <- (bws$bw[[j + 1L]]$ydati$all.dlev[[1L]])[as.integer(tmp.dat[, 1L])]
        } else {
          x.num.eval <- as.double(exdat[, j])
        }
        resx.eval[, j] <- x.num.eval - xhat.eval
      }
    }

    model <- lm(resy ~ resx - 1)
    B <- coef(model)

    train.fit <- as.vector(yhat.train + resx %*% B)
    Bvcov <- sum((tmp.ty - train.fit)^2) /
      (ntrain - ncol(xdat) - ncol(zdat)) *
      chol2inv(chol(t(model.matrix(model)) %*% model.matrix(model)))
    Berr <- sqrt(diag(Bvcov))

    RSQ <- RSQfunc(tmp.ty, train.fit)
    MSE <- MSEfunc(tmp.ty, train.fit)
    MAE <- MAEfunc(tmp.ty, train.fit)
    MAPE <- MAPEfunc(tmp.ty, train.fit)
    CORR <- CORRfunc(tmp.ty, train.fit)
    SIGN <- SIGNfunc(tmp.ty, train.fit)

    ply <- if (no.exz) {
      train.fit
    } else {
      as.vector(yhat.eval + resx.eval %*% B)
    }

    do.call(plregression, list(
      bws = bws,
      xcoef = B,
      xcoeferr = Berr,
      xcoefvcov = Bvcov,
      evalx = if (no.exz) xdat else exdat,
      evalz = if (no.exz) zdat else ezdat,
      mean = ply,
      ntrain = ntrain,
      trainiseval = no.exz,
      residuals = FALSE,
      xtra = c(RSQ, MSE, MAE, MAPE, CORR, SIGN)
    ))
  }


npplreg.plbandwidth <- 
  function(bws,
           txdat = stop("training data txdat missing"),
           tydat = stop("training data tydat missing"),
           tzdat = stop("training data tzdat missing"),
           exdat, eydat, ezdat, residuals = FALSE, ...){

    fit.start <- proc.time()[3]
    residuals <- npValidateScalarLogical(residuals, "residuals")
    dots <- list(...)
    fit.progress.handoff <- isTRUE(dots$.np_fit_progress_handoff)

    txdat = toFrame(txdat)
    tzdat = toFrame(tzdat)
    
    ## catch and destroy NA's, part 1
    keep.rows <- rep_len(TRUE, nrow(txdat))
    rows.omit <- attr(na.omit(data.frame(txdat,tydat,tzdat)), "na.action")
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Training data has no rows without NAs")

    txdat <- txdat[keep.rows,,drop = FALSE]
    tydat <- tydat[keep.rows]
    tzdat <- tzdat[keep.rows,,drop = FALSE]

    no.exz = missing(exdat)
    no.ey = missing(eydat)

    if (!no.exz){
      exdat = toFrame(exdat)
      ezdat = toFrame(ezdat)
      if (!no.ey)
        eydat = as.double(eydat)

      ## c& d NA's, part 2

      keep.eval <- rep_len(TRUE, nrow(exdat))
      eval.df <- data.frame(exdat, ezdat)
      if (!no.ey)
        eval.df <- data.frame(eval.df, eydat)
      rows.omit <- attr(na.omit(eval.df), "na.action")
      if (length(rows.omit) > 0L)
        keep.eval[as.integer(rows.omit)] <- FALSE

      if (!any(keep.eval))
        stop("Evaluation data has no rows without NAs")

      exdat <- exdat[keep.eval,,drop = FALSE]
      if (!no.ey)
        eydat <- eydat[keep.eval]
      ezdat <- ezdat[keep.eval,,drop = FALSE]
    }

    ## tmp.ty and tmp.ey are the numeric representations of tydat and eydat
    if (is.factor(tydat)){
      tmp.ty <- adjustLevels(data.frame(tydat), bws$bw$yzbw$ydati)
      tmp.ty <- (bws$bw$yzbw$ydati$all.dlev[[1]])[as.integer(tmp.ty)]
    } else {
      tmp.ty <- as.double(tydat)
    }

    if (!no.ey){
      if (is.factor(tydat)){
        tmp.ey <- adjustLevels(data.frame(eydat), bws$bw$yzbw$ydati)
        tmp.ey <- (bws$bw$yzbw$ydati$all.dlev[[1]])[as.integer(tmp.ey)]
      } else {
        tmp.ey <- as.double(eydat)
      }
    }
    
    ## y on z
    mmy = npreg(txdat = tzdat, tydat = tydat, bws = bws$bw$yzbw)

    resy <- tmp.ty - mmy$mean

    if (!no.exz)
      mmy.eval = npreg(txdat = tzdat, tydat = tydat, exdat = ezdat, bws = bws$bw$yzbw)

    
    ## x on z
    nrow = nrow(txdat)
    nrow.eval = (if (no.exz) 0 else nrow(exdat))
    ncol = ncol(txdat)
    B = double(ncol)
    resx = matrix(data = 0, nrow = nrow, ncol = ncol)
    resx.eval = matrix(data = 0, nrow = nrow.eval, ncol = ncol)
    fit.progress.targets <- .np_plreg_fit_progress_targets(names(txdat))
    fit.progress <- .np_plreg_fit_progress_begin(
      xnames = names(txdat),
      handoff = fit.progress.handoff
    )
    fit.progress.active <- TRUE
    on.exit({
      if (isTRUE(fit.progress.active))
        .np_progress_abort(fit.progress)
    }, add = TRUE)
    fit.progress <- .np_progress_step(
      fit.progress,
      done = 1L,
      detail = fit.progress.targets[1L]
    )

    for (i in seq_len(ncol)) {
      mm = npreg(txdat=tzdat, tydat=txdat[,i], bws = bws$bw[[i+1]])

      if (is.factor(txdat[1,i])){
        tmp.dat <- adjustLevels(txdat[,i, drop=FALSE], bws$bw[[i+1]]$ydati)
        resx[,i] <- (bws$bw[[i+1]]$ydati$all.dlev[[1]])[as.integer(tmp.dat[,1])] - mm$mean
      } else {
        resx[,i] <- txdat[,i] - mm$mean
      }

      if(!no.exz) {
        mm = npreg(txdat=tzdat, tydat=txdat[,i], exdat=ezdat, bws = bws$bw[[i+1]])

        if (is.factor(txdat[1,i])){
          tmp.dat <- adjustLevels(exdat[,i, drop=FALSE], bws$bw[[i+1]]$ydati)
          resx.eval[,i] <- (bws$bw[[i+1]]$ydati$all.dlev[[1]])[as.integer(tmp.dat[,1])] - mm$mean
        } else {
          resx.eval[,i] <- exdat[,i] - mm$mean
        }
      }

      fit.progress <- .np_progress_step(
        fit.progress,
        done = i + 1L,
        detail = fit.progress.targets[i + 1L]
      )
    }

    fit.progress <- .np_progress_step(
      fit.progress,
      done = ncol + 2L,
      detail = fit.progress.targets[ncol + 2L]
    )

    B = coef((model = lm(resy ~ resx - 1)))

    ## computes the standard errors of B using the model matrix
    ## and the MSE of the training data predictions

    Bvcov = sum((tmp.ty-(mmy$mean + resx  %*% B))^2)/
      (dim(txdat)[1]-dim(txdat)[2]-dim(tzdat)[2])*
      chol2inv(chol(t(model.matrix(model))%*%model.matrix(model)))

    Berr = sqrt(diag(Bvcov))

    train.ply =  mmy$mean + resx %*% B
    ply = if (no.exz) train.ply else mmy.eval$mean + resx.eval %*% B

    if (!no.ey) {
      RSQ = RSQfunc(tmp.ey, ply)
      MSE = MSEfunc(tmp.ey, ply)
      MAE = MAEfunc(tmp.ey, ply)
      MAPE = MAPEfunc(tmp.ey, ply)
      CORR = CORRfunc(tmp.ey, ply)
      SIGN = SIGNfunc(tmp.ey, ply)

    } else {
      RSQ = RSQfunc(tmp.ty, train.ply)
      MSE = MSEfunc(tmp.ty, train.ply)
      MAE = MAEfunc(tmp.ty, train.ply)
      MAPE = MAPEfunc(tmp.ty, train.ply)
      CORR = CORRfunc(tmp.ty, train.ply)
      SIGN = SIGNfunc(tmp.ty, train.ply)
    }

    ev.args <- list(
      bws = bws,
      xcoef = B,
      xcoeferr = Berr,
      xcoefvcov = Bvcov,
      evalx = if (no.exz) txdat else exdat,
      evalz = if (no.exz) tzdat else ezdat,
      mean = ply,
      ntrain = nrow,
      trainiseval = no.exz,
      residuals = residuals,
      xtra = c(RSQ, MSE, MAE, MAPE, CORR, SIGN)
    )
    if (residuals)
      ev.args$resid <- tmp.ty - train.ply
    ev <- do.call(plregression, ev.args)

    fit.elapsed <- proc.time()[3] - fit.start
    optim.time <- if (!is.null(bws$total.time) && is.finite(bws$total.time)) as.double(bws$total.time) else NA_real_
    total.time <- fit.elapsed + (if (is.na(optim.time)) 0.0 else optim.time)
    ev$timing <- bws$timing
    ev$total.time <- total.time
    ev$optim.time <- optim.time
    ev$fit.time <- fit.elapsed
    ev$nomad.time <- if (!is.null(bws$nomad.time) && is.finite(bws$nomad.time)) as.double(bws$nomad.time) else NA_real_
    ev$powell.time <- if (!is.null(bws$powell.time) && is.finite(bws$powell.time)) as.double(bws$powell.time) else NA_real_

    
    ev$call <- match.call(expand.dots = FALSE)
    environment(ev$call) <- parent.frame()
    fit.progress <- .np_progress_end(fit.progress)
    fit.progress.active <- FALSE
    return(ev)
  }


npplreg.default <- function(bws, txdat, tydat, tzdat, nomad = FALSE, ...) {
  sc <- sys.call()
  sc.names <- names(sc)
  nomad <- npValidateScalarLogical(nomad, "nomad")

  ## here we check to see if the function was called with tdat =
  ## if it was, we need to catch that and map it to dat =
  ## otherwise the call is passed unadulterated to npudensbw

  bws.named <- any(sc.names == "bws")
  txdat.named <- any(sc.names == "txdat")
  tydat.named <- any(sc.names == "tydat")
  tzdat.named <- any(sc.names == "tzdat")

  no.bws <- missing(bws)
  no.txdat <- missing(txdat)
  no.tydat <- missing(tydat)
  no.tzdat <- missing(tzdat)
  has.explicit.bws <- (!no.bws) && isa(bws, "plbandwidth")

  ## if bws was passed in explicitly, do not compute bandwidths
    
  if(txdat.named)
    txdat <- toFrame(txdat)

  if(tydat.named)
    tydat <- toFrame(tydat)

  if(tydat.named)
    tzdat <- toFrame(tzdat)

  sc.bw <- sc
  
  sc.bw[[1]] <- quote(npplregbw)

  bws.formula <- (!no.bws) && inherits(bws, "formula")
  if (bws.formula) {
    ib <- match("bws", names(sc.bw), nomatch = 0L)
    if (ib > 0L) names(sc.bw)[ib] <- "formula"
  }

  if(bws.named && !bws.formula){
    sc.bw$bandwidth.compute <- FALSE
  }

  ostxy <- c('txdat','tydat','tzdat')
  nstxy <- c('xdat','ydat','zdat')
  
  m.txy <- match(ostxy, names(sc.bw), nomatch = 0)

  if(any(m.txy > 0)) {
    names(sc.bw)[m.txy] <- nstxy[m.txy > 0]
  }
    
  use.outer.bandwidth.progress <- !.np_bw_call_uses_nomad_degree_search(
    sc.bw,
    caller_env = parent.frame()
  )

  tbw <- if (!has.explicit.bws) {
    if (use.outer.bandwidth.progress) {
      .np_progress_select_bandwidth_enhanced(
        "Selecting partially linear regression bandwidth",
        .np_eval_bw_call(sc.bw, caller_env = parent.frame())
      )
    } else {
      .np_eval_bw_call(sc.bw, caller_env = parent.frame())
    }
  } else {
    .np_eval_bw_call(sc.bw, caller_env = parent.frame())
  }
  
  call.args <- list(bws = tbw)
  if (no.bws) {
    call.args$txdat <- txdat
    call.args$tydat <- tydat
    call.args$tzdat <- tzdat
  } else {
    if (txdat.named) call.args$txdat <- txdat
    if (tydat.named) call.args$tydat <- tydat
    if (tzdat.named) call.args$tzdat <- tzdat
    if ((!bws.named) && (!txdat.named) && (!no.tzdat) && (!tzdat.named)) {
      call.args <- c(call.args, list(tzdat))
    }
  }
  if (!has.explicit.bws)
    call.args$.np_fit_progress_handoff <- TRUE
  do.call(npplreg, c(call.args, list(...)))
}

## This functions accepts the following arguments:

## y: univariate outcome
## z: endogenous predictors
## w: instruments
## x: exogenous predictors

## zeval: optional evaluation data for the endogenous predictors
## xeval: optional evaluation data for the exogenous predictors

## alpha.min: minimum value when conducting 1-dimensional search for
##            optimal Tikhonov regularization parameter alpha

## alpha.max: maximum value when conducting 1-dimensional search for
##            optimal Tikhonov regularization parameter alpha

## p: order of the local polynomial kernel estimator (p=0 is local
##    constant, p=1 local linear etc.)

## This function returns a list with at least the following elements:

## phi: the IV estimator of phi(z)
## convergence: a character string indicating whether/why iteration terminated

npregiv <- function(y,
                    z,
                    w,
                    x=NULL,
                    zeval=NULL,
                    xeval=NULL,
                    alpha=NULL,
                    alpha.iter=NULL,
                    alpha.max=1.0e-01,
                    alpha.min=1.0e-10,
                    alpha.tol=.Machine$double.eps^0.25,
                    bw=NULL,
                    constant=0.5,
                    iterate.diff.tol=1.0e-08,
                    iterate.max=1000,
                    iterate.Tikhonov=TRUE,
                    iterate.Tikhonov.num=1,
                    method=c("Landweber-Fridman","Tikhonov"),
                    nmulti=NULL,
                    optim.abstol=.Machine$double.eps,
                    optim.maxattempts = 10,
                    optim.maxit=500,
                    optim.method=c("Nelder-Mead", "BFGS", "CG"),
                    optim.reltol=sqrt(.Machine$double.eps),
                    p=1,
                    penalize.iteration=TRUE,
                    random.seed=42,
                    return.weights.phi=FALSE,
                    return.weights.phi.deriv.1=FALSE,
                    return.weights.phi.deriv.2=FALSE,
                    smooth.residuals=TRUE,
                    start.from=c("Eyz","EEywz"),
                    starting.values=NULL,
                    stop.on.increase=TRUE,
                    ...) {

  ptm.start <- proc.time()
  cl <- match.call()
  if(!is.null(nmulti)) nmulti <- npValidateNmulti(nmulti)

  ## This function was constructed initially by Samuele Centorrino
  ## <samuele.centorrino@univ-tlse1.fr> to reproduce illustrations in
  ## the following papers:

  ## A) Econometrica (2011), Volume 79, pp. 1541-1565
  ## "Nonparametric Instrumental Regression"
  ## S. Darolles, Y. Fan, J.P. Florens, E. Renault

  ## B) Econometrics Journal (2010), volume 13, pp. S1-S27. doi:
  ## 10.1111/j.1368-423X.2010.00314.x

  ## "The practice of non-parametric estimation by solving inverse
  ## problems: the example of transformation models"

  ## FREDERIQUE FEVE AND JEAN-PIERRE FLORENS
  ## IDEI and Toulouse School of Economics, Universite de Toulouse
  ## Capitole 21 alle de de Brienne, 31000 Toulouse, France. E-mails:
  ## feve@cict.fr, florens@cict.fr

  ## It was modified by Jeffrey S. Racine <racinej@mcmaster.ca> and all
  ## errors remain my responsibility. I am indebted to Samuele and the
  ## Toulouse School of Economics for their generous hospitality.

  ## First we require two functions, the first that conducts Regularized
  ## Tikhonov Regression' (aka Ridge Regression)

  ## This function conducts regularized Tikhonov regression which
  ## corresponds to (3.9) in Feve & Florens (2010).

  ## This function accepts as arguments

  ## alpha: penalty
  ## CZ:    row-normalized kernel weights for the `independent' variable
  ## CY:    row-normalized kernel weights for the `dependent' variable
  ## Cr:    row-normalized kernel weights for the `instrument/endogenous' variable (see NOTE below)
  ## r:     vector of conditional expectations (z can be E(Z|z) - see NOTE below)

  ## NOTE: for Cr, in the transformation model case treated in Feve &
  ## Florens (2010) this maps Z onto the Y space. In the IV case
  ## (Darrolles, Fan, Florens & Renault (2011, forthcoming Econometrica)
  ## it maps W (the instrument) onto the space of the endogenous
  ## regressor Z.

  ## NOTE: for r, in the transformation model it will be equivalent to
  ## the vector of exogenous covariates, and in the endogenous case r is
  ## the conditional mean of y given the instrument W.

  ## This function returns TBA (need better error checking!)

  ## phi:   the vector of estimated values for the unknown function at the evaluation points

  tikh <- function(alpha,CZ,CY,Cr.r,cholesky=FALSE){
      if(cholesky) {
          return(chol2inv(chol(diag(alpha, nrow(CY)) + CY%*%CZ)) %*% Cr.r)
      } else {
          return(solve(diag(alpha, nrow(CY)) + CY%*%CZ) %*% Cr.r)
      }
  }

  ## Samuele indicates alternate form for estimator (visit to SUNY
  ## Stony Brook Feb 24 2015) that can be used with evaluation
  ## data. There is no need to carry around two versions of the same
  ## function, so with some thought we could jettison the above and
  ## use this throughout.

  tikh.eval <- function(alpha,CZ,CY,CY.eval,r,cholesky=FALSE){
    if(cholesky) {
        return(CY.eval%*%chol2inv(chol(diag(alpha, nrow(CY)) + CZ%*%CY)) %*% r)
    } else {
        return(CY.eval%*%solve(diag(alpha, nrow(CY)) + CZ%*%CY) %*% r)
    }
  }

  ## This function applies the iterated Tikhonov approach which
  ## corresponds to (3.10) in Feve & Florens (2010).

  ## This function accepts as arguments

  ## alpha: penalty
  ## CZ:    row-normalized kernel weights for the `independent' variable
  ## CY:    row-normalized kernel weights for the `dependent' variable
  ## Cr:    row-normalized kernel weights for the `instrument/endogenous' variable (see NOTE below)
  ## r:     vector of conditional expectations (z can be E(Z|z) - see NOTE below)

  ## NOTE: for Cr, in the transformation model case treated in Feve &
  ## Florens (2010) this maps Z onto the Y space. In the IV case
  ## (Darrolles, Fan, Florens & Renault (2011, forthcoming Econometrica)
  ## it maps W (the instrument) onto the space of the endogenous
  ## regressor Z.

  ## NOTE: for r, in the transformation model it will be equivalent to
  ## the vector of exogenous covariates, and in the endogenous case r is
  ## the conditional mean of y given the instrument W.

  ## This function returns TBA (need better error checking!)

  ## phi:   the vector of estimated values for the unknown function at the evaluation points

  ## SSalpha: (scalar) value of the sum of square residuals criterion
  ## which is a function of alpha (see (3.10) of Feve & Florens (2010)

  ittik <- function(alpha,CZ,CY,Cr.r,r,cholesky=FALSE) {
      if(cholesky) {
          invmat <- chol2inv(chol(diag(alpha, nrow(CY)) + CY%*%CZ))
      } else {
          invmat <- solve(diag(alpha, nrow(CY)) + CY%*%CZ)
      }
      invmat.Cr.r <- invmat %*% Cr.r
      phi <- invmat.Cr.r + alpha * invmat %*% invmat.Cr.r
      return(sum((CZ%*%phi - r)^2)/NZD(alpha))
  }

  ## This function returns the weight matrix for a local polynomial,
  ## and was rewritten 14/1/15 in Toulouse while visiting JP. It
  ## supports mixed data types. Basic error checking is
  ## undertaken. deriv = 0, strips off weights for mean, = p partials
  ## up to order 2. No cross-partials in this one. Basically useful
  ## for univariate case when deriv > 0 though could be refined - the
  ## old function was slower but had more capability (that basically
  ## went unused).

  ## Update - from ?npksum, "The option permutation.operator= can be
  ## used to `mix and match' operator strings to create a `hybrid'
  ## kernel, in addition to the kernel sum with no operators applied,
  ## one for each continuous dimension in the data. For example, for a
  ## two-dimensional data frame of numeric datatypes,
  ## permutation.operator=c("derivative") will return the usual kernel
  ## sum as if operator = c("normal","normal") in the ksum member, and
  ## in the p.ksum member, it will return kernel sums for operator =
  ## c("derivative","normal"), and operator =
  ## c("normal","derivative"). This makes the computation of gradients
  ## much easier."

  ## So, the upshot is that I could, for the multivariate case, add
  ## the derivative stuff.

  Kmat.lp <- function(deriv=0,
                      mydata.train=NULL,
                      mydata.eval=NULL,
                      bws=NULL,
                      p=0,
                      shrink=TRUE,
                      warning.immediate=TRUE,
                      ...) {

      ## 14/1/15, Toulouse - note that the weights herein ** DO NOT **
      ## shrink towards the lc estimator (neither for the function nor
      ## derivatives), unlike the function returned in
      ## glpreg(). However, they all appear to agree with the previous
      ## Kmat.lp with ** also ** did not shrink towards the lc
      ## estimator. This is noticeably faster, which ought to render
      ## Tikhonov faster as well.

      ## Basic error checking...

      if(is.null(mydata.train)) stop("You must provide training data")
      if(is.null(mydata.eval)) mydata.eval <- mydata.train
      if(is.null(bws)) stop("You must provide bandwidths")

      n.train=nrow(mydata.train)
      n.eval=nrow(mydata.eval)

      X.train <- as.data.frame(mydata.train)
      X.eval <- as.data.frame(mydata.eval)

      ## Check whether it appears that training and evaluation data are
      ## conformable...

      if(ncol(X.train)!=ncol(X.eval))
          stop("Error: training and evaluation data have unequal number of columns\n")

      X.col.numeric <- sapply(seq_len(ncol(X.train)),function(i){is.numeric(X.train[,i])})

      ## k represents the number of numeric regressors, this will return
      ## zero if there are none

      k <- ncol(as.data.frame(X.train[,X.col.numeric]))

      if(k > 0) {
          X.train.numeric <- as.data.frame(X.train[,X.col.numeric])
          X.eval.numeric <- as.data.frame(X.eval[,X.col.numeric])
      }

      if(deriv<0||deriv>2)
          stop(paste("Error: deriv= (integer) is invalid\n[min = ", 0, ", max = ",  p, "]\n",sep=""))

      if(length(p) != 1) {
          p <- p[1]
      }

      if(p < 0)
          stop(paste("Error: p (order of polynomial) must be a non-negative integer\np is (", p, ")\n",sep=""))

      K.x <- npksum(txdat=X.train,
                    exdat=X.eval,
                    bws=bws,
                    return.kernel.weights=TRUE,
                    ...)$kw

      if(p==0) {

          ## No shrinking necessary for local constant estimator

          if(deriv==0) {

              Kmat <- t(K.x)/NZD(rowSums(t(K.x)))

          } else if(deriv==1) {

              ## Note this is not general XXX Feb 25 2015, for
              ## univariate z only

              K.x.deriv <- npksum(txdat=X.train,
                                  exdat=X.eval,
                                  bws=bws,
                                  return.kernel.weights=TRUE,
                                  operator="derivative",
                                  ...)$kw/NZD(bws)

              rSk <- NZD(rowSums(t(K.x)))

              Kmat <- t(K.x.deriv)/NZD(rSk)-t(K.x)/NZD(rSk)*(rowSums(t(K.x.deriv))/NZD(rSk))

          }

      }

      if(p > 0) {

          ## Re-use this matrix, shrinking occurs here

          W.z <- W.lp(xdat=X.train.numeric,
                             degree=rep(p,NCOL(X.train.numeric)))

          if(is.null(mydata.eval)) {
              ## Guess we could avoid copy with conditional statement below using either W.z or W.z.eval
              W.z.eval <- W.z
          } else {
              W.z.eval <- W.lp(xdat=X.train.numeric,
                                      exdat=as.data.frame(X.eval.numeric),
                                      degree=rep(p,NCOL(X.train.numeric)))
          }

          nc <- ncol(W.z)

          WzkWz.inv <- list()

          for(i in seq_len(ncol(K.x))) {

              WzkWz.inv.try <- tryCatch(as.matrix(chol2inv(chol(t(W.z)%*%(K.x[,i]*W.z)))),
                                        error = function(e){
                                            return(matrix(FALSE,nc,nc))
                                        })

              if(WzkWz.inv.try[1,1]!=FALSE) {
                  WzkWz.inv[[i]] <- WzkWz.inv.try
              } else {

                  if(shrink==FALSE) {

                      ## If we do not explicitly engage ridging then we do not fail
                      ## and terminate, rather, we return NA when Wmat.sum is
                      ## singular

                      Kmat <- NA

                  } else {

                      ## Ridging

                      epsilon <- 1/n.train
                      ridge <- 0

                      while(tryCatch(as.matrix(chol2inv(chol((chol(t(W.z)%*%(K.x[,i]*W.z)+diag(rep(ridge,nc))))))),
                                     error = function(e){
                                         return(matrix(FALSE,nc,nc))
                                     })[1,1]==FALSE) {
                          ridge <- ridge + epsilon
                      }

                      WzkWz.inv[[i]] <- as.matrix(chol2inv(chol(t(W.z)%*%(K.x[,i]*W.z)+diag(rep(ridge,nc)))))

                      .np_warning(paste("Ridging obs. ", i, ", ridge = ", signif(ridge,6),sep=""),
                              immediate.=warning.immediate,
                              call.=!warning.immediate)

                  }

              }

          }
      }

      if(p==1) {

          if(deriv==0) {
              Kmat <- matrix(NA, n.eval, n.train)
              for(i in seq_len(n.eval)) {
                  Kmat[i,] <- W.z.eval[i,,drop=FALSE]%*%WzkWz.inv[[i]]%*%t(W.z)*K.x[,i]
              }
          }
          if(deriv==1) {
              W.z.deriv.1 <- W.lp(xdat=X.train.numeric,
                                         exdat=as.matrix(X.eval.numeric),
                                         degree=rep(p,NCOL(X.train.numeric)),
                                         gradient.vec = 1)

              Kmat <- matrix(NA, n.eval, n.train)
              for(i in seq_len(n.eval)) {
                  Kmat[i,] <- W.z.deriv.1[i,,drop=FALSE]%*%WzkWz.inv[[i]]%*%t(W.z)*K.x[,i]
              }
          }

      }

      if(p >= 2) {

          Kmat <- matrix(NA, n.eval, n.train)
          if(deriv==0) {
              for(i in seq_len(n.eval)) {
                  Kmat[i,] <- W.z.eval[i,,drop=FALSE]%*%WzkWz.inv[[i]]%*%t(W.z)*K.x[,i]
              }
          }


          if(deriv==1) {

              W.z.deriv.1 <- W.lp(xdat=X.train.numeric,
                                         exdat=as.matrix(X.eval.numeric),
                                         degree=rep(p,NCOL(X.train.numeric)),
                                         gradient.vec = 1)

              for(i in seq_len(n.eval)) {
                  Kmat[i,] <- W.z.deriv.1[i,,drop=FALSE]%*%WzkWz.inv[[i]]%*%t(W.z)*K.x[,i]
              }

          }

          if(deriv==2) {

              W.z.deriv.2 <- W.lp(xdat=X.train.numeric,
                                         exdat=as.matrix(X.eval.numeric),
                                         degree=rep(p,NCOL(X.train.numeric)),
                                         gradient.vec = 2)

              for(i in seq_len(n.eval)) {
                  Kmat[i,] <- W.z.deriv.2[i,,drop=FALSE]%*%WzkWz.inv[[i]]%*%t(W.z)*K.x[,i]
              }

          }

      }

      return(Kmat)
  }

  glpreg <- function(tydat=NULL,
                     txdat=NULL,
                     exdat=NULL,
                     bws=NULL,
                     degree=NULL,
                     leave.one.out=FALSE,
                     deriv=1,
                     ...) {

    ## Don't think this error checking is robust

    if(is.null(tydat)) stop("Error: You must provide y data")
    if(is.null(txdat)) stop("Error: You must provide X data")
    if(is.null(bws)) stop("Error: You must provide a bandwidth object")
    if(is.null(degree) || any(degree < 0)) stop(paste("Error: degree vector must contain non-negative integers\ndegree is (", degree, ")\n",sep=""))
    if(p>0 && (any(deriv < 0) || any(deriv > degree)))
      stop("deriv must lie between 0 and degree")

    miss.ex = missing(exdat)

    if (miss.ex){
      exdat <- txdat
    }

    txdat <- as.data.frame(txdat)
    exdat <- as.data.frame(exdat)

    maxPenalty <- sqrt(.Machine$double.xmax)

    n.train <- nrow(txdat)
    n.eval <- nrow(exdat)

    ## Check whether it appears that training and evaluation data are
    ## conformable

    if(ncol(txdat)!=ncol(exdat))
      stop("Error: training and evaluation data have unequal number of columns\n")

    if(all(degree == 0)) {

      ## Local constant using only one call to npksum

      if(leave.one.out == TRUE) {

        ## exdat not supported with leave.one.out, but this is only used
        ## for cross-validation hence no exdat

        tww <- npksum(txdat = txdat,
                      weights = as.matrix(data.frame(1,tydat)),
                      tydat = rep(1,length(tydat)),
                      bws = bws,
                      bandwidth.divide = TRUE,
                      leave.one.out = leave.one.out,
                      ukertype="liracine",
                      okertype="liracine",
                      ...)$ksum

      } else {

        tww <- npksum(txdat = txdat,
                      exdat = exdat,
                      weights = as.matrix(data.frame(1,tydat)),
                      tydat = rep(1,length(tydat)),
                      bws = bws,
                      bandwidth.divide = TRUE,
                      leave.one.out = leave.one.out,
                      ukertype="liracine",
                      okertype="liracine",
                      ...)$ksum
      }

      ## Note that as bandwidth approaches zero the local constant
      ## estimator undersmooths and approaches each sample realization,
      ## so use the convention that when the sum of the kernel weights
      ## equals 0, return y. This is unique to this code.

      mhat <- tww[2,]/NZD(tww[1,])

      grad <- gradients(npreg(tydat=tydat,
                              txdat=txdat,
                              exdat = exdat,
                              bws = bws,
                              ukertype="liracine",
                              okertype="liracine",
                              gradients=TRUE,
                              ...))

      return(list(mean = mhat,
                  grad = grad))

    } else {

      W <- W.lp(xdat=txdat,
                       degree=degree)

      W.eval <- W.lp(xdat=txdat,
                            exdat=exdat,
                            degree=degree)

      W.eval.deriv <- W.lp(xdat=txdat,
                                  exdat=exdat,
                                  degree=degree,
                                  gradient.vec=rep(deriv,NCOL(txdat)))

      ## Local polynomial via smooth coefficient formulation and one
      ## call to npksum

      if(leave.one.out == TRUE) {

        ## exdat not supported with leave.one.out, but this is only used
        ## for cross-validation hence no exdat

        tww <- npksum(txdat = txdat,
                      tydat = as.matrix(cbind(tydat,W)),
                      weights = W,
                      bws = bws,
                      bandwidth.divide = TRUE,
                      leave.one.out = leave.one.out,
                      ukertype="liracine",
                      okertype="liracine",
                      ...)$ksum

      } else {

        tww <- npksum(txdat = txdat,
                      exdat = exdat,
                      tydat = as.matrix(cbind(tydat,W)),
                      weights = W,
                      bws = bws,
                      bandwidth.divide = TRUE,
                      leave.one.out = leave.one.out,
                      ukertype="liracine",
                      okertype="liracine",
                      ...)$ksum

      }

      tyw <- array(tww,dim = c(ncol(W)+1,ncol(W),n.eval))[1,,]
      tww <- array(tww,dim = c(ncol(W)+1,ncol(W),n.eval))[-1,,]

      coef.mat <- matrix(maxPenalty, ncol(W), n.eval)
      ridge.grid <- npRidgeSequenceAdditive(n.train = n.eval, cap = 1.0)
      ridge <- rep.int(ridge.grid[1L], n.eval)
      ridge.idx <- rep.int(1L, n.eval)
      ridge.lc.mean <- double(n.eval)
      doridge <- rep.int(TRUE, n.eval)

      nc <- ncol(tww[,,1])
      I.nc <- diag(nc)

      ## Test for singularity of the local polynomial (regtype = "lp")
      ## estimator, shrink the mean towards the local constant mean.

      while(any(doridge)){
        iloo <- seq_len(n.eval)[doridge]
        for(i in iloo) {
          doridge[i] <- FALSE
          ridge.lc.mean[i] <- tyw[1,i][1]/NZD(tww[,,i][1,1])
          coef.mat[,i] <- tryCatch(chol2inv(chol(tww[,,i] + ridge[i]*I.nc))%*%tyw[,i],
                                   error = function(e){
                                     ridge.idx[i] <- ridge.idx[i] + 1L
                                     if (ridge.idx[i] <= length(ridge.grid)) {
                                       ridge[i] <- ridge.grid[ridge.idx[i]]
                                       doridge[i] <- TRUE
                                     }
                                     return(rep(maxPenalty,nc))
                                   })
        }
      }

      ridge.alpha <- pmin(1.0, ridge)
      mhat <- sapply(seq_len(n.eval), function(i) {
        (1 - ridge.alpha[i]) * W.eval[i,, drop = FALSE] %*% coef.mat[,i] +
          ridge.alpha[i] * ridge.lc.mean[i]
      })

      grad <- sapply(seq_len(n.eval), function(i) {W.eval.deriv[i,-1, drop = FALSE] %*% coef.mat[-1,i]})

      return(list(mean = mhat,
                  grad = grad))

    }

  }

  minimand.cv.ls <- function(bws=NULL,
                             ydat=NULL,
                             xdat=NULL,
                             degree=NULL,
                             W=NULL,
                             ...) {

    ## Don't think this error checking is robust

    if(is.null(ydat)) stop("Error: You must provide y data")
    if(is.null(xdat)) stop("Error: You must provide X data")
    if(is.null(W)) stop("Error: You must provide a weighting matrix W")
    if(is.null(bws)) stop("Error: You must provide a bandwidth object")
    if(is.null(degree) || any(degree < 0)) stop(paste("Error: degree vector must contain non-negative integers\ndegree is (", degree, ")\n",sep=""))

    xdat <- as.data.frame(xdat)

    n <- length(ydat)

    maxPenalty <- sqrt(.Machine$double.xmax)

    if(any(bws<=0)) {

      return(maxPenalty)

    } else {

      if(all(degree == 0)) {

        ## Local constant via one call to npksum

        tww <- npksum(txdat = xdat,
                      weights = as.matrix(data.frame(1,ydat)),
                      tydat = rep(1,n),
                      bws = bws,
                      leave.one.out = TRUE,
                      bandwidth.divide = TRUE,
                      ukertype="liracine",
                      okertype="liracine",
                      ...)$ksum

        mean.loo <- tww[2,]/NZD(tww[1,])

        if (!any(is.nan(mean.loo)) && !any(mean.loo == maxPenalty)){
          fv <- mean((ydat-mean.loo)^2)
        } else {
          fv <- maxPenalty
        }

        return(if (is.finite(fv)) fv else maxPenalty)

      } else {

        ## Generalized local polynomial via smooth coefficient
        ## formulation and one call to npksum

        tww <- npksum(txdat = xdat,
                      tydat = as.matrix(cbind(ydat,W)),
                      weights = W,
                      bws = bws,
                      leave.one.out = TRUE,
                      bandwidth.divide = TRUE,
                      ukertype="liracine",
                      okertype="liracine",
                      ...)$ksum

        tyw <- array(tww,dim = c(ncol(W)+1,ncol(W),n))[1,,]
        tww <- array(tww,dim = c(ncol(W)+1,ncol(W),n))[-1,,]

        mean.loo <- rep(maxPenalty, n)
        ridge.grid <- npRidgeSequenceAdditive(n.train = n, cap = 1.0)
        ridge <- rep.int(ridge.grid[1L], n)
        ridge.idx <- rep.int(1L, n)
        ridge.lc.mean <- double(n)
        doridge <- rep.int(TRUE, n)

        nc <- ncol(tww[,,1])

        ## Test for singularity of the local polynomial (regtype = "lp")
        ## estimator, shrink the mean towards the local constant mean.

        while(any(doridge)){
          iloo <- seq_len(n)[doridge]
          for(i in iloo) {
            doridge[i] <- FALSE
            ridge.lc.mean[i] <- tyw[1,i][1]/NZD(tww[,,i][1,1])
            mean.i <- W[i,, drop = FALSE] %*% tryCatch(chol2inv(chol(tww[,,i]+diag(rep(ridge[i],nc))))%*%tyw[,i],
                                                       error = function(e){
                                                         ridge.idx[i] <- ridge.idx[i] + 1L
                                                         if (ridge.idx[i] <= length(ridge.grid)) {
                                                           ridge[i] <- ridge.grid[ridge.idx[i]]
                                                           doridge[i] <- TRUE
                                                         }
                                                         return(rep(maxPenalty,nc))
                                                       })
            alpha.i <- min(1.0, ridge[i])
            mean.loo[i] <- (1 - alpha.i) * mean.i + alpha.i * ridge.lc.mean[i]
          }
        }

        if (!any(is.nan(mean.loo)) && !any(mean.loo == maxPenalty)){
          fv <- mean((ydat-mean.loo)^2)
        } else {
          fv <- maxPenalty
        }

        return(if (is.finite(fv)) fv else maxPenalty)

      }

    }

  }

  minimand.cv.aic <- function(bws=NULL,
                              ydat=NULL,
                              xdat=NULL,
                              degree=NULL,
                              W=NULL,
                              ...) {

    ## Don't think this error checking is robust

    if(is.null(ydat)) stop("Error: You must provide y data")
    if(is.null(xdat)) stop("Error: You must provide X data")
    if(!all(degree==0)) if(is.null(W)) stop("Error: You must provide a weighting matrix W")
    if(is.null(bws)) stop("Error: You must provide a bandwidth object")
    if(is.null(degree) || any(degree < 0)) stop(paste("Error: degree vector must contain non-negative integers\ndegree is (", degree, ")\n",sep=""))

    xdat <- as.data.frame(xdat)

    n <- length(ydat)

    maxPenalty <- sqrt(.Machine$double.xmax)

    if(any(bws<=0)) {

      return(maxPenalty)

    } else {

      ## This computes the kernel function when i=j (i.e., K(0))

      kernel.i.eq.j <- npksum(txdat = xdat[1,],
                              weights = as.matrix(data.frame(1,ydat)[1,]),
                              tydat = 1,
                              bws = bws,
                              bandwidth.divide = TRUE,
                              ukertype="liracine",
                              okertype="liracine",
                              ...)$ksum[1,1]

      if(all(degree == 0)) {

        ## Local constant via one call to npksum

        tww <- npksum(txdat = xdat,
                      weights = as.matrix(data.frame(1,ydat)),
                      tydat = rep(1,n),
                      bws = bws,
                      bandwidth.divide = TRUE,
                      ukertype="liracine",
                      okertype="liracine",
                      ...)$ksum

        ghat <- tww[2,]/NZD(tww[1,])

        trH <- kernel.i.eq.j*sum(1/NZD(tww[1,]))

        aic.penalty <- (1+trH/n)/(1-(trH+2)/n)

        if (!any(ghat == maxPenalty) && (aic.penalty > 0)){
          fv <- log(mean((ydat-ghat)^2)) + aic.penalty
        } else {
          fv <- maxPenalty
        }

        return(if (is.finite(fv)) fv else maxPenalty)

      } else {

        ## Generalized local polynomial via smooth coefficient
        ## formulation and one call to npksum

        tww <- npksum(txdat = xdat,
                      tydat = as.matrix(cbind(ydat,W)),
                      weights = W,
                      bws = bws,
                      bandwidth.divide = TRUE,
                      ukertype="liracine",
                      okertype="liracine",
                      ...)$ksum

        tyw <- array(tww,dim = c(ncol(W)+1,ncol(W),n))[1,,]
        tww <- array(tww,dim = c(ncol(W)+1,ncol(W),n))[-1,,]

        ghat <- rep(maxPenalty, n)
        ridge.grid <- npRidgeSequenceAdditive(n.train = n, cap = 1.0)
        ridge <- rep.int(ridge.grid[1L], n)
        ridge.idx <- rep.int(1L, n)
        ridge.lc.mean <- double(n)
        doridge <- rep.int(TRUE, n)

        nc <- ncol(tww[,,1])

        ## Test for singularity of the local polynomial (regtype = "lp")
        ## estimator, shrink the mean towards the local constant mean.

        while(any(doridge)){
          ii <- seq_len(n)[doridge]
          for(i in ii) {
            doridge[i] <- FALSE
            ridge.lc.mean[i] <- tyw[1,i][1]/NZD(tww[,,i][1,1])
            ghat.i <- W[i,, drop = FALSE] %*% tryCatch(chol2inv(chol(tww[,,i]+diag(rep(ridge[i],nc))))%*%tyw[,i],
                                                       error = function(e){
                                                         ridge.idx[i] <- ridge.idx[i] + 1L
                                                         if (ridge.idx[i] <= length(ridge.grid)) {
                                                           ridge[i] <- ridge.grid[ridge.idx[i]]
                                                           doridge[i] <- TRUE
                                                         }
                                                         return(rep(maxPenalty,nc))
                                                       })
            alpha.i <- min(1.0, ridge[i])
            ghat[i] <- (1 - alpha.i) * ghat.i + alpha.i * ridge.lc.mean[i]
          }
        }

        if (any(ghat == maxPenalty))
          return(maxPenalty)

        ridge.alpha <- pmin(1.0, ridge)
        trH.comp <- sapply(seq_len(n), function(i) {
          tryCatch(
            (1 - ridge.alpha[i]) * W[i,, drop = FALSE] %*%
              chol2inv(chol(tww[,,i] + diag(rep(ridge[i], nc)))) %*%
              t(W[i,, drop = FALSE]) +
              ridge.alpha[i] / NZD(tww[,,i][1,1]),
            error = function(e) maxPenalty
          )
        })
        if (any(trH.comp == maxPenalty))
          return(maxPenalty)
        trH <- kernel.i.eq.j * sum(trH.comp)

        aic.penalty <- (1+trH/n)/(1-(trH+2)/n)

        if (!any(ghat == maxPenalty) && (aic.penalty > 0)){
          fv <- log(mean((ydat-ghat)^2)) + aic.penalty
        } else {
          fv <- maxPenalty
        }

        return(if (is.finite(fv)) fv else maxPenalty)

      }

    }

  }

  glpcv <- function(ydat=NULL,
                    xdat=NULL,
                    degree=NULL,
                    bwmethod=c("cv.ls","cv.aic"),
                    nmulti=nmulti,
                    random.seed=42,
                    optim.maxattempts = 10,
                    optim.method=c("Nelder-Mead", "BFGS", "CG"),
                    optim.reltol=sqrt(.Machine$double.eps),
                    optim.abstol=.Machine$double.eps,
                    optim.maxit=500,
                    debug=FALSE,
                    bw.init=NULL,
                    ...) {

    ## Save seed prior to setting
    seed.state <- .np_seed_enter(random.seed)
    on.exit(.np_seed_exit(seed.state, remove_if_absent = TRUE), add = TRUE)

    if(debug) {
      debug.dir <- tempfile("npregiv_debug_")
      dir.create(debug.dir, showWarnings = FALSE, recursive = TRUE)
      debug.file.optim <- file.path(debug.dir, "optim.debug")
      debug.file.bandwidth <- file.path(debug.dir, "bandwidth.out")
      debug.file.summary <- file.path(debug.dir, "optim.out")
    }

    ## Don't think this error checking is robust

    if(is.null(ydat)) stop("Error: You must provide y data")
    if(is.null(xdat)) stop("Error: You must provide X data")
    if(is.null(degree) || any(degree < 0)) stop(paste("Error: degree vector must contain non-negative integers\ndegree is (", degree, ")\n",sep=""))
    if(!is.null(nmulti)) nmulti <- npValidateNmulti(nmulti)

    bwmethod = match.arg(bwmethod)

    optim.method <- match.arg(optim.method)
    optim.control <- list(abstol = optim.abstol,
                          reltol = optim.reltol,
                          maxit = optim.maxit)

    maxPenalty <- sqrt(.Machine$double.xmax)

    xdat <- as.data.frame(xdat)

    num.bw <- ncol(xdat)

    if(is.null(nmulti)) nmulti <- npDefaultNmulti(num.bw)

    ## Which variables are categorical, which are discrete...

    xdat.numeric <- sapply(seq_len(ncol(xdat)),function(i){is.numeric(xdat[,i])})

    ## First initialize initial search values of the vector of
    ## bandwidths to lie in [0,1]

    if(debug) write(c("cv",paste(rep("x",num.bw),seq_len(num.bw),sep="")),file=debug.file.optim,ncolumns=(num.bw+1))

    ## Pass in the local polynomial weight matrix rather than
    ## recomputing with each iteration.

    W <- W.lp(xdat=xdat,
                     degree=degree)

    sum.lscv <- function(bw.gamma,...) {
      .np_progress_iv_activity_step()

      ## Note - we set the kernel for unordered and ordered regressors
      ## to the liracine kernel (0<=lambda<=1) and test for proper
      ## bounds in sum.lscv.

      if(all(bw.gamma>=0)&&all(bw.gamma[!xdat.numeric]<=1)) {
        lscv <- minimand.cv.ls(bws=bw.gamma,ydat=ydat,xdat=xdat,...)
      } else {
        lscv <- maxPenalty
      }

      if(debug) write(c(lscv,bw.gamma),file=debug.file.optim,ncolumns=(num.bw+1),append=TRUE)
      return(lscv)
    }

    sum.aicc <- function(bw.gamma,...) {
      .np_progress_iv_activity_step()

      ## Note - we set the kernel for unordered and ordered regressors
      ## to the liracine kernel (0<=lambda<=1) and test for proper
      ## bounds in sum.lscv.

      if(all(bw.gamma>=0)&&all(bw.gamma[!xdat.numeric]<=1)) {
        aicc <- minimand.cv.aic(bws=bw.gamma,ydat=ydat,xdat=xdat,...)
      } else {
        aicc <- maxPenalty
      }

      if(debug) write(c(aicc,bw.gamma),file=debug.file.optim,ncolumns=(num.bw+1),append=TRUE)
      return(aicc)
    }

    ## Multistarting

    fv.vec <- numeric(nmulti)

    ## Pass in the W matrix rather than recomputing it each time

    for(iMulti in seq_len(nmulti)) {

      num.numeric <- ncol(as.data.frame(xdat[,xdat.numeric]))

      if(iMulti == 1 && !is.null(bw.init)) {
        init.search.vals <- bw.init
      } else {

        ## First initialize to values for factors (`liracine' kernel)

        init.search.vals <- runif(ncol(xdat),0,1)

        for(i in seq_len(ncol(xdat))) {
          if(xdat.numeric[i]==TRUE) {
            init.search.vals[i] <- runif(1,.5,1.5)*EssDee(xdat[,i])*nrow(xdat)^{-1/(4+num.numeric)}
          }
        }
      }

      ## Initialize `best' values prior to search

      if(iMulti == 1) {
        fv <- maxPenalty
        numimp <- 0
        bw.opt <- init.search.vals
        best <- 1
      }

      if(bwmethod == "cv.ls" ) {

        suppressWarnings(optim.return <- optim(init.search.vals,
                                               fn=sum.lscv,
                                               method=optim.method,
                                               control=optim.control,
                                               degree=degree,
                                               W=W,
                                               ...))

        attempts <- 0
        while((optim.return$convergence != 0) && (attempts <= optim.maxattempts)) {
          init.search.vals <- runif(ncol(xdat),0,1)
          if(xdat.numeric[i]==TRUE) {
            init.search.vals[i] <- runif(1,.5,1.5)*EssDee(xdat[,i])*nrow(xdat)^{-1/(4+num.numeric)}
          }
          attempts <- attempts + 1
          optim.control$abstol <- optim.control$abstol * 10.0
          optim.control$reltol <- optim.control$reltol * 10.0
  #        optim.control <- lapply(optim.control, '*', 10.0) ## Perhaps do not want to keep increasing maxit??? Jan 31 2011
          suppressWarnings(optim.return <- optim(init.search.vals,
                                                 fn=sum.lscv,
                                                 method=optim.method,
                                                 control=optim.control,
                                                 degree=degree,
                                                 W=W,
                                                 ...))
        }

      } else {

        suppressWarnings(optim.return <- optim(init.search.vals,
                                               fn=sum.aicc,
                                               method=optim.method,
                                               control=optim.control,
                                               degree=degree,
                                               W=W,
                                               ...))

        attempts <- 0
        while((optim.return$convergence != 0) && (attempts <= optim.maxattempts)) {
          init.search.vals <- runif(ncol(xdat),0,1)
          if(xdat.numeric[i]==TRUE) {
            init.search.vals[i] <- runif(1,.5,1.5)*EssDee(xdat[,i])*nrow(xdat)^{-1/(4+num.numeric)}
          }
          attempts <- attempts + 1
          optim.control$abstol <- optim.control$abstol * 10.0
          optim.control$reltol <- optim.control$reltol * 10.0
  #        optim.control <- lapply(optim.control, '*', 10.0) ## Perhaps do not want to keep increasing maxit??? Jan 31 2011
          suppressWarnings(optim.return <- optim(init.search.vals,
                                                 fn = sum.aicc,
                                                 method=optim.method,
                                                 control = optim.control,
                                                 W=W,
                                                 ...))
        }
      }

      if(optim.return$convergence != 0) .np_warning(" optim failed to converge")

      fv.vec[iMulti] <- optim.return$value

      if(optim.return$value < fv) {
        bw.opt <- optim.return$par
        fv <- optim.return$value
        numimp <- numimp + 1
        best <- iMulti
        if(debug) {
          if(iMulti==1) {
            write(cbind(iMulti,t(bw.opt)),debug.file.bandwidth,ncolumns=(1+length(bw.opt)))
            write(cbind(iMulti,fv),debug.file.summary,ncolumns=2)
          } else {
            write(cbind(iMulti,t(bw.opt)),debug.file.bandwidth,ncolumns=(1+length(bw.opt)),append=TRUE)
            write(cbind(iMulti,fv),debug.file.summary,ncolumns=2,append=TRUE)
          }
        }
      }

    }

    return(list(bw=bw.opt,
                fv=fv,
                numimp=numimp,
                best=best,
                fv.vec=fv.vec))

  }

  ## Here is where the function `npregiv' really begins:

  ## Basic error checking

  if(!is.logical(penalize.iteration)) stop("penalize.iteration must be logical (TRUE/FALSE)")
  if(!is.logical(smooth.residuals)) stop("smooth.residuals must be logical (TRUE/FALSE)")
  if(!is.logical(stop.on.increase)) stop("stop.on.increase must be logical (TRUE/FALSE)")
  if(!is.logical(iterate.Tikhonov)) stop("iterate.Tikhonov must be logical (TRUE/FALSE)")
  if(iterate.Tikhonov.num < 1) stop("iterate.Tikhonov.num must be a positive integer")  

  if(missing(y)) stop("You must provide y")
  if(missing(z)) stop("You must provide z")
  if(missing(w)) stop("You must provide w")
  if(NCOL(y) > 1) stop("y must be univariate")
  if(NROW(y) != NROW(z) || NROW(y) != NROW(w)) stop("y, z, and w have differing numbers of rows")
  if(iterate.max < 2) stop("iterate.max must be at least 2")
  if(iterate.diff.tol < 0) stop("iterate.diff.tol must be non-negative")
  if(constant <= 0 || constant >=1) stop("constant must lie in (0,1)")
  if(p < 0) stop("p must be a non-negative integer")

  if(!is.null(alpha) && alpha <= 0) stop("alpha must be positive")
  if(!is.null(alpha.iter) && alpha.iter <= 0) stop("alpha.iter must be positive")

  if(return.weights.phi.deriv.1 && !return.weights.phi) stop("must use return.weights.phi=TRUE when using return.weights.phi.deriv.1=TRUE")
  if(return.weights.phi.deriv.2 && !return.weights.phi) stop("must use return.weights.phi=TRUE when using return.weights.phi.deriv.2=TRUE")
  if(return.weights.phi.deriv.2 && p<2) stop("must use p >= 2 when using return.weights.phi.deriv.2=TRUE")

  start.from <- match.arg(start.from)
  method <- match.arg(method)

  iv_set_stage <- function(label, iteration = NULL) {
    .np_progress_iv_set_object(label = label, iteration = iteration)
  }

  iv_start_label <- function() {
    if (identical(start.from, "Eyz")) "E[y|z]" else "E[E[y|w]|z]"
  }

  iv_residual_stage_label <- function(smooth.residuals) {
    if (smooth.residuals) "E[y-phi(z)|w]" else "E[phi(z)|w]"
  }

  iv_adjoint_stage_label <- function(smooth.residuals) {
    if (smooth.residuals) {
      "E[E[y-phi(z)|w]|z]"
    } else {
      "E[E[y|w]-E[phi(z)|w]|z]"
    }
  }

  ## Need to determine how many x, w, z are numeric

  z <- data.frame(z)
  w <- data.frame(w)
  if(!is.null(x)) {
      z <- data.frame(z,x)
      ## JP points out that, with exogenous predictors, they must be
      ## part of both z and the instruments. The line below was added
      ## 20/1/15 in Toulouse.
      w <- data.frame(w,x)
      ## Obviously, if you have exogenous variables that are only in
      ## the instrument set, you can trivially accommodate this
      ## (append to w before invoking the function - added to man
      ## page)
      if(!is.null(zeval)&&!is.null(xeval)) zeval <- data.frame(zeval,xeval)
  }

  nmulti.loop <- if(!is.null(nmulti)) nmulti else 1
  nmulti <- if(!is.null(nmulti)) nmulti else npDefaultNmulti(max(NCOL(z), NCOL(w)))

  z.numeric <- sapply(seq_len(NCOL(z)),function(i){is.numeric(z[,i])})
  num.z.numeric <- NCOL(as.data.frame(z[,z.numeric]))

  w.numeric <- sapply(seq_len(NCOL(w)),function(i){is.numeric(w[,i])})
  num.w.numeric <- NCOL(as.data.frame(w[,w.numeric]))

  result <- .np_progress_select_iv(.np_progress_iv_title(), {
  if(method=="Tikhonov") {

    ## Now y=phi(z) + u, hence E(y|w)=E(phi(z)|w) so we need two
    ## bandwidths, one for y on w and one for phi(z) on w (in the
    ## first step we use E(y|w) as a proxy for phi(z) and use
    ## bandwidths for y on w).

    ## Convergence value returned for Landweber-Fridman but value
    ## required

    ## convergence <- NULL

    iv_set_stage("E[y|w]")
    if(is.null(bw)) {
        hyw <- .np_progress_with_legacy_suppressed(glpcv(ydat=y,
                                                        xdat=w,
                                                        degree=rep(p, num.w.numeric),
                                                        nmulti=nmulti,
                                                        random.seed=random.seed,
                                                        optim.maxattempts=optim.maxattempts,
                                                        optim.method=optim.method,
                                                        optim.reltol=optim.reltol,
                                                        optim.abstol=optim.abstol,
                                                        optim.maxit=optim.maxit,
                                                        ...))

        bw.E.y.w <- hyw$bw
    } else {
        bw.E.y.w <- bw$bw.E.y.w
    }

    E.y.w <- .np_progress_with_legacy_suppressed(glpreg(tydat=y,
                                                        txdat=w,
                                                        bws=bw.E.y.w,
                                                        degree=rep(p, num.w.numeric),
                                                        ...))$mean

    KYW <- Kmat.lp(mydata.train=data.frame(w),
                   bws=bw.E.y.w,
                   p=rep(p, num.w.numeric),
                   ...)

    ## We conduct local polynomial kernel regression of E(y|w) on z

    iv_set_stage("E[E[y|w]|z]")
    if(is.null(bw)) {
        hywz <- .np_progress_with_legacy_suppressed(glpcv(ydat=E.y.w,
                                                         xdat=z,
                                                         degree=rep(p, num.z.numeric),
                                                         nmulti=nmulti,
                                                         random.seed=random.seed,
                                                         optim.maxattempts=optim.maxattempts,
                                                         optim.method=optim.method,
                                                         optim.reltol=optim.reltol,
                                                         optim.abstol=optim.abstol,
                                                         optim.maxit=optim.maxit,
                                                         ...))

        bw.E.E.y.w.z <- hywz$bw
    } else {
        bw.E.E.y.w.z <- bw$bw.E.E.y.w.z
    }

    E.E.y.w.z <- .np_progress_with_legacy_suppressed(glpreg(tydat=E.y.w,
                                                            txdat=z,
                                                            bws=bw.E.E.y.w.z,
                                                            degree=rep(p, num.z.numeric),
                                                            ...))$mean

    KYWZ <- Kmat.lp(mydata.train=data.frame(z),
                    bws=bw.E.E.y.w.z,
                    p=rep(p, num.z.numeric),
                    ...)

    ## Next, we minimize the function ittik to obtain the optimal value
    ## of alpha (here we use the iterated Tikhonov function) to
    ## determine the optimal alpha for the non-iterated scheme. Note
    ## that the function `optimize' accepts bounds on the search (in
    ## this case alpha.min to alpha.max))

    ## E(r|z)=E(E(phi(z)|w)|z)
    ## \phi^\alpha = (\alpha I+CzCw)^{-1}Cr x r

    if(!is.null(bw)) alpha <- bw$alpha

    if(is.null(alpha)&&is.null(bw)) {
      iv_set_stage("alpha")
      alpha <- optimize(ittik, c(alpha.min, alpha.max), tol = alpha.tol, CZ = KYW, CY = KYWZ, Cr.r = E.E.y.w.z, r = E.y.w)$minimum
    }

    ## Finally, we conduct regularized Tikhonov regression using this
    ## optimal alpha.

    iv_set_stage("phi(z)")
    phi <- as.vector(tikh(alpha, CZ = KYW, CY = KYWZ, Cr.r = E.E.y.w.z))
    phi.mat <- phi
    
    phi.eval.mat <- NULL
    if(!is.null(zeval)) {
        ## If there is evaluation data, KPHIWZ and KPHIWZ.eval will
        ## differ...
        
        KPHIWZ.eval <- Kmat.lp(mydata.train=data.frame(z),
                               mydata.eval=data.frame(z=zeval),
                               bws=bw.E.E.y.w.z,
                               p=rep(p, num.z.numeric),
                               ...)
        
        phi.eval <- as.vector(tikh.eval(alpha, CZ = KYW, CY = KYWZ, CY.eval = KPHIWZ.eval, r = E.y.w))
        phi.eval.mat <- cbind(phi.eval.mat,phi.eval)
    }    

    bw.E.phi.w <- NULL
    bw.E.E.phi.w.z <- NULL

    for(i in 1:iterate.Tikhonov.num) {
      iter.label <- if (iterate.Tikhonov.num > 1) i else NULL
      iv_set_stage("E[phi(z)|w]", iteration = iter.label)

      if(is.null(bw)) {
          hphiw <- .np_progress_with_legacy_suppressed(glpcv(ydat=phi, ## 23/1/15 phi is sample
                                                            xdat=w,
                                                            degree=rep(p, num.w.numeric),
                                                            nmulti=nmulti.loop,
                                                            random.seed=random.seed,
                                                            optim.maxattempts=optim.maxattempts,
                                                            optim.method=optim.method,
                                                            optim.reltol=optim.reltol,
                                                            optim.abstol=optim.abstol,
                                                            optim.maxit=optim.maxit,
                                                            bw.init=bw.E.phi.w,
                                                            ...))

          bw.E.phi.w <- hphiw$bw
      } else {
          bw.E.phi.w <- bw$bw.E.phi.w
      }

      E.phi.w <- .np_progress_with_legacy_suppressed(glpreg(tydat=phi,
                                                            txdat=w,
                                                            bws=bw.E.phi.w,
                                                            degree=rep(p, num.w.numeric),
                                                            ...))$mean

      KPHIW <- Kmat.lp(mydata.train=data.frame(w),
                       bws=bw.E.phi.w,
                       p=rep(p, num.w.numeric),
                       ...)

      iv_set_stage("E[E[phi(z)|w]|z]", iteration = iter.label)
      if(is.null(bw)) {
          hphiwz <- .np_progress_with_legacy_suppressed(glpcv(ydat=E.phi.w,
                                                             xdat=z,
                                                             degree=rep(p, num.z.numeric),
                                                             nmulti=nmulti.loop,
                                                             random.seed=random.seed,
                                                             optim.maxattempts=optim.maxattempts,
                                                             optim.method=optim.method,
                                                             optim.reltol=optim.reltol,
                                                             optim.abstol=optim.abstol,
                                                             optim.maxit=optim.maxit,
                                                             bw.init=bw.E.E.phi.w.z,
                                                             ...))

          bw.E.E.phi.w.z <- hphiwz$bw
      } else {
          bw.E.E.phi.w.z <- bw$bw.E.E.phi.w.z
      }

      E.E.phi.w.z <- .np_progress_with_legacy_suppressed(glpreg(tydat=E.y.w,
                                                                txdat=z,
                                                                bws=bw.E.E.phi.w.z,
                                                                degree=rep(p, num.z.numeric),
                                                                ...))$mean
      
      KPHIW <- Kmat.lp(mydata.train=data.frame(w),
                       bws=bw.E.phi.w,
                       p=rep(p, num.w.numeric),
                       ...)

      KPHIWZ <- Kmat.lp(mydata.train=data.frame(z),
                        bws=bw.E.E.phi.w.z,
                        p=rep(p, num.z.numeric),
                        ...)

      ## Next, we minimize the function ittik to obtain the optimal value
      ## of alpha (here we use the iterated Tikhonov approach) to
      ## determine the optimal alpha for the non-iterated scheme.

      if(!is.null(bw)) alpha.iter <- bw$alpha.iter

      if(!iterate.Tikhonov) {
          alpha.iter <- alpha
      } else {

          if(is.null(alpha.iter)&&is.null(bw)) {
              iv_set_stage("alpha", iteration = iter.label)
              alpha.iter <- optimize(ittik, c(alpha.min, alpha.max), tol = alpha.tol, CZ = KPHIW, CY = KPHIWZ, Cr.r = E.E.phi.w.z, r = E.y.w)$minimum
          }
      }

      ## Finally, we conduct regularized Tikhonov regression using this
      ## optimal alpha and the updated bandwidths.

      iv_set_stage("phi(z)", iteration = iter.label)
      phi <- as.vector(tikh.eval(alpha.iter, CZ = KPHIW, CY = KPHIWZ, CY.eval = KPHIWZ, r = E.y.w))
      phi.mat <- cbind(phi.mat,phi)

      H <- NULL
      if(return.weights.phi) {
          H <- KPHIWZ%*%solve(alpha.iter*diag(nrow(KPHIWZ)) + KPHIW%*%KPHIWZ)%*%KYW
      }
      
      ## First derivative
      
      KPHIWZ.deriv.1 <- Kmat.lp(deriv=1,
                                mydata.train=data.frame(z),
                                bws=bw.E.E.phi.w.z,
                                p=rep(p, num.z.numeric),
                                ...)
      
      phi.deriv.1 <- as.vector(tikh.eval(alpha.iter, CZ = KPHIW, CY = KPHIWZ, CY.eval = KPHIWZ.deriv.1, r = E.y.w))
      
      H.deriv.1 <- NULL
      if(return.weights.phi.deriv.1) {
          H.deriv.1 <- KPHIWZ.deriv.1%*%solve(alpha.iter*diag(nrow(KPHIWZ)) + KPHIW%*%KPHIWZ)%*%KYW
      }
      
      ## Second derivative
      
      phi.deriv.2 <- NULL
      H.deriv.2 <- NULL
      
      if(p >= 2) {
          
          KPHIWZ.deriv.2 <- Kmat.lp(deriv=2,
                                    mydata.train=data.frame(z),
                                    bws=bw.E.E.phi.w.z,
                                    p=rep(p, num.z.numeric),
                                    ...)
          
          phi.deriv.2 <- as.vector(tikh.eval(alpha.iter, CZ = KPHIW, CY = KPHIWZ, CY.eval = KPHIWZ.deriv.2, r = E.y.w))
          
          
          if(return.weights.phi.deriv.2) {
              H.deriv.2 <- KPHIWZ.deriv.2%*%solve(alpha.iter*diag(nrow(KPHIWZ)) + KPHIW%*%KPHIWZ)%*%KYW
          }
          
      }
      
      ## If evaluation data are provided...
      
      phi.eval <- NULL
      phi.deriv.eval.1 <- NULL
      phi.deriv.eval.2 <- NULL
      H.eval <- NULL
      H.deriv.eval.1 <- NULL
      H.deriv.eval.2 <- NULL
      
      if(!is.null(zeval)) {
          ## If there is evaluation data, KPHIWZ and KPHIWZ.eval will
          ## differ...
          
          KPHIWZ.eval <- Kmat.lp(mydata.train=data.frame(z),
                                 mydata.eval=data.frame(z=zeval),
                                 bws=bw.E.E.phi.w.z,
                                 p=rep(p, num.z.numeric),
                                 ...)
          
          phi.eval <- as.vector(tikh.eval(alpha.iter, CZ = KPHIW, CY = KPHIWZ, CY.eval = KPHIWZ.eval, r = E.y.w))
          phi.eval.mat <- cbind(phi.eval.mat,phi.eval)
          
          if(return.weights.phi) {
              H.eval <- KPHIWZ.eval%*%solve(alpha.iter*diag(nrow(KPHIWZ)) + KPHIW%*%KPHIWZ)%*%KYW
          }
          
          KPHIWZ.eval.deriv.1 <- Kmat.lp(deriv=1,
                                         mydata.train=data.frame(z),
                                         mydata.eval=data.frame(z=zeval),
                                         bws=bw.E.E.phi.w.z,
                                         p=rep(p, num.z.numeric),
                                         ...)
          
          phi.deriv.eval.1 <- as.vector(tikh.eval(alpha.iter, CZ = KPHIW, CY = KPHIWZ, CY.eval = KPHIWZ.eval.deriv.1, r = E.y.w))
          
          if(return.weights.phi.deriv.1) {
              H.deriv.eval.1 <- KPHIWZ.eval.deriv.1%*%solve(alpha.iter*diag(nrow(KPHIWZ)) + KPHIW%*%KPHIWZ)%*%KYW
          }
          
          if(p >= 2) {
              
              KPHIWZ.eval.deriv.2 <- Kmat.lp(deriv=2,
                                             mydata.train=data.frame(z),
                                             mydata.eval=data.frame(z=zeval),
                                             bws=bw.E.E.phi.w.z,
                                             p=rep(p, num.z.numeric),
                                             ...)
              
              phi.deriv.eval.2 <- as.vector(tikh.eval(alpha.iter, CZ = KPHIW, CY = KPHIWZ, CY.eval = KPHIWZ.eval.deriv.2, r = E.y.w))
              
              if(return.weights.phi.deriv.2) {
                  H.deriv.eval.2 <- KPHIWZ.eval.deriv.2%*%solve(alpha.iter*diag(nrow(KPHIWZ)) + KPHIW%*%KPHIWZ)%*%KYW
              }
              
          }
      }
      
    }
    
    if((alpha.iter-alpha.min)/NZD(alpha.min) < 0.01) .np_warning(paste("Tikhonov parameter alpha (",formatC(alpha.iter,digits=4,format="f"),") is close to the search minimum (",alpha.min,")",sep=""))
    if((alpha.max-alpha.iter)/NZD(alpha.max) < 0.01) .np_warning(paste("Tikhonov parameter alpha (",formatC(alpha.iter,digits=4,format="f"),") is close to the search maximum (",alpha.max,")",sep=""))

    ret <- list(phi=phi,
                phi.eval=phi.eval,
                phi.mat=phi.mat,
                phi.eval.mat=phi.eval.mat,
                phi.deriv.1=as.matrix(phi.deriv.1),
                phi.deriv.eval.1=if(!is.null(phi.deriv.eval.1)){as.matrix(phi.deriv.eval.1)}else{NULL},
                phi.deriv.2=if(!is.null(phi.deriv.2)){as.matrix(phi.deriv.2)}else{NULL},
                phi.deriv.eval.2=if(!is.null(phi.deriv.eval.2)){as.matrix(phi.deriv.eval.2)}else{NULL},
                phi.weights=H,
                phi.deriv.1.weights=H.deriv.1,
                phi.deriv.2.weights=H.deriv.2,
                phi.eval.weights=H.eval,
                phi.deriv.eval.1.weights=H.deriv.eval.1,
                phi.deriv.eval.2.weights=H.deriv.eval.2,
                alpha=alpha,
                alpha.iter=alpha.iter,
                bw.E.y.w=bw.E.y.w,
                bw.E.E.y.w.z=bw.E.E.y.w.z,
                bw.E.phi.w=bw.E.phi.w,
                bw.E.E.phi.w.z=bw.E.E.phi.w.z,
                call=cl,
                y=y,
                z=z,
                w=w,
                x=x,
                zeval=zeval,
                xeval=xeval,
                p=p,
                nmulti=nmulti,
                method=method,
                ptm=proc.time() - ptm.start)
    class(ret) <- "npregiv"
    ret

  } else {

    ## Landweber-Fridman

    norm.stop <- numeric()

    iv_set_stage("E[y|w]")
    if(is.null(bw)) {
        h <- .np_progress_with_legacy_suppressed(glpcv(ydat=y,
                                                      xdat=w,
                                                      degree=rep(p, num.w.numeric),
                                                      nmulti=nmulti,
                                                      random.seed=random.seed,
                                                      optim.maxattempts=optim.maxattempts,
                                                      optim.method=optim.method,
                                                      optim.reltol=optim.reltol,
                                                      optim.abstol=optim.abstol,
                                                      optim.maxit=optim.maxit,
                                                      ...))
        bw.E.y.w <- h$bw
    } else {
        bw.E.y.w <- bw$bw.E.y.w
    }

    E.y.w <- .np_progress_with_legacy_suppressed(glpreg(tydat=y,
                                                        txdat=w,
                                                        bws=bw.E.y.w,
                                                        degree=rep(p, num.w.numeric),
                                                        ...))$mean

    if(return.weights.phi) {

        if(p<0) stop("glp return weights not supported")
        if(NCOL(z) > 1) stop("dimension of z must be one for currently supported return weights")

        T.mat.r <- Kmat.lp(mydata.train=data.frame(w),
                           bws=bw.E.y.w,
                           p=rep(p, num.w.numeric),
                           ...)

    }

    ## We begin the iteration computing phi.0 and phi.1 directly, then
    ## iterate.

    if(is.null(starting.values)) {
      iv_set_stage(iv_start_label())

      if(is.null(bw)) {
          h <- .np_progress_with_legacy_suppressed(glpcv(ydat=if(start.from=="Eyz") y else E.y.w,
                                                        xdat=z,
                                                        degree=rep(p, num.z.numeric),
                                                        nmulti=nmulti,
                                                        random.seed=random.seed,
                                                        optim.maxattempts=optim.maxattempts,
                                                        optim.method=optim.method,
                                                        optim.reltol=optim.reltol,
                                                        optim.abstol=optim.abstol,
                                                        optim.maxit=optim.maxit,
                                                        ...))
          bw.E.y.z <- h$bw
      } else {
          bw.E.y.z <- bw$bw.E.y.z
      }

      g <- .np_progress_with_legacy_suppressed(glpreg(tydat=if(start.from=="Eyz") y else E.y.w,
                                                      txdat=z,
                                                      bws=bw.E.y.z,
                                                      degree=rep(p, num.z.numeric),
                                                      ...))

      phi.0 <- g$mean
      phi.0.deriv.1 <- g$grad
      if(p >= 2) {
          phi.0.deriv.2 <- .np_progress_with_legacy_suppressed(glpreg(tydat=if(start.from=="Eyz") y else E.y.w,
                                                                      txdat=z,
                                                                      bws=bw.E.y.z,
                                                                      degree=rep(p, num.z.numeric),
                                                                      deriv=2,
                                                                      ...))$grad
      }

      if(!is.null(zeval)) {
          g <- .np_progress_with_legacy_suppressed(glpreg(tydat=if(start.from=="Eyz") y else E.y.w,
                                                          txdat=z,
                                                          exdat=zeval,
                                                          bws=bw.E.y.z,
                                                          degree=rep(p, num.z.numeric),
                                                          ...))

          phi.eval.0 <- g$mean
          phi.eval.0.deriv.1 <- g$grad
          if(p >= 2) {
              phi.eval.0.deriv.2 <- .np_progress_with_legacy_suppressed(glpreg(tydat=if(start.from=="Eyz") y else E.y.w,
                                                                               txdat=z,
                                                                               exdat=zeval,
                                                                               bws=bw.E.y.z,
                                                                               degree=rep(p, num.z.numeric),
                                                                               deriv=2,
                                                                               ...))$grad
          }
      } else {
          phi.eval.0 <- NULL
          phi.eval.0.deriv.1 <- NULL
          phi.eval.0.deriv.2 <- NULL
      }

      if(return.weights.phi) {

          H <- Kmat.lp(mydata.train=data.frame(z),
                       bws=bw.E.y.z,
                       p=rep(p, num.z.numeric),
                       ...)

          if(!is.null(zeval)) H.eval <- Kmat.lp(mydata.train=data.frame(z),
                                                mydata.eval=data.frame(z=zeval),
                                                bws=bw.E.y.z,
                                                p=rep(p, num.z.numeric),
                                                ...)

          if(p==0 || p==1) {

              if(return.weights.phi.deriv.1) {

                  H.deriv.1 <- Kmat.lp(deriv=1,
                                       mydata.train=data.frame(z),
                                       bws=bw.E.y.z,
                                       p=rep(p, num.z.numeric),
                                       ...)

                  if(!is.null(zeval)) H.deriv.eval.1 <- Kmat.lp(deriv=1,
                                                                mydata.train=data.frame(z),
                                                                mydata.eval=data.frame(z=zeval),
                                                                bws=bw.E.y.z,
                                                                p=rep(p, num.z.numeric),
                                                                ...)
              }

          } else {

              if(return.weights.phi.deriv.1) {

                  H.deriv.1 <- Kmat.lp(deriv=1,
                                       mydata.train=data.frame(z),
                                       bws=bw.E.y.z,
                                       p=rep(p, num.z.numeric),
                                       ...)
                  if(!is.null(zeval)) {

                      H.deriv.eval.1 <- Kmat.lp(deriv=1,
                                                mydata.train=data.frame(z),
                                                mydata.eval=data.frame(z=zeval),
                                                bws=bw.E.y.z,
                                                p=rep(p, num.z.numeric),
                                                ...)

                  }

              }

              if(return.weights.phi.deriv.2) {

                  H.deriv.2 <- Kmat.lp(deriv=2,
                                       mydata.train=data.frame(z),
                                       bws=bw.E.y.z,
                                       p=rep(p, num.z.numeric),
                                       ...)

                  if(!is.null(zeval)) {

                      H.deriv.eval.2 <- Kmat.lp(deriv=2,
                                                mydata.train=data.frame(z),
                                                mydata.eval=data.frame(z=zeval),
                                                bws=bw.E.y.z,
                                                p=rep(p, num.z.numeric),
                                                ...)

                  }

              }

          }

      }

    } else {

      ## Starting values input by user

      phi.0 <- starting.values

      if(return.weights.phi)  H <- NULL
      bw.E.y.z <- NULL

    }

    starting.values.phi <- phi.0

    iv_set_stage(iv_residual_stage_label(smooth.residuals), iteration = 1L)

    if(smooth.residuals) {

      resid <- y - phi.0

      if(is.null(bw)) {
          h <- .np_progress_with_legacy_suppressed(glpcv(ydat=resid,
                                                        xdat=w,
                                                        degree=rep(p, num.w.numeric),
                                                        nmulti=nmulti,
                                                        random.seed=random.seed,
                                                        optim.maxattempts=optim.maxattempts,
                                                        optim.method=optim.method,
                                                        optim.reltol=optim.reltol,
                                                        optim.abstol=optim.abstol,
                                                        optim.maxit=optim.maxit,
                                                        ...))
          bw.resid.w <- h$bw
      } else {
          bw.resid.w <- bw$bw.resid.w[1,]
      }

      resid.fitted <- .np_progress_with_legacy_suppressed(glpreg(tydat=resid,
                                                                 txdat=w,
                                                                 bws=bw.resid.w,
                                                                 degree=rep(p, num.w.numeric),
                                                                 ...))$mean

    } else {

      if(is.null(bw)) {
          h <- .np_progress_with_legacy_suppressed(glpcv(ydat=phi.0,
                                                        xdat=w,
                                                        degree=rep(p, num.w.numeric),
                                                        nmulti=nmulti,
                                                        random.seed=random.seed,
                                                        optim.maxattempts=optim.maxattempts,
                                                        optim.method=optim.method,
                                                        optim.reltol=optim.reltol,
                                                        optim.abstol=optim.abstol,
                                                        optim.maxit=optim.maxit,
                                                        ...))
          bw.resid.w <- h$bw
      } else {
          bw.resid.w <- bw$bw.resid.w[1,]
      }

      resid.fitted <- E.y.w - .np_progress_with_legacy_suppressed(glpreg(tydat=phi.0,
                                                                         txdat=w,
                                                                         bws=bw.resid.w,
                                                                         degree=rep(p, num.w.numeric),
                                                                         ...))$mean

    }

    if(return.weights.phi) {

        T.mat <- Kmat.lp(mydata.train=data.frame(w),
                         bws=bw.resid.w,
                         p=rep(p, num.z.numeric),
                         ...)

    }

    norm.stop[1] <- sum(resid.fitted^2)/NZD_pos(sum(E.y.w^2))

    iv_set_stage(iv_adjoint_stage_label(smooth.residuals), iteration = 1L)

    if(is.null(bw)) {
        h <- .np_progress_with_legacy_suppressed(glpcv(ydat=resid.fitted,
                                                      xdat=z,
                                                      degree=rep(p, num.z.numeric),
                                                      nmulti=nmulti,
                                                      random.seed=random.seed,
                                                      optim.maxattempts=optim.maxattempts,
                                                      optim.method=optim.method,
                                                      optim.reltol=optim.reltol,
                                                      optim.abstol=optim.abstol,
                                                      optim.maxit=optim.maxit,
                                                      ...))
        bw.resid.fitted.w.z <- h$bw
    } else {
        bw.resid.fitted.w.z <- bw$bw.resid.fitted.w.z[1,]
    }

    g <- .np_progress_with_legacy_suppressed(glpreg(tydat=resid.fitted,
                                                    txdat=z,
                                                    bws=bw.resid.fitted.w.z,
                                                    degree=rep(p, num.z.numeric),
                                                    ...))

    phi.deriv.1.list <- vector(mode="list", length=iterate.max)
    phi.deriv.eval.1.list <- vector(mode="list", length=iterate.max)

    phi <- phi.0 + constant*g$mean
    phi.deriv.1 <- phi.0.deriv.1 + constant*g$grad

    phi.mat <- matrix(NA, length(phi), iterate.max)
    phi.mat[,1] <- phi
    phi.deriv.1.list[[1]] <- phi.deriv.1

    phi.deriv.2.list <- vector(mode="list", length=iterate.max)
    phi.deriv.eval.2.list <- vector(mode="list", length=iterate.max)
    phi.deriv.2 <- NULL
    phi.deriv.eval.2 <- NULL

    if(p >= 2) {
        phi.deriv.2 <- phi.0.deriv.2 + constant*.np_progress_with_legacy_suppressed(glpreg(tydat=resid.fitted,
                                                                                            txdat=z,
                                                                                            bws=bw.resid.fitted.w.z,
                                                                                            degree=rep(p, num.z.numeric),
                                                                                            deriv=2,
                                                                                            ...))$grad
        phi.deriv.2.list[[1]] <- phi.deriv.2
    }

    if(!is.null(zeval)) {
        g <- .np_progress_with_legacy_suppressed(glpreg(tydat=resid.fitted,
                                                        txdat=z,
                                                        exdat=zeval,
                                                        bws=bw.resid.fitted.w.z,
                                                        degree=rep(p, num.z.numeric),
                                                        ...))
        phi.eval <- phi.eval.0 + constant*g$mean
        phi.eval.mat <- matrix(NA, length(phi.eval), iterate.max)
        phi.eval.mat[,1] <- phi.eval

        phi.deriv.eval.1 <- phi.eval.0.deriv.1 + constant*g$grad
        phi.deriv.eval.1.list[[1]] <- phi.deriv.eval.1

        if(p >= 2) {
            phi.deriv.eval.2 <- phi.eval.0.deriv.2 + constant*.np_progress_with_legacy_suppressed(glpreg(tydat=resid.fitted,
                                                                                                          txdat=z,
                                                                                                          bws=bw.resid.fitted.w.z,
                                                                                                          exdat=zeval,
                                                                                                          degree=rep(p, num.z.numeric),
                                                                                                          deriv=2,
                                                                                                          ...))$grad
            phi.deriv.eval.2.list[[1]] <- phi.deriv.eval.2
        }



    } else {
        phi.eval <- NULL
        phi.eval.mat <- NULL
        phi.deriv.eval.1 <- NULL
        phi.deriv.eval.1.list <- NULL
    }

    ## Need these list even when no weights for return

    phi.weights.list <- vector(mode="list", length=iterate.max)
    phi.deriv.1.weights.list <- vector(mode="list", length=iterate.max)
    phi.deriv.2.weights.list <- vector(mode="list", length=iterate.max)

    phi.eval.weights.list <- vector(mode="list", length=iterate.max)
    phi.deriv.eval.1.weights.list <- vector(mode="list", length=iterate.max)
    phi.deriv.eval.2.weights.list <- vector(mode="list", length=iterate.max)

    ## Pre-allocate bandwidth matrices
    bw.resid.w.mat <- matrix(NA, iterate.max, length(bw.resid.w))
    bw.resid.w.mat[1,] <- bw.resid.w
    bw.resid.fitted.w.z.mat <- matrix(NA, iterate.max, length(bw.resid.fitted.w.z))
    bw.resid.fitted.w.z.mat[1,] <- bw.resid.fitted.w.z

    if(return.weights.phi) {

        T.mat.adjoint <- Kmat.lp(mydata.train=data.frame(z),
                                 bws=bw.resid.fitted.w.z,
                                 p=rep(p, num.z.numeric),
                                 ...)

        if(!is.null(zeval)) T.mat.adjoint.eval <- Kmat.lp(mydata.train=data.frame(z),
                                                          mydata.eval=data.frame(z=zeval),
                                                          bws=bw.resid.fitted.w.z,
                                                          p=rep(p, num.z.numeric),
                                                          ...)
        if(p==0 || p==1) {

            if(return.weights.phi.deriv.1) {

                T.mat.adjoint.deriv.1 <- Kmat.lp(deriv=1,
                                                 mydata.train=data.frame(z),
                                                 bws=bw.resid.fitted.w.z,
                                                 p=rep(p, num.z.numeric),
                                                 ...)

                if(!is.null(zeval)) T.mat.adjoint.deriv.eval.1 <- Kmat.lp(deriv=1,
                                                                          mydata.train=data.frame(z),
                                                                          mydata.eval=data.frame(z=zeval),
                                                                          bws=bw.resid.fitted.w.z,
                                                                          p=rep(p, num.z.numeric),
                                                                          ...)
            }

        } else {

            if(return.weights.phi.deriv.1) {

                T.mat.adjoint.deriv.1 <- Kmat.lp(deriv=1,
                                                 mydata.train=data.frame(z),
                                                 bws=bw.resid.fitted.w.z,
                                                 p=rep(p, num.z.numeric),
                                                 ...)

                if(!is.null(zeval)) {

                    T.mat.adjoint.deriv.eval.1 <- Kmat.lp(deriv=1,
                                                          mydata.train=data.frame(z),
                                                          mydata.eval=data.frame(z=zeval),
                                                          bws=bw.resid.fitted.w.z,
                                                          p=rep(p, num.z.numeric),
                                                          ...)

                }

            }

            if(return.weights.phi.deriv.2) {

                T.mat.adjoint.deriv.2 <- Kmat.lp(deriv=2,
                                                 mydata.train=data.frame(z),
                                                 bws=bw.resid.fitted.w.z,
                                                 p=rep(p, num.z.numeric),
                                                 ...)

                if(!is.null(zeval)) {

                    T.mat.adjoint.deriv.eval.2 <- Kmat.lp(deriv=2,
                                                          mydata.train=data.frame(z),
                                                          mydata.eval=data.frame(z=zeval),
                                                          bws=bw.resid.fitted.w.z,
                                                          p=rep(p, num.z.numeric),
                                                          ...)

                }

            }

        }

        if(smooth.residuals) {
            if(!is.null(zeval)) {
                H.eval <- H.eval + constant*T.mat.adjoint.eval%*%(T.mat-T.mat%*%H)
                if(return.weights.phi.deriv.1) H.deriv.eval.1 <- H.deriv.eval.1 + constant*T.mat.adjoint.deriv.eval.1%*%(T.mat-T.mat%*%H)
                if(p>1 && return.weights.phi.deriv.2) H.deriv.eval.2 <- H.deriv.eval.2 + constant*T.mat.adjoint.deriv.eval.2%*%(T.mat-T.mat%*%H)
            }
            if(return.weights.phi.deriv.1) H.deriv.1 <- H.deriv.1 + constant*T.mat.adjoint.deriv.1%*%(T.mat-T.mat%*%H)
            if(p>1  && return.weights.phi.deriv.2) H.deriv.2 <- H.deriv.2 + constant*T.mat.adjoint.deriv.2%*%(T.mat-T.mat%*%H)
            H <- H + constant*T.mat.adjoint%*%(T.mat-T.mat%*%H)
        } else {
            if(!is.null(zeval)) {
                H.eval <- H.eval + constant*T.mat.adjoint.eval%*%(T.mat.r-T.mat%*%H)
                if(return.weights.phi.deriv.1) H.deriv.eval.1 <- H.deriv.eval.1 + constant*T.mat.adjoint.deriv.eval.1%*%(T.mat.r-T.mat%*%H)
                if(p>1 && return.weights.phi.deriv.2) H.deriv.eval.2 <- H.deriv.eval.2 + constant*T.mat.adjoint.deriv.eval.2%*%(T.mat.r-T.mat%*%H)
            }
            if(return.weights.phi.deriv.1) H.deriv.1 <- H.deriv.1 + constant*T.mat.adjoint.deriv.1%*%(T.mat.r-T.mat%*%H)
            if(p>1 && return.weights.phi.deriv.2) H.deriv.2 <- H.deriv.2 + constant*T.mat.adjoint.deriv.2%*%(T.mat.r-T.mat%*%H)
            H <- H + constant*T.mat.adjoint%*%(T.mat.r-T.mat%*%H)
        }

        phi.weights.list[[1]] <- H
        if(return.weights.phi.deriv.1) phi.deriv.1.weights.list[[1]] <- H.deriv.1
        if(p>1 && return.weights.phi.deriv.2) phi.deriv.2.weights.list[[1]] <- H.deriv.2
        if(!is.null(zeval)) {
            phi.eval.weights.list[[1]] <- H.eval
            if(return.weights.phi.deriv.1) phi.deriv.eval.1.weights.list[[1]] <- H.deriv.eval.1
            if(p>1 && return.weights.phi.deriv.2) phi.deriv.eval.2.weights.list[[1]] <- H.deriv.eval.2
        }

    }

    if(!is.null(bw)) iterate.max <- bw$norm.index

    ## In what follows we rbind() bandwidths to return and are careful
    ## about which ones are used when fed in, so we use h$bw below
    ## (but above all are named).

    for(j in 2:iterate.max) {
      iv_set_stage(iv_residual_stage_label(smooth.residuals), iteration = j)

      if(smooth.residuals) {

        resid <- y - phi

        if(is.null(bw)) {
            h <- .np_progress_with_legacy_suppressed(glpcv(ydat=resid,
                                                          xdat=w,
                                                          degree=rep(p, num.w.numeric),
                                                          nmulti=nmulti.loop,
                                                          random.seed=random.seed,
                                                          optim.maxattempts=optim.maxattempts,
                                                          optim.method=optim.method,
                                                          optim.reltol=optim.reltol,
                                                          optim.abstol=optim.abstol,
                                                          optim.maxit=optim.maxit,
                                                          bw.init=bw.resid.w.mat[j-1,],
                                                          ...))
        } else {
            h <- NULL
            h$bw <- bw$bw.resid.w[j,]
        }

        resid.fitted <- .np_progress_with_legacy_suppressed(glpreg(tydat=resid,
                                                                   txdat=w,
                                                                   bws=h$bw,
                                                                   degree=rep(p, num.w.numeric),
                                                                   ...))$mean


      } else {

        if(is.null(bw)) {
            h <- .np_progress_with_legacy_suppressed(glpcv(ydat=phi,
                                                          xdat=w,
                                                          degree=rep(p, num.w.numeric),
                                                          nmulti=nmulti.loop,
                                                          random.seed=random.seed,
                                                          optim.maxattempts=optim.maxattempts,
                                                          optim.method=optim.method,
                                                          optim.reltol=optim.reltol,
                                                          optim.abstol=optim.abstol,
                                                          optim.maxit=optim.maxit,
                                                          bw.init=bw.resid.w.mat[j-1,],
                                                          ...))
        } else {
            h <- NULL
            h$bw <- bw$bw.resid.w[j,]
        }

        resid.fitted <- E.y.w - .np_progress_with_legacy_suppressed(glpreg(tydat=phi,
                                                                           txdat=w,
                                                                           bws=h$bw,
                                                                           degree=rep(p, num.w.numeric),
                                                                           ...))$mean

      }

      bw.resid.w.mat[j,] <- h$bw

      if(return.weights.phi) {

          T.mat <- Kmat.lp(mydata.train=data.frame(w),
                           bws=h$bw,
                           p=rep(p, num.w.numeric),
                           ...)

      }

      norm.stop[j] <- if (penalize.iteration) j*sum(resid.fitted^2)/NZD_pos(sum(E.y.w^2)) else sum(resid.fitted^2)/NZD_pos(sum(E.y.w^2))

      iv_set_stage(iv_adjoint_stage_label(smooth.residuals), iteration = j)

      if(is.null(bw)) {
          h <- .np_progress_with_legacy_suppressed(glpcv(ydat=resid.fitted,
                                                        xdat=z,
                                                        degree=rep(p, num.z.numeric),
                                                        nmulti=nmulti.loop,
                                                        random.seed=random.seed,
                                                        optim.maxattempts=optim.maxattempts,
                                                        optim.method=optim.method,
                                                        optim.reltol=optim.reltol,
                                                        optim.abstol=optim.abstol,
                                                        optim.maxit=optim.maxit,
                                                        bw.init=bw.resid.fitted.w.z.mat[j-1,],
                                                        ...))
      } else {
          h$bw <- bw$bw.resid.fitted.w.z[j,]
      }

      g <- .np_progress_with_legacy_suppressed(glpreg(tydat=resid.fitted,
                                                      txdat=z,
                                                      bws=h$bw,
                                                      degree=rep(p, num.z.numeric),
                                                      ...))

      phi <- phi + constant*g$mean
      phi.mat[,j] <- phi
      phi.deriv.1 <- phi.deriv.1 + constant*g$grad
      phi.deriv.1.list[[j]] <- phi.deriv.1

      if(p >= 2) {
          phi.deriv.2 <- phi.deriv.2 + constant*.np_progress_with_legacy_suppressed(glpreg(tydat=resid.fitted,
                                                                                            txdat=z,
                                                                                            bws=h$bw,
                                                                                            degree=rep(p, num.z.numeric),
                                                                                            deriv=2,
                                                                                            ...))$grad
          phi.deriv.2.list[[j]] <- phi.deriv.2
      }

      if(!is.null(zeval)) {
          g <- .np_progress_with_legacy_suppressed(glpreg(tydat=resid.fitted,
                                                          txdat=z,
                                                          exdat=zeval,
                                                          bws=h$bw,
                                                          degree=rep(p, num.z.numeric),
                                                          ...))
          phi.eval <- phi.eval + constant*g$mean
          phi.eval.mat[,j] <- phi.eval

          phi.deriv.eval.1 <- phi.deriv.eval.1 + constant*g$grad
          phi.deriv.eval.1.list[[j]] <- phi.deriv.eval.1
          if(p >= 2) {
              phi.deriv.eval.2 <- phi.deriv.eval.2 + constant*.np_progress_with_legacy_suppressed(glpreg(tydat=resid.fitted,
                                                                                                          txdat=z,
                                                                                                          exdat=zeval,
                                                                                                          bws=h$bw,
                                                                                                          degree=rep(p, num.z.numeric),
                                                                                                          deriv=2,
                                                                                                          ...))$grad
              phi.deriv.eval.2.list[[j]] <- phi.deriv.eval.2
          }

      }

      bw.resid.fitted.w.z.mat[j,] <- h$bw

      if(return.weights.phi) {

          T.mat.adjoint <- Kmat.lp(mydata.train=data.frame(z),
                                   bws=h$bw,
                                   p=rep(p, num.z.numeric),
                                   ...)

          if(!is.null(zeval)) T.mat.adjoint.eval <- Kmat.lp(mydata.train=data.frame(z),
                                                            mydata.eval=data.frame(z=zeval),
                                                            bws=h$bw,
                                                            p=rep(p, num.z.numeric),
                                                            ...)
          if(p==0 || p==1) {

              if(return.weights.phi.deriv.1) {

                  T.mat.adjoint.deriv.1 <- Kmat.lp(deriv=1,
                                                   mydata.train=data.frame(z),
                                                   bws=h$bw,
                                                   p=rep(p, num.z.numeric),
                                                   ...)

                  if(!is.null(zeval)) T.mat.adjoint.deriv.eval.1 <- Kmat.lp(deriv=1,
                                                                            mydata.train=data.frame(z),
                                                                            mydata.eval=data.frame(z=zeval),
                                                                            bws=h$bw,
                                                                            p=rep(p, num.z.numeric),
                                                                            ...)
              }

          } else {

              if(return.weights.phi.deriv.1) {

                  T.mat.adjoint.deriv.1 <- Kmat.lp(deriv=1,
                                                   mydata.train=data.frame(z),
                                                   bws=h$bw,
                                                   p=rep(p, num.z.numeric),
                                                   ...)

                  if(!is.null(zeval)) {

                      T.mat.adjoint.deriv.eval.1 <- Kmat.lp(deriv=1,
                                                            mydata.train=data.frame(z),
                                                            mydata.eval=data.frame(z=zeval),
                                                            bws=h$bw,
                                                            p=rep(p, num.z.numeric),
                                                            ...)

                  }

              }

              if(return.weights.phi.deriv.2) {

                  T.mat.adjoint.deriv.2 <- Kmat.lp(deriv=2,
                                                   mydata.train=data.frame(z),
                                                   bws=h$bw,
                                                   p=rep(p, num.z.numeric),
                                                   ...)

                  if(!is.null(zeval)) {

                      T.mat.adjoint.deriv.eval.2 <- Kmat.lp(deriv=2,
                                                            mydata.train=data.frame(z),
                                                            mydata.eval=data.frame(z=zeval),
                                                            bws=h$bw,
                                                            p=rep(p, num.z.numeric),
                                                            ...)

                  }

              }

          }

          if(smooth.residuals) {
              if(!is.null(zeval)) {
                  H.eval <- H.eval + constant*T.mat.adjoint.eval%*%(T.mat-T.mat%*%H)
                  if(return.weights.phi.deriv.1) H.deriv.eval.1 <- H.deriv.eval.1 + constant*T.mat.adjoint.deriv.eval.1%*%(T.mat-T.mat%*%H)
                  if(p>1 && return.weights.phi.deriv.2) H.deriv.eval.2 <- H.deriv.eval.2 + constant*T.mat.adjoint.deriv.eval.2%*%(T.mat-T.mat%*%H)
              }
              if(return.weights.phi.deriv.1) H.deriv.1 <- H.deriv.1 + constant*T.mat.adjoint.deriv.1%*%(T.mat-T.mat%*%H)
              if(p>1  && return.weights.phi.deriv.2) H.deriv.2 <- H.deriv.2 + constant*T.mat.adjoint.deriv.2%*%(T.mat-T.mat%*%H)
              H <- H + constant*T.mat.adjoint%*%(T.mat-T.mat%*%H)
          } else {
              if(!is.null(zeval)) {
                  H.eval <- H.eval + constant*T.mat.adjoint.eval%*%(T.mat.r-T.mat%*%H)
                  if(return.weights.phi.deriv.1) H.deriv.eval.1 <- H.deriv.eval.1 + constant*T.mat.adjoint.deriv.eval.1%*%(T.mat.r-T.mat%*%H)
                  if(p>1 && return.weights.phi.deriv.2) H.deriv.eval.2 <- H.deriv.eval.2 + constant*T.mat.adjoint.deriv.eval.2%*%(T.mat.r-T.mat%*%H)
              }
              if(return.weights.phi.deriv.1) H.deriv.1 <- H.deriv.1 + constant*T.mat.adjoint.deriv.1%*%(T.mat.r-T.mat%*%H)
              if(p>1 && return.weights.phi.deriv.2) H.deriv.2 <- H.deriv.2 + constant*T.mat.adjoint.deriv.2%*%(T.mat.r-T.mat%*%H)
              H <- H + constant*T.mat.adjoint%*%(T.mat.r-T.mat%*%H)
          }

          phi.weights.list[[j]] <- H
          if(return.weights.phi.deriv.1) phi.deriv.1.weights.list[[j]] <- H.deriv.1
          if(p>1 && return.weights.phi.deriv.2) phi.deriv.2.weights.list[[j]] <- H.deriv.2
          if(!is.null(zeval)) {
              phi.eval.weights.list[[j]] <- H.eval
              if(return.weights.phi.deriv.1) phi.deriv.eval.1.weights.list[[j]] <- H.deriv.eval.1
              if(p>1 && return.weights.phi.deriv.2) phi.deriv.eval.2.weights.list[[j]] <- H.deriv.eval.2
          }

      }

      ## The number of iterations in LF is asymptotically equivalent
      ## to 1/alpha (where alpha is the regularization parameter in
      ## Tikhonov).  Plus the criterion function we use is increasing
      ## for very small number of iterations. So we need a threshold
      ## after which we can pretty much confidently say that the
      ## stopping criterion is decreasing.  In Darolles et al. (2011)
      ## \alpha ~ O(N^(-1/(min(beta,2)+2)), where beta is the so
      ## called qualification of your regularization method. Take the
      ## worst case in which beta = 0 and then the number of
      ## iterations is ~ N^0.5.

      if(is.null(bw))  {

          if(j > round(sqrt(nrow(z))) && !is.monotone.increasing(norm.stop)) {

              ## If stopping rule criterion increases or we are below stopping
              ## tolerance then break

              if(stop.on.increase && norm.stop[j] > norm.stop[j-1]) {
                  convergence <- "STOP_ON_INCREASE"
                  break()
              }
              if(abs(norm.stop[j-1]-norm.stop[j]) < iterate.diff.tol) {
                  convergence <- "ITERATE_DIFF_TOL"
                  break()
              }

          }

          convergence <- "ITERATE_MAX"

      }

    }

    ## Trim matrices and lists to the actual number of iterations performed
    if(j < iterate.max) {
        phi.mat <- phi.mat[, 1:j, drop=FALSE]
        if(!is.null(phi.eval.mat)) phi.eval.mat <- phi.eval.mat[, 1:j, drop=FALSE]
        norm.stop <- norm.stop[1:j]
        phi.deriv.1.list <- phi.deriv.1.list[1:j]
        phi.deriv.2.list <- phi.deriv.2.list[1:j]
        if(!is.null(phi.deriv.eval.1.list)) phi.deriv.eval.1.list <- phi.deriv.eval.1.list[1:j]
        if(!is.null(phi.deriv.eval.2.list)) phi.deriv.eval.2.list <- phi.deriv.eval.2.list[1:j]
        phi.weights.list <- phi.weights.list[1:j]
        phi.deriv.1.weights.list <- phi.deriv.1.weights.list[1:j]
        phi.deriv.2.weights.list <- phi.deriv.2.weights.list[1:j]
        phi.eval.weights.list <- phi.eval.weights.list[1:j]
        phi.deriv.eval.1.weights.list <- phi.deriv.eval.1.weights.list[1:j]
        phi.deriv.eval.2.weights.list <- phi.deriv.eval.2.weights.list[1:j]
        bw.resid.w.mat <- bw.resid.w.mat[1:j, , drop=FALSE]
        bw.resid.fitted.w.z.mat <- bw.resid.fitted.w.z.mat[1:j, , drop=FALSE]
    }

    bw.resid.w <- bw.resid.w.mat
    bw.resid.fitted.w.z <- bw.resid.fitted.w.z.mat

    ## Extract minimum, and check for monotone increasing function and
    ## then decreasing (and potentially increasing thereafter) portion
    ## of the stopping function, ignore the initial increasing portion,
    ## and take the min from where the initial inflection point occurs
    ## to the length of norm.stop

    phi.weights <- NULL
    phi.deriv.1.weights <- NULL
    phi.deriv.2.weights <- NULL
    
    phi.eval.weights <- NULL
    phi.deriv.eval.1.weights <- NULL
    phi.deriv.eval.2.weights <- NULL

    if(is.null(bw))  {

      norm.value <- norm.stop/seq_along(norm.stop)

      if(which.min(norm.stop) == 1 && is.monotone.increasing(norm.stop)) {
          .np_warning("Stopping rule increases monotonically (consult model$norm.stop):\nThis could be the result of an inspired initial value (unlikely)\nNote: we suggest manually choosing phi.0 and restarting (e.g. instead set `starting.values' to E[E(Y|w)|z])")
          convergence <- "FAILURE_MONOTONE_INCREASING"
          phi <- starting.values.phi
          j <- 1
          while(norm.value[j+1] > norm.value[j]) j <- j + 1
          j <- j-1 + which.min(norm.value[j:length(norm.value)])
          phi <- phi.mat[,j]
          if(p>0) phi.deriv.1 <- phi.deriv.1.list[[j]]
          if(p>=2) phi.deriv.2 <- phi.deriv.2.list[[j]]
          if(return.weights.phi) phi.weights <- phi.weights.list[[j]]
          if(return.weights.phi.deriv.1) phi.deriv.1.weights <- phi.deriv.1.weights.list[[j]]
          if(p>=2 && return.weights.phi.deriv.2) phi.deriv.2.weights <- phi.deriv.2.weights.list[[j]]
          if(!is.null(zeval)) {
              phi.eval <- phi.eval.mat[,j]
              if(p>0) phi.deriv.eval.1 <- phi.deriv.eval.1.list[[j]]
              if(p>=2) phi.deriv.eval.2 <- phi.deriv.eval.2.list[[j]]
              if(return.weights.phi) phi.eval.weights <- phi.eval.weights.list[[j]]
              if(return.weights.phi.deriv.1) phi.deriv.eval.1.weights <- phi.deriv.eval.1.weights.list[[j]]
              if(p>=2 && return.weights.phi.deriv.2) phi.deriv.eval.2.weights <- phi.deriv.eval.2.weights.list[[j]]    
          }
      } else {
          ## Ignore the initial increasing portion, take the min to the
          ## right of where the initial inflection point occurs
          j <- 1
          while(norm.stop[j+1] > norm.stop[j]) j <- j + 1
          j <- j-1 + which.min(norm.stop[j:length(norm.stop)])
          phi <- phi.mat[,j]
          if(p>0) phi.deriv.1 <- phi.deriv.1.list[[j]]
          if(p>=2) phi.deriv.2 <- phi.deriv.2.list[[j]]
          if(return.weights.phi) phi.weights <- phi.weights.list[[j]]
          if(return.weights.phi.deriv.1) phi.deriv.1.weights <- phi.deriv.1.weights.list[[j]]
          if(p>=2 && return.weights.phi.deriv.2) phi.deriv.2.weights <- phi.deriv.2.weights.list[[j]]
          if(!is.null(zeval)) {
              phi.eval <- phi.eval.mat[,j]
              if(p>0) phi.deriv.eval.1 <- phi.deriv.eval.1.list[[j]]
              if(p>=2) phi.deriv.eval.2 <- phi.deriv.eval.2.list[[j]]
              if(return.weights.phi) phi.eval.weights <- phi.eval.weights.list[[j]]
              if(return.weights.phi.deriv.1) phi.deriv.eval.1.weights <- phi.deriv.eval.1.weights.list[[j]]
              if(p>=2 && return.weights.phi.deriv.2) phi.deriv.eval.2.weights <- phi.deriv.eval.2.weights.list[[j]]    
          }
      }
      if(j == iterate.max) .np_warning("iterate.max reached: increase iterate.max or inspect norm.stop vector")      
    } else {
        ## bw passed in, set j to norm.index, push out weights etc.
        j <- bw$norm.index
        phi <- phi.mat[,j]
        if(p>0) phi.deriv.1 <- phi.deriv.1.list[[j]]
        if(p>=2) phi.deriv.2 <- phi.deriv.2.list[[j]]
        if(return.weights.phi) phi.weights <- phi.weights.list[[j]]
        if(return.weights.phi.deriv.1) phi.deriv.1.weights <- phi.deriv.1.weights.list[[j]]
        if(p>=2 && return.weights.phi.deriv.2) phi.deriv.2.weights <- phi.deriv.2.weights.list[[j]]
        if(!is.null(zeval)) {
            phi.eval <- phi.eval.mat[,j]
            if(p>0) phi.deriv.eval.1 <- phi.deriv.eval.1.list[[j]]
            if(p>=2) phi.deriv.eval.2 <- phi.deriv.eval.2.list[[j]]
            if(return.weights.phi) phi.eval.weights <- phi.eval.weights.list[[j]]
            if(return.weights.phi.deriv.1) phi.deriv.eval.1.weights <- phi.deriv.eval.1.weights.list[[j]]
            if(p>=2 && return.weights.phi.deriv.2) phi.deriv.eval.2.weights <- phi.deriv.eval.2.weights.list[[j]]    
        }
        norm.value <- NULL
        norm.stop <- NULL
        convergence <- NULL
    }
    
    ret <- list(phi=phi,
                phi.mat=phi.mat,
                phi.deriv.1=as.matrix(phi.deriv.1),
                phi.deriv.2=if(!is.null(phi.deriv.2)){as.matrix(phi.deriv.2)}else{NULL},
                phi.weights=phi.weights,
                phi.deriv.1.weights=phi.deriv.1.weights,
                phi.deriv.2.weights=phi.deriv.2.weights,
                phi.eval=phi.eval,
                phi.eval.mat=phi.eval.mat,
                phi.deriv.eval.1=if(!is.null(phi.deriv.eval.1)){as.matrix(phi.deriv.eval.1)}else{NULL},
                phi.deriv.eval.2=if(!is.null(phi.deriv.eval.2)){as.matrix(phi.deriv.eval.2)}else{NULL},
                phi.eval.weights=phi.eval.weights,
                phi.deriv.eval.1.weights=phi.deriv.eval.1.weights,
                phi.deriv.eval.2.weights=phi.deriv.eval.2.weights,
                norm.index=j,
                norm.stop=norm.stop,
                norm.value=norm.value,
                convergence=convergence,
                starting.values.phi=starting.values.phi,
                return.weights.phi=return.weights.phi,
                bw.E.y.w=bw.E.y.w,
                bw.E.y.z=bw.E.y.z,
                bw.resid.w=as.matrix(bw.resid.w),
                bw.resid.fitted.w.z=as.matrix(bw.resid.fitted.w.z),
                call=cl,
                y=y,
                z=z,
                w=w,
                x=x,
                zeval=zeval,
                xeval=xeval,
                p=p,
                nmulti=nmulti,
                method=method,
                ptm=proc.time() - ptm.start)
    class(ret) <- "npregiv"
    ret

  }
  })
  return(result)
}

print.npregiv <- function(x, ...) {
  summary.npregiv(x, ...)
  invisible(x)
}

summary.npregiv <- function(object, ...) {
  format_bw <- function(bw, label, names = NULL) {
    if(is.null(bw)) return()
    if(is.matrix(bw)) bw <- bw[nrow(bw), , drop = TRUE]
    bw <- as.numeric(bw)
    if(!is.null(names) && length(names) == length(bw)) {
      vals <- paste(paste(names, formatC(bw, digits=8, format="g"), sep=": "), collapse=", ")
    } else {
      vals <- paste(formatC(bw, digits=8, format="g"), collapse=", ")
    }
    cat(paste("\n", label, " ", vals, sep=""))
  }

  cat("Call:\n")
  print(object$call)

  if(is.null(object$alpha))
    cat("\nNonparametric Instrumental Kernel Regression\n",sep="")
  else
    cat("\nNonparametric Instrumental Kernel Regression (Tikhonov)\n",sep="")

  cat(paste("\nNumber of continuous endogenous predictors: ",format(NCOL(object$z)),sep=""),sep="")
  cat(paste("\nNumber of continuous instruments: ",format(NCOL(object$w)),sep=""),sep="")
  if(!is.null(object$x)) cat(paste("\nNumber of continuous exogenous predictors: ",format(NCOL(object$x)),sep=""),sep="")

  cat(paste("\nLocal polynomial order (p): ", format(object$p), sep=""))
  cat(paste("\nTraining observations: ", format(NROW(object$y)), sep=""))

  if(!is.null(object$alpha)) {
    cat(paste("\n\nRegularization method: Tikhonov",sep=""))
    cat(paste("\nTikhonov parameter (alpha): ", format(object$alpha,digits=8), sep=""))
    if(!is.null(object$alpha.iter)) cat(paste("\nIterated Tikhonov parameter (alpha.iter): ", format(object$alpha.iter,digits=8), sep=""))
  } else {
    cat(paste("\n\nRegularization method: Landweber-Fridman",sep=""))
    cat(paste("\nNumber of iterations: ", format(object$norm.index), sep=""))
    cat(paste("\nStopping rule value: ", format(object$norm.stop[length(object$norm.stop)],digits=8), sep=""))
  }

  w.names <- if(!is.null(object$w)) colnames(object$w) else NULL
  z.names <- if(!is.null(object$z)) colnames(object$z) else NULL

  if(is.null(object$alpha)) {
    format_bw(object$bw.E.y.w, "Bandwidth for E(y|w):", w.names)
    format_bw(object$bw.E.y.z, "Bandwidth for E(y|z):", z.names)
    format_bw(object$bw.resid.w, "Bandwidth for E(y-phi(z)|w):", w.names)
    format_bw(object$bw.resid.fitted.w.z, "Bandwidth for E(E(y-phi(z)|w)|z):", z.names)
  } else {
    format_bw(object$bw.E.y.w, "Bandwidth for E(y|w):", w.names)
    format_bw(object$bw.E.E.y.w.z, "Bandwidth for E(E(y|w)|z):", z.names)
    format_bw(object$bw.E.phi.w, "Bandwidth for E(phi(z)|w):", w.names)
    format_bw(object$bw.E.E.phi.w.z, "Bandwidth for E(E(phi(z)|w)|z):", z.names)
  }

  cat(paste("\nNumber of multistarts: ", format(object$nmulti), sep=""))
  cat(paste("\nEstimation time: ", formatC(object$ptm[1],digits=1,format="f"), " seconds",sep=""))
  cat("\n\n")
}

.np_plot_validate_npregiv_call <- function(call,
                                           method_args,
                                           context) {
  call.args <- as.list(call)[-1L]
  arg.names <- names(call.args)
  if (is.null(arg.names))
    arg.names <- rep.int("", length(call.args))
  if (sum(!nzchar(arg.names)) > 1L)
    stop(sprintf("unnamed plot arguments are not supported for %s", context),
         call. = FALSE)
  allowed <- unique(c("x", method_args, .np_plot_graphics_arg_names()))
  bad <- setdiff(arg.names[nzchar(arg.names)], allowed)
  .np_plot_stop_unused_args(bad, allowed)
  invisible(NULL)
}

plot.npregiv <- function(x,
                         plot.data = FALSE,
                         deriv = FALSE,
                         ...) {

  object <- x
  .np_plot_validate_npregiv_call(
    sys.call(),
    method_args = c("plot.data", "deriv"),
    context = "plot.npregiv"
  )
  dots <- list(...)
  take_arg <- function(name, default = NULL) {
    if (!is.null(dots[[name]])) {
      val <- dots[[name]]
      dots[[name]] <<- NULL
      return(val)
    }
    default
  }

  ## We only support univariate endogenous predictor z
  if(NCOL(object$z) > 1) stop(" only univariate z is supported")

  z <- object$z[,1]
  y <- object$y
  phi <- object$phi

  zname <- names(object$z)[1]
  yname <- "y" ## Default

  if(deriv) {
    phi.prime <- object$phi.deriv.1[,1]

    plot.type <- take_arg("type", "l")
    plot.xlab <- take_arg("xlab", zname)
    plot.ylab <- take_arg("ylab", paste("d", yname, "/d", zname, sep=""))
    do.call(plot, c(list(x = z[order(z)],
                         y = phi.prime[order(z)],
                         type = plot.type,
                         xlab = plot.xlab,
                         ylab = plot.ylab),
                    dots))

  } else {

    if(plot.data) {
      plot.type <- take_arg("type", "p")
      plot.xlab <- take_arg("xlab", zname)
      plot.ylab <- take_arg("ylab", yname)
      user.col <- take_arg("col", NULL)
      line.lwd <- take_arg("lwd", 2)
      plot.args <- c(list(x = z,
                          y = y,
                          xlab = plot.xlab,
                          ylab = plot.ylab,
                          type = plot.type,
                          col = if (is.null(user.col)) "lightgrey" else user.col),
                     dots)
      do.call(plot, plot.args)
      line.args <- list(x = z[order(z)],
                        y = phi[order(z)],
                        lwd = line.lwd)
      if (!is.null(user.col)) {
        line.args$col <- user.col
      }
      do.call(lines, c(line.args, dots))
    } else {
      plot.type <- take_arg("type", "l")
      plot.xlab <- take_arg("xlab", zname)
      plot.ylab <- take_arg("ylab", yname)
      user.col <- take_arg("col", NULL)
      line.lwd <- take_arg("lwd", 2)
      plot.args <- list(x = z[order(z)],
                        y = phi[order(z)],
                        type = plot.type,
                        xlab = plot.xlab,
                        ylab = plot.ylab,
                        lwd = line.lwd)
      if (!is.null(user.col)) {
        plot.args$col <- user.col
      }
      do.call(plot, c(plot.args, dots))
    }
  }

}

npconmode <-
  function(bws, ...){
    args <- list(...)

    if (!missing(bws)){
      if (is.recursive(bws)){
        if (!is.null(bws$formula) && is.null(args$txdat))
          UseMethod("npconmode",bws$formula)
        else if (!is.null(bws$call) && is.null(args$txdat))
          UseMethod("npconmode",bws$call)
        else if (!is.call(bws))
          UseMethod("npconmode",bws)
        else
          UseMethod("npconmode",NULL)
      } else {
        UseMethod("npconmode", NULL)
      }
    } else {
      UseMethod("npconmode", NULL)
    }
  }

.npConmodeValidateNewdataTerms <- function(newdata, xnames) {
  nd <- toFrame(newdata)
  missing.names <- setdiff(xnames, names(nd))
  if (length(missing.names))
    stop(sprintf(
      "newdata must contain columns: %s",
      paste(shQuote(xnames), collapse = ", ")
    ), call. = FALSE)
  invisible(TRUE)
}

.npConmodeValidateCategoricalResponse <- function(tydat) {
  tydat <- toFrame(tydat)
  if (NCOL(tydat) != 1L)
    stop("'tydat' must consist of one (1) factor or ordered factor response",
         call. = FALSE)
  y <- tydat[[1L]]
  if (!is.factor(y))
    stop("npconmode requires a categorical response: supply 'tydat' as a factor or ordered factor",
         call. = FALSE)
  invisible(TRUE)
}

npconmode.formula <-
  function(bws, data = NULL, newdata = NULL, ...){

    tt <- terms(bws)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)
    tmf <- bws$call[c(1,m)]
    tmf[[1]] <- as.name("model.frame")
    tmf[["formula"]] <- tt
    mf.args <- as.list(tmf)[-1L]
    umf <- tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

    tydat <- tmf[, bws$variableNames[["response"]], drop = FALSE]
    txdat <- tmf[, bws$variableNames[["terms"]], drop = FALSE]
    .npConmodeValidateCategoricalResponse(tydat)

    has.eval <- !is.null(newdata)
    if (has.eval) {
      .npConmodeValidateNewdataTerms(newdata, bws$variableNames[["terms"]])
      has.ey <- bws$variableNames[["response"]] %in% names(newdata)

      if (has.ey){
        umf.args <- list(formula = tt, data = newdata)
        umf <- do.call(stats::model.frame, umf.args, envir = parent.frame())
        emf <- umf
        eydat <- emf[, bws$variableNames[["response"]], drop = FALSE]
      } else {
        umf.args <- list(formula = formula(bws)[-2], data = newdata)
        umf <- do.call(stats::model.frame, umf.args, envir = parent.frame())
        emf <- umf
      }

      exdat <- emf[, bws$variableNames[["terms"]], drop = FALSE]
    }
    
    cm.args <- list(txdat = txdat, tydat = tydat)
    if (has.eval) {
      cm.args$exdat <- exdat
      if (has.ey)
        cm.args$eydat <- eydat
    }
    cm.args$bws <- bws
    ev <- do.call(npconmode, c(cm.args, list(...)))

    ev$omit <- attr(umf,"na.action")
    ev$rows.omit <- as.vector(ev$omit)
    ev$nobs.omit <- length(ev$rows.omit)

    ev$conmode <- napredict(ev$omit, ev$conmode)
    ev$condens <- napredict(ev$omit, ev$condens)
    ev$conderr <- napredict(ev$omit, ev$conderr)
    if (!is.null(ev$probabilities))
      ev$probabilities <- napredict(ev$omit, ev$probabilities)
    if (!is.null(ev$probability.errors))
      ev$probability.errors <- napredict(ev$omit, ev$probability.errors)
    if (!is.null(ev$probability.repaired.rows))
      ev$probability.repaired.rows <- napredict(ev$omit, ev$probability.repaired.rows)
    if (!is.null(ev$probability.gradients))
      ev$probability.gradients <- .npConmodeNapredictArray(ev$omit, ev$probability.gradients)

    return(ev)
  }

npconmode.call <-
  function(bws, ...) {
    npconmode(txdat = .np_eval_bws_call_arg(bws, "xdat"),
              tydat = .np_eval_bws_call_arg(bws, "ydat"),
              bws = bws, ...)
  }

.npConmodeEffectiveProper <- function(bws, proper) {
  if (!is.null(proper)) {
    if (!is.logical(proper) || length(proper) != 1L || is.na(proper))
      stop("'proper' must be TRUE, FALSE, or NULL")
    return(proper)
  }

  regtype <- bws$regtype
  if (!is.null(bws$regtype.engine))
    regtype <- bws$regtype.engine
  !identical(as.character(regtype)[1L], "lc")
}

.npConmodeProjectSimplexVector <- function(x) {
  u <- sort(x, decreasing = TRUE)
  cssv <- cumsum(u) - 1
  ind <- seq_along(u)
  keep <- which(u - cssv / ind > 0)
  if (!length(keep))
    return(rep(1 / length(x), length(x)))
  theta <- cssv[max(keep)] / max(keep)
  pmax(x - theta, 0)
}

.npConmodeNormalizeProperControl <- function(proper.control) {
  if (is.null(proper.control))
    return(list())
  if (!is.list(proper.control) || is.data.frame(proper.control))
    stop("'proper.control' must be a list")

  nms <- names(proper.control)
  if (is.null(nms))
    nms <- rep("", length(proper.control))
  bad <- nms == "" | !(nms %in% "tol")
  if (any(bad))
    stop(sprintf("unused argument%s in proper.control: %s",
                 if (sum(bad) == 1L) "" else "s",
                 paste(sQuote(nms[bad]), collapse = ", ")))

  proper.control
}

.npConmodeProperProbabilities <- function(pmat,
                                          levels,
                                          proper = TRUE,
                                          proper.control = list()) {
  if (!is.matrix(pmat) || !is.numeric(pmat))
    stop("internal error: conditional mode probabilities must be a numeric matrix")
  if (ncol(pmat) != length(levels))
    stop("internal error: probability matrix does not match the response support")

  proper.control <- .npConmodeNormalizeProperControl(proper.control)
  tol <- proper.control$tol
  if (is.null(tol))
    tol <- sqrt(.Machine$double.eps)
  if (!is.numeric(tol) || length(tol) != 1L || is.na(tol) || tol < 0)
    stop("'proper.control$tol' must be a non-negative scalar")

  raw <- pmat
  finite.row <- rowSums(is.finite(raw)) == ncol(raw)
  raw.rowsum <- rep(NA_real_, nrow(raw))
  raw.rowsum[finite.row] <- rowSums(raw[finite.row,, drop = FALSE])
  raw.min <- rep(NA_real_, nrow(raw))
  raw.min[finite.row] <- apply(raw[finite.row,, drop = FALSE], 1L, min)
  raw.neg.violation <- ifelse(is.na(raw.min), NA_real_, pmax(-raw.min, 0))

  repaired <- raw
  projection.distance <- rep(NA_real_, nrow(raw))
  invalid.rows <- !finite.row

  if (isTRUE(proper)) {
    for (i in which(finite.row)) {
      x <- raw[i,]
      if (any(x < -tol) || abs(sum(x) - 1) > tol) {
        repaired[i,] <- .npConmodeProjectSimplexVector(x)
      } else {
        repaired[i,] <- pmax(x, 0)
        repaired[i,] <- repaired[i,] / sum(repaired[i,])
      }
      projection.distance[i] <- max(abs(repaired[i,] - raw[i,]))
    }
  } else {
    projection.distance[finite.row] <- 0
  }

  repaired.rows <- isTRUE(proper) & is.finite(projection.distance) & projection.distance > tol
  repaired.rows[is.na(repaired.rows)] <- FALSE

  binary.complement.max <- NA_real_
  if (ncol(repaired) == 2L) {
    ok <- rowSums(is.finite(repaired)) == 2L
    if (any(ok))
      binary.complement.max <- max(abs(repaired[ok, 2L] - (1 - repaired[ok, 1L])))
  }

  colnames(repaired) <- as.character(levels)
  info <- list(
    reason = if (!isTRUE(proper)) "not_requested" else if (any(repaired.rows)) "projected" else "already_proper",
    tol = tol,
    nlevels = ncol(raw),
    invalid.rows = sum(invalid.rows),
    repaired.rows = sum(repaired.rows),
    max.negative.violation.raw = suppressWarnings(max(raw.neg.violation, na.rm = TRUE)),
    max.row.sum.deviation.raw = suppressWarnings(max(abs(raw.rowsum - 1), na.rm = TRUE)),
    max.projection.distance = suppressWarnings(max(projection.distance, na.rm = TRUE)),
    binary.complement.max.deviation = binary.complement.max
  )
  for (nm in c("max.negative.violation.raw", "max.row.sum.deviation.raw", "max.projection.distance")) {
    if (!is.finite(info[[nm]]))
      info[[nm]] <- NA_real_
  }

  list(
    probabilities = repaired,
    raw.probabilities = raw,
    repaired.rows = repaired.rows,
    proper.requested = isTRUE(proper),
    proper.applied = any(repaired.rows),
    proper.info = info
  )
}

.npConmodeSelect <- function(pmat, perr) {
  enrow <- nrow(pmat)
  nlev <- ncol(pmat)
  mdens <- double(enrow)
  mderr <- rep(NA_real_, enrow)
  indices <- integer(enrow)
  for (i in seq_len(nlev)) {
    tf <- pmat[, i] >= mdens
    tf[is.na(tf)] <- FALSE
    indices[tf] <- i
    mdens[tf] <- pmat[tf, i]
    mderr[tf] <- perr[tf, i]
  }
  list(indices = indices, condens = mdens, conderr = mderr)
}

.npConmodeNapredictArray <- function(omit, x) {
  if (is.null(x) || !length(omit))
    return(x)
  dx <- dim(x)
  if (is.null(dx) || length(dx) < 3L)
    return(napredict(omit, x))

  keep <- seq_len(dx[1L] + length(omit))[-as.integer(omit)]
  dn <- dimnames(x)
  if (!is.null(dn))
    dn[[1L]] <- NULL
  out <- array(NA_real_,
               dim = c(dx[1L] + length(omit), dx[-1L]),
               dimnames = dn)
  out[keep,,] <- x
  out
}


npconmode.conbandwidth <-
  function (bws,
            txdat = stop("invoked without training data 'txdat'"),
            tydat = stop("invoked without training data 'tydat'"),
            exdat, eydat,
            proper = NULL,
            proper.control = list(),
            probabilities = FALSE,
            gradients = FALSE,
            level = NULL,
            ...){

    probabilities <- npValidateScalarLogical(probabilities, "probabilities")
    gradients <- npValidateScalarLogical(gradients, "gradients")
    txdat = toFrame(txdat)
    tydat = toFrame(tydat)

    no.ex = missing(exdat)
    no.ey = missing(eydat)

    if (!no.ex)
      exdat = toFrame(exdat)

    if (!no.ey)
      eydat = toFrame(eydat)

    ## catch and destroy NA's
    keep.rows <- rep_len(TRUE, nrow(txdat))
    rows.omit <- attr(na.omit(data.frame(txdat, tydat)), "na.action")
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Training data has no rows without NAs")

    txdat <- txdat[keep.rows,,drop = FALSE]
    tydat <- tydat[keep.rows,,drop = FALSE]

    if (!no.ex){
      keep.eval <- rep_len(TRUE, nrow(exdat))
      eval.df <- data.frame(exdat)
      if (!no.ey)
        eval.df <- data.frame(eval.df, eydat)
      rows.omit <- attr(na.omit(eval.df), "na.action")
      if (length(rows.omit) > 0L)
        keep.eval[as.integer(rows.omit)] <- FALSE

      exdat <- exdat[keep.eval,,drop = FALSE]

      if (!no.ey)
        eydat <- eydat[keep.eval,,drop = FALSE]

      if (!any(keep.eval))
        stop("Evaluation data has no rows without NAs")
    }


    tnrow = dim(txdat)[1]
    enrow = if (no.ex) tnrow else dim(exdat)[1]

    if (!no.ey && no.ex)
      stop("npconmode: invalid invocation: 'eydat' provided but not 'exdat'")

    if (bws$yndim != 1 || bws$yncon > 0)
      stop("'tydat' must consist of one (1) discrete variable")

    if(no.ey)
      efac <- factor(bws$ydati$all.lev[[1]],levels = bws$ydati$all.lev[[1]], ordered = is.ordered(tydat[,1]))
    else
      efac <- factor(union(bws$ydati$all.lev[[1]], levels(eydat[,1])),
                     levels = union(bws$ydati$all.lev[[1]], levels(eydat[,1])), ordered = is.ordered(tydat[,1]))

    nlev <- nlevels(efac)
    level.values <- levels(efac)
    if (is.null(level)) {
      gradient.level <- level.values[1L]
    } else {
      if (length(level) != 1L || is.na(level) ||
          !(as.character(level) %in% level.values))
        stop("'level' must identify one response level in the fitted conmode object")
      gradient.level <- as.character(level)
    }
    gradient.level.index <- match(gradient.level, level.values)
    if (isTRUE(gradients)) {
      if (bws$xndim < 1L)
        stop("npconmode class-probability gradients/effects require at least one conditioning variable")
    }
    pmat <- matrix(NA_real_, enrow, nlev, dimnames = list(NULL, level.values))
    perr <- matrix(NA_real_, enrow, nlev, dimnames = list(NULL, level.values))
    pgrad <- if (isTRUE(gradients)) {
      matrix(NA_real_,
             nrow = enrow,
             ncol = bws$xndim,
             dimnames = list(NULL, bws$xnames))
    } else {
      NULL
    }

    for (i in seq_len(nlevels(efac))) {
        dens.obj <- npcdens(
          txdat = txdat,
          tydat = tydat,
          exdat = if (no.ex) txdat else exdat,
          eydat = rep(efac[i], enrow),
          bws = bws,
          gradients = isTRUE(gradients) && identical(i, gradient.level.index)
        )
        pmat[, i] <- dens.obj$condens
        perr[, i] <- dens.obj$conderr
        if (isTRUE(gradients) && identical(i, gradient.level.index)) {
          if (is.null(dens.obj$congrad))
            stop("internal error: conditional-density gradient was not returned")
          pgrad[,] <- dens.obj$congrad
        }
    }

    proper.effective <- .npConmodeEffectiveProper(bws, proper)
    proper.out <- .npConmodeProperProbabilities(
      pmat,
      levels = levels(efac),
      proper = proper.effective,
      proper.control = proper.control
    )
    select <- .npConmodeSelect(proper.out$probabilities, perr)
    indices <- select$indices
    mdens <- select$condens
    mderr <- select$conderr
    mderr[proper.out$repaired.rows] <- NA_real_
    cm.args <- list(
      bws = bws,
      xeval = if (no.ex) txdat else exdat,
      conmode = efac[indices],
      condens = mdens,
      conderr = mderr,
      ntrain = nrow(txdat),
      trainiseval = no.ex,
      proper.requested = proper.out$proper.requested,
      proper.applied = proper.out$proper.applied,
      proper.info = proper.out$proper.info,
      gradients = gradients
    )
    if (isTRUE(probabilities) || isTRUE(gradients)) {
      cm.args$probabilities <- proper.out$probabilities
      cm.args$probability.levels <- efac
      cm.args$probability.errors <- perr
      cm.args$probability.errors[proper.out$repaired.rows, ] <- NA_real_
      cm.args$probability.repaired.rows <- proper.out$repaired.rows
      if (!no.ex) {
        cm.args$xtrain <- txdat
        cm.args$ytrain <- tydat
      }
    }
    if (isTRUE(gradients)) {
      cm.args$probability.gradients <- pgrad
      cm.args$probability.gradient.level <- gradient.level
      cm.args$probability.gradient.names <- bws$xnames
      cm.args$probability.gradient.info <- list(
        target = "class probability",
        response = if (nlev == 2L) "binary" else "multinomial",
        level = gradient.level,
        default.level = level.values[1L],
        semantics = "smooth class-probability gradients/effects for one selected response level"
      )
    }
    if (!(no.ey && !no.ex))
      cm.args$yeval <- if (no.ey) tydat else eydat
    con.mode <- do.call(conmode, cm.args)
    
    if (!(no.ey && !no.ex)){
      confusion.matrix <- 
        table(factor(if (no.ex) tydat[,1] else eydat[,1], exclude = NULL),
              factor(con.mode$conmode,exclude = NULL), dnn=c("Actual", "Predicted"))

      cj <- match(levels(factor(if (no.ex) tydat[,1] else eydat[,1], exclude = NULL)),
                  levels(factor(con.mode$conmode,exclude = NULL)), nomatch = 0)
      rj <- cj > 0

      t.diag <- cj
      t.diag[rj] <-  diag(confusion.matrix[rj,cj,drop=FALSE])
      
      CCR.overall <- sum(t.diag)/enrow
      
      CCR.byoutcome <- t.diag/rowSums(confusion.matrix)
      names(CCR.byoutcome) <- levels(factor(if (no.ex) tydat[,1] else eydat[,1], exclude = NULL))

      con.mode$confusion.matrix <- confusion.matrix
      con.mode$CCR.overall <- CCR.overall
      con.mode$CCR.byoutcome <- CCR.byoutcome

      confusion.matrix <- confusion.matrix/enrow
      t.diag <- t.diag/enrow

      fit.mcfadden <- sum(t.diag) - (sum(confusion.matrix^2)-sum(t.diag^2))
      con.mode$fit.mcfadden <- fit.mcfadden
    }
    con.mode
  }

npconmode.default <- function(bws, txdat, tydat,
                              nomad = FALSE,
                              proper = NULL,
                              proper.control = list(),
                              probabilities = FALSE,
                              gradients = FALSE,
                              level = NULL,
                              ...){
  nomad <- npValidateScalarLogical(nomad, "nomad")
  probabilities <- npValidateScalarLogical(probabilities, "probabilities")
  gradients <- npValidateScalarLogical(gradients, "gradients")
  sc <- sys.call()
  sc.names <- names(sc)

  ## here we check to see if the function was called with tdat =
  ## if it was, we need to catch that and map it to dat =
  ## otherwise the call is passed unadulterated to npudensbw

  bws.named <- any(sc.names == "bws")
  txdat.named <- any(sc.names == "txdat")
  tydat.named <- any(sc.names == "tydat")

  no.bws <- missing(bws)
  no.txdat <- missing(txdat)
  no.tydat <- missing(tydat)
  has.explicit.bws <- (!no.bws) && isa(bws, "conbandwidth")

  ## if bws was passed in explicitly, do not compute bandwidths
    
  if(txdat.named)
    txdat <- toFrame(txdat)

  if(tydat.named)
    tydat <- toFrame(tydat)

  if(!no.tydat) {
    if(!tydat.named)
      tydat <- toFrame(tydat)
    .npConmodeValidateCategoricalResponse(tydat)
  }

  sc.bw <- sc
  
  sc.bw[[1]] <- quote(npcdensbw)
  sc.bw$proper <- NULL
  sc.bw$proper.control <- NULL
  sc.bw$probabilities <- NULL
  sc.bw$gradients <- NULL
  sc.bw$level <- NULL
  sc.bw$newdata <- NULL
  sc.bw$exdat <- NULL
  sc.bw$eydat <- NULL

  bws.formula <- (!no.bws) && inherits(bws, "formula")
  if (bws.formula) {
    ib <- match("bws", names(sc.bw), nomatch = 0L)
    if (ib > 0L) names(sc.bw)[ib] <- "formula"
  }

  if(bws.named && !bws.formula){
    sc.bw$bandwidth.compute <- FALSE
  }

  ostxy <- c('txdat','tydat')
  nstxy <- c('xdat','ydat')
  
  m.txy <- match(ostxy, names(sc.bw), nomatch = 0)

  if(any(m.txy > 0)) {
    names(sc.bw)[m.txy] <- nstxy[m.txy > 0]
  }

  if (bws.formula && no.tydat) {
    mf.call <- sc.bw
    mf.call[[1]] <- quote(stats::model.frame)
    keep <- match(c("formula", "data", "subset", "na.action"),
                  names(mf.call), nomatch = 0L)
    mf.call <- mf.call[c(1L, keep)]
    if (!("formula" %in% names(mf.call)))
      mf.call$formula <- bws
    mf <- eval(mf.call, parent.frame())
    y <- stats::model.response(mf)
    if (is.null(y))
      stop("npconmode requires a categorical response in the formula",
           call. = FALSE)
    .npConmodeValidateCategoricalResponse(data.frame(y))
  }

  use.outer.bandwidth.progress <- !.np_bw_call_uses_nomad_degree_search(
    sc.bw,
    caller_env = parent.frame()
  )

  tbw <- if (!has.explicit.bws) {
    if (use.outer.bandwidth.progress) {
      .np_progress_select_bandwidth_enhanced(
        "Selecting conditional density bandwidth",
        .np_eval_bw_call(sc.bw, caller_env = parent.frame())
      )
    } else {
      .np_eval_bw_call(sc.bw, caller_env = parent.frame())
    }
  } else {
    .np_eval_bw_call(sc.bw, caller_env = parent.frame())
  }

  call.args <- list(bws = tbw)
  if (no.bws) {
    call.args$txdat <- txdat
    call.args$tydat <- tydat
  } else {
    if (txdat.named) call.args$txdat <- txdat
    if (tydat.named) call.args$tydat <- tydat
    if ((!bws.named) && (!txdat.named) && (!no.tydat) && (!tydat.named)) {
      call.args <- c(call.args, list(tydat))
    }
  }
  call.args$proper <- proper
  call.args$proper.control <- proper.control
  call.args$probabilities <- probabilities
  call.args$gradients <- gradients
  call.args$level <- level
  do.call(npconmode, c(call.args, list(...)))
}

unitest <- function(Srho,
                       Srho.bootstrap,
                       P,
                       bootstrap,
                       boot.num,
                       bw.x,
                       bw.y) {
  
  tuni = list(Srho = Srho,
    Srho.bootstrap = Srho.bootstrap,
    P = P,
    bootstrap = bootstrap,
    boot.num = boot.num,
    bw.x = bw.x,
    bw.y = bw.y)

  if(bootstrap) {
  
    reject <- ' '
    
    if (P < 0.1)
      reject <- '.'
    
    if (P < 0.05)
      reject <- '*'
    
    if (P < 0.01)
      reject <- '**'
    
    if (P < 0.001)
      reject <- '***'
    
    tuni$reject <- reject
    tuni$rejectNum <- switch(reject,
                             ' ' = 100,
                             '.' = 10,
                             '*' = 5,
                             '**' = 1,
                             '***' = 0.1)

  }
  
  class(tuni) <- "unitest"

  tuni

}

print.unitest <- function(x, ...){
  if(x$bootstrap) {
    reject.msg <- if (x$reject == ' ') {
      "\nFail to reject the null of equality at the 10% level"
    } else {
      paste("\nNull of equality is rejected at the ", x$rejectNum, "% level", sep = "")
    }
    cat("\nConsistent Univariate Entropy Density Equality Test",
        paste("\n", format(x$boot.num), " Bootstrap Replications",sep=""),
        "\n\nTest Statistic ", sQuote("Srho"), ": ",
        format(x$Srho), "\tP Value: ", format.pval(x$P)," ", x$reject,
        "\n---\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1",
        reject.msg,
        "\n\n", sep="")
  } else {
    cat("\nConsistent Univariate Density Difference Metric Entropy",
        "\n\nMetric Entropy ", sQuote("Srho"),": ",
        format(x$Srho),"\n\n", sep="")
  }
}

summary.unitest <- function(object, ...){
  print(object)
}

.np_condens_engine_spec <- function(bws) {
  reg.engine <- if (is.null(bws$regtype.engine)) {
    if (is.null(bws$regtype)) "lc" else as.character(bws$regtype)
  } else {
    as.character(bws$regtype.engine)
  }

  degree.engine <- if (is.null(bws$degree.engine)) {
    if (isTRUE(bws$xncon > 0L)) {
      if (identical(reg.engine, "lc")) {
        rep.int(0L, bws$xncon)
      } else {
        as.integer(npValidateGlpDegree(
          regtype = "lp",
          degree = bws$degree,
          ncon = bws$xncon
        ))
      }
    } else {
      integer(0L)
    }
  } else {
    as.integer(bws$degree.engine)
  }

  list(
    reg.engine = reg.engine,
    degree.engine = degree.engine
  )
}

.np_condens_is_already_proper_by_design <- function(bws) {
  spec <- .np_condens_engine_spec(bws)

  identical(spec$reg.engine, "lc") ||
    (identical(spec$reg.engine, "lp") && all(spec$degree.engine == 0L))
}

.np_condens_slice_dispatch_enabled <- function() {
  !identical(getOption("np.condens.proper.slice.enable"), FALSE)
}

.np_condens_validate_nonnegative_finite_numeric <- function(value, argname) {
  value <- as.double(value)[1L]
  if (!is.finite(value) || value < 0)
    stop(sprintf("'%s' must be a non-negative finite numeric scalar", argname))
  value
}

npdensity <- 
    function(bws, eval, dens,
             derr = NA, ll = NA,
             ntrain, trainiseval = FALSE,
             rows.omit = NA,
             timing = NA, total.time = NA,
             optim.time = NA, fit.time = NA){

        if (missing(bws) || missing(eval) || missing(dens) || missing(ntrain))
            stop("improper invocation of density constructor")

        if (length(rows.omit) == 0)
            rows.omit <- NA


        d <- list(
            bw = bws$bw,
            bws = bws,
            xnames = bws$xnames,
            nobs = nrow(eval),
            ndim = bws$ndim,
            nord = bws$nord,
            nuno = bws$nuno,
            ncon = bws$ncon,
            pscaling = bws$pscaling,
            ptype = bws$ptype,
            pckertype = bws$pckertype,
            pukertype = bws$pukertype,
            pokertype = bws$pokertype,
            eval = eval,
            dens = dens,
            derr = derr,
            log_likelihood = ll,
            ntrain = ntrain,
            trainiseval = trainiseval,
            rows.omit = rows.omit,
            nobs.omit = if (identical(rows.omit, NA)) 0 else length(rows.omit),
            timing = timing, total.time = total.time,
            optim.time = optim.time, fit.time = fit.time)


        class(d) <- "npdensity"

        return(d)
    }

print.npdensity <- function(x, digits=NULL, ...){
  cat("\nDensity Data: ", x$ntrain, " training points,",
      if (x$trainiseval) "" else paste(" and ", x$nobs,
                                      " evaluation points,", sep=""),
      " in ",x$ndim," variable(s)\n",sep="")

  print(matrix(x$bw,ncol=x$ndim,dimnames=list(paste(x$pscaling,":",sep=""),x$xnames)))

  cat(genDenEstStr(x))
  
  cat(genBwKerStrs(x$bws))
  
  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

fitted.npdensity <- function(object, ...){
 object$dens 
}
se.npdensity <- function(x){ x$derr }

predict.npdensity <- function(object, se.fit = FALSE, ...) {
  se.fit <- npValidateScalarLogical(se.fit, "se.fit")
  dots <- list(...)
  has.formula.route <- !is.null(object$bws$formula)

  if (!is.null(dots$edat) && !is.null(dots$newdata)) {
    dots$newdata <- NULL
  } else if (!has.formula.route && is.null(dots$edat) && !is.null(dots$newdata)) {
    dots$edat <- dots$newdata
    dots$newdata <- NULL
  }

  tr <- do.call(npudens, c(list(bws = object$bws), dots))
  if(se.fit)
    return(list(fit = fitted(tr), se.fit = se(tr), 
                df = tr$nobs, log.likelihood = tr$log_likelihood))
  else
    return(fitted(tr))
}


summary.npdensity <- function(object, ...) {
  cat("\nDensity Data: ", object$ntrain, " training points,",
      if (object$trainiseval) "" else paste(" and ", object$nobs,
                                      " evaluation points,", sep=""),
      " in ",object$ndim," variable(s)\n",sep="")

  cat(genOmitStr(object))

  print(matrix(object$bw,ncol=object$ndim,dimnames=list(paste(object$pscaling,":",sep=""),object$xnames)))

  cat(genDenEstStr(object))

  cat(genBwKerStrs(object$bws))
  cat(genTimingStr(object))
  cat('\n\n')  

}

.np_plot_bandwidth_engine <-
  function(bws,
           xdat,
           data = NULL,
           xq = 0.5,
           xtrim = 0.0,
           neval = 50,
           common.scale = TRUE,
           perspective = TRUE,
           renderer = c("base", "rgl"),
           main = NULL,
           type = NULL,
           border = NULL,
           cex.axis = NULL,
           cex.lab = NULL,
           cex.main = NULL,
           cex.sub = NULL,
           col = NULL,
           ylab = NULL,
           xlab = NULL,
           zlab = NULL,
           sub = NULL,
           ylim = NULL,
           xlim = NULL,
           zlim = NULL,
           lty = NULL,
           lwd = NULL,
           theta = 0.0,
           phi = 20.0,
           view = c("rotate","fixed"),
           plot.behavior = c("plot","plot-data","data"),
           plot.errors.method = c("none","bootstrap","asymptotic"),
           plot.errors.boot.method = c("inid", "fixed", "geom"),
           plot.errors.boot.nonfixed = c("exact", "frozen"),
           plot.errors.boot.blocklen = NULL,
           plot.errors.boot.num = 1999,
           plot.errors.center = c("estimate","bias-corrected"),
           plot.errors.type = c("pmzsd","pointwise","bonferroni","simultaneous","all"),
           plot.errors.alpha = 0.05,
           plot.errors.style = c("band","bar"),
           plot.errors.bar = c("|","I"),
           plot.errors.bar.num = min(neval,25),
           plot.bxp = FALSE,
           plot.bxp.out = TRUE,
           plot.par.mfrow = TRUE,
           plot.rug = FALSE,
           ...,
           random.seed){

    engine.ctx <- .np_plot_engine_begin(plot.par.mfrow = plot.par.mfrow)
    on.exit(.np_plot_restore_par(engine.ctx$oldpar), add = TRUE)
    plot.par.mfrow <- engine.ctx$plot.par.mfrow
    scalar_default <- .np_plot_scalar_default

    dots <- list(...)
    plot.legend <- if (!is.null(dots$legend)) dots$legend else TRUE
    plot.user.args <- .np_plot_user_args(dots, "plot")
    bxp.user.args <- .np_plot_user_args(dots, "bxp")
    rgl.persp3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.persp3d", "rgl.persp3d.")
    rgl.view3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.view3d", "rgl.view3d.")
    rgl.par3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.par3d", "rgl.par3d.")
    rgl.grid3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.grid3d", "rgl.grid3d.")
    rgl.widget.user.args <- .np_plot_collect_rgl_args(dots, "rgl.widget", "rgl.widget.")
    rgl.legend3d.user.args <- .np_plot_collect_rgl_args(dots, "rgl.legend3d", "rgl.legend3d.")
    rgl.legend3d.user.args <- .np_plot_merge_rgl_legend_control(rgl.legend3d.user.args, plot.legend)
    rgl.surface3d.user.args <- .np_plot_extract_prefixed_args(dots, "rgl.surface3d.")
    bxp.args <- bxp.user.args
    if (!is.null(col)) bxp.args$col <- col
    if (!is.null(lty)) bxp.args$lty <- lty
    if (!is.null(lwd)) bxp.args$lwd <- lwd
    if (!is.null(border)) bxp.args$border <- border

    miss.x <- missing(xdat)

    if(miss.x && !is.null(bws$formula)){
      tt <- terms(bws)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)
    tmf <- bws$call[c(1,m)]
    tmf[[1]] <- as.name("model.frame")
    tmf[["formula"]] <- tt
    mf.args <- as.list(tmf)[-1L]
    umf <- tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

      xdat <- tmf[, attr(attr(tmf, "terms"),"term.labels"), drop = FALSE]
    } else {
      if(miss.x && !is.null(bws$call)){
        xdat <- data.frame(.np_eval_bws_call_arg(bws, "dat"))
      }
      xdat = toFrame(xdat)
      xdat = na.omit(xdat)
    }

    xq = double(bws$ndim)+xq
    xtrim = double(bws$ndim)+xtrim

    if (missing(plot.errors.method) &
        any(!missing(plot.errors.boot.num), !missing(plot.errors.boot.method),
            !missing(plot.errors.boot.nonfixed),
            !missing(plot.errors.boot.blocklen))){
      stop(
        "errors must be set to \"bootstrap\" when bootstrap controls are supplied",
        call. = FALSE
      )
    }
    normalized.opts <- .np_plot_normalize_common_options(
      plot.behavior = plot.behavior,
      plot.errors.method = plot.errors.method,
      plot.errors.boot.method = plot.errors.boot.method,
      plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
      plot.errors.boot.blocklen = plot.errors.boot.blocklen,
      plot.errors.center = plot.errors.center,
      plot.errors.type = plot.errors.type,
      plot.errors.alpha = plot.errors.alpha,
      plot.errors.style = plot.errors.style,
      plot.errors.bar = plot.errors.bar,
      xdat = xdat,
      common.scale = common.scale,
      ylim = ylim
    )

    plot.behavior <- normalized.opts$plot.behavior
    plot.errors.method <- normalized.opts$plot.errors.method
    plot.errors.boot.method <- normalized.opts$plot.errors.boot.method
    plot.errors.boot.nonfixed <- normalized.opts$plot.errors.boot.nonfixed
    plot.errors.boot.blocklen <- normalized.opts$plot.errors.boot.blocklen
    plot.errors.center <- normalized.opts$plot.errors.center
    plot.errors.type <- normalized.opts$plot.errors.type
    plot.errors.alpha <- normalized.opts$plot.errors.alpha
    plot.errors.style <- normalized.opts$plot.errors.style
    plot.errors.bar <- normalized.opts$plot.errors.bar
    common.scale <- normalized.opts$common.scale
    plot.errors <- normalized.opts$plot.errors
    first.render <- .np_plot_first_render_state()
    on.exit(.np_plot_activity_end(first.render$activity), add = TRUE)
    surface.supported <- isTRUE((bws$ncon + bws$nord == 2) &&
                                (bws$nuno == 0) &&
                                !any(xor(bws$xdati$iord, bws$xdati$inumord)))
    renderer <- .np_plot_validate_renderer_request(
      renderer = renderer,
      route = "plot.bandwidth()",
      perspective = perspective,
      supported.route = surface.supported,
      view = as.character(view)[1L],
      plot.errors.method = plot.errors.method,
      plot.behavior = plot.behavior,
      allow.plot.errors = TRUE
    )
    plot.rug <- .np_plot_validate_rug_request(
      plot.rug = plot.rug,
      route = "plot.bandwidth()",
      supported.route = if (identical(renderer, "rgl")) {
        isTRUE(surface.supported && perspective)
      } else {
        TRUE
      },
      renderer = renderer,
      reason = if (identical(renderer, "rgl")) {
        "supported rgl surface routes"
      } else {
        "supported base plot routes"
      }
    )

    if (surface.supported && perspective){
      view = match.arg(view)
      rotate = (view == "rotate")

      
      if (is.ordered(xdat[,1])){
        x1.eval = bws$xdati$all.ulev[[1]]
        x1.neval = length(x1.eval)
      } else {
        x1.neval = neval
        qi = trim.quantiles(xdat[,1], xtrim[1])
        x1.eval = seq(qi[1], qi[2], length.out = x1.neval)
      }

      if (is.ordered(xdat[,2])){
        x2.eval = bws$xdati$all.ulev[[2]]
        x2.neval = length(x2.eval)
      } else {
        x2.neval = neval
        qi = trim.quantiles(xdat[,2], xtrim[2])
        x2.eval = seq(qi[1], qi[2], length.out = x2.neval)
      }

      x.eval <- expand.grid(x1.eval, x2.eval)

      ##x.eval = as.data.frame(matrix(data = NA, ncol = 2, nrow = x1.neval*x2.neval))
      ##x.eval[,1] = rep(x1.eval, times = x2.neval)
      ##x.eval[,2] = rep(x2.eval, each = x1.neval)

      if (is.ordered(xdat[,1]))
        x1.eval <- (bws$xdati$all.dlev[[1]])[as.integer(x1.eval)]

      if (is.ordered(xdat[,2]))
        x2.eval <- (bws$xdati$all.dlev[[2]])[as.integer(x2.eval)]

      tobj <- .np_plot_unconditional_eval(
        xdat = xdat,
        exdat = x.eval,
        bws = bws,
        cdf = FALSE,
        need.asymptotic = identical(plot.errors.method, "asymptotic")
      )

      tdens = matrix(data = tobj$dens,
        nrow = x1.neval, ncol = x2.neval, byrow = FALSE)

      terr.payload <- .np_plot_interval_payload(
        estimate = tobj$dens,
        se = tobj$derr,
        plot.errors.method = plot.errors.method,
        plot.errors.alpha = plot.errors.alpha,
        plot.errors.type = plot.errors.type,
        plot.errors.center = plot.errors.center,
        bootstrap_raw = if (identical(plot.errors.method, "bootstrap")) {
          compute.bootstrap.errors(
            xdat = xdat,
            exdat = x.eval,
            cdf = FALSE,
            slice.index = 0,
            plot.errors.boot.method = plot.errors.boot.method,
            plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
            plot.errors.boot.blocklen = plot.errors.boot.blocklen,
            plot.errors.boot.num = plot.errors.boot.num,
            plot.errors.center = plot.errors.center,
            plot.errors.type = plot.errors.type,
            plot.errors.alpha = plot.errors.alpha,
            bws = bws
          )
        } else {
          NULL
        }
      )

      terr <- terr.payload$err
      terr.all <- terr.payload$all.err
      center.val <- terr.payload$center

      lerr.all <- NULL
      herr.all <- NULL

      if (plot.errors) {
        lerr = matrix(data = center.val - terr[,1],
          nrow = x1.neval, ncol = x2.neval, byrow = FALSE)

        herr = matrix(data = center.val + terr[,2],
          nrow = x1.neval, ncol = x2.neval, byrow = FALSE)

        if (plot.errors.type == "all" && !is.null(terr.all)) {
          lerr.all <- lapply(terr.all, function(te)
            matrix(data = center.val - te[,1], nrow = x1.neval, ncol = x2.neval, byrow = FALSE))
          herr.all <- lapply(terr.all, function(te)
            matrix(data = center.val + te[,2], nrow = x1.neval, ncol = x2.neval, byrow = FALSE))
        }
      }

      if(is.null(zlim)) {
          zlim =
              if (plot.errors){
                  if (plot.errors.type == "all" && !is.null(lerr.all))
                  c(min(c(unlist(lerr.all), lerr)), max(c(unlist(herr.all), herr)))
                  else
                      c(min(lerr), max(herr))
              } else
                  c(min(tobj$dens),max(tobj$dens))
      }

      if (plot.behavior != "plot"){
        d1 <- npdensity(bws = bws, eval = x.eval, dens = tobj$dens, derr = terr[,1:2], ntrain = nrow(xdat))
        d1$bias = NA

        if (plot.errors.center == "bias-corrected")
          d1$bias = terr[,3] - tobj$dens
        
        if (plot.behavior == "data")
          return ( list(d1 = d1) )
      }
      
      # rows = constant x2
      # cols = constant x1

      xlab.val <- scalar_default(xlab, gen.label(names(xdat)[1], "X1"))
      ylab.val <- scalar_default(ylab, gen.label(names(xdat)[2], "X2"))
      zlab.val <- scalar_default(zlab, "Joint Density")

      if (identical(renderer, "rgl")) {
        rgl.view <- .np_plot_rgl_view_angles(theta = theta, phi = phi)
        main.val <- if (!is.null(main)) main else NULL
        .np_plot_first_render_begin(first.render)
        rgl.out <- .np_plot_render_surface_rgl(
          x = x1.eval,
          y = x2.eval,
          z = tdens,
          zlim = zlim,
          col = col,
          border = scalar_default(border, "black"),
          xlab = xlab.val,
          ylab = ylab.val,
          zlab = zlab.val,
          theta = rgl.view$theta,
          phi = rgl.view$phi,
          main = main.val,
          par3d.args = rgl.par3d.user.args,
          view3d.args = rgl.view3d.user.args,
          persp3d.args = rgl.persp3d.user.args,
          grid3d.args = rgl.grid3d.user.args,
          widget.args = rgl.widget.user.args,
          draw.extras = function() {
            if (plot.rug) {
              .np_plot_draw_floor_rug_rgl(
                x1 = xdat[,1],
                x2 = xdat[,2],
                zlim = zlim
              )
            }
            if (plot.errors) {
              .np_plot_error_surfaces_rgl(
                x = x1.eval,
                y = x2.eval,
                plot.errors.type = plot.errors.type,
                lerr = lerr,
                herr = herr,
                lerr.all = lerr.all,
                herr.all = herr.all,
                surface3d.args = rgl.surface3d.user.args,
                legend3d.args = rgl.legend3d.user.args
              )
            }
          }
        )
        .np_plot_first_render_end(first.render)
        return(.np_plot_rgl_finalize(
          rgl.out = rgl.out,
          plot.behavior = plot.behavior,
          plot.data = list(d1 = d1)
        ))
      }

      rotate.defaults <- .np_plot_rotate_defaults()
      dtheta = rotate.defaults$dtheta
      persp.col = grDevices::adjustcolor(
        .np_plot_persp_surface_colors(z = tdens, col = col),
        alpha.f = 0.5
      )
      frame.theta <- (0:((360 %/% dtheta - 1L) * rotate)) * dtheta + theta
      rotation.progress <- .np_plot_rotation_progress_begin(length(frame.theta))
      on.exit(.np_plot_rotation_progress_end(rotation.progress), add = TRUE)
      
      for (frame.idx in seq_along(frame.theta)){
          i <- frame.theta[[frame.idx]]
          .np_plot_first_render_begin(first.render)
          persp.mat <- persp(x1.eval,
                             x2.eval,
                             tdens,
                             zlim = zlim,
                             col = persp.col,
                             border = scalar_default(border, "black"),
                             ticktype = "detailed",
                             cex.axis = scalar_default(cex.axis, par()$cex.axis),
                             cex.lab = scalar_default(cex.lab, par()$cex.lab),
                             cex.main = scalar_default(cex.main, par()$cex.main),
                             cex.sub = scalar_default(cex.sub, par()$cex.sub),
                             lwd = 0.8 * scalar_default(lwd, par()$lwd),
                             xlab = xlab.val,
                             ylab = ylab.val,
                             zlab = zlab.val,
                             theta = i,
                             phi = phi,
                             main = scalar_default(main, ""))
          .np_plot_first_render_end(first.render)
          .np_plot_draw_box_grid_persp(
            xlim = range(x1.eval, finite = TRUE),
            ylim = range(x2.eval, finite = TRUE),
            zlim = zlim,
            persp.mat = persp.mat
          )
          if (plot.rug) {
            .np_plot_draw_floor_rug_persp(
              x1 = xdat[,1],
              x2 = xdat[,2],
              zlim = zlim,
              persp.mat = persp.mat
            )
          }

          if (plot.errors){
            .np_plot_draw_error_wireframes_persp(
              x = x1.eval,
              y = x2.eval,
              persp.mat = persp.mat,
              plot.errors.type = plot.errors.type,
              lerr = lerr,
              herr = herr,
              lerr.all = lerr.all,
              herr.all = herr.all,
              border = scalar_default(border, "grey"),
              lwd = scalar_default(lwd, par()$lwd)
            )
            if (plot.errors.type == "all" && !is.null(lerr.all) && !is.null(herr.all)) {
              .np_plot_draw_all_band_legend(
                legend = plot.legend,
                x = "topright",
                lty = 1,
                lwd = 2.15 * scalar_default(lwd, par()$lwd)
              )
            }
          }

          rotation.progress <- .np_plot_rotation_progress_tick(rotation.progress, done = frame.idx)
          Sys.sleep(if (isTRUE(rotate)) rotate.defaults$sleep else 0.5)
      }
      

      if (plot.behavior == "plot-data")
        return ( list(d1 = d1) )        

    } else {

      plot.layout <- .np_plot_layout_begin(
        plot.behavior = plot.behavior,
        plot.par.mfrow = plot.par.mfrow,
        mfrow = n2mfrow(bws$ndim)
      )

      ev = xdat[1,,drop = FALSE]

      for (i in seq_len(bws$ndim))
        ev[1,i] = uocquantile(xdat[,i], prob=xq[i])

      maxneval = max(c(sapply(xdat,nlevels),neval))

      exdat = xdat[rep(1, maxneval), , drop = FALSE]

      for (i in seq_len(bws$ndim))
        exdat[,i] = ev[1,i]
      
      if (common.scale){
        data.eval = matrix(data = NA, nrow = maxneval, ncol = bws$ndim)
        data.err = matrix(data = NA, nrow = maxneval, ncol = 3*bws$ndim)
        data.err.all = vector("list", bws$ndim)
        allei = as.data.frame(matrix(data = NA, nrow = maxneval, ncol = bws$ndim))
        all.bxp = list()
      }

      plot.out = list()

      temp.err = matrix(data = NA, nrow = maxneval, ncol = 3)
      temp.dens = replicate(maxneval, NA)
      ## plotting controls
      plot.bootstrap <- plot.errors.method == "bootstrap"
      plotOnEstimate <- (plot.errors.center == "estimate")

      ## density / distribution expressions
      for (i in seq_len(bws$ndim)){
        temp.err[,] = NA
        temp.dens[] =  NA
        temp.boot = list()
        temp.all.err <- NULL

        xi.factor = is.factor(xdat[,i])

        if (xi.factor){
          ei = levels(xdat[,i])
          ei = factor(ei, levels = ei)
          xi.neval = length(ei)
        } else {
          xi.neval = neval
          qi = trim.quantiles(xdat[,i], xtrim[i])
          ei = seq(qi[1], qi[2], length.out = neval)
        }

        if (xi.neval < maxneval){
          ei[(xi.neval+1):maxneval] = NA
        }

        eval.slice <- subcol(exdat,ei,i)[seq_len(xi.neval),, drop = FALSE]
        tobj <- .np_plot_unconditional_eval(
          xdat = xdat,
          exdat = eval.slice,
          bws = bws,
          cdf = FALSE,
          need.asymptotic = identical(plot.errors.method, "asymptotic")
        )
        temp.dens[seq_len(xi.neval)] <- tobj$dens

        if (plot.errors){
          temp.payload <- .np_plot_interval_payload(
            estimate = temp.dens[seq_len(xi.neval)],
            se = tobj$derr,
            plot.errors.method = plot.errors.method,
            plot.errors.alpha = plot.errors.alpha,
            plot.errors.type = plot.errors.type,
            plot.errors.center = plot.errors.center,
            bootstrap_raw = if (identical(plot.errors.method, "bootstrap")) {
              compute.bootstrap.errors(
                xdat = xdat,
                exdat = subcol(exdat,ei,i)[seq_len(xi.neval),, drop = FALSE],
                cdf = FALSE,
                slice.index = i,
                plot.errors.boot.method = plot.errors.boot.method,
                plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
                plot.errors.boot.blocklen = plot.errors.boot.blocklen,
                plot.errors.boot.num = plot.errors.boot.num,
                plot.errors.center = plot.errors.center,
                plot.errors.type = plot.errors.type,
                plot.errors.alpha = plot.errors.alpha,
                bws = bws
              )
            } else {
              NULL
            }
          )
          temp.err[seq_len(xi.neval),] <- temp.payload$err
          temp.all.err <- temp.payload$all.err
          temp.boot <- temp.payload$bxp
          if (identical(plot.errors.method, "bootstrap") && !plot.bxp.out){
            temp.boot$out <- numeric()
            temp.boot$group <- integer()
          }
        }
        
        if (common.scale){
          allei[,i] = ei
          data.eval[,i] = temp.dens
          if (plot.errors){
            all.bxp[i] = NA
            all.bxp[[i]] = temp.boot

            data.err[,c(3*i-2,3*i-1,3*i)] = temp.err
            data.err.all[[i]] = temp.all.err
          }
        } else if (plot.behavior != "data") {
          plot.layout <- .np_plot_layout_activate(plot.layout)
          ## plot evaluation
          plot.fun <- if (xi.factor) {
            .np_plot_panel_fun(plot.bootstrap = plot.bootstrap, plot.bxp = plot.bxp)
          } else {
            plot
          }
          plot.args <- list()
          if (xi.factor) {
            if (plot.bootstrap && plot.bxp) plot.args$z <- temp.boot else plot.args$f <- ei
          } else {
            plot.args$x <- ei
          }
          if (!(xi.factor && plot.bootstrap && plot.bxp))
            plot.args$y <- temp.dens
          if (plot.errors)
            plot.args$ylim <- if (plot.errors.type == "all")
              compute.all.error.range(if (plotOnEstimate) temp.dens else temp.err[,3], temp.all.err)
            else
              compute.default.error.range(if (plotOnEstimate) temp.dens else temp.err[,3], temp.err)
          plot.args$xlab <- scalar_default(xlab, gen.label(bws$xnames[i], paste("X", i, sep = "")))
          plot.args$ylab <- scalar_default(ylab, "Density")
          if (!xi.factor) {
            plot.args$type <- scalar_default(type, "l")
            plot.args$lty <- scalar_default(lty, par()$lty)
            plot.args$col <- scalar_default(col, par()$col)
            plot.args$lwd <- scalar_default(lwd, par()$lwd)
            plot.args$cex.axis <- scalar_default(cex.axis, par()$cex.axis)
            plot.args$cex.lab <- scalar_default(cex.lab, par()$cex.lab)
            plot.args$cex.main <- scalar_default(cex.main, par()$cex.main)
            plot.args$cex.sub <- scalar_default(cex.sub, par()$cex.sub)
          } else {
            if (!is.null(col)) plot.args$col <- col
            if (!is.null(lty)) plot.args$lty <- lty
            if (!is.null(lwd)) plot.args$lwd <- lwd
          }
          plot.args$main <- scalar_default(main, "")
          plot.args$sub <- scalar_default(sub, "")
          plot.args <- .np_plot_merge_user_args(
            plot.args,
            if (xi.factor && plot.bootstrap && plot.bxp) bxp.args else plot.user.args
          )
          .np_plot_first_render_begin(first.render)
          do.call(plot.fun, plot.args)
          .np_plot_first_render_end(first.render)
          if (plot.rug && !xi.factor)
            .np_plot_draw_rug_1d(xdat[,i])

          ## error plotting evaluation
          if (plot.errors && !(xi.factor && plot.bootstrap && plot.bxp)){
            if (!xi.factor && !plotOnEstimate)
              lines(na.omit(ei), na.omit(temp.err[,3]), lty = 3)

            if (plot.errors.type == "all") {
              draw.all.error.types(
                ex = as.numeric(na.omit(ei)),
                center = as.numeric(na.omit(if (plotOnEstimate) temp.dens else temp.err[,3])),
                all.err = temp.all.err,
                plot.errors.style = if (xi.factor) "bar" else plot.errors.style,
                plot.errors.bar = if (xi.factor) "I" else plot.errors.bar,
                plot.errors.bar.num = plot.errors.bar.num,
                lty = 2,
                add.legend = TRUE,
                legend = plot.legend)
            } else {
              draw.args <- list(
                ex = as.numeric(na.omit(ei)),
                ely = if (plotOnEstimate) na.omit(temp.dens - temp.err[,1]) else na.omit(temp.err[,3] - temp.err[,1]),
                ehy = if (plotOnEstimate) na.omit(temp.dens + temp.err[,2]) else na.omit(temp.err[,3] + temp.err[,2]),
                plot.errors.style = if (xi.factor) "bar" else plot.errors.style,
                plot.errors.bar = if (xi.factor) "I" else plot.errors.bar,
                plot.errors.bar.num = plot.errors.bar.num,
                lty = scalar_default(lty, if (xi.factor) 1 else 2)
              )
              do.call(draw.errors, draw.args)
            }
          }
        }

        if (plot.behavior != "plot") {
          plot.out[i] = NA
          plot.out[[i]] <- npdensity(
            bws = bws,
            eval = eval.slice,
            dens = na.omit(temp.dens),
            derr = na.omit(cbind(-temp.err[,1], temp.err[,2])),
            ntrain = bws$nobs
          )
          plot.out[[i]]$bias = na.omit(temp.dens - temp.err[,3])
          plot.out[[i]]$bxp = temp.boot
        }
      }
      
      if (common.scale && (plot.behavior != "data")){
        jj = seq_len(bws$ndim)*3

        if (plot.errors && plot.errors.type == "all") {
          y.min <- Inf
          y.max <- -Inf
          for (k in seq_len(bws$ndim)) {
            if (is.null(data.err.all[[k]])) next
            nkeep.k <- nrow(data.err.all[[k]]$pointwise)
            center.k <- if (plot.errors.center == "estimate")
              data.eval[seq_len(nkeep.k),k]
            else
              data.err[seq_len(nkeep.k),3*k]
            range.k <- compute.all.error.range(center.k, data.err.all[[k]])
            y.min <- min(y.min, range.k[1], na.rm = TRUE)
            y.max <- max(y.max, range.k[2], na.rm = TRUE)
          }
          if (!is.finite(y.min) || !is.finite(y.max)) {
            if (plot.errors.center == "estimate") {
              y.max = max(na.omit(as.double(data.eval)) + na.omit(as.double(data.err[,jj-1])))
              y.min = min(na.omit(as.double(data.eval)) - na.omit(as.double(data.err[,jj-2])))
            } else {
              y.max = max(na.omit(as.double(data.err[,jj] + data.err[,jj-1])))
              y.min = min(na.omit(as.double(data.err[,jj] - data.err[,jj-2])))
            }
          }
        } else if (plot.errors.center == "estimate" || !plot.errors) {
          y.max = max(na.omit(as.double(data.eval)) +
            if (plot.errors) na.omit(as.double(data.err[,jj-1]))
            else 0
            )
          y.min = min(na.omit(as.double(data.eval)) -
            if (plot.errors) na.omit(as.double(data.err[,jj-2]))
            else 0
            )
        } else if (plot.errors.center == "bias-corrected") {
          y.max = max(na.omit(as.double(data.err[,jj] + data.err[,jj-1])))
          y.min = min(na.omit(as.double(data.err[,jj] - data.err[,jj-2])))
        }

        if(!is.null(ylim)){
          y.min = ylim[1]
          y.max = ylim[2]
        }
        
        for (i in seq_len(bws$ndim)){
          xi.factor = is.factor(xdat[,i])
          plot.layout <- .np_plot_layout_activate(plot.layout)

          ## plot evaluation
          plot.fun <- if (xi.factor) {
            .np_plot_panel_fun(plot.bootstrap = plot.bootstrap, plot.bxp = plot.bxp)
          } else {
            plot
          }
          plot.args <- list()
          if (xi.factor) {
            if (plot.bootstrap && plot.bxp) plot.args$z <- all.bxp[[i]] else plot.args$f <- allei[,i]
          } else {
            plot.args$x <- allei[,i]
          }
          if (!(xi.factor && plot.bootstrap && plot.bxp))
            plot.args$y <- data.eval[,i]
          plot.args$ylim <- c(y.min, y.max)
          plot.args$xlab <- scalar_default(xlab, gen.label(bws$xnames[i], paste("X", i, sep = "")))
          plot.args$ylab <- scalar_default(ylab, "Density")
          if (!xi.factor) {
            plot.args$type <- scalar_default(type, "l")
            plot.args$lty <- scalar_default(lty, par()$lty)
            plot.args$col <- scalar_default(col, par()$col)
            plot.args$lwd <- scalar_default(lwd, par()$lwd)
            plot.args$cex.axis <- scalar_default(cex.axis, par()$cex.axis)
            plot.args$cex.lab <- scalar_default(cex.lab, par()$cex.lab)
            plot.args$cex.main <- scalar_default(cex.main, par()$cex.main)
            plot.args$cex.sub <- scalar_default(cex.sub, par()$cex.sub)
          } else {
            if (!is.null(col)) plot.args$col <- col
            if (!is.null(lty)) plot.args$lty <- lty
            if (!is.null(lwd)) plot.args$lwd <- lwd
          }
          plot.args$main <- scalar_default(main, "")
          plot.args$sub <- scalar_default(sub, "")
          plot.args <- .np_plot_merge_user_args(
            plot.args,
            if (xi.factor && plot.bootstrap && plot.bxp) bxp.args else plot.user.args
          )
          .np_plot_first_render_begin(first.render)
          do.call(plot.fun, plot.args)
          .np_plot_first_render_end(first.render)
          if (plot.rug && !xi.factor)
            .np_plot_draw_rug_1d(xdat[,i])

          ## error plotting evaluation
          if (plot.errors && !(xi.factor && plot.bootstrap && plot.bxp)){
            if (!xi.factor && !plotOnEstimate)
              lines(na.omit(ei), na.omit(temp.err[,3]), lty = 3)

            if (plot.errors.type == "all") {
              draw.all.error.types(
                ex = as.numeric(na.omit(allei[,i])),
                center = as.numeric(na.omit(if (plotOnEstimate) data.eval[,i] else data.err[,3*i])),
                all.err = data.err.all[[i]],
                plot.errors.style = if (xi.factor) "bar" else plot.errors.style,
                plot.errors.bar = if (xi.factor) "I" else plot.errors.bar,
                plot.errors.bar.num = plot.errors.bar.num,
                lty = 2,
                add.legend = TRUE,
                legend = plot.legend)
            } else {
              draw.args <- list(
                ex = as.numeric(na.omit(allei[,i])),
                ely = if (plotOnEstimate) na.omit(data.eval[,i] - data.err[,3*i-2]) else na.omit(data.err[,3*i] - data.err[,3*i-2]),
                ehy = if (plotOnEstimate) na.omit(data.eval[,i] + data.err[,3*i-1]) else na.omit(data.err[,3*i] + data.err[,3*i-1]),
                plot.errors.style = if (xi.factor) "bar" else plot.errors.style,
                plot.errors.bar = if (xi.factor) "I" else plot.errors.bar,
                plot.errors.bar.num = plot.errors.bar.num,
                lty = scalar_default(lty, if (xi.factor) 1 else 2)
              )
              do.call(draw.errors, draw.args)
            }
          }
        }
      }

      if (plot.behavior != "data" && plot.par.mfrow)
        par(mfrow=c(1,1),cex=par()$cex)

      if (plot.behavior != "plot"){
        names(plot.out) = paste("d",seq_len(bws$ndim),sep="")
        return (plot.out)
      }
    }
  }

deptest <- function(Srho,
                    Srho.bootstrap.vec,
                    P,
                    bootstrap,
                    boot.num,
                    bw.data.x,
                    bw.data.y,
                    bw.joint) {
  
  tdep = list(Srho = Srho,
    Srho.bootstrap.vec = Srho.bootstrap.vec,
    P = P,
    bootstrap = bootstrap,
    boot.num = boot.num,
    bw.data.x = bw.data.x,
    bw.data.y = bw.data.y,
    bw.joint = bw.joint)

  if(is.null(P)) {

    tdep$reject <- NA
    tdep$rejectNum <- NA

  } else {

    reject <- ' '
    
    if (P < 0.1)
      reject <- '.'
    
    if (P < 0.05)
      reject <- '*'
    
    if (P < 0.01)
      reject <- '**'
    
    if (P < 0.001)
      reject <- '***'
    
    tdep$reject <- reject
    tdep$rejectNum <- switch(reject,
                             ' ' = 100,
                             '.' = 10,
                             '*' = 5,
                             '**' = 1,
                             '***' = 0.1)
    
  }
  
  class(tdep) <- "deptest"

  tdep
}

print.deptest <- function(x, ...){
  if(x$bootstrap) {
    reject.msg <- if (x$reject == ' ') {
      paste("\nFail to reject the null of independence at the 10% level\n\n", sep = "")
    } else {
      paste("\nNull of independence is rejected at the ", x$rejectNum, "% level\n\n", sep = "")
    }
    cat("\nConsistent Metric Entropy Test for Dependence",
        paste("\n", format(x$boot.num), " Bootstrap Replications",sep=""))
    cat(paste("\n\nTest Statistic ", sQuote("Srho"), ": ",
              format(x$Srho), "\tP Value: ", format.pval(x$P)," ", x$reject,sep=""))
    cat("\n---\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")
    cat(reject.msg)
  } else {
    cat("\nConsistent Dependence Metric Entropy",
        paste("\nStatistic ", sQuote("Srho"),sep=""), ": ",
        format(x$Srho),"\n\n",sep="")
  }
}

summary.deptest <- function(object, ...){
  print(object)
}

npcmstest <- function(formula,
                      data = NULL,
                      subset,
                      xdat,
                      ydat,
                      model = stop(paste(sQuote("model")," has not been provided")),
                      distribution = c("bootstrap", "asymptotic"),
                      boot.method=c("iid","wild","wild-rademacher"),
                      boot.num = 399,
                      pivot = TRUE,
                      density.weighted = TRUE,
                      random.seed = 42,
                      ...) {
  
  pcall = paste(deparse(model$call),collapse="")
  if(length(grep("x = (T|TRUE)[ ,)]", pcall)) == 0 || length(grep("y = (T|TRUE)[ ,)]", pcall)) == 0)
    stop(paste(sQuote("model")," is missing components ", sQuote("x"), " and ",
               sQuote("y"), ".\nTo fix this please invoke ", sQuote("lm"),
               " or ", sQuote("glm"),
               " with ", sQuote("x=TRUE"), " and ", sQuote("y=TRUE"),
               ".\nSee help for further info.", sep=""))

  if(boot.num < 9) stop("number of bootstrap replications must be >= 9")

  ## checking for consistent interface usage
  miss.xy = c(missing(xdat),missing(ydat))
  miss.f = missing(formula)
    
  if (any(miss.xy) && !all(miss.xy))
    stop("one of, but not both, xdat and ydat was specified")
  else if(all(miss.xy) & miss.f)
    stop("xdat, and ydat, are missing, and no formula is specified.")
  else if(all(miss.xy) & !miss.f){
    mf.args <- list(formula = formula, data = data, na.action = na.omit)
    if (!missing(subset))
      mf.args$subset <- subset
    mf <- do.call(model.frame, mf.args)
    
    ydat <- model.response(mf)
    xdat <- mf[, attr(attr(mf, "terms"),"term.labels"), drop = FALSE]

    na.index <- unclass(attr(xdat,"na.action"))
  } else if(!miss.f){
    stop(paste("A formula was specified along with xdat and ydat.\n",
               "Please see the documentation on proper interface usage."))
  } else {
    xdat = toFrame(xdat)

    ## catch and destroy NA's
    keep.rows <- rep_len(TRUE, nrow(xdat))
    rows.omit <- attr(na.omit(data.frame(xdat, ydat)), "na.action")
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Data has no rows without NAs")

    xdat <- xdat[keep.rows,,drop = FALSE]
    ydat <- ydat[keep.rows]

    na.index <- which(!keep.rows)
  }

  ## Save seed prior to setting

  seed.state <- .np_seed_enter(random.seed)


  distribution = match.arg(distribution)
  boot.method = match.arg(boot.method)

  ## Here we go...

  model.resid <- residuals(model, type = "response")

  n = length(model.resid)

  ## ydat is model's residuals, xdat all regressors with types

  ##  bw <- npregbw(xdat=xdat,ydat=model.resid)

  .np_progress_note("Computing bandwidths")

  bw <- .np_progress_with_legacy_suppressed(npregbw(xdat=xdat, ydat=model$y, ...))
  
  ## Now define the Jn test statistic that takes arguments xdat, the
  ## residual vector, the bandwidth object, and the number of bootstrap
  ## replications

  fhat <- 1

  prodh <- if (bw$ncon == 0) 1.0
  else
    prod(bw$bw[bw$icon])

  if (!density.weighted)
    fhat <- npksum(txdat = xdat,
                   bws = bw$bw, leave.one.out = TRUE,
                   bandwidth.divide = TRUE, ...)$ksum/n


  if(min(fhat) == 0)
  stop(paste(sep="","\nAttempt to divide by zero density.",
             "\nYou can try re-running the test with `density.weighted=TRUE'\n"))

  In <- function(xdat, model.resid, bw) {
    
    ## n is the number of observations

    n <- length(model.resid)

    ## Compute In (equation 2.10, Hsiao/Li/racine 2005)

    return( sum(model.resid*npksum(txdat=xdat,
                                   tydat=model.resid,
                                   bws=bw$bw,
                                   leave.one.out=TRUE,
                                   bandwidth.divide=TRUE, ...)$ksum/fhat)/n^2 )
  }

  Omega.hat <- function(xdat, model.resid, bw) {
  
    ## Variance of In (equation 2.11, Hsiao/Li/racine 2005)

    n <- length(model.resid)
    
    return( 2*prodh*
           sum(model.resid^2*
               npksum(txdat=xdat,
                      tydat=model.resid^2,
                      bws=bw$bw,
                      leave.one.out=TRUE,
                      kernel.pow=2,
                      bandwidth.divide=TRUE, ...)$ksum/fhat^2)/n^2 )
  }

  Jn <- function(xdat, model.resid, bw) {
    ## Compute the statistic, supposed to be N(0,1) asymptotically
    n <- length(model.resid)
    n*sqrt(prodh)*In(xdat, model.resid, bw)/sqrt(Omega.hat(xdat, model.resid, bw))
  }


  ## Now conduct a wild bootstrap.. yhat is the fitted model, and we have
  ## ols.resid above... these are external in scope to boot.wild

  yhat <- fitted(model)

  ## data is y,xdat for the OLS model...

  ## jracine March 8, 2006... not using boot() library (problematic I
  ## realized with [indices] hence unnecessary)

  draw.wild.mult <- function(n.obs, a, b, p.a) {
    u <- stats::runif(n.obs)
    mult <- rep.int(b, n.obs)
    mult[u <= p.a] <- a
    mult
  }

  boot.wild <- function(model.resid) {

    a <- -0.6180339887499 # (1-sqrt(5))/2
    P.a <-0.72360679774998 # (1+sqrt(5))/(2*sqrt(5))
    b <- 1.6180339887499 # (1+sqrt(5))/2

    ## Use the wild bootstrap to get a bootstrap vector for y under the
    ## null that the model is correct. Alternatively, we could pairwise
    ## resample Z={y,xdat}

    ## jracine removed [indices]

    y.star <- yhat + model.resid * draw.wild.mult(length(model.resid), a, b, P.a)
    resid <-
      if(is.null(model$family)) {
        residuals(glm(y.star~ model$x - 1), type = "response")
      } else {
        residuals(glm(y.star~ model$x - 1,family=model$family), type = "response")
      }
    
    return(if (pivot) Jn(xdat, resid, bw)
           else In(xdat, resid, bw))
  }

  boot.wild.rademacher <- function(model.resid) {

    a <- -1
    P.a <- 0.5
    b <- 1

    ## Use the wild bootstrap to get a bootstrap vector for y under
    ## the null that the model is correct, using Rademacher variables

    ## jracine removed [indices]

    y.star <- yhat + model.resid * draw.wild.mult(length(model.resid), a, b, P.a)
    resid <-
      if(is.null(model$family)) {
        residuals(glm(y.star~ model$x - 1), type = "response")
      } else {
        residuals(glm(y.star~ model$x - 1,family=model$family), type = "response")
      }
    
    return(if (pivot) Jn(xdat, resid, bw)
           else In(xdat, resid, bw))
  }

  boot.iid <- function(model.resid) {

    y.star <- yhat + model.resid[sample.int(length(model.resid), replace = TRUE)]
    resid <-
      if(is.null(model$family)) {
        residuals(glm(y.star~ model$x - 1), type = "response")
      } else {
        residuals(glm(y.star~ model$x - 1,family=model$family), type = "response")
      }
    
    return(if (pivot) Jn(xdat, resid, bw)
           else In(xdat, resid, bw))
  }

  if(distribution == "bootstrap"){
    Sn.bootstrap <- numeric(boot.num)
    progress <- .np_progress_begin("Bootstrap replications", total = boot.num, surface = "bootstrap")

    for (ii in seq_len(boot.num)) {
       if(boot.method == "iid"){
        Sn.bootstrap[ii] <- boot.iid(model.resid)
      } else if(boot.method == "wild"){
        Sn.bootstrap[ii] <- boot.wild(model.resid)
      } else if(boot.method == "wild-rademacher"){
        Sn.bootstrap[ii] <- boot.wild.rademacher(model.resid)
      }
      progress <- .np_progress_step(progress, done = ii)
    }
    progress <- .np_progress_end(progress)
    Sn.bootstrap <- sort(Sn.bootstrap)
    ##cat("\n")
  }


  ##  Return a list containing the test statistic etc.

  tIn = In(xdat, model.resid, bw)
  to.h = Omega.hat(xdat, model.resid, bw)

  s.d =
    if (pivot) 1.0
    else sqrt(to.h/prodh)/n

  if(distribution == "asymptotic") {
    
    tJn = list(
      Jn = n*sqrt(prodh)*tIn/sqrt(to.h),
      In = tIn,
      Omega.hat = to.h,
      q.90=qnorm(p = .90, sd = s.d),
      q.95=qnorm(p = .95, sd = s.d),
      q.99=qnorm(p = .99, sd = s.d),
      bw = bw,
      Jn.bootstrap = NA,
      In.bootstrap = NA,
      pivot = pivot)

    Sn = if (pivot) tJn$Jn else tIn

    tJn$P <- (1-pnorm(Sn, sd = s.d))

  } else {
    tJn = list(
      Jn = n*sqrt(prodh)*tIn/sqrt(to.h),
      In = tIn,
      Omega.hat = to.h,
      q.90=Sn.bootstrap[ceiling(0.90*boot.num)],
      q.95=Sn.bootstrap[ceiling(0.95*boot.num)],
      q.99=Sn.bootstrap[ceiling(0.99*boot.num)],
      bw=bw,
      Jn.bootstrap = if(pivot) Sn.bootstrap else NA,
      In.bootstrap = if(pivot) NA else Sn.bootstrap,
      pivot = pivot
      )

    Sn = if (pivot) tJn$Jn else tIn

    tJn$P <- mean(Sn.bootstrap > Sn)

    
  }
  
  ## Restore seed

  .np_seed_exit(seed.state)
  
  cmstest(Jn = tJn$Jn,
          In = tJn$In,
          Omega.hat = tJn$Omega.hat,
          sd = s.d,
          q.90 = tJn$q.90,
          q.95 = tJn$q.95,
          q.99 = tJn$q.99,
          P = tJn$P,
          bws = bw,
          distribution = distribution,
          Jn.bootstrap = tJn$Jn.bootstrap,
          In.bootstrap = tJn$In.bootstrap,
          pivot = pivot,
          model = model,
          boot.method = boot.method,
          boot.num = boot.num,
          na.index = na.index)
}

.np_plot_proto_check_npcdens_fixed_surface <- function(bws,
                                                       xdat,
                                                       ydat,
                                                       neval,
                                                       xtrim,
                                                       ytrim,
                                                       cdf = FALSE) {
  expected.class <- if (isTRUE(cdf)) "condbandwidth" else "conbandwidth"
  expected.label <- if (isTRUE(cdf)) "conditional distribution" else "conditional density"
  if (!inherits(bws, expected.class))
    stop(sprintf("prototype route requires a %s bandwidth object", expected.label), call. = FALSE)
  regtype.engine <- if (is.null(bws$regtype.engine)) {
    if (is.null(bws$regtype)) "lc" else as.character(bws$regtype)
  } else {
    as.character(bws$regtype.engine)
  }
  if (!is.element(regtype.engine, c("lc", "lp")))
    stop("prototype route currently supports regtype='lc', 'll', or 'lp' only", call. = FALSE)
  if (!is.element(as.character(bws$type), c("fixed", "generalized_nn", "adaptive_nn")))
    stop("prototype route currently supports fixed, generalized-NN, or adaptive-NN bandwidths only", call. = FALSE)
  if (bws$xndim != 1L || bws$yndim != 1L)
    stop("prototype route currently supports one x variable and one y variable", call. = FALSE)
  if (bws$xnuno + bws$ynuno != 0L)
    stop("prototype route currently supports continuous/ordered surface variables only", call. = FALSE)
  if (bws$xncon + bws$xnord + bws$yncon + bws$ynord != 2L)
    stop(sprintf("prototype route requires a two-dimensional %s surface", expected.label), call. = FALSE)
  if (!is.numeric(neval) || length(neval) != 1L || is.na(neval) || neval < 2L)
    stop("prototype route requires scalar neval >= 2", call. = FALSE)
  invisible(TRUE)
}

.np_plot_proto_check_npcdens_lc_fixed_none <- .np_plot_proto_check_npcdens_fixed_surface

.np_plot_proto_clean_conditional_data <- function(xdat, ydat) {
  ## Contract: align explicit training data for the first prototype route. This
  ## stage intentionally does not recover data from formula/call objects; that
  ## wider state-resolution contract belongs to a later slice.
  xdat <- toFrame(xdat)
  ydat <- toFrame(ydat)
  keep.rows <- rep_len(TRUE, nrow(xdat))
  rows.omit <- attr(na.omit(data.frame(xdat, ydat)), "na.action")
  if (length(rows.omit) > 0L)
    keep.rows[as.integer(rows.omit)] <- FALSE
  if (!any(keep.rows))
    stop("Data has no rows without NAs")
  xdat <- xdat[keep.rows, , drop = FALSE]
  ydat <- ydat[keep.rows, , drop = FALSE]

  list(xdat = xdat, ydat = ydat)
}

.np_plot_proto_conditional_surface_grid <- function(bws,
                                                    xdat,
                                                    ydat,
                                                    neval,
                                                    xtrim,
                                                    ytrim) {
  ## Contract: build the two-dimensional conditional surface grid. This helper
  ## owns target construction only; it must not evaluate the estimator or create
  ## plot objects.
  xtrim <- double(bws$xndim) + xtrim
  ytrim <- double(bws$yndim) + ytrim

  if (is.ordered(xdat[, 1L])) {
    x1.eval <- bws$xdati$all.ulev[[1L]]
    x1.neval <- length(x1.eval)
  } else {
    x1.neval <- as.integer(neval)
    qi <- trim.quantiles(xdat[, 1L], xtrim[1L])
    x1.eval <- seq(qi[1L], qi[2L], length.out = x1.neval)
  }

  tx2 <- ydat[, 1L]
  txdati <- bws$ydati
  txtrim <- ytrim

  if (txdati$iord[1L]) {
    x2.eval <- txdati$all.ulev[[1L]]
    x2.neval <- length(x2.eval)
  } else {
    x2.neval <- as.integer(neval)
    qi <- trim.quantiles(tx2, txtrim[1L])
    x2.eval <- seq(qi[1L], qi[2L], length.out = x2.neval)
  }

  x.eval <- expand.grid(x1.eval, x2.eval)
  if (bws$xdati$iord[1L])
    x1.eval <- bws$xdati$all.dlev[[1L]][as.integer(x1.eval)]
  if (txdati$iord[1L])
    x2.eval <- txdati$all.dlev[[1L]][as.integer(x2.eval)]

  list(
    x.eval = x.eval,
    x1.eval = x1.eval,
    x2.eval = x2.eval,
    x1.neval = x1.neval,
    x2.neval = x2.neval
  )
}

.np_plot_proto_npcdens_fixed_data <- function(bws,
                                              xdat,
                                              ydat,
                                              neval = 50,
                                              xtrim = 0.0,
                                              ytrim = 0.0,
                                              plot.errors.method = c("none", "asymptotic", "bootstrap"),
                                              plot.errors.boot.method = c("inid", "fixed", "geom"),
                                              plot.errors.boot.nonfixed = c("exact", "frozen"),
                                              plot.errors.boot.blocklen = NULL,
                                              plot.errors.boot.num = 1999,
                                              plot.errors.center = c("estimate", "bias-corrected"),
                                              plot.errors.type = c("pmzsd", "pointwise", "bonferroni",
                                                                   "simultaneous", "all"),
                                              plot.errors.alpha = 0.05,
                                              proper = FALSE,
                                              proper.method = c("project"),
                                              proper.control = list(),
                                              return.stages = FALSE,
                                              cdf = FALSE) {
  ## Contract: private conditional density/distribution data-only prototype.
  ## This owns explicit data cleanup, target construction, evaluator invocation,
  ## optional asymptotic/bootstrap interval construction, and old-compatible
  ## plot-data assembly.
  ## It must not draw graphics, bootstrap, change RNG state, or recover formula
  ## data until those stages receive their own slice.
  if (missing(xdat) || missing(ydat))
    stop("prototype route requires explicit xdat and ydat", call. = FALSE)
  plot.errors.method <- match.arg(plot.errors.method)
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  plot.errors.boot.nonfixed <- match.arg(plot.errors.boot.nonfixed)
  plot.errors.center <- match.arg(plot.errors.center)
  plot.errors.type <- match.arg(plot.errors.type)
  if (missing(proper.method))
    proper.method <- if (isTRUE(cdf)) "isotonic" else "project"
  proper.args <- if (isTRUE(cdf)) {
    .np_condist_validate_proper_args(
      proper = proper,
      proper.method = proper.method,
      proper.control = proper.control
    )
  } else {
    .np_condens_validate_proper_args(
      proper = proper,
      proper.method = proper.method,
      proper.control = proper.control
    )
  }
  dat <- .np_plot_proto_clean_conditional_data(xdat = xdat, ydat = ydat)
  xdat <- dat$xdat
  ydat <- dat$ydat
.np_plot_proto_check_npcdens_fixed_surface(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    ytrim = ytrim,
    cdf = cdf
  )

  grid <- .np_plot_proto_conditional_surface_grid(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    ytrim = ytrim
  )
  fit <- .np_plot_conditional_eval(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    exdat = grid$x.eval[, 1L, drop = FALSE],
    eydat = grid$x.eval[, 2L, drop = FALSE],
    cdf = isTRUE(cdf),
    gradients = FALSE,
    proper = isTRUE(proper.args$proper.requested),
    proper.method = proper.args$proper.method,
    proper.control = proper.args$proper.control
  )

  tcomp <- if (isTRUE(cdf)) fit$condist else fit$condens
  terr <- matrix(fit$conderr, nrow = length(tcomp), ncol = 3L)
  terr[, 3L] <- NA_real_
  interval <- NULL
  bootstrap <- NULL
  if (identical(plot.errors.method, "asymptotic")) {
    interval <- .np_plot_asymptotic_error_from_se(
      se = fit$conderr,
      alpha = plot.errors.alpha,
      band.type = plot.errors.type,
      m = nrow(grid$x.eval[, 1L])
    )
    terr[, 1:2] <- interval$err
  } else if (identical(plot.errors.method, "bootstrap")) {
    bootstrap <- compute.bootstrap.errors(
      xdat = xdat,
      ydat = ydat,
      exdat = grid$x.eval[, 1L],
      eydat = grid$x.eval[, 2L],
      cdf = isTRUE(cdf),
      quantreg = FALSE,
      tau = 0.5,
      gradients = FALSE,
      gradient.index = 0L,
      slice.index = 0L,
      plot.errors.boot.method = plot.errors.boot.method,
      plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
      plot.errors.boot.blocklen = plot.errors.boot.blocklen,
      plot.errors.boot.num = plot.errors.boot.num,
      plot.errors.center = plot.errors.center,
      plot.errors.type = plot.errors.type,
      plot.errors.alpha = plot.errors.alpha,
      progress.target = NULL,
      proper = isTRUE(proper.args$proper.requested),
      proper.method = proper.args$proper.method,
      proper.control = proper.args$proper.control,
      bws = bws
    )
    terr <- bootstrap$boot.err
    interval <- list(
      err = bootstrap$boot.err[, 1:2, drop = FALSE],
      all.err = bootstrap$boot.all.err
    )
  }

  cd.args <- list(
    bws = bws,
    xeval = grid$x.eval[, 1L],
    yeval = grid$x.eval[, 2L],
    ntrain = nrow(xdat),
    conderr = terr[, 1:2, drop = FALSE],
    proper.requested = fit$proper.requested,
    proper.applied = fit$proper.applied,
    proper.method = fit$proper.method,
    proper.info = fit$proper.info
  )
  if (isTRUE(cdf)) {
    cd.args$condist <- fit$condist
    cd.args$condist.raw <- fit$condist.raw
    cd1 <- do.call(condistribution, cd.args)
  } else {
    cd.args$condens <- fit$condens
    cd.args$condens.raw <- fit$condens.raw
    cd1 <- do.call(condensity, cd.args)
  }
  cd1$bias <- NA

  plot.data <- list(cd1 = cd1)
  if (!isTRUE(return.stages))
    return(plot.data)

  list(
    state = list(
      bws = bws,
      xdat = xdat,
      ydat = ydat,
      ntrain = nrow(xdat),
      family = if (isTRUE(cdf)) "npcdist" else "npcdens",
      cdf = isTRUE(cdf),
      gradients = FALSE,
      proper = proper.args
    ),
    target_grid = grid,
    evaluator = fit,
    intervals = if (is.null(interval)) NULL else list(
      method = plot.errors.method,
      type = plot.errors.type,
      alpha = plot.errors.alpha,
      err = interval$err,
      all.err = interval$all.err
    ),
    bootstrap = if (is.null(bootstrap)) NULL else list(
      method = plot.errors.boot.method,
      nonfixed = plot.errors.boot.nonfixed,
      blocklen = plot.errors.boot.blocklen,
      B = plot.errors.boot.num,
      center = plot.errors.center,
      boot.err = bootstrap$boot.err,
      boot.all.err = bootstrap$boot.all.err,
      bxp = bootstrap$bxp
    ),
    proper_projection = if (!isTRUE(proper.args$proper.requested)) NULL else list(
      requested = fit$proper.requested,
      applied = fit$proper.applied,
      method = fit$proper.method,
      info = fit$proper.info
    ),
    plot_data = plot.data
  )
}

.np_plot_proto_npcdens_lc_fixed_data <- .np_plot_proto_npcdens_fixed_data

.np_plot_proto_npcdens_lc_fixed_none_data <- function(bws,
                                                      xdat,
                                                      ydat,
                                                      neval = 50,
                                                      xtrim = 0.0,
                                                      ytrim = 0.0,
                                                      proper = FALSE,
                                                      proper.method = c("project"),
                                                      proper.control = list(),
                                                      return.stages = FALSE) {
  .np_plot_proto_npcdens_lc_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    ytrim = ytrim,
    plot.errors.method = "none",
    proper = proper,
    proper.method = proper.method,
    proper.control = proper.control,
    return.stages = return.stages
  )
}

.np_plot_proto_npcdens_fixed_none_data <- function(bws,
                                                   xdat,
                                                   ydat,
                                                   neval = 50,
                                                   xtrim = 0.0,
                                                   ytrim = 0.0,
                                                   proper = FALSE,
                                                   proper.method = c("project"),
                                                   proper.control = list(),
                                                   return.stages = FALSE) {
  .np_plot_proto_npcdens_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    ytrim = ytrim,
    plot.errors.method = "none",
    proper = proper,
    proper.method = proper.method,
    proper.control = proper.control,
    return.stages = return.stages
  )
}

.np_plot_proto_npcdist_fixed_none_data <- function(bws,
                                                   xdat,
                                                   ydat,
                                                   neval = 50,
                                                   xtrim = 0.0,
                                                   ytrim = 0.0,
                                                   proper = FALSE,
                                                   proper.method = c("isotonic"),
                                                   proper.control = list(),
                                                   return.stages = FALSE) {
  .np_plot_proto_npcdens_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    ytrim = ytrim,
    plot.errors.method = "none",
    proper = proper,
    proper.method = proper.method,
    proper.control = proper.control,
    return.stages = return.stages,
    cdf = TRUE
  )
}

.np_plot_proto_npcdist_fixed_bootstrap_inid_data <- function(bws,
                                                            xdat,
                                                            ydat,
                                                            neval = 50,
                                                            xtrim = 0.0,
                                                            ytrim = 0.0,
                                                            plot.errors.boot.num = 1999,
                                                            plot.errors.center = c("estimate", "bias-corrected"),
                                                            plot.errors.type = c("pmzsd", "pointwise",
                                                                                 "bonferroni", "simultaneous",
                                                                                 "all"),
                                                            plot.errors.alpha = 0.05,
                                                            proper = FALSE,
                                                            proper.method = c("isotonic"),
                                                            proper.control = list(),
                                                            return.stages = FALSE) {
  .np_plot_proto_npcdens_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    ytrim = ytrim,
    plot.errors.method = "bootstrap",
    plot.errors.boot.method = "inid",
    plot.errors.boot.num = plot.errors.boot.num,
    plot.errors.center = plot.errors.center,
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    proper = proper,
    proper.method = proper.method,
    proper.control = proper.control,
    return.stages = return.stages,
    cdf = TRUE
  )
}

.np_plot_proto_npcdens_lc_fixed_asymptotic_data <- function(bws,
                                                           xdat,
                                                           ydat,
                                                           neval = 50,
                                                           xtrim = 0.0,
                                                           ytrim = 0.0,
                                                           plot.errors.type = c("pmzsd", "pointwise",
                                                                                "bonferroni", "simultaneous",
                                                                                "all"),
                                                           plot.errors.alpha = 0.05,
                                                           proper = FALSE,
                                                           proper.method = c("project"),
                                                           proper.control = list(),
                                                           return.stages = FALSE) {
  .np_plot_proto_npcdens_lc_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    ytrim = ytrim,
    plot.errors.method = "asymptotic",
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    proper = proper,
    proper.method = proper.method,
    proper.control = proper.control,
    return.stages = return.stages
  )
}

.np_plot_proto_npcdens_fixed_bootstrap_inid_data <- function(bws,
                                                            xdat,
                                                            ydat,
                                                            neval = 50,
                                                            xtrim = 0.0,
                                                            ytrim = 0.0,
                                                            plot.errors.boot.num = 1999,
                                                            plot.errors.center = c("estimate", "bias-corrected"),
                                                            plot.errors.type = c("pmzsd", "pointwise",
                                                                                 "bonferroni", "simultaneous",
                                                                                 "all"),
                                                            plot.errors.alpha = 0.05,
                                                            proper = FALSE,
                                                            proper.method = c("project"),
                                                            proper.control = list(),
                                                            return.stages = FALSE) {
  .np_plot_proto_npcdens_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    ytrim = ytrim,
    plot.errors.method = "bootstrap",
    plot.errors.boot.method = "inid",
    plot.errors.boot.num = plot.errors.boot.num,
    plot.errors.center = plot.errors.center,
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    proper = proper,
    proper.method = proper.method,
    proper.control = proper.control,
    return.stages = return.stages
  )
}

.np_plot_proto_npcdens_lc_fixed_bootstrap_inid_data <- function(bws,
                                                               xdat,
                                                               ydat,
                                                               neval = 50,
                                                               xtrim = 0.0,
                                                               ytrim = 0.0,
                                                               plot.errors.boot.num = 1999,
                                                               plot.errors.center = c("estimate", "bias-corrected"),
                                                               plot.errors.type = c("pmzsd", "pointwise",
                                                                                    "bonferroni", "simultaneous",
                                                                                    "all"),
                                                               plot.errors.alpha = 0.05,
                                                               proper = FALSE,
                                                               proper.method = c("project"),
                                                               proper.control = list(),
                                                               return.stages = FALSE) {
  .np_plot_proto_npcdens_lc_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    ytrim = ytrim,
    plot.errors.method = "bootstrap",
    plot.errors.boot.method = "inid",
    plot.errors.boot.num = plot.errors.boot.num,
    plot.errors.center = plot.errors.center,
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    proper = proper,
    proper.method = proper.method,
    proper.control = proper.control,
    return.stages = return.stages
  )
}

.np_plot_proto_npcdens_lc_fixed_bootstrap_block_data <- function(bws,
                                                                xdat,
                                                                ydat,
                                                                neval = 50,
                                                                xtrim = 0.0,
                                                                ytrim = 0.0,
                                                                plot.errors.boot.method = c("fixed", "geom"),
                                                                plot.errors.boot.blocklen,
                                                                plot.errors.boot.num = 1999,
                                                                plot.errors.center = c("estimate", "bias-corrected"),
                                                                plot.errors.type = c("pmzsd", "pointwise",
                                                                                     "bonferroni", "simultaneous",
                                                                                     "all"),
                                                                plot.errors.alpha = 0.05,
                                                                proper = FALSE,
                                                                proper.method = c("project"),
                                                                proper.control = list(),
                                                                return.stages = FALSE) {
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  .np_plot_proto_npcdens_lc_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    ytrim = ytrim,
    plot.errors.method = "bootstrap",
    plot.errors.boot.method = plot.errors.boot.method,
    plot.errors.boot.blocklen = plot.errors.boot.blocklen,
    plot.errors.boot.num = plot.errors.boot.num,
    plot.errors.center = plot.errors.center,
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    proper = proper,
    proper.method = proper.method,
    proper.control = proper.control,
    return.stages = return.stages
  )
}

.np_plot_proto_npcdens_surface_base_render <- function(plot.data,
                                                       perspective = TRUE,
                                                       main = NULL,
                                                       xlab = NULL,
                                                       ylab = NULL,
                                                       zlab = NULL,
                                                       col = "lightblue",
                                                       theta = 0.0,
                                                       phi = 20.0,
                                                       ...) {
  ## Contract: renderer smoke for staged npcdens surface data. It consumes a
  ## plot-data object and must not re-enter public estimators, bandwidth
  ## constructors, bootstrap helpers, or target builders.
  if (!is.list(plot.data) || is.null(plot.data$cd1) ||
      !(inherits(plot.data$cd1, "condensity") || inherits(plot.data$cd1, "condistribution")))
    stop("renderer prototype requires plot-data with a conditional density/distribution 'cd1' element", call. = FALSE)
  cd <- plot.data$cd1
  value.name <- if (inherits(cd, "condistribution")) "condist" else "condens"
  value.label <- if (identical(value.name, "condist")) "Conditional distribution" else "Conditional density"
  x <- unique(as.numeric(cd$xeval))
  y <- unique(as.numeric(cd$yeval))
  z <- matrix(as.numeric(cd[[value.name]]), nrow = length(x), ncol = length(y), byrow = FALSE)
  if (length(x) * length(y) != length(cd[[value.name]]) ||
      any(!is.finite(x)) || any(!is.finite(y)) || any(!is.finite(z))) {
    stop("renderer prototype requires a finite rectangular conditional surface", call. = FALSE)
  }

  xlab <- if (is.null(xlab)) cd$xnames[1L] else xlab
  ylab <- if (is.null(ylab)) cd$ynames[1L] else ylab
  zlab <- if (is.null(zlab)) value.label else zlab
  main <- if (is.null(main)) value.label else main

  if (isTRUE(perspective)) {
    graphics::persp(
      x = x,
      y = y,
      z = z,
      theta = theta,
      phi = phi,
      xlab = xlab,
      ylab = ylab,
      zlab = zlab,
      main = main,
      col = col,
      ...
    )
  } else {
    graphics::image(
      x = x,
      y = y,
      z = z,
      xlab = xlab,
      ylab = ylab,
      main = main,
      col = grDevices::hcl.colors(64L, "YlOrRd", rev = TRUE),
      ...
    )
  }
  invisible(plot.data)
}

.np_plot_proto_check_npreg_fixed_surface <- function(bws,
                                                     xdat,
                                                     ydat,
                                                     neval,
                                                     xtrim) {
  if (!inherits(bws, "rbandwidth"))
    stop("prototype route requires a regression bandwidth object", call. = FALSE)
  regtype.engine <- if (is.null(bws$regtype.engine)) {
    if (is.null(bws$regtype)) "lc" else as.character(bws$regtype)
  } else {
    as.character(bws$regtype.engine)
  }
  if (!is.element(regtype.engine, c("lc", "lp")))
    stop("prototype route currently supports regtype='lc', 'll', or 'lp' only", call. = FALSE)
  if (!identical(as.character(bws$type), "fixed"))
    stop("prototype route currently supports fixed bandwidths only", call. = FALSE)
  if (bws$ndim != 2L)
    stop("prototype route currently supports two-dimensional regression surfaces only", call. = FALSE)
  if (bws$nuno != 0L)
    stop("prototype route currently supports continuous/ordered surface variables only", call. = FALSE)
  if (bws$ncon + bws$nord != 2L)
    stop("prototype route requires a two-dimensional regression surface", call. = FALSE)
  if (!is.numeric(neval) || length(neval) != 1L || is.na(neval) || neval < 2L)
    stop("prototype route requires scalar neval >= 2", call. = FALSE)
  invisible(TRUE)
}

.np_plot_proto_clean_regression_data <- function(xdat, ydat) {
  xdat <- toFrame(xdat)
  ydat <- as.vector(ydat)
  keep.rows <- rep_len(TRUE, nrow(xdat))
  rows.omit <- attr(na.omit(data.frame(xdat, ydat)), "na.action")
  if (length(rows.omit) > 0L)
    keep.rows[as.integer(rows.omit)] <- FALSE
  if (!any(keep.rows))
    stop("Data has no rows without NAs")
  list(
    xdat = xdat[keep.rows, , drop = FALSE],
    ydat = ydat[keep.rows]
  )
}

.np_plot_proto_regression_surface_grid <- function(bws,
                                                   xdat,
                                                   neval,
                                                   xtrim) {
  xtrim <- double(bws$ndim) + xtrim

  if (is.ordered(xdat[, 1L])) {
    x1.eval <- bws$xdati$all.ulev[[1L]]
    x1.neval <- length(x1.eval)
  } else {
    x1.neval <- as.integer(neval)
    qi <- trim.quantiles(xdat[, 1L], xtrim[1L])
    x1.eval <- seq(qi[1L], qi[2L], length.out = x1.neval)
  }

  if (is.ordered(xdat[, 2L])) {
    x2.eval <- bws$xdati$all.ulev[[2L]]
    x2.neval <- length(x2.eval)
  } else {
    x2.neval <- as.integer(neval)
    qi <- trim.quantiles(xdat[, 2L], xtrim[2L])
    x2.eval <- seq(qi[1L], qi[2L], length.out = x2.neval)
  }

  x.eval <- expand.grid(x1.eval, x2.eval)
  if (is.ordered(xdat[, 1L]))
    x1.eval <- bws$xdati$all.dlev[[1L]][as.integer(x1.eval)]
  if (is.ordered(xdat[, 2L]))
    x2.eval <- bws$xdati$all.dlev[[2L]][as.integer(x2.eval)]

  list(
    x.eval = x.eval,
    x1.eval = x1.eval,
    x2.eval = x2.eval,
    x1.neval = x1.neval,
    x2.neval = x2.neval
  )
}

.np_plot_proto_npreg_fixed_data <- function(bws,
                                            xdat,
                                            ydat,
                                            neval = 50,
                                            xtrim = 0.0,
                                            plot.errors.method = c("none", "asymptotic", "bootstrap"),
                                            plot.errors.boot.method = c("wild", "inid", "fixed", "geom"),
                                            plot.errors.boot.nonfixed = c("exact", "frozen"),
                                            plot.errors.boot.wild = c("rademacher", "mammen"),
                                            plot.errors.boot.blocklen = NULL,
                                            plot.errors.boot.num = 399L,
                                            plot.errors.center = c("estimate", "bias-corrected"),
                                            plot.errors.type = c("pmzsd", "pointwise", "bonferroni",
                                                                 "simultaneous", "all"),
                                            plot.errors.alpha = 0.05,
                                            return.stages = FALSE) {
  if (missing(xdat) || missing(ydat))
    stop("prototype route requires explicit xdat and ydat", call. = FALSE)
  plot.errors.method <- match.arg(plot.errors.method)
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  plot.errors.boot.nonfixed <- match.arg(plot.errors.boot.nonfixed)
  plot.errors.boot.wild <- match.arg(plot.errors.boot.wild)
  plot.errors.center <- match.arg(plot.errors.center)
  plot.errors.type <- match.arg(plot.errors.type)

  dat <- .np_plot_proto_clean_regression_data(xdat = xdat, ydat = ydat)
  xdat <- dat$xdat
  ydat <- dat$ydat
  .np_plot_proto_check_npreg_fixed_surface(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim
  )
  grid <- .np_plot_proto_regression_surface_grid(
    bws = bws,
    xdat = xdat,
    neval = neval,
    xtrim = xtrim
  )
  fit <- .np_plot_regression_eval(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    exdat = grid$x.eval,
    gradients = FALSE,
    need.asymptotic = identical(plot.errors.method, "asymptotic")
  )

  terr <- matrix(data = fit$merr, nrow = nrow(grid$x.eval), ncol = 3L)
  terr[, 3L] <- NA
  interval <- NULL
  bootstrap <- NULL
  treg <- matrix(data = fit$mean,
                 nrow = grid$x1.neval,
                 ncol = grid$x2.neval,
                 byrow = FALSE)
  if (identical(plot.errors.method, "bootstrap")) {
    bootstrap <- compute.bootstrap.errors(
      xdat = xdat,
      ydat = ydat,
      exdat = grid$x.eval,
      gradients = FALSE,
      gradient.order = 1L,
      slice.index = 0L,
      plot.errors.boot.method = plot.errors.boot.method,
      plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
      plot.errors.boot.wild = plot.errors.boot.wild,
      plot.errors.boot.blocklen = plot.errors.boot.blocklen,
      plot.errors.boot.num = plot.errors.boot.num,
      plot.errors.center = plot.errors.center,
      plot.errors.type = plot.errors.type,
      plot.errors.alpha = plot.errors.alpha,
      progress.target = NULL,
      bws = bws
    )
    terr <- bootstrap$boot.err
    interval <- list(
      err = bootstrap$boot.err[, 1:2, drop = FALSE],
      all.err = bootstrap$boot.all.err
    )
  } else if (identical(plot.errors.method, "asymptotic")) {
    interval <- .np_plot_asymptotic_error_from_se(
      se = fit$merr,
      alpha = plot.errors.alpha,
      band.type = plot.errors.type,
      m = nrow(grid$x.eval)
    )
    terr[, 1:2] <- interval$err
  }

  r1 <- npregression(
    bws = bws,
    eval = grid$x.eval,
    mean = as.double(treg),
    merr = terr[, 1:2, drop = FALSE],
    ntrain = nrow(xdat)
  )
  r1$bias <- NA
  if (identical(plot.errors.center, "bias-corrected"))
    r1$bias <- terr[, 3L] - treg

  plot.data <- list(r1 = r1)
  if (!isTRUE(return.stages))
    return(plot.data)

  list(
    state = list(
      bws = bws,
      xdat = xdat,
      ydat = ydat,
      ntrain = nrow(xdat),
      family = "npreg",
      gradients = FALSE
    ),
    target_grid = grid,
    evaluator = fit,
    intervals = if (is.null(interval)) NULL else list(
      method = plot.errors.method,
      type = plot.errors.type,
      alpha = plot.errors.alpha,
      err = interval$err,
      all.err = interval$all.err
    ),
    bootstrap = if (is.null(bootstrap)) NULL else list(
      method = plot.errors.boot.method,
      nonfixed = plot.errors.boot.nonfixed,
      wild = plot.errors.boot.wild,
      blocklen = plot.errors.boot.blocklen,
      B = plot.errors.boot.num,
      center = plot.errors.center,
      boot.err = bootstrap$boot.err,
      boot.all.err = bootstrap$boot.all.err,
      bxp = bootstrap$bxp
    ),
    plot_data = plot.data
  )
}

.np_plot_proto_npreg_fixed_none_data <- function(bws,
                                                 xdat,
                                                 ydat,
                                                 neval = 50,
                                                 xtrim = 0.0,
                                                 return.stages = FALSE) {
  .np_plot_proto_npreg_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    plot.errors.method = "none",
    return.stages = return.stages
  )
}

.np_plot_proto_npreg_fixed_asymptotic_data <- function(bws,
                                                       xdat,
                                                       ydat,
                                                       neval = 50,
                                                       xtrim = 0.0,
                                                       plot.errors.type = c("pmzsd", "pointwise",
                                                                            "bonferroni", "simultaneous",
                                                                            "all"),
                                                       plot.errors.alpha = 0.05,
                                                       return.stages = FALSE) {
  .np_plot_proto_npreg_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    plot.errors.method = "asymptotic",
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    return.stages = return.stages
  )
}

.np_plot_proto_npreg_fixed_bootstrap_data <- function(bws,
                                                     xdat,
                                                     ydat,
                                                     neval = 50,
                                                     xtrim = 0.0,
                                                     plot.errors.boot.method = c("wild", "inid", "fixed", "geom"),
                                                     plot.errors.boot.nonfixed = c("exact", "frozen"),
                                                     plot.errors.boot.wild = c("rademacher", "mammen"),
                                                     plot.errors.boot.blocklen = NULL,
                                                     plot.errors.boot.num = 399L,
                                                     plot.errors.center = c("estimate", "bias-corrected"),
                                                     plot.errors.type = c("pmzsd", "pointwise", "bonferroni",
                                                                          "simultaneous", "all"),
                                                     plot.errors.alpha = 0.05,
                                                     return.stages = FALSE) {
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  plot.errors.boot.nonfixed <- match.arg(plot.errors.boot.nonfixed)
  plot.errors.boot.wild <- match.arg(plot.errors.boot.wild)
  plot.errors.center <- match.arg(plot.errors.center)
  plot.errors.type <- match.arg(plot.errors.type)
  .np_plot_proto_npreg_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim,
    plot.errors.method = "bootstrap",
    plot.errors.boot.method = plot.errors.boot.method,
    plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
    plot.errors.boot.wild = plot.errors.boot.wild,
    plot.errors.boot.blocklen = plot.errors.boot.blocklen,
    plot.errors.boot.num = plot.errors.boot.num,
    plot.errors.center = plot.errors.center,
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    return.stages = return.stages
  )
}

.np_plot_proto_check_npindex_fixed_curve <- function(bws,
                                                    xdat,
                                                    ydat,
                                                    neval) {
  if (!inherits(bws, "sibandwidth"))
    stop("prototype route requires a single-index bandwidth object", call. = FALSE)
  regtype.engine <- if (is.null(bws$regtype.engine)) {
    if (is.null(bws$regtype)) "lc" else as.character(bws$regtype)
  } else {
    as.character(bws$regtype.engine)
  }
  if (!is.element(regtype.engine, c("lc", "lp")))
    stop("prototype route currently supports regtype='lc', 'll', or 'lp' only", call. = FALSE)
  if (!identical(as.character(bws$type), "fixed"))
    stop("prototype route currently supports fixed bandwidths only", call. = FALSE)
  if (bws$nuno != 0L)
    stop("prototype route currently supports continuous/ordered index variables only", call. = FALSE)
  if (bws$ncon + bws$nord < 1L)
    stop("prototype route requires at least one continuous/ordered index variable", call. = FALSE)
  if (!is.numeric(neval) || length(neval) != 1L || is.na(neval) || neval < 2L)
    stop("prototype route requires scalar neval >= 2", call. = FALSE)
  invisible(TRUE)
}

.np_plot_proto_npindex_payload <- function(bws,
                                           index,
                                           mean,
                                           ntrain,
                                           trainiseval = FALSE,
                                           merr = NA,
                                           bias = NULL) {
  payload <- singleindex(
    bws = bws,
    index = index,
    mean = mean,
    merr = merr,
    ntrain = ntrain,
    trainiseval = trainiseval,
    gradients = FALSE
  )
  if (!is.null(bias))
    payload$bias <- bias
  payload
}

.np_plot_proto_npindex_fixed_data <- function(bws,
                                             xdat,
                                             ydat,
                                             neval = 50,
                                             plot.errors.method = c("none", "asymptotic", "bootstrap"),
                                             plot.errors.boot.method = c("wild", "inid", "fixed", "geom"),
                                             plot.errors.boot.nonfixed = c("exact", "frozen"),
                                             plot.errors.boot.wild = c("rademacher", "mammen"),
                                             plot.errors.boot.blocklen = NULL,
                                             plot.errors.boot.num = 399L,
                                             plot.errors.center = c("estimate", "bias-corrected"),
                                             plot.errors.type = c("pmzsd", "pointwise", "bonferroni",
                                                                  "simultaneous", "all"),
                                             plot.errors.alpha = 0.05,
                                             return.stages = FALSE) {
  if (missing(xdat) || missing(ydat))
    stop("prototype route requires explicit xdat and ydat", call. = FALSE)
  plot.errors.method <- match.arg(plot.errors.method)
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  plot.errors.boot.nonfixed <- match.arg(plot.errors.boot.nonfixed)
  plot.errors.boot.wild <- match.arg(plot.errors.boot.wild)
  plot.errors.center <- match.arg(plot.errors.center)
  plot.errors.type <- match.arg(plot.errors.type)

  dat <- .np_plot_proto_clean_regression_data(xdat = xdat, ydat = ydat)
  xdat <- dat$xdat
  ydat <- dat$ydat
  .np_plot_proto_check_npindex_fixed_curve(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval
  )
  eval.info <- .np_plot_singleindex_eval_grid(
    bws = bws,
    xdat = xdat,
    neval = neval,
    trim = 0.0,
    where = "plot.sibandwidth"
  )
  maxneval <- nrow(eval.info$idx.eval)
  fit <- if (identical(plot.errors.method, "asymptotic")) {
    .np_plot_singleindex_asymptotic_eval(
      bws = bws,
      txdat = xdat,
      tydat = ydat,
      gradients = FALSE,
      index.eval = eval.info$index.eval
    )
  } else {
    .np_plot_singleindex_local_eval(
      bws = bws,
      idx.train = eval.info$idx.train,
      idx.eval = eval.info$idx.eval,
      ydat = ydat,
      gradients = FALSE
    )
  }

  temp.err <- matrix(data = NA_real_, nrow = maxneval, ncol = 3L)
  temp.mean <- as.vector(fit$mean)
  temp.all.err <- NULL
  interval <- NULL
  bootstrap <- NULL
  if (identical(plot.errors.method, "bootstrap")) {
    bootstrap <- compute.bootstrap.errors(
      xdat = xdat,
      ydat = ydat,
      gradients = FALSE,
      plot.errors.boot.method = plot.errors.boot.method,
      plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
      plot.errors.boot.wild = plot.errors.boot.wild,
      plot.errors.boot.blocklen = plot.errors.boot.blocklen,
      plot.errors.boot.num = plot.errors.boot.num,
      plot.errors.center = plot.errors.center,
      plot.errors.type = plot.errors.type,
      plot.errors.alpha = plot.errors.alpha,
      idx.eval = eval.info$idx.eval,
      bws = bws
    )
    temp.err[,] <- bootstrap$boot.err
    temp.all.err <- bootstrap$boot.all.err
    interval <- list(
      err = bootstrap$boot.err[, 1:2, drop = FALSE],
      all.err = bootstrap$boot.all.err
    )
  } else if (identical(plot.errors.method, "asymptotic")) {
    interval <- .np_plot_asymptotic_error_from_se(
      se = fit$merr,
      alpha = plot.errors.alpha,
      band.type = plot.errors.type,
      m = maxneval
    )
    temp.err[, 1:2] <- interval$err
    temp.all.err <- interval$all.err
  }

  merr <- NA
  bias <- NULL
  if (!identical(plot.errors.method, "none")) {
    merr <- cbind(-temp.err[, 1L], temp.err[, 2L])
    if (identical(plot.errors.center, "bias-corrected"))
      bias <- temp.err[, 3L] - temp.mean
  }
  plot.data <- list(si1 = .np_plot_proto_npindex_payload(
    bws = bws,
    index = fit$index,
    mean = fit$mean,
    ntrain = nrow(xdat),
    trainiseval = eval.info$trainiseval,
    merr = merr,
    bias = bias
  ))
  if (!isTRUE(return.stages))
    return(plot.data)

  list(
    state = list(
      bws = bws,
      xdat = xdat,
      ydat = ydat,
      ntrain = nrow(xdat),
      family = "npindex",
      gradients = FALSE
    ),
    target_grid = eval.info,
    evaluator = fit,
    intervals = if (is.null(interval)) NULL else list(
      method = plot.errors.method,
      type = plot.errors.type,
      alpha = plot.errors.alpha,
      err = interval$err,
      all.err = temp.all.err
    ),
    bootstrap = if (is.null(bootstrap)) NULL else list(
      method = plot.errors.boot.method,
      nonfixed = plot.errors.boot.nonfixed,
      wild = plot.errors.boot.wild,
      blocklen = plot.errors.boot.blocklen,
      B = plot.errors.boot.num,
      center = plot.errors.center,
      boot.err = bootstrap$boot.err,
      boot.all.err = bootstrap$boot.all.err,
      bxp = bootstrap$bxp
    ),
    plot_data = plot.data
  )
}

.np_plot_proto_npindex_fixed_none_data <- function(bws,
                                                  xdat,
                                                  ydat,
                                                  neval = 50,
                                                  return.stages = FALSE) {
  .np_plot_proto_npindex_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    plot.errors.method = "none",
    return.stages = return.stages
  )
}

.np_plot_proto_npindex_fixed_asymptotic_data <- function(bws,
                                                        xdat,
                                                        ydat,
                                                        neval = 50,
                                                        plot.errors.type = c("pmzsd", "pointwise",
                                                                             "bonferroni", "simultaneous",
                                                                             "all"),
                                                        plot.errors.alpha = 0.05,
                                                        return.stages = FALSE) {
  .np_plot_proto_npindex_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    plot.errors.method = "asymptotic",
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    return.stages = return.stages
  )
}

.np_plot_proto_npindex_fixed_bootstrap_data <- function(bws,
                                                       xdat,
                                                       ydat,
                                                       neval = 50,
                                                       plot.errors.boot.method = c("wild", "inid", "fixed", "geom"),
                                                       plot.errors.boot.nonfixed = c("exact", "frozen"),
                                                       plot.errors.boot.wild = c("rademacher", "mammen"),
                                                       plot.errors.boot.blocklen = NULL,
                                                       plot.errors.boot.num = 399L,
                                                       plot.errors.center = c("estimate", "bias-corrected"),
                                                       plot.errors.type = c("pmzsd", "pointwise", "bonferroni",
                                                                            "simultaneous", "all"),
                                                       plot.errors.alpha = 0.05,
                                                       return.stages = FALSE) {
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  plot.errors.boot.nonfixed <- match.arg(plot.errors.boot.nonfixed)
  plot.errors.boot.wild <- match.arg(plot.errors.boot.wild)
  plot.errors.center <- match.arg(plot.errors.center)
  plot.errors.type <- match.arg(plot.errors.type)
  .np_plot_proto_npindex_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    plot.errors.method = "bootstrap",
    plot.errors.boot.method = plot.errors.boot.method,
    plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
    plot.errors.boot.wild = plot.errors.boot.wild,
    plot.errors.boot.blocklen = plot.errors.boot.blocklen,
    plot.errors.boot.num = plot.errors.boot.num,
    plot.errors.center = plot.errors.center,
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    return.stages = return.stages
  )
}

.np_plot_proto_check_npscoef_fixed_surface <- function(bws,
                                                       xdat,
                                                       ydat,
                                                       zdat,
                                                       neval,
                                                       xtrim,
                                                       ztrim) {
  if (!inherits(bws, "scbandwidth"))
    stop("prototype route requires a smooth-coefficient bandwidth object", call. = FALSE)
  regtype <- if (is.null(bws$regtype)) "lc" else as.character(bws$regtype)
  if (!is.element(regtype, c("lc", "ll", "lp")))
    stop("prototype route currently supports regtype='lc', 'll', or 'lp' only", call. = FALSE)
  if (!identical(as.character(bws$type), "fixed"))
    stop("prototype route currently supports fixed bandwidths only", call. = FALSE)
  if (ncol(xdat) != 1L || ncol(zdat) != 1L)
    stop("prototype route currently supports one x variable and one z variable", call. = FALSE)
  if (bws$xdati$iuno[1L] || bws$zdati$iuno[1L])
    stop("prototype route currently supports continuous/ordered surface variables only", call. = FALSE)
  if (!is.numeric(neval) || length(neval) != 1L || is.na(neval) || neval < 2L)
    stop("prototype route requires scalar neval >= 2", call. = FALSE)
  invisible(TRUE)
}

.np_plot_proto_clean_scoef_data <- function(xdat, ydat, zdat) {
  xdat <- toFrame(xdat)
  zdat <- toFrame(zdat)
  ydat <- as.vector(ydat)
  keep.rows <- rep_len(TRUE, nrow(xdat))
  rows.omit <- attr(na.omit(data.frame(xdat, ydat, zdat)), "na.action")
  if (length(rows.omit) > 0L)
    keep.rows[as.integer(rows.omit)] <- FALSE
  if (!any(keep.rows))
    stop("Data has no rows without NAs")
  list(
    xdat = xdat[keep.rows, , drop = FALSE],
    ydat = ydat[keep.rows],
    zdat = zdat[keep.rows, , drop = FALSE]
  )
}

.np_plot_proto_scoef_surface_grid <- function(bws,
                                              xdat,
                                              zdat,
                                              neval,
                                              xtrim,
                                              ztrim) {
  xtrim <- double(ncol(xdat)) + xtrim
  ztrim <- double(ncol(zdat)) + ztrim

  if (is.ordered(xdat[, 1L])) {
    x1.eval <- bws$xdati$all.ulev[[1L]]
    x1.neval <- length(x1.eval)
  } else {
    x1.neval <- as.integer(neval)
    qi <- trim.quantiles(xdat[, 1L], xtrim[1L])
    x1.eval <- seq(qi[1L], qi[2L], length.out = x1.neval)
  }

  if (is.ordered(zdat[, 1L])) {
    x2.eval <- bws$zdati$all.ulev[[1L]]
    x2.neval <- length(x2.eval)
  } else {
    x2.neval <- as.integer(neval)
    qi <- trim.quantiles(zdat[, 1L], ztrim[1L])
    x2.eval <- seq(qi[1L], qi[2L], length.out = x2.neval)
  }

  x.eval <- expand.grid(x1.eval, x2.eval)
  colnames(x.eval) <- c(colnames(xdat)[1L], colnames(zdat)[1L])
  if (is.ordered(xdat[, 1L]))
    x1.eval <- bws$xdati$all.dlev[[1L]][as.integer(x1.eval)]
  if (is.ordered(zdat[, 1L]))
    x2.eval <- bws$zdati$all.dlev[[1L]][as.integer(x2.eval)]

  list(
    x.eval = x.eval,
    exdat = x.eval[, 1L, drop = FALSE],
    ezdat = x.eval[, 2L, drop = FALSE],
    x1.eval = x1.eval,
    x2.eval = x2.eval,
    x1.neval = x1.neval,
    x2.neval = x2.neval
  )
}

.np_plot_proto_npscoef_fixed_data <- function(bws,
                                             xdat,
                                             ydat,
                                             zdat,
                                             neval = 50,
                                             xtrim = 0.0,
                                             ztrim = 0.0,
                                             plot.errors.method = c("none", "asymptotic", "bootstrap"),
                                             plot.errors.boot.method = c("wild", "inid", "fixed", "geom"),
                                             plot.errors.boot.nonfixed = c("exact", "frozen"),
                                             plot.errors.boot.wild = c("rademacher", "mammen"),
                                             plot.errors.boot.blocklen = NULL,
                                             plot.errors.boot.num = 399L,
                                             plot.errors.center = c("estimate", "bias-corrected"),
                                             plot.errors.type = c("pmzsd", "pointwise", "bonferroni",
                                                                  "simultaneous", "all"),
                                             plot.errors.alpha = 0.05,
                                             return.stages = FALSE) {
  if (missing(xdat) || missing(ydat) || missing(zdat))
    stop("prototype route requires explicit xdat, ydat, and zdat", call. = FALSE)
  plot.errors.method <- match.arg(plot.errors.method)
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  plot.errors.boot.nonfixed <- match.arg(plot.errors.boot.nonfixed)
  plot.errors.boot.wild <- match.arg(plot.errors.boot.wild)
  plot.errors.center <- match.arg(plot.errors.center)
  plot.errors.type <- match.arg(plot.errors.type)

  dat <- .np_plot_proto_clean_scoef_data(xdat = xdat, ydat = ydat, zdat = zdat)
  xdat <- dat$xdat
  ydat <- dat$ydat
  zdat <- dat$zdat
  .np_plot_proto_check_npscoef_fixed_surface(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    neval = neval,
    xtrim = xtrim,
    ztrim = ztrim
  )
  grid <- .np_plot_proto_scoef_surface_grid(
    bws = bws,
    xdat = xdat,
    zdat = zdat,
    neval = neval,
    xtrim = xtrim,
    ztrim = ztrim
  )
  fit <- .np_scoef_fit_internal(
    bws = bws,
    txdat = xdat,
    tydat = ydat,
    tzdat = zdat,
    exdat = grid$exdat,
    ezdat = grid$ezdat,
    iterate = FALSE,
    errors = identical(plot.errors.method, "asymptotic"),
    betas = FALSE
  )

  treg <- matrix(data = fit$mean,
                 nrow = grid$x1.neval,
                 ncol = grid$x2.neval,
                 byrow = FALSE)
  terr <- matrix(data = fit$merr, nrow = nrow(grid$x.eval), ncol = 3L)
  terr[, 3L] <- NA_real_
  interval <- NULL
  bootstrap <- NULL
  if (identical(plot.errors.method, "bootstrap")) {
    bootstrap <- compute.bootstrap.errors(
      xdat = xdat,
      ydat = ydat,
      zdat = zdat,
      exdat = grid$exdat,
      ezdat = grid$ezdat,
      gradients = FALSE,
      slice.index = 0L,
      progress.target = NULL,
      plot.errors.boot.method = plot.errors.boot.method,
      plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
      plot.errors.boot.wild = plot.errors.boot.wild,
      plot.errors.boot.blocklen = plot.errors.boot.blocklen,
      plot.errors.boot.num = plot.errors.boot.num,
      plot.errors.center = plot.errors.center,
      plot.errors.type = plot.errors.type,
      plot.errors.alpha = plot.errors.alpha,
      bws = bws
    )
    terr <- bootstrap$boot.err
    interval <- list(
      err = bootstrap$boot.err[, 1:2, drop = FALSE],
      all.err = bootstrap$boot.all.err
    )
  } else if (identical(plot.errors.method, "asymptotic")) {
    interval <- .np_plot_asymptotic_error_from_se(
      se = fit$merr,
      alpha = plot.errors.alpha,
      band.type = plot.errors.type,
      m = nrow(grid$x.eval)
    )
    terr[, 1:2] <- interval$err
  }

  r1 <- smoothcoefficient(
    bws = bws,
    eval = list(exdat = grid$exdat, ezdat = grid$ezdat),
    mean = as.double(treg),
    merr = if (identical(plot.errors.method, "none")) {
      matrix(NA, nrow = nrow(grid$x.eval), ncol = 2L)
    } else {
      terr[, 1:2, drop = FALSE]
    },
    ntrain = nrow(xdat)
  )
  r1$bias <- NA
  if (identical(plot.errors.center, "bias-corrected"))
    r1$bias <- terr[, 3L] - treg

  plot.data <- list(r1 = r1)
  if (!isTRUE(return.stages))
    return(plot.data)

  list(
    state = list(
      bws = bws,
      xdat = xdat,
      ydat = ydat,
      zdat = zdat,
      ntrain = nrow(xdat),
      family = "npscoef",
      gradients = FALSE,
      coef = FALSE
    ),
    target_grid = grid,
    evaluator = fit,
    intervals = if (is.null(interval)) NULL else list(
      method = plot.errors.method,
      type = plot.errors.type,
      alpha = plot.errors.alpha,
      err = interval$err,
      all.err = interval$all.err
    ),
    bootstrap = if (is.null(bootstrap)) NULL else list(
      method = plot.errors.boot.method,
      nonfixed = plot.errors.boot.nonfixed,
      wild = plot.errors.boot.wild,
      blocklen = plot.errors.boot.blocklen,
      B = plot.errors.boot.num,
      center = plot.errors.center,
      boot.err = bootstrap$boot.err,
      boot.all.err = bootstrap$boot.all.err,
      bxp = bootstrap$bxp
    ),
    plot_data = plot.data
  )
}

.np_plot_proto_npscoef_fixed_none_data <- function(bws,
                                                  xdat,
                                                  ydat,
                                                  zdat,
                                                  neval = 50,
                                                  xtrim = 0.0,
                                                  ztrim = 0.0,
                                                  return.stages = FALSE) {
  .np_plot_proto_npscoef_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    neval = neval,
    xtrim = xtrim,
    ztrim = ztrim,
    plot.errors.method = "none",
    return.stages = return.stages
  )
}

.np_plot_proto_npscoef_fixed_asymptotic_data <- function(bws,
                                                        xdat,
                                                        ydat,
                                                        zdat,
                                                        neval = 50,
                                                        xtrim = 0.0,
                                                        ztrim = 0.0,
                                                        plot.errors.type = c("pmzsd", "pointwise",
                                                                             "bonferroni", "simultaneous",
                                                                             "all"),
                                                        plot.errors.alpha = 0.05,
                                                        return.stages = FALSE) {
  .np_plot_proto_npscoef_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    neval = neval,
    xtrim = xtrim,
    ztrim = ztrim,
    plot.errors.method = "asymptotic",
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    return.stages = return.stages
  )
}

.np_plot_proto_npscoef_fixed_bootstrap_data <- function(bws,
                                                       xdat,
                                                       ydat,
                                                       zdat,
                                                       neval = 50,
                                                       xtrim = 0.0,
                                                       ztrim = 0.0,
                                                       plot.errors.boot.method = c("wild", "inid", "fixed", "geom"),
                                                       plot.errors.boot.nonfixed = c("exact", "frozen"),
                                                       plot.errors.boot.wild = c("rademacher", "mammen"),
                                                       plot.errors.boot.blocklen = NULL,
                                                       plot.errors.boot.num = 399L,
                                                       plot.errors.center = c("estimate", "bias-corrected"),
                                                       plot.errors.type = c("pmzsd", "pointwise", "bonferroni",
                                                                            "simultaneous", "all"),
                                                       plot.errors.alpha = 0.05,
                                                       return.stages = FALSE) {
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  plot.errors.boot.nonfixed <- match.arg(plot.errors.boot.nonfixed)
  plot.errors.boot.wild <- match.arg(plot.errors.boot.wild)
  plot.errors.center <- match.arg(plot.errors.center)
  plot.errors.type <- match.arg(plot.errors.type)
  .np_plot_proto_npscoef_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    neval = neval,
    xtrim = xtrim,
    ztrim = ztrim,
    plot.errors.method = "bootstrap",
    plot.errors.boot.method = plot.errors.boot.method,
    plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
    plot.errors.boot.wild = plot.errors.boot.wild,
    plot.errors.boot.blocklen = plot.errors.boot.blocklen,
    plot.errors.boot.num = plot.errors.boot.num,
    plot.errors.center = plot.errors.center,
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    return.stages = return.stages
  )
}

.np_plot_proto_check_npplreg_fixed_surface <- function(bws,
                                                       xdat,
                                                       ydat,
                                                       zdat,
                                                       neval,
                                                       xtrim,
                                                       ztrim) {
  if (!inherits(bws, "plbandwidth"))
    stop("prototype route requires a partially linear bandwidth object", call. = FALSE)
  regtype <- if (is.null(bws$regtype)) "lc" else as.character(bws$regtype)
  if (!is.element(regtype, c("lc", "ll", "lp")))
    stop("prototype route currently supports regtype='lc', 'll', or 'lp' only", call. = FALSE)
  if (!identical(as.character(bws$type), "fixed"))
    stop("prototype route currently supports fixed bandwidths only", call. = FALSE)
  if (ncol(xdat) != 1L || ncol(zdat) != 1L)
    stop("prototype route currently supports one x variable and one z variable", call. = FALSE)
  if (bws$xdati$iuno[1L] || bws$zdati$iuno[1L])
    stop("prototype route currently supports continuous/ordered surface variables only", call. = FALSE)
  if (!is.numeric(neval) || length(neval) != 1L || is.na(neval) || neval < 2L)
    stop("prototype route requires scalar neval >= 2", call. = FALSE)
  invisible(TRUE)
}

.np_plot_proto_npplreg_fixed_data <- function(bws,
                                             xdat,
                                             ydat,
                                             zdat,
                                             neval = 50,
                                             xtrim = 0.0,
                                             ztrim = 0.0,
                                             plot.errors.method = c("none", "asymptotic", "bootstrap"),
                                             plot.errors.boot.method = c("wild", "inid", "fixed", "geom"),
                                             plot.errors.boot.nonfixed = c("exact", "frozen"),
                                             plot.errors.boot.wild = c("rademacher", "mammen"),
                                             plot.errors.boot.blocklen = NULL,
                                             plot.errors.boot.num = 399L,
                                             plot.errors.center = c("estimate", "bias-corrected"),
                                             plot.errors.type = c("pmzsd", "pointwise", "bonferroni",
                                                                  "simultaneous", "all"),
                                             plot.errors.alpha = 0.05,
                                             return.stages = FALSE) {
  if (missing(xdat) || missing(ydat) || missing(zdat))
    stop("prototype route requires explicit xdat, ydat, and zdat", call. = FALSE)
  plot.errors.method <- match.arg(plot.errors.method)
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  plot.errors.boot.nonfixed <- match.arg(plot.errors.boot.nonfixed)
  plot.errors.boot.wild <- match.arg(plot.errors.boot.wild)
  plot.errors.center <- match.arg(plot.errors.center)
  plot.errors.type <- match.arg(plot.errors.type)

  dat <- .np_plot_proto_clean_scoef_data(xdat = xdat, ydat = ydat, zdat = zdat)
  xdat <- dat$xdat
  ydat <- dat$ydat
  zdat <- dat$zdat
  .np_plot_proto_check_npplreg_fixed_surface(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    neval = neval,
    xtrim = xtrim,
    ztrim = ztrim
  )
  grid <- .np_plot_proto_scoef_surface_grid(
    bws = bws,
    xdat = xdat,
    zdat = zdat,
    neval = neval,
    xtrim = xtrim,
    ztrim = ztrim
  )
  fit <- if (identical(plot.errors.method, "asymptotic")) {
    .np_plot_plreg_asymptotic_fit(
      bws = bws,
      xdat = xdat,
      ydat = ydat,
      zdat = zdat,
      exdat = grid$exdat,
      ezdat = grid$ezdat
    )
  } else {
    .np_plot_plreg_local_fit(
      bws = bws,
      xdat = xdat,
      ydat = ydat,
      zdat = zdat,
      exdat = grid$exdat,
      ezdat = grid$ezdat
    )
  }

  treg <- matrix(data = fit$mean,
                 nrow = grid$x1.neval,
                 ncol = grid$x2.neval,
                 byrow = FALSE)
  terr <- matrix(data = NA, nrow = nrow(grid$x.eval), ncol = 3L)
  interval <- NULL
  bootstrap <- NULL
  if (identical(plot.errors.method, "bootstrap")) {
    bootstrap <- compute.bootstrap.errors(
      xdat = xdat,
      ydat = ydat,
      zdat = zdat,
      exdat = grid$exdat,
      ezdat = grid$ezdat,
      gradients = FALSE,
      slice.index = 0L,
      progress.target = NULL,
      plot.errors.boot.method = plot.errors.boot.method,
      plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
      plot.errors.boot.wild = plot.errors.boot.wild,
      plot.errors.boot.blocklen = plot.errors.boot.blocklen,
      plot.errors.boot.num = plot.errors.boot.num,
      plot.errors.center = plot.errors.center,
      plot.errors.type = plot.errors.type,
      plot.errors.alpha = plot.errors.alpha,
      bws = bws
    )
    terr <- bootstrap$boot.err
    interval <- list(
      err = bootstrap$boot.err[, 1:2, drop = FALSE],
      all.err = bootstrap$boot.all.err
    )
  } else if (identical(plot.errors.method, "asymptotic")) {
    interval <- .np_plot_asymptotic_error_from_se(
      se = fit$merr,
      alpha = plot.errors.alpha,
      band.type = plot.errors.type,
      m = nrow(grid$x.eval)
    )
    terr[, 1:2] <- interval$err
  }

  r1 <- plregression(
    bws = bws,
    xcoef = fit$xcoef,
    xcoefvcov = vcov(fit),
    xcoeferr = fit$xcoeferr,
    evalx = grid$exdat[, 1L],
    evalz = grid$ezdat[, 1L],
    mean = fit$mean,
    ntrain = nrow(xdat),
    trainiseval = FALSE,
    xtra = c(fit$RSQ, fit$MSE, 0, 0, 0, 0)
  )
  r1$merr <- NA
  r1$bias <- NA
  if (!identical(plot.errors.method, "none"))
    r1$merr <- terr[, 1:2, drop = FALSE]
  if (identical(plot.errors.center, "bias-corrected"))
    r1$bias <- terr[, 3L] - treg

  plot.data <- list(r1 = r1)
  if (!isTRUE(return.stages))
    return(plot.data)

  list(
    state = list(
      bws = bws,
      xdat = xdat,
      ydat = ydat,
      zdat = zdat,
      ntrain = nrow(xdat),
      family = "npplreg",
      gradients = FALSE,
      coef = FALSE
    ),
    target_grid = grid,
    evaluator = fit,
    intervals = if (is.null(interval)) NULL else list(
      method = plot.errors.method,
      type = plot.errors.type,
      alpha = plot.errors.alpha,
      err = interval$err,
      all.err = interval$all.err
    ),
    bootstrap = if (is.null(bootstrap)) NULL else list(
      method = plot.errors.boot.method,
      nonfixed = plot.errors.boot.nonfixed,
      wild = plot.errors.boot.wild,
      blocklen = plot.errors.boot.blocklen,
      B = plot.errors.boot.num,
      center = plot.errors.center,
      boot.err = bootstrap$boot.err,
      boot.all.err = bootstrap$boot.all.err,
      bxp = bootstrap$bxp
    ),
    plot_data = plot.data
  )
}

.np_plot_proto_npplreg_fixed_none_data <- function(bws,
                                                  xdat,
                                                  ydat,
                                                  zdat,
                                                  neval = 50,
                                                  xtrim = 0.0,
                                                  ztrim = 0.0,
                                                  return.stages = FALSE) {
  .np_plot_proto_npplreg_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    neval = neval,
    xtrim = xtrim,
    ztrim = ztrim,
    plot.errors.method = "none",
    return.stages = return.stages
  )
}

.np_plot_proto_npplreg_fixed_asymptotic_data <- function(bws,
                                                        xdat,
                                                        ydat,
                                                        zdat,
                                                        neval = 50,
                                                        xtrim = 0.0,
                                                        ztrim = 0.0,
                                                        plot.errors.type = c("pmzsd", "pointwise",
                                                                             "bonferroni", "simultaneous",
                                                                             "all"),
                                                        plot.errors.alpha = 0.05,
                                                        return.stages = FALSE) {
  .np_plot_proto_npplreg_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    neval = neval,
    xtrim = xtrim,
    ztrim = ztrim,
    plot.errors.method = "asymptotic",
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    return.stages = return.stages
  )
}

.np_plot_proto_npplreg_fixed_bootstrap_data <- function(bws,
                                                       xdat,
                                                       ydat,
                                                       zdat,
                                                       neval = 50,
                                                       xtrim = 0.0,
                                                       ztrim = 0.0,
                                                       plot.errors.boot.method = c("wild", "inid", "fixed", "geom"),
                                                       plot.errors.boot.nonfixed = c("exact", "frozen"),
                                                       plot.errors.boot.wild = c("rademacher", "mammen"),
                                                       plot.errors.boot.blocklen = NULL,
                                                       plot.errors.boot.num = 399L,
                                                       plot.errors.center = c("estimate", "bias-corrected"),
                                                       plot.errors.type = c("pmzsd", "pointwise", "bonferroni",
                                                                            "simultaneous", "all"),
                                                       plot.errors.alpha = 0.05,
                                                       return.stages = FALSE) {
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  plot.errors.boot.nonfixed <- match.arg(plot.errors.boot.nonfixed)
  plot.errors.boot.wild <- match.arg(plot.errors.boot.wild)
  plot.errors.center <- match.arg(plot.errors.center)
  plot.errors.type <- match.arg(plot.errors.type)
  .np_plot_proto_npplreg_fixed_data(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    neval = neval,
    xtrim = xtrim,
    ztrim = ztrim,
    plot.errors.method = "bootstrap",
    plot.errors.boot.method = plot.errors.boot.method,
    plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
    plot.errors.boot.wild = plot.errors.boot.wild,
    plot.errors.boot.blocklen = plot.errors.boot.blocklen,
    plot.errors.boot.num = plot.errors.boot.num,
    plot.errors.center = plot.errors.center,
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    return.stages = return.stages
  )
}

.np_plot_proto_check_npqreg_fixed_slices <- function(object,
                                                     xdat,
                                                     ydat,
                                                     neval,
                                                     xtrim) {
  if (!inherits(object, "qregression"))
    stop("prototype route requires a quantile regression object", call. = FALSE)
  bws <- object$bws
  if (!(inherits(bws, "condbandwidth") || inherits(bws, "conbandwidth")))
    stop("prototype route requires a conditional distribution bandwidth object", call. = FALSE)
  if (!identical(as.character(bws$type), "fixed"))
    stop("prototype route currently supports fixed bandwidths only", call. = FALSE)
  if (bws$xnuno != 0L)
    stop("prototype route currently supports continuous/ordered slice variables only", call. = FALSE)
  if (bws$xndim < 1L || bws$yndim != 1L)
    stop("prototype route requires at least one x variable and one y variable", call. = FALSE)
  if (!is.numeric(neval) || length(neval) != 1L || is.na(neval) || neval < 2L)
    stop("prototype route requires scalar neval >= 2", call. = FALSE)
  invisible(TRUE)
}

.np_plot_proto_npqreg_slice_targets <- function(bws,
                                                xdat,
                                                ydat,
                                                neval,
                                                xq,
                                                yq,
                                                xtrim) {
  xq <- double(bws$xndim) + xq
  xtrim <- double(bws$xndim) + xtrim
  x.ev <- xdat[1L, , drop = FALSE]
  y.ev <- ydat[1L, , drop = FALSE]
  for (i in seq_len(bws$xndim))
    x.ev[1L, i] <- uocquantile(xdat[, i], prob = xq[i])
  for (i in seq_len(bws$yndim))
    y.ev[1L, i] <- uocquantile(ydat[, i], prob = yq)

  maxneval <- max(c(sapply(xdat, nlevels), neval))
  exdat.base <- xdat[rep(1L, maxneval), , drop = FALSE]
  eydat.base <- as.data.frame(matrix(data = 0, nrow = maxneval, ncol = bws$yndim))
  colnames(eydat.base) <- colnames(ydat)
  for (i in seq_len(bws$xndim))
    exdat.base[, i] <- x.ev[1L, i]
  for (i in seq_len(bws$yndim))
    eydat.base[, i] <- y.ev[1L, i]

  slices <- vector("list", bws$xndim)
  for (i in seq_len(bws$xndim)) {
    if (is.factor(xdat[, i])) {
      ei <- levels(xdat[, i])
      ei <- factor(ei, levels = ei)
      xi.neval <- length(ei)
    } else {
      xi.neval <- as.integer(neval)
      qi <- trim.quantiles(xdat[, i], xtrim[i])
      ei <- seq(qi[1L], qi[2L], length.out = neval)
    }
    exdat.i <- subcol(exdat.base, ei, i)[seq_len(xi.neval), , drop = FALSE]
    eydat.i <- eydat.base[seq_len(xi.neval), , drop = FALSE]
    slices[[i]] <- list(
      index = i,
      eval = ei,
      neval = xi.neval,
      exdat = exdat.i,
      eydat = eydat.i
    )
  }

  list(
    base_x = x.ev,
    base_y = y.ev,
    maxneval = maxneval,
    slices = slices
  )
}

.np_plot_proto_npqreg_fixed_slices_data <- function(object,
                                                   xdat,
                                                   ydat,
                                                   neval = 50,
                                                   xq = 0.5,
                                                   yq = 0.5,
                                                   xtrim = 0.0,
                                                   plot.errors.method = c("none", "asymptotic", "bootstrap"),
                                                   plot.errors.boot.method = c("wild", "inid", "fixed", "geom"),
                                                   plot.errors.boot.nonfixed = c("exact", "frozen"),
                                                   plot.errors.boot.blocklen = NULL,
                                                   plot.errors.boot.num = 399L,
                                                   plot.errors.center = c("estimate", "bias-corrected"),
                                                   plot.errors.type = c("pmzsd", "pointwise", "bonferroni",
                                                                        "simultaneous", "all"),
                                                   plot.errors.alpha = 0.05,
                                                   gradients = FALSE,
                                                   return.stages = FALSE) {
  if (missing(xdat) || missing(ydat))
    stop("prototype route requires explicit xdat and ydat", call. = FALSE)
  plot.errors.method <- match.arg(plot.errors.method)
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  plot.errors.boot.nonfixed <- match.arg(plot.errors.boot.nonfixed)
  plot.errors.center <- match.arg(plot.errors.center)
  plot.errors.type <- match.arg(plot.errors.type)
  gradients <- npValidateScalarLogical(gradients, "gradients")
  bws <- object$bws
  dat <- .np_plot_proto_clean_conditional_data(xdat = xdat, ydat = ydat)
  xdat <- dat$xdat
  ydat <- dat$ydat
  .np_plot_proto_check_npqreg_fixed_slices(
    object = object,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xtrim = xtrim
  )
  targets <- .np_plot_proto_npqreg_slice_targets(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xq = xq,
    yq = yq,
    xtrim = xtrim
  )

  plot.data <- vector("list", length(targets$slices))
  evaluators <- vector("list", length(targets$slices))
  intervals <- vector("list", length(targets$slices))
  bootstraps <- vector("list", length(targets$slices))
  for (ii in seq_along(targets$slices)) {
    target <- targets$slices[[ii]]
    fit <- .np_plot_quantile_eval(
      txdat = xdat,
      tydat = ydat,
      exdat = target$exdat,
      tau = object$tau,
      gradients = gradients,
      bws = bws
    )
    temp.err <- matrix(data = NA, nrow = targets$maxneval, ncol = 3L)
    temp.quant <- rep(NA_real_, targets$maxneval)
    temp.quant[seq_len(target$neval)] <- fit$quantile
    interval <- NULL
    bootstrap <- NULL

    if (identical(plot.errors.method, "bootstrap") && !isTRUE(gradients)) {
      bootstrap <- compute.bootstrap.errors(
        xdat = xdat,
        ydat = ydat,
        exdat = target$exdat,
        eydat = target$eydat,
        cdf = TRUE,
        quantreg = TRUE,
        tau = object$tau,
        gradients = FALSE,
        gradient.index = 0L,
        slice.index = target$index,
        plot.errors.boot.method = plot.errors.boot.method,
        plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
        plot.errors.boot.blocklen = plot.errors.boot.blocklen,
        plot.errors.boot.num = plot.errors.boot.num,
        plot.errors.center = plot.errors.center,
        plot.errors.type = plot.errors.type,
        plot.errors.alpha = plot.errors.alpha,
        progress.target = .np_plot_conditional_bootstrap_target_label(
          bws = bws,
          slice.index = target$index,
          gradients = FALSE,
          gradient.index = 0L
        ),
        bws = bws
      )
      temp.err[seq_len(target$neval), ] <- bootstrap$boot.err
      interval <- list(
        err = bootstrap$boot.err[, 1:2, drop = FALSE],
        all.err = bootstrap$boot.all.err
      )
    } else if (identical(plot.errors.method, "bootstrap") && isTRUE(gradients)) {
      bootstrap <- vector("list", bws$xndim)
      interval <- vector("list", bws$xndim)
      for (jj in seq_len(bws$xndim)) {
        temp.err.j <- matrix(data = NA, nrow = targets$maxneval, ncol = 3L)
        bootstrap.j <- compute.bootstrap.errors(
          xdat = xdat,
          ydat = ydat,
          exdat = target$exdat,
          eydat = target$eydat,
          cdf = TRUE,
          quantreg = TRUE,
          tau = object$tau,
          gradients = TRUE,
          gradient.index = jj,
          slice.index = target$index,
          plot.errors.boot.method = plot.errors.boot.method,
          plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
          plot.errors.boot.blocklen = plot.errors.boot.blocklen,
          plot.errors.boot.num = plot.errors.boot.num,
          plot.errors.center = plot.errors.center,
          plot.errors.type = plot.errors.type,
          plot.errors.alpha = plot.errors.alpha,
          progress.target = .np_plot_conditional_bootstrap_target_label(
            bws = bws,
            slice.index = target$index,
            gradients = TRUE,
            gradient.index = jj
          ),
          bws = bws
        )
        temp.err.j[seq_len(target$neval), ] <- bootstrap.j$boot.err
        bootstrap[[jj]] <- bootstrap.j
        interval[[jj]] <- list(
          err = bootstrap.j$boot.err[, 1:2, drop = FALSE],
          all.err = bootstrap.j$boot.all.err
        )
      }
    } else if (identical(plot.errors.method, "asymptotic")) {
      asym <- .np_plot_asymptotic_error_from_se(
        se = fit$quanterr,
        alpha = plot.errors.alpha,
        band.type = plot.errors.type,
        m = target$neval
      )
      temp.err[seq_len(target$neval), 1:2] <- asym$err
      interval <- list(err = asym$err, all.err = asym$all.err)
    }

    payload <- fit
    if (!identical(plot.errors.method, "none") && !isTRUE(gradients)) {
      payload$quanterr <- na.omit(cbind(
        -temp.err[seq_len(target$neval), 1L],
        temp.err[seq_len(target$neval), 2L]
      ))
      payload$bias <- na.omit(
        temp.quant[seq_len(target$neval)] - temp.err[seq_len(target$neval), 3L]
      )
      payload$bxp <- if (is.null(bootstrap)) list() else bootstrap$bxp
    } else if (identical(plot.errors.method, "bootstrap") && isTRUE(gradients)) {
      for (jj in seq_len(bws$xndim)) {
        temp.err.j <- matrix(data = NA, nrow = targets$maxneval, ncol = 3L)
        temp.err.j[seq_len(target$neval), ] <- bootstrap[[jj]]$boot.err
        temp.grad.j <- rep(NA_real_, targets$maxneval)
        temp.grad.j[seq_len(target$neval)] <- fit$quantgrad[, jj]
        payload[[paste0("gc", jj, "err")]] <- na.omit(cbind(
          -temp.err.j[, 1L],
          temp.err.j[, 2L]
        ))
        payload[[paste0("gc", jj, "bias")]] <- na.omit(
          temp.grad.j - temp.err.j[, 3L]
        )
        payload$bxp <- bootstrap[[jj]]$bxp
      }
    }

    plot.data[[ii]] <- payload
    evaluators[[ii]] <- fit
    intervals[[ii]] <- interval
    bootstraps[[ii]] <- bootstrap
  }
  names(plot.data) <- paste0("cd", seq_along(plot.data))

  if (!isTRUE(return.stages))
    return(plot.data)

  list(
    state = list(
      bws = bws,
      xdat = xdat,
      ydat = ydat,
      ntrain = nrow(xdat),
      family = "npqreg",
      tau = object$tau,
      gradients = gradients
    ),
    target_grid = targets,
    evaluator = evaluators,
    intervals = intervals,
    bootstrap = bootstraps,
    plot_data = plot.data
  )
}

.np_plot_proto_npqreg_fixed_none_data <- function(object,
                                                 xdat,
                                                 ydat,
                                                 neval = 50,
                                                 xq = 0.5,
                                                 yq = 0.5,
                                                 xtrim = 0.0,
                                                 gradients = FALSE,
                                                 return.stages = FALSE) {
  .np_plot_proto_npqreg_fixed_slices_data(
    object = object,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xq = xq,
    yq = yq,
    xtrim = xtrim,
    plot.errors.method = "none",
    gradients = gradients,
    return.stages = return.stages
  )
}

.np_plot_proto_npqreg_fixed_asymptotic_data <- function(object,
                                                       xdat,
                                                       ydat,
                                                       neval = 50,
                                                       xq = 0.5,
                                                       yq = 0.5,
                                                       xtrim = 0.0,
                                                       plot.errors.type = c("pmzsd", "pointwise",
                                                                            "bonferroni", "simultaneous",
                                                                            "all"),
                                                       plot.errors.alpha = 0.05,
                                                       gradients = FALSE,
                                                       return.stages = FALSE) {
  .np_plot_proto_npqreg_fixed_slices_data(
    object = object,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xq = xq,
    yq = yq,
    xtrim = xtrim,
    plot.errors.method = "asymptotic",
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    gradients = gradients,
    return.stages = return.stages
  )
}

.np_plot_proto_npqreg_fixed_bootstrap_data <- function(object,
                                                      xdat,
                                                      ydat,
                                                      neval = 50,
                                                      xq = 0.5,
                                                      yq = 0.5,
                                                      xtrim = 0.0,
                                                      plot.errors.boot.method = c("wild", "inid", "fixed", "geom"),
                                                      plot.errors.boot.nonfixed = c("exact", "frozen"),
                                                      plot.errors.boot.blocklen = NULL,
                                                      plot.errors.boot.num = 399L,
                                                      plot.errors.center = c("estimate", "bias-corrected"),
                                                      plot.errors.type = c("pmzsd", "pointwise", "bonferroni",
                                                                           "simultaneous", "all"),
                                                      plot.errors.alpha = 0.05,
                                                      gradients = FALSE,
                                                      return.stages = FALSE) {
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  plot.errors.boot.nonfixed <- match.arg(plot.errors.boot.nonfixed)
  plot.errors.center <- match.arg(plot.errors.center)
  plot.errors.type <- match.arg(plot.errors.type)
  .np_plot_proto_npqreg_fixed_slices_data(
    object = object,
    xdat = xdat,
    ydat = ydat,
    neval = neval,
    xq = xq,
    yq = yq,
    xtrim = xtrim,
    plot.errors.method = "bootstrap",
    plot.errors.boot.method = plot.errors.boot.method,
    plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
    plot.errors.boot.blocklen = plot.errors.boot.blocklen,
    plot.errors.boot.num = plot.errors.boot.num,
    plot.errors.center = plot.errors.center,
    plot.errors.type = plot.errors.type,
    plot.errors.alpha = plot.errors.alpha,
    gradients = gradients,
    return.stages = return.stages
  )
}

.np_plot_proto_check_unconditional_fixed_surface <- function(bws,
                                                            xdat,
                                                            neval,
                                                            xtrim,
                                                            cdf = FALSE) {
  expected.class <- if (isTRUE(cdf)) "dbandwidth" else "bandwidth"
  expected.label <- if (isTRUE(cdf)) "unconditional distribution" else "unconditional density"
  if (!inherits(bws, expected.class))
    stop(sprintf("prototype route requires an %s bandwidth object", expected.label), call. = FALSE)
  if (!identical(as.character(bws$type), "fixed"))
    stop("prototype route currently supports fixed bandwidths only", call. = FALSE)
  if (bws$ndim != 2L)
    stop(sprintf("prototype route currently supports two-dimensional %s surfaces", expected.label), call. = FALSE)
  if (bws$nuno != 0L)
    stop("prototype route currently supports continuous/ordered surface variables only", call. = FALSE)
  if (bws$ncon + bws$nord != 2L)
    stop(sprintf("prototype route requires a two-dimensional %s surface", expected.label), call. = FALSE)
  if (!is.numeric(neval) || length(neval) != 1L || is.na(neval) || neval < 2L)
    stop("prototype route requires scalar neval >= 2", call. = FALSE)
  invisible(TRUE)
}

.np_plot_proto_clean_unconditional_data <- function(xdat) {
  xdat <- toFrame(xdat)
  keep.rows <- rep_len(TRUE, nrow(xdat))
  rows.omit <- attr(na.omit(xdat), "na.action")
  if (length(rows.omit) > 0L)
    keep.rows[as.integer(rows.omit)] <- FALSE
  if (!any(keep.rows))
    stop("Data has no rows without NAs")
  list(xdat = xdat[keep.rows, , drop = FALSE])
}

.np_plot_proto_unconditional_fixed_data <- function(bws,
                                                   xdat,
                                                   neval = 50,
                                                   xtrim = 0.0,
                                                   plot.errors.method = c("none", "asymptotic", "bootstrap"),
                                                   plot.errors.boot.method = c("inid", "fixed", "geom"),
                                                   plot.errors.boot.nonfixed = c("exact", "frozen"),
                                                   plot.errors.boot.blocklen = NULL,
                                                   plot.errors.boot.num = 1999,
                                                   plot.errors.center = c("estimate", "bias-corrected"),
                                                   plot.errors.type = c("pmzsd", "pointwise", "bonferroni",
                                                                        "simultaneous", "all"),
                                                   plot.errors.alpha = 0.05,
                                                   return.stages = FALSE,
                                                   cdf = FALSE) {
  if (missing(xdat))
    stop("prototype route requires explicit xdat", call. = FALSE)
  plot.errors.method <- match.arg(plot.errors.method)
  plot.errors.boot.method <- match.arg(plot.errors.boot.method)
  plot.errors.boot.nonfixed <- match.arg(plot.errors.boot.nonfixed)
  plot.errors.center <- match.arg(plot.errors.center)
  plot.errors.type <- match.arg(plot.errors.type)

  dat <- .np_plot_proto_clean_unconditional_data(xdat = xdat)
  xdat <- dat$xdat
  .np_plot_proto_check_unconditional_fixed_surface(
    bws = bws,
    xdat = xdat,
    neval = neval,
    xtrim = xtrim,
    cdf = cdf
  )
  grid <- .np_plot_proto_regression_surface_grid(
    bws = bws,
    xdat = xdat,
    neval = neval,
    xtrim = xtrim
  )
  fit <- .np_plot_unconditional_eval(
    xdat = xdat,
    exdat = grid$x.eval,
    bws = bws,
    cdf = isTRUE(cdf),
    need.asymptotic = identical(plot.errors.method, "asymptotic")
  )

  estimate <- if (isTRUE(cdf)) fit$dist else fit$dens
  se <- if (isTRUE(cdf)) fit$derr else fit$derr
  terr <- matrix(data = se, nrow = length(estimate), ncol = 3L)
  terr[, 3L] <- NA_real_
  interval <- NULL
  bootstrap <- NULL
  if (identical(plot.errors.method, "bootstrap")) {
    bootstrap.args <- list(
      xdat = xdat,
      exdat = grid$x.eval,
      slice.index = 0L,
      plot.errors.boot.method = plot.errors.boot.method,
      plot.errors.boot.nonfixed = plot.errors.boot.nonfixed,
      plot.errors.boot.blocklen = plot.errors.boot.blocklen,
      plot.errors.boot.num = plot.errors.boot.num,
      plot.errors.center = plot.errors.center,
      plot.errors.type = plot.errors.type,
      plot.errors.alpha = plot.errors.alpha,
      bws = bws
    )
    if (!isTRUE(cdf))
      bootstrap.args$cdf <- FALSE
    bootstrap <- do.call(compute.bootstrap.errors, bootstrap.args)
    terr <- bootstrap$boot.err
    interval <- list(
      err = bootstrap$boot.err[, 1:2, drop = FALSE],
      all.err = bootstrap$boot.all.err
    )
  } else if (identical(plot.errors.method, "asymptotic")) {
    interval <- .np_plot_asymptotic_error_from_se(
      se = se,
      alpha = plot.errors.alpha,
      band.type = plot.errors.type,
      m = nrow(grid$x.eval)
    )
    terr[, 1:2] <- interval$err
  }

  d1 <- if (isTRUE(cdf)) {
    npdistribution(
      bws = bws,
      eval = grid$x.eval,
      dist = estimate,
      derr = terr[, 1:2, drop = FALSE],
      ntrain = nrow(xdat)
    )
  } else {
    npdensity(
      bws = bws,
      eval = grid$x.eval,
      dens = estimate,
      derr = terr[, 1:2, drop = FALSE],
      ntrain = nrow(xdat)
    )
  }
  d1$bias <- NA
  if (identical(plot.errors.center, "bias-corrected"))
    d1$bias <- terr[, 3L] - estimate

  plot.data <- list(d1 = d1)
  if (!isTRUE(return.stages))
    return(plot.data)

  list(
    state = list(
      bws = bws,
      xdat = xdat,
      ntrain = nrow(xdat),
      family = if (isTRUE(cdf)) "npudist" else "npudens",
      cdf = isTRUE(cdf)
    ),
    target_grid = grid,
    evaluator = fit,
    intervals = if (is.null(interval)) NULL else list(
      method = plot.errors.method,
      type = plot.errors.type,
      alpha = plot.errors.alpha,
      err = interval$err,
      all.err = interval$all.err
    ),
    bootstrap = if (is.null(bootstrap)) NULL else list(
      method = plot.errors.boot.method,
      nonfixed = plot.errors.boot.nonfixed,
      blocklen = plot.errors.boot.blocklen,
      B = plot.errors.boot.num,
      center = plot.errors.center,
      boot.err = bootstrap$boot.err,
      boot.all.err = bootstrap$boot.all.err,
      bxp = bootstrap$bxp
    ),
    plot_data = plot.data
  )
}

.np_plot_proto_npudens_fixed_none_data <- function(bws,
                                                   xdat,
                                                   neval = 50,
                                                   xtrim = 0.0,
                                                   return.stages = FALSE) {
  .np_plot_proto_unconditional_fixed_data(
    bws = bws,
    xdat = xdat,
    neval = neval,
    xtrim = xtrim,
    plot.errors.method = "none",
    return.stages = return.stages
  )
}

.np_plot_proto_npudist_fixed_none_data <- function(bws,
                                                   xdat,
                                                   neval = 50,
                                                   xtrim = 0.0,
                                                   return.stages = FALSE) {
  .np_plot_proto_unconditional_fixed_data(
    bws = bws,
    xdat = xdat,
    neval = neval,
    xtrim = xtrim,
    plot.errors.method = "none",
    return.stages = return.stages,
    cdf = TRUE
  )
}

.np_plot_proto_rectangular_surface_base_render <- function(plot.data,
                                                          perspective = TRUE,
                                                          main = NULL,
                                                          xlab = NULL,
                                                          ylab = NULL,
                                                          zlab = NULL,
                                                          col = "lightblue",
                                                          theta = 0.0,
                                                          phi = 20.0,
                                                          ...) {
  if (!is.list(plot.data))
    stop("renderer prototype requires a plot-data list", call. = FALSE)

  obj <- NULL
  value.name <- NULL
  value.label <- NULL
  if (!is.null(plot.data$r1) && inherits(plot.data$r1, "npregression")) {
    obj <- plot.data$r1
    value.name <- "mean"
    value.label <- "Conditional mean"
  } else if (!is.null(plot.data$d1) && inherits(plot.data$d1, "npdensity")) {
    obj <- plot.data$d1
    value.name <- "dens"
    value.label <- "Density"
  } else if (!is.null(plot.data$d1) && inherits(plot.data$d1, "npdistribution")) {
    obj <- plot.data$d1
    value.name <- "dist"
    value.label <- "Distribution"
  }
  if (is.null(obj))
    stop("renderer prototype requires npregression, npdensity, or npdistribution plot data", call. = FALSE)

  xeval <- obj$eval
  if (!is.data.frame(xeval) || ncol(xeval) != 2L)
    stop("renderer prototype requires a two-column rectangular evaluation grid", call. = FALSE)
  x <- unique(as.numeric(xeval[, 1L]))
  y <- unique(as.numeric(xeval[, 2L]))
  z <- matrix(as.numeric(obj[[value.name]]), nrow = length(x), ncol = length(y), byrow = FALSE)
  if (length(x) * length(y) != length(obj[[value.name]]) ||
      any(!is.finite(x)) || any(!is.finite(y)) || any(!is.finite(z))) {
    stop("renderer prototype requires a finite rectangular surface", call. = FALSE)
  }

  xlab <- if (is.null(xlab)) names(xeval)[1L] else xlab
  ylab <- if (is.null(ylab)) names(xeval)[2L] else ylab
  zlab <- if (is.null(zlab)) value.label else zlab
  main <- if (is.null(main)) value.label else main

  if (isTRUE(perspective)) {
    graphics::persp(
      x = x,
      y = y,
      z = z,
      theta = theta,
      phi = phi,
      xlab = xlab,
      ylab = ylab,
      zlab = zlab,
      main = main,
      col = col,
      ...
    )
  } else {
    graphics::image(
      x = x,
      y = y,
      z = z,
      xlab = xlab,
      ylab = ylab,
      main = main,
      col = grDevices::hcl.colors(64L, "YlOrRd", rev = TRUE),
      ...
    )
  }
  invisible(plot.data)
}

## 

## Original code in Matlab by A. Patton, R translation and
## modifications by C. Parmeter and J. Racine.
##
## We are grateful to Andrew Patton and Dimitris Politis for their
## assistance and feedback. Kindly report features, deficiencies, and
## improvements to racinej@mcmaster.ca.
##
## The citation is A. Patton, D.N. Politis, and H. White (2008,
## forthcoming), "CORRECTION TO `Automatic Block-Length Selection for
## the Dependent Bootstrap' by D.N. Politis and H. White". This is
## based on the article by Politis, D.N., and H. White (2004),
## "Automatic block-length selection for the dependent bootstrap."
## Econometric Reviews, vol. 23.
##
## INPUTS:  data, an n x k matrix.
##
## OUTPUTS: b.star, a 2 x k vector of optimal bootstrap block lengths
## for the stationary bootstrap and circular bootstrap (BstarSB,
## BstarCB).

## The function lam() is used to construct a "flat-top" lag window for
## spectral estimation based on Politis, D.N. and J.P. Romano (1995),
## "Bias-Corrected Nonparametric Spectral Estimation", Journal of Time
## Series Analysis, vol. 16, No. 1.

lam <- function(s){
  return((abs(s)>=0)*(abs(s)<0.5)+2*(1-abs(s))*(abs(s)>=0.5)*(abs(s)<=1))
}

## The function b.star() returns the optimal bootstrap block
## lengths. Note that an example for usage appears at the bottom of
## this file. If you use this function as input into a routine such as
## tsboot() in the boot library (Angelo Canty and Brian Ripley
## (2008). boot: Bootstrap R (S-Plus) Functions. R package version
## 1.2-34.) you ought to use the option round=TRUE.

b.star <- function(data,
                   Kn = NULL,
                   mmax= NULL,
                   Bmax = NULL,
                   c = NULL,
                   round = FALSE){

  ## Convert the data object to a data frame to handle both vectors
  ## and matrices.

  data <- data.frame(data)
  n <- nrow(data)
  k <- ncol(data)

  ## Set Defaults. Note that in footnote c, page 59, for Kn Politis
  ## and White (2004) use max(5,log10(n)). Since this must be an
  ## integer we use ceiling(log10(n)).

  if (is.null(Kn)) Kn <- max(5,ceiling(log10(n)))
  if (is.null(mmax)) mmax <- ceiling(sqrt(n))+Kn
  if (is.null(Bmax)) Bmax <- ceiling(min(3*sqrt(n),n/3))
  if (is.null(c)) c <- qnorm(0.975)

  ## Create two vectors of length k in which we store results.

  BstarSB <- numeric(length=k)
  BstarCB <- numeric(length=k)

  ## Now we loop through each variable in data (i.e., column,
  ## data[,i]).

  for(i in seq_len(k)) {

    ## We first obtain the autocorrelations rho(1),...,rho(mmax) (we
    ## need to drop the first autocorrelation as it is rho(0), hence
    ## acf[-1]). This is the default in acf [type="correlation"]. Note
    ## that Patton uses sample correlations after dropping the first
    ## mmax observations, while we instead use the acf to obtain
    ## rho(k).

    rho.k <- acf(data[,i],
                 lag.max = mmax,
                 type = "correlation",
                 plot = FALSE)$acf[-1]

    ## Next we compute mhat. The use of c*sqrt(log10(n)/n) for
    ## critical values is given in footnote c of Politis and White
    ## (2004, page 59), and the approach for determining mhat is
    ## described in footnote c.

    rho.k.crit <- c*sqrt(log10(n)/n)

    ## Compute the number of insignificant runs following each rho(k),
    ## k=1,...,mmax.
    
    num.insignificant <- sapply(seq_len(max(mmax - Kn + 1L, 0L)),
                                function(j){
                                  sum((abs(rho.k) < rho.k.crit)[j:(j+Kn-1)])
                                })

    ## If there are any values of rho(k) for which the Kn proceeding
    ## values of rho(k+j), j=1,...,Kn are all insignificant, take the
    ## smallest rho(k) such that this holds (see footnote c for
    ## further details).

    if(any(num.insignificant==Kn)) {
      mhat <- which(num.insignificant==Kn)[1]
    } else {

      ## If no runs of length Kn are insignificant, take the smallest
      ## value of rho(k) that is significant.

      if(any(abs(rho.k) > rho.k.crit)) {
        
        lag.sig <- which(abs(rho.k) > rho.k.crit)
        k.sig <- length(lag.sig)
        
        if(k.sig == 1) {
          
          ## When only one lag is significant, mhat is the sole
          ## significant rho(k).
          
          mhat <- lag.sig
          
        } else {

          ## If there are more than one significant lags but no runs
          ## of length Kn, take the largest value of rho(k) that is
          ## significant.
          
          mhat <- max(lag.sig)
          
        }
        
      } else {
        
        ## When there are no significant lags, mhat must be the
        ## smallest positive integer (footnote c), hence mhat is set
        ## to one.
        
        mhat <- 1
        
      }

    }

    ## Compute M (mhat is at least one).

    M <- min(2 * mhat, mmax)

    ## We compute BstarSB and BstarCB using the formulas in the above
    ## references. Now we require the autocovariance R(k) (hence
    ## type="covariance" in the acf call). Note that Patton uses
    ## sample covariances after dropping the first mmax observations,
    ## while we instead use the acf with type="covariance" to obtain
    ## R(k). Note also that we require R(0) hence we do not drop it as
    ## we did for rho(k) via acf(...)$acf[-1].
      
    kk <- seq(-M,M)

    R.k <- ccf(data[,i], data[,i],
               lag.max = M,
               type = "covariance",
               plot = FALSE)$acf
    
    Ghat <- sum(lam(kk/M)*abs(kk)*R.k)
    DCBhat <- 4/3*sum(lam(kk/M)*R.k)^2
    DSBhat <- 2*sum(lam(kk/M)*R.k)^2
    BstarSB[i] <- ((2*Ghat^2)/DSBhat)^(1/3)*n^(1/3)
    BstarCB[i] <- ((2*(Ghat^2)/DCBhat)^(1/3))*(n^(1/3))
      
  }

  ## The user can choose whether they want rounded values returned or
  ## not. BstarCB is rounded up, BstarSB simply rounded but both must
  ## be positive integers.

  if(round == FALSE) {
    
    BstarSB <- pmin(BstarSB, Bmax)
    BstarCB <- pmin(BstarCB, Bmax)

  } else {

    BstarSB <- pmax(1, pmin(Bmax, round(BstarSB)))
    BstarCB <- pmax(1, pmin(Bmax, round(BstarCB)))

  }
  
  return(cbind(BstarSB,BstarCB)) 
  
}

## Here is a simple example with an n x 2 matrix containing n=10^5
## observations, where column 2 of x is more persistent than column
## 1. This requires that you first install the forecast library (i.e.,
## install.packages("forecast")).
##
##  library(forecast)
##  set.seed(123)
##  x <- cbind(arima.sim(n = 100000, list(ar = c(.5,.0), ma = c(0,0)),sd = 1),
##             arima.sim(n = 100000, list(ar = c(.5,.4), ma = c(0,0)),sd = 1))
##  b.star(x)  
##  b.star(x,round=TRUE)
##>   b.star(x)  
##       BstarSB   BstarCB
##[1,]  50.39272  57.68526
##[2,] 251.62894 288.04323
##>   b.star(x,round=TRUE)
##     BstarSB BstarCB
##[1,]      50      58
##[2,]     252     289

symtest <- function(Srho, Srho.bootstrap, P, boot.num, data.rotate, bw) {
  
  tsym = list(Srho = Srho,
    Srho.bootstrap = Srho.bootstrap,
    P = P,
    boot.num = boot.num,
    data.rotate = data.rotate,
    bw = bw)
  
  reject <- ' '
  
  if (P < 0.1)
    reject <- '.'

  if (P < 0.05)
    reject <- '*'

  if (P < 0.01)
    reject <- '**'

  if (P < 0.001)
    reject <- '***'

  tsym$reject <- reject
  tsym$rejectNum <- switch(reject,
                           ' ' = 100,
                           '.' = 10,
                           '*' = 5,
                           '**' = 1,
                           '***' = 0.1)
  
  class(tsym) <- "symtest"

  tsym
}

print.symtest <- function(x, ...){
  reject.msg <- if (x$reject == ' ') {
    "\nFail to reject the null of symmetry at the 10% level"
  } else {
    paste("\nNull of symmetry is rejected at the ", x$rejectNum, "% level", sep = "")
  }
  cat("\nConsistent Entropy Asymmetry Test",
      paste("\n", format(x$boot.num), " Bootstrap Replications",sep=""),
      "\n\nTest Statistic ", sQuote("Srho"), ": ",
      format(x$Srho), "\tP Value: ", format.pval(x$P)," ", x$reject,
      "\n---\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1",
      reject.msg,
      "\n\n", sep="")
}

summary.symtest <- function(object, ...){
  print(object)
}

npudisthat <- function(bws,
                       tdat = stop("training data 'tdat' missing"),
                       edat,
                       y = NULL,
                       output = c("matrix", "apply")) {
  output <- match.arg(output)

  if (!inherits(bws, "dbandwidth")) {
    stop("argument 'bws' must inherit from class 'dbandwidth' in npudisthat()")
  }

  tdat <- toFrame(tdat)
  no.e <- missing(edat)

  if (!no.e) {
    edat <- toFrame(edat)
    if (!(tdat %~% edat))
      stop("'tdat' and 'edat' are not similar data frames!")
  }

  if (!is.null(y)) {
    if (is.factor(y) || is.vector(y)) {
      y <- matrix(as.double(y), ncol = 1L)
    } else {
      y <- as.matrix(y)
      storage.mode(y) <- "double"
    }

    if (nrow(y) != nrow(tdat))
      stop("number of rows in 'y' must match the number of training rows in 'tdat'")
  }

  keep.rows <- rep_len(TRUE, nrow(tdat))
  rows.omit.train <- attr(stats::na.omit(tdat), "na.action")
  if (!is.null(y))
    rows.omit.train <- union(rows.omit.train, attr(stats::na.omit(as.data.frame(y)), "na.action"))

  if (length(rows.omit.train) > 0L)
    keep.rows[as.integer(rows.omit.train)] <- FALSE

  if (!any(keep.rows))
    stop("Training data has no rows without NAs")

  tdat <- tdat[keep.rows, , drop = FALSE]
  if (!is.null(y))
    y <- y[keep.rows, , drop = FALSE]

  if (!no.e) {
    keep.eval <- rep_len(TRUE, nrow(edat))
    rows.omit.eval <- attr(stats::na.omit(edat), "na.action")
    if (length(rows.omit.eval) > 0L)
      keep.eval[as.integer(rows.omit.eval)] <- FALSE
    if (!any(keep.eval))
      stop("Evaluation data has no rows without NAs")
    edat <- edat[keep.eval, , drop = FALSE]
  }

  n.train <- nrow(tdat)
  target.dist <- fitted(npudist(
    bws = bws,
    tdat = tdat,
    edat = if (no.e) tdat else edat
  ))

  scale_rows <- function(x, probe.sum) {
    x <- as.matrix(x)
    probe.sum <- as.vector(probe.sum)
    safe.sum <- probe.sum
    safe.sum[abs(safe.sum) < .Machine$double.xmin] <- .Machine$double.xmin
    scaled <- sweep(x, 1L, target.dist / safe.sum, "*")
    if (ncol(scaled) == 1L)
      scaled[, 1L, drop = FALSE]
    else
      scaled
  }

  if (identical(output, "apply")) {
    if (is.null(y))
      stop("argument 'y' is required when output='apply'")

    rhs <- if (ncol(y) == 1L) {
      y
    } else {
      lapply(seq_len(ncol(y)), function(j) y[, j, drop = FALSE])
    }

    apply_one <- function(rhs.col) {
      if (identical(bws$type, "fixed")) {
        out <- .np_direct_operator_apply(
          kbw = bws,
          txdat = tdat,
          exdat = if (no.e) tdat else edat,
          operator = "integral",
          rhs = rhs.col,
          where = "npudisthat direct operator apply"
        ) / n.train
        probe.sum <- .np_direct_operator_apply(
          kbw = bws,
          txdat = tdat,
          exdat = if (no.e) tdat else edat,
          operator = "integral",
          rhs = matrix(1.0, nrow = n.train, ncol = 1L),
          where = "npudisthat direct operator probe"
        ) / n.train
      } else {
        out <- .np_exact_operator_apply(
          kbw = bws,
          txdat = tdat,
          exdat = if (no.e) tdat else edat,
          operator = "integral",
          rhs = rhs.col,
          where = "npudisthat exact operator apply"
        ) / n.train
        probe.sum <- .np_exact_operator_apply(
          kbw = bws,
          txdat = tdat,
          exdat = if (no.e) tdat else edat,
          operator = "integral",
          rhs = matrix(1.0, nrow = n.train, ncol = 1L),
          where = "npudisthat exact operator probe"
        ) / n.train
      }

      scale_rows(out, probe.sum)
    }

    out <- if (is.list(rhs)) {
      do.call(cbind, lapply(rhs, apply_one))
    } else {
      apply_one(rhs)
    }

    if (ncol(out) == 1L)
      return(as.vector(out))
    return(out)
  }

  if (identical(bws$type, "fixed")) {
    H <- .np_direct_operator_matrix(
      kbw = bws,
      txdat = tdat,
      exdat = if (no.e) tdat else edat,
      operator = "integral",
      where = "npudisthat direct operator"
    ) / n.train
  } else {
    H <- .np_exact_operator_matrix(
      kbw = bws,
      txdat = tdat,
      exdat = if (no.e) tdat else edat,
      operator = "integral",
      where = "npudisthat exact operator"
    ) / n.train
  }

  H <- scale_rows(H, rowSums(H))

  class(H) <- c("npudisthat", "matrix")
  attr(H, "bws") <- bws
  attr(H, "tdat") <- tdat
  attr(H, "edat") <- if (no.e) tdat else edat
  attr(H, "trainiseval") <- no.e
  attr(H, "rows.omit.train") <- rows.omit.train
  attr(H, "call") <- match.call(expand.dots = FALSE)

  if (!is.null(y)) {
    Hy <- H %*% y
    if (ncol(Hy) == 1L)
      Hy <- as.vector(Hy)
    attr(H, "Hy") <- Hy
  }

  H
}

.np_plot_scalar_match <- function(value, choices, argname) {
  if (is.null(value))
    return(NULL)
  if (length(value) != 1L || is.na(value))
    stop(sprintf("%s must be one of %s",
                 argname,
                 paste(sprintf("\"%s\"", choices), collapse = ", ")),
         call. = FALSE)
  value <- as.character(value)
  if (!(value %in% choices))
    stop(sprintf("%s must be one of %s",
                 argname,
                 paste(sprintf("\"%s\"", choices), collapse = ", ")),
         call. = FALSE)
  value
}

np_boot_control <- function(nonfixed = c("exact", "frozen"),
                            wild = c("rademacher", "mammen"),
                            blocklen = NULL) {
  nonfixed <- match.arg(nonfixed)
  wild <- match.arg(wild)
  if (!is.null(blocklen) &&
      (!is.numeric(blocklen) || length(blocklen) != 1L ||
       is.na(blocklen) || blocklen <= 0))
    stop("blocklen must be a positive numeric scalar", call. = FALSE)
  structure(
    list(nonfixed = nonfixed, wild = wild, blocklen = blocklen),
    class = "np_boot_control"
  )
}

np_grid_control <- function(xtrim = NULL, xq = NULL, slices = NULL) {
  if (!is.null(xtrim) &&
      (!is.numeric(xtrim) || length(xtrim) != 2L ||
       any(is.na(xtrim)) || any(xtrim < 0) || any(xtrim > 1) ||
       xtrim[1L] >= xtrim[2L]))
    stop("xtrim must be a numeric length-two vector with 0 <= xtrim[1] < xtrim[2] <= 1",
         call. = FALSE)
  structure(
    list(xtrim = xtrim, xq = xq, slices = slices),
    class = "np_grid_control"
  )
}

np_render_control <- function(style = c("band", "bar"),
                              bar = c("|", "I"),
                              bar_num = NULL) {
  style <- match.arg(style)
  bar <- match.arg(bar)
  if (!is.null(bar_num) &&
      (!is.numeric(bar_num) || length(bar_num) != 1L ||
       is.na(bar_num) || bar_num < 1))
    stop("bar_num must be a positive numeric scalar", call. = FALSE)
  structure(
    list(style = style, bar = bar, bar_num = bar_num),
    class = "np_render_control"
  )
}

.np_plot_dot_names <- function(dots_call) {
  if (is.null(dots_call) || length(dots_call) == 0L)
    return(character())
  nms <- names(dots_call)
  if (is.null(nms))
    rep.int("", length(dots_call))
  else
    nms
}

.np_plot_stop_unused_args <- function(bad, allowed) {
  bad <- unique(bad[nzchar(bad)])
  if (!length(bad))
    return(invisible(NULL))
  msg <- if (length(bad) == 1L)
    sprintf("unused plot argument: %s", bad)
  else
    sprintf("unused plot arguments: %s", paste(bad, collapse = ", "))
  close <- vapply(bad, function(x) {
    d <- utils::adist(x, allowed, partial = FALSE, ignore.case = FALSE)
    if (!length(d))
      return(NA_character_)
    i <- which.min(d)
    if (is.finite(d[i]) && d[i] <= max(2L, floor(nchar(x) / 3L)))
      allowed[i]
    else
      NA_character_
  }, character(1L))
  close <- unique(stats::na.omit(close))
  if (length(close))
    msg <- paste0(msg, "; did you mean ",
                  paste(close, collapse = " or "), "?")
  stop(msg, call. = FALSE)
}

.np_plot_graphics_arg_names <- function() {
  unique(c(
    setdiff(names(formals(graphics::plot.default)), c("x", "y", "...")),
    names(graphics::par(no.readonly = TRUE)),
    "panel.first", "panel.last", "zlab", "zlim", "theta", "phi", "border",
    "view", "type", "lty", "lwd", "col", "pch", "cex", "main", "sub",
    "xlab", "ylab", "xlim", "ylim"
  ))
}

.np_plot_canonical_arg_names <- function() {
  c("errors", "band", "alpha", "bootstrap", "B", "center",
    "output", "data_overlay", "data_rug", "layout", "legend",
    "factor_boxplot", "boxplot_outliers", "coef_index",
    "gradient_order", "common_scale", "proper_method", "proper_control",
    "renderer", "neval", "perspective",
    "boot_control", "grid_control", "render_control")
}

.np_plot_legacy_arg_names <- function() {
  c("plot.errors.method", "plot.errors.type", "plot.errors.alpha",
    "plot.errors.boot.method", "plot.errors.boot.num",
    "plot.errors.boot.nonfixed", "plot.errors.boot.wild",
    "plot.errors.boot.blocklen", "plot.errors.center",
    "plot.errors.style", "plot.errors.bar", "plot.errors.bar.num",
    "plot.behavior", "gradient", "persp", "plot.par.mfrow",
    "plot.data.overlay", "plot.rug", "plot.bxp", "plot.bxp.out",
    "behavior")
}

.np_plot_engine_for_bws <- function(bws) {
  cls <- class(bws)
  if ("rbandwidth" %in% cls)
    return(.np_plot_rbandwidth_engine)
  if ("conbandwidth" %in% cls)
    return(.np_plot_conbandwidth_engine)
  if ("condbandwidth" %in% cls)
    return(.np_plot_condbandwidth_engine)
  if ("plbandwidth" %in% cls)
    return(.np_plot_plbandwidth_engine)
  if ("sibandwidth" %in% cls)
    return(.np_plot_sibandwidth_engine)
  if ("scbandwidth" %in% cls)
    return(.np_plot_scbandwidth_engine)
  if ("dbandwidth" %in% cls)
    return(.np_plot_dbandwidth_engine)
  if ("bandwidth" %in% cls)
    return(.np_plot_bandwidth_engine)
  NULL
}

.np_plot_allowed_engine_args <- function(method = NULL, bws = NULL) {
  if (!is.null(method) && identical(method, .np_plot_compat_dispatch))
    method <- .np_plot_engine_for_bws(bws)
  if (is.null(method))
    return(character())
  setdiff(names(formals(method)), c("bws", "..."))
}

.np_plot_validate_public_dots <- function(dots_call,
                                          method = NULL,
                                          bws = NULL,
                                          context = "plot") {
  dot.names <- .np_plot_dot_names(dots_call)
  if (any(!nzchar(dot.names)))
    stop(sprintf("unnamed plot arguments are not supported for %s", context),
         call. = FALSE)
  if ("intervals" %in% dot.names)
    stop("unused plot argument: intervals; did you mean errors?",
         call. = FALSE)
  if ("boot" %in% dot.names)
    stop("unused plot argument: boot; did you mean bootstrap?",
         call. = FALSE)
  if ("bands" %in% dot.names)
    stop("unused plot argument: bands; did you mean band?",
         call. = FALSE)
  canonical <- .np_plot_canonical_arg_names()
  legacy <- .np_plot_legacy_arg_names()
  engine <- .np_plot_allowed_engine_args(method = method, bws = bws)
  dispatcher <- "random.seed"
  allowed <- unique(c(canonical, legacy, dispatcher, engine,
                      .np_plot_graphics_arg_names()))
  bad <- setdiff(dot.names[nzchar(dot.names)], allowed)
  .np_plot_stop_unused_args(bad, allowed)
  invisible(NULL)
}

.np_plot_set_normalized_arg <- function(dots, public, internal, value) {
  same <- !is.null(dots[[internal]]) &&
    isTRUE(all.equal(dots[[internal]], value, check.attributes = FALSE))
  if (!is.null(dots[[internal]]) && !same)
    stop(sprintf("conflicting plot arguments: %s and %s specify different values",
                 public, internal),
         call. = FALSE)
  dots[[internal]] <- value
  dots
}

.np_plot_match_layout <- function(value) {
  if (is.logical(value)) {
    if (length(value) != 1L || is.na(value))
      stop("layout must be TRUE/FALSE or one of \"auto\", \"current\"",
           call. = FALSE)
    return(isTRUE(value))
  }
  layout <- .np_plot_scalar_match(value, c("auto", "current"), "layout")
  identical(layout, "auto")
}

.np_plot_normalize_public_dots <- function(dots, context = "plot") {
  supplied <- names(dots)
  has <- function(x) x %in% supplied

  if (has("intervals"))
    stop("unused plot argument: intervals; did you mean errors?", call. = FALSE)
  if (has("boot"))
    stop("unused plot argument: boot; did you mean bootstrap?", call. = FALSE)
  if (has("bands"))
    stop("unused plot argument: bands; did you mean band?", call. = FALSE)

  if (has("errors")) {
    errors <- .np_plot_scalar_match(dots$errors,
                                    c("none", "bootstrap", "asymptotic"),
                                    "errors")
    dots$errors <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "errors",
                                        "plot.errors.method", errors)
  }
  if (has("band")) {
    band <- .np_plot_scalar_match(dots$band,
                                  c("pmzsd", "pointwise", "bonferroni",
                                    "simultaneous", "all"),
                                  "band")
    dots$band <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "band",
                                        "plot.errors.type", band)
  }
  if (has("alpha")) {
    alpha <- dots$alpha
    if (!is.numeric(alpha) || length(alpha) != 1L ||
        is.na(alpha) || alpha <= 0 || alpha >= 0.5)
      stop("alpha must lie in (0, 0.5)", call. = FALSE)
    dots$alpha <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "alpha",
                                        "plot.errors.alpha", alpha)
  }
  if (has("bootstrap")) {
    if (is.list(dots$bootstrap))
      stop("unused plot argument: bootstrap list; use scalar bootstrap, B, and boot_control",
           call. = FALSE)
    bootstrap <- .np_plot_scalar_match(dots$bootstrap,
                                       c("wild", "inid", "fixed", "geom"),
                                       "bootstrap")
    dots$bootstrap <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "bootstrap",
                                        "plot.errors.boot.method", bootstrap)
  }
  if (has("B")) {
    B <- dots$B
    if (!is.numeric(B) || length(B) != 1L || is.na(B) || B < 1)
      stop("B must be a positive numeric scalar", call. = FALSE)
    dots$B <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "B",
                                        "plot.errors.boot.num", as.integer(B))
  }
  if (has("center")) {
    center <- .np_plot_scalar_match(dots$center,
                                    c("estimate", "bias-corrected"),
                                    "center")
    dots$center <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "center",
                                        "plot.errors.center", center)
  }
  if (has("output")) {
    output <- .np_plot_scalar_match(dots$output,
                                    c("plot", "data", "both", "plot-data"),
                                    "output")
    if (identical(output, "both"))
      output <- "plot-data"
    dots$output <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "output",
                                        "plot.behavior", output)
  }
  if (has("behavior")) {
    behavior <- .np_plot_scalar_match(dots$behavior,
                                      c("plot", "plot-data", "data"),
                                      "behavior")
    dots$behavior <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "behavior",
                                        "plot.behavior", behavior)
  }
  if (has("data_overlay")) {
    data_overlay <- .np_plot_match_flag(dots$data_overlay, "data_overlay")
    dots$data_overlay <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "data_overlay",
                                        "plot.data.overlay", data_overlay)
  }
  if (has("data_rug")) {
    data_rug <- .np_plot_match_flag(dots$data_rug, "data_rug")
    dots$data_rug <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "data_rug",
                                        "plot.rug", data_rug)
  }
  if (has("layout")) {
    layout <- .np_plot_match_layout(dots$layout)
    dots$layout <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "layout",
                                        "plot.par.mfrow", layout)
  }
  if (has("factor_boxplot")) {
    factor_boxplot <- .np_plot_match_flag(dots$factor_boxplot,
                                          "factor_boxplot")
    dots$factor_boxplot <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "factor_boxplot",
                                        "plot.bxp", factor_boxplot)
  }
  if (has("boxplot_outliers")) {
    boxplot_outliers <- .np_plot_match_flag(dots$boxplot_outliers,
                                            "boxplot_outliers")
    dots$boxplot_outliers <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "boxplot_outliers",
                                        "plot.bxp.out", boxplot_outliers)
  }
  if (has("coef_index")) {
    coef_index <- dots$coef_index
    if (!is.numeric(coef_index) || length(coef_index) != 1L ||
        is.na(coef_index) || coef_index < 1L)
      stop("coef_index must be a positive numeric scalar", call. = FALSE)
    dots$coef_index <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "coef_index",
                                        "coef.index", as.integer(coef_index))
  }
  if (has("gradient_order")) {
    gradient_order <- dots$gradient_order
    if (!is.numeric(gradient_order) || any(is.na(gradient_order)) ||
        any(gradient_order < 1L))
      stop("gradient_order must contain positive numeric values",
           call. = FALSE)
    dots$gradient_order <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "gradient_order",
                                        "gradient.order",
                                        as.integer(gradient_order))
  }
  if (has("common_scale")) {
    common_scale <- .np_plot_match_flag(dots$common_scale, "common_scale")
    dots$common_scale <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "common_scale",
                                        "common.scale", common_scale)
  }
  if (has("proper_method")) {
    proper_method <- dots$proper_method
    if (!is.character(proper_method) || length(proper_method) != 1L ||
        is.na(proper_method))
      stop("proper_method must be a character scalar", call. = FALSE)
    dots$proper_method <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "proper_method",
                                        "proper.method", proper_method)
  }
  if (has("proper_control")) {
    proper_control <- dots$proper_control
    if (!is.list(proper_control))
      stop("proper_control must be a list", call. = FALSE)
    dots$proper_control <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "proper_control",
                                        "proper.control", proper_control)
  }
  if (has("perspective")) {
    perspective <- isTRUE(dots$perspective)
    dots$perspective <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "perspective",
                                        "perspective", perspective)
  }
  if (has("boot_control")) {
    if (!inherits(dots$boot_control, "np_boot_control"))
      stop("boot_control must be created by np_boot_control()", call. = FALSE)
    ctrl <- dots$boot_control
    dots$boot_control <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "boot_control$nonfixed",
                                        "plot.errors.boot.nonfixed",
                                        ctrl$nonfixed)
    dots <- .np_plot_set_normalized_arg(dots, "boot_control$wild",
                                        "plot.errors.boot.wild",
                                        ctrl$wild)
    if (!is.null(ctrl$blocklen))
      dots <- .np_plot_set_normalized_arg(dots, "boot_control$blocklen",
                                          "plot.errors.boot.blocklen",
                                          ctrl$blocklen)
  }
  if (has("grid_control")) {
    if (!inherits(dots$grid_control, "np_grid_control"))
      stop("grid_control must be created by np_grid_control()", call. = FALSE)
    ctrl <- dots$grid_control
    dots$grid_control <- NULL
    if (!is.null(ctrl$xtrim))
      dots <- .np_plot_set_normalized_arg(dots, "grid_control$xtrim",
                                          "xtrim", ctrl$xtrim)
    if (!is.null(ctrl$xq))
      dots <- .np_plot_set_normalized_arg(dots, "grid_control$xq",
                                          "xq", ctrl$xq)
    if (!is.null(ctrl$slices))
      stop(sprintf("grid_control$slices is not yet supported for %s", context),
           call. = FALSE)
  }
  if (has("render_control")) {
    if (!inherits(dots$render_control, "np_render_control"))
      stop("render_control must be created by np_render_control()", call. = FALSE)
    ctrl <- dots$render_control
    dots$render_control <- NULL
    dots <- .np_plot_set_normalized_arg(dots, "render_control$style",
                                        "plot.errors.style", ctrl$style)
    dots <- .np_plot_set_normalized_arg(dots, "render_control$bar",
                                        "plot.errors.bar", ctrl$bar)
    if (!is.null(ctrl$bar_num))
      dots <- .np_plot_set_normalized_arg(dots, "render_control$bar_num",
                                          "plot.errors.bar.num", ctrl$bar_num)
  }

  method <- if (!is.null(dots$plot.errors.method))
    as.character(dots$plot.errors.method)[1L]
  else
    "none"
  boot.only <- c("plot.errors.boot.method", "plot.errors.boot.num",
                 "plot.errors.boot.nonfixed", "plot.errors.boot.wild",
                 "plot.errors.boot.blocklen")
  boot.supplied <- any(c("bootstrap", "B", "boot_control", "center",
                         "plot.errors.center", boot.only) %in% supplied)
  if (!identical(method, "bootstrap") && boot.supplied)
    stop("bootstrap controls require errors = \"bootstrap\"",
         call. = FALSE)
  error.only <- c("plot.errors.type", "plot.errors.alpha",
                  "plot.errors.style", "plot.errors.bar",
                  "plot.errors.bar.num")
  error.supplied <- any(c("band", "alpha", "render_control",
                          error.only) %in% supplied)
  if (identical(method, "none") && error.supplied)
    stop("band, alpha, and interval rendering controls require errors != \"none\"",
         call. = FALSE)

  dots
}

.np_plot_call_method <- function(method, bws, ..., .plot_dots_call = NULL,
                                 .plot_context = "plot") {
  .np_plot_validate_public_dots(
    .plot_dots_call,
    method = method,
    bws = bws,
    context = .plot_context
  )
  dots <- list(...)
  dots <- .np_plot_normalize_public_dots(dots, context = .plot_context)
  random.seed <- if (!is.null(dots$random.seed)) dots$random.seed else 42L
  dots$random.seed <- NULL

  # Keep backward compatibility with legacy plot argument aliases.
  if (!is.null(dots$gradient) && is.null(dots$gradients)) {
    dots$gradients <- dots$gradient
  }
  dots$gradient <- NULL

  if (!is.null(dots$persp) && is.null(dots$perspective)) {
    dots$perspective <- dots$persp
  }
  dots$persp <- NULL

  if (!is.null(dots$plot.rug)) {
    dots$plot.rug <- .np_plot_match_flag(dots$plot.rug, "plot.rug")
    if (isTRUE(dots$plot.rug) &&
        !identical(method, .np_plot_rbandwidth_engine) &&
        !identical(method, .np_plot_bandwidth_engine) &&
        !identical(method, .np_plot_dbandwidth_engine) &&
        !identical(method, .np_plot_conbandwidth_engine) &&
        !identical(method, .np_plot_condbandwidth_engine) &&
        !identical(method, .np_plot_plbandwidth_engine) &&
        !identical(method, .np_plot_scbandwidth_engine) &&
        !identical(method, .np_plot_sibandwidth_engine) &&
        !identical(method, .np_plot_compat_dispatch)) {
      stop("plot.rug=TRUE is not yet implemented for this plot route.",
           call. = FALSE)
    }
  }

  if (!is.null(dots$renderer)) {
    dots$renderer <- .np_plot_match_renderer(dots$renderer)
    if (identical(dots$renderer, "rgl") &&
        !identical(method, .np_plot_rbandwidth_engine) &&
        !identical(method, .np_plot_bandwidth_engine) &&
        !identical(method, .np_plot_dbandwidth_engine) &&
        !identical(method, .np_plot_conbandwidth_engine) &&
        !identical(method, .np_plot_condbandwidth_engine) &&
        !identical(method, .np_plot_scbandwidth_engine) &&
        !identical(method, .np_plot_plbandwidth_engine) &&
        !identical(method, .np_plot_compat_dispatch)) {
      stop("renderer='rgl' is not yet implemented for this plot route. Use renderer='base'.",
           call. = FALSE)
    }
  }

  .np_with_seed(random.seed, do.call(method, c(list(bws = bws), dots)))
}

.np_plot_compat_dispatch <- function(bws, ...) {
  cls <- class(bws)
  dots <- list(...)

  if (!is.null(dots$renderer)) {
    dots$renderer <- .np_plot_match_renderer(dots$renderer)
    if (identical(dots$renderer, "rgl") &&
        !any(c("rbandwidth", "bandwidth", "dbandwidth",
               "conbandwidth", "condbandwidth", "scbandwidth", "plbandwidth") %in% cls)) {
      stop("renderer='rgl' is not yet implemented for this plot route. Use renderer='base'.",
           call. = FALSE)
    }
  }
  if (!is.null(dots$plot.rug)) {
    dots$plot.rug <- .np_plot_match_flag(dots$plot.rug, "plot.rug")
    if (isTRUE(dots$plot.rug) &&
        !any(c("rbandwidth", "bandwidth", "dbandwidth",
               "conbandwidth", "condbandwidth",
               "plbandwidth", "scbandwidth", "sibandwidth") %in% cls)) {
      stop("plot.rug=TRUE is not yet implemented for this plot route.",
           call. = FALSE)
    }
  }

  if (!is.null(dots$plot.data.overlay) &&
      !any(c("rbandwidth", "plbandwidth", "scbandwidth") %in% cls) &&
      !("condbandwidth" %in% cls && isTRUE(dots$quantreg))) {
    dots$plot.data.overlay <- .np_plot_match_flag(dots$plot.data.overlay,
                                                  "plot.data.overlay")
    if (isTRUE(dots$plot.data.overlay)) {
      stop("plot.data.overlay=TRUE is available only for regression, quantile regression, partially linear, and smooth coefficient plot surfaces.",
           call. = FALSE)
    }
    dots$plot.data.overlay <- NULL
  }

  if ("rbandwidth" %in% cls)
    return(do.call(.np_plot_rbandwidth_engine, c(list(bws = bws), dots)))
  if ("conbandwidth" %in% cls)
    return(do.call(.np_plot_conbandwidth_engine, c(list(bws = bws), dots)))
  if ("condbandwidth" %in% cls)
    return(do.call(.np_plot_condbandwidth_engine, c(list(bws = bws), dots)))
  if ("plbandwidth" %in% cls)
    return(do.call(.np_plot_plbandwidth_engine, c(list(bws = bws), dots)))
  if ("sibandwidth" %in% cls)
    return(do.call(.np_plot_sibandwidth_engine, c(list(bws = bws), dots)))
  if ("scbandwidth" %in% cls)
    return(do.call(.np_plot_scbandwidth_engine, c(list(bws = bws), dots)))
  if ("dbandwidth" %in% cls)
    return(do.call(.np_plot_dbandwidth_engine, c(list(bws = bws), dots)))
  if ("bandwidth" %in% cls)
    return(do.call(.np_plot_bandwidth_engine, c(list(bws = bws), dots)))

  stop("unsupported bandwidth class for plotting")
}

.np_plot_from_slot <- function(object, slot = "bws", ...,
                               .plot_dots_call = NULL,
                               .plot_context = "plot") {
  bws <- object[[slot]]
  if (is.null(bws))
    stop("plot object does not contain expected bandwidth slot")
  .np_plot_call_method(.np_plot_compat_dispatch, bws = bws, ...,
                       .plot_dots_call = .plot_dots_call,
                       .plot_context = .plot_context)
}

.np_plot_restore_bandwidth_from_call <- function(object, bws, caller_env = parent.frame()) {
  if (!is.null(bws$formula) || is.null(object$call))
    return(bws)

  bws.orig <- tryCatch(
    .np_eval_call_arg(object$call, "bws", caller_env = caller_env),
    error = function(e) NULL
  )

  if (!is.null(bws.orig) && any(grepl("bandwidth$", class(bws.orig))))
    return(bws.orig)

  bws
}

.np_plot_plreg_training_data <- function(bws) {
  out <- list(xdat = NULL, ydat = NULL, zdat = NULL)

  if (is.null(bws))
    return(out)

  if (!is.null(bws$formula) && !is.null(bws$call)) {
    tt <- terms(bws)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)

    tmf.xf <- tmf <- bws$call[c(1, m)]
    tmf[[1]] <- tmf.xf[[1]] <- as.name("model.frame")
    tmf[["formula"]] <- tt

    mf.args <- as.list(tmf)[-1L]
    tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

    bronze <- lapply(bws$chromoly, paste, collapse = " + ")
    tmf.xf[["formula"]] <- as.formula(paste(" ~ ", bronze[[2]]),
                                      env = environment(tt))
    mf.xf.args <- as.list(tmf.xf)[-1L]
    tmf.xf <- do.call(stats::model.frame, mf.xf.args, envir = environment(tt))

    out$ydat <- model.response(tmf)
    out$xdat <- tmf.xf
    out$zdat <- tmf[, bws$chromoly[[3]], drop = FALSE]
    return(out)
  }

  if (!is.null(bws$call)) {
    out$xdat <- tryCatch(data.frame(.np_eval_bws_call_arg(bws, "xdat")),
                         error = function(e) NULL)
    out$ydat <- tryCatch(.np_eval_bws_call_arg(bws, "ydat"),
                         error = function(e) NULL)
    out$zdat <- tryCatch(data.frame(.np_eval_bws_call_arg(bws, "zdat")),
                         error = function(e) NULL)
  }

  out
}

.np_plot_npregression <- function(object, ..., .plot_dots_call = NULL) {
  .np_plot_validate_public_dots(
    .plot_dots_call,
    method = .np_plot_compat_dispatch,
    bws = object$bws,
    context = "plot.npregression"
  )
  dots <- list(...)
  dots <- .np_plot_normalize_public_dots(dots, context = "plot.npregression")
  if (is.null(dots$xdat) && is.null(dots$ydat) &&
      is.null(object$bws$formula) &&
      !is.null(object$call)) {
    bws.orig <- tryCatch(.np_eval_call_arg(object$call, "bws", caller_env = parent.frame(2L)),
                         error = function(e) NULL)
    if (!is.null(bws.orig) && any(grepl("bandwidth$", class(bws.orig))))
      object$bws <- bws.orig
  }

  if (is.null(dots$xdat) && is.null(dots$ydat) &&
      isTRUE(object$trainiseval) &&
      !is.null(object$eval) &&
      !is.null(object$mean) &&
      !is.null(object$resid) &&
      NROW(object$eval) == length(object$mean) &&
      length(object$mean) == length(object$resid)) {
    dots$xdat <- object$eval
    dots$ydat <- object$mean + object$resid
  }

  do.call(.np_plot_from_slot, c(list(object = object, slot = "bws"), dots))
}
.np_plot_npdensity <- function(object, ..., .plot_dots_call = NULL) {
  .np_plot_validate_public_dots(
    .plot_dots_call,
    method = .np_plot_compat_dispatch,
    bws = object$bws,
    context = "plot.npdensity"
  )
  dots <- list(...)
  dots <- .np_plot_normalize_public_dots(dots, context = "plot.npdensity")
  plot.rug <- FALSE
  if (!is.null(dots$plot.rug)) {
    plot.rug <- .np_plot_match_flag(dots$plot.rug, "plot.rug")
    dots$plot.rug <- NULL
  }

  direct.args <- c("plot.behavior", "plot.errors.method", "plot.errors.type",
                   "plot.errors.boot.num", "plot.errors.boot.method",
                   "plot.errors.boot.nonfixed",
                   "plot.errors.alpha", "perspective", "gradients",
                   "xdat", "data", "neval", "xtrim", "xq")
  use.direct <- isTRUE(object$ndim == 1) &&
    isTRUE(object$trainiseval) &&
    !any(direct.args %in% names(dots))

  if (use.direct) {
    ex <- object$eval[[1]]
    if (!is.factor(ex)) {
      ord <- order(ex)
      xlab.val <- if (!is.null(dots$xlab)) dots$xlab else gen.label(object$xnames[1], "X1")
      ylab.val <- if (!is.null(dots$ylab)) dots$ylab else "Density"
      type.val <- if (!is.null(dots$type)) dots$type else "l"

      dots$xlab <- NULL
      dots$ylab <- NULL
      dots$type <- NULL

      do.call(plot, c(list(x = as.numeric(ex[ord]),
                           y = as.numeric(object$dens[ord]),
                           type = type.val,
                           xlab = xlab.val,
                           ylab = ylab.val),
                      dots))
      if (isTRUE(plot.rug)) {
        .np_plot_validate_rug_request(
          plot.rug = TRUE,
          route = "plot.npdensity()",
          supported.route = TRUE,
          renderer = "base"
        )
        .np_plot_draw_rug_1d(as.numeric(ex))
      }
      return(invisible(object))
    }
  }

  do.call(.np_plot_from_slot, c(list(object = object, slot = "bws"), dots))
}
.np_plot_condensity <- function(object, ..., .plot_dots_call = NULL) {
  .np_plot_validate_public_dots(
    .plot_dots_call,
    method = .np_plot_compat_dispatch,
    bws = object$bws,
    context = "plot.condensity"
  )
  dots <- list(...)
  dots <- .np_plot_normalize_public_dots(dots, context = "plot.condensity")
  if (is.null(dots$proper) && isTRUE(object$proper.requested))
    dots$proper <- TRUE
  do.call(.np_plot_from_slot, c(list(object = object, slot = "bws"), dots))
}
.np_plot_condistribution <- function(object, ..., .plot_dots_call = NULL) {
  .np_plot_validate_public_dots(
    .plot_dots_call,
    method = .np_plot_compat_dispatch,
    bws = object$bws,
    context = "plot.condistribution"
  )
  dots <- list(...)
  dots <- .np_plot_normalize_public_dots(dots, context = "plot.condistribution")
  if (is.null(dots$proper) && isTRUE(object$proper.requested))
    dots$proper <- TRUE
  do.call(.np_plot_from_slot, c(list(object = object, slot = "bws"), dots))
}
.np_plot_npdistribution <- function(object, ..., .plot_dots_call = NULL)
  .np_plot_from_slot(object, "bws", ...,
                     .plot_dots_call = .plot_dots_call,
                     .plot_context = "plot.npdistribution")
.np_plot_qregression <- function(object, ..., .plot_dots_call = NULL) {
  .np_plot_validate_public_dots(
    .plot_dots_call,
    method = .np_plot_compat_dispatch,
    bws = object$bws,
    context = "plot.qregression"
  )
  dots <- list(...)
  dots <- .np_plot_normalize_public_dots(dots, context = "plot.qregression")
  if (is.null(dots$quantreg))
    dots$quantreg <- TRUE
  if (is.null(dots$tau) && !is.null(object$tau))
    dots$tau <- object$tau
  if (is.null(dots$plot.data.overlay) && !isTRUE(dots$gradients))
    dots$plot.data.overlay <- TRUE
  do.call(.np_plot_from_slot, c(list(object = object, slot = "bws"), dots))
}
.np_plot_conmode_data <- function(object,
                                  gradients = FALSE,
                                  level = NULL,
                                  errors = "none",
                                  alpha = 0.05,
                                  band = "pointwise",
                                  center = "estimate",
                                  bootstrap = "inid",
                                  B = 399L,
                                  blocklen = NULL) {
  lev <- if (!is.null(object$probability.levels)) {
    as.character(object$probability.levels)
  } else if (!is.null(object$probability.gradient.level)) {
    as.character(object$probability.gradient.level)
  } else {
    levels(object$conmode)
  }
  lev <- unique(lev)
  if (!length(lev))
    stop("conmode object does not contain response-level metadata", call. = FALSE)

  if (is.null(level)) {
    if (isTRUE(gradients) && !is.null(object$probability.gradient.level)) {
      level <- as.character(object$probability.gradient.level)
    } else {
      level <- lev[1L]
    }
  }
  if (length(level) != 1L || is.na(level) || !(as.character(level) %in% lev))
    stop("'level' must identify one response level in the fitted conmode object",
         call. = FALSE)
  level <- as.character(level)
  level.idx <- match(level, lev)

  if (isTRUE(gradients)) {
    values <- object$probability.gradients
    if (is.null(values))
      stop("class-probability gradients/effects are not available: fit with gradients=TRUE",
           call. = FALSE)
    stored.level <- object$probability.gradient.level
    if (!identical(as.character(level), as.character(stored.level)))
      stop(sprintf("stored class-probability gradients are for level %s; refit with level=%s to plot that level",
                   sQuote(as.character(stored.level)),
                   sQuote(as.character(level))),
           call. = FALSE)
    p <- ncol(values)
    ymat <- as.matrix(values)
    value.name <- "effect"
  } else {
    values <- object$probabilities
    if (is.null(values))
      stop("class probabilities are not available: fit with probabilities=TRUE or gradients=TRUE",
           call. = FALSE)
    p <- object$xndim
    ymat <- matrix(values[, level.idx], nrow = nrow(values), ncol = 1L)
    if (p > 1L)
      ymat <- ymat[, rep(1L, p), drop = FALSE]
    value.name <- "probability"
  }

  xeval <- as.data.frame(object$xeval)
  boot.raw <- NULL
  boot.se <- NULL
  if (!isTRUE(gradients) && identical(errors, "bootstrap")) {
    training <- .np_plot_conmode_training_data(object)
    boot.raw <- .np_plot_conmode_bootstrap_raw(
      object = object,
      xtrain = training$xtrain,
      ytrain = training$ytrain,
      exdat = xeval,
      level = level,
      levels = lev,
      t0 = ymat[, 1L],
      alpha = alpha,
      band = band,
      center = center,
      method = bootstrap,
      B = B,
      blocklen = blocklen
    )
    boot.se <- rep(NA_real_, nrow(xeval))
  }
  out <- vector("list", p)
  xnames <- object$xnames
  if (is.null(xnames) || length(xnames) != p)
    xnames <- names(xeval)[seq_len(p)]
  for (j in seq_len(p)) {
    out[[j]] <- data.frame(
      variable = xnames[j],
      x = xeval[[j]],
      value = ymat[, j],
      level = level,
      gradients = isTRUE(gradients),
      stringsAsFactors = FALSE
    )
    names(out[[j]])[3L] <- value.name
    if (!isTRUE(gradients) && !identical(errors, "none")) {
      se <- rep(NA_real_, nrow(out[[j]]))
      if (identical(errors, "asymptotic")) {
        se <- object$probability.errors
        if (is.null(se))
          stop("class-probability standard errors are not available: refit with probabilities=TRUE",
               call. = FALSE)
        se <- as.matrix(se)[, level.idx]
        repaired <- object$probability.repaired.rows
        if (!is.null(repaired)) {
          repaired <- as.logical(repaired)
          se[is.na(repaired) | repaired] <- NA_real_
        }
      }
      out[[j]] <- .np_plot_conmode_add_interval_columns(
        out[[j]],
        se = if (identical(errors, "bootstrap")) boot.se else se,
        errors = errors,
        alpha = alpha,
        band = band,
        center = center,
        bootstrap_raw = boot.raw
      )
    }
  }
  names(out) <- xnames
  out
}

.np_plot_conmode_cast_like <- function(x, template) {
  if (is.factor(template))
    return(factor(as.character(x),
                  levels = levels(template),
                  ordered = is.ordered(template)))
  if (is.integer(template))
    return(as.integer(x))
  if (is.numeric(template))
    return(as.numeric(x))
  x
}

.np_plot_conmode_grid_values <- function(x, neval, xtrim) {
  if (is.factor(x)) {
    return(factor(levels(x), levels = levels(x), ordered = is.ordered(x)))
  }
  rng <- stats::quantile(x, probs = xtrim, names = FALSE, na.rm = TRUE)
  seq(rng[1L], rng[2L], length.out = neval)
}

.np_plot_conmode_base_row <- function(xtrain, xq) {
  out <- xtrain[1L, , drop = FALSE]
  for (j in seq_along(xtrain)) {
    val <- uocquantile(xtrain[[j]], prob = xq[j])
    out[[j]] <- .np_plot_conmode_cast_like(val, xtrain[[j]])
  }
  out
}

.np_plot_conmode_level_factor <- function(ytrain, level, n) {
  y <- ytrain[[1L]]
  factor(rep(level, n),
         levels = levels(y),
         ordered = is.ordered(y))
}

.np_plot_conmode_training_data <- function(object) {
  xtrain <- object$xtrain
  ytrain <- object$ytrain
  if ((is.null(xtrain) || is.null(ytrain)) &&
      isTRUE(object$trainiseval) &&
      !is.null(object$xeval) &&
      !is.null(object$yeval)) {
    xtrain <- object$xeval
    ytrain <- object$yeval
  }
  if (is.null(xtrain) || is.null(ytrain))
    stop("fixed-grid plot.conmode displays require an object fitted with probabilities=TRUE or gradients=TRUE",
         call. = FALSE)
  list(xtrain = as.data.frame(xtrain), ytrain = as.data.frame(ytrain))
}

.np_plot_conmode_levels <- function(object, ytrain) {
  lev <- if (!is.null(object$probability.levels)) {
    as.character(object$probability.levels)
  } else {
    levels(ytrain[[1L]])
  }
  lev <- unique(lev)
  if (!length(lev))
    stop("conmode object does not contain response-level metadata", call. = FALSE)
  lev
}

.np_plot_conmode_resolve_level <- function(object, ytrain, gradients, level) {
  lev <- .np_plot_conmode_levels(object, ytrain)
  if (is.null(level)) {
    if (isTRUE(gradients) && !is.null(object$probability.gradient.level)) {
      level <- as.character(object$probability.gradient.level)
    } else {
      level <- lev[1L]
    }
  }
  if (length(level) != 1L || is.na(level) || !(as.character(level) %in% lev))
    stop("'level' must identify one response level in the fitted conmode object",
         call. = FALSE)
  level <- as.character(level)
  if (isTRUE(gradients) &&
      !identical(level, as.character(object$probability.gradient.level)))
    stop(sprintf("stored class-probability gradients are for level %s; refit with level=%s to plot that level",
                 sQuote(as.character(object$probability.gradient.level)),
                 sQuote(level)),
         call. = FALSE)
  list(level = level, levels = lev)
}

.np_plot_conmode_resolve_neval <- function(neval) {
  if (!is.numeric(neval) || length(neval) != 1L || is.na(neval) || neval < 2L)
    stop("neval must be a numeric scalar at least 2 for fixed-grid plot.conmode displays",
         call. = FALSE)
  as.integer(neval)
}

.np_plot_conmode_resolve_xtrim <- function(xtrim) {
  if (!is.numeric(xtrim) || length(xtrim) != 2L ||
      any(is.na(xtrim)) || any(xtrim < 0) || any(xtrim > 1) ||
      xtrim[1L] >= xtrim[2L])
    stop("xtrim must be a numeric length-two vector with 0 <= xtrim[1] < xtrim[2] <= 1",
         call. = FALSE)
  xtrim
}

.np_plot_conmode_resolve_xq <- function(xq, p) {
  if (is.null(xq))
    xq <- rep(0.5, p)
  if (!is.numeric(xq) || any(is.na(xq)) || any(xq < 0) || any(xq > 1))
    stop("xq must contain probabilities in [0,1]", call. = FALSE)
  if (length(xq) == 1L)
    xq <- rep(xq, p)
  if (length(xq) != p)
    stop("xq must be a scalar or have one entry for each conditioning variable",
         call. = FALSE)
  xq
}

.np_plot_conmode_probability_se <- function(object,
                                            xtrain,
                                            ytrain,
                                            exdat,
                                            level) {
  dens.obj <- npcdens(
    txdat = xtrain,
    tydat = ytrain,
    exdat = exdat,
    eydat = .np_plot_conmode_level_factor(ytrain, level, nrow(exdat)),
    bws = object$bws
  )
  as.vector(dens.obj$conderr)
}

.np_plot_conmode_proper_control <- function(object) {
  out <- list()
  if (!is.null(object$proper.info$tol))
    out$tol <- object$proper.info$tol
  out
}

.np_plot_conmode_probability_matrix <- function(object,
                                                xtrain,
                                                ytrain,
                                                exdat,
                                                levels) {
  rhs <- rep.int(1.0, nrow(xtrain))
  pmat <- matrix(NA_real_, nrow(exdat), length(levels),
                 dimnames = list(NULL, levels))
  for (k in seq_along(levels)) {
    pmat[, k] <- npcdenshat(
      bws = object$bws,
      txdat = xtrain,
      tydat = ytrain,
      exdat = exdat,
      eydat = .np_plot_conmode_level_factor(ytrain, levels[k], nrow(exdat)),
      y = rhs,
      output = "apply"
    )
  }
  .npConmodeProperProbabilities(
    pmat,
    levels = levels,
    proper = isTRUE(object$proper.requested),
    proper.control = .np_plot_conmode_proper_control(object)
  )
}

.np_plot_conmode_boot_counts <- function(n, B, method, blocklen) {
  if (identical(method, "inid"))
    return(.np_inid_counts_matrix(n = n, B = B))
  drawer <- .np_block_counts_drawer(
    n = n,
    sim = method,
    blocklen = blocklen,
    B = B
  )
  .np_inid_counts_matrix(n = n, B = B, counts = drawer(1L, B))
}

.np_plot_conmode_bootstrap_values <- function(object,
                                              xtrain,
                                              ytrain,
                                              exdat,
                                              level,
                                              levels,
                                              t0,
                                              B,
                                              method,
                                              blocklen) {
  if (identical(method, "wild"))
    stop("wild bootstrap is not available for class-probability plot.conmode intervals; use bootstrap=\"inid\", \"fixed\", or \"geom\"",
         call. = FALSE)
  B <- as.integer(B)
  if (!is.numeric(B) || length(B) != 1L || is.na(B) || B < 1L)
    stop("plot.errors.boot.num must be a positive integer", call. = FALSE)
  n <- nrow(xtrain)
  counts <- .np_plot_conmode_boot_counts(
    n = n,
    B = B,
    method = method,
    blocklen = blocklen
  )
  level.idx <- match(level, levels)
  tmat <- matrix(NA_real_, nrow = B, ncol = length(t0))
  progress <- .np_plot_bootstrap_progress_begin(
    total = B,
    label = "Conmode probability bootstrap"
  )
  on.exit(.np_plot_progress_end(progress), add = TRUE)
  for (bb in seq_len(B)) {
    idx <- rep.int(seq_len(n), counts[, bb])
    proper.out <- .np_plot_conmode_probability_matrix(
      object = object,
      xtrain = xtrain[idx, , drop = FALSE],
      ytrain = ytrain[idx, , drop = FALSE],
      exdat = exdat,
      levels = levels
    )
    tmat[bb, ] <- proper.out$probabilities[, level.idx]
    progress <- .np_plot_progress_tick(progress, bb)
  }
  list(t = tmat, t0 = t0)
}

.np_plot_conmode_bootstrap_raw <- function(object,
                                           xtrain,
                                           ytrain,
                                           exdat,
                                           level,
                                           levels,
                                           t0,
                                           alpha,
                                           band,
                                           center,
                                           method,
                                           B,
                                           blocklen) {
  boot.out <- .np_plot_conmode_bootstrap_values(
    object = object,
    xtrain = xtrain,
    ytrain = ytrain,
    exdat = exdat,
    level = level,
    levels = levels,
    t0 = t0,
    B = B,
    method = method,
    blocklen = blocklen
  )
  if (identical(band, "pmzsd")) {
    boot.err <- matrix(NA_real_, nrow = length(t0), ncol = 3L)
    boot.sd <- .np_plot_bootstrap_col_sds(boot.out$t)
    boot.err[, 1:2] <- stats::qnorm(alpha / 2, lower.tail = FALSE) * boot.sd
    boot.all.err <- NULL
  } else {
    interval.summary <- .np_plot_bootstrap_interval_summary(
      boot.t = boot.out$t,
      t0 = boot.out$t0,
      alpha = alpha,
      band.type = band
    )
    boot.err <- matrix(NA_real_, nrow = length(t0), ncol = 3L)
    boot.err[, 1:2] <- interval.summary$err
    boot.all.err <- interval.summary$all.err
  }
  if (identical(center, "bias-corrected"))
    boot.err[, 3L] <- 2 * boot.out$t0 - colMeans(boot.out$t)
  list(
    boot.err = boot.err,
    boot.all.err = boot.all.err,
    bxp = list(),
    boot.t = boot.out$t
  )
}

.np_plot_conmode_add_interval_columns <- function(dat,
                                                  se,
                                                  errors,
                                                  alpha,
                                                  band,
                                                  center = "estimate",
                                                  bootstrap_raw = NULL) {
  payload <- .np_plot_interval_payload(
    estimate = dat$probability,
    se = se,
    plot.errors.method = errors,
    plot.errors.alpha = alpha,
    plot.errors.type = band,
    plot.errors.center = center,
    bootstrap_raw = bootstrap_raw
  )
  dat$stderr <- if (identical(errors, "bootstrap") &&
                    !is.null(bootstrap_raw$boot.t)) {
    .np_plot_bootstrap_col_sds(bootstrap_raw$boot.t)
  } else {
    as.numeric(se)
  }
  dat$center <- payload$center
  dat$lower <- payload$center - payload$err[, 1L]
  dat$upper <- payload$center + payload$err[, 2L]
  if (!is.null(payload$all.err)) {
    for (nm in names(payload$all.err)) {
      dat[[paste0(nm, ".lower")]] <- payload$center - payload$all.err[[nm]][, 1L]
      dat[[paste0(nm, ".upper")]] <- payload$center + payload$all.err[[nm]][, 2L]
    }
  }
  dat
}

.np_plot_conmode_grid_data <- function(object,
                                       gradients = FALSE,
                                       level = NULL,
                                       neval = 50L,
                                       xtrim = c(0, 1),
                                       xq = NULL,
                                       errors = "none",
                                       alpha = 0.05,
                                       band = "pointwise",
                                       center = "estimate",
                                       bootstrap = "inid",
                                       B = 399L,
                                       blocklen = NULL) {
  if (!isTRUE(gradients) && is.null(object$probabilities))
    stop("class probabilities are not available: fit with probabilities=TRUE or gradients=TRUE",
         call. = FALSE)
  if (isTRUE(gradients) && is.null(object$probability.gradients))
    stop("class-probability gradients/effects are not available: fit with gradients=TRUE",
         call. = FALSE)

  training <- .np_plot_conmode_training_data(object)
  xtrain <- training$xtrain
  ytrain <- training$ytrain
  resolved <- .np_plot_conmode_resolve_level(object, ytrain, gradients, level)
  level <- resolved$level
  lev <- resolved$levels
  neval <- .np_plot_conmode_resolve_neval(neval)
  xtrim <- .np_plot_conmode_resolve_xtrim(xtrim)

  p <- ncol(xtrain)
  xnames <- object$xnames
  if (is.null(xnames) || length(xnames) != p)
    xnames <- names(xtrain)
  xq <- .np_plot_conmode_resolve_xq(xq, p)

  rhs <- rep.int(1.0, nrow(xtrain))
  base <- .np_plot_conmode_base_row(xtrain, xq)
  proper.control <- list()
  if (!is.null(object$proper.info$tol))
    proper.control$tol <- object$proper.info$tol
  out <- vector("list", p)
  names(out) <- xnames
  ixcon <- object$bws$ixcon
  if (is.null(ixcon) || length(ixcon) != p)
    ixcon <- vapply(xtrain, is.numeric, logical(1L))

  for (j in seq_len(p)) {
    grid <- .np_plot_conmode_grid_values(xtrain[[j]], neval = neval,
                                         xtrim = xtrim)
    exdat <- base[rep(1L, length(grid)), , drop = FALSE]
    exdat[[j]] <- .np_plot_conmode_cast_like(grid, xtrain[[j]])

    if (isTRUE(gradients)) {
      if (isTRUE(ixcon[j])) {
        s <- integer(sum(ixcon))
        names(s) <- xnames[ixcon]
        s[xnames[j]] <- 1L
        vals <- npcdenshat(
          bws = object$bws,
          txdat = xtrain,
          tydat = ytrain,
          exdat = exdat,
          eydat = .np_plot_conmode_level_factor(ytrain, level, nrow(exdat)),
          y = rhs,
          output = "apply",
          s = s
        )
      } else {
        vals <- rep(0, nrow(exdat))
      }
      out[[j]] <- data.frame(
        variable = xnames[j],
        x = exdat[[j]],
        effect = as.vector(vals),
        level = level,
        gradients = TRUE,
        view = "fixed",
        stringsAsFactors = FALSE
      )
    } else {
      pmat <- matrix(NA_real_, nrow(exdat), length(lev),
                     dimnames = list(NULL, lev))
      for (k in seq_along(lev)) {
        pmat[, k] <- npcdenshat(
          bws = object$bws,
          txdat = xtrain,
          tydat = ytrain,
          exdat = exdat,
          eydat = .np_plot_conmode_level_factor(ytrain, lev[k], nrow(exdat)),
          y = rhs,
          output = "apply"
        )
      }
      proper.out <- .npConmodeProperProbabilities(
        pmat,
        levels = lev,
        proper = isTRUE(object$proper.requested),
        proper.control = proper.control
      )
      dat <- data.frame(
        variable = xnames[j],
        x = exdat[[j]],
        probability = proper.out$probabilities[, match(level, lev)],
        level = level,
        gradients = FALSE,
        view = "fixed",
        stringsAsFactors = FALSE
      )
      if (!identical(errors, "none")) {
        boot.raw <- NULL
        if (identical(errors, "bootstrap")) {
          boot.raw <- .np_plot_conmode_bootstrap_raw(
            object = object,
            xtrain = xtrain,
            ytrain = ytrain,
            exdat = exdat,
            level = level,
            levels = lev,
            t0 = dat$probability,
            alpha = alpha,
            band = band,
            center = center,
            method = bootstrap,
            B = B,
            blocklen = blocklen
          )
          se <- rep(NA_real_, nrow(exdat))
        } else {
          se <- .np_plot_conmode_probability_se(
            object = object,
            xtrain = xtrain,
            ytrain = ytrain,
            exdat = exdat,
            level = level
          )
          se[proper.out$repaired.rows] <- NA_real_
        }
        dat <- .np_plot_conmode_add_interval_columns(
          dat,
          se = se,
          errors = errors,
          alpha = alpha,
          band = band,
          center = center,
          bootstrap_raw = boot.raw
        )
      }
      out[[j]] <- dat
    }
  }
  out
}

.np_plot_conmode_resolve_surface_vars <- function(xtrain, object, plot.vars = NULL) {
  p <- ncol(xtrain)
  xnames <- object$xnames
  if (is.null(xnames) || length(xnames) != p)
    xnames <- names(xtrain)
  ixcon <- object$bws$ixcon
  if (is.null(ixcon) || length(ixcon) != p)
    ixcon <- vapply(xtrain, is.numeric, logical(1L))
  con.names <- xnames[ixcon]
  if (is.null(plot.vars)) {
    if (length(con.names) != 2L)
      stop("surface plot.conmode payloads require exactly two continuous conditioning variables, or a length-two 'plot.vars' selection",
           call. = FALSE)
    return(con.names)
  }
  if (!is.character(plot.vars) || length(plot.vars) != 2L ||
      any(is.na(plot.vars)) || any(!nzchar(plot.vars)))
    stop("'plot.vars' must be a character vector naming exactly two conditioning variables",
         call. = FALSE)
  if (any(!(plot.vars %in% xnames)))
    stop("'plot.vars' must name conditioning variables in the fitted conmode object",
         call. = FALSE)
  if (any(!(plot.vars %in% con.names)))
    stop("surface plot.conmode payloads currently require continuous variables in 'plot.vars'",
         call. = FALSE)
  plot.vars
}

.np_plot_conmode_surface_data <- function(object,
                                          level = NULL,
                                          neval = 50L,
                                          xtrim = c(0, 1),
                                          xq = NULL,
                                          plot.vars = NULL,
                                          errors = "none",
                                          alpha = 0.05,
                                          band = "pointwise") {
  if (is.null(object$probabilities))
    stop("class probabilities are not available: fit with probabilities=TRUE or gradients=TRUE",
         call. = FALSE)
  training <- .np_plot_conmode_training_data(object)
  xtrain <- training$xtrain
  ytrain <- training$ytrain
  resolved <- .np_plot_conmode_resolve_level(object, ytrain, FALSE, level)
  level <- resolved$level
  lev <- resolved$levels
  neval <- .np_plot_conmode_resolve_neval(neval)
  xtrim <- .np_plot_conmode_resolve_xtrim(xtrim)

  p <- ncol(xtrain)
  xnames <- object$xnames
  if (is.null(xnames) || length(xnames) != p)
    xnames <- names(xtrain)
  xq <- .np_plot_conmode_resolve_xq(xq, p)
  plot.vars <- .np_plot_conmode_resolve_surface_vars(
    xtrain = xtrain,
    object = object,
    plot.vars = plot.vars
  )
  ivars <- match(plot.vars, xnames)

  grid1 <- .np_plot_conmode_grid_values(xtrain[[ivars[1L]]],
                                        neval = neval,
                                        xtrim = xtrim)
  grid2 <- .np_plot_conmode_grid_values(xtrain[[ivars[2L]]],
                                        neval = neval,
                                        xtrim = xtrim)
  base <- .np_plot_conmode_base_row(xtrain, xq)
  grid.index <- expand.grid(
    i1 = seq_along(grid1),
    i2 = seq_along(grid2),
    KEEP.OUT.ATTRS = FALSE
  )
  exdat <- base[rep(1L, nrow(grid.index)), , drop = FALSE]
  exdat[[ivars[1L]]] <- .np_plot_conmode_cast_like(grid1[grid.index$i1],
                                                   xtrain[[ivars[1L]]])
  exdat[[ivars[2L]]] <- .np_plot_conmode_cast_like(grid2[grid.index$i2],
                                                   xtrain[[ivars[2L]]])

  rhs <- rep.int(1.0, nrow(xtrain))
  pmat <- matrix(NA_real_, nrow(exdat), length(lev),
                 dimnames = list(NULL, lev))
  for (k in seq_along(lev)) {
    pmat[, k] <- npcdenshat(
      bws = object$bws,
      txdat = xtrain,
      tydat = ytrain,
      exdat = exdat,
      eydat = .np_plot_conmode_level_factor(ytrain, lev[k], nrow(exdat)),
      y = rhs,
      output = "apply"
    )
  }
  proper.control <- list()
  if (!is.null(object$proper.info$tol))
    proper.control$tol <- object$proper.info$tol
  proper.out <- .npConmodeProperProbabilities(
    pmat,
    levels = lev,
    proper = isTRUE(object$proper.requested),
    proper.control = proper.control
  )

  surface <- data.frame(
    variable1 = plot.vars[1L],
    variable2 = plot.vars[2L],
    grid_index1 = grid.index$i1,
    grid_index2 = grid.index$i2,
    x1 = exdat[[ivars[1L]]],
    x2 = exdat[[ivars[2L]]],
    probability = proper.out$probabilities[, match(level, lev)],
    level = level,
    view = "fixed",
    perspective = TRUE,
    stringsAsFactors = FALSE
  )
  if (!identical(errors, "none")) {
    se <- .np_plot_conmode_probability_se(
      object = object,
      xtrain = xtrain,
      ytrain = ytrain,
      exdat = exdat,
      level = level
    )
    se[proper.out$repaired.rows] <- NA_real_
    surface <- .np_plot_conmode_add_interval_columns(
      surface,
      se = se,
      errors = errors,
      alpha = alpha,
      band = band
    )
  }
  held.vars <- setdiff(xnames, plot.vars)
  held <- if (length(held.vars)) base[, held.vars, drop = FALSE] else data.frame()
  structure(
    list(
      surface = surface,
      grid = list(
        variables = plot.vars,
        values = setNames(list(grid1, grid2), plot.vars),
        dims = c(length(grid1), length(grid2))
      ),
      held = held,
      level = level,
      levels = lev,
      proper = proper.out$proper.info
    ),
    class = "np_conmode_surface_data"
  )
}

.np_plot_conmode_surface_matrix <- function(payload) {
  dims <- payload$grid$dims
  matrix(payload$surface$probability, nrow = dims[1L], ncol = dims[2L])
}

.np_plot_conmode_validate_zlim <- function(zlim) {
  if (is.null(zlim))
    return(NULL)
  if (!is.numeric(zlim) || length(zlim) != 2L || any(is.na(zlim)) ||
      zlim[1L] >= zlim[2L])
    stop("zlim must be a numeric length-two vector with zlim[1] < zlim[2]",
         call. = FALSE)
  zlim
}

.np_plot_conmode_surface_panel <- function(payload, dots, renderer = "base") {
  vars <- payload$grid$variables
  x1 <- payload$grid$values[[1L]]
  x2 <- payload$grid$values[[2L]]
  z <- .np_plot_conmode_surface_matrix(payload)
  zlim <- .np_plot_conmode_validate_zlim(dots$zlim)
  if (is.null(zlim))
    zlim <- c(0, 1)

  theta <- .np_plot_scalar_default(dots$theta, 0)
  phi <- .np_plot_scalar_default(dots$phi, 20)
  main <- .np_plot_scalar_default(
    dots$main,
    paste0("Pr(Y=", payload$level, "|X=x)")
  )
  xlab <- .np_plot_scalar_default(dots$xlab, vars[1L])
  ylab <- .np_plot_scalar_default(dots$ylab, vars[2L])
  zlab <- .np_plot_scalar_default(dots$zlab, "Probability")
  border <- .np_plot_scalar_default(dots$border, "black")

  if (identical(renderer, "rgl")) {
    rgl.view <- .np_plot_rgl_view_angles(theta = theta, phi = phi)
    return(.np_plot_render_surface_rgl(
      x = x1,
      y = x2,
      z = z,
      xlab = xlab,
      ylab = ylab,
      zlab = zlab,
      main = main,
      theta = rgl.view$theta,
      phi = rgl.view$phi,
      col = dots$col,
      border = border,
      zlim = zlim,
      par3d.args = .np_plot_collect_rgl_args(dots, "rgl.par3d", "rgl.par3d."),
      view3d.args = .np_plot_collect_rgl_args(dots, "rgl.view3d", "rgl.view3d."),
      persp3d.args = .np_plot_collect_rgl_args(dots, "rgl.persp3d", "rgl.persp3d."),
      grid3d.args = .np_plot_collect_rgl_args(dots, "rgl.grid3d", "rgl.grid3d."),
      widget.args = .np_plot_collect_rgl_args(dots, "rgl.widget", "rgl.widget.")
    ))
  }

  persp.args <- list(
    x = x1,
    y = x2,
    z = z,
    theta = theta,
    phi = phi,
    ticktype = "detailed",
    xlab = xlab,
    ylab = ylab,
    zlab = zlab,
    zlim = zlim,
    main = main,
    col = grDevices::adjustcolor(
      .np_plot_persp_surface_colors(z = z, col = dots$col),
      alpha.f = 0.5
    ),
    border = border
  )
  persp.args <- .np_plot_merge_user_args(
    persp.args,
    .np_plot_user_args(dots, "persp")
  )
  do.call(graphics::persp, persp.args)
  invisible(NULL)
}

.np_plot_conmode_panel <- function(dat,
                                   gradients = FALSE,
                                   level,
                                   plot.user.args = list(),
                                   line.user.args = list(),
                                   data_rug = FALSE,
                                   main = NULL,
                                   xlab = NULL,
                                   ylab = NULL) {
  x <- dat$x
  y <- if (isTRUE(gradients)) dat$effect else dat$probability
  vname <- dat$variable[1L]
  if (is.null(main))
    main <- if (isTRUE(gradients)) {
      paste0("Effect on Pr(Y=", level, "|X=x)")
    } else {
      paste0("Pr(Y=", level, "|X=x)")
    }
  if (is.null(xlab))
    xlab <- vname
  if (is.null(ylab))
    ylab <- if (isTRUE(gradients)) "Effect" else "Probability"

  fixed.grid <- identical(dat$view[1L], "fixed")
  if (is.factor(x) && !is.ordered(x) && !isTRUE(fixed.grid)) {
    args <- .np_plot_merge_user_args(
      list(formula = y ~ x, xlab = xlab, ylab = ylab, main = main),
      plot.user.args
    )
    do.call(graphics::boxplot, args)
    return(invisible(NULL))
  }

  xplot <- if (is.factor(x)) as.numeric(x) else x
  ord <- order(xplot)
  interval.cols <- !isTRUE(gradients) && all(c("lower", "upper") %in% names(dat))
  ylim.interval <- NULL
  if (interval.cols) {
    ylim.interval <- range(c(y, dat$lower, dat$upper), finite = TRUE)
    if (!all(is.finite(ylim.interval)))
      ylim.interval <- NULL
  }
  args <- .np_plot_merge_user_args(
    list(x = xplot[ord], y = y[ord],
         type = if (is.factor(x) && !is.ordered(x)) "p" else "l",
         xlab = xlab, ylab = ylab, main = main),
    line.user.args
  )
  if (is.null(args$ylim) && !is.null(ylim.interval))
    args$ylim <- ylim.interval
  args <- .np_plot_merge_user_args(
    args,
    plot.user.args
  )
  if (is.factor(x)) {
    args$xaxt <- "n"
    do.call(graphics::plot, args)
    graphics::axis(1L, at = seq_along(levels(x)), labels = levels(x))
  } else {
    do.call(graphics::plot, args)
  }
  if (isTRUE(data_rug) && is.numeric(xplot))
    graphics::rug(xplot)
  if (interval.cols) {
    lower <- dat$lower[ord]
    upper <- dat$upper[ord]
    xord <- xplot[ord]
    good <- is.finite(xord) & is.finite(lower) & is.finite(upper)
    if (any(good)) {
      if (is.factor(x) && !is.ordered(x)) {
        graphics::segments(xord[good], lower[good], xord[good], upper[good],
                           col = "gray45", lty = 2)
      } else {
        graphics::lines(xord[good], lower[good], col = "gray45", lty = 2)
        graphics::lines(xord[good], upper[good], col = "gray45", lty = 2)
      }
    }
  }
  invisible(NULL)
}

.np_plot_conmode <- function(object, ...,
                             gradients = FALSE,
                             level = NULL,
                             output = c("plot", "data", "plot-data", "both"),
                             data_rug = FALSE,
                             layout = TRUE,
                             legend = TRUE,
                             .plot_dots_call = NULL) {
  dots <- list(...)
  dot.names <- .np_plot_dot_names(.plot_dots_call)
  if (any(!nzchar(dot.names)))
    stop("unnamed plot arguments are not supported for plot.conmode",
         call. = FALSE)
  interval.args <- c("errors", "band", "alpha", "bootstrap", "B", "center",
                     "render_control",
                     "boot_control", "plot.errors.method", "plot.errors.type",
                     "plot.errors.alpha", "plot.errors.boot.method",
                     "plot.errors.boot.num", "plot.errors.boot.nonfixed",
                     "plot.errors.boot.wild", "plot.errors.boot.blocklen",
                     "plot.errors.center", "plot.errors.style",
                     "plot.errors.bar", "plot.errors.bar.num")
  surface.args <- c("renderer", "perspective", "persp", "plot.vars")
  grid.args <- c("neval", "grid_control", "view", "xtrim", "xq")
  grid.supplied <- intersect(dot.names[nzchar(dot.names)], grid.args)
  rgl.prefixed.args <- dot.names[startsWith(dot.names, "rgl.")]
  allowed <- unique(c("gradients", "level", "output", "data_rug",
                      "layout", "legend", grid.args, surface.args,
                      interval.args, rgl.prefixed.args,
                      .np_plot_graphics_arg_names()))
  .np_plot_stop_unused_args(setdiff(dot.names[nzchar(dot.names)], allowed),
                            allowed)
  perspective.raw <- dots$perspective
  persp.raw <- dots$persp
  dots$perspective <- NULL
  dots$persp <- NULL
  dots <- .np_plot_normalize_public_dots(dots, context = "plot.conmode")
  if (!is.null(perspective.raw))
    dots$perspective <- perspective.raw
  if (!is.null(persp.raw))
    dots$persp <- persp.raw
  if (!is.null(dots$plot.behavior)) {
    output <- dots$plot.behavior
    dots$plot.behavior <- NULL
  }
  if (!is.null(dots$plot.rug)) {
    data_rug <- dots$plot.rug
    dots$plot.rug <- NULL
  }
  if (!is.null(dots$plot.par.mfrow)) {
    layout <- dots$plot.par.mfrow
    dots$plot.par.mfrow <- NULL
  }
  errors <- if (is.null(dots$plot.errors.method)) "none" else
    .np_plot_scalar_match(dots$plot.errors.method,
                          c("none", "bootstrap", "asymptotic"),
                          "plot.errors.method")
  band <- if (is.null(dots$plot.errors.type)) "pointwise" else
    .np_plot_scalar_match(dots$plot.errors.type,
                          c("pmzsd", "pointwise", "bonferroni",
                            "simultaneous", "all"),
                          "plot.errors.type")
  alpha <- if (is.null(dots$plot.errors.alpha)) 0.05 else
    dots$plot.errors.alpha
  center <- if (is.null(dots$plot.errors.center)) "estimate" else
    .np_plot_scalar_match(dots$plot.errors.center,
                          c("estimate", "bias-corrected"),
                          "plot.errors.center")
  if (!is.numeric(alpha) || length(alpha) != 1L ||
      is.na(alpha) || alpha <= 0 || alpha >= 0.5)
    stop("plot.errors.alpha must lie in (0, 0.5)", call. = FALSE)
  bootstrap <- if (is.null(dots$plot.errors.boot.method)) "inid" else
    .np_plot_scalar_match(dots$plot.errors.boot.method,
                          c("wild", "inid", "fixed", "geom"),
                          "plot.errors.boot.method")
  B <- if (is.null(dots$plot.errors.boot.num)) 399L else
    dots$plot.errors.boot.num
  if (!is.numeric(B) || length(B) != 1L || is.na(B) || B < 1)
    stop("plot.errors.boot.num must be a positive numeric scalar",
         call. = FALSE)
  B <- as.integer(B)
  boot.blocklen <- dots$plot.errors.boot.blocklen
  if (!identical(center, "estimate") && !identical(errors, "bootstrap"))
    stop("bias-corrected interval centering requires errors=\"bootstrap\"",
         call. = FALSE)
  unsupported.interval <- intersect(
    names(dots),
    c("plot.errors.style", "plot.errors.bar", "plot.errors.bar.num")
  )
  if (length(unsupported.interval))
    stop("plot.conmode probability intervals currently use simple asymptotic line/error-bar rendering; render_control/style arguments are not yet implemented",
         call. = FALSE)
  gradients <- npValidateScalarLogical(gradients, "gradients")
  if (!identical(errors, "none") && isTRUE(gradients))
    stop("plot.conmode intervals are available for class probabilities, not probability gradients/effects",
         call. = FALSE)
  data_rug <- npValidateScalarLogical(data_rug, "data_rug")
  renderer <- "base"
  if (!is.null(dots$renderer))
    renderer <- .np_plot_match_renderer(dots$renderer)
  perspective <- FALSE
  perspective.supplied <- !is.null(dots$perspective)
  persp.supplied <- !is.null(dots$persp)
  if (perspective.supplied)
    perspective <- npValidateScalarLogical(dots$perspective, "perspective")
  if (persp.supplied) {
    persp.value <- npValidateScalarLogical(dots$persp, "persp")
    if (perspective.supplied && !identical(perspective, persp.value))
      stop("conflicting plot.conmode arguments: 'perspective' and 'persp'",
           call. = FALSE)
    perspective <- persp.value
  }
  if (!is.null(dots$plot.vars) && !isTRUE(perspective))
    stop("'plot.vars' is only supported for fixed surface payloads with perspective=TRUE",
         call. = FALSE)
  if (!identical(renderer, "base") && !isTRUE(perspective))
    stop("renderer='rgl' is supported only for fixed surface payloads with perspective=TRUE",
         call. = FALSE)
  if (isTRUE(perspective) && isTRUE(gradients))
    stop("surface plot.conmode payloads are currently available for probabilities, not gradients",
         call. = FALSE)
  view <- dots$view
  if (is.null(view))
    view <- if (isTRUE(perspective) || length(grid.supplied)) "fixed" else "sample"
  view <- .np_plot_scalar_match(view, c("sample", "fixed"), "view")
  if (isTRUE(perspective) && !identical(view, "fixed"))
    stop("surface plot.conmode payloads require view='fixed'",
         call. = FALSE)
  neval <- dots$neval
  if (is.null(neval))
    neval <- 50L
  xtrim <- dots$xtrim
  xq <- dots$xq
  if (!is.null(dots$grid_control)) {
    if (!inherits(dots$grid_control, "np_grid_control"))
      stop("grid_control must be created by np_grid_control()", call. = FALSE)
    if (!is.null(dots$grid_control$slices))
      stop("grid_control$slices is not yet supported for plot.conmode",
           call. = FALSE)
    if (!is.null(dots$grid_control$xtrim))
      xtrim <- dots$grid_control$xtrim
    if (!is.null(dots$grid_control$xq))
      xq <- dots$grid_control$xq
  }
  if (is.null(xtrim))
    xtrim <- c(0, 1)
  plot.par.mfrow <- .np_plot_match_layout(layout)
  output <- match.arg(output)
  if (identical(output, "both"))
    output <- "plot-data"
  output <- .np_plot_scalar_match(output, c("plot", "data", "plot-data"), "output")
  if (!identical(errors, "none") && isTRUE(perspective) &&
      !identical(output, "data"))
    stop("surface probability intervals are available as output=\"data\" only; surface band rendering is not yet implemented for plot.conmode",
         call. = FALSE)
  if (identical(errors, "bootstrap") && isTRUE(perspective))
    stop("surface bootstrap intervals are not yet implemented for plot.conmode; use one-dimensional views or errors=\"asymptotic\" with output=\"data\"",
         call. = FALSE)
  plot.user.args <- .np_plot_user_args(dots, "plot")
  line.user.args <- .np_plot_user_args(dots, "lines")
  if (!is.null(dots$type))
    line.user.args$type <- dots$type

  plot.data <- if (isTRUE(perspective)) {
    .np_plot_conmode_surface_data(
      object,
      level = level,
      neval = neval,
      xtrim = xtrim,
      xq = xq,
      plot.vars = dots$plot.vars,
      errors = errors,
      alpha = alpha,
      band = band
    )
  } else if (identical(view, "fixed")) {
    .np_plot_conmode_grid_data(
      object,
      gradients = gradients,
      level = level,
      neval = neval,
      xtrim = xtrim,
      xq = xq,
      errors = errors,
      alpha = alpha,
      band = band,
      center = center,
      bootstrap = bootstrap,
      B = B,
      blocklen = boot.blocklen
    )
  } else {
    .np_plot_conmode_data(
      object,
      gradients = gradients,
      level = level,
      errors = errors,
      alpha = alpha,
      band = band,
      center = center,
      bootstrap = bootstrap,
      B = B,
      blocklen = boot.blocklen
    )
  }

  if (isTRUE(perspective)) {
    if (identical(output, "data"))
      return(plot.data)
    render.out <- .np_plot_conmode_surface_panel(plot.data, dots,
                                                 renderer = renderer)
    if (identical(renderer, "rgl"))
      return(.np_plot_rgl_finalize(
        rgl.out = render.out,
        plot.behavior = output,
        plot.data = plot.data
      ))
    if (identical(output, "plot-data"))
      return(plot.data)
    return(invisible(object))
  }
  level <- plot.data[[1L]]$level[1L]

  if (identical(output, "data"))
    return(plot.data)

  oldpar <- NULL
  if (isTRUE(plot.par.mfrow) && length(plot.data) > 1L) {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(oldpar), add = TRUE)
    graphics::par(mfrow = grDevices::n2mfrow(length(plot.data)))
  }

  for (i in seq_along(plot.data)) {
    panel.dots <- plot.user.args
    if (!is.null(dots$main) && length(plot.data) > 1L)
      panel.dots$main <- paste(dots$main, names(plot.data)[i], sep = ": ")
    .np_plot_conmode_panel(
      plot.data[[i]],
      gradients = gradients,
      level = level,
      plot.user.args = panel.dots,
      line.user.args = line.user.args,
      data_rug = data_rug
    )
  }

  if (identical(output, "plot-data"))
    return(plot.data)
  invisible(object)
}
.np_plot_singleindex <- function(object, ..., .plot_dots_call = NULL)
  .np_plot_from_slot(object, "bws", ...,
                     .plot_dots_call = .plot_dots_call,
                     .plot_context = "plot.singleindex")
.np_plot_smoothcoefficient <- function(object, ..., .plot_dots_call = NULL) {
  obj.bws <- .np_plot_restore_bandwidth_from_call(
    object = object,
    bws = object$bws,
    caller_env = parent.frame(2L)
  )
  .np_plot_validate_public_dots(
    .plot_dots_call,
    method = .np_plot_compat_dispatch,
    bws = obj.bws,
    context = "plot.smoothcoefficient"
  )
  dots <- list(...)
  dots <- .np_plot_normalize_public_dots(dots, context = "plot.smoothcoefficient")

  if (is.null(dots$xdat) && is.null(dots$ydat) && is.null(dots$zdat) &&
      isTRUE(object$trainiseval) &&
      !is.null(object$eval) &&
      !is.null(object$mean) &&
      !is.null(object$resid) &&
      length(object$mean) == length(object$resid) &&
      all(is.finite(object$mean)) &&
      all(is.finite(object$resid))) {
    if (is.list(object$eval) && !is.null(object$eval$exdat)) {
      dots$xdat <- object$eval$exdat
      if (!is.null(object$eval$ezdat))
        dots$zdat <- object$eval$ezdat
    } else {
      dots$xdat <- object$eval
    }
    dots$ydat <- object$mean + object$resid
  }

  do.call(.np_plot_call_method,
          c(list(method = .np_plot_compat_dispatch, bws = obj.bws), dots))
}
.np_plot_plregression <- function(object, ..., .plot_dots_call = NULL) {
  obj.bws <- .np_plot_restore_bandwidth_from_call(
    object = object,
    bws = .np_plreg_bws(object, where = "plot.plregression"),
    caller_env = parent.frame(2L)
  )
  .np_plot_validate_public_dots(
    .plot_dots_call,
    method = .np_plot_compat_dispatch,
    bws = obj.bws,
    context = "plot.plregression"
  )
  dots <- list(...)
  dots <- .np_plot_normalize_public_dots(dots, context = "plot.plregression")

  if (is.null(dots$xdat) && is.null(dots$ydat) && is.null(dots$zdat) &&
      isTRUE(object$trainiseval) &&
      !is.null(object$evalx) &&
      !is.null(object$evalz) &&
      !is.null(object$mean) &&
      !is.null(object$resid) &&
      NROW(object$evalx) == NROW(object$evalz) &&
      NROW(object$evalx) == length(object$mean) &&
      length(object$mean) == length(object$resid) &&
      all(is.finite(object$mean)) &&
      all(is.finite(object$resid))) {
    dots$xdat <- object$evalx
    dots$ydat <- object$mean + object$resid
    dots$zdat <- object$evalz
  }

  if (is.null(dots$xdat) && is.null(dots$ydat) && is.null(dots$zdat)) {
    training <- .np_plot_plreg_training_data(obj.bws)
    if (!is.null(training$xdat) &&
        !is.null(training$ydat) &&
        !is.null(training$zdat)) {
      dots$xdat <- training$xdat
      dots$ydat <- training$ydat
      dots$zdat <- training$zdat
    }
  }

  do.call(.np_plot_call_method,
          c(list(method = .np_plot_compat_dispatch, bws = obj.bws), dots))
}

plot.bandwidth <- function(x, ...)
  .np_plot_call_method(.np_plot_bandwidth_engine, bws = x, ...,
                       .plot_dots_call = match.call(expand.dots = FALSE)$...,
                       .plot_context = "plot.bandwidth")
plot.rbandwidth <- function(x, ...)
  .np_plot_call_method(.np_plot_rbandwidth_engine, bws = x, ...,
                       .plot_dots_call = match.call(expand.dots = FALSE)$...,
                       .plot_context = "plot.rbandwidth")
plot.dbandwidth <- function(x, ...)
  .np_plot_call_method(.np_plot_dbandwidth_engine, bws = x, ...,
                       .plot_dots_call = match.call(expand.dots = FALSE)$...,
                       .plot_context = "plot.dbandwidth")
plot.conbandwidth <- function(x, ...)
  .np_plot_call_method(.np_plot_conbandwidth_engine, bws = x, ...,
                       .plot_dots_call = match.call(expand.dots = FALSE)$...,
                       .plot_context = "plot.conbandwidth")
plot.condbandwidth <- function(x, ...)
  .np_plot_call_method(.np_plot_condbandwidth_engine, bws = x, ...,
                       .plot_dots_call = match.call(expand.dots = FALSE)$...,
                       .plot_context = "plot.condbandwidth")
plot.plbandwidth <- function(x, ...)
  .np_plot_call_method(.np_plot_plbandwidth_engine, bws = x, ...,
                       .plot_dots_call = match.call(expand.dots = FALSE)$...,
                       .plot_context = "plot.plbandwidth")
plot.sibandwidth <- function(x, ...)
  .np_plot_call_method(.np_plot_sibandwidth_engine, bws = x, ...,
                       .plot_dots_call = match.call(expand.dots = FALSE)$...,
                       .plot_context = "plot.sibandwidth")
plot.scbandwidth <- function(x, ...)
  .np_plot_call_method(.np_plot_scbandwidth_engine, bws = x, ...,
                       .plot_dots_call = match.call(expand.dots = FALSE)$...,
                       .plot_context = "plot.scbandwidth")

plot.npregression <- function(x, ...)
  .np_plot_npregression(x, ...,
                        .plot_dots_call = match.call(expand.dots = FALSE)$...)
plot.npdensity <- function(x, ...)
  .np_plot_npdensity(x, ...,
                     .plot_dots_call = match.call(expand.dots = FALSE)$...)
plot.condensity <- function(x, ...)
  .np_plot_condensity(x, ...,
                      .plot_dots_call = match.call(expand.dots = FALSE)$...)
plot.condistribution <- function(x, ...)
  .np_plot_condistribution(x, ...,
                           .plot_dots_call = match.call(expand.dots = FALSE)$...)
plot.npdistribution <- function(x, ...)
  .np_plot_npdistribution(x, ...,
                          .plot_dots_call = match.call(expand.dots = FALSE)$...)
plot.qregression <- function(x, ...)
  .np_plot_qregression(x, ...,
                       .plot_dots_call = match.call(expand.dots = FALSE)$...)
plot.conmode <- function(x, ...)
  .np_plot_conmode(x, ...,
                   .plot_dots_call = match.call(expand.dots = FALSE)$...)
plot.singleindex <- function(x, ...)
  .np_plot_singleindex(x, ...,
                       .plot_dots_call = match.call(expand.dots = FALSE)$...)
plot.smoothcoefficient <- function(x, ...)
  .np_plot_smoothcoefficient(x, ...,
                             .plot_dots_call = match.call(expand.dots = FALSE)$...)
plot.plregression <- function(x, ...)
  .np_plot_plregression(x, ...,
                        .plot_dots_call = match.call(expand.dots = FALSE)$...)

test_that("bootstrap coverage warning uses current public plot option names", {
  quantile_bounds <- getFromNamespace("compute.bootstrap.quantile.bounds", "np")

  set.seed(20260507L)
  boot.t <- matrix(rnorm(25L * 3L), nrow = 25L, ncol = 3L)
  msg <- NULL

  withCallingHandlers(
    quantile_bounds(
      boot.t = boot.t,
      alpha = 0.05,
      band.type = "all",
      warn.coverage = TRUE
    ),
    warning = function(w) {
      msg <<- conditionMessage(w)
      invokeRestart("muffleWarning")
    }
  )

  expect_type(msg, "character")
  expect_match(msg, "B=25", fixed = TRUE)
  expect_match(msg, "band=\"bonferroni/simultaneous/all\"", fixed = TRUE)
  expect_match(msg, "B >= ceiling(2*m/alpha - 1)", fixed = TRUE)
  expect_false(grepl("plot.errors.boot.num", msg, fixed = TRUE))
  expect_false(grepl("plot.errors.type", msg, fixed = TRUE))
})

test_that("common plot option validation reports current public names", {
  normalize <- getFromNamespace(".np_plot_normalize_common_options", "np")

  expect_error(
    normalize(
      plot.behavior = "data",
      plot.errors.method = "none",
      plot.errors.boot.method = "wild",
      plot.errors.boot.wild = "rademacher",
      plot.errors.boot.blocklen = NULL,
      plot.errors.center = "estimate",
      plot.errors.type = "pointwise",
      plot.errors.alpha = 0.5,
      plot.errors.style = "band",
      plot.errors.bar = "|",
      xdat = data.frame(x = 1:3),
      common.scale = FALSE,
      ylim = NULL
    ),
    "alpha must lie in \\(0, 0\\.5\\)"
  )

  msg <- NULL
  out <- withCallingHandlers(
    normalize(
      plot.behavior = "data",
      plot.errors.method = "none",
      plot.errors.boot.method = "wild",
      plot.errors.boot.wild = "rademacher",
      plot.errors.boot.blocklen = NULL,
      plot.errors.center = "estimate",
      plot.errors.type = "all",
      plot.errors.alpha = 0.05,
      plot.errors.style = "band",
      plot.errors.bar = "|",
      xdat = data.frame(x = 1:3),
      common.scale = FALSE,
      ylim = NULL
    ),
    warning = function(w) {
      msg <<- conditionMessage(w)
      invokeRestart("muffleWarning")
    }
  )

  expect_identical(out$plot.errors.method, "bootstrap")
  expect_match(msg, "band=\"all\" requires bootstrap errors", fixed = TRUE)
  expect_match(msg, "errors=\"bootstrap\"", fixed = TRUE)
  expect_false(grepl("plot.errors.type", msg, fixed = TRUE))
  expect_false(grepl("plot.errors.method", msg, fixed = TRUE))
})

test_that("snake_case plot controls normalize to engine controls", {
  normalize <- getFromNamespace(".np_plot_normalize_public_dots", "np")

  dots <- normalize(
    list(
      output = "both",
      data_overlay = FALSE,
      data_rug = TRUE,
      layout = "current",
      factor_boxplot = TRUE,
      boxplot_outliers = FALSE,
      coef_index = 2L,
      gradient_order = c(1L, 2L),
      common_scale = FALSE,
      proper_method = "isotonic",
      proper_control = list(mode = "slice"),
      errors = "asymptotic",
      render_control = np_render_control(style = "bar", bar = "I", bar_num = 7L)
    )
  )

  expect_identical(dots$plot.errors.method, "asymptotic")
  expect_identical(dots$plot.behavior, "plot-data")
  expect_false(dots$plot.data.overlay)
  expect_true(dots$plot.rug)
  expect_false(dots$plot.par.mfrow)
  expect_true(dots$plot.bxp)
  expect_false(dots$plot.bxp.out)
  expect_identical(dots$coef.index, 2L)
  expect_identical(dots$gradient.order, c(1L, 2L))
  expect_false(dots$common.scale)
  expect_identical(dots$proper.method, "isotonic")
  expect_identical(dots$proper.control, list(mode = "slice"))
  expect_identical(dots$plot.errors.style, "bar")
  expect_identical(dots$plot.errors.bar, "I")
  expect_identical(dots$plot.errors.bar.num, 7L)
})

test_that("snake_case plot controls fail cleanly on conflicts", {
  normalize <- getFromNamespace(".np_plot_normalize_public_dots", "np")

  expect_error(
    normalize(list(output = "data", plot.behavior = "plot")),
    "conflicting plot arguments: output and plot.behavior"
  )
  expect_error(
    normalize(list(data_rug = TRUE, plot.rug = FALSE)),
    "conflicting plot arguments: data_rug and plot.rug"
  )
  expect_error(
    normalize(list(layout = "maybe")),
    "layout must be one of"
  )
  expect_error(
    np_render_control(bar.num = 3L),
    "unused argument"
  )
})

library(np)

quiet_eval <- function(expr) {
  value <- NULL
  capture.output(value <- force(expr))
  value
}

test_that("npreg nonfixed frozen plot bootstrap supports continuous gradients", {
  run_case <- function(bwtype, boot.method) {
    set.seed(42)
    n <- 80L
    x <- runif(n)
    y <- x + rnorm(n)

    g <- quiet_eval(
      npreg(
        y ~ x,
        regtype = "ll",
        degree = c(1),
        bernstein = TRUE,
        bwtype = bwtype,
        nmulti = 1
      )
    )

    tf <- tempfile(fileext = ".pdf")
    grDevices::pdf(tf)
    on.exit({
      grDevices::dev.off()
      unlink(tf)
    }, add = TRUE)

    args <- list(
      x = g,
      neval = 20L,
      errors = "bootstrap",
      bootstrap = boot.method,
      boot_control = np_boot_control(nonfixed = "frozen"),
      B = 41L,
      band = "pointwise",
      gradients = TRUE
    )
    if (identical(boot.method, "geom"))
      args$boot_control <- np_boot_control(nonfixed = "frozen", blocklen = 4L)

    expect_no_error(do.call(plot, args))
  }

  for (bwtype in c("generalized_nn", "adaptive_nn")) {
    run_case(bwtype = bwtype, boot.method = "inid")
    run_case(bwtype = bwtype, boot.method = "geom")
  }
})

test_that("npcdens plot-data centers stay on predict and bootstrap offsets remain ordered", {
  skip_if_not_installed("np")

  check_mode <- function(bwtype, nonfixed_mode) {
    set.seed(42)
    n <- 100L
    x <- runif(n, -1, 1)
    y <- x^2 + rnorm(n, sd = 0.25 * stats::sd(x))

    bw <- npcdensbw(y ~ x, nmulti = 1L, bwtype = bwtype)
    fit <- npcdens(bws = bw)

    out <- suppressWarnings(plot(
      fit,
      output = "data",
      view = "fixed",
      perspective = FALSE,
      neval = 40L,
      errors = "bootstrap",
      bootstrap = "inid",
      B = 39L,
      band = "pointwise",
      boot_control = np_boot_control(nonfixed = nonfixed_mode)
    ))

    expect_true(length(out) > 0L)
    for (obj in out) {
      expect_s3_class(obj, "condensity")
      pred <- predict(fit, exdat = obj$xeval, eydat = obj$yeval)
      lower <- obj$condens + obj$conderr[, 1L]
      upper <- obj$condens + obj$conderr[, 2L]

      expect_equal(obj$condens, as.numeric(pred), tolerance = 1e-8)
      expect_true(all(lower <= upper + 1e-10, na.rm = TRUE))
    }

    out
  }

  fixed.exact <- check_mode("fixed", "exact")
  generalized.exact <- check_mode("generalized_nn", "exact")
  generalized.frozen <- check_mode("generalized_nn", "frozen")
  adaptive.exact <- check_mode("adaptive_nn", "exact")
  adaptive.frozen <- check_mode("adaptive_nn", "frozen")

  expect_equal(
    lapply(generalized.exact, `[[`, "condens"),
    lapply(generalized.frozen, `[[`, "condens"),
    tolerance = 1e-12
  )
  expect_equal(
    lapply(adaptive.exact, `[[`, "condens"),
    lapply(adaptive.frozen, `[[`, "condens"),
    tolerance = 1e-12
  )
  expect_true(all(vapply(fixed.exact, function(obj) is.matrix(obj$conderr), logical(1))))
})

.np_plbandwidth_child_degree <- function(bw, engine = FALSE) {
  field <- if (isTRUE(engine)) "degree.engine" else "degree"
  deg <- bw[[field]]
  if ((is.null(deg) || !length(deg)) && !is.null(bw$regtype)) {
    ncon <- if (is.null(bw$ncon) || !length(bw$ncon)) 0L else as.integer(bw$ncon[1L])
    if (identical(bw$regtype, "lc") && ncon > 0L)
      return(rep.int(0L, ncon))
    if (identical(bw$regtype, "ll") && ncon > 0L)
      return(rep.int(1L, ncon))
  }
  if (is.null(deg) || !length(deg))
    return(integer(0L))
  as.integer(deg)
}

.np_plbandwidth_format_degree <- function(degree) {
  degree <- as.integer(degree)
  if (!length(degree))
    return("none")
  paste(degree, collapse = ", ")
}

.np_plbandwidth_child_degree_block <- function(bw) {
  paste0("\nDegree: ", .np_plbandwidth_format_degree(
    .np_plbandwidth_child_degree(bw)
  ))
}

plbandwidth <-
  function(bws = stop("plbandwidth: bws missing"),
           regtype = c("lc","ll","lp"),
           basis = c("glp","additive","tensor"),
           degree = NULL,
           bernstein.basis = FALSE,
           bwmethod = c("cv.ls","cv.aic"),
           bwscaling = FALSE,
           bwtype = c("fixed","generalized_nn","adaptive_nn"),
           ckertype = c("gaussian","truncated gaussian","epanechnikov","uniform"), 
           ckerorder = c(2,4,6,8),
           ckerbound = c("none","range","fixed"),
           ckerlb = NULL,
           ckerub = NULL,
           ukertype = c("aitchisonaitken", "liracine"),
           okertype = c("liracine","wangvanryzin","racineliyan"),
           xdati, ydati, zdati,
           xnames, ynames, znames,
           nobs = NA,
           fval = NA,
           ifval = NA,
           num.feval = NA,
           num.feval.fast = NA,
           rows.omit = NA,
           bandwidth.compute = TRUE,
           total.time = NA,...){

    npRejectLegacyLpArgs(names(list(...)), where = "plbandwidth")
    spec <- npCanonicalConditionalRegSpec(
      regtype = regtype,
      basis = basis,
      degree = degree,
      bernstein.basis = bernstein.basis,
      ncon = sum(zdati$icon),
      where = "plbandwidth"
    )
    regtype <- spec$regtype
    basis <- spec$basis
    degree <- spec$degree
    bernstein.basis <- spec$bernstein.basis
    bwmethod = match.arg(bwmethod)
    bwtype = match.arg(bwtype)
    ckertype = match.arg(ckertype)
    ckerbound = match.arg(ckerbound)

    if(missing(ckerorder))
      ckerorder = 2
    else if (ckertype == "uniform")
      .np_warning("ignoring kernel order specified with uniform kernel type")
    else {
      kord = c(2,4,6,8) 
      if (!any(kord == ckerorder))
        stop("ckerorder must be one of ", paste(kord,collapse=" "))
    }

    if (ckertype == "truncated gaussian" && ckerorder != 2)
      .np_warning("using truncated gaussian of order 2, higher orders not yet implemented")

    ukertype = match.arg(ukertype)
    okertype = match.arg(okertype)
    cbounds <- npKernelBoundsResolve(
      dati = zdati,
      varnames = znames,
      kerbound = ckerbound,
      kerlb = ckerlb,
      kerub = ckerub,
      argprefix = "cker")
    ncon <- sum(zdati$icon)
    if (identical(spec$regtype.engine, "lp") && ncon > 0L && is.finite(nobs)) {
      lp.dim <- dim_basis(basis = spec$basis.engine,
                          kernel = TRUE,
                          degree = spec$degree.engine,
                          segments = rep.int(1L, ncon))
      if (is.finite(lp.dim) && lp.dim > (nobs - 1.0))
        stop(sprintf("LP basis dimension (%s) exceeds nobs - 1 (%s); reduce degree",
                     format(lp.dim, trim = TRUE, scientific = FALSE),
                     format(nobs - 1.0, trim = TRUE, scientific = FALSE)))
    }

    porder = switch( ckerorder/2, "Second-Order", "Fourth-Order", "Sixth-Order",
      "Eighth-Order" )
    ## chug chug
    sfactor <- lapply(seq_along(bws), function(i) { unlist(bws[[i]]$sfactor) })
    bandwidth <- lapply(seq_along(bws), function(i) { unlist(bws[[i]]$bandwidth) })
    sumNum <- lapply(seq_along(bws), function(i) { unlist(bws[[i]]$sumNum) })
    child.degree <- lapply(bws, .np_plbandwidth_child_degree)
    child.degree.engine <- lapply(bws, .np_plbandwidth_child_degree, engine = TRUE)
    degree.keys <- vapply(child.degree, paste, collapse = ",", character(1L))
    child.degree.common <- length(unique(degree.keys)) <= 1L

    names(sfactor) <- names(bandwidth) <- names(sumNum) <- rep("z", length(bws))

    if (length(rows.omit) == 0)
      rows.omit <- NA

    mybw = list(
      bw=bws,
      regtype = regtype,
      pregtype = switch(regtype,
        lc = "Local-Constant",
        ll = "Local-Linear",
        lp = "Local-Polynomial"),
      basis = basis,
      degree = degree,
      child.degree = child.degree,
      child.degree.engine = child.degree.engine,
      child.degree.common = child.degree.common,
      bernstein.basis = bernstein.basis,
      regtype.engine = spec$regtype.engine,
      basis.engine = spec$basis.engine,
      degree.engine = spec$degree.engine,
      bernstein.basis.engine = spec$bernstein.basis.engine,
      method = bwmethod,
      pmethod = bwmToPrint(bwmethod),
      scaling = bwscaling,
      pscaling = npBandwidthSummaryLabel(bwtype = bwtype, bwscaling = bwscaling),
      type = bwtype,
      ptype = bwtToPrint(bwtype),
      ckertype = ckertype,    
      ckerorder = ckerorder,
      ckerbound = cbounds$bound,
      ckerlb = cbounds$lb,
      ckerub = cbounds$ub,
      pckertype = cktToPrint(ckertype, order = porder, kerbound = cbounds$bound),
      ukertype = ukertype,
      pukertype = uktToPrint(ukertype),
      okertype = okertype,
      pokertype = oktToPrint(okertype),
      nobs = nobs,
      zndim = bws$yzbw$ndim,
      xndim = length(bws)-1,
      ncon = ncon,
      nuno = sum(zdati$iuno),
      nord = sum(zdati$iord),
      icon = zdati$icon,
      iuno = zdati$iuno,
      iord = zdati$iord,
      xdati = xdati,
      ydati = ydati,
      zdati = zdati,
      xnames = xnames,
      ynames = ynames,
      znames = znames,
      fval = fval,
      ifval = ifval,
      num.feval = num.feval,
      num.feval.fast = num.feval.fast,
      sfactor = sfactor,
      bandwidth = bandwidth,
      sumNum = sumNum,
      varnames = list(x = xnames, y = ynames, z = znames),
      vartitle = list(x = "Explanatory", y = "Dependent", z = "Explanatory"),
      vartitleabb = list(x = "Exp.", y = "Dep.", z = "Exp."),
      dati = list(x = xdati, y = ydati, z = zdati),
      rows.omit = rows.omit,
      nobs.omit = if (identical(rows.omit, NA)) 0 else length(rows.omit),
      total.time = total.time)

    mybw$klist = list(z = list(ckertype = ckertype,
                        ckerbound = cbounds$bound,
                        ckerlb = cbounds$lb,
                        ckerub = cbounds$ub,
                        pckertype = mybw$pckertype,
                        ukertype = ukertype,
                        pukertype = mybw$pukertype,
                        okertype = okertype,
                        pokertype = mybw$pokertype))

    if(!bandwidth.compute)
      mybw$pmethod <- "Manual"

    class(mybw) = "plbandwidth"
    if(!any(is.na(mybw$bandwidth)))
      validateBandwidth(mybw)
    mybw
    
  }

print.plbandwidth <- function(x, digits=NULL, ...){
  cat("\nPartially Linear Model",
      "\nRegression data (",x$nobs," observations,\n",
      x$xndim," linear parametric variable(s) ",
      x$zndim," nonparametric variable(s):\n",sep="")

  bwmat = matrix(data = 0, nrow = x$xndim+1, ncol = x$bw$yzbw$ndim)
  
  for (i in seq_along(x$bw))
    bwmat[i,] = x$bw[[i]]$bw

  child.labels <- c(x$ynames, x$xnames)
  child.degrees <- vapply(x$bw, function(bwi) {
    .np_plbandwidth_format_degree(.np_plbandwidth_child_degree(bwi))
  }, character(1L))
  
  ## perhaps add a column for objective function value?
  print(matrix(bwmat[1,], ncol=x$zndim,
               dimnames=list(paste(x$pscaling,":",sep=""),
                 c("y(z)", replicate(x$zndim-1,"")))))
  cat("\n")
  print(matrix(bwmat[2:(1+x$xndim),], ncol=x$zndim,
               dimnames=list(c(paste(x$pscaling,":",sep=""), replicate(x$xndim-1,"")),
                 c("x(z)", replicate(x$zndim-1,"")))))
  cat("\n")
  print(matrix(child.degrees,
               ncol = 1L,
               dimnames = list(child.labels, "Degree")))

  cat(genBwSelStr(x))
  cat(genBwKerStrs(x))

  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

predict.plbandwidth <- function(object, ...) { npplreg(bws = object, ...) }

summary.plbandwidth <- function(object, ...){
  cat("\nPartially Linear Model",
      "\nRegression data (",object$nobs," observations,\n",
      object$xndim," linear parametric variable(s) ",
      object$zndim," nonparametric variable(s):\n",sep="")

  cat(genOmitStr(object))
  cat(genBwSelStr(object))

  child.labels <- c(object$ynames, object$xnames)
  scale.blocks <- vapply(
    seq_along(object$bw),
    function(i) {
      paste0("\n\n", child.labels[[i]], " on z:",
             .np_plbandwidth_child_degree_block(object$bw[[i]]),
             paste(genBwScaleStrs(object$bw[[i]]), collapse = ""))
    },
    character(1L)
  )

  cat(scale.blocks)
  
  cat(genBwKerStrs(object))
  cat(genTimingStr(object))
  cat("\n\n")  
}

np_namespace_hygiene_root <- function() {
  starts <- unique(Filter(
    nzchar,
    c(
      tryCatch(normalizePath(testthat::test_path("..", ".."), mustWork = TRUE), error = function(e) ""),
      tryCatch(normalizePath(getwd(), mustWork = TRUE), error = function(e) ""),
      tryCatch(normalizePath(file.path(getwd(), ".."), mustWork = TRUE), error = function(e) ""),
      tryCatch(normalizePath(file.path(getwd(), "..", ".."), mustWork = TRUE), error = function(e) "")
    )
  ))

  seen <- character()
  for (start in starts) {
    candidate <- start
    repeat {
      if (!nzchar(candidate) || candidate %in% seen)
        break
      search_paths <- unique(c(candidate, file.path(candidate, "00_pkg_src", "np")))

      for (path in search_paths) {
        if (!nzchar(path) || path %in% seen)
          next
        seen <- c(seen, path)

        desc <- file.path(path, "DESCRIPTION")
        if (!file.exists(desc))
          next
        dcf <- tryCatch(read.dcf(desc), error = function(e) NULL)
        if (!is.null(dcf) &&
            nrow(dcf) &&
            identical(unname(dcf[1L, "Package"]), "np")) {
          return(path)
        }
      }

      parent <- dirname(candidate)
      if (identical(parent, candidate))
        break
      candidate <- parent
    }
  }

  stop("Could not locate np package root for namespace hygiene checks", call. = FALSE)
}

np_namespace_hygiene_scan <- function(root = np_namespace_hygiene_root(), pkg = "np") {
  root <- normalizePath(root, mustWork = TRUE)

  surface_for <- function(rel) {
    if (grepl("^tests/", rel))
      return("tests")
    if (grepl("^(R|inst|demo)/", rel))
      return("runtime")
    "other"
  }

  is_r_like_file <- function(rel) {
    grepl("\\.R$", rel) | basename(rel) %in% c("Rprofile", ".Rprofile")
  }

  all <- list.files(root, recursive = TRUE, full.names = TRUE, all.files = FALSE, no.. = TRUE)
  all <- all[file.info(all)$isdir %in% FALSE]
  rel <- sub(paste0("^", root, "/?"), "", normalizePath(all, mustWork = TRUE))
  keep <- is_r_like_file(rel) & grepl("^(R|inst|demo|tests)/", rel)

  files <- data.frame(
    path = normalizePath(all[keep], mustWork = TRUE),
    rel = rel[keep],
    surface = vapply(rel[keep], surface_for, character(1L)),
    stringsAsFactors = FALSE
  )

  parse_ops <- function(path) {
    expr <- parse(path, keep.source = TRUE)
    pd <- getParseData(expr, includeText = TRUE)
    if (is.null(pd) || !nrow(pd))
      return(pd[FALSE, , drop = FALSE])
    pd
  }

  out <- lapply(seq_len(nrow(files)), function(i) {
    info <- files[i, , drop = FALSE]
    pd <- parse_ops(info$path)
    if (!nrow(pd))
      return(list(ns = NULL, str = NULL))

    ns_rows <- pd[pd$token %in% c("NS_GET", "NS_GET_INT"), , drop = FALSE]
    str_rows <- pd[pd$token == "STR_CONST" & grepl("::", pd$text, fixed = TRUE), , drop = FALSE]

    ns_out <- lapply(seq_len(nrow(ns_rows)), function(j) {
      row <- ns_rows[j, , drop = FALSE]
      siblings <- pd[pd$parent == row$parent, , drop = FALSE]
      pkg_row <- siblings[siblings$token == "SYMBOL_PACKAGE", , drop = FALSE]
      sym_row <- siblings[siblings$token %in% c("SYMBOL", "SYMBOL_FUNCTION_CALL", "STR_CONST"), , drop = FALSE]
      if (!nrow(pkg_row) || !nrow(sym_row))
        return(NULL)
      data.frame(
        rel = info$rel,
        surface = info$surface,
        token = row$token,
        pkg = pkg_row$text[[1L]],
        sym = sym_row$text[[1L]],
        line = row$line1[[1L]],
        col = row$col1[[1L]],
        stringsAsFactors = FALSE
      )
    })
    ns_out <- do.call(rbind, ns_out[!vapply(ns_out, is.null, logical(1L))])

    str_out <- if (nrow(str_rows)) {
      data.frame(
        rel = info$rel,
        surface = info$surface,
        token = str_rows$token,
        pkg = "",
        sym = str_rows$text,
        line = str_rows$line1,
        col = str_rows$col1,
        stringsAsFactors = FALSE
      )
    } else {
      NULL
    }

    list(ns = ns_out, str = str_out)
  })

  ns_df <- do.call(rbind, lapply(out, `[[`, "ns"))
  str_df <- do.call(rbind, lapply(out, `[[`, "str"))

  if (is.null(ns_df)) {
    ns_df <- data.frame(
      rel = character(),
      surface = character(),
      token = character(),
      pkg = character(),
      sym = character(),
      line = integer(),
      col = integer(),
      stringsAsFactors = FALSE
    )
  }

  if (is.null(str_df)) {
    str_df <- data.frame(
      rel = character(),
      surface = character(),
      token = character(),
      pkg = character(),
      sym = character(),
      line = integer(),
      col = integer(),
      stringsAsFactors = FALSE
    )
  }

  runtime_ns <- ns_df[ns_df$surface == "runtime", , drop = FALSE]
  tests_ns <- ns_df[ns_df$surface == "tests", , drop = FALSE]

  list(
    files = files,
    runtime_same_package = runtime_ns[runtime_ns$pkg == pkg, , drop = FALSE],
    runtime_external_triple = runtime_ns[runtime_ns$pkg != pkg & runtime_ns$token == "NS_GET_INT", , drop = FALSE],
    runtime_external_double = runtime_ns[runtime_ns$pkg != pkg & runtime_ns$token == "NS_GET", , drop = FALSE],
    runtime_parser_literals = str_df[str_df$surface == "runtime", , drop = FALSE],
    tests_same_package = tests_ns[tests_ns$pkg == pkg, , drop = FALSE],
    tests_external_triple = tests_ns[tests_ns$pkg != pkg & tests_ns$token == "NS_GET_INT", , drop = FALSE]
  )
}

np_namespace_hygiene_format <- function(df) {
  if (!nrow(df))
    return("<none>")
  paste(
    sprintf(
      "%s:%d:%d %s%s%s",
      df$rel, df$line, df$col,
      ifelse(nzchar(df$pkg), paste0(df$pkg, ifelse(df$token == "NS_GET_INT", ":::", "::")), ""),
      ifelse(df$token == "STR_CONST", "", df$sym),
      ifelse(df$token == "STR_CONST", df$sym, "")
    ),
    collapse = "\n"
  )
}

test_that("namespace hygiene scanner distinguishes defects from allowed patterns", {
  mkroot <- function() {
    root <- tempfile("np-namespace-hygiene-")
    dir.create(root, recursive = TRUE, showWarnings = FALSE)
    writeLines(c("Package: np", "Version: 0.0.0"), file.path(root, "DESCRIPTION"))
    root
  }

  good <- mkroot()
  dir.create(file.path(good, "R"), recursive = TRUE, showWarnings = FALSE)
  writeLines(
    c(
      "f <- function(x) stats::na.omit(x)",
      "msg <- 'parser literal ::: should not fail the gate'"
    ),
    file.path(good, "R", "good.R")
  )
  good.scan <- np_namespace_hygiene_scan(good)
  expect_identical(nrow(good.scan$runtime_same_package), 0L)
  expect_identical(nrow(good.scan$runtime_external_triple), 0L)
  expect_identical(nrow(good.scan$runtime_external_double), 1L)
  expect_identical(nrow(good.scan$runtime_parser_literals), 1L)

  bad.same <- mkroot()
  dir.create(file.path(bad.same, "R"), recursive = TRUE, showWarnings = FALSE)
  writeLines(
    "np:::.hidden_helper()",
    file.path(bad.same, "R", "bad_same.R")
  )
  bad.same.scan <- np_namespace_hygiene_scan(bad.same)
  expect_identical(nrow(bad.same.scan$runtime_same_package), 1L)
  expect_match(np_namespace_hygiene_format(bad.same.scan$runtime_same_package), "R/bad_same.R")

  bad.external <- mkroot()
  dir.create(file.path(bad.external, "demo"), recursive = TRUE, showWarnings = FALSE)
  writeLines(
    "f <- function(x) stats:::na.omit(x)",
    file.path(bad.external, "demo", "bad_external.R")
  )
  bad.external.scan <- np_namespace_hygiene_scan(bad.external)
  expect_identical(nrow(bad.external.scan$runtime_external_triple), 1L)
  expect_match(np_namespace_hygiene_format(bad.external.scan$runtime_external_triple), "demo/bad_external.R")
})

test_that("runtime package surfaces avoid forbidden namespace qualification", {
  scan <- np_namespace_hygiene_scan()

  expect_identical(
    nrow(scan$runtime_same_package),
    0L,
    info = np_namespace_hygiene_format(scan$runtime_same_package)
  )
  expect_identical(
    nrow(scan$runtime_external_triple),
    0L,
    info = np_namespace_hygiene_format(scan$runtime_external_triple)
  )
})

npreg <-
  function(bws, ...){
    args <- list(...)

    if (!missing(bws)){
      if (inherits(bws, "formula") && is.null(args$txdat))
        UseMethod("npreg", bws)
      if (is.recursive(bws)){
        if (!is.null(bws$formula) && is.null(args$txdat))
          UseMethod("npreg",bws$formula)
        else if (!is.null(bws$call) && is.null(args$txdat))
          UseMethod("npreg",bws$call)
        else if (!is.call(bws))
          UseMethod("npreg",bws)
        else
          UseMethod("npreg",NULL)
      } else {
        UseMethod("npreg", NULL)
      }
    } else {
      UseMethod("npreg", NULL)
    }
  }

npreg.formula <-
  function(bws, data = NULL, newdata = NULL, y.eval = FALSE, ...){

    tt <- terms(bws)
    tmf <- if (!is.null(bws$call)) {
      m <- match(c("formula", "data", "subset", "na.action"),
                 names(bws$call), nomatch = 0)
      bws$call[c(1, m)]
    } else {
      mc <- match.call(expand.dots = FALSE)
      m <- match(c("bws", "data", "subset", "na.action"),
                 names(mc), nomatch = 0)
      tmf <- mc[c(1, m)]
      if ("bws" %in% names(tmf))
        names(tmf)[names(tmf) == "bws"] <- "formula"
      tmf
    }
    tmf[[1]] <- as.name("model.frame")
    tmf[["formula"]] <- tt
    mf.args <- as.list(tmf)[-1L]
    umf <- tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

    response.name <- attr(tmf, "names")[attr(attr(tmf, "terms"), "response")]
    tydat <- model.response(tmf)
    txdat <- tmf[, attr(attr(tmf, "terms"),"term.labels"), drop = FALSE]
    has.eval <- !is.null(newdata)
    if (has.eval) {
      if (!y.eval){
        npValidateNewdataFormula(newdata, tt, include.response = FALSE)
        tt <- delete.response(tt)

        orig.ts <- .np_terms_ts_mask(terms_obj = tt, data = newdata)
        
        ## delete.response clobbers predvars, which is used for timeseries objects
        ## so we need to reconstruct it

        if(all(orig.ts)){
          args <- (as.list(attr(tt, "variables"))[-1])
          attr(tt, "predvars") <- as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), args))))
        }else if(any(orig.ts)){
          arguments <- (as.list(attr(tt, "variables"))[-1])
          arguments.normal <- arguments[which(!orig.ts)]
          arguments.timeseries <- arguments[which(orig.ts)]

          ix <- sort(c(which(orig.ts),which(!orig.ts)),index.return = TRUE)$ix
          attr(tt, "predvars") <- bquote(.(as.call(c(quote(cbind),as.call(c(quote(as.data.frame),as.call(c(quote(ts.intersect), arguments.timeseries)))),arguments.normal,check.rows = TRUE)))[,.(ix)])
        }else{
          attr(tt, "predvars") <- attr(tt, "variables")
        }
      }
      
      if (y.eval)
        npValidateNewdataFormula(newdata, tt, include.response = TRUE)
      umf.args <- list(formula = tt, data = newdata)
      umf <- do.call(stats::model.frame, umf.args, envir = parent.frame())
      emf <- umf

      if (y.eval)
        eydat <- model.response(emf)
      
      exdat <- emf[, attr(attr(emf, "terms"),"term.labels"), drop = FALSE]
    }

    reg.args <- list(txdat = txdat, tydat = tydat, bws = bws)
    if (has.eval) {
      reg.args$exdat <- exdat
      if (y.eval)
        reg.args$eydat <- eydat
    }
    ev <- do.call(npreg, c(reg.args, list(...)))
    ev$call <- match.call(expand.dots = FALSE)
    environment(ev$call) <- parent.frame()

    if (length(response.name) == 1L && !is.na(response.name) && nzchar(response.name)) {
      if (!is.null(ev$bws))
        ev$bws$ynames <- response.name
    }

    ev$omit <- attr(umf,"na.action")
    ev$rows.omit <- as.vector(ev$omit)
    ev$nobs.omit <- length(ev$rows.omit)

    ev$mean <- napredict(ev$omit, ev$mean)
    ev$merr <- napredict(ev$omit, ev$merr)

    if(ev$gradients){
        ev$grad <- napredict(ev$omit, ev$grad)
        ev$gerr <- napredict(ev$omit, ev$gerr)
    }

    if(ev$residuals){
        ev$resid <- naresid(ev$omit, ev$resid)
    }    
    return(ev)
  }

npreg.call <-
  function(bws, ...) {
    ev <- npreg(txdat = .np_eval_bws_call_arg(bws, "xdat"),
                tydat = .np_eval_bws_call_arg(bws, "ydat"),
                bws = bws, ...)
    ev$call <- match.call(expand.dots = FALSE)
    environment(ev$call) <- parent.frame()
    return(ev)
  }

.np_reg_fit_total <- function(bws, tnrow, enrow) {
  if (identical(as.character(bws$type)[1L], "adaptive_nn")) {
    as.integer(tnrow)
  } else {
    as.integer(enrow)
  }
}

npreg.rbandwidth <-
  function(bws,
           txdat = stop("training data 'txdat' missing"),
           tydat = stop("training data 'tydat' missing"),
           exdat, eydat, gradient.order = 1L, gradients = FALSE,
           residuals = FALSE,
           ...){
    fit.start <- proc.time()[3]

    no.ex = missing(exdat)
    no.ey = missing(eydat)
    dots <- list(...)
    fit.progress.handoff <- isTRUE(dots$.np_fit_progress_handoff)
    if ("remin" %in% names(dots)) {
      warning("npreg: bandwidth-selection argument 'remin' is ignored when a bandwidth object is supplied",
              call. = FALSE)
      dots$remin <- NULL
    }
    npRejectLegacyLpArgs(names(dots), where = "npreg")
    warn.glp.gradient <- if (is.null(dots$warn.glp.gradient)) TRUE else isTRUE(dots$warn.glp.gradient)

    gradients <- npValidateScalarLogical(gradients, "gradients")
    residuals <- npValidateScalarLogical(residuals, "residuals")

    txdat = toFrame(txdat)

    if (!(is.vector(tydat) || is.factor(tydat)))
      stop("'tydat' must be a vector or a factor")


    ## if no.ex then if !no.ey then ey and tx must match, to get oos errors
    ## alternatively if no.ey you get is errors
    ## if !no.ex then if !no.ey then ey and ex must match, to get oos errors
    ## alternatively if no.ey you get NO errors since we don't evaluate on the training
    ## data
    
     
    if (!no.ex){
      exdat = toFrame(exdat)

      if (! txdat %~% exdat )
        stop("'txdat' and 'exdat' are not similar data frames!")

      if (!no.ey){
        if (!(is.vector(eydat) || is.factor(eydat)))
          stop("'eydat' must be a vector or a factor")
        if (dim(exdat)[1] != length(eydat))
          stop("number of evaluation data 'exdat' and dependent data 'eydat' do not match")
        if (!identical(coarseclass(eydat),coarseclass(tydat)))
          stop("type of evaluation data 'eydat' does not match that of 'tydat'")
      }
      
    } else if(!no.ey) {
      if (dim(txdat)[1] != length(eydat))
        stop("number of training data 'txdat' and dependent data 'eydat' do not match")
    }

    if (length(bws$bw) != length(txdat))
      stop("length of bandwidth vector does not match number of columns of 'txdat'")

    npValidateRegressionNnLowerBound(bws, where = "npreg")

    bws$basis <- npValidateLpBasis(regtype = bws$regtype,
                                   basis = bws$basis)
    bws$degree <- npValidateGlpDegree(regtype = bws$regtype,
                                      degree = bws$degree,
                                      ncon = bws$ncon)
    bws$bernstein.basis <- npValidateGlpBernstein(regtype = bws$regtype,
                                                  bernstein.basis = bws$bernstein.basis)
    reg.spec <- npCanonicalConditionalRegSpec(
      regtype = bws$regtype,
      basis = bws$basis,
      degree = bws$degree,
      bernstein.basis = bws$bernstein.basis,
      ncon = bws$ncon,
      where = "npreg"
    )
    glp.gradient.order <- if (identical(reg.spec$regtype.engine, "lp")) {
      if (identical(bws$regtype, "lp")) {
        npValidateGlpGradientOrder(regtype = bws$regtype,
                                   gradient.order = gradient.order,
                                   ncon = bws$ncon)
      } else if (bws$ncon > 0L) {
        rep.int(1L, bws$ncon)
      } else {
        integer(0)
      }
    } else {
      NULL
    }
    if (isTRUE(gradients) &&
        identical(reg.spec$regtype.engine, "lp") &&
        (bws$ncon > 0L) &&
        all(reg.spec$degree.engine == 0L)) {
      stop("regtype='lp' with degree=0 does not support derivatives; use gradients=FALSE for fitted/predicted values")
    }

    reg.c <- npRegtypeToC(regtype = reg.spec$regtype.engine,
                          degree = reg.spec$degree.engine,
                          ncon = bws$ncon,
                          context = "npreg")
    degree.c <- if (bws$ncon > 0) {
      as.integer(if (is.null(reg.c$degree)) rep.int(0L, bws$ncon) else reg.c$degree)
    } else {
      integer(1)
    }

    if ((any(bws$icon) &&
         !all(vapply(txdat[, bws$icon, drop = FALSE], inherits, logical(1), c("integer", "numeric")))) ||
        (any(bws$iord) &&
         !all(vapply(txdat[, bws$iord, drop = FALSE], inherits, logical(1), "ordered"))) ||
        (any(bws$iuno) &&
         !all(vapply(txdat[, bws$iuno, drop = FALSE], inherits, logical(1), "factor"))))
      stop("supplied bandwidths do not match 'txdat' in type")

    if (dim(txdat)[1] != length(tydat))
      stop("number of explanatory data 'txdat' and dependent data 'tydat' do not match")

    ## catch and destroy NA's
    keep.rows <- rep_len(TRUE, nrow(txdat))
    rows.omit <- attr(na.omit(data.frame(txdat,tydat)), "na.action")
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Training data has no rows without NAs")

    txdat <- txdat[keep.rows,,drop = FALSE]
    tydat <- tydat[keep.rows]

    ## no.ex = missing(exdat)
    ## no.ey = missing(eydat)

    if (!no.ex){
      keep.eval <- rep_len(TRUE, nrow(exdat))
      eval.df <- data.frame(exdat)
      if (!no.ey)
        eval.df <- data.frame(eval.df, eydat)
      rows.omit <- attr(na.omit(eval.df), "na.action")
      if (length(rows.omit) > 0L)
        keep.eval[as.integer(rows.omit)] <- FALSE

      if (!any(keep.eval))
        stop("Evaluation data has no rows without NAs")

      exdat <- exdat[keep.eval,,drop = FALSE]
      if (!no.ey)
        eydat <- eydat[keep.eval]
    }

    if (identical(bws$regtype, "lp") &&
        isTRUE(bws$bernstein.basis) &&
        !no.ex &&
        any(bws$icon)) {
      out.of.support <- character(0)
      for (ii in which(bws$icon)) {
        tr <- range(as.numeric(txdat[[ii]]))
        ex <- as.numeric(exdat[[ii]])
        if (any(ex < tr[1] | ex > tr[2], na.rm = TRUE))
          out.of.support <- c(out.of.support, colnames(txdat)[ii])
      }
      if (length(out.of.support)) {
        .np_warning(
          "bernstein.basis=TRUE: evaluation continuous predictor(s) outside training support (",
          paste(unique(out.of.support), collapse = ", "),
          "); proceeding, but interpret extrapolated results with care",
          call. = FALSE
        )
      }
    }

    ## evaluate residuals before data conversion ...

    if (residuals){
      resid <- tydat - npreg(txdat = txdat, tydat = tydat, bws = bws)$mean
    }


    tnrow = dim(txdat)[1]
    enrow = (if (no.ex) tnrow else dim(exdat)[1])
    ncol = dim(txdat)[2]

    ## convert tydat, eydat to numeric, from a factor with levels from the y-data
    ## used during bandwidth selection.
    
    if (is.factor(tydat)){
      tydat <- adjustLevels(data.frame(tydat), bws$ydati)[,1]
      tydat <- (bws$ydati$all.dlev[[1]])[as.integer(tydat)]
    }
    else
      tydat <- as.double(tydat)


    if (no.ey)
      eydat <- double()
    else {
      if (is.factor(eydat)){
        eydat <- adjustLevels(data.frame(eydat), bws$ydati, allowNewCells = TRUE)
        eydat <- toMatrix(eydat)[,1]
      }
      else
        eydat <- as.double(eydat)
    }

    ## re-assign levels in training and evaluation data to ensure correct
    ## conversion to numeric type.
    
    txdat <- adjustLevels(txdat, bws$xdati)
      
    if (!no.ex)
      exdat <- adjustLevels(exdat, bws$xdati, allowNewCells = TRUE)

    if (!no.ex)
      npKernelBoundsCheckEval(exdat, bws$icon, bws$ckerlb, bws$ckerub, argprefix = "cker")

    ## grab the evaluation data before it is converted to numeric
    if(no.ex)
      teval <- txdat
    else
      teval <- exdat

    ## put the unordered, ordered, and continuous data in their own objects
    ## data that is not a factor is continuous.

    txdat = toMatrix(txdat)

    tuno = txdat[, bws$iuno, drop = FALSE]
    tcon = txdat[, bws$icon, drop = FALSE]
    tord = txdat[, bws$iord, drop = FALSE]

    npCheckRegressionDesignCondition(reg.code = reg.c$code,
                                     xcon = tcon,
                                     basis = reg.spec$basis.engine,
                                     degree = reg.spec$degree.engine,
                                     bernstein.basis = reg.spec$bernstein.basis.engine,
                                     where = "npreg")

    if (!no.ex){
      exdat = toMatrix(exdat)

      euno = exdat[, bws$iuno, drop = FALSE]
      econ = exdat[, bws$icon, drop = FALSE]
      eord = exdat[, bws$iord, drop = FALSE]

    } else {
      euno = data.frame()
      eord = data.frame()
      econ = data.frame()
    }

    myopti = list(
      num_obs_train = tnrow,
      num_obs_eval = enrow,
      num_uno = bws$nuno, num_ord = bws$nord,
      num_con = bws$ncon,
      int_LARGE_SF = (if (bws$scaling) SF_NORMAL else SF_ARB),
      BANDWIDTH_reg_extern = switch(bws$type,
        fixed = BW_FIXED,
        generalized_nn = BW_GEN_NN,
        adaptive_nn = BW_ADAP_NN),
      int_MINIMIZE_IO=if (isTRUE(getOption("np.messages"))) IO_MIN_FALSE else IO_MIN_TRUE, 
      kerneval = switch(bws$ckertype,
        gaussian = CKER_GAUSS + bws$ckerorder/2 - 1,
        epanechnikov = CKER_EPAN + bws$ckerorder/2 - 1,
        uniform = CKER_UNI,
        "truncated gaussian" = CKER_TGAUSS),
      ukerneval = switch(bws$ukertype,
        aitchisonaitken = UKER_AIT,
        liracine = UKER_LR),
      okerneval = switch(bws$okertype,
        wangvanryzin = OKER_WANG,
        liracine = OKER_LR,
        "racineliyan" = OKER_RLY),
      ey_is_ty = no.ey,
      do_grad = gradients,
      regtype = reg.c$code,
      no.ex = no.ex,
      mcv.numRow = attr(bws$xmcv, "num.row"),
      int_do_tree = if (isTRUE(getOption("np.tree"))) DO_TREE_YES else DO_TREE_NO,
      old.reg = FALSE)

    cker.bounds.c <- npKernelBoundsMarshal(bws$ckerlb[bws$icon], bws$ckerub[bws$icon])
    
   asDouble <- function(data){
	   if (is.null(data)){
	 	   result <- as.double(0.0)
	   }
	   else {
		   result <- as.double(data)
	   }
	   return(result)
   }


    glp.gradient.order.c <- if (bws$ncon > 0L) {
      as.integer(if (is.null(glp.gradient.order)) rep.int(1L, bws$ncon) else glp.gradient.order)
    } else {
      integer(1)
    }

    myout <- .np_with_compiled_fit_progress(
      label = "Fitting regression",
      total = .np_reg_fit_total(bws = bws, tnrow = tnrow, enrow = enrow),
      handoff = fit.progress.handoff,
      handoff.detail = if (fit.progress.handoff) "starting" else NULL,
      .Call("C_np_regression",
            asDouble(tuno), asDouble(tord), asDouble(tcon), asDouble(tydat),
            asDouble(euno), asDouble(eord), asDouble(econ), asDouble(eydat),
            asDouble(c(bws$bw[bws$icon], bws$bw[bws$iuno], bws$bw[bws$iord])),
            asDouble(bws$xmcv), asDouble(attr(bws$xmcv, "pad.num")),
            asDouble(bws$nconfac), asDouble(bws$ncatfac), asDouble(bws$sdev),
            as.integer(myopti),
            as.integer(degree.c),
            as.integer(glp.gradient.order.c),
            as.integer(isTRUE(reg.spec$bernstein.basis.engine)),
            as.integer(npLpBasisCode(reg.spec$basis.engine)),
            as.integer(enrow),
            as.integer(ncol),
            as.logical(gradients),
            as.double(cker.bounds.c$lb),
            as.double(cker.bounds.c$ub),
            PACKAGE = "np")
    )

    if (gradients){
      myout$g = matrix(data=myout$g, nrow = enrow, ncol = ncol, byrow = FALSE) 
      rorder = numeric(ncol)
      ord_idx <- seq_len(ncol)
      rorder[c(ord_idx[bws$icon], ord_idx[bws$iuno], ord_idx[bws$iord])] <- ord_idx
      myout$g = as.matrix(myout$g[,rorder])

      myout$gerr = matrix(data=myout$gerr, nrow = enrow, ncol = ncol, byrow = FALSE) 
      myout$gerr = as.matrix(myout$gerr[,rorder])

      if (identical(bws$regtype, "lp")) {
        cont.idx <- which(bws$icon)
        if (length(cont.idx)) {
          invalid.order <- glp.gradient.order > bws$degree
          if (any(invalid.order)) {
            bad.idx <- cont.idx[invalid.order]
            myout$g[, bad.idx] <- NA_real_
            myout$gerr[, bad.idx] <- NA_real_
            if (warn.glp.gradient)
              .np_warning("some requested glp derivatives exceed polynomial degree; returning NA for those components")
          }
        }
      }
    }


    fit.elapsed <- proc.time()[3] - fit.start
    optim.time <- if (!is.null(bws$total.time) && is.finite(bws$total.time)) as.double(bws$total.time) else NA_real_
    total.time <- fit.elapsed + (if (is.na(optim.time)) 0.0 else optim.time)

    ev.args <- list(
      bws = bws,
      eval = teval,
      mean = myout$mean,
      merr = myout$merr,
      ntrain = tnrow,
      trainiseval = no.ex,
      gradients = gradients,
      residuals = residuals,
      xtra = myout$xtra,
      rows.omit = rows.omit,
      timing = bws$timing,
      total.time = total.time,
      optim.time = optim.time,
      fit.time = fit.elapsed
    )
    if (gradients) {
      ev.args$grad <- myout$g
      ev.args$gerr <- myout$gerr
    }
    if (residuals)
      ev.args$resid <- resid
    if (identical(bws$regtype, "lp"))
      ev.args$gradient.order <- glp.gradient.order
    ev <- do.call(npregression, ev.args)
    ev$nomad.time <- if (!is.null(bws$nomad.time) && is.finite(bws$nomad.time)) as.double(bws$nomad.time) else NA_real_
    ev$powell.time <- if (!is.null(bws$powell.time) && is.finite(bws$powell.time)) as.double(bws$powell.time) else NA_real_


    ev$call <- match.call(expand.dots = FALSE)
    environment(ev$call) <- parent.frame()
    return(ev)
  }

npreg.default <- function(bws, txdat, tydat, nomad = FALSE, ...){
  sc <- sys.call()
  sc.names <- names(sc)
  nomad <- npValidateScalarLogical(nomad, "nomad")

  if (!missing(bws) &&
      !isa(bws, "rbandwidth") &&
      (inherits(bws, "formula") || is.call(bws))) {
    dots <- list(...)
    dots$nomad <- nomad
    bw.args <- if (missing(txdat) && missing(tydat)) {
      list(formula = bws)
    } else {
      list(xdat = txdat, ydat = tydat)
    }
    tbw <- do.call(npregbw, c(bw.args, dots))
    reg.args <- list(bws = tbw)
    if (!missing(txdat))
      reg.args$txdat <- txdat
    if (!missing(tydat))
      reg.args$tydat <- tydat
    fit.dots <- dots
    fit.dots$remin <- NULL
    fit.dots$.np_fit_progress_handoff <- TRUE
    return(do.call(npreg, c(reg.args, fit.dots)))
  }

  ## here we check to see if the function was called with tdat =
  ## if it was, we need to catch that and map it to dat =
  ## otherwise the call is passed unadulterated to npudensbw

  bws.named <- any(sc.names == "bws")
  txdat.named <- any(sc.names == "txdat")
  tydat.named <- any(sc.names == "tydat")

  no.bws <- missing(bws)
  no.txdat <- missing(txdat)
  no.tydat <- missing(tydat)
  has.explicit.bws <- (!no.bws) && isa(bws, "rbandwidth")

  ## match.call() can normalize an originally unnamed formula first
  ## argument into named bws=... . Rewrite the bw call so formula-driven
  ## fits continue down the formula route instead of disabling search.
  if (bws.named && no.txdat && no.tydat && inherits(bws, "formula")) {
    sc$`bws` <- NULL
    sc$formula <- bws
    sc.bw <- sc
    sc.bw[[1]] <- quote(npregbw)
    bws.named <- FALSE
  } else {
    sc.bw <- sc
    sc.bw[[1]] <- quote(npregbw)
  }

  ## if bws was passed in explicitly, do not compute bandwidths
    
  if(txdat.named)
    txdat <- toFrame(txdat)

  if(bws.named){
    sc.bw$bandwidth.compute <- FALSE
  }

  ostxy <- c('txdat','tydat')
  nstxy <- c('xdat','ydat')
  
  m.txy <- match(ostxy, names(sc.bw), nomatch = 0)

  if(any(m.txy > 0)) {
    names(sc.bw)[m.txy] <- nstxy[m.txy > 0]
  }
    
  use.outer.bandwidth.progress <- !.np_bw_call_uses_nomad_degree_search(
    sc.bw,
    caller_env = parent.frame()
  )

  tbw <- if (!has.explicit.bws) {
    if (use.outer.bandwidth.progress) {
      .np_progress_select_bandwidth_enhanced(
        "Selecting regression bandwidth",
        .np_eval_bw_call(sc.bw, caller_env = parent.frame())
      )
    } else {
      .np_eval_bw_call(sc.bw, caller_env = parent.frame())
    }
  } else {
    .np_eval_bw_call(sc.bw, caller_env = parent.frame())
  }

  call.args <- list(bws = tbw)
  if (no.bws) {
    call.args$txdat <- txdat
    call.args$tydat <- tydat
  } else {
    if (txdat.named) call.args$txdat <- txdat
    if (tydat.named) call.args$tydat <- tydat
    if ((!bws.named) && (!txdat.named) && (!no.tydat) && (!tydat.named)) {
      call.args <- c(call.args, list(tydat))
    }
  }
  if (!has.explicit.bws)
    call.args$.np_fit_progress_handoff <- TRUE
  ev <- do.call(npreg, c(call.args, list(...)))

  ev$call <- match.call(expand.dots = FALSE)
  environment(ev$call) <- parent.frame()
  return(ev)
}

capture_rbandwidth_panel_ylabs <- function(object, xdat, ydat, ...) {
  captured <- new.env(parent = emptyenv())
  captured$ylab <- character()

  pdf(file = tempfile(fileext = ".pdf"))
  on.exit(dev.off(), add = TRUE)

  trace(
    what = "plot.default",
    where = asNamespace("graphics"),
    tracer = bquote({
      assign(
        "ylab",
        c(
          get("ylab", envir = .(captured), inherits = FALSE),
          as.character(ylab)
        ),
        envir = .(captured)
      )
    }),
    print = FALSE
  )
  on.exit(
    try(untrace("plot.default", where = asNamespace("graphics")), silent = TRUE),
    add = TRUE
  )

  suppressWarnings(plot(
    object,
    xdat = xdat,
    ydat = ydat,
    perspective = FALSE,
    errors = "none",
    data_overlay = FALSE,
    data_rug = FALSE,
    ...
  ))

  captured$ylab
}

make_regression_fixture <- function(predictors = c("g", "x"),
                                    regtype = "ll",
                                    degree = NULL,
                                    gradient.order = 1L) {
  set.seed(20260404)
  dat <- data.frame(
    y = NA_real_,
    g = factor(rep(c("a", "b"), each = 20L)),
    x = seq(0.05, 0.95, length.out = 40L)
  )
  dat$y <- 1 + 0.8 * (dat$g == "b") + sin(2 * pi * dat$x)
  formula <- stats::as.formula(paste("y ~", paste(predictors, collapse = " + ")))

  bw.args <- list(
    formula = formula,
    data = dat,
    regtype = regtype,
    bwtype = "fixed",
    bws = rep_len(c(0.4, 0.25), length(predictors)),
    bandwidth.compute = FALSE
  )
  if (!is.null(degree))
    bw.args$degree <- degree

  bw <- do.call(npregbw, bw.args)
  fit <- suppressWarnings(npreg(
    bws = bw,
    txdat = dat[predictors],
    tydat = dat$y,
    gradients = TRUE,
    gradient.order = gradient.order,
    warn.glp.gradient = FALSE
  ))

  list(
    fit = fit,
    xdat = dat[predictors],
    ydat = dat$y
  )
}

test_that("rbandwidth factor gradient panels use Delta labels", {
  mixed.factor.first <- make_regression_fixture(predictors = c("g", "x"))

  mixed.labels <- capture_rbandwidth_panel_ylabs(
    mixed.factor.first$fit,
    xdat = mixed.factor.first$xdat,
    ydat = mixed.factor.first$ydat,
    gradients = TRUE,
    common_scale = FALSE
  )
  expect_true("Delta y / Delta g" %in% mixed.labels)
})

test_that("rbandwidth non-gradient default ylab is unchanged", {
  fixture <- make_regression_fixture(predictors = c("x", "g"))
  labels <- capture_rbandwidth_panel_ylabs(
    fixture$fit,
    xdat = fixture$xdat,
    ydat = fixture$ydat,
    gradients = FALSE,
    common_scale = TRUE
  )

  expect_true(length(labels) >= 1L)
  expect_true(all(labels == paste("", "y")))
})

test_that("rbandwidth explicit ylab overrides remain unchanged", {
  fixture <- make_regression_fixture(predictors = c("x", "g"))

  labels <- capture_rbandwidth_panel_ylabs(
    fixture$fit,
    xdat = fixture$xdat,
    ydat = fixture$ydat,
    gradients = TRUE,
    ylab = "custom",
    common_scale = TRUE
  )

  expect_true(length(labels) >= 1L)
  expect_true(all(labels == "custom"))
})

test_that("rbandwidth mixed gradient plot still returns data payloads", {
  fixture <- make_regression_fixture(predictors = c("g", "x"))

  out <- expect_no_error(suppressWarnings(plot(
    fixture$fit,
    xdat = fixture$xdat,
    ydat = fixture$ydat,
    gradients = TRUE,
    perspective = FALSE,
    common_scale = FALSE,
    output = "data",
    errors = "none",
    data_overlay = FALSE
  )))

  expect_type(out, "list")
  expect_length(out, 2L)
  expect_true(all(vapply(out, inherits, logical(1), "npregression")))
})

test_that("npreg formula fits preserve the response name for plotting metadata", {
  set.seed(42)
  n <- 60L
  dat <- data.frame(x = runif(n, -1, 1))
  dat$y <- dat$x^2 + rnorm(n, sd = 0.25 * stats::sd(dat$x))

  fit <- np::npreg(
    y ~ x,
    data = dat,
    nmulti = 1,
    regtype = "lp",
    degree = 1,
    bwtype = "adaptive_nn"
  )

  expect_identical(fit$bws$ynames, "y")

  path <- tempfile(fileext = ".png")
  grDevices::png(path, width = 800, height = 600)
  on.exit({
    grDevices::dev.off()
    unlink(path)
  }, add = TRUE)

  expect_no_error(
    plot(fit, errors = "asymptotic", view = "fixed")
  )
})

library(np)

make_npreg_cv_lp_data <- function(n = 120L, seed = 20260305L) {
  set.seed(seed)
  x1 <- runif(n)
  x2 <- runif(n)
  z1 <- factor(rbinom(n, 1L, 0.5))
  z2 <- ordered(rbinom(n, 1L, 0.5))
  y <- cos(2 * pi * x1) + 0.5 * sin(2 * pi * x2) + as.numeric(z1) + rnorm(n, sd = 0.18)
  data.frame(y = y, x1 = x1, x2 = x2, z1 = z1, z2 = z2)
}

make_npreg_cv_lp_cont_data <- function(n = 120L, seed = 20260305L) {
  set.seed(seed)
  x <- runif(n)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.12)
  data.frame(y = y, x = x)
}

run_npreg_cv_lp_case <- function(dat, formula, regtype, tree, degree = NULL, basis = NULL) {
  old_opts <- options(np.messages = FALSE, np.tree = tree)
  on.exit(options(old_opts), add = TRUE)

  args <- list(
    formula = formula,
    data = dat,
    regtype = regtype,
    bwmethod = "cv.ls",
    nmulti = 1
  )
  if (!is.null(degree))
    args$degree <- degree
  if (!is.null(basis))
    args$basis <- basis

  do.call(npregbw, args)
}

test_that("npregbw extracted LP CV path preserves ll == lp degree-1 under tree on/off", {
  dat <- make_npreg_cv_lp_cont_data()

  for (tree in c(FALSE, TRUE)) {
    set.seed(404)
    bw.ll <- run_npreg_cv_lp_case(dat, formula = y ~ x, regtype = "ll", tree = tree)
    set.seed(404)
    bw.lp <- run_npreg_cv_lp_case(dat, formula = y ~ x, regtype = "lp", tree = tree, degree = 1L, basis = "glp")

    expect_equal(as.numeric(bw.ll$fval), as.numeric(bw.lp$fval), tolerance = 1e-12)
    expect_equal(as.numeric(bw.ll$bw), as.numeric(bw.lp$bw), tolerance = 1e-10)
    expect_equal(as.integer(bw.ll$num.feval), as.integer(bw.lp$num.feval))
  }
})

test_that("npregbw extracted LP CV path is deterministic and tree-invariant for degree-1 GLP", {
  dat <- make_npreg_cv_lp_data(seed = 20260306L)

  set.seed(505)
  bw.f0.a <- run_npreg_cv_lp_case(dat, formula = y ~ x1 + x2 + z1 + z2, regtype = "lp", tree = FALSE, degree = c(1L, 1L), basis = "glp")
  set.seed(505)
  bw.f0.b <- run_npreg_cv_lp_case(dat, formula = y ~ x1 + x2 + z1 + z2, regtype = "lp", tree = FALSE, degree = c(1L, 1L), basis = "glp")
  set.seed(505)
  bw.t1 <- run_npreg_cv_lp_case(dat, formula = y ~ x1 + x2 + z1 + z2, regtype = "lp", tree = TRUE, degree = c(1L, 1L), basis = "glp")
  fit.f0 <- npreg(bws = bw.f0.a, bandwidth.compute = FALSE)
  fit.t1 <- npreg(bws = bw.t1, bandwidth.compute = FALSE)

  expect_equal(as.numeric(bw.f0.a$fval), as.numeric(bw.f0.b$fval), tolerance = 1e-12)
  expect_equal(as.numeric(bw.f0.a$bw), as.numeric(bw.f0.b$bw), tolerance = 1e-12)
  expect_lt(abs(as.numeric(bw.f0.a$fval) - as.numeric(bw.t1$fval)), 2e-9)
  expect_equal(fitted(fit.f0), fitted(fit.t1), tolerance = 1e-4)
})

npqreg <-
  function(bws, ...){
    args <- list(...)

    if (!missing(bws)){
      if (is.recursive(bws)){
        if (!is.null(bws$formula) && is.null(args$txdat))
          UseMethod("npqreg",bws$formula)
        else if (!is.null(bws$call) && is.null(args$txdat))
          UseMethod("npqreg",bws$call)
        else if (!is.call(bws))
          UseMethod("npqreg",bws)
        else
          UseMethod("npqreg",NULL)
      } else {
        UseMethod("npqreg", NULL)
      }
    } else {
      UseMethod("npqreg", NULL)
    }
  }

.npqreg.fit.control.names <- c("data", "newdata", "exdat", "tau", "gradients", "tol", "small", "itmax")
.npqreg.removed.solver.controls <- c("ftol",
                                     "lbc.dir", "dfc.dir", "cfac.dir", "initc.dir",
                                     "lbd.dir", "hbd.dir", "dfac.dir", "initd.dir")

.npqreg_validate_tau <- function(tau) {
  if (!is.numeric(tau) || !length(tau) || anyNA(tau) ||
      any(!is.finite(tau)) || any(tau <= 0) || any(tau >= 1))
    stop("'tau' must contain numeric values in (0,1)")
  as.double(tau)
}

.npqreg_tau_labels <- function(tau) {
  paste0("tau=", format(tau, trim = TRUE, scientific = FALSE))
}

.npqreg_napredict_eval <- function(omit, x) {
  if (!length(omit))
    return(x)
  if (length(dim(x)) <= 2L)
    return(napredict(omit, x))
  d <- dim(x)
  dn <- dimnames(x)
  if (!is.null(dn))
    dn[[1L]] <- NULL
  out <- array(NA_real_,
               dim = c(d[1L] + length(omit), d[-1L]),
               dimnames = dn)
  keep <- seq_len(dim(out)[1L])[-as.integer(omit)]
  out[keep, , ] <- x
  out
}

.npqreg_validate_newdata_terms <- function(newdata, xnames) {
  nd <- toFrame(newdata)
  missing.names <- setdiff(xnames, names(nd))
  if (length(missing.names))
    stop(sprintf(
      "newdata must contain columns: %s",
      paste(shQuote(xnames), collapse = ", ")
    ), call. = FALSE)
  invisible(TRUE)
}

.npqreg_fit_dots <- function(dots, allow.bandwidth.controls = FALSE) {
  dot.names <- names(dots)
  if (is.null(dot.names))
    dot.names <- rep("", length(dots))

  stale <- intersect(dot.names[nzchar(dot.names)], .npqreg.removed.solver.controls)
  if (length(stale) && !allow.bandwidth.controls) {
    stop(sprintf(
      "'%s' %s no longer accepted by npqreg; the canonical one-dimensional quantile extractor is controlled by 'tol', 'small', and 'itmax'",
      paste(stale, collapse = "', '"),
      if (length(stale) == 1L) "is" else "are"
    ))
  }

  if (!allow.bandwidth.controls) {
    bad <- dot.names == "" | !(dot.names %in% .npqreg.fit.control.names)
    if (any(bad))
      .np_reject_unused_dots(dots[bad], "npqreg")
  }

  keep <- (!nzchar(dot.names)) | (dot.names %in% .npqreg.fit.control.names)
  dots[keep]
}

.npqreg_strip_fit_controls_from_bw_call <- function(call) {
  for (nm in c("tau", "gradients", "tol", "small", "itmax", "newdata", "exdat")) {
    if (nm %in% names(call))
      call[[nm]] <- NULL
  }
  call
}

.npqreg_quantile_delta_from_conditional <- function(bws,
                                                    xdat,
                                                    ydat,
                                                    exdat,
                                                    quantile,
                                                    gradients = FALSE) {
  xdat <- toFrame(xdat)
  ydat <- toFrame(ydat)
  exdat <- toFrame(exdat)
  quantile <- as.double(quantile)
  gradients <- npValidateScalarLogical(gradients, "gradients")

  if (length(quantile) != nrow(exdat))
    stop("quantile delta helper requires one quantile per evaluation row")
  if (ncol(ydat) != 1L)
    stop("quantile delta helper requires a single response")

  eydat <- stats::setNames(data.frame(quantile), names(ydat)[1L])
  cdf.obj <- .np_conditional_eval_selected(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    exdat = exdat,
    eydat = eydat,
    cdf = TRUE,
    gradients = gradients
  )
  dens.obj <- .np_conditional_eval_selected(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    exdat = exdat,
    eydat = eydat,
    cdf = FALSE,
    gradients = FALSE
  )

  dens <- as.double(dens.obj$condens)
  quanterr <- as.double(cdf.obj$conderr) / NZD(dens)
  quanterr[!is.finite(quanterr) | quanterr < 0.0] <- NA_real_

  if (!gradients) {
    return(list(
      quanterr = quanterr,
      quantgrad = NA,
      quantgerr = NA,
      cdf = cdf.obj,
      dens = dens.obj
    ))
  }

  dens.mat <- matrix(NZD(dens),
                     nrow = nrow(cdf.obj$congrad),
                     ncol = ncol(cdf.obj$congrad))
  grad <- -cdf.obj$congrad / dens.mat
  grad[!is.finite(grad)] <- NA_real_

  gerr <- cdf.obj$congerr / dens.mat
  gerr[!is.finite(gerr) | gerr < 0.0] <- NA_real_

  list(
    quanterr = quanterr,
    quantgrad = grad,
    quantgerr = gerr,
    cdf = cdf.obj,
    dens = dens.obj
  )
}

.npqreg_selected_cdf_values <- function(bws,
                                        xdat,
                                        ydat,
                                        exdat,
                                        ycand) {
  ydat <- toFrame(ydat)
  yname <- names(ydat)[1L]
  eydat <- stats::setNames(data.frame(as.double(ycand)), yname)
  as.double(.np_conditional_eval_selected(
    bws = bws,
    xdat = xdat,
    ydat = ydat,
    exdat = exdat,
    eydat = eydat,
    cdf = TRUE,
    gradients = FALSE
  )$condist)
}

.npqreg_assert_selected_cdf_metadata <- function(bws) {
  reg.engine <- if (is.null(bws$regtype.engine)) {
    if (is.null(bws$regtype)) "lc" else as.character(bws$regtype)
  } else {
    as.character(bws$regtype.engine)
  }
  if (!identical(reg.engine, "lp"))
    return(invisible(TRUE))

  if (is.null(bws$degree.engine) && is.null(bws$degree))
    stop("selected LP conditional distribution metadata missing from bandwidth object: degree")
  invisible(TRUE)
}

.npqreg_invert_selected_cdf <- function(bws,
                                        xdat,
                                        ydat,
                                        exdat,
                                        tau,
                                        tol,
                                        small,
                                        itmax) {
  .npqreg_assert_selected_cdf_metadata(bws)

  xdat <- toFrame(xdat)
  ydat <- toFrame(ydat)
  exdat <- toFrame(exdat)
  y <- as.double(ydat[[1L]])
  y <- y[is.finite(y)]
  if (!length(y))
    stop("npqreg selected-CDF inversion requires finite response support")

  n.eval <- nrow(exdat)
  y.min <- min(y)
  y.max <- max(y)
  if (!is.finite(y.min) || !is.finite(y.max))
    stop("npqreg selected-CDF inversion found non-finite response support")
  if (identical(y.min, y.max))
    return(rep.int(y.min, n.eval))

  lo <- rep.int(y.min, n.eval)
  hi <- rep.int(y.max, n.eval)

  flo <- .npqreg_selected_cdf_values(bws, xdat, ydat, exdat, lo)
  fhi <- .npqreg_selected_cdf_values(bws, xdat, ydat, exdat, hi)
  if (any(!is.finite(flo)) || any(!is.finite(fhi)))
    stop("npqreg selected-CDF inversion encountered non-finite bracket values")

  done.low <- flo >= tau
  done.high <- fhi < tau
  active <- !(done.low | done.high)

  maxiter <- min(as.integer(itmax), 1000L)
  iter <- 0L
  while (any(active) && iter < maxiter) {
    iter <- iter + 1L
    mid <- (lo[active] + hi[active]) / 2.0
    fmid <- .npqreg_selected_cdf_values(
      bws = bws,
      xdat = xdat,
      ydat = ydat,
      exdat = exdat[active, , drop = FALSE],
      ycand = mid
    )
    if (any(!is.finite(fmid)))
      stop("npqreg selected-CDF inversion encountered non-finite refinement values")

    active.idx <- which(active)
    upper <- fmid >= tau
    hi[active.idx[upper]] <- mid[upper]
    lo[active.idx[!upper]] <- mid[!upper]

    width <- hi[active.idx] - lo[active.idx]
    scale <- pmax(abs(hi[active.idx]), abs(lo[active.idx]), 1.0)
    active[active.idx] <- width > (tol * scale + small)
  }

  if (any(active))
    stop("npqreg selected-CDF inversion failed to converge within 'itmax'")

  out <- hi
  out[done.low] <- y.min
  out[done.high] <- y.max
  out
}

npqreg.formula <-
  function(bws, data = NULL, newdata = NULL, ...){

    tt <- terms(bws)
    m <- match(c("formula", "data", "subset", "na.action"),
               names(bws$call), nomatch = 0)
    tmf <- bws$call[c(1,m)]
    tmf[[1]] <- as.name("model.frame")
    tmf[["formula"]] <- tt
    mf.args <- as.list(tmf)[-1L]
    umf <- tmf <- do.call(stats::model.frame, mf.args, envir = environment(tt))

    tydat <- tmf[, bws$variableNames[["response"]], drop = FALSE]
    txdat <- tmf[, bws$variableNames[["terms"]], drop = FALSE]

    has.eval <- !is.null(newdata)
    if (has.eval) {
      .npqreg_validate_newdata_terms(newdata, bws$variableNames[["terms"]])
      tt <- drop.terms(tt, match(bws$variableNames$response, attr(tt, 'term.labels')))
      umf.args <- list(formula = tt, data = newdata)
      umf <- do.call(stats::model.frame, umf.args, envir = parent.frame())
      emf <- umf
      exdat <- emf[, bws$variableNames[["terms"]], drop = FALSE]
    }

    q.args <- list(txdat = txdat, tydat = tydat)
    if (has.eval)
      q.args$exdat <- exdat
    q.args$bws <- bws
    tbw <- do.call(npqreg, c(q.args, .npqreg_fit_dots(list(...))))

    tbw$omit <- attr(umf,"na.action")
    tbw$rows.omit <- as.vector(tbw$omit)
    tbw$nobs.omit <- length(tbw$rows.omit)

    tbw$quantile <- .npqreg_napredict_eval(tbw$omit, tbw$quantile)
    tbw$quanterr <- .npqreg_napredict_eval(tbw$omit, tbw$quanterr)

    if(tbw$gradients){
        tbw$quantgrad <- .npqreg_napredict_eval(tbw$omit, tbw$quantgrad)
        tbw$quantgerr <- .npqreg_napredict_eval(tbw$omit, tbw$quantgerr)
    }

    return(tbw)
  }

npqreg.call <-
  function(bws, ...) {
    npqreg(txdat = .np_eval_bws_call_arg(bws, "xdat"),
           tydat = .np_eval_bws_call_arg(bws, "ydat"),
           bws = bws, ...)
  }

npqreg.conbandwidth <-
  function(bws, ...){
    stop("incorrect bandwidth type: expected conditional distribution bandwidths instead of conditional density bandwidths")
  }

npqreg.condbandwidth <-
  function(bws,
           txdat = stop("training data 'txdat' missing"),
           tydat = stop("training data 'tydat' missing"),
           exdat,
           tau = 0.5,
           gradients = FALSE,
           tol = 1.490116e-04,
           small = 1.490116e-05, itmax = 10000,
           ...){

    fit.start <- proc.time()[3]
    fit.dots <- .npqreg_fit_dots(list(...))
    if (length(fit.dots))
      stop(sprintf("unused npqreg fit argument '%s'", names(fit.dots)[1L]))
    gradients <- npValidateScalarLogical(gradients, "gradients")
    if (!is.numeric(itmax) || length(itmax) != 1L || is.na(itmax) ||
        !is.finite(itmax) || itmax < 1 || itmax != floor(itmax))
      stop("'itmax' must be a positive integer")
    if (!is.numeric(tol) || length(tol) != 1L || is.na(tol) ||
        !is.finite(tol) || tol <= 0)
      stop("'tol' must be a positive finite numeric scalar")
    if (!is.numeric(small) || length(small) != 1L || is.na(small) ||
        !is.finite(small) || small <= 0)
      stop("'small' must be a positive finite numeric scalar")
    itmax <- as.integer(itmax)
    tol <- as.double(tol)
    small <- as.double(small)

    no.ex = missing(exdat)

    txdat = toFrame(txdat)
    tydat = toFrame(tydat)

    tau <- .npqreg_validate_tau(tau)
    ntau <- length(tau)
    tau.labels <- .npqreg_tau_labels(tau)

    if (dim(tydat)[2] != 1)
      stop("'tydat' has more than one column")

    if (!no.ex){
      exdat = toFrame(exdat)
      
      if (! txdat %~% exdat )
        stop("'txdat' and 'exdat' are not similar data frames!")
    }

    if (length(bws$xbw) != length(txdat))
      stop("length of bandwidth vector does not match number of columns of 'txdat'")

    if (length(bws$ybw) != 1)
      stop("length of bandwidth vector does not match number of columns of 'tydat'")

    if (any(bws$iyord) || any(bws$iyuno) || coarseclass(tydat[,1]) != "numeric")
      stop("'tydat' is not continuous")

    if ((any(bws$ixcon) &&
         !all(vapply(txdat[, bws$ixcon, drop = FALSE], inherits, logical(1), c("integer", "numeric")))) ||
        (any(bws$ixord) &&
         !all(vapply(txdat[, bws$ixord, drop = FALSE], inherits, logical(1), "ordered"))) ||
        (any(bws$ixuno) &&
         !all(vapply(txdat[, bws$ixuno, drop = FALSE], inherits, logical(1), "factor"))))
      stop("supplied bandwidths do not match 'txdat' in type")

    ## catch and destroy NA's
    keep.rows <- rep_len(TRUE, nrow(txdat))
    rows.omit <- attr(na.omit(data.frame(txdat, tydat)), "na.action")
    if (length(rows.omit) > 0L)
      keep.rows[as.integer(rows.omit)] <- FALSE

    if (!any(keep.rows))
      stop("Training data has no rows without NAs")

    txdat <- txdat[keep.rows,,drop = FALSE]
    tydat <- tydat[keep.rows,,drop = FALSE]

    if (!no.ex){
      keep.eval <- rep_len(TRUE, nrow(exdat))
      rows.omit <- attr(na.omit(exdat), "na.action")
      if (length(rows.omit) > 0L)
        keep.eval[as.integer(rows.omit)] <- FALSE
      exdat <- exdat[keep.eval,,drop = FALSE]
    }
    
    tnrow = dim(txdat)[1]
    enrow = (if (no.ex) tnrow else dim(exdat)[1])

    ## re-assign levels in training and evaluation data to ensure correct
    ## conversion to numeric type.
    
    txdat <- adjustLevels(txdat, bws$xdati)
    tydat <- adjustLevels(tydat, bws$ydati)
    
    if (!no.ex){
      exdat <- adjustLevels(exdat, bws$xdati)
    }

    ## grab the evaluation data before it is converted to numeric
    if(no.ex){
      txeval <- txdat
    } else {
      txeval <- exdat
    }
    txdat.df <- txdat
    tydat.df <- tydat
    if (!no.ex)
      exdat.df <- exdat

    fit_one_tau <- function(tau_i) {
      yq <- .npqreg_invert_selected_cdf(
        bws = bws,
        xdat = txdat.df,
        ydat = tydat.df,
        exdat = txeval,
        tau = tau_i,
        tol = tol,
        small = small,
        itmax = itmax
      )
      qdelta <- .npqreg_quantile_delta_from_conditional(
        bws = bws,
        xdat = txdat.df,
        ydat = tydat.df,
        exdat = txeval,
        quantile = yq,
        gradients = gradients
      )
      list(
        yq = yq,
        yqerr = qdelta$quanterr,
        yqgrad = if (gradients) qdelta$quantgrad else NA,
        yqgerr = if (gradients) qdelta$quantgerr else NA
      )
    }

    tau.out <- lapply(tau, fit_one_tau)

    if (ntau == 1L) {
      myout <- tau.out[[1L]]
    } else {
      myout <- list(
        yq = do.call(cbind, lapply(tau.out, `[[`, "yq")),
        yqerr = do.call(cbind, lapply(tau.out, `[[`, "yqerr")),
        yqgrad = NA,
        yqgerr = NA
      )
      colnames(myout$yq) <- tau.labels
      colnames(myout$yqerr) <- tau.labels
      if (gradients) {
        p <- ncol(tau.out[[1L]]$yqgrad)
        grad.names <- colnames(tau.out[[1L]]$yqgrad)
        myout$yqgrad <- array(NA_real_,
                              dim = c(enrow, p, ntau),
                              dimnames = list(NULL, grad.names, tau.labels))
        myout$yqgerr <- array(NA_real_,
                              dim = c(enrow, p, ntau),
                              dimnames = list(NULL, grad.names, tau.labels))
        for (j in seq_len(ntau)) {
          myout$yqgrad[, , j] <- tau.out[[j]]$yqgrad
          myout$yqgerr[, , j] <- tau.out[[j]]$yqgerr
        }
      }
    }


    fit.elapsed <- proc.time()[3] - fit.start
    optim.time <- if (!is.null(bws$total.time) && is.finite(bws$total.time)) as.double(bws$total.time) else NA_real_
    total.time <- fit.elapsed + (if (is.na(optim.time)) 0.0 else optim.time)

    qregression(bws = bws,
                xeval = txeval,
                tau = tau,
                quantile = myout$yq,
                quanterr = myout$yqerr,
                quantgrad = myout$yqgrad,
                quantgerr = myout$yqgerr,
                ntrain = tnrow,
                trainiseval = no.ex,
                gradients = gradients,
                timing = bws$timing, total.time = total.time,
                optim.time = optim.time, fit.time = fit.elapsed)
  }


npqreg.default <- function(bws, txdat, tydat, nomad = FALSE, ...){
  nomad <- npValidateScalarLogical(nomad, "nomad")
  sc <- sys.call()
  sc.names <- names(sc)

  ## here we check to see if the function was called with tdat =
  ## if it was, we need to catch that and map it to dat =
  ## otherwise the call is passed unadulterated to npudensbw

  bws.named <- any(sc.names == "bws")
  txdat.named <- any(sc.names == "txdat")
  tydat.named <- any(sc.names == "tydat")

  no.bws <- missing(bws)
  no.txdat <- missing(txdat)
  no.tydat <- missing(tydat)
  has.explicit.bws <- (!no.bws) && isa(bws, "condbandwidth")
  bws.formula <- (!no.bws) && inherits(bws, "formula")

  if (bws.named && no.txdat && no.tydat && bws.formula) {
    sc$`bws` <- NULL
    sc$formula <- bws
    sc.bw <- sc
    sc.bw[[1]] <- quote(npcdistbw)
    sc.bw <- .npqreg_strip_fit_controls_from_bw_call(sc.bw)
    bws.named <- FALSE
  } else if (bws.formula) {
    sc.bw <- sc
    sc.bw$`bws` <- NULL
    sc.bw[[1]] <- quote(npcdistbw)
    sc.bw <- .npqreg_strip_fit_controls_from_bw_call(sc.bw)
    bws.named <- FALSE
  } else {
    sc.bw <- sc
    sc.bw[[1]] <- quote(npcdistbw)
    sc.bw <- .npqreg_strip_fit_controls_from_bw_call(sc.bw)
  }

  ## if bws was passed in explicitly, do not compute bandwidths
    
  if(txdat.named)
    txdat <- toFrame(txdat)

  if(tydat.named)
    tydat <- toFrame(tydat)

  if(bws.named){
    sc.bw$bandwidth.compute <- FALSE
  }

  ostxy <- c('txdat','tydat')
  nstxy <- c('xdat','ydat')
  
  m.txy <- match(ostxy, names(sc.bw), nomatch = 0)

  if(any(m.txy > 0)) {
    names(sc.bw)[m.txy] <- nstxy[m.txy > 0]
  }

  use.outer.bandwidth.progress <- !.np_bw_call_uses_nomad_degree_search(
    sc.bw,
    caller_env = parent.frame()
  )

  tbw <- if (!has.explicit.bws) {
    if (use.outer.bandwidth.progress) {
      .np_progress_select_bandwidth_enhanced(
        "Selecting conditional distribution bandwidth",
        .np_eval_bw_call(sc.bw, caller_env = parent.frame())
      )
    } else {
      .np_eval_bw_call(sc.bw, caller_env = parent.frame())
    }
  } else {
    .np_eval_bw_call(sc.bw, caller_env = parent.frame())
  }

  call.args <- list(bws = tbw)
  if (no.bws) {
    call.args$txdat <- txdat
    call.args$tydat <- tydat
  } else {
    if (txdat.named) call.args$txdat <- txdat
    if (tydat.named) call.args$tydat <- tydat
    if ((!bws.named) && (!txdat.named) && (!no.tydat) && (!tydat.named)) {
      call.args <- c(call.args, list(tydat))
    }
  }
  dots <- list(...)
  if (has.explicit.bws)
    fit.dots <- .npqreg_fit_dots(dots)
  else
    fit.dots <- .npqreg_fit_dots(dots, allow.bandwidth.controls = TRUE)
  do.call(npqreg, c(call.args, fit.dots))
}

library(np)

make_adaptive_nn_conditional_data <- function() {
  set.seed(20260508L)
  n <- 120L
  x <- sort(runif(n))
  y <- sin(5 * x) + 0.3 * x^2 + rnorm(n, sd = 0.08)
  list(
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = data.frame(x = seq(0.1, 0.9, length.out = 40L)),
    eydat = data.frame(y = sin(5 * seq(0.1, 0.9, length.out = 40L)) +
      0.3 * seq(0.1, 0.9, length.out = 40L)^2)
  )
}

make_adaptive_nn_conditional_bw <- function(family, dat, regtype, degree = NULL) {
  fun <- switch(family,
    dens = npcdensbw,
    dist = npcdistbw
  )
  args <- list(
    xdat = dat$txdat,
    ydat = dat$tydat,
    bws = c(12, 12),
    bandwidth.compute = FALSE,
    bwtype = "adaptive_nn",
    regtype = regtype
  )
  if (!is.null(degree)) {
    args$degree <- degree
    args$basis <- "glp"
  }
  do.call(fun, args)
}

eval_adaptive_nn_conditional <- function(family, bw, dat) {
  fun <- switch(family,
    dens = npcdens,
    dist = npcdist
  )
  do.call(fun, list(
    bws = bw,
    txdat = dat$txdat,
    tydat = dat$tydat,
    exdat = dat$exdat,
    eydat = dat$eydat,
    gradients = TRUE
  ))
}

expect_adaptive_nn_lp_contract <- function(family) {
  dat <- make_adaptive_nn_conditional_data()
  fit.ll <- eval_adaptive_nn_conditional(
    family,
    make_adaptive_nn_conditional_bw(family, dat, "ll"),
    dat
  )
  fit.lp1 <- eval_adaptive_nn_conditional(
    family,
    make_adaptive_nn_conditional_bw(family, dat, "lp", 1L),
    dat
  )
  fit.lp2 <- eval_adaptive_nn_conditional(
    family,
    make_adaptive_nn_conditional_bw(family, dat, "lp", 2L),
    dat
  )

  expect_true(all(is.finite(fitted(fit.ll))))
  expect_true(all(is.finite(fitted(fit.lp1))))
  expect_true(all(is.finite(fitted(fit.lp2))))
  expect_equal(fitted(fit.ll), fitted(fit.lp1), tolerance = 1e-10)
  expect_equal(fit.ll$congrad, fit.lp1$congrad, tolerance = 1e-10)

  expect_gt(max(abs(fitted(fit.lp2) - fitted(fit.lp1))), 1e-4)
  expect_gt(max(abs(fit.lp2$congrad - fit.lp1$congrad)), 1e-4)
}

test_that("adaptive-nn conditional distribution lp propagates into npqreg", {
  dat <- make_adaptive_nn_conditional_data()
  q.ll <- npqreg(
    bws = make_adaptive_nn_conditional_bw("dist", dat, "ll"),
    txdat = dat$txdat,
    tydat = dat$tydat,
    exdat = dat$exdat,
    tau = 0.5,
    gradients = TRUE
  )
  q.lp1 <- npqreg(
    bws = make_adaptive_nn_conditional_bw("dist", dat, "lp", 1L),
    txdat = dat$txdat,
    tydat = dat$tydat,
    exdat = dat$exdat,
    tau = 0.5,
    gradients = TRUE
  )
  q.lp2 <- npqreg(
    bws = make_adaptive_nn_conditional_bw("dist", dat, "lp", 2L),
    txdat = dat$txdat,
    tydat = dat$tydat,
    exdat = dat$exdat,
    tau = 0.5,
    gradients = TRUE
  )

  expect_equal(q.ll$quantile, q.lp1$quantile, tolerance = 1e-10)
  expect_equal(q.ll$quanterr, q.lp1$quanterr, tolerance = 1e-10)
  expect_equal(q.ll$quantgrad, q.lp1$quantgrad, tolerance = 1e-10)
  expect_equal(q.ll$quantgerr, q.lp1$quantgerr, tolerance = 1e-10)

  expect_gt(max(abs(q.lp2$quantile - q.lp1$quantile)), 1e-4)
  expect_gt(max(abs(q.lp2$quanterr - q.lp1$quanterr)), 1e-4)
  expect_gt(max(abs(q.lp2$quantgrad - q.lp1$quantgrad)), 1e-4)
  expect_gt(max(abs(q.lp2$quantgerr - q.lp1$quantgerr)), 1e-4)
})

test_that("adaptive-nn conditional density lp honors degree metadata", {
  expect_adaptive_nn_lp_contract("dens")
})

test_that("adaptive-nn conditional distribution lp honors degree metadata", {
  expect_adaptive_nn_lp_contract("dist")
})

npcdenshat <- function(bws,
                       txdat = stop("training data 'txdat' missing"),
                       tydat = stop("training data 'tydat' missing"),
                       exdat,
                       eydat,
                       y = NULL,
                       output = c("matrix", "apply"),
                       deriv = NULL,
                       s = NULL) {
  if (!inherits(bws, "conbandwidth")) {
    stop("argument 'bws' must inherit from class 'conbandwidth' in npcdenshat()")
  }

  exdat.arg <- NULL
  eydat.arg <- NULL
  if (!missing(exdat))
    exdat.arg <- exdat
  if (!missing(eydat))
    eydat.arg <- eydat

  .npcdhat_core(
    bws = bws,
    txdat = txdat,
    tydat = tydat,
    exdat = exdat.arg,
    eydat = eydat.arg,
    y = y,
    output = output,
    operator = "normal",
    x.deriv = deriv,
    x.s = s,
    class_name = "npcdenshat",
    where = "npcdenshat"
  )
}

library(np)

test_that("npcdenshat matches npcdens and preserves matrix/apply parity across supported bwtype/regtype cells", {
  npcdenshat <- getFromNamespace("npcdenshat", "np")

  set.seed(20260310)
  n <- 18
  x <- data.frame(x = sort(runif(n)))
  y <- data.frame(y = rnorm(n))
  ex <- data.frame(x = seq(0.1, 0.9, length.out = 8))
  ey <- data.frame(y = seq(min(y$y), max(y$y), length.out = 8))
  iota <- rep.int(1, n)

  cfgs <- expand.grid(
    bwtype = c("fixed", "generalized_nn", "adaptive_nn"),
    regtype = c("lc", "ll", "lp"),
    stringsAsFactors = FALSE
  )

  for (ii in seq_len(nrow(cfgs))) {
    cfg <- cfgs[ii, ]
    bw.args <- list(
      xdat = x,
      ydat = y,
      regtype = cfg$regtype,
      bwtype = cfg$bwtype,
      bws = if (identical(cfg$bwtype, "fixed")) c(0.24, 0.28) else c(7, 7),
      bandwidth.compute = FALSE
    )
    if (identical(cfg$regtype, "lp")) {
      bw.args$basis <- "glp"
      bw.args$degree <- 2L
    }

    bw <- do.call(npcdensbw, bw.args)
    fit <- npcdens(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey)
    H <- npcdenshat(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey, output = "matrix")
    hy <- npcdenshat(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey, y = iota, output = "apply")

    expect_s3_class(H, "npcdenshat")
    expect_equal(
      as.vector(H %*% iota),
      as.vector(fit$condens),
      tolerance = 1e-10,
      info = paste(cfg$bwtype, cfg$regtype)
    )
    expect_equal(
      as.vector(hy),
      as.vector(H %*% iota),
      tolerance = 1e-12,
      info = paste(cfg$bwtype, cfg$regtype)
    )
  }
})

test_that("npcdenshat apply mode matches matrix RHS multiplication", {
  npcdenshat <- getFromNamespace("npcdenshat", "np")

  set.seed(20260311)
  n <- 14
  x <- data.frame(x = sort(runif(n)))
  y <- data.frame(y = rnorm(n))
  ex <- data.frame(x = seq(0.15, 0.85, length.out = 6))
  ey <- data.frame(y = seq(min(y$y), max(y$y), length.out = 6))

  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = 2L,
    bwtype = "generalized_nn",
    bws = c(7, 7),
    bandwidth.compute = FALSE
  )

  H <- npcdenshat(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey, output = "matrix")
  rhs <- cbind(rep.int(1, n), seq_len(n) / n)
  hy <- npcdenshat(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey, y = rhs, output = "apply")

  expect_equal(hy, H %*% rhs, tolerance = 1e-12)
})

test_that("npcdenshat derivative hat operators match npcdens gradients", {
  npcdenshat <- getFromNamespace("npcdenshat", "np")

  set.seed(20260514)
  n <- 24
  x <- data.frame(x = sort(runif(n)))
  y <- data.frame(y = sin(2 * pi * x$x) + rnorm(n, sd = 0.2))
  ex <- data.frame(x = seq(0.1, 0.9, length.out = 8))
  ey <- data.frame(y = seq(min(y$y), max(y$y), length.out = 8))
  iota <- rep.int(1, n)

  cfgs <- expand.grid(
    bwtype = c("fixed", "generalized_nn", "adaptive_nn"),
    regtype = c("lc", "ll", "lp"),
    stringsAsFactors = FALSE
  )

  for (ii in seq_len(nrow(cfgs))) {
    cfg <- cfgs[ii, ]
    bw.args <- list(
      xdat = x,
      ydat = y,
      regtype = cfg$regtype,
      bwtype = cfg$bwtype,
      bws = if (identical(cfg$bwtype, "fixed")) c(0.24, 0.28) else c(7, 7),
      bandwidth.compute = FALSE
    )
    if (identical(cfg$regtype, "lp")) {
      bw.args$basis <- "glp"
      bw.args$degree <- 2L
    }

    bw <- do.call(npcdensbw, bw.args)
    fit <- npcdens(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey, gradients = TRUE)
    H <- npcdenshat(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey, s = 1L)
    hy <- npcdenshat(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey,
                     y = iota, output = "apply", deriv = 1L)

    expect_s3_class(H, "npcdenshat")
    expect_equal(attr(H, "s"), 1L)
    expect_equal(
      as.vector(H %*% iota),
      as.vector(fit$congrad[, 1L]),
      tolerance = 1e-10,
      info = paste(cfg$bwtype, cfg$regtype)
    )
    expect_equal(
      as.vector(hy),
      as.vector(fit$congrad[, 1L]),
      tolerance = 1e-10,
      info = paste(cfg$bwtype, cfg$regtype, "apply")
    )
  }
})

test_that("npcdenshat lp Bernstein derivative operator matches npcdens gradients", {
  npcdenshat <- getFromNamespace("npcdenshat", "np")

  set.seed(20260516)
  n <- 36
  x <- data.frame(x = seq(0, 1, length.out = n))
  y <- data.frame(y = x$x^2 + 0.05 * sin(2 * pi * x$x))
  ex <- data.frame(x = seq(0.1, 0.9, length.out = 10))
  ey <- data.frame(y = seq(min(y$y) + 0.01, max(y$y) - 0.01, length.out = 10))
  iota <- rep.int(1, n)

  for (basis in c("glp", "additive", "tensor")) {
    bw <- npcdensbw(
      xdat = x,
      ydat = y,
      regtype = "lp",
      basis = basis,
      degree = 2L,
      bernstein.basis = TRUE,
      bwtype = "fixed",
      bws = c(0.24, 0.28),
      bandwidth.compute = FALSE
    )

    fit <- npcdens(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey, gradients = TRUE)
    H <- npcdenshat(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey,
                    output = "matrix", s = 1L)
    hy <- npcdenshat(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey,
                     y = iota, output = "apply", deriv = 1L)

    expect_equal(
      as.vector(H %*% iota),
      as.vector(fit$congrad[, 1L]),
      tolerance = 1e-10,
      info = paste("matrix", basis)
    )
    expect_equal(
      as.vector(hy),
      as.vector(fit$congrad[, 1L]),
      tolerance = 1e-10,
      info = paste("apply", basis)
    )
  }
})

test_that("npcdenshat named derivative selectors follow continuous x variables", {
  npcdenshat <- getFromNamespace("npcdenshat", "np")

  set.seed(20260515)
  n <- 18
  x <- data.frame(
    x1 = sort(runif(n)),
    x2 = sort(runif(n)),
    f = factor(rep(c("a", "b"), length.out = n))
  )
  y <- data.frame(y = rnorm(n))
  ex <- data.frame(
    x1 = seq(0.1, 0.9, length.out = 6),
    x2 = seq(0.2, 0.8, length.out = 6),
    f = factor(rep(c("a", "b"), length.out = 6), levels = levels(x$f))
  )
  ey <- data.frame(y = seq(min(y$y), max(y$y), length.out = 6))

  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    regtype = "ll",
    bwtype = "generalized_nn",
    bws = c(7, 7, 7, 0.4),
    bandwidth.compute = FALSE
  )

  fit <- npcdens(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey, gradients = TRUE)
  H <- npcdenshat(bws = bw, txdat = x, tydat = y, exdat = ex, eydat = ey,
                  s = c(x2 = 1L))

  expect_equal(
    as.vector(H %*% rep.int(1, n)),
    as.vector(fit$congrad[, 2L]),
    tolerance = 1e-10
  )
})

test_that("npcdenshat matches npcdens on multivariate and mixed common-use evaluation cells", {
  npcdenshat <- getFromNamespace("npcdenshat", "np")

  set.seed(20260315)
  n <- 20
  x.cont <- data.frame(
    x1 = sort(runif(n)),
    x2 = sort(runif(n))
  )
  y.cont <- data.frame(y = rnorm(n))
  ex.cont <- data.frame(
    x1 = seq(0.1, 0.9, length.out = 7),
    x2 = seq(0.2, 0.8, length.out = 7)
  )
  ey.cont <- data.frame(y = seq(min(y.cont$y), max(y.cont$y), length.out = 7))

  x.mixed <- data.frame(
    x1 = runif(n),
    x2 = runif(n),
    uf = factor(sample(c("a", "b", "c"), n, replace = TRUE)),
    of = ordered(sample(c("low", "mid", "high"), n, replace = TRUE),
      levels = c("low", "mid", "high")
    )
  )
  y.mixed <- data.frame(y = rnorm(n))
  ex.mixed <- data.frame(
    x1 = seq(0.1, 0.9, length.out = 7),
    x2 = seq(0.15, 0.85, length.out = 7),
    uf = factor(rep(c("a", "b"), length.out = 7), levels = levels(x.mixed$uf)),
    of = ordered(rep(c("low", "high"), length.out = 7), levels = levels(x.mixed$of))
  )
  ey.mixed <- data.frame(y = seq(min(y.mixed$y), max(y.mixed$y), length.out = 7))

  cases <- list(
    list(
      label = "cont2 generalized ll",
      bw = npcdensbw(
        xdat = x.cont,
        ydat = y.cont,
        regtype = "ll",
        bwtype = "generalized_nn",
        bws = c(6, 7, 7),
        bandwidth.compute = FALSE
      ),
      txdat = x.cont,
      tydat = y.cont,
      exdat = ex.cont,
      eydat = ey.cont
    ),
    list(
      label = "mixed generalized lp2",
      bw = npcdensbw(
        xdat = x.mixed,
        ydat = y.mixed,
        regtype = "lp",
        basis = "glp",
        degree = c(2L, 2L),
        bwtype = "generalized_nn",
        bws = c(6, 7, 7, 0.45, 0.55),
        bandwidth.compute = FALSE
      ),
      txdat = x.mixed,
      tydat = y.mixed,
      exdat = ex.mixed,
      eydat = ey.mixed
    )
  )

  for (case in cases) {
    fit <- npcdens(
      bws = case$bw,
      txdat = case$txdat,
      tydat = case$tydat,
      exdat = case$exdat,
      eydat = case$eydat
    )
    H <- npcdenshat(
      bws = case$bw,
      txdat = case$txdat,
      tydat = case$tydat,
      exdat = case$exdat,
      eydat = case$eydat,
      output = "matrix"
    )
    rhs <- cbind(rep.int(1, nrow(case$txdat)), seq_len(nrow(case$txdat)) / nrow(case$txdat))
    hy <- npcdenshat(
      bws = case$bw,
      txdat = case$txdat,
      tydat = case$tydat,
      exdat = case$exdat,
      eydat = case$eydat,
      y = rhs,
      output = "apply"
    )

    expect_equal(
      as.vector(H %*% rhs[, 1L]),
      as.vector(fit$condens),
      tolerance = 1e-10,
      info = case$label
    )
    expect_equal(hy, H %*% rhs, tolerance = 1e-12, info = case$label)
  }
})

test_that("npcdenshat preserves bounded gaussian manual-bandwidth semantics", {
  npcdenshat <- getFromNamespace("npcdenshat", "np")

  x <- data.frame(x = sort(c(5, 11, 21, 31, 46, 75, 98, 122, 145, 165, 195, 224)))
  y <- data.frame(y = sort(c(7, 15, 22, 35, 48, 59, 77, 95, 121, 141, 166, 188)))
  iota <- rep.int(1, nrow(x))

  bw <- npcdensbw(
    xdat = x,
    ydat = y,
    regtype = "lc",
    bws = c(100, 100),
    bandwidth.compute = FALSE,
    cxkertype = "gaussian",
    cykertype = "gaussian",
    cxkerbound = "range",
    cykerbound = "range"
  )

  fit <- npcdens(bws = bw, txdat = x, tydat = y)
  H <- npcdenshat(bws = bw, txdat = x, tydat = y, output = "matrix")
  hy <- npcdenshat(bws = bw, txdat = x, tydat = y, y = iota, output = "apply")

  expect_equal(as.vector(H %*% iota), as.vector(fit$condens), tolerance = 1e-8)
  expect_equal(as.vector(hy), as.vector(fit$condens), tolerance = 1e-8)
})

.npConmodeMetadataFromBws <- function(bws) {
  fields <- c(
    "regtype",
    "pregtype",
    "basis",
    "degree",
    "bernstein.basis",
    "regtype.engine",
    "basis.engine",
    "degree.engine",
    "bernstein.basis.engine",
    "degree.search",
    "nomad.shortcut",
    "nomad.time",
    "powell.time"
  )
  out <- setNames(vector("list", length(fields)), fields)
  for (field in fields) {
    if (!is.null(bws[[field]]))
      out[[field]] <- bws[[field]]
  }
  out$nomad <- if (!is.null(bws$nomad)) {
    isTRUE(bws$nomad)
  } else if (!is.null(bws$nomad.shortcut)) {
    isTRUE(bws$nomad.shortcut$enabled)
  } else {
    NULL
  }
  out$search.engine <- if (!is.null(bws$search.engine)) {
    as.character(bws$search.engine)[1L]
  } else if (!is.null(bws$degree.search$mode)) {
    as.character(bws$degree.search$mode)[1L]
  } else {
    NULL
  }
  out
}

conmode =
  function(bws, xeval, yeval = NA,
           conmode,
           condens, conderr = NA,
           confusion.matrix = NA,
           CCR.overall = NA,
           CCR.byoutcome = NA,
           fit.mcfadden = NA,
           ntrain, trainiseval = FALSE,
           proper.requested = FALSE,
           proper.applied = FALSE,
           proper.info = NULL,
           probabilities = NULL,
           probability.levels = NULL,
           probability.errors = NULL,
           probability.repaired.rows = NULL,
           probability.gradients = NULL,
           probability.gradient.level = NULL,
           probability.gradient.names = NULL,
           probability.gradient.info = NULL,
           xtrain = NULL,
           ytrain = NULL,
           gradients = FALSE){

    if (missing(bws) || missing(xeval) || missing(conmode) || missing(condens) || missing(ntrain))
      stop("improper invocation of conmode constructor")

    d = list(
      xbw = bws$xbw,
      ybw = bws$ybw,
      bws = bws,
      xnames = bws$xnames,
      ynames = bws$ynames,
      nobs = nrow(xeval),
      xndim = bws$xndim,
      yndim = bws$yndim,
      xnord = bws$xnord,
      xnuno = bws$xnuno,
      xncon = bws$xncon,
      ynord = bws$ynord,
      ynuno = bws$ynuno,
      yncon = bws$yncon,
      pscaling = bws$pscaling,
      ptype = bws$ptype,
      pcxkertype = bws$pcxkertype,
      puxkertype = bws$puxkertype,
      poxkertype = bws$poxkertype,
      pcykertype = bws$pcykertype,
      puykertype = bws$puykertype,
      poykertype = bws$poykertype,
      xeval = xeval,
      yeval = yeval,
      conmode = conmode,
      condens = condens,
      conderr = conderr,
      proper.requested = proper.requested,
      proper.applied = proper.applied,
      proper.info = proper.info,
      confusion.matrix = confusion.matrix,
      CCR.overall = CCR.overall,
      CCR.byoutcome = CCR.byoutcome,
      fit.mcfadden = fit.mcfadden,
      ntrain = ntrain,
      trainiseval = trainiseval,
      gradients = gradients)

    if (!is.null(probabilities)) {
      d$probabilities <- probabilities
      d$probability.levels <- probability.levels
      d$probability.errors <- probability.errors
      d$probability.repaired.rows <- probability.repaired.rows
    }
    if (!is.null(probability.gradients)) {
      d$probability.gradients <- probability.gradients
      d$probability.gradient.level <- probability.gradient.level
      d$probability.gradient.names <- probability.gradient.names
      d$probability.gradient.info <- probability.gradient.info
    }
    if (!is.null(xtrain) && !is.null(ytrain)) {
      d$xtrain <- xtrain
      d$ytrain <- ytrain
    }

    metadata <- .npConmodeMetadataFromBws(bws)
    d[names(metadata)] <- metadata

    class(d) = "conmode"

    d
  }


print.conmode <- function(x, ...){
  cat("\nConditional Mode data: ", x$ntrain, " training points,",
      if (x$trainiseval) "" else paste(" and ", x$nobs, " evaluation points,\n", sep=""),
      " in ", x$xndim + x$yndim, " variable(s)",
      "\n(", x$yndim, " dependent variable(s), and ", x$xndim, " explanatory variable(s))\n\n",
      sep="")
  print(matrix(x$ybw,ncol=x$yndim,dimnames=list(paste("Dep. Var. ",x$pscaling,":",sep=""),x$ynames)))

  print(matrix(x$xbw,ncol=x$xndim,dimnames=list(paste("Exp. Var. ",x$pscaling,":",sep=""),x$xnames)))

  cat(genBwSelStr(x$bws))

  cat(genBwKerStrs(x$bws))
  cat(genTimingStr(x$bws))
  if (!is.null(x$proper.info) && isTRUE(x$proper.requested)) {
    cat("Proper conditional probabilities: ",
        if (isTRUE(x$proper.applied)) "projected" else "already proper",
        "\n", sep="")
  }
  if (isTRUE(x$gradients)) {
    cat("Class-probability gradients/effects: available",
        if (!is.null(x$probability.gradient.level))
          paste0(" for level ", as.character(x$probability.gradient.level))
        else "",
        "\n", sep="")
  }
  
  cat("\n\n")
  if(!missing(...))
    print(...)
  invisible(x)
}

fitted.conmode <- function(object, ...) {
 object$condens 
}

.npConmodePredictHasNativeEvalArgs <- function(dots) {
  "exdat" %in% names(dots) && !is.null(dots$exdat)
}

.npConmodePredictSetNewdata <- function(dots, object, newdata) {
  if (.npConmodePredictHasNativeEvalArgs(dots))
    return(dots)

  nd <- toFrame(newdata)
  xnames <- object$xnames
  ynames <- object$ynames

  if (!is.null(xnames) && length(xnames)) {
    missing.names <- setdiff(xnames, names(nd))
    if (length(missing.names))
      stop(sprintf(
        "newdata must contain columns: %s",
        paste(shQuote(xnames), collapse = ", ")
      ), call. = FALSE)
    dots$exdat <- nd[, xnames, drop = FALSE]
    if (is.null(dots$eydat) &&
        !is.null(ynames) && length(ynames) && all(ynames %in% names(nd)))
      dots$eydat <- nd[, ynames, drop = FALSE]
  } else {
    dots$exdat <- nd
  }

  dots
}

predict.conmode <- function(object,
                            newdata = NULL,
                            type = c("class", "prob"),
                            se.fit = FALSE,
                            ...) {
  type <- match.arg(type)
  se.fit <- npValidateScalarLogical(se.fit, "se.fit")
  if (isTRUE(se.fit) && !identical(type, "prob"))
    stop("predict.conmode(se.fit=TRUE) is available only for type=\"prob\"",
         call. = FALSE)
  dots <- list(...)
  has.native.eval <- .npConmodePredictHasNativeEvalArgs(dots)
  has.native.ey <- "eydat" %in% names(dots) && !is.null(dots$eydat)
  if (isTRUE(has.native.ey) && !isTRUE(has.native.eval) && is.null(newdata))
    stop("predict.conmode: 'eydat' requires 'exdat' for native evaluation",
         call. = FALSE)
  has.eval <- !is.null(newdata) || has.native.eval

  if (!has.eval) {
    if (identical(type, "class"))
      return(object$conmode)
    probs <- object$probabilities
    if (is.null(probs))
      stop("class probabilities are not stored: fit with probabilities=TRUE")
    if (isTRUE(se.fit)) {
      se <- object$probability.errors
      if (is.null(se))
        stop("class-probability standard errors are not stored: refit with probabilities=TRUE")
      return(list(fit = probs, se.fit = se))
    }
    return(probs)
  }

  if (!is.null(newdata) && !has.native.eval) {
    if (!is.null(object$bws$formula))
      dots$newdata <- newdata
    else
      dots <- .npConmodePredictSetNewdata(dots, object, newdata)
  }

  if (identical(type, "prob")) {
    if (!is.null(dots$probabilities) && !isTRUE(dots$probabilities))
      stop("predict.conmode(type = \"prob\") requires probabilities=TRUE for evaluation")
    dots$probabilities <- TRUE
  }

  ev <- do.call(npconmode, c(list(bws = object$bws), dots))
  if (identical(type, "class"))
    return(ev$conmode)
  probs <- ev$probabilities
  if (is.null(probs))
    stop("class probabilities were not returned by the evaluation path")
  if (isTRUE(se.fit)) {
    se <- ev$probability.errors
    if (is.null(se))
      stop("class-probability standard errors were not returned by the evaluation path")
    return(list(fit = probs, se.fit = se))
  }
  probs
}
gradients.conmode <- function(x, level = NULL, errors = FALSE, ...) {
  errors <- npValidateScalarLogical(errors, "errors")
  if (isTRUE(errors))
    stop("gradient standard errors are not available for conmode objects")
  gout <- x$probability.gradients
  if (is.null(gout))
    stop("class-probability gradients/effects are not available: fit with gradients=TRUE")
  stored.level <- x$probability.gradient.level
  if (!is.null(level) && !identical(as.character(level), as.character(stored.level)))
    stop(sprintf("stored class-probability gradients are for level %s; refit with level=%s to obtain that level",
                 sQuote(as.character(stored.level)),
                 sQuote(as.character(level))))
  gout
}

summary.conmode <- function(object, ...){
  cat("\nConditional Mode data: ", object$ntrain, " training points,",
      if (object$trainiseval) "" else paste(" and ", object$nobs, " evaluation points,\n", sep=""),
      " in ", object$xndim + object$yndim, " variable(s)",
      "\n(", object$yndim, " dependent variable(s), and ", object$xndim, " explanatory variable(s))\n\n",
      sep="")

  cat(genOmitStr(object))
  print(matrix(object$ybw,ncol=object$yndim,dimnames=list(paste("Dep. Var. ",object$pscaling,":",sep=""),object$ynames)))

  print(matrix(object$xbw,ncol=object$xndim,dimnames=list(paste("Exp. Var. ",object$pscaling,":",sep=""),object$xnames)))

  cat(genBwSelStr(object$bws))
  cat('\n')
  pCatGofStr(object)

  cat(genBwKerStrs(object$bws))
  cat(genTimingStr(object$bws))
  if (!is.null(object$proper.info)) {
    cat("Proper conditional probabilities: ",
        if (isTRUE(object$proper.requested)) {
          if (isTRUE(object$proper.applied)) "projected" else "already proper"
        } else {
          "not requested"
        },
        "\n", sep="")
    if (!is.null(object$proper.info$max.negative.violation.raw))
      cat("  Max negative violation (raw): ",
          signif(object$proper.info$max.negative.violation.raw, 6), "\n", sep="")
    if (!is.null(object$proper.info$max.row.sum.deviation.raw))
      cat("  Max row-sum deviation (raw): ",
          signif(object$proper.info$max.row.sum.deviation.raw, 6), "\n", sep="")
    if (!is.null(object$proper.info$repaired.rows))
      cat("  Repaired rows: ", object$proper.info$repaired.rows, "\n", sep="")
  }
  if (isTRUE(object$gradients)) {
    cat("Class-probability gradients/effects: available",
        if (!is.null(object$probability.gradient.level))
          paste0(" for level ", as.character(object$probability.gradient.level))
        else "",
        "\n", sep="")
  }
  cat('\n\n')
  
}

with_np_progress_bindings <- function(bindings, code) {
  code <- substitute(code)
  ns <- asNamespace("np")
  old <- lapply(names(bindings), function(name) get(name, envir = ns, inherits = FALSE))
  names(old) <- names(bindings)

  for (name in names(bindings)) {
    was_locked <- bindingIsLocked(name, ns)
    if (was_locked) {
      unlockBinding(name, ns)
    }
    assign(name, bindings[[name]], envir = ns)
    if (was_locked) {
      lockBinding(name, ns)
    }
  }

  on.exit({
    for (name in names(old)) {
      was_locked <- bindingIsLocked(name, ns)
      if (was_locked) {
        unlockBinding(name, ns)
      }
      assign(name, old[[name]], envir = ns)
      if (was_locked) {
        lockBinding(name, ns)
      }
    }
  }, add = TRUE)

  eval(code, envir = parent.frame())
}

capture_progress_shadow_trace <- function(expr,
                                          force_renderer = NULL,
                                          now = function() 0,
                                          interactive = TRUE) {
  expr_env <- parent.frame()
  expr <- substitute(expr)
  trace <- list()

  recorder <- function(snapshot, event = c("render", "finish", "abort")) {
    event <- match.arg(event)
    if (isTRUE(getFromNamespace(".np_progress_is_message_muffled", "np")())) {
      return(invisible(snapshot))
    }
    trace[[length(trace) + 1L]] <<- list(
      event = event,
      renderer = force_renderer,
      id = snapshot$id,
      kind = snapshot$kind,
      current = snapshot$current,
      total = snapshot$total,
      detail = snapshot$detail,
      line = snapshot$line,
      started_at = snapshot$started_at,
      now = snapshot$now,
      last_width = snapshot$last_width
    )
    invisible(snapshot)
  }

  value <- with_np_progress_bindings(
    c(
      if (!is.null(force_renderer)) {
        list(.np_progress_renderer_for_surface = function(surface, capability) force_renderer)
      } else {
        list()
      },
      list(
        .np_progress_render_legacy = recorder,
        .np_progress_render_single_line = recorder,
        .np_progress_is_interactive = function() interactive,
        .np_progress_now = now
      )
    ),
    {
      reset <- getFromNamespace(".np_progress_reset_registry", "np")
      reset()
      on.exit(reset(), add = TRUE)
      eval(expr, envir = expr_env)
    }
  )

  list(
    value = value,
    trace = trace,
    final_line = if (length(trace)) trace[[length(trace)]]$line else NULL
  )
}

with_np_bindings <- function(bindings, code) {
  code <- substitute(code)
  ns <- asNamespace("np")
  old <- lapply(names(bindings), function(name) get(name, envir = ns, inherits = FALSE))
  names(old) <- names(bindings)

  for (name in names(bindings)) {
    was_locked <- bindingIsLocked(name, ns)
    if (was_locked) {
      unlockBinding(name, ns)
    }
    assign(name, bindings[[name]], envir = ns)
    if (was_locked) {
      lockBinding(name, ns)
    }
  }

  on.exit({
    for (name in names(old)) {
      was_locked <- bindingIsLocked(name, ns)
      if (was_locked) {
        unlockBinding(name, ns)
      }
      assign(name, old[[name]], envir = ns)
      if (was_locked) {
        lockBinding(name, ns)
      }
    }
  }, add = TRUE)

  eval(code, envir = parent.frame())
}

capture_messages_only <- function(expr) {
  messages <- character()
  withCallingHandlers(
    expr,
    message = function(m) {
      messages <<- c(messages, conditionMessage(m))
      invokeRestart("muffleMessage")
    }
  )
  messages
}

normalize_messages <- function(x) {
  sub("\n$", "", x)
}

progress_time_counter <- function(start = 0, by = 0.6) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

progress_time_values <- function(values) {
  force(values)
  i <- 0L
  function() {
    i <<- min(i + 1L, length(values))
    values[[i]]
  }
}

test_that("plot helper progress emits append-only bounded messages", {
  begin <- getFromNamespace(".np_plot_progress_begin", "np")
  tick <- getFromNamespace(".np_plot_progress_tick", "np")
  finish <- getFromNamespace(".np_plot_progress_end", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.interval.sec = 0,
    np.plot.progress.start.grace.sec = 0,
    np.plot.progress.max.intermediate = 3
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      state <- begin(total = 12, label = "Plot bootstrap wild")
      for (i in seq_len(12L)) {
        state <- tick(state, done = i)
      }
      finish(state)
    },
    now = progress_time_counter()
  )

  lines <- vapply(actual$trace, `[[`, character(1L), "line")

  expect_identical(
    vapply(actual$trace, `[[`, character(1L), "event"),
    c(rep("render", 13L), "finish")
  )
  expect_identical(lines[[1L]], "[np] Plot bootstrap wild...")
  expect_true(any(grepl("^\\[np\\] Plot bootstrap wild 1/12 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap wild 3/12 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap wild 6/12 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap wild 9/12 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap wild 12/12 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_equal(length(lines), 14L)
})

test_that("plot helper stays silent for instant runs below start grace", {
  begin <- getFromNamespace(".np_plot_progress_begin", "np")
  finish <- getFromNamespace(".np_plot_progress_end", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.interval.sec = 0,
    np.plot.progress.start.grace.sec = 1,
    np.plot.progress.max.intermediate = 3
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      state <- begin(total = 5, label = "Plot bootstrap wild")
      finish(state)
    },
    now = progress_time_counter(start = 0, by = 0.2)
  )

  expect_length(actual$trace, 0)
})

test_that("bootstrap execution stage surfaces immediately on begin", {
  begin <- getFromNamespace(".np_plot_bootstrap_progress_begin", "np")
  finish <- getFromNamespace(".np_plot_progress_end", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.interval.sec = 2,
    np.plot.progress.start.grace.sec = 0.75
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      state <- begin(total = 12L, label = "Plot bootstrap (surf 1/1)")
      finish(state)
    },
    now = progress_time_values(c(0, 0.2))
  )

  lines <- vapply(actual$trace, `[[`, character(1L), "line")
  expect_identical(
    lines[[1L]],
    "[np] Plot bootstrap (surf 1/1) 0/12 (0.0%, elapsed 0.0s, eta 0.0s)"
  )
  expect_identical(vapply(actual$trace, `[[`, character(1L), "event")[[1L]], "render")
})

test_that("single-surface plot targets omit 1/1 suffixes", {
  reg_label <- getFromNamespace(".np_plot_regression_bootstrap_target_label", "np")
  idx_label <- getFromNamespace(".np_plot_singleindex_bootstrap_target_label", "np")
  stage_label <- getFromNamespace(".np_plot_bootstrap_stage_label", "np")

  expect_null(reg_label(list(ndim = 1L, xnames = list("x")), slice.index = 1L))
  expect_null(idx_label(FALSE))
  expect_identical(
    stage_label("Plot bootstrap", method_label = "inid", target_label = idx_label(FALSE)),
    "Plot bootstrap inid"
  )
})

test_that("multi-surface plot targets retain i/n suffixes", {
  reg_label <- getFromNamespace(".np_plot_regression_bootstrap_target_label", "np")
  scoef_label <- getFromNamespace(".np_plot_scoef_bootstrap_target_label", "np")

  expect_identical(
    reg_label(list(ndim = 3L, xnames = list("x", "z", "w")), slice.index = 2L),
    "z 2/3"
  )
  expect_identical(
    scoef_label(list(xndim = 1L, zndim = 1L, xnames = list("x"), znames = list("z")), slice.index = 2L),
    "z 2/2"
  )
})

test_that("plot helper activity renders immediately for long blocking work", {
  begin <- getFromNamespace(".np_plot_activity_begin", "np")
  finish <- getFromNamespace(".np_plot_activity_end", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.start.grace.sec = 0.75
  )
  on.exit(options(old_opts), add = TRUE)

  legacy <- capture_progress_shadow_trace(
    {
      activity <- begin("Constructing bootstrap bands")
      finish(activity)
    },
    force_renderer = "legacy",
    now = progress_time_values(c(0, 1.0))
  )

  actual <- capture_progress_shadow_trace(
    {
      activity <- begin("Constructing bootstrap bands")
      finish(activity)
    },
    now = progress_time_values(c(0, 1.0))
  )

  actual_render_lines <- vapply(
    actual$trace[vapply(actual$trace, `[[`, character(1L), "event") == "render"],
    `[[`,
    character(1L),
    "line"
  )
  legacy_render_lines <- vapply(
    legacy$trace[vapply(legacy$trace, `[[`, character(1L), "event") == "render"],
    `[[`,
    character(1L),
    "line"
  )

  expect_identical(
    actual_render_lines,
    legacy_render_lines
  )
  expect_identical(
    actual_render_lines,
    "[np] Constructing bootstrap bands... elapsed 0.0s"
  )
  expect_identical(vapply(actual$trace, `[[`, character(1L), "event"), c("render", "finish"))
})

test_that("plot helper activity no longer waits for grace before first render", {
  begin <- getFromNamespace(".np_plot_activity_begin", "np")
  finish <- getFromNamespace(".np_plot_activity_end", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.start.grace.sec = 0.75
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      activity <- begin("Constructing bootstrap bands")
      finish(activity)
    },
    now = progress_time_values(c(0, 0.2))
  )

  expect_identical(vapply(actual$trace, `[[`, character(1L), "event"), c("render", "finish"))
  expect_identical(
    vapply(actual$trace, `[[`, character(1L), "line"),
    rep("[np] Constructing bootstrap bands... elapsed 0.0s", 2L)
  )
})

test_that("local-polynomial regression bootstrap advances when counts are precomputed", {
  boot_lp <- getFromNamespace(".np_inid_boot_from_regression", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.interval.sec = 0,
    np.plot.progress.start.grace.sec = 0,
    np.plot.progress.max.intermediate = 3
  )
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260508)
  n <- 8L
  B <- 12L
  x <- data.frame(z = seq(-1, 1, length.out = n))
  y <- sin(x$z)
  bw <- get("npregbw", envir = asNamespace("np"), inherits = FALSE)(
    xdat = x,
    ydat = y,
    bws = 0.75,
    regtype = "lp",
    degree = 1L,
    bwtype = "fixed",
    bandwidth.compute = FALSE
  )
  counts <- matrix(0L, nrow = n, ncol = B)
  for (b in seq_len(B))
    counts[((seq_len(n) + b - 2L) %% n) + 1L, b] <- 1L

  actual <- capture_progress_shadow_trace(
    {
      boot_lp(
        xdat = x,
        exdat = x,
        bws = bw,
        ydat = y,
        B = B,
        counts = counts,
        progress.label = "Plot bootstrap lp-counts"
      )
    },
    now = progress_time_counter()
  )

  lines <- vapply(actual$trace, `[[`, character(1L), "line")
  expect_true(any(grepl("^\\[np\\] Plot bootstrap lp-counts 3/12 ", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap lp-counts 6/12 ", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap lp-counts 9/12 ", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap lp-counts 12/12 ", lines)))
})

test_that("first-render helper emits exactly one render activity line", {
  init <- getFromNamespace(".np_plot_first_render_state", "np")
  begin <- getFromNamespace(".np_plot_first_render_begin", "np")
  finish <- getFromNamespace(".np_plot_first_render_end", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.start.grace.sec = 0.75
  )
  on.exit(options(old_opts), add = TRUE)

  state <- init()
  actual <- capture_progress_shadow_trace(
    {
      begin(state)
      finish(state)
      begin(state)
      finish(state)
    },
    now = progress_time_values(c(0, 0.2))
  )

  expect_false(isTRUE(state$pending))
  expect_identical(vapply(actual$trace, `[[`, character(1L), "event"), c("render", "finish"))
  expect_identical(
    vapply(actual$trace, `[[`, character(1L), "line"),
    rep("[np] Rendering plot surface... elapsed 0.0s", 2L)
  )
})

test_that("plot engine setup does not open a graphics device on the null device", {
  capture.par <- getFromNamespace(".np_plot_capture_par", "np")
  engine.begin <- getFromNamespace(".np_plot_engine_begin", "np")

  skip_if_not(isTRUE(unname(as.integer(dev.cur())) == 1L))

  before <- unname(as.integer(dev.cur()))
  on.exit({
    while (!isTRUE(unname(as.integer(dev.cur())) == 1L)) {
      dev.off()
    }
  }, add = TRUE)

  expect_identical(capture.par(c("mfrow", "cex")), list())
  expect_identical(unname(as.integer(dev.cur())), before)

  state <- engine.begin(plot.par.mfrow = TRUE)
  expect_true(is.list(state))
  expect_identical(unname(as.integer(dev.cur())), before)
})

test_that("plot helper activity yields the line to nested bounded progress", {
  activity.begin <- getFromNamespace(".np_plot_activity_begin", "np")
  activity.end <- getFromNamespace(".np_plot_activity_end", "np")
  progress.begin <- getFromNamespace(".np_plot_progress_begin", "np")
  progress.tick <- getFromNamespace(".np_plot_progress_tick", "np")
  progress.end <- getFromNamespace(".np_plot_progress_end", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.interval.sec = 0,
    np.plot.progress.start.grace.sec = 0,
    np.plot.progress.max.intermediate = 3
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      activity <- activity.begin("Preparing plot bootstrap inid")
      state <- progress.begin(total = 4L, label = "Plot bootstrap inid")
      for (i in seq_len(4L)) {
        state <- progress.tick(state, done = i)
      }
      progress.end(state)
      activity.end(activity)
    },
    now = progress_time_counter(start = 0, by = 0.6)
  )

  lines <- vapply(actual$trace, `[[`, character(1L), "line")

  expect_true(any(grepl("^\\[np\\] Preparing plot bootstrap inid\\.\\.\\. elapsed 0\\.0s$", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap inid 1/4 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap inid 4/4 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
})

test_that("block-style plot bootstrap chunking is capped only when plot progress is active", {
  chunk_size <- getFromNamespace(".np_inid_chunk_size", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.max.intermediate = 3L,
    np.plot.inid.chunk.size = NULL
  )
  on.exit(options(old_opts), add = TRUE)

  expect_identical(chunk_size(n = 1000L, B = 9999L, progress_cap = FALSE), 8388L)
  expect_identical(chunk_size(n = 1000L, B = 9999L, progress_cap = TRUE), 8388L)

  capped <- with_np_progress_bindings(
    list(.np_progress_is_interactive = function() TRUE),
    chunk_size(n = 1000L, B = 9999L, progress_cap = TRUE)
  )
  expect_true(capped < 2500L)
  expect_true(capped > 0L)
})

test_that("ordinary inid plot bootstrap chunking warms up early when plot progress is active", {
  chunk_size <- getFromNamespace(".np_inid_chunk_size", "np")
  warmup_max <- getFromNamespace(".np_plot_progress_warmup_max_reps", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.max.intermediate = 3L,
    np.plot.progress.warmup.max.reps = 16L,
    np.plot.inid.chunk.size = NULL
  )
  on.exit(options(old_opts), add = TRUE)

  uncapped <- chunk_size(n = 500L, B = 999L, progress_cap = FALSE, progress_enabled = FALSE)
  expect_identical(uncapped, 999L)

  capped <- chunk_size(n = 500L, B = 999L, progress_cap = FALSE, progress_enabled = TRUE)
  expect_true(capped > 0L)
  expect_true(capped <= warmup_max())
})

test_that("ordinary inid plot bootstrap emits intermediate progress updates", {
  helper <- getFromNamespace(".np_inid_lc_boot_from_hat", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.interval.sec = 0,
    np.plot.progress.start.grace.sec = 0,
    np.plot.progress.max.intermediate = 3L,
    np.plot.progress.warmup.max.reps = 16L,
    np.plot.inid.chunk.size = NULL
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    helper(
      H = diag(4L),
      ydat = c(1, 2, 3, 4),
      B = 9L
    ),
    now = progress_time_counter()
  )

  lines <- vapply(actual$trace, `[[`, character(1L), "line")
  expect_true(any(grepl("^\\[np\\] Plot bootstrap inid 3/9 \\(", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap inid 6/9 \\(", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap inid 9/9 \\(", lines)))
})

test_that("block-style unconditional bootstrap emits intermediate progress updates", {
  helper <- getFromNamespace(".np_inid_lc_boot_from_hat", "np")
  counts.drawer <- function(start, stopi) {
    matrix(1L, nrow = 4L, ncol = stopi - start + 1L)
  }

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.interval.sec = 0,
    np.plot.progress.start.grace.sec = 0,
    np.plot.progress.max.intermediate = 3L,
    np.plot.inid.chunk.size = NULL
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    helper(
      H = diag(4L),
      ydat = c(1, 2, 3, 4),
      B = 9L,
      counts.drawer = counts.drawer
    ),
    now = progress_time_counter()
  )

  lines <- vapply(actual$trace, `[[`, character(1L), "line")
  expect_true(any(grepl("^\\[np\\] Plot bootstrap block 3/9 \\(", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap block 6/9 \\(", lines)))
  expect_true(any(grepl("^\\[np\\] Plot bootstrap block 9/9 \\(", lines)))
})

test_that("block bootstrap drawer uses iid fast path when block length is one", {
  drawer_factory <- getFromNamespace(".np_block_counts_drawer", "np")
  boot.ns <- asNamespace("boot")
  calls <- 0L

  trace(
    what = "ts.array",
    where = boot.ns,
    tracer = quote(calls <<- calls + 1L),
    print = FALSE
  )
  on.exit(untrace("ts.array", where = boot.ns), add = TRUE)

  set.seed(20260313)
  drawer <- drawer_factory(n = 8L, B = 11L, blocklen = 1L, sim = "geom", n.sim = 8L)
  out <- drawer(1L, 3L)

  expect_identical(calls, 0L)
  expect_true(is.matrix(out))
  expect_identical(dim(out), c(8L, 3L))
  expect_true(all(colSums(out) == 8L))
  expect_true(all(out >= 0))
})

test_that("block bootstrap drawer defers ts.array setup until chunk demand", {
  drawer_factory <- getFromNamespace(".np_block_counts_drawer", "np")
  boot.ns <- asNamespace("boot")
  assign(".np_test_ts_array_calls", 0L, envir = .GlobalEnv)

  trace(
    what = "ts.array",
    where = boot.ns,
    tracer = quote(assign(
      ".np_test_ts_array_calls",
      get(".np_test_ts_array_calls", envir = .GlobalEnv) + 1L,
      envir = .GlobalEnv
    )),
    print = FALSE
  )
  on.exit({
    untrace("ts.array", where = boot.ns)
    rm(".np_test_ts_array_calls", envir = .GlobalEnv)
  }, add = TRUE)

  set.seed(20260313)
  drawer <- drawer_factory(n = 12L, B = 9L, blocklen = 3L, sim = "geom", n.sim = 12L)

  expect_identical(get(".np_test_ts_array_calls", envir = .GlobalEnv), 0L)

  out1 <- drawer(1L, 2L)
  expect_identical(get(".np_test_ts_array_calls", envir = .GlobalEnv), 1L)
  expect_identical(dim(out1), c(12L, 2L))
  expect_true(all(colSums(out1) == 12L))

  out2 <- drawer(3L, 4L)
  expect_identical(get(".np_test_ts_array_calls", envir = .GlobalEnv), 2L)
  expect_identical(dim(out2), c(12L, 2L))
  expect_true(all(colSums(out2) == 12L))
})

test_that("plot progress chunk controller adapts chunk size toward throttle interval", {
  make_controller <- getFromNamespace(".np_plot_progress_chunk_controller", "np")
  observe_controller <- getFromNamespace(".np_plot_progress_chunk_observe", "np")

  old_opts <- options(np.messages = TRUE, np.plot.progress = TRUE)
  on.exit(options(old_opts), add = TRUE)

  controller <- make_controller(chunk.size = 1000L, progress = list(throttle_sec = 0.5))
  controller$adaptive <- TRUE
  slower <- observe_controller(controller = controller, bsz = 1000L, elapsed.sec = 10)
  expect_true(slower$chunk.size < 1000L)
  expect_true(slower$chunk.size >= 250L)

  slower$adaptive <- TRUE
  faster <- observe_controller(controller = slower, bsz = slower$chunk.size, elapsed.sec = 0.05)
  expect_true(faster$chunk.size > slower$chunk.size)
})

test_that("heavy plot helpers invoke delayed activity notifications", {
  regression.eval <- getFromNamespace(".np_plot_regression_eval", "np")
  unconditional.eval <- getFromNamespace(".np_plot_unconditional_eval", "np")
  singleindex.local <- getFromNamespace(".np_plot_singleindex_local_eval", "np")
  singleindex.asym <- getFromNamespace(".np_plot_singleindex_asymptotic_eval", "np")

  labels <- character()
  record_activity <- function(label) {
    labels <<- c(labels, label)
    TRUE
  }

  dummy_bws <- list(
    type = "fixed",
    xdati = list(icon = TRUE),
    beta = 1,
    bw = 1,
    ckertype = "gaussian",
    ckerorder = 2L,
    xbw = 1,
    ybw = 1,
    ixcon = TRUE,
    ixord = FALSE,
    ixuno = FALSE,
    iycon = TRUE,
    iyord = FALSE,
    iyuno = FALSE,
    xncon = 1L,
    xnord = 0L,
    xnuno = 0L,
    yncon = 1L,
    ynord = 0L,
    ynuno = 0L,
    xndim = 1L,
    ydati = list(all.dlev = list()),
    total.time = 0,
    timing = FALSE,
    xmcv = structure(1, num.row = 1L),
    nconfac = 1,
    ncatfac = 1,
    sdev = 1,
    cxkertype = "gaussian",
    cxkerorder = 2L,
    cykertype = "gaussian",
    cykerorder = 2L,
    cxkerlb = -Inf,
    cxkerub = Inf,
    cykerlb = -Inf,
    cykerub = Inf,
    uxkertype = "aitchisonaitken",
    uykertype = "aitchisonaitken",
    oxkertype = "wangvanryzin",
    oykertype = "wangvanryzin"
  )
  x <- data.frame(x = 1:2)

  with_np_bindings(
    list(
      .np_plot_activity_begin = record_activity,
      .np_plot_activity_end = function(state) invisible(NULL),
      .np_regression_direct = function(...) list(mean = c(1, 2), grad = matrix(c(1, 2), ncol = 1L)),
      .np_ksum_unconditional_eval_exact = function(...) c(1, 2),
      npreg = function(...) list(mean = c(1, 2), merr = c(0.1, 0.2), grad = matrix(c(1, 2), ncol = 1L), gerr = matrix(c(0.1, 0.2), ncol = 1L)),
      npudens = function(...) list(dens = c(1, 2), derr = c(NA_real_, NA_real_)),
      npudist = function(...) list(dist = c(1, 2), derr = c(NA_real_, NA_real_)),
      .np_indexhat_rbw = function(...) dummy_bws,
      .np_plot_singleindex_hat_apply_index = function(...) c(1, 2),
      .npindex_resolve_spec = function(...) list(regtype.engine = "lc"),
      npValidateScalarLogical = function(x, ...) isTRUE(x),
      adjustLevels = function(dat, ...) dat,
      toFrame = function(x) if (is.data.frame(x)) x else data.frame(x = x),
      toMatrix = function(x) as.matrix(x)
    ),
    {
      regression.eval(bws = dummy_bws, xdat = x, ydat = c(1, 2), exdat = x)
      unconditional.eval(xdat = x, exdat = x, bws = dummy_bws, cdf = FALSE)
      singleindex.local(bws = dummy_bws, idx.train = x, idx.eval = x, ydat = c(1, 2))
      singleindex.asym(bws = dummy_bws, txdat = x, tydat = c(1, 2), index.eval = c(1, 2))
    }
  )

  expect_true("Computing regression plot fit" %in% labels)
  expect_true("Computing unconditional density plot fit" %in% labels)
  expect_true("Computing single-index plot fit" %in% labels)
  expect_true("Computing single-index plot asymptotic fit" %in% labels)
})

test_that("plot helper progress is silent by default in noninteractive mode", {
  begin <- getFromNamespace(".np_plot_progress_begin", "np")
  tick <- getFromNamespace(".np_plot_progress_tick", "np")
  finish <- getFromNamespace(".np_plot_progress_end", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.interval.sec = 0,
    np.plot.progress.start.grace.sec = 0,
    np.plot.progress.max.intermediate = 3
  )
  on.exit(options(old_opts), add = TRUE)

  messages <- with_np_bindings(
    list(
      .np_progress_is_interactive = function() FALSE,
      .np_progress_now = progress_time_counter()
    ),
    capture_messages_only({
      state <- begin(total = 12, label = "Plot bootstrap wild")
      state <- tick(state, done = 12)
      finish(state)
    })
  )

  expect_length(messages, 0)
})

test_that("plot helper progress respects suppressMessages", {
  begin <- getFromNamespace(".np_plot_progress_begin", "np")
  tick <- getFromNamespace(".np_plot_progress_tick", "np")
  finish <- getFromNamespace(".np_plot_progress_end", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.interval.sec = 0,
    np.plot.progress.start.grace.sec = 0,
    np.plot.progress.max.intermediate = 3
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    suppressMessages({
        state <- begin(total = 12, label = "Plot bootstrap wild")
        state <- tick(state, done = 12)
        finish(state)
      }),
    now = progress_time_counter()
  )

  expect_length(actual$trace, 0)
})

test_that("plot helper progress caps intermediate heartbeats", {
  begin <- getFromNamespace(".np_plot_progress_begin", "np")
  tick <- getFromNamespace(".np_plot_progress_tick", "np")
  finish <- getFromNamespace(".np_plot_progress_end", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.interval.sec = 10,
    np.plot.progress.start.grace.sec = 0,
    np.plot.progress.max.intermediate = 2
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      state <- begin(total = 12, label = "Plot bootstrap wild")
      for (i in seq_len(12L)) {
        state <- tick(state, done = i)
      }
      finish(state)
    },
    now = progress_time_counter()
  )

  lines <- vapply(actual$trace, `[[`, character(1L), "line")

  expect_equal(length(lines), 2L)
  expect_identical(lines[[1L]], "[np] Plot bootstrap wild...")
  expect_true(any(grepl("^\\[np\\] Plot bootstrap wild 12/12 ", lines)))
})

test_that("rotation helper emits bounded progress with eta", {
  begin <- getFromNamespace(".np_plot_rotation_progress_begin", "np")
  tick <- getFromNamespace(".np_plot_rotation_progress_tick", "np")
  finish <- getFromNamespace(".np_plot_rotation_progress_end", "np")

  old_opts <- options(
    np.messages = TRUE,
    np.plot.progress = TRUE,
    np.plot.progress.interval.sec = 0,
    np.plot.progress.start.grace.sec = 0,
    np.plot.progress.max.intermediate = 3
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    {
      state <- begin(total_frames = 8L)
      for (i in seq_len(8L)) {
        state <- tick(state, done = i)
      }
      finish(state)
    },
    now = progress_time_counter()
  )

  lines <- vapply(actual$trace, `[[`, character(1L), "line")
  expect_identical(lines[[1L]], "[np] Rotating plot...")
  expect_true(any(grepl("^\\[np\\] Rotating plot 1/8 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Rotating plot 8/8 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
})

npplreg_fit_progress_time_counter <- function(start = 0, by = 1.7) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

npplreg_fit_progress_time_values <- function(values) {
  force(values)
  i <- 0L
  function() {
    i <<- min(i + 1L, length(values))
    values[[i]]
  }
}

npplreg_fit_progress_lines <- function(shadow) {
  vapply(shadow$trace, `[[`, character(1L), "line")
}

make_npplreg_fit_progress_fixture <- function() {
  set.seed(20260404)
  n <- 32L
  dat <- data.frame(
    x = seq(-0.9, 0.9, length.out = n),
    z = seq(0.2, 1.8, length.out = n)
  )
  dat$y <- 1.5 * dat$x + sin(2 * pi * dat$z) + 0.1 * cos(pi * dat$x)

  list(
    dat = dat,
    tx = dat["x"],
    tz = dat["z"],
    y = dat$y,
    bw = npplregbw(
      xdat = dat["x"],
      zdat = dat["z"],
      ydat = dat$y,
      bws = matrix(c(0.35, 0.35), nrow = 2L),
      bandwidth.compute = FALSE
    )
  )
}

test_that("npplreg fit progress uses the bandwidth single-line surface", {
  begin <- getFromNamespace(".np_plreg_fit_progress_begin", "np")

  old_opts <- options(np.messages = TRUE)
  on.exit(options(old_opts), add = TRUE)

  state <- with_np_progress_bindings(
    list(
      .np_progress_is_interactive = function() TRUE,
      .np_progress_now = function() 0
    ),
    begin(xnames = "x")
  )

  expect_identical(state$surface, "bandwidth")
  expect_identical(state$renderer, "single_line")
  expect_equal(state$total, 2L)
})

test_that("npplreg direct bws fit emits known-total fit progress", {
  fixture <- make_npplreg_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npplreg(
      bws = fixture$bw,
      txdat = fixture$tx,
      tzdat = fixture$tz,
      tydat = fixture$y
    ),
    force_renderer = "single_line",
    now = npplreg_fit_progress_time_counter()
  )

  lines <- npplreg_fit_progress_lines(actual)

  expect_s3_class(actual$value, "plregression")
  expect_true(any(grepl(
    "^\\[np\\] Fitting partially linear regression 1/2 \\(50\\.0%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\): y~z$",
    lines
  )))
  expect_true(any(grepl(
    "^\\[np\\] Fitting partially linear regression 2/2 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\): x~z$",
    lines
  )))
})

test_that("npplreg direct bws fit stays silent below start grace without handoff", {
  fixture <- make_npplreg_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0.75,
    np.progress.interval.known.sec = 0.5
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npplreg(
      bws = fixture$bw,
      txdat = fixture$tx,
      tzdat = fixture$tz,
      tydat = fixture$y
    ),
    force_renderer = "single_line",
    now = npplreg_fit_progress_time_values(c(0, 0.2, 0.4, 0.6))
  )

  expect_length(actual$trace, 0L)
})

test_that("npplreg formula bw to fit route hands off immediately into fit progress", {
  fixture <- make_npplreg_fit_progress_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npplreg(
      y ~ x | z,
      data = fixture$dat,
      nmulti = 1
    ),
    force_renderer = "single_line",
    now = npplreg_fit_progress_time_counter()
  )

  lines <- npplreg_fit_progress_lines(actual)
  fit.zero.pos <- grep(
    "^\\[np\\] Fitting partially linear regression 0/2 \\(0\\.0%, elapsed 0\\.0s, eta 0\\.0s\\): starting y~z$",
    lines
  )
  fit.one.pos <- grep(
    "^\\[np\\] Fitting partially linear regression 1/2 \\(50\\.0%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\): y~z$",
    lines
  )
  fit.two.pos <- grep(
    "^\\[np\\] Fitting partially linear regression 2/2 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\): x~z$",
    lines
  )
  bandwidth.pos <- grep("^\\[np\\] Bandwidth selection \\(", lines)

  expect_s3_class(actual$value, "plregression")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(fit.zero.pos) == 1L)
  expect_true(length(fit.one.pos) >= 1L)
  expect_true(length(fit.two.pos) >= 1L)
  expect_lt(max(bandwidth.pos), fit.zero.pos[[1L]])
  expect_lt(fit.zero.pos[[1L]], fit.one.pos[[1L]])
  expect_lt(fit.one.pos[[1L]], fit.two.pos[[1L]])
})

test_that("npplreg nomad to powell to fit route preserves single-line fit handoff", {
  skip_if_not_installed("crs")

  set.seed(20260404)
  n <- 18L
  dat <- data.frame(x = rnorm(n), z = sort(runif(n)))
  dat$y <- 1 + 0.5 * dat$x + sin(2 * pi * dat$z) + rnorm(n, sd = 0.05)

  old_opts <- options(
    np.messages = TRUE,
    np.tree = FALSE,
    np.progress.start.grace.known.sec = 0,
    np.progress.start.grace.unknown.sec = 0,
    np.progress.interval.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  actual <- capture_progress_shadow_trace(
    npplreg(
      y ~ x | z,
      data = dat,
      nomad = TRUE,
      degree.max = 1L,
      nmulti = 1L
    ),
    force_renderer = "single_line",
    now = npplreg_fit_progress_time_counter()
  )

  lines <- npplreg_fit_progress_lines(actual)
  fit.zero.pos <- grep(
    "^\\[np\\] Fitting partially linear regression 0/2 \\(0\\.0%, elapsed 0\\.0s, eta 0\\.0s\\): starting y~z$",
    lines
  )
  fit.one.pos <- grep(
    "^\\[np\\] Fitting partially linear regression 1/2 \\(50\\.0%, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\): y~z$",
    lines
  )
  fit.two.pos <- grep(
    "^\\[np\\] Fitting partially linear regression 2/2 \\(100\\.0%, elapsed [0-9]+\\.[0-9]s, eta 0\\.0s\\): x~z$",
    lines
  )
  powell.pos <- grep("^\\[np\\] Refining bandwidth \\(", lines)
  bandwidth.pos <- grep("^\\[np\\] Selecting degree and bandwidth", lines)

  expect_s3_class(actual$value, "plregression")
  expect_true(length(bandwidth.pos) > 0L)
  expect_true(length(powell.pos) > 0L)
  expect_true(length(fit.zero.pos) == 1L)
  expect_true(length(fit.one.pos) >= 1L)
  expect_true(length(fit.two.pos) >= 1L)
  expect_lt(max(powell.pos), fit.zero.pos[[1L]])
  expect_lt(max(bandwidth.pos), fit.zero.pos[[1L]])
  expect_lt(fit.zero.pos[[1L]], fit.one.pos[[1L]])
  expect_lt(fit.one.pos[[1L]], fit.two.pos[[1L]])
})

progress_shadow_times <- function(values) {
  force(values)
  i <- 0L
  function() {
    i <<- min(i + 1L, length(values))
    values[[i]]
  }
}

test_that("shadow harness captures known-total render events and final line", {
  begin <- getFromNamespace(".np_progress_begin", "np")
  step <- getFromNamespace(".np_progress_step", "np")
  finish <- getFromNamespace(".np_progress_end", "np")

  old_opts <- options(np.messages = TRUE, np.progress.start.grace.known.sec = 0)
  on.exit(options(old_opts), add = TRUE)

  shadow <- capture_progress_shadow_trace(
    {
      state <- begin("Bootstrap replications", total = 2)
      state <- step(state, done = 1)
      finish(state)
    },
    now = progress_shadow_times(c(0, 0.6, 1.2))
  )

  expect_equal(vapply(shadow$trace, `[[`, character(1L), "event"), c("render", "render", "finish"))
  expect_equal(vapply(shadow$trace, `[[`, character(1L), "kind"), rep("known", 3L))
  expect_identical(shadow$trace[[1L]]$line, "[np] Bootstrap replications...")
  expect_match(shadow$trace[[2L]]$line, "^\\[np\\] Bootstrap replications 1/2 ")
  expect_match(shadow$final_line, "^\\[np\\] Bootstrap replications 2/2 ")
})

test_that("shadow harness captures unknown-total delayed activity semantics", {
  begin <- getFromNamespace(".np_progress_begin", "np")
  step <- getFromNamespace(".np_progress_step", "np")

  old_opts <- options(np.messages = TRUE, np.progress.start.grace.unknown.sec = 1.0)
  on.exit(options(old_opts), add = TRUE)

  shadow <- capture_progress_shadow_trace(
    {
      state <- begin("Iterating Landweber-Fridman solve")
      state <- step(state, done = 3, detail = "updating residual smoothing")
      step(state, done = 4, detail = "evaluating stopping rule")
    },
    now = progress_shadow_times(c(0, 0.4, 1.2))
  )

  expect_equal(vapply(shadow$trace, `[[`, character(1L), "event"), c("render", "render"))
  expect_equal(vapply(shadow$trace, `[[`, character(1L), "kind"), rep("unknown", 2L))
  expect_identical(shadow$trace[[1L]]$line, "[np] Iterating Landweber-Fridman solve...")
  expect_identical(
    shadow$final_line,
    "[np] Iterating Landweber-Fridman solve... iteration 4, elapsed 1.2s: evaluating stopping rule"
  )
})

progress_time_counter <- function(start = 0, by = 0.6) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

shadow_bootstrap_signature <- function(shadow) {
  lines <- vapply(shadow$trace, `[[`, character(1L), "line")
  events <- vapply(shadow$trace, `[[`, character(1L), "event")
  keep <- grepl("^\\[np\\] Bootstrap replications", lines)

  data.frame(
    event = events[keep],
    line = lines[keep],
    stringsAsFactors = FALSE
  )
}

shadow_lines <- function(shadow) {
  shadow_bootstrap_signature(shadow)$line
}

npqcmstest_fun <- function(...) {
  getFromNamespace("npqcmstest", "np")(...)
}

test_that("npqcmstest single-line bootstrap progress matches legacy semantics", {
  skip_if_not_installed("quantreg")
  library(quantreg)

  set.seed(42)
  n <- 40
  x <- rnorm(n)
  y <- 1 + x + rnorm(n, sd = 0.1)
  model <- rq(y ~ x, tau = 0.5, model = TRUE)

  old_opts <- options(np.messages = TRUE)
  on.exit(options(old_opts), add = TRUE)

  legacy <- capture_progress_shadow_trace(
    npqcmstest_fun(
      model = model,
      xdat = x,
      ydat = y,
      distribution = "bootstrap",
      boot.num = 9
    ),
    force_renderer = "legacy",
    now = progress_time_counter()
  )

  set.seed(42)
  single_line <- capture_progress_shadow_trace(
    npqcmstest_fun(
      model = model,
      xdat = x,
      ydat = y,
      distribution = "bootstrap",
      boot.num = 9
    ),
    force_renderer = "single_line",
    now = progress_time_counter()
  )

  lines <- shadow_lines(single_line)

  expect_s3_class(single_line$value, "cmstest")
  expect_equal(shadow_bootstrap_signature(single_line), shadow_bootstrap_signature(legacy))
  expect_true(any(grepl("^\\[np\\] Bootstrap replications [0-9]+/9 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bootstrap replications 9/9 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
})

test_that("npqcmstest progress respects np.messages FALSE", {
  skip_if_not_installed("quantreg")
  library(quantreg)

  set.seed(42)
  n <- 40
  x <- rnorm(n)
  y <- 1 + x + rnorm(n, sd = 0.1)
  model <- rq(y ~ x, tau = 0.5, model = TRUE)

  old_opts <- options(np.messages = FALSE)
  on.exit(options(old_opts), add = TRUE)

  res <- capture_progress_shadow_trace(
    npqcmstest_fun(
      model = model,
      xdat = x,
      ydat = y,
      distribution = "bootstrap",
      boot.num = 9
    ),
    now = progress_time_counter()
  )

  expect_length(res$trace, 0)
})

test_that("npqcmstest progress respects suppressMessages", {
  skip_if_not_installed("quantreg")
  library(quantreg)

  set.seed(42)
  n <- 40
  x <- rnorm(n)
  y <- 1 + x + rnorm(n, sd = 0.1)
  model <- rq(y ~ x, tau = 0.5, model = TRUE)

  old_opts <- options(np.messages = TRUE)
  on.exit(options(old_opts), add = TRUE)

  res <- capture_progress_shadow_trace(
    suppressMessages(
      npqcmstest_fun(
        model = model,
        xdat = x,
        ydat = y,
        distribution = "bootstrap",
        boot.num = 9
      )
    ),
    now = progress_time_counter()
  )

  expect_length(res$trace, 0)
})

progress_time_counter <- function(start = 0, by = 0.6) {
  current <- start
  function() {
    current <<- current + by
    current
  }
}

shadow_bootstrap_signature <- function(shadow) {
  lines <- vapply(shadow$trace, `[[`, character(1L), "line")
  events <- vapply(shadow$trace, `[[`, character(1L), "event")
  keep <- grepl("^\\[np\\] Bootstrap replications", lines)

  data.frame(
    event = events[keep],
    line = lines[keep],
    stringsAsFactors = FALSE
  )
}

shadow_lines <- function(shadow) {
  shadow_bootstrap_signature(shadow)$line
}

npsigtest_fun <- function(...) {
  getFromNamespace("npsigtest", "np")(...)
}

make_sigtest_fixture <- function(seed = 42, n = 30) {
  set.seed(seed)
  x1 <- runif(n)
  x2 <- runif(n)
  y <- x1 + rnorm(n, sd = 0.1)
  bw <- getFromNamespace("npregbw", "np")(
    y ~ x1 + x2,
    bws = c(0.2, 0.4),
    bandwidth.compute = FALSE
  )
  list(bw = bw)
}

test_that("npsigtest joint single-line bootstrap progress matches legacy semantics", {
  fixture <- make_sigtest_fixture()

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  legacy <- capture_progress_shadow_trace(
    npsigtest_fun(bws = fixture$bw, boot.num = 9, joint = TRUE, index = 1),
    force_renderer = "legacy",
    now = progress_time_counter()
  )

  single_line <- capture_progress_shadow_trace(
    npsigtest_fun(bws = fixture$bw, boot.num = 9, joint = TRUE, index = 1),
    force_renderer = "single_line",
    now = progress_time_counter()
  )

  lines <- shadow_lines(single_line)

  expect_s3_class(single_line$value, "sigtest")
  expect_equal(shadow_bootstrap_signature(single_line), shadow_bootstrap_signature(legacy))
  expect_true(any(grepl("^\\[np\\] Bootstrap replications [0-9]+/9 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
  expect_true(any(grepl("^\\[np\\] Bootstrap replications 9/9 \\([0-9]+\\.[0-9]%.*, elapsed [0-9]+\\.[0-9]s, eta [0-9]+\\.[0-9]s\\)$", lines)))
})

test_that("npsigtest individual single-line bootstrap progress matches legacy semantics", {
  fixture <- make_sigtest_fixture(seed = 99)

  old_opts <- options(
    np.messages = TRUE,
    np.progress.start.grace.known.sec = 0
  )
  on.exit(options(old_opts), add = TRUE)

  legacy <- capture_progress_shadow_trace(
    npsigtest_fun(bws = fixture$bw, boot.num = 9, joint = FALSE, index = c(1, 2)),
    force_renderer = "legacy",
    now = progress_time_counter()
  )

  single_line <- capture_progress_shadow_trace(
    npsigtest_fun(bws = fixture$bw, boot.num = 9, joint = FALSE, index = c(1, 2)),
    force_renderer = "single_line",
    now = progress_time_counter()
  )

  lines <- shadow_lines(single_line)

  expect_s3_class(single_line$value, "sigtest")
  expect_equal(shadow_bootstrap_signature(single_line), shadow_bootstrap_signature(legacy))
  expect_true(sum(grepl("^\\[np\\] Bootstrap replications [0-9]+/9 ", lines)) >= 2L)
  expect_true(sum(grepl("^\\[np\\] Bootstrap replications 9/9 ", lines)) >= 2L)
})

test_that("npsigtest progress respects np.messages FALSE", {
  fixture <- make_sigtest_fixture(seed = 17)

  old_opts <- options(np.messages = FALSE)
  on.exit(options(old_opts), add = TRUE)

  res <- capture_progress_shadow_trace(
    npsigtest_fun(bws = fixture$bw, boot.num = 9, joint = TRUE, index = 1),
    now = progress_time_counter()
  )

  expect_length(res$trace, 0)
})

test_that("npsigtest progress respects suppressMessages", {
  fixture <- make_sigtest_fixture(seed = 23)

  old_opts <- options(np.messages = TRUE)
  on.exit(options(old_opts), add = TRUE)

  res <- capture_progress_shadow_trace(
    suppressMessages(npsigtest_fun(bws = fixture$bw, boot.num = 9, joint = TRUE, index = 1)),
    now = progress_time_counter()
  )

  expect_length(res$trace, 0)
})

cmstest <- function(Jn, In, Omega.hat, sd,
                    q.90, q.95, q.99, P,
                    bws,
                    distribution,
                    Jn.bootstrap = NA, In.bootstrap = NA,
                    pivot,
                    model,
                    boot.method,
                    boot.num,
                    na.index = NULL){

  tcms = list(Jn = Jn,
    In = In,
    Omega.hat = Omega.hat,
    sd = sd,
    q.90 = q.90,
    q.95 = q.95,
    q.99 = q.99,
    P = P,
    bws = bws,
    Jn.bootstrap = Jn.bootstrap,
    In.bootstrap = In.bootstrap,
    pivot = pivot,
    pdistribution = switch(distribution,
      bootstrap = "Bootstrap",
      asymptotic = "Asymptotic"),
    pcall = paste(deparse(model$call),collapse=""),
    pmethod = switch(boot.method,
      "iid" = "IID",
      "wild" ="Wild",
      "wild-rademacher" = "Rademacher Wild"),
    boot.num = boot.num,
    na.index = na.index)

  ##Sn = if(pivot) Jn else In

  reject <- ' '
  
  if (P < 0.1)
    reject <- '.'

  if (P < 0.05)
    reject <- '*'

  if (P < 0.01)
    reject <- '**'

  if (P < 0.001)
    reject <- '***'

  tcms$reject <- reject
  tcms$rejectNum <- switch(reject,
                           ' ' = 100,
                           '.' = 10,
                           '*' = 5,
                           '**' = 1,
                           '***' = 0.1)
  
  class(tcms) <- "cmstest"

  tcms
}

print.cmstest <- function(x, ...){
  stat.label <- if (x$pivot) "Jn" else "In"
  stat.value <- if (x$pivot) x$Jn else x$In
  reject.msg <- if (x$reject == ' ') {
    "\nFail to reject the null of correct specification at the 10% level"
  } else {
    paste("\nNull of correct specification is rejected at the ", x$rejectNum, "% level", sep = "")
  }
  cat("\nConsistent Model Specification Test\n")
  writeLines(strwrap(paste("Parametric null model: ", x$pcall, sep=""),exdent=26,width=80))
  cat("Number of regressors: ", length(x$bws$bw),"\n",
      if(x$pdistribution == "Bootstrap"){
        paste(x$pmethod, " Bootstrap ",
              "(", x$boot.num, " replications)",sep="")
      } else paste(x$pdistribution, "Distribution"),
      "\n\nTest Statistic ", sQuote(stat.label), ": ",
      format(stat.value), "\tP Value: ", format.pval(x$P)," ", x$reject,
      "\n---\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1",
      reject.msg,
      "\n\n", sep="")
}

summary.cmstest <- function(object, ...){
  print(object)
}

singleindex = 
  function(bws,  betavcov = NULL, index, mean, merr = NA,
           grad = NA, gerr = NA,
           mean.grad = NA, mean.gerr = NA,
           resid = NA,
           ntrain, trainiseval = FALSE, residuals = FALSE,
           gradients = FALSE, xtra = NA,
           confusion.matrix = NA, CCR.overall = NA,
           CCR.byoutcome =  NA, fit.mcfadden = NA,
           timing = NA, total.time = NA,
           optim.time = NA, fit.time = NA
           ){

    if (missing(bws) || missing(index) || missing(mean) || missing(ntrain))
      stop("improper invocation of singleindex constructor")

    d = list(
      beta = bws$beta,
      betavcov = betavcov,
      bw = bws$bw,
      bws = bws,
      pregtype = bws$pregtype,
      pmethod = bws$pmethod,
      xnames = bws$xnames,
      ynames = bws$ynames,
      nobs = dim(index)[1],
      ndim = bws$ndim,
      nord = bws$nord,
      nuno = bws$nuno,
      ncon = bws$ncon,
      pckertype = bws$pckertype,
      index = index,
      mean = mean,
      merr = merr,
      grad = grad,
      gerr = gerr,
      mean.grad = mean.grad,
      mean.gerr = mean.gerr,
      resid = resid,
      ntrain = ntrain,
      trainiseval = trainiseval,
      residuals = residuals,
      gradients = gradients,
      R2 = xtra[1],
      MSE = xtra[2],
      MAE = xtra[3],
      MAPE = xtra[4],
      CORR = xtra[5],
      SIGN = xtra[6],
      confusion.matrix = confusion.matrix,
      CCR.byoutcome = CCR.byoutcome,
      CCR.overall = CCR.overall,
      fit.mcfadden = fit.mcfadden,
      timing = timing,
      total.time = total.time,
      optim.time = optim.time,
      fit.time = fit.time)

    class(d) = "singleindex"

    d
  }

print.singleindex <- function(x, digits=NULL, ...){
  cat("\nSingle Index Model",
      "\nRegression data: ", x$ntrain, " training points,",
      if (x$trainiseval) "" else paste(" and ", x$nobs,
                                      " evaluation points,", sep=""),
      " in ",x$ndim," variable(s)\n",sep="")

  print(matrix(x$beta,ncol=x$ndim,dimnames=list(paste("Beta",":",sep=""),x$xnames)))
  
  cat("Bandwidth:", x$bw)
  cat(genRegEstStr(x))
  cat(genBwKerStrs(x$bws))
  
  cat("\n\n")
  if(!missing(...))
    print(...,digits=digits)
  invisible(x)
}

coef.singleindex <- function(object, ...) {
  tc <- object$beta
  names(tc) <- object$xnames
  return(tc)
}
vcov.singleindex <- function(object, ...) {
  tc <- object$betavcov
  if(!is.null(tc)) {
    return(tc)
  } else {
    .np_warning("variance-covariance matrix does not exist: verify gradients=TRUE")
  }
}
fitted.singleindex <- function(object, ...){
 object$mean 
}
residuals.singleindex <- function(object, ...) {
 if(object$residuals) { return(object$resid) } else { return(npindex(bws = object$bws, residuals =TRUE)$resid) } 
}
predict.singleindex <- function(object, se.fit = FALSE, ...) {
  se.fit <- npValidateScalarLogical(se.fit, "se.fit")
  dots <- list(...)
  has.formula.route <- !is.null(object$bws$formula)

  if (!is.null(dots$exdat) && !is.null(dots$newdata)) {
    dots$newdata <- NULL
  } else if (!has.formula.route && is.null(dots$exdat) && !is.null(dots$newdata)) {
    dots$exdat <- dots$newdata
    dots$newdata <- NULL
  }

  ## When no new evaluation inputs are supplied, reuse stored fit
  ## directly. This avoids unnecessary recomputation.
  if (length(dots) == 0L) {
    if (se.fit) {
      return(list(
        fit = fitted(object),
        se.fit = se(object),
        df = object$nobs,
        residual.scale = object$MSE
      ))
    }
    return(fitted(object))
  }

  tr <- do.call(npindex, c(list(bws = object$bws, errors = se.fit), dots))
  if (se.fit)
    return(list(fit = fitted(tr), se.fit = se(tr),
                df = tr$nobs, residual.scale = tr$MSE))
  fitted(tr)
}
se.singleindex <- function(x){ x$merr }
gradients.singleindex <- function(x, errors = FALSE, ...) {
  errors <- npValidateScalarLogical(errors, "errors")
  gout <- if (!errors) x$grad else x$gerr
  if (is.null(gout) || (length(gout) == 1L && is.logical(gout) && is.na(gout)))
    stop(if (!errors)
      "gradients are not available: fit the model with gradients=TRUE"
    else
      "gradient standard errors are not available: fit the model with gradients=TRUE")
  gout
}

summary.singleindex <- function(object, ...){
  cat("\nSingle Index Model",
      "\nRegression Data: ", object$ntrain, " training points,",
      if (object$trainiseval) "" else paste(" and ", object$nobs,
                                      " evaluation points,", sep=""),
      " in ",object$ndim," variable(s)\n\n",sep="")

  cat(genOmitStr(object))
  print(matrix(object$beta,ncol=object$ndim,dimnames=list(paste("Beta",":",sep=""),object$xnames)))
  
  cat("Bandwidth:", object$bw)
  cat(genRegEstStr(object))
  cat("\n")
  cat(genGofStr(object))
  pCatGofStr(object)
  cat(genBwKerStrs(object$bws))
  cat(genTimingStr(object))
  
  cat("\n\n")
}

npcdens_conbandwidth <- getFromNamespace("npcdens.conbandwidth", "np")
npcdist_condbandwidth <- getFromNamespace("npcdist.condbandwidth", "np")
npqreg_condbandwidth <- getFromNamespace("npqreg.condbandwidth", "np")
npindex_sibandwidth <- getFromNamespace("npindex.sibandwidth", "np")
npscoef_scbandwidth <- getFromNamespace("npscoef.scbandwidth", "np")
npplreg_plbandwidth <- getFromNamespace("npplreg.plbandwidth", "np")
np_validate_scalar_logical <- getFromNamespace("npValidateScalarLogical", "np")

test_that("npValidateScalarLogical enforces scalar logical flags", {
  expect_identical(np_validate_scalar_logical(TRUE, "x"), TRUE)
  expect_identical(np_validate_scalar_logical(1, "x"), TRUE)
  expect_identical(np_validate_scalar_logical(0, "x"), FALSE)
  expect_error(np_validate_scalar_logical(c(TRUE, FALSE), "x"), "'x' must be TRUE or FALSE")
  expect_error(np_validate_scalar_logical("foo", "x"), "'x' must be TRUE or FALSE")
})

test_that("core estimator methods reject non-scalar logical control flags", {
  tx <- data.frame(x = c(0.1, 0.2))
  ty <- c(0.3, 0.4)
  tydf <- data.frame(y = ty)

  expect_error(
    npcdens_conbandwidth(structure(list(), class = "conbandwidth"),
                         txdat = tx, tydat = tydf, gradients = c(TRUE, FALSE)),
    "'gradients' must be TRUE or FALSE"
  )
  expect_error(
    npcdist_condbandwidth(structure(list(), class = "condbandwidth"),
                          txdat = tx, tydat = tydf, gradients = c(TRUE, FALSE)),
    "'gradients' must be TRUE or FALSE"
  )
  expect_error(
    npqreg_condbandwidth(structure(list(), class = "condbandwidth"),
                         txdat = tx, tydat = ty, gradients = c(TRUE, FALSE)),
    "'gradients' must be TRUE or FALSE"
  )
  expect_error(
    npqreg_condbandwidth(structure(list(), class = "condbandwidth"),
                         txdat = tx, tydat = ty, itmax = 0),
    "'itmax' must be a positive integer"
  )
  expect_error(
    npqreg_condbandwidth(structure(list(), class = "condbandwidth"),
                         txdat = tx, tydat = ty, ftol = 0),
    "'ftol' is no longer accepted by npqreg"
  )
  expect_error(
    npqreg_condbandwidth(structure(list(), class = "condbandwidth"),
                         txdat = tx, tydat = ty, foo = TRUE),
    "unused argument in npqreg: 'foo'"
  )
  expect_error(
    npindex_sibandwidth(structure(list(), class = "sibandwidth"),
                        txdat = tx, tydat = ty, gradients = c(TRUE, FALSE)),
    "'gradients' must be TRUE or FALSE"
  )
  expect_error(
    npindex_sibandwidth(structure(list(), class = "sibandwidth"),
                        txdat = tx, tydat = ty, boot.num = 0),
    "'boot.num' must be a positive integer"
  )
  expect_error(
    npscoef_scbandwidth(structure(list(), class = "scbandwidth"),
                        txdat = tx, tydat = ty, iterate = c(TRUE, FALSE)),
    "'iterate' must be TRUE or FALSE"
  )
  expect_error(
    npscoef_scbandwidth(structure(list(), class = "scbandwidth"),
                        txdat = tx, tydat = ty, maxiter = 0),
    "'maxiter' must be a positive integer"
  )
  expect_error(
    npscoef_scbandwidth(structure(list(), class = "scbandwidth"),
                        txdat = tx, tydat = ty, tol = -1),
    "'tol' must be a finite numeric scalar >= 0"
  )
  expect_error(
    npplreg_plbandwidth(structure(list(), class = "plbandwidth"),
                        txdat = tx, tydat = ty, tzdat = tx, residuals = c(TRUE, FALSE)),
    "'residuals' must be TRUE or FALSE"
  )
})

test_that("plot.par.mfrow=FALSE is honored for npcdens auto and manual plots", {
  skip_if_not_installed("grDevices")

  old.opts <- options(plot.par.mfrow = FALSE)
  on.exit(options(old.opts), add = TRUE)

  set.seed(101)
  n <- 40
  x <- runif(n)
  y <- rbeta(n, 1, 1)

  f.auto <- npcdens(
    y ~ x,
    cykerbound = "range",
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "cell",
    degree.max = 1,
    degree.verify = FALSE,
    bwmethod = "cv.ml",
    nmulti = 1
  )

  f.manual <- npcdens(
    y ~ x,
    cykerbound = "range",
    regtype = "lp",
    degree = 1,
    bwmethod = "cv.ml",
    nmulti = 1
  )

  grDevices::pdf(NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  par(mfrow = c(2, 2))
  plot(1:10)
  plot(1:10)
  suppressWarnings(
    plot(f.auto, theta = 70, phi = 10, view = "fixed", zlim = c(0, 2))
  )
  expect_identical(par("mfrow"), c(2L, 2L))
  expect_identical(par("mfg"), c(2L, 1L, 2L, 2L))

  suppressWarnings(
    plot(f.manual, theta = 70, phi = 10, view = "fixed", zlim = c(0, 2))
  )
  expect_identical(par("mfrow"), c(2L, 2L))
  expect_identical(par("mfg"), c(2L, 2L, 2L, 2L))
})

library(np)

gaussian_row_weights <- function(x, j, bw, extra = rep(1, length(x))) {
  idx <- setdiff(seq_along(x), j)
  u <- (x[idx] - x[[j]]) / bw
  exp(-0.5 * u * u) * extra[idx]
}

degree1_raw_moments <- function(x, y, j, weights) {
  idx <- setdiff(seq_along(x), j)
  xi <- x[idx]
  yi <- y[idx]
  bi <- cbind(1, xi)
  s <- crossprod(bi * weights, bi)
  t <- crossprod(bi * weights, yi)
  list(s = unclass(s), t = as.numeric(t))
}

degree1_centered_moments <- function(xj, raw) {
  s <- raw$s
  t <- raw$t
  k <- nrow(s)
  out_s <- matrix(0, nrow = k, ncol = k)
  out_t <- numeric(k)

  out_t[1] <- t[1]
  out_s[1, 1] <- s[1, 1]

  for (a in 2:k) {
    out_t[a] <- t[a] - xj[a - 1L] * t[1]
    out_s[1, a] <- s[1, a] - xj[a - 1L] * s[1, 1]
    out_s[a, 1] <- out_s[1, a]
  }

  for (a in 2:k) {
    sa0 <- s[a, 1]
    for (b in 2:k) {
      out_s[a, b] <- s[a, b] -
        xj[a - 1L] * s[1, b] -
        xj[b - 1L] * sa0 +
        xj[a - 1L] * xj[b - 1L] * s[1, 1]
    }
  }

  list(s = out_s, t = out_t)
}

solve_degree1_pair <- function(x, y, bw, extra = rep(1, length(x))) {
  n <- length(x)
  raw_fit <- numeric(n)
  centered_fit <- numeric(n)
  raw_cv <- numeric(n)
  centered_cv <- numeric(n)

  for (j in seq_len(n)) {
    w <- gaussian_row_weights(x, j, bw, extra = extra)
    raw <- degree1_raw_moments(x, y, j, w)
    beta_raw <- solve(raw$s, raw$t)
    centered <- degree1_centered_moments(x[j], raw)
    beta_centered <- solve(centered$s, centered$t)

    raw_fit[j] <- c(1, x[j]) %*% beta_raw
    centered_fit[j] <- beta_centered[1]
    raw_cv[j] <- (y[j] - raw_fit[j])^2
    centered_cv[j] <- (y[j] - centered_fit[j])^2
  }

  list(
    raw_fit = raw_fit,
    centered_fit = centered_fit,
    raw_cv = raw_cv,
    centered_cv = centered_cv
  )
}

solve_degree1_one <- function(x, y, j, bw, extra = rep(1, length(x))) {
  w <- gaussian_row_weights(x, j, bw, extra = extra)
  raw <- degree1_raw_moments(x, y, j, w)
  beta_raw <- solve(raw$s, raw$t)
  centered <- degree1_centered_moments(x[j], raw)
  beta_centered <- solve(centered$s, centered$t)

  list(
    raw_fit = as.numeric(c(1, x[j]) %*% beta_raw),
    centered_fit = as.numeric(beta_centered[1]),
    raw_cv = (y[j] - as.numeric(c(1, x[j]) %*% beta_raw))^2,
    centered_cv = (y[j] - as.numeric(beta_centered[1]))^2
  )
}

build_glp_terms_test <- function(degree) {
  k <- length(degree)
  if (k == 0L)
    return(matrix(integer(0), nrow = 1L, ncol = 0L))

  degree <- as.integer(degree)
  z <- as.matrix(do.call(expand.grid, lapply(degree, function(d) 0:d)))
  s <- rowSums(z)
  z <- z[(s > 0L) & (s <= max(degree)), , drop = FALSE]

  if (!all(degree == max(degree))) {
    for (j in seq_along(degree)) {
      d <- degree[j]
      if ((d < max(degree)) && (d > 0L)) {
        s <- rowSums(z)
        dropj <- (s > d) &
          (z[, j, drop = FALSE] == matrix(d, nrow(z), 1L, byrow = TRUE))
        z <- z[!dropj, , drop = FALSE]
      }
    }
  }

  rbind(matrix(0L, nrow = 1L, ncol = k), z)
}

eval_monomial_basis_test <- function(X, terms) {
  X <- as.matrix(X)
  out <- matrix(1, nrow = nrow(X), ncol = nrow(terms))
  for (j in seq_len(nrow(terms))) {
    exponents <- terms[j, ]
    nz <- which(exponents > 0L)
    if (length(nz)) {
      out[, j] <- apply(
        X[, nz, drop = FALSE]^matrix(exponents[nz], nrow(X), length(nz), byrow = TRUE),
        1L,
        prod
      )
    }
  }
  out
}

build_shift_raw_from_center_test <- function(terms, xj) {
  keys <- setNames(seq_len(nrow(terms)), apply(terms, 1L, paste, collapse = ","))
  out <- matrix(0, nrow = nrow(terms), ncol = nrow(terms))

  recurse <- function(alpha, pos = 1L, beta = integer(length(alpha)), acc = list()) {
    if (pos > length(alpha)) {
      acc[[length(acc) + 1L]] <- beta
      return(acc)
    }
    for (v in 0:alpha[pos]) {
      beta[pos] <- v
      acc <- recurse(alpha, pos + 1L, beta, acc)
    }
    acc
  }

  for (r in seq_len(nrow(terms))) {
    alpha <- terms[r, ]
    for (beta in recurse(alpha)) {
      cidx <- keys[[paste(beta, collapse = ",")]]
      out[r, cidx] <- out[r, cidx] + prod(choose(alpha, beta) * xj^(alpha - beta))
    }
  }

  out
}

solve_glp_raw_center_pair <- function(X, y, degree, bw, extra = rep(1, nrow(X))) {
  terms <- build_glp_terms_test(degree)
  B <- eval_monomial_basis_test(X, terms)
  n <- nrow(B)
  raw_fit <- numeric(n)
  centered_fit <- numeric(n)
  transformed_fit <- numeric(n)
  press_fit <- numeric(n)

  for (j in seq_len(n)) {
    idx <- setdiff(seq_len(n), j)
    u <- sweep(X[idx, , drop = FALSE], 2L, X[j, ], "-")
    w <- exp(-0.5 * rowSums((u / matrix(bw, nrow(u), ncol(u), byrow = TRUE))^2)) * extra[idx]
    Bj <- B[idx, , drop = FALSE]
    sys_raw <- list(
      s = crossprod(Bj, Bj * w),
      t = crossprod(Bj, y[idx] * w)
    )
    beta_raw <- solve(sys_raw$s, sys_raw$t)
    raw_fit[j] <- drop(B[j, ] %*% beta_raw)

    Cj <- eval_monomial_basis_test(u, terms)
    sys_center <- list(
      s = crossprod(Cj, Cj * w),
      t = crossprod(Cj, y[idx] * w)
    )
    beta_center <- solve(sys_center$s, sys_center$t)
    centered_fit[j] <- beta_center[1]

    shift <- build_shift_raw_from_center_test(terms, X[j, ])
    inv_shift <- solve(shift)
    s_transform <- inv_shift %*% sys_raw$s %*% t(inv_shift)
    t_transform <- inv_shift %*% sys_raw$t
    beta_transform <- solve(s_transform, t_transform)
    transformed_fit[j] <- beta_transform[1]
  }

  list(
    raw_fit = raw_fit,
    centered_fit = centered_fit,
    transformed_fit = transformed_fit,
    raw_cv = (y - raw_fit)^2,
    centered_cv = (y - centered_fit)^2,
    transformed_cv = (y - transformed_fit)^2
  )
}

press_global_glp_test <- function(X, y, degree) {
  terms <- build_glp_terms_test(degree)
  B <- eval_monomial_basis_test(X, terms)
  XtX_inv <- solve(crossprod(B))
  beta <- XtX_inv %*% crossprod(B, y)
  fit <- drop(B %*% beta)
  hdiag <- rowSums((B %*% XtX_inv) * B)
  sum(((y - fit) / (1 - hdiag))^2)
}

const_weight_loocv_glp_test <- function(X, y, degree) {
  terms <- build_glp_terms_test(degree)
  B <- eval_monomial_basis_test(X, terms)
  n <- nrow(B)
  fit <- numeric(n)
  for (j in seq_len(n)) {
    idx <- setdiff(seq_len(n), j)
    beta <- solve(crossprod(B[idx, , drop = FALSE]), crossprod(B[idx, , drop = FALSE], y[idx]))
    fit[j] <- drop(B[j, ] %*% beta)
  }
  sum((y - fit)^2)
}

test_that("degree-1 raw-basis and centered solves are identical for continuous fixed-bandwidth LOO fits", {
  set.seed(20260306)
  x <- sort(runif(40L))
  y <- sin(2 * pi * x) + rnorm(40L, sd = 0.03)

  res <- solve_degree1_pair(x, y, bw = 0.19)

  expect_equal(res$raw_fit, res$centered_fit, tolerance = 1e-12)
  expect_equal(res$raw_cv, res$centered_cv, tolerance = 1e-12)
})

test_that("degree-1 basis-change invariant holds under mixed-data style scalar kernel modifiers", {
  set.seed(20260306)
  x <- sort(runif(45L))
  y <- cos(2 * pi * x) + rnorm(45L, sd = 0.04)
  u <- factor(sample(c("a", "b", "c"), length(x), replace = TRUE))
  o <- ordered(sample(1:4, length(x), replace = TRUE))

  extra <- numeric(length(x))
  for (j in seq_along(x)) {
    same_u <- ifelse(u == u[j], 1.0, 0.35)
    ord_gap <- abs(as.integer(o) - as.integer(o[j]))
    mixed_extra <- same_u * (0.8 ^ ord_gap)
    res_j <- solve_degree1_one(x, y, j, bw = 0.17, extra = mixed_extra)
    extra[j] <- abs(res_j$raw_fit - res_j$centered_fit)
    expect_equal(res_j$raw_fit, res_j$centered_fit, tolerance = 1e-12)
    expect_equal(res_j$raw_cv, res_j$centered_cv, tolerance = 1e-12)
  }

  expect_lte(max(extra), 1e-12)
})

test_that("higher-degree GLP raw, centered, and transformed CVLS fits agree", {
  set.seed(20260306)
  X <- cbind(
    x1 = runif(24L, min = -1, max = 1),
    x2 = runif(24L, min = -1, max = 1),
    x3 = runif(24L, min = -1, max = 1)
  )
  y <- 0.3 + X[, 1] - 0.4 * X[, 2] * X[, 3] + 0.25 * X[, 1]^2 + rnorm(24L, sd = 0.02)

  res <- solve_glp_raw_center_pair(X, y, degree = c(2L, 1L, 1L), bw = c(0.75, 0.75, 0.75))

  expect_equal(res$raw_fit, res$centered_fit, tolerance = 1e-10)
  expect_equal(res$raw_fit, res$transformed_fit, tolerance = 1e-10)
  expect_equal(res$raw_cv, res$centered_cv, tolerance = 1e-10)
  expect_equal(res$raw_cv, res$transformed_cv, tolerance = 1e-10)
})

test_that("higher-degree GLP basis-change invariant survives mixed-data style scalar modifiers", {
  set.seed(20260306)
  X <- cbind(
    x1 = runif(22L, min = -1, max = 1),
    x2 = runif(22L, min = -1, max = 1)
  )
  y <- 0.5 + X[, 1] + 0.2 * X[, 2]^2 + rnorm(22L, sd = 0.03)
  u <- factor(sample(c("a", "b"), nrow(X), replace = TRUE))
  o <- ordered(sample(1:3, nrow(X), replace = TRUE))
  extra <- ifelse(u == "a", 1, 0.45) * (0.8 ^ abs(as.integer(o) - 2L))

  res <- solve_glp_raw_center_pair(X, y, degree = c(2L, 1L), bw = c(0.7, 0.7), extra = extra)

  expect_equal(res$raw_fit, res$centered_fit, tolerance = 1e-10)
  expect_equal(res$raw_fit, res$transformed_fit, tolerance = 1e-10)
})

test_that("large-h GLP CVLS collapses to global polynomial PRESS", {
  set.seed(20260306)
  X <- cbind(
    x1 = runif(20L, min = -1, max = 1),
    x2 = runif(20L, min = -1, max = 1)
  )
  y <- 0.4 + 0.8 * X[, 1] - 0.3 * X[, 2] + 0.25 * X[, 1]^2 + rnorm(20L, sd = 0.01)

  expect_equal(
    const_weight_loocv_glp_test(X, y, degree = c(2L, 2L)),
    press_global_glp_test(X, y, degree = c(2L, 2L)),
    tolerance = 1e-10
  )
})

test_that("conditional bootstrap target labels format compactly", {
  fmt <- getFromNamespace(".np_plot_conditional_bootstrap_target_label", "np")

  bws <- list(
    xndim = 2L,
    yndim = 1L,
    xnames = c("x1", "x2"),
    ynames = "y"
  )

  expect_identical(fmt(bws = bws, slice.index = 0L, gradients = FALSE), "surf 1/3")
  expect_identical(fmt(bws = bws, slice.index = 1L, gradients = FALSE), "x1 1/3")
  expect_identical(fmt(bws = bws, slice.index = 3L, gradients = FALSE), "y 3/3")
  expect_identical(fmt(bws = bws, slice.index = 2L, gradients = TRUE, gradient.index = 1L), "grad x1 on x2 2/3")
})

test_that("conditional helper labels carry target context for block bootstrap phases", {
  fn <- getFromNamespace("compute.bootstrap.errors.conbandwidth", "np")

  captured <- new.env(parent = emptyenv())
  captured$prep <- character()
  captured$progress <- character()
  captured$interval <- character()

  orig_activity_begin <- getFromNamespace(".np_plot_activity_begin", "np")
  orig_activity_end <- getFromNamespace(".np_plot_activity_end", "np")
  orig_grad <- getFromNamespace(".np_inid_boot_from_conditional_gradient_local", "np")
  orig_interval <- getFromNamespace(".np_plot_bootstrap_interval_summary", "np")

  assignInNamespace(".np_plot_activity_begin", function(label) {
    captured$prep <- c(captured$prep, label)
    list(label = label)
  }, ns = "np")
  assignInNamespace(".np_plot_activity_end", function(state) invisible(NULL), ns = "np")
  assignInNamespace(".np_inid_boot_from_conditional_gradient_local", function(..., progress.label = NULL) {
    captured$progress <- c(captured$progress, progress.label)
    list(t = matrix(0, nrow = 2L, ncol = 3L), t0 = c(0, 0, 0))
  }, ns = "np")
  assignInNamespace(".np_plot_bootstrap_interval_summary", function(boot.t, t0, alpha, band.type, progress.label = NULL) {
    captured$interval <- c(captured$interval, progress.label)
    list(err = matrix(0, nrow = ncol(boot.t), ncol = 2L), all.err = list())
  }, ns = "np")
  on.exit({
    assignInNamespace(".np_plot_activity_begin", orig_activity_begin, ns = "np")
    assignInNamespace(".np_plot_activity_end", orig_activity_end, ns = "np")
    assignInNamespace(".np_inid_boot_from_conditional_gradient_local", orig_grad, ns = "np")
    assignInNamespace(".np_plot_bootstrap_interval_summary", orig_interval, ns = "np")
  }, add = TRUE)

  xdat <- data.frame(x1 = c(0, 1, 2), x2 = c(1, 2, 3))
  ydat <- data.frame(y = c(0.1, 0.2, 0.3))
  bws <- list(
    type = "fixed",
    xndim = 2L,
    yndim = 1L,
    xnames = c("x1", "x2"),
    ynames = "y",
    xdati = list(),
    ydati = list()
  )

  out <- fn(
    xdat = xdat,
    ydat = ydat,
    exdat = xdat,
    eydat = ydat,
    cdf = FALSE,
    quantreg = FALSE,
    tau = NULL,
    gradients = TRUE,
    gradient.index = 1L,
    slice.index = 2L,
    plot.errors.boot.method = "geom",
    plot.errors.boot.nonfixed = "exact",
    plot.errors.boot.blocklen = 2L,
    plot.errors.boot.num = 2L,
    plot.errors.center = "estimate",
    plot.errors.type = "all",
    plot.errors.alpha = 0.05,
    progress.target = "grad x1 on x2 2/3",
    bws = bws
  )

  expect_true(is.list(out))
  expect_true(any(grepl("Preparing plot bootstrap geom \\(grad x1 on x2 2/3\\)", captured$prep)))
  expect_true(any(grepl("Plot bootstrap \\(grad x1 on x2 2/3\\)", captured$progress)))
  expect_true(any(grepl("Constructing bootstrap all bands \\(grad x1 on x2 2/3\\)", captured$interval)))
})

test_that("npcdensbw NOMAD degree search backend improves over the baseline", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(x = sort(runif(14)))
  dat$y <- dat$x + rnorm(nrow(dat), sd = 0.08)

  bw <- np::npcdensbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "nomad",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_s3_class(bw, "conbandwidth")
  expect_identical(bw$degree.search$mode, "nomad")
  expect_true(isTRUE(bw$degree.search$completed))
  expect_gte(bw$degree.search$n.unique, 1L)
  expect_gte(bw$degree.search$best.fval, bw$degree.search$baseline.fval - 1e-10)
})

library(np)

public_shadow_empty <- function(n) {
  matrix(numeric(0), nrow = n, ncol = 0)
}

public_shadow_cker <- function(kernel) {
  switch(kernel,
    gaussian = 0L,
    epanechnikov = 4L,
    uniform = 8L,
    truncated = 9L,
    stop("unsupported continuous kernel")
  )
}

public_shadow_uker <- function(kernel) {
  switch(kernel,
    aitchisonaitken = 0L,
    liracine = 1L,
    stop("unsupported unordered kernel")
  )
}

public_shadow_oker <- function(kernel) {
  switch(kernel,
    wangvanryzin = 0L,
    liracine = 2L,
    racineliyan = 3L,
    stop("unsupported ordered kernel")
  )
}

public_shadow_rbw <- function(bw) {
  c(
    bw$xbw[bw$ixcon],
    bw$ybw[bw$iycon],
    bw$ybw[bw$iyuno],
    bw$ybw[bw$iyord],
    bw$xbw[bw$ixuno],
    bw$xbw[bw$ixord]
  )
}

public_shadow_bwtype <- function(bw) {
  switch(bw$type,
    fixed = 0L,
    generalized_nn = 1L,
    adaptive_nn = 2L,
    stop("unsupported bandwidth type")
  )
}

public_shadow_regtype <- function(bw) {
  if (identical(bw$regtype.engine, "lp")) 2L else 0L
}

public_shadow_degree <- function(bw) {
  if (identical(bw$regtype.engine, "lp")) as.integer(bw$degree.engine) else integer(0)
}

public_shadow_basis <- function(basis_engine, regtype_engine) {
  if (!identical(regtype_engine, "lp")) {
    return(0L)
  }
  switch(basis_engine,
    additive = 0L,
    glp = 1L,
    tensor = 2L,
    stop("unsupported LP basis")
  )
}

public_shadow_safe_call <- function(name, ...) {
  on.exit(
    tryCatch(.Call("C_np_shadow_reset_state", PACKAGE = "np"),
             error = function(e) NULL),
    add = TRUE
  )
  .Call(name, ..., PACKAGE = "np")
}

call_public_cvml_shadow <- function(bw, x, y, tree = FALSE, compare_old = TRUE) {
  n <- nrow(x)
  public_shadow_safe_call(
    "C_np_shadow_cv_density_conditional",
    public_shadow_empty(n), public_shadow_empty(n), as.matrix(y),
    public_shadow_empty(n), public_shadow_empty(n), as.matrix(x),
    as.double(public_shadow_rbw(bw)),
    public_shadow_bwtype(bw),
    public_shadow_cker(bw$cykertype),
    public_shadow_uker(bw$uykertype),
    public_shadow_oker(bw$oykertype),
    public_shadow_cker(bw$cxkertype),
    public_shadow_uker(bw$uxkertype),
    public_shadow_oker(bw$oxkertype),
    tree,
    0L,
    public_shadow_regtype(bw),
    public_shadow_degree(bw),
    isTRUE(bw$bernstein.basis.engine),
    public_shadow_basis(bw$basis.engine, bw$regtype.engine),
    compare_old
  )
}

test_that("public npcdensbw cv.ml keeps lc on the legacy objective", {
  set.seed(202)
  n <- 32L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = rnorm(n))

  bw.lc <- npcdensbw(xdat = x, ydat = y, regtype = "lc", bwmethod = "cv.ml", nmulti = 1)
  shadow <- call_public_cvml_shadow(bw.lc, x, y)

  expect_equal(-bw.lc$fval, shadow$old, tolerance = 1e-10)
})

test_that("public npcdensbw cv.ml fixed LP/LL route activates with ll == lp parity", {
  set.seed(101)
  n <- 36L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = x$x1 - x$x2 + rnorm(n, sd = 0.2))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdensbw(
    xdat = x,
    ydat = y,
    regtype = "ll",
    bwmethod = "cv.ml",
    nmulti = 1
  )
  bw.lp <- npcdensbw(
    xdat = x,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = degree,
    bwmethod = "cv.ml",
    nmulti = 1
  )
 
  expect_equal(-bw.ll$fval, -bw.lp$fval, tolerance = 1e-8)
})

test_that("public npcdensbw cv.ml fixed LP route preserves tree parity", {
  set.seed(102)
  n <- 34L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = sin(2 * pi * x$x1) + x$x2 + rnorm(n, sd = 0.15))
  degree <- rep.int(1L, ncol(x))

  bw.serial <- npcdensbw(
    xdat = x,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = degree,
    bwmethod = "cv.ml",
    nmulti = 1
  )

  old_opt <- getOption("np.tree")
  on.exit(options(np.tree = old_opt), add = TRUE)
  options(np.tree = TRUE)

  bw.tree <- npcdensbw(
    xdat = x,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = degree,
    bwmethod = "cv.ml",
    nmulti = 1
  )

  expect_equal(-bw.tree$fval, -bw.serial$fval, tolerance = 1e-8)
})

test_that("public npcdensbw cv.ml generalized-nn LP route activates with ll == lp parity", {
  set.seed(103)
  n <- 36L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = x$x1 - x$x2 + rnorm(n, sd = 0.2))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdensbw(
    xdat = x,
    ydat = y,
    regtype = "ll",
    bwtype = "generalized_nn",
    bwmethod = "cv.ml",
    nmulti = 1
  )
  bw.lp <- npcdensbw(
    xdat = x,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = degree,
    bwtype = "generalized_nn",
    bwmethod = "cv.ml",
    nmulti = 1
  )

  expect_true(is.finite(bw.ll$fval))
  expect_true(is.finite(bw.lp$fval))
  expect_equal(bw.ll$fval, bw.lp$fval, tolerance = 1e-8)
})

test_that("public npcdensbw cv.ml adaptive-nn LP route activates with ll == lp parity", {
  set.seed(104)
  n <- 36L
  x <- data.frame(x1 = runif(n), x2 = runif(n))
  y <- data.frame(y1 = x$x1 - x$x2 + rnorm(n, sd = 0.2))
  degree <- rep.int(1L, ncol(x))

  bw.ll <- npcdensbw(
    xdat = x,
    ydat = y,
    regtype = "ll",
    bwtype = "adaptive_nn",
    bwmethod = "cv.ml",
    nmulti = 1
  )
  bw.lp <- npcdensbw(
    xdat = x,
    ydat = y,
    regtype = "lp",
    basis = "glp",
    degree = degree,
    bwtype = "adaptive_nn",
    bwmethod = "cv.ml",
    nmulti = 1
  )

  expect_true(is.finite(bw.ll$fval))
  expect_true(is.finite(bw.lp$fval))
  expect_equal(bw.ll$fval, bw.lp$fval, tolerance = 1e-8)
})

test_that("npcdens stores continuous scale factors distinctly from bandwidths", {
  set.seed(42)
  n <- 80
  x <- runif(n)
  y <- rbeta(n, 1, 1)

  fit <- npcdens(
    y ~ x,
    cxkerbound = "range",
    cykerbound = "range",
    regtype = "lp",
    degree = 3,
    bwmethod = "cv.ml",
    nmulti = 1
  )

  bw <- fit$bws
  expect_false(isTRUE(all.equal(unname(bw$bandwidth$x), unname(bw$sfactor$x))))
  expect_false(isTRUE(all.equal(unname(bw$bandwidth$y), unname(bw$sfactor$y))))

  expected.x <- unname(bw$xbw / (as.numeric(bw$sdev[seq_len(bw$xncon)]) * bw$nconfac))
  expected.y <- unname(bw$ybw / (as.numeric(bw$sdev[bw$xncon + seq_len(bw$yncon)]) * bw$nconfac))

  expect_equal(unname(bw$sfactor$x), expected.x, tolerance = 1e-10)
  expect_equal(unname(bw$sfactor$y), expected.y, tolerance = 1e-10)
  expect_equal(unname(bw$sumNum$x), expected.x, tolerance = 1e-10)
  expect_equal(unname(bw$sumNum$y), expected.y, tolerance = 1e-10)
})

test_that("npcdens cell cv.ml exhaustive search maximizes over degrees", {
  set.seed(42)
  n <- 50
  x <- runif(n)
  y <- rbeta(n, 1, 1)

  manual <- lapply(0:3, function(d) {
    npcdens(
      txdat = data.frame(x = x),
      tydat = data.frame(y = y),
      cxkerbound = "range",
      cykerbound = "range",
      regtype = "lp",
      degree = d,
      bwtype = "fixed",
      bwmethod = "cv.ml",
      nmulti = 1
    )$bws
  })
  manual.fval <- vapply(manual, function(bw) as.numeric(bw$fval[1L]), numeric(1L))
  expected.degree <- which.max(manual.fval) - 1L

  auto <- npcdens(
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    cxkerbound = "range",
    cykerbound = "range",
    regtype = "lp",
    degree.min = 0,
    degree.max = 3,
    degree.select = "exhaustive",
    search.engine = "cell",
    bwtype = "fixed",
    bwmethod = "cv.ml",
    nmulti = 1
  )

  expect_true(abs(manual.fval[auto$bws$degree + 1L] - max(manual.fval)) <= 1e-8)
  expect_equal(as.numeric(auto$bws$fval[1L]), max(manual.fval), tolerance = 1e-8)
})

test_that("npcdens nomad cv.ml retains the best objective found", {
  set.seed(42)
  n <- 60
  x <- runif(n)
  y <- rbeta(n, 1, 1)

  auto <- npcdens(
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    cxkerbound = "range",
    cykerbound = "range",
    regtype = "lp",
    degree.min = 0,
    degree.max = 3,
    degree.select = "coordinate",
    search.engine = "nomad+powell",
    degree.verify = FALSE,
    bwtype = "fixed",
    bwmethod = "cv.ml",
    nmulti = 2
  )

  expect_true(is.finite(auto$bws$degree.search$baseline.fval))
  expect_true(is.finite(auto$bws$degree.search$best.fval))
  expect_true(auto$bws$degree.search$best.fval >= auto$bws$degree.search$baseline.fval - 1e-8)
  expect_equal(as.numeric(auto$bws$fval[1L]), as.numeric(auto$bws$degree.search$best.fval), tolerance = 1e-8)
  expect_identical(auto$bws$degree.search$direction, "max")
  expect_identical(
    as.numeric(auto$bws$nomad.restart.fval),
    as.numeric(vapply(auto$bws$nomad.restart.results, `[[`, numeric(1L), "objective"))
  )
  expect_equal(as.integer(auto$bws$nomad.best.restart), as.integer(which.max(auto$bws$nomad.restart.fval)))
  expect_lte(
    auto$bws$nomad.restart.fval[auto$bws$nomad.best.restart],
    as.numeric(auto$bws$degree.search$best.fval) + 1e-6
  )
})

test_that("regression bootstrap target labels format compactly", {
  fmt <- getFromNamespace(".np_plot_regression_bootstrap_target_label", "np")

  bws <- list(
    ndim = 2L,
    xnames = c("x1", "x2")
  )

  expect_identical(fmt(bws = bws, slice.index = 0L, gradients = FALSE), "surf 1/2")
  expect_identical(fmt(bws = bws, slice.index = 1L, gradients = FALSE), "x1 1/2")
  expect_identical(fmt(bws = bws, slice.index = 2L, gradients = TRUE), "grad x2 2/2")
})

test_that("regression helper labels carry target context for block bootstrap phases", {
  fn <- getFromNamespace("compute.bootstrap.errors.rbandwidth", "np")

  captured <- new.env(parent = emptyenv())
  captured$prep <- character()
  captured$progress <- character()
  captured$interval <- character()

  orig_activity_begin <- getFromNamespace(".np_plot_activity_begin", "np")
  orig_activity_end <- getFromNamespace(".np_plot_activity_end", "np")
  orig_reg_boot <- getFromNamespace(".np_inid_boot_from_regression", "np")
  orig_interval <- getFromNamespace(".np_plot_bootstrap_interval_summary", "np")

  assignInNamespace(".np_plot_activity_begin", function(label) {
    captured$prep <- c(captured$prep, label)
    list(label = label)
  }, ns = "np")
  assignInNamespace(".np_plot_activity_end", function(state) invisible(NULL), ns = "np")
  assignInNamespace(".np_inid_boot_from_regression", function(..., prep.label = NULL, progress.label = NULL) {
    captured$progress <- c(captured$progress, prep.label, progress.label)
    list(t = matrix(0, nrow = 2L, ncol = 3L), t0 = c(0, 0, 0))
  }, ns = "np")
  assignInNamespace(".np_plot_bootstrap_interval_summary", function(boot.t, t0, alpha, band.type, progress.label = NULL) {
    captured$interval <- c(captured$interval, progress.label)
    list(err = matrix(0, nrow = ncol(boot.t), ncol = 2L), all.err = list())
  }, ns = "np")
  on.exit({
    assignInNamespace(".np_plot_activity_begin", orig_activity_begin, ns = "np")
    assignInNamespace(".np_plot_activity_end", orig_activity_end, ns = "np")
    assignInNamespace(".np_inid_boot_from_regression", orig_reg_boot, ns = "np")
    assignInNamespace(".np_plot_bootstrap_interval_summary", orig_interval, ns = "np")
  }, add = TRUE)

  xdat <- data.frame(x1 = c(0, 1, 2), x2 = c(1, 2, 3))
  bws <- list(
    type = "fixed",
    ndim = 2L,
    xnames = c("x1", "x2"),
    xdati = list(
      icon = c(TRUE, TRUE),
      iord = c(FALSE, FALSE),
      iuno = c(FALSE, FALSE)
    )
  )

  out <- fn(
    xdat = xdat,
    ydat = c(1, 2, 3),
    exdat = xdat,
    gradients = TRUE,
    gradient.order = 1L,
    slice.index = 2L,
    plot.errors.boot.method = "geom",
    plot.errors.boot.blocklen = 2L,
    plot.errors.boot.num = 2L,
    plot.errors.center = "estimate",
    plot.errors.type = "all",
    plot.errors.alpha = 0.05,
    progress.target = "grad x2 2/2",
    bws = bws
  )

  expect_true(is.list(out))
  expect_true(any(grepl("Preparing plot bootstrap geom \\(grad x2 2/2\\)", captured$prep, fixed = FALSE)))
  expect_true(any(grepl("Preparing plot bootstrap geom \\(grad x2 2/2\\)", captured$progress, fixed = FALSE)))
  expect_true(any(grepl("Plot bootstrap \\(grad x2 2/2\\)", captured$progress, fixed = FALSE)))
  expect_true(any(grepl("Constructing bootstrap all bands \\(grad x2 2/2\\)", captured$interval, fixed = FALSE)))
})

test_that("regression helper labels carry target context for wild bootstrap phases", {
  fn <- getFromNamespace("compute.bootstrap.errors.rbandwidth", "np")

  captured <- new.env(parent = emptyenv())
  captured$prep <- character()
  captured$progress <- character()
  captured$interval <- character()

  orig_activity_begin <- getFromNamespace(".np_plot_activity_begin", "np")
  orig_activity_end <- getFromNamespace(".np_plot_activity_end", "np")
  orig_wild <- getFromNamespace(".np_plot_boot_from_hat_wild", "np")
  orig_interval <- getFromNamespace(".np_plot_bootstrap_interval_summary", "np")
  orig_npreghat <- getFromNamespace("npreghat", "np")

  assignInNamespace(".np_plot_activity_begin", function(label) {
    captured$prep <- c(captured$prep, label)
    list(label = label)
  }, ns = "np")
  assignInNamespace(".np_plot_activity_end", function(state) invisible(NULL), ns = "np")
  assignInNamespace(".np_plot_boot_from_hat_wild", function(H, ydat, fit.mean, B, wild, progress.label = NULL) {
    captured$progress <- c(captured$progress, progress.label)
    list(t = matrix(0, nrow = 2L, ncol = 3L), t0 = c(0, 0, 0))
  }, ns = "np")
  assignInNamespace(".np_plot_bootstrap_interval_summary", function(boot.t, t0, alpha, band.type, progress.label = NULL) {
    captured$interval <- c(captured$interval, progress.label)
    list(err = matrix(0, nrow = ncol(boot.t), ncol = 2L), all.err = list())
  }, ns = "np")
  assignInNamespace("npreghat", function(...) {
    args <- list(...)
    if (identical(args$output, "apply")) {
      return(rep(0, nrow(args$txdat)))
    }
    matrix(0, nrow = nrow(args$exdat), ncol = nrow(args$txdat))
  }, ns = "np")
  on.exit({
    assignInNamespace(".np_plot_activity_begin", orig_activity_begin, ns = "np")
    assignInNamespace(".np_plot_activity_end", orig_activity_end, ns = "np")
    assignInNamespace(".np_plot_boot_from_hat_wild", orig_wild, ns = "np")
    assignInNamespace(".np_plot_bootstrap_interval_summary", orig_interval, ns = "np")
    assignInNamespace("npreghat", orig_npreghat, ns = "np")
  }, add = TRUE)

  xdat <- data.frame(x1 = c(0, 1, 2), x2 = c(1, 2, 3))
  bws <- list(
    type = "fixed",
    ndim = 2L,
    xnames = c("x1", "x2"),
    xdati = list(
      icon = c(TRUE, TRUE),
      iord = c(FALSE, FALSE),
      iuno = c(FALSE, FALSE)
    )
  )

  out <- fn(
    xdat = xdat,
    ydat = c(1, 2, 3),
    exdat = xdat,
    gradients = TRUE,
    gradient.order = 1L,
    slice.index = 1L,
    plot.errors.boot.method = "wild",
    plot.errors.boot.wild = "rademacher",
    plot.errors.boot.blocklen = NULL,
    plot.errors.boot.num = 2L,
    plot.errors.center = "estimate",
    plot.errors.type = "all",
    plot.errors.alpha = 0.05,
    progress.target = "grad x1 1/2",
    bws = bws
  )

  expect_true(is.list(out))
  expect_true(any(grepl("Preparing plot bootstrap wild \\(grad x1 1/2\\)", captured$prep, fixed = FALSE)))
  expect_true(any(grepl("Plot bootstrap \\(grad x1 1/2\\)", captured$progress, fixed = FALSE)))
  expect_true(any(grepl("Constructing bootstrap all bands \\(grad x1 1/2\\)", captured$interval, fixed = FALSE)))
})

library(np)

bounded_scalar_kernel <- function(y.eval, y.train, h, lb, ub) {
  denom <- h * (pnorm((ub - y.eval) / h) - pnorm((lb - y.eval) / h))
  dnorm((y.eval - y.train) / h) / denom
}

test_that("bounded conditional cv.ls fixed-point objective matches direct delete-one quadrature", {
  set.seed(20260320)
  n <- 24L
  x <- runif(n)
  y <- rbeta(n, 1, 1)
  h <- c(0.15, 0.12)
  x.lb <- min(x)
  x.ub <- max(x)
  y.lb <- min(y)
  y.ub <- max(y)

  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = h,
    bwmethod = "cv.ls",
    regtype = "lc",
    bandwidth.compute = FALSE,
    cxkerbound = "range",
    cykerbound = "range",
    cvls.quadrature.grid = "uniform"
  )

  Kx <- outer(seq_len(n), seq_len(n), Vectorize(function(i, j) {
    denom <- h[2] * (pnorm((x.ub - x[i]) / h[2]) - pnorm((x.lb - x[i]) / h[2]))
    dnorm((x[i] - x[j]) / h[2]) / denom
  }))
  Ky <- outer(seq_len(n), seq_len(n), Vectorize(function(i, j) {
    denom <- h[1] * (pnorm((y.ub - y[i]) / h[1]) - pnorm((y.lb - y[i]) / h[1]))
    dnorm((y[i] - y[j]) / h[1]) / denom
  }))

  w.loo <- Kx
  diag(w.loo) <- 0
  w.loo <- w.loo / rowSums(w.loo)

  f.loo <- rowSums(Ky * w.loo)
  q <- bw$cvls.quadrature.points[1]
  grid <- seq(y.lb, y.ub, length.out = q)
  trap <- diff(grid)[1] * c(0.5, rep(1, q - 2L), 0.5)
  Ky.grid <- outer(
    grid,
    y,
    Vectorize(function(y.eval, y.train) {
      bounded_scalar_kernel(y.eval, y.train, h[1], y.lb, y.ub)
    })
  )

  intsq <- vapply(
    seq_len(n),
    function(i) {
      fit <- as.numeric(Ky.grid %*% w.loo[i, ])
      sum(trap * fit^2)
    },
    numeric(1)
  )

  exact.obj <- -(mean(intsq) - 2 * mean(f.loo))
  np.obj <- np:::.npcdensbw_eval_only(data.frame(x = x), y, bw)$objective

  expect_equal(np.obj, exact.obj, tolerance = 1e-6)
})

npscoefbw_default <- getFromNamespace("npscoefbw.default", "np")
npscoefbw_scbandwidth <- getFromNamespace("npscoefbw.scbandwidth", "np")
npscoefbw_start_controls <- getFromNamespace(".npscoefbw_start_controls", "np")
npscoef_default_start_bandwidth <- getFromNamespace(".npscoef_default_start_bandwidth", "np")
npscoef_random_start_bandwidth <- getFromNamespace(".npscoef_random_start_bandwidth", "np")

test_that("npscoefbw surfaces fixed-start controls as formal arguments", {
  expect_true(all(c("scale.factor.init.lower", "scale.factor.init.upper", "scale.factor.init", "lbd.init", "hbd.init", "dfac.init") %in%
                    names(formals(npscoefbw_default))))
  expect_true(all(c("scale.factor.init.lower", "scale.factor.init.upper", "scale.factor.init", "lbd.init", "hbd.init", "dfac.init") %in%
                    names(formals(npscoefbw_scbandwidth))))
})

test_that("npscoefbw fixed start helpers replay legacy defaults", {
  param <- c(1.25, 0.6, 0.45)
  icon <- c(TRUE, FALSE, TRUE)
  iord <- c(FALSE, TRUE, FALSE)
  iuno <- c(FALSE, FALSE, FALSE)
  controls <- npscoefbw_start_controls()

  expect_equal(
    npscoef_default_start_bandwidth(
      param = param,
      bwtype = "fixed",
      nobs = 80L,
      start.controls = controls,
      icon = icon,
      iord = iord,
      iuno = iuno
    ),
    param
  )

  set.seed(20260405)
  got <- npscoef_random_start_bandwidth(
    param = param,
    bwtype = "fixed",
    nobs = 80L,
    start.controls = controls,
    icon = icon,
    iord = iord,
    iuno = iuno
  )
  set.seed(20260405)
  expect_equal(got, runif(length(param), min = 0.5, max = 1.5) * param)
})

test_that("npscoefbw fixed start helpers split continuous and categorical controls", {
  param <- c(1.25, 0.6, 0.45)
  icon <- c(TRUE, FALSE, TRUE)
  iord <- c(FALSE, TRUE, FALSE)
  iuno <- c(FALSE, FALSE, FALSE)
  controls <- npscoefbw_start_controls(
    scale.factor.init.lower = 0.7,
    scale.factor.init.upper = 0.9,
    scale.factor.init = 1.2,
    lbd.init = 1.1,
    hbd.init = 1.3,
    dfac.init = 0.8
  )

  expect_equal(
    npscoef_default_start_bandwidth(
      param = param,
      bwtype = "fixed",
      nobs = 80L,
      start.controls = controls,
      icon = icon,
      iord = iord,
      iuno = iuno
    ),
    c(1.2, 0.8, 1.2) * param
  )

  set.seed(11)
  got <- npscoef_random_start_bandwidth(
    param = param,
    bwtype = "fixed",
    nobs = 80L,
    start.controls = controls,
    icon = icon,
    iord = iord,
    iuno = iuno
  )
  set.seed(11)
  expected <- c(
    runif(1, min = 0.7, max = 0.9),
    runif(1, min = 1.1, max = 1.3),
    runif(1, min = 0.7, max = 0.9)
  ) * param
  expect_equal(got, expected)
})

test_that("npscoefbw nearest-neighbor start helpers ignore fixed-start controls", {
  param <- c(8, 8, 8)
  controls <- npscoefbw_start_controls(
    scale.factor.init.lower = 0.7,
    scale.factor.init.upper = 0.9,
    scale.factor.init = 1.2,
    lbd.init = 1.1,
    hbd.init = 1.3,
    dfac.init = 0.8
  )

  expect_equal(
    npscoef_default_start_bandwidth(
      param = param,
      bwtype = "generalized_nn",
      nobs = 80L,
      start.controls = controls
    ),
    npscoef_default_start_bandwidth(param = param, bwtype = "generalized_nn", nobs = 80L)
  )

  set.seed(20260405)
  got <- npscoef_random_start_bandwidth(
    param = param,
    bwtype = "generalized_nn",
    nobs = 80L,
    start.controls = controls
  )
  set.seed(20260405)
  expect_equal(got, npscoef_random_start_bandwidth(param = param, bwtype = "generalized_nn", nobs = 80L))
})

library(np)

quadrature_control_fixture <- function(n = 48L, seed = 20260424L) {
  set.seed(seed)
  x <- data.frame(x = runif(n))
  y <- data.frame(y = 0.25 + x$x + rchisq(n, df = 3))
  list(x = x, y = y)
}

quadrature_control_bw <- function(dat,
                                  cykerlb = 0,
                                  cykerub = Inf,
                                  ...) {
  npcdensbw(
    xdat = dat$x,
    ydat = dat$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    regtype = "lp",
    degree = 0,
    cxkerbound = "fixed",
    cxkerlb = 0,
    cxkerub = 1,
    cykerbound = "fixed",
    cykerlb = cykerlb,
    cykerub = cykerub,
    ...
  )
}

test_that("npcdensbw validates cv.ls quadrature controls", {
  dat <- quadrature_control_fixture(n = 12L)

  expect_no_error(quadrature_control_bw(dat, cvls.quadrature.extend.factor = 0.5))
  expect_no_error(quadrature_control_bw(dat, cvls.quadrature.extend.factor = 1))
  expect_no_error(quadrature_control_bw(dat, cvls.quadrature.extend.factor = 2))
  expect_no_error(quadrature_control_bw(dat, cvls.quadrature.points = c(41L, 17L)))
  expect_no_error(quadrature_control_bw(dat, cvls.quadrature.grid = "hybrid"))
  expect_no_error(quadrature_control_bw(dat, cvls.quadrature.grid = "uniform"))
  expect_no_error(quadrature_control_bw(dat, cvls.quadrature.grid = "sample"))

  removed_args <- list(
    cvls.i1.rescue = FALSE,
    cvls.quadrature.adaptive = TRUE,
    cvls.quadrature.adaptive.tol = 0,
    cvls.quadrature.adaptive.grid.hy.ratio = 0,
    cvls.quadrature.adaptive.floor.tol = 0
  )
  for (nm in names(removed_args)) {
    args <- list(dat)
    args[[nm]] <- removed_args[[nm]]
    expect_error(
      do.call(quadrature_control_bw, args),
      sprintf("%s has been removed", nm),
      fixed = TRUE
    )
  }

  bad_extend <- list(0, -1, NA_real_, NaN, Inf, "2", c(1, 2))
  for (value in bad_extend) {
    expect_error(
      quadrature_control_bw(dat, cvls.quadrature.extend.factor = value),
      "cvls.quadrature.extend.factor"
    )
  }

  bad_points <- list(1, c(81), c(81, 1), c(81, NA), c(81, Inf), c(81.5, 31), "81", c(81, 31, 21))
  for (value in bad_points) {
    expect_error(
      quadrature_control_bw(dat, cvls.quadrature.points = value),
      "cvls.quadrature.points"
    )
  }

  bad_grid <- list(NA, c("hybrid", "uniform"), TRUE, "adaptive")
  for (value in bad_grid) {
    expect_error(
      quadrature_control_bw(dat, cvls.quadrature.grid = value),
      "cvls.quadrature.grid"
    )
  }
})

test_that("npcdensbw stores cv.ls quadrature controls and old objects use defaults", {
  dat <- quadrature_control_fixture(n = 16L)

  bw_default <- quadrature_control_bw(dat)
  bw_explicit <- quadrature_control_bw(
    dat,
    cvls.quadrature.grid = "sample",
    cvls.quadrature.extend.factor = 1.5,
    cvls.quadrature.points = c(43L, 19L)
  )

  expect_identical(bw_default$cvls.quadrature.grid, "hybrid")
  expect_equal(bw_default$cvls.quadrature.extend.factor, 1)
  expect_identical(unname(bw_default$cvls.quadrature.points), c(100L, 50L))
  expect_identical(bw_explicit$cvls.quadrature.grid, "sample")
  expect_equal(bw_explicit$cvls.quadrature.extend.factor, 1.5)
  expect_identical(unname(bw_explicit$cvls.quadrature.points), c(43L, 19L))

  bw_old <- bw_default
  bw_old$cvls.quadrature.grid <- NULL
  bw_old$cvls.quadrature.extend.factor <- NULL
  bw_old$cvls.quadrature.points <- NULL
  expect_true(is.finite(np:::.npcdensbw_eval_only(dat$x, dat$y, bw_old)$objective))
})

test_that("explicit infinite response bounds warn when quadrature points are implicit", {
  dat <- quadrature_control_fixture(n = 12L)

  expect_warning(
    np:::.npcdensbw_warn_infinite_response_quadrature(
      kerlb = 0,
      kerub = Inf,
      kerbound = "fixed",
      points.supplied = FALSE,
      where = "npcdensbw()"
    ),
    "fixed infinite response bounds",
    fixed = TRUE
  )
  expect_silent(
    np:::.npcdensbw_warn_infinite_response_quadrature(
      kerlb = 0,
      kerub = Inf,
      kerbound = "fixed",
      points.supplied = TRUE,
      where = "npcdensbw()"
    )
  )
  expect_silent(
    np:::.npcdensbw_warn_infinite_response_quadrature(
      kerlb = 0,
      kerub = max(dat$y$y),
      kerbound = "fixed",
      points.supplied = FALSE,
      where = "npcdensbw()"
    )
  )
})

test_that("finite response bounds are invariant to cv.ls quadrature extend factor", {
  dat <- quadrature_control_fixture(n = 44L)
  finite_ub <- max(dat$y$y) + 0.5

  bw_factor1 <- quadrature_control_bw(
    dat,
    cykerub = finite_ub,
    cvls.quadrature.extend.factor = 1
  )
  bw_factor2 <- bw_factor1
  bw_factor2$cvls.quadrature.extend.factor <- 2

  obj_factor1 <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_factor1)$objective
  obj_factor2 <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_factor2)$objective

  expect_true(is.finite(obj_factor1))
  expect_equal(obj_factor2, obj_factor1, tolerance = 1e-12)
})

test_that("cv.ls quadrature point vector controls one- and two-dimensional grids", {
  dat <- quadrature_control_fixture(n = 40L)

  bw_1d_default <- quadrature_control_bw(dat, cvls.quadrature.points = c(100L, 50L))
  bw_1d_coarse <- quadrature_control_bw(dat, cvls.quadrature.points = c(41L, 31L))
  obj_1d_default <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_1d_default)$objective
  obj_1d_coarse <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_1d_coarse)$objective

  expect_true(is.finite(obj_1d_default))
  expect_true(is.finite(obj_1d_coarse))
  expect_gt(abs(obj_1d_default - obj_1d_coarse), 1e-10)

  set.seed(20260424)
  n2 <- 16L
  x2 <- data.frame(x = runif(n2))
  y2 <- data.frame(y1 = rbeta(n2, 2, 4), y2 = rbeta(n2, 3, 3))
  bw_2d_default <- npcdensbw(
    xdat = x2,
    ydat = y2,
    bws = c(0.16, 0.18, 0.22),
    bandwidth.compute = FALSE,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    regtype = "lc",
    cxkerbound = "range",
    cykerbound = "range",
    cvls.quadrature.points = c(100L, 50L)
  )
  bw_2d_coarse <- bw_2d_default
  bw_2d_coarse$cvls.quadrature.points <- c(81L, 17L)

  obj_2d_default <- np:::.npcdensbw_eval_only(x2, y2, bw_2d_default)$objective
  obj_2d_coarse <- np:::.npcdensbw_eval_only(x2, y2, bw_2d_coarse)$objective

  expect_identical(bw_2d_default$cvls.quadrature.grid, "uniform")
  expect_true(is.finite(obj_2d_default))
  expect_true(is.finite(obj_2d_coarse))
  expect_gt(abs(obj_2d_default - obj_2d_coarse), 1e-10)
  expect_error(
    npcdensbw(
      xdat = x2,
      ydat = y2,
      bws = c(0.16, 0.18, 0.22),
      bandwidth.compute = FALSE,
      bwmethod = "cv.ls",
      bwtype = "fixed",
      regtype = "lc",
      cxkerbound = "range",
      cykerbound = "range",
      cvls.quadrature.grid = "hybrid"
    ),
    "scalar continuous responses"
  )
})

library(np)

locate_r_source <- function(filename) {
  candidates <- c(
    test_path("..", "..", "R", filename),
    test_path("..", "..", "..", "R", filename),
    file.path(Sys.getenv("R_PACKAGE_DIR", ""), "R", filename),
    file.path(Sys.getenv("R_PACKAGE_SOURCE", ""), "R", filename),
    file.path(getwd(), "R", filename),
    file.path(getwd(), "..", "R", filename)
  )
  candidates <- unique(candidates[nzchar(candidates)])
  hits <- candidates[file.exists(candidates)]
  if (length(hits) == 0L)
    return("")
  hits[[1L]]
}

test_that("timing producers avoid user CPU summaries for touched families", {
  files <- c(
    "np.density.bw.R",
    "np.distribution.bw.R",
    "np.condensity.bw.R",
    "np.condistribution.bw.R",
    "np.plregression.bw.R",
    "np.singleindex.bw.R",
    "np.smoothcoef.bw.R"
  )

  paths <- vapply(files, locate_r_source, character(1), USE.NAMES = TRUE)
  if (any(!nzchar(paths)))
    skip("R source files unavailable in this test context")

  for (nm in names(paths)) {
    lines <- readLines(paths[[nm]], warn = FALSE)
    expect_false(any(grepl("\\}\\)\\[1\\]", lines)), info = nm)
  }
})

test_that("reference regression bandwidth timing remains full-stage elapsed", {
  src_file <- locate_r_source("np.regression.bw.R")
  skip_if(!nzchar(src_file) || !file.exists(src_file),
          "R source file np.regression.bw.R unavailable in this test context")

  lines <- readLines(src_file, warn = FALSE)
  if (!any(grepl("npregbw", lines, fixed = TRUE)))
    skip("R source file np.regression.bw.R unavailable in this test context")
  expect_true(any(grepl("elapsed.start <- proc.time\\(\\)\\[3\\]", lines)), info = "np.regression.bw.R")
  expect_true(any(grepl("tbw\\$total.time <- proc.time\\(\\)\\[3\\] - elapsed.start", lines)),
              info = "np.regression.bw.R")
})

test_that("npcdens basic functionality works", {
  data("faithful")
  # waiting is Y, eruptions is X
  bw <- npcdensbw(xdat=faithful$eruptions, ydat=faithful$waiting, 
                  bws=c(0.5, 5), bandwidth.compute=FALSE)
  
  fit <- npcdens(bws=bw)
  
  expect_s3_class(fit, "condensity")
  expect_type(predict(fit), "double")
  expect_equal(length(predict(fit)), nrow(faithful))
  
  expect_output(summary(fit))
})

test_that("npcdens works with formula", {
  data("Italy")
  Italy_sub <- Italy[1:50, ]
  bw <- npcdensbw(gdp~ordered(year), data=Italy_sub, 
                  bws=c(0.5, 1.0), bandwidth.compute=FALSE)
  
  fit <- npcdens(bws=bw)
  expect_s3_class(fit, "condensity")
})

test_that("npcdensbw exhaustive degree search matches manual profile minimum", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(x = sort(runif(24)))
  dat$y <- dat$x + rnorm(nrow(dat), sd = 0.08)

  bw0 <- np::npcdensbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree = 0L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  bw1 <- np::npcdensbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree = 1L,
    bernstein.basis = TRUE,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  auto <- np::npcdensbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_s3_class(auto, "conbandwidth")
  expect_true(isTRUE(auto$bernstein.basis))
  expect_identical(auto$degree.search$mode, "exhaustive")
  expect_true(isTRUE(auto$degree.search$completed))
  expect_true(isTRUE(auto$degree.search$certified))
  expect_gte(auto$fval, max(bw0$fval, bw1$fval) - 1e-10)
  expect_gte(auto$degree.search$best.fval, auto$degree.search$baseline.fval - 1e-10)
  expect_true(all(c("degree", "fval", "status", "cached") %in% names(auto$degree.search$trace)))
  expect_identical(nrow(auto$degree.search$trace), auto$degree.search$n.unique)
  expect_identical(auto$degree.search$n.cached, auto$degree.search$n.visits - auto$degree.search$n.unique)

  manual <- np::npcdensbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  expect_null(manual$degree.search)
})

test_that("npcdensbw coordinate search can be exhaustively certified on a small grid", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(
    x1 = runif(22),
    x2 = runif(22)
  )
  dat$y <- dat$x1 + dat$x2^2 + rnorm(nrow(dat), sd = 0.08)

  exhaustive <- np::npcdensbw(
    y ~ x1 + x2,
    data = dat,
    regtype = "lp",
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )
  coordinate <- np::npcdensbw(
    y ~ x1 + x2,
    data = dat,
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 1L,
    degree.verify = TRUE,
    degree.restarts = 1L,
    degree.max.cycles = 4L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_identical(coordinate$degree.search$mode, "coordinate")
  expect_true(isTRUE(coordinate$degree.search$completed))
  expect_true(isTRUE(coordinate$degree.search$certified))
  expect_equal(as.integer(coordinate$degree), as.integer(exhaustive$degree))
  expect_equal(coordinate$fval, exhaustive$fval, tolerance = 1e-10)
})

test_that("npcdensbw automatic degree search enforces pilot guardrails", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(y = rnorm(20), x = runif(20))

  expect_error(
    np::npcdensbw(
      y ~ x,
      data = dat,
      regtype = "lc",
      degree.select = "exhaustive",
      degree.min = 0L,
      degree.max = 1L,
      bwtype = "fixed",
      bwmethod = "cv.ls",
      nmulti = 1L
    ),
    "automatic degree search currently requires regtype='lp'"
  )

  bw <- np::npcdensbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    bernstein.basis = FALSE,
    degree.select = "exhaustive",
    search.engine = "cell",
    degree.min = 0L,
    degree.max = 4L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_s3_class(bw, "conbandwidth")
  expect_false(isTRUE(bw$bernstein.basis))
  expect_lte(max(as.integer(bw$degree)), 4L)
})

test_that("npcdens forwards automatic LP degree search through npcdensbw", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(x = runif(20))
  dat$y <- dat$x + rnorm(nrow(dat), sd = 0.08)

  fit <- local({
    suppressPackageStartupMessages(library(np))
    npcdens(
      y ~ x,
      data = dat,
      regtype = "lp",
      degree.select = "exhaustive",
      search.engine = "cell",
      degree.min = 0L,
      degree.max = 1L,
      bwtype = "fixed",
      bwmethod = "cv.ls",
      nmulti = 1L
    )
  })

  expect_false(is.null(fit$bws))
  expect_s3_class(fit$bws, "conbandwidth")
  expect_false(is.null(fit$bws$degree.search))
  expect_identical(fit$bws$degree.search$mode, "exhaustive")
})

test_that("plot helper runtime guard requires non-null bws", {
  guard <- getFromNamespace(".np_plot_require_bws", "np")
  expect_error(
    guard(NULL, "unit-test"),
    "required argument 'bws' is missing or NULL"
  )
  expect_invisible(guard(list(type = "fixed"), "unit-test"))
})

test_that("plot runtime files avoid forbidden *bw( calls", {
  root <- normalizePath(testthat::test_path("..", ".."), mustWork = TRUE)
  files <- c(
    file.path(root, "R", "np.plot.helpers.R"),
    Sys.glob(file.path(root, "R", "np.plot.engine*.R"))
  )
  files <- unique(files[file.exists(files)])
  skip_if(length(files) == 0L, "source R files unavailable in installed test context")

  forbidden <- "\\b[A-Za-z0-9._]+bw\\s*\\("
  allowed <- c(
    "\\.np_indexhat_rbw\\s*\\(",
    "\\.np_indexhat_kbw\\s*\\(",
    "\\.npcdhat_make_xbw\\s*\\(",
    "\\.npcdhat_make_xkbw\\s*\\(",
    "\\.npcdhat_make_ybw\\s*\\("
  )
  offenders <- character()

  for (f in files) {
    raw <- readLines(f, warn = FALSE)
    code <- sub("#.*$", "", raw)
    hit_idx <- which(grepl(forbidden, code, perl = TRUE))
    if (!length(hit_idx)) next

    for (i in hit_idx) {
      ok <- any(vapply(allowed, function(p) grepl(p, code[[i]], perl = TRUE), logical(1)))
      if (!ok) {
        offenders <- c(offenders, sprintf("%s:%d: %s", basename(f), i, trimws(raw[[i]])))
      }
    }
  }

  expect_equal(length(offenders), 0, info = paste(offenders, collapse = "\n"))
})

test_that("plot runtime files avoid silent remap/downgrade patterns", {
  root <- normalizePath(testthat::test_path("..", ".."), mustWork = TRUE)
  files <- c(
    file.path(root, "R", "np.plot.helpers.R"),
    Sys.glob(file.path(root, "R", "np.plot.engine*.R"))
  )
  files <- unique(files[file.exists(files)])
  skip_if(length(files) == 0L, "source R files unavailable in installed test context")

  offenders <- character()
  pat.assign <- "plot\\.errors\\.method\\s*=\\s*\"none\""
  pat.warn <- "Proceeding without"

  for (f in files) {
    raw <- readLines(f, warn = FALSE)
    code <- sub("#.*$", "", raw)
    idx <- which(grepl(pat.assign, code, perl = TRUE) | grepl(pat.warn, raw, fixed = TRUE))
    if (length(idx) && identical(basename(f), "np.plot.helpers.R")) {
      idx <- idx[!vapply(idx, function(i) {
        context <- raw[pmax(1L, i - 10L):i]
        any(grepl("\\.np_plot_validate_renderer_request\\s*<-\\s*function", context, perl = TRUE))
      }, logical(1L))]
    }
    if (length(idx)) {
      offenders <- c(offenders, sprintf("%s:%d: %s", basename(f), idx, trimws(raw[idx])))
    }
  }

  expect_equal(length(offenders), 0, info = paste(offenders, collapse = "\n"))
})

update_bw_name_metadata <- getFromNamespace("updateBwNameMetadata", "np")

test_that("updateBwNameMetadata updates bandwidth metadata deterministically", {
  bws <- list(varnames = list(x = "oldx", y = "oldy"))
  names_in <- list(xnames = "x_new", ynames = "y_new")

  out <- update_bw_name_metadata(nameList = names_in, bws = bws)

  expect_equal(out$xnames, "x_new")
  expect_equal(out$ynames, "y_new")
  expect_equal(out$varnames$x, "x_new")
  expect_equal(out$varnames$y, "y_new")
  expect_equal(bws$varnames$x, "oldx")
  expect_equal(bws$varnames$y, "oldy")
})

test_that("updateBwNameMetadata is a no-op for empty metadata input", {
  bws <- list(varnames = list(x = "x0"))
  out <- update_bw_name_metadata(nameList = list(), bws = bws)
  expect_equal(out, bws)
})

np_seed_enter <- getFromNamespace(".np_seed_enter", "np")
np_seed_exit <- getFromNamespace(".np_seed_exit", "np")

test_that(".np_seed_exit restores prior seed when one exists", {
  had_seed <- exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  if (had_seed) {
    seed_orig <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  }
  on.exit({
    if (had_seed) {
      assign(".Random.seed", seed_orig, envir = .GlobalEnv)
    } else if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      rm(".Random.seed", envir = .GlobalEnv)
    }
  }, add = TRUE)

  set.seed(123)
  seed_before <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  state <- np_seed_enter(42)
  np_seed_exit(state)
  seed_after <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  expect_identical(seed_after, seed_before)
})

test_that(".np_seed_exit remove_if_absent drops synthetic seed", {
  had_seed <- exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  if (had_seed) {
    seed_orig <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
    rm(".Random.seed", envir = .GlobalEnv)
  }
  on.exit({
    if (had_seed) {
      assign(".Random.seed", seed_orig, envir = .GlobalEnv)
    } else if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      rm(".Random.seed", envir = .GlobalEnv)
    }
  }, add = TRUE)

  expect_false(exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
  state <- np_seed_enter(42)
  expect_true(exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
  np_seed_exit(state, remove_if_absent = TRUE)
  expect_false(exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
})

library(np)

compact_window_route_fixture <- function() {
  set.seed(20260423)
  n <- 60L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = 0.4 + 0.8 * x$x + rchisq(n, df = 4))
  list(x = x, y = y)
}

compact_window_span_bounds <- function(vals, lb, ub, factor = 2) {
  rng <- range(vals)
  span <- diff(rng)
  if (!is.finite(span) || span <= 0)
    span <- 1
  c(
    lb = if (is.finite(lb)) lb else rng[1L] - factor * span,
    ub = if (is.finite(ub)) ub else rng[2L] + factor * span
  )
}

compact_window_clone_bw <- function(bw, bound, lb = NULL, ub = NULL) {
  if (bound == "range") {
    yvals <- as.numeric(bw$ydati)
    lb <- min(yvals)
    ub <- max(yvals)
  }
  if (bound == "none") {
    lb <- -Inf
    ub <- Inf
  }

  out <- np:::conbandwidth(
    xbw = bw$xbw,
    ybw = bw$ybw,
    bwmethod = bw$method,
    bwscaling = bw$scaling,
    bwtype = bw$type,
    cxkertype = bw$cxkertype,
    cxkerorder = bw$cxkerorder,
    cxkerbound = bw$cxkerbound,
    cxkerlb = bw$cxkerlb,
    cxkerub = bw$cxkerub,
    uxkertype = bw$uxkertype,
    oxkertype = bw$oxkertype,
    cykertype = bw$cykertype,
    cykerorder = bw$cykerorder,
    cykerbound = bound,
    cykerlb = lb,
    cykerub = ub,
    uykertype = bw$uykertype,
    oykertype = bw$oykertype,
    fval = bw$fval,
    ifval = bw$ifval,
    num.feval = bw$num.feval,
    num.feval.fast = bw$num.feval.fast,
    fval.history = bw$fval.history,
    eval.history = bw$eval.history,
    invalid.history = bw$invalid.history,
    nobs = bw$nobs,
    xdati = bw$xdati,
    ydati = bw$ydati,
    xnames = bw$xnames,
    ynames = bw$ynames,
    sfactor = bw$sfactor,
    bandwidth = bw$bandwidth,
    rows.omit = bw$rows.omit,
    nconfac = bw$nconfac,
    ncatfac = bw$ncatfac,
    sdev = bw$sdev,
    bandwidth.compute = TRUE,
    timing = bw$timing,
    total.time = bw$total.time,
    regtype = bw$regtype,
    pregtype = bw$pregtype,
    basis = bw$basis,
    degree = bw$degree,
    bernstein.basis = bw$bernstein.basis,
    regtype.engine = bw$regtype.engine,
    basis.engine = bw$basis.engine,
    degree.engine = bw$degree.engine,
    bernstein.basis.engine = bw$bernstein.basis.engine
  )
  out$scale.factor.search.lower <- bw$scale.factor.search.lower
  out$cvls.quadrature.grid <- bw$cvls.quadrature.grid
  out$cvls.quadrature.extend.factor <- bw$cvls.quadrature.extend.factor
  out$cvls.quadrature.points <- bw$cvls.quadrature.points
  out
}

test_that("one-sided fixed infinite bounds use the configured span surrogate", {
  dat <- compact_window_route_fixture()

  bw_upper <- npcdensbw(
    ydat = dat$y,
    xdat = dat$x,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    regtype = "lp",
    degree = 0,
    nmulti = 1,
    cxkerbound = "fixed",
    cxkerlb = 0,
    cxkerub = 1,
    cykerbound = "fixed",
    cykerlb = 0,
    cykerub = Inf,
    cvls.quadrature.points = c(81L, 31L)
  )

  bw_lower <- npcdensbw(
    ydat = dat$y,
    xdat = dat$x,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    regtype = "lp",
    degree = 0,
    nmulti = 1,
    cxkerbound = "fixed",
    cxkerlb = 0,
    cxkerub = 1,
    cykerbound = "fixed",
    cykerlb = -Inf,
    cykerub = max(dat$y$y) + 0.25,
    cvls.quadrature.points = c(81L, 31L)
  )
  bw_upper_span2 <- bw_upper
  bw_upper_span2$cvls.quadrature.extend.factor <- 2
  bw_lower_span2 <- bw_lower
  bw_lower_span2$cvls.quadrature.extend.factor <- 2

  upper_span1 <- compact_window_span_bounds(dat$y$y, 0, Inf, factor = 1)
  upper_span2 <- compact_window_span_bounds(dat$y$y, 0, Inf, factor = 2)
  lower_span1 <- compact_window_span_bounds(dat$y$y, -Inf, max(dat$y$y) + 0.25, factor = 1)
  lower_span2 <- compact_window_span_bounds(dat$y$y, -Inf, max(dat$y$y) + 0.25, factor = 2)

  upper_obj <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_upper)$objective
  lower_obj <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_lower)$objective
  upper_obj_factor2 <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_upper_span2)$objective
  lower_obj_factor2 <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_lower_span2)$objective

  upper_obj_span1 <- np:::.npcdensbw_eval_only(
    dat$x, dat$y, compact_window_clone_bw(bw_upper, "fixed", upper_span1[["lb"]], upper_span1[["ub"]])
  )$objective
  upper_obj_span2 <- np:::.npcdensbw_eval_only(
    dat$x, dat$y, compact_window_clone_bw(bw_upper, "fixed", upper_span2[["lb"]], upper_span2[["ub"]])
  )$objective

  lower_obj_span1 <- np:::.npcdensbw_eval_only(
    dat$x, dat$y, compact_window_clone_bw(bw_lower, "fixed", lower_span1[["lb"]], lower_span1[["ub"]])
  )$objective
  lower_obj_span2 <- np:::.npcdensbw_eval_only(
    dat$x, dat$y, compact_window_clone_bw(bw_lower, "fixed", lower_span2[["lb"]], lower_span2[["ub"]])
  )$objective

  expect_equal(upper_obj, upper_obj_span1, tolerance = 1e-12)
  expect_equal(upper_obj_factor2, upper_obj_span2, tolerance = 1e-12)
  expect_gt(abs(upper_obj - upper_obj_factor2), 1e-8)

  expect_equal(lower_obj, lower_obj_span1, tolerance = 1e-12)
  expect_equal(lower_obj_factor2, lower_obj_span2, tolerance = 1e-12)
  expect_gt(abs(lower_obj - lower_obj_factor2), 1e-8)
})

test_that("explicit fixed [-Inf, Inf] survives and uses the configured span surrogate", {
  dat <- compact_window_route_fixture()

  bw_two_inf <- npcdensbw(
    ydat = dat$y,
    xdat = dat$x,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    regtype = "lp",
    degree = 0,
    nmulti = 1,
    cxkerbound = "fixed",
    cxkerlb = 0,
    cxkerub = 1,
    cykerbound = "fixed",
    cykerlb = -Inf,
    cykerub = Inf,
    cvls.quadrature.points = c(81L, 31L)
  )

  bw_two_inf_span2 <- bw_two_inf
  bw_two_inf_span2$cvls.quadrature.extend.factor <- 2

  span1 <- compact_window_span_bounds(dat$y$y, -Inf, Inf, factor = 1)
  span2 <- compact_window_span_bounds(dat$y$y, -Inf, Inf, factor = 2)
  obj_two_inf <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_two_inf)$objective
  obj_two_inf_factor2 <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_two_inf_span2)$objective
  obj_none <- np:::.npcdensbw_eval_only(
    dat$x,
    dat$y,
    compact_window_clone_bw(bw_two_inf, "none")
  )$objective
  obj_span1 <- np:::.npcdensbw_eval_only(
    dat$x,
    dat$y,
    compact_window_clone_bw(bw_two_inf, "fixed", span1[["lb"]], span1[["ub"]])
  )$objective
  obj_span2 <- np:::.npcdensbw_eval_only(
    dat$x,
    dat$y,
    compact_window_clone_bw(bw_two_inf, "fixed", span2[["lb"]], span2[["ub"]])
  )$objective

  expect_identical(as.character(bw_two_inf$cykerbound), "fixed")
  expect_true(is.infinite(bw_two_inf$cykerlb[which(bw_two_inf$iycon)][1L]))
  expect_true(is.infinite(bw_two_inf$cykerub[which(bw_two_inf$iycon)][1L]))
  expect_gt(abs(obj_two_inf - obj_none), 1e-8)
  expect_equal(obj_two_inf, obj_span1, tolerance = 1e-12)
  expect_equal(obj_two_inf_factor2, obj_span2, tolerance = 1e-12)
})

test_that("npcdensbw bounded cv.ls nomad search prefers valid positive-score solutions", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(42)
  dat <- data.frame(x = runif(350L))
  dat$y <- rbeta(nrow(dat), 1, 1)

  bw <- np::npcdensbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "nomad+powell",
    degree.min = 0L,
    degree.max = 3L,
    degree.verify = FALSE,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 3L,
    cxkerbound = "range",
    cykerbound = "range"
  )

  expect_true(is.finite(bw$fval))
  expect_gt(bw$fval, 0)
  expect_equal(
    bw$fval,
    np:::.npcdensbw_eval_only(data.frame(x = dat$x), data.frame(y = dat$y), bw)$objective,
    tolerance = 1e-12
  )
})

test_that("genBwSelStr suppresses unknown fast counts but prints known ones", {
  gen_bw_sel <- getFromNamespace("genBwSelStr", "np")

  s.na <- gen_bw_sel(list(num.feval = 26L, num.feval.fast = NA_integer_))
  s.zero <- gen_bw_sel(list(num.feval = 26L, num.feval.fast = 0L))
  s.pos <- gen_bw_sel(list(num.feval = 26L, num.feval.fast = 3L))

  expect_match(s.na, "Number of Function Evaluations: 26")
  expect_false(grepl("fast =", s.na, fixed = TRUE))
  expect_true(grepl("Number of Function Evaluations: 26 (fast = 0)", s.zero, fixed = TRUE))
  expect_true(grepl("Number of Function Evaluations: 26 (fast = 3)", s.pos, fixed = TRUE))
})

fake_unconditional_density_bws <- function(ndim,
                                           ncon = ndim,
                                           nord = 0L,
                                           nuno = 0L,
                                           iord = rep(FALSE, ndim),
                                           inumord = rep(FALSE, ndim),
                                           xnames = paste0("x", seq_len(ndim))) {
  structure(
    list(
      ndim = as.integer(ndim),
      ncon = as.integer(ncon),
      nord = as.integer(nord),
      nuno = as.integer(nuno),
      xnames = xnames,
      xdati = list(
        iord = as.logical(iord),
        inumord = as.logical(inumord)
      )
    ),
    class = "bandwidth"
  )
}

test_that("unconditional density bootstrap planner emits one slice target per dimension", {
  planner <- getFromNamespace(".np_plot_bootstrap_plan_unconditional_density", "np")

  xdat <- data.frame(x1 = runif(5), x2 = runif(5), x3 = runif(5))
  bws <- fake_unconditional_density_bws(ndim = 3L)

  plan <- planner(
    bws = bws,
    xdat = xdat,
    neval = 25L,
    perspective = FALSE,
    plot.errors.boot.method = "inid",
    plot.errors.type = "simultaneous"
  )

  expect_identical(plan$family, "unconditional_density")
  expect_identical(plan$method_label, "inid")
  expect_identical(plan$band_type, "simultaneous")
  expect_identical(plan$target_total, 3L)
  expect_identical(vapply(plan$targets, `[[`, character(1L), "kind"), rep("slice", 3L))
  expect_identical(vapply(plan$targets, `[[`, character(1L), "label"), c("x1", "x2", "x3"))
  expect_identical(vapply(plan$targets, `[[`, integer(1L), "eval_size"), rep(25L, 3L))
  expect_identical(unname(plan$phase_labels[["constructing"]]), "constructing simultaneous bands")
})

test_that("unconditional density bootstrap planner collapses eligible 2D perspective work to one surface target", {
  planner <- getFromNamespace(".np_plot_bootstrap_plan_unconditional_density", "np")

  xdat <- data.frame(x1 = runif(6), x2 = runif(6))
  bws <- fake_unconditional_density_bws(ndim = 2L, ncon = 2L, nord = 0L, nuno = 0L)

  plan <- planner(
    bws = bws,
    xdat = xdat,
    neval = 40L,
    perspective = TRUE,
    plot.errors.boot.method = "geom",
    plot.errors.type = "all"
  )

  expect_identical(plan$target_total, 1L)
  expect_identical(plan$targets[[1L]]$kind, "surface")
  expect_identical(plan$targets[[1L]]$label, "x1,x2 surface")
  expect_identical(plan$targets[[1L]]$eval_size, 1600L)
  expect_identical(unname(plan$phase_labels[["constructing"]]), "constructing all bands")
})

test_that("unconditional density bootstrap planner uses factor cardinality for slice eval sizes", {
  planner <- getFromNamespace(".np_plot_bootstrap_plan_unconditional_density", "np")

  xdat <- data.frame(
    x1 = factor(c("a", "b", "c", "a")),
    x2 = runif(4)
  )
  bws <- fake_unconditional_density_bws(ndim = 2L)

  plan <- planner(
    bws = bws,
    xdat = xdat,
    neval = 30L,
    perspective = FALSE,
    plot.errors.boot.method = "fixed",
    plot.errors.type = "pointwise"
  )

  expect_identical(vapply(plan$targets, `[[`, integer(1L), "eval_size"), c(3L, 30L))
  expect_identical(vapply(plan$targets, `[[`, character(1L), "label"), c("x1", "x2"))
})

test_that("unconditional density bootstrap planner stays agnostic to execution-path details", {
  planner <- getFromNamespace(".np_plot_bootstrap_plan_unconditional_density", "np")

  planner.formals <- names(formals(planner))

  expect_false("plot.errors.boot.nonfixed" %in% planner.formals)
  expect_false("counts.drawer" %in% planner.formals)
  expect_false("H" %in% planner.formals)
  expect_false("helper" %in% planner.formals)
  expect_false("use.frozen.nonfixed" %in% planner.formals)
})

library(np)

locate_statmods_c <- function() {
  candidates <- c(
    test_path("..", "..", "src", "statmods.c"),
    test_path("..", "..", "..", "src", "statmods.c"),
    file.path(Sys.getenv("R_PACKAGE_DIR", ""), "src", "statmods.c"),
    file.path(Sys.getenv("R_PACKAGE_SOURCE", ""), "src", "statmods.c"),
    file.path(getwd(), "src", "statmods.c"),
    file.path(getwd(), "..", "src", "statmods.c")
  )
  candidates <- unique(candidates[nzchar(candidates)])
  hits <- candidates[file.exists(candidates)]
  if (length(hits) == 0L) {
    return(NULL)
  }
  hits[[1L]]
}

test_that("statmods uses observation-count support for adaptive and generalized radii", {
  src_file <- locate_statmods_c()
  skip_if(is.null(src_file), "source file src/statmods.c unavailable in this test context")

  lines <- readLines(src_file, warn = FALSE)

  expect_true(any(grepl("compute_nn_distance_observation_support_subset", lines, fixed = TRUE)))
  expect_true(any(grepl("compute_nn_distance_train_eval_observation_support_subset", lines, fixed = TRUE)))
  expect_true(any(grepl("kth_observation_radius_from_support", lines, fixed = TRUE)))
  expect_true(any(grepl("nearest_positive_radius_from_support", lines, fixed = TRUE)))
  expect_true(any(grepl("kth_observation_radius_for_eval_from_support", lines, fixed = TRUE)))

  expect_equal(sum(grepl("sort_safe(num_obs_train, vector_dist)", lines, fixed = TRUE)), 0L)
  expect_equal(sum(grepl("vector_unique_dist", lines, fixed = TRUE)), 0L)
  expect_equal(sum(grepl("pointer_nndi", lines, fixed = TRUE)), 0L)

  expect_gte(sum(grepl("compute_nn_distance_observation_support_subset\\(", lines)), 3L)
  expect_gte(sum(grepl("compute_nn_distance_train_eval_observation_support_subset\\(", lines)), 3L)
})

library(np)

test_that("formula npcdens keeps estimator-only proper arguments out of bandwidth selection", {
  set.seed(42)
  x <- rnorm(40)
  y <- x + rnorm(40)

  fit <- npcdens(
    y ~ x,
    regtype = "lp",
    degree = 1L,
    bernstein = TRUE,
    proper = TRUE,
    gradients = TRUE,
    nmulti = 1
  )

  expect_true(inherits(fit, "condensity"))
  expect_true(isTRUE(fit$proper.requested))
  expect_equal(dim(fit$congrad), c(40L, 1L))
})

test_that("formula npcdens keeps native evaluation arguments out of bandwidth selection", {
  set.seed(42)
  x <- rnorm(40)
  y <- x + rnorm(40)
  nd <- data.frame(
    y = seq(min(y), max(y), length.out = 7L),
    x = seq(min(x), max(x), length.out = 7L)
  )

  fit <- npcdens(
    y ~ x,
    exdat = nd["x"],
    eydat = nd["y"],
    nmulti = 1
  )

  expect_true(inherits(fit, "condensity"))
  expect_equal(length(fitted(fit)), nrow(nd))
})

test_that("proper helper weights and projection satisfy core invariants", {
  w.fun <- getFromNamespace(".np_condens_trapezoid_weights", "np")
  p.fun <- getFromNamespace(".np_condens_project_weighted_simplex", "np")

  w <- w.fun(c(0, 1, 2))
  expect_equal(w, c(0.5, 1, 0.5), tolerance = 1e-12)

  f <- c(-0.2, 0.4, 1.8)
  g <- p.fun(f = f, w = w, mass = 1, tol = 1e-12)
  expect_true(all(g >= -1e-10))
  expect_equal(sum(w * g), 1, tolerance = 1e-8)
})

test_that("proper helpers support shadow-validation metadata path", {
  old.opt <- getOption("np.condens.proper.shadow")
  options(np.condens.proper.shadow = TRUE)
  on.exit(options(np.condens.proper.shadow = old.opt), add = TRUE)

  set.seed(20260307)
  x <- runif(30)
  y <- rnorm(30)
  y.grid <- seq(min(y) - 0.2, max(y) + 0.2, length.out = 15)
  x.grid <- c(0.2, 0.8)
  nd <- do.call(rbind, lapply(x.grid, function(xx) data.frame(y = y.grid, x = xx)))

  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.3, 0.3),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )

  fit <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"],
    proper = TRUE
  )

  expect_true(isTRUE(fit$proper.requested))
  expect_false(isTRUE(fit$proper.applied))
  expect_identical(fit$proper.info$reason, "repair_disabled_shadow_mode")
})

test_that("proper=FALSE preserves legacy npcdens output on explicit grids", {
  set.seed(20260307)
  x <- runif(50, -1, 1)
  y <- sin(2 * pi * x) + rnorm(50, sd = 0.2)
  y.grid <- seq(min(y) - 0.3, max(y) + 0.3, length.out = 25)
  x.grid <- c(-0.5, 0.5)
  nd <- do.call(rbind, lapply(x.grid, function(xx) data.frame(y = y.grid, x = xx)))

  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.28, 0.22),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )

  fit.base <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"]
  )
  fit.flag <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"],
    proper = FALSE
  )

  expect_equal(fit.flag$condens, fit.base$condens, tolerance = 1e-12)
  expect_false(isTRUE(fit.flag$proper.requested))
  expect_false(isTRUE(fit.flag$proper.applied))
})

test_that("proper explicit-grid fit repairs slices and preserves raw values", {
  set.seed(2)
  n <- 80L
  x <- runif(n, -1, 1)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.20)
  x.grid <- unname(quantile(x, probs = c(0.2, 0.5, 0.8)))
  y.grid <- seq(min(y) - 0.3, max(y) + 0.3, length.out = 100L)
  nd <- do.call(rbind, lapply(x.grid, function(xx) data.frame(y = y.grid, x = xx)))

  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.28, 0.22),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )

  fit.raw <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"]
  )
  fit.proper <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"],
    proper = TRUE
  )

  expect_true(isTRUE(fit.proper$proper.requested))
  expect_true(isTRUE(fit.proper$proper.applied))
  expect_identical(fit.proper$proper.method, "project")
  expect_equal(fit.proper$condens.raw, fit.raw$condens, tolerance = 1e-12)
  expect_true(all(fit.proper$condens >= -1e-8))

  w.fun <- getFromNamespace(".np_condens_trapezoid_weights", "np")
  w <- w.fun(y.grid)
  split.idx <- split(seq_len(nrow(nd)), nd$x)
  mass <- vapply(split.idx, function(idx) sum(w * fit.proper$condens[idx]), numeric(1))
  expect_equal(unname(mass), rep(1, length(mass)), tolerance = 1e-6)

  expect_error(se(fit.proper), "unavailable for repaired conditional densities")
  expect_true(any(grepl("Proper density repair", capture.output(print(fit.proper)), fixed = TRUE)))
  expect_true(any(grepl("Proper density repair", capture.output(summary(fit.proper)), fixed = TRUE)))
})

test_that("proper request is a no-op for local-constant explicit-grid fits", {
  set.seed(20260322)
  x <- runif(60, -1, 1)
  y <- sin(2 * pi * x) + rnorm(60, sd = 0.15)
  y.grid <- seq(min(y) - 0.2, max(y) + 0.2, length.out = 40L)
  x.grid <- c(-0.4, 0.4)
  nd <- do.call(rbind, lapply(x.grid, function(xx) data.frame(y = y.grid, x = xx)))

  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.3, 0.24),
    bandwidth.compute = FALSE,
    regtype = "lc"
  )

  fit.raw <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"]
  )
  fit.req <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"],
    proper = TRUE
  )

  expect_true(isTRUE(fit.req$proper.requested))
  expect_false(isTRUE(fit.req$proper.applied))
  expect_identical(fit.req$proper.info$reason, "already_proper")
  expect_equal(fit.req$condens, fit.raw$condens, tolerance = 1e-12)
  expect_null(fit.req$condens.raw)
})

test_that("proper request is a no-op for degree-zero lp explicit-grid fits", {
  set.seed(20260322)
  x <- runif(60, -1, 1)
  y <- cos(2 * pi * x) + rnorm(60, sd = 0.15)
  y.grid <- seq(min(y) - 0.2, max(y) + 0.2, length.out = 40L)
  x.grid <- c(-0.25, 0.25)
  nd <- do.call(rbind, lapply(x.grid, function(xx) data.frame(y = y.grid, x = xx)))

  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.27, 0.21),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 0L
  )

  fit.raw <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"]
  )
  fit.req <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"],
    proper = TRUE
  )

  expect_true(isTRUE(fit.req$proper.requested))
  expect_false(isTRUE(fit.req$proper.applied))
  expect_identical(fit.req$proper.info$reason, "already_proper")
  expect_equal(fit.req$condens, fit.raw$condens, tolerance = 1e-12)
  expect_null(fit.req$condens.raw)
})

test_that("proper request on paired evaluation stores metadata without altering fit", {
  set.seed(20260307)
  x <- runif(60)
  y <- rnorm(60)
  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.25, 0.25),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )

  fit.raw <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y)
  )
  fit.req <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE
  )

  expect_true(isTRUE(fit.req$proper.requested))
  expect_false(isTRUE(fit.req$proper.applied))
  expect_identical(fit.req$proper.info$reason, "no_eval_grid")
  expect_equal(fit.req$condens, fit.raw$condens, tolerance = 1e-12)
})

test_that("slice mode can properize fitted npcdens values when apply='fitted'", {
  set.seed(20260322)
  x <- runif(36, -1, 1)
  y <- sin(2 * pi * x) + rnorm(36, sd = 0.18)

  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.29, 0.23),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )

  ctrl <- list(mode = "slice", apply = "fitted", slice.grid.size = 31L, slice.extend.factor = 0)

  fit.raw <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y)
  )
  fit.slice <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE,
    proper.control = ctrl
  )

  build.grid <- getFromNamespace(".np_condens_build_slice_eval_grid", "np")
  grid.eval <- build.grid(
    object = fit.raw,
    slice.context = list(
      txdat = data.frame(x = x),
      tydat = data.frame(y = y),
      exdat = data.frame(x = x),
      eydat = data.frame(y = y)
    ),
    proper.control = ctrl
  )

  grid.fit <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = grid.eval$exdat,
    eydat = grid.eval$eydat
  )
  grid.proper <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = grid.eval$exdat,
    eydat = grid.eval$eydat,
    proper = TRUE
  )

  ypos <- match(y, grid.eval$y.grid)
  oracle <- numeric(length(y))
  for (i in seq_along(grid.eval$groups)) {
    idx.req <- grid.eval$groups[[i]]
    idx.grid <- grid.eval$grid.slices[[i]]
    oracle[idx.req] <- grid.proper$condens[idx.grid[ypos[idx.req]]]
  }

  expect_true(isTRUE(fit.slice$proper.requested))
  expect_true(isTRUE(fit.slice$proper.applied))
  expect_identical(fit.slice$proper.info$route, "slice")
  expect_identical(fit.slice$proper.info$apply.scope, "fitted")
  expect_equal(fit.slice$condens.raw, fit.raw$condens, tolerance = 1e-12)
  expect_equal(fit.slice$condens, oracle, tolerance = 1e-10)
  expect_true(any(abs(fit.slice$condens - fit.raw$condens) > 1e-8))
})

test_that("apply='fitted' leaves explicit-evaluation npcdens objects unchanged", {
  set.seed(20260322)
  x <- runif(70, -1, 1)
  y <- cos(2 * pi * x) + rnorm(70, sd = 0.2)
  nd <- rbind(
    data.frame(y = c(-0.8, -0.1, 0.55), x = rep(-0.4, 3L)),
    data.frame(y = c(-0.35, 0.15, 0.8, 1.1), x = rep(0.45, 4L))
  )

  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.26, 0.2),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )

  fit.raw <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"]
  )
  fit.req <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"],
    proper = TRUE,
    proper.control = list(mode = "slice", apply = "fitted")
  )

  expect_true(isTRUE(fit.req$proper.requested))
  expect_false(isTRUE(fit.req$proper.applied))
  expect_identical(fit.req$proper.info$reason, "scope_not_selected")
  expect_true(isTRUE(fit.req$proper.info$supported))
  expect_equal(fit.req$condens, fit.raw$condens, tolerance = 1e-12)
})

test_that("apply='fitted' also leaves exact-grid npcdens evaluation objects unchanged", {
  set.seed(20260322)
  x <- runif(60, -1, 1)
  y <- cos(2 * pi * x) + rnorm(60, sd = 0.15)
  y.grid <- seq(min(y) - 0.2, max(y) + 0.2, length.out = 40L)
  x.grid <- c(-0.25, 0.25)
  nd <- do.call(rbind, lapply(x.grid, function(xx) data.frame(y = y.grid, x = xx)))

  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.27, 0.21),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )

  fit.raw <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"]
  )
  fit.req <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"],
    proper = TRUE,
    proper.control = list(mode = "slice", apply = "fitted")
  )

  expect_true(isTRUE(fit.req$proper.requested))
  expect_false(isTRUE(fit.req$proper.applied))
  expect_identical(fit.req$proper.info$reason, "scope_not_selected")
  expect_true(isTRUE(fit.req$proper.info$supported))
  expect_equal(fit.req$condens, fit.raw$condens, tolerance = 1e-12)
})

test_that("slice mode defers to exact-grid repair when common grids are supplied", {
  set.seed(20260322)
  x <- runif(70, -1, 1)
  y <- sin(2 * pi * x) + rnorm(70, sd = 0.18)
  y.grid <- seq(min(y) - 0.25, max(y) + 0.25, length.out = 60L)
  x.grid <- c(-0.35, 0.35)
  nd <- do.call(rbind, lapply(x.grid, function(xx) data.frame(y = y.grid, x = xx)))

  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.28, 0.22),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )

  fit.grid <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"],
    proper = TRUE
  )
  fit.slice <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"],
    proper = TRUE,
    proper.control = list(mode = "slice")
  )

  expect_true(isTRUE(fit.slice$proper.applied))
  expect_equal(fit.slice$condens, fit.grid$condens, tolerance = 1e-12)
})

test_that("slice mode on paired rows matches the internal explicit-grid oracle", {
  set.seed(20260322)
  x <- runif(80, -1, 1)
  y <- cos(2 * pi * x) + rnorm(80, sd = 0.2)
  nd <- rbind(
    data.frame(y = c(-0.8, -0.1, 0.55), x = rep(-0.4, 3L)),
    data.frame(y = c(-0.35, 0.15, 0.8, 1.1), x = rep(0.45, 4L))
  )

  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.26, 0.2),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )

  ctrl <- list(mode = "slice", slice.grid.size = 21L, slice.extend.factor = 0)

  fit.slice <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = nd["x"],
    eydat = nd["y"],
    proper = TRUE,
    proper.control = ctrl
  )

  build.grid <- getFromNamespace(".np_condens_build_slice_eval_grid", "np")
  grid.eval <- build.grid(
    object = fit.slice,
    slice.context = list(
      txdat = data.frame(x = x),
      tydat = data.frame(y = y),
      exdat = nd["x"],
      eydat = nd["y"]
    ),
    proper.control = ctrl
  )
  oracle.fit <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    exdat = grid.eval$exdat,
    eydat = grid.eval$eydat,
    proper = TRUE
  )

  ypos <- match(nd$y, grid.eval$y.grid)
  oracle.values <- numeric(nrow(nd))
  for (i in seq_along(grid.eval$groups)) {
    idx.req <- grid.eval$groups[[i]]
    idx.grid <- grid.eval$grid.slices[[i]]
    oracle.values[idx.req] <- oracle.fit$condens[idx.grid[ypos[idx.req]]]
  }

  expect_true(isTRUE(fit.slice$proper.applied))
  expect_identical(fit.slice$proper.info$route, "slice")
  expect_true(all(fit.slice$condens >= -1e-8))
  expect_equal(fit.slice$condens, oracle.values, tolerance = 1e-10)
})

test_that("predict inherits proper request and validates newdata geometry", {
  set.seed(2)
  n <- 80L
  x <- runif(n, -1, 1)
  y <- sin(2 * pi * x) + rnorm(n, sd = 0.20)
  bw <- npcdensbw(
    xdat = data.frame(x = x),
    ydat = data.frame(y = y),
    bws = c(0.28, 0.22),
    bandwidth.compute = FALSE,
    regtype = "lp",
    degree = 3L
  )
  fit.req <- npcdens(
    bws = bw,
    txdat = data.frame(x = x),
    tydat = data.frame(y = y),
    proper = TRUE
  )

  x.grid <- unname(quantile(x, probs = c(0.25, 0.75)))
  y.grid <- seq(min(y) - 0.2, max(y) + 0.2, length.out = 50L)
  nd.good <- do.call(rbind, lapply(x.grid, function(xx) data.frame(y = y.grid, x = xx)))
  pred <- predict(fit.req, newdata = nd.good)
  expect_true(all(pred >= -1e-8))

  nd.bad <- data.frame(y = y.grid[1:4], x = c(0.1, 0.2, 0.3, 0.4))
  expect_error(
    predict(fit.req, newdata = nd.bad),
    "requires repeated fixed-x slices|requires every fixed-x slice"
  )
})

library(np)

chisq_support_fixture <- function(n, seed) {
  set.seed(seed)
  x <- runif(n, 0, 1)
  y <- rchisq(n, df = 2 + 4 * (x - 0.5)^2)
  list(x = data.frame(x = x), y = data.frame(y = y))
}

test_that("npcdensbw stores the cv.ls quadrature grid mode on conditional bandwidth objects", {
  dat <- chisq_support_fixture(n = 40L, seed = 20260423L)

  bw_default <- npcdensbw(
    xdat = dat$x,
    ydat = dat$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    bwmethod = "cv.ls",
    bwtype = "fixed"
  )
  bw_hybrid <- npcdensbw(
    xdat = dat$x,
    ydat = dat$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    cvls.quadrature.grid = "hybrid"
  )
  bw_uniform <- npcdensbw(
    xdat = dat$x,
    ydat = dat$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    cvls.quadrature.grid = "uniform"
  )

  expect_false("cvls.i1.rescue" %in% names(formals(getS3method("npcdensbw", "default"))))
  expect_false("cvls.quadrature.adaptive" %in% names(formals(getS3method("npcdensbw", "default"))))
  expect_false("cvls.i1.rescue" %in% names(bw_default))
  expect_false("cvls.quadrature.adaptive" %in% names(bw_default))
  expect_equal(
    eval(formals(getS3method("npcdensbw", "default"))$cvls.quadrature.ratios),
    c(0.20, 0.55, 0.25)
  )
  expect_error(
    npcdensbw(
      xdat = dat$x,
      ydat = dat$y,
      bws = c(0.35, 0.35),
      bandwidth.compute = FALSE,
      bwmethod = "cv.ls",
      bwtype = "fixed",
      cvls.i1.rescue = FALSE
    ),
    "cvls.i1.rescue has been removed"
  )
  expect_identical(bw_default$cvls.quadrature.grid, "hybrid")
  expect_equal(bw_default$cvls.quadrature.ratios, c(0.20, 0.55, 0.25))
  expect_identical(bw_hybrid$cvls.quadrature.grid, "hybrid")
  expect_identical(bw_uniform$cvls.quadrature.grid, "uniform")
})

test_that("cv.ls hybrid quadrature ratios validate and persist", {
  dat <- chisq_support_fixture(n = 40L, seed = 20260426L)

  bw <- npcdensbw(
    xdat = dat$x,
    ydat = dat$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    cvls.quadrature.ratios = c(0.2, 0.5, 0.3)
  )

  expect_equal(bw$cvls.quadrature.ratios, c(0.2, 0.5, 0.3))
  expect_error(
    npcdensbw(
      xdat = dat$x,
      ydat = dat$y,
      bws = c(0.35, 0.35),
      bandwidth.compute = FALSE,
      bwmethod = "cv.ls",
      bwtype = "fixed",
      cvls.quadrature.ratios = c(0.5, 0.5)
    ),
    "three-element"
  )
  expect_error(
    npcdensbw(
      xdat = dat$x,
      ydat = dat$y,
      bws = c(0.35, 0.35),
      bandwidth.compute = FALSE,
      bwmethod = "cv.ls",
      bwtype = "fixed",
      cvls.quadrature.ratios = c(0.5, -0.1, 0.6)
    ),
    "non-negative"
  )
  expect_error(
    npcdensbw(
      xdat = dat$x,
      ydat = dat$y,
      bws = c(0.35, 0.35),
      bandwidth.compute = FALSE,
      bwmethod = "cv.ls",
      bwtype = "fixed",
      cvls.quadrature.ratios = c(0.5, 0.25, 0.2)
    ),
    "summing to one"
  )
})

test_that("cv.ls quadrature grid modes are stable finite objective controls", {
  dat <- chisq_support_fixture(n = 80L, seed = 20260423L)

  bw_uniform <- npcdensbw(
    xdat = dat$x,
    ydat = dat$y,
    bws = c(0.35, 0.35),
    bandwidth.compute = FALSE,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    regtype = "lp",
    degree = 0,
    cxkerbound = "fixed",
    cxkerlb = 0,
    cxkerub = 1,
    cykerbound = "fixed",
    cykerlb = 0,
    cykerub = Inf,
    cvls.quadrature.grid = "uniform",
    cvls.quadrature.points = c(81L, 31L)
  )
  bw_hybrid <- bw_uniform
  bw_hybrid$cvls.quadrature.grid <- "hybrid"
  bw_sample <- bw_uniform
  bw_sample$cvls.quadrature.grid <- "sample"

  obj_uniform <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_uniform)$objective
  obj_hybrid <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_hybrid)$objective
  obj_sample <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_sample)$objective

  expect_true(is.finite(obj_uniform))
  expect_true(is.finite(obj_hybrid))
  expect_true(is.finite(obj_sample))
  expect_gt(abs(obj_uniform - obj_hybrid), 1e-10)
  expect_gt(abs(obj_uniform - obj_sample), 1e-10)
})

test_that("hybrid cv.ls grid is honored for scalar continuous plus discrete responses", {
  set.seed(20260425)
  n <- 42L
  xdat <- data.frame(x = runif(n))
  ydat <- data.frame(
    y = 0.25 + rchisq(n, df = 3),
    z = factor(sample(letters[1:3], n, replace = TRUE))
  )

  bw_uniform <- npcdensbw(
    xdat = xdat,
    ydat = ydat,
    bws = c(0.36, 0.24, 0.31),
    bandwidth.compute = FALSE,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    regtype = "lp",
    degree = 0,
    cxkerbound = "fixed",
    cxkerlb = 0,
    cxkerub = 1,
    cykerbound = "fixed",
    cykerlb = 0,
    cykerub = Inf,
    cvls.quadrature.grid = "uniform",
    cvls.quadrature.points = c(41L, 17L)
  )
  bw_hybrid <- bw_uniform
  bw_hybrid$cvls.quadrature.grid <- "hybrid"
  bw_sample <- bw_uniform
  bw_sample$cvls.quadrature.grid <- "sample"

  obj_uniform <- np:::.npcdensbw_eval_only(xdat, ydat, bw_uniform)$objective
  obj_hybrid <- np:::.npcdensbw_eval_only(xdat, ydat, bw_hybrid)$objective
  obj_sample <- np:::.npcdensbw_eval_only(xdat, ydat, bw_sample)$objective

  expect_true(is.finite(obj_uniform))
  expect_true(is.finite(obj_hybrid))
  expect_true(is.finite(obj_sample))
  expect_gt(abs(obj_uniform - obj_hybrid), 1e-10)
  expect_gt(abs(obj_uniform - obj_sample), 1e-10)
})

test_that("hybrid cv.ls grid improves the known bad one-sided tiny-hy candidate", {
  dat <- chisq_support_fixture(n = 400L, seed = 600007L)

  bw_uniform <- npcdensbw(
    xdat = dat$x,
    ydat = dat$y,
    bws = c(1.94042638343838e-05, 2455873.66968089),
    bandwidth.compute = FALSE,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    regtype = "lp",
    degree = 0,
    cxkerbound = "fixed",
    cxkerlb = 0,
    cxkerub = 1,
    cykerbound = "fixed",
    cykerlb = 0,
    cykerub = Inf,
    cvls.quadrature.grid = "uniform",
    cvls.quadrature.extend.factor = 2,
    cvls.quadrature.points = c(81L, 31L)
  )
  bw_hybrid <- bw_uniform
  bw_hybrid$cvls.quadrature.grid <- "hybrid"

  obj_uniform <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_uniform)$objective
  obj_hybrid <- np:::.npcdensbw_eval_only(dat$x, dat$y, bw_hybrid)$objective

  expect_true(is.finite(obj_uniform))
  expect_true(is.finite(obj_hybrid))
  expect_gt(obj_uniform, 1)
  expect_lt(obj_hybrid, obj_uniform)
})

test_that("npplreg generalized-nn cv.aic bootstrap plot contract holds across regtypes and methods", {
  set.seed(20260311)

  n <- 60L
  tx <- data.frame(x1 = runif(n))
  tz <- data.frame(z1 = runif(n))
  y <- 0.7 * tx$x1 + sin(2 * pi * tz$z1) + rnorm(n, sd = 0.06)

  boot_methods <- c("inid", "fixed", "geom", "wild")
  regtypes <- c("lc", "ll", "lp")
  bws.nn <- matrix(c(2, 9), nrow = 2L)

  for (regtype in regtypes) {
    bw.args <- list(
      xdat = tx,
      ydat = y,
      zdat = tz,
      regtype = regtype,
      bwmethod = "cv.aic",
      bwtype = "generalized_nn",
      bws = bws.nn,
      bandwidth.compute = FALSE
    )

    if (identical(regtype, "lp")) {
      bw.args$basis <- "glp"
      bw.args$degree <- 1L
      bw.args$bernstein.basis <- FALSE
    }

    bw <- do.call(npplregbw, bw.args)

    for (boot_method in boot_methods) {
      out <- plot(
        bw,
        xdat = tx,
        ydat = y,
        zdat = tz,
        output = "data",
        errors = "bootstrap",
        bootstrap = boot_method,
        B = 7L,
        alpha = 0.1,
        perspective = FALSE
      )

      expect_true(is.list(out), info = sprintf("regtype=%s boot=%s", regtype, boot_method))
      expect_true(length(out) > 0L, info = sprintf("regtype=%s boot=%s", regtype, boot_method))
      expect_true(all(c("plr1", "plr2") %in% names(out)),
                  info = sprintf("regtype=%s boot=%s", regtype, boot_method))
    }
  }
})

np_internal <- function(name) {
  get(name, envir = asNamespace("np"), inherits = FALSE)
}

test_that(".np_nomad_require_crs fails fast when crs is missing", {
  require_crs <- np_internal(".np_nomad_require_crs")
  loaded <- FALSE

  expect_error(
    require_crs(
      version_fn = function(pkg) stop(sprintf("missing %s", pkg), call. = FALSE),
      load_namespace = function(pkg) {
        loaded <<- TRUE
        NULL
      }
    ),
    paste(
      "requires the suggested package 'crs'",
      "\\(>= 0\\.15-41\\) to provide the NOMAD backend;",
      "install\\.packages\\('crs'\\)"
    )
  )

  expect_false(loaded)
})

test_that(".np_nomad_require_crs fails fast when crs is too old", {
  require_crs <- np_internal(".np_nomad_require_crs")
  loaded <- FALSE

  expect_error(
    require_crs(
      version_fn = function(pkg) package_version("0.15-40"),
      load_namespace = function(pkg) {
        loaded <<- TRUE
        NULL
      }
    ),
    "requires 'crs' \\(>= 0\\.15-41\\); installed version is 0\\.15\\.40"
  )

  expect_false(loaded)
})

test_that(".np_nomad_require_crs succeeds when crs version is sufficient", {
  require_crs <- np_internal(".np_nomad_require_crs")
  loaded <- FALSE

  observed <- withVisible(
    require_crs(
      version_fn = function(pkg) package_version("0.15-41"),
      load_namespace = function(pkg) {
        loaded <<- identical(pkg, "crs")
        baseenv()
      }
    )
  )

  expect_true(isTRUE(observed$value))
  expect_false(observed$visible)
  expect_true(loaded)
})

test_that(".np_nomad_require_crs reports namespace load failures after version check", {
  require_crs <- np_internal(".np_nomad_require_crs")

  expect_error(
    require_crs(
      version_fn = function(pkg) package_version("0.15-41"),
      load_namespace = function(pkg) stop("broken namespace", call. = FALSE)
    ),
    "requires 'crs' \\(>= 0\\.15-41\\); failed to load installed version 0\\.15\\.41: broken namespace"
  )
})

test_that("progress core avoids legacy emitters", {
  src_path <- testthat::test_path("..", "..", "R", "progress.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_false(grepl("\\bprint\\(", src))
  expect_true(grepl("\\.np_progress_select_bandwidth <- function\\(", src))
  expect_true(grepl("\\.np_warning <- function\\(", src))
  expect_true(grepl("\\.np_progress_render_single_line <- function\\(", src))
  expect_true(grepl("base::cat\\(", src))
})

test_that("R layer routes warnings through unified helper", {
  r_dir <- testthat::test_path("..", "..", "R")
  skip_if_not(dir.exists(r_dir), "source R files unavailable in installed test context")
  files <- setdiff(list.files(r_dir, pattern = "\\.[Rr]$", full.names = TRUE), file.path(r_dir, "progress.R"))

  for (src_path in files) {
    src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")
    expect_false(grepl("\\bwarning\\(", src), info = basename(src_path))
    expect_false(grepl("\\bmessage\\(", src), info = basename(src_path))
  }
})

test_that("core estimator wrappers emit top-level bandwidth-selection notes", {
  cases <- list(
    list(file = "np.regression.R", label = "Selecting regression bandwidth", helper = ".np_progress_select_bandwidth_enhanced"),
    list(file = "np.density.R", label = "Selecting density bandwidth", helper = ".np_progress_select_bandwidth_enhanced"),
    list(file = "np.distribution.R", label = "Selecting distribution bandwidth", helper = ".np_progress_select_bandwidth_enhanced"),
    list(file = "np.condensity.R", label = "Selecting conditional density bandwidth", helper = ".np_progress_select_bandwidth_enhanced"),
    list(file = "np.condistribution.R", label = "Selecting conditional distribution bandwidth", helper = ".np_progress_select_bandwidth_enhanced"),
    list(file = "np.plregression.R", label = "Selecting partially linear regression bandwidth", helper = ".np_progress_select_bandwidth_enhanced"),
    list(file = "np.singleindex.R", label = "Selecting single-index bandwidth", helper = ".np_progress_select_bandwidth_enhanced"),
    list(file = "np.smoothcoef.R", label = "Selecting smooth coefficient bandwidth", helper = ".np_progress_select_bandwidth_enhanced"),
    list(file = "np.conmode.R", label = "Selecting conditional density bandwidth", helper = ".np_progress_select_bandwidth"),
    list(file = "np.qregression.R", label = "Selecting conditional distribution bandwidth", helper = ".np_progress_select_bandwidth")
  )

  for (case in cases) {
    src_path <- testthat::test_path("..", "..", "R", case$file)
    skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
    src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

    expect_true(
      grepl(sprintf("%s\\(\\s*\"%s\"", case$helper, case$label), src),
      info = case$file
    )
  }
})

test_that("npregivderiv no longer uses legacy console helpers", {
  src_path <- testthat::test_path("..", "..", "R", "npregivderiv.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_true(grepl("\\.np_progress_begin\\(\"Iterating Landweber-Fridman derivative solve\"", src))
})

test_that("npregiv no longer uses legacy console helpers", {
  src_path <- testthat::test_path("..", "..", "R", "npregiv.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_true(grepl("\\.np_progress_select_iv\\(", src))
  expect_true(grepl("\\.np_progress_iv_set_object\\(", src))
  expect_true(grepl("\\.np_progress_with_legacy_suppressed\\(", src))
})

test_that("npunitest no longer uses legacy console helpers", {
  src_path <- testthat::test_path("..", "..", "R", "np.unitest.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_true(grepl("\\.np_progress_note\\(\"Computing bandwidths\"\\)", src))
  expect_true(grepl("\\.np_progress_begin\\(\"Bootstrap replications\"", src))
})

test_that("npdeneqtest no longer uses legacy console helpers", {
  src_path <- testthat::test_path("..", "..", "R", "np.deneqtest.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_true(grepl("\\.np_progress_note\\(\"Computing bandwidths\"\\)", src))
  expect_true(grepl("\\.np_progress_begin\\(\"Bootstrap replications\"", src))
  expect_true(grepl("\\.np_progress_with_legacy_suppressed\\(", src))
})

test_that("npdeptest no longer uses legacy console helpers", {
  src_path <- testthat::test_path("..", "..", "R", "np.deptest.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_true(grepl("\\.np_progress_note\\(\"Computing bandwidths\"\\)", src))
  expect_true(grepl("\\.np_progress_note\\(\"Constructing metric entropy\"\\)", src))
  expect_true(grepl("\\.np_progress_begin\\(\"Bootstrap replications\"", src))
  expect_true(grepl("\\.np_progress_with_legacy_suppressed\\(", src))
})

test_that("npcmstest no longer uses legacy console helpers", {
  src_path <- testthat::test_path("..", "..", "R", "np.cmstest.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_true(grepl("\\.np_progress_note\\(\"Computing bandwidths\"\\)", src))
  expect_true(grepl("\\.np_progress_begin\\(\"Bootstrap replications\"", src))
  expect_true(grepl("\\.np_progress_with_legacy_suppressed\\(", src))
})

test_that("npqcmstest no longer uses legacy console helpers", {
  src_path <- testthat::test_path("..", "..", "R", "np.qcmstest.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_false(grepl("\\bcat\\(", src))
  expect_true(grepl("\\.np_progress_note\\(\"Computing bandwidths\"\\)", src))
  expect_true(grepl("\\.np_progress_begin\\(\"Bootstrap replications\"", src))
  expect_true(grepl("\\.np_progress_with_legacy_suppressed\\(", src))
})

test_that("npsymtest no longer uses legacy console helpers", {
  src_path <- testthat::test_path("..", "..", "R", "np.symtest.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_true(grepl("\\.np_progress_note\\(\"Computing bandwidths\"\\)", src))
  expect_true(grepl("\\.np_progress_begin\\(\"Bootstrap replications\"", src))
})

test_that("npsdeptest no longer uses legacy console helpers", {
  src_path <- testthat::test_path("..", "..", "R", "np.sdeptest.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_true(grepl("\\.np_progress_note\\(\"Computing bandwidths\"\\)", src))
  expect_true(grepl("\\.np_progress_begin\\(\"Constructing metric entropy by lag\"", src))
  expect_true(grepl("\\.np_progress_begin\\(\"Bootstrap replications\"", src))
  expect_true(grepl("\\.np_progress_with_legacy_suppressed\\(", src))
})

test_that("npsigtest no longer uses legacy console helpers", {
  src_path <- testthat::test_path("..", "..", "R", "np.sigtest.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_true(grepl("\\.np_progress_note\\(\"Testing joint significance\"\\)", src))
  expect_true(grepl("\\.np_progress_note\\(sprintf\\(\"Testing variable %s of \\(%s\\)\",", src))
  expect_true(grepl("\\.np_progress_begin\\(\"Bootstrap replications\"", src))
  expect_true(grepl("\\.np_progress_with_legacy_suppressed\\(", src))
})

test_that("npcopula no longer uses legacy console helpers", {
  src_path <- testthat::test_path("..", "..", "R", "np.copula.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_true(grepl("\\.np_progress_begin\\(", src))
  expect_true(grepl("\\.np_progress_step_at\\(", src))
  expect_true(grepl("Copula %s %s", src, fixed = TRUE))
  expect_true(grepl("quasi-inverse marginal %s", src, fixed = TRUE))
  expect_true(grepl("density marginal %s on expanded grid", src, fixed = TRUE))
})

test_that("npindexbw no longer uses legacy console helpers", {
  src_path <- testthat::test_path("..", "..", "R", "np.singleindex.bw.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_true(grepl("\\.np_progress_select_bandwidth_enhanced\\(\\s*\"Selecting single-index bandwidth\"", src))
  expect_true(grepl("\\.np_progress_bandwidth_multistart_step\\(done = i, total = nmulti\\)", src))
})

test_that("npscoefbw multistart path uses central progress core", {
  src_path <- testthat::test_path("..", "..", "R", "np.smoothcoef.bw.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_true(grepl("Optimizing smooth coefficient bandwidth", src, fixed = TRUE))
  expect_true(grepl("\\.np_progress_select_bandwidth_enhanced\\(\\s*\"Selecting smooth coefficient bandwidth\"", src))
  expect_true(grepl("\\.np_progress_bandwidth_multistart_step\\(done = i, total = nmulti\\)", src))
  expect_true(grepl("Backfitting smooth coefficient bandwidth", src, fixed = TRUE))
  expect_true(grepl("Optimizing partial residual bandwidth", src, fixed = TRUE))
  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
})

test_that("npplregbw direct path uses coordinated bandwidth selector", {
  src_path <- testthat::test_path("..", "..", "R", "np.plregression.bw.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_true(grepl("\\.np_progress_select_bandwidth_enhanced\\(\\s*\"Selecting partially linear regression bandwidth\"", src))
  expect_true(grepl("\\.np_progress_bandwidth_set_coordinator\\(", src))
  expect_true(grepl("\\.np_progress_bandwidth_set_coordinator_group\\(", src))
})

test_that("plot helpers use central progress core", {
  src_path <- testthat::test_path("..", "..", "R", "np.plot.helpers.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("printPush\\(|printPop\\(|printClear\\(|newLineConsole\\(", src))
  expect_true(grepl("\\.np_progress_begin\\(", src))
  expect_true(grepl("\\.np_progress_step\\(", src))
  expect_true(grepl("\\.np_progress_end\\(", src))
  expect_true(grepl("\\.np_plot_bootstrap_progress_begin\\(", src))
  expect_true(grepl("\\.np_plot_activity_begin\\(", src))
})

test_that("compiled multistart spinner no longer emits legacy redraw output", {
  src_path <- testthat::test_path("..", "..", "src", "np.c")
  skip_if_not(file.exists(src_path), "source C files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("Rprintf\\(\"\\\\rMultistart", src))
  expect_false(grepl("Rprintf\\(\"\\\\r +\\\\r\"", src))
  expect_true(grepl("SEXP C_np_progress_signal\\(", src))
  expect_true(grepl("\\.np_progress_signal_from_c", src))
  expect_true(grepl("np_progress_bandwidth_multistart_step\\(", src))
  expect_true(grepl("np_progress_bandwidth_activity_step\\(", src))
  expect_true(grepl("bwm_maybe_signal_activity\\(", src))
  expect_true(grepl("np_progress_signal\\(\"bandwidth_activity_step\",\\s*\"bandwidth\"", src))
})

test_that("compiled progress bridge is registered narrowly", {
  src_path <- testthat::test_path("..", "..", "src", "np_init.c")
  skip_if_not(file.exists(src_path), "source C files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_true(grepl("\"C_np_progress_signal\"", src, fixed = TRUE))
})

test_that("compiled working-status redraw is removed", {
  src_path <- testthat::test_path("..", "..", "src", "kernele.c")
  skip_if_not(file.exists(src_path), "source C files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")

  expect_false(grepl("REprintf\\(\"\\\\rWorking\\.\\.\\.", src))
})

test_that("npscoef plot bootstrap inid supports ll/lp basis variants", {
  skip_if_not_installed("np")

  set.seed(3233)
  n <- 70
  x <- runif(n)
  z <- runif(n)
  y <- (0.6 + x) * sin(2 * pi * z) + rnorm(n, sd = 0.08)
  tx <- data.frame(x = x)
  tz <- data.frame(z = z)

  cfgs <- list(
    list(regtype = "ll", basis = NULL, label = "ll"),
    list(regtype = "lp", basis = "glp", label = "lp-glp"),
    list(regtype = "lp", basis = "additive", label = "lp-additive"),
    list(regtype = "lp", basis = "tensor", label = "lp-tensor")
  )

  for (cfg in cfgs) {
    bw.args <- list(
      xdat = tx,
      zdat = tz,
      ydat = y,
      bws = 0.22,
      bandwidth.compute = FALSE,
      regtype = cfg$regtype
    )
    if (!is.null(cfg$basis)) {
      bw.args$basis <- cfg$basis
      bw.args$degree <- 2L
    }
    bw <- do.call(npscoefbw, bw.args)
    out <- suppressWarnings(
      plot(
        bw,
        xdat = tx,
        ydat = y,
        zdat = tz,
        perspective = FALSE,
        output = "data",
        errors = "bootstrap",
        bootstrap = "inid",
        B = 7
      )
    )
    expect_type(out, "list")
    expect_true(length(out) > 0, info = cfg$label)
  }
})

test_that("npcdensbw direct nomad payload preserves CV metadata", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260320)
  dat <- data.frame(x = runif(24))
  dat$y <- rbeta(nrow(dat), 1, 1)

  bw <- np::npcdensbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree.select = "coordinate",
    search.engine = "nomad",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L,
    cxkerbound = "range",
    cykerbound = "range"
  )

  expect_identical(bw$method, "cv.ls")
  expect_identical(bw$pmethod, "Least Squares Cross-Validation")
  expect_true(length(bw$ifval) == 1L && is.na(bw$ifval))
  expect_true(length(bw$fval.history) == 1L && is.na(bw$fval.history))
  expect_true(length(bw$eval.history) == 1L && is.na(bw$eval.history))
  expect_true(length(bw$invalid.history) == 1L && is.na(bw$invalid.history))
  expect_equal(
    bw$fval,
    np:::.npcdensbw_eval_only(data.frame(x = dat$x), data.frame(y = dat$y), bw)$objective,
    tolerance = 1e-12
  )

  printed <- paste(capture.output(print(bw)), collapse = "\n")
  expect_false(grepl("Manual", printed, fixed = TRUE))
  expect_match(printed, "achieved on multistart 1", fixed = TRUE)
})

test_that("npcdens direct nomad payload retains summary bandwidth metadata", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)
  suppressPackageStartupMessages(library(np))

  set.seed(42)
  n <- 500L
  x <- runif(n)
  y <- rbeta(n, 1 + x, 2 - x)

  fit <- np::npcdens(
    y ~ x,
    nomad = TRUE,
    search.engine = "nomad",
    cykerbound = "range"
  )

  expect_false(identical(fit$bws$sumNum, NA))
  expect_false(identical(fit$bws$bandwidth, NA))
  expect_false(identical(fit$bws$sfactor, NA))
  expect_gt(as.numeric(fit$bws$num.feval.fast[1L]), 0)

  summary_text <- expect_warning(capture.output(summary(fit$bws)), NA)
  summary_text <- paste(summary_text, collapse = "\n")
  expect_match(summary_text, "Exp\\. Var\\. Name:", perl = TRUE)
  expect_match(summary_text, "Dep\\. Var\\. Name:", perl = TRUE)
})

test_that("npcdens NOMAD accounting is owner-level on fast-path fits", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)
  suppressPackageStartupMessages(library(np))

  set.seed(42)
  n <- 250L
  x <- runif(n)
  y <- runif(n)

  fit <- np::npcdens(
    y ~ x,
    nomad = TRUE,
    search.engine = "nomad",
    cykerbound = "range"
  )

  expect_gt(as.numeric(fit$bws$num.feval.fast[1L]), 0)
  expect_lte(
    as.numeric(fit$bws$num.feval.fast[1L]),
    as.numeric(fit$bws$num.feval[1L])
  )

  ev <- np:::.npcdensbw_eval_only(
    data.frame(x = x),
    data.frame(y = y),
    fit$bws,
    invalid.penalty = "baseline",
    penalty.multiplier = 10
  )

  expect_equal(as.numeric(ev$num.feval), 1)
  expect_equal(as.numeric(ev$num.feval.fast), 1)
})

library(np)

chisq_support_fixture <- function(n, seed) {
  set.seed(seed)
  x <- runif(n, 0, 1)
  y <- rchisq(n, df = 2 + 4 * (x - 0.5)^2)
  list(x = data.frame(x = x), y = data.frame(y = y))
}

make_bad_seed_bandwidth <- function(scale.factor.search.lower = NULL) {
  dat <- chisq_support_fixture(n = 400L, seed = 600007L)
  args <- list(
    xdat = dat$x,
    ydat = dat$y,
    bws = c(1.94042638343838e-05, 2455873.66968089),
    bandwidth.compute = FALSE,
    bwmethod = "cv.ls",
    bwtype = "fixed",
    regtype = "lp",
    degree = 0,
    cxkerbound = "fixed",
    cxkerlb = 0,
    cxkerub = 1,
    cykerbound = "fixed",
    cykerlb = 0,
    cykerub = Inf
  )
  if (!is.null(scale.factor.search.lower)) {
    args$scale.factor.search.lower <- scale.factor.search.lower
  }
  list(
    data = dat,
    bw = do.call(npcdensbw, args)
  )
}

test_that("omitted scale-factor floor matches explicit default 0.1", {
  default_case <- make_bad_seed_bandwidth()
  strict_case <- make_bad_seed_bandwidth(scale.factor.search.lower = 0.1)
  legacy_case <- make_bad_seed_bandwidth(scale.factor.search.lower = 0.01)

  obj_default <- np:::.npcdensbw_eval_only(default_case$data$x, default_case$data$y, default_case$bw)$objective
  obj_strict <- np:::.npcdensbw_eval_only(strict_case$data$x, strict_case$data$y, strict_case$bw)$objective
  obj_legacy <- np:::.npcdensbw_eval_only(legacy_case$data$x, legacy_case$data$y, legacy_case$bw)$objective

  expect_equal(default_case$bw$scale.factor.search.lower, 0.1, tolerance = 0)
  expect_equal(strict_case$bw$scale.factor.search.lower, 0.1, tolerance = 0)
  expect_equal(legacy_case$bw$scale.factor.search.lower, 0.01, tolerance = 0)
  expect_equal(obj_default, obj_strict, tolerance = 1e-12)
  expect_equal(obj_default, obj_legacy, tolerance = 1e-12)
})

test_that("explicit bandwidth storage is unchanged when bandwidth.compute is FALSE", {
  strict_case <- make_bad_seed_bandwidth(scale.factor.search.lower = 0.1)

  expect_equal(strict_case$bw$scale.factor.search.lower, 0.1, tolerance = 0)
  expect_equal(strict_case$bw$ybw[1L], 1.94042638343838e-05, tolerance = 0)
  expect_equal(strict_case$bw$xbw[1L], 2455873.66968089, tolerance = 0)
})

test_that("explicit 0.1 floor does not reinterpret eval-only objective values", {
  default_case <- make_bad_seed_bandwidth()
  strict_case <- make_bad_seed_bandwidth(scale.factor.search.lower = 0.1)

  obj_default <- np:::.npcdensbw_eval_only(default_case$data$x, default_case$data$y, default_case$bw)$objective
  obj_strict <- np:::.npcdensbw_eval_only(strict_case$data$x, strict_case$data$y, strict_case$bw)$objective

  expect_equal(obj_strict, obj_default, tolerance = 1e-12)
})

test_that("explicit 0.1 floor keeps nomad and nomad+powell aligned on bad seed", {
  dat <- chisq_support_fixture(n = 400L, seed = 600007L)

  common_args <- list(
    xdat = dat$x,
    ydat = dat$y,
    bwmethod = "cv.ls",
    regtype = "lp",
    bwtype = "fixed",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 10L,
    bernstein.basis = TRUE,
    nmulti = 2,
    scale.factor.search.lower = 0.1,
    cvls.quadrature.points = c(81L, 31L),
    cxkerbound = "fixed", cxkerlb = 0, cxkerub = 1,
    cykerbound = "fixed", cykerlb = 0, cykerub = Inf
  )

  nomad <- do.call(npcdensbw, c(common_args, list(search.engine = "nomad")))
  hot <- do.call(npcdensbw, c(common_args, list(search.engine = "nomad+powell")))

  expect_equal(hot$degree, nomad$degree, tolerance = 0)
  expect_gte(hot$sfactor$y[1L], hot$scale.factor.search.lower)
  expect_true(is.finite(hot$fval[1L]))
})

test_that("explicit high floor is enforced during conditional-density Powell search", {
  set.seed(42)
  n <- 100L
  dat <- data.frame(x = rnorm(n), y = rnorm(n))

  out <- npcdensbw(
    y ~ x,
    data = dat,
    bwmethod = "cv.ls",
    scale.factor.search.lower = 1
  )

  expect_true(is.finite(out$fval[1L]))
  expect_gte(out$sfactor$x[1L], out$scale.factor.search.lower)
  expect_gte(out$sfactor$y[1L], out$scale.factor.search.lower)
})

library(np)

gen_bw_sel_str <- getFromNamespace("genBwSelStr", "np")

test_that("bandwidth summaries map base-run multistart labels to one", {
  expect_match(
    gen_bw_sel_str(list(fval = 1.23, ifval = 0, num.feval = NA, num.feval.fast = NA)),
    "achieved on multistart 1",
    fixed = TRUE
  )

  expect_match(
    gen_bw_sel_str(list(fval = 1.23, ifval = 2, num.feval = NA, num.feval.fast = NA)),
    "achieved on multistart 2",
    fixed = TRUE
  )
})

test_that("bandwidth summaries ignore malformed ifval and prefer NOMAD best restart", {
  expect_false(grepl(
    "achieved on multistart",
    gen_bw_sel_str(list(fval = 1.23, ifval = 4.81432545643515e-06, num.feval = NA, num.feval.fast = NA)),
    fixed = TRUE
  ))

  expect_match(
    gen_bw_sel_str(list(
      fval = 1.23,
      ifval = 4.81432545643515e-06,
      nomad.best.restart = 3L,
      num.feval = NA,
      num.feval.fast = NA
    )),
    "achieved on multistart 3",
    fixed = TRUE
  )
})

test_that("single-start progress avoids legacy redraw output", {
  set.seed(1)
  x <- runif(20)
  y <- x + rnorm(20, sd = 0.1)

  out <- capture.output(
    npcdensbw(y ~ x, regtype = "ll", bwtype = "adaptive_nn", nmulti = 1, itmax = 1)
  )

  expect_false(any(grepl("Multistart 1 of 1", out, fixed = TRUE)))
  expect_false(any(grepl("Multistart 1 of 0", out, fixed = TRUE)))
})

test_that("single-index bootstrap target labels format compactly", {
  fmt <- getFromNamespace(".np_plot_singleindex_bootstrap_target_label", "np")

  expect_null(fmt(gradients = FALSE))
  expect_null(fmt(gradients = TRUE))
})

test_that("single-index helper labels carry target context for bootstrap phases", {
  fn <- getFromNamespace("compute.bootstrap.errors.sibandwidth", "np")

  captured <- new.env(parent = emptyenv())
  captured$prep <- character()
  captured$progress <- character()
  captured$interval <- character()

  orig_activity_begin <- getFromNamespace(".np_plot_activity_begin", "np")
  orig_activity_end <- getFromNamespace(".np_plot_activity_end", "np")
  orig_index <- getFromNamespace(".np_inid_boot_from_index", "np")
  orig_interval <- getFromNamespace(".np_plot_bootstrap_interval_summary", "np")

  assignInNamespace(".np_plot_activity_begin", function(label) {
    captured$prep <- c(captured$prep, label)
    list(label = label)
  }, ns = "np")
  assignInNamespace(".np_plot_activity_end", function(state) invisible(NULL), ns = "np")
  assignInNamespace(".np_inid_boot_from_index", function(..., progress.label = NULL) {
    captured$progress <- c(captured$progress, progress.label)
    list(t = matrix(0, nrow = 2L, ncol = 3L), t0 = c(0, 0, 0))
  }, ns = "np")
  assignInNamespace(".np_plot_bootstrap_interval_summary", function(boot.t, t0, alpha, band.type, progress.label = NULL) {
    captured$interval <- c(captured$interval, progress.label)
    list(err = matrix(0, nrow = ncol(boot.t), ncol = 2L), all.err = list())
  }, ns = "np")
  on.exit({
    assignInNamespace(".np_plot_activity_begin", orig_activity_begin, ns = "np")
    assignInNamespace(".np_plot_activity_end", orig_activity_end, ns = "np")
    assignInNamespace(".np_inid_boot_from_index", orig_index, ns = "np")
    assignInNamespace(".np_plot_bootstrap_interval_summary", orig_interval, ns = "np")
  }, add = TRUE)

  xdat <- data.frame(x1 = c(0, 1, 2), x2 = c(1, 2, 3))
  ydat <- c(0.1, 0.2, 0.3)
  bws <- list(
    type = "fixed",
    beta = c(1, 1)
  )

  out <- fn(
    xdat = xdat,
    ydat = ydat,
    gradients = FALSE,
    plot.errors.boot.method = "geom",
    plot.errors.boot.blocklen = 2L,
    plot.errors.boot.num = 2L,
    plot.errors.center = "estimate",
    plot.errors.type = "all",
    plot.errors.alpha = 0.05,
    bws = bws
  )

  expect_true(is.list(out))
  expect_true(any(grepl("^Preparing plot bootstrap geom$", captured$prep)))
  expect_true(any(grepl("^Plot bootstrap$", captured$progress)))
  expect_true(any(grepl("^Constructing bootstrap all bands$", captured$interval)))
})

test_that("native progress signal accepts scalar payloads safely", {
  library(np)
  expect_silent(.Call(
    "C_np_progress_signal",
    "bandwidth_activity_step",
    "bandwidth",
    1L,
    2L,
    PACKAGE = "np"
  ))
})

library(np)

locate_native_source <- function(filename) {
  candidates <- c(
    test_path("..", "..", "src", filename),
    test_path("..", "..", "..", "src", filename),
    file.path(Sys.getenv("R_PACKAGE_DIR", ""), "src", filename),
    file.path(Sys.getenv("R_PACKAGE_SOURCE", ""), "src", filename),
    file.path(getwd(), "src", filename),
    file.path(getwd(), "..", "src", filename)
  )
  candidates <- unique(candidates[nzchar(candidates)])
  hits <- candidates[file.exists(candidates)]
  if (length(hits) == 0L) {
    return(NULL)
  }
  hits[[1L]]
}

test_that("tree allocation hotspots use overflow-checked helpers", {
  src_file <- locate_native_source("tree.c")
  skip_if(is.null(src_file), "source file src/tree.c unavailable in this test context")

  src <- paste(readLines(src_file, warn = FALSE), collapse = "\n")

  expect_true(grepl("np_tree_size_mul_or_die", src, fixed = TRUE))
  expect_true(grepl("realloc(ptr, np_tree_size_mul_or_die(", src, fixed = TRUE))
  expect_true(grepl("np_tree_malloc_array_or_die((size_t)numnode, sizeof(KDN), \"build_kdtree kdn\")", src, fixed = TRUE))
  expect_true(grepl("np_tree_size_mul3_or_die", src, fixed = TRUE))
  expect_false(grepl("kdx->kdn = \\(KDN \\*\\)malloc\\(numnode\\*sizeof\\(KDN\\)\\);", src))
  expect_false(grepl("kdx->bb = \\(double \\*\\)malloc\\(numnode\\*2\\*ndim\\*sizeof\\(double\\)\\);", src))
})

test_that("audited jksum allocation hotspots use checked helpers", {
  src_file <- locate_native_source("jksum.c")
  skip_if(is.null(src_file), "source file src/jksum.c unavailable in this test context")

  src <- paste(readLines(src_file, warn = FALSE), collapse = "\n")

  expect_true(grepl("np_jksum_malloc_array_or_die", src, fixed = TRUE))
  expect_true(grepl("np_jksum_malloc_array3_or_die", src, fixed = TRUE))
  expect_true(grepl("np_kernel_estimate_density_categorical_leave_one_out_cv kwx", src, fixed = TRUE))
  expect_true(grepl("np_kernel_estimate_density_categorical_convolution_cv operator", src, fixed = TRUE))
  expect_true(grepl("np_kernel_estimate_con_dens_dist_categorical permn", src, fixed = TRUE))
  expect_false(grepl("permn = \\(double \\*\\)malloc\\(num_X\\*num_obs_eval_alloc\\*sizeof\\(double\\)\\);", src))
  expect_false(grepl("permd = \\(double \\*\\)malloc\\(num_X\\*num_obs_eval_alloc\\*sizeof\\(double\\)\\);", src))
})

npindexbw_default <- getFromNamespace("npindexbw.default", "np")
npindexbw_sibandwidth <- getFromNamespace("npindexbw.sibandwidth", "np")
npindexbw_h_start_controls <- getFromNamespace(".npindexbw_h_start_controls", "np")
npindex_default_start_bandwidth <- getFromNamespace(".npindex_default_start_bandwidth", "np")
npindex_random_start_bandwidth <- getFromNamespace(".npindex_random_start_bandwidth", "np")
npindex_nomad_fixed_start_setup <- getFromNamespace(".npindexbw_nomad_fixed_start_setup", "np")

test_that("npindexbw surfaces fixed-h start controls as formal arguments", {
  expect_true(all(c("scale.factor.init.lower", "scale.factor.init.upper", "scale.factor.init") %in% names(formals(npindexbw_default))))
  expect_true(all(c("scale.factor.init.lower", "scale.factor.init.upper", "scale.factor.init") %in% names(formals(npindexbw_sibandwidth))))
})

test_that("npindexbw fixed-h helper replays legacy defaults", {
  fit <- c(-1.5, -0.25, 0.1, 0.8, 1.7)
  controls <- npindexbw_h_start_controls()

  expect_equal(
    npindex_default_start_bandwidth(fit = fit, bwtype = "fixed", nobs = 80L, start.controls = controls),
    1.059224 * getFromNamespace("EssDee", "np")(fit) * 80^(-1 / 5)
  )

  set.seed(20260405)
  got <- npindex_random_start_bandwidth(
    fit = fit,
    bwtype = "fixed",
    nobs = 80L,
    start.controls = controls
  )
  set.seed(20260405)
  expect_equal(got, runif(1, min = 0.5, max = 1.5) * getFromNamespace("EssDee", "np")(fit) * 80^(-1 / 5))
})

test_that("npindexbw fixed-h helper honors explicit overrides", {
  fit <- c(-1.5, -0.25, 0.1, 0.8, 1.7)
  controls <- npindexbw_h_start_controls(scale.factor.init.lower = 0.7, scale.factor.init.upper = 0.9, scale.factor.init = 1.2)

  expect_equal(
    npindex_default_start_bandwidth(fit = fit, bwtype = "fixed", nobs = 80L, start.controls = controls),
    1.2 * getFromNamespace("EssDee", "np")(fit) * 80^(-1 / 5)
  )

  set.seed(11)
  got <- npindex_random_start_bandwidth(
    fit = fit,
    bwtype = "fixed",
    nobs = 80L,
    start.controls = controls
  )
  set.seed(11)
  expect_equal(got, runif(1, min = 0.7, max = 0.9) * getFromNamespace("EssDee", "np")(fit) * 80^(-1 / 5))
})

test_that("npindexbw nearest-neighbor h starts ignore fixed-h controls", {
  fit <- c(-1.5, -0.25, 0.1, 0.8, 1.7)
  controls <- npindexbw_h_start_controls(scale.factor.init.lower = 0.7, scale.factor.init.upper = 0.9, scale.factor.init = 1.2)

  expect_equal(
    npindex_default_start_bandwidth(fit = fit, bwtype = "generalized_nn", nobs = 80L, start.controls = controls),
    npindex_default_start_bandwidth(fit = fit, bwtype = "generalized_nn", nobs = 80L)
  )

  set.seed(20260405)
  got <- npindex_random_start_bandwidth(
    fit = fit,
    bwtype = "generalized_nn",
    nobs = 80L,
    start.controls = controls
  )
  set.seed(20260405)
  expect_equal(got, npindex_random_start_bandwidth(fit = fit, bwtype = "generalized_nn", nobs = 80L))
})

test_that("npindexbw fixed NOMAD start setup preserves beta starts and replays default h starts", {
  set.seed(42)
  n <- 120L
  x1 <- runif(n, -1, 1)
  x2 <- runif(n, -1, 1)
  y <- x1 - x2 + rnorm(n)
  xmat <- cbind(x1 = x1, x2 = x2)
  degree.search <- list(start.degree = 1L, lower = 0L, upper = 3L, basis = "glp", nobs = n, start.user = FALSE)
  baseline.bws <- list(beta = c(1, 0), bw = 0)

  legacy <- npindex_nomad_fixed_start_setup(
    xmat = xmat,
    ydat = y,
    baseline.bws = baseline.bws,
    degree.search = degree.search,
    nmulti = 4L,
    random.seed = 42L
  )
  replay <- npindex_nomad_fixed_start_setup(
    xmat = xmat,
    ydat = y,
    baseline.bws = baseline.bws,
    degree.search = degree.search,
    nmulti = 4L,
    random.seed = 42L,
    h.start.controls = npindexbw_h_start_controls()
  )

  expect_equal(replay$start_matrix.raw, legacy$start_matrix.raw)
  expect_equal(replay$start_matrix.point, legacy$start_matrix.point)

  custom <- npindex_nomad_fixed_start_setup(
    xmat = xmat,
    ydat = y,
    baseline.bws = baseline.bws,
    degree.search = degree.search,
    nmulti = 4L,
    random.seed = 42L,
    h.start.controls = npindexbw_h_start_controls(scale.factor.init.lower = 0.7, scale.factor.init.upper = 0.9, scale.factor.init = 1.2)
  )

  expect_equal(custom$start_matrix.raw[, legacy$beta.free, drop = FALSE],
               legacy$start_matrix.raw[, legacy$beta.free, drop = FALSE])
  expect_false(isTRUE(all.equal(custom$start_matrix.raw[, legacy$h.col], legacy$start_matrix.raw[, legacy$h.col])))
})

test_that("bootstrap quantile bounds preserve legacy numerical results", {
  skip_if_not_installed("np")

  new_bounds <- getFromNamespace("compute.bootstrap.quantile.bounds", "np")
  new_scs <- getFromNamespace("np.plot.SCSrank", "np")

  old_scs <- function(x, conf.level = 0.95, alternative = "two.sided") {
    alternative <- match.arg(alternative, choices = c("two.sided", "less", "greater"))
    DataMatrix <- x
    N <- nrow(DataMatrix)
    k <- round(conf.level * N, 0)
    RankDat <- apply(DataMatrix, 2, rank)

    switch(alternative,
      "two.sided" = {
        W1 <- apply(RankDat, 1, max)
        W2 <- N + 1 - apply(RankDat, 1, min)
        Wmat <- cbind(W1, W2)
        w <- apply(Wmat, 1, max)
        tstar <- round(sort(w)[k], 0)
        SCI <- function(x) {
          sortx <- sort(x)
          cbind(sortx[N + 1 - tstar], sortx[tstar])
        }
        SCS <- t(apply(DataMatrix, 2, SCI))
      },
      "less" = {
        W1 <- apply(RankDat, 1, max)
        tstar <- round(sort(W1)[k], 0)
        SCI <- function(x) {
          sortx <- sort(x)
          cbind(-Inf, sortx[tstar])
        }
        SCS <- t(apply(DataMatrix, 2, SCI))
      },
      "greater" = {
        W2 <- N + 1 - apply(RankDat, 1, min)
        tstar <- round(sort(W2)[k], 0)
        SCI <- function(x) {
          sortx <- sort(x)
          cbind(sortx[N + 1 - tstar], Inf)
        }
        SCS <- t(apply(DataMatrix, 2, SCI))
      }
    )

    colnames(SCS) <- c("lower", "upper")
    attr(SCS, which = "k") <- k
    attr(SCS, which = "N") <- N
    list(conf.int = SCS, conf.level = conf.level, alternative = alternative)
  }

  old_bounds <- function(boot.t, alpha, band.type) {
    B <- nrow(boot.t)
    neval <- ncol(boot.t)
    if (band.type == "pointwise") {
      probs <- c(alpha / 2.0, 1.0 - alpha / 2.0)
      return(t(apply(boot.t, 2, quantile, probs = probs)))
    }
    if (band.type == "bonferroni") {
      probs <- c(alpha / (2.0 * neval), 1.0 - alpha / (2.0 * neval))
      return(t(apply(boot.t, 2, quantile, probs = probs)))
    }
    if (band.type == "simultaneous") {
      return(old_scs(boot.t, conf.level = 1.0 - alpha)$conf.int)
    }
    if (band.type == "all") {
      return(list(
        pointwise = old_bounds(boot.t, alpha, "pointwise"),
        bonferroni = old_bounds(boot.t, alpha, "bonferroni"),
        simultaneous = old_bounds(boot.t, alpha, "simultaneous")
      ))
    }
    stop("unsupported band type")
  }

  set.seed(20260310L)
  boot.t <- matrix(sample(c(-2, -1, 0, 0, 1, 2, 3), size = 257L * 9L, replace = TRUE), nrow = 257L, ncol = 9L)
  alpha <- 0.1

  expect_equal(new_bounds(boot.t, alpha, "pointwise", warn.coverage = FALSE), old_bounds(boot.t, alpha, "pointwise"))
  expect_equal(new_bounds(boot.t, alpha, "bonferroni", warn.coverage = FALSE), old_bounds(boot.t, alpha, "bonferroni"))
  expect_equal(new_bounds(boot.t, alpha, "simultaneous", warn.coverage = FALSE), old_bounds(boot.t, alpha, "simultaneous"))

  all.new <- new_bounds(boot.t, alpha, "all", warn.coverage = FALSE)
  all.old <- old_bounds(boot.t, alpha, "all")
  expect_equal(all.new$pointwise, all.old$pointwise)
  expect_equal(all.new$bonferroni, all.old$bonferroni)
  expect_equal(all.new$simultaneous, all.old$simultaneous)

  expect_equal(new_scs(boot.t, conf.level = 0.9, alternative = "two.sided")$conf.int,
               old_scs(boot.t, conf.level = 0.9, alternative = "two.sided")$conf.int)
  expect_equal(new_scs(boot.t, conf.level = 0.9, alternative = "less")$conf.int,
               old_scs(boot.t, conf.level = 0.9, alternative = "less")$conf.int)
  expect_equal(new_scs(boot.t, conf.level = 0.9, alternative = "greater")$conf.int,
               old_scs(boot.t, conf.level = 0.9, alternative = "greater")$conf.int)
})

library(np)

locate_native_source <- function(filename) {
  candidates <- c(
    test_path("..", "..", "src", filename),
    test_path("..", "..", "..", "src", filename),
    file.path(Sys.getenv("R_PACKAGE_DIR", ""), "src", filename),
    file.path(Sys.getenv("R_PACKAGE_SOURCE", ""), "src", filename),
    file.path(getwd(), "src", filename),
    file.path(getwd(), "..", "src", filename)
  )
  candidates <- unique(candidates[nzchar(candidates)])
  hits <- candidates[file.exists(candidates)]
  if (length(hits) == 0L) {
    return(NULL)
  }
  hits[[1L]]
}

test_that("mat_vec allocation helpers free partial state before error", {
  src_file <- locate_native_source("mat_vec.c")
  skip_if(is.null(src_file), "source file src/mat_vec.c unavailable in this test context")

  src <- paste(readLines(src_file, warn = FALSE), collapse = "\n")

  expect_true(grepl("free(m);", src, fixed = TRUE))
  expect_true(grepl("free_mat(m, i);", src, fixed = TRUE))
})

test_that("linalg helpers clean up audited allocations before malloc errors", {
  src_file <- locate_native_source("linalg.c")
  skip_if(is.null(src_file), "source file src/linalg.c unavailable in this test context")

  src <- paste(readLines(src_file, warn = FALSE), collapse = "\n")

  expect_true(grepl("while \\(--i >= 0\\)\\s*free\\(\\*\\(\\(double \\*\\*\\)\\(&mat->matrix\\) \\+ i\\)\\);\\s*free\\(mat\\);\\s*error\\(\"mat: malloc error\\\\n\" \\);", src))
  expect_true(grepl("free\\(A\\);\\s*free\\(ipiv\\);\\s*error\\(\"mat_inv: malloc error\\\\n\"\\);", src))
  expect_true(grepl("free\\(Ac\\);\\s*free\\(Bc\\);\\s*free\\(ipiv\\);\\s*error\\(\"mat_solve: malloc error\\\\n\"\\);", src))
  expect_true(grepl("free\\(Ac\\);\\s*free\\(ipiv\\);\\s*error\\(\"mat_is_nonsingular: malloc error\\\\n\"\\);", src))
  expect_true(grepl("free\\(Ac\\);\\s*free\\(bc\\);\\s*free\\(ipiv\\);\\s*error\\(\"mat_inv00: malloc error\\\\n\"\\);", src))
})

library(np)

test_that("npcdens formula fit keeps lc ll and lp estimator paths live", {
  set.seed(20260307)
  n <- 60L
  x <- rnorm(n)
  y <- x + rnorm(n)

  fit.lc <- npcdens(y ~ x, regtype = "lc", nmulti = 1)
  fit.ll <- npcdens(y ~ x, regtype = "ll", nmulti = 1)
  fit.lp0 <- npcdens(y ~ x, regtype = "lp", degree = 0L, nmulti = 1)
  fit.lp2 <- npcdens(y ~ x, regtype = "lp", degree = 2L, nmulti = 1)

  expect_true(all(is.finite(fitted(fit.lc))))
  expect_true(all(is.finite(fitted(fit.ll))))
  expect_true(all(is.finite(fitted(fit.lp0))))
  expect_true(all(is.finite(fitted(fit.lp2))))

  expect_equal(fitted(fit.lc), fitted(fit.lp0), tolerance = 1e-8)

  for (fit in list(fit.lc, fit.ll, fit.lp0, fit.lp2)) {
    expect_true(length(capture.output(summary(fit))) > 0L)

    pdf(file = tempfile(fileext = ".pdf"))
    on.exit(dev.off(), add = TRUE)
    expect_error(plot(fit, view = "fixed"), NA)
  }
})

test_that("npregiv ridging loops avoid superassignment", {
  src_path <- testthat::test_path("..", "..", "R", "npregiv.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")
  expect_false(grepl("<<-", src, fixed = TRUE))
})

test_that("npregiv uses shared seed enter helper", {
  src_path <- testthat::test_path("..", "..", "R", "npregiv.R")
  skip_if_not(file.exists(src_path), "source R files unavailable in installed test context")
  src <- paste(readLines(src_path, warn = FALSE), collapse = "\n")
  expect_true(grepl("\\.np_seed_enter\\(random\\.seed\\)", src))
  expect_false(grepl("exists\\(\"\\.Random\\.seed\"", src))
  expect_true(grepl("\\.np_seed_exit\\(seed\\.state, remove_if_absent = TRUE\\)", src))
})

test_that("bounded adaptive_nn is available for conditional density", {
  set.seed(20260325)
  n <- 36
  x <- runif(n)
  y <- runif(n)
  xdf <- data.frame(x = x)
  ydf <- data.frame(y = y)

  bw.fixed <- npcdensbw(
    xdat = xdf,
    ydat = ydf,
    bwmethod = "cv.ml",
    bwtype = "fixed",
    cxkerorder = 2,
    cykerorder = 2,
    cxkerbound = "range",
    cykerbound = "range",
    nmulti = 1,
    powell.remin = FALSE,
    itmax = 50,
    tol = 0.1
  )

  bw.gnn <- npcdensbw(
    xdat = xdf,
    ydat = ydf,
    bwmethod = "cv.ml",
    bwtype = "generalized_nn",
    cxkerorder = 2,
    cykerorder = 2,
    cxkerbound = "range",
    cykerbound = "range",
    nmulti = 1,
    powell.remin = FALSE,
    itmax = 50,
    tol = 0.1
  )

  bw.ad <- npcdensbw(
    xdat = xdf,
    ydat = ydf,
    bwmethod = "cv.ml",
    bwtype = "adaptive_nn",
    cxkerorder = 2,
    cykerorder = 2,
    cxkerbound = "range",
    cykerbound = "range",
    nmulti = 1,
    powell.remin = FALSE,
    itmax = 50,
    tol = 0.1
  )

  fit.ad <- npcdens(bws = bw.ad)

  expect_true(is.finite(bw.fixed$fval))
  expect_true(is.finite(bw.gnn$fval))
  expect_true(is.finite(bw.ad$fval))
  expect_true(all(is.finite(as.numeric(bw.ad$xbw))))
  expect_true(all(is.finite(as.numeric(bw.ad$ybw))))
  expect_true(all(is.finite(as.numeric(fitted(fit.ad)))))
  expect_true(all(as.numeric(fitted(fit.ad)) >= 0))
  expect_false(isTRUE(all.equal(as.numeric(bw.ad$xbw), as.numeric(bw.fixed$xbw))))
})

test_that("nonfixed unconditional exact bootstrap matches duplicate-row oracle", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  xdat <- data.frame(x = c(0, 0, 1, 3, 3, 10))
  exdat <- data.frame(x = c(0, 2, 5, 9))
  counts <- cbind(
    c(1, 0, 1, 2, 0, 2),
    c(0, 2, 1, 0, 1, 2)
  )
  storage.mode(counts) <- "double"

  for (bt in c("generalized_nn", "adaptive_nn")) {
    bw.val <- if (identical(bt, "adaptive_nn")) 2 else 1
    bw <- npudensbw(
      dat = xdat,
      bwtype = bt,
      bws = bw.val,
      bandwidth.compute = FALSE
    )

    helper <- np:::.np_inid_boot_from_ksum_unconditional_exact(
      xdat = xdat,
      exdat = exdat,
      bws = bw,
      B = ncol(counts),
      operator = "normal",
      counts = counts
    )

    manual <- vapply(seq_len(ncol(counts)), function(j) {
      idx <- np:::.np_counts_to_indices(counts[, j])
      np:::.np_ksum_unconditional_eval_exact(
        xdat = xdat[idx, , drop = FALSE],
        exdat = exdat,
        bws = bw,
        operator = "normal"
      )
    }, numeric(nrow(exdat)))

    expect_equal(helper$t0, np:::.np_ksum_unconditional_eval_exact(
      xdat = xdat,
      exdat = exdat,
      bws = bw,
      operator = "normal"
    ), info = bt)
    expect_equal(helper$t, t(manual), tolerance = 1e-14, info = bt)
  }
})

test_that("active bootstrap sample preserves exact row-sliced training payload", {
  xdat <- data.frame(
    x = c(0, 1, 3, 7),
    xf = factor(c("a", "b", "b", "c"))
  )
  ydat <- data.frame(y = c(2, 4, 6, 8))
  counts <- c(0, 2, 0, 1)

  out <- np:::.np_active_boot_sample(xdat = xdat, ydat = ydat, counts.col = counts)

  expect_identical(out$xdat, xdat[c(2, 4), , drop = FALSE])
  expect_identical(out$ydat, ydat[c(2, 4), , drop = FALSE])
  expect_identical(out$weights, matrix(c(2, 1), ncol = 1L))
  expect_identical(out$n.total, sum(counts))
})

test_that("fast exact frame bind matches data.frame semantics", {
  xdat <- data.frame("a b" = c(1, 2), xf = factor(c("a", "b")), check.names = FALSE)
  ydat <- data.frame("a b" = c(3, 4), y = c(5, 6), check.names = FALSE)

  expect_identical(
    np:::.np_bind_data_frames_fast(xdat, ydat),
    data.frame(xdat, ydat)
  )
})

test_that("nonfixed unconditional exact helper matches direct kbandwidth precompute", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  xdat <- data.frame(x = c(0, 0, 1, 3, 3, 10))
  exdat <- data.frame(x = c(0, 2, 5, 9))
  counts <- c(1, 1, 1, 2, 1, 2)
  active <- counts > 0

  for (bt in c("generalized_nn", "adaptive_nn")) {
    bw.val <- if (identical(bt, "adaptive_nn")) 2 else 1
    bw <- npudensbw(
      dat = xdat,
      bwtype = bt,
      bws = bw.val,
      bandwidth.compute = FALSE
    )

    kb <- np:::.np_make_kbandwidth_unconditional(bws = bw, xdat = xdat)
    weighted.active <- matrix(as.double(counts[active]), ncol = 1L)

    direct.bw <- np:::.np_ksum_unconditional_eval_exact(
      xdat = xdat[active, , drop = FALSE],
      exdat = exdat,
      bws = bw,
      operator = "normal",
      weights = weighted.active,
      n.total = sum(counts)
    )

    direct.kb <- np:::.np_ksum_unconditional_eval_exact(
      xdat = xdat[active, , drop = FALSE],
      exdat = exdat,
      bws = kb,
      operator = "normal",
      weights = weighted.active,
      n.total = sum(counts)
    )

    expect_equal(direct.kb, direct.bw, tolerance = 0, info = bt)
  }
})

test_that("nonfixed conditional exact bootstrap matches duplicate-row oracle", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  xdat <- data.frame(x = c(0, 0, 1, 3, 3, 10))
  ydat <- data.frame(y = c(0, 0, 1, 2, 2, 4))
  exdat <- data.frame(x = c(0, 2, 5, 9))
  eydat <- data.frame(y = c(0, 1, 2, 4))
  counts <- cbind(
    c(1, 0, 1, 2, 0, 2),
    c(0, 2, 1, 0, 1, 2)
  )
  storage.mode(counts) <- "double"

  for (bt in c("generalized_nn", "adaptive_nn")) {
    bw.val <- if (identical(bt, "adaptive_nn")) c(2, 2) else c(1, 1)
    bw <- npcdensbw(
      xdat = xdat,
      ydat = ydat,
      bwtype = bt,
      bws = bw.val,
      bandwidth.compute = FALSE
    )

    helper <- np:::.np_inid_boot_from_ksum_conditional_exact(
      xdat = xdat,
      ydat = ydat,
      exdat = exdat,
      eydat = eydat,
      bws = bw,
      B = ncol(counts),
      cdf = FALSE,
      counts = counts
    )

    kbx <- np:::.np_con_make_kbandwidth_x(bws = bw, xdat = xdat)
    kbxy <- np:::.np_con_make_kbandwidth_xy(bws = bw, xdat = xdat, ydat = ydat)

    manual <- vapply(seq_len(ncol(counts)), function(j) {
      idx <- np:::.np_counts_to_indices(counts[, j])
      np:::.np_ksum_conditional_eval_exact(
        xdat = xdat[idx, , drop = FALSE],
        ydat = ydat[idx, , drop = FALSE],
        exdat = exdat,
        eydat = eydat,
        kbx = kbx,
        kbxy = kbxy,
        cdf = FALSE
      )
    }, numeric(nrow(exdat)))

    expect_equal(helper$t0, np:::.np_ksum_conditional_eval_exact(
      xdat = xdat,
      ydat = ydat,
      exdat = exdat,
      eydat = eydat,
      kbx = kbx,
      kbxy = kbxy,
      cdf = FALSE
    ), info = bt)
    expect_equal(helper$t, t(manual), tolerance = 1e-14, info = bt)
  }
})

test_that("generalized conditional exact state apply matches weighted active-support oracle", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  xdat <- data.frame(x = c(0, 0, 1, 3, 3, 10))
  ydat <- data.frame(y = c(0, 0, 1, 2, 2, 4))
  exdat <- data.frame(x = c(0, 2, 5, 9))
  eydat <- data.frame(y = c(0, 1, 2, 4))
  counts <- c(1, 0, 1, 2, 0, 2)
  active <- counts > 0

  bw <- npcdensbw(
    xdat = xdat,
    ydat = ydat,
    bwtype = "generalized_nn",
    bws = c(1, 1),
    bandwidth.compute = FALSE
  )

  den.state <- np:::.np_ksum_exact_state_build(
    bws = np:::.np_con_make_kbandwidth_x(bws = bw, xdat = xdat),
    exdat = exdat,
    operator = "normal"
  )
  num.state <- np:::.np_ksum_exact_state_build(
    bws = np:::.np_con_make_kbandwidth_xy(bws = bw, xdat = xdat, ydat = ydat),
    exdat = np:::.np_bind_data_frames_fast(exdat, eydat),
    operator = c("normal", "normal")
  )

  sample <- np:::.np_active_boot_sample_matrix(
    xmat = data.matrix(xdat),
    ymat = cbind(data.matrix(xdat), data.matrix(ydat)),
    counts.col = counts
  )

  den <- as.numeric(np:::.np_ksum_eval_exact_state(
    state = den.state,
    txdat = sample$xmat,
    weights = sample$weights
  )) / sample$n.total

  num <- as.numeric(np:::.np_ksum_eval_exact_state(
    state = num.state,
    txdat = sample$ymat,
    weights = sample$weights
  )) / sample$n.total

  expect_equal(
    num / pmax(den, .Machine$double.eps),
    np:::.np_ksum_conditional_eval_exact(
      xdat = xdat[active, , drop = FALSE],
      ydat = ydat[active, , drop = FALSE],
      exdat = exdat,
      eydat = eydat,
      kbx = np:::.np_con_make_kbandwidth_x(bws = bw, xdat = xdat),
      kbxy = np:::.np_con_make_kbandwidth_xy(bws = bw, xdat = xdat, ydat = ydat),
      cdf = FALSE,
      weights = counts[active],
      n.total = sum(counts)
    ),
    tolerance = 0
  )
})

test_that("adaptive conditional exact handles tiny-support resamples consistently", {
  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  xdat <- data.frame(x = c(0, 0, 1, 3, 3, 10))
  ydat <- data.frame(y = c(0, 0, 1, 2, 2, 4))
  exdat <- data.frame(x = c(0, 2, 5, 9))
  eydat <- data.frame(y = c(0, 1, 2, 4))
  counts <- matrix(c(0, 2, 1, 0, 3, 0), ncol = 1L)
  storage.mode(counts) <- "double"

  bw <- npcdensbw(
    xdat = xdat,
    ydat = ydat,
    bwtype = "adaptive_nn",
    bws = c(2, 2),
    bandwidth.compute = FALSE
  )

  idx <- np:::.np_counts_to_indices(counts[, 1L])
  explicit <- npcdens(
    txdat = xdat[idx, , drop = FALSE],
    tydat = ydat[idx, , drop = FALSE],
    exdat = exdat,
    eydat = eydat,
    bws = npcdensbw(
      xdat = xdat[idx, , drop = FALSE],
      ydat = ydat[idx, , drop = FALSE],
      bwtype = "adaptive_nn",
      bws = c(2, 2),
      bandwidth.compute = FALSE
    )
  )

  helper <- np:::.np_inid_boot_from_ksum_conditional_exact(
    xdat = xdat,
    ydat = ydat,
    exdat = exdat,
    eydat = eydat,
    bws = bw,
    B = 1L,
    cdf = FALSE,
    counts = counts
  )

  expect_equal(as.numeric(helper$t[1L, ]), as.numeric(explicit$condens), tolerance = 1e-15)
})

test_that("plot contract: single-index asymptotic consumer payloads are supported", {
  skip_if_not_installed("np")

  set.seed(921)
  n <- 60
  x1 <- runif(n)
  x2 <- runif(n)
  y.cont <- sin(2 * pi * x1) + 0.35 * x2 + rnorm(n, sd = 0.08)
  y.bin <- as.numeric(x1 + x2 + rnorm(n, sd = 0.2) > 1)
  xdat <- data.frame(x1 = x1, x2 = x2)

  reg.cfgs <- list(
    list(regtype = "lc"),
    list(regtype = "ll"),
    list(regtype = "lp", basis = "tensor", degree = 2L)
  )
  neval <- 50L

  for (method in c("ichimura", "kleinspady")) {
    y.use <- if (identical(method, "kleinspady")) y.bin else y.cont
    for (bwtype in c("fixed", "generalized_nn", "adaptive_nn")) {
      h <- if (identical(bwtype, "fixed")) 0.30 else 6L
      for (cfg in reg.cfgs) {
        bw.args <- list(
          xdat = xdat,
          ydat = y.use,
          bws = c(1, h, h),
          bandwidth.compute = FALSE,
          method = method,
          bwtype = bwtype,
          regtype = cfg$regtype
        )
        if (!is.null(cfg$basis)) {
          bw.args$basis <- cfg$basis
          bw.args$degree <- cfg$degree
        }

        bw <- do.call(npindexbw, bw.args)
        out <- suppressWarnings(plot(
          bw,
          xdat = xdat,
          ydat = y.use,
          neval = neval,
          output = "data",
          errors = "asymptotic",
          band = "pointwise",
          perspective = FALSE
        ))[[1]]

        expect_s3_class(out, "singleindex")
        expect_equal(length(out$mean), neval, info = paste(method, bwtype, cfg$regtype, "mean length"))
        expect_equal(dim(out$merr), c(neval, 2L), info = paste(method, bwtype, cfg$regtype, "merr shape"))
        expect_true(all(is.finite(out$mean)), info = paste(method, bwtype, cfg$regtype, "mean finite"))
        expect_true(all(is.finite(out$merr)), info = paste(method, bwtype, cfg$regtype, "merr finite"))
      }
    }
  }
})

test_that("plot contract: partially linear asymptotic consumer payloads are supported", {
  skip_if_not_installed("np")

  set.seed(922)
  n <- 55
  x <- runif(n)
  z <- runif(n)
  y <- sin(2 * pi * z) + 0.5 * x + rnorm(n, sd = 0.08)

  reg.cfgs <- list(
    list(regtype = "lc"),
    list(regtype = "ll"),
    list(regtype = "lp", basis = "tensor", degree = 2L)
  )

  for (bwmethod in c("cv.ls", "cv.aic")) {
    for (bwtype in c("fixed", "generalized_nn", "adaptive_nn")) {
      h <- if (identical(bwtype, "fixed")) 0.30 else 6L
      for (cfg in reg.cfgs) {
        bw.args <- list(
          xdat = data.frame(x = x),
          zdat = data.frame(z = z),
          ydat = y,
          bws = matrix(c(h, h), nrow = 2L),
          bandwidth.compute = FALSE,
          bwmethod = bwmethod,
          bwtype = bwtype,
          regtype = cfg$regtype
        )
        if (!is.null(cfg$basis)) {
          bw.args$basis <- cfg$basis
          bw.args$degree <- cfg$degree
        }

        bw <- do.call(npplregbw, bw.args)
        out <- suppressWarnings(plot(
          bw,
          xdat = data.frame(x = x),
          ydat = y,
          zdat = data.frame(z = z),
          output = "data",
          errors = "asymptotic",
          band = "pointwise",
          perspective = FALSE
        ))[[1]]

        expect_s3_class(out, "plregression")
        expect_true(length(out$mean) > 0L, info = paste(bwmethod, bwtype, cfg$regtype, "mean length"))
        expect_equal(dim(out$merr), c(length(out$mean), 2L), info = paste(bwmethod, bwtype, cfg$regtype, "merr shape"))
        expect_true(all(is.finite(out$mean)), info = paste(bwmethod, bwtype, cfg$regtype, "mean finite"))
        expect_true(all(is.finite(out$merr)), info = paste(bwmethod, bwtype, cfg$regtype, "merr finite"))
      }
    }
  }
})

gsl_bs_np <- getFromNamespace("gsl.bs", "np")
predict_gsl_bs_np <- getFromNamespace("predict.gsl.bs", "np")

capture_gsl_bs_eval <- function(expr) {
  warnings <- character()

  value <- tryCatch(
    withCallingHandlers(
      expr,
      warning = function(w) {
        warnings <<- c(warnings, conditionMessage(w))
        invokeRestart("muffleWarning")
      }
    ),
    error = function(e) {
      structure(
        list(
          message = conditionMessage(e),
          class = class(e)
        ),
        class = "captured_error"
      )
    }
  )

  if (inherits(value, "captured_error")) {
    return(list(
      ok = FALSE,
      warnings = warnings,
      error_message = value$message,
      error_class = value$class
    ))
  }

  list(
    ok = TRUE,
    warnings = warnings,
    value = value
  )
}

predict_contract_cases <- list(
  uniform_intercept = list(
    x = c(-1, -0.5, 0, 0.5, 1),
    degree = 3,
    nbreak = 4,
    deriv = 0,
    intercept = TRUE,
    x.min = NULL,
    x.max = NULL,
    knots = NULL
  ),
  deriv_knots_no_intercept = list(
    x = c(-1, -0.5, 0, 0.5, 1),
    degree = 3,
    nbreak = 4,
    deriv = 1,
    intercept = FALSE,
    x.min = -1,
    x.max = 1,
    knots = c(-1, -0.3, 0.2, 1)
  )
)

predict_contract_newx_cases <- list(
  newx_null = NULL,
  numeric_vector = c(-1, -0.25, 0.25, 1),
  integer_vector = as.integer(c(-1, 0, 1)),
  scalar_numeric = 0.2,
  numeric_matrix = matrix(c(-1, -0.25, 0.25, 1), ncol = 1),
  integer_matrix = matrix(as.integer(c(-1, 0, 1)), ncol = 1),
  named_vector = setNames(c(-1, 0, 1), c("a", "b", "c")),
  duplicates_outside = c(-2, -1, -1, 0, 1, 2),
  zero_length = numeric(0),
  na_nan_inf = c(NA_real_, NaN, Inf, -Inf, 0)
)

test_that("np gsl.bs uses package-specific first class with compatibility superclass", {
  obj <- gsl_bs_np(c(-1, -0.5, 0, 0.5, 1), degree = 3, nbreak = 4, intercept = TRUE)

  expect_identical(class(obj), c("np_gsl.bs", "gsl.bs", "matrix"))
  expect_true(inherits(obj, "gsl.bs"))
  expect_true(inherits(obj, "matrix"))
  expect_true(is.function(getS3method("predict", "np_gsl.bs")))
})

test_that("np gsl.bs rejects malformed derivative and knot inputs before native evaluation", {
  x <- c(-1, -0.5, 0, 0.5, 1)

  expect_error(gsl_bs_np(x, deriv = NA_integer_), "deriv")
  expect_error(gsl_bs_np(x, deriv = -1L), "deriv")
  expect_error(gsl_bs_np(x, deriv = 1.5), "deriv")
  expect_error(gsl_bs_np(x, nbreak = NA_integer_), "nbreak")
  expect_error(gsl_bs_np(x, nbreak = 1.5), "nbreak")
  expect_error(gsl_bs_np(x, knots = NA_real_), "knots")
  expect_error(gsl_bs_np(x, knots = 0.5), "knots")

  bs_des_np <- getFromNamespace("bs.des", "np")
  expect_error(
    bs_des_np(x, degree = 3, nbreak = 4, deriv = c(0L, 1L)),
    "length equal to x"
  )
  expect_error(
    bs_des_np(x, degree = 3, nbreak = 4,
              deriv = c(0L, 1L, NA_integer_, 0L, 1L)),
    "deriv"
  )
  expect_error(
    bs_des_np(x, degree = 3, nbreak = 4, deriv = rep(5L, length(x))),
    "degree plus 2"
  )
})

test_that("np gsl.bs predict dispatch matches legacy wrapper across supported newx forms", {
  method <- getS3method("predict", "np_gsl.bs")

  for (case_name in names(predict_contract_cases)) {
    args_case <- predict_contract_cases[[case_name]]
    obj <- do.call(gsl_bs_np, args_case)

    expect_identical(class(obj), c("np_gsl.bs", "gsl.bs", "matrix"))

    for (newx_name in names(predict_contract_newx_cases)) {
      newx <- predict_contract_newx_cases[[newx_name]]

      dispatch_res <- capture_gsl_bs_eval(predict(obj, newx = newx))
      method_res <- capture_gsl_bs_eval(method(obj, newx = newx))
      legacy_res <- capture_gsl_bs_eval(predict_gsl_bs_np(obj, newx = newx))

      expect_identical(
        dispatch_res,
        method_res,
        info = sprintf("%s/%s dispatch vs method", case_name, newx_name)
      )
      expect_identical(
        dispatch_res,
        legacy_res,
        info = sprintf("%s/%s dispatch vs legacy", case_name, newx_name)
      )

      if (dispatch_res$ok) {
        if (is.null(newx)) {
          expected <- obj
        } else {
          expected <- gsl_bs_np(
            x = as.numeric(newx),
            degree = attr(obj, "degree"),
            nbreak = attr(obj, "nbreak"),
            deriv = attr(obj, "deriv"),
            x.min = attr(obj, "x.min"),
            x.max = attr(obj, "x.max"),
            intercept = attr(obj, "intercept"),
            knots = attr(obj, "knots")
          )
        }

        attr(expected, "newx") <- if (is.null(newx)) NULL else as.numeric(newx)
        attr(expected, "newx.trimmed") <- NULL

        expect_identical(
          dispatch_res$value,
          expected,
          info = sprintf("%s/%s semantic parity", case_name, newx_name)
        )
      }
    }
  }
})

library(np)

locate_jksum_c <- function() {
  candidates <- c(
    test_path("..", "..", "src", "jksum.c"),
    test_path("..", "..", "..", "src", "jksum.c"),
    file.path(Sys.getenv("R_PACKAGE_DIR", ""), "src", "jksum.c"),
    file.path(Sys.getenv("R_PACKAGE_SOURCE", ""), "src", "jksum.c"),
    file.path(getwd(), "src", "jksum.c"),
    file.path(getwd(), "..", "src", "jksum.c")
  )
  candidates <- unique(candidates[nzchar(candidates)])
  hits <- candidates[file.exists(candidates)]
  if (length(hits) == 0L) {
    return(NULL)
  }
  hits[[1L]]
}

locate_kernelcv_c <- function() {
  candidates <- c(
    test_path("..", "..", "src", "kernelcv.c"),
    test_path("..", "..", "..", "src", "kernelcv.c"),
    file.path(Sys.getenv("R_PACKAGE_DIR", ""), "src", "kernelcv.c"),
    file.path(Sys.getenv("R_PACKAGE_SOURCE", ""), "src", "kernelcv.c"),
    file.path(getwd(), "src", "kernelcv.c"),
    file.path(getwd(), "..", "src", "kernelcv.c")
  )
  candidates <- unique(candidates[nzchar(candidates)])
  hits <- candidates[file.exists(candidates)]
  if (length(hits) == 0L) {
    return(NULL)
  }
  hits[[1L]]
}

test_that("CVLS LL/LP route predicate is centralized in one helper", {
  src_file <- locate_jksum_c()
  skip_if(is.null(src_file), "source file src/jksum.c unavailable in this test context")

  lines <- readLines(src_file, warn = FALSE)

  expect_true(any(grepl("np_reg_cv_use_symmetric_dropone_path", lines, fixed = TRUE)))
  expect_true(any(grepl("np_reg_cv_use_canonical_glp_fixed_kernel", lines, fixed = TRUE)))
  expect_equal(sum(grepl("np_reg_cv_use_degree1_rawbasis_kernel", lines, fixed = TRUE)), 0L)
  expect_equal(sum(grepl("np_regression_cv_degree1_rawbasis_fixed", lines, fixed = TRUE)), 0L)
  expect_equal(sum(grepl("np_reg_degree1_center_raw_moments_at_eval", lines, fixed = TRUE)), 0L)

  raw_pred <- "(bwm == RBWM_CVLS) || ks_tree_use || (BANDWIDTH_reg == BW_ADAP_NN)"
  # Exactly one raw predicate instance is allowed (the helper definition itself).
  expect_equal(sum(grepl(raw_pred, lines, fixed = TRUE)), 1L)
  # The two-term suffix should also only appear inside the helper definition.
  expect_equal(sum(grepl("ks_tree_use || (BANDWIDTH_reg == BW_ADAP_NN)", lines, fixed = TRUE)), 1L)
  # No inline ad hoc route predicates should remain in if-statements.
  expect_equal(sum(grepl("if\\(\\(bwm == RBWM_CVLS\\)", lines)), 0L)
  expect_equal(sum(grepl("if\\(ks_tree_use \\|\\| \\(BANDWIDTH_reg == BW_ADAP_NN\\)\\)", lines)), 0L)

  helper_calls <- sum(grepl("np_reg_cv_use_symmetric_dropone_path\\(", lines))
  expect_gte(helper_calls, 3L)

  canonical_glp_calls <- sum(grepl("np_reg_cv_use_canonical_glp_fixed_kernel\\(", lines))
  expect_gte(canonical_glp_calls, 3L)
})

test_that("density CV tree-bypass predicate is centralized in one helper", {
  src_file <- locate_jksum_c()
  skip_if(is.null(src_file), "source file src/jksum.c unavailable in this test context")

  lines <- readLines(src_file, warn = FALSE)
  expect_true(any(grepl("np_den_cv_use_tree_bypass_path", lines, fixed = TRUE)))

  raw_pred <- "gate_x_all_large_fixed || !int_TREE_XY || (BANDWIDTH_den == BW_ADAP_NN)"
  # One raw predicate instance is allowed (the helper definition itself).
  expect_equal(sum(grepl(raw_pred, lines, fixed = TRUE)), 1L)
  # No inline ad hoc usage in if-statements.
  expect_equal(sum(grepl("if\\(gate_x_all_large_fixed \\|\\| !int_TREE_XY \\|\\| \\(BANDWIDTH_den == BW_ADAP_NN\\)\\)", lines)), 0L)
})

test_that("conditional CV large-kernel gating keeps Y and XY activation routes live", {
  src_file <- locate_jksum_c()
  skip_if(is.null(src_file), "source file src/jksum.c unavailable in this test context")

  lines <- readLines(src_file, warn = FALSE)

  expect_false(any(grepl("Canonical conditional CV gate policy: large-kernel shortcuts are X-only", lines, fixed = TRUE)))
  expect_gte(sum(grepl("^\\s*gate_y_active = 1;\\s*$", lines)), 2L)
  expect_gte(sum(grepl("^\\s*gate_xy_active = 1;\\s*$", lines)), 1L)
  expect_equal(sum(grepl("^\\s*gate_y_active = 0;\\s*$", lines)), 0L)
  expect_equal(sum(grepl("^\\s*gate_xy_active = 0;\\s*$", lines)), 0L)
})

test_that("conditional public LP CV routes stay thin kernelcv dispatches", {
  src_file <- locate_kernelcv_c()
  skip_if(is.null(src_file), "source file src/kernelcv.c unavailable in this test context")

  lines <- readLines(src_file, warn = FALSE)

  expect_equal(sum(grepl("^double cv_func_con_density_categorical_ml\\(", lines)), 1L)
  expect_equal(sum(grepl("^double cv_func_con_density_categorical_ls\\(", lines)), 1L)
  expect_equal(sum(grepl("^double cv_func_con_distribution_categorical_ls\\(", lines)), 1L)
  expect_equal(sum(grepl("^double np_cv_func_con_density_categorical_ml\\(", lines)), 1L)
  expect_equal(sum(grepl("^double np_cv_func_con_density_categorical_ls\\(", lines)), 1L)
  expect_false(any(grepl("np_shadow_proof_cv_con_density_ml\\(", lines)))
  expect_false(any(grepl("np_shadow_proof_cv_con_density_ls\\(", lines)))
  expect_false(any(grepl("np_shadow_proof_cv_con_distribution_ls\\(", lines)))
  expect_equal(sum(grepl("kernel_estimate_regression_categorical_tree_np", lines, fixed = TRUE)), 0L)
})

np_eval_call_arg <- getFromNamespace(".np_eval_call_arg", "np")
np_eval_bws_call_arg <- getFromNamespace(".np_eval_bws_call_arg", "np")
explode_pipe <- getFromNamespace("explodePipe", "np")

test_that(".np_eval_call_arg prefers eval env and falls back to caller symbols", {
  call_obj <- as.call(list(as.name("f"), x = as.name("x")))
  eval_env <- new.env(parent = emptyenv())
  eval_env$x <- 11L
  environment(call_obj) <- eval_env

  caller_env <- new.env(parent = emptyenv())
  caller_env$x <- 22L
  expect_identical(np_eval_call_arg(call_obj, "x", caller_env = caller_env), 11L)

  rm("x", envir = eval_env)
  expect_identical(np_eval_call_arg(call_obj, "x", caller_env = caller_env), 22L)
})

test_that(".np_eval_bws_call_arg preserves bws fallback for missing symbols", {
  call_obj <- as.call(list(as.name("f"), txdat = as.name("txdat")))
  environment(call_obj) <- new.env(parent = emptyenv())

  bws <- list(call = call_obj, txdat = 1:4)
  expect_identical(np_eval_bws_call_arg(bws, "txdat"), 1:4)
})

test_that("call arg helpers preserve explicit NULL symbol values", {
  call_obj <- as.call(list(as.name("f"), x = as.name("x")))
  eval_env <- new.env(parent = emptyenv())
  eval_env$x <- NULL
  environment(call_obj) <- eval_env

  caller_env <- new.env(parent = emptyenv())
  caller_env$x <- 22L
  expect_null(np_eval_call_arg(call_obj, "x", caller_env = caller_env))

  eval_env$txdat <- NULL
  call_obj2 <- as.call(list(as.name("f"), txdat = as.name("txdat")))
  environment(call_obj2) <- eval_env
  bws <- list(call = call_obj2, txdat = 1:4)
  expect_null(np_eval_bws_call_arg(bws, "txdat"))
})

test_that(".np_eval_call_arg prefers explicit caller_env over ambient stack frames", {
  call_obj <- as.call(list(as.name("f"), x = as.name("x")))
  eval_env <- new.env(parent = emptyenv())
  environment(call_obj) <- eval_env

  caller_env <- new.env(parent = emptyenv())
  caller_env$x <- 22L

  out <- local({
    x <- 99L
    np_eval_call_arg(call_obj, "x", caller_env = caller_env)
  })
  expect_identical(out, 22L)
})

test_that("call arg helpers route evaluation through shared frame helper", {
  call.body <- paste(deparse(body(np_eval_call_arg), width.cutoff = 500L), collapse = " ")
  bws.body <- paste(deparse(body(np_eval_bws_call_arg), width.cutoff = 500L), collapse = " ")

  expect_match(call.body, "\\.np_try_eval_in_frames\\(expr, eval_env = eval\\.env\\)")
  expect_match(bws.body, "\\.np_try_eval_in_frames\\(expr, eval_env = eval\\.env, search_frames = FALSE\\)")
})

test_that("explodePipe resolves formula symbols from provided env", {
  env <- new.env(parent = baseenv())
  env$fml <- y ~ x | z

  expect_identical(explode_pipe(quote(fml), env = env),
                   explode_pipe(env$fml, env = env))
})

test_that("explodePipe evaluates non-symbol formulas through strict frame helper", {
  fn.body <- paste(deparse(body(explode_pipe), width.cutoff = 500L), collapse = " ")
  expect_match(fn.body, "\\.np_try_eval_in_frames\\(formula, eval_env = env, search_frames = FALSE\\)")
})

test_that("bounded conditional kernels change npcdens on bounded-support DGP", {
  library(np)
  set.seed(20260216)

  n <- 120
  x <- runif(n)
  y <- rbeta(n, shape1 = 1 + 4 * (x - 0.5)^2, shape2 = 2 - 4 * (x - 0.5)^2)
  dat <- data.frame(y = y, x = x)

  b_emp <- npcdensbw(
    y ~ x,
    data = dat,
    bwmethod = "cv.ls",
    nmulti = 1,
    cxkertype = "gaussian",
    cykertype = "gaussian",
    cykerbound = "range"
  )
  b_inf <- npcdensbw(
    y ~ x,
    data = dat,
    bwmethod = "cv.ls",
    nmulti = 1,
    cxkertype = "gaussian",
    cykertype = "gaussian",
    cykerbound = "fixed",
    cykerlb = -Inf,
    cykerub = Inf
  )

  f_emp <- npcdens(bws = b_emp)
  f_inf <- npcdens(bws = b_inf)

  grid <- seq(min(y), max(y), length.out = 120)
  nd <- data.frame(y = grid, x = rep(0.5, length(grid)))
  p_emp <- predict(f_emp, newdata = nd)
  p_inf <- predict(f_inf, newdata = nd)

  expect_true(max(abs(p_emp - p_inf)) > 1e-6)
})

library(np)

test_that("public npcdensbw cv.ls bounded LP does not reward numerically singular tiny hx", {
  set.seed(42)
  n <- 350L
  x <- data.frame(x = runif(n))
  y <- data.frame(y = rbeta(n, 1, 1))

  bw.bad <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(0.0610566549, 0.0009802178),
    bwtype = "fixed",
    bwmethod = "cv.ls",
    regtype = "lp",
    degree = 3,
    basis = "glp",
    bandwidth.compute = FALSE,
    cxkerbound = "range",
    cykerbound = "range"
  )
  bw.good <- npcdensbw(
    xdat = x,
    ydat = y,
    bws = c(1.20411811, 0.05536272),
    bwtype = "fixed",
    bwmethod = "cv.ls",
    regtype = "lp",
    degree = 3,
    basis = "glp",
    bandwidth.compute = FALSE,
    cxkerbound = "range",
    cykerbound = "range"
  )

  obj.bad <- np:::.npcdensbw_eval_only(x, y, bw.bad)$objective
  obj.good <- np:::.npcdensbw_eval_only(x, y, bw.good)$objective

  expect_true(is.finite(obj.bad))
  expect_true(is.finite(obj.good))
  expect_lte(obj.bad, obj.good + 1e-10)
})

test_that("bootstrap interval summary avoids recursive pointwise recomputation for all-type requests", {
  skip_if_not_installed("np")

  interval_summary <- getFromNamespace(".np_plot_bootstrap_interval_summary", "np")
  quantile_bounds <- getFromNamespace("compute.bootstrap.quantile.bounds", "np")

  set.seed(20260310L)
  boot.t <- matrix(rnorm(128L * 5L), nrow = 128L, ncol = 5L)
  t0 <- colMeans(boot.t) + seq_len(ncol(boot.t)) / 10
  alpha <- 0.1

  assign("np_plot_band_type_calls_test", character(), envir = .GlobalEnv)
  trace(
    "compute.bootstrap.quantile.bounds",
    where = asNamespace("np"),
    print = FALSE,
    tracer = quote({
      assign(
        "np_plot_band_type_calls_test",
        c(get("np_plot_band_type_calls_test", envir = .GlobalEnv, inherits = FALSE), band.type),
        envir = .GlobalEnv
      )
    })
  )
  on.exit({
    untrace("compute.bootstrap.quantile.bounds", where = asNamespace("np"))
    rm("np_plot_band_type_calls_test", envir = .GlobalEnv)
  }, add = TRUE)

  out <- interval_summary(
    boot.t = boot.t,
    t0 = t0,
    alpha = alpha,
    band.type = "all"
  )

  expect_identical(
    get("np_plot_band_type_calls_test", envir = .GlobalEnv, inherits = FALSE),
    c("all", "simultaneous")
  )

  pointwise <- quantile_bounds(boot.t = boot.t, alpha = alpha, band.type = "pointwise", warn.coverage = FALSE)
  bonferroni <- quantile_bounds(boot.t = boot.t, alpha = alpha, band.type = "bonferroni", warn.coverage = FALSE)
  simultaneous <- quantile_bounds(boot.t = boot.t, alpha = alpha, band.type = "simultaneous", warn.coverage = FALSE)

  expect_equal(out$err, cbind(t0 - pointwise[, 1L], pointwise[, 2L] - t0))
  expect_equal(out$all.err$pointwise, cbind(t0 - pointwise[, 1L], pointwise[, 2L] - t0))
  expect_equal(out$all.err$bonferroni, cbind(t0 - bonferroni[, 1L], bonferroni[, 2L] - t0))
  expect_equal(out$all.err$simultaneous, cbind(t0 - simultaneous[, 1L], simultaneous[, 2L] - t0))
})

test_that("pmzsd helper matches covariance-diagonal standard deviations", {
  skip_if_not_installed("np")

  boot_col_sds <- getFromNamespace(".np_plot_bootstrap_col_sds", "np")

  set.seed(20260311L)
  boot.t <- matrix(rnorm(257L * 11L), nrow = 257L, ncol = 11L)

  expect_equal(
    boot_col_sds(boot.t),
    sqrt(diag(cov(boot.t))),
    tolerance = 1e-12
  )
})

test_that("npcdensbw automatic degree search defaults to NOMAD plus Powell", {
  skip_if_not_installed("crs")

  old_opts <- options(np.messages = FALSE, np.tree = FALSE)
  on.exit(options(old_opts), add = TRUE)

  set.seed(20260319)
  dat <- data.frame(x = sort(runif(14)))
  dat$y <- dat$x + rnorm(nrow(dat), sd = 0.08)

  bw <- np::npcdensbw(
    y ~ x,
    data = dat,
    regtype = "lp",
    degree.select = "coordinate",
    degree.min = 0L,
    degree.max = 1L,
    bwtype = "fixed",
    bwmethod = "cv.ls",
    nmulti = 1L
  )

  expect_identical(bw$degree.search$mode, "nomad+powell")
  expect_true(isTRUE(bw$degree.search$completed))
  expect_gte(bw$degree.search$best.fval, bw$degree.search$baseline.fval - 1e-8)
})

test_that("smooth coefficient bootstrap target labels format compactly", {
  fmt <- getFromNamespace(".np_plot_scoef_bootstrap_target_label", "np")

  bws <- list(
    xndim = 1L,
    zndim = 1L,
    xnames = "x",
    znames = "z"
  )

  expect_identical(fmt(bws = bws, slice.index = 0L), "surf 1/2")
  expect_identical(fmt(bws = bws, slice.index = 1L), "x 1/2")
  expect_identical(fmt(bws = bws, slice.index = 2L), "z 2/2")
})

test_that("smooth coefficient helper labels carry target context for bootstrap phases", {
  fn <- getFromNamespace("compute.bootstrap.errors.scbandwidth", "np")

  captured <- new.env(parent = emptyenv())
  captured$prep <- character()
  captured$progress <- character()
  captured$interval <- character()

  orig_activity_begin <- getFromNamespace(".np_plot_activity_begin", "np")
  orig_activity_end <- getFromNamespace(".np_plot_activity_end", "np")
  orig_scoef <- getFromNamespace(".np_inid_boot_from_scoef", "np")
  orig_interval <- getFromNamespace(".np_plot_bootstrap_interval_summary", "np")

  assignInNamespace(".np_plot_activity_begin", function(label) {
    captured$prep <- c(captured$prep, label)
    list(label = label)
  }, ns = "np")
  assignInNamespace(".np_plot_activity_end", function(state) invisible(NULL), ns = "np")
  assignInNamespace(".np_inid_boot_from_scoef", function(..., progress.label = NULL) {
    captured$progress <- c(captured$progress, progress.label)
    list(t = matrix(0, nrow = 2L, ncol = 3L), t0 = c(0, 0, 0))
  }, ns = "np")
  assignInNamespace(".np_plot_bootstrap_interval_summary", function(boot.t, t0, alpha, band.type, progress.label = NULL) {
    captured$interval <- c(captured$interval, progress.label)
    list(err = matrix(0, nrow = ncol(boot.t), ncol = 2L), all.err = list())
  }, ns = "np")
  on.exit({
    assignInNamespace(".np_plot_activity_begin", orig_activity_begin, ns = "np")
    assignInNamespace(".np_plot_activity_end", orig_activity_end, ns = "np")
    assignInNamespace(".np_inid_boot_from_scoef", orig_scoef, ns = "np")
    assignInNamespace(".np_plot_bootstrap_interval_summary", orig_interval, ns = "np")
  }, add = TRUE)

  xdat <- data.frame(x = c(0, 1, 2))
  zdat <- data.frame(z = c(1, 2, 3))
  ydat <- c(0.1, 0.2, 0.3)
  bws <- list(
    xndim = 1L,
    zndim = 1L,
    xnames = "x",
    znames = "z"
  )

  out <- fn(
    xdat = xdat,
    ydat = ydat,
    zdat = zdat,
    exdat = xdat,
    ezdat = zdat,
    gradients = FALSE,
    slice.index = 2L,
    progress.target = "z 2/2",
    plot.errors.boot.method = "geom",
    plot.errors.boot.blocklen = 2L,
    plot.errors.boot.num = 2L,
    plot.errors.center = "estimate",
    plot.errors.type = "all",
    plot.errors.alpha = 0.05,
    bws = bws
  )

  expect_true(is.list(out))
  expect_true(any(grepl("Preparing plot bootstrap geom \\(z 2/2\\)", captured$prep)))
  expect_true(any(grepl("Plot bootstrap \\(z 2/2\\)", captured$progress)))
  expect_true(any(grepl("Constructing bootstrap all bands \\(z 2/2\\)", captured$interval)))
})

% $Id: jss_np.Rnw,v 1.69 2008/07/22 19:01:59 jracine Exp jracine $

%\VignetteIndexEntry{The np Package}
%\VignetteDepends{np,boot,MASS}
%\VignetteKeywords{nonparametric, kernel, econometrics, qualitative,
%categorical}
%\VignettePackage{np}

\documentclass[nojss]{jss}
%\usepackage{Sweave}

\usepackage{amsmath}
\usepackage[utf8]{inputenc}

\newcommand{\field}[1]{\mathbb{#1}}
\newcommand{\R}{\field{R}}

\author{Tristen Hayfield\\ETH Z\"urich
  \And Jeffrey S.~Racine\\McMaster University}

\title{The \pkg{np} Package}

\Plainauthor{Tristen Hayfield, Jeffrey S.~Racine}

\Plaintitle{The np Package}

%\Shorttitle{The np Package}

\Abstract{ 
  
  We describe the \proglang{R} \pkg{np} package via a series of
  applications that may be of interest to applied econometricians.
  This vignette is based on \citet{JSSv027i05}.  The
  \pkg{np} package implements a variety of nonparametric and
  semiparametric kernel-based estimators that are popular among
  econometricians. There are also procedures for nonparametric tests
  of significance and consistent model specification tests for
  parametric mean regression models and parametric quantile regression
  models, among others. The \pkg{np} package focuses on kernel methods
  appropriate for the mix of continuous, discrete, and categorical
  data often found in applied settings. Data-driven methods of
  bandwidth selection are emphasized throughout, though we caution the
  user that data-driven bandwidth selection methods can be
  computationally demanding.

}

\Keywords{nonparametric, semiparametric, kernel smoothing,
  categorical data.}

\Plainkeywords{Nonparametric, kernel, econometrics, qualitative,
  categorical}

%% \Volume{13}
%% \Issue{9}
%% \Month{September}
%% \Year{2004}
%% \Submitdate{2004-09-29}
%% \Acceptdate{2004-09-29}

\Address{Tristen Hayfield\\
  ETH H\"onggerberg Campus\\
  Physics Department\\
  CH-8093 Z\"urich, Switzerland\\
  E-mail: \email{hayfield@phys.ethz.ch}
%%  URL: \url{http://www.exp-astro.phys.ethz.ch/hayfield/}\\
  ~\\
  Jeffrey S.~Racine\\
  Department of Economics\\
  McMaster University\\
  Hamilton, Ontario, Canada, L8S 4L8\\
  E-mail: \email{racinej@mcmaster.ca}\\
  URL: \url{https://experts.mcmaster.ca/people/racinej}\\
}

\begin{document}
%\SweaveOpts{concordance=TRUE}

%% Note that you should use the \pkg{}, \proglang{} and \code{} commands.

%% Note - fragile using \label{} in \section{} - must be outside

%% For graphics

\setkeys{Gin}{width=\textwidth}

%% Achim's request for R prompt set invisibly at the beginning of the
%% paper

<<eval=TRUE,echo=FALSE,keep.source=TRUE>>=
options(prompt = "R> ", np.messages = FALSE, digits = 3, warn = -1)
@ 

%% Achim asks for making use of <<..., height= ..., width = ...>>= for
%% better aspect ratios and better pen-to-paper ratios...

%% <<fig=TRUE,eval=TRUE, height=, width=>>=

\section{Introduction}

Devotees of \proglang{R} \citep{R} are likely to be aware of a number
of nonparametric kernel\footnote{A `kernel' is simply a weighting
  function.} smoothing methods that exist in \proglang{R} base (e.g.,
\code{density}) and in certain \proglang{R} packages (e.g.,
\code{locpoly} in the \pkg{KernSmooth} package
\citep{R-KernSmooth-package}). These routines deliver nonparametric
smoothing methods to a wide audience, allowing \proglang{R} users to
nonparametrically model a density or to conduct nonparametric local
polynomial regression, by way of example.

The appeal of nonparametric methods, for applied researchers at least,
lies in their ability to reveal structure in data that might be missed
by classical parametric methods. Nonparametric kernel smoothing
methods are often, however, much more computationally demanding than
their parametric counterparts.

In applied settings we often encounter a combination of categorical
and continuous datatypes.  Those familiar with traditional
nonparametric kernel smoothing methods will appreciate that these
methods presume that the underlying data is continuous in nature,
which is frequently not the case.  One approach towards handling the
presence of both continuous and categorical data is called a
`frequency' approach, whereby data is broken up into subsets (`cells')
corresponding to the values assumed by the categorical variables, and
only then do you apply say \code{density} or \code{locpoly} to the
continuous data remaining in each cell.  Nonparametric frequency
approaches are widely acknowledged to be unsatisfactory as they often
lead to substantial efficiency losses arising from the use of sample
splitting, particularly when the number of cells is large.

Recent theoretical developments offer practitioners a variety of
kernel-based methods for categorical data only (i.e., unordered and
ordered factors), or for a mix of continuous and categorical
data. These methods have the potential to recapture the efficiency
losses associated with nonparametric frequency approaches as they do
not rely on sample splitting, rather, they smooth the categorical
variables in an appropriate manner; see \citet{LI_RACINE:2007} and the
references therein for an in-depth treatment of these methods, and see
also the references listed in the bibliography.

The \pkg{np} package implements recently developed kernel methods that
seamlessly handle the mix of continuous, unordered, and ordered factor
datatypes often found in applied settings. The package also allows the
user to create their own routines using high-level function calls
rather than writing their own \proglang{C} or \proglang{Fortran}
code.\footnote{The high-level functions found in the package in turn
  call compiled \proglang{C} code allowing the user to focus on the
  application rather than the implementation details.}  The design
philosophy underlying \pkg{np} aims to provide an intuitive, flexible,
and extensible environment for applied kernel estimation. We
appreciate that there exists tension among these goals, and have tried
to balance these competing ends, with varying degrees of success.
\pkg{np} is available from the Comprehensive \proglang{R} Archive
Network at \url{https://CRAN.R-project.org/package=np}.

Currently, a range of methods can be found in the \pkg{np} package
including unconditional (\citeauthor{LI_RACINE:2003a}
\citeyear{LI_RACINE:2003a}, \citeauthor{OUYANG_LI_RACINE:2006}
\citeyear{OUYANG_LI_RACINE:2006}) and conditional
(\citeauthor{HALL_RACINE_LI:2004} \citeyear{HALL_RACINE_LI:2004},
\citeauthor{RACINE_LI_ZHU:2004} \citeyear{RACINE_LI_ZHU:2004}) density
estimation and bandwidth selection, conditional mean and gradient
estimation (local constant \citeauthor{RACINE_LI:2004}
\citeyear{RACINE_LI:2004}, \citeauthor{HALL_LI_RACINE:2007}
\citeyear{HALL_LI_RACINE:2007} and local polynomial
\citeauthor{LI_RACINE:2004} \citeyear{LI_RACINE:2004}), conditional
quantile and gradient estimation (\citeauthor{LI_RACINE:2008}
\citeyear{LI_RACINE:2008}), model specification tests (regression
\citeauthor{HSIAO_LI_RACINE:2007} \citeyear{HSIAO_LI_RACINE:2007},
quantile), significance tests (\citeauthor{RACINE:1997}
\citeyear{RACINE:1997}, \citeauthor{RACINE_HART_LI:2006}
\citeyear{RACINE_HART_LI:2006}), semiparametric regression (partially
linear \citeauthor{ROBINSON:1988} \citeyear{ROBINSON:1988},
\citeauthor{GAO_LIU_RACINE:2012} \citeyear{GAO_LIU_RACINE:2012}, index models
\citeauthor{KLEIN_SPADY:1993} \citeyear{KLEIN_SPADY:1993},
\citeauthor{ICHIMURA:1993} \citeyear{ICHIMURA:1993}, average
derivative estimation, varying/smooth coefficient models
\citeauthor{LI_RACINE:2010} \citeyear{LI_RACINE:2010},
\citeauthor{LI_OUYANG_RACINE:2012} \citeyear{LI_OUYANG_RACINE:2012}),
among others. The various functions in the \pkg{np} package are
described in Table~\ref{np function table}.


\begin{table}[!ht]
  \centering\raggedright
  {\small
    \begin{tabular}{p{.75in}|p{3in}|p{2in}}
			
			\hline
			\textbf{Function} & \textbf{Description} & \textbf{Reference} \\ 
			\hline
			
\code{npcdens} & Nonparametric Conditional Density Estimation & \citet{HALL_RACINE_LI:2004} \\
\code{npcdensbw} &  Nonparametric Conditional Density Bandwidth Selection & \citet{HALL_RACINE_LI:2004}\\
\code{npcdist} &  Nonparametric Conditional Distribution Estimation & \citet{LI_RACINE:2008}\\
\code{npcmstest} &   Parametric Model Specification Test & \citet{HSIAO_LI_RACINE:2007}\\
\code{npconmode} &  Nonparametric Modal Regression & \\
\code{npdeneqtest} &  Nonparametric Test for Equality of Densities & \citet{LI_MAASOUMI_RACINE:2009} \\
\code{npdeptest} &  Nonparametric Entropy Test for Pairwise Dependence & \citet{MAASOUMI_RACINE:2002} \\
\code{npindex} & Semiparametric Single Index Model & \citet{ICHIMURA:1993}, \citet{KLEIN_SPADY:1993}\\
\code{npindexbw} & Semiparametric Single Index Model Parameter and Bandwidth Selection& \citet{ICHIMURA:1993}, \citet{KLEIN_SPADY:1993}\\
\code{npksum} &  Nonparametric Kernel Sums & \\
\code{npplot} & General Purpose Plotting of Nonparametric Objects& \\
\code{npplreg} & Semiparametric Partially Linear  Regression &
\citet{ROBINSON:1988}, \citet{GAO_LIU_RACINE:2012} \\
\code{npplregbw} & Semiparametric Partially Linear  Regression Bandwidth Selection & \citet{ROBINSON:1988}, \citet{GAO_LIU_RACINE:2012} \\
\code{npqcmstest} &  Parametric Quantile Regression Model Specification Test
& \citet{ZHENG:1998}, \citet{RACINE:2006}\\
\code{npqreg} & Nonparametric Quantile Regression & \citet{LI_RACINE:2008} \\
\code{npreg} &  Nonparametric Regression & \citet{RACINE_LI:2004}, \citet{LI_RACINE:2004}\\
\code{npregbw} &  Nonparametric Regression Bandwidth Selection & \citet{HURVICH_SIMONOFF_TSAI:1998}, \citet{RACINE_LI:2004}, \citet{LI_RACINE:2004} \\
\code{npscoef} & Semiparametric Smooth Coefficient  Regression& \citet{LI_RACINE:2010}\\
\code{npscoefbw} & Semiparametric Smooth Coefficient  Regression Bandwidth Selection& \citet{LI_RACINE:2010}\\
\code{npsdeptest} & Nonparametric Entropy Test for Serial Nonlinear Dependence& \citet{GRANGER_MAASOUMI_RACINE:2004}\\
\code{npsigtest} &  Nonparametric Regression Significance Test & \citet{RACINE:1997}, \citet{RACINE_HART_LI:2006}\\
\code{npsymtest} & Nonparametric Entropy Test for Asymmetry& \citet{MAASOUMI_RACINE:2009}\\
\code{npudens} &  Nonparametric Density Estimation & \citet{PARZEN:1962}, \citet{ROSENBLATT:1956}, \citet{LI_RACINE:2003a}\\
\code{npudensbw} &  Nonparametric Density Bandwidth Selection & \citet{PARZEN:1962}, \citet{ROSENBLATT:1956}, \citet{LI_RACINE:2003a}\\
\code{npudist} &  Nonparametric Distribution Estimation & \citet{PARZEN:1962}, \citet{ROSENBLATT:1956}, \citet{LI_RACINE:2003a}\\
\code{npunitest} &  Nonparametric Entropy Test for Univariate Density Equality & \citet{MAASOUMI_RACINE:2002} \\
-- Utilities --&&\\
\code{gradients} & Extract Gradients& \\
\code{se} &	Extract Standard Errors&\\
\code{uocquantile} & Compute Quantiles/Modes for Unordered, Ordered, and Numeric Data& \\
    \end{tabular}
    \label{np function table}
  }
  \caption{\textbf{\pkg{np} functions.}}
\end{table}
	
In this article, we illustrate the use of the \pkg{np} package via a
number of empirical applications. Each application is chosen to
highlight a specific econometric method in an applied setting. We
begin first with a discussion of some of the features and
implementation details of the \pkg{np} package in
Section~\ref{implementation section}. We then proceed to illustrate
the functionality of the package via a series of applications,
sometimes beginning with a classical parametric method that will
likely be familiar to the reader, and then performing the same
analysis in a semi- or nonparametric framework. It is hoped that such
comparison helps the reader quickly gauge whether or not there is any
value added by moving towards a nonparametric framework for the
application they have at hand. We commence with the workhorse of
applied data analysis (regression) in Section~\ref{regression
  section}, beginning with a simple univariate regression example and
then moving on to a multivariate example. We then proceed to
nonparametric methods for binary and multinomial outcome models in
Section~\ref{count section}. Section~\ref{pdf and cdf section}
considers nonparametric methods for unconditional probability density
function (PDF) and cumulative distribution function (CDF) estimation,
while Section~\ref{cpdf and ccdf section} considers conditional PDF
and CDF estimation, and nonparametric estimators of quantile models
are considered in Section~\ref{quantile section}. A range of
semiparametric models are then considered, including partially linear
models in Section~\ref{partially linear section}, single-index models
in Section~\ref{single index section}, and finally varying coefficient
models are considered in Section~\ref{varying coefficient section}.

\section{Important implementation details}\label{implementation section}

In this section we describe some implementation details that may help
users navigate the methods that reside in the \pkg{np} package. We
shall presume that the user is familiar with the traditional kernel
estimation of, say, density functions (e.g.,
\citet{ROSENBLATT:1956}, \citet{PARZEN:1962}) and regression
functions (e.g., \citet{NADARAYA:1965}, \citet{WATSON:1964})
when the underlying data is continuous in nature. However, we do not
presume familiarity with mixed-data kernel methods hence briefly
describe modifications to the kernel function that are necessary to
handle the mix of categorical and continuous data often encountered in
applied settings. These methods, of course, collapse to the familiar
estimators when all variables are continuous.

\subsection{The primacy of the bandwidth}

Bandwidth selection is a key aspect of sound nonparametric and
semiparametric kernel estimation. It is the direct counterpart of
model selection for parametric approaches, and should therefore not be
taken lightly. \pkg{np} is designed from the ground up to make
bandwidth selection the focus of attention. To this end, one typically
begins by creating a `bandwidth object' which embodies all aspects of
the method, including specific kernel functions, data names,
datatypes, and the like. One then passes these bandwidth objects to
other functions, and those functions can grab the specifics from the
bandwidth object thereby removing potential inconsistencies and
unnecessary repetition. For convenience these steps can be combined
should the user so choose, i.e., if the first step (bandwidth
selection) is not performed explicitly then the second step will
automatically call the omitted first step bandwidth selection using
defaults unless otherwise specified, and the bandwidth object could
then be retrieved retroactively if so desired. Note that the combined
approach would not be a wise choice for certain applications such as
when bootstrapping (as it would involve unnecessary computation since
the bandwidths would properly be those for the original sample and not
the bootstrap resamples) or when conducting quantile regression (as it
would involve unnecessary computation when different quantiles are
computed from the same conditional cumulative distribution estimate).

Work flow therefore typically proceeds as follows:

\begin{enumerate}

\item compute data-driven bandwidths;

\item using the bandwidth object, proceed to estimate a model and
  extract fitted or predicted values, standard errors, etc.;

\item optionally, plot the object.

\end{enumerate}

In order to streamline the creation of a set of complicated graphics
objects, \code{plot} (which calls \code{npplot}) is dynamic; i.e., you
can specify, say, bootstrapped error bounds and the appropriate
routines will be called in real time. Be aware, however, that
bootstrap methods can be computationally demanding hence some plots
may not appear immediately in the graphics window.

\subsection{Data-driven bandwidth selection methods}

We caution the reader that data-driven bandwidth selection methods can
be computationally demanding.  We ought to also point out that
data-driven (i.e., automatic) bandwidth selection procedures are not
guaranteed always to produce good results due to perhaps the presence
of outliers or the rounding/discretization of continuous data, among
others. For this reason, we advise the reader to interrogate their
bandwidth objects with the \code{summary} command which produces a
table of bandwidths for the continuous variables along with a constant
multiple of $\sigma_x n^\alpha$, where $\sigma_x$ is a variable's
standard deviation, $n$ the number of observations, and $\alpha$ a
known constant that depends on the method, kernel order, and number of
continuous variables involved, e.g., $\alpha=-1/5$ for univariate
density estimation with one continuous variable and a second order
kernel. Seasoned practitioners can immediately assess whether
undersmoothing or oversmoothing may be present by examining these
constants, as the appropriate constant (called the `scaling factor')
that is multiplied by $\sigma_x n^\alpha$ often ranges from between
0.5 to 1.5 for some though not all methods, and it is this constant
that is computed and reported by \code{summary}. Also, the admissible
range for the bandwidths for the categorical variables is provided
when \code{summary} is used, which some readers may also find helpful.

We caution users to use multistarting for any serious application
(multistarting refers to restarting numerical search methods from
different initial values to avoid the presence of local minima - the
default is the minimum of the number of variables or 5 and can be
changed via the argument \code{nmulti =}), and do not recommend
overriding default search tolerances (unless increasing \code{nmulti =}
beyond its default value).

We direct the interested reader to the frequently asked questions
document 
(\url{https://CRAN.R-project.org/package=np/vignettes/np_faq.pdf}) for a range of
potentially helpful tips and suggestions surrounding bandwidth
selection and the \pkg{np} package.

\subsection[Interacting with np functions]{Interacting with \pkg{np} functions}

A few words about the \proglang{R} \code{data.frame} construct are in
order. Data frames are fundamental objects in \proglang{R}, defined as
``tightly coupled collections of variables which share many of the
properties of matrices and of lists, used as the fundamental data
structure by most of R's modeling software.'' A data frame is ``a
matrix-like structure whose columns may be of differing types
(numeric, logical, factor and character and so on).'' Seasoned
\proglang{R} users would, prior to estimation or inference, transform
a given data frame into one with appropriately classed elements (the
\pkg{np} package contains a number of datasets whose variables have
already been classed appropriately). It will be seen that appropriate
classing of variables is crucial in order for functions in the
\pkg{np} package to {\em automatically} use appropriate weighting
functions which differ according to a variable's class. If your data
frame contains variables that have not been classed appropriately, you
can do this `on the fly' by re-classing the variable upon invocation
of an \pkg{np} function, however, it is preferable to begin with a
data frame having appropriately classed elements.

There are two ways in which you can interact with functions in
\pkg{np}, namely using data frames or using a formula interface, where
appropriate.  Every function in \pkg{np} supports both interfaces,
where appropriate.

To some, it may be natural to use the data frame interface.  If you
find this most natural for your project, you first create a data frame
casting data according to their type (i.e., one of continuous
(default), \code{factor}, \code{ordered}), as in
\begin{verbatim} 
R> data.object <- data.frame(x1 = factor(x1), x2, x3 = ordered(x3)) 
\end{verbatim} 
where $x_1$ is, say, a binary \code{factor}, $x_2$ continuous, and
$x_3$ an \code{ordered} \code{factor}. Then you could pass this data
frame to the appropriate \pkg{np} function, say
\begin{verbatim} 
R> bw <- npudensbw(dat = data.object)
\end{verbatim}

To others, however, it may be natural to use the formula interface
that is used for the regression example outlined below. For
nonparametric regression functions such as \code{npregbw}, you would
proceed as you might using \code{lm}, e.g.,
\begin{verbatim}
R> bw <- npregbw(y ~ x1 + x2)
\end{verbatim}
except that you would of course not need to specify, e.g., polynomials
in variables, interaction terms, or create a number of dummy variables
for a factor. Of course, if the variables in your data frame have not
been classed appropriately, then you would need to explicitly cast,
say, \code{x1} via \code{npregbw(y ~ factor(x1) + x2)} above.

A few words on the formula interface are in order.  We use the
standard formula interface as it provides capabilities for handling
missing observations and so forth.  This interface, however, is simply
a convenient device for telling a routine which variable is, say, the
outcome and which are, say, the covariates.  That is, just because one
writes $x_1+x_2$ in no way means or is meant to imply that the model
will be linear and additive (why use fully nonparametric methods to
estimate such models in the first place?). It simply means that there
are, say, two covariates in the model, the first being $x_1$ and the
second $x_2$, we are passing them to a routine with the formula
interface, and nothing more is presumed nor implied. This will likely
be obvious to most \proglang{R} users, but we point it out simply to
avoid any potential confusion for those unfamiliar with kernel
smoothing methods.

\subsection{Writing your own functions}

We have tried to make \pkg{np} flexible enough to be of use to a wide
range of users.  All options can be tweaked by the user (kernel
function, kernel order, bandwidth type, estimator type and so forth).
One function, \code{npksum}, allows you to create your own estimators,
tests, etc.  The function \code{npksum} is simply a call to
specialized \proglang{C} code, so you get the benefits of compiled
code along with the power and flexibility of the \proglang{R}
language. We hope that incorporating the \code{npksum} function
renders the package suitable for teaching and research alike.

\subsection{Generalized product kernels}\label{generalized kernel
  section}

As noted above, traditional nonparametric kernel methods presume that
the underlying data is continuous in nature, which is frequently not
the case.  The basic idea underlying the treatment of kernel methods
in the presence of a mix of categorical and continuous data lies in
the use of what we call `generalized product kernels', which we
briefly summarize.

Suppose that you are interested in kernel estimation for `unordered'
categorical data, i.e., you are presented with discrete data $X^d\in
{\mathcal S}^d$, where ${\mathcal S}^d$ denotes the support of $X^d$.
We use $x_s^d$ and $X_{is}^d$ to denote the $s$th component of $x^d$
and $X_i^d$ ($i=1,\dots,n$), respectively.  Following
\citet{AITCHISON_AITKEN:1976}, for $x_s^d$, $X^d_{is}\in \{0,1,\dots,
c_s-1\}$, we define a discrete univariate kernel function
\begin{equation}
  \label{mixed_smooth:eq:l(.)}
  l^u(X^d_{is},x^d_s,\lambda_s) = \left\{ \begin{array}{ll}
      1 - \lambda_s &\text{ if } X^d_{is} = x^d_s \\
      \lambda_s/(c_s-1) &\text{ if } X^d_{is} \neq x^d_s.
    \end{array}
  \right.
\end{equation}

Note that when $\lambda_s = 0$, $l(X^d_{is},x^d_s,0)={\bf
  1}(X_{is}^d=x^d_s)$ becomes an indicator function, and if $\lambda_s
= (c_s - 1)/c_s$, then $l\left(X^d_{is},x^d_s,
  \frac{c_s-1}{c_s}\right) = 1/c_s$ is a constant for {\it all} values
of $X_{is}^d$ and $x^d_s$.  The range of $\lambda_s$ is $[0,
(c_s-1)/c_s]$. Observe that these weights sum to 1.

If, instead, some of the discrete variables are ordered (i.e., are
`ordinal categorical variables'), then we should use a kernel function
which is capable of reflecting their ordered status. Assuming that
$x^d_s$ can take on $c_s$ different ordered values, $\{0,1,\dots
,c_s-1\}$, \citet[p.~29]{AITCHISON_AITKEN:1976} suggested using the
kernel function given by
\begin{equation}
l^o(x^d_{s},v^d_{s},\lambda_s) =
\begin{pmatrix}c_s\cr j\end{pmatrix}
\lambda_s^{j}(1-\lambda_s)^{c_s-j}\text{ when }|x^d_{s}-v^d_s|=j\qquad (0\leq
s\leq c_s),
\end{equation}
 where
\begin{equation}
\begin{pmatrix}
c_s\cr j
\end{pmatrix}
=\frac{c_s!}{j!(c_s-j)!}.
\end{equation}
Observe that these weights sum to 1.

If instead a variable was continuous, you could use, say, the second
order Gaussian kernel, namely
\begin{equation}
w(x^c,X^c_i,h)=\frac{1}{\sqrt{2\pi}}\exp\left\{-\frac{1}{2}\left(\frac{X^c_i-x^c}{h}\right)^2\right\}.
\end{equation}
Observe that these weights integrate to 1.

The `generalized product kernel function' for a vector of, say, one
unordered, one ordered, and one continuous variable is simply the
product of $l^u(\cdot)$, $l^o(\cdot)$, and $w(\cdot)$, and
multivariate versions with differing numbers of each datatype are
created in the same fashion. We naturally allow the bandwidths to
differ for each variable. For what follows, for a vector of mixed
variables, we define $K_{\gamma}(x,X_i)$ to be this product, where
$\gamma$ is the vector of bandwidths (e.g., the $h$s and $\lambda$s)
and $x$ the vector of mixed datatypes.  For further details see
\citet{LI_RACINE:2003a} who proposed the use of these generalized
product kernels for unconditional density estimation and developed the
underlying theory for a data-driven method of bandwidth selection for
this class of estimators.  The use of such kernels offers a seamless
framework for kernel methods with mixed data. For further details on a
range of kernel methods that employ this approach, we direct the
interested reader to \citet[Chapter~4]{LI_RACINE:2007}.

If the user wishes to apply kernel functions other than those provided
by the default settings, the kernel function can be changed by passing
the appropriate arguments to \code{ukertype}, \code{okertype}, and
\code{ckertype} (the first letter of each representing unordered,
ordered, and continuous, respectively), while the `order' for the
continuous kernel (i.e., the first non-zero moment) can be changed by
passing the appropriate (even) integer to \code{ckerorder}. We support
a variety of unordered, ordered, and continuous kernels along with a
variety of high-order kernels, the default being \code{ckerorder =
  2}. Using these arguments, the user could select an eighth order
Epanechnikov kernel, a fourth order Gaussian kernel, or even a uniform
kernel for the continuous variables in their application. See
\code{?np} for further details.

\section{Nonparametric regression}\label{regression section}

We shall start with the workhorse of applied data analysis, namely,
regression models. In order to introduce nonparametric regression, we
shall first consider the simplest possible regression model, one that
involves one continuous dependent variable, $y$, and one continuous
explanatory variable, $x$.  We shall begin with a popular parametric
model of a wage equation, and then move on to a fully nonparametric
regression model. Having compared and contrasted the parametric and
nonparametric approach towards univariate regression, we then proceed
to multivariate regression.

\subsection{Univariate regression}

We begin with a classic dataset taken from
\citet[p.~155]{PAGAN_ULLAH:1999} who consider Canadian cross-section
wage data consisting of a random sample taken from the 1971 Canadian
Census Public Use Tapes for male individuals having common education
(Grade 13). There are $n=205$ observations in total, and 2 variables,
the logarithm of the individual's wage (\code{logwage}) and their age
(\code{age}).  The traditional wage equation is typically modelled as
a quadratic in \code{age}.

First, we begin with a simple parametric model for this relationship.

<<eval=TRUE,keep.source=TRUE>>=
library("np")
data("cps71")
model.par <- lm(logwage ~ age + I(age^2), data = cps71)
summary(model.par)
@ 

If we first consider the model's goodness of fit, the model has an
unadjusted $R^2$ of \Sexpr{summary(model.par)$r.squared}\%.  Next, we
consider the local linear nonparametric method proposed by
\citet{LI_RACINE:2004} employing cross-validated bandwidth selection
using the method of \citet{HURVICH_SIMONOFF_TSAI:1998}. Note that in
this example we conduct bandwidth selection and estimation in one
step.


<<eval=TRUE,keep.source=TRUE>>=
model.np <- npreg(logwage ~ age, 
                  regtype = "ll", 
                  bwmethod = "cv.aic",
                  gradients = TRUE,
                  data = cps71)
summary(model.np)
@ 

Using the measure of goodness of fit introduced in the next section,
we see that this method produces a better in-sample model, at least as
measured by the $R^2$ criterion, having an $R^2$ of
\Sexpr{model.np$R2}\%.\footnote{See Section~\ref{r-squared section}
  for details on computation of the nonparametric $R^2$ measure.}

So far we have summarized the model's goodness-of-fit. However,
econometricians also routinely report the results from tests of
significance. There exist nonparametric counterparts to these tests
that were proposed by \citet{RACINE:1997}, and extended to admit
categorical variables by \citet{RACINE_HART_LI:2006}, which we conduct
below.

<<eval=TRUE,keep.source=TRUE>>=
npsigtest(model.np)
@ 

Having conducted this test we observe that, as was the case for the
linear parametric model, the explanatory variable \code{age} is
significant at all conventional levels in the local linear
nonparametric model.

\subsubsection{Assessing goodness of fit for nonparametric
  models}\label{r-squared section}

The reader may have wondered what the difference is between the $R^2$
measures reported by the linear and nonparametric models summarized
above, or perhaps how the $R^2$ was generated for the nonparametric
model.  It is desirable to use a unit-free measure of goodness-of-fit
for nonparametric regression models which is comparable to that used
for parametric regression models, namely $R^2$. Note that this will be
a {\em within-sample} measure of goodness-of-fit. Given the known
drawbacks of computing $R^2$ based on the decomposition of the sum of
squares (such as possible negative values), there is an alternative
definition and method for computing $R^2$ which can be used that is
directly applicable to any model, linear or nonlinear.  Letting $y_i$
denote the outcome and $\hat y_i$ the fitted value for observation
$i$, we may define $R^2$ as follows:
\begin{equation*}
R^2=\frac{\left[\sum_{i=1}^n(y_i-\bar y)(\hat y_i-\bar y)\right]^2}
{\sum_{i=1}^n(y_i-\bar y)^2\sum_{i=1}^n(\hat y_i-\bar y)^2}
\end{equation*}
and this measure will {\em always} lie in the range $[0,1]$ with the
value 1 denoting a perfect fit to the sample data and 0 denoting no
predictive power above that given by the unconditional mean of the
outcome. It can be demonstrated that this method of computing $R^2$ is
identical to the standard measure computed as $\sum_{i=1}^n(\hat
y_i-\bar y)^2/\sum_{i=1}^n(y_i-\bar y)^2$ {\em when the model is
  linear, fitted with least squares, and includes an intercept
  term}. This useful measure will permit direct comparison of
within-sample goodness-of-fit subject to the obvious qualification
that this is by no means a model selection criterion, rather, simply a
summary measure that some may wish to report. This measure can, of
course, also be computed using out-of-sample predictions and
out-of-sample realizations.  Were we to consider models estimated on a
randomly selected subset of data and evaluated on an independent
sample of hold-out data, this measure computed for the hold-out
observations might serve to guide model selection, particularly when
averaged over a number of independent hold-out datasets.

\subsubsection{Graphical comparison of parametric and nonparametric
  models}

Often, a suitable graphical comparison of models allows the user to
immediately appreciate features present in both the data and the
estimated models.\footnote{By suitable, we mean those that display
  both point estimates and variability estimates.}  The upper left
plot in Figure~\ref{age profile} presents the fitted parametric and
nonparametric regression models along with the data for the
\code{cps71} example. The following code can be used to construct the
plots appearing in Figure~\ref{age profile}.

<<eval=FALSE, keep.source=TRUE>>=
plot(cps71$age, cps71$logwage, xlab = "age", ylab = "log(wage)", cex=.1)
lines(cps71$age, fitted(model.np), lty = 1, col = "blue")
lines(cps71$age, fitted(model.par), lty = 2, col = " red")
plot(model.np, plot.errors.method = "asymptotic")

plot(model.np, gradients = TRUE)
lines(cps71$age, coef(model.par)[2]+2*cps71$age*coef(model.par)[3],
      lty = 2,
      col = "red")
plot(model.np, gradients = TRUE, plot.errors.method = "asymptotic")
@ 

\begin{figure}[!ht]
\begin{center}
<<fig=TRUE, png=TRUE, resolution=100, multifig=TRUE, eval=TRUE, echo=FALSE, keep.source=TRUE, height=8, width=8>>=
options(SweaveHooks = list(multifig = function() par(mfrow=c(2,2),mar=c(4,4,3,2))))

# Plot 1

plot(cps71$age,cps71$logwage,xlab="age",ylab="log(wage)",cex=.1)
lines(cps71$age,fitted(model.np),lty=1,col="blue")
lines(cps71$age,fitted(model.par),lty=2,col="red")

# Plot 2

plot(cps71$age,gradients(model.np),xlab="age",ylab="gradient",type="l",lty=1,col="blue")
lines(cps71$age,coef(model.par)[2]+2*cps71$age*coef(model.par)[3],lty=2,col="red")

# Plot 3

plot(cps71$age,fitted(model.np),xlab="age",ylab="log(wage)",ylim=c(min(fitted(model.np)-2*model.np$merr),max(fitted(model.np)+2*model.np$merr)),type="l")
lines(cps71$age,fitted(model.np)+2*model.np$merr,lty=2,col="red")
lines(cps71$age,fitted(model.np)-2*model.np$merr,lty=2,col="red")

# Plot 4

plot(cps71$age,gradients(model.np),xlab="age",ylab="gradient",ylim=c(min(gradients(model.np)-2*model.np$gerr),max(gradients(model.np)+2*model.np$gerr)),type="l",lty=1,col="blue")
lines(cps71$age,gradients(model.np)+2*model.np$gerr,lty=2,col="red")
lines(cps71$age,gradients(model.np)-2*model.np$gerr,lty=2,col="red")

@ 
\caption{\label{age profile}The figure on the upper left presents the
  parametric (quadratic, dashed line) and the nonparametric estimates
  (solid line) of the regression function for the \code{cps71} data. The
  figure on the upper right presents the parametric (dashed line) and
  nonparametric estimates (solid line) of the gradient. The figures on
  the lower left and lower right present the nonparametric estimates
  of the regression function and gradient along with their variability
  bounds, respectively.}
\end{center}
\end{figure}

The upper right plot in Figure~\ref{age profile} compares the
gradients for the parametric and nonparametric models. Note that the
gradient for the parametric model will be given by $\partial
\text{logwage}_i/\partial\text{age}_i=\partial\hat\beta_2+2\hat\beta_3\text{age}_i$,
$i=1,\dots,n$, as the model is quadratic in age.

Of course, it might be preferable to also plot error bars for the
estimates, either asymptotic or resampled. We have automated this in
the function \code{npplot} which is automatically called by
\code{plot}. The lower left and lower right plots in Figure~\ref{age
  profile} present pointwise error bars using asymptotic standard
error formulas for the regression function and gradient, respectively.

Often, however, distribution free (bootstrapped) error bounds may be
desired, and we allow the user to readily do so as we have written
\pkg{np} to leverage the \pkg{boot} package
(\citeauthor{R-boot-package} \citeyear{R-boot-package}). By default
(`\code{iid}'), bootstrap resampling is conducted pairwise on
$(y,X,Z)$ (i.e., by resampling from rows of the $(y,X)$ data or
$(y,X,Z)$ data where appropriate). Specifying the type of
bootstrapping as `\code{inid}' admits general heteroskedasticity of
unknown form via the wild bootstrap \citep{LIU:1988}, though it does
not allow for dependence. `\code{fixed}' conducts the block bootstrap
\citep{KUNSCH:1989} for dependent data, while `\code{geom}' conducts
the stationary bootstrap \citep{POLITIS_ROMANO:1994}.

\subsubsection{Generating predictions from nonparametric models}

Once you have obtained appropriate bandwidths and estimated a
nonparametric model, generating predictions is straightforward
involving nothing more than creating a set of explanatory variables
for which you wish to generate predictions. These can lie inside the
support of the original data or outside should the user so choose. We
have written our routines to support the \code{predict} function in R,
so by using the \code{newdata =} option one can readily generate
predictions. It is important to note that typically you do not have
the outcome for the evaluation data, hence you need only provide the
explanatory variables.  However, if by chance you do have the outcome
and you provide it, the routine will compute the out-of-sample summary
measures of predictive ability. This would be useful when one splits a
dataset into two independent samples, estimates a model on one sample,
and wishes to assess its performance on the independent hold-out data.

By way of example we consider the \code{cps71} data, generate a set of
values for \code{age}, two of which lie outside of the support of the
data (10 and 70 years of age), and generate the parametric and
nonparametric predictions using the generic \code{predict} function.

<<eval=TRUE,keep.source=TRUE>>=
cps.eval <- data.frame(age = seq(10,70, by=10))
predict(model.par, newdata = cps.eval)
predict(model.np, newdata = cps.eval)
@ 

Note that if you provide \code{predict} with the argument
\code{se.fit = TRUE}, it will also return pointwise asymptotic standard
errors where appropriate.

\subsection{Multivariate regression with qualitative and quantitative
  data}\label{wage1 section}

Based on the presumption that some readers will be unfamiliar with the
kernel smoothing of qualitative data, we next consider a multivariate
regression example that highlights the potential benefits arising from
the use of kernel smoothing methods that smooth both the qualitative
and quantitative variables in a particular manner. For what follows,
we consider an application taken from \citet[p.~226]{WOOLDRIDGE:2003}
that involves multiple regression analysis with qualitative
information.

We consider modeling an hourly wage equation for which the dependent
variable is $\log$(\code{wage}) (\code{lwage}) while the explanatory
variables include three continuous variables, namely \code{educ}
(years of education), \code{exper} (the number of years of potential
experience), and \code{tenure} (the number of years with their current
employer) along with two qualitative variables, \code{female}
(`\code{Female}'/`\code{Male}') and \code{married}
(`\code{Married}'/`\code{Notmarried}').  For this example there are
$n=526$ observations.

The classical parametric approach towards estimating such
relationships requires that one first specify the functional form of
the underlying relationship.  We start by first modelling this
relationship using a simple parametric linear model. By way of
example, \citet[p.~222]{WOOLDRIDGE:2003} presents the following
model:\footnote{We would like to thank Jeffrey Wooldridge for allowing
  us to incorporate his data in the \pkg{np} package. Also, we would
  like to point out that Wooldridge starts out with this na\"ive
  linear model, but quickly moves on to a more realistic model
  involving nonlinearities in the continuous variables and so
  forth. The purpose of this example is simply to demonstrate how
  nonparametric models can outperform misspecified parametric models
  in multivariate finite-sample settings.}
<<eval=TRUE,keep.source=TRUE>>=
data("wage1")
model.ols <- lm(lwage ~ female +
                married +
                educ +
                exper +
                tenure,
                data = wage1)
summary(model.ols)
@

This model is, however, restrictive in a number of ways. First, the
analyst must specify the functional form (in this case linear) for the
continuous variables (\code{educ}, \code{exper}, and \code{tenure}).
Second, the analyst must specify how the qualitative variables
(\code{female} and \code{married}) enter the model (in this case they
affect the model's intercepts only). Third, the analyst must specify
the nature of any interactions among all variables, quantitative and
qualitative (in this case, there are none).  Should any of these
presumptions be incorrect, then the estimated model will be biased and
inconsistent potentially leading to faulty inference.

One might next test the null hypothesis that this parametric linear
model is correctly specified using the consistent model specification
test described in \citet{HSIAO_LI_RACINE:2007} that admits both
categorical and continuous data (we have overridden the default search
tolerances for this example as the objective function is well-behaved
via \code{tol = 0.1} and \code{ftol = 0.1}, which should {\em not} be done
in general - this example will likely take a few minutes on a desktop
computer as it uses bootstrapping and cross-validated bandwidth
selection):
<<eval=TRUE,keep.source=TRUE>>=
model.ols <- lm(lwage ~ female +
                married +
                educ +
                exper +
                tenure,
                x = TRUE,
                y = TRUE,
                data = wage1)
X <- data.frame(wage1$female,
                wage1$married,
                wage1$educ,
                wage1$exper,
                wage1$tenure)
output <- npcmstest(model = model.ols, 
                    xdat = X, 
                    ydat = wage1$lwage, 
                    nmulti = 1,
                    tol = 0.1, 
                    ftol = 0.1)
summary(output)
@

Note that it might appear at first blush that the user needs to do
some redundant typing as the $X$ data in this example is the same as
the regressors in the model. However, in general $X$ could differ
which is why the user must specify this object separately as we cannot
assume that the explanatory variables in the model will be equal to
$X$.

This na\"ive linear model is rejected by the data (the $p$-value for
the null of correct specification is $<0.001$), hence one might
proceed instead to model this relationship using kernel methods.

As noted, the traditional nonparametric approach towards modeling
relationships in the presence of qualitative variables requires that
you first split your data into subsets containing only the continuous
variables of interest (\code{lwage}, \code{exper}, and \code{tenure}).
For instance, we would have four such subsets, a) $n=132$ observations
for married females, b) $n=120$ observations for single females, c)
$n=86$ observations for single males, and d) $n=188$ observations for
married males. One would then construct smooth nonparametric
regression models for each of these subsets and proceed with analysis
in this fashion. However, this may lead to a loss in efficiency due to
a reduction in the sample size leading to overly variable estimates of
the underlying relationship.

Instead, however, we could construct smooth nonparametric regression
models by i) using a kernel function that is appropriate for the
qualitative variables as outlined in Section~\ref{generalized kernel
  section} and ii) modifying the nonparametric regression model as was
done by \citet{LI_RACINE:2004}. One can then conduct sound
nonparametric estimation based on the $n=526$ observations rather than
resorting to sample splitting. The rationale for this lies in the fact
that doing so may introduce potential bias, however it will always
reduce variability thus leading to potential finite-sample efficiency
gains. Our experience has been that the potential benefits arising
from this approach more than offset the potential costs in
finite-sample settings.

Next, we consider using the local-linear nonparametric method
described in \citet{LI_RACINE:2004}. For the reader's convenience we
supply precomputed cross-validated bandwidths which are automatically
loaded when one loads the \code{wage1} dataset (recall being cautioned
about the computational burden associated with multivariate
data-driven bandwidth methods).

% The commented out code that generates the cross-validated bandwidths
% is also provided should the reader wish to fully replicate the
% example

In the example that follows, we use the precomputed bandwidth object
\code{bw.all} that contains the data-driven bandwidths for the local
linear regression produced below using all observations in the sample.

<<eval=TRUE,keep.source=TRUE>>=
#bw.all <- npregbw(formula = lwage ~ female +
#                  married +
#                  educ +
#                  exper +
#                  tenure,
#                  regtype = "ll",
#                  bwmethod = "cv.aic",
#                  data = wage1)
model.np <- npreg(bws = bw.all)
summary(model.np)
@

Note again that the bandwidth object is the only thing you need to
pass to \code{npreg} as it encapsulates the kernel types, regression
method, and so forth. You could also use \code{npreg} and manually
specify the bandwidths using a bandwidth vector if you so
choose.\footnote{For example, attach a dataset via \code{data(cps71)}
  then \code{attach(cps71)} then enter, say, \code{plot(age,logwage)}
  and \code{lines(age,fitted(npreg(logwage~age,bws=2)))} to plot the
  local constant estimator with a bandwidth of 2 years).} We have
tried to make each function as flexible as possible to meet the needs
of a variety of users.

The goodness of fit of the nonparametric model ($R^2=51.5\%$) is
better than that for the parametric model ($R^2=40.4\%$).  In order to
investigate whether this apparent improvement reflects overfitting or
simply that the nonparametric model is in fact more faithful to the
underlying data generating process, we shuffled the data and created
two independent samples, one of size $n_1=400$ and one of size
$n_2=126$. We fit the models on the $n_1$ training observations then
evaluate the models on the $n_2$ independent hold-out observations
using the predicted square error criterion, namely
$n_2^{-1}\sum_{i=1}^{n_2}(y_i-\hat y_i)^2$, where the $y_i$s are the
\code{lwage} values for the hold-out observations and the $\hat y_i$s
are the predicted values. Finally, we compare the parametric model,
the nonparametric model that smooths both the categorical and
continuous variables, and the traditional frequency nonparametric
model that breaks the data into subsets and smooths the continuous
data only. For this example we use the precomputed bandwidth object
\code{bw.subset} which contains the data-driven bandwidths for a
random subset of data.

<<eval=TRUE,keep.source=TRUE>>=
set.seed(123)
ii <- sample(seq(1, nrow(wage1)), replace=FALSE)
wage1.train <- wage1[ii[1:400],]
wage1.eval <- wage1[ii[401:nrow(wage1)],]
model.ols <- lm(lwage ~ female +
                married +
                educ +
                exper +
                tenure,
                data = wage1.train)
fit.ols <- predict(model.ols,
                   data = wage1.train,
                   newdata = wage1.eval)
pse.ols <- mean((wage1.eval$lwage - fit.ols)^2)
#bw.subset <- npregbw(formula = lwage ~ female +
#                     married +
#                     educ +
#                     exper +
#                     tenure,
#                     regtype = "ll",
#                     bwmethod = "cv.aic",
#                     data = wage1.train)
model.np <- npreg(bws = bw.subset)
fit.np <- predict(model.np,
                  data = wage1.train,
                  newdata = wage1.eval)
pse.np <- mean((wage1.eval$lwage - fit.np)^2)
bw.freq <- bw.subset
bw.freq$bw[1] <- 0
bw.freq$bw[2] <- 0
model.np.freq <- npreg(bws = bw.freq)
fit.np.freq <- predict(model.np.freq,
                       data = wage1.train,
                       newdata = wage1.eval)
pse.np.freq <- mean((wage1.eval$lwage - fit.np.freq)^2)
@

The predicted square error on the hold-out data was
\Sexpr{formatC(pse.ols,format="f",digits=3)} for the parametric linear
model, \Sexpr{formatC(pse.np.freq,format="f",digits=3)} for the
traditional nonparametric estimator that splits the data into subsets,
and \Sexpr{formatC(pse.np,format="f",digits=3)} for the nonparametric
estimator that uses the full sample but smooths both the qualitative
and quantitative data. The nonparametric model that smooths both the
quantitative and qualitative data is more than
\Sexpr{formatC(((pse.ols-pse.np)/pse.ols)*100,format="f",digits=1)}\%
more efficient in terms of out-of-sample predictive ability than the
parametric or frequency nonparametric model, and therefore appears to
provide a better description of the underlying data generating process
than either.

Note that for this example we have only four cells. If one used all
qualitative and binary variables included in the dataset (sixteen in
total), one would have $65,536$ cells, many of which would be empty,
and most having far too few observations to provide meaningful
nonparametric estimates.  As the number of qualitative variables
increases, the difference between the estimator that smooths both
continuous and discrete variables in a particular manner and the
traditional estimator that relies upon sample splitting will become
even more pronounced.

Next, we display partial regression plots in Figure~\ref{wage1
  profile}.\footnote{A `partial regression plot' is simply a 2D plot
  of the outcome $y$ versus one covariate $x_j$ when all other
  covariates are held constant at their respective medians/modes (this
  can be changed by the user).} We also plot bootstrapped variability
bounds, where the bootstrapping is done via the \pkg{boot} package
thereby facilitating a variety of bootstrap methods. The following
code will generate Figure~\ref{wage1 profile}.

<<eval=FALSE, keep.source=TRUE>>=
plot(model.np,
     plot.errors.method = "bootstrap",
     plot.errors.boot.num = 25)
@

\begin{figure}[!ht]
\begin{center}
<<fig=TRUE,eval=TRUE,echo=FALSE,keep.source=TRUE>>=
plot(model.np,
     plot.errors.method = "bootstrap",
     plot.errors.boot.num = 25)
@
\caption{\label{wage1 profile}Partial local linear nonparametric
  regression plots with bootstrapped pointwise error bounds for the
  \code{wage1} dataset.}
\end{center}
\end{figure}

Figure~\ref{wage1 profile} reveals that (holding other regressors
constant at their median/mode), males have higher expected wages than
females, there does not appear to be a significant difference between
the expected wages of married and nonmarried individuals, wages
increase with education and tenure and first rise then fall as
experience increases, other things equal.

\section{Nonparametric binary outcome and count data
  models}\label{count section}

For what follows, we adopt the conditional probability estimator
proposed in \citet{HALL_RACINE_LI:2004} to estimate a nonparametric
model of a binary outcome when there exist a number of categorical
covariates.

For this example, we use the \code{birthwt} data taken from the
\pkg{MASS} package, and compute a parametric Logit model and a
nonparametric conditional mode model. We then compare their confusion
matrices\footnote{A `confusion matrix' is simply a tabulation of the
  actual outcomes versus those predicted by a model. The diagonal
  elements contain correctly predicted outcomes while the off-diagonal
  ones contain incorrectly predicted (confused) outcomes.} and assess
their classification ability. Note that many of the variables in this
dataset have not been classed so we must do this upon invocation of a
function. The outcome is an indicator of low infant birthweight (0/1).
The method can handle unordered and ordered multinomial outcomes
without modification (we have overridden the default search tolerances
for this example as the objective function is well-behaved via
\code{tol = 0.1} and \code{ftol = 0.1}, which should {\em not} be done in
general).  In this example we first model this relationship using a
simple parametric Logit model, then we model this with a nonparametric
conditional density estimator and compute the conditional mode, and
finally, we compare confusion matrices from the logit and
nonparametric models. Note that for the nonparametric estimator we
conduct bandwidth selection and estimation in one step, and we first
convert the data frame to one with appropriately classed elements.

<<eval=TRUE,keep.source=TRUE>>=
data("birthwt", package = "MASS")
birthwt$low <- factor(birthwt$low)
birthwt$smoke <- factor(birthwt$smoke)
birthwt$race <- factor(birthwt$race)
birthwt$ht <- factor(birthwt$ht)
birthwt$ui <- factor(birthwt$ui)
birthwt$ftv <- factor(birthwt$ftv)
model.logit <- glm(low ~ smoke +
                   race +
                   ht +
                   ui +
                   ftv +
                   age +
                   lwt, 
                   family = binomial(link = logit),
                   data = birthwt)
model.np <- npconmode(low ~
                      smoke +
                      race +
                      ht +
                      ui +
                      ftv +
                      age +
                      lwt,
                      tol = 0.1,
                      ftol = 0.1,
                      data = birthwt)
cm <- table(birthwt$low, 
            ifelse(fitted(model.logit) > 0.5, 1, 0))
cm
summary(model.np)
@

For this example the nonparametric model is better able to predict low
birthweight infants than its parametric counterpart, correctly
predicting
\Sexpr{sum(diag(model.np$confusion.matrix))}/\Sexpr{sum(model.np$confusion.matrix)}
birthweights compared with \Sexpr{sum(diag(cm))}/\Sexpr{sum(cm)} for
the parametric model.

\section{Nonparametric unconditional PDF and CDF estimation}\label{pdf
  and cdf section}

The Old Faithful Geyser is a tourist attraction located in Yellowstone
National Park. This famous dataset containing $n=272$ observations
consists of two variables, eruption duration in minutes
(\code{eruptions}) and waiting time until the next eruption in minutes
(\code{waiting}). This dataset is used by the park service to model,
among other things, expected duration conditional upon the amount of
time that has elapsed since the previous eruption. Modeling the joint
distribution is, however, of interest in its own right, and the
underlying bimodal nature of the joint PDF and CDF is readily revealed
by the kernel estimator.  For this example, we load the old faithful
geyser data and compute the density and distribution
functions. Results are presented in Figure~\ref{faithful plot}. Note
that in this example we conduct bandwidth selection and estimation in
one step.

<<eval=TRUE,keep.source=TRUE>>=
data("faithful", package = "datasets")
f.faithful <- npudens(~ eruptions + waiting, data = faithful)
F.faithful <- npudist(~ eruptions + waiting, data = faithful)
summary(f.faithful)
summary(F.faithful)
@

The following code will generate Figure~\ref{faithful plot}.

<<eval=FALSE, keep.source=TRUE>>=
plot(f.faithful, xtrim = -0.2, view = "fixed", main = "")
plot(F.faithful, xtrim = -0.2, view = "fixed", main = "")
@

\begin{figure}[!ht]
\begin{center}
<<fig=TRUE, png=TRUE, resolution=100, multifig=TRUE, eval=TRUE, echo=FALSE, keep.source=TRUE, height=4, width=8>>=
options(SweaveHooks = list(multifig = function() par(mfrow=c(1,2),mar=rep(1.5,4))))
plot.output <- plot(f.faithful, xtrim=-0.2, view="fixed", main = "",plot.behavior="data")
# Retrieve data from plot() to create multiple figures
f <- matrix(plot.output$d1$dens, 50, 50)
plot.x1 <- unique(plot.output$d1$eval[,1])
plot.x2 <- unique(plot.output$d1$eval[,2])
persp(plot.x1,plot.x2,f,xlab="eruptions",ylab="waiting",zlab="Joint Density",col="lightblue", ticktype="detailed")

plot.output <- plot(F.faithful, xtrim = -0.2, view = "fixed", main="",plot.behavior="data")
# Retrieve data from plot() to create multiple figures
F <- matrix(plot.output$d1$dist, 50, 50)
plot.x1 <- unique(plot.output$d1$eval[,1])
plot.x2 <- unique(plot.output$d1$eval[,2])
persp(plot.x1,plot.x2,F,xlab="eruptions",ylab="waiting",zlab="Joint Distribution",col="lightblue", ticktype="detailed")
@
\caption{\label{faithful plot}Nonparametric multivariate PDF and CDF
  estimates for the Old Faithful data.}
\end{center}
\end{figure}

If one were to instead model this density with a parametric model such
as the bivariate normal (being symmetric, unimodal, and monotonically
decreasing away from the mode), one would of course fail to uncover
the underlying structure readily revealed by the kernel estimate.

\section{Nonparametric conditional PDF and CDF estimation}\label{cpdf
  and ccdf section}

We consider Giovanni Baiocchi's (\citeauthor{BAIOCCHI:2006}
\citeyear{BAIOCCHI:2006}) Italian GDP growth panel for 21 regions
covering the period 1951-1998 (millions of Lire, 1990=base).  There
are $n=1,008$ observations in total, and two variables, \code{gdp} and
\code{year}.  First, we compute the bandwidths. Note that this may
take a minute or two depending on the speed of your computer. We
override the default tolerances for the search method as the objective
function is well-behaved (do not of course do this in general), then
we compute the \code{npcdens} object. Note that in this
example we conduct bandwidth selection and estimation in one step.

<<eval=TRUE,keep.source=TRUE>>=
data("Italy")
fhat <- npcdens(gdp ~ year,
                tol = 0.1,
                ftol = 0.1,
                data = Italy)
summary(fhat)
Fhat <- npcdist(gdp ~ year,
                tol = 0.1,
                ftol = 0.1,
                data = Italy)
summary(Fhat)
@ 

Figure~\ref{italy conditional plot} plots the resulting conditional
PDF and CDF for the \code{Italy} GDP panel. The following code will
generate Figure~\ref{italy conditional plot}.

<<eval=FALSE, keep.source=TRUE>>=
plot(fhat, view = "fixed", main = "", theta = 300, phi = 50)
plot(Fhat, view = "fixed", main = "", theta = 300, phi = 50)
@

\begin{figure}[!ht]
\begin{center}
<<fig=TRUE, png=TRUE, resolution=100, multifig=TRUE, eval=TRUE, echo=FALSE, keep.source=TRUE, height=4, width=8>>=
options(SweaveHooks = list(multifig = function() par(mfrow=c(1,2),mar=rep(1.25,4))))
plot.output <- plot(fhat, view="fixed", main="",plot.behavior="data")
# Retrieve data from plot() to create multiple figures
f <- matrix(plot.output$cd1$condens, 48, 50)
plot.y1 <- unique(plot.output$cd1$yeval)
plot.x1 <- unique(plot.output$cd1$xeval)
persp(as.integer(levels(plot.x1)),plot.y1,f,xlab="year",ylab="gdp",zlab="Conditional Density",col="lightblue", ticktype="detailed",theta=300,phi=50)

plot.output <- plot(Fhat, view="fixed", main="",plot.behavior="data")
# Retrieve data from plot() to create multiple figures
F <- matrix(plot.output$cd1$condist, 48, 50)
plot.y1 <- unique(plot.output$cd1$yeval)
plot.x1 <- unique(plot.output$cd1$xeval)
persp(as.numeric(plot.x1)+1951,plot.y1,F,xlab="year",ylab="gdp",zlab="Conditional Distribution",col="lightblue", ticktype="detailed",theta=300,phi=50)

@
\caption{\label{italy conditional plot}Nonparametric conditional PDF
  and CDF estimates for the Italian GDP panel.}
\end{center}
\end{figure}

Figure~\ref{italy conditional plot} reveals that the distribution of
income has evolved from a unimodal one in the early 1950s to a
markedly bimodal one in the 1990s. This result is robust to bandwidth
choice, and is observed whether using simple rules-of-thumb or
data-driven methods such as likelihood cross-validation.  The kernel
method readily reveals this evolution which might easily be missed
were one to use parametric models of the income distribution (e.g.,
the unimodal lognormal distribution is commonly used to model income
distributions).

\section{Nonparametric quantile regression}\label{quantile section}

We again consider Giovanni Baiocchi's Italian GDP growth panel.
First, we compute the likelihood cross-validation bandwidths
(default). We override the default tolerances for the search method as
the objective function is well-behaved (do not of course do this in
general). Then we compute the resulting conditional quantile estimates
using the method of \citet{LI_RACINE:2008}. By way of example, we
compute the 25th, 50th, and 75th conditional quantiles. Note that this
may take a minute or two depending on the speed of your computer.
Note that for this example we first call \code{npcdistbw} to avoid
unnecessary re-computation of the bandwidth object.

<<eval=TRUE,keep.source=TRUE>>=
bw <- npcdistbw(formula = gdp ~ year,
                tol = 0.1,
                ftol = 0.1,
                data = Italy)
model.q0.25 <- npqreg(bws = bw, tau = 0.25)
model.q0.50 <- npqreg(bws = bw, tau = 0.50)
model.q0.75 <- npqreg(bws = bw, tau = 0.75)
@ 

Figure~\ref{italy quantile plot} plots the resulting quantile
estimates. The following code will generate Figure~\ref{italy quantile
  plot}.

<<eval=FALSE,keep.source=TRUE>>=
plot(Italy$year, Italy$gdp, main = "", 
     xlab = "Year", ylab = "GDP Quantiles")
lines(Italy$year, model.q0.25$quantile, col = "red", lty = 1, lwd = 2)
lines(Italy$year, model.q0.50$quantile, col = "blue", lty = 2, lwd = 2)
lines(Italy$year, model.q0.75$quantile, col = "red", lty = 3, lwd = 2)
legend(ordered(1951), 32, c("tau = 0.25", "tau = 0.50", "tau = 0.75"), 
       lty = c(1, 2, 3), col = c("red", "blue", "red"))
@

%% Reset sizing to admit smaller graphs

\setkeys{Gin}{width=0.75\textwidth}

\begin{figure}[!ht]
\begin{center}
<<fig=TRUE,eval=TRUE,echo=FALSE,keep.source=TRUE,width=6,height=5>>=
plot(Italy$year, Italy$gdp, 
     main = "",
     xlab = "Year", 
     ylab = "GDP Quantiles")
lines(Italy$year, model.q0.25$quantile, col = "red", lty = 1, lwd = 2)
lines(Italy$year, model.q0.50$quantile, col = "blue", lty = 2, lwd = 2)
lines(Italy$year, model.q0.75$quantile, col = "red", lty = 3, lwd = 2)
legend(ordered(1951), 32, c("tau = 0.25", "tau = 0.50", "tau = 0.75"), 
       lty = c(1, 2, 3), col = c("red", "blue", "red"))
@
\caption{\label{italy quantile plot}Nonparametric quantile regression
  on the Italian GDP panel.}
\end{center}
\end{figure}

One nice feature of this application is that the explanatory variable
is ordered and there exist multiple observations per year. Using the
\code{plot} function with ordered data produces a boxplot which
readily reveals the non-smooth 25th, 50th, and 75th quantiles. These
non-smooth quantile estimates can then be directly compared to those
obtained via direct estimation of the smooth CDF which are plotted in
Figure~\ref{italy quantile plot}.

\section{Semiparametric partially linear models}\label{partially
  linear section}

Suppose that we consider the \code{wage1} dataset from
\citet[p.~222]{WOOLDRIDGE:2003}, but now assume that the researcher
is unwilling to presume the nature of the relationship between
\code{exper} and \code{lwage}, hence relegates \code{exper} to the
nonparametric part of a semiparametric partially linear model.  The
partially linear model was proposed by \citet{ROBINSON:1988} and
extended to handle the presence of categorical covariates by
\citet{GAO_LIU_RACINE:2012}.

Before proceeding, we ought to clarify a common misunderstanding about
partially linear models. Many believe that, as the model is apparently
simple, its computation ought to also be simple. However, the apparent
simplicity hides the perhaps under-appreciated fact that bandwidth
selection for partially linear models can be orders of magnitude more
computationally burdensome than that for fully nonparametric models,
for one simple reason. Data-driven bandwidth selection methods such as
cross-validation are being used, and the partially linear model
involves cross-validation to regress $y$ on $Z$ ($Z$ is multivariate)
then {\em each column of $X$} on $Z$, whereas fully nonparametric
regression involves cross-validation of $y$ on $X$ only. The
computational burden associated with partially linear models is
therefore much more demanding than for nonparametric models, so be
forewarned. Note that in this example we conduct bandwidth selection
and estimation in one step.

<<eval=TRUE,keep.source=TRUE>>=
model.pl <- npplreg(lwage ~ female +
                    married +
                    educ +
                    tenure | exper,
                    data = wage1)
summary(model.pl)
@ 

A comparison of this model with the parametric and nonparametric
models presented in Section~\ref{wage1 section} indicates an in-sample
fit ($44.9\%$) that lies in between the misspecified parametric model
($40.4\%$) and the fully nonparametric model ($51.5\%$).

\section{Semiparametric single-index models}\label{single index section}

\subsection{Binary outcomes (Klein-Spady with cross-validation)}

We could again consider the \code{birthwt} data taken from the
\pkg{MASS} package, and this time compute a semiparametric index
model. We then compare confusion matrices and assess classification
ability. The outcome is an indicator of low infant birthweight
(0/1). We apply the method of \citet{KLEIN_SPADY:1993} with bandwidths
selected via cross-validation. Note that for this example we conduct
bandwidth selection and estimation in one step

<<eval=TRUE,keep.source=TRUE>>=
model.index <- npindex(low ~
                       smoke +
                       race +
                       ht +
                       ui +
                       ftv +
                       age +
                       lwt,
                       method = "kleinspady",
                       gradients = TRUE,
                       data = birthwt)
summary(model.index)
@ 

It is interesting to compare this with the parametric Logit model's
confusion matrix presented in Section~\ref{count section}.  A
comparison of this model with the parametric model presented in
Section~\ref{count section} reveals that it correctly classifies an
additional 5/189 observations.

\subsection{Continuous outcomes (Ichimura with cross-validation)}

Next, we consider applying \citet{ICHIMURA:1993}'s single-index method
which is appropriate for continuous outcomes, unlike that of
\citet{KLEIN_SPADY:1993} (we override the default number of
multistarts for the user's convenience as the global minimum appears
to have been located in the first attempt). We again consider the
\code{wage1} dataset found in \citet[p.~222]{WOOLDRIDGE:2003}.  Note
that in this example we conduct bandwidth selection and estimation in
one step.

<<eval=TRUE,keep.source=TRUE>>=
model <- npindex(lwage ~ female +
                 married +
                 educ +
                 exper +
                 tenure,
                 data = wage1,
                 nmulti = 1)
summary(model)
@ 

It is interesting to compare this model with the parametric and
nonparametric models presented in Section~\ref{wage1 section} as it
provides an in-sample fit ($43.1\%$) that lies in between the
misspecified parametric model ($40.4\%$) and fully nonparametric model
($51.5\%$). Whether this model yields improved out-of-sample
predictions could also be explored.

\section{Semiparametric varying coefficient models}\label{varying
  coefficient section}

We revisit the \code{wage1} dataset found in
\citet[p.~222]{WOOLDRIDGE:2003}, but assume that the researcher
believes that the model's parameters may differ depending on one's
sex. In this case, one might adopt a varying coefficient approach such
as that found in \citet{LI_RACINE:2010} and
\citet{LI_OUYANG_RACINE:2012}. We compare a simple linear model with
the semiparametric varying coefficient model.  Note that the $X$ data
in the varying coefficient model must be of type \code{numeric}, so we
create a 0/1 dummy variable from the qualitative variable for $X$, but
of course for the nonparametric component we can simply treat these as
unordered factors.  Note that we do bandwidth selection and estimation
in one step.

<<eval=TRUE,keep.source=TRUE>>=
model.ols <- lm(lwage ~ female +
                married +
                educ +
                exper +
                tenure,
                data = wage1)
wage1.augmented <- wage1
wage1.augmented$dfemale <- as.integer(wage1$female == "Male")
wage1.augmented$dmarried <- as.integer(wage1$married == "Notmarried")
model.scoef <- npscoef(lwage ~ dfemale +
                       dmarried +
                       educ +
                       exper +
                       tenure | female,
                       betas = TRUE,
                       data = wage1.augmented)
summary(model.scoef)
colMeans(coef(model.scoef))
coef(model.ols)
@

It is again interesting to compare this model with the parametric and
nonparametric models presented in Section~\ref{wage1 section}. It can be
seen that the {\em average} values of the model's coefficients are in
agreement with those from the linear specification, while the
in-sample goodness of fit ($42.6\%$) lies in between the misspecified
parametric model ($40.4\%$) and fully nonparametric model ($51.5\%$).

\section{Writing your own kernel-based functions}

The function \code{npksum} exists so that you can create your own
kernel objects with or without a variable to be weighted (defaults to
$1$). With the options available, you could create new nonparametric
tests or even new kernel estimators. The convolution kernel option
would allow you to create, say, the least squares cross-validation
function for kernel density estimation.  \code{npksum} implements a
variety of methods for computing multivariate kernel sums
($p$-variate) defined over a set of possibly continuous and/or
discrete data.

By way of example, we construct a local constant kernel estimator with
a bandwidth of, say, two years. Figure~\ref{lc npksum profile} plots
the resulting estimate.

<<eval=TRUE,keep.source=TRUE>>=
fit.lc <- npksum(txdat = cps71$age, tydat = cps71$logwage, bws = 2)$ksum/
  npksum(txdat = cps71$age, bws = 2)$ksum
@ 

The following code will generate Figure~\ref{lc npksum profile}.

<<eval=FALSE, keep.source=TRUE>>=
plot(cps71$age, cps71$logwage, xlab = "Age", ylab = "log(wage)")
lines(cps71$age, fit.lc, col = "blue")
@ 

\begin{figure}[!ht]
\begin{center}
<<fig=TRUE,eval=TRUE,echo=FALSE,keep.source=TRUE>>=
plot(cps71$age, cps71$logwage, xlab = "Age", ylab = "log(wage)", cex=.1)
lines(cps71$age, fit.lc, col = "blue")
@ 
\caption{\label{lc npksum profile}A local constant kernel estimator
  generated with \code{npksum} for the \code{cps71} dataset.}
\end{center}
\end{figure}

\code{npksum} is exceedingly flexible, allowing for leave-one-out
sums, weighted sums of matrices, raising the kernel function to
different powers, the use of convolution kernels, and so forth. See
\code{?npksum} for further details.

\section{A parallel implementation}

Data-driven bandwidth selection methods are, by their nature,
computationally burdensome. However, many bandwidth selection methods
lend themselves well to parallel computing approaches. High
performance computing resources are becoming widely available, and
multiple CPU desktop systems have become the norm and will continue to
make inroads.

When users have a large amount of data, serial bandwidth selection
procedures can be infeasible. For this reason, we have developed an
\proglang{MPI}-aware version of the \pkg{np} package that uses some of
the functionality of the \pkg{Rmpi} package which we have tentatively
called the \pkg{npRmpi} package and is available from the authors upon
request (the Message Passing Interface (MPI) is an open library
specification for parallel computation available for a range of
computing platforms).  The functionality of \pkg{np} and \pkg{npRmpi}
is identical, however, using \pkg{npRmpi} you could take advantage of
a cluster computing environment or a multi-core/multi-CPU desktop
machine thereby alleviating the computational burden associated with
the nonparametric analysis of large datasets. Installation of this
package, however, requires knowledge that goes beyond that which even
seasoned \proglang{R} users may possess.  Having access to local
expertise would be necessary for many users therefore this package is
not available via CRAN.

\section{Summary}

The \pkg{np} package offers users of \proglang{R} a variety of
nonparametric and semiparametric kernel-based methods that are capable
of handling the mix of categorical and continuous data typically
encountered by applied researchers. In this article we have described
the functionality of the \pkg{np} package via a series of illustrative
applications that may be of interest to applied econometricians
interested in becoming familiar with these methods. We do not delve
into details of the underlying estimators, rather we provide
references where appropriate and direct the interested reader to those
resources.

The help files accompanying many functions found in the \pkg{np}
package contain numerous examples which may be of interest to some
readers, and we encourage readers to experiment with these examples in
addition to those contained herein.

Finally, we encourage readers who have successfully implemented new
kernel-based methods using the \code{npksum} function to send such
functions to us so that they can be included in future versions of the
\pkg{np} package with appropriate acknowledgment of course.

\section*{Acknowledgments}

We are indebted to Achim Zeileis and to two anonymous referees whose
comments and feedback have resulted in numerous improvements to the
\pkg{np} package and to the exposition of this article. Racine would
like to gratefully acknowledge support from the Natural Sciences and
Engineering Research Council of Canada (\url{https://nserc-crsng.canada.ca/}),
the Social Sciences and Humanities Research Council of Canada
(\url{https://sshrc-crsh.canada.ca/}), and the Shared Hierarchical Academic
Research Computing Network (\url{https://www.sharcnet.ca/}).

\bibliography{np}

\end{document}

%% $Id: entropy_np.Rnw,v 1.43 2010/02/18 14:43:37 jracine Exp jracine $

%\VignetteIndexEntry{Entropy-based Inference Using the np Package}
%\VignetteDepends{np,boot,cubature,MASS}
%\VignetteKeywords{nonparametric, kernel, entropy, econometrics, qualitative,
%categorical}
%\VignettePackage{np}

\documentclass[nojss]{jss}

%% need no \usepackage{Sweave.sty}

\usepackage{amsmath,amsfonts}
\usepackage[utf8]{inputenc}

\newcommand{\field}[1]{\mathbb{#1}}
\newcommand{\N}{\field{N}}
\newcommand{\bbR}{\field{R}} %% Blackboard R
\newcommand{\bbS}{\field{S}} %% Blackboard S

\author{Jeffrey S.~Racine\\McMaster University}

\title{Entropy-Based Inference using \proglang{R} and the \pkg{np}
  Package: A Primer}

\Plainauthor{Jeffrey S.~Racine}

\Plaintitle{Entropy-Based Inference using R and the np Package: A
  Primer}

%\Shorttitle{The np Package}

\Abstract{ 
  
  We describe new functionality in the \proglang{R} (\citet{R})
  package \pkg{np} (\citet{np}) by providing a brief overview of
  each approach and its implementation followed by brief illustrative
  examples. A simple demonstration outlining how practitioners can
  implement their own entropy functions is also provided.

}

\Keywords{nonparametric, semiparametric, kernel smoothing,
  categorical data.}

\Plainkeywords{Nonparametric, kernel, econometrics, qualitative,
  categorical}

\Address{Jeffrey S.~Racine\\
  Department of Economics\\
  McMaster University\\
  Hamilton, Ontario, Canada, L8S 4L8\\
  E-mail: \email{racinej@mcmaster.ca}\\
  URL: \url{https://experts.mcmaster.ca/people/racinej}\\
}

\begin{document}

%% Note that you should use the \pkg{}, \proglang{} and \code{} commands.

%% Note - fragile using \label{} in \section{} - must be outside

%% For graphics

\setkeys{Gin}{width=\textwidth}

%% Achim's request for R prompt set invisibly at the beginning of the
%% paper

<<eval=TRUE,echo=FALSE,results=hide,keep.source=TRUE>>=
library(np)
options(prompt = "R> ", np.messages = FALSE, digits = 3, warn = -1)
@ 

%% Achim asks for making use of <<..., height= ..., width = ...>>= for
%% better aspect ratios and better pen-to-paper ratios...

\section{Overview}

In versions 0.30-4 through 0.30-7 of the \pkg{np} package
(\citet{np}) we introduced a range of new functions that may be
of interest to those conducting applied research. The five new
functions are summarized in Table \ref{np function table} below.

\begin{table}[!ht]
  \centering\raggedright
  {\small
    \begin{tabular}{p{.75in}|p{2.75in}|p{2in}}
			\hline
			\textbf{Function} & \textbf{Description} & \textbf{Reference} \\ 
			\hline
\code{npdeneqtest} &  Nonparametric Test for Equality of Densities & \citet{LI_MAASOUMI_RACINE:2009} \\
\code{npdeptest} &  Nonparametric Entropy Test for Pairwise Dependence & \citet{MAASOUMI_RACINE:2002} \\
\code{npsdeptest} & Nonparametric Entropy Test for Serial Nonlinear Dependence& \citet{GRANGER_MAASOUMI_RACINE:2004}\\
\code{npsymtest} & Nonparametric Entropy Test for Asymmetry&\citet{RACINE_MAASOUMI:2007}, \citet{MAASOUMI_RACINE:2009}\\
\code{npunitest} &  Nonparametric Entropy Test for Univariate Density Equality & \citet{MAASOUMI_RACINE:2002}\\
\hline
    \end{tabular}
    \label{np function table}
  }
  \caption{New \pkg{np} functions introduced in versions 0.30-4 through 0.30-7.}
\end{table}

In what follows we briefly describe each new function listed in Table
\ref{np function table} and provide illustrative examples of their
use. Be aware that many of these functions rely on numerical
integration and can be computationally demanding. Though we provide
moment-based versions of the most computationally demanding functions,
we advocate the use of the integral-based versions which are the
default. In the last section of this overview we describe in detail
how the user can implement their own custom entropy functions with
minimal effort using the \pkg{np} package.

\section{Testing Equality of Multivariate Densities}

\citet{LI_MAASOUMI_RACINE:2009} proposed a nonparametric test for
equality of multivariate densities comprised of continuous and
categorical data. This test can be accessed via the \code{npdeneqtest}
function.

Let $X$ and $Y$ be multivariate vectors of dimension $q+r$ where $q$
denotes the number of continuous variables and $r$ the number of
discrete/categorical variables.  A test statistic can be constructed
based on the integrated squared density difference given by $I=\int [
f(x) - g(x)]^2 dx = \int [ f(x) dF(x) + g(x) dG(x) - f(x) dG(x) - g(x)
dF(x)]$, where $F(\cdot)$ and $G(\cdot)$ are the cumulative
distribution functions for $X$ and $Y$, respectively, and where $\int
dx = \sum_{x^d\in \bbS^d} \int dx^c$.  Replacing the first occurrences
of $f(\cdot)$ and $g(\cdot)$ by their leave-one-out kernel estimates,
and replacing $F(\cdot)$ and $G(\cdot)$ by their empirical
distribution functions, we obtain the following test statistic,
\begin{equation*}
  I_n = \frac{1}{n_1(n_1-1)} \sum_{i=1}^{n_1} \sum_{j\neq i}^{n_1}
  K_{\gamma,x_i,x_j} + \frac{1}{n_2(n_2-1)} \sum_{i=1}^{n_2}
  \sum_{j\neq i}^{n_2} K_{\gamma,y_i,y_j}
  - \frac{2}{n_1 n_2} \sum_{i=1}^{n_1} \sum_{j =
    i}^{n_2} K_{\gamma, x_i,y_j}.
\end{equation*}
	
\citet{LI_MAASOUMI_RACINE:2009} demonstrate that, under the null
of equality,
\begin{equation*}
  T_n = (n_1n_2 h_1\dots h_q)^{1/2}I_n/\sigma_n \to N(0,1)
  \mbox{ in
    distribution},
\end{equation*}
where
\begin{align*}
  \sigma^2_n = 2(n_1n_2 h_1\dots h_q) \left[
  \frac{1}{n_1^2(n_1-1)^2} \sum_{i=1}^{n_1} \sum_{j \neq i}^{n_1}
  (K_{\gamma,x_i,x_j})^2\right. & + \frac{1}{n_2^2(n_2-1)^2}
  \sum_{i=1}^{n_2}\sum_{j\neq i}^{n_2} (K_{\gamma,y_i,y_j})^2\cr
 &\left.  + \frac{2}{n_1^2 n_2^2} \sum_{i=1}^{n_1}\sum_{j= 1}^{n_2}
  (K_{\gamma,x_i,y_j})^2\right],
\end{align*}
which is a consistent estimator of $\sigma^2_{0} = 2[\delta^{-1} +
\delta + 2][E[f(X_i)]][\int W^2(v)dv]$, the asymptotic variance of
$(n_1n_2 h_1\dots h_q)^{1/2}I_n$, where $\delta =
\lim_{\min\{n_1,n_2\}\to \infty} (n_1/n_2)$.  Under the alternative
the statistic diverges to $+\infty$, so the test is one-sided
rejecting when the test statistic is sufficiently large. 

The test that uses critical values taken from the asymptotic
distribution displays finite-sample size distortions, so the
\code{npdeneqtest} function employs bootstrap resampling to obtain the
finite-sample distribution of the statistic (this provides a test
having correct size). The bootstrap resamples are obtained by
resampling from the empirical distribution of the pooled data
(i.e.~are drawn from a common multivariate distribution under the
null). Bandwidths are obtained via likelihood cross-validation by
default.

The following code snippet provides two simple examples using a mix of
continuous and discrete data. Note that the two samples must be data
frames with identically named variables (e.g., the variables `wages'
and `experience' must be common to both data frames while sample A
could be that for females, sample B for males).

<<keep.source=TRUE>>=
set.seed(1234)
n <- 250
## Distributions are equal

sample.A <- data.frame(a=rnorm(n),b=factor(rbinom(n,2,.5)))
sample.B <- data.frame(a=rnorm(n),b=factor(rbinom(n,2,.5)))

npdeneqtest(sample.A,sample.B,boot.num=99)

## Distributions are unequal

sample.A <- data.frame(a=rnorm(n),b=factor(rbinom(n,2,.5)))
sample.B <- data.frame(a=rnorm(n,sd=10),b=factor(rbinom(n,2,.25)))

npdeneqtest(sample.A,sample.B,boot.num=99)
@ 
  
\section{Testing Equality of Univariate Densities}

\citet{MAASOUMI_RACINE:2002} consider a metric entropy useful for
testing for equality of densities for two univariate random variables
$X$ and $Y$. The function \code{npunitest} computes the nonparametric
metric entropy (normalized Hellinger of
\citet{GRANGER_MAASOUMI_RACINE:2004}) for testing the null of
equality of two univariate density (or probability) functions.  For
continuous variables we construct
\begin{align*}
S_{\rho}&=\frac{1}{2}\int\left(f_{1}^{1/2}-f_{2}^{1/2}\right)^{2}\,dx\cr
&=\frac{1}{2}\int\left(1-\frac{f_{2}^{1/2}}{f_1^{1/2}}\right)^2 dF_{1}(x),
\end{align*}
where $f_1=f(x)$ and $f_2=f(y)$ are the marginal densities of the
random variables $X$ and $Y$. The second expression is in a moment
form which is often replaced with a sample average, especially for
theoretical developments. When $X$ and $Y$ are discrete/categorical,
we replace integration with the sum over all possible outcomes. The
unknown density/probability functions are replaced with nonparametric
kernel estimates.

The bootstrap is conducted by resampling with replacement from the
pooled empirical distribution of $X$ and $Y$ ($X$ only for the moment
version). Default bandwidths are of the plug-in variety (‘bw.SJ’ for
continuous variables and direct plug-in for discrete variables).

The following code snippet provides three simple examples for both
continuous and discrete data.

<<keep.source=TRUE>>=
set.seed(1234)
n <- 1000
     
## Compute the statistic only, different distributions
     
x <- rchisq(n,df=10)
y <- rnorm(n,sd=10000,mean=-10000)
     
npunitest(x,y,bootstrap=FALSE)

## Data drawn from same continuous distribution

x <- rnorm(n)
y <- rnorm(n)
npunitest(x,y,boot.num=99)

## Data drawn from different continuous distributions having the
## same mean and variance

x <- rchisq(n,df=5)
y <- rnorm(n,mean=5,sd=sqrt(10))
npunitest(x,y,boot.num=99)

## Data drawn from different discrete distributions
     
x <- factor(rbinom(n,2,.5))
y <- factor(rbinom(n,2,.1))
npunitest(x,y,boot.num=99)
@ 

\section{Testing Univariate Asymmetry}

Consider a (strictly) stationary series $\{Y_{t}\}_{t=1}^{T}$. Let
$\mu_{y}$ denote a measure of central tendency, say
$\mu_{y}=E[Y_{t}]$, let $f(y)$ denote the density function of the
random variable $Y_{t}$, let $\tilde Y_{t}=-Y_{t}+2\mu_{y}$ denote a
rotation of $Y_{t}$ about its mean, and let $f(\tilde y)$ denote the
density function of the random variable $\tilde Y_{t}$. Note that if
$\mu_{y}=0$ then $\tilde Y_{t}=-Y_{t}$, though in general this will
not be so.

We say a series is \emph{symmetric about the mean} (median, mode) if
$f(y)\equiv f(\tilde y)$ almost surely. Tests for asymmetry about the
mean therefore naturally involve testing the following null:
\begin{equation*}
  H_{0}:f(y)=f(\tilde y)\mbox{ almost everywhere (a.e.)}\nonumber
\end{equation*}
against the alternative:
\begin{equation*}
  H_{1}:f(y)\ne f(\tilde y)\mbox{ on a set with positive measure}.\nonumber
\end{equation*}

The function \code{npsymtest} computes the nonparametric metric
entropy (normalized Hellinger of \citet{GRANGER_MAASOUMI_RACINE:2004})
outlined in \citet{MAASOUMI_RACINE:2009} for testing the null of
symmetry using the densities/probabilities of the data and the rotated
data, $f(y)$ and $f(\tilde y)$, respectively, and in
\citet{RACINE_MAASOUMI:2007}. $Y$ must be univariate and can be a time
series, continuous, or even categorical valued so long as the outcomes
are not character strings.

For bootstrapping the null distribution of the statistic, ‘iid’
conducts simple random resampling, while ‘geom’ conducts stationary
bootstrapping using automatic block length selection via the ‘b.star’
function in the ‘np’ package (\citet{POLITIS_ROMANO:1994},
\citet{POLITIS_WHITE:2004},
\citet{PATTON_POLITIS_WHITE:2009}). Bootstrapping is conducted by
resampling from the empirical distribution of the pooled data and
rotated data. Default bandwidths are of the plug-in variety (‘bw.SJ’
for continuous variables and direct plug-in for discrete variables).

For continuous variables we use
\begin{align*}
S_{\rho}&=\frac{1}{2}\int\left(f_{1}^{1/2}-f_{2}^{1/2}\right)^{2}\,dx\cr
&=\frac{1}{2}\int\left(1-\frac{f_{2}^{1/2}}{f_1^{1/2}}\right)^2 dF_{1}(x),
\end{align*}
where $f_1$ and $f_2$ are the marginal densities of the data and
rotated data, respectively. The second expression is in a moment form
which is often replaced with a sample average, especially for
theoretical developments. When $Y$ is discrete/categorical, we replace
integration with the sum over all possible outcomes.

The following code snippet provides two simple examples for both
continuous and discrete data.

<<keep.source=TRUE>>=
set.seed(1234)
     
n <- 100

## Asymmetric discrete probability distribution

x <- factor(rbinom(n,2,.8))
npsymtest(x,boot.num=99)

## Symmetric continuous distribution
     
y <- rnorm(n)
npsymtest(y,boot.num=99)
@ 

\section{Testing Nonlinear Pairwise Independence}

\citet{MAASOUMI_RACINE:2002} consider a metric entropy useful for
testing for pairwise independence of two random variables $X$ and
$Y$. The function \code{npdeptest} computes the nonparametric metric
entropy (normalized Hellinger of
\citet{GRANGER_MAASOUMI_RACINE:2004}) for testing the null of
pairwise independence of two univariate density (or probability)
functions.  For continuous variables we construct
\begin{align*}
S_{\rho}&=\frac{1}{2}\int_{-\infty}^{\infty}\int_{-\infty}^{\infty}\left(f_{1}^{1/2}-f_{2}^{1/2}\right)^{2}\,dx\,dy\cr
&=\frac{1}{2}\int\int\left(1-\frac{f_{2}^{1/2}}{f_1^{1/2}}\right)^2 dF_{1}(x,y),
\end{align*}
where $f_{1}=f(x_i,y_i)$ is the joint density and $f_{2}=g(x_i)\times
h(y_i)$ is the product of the marginal densities of the random
variables $X_i$ and $Y_i$. The unknown density/probability functions
are replaced with nonparametric kernel estimates.

The bootstrap distribution is obtained by resampling with replacement
from the empirical distribution of $X$ delivering $\{X_i,Y_i\}$ pairs
under the null generated as $\{X_i^*,Y_i\}$ where $X^*$ is the
bootstrap resample (i.e.~we `shuffle' $X$ leaving $Y$ unchanged
thereby breaking any pairwise dependence to generate resamples under
the null). Bandwidths are obtained via likelihood cross-validation by
default for the marginal and joint densities.

Examples include, (a) a measure/test of ``fit'', for in-sample values
of a variable $y$ and its fitted values, $\hat y$, and (b) a measure
of ``predictability'' for a variable $y$ and its predicted values
$\hat y$ (from a user implemented model).

The following code snippet provides a simple example using the actual
and fitted values from a regression model. Note that we strongly
advocate the use of the integration (default) version of the statistic
in applied settings but use the summation (i.e.~moment) version below
purely by way of demonstration as it is computationally faster.

<<keep.source=TRUE>>=
set.seed(123)
## Test/measure lack of fit between y and its fitted value from a
## regression model when x is relevant.
n <- 100
x <- rnorm(n)
y <- 1 + x + rnorm(n)
model <- lm(y~x)
y.fit <- fitted(model)
npdeptest(y,y.fit,boot.num=99,method="summation")
@ 

\section{Testing Nonlinear Serial Independence}

\citet{GRANGER_MAASOUMI_RACINE:2004} consider a metric entropy
useful for testing for nonlinear serial independence in a univariate
random variable $Y$. The function \code{npsdeptest} computes the
nonparametric metric entropy (normalized Hellinger) for testing the
null of nonlinear serial independence of such a series. We construct
\begin{align*}
S_{\rho}&=\frac{1}{2}\int_{-\infty}^{\infty}\int_{-\infty}^{\infty}\left(f_{1}^{1/2}-f_{2}^{1/2}\right)^{2}\,dx\,dy\cr
&=\frac{1}{2}\int\int\left(1-\frac{f_{2}^{1/2}}{f_1^{1/2}}\right)^2 dF_{1}(x,y),
\end{align*}
where $f_{1}=f(y_t,y_{t-k})$ is the joint density and
$f_{2}=g(y_t)\times h(y_{t-k})$ is the product of the marginal
densities of the random variables $Y_t$ and $Y_{t-k}$. The second
expression is in a moment form which is often replaced with a sample
average, especially for theoretical developments.  The unknown
density/probability functions are replaced with nonparametric kernel
estimates.

The bootstrap distribution is obtained by resampling with replacement
from the empirical distribution of $Y_t$ delivering $Y^*_t$ under the
null of nonlinear serial independence.  Bandwidths are obtained via
likelihood cross-validation by default for the marginal and joint
densities.

The following code snippet provides a simple example for a continuous
time series. Note that we strongly advocate the use of the integration
(default) version of the statistic in applied settings but use the
summation (i.e.~moment) version below purely by way of demonstration
as it is computationally faster.

<<keep.source=TRUE>>=
set.seed(123)
## A function to create a time series
ar.series <- function(phi,epsilon) {
  n <- length(epsilon)
  series <- numeric(n)
  series[1] <- epsilon[1]/(1-phi)
  for(i in 2:n) {
    series[i] <- phi*series[i-1] + epsilon[i]
  }
  return(series)
}
n <- 100
## Stationary persistent time-series
yt <- ar.series(0.95,rnorm(n))
npsdeptest(yt,lag.num=2,boot.num=99,method="summation")
@ 

\section{Rolling Your Own Entropy Function}

The functions outlined above allow for a range of entropy-based tests
where the underlying distributions are modelled
nonparametrically. However, practitioners may wish to construct their
own entropy measures that rely on nonparametric estimates of the
underlying distributions. By way of example we consider a metric
entropy that measures distance (Hellinger) of an unknown univariate
density from a specific parametric density. The parametric density for
this example will be Gaussian and the unknown density will be
estimated via kernel methods.

In the following code snippet we construct a simple function in R that
accomplishes this. You could of course use nonparametric methods for
both distributions or parametric methods for both. The function
\code{npudens} computes unconditional kernel density estimates, and
\code{fitted} computes the fitted density for the sample
realizations. The function \code{dnorm} provides the normal density
function. The \code{integrate} function performs univariate numerical
integration.

<<keep.source=TRUE>>=
Srho <- function(x,y,...) {
  ## First write a function to compute the integrand (this is fed to
  ## the `integrate' function). This function's first argument is the
  ## point at which the two densities are computed (the remaining
  ## arguments are the data vectors).
  integrand <- function(t,x,y) {
    ## First, nonparametrically estimate the density of the x data
    ## using a plug-in bandwidth and evaluate the density at the point
    ## `t'.
    f.x <- fitted(npudens(tdat=x,edat=t,bws=bw.SJ(x),...))
    ## Next, estimate the parametric density of the data y using the
    ## Gaussian distribution and evaluate the density at the point
    ## `t'.
    f.y <- dnorm(t,mean=mean(y),sd=sd(y))
    ## Compute and return the integrand evaluated at the point `t'.
    return(0.5*(sqrt(f.x)-sqrt(f.y))**2)
  }
  ## Feed the integrand function to integrate() and return the value
  ## of the integral.
  return(integrate(integrand,-Inf,Inf,x=x,y=y)$value)
}
set.seed(123)
n <- 1000
## Data drawn from the same distribution
x <- rnorm(n)
y <- rnorm(n)
Srho(x,y)
## Data drawn from different distributions
y <- rnorm(n,sd=100)
Srho(x,y)
@ 

Should the reader be interested in multivariate measures, multivariate
numerical integration is available in the \proglang{R} packages
\pkg{cubature} and \pkg{R2Cuba} (\citet{CUBATURE}, \citet{R2CUBA}).

\section{Summary}

The \pkg{np} package (\citet{np}) contains new functionality
introduced in versions 0.30-4 through 0.30-7 that implements a range
of entropy-based inferential procedures. We hope you find these to be
easy to use. Please report any issues encountered to
\email{racinej@mcmaster.ca}. Any suggestions for improvements to this
document would be greatly appreciated. As always, if you use these
functions for your work we would sincerely appreciate your citing both
the relevant published research papers and the \pkg{np} package
(\citet{np}).

\section*{Acknowledgments}

We would like to gratefully acknowledge support from the Natural Sciences and
Engineering Research Council of Canada (\url{https://nserc-crsng.canada.ca/}),
the Social Sciences and Humanities Research Council of Canada
(\url{https://sshrc-crsh.canada.ca/}), and the Shared Hierarchical Academic
Research Computing Network (\url{https://www.sharcnet.ca/}).

\bibliography{entropy_np}

\end{document}

# `kernels.max` Developer Notes

This file is a Maxima script used for symbolic derivation of higher-order
kernel objects (kernel, derivative, CDF, and convolution expressions) and
associated constants.

## Purpose

- `inst/kernels.max` is **developer tooling**, not runtime package code.
- It is useful when extending or validating higher-order kernel algebra.
- It is not sourced by R code during normal package execution.

## Typical Use

Install/launch Maxima from the official project page:
<https://maxima.sourceforge.io/>

1. Open Maxima.
2. Load `kernels.max`.
3. Run `computeKernels(<base_kernel>, <support>)` for the desired kernel.
4. Capture symbolic/numeric output for:
   - order-2/4/6/8 kernel forms,
   - derivatives,
   - CDF forms,
   - convolution forms,
   - kernel constants.

Example Maxima session sketch:

```maxima
load("kernels.max")$
computeKernels(epanechnikov, sqrt(5))$
computeKernels(gaussian, inf)$
```

## Maintenance Guidance

- Treat outputs as algebraic reference material.
- Before integrating derived expressions into C/R code, verify:
  - numerical parity against existing implementations,
  - boundary/support behavior,
  - objective-value parity in benchmark smoke tests.
- Keep any production code changes separate from symbolic scratch edits.

## Scope Boundaries

- Keep this file for reproducibility of derivations.
- Do not rely on it as an executable part of package installation/runtime.

## Canonical Implementation Directive (2026-03-05)

This repository follows a strict canonical execution rule:

1. One canonical implementation per method (outside explicit `np.tree` branching).
2. Unsupported configurations must fail fast with explicit `stop(...)` diagnostics.
3. No silent remap/coercion of user-selected options (for example `bwmethod`, `regtype`, kernels, `cv.iterate`, or bounds transforms).
4. No hidden alternate execution paths for the same method semantics.
5. All fit-defining options (for example `degree`, `basis`, `bernstein.basis`, kernels, and bounds) must be propagated and used by the canonical path.
6. `np.tree=FALSE` is the default; when `np.tree=TRUE`, behavior must remain semantics-preserving and option-compatible with the canonical path.
7. Remove or reject legacy/debug compatibility branches that add redundant runtime overhead once canonical behavior exists.

---
title: "Getting Started with np"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with np}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
options(np.messages = FALSE)
```

This vignette is meant to be the smallest useful package-side introduction to
`np`. The emphasis is on one clean workflow that users can run after
installation: choose a bandwidth, fit a model, inspect the result, and plot it.

Broader worked examples, package comparisons, and method-specific articles are
better carried by the gallery site:

- <https://jeffreyracine.github.io/gallery/np_npRmpi.html>
- <https://jeffreyracine.github.io/gallery/quickstarts.html>

## The basic workflow

In `np`, the bandwidth object is often the key object in the analysis.

1. compute or inspect a bandwidth object,
2. fit the model,
3. summarize or plot the result.

## A simple regression example

```{r}
library(np)
data(cps71, package = "np")

bw <- npregbw(logwage ~ age, data = cps71)
summary(bw)

fit <- npreg(bws = bw)
summary(fit)
```

## Plotting the fitted relationship

```{r, fig.width = 6, fig.height = 4}
plot(cps71$age, cps71$logwage, cex = 0.25, col = "grey")
lines(cps71$age, fitted(fit), col = 2, lwd = 2)
```

## Mixed data

One important feature of `np` is that it handles mixed data directly. Variable
class matters: unordered categorical variables should be factors, and ordered
categorical variables should be ordered factors when appropriate.

```{r}
set.seed(42)
mydat <- data.frame(
  y = rnorm(200),
  x_cont = runif(200),
  x_unordered = factor(sample(c("a", "b", "c"), 200, replace = TRUE)),
  x_ordered = ordered(sample(1:4, 200, replace = TRUE))
)

bw_mixed <- npregbw(y ~ x_cont + x_unordered + x_ordered, data = mydat)
fit_mixed <- npreg(bws = bw_mixed)
summary(fit_mixed)
```

## A note on modern local-polynomial search

For local-polynomial-capable methods, `np` now supports joint selection of
polynomial order and bandwidth. The modern route is to use
`search.engine = "nomad+powell"` when you want the search to choose both
together.

If you want the recommended route without spelling out all of the LP tuning
arguments, use `nomad = TRUE`. This is a documented convenience preset, not a
generic optimizer alias: it fills only missing values among the LP degree-search
controls and leaves compatible explicit overrides in place. This route uses the
optional NOMAD backend provided by the suggested package `crs`, so install
`crs` first if you want to use `nomad = TRUE` or
`search.engine = "nomad"`/`"nomad+powell"`.

```{r}
if (requireNamespace("crs", quietly = TRUE) &&
    utils::packageVersion("crs") >= package_version("0.15-41")) {
  set.seed(7)
  n <- 120
  x <- runif(n, -1, 1)
  y <- x + 0.4 * x^2 + rnorm(n, sd = 0.18)

  fit_nomad <- npreg(y ~ x, nomad = TRUE, degree.max = 1L, nmulti = 1L)
  fit_nomad$bws$nomad.shortcut

  # Tune one component explicitly while leaving the rest of the preset in place.
  fit_nomad_direct <- npreg(
    y ~ x,
    nomad = TRUE,
    search.engine = "nomad",
    degree.max = 1L,
    nmulti = 1L
  )
}
```

The same convenience entry point is available for the other LP-capable families:
`npcdens`, `npcdist`, `npplreg`, `npscoef`, and `npindex`, together with their
corresponding `*bw` constructors.

Keep the first run modest and runnable. Fuller worked examples belong on the
gallery rather than in this package vignette.

## Data preparation matters

In `np`, the formula interface tells the function which variables are the
response and regressors. It is not imposing an ordinary linear-additive model.

It is also important not to pass blocks of 0/1 dummies as if this were a
standard linear-model workflow. If the underlying variable is categorical, it
is usually better to keep it as one `factor` or `ordered` variable.

## Other common starting points

This vignette keeps the package-side introduction intentionally narrow. Other
common first routes are:

- `?npudens` and `?npudist` for unconditional density and distribution work,
- `?npcdens`, `?npcdist`, and `?npqreg` for conditional density, distribution,
  and quantiles,
- `?npconmode` for classification and conditional mode estimation,
- `?npplreg`, `?npindex`, and `?npscoef` for semiparametric models.

Those broader branches are better carried by help pages and website articles
than by a single shipped vignette.

## Where to go next

- `vignette("np_entropy_tests", package = "np")` for a compact package-side
  testing overview
- `?npreg`, `?npregbw`, `?npudens`, and `?npcdens` for core help pages
- <https://jeffreyracine.github.io/gallery/kernel_primer.html> for the
  conceptual kernel overview
- <https://jeffreyracine.github.io/gallery/density_distribution_quantiles.html>
  for density, distribution, and quantile workflows
- <https://jeffreyracine.github.io/gallery/semiparametric_models.html> for
  partially linear, single-index, and varying-coefficient routes

---
title: "Entropy and Testing with np"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Entropy and Testing with np}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
options(np.messages = FALSE)
```

This vignette is meant to be a small package-side introduction to the
entropy-based testing tools in `np`. It is intentionally much shorter than the
legacy article-style document and focuses on what the functions are for and one
small runnable example.

The fuller narrative treatment now belongs on the gallery site rather than in a
shipped package vignette:

- <https://jeffreyracine.github.io/gallery/entropy_tests.html>

## Main functions

The main entropy-based testing functions are:

- `npdeneqtest`: equality of multivariate densities
- `npunitest`: equality of univariate densities
- `npsymtest`: asymmetry in a univariate variable or series
- `npdeptest`: nonlinear pairwise dependence
- `npsdeptest`: nonlinear serial dependence

These functions can be computationally demanding, especially when integration
and bootstrap resampling are involved.

## A small example

For a first run, it is reasonable to begin with a simple univariate comparison
and keep the example small enough that bootstrapping remains practical.

```{r}
library(np)
set.seed(42)

n <- 250
x <- rnorm(n)
y <- rnorm(n)

npunitest(x, y, bootstrap = TRUE)
```

## Practical guidance

- start with the smallest example that answers your question,
- use the default integral-based versions for serious work unless you have a
  reason not to,
- expect runtime to grow quickly when bootstrap resampling is involved,
- if the testing workflow is correct but the runtime becomes burdensome, move to
  `npRmpi` rather than rewriting the statistical problem.

## Where to go next

- `?npunitest`, `?npdeneqtest`, `?npdeptest`, `?npsdeptest`, `?npsymtest`
- <https://jeffreyracine.github.io/gallery/entropy_tests.html> for the longer
  website article
- <https://jeffreyracine.github.io/gallery/mpi_large_data.html> if the testing
  workflow is correct but runtime now points to `npRmpi`

\name{npseed}
\title{Set Random Seed}
\alias{npseed}

\usage{
npseed(seed)
}

\arguments{
\subsection{Seed Input}{
Seed value used to reset the package C backend random number generator.
}

  \item{seed}{a single finite numeric value that is representable as a non-negative integer after applying \code{abs()}.}
}
\value{
  None.
}
\description{
  \code{npseed} is a function which sets the random seed in the
  \code{\link{np}} C backend, resetting the random number generator.
}

\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.

  \code{npseed} provides an interface for setting the random seed (and
  resetting the random number generator) used
  by \code{\link{np}}. The random number generator is used during the
  bandwidth search procedure to set the search starting point, and in
  subsequent searches when using multistarting, to avoid being trapped
  in local minima if the objective function is not globally concave.

  Calling \code{npseed} will only affect the numerical search if it is
  performed by the C backend. The affected functions include:
  \code{\link{npudensbw}}, \code{\link{npcdensbw}},
  \code{\link{npregbw}}, \code{\link{npplregbw}}, \code{\link{npqreg}},
  \code{\link{npcmstest}} (via \code{\link{npregbw}}),
  \code{\link{npqcmstest}} (via \code{\link{npregbw}}),
  \code{\link{npsigtest}} (via \code{\link{npregbw}}).

  Invalid inputs such as missing values, non-finite values, non-scalars,
  and values that cannot be represented as integers are rejected with an
  explicit error.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}
  \code{\link{set.seed}}
}

\references{
    Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics:
      Theory and Practice,} Princeton University Press.
}

\author{ Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey
  S. Racine \email{racinej@mcmaster.ca}
}

\section{Note}{
  This method currently only supports objects from the \code{\link{np}} library.
}
\examples{
npseed(712)
x <- runif(10)
y <- x + rnorm(10, sd = 0.1)
bw <- npregbw(y~x)
}
\keyword{nonparametric}

\name{npcdist}
\alias{npcdist}
\alias{npcdist.condbandwidth}
\alias{npcdist.default}
\alias{npcdist.formula}

\title{Kernel Conditional Distribution Estimation with Mixed Data Types}

\description{
  \code{npcdist} computes kernel cumulative conditional distribution
  estimates on \eqn{p+q}-variate evaluation data, given a set of
  training data (both explanatory and dependent) and a bandwidth
  specification (a \code{condbandwidth} object or a bandwidth vector,
  bandwidth type, and kernel type) using the method of Li and Racine
  (2008) and Li, Lin, and Racine (2013).  The data may be continuous,
  discrete (unordered and ordered factors), or some combination thereof.
}

\usage{
npcdist(bws, \dots)

\method{npcdist}{formula}(bws, data = NULL, newdata = NULL, \dots)


\method{npcdist}{condbandwidth}(bws,
        txdat = stop("invoked without training data 'txdat'"),
        tydat = stop("invoked without training data 'tydat'"),
        exdat,
        eydat,
        gradients = FALSE,
        gradient.order = 1L,
        proper = FALSE,
        proper.method = c("isotonic"),
        proper.control = list(),
        \dots)

\method{npcdist}{default}(bws, txdat, tydat, nomad = FALSE, \dots)

}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the bandwidth specification, formula/data interface, and training data.
}

  \item{bws}{
    a bandwidth specification. This can be set as a \code{condbandwidth}
    object returned from a previous invocation of
    \code{\link{npcdistbw}}, or as a \eqn{p+q}-vector of bandwidths,
    with each element \eqn{i} up to \eqn{i=q} corresponding to the
    bandwidth for column \eqn{i} in \code{tydat}, and each element
    \eqn{i} from \eqn{i=q+1} to \eqn{i=p+q} corresponding to the
    bandwidth for column \eqn{i-q} in \code{txdat}. If specified as a
    vector, then additional arguments will need to be supplied as
    necessary to specify the bandwidth type, kernel types, training
    data, and so on.
  }


  \item{data}{
    an optional data frame, list or environment (or object coercible to
    a data frame by \code{\link{as.data.frame}}) containing the
    variables in the model. If not found in data, the variables are
    taken from \code{environment(bws)}, typically the environment from
    which \code{\link{npcdistbw}} was called.
 }


  \item{txdat}{
    a \eqn{p}-variate data frame of sample realizations of explanatory
    data (training data). Defaults to the training data used to
    compute the bandwidth object.
  }


  \item{tydat}{
    a \eqn{q}-variate data frame of sample realizations of dependent
    data (training data). Defaults to the training data used to
    compute the bandwidth object.
  }


\subsection{Local-Polynomial Degree And Bandwidth Search}{
This argument controls the recommended automatic local-polynomial NOMAD route, which jointly selects continuous polynomial degree and bandwidths when these are computed inside \code{npcdist}.
}

  \item{nomad}{
    logical shortcut passed through to \code{\link{npcdistbw}} when
    bandwidths are computed inside \code{npcdist}. When \code{TRUE},
    the bandwidth route fills any missing values among \code{regtype},
    \code{search.engine}, \code{degree.select},
    \code{bernstein.basis}, \code{degree.min}, \code{degree.max},
    \code{degree.verify}, and \code{bwtype} with the recommended
    automatic local-polynomial degree-and-bandwidth NOMAD preset documented in \code{\link{npcdistbw}}.
    Additional NOMAD tuning arguments such as \code{nomad.nmulti} may
    also be supplied through \code{\dots}; \code{nmulti} remains the
    outer restart count while \code{nomad.nmulti} controls inner
    \code{crs::snomadr()} multistarts within each outer restart.
    After fitting, inspect \code{fit$bws$nomad.shortcut} on the
    returned object \code{fit} to see the normalized shortcut metadata.
  }


\subsection{Evaluation Data And Returned Quantities}{
These arguments control where the fitted conditional distribution is evaluated and which estimates are returned.
}

  \item{exdat}{
    a \eqn{p}-variate data frame of explanatory data on
    which cumulative conditional distributions will be evaluated. By
    default, evaluation takes place on the data provided by
    \code{txdat}.
  }


  \item{eydat}{
    a \eqn{q}-variate data frame of dependent data on which
    cumulative conditional distributions will be evaluated. By default,
    evaluation takes place on the data provided by \code{tydat}.
  }


  \item{gradients}{
    a logical value specifying whether to return estimates of the
    gradients at the evaluation points. Defaults to \code{FALSE}.
  }


  \item{gradient.order}{
    derivative order for continuous explanatory-variable gradients when
    \code{gradients = TRUE} and \code{regtype = "lp"}. A scalar is
    recycled across continuous explanatory variables, or one value may
    be supplied per continuous explanatory variable. The default
    \code{1L} returns first derivatives. Higher orders are available
    for continuous explanatory variables only and must not exceed the
    corresponding local-polynomial degree; unordered and ordered
    predictors retain their usual first-order discrete effects.
  }


  \item{newdata}{
    An optional data frame in which to look for evaluation data. If
    omitted, the training data are used.
  }



\subsection{Fit Properization Controls}{
These arguments control optional post-estimation properization of the fitted conditional distribution.
}

  \item{proper}{
    a logical value specifying whether to apply post-estimation
    properization to the conditional distribution estimate. Defaults to
    \code{FALSE}.
  }


  \item{proper.control}{
    a list of controls for properization. Supported entries are
    \code{tol}, \code{grid.check}, \code{store.raw}, and
    \code{fail.on.unsupported}.
  }



  \item{proper.method}{
    a character string specifying the properization method. Currently
    \code{"isotonic"} is supported.
  }


\subsection{Additional Arguments}{
Further arguments are passed to \code{\link{npcdistbw}} when bandwidths are computed internally, or used to interpret a numeric \code{bws} vector.
}

  \item{\dots}{
    additional arguments supplied to \code{\link{npcdistbw}} when
    \code{npcdist} computes bandwidths internally, or arguments needed
    to interpret a numeric \code{bws} vector. This is where bandwidth
    selection controls such as \code{bwmethod}, \code{bwtype},
    kernel/support controls such as \code{cxkertype},
    \code{cykertype}, \code{cxkerbound}, and \code{cykerbound},
    search controls such as \code{nmulti},
    \code{scale.factor.search.lower}, and \code{nomad.nmulti}, and
    local-polynomial controls such as \code{regtype}, \code{degree},
    \code{basis}, and \code{bernstein.basis} are supplied. See
    \code{\link{npcdistbw}} for the complete bandwidth-selection
    argument surface.
  }

}
\details{
  Documentation guide: see \code{\link{npcdistbw}} for bandwidth selection and search controls, \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}}, \code{\link{plot.np}} for plotting options.

  When \code{bws} is omitted, the formula and default methods call
  \code{\link{npcdistbw}} first and pass bandwidth-selection arguments
  from \code{\dots} to that call. When \code{bws} is already a
  \code{condbandwidth} object, \code{npcdist} estimates with the stored
  bandwidth metadata in that object.

  Argument groups for bandwidth selection are documented on
  \code{\link{npcdistbw}}. The most common workflow is to choose data
  and bandwidth inputs first, then bandwidth criterion and
  representation, then kernel/support controls, numerical search
  controls, and finally local-polynomial/NOMAD controls for
  polynomial-adaptive fits.

  For S3 plotting help, see \code{\link{plot.np}}. You can list
  available plot methods with \code{methods("plot")}.

  \code{npcdist} implements a variety of methods for estimating
  multivariate conditional cumulative distributions (\eqn{p+q}-variate)
  defined over a set of possibly continuous and/or discrete (unordered,
  ordered) data. The approach is based on Li and Racine (2004) who
  employ \sQuote{generalized product kernels} that admit a mix of
  continuous and discrete data types.

  Three classes of kernel estimators for the continuous data types are
  available: fixed, adaptive nearest-neighbor, and generalized
  nearest-neighbor. Adaptive nearest-neighbor bandwidths change with
  each sample realization in the set, \eqn{x_i}{x[i]}, when estimating
  the cumulative conditional distribution at the point
  \eqn{x}. Generalized nearest-neighbor bandwidths change with the point
  at which the cumulative conditional distribution is estimated,
  \eqn{x}. Fixed bandwidths are constant over the support of \eqn{x}.

  Training and evaluation input data  may be a
  mix of continuous (default), unordered discrete (to be specified in
  the data frames using \code{\link{factor}}), and ordered discrete (to be
  specified in the data frames using \code{\link{ordered}}). Data can be
  entered in an arbitrary order and data types will be detected
  automatically by the routine (see \code{\link{np}} for details).

  A variety of kernels may be specified by the user. Kernels implemented
  for continuous data types include the second, fourth, sixth, and eighth
  order Gaussian and Epanechnikov kernels, and the uniform
  kernel. Unordered discrete data types use a variation on Aitchison and
  Aitken's (1976) kernel, while ordered data types use a variation of the
  Wang and van Ryzin (1981) kernel.

  For practitioners who want the recommended automatic local-polynomial degree-and-bandwidth NOMAD route
  without spelling out all LP tuning arguments,
  \code{npcdist(..., nomad=TRUE)} and \code{npcdistbw(..., nomad=TRUE)}
  expand missing settings to the same documented preset. Explicit
  incompatible settings fail fast rather than being silently rewritten.

  With \code{regtype = "lp"}, \code{gradients = TRUE} and
  \code{gradient.order} greater than one expose higher-order derivative
  estimates with respect to continuous explanatory variables. These
  derivatives use the same local-polynomial basis, degree, Bernstein
  option, bandwidths, and kernels as the fitted conditional
  distribution. Asymptotic standard errors for these higher-order
  derivative columns are not currently reported; the corresponding
  entries of \code{congerr} are \code{NA}. Use bootstrap intervals when
  inference on higher-order derivatives is required.

}
\value{
  \code{npcdist} returns a \code{condistribution} object. The generic
  accessor functions \code{\link{fitted}}, \code{\link{se}}, and
  \code{\link{gradients}}, extract estimated values, asymptotic standard
  errors on estimates, and gradients, respectively, from
  the returned object. Furthermore, the functions \code{\link{predict}},
  \code{\link{summary}}
  and \code{\link{plot}} support objects of both classes. The returned objects
  have the following components:

  \item{xbw}{ bandwidth(s), scale factor(s) or nearest neighbours for the
    explanatory data, \code{txdat} }
  \item{ybw}{ bandwidth(s), scale factor(s) or nearest neighbours for the
    dependent data, \code{tydat} }
  \item{xeval}{ the evaluation points of the explanatory data }
  \item{yeval}{ the evaluation points of the dependent data }
  \item{condist}{ estimates of the conditional cumulative
    distribution at the evaluation points }
  \item{conderr}{ standard errors of the cumulative conditional distribution
    estimates }
  \item{congrad}{ if invoked with \code{gradients = TRUE}, estimates of
    the gradients at the evaluation points. For local-polynomial fits,
    continuous-coordinate derivative orders are controlled by
    \code{gradient.order}. }
  \item{congerr}{ if invoked with \code{gradients = TRUE}, standard
    errors of the gradients at the evaluation points. Higher-order
    continuous-coordinate derivative standard errors are currently
    returned as \code{NA}; bootstrap inference is preferred for those
    targets. }
  \item{log_likelihood}{ log likelihood of the cumulative conditional distribution estimate }

}

\section{Book And Method Pointers}{
  The conditional distribution target is
  \eqn{F(y\mid x)=\Pr(Y\le y\mid X=x)}. The estimator uses the selected
  mixed-data conditional distribution bandwidths and kernels for the
  response and conditioning coordinates; local-polynomial routes use the
  selected continuous-coordinate polynomial degree. These fitted
  conditional CDFs are also the object inverted by \code{\link{npqreg}}
  when computing conditional quantiles.

  For book-length derivations, see Li and Racine (2007), Chapter 6
  \emph{Conditional CDF and Quantile Estimation}, especially Sections 6.1,
  6.2, and 6.5, together with Chapter 5 \emph{Conditional Density
  Estimation}. The later workflow treatment is Racine (2019), Chapter
  4 \emph{Conditional Probability Density and Cumulative Distribution
  Functions}.
}

\references{

  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Hall, P. and J.S. Racine and Q. Li (2004), \dQuote{Cross-validation and the
    estimation of conditional probability densities,} Journal of the
  American Statistical Association, 99, 1015-1026.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
    and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2008), \dQuote{Nonparametric estimation of
  conditional CDF and quantile functions with mixed categorical and
  continuous data,} Journal of Business and Economic Statistics, 26,
  423-434.

  Li, Q. and J. Lin and J.S. Racine (2013), \dQuote{Optimal bandwidth
  selection for nonparametric conditional distribution and quantile
  functions}, Journal of Business and Economic Statistics, 31, 57-65.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,} Cambridge
  University Press.

  Scott, D.W. (1992), \emph{Multivariate Density Estimation. Theory,
  Practice and Visualization,} New York: Wiley.

  Silverman, B.W. (1986), \emph{Density Estimation,} London: Chapman and
  Hall.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth estimators
    for discrete distributions,}  Biometrika, 68, 301-309.

}
\author{
  Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
  \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}, \code{\link{plot.np}}
  \code{\link{npudens}}
}
\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): For this example, we load Giovanni
# Baiocchi's Italian GDP panel (see Italy for details), and compute the
# cross-validated bandwidths (default) using a second-order Gaussian
# kernel (default). Note - this may take a minute or two depending on
# the speed of your computer.

data("Italy")
Italy <- Italy[seq_len(min(300, nrow(Italy))), ]
with(Italy, {

# First, compute the bandwidths.

bw <- npcdistbw(formula=gdp~ordered(year), nmulti=1)

# Next, compute the condistribution object...

Fhat <- npcdist(bws=bw)

# The object Fhat now contains results such as the estimated cumulative
# conditional distribution function (Fhat$condist) and so on...

summary(Fhat)

# Call the plot() function to visualize the results (<ctrl>-C will
# interrupt on *NIX systems, <esc> will interrupt on MS Windows
# systems).

if (interactive()) plot(bw)

})

# EXAMPLE 1 (INTERFACE=DATA FRAME): For this example, we load Giovanni
# Baiocchi's Italian GDP panel (see Italy for details), and compute the
# cross-validated bandwidths (default) using a second-order Gaussian
# kernel (default). Note - this may take a minute or two depending on
# the speed of your computer.

data("Italy")
Italy <- Italy[seq_len(min(300, nrow(Italy))), ]
with(Italy, {

# First, compute the bandwidths.

# Note - we cast `X' and `y' as data frames so that plot() can
# automatically grab names (this looks like overkill, but in
# multivariate settings you would do this anyway, so may as well get in
# the habit).

X <- data.frame(year=ordered(year))
y <- data.frame(gdp)

bw <- npcdistbw(xdat=X, ydat=y, nmulti=1)

# Next, compute the condistribution object...

Fhat <- npcdist(bws=bw)

# The object Fhat now contains results such as the estimated cumulative
# conditional distribution function (Fhat$condist) and so on...

summary(Fhat)

# Call the plot() function to visualize the results (<ctrl>-C will
# interrupt on *NIX systems, <esc> will interrupt on MS Windows systems).

if (interactive()) plot(bw)

})

# EXAMPLE 2 (INTERFACE=FORMULA): For this example, we load the old
# faithful geyser data from the R `datasets' library and compute the
# conditional distribution function.

library("datasets")
data("faithful")
with(faithful, {

# Note - this may take a few minutes depending on the speed of your
# computer...

bw <- npcdistbw(formula=eruptions~waiting, nmulti=1)

summary(bw)

# Plot the conditional cumulative distribution function (<ctrl>-C will
# interrupt on *NIX systems, <esc> will interrupt on MS Windows
# systems).

if (interactive()) plot(bw)

})

# EXAMPLE 2 (INTERFACE=DATA FRAME): For this example, we load the old
# faithful geyser data from the R `datasets' library and compute the
# cumulative conditional distribution function.

library("datasets")
data("faithful")
with(faithful, {

# Note - this may take a few minutes depending on the speed of your
# computer...

# Note - we cast `X' and `y' as data frames so that plot() can
# automatically grab names (this looks like overkill, but in
# multivariate settings you would do this anyway, so may as well get in
# the habit).

X <- data.frame(waiting)
y <- data.frame(eruptions)

bw <- npcdistbw(xdat=X, ydat=y, nmulti=1)

summary(bw)

# Plot the conditional cumulative distribution function (<ctrl>-C will
# interrupt on *NIX systems, <esc> will interrupt on MS Windows systems)

if (interactive()) plot(bw)

})

# EXAMPLE 3: Variations on local polynomial conditional distribution
# estimation with proper = TRUE.

data("Italy")

Italy2 <- within(Italy, {
  year <- as.numeric(as.character(year))
})

# Plot only: make the plotted surface proper on the plot evaluation grid.

Fhat <- npcdist(gdp ~ year, data = Italy2,
                regtype = "lp", degree = 3, nmulti = 1)

plot(Fhat, proper = TRUE)

# Fit an object whose fitted values are themselves proper.

ctrl_fit <- list(
  mode = "slice",
  apply = "fitted",
  slice.grid.size = 101L,
  slice.extend.factor = 0.1
)

Fhat_fit <- npcdist(
  gdp ~ year,
  data = Italy2,
  regtype = "lp",
  degree = 3,
  nmulti = 1,
  proper = TRUE,
  proper.control = ctrl_fit
)

fit_proper <- fitted(Fhat_fit)
fit_raw <- Fhat_fit$condist.raw

# Predict on a common explicit y-grid for several years, and render
# those predictions proper.

g.grid <- seq(min(Italy2$gdp), max(Italy2$gdp), length.out = 200)

nd_grid <- expand.grid(
  gdp = g.grid,
  year = c(1955, 1975, 1995)
)

pred_grid <- predict(Fhat, newdata = nd_grid, proper = TRUE)

# Predict on paired rows with different gdp grids by year, and still
# make the predictions proper via slice mode.

g1 <- seq(quantile(Italy2$gdp, 0.10),
          quantile(Italy2$gdp, 0.60), length.out = 60)
g2 <- seq(quantile(Italy2$gdp, 0.30),
          quantile(Italy2$gdp, 0.90), length.out = 35)

nd_slice <- rbind(
  data.frame(gdp = g1, year = rep(1960, length(g1))),
  data.frame(gdp = g2, year = rep(1985, length(g2)))
)

pred_slice <- predict(
  Fhat,
  newdata = nd_slice,
  proper = TRUE,
  proper.control = list(mode = "slice")
)

# One object that carries properization for fitted values and for later
# predict() calls.

ctrl_both <- list(
  mode = "slice",
  apply = "both",
  slice.grid.size = 101L,
  slice.extend.factor = 0.1
)

Fhat_both <- npcdist(
  gdp ~ year,
  data = Italy2,
  regtype = "lp",
  degree = 3,
  nmulti = 1,
  proper = TRUE,
  proper.control = ctrl_both
)

fit_both <- fitted(Fhat_both)
pred_both <- predict(
  Fhat_both,
  newdata = nd_slice,
  proper.control = ctrl_both
)

} % enddontrun
}
\keyword{nonparametric}

\name{b.star}
\alias{b.star}
\title{Compute Optimal Block Length for Stationary and Circular Bootstrap}
\usage{
b.star(data,
       Kn = NULL,
       mmax= NULL,
       Bmax = NULL,
       c = NULL,
       round = FALSE)
}
\arguments{
\subsection{Data Input}{
Time-series data used for automatic block-length selection.
}

  \item{data}{data, an n x k matrix, each column being a data series.}
  
\subsection{Block-Length Selection Controls}{
Tuning constants from Politis and White (2004) and Patton, Politis, and
White (2009).
}

  \item{Kn}{ See footnote c, page 59, Politis and White (2004). Defaults
    to \code{ceiling(log10(n))}.}

  \item{mmax}{See Politis and White (2004). Defaults to
  \code{ceiling(sqrt(n))+Kn}.}

  \item{Bmax}{See Politis and White (2004). Defaults to
    \code{ceiling(min(3*sqrt(n),n/3))}.}

  \item{c}{See Politis and White (2004). Defaults to
  \code{qnorm(0.975)}.}

\subsection{Output Rounding}{
Control for rounding the selected block lengths.
}

  \item{round}{whether to round the result or not. Defaults to FALSE.}
 
}
\value{
  
  A kx2 matrix of optimal bootstrap block lengths computed from
 \code{data} for the stationary bootstrap and circular bootstrap (column
 1 is for the stationary bootstrap, column 2 the circular).

}
\description{
  \code{b.star} is a function which computes the optimal block length
  for the continuous variable \code{data} using the method described in
  Patton, Politis and White (2009).
}
\details{
  \code{b.star} is a function which computes optimal block lengths for
  the stationary and circular bootstraps. This allows the use of
  \code{tsboot} from the \pkg{boot} package to be fully
  automatic by using the output from \code{b.star} as an input to the
  argument \code{l = } in \code{tsboot}. See below for an example.
}

\references{

  Patton, A. and D.N. Politis and H. White (2009), \dQuote{CORRECTION TO
    "Automatic block-length selection for the dependent bootstrap" by
    D. Politis and H. White}, Econometric Reviews 28(4), 372-375.

  Politis, D.N. and J.P. Romano (1994), \dQuote{Limit theorems for
    weakly dependent Hilbert space valued random variables with
    applications to the stationary bootstrap}, Statistica Sinica 4,
    461-476.

  Politis, D.N. and H. White (2004), \dQuote{Automatic block-length
    selection for the dependent bootstrap}, Econometric Reviews 23(1),
    53-70.

}

\author{
  Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine \email{racinej@mcmaster.ca}
}

\examples{
set.seed(12345)

# Function to generate an AR(1) series

ar.series <- function(phi,epsilon) {
  n <- length(epsilon)
  series <- numeric(n)
  series[1] <- epsilon[1]/(1-phi)
  for(i in 2:n) {
    series[i] <- phi*series[i-1] + epsilon[i]
  }
  return(series)
}

yt <- ar.series(0.1,rnorm(10000))
b.star(yt,round=TRUE)

yt <- ar.series(0.9,rnorm(10000))
b.star(yt,round=TRUE)
}

\keyword{univar}
\keyword{nonparametric}

\name{npcdistbw}
\alias{npcdistbw}
\alias{npcdistbw.condbandwidth}
\alias{npcdistbw.default}
\alias{npcdistbw.formula}
\title{Kernel Conditional Distribution Bandwidth Selection with Mixed
Data Types}

\description{ 
  \code{npcdistbw} computes a \code{condbandwidth} object for estimating
  a \eqn{p+q}-variate kernel conditional cumulative distribution
  estimator defined over mixed continuous and discrete (unordered
  \code{xdat}, ordered \code{xdat} and \code{ydat}) data using either
  the normal-reference rule-of-thumb or least-squares cross validation
  method of Li and Racine (2008) and Li, Lin and Racine
  (2013).
}

\usage{
npcdistbw(\dots)

\method{npcdistbw}{formula}(formula, 
          data, 
          subset, 
          na.action, 
          call, 
          gdata = NULL,
          \dots)

\method{npcdistbw}{condbandwidth}(xdat = stop("data 'xdat' missing"),
          ydat = stop("data 'ydat' missing"),
          gydat = NULL,
          bws,
          bandwidth.compute = TRUE,
          cfac.dir = 2.5*(3.0-sqrt(5)),
          scale.factor.init = 0.5,
          dfac.dir = 0.25*(3.0-sqrt(5)),
          dfac.init = 0.375,
          dfc.dir = 3,
          do.full.integral = FALSE,
          ftol = 1.490116e-07,
          scale.factor.init.upper = 2.0,
          hbd.dir = 1,
          hbd.init = 0.9,
          initc.dir = 1.0,
          initd.dir = 1.0,
          invalid.penalty = c("baseline","dbmax"),
          itmax = 10000,
          lbc.dir = 0.5,
          scale.factor.init.lower = 0.1,
          lbd.dir = 0.1,
          lbd.init = 0.1,
          memfac = 500.0,
          ngrid = 100,
          nmulti,
          penalty.multiplier = 10,
          powell.remin = TRUE,
          scale.init.categorical.sample = FALSE,
          scale.factor.search.lower = NULL,
          small = 1.490116e-05,
          tol = 1.490116e-04,
          transform.bounds = FALSE,
          \dots)

\method{npcdistbw}{default}(xdat = stop("data 'xdat' missing"),
          ydat = stop("data 'ydat' missing"),
          gydat,
          bws,
          bandwidth.compute = TRUE,
          bwmethod,
          bwscaling,
          bwtype,
          cfac.dir,
          scale.factor.init,
          cxkerbound,
          cxkerlb,
          cxkerorder,
          cxkertype,
          cxkerub,
          cykerbound,
          cykerlb,
          cykerorder,
          cykertype,
          cykerub,
          dfac.dir,
          dfac.init,
          dfc.dir,
          do.full.integral,
          ftol,
          scale.factor.init.upper,
          hbd.dir,
          hbd.init,
          initc.dir,
          initd.dir,
          invalid.penalty,
          itmax,
          lbc.dir,
          scale.factor.init.lower,
          lbd.dir,
          lbd.init,
          memfac,
          ngrid,
          nmulti,
          oxkertype,
          oykertype,
          penalty.multiplier,
          nomad.remin = FALSE,
          powell.remin,
          scale.init.categorical.sample,
          scale.factor.search.lower = NULL,
          small,
          tol,
          transform.bounds,
          uxkertype,
          regtype = c("lc", "ll", "lp"),
          basis = c("glp", "additive", "tensor"),
          degree = NULL,
          degree.select = c("manual", "coordinate", "exhaustive"),
          search.engine = c("nomad+powell", "cell", "nomad"),
          nomad = FALSE,
          nomad.nmulti = 0L,
            degree.min = NULL,
          degree.max = NULL,
          degree.start = NULL,
          degree.restarts = 0L,
          degree.max.cycles = 20L,
          degree.verify = FALSE,
          bernstein.basis = FALSE,
          \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the data, formula interface, optional distribution grid, and whether bandwidths are supplied or computed.
}

  \item{bandwidth.compute}{
    a logical value which specifies whether to do a numerical search for
    bandwidths or not. If set to \code{FALSE}, a \code{condbandwidth} object
    will be returned with bandwidths set to those specified
    in \code{bws}. Defaults to \code{TRUE}.
  }

  \item{bws}{
    a bandwidth specification. This can be set as a \code{condbandwidth}
    object returned from a previous invocation, or as a \eqn{p+q}-vector
    of bandwidths, with each element \eqn{i} up to \eqn{i=q}
    corresponding to the bandwidth for column \eqn{i} in \code{ydat},
    and each element \eqn{i} from \eqn{i=q+1} to \eqn{i=p+q}
    corresponding to the bandwidth for column \eqn{i-q} in
    \code{xdat}. In either case, the bandwidth supplied will serve as a
    starting point in the numerical search for optimal bandwidths. If
    specified as a vector, then additional arguments will need to be
    supplied as necessary to specify the bandwidth type, kernel types,
    selection methods, and so on. This can be left unset.
  }


    
  \item{call}{
    the original function call. This is passed internally by
    \code{\link{np}} when a bandwidth search has been implied by a call to
    another function. It is not recommended that the user set this.
  }

  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(formula)}, typically the environment from which the
    function is called.
  }

  \item{formula}{
    a symbolic description of variables on which bandwidth selection is
    to be performed. The details of constructing a formula are
    described below.
  }

  \item{gdata}{
    a grid of data on which the indicator function for
    least-squares cross-validation is to be computed (can be the sample
    or a grid of quantiles).
  }

  \item{gydat}{
    a grid of data on which the indicator function for
    least-squares cross-validation is to be computed (can be the sample
    or a grid of quantiles for \code{ydat}).
  }

  \item{na.action}{
    a function which indicates what should happen when the data contain
   \code{NA}s. The default is set by the \code{\link{na.action}} setting of options, and is
    \code{\link{na.fail}} if that is unset. The (recommended) default is
    \code{\link{na.omit}}.  
  }
  
  \item{subset}{
    an optional vector specifying a subset of observations to be used in
    the fitting process. 
  }

  \item{xdat}{
    a \eqn{p}-variate data frame of explanatory data on which bandwidth selection will
    be performed. The data types may be continuous, discrete (unordered
    and ordered factors), or some combination thereof.
  }

  \item{ydat}{
    a \eqn{q}-variate data frame of dependent data on which bandwidth
    selection will be performed. The data types may be continuous,
    discrete (ordered factors), or some combination thereof.
  }


\subsection{Automatic Degree Search Controls}{
These arguments control automatic local-polynomial degree search when \code{regtype="lp"}.
}

  \item{degree.max}{
    optional scalar or integer vector giving upper bounds for automatic
    degree search over continuous \code{xdat} predictors when
    \code{degree.select != "manual"}.
  }

  \item{degree.max.cycles}{
    positive integer giving the maximum number of coordinate-search
    sweeps over the degree vector. Ignored for \code{"manual"} and
    \code{"exhaustive"} degree selection.
  }

  \item{degree.min}{
    optional scalar or integer vector giving lower bounds for automatic
    degree search over continuous \code{xdat} predictors when
    \code{degree.select != "manual"}.
  }

  \item{degree.restarts}{
    non-negative integer giving the number of additional deterministic
    coordinate-search restarts. Ignored for \code{"manual"} and
    \code{"exhaustive"} degree selection.
  }

  \item{degree.select}{
    character string controlling local-polynomial degree handling when
    \code{regtype="lp"}. \code{"manual"} (default) treats
    \code{degree} as fixed. \code{"coordinate"} performs cached
    coordinate-wise search over admissible degree vectors for the
    continuous \code{xdat} predictors. \code{"exhaustive"} evaluates
    the full admissible degree grid when \code{search.engine="cell"}.
    For NOMAD-based search engines, any non-\code{"manual"} value
    requests direct joint search over degree and bandwidth
    coordinates.
  }

  \item{degree.start}{
    optional starting degree vector for automatic coordinate search. If
    omitted, the search starts from the degree-zero local-constant
    baseline on the continuous \code{xdat} predictors.
  }

  \item{degree.verify}{
    logical value indicating whether a coordinate-search solution should
    be exhaustively verified over the admissible degree grid after the
    heuristic phase completes. Available only for
    \code{search.engine="cell"}.
  }


\subsection{Bandwidth Criterion And Representation}{
These arguments choose the selection criterion and the way continuous bandwidths are represented.
}

  \item{bwmethod}{ which method to use to select bandwidths.
    \code{cv.ls} specifies least-squares cross-validation (Li, Lin and
    Racine (2013), and \code{normal-reference} just computes the
    \sQuote{rule-of-thumb} bandwidth \eqn{h_j}{h[j]} using the standard
    formula \eqn{h_j = 1.06 \sigma_j n^{-1/(2P+l)}}{h[j] =
    1.06*sigma[j]*n^(-1.0/(2.0*P+l))}, where \eqn{\sigma_j}{sigma[j]} is
    an adaptive measure of spread of the \eqn{j}th continuous variable
    defined as min(standard deviation, mean absolute deviation/1.4826,
    interquartile range/1.349), \eqn{n} the number of observations,
    \eqn{P} the order of the kernel, and \eqn{l} the number of
    continuous variables. Note that when there exist factors and the
    normal-reference rule is used, there is zero smoothing of the
    factors. Defaults to \code{cv.ls}.  }

  \item{bwscaling}{ a logical value that when set to \code{TRUE} the
    supplied bandwidths are interpreted as \sQuote{scale factors}
    (\eqn{c_j}{c[j]}), otherwise when the value is \code{FALSE} they are
    interpreted as \sQuote{raw bandwidths} (\eqn{h_j}{h[j]} for continuous data
    types, \eqn{\lambda_j}{lambda[j]} for discrete data types). For
    continuous data types, \eqn{c_j}{c[j]} and \eqn{h_j}{h[j]} are
    related by the formula \eqn{h_j = c_j \sigma_j n^{-1/(2P+l)}}{h[j] =
    c[j]*sigma[j]*n^(-1/(2*P+l))}, where \eqn{\sigma_j}{sigma[j]} is an
    adaptive measure of spread of continuous variable \eqn{j} defined as
    min(standard deviation, mean absolute deviation/1.4826, interquartile
    range/1.349), \eqn{n} the number of observations, \eqn{P}{P} the
    order of the kernel, and \eqn{l}{l} the number of continuous
    variables. For discrete data types, \eqn{c_j}{c[j]} and
    \eqn{h_j}{h[j]} are related by the formula \eqn{h_j =
    c_jn^{-2/(2P+l)}}{h[j] = c[j]*n^(-2/(2*P+l))}, where here
    \eqn{j}{[j]} denotes discrete variable \eqn{j}{j}.  Defaults to
    \code{FALSE}.  }

  \item{bwtype}{
    character string used for the continuous variable bandwidth type,
    specifying the type of bandwidth to compute and return in the
    \code{condbandwidth} object. Defaults to \code{fixed}. Option
    summary:\cr
    \code{fixed}: compute fixed bandwidths \cr
    \code{generalized_nn}: compute generalized nearest neighbors \cr
    \code{adaptive_nn}: compute adaptive nearest neighbors
  }


\subsection{Categorical Search Initialization}{
These controls set categorical search starts and categorical direction-set initialization.
}

  \item{dfac.dir}{stretch factor for direction set search for Powell's algorithm for categorical variables. See Details}

  \item{dfac.init}{non-random initial values for scale factors for categorical variables for Powell's algorithm. See Details}

  \item{hbd.dir}{upper bound for direction set search for Powell's algorithm for categorical variables. See Details}

  \item{hbd.init}{upper bound for scale factors for categorical variables for Powell's algorithm. See Details}

  \item{initd.dir}{initial non-random values for direction set search for Powell's algorithm for categorical variables. See Details}

  \item{lbd.dir}{lower bound for direction set search for Powell's algorithm for categorical variables. See Details}

  \item{lbd.init}{lower bound for scale factors for categorical variables for Powell's algorithm. See Details}

  \item{scale.init.categorical.sample}{ a logical value that when set
    to \code{TRUE} scales \code{lbd.dir}, \code{hbd.dir},
    \code{dfac.dir}, and \code{initd.dir} by \eqn{n^{-2/(2P+l)}},
    \eqn{n} the number of observations, \eqn{P}{P} the order of the
    kernel, and \eqn{l}{l} the number of \code{numeric} variables. See
    Details}


\subsection{Continuous Direction-Set Search Controls}{
These controls set Powell direction-set initialization for continuous variables.
}

  \item{cfac.dir}{stretch factor for direction set search for Powell's algorithm for \code{numeric} variables. See Details}

  \item{dfc.dir}{chi-square degrees of freedom for direction set search for Powell's algorithm for \code{numeric} variables. See Details}

  \item{initc.dir}{initial non-random values for direction set search for Powell's algorithm for \code{numeric} variables. See Details}

  \item{lbc.dir}{lower bound for direction set search for Powell's algorithm for \code{numeric} variables. See Details}


\subsection{Continuous Kernel Support Controls}{
These controls choose and parameterize bounded support for continuous kernels.
}

  \item{cxkerbound}{
    character string controlling continuous-kernel support handling for
    \code{xdat}. Can be set as \code{none} (default kernel on full support),
    \code{range} (use sample min/max), or \code{fixed} (use
    \code{cxkerlb}/\code{cxkerub}). The bounded-kernel route reuses the selected continuous kernel and renormalizes it on the chosen support; see \code{\link{np.kernels}}.
  }

  \item{cxkerlb}{
    numeric scalar/vector of lower bounds for continuous \code{xdat}
    variables used when \code{cxkerbound="fixed"}. Must satisfy
    lower-bound validity for each variable (e.g.,
    \code{<= min(variable)}). Use \code{-Inf} for unbounded below. See \code{\link{np.kernels}} for bounded-kernel normalization details.
  }

  \item{cxkerub}{
    numeric scalar/vector of upper bounds for continuous \code{xdat}
    variables used when \code{cxkerbound="fixed"}. Must satisfy
    upper-bound validity for each variable (e.g.,
    \code{>= max(variable)}). Use \code{Inf} for unbounded above. See \code{\link{np.kernels}} for bounded-kernel normalization details.
  }

  \item{cykerbound}{
    character string controlling continuous-kernel support handling for
    \code{ydat}. Can be set as \code{none} (default kernel on full support),
    \code{range} (use sample min/max), or \code{fixed} (use
    \code{cykerlb}/\code{cykerub}). The bounded-kernel route reuses the selected continuous kernel and renormalizes it on the chosen support; see \code{\link{np.kernels}}.
  }

  \item{cykerlb}{
    numeric scalar/vector of lower bounds for continuous \code{ydat}
    variables used when \code{cykerbound="fixed"}. Must satisfy
    lower-bound validity for each variable (e.g.,
    \code{<= min(variable)}). Use \code{-Inf} for unbounded below. See \code{\link{np.kernels}} for bounded-kernel normalization details.
  }

  \item{cykerub}{
    numeric scalar/vector of upper bounds for continuous \code{ydat}
    variables used when \code{cykerbound="fixed"}. Must satisfy
    upper-bound validity for each variable (e.g.,
    \code{>= max(variable)}). Use \code{Inf} for unbounded above. See \code{\link{np.kernels}} for bounded-kernel normalization details.
  }


\subsection{Continuous Scale-Factor Search Initialization}{
These controls define deterministic and random continuous scale-factor starts and the lower admissibility floor for fixed-bandwidth search.
}

  \item{scale.factor.init}{
    deterministic initial scale factor for continuous fixed-bandwidth
    search. Defaults to \code{0.5}. The value supplied by the user is
    not rewritten, but the effective first start passed to the optimizer
    is \code{max(scale.factor.init, scale.factor.search.lower)}. See
    Details.
  }
  \item{scale.factor.init.lower}{
    lower endpoint for random continuous scale-factor starts. Defaults
    to \code{0.1}. The value supplied by the user is not rewritten, but
    the effective random-start lower endpoint is
    \code{max(scale.factor.init.lower, scale.factor.search.lower)}. See
    Details.
  }
  \item{scale.factor.init.upper}{
    upper endpoint for random continuous scale-factor starts. Defaults
    to \code{2.0}. It must be greater than or equal to the effective
    lower endpoint,
    \code{max(scale.factor.init.lower, scale.factor.search.lower)};
    otherwise bandwidth search errors rather than silently expanding the
    interval. See Details.
  }
  \item{scale.factor.search.lower}{
    optional nonnegative scalar giving the hard lower admissibility
    bound for continuous fixed-bandwidth search candidates. Defaults to
    \code{NULL}. If \code{NULL}, an existing bandwidth object's stored
    value is inherited when available; otherwise the package default
    \code{0.1} is used. This floor applies to computed/search bandwidth
    candidates and to effective search starts only. It does not rewrite
    explicit bandwidths supplied for storage with
    \code{bandwidth.compute = FALSE}. Final fixed-bandwidth search
    candidates must also have a finite valid raw objective value.
  }

\subsection{Distribution Integral And Grid Controls}{
These controls tune the conditional distribution-function integral and grid calculations.
}

  \item{do.full.integral}{
    a logical value which when set as \code{TRUE} evaluates the
    moment-based integral on the entire sample.
  }

  \item{memfac}{
    The algorithm to compute the least-squares objective function uses
    a block-based algorithm to eliminate or minimize redundant kernel
    evaluations. Due to memory, hardware and software constraints, a
    maximum block size must be imposed by the algorithm. This block size
    is roughly equal to memfac*10^5 elements. Empirical tests on
    modern hardware find that a memfac of around 500 performs well. If
    you experience out of memory errors, or strange behaviour for
    large data sets (>100k elements) setting memfac to a lower value may
    fix the problem.
  }

  \item{ngrid}{
    integer number of grid points to use when computing the moment-based
    integral. Defaults to \code{100}.
  }


\subsection{Kernel Type Controls}{
These controls choose continuous, unordered, and ordered kernels for \code{xdat} and \code{ydat}.
}

  \item{cxkerorder}{
    numeric value specifying kernel order for
    \code{xdat} (one of
    \code{(2,4,6,8)}). Kernel order specified along with a
  \code{uniform} continuous kernel type will be ignored. Defaults to
  \code{2}.
  }

  \item{cxkertype}{
    character string used to specify the continuous kernel type for
    \code{xdat}.  Can be set as \code{gaussian},
    \code{epanechnikov}, or \code{uniform}. Defaults to \code{gaussian}.
  }

  \item{cykerorder}{
    numeric value specifying kernel order for
    \code{ydat} (one of
    \code{(2,4,6,8)}). Kernel order specified along with a
  \code{uniform} continuous kernel type will be ignored. Defaults to
  \code{2}.
  }

  \item{cykertype}{
    character string used to specify the continuous kernel type for
    \code{ydat}.
    Can be set as \code{gaussian}, \code{epanechnikov}, or
    \code{uniform}. Defaults to \code{gaussian}.
  }

  \item{oxkertype}{
    character string used to specify the ordered categorical kernel
    type for \code{xdat}. Can be set as \code{wangvanryzin},
    \code{liracine}, or \code{racineliyan}. Defaults to
    \code{liracine}.
  }
  \item{oykertype}{
    character string used to specify the ordered categorical kernel
    type for \code{ydat}. Can be set as \code{wangvanryzin},
    \code{liracine}, or \code{racineliyan}. Defaults to
    \code{liracine}.
  }
  \item{uxkertype}{
    character string used to specify the unordered categorical kernel
    type for \code{xdat}. Can be set as \code{aitchisonaitken} or
    \code{liracine}. Defaults to \code{aitchisonaitken}.
  }

\subsection{Local-Polynomial Model Specification}{
These arguments control the local-polynomial estimator, basis, and fixed degree specification.
}

  \item{basis}{
    character string specifying the polynomial basis used when
    \code{regtype="lp"}. Options are \code{"glp"},
    \code{"additive"}, and \code{"tensor"}.
  }

  \item{bernstein.basis}{
    logical value controlling Bernstein basis evaluation for
    \code{regtype="lp"}. When automatic degree search is requested and
    \code{bernstein.basis} is not explicitly supplied, the search route
    defaults to \code{TRUE} for numerical stability. Explicit
    \code{bernstein.basis=FALSE} is honored, but raw-polynomial search
    can be poorly conditioned at higher degrees.
  }

  \item{degree}{
    integer scalar or integer vector of polynomial degrees for
    continuous \code{xdat} variables when \code{regtype="lp"}.
    If scalar, the value is recycled to all continuous \code{xdat}
    variables. With no continuous \code{xdat} predictors,
    \code{regtype="lp"} is only admissible when \code{degree=0}, the
    local-constant equivalent.
  }

  \item{regtype}{
    character string specifying the conditional local method used for
    the \code{xdat} regression weight operator. Options are
    \code{"lc"}, \code{"ll"}, and \code{"lp"}.
    For \code{npc*} methods, \code{"ll"} is implemented via the
    canonical local polynomial engine with \code{degree = 1} and
    \code{basis = "glp"}. \code{"ll"} and positive-degree
    \code{"lp"} require at least one continuous \code{xdat} predictor;
    for categorical-only \code{xdat}, use \code{"lc"} or
    \code{"lp"} with \code{degree=0}. If local-linear \code{cv.ls} search fails
    while using this canonical raw basis, retrying explicitly with
    \code{regtype="lp"}, \code{degree=1}, and
    \code{bernstein.basis=TRUE}, or centering/scaling the continuous
    regressors, can improve numerical conditioning without changing package
    defaults or invoking an automatic fallback.
  }


\subsection{NOMAD Search Controls}{
These arguments control the optional NOMAD direct-search route for local-polynomial degree and bandwidth search.
}

  \item{nomad}{
    logical shortcut for the recommended automatic local-polynomial
    NOMAD route. When \code{TRUE}, any missing values among
    \code{regtype}, \code{search.engine}, \code{degree.select},
    \code{bernstein.basis}, \code{degree.min}, \code{degree.max},
    \code{degree.verify}, and \code{bwtype} are filled with
    \code{regtype="lp"}, \code{search.engine="nomad+powell"},
    \code{degree.select="coordinate"},
    \code{bernstein.basis=TRUE}, \code{degree.min=0L},
    \code{degree.max=10L}, \code{degree.verify=FALSE}, and
    \code{bwtype="fixed"}. Explicit incompatible settings error
    immediately; in particular, \code{nomad=TRUE} currently requires
    \code{regtype="lp"}, \code{bwtype="fixed"}, automatic degree
    search, \code{bernstein.basis=TRUE}, no explicit \code{degree},
    at least one continuous \code{xdat} predictor, and
    \code{search.engine \%in\% c("nomad", "nomad+powell")}.
    This shortcut does not change the meaning of \code{nmulti} or
    \code{nomad.nmulti}: \code{nmulti} remains the outer restart count,
    while \code{nomad.nmulti} controls inner
    \code{crs::snomadr()} multistarts within each outer restart.
    Returned bandwidth objects retain this normalized preset metadata in
    \code{bw$nomad.shortcut} for a returned object \code{bw}; when
    available, \code{nomad.time} and \code{powell.time} record the
    direct-search and Powell-polish timing components.
  }

  \item{nomad.nmulti}{
    non-negative integer controlling the inner
    \code{crs::snomadr()} multistart count used within each outer
    NOMAD restart when \code{regtype="lp"} and automatic degree search
    uses \code{search.engine="nomad"} or \code{"nomad+powell"}.
    Defaults to \code{0L}, which preserves the current one-start-per-
    restart behavior. This does not replace \code{nmulti}:
    \code{nmulti} controls outer restarts, while \code{nomad.nmulti}
    controls inner NOMAD multistarts within each outer restart.
  }

  \item{nomad.remin}{
    logical flag controlling the optional second NOMAD hot start. When
    \code{TRUE}, NOMAD is restarted once from the best full candidate found,
    including both bandwidth and degree coordinates. Defaults to \code{FALSE};
    current simulation evidence favors the one-pass NOMAD default for routine
    use, while leaving this switch available for sensitivity checks.
  }


  \item{search.engine}{
    character string controlling the automatic local-polynomial search
    backend when \code{regtype="lp"} and
    \code{degree.select != "manual"}. \code{"nomad+powell"} (default)
    performs direct joint search over the \code{xdat}-side continuous
    bandwidth coordinates and degree vector using
    \code{crs::snomadr()}, then applies one Powell hot start from the
    NOMAD solution. \code{"nomad"} omits the Powell refinement.
    \code{"cell"} profiles the criterion over the admissible degree
    grid using repeated fixed-degree bandwidth solves. NOMAD-based
    search currently requires \code{bwtype="fixed"},
    \code{degree.verify=FALSE}, and the suggested package \pkg{crs} to
    be installed.
  }


\subsection{Numerical Search And Tolerance Controls}{
These controls set optimizer tolerances, restart behavior, invalid-candidate penalties, memory blocking, and bounded search transformations.
}

 \item{ftol}{
    fractional tolerance on the value of the cross-validation function
    evaluated at located minima (of order the machine precision or
    perhaps slightly larger so as not to be diddled by
    roundoff). Defaults to \code{1.490116e-07}
    (1.0e+01*sqrt(.Machine$double.eps)).
  }

  \item{invalid.penalty}{ a character string specifying the penalty
    used when the optimizer encounters invalid bandwidths.
    \code{"baseline"} returns a finite penalty based on a baseline
    objective; \code{"dbmax"} returns \code{DBL\_MAX}. Defaults to
    \code{"baseline"}.}

  \item{itmax}{
    integer number of iterations before failure in the numerical
    optimization routine. Defaults to \code{10000}.
  }

  \item{nmulti}{
    integer number of times to restart the process of finding extrema of
    the cross-validation function from different (random) initial points
  }

  \item{penalty.multiplier}{ a numeric multiplier applied to the
    baseline penalty when \code{invalid.penalty="baseline"}.
    Defaults to \code{10}.}

  \item{powell.remin}{
    logical flag controlling Powell restart-from-minimum behavior. For
    ordinary fixed-degree Powell-style search, \code{TRUE} restarts the
    local search from the located minimum. For \code{search.engine="nomad+powell"},
    this controls only the final Powell bandwidth-polish step. The default
    is \code{TRUE} for ordinary Powell routes and \code{FALSE} for the
    Powell polish after NOMAD unless explicitly supplied.
  }

  \item{small}{
    a small number used to bracket a minimum (it is hopeless to ask for
    a bracketing interval of width less than sqrt(epsilon) times its
    central value, a fractional width of only about 10-04 (single
    precision) or 3x10-8 (double precision)). Defaults to \code{small
      = 1.490116e-05 (1.0e+03*sqrt(.Machine$double.eps))}.
  }

  \item{tol}{
    tolerance on the position of located minima of the cross-validation
    function (tol should generally be no smaller than the square root of
    your machine's floating point precision). Defaults to \code{
      1.490116e-04 (1.0e+04*sqrt(.Machine$double.eps))}.
  }

  \item{transform.bounds}{ a logical value that when set to \code{TRUE}
    applies an internal transformation that maps the unconstrained
    search to the feasible bandwidth domain. Defaults to
    \code{FALSE}.}


\subsection{Additional Arguments}{
These arguments collect remaining controls passed through S3 methods.
}

  \item{\dots}{
    additional arguments supplied to specify the bandwidth type,
    kernel types, selection methods, and so on, detailed below.
  }

}

\details{

The \code{scale.factor.*} controls are dimensionless search
controls. The package converts scale factors to bandwidths using the
estimator-specific scaling encoded in the bandwidth object, including
kernel order and the number of continuous variables relevant for the
estimator. Users should not pre-multiply these controls by sample-size
or standard-deviation factors.

  \code{scale.factor.init} controls the deterministic first search
  start. \code{scale.factor.init.lower} and
  \code{scale.factor.init.upper} define the random multistart interval.
  \code{scale.factor.search.lower} is the lower admissibility bound for
  continuous fixed-bandwidth search candidates. The effective first
  start is \code{max(scale.factor.init, scale.factor.search.lower)},
  and the effective random-start lower endpoint is
  \code{max(scale.factor.init.lower, scale.factor.search.lower)}.
  \code{scale.factor.init.upper} must be at least that effective lower
  endpoint; the package errors rather than silently expanding the user's
  interval.

  When \code{scale.factor.search.lower} is \code{NULL}, an existing
  bandwidth object's stored floor is inherited when available;
  otherwise the package default \code{0.1} is used. Explicit bandwidths
  supplied for storage with \code{bandwidth.compute = FALSE} are not
  rewritten by the search floor.

  Categorical search-start controls such as \code{dfac.init},
  \code{lbd.init}, and \code{hbd.init} have separate semantics and are
  not affected by \code{scale.factor.search.lower}.

  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}}, \code{\link{plot.np}} for plotting options.

  The bandwidth-selection argument surface is easiest to read by
  decision group. Start with the data and bandwidth inputs
  (\code{xdat}, \code{ydat}, \code{gydat}, \code{bws}, and
  \code{bandwidth.compute}), then choose the bandwidth criterion and
  representation (\code{bwmethod}, \code{bwscaling}, and
  \code{bwtype}). Next choose continuous kernel and support controls
  (\code{cxker*} and \code{cyker*}), categorical kernel controls
  (\code{uxkertype}, \code{oxkertype}, and \code{oykertype}), and
  numerical search controls including \code{nmulti}, tolerances,
  penalties, and the \code{scale.factor.*} search-start and
  admissibility controls. Local-polynomial and NOMAD controls
  (\code{regtype}, \code{basis}, \code{degree*},
  \code{search.engine}, \code{nomad}, \code{nomad.nmulti}, and
  \code{bernstein.basis}) are relevant when using the explicit
  local-polynomial route.

  For S3 plotting help, see \code{\link{plot.np}}. You can list
  available plot methods with \code{methods("plot")}.

  \code{npcdistbw} implements a variety of methods for choosing
  bandwidths for multivariate distributions (\eqn{p+q}-variate) defined
  over a set of possibly continuous and/or discrete (unordered
  \code{xdat}, ordered \code{xdat} and \code{ydat}) data. The approach
  is based on Li and Racine (2004) who employ \sQuote{generalized
  product kernels} that admit a mix of continuous and discrete data
  types.

  The cross-validation methods employ multivariate numerical search
  algorithms. For fixed local-constant/local-linear fits, and for
  local-polynomial fits with \code{degree.select="manual"}, bandwidth
  search uses multidimensional Powell direction-set optimization.

  Bandwidths can (and will) differ for each variable which is, of
  course, desirable.

  Three classes of kernel estimators for the continuous data types are
  available: fixed, adaptive nearest-neighbor, and generalized
  nearest-neighbor. Adaptive nearest-neighbor bandwidths change with
  each sample realization in the set, \eqn{x_i}{x[i]}, when estimating
  the cumulative distribution at the point \eqn{x}. Generalized nearest-neighbor
  bandwidths change with the point at which the cumulative distribution is estimated,
  \eqn{x}. Fixed bandwidths are constant over the support of \eqn{x}.

  \code{npcdistbw} may be invoked \emph{either} with a formula-like
  symbolic 
  description of variables on which bandwidth selection is to be
  performed \emph{or} through a simpler interface whereby data is passed
  directly to the function via the \code{xdat} and \code{ydat}
  parameters. Use of these two interfaces is \bold{mutually exclusive}.

  Data contained in the data frame \code{xdat} may be a mix of
  continuous (default), unordered discrete (to be specified in the data
  frames using \code{\link{factor}}), and ordered discrete (to be
  specified in the data frames using \code{\link{ordered}}).  Data
  contained in the data frame \code{ydat} may be a mix of continuous
  (default) and ordered discrete (to be specified in the data frames
  using \code{\link{ordered}}). Data can be entered in an arbitrary
  order and data types will be detected automatically by the routine
  (see \code{\link{np}} for details).

  Data for which bandwidths are to be estimated may be specified
  symbolically. A typical description has the form \code{dependent data
  ~ explanatory data},
  where \code{dependent data} and \code{explanatory data} are both
  series of variables specified by name, separated by 
  the separation character '+'. For example, \code{ y1 + y2 ~ x1 + x2 }
  specifies that the bandwidths for the joint distribution of variables
  \code{y1} and \code{y2} conditioned on \code{x1} and \code{x2} are to
  be estimated. See below for further examples. 

  A variety of kernels may be specified by the user. Kernels implemented
  for continuous data types include the second, fourth, sixth, and eighth
  order Gaussian and Epanechnikov kernels, and the uniform
  kernel. Unordered discrete data types use a variation on Aitchison and
  Aitken's (1976) kernel, while ordered data types use a variation of the
  Wang and van Ryzin (1981) kernel.

  When \code{regtype="lp"} and \code{degree.select != "manual"},
  \code{npcdistbw} can jointly determine the \code{xdat}-side local
  polynomial degree vector and the fixed bandwidth coordinates entering
  the conditional distribution criterion. With
  \code{search.engine="cell"}, the criterion is profiled over the
  admissible degree grid using cached coordinate-wise or exhaustive
  search. With \code{search.engine="nomad"} or
  \code{"nomad+powell"}, the criterion is optimized directly over the
  joint degree/bandwidth space using \code{crs::snomadr()};
  \code{"nomad+powell"} then performs one Powell hot start from the
  NOMAD solution and keeps the better of the direct NOMAD and polished
  answers. This polynomial-adaptive joint-search route is motivated by
  Hall and Racine (2015) together with Li, Li, and Racine (under
  revision). When \code{bernstein.basis} is not explicitly supplied,
  the automatic search route defaults to \code{bernstein.basis=TRUE}
  for numerical stability.

  Setting \code{nomad=TRUE} is a convenience preset for this automatic
  LP route, not a generic optimizer alias. For conditional distribution
  bandwidth selection it expands any missing values to the equivalent
  long-form call
  \preformatted{
    npcdistbw(...,
              regtype = "lp",
              search.engine = "nomad+powell",
              degree.select = "coordinate",
              bernstein.basis = TRUE,
              degree.min = 0L,
              degree.max = 10L,
              degree.verify = FALSE,
              bwtype = "fixed")
  }
  Compatible explicit tuning arguments are respected. Incompatible
  explicit settings fail fast so the shortcut never silently changes
  user-selected semantics.

  The optimizer invoked for search is Powell's conjugate direction
  method which requires the setting of (non-random) initial values and
  search directions for bandwidths, and, when restarting, random values
  for successive invocations. Bandwidths for \code{numeric} variables
  are scaled by robust measures of spread, the sample size, and the
  number of \code{numeric} variables where appropriate. Two sets of
  parameters for bandwidths for \code{numeric} can be modified, those
  for initial values for the parameters themselves, and those for the
  directions taken (Powell's algorithm does not involve explicit
  computation of the function's gradient). The default values are set by
  considering search performance for a variety of difficult test cases
  and simulated cases. We highly recommend restarting search a large
  number of times to avoid the presence of local minima (achieved by
  modifying \code{nmulti}). Further refinement for difficult cases can
  be achieved by modifying these sets of parameters. However, these
  parameters are intended more for the authors of the package to enable
  \sQuote{tuning} for various methods rather than for the user themselves.

}
\value{

  \code{npcdistbw} returns a \code{condbandwidth} object, with the
  following components:

  \item{xbw}{ bandwidth(s), scale factor(s) or nearest neighbours for the
    explanatory data, \code{xdat} }
  \item{ybw}{ bandwidth(s), scale factor(s) or nearest neighbours for the
    dependent data, \code{ydat} }
  \item{fval}{ objective function value at minimum }

  if \code{bwtype} is set to \code{fixed}, an object containing
  bandwidths (or scale factors if \code{bwscaling = TRUE}) is
  returned. If it is set to \code{generalized_nn} or \code{adaptive_nn},
  then instead the \eqn{k}th nearest neighbors are returned for the
  continuous variables while the discrete kernel bandwidths are returned
  for the discrete variables.

  The functions \code{\link{predict}}, \code{\link{summary}} and \code{\link{plot}} support
  objects of type \code{condbandwidth}.
}

\section{Book And Method Pointers}{
  \code{npcdistbw} selects response and conditioning bandwidths, and
  when requested local-polynomial degree/search metadata, for the
  conditional distribution target
  \eqn{F(y\mid x)=\Pr(Y\le y\mid X=x)} estimated by
  \code{\link{npcdist}} and inverted by \code{\link{npqreg}}.

  For book-length derivations, see Li and Racine (2007), Chapter 6
  \emph{Conditional CDF and Quantile Estimation}, especially Sections 6.1,
  6.2, 6.3, and 6.5, and Chapter 5 \emph{Conditional Density Estimation}.
  The later workflow treatment is Racine (2019), Chapter 4.
}

\references{

  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
  discrimination by the kernel method,} Biometrika, 63, 413-420.

  Hall, P. and J.S. Racine and Q. Li (2004), \dQuote{Cross-validation and
  the estimation of conditional probability densities,} Journal of the
  American Statistical Association, 99, 1015-1026.

  Hall, P. and J.S. Racine (2015), \dQuote{Infinite Order Cross-Validated
  Local Polynomial Regression,} Journal of Econometrics, 185, 510-525.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
  and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2008), \dQuote{Nonparametric estimation of
  conditional CDF and quantile functions with mixed categorical and
  continuous data,} Journal of Business and Economic Statistics, 26,
  423-434.
  
  Li, Q. and J. Lin and J.S. Racine (2013), \dQuote{Optimal bandwidth
  selection for nonparametric conditional distribution and quantile
  functions}, Journal of Business and Economic Statistics, 31, 57-65.

  Li, A. and Q. Li and J.S. Racine (under revision), \dQuote{Boundary
  Adjusted, Polynomial Adaptive, Nonparametric Kernel Conditional
  Density Estimation,} Econometric Reviews.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,}
  Cambridge University Press.

  Scott, D.W. (1992), \emph{Multivariate Density Estimation. Theory,
  Practice and Visualization,} New York: Wiley.

  Silverman, B.W. (1986), \emph{Density Estimation,} London: Chapman and
  Hall.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth estimators
  for discrete distributions,}  Biometrika, 68, 301-309.

}
\author{
  Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
  \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.

  Caution: multivariate data-driven bandwidth selection methods are, by
  their nature, \emph{computationally intensive}. Virtually all methods
  require dropping the \eqn{i}th observation from the data set, computing an
  object, repeating this for all observations in the sample, then
  averaging each of these leave-one-out estimates for a \emph{given}
  value of the bandwidth vector, and only then repeating this a large
  number of times in order to conduct multivariate numerical
  minimization/maximization. Furthermore, due to the potential for local
  minima/maxima, \emph{restarting this procedure a large number of times may
  often be necessary}. This can be frustrating for users possessing
  large datasets. For exploratory purposes, you may wish to override the
  default search tolerances, say, setting ftol=.01 and tol=.01 and
  conduct multistarting (the default is to restart min(2, ncol(xdat,ydat))
  times) as is done for a number of examples. Once the procedure
  terminates, you can restart search with default tolerances using those
  bandwidths obtained from the less rigorous search (i.e., set
  \code{bws=bw} on subsequent calls to this routine where \code{bw} is
  the initial bandwidth object).  A version of this package using the
  \code{Rmpi} wrapper is under development that allows one to deploy
  this software in a clustered computing environment to facilitate
  computation involving large datasets.

}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}, \code{\link{plot.np}}
  \code{\link{bw.nrd}}, \code{\link{bw.SJ}}, \code{\link{hist}},
  \code{\link{npudens}}, \code{\link{npudist}}
}

\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): For this example, we compute the
# cross-validated bandwidths (default) using a second-order Gaussian
# kernel (default). Note - this may take a minute or two depending on
# the speed of your computer.

data("Italy")
Italy <- Italy[seq_len(min(300, nrow(Italy))), ]
with(Italy, {

bw <- npcdistbw(formula=gdp~ordered(year), nmulti=1)

# The object bw can be used for further estimation using
# npcdist(), plotting using plot() etc. Entering the name of
# the object provides useful summary information, and names() will also
# provide useful information.

summary(bw)

# Note - see the example for npudensbw() for multiple illustrations
# of how to change the kernel function, kernel order, and so forth.

})

# EXAMPLE 1 (INTERFACE=DATA FRAME): For this example, we compute the
# cross-validated bandwidths (default) using a second-order Gaussian
# kernel (default). Note - this may take a minute or two depending on
# the speed of your computer.

data("Italy")
Italy <- Italy[seq_len(min(300, nrow(Italy))), ]
with(Italy, {

bw <- npcdistbw(xdat=ordered(year), ydat=gdp, nmulti=1)

# The object bw can be used for further estimation using npcdist(),
# plotting using plot() etc. Entering the name of the object provides
# useful summary information, and names() will also provide useful
# information.

summary(bw)

# Note - see the example for npudensbw() for multiple illustrations
# of how to change the kernel function, kernel order, and so forth.

})
} % enddontrun
}
\keyword{nonparametric}

%
\name{npdeneqtest}
\alias{npdeneqtest}
\title{ Kernel Consistent Density Equality Test with Mixed Data Types }
\description{
  \code{npdeneqtest} implements a consistent integrated squared
  difference test for equality of densities as described in Li, Maasoumi,
  and Racine (2009).
}
\usage{
npdeneqtest(x = NULL,
            y = NULL,
            bw.x = NULL,
            bw.y = NULL,
            boot.num = 399,
            random.seed = 42,
            \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the two samples and any supplied bandwidths.
}

  \item{bw.x,bw.y}{
    optional bandwidth objects for \code{x,y}
  }


  \item{x,y}{
    data frames for the two samples for which one wishes to
    test equality of densities. The variables in each data
    frame must be the same (i.e. have identical names).
  }


\subsection{Bootstrap Controls}{
These arguments control bootstrap replication and reproducibility settings.
}

  \item{boot.num}{
    an integer value specifying the number of bootstrap
    replications to use. Defaults to \code{399}.
  }


  \item{random.seed}{
    an integer used to seed R's random number generator. This is to
    ensure replicability. Defaults to 42.
  }


\subsection{Additional Arguments}{
Further arguments are passed to the bandwidth-selection routines used by the test.
}

  \item{\dots}{ additional arguments supplied to specify the bandwidth
    type, kernel types, and so on.  This is used if you do not pass in
    bandwidth objects and you do not desire the default behaviours. To
    do this, you may specify any of \code{bwscaling}, \code{bwtype},
    \code{ckertype}, \code{ckerorder}, \code{ukertype},
    \code{okertype}.}


}
\value{
  \code{npdeneqtest} returns an object of type \code{deneqtest} with the
  following components

  \item{Tn}{ the (standardized) statistic \code{Tn} }
  \item{In}{ the (unstandardized) statistic \code{In} }
  \item{Tn.bootstrap}{ contains the bootstrap replications of \code{Tn} }
  \item{In.bootstrap}{ contains the bootstrap replications of \code{In} }
  \item{Tn.P}{ the P-value of the \code{Tn} statistic }
  \item{In.P}{ the P-value of the \code{In} statistic }
  \item{boot.num}{ number of bootstrap replications }

  \code{\link{summary}} supports object of type \code{deneqtest}.
}
\section{Book And Method Pointers}{
  \code{npdeneqtest} compares two distributions by estimating the
  integrated squared difference between their density or probability
  functions. The two samples must have matching variables and data
  types because the statistic compares like with like.

  For book-length background, see Li and Racine (2007), Chapter 12
  \emph{Model Specification Tests}, especially Section 12.2, and Chapter
  13 \emph{Nonsmoothing Tests}, especially Section 13.2. For density and
  mixed-data density context, see Racine (2019), Chapters 2 and 3.
}

\references{
  Li, Q. and E. Maasoumi and J.S. Racine (2009), \dQuote{A Nonparametric
    Test for Equality of Distributions with Mixed Categorical and
    Continuous Data,} Journal of Econometrics, 148, pp 186-200.
}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.


  \code{npdeneqtest} computes the integrated squared density difference
  between the estimated densities/probabilities of two samples having
  identical variables/datatypes. See Li, Maasoumi, and Racine (2009) for
  details.

}

\section{Usage Issues}{

  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.

  It is crucial that both data frames have the same variable names.

}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}
  \code{\link{npdeptest},\link{npsdeptest},\link{npsymtest},\link{npunitest}}
}

\examples{
\dontrun{
set.seed(1234)

## Distributions are equal

n <- 250

sample.A <- data.frame(x=rnorm(n))
sample.B <- data.frame(x=rnorm(n))

npdeneqtest(sample.A,sample.B,boot.num=99)

if (interactive()) Sys.sleep(5)

## Distributions are unequal

sample.A <- data.frame(x=rnorm(n))
sample.B <- data.frame(x=rchisq(n,df=5))

npdeneqtest(sample.A,sample.B,boot.num=99)

## Mixed datatypes, distributions are equal

sample.A <- data.frame(a=rnorm(n),b=factor(rbinom(n,2,.5)))
sample.B <- data.frame(a=rnorm(n),b=factor(rbinom(n,2,.5)))

npdeneqtest(sample.A,sample.B,boot.num=99)

if (interactive()) Sys.sleep(5)

## Mixed datatypes, distributions are unequal

sample.A <- data.frame(a=rnorm(n),b=factor(rbinom(n,2,.5)))
sample.B <- data.frame(a=rnorm(n,sd=10),b=factor(rbinom(n,2,.25)))

npdeneqtest(sample.A,sample.B,boot.num=99)
} % enddontrun
}

\keyword{ nonparametric }

\name{npunitest}
\alias{npunitest}
\title{ Kernel Consistent Univariate Density Equality Test with Mixed Data Types}
\description{
  \code{npunitest} implements the consistent metric entropy test of
  Maasoumi and Racine (2002) for two arbitrary, stationary
  univariate nonparametric densities on common support.
}
\usage{
npunitest(data.x = NULL,
          data.y = NULL,
          method = c("integration","summation"),
          bootstrap = TRUE,
          boot.num = 399,
          bw.x = NULL,
          bw.y = NULL,
          random.seed = 42,
          \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the two samples, statistic variant, and any supplied bandwidths.
}

  \item{bw.x, bw.y}{
    numeric (scalar) bandwidths. Defaults to plug-in (see details below).
  }


  \item{data.x, data.y}{
    common support univariate vectors containing the variables.
  }


  \item{method}{
    a character string used to specify whether to compute
    the integral version or the summation version of the statistic. Can
    be set as \code{integration} or \code{summation}. Defaults to
    \code{integration}. See \sQuote{Details} below for important information
    regarding the use of \code{summation} when \code{data.x} and
    \code{data.y} lack common support and/or are sparse.
  }


\subsection{Bootstrap Controls}{
These arguments control bootstrap execution and reproducibility settings.
}

  \item{boot.num}{
    an integer value specifying the number of bootstrap
    replications to use. Defaults to \code{399}.
  }


  \item{bootstrap}{
    a logical value which specifies whether to conduct the bootstrap
    test or not. If set to \code{FALSE}, only the statistic will be
    computed. Defaults to \code{TRUE}.
  }


  \item{random.seed}{
    an integer used to seed R's random number generator. This is to
    ensure replicability. Defaults to 42.
  }


\subsection{Additional Arguments}{
Further arguments are passed to the bandwidth-selection routines used by the test.
}

  \item{\dots}{ additional arguments supplied to specify the bandwidth
    type, kernel types, and so on.  This is used since we specify bw as
    a numeric scalar and not a \code{bandwidth} object, and is of
    interest if you do not desire the default behaviours. To change the
    defaults, you may specify any of \code{bwscaling}, \code{bwtype},
    \code{ckertype}, \code{ckerorder}, \code{ukertype},
    \code{okertype}.}


}
\value{
  \code{npunitest} returns an object of type \code{unitest} with the
  following components

  \item{Srho}{ the statistic \code{Srho} }
  \item{Srho.bootstrap}{ contains the bootstrap replications of \code{Srho} }
  \item{P}{ the P-value of the statistic }
  \item{boot.num}{ number of bootstrap replications }
  \item{bw.x, bw.y}{ scalar bandwidths for \code{data.x, data.y} }

  \code{\link{summary}} supports object of type \code{unitest}.
}
\section{Book And Method Pointers}{
  \code{npunitest} uses an entropy-distance measure to compare
  univariate density or probability functions. The statistic is based on
  the normalized Hellinger-type distance described in the entropy-test
  literature referenced below.

  For book-length background, see Racine (2019), Chapter 2
  \emph{Continuous Density and Cumulative Distribution Functions},
  especially the entropy and density-test material. For related density
  and testing background, see Li and Racine (2007), Chapters 1, 12, and
  13.
}

\references{

  Granger, C.W. and E. Maasoumi and J.S. Racine (2004), \dQuote{A
  dependence metric for possibly nonlinear processes}, Journal of Time
  Series Analysis, 25, 649-669.

  Maasoumi, E. and J.S. Racine (2002), \dQuote{Entropy and
    predictability of stock market returns,} Journal of Econometrics,
    107, 2, pp 291-312.

}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.


  \code{npunitest} computes the nonparametric metric entropy (normalized
   Hellinger of Granger, Maasoumi and Racine (2004)) for testing
   equality of two univariate density/probability functions,
   \eqn{D[f(x), f(y)]}{D[f(x), f(y)]}. See Maasoumi and Racine (2002)
   for details. Default bandwidths are of the plug-in variety
   (\code{\link{bw.SJ}} for continuous variables and direct plug-in for
   discrete variables). The bootstrap is conducted via simple resampling
   with replacement from the pooled \code{data.x} and \code{data.y}
   (\code{data.x} only for \code{summation}).

   The summation version of this statistic can be numerically unstable
   when \code{data.x} and \code{data.y} lack common support or when the
   overlap is sparse (the summation version involves division of
   densities while the integration version involves differences, and the
   statistic in such cases can be reported as exactly 0.5 or 0). Warning
   messages are produced when this occurs (\sQuote{integration recommended})
   and should be heeded.

   Numerical integration can occasionally fail when the \code{data.x}
   and \code{data.y} distributions lack common support and/or lie an
   extremely large distance from one another (the statistic in such
   cases will be reported as exactly 0.5 or 0). However, in these
   extreme cases, simple tests will reveal the obvious differences in
   the distributions and entropy-based tests for equality will be
   clearly unnecessary.

}

\section{Usage Issues}{

  See the example below for proper usage.

}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}
  \code{\link{npdeneqtest},\link{npdeptest},\link{npsdeptest},\link{npsymtest}}
}

\examples{
\dontrun{
set.seed(1234)
n <- 1000

## Compute the statistic only for data drawn from same distribution

x <- rnorm(n)
y <- rnorm(n)

npunitest(x,y,bootstrap=FALSE)

if (interactive()) Sys.sleep(5)

## Conduct the test for this data

npunitest(x,y,boot.num=99)

if (interactive()) Sys.sleep(5)

## Conduct the test for data drawn from different distributions having
## the same mean and variance

x <- rchisq(n,df=5)
y <- rnorm(n,mean=5,sd=sqrt(10))
mean(x)
mean(y)
sd(x)
sd(y)

npunitest(x,y,boot.num=99)

if (interactive()) Sys.sleep(5)

## Two sample t-test for equality of means
t.test(x,y)
## F test for equality of variances and asymptotic
## critical values
Fstat <- var(x)/var(y)
qf(c(0.025,0.975),df1=n-1,df2=n-1)

## Plot the nonparametric density estimates on the same axes

fx <- density(x)
fy <- density(y)
xlim <- c(min(fx$x,fy$x),max(fx$x,fy$x))
ylim <- c(min(fx$y,fy$y),max(fx$y,fy$y))
if (interactive()) plot(fx,xlim=xlim,ylim=ylim,xlab="Data",main="f(x), f(y)")
lines(fy$x,fy$y,col="red")

if (interactive()) Sys.sleep(5)

## Test for equality of log(wage) distributions

data(wage1)
with(wage1, {
lwage.male <- lwage[female=="Male"]
lwage.female <- lwage[female=="Female"]

npunitest(lwage.male,lwage.female,boot.num=99)

if (interactive()) Sys.sleep(5)

## Plot the nonparametric density estimates on the same axes

f.m <- density(lwage.male)
f.f <- density(lwage.female)
xlim <- c(min(f.m$x,f.f$x),max(f.m$x,f.f$x))
ylim <- c(min(f.m$y,f.f$y),max(f.m$y,f.f$y))
plot(f.m,xlim=xlim,ylim=ylim,
     xlab="log(wage)",
     main="Male/Female log(wage) Distributions")
lines(f.f$x,f.f$y,col="red",lty=2)
rug(lwage.male)
legend(-1,1.2,c("Male","Female"),lty=c(1,2),col=c("black","red"))

})

if (interactive()) Sys.sleep(5)

## Conduct the test for data drawn from different discrete probability
## distributions

x <- factor(rbinom(n,2,.5))
y <- factor(rbinom(n,2,.1))

npunitest(x,y,boot.num=99)
} % enddontrun
}
\keyword{ univar }
\keyword{ nonparametric }

\name{Engel95}
\docType{data}
\alias{Engel95}
\title{ 1995 British Family Expenditure Survey  }
\description{
British cross-section data consisting of a random sample taken from
the British Family Expenditure Survey for 1995. The households consist
of married couples with an employed head-of-household between the ages
of 25 and 55 years. There are 1655 household-level observations in
total.
}
\usage{data("Engel95")}
\format{ A data frame with 10 columns, and 1655 rows.
  \describe{
 \item{food}{ expenditure share on food, of type \code{numeric}}
 \item{catering}{ expenditure share on catering, of type \code{numeric}}
 \item{alcohol}{ expenditure share on alcohol, of type \code{numeric}}
 \item{fuel}{ expenditure share on fuel, of type \code{numeric}}
 \item{motor}{ expenditure share on motor, of type \code{numeric}}
 \item{fares}{ expenditure share on fares, of type \code{numeric}}
 \item{leisure}{ expenditure share on leisure, of type \code{numeric}}
 \item{logexp}{ logarithm of total expenditure, of type \code{numeric}}
 \item{logwages}{ logarithm of total earnings, of type \code{numeric}}
 \item{nkids}{ number of children, of type \code{numeric}}  
  }
}
\source{ Richard Blundell and Dennis Kristensen }
\references{

  Blundell, R. and X. Chen and D. Kristensen (2007),
  \dQuote{Semi-Nonparametric IV Estimation of Shape-Invariant Engel
  Curves,} Econometrica, 75, 1613-1669.
    
  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics:
  Theory and Practice,} Princeton University Press.
  
}

\keyword{datasets}

\examples{
\dontrun{
## Example - compute nonparametric instrumental regression using
## Landweber-Fridman iteration of Fredholm integral equations of the
## first kind.

## We consider an equation with an endogenous regressor (`z') and an
## instrument (`w'). Let y = phi(z) + u where phi(z) is the function of
## interest. Here E(u|z) is not zero hence the conditional mean E(y|z)
## does not coincide with the function of interest, but if there exists
## an instrument w such that E(u|w) = 0, then we can recover the
## function of interest by solving an ill-posed inverse problem.

data(Engel95)

## Sort on logexp (the endogenous regressor) for plotting purposes

Engel95 <- Engel95[order(Engel95$logexp),] 
Engel95 <- Engel95[seq_len(min(120, nrow(Engel95))), ]

with(Engel95, {

model.iv <- npregiv(y=food,
                    z=logexp,
                    w=logwages,
                    method="Landweber-Fridman",
                    iterate.max=3,
                    iterate.diff.tol=1.0e-04)
phi <- model.iv$phi

## Compute the non-IV regression (i.e. regress y on z)

ghat <- npreg(food~logexp,regtype="ll")

## For the plots, restrict focal attention to the bulk of the data
## (i.e. for the plotting area trim out 1/4 of one percent from each
## tail of y and z)

trim <- 0.0025

plot(logexp,food,
     ylab="Food Budget Share",
     xlab="log(Total Expenditure)",
     xlim=quantile(logexp,c(trim,1-trim)),
     ylim=quantile(food,c(trim,1-trim)),
     main="Nonparametric Instrumental Kernel Regression",
     type="p",
     cex=.5,
     col="lightgrey")

lines(logexp,phi,col="blue",lwd=2,lty=2)

lines(logexp,fitted(ghat),col="red",lwd=2,lty=4)

legend(quantile(logexp,trim),quantile(food,1-trim),
       c(expression(paste("Nonparametric IV: ",hat(varphi)(logexp))),
         "Nonparametric Regression: E(food | logexp)"),
       lty=c(2,4),
       col=c("blue","red"),
       lwd=c(2,2))
})
} % enddontrun
}

% 
\name{npksum}
\alias{npksum}
\alias{npksum.default}
\alias{npksum.formula}
\alias{npksum.numeric}
\title{ Kernel Sums with Mixed Data Types }
\description{
  \code{npksum} computes kernel sums on evaluation
  data, given a set of training data, data to be weighted (optional), and a
  bandwidth specification (any bandwidth object).

}
\usage{
npksum(\dots)

\method{npksum}{formula}(formula, 
       data, 
       newdata, 
       subset, 
       na.action, 
       \dots)

\method{npksum}{default}(bws,
       txdat = stop("training data 'txdat' missing"),
       tydat = NULL,
       exdat = NULL,
       weights = NULL,
       bandwidth.divide = FALSE,
       compute.ocg = FALSE,
       compute.score = FALSE,
       kernel.pow = 1.0,
       leave.one.out = FALSE,
       operator = names(ALL_OPERATORS),
       permutation.operator = names(PERMUTATION_OPERATORS),
       return.kernel.weights = FALSE,
       \dots)

\method{npksum}{numeric}(bws,
       txdat = stop("training data 'txdat' missing"),
       tydat,
       exdat,
       weights,
       bandwidth.divide,
       compute.ocg,
       compute.score,
       kernel.pow,
       leave.one.out,
       operator,
       permutation.operator,
       return.kernel.weights,
       \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
Core inputs defining the symbolic interface, training data, optional weighted
response, and evaluation points for the generalized product kernel sum.
}

  \item{bws}{
    a bandwidth specification. This can be set as any suitable bandwidth
    object returned from a bandwidth-generating function, or a numeric vector.
  }

  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(formula)}, typically the environment from which the
    function is called.
  }

  \item{exdat}{
    a \eqn{p}-variate data frame of sum evaluation points (if omitted,
    defaults to the training data itself).
  }

  \item{formula}{
    a symbolic description of variables on which the sum is
    to be performed. The details of constructing a formula are
    described below.
  }

  \item{na.action}{
    a function which indicates what should happen when the data contain
   \code{NA}s. The default is set by the \code{\link{na.action}} setting of options, and is
    \code{\link{na.fail}} if that is unset. The (recommended) default is
    \code{\link{na.omit}}.
  }

  \item{newdata}{
    an optional data frame in which to look for evaluation data. If
    omitted, \code{data} is used.
  }

  \item{subset}{
    an optional vector specifying a subset of observations to be used.
  }
  
  \item{txdat}{
    a \eqn{p}-variate data frame of sample realizations (training data) used to
    compute the sum.
  }
  
  \item{tydat}{
    a numeric vector of data to be weighted. The \eqn{i}th kernel weight
    is applied to the \eqn{i}th element. Defaults to  \code{1}.
  }
  
  \item{weights}{
    a \eqn{n} by \eqn{q}  matrix of weights which can optionally be
    applied to \code{tydat} in the sum. See details.
  }
  
\subsection{Kernel-Sum Operators And Output}{
Controls for bandwidth normalization, kernel powers, leave-one-out
evaluation, operator variants, and returned kernel weights.
}

  \item{bandwidth.divide}{
    a logical specifying whether or not to divide continuous kernel
    weights by their bandwidths. Use this with nearest-neighbor
    methods. Defaults to \code{FALSE}.
  }

  \item{compute.ocg}{
    a logical specifying whether or not to return a separate result for
    each unordered and ordered dimension, where the product kernel term
    for that dimension is evaluated at an appropriate reference
    category. This is used primarily in \code{np} to compute ordered and
    unordered categorical gradients. See details.
  }

  \item{compute.score}{
    a logical specifying whether or not to return the score
    (the \sQuote{grad h} terms) for each dimension in addition to the kernel
    sum. Cannot be \code{TRUE} if a permutation operator other than
    \code{"none"} is selected. 
  }

  \item{kernel.pow}{
    an integer specifying the power to which the kernels will be raised
    in the sum. Defaults to \code{1}.
  }

  \item{leave.one.out}{
    a logical value to specify whether or not to compute the leave one
    out sums. Will not work if \code{exdat} is specified. Defaults to
    \code{FALSE}.
  }

  \item{operator}{
    a string specifying whether the \code{normal}, \code{convolution},
    \code{derivative}, or \code{integral} kernels are to be
    used. Operators scale results by factors of \eqn{h} or \eqn{1/h} where
    appropriate. Defaults to \code{normal} and applies to all elements in a
    multivariate object. See details.
  }

  \item{permutation.operator}{
    a string which can have a value of \code{none}, \code{normal},
    \code{derivative}, or \code{integral}. If set to something other
    than \code{none} (the default), then a separate result will be returned for each
    term in the product kernel, with the operator applied to that term.
    Permutation operators scale results by factors of \eqn{h} or \eqn{1/h} where
    appropriate. This is useful for computing gradients, for example.
  }

  \item{return.kernel.weights}{
    a logical specifying whether or not to return the matrix of
    generalized product kernel weights. Defaults to \code{FALSE}. See
    details.
  }


\subsection{Additional Arguments}{
Further arguments passed to the default kernel-sum method.
}

  \item{\dots}{
    additional arguments supplied to specify the parameters to the
    \code{default} S3 method, which is called during estimation. 
  }
}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.

  \code{npksum}
  exists so that you can create your own kernel objects with
  or without a variable to be weighted (default \eqn{Y=1}). With the options
  available, you could create new nonparametric tests or even new kernel
  estimators. The convolution kernel option would allow you to create,
  say, the least squares cross-validation function for kernel density
  estimation.
  
  \code{npksum} uses highly-optimized C code that strives to minimize
  its \sQuote{memory footprint}, while there is low overhead involved
  when using repeated calls to this function (see, by way of
  illustration, the example below that conducts leave-one-out
  cross-validation for a local constant regression estimator via calls
  to the \code{R} function \code{\link{nlm}}, and compares this to the
  \code{\link{npregbw}} function).

  \code{npksum} implements a variety of methods for computing
  multivariate kernel sums (\eqn{p}-variate) defined over a set of
  possibly continuous and/or discrete (unordered, ordered) data. The
  approach is based on Li and Racine (2003) who employ
  \sQuote{generalized product kernels} that admit a mix of continuous
  and discrete data types.

  Three classes of kernel estimators for the continuous data types are
  available: fixed, adaptive nearest-neighbor, and generalized
  nearest-neighbor. Adaptive nearest-neighbor bandwidths change with
  each sample realization in the set, \eqn{x_i}{x[i]}, when estimating
  the kernel sum at the point \eqn{x}. Generalized nearest-neighbor
  bandwidths change with the point at which the sum is computed,
  \eqn{x}. Fixed bandwidths are constant over the support of \eqn{x}.

  \code{npksum} computes \eqn{\sum_{j=1}^{n}{W_j^\prime Y_j
  K(X_j)}}{sum(t(W[j]) \%*\% Y[j] * K(X[j]))}, where \eqn{W_j}{W[j]}
  represents a row vector extracted from \eqn{W}.  That is, it computes
  the kernel weighted sum of the outer product of the rows of \eqn{W}
  and \eqn{Y}. In the examples, the uses of such sums are illustrated.

  \code{npksum} may be invoked \emph{either} with a formula-like
  symbolic 
  description of variables on which the sum is to be
  performed \emph{or} through a simpler interface whereby data is passed
  directly to the function via the \code{txdat} and \code{tydat}
  parameters. Use of these two interfaces is \bold{mutually exclusive}.

  Data contained in the data frame \code{txdat} (and also \code{exdat})
  may be a mix of continuous (default), unordered discrete (to be
  specified in the data frame \code{txdat} using the
  \code{\link{factor}} command), and ordered discrete (to be specified
  in the data frame \code{txdat} using the \code{\link{ordered}}
  command). Data can be entered in an arbitrary order and data types
  will be detected automatically by the routine (see \code{\link{np}}
  for details).

  Data for which bandwidths are to be estimated may be specified
  symbolically. A typical description has the form \code{dependent data
  ~ explanatory data}, where \code{dependent data} and \code{explanatory
  data} are both series of variables specified by name, separated by the
  separation character '+'. For example, \code{ y1 ~ x1 + x2 } specifies
  that \code{y1} is to be kernel-weighted by \code{x1} and \code{x2}
  throughout the sum. See below for further examples.

  A variety of kernels may be specified by the user. Kernels implemented
  for continuous data types include the second, fourth, sixth, and
  eighth order Gaussian and Epanechnikov kernels, and the uniform
  kernel. Unordered discrete data types use a variation on Aitchison and
  Aitken's (1976) kernel, while ordered data types use a variation of
  the Wang and van Ryzin (1981) kernel (see \code{\link{np}} for
  details).

  The option \code{operator=} can be used to \sQuote{mix and match}
  operator strings to create a \sQuote{hybrid} kernel provided they
  match the dimension of the data. For example, for a two-dimensional
  data frame of \link{numeric} datatypes,
  \code{operator=c("normal","derivative")} will use the normal
  (i.e. PDF) kernel for variable one and the derivative of the PDF
  kernel for variable two. Please note that applying operators will scale the
  results by factors of \eqn{h} or \eqn{1/h} where appropriate.

  The option \code{permutation.operator=} can be used to \sQuote{mix and match}
  operator strings to create a \sQuote{hybrid} kernel, in addition to
  the kernel sum with no operators applied, one for each continuous
  dimension in the data. For example, for a two-dimensional
  data frame of \link{numeric} datatypes,
  \code{permutation.operator=c("derivative")} will return the usual
  kernel sum as if \code{operator = c("normal","normal")} in the
  \code{ksum} member, and in the \code{p.ksum} member, it will return
  kernel sums for \code{operator = c("derivative","normal")}, and
  \code{operator = c("normal","derivative")}. This makes the computation
  of gradients much easier.

  The option \code{compute.score=} can be used to compute the gradients
  with respect to \eqn{h} in addition to the normal kernel sum. Like
  permutations, the additional results are returned in the
  \code{p.ksum}. This option does not work in conjunction with
  \code{permutation.operator}.

  The option \code{compute.ocg=} works much like \code{permutation.operator},
  but for discrete variables. The kernel is evaluated at a reference
  category in each dimension: for ordered data, the next lowest category
  is selected, except in the case of the lowest category, where the
  second lowest category is selected; for unordered data, the first
  category is selected. These additional data are returned in the
  \code{p.ksum} member. This option can be set simultaneously with
  \code{permutation.operator}.
  
  The option \code{return.kernel.weights=TRUE} returns a matrix of
  dimension \sQuote{number of training observations} by \sQuote{number
  of evaluation observations} and contains only the generalized product
  kernel weights ignoring all other objects and options that may be
  provided to \code{npksum} (e.g. \code{bandwidth.divide=TRUE} will be
  ignored, etc.). Summing the columns of the weight matrix and dividing
  by \sQuote{number of training observations} times the product of the
  bandwidths (i.e. \code{\link{colMeans}(foo$kw)/prod(h)}) would produce
  the kernel estimator of a (multivariate) density
  (\code{operator="normal"}) or multivariate cumulative distribution
  (\code{operator="integral"}).

}
\value{
  \code{npksum} returns a \code{npkernelsum} object
  with the following components:

  \item{eval}{ the evaluation points }
  \item{ksum}{ the sum at the evaluation points }
  \item{kw}{ the kernel weights (when \code{return.kernel.weights=TRUE}
  is specified)}
}
\references{
  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{ Multivariate binary
  discrimination by the kernel method,} Biometrika, 63, 413-420.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
  and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2003), \dQuote{Nonparametric estimation of
  distributions with categorical and continuous data,} Journal
  of Multivariate Analysis, 86, 266-292.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,} Cambridge
  University Press. 

  Scott, D.W. (1992), \emph{Multivariate Density Estimation. Theory,
  Practice and Visualization,} New York: Wiley.

  Silverman, B.W. (1986), \emph{Density Estimation,} London: Chapman and
  Hall.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth
  estimators for discrete distributions,} Biometrika, 68, 301-309.  }

\author{
  Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
  \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.
}

\examples{
\dontrun{
# EXAMPLE 1: For this example, we generate 100,000 observations from a
# N(0, 1) distribution, then evaluate the kernel density on a grid of 50
# equally spaced points using the npksum() function, then compare
# results with the (identical) npudens() function output

n <- 100000
x <- rnorm(n)
x.eval <- seq(-4, 4, length=50)

# Compute the bandwidth with the normal-reference rule-of-thumb

bw <- npudensbw(dat=x, bwmethod="normal-reference")

# Compute the univariate kernel density estimate using the 100,000
# training data points evaluated on the 50 evaluation data points, 
# i.e., 1/nh times the sum of the kernel function evaluated at each of
# the 50 points

den.ksum <- npksum(txdat=x, exdat=x.eval, bws=bw$bw)$ksum/(n*bw$bw[1])

# Note that, alternatively (easier perhaps), you could use the
# bandwidth.divide=TRUE argument and drop the *bw$bw[1] term in the
# denominator, as in
# npksum(txdat=x, exdat=x.eval, bws=bw$bw, bandwidth.divide=TRUE)$ksum/n

# Compute the density directly with the npudens() function...

den <- fitted(npudens(tdat=x, edat=x.eval, bws=bw$bw))

# Plot the true DGP, the npksum()/(nh) estimate and (identical)
# npudens() estimate

if (interactive()) plot(x.eval, dnorm(x.eval), xlab="X", ylab="Density", type="l")
points(x.eval, den.ksum, col="blue")
points(x.eval, den, col="red", cex=0.2)
legend(1, .4, 
       c("DGP", "npksum()", 
       "npudens()"), 
       col=c("black", "blue", "red"), 
       lty=c(1, 1, 1))

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# EXAMPLE 2: For this example, we first obtain residuals from a
# parametric regression model, then compute a vector of leave-one-out
# kernel weighted sums of squared residuals where the kernel function is
# raised to the power 2. Note that this returns a vector of kernel
# weighted sums, one for each element of the error vector. Note also
# that you can specify the bandwidth type, kernel function, kernel order
# etc.

data("cps71")
with(cps71, {

error <- residuals(lm(logwage~age+I(age^2)))

bw <- npreg(error~age)

ksum <- npksum(txdat=age, 
               tydat=error^2, 
               bws=bw$bw,
               leave.one.out=TRUE, 
               kernel.pow=2)

ksum

# Obviously, if we wanted the sum of these weighted kernel sums then, 
# trivially, 

sum(ksum$ksum)

})

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Note that weighted leave-one-out sums of squared residuals are used to
# construct consistent model specification tests. In fact, the
# npcmstest() routine in this package is constructed purely from calls
# to npksum(). You can type npcmstest to see the npcmstest()
# code and also examine some examples of the many uses of
# npksum().

# EXAMPLE 3: For this example, we conduct local-constant (i.e., 
# Nadaraya-Watson) kernel regression. We shall use cross-validated
# bandwidths using npregbw() by way of example. Note we extract
# the kernel sum from npksum() via the `$ksum' argument in both
# the numerator and denominator.

data("cps71")
with(cps71, {

bw <- npregbw(xdat=age, ydat=logwage)

fit.lc <- npksum(txdat=age, tydat=logwage, bws=bw$bw)$ksum/
          npksum(txdat=age, bws=bw$bw)$ksum

if (interactive()) plot(age, logwage, xlab="Age", ylab="log(wage)")
lines(age, fit.lc)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# If you wished to compute the kernel sum for a set of evaluation points, 
# you first generate the set of points then feed this to npksum, 
# e.g., for the set (20, 30, ..., 60) we would use

age.seq <- seq(20, 60, 10)

fit.lc <- npksum(txdat=age, exdat=age.seq, tydat=logwage, bws=bw$bw)$ksum/
          npksum(txdat=age, exdat=age.seq, bws=bw$bw)$ksum

# Note that now fit.lc contains the 5 values of the local constant
# estimator corresponding to age.seq...

fit.lc

})

# EXAMPLE 4: For this example, we conduct least-squares cross-validation
# for the local-constant regression estimator. We first write an R
# function `ss' that computes the leave-one-out sum of squares using the
# npksum() function, and then feed this function, along with
# random starting values for the bandwidth vector, to the nlm() routine
# in R (nlm = Non-Linear Minimization). Finally, we compare results with
# the function npregbw() that is written solely in C and calls
# a tightly coupled C-level search routine.  Note that one could make
# repeated calls to nlm() using different starting values for h (highly
# recommended in general).

# Increase the number of digits printed out by default in R and avoid
# using scientific notation for this example (we wish to compare
# objective function minima), then restore the user's options on exit.

old <- options(scipen=100, digits=12)
on.exit(options(old), add = TRUE)

# Generate 100 observations from a simple DGP where one explanatory
# variable is irrelevant.

n <- 100

x1 <- runif(n)
x2 <- rnorm(n)
x3 <- runif(n)

txdat <- data.frame(x1, x2, x3)

# Note - x3 is irrelevant

tydat <- x1 + sin(x2) + rnorm(n)

# Write an R function that returns the average leave-one-out sum of
# squared residuals for the local constant estimator based upon
# npksum(). This function accepts one argument and presumes that
# txdat and tydat have been defined already.

ss <- function(h) {

# Test for valid (non-negative) bandwidths - return infinite penalty
# when this occurs

  if(min(h)<=0) {

    return(.Machine$double.xmax)

  } else {

    mean <-  npksum(txdat, 
                    tydat, 
                    leave.one.out=TRUE, 
                    bandwidth.divide=TRUE, 
                    bws=h)$ksum/
             npksum(txdat, 
                    leave.one.out=TRUE, 
                    bandwidth.divide=TRUE, 
                    bws=h)$ksum
  
    return(sum((tydat-mean)^2)/length(tydat))

  }

}

# Now pass this function to R's nlm() routine along with random starting
# values and place results in `nlm.return'.

nlm.return <- nlm(ss, runif(length(txdat)))

bw <- npregbw(xdat=txdat, ydat=tydat)

# Bandwidths from nlm()

nlm.return$estimate

# Bandwidths from npregbw()

bw$bw

# Function value (minimum) from nlm()

nlm.return$minimum

# Function value (minimum) from npregbw()

bw$fval

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# EXAMPLE 5: For this example, we use npksum() to plot the kernel
# function itself. Our `training data' is the singleton, 0, and our
# evaluation data a grid in [-4,4], while we use a bandwidth of 1. By
# way of example we plot a sixth order Gaussian kernel (note that this
# kernel function can assume negative values)

x <- 0
x.eval <- seq(-4,4,length=500)

kernel <- npksum(txdat=x,exdat=x.eval,bws=1,
                 bandwidth.divide=TRUE,
                 ckertype="gaussian",
                 ckerorder=6)$ksum

plot(x.eval,kernel,type="l",xlab="X",ylab="Kernel Function") 
abline(0,0)
} % enddontrun
}
\keyword{ nonparametric }

\name{wage1}
\docType{data}
\alias{wage1}
\alias{bw.subset}
\alias{bw.all}
\title{ Cross-Sectional Data on Wages }
\description{ Cross-section wage data consisting of a random sample
taken from the U.S. Current Population Survey for the year 1976. There
are 526 observations in total. \code{data("wage1")} makes available the
dataset \code{"wage"} plus additional objects \code{"bw.all"} and
\code{"bw.subset"}.}

\usage{data("wage1")}

\format{ A data frame with 24 columns, and 526 rows.

    Two local-linear \code{rbandwidth} objects (\code{bw.all} and
    \code{bw.subset}) have been computed for the user's convenience
    which can be used to visualize this dataset using
    \code{\link{plot}(bw.all)}
  
\describe{
    \item{wage}{column 1, of type \code{numeric}, average hourly earnings}
    \item{educ}{column 2, of type \code{numeric}, years of education}
    \item{exper}{column 3, of type \code{numeric}, years potential experience}
    \item{tenure}{column 4, of type \code{numeric}, years with current employer}
    \item{nonwhite}{column 5, of type \code{factor}, =\dQuote{Nonwhite} if nonwhite, \dQuote{White} otherwise}
    \item{female}{column 6, of type \code{factor}, =\dQuote{Female} if female, \dQuote{Male} otherwise}
    \item{married}{column 7, of type \code{factor}, =\dQuote{Married} if Married, \dQuote{Nonmarried} otherwise}
    \item{numdep}{column 8, of type \code{numeric}, number of dependants}
    \item{smsa}{column 9, of type \code{numeric}, =1 if live in SMSA}
    \item{northcen}{column 10, of type \code{numeric}, =1 if live in north central U.S}
    \item{south}{column 11, of type \code{numeric}, =1 if live in southern region}
    \item{west}{column 12, of type \code{numeric}, =1 if live in western region}
    \item{construc}{column 13, of type \code{numeric}, =1 if work in construction industry}
    \item{ndurman}{column 14, of type \code{numeric}, =1 if in non-durable manufacturing industry}
    \item{trcommpu}{column 15, of type \code{numeric}, =1 if in transportation, communications, public utility}
    \item{trade}{column 16, of type \code{numeric}, =1 if in wholesale or retail}
    \item{services}{column 17, of type \code{numeric}, =1 if in services industry}
    \item{profserv}{column 18, of type \code{numeric}, =1 if in professional services industry}
    \item{profocc}{column 19, of type \code{numeric}, =1 if in professional occupation}
    \item{clerocc}{column 20, of type \code{numeric}, =1 if in clerical occupation}
    \item{servocc}{column 21, of type \code{numeric}, =1 if in service occupation}
    \item{lwage}{column 22, of type \code{numeric}, log(wage)}
    \item{expersq}{column 23, of type \code{numeric}, exper\eqn{^2}{**2}}
    \item{tenursq}{column 24, of type \code{numeric}, tenure\eqn{^2}{**2}}
  }
}
\source{ Jeffrey M. Wooldridge }
\examples{
data("wage1")
with(wage1, {
summary(wage1)
})
}


\references{

  Wooldridge, J.M. (2000), \emph{Introductory Econometrics: A Modern
  Approach}, South-Western College Publishing.

}

\keyword{datasets}

\name{uocquantile}
\title{Compute Quantiles}
\usage{
uocquantile(x, prob)
}
\alias{uocquantile}
\arguments{
\subsection{Variable And Probability Inputs}{
Variable and requested probability for mixed-type quantile extraction.
}

  \item{x}{an ordered, unordered or continuous variable.}
  \item{prob}{quantile to compute.}
}
\value{
  A quantile computed from \code{x}.
}
\description{
  \code{uocquantile} is a function which computes quantiles of an
  unordered, ordered or continuous variable  \code{x}.
}
\details{
  \code{uocquantile} is a function which computes quantiles of
  an unordered, ordered or continuous variable \code{x}. If \code{x}
  is unordered, the mode is returned. If \code{x} is ordered, the level
  for which the cumulative distribution is >= prob is returned. If
  \code{x} is continuous, \code{\link{quantile}} is invoked and the
  result returned. 
}
\seealso{
  \code{\link{quantile}}
}
\author{
  Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine \email{racinej@mcmaster.ca}
}

\examples{
x <- rbinom(n = 100, size = 10, prob = 0.5)
uocquantile(x, 0.5)
}

\keyword{univar}
\keyword{nonparametric}

%
\name{plot.np}
\alias{plot.np}

\alias{plot.bandwidth}
\alias{plot.conbandwidth}
\alias{plot.plbandwidth}
\alias{plot.rbandwidth}
\alias{plot.scbandwidth}
\alias{plot.sibandwidth}

\title{General Purpose Plotting of Nonparametric Objects}

\description{
  Plotting is provided via \code{\link{plot}} S3 methods, which generate
  plots of
  nonparametric statistical objects such as regressions, quantile
  regressions, partially linear regressions, single-index models,
  densities and distributions, given training data and a bandwidth
  object. \code{plot(...)} is the supported public interface.
}

\usage{
\method{plot}{bandwidth}(x, \dots)

\method{plot}{conbandwidth}(x, \dots)

\method{plot}{plbandwidth}(x, \dots)

\method{plot}{rbandwidth}(x, \dots)

\method{plot}{scbandwidth}(x, \dots)

\method{plot}{sibandwidth}(x, \dots)

}
\arguments{
\subsection{Plot Object}{
This argument identifies the object to plot.
}

  \item{x}{
    an object generated by the \pkg{np} package. The most convenient
    inputs are fitted estimator objects returned by invocations of
    \code{\link{npudens}}, \code{\link{npcdens}}, \code{\link{npreg}},
    \code{\link{npconmode}}, \code{\link{npplreg}}, \code{\link{npindex}}, or
    \code{\link{npscoef}}. Bandwidth objects returned by
    \code{\link{npudensbw}}, \code{\link{npcdensbw}},
    \code{\link{npregbw}}, \code{\link{npplregbw}},
    \code{\link{npindexbw}}, or \code{\link{npscoefbw}} are also
    supported when plotting should construct the fitted object.
  }


\subsection{Plot Options Passed Through \code{\dots}}{
Named options passed via \code{\dots} include:
}

    \strong{Data and evaluation controls}
    \describe{
      \item{\code{data}}{
        an optional \code{\link{data.frame}} used to resolve stored
        formula calls when explicit training data arguments are not
        supplied.
      }
      \item{\code{grid_control}}{
        optional detailed grid controls created by
        \code{\link{np_grid_control}}.
      }
      \item{\code{neval}}{
        the number of evaluation points used for each plotted continuous
        margin.
      }
      \item{\code{xdat}, \code{ydat}, \code{zdat}}{
        training data used when plotting bandwidth objects. \code{xdat}
        is the regressor/conditioning data, \code{ydat} is the response
        or conditional outcome data, and \code{zdat} is the auxiliary
        semiparametric covariate block when relevant.
      }
      \item{\code{xq}, \code{yq}, \code{zq}}{
        quantiles used to hold non-plotted variables fixed when
        constructing slices.
      }
      \item{\code{xtrim}, \code{ytrim}, \code{ztrim}}{
        trimming/expansion controls for the evaluation support.
      }
    }

    \strong{Plot content controls}
    \describe{
      \item{\code{coef}, \code{coef_index}}{
        coefficient-function plotting controls for semiparametric
        families where supported.
      }
      \item{\code{common_scale}}{
        a logical value specifying whether multiple panels should share a
        common plotting scale when feasible.
      }
      \item{\code{gradient_order}, \code{gradients}}{
        derivative plotting controls for families that support
        gradients/derivatives. For \code{npreg} local-polynomial
        objects, \code{gradient_order} requests the continuous-predictor
        derivative order to generate on the plot evaluation grid; if
        omitted the first derivative/effect is plotted. For
        \code{npconmode} objects,
        \code{gradients=TRUE} plots stored class-probability gradients/effects
        for the selected response level, computed by
        \code{npconmode(..., gradients=TRUE, level=...)}.
      }
      \item{\code{level}}{
        response level to plot for class-probability displays when supported,
        notably \code{plot.conmode}. If omitted, the default for
        \code{npconmode} objects is the base/reference response level
        \code{levels(y)[1]}. These plots require probabilities stored by
        \code{npconmode(..., probabilities=TRUE)}. Currently
        \code{plot.conmode} displays stored object-fed probability/effect
        payloads only; probability/effect interval bands and surface grids are
        not yet implemented for this route.
      }
      \item{\code{layout}}{
        \code{"auto"} or \code{TRUE} requests automatic
        \code{par(mfrow=...)} management for multi-panel plots.
        \code{"current"} or \code{FALSE} preserves the user's current
        graphics layout.
      }
      \item{\code{output}}{
        one of \code{"plot"}, \code{"data"}, or \code{"both"}.
        \code{"both"} is the public spelling for the historical
        \code{"plot-data"} behavior.
      }
      \item{\code{proper}, \code{proper_control}, \code{proper_method}}{
        controls for proper conditional density/distribution handling
        where implemented.
      }
      \item{\code{tau}}{
        the quantile index or indices used for quantile-regression-related
        plotting routes when relevant. A vector of \code{tau} values overlays
        the corresponding conditional quantile curves and reuses the stored
        bandwidth object. With \code{errors="asymptotic"} or
        \code{errors="bootstrap"}, intervals are constructed separately for
        each requested \code{tau}; bootstrap work is shared across \code{tau}
        values within each resample. For \code{npqreg} objects whose
        conditional-distribution bandwidths were selected with
        \code{nomad=TRUE}, plotting reuses the stored LP
        \code{regtype.engine} and selected \code{degree.engine} metadata.
      }
    }

    \strong{Interval and bootstrap controls}
    \describe{
      \item{\code{alpha}}{
        the nominal significance level used for interval/band
        construction.
      }
      \item{\code{B}}{
        the bootstrap replication count.
      }
      \item{\code{band}}{
        one of \code{"pmzsd"}, \code{"pointwise"},
        \code{"bonferroni"}, \code{"simultaneous"}, or \code{"all"}.
      }
      \item{\code{boot_control}}{
        optional detailed bootstrap controls created by
        \code{\link{np_boot_control}}. This controls non-fixed bootstrap
        behavior, wild-bootstrap multipliers, and dependent-bootstrap
        block length.
      }
      \item{\code{bootstrap}}{
        the bootstrap resampling scheme. Regression-family plots support
        \code{"wild"}, \code{"inid"}, \code{"fixed"}, and
        \code{"geom"}; density/distribution-family plots support
        \code{"inid"}, \code{"fixed"}, and \code{"geom"}.
      }
      \item{\code{boxplot_outliers}, \code{factor_boxplot}}{
        boxplot-style bootstrap summary controls for factor/categorical
        displays when available.
      }
      \item{\code{center}}{
        one of \code{"estimate"} or \code{"bias-corrected"}.
      }
      \item{\code{errors}}{
        one of \code{"none"}, \code{"bootstrap"}, or
        \code{"asymptotic"}.
      }
    }

    \strong{Rendering and view controls}
    \describe{
      \item{\code{data_overlay}}{
        logical; when \code{TRUE}, overlay raw data on regression,
        quantile-regression, partially linear, and smooth coefficient
        plot surfaces and expand plot limits to include the data. For
        \code{npqreg} level plots the data layer is drawn first by
        default, so quantile curves are overlaid on top; ordered and
        factor conditioning variables use the natural boxplot-style
        display from \code{plot()}. The option is ignored for
        derivative/gradient plots and coefficient plots. Overlay points
        default to small, light markers and can be customized with
        standard \code{points()} arguments passed via \code{...} (for
        example \code{pch}, \code{cex}, \code{col}, and \code{bg}).
        Default is \code{TRUE}.
      }
      \item{\code{legend}}{
        legend control for legends drawn by the package, including
        vector-\code{tau} \code{npqreg} level plots and
        \code{band="all"} uncertainty legends. \code{TRUE} draws the
        default legend,
        \code{FALSE}, \code{NULL}, or \code{NA} suppresses it, a
        single character string supplies the legend position, and a
        list is merged into the default \code{\link{legend}} call for
        custom labels, placement, or styling. For plot routes with
        more than one package-drawn legend, named sublists such as
        \code{legend=list(tau=list(x="topright"), bands=FALSE)} may
        be used when a route distinguishes those legend roles.
      }
      \item{\code{data_rug}}{
        logical; when \code{TRUE}, draw a rug-style support cue for the
        plotted coordinates. For ordinary 1D base plots this behaves
        like a standard rug. For eligible 2D surface plots, the base
        renderer draws short floor-rug ticks and
        \code{renderer="rgl"} draws the analogous 3D floor-rug ticks
        on supported surface routes. In this rollout tranche,
        \code{data_rug=TRUE} is implemented across the current base
        plot families and across supported \code{rgl} surface routes;
        unsupported renderer/geometry combinations still fail clearly
        rather than being ignored silently. On mixed base plot layouts,
        rug marks are drawn for continuous panels and omitted for
        categorical panels.
        This option is distinct from \code{data_overlay}:
        \code{data_rug} shows where observations lie in the plotted
        covariate support, while \code{data_overlay} is the
        regression-only 3D response overlay used on supported surface
        plots. On supported regression surfaces, both options may be
        used together. Default is \code{FALSE}.
      }
      \item{\code{perspective}}{
        a logical value specifying whether bivariate continuous surfaces
        are displayed with \code{\link{persp}}-style plots when relevant.
      }
      \item{\code{phi}, \code{theta}, \code{view}}{
        view controls for bivariate continuous surface plots.
      }
      \item{\code{render_control}}{
        optional detailed rendering controls created by
        \code{\link{np_render_control}}, including band-versus-bar
        uncertainty display controls.
      }
      \item{\code{renderer}}{
        the surface-rendering backend for eligible bivariate continuous
        plots. \code{"base"} uses the existing base R
        \code{\link{persp}} path. \code{"rgl"} is currently a
        fail-closed backend for supported plain 2D surface routes in
        \code{npreg}/\code{npregbw},
        \code{npplreg}/\code{npplregbw},
        \code{npscoef}/\code{npscoefbw},
        \code{npudens}/\code{npudensbw},
        \code{npudist}/\code{npudistbw},
        \code{npcdens}/\code{npcdensbw}, and
        \code{npcdist}/\code{npcdistbw}. In this tranche,
        \code{renderer="rgl"} requires \code{perspective=TRUE},
        \code{view="fixed"} and \code{output="plot"};
        unsupported combinations still error rather than falling back
        silently. The suggested package
        \code{rgl} must be installed when \code{renderer="rgl"} is
        requested. When drawn, the \code{rgl} surface is displayed via
        the active \code{rgl} backend, which on this rollout path is
        typically the RStudio Viewer or a browser/WebGL window rather
        than the base graphics plot pane. When \code{col} is omitted,
        the \code{rgl} surface uses a value-based
        \code{\link[grDevices]{hcl.colors}(..., "viridis")} gradient. When the
        inherited base-view defaults \code{theta=0} and \code{phi=20}
        are left unchanged, the \code{rgl} backend remaps that default
        camera to a more informative oblique view for this rollout
        tranche. For the currently supported \code{npreg},
        \code{npplreg}, and plain mean-surface
        \code{npscoef} surface routes, \code{data_overlay=TRUE}
        adds observed data as an interactive point cloud. For all
        currently supported \code{rgl} surface routes,
        \code{errors != "none"} adds interval surfaces;
        \code{band="all"} draws pointwise,
        simultaneous, and bonferroni band families with distinct,
        colorblind-safer colors when those bands are available for the route.
        Unsupervised and conditional density/distribution surface
        routes do not use observed-data point overlays in this tranche.
        For \code{npscoef}/\code{npscoefbw}, \code{renderer="rgl"}
        currently applies only to the plain mean-surface route;
        coefficient-mode plots remain on the base renderer.
        Common
        \code{rgl} display controls may be supplied
        directly in \code{\dots} on supported routes, including surface
        arguments such as \code{alpha}, \code{back}, and
        \code{front}, and view arguments such as \code{fov} and
        \code{zoom}. Additional backend-specific arguments may be
        passed through using prefixed names in \code{\dots}, namely
        \code{rgl.persp3d.<arg>}, \code{rgl.view3d.<arg>},
        \code{rgl.par3d.<arg>}, \code{rgl.grid3d.<arg>}, and
        \code{rgl.widget.<arg>}, and \code{rgl.legend3d.<arg>}. For
        the new overlay layers,
        \code{rgl.points3d.<arg>} customizes observed-point overlays
        and \code{rgl.surface3d.<arg>} customizes interval surfaces.
        When \code{band="all"} is used on supported
        \code{rgl} routes, a matching interactive legend is drawn, and
        \code{rgl.legend3d.<arg>} may be used to customize it, for
        example \code{rgl.legend3d.bty="o"} to restore a legend box.
        By default, these legends follow the base-graphics uncertainty
        legend convention and are drawn without a surrounding box.
      }
    }

    \strong{Graphical parameters}
    \describe{
      \item{\code{border}, \code{cex.axis}, \code{cex.lab},
      \code{cex.main}, \code{cex.sub}, \code{col}, \code{lty},
      \code{lwd}, \code{main}, \code{sub}, \code{type}, \code{xlab},
      \code{xlim}, \code{ylab}, \code{ylim}, \code{zlab}, \code{zlim}}{
        standard graphical controls forwarded to the underlying plotting
        primitives when relevant.
      }
    }

    \strong{Reproducibility controls}
    \describe{
      \item{\code{random.seed}}{
        an optional random seed used to make bootstrap plotting
        reproducible.
      }
    }

\subsection{Additional Arguments}{
Further plotting controls are passed through \code{\dots} to the relevant plot method.
}

  \item{\dots}{
    additional arguments supplied to control plotting behavior or passed
    through to underlying plotting helpers where supported. Supported named
    plotting controls are summarized above; standard graphical parameters
    are forwarded where relevant.
  }

}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels and
  \code{\link{np.options}} for global options.

  The preferred public interface is \code{\link{plot}} on fitted or
  bandwidth objects (e.g., \code{plot(fit)} or \code{plot(bw)}).

  \code{plot} is a general purpose plotting routine for visually
  exploring objects generated by the \code{np} library, such as
  regressions, quantile regressions, partially linear regressions,
  single-index models, densities and distributions. There is no need to
  call \code{plot} directly: plotting is handled by class-specific S3
  \code{\link{plot}} methods for objects generated by the \pkg{np}
  package.

  Visualizing one and two dimensional datasets is a straightforward
  process. The default behavior of \code{plot} is to generate a
  standard 2D plot to visualize univariate data, and a perspective plot
  for bivariate data. When visualizing higher dimensional data,
  \code{plot} resorts to plotting a series of 1D slices of the
  data. For a slice along dimension \eqn{i}, all other variables at
  indices \eqn{j \ne i}{j != i} are held constant at the quantiles
  specified in the \eqn{j}th element of \code{xq}. The default is the
  median.

  The slice itself is evaluated on a uniformly spaced sequence of
  \eqn{neval} points. The interval of evaluation is determined by the
  training data. The default behavior is to evaluate from
  \code{min(txdat[,i])} to \code{max(txdat[,i])}. The \code{xtrim}
  variable allows for control over this behavior. When \code{xtrim} is
  set, data is evaluated from the \code{xtrim[i]}th quantile of
  \code{txdat[,i]} to the \code{1.0-xtrim[i]}th quantile of
  \code{txdat[,i]}.

  Furthermore, \code{xtrim} can be set to a negative
  value in which case it will expand the limits of the evaluation
  interval beyond the support of the training data, by measuring the
  distance between \code{min(txdat[,i])} and the \code{xtrim[i]}th
  quantile of \code{txdat[,i]}, and extending the support by that
  distance on the lower limit of the interval. \code{plot} uses an
  analogous procedure to extend the upper limit of the interval.

  Plot interval/error types are:
  \preformatted{
  "pmzsd"        : point estimate +/- z_(1-alpha/2) * standard error
  "pointwise"    : per-point two-sided interval from N(0,1) quantiles
  "bonferroni"   : pointwise interval with alpha replaced by alpha/m
  "simultaneous" : joint-band interval (bootstrap route)
  }
  where \eqn{m} is the number of evaluation points used in the plotted
  curve/surface (\eqn{m=\texttt{neval}}{m=neval} for univariate curves,
  typically \eqn{m=\texttt{neval}^2}{m=neval^2} for full 2D perspective
  surfaces).

  For asymptotic intervals, let \eqn{T(x)} denote the plotted functional
  (mean, gradient, density, distribution, etc.) and \eqn{\widehat{se}(x)}
  its asymptotic standard error:
  \eqn{T(x)\pm z_{1-\alpha/2}\widehat{se}(x)} for \code{"pmzsd"} and
  \eqn{[T(x)+z_{\alpha/2}\widehat{se}(x),\ T(x)+z_{1-\alpha/2}\widehat{se}(x)]}
  for \code{"pointwise"}.
  \code{"bonferroni"} applies the same pointwise construction with
  \eqn{\alpha/m} in place of \eqn{\alpha}. For the kernel estimators in
  this package, asymptotic simultaneous bands are not generally
  available, so \code{"simultaneous"} with
  \code{errors="asymptotic"} returns \code{NA} bands.
  Asymptotic standard errors are taken from fitted-object components such as
  \code{merr}, \code{gerr}, \code{derr}, \code{conderr}, and
  \code{congerr} where implemented.

  Bootstrap resampling is conducted pairwise on \eqn{(y,X,Z)} (i.e., by
  resampling rows of \eqn{(y,X)} or \eqn{(y,X,Z)} as appropriate).
  Bootstrap method support differs by estimator family:
  \preformatted{
  Regression-family (npreg/npindex/npscoef/npplreg): wild, inid, fixed, geom
  Density/distribution-family (npudens/npudist/npcdens/npcdist): inid, fixed, geom
  }
  hence \code{"wild"} is only available for regression-family plotting.

  Implementation notes for speed:
  \preformatted{
  wild            : fast np*hat linear-operator bootstrap path
  inid/fixed/geom : fast direct helper path (no internal bandwidth search)
  }
  For non-fixed density/distribution bootstrap, an explicit experimental
  approximation is available via
  \code{boot_control = np_boot_control(nonfixed = "exact")} or
  \code{boot_control = np_boot_control(nonfixed = "frozen")}. The default
  \code{"exact"} route recomputes the non-fixed geometry for each resample.
  The experimental \code{"frozen"} route reuses the original-sample
  non-fixed geometry throughout the bootstrap run. This option is currently
  implemented only for unconditional and conditional density/distribution
  bootstrap routes and remains off by default. For generalized/adaptive
  nearest-neighbor runs, \code{"frozen"} is an approximation that can alter
  interval/band width by holding the original-sample nearest-neighbor
  geometry fixed across bootstrap resamples; \code{"exact"} remains the
  recommended setting for production inference. This approximation can be
  more noticeable for conditional density/distribution plotting than for the
  regression-style plot families because the conditional bootstrap paths
  freeze both numerator and denominator nearest-neighbor geometry before
  recombining them. In practice, conditional distribution bands are often
  closer, while conditional density bands can differ more materially from
  \code{"exact"} under generalized/adaptive nearest-neighbor bandwidths.
  For smooth coefficient plots (\code{npscoef}) under non-fixed bandwidths,
  \code{"exact"} can also be much more expensive than \code{"frozen"} on
  large jobs, because the coefficient field must be recomputed for each
  bootstrap resample rather than reusing the original-sample geometry. This
  recomputation cannot in general be avoided without a more aggressive
  approximation: for \code{npscoef} the local weighted systems that define the
  coefficient vector depend on the bootstrap resample weights/counts at each
  evaluation point, so unlike \code{npplreg} there is no single global
  coefficient vector that can be updated once per draw.
  \code{inid} admits general heteroskedasticity of unknown form, though
  it does not allow for dependence. \code{fixed} conducts Kunsch's (1988)
  block bootstrap for dependent data, while \code{geom} conducts Politis
  and Romano's (1994) stationary bootstrap.

  For local polynomial conditional density/distribution plotting
  (\code{npcdens}/\code{npcdist} with \code{regtype="ll"} or
  \code{regtype="lp"}) and \code{proper=TRUE}, the plotted estimate is
  rendered proper slice-by-slice on the fixed evaluation grid: each
  conditional density slice is projected to be nonnegative and to integrate
  to one using trapezoidal quadrature weights from the evaluation
  \eqn{y}-grid, while each conditional distribution slice is projected to
  be monotone and bounded in \eqn{[0,1]}. When
  \code{errors="bootstrap"}, the bootstrap resample surfaces
  are computed first on that same fixed grid and then properized
  resample-by-resample using the same grid geometry before
  \code{"pointwise"}, \code{"bonferroni"}, \code{"simultaneous"}, and
  \code{"all"} bands are constructed. Thus the bootstrap distribution used
  to form these bands is built from properized resample surfaces. The final
  lower/upper band surfaces are interval envelopes and are not themselves
  separately re-projected to satisfy the density/distribution shape
  constraints.

  For consistency of the block and stationary bootstrap, the (mean)
  block length \eqn{b} should grow with the sample size \eqn{n} at an
  appropriate rate. If \eqn{b} is not given, then a default growth rate
  of \eqn{const \times n^{1/3}}{const n^{1/3}} is used. This rate is
  \dQuote{optimal} under certain conditions (see Politis and Romano
  (1994) for more details). However, in general the growth rate depends on
  the specific properties of the DGP. A default value for \eqn{const}
  (\eqn{3.15}) has been determined by a Monte Carlo simulation using a
  Gaussian AR(1) process (AR(1)-parameter of 0.5, 500
  observations). \eqn{const} has been chosen such that the mean square
  error for the bootstrap estimate of the variance of the empirical mean
  is minimized.

  The default bootstrap replication count is
  \code{B=1999}. For pointwise tails, ensure
  \eqn{B \ge \lceil 2/\alpha - 1 \rceil}{B >= ceiling(2/alpha - 1)} so
  \eqn{\alpha(B+1)} is feasible on the bootstrap rank grid. For interval
  types \code{"bonferroni"}, \code{"simultaneous"}, and \code{"all"},
  the minimum recommended count is
  \eqn{B_{\min}=\lceil 2m/\alpha-1 \rceil}{Bmin=ceiling(2*m/alpha-1)},
  where \eqn{m} is the number of evaluation points used by the plotted
  curve/surface. For full 2D perspective grids this is typically
  \eqn{m=\texttt{neval}^2}{m=neval^2}. When \code{B} is below these
  thresholds, plotting proceeds but warning guidance is reported.

  Typical plotting calls:
  \preformatted{
  ## Asymptotic pointwise/bonferroni intervals
  plot(fit, errors="asymptotic", band="pointwise")
  plot(fit, errors="asymptotic", band="bonferroni")

  ## Regression-family bootstrap (wild available)
  plot(fit, errors="bootstrap", bootstrap="wild")

  ## Density/distribution-family bootstrap (use inid/fixed/geom)
  plot(fit, errors="bootstrap", bootstrap="inid")
  }

  New public plot controls use \code{snake_case}. Dotted names are reserved for
  S3 methods and internal plot-engine plumbing. Use \code{errors},
  \code{bootstrap}, \code{B}, \code{band}, \code{alpha}, \code{center},
  \code{output}, \code{data_overlay}, \code{data_rug}, \code{layout}, and the
  control constructors documented in \code{\link{np_boot_control}},
  \code{\link{np_grid_control}}, and \code{\link{np_render_control}}.

}

\value{
  Setting \code{output} will instruct \code{plot} what data
  to return. Option summary:\cr
  \code{plot}: instruct \code{plot} to just plot the data and
  return \code{NULL} \cr
  \code{both}: instruct \code{plot} to plot the data and return
  the data used to generate the plots. The data will be a \code{list} of
  objects of the appropriate type, with one object per plot. For
  example, invoking \code{plot} on 3D density data will have it
  return a list of three npdensity objects. If biases were calculated,
  they are stored in a component named \code{bias}\cr
  \code{data}: instruct \code{plot} to generate data only and no plots
}

\references{

  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Hall, P. and J.S. Racine and Q. Li (2004), \dQuote{Cross-validation and the
    estimation of conditional probability densities,} Journal of the
  American Statistical Association, 99, 1015-1026.

  Kunsch, H.R. (1989), \dQuote{The jackknife and the bootstrap for
    general stationary observations,} The Annals of Statistics, 17,
  1217-1241.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
  and Practice,} Princeton University Press.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,} Cambridge
  University Press.

  Politis, D.N. and J.P. Romano (1994), \dQuote{The stationary
    bootstrap,} Journal of the American Statistical Association, 89,
  1303-1313.

  Scott, D.W. (1992), \emph{Multivariate Density Estimation. Theory,
  Practice and Visualization,} New York: Wiley.

  Silverman, B.W. (1986), \emph{Density Estimation,} London: Chapman and
  Hall.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth
    estimators for discrete distributions,} Biometrika, 68, 301-309.

}


\author{
  Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
  \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.
}

\examples{
\dontrun{
# EXAMPLE 1: For this example, we load Giovanni Baiocchi's Italian GDP
# panel (see Italy for details), then create a data frame in which year
# is an ordered factor, GDP is continuous, compute bandwidths using
# likelihood cross-validation, then create a grid of data on which the
# density will be evaluated for plotting purposes

data("Italy")
with(Italy, {

data <- data.frame(ordered(year), gdp)

# Compute bandwidths using likelihood cross-validation (default). Note
# that this may take a minute or two depending on the speed of your
# computer...

bw <- npudensbw(dat=data)
fit <- npudens(tdat=data, bws=bw)

# You can always do things manually, as the following example demonstrates

# Create an evaluation data matrix

year.seq <- sort(unique(year))
gdp.seq <- seq(1,36,length=50)
data.eval <- expand.grid(year=year.seq,gdp=gdp.seq)

# Generate the estimated density computed for the evaluation data

fhat <- fitted(npudens(tdat = data, edat = data.eval, bws=bw))

# Coerce the data into a matrix for plotting with persp()

f <- matrix(fhat, length(unique(year)), 50)

# Next, create a 3D perspective plot of the PDF f

persp(as.integer(levels(year.seq)), gdp.seq, f, col="lightblue",
      ticktype="detailed", ylab="GDP", xlab="Year", zlab="Density",
      theta=300, phi=50)

# Sleep for 5 seconds so that we can examine the output...

Sys.sleep(5)

# However, plot simply streamlines this process and aids in the
# visualization process (<ctrl>-C will interrupt on *NIX systems, <esc>
# will interrupt on MS Windows systems).

plot(fit)

# plot also streamlines construction of variability bounds (<ctrl>-C
# will interrupt on *NIX systems, <esc> will interrupt on MS Windows
# systems)

plot(fit, errors = "asymptotic")

# EXAMPLE 2: For this example, we simulate multivariate data, and plot the
# partial regression surfaces for a locally linear estimator and its
# derivatives.

set.seed(123)

n <- 100

x1 <- runif(n)
x2 <- runif(n)
x3 <- runif(n)
x4 <- rbinom(n, 2, .3)

y <- 1 + x1 + x2 + x3 + x4 + rnorm(n)

X <- data.frame(x1, x2, x3, ordered(x4))

bw <- npregbw(xdat=X, ydat=y, regtype="ll", bwmethod="cv.aic")
fit <- npreg(bws=bw)

plot(fit)

# Sleep for 5 seconds so that we can examine the output...

Sys.sleep(5)

# Now plot the gradients...

plot(fit, gradients=TRUE)

# Plot the partial regression surfaces with bias-corrected bootstrapped
# nonparametric confidence intervals... this may take a minute or two
# depending on the speed of your computer as the bootstrapping must be
# completed prior to results being displayed...

plot(fit,
     errors="bootstrap",
     center="bias-corrected",
     band="simultaneous")

# EXAMPLE 3: This example demonstrates how to retrieve plotting data from
# plot(). When plot() is called with the arguments
# `output="both"' (or "data"), it returns plotting objects
# named r1, r2, and so on (rg1, rg2, and so on when `gradients=TRUE' is
# set).  Each plotting object's index (1,2,...) corresponds to the index
# of the explanatory data data frame xdat (and zdat if appropriate).

# Take the cps71 data by way of example. In this case, there is only one
# object returned by default, `r1', since xdat is univariate.

data("cps71", package = "np")

# Compute bandwidths for local linear regression using cv.aic...

bw <- npregbw(xdat=cps71$age, ydat=cps71$logwage,
              regtype="ll", bwmethod="cv.aic")
fit <- npreg(bws=bw)

# Generate the plot and return plotting data, and store output in
# `plot.out' (NOTE: the call to `output' is necessary).

plot.out <- plot(fit,
                 perspective=FALSE,
                 errors="bootstrap",
                 B=25,
                 output="both")

# Now grab the r1 object that plot plotted on the screen, and take
# what you need. First, take the output, lower error bound and upper
# error bound...

logwage.eval <- fitted(plot.out$r1)
logwage.se <- se(plot.out$r1)
logwage.lower.ci <- logwage.eval + logwage.se[,1]
logwage.upper.ci <- logwage.eval + logwage.se[,2]

# Next grab the x data evaluation data. xdat is a data.frame(), so we
# need to coerce it into a vector (take the `first column' of data frame
# even though there is only one column)

age.eval <- plot.out$r1$eval[,1]

# Now we could plot this if we wished, or direct it to whatever end use
# we envisioned. We plot the results using R's plot() routines...

with(cps71, plot(age, logwage, cex=0.2, xlab="Age", ylab="log(Wage)"))
lines(age.eval,logwage.eval)
lines(age.eval,logwage.lower.ci,lty=3)
lines(age.eval,logwage.upper.ci,lty=3)

# If you wanted plot() data for gradients, you would use the argument
# `gradients=TRUE' in the call to plot() as the following
# demonstrates...

plot.out <- plot(fit,
                 perspective=FALSE,
                 errors="bootstrap",
                 B=25,
                 output="both",
                 gradients=TRUE)

# Now grab object that plot() plotted on the screen. First, take the
# output, lower error bound and upper error bound... note that gradients
# are stored in objects rg1, rg2 etc.

grad.eval <- gradients(plot.out$rg1)
grad.se <- gradients(plot.out$rg1, errors = TRUE)
grad.lower.ci <- grad.eval + grad.se[,1]
grad.upper.ci <- grad.eval + grad.se[,2]

# Next grab the x evaluation data. xdat is a data.frame(), so we need to
# coerce it into a vector (take `first column' of data frame even though
# there is only one column)

age.eval <- plot.out$rg1$eval[,1]

# We plot the results using R's plot() routines...

plot(age.eval,grad.eval,cex=0.2,
     ylim=c(min(grad.lower.ci),max(grad.upper.ci)),
     xlab="Age",ylab="d log(Wage)/d Age",type="l")
lines(age.eval,grad.lower.ci,lty=3)
lines(age.eval,grad.upper.ci,lty=3)

# EXAMPLE 4: Variations on local polynomial conditional density
# estimation with proper = TRUE.

data("Italy")

Italy2 <- within(Italy, {
  year <- as.numeric(as.character(year))
})

# Plot only: make the plotted surface proper on the plot evaluation grid.

fhat <- npcdens(gdp ~ year, data = Italy2,
                regtype = "lp", degree = 3, nmulti = 1)

plot(fhat, proper = TRUE)

# Fit an object whose fitted values are themselves proper.

ctrl_fit <- list(
  mode = "slice",
  apply = "fitted",
  slice.grid.size = 101L,
  slice.extend.factor = 0.1
)

fhat_fit <- npcdens(
  gdp ~ year,
  data = Italy2,
  regtype = "lp",
  degree = 3,
  nmulti = 1,
  proper = TRUE,
  proper.control = ctrl_fit
)

fit_proper <- fitted(fhat_fit)
fit_raw <- fhat_fit$condens.raw

# Display the repaired and raw fitted values for cases where the raw
# fitted density is negative.

head(cbind(fit_proper, fit_raw)[which(fit_raw < 0), ])

# Predict on a common explicit y-grid for several years, and render
# those predictions proper.

g.grid <- seq(min(Italy2$gdp), max(Italy2$gdp), length.out = 200)

nd_grid <- expand.grid(
  gdp = g.grid,
  year = c(1955, 1975, 1995)
)

pred_grid <- predict(fhat, newdata = nd_grid, proper = TRUE)

# Predict on paired rows with different gdp grids by year, and still
# make the predictions proper via slice mode.

g1 <- seq(quantile(Italy2$gdp, 0.10),
          quantile(Italy2$gdp, 0.60), length.out = 60)
g2 <- seq(quantile(Italy2$gdp, 0.30),
          quantile(Italy2$gdp, 0.90), length.out = 35)

nd_slice <- rbind(
  data.frame(gdp = g1, year = rep(1960, length(g1))),
  data.frame(gdp = g2, year = rep(1985, length(g2)))
)

pred_slice <- predict(
  fhat,
  newdata = nd_slice,
  proper = TRUE,
  proper.control = list(mode = "slice")
)

# One object that carries properization for fitted values and for later
# predict() calls.

ctrl_both <- list(
  mode = "slice",
  apply = "both",
  slice.grid.size = 101L,
  slice.extend.factor = 0.1
)

fhat_both <- npcdens(
  gdp ~ year,
  data = Italy2,
  regtype = "lp",
  degree = 3,
  nmulti = 1,
  proper = TRUE,
  proper.control = ctrl_both
)

fit_both <- fitted(fhat_both)
pred_both <- predict(
  fhat_both,
  newdata = nd_slice,
  proper.control = ctrl_both
)
plot(fhat_both)
})
} % enddontrun
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}
  \code{\link{np.options}}
}

\keyword{nonparametric}

\name{np_plot_controls}
\alias{np_boot_control}
\alias{np_grid_control}
\alias{np_render_control}
\title{Validated plot control helpers}
\description{
  Constructors for detailed \code{plot} controls used by the redesigned
  \pkg{np} plotting interface. Common choices such as \code{errors},
  \code{band}, \code{alpha}, \code{bootstrap}, and \code{B} remain direct
  \code{plot} arguments; these helpers collect less common options.
}
\usage{
np_boot_control(nonfixed = c("exact", "frozen"),
                wild = c("rademacher", "mammen"),
                blocklen = NULL)

np_grid_control(xtrim = NULL, xq = NULL, slices = NULL)

np_render_control(style = c("band", "bar"),
                  bar = c("|", "I"),
                  bar_num = NULL)
}
\arguments{
  \item{nonfixed}{Bootstrap refit behavior for non-fixed bandwidth routes.}
  \item{wild}{Wild-bootstrap multiplier distribution.}
  \item{blocklen}{Optional block length for block bootstrap methods.}
  \item{xtrim}{Optional two-element trim range for evaluation grids.}
  \item{xq}{Optional evaluation quantiles.}
  \item{slices}{Optional slice specification for routes that support it.}
  \item{style}{Interval display style.}
  \item{bar}{Bar style when \code{style = "bar"}.}
  \item{bar_num}{Optional number of bars.}
}
\details{
  These helpers validate names and values early. Raw lists are deliberately not
  accepted as substitutes, so misspelled detailed controls fail before plot
  computation begins.

  New public plot arguments use \code{snake_case}. Dotted argument names are
  reserved for S3 methods and internal plot-engine plumbing. Common rendering
  controls such as \code{data_overlay}, \code{data_rug}, \code{layout}, and
  \code{output} are supplied directly to \code{plot(...)} rather than through
  \code{np_render_control()}.
}
\value{
  A small classed list consumed by \code{plot} methods.
}
\examples{
\dontrun{
plot(fit,
     errors = "bootstrap",
     bootstrap = "inid",
     B = 1999,
     band = "pointwise",
     boot_control = np_boot_control(nonfixed = "exact"))
}
}
\keyword{hplot}

\name{Italy}
\docType{data}
\alias{Italy}
\title{ Italian GDP Panel }
\description{
  Italian GDP growth panel for 21 regions covering the period 1951-1998
  (millions of Lire, 1990=base). There are 1008 observations in total.
}
\usage{data("Italy")}
\format{ A data frame with 2 columns, and 1008 rows.
  \describe{
    \item{year}{ the first column, of type \code{ordered}}

    \item{gdp}{ the second column, of type \code{numeric}: millions of
    Lire, 1990=base}
  }
}
\source{Giovanni Baiocchi}
\examples{
data("Italy")
with(Italy, {

plot(ordered(year), gdp, xlab="Year (ordered factor)",
     ylab="GDP (millions of Lire, 1990=base)")

})
}
\references{
  Baiocchi, G. (2006), \dQuote{Economic Applications of Nonparametric
    Methods,} Ph.D. Thesis, University of York.
}
\keyword{datasets}

\name{npudist}
\alias{npudist}
\alias{npudist.dbandwidth}
\alias{npudist.default}
\alias{npudist.formula}

\title{ Kernel Distribution Estimation with Mixed Data Types }
\description{

  \code{npudist} computes kernel unconditional cumulative distribution
  estimates on evaluation data, given a set of training data and a
  bandwidth specification (a \code{dbandwidth} object or a bandwidth
  vector, bandwidth type, and kernel type) using the method of Li, Li
  and Racine (2017).

}
\usage{
npudist(bws, \dots)

\method{npudist}{formula}(bws, data = NULL, newdata = NULL, \dots)

\method{npudist}{dbandwidth}(bws,
        tdat = stop("invoked without training data 'tdat'"),
        edat,
        \dots)


\method{npudist}{default}(bws, tdat, \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the bandwidth specification, formula/data interface, and training data.
}

  \item{bws}{
    a \code{dbandwidth} specification. This can be set as a \code{dbandwidth}
    object returned from an invocation of \code{\link{npudistbw}}, or as a
    \eqn{p}-vector of bandwidths, with an element for each variable in the
    training data. If specified as a vector, then additional arguments
    will need to be supplied as necessary to change them from the
    defaults to specify the bandwidth type, kernel types, training data,
    and so on.
  }



  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(bws)}, typically the environment from which
    \code{\link{npudistbw}} was called.
  }


  \item{tdat}{
    a \eqn{p}-variate data frame of sample realizations (training data)
    used to estimate the cumulative distribution. Defaults to the training data used to
    compute the bandwidth object.
  }


\subsection{Evaluation Data}{
These arguments control where the fitted cumulative distribution is evaluated.
}

  \item{edat}{
    a \eqn{p}-variate data frame of cumulative distribution evaluation points. By default,
    evaluation takes place on the data provided by \code{tdat}.
  }


  \item{newdata}{
    An optional data frame in which to look for evaluation data. If
    omitted, the training data are used.
  }


\subsection{Additional Arguments}{
Further arguments are passed to \code{\link{npudistbw}} when bandwidths are computed internally, or used to interpret a numeric \code{bws} vector.
}

  \item{\dots}{
    additional arguments supplied to \code{\link{npudistbw}} when
    \code{npudist} computes bandwidths internally, or arguments needed
    to interpret a numeric \code{bws} vector. This is where bandwidth
    selection controls such as \code{bwmethod}, \code{bwtype},
    \code{bwscaling}, kernel/support controls such as \code{ckertype},
    \code{ckerorder}, and \code{ckerbound}, categorical kernel controls
    such as \code{okertype}, and search controls such as
    \code{nmulti}, \code{ngrid}, and
    \code{scale.factor.search.lower} are supplied. See
    \code{\link{npudistbw}} for the complete bandwidth-selection
    argument surface.
  }
}
\details{
  Documentation guide: see \code{\link{npudistbw}} for bandwidth
  selection and search controls, \code{\link{np.kernels}} for kernels,
  \code{\link{np.options}} for global options, and \code{\link{plot}},
  \code{\link{plot.np}} for plotting options.

  When \code{bws} is omitted, the formula and default methods call
  \code{\link{npudistbw}} first and pass bandwidth-selection arguments
  from \code{\dots} to that call. When \code{bws} is already a
  \code{dbandwidth} object, \code{npudist} estimates with the stored
  bandwidth metadata in that object.

  Argument groups for bandwidth selection are documented on
  \code{\link{npudistbw}}. The most common workflow is to choose data
  and bandwidth inputs first, then bandwidth criterion and
  representation, then kernel/support controls, distribution-specific
  integral/grid controls, and numerical search controls if defaults
  need to be changed.

  For S3 plotting help, see \code{\link{plot.np}}. You can list
  available plot methods with \code{methods("plot")}.


  Typical usages are (see below for a complete list of options and also
  the examples at the end of this help file)
  \preformatted{
    Usage 1: first compute the bandwidth object via npudistbw and then
    compute the cumulative distribution:

    bw <- npudistbw(~y)
    Fhat <- npudist(bw)

    Usage 2: alternatively, compute the bandwidth object indirectly:

    Fhat <- npudist(~y)

    Usage 3: modify the default kernel and order:

    Fhat <- npudist(~y, ckertype="epanechnikov", ckerorder=4)

    Usage 4: use the data frame interface rather than the formula
    interface:

    Fhat <- npudist(tdat = y, ckertype="epanechnikov", ckerorder=4)
  }

  \code{npudist} implements a variety of methods for estimating
  multivariate cumulative distributions (\eqn{p}-variate) defined over a
  set of possibly continuous and/or discrete (ordered) data. The
  approach is based on Li and Racine (2003) who employ
  \sQuote{generalized product kernels} that admit a mix of continuous
  and discrete data types.

  Three classes of kernel estimators for the continuous data types are
  available: fixed, adaptive nearest-neighbor, and generalized
  nearest-neighbor. Adaptive nearest-neighbor bandwidths change with
  each sample realization in the set, \eqn{x_i}{x[i]}, when estimating
  the cumulative distribution at the point \eqn{x}. Generalized nearest-neighbor
  bandwidths change with the point at which the cumulative distribution is estimated,
  \eqn{x}. Fixed bandwidths are constant over the support of \eqn{x}.

  Data contained in the data frame \code{tdat} (and also \code{edat})
  may be a mix of continuous (default) and ordered discrete (to be
  specified in the data frame \code{tdat} using the
  \code{\link{ordered}} command). Data can be entered in an arbitrary
  order and data types will be detected automatically by the routine
  (see \code{\link{np}} for details).

  A variety of kernels may be specified by the user. Kernels implemented
  for continuous data types include the second, fourth, sixth, and
  eighth-order Gaussian and Epanechnikov kernels, and the uniform
  kernel. Ordered data types use a variation of the Wang and van Ryzin
  (1981) kernel.

}
\value{

  \code{\link{npudist}} returns a \code{npdistribution} object. The
  generic accessor functions \code{\link{fitted}} and \code{\link{se}}
  extract estimated values and asymptotic standard errors on estimates,
  respectively, from the returned object. Furthermore, the functions
  \code{\link{predict}}, \code{\link{summary}} and \code{\link{plot}}
  support objects of both classes. The returned objects have the
  following components:

  \item{eval}{ the evaluation points. }
  \item{dist}{ estimate of the cumulative distribution at the evaluation points }
  \item{derr}{ standard errors of the cumulative distribution estimates }

}
\section{Book And Method Pointers}{
  The unconditional distribution target is \eqn{F(x)=\Pr(X\le x)}.
  The estimator replaces the density kernel factors by integrated
  distribution-kernel factors where appropriate, while retaining the
  same mixed-data product-kernel logic for continuous and ordered
  components. The returned standard errors are asymptotic standard
  errors for the estimated cumulative distribution values.

  For book-length derivations, see Li and Racine (2007), Chapter 1
  \emph{Density Estimation}, especially Sections 1.4 and 1.5 for CDF
  estimation and bandwidth selection, together with Chapter 4
  \emph{Kernel Estimation with Mixed Data}. The later workflow
  treatment is Racine (2019), Chapter 2 \emph{Continuous Density and
  Cumulative Distribution Functions} and Chapter 3 \emph{Mixed-Data
  Probability Density and Cumulative Distribution Functions}.
}

\references{
  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{ Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
  and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2003), \dQuote{Nonparametric estimation of
    distributions with categorical and continuous data,} Journal
  of Multivariate Analysis, 86, 266-292.

  Li, C. and H. Li and J.S. Racine (2017), \dQuote{Cross-Validated Mixed
  Datatype Bandwidth Selection for Nonparametric Cumulative
  Distribution/Survivor Functions,} Econometric Reviews, \bold{36},
  970-987.

  Ouyang, D. and Q. Li and J.S. Racine (2006), \dQuote{Cross-validation
    and the estimation of probability distributions with categorical
    data,} Journal of Nonparametric Statistics, 18, 69-100.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,} Cambridge
  University Press.

  Scott, D.W. (1992), \emph{Multivariate Density Estimation. Theory,
  Practice and Visualization,} New York: Wiley.

  Silverman, B.W. (1986), \emph{Density Estimation,} London: Chapman and
  Hall.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth estimators
    for discrete distributions,}  Biometrika, 68, 301-309.
}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}, \code{\link{plot.np}} \code{\link{npudistbw}} , \code{\link{density}} }

\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): For this example, we load Giovanni
# Baiocchi's Italian GDP panel (see Italy for details), then create a
# data frame in which year is an ordered factor, GDP is continuous,
# compute bandwidths using cross-validation, then create a grid of data
# on which the cumulative distribution will be evaluated for plotting
# purposes.

data("Italy")
with(Italy, {

# Compute bandwidths using cross-validation (default).

bw <- npudistbw(formula=~ordered(year)+gdp)

# At this stage you could use npudist() to do a variety of things. Here
# we compute the npudist() object and place it in Fhat.

Fhat <- npudist(bws=bw)

# Note that simply typing the name of the object returns some useful
# information. For more info, one can call summary:

summary(Fhat)

# Next, we illustrate how to create a grid of `evaluation data' and feed
# it to the perspective plotting routines in R, among others.

# Create an evaluation data matrix

year.seq <- sort(unique(year))
gdp.seq <- seq(1,36,length=50)
data.eval <- expand.grid(year=year.seq,gdp=gdp.seq)

# Generate the estimated cumulative distribution computed for the
# evaluation data

Fhat <- fitted(npudist(bws=bw, newdata=data.eval))

# Coerce the data into a matrix for plotting with persp()

Fhat <- matrix(Fhat, length(unique(year)), 50)

# Next, create a 3D perspective plot of the CDF F, and a 2D
# contour plot.

persp(as.integer(levels(year.seq)), gdp.seq, Fhat, col="lightblue",
      ticktype="detailed", ylab="GDP", xlab="Year", zlab="Density",
      theta=300, phi=50)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

contour(as.integer(levels(year.seq)),
        gdp.seq,
        Fhat,
        xlab="Year",
        ylab="GDP",
        main = "Cumulative Distribution Contour Plot",
        col=topo.colors(100))

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Alternatively, you could use the plot() command (<ctrl>-C will
# interrupt on *NIX systems, <esc> will interrupt on MS Windows
# systems).

if (interactive()) plot(bw)

})

# EXAMPLE 1 (INTERFACE=DATA FRAME): For this example, we load Giovanni
# Baiocchi's Italian GDP panel (see Italy for details), then create a
# data frame in which year is an ordered factor, GDP is continuous,
# compute bandwidths using cross-validation, then create a grid of data
# on which the cumulative distribution will be evaluated for plotting
# purposes.

data("Italy")
with(Italy, {

data <- data.frame(year=ordered(year), gdp)

# Compute bandwidths using cross-validation (default).

bw <- npudistbw(dat=data)

# At this stage you could use npudist() to do a variety of
# things. Here we compute the npudist() object and place it in Fhat.

Fhat <- npudist(bws=bw)

# Note that simply typing the name of the object returns some useful
# information. For more info, one can call summary:

summary(Fhat)

# Next, we illustrate how to create a grid of `evaluation data' and feed
# it to the perspective plotting routines in R, among others.

# Create an evaluation data matrix

year.seq <- sort(unique(year))
gdp.seq <- seq(1,36,length=50)
data.eval <- expand.grid(year=year.seq,gdp=gdp.seq)

# Generate the estimated cumulative distribution computed for the
# evaluation data

Fhat <- fitted(npudist(edat = data.eval, bws=bw))

# Coerce the data into a matrix for plotting with persp()

Fhat <- matrix(Fhat, length(unique(year)), 50)

# Next, create a 3D perspective plot of the CDF F, and a 2D
# contour plot.

persp(as.integer(levels(year.seq)), gdp.seq, Fhat, col="lightblue",
      ticktype="detailed", ylab="GDP", xlab="Year",
      zlab="Cumulative Distribution",
      theta=300, phi=50)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

contour(as.integer(levels(year.seq)),
        gdp.seq,
        Fhat,
        xlab="Year",
        ylab="GDP",
        main = "Cumulative Distribution Contour Plot",
        col=topo.colors(100))

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Alternatively, you could use the plot() command (<ctrl>-C will
# interrupt on *NIX systems, <esc> will interrupt on MS Windows
# systems).

if (interactive()) plot(bw)

})

# EXAMPLE 2 (INTERFACE=FORMULA): For this example, we load the old
# faithful geyser data and compute the cumulative distribution function.

library("datasets")
data("faithful")
with(faithful, {

# Note - this may take a few minutes depending on the speed of your
# computer...

bw <- npudistbw(formula=~eruptions+waiting)

summary(bw)

# Plot the cumulative distribution function (<ctrl>-C will interrupt on
# *NIX systems, <esc> will interrupt on MS Windows systems). Note that
# we use xtrim = -0.2 to extend the plot outside the support of the data
# (i.e., extend the tails of the estimate to meet the horizontal axis).

if (interactive()) plot(bw, xtrim=-0.2)

})

# EXAMPLE 2 (INTERFACE=DATA FRAME): For this example, we load the old
# faithful geyser data and compute the cumulative distribution function.

library("datasets")
data("faithful")
with(faithful, {

# Note - this may take a few minutes depending on the speed of your
# computer...

bw <- npudistbw(dat=faithful)

summary(bw)

# Plot the cumulative distribution function (<ctrl>-C will interrupt on
# *NIX systems, <esc> will interrupt on MS Windows systems). Note that
# we use xtrim = -0.2 to extend the plot outside the support of the data
# (i.e., extend the tails of the estimate to meet the horizontal axis).

if (interactive()) plot(bw, xtrim=-0.2)

})
} % enddontrun
}
\keyword{ nonparametric }% at least one, from doc/KEYWORDS

\name{np.pairs}
\alias{np.pairs}
\alias{np.pairs.plot}
\title{Cross-Validated Pairs Plot (Helper Functions)}
\description{
Compute pairwise nonparametric regressions and densities for a set of
variables, then plot a pairs-style display with fitted smoothers.
}
\usage{
np.pairs(y_vars, y_dat, ...)
np.pairs.plot(pair_list)
}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the variables, data, and pair specifications to plot.
}

  \item{pair_list}{
    list returned by \code{np.pairs}.
  }


  \item{y_dat}{
    data frame containing the variables listed in \code{y_vars}.
  }


  \item{y_vars}{
    character vector of column names in \code{y_dat}. If \code{y_vars}
    is named, the names are used as plot labels.
  }


\subsection{Additional Arguments}{
Further graphical arguments are passed through to plotting methods.
}

  \item{...}{
    additional arguments passed to \code{\link{npudens}} and
    \code{\link{npreg}}.
  }

}
\value{
  \code{np.pairs} returns a list with components
  \code{y_vars}, \code{pair_names}, and \code{pair_kerns}.
  \code{np.pairs.plot} returns \code{NULL} (invisibly).
}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.

On the diagonal, \code{npudens} is used to compute kernel density
estimates. Off-diagonal panels use \code{npreg} with residuals to draw
scatterplots and smoothers.
}
\examples{
\dontrun{
data("USArrests")
y_vars <- c("Murder", "UrbanPop")
names(y_vars) <- c("Murder Arrests per 100K", "Pop. Percent Urban")

pair_list <- np.pairs(
  y_vars = y_vars,
  y_dat = USArrests,
  ckertype = "epanechnikov",
  bwscaling = TRUE
)

np.pairs.plot(pair_list)
}
}
\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}\code{\link{npudens}}, \code{\link{npreg}}}
\keyword{hplot}

\name{np.kernels}
\alias{np.kernels}
\title{Kernel Functions Used In \pkg{np}}
\description{
  Summary of continuous, unordered-categorical, and ordered-categorical
  kernels used by \pkg{np} (including higher-order continuous kernels and
  compact-support variants used in C-level code paths).
}
\details{
  Documentation guide: see \code{\link{np.options}} for global options
  and \code{\link{plot}} for plotting options.

  Kernel option names used in \pkg{np}:

  \itemize{
    \item Continuous kernels:
    \code{ckertype} (and \code{ckerorder}, \code{ckerbound} where applicable).
    \item Unordered kernels:
    \code{ukertype}.
    \item Ordered kernels:
    \code{okertype}.
    \item Conditional density/distribution bandwidth objects split kernel
    choices by response and regressor blocks:
    \code{cykertype}/\code{cxkertype},
    \code{uykertype}/\code{uxkertype},
    \code{oykertype}/\code{oxkertype}
    (with matching order/bound options for continuous kernels).
  }

  Let \eqn{u = (x_i-x)/h} for continuous variables.

  Continuous kernels (called via \code{ckertype}):
  \deqn{K_{G,2}(u)=\phi(u)}
  \deqn{K_{G,4}(u)=\left(\frac{3}{2}-\frac{1}{2}u^2\right)\phi(u)}
  \deqn{K_{G,6}(u)=\left(\frac{15}{8}-\frac{5}{4}u^2+\frac{1}{8}u^4\right)\phi(u)}
  \deqn{K_{G,8}(u)=\left(\frac{35}{16}-\frac{35}{16}u^2+\frac{7}{16}u^4-\frac{1}{48}u^6\right)\phi(u)}
  where \eqn{\phi(u)} is the standard normal density.

  \code{ckertype="gaussian"} with \code{ckerorder=2,4,6,8}.

  The compact-support Epanechnikov-family kernels implemented in C use
  support \eqn{|u|<\sqrt{5}}:
  \deqn{K_{E,2}(u)=\frac{3}{4\sqrt{5}}\left(1-\frac{u^2}{5}\right)\mathbf{1}(|u|<\sqrt{5})}
  \deqn{K_{E,4}(u)=0.008385254916(-15+7u^2)(-5+u^2)\mathbf{1}(u^2<5)}
  \deqn{K_{E,6}(u)=0.33541019662496845446\left(2.734375-3.28125u^2+0.721875u^4\right)\left(1-0.2u^2\right)\mathbf{1}(u^2<5)}
  \deqn{K_{E,8}(u)=0.33541019662496845446\left(3.5888671875-7.8955078125u^2+4.1056640625u^4-0.5865234375u^6\right)\left(1-0.2u^2\right)\mathbf{1}(u^2<5)}

  \code{ckertype="epanechnikov"} with \code{ckerorder=2,4,6,8}.

  Uniform (rectangular) kernel:
  \deqn{K_U(u)=\frac{1}{2}\mathbf{1}(|u|<1)}
  via \code{ckertype="uniform"} (order ignored).

  Truncated-Gaussian (second-order) kernel via
  \code{ckertype="truncated gaussian"}:
  \deqn{K_{TG,2}(u)=\left[\alpha\phi(u)-c_0\right]\mathbf{1}(|u|<b)}
  with defaults \eqn{b=3} and internal constants calibrated in C.

  Bounded continuous-kernel normalization (\code{ckerbound} and, for
  conditional objects, \code{cxkerbound}/\code{cykerbound}) reuses the
  selected continuous kernel and renormalizes it on the declared support.
  For a base kernel \eqn{K} and support \eqn{[a,b]}, the bounded kernel is
  \deqn{K_{[a,b]}(u;x,h)=\frac{K(u)}{\int_{(a-x)/h}^{(b-x)/h}K(t)dt}}
  with \eqn{u=(x_i-x)/h}. Option \code{ckerbound="range"} uses
  sample bounds for \eqn{a,b}; \code{ckerbound="fixed"} uses user-supplied
  bounds via \code{ckerlb}/\code{ckerub} (or the corresponding
  \code{cx*}/\code{cy*} arguments). Infinite bounds recover the unbounded
  kernel. This support-normalization strategy follows the same Racine-Li-Yan
  finite-support normalization principle and is useful when data exhibit
  non-negligible probability mass near boundaries.

  Typical bounded-kernel calls:
  \preformatted{
  ## Unconditional density on [0,1]
  bw <- npudensbw(dat=data.frame(x),
                  ckertype="gaussian",
                  ckerbound="fixed", ckerlb=0, ckerub=1)

  ## Regression with automatic sample-range bounds
  bw <- npregbw(xdat=data.frame(x), ydat=y, ckerbound="range")

  ## Conditional density with separate x/y support controls
  bw <- npcdensbw(xdat=data.frame(x), ydat=data.frame(y),
                  cxkerbound="fixed", cxkerlb=0, cxkerub=1,
                  cykerbound="range")
  }

  Unordered-categorical kernels (called via \code{ukertype}; for category
  count \eqn{c}):
  \deqn{L_{AA}(x_i,x;\lambda)=\mathbf{1}(x_i=x)(1-\lambda)+\mathbf{1}(x_i\neq x)\frac{\lambda}{c-1}}
  (Aitchison-Aitken)
  via \code{ukertype="aitchisonaitken"}.

  \deqn{L_{LR,u}(x_i,x;\lambda)=\mathbf{1}(x_i=x)+\mathbf{1}(x_i\neq x)\lambda}
  (Li-Racine unordered kernel)
  via \code{ukertype="liracine"}.

  Ordered-categorical kernels (called via \code{okertype}):
  \deqn{L_{WvR}(x_i,x;\lambda)=
    \begin{cases}
      1-\lambda, & x_i=x\\
      \frac{1-\lambda}{2}\lambda^{|x_i-x|}, & x_i\neq x
    \end{cases}}
  (Wang-van Ryzin)
  via \code{okertype="wangvanryzin"}.

  \deqn{L_{LR,o}(x_i,x;\lambda)=\lambda^{|x_i-x|}}
  (Li-Racine ordered kernel)
  via \code{okertype="liracine"}.

  \deqn{L_{NLR,o}(x_i,x;\lambda)=\lambda^{|x_i-x|}\frac{1-\lambda}{1+\lambda}}
  (normalized Li-Racine ordered kernel; used internally)

  \deqn{L_{RLY}(x_i,x;\lambda)=\frac{\lambda^{|x_i-x|}}{\sum_{z\in\mathcal{S}(x)}\lambda^{|x_i-z|}}}
  (Racine-Li-Yan ordered kernel, normalized on support \eqn{\mathcal{S}(x)}).
  exposed as \code{okertype="racineliyan"}.

  These univariate kernels are combined as generalized product kernels over
  mixed data types in the estimators and cross-validation criteria.
}
\references{
  Aitchison, J. and Aitken, C. G. G. (1976),
  \dQuote{Multivariate binary discrimination by the kernel method,}
  \emph{Biometrika}, \bold{63}, 413--420.

  Wang, M. C. and Van Ryzin, J. (1981),
  \dQuote{A class of smooth estimators for discrete distributions,}
  \emph{Biometrika}, \bold{68}, 301--309.

  Li, Q. and Racine, J. S. (2007),
  \emph{Nonparametric Econometrics: Theory and Practice},
  Princeton University Press.

  Racine, J. S. and Li, Q. (2004),
  \dQuote{Nonparametric estimation of regression functions with both
  categorical and continuous data,}
  \emph{Journal of Econometrics}, \bold{119}, 99--130.

  Racine, J. S., Li, Q., and Yan, K. X. (2020),
  \dQuote{Kernel Smoothed Probability Mass Functions for Ordered Datatypes,}
  \emph{Journal of Nonparametric Statistics}, \bold{32}(3), 563--586.
  doi:10.1080/10485252.2020.1759595

  Hall, P., Racine, J. S., and Li, Q. (2004),
  \dQuote{Cross-validation and the estimation of conditional probability
  densities,}
  \emph{Journal of the American Statistical Association}, \bold{99}, 1015--1026.
}
\seealso{
  \code{\link{np.options}}, \code{\link{plot}}
  \code{\link{npregbw}},
  \code{\link{npudensbw}},
  \code{\link{npudistbw}},
  \code{\link{npcdensbw}},
  \code{\link{npcdistbw}},
  \code{\link{npksum}},
  \code{\link{np.options}}.
}
\keyword{nonparametric}
\keyword{smooth}

% 
\name{cps71}
\docType{data}
\alias{cps71}
\title{ Canadian High School Graduate Earnings  }
\description{
	Canadian cross-section wage data consisting of a random sample taken
  from the 1971 Canadian Census Public Use Tapes for male individuals
  having common education (grade 13). There are 205 observations in total.
}
\usage{data("cps71")}
\format{ A data frame with 2 columns, and 205 rows.
  \describe{
    \item{logwage}{ the first column, of type \code{numeric}}
    \item{age}{ the second column, of type \code{integer}}
  }
}
\source{ Aman Ullah }
\examples{
data("cps71", package = "np")

if (interactive()) with(cps71, plot(age, logwage, xlab="Age", ylab="log(wage)"))
}
\references{
  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,}
  Cambridge	University Press.
}
\keyword{datasets}

% 
\name{npscoefbw}
\alias{npscoefbw}
\alias{npscoefbw.default}
\alias{npscoefbw.formula}
\alias{npscoefbw.scbandwidth}
\title{Smooth Coefficient Kernel Regression Bandwidth Selection}

\description{
  \code{npscoefbw} computes a bandwidth object for a smooth
  coefficient kernel regression estimate of a one (1) dimensional
  dependent variable on 
  \eqn{p+q}-variate explanatory data, using the model
  \eqn{Y_i = W_{i}^{\prime} \gamma (Z_i) + u_i}{Y_i = t(W_i) * gamma(Z_i) +
    u_i} where \eqn{W_i'=(1,X_i')}{t(W_i) = (1,t(X_i))}
  given training points (consisting of explanatory data and dependent
  data), and a bandwidth specification, which can be a \code{rbandwidth}
  object, or a bandwidth vector, bandwidth type and kernel type. 
}

\usage{
npscoefbw(\dots)

\method{npscoefbw}{formula}(formula, 
          data, 
          subset, 
          na.action, 
          call, 
          \dots)

\method{npscoefbw}{default}(xdat = stop("invoked without data 'xdat'"),
          ydat = stop("invoked without data 'ydat'"),
          zdat = NULL,
          bws,
          backfit.iterate,
          backfit.maxiter,
          backfit.tol,
          bandwidth.compute = TRUE,
          basis,
          bernstein.basis,
          bwmethod,
          bwscaling,
          bwtype,
          ckerbound,
          ckerlb,
          ckerorder,
          ckertype,
          ckerub,
          cv.iterate,
          cv.num.iterations,
          degree,
          degree.select = c("manual", "coordinate", "exhaustive"),
          search.engine = c("nomad+powell", "cell", "nomad"),
          nomad = FALSE,
          nomad.nmulti = 0L,
            degree.min = NULL,
          degree.max = NULL,
          degree.start = NULL,
          degree.restarts = 0L,
          degree.max.cycles = 20L,
          degree.verify = FALSE,
          nmulti,
          nomad.remin = FALSE,
          powell.remin = TRUE,
          okertype,
          optim.abstol,
          optim.maxattempts,
          optim.maxit,
          optim.method,
          optim.reltol,
          random.seed,
          regtype,
          ukertype,
          scale.factor.init.lower = 0.1,
          scale.factor.init.upper = 2.0,
          scale.factor.init = 0.5,
          lbd.init = 0.5,
          hbd.init = 1.5,
          dfac.init = 1.0,
          scale.factor.search.lower = NULL,
          \dots)

\method{npscoefbw}{scbandwidth}(xdat = stop("invoked without data 'xdat'"),
          ydat = stop("invoked without data 'ydat'"),
          zdat = NULL,
          bws,
          backfit.iterate = FALSE,
          backfit.maxiter = 100,
          backfit.tol = .Machine$double.eps,
          bandwidth.compute = TRUE,
          cv.iterate = FALSE,
          cv.num.iterations = 1,
          nmulti,
          optim.abstol = .Machine$double.eps,
          optim.maxattempts = 10,
          optim.maxit = 500,
          optim.method = c("Nelder-Mead", "BFGS", "CG"),
          optim.reltol = sqrt(.Machine$double.eps),
          random.seed = 42,
          scale.factor.init.lower = 0.1,
          scale.factor.init.upper = 2.0,
          scale.factor.init = 0.5,
          lbd.init = 0.5,
          hbd.init = 1.5,
          dfac.init = 1.0,
          scale.factor.search.lower = NULL,
          \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the smooth-coefficient data, formula interface, and whether bandwidths are supplied or computed.
}

  \item{bandwidth.compute}{
    a logical value which specifies whether to do a numerical search for
    bandwidths or not. If set to \code{FALSE}, a \code{scbandwidth} object
    will be returned with bandwidths set to those specified
    in \code{bws}. Defaults to \code{TRUE}.
  }

  \item{bws}{
    a bandwidth specification. This can be set as a \code{scbandwidth}
    object returned from a previous invocation, or as a vector of
    bandwidths, with each element \eqn{i} corresponding to the bandwidth
    for column \eqn{i} in \code{xdat}. In either case, the bandwidth
    supplied will serve as a starting point in the numerical search for
    optimal bandwidths. If specified as a vector, then additional
    arguments will need to be supplied as necessary to specify the
    bandwidth type, kernel types, selection methods, and so on. This can
    be left unset.
  }

  
  \item{call}{
    the original function call. This is passed internally by
    \code{\link{np}} when a bandwidth search has been implied by a call to
    another function. It is not recommended that the user set this.
  }

  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(formula)}, typically the environment from which the
    function is called.
  }

  \item{formula}{
    a symbolic description of variables on which bandwidth selection is
    to be performed. The details of constructing a formula are
    described below.
  }

  \item{na.action}{
    a function which indicates what should happen when the data contain
   \code{NA}s. The default is set by the \code{\link{na.action}} setting of options, and is
    \code{\link{na.fail}} if that is unset. The (recommended) default is
    \code{\link{na.omit}}.  
  }
  
  \item{subset}{
    an optional vector specifying a subset of observations to be used in
    the fitting process. 
  }

  \item{xdat}{
    a \eqn{p}-variate data frame of explanatory data (training data),
    which, by default, populates the columns \eqn{2} through \eqn{p+1}
    of \eqn{W} in the model equation, and in the 
    absence of \code{zdat}, will also correspond to 
    \eqn{Z} from the model equation.
  }

  \item{ydat}{
    a one (1) dimensional numeric or integer vector of dependent data, each
    element \eqn{i} corresponding to each observation (row) \eqn{i} of
    \code{xdat}. 
  }

  \item{zdat}{
    an optionally specified \eqn{q}-variate data frame of explanatory
    data (training data), which corresponds to \eqn{Z}
    in the model equation. Defaults to be the same as \code{xdat}.
  }



\subsection{Automatic Degree Search Controls}{
These arguments control automatic local-polynomial degree search.
}

  \item{degree.max}{
    optional scalar or integer vector giving upper bounds for automatic
    degree search when \code{degree.select != "manual"}.
  }

  \item{degree.max.cycles}{
    positive integer giving the maximum number of coordinate-search
    sweeps over the degree vector. Ignored for \code{"manual"} and
    \code{"exhaustive"} degree selection.
  }

  \item{degree.min}{
    optional scalar or integer vector giving lower bounds for automatic
    degree search when \code{degree.select != "manual"}.
  }

  \item{degree.restarts}{
    non-negative integer giving the number of additional deterministic
    coordinate-search restarts. Ignored for \code{"manual"} and
    \code{"exhaustive"} degree selection.
  }

  \item{degree.select}{
    character string controlling local-polynomial degree handling when
    \code{regtype="lp"}. \code{"manual"} (default) treats
    \code{degree} as fixed. \code{"coordinate"} performs cached
    coordinate-wise search over admissible degree vectors for the
    continuous \code{z} variables. \code{"exhaustive"} evaluates the
    full admissible degree grid when \code{search.engine="cell"}.
    For NOMAD-based search engines, any non-\code{"manual"} value
    requests direct joint search over degree and bandwidth
    coordinates.
  }

  \item{degree.start}{
    optional starting degree vector for automatic coordinate search. If
    omitted, the search starts from the degree-zero local-constant
    baseline on the continuous \code{z} variables.
  }

  \item{degree.verify}{
    logical value indicating whether a coordinate-search solution should
    be exhaustively verified over the admissible degree grid after the
    heuristic phase completes. Available only for
    \code{search.engine="cell"}.
  }


\subsection{Backfitting Controls}{
These controls tune the optional smooth-coefficient backfitting iterations.
}

  \item{backfit.iterate}{
    boolean value specifying whether or not to iterate evaluations of
    the smooth coefficient estimator, for extra accuracy, during the
    cross-validated backfitting procedure. Defaults to \code{FALSE}.
  }

  \item{backfit.maxiter}{
    integer specifying the maximum number of times to iterate the
    evaluation of the smooth coefficient estimator in the attempt to
    obtain the desired accuracy. Defaults to \code{100}.
  }

  \item{backfit.tol}{
    tolerance to determine convergence of iterated evaluations of the
    smooth coefficient estimator. Defaults to \code{.Machine$double.eps}.
  }


\subsection{Bandwidth Criterion And Representation}{
These arguments choose the selection criterion and the way continuous bandwidths are represented.
}

  \item{bwmethod}{
    which method was used to select bandwidths. \code{cv.ls}
    specifies least-squares cross-validation, which is all that is
    currently supported. Defaults to \code{cv.ls}.
  }

  \item{bwscaling}{ a logical value that when set to \code{TRUE} the
    supplied bandwidths are interpreted as \sQuote{scale factors}
    (\eqn{c_j}{c[j]}), otherwise when the value is \code{FALSE} they are
    interpreted as \sQuote{raw bandwidths} (\eqn{h_j}{h[j]} for continuous data
    types, \eqn{\lambda_j}{lambda[j]} for discrete data types). For
    continuous data types, \eqn{c_j}{c[j]} and \eqn{h_j}{h[j]} are
    related by the formula \eqn{h_j = c_j \sigma_j n^{-1/(2P+l)}}{h[j] =
    c[j]*sigma[j]*n^(-1/(2*P+l))}, where \eqn{\sigma_j}{sigma[j]} is an
    adaptive measure of spread of continuous variable \eqn{j} defined as
    min(standard deviation, mean absolute deviation, interquartile
    range/1.349), \eqn{n} the number of observations, \eqn{P}{P} the
    order of the kernel, and \eqn{l}{l} the number of continuous
    variables. For discrete data types, \eqn{c_j}{c[j]} and
    \eqn{h_j}{h[j]} are related by the formula \eqn{h_j =
    c_jn^{-2/(2P+l)}}{h[j] = c[j]*n^(-2/(2*P+l))}, where here
    \eqn{j}{[j]} denotes discrete variable \eqn{j}{j}.  Defaults to
    \code{FALSE}.  }

  \item{bwtype}{
    character string used for the continuous variable bandwidth type,
    specifying the type of bandwidth provided. Defaults to
    \code{fixed}. Option summary:\cr
    \code{fixed}: fixed bandwidths or scale factors \cr
    \code{generalized_nn}: generalized nearest neighbors \cr
    \code{adaptive_nn}:  adaptive nearest neighbors
  }
  

\subsection{Categorical Search Initialization}{
These controls set categorical search starts.
}

  \item{dfac.init}{
    deterministic fixed-bandwidth start factor for ordered and
    unordered categorical coordinates. Used only when
    \code{bwtype="fixed"}. Defaults to \code{1.0}. Values must not
    exceed \code{2}.
  }

  \item{hbd.init}{
    upper bound for random fixed-bandwidth start factors for ordered
    and unordered categorical coordinates. Used only when
    \code{bwtype="fixed"}. Defaults to \code{1.5}. Must be greater than
    or equal to \code{lbd.init} and must not exceed \code{2}.
  }

  \item{lbd.init}{
    lower bound for random fixed-bandwidth start factors for ordered
    and unordered categorical coordinates. Used only when
    \code{bwtype="fixed"}. Defaults to \code{0.5}. Values must not
    exceed \code{2} so that categorical fixed starts remain within
    lawful bandwidth bounds.
  }


\subsection{Continuous Kernel Support Controls}{
These controls choose and parameterize bounded support for continuous kernels.
}

  \item{ckerbound}{
    character string controlling continuous-kernel support handling.
    Can be set as \code{none} (default kernel on full support),
    \code{range} (use sample min/max), or \code{fixed} (use
    \code{ckerlb}/\code{ckerub}). The bounded-kernel route reuses the selected continuous kernel and renormalizes it on the chosen support; see \code{\link{np.kernels}}.
  }

  \item{ckerlb}{
    numeric scalar/vector of lower bounds for continuous variables used
    when \code{ckerbound="fixed"}. Must satisfy lower-bound validity for
    each continuous variable (e.g., \code{<= min(variable)}). Use \code{-Inf} for unbounded below. See \code{\link{np.kernels}} for bounded-kernel normalization details.
  }

  \item{ckerub}{
    numeric scalar/vector of upper bounds for continuous variables used
    when \code{ckerbound="fixed"}. Must satisfy upper-bound validity for
    each continuous variable (e.g., \code{>= max(variable)}). Use \code{Inf} for unbounded above. See \code{\link{np.kernels}} for bounded-kernel normalization details.
  }


\subsection{Continuous Scale-Factor Search Initialization}{
These controls define deterministic and random continuous scale-factor starts and the lower admissibility floor for fixed-bandwidth search.
}

  \item{scale.factor.init}{
    deterministic initial scale factor for continuous fixed-bandwidth
    search. Defaults to \code{0.5}. The value supplied by the user is
    not rewritten, but the effective first start passed to the optimizer
    is \code{max(scale.factor.init, scale.factor.search.lower)}. See
    Details.
  }
  \item{scale.factor.init.lower}{
    lower endpoint for random continuous scale-factor starts. Defaults
    to \code{0.1}. The value supplied by the user is not rewritten, but
    the effective random-start lower endpoint is
    \code{max(scale.factor.init.lower, scale.factor.search.lower)}. See
    Details.
  }
  \item{scale.factor.init.upper}{
    upper endpoint for random continuous scale-factor starts. Defaults
    to \code{2.0}. It must be greater than or equal to the effective
    lower endpoint,
    \code{max(scale.factor.init.lower, scale.factor.search.lower)};
    otherwise bandwidth search errors rather than silently expanding the
    interval. See Details.
  }
  \item{scale.factor.search.lower}{
    optional nonnegative scalar giving the hard lower admissibility
    bound for continuous fixed-bandwidth search candidates. Defaults to
    \code{NULL}. If \code{NULL}, an existing bandwidth object's stored
    value is inherited when available; otherwise the package default
    \code{0.1} is used. This floor applies to computed/search bandwidth
    candidates and to effective search starts only. It does not rewrite
    explicit bandwidths supplied for storage with
    \code{bandwidth.compute = FALSE}. Final fixed-bandwidth search
    candidates must also have a finite valid raw objective value.
  }

\subsection{Cross-Validation Iteration Controls}{
These controls tune iterative cross-validation behavior.
}

  \item{cv.iterate}{
    boolean value specifying whether or not to perform iterative,
    cross-validated backfitting on the data. See details for limitations
    of the backfitting procedure. Defaults to \code{FALSE}.
  }

  \item{cv.num.iterations}{
    integer specifying the number of times to iterate the backfitting
    process over all covariates. Defaults to \code{1}.
  }


\subsection{Kernel Type Controls}{
These controls choose continuous, unordered, and ordered kernels.
}

  \item{ckerorder}{
    numeric value specifying kernel order (one of
    \code{(2,4,6,8)}). Kernel order specified along with a
  \code{uniform} continuous kernel type will be ignored. Defaults to
  \code{2}.
  }

  \item{ckertype}{
    character string used to specify the continuous kernel type.
    Can be set as \code{gaussian}, \code{epanechnikov}, or
    \code{uniform}. Defaults to \code{gaussian}.
  }

  \item{okertype}{
    character string used to specify the ordered categorical kernel type.
    Can be set as \code{wangvanryzin}, \code{liracine}, or
    \code{racineliyan}. Defaults to \code{liracine}.
  }
  \item{ukertype}{
    character string used to specify the unordered categorical kernel type.
    Can be set as \code{aitchisonaitken} or \code{liracine}. Defaults
    to \code{aitchisonaitken}.
  }

\subsection{Local-Polynomial Model Specification}{
These arguments control the local-polynomial estimator, basis, and fixed degree specification.
}

  \item{basis}{
    for \code{regtype="lp"}, the polynomial basis family used for local
    polynomial fitting. Options are \code{"glp"} (default),
    \code{"additive"}, and \code{"tensor"}.
  }

  \item{bernstein.basis}{
    for \code{regtype="lp"}, logical flag selecting Bernstein-basis
    representation where supported. When automatic degree search is
    requested and \code{bernstein.basis} is not explicitly supplied,
    the search route defaults to \code{TRUE} for numerical stability.
    Explicit \code{bernstein.basis=FALSE} is honored, but raw-polynomial
    search can be poorly conditioned at higher degrees.
  }

  \item{degree}{
    for \code{regtype="lp"}, polynomial degree specification for each
    continuous \code{z} variable.
  }

  \item{regtype}{
    a character string specifying local smoothing type for the \code{z}
    surface. Options are \code{"lc"} (default), \code{"ll"}, and
    \code{"lp"}.
  }


\subsection{NOMAD Search Controls}{
These arguments control the optional NOMAD direct-search route for local-polynomial degree and bandwidth search.
}

  \item{nomad}{
    logical shortcut for the recommended automatic local-polynomial
    NOMAD route. When \code{TRUE}, any missing values among
    \code{regtype}, \code{search.engine}, \code{degree.select},
    \code{bernstein.basis}, \code{degree.min}, \code{degree.max},
    \code{degree.verify}, and \code{bwtype} are filled with
    \code{regtype="lp"}, \code{search.engine="nomad+powell"},
    \code{degree.select="coordinate"},
    \code{bernstein.basis=TRUE}, \code{degree.min=0L},
    \code{degree.max=10L}, \code{degree.verify=FALSE}, and
    \code{bwtype="fixed"}. Explicit incompatible settings error
    immediately; in particular, \code{nomad=TRUE} currently requires
    \code{regtype="lp"}, \code{bwtype="fixed"}, automatic degree
    search, \code{bernstein.basis=TRUE}, no explicit \code{degree},
    and \code{search.engine \%in\% c("nomad", "nomad+powell")}.
    This shortcut does not change the meaning of \code{nmulti} or
    \code{nomad.nmulti}: \code{nmulti} remains the outer restart count,
    while \code{nomad.nmulti} controls inner
    \code{crs::snomadr()} multistarts within each outer restart.
    Returned bandwidth objects retain this normalized preset metadata in
    \code{bw$nomad.shortcut} for a returned object \code{bw}; when
    available, \code{nomad.time} and \code{powell.time} record the
    direct-search and Powell-polish timing components.
  }

  \item{nomad.nmulti}{
    non-negative integer controlling the inner
    \code{crs::snomadr()} multistart count used within each outer
    NOMAD restart when \code{regtype="lp"} and automatic degree search
    uses \code{search.engine="nomad"} or \code{"nomad+powell"}.
    Defaults to \code{0L}, which preserves the current one-start-per-
    restart behavior. This does not replace \code{nmulti}:
    \code{nmulti} controls outer restarts, while \code{nomad.nmulti}
    controls inner NOMAD multistarts within each outer restart.
  }

  \item{nomad.remin}{
    logical flag controlling the optional second NOMAD hot start. When
    \code{TRUE}, NOMAD is restarted once from the best full candidate found,
    including both bandwidth and degree coordinates. Defaults to \code{FALSE};
    current simulation evidence favors the one-pass NOMAD default for routine
    use, while leaving this switch available for sensitivity checks.
  }


  \item{search.engine}{
    character string controlling the automatic local-polynomial search
    backend when \code{regtype="lp"} and
    \code{degree.select != "manual"}. \code{"nomad+powell"} (default)
    performs direct joint search over the \code{zdat}-side continuous
    bandwidth coordinates and degree vector using
    \code{crs::snomadr()}, then applies one Powell hot start from the
    NOMAD solution. \code{"nomad"} omits the Powell refinement.
    \code{"cell"} profiles the criterion over the admissible degree
    grid using repeated fixed-degree bandwidth solves. NOMAD-based
    search currently requires \code{bwtype="fixed"},
    \code{degree.verify=FALSE}, and the suggested package \pkg{crs} to
    be installed.
  }


\subsection{Numerical Search Controls}{
These controls set search restart behavior.
}

  \item{nmulti}{
    integer number of times to restart the process of finding extrema of
    the cross-validation function from different (random) initial
    points. Defaults to \code{min(2,ncol(xdat))}.
  }

  \item{powell.remin}{
    logical flag controlling Powell restart-from-minimum behavior. For
    ordinary fixed-degree Powell-style search, \code{TRUE} restarts the
    local search from the located minimum. For \code{search.engine="nomad+powell"},
    this controls only the final Powell bandwidth-polish step. The default
    is \code{TRUE} for ordinary Powell routes and \code{FALSE} for the
    Powell polish after NOMAD unless explicitly supplied.
  }


\subsection{Optimization Controls}{
These arguments control outer optimization behavior for the semiparametric search.
}

  \item{optim.abstol}{
    the absolute convergence tolerance used by \code{\link{optim}}. Only useful
    for non-negative functions, as a tolerance for reaching
    zero. Defaults to \code{.Machine$double.eps}.
  }

  \item{optim.maxattempts}{
    maximum number of attempts taken trying to achieve successful
    convergence in \code{\link{optim}}. Defaults to \code{100}.
  }

  \item{optim.maxit}{
    maximum number of iterations used by \code{\link{optim}}. Defaults
     to \code{500}.
   }

  \item{optim.method}{ method used by \code{\link{optim}} for minimization of
    the objective function. See \code{?optim} for references. Defaults
    to \code{"Nelder-Mead"}.

    the default method is an implementation of that of Nelder and Mead
    (1965), that uses only function values and is robust but relatively
    slow.  It will work reasonably well for non-differentiable
    functions.

    method \code{"BFGS"} is a quasi-Newton method (also known as a
    variable metric algorithm), specifically that published
    simultaneously in 1970 by Broyden, Fletcher, Goldfarb and Shanno.
    This uses function values and gradients to build up a picture of the
    surface to be optimized.

    method \code{"CG"} is a conjugate gradients method based
    on that by Fletcher and Reeves (1964) (but with the option of
    Polak-Ribiere or Beale-Sorenson updates).  Conjugate gradient
    methods will generally be more fragile than the BFGS method, but as
    they do not store a matrix they may be successful in much larger
    optimization problems.
  }

  \item{optim.reltol}{
    relative convergence tolerance used by \code{\link{optim}}.  The algorithm
    stops if it is unable to reduce the value by a factor of 'reltol *
    (abs(val) + reltol)' at a step.  Defaults to
    \code{sqrt(.Machine$double.eps)}, typically about \code{1e-8}.
  }

  \item{random.seed}{
    an integer used to seed R's random number generator. This ensures
    replicability of the numerical search. Defaults to 42.
  }


\subsection{Additional Arguments}{
These arguments collect remaining controls passed through S3 methods.
}

  \item{\dots}{
  additional arguments supplied to specify the regression type,
  bandwidth type, kernel types, selection methods, and so on, detailed
  below. 
  }

}

\details{

The \code{scale.factor.*} controls are dimensionless search
controls. The package converts scale factors to bandwidths using the
estimator-specific scaling encoded in the bandwidth object, including
kernel order and the number of continuous variables relevant for the
estimator. Users should not pre-multiply these controls by sample-size
or standard-deviation factors.

  \code{scale.factor.init} controls the deterministic first search
  start when that control is exposed. \code{scale.factor.init.lower}
  and \code{scale.factor.init.upper} define the random multistart
  interval when exposed. \code{scale.factor.search.lower} is the lower
  admissibility bound for continuous fixed-bandwidth search candidates.
  The effective first start is \code{max(scale.factor.init,
  scale.factor.search.lower)} when both controls are present, and the
  effective random-start lower endpoint is
  \code{max(scale.factor.init.lower, scale.factor.search.lower)}.
  \code{scale.factor.init.upper} must be at least that effective lower
  endpoint; the package errors rather than silently expanding the user's
  interval.

  When \code{scale.factor.search.lower} is \code{NULL}, an existing
  bandwidth object's stored floor is inherited when available;
  otherwise the package default \code{0.1} is used. Explicit bandwidths
  supplied for storage with \code{bandwidth.compute = FALSE} are not
  rewritten by the search floor.

  Categorical search-start controls such as \code{dfac.init},
  \code{lbd.init}, and \code{hbd.init} have separate semantics and are
  not affected by \code{scale.factor.search.lower}.

  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}}, \code{\link{plot.np}} for plotting options.

  For S3 plotting help, see \code{\link{plot.np}}. You can list
  available plot methods with \code{methods("plot")}.

  \code{npscoefbw} implements a variety of methods for semiparametric
 regression on multivariate (\eqn{p+q}-variate) explanatory data defined
 over a set of possibly continuous data. The approach is based on Li and
 Racine (2003) who employ \sQuote{generalized product kernels} that
 admit a mix of continuous and discrete data types.

 Three classes of kernel estimators for the continuous data types are
 available: fixed, adaptive nearest-neighbor, and generalized
 nearest-neighbor. Adaptive nearest-neighbor bandwidths change with
 each sample realization in the set, \eqn{x_i}{x[i]}, when estimating the
 density at the point \eqn{x}. Generalized nearest-neighbor bandwidths change
 with the point at which the density is estimated, \eqn{x}. Fixed bandwidths
 are constant over the support of \eqn{x}.

 \code{npscoefbw} may be invoked \emph{either} with a formula-like
 symbolic description of variables on which bandwidth selection is to be
 performed \emph{or} through a simpler interface whereby data is passed
 directly to the function via the \code{xdat}, \code{ydat}, and
 \code{zdat} parameters. Use of these two interfaces is \bold{mutually
 exclusive}.

 Data contained in the data frame \code{xdat} may be continuous and in
 \code{zdat} may be of mixed type. Data can be entered in an arbitrary
 order and data types will be detected automatically by the routine (see
 \code{\link{np}} for details).

 Data for which bandwidths are to be estimated may be specified
 symbolically. A typical description has the form \code{dependent}
 \code{data} \code{~} \code{parametric} \code{explanatory} \code{data}
 \code{|} \code{nonparametric} \code{explanatory} \code{data}, where
 \code{dependent} \code{data} is a univariate response, and
 \code{parametric} \code{explanatory} \code{data} and
 \code{nonparametric} \code{explanatory} \code{data} are both series of
 variables specified by name, separated by the separation character
 '+'. For example, \code{ y1 ~ x1 + x2 | z1 } specifies that the
 bandwidth object for the smooth coefficient model with response
 \code{y1}, linear parametric regressors \code{x1} and \code{x2}, and
 nonparametric regressor (that is, the slope-changing variable)
 \code{z1} is to be estimated. See below for further examples.  In the
 case where the nonparametric (slope-changing) variable is not
 specified, it is assumed to be the same as the parametric variable.

 A variety of kernels may be specified by the user. Kernels implemented
 for continuous data types include the second, fourth, sixth, and eighth
 order Gaussian and Epanechnikov kernels, and the uniform
 kernel. Unordered discrete data types use a variation on Aitchison and
 Aitken's (1976) kernel, while ordered data types use a variation of the
 Wang and van Ryzin (1981) kernel.

 Setting \code{nomad=TRUE} is a convenience preset for this automatic
 LP route, not a generic optimizer alias. For smooth coefficient
 regression it expands any missing values to the equivalent long-form
 call
 \preformatted{
   npscoefbw(...,
             regtype = "lp",
             search.engine = "nomad+powell",
             degree.select = "coordinate",
             bernstein.basis = TRUE,
             degree.min = 0L,
             degree.max = 10L,
             degree.verify = FALSE,
             bwtype = "fixed")
 }
 Compatible explicit tuning arguments are respected. Incompatible
 explicit settings fail fast so the shortcut never silently changes
 user-selected semantics.

  When \code{regtype="lp"} and \code{degree.select != "manual"},
  \code{npscoefbw} can jointly determine the \code{zdat}-side local
  polynomial degree vector together with the associated bandwidth
  coordinates. With \code{search.engine="cell"}, the criterion is
  profiled over the admissible degree grid using cached
  coordinate-wise or exhaustive search together with repeated
  fixed-degree bandwidth solves. With \code{search.engine="nomad"} or
  \code{"nomad+powell"}, the criterion is optimized directly over the
  joint degree/bandwidth space using \code{crs::snomadr()};
  \code{"nomad+powell"} then performs one Powell hot start from the
  NOMAD solution and keeps the better of the direct NOMAD and polished
  answers. This polynomial-adaptive joint-search route is motivated by
  Hall and Racine (2015). When \code{bernstein.basis} is not explicitly
  supplied, the automatic search route defaults to
  \code{bernstein.basis=TRUE} for numerical stability.

} 

\value{

 if \code{bwtype} is set to \code{fixed}, an object containing
 bandwidths (or scale factors if \code{bwscaling = TRUE}) is
 returned. If it is set to \code{generalized_nn} or \code{adaptive_nn},
 then instead the \eqn{k}th nearest neighbors are returned for the
 continuous variables while the discrete kernel bandwidths are returned
 for the discrete variables. Bandwidths are stored in a vector under the
 component name \code{bw}. Backfitted bandwidths are stored under the
 component name \code{bw.fitted}.

  The functions \code{\link{predict}}, \code{\link{summary}}, and
  \code{\link{plot}} support 
  objects of this class.

}

\section{Book And Method Pointers}{
  \code{npscoefbw} selects bandwidths for the smooth-coefficient model
  \eqn{Y=X^\prime\beta(Z)+\epsilon}. The selected bandwidths control
  the mixed-data smoothing over \eqn{Z} used to estimate the coefficient
  functions \eqn{\beta_j(z)}.

  For book-length derivations, see Li and Racine (2007), Chapter 9
  \emph{Additive and Smooth (Varying) Coefficient Semiparametric Models},
  especially Sections 9.3-9.3.4, and Racine (2019), Chapter 8
  \emph{Semiparametric Conditional Mean Function Estimation}.
}

\references{

 Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
   discrimination by the kernel method,} Biometrika, 63, 413-420.

 Cai Z. (2007), \dQuote{Trending time-varying coefficient time series
   models with serially correlated errors,} Journal of Econometrics,
 136, 163-188.

 Hastie, T. and R. Tibshirani (1993), \dQuote{Varying-coefficient
  models,} Journal of the Royal Statistical Society, B 55, 757-796.

 Hall, P. and J.S. Racine (2015), \dQuote{Infinite Order Cross-Validated
 Local Polynomial Regression,} Journal of Econometrics, 185, 510-525.

 Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
 and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2010), \dQuote{Smooth varying-coefficient
    estimation and inference for qualitative and quantitative data,}
    Econometric Theory, 26, 1-31.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,}
 Cambridge University Press.
 
 Li, Q. and D. Ouyang and J.S. Racine (2013), \dQuote{Categorical
 semiparametric varying-coefficient models,} Journal of Applied
 Econometrics, 28, 551-589.

 Li, A. and Q. Li and J.S. Racine (under revision), \dQuote{Boundary
 Adjusted, Polynomial Adaptive, Nonparametric Kernel Conditional
 Density Estimation,} Econometric Reviews.

 Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth estimators
 for discrete distributions,}  Biometrika, 68, 301-309.

}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.

  Caution: multivariate data-driven bandwidth selection methods are, by
  their nature, \emph{computationally intensive}. Virtually all methods
  require dropping the \eqn{i}th observation from the data set,
  computing an object, repeating this for all observations in the
  sample, then averaging each of these leave-one-out estimates for a
  \emph{given} value of the bandwidth vector, and only then repeating
  this a large number of times in order to conduct multivariate
  numerical minimization/maximization. Furthermore, due to the potential
  for local minima/maxima, \emph{restarting this procedure a large
  number of times may often be necessary}. This can be frustrating for
  users possessing large datasets. For exploratory purposes, you may
  wish to override the default search tolerances, say, setting
  optim.reltol=.1 and conduct multistarting (the default is to restart
  min(2,ncol(zdat)) times). Once the procedure terminates, you can restart
  search with default tolerances using those bandwidths obtained from
  the less rigorous search (i.e., set \code{bws=bw} on subsequent calls
  to this routine where \code{bw} is the initial bandwidth object).  A
  version of this package using the \code{Rmpi} wrapper is under
  development that allows one to deploy this software in a clustered
  computing environment to facilitate computation involving large
  datasets.

  Support for backfitted bandwidths is experimental and is limited in
  functionality. The code does not support asymptotic standard errors
  or out of sample estimates with backfitting.

}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}, \code{\link{plot.np}}
  \code{\link{npregbw}}, \code{\link{npreg}} }
\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA):
set.seed(42)

n <- 100
x <- runif(n)
z <- runif(n, min=-2, max=2)
y <- x*exp(z)*(1.0+rnorm(n,sd = 0.2))
bw <- npscoefbw(formula=y~x|z)
summary(bw)

# EXAMPLE 1 (INTERFACE=DATA FRAME): 

n <- 100
x <- runif(n)
z <- runif(n, min=-2, max=2)
y <- x*exp(z)*(1.0+rnorm(n,sd = 0.2))
bw <- npscoefbw(xdat=x, ydat=y, zdat=z)
summary(bw)
} % enddontrun
}
\keyword{nonparametric}

\name{dimBS}
\alias{dimBS}
\title{Local-Polynomial Basis Dimension Helper}
\description{
  \code{dimBS} returns the number of columns implied by an additive,
  generalized local-polynomial, or tensor-product basis specification.
  It is a compatibility wrapper around the internal \code{dim_basis()}
  helper used by \pkg{np}.
}
\usage{
dimBS(basis = "additive",
      kernel = TRUE,
      degree = NULL,
      segments = NULL,
      include = NULL,
      categories = NULL)
}
\arguments{
\subsection{Basis Specification}{
Basis family, continuous-kernel counting mode, polynomial degree, and
segment controls.
}

  \item{basis}{
    basis family. One of \code{"additive"}, \code{"glp"}, or
    \code{"tensor"}.
  }
  \item{kernel}{
    logical indicating whether only the continuous-kernel basis should
    be counted. When \code{FALSE}, optional categorical augmentation
    controlled by \code{include} and \code{categories} is included.
  }
  \item{degree}{
    non-negative integer vector of local-polynomial degrees.
  }
  \item{segments}{
    positive integer vector giving the number of segments for each
    continuous predictor. Defaults to one segment per degree entry.
  }

\subsection{Categorical Augmentation}{
Optional categorical-component controls used when \code{kernel = FALSE}.
}

  \item{include}{
    non-negative integer vector indicating which categorical
    components are included when \code{kernel = FALSE}.
  }
  \item{categories}{
    non-negative integer vector giving category counts for included
    categorical components when \code{kernel = FALSE}.
  }
}
\value{
  A numeric scalar giving the implied basis dimension.
}
\details{
  \code{dimBS()} is provided for compatibility with \pkg{crs}. In
  \pkg{np}, the underlying implementation lives in
  \code{dim_basis()}, which is used internally for LP basis-dimension
  checks and safe NOMAD restart initialization.
}
\examples{
dimBS(basis = "tensor", degree = c(2, 2))
dimBS(basis = "glp", degree = c(3, 1, 0))
}

\name{npcdisthat}
\alias{npcdisthat}
\title{Conditional Distribution Hat Operator}
\description{
Constructs the conditional distribution hat operator associated with
\code{\link{npcdist}} bandwidth objects. The returned operator maps a
right-hand side \code{y} to \eqn{H y}; with \code{y = 1} this reproduces the
fitted conditional distribution function.
}
\usage{
npcdisthat(bws,
           txdat = stop("training data 'txdat' missing"),
           tydat = stop("training data 'tydat' missing"),
           exdat,
           eydat,
           y = NULL,
           output = c("matrix", "apply"),
           deriv = NULL,
           s = NULL)
}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the fitted bandwidth object, training data, and evaluation data.
}

  \item{bws}{A fitted conditional distribution bandwidth object of class \code{"condbandwidth"}.}


  \item{exdat}{Optional evaluation conditioning data. If omitted, the operator is built on the training conditioning data.}


  \item{eydat}{Optional evaluation response data. If omitted, the operator is built on the training response data.}


  \item{txdat}{Training conditioning data used to construct the operator.}


  \item{tydat}{Training response data used to construct the operator.}


\subsection{Operator Output}{
These arguments control whether the operator is returned as a matrix or applied directly.
}

  \item{output}{Either \code{"matrix"} for the hat matrix or \code{"apply"} for direct application to \code{y}.}

  \item{y}{Optional right-hand side vector or matrix with one row per training observation.}


\subsection{Derivative Controls}{
These arguments request continuous-conditioning derivative hat operators.
}

  \item{deriv}{Optional derivative selector passed through to the continuous-conditioning derivative resolver.}

  \item{s}{Optional derivative multi-index for continuous conditioning variables. The default \code{s = NULL} returns the ordinary conditional-distribution hat operator.}


}
\details{
For \code{output = "matrix"}, the return value is a matrix with class
\code{c("npcdisthat", "matrix")} and attributes storing the bandwidth object,
training data, evaluation data, and call metadata.

For \code{output = "apply"}, the function returns \code{H y} directly. Matrix
right-hand sides are applied column-wise.

The optional \code{s} argument follows the derivative multi-index convention
used by \code{\link{npreghat}} for continuous conditioning variables. The
default \code{s = NULL} returns the ordinary conditional-distribution hat
operator. Supplying one first derivative, for example \code{s = 1L} in a
univariate continuous-\eqn{x} problem or \code{s = c(x2 = 1L)} in a
multivariate problem, returns the hat operator for that component of the
conditional-distribution gradient. This derivative route is intended to match
\code{npcdist(..., gradients = TRUE)$congrad}.

This helper is intended for object-fed repeated evaluation once a bandwidth
object has already been constructed. It does not perform bandwidth selection.
}
\value{
Either a hat matrix of class \code{"npcdisthat"} or the applied result
\code{H y}, depending on \code{output}.
}
\examples{
\dontrun{
data(cps71)
tx <- data.frame(age = cps71$age)
ty <- data.frame(logwage = cps71$logwage)

bw <- npcdistbw(xdat = tx, ydat = ty, bwtype = "fixed",
                bandwidth.compute = FALSE, bws = c(1.0, 1.0))

H <- npcdisthat(bws = bw, txdat = tx, tydat = ty)
dist.hat <- npcdisthat(bws = bw, txdat = tx, tydat = ty,
                       y = rep(1, nrow(tx)),
                       output = "apply")
dist.core <- fitted(npcdist(bws = bw, txdat = tx, tydat = ty))

head(cbind(dist.core, dist.hat), n = 2L)
}
}
\keyword{nonparametric}
\keyword{distribution}

\name{npsemihat}
\alias{npindexhat}
\alias{npplreghat}
\alias{npscoefhat}
\title{Experimental Hat Operators for Semiparametric Estimators}
\description{
Constructs hat operators for semiparametric estimators so that fitted values
or bootstrap draws can be computed by matrix application in one step.
These interfaces are currently experimental.
}
\usage{
npindexhat(bws,
           txdat = stop("training data 'txdat' missing"),
           exdat = txdat,
           y = NULL,
           output = c("matrix", "apply", "constraint"),
           s = 0L,
           fd.step = NULL,
           ...)

npplreghat(bws,
           txdat = stop("training data 'txdat' missing"),
           tzdat = stop("training data 'tzdat' missing"),
           exdat = txdat,
           ezdat = tzdat,
           y = NULL,
           output = c("apply", "matrix", "constraint"),
           ...)

npscoefhat(bws,
           txdat = stop("training data 'txdat' missing"),
           tzdat = NULL,
           exdat = txdat,
           ezdat = tzdat,
           y = NULL,
           output = c("matrix", "apply", "constraint"),
           ridge = 0,
           iterate = FALSE,
           leave.one.out = FALSE,
           ...)
}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the fitted bandwidth object, training data, and evaluation data.
}

  \item{bws}{A fitted bandwidth object. \code{npindexhat()} requires class \code{sibandwidth}, \code{npplreghat()} requires class \code{plbandwidth}, and \code{npscoefhat()} requires class \code{scbandwidth}.}


  \item{exdat}{Evaluation \code{x} data.}


  \item{txdat}{Training \code{x} data.}


\subsection{Operator Output And Derivatives}{
These arguments control operator output, derivative selection, finite-difference compatibility, and apply-mode right-hand sides.
}

  \item{fd.step}{Compatibility argument for \code{npindexhat(..., s=1)}. If supplied it must be a positive finite scalar, but the current \code{s=1} route uses the canonical exact derivative operator rather than finite-differencing the fit operator.}


  \item{output}{Either \code{"matrix"} for the hat matrix, \code{"apply"} for direct application to \code{y}, or \code{"constraint"} for the row-weighted transpose \code{t(H) * y}. Note that \code{npplreghat()} retains \code{"apply"} as its default for historical compatibility.}


  \item{s}{For \code{npindexhat}, \code{s=0} returns fit operator and \code{s=1} returns index-derivative operator.}


  \item{y}{Optional response vector or matrix for apply mode. For \code{output = "constraint"}, \code{y} is required and must be a vector or one-column object.}


\subsection{Partially Linear And Smooth-Coefficient Data}{
These arguments supply the additional \code{z} data used by partially linear and smooth-coefficient models.
}

  \item{ezdat}{Evaluation \code{z} data.}


  \item{tzdat}{Training \code{z} data for partially linear and smooth-coefficient models.}


\subsection{Smooth-Coefficient Controls}{
These arguments control smooth-coefficient iteration, leave-one-out behavior, and ridge stabilization.
}

  \item{iterate}{Logical; \code{npscoefhat()} currently supports \code{iterate = FALSE} only.}


  \item{leave.one.out}{Logical; leave-one-out kernel weights for \code{npscoefhat}. This currently requires evaluation \code{z} data to match training \code{z} data.}


  \item{ridge}{Base ridge term for local linear solves in \code{npscoefhat}; must be a non-negative finite scalar. The ridge sequence starts at \code{0} (no regularization) and then increments by \code{1/n.train} as needed for stable solves.}


\subsection{Additional Arguments}{
Reserved for future extensions.
}

  \item{...}{Reserved for future extensions.}

}
\value{
If \code{output = "matrix"}, returns a hat matrix \eqn{H}. If
\code{output = "apply"}, returns \eqn{H y} (or \eqn{H Y} for matrix
right-hand-side input). If \code{output = "constraint"}, returns the
constraint design matrix \eqn{\operatorname{diag}(y)H'} as represented in R by
\code{t(H) * y}.
}
\details{
These operators are intended for fixed-\eqn{X} workflows such as one-shot wild
bootstrap calculations where many response draws are projected through the same
operator. The implementation is intentionally conservative: class and scalar
argument contracts are validated explicitly, and unsupported iterative
\code{npscoefhat()} paths fail fast. \code{npscoefhat()} inherits
\code{regtype}/LP-basis controls from the supplied \code{scbandwidth}
object. For non-fixed \code{npscoef} bootstrap plotting, these operators can
support a frozen approximation, but they do not remove the need to recompute
the local smooth-coefficient vector itself for each resample: the local
weighted systems depend on the resample weights/counts at each evaluation
point, so unlike \code{npplreg} there is no single global coefficient vector
to update once per draw.

Method-specific argument map:
\code{npindexhat()} uses \code{s}; \code{fd.step} is accepted for compatibility but the current \code{s=1} route uses the canonical exact derivative operator;
\code{npplreghat()} and \code{npscoefhat()} use \code{tzdat}/\code{ezdat};
\code{npscoefhat()} additionally uses \code{ridge}, \code{iterate}, and
\code{leave.one.out}.

For \code{output = "constraint"}, the selected operator is first defined exactly
as in \code{output = "matrix"} and the return value is \code{t(H) * y}. This is
a convenience route for constrained quadratic-programming workflows; it does
not solve a constrained estimation problem and it does not change the meaning
of \code{output = "apply"}.
}
\examples{
\dontrun{
set.seed(42)
n <- 100
x <- runif(n)
z <- runif(n)
y <- sin(2*pi*x) + 0.5 * z + rnorm(n, sd = 0.1)

tx <- data.frame(x = x)
tz <- data.frame(z = z)

ibw <- npindexbw(xdat = data.frame(x, x2 = x^2), ydat = y,
                 bws = c(0.5, 1.0, 1.0), bandwidth.compute = FALSE)
iH <- npindexhat(bws = ibw, txdat = data.frame(x, x2 = x^2), output = "matrix")
iH.fitted <- iH %*% y
iA <- npindexhat(bws = ibw, txdat = data.frame(x, x2 = x^2),
                 y = y, output = "constraint")
all.equal(iA, t(iH) * y)
ifit <- npindex(bws = ibw, txdat = data.frame(x, x2 = x^2), tydat = y)
head(cbind(fitted(ifit), iH.fitted), n = 2L)

pbw <- npplregbw(xdat = tx, zdat = tz, ydat = y,
                 bws = matrix(c(0.2, 0.2), nrow = 2L, ncol = 1L),
                 bandwidth.compute = FALSE)
pH <- npplreghat(bws = pbw, txdat = tx, tzdat = tz, output = "matrix")
pH.fitted <- pH %*% y
pfit <- npplreg(bws = pbw, txdat = tx, tydat = y, tzdat = tz)
head(cbind(fitted(pfit), pH.fitted), n = 2L)

sbw <- npscoefbw(xdat = tx, zdat = tz, ydat = y,
                 bws = 0.2, bandwidth.compute = FALSE)
sH <- npscoefhat(bws = sbw, txdat = tx, tzdat = tz,
                 output = "matrix", iterate = FALSE)
sH.fitted <- sH %*% y
sfit <- npscoef(bws = sbw, txdat = tx, tydat = y, tzdat = tz,
                iterate = FALSE)
head(cbind(fitted(sfit), sH.fitted), n = 2L)
}
}
\keyword{nonparametric}
\keyword{semiparametric}
\keyword{regression}

\name{npregiv}
\alias{npregiv}

\title{
Nonparametric Instrumental Regression
}
\description{
\code{npregiv} computes nonparametric estimation of an instrumental
regression function \eqn{\varphi}{phi} defined by conditional moment
restrictions stemming from a structural econometric model: \eqn{E [Y -
\varphi (Z,X) | W ] = 0}{E [Y - phi (Z,X) | W ] = 0}, and involving
endogenous variables \eqn{Y} and \eqn{Z} and exogenous variables
\eqn{X} and instruments \eqn{W}. The function \eqn{\varphi}{phi} is the
solution of an ill-posed inverse problem.

When \code{method="Tikhonov"}, \code{npregiv} uses the approach of
Darolles, Fan, Florens and Renault (2011) modified for local
polynomial kernel regression of any order (Darolles et al use local
constant kernel weighting which corresponds to setting \code{p=0}; see
below for details). When \code{method="Landweber-Fridman"},
\code{npregiv} uses the approach of Horowitz (2011) again using local
polynomial kernel regression (Horowitz uses B-spline weighting).
}
\usage{
npregiv(y,
        z,
        w,
        x = NULL,
        zeval = NULL,
        xeval = NULL,
        alpha = NULL,
        alpha.iter = NULL,
        alpha.max = 1e-01,
        alpha.min = 1e-10,
        alpha.tol = .Machine$double.eps^0.25,
        bw = NULL,
        constant = 0.5,
        iterate.diff.tol = 1.0e-08,
        iterate.max = 1000,
        iterate.Tikhonov = TRUE,
        iterate.Tikhonov.num = 1,
        method = c("Landweber-Fridman","Tikhonov"),
        nmulti = NULL,
        optim.abstol = .Machine$double.eps,
        optim.maxattempts = 10,
        optim.maxit = 500,
        optim.method = c("Nelder-Mead", "BFGS", "CG"),
        optim.reltol = sqrt(.Machine$double.eps),
        p = 1,
        penalize.iteration = TRUE,
        random.seed = 42,
        return.weights.phi = FALSE,
        return.weights.phi.deriv.1 = FALSE,
        return.weights.phi.deriv.2 = FALSE,
        smooth.residuals = TRUE,
        start.from = c("Eyz","EEywz"),
        starting.values  = NULL,
        stop.on.increase = TRUE,
        \dots)
}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the response, endogenous variables, instruments, exogenous covariates, and evaluation data.
}

  \item{w}{
    a \eqn{q}-variate data frame of instruments. The data types may be
    continuous, discrete (unordered and ordered factors), or some
    combination thereof.
  }


  \item{x}{
    an \eqn{r}-variate data frame of exogenous regressors. The data
    types may be continuous, discrete (unordered and ordered factors),
    or some combination thereof.
  }


  \item{xeval}{
    an \eqn{r}-variate data frame of exogenous regressors on which the
    regression will be estimated (evaluation data). By default,
    evaluation takes place on the data provided by \code{x}.
  }



  \item{y}{
    a one (1) dimensional numeric or integer vector of dependent data, each
    element \eqn{i} corresponding to each observation (row) \eqn{i} of
    \code{z}.
  }



  \item{z}{
    a \eqn{p}-variate data frame of endogenous regressors. The data
    types may be continuous, discrete (unordered and ordered factors),
    or some combination thereof.
  }


  \item{zeval}{
    a \eqn{p}-variate data frame of endogenous regressors on which the
    regression will be estimated (evaluation data). By default, evaluation
    takes place on the data provided by \code{z}.
  }


\subsection{Landweber-Fridman Iteration Controls}{
These arguments control the Landweber-Fridman iteration path.
}

  \item{constant}{
    the constant to use when using  \code{method="Landweber-Fridman"}.
  }


  \item{iterate.diff.tol}{
    the search tolerance for the difference in the stopping rule from
    iteration to iteration when using \code{method="Landweber-Fridman"}
    (disable by setting to zero).
  }


  \item{iterate.max}{
    an integer indicating the maximum number of iterations permitted
    before termination occurs when using \code{method="Landweber-Fridman"}.
  }


  \item{iterate.Tikhonov}{
    a logical value indicating whether to use iterated Tikhonov (one
    iteration) or not when using \code{method="Tikhonov"}.
  }


  \item{iterate.Tikhonov.num}{
    an integer indicating the number of iterations to conduct when using
    \code{method="Tikhonov"}.
  }


  \item{method}{
    the regularization method employed (defaults to
    \code{"Landweber-Fridman"}, see Horowitz (2011); see Darolles,
    Fan, Florens and Renault (2011) for details for
    \code{"Tikhonov"}).
  }


  \item{nmulti}{
    integer number of times to restart the process of finding extrema of
    the cross-validation function from different (random) initial
    points.
  }



\subsection{Optimization Controls}{
These arguments control numerical optimization for the inverse problem.
}

  \item{optim.abstol}{
    the absolute convergence tolerance used by \code{\link{optim}}. Only useful
    for non-negative functions, as a tolerance for reaching
    zero. Defaults to \code{.Machine$double.eps}.
  }


  \item{optim.maxattempts}{
    maximum number of attempts taken trying to achieve successful
    convergence in \code{\link{optim}}. Defaults to \code{100}.
  }


  \item{optim.maxit}{
    maximum number of iterations used by \code{\link{optim}}. Defaults
     to \code{500}.
   }


  \item{optim.method}{ method used by \code{\link{optim}} for minimization of
    the objective function. See \code{?optim} for references. Defaults
    to \code{"Nelder-Mead"}.

    the default method is an implementation of that of Nelder and Mead
    (1965), that uses only function values and is robust but relatively
    slow.  It will work reasonably well for non-differentiable
    functions.

    method \code{"BFGS"} is quasi-Newton method (also known as a
    variable metric algorithm), specifically that published
    simultaneously in 1970 by Broyden, Fletcher, Goldfarb and Shanno.
    This uses function values and gradients to build up a picture of the
    surface to be optimized.

    method \code{"CG"} is a conjugate gradients method based
    on that by Fletcher and Reeves (1964) (but with the option of
    Polak-Ribiere or Beale-Sorenson updates).  Conjugate gradient
    methods will generally be more fragile than the BFGS method, but as
    they do not store a matrix they may be successful in much larger
    optimization problems.
  }


  \item{optim.reltol}{
    relative convergence tolerance used by \code{\link{optim}}.  The algorithm
    stops if it is unable to reduce the value by a factor of 'reltol *
    (abs(val) + reltol)' at a step.  Defaults to
    \code{sqrt(.Machine$double.eps)}, typically about \code{1e-8}.
  }


  \item{p}{
    the order of the local polynomial regression (defaults to
    \code{p=1}, i.e. local linear).
  }



\subsection{Returned Weights And Smooth Residuals}{
These arguments control returned kernel weights, starting values, residual smoothing, and iteration stopping behavior.
}

  \item{penalize.iteration}{ a logical value indicating whether to
    penalize the norm by the number of iterations or not (default
    \code{TRUE})
  }    


  \item{random.seed}{
    an integer used to seed R's random number generator. This ensures
    replicability of the numerical search. Defaults to 42.
  }


  \item{return.weights.phi}{
    a logical value (defaults to \code{FALSE}) indicating whether to
    return the weight matrix which when postmultiplied by the response
    \eqn{y} delivers the instrumental regression
  }


  \item{return.weights.phi.deriv.1}{
    a logical value (defaults to \code{FALSE}) indicating whether to
    return the weight matrix which when postmultiplied by the response
    \eqn{y} delivers the first partial derivative of the instrumental
    regression with respect to \eqn{z}
  }


  \item{return.weights.phi.deriv.2}{
    a logical value (defaults to \code{FALSE}) indicating whether to
    return the weight matrix which when postmultiplied by the response
    \eqn{y} delivers the second partial derivative of the instrumental
    regression with respect to \eqn{z}
  }


  \item{smooth.residuals}{ a logical value indicating whether to
    optimize bandwidths for the regression of
    \eqn{(y-\varphi(z))}{(y-phi(z))} on \eqn{w} (defaults to
    \code{TRUE}) or for the regression of \eqn{\varphi(z)}{phi(z)} on
    \eqn{w} during iteration
  }


  \item{start.from}{ a character string indicating whether to start from
    \eqn{E(Y|z)} (default, \code{"Eyz"}) or from \eqn{E(E(Y|z)|z)} (this can
    be overridden by providing \code{starting.values} below)
  }  


  \item{starting.values}{ a value indicating whether to commence
    Landweber-Fridman assuming
    \eqn{\varphi_{-1}=starting.values}{phi[-1]=starting.values} (proper
    Landweber-Fridman) or instead begin from \eqn{E(y|z)} (defaults to
    \code{NULL}, see details below)
  }


  \item{stop.on.increase}{
    a logical value (defaults to \code{TRUE}) indicating whether to halt
    iteration if the stopping criterion (see below) increases over the
    course of one iteration (i.e. it may be above the iteration tolerance
    but increased)
  }



\subsection{Tikhonov Regularization Controls}{
These arguments control Tikhonov regularization and its bandwidth.
}

  \item{alpha}{
    a numeric scalar that, if supplied, is used rather than numerically
    solving for \code{alpha}, when using \code{method="Tikhonov"}.
    }


  \item{alpha.iter}{
    a numeric scalar that, if supplied, is used for iterated Tikhonov
    rather than numerically solving for \code{alpha}, when using
    \code{method="Tikhonov"}.
  }


  \item{alpha.max}{
    maximum of search range for \eqn{\alpha}{alpha}, the Tikhonov
    regularization parameter, when using  \code{method="Tikhonov"}.
  }


  \item{alpha.min}{
    minimum of search range for \eqn{\alpha}{alpha}, the Tikhonov
    regularization parameter, when using \code{method="Tikhonov"}.
  }


  \item{alpha.tol}{
    the search tolerance for \code{optimize} when solving for
    \eqn{\alpha}{alpha}, the Tikhonov regularization parameter,
    when using \code{method="Tikhonov"}.
  }


  \item{bw}{
    an object which, if provided, contains bandwidths and parameters
    (obtained from a previous invocation of \code{npregiv}) required to
    re-compute the estimator without having to re-run cross-validation
    and/or numerical optimization which is particularly costly in this
    setting (see details below for an illustration of its use)
  }



\subsection{Additional Arguments}{
Further arguments are passed to lower-level kernel-sum and estimation routines.
}

  \item{\dots}{
    additional arguments supplied to \code{\link{npksum}}.
  }


}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.

  Tikhonov regularization requires computation of weight matrices of
  dimension \eqn{n\times n}{n x n} which can be computationally costly
  in terms of memory requirements and may be unsuitable for large
  datasets. Landweber-Fridman will be preferred in such settings as it
  does not require construction and storage of these weight matrices
  while it also avoids the need for numerical optimization methods to
  determine \eqn{\alpha}{alpha}.
  
  \code{method="Landweber-Fridman"} uses an optimal stopping rule based
  upon \eqn{||E(y|w)-E(\varphi_k(z,x)|w)||^2
  }{||E(y|w)-E(phi(z,x)|w)||^2 }. However, if local rather than global
  optima are encountered the resulting estimates can be overly noisy. To
  best guard against this eventuality set \code{nmulti} to a larger
  number than the default \code{nmulti=min(2,p)} for the first
  iteration, where \code{p} is the dimension of the current smoothing
  problem.

  Note that for subsequent Landweber-Fridman iterations, a \dQuote{warm
  start} strategy is employed. The optimal bandwidths from the previous
  iteration are used as starting values for the current iteration. The
  user-supplied \code{nmulti} is respected for all iterations. For
  iterations after the first successful one, these optimal bandwidths
  serve as the first of the multiple initial points (a warm start),
  while any remaining restarts are cold starts. If \code{nmulti} is not
  explicitly supplied by the user, it defaults to \code{min(2,p)} for the first
  iteration and to 1 for all subsequent iterations. This strategy
  provides a balance between computational efficiency and robustness,
  allowing the numerical optimizer to refine the structural bandwidths
  as the residuals evolve incrementally while still guarding against
  local optima.

  When using \code{method="Landweber-Fridman"}, iteration will terminate
  when either the change in the value of
  \eqn{||(E(y|w)-E(\varphi_k(z,x)|w))/E(y|w)||^2
  }{||(E(y|w)-E(phi(z,x)|w))/E(y|w)||^2 } from iteration to iteration is
  less than \code{iterate.diff.tol} or we hit \code{iterate.max} or
  \eqn{||(E(y|w)-E(\varphi_k(z,x)|w))/E(y|w)||^2
  }{||(E(y|w)-E(phi(z,x)|w))/E(y|w)||^2 } stops falling in value and
  starts rising.

  The option \code{bw=} would be useful, say, when bootstrapping is
  necessary. Note that when passing \code{bw}, it must be obtained from
  a previous invocation of \code{npregiv}. For instance, if
  \code{model.iv} was obtained from an invocation of \code{npregiv} with
  \code{method="Landweber-Fridman"}, then the following needs to be fed
  to the subsequent invocation of \code{npregiv}:

  \preformatted{

    model.iv <- npregiv(\dots)

    bw <- NULL
    bw$bw.E.y.w <- model.iv$bw.E.y.w
    bw$bw.E.y.z <- model.iv$bw.E.y.z
    bw$bw.resid.w <- model.iv$bw.resid.w
    bw$bw.resid.fitted.w.z <- model.iv$bw.resid.fitted.w.z
    bw$norm.index <- model.iv$norm.index

    foo <- npregiv(\dots,bw=bw)
  }

  If, on the other hand \code{model.iv} was obtained from an invocation
  of \code{npregiv} with \code{method="Tikhonov"}, then the following
  needs to be fed to the subsequent invocation of \code{npregiv}:

  \preformatted{

    model.iv <- npregiv(\dots)

    bw <- NULL
    bw$alpha <- model.iv$alpha
    bw$alpha.iter <- model.iv$alpha.iter
    bw$bw.E.y.w <- model.iv$bw.E.y.w
    bw$bw.E.E.y.w.z <- model.iv$bw.E.E.y.w.z
    bw$bw.E.phi.w <- model.iv$bw.E.phi.w
    bw$bw.E.E.phi.w.z <- model.iv$bw.E.E.phi.w.z

    foo <- npregiv(\dots,bw=bw)    

  }

  Or, if \code{model.iv} was obtained from an invocation of
  \code{npregiv} with either \code{method="Landweber-Fridman"} or
  \code{method="Tikhonov"}, then the following would also work:

  \preformatted{

    model.iv <- npregiv(\dots)

    foo <- npregiv(\dots,bw=model.iv)    

  }

  
  When exogenous predictors \code{x} (\code{xeval}) are passed, they are
  appended to both the endogenous predictors \code{z} and the
  instruments \code{w} as additional columns. If this is not desired,
  one can manually append the exogenous variables to \code{z} (or
  \code{w}) prior to passing \code{z} (or \code{w}), and then they will
  only appear among the \code{z} or \code{w} as desired.
  
}

\value{

  \code{npregiv} returns a \code{npregiv} object.  The generic
    functions \code{\link{print}}, \code{\link{summary}}, and
    \code{\link{plot}} support objects of this type.

  \code{npregiv} returns a list with components \code{phi},
  \code{phi.mat} and either \code{alpha} when \code{method="Tikhonov"}
  or \code{norm.index}, \code{norm.stop} and \code{convergence} when
  \code{method="Landweber-Fridman"}, among others.

  In addition, if any of \code{return.weights.*} are invoked
  (\code{*=1,2}), then \code{phi.weights} and \code{phi.deriv.*.weights}
  return weight matrices for computing the instrumental regression and
  its partial derivatives. Note that these weights, post multiplied by
  the response vector \eqn{y}, will deliver the estimates returned in
  \code{phi}, \code{phi.deriv.1}, and \code{phi.deriv.2} (the latter
  only being produced when \code{p} is 2 or greater). When invoked with
  evaluation data, similar matrices are returned but named
  \code{phi.eval.weights} and \code{phi.deriv.eval.*.weights}. These
  weights can be used for constrained estimation, among others.

  When \code{method="Landweber-Fridman"} is invoked, bandwidth objects
  are returned in \code{bw.E.y.w} (scalar/vector), \code{bw.E.y.z}
  (scalar/vector), and \code{bw.resid.w} (matrix) and
  \code{bw.resid.fitted.w.z}, the latter matrices containing bandwidths
  for each iteration stored as rows. When \code{method="Tikhonov"} is
  invoked, bandwidth objects are returned in \code{bw.E.y.w},
  \code{bw.E.E.y.w.z}, and \code{bw.E.phi.w} and \code{bw.E.E.phi.w.z}.
  
}

\references{
  
  Carrasco, M. and J.P. Florens and E. Renault (2007), \dQuote{Linear
  Inverse Problems in Structural Econometrics Estimation Based on
  Spectral Decomposition and Regularization,} In: James J. Heckman and
  Edward E. Leamer, Editor(s), Handbook of Econometrics, Elsevier, 2007,
  Volume 6, Part 2, Chapter 77, Pages 5633-5751

  Darolles, S. and Y. Fan and J.P. Florens and E. Renault (2011),
  \dQuote{Nonparametric instrumental regression,} Econometrica, 79,
  1541-1565.
  
  Feve, F. and J.P. Florens (2010), \dQuote{The practice of
  non-parametric estimation by solving inverse problems: the example of
  transformation models,} Econometrics Journal, 13, S1-S27.

  Florens, J.P. and J.S. Racine and S. Centorrino (2018),
  \dQuote{Nonparametric instrumental derivatives,} Journal of
  Nonparametric Statistics, 30 (2), 368-391.

  Fridman, V. M. (1956), \dQuote{A method of successive approximations
  for Fredholm integral equations of the first kind,} Uspeskhi,
  Math. Nauk., 11, 233-334, in Russian.

  Horowitz, J.L. (2011), \dQuote{Applied nonparametric instrumental
  variables estimation,} Econometrica, 79, 347-394.
  
  Landweber, L. (1951), \dQuote{An iterative formula for Fredholm
  integral equations of the first kind,} American Journal of
  Mathematics, 73, 615-24.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics:
  Theory and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2004), \dQuote{Cross-validated Local Linear
  Nonparametric Regression,} Statistica Sinica, 14, 485-512.

}
\author{
  Jeffrey S. Racine \email{racinej@mcmaster.ca}, Samuele Centorrino
 \email{samuele.centorrino@univ-tlse1.fr}
}
\note{
This function should be considered to be in \sQuote{beta test} status until further notice.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}
\code{\link{npregivderiv},\link{npreg}}
}
\examples{
\dontrun{
## This illustration was made possible by Samuele Centorrino
## <samuele.centorrino@univ-tlse1.fr>

set.seed(42)
n <- 500

## The DGP is as follows:

## 1) y = phi(z) + u

## 2) E(u|z) != 0 (endogeneity present)

## 3) Suppose there exists an instrument w such that z = f(w) + v and
## E(u|w) = 0

## 4) We generate v, w, and generate u such that u and z are
## correlated. To achieve this we express u as a function of v (i.e. u =
## gamma v + eps)

v <- rnorm(n,mean=0,sd=0.27)
eps <- rnorm(n,mean=0,sd=0.05)
u <- -0.5*v + eps
w <- rnorm(n,mean=0,sd=1)

## In Darolles et al (2011) there exist two DGPs. The first is
## phi(z)=z^2 and the second is phi(z)=exp(-abs(z)) (which is
## discontinuous and has a kink at zero).

fun1 <- function(z) { z^2 }
fun2 <- function(z) { exp(-abs(z)) }

z <- 0.2*w + v

## Generate two y vectors for each function.

y1 <- fun1(z) + u
y2 <- fun2(z) + u

## You set y to be either y1 or y2 (ditto for phi) depending on which
## DGP you are considering:

y <- y1
phi <- fun1

## Sort on z (for plotting)

ivdata <- data.frame(y,z,w)
ivdata <- ivdata[order(ivdata$z),]
rm(y,z,w)
with(ivdata, {

model.iv <- npregiv(y=y,z=z,w=w)
phi.iv <- model.iv$phi

## Now the non-iv local linear estimator of E(y|z)

ll.mean <- fitted(npreg(y~z,regtype="ll"))

## For the plots, restrict focal attention to the bulk of the data
## (i.e. for the plotting area trim out 1/4 of one percent from each
## tail of y and z)

trim <- 0.0025

curve(phi,min(z),max(z),
      xlim=quantile(z,c(trim,1-trim)),
      ylim=quantile(y,c(trim,1-trim)),
      ylab="Y",
      xlab="Z",
      main="Nonparametric Instrumental Kernel Regression",
      lwd=2,lty=1)

points(z,y,type="p",cex=.25,col="grey")

lines(z,phi.iv,col="blue",lwd=2,lty=2)

lines(z,ll.mean,col="red",lwd=2,lty=4)

legend("topright",
       c(expression(paste(varphi(z))),
         expression(paste("Nonparametric ",hat(varphi)(z))),
         "Nonparametric E(y|z)"),
       lty=c(1,2,4),
       col=c("black","blue","red"),
       lwd=c(2,2,2),
       bty="n")
})
} % enddontrun
}
\keyword{ instrument }

%
\name{npscoef}
\alias{npscoef}
\alias{npscoef.default}
\alias{npscoef.formula}
\alias{npscoef.scbandwidth}
\title{Smooth Coefficient Kernel Regression}

\description{
  \code{npscoef} computes a kernel regression estimate of a one (1)
  dimensional dependent variable on \eqn{p}-variate explanatory data,
  using the model \eqn{Y_i = W_{i}^{\prime} \gamma (Z_i) + u_i}{Y_i =
  t(W_i) * gamma(Z_i) + u_i} where
  \eqn{W_i'=(1,X_i')}{t(W_i)=(1,t(X_i))}, given a set of evaluation
  points, training points (consisting of explanatory data and dependent
  data), and a bandwidth specification. A bandwidth specification can be
  a \code{scbandwidth} object, or a bandwidth vector, bandwidth type and
  kernel type.
}

\usage{
npscoef(bws, \dots)

\method{npscoef}{formula}(bws, 
        data = NULL, 
        newdata = NULL, 
        y.eval = FALSE, 
        \dots)

\method{npscoef}{default}(bws, 
        txdat, 
        tydat, 
        tzdat, 
        nomad = FALSE, 
        \dots)

\method{npscoef}{scbandwidth}(bws,
        txdat = stop("training data 'txdat' missing"),
        tydat = stop("training data 'tydat' missing"),
        tzdat = NULL,
        exdat,
        eydat,
        ezdat,
        betas = FALSE,
        errors = TRUE,
        iterate = TRUE,
        leave.one.out = FALSE,
        maxiter = 100,
        residuals = FALSE,
        tol = .Machine$double.eps,
        \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the bandwidth specification, formula/data interface, and smooth-coefficient training data.
}

  \item{bws}{
  a bandwidth specification. This can be set as a \code{scbandwidth}
  object returned from an invocation of \code{\link{npscoefbw}}, or
    as a vector of bandwidths, with each element \eqn{i} corresponding
    to the bandwidth for column \eqn{i} in \code{tzdat}. If specified as
    a vector
    additional arguments will need to be supplied as necessary to
    specify the bandwidth type, kernel types, training data, and so on.

  }



  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing
    the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(bws)}, typically the environment from which
   \code{\link{npscoefbw}} was called.
  }


  \item{txdat}{
    a \eqn{p}-variate data frame of explanatory data (training data),
    which, by default, populates the columns \eqn{2} through \eqn{p+1}
    of \eqn{W} in the model equation, and in the
    absence of \code{zdat}, will also correspond to
    \eqn{Z} from the model equation. Defaults to
    the training data used to
    compute the bandwidth object.
  }



  \item{tydat}{
    a one (1) dimensional numeric or integer vector of dependent data, each
    element \eqn{i} corresponding to each observation (row) \eqn{i} of
    \code{txdat}. Defaults to
    the training data used to
    compute the bandwidth object.
  }


  \item{tzdat}{
    an optionally specified \eqn{q}-variate data frame of explanatory data
    (training data), which corresponds to \eqn{Z} in the model
    equation. Defaults to
    the training data used to
    compute the bandwidth object.
  }


\subsection{Local-Polynomial Degree And Bandwidth Search}{
This argument controls the recommended automatic local-polynomial NOMAD route, which jointly selects continuous polynomial degree and bandwidths when these are computed inside \code{npscoef}.
}

  \item{nomad}{
    logical shortcut passed through to \code{\link{npscoefbw}} when
    bandwidths are computed inside \code{npscoef}. When \code{TRUE},
    the smooth-coefficient bandwidth route fills any missing values
    among \code{regtype}, \code{search.engine},
    \code{degree.select}, \code{bernstein.basis},
    \code{degree.min}, \code{degree.max}, \code{degree.verify}, and
    \code{bwtype} with the recommended automatic local-polynomial degree-and-bandwidth NOMAD preset
    documented in \code{\link{npscoefbw}}. Additional NOMAD tuning
    arguments such as \code{nomad.nmulti} may also be supplied through
    \code{\dots}; \code{nmulti} remains the outer restart count while
    \code{nomad.nmulti} controls inner
    \code{crs::snomadr()} multistarts within each outer restart. After fitting, inspect
    \code{fit$bws$nomad.shortcut} on the returned object \code{fit} to
    see the normalized shortcut metadata.
  }


\subsection{Evaluation Data And Returned Quantities}{
These arguments control where the smooth-coefficient fit is evaluated and which evaluation quantities are returned.
}

  \item{exdat}{
    a \eqn{p}-variate data frame of points on which the regression will be
    estimated (evaluation data).By default,
    evaluation takes place on the data provided by \code{txdat}.
  }


  \item{eydat}{
    a one (1) dimensional numeric or integer vector of the true values
    of the dependent variable. Optional, and used only to calculate the
    true errors.
  }


  \item{ezdat}{
    an optionally specified \eqn{q}-variate data frame of points on
    which
    the regression will be estimated
    (evaluation data), which corresponds to \eqn{Z}
    in the model equation. Defaults to be the same as \code{txdat}.
  }


  \item{newdata}{
    An optional data frame in which to look for evaluation data. If
    omitted, the training data are used.
  }


  \item{y.eval}{
    If \code{newdata} contains dependent data and \code{y.eval = TRUE},
    \code{\link{np}} will compute goodness of fit statistics on these
    data and return them. Defaults to \code{FALSE}.
  }


\subsection{Fitted Quantities And Backfitting}{
These arguments control returned coefficient estimates, errors, residuals, and iterative backfitting.
}

  \item{betas}{
    a logical value indicating whether or not estimates of the
    components of \eqn{\gamma}{gamma} should be returned in the
    \code{smoothcoefficient} object along with the regression estimates.
    Defaults to \code{FALSE}.
  }


  \item{errors}{
    a logical value indicating whether or not asymptotic standard errors
    should be computed and returned in the resulting
    \code{smoothcoefficient} object. Defaults to \code{TRUE}.
  }


  \item{iterate}{
    a logical value indicating whether or not backfitted estimates
    should be iterated for self-consistency. Defaults to \code{TRUE}.
  }


  \item{leave.one.out}{
    a logical value to specify whether or not to compute the leave one
    out estimates. Will not work if \code{e[xyz]dat} is specified. Defaults to
    \code{FALSE}.
  }


  \item{maxiter}{
    integer specifying the maximum number of times to iterate the
    backfitted estimates while attempting to make the backfitted estimates
    converge to the desired tolerance. Defaults to \code{100}.
  }


  \item{residuals}{
    a logical value indicating that you want residuals computed and
    returned in the resulting \code{smoothcoefficient} object. Defaults to
    \code{FALSE}.
  }


  \item{tol}{
    desired tolerance on the relative convergence of backfit
    estimates. Defaults to \code{.Machine$double.eps}.
  }


\subsection{Additional Arguments}{
Further arguments are passed to the bandwidth-selection counterpart when bandwidths are not supplied.
}

  \item{\dots}{
    additional arguments supplied to specify the regression type,
    bandwidth type, kernel types, selection methods, and so on.
    To do this, you may specify any of \code{bwscaling}, \code{bwtype},
    \code{ckertype}, \code{ckerorder}, as described in
    \code{\link{npscoefbw}}.

  }
}
\value{
  \code{npscoef} returns a \code{smoothcoefficient} object.   The generic
  functions \code{\link{fitted}}, \code{\link{residuals}}, \code{\link{coef}},
  \code{\link{se}}, and \code{\link{predict}},
  extract (or generate) estimated values,
  residuals, coefficients, bootstrapped standard
  errors on estimates, and predictions, respectively, from
  the returned object. Furthermore, the functions \code{\link{summary}}
  and \code{\link{plot}} support objects of this type. The returned object
  has the following components:

  \item{eval}{ evaluation points }
  \item{mean}{ estimation of the regression function (conditional mean) at the
    evaluation points }
  \item{merr}{ if \code{errors = TRUE}, standard errors of the
    regression estimates }
  \item{beta}{ if \code{betas = TRUE}, estimates of the coefficients
    \eqn{\gamma}{gamma} at the
    evaluation points }
  \item{grad}{ estimated derivatives of the conditional mean with
    respect to the regressors in \code{xdat}; these correspond to the
    non-intercept smooth coefficient estimates at each evaluation point}
  \item{gerr}{ if \code{errors = TRUE}, asymptotic standard errors for
    \code{grad} }
  \item{resid}{ if \code{residuals = TRUE}, in-sample or out-of-sample
    residuals where appropriate (or possible) }
  \item{R2}{ coefficient of determination (Doksum and Samarov (1995))}
  \item{MSE}{ mean squared error }
  \item{MAE}{ mean absolute error }
  \item{MAPE}{ mean absolute percentage error }
  \item{CORR}{ absolute value of Pearson's correlation coefficient }
  \item{SIGN}{ fraction of observations where fitted and observed values
  agree in sign }

}
\section{Book And Method Pointers}{
  The smooth-coefficient model lets slopes vary with conditioning
  variables, typically written
  \eqn{Y=X^\prime\beta(Z)+\epsilon}. The functions estimate the
  coefficient functions \eqn{\beta_j(z)} using mixed-data kernel
  smoothing over \eqn{Z}, with the reported fitted values obtained by
  combining the estimated coefficient functions with the corresponding
  columns of \eqn{X}.

  For book-length derivations, see Li and Racine (2007), Chapter 9
  \emph{Additive and Smooth (Varying) Coefficient Semiparametric Models},
  especially Sections 9.3-9.3.4, and Racine (2019), Chapter 8
  \emph{Semiparametric Conditional Mean Function Estimation}, especially
  the varying-coefficient material.
}

\references{

  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Cai Z. (2007), \dQuote{Trending time-varying coefficient time series
    models with serially correlated errors,} Journal of Econometrics,
  136, 163-188.

  Doksum, K. and A. Samarov (1995), \dQuote{Nonparametric estimation of
  global functionals and a measure of the explanatory power of
  covariates in regression,} The Annals of Statistics, 23 1443-1473.

  Hastie, T. and R. Tibshirani (1993), \dQuote{Varying-coefficient
    models,} Journal of the Royal Statistical Society, B 55, 757-796.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics:
    Theory and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2010), \dQuote{Smooth varying-coefficient
    estimation and inference for qualitative and quantitative data,}
    Econometric Theory, 26, 1-31.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,}
  Cambridge University Press.

 Li, Q. and D. Ouyang and J.S. Racine (2013), \dQuote{Categorical
 semiparametric varying-coefficient models,} Journal of Applied
 Econometrics, 28, 551-589.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth
    estimators for discrete distributions,}  Biometrika, 68, 301-309.

}

\author{
  Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.

  For practitioners who want the recommended automatic local-polynomial degree-and-bandwidth NOMAD route
  without spelling out all LP tuning arguments,
  \code{npscoef(..., nomad=TRUE)} and \code{npscoefbw(..., nomad=TRUE)}
  expand missing settings to the same documented preset. Explicit
  incompatible settings fail fast rather than being silently rewritten.

  For plotting options for fitted smooth-coefficient objects and their
  bandwidth objects, see \code{\link{plot.np}}.

  Support for backfitted bandwidths is experimental and is limited in
  functionality. The code does not support asymptotic standard errors
  or out of sample estimates with backfitting.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}, \code{\link{plot.np}},
  \code{\link{bw.nrd}}, \code{\link{bw.SJ}}, \code{\link{hist}},
  \code{\link{npudens}}, \code{\link{npudist}},
  \code{\link{npudensbw}}, \code{\link{npscoefbw}} }

\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA):

n <- 250
x <- runif(n)
z <- runif(n, min=-2, max=2)
y <- x*exp(z)*(1.0+rnorm(n,sd = 0.2))
bw <- npscoefbw(y~x|z)
model <- npscoef(bw)
if (interactive()) plot(model)

# EXAMPLE 1 (INTERFACE=DATA FRAME):

n <- 250
x <- runif(n)
z <- runif(n, min=-2, max=2)
y <- x*exp(z)*(1.0+rnorm(n,sd = 0.2))
bw <- npscoefbw(xdat=x, ydat=y, zdat=z)
model <- npscoef(bw)
if (interactive()) plot(model)
} % enddontrun
}
\keyword{nonparametric}

\name{npuniden.reflect}
\alias{npuniden.reflect}

\title{ Kernel Bounded Univariate Density Estimation Via Data-Reflection }

\description{

  \code{npuniden.reflect} computes kernel univariate unconditional
  density estimates given a vector of continuously distributed training
  data and, optionally, a bandwidth (otherwise likelihood
  cross-validation is used for its selection). Lower and upper bounds
  [\code{a},\code{b}] can be supplied (default is [0,1]) and if \code{a}
  is set to \code{-Inf} there is only one bound on the right, while if
  \code{b} is set to \code{Inf} there is only one bound on the left.

}

\usage{
npuniden.reflect(X = NULL,
                 Y = NULL,
                 h = NULL,
                 a = 0,
                 b = 1,
                 ...)
}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify evaluation points, observations, support bounds, and optional bandwidths.
}

  \item{a}{
an optional lower bound (defaults to 0)
}

  \item{b}{
an optional upper bound (defaults to 1)
}


  \item{h}{
an optional bandwidth (>0)
}

  \item{X}{
a required numeric vector of training data lying in \eqn{[a,b]}
}

  \item{Y}{
an optional numeric vector of evaluation data lying in \eqn{[a,b]}
}

\subsection{Additional Arguments}{
Further arguments are passed to \code{\link{npudensbw}} and \code{\link{npudens}}.
}

 \item{...}{
optional arguments passed to \code{npudensbw} and \code{npudens}
}

}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.

  Typical usages are (see below for a complete list of options and also
  the examples at the end of this help file)
  \preformatted{
    model <- npuniden.reflect(X,a=-2,b=3)
  }

  \code{npuniden.reflect} implements the data-reflection method for
  estimating a univariate density function defined over a continuous
  random variable in the presence of bounds.

  Note that data-reflection imposes a zero derivative at the boundary,
  i.e., \eqn{f'(a)=f'(b)=0}.
  
}
\value{
  \code{npuniden.reflect} returns the following components:
  
  \item{f}{ estimated density at the points X }
  \item{F}{ estimated distribution at the points X (numeric integral of f) }
  \item{sd.f}{ asymptotic standard error of the estimated density at the points X }
  \item{sd.F}{ asymptotic standard error of the estimated distribution at the points X }
  \item{h}{ bandwidth used }
  \item{nmulti}{ number of multi-starts used }
}
\references{

  Boneva, L. I., Kendall, D., and Stefanov, I. (1971). \dQuote{Spline
  transformations: Three new diagnostic aids for the statistical data-
  analyst,} Journal of the Royal Statistical Society. Series B
  (Methodological), 33(1):1-71.

  Cline, D. B. H. and Hart, J. D. (1991). \dQuote{Kernel estimation of
  densities with discontinuities or discontinuous
  derivatives,} Statistics, 22(1):69-84.

  Hall, P. and Wehrly, T. E. (1991). \dQuote{A geometrical method for
  removing edge effects from kernel- type nonparametric regression
  estimators,} Journal of the American Statistical Association,
  86(415):665-672.
  
}
\author{
Jeffrey S. Racine \email{racinej@mcmaster.ca}
}
\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}
The \pkg{Ake}, \pkg{bde}, and \pkg{Conake} packages and the function \code{\link{npuniden.boundary}}.
}
\examples{
\dontrun{
## Example 1: f(0)=0, f(1)=1, plot boundary corrected density,
## unadjusted density, and DGP
set.seed(42)
n <- 100
X <- sort(rbeta(n,5,1))
dgp <- dbeta(X,5,1)
model <- npuniden.reflect(X)
model.unadjusted <- npuniden.boundary(X,a=-Inf,b=Inf)
ylim <- c(0,max(c(dgp,model$f,model.unadjusted$f)))
if (interactive()) plot(X,model$f,ylab="Density",ylim=ylim,type="l")
lines(X,model.unadjusted$f,lty=2,col=2)
lines(X,dgp,lty=3,col=3)
rug(X)
legend("topleft",c("Data-Reflection","Unadjusted","DGP"),col=1:3,lty=1:3,bty="n")

## Example 2: f(0)=0, f(1)=0, plot density, distribution, DGP, and
## asymptotic point-wise confidence intervals
set.seed(42)
X <- sort(rbeta(100,5,3))
model <- npuniden.reflect(X)
oldpar <- par(no.readonly = TRUE)
on.exit(par(oldpar), add = TRUE)
par(mfrow=c(1,2))
ylim=range(c(model$f,model$f+1.96*model$sd.f,model$f-1.96*model$sd.f,dbeta(X,5,3)))
if (interactive()) plot(X,model$f,ylim=ylim,ylab="Density",type="l",)
lines(X,model$f+1.96*model$sd.f,lty=2)
lines(X,model$f-1.96*model$sd.f,lty=2)
lines(X,dbeta(X,5,3),col=2)
rug(X)
legend("topleft",c("Density","DGP"),lty=c(1,1),col=1:2,bty="n")

if (interactive()) plot(X,model$F,ylab="Distribution",type="l")
lines(X,model$F+1.96*model$sd.F,lty=2)
lines(X,model$F-1.96*model$sd.F,lty=2)
lines(X,pbeta(X,5,3),col=2)
rug(X)
legend("topleft",c("Distribution","DGP"),lty=c(1,1),col=1:2,bty="n")


## Example 3: Age for working age males in the cps71 data set bounded
## below by 21 and above by 65
data(cps71)
with(cps71, {
model <- npuniden.reflect(age,a=21,b=65)
par(mfrow=c(1,1))
hist(age,prob=TRUE,main="",ylim=c(0,max(model$f)))
lines(age,model$f)
lines(density(age,bw=model$h),col=2)
legend("topright",c("Data-Reflection","Unadjusted"),lty=c(1,1),col=1:2,bty="n")
})
} % enddontrun
}
\keyword{ nonparametric }
\keyword{ smooth }

%
\name{npudens}
\alias{npudens}
\alias{npudens.bandwidth}
\alias{npudens.default}
\alias{npudens.formula}

\title{ Kernel Density Estimation with Mixed Data Types }
\description{

  \code{npudens} computes  kernel unconditional density estimates on
  evaluation data, given a set of training data and a bandwidth
  specification (a \code{bandwidth} object or a bandwidth vector,
  bandwidth type, and kernel type) using the method of Li and Racine
  (2003).

}
\usage{
npudens(bws, \dots)

\method{npudens}{formula}(bws, data = NULL, newdata = NULL, \dots)

\method{npudens}{bandwidth}(bws,
        tdat = stop("invoked without training data 'tdat'"),
        edat,
        \dots)


\method{npudens}{default}(bws, tdat, \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the bandwidth specification, formula/data interface, and training data.
}

  \item{bws}{
    a bandwidth specification. This can be set as a \code{bandwidth}
    object returned from an invocation of \code{\link{npudensbw}}, or as a
    \eqn{p}-vector of bandwidths, with an element for each variable in the
    training data. If specified as a vector, then additional arguments
    will need to be supplied as necessary to change them from the
    defaults to specify the bandwidth type, kernel types, training data,
    and so on.
  }



  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(bws)}, typically the environment from which
    \code{\link{npudensbw}} was called.
  }


  \item{tdat}{
    a \eqn{p}-variate data frame of sample realizations (training data)
    used to estimate the density. Defaults to the training data used to
    compute the bandwidth object.
  }


\subsection{Evaluation Data}{
These arguments control where the fitted density is evaluated.
}

  \item{edat}{
    a \eqn{p}-variate data frame of density evaluation points. By default,
    evaluation takes place on the data provided by \code{tdat}.
  }


  \item{newdata}{
    An optional data frame in which to look for evaluation data. If
    omitted, the training data are used.
  }


\subsection{Additional Arguments}{
Further arguments are passed to \code{\link{npudensbw}} when bandwidths are computed internally, or used to interpret a numeric \code{bws} vector.
}

  \item{\dots}{
    additional arguments supplied to \code{\link{npudensbw}} when
    \code{npudens} computes bandwidths internally, or arguments needed
    to interpret a numeric \code{bws} vector. This is where bandwidth
    selection controls such as \code{bwmethod}, \code{bwtype},
    \code{bwscaling}, kernel/support controls such as \code{ckertype},
    \code{ckerorder}, and \code{ckerbound}, categorical kernel controls
    such as \code{ukertype} and \code{okertype}, and search controls
    such as \code{nmulti} and \code{scale.factor.search.lower} are
    supplied. See \code{\link{npudensbw}} for the complete
    bandwidth-selection argument surface.
  }
}
\details{
  Documentation guide: see \code{\link{npudensbw}} for bandwidth
  selection and search controls, \code{\link{np.kernels}} for kernels,
  \code{\link{np.options}} for global options, and \code{\link{plot}},
  \code{\link{plot.np}} for plotting options.

  When \code{bws} is omitted, the formula and default methods call
  \code{\link{npudensbw}} first and pass bandwidth-selection arguments
  from \code{\dots} to that call. When \code{bws} is already a
  \code{bandwidth} object, \code{npudens} estimates with the stored
  bandwidth metadata in that object.

  Argument groups for bandwidth selection are documented on
  \code{\link{npudensbw}}. The most common workflow is to choose data
  and bandwidth inputs first, then bandwidth criterion and
  representation, then kernel/support controls and numerical search
  controls if defaults need to be changed.

  For S3 plotting help, see \code{\link{plot.np}}. You can list
  available plot methods with \code{methods("plot")}.


  Typical usages are (see below for a complete list of options and also
  the examples at the end of this help file)
  \preformatted{
    Usage 1: first compute the bandwidth object via npudensbw and then
    compute the density:

    bw <- npudensbw(~y)
    fhat <- npudens(bw)

    Usage 2: alternatively, compute the bandwidth object indirectly:

    fhat <- npudens(~y)

    Usage 3: modify the default kernel and order:

    fhat <- npudens(~y, ckertype="epanechnikov", ckerorder=4)

    Usage 4: use the data frame interface rather than the formula
    interface:

    fhat <- npudens(tdat = y, ckertype="epanechnikov", ckerorder=4)
  }

  \code{npudens} implements a variety of methods for estimating
  multivariate density functions (\eqn{p}-variate) defined over a set of
  possibly continuous and/or discrete (unordered, ordered) data. The
  approach is based on Li and Racine (2003) who employ
  \sQuote{generalized product kernels} that admit a mix of continuous
  and discrete data types.

  Three classes of kernel estimators for the continuous data types are
  available: fixed, adaptive nearest-neighbor, and generalized
  nearest-neighbor. Adaptive nearest-neighbor bandwidths change with
  each sample realization in the set, \eqn{x_i}{x[i]}, when estimating
  the density at the point \eqn{x}. Generalized nearest-neighbor
  bandwidths change with the point at which the density is estimated,
  \eqn{x}. Fixed bandwidths are constant over the support of \eqn{x}.

  Data contained in the data frame \code{tdat} (and also \code{edat})
  may be a mix of continuous (default), unordered discrete (to be
  specified in the data frame \code{tdat} using the \code{\link{factor}}
  command), and ordered discrete (to be specified in the data frame
  \code{tdat} using the \code{\link{ordered}} command). Data can be
  entered in an arbitrary order and data types will be detected
  automatically by the routine (see \code{\link{np}} for details).

  A variety of kernels may be specified by the user. Kernels implemented
  for continuous data types include the second, fourth, sixth, and
  eighth order Gaussian and Epanechnikov kernels, and the uniform
  kernel. Unordered discrete data types use a variation on Aitchison and
  Aitken's (1976) kernel, while ordered data types use a variation of
  the Wang and van Ryzin (1981) kernel.

}
\value{

  \code{npudens} returns a \code{npdensity} object. The generic accessor
  functions \code{\link{fitted}}, and \code{\link{se}}, extract
  estimated values and asymptotic standard errors on estimates,
  respectively, from the returned object. Furthermore, the functions
  \code{\link{predict}}, \code{\link{summary}} and \code{\link{plot}}
  support objects of both classes. The returned objects have the
  following components:

  \item{eval}{ the evaluation points. }
  \item{dens}{ estimation of the density at the evaluation points }
  \item{derr}{ standard errors of the density estimates }
  \item{log_likelihood}{ log likelihood of the density estimates }

}
\section{Book And Method Pointers}{
  The unconditional density target is \eqn{f(x)}. In the continuous
  fixed-bandwidth case the estimator has the usual kernel form
  \eqn{\hat f(x)=n^{-1}\sum_i K_h(X_i,x)}; with mixed data,
  \code{npudens} uses generalized product kernels so that the continuous,
  unordered, and ordered components each contribute their appropriate
  kernel factor. Nearest-neighbor bandwidth types change how the
  continuous bandwidth is indexed, not the statistical target.

  For book-length derivations, see Li and Racine (2007), Chapter 1
  \emph{Density Estimation}, especially Sections 1.1, 1.3, 1.6, 1.8, and
  1.9, and Chapter 4 \emph{Kernel Estimation with Mixed Data}. The later
  workflow treatment is Racine (2019), Chapter 2 \emph{Continuous
  Density and Cumulative Distribution Functions} and Chapter 3
  \emph{Mixed-Data Probability Density and Cumulative Distribution
  Functions}.
}

\references{
  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{ Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
  and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2003), \dQuote{Nonparametric estimation of
    distributions with categorical and continuous data,} Journal
  of Multivariate Analysis, 86, 266-292.

  Ouyang, D. and Q. Li and J.S. Racine (2006), \dQuote{Cross-validation
    and the estimation of probability distributions with categorical
    data,} Journal of Nonparametric Statistics, 18, 69-100.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,} Cambridge
  University Press.

  Scott, D.W. (1992), \emph{Multivariate Density Estimation: Theory,
  Practice and Visualization,} New York: Wiley.

  Silverman, B.W. (1986), \emph{Density Estimation,} London: Chapman and
  Hall.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth estimators
    for discrete distributions,}  Biometrika, 68, 301-309.
}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}, \code{\link{plot.np}} \code{\link{npudensbw}} , \code{\link{density}} }

\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): For this example, we load Giovanni
# Baiocchi's Italian GDP panel (see Italy for details), then create a
# data frame in which year is an ordered factor, GDP is continuous,
# compute bandwidths using likelihood cross-validation, then create a
# grid of data on which the density will be evaluated for plotting
# purposes.

data("Italy")
with(Italy, {

# Compute bandwidths using likelihood cross-validation (default).

bw <- npudensbw(formula=~ordered(year)+gdp)

# At this stage you could use npudens() to do a variety of
# things. Here we compute the npudens() object and place it in fhat.

fhat <- npudens(bws=bw)

# Note that simply typing the name of the object returns some useful
# information. For more info, one can call summary:

summary(fhat)

# Next, we illustrate how to create a grid of `evaluation data' and feed
# it to the perspective plotting routines in R, among others.

# Create an evaluation data matrix

year.seq <- sort(unique(year))
gdp.seq <- seq(1,36,length=50)
data.eval <- expand.grid(year=year.seq,gdp=gdp.seq)

# Generate the estimated density computed for the evaluation data

fhat <- fitted(npudens(bws=bw, newdata=data.eval))

# Coerce the data into a matrix for plotting with persp()

f <- matrix(fhat, length(unique(year)), 50)

# Next, create a 3D perspective plot of the PDF f, and a 2D
# contour plot.

persp(as.integer(levels(year.seq)), gdp.seq, f, col="lightblue",
      ticktype="detailed", ylab="GDP", xlab="Year", zlab="Density",
      theta=300, phi=50)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

contour(as.integer(levels(year.seq)),
        gdp.seq,
        f,
        xlab="Year",
        ylab="GDP",
        main = "Density Contour Plot",
        col=topo.colors(100))

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Alternatively, you could use the plot() command (<ctrl>-C will
# interrupt on *NIX systems, <esc> will interrupt on MS Windows
# systems).

if (interactive()) plot(bw)

})

# EXAMPLE 1 (INTERFACE=DATA FRAME): For this example, we load Giovanni
# Baiocchi's Italian GDP panel (see Italy for details), then create a
# data frame in which year is an ordered factor, GDP is continuous,
# compute bandwidths using likelihood cross-validation, then create a
# grid of data on which the density will be evaluated for plotting
# purposes.

data("Italy")
with(Italy, {

data <- data.frame(year=ordered(year), gdp)

# Compute bandwidths using likelihood cross-validation (default).

bw <- npudensbw(dat=data)

# At this stage you could use npudens() to do a variety of
# things. Here we compute the npudens() object and place it in fhat.

fhat <- npudens(bws=bw)

# Note that simply typing the name of the object returns some useful
# information. For more info, one can call summary:

summary(fhat)

# Next, we illustrate how to create a grid of `evaluation data' and feed
# it to the perspective plotting routines in R, among others.

# Create an evaluation data matrix

year.seq <- sort(unique(year))
gdp.seq <- seq(1,36,length=50)
data.eval <- expand.grid(year=year.seq,gdp=gdp.seq)

# Generate the estimated density computed for the evaluation data

fhat <- fitted(npudens(edat = data.eval, bws=bw))

# Coerce the data into a matrix for plotting with persp()

f <- matrix(fhat, length(unique(year)), 50)

# Next, create a 3D perspective plot of the PDF f, and a 2D
# contour plot.

persp(as.integer(levels(year.seq)), gdp.seq, f, col="lightblue",
      ticktype="detailed", ylab="GDP", xlab="Year", zlab="Density",
      theta=300, phi=50)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

contour(as.integer(levels(year.seq)),
        gdp.seq,
        f,
        xlab="Year",
        ylab="GDP",
        main = "Density Contour Plot",
        col=topo.colors(100))

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Alternatively, you could use the plot() command (<ctrl>-C will
# interrupt on *NIX systems, <esc> will interrupt on MS Windows
# systems).

if (interactive()) plot(bw)

})

# EXAMPLE 2 (INTERFACE=FORMULA): For this example, we load the old
# faithful geyser data and compute the density and distribution
# functions.

library("datasets")
data("faithful")
with(faithful, {

# Note - this may take a few minutes depending on the speed of your
# computer...

bw <- npudensbw(formula=~eruptions+waiting)

summary(bw)

# Plot the density function (<ctrl>-C will interrupt on *NIX systems,
# <esc> will interrupt on MS Windows systems). Note that we use xtrim =
# -0.2 to extend the plot outside the support of the data (i.e., extend
# the tails of the estimate to meet the horizontal axis).

if (interactive()) plot(bw, xtrim=-0.2)

})

# EXAMPLE 2 (INTERFACE=DATA FRAME): For this example, we load the old
# faithful geyser data and compute the density and distribution
# functions.

library("datasets")
data("faithful")
with(faithful, {

# Note - this may take a few minutes depending on the speed of your
# computer...

bw <- npudensbw(dat=faithful)

summary(bw)

# Plot the density function (<ctrl>-C will interrupt on *NIX systems,
# <esc> will interrupt on MS Windows systems). Note that we use xtrim =
# -0.2 to extend the plot outside the support of the data (i.e., extend
# the tails of the estimate to meet the horizontal axis).

if (interactive()) plot(bw, xtrim=-0.2)

})
} % enddontrun
}
\keyword{ nonparametric }% at least one, from doc/KEYWORDS

\name{npreghat}
\alias{npreghat}
\alias{npreghat.formula}
\alias{npreghat.rbandwidth}
\alias{npreghat.npregression}
\alias{predict.npreghat}
\alias{print.npreghat}
\title{Nonparametric Regression Hat Operator}
\description{
Constructs nonparametric regression hat operators for \code{npreg}-compatible
bandwidth objects. The returned operator \eqn{H^{(s)}} maps responses to fitted
values or derivative estimates via \eqn{H^{(s)} y}.
}
\usage{
npreghat(bws, ...)

\method{npreghat}{formula}(bws, 
         data = NULL, 
         newdata = NULL, 
         ...)

\method{npreghat}{rbandwidth}(bws,
         txdat = stop("training data 'txdat' missing"),
         exdat, y = NULL,
         output = c("matrix", "apply", "constraint"),
         basis = NULL,
         bernstein.basis = NULL,
         degree = NULL,
         deriv = NULL,
         leave.one.out = FALSE,
         ridge = 0,
         s = NULL, 
         ...)

\method{npreghat}{npregression}(bws, 
         txdat, 
         y, 
         ...)

\method{predict}{npreghat}(object, 
        newdata = NULL, y = NULL,
        output = c("matrix", "apply", "constraint"),
        s = attr(object, "s"),
        leave.one.out = attr(object, "leave.one.out"),
        deriv = NULL, 
        ...)
}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the fitted bandwidth object, formula/data interface, training data, and evaluation data.
}

  \item{bws}{An object of class \code{rbandwidth} or \code{npregression}.}


  \item{data}{A data frame used with the formula interface.}


  \item{exdat}{Optional evaluation predictors.}


  \item{newdata}{Optional evaluation data for formula and predict methods.}


  \item{txdat}{Training predictors.}


\subsection{Local-Polynomial Controls}{
These arguments control local-polynomial basis, degree, derivatives, leave-one-out behavior, and ridge stabilization.
}

  \item{basis}{Local polynomial basis: \code{"glp"}, \code{"additive"}, or \code{"tensor"}.}


  \item{bernstein.basis}{Logical; use Bernstein basis for LP terms.}


  \item{degree}{Optional local polynomial degree vector override (LP path).}


  \item{deriv}{Convenience alias for \code{s}.}


  \item{leave.one.out}{Logical; if \code{TRUE}, compute in-sample leave-one-out hat weights. This cannot be combined with explicit \code{exdat}/\code{newdata}.}


  \item{ridge}{Base diagonal regularization used when local systems are ill-conditioned. The ridge sequence starts at \code{0} (no regularization) and then increments by \code{1/n.train} as needed for stable solves.}


  \item{s}{Derivative multi-index over continuous predictors.}


\subsection{Method Objects}{
This argument identifies a fitted hat-operator object supplied to an S3 method.
}

  \item{object}{An object returned by \code{npreghat}.}


\subsection{Operator Output}{
These arguments control whether the operator is returned as a matrix, applied directly, or returned as a quadratic-programming constraint design matrix.
}

  \item{output}{Either \code{"matrix"} for the hat matrix, \code{"apply"} for direct application to \code{y}, or \code{"constraint"} for the row-weighted transpose \code{t(H) * y}.}


  \item{y}{Optional response vector or matrix for apply mode. For \code{output = "constraint"}, \code{y} is required and must be a vector or one-column object.}


\subsection{Additional Arguments}{
Further arguments are passed to methods.
}

  \item{...}{Additional arguments passed to methods.}

}
\details{
For \code{output = "matrix"}, the return value is a matrix with class
\code{c("npreghat", "matrix")} so it can be used directly in matrix products,
e.g. \code{H \%*\% y}. Attributes on the matrix store metadata used by
\code{predict.npreghat}.

For \code{output = "apply"}, the function returns \code{H^{(s)} y} directly and
accepts matrix right-hand sides for one-shot bootstrap-style calculations.

For \code{output = "constraint"}, the function returns \code{t(H^{(s)}) * y},
the row-weighted transpose commonly used as the design matrix in
shape-constrained quadratic-programming examples. This is a convenience route
that is exactly equivalent to obtaining \code{H} with \code{output = "matrix"}
and then computing \code{t(H) * y}; it does not solve a constrained estimation
problem.
}
\value{
Either a hat matrix (class \code{"npreghat"}) or the applied result
\code{H^{(s)} y}, or the constraint design matrix \code{t(H^{(s)}) * y},
depending on \code{output}.
}
\examples{
\dontrun{
data(cps71)
bw <- npregbw(xdat = cps71$age, ydat = cps71$logwage,
              regtype = "ll", bandwidth.compute = FALSE, bws = 1.0)
H <- npreghat(bws = bw, txdat = data.frame(age = cps71$age))
H.fitted <- H %*% cps71$logwage
A <- npreghat(bws = bw, txdat = data.frame(age = cps71$age),
              y = cps71$logwage, output = "constraint")
all.equal(A, t(H) * cps71$logwage)
ghat <- npreg(bws = bw)
head(cbind(fitted(ghat), H.fitted), n = 2L)
}
}
\keyword{nonparametric}
\keyword{regression}

%
\name{npdeptest}
\alias{npdeptest}
\title{ Kernel Consistent Pairwise Nonlinear Dependence Test for Univariate Processes }
\description{

  \code{npdeptest} implements the consistent metric entropy test of
  pairwise independence as described in Maasoumi and Racine (2002).

}
\usage{

npdeptest(data.x = NULL,
          data.y = NULL,
          method = c("integration","summation"),
          bootstrap = TRUE,
          boot.num = 399,
          random.seed = 42)
  
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the paired samples being tested.
}

  \item{data.x, data.y}{

    two univariate vectors containing two variables that are of type
    \code{\link{numeric}}.

  }

\item{method}{

    a character string used to specify whether to compute the integral
    version or the summation version of the statistic. Can be set as
    \code{integration} or \code{summation} (see below for
    details). Defaults to \code{integration}.

  }


\subsection{Bootstrap Controls}{
These arguments control bootstrap execution and reproducibility settings.
}

  \item{boot.num}{
    an integer value specifying the number of bootstrap
    replications to use. Defaults to \code{399}.
  }


  \item{bootstrap}{
    a logical value which specifies whether to conduct
    the bootstrap test or not. If set to \code{FALSE}, only the
    statistic will be computed. Defaults to \code{TRUE}.
  }


  \item{random.seed}{
    an integer used to seed R's random number generator. This is to
    ensure replicability. Defaults to 42.
  }

}
\value{
  \code{npdeptest} returns an object of type \code{deptest} with the
  following components

  \item{Srho}{ the statistic \code{Srho} }
  \item{Srho.bootstrap.vec}{ contains the bootstrap replications of
    \code{Srho} }
  \item{P}{ the P-value of the Srho statistic }
  \item{bootstrap}{ a logical value indicating whether bootstrapping was
  performed }
  \item{boot.num}{ number of bootstrap replications }
  \item{bw.data.x}{ the numeric bandwidth for \code{data.x} marginal
    density}
  \item{bw.data.y}{ the numeric bandwidth for
    \code{data.y} marginal density}
  \item{bw.joint}{ the numeric matrix of bandwidths for \code{data}
  and lagged \code{data} joint density  at lag \code{num.lag}}

\code{\link{summary}} supports object of type \code{deptest}.
}
\section{Book And Method Pointers}{
  \code{npdeptest} compares the joint density or probability law of two
  variables with the product of their marginals. Departures from the
  product form indicate dependence, lack of fit, or predictability
  depending on how the two inputs are constructed.

  For book-length background, see Li and Racine (2007), Chapter 13
  \emph{Nonsmoothing Tests}, especially Sections 13.4 and 13.5, and
  Chapter 12 \emph{Model Specification Tests}, especially Sections 12.4.1
  and 12.4.3. For entropy and density context, see Racine (2019),
  Chapter 2.
}

\references{

  Granger, C.W. and E. Maasoumi and J.S. Racine (2004), \dQuote{A
  dependence metric for possibly nonlinear processes}, Journal of Time
  Series Analysis, 25, 649-669.

  Maasoumi, E. and J.S. Racine (2002), \dQuote{Entropy and
    Predictability of Stock Market Returns,} Journal of Econometrics,
    107, 2, pp 291-312.

}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.


  \code{npsdeptest} computes the nonparametric metric entropy
   (normalized Hellinger of Granger, Maasoumi and Racine (2004)) for
   testing pairwise nonlinear dependence between the densities of two
   data series. See Maasoumi and Racine (2002) for details. Default
   bandwidths are of the Kullback-Leibler variety obtained via
   likelihood cross-validation. The null distribution is obtained via
   bootstrap resampling under the null of pairwise independence.

  \code{npdeptest} computes the distance between the joint distribution
  and the product of marginals (i.e. the joint distribution under the
  null), \eqn{D[f(y, \hat y), f(y)\times f(\hat y)]}{D[f(y, y-hat),
  f(y)*f(y-hat)]}. Examples include, (a) a measure/test of \dQuote{fit},
  for in-sample values of a variable \eqn{y}{y} and its fitted values,
  \eqn{\hat y}{y-hat}, and (b) a measure of \dQuote{predictability} for
  a variable \eqn{y} and its predicted values \eqn{\hat y}{y-hat} (from
  a user implemented model).

  The summation version of this statistic will be numerically unstable
  when \code{data.x} and \code{data.y} lack common support or are sparse
  (the summation version involves division of densities while the
  integration version involves differences). Warning messages are
  produced should this occur (\sQuote{integration recommended}) and should be
  heeded.

}

\section{Usage Issues}{

  The \code{integration} version of the statistic uses multidimensional
  numerical methods from the \pkg{cubature} package. See
  \pkg{adaptIntegrate} for details. The \code{integration}
  version of the statistic will be substantially slower than the
  \code{summation} version, however, it will likely be both more
  accurate and powerful.

}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}
  \code{\link{npdeneqtest},\link{npsdeptest},\link{npsymtest},\link{npunitest}}
}

\examples{
\dontrun{
set.seed(1234)

## Test/measure lack of fit between y and its fitted value from a
## regression model when x is relevant using the `summation' version.

n <- 100
x <- rnorm(n)
y <- 1 + x + rnorm(n)
model <- lm(y~x)
y.fit <- fitted(model)

npdeptest(y,y.fit,boot.num=99,method="summation")

if (interactive()) Sys.sleep(5)

## Test/measure lack of fit between y and its fitted value from a
## regression model when x is irrelevant using the `summation' version.

n <- 100
x <- runif(n,-2,2)
y <- 1 + rnorm(n)
model <- lm(y~x)
y.fit <- fitted(model)

npdeptest(y,y.fit,boot.num=99,method="summation")

## Test/measure lack of fit between y and its fitted value from a
## regression model when x is relevant using the `integration'
## version (default, slower than summation version).

n <- 100
x <- rnorm(n)
y <- 1 + x + rnorm(n)
model <- lm(y~x)
y.fit <- fitted(model)

npdeptest(y,y.fit,boot.num=99)

if (interactive()) Sys.sleep(5)

## Test/measure lack of fit between y and its fitted value from a
## regression model when x is irrelevant using the `integration'
## version (default, slower than summation version).

n <- 100
x <- runif(n,-2,2)
y <- 1 + rnorm(n)
model <- lm(y~x)
y.fit <- fitted(model)

npdeptest(y,y.fit,boot.num=99)
} % enddontrun
}
\keyword{ univar }
\keyword{ nonparametric }

%
\name{npreg}
\alias{npreg}
\alias{npreg.default}
\alias{npreg.formula}
\alias{npreg.rbandwidth}

\title{Kernel Regression with Mixed Data Types}

\description{
  \code{npreg} computes a kernel regression estimate of a one
  (1) dimensional dependent variable on \eqn{p}-variate explanatory
  data, given a set of evaluation points, training points (consisting of
  explanatory data and dependent data), and a bandwidth specification
  using the method of Racine and Li (2004) and Li and Racine (2004). A
  bandwidth specification can be a \code{rbandwidth} object, or a
  bandwidth vector, bandwidth type and kernel type.
}

\usage{
npreg(bws, \dots)

\method{npreg}{formula}(bws, 
      data = NULL, 
      newdata = NULL, 
      y.eval = FALSE, 
      \dots)

\method{npreg}{default}(bws, 
      txdat, 
      tydat, 
      nomad = FALSE, 
      \dots)

\method{npreg}{rbandwidth}(bws,
      txdat = stop("training data 'txdat' missing"),
      tydat = stop("training data 'tydat' missing"),
      exdat,
      eydat,
      gradient.order = 1L,
      gradients = FALSE,
      residuals = FALSE,
      \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the bandwidth specification, formula/data interface, and training data.
}

  \item{bws}{
    a bandwidth specification. This can be set as a \code{rbandwidth}
    object returned from an invocation of \code{\link{npregbw}}, or
    as a vector of bandwidths, with each element \eqn{i} corresponding
    to the bandwidth for column \eqn{i} in \code{txdat}. If specified as
    a vector, then additional arguments will need to be supplied as
    necessary to specify the bandwidth type, kernel types, and so on.
  }


  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(bws)}, typically the environment from which
    \code{\link{npregbw}} was called.
  }


  \item{txdat}{
    a \eqn{p}-variate data frame of explanatory data (training data) used to
    calculate the regression estimators. Defaults to the training data used to
    compute the bandwidth object.
  }


  \item{tydat}{
    a one (1) dimensional numeric or integer vector of dependent data, each
    element \eqn{i} corresponding to each observation (row) \eqn{i} of
    \code{txdat}. Defaults to the training data used to
    compute the bandwidth object.
  }


\subsection{Local-Polynomial Degree And Bandwidth Search}{
This argument controls the recommended automatic local-polynomial NOMAD route, which jointly selects continuous polynomial degree and bandwidths when these are computed inside \code{npreg}.
}

  \item{nomad}{
    logical shortcut passed through to \code{\link{npregbw}} when
    bandwidths are computed inside \code{npreg}. When \code{TRUE}, the
    regression bandwidth route fills any missing values among
    \code{regtype}, \code{search.engine}, \code{degree.select},
    \code{bernstein.basis}, \code{degree.min}, \code{degree.max},
    \code{degree.verify}, and \code{bwtype} with the recommended
    automatic local-polynomial degree-and-bandwidth NOMAD preset documented in \code{\link{npregbw}}.
    Additional NOMAD tuning arguments such as \code{nomad.nmulti} may
    also be supplied through \code{\dots}; \code{nmulti} remains the
    outer restart count while \code{nomad.nmulti} controls inner
    \code{crs::snomadr()} multistarts within each outer restart.
    After fitting, inspect \code{fit$bws$nomad.shortcut} on the
    returned object \code{fit} to see the normalized shortcut metadata.
  }


\subsection{Evaluation Data And Returned Quantities}{
These arguments control where the regression is evaluated and which fitted quantities are returned.
}

  \item{exdat}{
    a \eqn{p}-variate data frame of points on which the regression will be
    estimated (evaluation data). By default,
    evaluation takes place on the data provided by \code{txdat}.
  }


  \item{eydat}{
    a one (1) dimensional numeric or integer vector of the true values
    of the dependent variable. Optional, and used only to calculate the
    true errors.
  }


  \item{gradient.order}{
    for \code{regtype="lp"} with \code{gradients=TRUE}, a positive
    integer (or integer vector with one entry per continuous predictor)
    specifying derivative order(s). Defaults to \code{1L}. Orders
    exceeding \code{degree} for a variable are returned as
    \code{NA}. Higher-order derivatives are available for continuous
    predictors when the requested order does not exceed the
    corresponding local-polynomial degree.
  }


  \item{gradients}{
    a logical value indicating that you want gradients computed and
    returned in the resulting \code{npregression} object. Defaults to
    \code{FALSE}. For \code{regtype="lp"}, derivative components that
    are not defined (requested order exceeds the variable-specific
    polynomial degree) are returned as \code{NA}. For ordered and
    unordered predictors the gradient columns are first-order
    effects/contrasts where supported; \code{gradient.order} controls
    derivative order for continuous predictors.
  }


  \item{newdata}{
    An optional data frame in which to look for evaluation data. If
    omitted, the training data are used.
  }


  \item{residuals}{
    a logical value indicating that you want residuals computed and
    returned in the resulting \code{npregression} object. Defaults to
    \code{FALSE}.
  }


  \item{y.eval}{
    If \code{newdata} contains dependent data and \code{y.eval = TRUE},
    \code{\link{np}} will compute goodness of fit statistics on these
    data and return them. Defaults to \code{FALSE}.
  }


\subsection{Additional Arguments}{
Further arguments are passed to \code{\link{npregbw}} when bandwidths are computed internally, or used to interpret a numeric \code{bws} vector.
}

  \item{\dots}{
    additional arguments supplied to \code{\link{npregbw}} when
    \code{npreg} computes bandwidths internally, or arguments needed to
    interpret a numeric \code{bws} vector. This is where bandwidth
    selection controls such as \code{bwmethod}, \code{bwtype},
    \code{bwscaling}, kernel/support controls such as \code{ckertype},
    \code{ckerorder}, and \code{ckerbound}, categorical kernel controls
    such as \code{ukertype} and \code{okertype}, search controls such
    as \code{nmulti} and \code{scale.factor.search.lower}, and
    local-polynomial/NOMAD controls such as \code{regtype},
    \code{degree}, \code{basis}, \code{bernstein.basis},
    \code{degree.select}, and \code{nomad.nmulti} are supplied. See
    \code{\link{npregbw}} for the complete bandwidth-selection argument
    surface.
  }
}
\details{
  Documentation guide: see \code{\link{npregbw}} for bandwidth
  selection and search controls, \code{\link{np.kernels}} for kernels,
  \code{\link{np.options}} for global options, and \code{\link{plot}},
  \code{\link{plot.np}} for plotting options.

  When \code{bws} is omitted, the formula and default methods call
  \code{\link{npregbw}} first and pass bandwidth-selection arguments
  from \code{\dots} to that call. When \code{bws} is already an
  \code{rbandwidth} object, \code{npreg} estimates with the stored
  bandwidth metadata in that object.

  Argument groups for bandwidth selection are documented on
  \code{\link{npregbw}}. The most common workflow is to choose data and
  bandwidth inputs first, then bandwidth criterion and representation,
  then kernel/support controls, and finally local-polynomial/NOMAD
  controls when using polynomial-adaptive fits.

  For S3 plotting help, see \code{\link{plot.np}}. You can list
  available plot methods with \code{methods("plot")}.


  Typical usages are (see below for a complete list of options and also
  the examples at the end of this help file)
  \preformatted{
    Usage 1: first compute the bandwidth object via npregbw and then
    compute the conditional mean:

    bw <- npregbw(y~x)
    ghat <- npreg(bw)

    Usage 2: alternatively, compute the bandwidth object indirectly:

    ghat <- npreg(y~x)

    Usage 3: modify the default kernel and order:

    ghat <- npreg(y~x, ckertype="epanechnikov", ckerorder=4)

    Usage 4: use the data frame interface rather than the formula
    interface:

    ghat <- npreg(tydat=y, txdat=x, ckertype="epanechnikov", ckerorder=4)
  }



  \code{npreg} implements a variety of methods for regression on
  multivariate (\eqn{p}-variate) data, the types of which are possibly
  continuous and/or discrete (unordered, ordered). The approach is
  based on Li and Racine (2003) who employ \sQuote{generalized product kernels}
  that admit a mix of continuous and discrete data types.

  Three classes of kernel estimators for the continuous data types are
  available: fixed, adaptive nearest-neighbor, and generalized
  nearest-neighbor. Adaptive nearest-neighbor bandwidths change with
  each sample realization in the set, \eqn{x_i}{x[i]}, when estimating the
  density at the point \eqn{x}. Generalized nearest-neighbor bandwidths change
  with the point at which the density is estimated, \eqn{x}. Fixed bandwidths
  are constant over the support of \eqn{x}.

  Data contained in the data frame \code{txdat} may be a mix of
  continuous (default), unordered discrete (to be specified in the data
  frame \code{txdat} using \code{\link{factor}}), and ordered discrete
  (to be specified in the data frame \code{txdat} using
  \code{\link{ordered}}). Data can be entered in an arbitrary order and
  data types will be detected automatically by the routine (see
  \code{\link{np}} for details).

  A variety of kernels may be specified by the user. Kernels implemented
  for continuous data types include the second, fourth, sixth, and eighth
  order Gaussian and Epanechnikov kernels, and the uniform
  kernel. Unordered discrete data types use a variation on Aitchison and
  Aitken's (1976) kernel, while ordered data types use a variation of the
  Wang and van Ryzin (1981) kernel.

  When bandwidths are obtained with \code{regtype="lp"}, C-level
  \code{npreg} supports heterogeneous continuous polynomial degrees via
  \code{degree}. The \code{basis} selector currently supports
  \code{basis="glp"}, \code{"additive"}, and \code{"tensor"}.
  For continuous predictors with degree vector
  \eqn{d}{d}, additive basis size is
  \eqn{1+\sum_j d_j}{1+sum(d[j])}, tensor basis size is
  \eqn{\prod_j (d_j+1)}{prod(d[j]+1)}, and GLP uses admissible
  multi-indices \eqn{\alpha}{alpha} with
  \eqn{\alpha_j \le d_j}{alpha[j] <= d[j]} and
  \eqn{0<\sum_j \alpha_j \le \max_j d_j}{0<sum(alpha)<=max(d)} plus
  an intercept. The optional flag \code{bernstein.basis}
  controls basis construction: \code{FALSE} (default) uses raw
  local-polynomial powers, while \code{TRUE} uses a Bernstein/B-spline basis. The
  homogeneous degree-0 and degree-1 cases remain
  equivalent to \code{lc} and \code{ll}, respectively. For continuous
  predictors, \code{gradients=TRUE} returns the derivative order(s)
  requested by \code{gradient.order}; the associated \code{gerr}
  entries are standard errors for those same derivative orders. Both
  raw and Bernstein/B-spline bases report fitted values and derivatives
  on the original response/predictor scale. In mixed-data GLP settings,
  unordered and ordered predictors contribute first-order
  effects/contrasts rather than higher-order derivatives.
  When \code{npregbw(..., regtype="lp")} is used with
  \code{degree.select="manual"}, the degree vector remains fixed user
  input. When \code{degree.select != "manual"}, \code{npregbw} can
  jointly select polynomial degree and bandwidth using either the
  cached cell-search backend or the direct
  \code{search.engine="nomad"}/\code{"nomad+powell"} route described in
  \code{\link{npregbw}}; the latter follows Hall and Racine (2015). For
  practitioners who want that recommended route without spelling out all
  LP tuning arguments, \code{npreg(..., nomad=TRUE)} and
  \code{npregbw(..., nomad=TRUE)} expand missing settings to the same
  documented automatic-LP NOMAD preset. Explicit incompatible settings
  fail fast rather than being silently rewritten. The direct NOMAD
  backend is provided by the suggested package \code{crs}, so install
  \code{crs} before using \code{search.engine="nomad"},
  \code{"nomad+powell"}, or \code{nomad=TRUE}. For
  \code{bernstein.basis=TRUE}, evaluation points for continuous predictors
  must lie within training support; use \code{bernstein.basis=FALSE} for
  extrapolation. For \code{regtype="ll"} and \code{regtype="lp"}, the
  training continuous design is checked for rank deficiency and extreme
  condition number before estimation proceeds.

  The use of compactly supported kernels or the occurrence of small
  bandwidths can lead to numerical problems for the local linear
  estimator when computing the locally weighted least squares
  solution. To overcome this problem we rely on a form or
  \sQuote{ridging} proposed by Cheng, Hall, and Titterington (1997),
  modified so that we solve the problem pointwise rather than globally
  (i.e. only when it is needed).

}

\value{

  \code{npreg} returns a \code{npregression} object.
  The generic
  functions \code{\link{fitted}}, \code{\link{residuals}},
  \code{\link{se}}, \code{\link{predict}}, and
  \code{\link{gradients}}, extract (or generate) estimated values,
  residuals, asymptotic standard
  errors on estimates, predictions, and gradients, respectively, from
  the returned object. Furthermore, the functions \code{\link{summary}}
  and \code{\link{plot}} support objects of this type. The returned object
  has the following components:


  \item{eval}{ evaluation points }
  \item{mean}{ estimates of the regression function (conditional mean) at the
  evaluation points }
  \item{merr}{ standard errors of the regression function estimates }
  \item{grad}{ estimates of the gradients at each evaluation point }
  \item{gerr}{ standard errors of the gradient estimates }
  \item{resid}{ if \code{residuals = TRUE}, in-sample or out-of-sample
    residuals where appropriate (or possible) }
  \item{R2}{ coefficient of determination (Doksum and Samarov (1995))}
  \item{MSE}{ mean squared error }
  \item{MAE}{ mean absolute error }
  \item{MAPE}{ mean absolute percentage error }
  \item{CORR}{ absolute value of Pearson's correlation coefficient }
  \item{SIGN}{ fraction of observations where fitted and observed values
  agree in sign }

}

\section{Book And Method Pointers}{
  The regression target is the conditional mean
  \eqn{m(x)=E[Y\mid X=x]}. Local-constant estimation can be written
  schematically as \eqn{\hat m(x)=\sum_i W_i(x)Y_i}, where the weights
  are normalized generalized product-kernel weights over the continuous,
  unordered, and ordered components of \eqn{X}. Local-polynomial
  regression replaces this scalar weighted average by a weighted least
  squares fit in powers of the continuous coordinates around \eqn{x};
  the fitted intercept is the estimate of \eqn{m(x)} and the fitted
  slope terms give the reported gradients where available.

  For book-length derivations, see Li and Racine (2007), Chapter 2
  \emph{Regression}, especially Sections 2.1, 2.2, 2.4, and 2.5, and
  Chapter 4 \emph{Kernel Estimation with Mixed Data}, especially Sections
  4.2 and 4.4. The later workflow treatment is Racine (2019), Chapter 6
  \emph{Conditional Mean Function Estimation}, including the local
  polynomial and mixed-data marginal-effect discussion.
}

\references{

  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
  discrimination by the kernel method,} Biometrika, 63, 413-420.

  Cheng, M.-Y. and P. Hall and D.M. Titterington (1997), \dQuote{On the
  shrinkage of local linear curve estimators,} Statistics and Computing,
  7, 11-17.

  Fan, J. and I. Gijbels (1996), \emph{Local Polynomial Modelling and
  Its Applications,} Chapman and Hall.

  Doksum, K. and A. Samarov (1995), \dQuote{Nonparametric estimation of
  global functionals and a measure of the explanatory power of
  covariates in regression,} The Annals of Statistics, 23 1443-1473.

  Hall, P. and Q. Li and J.S. Racine (2007), \dQuote{Nonparametric
  estimation of regression functions in the presence of irrelevant
  regressors,} The Review of Economics and Statistics, 89, 784-789.

  Hall, P. and J.S. Racine (2015), \dQuote{Infinite Order Cross-Validated
  Local Polynomial Regression,} Journal of Econometrics, 185, 510-525.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
  and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2004), \dQuote{Cross-validated local linear
  nonparametric regression,} Statistica Sinica, 14, 485-512.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,}
  Cambridge University Press.

  Racine, J.S. and Q. Li (2004), \dQuote{Nonparametric estimation of
    regression functions with both categorical and continuous data,}
  Journal of Econometrics, 119, 99-130.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth
    estimators for discrete distributions,} Biometrika, 68, 301-309.

}

\author{
  Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}, \code{\link{plot.np}}
  \code{\link{loess}}
}
\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): For this example, we compute a
# bivariate nonparametric regression estimate for Giovanni Baiocchi's
# Italian income panel (see Italy for details)

data("Italy")
with(Italy, {

# First, compute the least-squares cross-validated bandwidths for the
# local constant estimator (default).

bw <- npregbw(formula=gdp~ordered(year))

# Now take these bandwidths and fit the model and gradients

model <- npreg(bws = bw, gradients = TRUE)

summary(model)

# Use plot() to visualize the regression function, add bootstrap
# error bars, and overlay the data on the same plot.

# Note - this may take a minute or two depending on the speed of your
# computer due to bootstrapping being conducted (<ctrl>-C will
# interrupt). Note - nothing will appear in the graphics window until
# all computations are completed (if you use
# errors="asymptotic" the figure will instantly appear).

if (interactive()) plot(bw, errors="bootstrap")
points(ordered(year), gdp, cex=.2, col="red")

})

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# EXAMPLE 1 (INTERFACE=DATA FRAME): For this example, we compute a
# bivariate nonparametric regression estimate for Giovanni Baiocchi's
# Italian income panel (see Italy for details)

data("Italy")
with(Italy, {

# First, compute the least-squares cross-validated bandwidths for the
# local constant estimator (default).

bw <- npregbw(xdat=ordered(year), ydat=gdp)

# Now take these bandwidths and fit the model and gradients

model <- npreg(bws = bw, gradients = TRUE)

summary(model)

# Use plot() to visualize the regression function, add bootstrap
# error bars, and overlay the data on the same plot.

# Note - this may take a minute or two depending on the speed of your
# computer due to bootstrapping being conducted (<ctrl>-C will
# interrupt). Note - nothing will appear in the graphics window until
# all computations are completed (if you use
# errors="asymptotic" the figure will instantly appear).

if (interactive()) plot(bw, errors="bootstrap")
points(ordered(year), gdp, cex=.2, col="red")

})

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# EXAMPLE 2 (INTERFACE=FORMULA): For this example, we compute a local
# linear fit using the AIC_c bandwidth selection criterion. We then plot
# the estimator and its gradient using asymptotic standard errors.

data("cps71", package = "np")

bw <- npregbw(logwage~age, regtype="ll", bwmethod="cv.aic", data = cps71)

# Next, plot the regression function...

if (interactive()) plot(bw, errors="asymptotic")
with(cps71, points(age, logwage, cex=.2, col="red"))

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Next, plot the derivative...

if (interactive()) plot(bw, errors="asymptotic", gradient=TRUE)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# EXAMPLE 2 (INTERFACE=DATA FRAME): For this example, we compute a local
# linear fit using the AIC_c bandwidth selection criterion. We then plot
# the estimator and its gradient using asymptotic standard errors.

data("cps71", package = "np")

bw <- npregbw(xdat=cps71$age, ydat=cps71$logwage, regtype="ll", bwmethod="cv.aic")

# Next, plot the regression function...

if (interactive()) plot(bw, errors="asymptotic")
with(cps71, points(age, logwage, cex=.2, col="red"))

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Next, plot the derivative...

if (interactive()) plot(bw, errors="asymptotic", gradient=TRUE)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# EXAMPLE 3 (INTERFACE=FORMULA): For this example, we replicate the
# nonparametric regression in Maasoumi, Racine, and Stengos
# (2007) (see oecdpanel for details). Note that X is multivariate
# containing a mix of unordered, ordered, and continuous data types. Note
# - this may take a few minutes depending on the speed of your computer.

data("oecdpanel")
with(oecdpanel, {

bw <- npregbw(formula=growth~
              factor(oecd)+
              factor(year)+
              initgdp+
              popgro+
              inv+
              humancap)

if (interactive()) plot(bw, errors="asymptotic")

})

# EXAMPLE 3 (INTERFACE=DATA FRAME): For this example, we replicate the
# nonparametric regression in Maasoumi, Racine, and Stengos
# (2007) (see oecdpanel for details). Note that X is multivariate
# containing a mix of unordered, ordered, and continuous data types. Note
# - this may take a few minutes depending on the speed of your computer.

data("oecdpanel")
with(oecdpanel, {

y <- growth
X <- data.frame(factor(oecd), factor(year), initgdp, popgro, inv, humancap)

bw <- npregbw(xdat=X, ydat=y)

if (interactive()) plot(bw, errors="asymptotic")

})


# EXAMPLE 4 (INTERFACE=FORMULA): Experimental data - the effect of
# vitamin C on tooth growth in guinea pigs
#
# Description:
#
#     The response is the length of odontoblasts (teeth) in each of 10
#     guinea pigs at each of three dose levels of Vitamin C (0.5, 1, and
#     2 mg) with each of two delivery methods (orange juice or ascorbic
#     acid).
#
# Usage:
#
#     ToothGrowth
#
# Format:
#
#     A data frame with 60 observations on 3 variables.
#
#       [,1]  len   numeric  Tooth length
#       [,2]  supp  factor   Supplement type (VC or OJ).
#       [,3]  dose  numeric  Dose in milligrams.

library("datasets")
with(ToothGrowth, {

# Note - in this example, there are six cells.

bw <- npregbw(formula=len~factor(supp)+ordered(dose))

# Now plot the partial regression surfaces with bootstrapped
# nonparametric confidence bounds

if (interactive()) plot(bw, errors="bootstrap", band="simultaneous")

})

# EXAMPLE 4 (INTERFACE=DATA FRAME): Experimental data - the effect of
# vitamin C on tooth growth in guinea pigs
#
# Description:
#
#     The response is the length of odontoblasts (teeth) in each of 10
#     guinea pigs at each of three dose levels of Vitamin C (0.5, 1, and
#     2 mg) with each of two delivery methods (orange juice or ascorbic
#     acid).
#
# Usage:
#
#     ToothGrowth
#
# Format:
#
#     A data frame with 60 observations on 3 variables.
#
#       [,1]  len   numeric  Tooth length
#       [,2]  supp  factor   Supplement type (VC or OJ).
#       [,3]  dose  numeric  Dose in milligrams.

library("datasets")
with(ToothGrowth, {

# Note - in this example, there are six cells.

y <- len
X <- data.frame(supp=factor(supp), dose=ordered(dose))

bw <- npregbw(X, y)

# Now plot the partial regression surfaces with bootstrapped
# nonparametric confidence bounds

if (interactive()) plot(bw, errors="bootstrap", band="simultaneous")

})

# EXAMPLE 5 (INTERFACE=FORMULA): a pretty 2-d smoothing example adapted
# from the R mgcv library which was written by Simon N. Wood.

set.seed(12345)

# This function generates a smooth nonlinear DGP

dgp.func <- function(x, z, sx=0.3, sz=0.4)
  { (pi**sx*sz)*(1.2*exp(-(x-0.2)^2/sx^2-(z-0.3)^2/sz^2)+
                 0.8*exp(-(x-0.7)^2/sx^2-(z-0.8)^2/sz^2))
  }

# Generate 500 observations, compute the true DGP (i.e., no noise),
# then a noisy sample

n <- 500

x <- runif(n)
z <- runif(n)

xs <- seq(0, 1, length=30)
zs <- seq(0, 1, length=30)

X.eval <- data.frame(x=rep(xs, 30), z=rep(zs, rep(30, 30)))

dgp <- matrix(dgp.func(X.eval$x, X.eval$z), 30, 30)

y <- dgp.func(x, z)+rnorm(n)*0.1

# Prepare the screen for output... first, plot the true DGP

split.screen(c(2, 1))

screen(1)

persp(xs, zs, dgp, xlab="x1", ylab="x2", zlab="y", main="True DGP")

# Next, compute a local linear fit and plot that

bw <- npregbw(formula=y~x+z, regtype="ll", bwmethod="cv.aic")
fit <- fitted(npreg(bws=bw, newdata=X.eval))
fit.mat <- matrix(fit, 30, 30)

screen(2)

persp(xs, zs, fit.mat, xlab="x1", ylab="x2", zlab="y",
      main="Local linear estimate")

# EXAMPLE 5 (INTERFACE=DATA FRAME): a pretty 2-d smoothing example
# adapted from the R mgcv library which was written by Simon N. Wood.

set.seed(12345)

# This function generates a smooth nonlinear DGP

dgp.func <- function(x, z, sx=0.3, sz=0.4)
  { (pi**sx*sz)*(1.2*exp(-(x-0.2)^2/sx^2-(z-0.3)^2/sz^2)+
                 0.8*exp(-(x-0.7)^2/sx^2-(z-0.8)^2/sz^2))
  }

# Generate 500 observations, compute the true DGP (i.e., no noise),
# then a noisy sample

n <- 500

x <- runif(n)
z <- runif(n)

xs <- seq(0, 1, length=30)
zs <- seq(0, 1, length=30)

X <- data.frame(x, z)
X.eval <- data.frame(x=rep(xs, 30), z=rep(zs, rep(30, 30)))

dgp <- matrix(dgp.func(X.eval$x, X.eval$z), 30, 30)

y <- dgp.func(x, z)+rnorm(n)*0.1

# Prepare the screen for output... first, plot the true DGP

split.screen(c(2, 1))

screen(1)

persp(xs, zs, dgp, xlab="x1", ylab="x2", zlab="y", main="True DGP")

# Next, compute a local linear fit and plot that

bw <- npregbw(xdat=X, ydat=y, regtype="ll", bwmethod="cv.aic")
fit <- fitted(npreg(exdat=X.eval, bws=bw))
fit.mat <- matrix(fit, 30, 30)

screen(2)

persp(xs, zs, fit.mat, xlab="x1", ylab="x2", zlab="y",
      main="Local linear estimate")
} % enddontrun
}
\keyword{nonparametric}

% 
\name{npregbw}
\alias{npregbw}
\alias{npregbw.default}
\alias{npregbw.formula}
\alias{npregbw.rbandwidth}

\title{Kernel Regression Bandwidth Selection with Mixed Data Types}

\description{
  \code{npregbw} computes a bandwidth object for a
  \eqn{p}-variate kernel regression estimator defined over mixed
  continuous and discrete (unordered, ordered) data using expected
  Kullback-Leibler cross-validation, or least-squares cross validation
  using the method of Racine and Li (2004) and Li and Racine (2004).
}

\usage{
npregbw(\dots)

\method{npregbw}{formula}(formula, 
        data, 
        subset, 
        na.action, 
        call, 
        \dots)

\method{npregbw}{default}(xdat = stop("invoked without data 'xdat'"),
        ydat = stop("invoked without data 'ydat'"),
        bws,
        bandwidth.compute = TRUE,
        basis,
        bernstein.basis,
        bwmethod,
        bwscaling,
        bwtype,
        cfac.dir,
        scale.factor.init,
        ckerbound,
        ckerlb,
        ckerorder,
        ckertype,
        ckerub,
        degree,
        degree.select = c("manual", "coordinate", "exhaustive"),
        search.engine = c("nomad+powell", "cell", "nomad"),
        nomad = FALSE,
        nomad.nmulti = 0L,
        degree.min = NULL,
        degree.max = NULL,
        degree.start = NULL,
        degree.restarts = 0L,
        degree.max.cycles = 20L,
        degree.verify = FALSE,
        dfac.dir,
        dfac.init,
        dfc.dir,
        ftol,
        scale.factor.init.upper,
        hbd.dir,
        hbd.init,
        initc.dir,
        initd.dir,
        invalid.penalty = c("baseline","dbmax"),
        itmax,
        lbc.dir,
        scale.factor.init.lower,
        lbd.dir,
        lbd.init,
        nmulti,
        okertype,
        penalty.multiplier = 10,
        regtype,
        nomad.remin = FALSE,
        powell.remin,
        scale.init.categorical.sample,
        scale.factor.search.lower = NULL,
        small,
        tol,
        transform.bounds = FALSE,
        ukertype,
        \dots)

\method{npregbw}{rbandwidth}(xdat = stop("invoked without data 'xdat'"),
        ydat = stop("invoked without data 'ydat'"),
        bws,
        bandwidth.compute = TRUE,
        cfac.dir = 2.5*(3.0-sqrt(5)),
        scale.factor.init = 0.5,
        dfac.dir = 0.25*(3.0-sqrt(5)),
        dfac.init = 0.375,
        dfc.dir = 3,
        ftol = 1.490116e-07,
        scale.factor.init.upper = 2.0,
        hbd.dir = 1,
        hbd.init = 0.9,
        initc.dir = 1.0,
        initd.dir = 1.0,
        invalid.penalty = c("baseline","dbmax"),
        itmax = 10000,
        lbc.dir = 0.5,
        scale.factor.init.lower = 0.1,
        lbd.dir = 0.1,
        lbd.init = 0.1,
        nmulti,
        penalty.multiplier = 10,
        powell.remin = TRUE,
        scale.init.categorical.sample = FALSE,
        scale.factor.search.lower = NULL,
        small = 1.490116e-05,
        tol = 1.490116e-04,
        transform.bounds = FALSE,
        \dots)
}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the data, formula interface, and whether bandwidths are supplied or computed.
}

  \item{bandwidth.compute}{
    a logical value which specifies whether to do a numerical search for
    bandwidths or not. If set to \code{FALSE}, a \code{rbandwidth} object
    will be returned with bandwidths set to those specified
    in \code{bws}. Defaults to \code{TRUE}.
  }

  \item{bws}{
    a bandwidth specification. This can be set as a \code{rbandwidth}
    object returned from a previous invocation, or as a vector of
    bandwidths, with each element \eqn{i} corresponding to the bandwidth
    for column \eqn{i} in \code{xdat}. In either case, the bandwidth
    supplied will serve as a starting point in the numerical search for
    optimal bandwidths. If specified as a vector, then additional
    arguments will need to be supplied as necessary to specify the
    bandwidth type, kernel types, selection methods, and so on. This can
    be left unset.
  }

  \item{call}{
    the original function call. This is passed internally by
    \code{\link{np}} when a bandwidth search has been implied by a call to
    another function. It is not recommended that the user set this.
  }

  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(formula)}, typically the environment from which the
    function is called.
  }

  \item{formula}{
    a symbolic description of variables on which bandwidth selection is
    to be performed. The details of constructing a formula are
    described below.
  }

  \item{na.action}{
    a function which indicates what should happen when the data contain
   \code{NA}s. The default is set by the \code{\link{na.action}} setting of options, and is
    \code{\link{na.fail}} if that is unset. The (recommended) default is
    \code{\link{na.omit}}.  
  }

  \item{subset}{
    an optional vector specifying a subset of observations to be used in
    the fitting process. 
  }

  \item{xdat}{
    a \eqn{p}-variate data frame of regressors on which bandwidth selection will
    be performed. The data types may be continuous, discrete (unordered
    and ordered factors), or some combination thereof.
  }

  \item{ydat}{
    a one (1) dimensional numeric or integer vector of dependent data, each
    element \eqn{i} corresponding to each observation (row) \eqn{i} of
    \code{xdat}. 
  }

\subsection{Automatic Degree Search Controls}{
These arguments control automatic local-polynomial degree search when \code{regtype="lp"}.
}

  \item{degree.max}{
    optional scalar or integer vector giving upper bounds for automatic
    degree search when \code{degree.select != "manual"}. If scalar, the
    value is recycled over continuous predictors.
  }

  \item{degree.max.cycles}{
    positive integer giving the maximum number of coordinate-search
    sweeps over the continuous-predictor degree vector. Ignored for
    \code{degree.select="manual"} and \code{"exhaustive"}.
  }

  \item{degree.min}{
    optional scalar or integer vector giving lower bounds for automatic
    degree search when \code{degree.select != "manual"}. If scalar, the
    value is recycled over continuous predictors.
  }

  \item{degree.restarts}{
    non-negative integer giving the number of additional deterministic
    restarts used by coordinate search. Ignored for
    \code{degree.select="manual"} and \code{"exhaustive"}.
  }

  \item{degree.select}{
    character string controlling local-polynomial degree handling when
    \code{regtype="lp"}. \code{"manual"} (default) treats
    \code{degree} as fixed. \code{"coordinate"} performs cached
    coordinate-wise search over admissible degree vectors.
    \code{"exhaustive"} evaluates the full admissible degree grid when
    \code{search.engine="cell"}. For NOMAD-based search engines, any
    non-\code{"manual"} value requests direct joint search over degree
    and bandwidth coordinates.
  }

  \item{degree.start}{
    optional starting degree vector for automatic degree search when
    \code{degree.select="coordinate"}. If omitted, cell-based search starts
    from the degree-zero local-constant baseline on the continuous
    predictors, while NOMAD-based search starts from a clipped
    degree-one vector on the searchable continuous predictors. For
    NOMAD multistarts, later restart starts are generated
    reproducibly from a conservative proposal box and screened using
    \code{dim_basis()} so that the initial basis dimension remains well
    below the training-sample limit. This avoids wasting starts on flat
    penalty or heavily ridged designs while leaving the full user
    requested degree search region unchanged.
  }

  \item{degree.verify}{
    logical value indicating whether a coordinate-search solution should
    be exhaustively verified over the admissible degree grid after the
    heuristic phase completes. Available only for
    \code{search.engine="cell"}.
  }

\subsection{Bandwidth Criterion And Representation}{
These arguments choose the selection criterion and the way continuous bandwidths are represented.
}

  \item{bwmethod}{
    which method to use to select bandwidths. \code{cv.aic} specifies
    expected Kullback-Leibler cross-validation (Hurvich, Simonoff, and
    Tsai (1998)), and \code{cv.ls} specifies least-squares
    cross-validation. Defaults to \code{cv.ls}.
  }

  \item{bwscaling}{ a logical value that when set to \code{TRUE} the
    supplied bandwidths are interpreted as \sQuote{scale factors}
    (\eqn{c_j}{c[j]}), otherwise when the value is \code{FALSE} they are
    interpreted as \sQuote{raw bandwidths} (\eqn{h_j}{h[j]} for
    continuous data types, \eqn{\lambda_j}{lambda[j]} for discrete data
    types). For continuous data types, \eqn{c_j}{c[j]} and
    \eqn{h_j}{h[j]} are related by the formula \eqn{h_j = c_j \sigma_j
    n^{-1/(2P+l)}}{h[j] = c[j]*sigma[j]*n^(-1/(2*P+l))}, where
    \eqn{\sigma_j}{sigma[j]} is an adaptive measure of spread of
    continuous variable \eqn{j} defined as min(standard deviation, mean
    absolute deviation/1.4826, interquartile range/1.349), \eqn{n} the
    number of observations, \eqn{P}{P} the order of the kernel, and
    \eqn{l}{l} the number of continuous variables. For discrete data
    types, \eqn{c_j}{c[j]} and \eqn{h_j}{h[j]} are related by the
    formula \eqn{h_j = c_jn^{-2/(2P+l)}}{h[j] = c[j]*n^(-2/(2*P+l))},
    where here \eqn{j}{[j]} denotes discrete variable \eqn{j}{j}.
    Defaults to \code{FALSE}.  }

  \item{bwtype}{
    character string used for the continuous variable bandwidth type,
    specifying the type of bandwidth to compute and return in the
    \code{bandwidth} object. Defaults to \code{fixed}. Option
    summary:\cr
    \code{fixed}: compute fixed bandwidths \cr
    \code{generalized_nn}: compute generalized nearest neighbors \cr
    \code{adaptive_nn}: compute adaptive nearest neighbors

  }

\subsection{Categorical Search Initialization}{
These controls set categorical search starts and categorical direction-set initialization.
}

  \item{dfac.dir}{stretch factor for direction set search for Powell's algorithm for categorical variables. See Details}

  \item{dfac.init}{non-random initial values for scale factors for categorical variables for Powell's algorithm. See Details}

  \item{hbd.dir}{upper bound for direction set search for Powell's algorithm for categorical variables. See Details}

  \item{hbd.init}{upper bound for scale factors for categorical variables for Powell's algorithm. See Details}

  \item{initd.dir}{initial non-random values for direction set search for Powell's algorithm for categorical variables. See Details}

  \item{lbd.dir}{lower bound for direction set search for Powell's algorithm for categorical variables. See Details}

  \item{lbd.init}{lower bound for scale factors for categorical variables for Powell's algorithm. See Details}

  \item{scale.init.categorical.sample}{ a logical value that when set
    to \code{TRUE} scales \code{lbd.dir}, \code{hbd.dir},
    \code{dfac.dir}, and \code{initd.dir} by \eqn{n^{-2/(2P+l)}},
    \eqn{n} the number of observations, \eqn{P}{P} the order of the
    kernel, and \eqn{l}{l} the number of \code{numeric} variables. See
    Details}

\subsection{Continuous Direction-Set Search Controls}{
These controls set Powell direction-set initialization for continuous variables.
}

  \item{cfac.dir}{stretch factor for direction set search for Powell's algorithm for \code{numeric} variables. See Details}

  \item{dfc.dir}{chi-square degrees of freedom for direction set search for Powell's algorithm for \code{numeric} variables. See Details}

  \item{initc.dir}{initial non-random values for direction set search for Powell's algorithm for \code{numeric} variables. See Details}

  \item{lbc.dir}{lower bound for direction set search for Powell's algorithm for \code{numeric} variables. See Details}

\subsection{Continuous Kernel Support Controls}{
These controls choose and parameterize bounded support for continuous kernels.
}

  \item{ckerbound}{
    character string controlling continuous-kernel support handling.
    Can be set as \code{none} (default kernel on full support),
    \code{range} (use sample min/max), or \code{fixed} (use
    \code{ckerlb}/\code{ckerub}). The bounded-kernel route reuses the selected continuous kernel and renormalizes it on the chosen support; see \code{\link{np.kernels}}.
  }

  \item{ckerlb}{
    numeric scalar/vector of lower bounds for continuous variables used
    when \code{ckerbound="fixed"}. Must satisfy lower-bound validity for
    each continuous variable (e.g., \code{<= min(variable)}). Use \code{-Inf} for unbounded below. See \code{\link{np.kernels}} for bounded-kernel normalization details.
  }

  \item{ckerub}{
    numeric scalar/vector of upper bounds for continuous variables used
    when \code{ckerbound="fixed"}. Must satisfy upper-bound validity for
    each continuous variable (e.g., \code{>= max(variable)}). Use \code{Inf} for unbounded above. See \code{\link{np.kernels}} for bounded-kernel normalization details.
  }

\subsection{Continuous Scale-Factor Search Initialization}{
These controls define deterministic and random continuous scale-factor starts and the lower admissibility floor for fixed-bandwidth search.
}

  \item{scale.factor.init}{
    deterministic initial scale factor for continuous fixed-bandwidth
    search. Defaults to \code{0.5}. The value supplied by the user is
    not rewritten, but the effective first start passed to the optimizer
    is \code{max(scale.factor.init, scale.factor.search.lower)}. See
    Details.
  }

  \item{scale.factor.init.lower}{
    lower endpoint for random continuous scale-factor starts. Defaults
    to \code{0.1}. The value supplied by the user is not rewritten, but
    the effective random-start lower endpoint is
    \code{max(scale.factor.init.lower, scale.factor.search.lower)}. See
    Details.
  }

  \item{scale.factor.init.upper}{
    upper endpoint for random continuous scale-factor starts. Defaults
    to \code{2.0}. It must be greater than or equal to the effective
    lower endpoint,
    \code{max(scale.factor.init.lower, scale.factor.search.lower)};
    otherwise bandwidth search errors rather than silently expanding the
    interval. See Details.
  }

  \item{scale.factor.search.lower}{
    optional nonnegative scalar giving the hard lower admissibility
    bound for continuous fixed-bandwidth search candidates. Defaults to
    \code{NULL}. If \code{NULL}, an existing bandwidth object's stored
    value is inherited when available; otherwise the package default
    \code{0.1} is used. This floor applies to computed/search bandwidth
    candidates and to effective search starts only. It does not rewrite
    explicit bandwidths supplied for storage with
    \code{bandwidth.compute = FALSE}. Final fixed-bandwidth search
    candidates must also have a finite valid raw objective value.
  }

\subsection{Kernel Type Controls}{
These controls choose continuous, unordered, and ordered kernels.
}

  \item{ckerorder}{
    numeric value specifying kernel order (one of
    \code{(2,4,6,8)}). Kernel order specified along with a
  \code{uniform} continuous kernel type will be ignored. Defaults to
  \code{2}.
  }

  \item{ckertype}{
    character string used to specify the continuous kernel type.
    Can be set as \code{gaussian}, \code{epanechnikov}, or
    \code{uniform}. Defaults to \code{gaussian}.
  }

  \item{okertype}{
    character string used to specify the ordered categorical kernel type.
    Can be set as \code{wangvanryzin}, \code{liracine}, or \code{racineliyan}. Defaults to
    \code{liracine}.
  }

  \item{ukertype}{
    character string used to specify the unordered categorical kernel type.
    Can be set as \code{aitchisonaitken} or \code{liracine}. Defaults to
    \code{aitchisonaitken}.
  }

\subsection{Local-Polynomial Model Specification}{
These arguments control the local-polynomial estimator, basis, and fixed degree specification.
}

  \item{basis}{
    basis selector relevant only when \code{regtype="lp"}.
    Supported values are \code{"glp"}, \code{"additive"}, and
    \code{"tensor"}.
    Let \eqn{d_j}{d[j]} denote the degree for continuous predictor
    \eqn{j}{j} and \eqn{q}{q} the number of continuous predictors.
    With one segment per predictor (no internal knots), basis
    dimensions are:
    \eqn{1+\sum_{j=1}^q d_j}{1+sum(d[j])} for additive,
    \eqn{\prod_{j=1}^q (d_j+1)}{prod(d[j]+1)} for tensor, and
    \eqn{1 + |\{\alpha:\alpha_j \le d_j,\ 0<\sum_j \alpha_j \le \max_j d_j\}|}{1 + number of admissible multi-indices}
    for generalized local-polynomial (GLP) basis construction.
  }

  \item{bernstein.basis}{
    logical flag relevant only when \code{regtype="lp"}. If
    \code{FALSE} (default), the GLP basis uses raw local-polynomial
    powers (stable for extrapolation). If \code{TRUE}, a Bernstein
    (B-spline) basis is used for continuous predictors. For
    \code{bernstein.basis=TRUE}, prediction/evaluation points must lie
    within the training support of each continuous predictor. For
    automatic degree search, if \code{bernstein.basis} is not
    explicitly supplied, the search route defaults to \code{TRUE} for
    numerical stability. Explicit \code{bernstein.basis=FALSE} is
    honored, but raw-polynomial search can be poorly conditioned at
    higher degrees. For
    \code{regtype="ll"} and \code{regtype="lp"}, a pre-optimization
    design-conditioning check is performed on the training continuous
    design: rank deficiency triggers an error, and large condition
    number (\eqn{\kappa(B)}{kappa(B)}) triggers warning/error thresholds
    to avoid unstable optimization dominated by ridging.
  }

  \item{degree}{
    a user-supplied vector of fixed polynomial degrees for the
    continuous predictors (exactly one degree per continuous predictor),
    relevant only when \code{regtype="lp"}. When
    \code{degree.select="manual"}, this must be supplied explicitly.
    Entries must be non-negative integers in \code{[0,12]}.
    With no continuous predictors, \code{regtype="lp"} is only
    admissible when \code{degree=0}, the local-constant equivalent.
    Bandwidth optimization treats this vector as fixed input and
    optimizes only bandwidths.
  }

  \item{regtype}{
    a character string specifying which type of kernel regression
    estimator to use. \code{lc} specifies a local-constant estimator
    (Nadaraya-Watson) and \code{ll} specifies a local-linear
    estimator. \code{lp} specifies a local polynomial
    estimator with polynomial degree(s) given by \code{degree} for
    continuous predictors, or selected automatically when
    \code{degree.select != "manual"}. \code{ll} and positive-degree
    \code{lp} require at least one continuous predictor; for
    categorical-only predictors use \code{lc} or \code{lp} with
    \code{degree=0}. Defaults to \code{lc}.
  }

\subsection{NOMAD Search Controls}{
These arguments control the optional NOMAD direct-search route for local-polynomial degree and bandwidth search.
  }

  \item{nomad}{
    logical shortcut for the recommended automatic local-polynomial
    NOMAD route. When \code{TRUE}, any missing values among
    \code{regtype}, \code{search.engine}, \code{degree.select},
    \code{bernstein.basis}, \code{degree.min}, \code{degree.max},
    \code{degree.verify}, and \code{bwtype} are filled with
    \code{regtype="lp"}, \code{search.engine="nomad+powell"},
    \code{degree.select="coordinate"},
    \code{bernstein.basis=TRUE}, \code{degree.min=0L},
    \code{degree.max=10L}, \code{degree.verify=FALSE}, and
    \code{bwtype="fixed"}. Explicit incompatible settings error
    immediately; in particular, \code{nomad=TRUE} currently requires
    \code{regtype="lp"}, \code{bwtype="fixed"}, automatic degree
    search, \code{bernstein.basis=TRUE}, no explicit \code{degree},
    at least one continuous predictor, and
    \code{search.engine \%in\% c("nomad", "nomad+powell")}.
    This shortcut does not change the meaning of \code{nmulti} or
    \code{nomad.nmulti}: \code{nmulti} remains the outer restart count,
    while \code{nomad.nmulti} controls inner
    \code{crs::snomadr()} multistarts within each outer restart.
    Returned bandwidth objects retain this normalized preset metadata in
    \code{bw$nomad.shortcut} for a returned object \code{bw}; when
    available, \code{nomad.time} and \code{powell.time} record the
    direct-search and Powell-polish timing components.
  }

  \item{nomad.nmulti}{
    non-negative integer controlling the inner
    \code{crs::snomadr()} multistart count used within each outer
    NOMAD restart when \code{regtype="lp"} and automatic degree search
    uses \code{search.engine="nomad"} or \code{"nomad+powell"}.
    Defaults to \code{0L}, which preserves the current one-start-per-
    restart behavior. This does not replace \code{nmulti}:
    \code{nmulti} controls outer restarts, while \code{nomad.nmulti}
    controls inner NOMAD multistarts within each outer restart.
  }

  \item{nomad.remin}{
    logical flag controlling the optional second NOMAD hot start. When
    \code{TRUE}, NOMAD is restarted once from the best full candidate found,
    including both bandwidth and degree coordinates. Defaults to \code{FALSE};
    current simulation evidence favors the one-pass NOMAD default for routine
    use, while leaving this switch available for sensitivity checks.
  }


  \item{search.engine}{
    character string controlling the automatic local-polynomial search
    backend when \code{regtype="lp"} and
    \code{degree.select != "manual"}. \code{"nomad+powell"} (default)
    performs direct joint mixed discrete/continuous search over fixed
    bandwidths and the degree vector using \code{crs::snomadr()},
    followed by one Powell hot start from the NOMAD solution.
    \code{"nomad"} performs the direct joint NOMAD search without the
    Powell refinement. \code{"cell"} uses the legacy profiled
    degree-grid search built from repeated fixed-degree bandwidth
    solves. NOMAD-based search currently requires \code{bwtype="fixed"},
    \code{degree.verify=FALSE}, and the suggested package
    \pkg{crs} to be installed.
  }

\subsection{Numerical Search And Tolerance Controls}{
These controls set optimizer tolerances, restart behavior, invalid-candidate penalties, and bounded search transformations.
}

  \item{ftol}{
    fractional tolerance on the value of the cross-validation function
    evaluated at located minima (of order the machine precision or
    perhaps slightly larger so as not to be diddled by
    roundoff). Defaults to \code{1.490116e-07}
    (1.0e+01*sqrt(.Machine$double.eps)).
  }

  \item{invalid.penalty}{ a character string specifying the penalty
    used when the optimizer encounters invalid bandwidths.
    \code{"baseline"} returns a finite penalty based on a baseline
    objective; \code{"dbmax"} returns \code{DBL\_MAX}. Defaults to
    \code{"baseline"}.}

  \item{itmax}{
    integer number of iterations before failure in the numerical
    optimization routine. Defaults to \code{10000}.
  }

  \item{nmulti}{
    integer number of times to restart the process of finding extrema of
    the cross-validation function from different (random) initial
    points. Defaults to \code{min(2,ncol(xdat))}.
  }

  \item{penalty.multiplier}{ a numeric multiplier applied to the
    baseline penalty when \code{invalid.penalty="baseline"}.
    Defaults to \code{10}.}

  \item{powell.remin}{
    logical flag controlling Powell restart-from-minimum behavior. For
    ordinary fixed-degree Powell-style search, \code{TRUE} restarts the
    local search from the located minimum. For \code{search.engine="nomad+powell"},
    this controls only the final Powell bandwidth-polish step. The default
    is \code{TRUE} for ordinary Powell routes and \code{FALSE} for the
    Powell polish after NOMAD unless explicitly supplied.
  }

  \item{small}{
    a small number used to bracket a minimum (it is hopeless to ask for
    a bracketing interval of width less than sqrt(epsilon) times its
    central value, a fractional width of only about 10-04 (single
    precision) or 3x10-8 (double precision)). Defaults to \code{small
      = 1.490116e-05 (1.0e+03*sqrt(.Machine$double.eps))}.
  }

  \item{tol}{
    tolerance on the position of located minima of the cross-validation
    function (tol should generally be no smaller than the square root of
    your machine's floating point precision). Defaults to \code{
      1.490116e-04 (1.0e+04*sqrt(.Machine$double.eps))}.
  }

  \item{transform.bounds}{ a logical value that when set to \code{TRUE}
    applies an internal transformation that maps the unconstrained
    search to the feasible bandwidth domain. Defaults to
    \code{FALSE}.}

\subsection{Additional Arguments}{
These arguments collect remaining controls passed through S3 methods.
}

  \item{\dots}{
    additional arguments supplied to specify the bandwidth type,
    kernel types, selection methods, and so on, detailed below.
  }
}
\details{

The \code{scale.factor.*} controls are dimensionless search
controls. The package converts scale factors to bandwidths using the
estimator-specific scaling encoded in the bandwidth object, including
kernel order and the number of continuous variables relevant for the
estimator. Users should not pre-multiply these controls by sample-size
or standard-deviation factors.

  \code{scale.factor.init} controls the deterministic first search
  start. \code{scale.factor.init.lower} and
  \code{scale.factor.init.upper} define the random multistart interval.
  \code{scale.factor.search.lower} is the lower admissibility bound for
  continuous fixed-bandwidth search candidates. The effective first
  start is \code{max(scale.factor.init, scale.factor.search.lower)},
  and the effective random-start lower endpoint is
  \code{max(scale.factor.init.lower, scale.factor.search.lower)}.
  \code{scale.factor.init.upper} must be at least that effective lower
  endpoint; the package errors rather than silently expanding the user's
  interval.

  When \code{scale.factor.search.lower} is \code{NULL}, an existing
  bandwidth object's stored floor is inherited when available;
  otherwise the package default \code{0.1} is used. Explicit bandwidths
  supplied for storage with \code{bandwidth.compute = FALSE} are not
  rewritten by the search floor.

  Categorical search-start controls such as \code{dfac.init},
  \code{lbd.init}, and \code{hbd.init} have separate semantics and are
  not affected by \code{scale.factor.search.lower}.

  Documentation guide: see \code{\link{np.kernels}} for kernels,
  \code{\link{np.options}} for global options, and
  \code{\link{plot}}, \code{\link{plot.np}} for plotting options.

  The bandwidth-selection argument surface is easiest to read by
  decision group: data and existing bandwidth inputs;
  local-polynomial/NOMAD controls when polynomial-adaptive regression
  is requested; bandwidth criterion and representation; continuous
  kernel and support controls beginning with \code{cker*};
  categorical kernel controls \code{ukertype} and \code{okertype}; and
  numerical search initialization, tolerances, and feasibility
  controls. Users who call \code{\link{npreg}} without a bandwidth
  object can pass these same bandwidth-selection controls through that
  function's \code{\dots}.

  For S3 plotting help, see \code{\link{plot.np}}. You can list
  available plot methods with \code{methods("plot")}.

  \code{npregbw} implements a variety of methods for choosing
  bandwidths for multivariate (\eqn{p}-variate) regression data defined
  over a set of possibly continuous and/or discrete (unordered, ordered)
  data. The approach is based on Li and Racine (2003) who employ
  \sQuote{generalized product kernels} that admit a mix of continuous
  and discrete data types.

  The cross-validation methods employ multivariate numerical search
  algorithms. For fixed-degree local-constant/local-linear regression,
  and for local-polynomial regression with \code{degree.select="manual"},
  the bandwidth search uses multidimensional Powell direction-set
  optimization.

  Bandwidths can (and will) differ for each variable which is, of
  course, desirable.

  Three classes of kernel estimators for the continuous data types are
  available: fixed, adaptive nearest-neighbor, and generalized
  nearest-neighbor. Adaptive nearest-neighbor bandwidths change with
  each sample realization in the set, \eqn{x_i}{x[i]}, when estimating the
  density at the point \eqn{x}. Generalized nearest-neighbor bandwidths change
  with the point at which the density is estimated, \eqn{x}. Fixed bandwidths
  are constant over the support of \eqn{x}.

  \code{npregbw} may be invoked \emph{either} with a formula-like
  symbolic 
  description of variables on which bandwidth selection is to be
  performed \emph{or} through a simpler interface whereby data is passed
  directly to the function via the \code{xdat} and \code{ydat}
  parameters. Use of these two interfaces is \bold{mutually exclusive}.

  Data contained in the data frame \code{xdat} may be a mix of
  continuous (default), unordered discrete (to be specified in the data
  frame \code{xdat} using \code{\link{factor}}), and ordered discrete
  (to be specified in the data frame \code{xdat} using
  \code{\link{ordered}}). Data can be entered in an arbitrary order and
  data types will be detected automatically by the routine (see
  \code{\link{np}} for details).

  Data for which bandwidths are to be estimated may be specified
  symbolically. A typical description has the form \code{dependent data
    ~ explanatory data},
  where \code{dependent data} is a univariate response, and
  \code{explanatory data} is a
  series of variables specified by name, separated by 
  the separation character '+'. For example, \code{ y1 ~ x1 + x2 }
  specifies that the bandwidths for the regression of response \code{y1}
  and 
  nonparametric regressors \code{x1} and \code{x2} are to be estimated.
  See below for further examples. 

  A variety of kernels may be specified by the user. Kernels implemented
  for continuous data types include the second, fourth, sixth, and eighth
  order Gaussian and Epanechnikov kernels, and the uniform
  kernel. Unordered discrete data types use a variation on Aitchison and
  Aitken's (1976) kernel, while ordered data types use a variation of the
  Wang and van Ryzin (1981) kernel.

  When \code{regtype="lp"} and \code{degree.select != "manual"},
  \code{npregbw} can jointly determine the continuous-predictor degree
  vector and bandwidth coordinates. With \code{search.engine="cell"},
  the objective is profiled over the degree grid using cached
  coordinate-wise or exhaustive search together with the existing
  fixed-degree bandwidth optimizer. With
  \code{search.engine="nomad"} or \code{"nomad+powell"}, the package
  instead evaluates the cross-validation criterion directly over the
  joint space of fixed bandwidths and polynomial degrees using
  \code{crs::snomadr()}. \code{"nomad+powell"} then performs one Powell
  hot start from the NOMAD solution and retains the better of the
  direct NOMAD and polished solutions. This direct joint-search route
  follows the polynomial-adaptive cross-validation rationale of Hall
  and Racine (2015). When \code{bernstein.basis} is not explicitly
  supplied, the automatic search route defaults to
  \code{bernstein.basis=TRUE} for numerical stability; explicit
  \code{bernstein.basis=FALSE} is honored but can be poorly conditioned
  at higher degrees. NOMAD multistarts are initialized more
  conservatively than the full degree search box: start 1 is the
  user-supplied degree/bandwidth vector when provided and otherwise a
  clipped degree-one vector, while later starts are reproducible random
  draws from a reduced degree proposal box whose candidates are screened
  using \code{dim_basis()}. This heuristic is used only to obtain
  feasible, numerically safer, and quicker initial evaluations; it does
  not restrict the admissible degree region searched by NOMAD. The
  direct NOMAD backend is provided by the suggested package
  \code{crs}, so install \code{crs} before using
  \code{search.engine="nomad"}, \code{"nomad+powell"}, or
  \code{nomad=TRUE}.

  Setting \code{nomad=TRUE} is a convenience preset for this automatic
  LP route, not a generic optimizer alias. For regression it expands any
  missing values to the equivalent long-form call
  \preformatted{
    npregbw(...,
            regtype = "lp",
            search.engine = "nomad+powell",
            degree.select = "coordinate",
            bernstein.basis = TRUE,
            degree.min = 0L,
            degree.max = 10L,
            degree.verify = FALSE,
            bwtype = "fixed")
  }
  Compatible explicit tuning arguments are respected. Incompatible
  explicit settings fail fast so the shortcut never silently changes
  user-selected semantics.
  When the direct NOMAD route is active, \code{nmulti} controls the
  package-level outer restart count while \code{nomad.nmulti}
  controls the inner \code{crs::snomadr()} multistart count used within
  each outer restart. The default \code{nomad.nmulti=0L} preserves the
  current single-start inner NOMAD behavior.

  The use of compactly supported kernels or the occurrence of small
  bandwidths during cross-validation can lead to numerical problems for
  the local linear estimator when computing the locally weighted least
  squares solution. To overcome this problem we rely on a form or
  \sQuote{ridging} proposed by Cheng, Hall, and Titterington (1997),
  modified so that we solve the problem pointwise rather than globally
  (i.e. only when it is needed).

  The optimizer invoked for search is Powell's conjugate direction
  method which requires the setting of (non-random) initial values and
  search directions for bandwidths, and, when restarting, random values
  for successive invocations. Bandwidths for \code{numeric} variables
  are scaled by robust measures of spread, the sample size, and the
  number of \code{numeric} variables where appropriate. Two sets of
  parameters for bandwidths for \code{numeric} can be modified, those
  for initial values for the parameters themselves, and those for the
  directions taken (Powell's algorithm does not involve explicit
  computation of the function's gradient). The default values are set by
  considering search performance for a variety of difficult test cases
  and simulated cases. We highly recommend restarting search a large
  number of times to avoid the presence of local minima (achieved by
  modifying \code{nmulti}). Further refinement for difficult cases can
  be achieved by modifying these sets of parameters. However, these
  parameters are intended more for the authors of the package to enable
  \sQuote{tuning} for various methods rather than for the user
  themselves.

}
\value{
  \code{npregbw} returns a \code{rbandwidth} object, with the
  following components:

  \item{bw}{ bandwidth(s), scale factor(s) or nearest neighbours for the
    data, \code{xdat} }
  \item{fval}{ objective function value at minimum }

  if \code{bwtype} is set to \code{fixed}, an object containing bandwidths
  (or scale factors if \code{bwscaling = TRUE}) is returned. If it is set to
  \code{generalized_nn} or \code{adaptive_nn}, then instead the \eqn{k}th nearest
  neighbors are returned for the continuous variables while the discrete
  kernel bandwidths are returned for the discrete variables. Bandwidths
  are stored under the component name \code{bw}, with each
  element \eqn{i} corresponding to column \eqn{i} of input data
  \code{xdat}.

  The functions \code{\link{predict}}, \code{\link{summary}}, and \code{\link{plot}} support
  objects of this class.
}

\section{Book And Method Pointers}{
  \code{npregbw} selects bandwidths and, when requested, local-polynomial
  degree/search metadata for estimating the conditional mean
  \eqn{m(x)=E[Y\mid X=x]}. For the estimator target and local-polynomial
  interpretation, see \code{\link{npreg}}.

  For book-length derivations, see Li and Racine (2007), Chapter 2
  \emph{Regression}, especially Sections 2.2, 2.4, and 2.5, and Chapter 4
  \emph{Kernel Estimation with Mixed Data}, especially Sections 4.2 and
  4.4. The later workflow treatment is Racine (2019), Chapter 6
  \emph{Conditional Mean Function Estimation}.
}

\references{
  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Cheng, M.-Y. and P. Hall and D.M. Titterington (1997), \dQuote{On the
  shrinkage of local linear curve estimators,} Statistics and Computing,
  7, 11-17.

  Fan, J. and I. Gijbels (1996), \emph{Local Polynomial Modelling and
  Its Applications,} Chapman and Hall.

  Hall, P. and J.S. Racine (2015), \dQuote{Infinite Order Cross-Validated
  Local Polynomial Regression,} Journal of Econometrics, 185, 510-525.

  Hall, P. and Q. Li and J.S. Racine (2007), \dQuote{Nonparametric
  estimation of regression functions in the presence of irrelevant
  regressors,} The Review of Economics and Statistics, 89, 784-789.

  Hurvich, C.M. and J.S. Simonoff and C.L. Tsai (1998),
  \dQuote{Smoothing parameter selection in nonparametric regression
    using an improved Akaike information criterion,} Journal of the
  Royal Statistical Society B, 60, 271-293.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
  and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2004), \dQuote{Cross-validated local linear
    nonparametric regression,} Statistica Sinica, 14, 485-512.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,}
  Cambridge University Press.
  
  Racine, J.S. and Q. Li (2004), \dQuote{Nonparametric estimation of regression
    functions with both categorical and continuous data,} Journal of
  Econometrics, 119, 99-130.
  
  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth estimators
    for discrete distributions,}  Biometrika, 68, 301-309.
}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.

  Caution: multivariate data-driven bandwidth selection methods are, by
  their nature, \emph{computationally intensive}. Virtually all methods
  require dropping the \eqn{i}th observation from the data set, computing an
  object, repeating this for all observations in the sample, then
  averaging each of these leave-one-out estimates for a \emph{given}
  value of the bandwidth vector, and only then repeating this a large
  number of times in order to conduct multivariate numerical
  minimization/maximization. Furthermore, due to the potential for local
  minima/maxima, \emph{restarting this procedure a large number of times may
  often be necessary}. This can be frustrating for users possessing
  large datasets. For exploratory purposes, you may wish to override the
  default search tolerances, say, setting ftol=.01 and tol=.01 and
  conduct multistarting (the default is to restart min(2, ncol(xdat))
  times) as is done for a number of examples. Once the procedure
  terminates, you can restart search with default tolerances using those
  bandwidths obtained from the less rigorous search (i.e., set
  \code{bws=bw} on subsequent calls to this routine where \code{bw} is
  the initial bandwidth object).  A version of this package using the
  \code{Rmpi} wrapper is under development that allows one to deploy
  this software in a clustered computing environment to facilitate
  computation involving large datasets.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}, \code{\link{plot.np}}
  \code{\link{npreg}}
}
\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): For this example, we compute a
# Bivariate nonparametric regression estimate for Giovanni Baiocchi's
# Italian income panel (see Italy for details)

data("Italy")
with(Italy, {

# Compute the least-squares cross-validated bandwidths for the local
# constant estimator (default)

bw <- npregbw(formula=gdp~ordered(year))

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Supply your own bandwidth...

bw <- npregbw(formula=gdp~ordered(year), bws=c(0.75),
              bandwidth.compute=FALSE)

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Treat year as continuous and supply your own scaling factor c in
# c sigma n^{-1/(2p+q)}

bw <- npregbw(formula=gdp~year, bws=c(1.06),
              bandwidth.compute=FALSE, 
              bwscaling=TRUE)

summary(bw)

# Note - see also the example for npudensbw() for more extensive
# multiple illustrations of how to change the kernel function, kernel
# order, bandwidth type and so forth.

})

# EXAMPLE 1 (INTERFACE=DATA FRAME): For this example, we compute a
# Bivariate nonparametric regression estimate for Giovanni Baiocchi's
# Italian income panel (see Italy for details)

data("Italy")
with(Italy, {

# Compute the least-squares cross-validated bandwidths for the local
# constant estimator (default)

bw <- npregbw(xdat=ordered(year), ydat=gdp)

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Supply your own bandwidth...

bw <- npregbw(xdat=ordered(year), ydat=gdp, bws=c(0.75),
              bandwidth.compute=FALSE)

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Treat year as continuous and supply your own scaling factor c in
# c sigma n^{-1/(2p+q)}

bw <- npregbw(xdat=year, ydat=gdp, bws=c(1.06),
              bandwidth.compute=FALSE, 
              bwscaling=TRUE)

summary(bw)

# Note - see also the example for npudensbw() for more extensive
# multiple illustrations of how to change the kernel function, kernel
# order, bandwidth type and so forth.

})
} % enddontrun
}
\keyword{nonparametric}

\name{npquantile}
\alias{npquantile}

\title{ Kernel Univariate Quantile Estimation }
\description{

  \code{npquantile} computes smooth quantiles from a univariate
  unconditional kernel cumulative distribution estimate given data and,
  optionally, a bandwidth specification i.e. a \code{dbandwidth} object
  using the bandwidth selection method of Li, Li and Racine (2017).

}
\usage{
npquantile(x = NULL,
           tau = c(0.01,0.05,0.25,0.50,0.75,0.95,0.99),
           num.eval = 10000,
           bws = NULL,
           f = 1,
           \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the distribution object, data, and bandwidth controls used for quantile extraction.
}

  \item{bws}{
    an optional \code{dbandwidth} specification (if already computed
    avoid unnecessary computation inside \code{npquantile}). This must
    be set as a \code{dbandwidth} object returned from an invocation of
    \code{\link{npudistbw}}. If not provided \code{\link{npudistbw}} is
    invoked with optional arguments passed via \code{\dots}.
  }


  \item{x}{
    a univariate vector of type \code{\link{numeric}} containing sample
    realizations (training data) used to estimate the cumulative
    distribution (must be the same training data used to compute the
    bandwidth object \code{bws} passed in).
  }

\subsection{Evaluation And Quantile Controls}{
These arguments control the target quantile level, evaluation size, and distribution interpolation.
}

  \item{f}{
    an optional argument fed to \code{\link{extendrange}}. Defaults to
    \code{1}. See \code{?\link{extendrange}} for details.
  }


  \item{num.eval}{
    an optional integer specifying the length of the grid on which the
    quasi-inverse is computed. Defaults to \code{10000}.
  }

  \item{tau}{
    an optional vector containing the probabilities for quantile(s) to
    be estimated (must contain numbers in \eqn{[0,1]}). Defaults to
    \code{c(0.01,0.05,0.25,0.50,0.75,0.95,0.99)}.
  }

\subsection{Additional Arguments}{
Further arguments are passed to bandwidth or distribution routines as needed.
}

  \item{\dots}{
    additional arguments supplied to specify the bandwidth type, kernel
    types, bandwidth selection methods, and so on. See
    \code{?\link{npudistbw}} for details.
  }


}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.


  Typical usage is
  \preformatted{
    x <- rchisq(100,df=10)
    npquantile(x)
  }

  The quantile function \eqn{q_\tau} is defined to be the
  left-continuous inverse of the distribution function \eqn{F(x)},
  i.e. \eqn{q_\tau = \inf\{x: F(x) \ge \tau\}}.

  A traditional estimator of \eqn{q_\tau} is the \eqn{\tau}th sample
  quantile. However, these estimates suffer from lack of efficiency
  arising from variability of individual order statistics; see Sheather
  and Marron (1990) and Hyndman and Fan (1996) for methods that
  interpolate/smooth the order statistics, each of which discussed in
  the latter can be invoked through \code{\link{quantile}} via
  \code{type=j}, \code{j=1,\dots,9}.

  The function \code{npquantile} implements a method for estimating
  smooth quantiles based on the quasi-inverse of a \code{\link{npudist}}
  object where \eqn{F(x)} is replaced with its kernel estimator and
  bandwidth selection is that appropriate for such objects; see
  Definition 2.3.6, page 21, Nelsen 2006 for a definition of the
  quasi-inverse of \eqn{F(x)}.

  For construction of the quasi-inverse we create a grid of evaluation
  points based on the function \code{\link{extendrange}} along with the
  sample quantiles themselves computed from invocation of
  \code{\link{quantile}}. The coarseness of the grid defined by
  \code{\link{extendrange}} (which has been passed the option
  \code{f=1}) is controlled by \code{num.eval}.

  Note that for any value of \eqn{\tau} less/greater than the
  smallest/largest value of \eqn{F(x)} computed for the evaluation data
  (i.e. that outlined in the paragraph above), the quantile returned for
  such values is that associated with the smallest/largest value of
  \eqn{F(x)}, respectively.

}
\value{

  \code{\link{npquantile}} returns a vector of quantiles corresponding
  to \code{tau}.

}
\references{


  Cheng, M.-Y. and Sun, S. (2006), \dQuote{Bandwidth selection for
  kernel quantile estimation,} Journal of the Chinese Statistical
  Association, \bold{44}, 271-295.

  Hyndman, R.J. and Fan, Y. (1996), \dQuote{Sample quantiles in
  statistical packages,} American Statistician, \bold{50}, 361-365.

  Li, Q. and J.S. Racine (2017), \dQuote{Smooth Unconditional Quantile
  Estimation,} Manuscript.

  Li, C. and H. Li and J.S. Racine (2017), \dQuote{Cross-Validated Mixed
  Datatype Bandwidth Selection for Nonparametric Cumulative
  Distribution/Survivor Functions,} Econometric Reviews, \bold{36},
  970-987.

  Nelsen, R.B. (2006), \emph{An Introduction to Copulas,} Second
  Edition, Springer-Verlag.

  Sheather, S. and J.S. Marron (1990), \dQuote{Kernel quantile
  estimators,} Journal of the American Statistical Association, Vol. 85,
  No. 410, 410-416.

  Yang, S.-S. (1985), \dQuote{A Smooth Nonparametric Estimator of a
  Quantile Function,} Journal of the American Statistical Association,
  \bold{80}, 1004-1011.

}
\section{Usage Issues}{
  Cross-validated bandwidth selection is used by default
  (\code{\link{npudistbw}}). For large datasets this can be
  computationally demanding. In such cases one might instead consider a
  rule-of-thumb bandwidth (\code{bwmethod="normal-reference"}) or,
  alternatively, use kd-trees (\code{options(np.tree=TRUE)} along with a
  bounded kernel (\code{ckertype="epanechnikov"})), both of which will
  reduce the computational burden appreciably.
}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}

  \code{\link{quantile}} for various types of sample quantiles;
  \code{\link{ecdf}} for empirical distributions of which
  \code{\link{quantile}} is an inverse; \code{\link{boxplot.stats}} and
  \code{\link{fivenum}} for computing other versions of quartiles;
  \code{\link[logspline]{qlogspline}} for logspline density quantiles;
  \code{\link[ks]{qkde}} for alternative kernel quantiles, etc.

}

\examples{
\dontrun{
## Simulate data from a chi-square distribution
df <- 50
x <- rchisq(100,df=df)

## Vector of quantiles desired
tau <- c(0.01,0.05,0.25,0.50,0.75,0.95,0.99)

## Compute kernel smoothed sample quantiles
npquantile(x,tau)

## Compute sample quantiles using the default method in R (Type 7)
quantile(x,tau)

## True quantiles based on known distribution
qchisq(tau,df=df)
} % enddontrun
}
\keyword{ nonparametric }% at least one, from doc/KEYWORDS

%

\name{npsigtest}
\alias{npsigtest}
\alias{npsigtest.default}
\alias{npsigtest.formula}
\alias{npsigtest.rbandwidth}
\alias{npsigtest.npregression}

\title{Kernel Regression Significance Test with Mixed Data Types}
\description{
  \code{npsigtest} implements a consistent test of significance of an
  explanatory variable(s) in a nonparametric regression setting that is
  analogous to a simple \eqn{t}-test (\eqn{F}-test) in a parametric
  regression setting. The test is based on Racine, Hart, and Li (2006)
  and Racine (1997).
}
\usage{
npsigtest(bws, 
          \dots)

\method{npsigtest}{formula}(bws, 
          data = NULL, 
          \dots)

\method{npsigtest}{npregression}(bws, \dots)

\method{npsigtest}{default}(bws, 
          xdat, 
          ydat, 
          \dots)

\method{npsigtest}{rbandwidth}(bws,
          xdat = stop("data xdat missing"),
          ydat = stop("data ydat missing"),
          boot.num = 399,
          boot.method = c("iid","wild","wild-rademacher","pairwise"),
          boot.type = c("I","II"),
          pivot=TRUE,
          joint=FALSE,
          index = seq_len(ncol(xdat)),
          random.seed = 42,
          \dots)

}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the bandwidth object, formula/data interface, and explicit data inputs.
}

  \item{bws}{
    a bandwidth specification. This can be set as a \code{rbandwidth}
    object returned from a previous invocation, or as a vector of
    bandwidths, with each element \eqn{i} corresponding to the bandwidth
    for column \eqn{i} in \code{xdat}. In either case, the bandwidth
    supplied will serve as a starting point in the numerical search for
    optimal bandwidths when using \code{boot.type="II"}. If specified
    as a vector, then additional arguments will need to be supplied as
    necessary to specify the bandwidth type, kernel types, selection
    methods, and so on.
  }


  \item{data}{
    an optional data frame, list or environment (or object coercible to
    a data frame by \code{\link{as.data.frame}}) containing the
    variables in the model. If not found in data, the variables are
    taken from \code{environment(bws)}, typically the environment from
    which \code{\link{npregbw}} was called.
  }


  \item{xdat}{
    a \eqn{p}-variate data frame of explanatory data (training data)
    used to calculate the regression estimators.
  }


  \item{ydat}{
    a one (1) dimensional numeric or integer vector of dependent data,
    each element \eqn{i} corresponding to each observation (row) \eqn{i}
    of \code{xdat}.
  }


\subsection{Bootstrap And Test Controls}{
These arguments control bootstrap execution, tested effects, and reproducibility settings.
}

  \item{boot.method}{
    a character string used to specify the bootstrap method for
    determining the null distribution. \code{pairwise} resamples
    pairwise, while the remaining methods use a residual bootstrap
    procedure.  \code{iid} will generate independent identically
    distributed draws. \code{wild} will use a wild
    bootstrap. \code{wild-rademacher} will use a wild bootstrap with
    Rademacher variables. Defaults to \code{iid}.
  }


  \item{boot.num}{
    an integer value specifying the number of bootstrap replications to
    use. Defaults to \code{399}.
  }


  \item{boot.type}{
    a character string specifying whether to use a \sQuote{Bootstrap I} or
    \sQuote{Bootstrap II} method (see Racine, Hart, and Li (2006) for
    details). The \sQuote{Bootstrap II} method re-runs cross-validation for
    each bootstrap replication and uses the new cross-validated
    bandwidth for variable \eqn{i} and the original ones for the
    remaining variables. Defaults to \code{boot.type="I"}.
  }


  \item{index}{
    a vector of indices for the columns of \code{xdat} for which the
    test of significance is to be conducted. Defaults to
    (1,2,\dots,\eqn{p}) where \eqn{p} is the number of columns in
    \code{xdat}.
  }


  \item{joint}{
    a logical value which specifies whether to conduct a joint test or
    individual test. This is to be used in conjunction with \code{index}
    where \code{index} contains two or more integers corresponding to
    the variables being tested, where the integers correspond to the
    variables in the order in which they appear among the set of
    explanatory variables in the function call to
    \code{npreg}/\code{npregbw}. Defaults to \code{FALSE}.
  }


  \item{pivot}{
    a logical value which specifies whether to bootstrap a pivotal
    statistic or not (pivoting is achieved by dividing gradient
    estimates by their asymptotic standard errors). Defaults to
    \code{TRUE}.
  }


  \item{random.seed}{
    an integer used to seed R's random number generator. This is to
    ensure replicability. Defaults to 42.
  }


\subsection{Additional Arguments}{
Further arguments are passed to the bandwidth-selection routines used by the test.
}

  \item{\dots}{
    additional arguments supplied to specify the bandwidth type,
    kernel types, selection methods, and so on, detailed below.
  }


}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.


  \code{npsigtest} implements a variety of methods for computing the
  null distribution of the test statistic and allows the user to
  investigate the impact of a variety of default settings including
  whether or not to pivot the statistic (\code{pivot}), whether pairwise
  or residual resampling is to be used (\code{boot.method}), and whether
  or not to recompute the bandwidths for the variables being tested
  (\code{boot.type}), among others.

  Defaults are chosen so as to provide reasonable behaviour in a broad
  range of settings and this involves a trade-off between computational
  expense and finite-sample performance. However, the default
  \code{boot.type="I"}, though computationally expedient, can deliver a
  test that can be slightly over-sized in small sample settings (e.g.
  at the 5\% level the test might reject 8\% of the time for samples of
  size \eqn{n=100} for some data generating processes).  If the default
  setting (\code{boot.type="I"}) delivers a P-value that is in the
  neighborhood (i.e. slightly smaller) of any classical level
  (e.g. 0.05) and you only have a modest amount of data, it might be
  prudent to re-run the test using the more computationally intensive
  \code{boot.type="II"} setting to confirm the original result. Note
  also that \code{boot.method="pairwise"} is not recommended for the
  multivariate local linear estimator due to substantial size
  distortions that may arise in certain cases.

}

\value{
  \code{npsigtest} returns an object of type
  \code{sigtest}. \code{\link{summary}} supports \code{sigtest} objects. It
  has the
  following components:

  \item{In}{ the vector of statistics \code{In} }
  \item{P}{ the vector of P-values for each statistic in \code{In}}
  \item{In.bootstrap}{ contains a matrix of the bootstrap
    replications of the vector \code{In}, each column corresponding to
    replications associated with explanatory variables in \code{xdat}
    indexed by \code{index} (e.g., if you selected \code{index = c(1,4)}
    then In.bootstrap will have two columns, the first being the
    bootstrap replications of \code{In} associated with variable
    \code{1}, the second with variable \code{4}).}

}
\section{Book And Method Pointers}{
  \code{npsigtest} assesses predictor relevance in nonparametric
  regression by comparing the fitted unrestricted regression with
  restrictions implied by excluding or smoothing over selected
  regressors. Bootstrap choices trade computational cost against
  finite-sample behavior.

  For book-length background, see Li and Racine (2007), Chapter 12
  \emph{Model Specification Tests}, especially Sections 12.3 and 12.3.5,
  and Chapter 13 \emph{Nonsmoothing Tests}, especially Section 13.3 where
  relevant. For the underlying conditional-mean setting, see Racine
  (2019), Chapter 6 \emph{Conditional Mean Function Estimation}.
}

\references{
  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics:
    Theory and Practice,} Princeton University Press.

  Racine, J.S., J. Hart, and Q. Li (2006), \dQuote{Testing the
    significance of categorical predictor variables in nonparametric
    regression models,} Econometric Reviews, 25, 523-544.

  Racine, J.S. (1997), \dQuote{Consistent significance testing for
    nonparametric regression,} Journal of Business and Economic
  Statistics 15, 369-379.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth estimators
    for discrete distributions,}  Biometrika, 68, 301-309.

}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{

  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work
  as intended on mixed data types and will coerce the data to the same
  type.

  Caution: bootstrap methods are, by their nature, \emph{computationally
  intensive}. This can be frustrating for users possessing large
  datasets. For exploratory purposes, you may wish to override the
  default number of bootstrap replications, say, setting them to
  \code{boot.num=99}. A version of this package using the \code{Rmpi}
  wrapper is under development that allows one to deploy this software
  in a clustered computing environment to facilitate computation
  involving large datasets.

}

\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): For this example, we simulate 100 draws
# from a DGP in which z, the first column of X, is an irrelevant
# discrete variable

set.seed(12345)

n <- 100

z <- rbinom(n,1,.5)
x1 <- rnorm(n)
x2 <- runif(n,-2,2)

y <- x1 + x2 + rnorm(n)

# Next, we must compute bandwidths for our regression model. In this
# case we conduct local linear regression. Note - this may take a few
# minutes depending on the speed of your computer...

bw <- npregbw(formula=y~factor(z)+x1+x2,regtype="ll",bwmethod="cv.aic")

# We then compute a vector of tests corresponding to the columns of
# X. Note - this may take a few minutes depending on the speed of your
# computer... we have to generate the null distribution of the statistic
# for each variable whose significance is being tested using 399
# bootstrap replications for each...

npsigtest(bws=bw)

# If you wished, you could conduct the test for, say, variables 1 and 3
# only, as in

npsigtest(bws=bw,index=c(1,3))

# EXAMPLE 1 (INTERFACE=DATA FRAME): For this example, we simulate 100
# draws from a DGP in which z, the first column of X, is an irrelevant
# discrete variable

set.seed(12345)

n <- 100

z <- rbinom(n,1,.5)
x1 <- rnorm(n)
x2 <- runif(n,-2,2)

X <- data.frame(factor(z),x1,x2)

y <- x1 + x2 + rnorm(n)

# Next, we must compute bandwidths for our regression model. In this
# case we conduct local linear regression. Note - this may take a few
# minutes depending on the speed of your computer...

bw <- npregbw(xdat=X,ydat=y,regtype="ll",bwmethod="cv.aic")

# We then compute a vector of tests corresponding to the columns of
# X. Note - this may take a few minutes depending on the speed of your
# computer... we have to generate the null distribution of the statistic
# for each variable whose significance is being tested using 399
# bootstrap replications for each...

npsigtest(bws=bw)

# If you wished, you could conduct the test for, say, variables 1 and 3
# only, as in

npsigtest(bws=bw,index=c(1,3))
} % enddontrun
}
\keyword{ nonparametric }

\name{npuniden.boundary}
\alias{npuniden.boundary}

\title{ Kernel Bounded Univariate Density Estimation Via Boundary Kernel
  Functions }

\description{

  \code{npuniden.boundary} computes kernel univariate unconditional
  density estimates given a vector of continuously distributed training
  data and, optionally, a bandwidth (otherwise least squares
  cross-validation is used for its selection). Lower and upper bounds
  [\code{a},\code{b}] can be supplied (default is the empirical support 
  \eqn{[\min(X),\max(X)]}) and if \code{a}
  is set to \code{-Inf} there is only one bound on the right, while if
  \code{b} is set to \code{Inf} there is only one bound on the left. If
  \code{a} is set to \code{-Inf} and \code{b} to \code{Inf} and the
  Gaussian type 1 kernel function is used, this will deliver the
  standard unadjusted kernel density estimate.

}

\usage{
npuniden.boundary(X = NULL,
                  Y = NULL,
                  h = NULL,
                  a = min(X),
                  b = max(X),
                  bwmethod = c("cv.ls","cv.ml"),
                  cv = c("grid-hybrid","numeric"),
                  grid = NULL,
                  kertype = c("gaussian1","gaussian2",
                              "beta1","beta2",
                              "fb","fbl","fbu",
                              "rigaussian","gamma"),
                  nmulti = 1,
                  proper = FALSE)
}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify evaluation points, observations, support bounds, and optional bandwidths.
}

  \item{a}{
an optional lower bound (defaults to lower bound of empirical support \eqn{\min(X)})
}

  \item{b}{
an optional upper bound (defaults to upper bound of empirical support \eqn{\max(X)})
}


  \item{h}{
an optional bandwidth (>0)
}

  \item{X}{
a required numeric vector of training data lying in \eqn{[a,b]}
}

  \item{Y}{
an optional numeric vector of evaluation data lying in \eqn{[a,b]}
}

\subsection{Bandwidth Search Controls}{
These arguments control the boundary-corrected bandwidth search.
}

  \item{bwmethod}{
whether to conduct bandwidth search via least squares cross-validation
(\code{"cv.ls"}) or likelihood cross-validation (\code{"cv.ml"})
}

  \item{cv}{
an optional argument for search (default is likely more reliable in the
presence of local maxima)
}

  \item{grid}{
an optional grid used for the initial grid search when \code{cv="grid-hybrid"}
}

  \item{kertype}{
an optional kernel specification (defaults to "gaussian1")
}

  \item{nmulti}{
number of multi-starts used when \code{cv="numeric"} (defaults to 1 for
the univariate boundary problem)
}


\subsection{Proper Density Control}{
This argument controls optional proper-density repair.
}

  \item{proper}{
an optional logical value indicating whether to enforce proper density
and distribution function estimates over the range \eqn{[a,b]}
}

}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.

  Typical usages are (see below for a complete list of options and also
  the examples at the end of this help file)
  \preformatted{
    model <- npuniden.boundary(X,a=-2,b=3)
  }

  \code{npuniden.boundary} implements a variety of methods for
  estimating a univariate density function defined over a continuous
  random variable in the presence of bounds via the use of so-called
  boundary or edge kernel functions.

  The kernel functions \code{"beta1"} and \code{"beta2"} are Chen's
  (1999) type 1 and 2 kernel functions with biases of \eqn{O(h)}, the
  \code{"gamma"} kernel function is from Chen (2000) with a bias of
  \eqn{O(h)}, \code{"rigaussian"} is the reciprocal inverse Gaussian
  kernel function (Scaillet (2004), Igarashi & Kakizawa (2014)) with
  bias of \eqn{O(h)}, and \code{"gaussian1"} and \code{"gaussian2"} are
  truncated Gaussian kernel functions with biases of \eqn{O(h)} and
  \eqn{O(h^2)}, respectively. The kernel functions \code{"fb"},
  \code{"fbl"} and \code{"fbu"} are floating boundary polynomial
  biweight kernels with biases of \eqn{O(h^2)} (Scott (1992), Page
  146). Without exception, these kernel functions are asymmetric in
  general with shape that changes depending on where the density is
  being estimated (i.e., how close the estimation point \eqn{x} in
  \eqn{\hat f(x)} is to a boundary). This function is written purely in
  R, so to see the exact form for each of these kernel functions, simply
  enter the name of this function in R (i.e., enter
  \code{npuniden.boundary} after loading this package) and scroll up for
  their definitions.

  The kernel functions \code{"gamma"}, \code{"rigaussian"}, and
  \code{"fbl"} have support \eqn{[a,\infty]}. The kernel function
  \code{"fbu"} has support \eqn{[-\infty,b]}. The rest have support on
  \eqn{[a,b]}.  Note that the two sided support default values are
  \code{a=min(X)} and \code{b=max(X)}. 

  Note that data-driven bandwidth selection is more nuanced in bounded
  settings, therefore it would be prudent to manually select a bandwidth
  that is, say, 1/25th of the range of the data and manually inspect the
  estimate (say \code{h=0.05} when \eqn{X\in [0,1]}). Also, it may be
  wise to compare the density estimate with that from a histogram with
  the option \code{breaks=25}. Note also that the kernel functions
  \code{"gaussian2"}, \code{"fb"}, \code{"fbl"} and \code{"fbu"} can
  assume negative values leading to potentially negative density
  estimates, and must be trimmed when conducting likelihood
  cross-validation which can lead to oversmoothing. Least squares
  cross-validation is unaffected and appears to be more reliable in such
  instances hence is the default here.

  Scott (1992, Page 149) writes \dQuote{While boundary kernels can be
  very useful, there are potentially serious problems with real
  data. There are an infinite number of boundary kernels reflecting the
  spectrum of possible design constraints, and these kernels are not
  interchangeable. Severe artifacts can be introduced by any one of them
  in inappropriate situations. Very careful examination is required to
  avoid being victimized by the particular boundary kernel
  chosen. Artifacts can unfortunately be introduced by the choice of the
  support interval for the boundary kernel.}

  Note that since some kernel functions can assume negative values, this
  can lead to improper density estimates. The estimated distribution
  function is obtained via numerical integration of the estimated
  density function and may itself not be proper even when evaluated on
  the full range of the data \eqn{[a,b]}. Setting the option
  \code{proper=TRUE} will render the density and distribution estimates
  proper over the full range of the data, though this may not in
  general be a mean square error optimal strategy.

  Finally, note that this function is pretty bare-bones relative to
  other functions in this package. For one, at this time there is no
  automatic print support so kindly see the examples for illustrations
  of its use, among other differences.

}
\value{
  \code{npuniden.boundary} returns the following components:
  
  \item{f}{ estimated density at the points X }
  \item{F}{ estimated distribution at the points X (numeric integral of f) }
  \item{sd.f}{ asymptotic standard error of the estimated density at the points X }
  \item{sd.F}{ asymptotic standard error of the estimated distribution at the points X }
  \item{h}{ bandwidth used }
  \item{nmulti}{ number of multi-starts used }
}
\references{

  Bouezmarni, T. and Rolin, J.-M. (2003). \dQuote{Consistency of the
  beta kernel density function estimator,}  The Canadian Journal of
  Statistics / La Revue Canadienne de Statistique, 31(1):89-98.

  Chen, S. X. (1999). \dQuote{Beta kernel estimators for density
  functions,} Computational Statistics & Data Analysis, 31(2):131-145.

  Chen, S. X. (2000). \dQuote{Probability density function estimation
  using gamma kernels,} Annals of the Institute of Statistical
  Mathematics, 52(3):471-480.

  Diggle, P. (1985). \dQuote{A kernel method for smoothing point process
  data,} Journal of the Royal Statistical Society. Series C (Applied
  Statistics), 34(2):138-147.

  Igarashi, G. and Y. Kakizawa (2014). \dQuote{Re-formulation of inverse
  Gaussian, reciprocal inverse Gaussian, and Birnbaum-Saunders kernel
  estimators,} Statistics & Probability Letters, 84:235-246.

  Igarashi, G. and Y. Kakizawa (2015). \dQuote{Bias corrections for some
  asymmetric kernel estimators,} Journal of Statistical Planning and
  Inference, 159:37-63.
  
  Igarashi, G. (2016). \dQuote{Bias reductions for beta kernel
  estimation,} Journal of Nonparametric Statistics, 28(1):1-30.
  
  Racine, J. S. and Q. Li and Q. Wang, \dQuote{Boundary-adaptive 
  kernel density estimation: the case of (near) uniform density}, 
  Journal of Nonparametric Statistics, 2024, 36 (1), 146-164, 
  https://doi.org/10.1080/10485252.2023.2250011.

  Scaillet, O. (2004). \dQuote{Density estimation using inverse and
  reciprocal inverse Gaussian kernels,} Journal of Nonparametric
  Statistics, 16(1-2):217-226.

  Scott, D. W. (1992). \dQuote{Multivariate density estimation: Theory,
  practice, and visualization,} New York: Wiley.

  Zhang, S. and R. J. Karunamuni (2010). \dQuote{Boundary performance of
  the beta kernel estimators,} Journal of Nonparametric Statistics,
  22(1):81-104.
  
}
\author{
Jeffrey S. Racine \email{racinej@mcmaster.ca}
}
\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}
The \pkg{Ake}, \pkg{bde}, and \pkg{Conake} packages and the function \code{\link{npuniden.reflect}}.
}
\examples{
\dontrun{
## Example 1: f(0)=0, f(1)=1, plot boundary corrected density,
## unadjusted density, and DGP
set.seed(42)
n <- 100
X <- sort(rbeta(n,5,1))
dgp <- dbeta(X,5,1)
model.g1 <- npuniden.boundary(X,kertype="gaussian1")
model.g2 <- npuniden.boundary(X,kertype="gaussian2")
model.b1 <- npuniden.boundary(X,kertype="beta1")
model.b2 <- npuniden.boundary(X,kertype="beta2")
model.fb <- npuniden.boundary(X,kertype="fb")
model.unadjusted <- npuniden.boundary(X,a=-Inf,b=Inf)
ylim <- c(0,max(c(dgp,model.g1$f,model.g2$f,model.b1$f,model.b2$f,model.fb$f)))
if (interactive()) plot(X,dgp,ylab="Density",ylim=ylim,type="l")
lines(X,model.g1$f,lty=2,col=2)
lines(X,model.g2$f,lty=3,col=3)
lines(X,model.b1$f,lty=4,col=4)
lines(X,model.b2$f,lty=5,col=5)
lines(X,model.fb$f,lty=6,col=6)
lines(X,model.unadjusted$f,lty=7,col=7)
rug(X)
legend("topleft",c("DGP",
                   "Boundary Kernel (gaussian1)",
                   "Boundary Kernel (gaussian2)",
                   "Boundary Kernel (beta1)",
                   "Boundary Kernel (beta2)",
                   "Boundary Kernel (floating boundary)",
                   "Unadjusted"),col=1:7,lty=1:7,bty="n")

## Example 2: f(0)=0, f(1)=0, plot density, distribution, DGP, and
## asymptotic point-wise confidence intervals
set.seed(42)
X <- sort(rbeta(100,5,3))
model <- npuniden.boundary(X)
oldpar <- par(no.readonly = TRUE)
on.exit(par(oldpar), add = TRUE)
par(mfrow=c(1,2))
ylim=range(c(model$f,model$f+1.96*model$sd.f,model$f-1.96*model$sd.f,dbeta(X,5,3)))
if (interactive()) plot(X,model$f,ylim=ylim,ylab="Density",type="l",)
lines(X,model$f+1.96*model$sd.f,lty=2)
lines(X,model$f-1.96*model$sd.f,lty=2)
lines(X,dbeta(X,5,3),col=2)
rug(X)
legend("topleft",c("Density","DGP"),lty=c(1,1),col=1:2,bty="n")

if (interactive()) plot(X,model$F,ylab="Distribution",type="l")
lines(X,model$F+1.96*model$sd.F,lty=2)
lines(X,model$F-1.96*model$sd.F,lty=2)
lines(X,pbeta(X,5,3),col=2)
rug(X)
legend("topleft",c("Distribution","DGP"),lty=c(1,1),col=1:2,bty="n")

## Example 3: Age for working age males in the cps71 data set bounded
## below by 21 and above by 65
data(cps71)
with(cps71, {
model <- npuniden.boundary(age,a=21,b=65)
par(mfrow=c(1,1))
hist(age,prob=TRUE,main="")
lines(age,model$f)
lines(density(age,bw=model$h),col=2)
legend("topright",c("Boundary Kernel","Unadjusted"),lty=c(1,1),col=1:2,bty="n")
})
} % enddontrun
}
\keyword{ nonparametric }
\keyword{ smooth }

\name{npcopula}
\alias{npcopula}
\alias{npcopula.default}
\alias{npcopula.formula}
\alias{plot.npcopula}
\alias{print.npcopula}
\alias{summary.npcopula}
\alias{fitted.npcopula}
\alias{predict.npcopula}
\alias{se.npcopula}
\alias{as.data.frame.npcopula}
\title{Kernel Copula Estimation with Mixed Data Types}

\description{
  \code{npcopula} estimates a mixed-data kernel copula distribution or
  copula density. It can be called with an existing unconditional
  distribution/density bandwidth object, or with a one-sided formula in
  which case the appropriate bandwidth object is selected internally.
}

\usage{
npcopula(bws, \dots)

\method{npcopula}{formula}(bws,
         data = NULL,
         u = NULL,
         target = c("distribution", "density"),
         evaluation = c("grid", "sample"),
         neval = 30,
         n.quasi.inv = 1000,
         er.quasi.inv = 1,
         \dots)

\method{npcopula}{default}(bws,
         data,
         u = NULL,
         target = NULL,
         evaluation = c("sample", "grid"),
         neval = 30,
         n.quasi.inv = 1000,
         er.quasi.inv = 1,
         \dots)

\method{predict}{npcopula}(object,
        newdata = NULL,
        u = NULL,
        se.fit = FALSE,
        output = c("vector", "object", "data"),
        \dots)

\method{plot}{npcopula}(x,
     perspective = TRUE,
     view = c("rotate", "fixed", "surface", "contour", "image",
              "empirical", "all"),
     renderer = c("base", "rgl"),
     errors = c("none", "bootstrap", "asymptotic"),
     band = c("pointwise", "pmzsd", "bonferroni",
              "simultaneous", "all"),
     alpha = 0.05,
     bootstrap = c("inid", "fixed", "geom"),
     B = 1999,
     center = c("estimate", "bias-corrected"),
     boot_control = np_boot_control(),
     output = c("plot", "data", "plot-data", "both"),
     legend = TRUE,
     theta = 0.0,
     phi = 20.0,
     xlab = "u1",
     ylab = "u2",
     zlab = NULL,
     main = NULL,
     col = NULL,
     border = "black",
     zlim = NULL,
     \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the bandwidth specification, source data, and
formula route.
}

  \item{bws}{
    bandwidth specification or one-sided formula. A bandwidth
    specification is either an unconditional distribution bandwidth
    object returned by \code{\link{npudistbw}}, for a copula
    distribution, or an unconditional density bandwidth object returned
    by \code{\link{npudensbw}}, for a copula density. If \code{bws} is a
    formula such as \code{~ x + y}, \code{npcopula} first calls
    \code{\link{npudistbw}} when \code{target="distribution"} and
    \code{\link{npudensbw}} when \code{target="density"}; additional
    bandwidth-selection arguments in \code{\dots} are forwarded to that
    bandwidth selector.
  }

  \item{data}{
    data frame containing the variables used to construct \code{bws};
    when \code{bws} is a formula, \code{data} is passed to the
    bandwidth selector. Copulas are defined here for numeric and ordered
    variables; unordered factors are rejected.
  }

\subsection{Copula Target And Evaluation Grid}{
These arguments control whether a copula distribution or density is
estimated and where it is evaluated.
}

  \item{er.quasi.inv}{
    fraction passed to \code{\link{extendrange}} when constructing the
    marginal quasi-inverse grid used for supplied probability values
    \code{u}. See Details.
  }

  \item{evaluation}{
    evaluation route used when \code{bws} is a formula. The default
    \code{"grid"} constructs a plot-ready two-dimensional probability
    grid when \code{u} is omitted. Use \code{"sample"} to evaluate at
    the sample realizations.
  }

  \item{n.quasi.inv}{
    number of grid points used to compute each marginal quasi-inverse
    when \code{u} is supplied or automatically generated.
  }

  \item{newdata}{
    optional prediction data for \code{predict.npcopula}. This is a
    compatibility alias for \code{u}: it should contain marginal
    probability values in \eqn{[0,1]}, either in columns named as the
    original variables or in columns named \code{u1}, \code{u2}, ....
    If both \code{u} and \code{newdata} are supplied, \code{u} takes
    precedence.
  }

  \item{neval}{
    number of probability values per margin in the automatically
    generated two-dimensional grid used when \code{bws} is a formula,
    \code{evaluation="grid"}, and \code{u} is omitted.
  }

  \item{target}{
    target used when \code{bws} is a formula. The default
    \code{"distribution"} estimates a copula distribution using
    \code{\link{npudistbw}}; \code{"density"} estimates a copula
    density using \code{\link{npudensbw}}. When \code{bws} is already a
    bandwidth object, the target is inferred from that object and a
    conflicting explicit \code{target} is rejected.
  }

  \item{u}{
    optional matrix or data frame of marginal probability values in
    \eqn{[0,1]}. Each column corresponds to one variable in the copula.
    If supplied, \code{npcopula} evaluates on the Cartesian product
    \code{expand.grid(u)}. For two-dimensional displays, the clearest
    spelling is often \code{data.frame(u1 = u1.seq, u2 = u2.seq)};
    names matching the original variables are also accepted for
    compatibility. If omitted with a formula route and
    \code{evaluation="grid"}, a two-dimensional grid is generated
    automatically.
  }

  \item{object}{
    an object of class \code{"npcopula"} returned by \code{npcopula}.
  }

  \item{se.fit}{
    logical value. If \code{TRUE}, \code{predict.npcopula} returns a
    list with fitted values and the stored standard-error slot.
  }

\subsection{Plot Display Controls}{
These arguments control how a two-dimensional grid copula object is
displayed.
}

  \item{x}{
    an object of class \code{"npcopula"} returned by \code{npcopula}.
  }

  \item{border}{
    border color for surface facets and interval wireframes when
    \code{view="surface"}.
  }

  \item{col}{
    optional surface or image colors. If omitted, \code{plot.npcopula}
    uses the same viridis \code{\link{hcl.colors}} palette used by the
    other modern surface plot methods.
  }

  \item{legend}{
    logical or legend-control value used for \code{band="all"} interval
    overlays. Use \code{FALSE}, \code{NULL}, or \code{NA} to suppress
    the interval legend.
  }

  \item{main, xlab, ylab, zlab}{
    plot titles and axis labels.
  }

  \item{output}{
    return mode. For \code{plot.npcopula}, \code{"plot"} draws the
    plot, \code{"data"} returns the plotted data with interval columns
    when requested, and \code{"plot-data"} draws and returns the
    plotted data. \code{"both"} is accepted as an alias for
    \code{"plot-data"}. For \code{predict.npcopula}, \code{"vector"}
    returns fitted copula values, \code{"object"} returns the evaluated
    \code{"npcopula"} object, and \code{"data"} returns
    \code{as.data.frame()} on that object.
  }

  \item{perspective}{
    logical value. If \code{TRUE}, draw a surface display; otherwise
    use \code{view} to choose \code{"contour"} or \code{"image"}.
    The \code{"empirical"} and \code{"all"} views ignore this argument.
  }

  \item{phi, theta}{
    viewing angles passed to \code{\link{persp}} or to the shared
    \pkg{rgl} surface renderer. The defaults match the package-wide
    surface-plot defaults used by the other perspective plot methods.
    As with those methods, the exact default pair \code{theta = 0} and
    \code{phi = 20} is remapped internally for \code{renderer="rgl"} to
    account for the different viewing-angle convention used by \pkg{rgl};
    explicitly supplied non-default angles are passed through.
  }

  \item{renderer}{
    plotting renderer for surface displays. \code{"base"} uses
    \code{\link{persp}}. \code{"rgl"} uses the shared interactive
    \pkg{rgl} surface renderer when the suggested package \pkg{rgl} is
    installed.
  }

  \item{view}{
    display type for grid output. The default \code{"rotate"} draws a
    rotating base \code{\link{persp}} surface using the same frame step
    and delay as the other package perspective plots. Use
    \code{"fixed"} for a single fixed surface, \code{"surface"} as a
    backward-compatible fixed-surface alias, or \code{"contour"} and
    \code{"image"} with \code{perspective=FALSE}. Use
    \code{"empirical"} to plot the empirical copula coordinates. Use
    \code{"all"} for a base-graphics four-panel display containing the
    copula contour, copula surface, empirical copula coordinates, and a
    copula-density surface. The \code{"all"} view is not currently
    supported with \code{renderer="rgl"}.
  }

  \item{zlim}{
    optional z-axis limits for surface displays.
  }

\subsection{Plot Interval Controls}{
These arguments add asymptotic or bootstrap intervals to two-dimensional
surface plots.
}

  \item{alpha}{
    nominal size used for asymptotic or bootstrap intervals when
    \code{errors!="none"}.
  }

  \item{B}{
    number of bootstrap replications when \code{errors="bootstrap"}.
  }

  \item{band}{
    interval type for plotted surfaces. Supported values are
    \code{"pointwise"}, \code{"pmzsd"}, \code{"bonferroni"},
    \code{"simultaneous"}, and \code{"all"}. The \code{"all"} option
    overlays pointwise, simultaneous, and Bonferroni wireframes where
    available.
  }

  \item{bootstrap}{
    bootstrap resampling method used when \code{errors="bootstrap"};
    supported values are \code{"inid"}, \code{"fixed"}, and
    \code{"geom"}. Wild bootstrap is not defined for copula surfaces.
  }

  \item{boot_control}{
    optional \code{\link{np_boot_control}} object. For copula surfaces
    the block length is used by the \code{"fixed"} and \code{"geom"}
    block bootstrap routes.
  }

  \item{center}{
    centering convention for bootstrap intervals, either
    \code{"estimate"} or \code{"bias-corrected"}.
  }

  \item{errors}{
    interval route for \code{plot.npcopula}: \code{"none"},
    \code{"asymptotic"}, or \code{"bootstrap"}. Intervals are available
    for two-dimensional grid evaluation output and are drawn as
    transparent wireframes over the copula surface.
  }

\subsection{Additional Arguments}{
Further arguments are passed to the bandwidth-selection counterpart,
prediction/evaluation route, or graphics renderer as appropriate.
}

  \item{\dots}{
    additional arguments supplied to \code{\link{npudistbw}} or
    \code{\link{npudensbw}} when \code{npcopula} computes bandwidths
    internally, or arguments needed to interpret a numeric \code{bws}
    vector. This is where bandwidth-selection controls such as
    \code{bwmethod}, \code{bwtype}, and \code{bwscaling},
    kernel/support controls such as \code{ckertype}, \code{ckerorder},
    and \code{ckerbound}, categorical kernel controls such as
    \code{ukertype} and \code{okertype}, and search controls such as
    \code{nmulti} and \code{scale.factor.search.lower} are supplied.
    In \code{predict.npcopula}, additional arguments are passed to
    \code{npcopula} for evaluation with the stored bandwidth object and
    training data. In \code{plot.npcopula}, additional arguments are
    passed to the selected graphics routine, such as
    \code{\link{persp}}, \code{\link{contour}}, \code{\link{image}}, or
    the shared \pkg{rgl} renderer.
  }
}

\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels,
  \code{\link{np.options}} for global options, and \code{\link{plot}}
  for plotting options.

  \code{npcopula} computes the nonparametric copula distribution or
  copula density using marginal quasi-inversion. For the distribution
  target, Sklar's theorem gives
  \deqn{
    C(u_1,\ldots,u_d)
      =
    H(F_1^{-1}(u_1),\ldots,F_d^{-1}(u_d)),
  }
  where \eqn{H} is the joint distribution and \eqn{F_j^{-1}} is the
  quasi-inverse of marginal distribution \eqn{F_j}. For the density
  target, the estimated copula density is
  \deqn{
    c(u)
      =
    \frac{f(x_u)}
         {\prod_{j=1}^d f_j(x_{u,j})},
    \quad x_{u,j}=F_j^{-1}(u_j),
  }
  with numerator and marginal denominators estimated using the selected
  mixed-data kernel bandwidths.

  If \code{u} is provided, \code{\link{expand.grid}} is called on
  \code{u}. As the dimension increases this can become unwieldy because
  a grid with \eqn{m} points in each of \eqn{d} margins has
  \eqn{m^d} rows. Therefore the formula route automatically generates a
  probability grid only for two-dimensional copulas. For
  higher-dimensional copulas, supply \code{u} explicitly or use
  \code{evaluation="sample"}.

  The \sQuote{quasi-inverse} is computed via Definition 2.3.6 from
  Nelsen (2006). An equi-quantile grid on the data range of length
  \code{n.quasi.inv/2} is combined with an equi-spaced grid on the data
  range extended by \code{er.quasi.inv}; the sorted union forms the
  grid used for marginal inversion. If requested probability values lie
  outside the attainable estimated marginal distribution range, they
  are reset to the nearest attainable endpoint. Inspect the returned
  \code{u} columns when endpoint behavior matters.

  The \code{plot.npcopula} method supports base \code{\link{persp}},
  \code{\link{contour}}, and \code{\link{image}} displays for
  two-dimensional grid output. Surface plots use the package-wide
  viridis default palette, detailed perspective ticks, and the same
  default viewing angles and base-graphics rotation cadence as the
  other surface plot methods. \code{renderer="rgl"} requests the shared
  interactive \pkg{rgl} surface renderer, using the same default-angle
  remapping used by the other package surface plots. For mixed ordered
  margins, grid displays are drawn against the requested probability
  grid, while the returned \code{u} columns retain the attainable
  marginal probability values produced by quasi-inversion. The
  \code{"empirical"} view plots empirical copula coordinates, and
  \code{"all"} gives a base-graphics four-panel diagnostic display.

  For grid surfaces, \code{plot.npcopula} can add asymptotic or
  bootstrap intervals. Distribution-copula asymptotic intervals use the
  joint distribution standard error evaluated at the marginal
  quasi-inverse grid. Density-copula asymptotic intervals use the
  plug-in delta-method denominator correction corresponding to
  \eqn{c(u)=f(x_u)/\prod_j f_j(x_{u,j})}. Bootstrap intervals resample
  rows and recompute the plotted copula surface on the same probability
  grid; \code{band="all"} overlays transparent pointwise,
  simultaneous, and Bonferroni wireframes.
}

\value{
  \code{npcopula} returns an object of class \code{"npcopula"}. The main
  components are:

  \item{copula}{
    estimated copula distribution value or copula density value.
  }

  \item{u1, u2, ...}{
    marginal probability coordinates associated with the sample
    realizations or evaluation grid.
  }

  \item{x, y, ...}{
    marginal quasi-inverse coordinates corresponding to the requested
    probability grid when grid evaluation is used.
  }

  \item{bws}{
    selected unconditional distribution or density bandwidth object.
  }

  \item{eval}{
    data frame containing the copula values, probability coordinates, and
    quasi-inverse coordinates. \code{as.data.frame(object)} returns this
    component for data-frame workflows.
  }

  \item{copulaerr}{
    asymptotic or bootstrap standard-error slot. The fitted object stores
    \code{NA} unless an interval-producing plotting route constructs
    evaluation-specific intervals.
  }

  The source data, target, evaluation route, grid dimensions, and timing
  metadata are retained as list components. The functions
  \code{\link{fitted}}, \code{\link{predict}}, \code{\link{se}},
  \code{\link{summary}}, \code{\link{as.data.frame}}, and
  \code{\link{plot}} support \code{"npcopula"} objects.
}

\section{Book And Method Pointers}{
  The copula distribution target is
  \eqn{C(u)=H(F_1^{-1}(u_1),\ldots,F_d^{-1}(u_d))}; the copula density
  target is \eqn{c(u)=f(x_u)/\prod_j f_j(x_{u,j})}. The mixed-data
  kernel implementation follows Racine (2015), with quasi-inversion in
  the sense of Nelsen (2006). For the underlying mixed-data density and
  distribution estimators, see Li and Racine (2007), Chapter 1
  \emph{Density Estimation}, Chapter 3 \emph{Kernel Estimation with
  Mixed Data}, and Racine (2019), Chapter 2 \emph{Continuous Density
  and Cumulative Distribution Functions}.
}

\references{
  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics:
  Theory and Practice}, Princeton University Press.

  Nelsen, R.B. (2006), \emph{An Introduction to Copulas}, Second
  Edition, Springer.

  Racine, J.S. (2015), \dQuote{Mixed Data Kernel Copulas}, \emph{Empirical
  Economics}, 48, 37--59.

  Racine, J.S. (2019), \emph{An Introduction to the Advanced Theory and
  Practice of Nonparametric Econometrics: A Replicable Approach Using R}.
}

\author{
  Jeffrey S. Racine \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  Use a \code{\link{data.frame}} rather than \code{\link{cbind}} for
  mixed data so that ordered variables remain ordered. Unordered
  factors are not valid for copula estimation in this implementation.
}

\seealso{
  \code{\link{npudistbw}}, \code{\link{npudist}},
  \code{\link{npudensbw}}, \code{\link{npudens}},
  \code{\link{np.kernels}}, \code{\link{np.options}},
  \code{\link{plot}}
}

\examples{
\dontrun{
library("MASS")

## Example 1: bivariate mixed data, continuous x and ordered y.

set.seed(42)
n <- 1000
n.eval <- 30
rho <- 0.99
mu <- c(0, 0)
Sigma <- matrix(c(1, rho, rho, 1), 2, 2)
xy <- mvrnorm(n = n, mu = mu, Sigma = Sigma)
mydat <- data.frame(
  x = xy[, 1],
  y = ordered(as.integer(cut(xy[, 2],
    quantile(xy[, 2], seq(0, 1, by = .1)),
    include.lowest = TRUE)) - 1)
)

grid.seq <- seq(0, 1, length.out = n.eval)
grid.dat <- data.frame(u1 = grid.seq, u2 = grid.seq)

## Estimate the copula distribution from an npudistbw() object.

bw.cdf <- npudistbw(~ x + y, data = mydat, nmulti = 1)
copula <- npcopula(bws = bw.cdf, data = mydat, u = grid.dat)
summary(copula)

## Native plotting replaces the older manual contour(), persp(), and
## empirical scatterplot calls.

plot(copula, perspective = FALSE, view = "contour")
plot(copula, perspective = FALSE, view = "image")
plot(copula, view = "fixed", zlim = c(0, 1))
if (requireNamespace("rgl", quietly = TRUE))
  plot(copula, view = "fixed", renderer = "rgl", zlim = c(0, 1))
plot(copula)

## Plot empirical copula coordinates from the retained sample data.

plot(copula, view = "empirical")

## Or request the four-panel base-graphics diagnostic display.

plot(copula, view = "all")

## Estimate and plot the copula density from an npudensbw() object.

bw.pdf <- npudensbw(~ x + y, data = mydat, nmulti = 1)
copula.dens <- npcopula(bws = bw.pdf, data = mydat, u = grid.dat)
summary(copula.dens)
plot(copula.dens, view = "fixed")
if (requireNamespace("rgl", quietly = TRUE))
  plot(copula.dens, view = "fixed", renderer = "rgl")
plot(copula.dens)

## Intervals are available for two-dimensional grid surfaces.

plot(copula, errors = "asymptotic", band = "pointwise")
plot(copula, errors = "bootstrap", bootstrap = "inid", B = 399,
     band = "pointwise")

## Prediction evaluates the retained bandwidth object on a supplied
## probability grid.

predict(copula, u = data.frame(x = c(0.25, 0.75),
                               y = c(0.25, 0.75)))
predict(copula, newdata = data.frame(u1 = c(0.25, 0.75),
                                     u2 = c(0.25, 0.75)))

## Example 2: bivariate continuous data.

set.seed(42)
n <- 1000
n.eval <- 30
rho <- 0.99
mu <- c(0, 0)
Sigma <- matrix(c(1, rho, rho, 1), 2, 2)
xy <- mvrnorm(n = n, mu = mu, Sigma = Sigma)
mydat <- data.frame(x = xy[, 1], y = xy[, 2])

grid.seq <- seq(0, 1, length.out = n.eval)
grid.dat <- data.frame(u1 = grid.seq, u2 = grid.seq)

bw.cdf <- npudistbw(~ x + y, data = mydat, nmulti = 1)
copula <- npcopula(bws = bw.cdf, data = mydat, u = grid.dat)
summary(copula)

plot(copula, perspective = FALSE, view = "contour")
plot(copula, perspective = FALSE, view = "image")
plot(copula, view = "fixed", zlim = c(0, 1))
if (requireNamespace("rgl", quietly = TRUE))
  plot(copula, view = "fixed", renderer = "rgl", zlim = c(0, 1))
plot(copula)
plot(copula, view = "empirical")
plot(copula, view = "all")

bw.pdf <- npudensbw(~ x + y, data = mydat, nmulti = 1)
copula.dens <- npcopula(bws = bw.pdf, data = mydat, u = grid.dat)
summary(copula.dens)
plot(copula.dens, view = "fixed", zlim = c(0, 40))
if (requireNamespace("rgl", quietly = TRUE))
  plot(copula.dens, view = "fixed", renderer = "rgl",
       zlim = c(0, 40))
plot(copula.dens, zlim = c(0, 40))

## The formula interface is a shorter route when bandwidths do not need
## to be reused explicitly.

copula.short <- npcopula(~ x + y, data = mydat, neval = n.eval,
                         nmulti = 1)
plot(copula.short, view = "all")
} % enddontrun
}

\keyword{nonparametric}

%
\name{npplreg}
\alias{npplreg}
\alias{npplreg.default}
\alias{npplreg.formula}
\alias{npplreg.plbandwidth}

\title{Partially Linear Kernel Regression with Mixed Data Types}

\description{
  \code{npplreg} computes a partially linear kernel regression estimate
  of a one (1) dimensional dependent variable on \eqn{p+q}-variate
  explanatory data, using the model \eqn{Y = X\beta + \Theta (Z) +
  \epsilon}{Y = XB + theta(Z) + epsilon} given a set of estimation
  points, training points (consisting of explanatory data and dependent
  data), and a bandwidth specification, which can be a \code{rbandwidth}
  object, or a bandwidth vector, bandwidth type and kernel type.
}

\usage{
npplreg(bws, 
        \dots)

\method{npplreg}{formula}(bws, 
       data = NULL, 
       newdata = NULL, 
       y.eval = FALSE, 
       \dots)

\method{npplreg}{default}(bws,
        txdat,
        tydat,
        tzdat,
        nomad = FALSE,
        \dots)

\method{npplreg}{plbandwidth}(bws,
        txdat = stop("training data txdat missing"),
        tydat = stop("training data tydat missing"),
        tzdat = stop("training data tzdat missing"),
        exdat,
        eydat,
        ezdat,
        residuals = FALSE,
        \dots)
}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the bandwidth specification, formula/data interface, and partially linear training data.
}

  \item{bws}{
  a bandwidth specification. This can be set as a \code{plbandwidth}
  object returned from an invocation of \code{\link{npplregbw}}, or as
  a matrix of bandwidths, where each row is a set of bandwidths for \eqn{Z},
  with a column for each variable \eqn{Z_i}{Z[i]}. In the first row are
  the bandwidths for the regression of \eqn{Y} on \eqn{Z}, the following
  rows contain the bandwidths for the regressions of the columns of
  \eqn{X} on \eqn{Z}.  If specified as a matrix
  additional arguments will need to be supplied as necessary to
  specify the bandwidth type, kernel types, training data, and so on.

  }



  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(bws)}, typically the environment from which
   \code{\link{npplregbw}} was called.
  }


  \item{txdat}{
  a \eqn{p}-variate data frame of explanatory data (training data),
  corresponding to \eqn{X} in the model equation, whose linear
  relationship with the dependent data \eqn{Y} is posited. Defaults to
  the training data used to
    compute the bandwidth object.
  }


  \item{tydat}{
  a one (1) dimensional numeric or integer vector of dependent data, each
  element \eqn{i} corresponding to each observation (row) \eqn{i} of
  \code{txdat}. Defaults to the training data used to
    compute the bandwidth object.
  }


  \item{tzdat}{
  a \eqn{q}-variate data frame of explanatory data (training data),
  corresponding to \eqn{Z} in the model equation, whose relationship
  to the dependent variable is unspecified (nonparametric). Defaults to
  the training data used to
    compute the bandwidth object.
  }


\subsection{Local-Polynomial Degree And Bandwidth Search}{
This argument controls the recommended automatic local-polynomial NOMAD route, which jointly selects continuous polynomial degree and bandwidths when these are computed inside \code{npplreg}.
}

  \item{nomad}{
    logical shortcut passed through to \code{\link{npplregbw}} when
    bandwidths are computed inside \code{npplreg}. When \code{TRUE},
    the partially linear bandwidth route fills any missing values among
    \code{regtype}, \code{search.engine}, \code{degree.select},
    \code{bernstein.basis}, \code{degree.min}, \code{degree.max},
    \code{degree.verify}, and \code{bwtype} with the recommended
    automatic local-polynomial degree-and-bandwidth NOMAD preset documented in \code{\link{npplregbw}}.
    The preset selects local-polynomial degrees and bandwidths
    separately for the nuisance regressions \eqn{E[Y \mid Z]} and
    \eqn{E[X_j \mid Z]} before the final partially linear solve.
    Additional NOMAD tuning arguments such as \code{nomad.nmulti} may
    also be supplied through \code{\dots}; \code{nmulti} remains the
    outer restart count while \code{nomad.nmulti} controls inner
    \code{crs::snomadr()} multistarts within each outer restart.
    After fitting, inspect \code{fit$bws$nomad.shortcut} on the
    returned object \code{fit} to see the normalized shortcut metadata.
  }


\subsection{Evaluation Data And Returned Quantities}{
These arguments control where the partially linear regression is evaluated and which fitted quantities are returned.
}

  \item{exdat}{
  a \eqn{p}-variate data frame of points on which the regression will be
  estimated (evaluation data).  By default,
    evaluation takes place on the data provided by \code{txdat}.
  }


  \item{eydat}{
  a one (1) dimensional numeric or integer vector of the true values
  of the dependent variable. Optional, and used only to calculate the
  true errors.  By default,
    evaluation takes place on the data provided by \code{tydat}.
  }


  \item{ezdat}{
    a \eqn{q}-variate data frame of points on which the regression will
    be estimated (evaluation data).  By default,
    evaluation takes place on the data provided by \code{tzdat}.
  }


  \item{newdata}{
    An optional data frame in which to look for evaluation data. If
    omitted, the training data are used.
  }


  \item{residuals}{
    a logical value indicating that you want residuals computed and
    returned in the resulting \code{plregression} object. Defaults to
    \code{FALSE}.
  }


  \item{y.eval}{
    If \code{newdata} contains dependent data and \code{y.eval = TRUE},
    \code{\link{np}} will compute goodness of fit statistics on these
    data and return them. Defaults to \code{FALSE}.
  }


\subsection{Additional Arguments}{
Further arguments are passed to \code{\link{npplregbw}} and its component \code{\link{npregbw}} searches when bandwidths are computed internally.
}

  \item{\dots}{
    additional arguments supplied to \code{\link{npplregbw}} when
    \code{npplreg} computes bandwidths internally, or arguments needed
    to interpret a numeric or matrix \code{bws} specification. This is
    where bandwidth selection controls such as \code{bwmethod},
    \code{bwtype}, and \code{bwscaling}, kernel/support controls such
    as \code{ckertype}, \code{ckerorder}, and \code{ckerbound},
    categorical kernel controls such as \code{ukertype} and
    \code{okertype}, search controls such as \code{nmulti} and
    \code{scale.factor.search.lower}, and local-polynomial/NOMAD
    controls such as \code{regtype}, \code{degree},
    \code{bernstein.basis}, \code{degree.select}, and
    \code{nomad.nmulti} are supplied. See \code{\link{npplregbw}} and
    \code{\link{npregbw}} for the complete bandwidth-selection
    argument surface, including the child-indexed matrix/list
    \code{degree} form for fixed local-polynomial partially linear
    fits.
  }
}
\details{
  Documentation guide: see \code{\link{npplregbw}} for partially linear
  bandwidth selection, \code{\link{npregbw}} for the component
  nonparametric regression search controls, \code{\link{np.kernels}}
  for kernels, \code{\link{np.options}} for global options, and
  \code{\link{plot}}, \code{\link{plot.np}} for plotting options.

  When \code{bws} is omitted, the formula and default methods call
  \code{\link{npplregbw}} first and pass bandwidth-selection arguments
  from \code{\dots} to that call. When \code{bws} is already a
  \code{plbandwidth} object, \code{npplreg} estimates with the stored
  bandwidth metadata in that object.

  Argument groups for bandwidth selection are documented on
  \code{\link{npplregbw}} and, for the component nonparametric
  regressions, \code{\link{npregbw}}. The most common workflow is to
  choose the linear \code{X} variables and nonparametric \code{Z}
  variables first, then bandwidth/search controls for the
  \code{Z}-side nonparametric regressions, and finally
  local-polynomial/NOMAD controls when using polynomial-adaptive fits.

  For S3 plotting help, see \code{\link{plot.np}}. You can list
  available plot methods with \code{methods("plot")}.

 \code{npplreg} uses a combination of OLS and nonparametric
 regression to estimate the parameter \eqn{\beta}{B} in the model
 \eqn{Y = X\beta + \Theta (Z) + \epsilon}{Y = XB + theta(Z) + epsilon}.

 \code{npplreg} implements a variety of methods for
 nonparametric regression on multivariate (\eqn{q}-variate) explanatory
 data defined over a set of possibly continuous and/or discrete
 (unordered, ordered) data. The approach is based on Li and Racine
 (2003) who employ \sQuote{generalized product kernels} that admit a mix
 of continuous and discrete data types.

 Three classes of kernel estimators for the continuous data types are
 available: fixed, adaptive nearest-neighbor, and generalized
 nearest-neighbor. Adaptive nearest-neighbor bandwidths change with
 each sample realization in the set, \eqn{x_i}{x[i]}, when estimating the
 density at the point \eqn{x}. Generalized nearest-neighbor bandwidths change
 with the point at which the density is estimated, \eqn{x}. Fixed bandwidths
 are constant over the support of \eqn{x}.

 Data contained in the data frame \code{tzdat} may be a mix of
 continuous (default), unordered discrete (to be specified in the data
 frame \code{tzdat} using \code{\link{factor}}), and ordered discrete
 (to be specified in the data frame \code{tzdat} using
 \code{\link{ordered}}). Data can be entered in an arbitrary order and
 data types will be detected automatically by the routine (see
 \code{\link{np}} for details).

 A variety of kernels may be specified by the user. Kernels implemented
 for continuous data types include the second, fourth, sixth, and eighth
 order Gaussian and Epanechnikov kernels, and the uniform
 kernel. Unordered discrete data types use a variation on Aitchison and
 Aitken's (1976) kernel, while ordered data types use a variation of the
 Wang and van Ryzin (1981) kernel.

 For practitioners who want the recommended automatic local-polynomial degree-and-bandwidth NOMAD route
 without spelling out all LP tuning arguments,
 \code{npplreg(..., nomad=TRUE)} and \code{npplregbw(..., nomad=TRUE)}
 expand missing settings to the same documented preset. Explicit
 incompatible settings fail fast rather than being silently rewritten.
}

\value{
  \code{npplreg} returns a \code{plregression} object. The generic
  accessor functions \code{\link{coef}}, \code{\link{fitted}},
  \code{\link{residuals}}, \code{\link{predict}}, and
  \code{\link{vcov}}, extract (or
  estimate) coefficients, estimated values, residuals,
  predictions, and variance-covariance matrices,
  respectively, from
  the returned object. Furthermore, the functions \code{summary}
  and \code{plot} support objects of this type. The returned object
  has the following components:

  For \code{plregression} objects, \code{predict(object, se.fit = TRUE)}
  returns a list with components \code{fit} and \code{se.fit}. The
  \code{se.fit} component contains asymptotic prediction standard errors
  computed from the partially linear fitted-value linear operator and
  the squared residuals from the training fit.

  \item{evalx}{ evaluation points }
  \item{evalz}{ evaluation points }
  \item{mean}{ estimation of the regression, or conditional mean, at the
    evaluation points }
  \item{xcoef}{ coefficient(s) corresponding to the components
    \eqn{\beta_i}{B[i]} in the model }
  \item{xcoeferr}{ standard errors of the coefficients }
  \item{xcoefvcov}{ covariance matrix of the coefficients }
  \item{bws}{ the canonical bandwidth object, stored as a
    \code{plbandwidth} object }
  \item{bw}{ backward-compatible alias for \code{bws} }
  \item{resid}{ if \code{residuals = TRUE}, in-sample or out-of-sample
    residuals where appropriate (or possible) }
  \item{R2}{ coefficient of determination (Doksum and Samarov (1995))}
  \item{MSE}{ mean squared error }
  \item{MAE}{ mean absolute error }
  \item{MAPE}{ mean absolute percentage error }
  \item{CORR}{ absolute value of Pearson's correlation coefficient }
  \item{SIGN}{ fraction of observations where fitted and observed values
    agree in sign }

}

\section{Book And Method Pointers}{
  The partially linear model is
  \eqn{Y=X\beta+\Theta(Z)+\epsilon}. The Robinson-style construction
  residualizes \eqn{Y} and each column of \eqn{X} on \eqn{Z}, then
  solves the final least-squares problem using those residualized
  variables. Thus the nonparametric part of the problem is a sequence
  of conditional-mean regressions \eqn{E[Y\mid Z]} and
  \eqn{E[X_j\mid Z]}, followed by the parametric solve for
  \eqn{\beta}. In local-polynomial/NOMAD fits, the nuisance regressions
  may select child-specific continuous polynomial degrees.

  For book-length derivations, see Li and Racine (2007), Chapter 7
  \emph{Semiparametric Partially Linear Models}, especially Sections 7.1,
  7.2, and 7.4, and Racine (2019), Chapter 8 \emph{Semiparametric
  Conditional Mean Function Estimation}, especially Robinson's
  partially linear model.
}

\references{

 Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
   discrimination by the kernel method,} Biometrika, 63, 413-420.

 Doksum, K. and A. Samarov (1995), \dQuote{Nonparametric estimation of
 global functionals and a measure of the explanatory power of covariates
 in regression,} The Annals of Statistics, 23 1443-1473.

 Gao, Q. and L. Liu and J.S. Racine (2015), \dQuote{A partially linear
 kernel estimator for categorical data,} Econometric Reviews, 34 (6-10),
 958-977.

 Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
 and Practice,} Princeton University Press.

 Li, Q. and J.S. Racine (2004), \dQuote{Cross-validated local linear
   nonparametric regression,} Statistica Sinica, 14, 485-512.

 Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,}
 Cambridge University Press.

 Racine, J.S. and Q. Li (2004), \dQuote{Nonparametric estimation of regression
 functions with both categorical and continuous data,} Journal of
 Econometrics, 119, 99-130.

 Robinson, P.M. (1988), \dQuote{Root-n-consistent semiparametric regression,}
 Econometrica, 56, 931-954.

 Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth estimators
 for discrete distributions,}  Biometrika, 68, 301-309.

}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}, \code{\link{plot.np}}
  \code{\link{npregbw}}, \code{\link{npreg}} }
\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): For this example, we simulate an
# example for a partially linear model and compare the coefficient
# estimates from the partially linear model with those from a correctly
# specified parametric model...

set.seed(42)

n <- 250
x1 <- rnorm(n)
x2 <- rbinom(n, 1, .5)

z1 <- rbinom(n, 1, .5)
z2 <- rnorm(n)

y <- 1 + x1 + x2 + z1 + sin(z2) + rnorm(n)

# First, compute data-driven bandwidths. This may take a few minutes
# depending on the speed of your computer...

bw <- npplregbw(formula=y~x1+factor(x2)|factor(z1)+z2)

# Next, compute the partially linear fit

pl <- npplreg(bws=bw)

# Print a summary of the model...

summary(pl)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Retrieve the coefficient estimates and their standard errors...

coef(pl)
coef(pl, errors = TRUE)

# Compare the partially linear results to those from a correctly
# specified model's coefficients for x1 and x2

ols <- lm(y~x1+factor(x2)+factor(z1)+I(sin(z2)))

# The intercept is coef()[1], and those for x1 and x2 are coef()[2] and
# coef()[3]. The standard errors are the square root of the diagonal of
# the variance-covariance matrix (elements 2 and 3)

coef(ols)[2:3]
sqrt(diag(vcov(ols)))[2:3]

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Plot the regression surfaces via plot() (i.e., plot the `partial
# regression surface plots').

if (interactive()) plot(bw)

# Note - to plot regression surfaces with variability bounds constructed
# from bootstrapped standard errors, use the following (note that this
# may take a minute or two depending on the speed of your computer as
# the bootstrapping is done in real time, and note also that we override
# the default number of bootstrap replications (399) reducing them to 25
# in order to quickly compute standard errors in this instance - don't
# of course do this in general)

plot(bw,
     B=25,
     errors="bootstrap")


# EXAMPLE 1 (INTERFACE=DATA FRAME): For this example, we simulate an
# example for a partially linear model and compare the coefficient
# estimates from the partially linear model with those from a correctly
# specified parametric model...

set.seed(42)

n <- 250
x1 <- rnorm(n)
x2 <- rbinom(n, 1, .5)

z1 <- rbinom(n, 1, .5)
z2 <- rnorm(n)

y <- 1 + x1 + x2 + z1 + sin(z2) + rnorm(n)

X <- data.frame(x1, factor(x2))
Z <- data.frame(factor(z1), z2)

# First, compute data-driven bandwidths. This may take a few minutes
# depending on the speed of your computer...

bw <- npplregbw(xdat=X, zdat=Z, ydat=y)

# Next, compute the partially linear fit

pl <- npplreg(bws=bw)

# Print a summary of the model...

summary(pl)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Retrieve the coefficient estimates and their standard errors...

coef(pl)
coef(pl, errors = TRUE)

# Compare the partially linear results to those from a correctly
# specified model's coefficients for x1 and x2

ols <- lm(y~x1+factor(x2)+factor(z1)+I(sin(z2)))

# The intercept is coef()[1], and those for x1 and x2 are coef()[2] and
# coef()[3]. The standard errors are the square root of the diagonal of
# the variance-covariance matrix (elements 2 and 3)

coef(ols)[2:3]
sqrt(diag(vcov(ols)))[2:3]

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Plot the regression surfaces via plot() (i.e., plot the `partial
# regression surface plots').

if (interactive()) plot(bw)

# Note - to plot regression surfaces with variability bounds constructed
# from bootstrapped standard errors, use the following (note that this
# may take a minute or two depending on the speed of your computer as
# the bootstrapping is done in real time, and note also that we override
# the default number of bootstrap replications (399) reducing them to 25
# in order to quickly compute standard errors in this instance - don't
# of course do this in general)

plot(bw,
     B=25,
     errors="bootstrap")
} % enddontrun
}
\keyword{nonparametric}

% 
\name{npindexbw}
\alias{npindexbw}
\alias{npindexbw.default}
\alias{npindexbw.formula}
\alias{npindexbw.sibandwidth}
\title{Semiparametric Single Index Model Parameter and Bandwidth Selection}

\description{
  \code{npindexbw} computes a \code{npindexbw} bandwidth specification
  using the model \eqn{Y = G(X\beta) + \epsilon}{Y = G(XB) +
  epsilon}. For continuous \eqn{Y}, the approach is that of Hardle, Hall
  and Ichimura (1993) which jointly minimizes a least-squares
  cross-validation function with respect to the parameters and
  bandwidth. For binary \eqn{Y}, a likelihood-based cross-validation
  approach is employed which jointly maximizes a likelihood
  cross-validation function with respect to the parameters and
  bandwidth. The bandwidth object contains parameters for the single
  index model and the (scalar) bandwidth for the index function.
}

\usage{
npindexbw(\dots)

\method{npindexbw}{formula}(formula, 
          data, 
          subset, 
          na.action, 
          call, 
          \dots)

\method{npindexbw}{default}(xdat = stop("training data xdat missing"),
          ydat = stop("training data ydat missing"),
          bws,
          bandwidth.compute = TRUE,
          basis = c("glp", "additive", "tensor"),
          bernstein.basis = FALSE,
          degree = NULL,
          degree.select = c("manual", "coordinate", "exhaustive"),
          search.engine = c("nomad+powell", "cell", "nomad"),
          nomad = FALSE,
          nomad.nmulti = 0L,
            degree.min = NULL,
          degree.max = NULL,
          degree.start = NULL,
          degree.restarts = 0L,
          degree.max.cycles = 20L,
          degree.verify = FALSE,
          nmulti,
          nomad.remin = FALSE,
          powell.remin = TRUE,
          only.optimize.beta,
          optim.abstol,
          optim.maxattempts,
          optim.maxit,
          optim.method,
          optim.reltol,
          random.seed,
          regtype = c("lc", "ll", "lp"),
          scale.factor.init.lower = 0.1,
          scale.factor.init.upper = 2.0,
          scale.factor.init = 0.5,
          scale.factor.search.lower = NULL,
          \dots)

\method{npindexbw}{sibandwidth}(xdat = stop("training data xdat missing"),
          ydat = stop("training data ydat missing"),
          bws,
          bandwidth.compute = TRUE,
          nmulti,
          only.optimize.beta = FALSE,
          optim.abstol = .Machine$double.eps,
          optim.maxattempts = 10,
          optim.maxit = 500,
          optim.method = c("Nelder-Mead", "BFGS", "CG"),
          optim.reltol = sqrt(.Machine$double.eps),
          random.seed = 42,
          scale.factor.init.lower = 0.1,
          scale.factor.init.upper = 2.0,
          scale.factor.init = 0.5,
          scale.factor.search.lower = NULL,
          \dots)

}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the data, formula interface, method label, and whether bandwidths are supplied or computed.
}

  \item{bandwidth.compute}{
    a logical value which specifies whether to do a numerical search for
    bandwidths or not. If set to \code{FALSE}, a \code{bandwidth} object
    will be returned with bandwidths set to those specified
    in \code{bws}. Defaults to \code{TRUE}.
  }

  \item{bws}{
    a bandwidth specification. This can be set as a
    \code{singleindexbandwidth}
    object returned from an invocation of \code{npindexbw}, or
    as a vector of parameters (beta) with each element \eqn{i}
    corresponding to the coefficient for column \eqn{i} in \code{xdat}
    where the first element is normalized to 1, and a scalar bandwidth
    (h). If specified as a vector, then additional arguments will need
    to be supplied as necessary to specify the bandwidth type, kernel
    types, and so on.
  }

  \item{call}{
    the original function call. This is passed internally by
    \code{\link{np}} when a bandwidth search has been implied by a call to
    another function. It is not recommended that the user set this.
  }

  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing
    the variables 
    in the model. If not found in data, the variables are taken from
    \code{environment(formula)}, typically the environment from which the
    function is called.
  }

  \item{formula}{
    a symbolic description of variables on which bandwidth selection is
    to be performed. The details of constructing a formula are
    described below.
  }

  \item{method}{
    the single index model method, one of either \dQuote{ichimura}
    (Ichimura (1993)) or \dQuote{kleinspady} (Klein and Spady
    (1993)). Defaults to
    \code{ichimura}.
  }

  \item{na.action}{
    a function which indicates what should happen when the data contain
    \code{NA}s. The default is set by the
    \code{\link{na.action}} setting of options, and is
    \code{\link{na.fail}} if that is unset. The (recommended) default is
    \code{\link{na.omit}}.  
  }
  
  \item{subset}{
    an optional vector specifying a subset of observations to be used in
    the fitting process. 
  }

  \item{xdat}{
    a \eqn{p}-variate data frame of explanatory data (training data) used to
    calculate the regression estimators.
  }

  \item{ydat}{
    a one (1) dimensional numeric or integer vector of dependent data, each
    element \eqn{i} corresponding to each observation (row) \eqn{i} of
    \code{xdat}. 
  }


\subsection{Automatic Degree Search Controls}{
These arguments control automatic local-polynomial degree search.
}

  \item{degree.max}{
    optional scalar or integer vector giving upper bounds for automatic
    degree search when \code{degree.select != "manual"}.
  }

  \item{degree.max.cycles}{
    positive integer giving the maximum number of coordinate-search
    sweeps over the degree vector. Ignored for \code{"manual"} and
    \code{"exhaustive"} degree selection.
  }

  \item{degree.min}{
    optional scalar or integer vector giving lower bounds for automatic
    degree search when \code{degree.select != "manual"}.
  }

  \item{degree.restarts}{
    non-negative integer giving the number of additional deterministic
    coordinate-search restarts. Ignored for \code{"manual"} and
    \code{"exhaustive"} degree selection.
  }

  \item{degree.select}{
    character string controlling local-polynomial degree handling when
    \code{regtype="lp"}. \code{"manual"} (default) treats
    \code{degree} as fixed. \code{"coordinate"} performs cached
    coordinate-wise search over admissible degree values for the index
    smoother. \code{"exhaustive"} evaluates the full admissible degree
    grid when \code{search.engine="cell"}. For NOMAD-based search
    engines, any non-\code{"manual"} value requests direct joint search
    over the degree and bandwidth coordinates.
  }

  \item{degree.start}{
    optional starting degree vector for automatic coordinate search. If
    omitted, the search starts from the degree-zero local-constant
    baseline for the index smoother.
  }

  \item{degree.verify}{
    logical value indicating whether a coordinate-search solution should
    be exhaustively verified over the admissible degree grid after the
    heuristic phase completes. Available only for
    \code{search.engine="cell"}.
  }


\subsection{Continuous Scale-Factor Search Initialization}{
These controls define deterministic and random continuous scale-factor starts and the lower admissibility floor for fixed-bandwidth search.
}

  \item{scale.factor.init}{
    deterministic initial scale factor for continuous fixed-bandwidth
    search. Defaults to \code{0.5}. The value supplied by the user is
    not rewritten, but the effective first start passed to the optimizer
    is \code{max(scale.factor.init, scale.factor.search.lower)}. See
    Details.
  }
  \item{scale.factor.init.lower}{
    lower endpoint for random continuous scale-factor starts. Defaults
    to \code{0.1}. The value supplied by the user is not rewritten, but
    the effective random-start lower endpoint is
    \code{max(scale.factor.init.lower, scale.factor.search.lower)}. See
    Details.
  }
  \item{scale.factor.init.upper}{
    upper endpoint for random continuous scale-factor starts. Defaults
    to \code{2.0}. It must be greater than or equal to the effective
    lower endpoint,
    \code{max(scale.factor.init.lower, scale.factor.search.lower)};
    otherwise bandwidth search errors rather than silently expanding the
    interval. See Details.
  }
  \item{scale.factor.search.lower}{
    optional nonnegative scalar giving the hard lower admissibility
    bound for continuous fixed-bandwidth search candidates. Defaults to
    \code{NULL}. If \code{NULL}, an existing bandwidth object's stored
    value is inherited when available; otherwise the package default
    \code{0.1} is used. This floor applies to computed/search bandwidth
    candidates and to effective search starts only. It does not rewrite
    explicit bandwidths supplied for storage with
    \code{bandwidth.compute = FALSE}. Final fixed-bandwidth search
    candidates must also have a finite valid raw objective value.
  }

\subsection{Local-Polynomial Model Specification}{
These arguments control the index smoother, local-polynomial basis, and fixed degree specification.
}

  \item{basis}{
    local polynomial basis selector used when
    \code{regtype="lp"}: one of \code{"glp"}, \code{"additive"}, or
    \code{"tensor"}. Ignored for \code{"lc"} and \code{"ll"}.
  }

  \item{bernstein.basis}{
    logical flag used when \code{regtype="lp"}; if \code{TRUE}, use a
    Bernstein/B-spline basis for local polynomial terms. When automatic
    degree search is requested and \code{bernstein.basis} is not
    explicitly supplied, the search route defaults to \code{TRUE} for
    numerical stability. Explicit \code{bernstein.basis=FALSE} is
    honored, but raw-polynomial search can be poorly conditioned at
    higher degrees.
  }

  \item{degree}{
    integer degree vector for continuous predictors when
    \code{regtype="lp"}. When \code{degree.select="manual"}, this must
    be supplied explicitly. For single-index regression this is
    typically length one.
  }

  \item{regtype}{
    a character string specifying local smoothing type for the
    nonparametric index regression fit used downstream in
    \code{npindex}. Supported values are \code{"lc"}, \code{"ll"}, and
    \code{"lp"}.
  }


\subsection{NOMAD Search Controls}{
These arguments control the optional NOMAD direct-search route for local-polynomial degree and bandwidth search.
}

  \item{nomad}{
    logical shortcut for the recommended automatic local-polynomial
    NOMAD route. When \code{TRUE}, any missing values among
    \code{regtype}, \code{search.engine}, \code{degree.select},
    \code{bernstein.basis}, \code{degree.min}, \code{degree.max},
    \code{degree.verify}, and \code{bwtype} are filled with
    \code{regtype="lp"}, \code{search.engine="nomad+powell"},
    \code{degree.select="coordinate"},
    \code{bernstein.basis=TRUE}, \code{degree.min=0L},
    \code{degree.max=10L}, \code{degree.verify=FALSE}, and
    \code{bwtype="fixed"}. Explicit incompatible settings error
    immediately; in particular, \code{nomad=TRUE} currently requires
    \code{regtype="lp"}, \code{bwtype="fixed"}, automatic degree
    search, \code{bernstein.basis=TRUE}, no explicit \code{degree},
    and \code{search.engine \%in\% c("nomad", "nomad+powell")}.
    This shortcut does not change the meaning of \code{nmulti} or
    \code{nomad.nmulti}: \code{nmulti} remains the outer restart count,
    while \code{nomad.nmulti} controls inner
    \code{crs::snomadr()} multistarts within each outer restart.
    Returned bandwidth objects retain this normalized preset metadata in
    \code{bw$nomad.shortcut} for a returned object \code{bw}; when
    available, \code{nomad.time} and \code{powell.time} record the
    direct-search and Powell-polish timing components.
  }

  \item{nomad.nmulti}{
    non-negative integer controlling the inner
    \code{crs::snomadr()} multistart count used within each outer
    NOMAD restart when \code{regtype="lp"} and automatic degree search
    uses \code{search.engine="nomad"} or \code{"nomad+powell"}.
    Defaults to \code{0L}, which preserves the current one-start-per-
    restart behavior. This does not replace \code{nmulti}:
    \code{nmulti} controls outer restarts, while \code{nomad.nmulti}
    controls inner NOMAD multistarts within each outer restart.
  }

  \item{nomad.remin}{
    logical flag controlling the optional second NOMAD hot start. When
    \code{TRUE}, NOMAD is restarted once from the best full candidate found,
    including both bandwidth and degree coordinates. Defaults to \code{FALSE};
    current simulation evidence favors the one-pass NOMAD default for routine
    use, while leaving this switch available for sensitivity checks.
  }


  \item{search.engine}{
    character string controlling the automatic local-polynomial search
    backend when \code{regtype="lp"} and
    \code{degree.select != "manual"}. \code{"nomad+powell"} (default)
    performs direct joint search over the index-smoother bandwidth and
    degree using \code{crs::snomadr()}, then applies one Powell hot
    start from the NOMAD solution. \code{"nomad"} omits the Powell
    refinement. \code{"cell"} profiles the criterion over the
    admissible degree grid using repeated fixed-degree solves.
    NOMAD-based search currently requires \code{degree.verify=FALSE}
    and the suggested package \pkg{crs} to be installed.
  }

\subsection{Numerical Search Controls}{
These controls set search restart behavior.
}

  \item{nmulti}{
    integer number of times to restart the process of finding extrema of
    the cross-validation function from different (random) initial
    points. Defaults to \code{min(2,ncol(xdat))}.
  }

  \item{powell.remin}{
    logical flag controlling Powell restart-from-minimum behavior. For
    ordinary fixed-degree Powell-style search, \code{TRUE} restarts the
    local search from the located minimum. For \code{search.engine="nomad+powell"},
    this controls only the final Powell bandwidth-polish step. The default
    is \code{TRUE} for ordinary Powell routes and \code{FALSE} for the
    Powell polish after NOMAD unless explicitly supplied.
  }


\subsection{Optimization Controls}{
These arguments control outer optimization behavior for the semiparametric search.
}

   \item{only.optimize.beta}{
     signals the routine to only minimize the objective function with
     respect to beta
   }

  \item{optim.abstol}{
    the absolute convergence tolerance used by \code{\link{optim}}. Only useful
    for non-negative functions, as a tolerance for reaching
    zero. Defaults to \code{.Machine$double.eps}.
  }

  \item{optim.maxattempts}{
    maximum number of attempts taken trying to achieve successful
    convergence in \code{\link{optim}}. Defaults to \code{100}.
  }

  \item{optim.maxit}{
    maximum number of iterations used by \code{\link{optim}}. Defaults
     to \code{500}.
   }

  \item{optim.method}{ method used by \code{\link{optim}} for minimization of
    the objective function. See \code{?optim} for references. Defaults
    to \code{"Nelder-Mead"}.

    the default method is an implementation of that of Nelder and Mead
    (1965), that uses only function values and is robust but relatively
    slow.  It will work reasonably well for non-differentiable
    functions.

    method \code{"BFGS"} is a quasi-Newton method (also known as a
    variable metric algorithm), specifically that published
    simultaneously in 1970 by Broyden, Fletcher, Goldfarb and Shanno.
    This uses function values and gradients to build up a picture of the
    surface to be optimized.

    method \code{"CG"} is a conjugate gradients method based
    on that by Fletcher and Reeves (1964) (but with the option of
    Polak-Ribiere or Beale-Sorenson updates).  Conjugate gradient
    methods will generally be more fragile than the BFGS method, but as
    they do not store a matrix they may be successful in much larger
    optimization problems.
  }

  \item{optim.reltol}{
    relative convergence tolerance used by \code{\link{optim}}.  The algorithm
    stops if it is unable to reduce the value by a factor of 'reltol *
    (abs(val) + reltol)' at a step.  Defaults to
    \code{sqrt(.Machine$double.eps)}, typically about \code{1e-8}.
  }

  \item{random.seed}{
    an integer used to seed R's random number generator. This ensures
    replicability of the numerical search. Defaults to 42.
  }


\subsection{Additional Arguments}{
These arguments collect remaining controls passed through S3 methods.
}

  \item{\dots}{
    additional arguments supplied to specify the parameters to the
    \code{sibandwidth} S3 method, which is called during the numerical
    search. 
  }

  


}

\details{

The \code{scale.factor.*} controls are dimensionless search
controls. The package converts scale factors to bandwidths using the
estimator-specific scaling encoded in the bandwidth object, including
kernel order and the number of continuous variables relevant for the
estimator. Users should not pre-multiply these controls by sample-size
or standard-deviation factors.

  \code{scale.factor.init} controls the deterministic first search
  start when that control is exposed. \code{scale.factor.init.lower}
  and \code{scale.factor.init.upper} define the random multistart
  interval when exposed. \code{scale.factor.search.lower} is the lower
  admissibility bound for continuous fixed-bandwidth search candidates.
  The effective first start is \code{max(scale.factor.init,
  scale.factor.search.lower)} when both controls are present, and the
  effective random-start lower endpoint is
  \code{max(scale.factor.init.lower, scale.factor.search.lower)}.
  \code{scale.factor.init.upper} must be at least that effective lower
  endpoint; the package errors rather than silently expanding the user's
  interval.

  When \code{scale.factor.search.lower} is \code{NULL}, an existing
  bandwidth object's stored floor is inherited when available;
  otherwise the package default \code{0.1} is used. Explicit bandwidths
  supplied for storage with \code{bandwidth.compute = FALSE} are not
  rewritten by the search floor.

  Categorical search-start controls such as \code{dfac.init},
  \code{lbd.init}, and \code{hbd.init} have separate semantics and are
  not affected by \code{scale.factor.search.lower}.

  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}}, \code{\link{plot.np}} for plotting options.

  For S3 plotting help, see \code{\link{plot.np}}. You can list
  available plot methods with \code{methods("plot")}.

  We implement Ichimura's (1993) method via
  joint estimation of the bandwidth and coefficient vector using
  leave-one-out nonlinear least squares. We implement Klein and Spady's
  (1993) method maximizing the leave-one-out log likelihood function
  jointly with respect to the bandwidth and coefficient vector. Note
  that Klein and Spady's (1993) method is for \emph{binary outcomes
    only}, while Ichimura's (1993) method can be applied for any outcome
  data type (i.e., continuous or discrete).

  We impose the identification condition that the first element of the
  coefficient vector beta is equal to one, while identification also
  requires that the explanatory variables contain \emph{at least one}
  continuous variable.

  \code{npindexbw} may be invoked \emph{either} with a formula-like
  symbolic description of variables on which bandwidth selection is to
  be performed \emph{or} through a simpler interface whereby data is
  passed directly to the function via the \code{xdat} and \code{ydat}
  parameters. Use of these two interfaces is \bold{mutually exclusive}.
  
  Note that, unlike most other bandwidth methods in the \code{np}
  package, this implementation uses the R \code{\link{optim}} nonlinear
  minimization routines and \code{\link{npksum}}. We have implemented
  multistarting and strongly encourage its use in practice. For
  exploratory purposes, you may wish to override the default search
  tolerances, say, setting \code{optim.reltol=.1} and conduct
  multistarting (the default is to restart min(2, ncol(xdat)) times) as is done
  for a number of examples.

  Data for which bandwidths are to be estimated may be specified
  symbolically. A typical description has the form \code{dependent data
  ~ explanatory data}, where \code{dependent data} is a univariate
  response, and \code{explanatory data} is a series of variables
  specified by name, separated by the separation character '+'. For
  example \code{ y1 ~ x1 + x2 } specifies that the bandwidth object for
  the regression of response \code{y1} and semiparametric regressors
  \code{x1} and \code{x2} are to be estimated.  See below for further
  examples.

  When \code{regtype="lp"} and \code{degree.select != "manual"},
  \code{npindexbw} can jointly determine the local-polynomial degree for
  the index smoother together with its bandwidth coordinate. With
  \code{search.engine="cell"}, the criterion is profiled over the
  admissible degree grid using cached coordinate-wise or exhaustive
  search. With \code{search.engine="nomad"} or
  \code{"nomad+powell"}, the criterion is optimized directly over the
  joint degree/bandwidth space using \code{crs::snomadr()};
  \code{"nomad+powell"} then performs one Powell hot start and retains
  the better of the direct NOMAD and polished solutions. For the
  index-smoother local-polynomial component, this polynomial-adaptive
  joint-search route follows Hall and Racine (2015).

  Setting \code{nomad=TRUE} is a convenience preset for this automatic
  LP route, not a generic optimizer alias. For single-index bandwidth
  selection it expands any missing values to the equivalent long-form
  call
  \preformatted{
    npindexbw(...,
              regtype = "lp",
              search.engine = "nomad+powell",
              degree.select = "coordinate",
              bernstein.basis = TRUE,
              degree.min = 0L,
              degree.max = 10L,
              degree.verify = FALSE,
              bwtype = "fixed")
  }
  Compatible explicit tuning arguments are respected. Incompatible
  explicit settings fail fast so the shortcut never silently changes
  user-selected semantics.
  
}

\value{
  \code{npindexbw} returns a \code{sibandwidth} object, with the
  following components:

  \item{bw}{ bandwidth(s), scale factor(s) or nearest neighbours for the
    data, \code{xdat} }
  \item{beta}{ coefficients of the model }
  \item{fval}{ objective function value at minimum }

  If \code{bwtype} is set to \code{fixed}, an object containing a scalar
  bandwidth for the function \eqn{G(X\beta)}{G(XB)} and an estimate of
  the parameter vector \eqn{\beta}{B} is returned.

  If \code{bwtype} is set to \code{generalized_nn} or
  \code{adaptive_nn}, then instead the scalar \eqn{k}th nearest neighbor
  is returned.

  The functions \code{\link{coef}}, \code{\link{predict}},
  \code{\link{summary}}, and  \code{\link{plot}} support
  objects of this class.
}

\section{Book And Method Pointers}{
  \code{npindexbw} selects the index direction, bandwidth, and when
  requested local-polynomial degree/search metadata for the single-index
  model. The continuous-outcome target is typically written
  \eqn{E[Y\mid X]=g(X^\prime\beta)}, with \eqn{\beta} normalized for
  identification; binary outcomes use the Klein-Spady route.

  For book-length derivations, see Li and Racine (2007), Chapter 8
  \emph{Semiparametric Single Index Models}, especially Sections 8.1,
  8.2.1, 8.4, 8.5, and 8.10, and Racine (2019), Chapter 8.
}

\references{

  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Hardle, W. and P. Hall and H. Ichimura (1993), \dQuote{Optimal
  Smoothing in Single-Index Models,} The Annals of Statistics, 21,
  157-178.

  Ichimura, H., (1993), \dQuote{Semiparametric least squares (SLS) and
  weighted SLS estimation of single-index models,} Journal of
  Econometrics, 58, 71-120.

  Klein, R. W. and R. H. Spady (1993), \dQuote{An efficient semiparametric
  estimator for binary response models,} Econometrica, 61, 387-421.

  Hall, P. and J.S. Racine (2015), \dQuote{Infinite Order Cross-Validated
  Local Polynomial Regression,} Journal of Econometrics, 185, 510-525.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics:
  Theory and Practice,} Princeton University Press.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth
    estimators for discrete distributions,} Biometrika, 68, 301-309.
}

    

\author{
  Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.

  Caution: multivariate data-driven bandwidth selection methods are, by
  their nature, \emph{computationally intensive}. Virtually all methods
  require dropping the \eqn{i}th observation from the data set,
  computing an object, repeating this for all observations in the
  sample, then averaging each of these leave-one-out estimates for a
  \emph{given} value of the bandwidth vector, and only then repeating
  this a large number of times in order to conduct multivariate
  numerical minimization/maximization. Furthermore, due to the potential
  for local minima/maxima, \emph{restarting this procedure a large
  number of times may often be necessary}. This can be frustrating for
  users possessing large datasets. For exploratory purposes, you may
  wish to override the default search tolerances, say, setting
  \code{optim.reltol=.1} and conduct multistarting (the default is to
  restart min(2, ncol(xdat)) times). Once the procedure terminates, you can
  restart search with default tolerances using those bandwidths obtained
  from the less rigorous search (i.e., set \code{bws=bw} on subsequent
  calls to this routine where \code{bw} is the initial bandwidth
  object).  A version of this package using the \code{Rmpi} wrapper is
  under development that allows one to deploy this software in a
  clustered computing environment to facilitate computation involving
  large datasets.  }

\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): Generate a simple linear model then
# compute coefficients and the bandwidth using Ichimura's nonlinear
# least squares approach.

set.seed(12345)

n <- 100

x1 <- runif(n, min=-1, max=1)
x2 <- runif(n, min=-1, max=1)

y <- x1 - x2 + rnorm(n)

# Note - this may take a minute or two depending on the speed of your
# computer. Note also that the first element of the vector beta is
# normalized to one for identification purposes, and that X must contain
# at least one continuous variable.

bw <- npindexbw(formula=y~x1+x2, method="ichimura")

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# EXAMPLE 1 (INTERFACE=DATA FRAME): Generate a simple linear model then
# compute coefficients and the bandwidth using Ichimura's nonlinear
# least squares approach.

set.seed(12345)

n <- 100

x1 <- runif(n, min=-1, max=1)
x2 <- runif(n, min=-1, max=1)

y <- x1 - x2 + rnorm(n)

X <- cbind(x1, x2)

# Note - this may take a minute or two depending on the speed of your
# computer. Note also that the first element of the vector beta is
# normalized to one for identification purposes, and that X must contain
# at least one continuous variable.

bw <- npindexbw(xdat=X, ydat=y, method="ichimura")

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# EXAMPLE 2 (INTERFACE=DATA FRAME): Generate a simple binary outcome
# model then compute coefficients and the bandwidth using Klein and
# Spady's likelihood-based approach.

n <- 100

x1 <- runif(n, min=-1, max=1)
x2 <- runif(n, min=-1, max=1)

y <- ifelse(x1 + x2 + rnorm(n) > 0, 1, 0)

# Note that the first element of the vector beta is normalized to one
# for identification purposes, and that X must contain at least one
# continuous variable.

bw <- npindexbw(formula=y~x1+x2, method="kleinspady")

summary(bw)

# EXAMPLE 2 (INTERFACE=DATA FRAME): Generate a simple binary outcome
# model then compute coefficients and the bandwidth using Klein and
# Spady's likelihood-based approach.

n <- 100

x1 <- runif(n, min=-1, max=1)
x2 <- runif(n, min=-1, max=1)

y <- ifelse(x1 + x2 + rnorm(n) > 0, 1, 0)

X <- cbind(x1, x2)

# Note that the first element of the vector beta is normalized to one
# for identification purposes, and that X must contain at least one
# continuous variable.

bw <- npindexbw(xdat=X, ydat=y, method="kleinspady")

summary(bw)
} % enddontrun
}
\keyword{nonparametric}

\name{npudenshat}
\alias{npudenshat}
\title{Unconditional Density Hat Operator}
\description{
Constructs the unconditional density hat operator associated with
\code{\link{npudens}} bandwidth objects. The returned operator maps a
right-hand side \code{y} to \eqn{H y}; with \code{y = 1} this reproduces the
fitted unconditional density.
}
\usage{
npudenshat(bws,
           tdat = stop("training data 'tdat' missing"),
           edat,
           y = NULL,
           output = c("matrix", "apply"))
}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the fitted bandwidth object, training data, and evaluation data.
}

  \item{bws}{A fitted unconditional density bandwidth object of class \code{"bandwidth"}.}


  \item{edat}{Optional evaluation data. If omitted, the operator is built on the training data.}


  \item{tdat}{Training data used to construct the operator.}


\subsection{Operator Output}{
These arguments control whether the operator is returned as a matrix or applied directly.
}

  \item{output}{Either \code{"matrix"} for the hat matrix or \code{"apply"} for direct application to \code{y}.}

  \item{y}{Optional right-hand side vector or matrix with one row per training observation.}


}
\details{
For \code{output = "matrix"}, the return value is a matrix with class
\code{c("npudenshat", "matrix")} and attributes storing the bandwidth object,
training data, evaluation data, and call metadata.

For \code{output = "apply"}, the function returns \code{H y} directly. Matrix
right-hand sides are applied column-wise.

This helper is intended for object-fed repeated evaluation once a bandwidth
object has already been constructed. It does not perform bandwidth selection.
}
\value{
Either a hat matrix of class \code{"npudenshat"} or the applied result
\code{H y}, depending on \code{output}.
}
\examples{
\dontrun{
data(cps71)
tx <- data.frame(age = cps71$age)

bw <- npudensbw(dat = tx, bwtype = "fixed",
                bandwidth.compute = FALSE, bws = 1.0)

H <- npudenshat(bws = bw, tdat = tx)
dens.hat <- npudenshat(bws = bw, tdat = tx,
                       y = rep(1, nrow(tx)),
                       output = "apply")
dens.core <- fitted(npudens(bws = bw, tdat = tx))

head(cbind(dens.core, dens.hat), n = 2L)
}
}
\keyword{nonparametric}
\keyword{density}

\name{npcdenshat}
\alias{npcdenshat}
\title{Conditional Density Hat Operator}
\description{
Constructs the conditional density hat operator associated with
\code{\link{npcdens}} bandwidth objects. The returned operator maps a
right-hand side \code{y} to \eqn{H y}; with \code{y = 1} this reproduces the
fitted conditional density.
}
\usage{
npcdenshat(bws,
           txdat = stop("training data 'txdat' missing"),
           tydat = stop("training data 'tydat' missing"),
           exdat,
           eydat,
           y = NULL,
           output = c("matrix", "apply"),
           deriv = NULL,
           s = NULL)
}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the fitted bandwidth object, training data, and evaluation data.
}

  \item{bws}{A fitted conditional density bandwidth object of class \code{"conbandwidth"}.}


  \item{exdat}{Optional evaluation conditioning data. If omitted, the operator is built on the training conditioning data.}


  \item{eydat}{Optional evaluation response data. If omitted, the operator is built on the training response data.}


  \item{txdat}{Training conditioning data used to construct the operator.}


  \item{tydat}{Training response data used to construct the operator.}


\subsection{Operator Output}{
These arguments control whether the operator is returned as a matrix or applied directly.
}

  \item{deriv}{Convenience alias for \code{s}.}

  \item{output}{Either \code{"matrix"} for the hat matrix or \code{"apply"} for direct application to \code{y}.}

  \item{s}{Optional first-derivative multi-index over continuous conditioning variables. By default the conditional-density operator itself is returned. When \code{s} requests one first derivative, the returned operator maps \code{y} to the corresponding \code{npcdens(..., gradients = TRUE)} conditioning-variable gradient component.}

  \item{y}{Optional right-hand side vector or matrix with one row per training observation.}


}
\details{
For \code{output = "matrix"}, the return value is a matrix with class
\code{c("npcdenshat", "matrix")} and attributes storing the bandwidth object,
training data, evaluation data, and call metadata.

For \code{output = "apply"}, the function returns \code{H y} directly. Matrix
right-hand sides are applied column-wise.

The optional \code{s} argument follows the derivative multi-index convention
used by \code{\link{npreghat}} for continuous conditioning variables. The
default \code{s = NULL} returns the ordinary conditional-density hat operator.
Supplying one first derivative, for example \code{s = 1L} in a univariate
continuous-\eqn{x} problem or \code{s = c(x2 = 1L)} in a multivariate problem,
returns the hat operator for that component of the conditional-density
gradient. This derivative route is intended to match
\code{npcdens(..., gradients = TRUE)$congrad}.

This helper is intended for object-fed repeated evaluation once a bandwidth
object has already been constructed. It does not perform bandwidth selection.
}
\value{
Either a hat matrix of class \code{"npcdenshat"} or the applied result
\code{H y}, depending on \code{output}.
}
\examples{
\dontrun{
data(cps71)
tx <- data.frame(age = cps71$age)
ty <- data.frame(logwage = cps71$logwage)

bw <- npcdensbw(xdat = tx, ydat = ty, bwtype = "fixed",
                bandwidth.compute = FALSE, bws = c(1.0, 1.0))

H <- npcdenshat(bws = bw, txdat = tx, tydat = ty)
dens.hat <- npcdenshat(bws = bw, txdat = tx, tydat = ty,
                       y = rep(1, nrow(tx)),
                       output = "apply")
dens.core <- fitted(npcdens(bws = bw, txdat = tx, tydat = ty))

head(cbind(dens.core, dens.hat), n = 2L)

grad.hat <- npcdenshat(bws = bw, txdat = tx, tydat = ty,
                       y = rep(1, nrow(tx)),
                       output = "apply", s = 1L)
grad.core <- gradients(npcdens(bws = bw, txdat = tx, tydat = ty,
                               gradients = TRUE))[, 1L]
head(cbind(grad.core, grad.hat), n = 2L)
}
}
\keyword{nonparametric}
\keyword{density}

\name{np.options}
\alias{np.options}
\title{Global Package Options for \pkg{np}}
\description{
  Global options controlling selected computational and display behavior
  for the \pkg{np} package.
}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels and
  \code{\link{plot}} for plotting options.

  The following options are recognized by \pkg{np}.

  \itemize{
    \item \code{np.messages} (logical): controls console/progress output.
    Default is \code{TRUE}.

    \item \code{np.plot.progress} (logical): controls bounded
    plot/bootstrap progress heartbeats. Default is \code{TRUE}.

    \item \code{np.plot.progress.start.grace.sec} (numeric): delay before
    the first plot/bootstrap progress line is shown. Default is
    \code{0.75}.

    \item \code{np.plot.progress.interval.sec} (numeric): minimum elapsed
    time between plot/bootstrap heartbeat lines once progress reporting has
    started. Default is \code{0.5}.

    \item \code{np.plot.progress.max.intermediate} (integer): maximum
    number of mid-run plot/bootstrap heartbeat lines emitted between the
    initial start notice and final completion line. Default is \code{3}.

    \item \code{np.tree} (logical): enables kd-tree acceleration when
    supported by the selected kernel/operator combination. Default is
    \code{FALSE}.

    \item \code{np.largeh.rel.tol} (numeric): relative tolerance used by the
    continuous large-\eqn{h} shortcut. When all standardized distances for a
    continuous predictor are sufficiently close to zero, the corresponding
    kernel factor is approximated by \eqn{K(0)} to reduce repeated kernel
    evaluations. Default is \code{1e-3}. Valid range is \code{(0, 0.1)}.

    \item \code{np.disc.upper.rel.tol} (numeric): relative tolerance used by
    the discrete upper-bound shortcut for bandwidths near their feasible upper
    bounds. The near-upper check is applied relative to each kernel's own
    feasible upper bound (e.g., Aitchison-Aitken depends on category
    cardinality), with a tiny machine-precision floor for numerical robustness.
    When same/different-category kernel values are numerically close, the
    corresponding discrete kernel factor is treated as constant to reduce
    repeated category comparisons. Default is \code{1e-2}. Valid range is
    \code{(0, 0.5)}.

    \item \code{plot.par.mfrow} (logical): used by \code{\link{plot}} to
    determine whether plotting layout is automatically managed via
    \code{par(mfrow=...)}. If \code{NULL} (default behavior), \pkg{np} uses
    its internal plotting defaults.
  }

  Option values can be set globally via \code{\link{options}} and restored
  with \code{\link{on.exit}} in scripts/functions for reproducibility.
}
\author{
  Jeffrey S. Racine \email{racinej@mcmaster.ca}
}
\seealso{
  \code{\link{np.kernels}}, \code{\link{plot}}
  \code{\link{np}}, \code{\link{plot}}, \code{\link{options}}
}
\examples{
\dontrun{
old <- options(
  np.tree = TRUE,
  np.messages = FALSE,
  np.largeh.rel.tol = 1e-3,
  np.disc.upper.rel.tol = 1e-2
)
on.exit(options(old), add = TRUE)

## ... run bandwidth selection / estimation ...
}
}
\keyword{misc}

\name{np}
\alias{np}
\alias{np-package}
\docType{package}
\title{Nonparametric Kernel Smoothing Methods for Mixed Data Types}
\description{

	This package provides a variety of nonparametric and semiparametric
	kernel methods that seamlessly handle a mix of continuous, unordered,
	and ordered factor data types (unordered and ordered factors are often
	referred to as \sQuote{nominal} and \sQuote{ordinal} categorical
	variables respectively). A getting-started vignette containing a
	short introduction to the \pkg{np} package can be accessed via
	\code{vignette("np_getting_started", package = "np")}. An installed
	vignette for the entropy-based test family can be accessed via
	\code{vignette("np_entropy_tests", package = "np")}.

  For a listing of all routines in the \pkg{np} package type:
  \sQuote{library(help="np")}.  

  Bandwidth selection is a key aspect of sound nonparametric and
  semiparametric kernel estimation. \code{np} is designed from the
  ground up to make bandwidth selection the focus of attention. To this
  end, one typically begins by creating a \sQuote{bandwidth object}
  which embodies all aspects of the method, including specific kernel
  functions, data names, data types, and the like. One then passes these
  bandwidth objects to other functions, and those functions can grab the
  specifics from the bandwidth object thereby removing potential
  inconsistencies and unnecessary repetition. Furthermore, many
  functions such as \code{\link{plot}} (via class-specific S3 methods)
  can work with the bandwidth object directly without
  having to do the subsequent companion function evaluation.

  The user may also combine these steps. If the first step (bandwidth
  selection) is not performed explicitly then the second step will
  automatically call the omitted first-step bandwidth selector using
  defaults unless otherwise specified, and the bandwidth object can be
  retrieved retroactively if so desired via \code{objectname$bws}.
  Furthermore, options for bandwidth selection will be passed directly
  to the bandwidth selector function. Note that the combined approach
  would not be a wise choice for certain applications such as when
  bootstrapping (as it would involve unnecessary computation since the
  bandwidths would properly be those for the original sample and not the
  bootstrap resamples) or when conducting quantile regression (as it
  would involve unnecessary computation when different quantiles are
  computed from the same conditional cumulative distribution estimate).
  
  There are two ways in which you can interact with functions in
  \code{np}, either i) using data frames, or ii) using a formula
  interface, where appropriate.

  To some, it may be natural to use the data frame interface.  The R
  \code{\link{data.frame}} function preserves a variable's type once it
  has been cast (unlike \code{\link{cbind}}, which we avoid for this
  reason).  If you find this most natural for your project, you first
  create a data frame casting data according to their type (i.e., one of
  continuous (default, \code{\link{numeric}}), \code{\link{factor}},
  \code{\link{ordered}}). Then you would simply pass this data frame to
  the appropriate \code{np} function, for example
  \code{npudensbw(dat=data)}.

  To others, however, it may be natural to use the formula interface
  that is used for the regression examples, among others. For
  nonparametric regression functions such as \code{\link{npreg}}, you
  would proceed as you would using \code{\link{lm}} (e.g., \code{bw <-
  npregbw(y~factor(x1)+x2))} except that you would of course not need to
  specify, e.g., polynomials in variables, interaction terms, or create
  a number of dummy variables for a factor.  Every function in np
  supports both interfaces, where appropriate.

  Note that if your factor is in fact a character string such as, say,
	\code{X} being either \code{"MALE"} or \code{"FEMALE"}, np will handle
	this directly, i.e., there is no need to map the string values into
	unique integers such as (0,1). Once the user casts a variable as a
	particular data type (i.e., \code{\link{factor}},
	\code{\link{ordered}}, or continuous (default,
	\code{\link{numeric}})), all subsequent methods automatically detect
	the type and use the appropriate kernel function and method where
	appropriate.

	All estimation methods are fully multivariate, i.e., there are no
	limitations on the number of variables one can model (or number of
	observations for that matter). Execution time for most routines is,
	however, exponentially increasing in the number of observations and
	increases with the number of variables involved.

	Nonparametric methods include unconditional density (distribution),
	conditional density (distribution), regression, mode, and quantile
	estimators along with gradients where appropriate, while
	semiparametric methods include single index, partially linear, and
	smooth (i.e., varying) coefficient models.
	
	A number of tests are included such as consistent specification tests
	for parametric regression and quantile regression models along with
	tests of significance for nonparametric regression.
	
	A variety of bootstrap methods for computing standard errors,
	nonparametric confidence bounds, and bias-corrected bounds are
	implemented.

	A variety of bandwidth methods are implemented including fixed,
	nearest-neighbor, and adaptive nearest-neighbor.

	A variety of data-driven methods of bandwidth selection are
	implemented, while the user can specify their own bandwidths should
	they so choose (either a raw bandwidth or scaling factor).

	A flexible plotting utility, via class-specific S3
	\code{\link{plot}} methods, facilitates graphing of multivariate
	objects. An example for creating postscript graphs and pulling this
	into a LaTeX document is provided.

	The function \code{\link{npksum}} allows users to create or implement
	their own kernel estimators or tests should they so desire.

	The underlying functions are written in C for computational
	efficiency. Despite this, due to their nature, data-driven bandwidth
	selection methods involving multivariate numerical search can be
	time-consuming, particularly for large datasets. For MPI-oriented
	workflows on larger jobs, see the companion package \pkg{npRmpi},
	which extends the same mixed-data kernel methodology to clustered
	computing environments.

  To cite the \code{np} package, type \code{citation("np")} from within
  \code{R} for details.

}

\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.
 The kernel methods in \code{np} employ the so-called
	\sQuote{generalized product kernels} found in Hall, Racine,
	and Li (2004), Li, Lin, and Racine (2013), Li, Ouyang, and
	Racine (2013), Li and Racine (2003), Li and Racine (2004), Li
	and Racine (2007), Li and Racine (2010), Ouyang, Li, and
	Racine (2006), and Racine and Li (2004), among others. For
	details on a particular method, kindly refer to the original
	references listed above.

  We briefly describe the particulars of various univariate kernels used
  to generate the generalized product kernels that underlie the kernel
  estimators implemented in the \code{np} package. In a nutshell, the
  generalized kernel functions that underlie the kernel estimators in
  \code{np} are formed by taking the product of univariate kernels such
  as those listed below. When you cast your data as a particular type
  (continuous, factor, or ordered factor) in a data frame or formula,
  the routines will automatically recognize the type of variable being
  modelled and use the appropriate kernel type for each variable in the
  resulting estimator.

	\describe{

		\item{Second Order Gaussian (\eqn{x} is continuous)}{
			\eqn{k(z) = \exp(-z^2/2)/\sqrt{2\pi}}{k(z) =
				exp(-z^2/2)/\sqrt 2\pi} where \eqn{z=(x_i-x)/h}{z=(x[i]-x)/h},
      and \eqn{h>0}.
    }
    
    \item{Second Order Truncated Gaussian (\eqn{x} is continuous)}{
			\eqn{k(z) = (\exp(-z^2/2)-\exp(-b^2/2))/(\textrm{erf}(b/\sqrt{2})\sqrt{2\pi}-2b\exp(-b^2/2))}{k(z) =
				(exp(-z^2/2)-exp(-b^2/2))/(erf(b/\sqrt{2})\sqrt{2\pi}-2b\exp(-b^2/2))} where \eqn{z=(x_i-x)/h}{z=(x[i]-x)/h}, \eqn{b>0}, \eqn{|z|\le b}
      and \eqn{h>0}.
      
      See \code{\link{nptgauss}} for details on modifying \eqn{b}.
    }

    \item{Second Order Epanechnikov (\eqn{x} is continuous)}{
      \eqn{k(z) =  3\left(1 - z^2/5\right)/(4\sqrt{5})}{
        k(z) =  3(1 - z^2/5)/(4\sqrt 5)}
      if \eqn{z^2<5}, \eqn{0} otherwise, where
      \eqn{z=(x_i-x)/h}{z=(x[i]-x)/h}, and \eqn{h>0}.
    }

    \item{Uniform (\eqn{x} is continuous)}{
      \eqn{k(z) = 1/2} if \eqn{|z|<1}, \eqn{0} otherwise, where
      \eqn{z=(x_i-x)/h}{z=(x[i]-x)/h}, and \eqn{h>0}.
    }

    \item{Aitchison and Aitken (\eqn{x} is a (discrete) factor)}{
      \eqn{l(x_i,x,\lambda) = 1 - \lambda}{l(x[i],x,lambda) = 1 -
        lambda} if \eqn{x_i=x}{x[i] = x}, and
      \eqn{\lambda/(c-1)}{lambda/(c-1)} if \eqn{x_i \neq x}{x[i] != x},
      where \eqn{c} is the number of (discrete) outcomes assumed by the
      factor \eqn{x}.

      Note that \eqn{\lambda}{lambda} must lie between \eqn{0} and
      \eqn{(c-1)/c}.
   }

    \item{Wang and van Ryzin (\eqn{x} is a (discrete) ordered factor)}{
      \eqn{l(x_i,x,\lambda) = 1 - \lambda}{l(x[i],x,lambda) = 1 -
        lambda} if \eqn{|x_i-x|=0}{|x[i] - x| = 0}, and
      \eqn{((1-\lambda)/2)\lambda^{|x_i-x|}}{((1-lambda)/2)lambda^|x_i-x|}
      if \eqn{|x_i - x|\ge1}{|x[i] - x|>=1}.

      Note that \eqn{\lambda}{lambda} must lie between \eqn{0} and
      \eqn{1}.

    }
      

    \item{Li and Racine (\eqn{x} is a (discrete) factor)}{
      \eqn{l(x_i,x,\lambda) = 1 }{l(x[i],x,lambda) =
        1} if \eqn{x_i=x}{x[i] = x}, and
      \eqn{\lambda}{lambda} if \eqn{x_i \neq x}{x[i] != x}.

      Note that \eqn{\lambda}{lambda} must lie between \eqn{0} and
      \eqn{1}.
    }
    
        \item{Li and Racine Normalised for Unconditional Objects (\eqn{x} is a (discrete) factor)}{
      \eqn{l(x_i,x,\lambda) = 1/(1+(c-1)\lambda) }{l(x[i],x,lambda) =
        1/(1+(c-1)lambda)} if \eqn{x_i=x}{x[i] = x}, and
      \eqn{\lambda/(1+(c-1)\lambda)}{lambda/(1+(c-1)lambda)} if \eqn{x_i \neq x}{x[i] != x}.

      Note that \eqn{\lambda}{lambda} must lie between \eqn{0} and
      \eqn{1}.
    }

    \item{Li and Racine (\eqn{x} is a (discrete) ordered factor)}{
      \eqn{l(x_i,x,\lambda) = 1}{l(x[i],x,lambda) = 1} if
      \eqn{|x_i-x|=0}{|x[i] - x| = 0}, and
      \eqn{\lambda^{|x_i-x|}}{lambda^|x_i-x|} if \eqn{|x_i -
        x|\ge1}{|x[i] - x|>=1}.

      Note that \eqn{\lambda}{lambda} must lie between \eqn{0} and
      \eqn{1}.
    }
    
    \item{Li and Racine Normalised for Unconditional Objects (\eqn{x} is a (discrete) ordered factor)}{
      \eqn{l(x_i,x,\lambda) = (1-\lambda)/(1+\lambda)}{l(x[i],x,lambda) = (1-lambda)/(1+lambda)} if
      \eqn{|x_i-x|=0}{|x[i] - x| = 0}, and
      \eqn{(1-\lambda)/(1+\lambda)\lambda^{|x_i-x|}}{ (1-lambda)/(1+lambda)lambda^|x_i-x|} if \eqn{|x_i -
        x|\ge1}{|x[i] - x|>=1}.

      Note that \eqn{\lambda}{lambda} must lie between \eqn{0} and
      \eqn{1}.
    }

    \item{Racine, Li, and Yan (\eqn{x} is a (discrete) ordered factor)}{
      \eqn{l(x_i,x,\lambda)=\lambda^{|x_i-x|}/\sum_{z \in D}\lambda^{|x_i-z|}}{l(x[i],x,lambda)=lambda^|x[i]-x|/sum_{z in D}lambda^|x[i]-z|},
      where \eqn{D} is the ordered support.

      Note that \eqn{\lambda}{lambda} must lie between \eqn{0} and
      \eqn{1}.
    }

  }

  So, if you had two variables, \eqn{x_{i1}}{x1[i]} and
  \eqn{x_{i2}}{x2[i]}, and \eqn{x_{i1}}{x1[i]} was continuous while
  \eqn{x_{i2}}{x2[i]} was, say, binary (0/1), and you created a data
  frame of the form \code{X <- data.frame(x1,x2=factor(x2))}, then the
  kernel function used by \code{np} would be
  \eqn{K(\cdot)=k(\cdot)\times l(\cdot)}{K(.)=k(.)*l(.)} where the
  particular kernel functions \eqn{k(\cdot)}{k(.)} and
  \eqn{l(\cdot)}{l(.)} would be, say, the second order Gaussian
  (\code{ckertype="gaussian"}) and Aitchison and Aitken
  (\code{ukertype="aitchisonaitken"}) kernels by default, respectively
  (note that for conditional density and distribution objects we can
  specify kernels for the left-hand side and right-hand side variables
  in this manner using \code{cykertype="gaussian"},
  \code{cxkertype="gaussian"} and \code{uykertype="aitchisonaitken"},
  \code{uxkertype="aitchisonaitken"}).

  Note that higher order continuous kernels (i.e., fourth, sixth, and
  eighth order) are derived from the second order kernels given above
  (see Li and Racine (2007) for details).

  For particulars on any given method, kindly see the references listed
  for the method in question.

}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}
  \code{\link{np.options}}
}

\author{
  Tristen Hayfield <tristen.hayfield@gmail.com>, Jeffrey S. Racine
  <racinej@mcmaster.ca>

  Maintainer: Jeffrey S. Racine <racinej@mcmaster.ca>

  We are grateful to John Fox and Achim Zeleis for their valuable input
  and encouragement. We would like to gratefully acknowledge support
  from the Natural Sciences and Engineering Research Council of Canada
  (NSERC:www.nserc.ca), the Social Sciences and Humanities Research
  Council of Canada (SSHRC:www.sshrc.ca), and the Shared Hierarchical
  Academic Research Computing Network (SHARCNET:www.sharcnet.ca)

  
}
\references{

  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
  discrimination by the kernel method,} Biometrika, 63, 413-420.

  Hall, P. and J.S. Racine and Q. Li (2004), \dQuote{Cross-validation
  and the estimation of conditional probability densities,} Journal of
  the American Statistical Association, 99, 1015-1026.

  Li, Q. and J. Lin and J.S. Racine (2013), \dQuote{Optimal bandwidth
  selection for nonparametric conditional distribution and quantile
  functions}, Journal of Business and Economic Statistics, 31, 57-65.

  Li, Q. and D. Ouyang and J.S. Racine (2013), \dQuote{Categorical
  Semiparametric Varying-Coefficient Models,} Journal of Applied
  Econometrics, 28, 551-589.

  Li, Q. and J.S. Racine (2003), \dQuote{Nonparametric estimation of
  distributions with categorical and continuous data,} Journal of
  Multivariate Analysis, 86, 266-292.

  Li, Q. and J.S. Racine (2004), \dQuote{Cross-validated local linear
  nonparametric regression,} Statistica Sinica, 14, 485-512.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics:
  Theory and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2010), \dQuote{Smooth varying-coefficient
  estimation and inference for qualitative and quantitative data,}
  Econometric Theory, 26, 1-31.

  Ouyang, D. and Q. Li and J.S. Racine (2006), \dQuote{Cross-validation
  and the estimation of probability distributions with categorical
  data,} Journal of Nonparametric Statistics, 18, 69-100.

  Racine, J.S. and Q. Li (2004), \dQuote{Nonparametric estimation of
	regression functions with both categorical and continuous data,}
	Journal of Econometrics, 119, 99-130.

  Racine, J.S., Q. Li, and K.X. Yan (2020), \dQuote{Kernel Smoothed
  Probability Mass Functions for Ordered Datatypes,} Journal of
  Nonparametric Statistics, 32(3), 563-586.
  
  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,}
  Cambridge University Press.

  Scott, D.W. (1992), \emph{Multivariate Density Estimation: Theory,
  Practice and Visualization,} New York: Wiley.

  Silverman, B.W. (1986), \emph{Density Estimation,} London: Chapman and
  Hall.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth
  estimators for discrete distributions,} Biometrika, 68, 301-309.

}
\keyword{package}

\name{npuniden.sc}
\alias{npuniden.sc}

\title{
  Kernel Shape Constrained Bounded Univariate Density Estimation
}
\description{

  \code{npuniden.sc} computes shape constrained kernel univariate
  unconditional density estimates given a vector of continuously
  distributed training data and a bandwidth. Lower and upper bounds
  [\code{a},\code{b}] can be supplied (default is [0,1]) and if \code{a}
  is set to \code{-Inf} there is only one bound on the right, while if
  \code{b} is set to \code{Inf} there is only one bound on the left.

}
\usage{
npuniden.sc(X = NULL,
            Y = NULL,
            h = NULL,
            a = 0,
            b = 1,
            lb = NULL,
            ub = NULL,
            extend.range = 0,
            num.grid = 0,
            function.distance = TRUE,
            integral.equal = FALSE,
            constraint = c("density",
                           "mono.incr",
                           "mono.decr",
                           "concave",
                           "convex",
                           "log-concave",
                           "log-convex"))
}
%- maybe also 'usage' for other objects documented here.
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify evaluation points, observations, support bounds, and optional bandwidths.
}

  \item{a}{
an optional lower bound on the support of \code{X} or \code{Y} (defaults to 0)
}

  \item{b}{
an optional upper bound on the support of \code{X} or \code{Y} (defaults to 1)
}


  \item{h}{
a bandwidth (\eqn{>0})
}

  \item{X}{
a required numeric vector of training data lying in \eqn{[a,b]}
}

  \item{Y}{
an optional numeric vector of evaluation data lying in \eqn{[a,b]}
}

\subsection{Density Bounds}{
These arguments set optional lower and upper density bounds used with \code{constraint = \"density\"}.
}

\item{lb}{
  a scalar lower bound (\eqn{\ge 0}) to be used in conjunction with
  \code{constraint="density"}
}

\item{ub}{
  a scalar upper bound (\eqn{\ge 0} and \eqn{\ge} \code{lb}) to be used
  in conjunction with \code{constraint="density"}
}


\subsection{Grid And Distance Controls}{
These arguments control grid construction, distance metric, and mass preservation.
}

\item{extend.range}{
  number specifying the fraction by which the range of the training data
  should be extended for the additional grid points (passed to the
  function \code{extendrange})
}

\item{function.distance}{
  a logical value that, if \code{TRUE}, minimizes the squared deviation between
  the constrained and unconstrained estimates, otherwise, minimizes the
  squared deviation between the constrained and unconstrained weights
}

\item{integral.equal}{
  a logical value, that, if \code{TRUE}, adjusts the constrained estimate
  to have the same probability mass over the range \code{X}, \code{Y},
  and the additional grid points
}


\item{num.grid}{
  number of additional grid points (in addition to \code{X} and
  \code{Y}) placed on an equi-spaced grid spanning
  \code{extendrange(c(X,Y),f=extend.range)} (if \code{num.grid=0} no
  additional grid points will be used regardless of the value of
  \code{extend.range})
}

\subsection{Shape Constraint}{
This argument chooses the monotonicity, convexity, or log-shape constraint.
}

\item{constraint}{
  a character string indicating whether the estimate is to be
  constrained to be monotonically increasing
  (\code{constraint="mono.incr"}), decreasing
  (\code{constraint="mono.decr"}), convex (\code{constraint="convex"}),
  concave (\code{constraint="concave"}), log-convex
  (\code{constraint="log-convex"}), or log-concave
  (\code{constraint="log-concave"})
}

}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.

  Typical usages are (see below for a complete list of options and also
  the examples at the end of this help file)
  \preformatted{
    model <- npuniden.sc(X,a=-2,b=3)
  }

  \code{npuniden.sc} implements a methods for estimating a univariate
  density function defined over a continuous random variable in the
  presence of bounds subject to a variety of shape constraints.  The
  bounded estimates use the truncated Gaussian kernel function.

  Note that for the log-constrained estimates, the derivative estimate
  returned is that for the log-constrained estimate not the non-log
  value of the estimate returned by the function. See Example 5 below
  hat manually plots the log-density and returned derivative (no
  transformation is needed when plotting the density estimate itself).

  If the quadratic program solver fails to find a solution, the
  unconstrained estimate is returned with an immediate warning. Possible
  causes to be investigated are undersmoothing, sparsity, and the
  presence of non-sample grid points. To investigate the possibility of
  undersmoothing try using a larger bandwidth, to investigate sparsity
  try decreasing \code{extend.range}, and to investigate non-sample grid
  points try setting \code{num.grid} to \code{0}.

  Mean square error performance seems to improve generally when using
  additional grid points in the empirical support of \code{X} and
  \code{Y} (i.e., in the observed range of the data sample) but appears
  to deteriorate when imposing constraints beyond the empirical support
  (i.e., when \code{extend.range} is positive). Increasing the number of
  additional points beyond a hundred or so appears to have a limited
  impact.

  The option \code{function.distance=TRUE} appears to perform better for
  imposing convexity, concavity, log-convexity and log-concavity, while
  \code{function.distance=FALSE} appears to perform better for imposing
  monotonicity, whether increasing or decreasing (based on simulations
  for the Beta(s1,s2) distribution with sample size \eqn{n=100}).

}
\value{
A list with the following elements:
\item{f }{unconstrained density estimate}
\item{f.sc }{shape constrained density estimate}
\item{se.f }{asymptotic standard error of the unconstrained density estimate}
\item{se.f.sc }{asymptotic standard error of the shape constrained density estimate}
\item{f.deriv }{unconstrained derivative estimate (of order 1 or 2 or log thereof)}
\item{f.sc.deriv }{shape constrained derivative estimate (of order 1 or 2 or log thereof)}
\item{F }{unconstrained distribution estimate}
\item{F.sc }{shape constrained distribution estimate}
\item{integral.f}{the integral of the unconstrained estimate over \code{X}, \code{Y}, and the additional grid points}
\item{integral.f.sc}{the integral of the constrained estimate over \code{X}, \code{Y}, and the additional grid points}
\item{solve.QP }{logical, if \code{TRUE} \code{solve.QP} succeeded, otherwise failed}
\item{attempts }{number of attempts when \code{solve.QP} fails (max = 9)}
}
\references{
  Du, P. and C. Parmeter and J. Racine (2024), \dQuote{Shape Constrained Kernel PDF and PMF Estimation}, Statistica Sinica, 34 (1), 257-289, \doi{10.5705/ss.202021.0112}
}
\author{
Jeffrey S. Racine \email{racinej@mcmaster.ca}
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}
  The \pkg{logcondens}, \pkg{LogConDEAD}, and \pkg{scdensity} packages,
  and the function \code{\link{npuniden.boundary}}.
}
\examples{
\dontrun{
n <- 100
set.seed(42)

## Example 1: N(0,1), constrain the density to lie within lb=.1 and ub=.2

X <- sort(rnorm(n))
h <- npuniden.boundary(X,a=-Inf,b=Inf)$h
foo <- npuniden.sc(X,h=h,constraint="density",a=-Inf,b=Inf,lb=.1,ub=.2)
ylim <- range(c(foo$f.sc,foo$f))
if (interactive()) plot(X,foo$f.sc,type="l",ylim=ylim,xlab="X",ylab="Density")
lines(X,foo$f,col=2,lty=2)
rug(X)
legend("topleft",c("Constrained","Unconstrained"),lty=1:2,col=1:2,bty="n")

## Example 2: Beta(5,1), DGP is monotone increasing, impose valid
## restriction

X <- sort(rbeta(n,5,1))
h <- npuniden.boundary(X)$h

foo <- npuniden.sc(X=X,h=h,constraint=c("mono.incr"))

oldpar <- par(no.readonly = TRUE)
on.exit(par(oldpar), add = TRUE)
par(mfrow=c(1,2))
ylim <- range(c(foo$f.sc,foo$f))
if (interactive()) plot(X,foo$f.sc,type="l",ylim=ylim,xlab="X",ylab="Density")
lines(X,foo$f,col=2,lty=2)
rug(X)
legend("topleft",c("Constrained","Unconstrained"),lty=1:2,col=1:2,bty="n")

ylim <- range(c(foo$f.sc.deriv,foo$f.deriv))
if (interactive()) plot(X,foo$f.sc.deriv,type="l",ylim=ylim,xlab="X",ylab="First Derivative")
lines(X,foo$f.deriv,col=2,lty=2)
abline(h=0,lty=2)
rug(X)
legend("topleft",c("Constrained","Unconstrained"),lty=1:2,col=1:2,bty="n")

## Example 3: Beta(1,5), DGP is monotone decreasing, impose valid
## restriction

X <- sort(rbeta(n,1,5))
h <- npuniden.boundary(X)$h

foo <- npuniden.sc(X=X,h=h,constraint=c("mono.decr"))

par(mfrow=c(1,2))
ylim <- range(c(foo$f.sc,foo$f))
if (interactive()) plot(X,foo$f.sc,type="l",ylim=ylim,xlab="X",ylab="Density")
lines(X,foo$f,col=2,lty=2)
rug(X)
legend("topleft",c("Constrained","Unconstrained"),lty=1:2,col=1:2,bty="n")

ylim <- range(c(foo$f.sc.deriv,foo$f.deriv))
if (interactive()) plot(X,foo$f.sc.deriv,type="l",ylim=ylim,xlab="X",ylab="First Derivative")
lines(X,foo$f.deriv,col=2,lty=2)
abline(h=0,lty=2)
rug(X)
legend("topleft",c("Constrained","Unconstrained"),lty=1:2,col=1:2,bty="n")
 
## Example 4: N(0,1), DGP is log-concave, impose invalid concavity
## restriction

X <- sort(rnorm(n))
h <- npuniden.boundary(X,a=-Inf,b=Inf)$h

foo <- npuniden.sc(X=X,h=h,a=-Inf,b=Inf,constraint=c("concave"))

par(mfrow=c(1,2))
ylim <- range(c(foo$f.sc,foo$f))
if (interactive()) plot(X,foo$f.sc,type="l",ylim=ylim,xlab="X",ylab="Density")
lines(X,foo$f,col=2,lty=2)
rug(X)
legend("topleft",c("Constrained","Unconstrained"),lty=1:2,col=1:2,bty="n")
ylim <- range(c(foo$f.sc.deriv,foo$f.deriv))

if (interactive()) plot(X,foo$f.sc.deriv,type="l",ylim=ylim,xlab="X",ylab="Second Derivative")
lines(X,foo$f.deriv,col=2,lty=2)
abline(h=0,lty=2)
rug(X)
legend("topleft",c("Constrained","Unconstrained"),lty=1:2,col=1:2,bty="n")

## Example 45: Beta(3/4,3/4), DGP is convex, impose valid restriction

X <- sort(rbeta(n,3/4,3/4))
h <- npuniden.boundary(X)$h

foo <- npuniden.sc(X=X,h=h,constraint=c("convex"))

par(mfrow=c(1,2))
ylim <- range(c(foo$f.sc,foo$f))
if (interactive()) plot(X,foo$f.sc,type="l",ylim=ylim,xlab="X",ylab="Density")
lines(X,foo$f,col=2,lty=2)
rug(X)
legend("topleft",c("Constrained","Unconstrained"),lty=1:2,col=1:2,bty="n")

ylim <- range(c(foo$f.sc.deriv,foo$f.deriv))
if (interactive()) plot(X,foo$f.sc.deriv,type="l",ylim=ylim,xlab="X",ylab="Second Derivative")
lines(X,foo$f.deriv,col=2,lty=2)
abline(h=0,lty=2)
rug(X)
legend("topleft",c("Constrained","Unconstrained"),lty=1:2,col=1:2,bty="n")

## Example 6: N(0,1), DGP is log-concave, impose log-concavity
## restriction

X <- sort(rnorm(n))
h <- npuniden.boundary(X,a=-Inf,b=Inf)$h

foo <- npuniden.sc(X=X,h=h,a=-Inf,b=Inf,constraint=c("log-concave"))

par(mfrow=c(1,2))

ylim <- range(c(log(foo$f.sc),log(foo$f)))
if (interactive()) plot(X,log(foo$f.sc),type="l",ylim=ylim,xlab="X",ylab="Log-Density")
lines(X,log(foo$f),col=2,lty=2)
rug(X)
legend("topleft",c("Constrained-log","Unconstrained-log"),lty=1:2,col=1:2,bty="n")

ylim <- range(c(foo$f.sc.deriv,foo$f.deriv))
if (interactive()) plot(X,
                        foo$f.sc.deriv,
                        type="l",
                        ylim=ylim,
                        xlab="X",
                        ylab="Second Derivative of Log-Density")
lines(X,foo$f.deriv,col=2,lty=2)
abline(h=0,lty=2)
rug(X)
legend("topleft",c("Constrained-log","Unconstrained-log"),lty=1:2,col=1:2,bty="n")
} % enddontrun
}
\keyword{ nonparametric }
\keyword{ smooth }

%
\name{npconmode}
\alias{npconmode}
\alias{npconmode.default}
\alias{npconmode.formula}
\alias{npconmode.conbandwidth}
\alias{predict.conmode}
\alias{plot.conmode}
\title{ Kernel Modal Regression with Mixed Data Types }
\description{
  \code{npconmode} performs kernel modal regression on mixed data,
  and finds the
  conditional mode given a set of training data, consisting of
  explanatory data and dependent data, and possibly evaluation data.
  Automatically computes various in sample and out of sample measures of
  accuracy.
}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels,
  \code{\link{np.options}} for global options, and
  \code{\link{plot}} for plotting options.

  If \code{bws} is not an explicit \code{conbandwidth} object,
  \code{npconmode} computes conditional density bandwidths by forwarding the
  call to \code{\link{npcdensbw}}. The resulting \code{conbandwidth} object is
  stored unchanged in the returned object's \code{bws} component. Regression
  type, local-polynomial degree, and NOMAD/search metadata are also mirrored on
  the returned \code{conmode} object for convenient inspection and summary
  reporting; the bandwidth object remains the canonical source.

  For non-local-constant conditional-mode fits, \code{npconmode} constructs
  the full set of fitted probabilities over the discrete response support and
  projects them onto the probability simplex before selecting and reporting
  the modal outcome. Thus the probabilities used for modal selection are
  non-negative and sum to one over the discrete support. For binary outcomes
  this is the complement contract
  \eqn{\Pr(Y=\ell_2\mid X=x)=1-\Pr(Y=\ell_1\mid X=x)}. Local-constant fits
  are already proper by construction, so omitted \code{proper} resolves to
  \code{FALSE} for \code{regtype="lc"} and to \code{TRUE} otherwise.

  The \code{\link{predict}} method follows the usual S3
  \code{newdata} convention. With no evaluation arguments,
  \code{predict(fit)} extracts the stored modal class and
  \code{predict(fit, type="prob")} extracts stored class probabilities
  when the object was fitted with \code{probabilities=TRUE}. With
  \code{newdata}, \code{predict} evaluates the conditional mode at the
  supplied rows using the stored bandwidth object. Use
  \code{type="prob"} to return the full matrix of fitted class
  probabilities at the evaluation rows. Use \code{se.fit=TRUE} with
  \code{type="prob"} to return a list containing the class-probability matrix
  and the matching asymptotic standard-error matrix. Native evaluator
  arguments \code{exdat} and \code{eydat} remain available for advanced
  workflows and take precedence if supplied.
}

\usage{
npconmode(bws, 
          \dots)

\method{npconmode}{formula}(bws, 
          data = NULL,
          newdata = NULL, 
          \dots)

\method{npconmode}{default}(bws, 
          txdat, 
          tydat, 
          nomad = FALSE,
          proper = NULL,
          proper.control = list(),
          probabilities = FALSE,
          gradients = FALSE,
          level = NULL,
          \dots)

\method{npconmode}{conbandwidth}(bws,
          txdat = stop("invoked without training data 'txdat'"),
          tydat = stop("invoked without training data 'tydat'"),
          exdat,
          eydat,
          proper = NULL,
          proper.control = list(),
          probabilities = FALSE,
          gradients = FALSE,
          level = NULL,
          \dots)

\method{predict}{conmode}(object,
          newdata = NULL,
          type = c("class", "prob"),
          se.fit = FALSE,
          \dots)

\method{plot}{conmode}(x, \dots)

}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the bandwidth specification, formula/data interface, and training data.
}

  \item{bws}{
    a bandwidth specification. This can be set as a \code{conbandwidth}
    object returned from an invocation of \code{\link{npcdensbw}}
  }



  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(bws)}, typically the environment from which
    \code{\link{npcdensbw}} was called.
  }


  \item{txdat}{
    a \eqn{p}-variate data frame of explanatory data (conditioning data) used to
    calculate the regression estimators. Defaults to the training data used to
    compute the bandwidth object.
  }


  \item{tydat}{
    a one (1) dimensional vector of unordered or ordered factors,
    containing the dependent data. Defaults to the training data used to
    compute the bandwidth object.
  }

  \item{object}{
    an object of class \code{"conmode"} returned by \code{npconmode}.
  }

  \item{x}{
    an object of class \code{"conmode"} returned by \code{npconmode}.
  }


\subsection{Evaluation Data}{
These arguments control where the conditional mode is evaluated.
}

  \item{exdat}{
    a \eqn{p}-variate data frame of points on which the regression will be
    estimated (evaluation data). By default,
    evaluation takes place on the data provided by \code{txdat}.
  }


  \item{eydat}{
    a one (1) dimensional numeric or integer vector of the true values
    (outcomes) of the dependent variable. By default,
    evaluation takes place on the data provided by \code{tydat}.
  }



  \item{newdata}{
    An optional data frame in which to look for evaluation data. If
    omitted, the training data are used.
  }

  \item{type}{
    prediction type for \code{predict.conmode}. Use \code{"class"} to
    return modal class predictions and \code{"prob"} to return the full
    class-probability matrix. Probability prediction requires either a
    stored probability matrix or an evaluation path with
    \code{probabilities=TRUE}.
  }

  \item{se.fit}{
    a logical value for \code{predict.conmode}. If \code{TRUE} and
    \code{type="prob"}, return a list with components \code{fit} and
    \code{se.fit}, containing the class-probability matrix and matching
    asymptotic standard-error matrix. Standard errors are for the smooth
    class-probability target, not for the discrete modal class label.
  }


\subsection{Probability And Gradient Controls}{
These arguments control stored class probabilities, proper probability
normalization, and level-specific class-probability effects.
}

  \item{probabilities}{
    a logical value. If \code{TRUE}, store the full matrix of fitted
    probabilities over the discrete response support in the returned object.
    The default \code{FALSE} keeps returned objects compact. If
    \code{gradients=TRUE}, probabilities are stored automatically because
    they are the object-fed target for \code{\link{plot.conmode}}.
  }

  \item{gradients}{
    a logical value. If \code{TRUE}, compute and store class-probability
    gradients/effects for one response level with respect to the conditioning
    variables. These are derivatives for continuous conditioning variables and
    the corresponding smooth finite-change effects for discrete conditioning
    variables as produced by the underlying \code{\link{npcdens}} gradient
    route. The target is the smooth class probability
    \eqn{\Pr(Y=\ell\mid X=x)}, not the derivative of the discrete modal class
    label.
  }

  \item{level}{
    response level for class-probability gradients/effects and
    \code{plot.conmode} class-probability displays. If omitted, the default is
    the base/reference response level, \code{levels(y)[1]}, for both factor and
    ordered responses. Users should supply \code{level} explicitly when a
    particular binary or multinomial outcome probability is of substantive
    interest.
  }

  \item{proper}{
    a logical value or \code{NULL}. If \code{TRUE}, fitted probabilities over
    the discrete response support are projected onto the probability simplex
    before modal selection. If \code{FALSE}, legacy raw per-level values are
    used. If \code{NULL}, the default is estimator-aware:
    \code{FALSE} for \code{regtype="lc"} and \code{TRUE} otherwise.
  }

  \item{proper.control}{
    a list of controls for discrete probability properization. Currently
    \code{tol} may be supplied to set the numerical tolerance for
    non-negativity and unit-mass checks.
  }


\subsection{Local-Polynomial Degree And Bandwidth Search}{
This argument controls the recommended automatic local-polynomial NOMAD route, which jointly selects continuous polynomial degree and bandwidths when conditional-density bandwidths are computed inside \code{npconmode}.
}

  \item{nomad}{
    logical shortcut passed through to \code{\link{npcdensbw}} when
    bandwidths are computed inside \code{npconmode}. When \code{TRUE},
    the conditional-density bandwidth route fills any missing values
    among \code{regtype}, \code{search.engine}, \code{degree.select},
    \code{bernstein.basis}, \code{degree.min}, \code{degree.max},
    \code{degree.verify}, and \code{bwtype} with the recommended
    automatic local-polynomial degree-and-bandwidth NOMAD preset documented in \code{\link{npcdensbw}}.
    Additional NOMAD tuning arguments such as \code{nomad.nmulti} may
    also be supplied through \code{\dots}; \code{nmulti} remains the
    outer restart count while \code{nomad.nmulti} controls inner
    \code{crs::snomadr()} multistarts within each outer restart.
    Automatic NOMAD degree search requires at least one continuous
    explanatory variable. After fitting, inspect
    \code{fit$bws$nomad.shortcut} on the returned object \code{fit} to
    see the normalized shortcut metadata.
  }

\subsection{Additional Arguments}{
Further arguments are passed to the bandwidth-selection counterpart,
prediction/evaluation route, or plot route as appropriate.
}

  \item{\dots}{
    additional arguments supplied to \code{\link{npcdensbw}} when bandwidths
    are computed internally, or arguments needed to interpret a numeric
    \code{bws} vector. This is where bandwidth-selection controls such
    as \code{bwmethod}, \code{bwtype}, and \code{bwscaling},
    kernel/support controls such as \code{cxkertype}, \code{cykertype},
    \code{cxkerorder}, \code{cykerorder}, \code{cxkerbound}, and
    \code{cykerbound}, categorical kernel controls such as
    \code{uxkertype}, \code{uykertype}, \code{oxkertype}, and
    \code{oykertype}, search controls such as \code{nmulti} and
    \code{scale.factor.search.lower}, and local-polynomial/NOMAD
    controls such as \code{regtype}, \code{degree},
    \code{bernstein.basis}, \code{degree.select}, and
    \code{nomad.nmulti} are supplied. In \code{predict.conmode},
    additional arguments are passed to \code{npconmode} for evaluation
    with the stored bandwidth object; native \code{exdat} and
    \code{eydat} take precedence over \code{newdata}. In
    \code{plot.conmode}, additional arguments control the object-fed
    probability/effect display; common controls include
    \code{gradients}, \code{level}, \code{output}, \code{data_rug},
    \code{layout}, \code{legend}, \code{view}, \code{neval},
    \code{grid_control}, \code{perspective}, \code{plot.vars},
    \code{renderer}, and graphics arguments. The default
    \code{view="sample"} draws the stored fitted probabilities/effects at the
    fitted evaluation points. The \code{view="fixed"} route constructs
    one-dimensional object-fed slices over each conditioning variable, holding
    the remaining variables at their \code{uocquantile} median/mode values;
    \code{np_grid_control(xtrim=..., xq=...)} may be used to adjust the slice
    range and holdout quantiles. With \code{perspective=TRUE}, the same
    fixed-grid route draws a base-graphics probability surface for one selected
    response level over two continuous conditioning variables; use
    \code{plot.vars} to name the displayed pair when more than two continuous
    conditioning variables are present; \code{renderer="rgl"} requests the
    corresponding interactive surface when \pkg{rgl} is available. The public
    \code{output="both"} spelling is
    accepted as an alias for the historical \code{output="plot-data"}
    behavior. For class-probability plots, \code{errors="asymptotic"}
    adds probability-level asymptotic standard errors and intervals for the
    selected response level. These intervals target the smooth probability
    \eqn{\Pr(Y=\ell\mid X=x)}, not the discrete modal class label. Rows whose
    probabilities were materially repaired by simplex projection report
    \code{NA} standard errors rather than treating raw conditional-density
    standard errors as projected-probability standard errors. Bootstrap
    intervals are available for one-dimensional \code{view="sample"} and
    \code{view="fixed"} probability plots and use the stored bandwidth object
    without recomputing bandwidths. Surface asymptotic intervals are available
    as \code{output="data"} payload columns; surface band rendering and
    surface bootstrap intervals are intentionally deferred.
    See \code{\link{npcdensbw}} and \code{\link{plot.np}} for the complete
    bandwidth-selection and plotting argument surfaces.
  }

}
\value{
  \code{npconmode} returns a \code{conmode} object with the following
  components:

  \item{bws}{ the \code{conbandwidth} object used to compute the conditional
    mode }
  \item{conmode}{ a vector of type \code{factor} (or \code{ordered
      factor}) containing the conditional mode at each evaluation
    point}
  \item{condens}{ a vector of numeric type containing the modal density
    estimates at each evaluation point}
  \item{conderr}{ a vector of numeric type containing asymptotic
    standard errors for the modal density estimates at each evaluation point.
    If a row is materially properized by probability projection, this value is
    set to \code{NA} for that row rather than reporting a standard error for
    the unrepaired raw probability.}
  \item{xeval}{ a data frame of evaluation points }
  \item{yeval}{ a vector of type \code{factor} (or \code{ordered
      factor}) containing the actual outcomes, or \code{NA} if not
    provided }
  \item{confusion.matrix}{ the confusion matrix or \code{NA} if outcomes
    are not available }
  \item{CCR.overall}{ the overall correct
    classification ratio, or \code{NA} if outcomes are not available }
  \item{CCR.byoutcome}{ a numeric vector containing the correct
    classification ratio by outcome, or \code{NA} if outcomes are not
    available }
  \item{fit.mcfadden}{ the McFadden-Puig-Kerschner performance measure
    or \code{NA} if outcomes are not available }
  \item{probabilities}{ if requested, a matrix of fitted probabilities over
    the discrete response support }
  \item{probability.levels}{ if \code{probabilities} are stored, the response
    support corresponding to the probability matrix columns }
  \item{probability.errors}{ if \code{probabilities} are stored, a matrix of
    asymptotic standard errors for the raw selected-level conditional
    probability estimates, with rows set to \code{NA} where simplex projection
    materially repaired the fitted probability vector }
  \item{probability.repaired.rows}{ if \code{probabilities} are stored, a
    logical vector identifying rows whose fitted probability vector was
    materially repaired by proper probability projection }
  \item{probability.gradients}{ if \code{gradients=TRUE}, a matrix whose
    entries are class-probability gradients/effects for one selected response
    level, indexed by evaluation row and conditioning variable }
  \item{probability.gradient.level, probability.gradient.names,
    probability.gradient.info}{ metadata describing the response support,
    conditioning variables, and interpretation of
    \code{probability.gradients} }
  \item{proper.requested, proper.applied, proper.info}{ metadata describing
    whether discrete probability properization was requested, whether any row
    was materially projected, and diagnostics for non-negativity and unit-mass
    checks }
  \item{regtype, degree, nomad, search.engine}{ metadata mirrored from
    \code{bws} describing the conditional density regression type,
    local-polynomial degree, whether NOMAD automatic degree search was used,
    and the search engine when available }
  \item{degree.search, nomad.shortcut, nomad.time, powell.time}{ detailed
    search metadata mirrored from \code{bws} when automatic degree search or
    NOMAD/Powell refinement was used }

  The function \code{\link{predict}} may be used to extract conditional mode
  class predictions, while \code{\link{fitted}} extracts the conditional
  density estimates at the conditional mode from the resulting object. The
  function \code{\link{gradients}} extracts
  class-probability gradients/effects when \code{gradients=TRUE}. Also,
  \code{\link{summary}} and \code{\link{plot}} support \code{conmode}
  objects. For \code{plot.conmode}, first fit with
  \code{probabilities=TRUE} so the plot can remain object-fed. The default
  plot displays the fitted probability for the base/reference response level
  \code{levels(y)[1]}; use \code{plot(fit, level=...)} to select another
  outcome and \code{plot(fit, gradients=TRUE)} to display stored
  class-probability effects. The default \code{view="sample"} draws stored
  object-fed probability/effect payloads at the fitted evaluation points. Use
  \code{view="fixed"} and \code{neval} to draw object-fed one-dimensional
  slices over each conditioning variable, with other variables held at their
  median/mode values. Use \code{perspective=TRUE} to draw a base-graphics
  probability surface for one selected response level over two continuous
  conditioning variables, and \code{renderer="rgl"} for the corresponding
  interactive surface. Use \code{errors="asymptotic"} for probability-level
  standard errors and intervals in one-dimensional plots, or with
  \code{output="data"} for surface interval payloads. Bootstrap intervals and
  surface band rendering are not yet implemented for \code{plot.conmode}.
}
\section{Book And Method Pointers}{
  The conditional-mode target is
  \eqn{\arg\max_y \Pr(Y=y\mid X=x)} for a discrete response. In practice
  \code{npconmode} estimates the conditional probability of each
  response support point using \code{\link{npcdensbw}} /
  \code{\link{npcdens}}, optionally projects non-local-constant fitted
  probabilities onto the probability simplex, and then reports the
  modal support point.

  Setting \code{gradients=TRUE} stores class-probability effects for one
  response level. If \code{level} is omitted, the base/reference response level
  \code{levels(y)[1]} is used. These effects are useful for asking how the
  fitted probability of a selected class changes with each covariate; they are
  not gradients of the \code{arg max} classification rule itself.

  For book-length background, see Racine (2019), Chapter 4
  \emph{Conditional Probability Density and Cumulative Distribution
  Functions}, especially the binary and multinomial choice material,
  and Li and Racine (2007), Chapter 5 \emph{Conditional Density
  Estimation} together with Chapter 4 \emph{Kernel Estimation with Mixed
  Data}.
}

\references{
  Aitchison, J. and  C.G.G. Aitken (1976), \dQuote{Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Hall, P. and J.S. Racine and Q. Li (2004), \dQuote{Cross-validation and the
    estimation of conditional probability densities,} Journal of the
  American Statistical Association, 99, 1015-1026.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
  and Practice,} Princeton University Press.

  McFadden, D. and C. Puig and D. Kerschner (1977), \dQuote{Determinants
    of the long-run demand for electricity,} Proceedings of the
  American Statistical Association (Business and Economics Section),
  109-117.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,}
  Cambridge University Press.

  Scott, D.W. (1992), \emph{Multivariate Density Estimation. Theory,
  Practice and Visualization,} New York: Wiley.

  Silverman, B.W. (1986), \emph{Density Estimation,} London: Chapman and
  Hall.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth estimators
    for discrete distributions,}  Biometrika, 68, 301-309.
}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}},
  \code{\link{plot}}, \code{\link{npcdensbw}}.
}

\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): For this example, we use the
# birthweight data taken from the MASS library, and compute a parametric
# logit model and a nonparametric conditional mode model. We then
# compare their confusion matrices and summary measures of
# classification ability.

library("MASS")
data("birthwt")

birthwt$low <- factor(birthwt$low)
birthwt$smoke <- factor(birthwt$smoke)
birthwt$race <- factor(birthwt$race)
birthwt$ht <- factor(birthwt$ht)
birthwt$ui <- factor(birthwt$ui)
birthwt$ftv <- ordered(birthwt$ftv)

with(birthwt, {

# Fit a parametric logit model with low (0/1) as the dependent
# variable and age, lwt, and smoke (0/1) as the covariates

# From ?birthwt
# 'low' indicator of birth weight less than 2.5kg
# 'smoke' smoking status during pregnancy
# 'race' mother's race ('1' = white, '2' = black, '3' = other)
# 'ht' history of hypertension
# 'ui' presence of uterine irritability
# 'ftv' number of physician visits during the first trimester
# 'age' mother's age in years
# 'lwt' mother's weight in pounds at last menstrual period

model.logit <- glm(low~smoke+
                   race+
                   ht+
                   ui+
                   ftv+
                   age+
                   lwt,
                   family=binomial(link=logit))

# Generate the confusion matrix and correct classification ratio

cm <- table(low, ifelse(fitted(model.logit)>0.5, 1, 0))
ccr <- sum(diag(cm))/sum(cm)

# Now do the same with a nonparametric model.  Note - this may take a
# few minutes depending on the speed of your computer...

bw <- npcdensbw(formula=low~smoke+
                race+
                ht+
                ui+
                ftv+
                age+
                lwt,
                data=birthwt)

model.np <- npconmode(bws=bw)

# Compare confusion matrices from the logit and nonparametric model

# Logit

cm
ccr

# Nonparametric
summary(model.np)

# Predict modal classes and fitted class probabilities at selected rows
new.birthwt <- birthwt[1:5, c("smoke", "race", "ht", "ui", "ftv", "age", "lwt")]
predict(model.np, newdata=new.birthwt)
predict(npconmode(bws=bw, probabilities=TRUE),
        newdata=new.birthwt, type="prob")

})

# EXAMPLE 1 (INTERFACE=DATA FRAME): For this example, we use the
# birthweight data taken from the MASS library, and compute a parametric
# logit model and a nonparametric conditional mode model. We then
# compare their confusion matrices and summary measures of
# classification ability.

library("MASS")
data("birthwt")
with(birthwt, {

# Fit a parametric logit model with low (0/1) as the dependent
# variable and age, lwt, and smoke (0/1) as the covariates

# From ?birthwt
# 'low' indicator of birth weight less than 2.5kg
# 'smoke' smoking status during pregnancy
# 'race' mother's race ('1' = white, '2' = black, '3' = other)
# 'ht' history of hypertension
# 'ui' presence of uterine irritability
# 'ftv' number of physician visits during the first trimester
# 'age' mother's age in years
# 'lwt' mother's weight in pounds at last menstrual period

model.logit <- glm(low~factor(smoke)+
                   factor(race)+
                   factor(ht)+
                   factor(ui)+
                   ordered(ftv)+
                   age+
                   lwt,
                   family=binomial(link=logit))

# Generate the confusion matrix and correct classification ratio

cm <- table(low, ifelse(fitted(model.logit)>0.5, 1, 0))
ccr <- sum(diag(cm))/sum(cm)

# Now do the same with a nonparametric model...

X <- data.frame(factor(smoke),
                factor(race),
                factor(ht),
                factor(ui),
                ordered(ftv),
                age,
                lwt)

y <- factor(low)

# Note - this may take a few minutes depending on the speed of your
# computer...

bw <- npcdensbw(xdat=X, ydat=y)

model.np <- npconmode(bws=bw)

# Compare confusion matrices from the logit and nonparametric model

# Logit

cm
ccr

# Nonparametric
summary(model.np)

})

# EXAMPLE 3 (CLASS PROBABILITY EFFECTS): compute and plot
# class-probability gradients/effects for one selected response level.
# This example uses a small artificial sample so that it runs quickly.

set.seed(42)
n <- 100
x <- seq(-1, 1, length.out=n)
y <- factor(rbinom(n, 1, plogis(1.5*x)), levels=0:1)

model.effects <- npconmode(y~x,
                           regtype="ll",
                           bwmethod="cv.ls",
                           nmulti=1,
                           probabilities=TRUE,
                           gradients=TRUE)

gradients(model.effects)
plot(model.effects)
plot(model.effects, view="fixed", neval=25)
plot(model.effects, gradients=TRUE)
plot(model.effects, gradients=TRUE, view="fixed", neval=25)

## A two-continuous-predictor fit can be displayed as a probability surface.
## The default level is the base/reference response level.

z <- runif(n, -1, 1)
y2 <- factor(rbinom(n, 1, plogis(1.5*x - z)), levels=0:1)
model.surface <- npconmode(y2~x+z,
                           regtype="ll",
                           bwmethod="cv.ls",
                           nmulti=1,
                           probabilities=TRUE)
plot(model.surface, perspective=TRUE, neval=15)
} % enddontrun
}
\keyword{ nonparametric }% at least one, from doc/KEYWORDS

%
\name{npcmstest}
\alias{npcmstest}
\title{ Kernel Consistent Model Specification Test with Mixed Data Types }
\description{
  \code{npcmstest} implements a consistent test for correct
  specification of parametric regression models (linear or nonlinear) as
  described in Hsiao, Li, and Racine (2007).
}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels,
  \code{\link{np.options}} for global options, and
  \code{\link{plot}} for plotting options.
}
\usage{
npcmstest(formula,
          data = NULL,
          subset,
          xdat,
          ydat,
          model = stop(paste(sQuote("model")," has not been provided")),
          distribution = c("bootstrap", "asymptotic"),
          boot.method = c("iid","wild","wild-rademacher"),
          boot.num = 399,
          pivot = TRUE,
          density.weighted = TRUE,
          random.seed = 42,
          \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the model formula/data interface and explicit data inputs.
}

  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(formula)}, typically the environment from which the
    function is called.
  }

  \item{formula}{
    a symbolic description of the model to be tested. If \code{xdat}
    and \code{ydat} are omitted, the data are extracted from this
    formula and \code{data}.
  }


  \item{model}{
    a model object obtained from a call to \code{\link{lm}} (or
    \code{\link{glm}}). Important: the
    call to either \code{\link{glm}} or \code{\link{lm}} must have the arguments
    \code{x=TRUE} and
    \code{y=TRUE} or \code{npcmstest} will not work. Also, the test is
    based on residual bootstrapping hence the outcome must be continuous
    (which rules out Logit, Probit, and Count models).
  }


  \item{subset}{
    an optional vector specifying a subset of observations to be used.
  }


  \item{xdat}{
    a \eqn{p}-variate data frame of explanatory data (training data) used to
    calculate the regression estimators.
  }


  \item{ydat}{
    a one (1) dimensional numeric or integer vector of dependent data, each
    element \eqn{i} corresponding to each observation (row) \eqn{i} of
    \code{xdat}.
  }


\subsection{Bootstrap And Test Controls}{
These arguments control the test statistic, bootstrap procedure, and reproducibility settings.
}

  \item{boot.method}{
    a character string used to specify the bootstrap method.
    \code{iid} will generate independent identically distributed
    draws. \code{wild} will use a wild bootstrap. \code{wild-rademacher}
    will use a wild bootstrap with Rademacher variables. Defaults to
    \code{iid}.
  }


  \item{boot.num}{
    an integer value specifying the number of bootstrap replications to
    use. Defaults to \code{399}.
  }


  \item{density.weighted}{
    a logical value specifying whether the statistic should be
    weighted by the density of \code{xdat}. Defaults to \code{TRUE}.
  }


  \item{distribution}{
    a character string used to specify the method of estimating the
    distribution of the statistic to be calculated. \code{bootstrap}
    will conduct bootstrapping. \code{asymptotic} will use the normal
    distribution. Defaults to \code{bootstrap}.
  }


  \item{pivot}{
    a logical value specifying whether the statistic should be
    normalised such that it approaches \eqn{N(0,1)} in
    distribution. Defaults to \code{TRUE}.
  }


  \item{random.seed}{
    an integer used to seed R's random number generator. This is to
    ensure replicability. Defaults to 42.
  }


\subsection{Additional Arguments}{
Further arguments are passed to the bandwidth-selection routines used by the test.
}

  \item{\dots}{
    additional arguments supplied to control bandwidth selection on the
    residuals. One can specify the bandwidth type,
    kernel types, and so on. To do this, you may specify any of \code{bwscaling},
    \code{bwtype}, \code{ckertype}, \code{ckerorder}, \code{ukertype},
    \code{okertype}, as described in \code{\link{npregbw}}.
    This is necessary if you specify \code{bws} as a \eqn{p}-vector and not
    a \code{bandwidth} object, and you do not desire the default behaviours.
  }


}
\value{
  \code{npcmstest} returns an object of type \code{cmstest} with the
  following components, components will contain information
  related to \code{Jn} or \code{In} depending on the value of \code{pivot}:

  \item{Jn}{ the statistic \code{Jn} }
  \item{In}{ the statistic \code{In} }
  \item{Omega.hat}{ as described in Hsiao, C. and Q. Li and J.S. Racine. }
  \item{q.*}{ the various quantiles of the statistic \code{Jn} (or
    \code{In} if
    \code{pivot=FALSE})  are in
    components \code{q.90},
    \code{q.95}, \code{q.99} (one-sided 1\%, 5\%, 10\% critical values) }
  \item{P}{ the P-value of the statistic }
  \item{Jn.bootstrap}{ if \code{pivot=TRUE} contains the bootstrap
    replications of \code{Jn} }
  \item{In.bootstrap}{ if \code{pivot=FALSE} contains the bootstrap
    replications of \code{In} }

  \code{\link{summary}} supports object of type \code{cmstest}.
}
\section{Book And Method Pointers}{
  \code{npcmstest} tests a parametric conditional-mean specification
  against a nonparametric alternative, using residual-based bootstrap or
  asymptotic calibration as selected by \code{distribution}. The
  statistic is designed for continuous outcomes and fitted
  \code{\link{lm}} or \code{\link{glm}} objects whose model frame, model
  matrix, and response are available.

  For book-length background, see Li and Racine (2007), Chapter 12
  \emph{Model Specification Tests}, especially Section 12.1 and Section
  12.1.2. For the conditional-mean estimation context used by the test,
  see Racine (2019), Chapter 6 \emph{Conditional Mean Function Estimation}.
}

\references{
  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Hsiao, C. and Q. Li and J.S. Racine (2007), \dQuote{A consistent
    model specification test with mixed categorical and continuous
    data,} Journal of Econometrics, 140, 802-826.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
    and Practice,} Princeton University Press.

  Maasoumi, E. and J.S. Racine and T. Stengos (2007),
  \dQuote{Growth and convergence: a profile of distribution dynamics and
    mobility,} Journal of Econometrics, 136, 483-508.

  Murphy, K. M. and F. Welch (1990), \dQuote{Empirical age-earnings
    profiles,} Journal of Labor Economics, 8, 202-229.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,}
  Cambridge University Press.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth estimators
    for discrete distributions,}  Biometrika, 68, 301-309.

}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{

  \code{npcmstest} supports regression objects generated by
  \code{\link{lm}} and uses features specific to objects of type
  \code{\link{lm}} hence if you attempt to pass objects of a different
  type the function cannot be expected to work.

  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}},
  \code{\link{plot}}, \code{\link{npregbw}}.
}

\examples{
\dontrun{
# EXAMPLE 1: For this example, we conduct a consistent model
# specification test for a parametric wage regression model that is
# quadratic in age. The work of Murphy and Welch (1990) would suggest
# that this parametric regression model is misspecified.

data("cps71")
with(cps71, {

model <- lm(logwage~age+I(age^2), x=TRUE, y=TRUE)

if (interactive()) plot(age, logwage)
lines(age, fitted(model))

# Note - this may take a few minutes depending on the speed of your
# computer...

npcmstest(model = model, xdat = age, ydat = logwage, boot.num=9, nmulti = 1)

# Next try Murphy & Welch's (1990) suggested quintic specification.

model <- lm(logwage~age+I(age^2)+I(age^3)+I(age^4)+I(age^5), x=TRUE, y=TRUE)

if (interactive()) plot(age, logwage)
lines(age, fitted(model))

X <- data.frame(age)

# Note - this may take a few minutes depending on the speed of your
# computer...

npcmstest(model = model, xdat = age, ydat = logwage, boot.num=9, nmulti = 1)

# Note - you can pass in multiple arguments to this function. For
# instance, to use local linear rather than local constant regression,
# you would use npcmstest(model, X, regtype="ll"), while you could also
# change the kernel type (default is second order Gaussian), numerical
# search tolerance, or feed in your own vector of bandwidths and so
# forth.

})

# EXAMPLE 2: For this example, we replicate the application in Maasoumi,
# Racine, and Stengos (2007) (see oecdpanel for details). We
# estimate a parametric model that is used in the literature, then
# subject it to the model specification test.

data("oecdpanel")
with(oecdpanel, {

model <- lm(growth ~ oecd +
            factor(year) +
            initgdp +
            I(initgdp^2) +
            I(initgdp^3) +
            I(initgdp^4) +
            popgro +
            inv +
            humancap +
            I(humancap^2) +
            I(humancap^3) - 1,
            x=TRUE,
            y=TRUE)

X <- data.frame(factor(oecd), factor(year), initgdp, popgro, inv, humancap)

npcmstest(model = model, xdat = X, ydat = growth, boot.num=9, nmulti = 1)

})
} % enddontrun
}
\keyword{ nonparametric }

\name{npsymtest}
\alias{npsymtest}
\title{ Kernel Consistent Density Asymmetry Test with Mixed Data Types }
\description{
  \code{npsymtest} implements the consistent metric entropy test of
  asymmetry as described in Maasoumi and Racine (2009).
}
\usage{
npsymtest(data = NULL,
          method = c("integration","summation"),
          boot.num = 399,
          bw = NULL,
          boot.method = c("iid", "geom"),
          random.seed = 42,
          \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the data, statistic variant, and any supplied bandwidth.
}

  \item{bw}{
    a numeric (scalar) bandwidth. Defaults to plug-in (see details below).
  }


  \item{data}{
    a vector containing the variable.
  }


  \item{method}{
    a character string used to specify whether to compute the integral
    version or the summation version of the statistic. Can be set as
    \code{integration} or \code{summation} (see below for
    details). Defaults to \code{integration}.
  }


\subsection{Bootstrap Controls}{
These arguments control bootstrap execution and reproducibility settings.
}

  \item{boot.method}{ a character string used to specify the
    bootstrap method. Can be set as \code{iid} or \code{geom} (see below
    for details). Defaults to \code{iid}.
  }


  \item{boot.num}{
    an integer value specifying the number of bootstrap
    replications to use. Defaults to \code{399}.
  }


  \item{random.seed}{
    an integer used to seed R's random number generator. This is to
    ensure replicability. Defaults to 42.
  }


\subsection{Additional Arguments}{
Further arguments are passed to the bandwidth-selection routines used by the test.
}

  \item{\dots}{ additional arguments supplied to specify the bandwidth
    type, kernel types, and so on.  This is used since we specify bw as
    a numeric scalar and not a \code{bandwidth} object, and is of
    interest if you do not desire the default behaviours. To change the
    defaults, you may specify any of \code{bwscaling}, \code{bwtype},
    \code{ckertype}, \code{ckerorder}, \code{ukertype},
    \code{okertype}.}


}
\value{
  \code{npsymtest} returns an object of type \code{symtest} with the
  following components

  \item{Srho}{ the statistic \code{Srho} }
  \item{Srho.bootstrap}{ contains the bootstrap replications of \code{Srho} }
  \item{P}{ the P-value of the statistic }
  \item{boot.num}{ number of bootstrap replications }
  \item{data.rotate}{ the rotated data series }
  \item{bw}{ the numeric (scalar) bandwidth }

  \code{\link{summary}} supports object of type \code{symtest}.
}
\section{Book And Method Pointers}{
  \code{npsymtest} tests symmetry by comparing the density or
  probability function of the observed data with that of the data
  reflected around the relevant center. The same entropy-distance logic
  used by the univariate density tests underlies the statistic.

  For book-length background, see Racine (2019), Chapter 2
  \emph{Continuous Density and Cumulative Distribution Functions} and Li
  and Racine (2007), Chapter 13 \emph{Nonsmoothing Tests} and Chapter 12
  \emph{Model Specification Tests} where related density-feature tests are
  treated.
}

\references{

  Granger, C.W. and E. Maasoumi and J.S. Racine (2004), \dQuote{A
  dependence metric for possibly nonlinear processes}, Journal of Time
  Series Analysis, 25, 649-669.

  Maasoumi, E. and J.S. Racine (2009), \dQuote{A robust entropy-based
  test of asymmetry for discrete and continuous processes,} Econometric
  Reviews, 28, 246-261.

  Politis, D.N. and J.P. Romano (1994), \dQuote{The stationary
    bootstrap,} Journal of the American Statistical Association, 89,
    1303-1313.

}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.


  \code{npsymtest} computes the nonparametric metric entropy (normalized
   Hellinger of Granger, Maasoumi and Racine (2004)) for testing
   symmetry using the densities/probabilities of the data and the
   rotated data, \eqn{D[f(y), f(\tilde y)]}{D[f(y), f(y-tilde)]}. See
   Maasoumi and Racine (2009) for details. Default bandwidths are of the
   plug-in variety (\code{\link{bw.SJ}} for continuous variables and
   direct plug-in for discrete variables).

  For bootstrapping the null distribution of the statistic, \code{iid}
  conducts simple random resampling, while \code{geom} conducts Politis
  and Romano's (1994) stationary bootstrap using automatic block length
  selection via the \code{\link{b.star}} function in the
  \code{\link{np}} package. See the \pkg{boot} package for
  details.

  The summation version of this statistic may be numerically unstable
  when \code{y} is sparse (the summation version involves division of
  densities while the integration version involves differences). Warning
  messages are produced should this occur (\sQuote{integration recommended})
  and should be heeded.

}

\section{Usage Issues}{

  When using data of type \code{\link{factor}} it is crucial that the
  variable not be an alphabetic character string (i.e. the factor must
  be integer-valued). The rotation is conducted about the median after
  conversion to type \code{\link{numeric}} which is then converted back
  to type \code{\link{factor}}. Failure to do so will have unpredictable
  results. See the example below for proper usage.

}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}
  \code{\link{npdeneqtest},\link{npdeptest},\link{npsdeptest},\link{npunitest}}
}

\examples{
\dontrun{
set.seed(1234)

n <- 100

## Asymmetric discrete probability distribution function

x <- factor(rbinom(n,2,.8))
npsymtest(x,boot.num=99)

if (interactive()) Sys.sleep(5)

## Symmetric discrete probability distribution function

x <- factor(rbinom(n,2,.5))
npsymtest(x,boot.num=99)

if (interactive()) Sys.sleep(5)

## Asymmetric continuous distribution function

y <- rchisq(n,df=2)
npsymtest(y,boot.num=99)

if (interactive()) Sys.sleep(5)

## Symmetric continuous distribution function

y <- rnorm(n)
npsymtest(y,boot.num=99)

## Time-series bootstrap

ar.series <- function(phi,epsilon) {
  n <- length(epsilon)
  series <- numeric(n)
  series[1] <- epsilon[1]/(1-phi)
  for(i in 2:n) {
    series[i] <- phi*series[i-1] + epsilon[i]
  }
  return(series)
}

## Asymmetric time-series

yt <- ar.series(0.5,rchisq(n,df=3))
npsymtest(yt,boot.num=99,boot.method="geom")

## Symmetric time-series

yt <- ar.series(0.5,rnorm(n))
npsymtest(yt,boot.num=99,boot.method="geom")

} % enddontrun

}
\keyword{ univar }
\keyword{ nonparametric }

\name{npudistbw}
\alias{npudistbw}
\alias{npudistbw.formula}
\alias{npudistbw.default}
\alias{npudistbw.dbandwidth}
\title{Kernel Distribution Bandwidth Selection with Mixed Data Types}

\description{
  \code{npudistbw} computes a bandwidth object for a \eqn{p}-variate
  kernel cumulative distribution estimator defined over mixed continuous
  and discrete (ordered) data using either the normal reference
  rule-of-thumb or least-squares cross validation using the method of
  Li, Li and Racine (2017).
}

\usage{

npudistbw(\dots)

\method{npudistbw}{formula}(formula, 
          data, 
          subset, 
          na.action, 
          call, 
          gdata = NULL,
          \dots)

\method{npudistbw}{dbandwidth}(dat = stop("invoked without input data 'dat'"),
          bws,
          gdat = NULL,
          bandwidth.compute = TRUE,
          cfac.dir = 2.5*(3.0-sqrt(5)),
          scale.factor.init = 0.5,
          dfac.dir = 0.25*(3.0-sqrt(5)),
          dfac.init = 0.375,
          dfc.dir = 3,
          do.full.integral = FALSE,
          ftol = 1.490116e-07,
          scale.factor.init.upper = 2.0,
          hbd.dir = 1,
          hbd.init = 0.9,
          initc.dir = 1.0,
          initd.dir = 1.0,
          invalid.penalty = c("baseline","dbmax"),
          itmax = 10000,
          lbc.dir = 0.5,
          scale.factor.init.lower = 0.1,
          lbd.dir = 0.1,
          lbd.init = 0.1,
          memfac = 500.0,
          ngrid = 100,
          nmulti,
          penalty.multiplier = 10,
          powell.remin = TRUE,
          scale.init.categorical.sample = FALSE,
          scale.factor.search.lower = NULL,
          small = 1.490116e-05,
          tol = 1.490116e-04,
          transform.bounds = FALSE,
          \dots)

\method{npudistbw}{default}(dat = stop("invoked without input data 'dat'"),
          bws,
          gdat,
          bandwidth.compute = TRUE,
          bwmethod,
          bwscaling,
          bwtype,
          cfac.dir,
          scale.factor.init,
          ckerbound,
          ckerlb,
          ckerorder,
          ckertype,
          ckerub,
          dfac.dir,
          dfac.init,
          dfc.dir,
          do.full.integral,
          ftol,
          scale.factor.init.upper,
          hbd.dir,
          hbd.init,
          initc.dir,
          initd.dir,
          invalid.penalty,
          itmax,
          lbc.dir,
          scale.factor.init.lower,
          lbd.dir,
          lbd.init,
          memfac,
          ngrid,
          nmulti,
          okertype,
          penalty.multiplier,
          powell.remin,
          scale.init.categorical.sample,
          scale.factor.search.lower = NULL,
          small,
          tol,
          transform.bounds,
          \dots) 

}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the data, formula interface, optional integration grid, and whether bandwidths are supplied or computed.
}

  \item{bandwidth.compute}{
    a logical value which specifies whether to do a numerical search for
    bandwidths or not. If set to \code{FALSE}, a \code{bandwidth} object
    will be returned with bandwidths set to those specified
    in \code{bws}. Defaults to \code{TRUE}.
  }
  
  \item{bws}{
    a bandwidth specification. This can be set as a bandwidth object
    returned from a previous invocation, or as a vector of bandwidths,
    with each element \eqn{i} corresponding to the bandwidth for column
    \eqn{i} in \code{dat}. In either case, the bandwidth supplied will
    serve as a starting point in the numerical search for optimal
    bandwidths. If specified as a vector, then additional arguments will
    need to be supplied as necessary to specify the bandwidth type,
    kernel types, selection methods, and so on. This can be left unset.
  }

  \item{call}{
    the original function call. This is passed internally by
    \code{\link{np}} when a bandwidth search has been implied by a call to
    another function. It is not recommended that the user set this.
  }
    
  \item{dat}{
    a \eqn{p}-variate data frame on which bandwidth selection will be
    performed. The data types may be continuous, discrete (ordered
    factors), or some combination thereof.
  }

  \item{data}{
    an optional data frame, list or environment (or object coercible to
    a data frame by \code{\link{as.data.frame}}) containing the
    variables in the model. If not found in data, the variables are
    taken from \code{environment(formula)}, typically the environment
    from which the function is called.
  }

  \item{formula}{
    a symbolic description of variables on which bandwidth selection is
    to be performed. The details of constructing a formula are
    described below.
  }

  \item{gdat}{
    a grid of data on which the indicator function for
    least-squares cross-validation is to be computed (can be the sample
    or a grid of quantiles).
  }

  
  \item{gdata}{
    a grid of data on which the indicator function for
    least-squares cross-validation is to be computed (can be the sample
    or a grid of quantiles).
  }

  \item{na.action}{
    a function which indicates what should happen when the data contain
   \code{NA}s. The default is set by the \code{\link{na.action}} setting
   of options, and is \code{\link{na.fail}} if that is unset. The
   (recommended) default is \code{\link{na.omit}}.
  }
  
  \item{subset}{
    an optional vector specifying a subset of observations to be used in
    the fitting process. 
  }


\subsection{Bandwidth Criterion And Representation}{
These arguments choose the selection criterion and the way continuous bandwidths are represented.
}

  \item{bwmethod}{ a character string specifying the bandwidth selection
    method.  \code{cv.cdf} specifies least-squares cross-validation for
    cumulative distribution functions (Li, Li and Racine (2017)), and
    \code{normal-reference} just computes the \sQuote{rule-of-thumb}
    bandwidth \eqn{h_j}{h[j]} using the standard formula \eqn{h_j =
    1.587 \sigma_j n^{-1/(P+l)}}{h[j] = 1.587*sigma[j]*n^(-1.0/(P+l))},
    where \eqn{\sigma_j}{sigma[j]} is an adaptive measure of spread of
    the \eqn{j}th continuous variable defined as min(standard deviation,
    mean absolute deviation/1.4826, interquartile range/1.349), \eqn{n}
    the number of observations, \eqn{P} the order of the kernel, and
    \eqn{l} the number of continuous variables. Note that when there
    exist factors and the normal-reference rule is used, there is zero
    smoothing of the factors. Defaults to \code{cv.cdf}.  }

  \item{bwscaling}{ a logical value that when set to \code{TRUE} the
    supplied bandwidths are interpreted as \sQuote{scale factors}
    (\eqn{c_j}{c[j]}), otherwise when the value is \code{FALSE} they are
    interpreted as \sQuote{raw bandwidths} (\eqn{h_j}{h[j]} for continuous data
    types, \eqn{\lambda_j}{lambda[j]} for discrete data types). For
    continuous data types, \eqn{c_j}{c[j]} and \eqn{h_j}{h[j]} are
    related by the formula \eqn{h_j = c_j \sigma_j n^{-1/(P+l)}}{h[j] =
    c[j]*sigma[j]*n^(-1/(P+l))}, where \eqn{\sigma_j}{sigma[j]} is an
    adaptive measure of spread of the \eqn{j}th continuous variable
    defined as min(standard deviation, mean absolute deviation/1.4826,
    interquartile range/1.349), \eqn{n} the number of observations,
    \eqn{P}{P} the order of the kernel, and \eqn{l}{l} the number of
    continuous variables. For discrete data types, \eqn{c_j}{c[j]} and
    \eqn{h_j}{h[j]} are related by the formula \eqn{h_j =
    c_jn^{-2/(P+l)}}{h[j] = c[j]*n^(-2/(P+l))}, where here \eqn{j}{[j]}
    denotes discrete variable \eqn{j}{j}.  Defaults to \code{FALSE}.  }

  \item{bwtype}{
    character string used for the continuous variable bandwidth type,
    specifying the type of bandwidth to compute and return in the
    \code{bandwidth} object. Defaults to \code{fixed}. Option
    summary:\cr
    \code{fixed}: compute fixed bandwidths \cr
    \code{generalized_nn}: compute generalized nearest neighbors \cr
    \code{adaptive_nn}: compute adaptive nearest neighbors
    
  }


\subsection{Categorical Search Initialization}{
These controls set categorical search starts and categorical direction-set initialization.
}

  \item{dfac.dir}{stretch factor for direction set search for Powell's algorithm for categorical variables. See Details}

  \item{dfac.init}{non-random initial values for scale factors for categorical variables for Powell's algorithm. See Details}

  \item{hbd.dir}{upper bound for direction set search for Powell's algorithm for categorical variables. See Details}

  \item{hbd.init}{upper bound for scale factors for categorical variables for Powell's algorithm. See Details}

  \item{initd.dir}{initial non-random values for direction set search for Powell's algorithm for categorical variables. See Details}

  \item{lbd.dir}{lower bound for direction set search for Powell's algorithm for categorical variables. See Details}

  \item{lbd.init}{lower bound for scale factors for categorical variables for Powell's algorithm. See Details}

  \item{scale.init.categorical.sample}{ a logical value that when set
    to \code{TRUE} scales \code{lbd.dir}, \code{hbd.dir},
    \code{dfac.dir}, and \code{initd.dir} by \eqn{n^{-2/(2P+l)}},
    \eqn{n} the number of observations, \eqn{P}{P} the order of the
    kernel, and \eqn{l}{l} the number of \code{numeric} variables. See
    Details}


\subsection{Continuous Direction-Set Search Controls}{
These controls set Powell direction-set initialization for continuous variables.
}

  \item{cfac.dir}{stretch factor for direction set search for Powell's algorithm for \code{numeric} variables. See Details}

  \item{dfc.dir}{chi-square degrees of freedom for direction set search for Powell's algorithm for \code{numeric} variables. See Details}

  \item{initc.dir}{initial non-random values for direction set search for Powell's algorithm for \code{numeric} variables. See Details}

  \item{lbc.dir}{lower bound for direction set search for Powell's algorithm for \code{numeric} variables. See Details}


\subsection{Continuous Kernel Support Controls}{
These controls choose and parameterize bounded support for continuous kernels.
}

  \item{ckerbound}{
    character string controlling continuous-kernel support handling.
    Can be set as \code{none} (default kernel on full support),
    \code{range} (use sample min/max), or \code{fixed} (use
    \code{ckerlb}/\code{ckerub}). The bounded-kernel route reuses the selected continuous kernel and renormalizes it on the chosen support; see \code{\link{np.kernels}}.
  }

  \item{ckerlb}{
    numeric scalar/vector of lower bounds for continuous variables used
    when \code{ckerbound="fixed"}. Must satisfy lower-bound validity for
    each continuous variable (e.g., \code{<= min(variable)}). Use \code{-Inf} for unbounded below. See \code{\link{np.kernels}} for bounded-kernel normalization details.
  }

  \item{ckerub}{
    numeric scalar/vector of upper bounds for continuous variables used
    when \code{ckerbound="fixed"}. Must satisfy upper-bound validity for
    each continuous variable (e.g., \code{>= max(variable)}). Use \code{Inf} for unbounded above. See \code{\link{np.kernels}} for bounded-kernel normalization details.
  }


\subsection{Continuous Scale-Factor Search Initialization}{
These controls define deterministic and random continuous scale-factor starts and the lower admissibility floor for fixed-bandwidth search.
}

  \item{scale.factor.init}{
    deterministic initial scale factor for continuous fixed-bandwidth
    search. Defaults to \code{0.5}. The value supplied by the user is
    not rewritten, but the effective first start passed to the optimizer
    is \code{max(scale.factor.init, scale.factor.search.lower)}. See
    Details.
  }
  \item{scale.factor.init.lower}{
    lower endpoint for random continuous scale-factor starts. Defaults
    to \code{0.1}. The value supplied by the user is not rewritten, but
    the effective random-start lower endpoint is
    \code{max(scale.factor.init.lower, scale.factor.search.lower)}. See
    Details.
  }
  \item{scale.factor.init.upper}{
    upper endpoint for random continuous scale-factor starts. Defaults
    to \code{2.0}. It must be greater than or equal to the effective
    lower endpoint,
    \code{max(scale.factor.init.lower, scale.factor.search.lower)};
    otherwise bandwidth search errors rather than silently expanding the
    interval. See Details.
  }
  \item{scale.factor.search.lower}{
    optional nonnegative scalar giving the hard lower admissibility
    bound for continuous fixed-bandwidth search candidates. Defaults to
    \code{NULL}. If \code{NULL}, an existing bandwidth object's stored
    value is inherited when available; otherwise the package default
    \code{0.1} is used. This floor applies to computed/search bandwidth
    candidates and to effective search starts only. It does not rewrite
    explicit bandwidths supplied for storage with
    \code{bandwidth.compute = FALSE}. Final fixed-bandwidth search
    candidates must also have a finite valid raw objective value.
  }

\subsection{Distribution Integral And Grid Controls}{
These controls tune the distribution-function integral and grid calculations.
}

  \item{do.full.integral}{
    a logical value which when set as \code{TRUE} evaluates the
    moment-based integral on the entire sample. Defaults
    to \code{FALSE}.
  }

  \item{memfac}{
    The algorithm to compute the least-squares objective function uses
    a block-based algorithm to eliminate or minimize redundant kernel
    evaluations. Due to memory, hardware and software constraints, a
    maximum block size must be imposed by the algorithm. This block size
    is roughly equal to memfac*10^5 elements. Empirical tests on
    modern hardware find that a memfac of 500 performs well. If
    you experience out of memory errors, or strange behaviour for
    large data sets (>100k elements) setting memfac to a lower value may
    fix the problem.
  }

  \item{ngrid}{
    integer number of grid points to use when computing the moment-based
    integral. Defaults to \code{100}.
  }


\subsection{Kernel Type Controls}{
These controls choose continuous, unordered, and ordered kernels.
}

  \item{ckerorder}{
    numeric value specifying kernel order (one of
    \code{(2,4,6,8)}). Kernel order specified along with a
  \code{uniform} continuous kernel type will be ignored. Defaults to
  \code{2}.
  }

  \item{ckertype}{
    character string used to specify the continuous kernel type.
    Can be set as \code{gaussian}, \code{epanechnikov}, or
    \code{uniform}. Defaults to \code{gaussian}.
  }

  \item{okertype}{
    character string used to specify the ordered categorical kernel type.
    Can be set as \code{wangvanryzin}, \code{liracine}, or
    \code{racineliyan}. Defaults to \code{liracine}.
  }

\subsection{Numerical Search And Tolerance Controls}{
These controls set optimizer tolerances, restart behavior, invalid-candidate penalties, and bounded search transformations.
}

  \item{ftol}{
    fractional tolerance on the value of the cross-validation function
    evaluated at located minima (of order the machine precision or
    perhaps slightly larger so as not to be diddled by
    roundoff). Defaults to \code{1.490116e-07}
    (1.0e+01*sqrt(.Machine$double.eps)).
  }

  \item{invalid.penalty}{ a character string specifying the penalty
    used when the optimizer encounters invalid bandwidths.
    \code{"baseline"} returns a finite penalty based on a baseline
    objective; \code{"dbmax"} returns \code{DBL\_MAX}. Defaults to
    \code{"baseline"}.}

  \item{itmax}{
    integer number of iterations before failure in the numerical
    optimization routine. Defaults to \code{10000}.
  }

  \item{nmulti}{
    integer number of times to restart the process of finding extrema of
    the cross-validation function from different (random) initial points.
  }

  \item{penalty.multiplier}{ a numeric multiplier applied to the
    baseline penalty when \code{invalid.penalty="baseline"}.
    Defaults to \code{10}.}

  \item{powell.remin}{
    logical flag controlling Powell restart-from-minimum behavior. When
    \code{TRUE}, the Powell-style search restarts from the located minimum
    for a minor gain in accuracy. Defaults to \code{TRUE}.
  }

  \item{small}{
    a small number used to bracket a minimum (it is hopeless to ask for
    a bracketing interval of width less than sqrt(epsilon) times its
    central value, a fractional width of only about 10-04 (single
    precision) or 3x10-8 (double precision)). Defaults to \code{small
      = 1.490116e-05 (1.0e+03*sqrt(.Machine$double.eps))}.
  }

  \item{tol}{
    tolerance on the position of located minima of the cross-validation
    function (tol should generally be no smaller than the square root of
    your machine's floating point precision). Defaults to \code{
      1.490116e-04 (1.0e+04*sqrt(.Machine$double.eps))}.
  }

  \item{transform.bounds}{ a logical value that when set to \code{TRUE}
    applies an internal transformation that maps the unconstrained
    search to the feasible bandwidth domain. Defaults to
    \code{FALSE}.}


\subsection{Additional Arguments}{
These arguments collect remaining controls passed through S3 methods.
}

  \item{\dots}{
    additional arguments supplied to specify the bandwidth type,
    kernel types, selection methods, and so on, detailed below.
  }

}

\details{

The \code{scale.factor.*} controls are dimensionless search
controls. The package converts scale factors to bandwidths using the
estimator-specific scaling encoded in the bandwidth object, including
kernel order and the number of continuous variables relevant for the
estimator. Users should not pre-multiply these controls by sample-size
or standard-deviation factors.

  \code{scale.factor.init} controls the deterministic first search
  start. \code{scale.factor.init.lower} and
  \code{scale.factor.init.upper} define the random multistart interval.
  \code{scale.factor.search.lower} is the lower admissibility bound for
  continuous fixed-bandwidth search candidates. The effective first
  start is \code{max(scale.factor.init, scale.factor.search.lower)},
  and the effective random-start lower endpoint is
  \code{max(scale.factor.init.lower, scale.factor.search.lower)}.
  \code{scale.factor.init.upper} must be at least that effective lower
  endpoint; the package errors rather than silently expanding the user's
  interval.

  When \code{scale.factor.search.lower} is \code{NULL}, an existing
  bandwidth object's stored floor is inherited when available;
  otherwise the package default \code{0.1} is used. Explicit bandwidths
  supplied for storage with \code{bandwidth.compute = FALSE} are not
  rewritten by the search floor.

  Categorical search-start controls such as \code{dfac.init},
  \code{lbd.init}, and \code{hbd.init} have separate semantics and are
  not affected by \code{scale.factor.search.lower}.

  Documentation guide: see \code{\link{np.kernels}} for kernels,
  \code{\link{np.options}} for global options, and
  \code{\link{plot}}, \code{\link{plot.np}} for plotting options.

  The bandwidth-selection argument surface is easiest to read by
  decision group: data, grid, and existing bandwidth inputs; bandwidth
  criterion and representation; continuous kernel and support controls
  beginning with \code{cker*}; ordered categorical kernel controls
  such as \code{okertype}; distribution-specific integral/grid controls
  such as \code{gdat}, \code{gdata}, \code{do.full.integral}, and
  \code{ngrid}; and numerical search initialization, tolerances, and
  feasibility controls. Users who call \code{\link{npudist}} without a
  bandwidth object can pass these same bandwidth-selection controls
  through that function's \code{\dots}.

  For S3 plotting help, see \code{\link{plot.np}}. You can list
  available plot methods with \code{methods("plot")}.


  Typical usages are (see below for a complete list of options and also
  the examples at the end of this help file)
  \preformatted{
    
    Usage 1: compute a bandwidth object using the formula interface:
    
    bw <- npudistbw(~y)
    
    Usage 2: compute a bandwidth object using the data frame interface
    and change the default kernel and order:

    Fhat <- npudistbw(tdat = y, ckertype="epanechnikov", ckerorder=4)
    
  }

  
  \code{npudistbw} implements a variety of methods for choosing
  bandwidths for multivariate (\eqn{p}-variate) distributions defined
  over a set of possibly continuous and/or discrete (ordered) data. The
  approach is based on Li and Racine (2003) who employ
  \sQuote{generalized product kernels} that admit a mix of continuous
  and discrete data types.

  The cross-validation methods employ multivariate numerical search
  algorithms (direction set (Powell's) methods in multidimensions).

  Bandwidths can (and will) differ for each variable which is, of
  course, desirable.

  Three classes of kernel estimators for the continuous data types are
  available: fixed, adaptive nearest-neighbor, and generalized
  nearest-neighbor. Adaptive nearest-neighbor bandwidths change with
  each sample realization in the set, \eqn{x_i}{x[i]}, when estimating the
  cumulative distribution at the point \eqn{x}. Generalized nearest-neighbor bandwidths change
  with the point at which the cumulative distribution is estimated, \eqn{x}. Fixed bandwidths
  are constant over the support of \eqn{x}.

  \code{npudistbw} may be invoked \emph{either} with a formula-like
  symbolic description of variables on which bandwidth selection is to
  be performed \emph{or} through a simpler interface whereby data is
  passed directly to the function via the \code{dat} parameter. Use of
  these two interfaces is \bold{mutually exclusive}.

  Data contained in the data frame \code{dat} may be a mix of continuous
  (default) and ordered discrete (to be specified in the data frame
  \code{dat} using \code{\link{ordered}}). Data can be entered in an
  arbitrary order and data types will be detected automatically by the
  routine (see \code{\link{np}} for details).

  Data for which bandwidths are to be estimated may be specified
  symbolically. A typical description has the form \code{~ data}, where
  \code{data} is a series of variables specified by name, separated by
  the separation character '+'. For example, \code{ ~ x + y } specifies
  that the bandwidths for the joint distribution of variables \code{x}
  and \code{y} are to be estimated. See below for further examples.

  A variety of kernels may be specified by the user. Kernels implemented
  for continuous data types include the second, fourth, sixth, and
  eighth-order Gaussian and Epanechnikov kernels, and the uniform
  kernel. Ordered data types use a variation of the Wang and van Ryzin
  (1981) kernel.

  The optimizer invoked for search is Powell's conjugate direction
  method which requires the setting of (non-random) initial values and
  search directions for bandwidths, and when restarting, random values
  for successive invocations. Bandwidths for \code{numeric} variables
  are scaled by robust measures of spread, the sample size, and the
  number of \code{numeric} variables where appropriate. Two sets of
  parameters for bandwidths for \code{numeric} can be modified, those
  for initial values for the parameters themselves, and those for the
  directions taken (Powell's algorithm does not involve explicit
  computation of the function's gradient). The default values are set by
  considering search performance for a variety of difficult test cases
  and simulated cases. We highly recommend restarting search a large
  number of times to avoid the presence of local minima (achieved by
  modifying \code{nmulti}). Further refinement for difficult cases can
  be achieved by modifying these sets of parameters. However, these
  parameters are intended more for the authors of the package to enable
  \sQuote{tuning} for various methods rather than for the user them
  self.

}
\value{
  \code{npudistbw} returns a \code{bandwidth} object with the
  following components:

  \item{bw}{ bandwidth(s), scale factor(s) or nearest neighbours for the
    data, \code{dat} }
  \item{fval}{ objective function value at minimum }

  if \code{bwtype} is set to \code{fixed}, an object containing
  bandwidths, of class \code{bandwidth}
  (or scale factors if \code{bwscaling = TRUE}) is returned. If it is set to
  \code{generalized_nn} or \code{adaptive_nn}, then instead the
  \eqn{k}th nearest 
  neighbors are returned for the continuous variables while the discrete
  kernel bandwidths are returned for the discrete variables. Bandwidths
  are stored under the component name \code{bw}, with each
  element \eqn{i} corresponding to column \eqn{i} of input data
  \code{dat}.

  The functions  \code{\link{predict}}, \code{\link{summary}} and \code{\link{plot}} support
  objects of type \code{bandwidth}.
}
\section{Book And Method Pointers}{
  \code{npudistbw} selects bandwidths for the unconditional distribution
  target \eqn{F(x)=\Pr(X\le x)} estimated by \code{\link{npudist}}.
  The selected bandwidths control the integrated kernel factors used in
  the mixed-data CDF estimator.

  For book-length derivations, see Li and Racine (2007), Chapter 1
  \emph{Density Estimation}, especially Sections 1.4 and 1.5, and Chapter
  4 \emph{Kernel Estimation with Mixed Data}. The later workflow treatment
  is Racine (2019), Chapters 2 and 3.
}

\references{

  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Bowman, A. and P. Hall and T. Prvan (1998), \dQuote{Bandwidth
    selection for the smoothing of distribution functions,} Biometrika,
    85, 799-808.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
  and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2003), \dQuote{Nonparametric estimation of
   distributions with categorical and continuous data,} Journal
  of Multivariate Analysis, 86, 266-292.

  Li, C. and H. Li and J.S. Racine (2017), \dQuote{Cross-Validated Mixed
  Datatype Bandwidth Selection for Nonparametric Cumulative
  Distribution/Survivor Functions,} Econometric Reviews, \bold{36},
  970-987.

  Ouyang, D. and Q. Li and J.S. Racine (2006), \dQuote{Cross-validation
    and the estimation of probability distributions with categorical
    data,} Journal of Nonparametric Statistics, 18, 69-100.

  Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,}
  Cambridge University Press.

  Scott, D.W. (1992), \emph{Multivariate Cumulative Distribution
  Estimation: Theory, Practice and Visualization,} New York: Wiley.

  Silverman, B.W. (1986), \emph{Density Estimation,} London: Chapman and
  Hall.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth estimators
   for discrete distributions,}  Biometrika, 68, 301-309.

}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.

  Caution: multivariate data-driven bandwidth selection methods are, by
  their nature, \emph{computationally intensive}. Virtually all methods
  require dropping the \eqn{i}th observation from the data set, computing an
  object, repeating this for all observations in the sample, then
  averaging each of these leave-one-out estimates for a \emph{given}
  value of the bandwidth vector, and only then repeating this a large
  number of times in order to conduct multivariate numerical
  minimization/maximization. Furthermore, due to the potential for local
  minima/maxima, \emph{restarting this procedure a large number of times may
  often be necessary}. This can be frustrating for users possessing
  large datasets. For exploratory purposes, you may wish to override the
  default search tolerances, say, setting ftol=.01 and tol=.01 and
  conduct multistarting (the default is to restart min(2, ncol(dat))
  times) as is done for a number of examples. Once the procedure
  terminates, you can restart search with default tolerances using those
  bandwidths obtained from the less rigorous search (i.e., set
  \code{bws=bw} on subsequent calls to this routine where \code{bw} is
  the initial bandwidth object).  A version of this package using the
  \code{Rmpi} wrapper is under development that allows one to deploy
  this software in a clustered computing environment to facilitate
  computation involving large datasets.
 
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}, \code{\link{plot.np}}
  \code{\link{bw.nrd}}, \code{\link{bw.SJ}},  \code{\link{hist}},
  \code{\link{npudist}}, \code{\link{npudist}} }
\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): For this example, we load Giovanni
# Baiocchi's Italian GDP panel (see Italy for details), then create a
# data frame in which year is an ordered factor, GDP is continuous.

data("Italy")
with(Italy, {

data <- data.frame(ordered(year), gdp)

# We compute bandwidths for the kernel cumulative distribution estimator
# using the normal-reference rule-of-thumb. Otherwise, we use the
# defaults (second order Gaussian kernel, fixed bandwidths). Note that
# the bandwidth object you compute inherits all properties of the
# estimator (kernel type, kernel order, estimation method) and can be
# fed directly into the plotting utility plot() or into the npudist()
# function.

bw <- npudistbw(formula=~ordered(year)+gdp, bwmethod="normal-reference")

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Next, specify a value for the bandwidths manually (0.5 for the first
# variable, 1.0 for the second)...

bw <- npudistbw(formula=~ordered(year)+gdp, bws=c(0.5, 1.0),
                bandwidth.compute=FALSE)

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Next, if you wanted to use the 1.587 sigma n^{-1/(2p+q)} rule-of-thumb
# for the bandwidth for the continuous variable and, say, no smoothing
# for the discrete variable, you would use the bwscaling=TRUE argument
# and feed in the values 0 for the first variable (year) and 1.587 for
# the second (gdp). Note that in the printout it reports the `scale
# factors' rather than the `bandwidth' as reported in some of the
# previous examples.

bw <- npudistbw(formula=~ordered(year)+gdp, bws=c(0, 1.587),
                bwscaling=TRUE, 
                bandwidth.compute=FALSE)

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# If you wished to use, say, an eighth-order Epanechnikov kernel for the
# continuous variables and specify your own bandwidths, you could do
# that as follows.

bw <- npudistbw(formula=~ordered(year)+gdp, bws=c(0.5, 1.0),
                bandwidth.compute=FALSE, 
                ckertype="epanechnikov",
                ckerorder=8)

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# If you preferred, say, nearest-neighbor bandwidths and a generalized
# kernel estimator for the continuous variable, you would use the
# bwtype="generalized_nn" argument.

bw <- npudistbw(formula=~ordered(year)+gdp, bwtype = "generalized_nn")

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Next, compute bandwidths using cross-validation, fixed bandwidths, and
# a second-order Gaussian kernel for the continuous variable (default).
# Note - this may take a few minutes depending on the speed of your
# computer.

bw <- npudistbw(formula=~ordered(year)+gdp)

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Finally, if you wish to use initial values for numerical search, you
# can either provide a vector of bandwidths as in bws=c(...) or a
# bandwidth object from a previous run, as in

bw <- npudistbw(formula=~ordered(year)+gdp, bws=c(1, 1))

summary(bw)

})

# EXAMPLE 1 (INTERFACE=DATA FRAME): For this example, we load Giovanni
# Baiocchi's Italian GDP panel (see Italy for details), then create a
# data frame in which year is an ordered factor, GDP is continuous.

data("Italy")
with(Italy, {

data <- data.frame(ordered(year), gdp)

# We compute bandwidths for the kernel cumulative distribution estimator
# using the normal-reference rule-of-thumb. Otherwise, we use the
# defaults (second-order Gaussian kernel, fixed bandwidths). Note that
# the bandwidth object you compute inherits all properties of the
# estimator (kernel type, kernel order, estimation method) and can be
# fed directly into the plotting utility plot() or into the npudist()
# function.

bw <- npudistbw(dat=data, bwmethod="normal-reference")

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Next, specify a value for the bandwidths manually (0.5 for the first
# variable, 1.0 for the second)...

bw <- npudistbw(dat=data, bws=c(0.5, 1.0), bandwidth.compute=FALSE)

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Next, if you wanted to use the 1.587 sigma n^{-1/(2p+q)} rule-of-thumb
# for the bandwidth for the continuous variable and, say, no smoothing
# for the discrete variable, you would use the bwscaling=TRUE argument
# and feed in the values 0 for the first variable (year) and 1.587 for
# the second (gdp). Note that in the printout it reports the `scale
# factors' rather than the `bandwidth' as reported in some of the
# previous examples.

bw <- npudistbw(dat=data, bws=c(0, 1.587),
                bwscaling=TRUE, 
                bandwidth.compute=FALSE)

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# If you wished to use, say, an eighth-order Epanechnikov kernel for the
# continuous variables and specify your own bandwidths, you could do
# that as follows:

bw <- npudistbw(dat=data, bws=c(0.5, 1.0),
                bandwidth.compute=FALSE, 
                ckertype="epanechnikov",
                ckerorder=8)

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# If you preferred, say, nearest-neighbor bandwidths and a generalized
# kernel estimator for the continuous variable, you would use the
# bwtype="generalized_nn" argument.

bw <- npudistbw(dat=data, bwtype = "generalized_nn")

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Next, compute bandwidths using cross-validation, fixed bandwidths, and
# a second order Gaussian kernel for the continuous variable (default).
# Note - this may take a few minutes depending on the speed of your
# computer.

bw <- npudistbw(dat=data)

summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Finally, if you wish to use initial values for numerical search, you
# can either provide a vector of bandwidths as in bws=c(...) or a
# bandwidth object from a previous run, as in

bw <- npudistbw(dat=data, bws=c(1, 1))

summary(bw)

})
} % enddontrun
}
\keyword{nonparametric}

\name{oecdpanel}
\docType{data}
\alias{oecdpanel}
\alias{bw}
\title{ Cross Country Growth Panel }
\description{
  Cross country GDP growth panel covering the period 1960-1995 used by
  Liu and Stengos (2000) and Maasoumi, Racine, and Stengos (2007). There
  are 616 observations in total. \code{data("oecdpanel")} makes available
  the dataset \code{"oecdpanel"} plus an additional object \code{"bw"}.
}
\usage{data("oecdpanel")}
\format{

  A data frame with 7 columns, and 616 rows.  This panel covers 7
  5-year periods: 1960-1964, 1965-1969, 1970-1974, 1975-1979, 1980-1984,
  1985-1989 and 1990-1994.

  A separate local-linear \code{rbandwidth} object (\code{bw}) has
  been computed for the user's convenience which can be used to
  visualize this dataset using \code{\link{plot}(bw)}.
  
  \describe{

    \item{growth}{ the first column, of type \code{numeric}: growth rate
    of real GDP per capita for each 5-year period}

    \item{oecd}{ the second column, of type \code{factor}: equal to
    1 for OECD members, 0 otherwise}

    \item{year}{ the third column, of type \code{integer}}

    \item{initgdp}{ the fourth column, of type \code{numeric}: per capita
    real GDP at the beginning of each 5-year period}

    \item{popgro}{ the fifth column, of type \code{numeric}: average
    annual population growth rate for each 5-year period}

    \item{inv}{ the sixth column, of type \code{numeric}: average
    investment/GDP ratio for each 5-year period}

    \item{humancap}{ the seventh column, of type \code{numeric}:
      average secondary school enrolment rate for each 5-year period}

    }

}
\source{ Thanasis Stengos }
\examples{
data("oecdpanel")
with(oecdpanel, {
summary(oecdpanel)
})
}
\references{

  Liu, Z. and T. Stengos (1999), \dQuote{Non-linearities in cross
  country growth regressions: a semiparametric approach,}
  Journal of Applied Econometrics, 14, 527-538.

  Maasoumi, E. and J.S. Racine and T. Stengos (2007),
  \dQuote{Growth and convergence: a profile of distribution dynamics and
  mobility,} Journal of Econometrics, 136, 483-508 
  
}
\keyword{datasets}

\name{se}
\alias{se}
\title{Extract Standard Errors}
\usage{
se(x)
}
\arguments{
\subsection{Object Input}{
Object to interrogate for standard errors.
}

  \item{x}{an object for which the extraction of standard errors is
    meaningful.}
}
\value{
  Standard errors extracted from the model object \code{x}.
}
\description{
  \code{se} is a generic function which extracts standard errors 
  from objects. 
}
\details{
  This function provides a generic interface for extraction of standard
  errors from objects.
}
\seealso{
  \code{\link{fitted}}, \code{\link{residuals}}, \code{\link{coef}},
  and \code{\link{gradients}}, for related methods;
  \code{\link{np}} for supported objects. 
}

\author{
  Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine \email{racinej@mcmaster.ca}
}

\section{Note}{
  This method currently only supports objects from the \code{\link{np}} library.
}
\examples{
x <- rnorm(10)
se(npudens(npudensbw(~x)))
}
\keyword{nonparametric}

%
\name{npsdeptest}
\alias{npsdeptest}
\title{ Kernel Consistent Serial Dependence Test for Univariate Nonlinear Processes }
\description{
  \code{npsdeptest} implements the consistent metric entropy test of
  nonlinear serial dependence as described in Granger, Maasoumi and
  Racine (2004).
}
\usage{

npsdeptest(data = NULL,
           lag.num = 1,
           method = c("integration","summation"),
           bootstrap = TRUE,
           boot.num = 399,
           random.seed = 42)

}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the data series, lag count, and statistic variant.
}

  \item{data}{
    a vector containing the variable that can be of type
    \code{\link{numeric}} or \code{\link{ts}}.
  }


  \item{lag.num}{
    an integer value specifying the maximum number of lags to
    use. Defaults to \code{1}.
  }


  \item{method}{
    a character string used to specify whether to compute the integral
    version or the summation version of the statistic. Can be set as
    \code{integration} or \code{summation} (see below for
    details). Defaults to \code{integration}.
  }


\subsection{Bootstrap Controls}{
These arguments control bootstrap execution and reproducibility settings.
}

  \item{boot.num}{
    an integer value specifying the number of bootstrap
    replications to use. Defaults to \code{399}.
  }


  \item{bootstrap}{
    a logical value which specifies whether to conduct
    the bootstrap test or not. If set to \code{FALSE}, only the
    statistic will be computed. Defaults to \code{TRUE}.
  }


  \item{random.seed}{
    an integer used to seed R's random number generator. This is to
    ensure replicability. Defaults to 42.
  }


}
\value{
  \code{npsdeptest} returns an object of type \code{deptest} with the
  following components

  \item{Srho}{ the statistic vector \code{Srho} }
  \item{Srho.cumulant}{ the cumulant statistic vector \code{Srho.cumulant} }
  \item{Srho.bootstrap.mat}{ contains the bootstrap replications of
    \code{Srho} }
  \item{Srho.cumulant.bootstrap.mat}{ contains the bootstrap
    replications of \code{Srho.cumulant} }
  \item{P}{ the P-value vector of the Srho statistic vector }
  \item{P.cumulant}{ the P-value vector of the cumulant Srho statistic vector }
  \item{bootstrap}{ a logical value indicating whether bootstrapping was
  performed }
  \item{boot.num}{ number of bootstrap replications }
  \item{lag.num}{ the number of lags }
  \item{bw.y}{ the numeric vector of bandwidths for \code{data}
  marginal density at lag \code{num.lag}}
  \item{bw.y.lag}{ the numeric vector of bandwidths for lagged
    \code{data} marginal density  at lag \code{num.lag}}
  \item{bw.joint}{ the numeric matrix of bandwidths for \code{data}
  and lagged \code{data} joint density  at lag \code{num.lag}}

\code{\link{summary}} supports object of type \code{deptest}.
}
\section{Book And Method Pointers}{
  \code{npsdeptest} applies the dependence-test idea to a time series by
  comparing the joint law of current and lagged values with the product
  of their marginal laws. The lag and bootstrap choices determine the
  serial-dependence alternatives being probed and the null calibration.

  For book-length background, see Li and Racine (2007), Chapter 13
  \emph{Nonsmoothing Tests}, especially Sections 13.4 and 13.5, and
  Chapter 12 \emph{Model Specification Tests} for related distributional
  tests. For entropy and density context, see Racine (2019), Chapter 2.
}

\references{

  Granger, C.W. and E. Maasoumi and J.S. Racine (2004), \dQuote{A
  dependence metric for possibly nonlinear processes}, Journal of Time
  Series Analysis, 25, 649-669.

}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.


  \code{npsdeptest} computes the nonparametric metric entropy
   (normalized Hellinger of Granger, Maasoumi and Racine (2004)) for
   testing for nonlinear serial dependence, \eqn{D[f(y_t, \hat y_{t-k}),
   f(y_t)\times f(\hat y_{t-k})]}{D[f(y[t], y[t-k]),
   f(y[t])*f(y[t-k])]}.  Default bandwidths are of the Kullback-Leibler
   variety obtained via likelihood cross-validation.

  The test may be applied to a raw data series or to residuals of user
  estimated models.

  The summation version of this statistic may be numerically unstable
  when \code{data} is sparse (the summation version involves division of
  densities while the integration version involves differences). Warning
  messages are produced should this occur (\sQuote{integration recommended})
  and should be heeded.

}

\section{Usage Issues}{

  The \code{integration} version of the statistic uses multidimensional
  numerical methods from the \pkg{cubature} package. See
  \pkg{adaptIntegrate} for details. The \code{integration}
  version of the statistic will be substantially slower than the
  \code{summation} version, however, it will likely be both more
  accurate and powerful.

}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}
  \code{\link{npdeptest},\link{npdeneqtest},\link{npsymtest},\link{npunitest}}
}

\examples{
\dontrun{
set.seed(1234)

## A function to create a time series

ar.series <- function(phi,epsilon) {
  n <- length(epsilon)
  series <- numeric(n)
  series[1] <- epsilon[1]/(1-phi)
  for(i in 2:n) {
    series[i] <- phi*series[i-1] + epsilon[i]
  }
  return(series)
}

n <- 100

## Stationary persistent time-series

yt <- ar.series(0.95,rnorm(n))
npsdeptest(yt,lag.num=2,boot.num=99,method="summation")

if (interactive()) Sys.sleep(5)

## Stationary independent time-series

yt <- ar.series(0.0,rnorm(n))
npsdeptest(yt,lag.num=2,boot.num=99,method="summation")

## Stationary persistent time-series

yt <- ar.series(0.95,rnorm(n))
npsdeptest(yt,lag.num=2,boot.num=99,method="integration")

if (interactive()) Sys.sleep(5)

## Stationary independent time-series

yt <- ar.series(0.0,rnorm(n))
npsdeptest(yt,lag.num=2,boot.num=99,method="integration")

} % enddontrun

}
\keyword{ univar }
\keyword{ nonparametric }

\name{nptgauss}
\alias{nptgauss}
\title{Truncated Second-order Gaussian Kernels}
\description{
  \code{nptgauss} provides an interface for setting the truncation
  radius of the truncated second-order Gaussian kernel used
  by \pkg{np}.
}
\usage{
nptgauss(b)
}

\arguments{
\subsection{Kernel Truncation}{
Truncation radius for the truncated Gaussian kernel helper.
}

  \item{b}{
    Truncation radius of the kernel.
  }
}

\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.

  \code{nptgauss} allows one to set the truncation radius of the truncated Gaussian kernel used by \pkg{np}, which defaults to 3. It automatically computes the constants describing the truncated gaussian kernel for the user.
  
We define the truncated gaussion kernel on the interval \eqn{[-b,b]} as:
\deqn{K = \frac{\alpha}{\sqrt{2\pi}}\left(e^{-z^2/2} - e^{-b^2/2}\right)}{K = alpha/sqrt(2*pi)*exp(-b^2/2)*(exp(-z^2/2) - exp(-b^2/2))}

The constant \eqn{\alpha} is computed as:
\deqn{\alpha = \left[\int_{-b}^{b} \frac{1}{\sqrt{2\pi}}\left(e^{-z^2/2} - e^{-b^2/2}\right)\right]^{-1}}{1/integrate(1/sqrt(2*pi)*exp(-b^2/2)*(exp(-z^2/2) - exp(-b^2/2)), -b, b)}

Given these definitions, the derivative kernel is simply:
\deqn{K' = (-z)\frac{\alpha}{\sqrt{2\pi}}e^{-z^2/2}}{-z*alpha/(sqrt(2*pi))*exp(-z^2/2)}

The CDF kernel is:
\deqn{G = \frac{\alpha}{2}\mathrm{erf}(z/\sqrt{2}) + \frac{1}{2} - c_0z}{alpha/2*erf(z/sqrt(2))+0.5-c0*z}

The convolution kernel on \eqn{[-2b,0]} has the general form:
\deqn{H_- = a_0\,\mathrm{erf}(z/2 + b) e^{-z^2/4} + a_1z + a_2\,\mathrm{erf}((z+b)/\sqrt{2}) - c_0}{a0*erf(z/2 + b)*exp(-0.25*z^2) + a1*z + a2*erf((z+b)/sqrt(2)) - c0}

and on \eqn{[0,2b]} it is:
\deqn{H_+ = -a_0\,\mathrm{erf}(z/2 - b) e^{-z^2/4} - a_1z - a_2\,\mathrm{erf}((z-b)/\sqrt{2}) - c_0}{-a0*erf(z/2 - b)*exp(-0.25*z^2) - a1*z - a2*erf((z-b)/sqrt(2)) - c0}

where \eqn{a_0}{a0} is determined by the normalisation condition on H,
\eqn{a_2}{a2} is determined by considering the value of the kernel at
\eqn{z = 0} and \eqn{a_1}{a1} is determined by the requirement that \eqn{H = 0} at \eqn{[-2b,2b]}.
}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\examples{
## The default kernel, a gaussian truncated at +- 3
nptgauss(b = 3.0)

}

%
\name{npindex}
\alias{npindex}
\alias{npindex.default}
\alias{npindex.formula}
\alias{npindex.sibandwidth}

\title{Semiparametric Single Index Model}

\description{
  \code{npindex} computes a semiparametric single index model
  for a dependent variable and \eqn{p}-variate explanatory data using
  the model \eqn{Y = G(X\beta) + \epsilon}{Y = G(XB) + epsilon}, given a
  set of evaluation points, training points (consisting of explanatory
  data and dependent data), and a \code{npindexbw} bandwidth
  specification. Note that for this semiparametric estimator, the
  bandwidth object contains parameters for the single index model and
  the (scalar) bandwidth for the index function.
}

\usage{
npindex(bws, \dots)

\method{npindex}{formula}(bws,
        data = NULL,
        newdata = NULL,
        y.eval = FALSE,
        \dots)


\method{npindex}{default}(bws,
        txdat,
        tydat,
        nomad = FALSE,
        \dots) 

\method{npindex}{sibandwidth}(bws,
        txdat = stop("training data 'txdat' missing"),
        tydat = stop("training data 'tydat' missing"),
        exdat,
        eydat,
        boot.num = 399,
        errors = FALSE,
        gradients = FALSE,
        residuals = FALSE,
        \dots)
}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the bandwidth specification, formula/data interface, and training data.
}

  \item{bws}{
    a bandwidth specification. This can be set as a
    \code{sibandwidth}
    object returned from an invocation of \code{npindexbw}, or
    as a vector of parameters (beta) with each element \eqn{i}
    corresponding to the coefficient for column \eqn{i} in \code{txdat}
    where the first element is normalized to 1, and a scalar bandwidth
    (h).
  }


  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(bws)}, typically the environment from which
    \code{\link{npindexbw}} was called.
  }


  \item{txdat}{
    a \eqn{p}-variate data frame of explanatory data (training data) used to
    calculate the regression estimators. Defaults to the training data used to
    compute the bandwidth object.
  }


  \item{tydat}{
    a one (1) dimensional numeric or integer vector of dependent data, each
    element \eqn{i} corresponding to each observation (row) \eqn{i} of
    \code{txdat}. Defaults to the training data used to
    compute the bandwidth object.
  }


\subsection{Local-Polynomial Degree And Bandwidth Search}{
This argument controls the recommended automatic local-polynomial NOMAD route, which jointly selects continuous polynomial degree and bandwidths when these are computed inside \code{npindex}.
}

  \item{nomad}{
    logical shortcut passed through to \code{\link{npindexbw}} when
    bandwidths are computed inside \code{npindex}. When \code{TRUE},
    the single-index bandwidth route fills any missing values among
    \code{regtype}, \code{search.engine}, \code{degree.select},
    \code{bernstein.basis}, \code{degree.min}, \code{degree.max},
    \code{degree.verify}, and \code{bwtype} with the recommended
    automatic local-polynomial degree-and-bandwidth NOMAD preset documented in \code{\link{npindexbw}}.
    Additional NOMAD tuning arguments such as \code{nomad.nmulti} may
    also be supplied through \code{\dots}; \code{nmulti} remains the
    outer restart count while \code{nomad.nmulti} controls inner
    \code{crs::snomadr()} multistarts within each outer restart.
    After fitting, inspect \code{fit$bws$nomad.shortcut} on the
    returned object \code{fit} to see the normalized shortcut metadata.
  }


\subsection{Evaluation Data And Returned Quantities}{
These arguments control where the single-index fit is evaluated and which evaluation quantities are returned.
}

  \item{exdat}{
    a \eqn{p}-variate data frame of points on which the regression will be
    estimated (evaluation data). By default,
    evaluation takes place on the data provided by \code{txdat}.
  }


  \item{eydat}{
    a one (1) dimensional numeric or integer vector of the true values
    of the dependent variable. Optional, and used only to calculate the
    true errors.
  }


  \item{newdata}{
    An optional data frame in which to look for evaluation data. If
    omitted, the training data are used.
  }


  \item{y.eval}{
    If \code{newdata} contains dependent data and \code{y.eval = TRUE},
    \code{\link{np}} will compute goodness of fit statistics on these
    data and return them. Defaults to \code{FALSE}.
  }


\subsection{Fitted Quantities And Inference}{
These arguments control residuals, gradients, and bootstrap standard errors.
}

  \item{boot.num}{
    an integer specifying the number of bootstrap replications to use
    when performing standard error calculations. Defaults to
    \code{399}.
  }


  \item{errors}{
    a logical value indicating that you want (bootstrapped)
    standard errors for the conditional mean, gradients (when
    \code{gradients=TRUE} is set), and average gradients (when
    \code{gradients=TRUE} is set), computed and returned in the
    resulting \code{singleindex} object. Defaults to \code{FALSE}.
  }


  \item{gradients}{
    a logical value indicating that you want gradients and the
    asymptotic covariance matrix for beta computed and returned in the
    resulting \code{singleindex} object. Defaults to \code{FALSE}.
  }


  \item{residuals}{
    a logical value indicating that you want residuals computed and
    returned in the resulting \code{singleindex} object. Defaults to
    \code{FALSE}.
  }


\subsection{Additional Arguments}{
Further arguments are passed to the bandwidth-selection counterpart when bandwidths are not supplied.
}

  \item{\dots}{
    additional arguments supplied to specify the parameters to the
    \code{sibandwidth} S3 method, which is called during estimation.
  }
}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}}, \code{\link{plot.np}} for plotting options.

  For S3 plotting help, see \code{\link{plot.np}}. You can list
  available plot methods with \code{methods("plot")}.


  A matrix of gradients along with average derivatives are computed and
  returned if \code{gradients=TRUE} is used.

  For practitioners who want the recommended automatic local-polynomial degree-and-bandwidth NOMAD route
  without spelling out all LP tuning arguments,
  \code{npindex(..., nomad=TRUE)} and \code{npindexbw(..., nomad=TRUE)}
  expand missing settings to the same documented preset. Explicit
  incompatible settings fail fast rather than being silently rewritten.

}

\value{

  \code{npindex} returns a \code{npsingleindex} object.  The generic
  functions \code{\link{fitted}}, \code{\link{residuals}},
  \code{\link{coef}}, \code{\link{vcov}}, \code{\link{se}},
  \code{\link{predict}}, and \code{\link{gradients}}, extract (or
  generate) estimated values, residuals, coefficients,
  variance-covariance matrix, bootstrapped standard errors on estimates,
  predictions, and gradients, respectively, from the returned
  object. Furthermore, the functions \code{\link{summary}} and
  \code{\link{plot}} support objects of this type. The returned object
  has the following components:

  \item{eval}{ evaluation points }
  \item{mean}{ estimates of the regression function (conditional mean) at the
  evaluation points }
  \item{beta}{ the model coefficients }
  \item{betavcov}{ the asymptotic covariance matrix for the model coefficients}
  \item{merr}{ standard errors of the regression function estimates }
  \item{grad}{ estimates of the gradients at each evaluation point }
  \item{gerr}{ standard errors of the gradient estimates }
  \item{mean.grad}{ mean (average) gradient over the evaluation points}
  \item{mean.gerr}{ bootstrapped standard error of the mean gradient estimates }
  \item{R2}{ if \code{method="ichimura"}, coefficient of determination
  (Doksum and Samarov (1995))}
  \item{MSE}{if \code{method="ichimura"}, mean squared error }
  \item{MAE}{if \code{method="ichimura"}, mean absolute error }
  \item{MAPE}{if \code{method="ichimura"}, mean absolute percentage error }
  \item{CORR}{if \code{method="ichimura"}, absolute value of Pearson's correlation coefficient }
  \item{SIGN}{if \code{method="ichimura"}, fraction of observations where fitted and observed values
    agree in sign }
  \item{confusion.matrix}{if \code{method="kleinspady"}, the confusion matrix or \code{NA} if outcomes
    are not available }
  \item{CCR.overall}{if \code{method="kleinspady"},  the overall correct
    classification ratio, or \code{NA} if outcomes are not available }
  \item{CCR.byoutcome}{if \code{method="kleinspady"},  a numeric vector containing the correct
    classification ratio by outcome, or \code{NA} if outcomes are not
    available }
  \item{fit.mcfadden}{if \code{method="kleinspady"},  the McFadden-Puig-Kerschner performance measure
    or \code{NA} if outcomes are not available }
}



\section{Book And Method Pointers}{
  The single-index model reduces a multivariate predictor to an index,
  typically written \eqn{E[Y\mid X]=g(X^\prime\beta)} for continuous
  outcomes, with a normalization on \eqn{\beta} for identification.
  For binary outcomes, the Klein-Spady route estimates the corresponding
  binary-choice probability through the index. The fitted index
  regression, gradients, and average derivatives are computed from the
  selected index direction and kernel regression fit.

  For book-length derivations, see Li and Racine (2007), Chapter 8
  \emph{Semiparametric Single Index Models}, especially Sections 8.1,
  8.2.1, 8.4, 8.5, and 8.10. The later workflow treatment is Racine
  (2019), Chapter 8 \emph{Semiparametric Conditional Mean Function
  Estimation}, especially the single-index material.
}

\references{

  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Doksum, K. and A. Samarov (1995), \dQuote{Nonparametric estimation of
  global functionals and a measure of the explanatory power of
  covariates  regression,} The Annals of Statistics, 23 1443-1473.

  Ichimura, H., (1993), \dQuote{Semiparametric least squares (SLS) and
  weighted SLS estimation of single-index models,} Journal of
  Econometrics, 58, 71-120.

  Klein, R. W. and R. H. Spady (1993), \dQuote{An efficient semiparametric
  estimator for binary response models,} Econometrica, 61, 387-421.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics:
  Theory and Practice,} Princeton University Press.

  McFadden, D. and C. Puig and D. Kerschner (1977), \dQuote{Determinants
    of the long-run demand for electricity,} Proceedings of the
  American Statistical Association (Business and Economics Section),
  109-117.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth
    estimators for discrete distributions,} Biometrika, 68, 301-309.
}



\author{
  Tristen Hayfield  \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.

  \code{\link{vcov}} requires that \code{gradients=TRUE} be set.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}},
  \code{\link{plot}}, \code{\link{plot.np}}, \code{\link{npindexbw}}
}

\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): Generate a simple linear model then
# estimate it using a semiparametric single index specification and
# Ichimura's nonlinear least squares coefficients and bandwidth
# (default). Also compute the matrix of gradients and average derivative
# estimates.

set.seed(12345)

n <- 100

x1 <- runif(n, min=-1, max=1)
x2 <- runif(n, min=-1, max=1)

y <- x1 - x2 + rnorm(n)

# Note - this may take a minute or two depending on the speed of your
# computer. Note also that the first element of the vector beta is
# normalized to one for identification purposes, and that X must contain
# at least one continuous variable.

bw <- npindexbw(formula=y~x1+x2)

summary(bw)

model <- npindex(bws=bw, gradients=TRUE)

summary(model)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Or you can visualize the input with plot.

if (interactive()) plot(bw)

if (interactive()) Sys.sleep(5)

# EXAMPLE 1 (INTERFACE=DATA FRAME): Generate a simple linear model then
# estimate it using a semiparametric single index specification and
# Ichimura's nonlinear least squares coefficients and bandwidth
# (default). Also compute the matrix of gradients and average derivative
# estimates.

set.seed(12345)

n <- 100

x1 <- runif(n, min=-1, max=1)
x2 <- runif(n, min=-1, max=1)

y <- x1 - x2 + rnorm(n)

X <- cbind(x1, x2)

# Note - this may take a minute or two depending on the speed of your
# computer. Note also that the first element of the vector beta is
# normalized to one for identification purposes, and that X must contain
# at least one continuous variable.

bw <- npindexbw(xdat=X, ydat=y)

summary(bw)

model <- npindex(bws=bw, gradients=TRUE)

summary(model)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# Or you can visualize the input with plot.

if (interactive()) plot(bw)

if (interactive()) Sys.sleep(5)

# EXAMPLE 2 (INTERFACE=FORMULA): Generate a simple binary outcome linear
# model then estimate it using a semiparametric single index
# specification and Klein and Spady's likelihood-based coefficients and
# bandwidth (default). Also compute the matrix of gradients and average
# derivative estimates.

n <- 100

x1 <- runif(n, min=-1, max=1)
x2 <- runif(n, min=-1, max=1)

y <- ifelse(x1 + x2 + rnorm(n) > 0, 1, 0)

# Note that the first element of the vector beta is normalized to one
# for identification purposes, and that X must contain at least one
# continuous variable.

bw <- npindexbw(formula=y~x1+x2, method="kleinspady")

summary(bw)

model <- npindex(bws=bw, gradients=TRUE)

# Note that, since the outcome is binary, we can assess model
# performance using methods appropriate for binary outcomes. We look at
# the confusion matrix, various classification ratios, and McFadden et
# al's measure of predictive performance.

summary(model)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# EXAMPLE 2 (INTERFACE=DATA FRAME): Generate a simple binary outcome
# linear model then estimate it using a semiparametric single index
# specification and Klein and Spady's likelihood-based coefficients and
# bandwidth (default). Also compute the matrix of gradients and average
# derivative estimates.

n <- 100

x1 <- runif(n, min=-1, max=1)
x2 <- runif(n, min=-1, max=1)

y <- ifelse(x1 + x2 + rnorm(n) > 0, 1, 0)

X <- cbind(x1, x2)

# Note that the first element of the vector beta is normalized to one
# for identification purposes, and that X must contain at least one
# continuous variable.

bw <- npindexbw(xdat=X, ydat=y, method="kleinspady")

summary(bw)

model <- npindex(bws=bw, gradients=TRUE)

# Note that, since the outcome is binary, we can assess model
# performance using methods appropriate for binary outcomes. We look at
# the confusion matrix, various classification ratios, and McFadden et
# al's measure of predictive performance.

summary(model)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# EXAMPLE 3 (INTERFACE=FORMULA): Replicate the DGP of Klein & Spady
# (1993) (see their description on page 405, pay careful attention to
# footnote 6 on page 405).

set.seed(123)

n <- 1000

# x1 is chi-squared having 3 df truncated at 6 standardized by
# subtracting 2.348 and dividing by 1.511

x <- rchisq(n, df=3)
x1 <- (ifelse(x < 6, x, 6) - 2.348)/1.511

# x2 is normal (0, 1) truncated at +- 2 divided by 0.8796

x <- rnorm(n)
x2 <- ifelse(abs(x) < 2 , x, 2) / 0.8796

# y is 1 if y* > 0, 0 otherwise.

y <- ifelse(x1 + x2 + rnorm(n) > 0, 1, 0)

# Compute the parameter vector and bandwidth. Note that the first
# element of the vector beta is normalized to one for identification
# purposes, and that X must contain at least one continuous variable.


bw <- npindexbw(formula=y~x1+x2, method="kleinspady")

# Next, create the evaluation data in order to generate a perspective
# plot

# Create an evaluation data matrix

x1.seq <- seq(min(x1), max(x1), length=50)
x2.seq <- seq(min(x2), max(x2), length=50)
X.eval <- expand.grid(x1=x1.seq, x2=x2.seq)

# Now evaluate the single index model on the evaluation data

fit <- fitted(npindex(exdat=X.eval,
               eydat=rep(1, nrow(X.eval)),
               bws=bw))

# Finally, coerce the fitted model into a matrix suitable for 3D
# plotting via persp()

fit.mat <- matrix(fit, 50, 50)

# Generate a perspective plot similar to Figure 2 b of Klein and Spady
# (1993)

persp(x1.seq,
      x2.seq,
      fit.mat,
      col="white",
      ticktype="detailed",
      expand=0.5,
      axes=FALSE,
      box=FALSE,
      main="Estimated Semiparametric Probability Perspective",
      theta=310,
      phi=25)

# EXAMPLE 3 (INTERFACE=DATA FRAME): Replicate the DGP of Klein & Spady
# (1993) (see their description on page 405, pay careful attention to
# footnote 6 on page 405).

set.seed(123)

n <- 1000

# x1 is chi-squared having 3 df truncated at 6 standardized by
# subtracting 2.348 and dividing by 1.511

x <- rchisq(n, df=3)
x1 <- (ifelse(x < 6, x, 6) - 2.348)/1.511

# x2 is normal (0, 1) truncated at +- 2 divided by 0.8796

x <- rnorm(n)
x2 <- ifelse(abs(x) < 2 , x, 2) / 0.8796

# y is 1 if y* > 0, 0 otherwise.

y <- ifelse(x1 + x2 + rnorm(n) > 0, 1, 0)

# Create the X matrix

X <- cbind(x1, x2)

# Compute the parameter vector and bandwidth. Note that the first
# element of the vector beta is normalized to one for identification
# purposes, and that X must contain at least one continuous variable.


bw <- npindexbw(xdat=X, ydat=y, method="kleinspady")

# Next, create the evaluation data in order to generate a perspective
# plot

# Create an evaluation data matrix

x1.seq <- seq(min(x1), max(x1), length=50)
x2.seq <- seq(min(x2), max(x2), length=50)
X.eval <- expand.grid(x1=x1.seq, x2=x2.seq)

# Now evaluate the single index model on the evaluation data

fit <- fitted(npindex(exdat=X.eval,
               eydat=rep(1, nrow(X.eval)),
               bws=bw))

# Finally, coerce the fitted model into a matrix suitable for 3D
# plotting via persp()

fit.mat <- matrix(fit, 50, 50)

# Generate a perspective plot similar to Figure 2 b of Klein and Spady
# (1993)

persp(x1.seq,
      x2.seq,
      fit.mat,
      col="white",
      ticktype="detailed",
      expand=0.5,
      axes=FALSE,
      box=FALSE,
      main="Estimated Semiparametric Probability Perspective",
      theta=310,
      phi=25)
} % enddontrun
}
\keyword{nonparametric}

\name{gradients}
\title{Extract Gradients}
\alias{gradients}
\alias{gradients.condensity}
\alias{gradients.condistribution}
\alias{gradients.npregression}
\alias{gradients.qregression}  
\alias{gradients.singleindex}
\usage{
gradients(x, \dots)

\method{gradients}{condensity}(x, errors = FALSE, gradient.order = NULL, \dots)

\method{gradients}{condistribution}(x, errors = FALSE, gradient.order = NULL, \dots)

\method{gradients}{npregression}(x, errors = FALSE, gradient.order = NULL, \dots)

\method{gradients}{qregression}(x, errors = FALSE, \dots)

\method{gradients}{singleindex}(x, errors = FALSE, \dots)
}

\arguments{
\subsection{Object And Output Controls}{
Object to interrogate and whether gradient standard errors are requested.
}

  \item{x}{an object for which the extraction of gradients is
    meaningful.}
  \item{errors}{ a logical value specifying whether or not standard
    errors of gradients are desired. Defaults to \code{FALSE}.}

\subsection{Derivative Order Controls}{
Optional local-polynomial derivative order controls.
}

  \item{gradient.order}{
    for \code{npregression}, \code{condensity}, and
    \code{condistribution} objects fitted with \code{regtype="lp"},
    optional derivative order request (scalar or one entry per
    continuous predictor). Orders exceeding the fitted polynomial degree
    are returned as \code{NA}. If a different admissible derivative order
    is desired, compute the fit, prediction, or evaluation with
    \code{gradients=TRUE} and the desired \code{gradient.order};
    \code{gradients()} extracts stored derivatives and does not
    recompute another derivative order after fitting.
  }

\subsection{Additional Arguments}{
Further method-specific arguments.
}

  \item{\dots}{other arguments.}
}
\value{
  Gradients extracted from the model object \code{x}.
}
\description{
  \code{gradients} is a generic function which extracts gradients 
  from objects. 
}
\details{
  This function provides a generic interface for extraction of gradients
  from objects. For \code{npregression}, \code{condensity}, and
  \code{condistribution} objects fitted with \code{regtype="lp"},
  \code{gradient.order} identifies the stored continuous-predictor
  derivative order to extract; it is not a post-fit recomputation
  control. For \code{qregression} objects,
  \code{errors=TRUE} returns
  asymptotic standard errors for the quantile gradients when the object was
  fitted with \code{gradients=TRUE}.
}
\seealso{
  \code{\link{fitted}}, \code{\link{residuals}}, \code{\link{coef}},
  and \code{\link{se}}, for related methods;
  \code{\link{np}} for supported objects. 
}
\references{
  See the references for the method being interrogated via
  \code{\link{gradients}} in the appropriate help file. For example, for
  the particulars of the gradients for nonparametric regression see the
  references in \code{\link{npreg}}
}
\author{
  Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine \email{racinej@mcmaster.ca}
}

\section{Note}{
  This method currently only supports objects from the \code{\link{np}} library.
}
\examples{
x <- runif(10)
y <- x + rnorm(10, sd = 0.1)
gradients(npreg(y~x, gradients=TRUE))
}
\keyword{nonparametric}

\name{npregivderiv}
\alias{npregivderiv}

\title{
Nonparametric Instrumental Derivatives
}
\description{
\code{npregivderiv} uses the approach of Florens, Racine and Centorrino
(2018) to compute the partial derivative of a nonparametric
estimation of an instrumental regression function \eqn{\varphi}{phi}
defined by conditional moment restrictions stemming from a structural
econometric model: \eqn{E [Y - \varphi (Z,X) | W ] = 0}{E [Y - phi (Z,X)
| W ] = 0}, and involving endogenous variables \eqn{Y} and \eqn{Z} and
exogenous variables \eqn{X} and instruments \eqn{W}. The derivative
function \eqn{\varphi'}{phi'} is the solution of an ill-posed inverse
problem, and is computed using Landweber-Fridman regularization.
}
\usage{
npregivderiv(y,
             z,
             w,
             x = NULL,
             zeval = NULL,
             weval = NULL,
             xeval = NULL,
             constant = 0.5,
             iterate.break = TRUE,
             iterate.max = 1000,
             nmulti = NULL,
             random.seed = 42,
             smooth.residuals = TRUE,
             start.from = c("Eyz","EEywz"),
             starting.values = NULL,
             stop.on.increase = TRUE,
             ...)
}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the response, endogenous variables, instruments, exogenous covariates, and evaluation data.
}

  \item{w}{
    a \eqn{q}-variate data frame of instruments. The data types may be
    continuous, discrete (unordered and ordered factors), or some
    combination thereof.
  }

  \item{weval}{
    a \eqn{q}-variate data frame of instruments on which the regression
    will be estimated (evaluation data). By default, evaluation
    takes place on the data provided by \code{w}.
  }

  \item{x}{
    an \eqn{r}-variate data frame of exogenous regressors. The data
    types may be continuous, discrete (unordered and ordered factors),
    or some combination thereof.
  }

  \item{xeval}{
    an \eqn{r}-variate data frame of exogenous regressors on which the
    regression will be estimated (evaluation data). By default,
    evaluation takes place on the data provided by \code{x}.
  }


  \item{y}{
    a one (1) dimensional numeric or integer vector of dependent data, each
    element \eqn{i} corresponding to each observation (row) \eqn{i} of
    \code{z}.
  }


  \item{z}{
    a \eqn{p}-variate data frame of endogenous regressors. The data
    types may be continuous, discrete (unordered and ordered factors),
    or some combination thereof.
  }

  \item{zeval}{
    a \eqn{p}-variate data frame of endogenous regressors on which the
    regression will be estimated (evaluation data). By default, evaluation
    takes place on the data provided by \code{z}.
  }

\subsection{Iteration Controls}{
These arguments control derivative iteration and reproducibility settings.
}

\item{constant}{
  the constant to use for Landweber-Fridman iteration.
}


  \item{iterate.break}{
    a logical value indicating whether to compute all objects up to
    \code{iterate.max} or to break when a potential optimum arises
    (useful for inspecting full stopping rule profile up to
    \code{iterate.max})
  }


  \item{iterate.max}{
    an integer indicating the maximum number of iterations permitted
    before termination occurs for Landweber-Fridman iteration.
  }


  \item{nmulti}{
    integer number of times to restart the process of finding extrema of
    the cross-validation function from different (random) initial
    points.
  }


  \item{random.seed}{
    an integer used to seed R's random number generator. This ensures
    replicability of the numerical search. Defaults to 42.
  }


\subsection{Residual Smoothing And Starting Values}{
These arguments control residual smoothing and the initial derivative path.
}

\item{smooth.residuals}{
  a logical value (defaults to \code{TRUE}) indicating whether to
  optimize bandwidths for the regression of \eqn{y-\varphi(z)}{y-phi(z)}
  on \eqn{w} or for the regression of \eqn{\varphi(z)}{phi(z)} on
  \eqn{w} during Landweber-Fridman iteration.
 }


\item{start.from}{ a character string indicating whether to start from
  \eqn{E(Y|z)} (default, \code{"Eyz"}) or from \eqn{E(E(Y|z)|z)} (this can
  be overridden by providing \code{starting.values} below)
}  

\item{starting.values}{ a value indicating whether to commence
  Landweber-Fridman assuming
  \eqn{\varphi'_{-1}=starting.values}{phi'[-1]=starting.values} (proper
  Landweber-Fridman) or instead begin from \eqn{E(y|z)} (defaults to
  \code{NULL}, see details below)
}

  \item{stop.on.increase}{
    a logical value (defaults to \code{TRUE}) indicating whether to halt
    iteration if the stopping criterion (see below) increases over the
    course of one iteration (i.e. it may be above the iteration tolerance
    but increased).
  }


\subsection{Additional Arguments}{
Further arguments are passed to \code{\link{npreg}} and \code{\link{npksum}}.
}

  \item{...}{ 
    additional arguments supplied to \code{\link{npreg}} and \code{\link{npksum}}.
  }

}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}} for plotting options.


  Note that Landweber-Fridman iteration presumes that
  \eqn{\varphi_{-1}=0}{phi[-1]=0}, and so for derivative estimation we
  commence iterating from a model having derivatives all equal to
  zero. Given this starting point it may require a fairly large number
  of iterations in order to converge. Other perhaps more reasonable
  starting values might present themselves. When \code{start.phi.zero}
  is set to \code{FALSE} iteration will commence instead using
  derivatives from the conditional mean model \eqn{E(y|z)}. Should the
  default iteration terminate quickly or you are concerned about your
  results, it would be prudent to verify that this alternative starting
  value produces the same result. Also, check the norm.stop vector for
  any anomalies (such as the error criterion increasing immediately).
  
  Landweber-Fridman iteration uses an optimal stopping rule based upon
  \eqn{||E(y|w)-E(\varphi_k(z,x)|w)||^2 }{||E(y|w)-E(phi(z,x)|w)||^2
  }. However, if local rather than global optima are encountered the
  resulting estimates can be overly noisy. To best guard against this
  eventuality set \code{nmulti} to a larger number than the default
  \code{nmulti=min(2,p)} for the first iteration, where \code{p} is
  the dimension of the current smoothing problem.

  Note that for subsequent Landweber-Fridman iterations, a \dQuote{warm
  start} strategy is employed. The optimal bandwidths from the previous
  iteration are used as starting values for the current iteration. The
  user-supplied \code{nmulti} is respected for all iterations. For
  iterations after the first successful one, these optimal bandwidths
  serve as the first of the multiple initial points (a warm start),
  while any remaining restarts are cold starts. If \code{nmulti} is not
  explicitly supplied by the user, it defaults to \code{min(2,p)} for the first
  iteration and to 1 for all subsequent iterations. This strategy
  provides a balance between computational efficiency and robustness,
  allowing the numerical optimizer to refine the structural bandwidths
  as the residuals evolve incrementally while still guarding against
  local optima.

  Iteration will terminate when either the change in the value of
  \eqn{||(E(y|w)-E(\varphi_k(z,x)|w))/E(y|w)||^2
  }{||(E(y|w)-E(phi(z,x)|w))/E(y|w)||^2 } from iteration to iteration is
  less than \code{iterate.diff.tol} or we hit \code{iterate.max} or
  \eqn{||(E(y|w)-E(\varphi_k(z,x)|w))/E(y|w)||^2
  }{||(E(y|w)-E(phi(z,x)|w))/E(y|w)||^2 } stops falling in value and
  starts rising.  }

\value{

  \code{npregivderiv} returns a \code{npregivderiv} object.  The
    generic functions \code{\link{print}}, \code{\link{summary}}, and
    \code{\link{plot}} support objects of this type.

  \code{npregivderiv} returns a list with components \code{phi.prime},
  \code{phi}, \code{num.iterations}, \code{norm.stop} and
  \code{convergence}.
  
}

\references{
  
  Carrasco, M. and J.P. Florens and E. Renault (2007), \dQuote{Linear
  Inverse Problems in Structural Econometrics Estimation Based on
  Spectral Decomposition and Regularization,} In: James J. Heckman and
  Edward E. Leamer, Editor(s), Handbook of Econometrics, Elsevier, 2007,
  Volume 6, Part 2, Chapter 77, Pages 5633-5751

  Darolles, S. and Y. Fan and J.P. Florens and E. Renault (2011),
  \dQuote{Nonparametric instrumental regression,} Econometrica, 79,
  1541-1565.
  
  Feve, F. and J.P. Florens (2010), \dQuote{The practice of
  non-parametric estimation by solving inverse problems: the example of
  transformation models,} Econometrics Journal, 13, S1-S27.

  Florens, J.P. and J.S. Racine and S. Centorrino (2018),
  \dQuote{Nonparametric instrumental derivatives,} Journal of
  Nonparametric Statistics, 30 (2), 368-391.

  Fridman, V. M. (1956), \dQuote{A method of successive approximations
  for Fredholm integral equations of the first kind,} Uspeskhi,
  Math. Nauk., 11, 233-334, in Russian.

  Horowitz, J.L. (2011), \dQuote{Applied nonparametric instrumental
  variables estimation,} Econometrica, 79, 347-394.
  
  Landweber, L. (1951), \dQuote{An iterative formula for Fredholm
  integral equations of the first kind,} American Journal of
  Mathematics, 73, 615-24.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics:
  Theory and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2004), \dQuote{Cross-validated Local Linear
  Nonparametric Regression,} Statistica Sinica, 14, 485-512.

}
\author{
  Jeffrey S. Racine \email{racinej@mcmaster.ca}
}
\note{
  This function currently supports univariate \code{z} only.  This
  function should be considered to be in \sQuote{beta test} status until
  further notice.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}
\code{\link{npregiv},\link{npreg}}
}
\examples{
\dontrun{
## This illustration was made possible by Samuele Centorrino
## <samuele.centorrino@univ-tlse1.fr>

set.seed(42)
n <- 1500

## For trimming the plot (trim .5\% from each tail)

trim <- 0.005

## The DGP is as follows:

## 1) y = phi(z) + u

## 2) E(u|z) != 0 (endogeneity present)

## 3) Suppose there exists an instrument w such that z = f(w) + v and
## E(u|w) = 0

## 4) We generate v, w, and generate u such that u and z are
## correlated. To achieve this we express u as a function of v (i.e. u =
## gamma v + eps)

v <- rnorm(n,mean=0,sd=0.27)
eps <- rnorm(n,mean=0,sd=0.05)
u <- -0.5*v + eps
w <- rnorm(n,mean=0,sd=1)

## In Darolles et al (2011) there exist two DGPs. The first is
## phi(z)=z^2 and the second is phi(z)=exp(-abs(z)) (which is
## discontinuous and has a kink at zero).

fun1 <- function(z) { z^2 }
fun2 <- function(z) { exp(-abs(z)) }

z <- 0.2*w + v

## Generate two y vectors for each function.

y1 <- fun1(z) + u
y2 <- fun2(z) + u

## You set y to be either y1 or y2 (ditto for phi) depending on which
## DGP you are considering:

y <- y1
phi <- fun1

## Sort on z (for plotting)

ivdata <- data.frame(y,z,w,u,v)
ivdata <- ivdata[order(ivdata$z),]
rm(y,z,w,u,v)
with(ivdata, {

model.ivderiv <- npregivderiv(y=y,z=z,w=w)

ylim <-c(quantile(model.ivderiv$phi.prime,trim),
         quantile(model.ivderiv$phi.prime,1-trim))

plot(z,model.ivderiv$phi.prime,
     xlim=quantile(z,c(trim,1-trim)),
     main="",
     ylim=ylim,
     xlab="Z",
     ylab="Derivative",
     type="l",
     lwd=2)
rug(z)     
})
} % enddontrun
}
\keyword{ instrument }

% 
\name{npplregbw}
\alias{npplregbw}
\alias{npplregbw.default}
\alias{npplregbw.formula}
\alias{npplregbw.plbandwidth}

\title{Partially Linear Kernel Regression Bandwidth Selection with Mixed Data Types}

\description{
  \code{npplregbw} computes a bandwidth object for a partially linear
  kernel regression estimate of a one (1) dimensional dependent variable
  on \eqn{p+q}-variate explanatory data, using the model \eqn{Y = X\beta
  + \Theta (Z) + \epsilon}{Y = XB + theta(Z) + epsilon} given a set of
  estimation points, training points (consisting of explanatory data and
  dependent data), and a bandwidth specification, which can be a
  \code{rbandwidth} object, or a bandwidth vector, bandwidth type and
  kernel type.
}

\usage{
npplregbw(\dots)

\method{npplregbw}{formula}(formula, data, subset, na.action, call, \dots)


\method{npplregbw}{default}(xdat = stop("invoked without data `xdat'"),
          ydat = stop("invoked without data `ydat'"),
          zdat = stop("invoked without data `zdat'"),
          bandwidth.compute = TRUE,
          bws,
          degree = NULL,
          degree.select = c("manual", "coordinate", "exhaustive"),
          search.engine = c("nomad+powell", "cell", "nomad"),
          nomad = FALSE,
          nomad.nmulti = 0L,
            degree.min = NULL,
          degree.max = NULL,
          degree.start = NULL,
          degree.restarts = 0L,
          degree.max.cycles = 20L,
          degree.verify = FALSE,
          scale.factor.search.lower = NULL,
          ftol,
          itmax,
          nmulti,
          nomad.remin,
          powell.remin,
          small,
          tol,
          \dots)

\method{npplregbw}{plbandwidth}(xdat = stop("invoked without data `xdat'"),
          ydat = stop("invoked without data `ydat'"),
          zdat = stop("invoked without data `zdat'"),
          bws,
          nmulti,
          \dots)

}
\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the linear, nonparametric, formula, and bandwidth inputs.
}

  \item{bandwidth.compute}{
    a logical value which specifies whether to do a numerical search for
    bandwidths or not. If set to \code{FALSE}, a \code{plbandwidth} object
    will be returned with bandwidths set to those specified
    in \code{bws}. Defaults to \code{TRUE}.
  }

  \item{bws}{
    a bandwidth specification. This can be set as a \code{plbandwidth}
    object returned from an invocation of \code{npplregbw}, or as
    a matrix of bandwidths, where each row is a set of bandwidths for \eqn{Z},
    with a column for each variable \eqn{Z_i}{Z[i]}. In the first row
    are the bandwidths for the regression of \eqn{Y} on \eqn{Z}. The
    following rows contain the bandwidths for the regressions of the
    columns of \eqn{X} on \eqn{Z}.  If specified as a matrix,
    additional arguments will need to be supplied as necessary to
    specify the bandwidth type, kernel types, and so on.

    If left unspecified, \code{npplregbw} will search for optimal
    bandwidths using \code{\link{npregbw}} in the course of
    calculations. If specified, \code{npplregbw} will use the given
    bandwidths as the starting point for the numerical search for optimal
    bandwidths, unless you specify bandwidth.compute = FALSE. 
  }

  \item{call}{
    the original function call. This is passed internally by
    \code{\link{np}} when a bandwidth search has been implied by a call to
    another function. It is not recommended that the user set this.
  }

  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(formula)}, typically the environment from which the
    function is called.
  }

  \item{formula}{
    a symbolic description of variables on which bandwidth selection is
    to be performed. The details of constructing a formula are
    described below.
  }

  \item{na.action}{
    a function which indicates what should happen when the data contain
   \code{NA}s. The default is set by the \code{\link{na.action}} setting of options, and is
    \code{\link{na.fail}} if that is unset. The (recommended) default is
    \code{\link{na.omit}}.  
  }
  
  \item{subset}{
    an optional vector specifying a subset of observations to be used in
    the fitting process. 
  }

  \item{xdat}{
  a \eqn{p}-variate data frame of explanatory data (training data),
  corresponding to \eqn{X} in the model equation, whose linear
  relationship with the dependent data \eqn{Y} is posited.
  }

  \item{ydat}{
    a one (1) dimensional numeric or integer vector of dependent data, each
    element \eqn{i} corresponding to each observation (row) \eqn{i} of
  \code{xdat}. 
  }

  \item{zdat}{
    a \eqn{q}-variate data frame of explanatory data (training data),
    corresponding to \eqn{Z} in the model equation, whose relationship
    to the dependent variable is unspecified (nonparametric)
  }


\subsection{Automatic Degree Search Controls}{
These arguments control automatic local-polynomial degree search.
}

  \item{degree.max}{
    optional scalar or integer vector giving upper bounds for automatic
    degree search over continuous \code{zdat} predictors when
    \code{degree.select != "manual"}.
  }

  \item{degree.max.cycles}{
    positive integer giving the maximum number of coordinate-search
    sweeps over the degree vector. Ignored for \code{"manual"} and
    \code{"exhaustive"} degree selection.
  }

  \item{degree.min}{
    optional scalar or integer vector giving lower bounds for automatic
    degree search over continuous \code{zdat} predictors when
    \code{degree.select != "manual"}.
  }

  \item{degree.restarts}{
    non-negative integer giving the number of additional deterministic
    coordinate-search restarts. Ignored for \code{"manual"} and
    \code{"exhaustive"} degree selection.
  }

  \item{degree.select}{
    character string controlling local-polynomial degree handling for
    the nuisance regressions of \eqn{Y} on \eqn{Z} and each column of
    \eqn{X} on \eqn{Z}. \code{"manual"} (default) treats \code{degree}
    as fixed. \code{"coordinate"} performs cached coordinate-wise
    search over admissible degree vectors.
    \code{"exhaustive"} evaluates the full admissible degree grid when
    \code{search.engine="cell"}. For NOMAD-based search engines, any
    non-\code{"manual"} value requests a separate degree and bandwidth
    search for each nuisance regression.
  }

  \item{degree.start}{
    optional starting degree vector for automatic coordinate search. If
    omitted, the search starts from the degree-zero local-constant
    baseline on the continuous \code{zdat} predictors.
  }

  \item{degree.verify}{
    logical value indicating whether a coordinate-search solution should
    be exhaustively verified over the admissible degree grid after the
    heuristic phase completes. Available only for
    \code{search.engine="cell"}.
  }


\subsection{Continuous Scale-Factor Search Controls}{
These controls define lower admissibility bounds for continuous fixed-bandwidth search.
}

  \item{scale.factor.search.lower}{
    optional nonnegative scalar giving the hard lower admissibility
    bound for continuous fixed-bandwidth search candidates. Defaults to
    \code{NULL}. If \code{NULL}, an existing bandwidth object's stored
    value is inherited when available; otherwise the package default
    \code{0.1} is used. This floor applies to computed/search bandwidth
    candidates and to effective search starts only. It does not rewrite
    explicit bandwidths supplied for storage with
    \code{bandwidth.compute = FALSE}. Final fixed-bandwidth search
    candidates must also have a finite valid raw objective value.
  }

\subsection{Local-Polynomial Model Specification}{
These arguments control fixed local-polynomial specification for the nonparametric component.
}

  \item{degree}{
    for local-polynomial partially linear fits, polynomial degree
    specification for each continuous nonparametric regressor in
    \code{zdat}. When supplied with \code{degree.select="manual"},
    bandwidth optimization treats this input as fixed. A numeric vector
    is used as a common degree vector for every nuisance regression.
    To specify child-specific fixed degrees, supply either a matrix with
    \eqn{p+1} rows and one column per continuous \code{zdat} predictor,
    or a list with \eqn{p+1} entries. The child order matches the
    bandwidth matrix/object order: first the regression of \eqn{Y} on
    \eqn{Z} (row or entry \code{yzbw}), followed by the regressions of
    the columns of \eqn{X} on \eqn{Z}. Named matrix rows or list entries
    are accepted and reordered to this canonical order.
  }


\subsection{NOMAD Search Controls}{
These arguments control the optional NOMAD direct-search route for local-polynomial degree and bandwidth search.
}

  \item{nomad}{
    logical shortcut for the recommended automatic local-polynomial
    NOMAD route for the partially linear nuisance regressions. When
    \code{TRUE}, any missing values among \code{regtype},
    \code{search.engine}, \code{degree.select},
    \code{bernstein.basis}, \code{degree.min}, \code{degree.max},
    \code{degree.verify}, and \code{bwtype} are filled with
    \code{regtype="lp"}, \code{search.engine="nomad+powell"},
    \code{degree.select="coordinate"},
    \code{bernstein.basis=TRUE}, \code{degree.min=0L},
    \code{degree.max=10L}, \code{degree.verify=FALSE}, and
    \code{bwtype="fixed"}. Explicit incompatible settings error
    immediately; in particular, \code{nomad=TRUE} currently requires
    \code{regtype="lp"}, \code{bwtype="fixed"}, automatic degree
    search, \code{bernstein.basis=TRUE}, no explicit \code{degree},
    and \code{search.engine \%in\% c("nomad", "nomad+powell")}.
    This shortcut does not change the meaning of \code{nmulti} or
    \code{nomad.nmulti}: \code{nmulti} remains the outer restart count,
    while \code{nomad.nmulti} controls inner
    \code{crs::snomadr()} multistarts within each outer restart.
    Returned bandwidth objects retain this normalized preset metadata in
    \code{bw$nomad.shortcut} for a returned object \code{bw}; when
    available, \code{nomad.time} and \code{powell.time} record the
    direct-search and Powell-polish timing components. Each nuisance
    regression selects its own admissible degree and bandwidths.
  }

  \item{nomad.nmulti}{
    non-negative integer controlling the inner
    \code{crs::snomadr()} multistart count used within each outer
    NOMAD restart when \code{regtype="lp"} and automatic degree search
    uses \code{search.engine="nomad"} or \code{"nomad+powell"}.
    Defaults to \code{0L}, which preserves the current one-start-per-
    restart behavior. This does not replace \code{nmulti}:
    \code{nmulti} controls outer restarts, while \code{nomad.nmulti}
    controls inner NOMAD multistarts within each outer restart.
  }

  \item{nomad.remin}{
    logical flag controlling the optional second NOMAD hot start. When
    \code{TRUE}, NOMAD is restarted once from the best full candidate found,
    including both bandwidth and degree coordinates. Defaults to \code{FALSE};
    current simulation evidence favors the one-pass NOMAD default for routine
    use, while leaving this switch available for sensitivity checks.
  }


  \item{search.engine}{
    character string controlling the automatic local-polynomial search
    backend for each nuisance regression when
    \code{degree.select != "manual"}. \code{"nomad+powell"} (default)
    performs direct search over fixed bandwidths and the degree vector
    using \code{crs::snomadr()}, then applies one Powell hot start from
    the NOMAD solution, separately for each nuisance regression.
    \code{"nomad"} omits the Powell refinement. \code{"cell"} profiles
    the criterion over the admissible degree grid using repeated
    fixed-degree bandwidth solves.
    NOMAD-based search currently requires fixed-bandwidth child
    templates, \code{degree.verify=FALSE}, and the suggested package
    \pkg{crs} to be installed.
  }


\subsection{Numerical Search And Tolerance Controls}{
These controls set optimizer tolerances and restart behavior.
}

  \item{ftol}{
    tolerance on the value of the cross-validation function 
    evaluated at located minima. Defaults to \code{1.19e-07
      (FLT_EPSILON)}
  }

  \item{itmax}{
    integer number of iterations before failure in the numerical
    optimization routine. Defaults to \code{10000}
  }

  \item{nmulti}{
    integer number of times to restart the process of finding extrema of
    the cross-validation function from different (random) initial
    points. Defaults to \code{min(2,ncol(zdat))}.
  }

  \item{powell.remin}{
    logical flag controlling Powell restart-from-minimum behavior. For
    ordinary fixed-degree Powell-style search, \code{TRUE} restarts the
    local search from the located minimum. For \code{search.engine="nomad+powell"},
    this controls only the final Powell bandwidth-polish step. The default
    is \code{TRUE} for ordinary Powell routes and \code{FALSE} for the
    Powell polish after NOMAD unless explicitly supplied.
  }

  \item{small}{ a small number, at about the precision of the data type
    used. Defaults to \code{2.22e-16 (DBL_EPSILON)}

  }

  \item{tol}{
    tolerance on the position of located minima of the 
    cross-validation function. Defaults to \code{1.49e-08
      (sqrt(DBL_EPSILON))}
  }


\subsection{Additional Arguments}{
These arguments collect remaining controls passed through S3 methods.
}

  \item{\dots}{
    additional arguments supplied to specify the regression type,
    bandwidth type, kernel types, selection methods, and so on. To do
    this, you may specify any of \code{regtype}, 
    \code{bwmethod}, \code{bwscaling}, \code{bwtype}, \code{ckertype},
    \code{ckerorder}, \code{ukertype}, \code{okertype},
    \code{bernstein.basis}, and \code{basis}, as described in
    \code{\link{npregbw}}.
  }

}

\details{

The \code{scale.factor.*} controls are dimensionless search
controls. The package converts scale factors to bandwidths using the
estimator-specific scaling encoded in the bandwidth object, including
kernel order and the number of continuous variables relevant for the
estimator. Users should not pre-multiply these controls by sample-size
or standard-deviation factors.

  \code{scale.factor.init} controls the deterministic first search
  start when that control is exposed. \code{scale.factor.init.lower}
  and \code{scale.factor.init.upper} define the random multistart
  interval when exposed. \code{scale.factor.search.lower} is the lower
  admissibility bound for continuous fixed-bandwidth search candidates.
  The effective first start is \code{max(scale.factor.init,
  scale.factor.search.lower)} when both controls are present, and the
  effective random-start lower endpoint is
  \code{max(scale.factor.init.lower, scale.factor.search.lower)}.
  \code{scale.factor.init.upper} must be at least that effective lower
  endpoint; the package errors rather than silently expanding the user's
  interval.

  When \code{scale.factor.search.lower} is \code{NULL}, an existing
  bandwidth object's stored floor is inherited when available;
  otherwise the package default \code{0.1} is used. Explicit bandwidths
  supplied for storage with \code{bandwidth.compute = FALSE} are not
  rewritten by the search floor.

  Categorical search-start controls such as \code{dfac.init},
  \code{lbd.init}, and \code{hbd.init} have separate semantics and are
  not affected by \code{scale.factor.search.lower}.

  Documentation guide: see \code{\link{npregbw}} for component
  nonparametric regression bandwidth controls, \code{\link{np.kernels}}
  for kernels, \code{\link{np.options}} for global options, and
  \code{\link{plot}}, \code{\link{plot.np}} for plotting options.

  The partially linear bandwidth-selection argument surface is easiest
  to read by decision group: linear \code{xdat} inputs,
  nonparametric \code{zdat} inputs, and existing bandwidth inputs;
  local-polynomial/NOMAD controls for the nonparametric component;
  numerical search and feasibility controls; formula-interface
  controls; and additional bandwidth, kernel, and support controls that
  are passed to the component \code{\link{npregbw}} searches.

  For S3 plotting help, see \code{\link{plot.np}}. You can list
  available plot methods with \code{methods("plot")}.

  \code{npplregbw} implements a variety of methods for nonparametric
 regression on multivariate (\eqn{q}-variate) explanatory data defined
 over a set of possibly continuous and/or discrete (unordered, ordered)
 data. The approach is based on Li and Racine (2003), who employ
 \sQuote{generalized product kernels} that admit a mix of continuous and
 discrete data types.

 Three classes of kernel estimators for the continuous data types are
 available: fixed, adaptive nearest-neighbor, and generalized
 nearest-neighbor. Adaptive nearest-neighbor bandwidths change with
 each sample realization in the set, \eqn{x_i}{x[i]}, when estimating the
 density at the point \eqn{x}. Generalized nearest-neighbor bandwidths change
 with the point at which the density is estimated, \eqn{x}. Fixed bandwidths
 are constant over the support of \eqn{x}.

 \code{npplregbw} may be invoked \emph{either} with a formula-like
 symbolic 
 description of variables on which bandwidth selection is to be
 performed \emph{or} through a simpler interface whereby data is passed
 directly to the function via the \code{xdat}, \code{ydat}, and
 \code{zdat} 
 parameters. Use of these two interfaces is \bold{mutually exclusive}.

 Data contained in the data frame \code{zdat} may be a mix of continuous
 (default), unordered discrete (to be specified in the data frame
 \code{zdat} using \code{\link{factor}}), and ordered discrete (to be
 specified in the data frame \code{zdat} using
 \code{\link{ordered}}). Data can be entered in an arbitrary order and
 data types will be detected automatically by the routine (see
 \code{\link{np}} for details).

 Data for which bandwidths are to be estimated may be specified
 symbolically. A typical description has the form \code{dependent}
 \code{data} \code{~} \code{parametric} \code{explanatory} \code{data}
 \code{|} \code{nonparametric} \code{explanatory} \code{data},
 where \code{dependent} \code{data} is a univariate response, and
 \code{parametric} \code{explanatory} \code{data} and
 \code{nonparametric} \code{explanatory}
 \code{data} are both series of variables specified by name, separated by
 the separation character '+'. For example, \code{ y1 ~ x1 + x2 | z1 }
 specifies that the bandwidth object for the partially linear model with
 response \code{y1}, linear parametric regressors \code{x1} and
 \code{x2}, and 
 nonparametric regressor \code{z1} is to be estimated. See below for
 further examples.  

 A variety of kernels may be specified by the user. Kernels implemented
 for continuous data types include the second, fourth, sixth, and eighth
 order Gaussian and Epanechnikov kernels, and the uniform
 kernel. Unordered discrete data types use a variation on Aitchison and
 Aitken's (1976) kernel, while ordered data types use a variation of the
 Wang and van Ryzin (1981) kernel.

 When the nonparametric component is estimated with
 \code{regtype="lp"} and \code{degree.select != "manual"},
 \code{npplregbw} determines the \code{zdat}-side degree vector and
 associated bandwidth coordinates separately for the nuisance
 regressions \eqn{E[Y \mid Z]} and \eqn{E[X_j \mid Z]}. With
 \code{search.engine="cell"}, the criterion is profiled over the degree
 grid using cached coordinate-wise or exhaustive search together with
 repeated fixed-degree bandwidth solves. With
 \code{search.engine="nomad"} or \code{"nomad+powell"}, each nuisance
 criterion is optimized directly over that child's degree/bandwidth
 space using \code{crs::snomadr()}; \code{"nomad+powell"} then performs
 one Powell hot start for that child and keeps the better of the direct
 NOMAD and polished solutions. For the nonparametric regression
 components, this
 polynomial-adaptive joint-search route follows Hall and Racine (2015).

 Setting \code{nomad=TRUE} is a convenience preset for this automatic
 LP route, not a generic optimizer alias. For partially linear
 regression it expands any missing values to the equivalent long-form
 call
 \preformatted{
   npplregbw(...,
             regtype = "lp",
             search.engine = "nomad+powell",
             degree.select = "coordinate",
             bernstein.basis = TRUE,
             degree.min = 0L,
             degree.max = 10L,
             degree.verify = FALSE,
             bwtype = "fixed")
 }
 Compatible explicit tuning arguments are respected. Incompatible
 explicit settings fail fast so the shortcut never silently changes
 user-selected semantics.
}
\value{
  if \code{bwtype} is set to \code{fixed}, an object containing bandwidths
  (or scale factors if \code{bwscaling = TRUE}) is returned. If it is set to
  \code{generalized_nn} or \code{adaptive_nn}, then instead the \eqn{k}th nearest
  neighbors are returned for the continuous variables while the discrete
  kernel bandwidths are returned for the discrete variables. Bandwidths
  are stored in a list under the component name \code{bw}. Each element
  is an \code{rbandwidth} object. The first
  element of the list corresponds to the regression of \eqn{Y} on \eqn{Z}.  
  Each subsequent element is the bandwidth object corresponding to the
  regression of the \eqn{i}th column of \eqn{X} on \eqn{Z}. See examples
  for more information.
}

\section{Book And Method Pointers}{
  \code{npplregbw} selects the bandwidths and, for local-polynomial
  routes, child-specific degree/search metadata used by
  \code{\link{npplreg}}. The target model is
  \eqn{Y=X\beta+\Theta(Z)+\epsilon}; the bandwidth object controls the
  nuisance regressions \eqn{E[Y\mid Z]} and \eqn{E[X_j\mid Z]} used in
  the final partially linear solve.

  For book-length derivations, see Li and Racine (2007), Chapter 7
  \emph{Semiparametric Partially Linear Models}, especially Sections 7.1,
  7.2, and 7.4, and Racine (2019), Chapter 8 \emph{Semiparametric
  Conditional Mean Function Estimation}.
}

\references{

 Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
 discrimination by the kernel method,} Biometrika, 63, 413-420.

 Gao, Q. and L. Liu and J.S. Racine (2015), \dQuote{A partially linear
 kernel estimator for categorical data,} Econometric Reviews, 34 (6-10),
 958-977.

 Hall, P. and J.S. Racine (2015), \dQuote{Infinite Order Cross-Validated
 Local Polynomial Regression,} Journal of Econometrics, 185, 510-525.

 Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics: Theory
 and Practice,} Princeton University Press.

 Li, Q. and J.S. Racine (2004), \dQuote{Cross-validated local linear
   nonparametric regression,} Statistica Sinica, 14, 485-512.

 Pagan, A. and A. Ullah (1999), \emph{Nonparametric Econometrics,}
 Cambridge University Press.
 
 Racine, J.S. and Q. Li (2004), \dQuote{Nonparametric estimation of regression
 functions with both categorical and continuous data,} Journal of
 Econometrics, 119, 99-130.

 Robinson, P.M. (1988), \dQuote{Root-n-consistent semiparametric regression,}
 Econometrica, 56, 931-954. 
 
 Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth estimators
 for discrete distributions,}  Biometrika, 68, 301-309.

}
\author{
    Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine
    \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.

  Caution: multivariate data-driven bandwidth selection methods are, by
  their nature, \emph{computationally intensive}. Virtually all methods
  require dropping the \eqn{i}th observation from the data set, computing an
  object, repeating this for all observations in the sample, then
  averaging each of these leave-one-out estimates for a \emph{given}
  value of the bandwidth vector, and only then repeating this a large
  number of times in order to conduct multivariate numerical
  minimization/maximization. Furthermore, due to the potential for local
  minima/maxima, \emph{restarting this procedure a large number of times may
  often be necessary}. This can be frustrating for users possessing
  large datasets. For exploratory purposes, you may wish to override the
  default search tolerances, say, setting ftol=.01 and tol=.01 and
  conduct multistarting (the default is to restart min(2, ncol(zdat))
  times) as is done for a number of examples. Once the procedure
  terminates, you can restart search with default tolerances using those
  bandwidths obtained from the less rigorous search (i.e., set
  \code{bws=bw} on subsequent calls to this routine where \code{bw} is
  the initial bandwidth object).  A version of this package using the
  \code{Rmpi} wrapper is under development that allows one to deploy
  this software in a clustered computing environment to facilitate
  computation involving large datasets.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}, \code{\link{plot.np}}
  \code{\link{npregbw}}, \code{\link{npreg}} }
\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): For this example, we simulate an
# example for a partially linear model and perform bandwidth selection

set.seed(42)

n <- 250
x1 <- rnorm(n)
x2 <- rbinom(n, 1, .5)

z1 <- rbinom(n, 1, .5)
z2 <- rnorm(n)

y <- 1 + x1 + x2 + z1 + sin(z2) + rnorm(n)

X <- data.frame(x1, factor(x2))
Z <- data.frame(factor(z1), z2)

# Compute data-driven bandwidths... this may take a minute or two
# depending on the speed of your computer...

bw <- npplregbw(formula=y~x1+factor(x2)|factor(z1)+z2)

summary(bw)

# Note - the default is to use the local constant estimator. If you wish
# to use instead a local linear estimator, this is accomplished via
# npplregbw(xdat=X, zdat=Z, ydat=y, regtype="ll")

# Note - see the example for npudensbw() for multiple illustrations
# of how to change the kernel function, kernel order, and so forth.

# You may want to manually specify your bandwidths
bw.mat <- matrix(data =  c(0.19, 0.34,  # y on Z
                           0.00, 0.74,  # X[,1] on Z
                           0.29, 0.23), # X[,2] on Z
                ncol = ncol(Z), byrow=TRUE)

bw <- npplregbw(formula=y~x1+factor(x2)|factor(z1)+z2, 
                          bws=bw.mat, bandwidth.compute=FALSE)
summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# You may want to tweak some of the bandwidths
bw$bw[[1]] # y on Z, alternatively bw$bw$yzbw
bw$bw[[1]]$bw <- c(0.17, 0.30)

bw$bw[[2]] # X[,1] on Z
bw$bw[[2]]$bw[1] <- 0.00054

summary(bw)

# EXAMPLE 1 (INTERFACE=DATA FRAME): For this example, we simulate an
# example for a partially linear model and perform bandwidth selection

set.seed(42)

n <- 250
x1 <- rnorm(n)
x2 <- rbinom(n, 1, .5)

z1 <- rbinom(n, 1, .5)
z2 <- rnorm(n)

y <- 1 + x1 + x2 + z1 + sin(z2) + rnorm(n)

X <- data.frame(x1, factor(x2))
Z <- data.frame(factor(z1), z2)

# Compute data-driven bandwidths... this may take a minute or two
# depending on the speed of your computer...

bw <- npplregbw(xdat=X, zdat=Z, ydat=y)

summary(bw)

# Note - the default is to use the local constant estimator. If you wish
# to use instead a local linear estimator, this is accomplished via
# npplregbw(xdat=X, zdat=Z, ydat=y, regtype="ll")

# Note - see the example for npudensbw() for multiple illustrations
# of how to change the kernel function, kernel order, and so forth.

# You may want to manually specify your bandwidths
bw.mat <- matrix(data =  c(0.19, 0.34,  # y on Z
                           0.00, 0.74,  # X[,1] on Z
                           0.29, 0.23), # X[,2] on Z
                ncol = ncol(Z), byrow=TRUE)

bw <- npplregbw(xdat=X, zdat=Z, ydat=y, 
                          bws=bw.mat, bandwidth.compute=FALSE)
summary(bw)

# Sleep for 5 seconds so that we can examine the output...

if (interactive()) Sys.sleep(5)

# You may want to tweak some of the bandwidths
bw$bw[[1]] # y on Z, alternatively bw$bw$yzbw
bw$bw[[1]]$bw <- c(0.17, 0.30)

bw$bw[[2]] # X[,1] on Z
bw$bw[[2]]$bw[1] <- 0.00054

summary(bw)
} % enddontrun
}
\keyword{nonparametric}

%
\name{npqreg}
\alias{npqreg}
\alias{npqreg.condbandwidth}
\alias{npqreg.default}
\alias{npqreg.formula}
\alias{predict.qregression}
\alias{plot.qregression}
\title{Kernel Quantile Regression with Mixed Data Types}

\description{
  \code{npqreg} computes a kernel quantile regression estimate of a one
  (1) dimensional dependent variable on \eqn{p}-variate explanatory
  data, given a set of evaluation points, training points (consisting of
  explanatory data and dependent data), and a bandwidth specification
  using the methods of Li and Racine (2008) and Li, Lin and Racine
  (2013). A bandwidth specification can be a \code{condbandwidth} object,
  or a bandwidth vector, bandwidth type and kernel type.
}

\usage{
npqreg(bws, \dots)

\method{npqreg}{formula}(bws, data = NULL, newdata = NULL, \dots)


\method{npqreg}{condbandwidth}(bws,
       txdat = stop("training data 'txdat' missing"),
       tydat = stop("training data 'tydat' missing"),
       exdat,
       tau = 0.5,
       gradients = FALSE,
       tol = 1.490116e-04,
       small = 1.490116e-05,
       itmax = 10000,
       \dots)

\method{npqreg}{default}(bws, txdat, tydat, nomad = FALSE, \dots)

\method{predict}{qregression}(object, se.fit = FALSE, \dots)

\method{plot}{qregression}(x, \dots)

}

\arguments{
\subsection{Data, Bandwidth Inputs And Formula Interface}{
These arguments identify the bandwidth specification, formula/data interface, and training data.
}

  \item{bws}{
    a bandwidth specification. This can be set as a \code{condbandwidth}
    object returned from an invocation of \code{\link{npcdistbw}}, or
    as a vector of bandwidths, with each element \eqn{i} corresponding
    to the bandwidth for column \eqn{i} in \code{txdat}. If specified as
    a vector, then additional arguments will need to be supplied as
    necessary to specify the bandwidth type, kernel types, and so on.
  }


  \item{data}{
    an optional data frame, list or environment (or object
    coercible to a data frame by \code{\link{as.data.frame}}) containing the variables
    in the model. If not found in data, the variables are taken from
    \code{environment(bws)}, typically the environment from which
    \code{\link{npcdistbw}} was called.
  }


  \item{txdat}{
    a \eqn{p}-variate data frame of explanatory data (training data) used to
    calculate the regression estimators. Defaults to the training data used to
    compute the bandwidth object.
  }


  \item{tydat}{
    a one (1) dimensional numeric or integer vector of dependent data, each
    element \eqn{i} corresponding to each observation (row) \eqn{i} of
    \code{txdat}. Defaults to the training data used to
    compute the bandwidth object.
  }

  \item{object}{
    an object of class \code{"qregression"} returned by \code{npqreg}.
  }

  \item{x}{
    an object of class \code{"qregression"} returned by \code{npqreg}.
  }


\subsection{Local-Polynomial Degree And Bandwidth Search}{
This argument controls the recommended automatic local-polynomial NOMAD route, which jointly selects continuous polynomial degree and bandwidths when conditional-distribution bandwidths are computed inside \code{npqreg}.
}

  \item{nomad}{
    logical shortcut passed through to \code{\link{npcdistbw}} when
    bandwidths are computed inside \code{npqreg}. When \code{TRUE}, the
    conditional-distribution bandwidth route fills any missing values
    among \code{regtype}, \code{search.engine}, \code{degree.select},
    \code{bernstein.basis}, \code{degree.min}, \code{degree.max},
    \code{degree.verify}, and \code{bwtype} with the recommended
    automatic local-polynomial degree-and-bandwidth NOMAD preset documented in \code{\link{npcdistbw}}.
    Additional NOMAD tuning arguments such as \code{nomad.nmulti} may
    also be supplied through \code{\dots}; \code{nmulti} remains the
    outer restart count while \code{nomad.nmulti} controls inner
    \code{crs::snomadr()} multistarts within each outer restart.
    After fitting, inspect \code{fit$bws$nomad.shortcut} on the
    returned object \code{fit} to see the normalized shortcut metadata.
  }


\subsection{Evaluation Data And Returned Quantities}{
These arguments control where the quantile regression is evaluated and which fitted quantities are returned.
}

  \item{exdat}{
    a \eqn{p}-variate data frame of points on which the regression will be
    estimated (evaluation data). By default,
    evaluation takes place on the data provided by \code{txdat}.
  }


  \item{gradients}{
    a logical value indicating that you want gradients of the conditional
    quantile with respect to the conditioning variables computed and returned
    in the resulting \code{qregression} object. Defaults to \code{FALSE}.
  }


  \item{newdata}{
    An optional data frame in which to look for evaluation data. If
    omitted, the training data are used.
  }

  \item{se.fit}{
    logical value. If \code{TRUE}, \code{predict.qregression} returns a
    list with components \code{fit} and \code{se.fit}; otherwise it
    returns fitted conditional quantiles.
  }


  \item{tau}{
    a numeric scalar or vector specifying the quantile probability or
    probabilities \eqn{\tau}{tau}. Defaults to \code{0.5}.
  }



\subsection{Quantile Solver Controls}{
These arguments control the one-dimensional numerical quantile extraction step.
}

  \item{itmax}{
    integer maximum number of iterations allowed in the one-dimensional
    quantile refinement. Defaults to \code{10000}.
  }


  \item{small}{
    minimum interval width used by the one-dimensional quantile
    refinement. Defaults to \code{1.490116e-05}
    (approximately \code{1000*sqrt(.Machine$double.eps)}).
  }


  \item{tol}{
    tolerance on the one-dimensional quantile location refinement.
    Defaults to \code{1.490116e-04}
    (approximately \code{10000*sqrt(.Machine$double.eps)}).
  }


\subsection{Additional Arguments}{
Further arguments are passed to the bandwidth-selection counterpart,
prediction/evaluation route, or plot route as appropriate.
}

  \item{\dots}{
    additional arguments supplied to \code{\link{npcdistbw}} when
    \code{npqreg} computes bandwidths internally, or arguments needed
    to interpret a numeric \code{bws} vector. This is where bandwidth
    selection controls such as \code{bwmethod}, \code{bwtype}, and
    \code{bwscaling}, kernel/support controls such as
    \code{cxkertype}, \code{cykertype}, \code{cxkerorder},
    \code{cykerorder}, \code{cxkerbound}, and \code{cykerbound},
    categorical kernel controls such as \code{uxkertype},
    \code{uykertype}, \code{oxkertype}, and \code{oykertype}, search
    controls such as \code{nmulti} and
    \code{scale.factor.search.lower}, and local-polynomial/NOMAD
    controls such as \code{regtype}, \code{degree},
    \code{bernstein.basis}, \code{degree.select}, and
    \code{nomad.nmulti} are supplied. In \code{predict.qregression},
    additional arguments are passed to \code{npqreg} for evaluation
    with the stored bandwidth object; common examples are
    \code{newdata}, native \code{exdat}, and \code{tau}. In
    \code{plot.qregression}, additional arguments are passed through
    the package plot route; common controls include \code{tau},
    \code{gradients}, \code{output}, \code{legend}, and graphics
    arguments. See \code{\link{npcdistbw}} and \code{\link{plot.np}}
    for the complete bandwidth-selection and plotting argument
    surfaces.
  }
}
\details{
  Documentation guide: see \code{\link{np.kernels}} for kernels, \code{\link{np.options}} for global options, and \code{\link{plot}}, \code{\link{plot.np}} for plotting options.


  Given a conditional distribution bandwidth object, \code{npqreg}
  estimates the conditional distribution function
  \eqn{F(y|x)} and extracts the requested conditional quantile. For
  \eqn{0 < \tau < 1}, the conditional quantile at probability
  \eqn{\tau}{tau} is
  \deqn{
    q_\tau(x) = \inf\{y : F(y|x) \ge \tau\}.
  }
  Equivalently, \eqn{q_\tau(x)} is a quasi-inverse of the conditional
  distribution in the sense of Nelsen (2006): an inverse agrees with
  \eqn{F} on the range of \eqn{F}, while outside that range the
  generalized inverse is defined by the lower endpoint at which
  \eqn{F} reaches or exceeds the requested probability. Numerically,
  \code{npqreg} inverts the selected conditional distribution estimator
  represented by \code{bws}. This includes the selected bandwidth type,
  kernels, local-polynomial regression type, selected polynomial degree,
  basis, and Bernstein-basis setting inherited from \code{\link{npcdistbw}}.
  If the bandwidth object was selected with \code{nomad=TRUE}, the returned
  conditional-distribution bandwidth object is an LP object: its
  \code{regtype}/\code{regtype.engine} metadata identify the selected
  local-polynomial route and its \code{degree}/\code{degree.engine} metadata
  record the selected continuous-coordinate polynomial degree. \code{npqreg},
  \code{\link{predict}}, and \code{\link{plot}} reuse this stored LP metadata;
  plotting additional \code{tau} values does not recompute or downgrade the
  selected degree.
  The one-dimensional inversion is carried out over the observed support
  of the dependent variable using the same selected conditional CDF
  estimator that is later used for quantile standard errors and gradients.
  The arguments \code{tol}, \code{small}, and \code{itmax} control this
  one-dimensional refinement.

  Let \eqn{f(y|x) = \partial F(y|x)/\partial y} denote the conditional
  density. The asymptotic standard error of the conditional quantile is
  computed by the first-order delta method,
  \deqn{
    se\{\hat q_\tau(x)\}
      =
    \frac{se\{\hat F(\hat q_\tau(x)|x)\}}
         {\hat f(\hat q_\tau(x)|x)} ,
  }
  using the selected conditional distribution standard-error machinery
  and the selected conditional density evaluated at the fitted quantile.
  This corresponds to the quantile variance expression in Li, Lin and
  Racine (2013).

  If \code{gradients=TRUE}, \code{npqreg} also computes gradients of
  the conditional quantile with respect to the conditioning variables
  for which gradients are defined. Differentiating
  \eqn{F(q_\tau(x)|x) = \tau} gives
  \deqn{
    \nabla_x q_\tau(x)
      =
    -\frac{F_x(q_\tau(x)|x)}{f(q_\tau(x)|x)},
  }
  where \eqn{F_x(y|x)} is the derivative of the same selected conditional
  distribution estimator with respect to \eqn{x}. For \code{regtype="lc"},
  this uses the local-constant conditional-gradient machinery; for
  \code{regtype="ll"} it uses the canonical local-polynomial degree-one
  route; and for \code{regtype="lp"} it uses the selected or supplied
  degree vector. The corresponding first-order gradient standard errors
  are computed componentwise as
  \deqn{
    se\{\nabla_x \hat q_\tau(x)\}
      =
    \frac{se\{\hat F_x(\hat q_\tau(x)|x)\}}
         {\hat f(\hat q_\tau(x)|x)} .
  }

  When \code{npqreg} is called without an explicit \code{bws} object, it
  first computes conditional distribution bandwidths using
  \code{\link{npcdistbw}} and stores them in the returned object's
  \code{bws} component. If a scalar \code{tau} was used initially and
  additional quantiles are later desired as fitted objects, reuse those
  selected bandwidths directly, for example
  \code{npqreg(bws = fit$bws, tau = c(0.25, 0.5, 0.75))}. If the goal
  is only to inspect additional quantiles graphically, use
  \code{plot(fit, tau = c(0.25, 0.5, 0.75))}; this reuses the stored
  bandwidths and recomputes only the one-dimensional quantile extraction
  step for the requested \code{tau} values. Vector-\code{tau} plots are
  overlaid and include a legend; use \code{legend=FALSE},
  \code{legend=NULL}, or a \code{legend=list(...)} control to suppress
  or customize it.

  The \code{\link{predict}} method follows the usual S3
  \code{newdata} convention. For formula fits, supply a data frame of
  evaluation covariates via \code{predict(fit, newdata=...)}. For
  non-formula fits, \code{newdata} is translated to the native
  evaluator argument \code{exdat} when \code{exdat} is not supplied.
  The native \code{exdat} argument remains available for advanced
  workflows and takes precedence if both \code{newdata} and
  \code{exdat} are supplied. If \code{tau} is omitted in
  \code{predict}, the fitted object's stored \code{tau} value is used.

}

\value{

  \code{npqreg} returns a \code{qregression} object.  The generic
  functions \code{\link{fitted}} (or \code{\link{quantile}}),
  \code{\link{se}}, \code{\link{predict}}, and \code{\link{gradients}}
  extract (or generate) estimated values, asymptotic standard errors on
  estimates, predictions, and gradients, respectively, from the returned
  object. \code{\link{predict}} uses the object's stored \code{tau}
  value by default; supply \code{tau=} to override it. Furthermore, the functions
  \code{\link{summary}} and \code{\link{plot}} support objects of this
  type. The returned object has the following components:

  \item{eval}{ evaluation points }
  \item{quantile}{ estimation of the quantile regression function
    (conditional quantile) at the evaluation points. If \code{tau} has
    length greater than one this is an evaluation-by-\code{tau} matrix. }
  \item{quanterr}{ asymptotic standard errors of the quantile
    regression estimates, obtained from the conditional distribution
    standard error and the estimated conditional density at the fitted
    quantile. If \code{tau} has length greater than one this is an
    evaluation-by-\code{tau} matrix. }
  \item{quantgrad}{ gradients of the conditional quantile with respect
    to the conditioning variables at each evaluation point, when
    \code{gradients=TRUE}. If \code{tau} has length greater than one
    this is an evaluation-by-gradient-by-\code{tau} array. }
  \item{quantgerr}{ asymptotic standard errors for gradients, when
    \code{gradients=TRUE}. If \code{tau} has length greater than one
    this is an evaluation-by-gradient-by-\code{tau} array. }
  \item{tau}{ the quantile probability or probabilities computed }


}

\section{Book And Method Pointers}{
  The conditional quantile target is the generalized inverse
  \eqn{q_\tau(x)=\inf\{y:F(y\mid x)\ge \tau\}} of the conditional
  distribution. The standard errors and gradients described above are
  first-order delta-method quantities evaluated using the same selected
  conditional CDF, conditional density, bandwidths, kernels, and
  local-polynomial degree inherited from the supplied
  \code{\link{npcdistbw}} object.

  For book-length derivations, see Li and Racine (2007), Chapter 6
  \emph{Conditional CDF and Quantile Estimation}, especially Sections
  6.3-6.5, and Racine (2019), Chapter 4 \emph{Conditional Probability
  Density and Cumulative Distribution Functions}. The quasi-inverse
  terminology follows Nelsen (2006).
}

\references{

  Aitchison, J. and C.G.G. Aitken (1976), \dQuote{Multivariate binary
    discrimination by the kernel method,} Biometrika, 63, 413-420.

  Hall, P. and J.S. Racine and Q. Li (2004), \dQuote{Cross-validation
    and the estimation of conditional probability densities,} Journal of
    the American Statistical Association, 99, 1015-1026.

  Koenker, R. W. and G.W. Bassett (1978), \dQuote{Regression
  quantiles,} Econometrica, 46, 33-50.

  Koenker, R. (2005), \emph{Quantile Regression,} Econometric Society
  Monograph Series, Cambridge University Press.

  Li, Q. and J.S. Racine (2007), \emph{Nonparametric Econometrics:
  Theory and Practice,} Princeton University Press.

  Li, Q. and J.S. Racine (2008), \dQuote{Nonparametric estimation of
  conditional CDF and quantile functions with mixed categorical and
  continuous data,} Journal of Business and Economic Statistics, 26,
  423-434.

  Li, Q. and J. Lin and J.S. Racine (2013), \dQuote{Optimal Bandwidth
  Selection for Nonparametric Conditional Distribution and Quantile
  Functions}, Journal of Business and Economic Statistics, 31, 57-65.

  Nelsen, R.B. (2006), \emph{An Introduction to Copulas}, Second
  Edition, Springer.

  Wang, M.C. and J. van Ryzin (1981), \dQuote{A class of smooth
    estimators for discrete distributions,} Biometrika, 68, 301-309.
}

\author{
  Tristen Hayfield \email{tristen.hayfield@gmail.com}, Jeffrey S. Racine \email{racinej@mcmaster.ca}
}

\section{Usage Issues}{
  If you are using data of mixed types, then it is advisable to use the
  \code{\link{data.frame}} function to construct your input data and not
  \code{\link{cbind}}, since \code{\link{cbind}} will typically not work as
  intended on mixed data types and will coerce the data to the same
  type.
}

\seealso{
  \code{\link{np.kernels}}, \code{\link{np.options}}, \code{\link{plot}}, \code{\link{plot.np}}, \pkg{quantreg} }

\examples{
\dontrun{
# EXAMPLE 1 (INTERFACE=FORMULA): For this example, we compute a
# bivariate nonparametric quantile regression estimate for Giovanni
# Baiocchi's Italian income panel (see Italy for details)

data("Italy")
with(Italy, {

# Compute conditional distribution bandwidths and extract three
# conditional quantiles using the same selected bandwidths.

model.q <- npqreg(gdp~ordered(year), tau=c(0.25, 0.50, 0.75))

# Plot the overlaid quantiles.

plot(model.q)

# If a scalar tau was used first, additional quantiles can reuse the
# selected bandwidths without recomputing cross-validation. Use npqreg()
# when the additional fitted values are needed as an object, or plot()
# when graphical inspection is all that is desired.

model.med <- npqreg(gdp~ordered(year), tau=0.50)
model.q <- npqreg(bws=model.med$bws, tau=c(0.25, 0.50, 0.75))
plot(model.med, tau=c(0.25, 0.50, 0.75))

})

# EXAMPLE 1 (INTERFACE=DATA FRAME): For this example, we compute a
# bivariate nonparametric quantile regression estimate for Giovanni
# Baiocchi's Italian income panel (see Italy for details)

data("Italy")
with(Italy, {
data <- data.frame(ordered(year), gdp)

# First, compute the likelihood cross-validation bandwidths (default).
# Note - this may take a few minutes depending on the speed of your
# computer...

bw <- npcdistbw(xdat=ordered(year), ydat=gdp)

# Note - numerical search for computing the quantiles will take a
# minute or so...

model.q <- npqreg(bws=bw, tau=c(0.25, 0.50, 0.75))

plot(model.q)

})
} % enddontrun
}
\keyword{nonparametric}
